#
# @(#)jreimage.sh	1.22 98/10/14
#
# Make a jre image. Called from TOPDIR/build/platform. It will create
# a directory named "jre-image-${platform}/". The contents of this
# directory is what the end-user install will look like.
#

set -e 
set -x

if [ $# != 1 ]; then
    echo "`basename $0`: must provide a destination directory argument."
    exit 1
fi

if [ ! -d classes ]; then
    echo "`basename $0`: not at BUILDDIR or JDK not built."
    exit 2
fi

DESTDIR="$1"
rm -fr $DESTDIR
mkdir -p $DESTDIR
BUILDDIR=`pwd`
SHARE_SRC=$BUILDDIR/../../src/share

#
# BINDIR
#
(cd $BUILDDIR; tar cf - bin) | (cd $DESTDIR; tar xf -)
# Remove _g files
rm -f `find $DESTDIR/bin -name '*_g*' -print`
# Remove tools that should not be part of JRE.
JRETOOLS=".java_wrapper java javaw rmiregistry rmid tnameserv keytool policytool"
k=""
for t in $JRETOOLS; do
    p="/$t\$|/$t.exe\$"
    if [ -z "$k" ]; then
	k="$p"
    else
	k="$k|$p"
    fi
done
k="$k|\.dll|\.ocx|\.txt"
find $DESTDIR/bin -type d -o -print | egrep -v "$k" | xargs rm -f

#
# LIBDIR
#
cp -r $BUILDDIR/lib $DESTDIR
# Remove _g files.
rm -f `find $DESTDIR/lib -name '*_g*' -print`
# Remove .lib files on windows
(cd $DESTDIR/lib; rm -f java.lib jvm.lib hpi.lib)

#
# Remove agent.
#
find $DESTDIR/lib -name 'libagent.so' -exec rm -f \{} \;
find $DESTDIR/bin -name 'agent.dll' -exec rm -f \{} \;

#
# rt.jar.
#     - sun.applet is included, till hotjava stops relying on it.
#     - sun.tools.jar is included, needs cleanup.
#     - CommandLine.class is include for jar as well.
#
NOTJRE='
com/sun/javadoc
com/sun/tools
sun/net/www/httpd
sun/rmi/rmic
sun/tools/agent
sun/tools/asm
sun/tools/debug
sun/tools/java
sun/tools/javac
sun/tools/javap
sun/tools/native2ascii
sun/tools/serialver
sun/tools/tree
sun/tools/ttydebug
sun/tools/util
'

UTIL="sun/tools/util"
MAINMANIFEST="$BUILDDIR/../share/tools/manifest.mf"
BEANMANIFEST="$BUILDDIR/javax/swing/beaninfo/manifest"
VERSION="1.2.0"

# Move what we don't want to "notjre/"
rm -fr $BUILDDIR/notjre; mkdir -p $BUILDDIR/notjre
(cd $BUILDDIR/classes; tar cf - $NOTJRE) | (cd $BUILDDIR/notjre; tar xf -)
(cd $BUILDDIR/classes; rm -fr $NOTJRE)

  # Temporary hack.
  mkdir -p $BUILDDIR/classes/$UTIL
  cp $BUILDDIR/notjre/$UTIL/CommandLine.class $BUILDDIR/classes/$UTIL
  
  # Create the jar file.
  ( sed "s/@@RELEASE@@/$VERSION/" $MAINMANIFEST; echo; cat $BEANMANIFEST ) >manifest.tmp
  cd $BUILDDIR/classes; $BUILDDIR/bin/jar cv0mf ../manifest.tmp ../rt.jar *; cd ..
  rm -f manifest.tmp
  mv rt.jar $DESTDIR/lib

# Move stuff back.
(cd $BUILDDIR/notjre; tar cf - .) | (cd $BUILDDIR/classes; tar xf -)
rm -fr $BUILDDIR/notjre

#
# Doc files.
#
for i in COPYRIGHT LICENSE README CHANGES; do
    rm -f $DESTDIR/$i
    sed 's/$//g' $SHARE_SRC/doc/jre/$i > $DESTDIR/$i
done
