#!/bin/sh -f

#  @(#)allsrcprods.sh	1.1 98/10/06
# 
#  Copyright 1998 by Sun Microsystems, Inc.,
#  901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
#  All rights reserved.
# 
#  This software is the confidential and proprietary information
#  of Sun Microsystems, Inc. ("Confidential Information").  You
#  shall not disclose such Confidential Information and shall use
#  it only in accordance with the terms of the license agreement
#  you entered into with Sun.

# Script to make the 6 licensable source products from the master workspace
# Products are JDK and JAE for Solaris, win32, or both platforms
# JAE source product is loosely defined as the JDK source product,
# less compiler, debugger and elements dependent on compiler or debugger
# This script is intended only for internal use by Sun, and expects to
# be executed on Solaris
# 
# Sample usage:
# allsrcprods.sh -build FCS_M  -destdir /srcprods -srcdir /java/ws/jdk12

# exit immediately on error conditions
set -e 

###############

usage() {

  echo "usage: allsrcprods.sh [-build \"version\"] [-srcdir \"dir\"] [-destdir \"dir\"]"
  exit 1;
}

###############

# Note the top level directory of the workspace we are building
SRCDIR=`pwd`
DESTDIR=$SRCDIR/../srcprods
BUILD=1_2

# crudely parse args, this could be better
while [ $# -gt 0 ]; do
  case "$1" in
    -build)    shift; BUILD="$1" ;;
    -srcdir)   shift; SRCDIR="$1" ;;
    -destdir)  shift; DESTDIR="$1" ;;
    *)  usage  ;;
  esac
  shift
done

TOOLDIR=$SRCDIR/build/share/tools/srcproduct

#Check the SRCDIR is valid with a fairly arbitrary heuristic
if [ ! -d $TOOLDIR ]; then
  echo "srcdir "$SRCDIR" is not a valid workspace"
  exit 1
fi

# Check we can create the src product in DESTDIR, and make sure its empty
if [ ! -f $DESTDIR ]; then
  mkdir -p $DESTDIR
elif [ ! -d $DESTIR -o ! -w $DESTDIR ]; then
  echo "can't create output in $DESTDIR"
  exit 1
fi
/usr/bin/rm -f $DESTDIR/\*

SRCPRODSH=${TOOLDIR}/srcproduct.sh
# we exit immediatelu if any of these return a failure code
sh $SRCPRODSH -type jae -os win32 -build $BUILD -srcdir $SRCDIR -destdir $DESTDIR
sh $SRCPRODSH -type jae -os solaris -build $BUILD -srcdir $SRCDIR -destdir $DESTDIR
sh $SRCPRODSH -type jae -os solwin32 -build $BUILD -srcdir $SRCDIR -destdir $DESTDIR
sh $SRCPRODSH -type jdk -os win32 -build $BUILD -srcdir $SRCDIR -destdir $DESTDIR
sh $SRCPRODSH -type jdk -os solaris -build $BUILD -srcdir $SRCDIR -destdir $DESTDIR
sh $SRCPRODSH -type jdk -os solwin32 -build $BUILD -srcdir $SRCDIR -destdir $DESTDIR

exit 0
