#!/bin/sh -f

#  @(#)srcproduct.sh	1.2 98/10/21
# 
#  Copyright 1998 by Sun Microsystems, Inc.,
#  901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
#  All rights reserved.
# 
#  This software is the confidential and proprietary information
#  of Sun Microsystems, Inc. ("Confidential Information").  You
#  shall not disclose such Confidential Information and shall use
#  it only in accordance with the terms of the license agreement
#  you entered into with Sun.

# Script to make the 6 licensable source products from the master workspace
# Products are JDK and JAE for Solaris, win32, or both platforms
# JAE source product is loosely defined as the JDK source product,
# less compiler, debugger and elements dependent on compiler or debugger
# This script is intended only for internal use by Sun, and expects to
# be executed on Solaris
# 
# Sample usage:
# srcproduct.sh -type jae -os solwin32 -build FCS_M  -destdir /srcprods -srcdir /java/ws/jdk12

# exit immediately on error conditions
set -e 

###############

cleanup() {

$RM -f $ALLSRCFILES
$RM -f $EXCLUDEDLIST
$RM -f $EXCLUDEDFILES

}

usage() {

  echo "usage: srcproduct.sh -type jdk|jae -os solaris|win32|solwin32"
  echo "         [-build \"version\"] [-srcdir \"dir\"] [-destdir \"dir\"]"
  cleanup
  exit 1;
}

###############

# Lets make sure we get the right versions of these utilities
ZIP=/usr/dist/local/exe/zip
TAR=/usr/bin/tar
FIND=/usr/bin/find
RM=/usr/bin/rm
CAT=/usr/bin/cat

# Note the top level directory of the workspace we are building
# OS to build is either solaris, win32, or solwin32 (ie both)
SRCDIR=`pwd`
DESTDIR=$SRDIR/../srcprods
SRC_PROD_OS=solwin32
SRC_PROD_TYPE=jdk
PROD_VERSION=1_2
BUILD=${PROD_VERSION}
DATE=`/usr/bin/date +%d_%b` 

# crudely parse args, this could be better
while [ $# -gt 0 ]; do
  case "$1" in
    -type)     shift; SRC_PROD_TYPE="$1" ;;
    -os)       shift; SRC_PROD_OS="$1" ;;
    -build)    shift; BUILD="$1" ;;
    -srcdir)   shift; SRCDIR="$1" ;;
    -destdir)  shift; DESTDIR="$1" ;;
    *)  usage  ;;
  esac
  shift
done

case $SRC_PROD_OS in
  'solaris')  BUNDLE_OS="solaris" ;;
  'win32')    BUNDLE_OS="win32" ;;
  'solwin32') BUNDLE_OS="win32solaris" ;;
  *) usage    ;;
esac

case $SRC_PROD_TYPE in
  'jae')   NAME_STUB="JAE" ;;
  'jdk')   NAME_STUB="JDK" ;;
  *) usage ;;
esac

TOOLDIR=$SRCDIR/build/share/tools/srcproduct

#Check the SRCDIR is valid with a fairly arbitrary heuristic
if [ ! -d $TOOLDIR ]; then
  echo "srcdir "$SRCDIR" is not a valid workspace"
  exit 1
fi

# Check we can create the src product in DESTDIR
if [ ! -f $DESTDIR ]; then
  mkdir -p $DESTDIR
elif [ ! -d $DESTIR -o ! -w $DESTDIR ]; then
  echo "can't create output in $DESTDIR"
  exit 1
fi

# There's a slightly new naming convention coming soon, so this may change
SRC_PROD_NAME=${SRC_PROD_TYPE}${PROD_VERSION}${BUILD}-src-${SRC_PROD_OS}-${DATE}.zip
$RM -f $SRC_PROD_NAME
# Need the old style name until other scripts updated.
SRC_BUNDLE_NAME=bundle.${BUNDLE_OS}.${SRC_PROD_TYPE}.zip
$RM -f $SRC_BUNDLE_NAME

# The files to include are a subset of those under codemgr control. The
# nametable is first stripped down so each line contains just the filename
# we also exclude any line which begins "test" (ie all the tests)
# This step could be done solely by the "exclude" mechanism, but there are
# a lot of files under "test" and we presently don't ship them with any
# source release, so its more efficient to hardwire their elimination

ALLSRCFILES=/tmp/allfiles.$$
$RM -f $ALLSRCFILES
/usr/bin/sed -e '1,1d' -e '/^test/d' -e'/^deleted_files/d' -e 's/ .*$//' $SRCDIR/Codemgr_wsdata/nametable > $ALLSRCFILES

# We refine a source product by excluding those elements not part of
# that specific source product, as defined by "exclude" files.
#
# exclude-all            : Not part of any source product
# exclude-jdk-all        : Not part of any JDK source product
# exclude-jdk-solaris    : Not part of the Solaris JDK source product
# exclude-jdk-win32      : Not part of the win32 JDK source product
# exclude-jae-all        : Not part of any JAE source product
# exclude-jae-solaris    : Not part of the Solaris JAE source product
# exclude-jae-win32      : Not part of the win32 JAE source product
#
# These files contain files and directories, one per line
# If just some files in a directory need to be included, then you need to
# remove the directory from the exclude list, and list all the files
# you don't want to included in the product
# Some of the exclude files may be non-existent or empty if nothing needs
# to be excluded.

EXCLUDEDLIST=/tmp/excluded-list.$$
$RM -f $EXCLUDEDLIST
touch $EXCLUDEDLIST
if [ -r $TOOLDIR/exclude-all ]; then
  $CAT $TOOLDIR/exclude-all >> $EXCLUDEDLIST
fi

# define JDK source product
if [ $SRC_PROD_TYPE = "jdk" ]; then

  if [ -r $TOOLDIR/exclude-jdk-all ]; then
    $CAT $TOOLDIR/exclude-jdk-all >> $EXCLUDEDLIST
  fi

  if [ $SRC_PROD_OS = "solaris" -a -r $TOOLDIR/exclude-jdk-solaris ]; then
     $CAT $TOOLDIR/exclude-jdk-solaris >> $EXCLUDEDLIST
  elif [ $SRC_PROD_OS = "win32" ]; then
     if [  -r $TOOLDIR/exclude-jdk-win32 ]; then
       $CAT $TOOLDIR/exclude-jdk-win32 >> $EXCLUDEDLIST
     fi
  fi

fi

# define JAE source product
if [ $SRC_PROD_TYPE = "jae" ]; then

  if [ -r $TOOLDIR/exclude-jae-all ]; then
    $CAT $TOOLDIR/exclude-jae-all >> $EXCLUDEDLIST
  fi

  if [ $SRC_PROD_OS = "solaris" -a -r $TOOLDIR/exclude-jae-solaris ]; then
     $CAT $TOOLDIR/exclude-jae-solaris >> $EXCLUDEDLIST
  elif [ $SRC_PROD_OS = "win32" ]; then
     if [  -r $TOOLDIR/exclude-jae-win32 ]; then
       $CAT $TOOLDIR/exclude-jae-win32 >> $EXCLUDEDLIST
     fi
  fi

fi


# We then exclude all files listed in the exclude files
# Because the "zip" we have doesn't provide a way to exclude a directory
# we need to use "find" to enumerate all the files in the excluded list of
# directories. This only takes a few seconds and is much better than
# placing the burden on a human to maintain such a list, which can be
# very long indeed.
#
# Enumerate every file which is under SCCS control and hence appears in
# nametable, but should not be in this source product
# This list may include some extraneous files which aren't on the
# include list in the first place, but I to try to avoid that overhead of
# requiring the zip program to check against an overly long exlcude list. 
# For example I try to ensure obj dirs aren't on the exclude list, 
# However there will still be a lot of files on this list for the
# sources which exclude all of one O/S
# For example I try to ensure obj dirs aren't on the exclude list,
# The same should be done for
# build/solaris/bin
# build/solaris/lib
# build/solaris/include
# build/solaris/classes
# build/solaris/demo
# build/win32/bin
# build/win32/lib
# build/win32/include
# build/win32/classes
# build/win32/demo
# Note, these won't be included in any event, this concern is just about
# reducing the length of the exclude list.


cd $SRCDIR
EXCLUDEDFILES=/tmp/excluded-files.$$
$RM -f $EXCLUDEDFILES
touch $EXCLUDEDLIST
for f in  `$CAT $EXCLUDEDLIST` ; do

  if [ -f $f ]; then
   echo $f >> $EXCLUDEDFILES
  elif [ -d $f ]; then
   $FIND $f \( -name SCCS -o -name obj  -o -name obj_g -o -name CClassHeaders \) -prune -o -type f -print >> $EXCLUDEDFILES
  fi
done

# Create the zip archive of all source files, except those explicitly excluded
$CAT $ALLSRCFILES | $ZIP -@ -r ${DESTDIR}/${SRC_PROD_NAME} -x `$CAT $EXCLUDEDFILES` > /dev/null
cd $DESTDIR
/usr/bin/ln -s ${SRC_PROD_NAME} ${SRC_BUNDLE_NAME}

# Add the OEM source README at the top-level.
/usr/bin/cp -p $SRCDIR/src/share/doc/README.OEM $DESTDIR/README.OEM
cd $DESTDIR
$ZIP -u ${DESTDIR}/${SRC_PROD_NAME} README.OEM > /dev/null
$RM -f README.OEM

cleanup
exit 0
