/*
 * @(#)GenDocletBeanInfo.java	1.12 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import com.sun.javadoc.*;
import java.beans.*;
import java.util.*;

/**
 * properties supported and tag syntax:
 * @beaninfo
 *  bound: flag
 *  constrained: flag
 *  expert: flag
 *  hidden: flag
 *  preferred: flag
 *  description: string
 *  displayname: string
 *  propertyeditorclass: string (with dots: foo.bar.MyPropertyEditor
 *  customizerclass: string (w/dots: foo.bar.MyCustomizer)
 *  attribute: key1 value1
 *  attribute: key2 value2
 */
public class GenDocletBeanInfo {

    static String[] propertyDesc = { "bound",
				     "constrained",
				     "expert",
				     "hidden",
				     "preferred",
				     "displayname",
				     "propertyeditorclass",
				     "customizerclass",
				     "displayname",
				     "description",
				     "enum",
				     "attribute" };

    public static final String TRUE = "true";
    public static DocBeanInfo[] dB = null;
    public static Vector v = new Vector();
    public static String[] key,keyvalue,ekey,ekeyvalue;

    public static String fileDir;
    public static String templateDir;

    /**
     * Doclet options:
     * 		-t templateFileName
     *		-d outputdir
     */
    public static int optionLength(String option) {
	if (option.equals("-t"))
	    return 2;
	if (option.equals("-d"))
	    return 2;
	return 0;
    }

    /** @beaninfo
     * bound:true
     * constrained:false
     * expert:true
     * hidden:true
     * preferred:false
     * description: the description of this method can
     *              do all sorts of funky things. if it \n
     *              is indented like this, we have to remove
     *              all char spaces greater than 2 and also any hard-coded \n
     *              newline characters and all newlines
     * displayname: theString
     * propertyeditorclass: foo.bar.MyPropertyEditorClass
     * customizerclass: foo.bar.MyCustomizerClass
     * attribute:key1 value1
     * attribute: key2  value2
     *
     */
    public static boolean start(RootDoc doc) {
	for (int i = 0; i < doc.options().length; i++){
	    if (doc.options()[i][0].equals("-t")) {
		templateDir = doc.options()[i][1];
	    } else if (doc.options()[i][0].equals("-d")) {
		fileDir = doc.options()[i][1];
	    }
	}
	if (templateDir == null) {
	    System.err.println("-t option not specified");
	    return false;
	}
	if (fileDir == null) {
	    System.err.println("-d option not specified");
	    return false;
	}

	Hashtable dochash;

	/* "javadoc Foo.java Bar.java" will return:
	 *         "Foo Foo.I1 Foo.I2 Bar Bar.I1 Bar.I2"
	 * i.e., with all the innerclasses of classes specified in the command
	 * line.  We don't want to generate BeanInfo for any of these inner
	 * classes, so we ignore these by remembering what the last outer
	 * class was.  A hack, I admit, but makes the build faster.
	 */
	String previousClass = null;

	ClassDoc[] classes = doc.classes();

	for (int cnt = 0; cnt < classes.length; cnt++) {
	    String className = classes[cnt].qualifiedName();
	    if (previousClass != null &&
		className.startsWith(previousClass) &&
		className.charAt(previousClass.length()) == '.') {
		continue;
	    }
	    previousClass = className;
	    System.out.println(">>> Generating beaninfo for " +
			       classes[cnt].qualifiedName() + " ... ");

	    // get class info first here
	    Tag[] ct = classes[cnt].tags();
	    for (int i = 0; i < ct.length; i++){
		if (ct[i].kind().equalsIgnoreCase("@beaninfo")){
		    genDocletInfo(ct[i].text(),classes[cnt].name());
		}
	    }
	    int startPos = -1;
	    MethodDoc[] methods = classes[cnt].methods();
	    Tag[] pt;
	    for (int j = 0; j< methods.length; j++){
		// actually don't "introspect" - look for all
		// methods with a @beaninfo tag
		pt = methods[j].tags();
		for (int x = 0; x < pt.length; x++){
		    if (pt[x].kind().equalsIgnoreCase("@beaninfo")){
			if ((methods[j].name().startsWith("get")) ||
			    (methods[j].name().startsWith("set")))
			    startPos = 3;
			else if (methods[j].name().startsWith("is"))
			    startPos = 2;
			else
			    startPos = 0;
			String propDesc =
			    Introspector.decapitalize((methods[j].name()).substring(startPos));
			genDocletInfo(pt[x].text(),propDesc);
		    }
		}
	    }
	    // create our hashtable
	    if (v.isEmpty()){ // no beaninfo tags found
		dochash = new Hashtable();
	    }
	    else {
		dB = new DocBeanInfo[v.size()];
		dochash = new Hashtable(v.size());
		for (int z = 0; z < v.size(); z++){
		    dB[z] = (DocBeanInfo)v.elementAt(z);
		    dochash.put(dB[z].name, dB[z]);
		}
	    }
	    GenSwingBeanInfo generator = new GenSwingBeanInfo();
	    generator.genSwingBeanInfo(classes[cnt].name(),
				       dochash, fileDir, templateDir);
	}
	return true;
    }

    public static void genDocletInfo(String tag, String name){
	int beanflags = 0;
	int numattributes = 0;
	int numenum = 0;
	String desc = "null";
	String displayname = "null";
	String propertyeditorclass = "null";
	String customizerclass = "null";

	int plen = propertyDesc.length;
	String value = "null";
	int index;

	for (int j = 0; j < plen; j++){
	    index = 0;
	    if ((index = tag.indexOf(propertyDesc[j])) != -1){
		value = getValue((tag).substring(index),propertyDesc[j]);
		if (propertyDesc[j].equalsIgnoreCase("attribute")){
		    numattributes = getAttributes(value);
		}
		if (propertyDesc[j].equalsIgnoreCase("enum")){
		    numenum = getEnumAttributes(value);
		}
		else if (propertyDesc[j].equals("displayname")){
		    displayname = value;
		}
		else if (propertyDesc[j].equalsIgnoreCase("propertyeditorclass")){
		    propertyeditorclass = value;
		}
		else if (propertyDesc[j].equalsIgnoreCase("customizerclass")){
		    customizerclass = value;
		}
		else if ((propertyDesc[j].equalsIgnoreCase("bound"))
			 && (value.equalsIgnoreCase(TRUE)))
		    beanflags = beanflags | BeanInfoConstants.BOUND;
		else if ((propertyDesc[j].equalsIgnoreCase("expert"))
			 && (value.equalsIgnoreCase(TRUE)))
		    beanflags = beanflags | BeanInfoConstants.EXPERT;
		else if ((propertyDesc[j].equalsIgnoreCase("constrained"))
			 && (value.equalsIgnoreCase(TRUE)))
		    beanflags = beanflags | BeanInfoConstants.CONSTRAINED;
		else if ((propertyDesc[j].equalsIgnoreCase("hidden"))
			 && (value.equalsIgnoreCase(TRUE)))
		    beanflags = beanflags | BeanInfoConstants.HIDDEN;
		else if ((propertyDesc[j].equalsIgnoreCase("preferred"))
			 && (value.equalsIgnoreCase(TRUE)))
		    beanflags = beanflags | BeanInfoConstants.PREFERRED;
		else if (propertyDesc[j].equalsIgnoreCase("description")){
		    desc = value;
		}
	    }
	}
	/** here we create our doclet-beaninfo data structure, which we read in
	 *  later if it has anything worthwhile
	 */

	// initialize hashtable and propogate values
	DocBeanInfo db = new DocBeanInfo(name, beanflags, desc,displayname,
					 propertyeditorclass, customizerclass,
					 numattributes,key,keyvalue,numenum,ekey,ekeyvalue);
	v.addElement(db);
    }

    public static String getValue(String substring, String prop){
	StringTokenizer t;
	String value = "null";

	try {
	    /** if the propertyDesc is NOT the description, then we
	     *  parse until newline
	     *  if it is the description we read until the next token
	     *  and then look for a match in the last MAXMATCH index
	     *  and truncate the description
	     *  if it is the attribute we wead until no more
	     */
	    if (prop.equalsIgnoreCase("attribute")){
		StringBuffer tmp = new StringBuffer();
		try {
		    t = new StringTokenizer(substring, " :\n");
		    t.nextToken().trim();//the prop
		    // we want to return : key1 value1 key2 value2
		    while (t.hasMoreTokens()){
			tmp.append(t.nextToken().trim()).append(" ");
			tmp.append(t.nextToken().trim()).append(" ");
			String test = t.nextToken().trim();
			if (!(test.equalsIgnoreCase("attribute")))
			    break;
		    }
		} catch (Exception e){
		}
		value = tmp.toString();
	    }
	    else if (prop.equalsIgnoreCase("enum")){
		t = new StringTokenizer(substring, ":");
		t.nextToken().trim(); // the prop we already know
		StringBuffer tmp = new StringBuffer(t.nextToken().trim());
		for (int i = 0; i < propertyDesc.length; i++){
		    if (tmp.toString().endsWith(propertyDesc[i])){
			int len = propertyDesc[i].length();
			// trim off that
			tmp.setLength(tmp.length() - len);
			break;
		    }
		}
		value = tmp.toString();
	    }
	    else if (prop.equalsIgnoreCase("description")){
		t = new StringTokenizer(substring, ":");
		t.nextToken().trim(); // the prop we already know
		StringBuffer tmp = new StringBuffer(t.nextToken().trim());
		for (int i = 0; i < propertyDesc.length; i++){
		    if (tmp.toString().endsWith(propertyDesc[i])){
			int len = propertyDesc[i].length();
			// trim off that
			tmp.setLength(tmp.length() - len);
			break;
		    }
		}
		value = hansalizeIt(tmp.toString());
	    }
	    else {
		t = new StringTokenizer(substring, ":\n");
		t.nextToken().trim(); // the prop we already know
		value = t.nextToken().trim();
	    }

	    // now we need to look for a match of any of the
	    // property
	    return value;
	}
	catch (Exception e){
	    return "invalidValue";
	}
    }

    // we keep the attributes in a key value vector
    //
    public static int getAttributes(String str){
	int num = 0;

	try {
	    StringTokenizer t = new StringTokenizer(str," ");
	    num = t.countTokens()/2;
	    key = new String[num];
	    keyvalue = new String[num];
	    for (int j = 0; j < num; j++){
		key[j] = t.nextToken().trim();
		keyvalue[j] = t.nextToken().trim();
		if (keyvalue[j].equalsIgnoreCase("false")){
		    keyvalue[j] = "Boolean.FALSE";
		}
		else if (keyvalue[j].equalsIgnoreCase("true")){
		    keyvalue[j] = "Boolean.TRUE";
		}
	    }
	}
	catch (Exception e){
	}
	return num;
    }

    // we keep the attributes in a key value vector
    //
    public static int getEnumAttributes(String str){
	int num = 0;

	try {
	    // support multiple lines??
	    StringTokenizer t = new StringTokenizer(str," \n");
	    num = t.countTokens()/2;
	    ekey = new String[num];
	    ekeyvalue = new String[num];
	    for (int j = 0; j < num; j++){
		ekey[j] = t.nextToken().trim();
		ekeyvalue[j] = t.nextToken().trim();
	    }
	}
	catch (Exception e){
	}
	return num;
    }


    // looks for extra spaces, \n hard-coded and invisible,etc
    public static String hansalizeIt(String from){
	char [] chars = from.toCharArray();
	int len = chars.length;
	int toss = 0;

	// remove double spaces
	for (int i = 0; i < len; i++){
	    if ((chars[i] == ' ')) {
		if (i+1 < len) {
		    if ((chars[i+1] == ' ' ) || (chars[i+1] == '\n'))
			{
			    --len;
			    System.arraycopy(chars,i+1,chars,i,len-i);
			    --i;
			}
		}
	    }

	    if (chars[i] == '\n'){
		chars[i] = ' ';
		i -= 2;
	    }

	    if (chars[i] == '\\') {
		if (i+1 < len) {
		    if (chars[i+1] == 'n'){
			chars[i+1] = ' ';
			--len;
			System.arraycopy(chars,i+1, chars,i, len-i);
			--i;
		    }
		}
	    }
	}
	return new String(chars,0,len);
    }

}


