/*
 * @(#)GenSwingBeanInfo.java	1.13 98/08/26
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.beans.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

      
/**
 * A utlity for generating a standard Swing BeanInfo class.  The code
 * generated by GenSwingBeanInfo should be used as a starting point, 
 * it's neccessary to write things like the per property descriptions 
 * by hand.  To run the application:
 * <pre>
 * java GenSwingBeanInfo <class name>
 * </pre>
 * Code for a bean info class is written to out.  If the class is 
 * swing package, you don't need to fully specify its name.
 * 
 * 1/13/98: TODO - add ability to createBeandescriptor info for the class
 *
 * @version 1.13 08/26/98
 * @author Hans Muller
 * @author Rich Schiavi
 */
public class GenSwingBeanInfo 
{

  PrintStream out;
  String enumcode = " ";  // quick and dirty
  String packageName;


  PrintStream initOutputFile(String fileDir, String classname) {
    try {
      OutputStream out = new FileOutputStream(fileDir + "/" + classname + "BeanInfo.java");
      BufferedOutputStream bout = new BufferedOutputStream(out);
      return new PrintStream(out);
    }
    catch (IOException e){
      //    System.err.println("GenSwingBeanInfo: " + e.toString());
    }
    return null;
  }


  static void messageAndExit(String msg) {
    System.err.println("\n" + msg);
    System.exit(1);
  }

  
    /**
     * Load the contents of "SwingBeanInfo.template" into a string and
     * return the string.
     */
    String loadTemplate(String dir) {
	String fileName = "SwingBeanInfo.template";
	String template = "<no template>";
	
	try {
	    //InputStream stream = 
	    //	getClass().getResource(fileName).openStream();
	    File file = new File(dir + "/SwingBeanInfo.template");
	    DataInputStream stream = 
		new DataInputStream(new FileInputStream(file));
	    BufferedReader reader = 
		new BufferedReader(new InputStreamReader(stream));
	    StringBuffer buffer = new StringBuffer();
	    int c;
	    
	    while((c = reader.read()) != -1) {
		buffer.append((char)c);
	    }
	    template = buffer.toString();
	    reader.close();
	} catch (IOException e) {
	    messageAndExit("Couldn't load template \"" + fileName + "\" " + e);
	}
	return template;
    }


  /**
     * Generate the createPropertyDescriptor() calls, one per property.
     * A fully specified createPropertyDescriptor() call looks like this:
     * <pre>
     *      createPropertyDescriptor("contentPane", new Object[] {
     *			   BOUND, Boolean.TRUE,
     *  	     CONSTRAINED, Boolean.TRUE,
     *	     PROPERTYEDITORCLASS, package.MyEditor.cl  
     *  	     WRITEMETHOD, "setContentPane",
     *  	     DISPLAYNAME, "contentPane",
     *			  EXPERT, Boolean.FALSE,
     *			  HIDDEN, Boolean.FALSE,
     *		       PREFERRED, Boolean.TRUE,
     *		SHORTDESCRIPTION, "A top level window with a window manager border",
     *	       "random attribute","random value" 
     *	      }
     *	   );
     * </pre>
     *
     */
  String genPropertyDescriptors(BeanInfo info, Hashtable dochash) 
  {
    String code = "";
    PropertyDescriptor[] pds = info.getPropertyDescriptors();
    boolean hash_match = false;
    DocBeanInfo dbi = null;

    for(int i = 0; i < pds.length; i++) {
      // REMIND: why are is methods being found??
      if (pds[i].getReadMethod() != null){

	code += "\ncreatePropertyDescriptor(\"" + pds[i].getName() + "\", new Object[] {\n";
	   
	if (dochash.size() > 0){
	  if (dochash.containsKey(pds[i].getName())){
	    dbi = (DocBeanInfo)dochash.remove(pds[i].getName());
	    setDocInfoProps(dbi,pds[i]);
	    hash_match = true;
	  }
	}
	// Do I need to 
	if (hash_match){
	  if ((dbi.beanflags & BeanInfoConstants.BOUND) != 0){ 
	    code += "               BOUND, Boolean.TRUE,\n";
	  }
	  else {
	    code += "               BOUND, Boolean.FALSE,\n";	  
	  }
	}

	if (pds[i].isConstrained()) {
	  code += "         CONSTRAINED, Boolean.TRUE,\n";
	}

	if (pds[i].getPropertyEditorClass() != null) {
	  String className = pds[i].getPropertyEditorClass().getName();
	  code += " PROPERTYEDITORCLASS, " + className + ".class,\n";
	}
	else if ((hash_match) && (!(dbi.propertyeditorclass.equals("null")))){
	  code += " PROPERTYEDITORCLASS, " + dbi.propertyeditorclass + ".class,\n";
	}

	if ((hash_match) && (!(dbi.customizerclass.equals("null")))){
	  code += " CUSTOMIZERCLASS, " + dbi.customizerclass + ".class,\n";
	}

	if ((hash_match) && (dbi.numenum > 0)){
	  enumcode += "\n\t\tObject[] " + pds[i].getName() +
	    "Enumeration = new Object[] { \n";
	  for (int z = 0; z< dbi.numenum; z++){
	    if (z == dbi.numenum-1)
	      enumcode += "\t\t\t\"" + dbi.ekey[z] + "\" ,   new Integer(" + dbi.evalue[z] + ") , \"" + dbi.evalue[z] + "\"" + "  \n\t\t};\n";
	    else
	      enumcode += "\t\t\t\"" + dbi.ekey[z] + "\" ,   new Integer(" + dbi.evalue[z] + ") , \""  + dbi.evalue[z] + "\"" + ",\n";
	  }		
	  code += "         \"enumerationValues\", " + pds[i].getName()+"Enumeration ,\n";
	}

	if (!pds[i].getDisplayName().equals(pds[i].getName())) {
	  code += "         DISPLAYNAME, \"" + pds[i].getDisplayName() + "\",\n";
	}

	if (pds[i].isExpert()) {
	  code += "              EXPERT, Boolean.TRUE,\n";
	}

	if (pds[i].isHidden()) {
	  code += "              HIDDEN, Boolean.TRUE,\n";
	}

	/* PENDING(hmuller) JDK1.2 only  */

	if (pds[i].isPreferred()) {
	  code += "           PREFERRED, Boolean.TRUE,\n";
	} 

	// random attributes 
	if (hash_match){
	  int na = dbi.numattributes;
	  for (int j = 0; j < dbi.numattributes; j++){
	    if ((dbi.value[j].equals("Boolean.FALSE")) ||
		(dbi.value[j].equals("Boolean.TRUE")))
	      code += "              \"" + dbi.key[j] + "\","  + dbi.value[j] + ",\n";
	    else
	      code += "              \"" + dbi.key[j] + "\"," + "\"" + dbi.value[j] + "\",\n";
	  }
	  hash_match = false;
	}

	code += "    SHORTDESCRIPTION, \"" + pds[i].getShortDescription() + "\",\n";

	/* Print the closing brackets.  If this is the last array initializer,
	 * don't print the trailing comma.
	 */
	if (i == (pds.length - 1)) {
	  code += "  }\n)\n";
	}
	else {
	  code += "  }\n),\n";
	}
      }
    }
    return code;
  }

  public void genSwingBeanInfo(String classname, Hashtable dochash, String fileDir, String templateDir)
  {
    /* The following initial values are just examples.  All of these
     * fields are initialized below.
	 */
    Class cls = null;

    String beanClassName = "JInternalFrame";
    String beanClassObject = "javax.swing.JInternalFrame.class";
    String beanDescription = "<A description of this component>.";
    String beanPropertyDescriptors = "<createSwingPropertyDescriptor code>";
    String classPropertyDescriptors = "<createSwingClassPropertyDescriptor code>";
      
      
      // REMIND: needed?? class.toString()
      // Introspector.setBeanInfoSearchPath(new String[] {classname}); ????
      // check out original code - what did this do? point it to source
      // directory or 1.1_classes??
      
    String prefixes[] = {
      "",
      "javax.swing",
      "javax.swing.plaf.basic",
      "javax.swing.text",
      "com.sun.java.swing.plaf.motif"
    };

    String suffixes[] = {
      "",
      ".class"
    };

    for(int i = 0; i < prefixes.length; i++) {
      for(int j = 0; j < suffixes.length; j++) {
	try {
	  cls = Class.forName(prefixes[i] + "." + classname + suffixes[j]);
	  packageName = prefixes[i];
	}
	catch (ClassNotFoundException e) {
	  // System.err.println("ClassNotFound: " + e.toString());
	}
      }
    }

    if (cls == null){
      messageAndExit("Can't find class: " + classname);
    } 

    out = initOutputFile(fileDir,classname); 

      /* Run The Java Beans Introspector And Initialize The String Beanfoo 
       * Variables.
       */

    BeanInfo beanInfo = null;
    BeanDescriptor beanDescriptor = null;
	
    try {
      beanInfo = Introspector.getBeanInfo(cls, cls.getSuperclass());
      beanDescriptor = beanInfo.getBeanDescriptor();
    }
    catch (IntrospectionException e) {
      messageAndExit("Introspection failed for " + cls.getName() + " " + e);
    }

    beanClassName = beanDescriptor.getName();
    beanClassObject = cls.getName() + ".class";

    if (dochash.size() > 0){
      if (dochash.containsKey(beanClassName)){
	DocBeanInfo dbi = (DocBeanInfo)dochash.remove(beanClassName);
	// REMIND: set props on the bean class/modify template
	classPropertyDescriptors = genClassInfoProps(dbi,beanClassName);
	// setDocInfoProps(dbi,pds[i]);
	if (!(dbi.desc.equals("null")))
	  beanDescription = dbi.desc;
      }
      else 
	beanDescription = beanDescriptor.getShortDescription();	    
    }
    else 
      beanDescription = beanDescriptor.getShortDescription();

    beanPropertyDescriptors = genPropertyDescriptors(beanInfo,dochash);

    /* Dump the template to out, substituting values for
	 * @(token) tokens as they're encountered.
	 */

    int currentIndex = 0;
    // not loading this to get around build issue for now
    String template = loadTemplate(templateDir);

    while(currentIndex < template.length()) {
      int tokenStart = template.indexOf("@(", currentIndex);
      if (tokenStart == -1) {
	out.print(template.substring(currentIndex, template.length()));
	break;
      }
      else {
	out.print(template.substring(currentIndex, tokenStart));

      }
      if (tokenStart != -1) {
	int tokenEnd = template.indexOf(")", tokenStart);
	if (tokenEnd == -1) {
	  messageAndExit("Bad @(<token>) beginning at " + tokenStart);
	}

	String token = template.substring(tokenStart+2, tokenEnd);

	if (token.equals("BeanClassName")) {
	  out.print(beanClassName);
	}
	else if (token.equals("ClassDescriptors")){
	  String s = classPropertyDescriptors; 
	  if (!(s.equals("<createSwingClassPropertyDescriptor code>"))){

	    String indent = "";

	    /* Find the newline that preceeds @(BeanPropertyDescriptors) to 
	     * calculate the indent.
	     */
	    for(int i = tokenStart; i >= 0; i--) {
	      if (template.charAt(i) == '\n') {
		char[] chars = new char[tokenStart - i];
		for(int j = 0; j < chars.length; j++) {
		  chars[j] = ' ';
		}
		indent = new String(chars);
		break;
	      }
	    }

	    int i = 0;
	    while(i < s.length()) {
	      int nlIndex = s.indexOf('\n', i);
	      out.print(s.substring(i, nlIndex+1));
	      out.print(indent);
	      i = nlIndex + 1;
	    }
	  }
	}
	else if (token.equals("BeanPackageName")){
	  out.print(packageName);
	}
	else if (token.equals("BeanClassObject")) {
	  out.print(beanClassObject);
	}
	else if (token.equals("BeanDescription")) {
	  out.print(beanDescription);
	}
	else if (token.equals("EnumVariables")){
	  out.print(enumcode);
	}
	else if (token.equals("BeanPropertyDescriptors")) {

	  String indent = "";

	  /* Find the newline that preceeds @(BeanPropertyDescriptors) to 
	   * calculate the indent.
	   */
	  for(int i = tokenStart; i >= 0; i--) {
	    if (template.charAt(i) == '\n') {
	      char[] chars = new char[tokenStart - i];
	      for(int j = 0; j < chars.length; j++) {
		chars[j] = ' ';
	      }
	      indent = new String(chars);
	      break;
	    }
	  }

	  /* Print beanPropertyDescriptors, replacing each newline with
	   * a newline followed by indent.
	   */
	  String s = beanPropertyDescriptors;
	  int i = 0;
	  while(i < s.length()) {
	    int nlIndex = s.indexOf('\n', i);
	    out.print(s.substring(i, nlIndex+1));
	    out.print(indent);
	    i = nlIndex + 1;
	  }
	}
	else if (token.equals("#")){
	  out.print(beanClassName + "BeanInfo");
	}
	else {
	  messageAndExit("Unrecognized token @(" + token + ")");
	}
	currentIndex = tokenEnd + 1;
      }
    }
    out.close();
  }

  public String genClassInfoProps(DocBeanInfo dbi, String beanClassName){
    String code = "";
    int beanflags = dbi.beanflags;

    // we support export, hidden, preferred
    if ((beanflags & BeanInfoConstants.EXPERT) != 0)
      code += " EXPERT, Boolean.TRUE,\n";
    if ((beanflags & BeanInfoConstants.HIDDEN) !=0)
      code += "                    HIDDEN, Boolean.TRUE,\n";
    /* 1.2 only - make sure build flag build using 1.2 */
    if ((beanflags & BeanInfoConstants.PREFERRED) !=0)
      code += "                 PREFERRED, Boolean.TRUE,\n";
    if (!(dbi.customizerclass.equals("null")))
      code += "            CUSTOMIZERCLASS, " + dbi.customizerclass + ".class,\n";
    int na = dbi.numattributes;
    for (int j = 0; j < dbi.numattributes; j++){
      if ((dbi.value[j].equals("Boolean.FALSE")) ||
	  (dbi.value[j].equals("Boolean.TRUE")))
	code += "              \"" + dbi.key[j] + "\","  + dbi.value[j] + ",\n";
      else
	code += "              \"" + dbi.key[j] + "\"," + "\"" + dbi.value[j] + "\",\n";
    }

    return code;
  }

  public void setDocInfoProps(DocBeanInfo dbi, PropertyDescriptor pds){
    int beanflags = dbi.beanflags;

    if ((beanflags & BeanInfoConstants.BOUND) != 0){
      // setting the Bounds to true wont work since the Beans spec
      // determines that for beans with a changelistener - every property is BOUND!!
      pds.setBound(true);
    }
    if ((beanflags & BeanInfoConstants.EXPERT) != 0)
      pds.setExpert(true);
    if ((beanflags & BeanInfoConstants.CONSTRAINED) != 0)
      pds.setConstrained(true);
    if ((beanflags & BeanInfoConstants.HIDDEN) !=0)
      pds.setHidden(true);
    /* 1.2 only - make sure build flag build using 1.2 */
    if ((beanflags & BeanInfoConstants.PREFERRED) !=0)
      pds.setPreferred(true); 
    if (!(dbi.desc.equals("null"))){
      pds.setShortDescription(dbi.desc);
    }
    if (!(dbi.displayname.equals("null"))){
      pds.setDisplayName(dbi.displayname);
    }
  }
}


