#
# @(#)check_offsets.nawk	1.2 96/05/28
#
# Ensure that opcode table is in the right order.
# This nawk script takes as input the file opcodes.list and produces
# as output an adb script. The output should be piped into adb, running
# against the java interpreter we just made using executeJava_sparc.s, thusly:
#	nawk -f check_offsets.nawk < blah/blah/opcodes.list | adb fu/bar/java -
# It will cause adb to check the offsets of all the symbols in the table.
# For each opcode in the opcodes.list file, there are two checks:
# 1) does it exist
#    this is checked by generating the adb command sequence
#	0>a
#	symbol\.0>a
#	<a,#<a="symbol UNDEFINED!"
#    If the symbol is defined, then <a will contain a non-zero value, and
#    so the third command will have a repeat count of zero so will print nothing.
#    If the symbol is NOT defined, the second command will fail, <a will contain
#    zero, and the third command will have a repeat count of 1 and will print.
# 2) does it have the right value.
#    This is checked using the single adb command
#	symbol\.0-nop\.0%0t64,##(symbol\.0-nop\.0-V)=" symbol has wrong opcode of " D
#    where V is 64 times the expected opcode value, derived from ordinal
#    position in the input file.
# Test 2 is not sufficient to test for non-existence of symbol only because
# of the cryptic way in which the command fails without printing out the 
# name of the symbol the non-existance of which is causing grief.
#

BEGIN{ i = 0 }
/^[a-z]/{
	print "0>a"
	print $1 "\\.0>a"
	print "<a,#<a=\"" $1 " UNDEFINED!\""
	print $1 "\\.0-nop\\.0%0t64,##(" $1 "\\.0-nop\\.0-0t" i ")=\"" $1 " has wrong opcode of \" Dn"
	i+=64
	next
}
/./{ next }
