#
# @(#)extract_offsets.nawk	1.4	98/08/03
#
# This program used, along with m4 macros, to extract from
# executeJava_sparc.s all references to C-language field references
# and defined constants.
#
# 1) run m4 -DEXTRACT_OFFSETS over executeJava_sparc.s. The result will
#    be records starting CONST or FIELD, for every use of
#    DEFINED_CONSTANT or STRUCT_OFFSET macros, respectively.
# 2) Pipe that through sort -u to yield only one instance of each record.
# 3) Pipe that through this program. When concatenated with enough header
#    information to #include all the required .h files, the product will
#    be a C program.
# 4) Compile and run the resulting C program. Redirect its output to
#    file struct_offsets.include.
# 5) Run the assembler with m4 preprocessing, as -m executeJava_sparc.s
#    This will include struct_offsets.include, where all the required
#    offset and constant information will be defined!
#
#
# Format of the input:
# Each line begins with CONST or FIELD
# 1) CONST lines contain only one other field, which is the name
#    of a C #define'd constant.
# 2) FIELD lines contain three fields: the first is a struct tag
#    and the second is a field name in that struct. The third is
#    the mangled name by which the offset will be known in the assembly
#    program.
# Both lines result in generation of a printf statement, which in turn
# generates an assembler symbol assignment.
#
BEGIN{
	print "#include \"oobj.h\""
	print "#include \"interpreter.h\""
	print "#include \"opcodes.h\""
	print "#include \"tree.h\""
	print "#include \"typecodes.h\""
	print "#include \"stddef.h\""
	print "#include \"vmprofiler.h\""
	print "main(int argc, char ** argv){"
	print "	puts(\"! File automagically generated by m4 and nawk\" );"
	print "	puts(\"! Do not bother editing this! Find the source!\" );"
}
/^CONST/{ print "	printf(\"	"$2" = %d\\n\", (int)"$2" );" }

/^FIELD/{ print "	printf(\"	"$4" = %d\\n\", (int) offsetof(struct "$2", "$3") );" }

END{
	print "	return(0);"
	print "}"
}
