%{
/*
 * @(#)lexer.l	1.5 98/08/20
 *
 * Copyright 1996 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
%}
%{
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <ctype.h>
enum _bool_t { FALSE, TRUE };
typedef enum _bool_t bool_t;
#include "y.tab.h"
char* mprintf(char *format, ...);
char* comment;
char* purgeWhiteSpace(char *, int length);

%}
%a  4000
%o  8000
%p  3000
R (ax|bx|cx|dx|bp|sp|si|di)
%%
^.*\n            { comment = purgeWhiteSpace(yytext, yyleng); REJECT; } 
e{R}             { yylval.sval = mprintf("%%%s", yytext); return(REG32); }
{R}              { yylval.sval = mprintf("%%%s", yytext); return(REG16); }
[a-d][hl]        { yylval.sval = mprintf("%%%s", yytext); return(REG8); }
st               { return(FREG); }
AND		 { return AND; }
OR               { return OR; }
SHL		 { return SHL; }
SHR              { return SHR; }
BYTE             { yylval.ival =  8; return(SIZE); }
byte             { yylval.ival =  8; return(SIZE); }
WORD             { yylval.ival = 16; return(SIZE); }
word             { yylval.ival = 16; return(SIZE); }
DWORD            { yylval.ival = 32; return(SIZE); }
dword            { yylval.ival = 32; return(SIZE); }
DD               { yylval.ival = 32; return(SIZE); }
QWORD            { yylval.ival = 64; return(SIZE); }
qword            { yylval.ival = 64; return(SIZE); }
TWORD            { yylval.ival = 80; return(SIZE); }
tword            { yylval.ival = 80; return(SIZE); }
NEAR             { yylval.ival = 0; return(DISTANCE); }
FAR              { yylval.ival = 1; return(DISTANCE); }
align            return(ALIGN);
[Ee][Qq][Uu]     return(EQU);
END              return(END);
ENDP             return(ENDP);
EXTERN           return(EXTERN);
EXTERNDEF        return(EXTERNDEF);
FLAT             return(FLAT);
model            return(MODEL);
OFFSET           return(OFFSET);
PTR              return(PTR);
ptr              return(PTR);
PROC             return(PROC);
PUBLIC           return(PUBLIC);
SEGMENT          return(SEGMENT);
SHORT            return(SHORT);
_DATA            { yylval.sval = "data"; return SEGMENT_TYPE; }
_TEXT            { yylval.sval = "text"; return SEGMENT_TYPE; }
_BSS             { yylval.sval = "bss"; return SEGMENT_TYPE; }
rep              { yylval.sval = "rep"; return(PREFIX); }
repe             { yylval.sval = "repz"; return(PREFIX); }
repne            { yylval.sval = "repnz"; return(PREFIX); }
[0-9]+           { yylval.ival = strtoll(yytext, NULL, 10); return NUMBER; }
[01]+[bB]        { yylval.ival = strtoll(yytext, NULL, 2); return NUMBER; }
[0-9][0-9a-fA-F]+[hH] { 
                   yylval.ival = strtoll(yytext, NULL, 16); return NUMBER; }
[0-7]+[oOqQ]     { yylval.ival = strtoll(yytext, NULL, 8); return NUMBER; }
@@               { yylval.sval = "1"; return IDENTIFIER; } 
@F               { yylval.sval = "1f"; return IDENTIFIER; }
@B               { yylval.sval = "1b"; return IDENTIFIER; }
[a-zA-Z0-9$_@]+  { if (yytext[0] == '$') 
                      yylval.sval = mprintf("L%s", yytext);
                   else 
                      yylval.sval = strdup(yytext);
                   return IDENTIFIER;
                 }
\"[^"]*\"        { yylval.sval = strdup(yytext); return(STRING); }
;.*\n            { yylval.sval = comment; comment = ""; return(NEWLINE); }
[ \t\r]          /* ignore white space */;
TITLE.*\n        { yylval.sval = comment; comment = ""; return(NEWLINE); }
\n               { yylval.sval = comment; comment = ""; return(NEWLINE); }
.	         return(yytext[0]);
%%


char* purgeWhiteSpace(char *comment, int length) { 
    int i;
    while (isspace(*comment)) { 
        comment++; length--;
    }
    while ((length > 0) && isspace(comment[length - 1])) { 
        length--; 
    }
    { 
        char *result = (char *)malloc(length + 1);
        memcpy(result, comment, length);
        result[length] = 0;
        comment = result;
    }
    for (i = 0; i < length; i++) 
        if (isspace(comment[i])) 
            comment[i] = ' ';
    return comment;
}

struct yytoktype { char *t_name; int t_val; };
extern struct yytoktype yytoks[];


#ifdef YYDEBUG
void debugLexer() {
    int i, yy_i;
    do {
        i = yylex();
        printf("%d: ", i);
	fflush(stdout);
        for ( yy_i = 0; yytoks[yy_i].t_val >= 0; yy_i++ ) {
            if ( yytoks[yy_i].t_val == i )
                break;
        }
        if (yytoks[yy_i].t_val == i) { 
            printf( "%s\n", yytoks[yy_i].t_name );
	} else if (i == ' ') { 
	    printf("space");
	} else if (isprint(i)) { 
	    printf("%c (%d)\n", i, i);
	} else if ((0 <= i) && (i <= 26)) {
	    printf("^%c\n", (i - 1) + 'A');
	} else { 
	    printf("\\%x\n", i);
	}
    } while (i > 0);
}

#endif

