%{
/*
 * @(#)parser.y	1.4 98/08/20
 *
 * Copyright 1996 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
%}

%token REG32, REG16, REG8, FREG
%token IDENTIFIER, PREFIX, NUMBER, SIZE, STRING, PROC, MODEL, END
%token PTR, EQU, SHORT, ALIGN, EXTERNDEF, OFFSET, FLAT
%token PUBLIC, EXTERN, DISTANCE, ENDP, NEWLINE
%token SEGMENT, SEGMENT_TYPE
%token ':', ',', '=', '.', '[', ']', '(', ')'

%left AND OR SHL SHR
%left '+'  '-'
%left '*'  '/'
%{
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/types.h>

enum _bool_t { FALSE, TRUE };
typedef enum _bool_t bool_t;
bool_t isBranch;		/* true if operand is a branch */

struct memval;
struct optval;
struct exprval;

#define NEW(xx) xx = malloc(sizeof(*xx))

/*PRINTFLIKE1*/
char* mprintf(char *format, ...);

static char* memvalToString(struct memval *mem);
static char* exprToString(struct exprval *expr);
static void join1Expression(struct exprval *, struct exprval *, long);
void join2Expressions(struct exprval *result, struct exprval *expr1, 
		      struct exprval *expr2, long op);
static bool_t memvalAdd(struct memval *old, struct memval *new);
static char* getSize1(char *opcode, int size);
static char* getSize2(char *opcode, int src, int dst);
static char* getSize3(char *opcode, int shift, int src, int dst);
static char *sizeToString(int x);

static bool_t isBranchOp(char *opcode) ;

static void assignValue(char *string, long value);
static bool_t hasValue(char *string, long *value);

static char *rewriteOpcode(char *name);

bool_t gcc = FALSE;
bool_t error = FALSE;
%}

%union {
   char *sval;
   long  ival;
   struct exprval { 
        bool_t isInt;
        union { 
	  long i;
	  char *s;
	} value;
   } eval;
   struct memval {
       struct exprval *offset;
       char *reg;
       char *index;
  } mval;
  struct optval {
        char *svalue;
        int size;
  } oval;
}

%type <sval> IDENTIFIER, STRING, REG32, REG16, REG8, PREFIX, NEWLINE, 
%type <sval> Opcode, SEGMENT_TYPE
%type <eval> Expression
%type <ival> SIZE, NUMBER
%type <oval> Operand, MemoryOp, RegisterOp, ExpressionOp
%type <mval> MemRef, MemRefBrackets  MemList  MemItem

%%

Program:    
	   /* empty */ 
	| Program OneLine NEWLINE 
	    {   if (strlen($3) > 0) 
		    printf(" // %s", $3); 
	        printf("\n"); 
	    }
        ;

OneLine: Instruction | RandomJunk | /* empty */ | error ;

Instruction:
	  Opcode 
            { printf("\t%s", $1); } 
	| Opcode Operand
	    { printf("\t%s%s\t%s", 
		     $1, getSize1($1, $2.size), $2.svalue); } 
        | Opcode Operand ',' Operand
	    { printf("\t%s%s\t%s, %s", 
		     $1, getSize2($1, $4.size, $2.size), 
		     $4.svalue, $2.svalue); } 
    	| Opcode Operand ',' Operand ',' Operand
	    { printf("\t%s%s\t%s, %s", /* elide the %cl */
		     $1, getSize3($1, $6.size, $4.size, $2.size), 
		     $4.svalue, $2.svalue); } 
	| PREFIX Opcode
	    { printf("\t%s;\t%s", $1, $2); } 
        | IDENTIFIER ':' 
            { printf("%s:", $1); } ;
        ;

RandomJunk:   
	  IDENTIFIER EQU Expression 
            { printf("%s = %s", $1, exprToString(&$3)); 
	      if ($3.isInt) assignValue($1, $3.value.i);
	    }
        | IDENTIFIER '=' Expression 
            { printf("%s = %s", $1, exprToString(&$3)); 
	      if ($3.isInt) assignValue($1, $3.value.i);
	    }
        | EXTERNDEF IDENTIFIER ':' PTR
	    { printf(".globl %s", $2); }
        | EXTERN  IDENTIFIER ':' SIZE
	    { printf(".globl %s", $2); }
        | EXTERN  IDENTIFIER ':' DISTANCE
	    { printf(".globl %s", $2); }
        | PUBLIC IDENTIFIER
	    { printf(".globl %s", $2); }
        | ALIGN NUMBER
            { printf("\t.align %ld", $2); }
        | SIZE Expression
            { printf("\t.%s %s", sizeToString($1), exprToString(&$2)); }
        | SIZE STRING ',' NUMBER
            { if ($1 == 8 && $4 == 0) 
		  printf("\t.string %s", $2); 
	      else { yyerror("syntax error"); YYERROR; } 
	    }
        | SIZE STRING 
            { if ($1 == 8)
		  printf("\t.ascii %s", $2); 
	       else { yyerror("syntax error"); YYERROR; } 
	    }
        | SIZE OFFSET FLAT ':' IDENTIFIER
            { printf("\t.%s %s", sizeToString($1), $5); }
        | IDENTIFIER PROC DISTANCE /* start of function */
            { printf("\t.type %s, @function\n", $1);
	      printf("%s:", $1);
	    }
        | SEGMENT_TYPE SEGMENT	/* start egment */
            { printf("\t.%s", $1); } 
	| IDENTIFIER ENDP	/* end of procedure */
            { printf("\t.size %s, . - %s", $1, $1); }
        | '.' NUMBER 
	| END
	| '.' MODEL FLAT
        ;


Opcode:	IDENTIFIER { $$ = rewriteOpcode($1); isBranch = isBranchOp($$); };

Operand:   RegisterOp   { if (isBranch) $$.svalue = mprintf("*%s", $1.svalue); }
	|  MemoryOp     { if (isBranch) $$.svalue = mprintf("*%s", $1.svalue); }
        |  ExpressionOp { if (!isBranch)$$.svalue = mprintf("$%s", $1.svalue); }
	;

RegisterOp: 
	  REG32    { $$.svalue = $1;    $$.size = 32; }
        | REG16    { $$.svalue = $1;    $$.size = 16; }
        | REG8     { $$.svalue = $1;    $$.size = 8; }
        | FREG     { $$.svalue = "%st"; $$.size = -1; }
        | FREG '(' NUMBER ')'
                   { $$.svalue = mprintf("%%st(%ld)", $3); $$.size = -1; }
        | SIZE PTR RegisterOp
                   { $$ = $3; 
		     if ($$.size != $1) { yyerror("BAD SIZE"); YYERROR; }
		   }
        ;

MemoryOp: MemRef          
             { $$.svalue = memvalToString(&$1); $$.size = 0; }
        | SIZE PTR MemRef 
             { $$.svalue = memvalToString(&$3); $$.size = $1; }
        ;

ExpressionOp:
	  Expression           { $$.svalue = exprToString(&$1); $$.size = 0; } 
        | SIZE PTR Expression  { $$.svalue = exprToString(&$3); $$.size = $1; } 
        | SHORT Expression     { $$.svalue = exprToString(&$2); $$.size = 0; }
        ;

MemRef:   MemRefBrackets
        | Expression MemRefBrackets           
             { NEW($$.offset); *$$.offset = $1; $$.index = $$.reg = NULL;
	       memvalAdd(&$$, &$2);
	    }
        ;

MemRefBrackets:
          MemRefBrackets '[' MemList ']' 
             { $$ = $1; 
 	       if (!memvalAdd(&$$, &$3)) { yyerror("memref error"); YYERROR; }
             }
        | '[' MemList ']' 
             { $$ = $2; }
        ;

MemList:   /* one or more things inside brackets */
	  MemItem		
        | MemList '+' MemItem 
            { $$ = $1;
	      if (!memvalAdd(&$$, &$3)) { yyerror("memref error"); YYERROR; }
            }
        | MemList '-' Expression
	    { struct memval temp;
	      $$ = $1;
	      NEW(temp.offset);
	      join1Expression(temp.offset, &$3, '-');
	      temp.reg = temp.index = NULL;
	      memvalAdd(&$$, &temp);
	    }   
        ;

MemItem:  Expression %prec '*'
             { NEW($$.offset); *$$.offset = $1; $$.index = $$.reg = NULL; }
	| REG32
             { $$.offset = NULL; $$.reg = $1; $$.index = NULL; }
        | REG32 '*' NUMBER
             { $$.offset = NULL; $$.reg = NULL; 
	       $$.index = mprintf("%s,%ld", $1, $3); }
        ;

Expression: 
	  IDENTIFIER        
             { long temp; 
	       $$.isInt = hasValue($1, &temp);
	       if ($$.isInt) $$.value.i = temp; else $$.value.s = $1;
	     }
        | NUMBER                     { $$.isInt = TRUE; $$.value.i = $1; }
        | Expression '+' Expression  { join2Expressions(&$$, &$1, &$3, '+'); }
        | Expression '-' Expression  { join2Expressions(&$$, &$1, &$3, '-'); }
        | Expression '*' Expression  { join2Expressions(&$$, &$1, &$3, '*'); }
        | Expression '/' Expression  { join2Expressions(&$$, &$1, &$3, '/'); }
        | Expression AND Expression  { join2Expressions(&$$, &$1, &$3, AND); }
        | Expression OR Expression   { join2Expressions(&$$, &$1, &$3, OR);  }
        | Expression SHL Expression  { join2Expressions(&$$, &$1, &$3, SHL); }
        | Expression SHR Expression  { join2Expressions(&$$, &$1, &$3, SHR); }
        | '(' Expression ')'         { join1Expression(&$$, &$2, '('); }
        | '-' Expression             { join1Expression(&$$, &$2, '-'); }
        | '+' Expression             { join1Expression(&$$, &$2, '+'); }
        ;
%%

/* Print out a memval as an appropriate string */
char *memvalToString(struct memval *mem) { 
    char *offset = (mem->offset == NULL) ? "" : exprToString(mem->offset);
    if (mem->index == NULL && mem->reg == NULL) {
        return offset;
    } else { 
        if (!strcmp(offset, "0")) 
	    offset = "";
	if (mem->reg == NULL) 
	    mem->reg = "";
	if (mem->index == NULL) 
	    return mprintf("%s(%s)", offset, mem->reg);
	else 
	    return mprintf("%s(%s,%s)", offset, mem->reg, mem->index);
    }
}

/* Add the information in the "new" memval structure to the information already
 * contained in the "old" memval structure
 * 
 * Return non-zero is successful, zero if unsuccessful.
 */
bool_t memvalAdd(struct memval *old, struct memval *new) { 
    if (new->offset) { 
        if (old->offset == NULL)
	    old->offset = new->offset;
	else { 
	    join2Expressions(old->offset, old->offset, new->offset, '+');
        }
    } 
    if (new->reg) { 
        /* A new register can either go into the register slot or the scaled
	 * index slot    */
        if (old->reg == NULL) 
	    old->reg = new->reg;
	else if (old->index == NULL) 
	    old->index = new->reg;
	else 
	    return FALSE;
    } 
    if (new->index) { 
        /* A new scaled index can only go into the index slot */
	if (old->index != NULL) 
	    return FALSE;
	old->index = new->index;
    }
    return TRUE;
}

/* Return the string representation of an expression */
char *exprToString(struct exprval *expr) { 
     if (expr == NULL) {
         return "0";
     } else if (expr->isInt) { 
         long value = expr->value.i;
	 /* Use hex for large values, decimal for small.  */
	 char *format = ((value >= 1024 || value <= -1024) ? "%#lx" : "%ld");
	 return mprintf(format, value);
     } else { 
         return expr->value.s;
     }
}

/* Apply an operator to an expression */
void join1Expression(struct exprval *result, struct exprval *expr, long op) {
    if (expr->isInt) {
        result->isInt = TRUE;
	switch(op) { 
	    case '-':           result->value.i = -(expr->value.i); break;
	    case '(': case '+': result->value.i = +(expr->value.i); break;
	    default:            printf("?????"); break;
	} 
    } else { 
        char *s1 = exprToString(expr);
	result->value.s = (op == '-') ? mprintf("-%s", s1) : s1;
	/* 
	char *joiner = (op == '-') ? "-" : "";
	result->value.s = gcc ? mprintf("(%s%s)", joiner, s1);
	                     : mprintf("[%s%s]", joiner, s1);
	*/
        result->isInt = FALSE;
    }
}

void join2Expressions(struct exprval *result, struct exprval *expr1, 
		      struct exprval *expr2, long op) {
    /* Don't assign yet, because result may be the same as expr1 or expr2 */
    if (expr1->isInt && expr2->isInt) { 
        result->isInt = TRUE;
	switch(op) { 
	    case '+': result->value.i = expr1->value.i + expr2->value.i; break;
	    case '-': result->value.i = expr1->value.i - expr2->value.i; break;
	    case '*': result->value.i = expr1->value.i * expr2->value.i; break;
	    case '/': result->value.i = expr1->value.i / expr2->value.i; break;
	    case AND: result->value.i = expr1->value.i & expr2->value.i; break;
	    case OR:  result->value.i = expr1->value.i | expr2->value.i; break;
	    case SHR: result->value.i = expr1->value.i >> expr2->value.i; break;
	    case SHL: result->value.i = expr1->value.i << expr2->value.i; break;
	    default:  printf("?????"); break;
	} 
    } else { 
        static count = 0;
        char *s1 = exprToString(expr1);
	char *s2 = exprToString(expr2);
	char *gensym = mprintf("GENSYM%5.5d", ++count);
        char *joiner;
        switch(op) { 
	    case '+':  joiner = "+";  break;
	    case '-':  joiner = "-";  break;
	    case '*':  joiner = gcc ? "*" : "\\*"; break;
	    case '/':  joiner = gcc ? "/" : "\\/"; break;
	    case AND:  joiner = "&";  break;
	    case OR:   joiner = "|";  break;
	    case SHR:  joiner = ">>"; break;
	    case SHL:  joiner = "<<";  break;
	    default:   joiner = "????"; break;
        }
	printf("%s = %s %s %s\n", gensym, s1, joiner, s2);
	result->value.s = gensym;
	result->isInt = FALSE;
	/*
	result->value.s = gcc ? mprintf("(%s %s %s)", s1, joiner, s2);
	                     : mprintf("[%s %s %s]", s1, joiner, s2);
        */				    
    }
}


/* Convert a size to the appropriate word */
char *sizeToString(int x) { 
    switch(x) { 
       case 8: return "byte";
       case 16: return "word";
       case 32: return "long";
       case 64: return "double";
       default: return mprintf("@@@@%d@@@@", x);
    }
}

char* mprintf(char *format, ...);


bool_t isBranchOp(char *opcode) {
    return (opcode[0] == 'j' || !strncmp(opcode, "loop", 4) 
	    || !strcmp(opcode, "call"));
}

char *
getSize1(char *opcode, int size) { 
    if (isBranch || !strcmp(opcode, "ret") || !strncmp(opcode, "set", 3)
         || !strcmp(opcode, "bswap"))
        return "";	
    if (opcode[0] == 'f') {
	if (opcode[1] == 'i') { 
	    switch(size) { 
	        case 64: return "ll";
	        case 32: return "l";
	        default: return mprintf("@@@@%d@@@@", size);
	    } 
	} else { 
	    if (   !strcmp(opcode, "fnstsw") || !strcmp(opcode, "fstsw") 
		|| !strcmp(opcode, "fnstcw") || !strcmp(opcode, "fldcw")
		|| !strcmp(opcode, "fstcw") 
	        || !strcmp(opcode, "frstor"))
		return "";		 
	    switch(size) { 
	        case 80: return "t";
	        case 64: return "l";
	        case 32: return "s";
	        case -1:  return "";
	        default: return mprintf("@@@@%d@@@@", size);
	    }
	}
    } else { 
	switch(size) { 
            case 32: return "l"; 
            case 16: return "h";
            case 8:  return "b";
            case 0:  if (!strcmp(opcode, "push")) 
		         return  "l";
	    default: return mprintf("@@@@%d@@@@", size);
	}
    } 
}

char *
getSize2(char *opcode, int src, int dst) { 
    int result;
    if (dst <= 0) {
	result = src;
    } else if (src <= 0) { 
	result = dst;
    } else if (dst == src) { 
	result = dst;
    } else if (opcode[0] == 's' && (opcode[1] == 'a' || opcode[1] == 'h') && 
	       (opcode[2] == 'l' || opcode[2] == 'r') && opcode[3] == '\0') {
	result = dst;
    } else if (opcode[0] == 'r' && (opcode[1] == 'l' || opcode[1] == 'c') && 
	       (opcode[2] == 'l' || opcode[2] == 'r') && opcode[3] == '\0') {
	result = dst;
    } else if (!strcmp(opcode, "movs") || !strcmp(opcode, "movz")) { 
	if (src == 8 && dst == 16) return "bw";
	if (src == 8 && (dst == 32 || dst == 0)) return "bl";
	if (src == 16 && (dst == 32 || dst == 0)) return "wl";
	return mprintf("@@@@%d/%d@@@@", src, dst);
    } else { 
	return mprintf("@@@@%d/%d@@@@", src, dst);
    }
    if (opcode[0] == 'f') {
	return (result == -1) ? "" : mprintf("@@@@%d@@@@", result);
    } else { 
	switch(result) { 
            case 32: return "l"; 
            case 16: return "w";
            case 8: return "b";
	    default: return mprintf("@@@@%d@@@@", result);
	}
    }
}

char *getSize3(char *opcode, int shift, int src, int dst) { 
    if (!strcmp(opcode, "shrd") || !strcmp(opcode, "shld")) { 
	return getSize2(opcode, dst, dst);
    } else { 
	return mprintf("@@@@%d/%d/%d@@@@", shift, src, dst);
    }
}

void yyerror(const char *message) { 
    printf("@@@@%s@@@@\n", message);
    error = TRUE;
}

/* This function is like sprintf(), except that it always mallocs a new buffer
 * into which to put the result.
 */

/*PRINTFLIKE1*/
char* mprintf(char *format, ...) {
    char outbuf[100];
    va_list args;
    va_start(args, format);
        vsprintf(outbuf, format, args);
    va_end(args);
    return strdup(outbuf);
}

extern int debugLexer(void);
extern int yyparse(void);

main(int argc, char **argv) {
    extern int yydebug;
    char *progname = argv[0];
    char **argp;
    bool_t lexOnly = FALSE;
    for (argp = argv + 1; *argp != 0; argp++) { 
        if (!strcmp(*argp, "-gcc")) { 
	    gcc = TRUE;
#ifdef DEBUG
	} else if (!strcmp(*argp, "-debug")) { 
	    yydebug = 1;
	} else if (!strcmp(*argp, "-lex")) { 
	    lexOnly = TRUE;
#endif
	} else { 
	    printf("%s [-gcc] [-debug] [-lex] < input > output\n", progname);
	    exit(1);
	}
    } 
    if (lexOnly) { 
#ifdef DEBUG
        debugLexer();
	return 0;
#else 
	printf("Not compiled -DEBUG.  Lexer not available.\n");
#endif
    } else { 
        error = FALSE;
	yyparse();
	return(error ? EXIT_FAILURE : EXIT_SUCCESS);
    }
}

struct ID { 
    char *name;
    int ivalue;
};


#include <search.h>

int IDcompare(const void *a, const void *b) { 
    return strcmp(((struct ID *)a)->name, ((struct ID *)b)->name);
} 

static void *root = NULL;

void assignValue(char *name, long value) { 
    struct ID *node = (struct ID *)malloc(sizeof(*node));
    struct ID *result;
    node->name = name;
    result = *(struct ID **)tsearch(node, &root, IDcompare);
    if (result != node) { 
	free(node);
    }
    result->ivalue = value;
}


bool_t hasValue(char *name, long *value) { 
    struct ID *node = (struct ID *)malloc(sizeof(*node));
    struct ID *result, **resultp;
    node->name = name;
    resultp = (struct ID **)tfind(node, &root, IDcompare);

    if (resultp == NULL) { 
         free(node);
	 return FALSE;
    } else { 
         result = *resultp;
         *value = result->ivalue;
	 return TRUE;
    }
}


struct node {            /* these are stored in the table */
    char *string;
    char *rewrite;
};

static struct node table[] = {     /* table to be searched */
    {"cbw",    "cbtw" },   /* AL -> AX */
    {"cdq",    "cltd" },   /* AX -> AX:DX */
    {"cwd",    "cwtd" },   /* AX -> EAX */
    {"cwde",   "cwtl" },   /* EAX -> EAX:EDX */
    {"movsx",  "movs" },
    {"movzx",  "movz" },
    {"popf",   "popfh"},
    {"popfd",  "popfl"},
    {"pushf",  "pushfh"},
    {"pushfd", "pushfl"},
};

char *rewriteOpcode(char *name) { 
    struct node node, *node_ptr;
    node.string = name;
    node_ptr = bsearch( &node, table, sizeof(table)/sizeof(struct node),
			sizeof(struct node), IDcompare);
    if (node_ptr != NULL) {
        return node_ptr->rewrite;
    } else { 
        return name;
    }
}
