@echo off

rem *
rem * Batch file to build InstallShield executable for JDK
rem *
rem * Usage: buildjdk [build-dir] [image-dir] [jre-image-dir] [dist-dir]
rem *
rem * Where
rem *    build-dir 	is <tree>\build\win32
rem *    image-dir      is a directory with an image of the JDK distribution
rem *    jre-image-dir  is a directory with an image of the JRE distribution
rem *    dist-dir       is a tmp directory for ishield to store files.
rem *
rem * Other environment settings:
rem *   ISHIELD is location of InstallShield3 directory
rem *	    (default C:\Program\InstallShield\InstallShield3)
rem *

setlocal

if cmdextversion 1 goto begin
echo "Must be running cmd.exe with command extensions enabled"
goto end

:begin

set BUILDDIR=%1
set IMAGE=%2
set JREIMAGE=%3
set DIST=%4

if /"%ISHIELD%"/ == /""/ set ISHIELD=C:\Program Files\InstallShield\InstallShield3
if /%BUILDDIR%/ == // set BUILDDIR=..
if /%IMAGE%/ == //    set IMAGE=..\jdk-image-win32
if /%DIST%/ == //     set DIST=c:\temp\jdk-ishield
if /%JREIMAGE%/ == // set JREIMAGE=..\jre-image-win32

: Do not use compression with ICOMP.EXE and let zip do all the work instead
set OPT=-sn

set PATH=%ISHIELD%\program;%ISHIELD%\utility\exebldr;%IBIN%;%PATH%

echo BUILDDIR is %BUILDDIR%
echo IMAGE    is %IMAGE%
echo DIST     is %DIST%
echo JREIMAGE is %JREIMAGE%
echo.

echo Removing original distribution...
echo.
del/f/q %BUILDDIR%\jdk12.exe
rmdir/s/q %DIST% 2>nul
mkdir %DIST%

echo Copying InstallShield support files...
echo.
copy "%ISHIELD%\program\setup.exe" %DIST%
copy "%ISHIELD%\program\_setup.dll" %DIST%
copy "%ISHIELD%\program\_setup.lib" %DIST%
copy "%ISHIELD%\program\_inst32i.ex_" %DIST%
copy "%ISHIELD%\program\_isdel.exe" %DIST%
copy "%ISHIELD%\program\uninst.exe" %DIST%

echo Adding LICENSE files for JDK and JRE...
echo.
icomp %IMAGE%\LICENSE %DIST%\_setup.lib %OPT%
icomp %JREIMAGE%\LICENSE %DIST%\_setup.lib jre %OPT%

echo Compiling setupjdk.rul...
echo.
compile setupjdk.rul
move setupjdk.ins %DIST%\setup.ins

echo Compiling setupjre.rul...
compile setupjre.rul
move setupjre.ins %DIST%\setupjre.ins

echo Adding LICENSE and LICENSE.JRE to setup.lib...
echo.
icomp %IMAGE%\LICENSE %DIST%\_setup.lib %OPT%
del/f/q LICENSE.JRE
copy %JREIMAGE%\LICENSE LICENSE.JRE
icomp LICENSE.JRE %DIST%\_setup.lib %OPT%
del/f/q LICENSE.JRE

echo Updating setup.lib...
echo.
: Update _setup.lib with patched version of _isres.dll
icomp "%BUILDDIR%\ishield\_isres.dll" %DIST%\_setup.lib -r
icomp "%BUILDDIR%\ishield\_isres.dll" %DIST%\_setup.lib %OPT%

echo Adding wait.bmp...
icomp wait.bmp %DIST%\_setup.lib %OPT%

echo Creating jdk.z...
echo.
icomp %IMAGE%\bin\*.* %DIST%\jdk.z bin -i %OPT%
icomp %IMAGE%\lib\*.* %DIST%\jdk.z lib -i %OPT%

echo Creating jre.z...
echo.
icomp %IMAGE%\jre\*.*      %DIST%\jre.z jre -i %OPT%
icomp %JREIMAGE%\LICENSE   %DIST%\jre.z jre %OPT%
icomp %JREIMAGE%\README    %DIST%\jre.z jre %OPT%
icomp %JREIMAGE%\CHANGES   %DIST%\jre.z jre %OPT%
icomp %JREIMAGE%\COPYRIGHT %DIST%\jre.z jre %OPT%

echo Creating docs.z...
echo.
icomp %IMAGE%\LICENSE     %DIST%\docs.z %OPT%
icomp %IMAGE%\COPYRIGHT   %DIST%\docs.z %OPT%
icomp %IMAGE%\README      %DIST%\docs.z %OPT%
icomp %IMAGE%\readme.html %DIST%\docs.z %OPT%
icomp %IMAGE%\docs\*.*    %DIST%\docs.z docs -i %OPT%

echo Creating redist.z...
echo.
icomp %BUILDDIR%\redist\msvcrt.dll %DIST%\redist.z -i %OPT%
icomp %BUILDDIR%\redist\ActPanel.dll %DIST%\redist.z -i %OPT%

echo Creating lib.z...
echo.
icomp %IMAGE%\include\*.* %DIST%\lib.z include -i %OPT%

echo Creating lib-old.z...
echo.
icomp %IMAGE%\include-old\*.* %DIST%\lib-old.z include-old -i %OPT%

echo Creating demo.z...
echo.
icomp %IMAGE%\demo\*.* %DIST%\demo.z demo -i %OPT%

echo Creating src.z...
echo.
icomp %IMAGE%\src.jar %DIST%\src.z -i %OPT%

echo Copying bitmap file...
echo.
copy setup.bmp %DIST%

echo Creating self-installing executable...
echo.

echo [EXE Builder] > %DIST%\exebuild.ini
echo SrcDir=%DIST% >> %DIST%\exebuild.ini
echo DestDir=%DIST% >> %DIST%\exebuild.ini
echo CmdLine=setup.exe >> %DIST%\exebuild.ini
echo Title=Java(tm) Development Kit 1.2 >> %DIST%\exebuild.ini
echo Compress=1 >> %DIST%\exebuild.ini
echo 32Bit=1 >> %DIST%\exebuild.ini

echo Generating packing list...
echo.
echo 1; >setup.lst
echo %DIST%\jdk.z >>setup.lst
echo %DIST%\jre.z >>setup.lst
echo %DIST%\docs.z >>setup.lst
echo %DIST%\lib.z >>setup.lst
echo %DIST%\lib-old.z >>setup.lst
echo %DIST%\demo.z >>setup.lst
echo %DIST%\src.z >>setup.lst
echo %DIST%\redist.z >>setup.lst
packlist setup.lst
move setup.pkg %DIST%
rem del /f setup.lst
start /wait exebuild.exe %DIST%\exebuild.ini
move %DIST%\setupex.exe %BUILDDIR%\jdk12.exe

echo Done
echo.
echo Cleaning up...
echo

rem del /f %DIST%\exebuild.ini

:end
endlocal
