@echo off

rem *
rem * Batch file to build InstallShield executable for JDK
rem *
rem * Usage: buildjre [build-dir] [jre-image-dir] [dist-dir]
rem *
rem * Where
rem *    build-dir      is <tree>\build\win32
rem *    jre-image-dir  is a directory with an image of the JRE distribution
rem *    dist-dir       is a tmp directory for ishield to store files.
rem *
rem * Other environment settings:
rem *   ISHIELD is location of InstallShield3 directory
rem *	    (default C:\Program\InstallShield\InstallShield3)
rem *

setlocal

if cmdextversion 1 goto begin
echo "Must be running cmd.exe with command extensions enabled"
goto end

:begin

set BUILDDIR=%1
set IMAGE=%2
set DIST=%3

if /"%ISHIELD%"/ == /""/ set ISHIELD=C:\Program Files\InstallShield\InstallShield3
if /%BUILDDIR%/ == // set BUILDDIR=..
if /%IMAGE%/ == //    set IMAGE=..\jre-image-win32
if /%DIST%/ == //     set DIST=c:\temp\jre-ishield

: Do not use compression with ICOMP.EXE and let zip do all the work instead
set OPT=-sn

set PATH=%ISHIELD%\program;%ISHIELD%\utility\exebldr;%IBIN%;%PATH%

echo BUILDDIR is %BUILDDIR%
echo IMAGE    is %IMAGE%
echo DIST     is %DIST%
echo.

echo Removing original distribution...
echo.
del/f/q %BUILDDIR%\jre12.exe
del/f/q %BUILDDIR%\jre12i.exe
rmdir/s/q %DIST% 2>nul
mkdir %DIST%

echo Copying InstallShield support files...
echo.
copy "%ISHIELD%\program\setup.exe" %DIST%
copy "%ISHIELD%\program\_setup.dll" %DIST%
copy "%ISHIELD%\program\_setup.lib" %DIST%
copy "%ISHIELD%\program\_inst32i.ex_" %DIST%
copy "%ISHIELD%\program\_isdel.exe" %DIST%
copy "%ISHIELD%\program\uninst.exe" %DIST%

echo Compiling setup.rul...
echo.
compile setupjre.rul
move setupjre.ins %DIST%\setup.ins

echo Adding LICENSE file into setup.lib...
echo.
del/f/q LICENSE.JRE
copy %IMAGE%\LICENSE LICENSE.JRE
icomp LICENSE.JRE %DIST%\_setup.lib %OPT%
del/f/q LICENSE.JRE

echo Updating setup.lib...
echo.
: Update _setup.lib with patched version of _isres.dll
icomp _isres.dll %BUILDDIR%\_setup.lib -r
icomp _isres.dll %BUILDDIR%\_setup.lib %OPT%

echo Creating jre.z...
echo.
icomp %IMAGE%\*.* %DIST%\jre.z jre %OPT%
icomp %IMAGE%\bin\*.* %DIST%\jre.z jre\bin -i %OPT%
icomp %IMAGE%\lib\*.* %DIST%\jre.z jre\lib -i %OPT% 

echo Creating redist.z...
icomp %BUILDDIR%\redist\msvcrt.dll %DIST%\redist.z -i %OPT%
icomp %BUILDDIR%\redist\ActPanel.dll %DIST%\redist.z -i %OPT%

echo Copying bitmap file...
echo.
copy setup.bmp %DIST%

echo Creating self-installing executables...
echo.

echo [EXE Builder] >exebuild.ini
echo SrcDir=%DIST% >>exebuild.ini
echo DestDir=%DIST% >>exebuild.ini
echo CmdLine=setup.exe >>exebuild.ini
echo Title=Java(tm) Runtime Environment 1.2 >>exebuild.ini
echo Compress=1 >>exebuild.ini
echo 32Bit=1 >>exebuild.ini

echo Creating %BUILDDIR%\jre12i.exe ...
echo.
echo 1; >setup.lst
echo %DIST%\jre.z >>setup.lst
echo %DIST%\redist.z >>setup.lst
packlist setup.lst
move setup.pkg %DIST%
del /f setup.lst 2>nul
start /wait exebuild.exe .\exebuild.ini
move %DIST%\setupex.exe %BUILDDIR%\jre12i.exe

echo Creating %BUILDDIR%\jre12.exe ...
echo.
icomp jre\lib\i18n.jar %DIST%\jre.z -r
icomp jre\lib\ext\iiimp.jar %DIST%\jre.z -r
echo 1; >setup.lst
echo %DIST%\jre.z >>setup.lst
echo %DIST%\redist.z >>setup.lst
packlist setup.lst
del /f %DIST%\setup.pkg
move setup.pkg %DIST%
del /f setup.lst 2>nul
start /wait exebuild.exe .\exebuild.ini
move %DIST%\setupex.exe %BUILDDIR%\jre12.exe

echo Cleaning up...
echo.
del /f exebuild.ini

:end
endlocal
