/*
 * @(#)includes.c	1.15 98/06/30
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Generate m4 definitions for the IA fast interpreter.
 */

#include <sys/types.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <assert.h>
#include <stddef.h>

#include "bool.h"
#include "memory.h"
#include "opcodes.h"
#include "oobj.h"			     
#include "interpreter.h"
#include "signature.h"
#include "typecodes.h"
#include "tree.h"
#include "threads.h"
#include "monitor.h"		    
#include "exceptions.h"
#include "javaString.h"
#include "path.h"
#include "byteorder.h"
#include "jmath.h"
#include "vmprofiler.h"

#define FMT1(s1,f1,m)  fprintf(f,""#m" EQU %d\n", offsetof(s1,f1));

int
main (int argc, char* argv[])
{
  /* Build interesting constants for inclusion by fastjava.m4 */
  FILE *f, *g;
  char path[2048], *p;
  int i;

  if (argc != 2) { 
      fprintf (stderr, "say %s <path>\n", argv[0]);
      exit (1);
  }

  strcpy (path, argv[1]);
  p = path + strlen(path);

  if (p[-1] != '/' && p[-1] != '\\') { 
      *p++ = '/';
      *p = '\0';
  }

  strcpy (p, "fjconst.inc");
  if ((f = fopen(path, "w")) == 0) { 
      fprintf (stderr, "%s: cannot open %s for write\n", argv[0], path);
      exit (1);
  }

  strcpy (p, "fjtables.inc");
  if ((g = fopen(path, "w")) == 0) { 
      fprintf (stderr, "%s: cannot open %s for write\n", argv[0], path);
      fclose(f);
      exit (1);
  }

  /* Various Constants */
  fprintf(f,"\n\n; Various Constants - machine generated by %s\n\n", argv[0]);
  fprintf(f,"JVMPI_EVENT_METHOD_ENTRY_ON		EQU	%d\n",
	  JVMPI_EVENT_METHOD_ENTRY_ON);
  fprintf(f,"JVMPI_EVENT_METHOD_EXIT_ON		EQU	%d\n",
	  JVMPI_EVENT_METHOD_EXIT_ON);
  fprintf(f,"METHOD_FLAG_BITS		EQU	%d\n",METHOD_FLAG_BITS);
  fprintf(f,"FJ_FLAG_MASK			EQU	%d\n",FLAG_MASK);
  fprintf(f,"FJ_T_NORMAL_OBJECT		EQU	%d\n",T_NORMAL_OBJECT);
  fprintf(f,"FJ_ACC_STATIC			EQU	%d\n",ACC_STATIC);
  fprintf(f,"FJ_ACC_PUBLIC			EQU	%d\n",ACC_PUBLIC);
  fprintf(f,"SIGNATURE_DOUBLE		EQU	%d\n",SIGNATURE_DOUBLE);
  fprintf(f,"SIGNATURE_LONG		EQU	%d\n",SIGNATURE_LONG);

  fprintf(f,"ClassMask EQU %d\n",
         (1<<CONSTANT_Class));

  fprintf(f,"WordMask  EQU %d\n",
         (1<<CONSTANT_Integer)|(1<<CONSTANT_Float)|(1<<CONSTANT_String));

  fprintf(f,"LongMask  EQU %d\n",
         (1<<CONSTANT_Double)|(1<<CONSTANT_Long));

  /* ClassClass structure (oobj.h) */
  fprintf(f,"\n\n; ClassClass structure (oobj.h) \n");

  FMT1 (Classjava_lang_Class, instance_size, 	ucbInstanceSize);
  FMT1 (Classjava_lang_Class, superclass, 	ucbSuperclass)	
  FMT1 (Classjava_lang_Class, constantpool, 	ucbConstantPool)
  FMT1 (Classjava_lang_Class, methodtable, 	ucbMethodTable)
  FMT1 (Classjava_lang_Class, imethodtable,	ucbIntfMethodTable)

			 
  /* ExecEnv structure (interpreter.h) */
  fprintf(f,"\n\n; ExecEnv structure (interpreter.h) \n");
  FMT1 (ExecEnv, initial_stack,			ee_inital_stack)
  FMT1 (ExecEnv, current_frame,			ee_current_frame)
  FMT1 (ExecEnv, thread,			ee_thread)
  FMT1 (ExecEnv, exceptionKind,			ee_exceptionKind)
  FMT1 (ExecEnv, exception,			ee_exception)
  FMT1 (ExecEnv, sys_thr,			ee_sys_thr)


  /* JavaFrame structure (interpreter.h) */
  fprintf(f,"\n\n; JavaFrame structure (interpreter.h) \n");
  FMT1 (JavaFrame, constant_pool,	 	javaframe_constant_pool)
  FMT1 (JavaFrame, returnpc, 			javaframe_returnpc)
  FMT1 (JavaFrame, optop,				javaframe_optop)
  FMT1 (JavaFrame, vars,				javaframe_vars)
  FMT1 (JavaFrame, prev,				javaframe_prev)
  FMT1 (JavaFrame, javastack,			javaframe_javastack)
  FMT1 (JavaFrame, lastpc,				javaframe_lastpc)
  FMT1 (JavaFrame, current_method,		javaframe_current_method)
  FMT1 (JavaFrame, monitor,				javaframe_monitor)
  FMT1 (JavaFrame, profiler_info,		javaframe_profiler_info)
  FMT1 (JavaFrame, ostack,				javaframe_ostack)


  /* methodblock structure (oobj.h) */
  fprintf(f,"\n\n; methodblock structure (oobj.h) \n");
  FMT1 (struct methodblock, fb, 	mb_fb)
  FMT1 (struct methodblock, code, 	mb_code)
  FMT1 (struct methodblock, invoker, 	mb_invoker)
  FMT1 (struct methodblock, args_size,	mb_args_size)
  FMT1 (struct methodblock, maxstack,	mb_maxstack)
  FMT1 (struct methodblock, nlocals, 	mb_nlocals)
										 

  /* fieldblock structure (oobj.h) */
  fprintf(f,"\n\n; fieldblock structure (oobj.h) \n");
  FMT1 (struct fieldblock, clazz,		fb_clazz)
  FMT1 (struct fieldblock, signature,		fb_signature)
  FMT1 (struct fieldblock, name,		fb_name)
  FMT1 (struct fieldblock, access,		fb_access)
  FMT1 (struct fieldblock, u,			fb_u)
  FMT1 (struct fieldblock, u.static_address,	fb_static_address)
  FMT1 (struct fieldblock, u.static_value,	fb_static_value)
  FMT1 (struct fieldblock, u.offset,		fb_offset)


  /* methodtable structure (oobj.h) */
  fprintf(f,"\n\n; methodtable structure (oobj.h) \n");
  FMT1 (struct methodtable, classdescriptor,	methodtable_classdescriptor)
  FMT1 (struct methodtable, methods,			methodtable_methods)


  /* imethodtable structure (oobj.h) */
  fprintf(f,"\n\n; imethodtable structure (oobj.h) \n");
  FMT1 (struct imethodtable, icount,		imethodtable_icount)
  FMT1 (struct imethodtable, itable,		imethodtable_itable)



  /* JavaStack structure (interpreter.h) */
  fprintf(f,"\n\n; JavaStack structure (interpreter.h) \n");
  FMT1 (JavaStack, execenv, 			javastack_execenv)
  FMT1 (JavaStack, prev, 				javastack_prev)
  FMT1 (JavaStack, next, 				javastack_next)
  FMT1 (JavaStack, end_data, 			javastack_end_data)
  FMT1 (JavaStack, stack_so_far,		javastack_stack_so_far)
  FMT1 (JavaStack, data,	 			javastack_data)

  /* HPI_ThreadInterface structure (hpi.h) */
  fprintf(f,"\n\n; HPI_ThreadInterface structure (hpi.h) \n");
  FMT1 (HPI_ThreadInterface, ThreadCPUTime,    hpi_thread_cputime)
  FMT1 (HPI_ThreadInterface, MonitorExit,      hpi_thread_monitorexit)
  FMT1 (HPI_ThreadInterface, ThreadCheckStack, hpi_thread_checkstack)

  /* java_lang_Object structure (oobj.h) */
  fprintf(f,"\n\n; java_lang_Object structure (oobj.h) \n");
  FMT1 (Hjava_lang_Object, obj,			o_obj)
  FMT1 (Hjava_lang_Object, methods,		o_methods)
  FMT1 (ClassArrayOfObject, body,		array_body)


  fprintf(f,"\n; opcodes\n");
  fprintf(g,"; jump tables - machine generated by %s\n", argv[0]);

  for (i = 0; i < 256; i++) {
      if (strcmp (opnames[i], "??")) {
	  fprintf(f,"opc_%s equ %d\n", opnames[i], i);
	  fprintf(g, "OPCODE(%s)\n", opnames[i]);
      } else { 
	  fprintf(g, "OPCODE(%s)\n", "unimplemented");	
      }
  }
  fclose(f);
  fclose(g);
  exit(0);				
}
