# 
# @(#)jdkimage.sh	1.24 98/11/06
#
# Make "release images" of JDK. A release image is what an unpacked
# installation looks like.
#

#
# Die on error. Echo what is going on.
#
set -e -x

if [ ! -d jre-image-win32 ]; then
    echo "ERROR: to make jdk-image-win32, you must first make a jre-image-win32"
    exit 1
fi

#
# Must be absolute path names.
#
BUILD=`pwd`
TOPDIR=$BUILD/../..
SHARE_SRC=$TOPDIR/src/share
PLATFORM_SRC=$TOPDIR/src/win32

if [ ! -f $BUILD/java/jvm/Makefile ]; then
    echo "ERROR: this script has not been called from a correct place."
    exit 1
fi

IMAGE=$BUILD/jdk-image-win32
IMAGE_g=$BUILD/jdk-debug-image-win32
JAR=$BUILD/bin/jar

#
# Cleanup.  Stuff below assumes fresh start.
#
rm -rf $IMAGE
mkdir -p $IMAGE

#
# jre/.  For now put agent.dll in JDK's private JRE's bindir
#
cp -r $BUILD/jre-image-win32 $IMAGE/jre
cp $BUILD/bin/agent.dll $IMAGE/jre/bin
(cd $IMAGE/jre; rm -f CHANGES COPYRIGHT LICENSE README)

#
# lib/
#
mkdir -p $IMAGE/lib
cp $BUILD/lib/jvm.lib $IMAGE/lib

#
# bin/
#
mkdir -p $IMAGE/bin
cp `echo $BUILD/bin/*.exe` $IMAGE/bin

#
# include/ include-old/
#
cp -r $BUILD/include $BUILD/include-old $IMAGE

#
# tools.jar
#
TOOLS='
sun/tools
sun/security/tools
sun/rmi/rmic
sun/rmi/rmid
sun/applet
com/sun/tools
com/sun/javadoc'
(cd $BUILD/classes; $JAR -cv0f $IMAGE/lib/tools.jar $TOOLS)

#
# src/
#
SOURCES='
java/applet
java/awt
java/beans
java/io
java/lang
java/math
java/net
java/rmi
java/security
java/sql
java/text
java/util
org/omg
javax/accessibility
javax/swing
com/sun/java/swing/plaf/windows
com/sun/java/swing/plaf/motif
com/sun/image/codec/jpeg
sunw
sun/tools/ttydebug
'
mkdir -p $IMAGE/src
(cd $SHARE_SRC/classes; \
 tar cf - `find $SOURCES -type d -name SCCS -prune -o -type f -name '*.java' -print`) | \
 (cd $IMAGE/src; tar xvf -)
# Use the munged swing source, tar fails in this case because the files
# already exit, so we use cp.
cp -rfp $BUILD/javax/swing/1.2_src/javax $IMAGE/src
# This should not be part of src.jar
rm -rf $IMAGE/src/javax/swing/pending
# LabelView2D is munged to LabelView in 1.2, so don't include LabelView2D.java
rm -f $IMAGE/src/javax/swing/text/LabelView2D.java
mkdir -p $IMAGE/src/launcher
cp $SHARE_SRC/bin/java.c $IMAGE/src/launcher
cp $SHARE_SRC/bin/java.h $IMAGE/src/launcher
cp $PLATFORM_SRC/bin/java_md.c $IMAGE/src/launcher
cp $PLATFORM_SRC/bin/java_md.h $IMAGE/src/launcher
(cd $IMAGE; $BUILD/bin/jar cvf0M src.jar src)
rm -fr $IMAGE/src

#
# demo/
#
cp -r $BUILD/demo $IMAGE/demo

#
# Docs.
#
(cd $SHARE_SRC/doc/jdk; tar cf - 				\
    `find . -type d -name SCCS -prune -o -type f -print`) |	\
    (cd $IMAGE; tar xvf -)
# User visible text files must not have ^M's in them.
for i in $IMAGE/README $IMAGE/LICENSE $IMAGE/COPYRIGHT; do
    sed 's/$//g' $i > docfile.tmp
    mv -f docfile.tmp $i
    chmod 444 $i
done

#
# Nuke _g.
#
rm -f `find $IMAGE/bin $IMAGE/lib -type f | grep _g`

#
# JDK debug.
#
rm -fr $IMAGE_g
mkdir -p $IMAGE_g
mkdir -p $IMAGE_g/bin;     cp `echo $BUILD/bin/*_g.exe` $IMAGE_g/bin
mkdir -p $IMAGE_g/jre/bin; cp `echo $BUILD/bin/*_g.dll` $IMAGE_g/jre/bin
mkdir -p $IMAGE_g/jre/bin/classic
cp $BUILD/bin/classic/jvm_g.dll $IMAGE_g/jre/bin/classic
mkdir -p $IMAGE_g/lib;     cp $BUILD/lib/jvm_g.lib $IMAGE_g/lib
