/*
 * @(#)IIIMPDebug.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

class IIIMPDebug {
    static boolean debug = true;

    static boolean isDebug() {
	return debug;
    }

    static void on() {
	debug = true;
    }

    static void off() {
	debug = false;
    }

    static String toString(Packet packet, String str) {
	if (debug == false) {
	    return null;
	}
	StringBuffer sb = new StringBuffer();
	sb.append("/**********************************\n");
	sb.append("Direction: " + str + "\n");
	try {
	    String proto = protocols[packet.protocol];
	    if (proto != null) {
		sb.append("OpCode: " + proto + "\n");
	    } else {
		sb.append("OpCode: " + packet.protocol + "\n");
	    }
	} catch (Exception e) {
	    sb.append("OpCode: " + packet.protocol + "\n");
	}
	sb.append("Length: " + packet.len + "\n");
	sb.append("Content:\n");
	sb.append(toHexString(packet.buf, packet.len));
	sb.append("**********************************/");
	return sb.toString();
    }

    static void println(Packet packet, String direction) {
	System.out.println(toString(packet, direction));
    }

    static void println(byte []b) {
	println(b, b.length);
    }

    static void println(byte []b, int length) {
	System.out.println(toHexString(b, length));
    }

    static String toHexString(byte []b, int length) {
	if (debug == false) {
	    return null;
	}

	StringBuffer sb = new StringBuffer();

	// print x-th for (x + y * 10)
	sb.append("       0  1  2  3  4  5  6  7  8  9\n");

	int num_space = 4;

	for (int i = 0; i < length; i++) {

	    // print y-th for (x + y * 10)
	    if ((i % 10) == 0) {
		// skip the first one
		if (i != 0) {
		    sb.append("\n");
		}
		// print y-th
		int ith = i / 10;
		sb.append("  " + ith);

		// print spaces depending on how many x-th characters are
		String s = Integer.toString(ith);
		int n = num_space - s.length();
		for (int j = 0; j < n; j++) {
		    sb.append(" ");
		}
	    }

	    // print a hex value
	    String s = Integer.toHexString((int) (b[i] & 0x000000FF));
	    if (s.length() == 1) {
		sb.append("0");
		sb.append(s);
	    } else {
		sb.append(s);
	    }

	    sb.append(" ");
	}

	// flush them out
	sb.append("\n");

	return sb.toString();
    }

    static String protocols[] = {
	"",					// 0
	"IM_CONNECT",				// 1
	"IM_CONNECT_REPLY",			// 2 
	"IM_DISCONNECT",			// 3
	"IM_DISCONNECT_REPLY",			// 4
	"IM_REGISTER_TRIGGER_KEYS",		// 5
	"IM_TRIGGER_NOTIFY",			// 6
	"IM_TRIGGER_NOTIFY_REPLY",		// 7
	"IM_SET_IM_VALUES",			// 8
	"IM_SET_IM_VALUES_REPLY",		// 9
	"IM_GET_IM_VALUES",			// 10
	"IM_GET_IM_VALUES_REPLY",		// 11
	"IM_FORWARD_EVENT",			// 12
	"IM_FORWARD_EVENT_REPLY",		// 13
	"IM_COMMIT_STRING",			// 14
	"",					// 15
	"",					// 16
	"",					// 17
	"",					// 18
	"",					// 19
	"IM_CREATEIC",				// 20
	"IM_CREATEIC_REPLY",			// 21
	"IM_DESTROYIC",				// 22
	"IM_DESTROYIC_REPLY",			// 23
	"IM_SETICVALUES",			// 24
	"IM_SETICVALUES_REPLY",			// 25
	"IM_GETICVALUES",			// 26
	"IM_GETICVALUES_REPLY",			// 27
	"IM_SETICFOCUS",			// 28
	"IM_SETICFOCUS_REPLY",			// 29
	"IM_UNSETICFOCUS",			// 30
	"IM_UNSETICFOCUS_REPLY",		// 31
	"",					// 32
	"",					// 33
	"",					// 34
	"",					// 35
	"",					// 36
	"",					// 37
	"",					// 38
	"",					// 39
	"IM_PREEDIT_START",			// 40
	"IM_PREEDIT_START_REPLY",		// 41
	"IM_PREEDIT_DRAW",			// 42
	"IM_PREEDIT_DRAW_REPLY",		// 43
	"IM_PREEDIT_CARET",			// 44
	"IM_PREEDIT_CARET_REPLY",		// 45
	"IM_PREEDIT_DONE",			// 46
	"IM_PREEDIT_DONE_REPLY",		// 47
	"",					// 48
	"",					// 49
	"IM_STATUS_START",			// 50
	"IM_STATUS_START_REPLY",		// 51
	"IM_STATUS_DRAW",			// 52
	"IM_STATUS_DRAW_REPLY",			// 53
	"IM_STATUS_DONE",			// 54
	"IM_STATUS_DONE_REPLY",			// 55
	"",					// 56
	"",					// 57
	"",					// 58
	"",					// 59
	"",					// 60
	"",					// 61
	"",					// 62
	"",					// 63
	"",					// 64
	"",					// 65
	"",					// 66
	"",					// 67
	"",					// 68
	"",					// 69
	"IM_LOOKUP_CHOICE_START",		// 70
	"IM_LOOKUP_CHOICE_START_REPLY",		// 71
	"IM_LOOKUP_CHOICE_DRAW",		// 72
	"IM_LOOKUP_CHOICE_DRAW_REPLY",		// 73
	"IM_LOOKUP_CHOICE_DONE",		// 74
	"IM_LOOKUP_CHOICE_DONE_REPLY",		// 75
	"IM_LOOKUP_CHOICE_PROCESS",		// 76
	"IM_LOOKUP_CHOICE_PROCESS_REPLY",	// 77
	"",					// 78
	"",					// 79
	"",					// 80
	"",					// 81
	"",					// 82
	"",					// 83
	"",					// 84
	"",					// 85
	"",					// 86
	"",					// 87
	"",					// 88
	"",					// 89
	"IM_AUX_START",				// 90
	"IM_AUX_START_REPLY",			// 91
	"IM_AUX_DRAW",				// 92
	"IM_AUX_DRAW_REPLY",			// 93
	"IM_AUX_DONE",				// 94
	"IM_AUX_DONE_REPLY",			// 95
    };
}
