/*
 * @(#)IIIMPDispatchEvent.java	1.3 98/03/23
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.awt.event.KeyEvent;
import java.io.IOException;

abstract class IIIMPCallbackRec implements IIIMProtocol {
    int opcode = IM_NO_PROTOCOL;

    byte arg[] = null;

    abstract IIIMPEvent getEvent(InputMethod im,
				 InputContext ic) throws IOException;

    Packet getReplyPacket(InputMethod im,
			  InputContext ic) throws IOException {
	return getReply(im, ic, opcode);
    }

    Packet getReply(InputMethod im,
		    InputContext ic,
		    int opcode,
		    int code) throws IOException {
	PacketData ps = new PacketData(opcode);
        ps.write2(im.getID());
        ps.write2(ic.getID());
        ps.write4(code);
	return ps.toPacket();
    }

    Packet getReply(InputMethod im,
		    InputContext ic,
		    int opcode) throws IOException {
        PacketData ps = new PacketData(opcode);
        ps.write2(im.getID());
        ps.write2(ic.getID());
	return ps.toPacket();
    }
}

class ForwardEventCB extends IIIMPCallbackRec {
    static final int opcode = IM_FORWARD_EVENT_REPLY;
    static final int STRING = 0;
    static final int TEXT = 1;
    static final int KEY_EVENT = 2;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	ForwardEvent ret = null;
        Data d = new Data(arg, arg.length);
        int type = d.read4(); // string or text or key event
        switch (type) {
          case STRING:
            break;
          case TEXT:
            break;
          case KEY_EVENT:
            int len = d.read4();
            if (len > 0) {
                byte[] buf = new byte[len];
                d.read(buf, 0, len);
                KeyEventData kd = new KeyEventData(buf, len);
                IIIMPKeyEvent[] kev = kd.toKeyEvents();
		ret = new ForwardEvent(ic, IM_FORWARD_EVENT, kev);
                break;
            }
        }
	return ret;
    }
}

class TriggerNotifyCB extends IIIMPCallbackRec {
    static final int opcode = IM_TRIGGER_NOTIFY_REPLY;
    static final int ON = 0;
    static final int OFF = 1;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
        Data d = new Data(arg, arg.length);
        int on_off = d.read2(); // on or off
        d.read2(); // padding
	return new TriggerEvent(ic, IM_TRIGGER_NOTIFY, (on_off == ON));
    }
}

class CommitStringCB extends IIIMPCallbackRec {
    static final int STRING = 0;
    static final int TEXT = 1;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	CommitEvent ret = null;
        Data d = new Data(arg, arg.length);
        int type = d.read4();
        if (type == STRING) {
	    ret = new CommitEvent(ic, d.readString());
        } else if (type == TEXT) {
	    ret = new CommitEvent(ic, IIIMPInputMethodText.
				  toInputMethodText((Data) d));
        }
	return ret;
    }
}

class PreeditStartCB extends IIIMPCallbackRec {
    static final int opcode = IM_PREEDIT_START_REPLY;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	return new PreeditEvent(ic, IM_PREEDIT_START);
    }
    Packet getReplyPacket(InputMethod im, InputContext ic) throws IOException {
	if (ic.preeditBufferSize == 0) {
	    ic.preeditBufferSize = 1024;
	}
	return getReply(im, ic, opcode, ic.preeditBufferSize);
    }
}

class PreeditDoneCB extends IIIMPCallbackRec {
    static final int opcode = IM_PREEDIT_DONE_REPLY;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	return new PreeditEvent(ic, IM_PREEDIT_DONE);
    }
}

class PreeditDrawCB extends IIIMPCallbackRec {
    static final int opcode = IM_PREEDIT_DRAW_REPLY;
    static final int STRING = 0;
    static final int TEXT = 1;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
        Data d = new Data(arg, arg.length);
        int caret_position = d.read4();
        int first_changed = d.read4();
        int length_changed = d.read4();
        int type = d.read4();
	InputMethodText text = null;
        if (type == STRING) {
        } else if (type == TEXT) {
            text = IIIMPInputMethodText.toInputMethodText(d);
        }
	InputMethodPreeditText preedit = new InputMethodPreeditText();
	preedit.setPosition(caret_position);
	preedit.setFirstChangedPosition(first_changed);
	preedit.setTextChangedLength(length_changed);
	preedit.setText(text);
        return new PreeditEvent(ic, IM_PREEDIT_DRAW, preedit);
    }
}

class PreeditCaretCB extends IIIMPCallbackRec {
    static final int opcode = IM_PREEDIT_CARET_REPLY;

    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
        Data d = new Data(arg, arg.length);
        int position = d.read4();
	int direction = InputMethodCaret.CARET_FORWARD_CHAR;
        switch (d.read4()) {
        case 0:
            direction = InputMethodCaret.CARET_FORWARD_CHAR;
            break;
        case 1:
            direction = InputMethodCaret.CARET_BACKWARD_CHAR;
            break;
        case 2:
            direction = InputMethodCaret.CARET_FORWARD_WORD;
            break;
        case 3:
            direction = InputMethodCaret.CARET_BACKWARD_WORD;
            break;
        case 4:
            direction = InputMethodCaret.CARET_UP;
            break;
        case 5:
            direction = InputMethodCaret.CARET_DOWN;
            break;
        case 6:
            direction = InputMethodCaret.CARET_NEXT_LINE;
            break;
        case 7:
            direction = InputMethodCaret.CARET_PREVIOUS_LINE;
            break;
        case 8:
            direction = InputMethodCaret.CARET_LINE_START;
            break;
        case 9:
            direction = InputMethodCaret.CARET_LINE_END;
            break;
        case 10:
            direction = InputMethodCaret.CARET_ABSOLUTE_POSITION;
            break;
        case 11:
            direction = InputMethodCaret.CARET_DONT_CHANGE;
            break;
        default:
            break;
        }
        int style = InputMethodCaret.CARET_INVISIBLE;;
        switch (d.read4()) {
        case 0:
            style = InputMethodCaret.CARET_INVISIBLE;
            break;
        case 1:
            style = InputMethodCaret.CARET_PRIMARY;
            break;
        case 2:
            style = InputMethodCaret.CARET_SECONDARY;
            break;
        default:
            break;
        }
	InputMethodCaret caret = new InputMethodCaret(position,
						      direction,
						      style);
        return new PreeditEvent(ic, IM_PREEDIT_CARET, caret);
    }

    Packet getReplyPacket(InputMethod im, InputContext ic) throws IOException {
	return getReply(im, ic, opcode, ic.preeditCaretPosition);
    }
}

class StatusStartCB extends IIIMPCallbackRec {
    static final int opcode = IM_STATUS_START_REPLY;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	return new StatusEvent(ic, IM_STATUS_START);
    }
}

class StatusDoneCB extends IIIMPCallbackRec {
    static final int opcode = IM_STATUS_DONE_REPLY;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	return new StatusEvent(ic, IM_STATUS_DONE);
    }
}

class StatusDrawCB extends IIIMPCallbackRec {
    static final int opcode = IM_STATUS_DRAW_REPLY;
    static final int STRING = 0;
    static final int TEXT = 1;
    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	StatusEvent ret = null;
        Data d = new Data(arg, arg.length);
        int type = d.read4();
        if (type == STRING) {
	    ret = new StatusEvent(ic, IM_STATUS_DRAW, d.readString());
        } else if (type == TEXT) {
	    ret = new StatusEvent(ic, IM_STATUS_DRAW,
				  IIIMPInputMethodText.toInputMethodText(d));
        }
	return ret;
    }
}

class LookupStartCB extends IIIMPCallbackRec {
    static final int opcode = IM_LOOKUP_CHOICE_START_REPLY;
    static final int SERVER_DRIVEN = 1;
    static final int CLIENT_DRIVEN = 2;
    static final int HORIZONTAL = 0;
    static final int VERTICAL = 1;
    static final int SERVER_LABEL = 0;
    static final int CLIENT_LABEL = 1;

    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
        if (arg == null) {
	    InputMethodLookupStart s =
		new InputMethodLookupStart(
		    InputMethodLookupStart.EVENT_SERVER_DRIVEN,
		    0, 0, 0, 0, true);
            return new LookupEvent(ic, IM_LOOKUP_CHOICE_START, s);
        }

        Data d = new Data(arg, arg.length);

        int type = d.read2();
        if (type == SERVER_DRIVEN) {
            type = InputMethodLookupStart.EVENT_SERVER_DRIVEN;
        } else if (type == CLIENT_DRIVEN) {
            type = InputMethodLookupStart.EVENT_CLIENT_DRIVEN;
        }

        int size = d.read2();
        int rows = d.read2();
        int columns = d.read2();

        int dir = d.read2();
        if (dir == HORIZONTAL) {
            dir = InputMethodLookupStart.DIRECTION_HORIZONTAL;
        } else if (dir == VERTICAL) {
            dir = InputMethodLookupStart.DIRECTION_VERTICAL;
        }

        boolean labeled = (d.read2() == SERVER_LABEL) ? true : false;

	InputMethodLookupStart s = new InputMethodLookupStart(type,
							      dir,
							      size,
							      rows,
							      columns,
							      labeled);

        return new LookupEvent(ic, IM_LOOKUP_CHOICE_START, s);
    }
}

class LookupDoneCB extends IIIMPCallbackRec {
    static final int opcode = IM_LOOKUP_CHOICE_DONE_REPLY;
    IIIMPEvent getEvent(InputMethod im,
			     InputContext ic) throws IOException {
	LookupEvent ret = new LookupEvent(ic, IM_LOOKUP_CHOICE_DONE);
	return ret;
    }
}

class LookupDrawCB extends IIIMPCallbackRec {
    static final int opcode = IM_LOOKUP_CHOICE_DRAW_REPLY;
    static final int STRING = 0;
    static final int TEXT = 1;

    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	 // initialize...
	InputMethodLookupText[] ret_value = null;
        int ret_count = 0;
        InputMethodText ret_label = null;

        Data d = new Data(arg, arg.length);

        // indeces of candidates
        int ret_first_index = d.read4();
        int ret_last_index = d.read4();
        int ret_current_index = d.read4();

        // List of candidates
        InputMethodText[] candidates = null;
        int len = d.read4();

        byte[] buf;
        if (len > 0) {
            buf =new byte[len];
            d.read(buf, 0, len);
            candidates = IIIMPInputMethodText.
                toListOfInputMethodText(new Data(buf, len));
        }

        // List of labels
        InputMethodText[] labels = null;
        len = d.read4();
        if (len > 0) {
            buf =new byte[len];
            d.read(buf, 0, len);
            labels = IIIMPInputMethodText.
                toListOfInputMethodText(new Data(buf, len));
        }

        InputMethodLookupText[] can = null;
        if (candidates != null && candidates.length > 0) {
            can = new InputMethodLookupText[candidates.length];
            for (int i = 0; i < candidates.length; i++) {
                can[i] = new InputMethodLookupText();
                can[i].label = ((labels != null) ? labels[i] : null);
                can[i].value = candidates[i];
            }
            ret_value = can;
            ret_count = candidates.length;
        }

        // Label/Title
        ret_label = IIIMPInputMethodText.toInputMethodText(d);

	InputMethodLookupDraw draw =
	    new InputMethodLookupDraw(ret_value,
				      ret_count,
				      ret_first_index,
				      ret_last_index,
				      ret_current_index,
				      ret_label);

	return new LookupEvent(ic, IM_LOOKUP_CHOICE_DRAW, draw);
    }
}

class LookupProcessCB extends IIIMPCallbackRec {
    static final int opcode = IM_LOOKUP_CHOICE_PROCESS_REPLY;
    static final int INDEX = 0;
    static final int PAGE = 1;
    static final int NEXT_PAGE = 1;
    static final int PREV_PAGE = 2;
    static final int FIRST_PAGE = 3;
    static final int LAST_PAGE = 4;

    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	Data d = new Data(arg, arg.length);

        int type = d.read2();
        if (type == INDEX) {
            type = InputMethodLookupProcess.TYPE_INDEX;
        } else if (type == PAGE) {
            type = InputMethodLookupProcess.TYPE_PAGE;
        }

	InputMethodLookupProcess p = null;

        int val = d.read2();
        if (type == INDEX) {
	    p = new InputMethodLookupProcess(type, val);
        } else if (type == PAGE) {
	    int page = InputMethodLookupProcess.PAGE_FIRST;
            switch (val) {
            case NEXT_PAGE:
                page = InputMethodLookupProcess.PAGE_NEXT;
                break;
            case PREV_PAGE:
                page = InputMethodLookupProcess.PAGE_PREV;
                break;
            case FIRST_PAGE:
                page = InputMethodLookupProcess.PAGE_FIRST;
                break;
            case LAST_PAGE:
                page = InputMethodLookupProcess.PAGE_LAST;
                break;
            }
	    p = new InputMethodLookupProcess(type, page);
        }

	 return new LookupEvent(ic, IM_LOOKUP_CHOICE_PROCESS, p);
    }
}

abstract class AuxCB extends IIIMPCallbackRec {
    int index;
    String name;
    Packet getReply(InputMethod im, InputContext ic, int opcode)
        throws IOException {
        PacketData ps = new PacketData(opcode);
        ps.write2(im.getID());
        ps.write2(ic.getID());
        ps.write4(index);
        ps.writeString(name);
	return ps.toPacket();
    }
}

class AuxStartCB extends AuxCB {
    static final int opcode = IM_AUX_START_REPLY;

    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	AuxEvent ret = new AuxEvent(ic, IM_AUX_START);

        Data d = new Data(arg, arg.length);
        ret.index = index = d.read4();
        ret.name = name = d.readString();

	return ret;
    }
}

class AuxDoneCB extends AuxCB {
    static final int opcode = IM_AUX_DONE_REPLY;

    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	AuxEvent ret = new AuxEvent(ic, IM_AUX_DONE);

        Data d = new Data(arg, arg.length);
        ret.index = index = d.read4();
        ret.name = name = d.readString();

	return ret;
    }
}

class AuxDrawCB extends AuxCB {
    static final int opcode = IM_AUX_DRAW_REPLY;

    IIIMPEvent getEvent(InputMethod im, InputContext ic) throws IOException {
	AuxEvent ret = new AuxEvent(ic, IM_AUX_DRAW);

	StringData d = new StringData(arg, arg.length);
        ret.index = index = d.read4();
        ret.name = name = d.readString();

        int len = d.read4() / 4;
        ret.idata = new int[len];
        for (int i = 0; i < len; i++) {
            ret.idata[i] = d.read4();
        }

        ret.sdata = d.toStringArray2();

	return ret;
    }
}

class IIIMPDispatchEvent implements IIIMProtocol {

    static IIIMPClient client = IIIMPClient.getInstance();

    static IIIMPCallbackRec[] table = null;

    static void init() {
	table = new IIIMPCallbackRec[2<<7];
	table[IM_FORWARD_EVENT] = new ForwardEventCB();
	table[IM_TRIGGER_NOTIFY] = new TriggerNotifyCB();
	table[IM_COMMIT_STRING] = new CommitStringCB();
	table[IM_PREEDIT_START] = new PreeditStartCB();
	table[IM_PREEDIT_DRAW] = new PreeditDrawCB();
	table[IM_PREEDIT_CARET] = new PreeditCaretCB();
	table[IM_PREEDIT_DONE] = new PreeditDoneCB();
	table[IM_STATUS_START] = new StatusStartCB();
	table[IM_STATUS_DRAW] = new StatusDrawCB();
	table[IM_STATUS_DONE] = new StatusDoneCB();
	table[IM_LOOKUP_CHOICE_START] = new LookupStartCB();
	table[IM_LOOKUP_CHOICE_DRAW] = new LookupDrawCB();
	table[IM_LOOKUP_CHOICE_DONE] = new LookupDoneCB();
	table[IM_LOOKUP_CHOICE_PROCESS] = new LookupProcessCB();
	table[IM_AUX_START] = new AuxStartCB();
	table[IM_AUX_DRAW] = new AuxDrawCB();
	table[IM_AUX_DONE] = new AuxDoneCB();
    }

    synchronized static boolean processProtocol(int imID,
						int icID,
						int protocol,
						PacketData ps) 
	throws IOException {

        InputMethod im = InputMethod.getInputMethod(imID);
        if (im == null) {
            return false;
        }

        InputContext ic = im.getInputContext(icID);
        if (ic == null) {
            return false;
        }

	if (table == null) {
	    IIIMPDispatchEvent.init();
	}

	IIIMPCallbackRec cb = table[protocol];
	if (cb == null) {
	    return false;
	}

        int num = ps.available();
        byte arg[] = null;
        if (num > 0) {
            arg = new byte[num];
            ps.read(arg, 0, num);
        }

        if (num > 0) {
            cb.arg = new byte[num];
            System.arraycopy(arg, 0, cb.arg, 0, num);
        }

	// need to verify try 
        try {
	    IIIMPEvent e = cb.getEvent(im, ic);
	    ic.processEvent(e);
        } catch (Exception e) {
            e.printStackTrace();
        }

	Packet packet = cb.getReplyPacket(im, ic);
	client.send(packet);

        return true;
    }
}
