/*
 * @(#)IIIMPICValues.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.io.IOException;
import java.util.Locale;

/**
 * Only the locale attribute in IC values is currently supported.
 * If another attributes are added, then this IiimpICValues needs
 * to be updated to support the attributes.
 */
class IIIMPICValues {
    final int InputLanguage = 1;

    Locale locale;

    IIIMPICValues(Locale locale) {
	this.locale = locale;
    }

    IIIMPICValues(byte[] buf) throws IOException {
	Data d = new Data(buf, buf.length);

	while (d.available() > 0) {
	    int id = d.read2();

	    int len = d.read2(); // byte length of attr

	    byte[] attr = new byte[len]; // attribute array
	    d.read(attr, 0, len);

	    int pad = d.paddings(len); // number of pad
	    d.skipBytes(pad); // skip pad

	    /*
	     * Currently only InputLanguage is supported.
	     */
	    if (id == InputLanguage) {
		setInputLanguage(attr);
	    }

	    /*
	     * switch (id) {
	     * case InputLanguage:
	     * getInputLocale(attr);
	     * break;
	     * }
	     */
	}
    }

    byte[] getListOfICAttrId() throws IOException {
	Data d = new Data();

	// byte length of attr id
	int len = 1 * 2;
	d.write2(len);

	// InputLanguage
	d.write2(InputLanguage);

	// pad it
	d.pad();

	byte[] ret = new byte[d.count];
	System.arraycopy(d.buf, 0, ret, 0, d.count);

	return ret;
    }

    byte[] getListOfICAttr() throws IOException {
	byte[] ret = getInputLanguage();
	return ret;
    }

    void setInputLanguage(byte[] buf) throws IOException {
	Data d = new Data(buf, buf.length);

	// The locale names are defined in the IIIMP spec.
	String str = d.readString();
	locale = IIIMPInputLanguage.toLocale(str);
    }

    byte[] getInputLanguage() throws IOException {
	Data d = new Data();

	// attribute id
	d.write2(InputLanguage);

	// input langauge value.
	Data val = new Data();
	String lang = IIIMPInputLanguage.toString(locale);
	val.writeString(lang);
	
	// value length
	d.write2(val.count);
	// value
	d.writeBytes(val.buf, val.count);
	// pad it
	d.pad();

	byte[] ret = new byte[d.count];
	System.arraycopy(d.buf, 0, ret, 0, d.count);
	return ret;
    }
}
