/*
 * @(#)IIIMPInputMethodText.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.io.IOException;

import sun.awt.im.iiimp.InputMethodText;
import sun.awt.im.iiimp.InputMethodChar;

class IIIMPInputMethodText extends InputMethodText {

    IIIMPInputMethodText() {
	super();
	count = 0;
	value = new InputMethodChar[8];
    }

    void append(InputMethodChar c) {
	ensureCapacity(count + 1);
	value[count++] = c;
    }

    void ensureCapacity(int minimumNum) {
	int maxNum = value.length;
	if (minimumNum > maxNum) {
	    int newNum = (maxNum + 1) * 2;
	    if (minimumNum > newNum) {
		newNum = minimumNum;
	    }
	    InputMethodChar newValue[] = new InputMethodChar[newNum];
	    System.arraycopy(value, 0, newValue, 0, count);
	    value = newValue;
	}
    }

    static InputMethodText toInputMethodText(Data d) throws IOException {
	int len = d.read4();
	if (len == 0) {
	    return (InputMethodText) null;
	}
	return toInputMethodTextImpl(d, len);
    }

    static private InputMethodText toInputMethodTextImpl(
	Data d, int len) throws IOException {
	IIIMPInputMethodText im = new IIIMPInputMethodText();

	int upto = d.pos + len;
	while (d.pos < upto) {
	    IIIMPInputMethodChar ic = new IIIMPInputMethodChar();
	    ic.c = (char) d.read2();
	    int fb_len = d.read2();
	    for (int i = 0; i < fb_len; i += 8) {
		int id = d.read4();
		int val = d.read4();
		ic.addInputMethodFeedback(id, val);
	    }

	    im.append(ic);
	}

	return (InputMethodText) im;
    }

    static InputMethodText[] toListOfInputMethodText(Data d)
	throws IOException {
	final int thread = 26; // number of alphabets

	int num_ret = thread;
	InputMethodText[] ret = new InputMethodText[thread];

	int num = 0;
	for (;; num++) {
	    if (d.available() == 0) {
		break;
	    }
	    int len = d.read4();
	    if (len == 0) {
		break;
	    }
	    if (num == num_ret) {
		InputMethodText[] tmp =
		    new InputMethodText[num_ret + thread];
		System.arraycopy(ret, 0, tmp, 0, num_ret);
		ret = tmp;
		num_ret += thread;
	    }
	    ret[num] = toInputMethodTextImpl(d, len);

	    // annotation, disregard it for a while...
	    len = d.read4();
	    if (len > 0) {
		d.skipBytes(len);
	    }
	}

	if (num > 0) {
	    InputMethodText[] tmp = new InputMethodText[num];
	    System.arraycopy(ret, 0, tmp, 0, num);
	    return tmp;
	}

	return (InputMethodText[]) null;
    }
}
