/*
 * @(#)IIIMPInputStream.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.io.EOFException;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.IOException;

class IIIMPInputStream extends DataInputStream implements IIIMProtocol {

    static int packet_size = FOURBYTES_PACKET;

    IIIMPClient client;

    IIIMPInputStream(IIIMPClient client, InputStream in) {
	this(client, in, FOURBYTES_PACKET);
    }
    
    IIIMPInputStream(IIIMPClient client, InputStream in, int packet_size) {
	super(in);
	this.client = client;
	this.packet_size = packet_size;
    }

    public int read() throws IOException {
	try {
	    int c = in.read();
	    if (c == -1) {
		client.lostConnect();
	    }
	    return c;
	} catch (Exception e) {
	    client.lostConnect();
	    throw new IOException("Connection lost");
	}
    }

    int readOpCode() throws IOException {
	int c = read();
	int ret = ((c >>> 7) & 0xFF);

	if (ret == 0) {
	    packet_size = FOURBYTES_PACKET;
	} else if (ret == 1) {
	    packet_size = EIGHTBYTES_PACKET;
	}

	return (c & 0x7f);
    }

    int readPacketLength() throws IOException {
	int ch1 = 0x00;
	int ch2 = read();
	if (ch2 == -1) {
	    return 0;
	}
	int ch3 = read();
	int ch4 = read();

	if (packet_size == FOURBYTES_PACKET) {
	    if ((ch1 | ch2 | ch3 | ch4) < 0) {
		throw new EOFException();
	    }

	    return (((ch1 << 24) + (ch2 << 16) + (ch3 << 8) + (ch4 << 0))
		<< 2);
	} else if (packet_size == EIGHTBYTES_PACKET) {
	    // not supported
	    throw new IOException("8 byte length of packet");
	}

	return 0;
    }

    Packet readPacket(int opcode, int len) throws IOException {
	byte buf[] = null;
	if (len != 0) {
	    buf = new byte[len];
	    int ret = 0;
	    while ((len - ret) > 0) {
		ret += read(buf, ret, len - ret);
		if (ret == -1) {
		    throw new IIIMProtocolException(
			opcode + ": unexpected end of the stream");
		}
	    }
	}
	return new Packet(opcode, buf, len);
    }

    static int packetSize() {
	return packet_size;
    }
}
