/*
 * @(#)IIIMPURLConnection.java	1.3 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;
import java.net.MalformedURLException;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;
import sun.net.NetworkClient;

//public 
class IIIMPURLConnection extends URLConnection {

    NetworkClient nc = null;

    protected IIIMPURLConnection(URL url) {
	super(url);
	// This only creates an instance for NetworkClient.
	nc = new NetworkClient();
    }

    public void connect() throws IOException {
	if (connected == true) {
	    return;
	}
	try {
	    final URL url = this.url;
	    AccessController.doPrivileged(new PrivilegedExceptionAction() {
		public Object run() throws IOException, UnknownHostException {
		    nc.openServer(url.getHost(), url.getPort());
		    return null;
		}
	    });
	    connected = true;
	} catch (PrivilegedActionException e) {
	    throw (IOException)e.getException();
	}
    }

    public void disconnect() throws IOException {
	if (connected == false) {
	    return;
	}
	nc.closeServer();
    }

    public boolean isConnected() {
	return nc.serverIsOpen();
    }
    
    public InputStream getInputStream() throws IOException {
	if (connected == false) {
	    connect();
	}
	return nc.serverInput;
    }

    public OutputStream getOutputStream() throws IOException {
	if (connected == false) {
	    connect();
	}
	return nc.serverOutput;
    }
}
