/*
 * @(#)InputMethod.java	1.7 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.awt.event.KeyEvent;
import java.util.Locale;
import java.util.Vector;
import java.util.Properties;
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.security.AccessController;
import java.security.PrivilegedAction;

//public 
class InputMethod {

    private Vector ichList = new Vector();

    private Locale locale = null;

    private int id = -1;

    private Locale localeList[] = null;

    private IIIMPKeyEvent onKey[] = null;

    private IIIMPKeyEvent offKey[] = null;

    private boolean dynamic_event_flow = false;

    private boolean connected = false;

    private IIIMPClient client = IIIMPClient.getInstance();

    private static InputMethod im = null;

    InputMethod() {
	// adds this.
	InputMethod.addInputMethod(this);
    }

    public static InputMethod getInstance() {
	if (im == null) {
	    im = new InputMethod();
	}
	return im;
    }

    boolean open(Locale locale) {
	if (connected == false) {
	    try {
		openIM();
	    } catch (IOException e) {}
	}
	if (isLocaleSupported(locale)) {
	    this.locale = locale;
	    return true;
	}
	return false;
    }

    boolean reopen(Locale locale) {
	if (connected == false) {
	    try {
		openIM();
	    } catch (IOException e) {}
	}
	if (connected == true) {
	    if (isLocaleSupported(locale)) {
		this.locale = locale;
		return true;
	    }
	    return false;
	}
	return false;
    }

    /**
     * If this is already connected to the server, it simply returns true.
     * If not, it tries to connect the server.  If the server location
     * is neither specified nor a correct URL, then it throws
     * IOException.  If the server to be connected does not run,
     * it returns false.  Otherwise, it returns true.
     */
    //private
    boolean openIM() throws IOException {
	if (connected == true) {
	    return true;
	}

	IIIMPIMValues d = client.connectIM();
	if (d == null) {
	    return false;
	}

	// seems nothing wrong happened, so it is connected
	connected = true;

	// copy IiimpIMValues
	id = d.id;
	dynamic_event_flow = d.dynamic_event_flow;
	onKey = d.onKey;
	offKey = d.offKey;
	localeList = d.localeList;

	return true;
    }

    public void dispose() {
	try {
	    client.closeIM(id);
	    client = null;
	} catch (Exception e) {}
	connected = false;
	InputMethod.removeInputMethod(this);
    }

    public Locale[] getAvailableLocales() {
	if (localeList == null) {
	    return null;
	}
        Locale[] tmp = new Locale[localeList.length];
	System.arraycopy(localeList, 0, tmp, 0, localeList.length);
	return tmp;
    }

    protected void finalize() throws Throwable {
	dispose();
    }
    
    void addInputContextHandler(InputContext ich) {
	ichList.addElement(ich);
    }

    void removeInputContextHandler(InputContext ich) {
	ichList.removeElement(ich);
    }

    InputContext getInputContext(int id) {
	for (int i = 0; i < ichList.size(); i++) {
	    InputContext ic = (InputContext) ichList.elementAt(i);
	    if (ic.getID() == id) {
		return ic;
	    }
	}
	return (InputContext) null;
    }

    public InputContext[] getInputContext() {
	if (ichList.size() == 0) {
	    return null;
	}
	InputContext[] ich = new InputContext[ichList.size()];
	ichList.copyInto(ich);
	return ich;
    }

    int getID() {
	return id;
    }

    boolean isConnected() {
	return connected;
    }

    boolean isDynamicEventFlow() {
	return dynamic_event_flow;
    }

    boolean isConversionOnKey(KeyEvent e) {
	int mod = e.getModifiers();
	int code = e.getKeyCode();
	if (onKey != null) {
	    for (int i = 0; i < onKey.length; i++) {
		if ((mod & onKey[i].modifier) != 0) {
		    if (code == onKey[i].keycode) {
			return true;
		    }
		}
	    }
	}
	return false;
    }

    boolean isConversionOffKey(KeyEvent e) {
	int mod = e.getModifiers();
	int code = e.getKeyCode();
	if (offKey != null) {
	    for (int i = 0; i < offKey.length; i++) {
		if ((mod & offKey[i].modifier) != 0) {
		    if (code == offKey[i].keycode) {
			return true;
		    }
		}
	    }
	}
	return false;
    }

    boolean isLocaleSupported(Locale locale) {
        if (localeList != null && localeList.length > 0) {
            for (int i = 0; i < localeList.length; i++) {
                if (Util.localeEquals(locale, localeList[i]) == true) {
                    return true;
                }
            }
        }
        return false;
    }

    void lostConnect() {
	if (connected == false) {
	    return;
	}
	connected = false;
	InputContext[] ic = getInputContext();
	for (int i = 0; i < ic.length; i++) {
	    ic[i].lostConnect();
	}
    }

    private static InputMethod imList[] = null;

    static void addInputMethod(InputMethod im) {
	if (imList == null) {
	    imList = new InputMethod[1];
	    imList[0] = im;
	    return;
	}
	for (int i = 0; i < imList.length; i++) {
	    InputMethod list = imList[i];
	    if (list == im) {
		return;
	    }
	}
	InputMethod[] tmp = new InputMethod[imList.length+1];
	System.arraycopy(imList, 0, tmp, 0, imList.length);
	tmp[imList.length] = im;
	imList = tmp;
    }

    static void removeInputMethod(InputMethod im) {
	if (imList == null) {
	    return;
	}
	int i = 0;
	for (; i < imList.length; i++) {
	    InputMethod list = imList[i];
	    if (list == im) {
		break;
	    }
	}
	if (i == imList.length) {
	    return;
	}
	InputMethod[] tmp = new InputMethod[imList.length-1];
	int j = 0;
	for (i = 0; i < imList.length; i++) {
	    if (imList[i] != im) {
		tmp[j++] = imList[i];
	    }
	}
	imList = tmp;
    }

    static InputMethod getInputMethod(int id) {
	if (imList[0].id == id) {
	    return imList[0];
	}
	for (int i = 0; i < imList.length; i++) {
	    if (imList[i].id == id) {
		return imList[i];
	    }
	}
	return (InputMethod) null;
    }

    static InputMethod[] getInputMethod() {
	InputMethod[] tmp = new InputMethod[imList.length];
	System.arraycopy(imList, 0, tmp, 0, imList.length);
	return tmp;
    }

    // Properties
    private static Properties props = null;

    private static void initializeProps() {
	Properties defaultProps = new Properties();
	defaultProps.put("iiimp.status.style", "client");
	props = new Properties(defaultProps);

	AccessController.doPrivileged(new PrivilegedAction() {
	    public Object run() {
		String javaHome = System.getProperty("java.home");
		String userHome = System.getProperty("user.home");
		File f;
		InputStream in;

		try {
		    // ./lib/iiimp.properties at JDK/JRE home
		    f = new File(javaHome + File.separator +
				 "lib" + File.separator + "iiimp.properties");
		    if (f != null && f.canRead()) {
			in = new BufferedInputStream(new FileInputStream(f.getPath()));
			props.load(in);
			in.close();
		    }

		    // $HOME/.iiimp
		    f = new File(userHome + File.separator +
				 ".iiimp");
		    if (f != null && f.canRead()) {
			in = new BufferedInputStream(new FileInputStream(f.getPath()));
			props.load(in);
			in.close();
		    }
		} catch (Exception e) {}
		return null;
	    }
	});
    }

    //public
    static String setProperty(String key, String value) {
	return (String) props.put(key, value);
    }

    //public
    static String getProperty(String key) {
	return props.getProperty(key);
    }

    //public
    static String getProperty(String key, String def) {
	return props.getProperty(key, def);
    }

    static {
	initializeProps();
    }
}
