/*
 * @(#)InputMethodLookupStart.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

//public 
class InputMethodLookupStart {

    public InputMethodLookupStart(int type,
				  int direction,
				  int size,
				  int rows,
				  int columns,
				  boolean labeled) {
	this.type = type;
	this.direction = direction;
	this.num_items = size;
	this.rows = rows;
	this.columns = columns;
	this.labeled = labeled;
    }

    /**
     * The type of lookup process.  If the type is SERVER_DRIVEN,
     * the input method server sends lookup draw data one page by
     * one page on demand.  Thus, the draw() method in LookupCallback 
     * can be called more than one time for displaying the candidates
     * one after another.
     * If the type is CLIENT_DRIVEN, the input method server sends
     * all lookup draw data at a time.  That is, the client has to make
     * the data a list of pages to fit each of them into the size of the 
     * lookup window and display the candidates page by page.
     * Also, in case of CLIENT_DRIVEN, the process() method in
     * LookupCallback may be called for changing the pages.
     */
    /** The server driven lookup process type. */
    public final static int EVENT_SERVER_DRIVEN = 0;
    /** The client driven lookup process type. */
    public final static int EVENT_CLIENT_DRIVEN = 1;
    
    int type;

    public int getType() {
	return type;
    }

    /** The horizontal drawing direction. */
    public final static int DIRECTION_HORIZONTAL = 10;

    /** The vertical drawing direction. */
    public final static int DIRECTION_VERTICAL = 11;

    /** The drawing direction. */
    int direction;

    public int getDirection() {
	return direction;
    }

    /**
     * The number of lookup candidates to be displayed per lookup window.
     */
    int num_items;

    public int getSize() {
	return num_items;
    }

    /** The number of rows for a lookup window. */
    int rows;

    public int getRows() {
	return rows;
    }

    /** The number of columns for a lookup window. */
    int columns;

    public int getColumns() {
	return columns;
    }

    /**
     * The flag indicating whether lookup candidates are labeled or not.
     * If not, it is the user's responsibility to label each of the
     * lookup candidates.
     */
    boolean labeled;

    public boolean isLabeled() {
	return labeled;
    } 
}
