/*
 * @(#)KeyEventData.java	1.3 98/03/23
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.iiimp;

import java.awt.event.KeyEvent;
import java.io.IOException;

class KeyEventData extends Data {

    KeyEventData(KeyEvent kev) throws IOException {
	super();
	int code = kev.getKeyCode();
	char c = kev.getKeyChar();
	int mod = kev.getModifiers();
	long when = kev.getWhen();

	write4(code);
	write4((int) c);
	write4(mod);
	write4((int) when);
    }

    KeyEventData(byte b[], int len) {
	super(b, len);
    }

    KeyEventData(Data d) {
	super(d);
    }

    IIIMPKeyEvent[] toKeyEvents() throws IOException {
	IIIMPKeyEvent[] kev = null;

	int num_kev = (count / 16);
	if (num_kev > 0) {
	    kev = new IIIMPKeyEvent[num_kev];
	    for (int i = 0; i < num_kev; i++) {
		int keycode = read4();
		char keychar = (char) read4();
		int modifier = read4();
		int timestamp = read4();
		kev[i] = new IIIMPKeyEvent(keycode, keychar, modifier);
	    }
	}

	return kev;
    }

    /**
     * Converts buf to the byte stream, which, if necessary, is padded.
     */
    byte[] toByteStream() throws IOException {
	// reserved 2 bytes space at the head of the stream.
	final int RESERVED = 4;

	if (count == 0) {
	    return null;
	}

	int pad = paddings(count);
	int total = count + pad;
	int real_total = RESERVED + count + pad;

	byte newBuf[] = new byte[real_total];
	System.arraycopy(buf, 0, newBuf, RESERVED, count);

	newBuf[0] = (byte) ((count >>> 24) & 0xFF);
	newBuf[1] = (byte) ((count >>> 16) & 0xFF);
	newBuf[2] = (byte) ((count >>> 8) & 0xFF);
	newBuf[3] = (byte) ((count >>> 0) & 0xFF);

	// padd to 0 just in case
	for (int i = (RESERVED + count); i < real_total; i++) {
	    newBuf[i] = 0;
	}

	// proceed the position to the end
	pos = count;

	return newBuf;
    }
}
