/*
 * @(#)IOR.java	1.40 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.INV_OBJREF;
import com.sun.CORBA.util.*;
import org.omg.CORBA.portable.*;

public final class IOR
{
    private String typeId;
    private int[] profileTags;
    private byte[][] profileData;
    private com.sun.CORBA.ORB factory;
    // Cached lookups
    private Profile iop;

    private static final int InternetIOPTag = 0;


    public IOR(com.sun.CORBA.ORB factory)
    {
	this.factory = factory;
    }

    public IOR(com.sun.CORBA.ORB factory, String typeId, String host,
	       int port, byte[] key)
    {
	this(factory);
        Profile p = new Profile(factory, host, port, key);
        this.typeId = typeId;

        //
        // Null java.lang.Object IOR?
        //
        if (typeId.length() == 0 && p == null) {
            //
            // Make write happy.
            //
            profileTags = new int[0];
            return;
        }

        putProfile(p);
    }

    public IOR(com.sun.CORBA.ORB factory, String typeId, Profile p)
    {
	this(factory);

	this.typeId = typeId;

	//
	// Null java.lang.Object IOR?
	//
	if (typeId.length() == 0 && p == null) {
	    //
	    // Make write happy.
	    //
	    profileTags = new int[0];
	    return;
	}

	putProfile(p);
    }

    public void read(InputStream s)
    {
        int len;

        typeId = s.read_string();
        len = s.read_long();
        profileTags = new int[len];
        profileData = new byte[len][];
        for (int i=0; i < len; i++) {
            profileTags[i] = s.read_long();
            profileData[i] = new byte[s.read_long()];
            s.read_octet_array(profileData[i], 0, profileData[i].length);
        }
    }

    public void write(MarshalOutputStream s)
    {
	s.write_string(typeId);
	s.write_long(profileTags.length);
	for (int i=0; i < profileTags.length; i++) {
	    s.write_long(profileTags[i]);
	    s.write_long(profileData[i].length);
	    s.write_octet_array(profileData[i], 0 , profileData[i].length);
	}
    }

    public synchronized Profile getProfile()
    {
	if (iop != null)
	    return iop;

	for (int i=0; i < profileTags.length; i++) {
            if (profileTags[i] == InternetIOPTag) {
		iop = new Profile(factory, profileData[i]);

		return iop;
	    }
	}

        // if we come to this point then no interesting Profile
        // is present.  Therefore, throw an INV_OBJREF exception.
        throw new INV_OBJREF( MinorCodes.NO_PROFILE_PRESENT,
	    CompletionStatus.COMPLETED_NO );
	// return null;
    }

    public void putProfile(Profile x)
    {
	int[] tags;
	byte[][] data;
	int len = 0;
	int i;

	if (profileTags != null)
	    len = profileTags.length;

	for (i=0; i < len; i++) {
            if (profileTags[i] == InternetIOPTag) {
		profileData[i] = x.getEncapsulation();
		iop = x;

		return;
	    }
	}

	tags = new int[len+1];
	data = new byte[len+1][];
	for (i=0; i < len; i++) {
	    tags[i] = profileTags[i];
	    data[i] = profileData[i];
	}
	tags[i] = InternetIOPTag;
	data[i] = x.getEncapsulation();

	profileTags = tags;
	profileData = data;
	iop = x;
    }

    public String getTypeId()
    {
	return typeId;
    }

    public String stringify()
    {
	ByteArrayOutputStream bs;

	MarshalOutputStream s = factory.newOutputStream();
	s.putEndian();
	write(s);
	bs = new ByteArrayOutputStream();
	try {
	    s.writeTo(new HexOutputStream(bs));
	} catch (IOException ex) {
	    throw new INTERNAL( MinorCodes.STRINGIFY_WRITE_ERROR,
		CompletionStatus.COMPLETED_NO );
	}

	return "IOR:" + bs;
    }

    public boolean is_nil()
    {
        //
        // The check for typeId length of 0 below is commented out
        // as a workaround for a bug in ORBs which send a
        // null objref with a non-empty typeId string.
        //
	return ((profileTags.length == 0) /* && (typeId.length() == 0) */);
    }

    public boolean isEquivalent(IOR ior)
    {
	return getProfile().isEquivalent(ior.getProfile());
    }
}
