/*
 * @(#)Profile.java	1.28 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA;

import org.omg.CORBA.SystemException;

public final class Profile
{
    private String host;
    private int port;
    private byte major;
    private byte minor;
    private byte[] objectKey;
    private byte[] data;
    private ORB factory;

    Profile(ORB factory, byte[] data)
	throws SystemException
    {
	this.factory = factory;

	MarshalInputStream s = factory.newInputStream(data, data.length);
	s.consumeEndian();
	major = s.read_octet();
	minor = s.read_octet();

	host = s.read_string();
	port = s.read_short() & 0xFFFF;
	objectKey = new byte[s.read_long()];
	s.read_octet_array(objectKey, 0, objectKey.length);

	//
	// In IIOP 1.1 there will be a tagged component sequence
	// here.
	//
	this.data = data;
    }

    Profile(ORB factory, String host, int port, byte[] key) {
	this.factory = factory;
	this.host = host;
	this.port = port;
	objectKey = key;
	major = 1;
	minor = 0;
    }

    byte[] getEncapsulation() throws SystemException {

	if (data != null)
	    return data;

	MarshalOutputStream s = factory.newOutputStream();
	s.putEndian();
	s.write_octet(major);
	s.write_octet(minor);
	s.write_string(host);
	s.write_short((short)port);
	s.write_long(objectKey.length);
	s.write_octet_array(objectKey, 0, objectKey.length);

	//
	// In IIOP 1.1 we would marshal a sequence of
	// tagged components here.
	//
	data = s.toByteArray();
	return data;
    }

    public String getHost() {
	return host;
    }

    public int getPort() {
	return port;
    }

    public byte[] getObjectKey() {
	return objectKey;
    }

    boolean isEquivalent(Profile p) {
	if (getPort() != p.getPort() ||
	  !getHost().equalsIgnoreCase(p.getHost()) ||
	  getObjectKey().length != p.getObjectKey().length)
	    return false;

	byte[] key = p.getObjectKey();
	for (int i=0; i < objectKey.length; i++)
	    if (objectKey[i] != key[i])
		return false;

	return true;
    }

}
