/*
 * @(#)ServerGIOP.java	1.8 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA;

import java.net.InetAddress;

/**
 * ServerGIOP contains the server-side GIOP APIs.
 */
public interface ServerGIOP
{

    /**
     * Get the default end point, if you don't care about an end point
     * on a particular port.
     *
     * NOTE: If getDefaultEndpoint returns null, initEndpoints() needs
     *       to be called after which getDefaultEndpoint will return
     *       the default end point.
     */
    public EndPoint getDefaultEndpoint();

    /**
     * Initialize the vector of end points.
     */
    public void initEndpoints();

    /**
     * Get an EndPoint for the specified type, listenPort and local address.
     * type must be one of CLEAR_TEXT or SSL.
     * If listenPort == 0, a listening port will be assigned.
     * If addr == null, InetAddress.getLocalHost() is used.
     * If an EndPoint at the specified port/address does not exist,
     * then it will be created.
     */
    public EndPoint getEndpoint(int type, int listenPort, InetAddress addr);

    /**
     * Get an EndPoint for the bootstrap naming service, at the specified port.
     * If listenPort == 0 and this is the first time getBootstrapEndpoint was
     * invoked, then an EndPoint at the default port of 900 will be created.
     * If listenPort == 0 and this is not the first time getBootstrapEndpoint
     * was called, then the existing bootstrap EndPoint will be returned.
     */
    public EndPoint getBootstrapEndpoint(int port);

    /**
     * Does this string correspond to my host?
     */
    public boolean isLocalHost(String hostname);

    /**
     * All incoming requests will be dispatched to the RequestHandler.
     */
    public void setRequestHandler(RequestHandler handler);

    /**
     * Return the RequestHandler.
     */
    public RequestHandler getRequestHandler();
}

