/*
 * @(#)ServerSubcontract.java	1.7 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA;

/**
 * Server subcontract adds behavior on the server-side -- specifically
 * on the dispatch path. A single server subcontract instance serves
 * many server objects.
 */
public interface ServerSubcontract {

    /**
     * This should be the first four bytes of IORs created by
     * JavaIDL and all derivative ORBs.
     */
    public final int JAVAMAGIC = 0xAFabCafe;

    /**
     * Create an object reference given a servant and/or user-key.
     *
     * @param key the user key part of the ORB key
     * @param servant the server object (may be null)
     */
    public Object createObjref(byte[] key, Object servant);

    /**
     * Create an object reference given an IOR.
     *
     * @param ior the ior to create an objref for.
     */
    public Object createObjref(IOR ior) throws Exception;

    /**
     * Destroy the object reference.
     *
     * XXX: should the argument be the servant or the
     *      object reference?
     */
    public void destroyObjref(Object objref);

    /**
     * Dispatch to the server object and return its
     * response.
     *
     * dispatch() should catch all exceptions and convert
     * them to a ServerResponse.
     *
     */
    public ServerResponse dispatch(ServerRequest request);

    /**
     * Handle a locate request.
     */
    public IOR locate(byte[] key);

    /**
     * Lookup and return the servant corresponding to the object key
     */
    public Object getServant(byte[] key);

    /**
     * Get the parallel client subcontract class
     */
    public Class getClientSubcontractClass();

    /**
     * Set the subcontract ID assigned by the ORB for this
     * class.
     */
    void setId(int id);

    /**
     * Set the ORB instance that created this SC
     */
    void setOrb(ORB orb);
}
