/*
 * @(#)SubcontractRegistry.java	1.9 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA;

import java.util.*;

/**
 * This is a registry of all client subcontract classes and
 * server subcontract instances.
 *
 * In order to allow reuse of this in more than one context
 * (that is, ORBs), we try not to throw any CORBA specific
 * exceptions. Most of the methods either throw Java exceptions
 * or return nulls or false on failure.
 */
public class SubcontractRegistry {
    private int size;
    private ServerSubcontract[] serverRegistry;
    private Class[] clientFactory;
    private int defaultId;

    public SubcontractRegistry(int defaultId) {
        this(defaultId, 256);
    }

    public SubcontractRegistry(int defaultId, int size) {
        this.size = size;
        this.defaultId = defaultId;
        serverRegistry = new ServerSubcontract[size];
        clientFactory = new Class[size];
    }

    public synchronized boolean registerClient(Class clientSC,
					       int scid)
    {
        if (scid < size) {
            clientFactory[scid] = clientSC;
	    return true;
	}
	return false;
    }

    public synchronized boolean registerServer(ServerSubcontract sc,
					       int scid) {
        if (scid < size) {
	    serverRegistry[scid] = sc;
	    sc.setId(scid);
	    Class client = sc.getClientSubcontractClass();
	    if (client != null)
		registerClient(client, scid);
	    return true;
        }
	return false;
    }

    protected ServerSubcontract bootstrapServer = null;
    public void registerBootstrapServer(ServerSubcontract sc)
    {
	this.bootstrapServer = sc;
    }

    public ServerSubcontract getServerSubcontract(int scid)
    {
	ServerSubcontract sc=null;
        if (scid < size)
            sc = serverRegistry[scid];

	if ( sc == null )
	    return serverRegistry[defaultId];
	else
	    return sc;
    }

    public ServerSubcontract getServerSubcontract(byte[] objKey)
    {
        int magic;
	ServerSubcontract sc=null;

        if (objKey.length == 4) {
            magic = getMagic(objKey);
	    if ((magic == 0x494e4954) || (magic == 0x54494e49))
		return bootstrapServer;
	} else if (objKey.length >= 8) {
            magic = getMagic(objKey);
            if (magic == ServerSubcontract.JAVAMAGIC) {
                int scid = getId(objKey);
                if (scid < size)
                    sc = serverRegistry[scid];
            }
        }

	if ( sc == null )
	    return serverRegistry[defaultId];
	else
	    return sc;
    }


    // Called to get Client side Subcontract
    public ClientSubcontract getClientSubcontract(byte[] key) {
        Class repClass = null;
	int scid = defaultId;

        try {
            if (key.length >= 8) {
                int magic = getMagic(key);
                if (magic == ServerSubcontract.JAVAMAGIC) {
                    scid = getId(key);
                    if (scid < size)
                        repClass = clientFactory[scid];
                }
            }
	    if ( repClass == null )
                repClass = clientFactory[defaultId];
	    ClientSubcontract rep =
		(ClientSubcontract) repClass.newInstance();
	    rep.setId(scid);
            return rep;
        }
        catch (Exception e1){
	    return null;
        }
    }

    protected int getMagic(byte[] key) {
	return bytesToInt(key, 0);
    }

    protected int getId(byte[] key) {
	int id = bytesToInt(key, 4);
	return id;
    }

    protected int getServerId(byte[] key) {
	return bytesToInt(key, 8);
    }

    private static int bytesToInt(byte[] array, int offset) {
	int b1, b2, b3, b4;

        b1 = (array[offset++] << 24) & 0xFF000000;
        b2 = (array[offset++] << 16) & 0x00FF0000;
        b3 = (array[offset++] << 8)  & 0x0000FF00;
        b4 = (array[offset++] << 0)  & 0x000000FF;

	return (b1 | b2 | b3 | b4);
    }
}
