/*
 * @(#)AnyImpl.java	1.30 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.idl;

import org.omg.CORBA.*;
import org.omg.CORBA.portable.InputStream;
import org.omg.CORBA.portable.OutputStream;
import org.omg.CORBA.portable.Streamable;

import com.sun.CORBA.*;
import com.sun.CORBA.iiop.CDRInputStream;
import com.sun.CORBA.iiop.CDROutputStream;



final class AnyInputStream extends CDRInputStream {
    AnyInputStream(org.omg.CORBA.ORB orb, byte[] data, int size,
      boolean littleEndian) {
	super(orb, data, size, littleEndian);
    }
}

final class AnyOutputStream extends CDROutputStream {

    public AnyOutputStream(org.omg.CORBA.ORB orb) {
        super(orb, false);
    }

    public AnyOutputStream(org.omg.CORBA.ORB orb, int size) {
        super(orb, false, size);
    }

    public InputStream create_input_stream() {
        return new AnyInputStream(orb, buf, size, littleEndian);
    }
}


public final class AnyImpl extends Any
{

  //
  // Always valid.
  //
  private TypeCodeImpl typeCode;
  private org.omg.CORBA.ORB orb;

  //
  // Validity depends upon typecode. The 'value' and 'object' instance
  // members are used to hold immutable types as defined by the
  // isStreamed[] table below. Otherwise, 'stream' is non-null and
  // holds the value in CDR marshaled format. As an optimization, the
  // stream type is an Any extension of CDR stream that is used to
  // detect an optimization in read_value().
  //
  private CDRInputStream stream;
  private long value;
  private java.lang.Object object;

  private static final int DEFAULT_BUFFER_SIZE = 32;


    /*
     * This boolean array tells us if a given typecode must be
     * streamed. Objects that are immutable don't have to be streamed.
     */
    static boolean isStreamed[] = {
	false, false, false, false, false, false, false, false, false, 
//      null   void   short  long   ushort ulong  float  double boolean 

	false, false, true, false,   true,     false, true,  true, true, 
//	char   octet  any   TypeCode Principal objref struct union enum 

	false, true,    true, true, true,  false,   false,    false,
//	string sequence array alias except longlong ulonglong longdouble 

	false, false, false
//	wchar wstring fixed 

    };

  ///////////////////////////////////////////////////////////////////////////
  // Constructors

  /**
   * A constructor that sets the Any to contain a null. It also marks
   * the value as being invalid so that extractions throw an exception
   * until an insertion has been performed. 
   */

  public AnyImpl(org.omg.CORBA.ORB orb)
  {
    // Note: orb could be an instance of ORBSingleton or the actual ORB
    this.orb = orb;
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_null);
  }


  //
  // Create a new AnyImpl which is a copy of obj.
  //
  AnyImpl(org.omg.CORBA.ORB orb, Any obj) {
    this(orb);

    if ((obj instanceof AnyImpl)) {
      AnyImpl objImpl = (AnyImpl)obj;
      typeCode = objImpl.typeCode;
      value = objImpl.value;
      object = objImpl.object;
      if (objImpl.stream != null)
        stream = objImpl.stream.dup();
    } else {
      read_value(obj.create_input_stream(), obj.type());
    }
  }


  ///////////////////////////////////////////////////////////////////////////
  // basic accessors

  /**
   * returns the type of the element contained in the Any.
   *
   * @result		the TypeCode for the element in the Any
   */
   
  public TypeCode type() 
  {
    return typeCode;
  }


  /**
   * sets the type of the element to be contained in the Any.
   *
   * @param tc		the TypeCode for the element in the Any
   */
   
  public void type(TypeCode tc) 
  {
    // set the typecode
    typeCode = TypeCodeImpl.convertToNative(orb, tc);

    stream = null;
    value = 0;
    object = null;
  }


  /**
   * checks for equality between Anys.
   *
   * @param otherAny	the Any to be compared with.
   * @result		true if the Anys are equal, false otherwise.
   */
   
  public boolean equal(Any otherAny) 
  {
    // first check for typecode equality.
    // note that this will take aliases into account
    if (!typeCode.equal(otherAny.type()))
      return false;

    switch (typeCode.kind().value()) {
      // handle primitive types
    case TCKind._tk_null:
    case TCKind._tk_void:
      return true;
    case TCKind._tk_short:
      return (extract_short() == otherAny.extract_short());
    case TCKind._tk_long:
      return (extract_long() == otherAny.extract_long());
    case TCKind._tk_ushort:
      return (extract_ushort() == otherAny.extract_ushort());
    case TCKind._tk_ulong:
      return (extract_ulong() == otherAny.extract_ulong());
    case TCKind._tk_float:
      return (extract_float() == otherAny.extract_float());
    case TCKind._tk_double:
      return (extract_double() == otherAny.extract_double());
    case TCKind._tk_boolean:
      return (extract_boolean() == otherAny.extract_boolean());
    case TCKind._tk_char:
      return (extract_char() == otherAny.extract_char());
    case TCKind._tk_octet:
      return (extract_octet() == otherAny.extract_octet());
    case TCKind._tk_any:
      return extract_any().equal(otherAny.extract_any());
    case TCKind._tk_TypeCode:
      return extract_TypeCode().equal(otherAny.extract_TypeCode());
    case TCKind._tk_string:
      return extract_string().equals(otherAny.extract_string());
    case TCKind._tk_longlong:
      return (extract_longlong() == otherAny.extract_longlong());
    case TCKind._tk_ulonglong:
      return (extract_ulonglong() == otherAny.extract_ulonglong());
    case TCKind._tk_enum:
      return (create_input_stream().read_long() == 
               otherAny.create_input_stream().read_long());
     
      // [TODO] not yet implemented for extended types

    case TCKind._tk_longdouble:
    case TCKind._tk_wchar:
    case TCKind._tk_wstring:
      throw new org.omg.CORBA.NO_IMPLEMENT();

      // [TODO] not yet implemented for complex types

    case TCKind._tk_objref:
    case TCKind._tk_Principal:
    case TCKind._tk_struct:
    case TCKind._tk_union:
    case TCKind._tk_sequence:
    case TCKind._tk_array:
    case TCKind._tk_alias:
    case TCKind._tk_except:
      throw new org.omg.CORBA.NO_IMPLEMENT();
    }
    return false;
  }

  ///////////////////////////////////////////////////////////////////////////
  // accessors for marshaling/unmarshaling

  /**
   * returns an output stream that an Any value can be marshaled into. 
   *
   * @result		the OutputStream to marshal value of Any into
   */
   
  public OutputStream create_output_stream()
  {
    return new AnyOutputStream(orb, DEFAULT_BUFFER_SIZE);
  }

  /**
   * returns an input stream that an Any value can be marshaled out of.
   *
   * @result		the InputStream to marshal value of Any out of.
   */

  //
  // We create a new InputStream so that multiple threads can call here
  // and read the streams in parallel without thread safety problems.
  //
  public InputStream create_input_stream()
  {
    if (stream != null) {
	return stream.dup();
    }

    CDROutputStream os = new CDROutputStream(orb);
    TCUtility.marshalIn(os, typeCode.kind().value(), value, object);

    return os.create_input_stream();
  }

  ///////////////////////////////////////////////////////////////////////////
  // marshaling/unmarshaling routines

  //
  // If the InputStream is a CDRInputStream then we can copy the bytes
  // since it is in our format and does not have alignment issues.
  //
  public void read_value(InputStream in, TypeCode tc)
  { 
    //
    // Assume that someone isn't going to think they can keep reading
    // from this stream after calling us. That would be likely for
    // an IIOPInputStream but if it is an AnyInputStream then they
    // presumably obtained it via our create_output_stream() so they could
    // write the contents of an IDL data type to it and then call
    // create_input_stream() for us to read it. This is how Helper classes
    // typically implement the insert() method.
    // We should probably document this behavior in the 1.1 revision
    // task force.
    //
    typeCode = TypeCodeImpl.convertToNative(orb, tc);

    int type = typeCode.kind().value();

    if (AnyImpl.isStreamed[type]) {
      if (in instanceof AnyInputStream) {
        stream = (CDRInputStream)in;
	if (stream.getORB() == orb)
	  return;
      }
      CDROutputStream out = new CDROutputStream(orb);
      typeCode.copy(in, out);
      stream = (CDRInputStream)out.create_input_stream();
    } else {
      java.lang.Object[] objholder = new java.lang.Object[1];
      objholder[0] = object;
      long[] longholder = new long[1];
      TCUtility.unmarshalIn(in, type, longholder, objholder);
      value = longholder[0];
      object = objholder[0];
      stream = null;
    }
  }


  //
  // We could optimize this by noticing whether the target stream
  // has ever had anything marshaled on it that required an
  // alignment of greater than 4 (was write_double() ever called on it).
  // If not, then we can just do a byte array copy without having to
  // drive the remarshaling through typecode interpretation.
  //
  public void write_value(OutputStream out)
  {
    if (stream != null) {
      typeCode.copy(stream.dup(), out);
    } else {
      TCUtility.marshalIn(out, typeCode.kind().value(), value, object);
    }
  }

  ///////////////////////////////////////////////////////////////////////////
  // insertion of streamables 
  //
  // We need to more clearly define the semantics of streamables
  // passed in here. The current implementation is safe. A faster
  // implementation would merely save the streamable reference and
  // defer using it until needed. Doing this would eliminate the
  // extra objects generated below in order to capture the state of the
  // streamable immediately. However, we need to document this behavior
  // since it has an impact on users of the API. This should be handled
  // as part of the OMG Java IDL mapping 1.1 revision task force.
  //

  /**
   * takes a streamable and inserts it into the any
   *
   * @param s		the streamable to insert
   */
  public void insert_Streamable(Streamable s)
  {
    AnyOutputStream os = new AnyOutputStream(orb);
    s._write(os);
    read_value(os.create_input_stream(), s._type());
  }
  
  ///////////////////////////////////////////////////////////////////////////
  // insertion/extraction/replacement for all basic types

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_short(short s)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_short);
    value = s;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public short extract_short()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_short)))
      throw new BAD_OPERATION();
    return (short)value;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_long(int l)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_long);
    value = l;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public int extract_long()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_long)))
      throw new BAD_OPERATION();
    return (int)value;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_ushort(short s)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_ushort);
    value = s;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public short extract_ushort()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_ushort)))
      throw new BAD_OPERATION();
    return (short)value;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_ulong(int l)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_ulong);
    value = l;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public int extract_ulong()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_ulong)))
      throw new BAD_OPERATION();
    return (int)value;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_float(float f)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_float);
    value = Float.floatToIntBits(f);
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public float extract_float()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_float)))
      throw new BAD_OPERATION();
    return Float.intBitsToFloat((int)value);
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_double(double d)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_double);
    value = Double.doubleToLongBits(d);
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public double extract_double()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_double)))
      throw new BAD_OPERATION();
    return Double.longBitsToDouble(value);
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_longlong(long l)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_longlong);
    value = l;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public long extract_longlong()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_longlong)))
      throw new BAD_OPERATION();
    return value;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_ulonglong(long l)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_ulonglong);
    value = l;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public long extract_ulonglong()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_ulonglong)))
      throw new BAD_OPERATION();
    return value;
  }


  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_boolean(boolean b)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_boolean);
    value = (b)? 1:0;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public boolean extract_boolean()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_boolean)))
      throw new BAD_OPERATION();
    return (value == 0)? false: true;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_char(char c)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_char);
    value = c;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public char extract_char()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_char)))
      throw new BAD_OPERATION();
    return (char)value;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_wchar(char c)
  {
    // [TODO] need to figure out what to do with wchars
    throw new org.omg.CORBA.NO_IMPLEMENT();
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public char extract_wchar()
  {
    // [TODO] need to figure out what to do with wchars
    throw new org.omg.CORBA.NO_IMPLEMENT();
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_octet(byte b)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_octet);
    value = b;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public byte extract_octet()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_octet)))
      throw new BAD_OPERATION();
    return (byte)value;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_string(String s)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_string);
    object = s;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public String extract_string()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_string)))
      throw new BAD_OPERATION();
    return (String)object;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_wstring(String s)
  {
    // [TODO] need to figure out what to do with wstrings
    throw new org.omg.CORBA.NO_IMPLEMENT();
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public String extract_wstring()
  {
    // [TODO] need to figure out what to do with wstrings
    throw new org.omg.CORBA.NO_IMPLEMENT();
  }

 /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_any(Any a)
  {
    //
    // Anys are mutable so we need to copy the data by marshaling.
    //
    AnyOutputStream os = new AnyOutputStream(orb);
    os.write_any(a);
    read_value(os.create_input_stream(), TypeCodeImpl.get_primitive_tc(TCKind.tk_any));
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public Any extract_any()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_any)))
      throw new BAD_OPERATION();
    return stream.dup().read_any();
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_Object(org.omg.CORBA.Object o)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_objref);
    object = o;
  }

  /**
   * A variant of the insertion operation that takes a typecode
   * argument as well.
   */

  public void insert_Object(org.omg.CORBA.Object o, TypeCode tc)
  {
    try {
      if ( tc.id().equals("IDL:omg.org/CORBA/Object:1.0") || o._is_a(tc.id()) )
      {
        typeCode = TypeCodeImpl.convertToNative(orb, tc);
        object = o;
      }
      else {
        throw new BAD_OPERATION();
      }
    } catch ( Exception ex ) {
      throw new BAD_OPERATION();
    }
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public org.omg.CORBA.Object extract_Object()
  {
    // Check if the object contained here is of the type in typeCode
    org.omg.CORBA.Object obj = null;
    try {
      obj = (org.omg.CORBA.Object) object;
      if ( typeCode.id().equals("IDL:omg.org/CORBA/Object:1.0") 
	   || obj._is_a(typeCode.id()) ) {
        return obj;
      }
      else {
        throw new BAD_OPERATION();
      }
    } catch ( Exception ex ) {
      throw new BAD_OPERATION();
    }
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_TypeCode(TypeCode tc)
  {
    typeCode = TypeCodeImpl.get_primitive_tc(TCKind.tk_TypeCode);
    object = tc;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public TypeCode extract_TypeCode()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_TypeCode)))
      throw new BAD_OPERATION();
    return (TypeCode)object;
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public void insert_Principal(Principal p)
  {
    //
    // Anys are mutable so we need to copy the data by marshaling.
    //
    AnyOutputStream os = new AnyOutputStream(orb);
    os.write_Principal(p);
    read_value(os.create_input_stream(), TypeCodeImpl.get_primitive_tc(TCKind.tk_Principal));
  }

  /**
   * See the description of the <a href="#anyOps">general Any operations.</a>
   */

  public Principal extract_Principal()
  {
    if (!typeCode.equal(TypeCodeImpl.get_primitive_tc(TCKind.tk_Principal)))
      throw new BAD_OPERATION();
    return stream.dup().read_Principal();
  }
}

