/*
 * @(#)AsynchInvoke.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.idl;


///////////////////////////////////////////////////////////////////////////
// helper class for deferred invocations

/*
 * The AsynchInvoke class allows for the support of asynchronous
 * invocations. Instances of these are created with a request object,
 * and when run, perform an invocation. The instance is also
 * responsible for waking up a client that might be waiting on the
 * 'get_response' method.
 */

class AsynchInvoke implements Runnable {

  private RequestImpl _req;
  private ORB         _orb;
  private boolean     _notifyORB;

  AsynchInvoke (ORB o, RequestImpl reqToInvokeOn, boolean n)
  {
    _orb = o;
    _req = reqToInvokeOn;
    _notifyORB = n;
  };


  /*
   * The run operation calls the invocation on the request object,
   * updates the RequestImpl state to indicate that the asynchronous
   * invocation is complete, and wakes up any client that might be
   * waiting on a 'get_response' call.
   *
   */

  public void run() 
  {

    // do the actual invocation
    _req.doInvocation();
    
    // for the asynchronous case, note that the response has been
    // received.
    synchronized (_req)
      {
        // update local boolean indicator
        _req.gotResponse = true;

        // notify any client waiting on a 'get_response'
        _req.notify();
      }
      
    if (_notifyORB == true) {
      synchronized(_orb._svResponseReceived) {
        _orb.notifyResponse();
      }
    }
  }

};

///////////////////////////////////////////////////////////////////////////
