/*
 * @(#)GenericCORBAClientSC.java	1.18 98/10/12
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.idl;

import java.io.IOException;
import java.net.InetAddress;
import java.util.Hashtable;
import java.util.StringTokenizer;

import org.omg.CORBA.ARG_IN;
import org.omg.CORBA.ARG_INOUT;
import org.omg.CORBA.ARG_OUT;
import org.omg.CORBA.Any;
import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.Context;
import org.omg.CORBA.ContextList;
import org.omg.CORBA.ExceptionList;
import org.omg.CORBA.Environment;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.NO_IMPLEMENT;
import org.omg.CORBA.NVList;
import org.omg.CORBA.NamedValue;
import org.omg.CORBA.Request;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.UNKNOWN;
import org.omg.CORBA.UnknownUserException;
import org.omg.CORBA.UserException;
import org.omg.CORBA.WrongTransaction;

import org.omg.CORBA.portable.*;

import com.sun.CORBA.*;
import com.sun.CORBA.util.*;



/**
 * The generic CORBA client subcontract (delegate) class.
 */

public class GenericCORBAClientSC extends Delegate implements ClientSubcontract 
{
    protected static final int MAX_RETRIES = 3;
    protected IOR ior;
    protected IOR locatedIOR;
    protected int scid;  // my subcontract id
    protected com.sun.CORBA.ORB orb;


    public GenericCORBAClientSC() {
    }

    public GenericCORBAClientSC(IOR myIor)
    {
        this.ior = myIor;
	this.locatedIOR = ior;
    }

    public IOR getIOR()
    {
        return ior;
    }

    public void setIOR(IOR ior)
    {
        this.ior = ior;
	this.locatedIOR = ior;
    }

    public com.sun.CORBA.ORB getOrb() {
        return orb;
    }  

    public void setOrb(com.sun.CORBA.ORB orb) {
        this.orb = orb;
    }

    public void setId(int scid)
    {
        this.scid = scid;
    }

    // Why is this called marshal ??
    public IOR marshal()
    {
	return getIOR();
    }


    // Why is this called unmarshal ??
    public void unmarshal(IOR ior)
    {
	setIOR(ior);
    }
 

    // Invoked by DII layer (from RequestImpl.doInvocation()).
    public ClientResponse invoke(ClientRequest request)
    {
	ClientResponse resp = request.invoke();

	if ( request.isOneWay() )
	    return null;	

	if ( resp.isLocationForward() ) {
	    IOR[] iorHolder = new IOR[1];
	    iorHolder[0] = resp.getForwardedIOR();
	    resp = request.reInvoke(iorHolder);
	    locatedIOR = iorHolder[0];
	}

	return resp;
    }


    // Invoked by DII layer (from RequestImpl.doInvocation()).
    // before marshaling arguments.
    public ClientRequest createRequest(String opName, boolean isOneWay)
    {
        // Create request object which is also the outputstream
        ClientGIOP giop = orb.getClientGIOP();
        return giop.createRequest(locatedIOR, opName, isOneWay, null);
    }

    // Invoked by DII layer (from RequestImpl.doInvocation())
    // after unmarshaling reply and exceptions.
    public void releaseReply(ClientResponse resp, String method,
                                Exception exception)
        throws WrongTransaction, SystemException
    {
    }




/*************************************************************************/
/* Implementation of CORBA::Object methods from Delegate *****************/
/*************************************************************************/

    public boolean is_equivalent(org.omg.CORBA.Object obj, 
				 org.omg.CORBA.Object ref)
    {
        if ( ref == null )
            return false;
        ObjectImpl oi = (ObjectImpl)ref;

        ClientSubcontract del = (ClientSubcontract)oi._get_delegate();
        if (del == this)
            return true;
        if (!(del instanceof GenericCORBAClientSC))
            return false;
        return ior.isEquivalent(del.marshal());
    }

    public int hash(org.omg.CORBA.Object obj, int maximum) 
    {
        int h = this.hashCode();
        if ( h > maximum )
            return 0;
        return h;
    }

    public org.omg.CORBA.Object duplicate(org.omg.CORBA.Object orig)
    {
        return orig;
    }

    public void release(org.omg.CORBA.Object obj)
    {
        // DONT clear out internal variables to release memory !!
        // This delegate may be pointed-to by other objrefs !
    }

    public org.omg.CORBA.Request create_request(org.omg.CORBA.Object obj, 
                                           org.omg.CORBA.Context ctx,
                                           String operation,
                                           org.omg.CORBA.NVList arg_list,
                                           org.omg.CORBA.NamedValue result)
    {
        return new RequestImpl(orb, obj, ctx, operation, arg_list,
                               result, null, null);
    }

    public org.omg.CORBA.Request request(org.omg.CORBA.Object obj,
                                         String operation) 
    {
        return new RequestImpl(orb, obj, null, operation, null, null, null,
                               null);
    }

    public org.omg.CORBA.ORB orb(org.omg.CORBA.Object self) 
    {
        return this.orb;
    }

    public org.omg.CORBA.Request create_request(
                                  org.omg.CORBA.Object obj, 
                                  org.omg.CORBA.Context ctx, 
                                  String operation, 
                                  org.omg.CORBA.NVList arg_list,
                                  org.omg.CORBA.NamedValue result, 
                                  org.omg.CORBA.ExceptionList exclist, 
                                  org.omg.CORBA.ContextList ctxlist)
    {
        return new RequestImpl(orb, obj, ctx, operation, arg_list, result,
                               exclist, ctxlist);
    }
  
    public org.omg.CORBA.Object get_interface_def(org.omg.CORBA.Object obj) 
    {
        // Invoke on server
        ClientRequest os = createRequest("_interface", false);
        ClientResponse is = invoke(os);

        // Process reply
        if ( is.isSystemException() ) 
            throw is.getSystemException();
        else if ( is.isLocationForward() ) {
	    locatedIOR = is.getForwardedIOR();
		
	    // Invoke again.
            return get_interface_def(obj);
	}
        else { // normal return
	    ObjectImpl objimpl = (ObjectImpl)is.read_Object();

            // check if returned object is of correct type
            if ( !objimpl._is_a("IDL:omg.org/CORBA/InterfaceDef:1.0") )
                throw new org.omg.CORBA.UNKNOWN("InterfaceDef object of wrong type returned by server", 0, CompletionStatus.COMPLETED_MAYBE);
    
            // instantiate the stub
            org.omg.CORBA.Object stub;
            try {
                stub = (org.omg.CORBA.Object)Class.forName("org.omg.CORBA._InterfaceDefStub", true, Thread.currentThread().getContextClassLoader()).newInstance();
            } catch ( Exception ex ) {
                throw new org.omg.CORBA.UNKNOWN("org.omg.CORBA._InterfaceDefStub class not available", 0, CompletionStatus.COMPLETED_NO);
            }
    
            // set the delegate in the stub
            org.omg.CORBA.portable.Delegate del = objimpl._get_delegate();
            org.omg.CORBA.portable.ObjectImpl stubimpl = (org.omg.CORBA.portable.ObjectImpl)stub;
            stubimpl._set_delegate(del);
    
            return stub;
	}
    }

    public boolean is_a(org.omg.CORBA.Object obj, String dest) 
    {
        // dest is the typeId of the interface to compare against.
        // repositoryIds is the list of typeIds that the stub knows about.
	// First we look for an answer using local information.
        // But repositoryIds may not be complete, so it may be necessary to
        // go to server.
        String [] repositoryIds = ((ObjectImpl)obj)._ids();
        String myid = ior.getTypeId();
        if ( dest.equals(myid) )
            return true;
        for ( int i=0; i<repositoryIds.length; i++ )
            if ( dest.equals(repositoryIds[i]) )
                return true;

        // Invoke on server
        ClientRequest os = createRequest("_is_a", false);
	os.write_string(dest);
        ClientResponse is = invoke(os);

        // Process reply
        if ( is.isSystemException() ) 
            throw is.getSystemException();
        else if ( is.isLocationForward() ) {
	    locatedIOR = is.getForwardedIOR();
	    // Invoke again.
            return is_a(obj, dest);
	}
        else // normal return
	    return is.read_boolean();
    }

    public boolean non_existent(org.omg.CORBA.Object obj) 
    {
        try {
            // Invoke on server. operation name *is* _not_existent!
            ClientRequest os = createRequest("_not_existent", false);
            ClientResponse is = invoke(os);

            // Process reply
            if ( is.isSystemException() ) 
                return true;
            else if ( is.isLocationForward() ) {
	        locatedIOR = is.getForwardedIOR();
		// Invoke again.
                return non_existent(obj);
	    }
            else // normal return
	        return is.read_boolean();
        } catch (SystemException ex) {
            return true;
        }
    }

}


