/*
 * @(#)GenericCORBAServerSC.java	1.16 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.idl;

import java.net.*;
import java.util.*;
//import java.io.*;

import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.UserException;
import org.omg.CORBA.Any;
import org.omg.CORBA.ARG_IN;
import org.omg.CORBA.ARG_OUT;
import org.omg.CORBA.ARG_INOUT;
import org.omg.CORBA.UNKNOWN;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.NO_IMPLEMENT;
import org.omg.CORBA.BAD_OPERATION;
import org.omg.CORBA.COMM_FAILURE;
import org.omg.CORBA.OBJECT_NOT_EXIST;
import org.omg.CORBA.Object;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.DynamicImplementation;
import org.omg.CORBA.ServerRequest;

import org.omg.CORBA.portable.*;

import com.sun.CORBA.*;
import com.sun.CORBA.util.*;



/** Generic CORBA server-side subcontract.
 *  This is the super class for all CORBA server-side subcontracts.
 */

public class GenericCORBAServerSC implements ServerSubcontract
{
    // These offsets are for the object key. Assume the first 8 bytes
    // contain the magic and the sc-id.
    public static final int SERVERID_OFFSET = 8;
    public static final int USERKEYLEN_OFFSET = 12;
    public static final int USERKEY_OFFSET = 16;

    protected com.sun.CORBA.ORB orb;
    protected int scid;
    private TransientObjectManager transientObjMgr;


    public GenericCORBAServerSC() 
    {
	transientObjMgr = new TransientObjectManager();
    }

    public GenericCORBAServerSC(com.sun.CORBA.ORB orb) 
    {
        this.orb = orb;
	transientObjMgr = new TransientObjectManager();
    }

    public com.sun.CORBA.ORB getORB() 
    {
        return orb;
    }

    public void setOrb(com.sun.CORBA.ORB orb) 
    {
        this.orb = orb;
    }

    public final void setId(int id) 
    {
        scid = id;
    }

    public final int getId() 
    {
        return scid;
    }


    // Need to signal one of OBJECT_HERE, OBJECT_FORWARD, OBJECT_NOT_EXIST.
    public IOR locate(byte[] objectKey)
    {
        // Check if the serverid matches this server's transientServerId.
        int sId = Utility.bytesToInt(objectKey, SERVERID_OFFSET);
        if ( sId != orb.getTransientServerId() ) 
            throw new OBJECT_NOT_EXIST(MinorCodes.BAD_SERVER_ID,
                            	       CompletionStatus.COMPLETED_NO);
     
	// Get the userKey
	int len = Utility.bytesToInt(objectKey, USERKEYLEN_OFFSET);
        byte[] userKey = new byte[len];
        for ( int i=0; i<userKey.length; i++ )
            userKey[i] = objectKey[i+USERKEY_OFFSET];

	// Get the servant
	DynamicImplementation servant = (DynamicImplementation)
					transientObjMgr.lookupServant(userKey);
	if ( servant == null ) 
	    throw new OBJECT_NOT_EXIST(MinorCodes.SERVANT_NOT_FOUND,
				       CompletionStatus.COMPLETED_NO);

	// If we reached here, it means we got the servant, 
	// so just return null to signal OBJECT_HERE.
	return null;
    }



    /** dispatch() is the start of the upcall for remote invocations.
     *  The InputStream cursor is at the end of the RequestHeader
     *  and the request's arguments can be read directly. On normal return
     *  the OutputStream should contain the entire reply message
     *  (Header + return arguments if any).
     */
    public ServerResponse dispatch(com.sun.CORBA.ServerRequest req)
    {
	InputStream is = (InputStream)req;

	// Get object key and its server id.
        byte[] objectKey = req.getObjectKey();
        int sId = Utility.bytesToInt(objectKey, SERVERID_OFFSET);

        // Check if the serverid matches this server's transientServerId.
        if ( sId != orb.getTransientServerId() ) 
            throw new OBJECT_NOT_EXIST(MinorCodes.BAD_SERVER_ID,
                            	       CompletionStatus.COMPLETED_NO);
     
	// Get the userKey
	int len = Utility.bytesToInt(objectKey, USERKEYLEN_OFFSET);
        byte[] userKey = new byte[len];
        for ( int i=0; i<userKey.length; i++ )
            userKey[i] = objectKey[i+USERKEY_OFFSET];

	ServerResponse resp;
	OutputStream os;
	
	try {
	    String method = req.getOperationName();
	    
	    // Get the servant
	    DynamicImplementation servant = (DynamicImplementation)
					transientObjMgr.lookupServant(userKey);
	    if ( servant == null )
		if (method.equals("_not_existent")) {
		    resp = req.createResponse(null);
		    os = (OutputStream)resp;
		    os.write_boolean(true);
		    return resp;
		}
		else
		    throw new OBJECT_NOT_EXIST(MinorCodes.SERVANT_NOT_FOUND,
					       CompletionStatus.COMPLETED_NO);
	    
	    

	    // First handle special methods
            if ( method.equals("_is_a") ) {
		String[] ids = servant._ids();
		String clientid = is.read_string();
        	boolean answer = false;
        	for ( int i=0; i<ids.length; i++ ) {
            	    if ( ids[i].equals(clientid) )
                	answer = true;
        	}
		resp = req.createResponse(null);
		os = (OutputStream)resp;
		os.write_boolean(answer);
            }
            else if ( method.equals("_interface") ) {
                throw new NO_IMPLEMENT(MinorCodes.GETINTERFACE_NOT_IMPLEMENTED,
                                            CompletionStatus.COMPLETED_NO);
            }
            else if ( method.equals("_not_existent") ) { // it is _not_existent!
	        // If we came here, there is a servant, so non_existent=false
		resp = req.createResponse(null);
		os = (OutputStream)resp;
		os.write_boolean(false);
	    }
	    else { 
		// Invoke on servant using DSI
		ServerRequestImpl sreq = new ServerRequestImpl(req, orb);

		servant.invoke(sreq);

                // Check if ServerRequestImpl.result() has been called
                Any excany = sreq.checkResultCalled();
    
                if ( excany == null ) { // normal return
		    resp = req.createResponse(null);

                    // Marshal out/inout/return parameters into the ReplyMessage
		    os = (OutputStream)resp;
                    sreq.marshalReplyParams(os);
                }
		else { 
		    // Check if the servant set a SystemException or 
		    // UserException
		    String repId = excany.type().id();
		    if ( Utility.isSystemException(repId) ) {
			// Get the exception object from the Any
			InputStream in = excany.create_input_stream();
			SystemException ex = Utility.readSystemException(in);
			// Marshal the exception back
		        resp = req.createSystemExceptionResponse(ex, null);
		    }
		    else {
		        resp = req.createUserExceptionResponse(null);
		        os = (OutputStream)resp;
		        excany.write_value(os);
		    }
		}
	    }
	}
	catch ( ThreadDeath ex ) {
	    throw ex;
	}
	catch ( Throwable ex ) {

	    SystemException ex2;
	    if ( ex instanceof SystemException ) 
		ex2 = (SystemException)ex;
	    else {
                ex2 = new UNKNOWN(
			    "Unknown Application exception/error on server: "
                            +ex.getClass().getName(),
                            MinorCodes.RUNTIMEEXCEPTION,
                            CompletionStatus.COMPLETED_NO);
	    }
	    resp = req.createSystemExceptionResponse(ex2, null);
        }

	return resp;
    }


    public java.lang.Object createObjref(byte[] userKey, 
					 java.lang.Object servant)
    {
	// Ignore the passed-in userKey since the objmgr generates one
	userKey = transientObjMgr.storeServant(servant, null);

        // Construct the object key
        byte[] objkey = new byte[USERKEY_OFFSET+userKey.length];
        Utility.intToBytes(JAVAMAGIC, objkey, 0);
        Utility.intToBytes(scid, objkey, 4);
        Utility.intToBytes(orb.getTransientServerId(), objkey, SERVERID_OFFSET);
        Utility.intToBytes(userKey.length, objkey, USERKEYLEN_OFFSET);
        for ( int i=0; i<userKey.length; i++ )
            objkey[i+USERKEY_OFFSET] = userKey[i];

	// Get the objref. For JavaIDL the servant is the objref.
        ObjectImpl objRef = (ObjectImpl) servant;

        // Create the IOR
        String[] ids = objRef._ids();
	
	EndPoint endpoint = orb.getServerGIOP().getDefaultEndpoint();
	if (endpoint == null) {
	    orb.getServerGIOP().initEndpoints();
	    endpoint = orb.getServerGIOP().getDefaultEndpoint();
	}
	
	int port = endpoint.getPort();
	String host = endpoint.getHostName();
	
        IOR ior = new IOR(orb, ids[0], host, port, objkey);

        // Create client subcontract (delegate)
        GenericCORBAClientSC clientSC = new GenericCORBAClientSC(ior);
        clientSC.setId(scid);
        clientSC.setOrb(orb);

        // Set the delegate in the objref.
        objRef._set_delegate(clientSC);
        return objRef;
    } 


    /** Create an objref given an IOR. This may be called while an
     *  objref is being unmarshalled. Note: the IOR may have been created by
     *  a foreign ORB.
     */
    public java.lang.Object createObjref(IOR ior)
	throws Exception
    {
	String host = ior.getProfile().getHost();
	byte[] objKey = ior.getProfile().getObjectKey();

	org.omg.CORBA.Object objref = null;

	SubcontractRegistry
	    subcontractRegistry = orb.getSubcontractRegistry();

	try {
	    int magic = Utility.bytesToInt(objKey, 0);
	    int sid = Utility.bytesToInt(objKey, SERVERID_OFFSET);
	    // First check if the ior has same host as myself
	    if ( magic == JAVAMAGIC 
		 && orb.getServerGIOP().isLocalHost(host)
		 && sid == orb.getTransientServerId() )
	    {
		// Now try and get a servant which is the objref
		objref = (org.omg.CORBA.Object)getServant(objKey);
		if ( objref != null )
		    return objref;
	    }
	}
	catch ( Exception ex )  // if any exception, just create default stub
	{}

	// This IOR does not correspond to a servant.
	// Create a stub/objref now.
		
	objref = new CORBAObjectImpl();
	
	// Create a client subcontract and stick it in the stub
	ClientSubcontract sc = subcontractRegistry.getClientSubcontract(objKey);
	sc.unmarshal(ior);
	sc.setOrb(orb);
	ObjectImpl oi = (ObjectImpl) objref;
	oi._set_delegate((org.omg.CORBA.portable.Delegate) sc);

	return objref;
    }

       
    public void destroyObjref(java.lang.Object objref)
    {
        // Get the delegate, then ior, then transientKey, then delete servant
        ObjectImpl oi = (ObjectImpl)objref;
        ClientSubcontract del = (ClientSubcontract)oi._get_delegate();
        IOR ior = del.marshal();
        byte[] objKey = ior.getProfile().getObjectKey();
        int len = Utility.bytesToInt(objKey, USERKEYLEN_OFFSET);
        byte[] transientKey = new byte[len];
        for ( int i=0; i<len; i++ )
            transientKey[i] = objKey[i+USERKEY_OFFSET];
        transientObjMgr.deleteServant(transientKey);
    }

    public java.lang.Object getServant(byte[] objKey)
    {
	// This is called only from createObjref, so no need to check serverid ?
        //int serverId = Utility.bytesToInt(objKey, SERVERID_OFFSET);
	//if ( serverId != orb.getTransientServerId() ) 
	//    return null;

	// Get the userkey and lookup transientObjMgr
        int len = Utility.bytesToInt(objKey, USERKEYLEN_OFFSET);
        byte[] transientKey = new byte[len];
        for ( int i=0; i<len; i++ )
            transientKey[i] = objKey[i+USERKEY_OFFSET];
        return transientObjMgr.lookupServant(transientKey);
    }

    public Class getClientSubcontractClass()
    {
	return com.sun.CORBA.idl.GenericCORBAClientSC.class;
    }
}
