/*
 * @(#)RequestImpl.java	1.43 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.idl;


import org.omg.CORBA.Any;
import org.omg.CORBA.ARG_IN;
import org.omg.CORBA.ARG_OUT;
import org.omg.CORBA.ARG_INOUT;
import org.omg.CORBA.Context;
import org.omg.CORBA.ContextList;
import org.omg.CORBA.Environment;
import org.omg.CORBA.ExceptionList;
import org.omg.CORBA.NVList;
import org.omg.CORBA.NamedValue;
import org.omg.CORBA.Request;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.TypeCodePackage.BadKind;
import org.omg.CORBA.UnknownUserException;
import org.omg.CORBA.Bounds;
import org.omg.CORBA.UNKNOWN;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.NO_IMPLEMENT;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.WrongTransaction;

import org.omg.CORBA.portable.*;

import com.sun.CORBA.*;
import com.sun.CORBA.util.*;


public class RequestImpl extends Request {

    ///////////////////////////////////////////////////////////////////////////
    // data members

    private ObjectImpl		 _target	= null;
    private String       		 _opName	= null;
    private NVList 		 _arguments 	= null;
    private ExceptionList		 _exceptions	= null;
    private NamedValue		 _result  	= null;
    private Environment		 _env		= null;
    private Context		 _ctx		= null;
    private ContextList		 _ctxList	= null;
    private com.sun.CORBA.ORB	 _orb		= null;

    // invocation-specific stuff

    private boolean 		 _isOneWay	= false;
    private int[]			 _paramCodes	= null;
    private long[]		 _paramLongs	= null;
    private java.lang.Object[] 	 _paramObjects	= null;

    // support for deferred invocations. 
    // protected instead of private since it needs to be set by the
    // thread object doing the asynchronous invocation.
    protected boolean 		 gotResponse 	= false;

    ///////////////////////////////////////////////////////////////////////////
    // constructor

    protected RequestImpl(com.sun.CORBA.ORB orb,
			  org.omg.CORBA.Object targetObject,
			  Context ctx,
			  String operationName,
			  NVList argumentList,
			  NamedValue resultContainer,
			  ExceptionList exceptionList,
			  ContextList ctxList)
    {

        // initialize the orb
        _orb 	= orb;

        // initialize target, context and operation name
        _target     = (ObjectImpl) targetObject;
        _ctx	= ctx;
        _opName	= operationName;

        // initialize argument list if not passed in
        if (argumentList == null)
            _arguments = new NVListImpl(_orb);
        else
            _arguments = argumentList;

        // set result container. 
        _result = resultContainer;

        // initialize exception list if not passed in
        if (exceptionList == null)
            _exceptions = new ExceptionListImpl();
        else
            _exceptions = exceptionList;

        // initialize context list if not passed in
        if (ctxList == null)
            _ctxList = new ContextListImpl(_orb);
        else
            _ctxList = ctxList;

        // initialize environment 
        _env	= new EnvironmentImpl();

    }

    public org.omg.CORBA.Object target()
    {
        return _target;
    }

    public String operation() 
    {
        return _opName;
    }

    public NVList arguments() 
    {
        return _arguments;
    }
    
    public NamedValue result() 
    {
        return _result;
    }
    
    public Environment env() 
    {
        return _env;
    }
    
    public ExceptionList exceptions() 
    {
        return _exceptions;
    }
    
    public ContextList contexts() 
    {
        return _ctxList;
    }
    
    public synchronized Context ctx() 
    {
        if (_ctx == null)
            _ctx = new ContextImpl(_orb);
        return _ctx;
    }
    
    public synchronized void ctx(Context newCtx) 
    {
        _ctx = newCtx;
    }

    public synchronized Any add_in_arg()
    {
        return _arguments.add(org.omg.CORBA.ARG_IN.value).value();
    }

    public synchronized Any add_named_in_arg(String name)
    {
        return _arguments.add_item(name, org.omg.CORBA.ARG_IN.value).value();
    }

    public synchronized Any add_inout_arg()
    {
        return _arguments.add(org.omg.CORBA.ARG_INOUT.value).value();
    }

    public synchronized Any add_named_inout_arg(String name)
    {
        return _arguments.add_item(name, org.omg.CORBA.ARG_INOUT.value).value();
    }

    public synchronized Any add_out_arg()
    {
        return _arguments.add(org.omg.CORBA.ARG_OUT.value).value();
    }

    public synchronized Any add_named_out_arg(String name)
    {
        return _arguments.add_item(name, org.omg.CORBA.ARG_OUT.value).value();
    }

    public synchronized void set_return_type(TypeCode tc)
    {
        if (_result == null)
            _result = new NamedValueImpl(_orb);
        _result.value().type(tc);
    }

    public synchronized Any return_value()
    {
        if (_result == null)
            _result = new NamedValueImpl(_orb);
        return _result.value();
    }

    public synchronized void add_exception(TypeCode exceptionType)
    {
        _exceptions.add(exceptionType);
    }
    
    public synchronized void invoke()
    {
        doInvocation();
    }

    public synchronized void send_oneway()
    {
        _isOneWay = true;
        doInvocation();
    }
    
    public synchronized void send_deferred()
    {
        AsynchInvoke invokeObject = new AsynchInvoke((com.sun.CORBA.idl.ORB)_orb, this, false);
        new Thread(invokeObject).start();
    }
    
    public synchronized boolean poll_response()
    {
        // this method has to be synchronized even though it seems
        // "readonly" since the thread object doing the asynchronous
        // invocation can potentially update this variable in parallel.
        // updates are currently simply synchronized againt the request
        // object. 
        return gotResponse;
    }
    
    public synchronized void get_response()
        throws org.omg.CORBA.WrongTransaction
    {
        while (gotResponse == false) {
            // release the lock. wait to be notified by the thread that is
            // doing the asynchronous invocation.
            try {
	        wait();
            } 
	    catch (InterruptedException e) {}
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////
    // private helper methods

    /*
     * The doInvocation operation is where the real mechanics of
     * performing the request invocation is done.
     */

    protected void doInvocation()
    {
        // Get delegate
        ClientSubcontract delegate = (ClientSubcontract)_target._get_delegate();
        ClientRequest req = delegate.createRequest(_opName, _isOneWay);
        OutputStream os = (OutputStream)req;

        // Marshal args
        try {
            for (int i=0; i<_arguments.count() ; i++) {
	        NamedValue nv = _arguments.item(i);
                switch( nv.flags() ) {
                    case ARG_IN.value:
		        nv.value().write_value(os);
		        break;
                    case ARG_OUT.value:
		        break;
                    case ARG_INOUT.value:
		        nv.value().write_value(os);
		        break;
	        }
            }
        } 
	catch ( org.omg.CORBA.Bounds ex ) {
	    // Cannot happen since we only iterate till _arguments.count()
        }


        ClientResponse resp=null;
	Exception exception=null;

	try { // this outer try block ensures we do a req.completeRequest.

	    // Invoke
	    try {
		resp = delegate.invoke(req);
	    } 
	    catch ( SystemException ex ) {
		// set the SystemException in the env and rethrow it.
		_env.exception(ex);
		exception = ex;
		throw ex;
	    }

	    if ( _isOneWay )
		return;

	    // Process reply
	    InputStream is = (InputStream)resp;

	    if ( resp.isSystemException() ) {
		// Unmarshal the SystemException, set it in the env and throw it.
		SystemException e = resp.getSystemException();
		exception = e;
		_env.exception(exception);
		throw e;
	    }
	    else if ( resp.isUserException() ) {
		// Unmarshal exception's repository id
		//String exid = resp.peekUserExceptionId();
		int pos = resp.mark();
		String exid = resp.read_string();
		resp.reset(pos);

		try {
		    // Find the typecode for the exception 
		    for (int i=0; i<_exceptions.count() ; i++) {
			TypeCode tc = _exceptions.item(i);
			if ( tc.id().equals(exid) ) {
			    // Since we dont have the actual user exception 
			    // class, the spec says we have to create an 
			    // UnknownUserException and put it in the 
			    // environment.
			    Any eany = _orb.create_any();
			    eany.read_value(is, (TypeCode)tc);
			    exception = new UnknownUserException(eany);
			    _env.exception(exception);
			    return;
			}
		    }
		} 
		catch ( Exception b ) {
		    // Only exceptions to be caught are Bounds and BadKind.
		    // Both cannot happen here.
		}

		// must be a truly unknown exception
		SystemException u = new UNKNOWN(MinorCodes.UNKNOWN_CORBA_EXC,
				 CompletionStatus.COMPLETED_MAYBE);
		exception = u;
		_env.exception(exception);
		throw u;
	    }
	    else if ( resp.isLocationForward() ) {
		// Location forwards are handled internally in the subcontract.
		throw new INTERNAL(MinorCodes.LOCATIONFORWARD_ERROR,
				   CompletionStatus.COMPLETED_NO);
	    }
	    else { // normal return
		// Unmarshal return args
		unmarshalParams(is);
	    }
	} 
	catch ( SystemException ex ) {
	    throw ex;
	}
	finally {
	    try {
	        delegate.releaseReply(resp, _opName, exception);
	    } catch ( WrongTransaction ex ) {
		// XXX return this for deferred sends
		throw new NO_IMPLEMENT(MinorCodes.SEND_DEFERRED_NOTIMPLEMENTED,
					CompletionStatus.COMPLETED_MAYBE);
	    } catch ( SystemException ex ) {
		// substitute this exception for original exception
		_env.exception(ex);	
		throw ex;
	    }
	}
    }


    private void unmarshalParams(InputStream is)
    {
        // First unmarshal the return value if it is not void
        if ( _result != null ) {
            Any returnAny = _result.value();
            TypeCode returnType = returnAny.type();
            if ( returnType.kind().value() != TCKind._tk_void )
                returnAny.read_value(is, returnType);
        }
        
        // Now unmarshal the out/inout args
        try {
            for ( int i=0; i<_arguments.count() ; i++) {
                NamedValue nv = _arguments.item(i);
                switch( nv.flags() ) {
                    case ARG_IN.value:
                        break;
                    case ARG_OUT.value:
                    case ARG_INOUT.value:
		        Any any = nv.value();	
		        any.read_value(is, any.type());
                        break;
                }
            }
        } 
	catch ( org.omg.CORBA.Bounds ex ) {
	    // Cannot happen since we only iterate till _arguments.count()
        }
    }
}
