/*
 * @(#)TypeCodeImpl.java	1.44 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.idl;

import org.omg.CORBA.*;
import org.omg.CORBA.TypeCodePackage.*;

import org.omg.CORBA.portable.InputStream;
import org.omg.CORBA.portable.OutputStream;

import com.sun.CORBA.*;
import com.sun.CORBA.iiop.CDRInputStream;
import com.sun.CORBA.iiop.CDROutputStream;



public final class TypeCodeImpl extends TypeCode 
{
  
  // the predefined typecode constants

  private static final TypeCodeImpl primitiveConstants[] = {
    new TypeCodeImpl(TCKind._tk_null),			// tk_null      
    new TypeCodeImpl(TCKind._tk_void),			// tk_void      
    new TypeCodeImpl(TCKind._tk_short),			// tk_short     
    new TypeCodeImpl(TCKind._tk_long),			// tk_long      
    new TypeCodeImpl(TCKind._tk_ushort),		// tk_ushort    
    new TypeCodeImpl(TCKind._tk_ulong),			// tk_ulong     
    new TypeCodeImpl(TCKind._tk_float),			// tk_float     
    new TypeCodeImpl(TCKind._tk_double),		// tk_double    
    new TypeCodeImpl(TCKind._tk_boolean),		// tk_boolean   
    new TypeCodeImpl(TCKind._tk_char),			// tk_char      
    new TypeCodeImpl(TCKind._tk_octet),			// tk_octet     
    new TypeCodeImpl(TCKind._tk_any),			// tk_any       
    new TypeCodeImpl(TCKind._tk_TypeCode),		// tk_typecode  
    new TypeCodeImpl(TCKind._tk_Principal),		// tk_principal 
    new TypeCodeImpl(TCKind._tk_objref),		// tk_objref    
    null,						// tk_struct    
    null,						// tk_union     
    null,						// tk_enum      
    new TypeCodeImpl(TCKind._tk_string),		// tk_string    
    null,						// tk_sequence  
    null,						// tk_array     
    null,						// tk_alias     
    null,						// tk_except    
    new TypeCodeImpl(TCKind._tk_longlong),		// tk_longlong  
    new TypeCodeImpl(TCKind._tk_ulonglong),		// tk_ulonglong 
    new TypeCodeImpl(TCKind._tk_longdouble),		// tk_longdouble
    new TypeCodeImpl(TCKind._tk_wchar),			// tk_wchar     
    new TypeCodeImpl(TCKind._tk_wstring)		// tk_wstring   
  };

  // the indirection TCKind, needed for recursive typecodes. 
  protected static final int tk_indirect = 0xFFFFFFFF;
  
  // typecode encodings have three different categories that determine
  // how the encoding should be done.
  
  private static final int EMPTY = 0;	// no parameters
  private static final int SIMPLE = 1;	// simple parameters.
  private static final int COMPLEX = 2; // complex parameters. need to
  					// use CDR encapsulation for
					// parameters 
  
  // a table storing the encoding category for the various typecodes.
  
  private static final int typeTable[] = {
    EMPTY,	// tk_null      
    EMPTY,	// tk_void      
    EMPTY,	// tk_short     
    EMPTY,	// tk_long      
    EMPTY,	// tk_ushort    
    EMPTY,	// tk_ulong     
    EMPTY,	// tk_float     
    EMPTY,	// tk_double    
    EMPTY,	// tk_boolean   
    EMPTY,	// tk_char      
    EMPTY,	// tk_octet     
    EMPTY,	// tk_any       
    EMPTY,	// tk_typecode  
    EMPTY,	// tk_principal 
    COMPLEX,	// tk_objref    
    COMPLEX,	// tk_struct    
    COMPLEX,	// tk_union     
    COMPLEX,	// tk_enum      
    SIMPLE,	// tk_string    
    COMPLEX,	// tk_sequence  
    COMPLEX,	// tk_array     
    COMPLEX,	// tk_alias     
    COMPLEX,	// tk_except    
    EMPTY,	// tk_longlong  
    EMPTY,	// tk_ulonglong 
    EMPTY,	// tk_longdouble
    EMPTY,	// tk_wchar     
    SIMPLE	// tk_wstring   
  };
  

  private int 		_kind		= 0;	// the typecode kind

  // data members for representing the various kinds of typecodes. 

  private String 	_id		= "";	// the typecode repository id
  private String 	_name		= "";	// the typecode name
  private int		_memberCount    = 0;    // member count
  private String 	_memberNames[]  = null; // names of members
  private TypeCodeImpl 	_memberTypes[]  = null; // types of members
  private AnyImpl 	_unionLabels[]  = null; // values of union labels
  private TypeCodeImpl  _discriminator  = null; // union discriminator type
  private int		_defaultIndex   = -1;	// union default index
  private int		_length		= 0;	// string/seq/array length
  private TypeCodeImpl	_contentType	= null;	// seq/array/alias type

  // the ORB instance: may be instanceof ORBSingleton or ORB
  private org.omg.CORBA.ORB _orb; 		

  ///////////////////////////////////////////////////////////////////////////
  // Constructors...

  public TypeCodeImpl(org.omg.CORBA.ORB orb) 
  {
      // initialized to tk_null

      _orb = orb;
  }
  
  public TypeCodeImpl(org.omg.CORBA.ORB orb, TypeCode tc)
    // to handle conversion of "remote" typecodes into "native" style.
    // also see the 'convertToNative(ORB orb, TypeCode tc)' function
  {
    // the orb object
    _orb = orb;

    // set up kind
    _kind 	= tc.kind().value();

    try {
      // set up parameters
      switch (_kind) {
      case TCKind._tk_except:
      case TCKind._tk_struct:
      case TCKind._tk_union:
	// set up member types
	_memberTypes = new TypeCodeImpl[tc.member_count()];
	for (int i=0; i < tc.member_count(); i++)
	  _memberTypes[i] = convertToNative(_orb, tc.member_type(i));
      case TCKind._tk_enum:
	// set up member names
	_memberNames = new String[tc.member_count()];
	for (int i=0; i < tc.member_count(); i++)
	  _memberNames[i] = tc.member_name(i);
	// set up member count
	_memberCount = tc.member_count();
      case TCKind._tk_objref:
      case TCKind._tk_alias:
	_id = tc.id();
	_name = tc.name();
	break;
      }
      
      // set up stuff for unions
      switch (_kind) {
      case TCKind._tk_union:
	_discriminator = convertToNative(_orb, tc.discriminator_type());
	_defaultIndex  = tc.default_index();
	_unionLabels = new AnyImpl[_memberCount];
	for (int i=0; i < _memberCount; i++)
	  _unionLabels[i] = new AnyImpl(_orb, tc.member_label(i));
	break;
      }
      
      // set up length
      switch (_kind) {
      case TCKind._tk_string:
      case TCKind._tk_wstring:
      case TCKind._tk_sequence:
      case TCKind._tk_array:
	_length = tc.length();
      }
      
      // set up content type
      switch (_kind) {
      case TCKind._tk_sequence:
      case TCKind._tk_array:
      case TCKind._tk_alias:
	_contentType = convertToNative(_orb, tc.content_type());
      }
    } catch (org.omg.CORBA.TypeCodePackage.Bounds e) {} catch (BadKind e) {};
    // dont have to worry about these since code ensures we dont step
    // out of bounds.
      
  }
    
  protected TypeCodeImpl(int creationKind)
    // for primitive types
  {  
    // the orb object
    _orb = null;

    // private API. dont bother checking that
    //     (creationKind < 0 || creationKind > typeTable.length)

    _kind = creationKind;

    // do initialization for special cases
    switch (_kind) {
      case TCKind._tk_objref:
	{
	  // this is being used to create typecode for CORBA::Object
	  _id   = "IDL:omg.org/CORBA/Object:1.0";
	  _name = "Object";
	  break;
	}
      case TCKind._tk_string:
      case TCKind._tk_wstring:
	_length =0;
	break;
      }
  }

  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      String id,
		      String name,
		      StructMember[] members)
  // for structs and exceptions
  {
    // the orb object
    _orb = orb;

    if ((creationKind == TCKind._tk_struct) || (creationKind == TCKind._tk_except))
      {
	
	_kind		= creationKind;
	_id		= id;
	_name		= name;
	_memberCount	= members.length;
	
	_memberNames = new String[_memberCount];
	_memberTypes = new TypeCodeImpl[_memberCount];

	for (int i = 0 ; i < _memberCount ; i++) {
	  _memberNames[i] = members[i].name;
	  _memberTypes[i] = convertToNative(_orb, members[i].type);
	}
      } // else initializes to null
  }

  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      String id,
		      String name,
		      TypeCode discriminator_type,
		      UnionMember[] members)
  // for unions
  {
    // the orb object
    _orb = orb;

    if (creationKind == TCKind._tk_union)
      {
	_kind		= creationKind;
	_id		= id;
	_name		= name;
	_memberCount	= members.length;
	_discriminator	= convertToNative(_orb, discriminator_type);

	_memberNames = new String[_memberCount];
	_memberTypes = new TypeCodeImpl[_memberCount];
	_unionLabels = new AnyImpl[_memberCount];

	for (int i = 0 ; i < _memberCount ; i++) {
	  _memberNames[i] = members[i].name;
	  _memberTypes[i] = convertToNative(_orb, members[i].type);
	  _unionLabels[i] = new AnyImpl(_orb, members[i].label);
	  if (_unionLabels[i].type().kind() == TCKind.tk_octet)
	    { 
	      // check whether this is the default branch.
	      // strictly speaking, should also check that the value
	      // of the octet is zero. 
	      _defaultIndex = i;
	    }
	}
      } // else initializes to null
  }
  
  
  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      String id,
		      String name,
		      String[] members)
  // for enums
  {
    // the orb object
    _orb = orb;

    if (creationKind == TCKind._tk_enum)
      {
	_kind		= creationKind;
	_id		= id;
	_name		= name;
	_memberCount	= members.length;

	_memberNames = new String[_memberCount];

	for (int i = 0 ; i < _memberCount ; i++)
	  _memberNames[i] = members[i];
      } // else initializes to null
  }

  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      String id,
		      String name,
		      TypeCode original_type)
  // for aliases
  {
    // the orb object
    _orb = orb;

    if (creationKind == TCKind._tk_alias)
      {
	_kind		= creationKind;
	_id		= id;
	_name		= name;
	_contentType	= convertToNative(_orb, original_type);
      } // else initializes to null
  }

  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      String id,
		      String name)
  {
    // the orb object
    _orb = orb;

    if (creationKind == TCKind._tk_objref)
      {
	_kind		= creationKind;
	_id		= id;
	_name		= name;
      } // else initializes to null
  }

  
  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      int bound)
  // for strings
  {
    // the orb object
    _orb = orb;

    if ((creationKind == TCKind._tk_string) || (creationKind == TCKind._tk_wstring))
      {
	_kind		= creationKind;
	_length		= bound;
      } // else initializes to null
  }

  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      int bound,
		      TypeCode element_type)
  // for sequences and arrays
  {
    // the orb object
    _orb = orb;

    if ((creationKind == TCKind._tk_sequence) || (creationKind == TCKind._tk_array))
      {
	_kind		= creationKind;
	_length		= bound;
	_contentType	= convertToNative(_orb, element_type);
      } // else initializes to null
  }
  
  public TypeCodeImpl(org.omg.CORBA.ORB orb, 
		      int creationKind,
		      int bound,
		      int offset)
  // for recursive sequences
  // [TODO] not implemented yet.
  {
    // the orb object
    _orb = orb;

    throw new NO_IMPLEMENT();
  }

  ///////////////////////////////////////////////////////////////////////////
  // Other creation functions...

  public static TypeCodeImpl get_primitive_tc(TCKind tcKind) 
  {
    return primitiveConstants[tcKind.value()];
  }

  //
  // If we checked for and returned constant primitive typecodes
  // here we could reduce object creation and also enable more
  // efficient typecode comparisons for primitive typecodes.
  //
  protected static TypeCodeImpl convertToNative(org.omg.CORBA.ORB orb, 
						TypeCode tc) 
  {
    if (tc instanceof TypeCodeImpl)
      return (TypeCodeImpl) tc;
    else
      return new TypeCodeImpl(orb, tc);
  }

  ///////////////////////////////////////////////////////////////////////////
  // TypeCode operations

  public final boolean equal(TypeCode tc)
  // [TODO] for all optional names/ids, we might want to check that
  // they are equal in case both are non-nil.
  {

    try {

	// make sure kinds are identical. 
	if (_kind != tc.kind().value()) {
	    return false;
	}

    
    switch (typeTable[_kind]) {
    case EMPTY:
      // no parameters to check.
      return true;
      
    case SIMPLE:
      switch (_kind) {
      case TCKind._tk_string:
      case TCKind._tk_wstring:
	// check for bound.
	if (_length == tc.length())
	  return true;
	else 
	  return false;
      }
      
    case COMPLEX:

      switch(_kind) {

      case TCKind._tk_objref:
	{
	  // check for logical id. 
	  if (_id.compareTo(tc.id()) != 0) {
	    return false;
	  }
	  // ignore name since its optional.
	  return true;
	}

      case TCKind._tk_struct:
	{
	  // check for member count
	  if (_memberCount != tc.member_count())
	    return false;
	  // check for repository id
	  if (_id.compareTo(tc.id()) != 0)
	    return false;
	  // check for member types. 
	  for (int i = 0 ; i < _memberCount ; i++)
	    if (! _memberTypes[i].equal(tc.member_type(i)))
	      return false;
	  // ignore id and names since those are optional.
	  return true;
	}

      case TCKind._tk_union:
	{
	  // check for member count
	  if (_memberCount != tc.member_count())
	    return false;
	  // check for repository id
	  if (_id.compareTo(tc.id()) != 0)
	    return false;
	  // check for default index
	  if (_defaultIndex != tc.default_index())
	    return false;
	  // check for discriminator type
	  if (!_discriminator.equal(tc.discriminator_type()))
	    return false;
	  // check for label types and values
	  for (int i = 0 ; i < _memberCount ; i++)
	    if (! _unionLabels[i].equal(tc.member_label(i)))
	      return false;
	  // check for branch types
	  for (int i = 0 ; i < _memberCount ; i++)
	    if (! _memberTypes[i].equal(tc.member_type(i)))
	      return false;
	  // ignore id and names since those are optional.
	  return true;
	}

      case TCKind._tk_enum:
	{
	  // check for repository id
	  if (_id.compareTo(tc.id()) != 0)
	    return false;
	  // check member count
	  if (_memberCount != tc.member_count())
	    return false;
	  // ignore names since those are optional.
	  return true;
	}

      case TCKind._tk_sequence:
      case TCKind._tk_array:
	{
	  // check bound/length
	    if (_length != tc.length()) {
		return false;
	    }
	  // check content type
	    if (! _contentType.equal(tc.content_type())) {
		return false;
	    }
	  // ignore id and name since those are optional.
	  return true;
	}

      case TCKind._tk_alias:
	{
	  // check for repository id
	    if (_id.compareTo(tc.id()) != 0) {
		return false;
	    }
	  // check for equality with the true type
	  return _contentType.equal(tc.content_type());
	}

      case TCKind._tk_except:
	{
	  // check for member count
	  if (_memberCount != tc.member_count())
	    return false;
	  // check for repository id
	  if (_id.compareTo(tc.id()) != 0)
	    return false;
	  // check for member types. ignore names since those are
	  // optional anyway.
	  for (int i = 0 ; i < _memberCount ; i++)
	    if (! _memberTypes[i].equal(tc.member_type(i)))
	      return false;
	  return true;
	}
      }
    }
    } catch (org.omg.CORBA.TypeCodePackage.Bounds e) {} catch (BadKind e) {}
    // dont have to worry about these since the code ensures these dont
    // arise.
    return false;
  }
  

  public TCKind kind() 
  {
      return TCKind.from_int(_kind);
  }
  
  public String id()
    throws BadKind
  {
    switch (_kind) {
    case TCKind._tk_except:
    case TCKind._tk_objref:
    case TCKind._tk_struct:
    case TCKind._tk_union:
    case TCKind._tk_enum:
    case TCKind._tk_alias:
      // exception and objref typecodes must have a repository id.
      // structs, unions, enums, and aliases may or may not.
      return _id;
    default:
      // all other typecodes throw the BadKind exception.
      throw new BadKind();
    }
  }
  
  public String name()
    throws BadKind
  {
    switch (_kind) {
    case TCKind._tk_except:
    case TCKind._tk_objref:
    case TCKind._tk_struct:
    case TCKind._tk_union:
    case TCKind._tk_enum:
    case TCKind._tk_alias:
      return _name;
    default:
      throw new BadKind();
    }
  }

  public int member_count()
    throws BadKind
  {
    switch (_kind) {
    case TCKind._tk_except:
    case TCKind._tk_struct:
    case TCKind._tk_union:
    case TCKind._tk_enum:
      return _memberCount;
    default:
      throw new BadKind();
    }
  }

  public String member_name(int index)
    throws BadKind, org.omg.CORBA.TypeCodePackage.Bounds
  {
    switch (_kind) {
    case TCKind._tk_except:
    case TCKind._tk_struct:
    case TCKind._tk_union:
    case TCKind._tk_enum:
      try {
	return _memberNames[index];
      } catch (ArrayIndexOutOfBoundsException e) {
	throw new org.omg.CORBA.TypeCodePackage.Bounds();
      };
    default:
      throw new BadKind();
    }
  }

  public TypeCode member_type(int index)
    throws BadKind, org.omg.CORBA.TypeCodePackage.Bounds
  {
    switch (_kind) {
    case TCKind._tk_except:
    case TCKind._tk_struct:
    case TCKind._tk_union:
      try {
	return _memberTypes[index];
      } catch (ArrayIndexOutOfBoundsException e) {
	throw new org.omg.CORBA.TypeCodePackage.Bounds();
      };
    default:
      throw new BadKind();
    }
  }
  
  public Any member_label(int index)
    throws BadKind, org.omg.CORBA.TypeCodePackage.Bounds
  {
    switch (_kind) {
    case TCKind._tk_union:
      try {
	return new AnyImpl(_orb, _unionLabels[index]);
      } catch (ArrayIndexOutOfBoundsException e) {
	throw new org.omg.CORBA.TypeCodePackage.Bounds();
      };
    default:
      throw new BadKind();
    }
  }

  public TypeCode discriminator_type()
    throws BadKind
  {
    switch (_kind) {
    case TCKind._tk_union:
      return _discriminator;
    default:
      throw new BadKind();
    }
  }

  public int default_index()
    throws BadKind
  {
    switch (_kind) {
    case TCKind._tk_union:
      return _defaultIndex;
    default:
      throw new BadKind();
    }
  }

  public int length()
    throws BadKind
  {
    switch (_kind) {
    case TCKind._tk_string:
    case TCKind._tk_wstring:
    case TCKind._tk_sequence:
    case TCKind._tk_array:
      return _length;
    default:
      throw new BadKind();
    }
  }
  
  public TypeCode content_type()
    throws BadKind
  {
    switch (_kind) {
    case TCKind._tk_sequence:
    case TCKind._tk_array:
    case TCKind._tk_alias:
      return _contentType;
    default:
      throw new BadKind();
    }
  }
  
  public void read_value(InputStream s)
  {
    CDRInputStream _encap = null;

    // unmarshal the kind
    _kind = s.read_long();

    // check validity of kind
    if (_kind < 0 || _kind > typeTable.length)
      throw new MARSHAL();
    
    // start unmarshaling the rest of the typecode, based on the
    // encoding (empty, simple or complex).

    switch (typeTable[_kind]) {
    case EMPTY:
      // nothing to unmarshal
      break;
      
    case SIMPLE:
      switch (_kind) {
      case TCKind._tk_string:
      case TCKind._tk_wstring:
	_length = s.read_long();
	break;
      default:
	throw new MARSHAL();
      }
      break;

    case COMPLEX: 
      {
	// pop off the encapsulation length
	int encapLength = s.read_long();

	// read off part of the buffer corresponding to the encapsulation
	byte[] encapBuffer = new byte[encapLength];
	s.read_octet_array(encapBuffer, 0, encapBuffer.length);

	// create an encapsulation using the marshal buffer
	_encap = new CDRInputStream(_orb, encapBuffer, encapBuffer.length);

	// read and set the endianness of the encapsulation
	_encap.consumeEndian();

	switch(_kind) {

	case TCKind._tk_objref:
	  {
	    // get the repository id
	    _id = _encap.read_string();
	    // get the name
	    _name = _encap.read_string();
	  }
	  break;

	case TCKind._tk_struct:
	  {
	    // get the repository id
	    _id = _encap.read_string();

	    // get the name
	    _name = _encap.read_string();

	    // get the number of members
	    _memberCount = _encap.read_long();

	    // create arrays for the names and types of members
	    _memberNames = new String[_memberCount];
	    _memberTypes = new TypeCodeImpl[_memberCount];

	    // read off member names and types
	    for (int i=0; i < _memberCount; i++) {
	      _memberNames[i] = _encap.read_string();
	      _memberTypes[i] = new TypeCodeImpl(_orb);
	      _memberTypes[i].read_value(_encap);
	    }
	  }
	  break;

	case TCKind._tk_union:
	  {
	    // get the repository id
	    _id = _encap.read_string();

	    // get the name
	    _name = _encap.read_string();

	    // discriminant typecode
	    _discriminator = new TypeCodeImpl(_orb);
	    _discriminator.read_value(_encap);

	    // default index
	    _defaultIndex = _encap.read_long();

	    // get the number of members
	    _memberCount = _encap.read_long();

	    // create arrays for the label values, names and types of members
	    _unionLabels = new AnyImpl[_memberCount];
	    _memberNames = new String[_memberCount];
	    _memberTypes = new TypeCodeImpl[_memberCount];

	    // read off label values, names and types
	    for (int i=0; i < _memberCount; i++) {
	      _unionLabels[i] = new AnyImpl(_orb);
	      if (i == _defaultIndex) 
		// for the default case, read off the zero octet
		_unionLabels[i].insert_octet(_encap.read_octet());
	      else {
		switch (_discriminator.kind().value()) {
		case TCKind._tk_short:
		  _unionLabels[i].insert_short(_encap.read_short());
		  break;
		case TCKind._tk_long:
		  _unionLabels[i].insert_long(_encap.read_long());
		  break;
		case TCKind._tk_ushort:
		  _unionLabels[i].insert_ushort(_encap.read_short());
		  break;
		case TCKind._tk_ulong:
		  _unionLabels[i].insert_ulong(_encap.read_long());
		  break;
		case TCKind._tk_float:
		  _unionLabels[i].insert_float(_encap.read_float());
		  break;
		case TCKind._tk_double:
		  _unionLabels[i].insert_double(_encap.read_double());
		  break;
		case TCKind._tk_boolean:
		  _unionLabels[i].insert_boolean(_encap.read_boolean());
		  break;
		case TCKind._tk_char:
		  _unionLabels[i].insert_char(_encap.read_char());
		  break;
		case TCKind._tk_enum:
		  int value = _encap.read_long();
		  org.omg.CORBA.portable.OutputStream out = 
                      _unionLabels[i].create_output_stream();
                  out.write_long(value);
                  _unionLabels[i].read_value(out.create_input_stream(),
                                             _discriminator);
		  break;
		case TCKind._tk_longlong:
		  _unionLabels[i].insert_longlong(_encap.read_longlong());
		  break;
		case TCKind._tk_ulonglong:
		  _unionLabels[i].insert_ulonglong(_encap.read_longlong());
		  break;
		  // [TODO] figure out long double mapping
		  // case TCKind.tk_longdouble:
		  // _unionLabels[i].insert_longdouble(_encap.getDouble());
		  // break;
//		case TCKind._tk_wchar:
//		  _unionLabels[i].insert_wchar(_encap.read_char());
//		  break;
		default:
		  throw new MARSHAL();
		}
	      }
	      _memberNames[i] = _encap.read_string();
	      _memberTypes[i] = new TypeCodeImpl(_orb);
	      _memberTypes[i].read_value(_encap);
	    }
	  }
	  break;

	case TCKind._tk_enum:
	  {
	    // get the repository id
	    _id = _encap.read_string();

	    // get the name
	    _name = _encap.read_string();

	    // get the number of members
	    _memberCount = _encap.read_long();

	    // create arrays for the identifier names
	    _memberNames = new String[_memberCount];

	    // read off identifier names
	    for (int i=0; i < _memberCount; i++)
	      _memberNames[i] = _encap.read_string();
	  }
	  break;

	case TCKind._tk_sequence:
	  {
	    // get the type of the sequence
	    _contentType = new TypeCodeImpl(_orb);
	    _contentType.read_value(_encap);
	    
	    // get the bound on the length of the sequence
	    _length = _encap.read_long();
	  }
	  break;

	case TCKind._tk_array:
	  {
	    // get the type of the array
	    _contentType = new TypeCodeImpl(_orb);
	    _contentType.read_value(_encap);
	    
	    // get the length of the array
	    _length = _encap.read_long();
	  }
	  break;

	case TCKind._tk_alias:
	  {
	    // get the repository id
	    _id = _encap.read_string();

	    // get the name
	    _name = _encap.read_string();

	    // get the type aliased
	    _contentType = new TypeCodeImpl(_orb);
	    _contentType.read_value(_encap);
	  }
	  break;

	case TCKind._tk_except:
	  {
	    // get the repository id
	    _id = _encap.read_string();

	    // get the name
	    _name = _encap.read_string();

	    // get the number of members
	    _memberCount = _encap.read_long();

	    // create arrays for the names and types of members
	    _memberNames = new String[_memberCount];
	    _memberTypes = new TypeCodeImpl[_memberCount];

	    // read off member names and types
	    for (int i=0; i < _memberCount; i++) {
	      _memberNames[i] = _encap.read_string();
	      _memberTypes[i] = new TypeCodeImpl(_orb);
	      _memberTypes[i].read_value(_encap);
	    }
	  }
	  break;
	  
	default:
	  throw new MARSHAL();
	}
	break;
      }
    }
  }
  
  public void write_value(OutputStream s)
  {
    CDROutputStream _encap = null;

    // marshal the kind
    s.write_long(_kind);

    switch (typeTable[_kind]) {
    case EMPTY:
      // nothing more to marshal
      break;
      
    case SIMPLE:
      switch (_kind) {
      case TCKind._tk_string:
      case TCKind._tk_wstring:
	// marshal the bound on string length
	s.write_long(_length);
	break;
      default:
	// unknown typecode kind
	throw new MARSHAL();
      }
      break;
      
    case COMPLEX:
      {
	// create an encapsulation
	_encap = new CDROutputStream(_orb);
	
	// stick the endianness of the encapsulation
	_encap.putEndian();
	
	switch(_kind) {
	  
	case TCKind._tk_objref:
	  {
	    // put the repository id
	    _encap.write_string(_id);
	    
	    // put the name
	    _encap.write_string(_name);
	  }
	  break;
	
	case TCKind._tk_struct:
	  {
	    // put the repository id
	    _encap.write_string(_id);

	    // put the name
	    _encap.write_string(_name);

	    // put the number of members
	    _encap.write_long(_memberCount);

	    // marshal the member names and types
	    for (int i=0; i < _memberCount; i++) {
	      _encap.write_string(_memberNames[i]);
	      _memberTypes[i].write_value(_encap);
	    }
	  }
	  break;

	case TCKind._tk_union:
	  {
	    // put the repository id
	     _encap.write_string(_id);

	    // put the name
	    _encap.write_string(_name);

	    // discriminant typecode
	    _discriminator.write_value(_encap);

	    // default index
	    _encap.write_long(_defaultIndex);

	    // put the number of members
	    _encap.write_long(_memberCount);

	    // marshal label values, names and types
	    for (int i=0; i < _memberCount; i++) {

	      // for the default case, marshal the zero octet
	      if (i == _defaultIndex)
		_encap.write_octet(_unionLabels[i].extract_octet());

	      else {
		switch (_discriminator.kind().value()) {
		case TCKind._tk_short:
		  _encap.write_short(_unionLabels[i].extract_short());
		  break;
		case TCKind._tk_long:
		  _encap.write_long(_unionLabels[i].extract_long());
		  break;
		case TCKind._tk_ushort:
		  _encap.write_short(_unionLabels[i].extract_ushort());
		  break;
		case TCKind._tk_ulong:
		  _encap.write_long(_unionLabels[i].extract_ulong());
		  break;
		case TCKind._tk_float:
		  _encap.write_float(_unionLabels[i].extract_float());
		  break;
		case TCKind._tk_double:
		  _encap.write_double(_unionLabels[i].extract_double());
		  break;
		case TCKind._tk_boolean:
		  _encap.write_boolean(_unionLabels[i].extract_boolean());
		  break;
		case TCKind._tk_char:
		  _encap.write_char(_unionLabels[i].extract_char());
		  break;
		case TCKind._tk_enum:
                  int value = _unionLabels[i].create_input_stream().
                                         read_long();
		  _encap.write_long(value);
                  break;
		case TCKind._tk_longlong:
		  _encap.write_longlong(_unionLabels[i].extract_longlong());
		  break;
		case TCKind._tk_ulonglong:
		  _encap.write_longlong(_unionLabels[i].extract_ulonglong());
		  break;
		  // [TODO] figure out long double mapping
		  // case TCKind.tk_longdouble:
		  // _encap.putDouble(_unionLabels[i].extract_longdouble());
		  // break;
//		case TCKind._tk_wchar:
//		  _encap.write_wchar(_unionLabels[i].extract_wchar());
//		  break;
		default:
		  throw new MARSHAL();
		}
	      }
	      _encap.write_string(_memberNames[i]);
	      _memberTypes[i].write_value(_encap);
	    }
	  }
	  break;

	case TCKind._tk_enum:
	  {
	    // put the repository id
	    _encap.write_string(_id);

	    // put the name
	    _encap.write_string(_name);

	    // put the number of members
	    _encap.write_long(_memberCount);

	    // marshal identifier names
	    for (int i=0; i < _memberCount; i++)
	      _encap.write_string(_memberNames[i]);
	  }
	  break;

	case TCKind._tk_sequence:
	  {
	    // put the type of the sequence
	    _contentType.write_value(_encap);
	    
	    // put the bound on the length of the sequence
	    _encap.write_long(_length);
	  }
	  break;

	case TCKind._tk_array:
	  {
	    // put the type of the array
	    _contentType.write_value(_encap);
	    
	    // put the length of the array
	    _encap.write_long(_length);
	  }
	  break;

	case TCKind._tk_alias:
	  {
	    // put the repository id
	    _encap.write_string(_id);

	    // put the name
	    _encap.write_string(_name);

	    // put the type aliased
	    _contentType.write_value(_encap);
	  }
	  break;

	case TCKind._tk_except:
	  {
	    // put the repository id
	    _encap.write_string(_id);

	    // put the name
	    _encap.write_string(_name);

	    // put the number of members
	    _encap.write_long(_memberCount);

	    // marshal member names and types
	    for (int i=0; i < _memberCount; i++) {
	      _encap.write_string(_memberNames[i]);
	      _memberTypes[i].write_value(_encap);
	    }
	  }
	  break;
	  
	default:
	  throw new MARSHAL();
	}

	// marshal the encapsulation
	_encap.writeOctetSequenceTo(s);
	break;
      }
    }
  }
  

  protected void copy(InputStream src, OutputStream dst)
  {
    switch (_kind) {

    case TCKind._tk_null:
    case TCKind._tk_void:
      break;

    case TCKind._tk_short:
    case TCKind._tk_ushort:
      dst.write_short(src.read_short());
      break;

    case TCKind._tk_long:
    case TCKind._tk_ulong:
      dst.write_long(src.read_long());
      break;

    case TCKind._tk_float:
      dst.write_float(src.read_float());
      break;

    case TCKind._tk_double:
      dst.write_double(src.read_double());
      break;

    case TCKind._tk_longlong:
    case TCKind._tk_ulonglong:
      dst.write_longlong(src.read_longlong());
      break;

    case TCKind._tk_longdouble:
      dst.write_double(src.read_double());
      break;

    case TCKind._tk_boolean:
      dst.write_boolean(src.read_boolean());
      break;
      
    case TCKind._tk_char:
//  case TCKind._tk_wchar:
      dst.write_char(src.read_char());
      break;
      
    case TCKind._tk_octet:
      dst.write_octet(src.read_octet());
      break;

    case TCKind._tk_string:
//  case TCKind._tk_wstring:
      {
	String s;
	s = src.read_string();
	// make sure length bound in typecode is not violated
	if ((_length != 0) && (s.length() > _length))
	  throw new MARSHAL();
	dst.write_string(s);
      }
      break;

    case TCKind._tk_any: 
      {
	Any tmp = new AnyImpl(_orb);
	TypeCodeImpl t = new TypeCodeImpl(_orb);
	t.read_value(src);
	t.write_value(dst);
	tmp.read_value(src, t);
	tmp.write_value(dst);
	break;
      }
    
    case TCKind._tk_TypeCode: 
      {
	dst.write_TypeCode(src.read_TypeCode());
	break;
      }
    
    case TCKind._tk_Principal: 
      {
	dst.write_Principal(src.read_Principal());
	break;
      }

    case TCKind._tk_objref:
      {
	dst.write_Object(src.read_Object());
	break;
      }

    case TCKind._tk_except:
	// Copy repositoryId
	dst.write_string(src.read_string());

	// Fall into ...
    case TCKind._tk_struct:
      {
	// copy each element, using the corresponding member type
	for (int i=0; i < _memberTypes.length; i++) {
	  _memberTypes[i].copy(src, dst);
	}
	break;
      }

    case TCKind._tk_union:
      {
	Any tagValue = new AnyImpl(_orb);

	switch  (_discriminator.kind().value()) {
	case TCKind._tk_short:
	  {
	    short value = src.read_short();
	    tagValue.insert_short(value);
	    dst.write_short(value);
	    break;
	  }
	case TCKind._tk_long:
	  {
	    int value = src.read_long();
	    tagValue.insert_long(value);
	    dst.write_long(value);
	    break;
	  }
	case TCKind._tk_ushort:
	  {
	    short value = src.read_short();
	    tagValue.insert_ushort(value);
	    dst.write_short(value);
	    break;
	  }
	case TCKind._tk_ulong:
	  {
	    int value = src.read_long();
	    tagValue.insert_ulong(value);
	    dst.write_long(value);
	    break;
	  }
	case TCKind._tk_float:
	  {
	    float value = src.read_float();
	    tagValue.insert_float(value);
	    dst.write_float(value);
	    break;
	  }
	case TCKind._tk_double:
	  {
	    double value = src.read_double();
	    tagValue.insert_double(value);
	    dst.write_double(value);
	    break;
	  }
	case TCKind._tk_boolean:
	  {
	    boolean value = src.read_boolean();
	    tagValue.insert_boolean(value);
	    dst.write_boolean(value);
	    break;
	  }
	case TCKind._tk_char:
	  {
	    char value = src.read_char();
	    tagValue.insert_char(value);
	    dst.write_char(value);
	    break;
	  }
	case TCKind._tk_enum:
	  {
	    int value = src.read_long();
	    tagValue.type(_discriminator);
	    org.omg.CORBA.portable.OutputStream out = 
                     tagValue.create_output_stream();
            out.write_long(value);
            tagValue.read_value(out.create_input_stream(), _discriminator);
            dst.write_long(value);
	    break;
	  }
	case TCKind._tk_longlong:
	  {
	    long value = src.read_longlong();
	    tagValue.insert_longlong(value);
	    dst.write_longlong(value);
	    break;
	  }
	case TCKind._tk_ulonglong:
	  {
	    long value = src.read_longlong();
	    tagValue.insert_ulonglong(value);
	    dst.write_longlong(value);
	    break;
	  }
	// [TODO] figure out long double mapping
	// case TCKind.tk_longdouble:
	// {
	// double value = src.read_double();
	//  tagValue.insert_longdouble(value);
	//  dst.putDouble(value);
	//  break;
	//}
//	case TCKind._tk_wchar:
//	  {
//	    char value = src.read_char();
//	    tagValue.insert_wchar(value);
//	    dst.write_char(value);
//	    break;
//	  }
	default:
	  throw new MARSHAL();
	}

	// using the value of the tag, find out the type of the value
	// following. 

	int labelIndex;
	for (labelIndex = 0; labelIndex < _unionLabels.length; labelIndex++) {
	  // use equality over anys
	  if (tagValue.equal(_unionLabels[labelIndex])) {
	    _memberTypes[labelIndex].copy(src, dst);
	    break;
	  }
	}

	if (labelIndex == _unionLabels.length) {
	  // check if label has not been found
	  if (_defaultIndex == -1)
	    // throw exception if default was not expected
	    throw new MARSHAL();
	  else 
	    // must be of the default branch type
	    _memberTypes[_defaultIndex].copy(src, dst);
	}
	break;
      }

    case TCKind._tk_enum:
      dst.write_long(src.read_long());
      break;
      
    case TCKind._tk_sequence:
      // get the length of the sequence
      int seqLength = src.read_long();

      // check for sequence bound violated
      if ((_length != 0) && (seqLength > _length))
	throw new MARSHAL();

      // write the length of the sequence
      dst.write_long(seqLength);

      // copy each element of the seq using content type
      for (int i=0; i < seqLength; i++)
	_contentType.copy(src, dst);
      break;

    case TCKind._tk_array:
      // copy each element of the array using content type
      for (int i=0; i < _length; i++)
	_contentType.copy(src, dst);
      break;

    case TCKind._tk_alias:
      // follow the alias
      _contentType.copy(src, dst);
      break;

      // case _tk_indirect:
      // need to follow offset, get unmarshal typecode from that
      // offset, and use that to do the copy
      // [TODO]
      // throw new NO_IMPLEMENT();
      // break;

    default:
      throw new MARSHAL();
    }
  }
  
}
