/*
 * @(#)CDRInputStream.java	1.9 98/07/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.io.IOException;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.DATA_CONVERSION;
import org.omg.CORBA.BAD_PARAM;
import org.omg.CORBA.Object;
import org.omg.CORBA.Principal;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.Any;

import org.omg.CORBA.portable.ObjectImpl;
import com.sun.CORBA.IOR;
import com.sun.CORBA.SubcontractRegistry;
import com.sun.CORBA.ServerSubcontract;
import com.sun.CORBA.ClientSubcontract;
import com.sun.CORBA.util.*;

import com.sun.CORBA.idl.PrincipalImpl;
import com.sun.CORBA.idl.TypeCodeImpl;


public class CDRInputStream
  extends org.omg.CORBA.portable.InputStream
  implements com.sun.CORBA.MarshalInputStream
{
    protected boolean littleEndian;
    protected int index;
    protected int size;
    protected byte[] buf;
    com.sun.CORBA.ORB orb;

    public CDRInputStream() {
    }

    public CDRInputStream(org.omg.CORBA.ORB orb, byte[] data, int size) {
	this(orb, data, size, false);
    }

    public CDRInputStream(org.omg.CORBA.ORB orb, byte[] data, int size, boolean littleEndian) {
	this.orb = (com.sun.CORBA.ORB)orb;
	this.littleEndian = littleEndian;
	this.size = size;
	buf = data;
    }

    public CDRInputStream(CDRInputStream is) {
	this.littleEndian = is.littleEndian;
	this.index = is.index;
	this.size = is.size;
	this.buf = is.buf;
	this.orb = is.orb;
    }

    // needed by AnyImpl
    public com.sun.CORBA.ORB getORB()
    {
	return orb;
    }

    public CDRInputStream dup() {
        return new CDRInputStream(this);
    }


    private final void alignAndCheck(int align, int n) {
	int cnt;

	cnt = ((index - 1) + align) & (~(align - 1));
	if (cnt + n <= size) {
	    index = cnt;
	    return;
	}

        grow(align, n);
    }


    //
    // This can be overridden....
    //
    protected void grow(int align, int n) {
	throw new MARSHAL();
    }

    public void setOffset(int offset) {
        index = offset;
    }


    //
    // Marshal primitives.
    //

    public final void consumeEndian() {
	littleEndian = read_boolean();
    }

    public final boolean read_boolean()
    {
	boolean result;

	result = (read_octet() != 0)? true: false;
	return result;
    }

    public final char read_char()
    {
        alignAndCheck(1, 1);

        char x = (char)(buf[index++] & 0xFF);
	if ( x > '\u00ff' )
	    throw new MARSHAL(MinorCodes.CHARACTER_OUTOFRANGE,
				  CompletionStatus.COMPLETED_NO);
        return x;
    }

    public final char read_wchar()
    {
	throw new MARSHAL( MinorCodes.UNSUPPORTED_IDLTYPE,
				CompletionStatus.COMPLETED_NO );
    }

    public final byte read_octet()
    {
	alignAndCheck(1, 1);

	return buf[index++];
    }

    public final short read_short()
    {
	int b1, b2;

	alignAndCheck(2, 2);

	if (littleEndian) {
	    b2 = (buf[index++] << 0) & 0x000000FF;
	    b1 = (buf[index++] << 8) & 0x0000FF00;
	} else {
	    b1 = (buf[index++] << 8) & 0x0000FF00;
	    b2 = (buf[index++] << 0) & 0x000000FF;
	}

	return (short)(b1 | b2);
    }

    public final short read_ushort()
    {
	return read_short();
    }

    public final int read_long()
    {
	int b1, b2, b3, b4;

	alignAndCheck(4, 4);

	if (littleEndian) {
	    b4 = (buf[index++] << 0)  & 0x000000FF;
	    b3 = (buf[index++] << 8)  & 0x0000FF00;
	    b2 = (buf[index++] << 16) & 0x00FF0000;
	    b1 = (buf[index++] << 24) & 0xFF000000;
	} else {
	    b1 = (buf[index++] << 24) & 0xFF000000;
	    b2 = (buf[index++] << 16) & 0x00FF0000;
	    b3 = (buf[index++] << 8)  & 0x0000FF00;
	    b4 = (buf[index++] << 0)  & 0x000000FF;
	}

	return (b1 | b2 | b3 | b4);
    }

    public final int read_ulong()
    {
	return read_long();
    }

    public final long read_longlong()
    {
	long i1, i2;

	alignAndCheck(8, 8);

	if (littleEndian) {
	    i2 = read_long() & 0xFFFFFFFFL;
	    i1 = (long)read_long() << 32;
	} else {
	    i1 = (long)read_long() << 32;
	    i2 = read_long() & 0xFFFFFFFFL;
	}

	return (i1 | i2);
    }

    public final long read_ulonglong()
    {
	return read_longlong();
    }

    public final float read_float()
    {
	return Float.intBitsToFloat(read_long());
    }

    public final double read_double()
    {
	return Double.longBitsToDouble(read_longlong());
    }


    public final String read_string()
    {
	byte[] b;
	String result;
	int len;

	len = read_long();

	//
	// Workaround for ORBs which send string lengths of
	// zero to mean empty string.
	//
	if (len == 0)
	    return "";

	b = new byte[len];
	read_octet_array(b,0,len);

	//
	// Strip out the embedded string null termination
	// byte.
	//
	result = new String (b, 0, 0, b.length - 1);

	// Do range checking
	for ( int i=0; i<result.length(); i++ ) {
	    if ( result.charAt(i) > '\u00ff' )
		throw new MARSHAL(MinorCodes.CHARACTER_OUTOFRANGE,
				  CompletionStatus.COMPLETED_MAYBE);
	}

	return (result);
    }

    // This is used ONLY while unmarshalling user exceptions,
    // to peek the repositoryId. Called from ReplyMessage.unmarshal();
    // NOTE : We assume that the entire string can be obtained from
    // the current message buffer. This may have to be changed for
    // IIOP 1.1.
    public final String peek_string()
    {
        byte[] b;
        String result;
        int length;

	// First peek a long for string length
	int b1, b2, b3, b4;
	// Its ok to align: the next read operation MUST be a read_long
	// so the alignment here will be done anyway.
	alignAndCheck(4, 4);
	if (littleEndian) {
	    b4 = (buf[index] << 0)  & 0x000000FF;
	    b3 = (buf[index+1] << 8)  & 0x0000FF00;
	    b2 = (buf[index+2] << 16) & 0x00FF0000;
	    b1 = (buf[index+3] << 24) & 0xFF000000;
	} else {
	    b1 = (buf[index] << 24) & 0xFF000000;
	    b2 = (buf[index+1] << 16) & 0x00FF0000;
	    b3 = (buf[index+2] << 8)  & 0x0000FF00;
	    b4 = (buf[index+3] << 0)  & 0x000000FF;
	}
        length = (b1 | b2 | b3 | b4);

        //
        // Workaround for ORBs which send string lengths of
        // zero to mean empty string.
        //
        if (length == 0)
            return "";

        b = new byte[length];
	int index4 = index + 4;  // the string starts after long

	// peek the string
	alignAndCheck(1, 1);
	int avail = size - index4;
	if ( length > avail ) {
	    throw new INTERNAL(MinorCodes.PEEKSTRING_FAILED,
				CompletionStatus.COMPLETED_MAYBE);
	}
	System.arraycopy(buf, index4, b, 0, length);

        //
        // Strip out the embedded string null termination
        // byte.
        //
        result = new String (b, 0, 0, b.length - 1);

        // Do range checking
        for ( int i=0; i<result.length(); i++ ) {
            if ( result.charAt(i) > '\u00ff' )
                throw new MARSHAL(MinorCodes.CHARACTER_OUTOFRANGE,
                                  CompletionStatus.COMPLETED_MAYBE);
	}

	return result;
    }

    public final String read_wstring()
    {
	throw new MARSHAL( MinorCodes.UNSUPPORTED_IDLTYPE,
				CompletionStatus.COMPLETED_NO );
    }

    public final void read_octet_array(byte[] b, int offset, int length)
    {
	if ( b == null )
	    throw new BAD_PARAM();

	int n = offset;
	while (n < length+offset) {
	    int avail;
	    int bytes;
	    int wanted;

	    alignAndCheck(1, 1);
	    avail = size - index;
	    wanted = (length + offset) - n;
	    bytes = (wanted < avail) ? wanted : avail;
	    System.arraycopy(buf, index, b, n, bytes);
	    index += bytes;
	    n += bytes;
	}
    }

    public final Principal   read_Principal()
    {
	int len = read_long();
	byte[] pvalue = new byte[len];
	read_octet_array(pvalue,0,len);

	Principal p = new PrincipalImpl();
	p.name(pvalue);
	return p;
    }

    public final TypeCode read_TypeCode()
    {
        TypeCodeImpl tc = new TypeCodeImpl(orb);
        tc.read_value(this);
	return tc;
    }

    public final Any read_any()
    {
        // create an Any
        Any any;
        any = orb.create_any();

        // read off the typecode
        TypeCodeImpl tc = new TypeCodeImpl(orb);
        tc.read_value(this);

        // read off the value of the any
        any.read_value(this, tc);

        return any;
    }

    public final org.omg.CORBA.Object read_Object() {
        try {
            IOR ior = new IOR(orb);
            ior.read(this);

            if (ior.is_nil())
	        return null;

	    byte[] objKey = ior.getProfile().getObjectKey();

	    ServerSubcontract sc
		    = orb.getSubcontractRegistry().getServerSubcontract(objKey);

	    return (org.omg.CORBA.Object)sc.createObjref(ior);
	}
	catch ( Exception ex ) {
	    throw new MARSHAL(MinorCodes.READ_OBJECT_EXCEPTION,
			      CompletionStatus.COMPLETED_NO);
	}
    }



    // ------------ End RMI related methods --------------------------

    public final void read_boolean_array(boolean[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_boolean();
	}
    }

    public final void read_char_array(char[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_char();
	}
    }

    public final void read_wchar_array(char[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_wchar();
	}
    }

    public final void read_short_array(short[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_short();
	}
    }

    public final void read_ushort_array(short[] value, int offset, int length) {
	read_short_array(value, offset, length);
    }

    public final void read_long_array(int[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_long();
	}
    }

    public final void read_ulong_array(int[] value, int offset, int length) {
	read_long_array(value, offset, length);
    }

    public final void read_longlong_array(long[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_longlong();
	}
    }

    public final void read_ulonglong_array(long[] value, int offset, int length) {
	read_longlong_array(value, offset, length);
    }

    public final void read_float_array(float[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_float();
	}
    }

    public final void read_double_array(double[] value, int offset, int length) {
	for(int i=0; i < length; i++) {
	    value[i+offset] = read_double();
	}
    }

    //--------------------------------------------------------------------//
    // CDRInputStream state management.
    //

    public final void rewind() {
	index = 0;
    }

    /**
     * Are we at the end of the input stream?
     */
    public final boolean isAtEnd() {
	return index == size;
    }


    // MarshalInputStream methods mark, reset
    public int mark() {
	return index;
    }

    public void reset(int mark) {
	index = mark;
    }
}

