/*
 * @(#)CDROutputStream.java	1.19 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.DATA_CONVERSION;
import org.omg.CORBA.BAD_PARAM;
import org.omg.CORBA.Object;
import org.omg.CORBA.Principal;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.Any;
import org.omg.CORBA.portable.ObjectImpl;
import org.omg.CORBA.portable.OutputStream;

import com.sun.CORBA.IOR;
import com.sun.CORBA.ClientSubcontract;
import com.sun.CORBA.util.*;

import com.sun.CORBA.idl.TypeCodeImpl;


public class CDROutputStream
   extends org.omg.CORBA.portable.OutputStream
   implements com.sun.CORBA.MarshalOutputStream
{
    static final int DEFAULT_BUFFER_SIZE = 1024;
    protected boolean littleEndian;
    protected int size;
    protected byte[] buf;
    protected com.sun.CORBA.ORB orb;

    public CDROutputStream(org.omg.CORBA.ORB orb) {
	this(orb, false);
    }

    public CDROutputStream(org.omg.CORBA.ORB orb, boolean littleEndian) {
	this(orb, littleEndian, DEFAULT_BUFFER_SIZE);
    }

    public CDROutputStream(org.omg.CORBA.ORB orb, boolean littleEndian, int size) {
	this.orb = (com.sun.CORBA.ORB)orb;
	this.littleEndian = littleEndian;
	this.buf = new byte[size];
	this.size = 0;
    }

    public final byte[] toByteArray() {
	byte[] it;

	it = new byte[size];
	System.arraycopy(buf, 0, it, 0, size);

	return it;
    }

    private final void alignAndReserve(int align, int n) {
	int cnt;

	cnt = ((size - 1) + align) & (~(align - 1));

	size = cnt;
	if (cnt + n < buf.length) {
	    return;
	}

        grow(align, n);
    }

    //
    // Default implementation of grow.  Subclassers may override this.
    // Always grow the single buffer. This needs to delegate
    // fragmentation policy for IIOP 1.1.
    //
    protected void grow(int align, int n) {

	if (buf == null) {
	    buf = new byte[DEFAULT_BUFFER_SIZE];
	    return;
        }

	byte[] old = buf;
	buf = new byte[old.length * 2];
	System.arraycopy(old, 0, buf, 0, old.length);
    }

    public final void putEndian() throws SystemException {
	write_boolean(littleEndian);
    }

    public org.omg.CORBA.portable.InputStream create_input_stream()
    {
	return new CDRInputStream(orb, buf, size, littleEndian);
    }

    public final void write_boolean(boolean x)
    {
	write_octet(x? (byte)1:(byte)0);
    }

    public final void write_char(char x)
    {
	alignAndReserve(1, 1);

	if ( x > '\u00ff' )
	    throw new MARSHAL(MinorCodes.CHARACTER_OUTOFRANGE,
				  CompletionStatus.COMPLETED_NO);
	buf[size++] = (byte)((x >>> 0) & 0xFF);
    }

    public final void write_wchar(char value)
    {
	throw new MARSHAL( MinorCodes.UNSUPPORTED_IDLTYPE,
				CompletionStatus.COMPLETED_NO );
    }


    public final void write_octet(byte x)
    {
	alignAndReserve(1, 1);

	buf[size++] = x;
    }

    public final void write_short(short x)
    {
	alignAndReserve(2, 2);

	if (littleEndian) {
	    buf[size++] = (byte)((x >>> 0) & 0xFF);
	    buf[size++] = (byte)((x >>> 8) & 0xFF);
	} else {
	    buf[size++] = (byte)((x >>> 8) & 0xFF);
	    buf[size++] = (byte)((x >>> 0) & 0xFF);
	}
    }

    public final void write_ushort(short x)
    {
	write_short(x);
    }

    public final void write_long(int x)
    {
	alignAndReserve(4, 4);

	if (littleEndian) {
	    buf[size++] = (byte)((x >>> 0) & 0xFF);
	    buf[size++] = (byte)((x >>> 8) & 0xFF);
	    buf[size++] = (byte)((x >>> 16) & 0xFF);
	    buf[size++] = (byte)((x >>> 24) & 0xFF);
	} else {
	    buf[size++] = (byte)((x >>> 24) & 0xFF);
	    buf[size++] = (byte)((x >>> 16) & 0xFF);
	    buf[size++] = (byte)((x >>> 8) & 0xFF);
	    buf[size++] = (byte)((x >>> 0) & 0xFF);
	}
    }

    public final void write_ulong(int x)
    {
	write_long(x);
    }

    public final void write_longlong(long x)
    {
	alignAndReserve(8, 8);

	if (littleEndian) {
	    buf[size++] = (byte)((x >>> 0) & 0xFF);
	    buf[size++] = (byte)((x >>> 8) & 0xFF);
	    buf[size++] = (byte)((x >>> 16) & 0xFF);
	    buf[size++] = (byte)((x >>> 24) & 0xFF);
	    buf[size++] = (byte)((x >>> 32) & 0xFF);
	    buf[size++] = (byte)((x >>> 40) & 0xFF);
	    buf[size++] = (byte)((x >>> 48) & 0xFF);
	    buf[size++] = (byte)((x >>> 56) & 0xFF);
	} else {
	    buf[size++] = (byte)((x >>> 56) & 0xFF);
	    buf[size++] = (byte)((x >>> 48) & 0xFF);
	    buf[size++] = (byte)((x >>> 40) & 0xFF);
	    buf[size++] = (byte)((x >>> 32) & 0xFF);
	    buf[size++] = (byte)((x >>> 24) & 0xFF);
	    buf[size++] = (byte)((x >>> 16) & 0xFF);
	    buf[size++] = (byte)((x >>> 8) & 0xFF);
	    buf[size++] = (byte)((x >>> 0) & 0xFF);
	}
    }

    public final void write_ulonglong(long x)
    {
	write_longlong(x);
    }

    public final void write_float(float x)
    {
	write_long(Float.floatToIntBits(x));
    }

    public final void write_double(double x)
    {
	write_longlong(Double.doubleToLongBits(x));
    }


    public final void write_string(String x)
    {
        if ( x == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	int len = x.length();
	write_long(len + 1);
	for (int i=0; i < len; i++) {
	    write_char(x.charAt(i));
	}

	//
	// Send the string null termination byte.
	//
	write_octet((byte)0);
    }

    public final void write_wstring(String value)
    {
	throw new MARSHAL( MinorCodes.UNSUPPORTED_IDLTYPE,
				CompletionStatus.COMPLETED_NO );
    }

    public final void write_octet_array(byte b[], int offset, int length)
    {
        if ( b == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	int n = offset;

	while (n < length+offset) {
	    int avail;
	    int bytes;
	    int wanted;

	    alignAndReserve(1, 1);
	    avail = buf.length - size;
	    wanted = (length + offset) - n;
	    bytes = (wanted < avail) ? wanted : avail;
	    System.arraycopy(b, n, buf, size, bytes);
	    size += bytes;
	    n += bytes;
	}
    }

    public final void write_Principal(Principal p)
    {
	write_long(p.name().length);
	write_octet_array(p.name(), 0, p.name().length);
    }

    public final void write_any(Any any)
    {
        if ( any == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	write_TypeCode(any.type());
	any.write_value(this);
    }

    public final void write_TypeCode(TypeCode tc)
    {
        if ( tc == null )
            throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

        TypeCodeImpl tci;
        if (tc instanceof TypeCodeImpl)
	    tci = (TypeCodeImpl)tc;
        else
	    tci = new TypeCodeImpl(orb, tc);
        tci.write_value(this);
    }

    public final void write_Object(org.omg.CORBA.Object ref)
    {
        if (ref == null) {
            new IOR(orb, "", null).write(this);
            return;
        }

        ObjectImpl oi = (ObjectImpl)ref;
	ClientSubcontract rep = null;
        try {
	  // throws BAD_OPERATION if no delegate
          rep = (ClientSubcontract) oi._get_delegate();
        } catch ( Exception ex ) {}
        if ( rep == null ) {
          // no delegate: servant was not connected to ORB
	  orb.connect(ref);
	  rep = (ClientSubcontract) oi._get_delegate();
        }

        IOR ior = rep.marshal();
	ior.write(this);
	return;
    }

    public final void write_boolean_array(boolean[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_boolean(value[i+offset]);
	}
    }

    public final void write_char_array(char[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_char(value[i+offset]);
	}
    }

    public final void write_wchar_array(char[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_wchar(value[i+offset]);
	}
    }

    public final void write_short_array(short[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_short(value[i+offset]);
	}
    }

    public final void write_ushort_array(short[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	write_short_array(value, offset, length);
    }

    public final void write_long_array(int[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_long(value[i+offset]);
	}
    }

    public final void write_ulong_array(int[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	write_long_array(value, offset, length);
    }

    public final void write_longlong_array(long[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_longlong(value[i+offset]);
	}
    }

    public final void write_ulonglong_array(long[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	write_longlong_array(value, offset, length);
    }

    public final void write_float_array(float[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_float(value[i+offset]);
	}
    }

    public final void write_double_array(double[]value, int offset, int length) {
        if ( value == null )
	    throw new BAD_PARAM(MinorCodes.NULL_PARAM, CompletionStatus.COMPLETED_MAYBE);

	for(int i=0; i < length; i++) {
	    write_double(value[i+offset]);
	}
    }


    //--------------------------------------------------------------------//
    // CDROutputStream state management.
    //

    public final void
    reset()
    {
	size = 0;
    }

    /**
     * Write the contents of the CDROutputStream to the specified output stream.
     * @param s The output stream to write to.
     */
    public void writeTo(java.io.OutputStream s) throws java.io.IOException {
	s.write(buf, 0, size);
    }

    public void writeOctetSequenceTo(org.omg.CORBA.portable.OutputStream s) {
	s.write_long(size);
	s.write_octet_array(buf, 0, size);
    }

    public final int getSize() {
	return size;
    }
}

