/*
 * @(#)ClientRequestImpl.java	1.5 98/07/09
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.COMM_FAILURE;

import com.sun.CORBA.Request;
import com.sun.CORBA.ClientRequest;
import com.sun.CORBA.ServiceContext;
import com.sun.CORBA.ClientResponse;
import com.sun.CORBA.IOR;
import com.sun.CORBA.util.MinorCodes;


class ClientRequestImpl extends IIOPOutputStream implements ClientRequest {

    ClientRequestImpl(byte[] objectKey, String operationName,
		      boolean oneway, ServiceContext[] svc, 
		      int requestId, Connection conn) 
    {
	super(conn.getORB(), conn);

	this.isOneway = oneway;
	boolean responseExpected = !isOneway;
	this.request = new RequestMessage(svc, requestId, responseExpected,
					  objectKey, operationName,
					  null);
	request.write(this);
	setMessage(request);

	// mark beginning of msg body for possible later use
	bodyBegin = size;
    }

    public int getRequestId() {
	return request.getRequestId();
    }

    public boolean isOneWay() {
	return isOneway;
    }

    public ServiceContext[] getServiceContextList() {
	return request.getServiceContextList();
    }

    public String getOperationName() {
	return request.getOperation();
    }

    public byte[] getObjectKey() {
	return request.getObjectKey();
    }

    public ClientResponse invoke() 
    {
        return (ClientResponse)super.invoke(isOneway);
    }

    /**
     * Resend the invocation to the new IOR, after a location forward.
     */
    public ClientResponse reInvoke(IOR[] iorHolder)
    {
	IOR ior = iorHolder[0];
	int[] bodyBeginHolder = new int[1];
	bodyBeginHolder[0] = bodyBegin;
	ClientResponse resp;
	while (true) { // Keep invoking if there are multiple location forwards
	    resp = (ClientResponse) super.reInvoke(ior, bodyBeginHolder);
	    if ( !resp.isLocationForward() )
		break;
	    ior = resp.getForwardedIOR();
	}
	iorHolder[0] = ior;
	return resp;
    }


    private RequestMessage request;
    private int bodyBegin;
    private boolean isOneway;
}
