/*
 * @(#)Connection.java	1.36 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.net.InetAddress;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.OBJECT_NOT_EXIST;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.INTERNAL;

import org.omg.CORBA.portable.*;

import com.sun.CORBA.ServerGIOP;
import com.sun.CORBA.IOR;
import com.sun.CORBA.util.*;

/**
 * Common connection base class.
 */
abstract public class Connection {

    // Connection close states
    static final int CONN_ABORT = 1;
    static final int CONN_REBIND = 2;

    protected boolean debug = false;
    protected ORB orb;
    protected Socket socket;    // The socket used for this connection.
    protected long timeStamp = 0;
    protected boolean isServer = false;
    protected ConnectionTable connectionTable = null;

    public void setDebug(boolean enable) {
        debug = enable;
    }

    void dprint(String msg) {
	Utility.dprint(this, msg);
    }

    public ORB getORB() {
        return orb;
    }

    Socket getSocket() {
	return socket;
    }


    public static Socket newSocket(ORB orb, String host, int port)
	throws IOException
    {
	return new Socket(host, port);
    }

    abstract public IIOPInputStream invoke(IIOPOutputStream s)
	throws SystemException;

    abstract public void delete();

    abstract public java.io.InputStream getInputStream();

    abstract ServerGIOP getServerGIOP();

    abstract public IIOPInputStream send(IIOPOutputStream s, boolean oneWay);

    abstract public void sendReply(IIOPOutputStream s) throws Exception;

    abstract public IIOPInputStream createInputStream() throws Exception;


    // All of Following abstract required for Connection cleanup

    abstract public void cleanUp() throws java.lang.Exception;

    abstract public boolean isBusy();

    abstract public void requestBegins();
    abstract public void requestEnds();

    abstract public void print();

    abstract public void setConnection(Socket _socket, ConnectionTable ctab)
                    throws java.lang.Exception;
    abstract public void abortConnection();

    public IOR locate(int id, byte [] key)
    {
        LocateRequestMessage msg;
        IIOPOutputStream os;
        IIOPInputStream is;

        msg = new LocateRequestMessage(id, key);
        os = orb.newOutputStream(this);
        msg.write(os);
        os.setMessage(msg);

        is = send(os, false);

        LocateReplyMessage reply;
        reply = (LocateReplyMessage)is.getMessage();
        switch (reply.getRequestStatus()) {
          case Message.GIOPLocateReplyUnknownObject:
            throw new OBJECT_NOT_EXIST( MinorCodes.LOCATE_UNKNOWN_OBJECT,
                CompletionStatus.COMPLETED_NO );

          case Message.GIOPLocateReplyObjectHere:
            return null;

          case Message.GIOPLocateReplyObjectForward: {
            IOR ior;

            ior = new IOR(orb);
            ior.read(is);

            return ior;
          }
        }

        throw new INTERNAL( MinorCodes.BAD_LOCATE_REQUEST_STATUS,
            		    CompletionStatus.COMPLETED_NO );
    }

    public static String getLocalHost() {
        try {
            return InetAddress.getLocalHost().getHostName();
        }
        catch (Exception ex) {
            throw new INTERNAL( MinorCodes.GET_LOCAL_HOST_FAILED,
				CompletionStatus.COMPLETED_NO );
        }
    }

    public void stampTime() {
	connectionTable.stampTime(this);
    }

}
