/*
 * @(#)GIOPImpl.java	1.13 98/07/09
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.net.*;
import java.util.*;

import com.sun.CORBA.ClientGIOP;
import com.sun.CORBA.ServerGIOP;
import com.sun.CORBA.IOR;
import com.sun.CORBA.RequestHandler;
import com.sun.CORBA.ServiceContext;
import com.sun.CORBA.ClientRequest;
import com.sun.CORBA.EndPoint;
import com.sun.CORBA.Profile;

import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.CompletionStatus;

import com.sun.CORBA.util.*;


public class GIOPImpl implements ClientGIOP, ServerGIOP {

    // Start at some value other than zero since this is a magic
    // value in some protocols.
    private int requestId = 5;

    synchronized int getNextRequestId() {
	return requestId++;
    }

    private ConnectionTable table;
    private ORB orb;

    public GIOPImpl(ORB orb, RequestHandler handler) {
	this.orb = orb;
	this.table = new ConnectionTable(orb, this);
	setRequestHandler(handler);
    }

    public ConnectionTable getConnectionTable() {
	return table;
    }

    public ClientRequest createRequest(IOR ior, String operationName,
				       boolean oneway,
				       ServiceContext[] svc)
    {
	return createRequest(ior, operationName, oneway,
				     svc, getNextRequestId());
    }

    /**
     * Prepare an invocation request with the given requestId.
     */
    public ClientRequest createRequest(IOR ior, String operationName,
                                       boolean oneway,
                                       ServiceContext[] svc, int id)
    {
	Profile iop = ior.getProfile();
	String host = iop.getHost();
	int    port = iop.getPort();
	byte[]  key = iop.getObjectKey();

	Connection c = table.get(host, port);

	return new ClientRequestImpl(key, operationName, oneway,
				     svc, id, c);
    }

    /**
     * Allocate a new request id.
     */
    public int allocateRequestId()
    {
	return getNextRequestId();
    }


    private RequestHandler handler;

    public void setRequestHandler(RequestHandler handler) {
	this.handler = handler;
    }

    public RequestHandler getRequestHandler() {
	return handler;
    }

    public IOR locate(IOR ior)
    {
	Profile iop = ior.getProfile();
	String host = iop.getHost();
	int    port = iop.getPort();
	byte[]  objKey = iop.getObjectKey();

	Connection c = table.get(host, port);
	IOR newIOR = c.locate(getNextRequestId(), objKey);
	if ( newIOR == null ) // we got an OBJECT_HERE reply
	    newIOR = ior;
	return newIOR;
    }


    public boolean isLocalHost(String hostname) {
	return
	    hostname.equals(orb.getORBServerHost()) ||
	    hostname.equals(Connection.getLocalHost());
    }


/****************************************************************************
  The following methods are implementations of the ServerGIOP's endpoint APIs.
 ****************************************************************************/

    private int listenerPort;
    private Vector endPoints = new Vector();
    private ListenerThread listenerThread;
    private EndPoint bootstrapEndpoint;


    public synchronized void initEndpoints() {
	if (endPoints.size() == 0) // create a default endpoint
	    getEndpoint(EndPoint.CLEAR_TEXT, 0, null);
    }

    public synchronized EndPoint getDefaultEndpoint() {
	if (endPoints.size() == 0)
	    return null;
	return (EndPoint) endPoints.elementAt(0);
    }

    /**
     * Get an EndPoint for the specified type, listenPort and local address.
     * type must be one of CLEAR_TEXT or SSL.
     * If listenPort == 0, a listening port will be assigned.
     * If addr == null, InetAddress.getLocalHost() is used.
     * If an EndPoint at the specified port/address does not exist,
     * then it will be created.
     */
    public synchronized EndPoint getEndpoint(int type, int port,
				             InetAddress addr)
    {
	// Check if this endpoint was already created
	Enumeration eps = endPoints.elements();
	while ( eps.hasMoreElements() ) {
	    EndPoint ep = (EndPoint)eps.nextElement();
	    if ( ep.getType() == type && ep.getPort() == port
		 		      && ep.getInetAddress() == addr )
		return ep;
	}

	if (port == 0 && orb.getORBServerPort() != 0)
	    port = orb.getORBServerPort();

	// Create the endpoint
	listenerThread = createListener(port);
	ServerSocket listenSocket = listenerThread.getSocket();
	listenerPort = listenSocket.getLocalPort();
	EndPoint ep = new EndPointImpl(type, listenerPort, addr,
				       orb.getORBServerHost());
	endPoints.addElement(ep);

	return ep;
    }

    /**
     * Get an EndPoint for the bootstrap naming service, at the specified port.
     * If listenPort == 0 and this is the first time getBootstrapEndpoint was
     * invoked, then an EndPoint at the default port of 900 will be created.
     * If listenPort == 0 and this is not the first time getBootstrapEndpoint
     * was called, then the existing bootstrap EndPoint will be returned.
     */
    public synchronized EndPoint getBootstrapEndpoint(int port)
    {
	if (bootstrapEndpoint == null)
	    bootstrapEndpoint = getEndpoint(EndPoint.CLEAR_TEXT, port, null);

	return bootstrapEndpoint;
    }

    private synchronized ListenerThread createListener(int port) {

	ServerSocket ss;
        ListenerThread lis;
        try {
            ss = new java.net.ServerSocket(port);
            lis = new ListenerThread(table, ss);
            lis.setDaemon(true);
            lis.start();
        }
        catch (java.lang.Exception e) {
            throw new INTERNAL(MinorCodes.CREATE_LISTENER_FAILED,
			       CompletionStatus.COMPLETED_NO);
        }
        return lis;
    }
}
