/*
 * @(#)IIOPInputStream.java	1.17 98/07/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.io.IOException;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.MARSHAL;
import org.omg.CORBA.DATA_CONVERSION;
import org.omg.CORBA.BAD_PARAM;
import org.omg.CORBA.Object;
import org.omg.CORBA.Principal;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.Any;
import org.omg.CORBA.CompletionStatus;
import com.sun.CORBA.util.MinorCodes;


public class IIOPInputStream extends CDRInputStream
{
    private static final int DEFAULT_BUFFER_SIZE = 1024;
    protected Connection conn;
    protected Message msg;


    public IIOPInputStream() {
    }

    public IIOPInputStream(ORB orb, Connection c) {
	super(orb, null, 0);
	this.conn = c;
    }

    public IIOPInputStream(ORB orb, byte[] msgbuf, int msgsize, boolean endian,
			   Message msg)
    {
	super(orb, msgbuf, msgsize, endian);
	this.msg = msg;
    }

    public IIOPInputStream(Connection c, byte[] msgbuf, Message header)
        throws java.io.IOException
    {
        this(c.getORB(), c);

        this.buf = msgbuf;
        this.msg = header;

        this.index = Message.GIOPMessageHeaderLength;
        this.littleEndian = msg.isLittleEndian();

        size = msg.getSize();
        if (size > buf.length) {
            byte[] old;

            old = buf;
            buf = new byte[size];
            System.arraycopy(old, 0, buf, 0, old.length);
        }

        Message.readFully(c.getInputStream(), buf, index, size - index);

        // Unmarshal the extended GIOP message from the buffer.
        msg.read(this);

        c.stampTime();
    }


    public final Connection getConnection() {
        return conn;
    }

    public final void setConnection(Connection c) {
        this.conn = c;
	orb = c.getORB();
    }

    public final Message getMessage() {
        return msg;
    }

    //
    // We ran out of things to read. We pre-read everything
    // off of the java.io.InputStream in IIOPConnection.createInputStream()
    // so this can't happen normally.
    //
    protected void grow(int align, int n) {
	throw new MARSHAL(MinorCodes.IIOPINPUTSTREAM_GROW,
			  CompletionStatus.COMPLETED_MAYBE);
    }
}

