/*
 * @(#)ListenerThread.java	1.20 98/07/09
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.ServerSocket;

public final class ListenerThread extends Thread 
{
    private static final int MAX_CLEANUP_RETRIES=5;
    private ServerSocket serverSocket;
    private ConnectionTable connectionTable;

    ListenerThread(ConnectionTable connectionTable,
		   ThreadGroup g, ServerSocket serverSocket) 
    {
	super(g, "JavaIDL Listener");
	this.serverSocket = serverSocket;
	this.connectionTable = connectionTable;
    }

    ListenerThread(ConnectionTable connectionTable, ServerSocket serverSocket) 
    {
	super("JavaIDL Listener");
	this.serverSocket = serverSocket;
	this.connectionTable = connectionTable;
    }

    public ServerSocket getSocket()
    {
	return serverSocket;
    }

    public void run() 
    {
        int tries = 0;

	// Loop forever listening for incoming connections.
	while (true) {
	    try {
		// Accept an incoming connection request
	        Socket socket = serverSocket.accept();

		// Create a Connection instance and process incoming request
		Connection conn = connectionTable.get(socket);

                tries = 0;

		// Check if #connections > HighWaterMark, if so cleanup.
		// This prevents incoming connections from being rejected
		// because of TCP/IP or OS network connection limits.
                connectionTable.checkConnectionTable();
            }
            catch (SocketException ex) {
		// An exception was thrown by the accept() call.
		// This probably means that we have run out of network 
		// connections. 

                if (tries == MAX_CLEANUP_RETRIES) {
		    // We tried closing existing connections but that didnt help
		    // So it must be some unknown networking problem.
                    continue;
		}

		// Clean up some existing connections.
                if (!connectionTable.cleanUp()) {
		    // Could not close any connections :-(
                    continue;
		}

                tries++;
            }
            catch (Exception ex) { }
	}
    }
}


