/*
 * @(#)ORB.java	1.156 98/07/13
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

// Import JDK stuff
import java.net.*;
import java.util.*;


// Import our stuff
import org.omg.CORBA.SystemException;
import org.omg.CORBA.OBJ_ADAPTER;
import org.omg.CORBA.CompletionStatus;

import com.sun.CORBA.ClientSubcontract;
import com.sun.CORBA.ServerSubcontract;
import com.sun.CORBA.SubcontractRegistry;
import com.sun.CORBA.ClientGIOP;
import com.sun.CORBA.ServerGIOP;
import com.sun.CORBA.MarshalInputStream;
import com.sun.CORBA.MarshalOutputStream;
import com.sun.CORBA.RequestHandler;
import com.sun.CORBA.ServerResponse;
import com.sun.CORBA.ServerRequest;

import com.sun.CORBA.IOR;
import com.sun.CORBA.util.*;

/**
 * The JavaIDL IIOP ORB implementation.
 */
public class ORB extends com.sun.CORBA.idl.ORB implements RequestHandler {
    // The following fields form our special little collection of global state.
    // We keep it bottled up here in the ORB class and attach a reference to
    // ourselves to every object reference that we create and handle.
    protected SubcontractRegistry subcontractRegistry;
    protected GIOPImpl giopTransport;

    // This is the unique id of this server (JVM). Multiple incarnations
    // of this server will get different ids.
    protected int transientServerId=0;

    // The thread group of the main thread (for applications) or applet.
    ThreadGroup threadGroup;

    // Synchronization variable for shutdown
    // XXX Is this necessary here?
    protected java.lang.Object shutdownObj = new java.lang.Object();

    /**
     * Create a new ORB. Should only be invoked from the
     * CORBA ORB superclass. Should be followed by the appropriate
     * set_parameters() call.
     */
    public ORB()
    {
        //
        // We attempt to create new threads in this thread group, if
        // possible. This avoids problems if the application/applet
        // creates a thread group, makes JavaIDL calls which create a new
        // connection and ReaderThread, and then destroys the thread
        // group. If our ReaderThread were part of this destroyed thread
        // group then it might get killed and cause other invoking threads
        // sharing the same connection to get a non-restartable
        // CommunicationFailure. We'd like to avoid that.
        //
        // Our solution is to create all of our threads in the same
        // thread group that we were initialized under.
        //
        threadGroup = Thread.currentThread().getThreadGroup();

        // Compute transientServerId = (milliseconds since Jan 1, 1970)/10.
	// Note: transientServerId will wrap in about 2^32 / 8640000 = 497 days.
        // Randomize time before currentTimeMillis() call to prevent multiple
        // processes on same (multiprocessor) machine from getting same serverID
        long rndVal = new java.util.Random().nextLong();
        int x = 0;
        while ( rndVal > 0 ) {
	    if ((rndVal & 1)!=0)
	        x++;
	    rndVal >>= 1;
        }
        x = x*100;
        while ( x > 0 )
            x--;
        transientServerId = (int)(System.currentTimeMillis()/10);

        subcontractRegistry
	    = new SubcontractRegistry(SubcontractList.defaultSubcontract);

	// Register subcontracts in subcontractRegistry

	// register the server subcontract *instance*

	for(int i = 0; i < SubcontractList.subcontracts.length; i++) {
	    try {
		Class serverClass =
		    SubcontractList.subcontracts[i].serverSC;
		ServerSubcontract sc = (ServerSubcontract)
		    serverClass.newInstance();
		sc.setOrb(this);
		subcontractRegistry.registerServer(sc,
						   SubcontractList.subcontracts[i].id);
                // register GenericCORBAServerSC as the default bootstrap SC
                // so that SC lookups with objkey = "INIT" will work
		// The BootstrapServer will override this with its own SC.
                // See bugid 4155455
                if ( serverClass == com.sun.CORBA.idl.GenericCORBAServerSC.class )
                     subcontractRegistry.registerBootstrapServer(sc);

		Class clientClass =
		    SubcontractList.subcontracts[i].clientSC;
		if (clientClass != null)
		    subcontractRegistry.registerClient(clientClass,
		        SubcontractList.subcontracts[i].id);
	    } catch (Exception ex) { }
	}

	// Initialize the GIOP transport
	giopTransport = new GIOPImpl(this, this);
    }

    public ClientGIOP getClientGIOP()
    {
	return giopTransport;
    }

    public ServerGIOP getServerGIOP()
    {
	return giopTransport;
    }

    public MarshalInputStream newInputStream() {
	return new CDRInputStream(this, new byte[1024], 1024);
    }

    public MarshalInputStream newInputStream(byte[] buffer, int size) {
	return new CDRInputStream(this, buffer, size);
    }

    public MarshalOutputStream newOutputStream() {
	return new CDROutputStream(this);
    }

    IIOPOutputStream newOutputStream(Connection c) {
	return new IIOPOutputStream(this, c);
    }



    public int getTransientServerId()
    {
        return transientServerId;
    }

    public SubcontractRegistry getSubcontractRegistry()
    {
        return subcontractRegistry;
    }

    public ServerResponse process(ServerRequest request) {
	ServerSubcontract sc =
	    subcontractRegistry.getServerSubcontract(request.getObjectKey());
	if (sc == null) {
	    SystemException ex =
		new OBJ_ADAPTER(MinorCodes.NO_SERVER_SC_IN_DISPATCH,
				CompletionStatus.COMPLETED_NO);

	    return request.createSystemExceptionResponse(ex, null);
	}
	return sc.dispatch(request);
    }

    public IOR locate(byte[] key) {
	ServerSubcontract sc =
	    subcontractRegistry.getServerSubcontract(key);
	if (sc == null)
	    // XXX What to do here??
	    return null;
	return sc.locate(key);
    }

} // Class ORB
