/*
 * @(#)ReplyMessage.java	1.30 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import org.omg.CORBA.SystemException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.portable.*;

import com.sun.CORBA.ServiceContext;
import com.sun.CORBA.IOR;
import com.sun.CORBA.util.*;

public final class ReplyMessage extends Message {
    private ServiceContext[] ctx;
    private int status;
    private String exceptionId;
    private String exClassName;
    private int minorCode;
    private CompletionStatus completionStatus;
    private IOR ior;

    public static final int NO_EXCEPTION = 0;
    public static final int USER_EXCEPTION = 1;
    public static final int SYSTEM_EXCEPTION = 2;
    public static final int LOCATION_FORWARD = 3;

    public ReplyMessage() {
    }

    public ReplyMessage(ServiceContext[] ctx, int requestId, int status) {
	super();
	setType(Message.GIOPReply);
	this.ctx = ctx;
	this.requestId = requestId;
	this.status = status;
    }

    public ServiceContext[] getServiceContextList() {
	return ctx;
    }

    public int getReplyStatus() {
	return status;
    }

    public String getExceptionId() {
	return exceptionId;
    }

    public int getMinorCode() {
	return minorCode;
    }

    public IOR getIOR() {
	return ior;
    }

    public SystemException getSystemException()
    {
	SystemException sysEx;
	try {
	    sysEx = (SystemException)Class.forName(exClassName).newInstance();
        }
        catch ( Exception someEx ) {
            throw new INTERNAL("BAD SystemException: " + exClassName,
                                0, CompletionStatus.COMPLETED_MAYBE);
        }
	sysEx.minor = minorCode;
	sysEx.completed = completionStatus;

	return sysEx;
    }

    public void write(OutputStream s) throws SystemException {
	//
	// Handle nulls for Service Context and
	// Principal in case someone doesn't set these
	// to meaning values yet (we don't care about
	// them typically).
	//
	super.write(s);
	if (ctx != null) {
	    s.write_long(ctx.length);
	    for (int i=0; i < ctx.length; i++) {
		ctx[i].write(s);
	    }
	} else
            s.write_long(0);

	s.write_long(requestId);
	s.write_long(status);
    }

    public void read(InputStream s) throws SystemException {
	ctx = new ServiceContext[s.read_long()];
	for (int i=0; i < ctx.length; i++) {
	    ctx[i] = new ServiceContext();
	    ctx[i].read(s);
	}
	requestId = s.read_long();
	status = s.read_long();

	if (status == SYSTEM_EXCEPTION) {
            String reposId = s.read_string();
            exClassName = Utility.classNameOf(reposId);
            minorCode = s.read_long();
            int status = s.read_long();
            switch ( status ) {
                case CompletionStatus._COMPLETED_YES:
                    completionStatus = CompletionStatus.COMPLETED_YES;
                    break;
                case CompletionStatus._COMPLETED_NO:
                    completionStatus = CompletionStatus.COMPLETED_NO;
                    break;
                case CompletionStatus._COMPLETED_MAYBE:
                    completionStatus = CompletionStatus.COMPLETED_MAYBE;
                    break;
                default:
                    throw new INTERNAL("BAD completion status: " + status,
                                        0,
                                        CompletionStatus.COMPLETED_MAYBE);
            }
	} else if (status == USER_EXCEPTION) {
	    // NOTE: at this point we should only PEEK the exceptionID
            // since it is actually read by exceptionHolder.__read().
	    IIOPInputStream iis = (IIOPInputStream)s;
	    exceptionId = iis.peek_string();
	} else if (status == LOCATION_FORWARD) {
	    CDRInputStream cdr = (CDRInputStream) s;
	    ior = new IOR(cdr.orb);
	    ior.read(s);
	}
    }
}
