/*
 * @(#)RequestMessage.java	1.31 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import org.omg.CORBA.Principal;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.portable.*;

import com.sun.CORBA.ServiceContext;

public final class RequestMessage extends Message {
    private ServiceContext[] ctx;
    private boolean responseExpected;
    private byte[] objectKey;
    private String operation;
    private Principal principal;

    public RequestMessage() {
    }

    //
    // Create an instance with initialized instance variables
    // and a filled in request header in the buffer ready to be
    // appended to.
    //
    public RequestMessage(ServiceContext[] ctx, int requestId,
	boolean response, byte[] objectKey, String operation,
	Principal principal)
    {
	setType(Message.GIOPRequest);
	this.ctx = ctx;
	this.requestId = requestId;
	responseExpected = response;
	this.objectKey = objectKey;
	this.operation = operation;
	this.principal = principal;
    }

    //
    // The common case, with no ServiceContext or Principal.
    //
    public RequestMessage(int requestId, boolean response, byte[] objectKey,
      String operation)
    {
	this((ServiceContext[])null, requestId, response, objectKey,
	  operation, (Principal)null);
    }

    public byte[] getObjectKey() {
	return objectKey;
    }

    public ServiceContext[] getServiceContextList() {
	return ctx;
    }

    public String getOperation() {
	return operation;
    }

    public boolean isResponseExpected() {
	return responseExpected;
    }

    public void write(OutputStream s) throws SystemException {
	//
	// Handle nulls for Service Context and
	// Principal in case someone doesn't set these
	// to meaning values yet (we don't care about
	// them typically).
	//
	super.write(s);
	if (ctx != null) {
	    s.write_long(ctx.length);
	    for (int i=0; i < ctx.length; i++) {
		ctx[i].write(s);
	    }
	} else
            s.write_long(0);

	s.write_long(requestId);
	s.write_boolean(responseExpected);
        s.write_long(objectKey.length);
	s.write_octet_array(objectKey,0,objectKey.length);
	s.write_string(operation);

	if (principal != null)
	    s.write_Principal(principal);
	else
	    s.write_long(0);
    }

    public void read(InputStream s) throws SystemException {
	ctx = new ServiceContext[s.read_long()];
	for (int i=0; i < ctx.length; i++) {
	    ctx[i] = new ServiceContext();
	    ctx[i].read(s);
	}
	requestId = s.read_long();
	responseExpected = s.read_boolean();
	int len = s.read_long();
	objectKey = new byte[len];
	s.read_octet_array(objectKey,0,len);
	operation = s.read_string();
	principal = s.read_Principal();
    }



    public void realignAndWrite(OutputStream os, byte[] newObjKey,
				boolean alignChangeNeeded)
    {
	// Note: Since the last component of the RequestMessage header
	// is the Principal which is always null, we will always end on
	// a 4-byte or 8-byte alignment.

	if ( newObjKey.length == objectKey.length || !alignChangeNeeded ) {
	    // this is the common case. e.g. all IDL servers do not change
	    // object key lengths while forwarding IORs.
	    // Also, alignChangeNeeded is true only if the message contains
	    // doubles / longlong / fixed.
	    objectKey = newObjKey;
	    write(os);
	}
	else {
	    // Find if the new marshaled msg header will end on a
	    // different alignment.

	    // alignment = 8-byte if len = 8*x + 5, 6, 7, 8
	    // alignment = 4-byte if len = 8*x + 1, 2, 3, 4

	    // First find current alignment. (Assume length != 0)
	    boolean currentAlign8;
	    if ( ((objectKey.length-1) % 8) >= 4 )
		currentAlign8 = true;
	    else
		currentAlign8 = false;

	    // Find new alignment.
	    boolean newAlign8;
	    if ( ((newObjKey.length-1) % 8) >= 4 )
		newAlign8 = true;
	    else
		newAlign8 = false;

	    if ( currentAlign8 != newAlign8 ) {
		// We will have alignment problems.
		// Add 4 bytes to end of ServiceContext to make alignment of
		// RequestMessage-end the same again.
		if ( ctx == null )
		    ctx = new ServiceContext[1];
		else {
		    ServiceContext[] newctx = new ServiceContext[ctx.length+1];
		    for ( int i=0; i<ctx.length; i++ )
			newctx[i] = ctx[i];
		    ctx = newctx;
		}
		byte[] data = new byte[4];
		ctx[ctx.length-1] = new ServiceContext(JAVAIDL_ALIGN_SERVICE_ID,
						       data);
	    }

	    objectKey = newObjKey;
            write(os);
	}
    }

    private static final int JAVAIDL_ALIGN_SERVICE_ID = 0xbe1345cd;
}
