/*
 * @(#)ServerResponseImpl.java	1.6 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import org.omg.CORBA.SystemException;

import com.sun.CORBA.ServerResponse;
import com.sun.CORBA.IOR;
import com.sun.CORBA.ServiceContext;

class ServerResponseImpl extends IIOPOutputStream implements ServerResponse
{
    ServerResponseImpl(ServerRequestImpl request, ServiceContext[] svc)
    {
	this(request,
	     new ReplyMessage(svc, request.getRequestId(),
			      ReplyMessage.NO_EXCEPTION),
	     null);
    }

    ServerResponseImpl(ServerRequestImpl request, ServiceContext[] svc,
		       boolean user)
    {
	this(request,
	     new ReplyMessage(svc, request.getRequestId(),
			      user ? ReplyMessage.USER_EXCEPTION :
			             ReplyMessage.SYSTEM_EXCEPTION),
	     null);
    }

    ServerResponseImpl(ServerRequestImpl request, ReplyMessage reply, IOR ior)
    {
	super(request.getConnection().getORB(), request.getConnection());
	reply.write(this);
	setMessage(reply);
	if (ior != null)
	    ior.write(this);

	this.reply = reply;
	this.ior = ior;
    }

    public boolean isSystemException() {
	if (reply != null)
	    return reply.getReplyStatus() == ReplyMessage.SYSTEM_EXCEPTION;
	return false;
    }

    public boolean isUserException() {
	if (reply != null)
	    return reply.getReplyStatus() == ReplyMessage.USER_EXCEPTION;
	return false;
    }

    public boolean isLocationForward() {
	if (ior != null)
	    return true;
	return false;
    }

    public IOR getForwardedIOR() {
	return ior;
    }

    public int getRequestId() {
	if (reply != null)
	    return reply.getRequestId();
	return -1;
    }

    public ServiceContext[] getServiceContextList() {
	if (reply != null)
	    return reply.getServiceContextList();
	return null;
    }

    public SystemException getSystemException() {
	if (reply != null)
	    return reply.getSystemException();
	return null;
    }

    private ReplyMessage reply;
    private IOR ior; // forwarded IOR
}
