/*
 * @(#)WorkerThread.java	1.12 98/07/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.iiop;

import java.io.*;
import java.net.*;

import org.omg.CORBA.SystemException;
import org.omg.CORBA.UNKNOWN;
import org.omg.CORBA.CompletionStatus;

import com.sun.CORBA.ServerGIOP;
import com.sun.CORBA.RequestHandler;
import com.sun.CORBA.IOR;
import com.sun.CORBA.ServerRequest;

import com.sun.CORBA.util.*;



/** This is a generic worker thread that dispatches a GIOP message
    to the ORB and sends a reply back.
    It may be used by multiple protocols (e.g. IIOP and HTTP)
 */

public class WorkerThread extends ORBThread {
    private Connection conn=null;
    private IIOPInputStream request = null;
    private IIOPOutputStream response = null;
    private int msgType;
    private RequestHandler handler;

    /** This constructor is called when the mb has already been filled
	by a reader-thread. E.g. for IIOP-Connections. */
    public WorkerThread(RequestHandler handler,
			Connection c, IIOPInputStream is) {
	this.conn = c;
	this.request = is;
	this.handler = handler;
    }

    public void run()
    {
	Message msg = request.getMessage();

        conn.requestBegins();

	switch ( msg.getType() ) {
        case Message.GIOPRequest: {
	    ServerRequest srequest = (ServerRequest)request;
	    try {
	        // Here is the actual dispatch to the ORB which
	        // dispatches to the subcontract which dispatches to the
		// skeleton which finally dispatches to the user's servant.

		response = (IIOPOutputStream)handler.process(srequest);

		if (srequest.isOneWay()) {
                    conn.requestEnds();
	            Thread.currentThread().stop();
	        }
            } catch (SystemException ex) {
	        // Caught a system exception marshal it back to the client
	        //
	        // ex.printStackTrace();
                // conn.dprint("run: got exception: " + ex);
                try {
		    response = (IIOPOutputStream)
			srequest.createSystemExceptionResponse(ex, null);
                } catch (Exception e) {
                    conn.requestEnds();
	            return; // marshaling exception error; return;
                }
    
	    } catch (ThreadDeath ex) {
		throw ex;

	    } catch (Throwable ex) {
	        //
	        // Caught an unknown Exception or Error, create a 
		// SystemException and marshal it back to the client
	        //

                // conn.dprint("run: got exception: " + ex);
                // ex.printStackTrace();
                SystemException exc = new UNKNOWN(MinorCodes.UNKNOWN_SERVER_ERROR,
						  CompletionStatus.COMPLETED_MAYBE);
                try {
		    response = (IIOPOutputStream)
			srequest.createSystemExceptionResponse(exc, null);
                } catch (Exception e) {
                    conn.requestEnds();
                    return; // marshaling exception error; return;
                }
            }
	    break;
	}
        case Message.GIOPLocateRequest: {
	    LocateRequestMessage lmsg = (LocateRequestMessage)msg;
	    LocateReplyMessage reply;
	    int requestId = msg.getRequestId();

	    IOR ior = null;
	    try {
		ior = handler.locate(lmsg.getObjectKey());

		if ( ior == null ) {
		    reply = new LocateReplyMessage(requestId,
					Message.GIOPLocateReplyObjectHere);
		}
		else {
		    reply = new LocateReplyMessage(requestId,
					Message.GIOPLocateReplyObjectForward);
	 	}
	    }
	    catch ( Exception ex ) {
		// This handles OBJECT_NOT_EXIST exceptions thrown in
		// the subcontract or obj manager. Send back UNKNOWN_OBJECT.
		reply = new LocateReplyMessage(requestId,
				     Message.GIOPLocateReplyUnknownObject);
	    }
	    response = new IIOPOutputStream(request.getConnection().getORB(),
				    	    request.getConnection());
	    reply.write(response);
	    response.setMessage(reply);
	    if (ior != null)
		ior.write(response);
	    break;
	}
        case Message.GIOPCancelRequest: {
            // Do nothing for now (actually legal IIOP).
            // Just return so the WorkerThread ends.
            conn.requestEnds();
            return;
        }

        default:
            //
            // Any other message type is unexpected and is a fatal
            // error for this connection ??
            //
            conn.requestEnds();
	    return;

	} // end switch (msgType)


	// Send the reply on the wire
	// Only RequestMessages and LocateRequestMessages have replies
        try {
	    conn.sendReply(response);
            conn.requestEnds();
        }
        catch(Exception e) {
            conn.requestEnds();
	    return;
        }
    }
}

