/*
 * @(#)HexOutputStream.java	1.11 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CORBA.util;

import java.io.OutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;

/**
 * Writes each input byte as a 2 byte hexidecimal output pair making it
 * possible to turn arbitrary binary data into an ASCII format.
 * The high 4 bits of the byte is translated into the first byte.
 *
 * @author	Jeff Nisewanger
 */
public class HexOutputStream extends FilterOutputStream
{
    static private final char hex[] = {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
    };


    /**
     * Creates a new HexOutputStream.
     * @param in 	The input stream.
     */
    public
    HexOutputStream(OutputStream out) {
        super(out);
    }


    /**
     * Writes a byte. Will block until the byte is actually
     * written.
     * param b The byte to write out.
     * @exception java.io.IOException I/O error occurred.
     */
    public synchronized void write(int b) throws IOException {
	out.write(hex[((b >> 4) & 0xF)]);
	out.write(hex[((b >> 0) & 0xF)]);
    }

    public synchronized void write(byte[] b) throws IOException {
	write(b, 0, b.length);
    }

    public synchronized void write(byte[] b, int off, int len)
    throws IOException
    {
	for(int i=0; i < len; i++) {
	    write(b[off + i]);
	}
    }
}

