/*
 * @(#)BindingIteratorImpl.java	1.26 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CosNaming;

// Import general CORBA classes
import org.omg.CORBA.ORB;
import org.omg.CORBA.Object;

// Import org.omg.CosNaming classes
import org.omg.CosNaming.Binding;
import org.omg.CosNaming.BindingType;
import org.omg.CosNaming.BindingHolder;
import org.omg.CosNaming.BindingListHolder;
import org.omg.CosNaming.BindingIteratorHolder;
import org.omg.CosNaming._BindingIteratorImplBase;

/**
 * Class BindingIteratorImpl implements the org.omg.CosNaming::BindingIterator
 * interface, but does not implement the method to retrieve the next
 * binding in the NamingContext for which it was created. This is left
 * to a subclass, which is why this class is abstract; BindingIteratorImpl
 * provides an implementation of the interface operations on top of two
 * subclass methods, allowing multiple implementations of iterators that
 * differ in storage and access to the contents of a NamingContext
 * implementation.
 * <p>
 * The operation next_one() is implemented by the subclass, whereas
 * next_n() is implemented on top of the next_one() implementation.
 * Destroy must also be implemented by the subclass.
 * <p>
 * A subclass must implement NextOne() and Destroy(); these
 * methods are invoked from synchronized methods and need therefore
 * not be synchronized themselves.
 */
public abstract class BindingIteratorImpl extends _BindingIteratorImplBase
{
  /**
   * Create a binding iterator servant.
   * runs the super constructor.
   * @param orb an ORB object.
   * @exception java.lang.Exception a Java exception.
   */
  public BindingIteratorImpl(org.omg.CORBA.ORB orb) throws java.lang.Exception {
    super();
  }

  /**
   * Return the next binding. It also returns true or false, indicating
   * whether there were more bindings.
   * @param b The Binding as an out parameter.
   * @return true if there were more bindings.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see NextOne
   */
  public synchronized boolean next_one(org.omg.CosNaming.BindingHolder b)
  {
    // NextOne actually returns the next one
    return NextOne(b);
  }

  /**
   * Return the next n bindings. It also returns true or false, indicating
   * whether there were more bindings.
   * @param how_many The number of requested bindings in the BindingList.
   * @param bl The BindingList as an out parameter.
   * @return true if there were more bindings.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see NextOne
   */
  public synchronized boolean next_n(int how_many, org.omg.CosNaming.BindingListHolder blh)
  {
    // Take the smallest of what's left and what's being asked for
    int numberToGet = Math.min(RemainingElements(),how_many);

    // Create a resulting BindingList
    Binding[] bl = new Binding[numberToGet];
    BindingHolder bh = new BindingHolder();
    int i = 0;
    // Keep iterating as long as there are entries
    while (i < numberToGet && this.NextOne(bh) == true) {
      bl[i] = bh.value;
      i++;
    }
    // Found any at all?
    if (i == 0) {
      // No
      blh.value = new Binding[0];
      return false;
    }

    // Set into holder
    blh.value = bl;

    return true;
  }

  /**
   * Destroy this BindingIterator object. The object corresponding to this
   * object reference is destroyed.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see Destroy
   */
  public synchronized void destroy()
  {
    // Destroy actually destroys
    this.Destroy();
  }

  /**
   * Abstract method for returning the next binding in the NamingContext
   * for which this BindingIterator was created.
   * @param b The Binding as an out parameter.
   * @return true if there were more bindings.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   */
  protected abstract boolean NextOne(org.omg.CosNaming.BindingHolder b);

  /**
   * Abstract method for destroying this BindingIterator.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   */
  protected abstract void Destroy();

  /**
   * Abstract method for returning the remaining number of elements.
   * @return the remaining number of elements in the iterator.
   */
  protected abstract int RemainingElements();
}
