/*
 * @(#)InternalBindingKey.java	1.9 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CosNaming;

import org.omg.CosNaming.NameComponent;

/**
 * Class InternalBindingKey implements the necessary wrapper code
 * around the org.omg.CosNaming::NameComponent class to implement the proper
 * equals() method and the hashCode() method for use in a hash table.
 * It computes the hashCode once and stores it, and also precomputes
 * the lengths of the id and kind strings for faster comparison.
 */
public class InternalBindingKey
{
  // A key contains a name
  public NameComponent name;
  public int idLen;
  public int kindLen;
  public int hashVal;

  // Default Constructor
  public InternalBindingKey() {}

  // Normal constructor
  public InternalBindingKey(NameComponent n)
  {
    setup(n);
  }

  // Setup the object
  protected void setup(NameComponent n) {
    this.name = n;
    // Precompute lengths and values since they will not change
    idLen = this.name.id.length();
    kindLen = this.name.kind.length();
    hashVal = 0;
    if (idLen > 0)
      hashVal += this.name.id.hashCode();
    if (kindLen > 0)
      hashVal += this.name.kind.hashCode();
  }

  // Compare the keys by comparing name's id and kind
  public boolean equals(java.lang.Object o) {
    if (o == null)
      return false;
    if (o instanceof InternalBindingKey) {
      InternalBindingKey that = (InternalBindingKey)o;
      // Both lengths must match
      if (this.idLen != that.idLen || this.kindLen != that.kindLen) {
	return false;
      }
      // If id is set is must be equal
      if (this.idLen > 0 && this.name.id.equals(that.name.id) == false) {
	return false;
      }
      // If kind is set it must be equal
      if (this.kindLen > 0 && this.name.kind.equals(that.name.kind) == false) {
	return false;
      }
      // Must be the same
      return true;
    } else {
      return false;
    }
  }
  // Return precomputed value
  public int hashCode() {
    return this.hashVal;
  }
}

