/*
 * @(#)NamingContextImpl.java	1.45 98/07/09
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CosNaming;

// Import general CORBA classes
import org.omg.CORBA.ORB;
import org.omg.CORBA.Object;
import org.omg.CORBA.BAD_PARAM;
import org.omg.CORBA.INTERNAL;
import org.omg.CORBA.CompletionStatus;

// Import org.omg.CosNaming classes
import org.omg.CosNaming.BindingType;
import org.omg.CosNaming.BindingTypeHolder;
import org.omg.CosNaming.BindingListHolder;
import org.omg.CosNaming.BindingIteratorHolder;
import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContextHelper;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextPackage.*;
import org.omg.CosNaming._NamingContextImplBase;

import com.sun.CosNaming.NamingContextDataStore;

import com.sun.CosNaming.MinorCodes;

/**
 * Class NamingContextImpl implements the org.omg.CosNaming::NamingContext
 * interface, but does not implement the methods associated with
 * maintaining the "table" of current bindings in a NamingContext.
 * Instead, this implementation assumes that the derived implementation
 * implements the NamingContextDataStore interface, which has the necessary
 * methods. This allows multiple
 * NamingContext implementations that differ in storage of the bindings,
 * as well as implementations of interfaces derived from
 * CosNaming::NamingContext that still reuses the implementation.
 * <p>
 * The operations bind(), rebind(), bind_context() and rebind_context()
 * are all really implemented by doBind(). resolve() is really implemented
 * by doResolve(), unbind() by doUnbind(). list(), new_context() and
 * destroy() uses the NamingContextDataStore interface directly. All the
 * doX() methods are public static.
 * They synchronize on the NamingContextDataStore object.
 * <p>
 * An implementation a NamingContext must extend this class and implement
 * the NamingContextDataStore interface with the operations:
 * Bind(), Resolve(),
 * Unbind(), List(), NewContext() and Destroy(). Calls
 * to these methods are synchronized; these methods should
 * therefore not be synchronized.
 */
public abstract class NamingContextImpl extends _NamingContextImplBase
{
  /**
   * Create a naming context servant.
   * Runs the super constructor.
   * @param orb an ORB object.
   * @exception java.lang.Exception a Java exception.
   */
  public NamingContextImpl(org.omg.CORBA.ORB orb) throws java.lang.Exception {
    super();
    this.orb = orb;
  }

  /**
   * Bind an object under a name in this NamingContext. If the name
   * contains multiple (n) components, n-1 will be resolved in this
   * NamingContext and the object bound in resulting NamingContext.
   * An exception is thrown if a binding with the supplied name already
   * exists. If the
   * object to be bound is a NamingContext it will not participate in
   * a recursive resolve.
   * @param n a sequence of NameComponents which is the name under which
   * the object will be bound.
   * @param obj the object reference to be bound.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CosNaming.NamingContextPackage.AlreadyBound An object is
   * already bound under the supplied name.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see doBind
   */
  public void bind(NameComponent[] n, org.omg.CORBA.Object obj)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
    org.omg.CosNaming.NamingContextPackage.CannotProceed,
    org.omg.CosNaming.NamingContextPackage.InvalidName,
    org.omg.CosNaming.NamingContextPackage.AlreadyBound
  {
    if (debug)
      dprint("bind " + nameToString(n) + " to " + obj);
    // doBind implements all four flavors of binding
    NamingContextDataStore impl = (NamingContextDataStore)this;
    doBind(impl,n,obj,false,BindingType.nobject);
  }

  /**
   * Bind a NamingContext under a name in this NamingContext. If the name
   * contains multiple (n) components, n-1 will be resolved in this
   * NamingContext and the object bound in resulting NamingContext.
   * An exception is thrown if a binding with the supplied name already
   * exists. The NamingContext will participate in recursive resolving.
   * @param n a sequence of NameComponents which is the name under which
   * the object will be bound.
   * @param obj the NamingContect object reference to be bound.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CosNaming.NamingContextPackage.AlreadyBound An object is
   * already bound under the supplied name.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see doBind
   */
  public void bind_context(NameComponent[] n, NamingContext nc)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
      org.omg.CosNaming.NamingContextPackage.CannotProceed,
      org.omg.CosNaming.NamingContextPackage.InvalidName,
      org.omg.CosNaming.NamingContextPackage.AlreadyBound
  {
    if (debug)
      dprint("bind_context " + nameToString(n) + " to " + nc);
    // doBind implements all four flavors of binding
    NamingContextDataStore impl = (NamingContextDataStore)this;
    doBind(impl,n,nc,false,BindingType.ncontext);
  }

  /**
   * Bind an object under a name in this NamingContext. If the name
   * contains multiple (n) components, n-1 will be resolved in this
   * NamingContext and the object bound in resulting NamingContext.
   * If a binding under the supplied name already exists it will be
   * unbound first. If the
   * object to be bound is a NamingContext it will not participate in
   * a recursive resolve.
   * @param n a sequence of NameComponents which is the name under which
   * the object will be bound.
   * @param obj the object reference to be bound.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see doBind
   */
  public  void rebind(NameComponent[] n, org.omg.CORBA.Object obj)
    throws       org.omg.CosNaming.NamingContextPackage.NotFound,
      org.omg.CosNaming.NamingContextPackage.CannotProceed,
      org.omg.CosNaming.NamingContextPackage.InvalidName
  {
    try {
      if (debug)
	dprint("rebind " + nameToString(n) + " to " + obj);
      // doBind implements all four flavors of binding
      NamingContextDataStore impl = (NamingContextDataStore)this;
      doBind(impl,n,obj,true,BindingType.nobject);
    } catch (org.omg.CosNaming.NamingContextPackage.AlreadyBound ex) {
      // This should not happen
      throw new org.omg.CORBA.INTERNAL(MinorCodes.NAMING_CTX_REBIND_ALREADY_BOUND,
				    CompletionStatus.COMPLETED_NO);
    }
  }

  /**
   * Bind a NamingContext under a name in this NamingContext. If the name
   * contains multiple (n) components, the first n-1 components will be
   * resolved in this
   * NamingContext and the object bound in resulting NamingContext.
   * If a binding under the supplied name already exists it will be
   * unbound first. The NamingContext will participate in recursive resolving.
   * @param n a sequence of NameComponents which is the name under which
   * the object will be bound.
   * @param obj the object reference to be bound.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see doBind
   */
  public  void rebind_context(NameComponent[] n, NamingContext nc)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
      org.omg.CosNaming.NamingContextPackage.CannotProceed,
      org.omg.CosNaming.NamingContextPackage.InvalidName
  {
    try {
      if (debug)
	dprint("rebind_context " + nameToString(n) + " to " + nc);
      // doBind implements all four flavors of binding
      NamingContextDataStore impl = (NamingContextDataStore)this;
      doBind(impl,n,nc,true,BindingType.ncontext);
    } catch (org.omg.CosNaming.NamingContextPackage.AlreadyBound ex) {
      // This should not happen
      throw new org.omg.CORBA.INTERNAL(MinorCodes.NAMING_CTX_REBINDCTX_ALREADY_BOUND,
				    CompletionStatus.COMPLETED_NO);
    }
  }

  /**
   * Resolve a name in this NamingContext and return the object reference
   * bound to the name. If the name contains multiple (n) components,
   * the first component will be resolved in this NamingContext and the
   * remaining components resolved in the resulting NamingContext, provided
   * that the NamingContext bound to the first component of the name was
   * bound with bind_context().
   * @param n a sequence of NameComponents which is the name to be resolved.
   * @return the object reference bound under the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see doResolve
   */
  public  org.omg.CORBA.Object resolve(NameComponent[] n)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
      org.omg.CosNaming.NamingContextPackage.CannotProceed,
      org.omg.CosNaming.NamingContextPackage.InvalidName
  {
    if (debug)
      dprint("resolve " + nameToString(n));
    // doResolve actually resolves
    NamingContextDataStore impl = (NamingContextDataStore)this;
    return doResolve(impl,n);
  }

  /**
   * Remove a binding from this NamingContext. If the name contains
   * multiple (n) components, the first n-1 components will be resolved
   * from this NamingContext and the final component unbound in
   * the resulting NamingContext.
   * @param n a sequence of NameComponents which is the name to be unbound.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see doUnbind
   */
  public  void unbind(NameComponent[] n)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
      org.omg.CosNaming.NamingContextPackage.CannotProceed,
      org.omg.CosNaming.NamingContextPackage.InvalidName
  {
    if (debug)
      dprint("unbind " + nameToString(n));
    // doUnbind actually unbinds
    NamingContextDataStore impl = (NamingContextDataStore)this;
    doUnbind(impl,n);
  }

  /**
   * List the contents of this NamingContest. A sequence of bindings
   * is returned (a BindingList) containing up to the number of requested
   * bindings, and a BindingIterator object reference is returned for
   * iterating over the remaining bindings.
   * @param how_many The number of requested bindings in the BindingList.
   * @param bl The BindingList as an out parameter.
   * @param bi The BindingIterator as an out parameter.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see BindingListHolder
   * @see BindingIteratorImpl
   */
  public  void list(int how_many, BindingListHolder bl, BindingIteratorHolder bi)
  {
    if (debug)
      dprint("list(" + how_many + ")");
    // List actually generates the list
    NamingContextDataStore impl = (NamingContextDataStore)this;
    synchronized (impl) {
      impl.List(how_many,bl,bi);
    }
    if (debug && bl.value != null)
      dprint("list(" + how_many + ") -> bindings[" + bl.value.length +
	     "] + iterator: " + bi.value);
  }

  /**
   * Create a NamingContext object and return its object reference.
   * @return an object reference for a new NamingContext object implemented
   * by this Name Server.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   */
  public synchronized NamingContext new_context()
  {
    // Create actually creates a new naming context
    if (debug)
      dprint("new_context()");
    NamingContextDataStore impl = (NamingContextDataStore)this;
    synchronized (impl) {
      return impl.NewContext();
    }
  }

  /**
   * Create a new NamingContext, bind it in this Naming Context and return
   * its object reference. This is equivalent to using new_context() followed
   * by bind_context() with the supplied name and the object reference for
   * the newly created NamingContext.
   * @param n a sequence of NameComponents which is the name to be unbound.
   * @return an object reference for a new NamingContext object implemented
   * by this Name Server, bound to the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.AlreadyBound An object is
   * already bound under the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see new_context
   * @see bind_context
   */
  public  NamingContext bind_new_context(NameComponent[] n)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
      org.omg.CosNaming.NamingContextPackage.AlreadyBound,
      org.omg.CosNaming.NamingContextPackage.CannotProceed,
      org.omg.CosNaming.NamingContextPackage.InvalidName
  {
    NamingContext nc = null;
    NamingContext rnc = null;
    try {
      if (debug)
	dprint("bind_new_context " + nameToString(n));
      // The obvious solution:
      nc = this.new_context();
      this.bind_context(n,nc);
      rnc = nc;
      nc = null;
    } finally {
      try {
	if(nc != null)
	  nc.destroy();
      } catch (org.omg.CosNaming.NamingContextPackage.NotEmpty e) {
      }
    }
    return rnc;
  }

  /**
   * Destroy this NamingContext object. If this NamingContext contains
   * no bindings, the NamingContext is deleted.
   * @exception org.omg.CosNaming.NamingContextPackage.NotEmpty This NamingContext
   * is not empty (i.e., contains bindings).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   */
  public  void destroy()
    throws org.omg.CosNaming.NamingContextPackage.NotEmpty
  {
    if (debug)
      dprint("destroy ");
    NamingContextDataStore impl = (NamingContextDataStore)this;
    synchronized (impl) {
      if (impl.IsEmpty() == true)
	// The context is empty so it can be destroyed
	impl.Destroy();
      else
	// This context is not empty!
	throw new NotEmpty();
    }
  }

  /**
   * Implements all four flavors of binding. It uses Resolve() to
   * check if a binding already exists (for bind and bind_context), and
   * unbind() to ensure that a binding does not already exist.
   * If the length of the name is 1, then Bind() is called with
   * the name and the object to bind. Otherwise, the first component
   * of the name is resolved in this NamingContext and the appropriate
   * form of bind passed to the resulting NamingContext.
   * This method is static for maximal reuse - even for extended naming
   * context implementations where the recursive semantics still apply.
   * @param impl an implementation of NamingContextDataStore
   * @param n a sequence of NameComponents which is the name under which
   * the object will be bound.
   * @param obj the object reference to be bound.
   * @param rebind Replace an existing binding or not.
   * @param bt Type of binding (as object or as context).
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the first component of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CosNaming.NamingContextPackage.AlreadyBound An object is
   * already bound under the supplied name.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see resolve
   * @see unbind
   * @see bind
   * @see bind_context
   * @see rebind
   * @see rebind_context
   */
  public static void doBind(NamingContextDataStore impl,
			    NameComponent[] n,
			    org.omg.CORBA.Object obj,
			    boolean rebind,
			    org.omg.CosNaming.BindingType bt)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
    org.omg.CosNaming.NamingContextPackage.CannotProceed,
    org.omg.CosNaming.NamingContextPackage.InvalidName,
    org.omg.CosNaming.NamingContextPackage.AlreadyBound
  {
    // Valid name?
    if (n.length < 1)
      throw new InvalidName();

    // At bottom level?
    if (n.length == 1) {
      // The identifier must be set
      if (n[0].id.length() == 0)
	throw new InvalidName();

      // Ensure synchronization of backend
      synchronized (impl) {
	// Yes: bind object in this context under the name
	BindingTypeHolder bth = new BindingTypeHolder();
	if (rebind) {
	  // Remove if there
	  impl.Unbind(n[0]);
	} else {
	  if (impl.Resolve(n[0],bth) != null)
	    // "Resistence is futile." [Borg pickup line]
	    throw new AlreadyBound();
	}

	// Now there are no other bindings under this name
	impl.Bind(n[0],obj,bt);
      }
    } else {
      // No: bind in a different context
      NamingContext context = resolveFirstAsContext(impl,n[0]);

      // Compute tail
      NameComponent[] tail = new NameComponent[n.length - 1];
      System.arraycopy(n,1,tail,0,n.length-1);

      // How should we propagate the bind
      switch (bt.value()) {
      case BindingType._nobject:
	{
	  // Bind as object
	  if (rebind)
	    context.rebind(tail,obj);
	  else
	    context.bind(tail,obj);
	}
      break;
      case BindingType._ncontext:
	{
	  // Narrow to a naming context using Java casts. It must work.
	  NamingContext objContext = (NamingContext)obj;
	  // Bind as context
	  if (rebind)
	    context.rebind_context(tail,objContext);
	  else
	    context.bind_context(tail,objContext);
	}
      break;
      default:
	// This should not happen
	throw new org.omg.CORBA.INTERNAL(MinorCodes.NAMING_CTX_BAD_BINDINGTYPE,
					 CompletionStatus.COMPLETED_NO);
      }
    }
  }

  /**
   * Implements resolving names in this NamingContext. The first component
   * of the supplied name is resolved in this NamingContext by calling
   * Resolve(). If there are no more components in the name, the
   * resulting object reference is returned. Otherwise, the resulting object
   * reference must have been bound as a context and be narrowable to
   * a NamingContext. If this is the case, the remaining
   * components of the name is resolved in the resulting NamingContext.
   * This method is static for maximal reuse - even for extended naming
   * context implementations where the recursive semantics still apply.
   * @param impl an implementation of NamingContextDataStore
   * @param n a sequence of NameComponents which is the name to be resolved.
   * @return the object reference bound under the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the first component of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see resolve
   */
  public static org.omg.CORBA.Object doResolve(NamingContextDataStore impl,
					       NameComponent[] n)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
    org.omg.CosNaming.NamingContextPackage.CannotProceed,
    org.omg.CosNaming.NamingContextPackage.InvalidName
  {
    org.omg.CORBA.Object obj = null;
    BindingTypeHolder bth = new BindingTypeHolder();

    // Length must be greater than 0
    if (n.length < 1)
      throw new InvalidName();

    // The identifier must be set
    if (n.length == 1) {
      synchronized (impl) {
	// Resolve first level in this context
	obj = impl.Resolve(n[0],bth);
      }
      if (obj == null) {
	// Object was not found
	throw new NotFound(NotFoundReason.missing_node,n);
      }
      return obj;
    } else {
      // n.length > 1
      if (n[1].id.length() == 0)
	throw new InvalidName();

      NamingContext context = resolveFirstAsContext(impl,n[0]);

      // Compute restOfName = name[1..length]
      NameComponent[] tail = new NameComponent[n.length -1];
      System.arraycopy(n,1,tail,0,n.length-1);

      // Resolve rest of name in context
      return context.resolve(tail);
    }
  }

  /**
   * Implements unbinding bound names in this NamingContext. If the
   * name contains only one component, the name is unbound in this
   * NamingContext using Unbind(). Otherwise, the first component
   * of the name is resolved in this NamingContext and
   * unbind passed to the resulting NamingContext.
   * This method is static for maximal reuse - even for extended naming
   * context implementations where the recursive semantics still apply.
   * @param impl an implementation of NamingContextDataStore
   * @param n a sequence of NameComponents which is the name to be unbound.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound A name with multiple
   * components was supplied, but the first component could not be
   * resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the n-1 components of the supplied name.
   * @exception org.omg.CosNaming.NamingContextPackage.InvalidName The supplied name
   * is invalid (i.e., has length less than 1).
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see resolve
   */
  public static void doUnbind(NamingContextDataStore impl,
			      NameComponent[] n)
    throws org.omg.CosNaming.NamingContextPackage.NotFound,
    org.omg.CosNaming.NamingContextPackage.CannotProceed,
    org.omg.CosNaming.NamingContextPackage.InvalidName
  {
    // Name valid?
    if (n.length < 1)
      throw new InvalidName();

    // Unbind here?
    if (n.length == 1) {
      // The identifier must be set
      if (n[0].id.length() == 0)
	throw new InvalidName();

      org.omg.CORBA.Object objRef = null;
      synchronized (impl) {
	// Yes: unbind in this context
	objRef = impl.Unbind(n[0]);
      }

      if (objRef == null)
	  // It was not bound
	throw new NotFound(NotFoundReason.missing_node,n);
      // Done
      return;
    } else {
      // No: unbind in a different context

      // Resolve first  - must be resolveable
      NamingContext context = resolveFirstAsContext(impl,n[0]);

      // Compute tail
      NameComponent[] tail = new NameComponent[n.length - 1];
      System.arraycopy(n,1,tail,0,n.length-1);

      // Propagate unbind to this context
      context.unbind(tail);
    }
  }

  /**
   * Implements resolving a NameComponent in this context and
   * narrowing it to CosNaming::NamingContext. It will throw appropriate
   * exceptions if not found or not narrowable.
   * @param impl an implementation of NamingContextDataStore
   * @param n a NameComponents which is the name to be found.
   * @exception org.omg.CosNaming.NamingContextPackage.NotFound The
   * first component could not be resolved.
   * @exception org.omg.CosNaming.NamingContextPackage.CannotProceed Could not proceed
   * in resolving the first component of the supplied name.
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   * @see resolve
   */
  protected static NamingContext resolveFirstAsContext(NamingContextDataStore impl,
						       NameComponent n)
    throws org.omg.CosNaming.NamingContextPackage.NotFound {
    NameComponent[] rest = new NameComponent[1];
    rest[0] = n;
    org.omg.CORBA.Object topRef = null;
    BindingTypeHolder bth = new BindingTypeHolder();
    NamingContext context = null;

    synchronized (impl) {
      // Resolve first  - must be resolveable
      topRef = impl.Resolve(n,bth);
      if (topRef == null) {
	// It was not bound
	throw new NotFound(NotFoundReason.missing_node,rest);
      }
    }

    // Was it bound as a context?
    if (bth.value != BindingType.ncontext) {
      // It was not a context
      throw new NotFound(NotFoundReason.not_context,rest);
    }

    // Narrow to a naming context
    try {
      context = NamingContextHelper.narrow(topRef);
    } catch (org.omg.CORBA.BAD_PARAM ex) {
      // It was not a context
      throw new NotFound(NotFoundReason.not_context,rest);
    }

    // Hmm. must be ok
    return context;
  }

  transient protected org.omg.CORBA.ORB orb;

  public static String nameToString(NameComponent[] name)
  {
    StringBuffer s = new StringBuffer("{");
    if (name != null || name.length > 0) {
      for (int i=0;i<name.length;i++) {
	if (i>0)
	  s.append(",");
	s.append("[").
	  append(name[i].id).
	  append(",").
	  append(name[i].kind).
	  append("]");
      }
    }
    s.append("}");
    return s.toString();
  }

  // Debugging aids.
  public static final boolean debug = false;

  private static void dprint(String msg) {
    NamingUtils.dprint("NamingContextImpl("  +
		       Thread.currentThread().getName() + " at " +
		       System.currentTimeMillis() +
		       " ems): " + msg);
  }
}
