/*
 * @(#)TransientBindingIterator.java	1.31 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CosNaming;

// Import general CORBA classes
import org.omg.CORBA.SystemException;
import org.omg.CORBA.ORB;

// Get org.omg.CosNaming Types
import org.omg.CosNaming.Binding;
import org.omg.CosNaming.BindingType;
import org.omg.CosNaming.BindingTypeHolder;
import org.omg.CosNaming.NameComponent;

// Get base implementation
import com.sun.CosNaming.NamingContextImpl;
import com.sun.CosNaming.InternalBindingValue;

// Get a hash table
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Class TransientBindingIterator implements the abstract methods
 * defined by BindingIteratorImpl, to use with the TransientNamingContext
 * implementation of the NamingContextImpl. The TransientBindingIterator
 * implementation receives a hash table of InternalBindingValues, and uses
 * an Enumeration to iterate over the contents of the hash table.
 * @see BindingIteratorImpl
 * @see TransientNamingContext
 */
public class TransientBindingIterator extends BindingIteratorImpl
{
  /**
   * Constructs a new TransientBindingIterator object.
   * @param orb a org.omg.CORBA.ORB object.
   * @param aTable A hashtable containing InternalBindingValues which is
   * the content of the TransientNamingContext.
   * @param java.lang.Exception a Java exception.
   * @exception Exception a Java exception thrown of the base class cannot
   * initialize.
   */
  public TransientBindingIterator(org.omg.CORBA.ORB orb, Hashtable aTable)
    throws java.lang.Exception
  {
    super(orb);
    this.orb = orb;
    theHashtable = aTable;
    theEnumeration = this.theHashtable.elements();
    currentSize = this.theHashtable.size();
    this.orb.connect(this);
  }

  /**
   * Returns the next binding in the NamingContext. Uses the enumeration
   * object to determine if there are more bindings and if so, returns
   * the next binding from the InternalBindingValue.
   * @param b The Binding as an out parameter.
   * @return true if there were more bindings.
   */
  final public boolean NextOne(org.omg.CosNaming.BindingHolder b)
  {
    // If there are more elements get the next element
    boolean hasMore = theEnumeration.hasMoreElements();
    if (hasMore) {
      b.value =
	((InternalBindingValue)theEnumeration.nextElement()).theBinding;
      currentSize--;
    } else {
      // Return empty but marshalable binding
      b.value = new Binding(new NameComponent[0],BindingType.nobject);
    }
    return hasMore;
  }

  /**
   * Destroys this BindingIterator by disconnecting from the ORB
   * @exception org.omg.CORBA.SystemException One of a fixed set of CORBA system exceptions.
   */
  final public void Destroy()
  {
    // Disconnect the object reference
    orb.disconnect(this);
  }

  /**
   * Returns the remaining number of elements in the iterator.
   * @return the remaining number of elements in the iterator.
   */
  public final int RemainingElements() {
    return currentSize;
  }

  private int currentSize;
  private Hashtable theHashtable;
  private Enumeration theEnumeration;
  private org.omg.CORBA.ORB orb;
}
