/*
 * @(#)TransientNameServer.java	1.15 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CosNaming;

import java.util.Properties;
import java.net.InetAddress;

import org.omg.CORBA.ORB;

import org.omg.CosNaming.NamingContext;

import com.sun.CosNaming.TransientNameService;
import com.sun.CosNaming.BootstrapServer;

/**
 * Class TransientNameServer is a standalone application which
 * implements a transient name service. It uses the TransientNameService
 * class for the name service implementation, and the BootstrapServer
 * for implementing bootstrapping, i.e., to get the initial NamingContext.
 * <p>
 * The BootstrapServer uses a Properties object specify the initial service
 * object references supported; such as Properties object is created containing
 * only a "NameService" entry together with the stringified object reference
 * for the initial NamingContext. The BootstrapServer's listening port
 * is set by first checking the supplied arguments to the name server
 * (-ORBInitialPort), and if not set, defaults to the standard port number.
 * The BootstrapServer is created supplying the Properties object, using no
 * external File object for storage, and the derived initial port number.
 * @see TransientNameService
 * @see BootstrapServer
 */
public class TransientNameServer
{
  /**
   * Main startup routine. It instantiates a TransientNameService
   * object and a BootstrapServer object, and then allows invocations to
   * happen.
   * @param args an array of strings representing the startup arguments.
   */
  public static void main(String args[])
  {
    try {
      // Create an ORB object
       com.sun.CORBA.iiop.ORB orb = (com.sun.CORBA.iiop.ORB)
	   ORB.init(args,null);

      // Create a name service
      TransientNameService nameService = new TransientNameService(orb);

      // Get its object reference
      NamingContext initialRef = nameService.initialNamingContext();

      // Stringify and print out
      String stringifiedIOR = orb.object_to_string(initialRef) ;
      System.out.println("Initial Naming Context:\n" + stringifiedIOR) ;

      // Create a properties object with the initial references
      // we know about
      Properties initialProps = new Properties();
      initialProps.put("NameService",stringifiedIOR);

      // Determine the initial bootstrap port to use
      int initialPort = 900; // by convention
      try {
	// Try environment
	String ips = System.getProperty("org.omg.CORBA.ORBInitialPort");
	if (ips != null && ips.length() > 0)
	  initialPort = java.lang.Integer.parseInt(ips);
      } catch (java.lang.NumberFormatException e) {
	// do nothing
      }
      // Let arguments override
      for (int i=0;i<args.length;i++) {
	// Was the initial port specified?
	if (args[i].equals("-ORBInitialPort") &&
	    i < args.length-1) {
	  initialPort = java.lang.Integer.parseInt(args[i+1]);
	}
      }

      System.out.println("TransientNameServer: setting port for " +
			 "initial object references to: " +
			 initialPort);

      // Create a bootstrap server object with no associated file
      // and the properties object just created
      BootstrapServer initialBoot = new BootstrapServer(orb,
							initialPort,
							null,
							initialProps);
      try {
	// Start the bootstrap server
	initialBoot.start();
      } catch (org.omg.CORBA.SystemException e) {
	NamingUtils.errprint("TransientNameServer: caught a exception while starting the bootstrap service on port " + initialPort + ".");
	NamingUtils.errprint("TransientNameServer: try using a different port with commandline arguments -ORBInitialPort <portno>");
	return;
      }

      // Serve objects.
      java.lang.Object sync = new java.lang.Object();
      synchronized (sync) {sync.wait();}

    } catch (Exception e) {
      NamingUtils.errprint("TransientNameServer: Caught exception: ");
      NamingUtils.printException(e);
    }
  }

  /**
   * Private constructor since no object of this type should be instantiated.
   */
  private TransientNameServer() {}
}
