/*
 * @(#)TransientNameService.java	1.25 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.CosNaming;

// Get CORBA type
import org.omg.CORBA.INITIALIZE;
import org.omg.CORBA.ORB;
import org.omg.CORBA.CompletionStatus;

// Get org.omg.CosNaming types
import org.omg.CosNaming.NamingContext;

// Import transient naming context
import com.sun.CosNaming.TransientNamingContext;

/**
 * Class TransientNameService implements a transient name service
 * using TransientNamingContexts and TransientBindingIterators, which
 * implement the org.omg.CosNaming::NamingContext and org.omg.CosNaming::BindingIterator
 * interfaces specfied by the OMG Common Object Services Specification.
 * <p>
 * The TransientNameService creates the initial NamingContext object.
 * @see NamingContextImpl
 * @see BindingIteratorImpl
 * @see TransientNamingContext
 * @see TransientBindingIterator
 */
public class TransientNameService
{
  /**
   * Constructs a new TransientNameService, and creates an initial
   * NamingContext, whose object
   * reference can be obtained by the initialNamingContext method.
   * @param orb The ORB object
   * @exception org.omg.CORBA.INITIALIZE Thrown if
   * the TransientNameService cannot initialize.
   */
  public TransientNameService(org.omg.CORBA.ORB orb)
    throws org.omg.CORBA.INITIALIZE
  {
    // Create impldefs for naming contexts and binding iterators
    try {
      // Create an initial context
      TransientNamingContext initialContext = new TransientNamingContext(orb,null);
      // Get its object ref
      theInitialNamingContext = (org.omg.CosNaming.NamingContext)initialContext;
      // Patch in the object ref into the servant
      initialContext.localRoot = theInitialNamingContext;
    } catch (org.omg.CORBA.SystemException e) {
      NamingUtils.printException(e);
      throw new org.omg.CORBA.INITIALIZE(MinorCodes.TRANS_NS_CANNOT_CREATE_INITIAL_NC_SYS,
				      CompletionStatus.COMPLETED_NO);
    } catch (Exception e) {
      NamingUtils.printException(e);
      throw new org.omg.CORBA.INITIALIZE(MinorCodes.TRANS_NS_CANNOT_CREATE_INITIAL_NC,
				      CompletionStatus.COMPLETED_NO);
    }
  }

  /**
   * Return the initial NamingContext.
   * @return the object reference for the initial NamingContext.
   */
  public org.omg.CosNaming.NamingContext initialNamingContext()
  {
    return theInitialNamingContext;
  }

  // The initial naming context for this name service
  private org.omg.CosNaming.NamingContext theInitialNamingContext;
}
