/*
 * @(#)PackageDoc.java	1.2 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.javadoc;

/**
 * Represents a java package.  Provides access to information
 * about the package, the package's comment and tags, and the
 * classes in the package.
 *
 * @since JDK1.2
 * @author Kaiyang Liu (original)
 * @author Robert Field (rewrite)
 */
public interface PackageDoc extends Doc {

    /**
     * Get all classes (including Exceptions and Errors)
     * and interfaces.
     *
     * @return all included classes and interfaces in this package.
     */
    ClassDoc[] allClasses();

    /**
     * Get ordinary classes (excluding Exceptions and Errors)
     * in this package.
     *
     * @return included classes in this package.
     */
    ClassDoc[] ordinaryClasses();

    /**
     * Get Exception classes in this package.
     *
     * @return included Exceptions in this package.
     */
    ClassDoc[] exceptions();

    /**
     * Get Error classes in this package.
     *
     * @return included Errors in this package.
     */
    ClassDoc[] errors();

    /**
     * Get Interfaces in this package.
     *
     * @return included Errors in this package.
     */
    ClassDoc[] interfaces();

    /**
     * Lookup for a class within this package.
     *
     * @return ClassDoc of found class, or null if not found.
     */
    ClassDoc findClass(String className);
}
