/*
 * @(#)Tag.java	1.3 98/07/09
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.javadoc;

/**
 * Represents a documentation tag, e.g. @since, @author, @version.
 * Given a tag (e.g. "@since 1.2"), holds tag name (e.g. "@since")
 * and tag text (e.g. "1.2").  Tags with structure or which require
 * special processing are handled by subclasses (ParamTag, SeeTag,
 * and ThrowsTag.
 *
 * @author Robert Field
 * @author Atul M Dambalkar
 * @see SeeTag
 * @see ParamTag
 * @see ThrowsTag
 * @see Doc#tags()
 *
 */
public interface Tag extends java.io.Serializable {

    /**
     * Return the name of this tag.
     */
    String name();

    /**
     * Return the kind of this tag.
     */
    String kind();

    /**
     * Return the text of this tag, that is, portion beyond tag name.
     */
    String text();

    /**
     * convert this object to a string.
     */
    String toString();

    /**
     * For documentation comment with embedded @link tags, return the array of
     * Tags consisting of SeeTag(s) and text containing Tag(s).
     * Within a comment string "This is an example of inline tags for a
     * documentaion comment {@link Doc {@link Doc commentlabel}}",
     * where inside the inner braces, the first "Doc" carries exctly the same
     * syntax as a SeeTag and the second "commentlabel" is label for the Html
     * Link, will return an array of Tag(s) with first element as Tag with
     * comment text "This is an example of inline tags for a documentation
     * comment" and second element as SeeTag with referenced class as "Doc"
     * and the label for the Html Link as "commentlabel".
     *
     * @return Tag[] Array of tags with inline SeeTags.
     * @see ParamTag
     * @see ThrowsTag
     */
    Tag[] inlineTags();

    /**
     * Return the first sentence of the comment as tags. 
     * Includes inline tags 
     * (i.e. {&#64link <i>reference</i>} tags)  but not
     * regular tags.  
     * Each section of plain text is represented as a {@link Tag} 
     * of kind "Text".
     * Inline tags are represented as a {@link SeeTag} of kind "link".
     * First sentence is determined by 
     * {@link java.text.BreakIterator#getSentenceInstance(Locale)}.
     *
     * @return an array of {@link Tag}s representing the 
     * first sentence of the comment
     */
    Tag[] firstSentenceTags();
}
