/*
 * @(#)Method11SubWriter.java	1.8 98/08/05
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.doclets.oneone;

import com.sun.javadoc.*;
import com.sun.tools.doclets.*;

/**
 * Write method information in the old JDK 1.1 style.
 *
 * @author Robert Field
 */
public class Method11SubWriter extends ExecutableMemberSub11Writer {

    Method11SubWriter(Class11Writer writer) {
        super(writer);
    }

    public ProgramElementDoc[] members(ClassDoc cd) {
        return cd.methods();
    }

    protected void printReturnTag(Tag[] returns) {
        if (returns.length > 0) {
            writer.dt();
            writer.boldText("doclet.Returns");
            writer.dd();
            writer.print(returns[0].text());
        }
    }

    protected void printTags(ProgramElementDoc member) {
        MethodDoc method = (MethodDoc)member;
        ParamTag[] params = method.paramTags();
        Tag[] returns = method.tags("return");
        ThrowsTag[] thrown = method.throwsTags();
        SeeTag[] sees = method.seeTags();
        ClassDoc[] intfacs = member.containingClass().interfaces();
        ClassDoc overridden = method.overriddenClass();
        if (params.length + returns.length + thrown.length 
            + intfacs.length + sees.length > 0 ||
            overridden != null) {
            writer.dd();
            writer.dl();
            printParamTags(params);
            printReturnTag(returns);
            printThrowsTags(thrown);
            printOverridden(overridden, method);
            writer.printSeeTags(method);
            writer.dlEnd();
            writer.ddEnd();
        }
    }
                               
    protected void printSignature(ExecutableMemberDoc member) {
        displayLength = 0;
	writer.pre();
	printModifiers(member);
	printReturnType((MethodDoc)member);
	bold(member.name());
	printParameters(member);
	printExceptions(member);
	writer.preEnd();
    }
      
    protected void printReturnType(MethodDoc method) {
        Type type = method.returnType();
        if (type != null) {
            printTypeLink(type);
            print(' ');
        }
    }
    
    protected void printMethodBall(MethodDoc method, boolean small) {
        if(method.isStatic()) {
            writer.printBall("green-ball", "Green Ball", small);
        } else {
            writer.printBall("red-ball", "Red Ball", small);
        }
    }

    public void printSummaryLabel(ClassDoc cd) {
        writer.printBanner("method-index", "doclet.Method_Index", 207, 38);
    }

    protected void printSummaryType(ProgramElementDoc member) {
        printMethodBall((MethodDoc)member, true);
    }

    protected void printSummaryLink(ClassDoc cd, ProgramElementDoc member) {
        MethodDoc method = (MethodDoc)member;
        String name = method.name();
        writer.printHyperLink("#" + name + method.signature(), name);
        writer.println(method.flatSignature());
    }

    /**
     * Reproduce 1.1 bug.  Remove soon.
     */
    protected void printOverridden(ClassDoc overridden, MethodDoc method) {
        if (overridden != null) {
            String name = method.name();
            writer.dt();
            writer.boldText("doclet.Overrides");
            writer.dd();
	    // bug: no test
	    writer.printHyperLink(overridden.qualifiedName() + ".html", 
                                  name + method.signature(), name);
            writer.print(' ');
            writer.printText("doclet.in_class");
            writer.print(' ');
            writer.printClassLink(overridden);
        }
    }
                               
    protected void printHeader(ClassDoc cd) {
        writer.anchor("methods");
        writer.printBanner("methods", "doclet.Methods", 151, 38);
    }

    protected void printFooter(ClassDoc cd) {
    }

    protected void printMember(ClassDoc cd, ProgramElementDoc member) {
        MethodDoc method = (MethodDoc)member;
        String name = method.name();
        writer.aName(name + method.signature());
        printMethodBall(method, false);
        writer.aEnd();

        writer.aName(name);
        writer.bold(name);
        writer.aEnd();

        // Get the signature,
        // generate link for each parameter if it is a class.
        printSignature(method);
        printFullComment(method);
    }
}  
    
    
