/*
 * @(#)MemberFrameWriter.java	1.4 98/07/17
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.doclets.standard;

import com.sun.tools.doclets.*;
import com.sun.javadoc.*;
import java.io.*;
import java.lang.*;
import java.util.*;

/**
 * Generate the Class Information Page.
     * @see com.sun.javadoc.ClassDoc
     * @see java.util.Collections
     * @see java.util.List
     * @see java.util.ArrayList
     * @see java.util.HashMap
 *
 * @author Atul M Dambalkar
 */
public class MemberFrameWriter extends SubWriterHolderWriter {

    protected ClassDoc classdoc;

    public MemberFrameWriter(String path, String filename, ClassDoc classdoc)
                       throws IOException, DocletAbortException {
        super(path, filename,
        DirectoryManager.getRelativePath(classdoc.containingPackage().name()));
        this.classdoc = classdoc;
    }

    /**
     * Generate a class page.
     *
     * @param prev the previous class to generated, or null if no previous.
     * @param classdoc the class to generate.
     * @param next the next class to be generated, or null if no next.
     */
    public static void generate(ClassDoc classdoc) throws DocletAbortException {
        MemberFrameWriter mfgen;
        String path =
           DirectoryManager.getDirectoryPath(classdoc.containingPackage());
        String filename = classdoc.name() + "-members.html";
        try {
            mfgen = new MemberFrameWriter(path, filename, classdoc);
            mfgen.generateMemberFrameFile();
            mfgen.close();
        } catch (IOException exc) {
            Standard.configuration().standardmessage.
                error("doclet.exception_encountered", exc.toString(), filename);
            throw new DocletAbortException();
        }
    }

    /**
     * Generate the class file contents.
     */
    public void generateMemberFrameFile() {
        String clname = classdoc.name();
        String label = clname + "-Members";

        printHeader(label);

	printAllMembers();

        printBodyHtmlEnd();
    }

    /**
     * Print summary and detail information for the specified members in the
     * class.
     */
    protected void printAllMembers() {
        MethodSubWriter methW = new MethodSubWriter(this);
        ConstructorSubWriter consW = new ConstructorSubWriter(this);
        FieldSubWriter fieldW = new FieldSubWriter(this);
        ClassSubWriter innerW = new ClassSubWriter(this);

	innerW.membersSummaryLink(classdoc);
	fieldW.membersSummaryLink(classdoc);
	consW.membersSummaryLink(classdoc);
	methW.membersSummaryLink(classdoc);

        p();

	fieldW.membersLink(classdoc);
	consW.membersLink(classdoc);
	methW.membersLink(classdoc);
    }
}


