/*
 * @(#)Env.java	1.8 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.*;
import sun.tools.tree.Node;
import sun.tools.java.Package;
import sun.tools.javac.BatchEnvironment;
import sun.tools.javac.ErrorConsumer;
import sun.tools.javac.SourceClass;

import java.util.*;
import java.io.*;
import sun.io.MalformedInputException;

/**
 * Main "compilation" environment of javadoc,
 * this needs more work - should be untangled from BatchEnvironment.
 */

class Env extends BatchEnvironment implements ErrorConsumer {

    /**
     * Factory
     */
    static Env create(String srcPathString,
                      String classPathString,
                      String sysClassPathString,
                      String extDirsString,
                      int flags,
                      String encoding){
        ClassPath[] classPaths = classPaths(srcPathString, classPathString,
                                            sysClassPathString, extDirsString);
        // Create environment
        Env env = new Env(classPaths[0], classPaths[1]);
        env.flags |= flags;
	env.setCharacterEncoding(encoding);

        return env;
    }

    /**
     * Constructor
     */
    private Env(ClassPath sourceClassPath, ClassPath binaryClassPath) {
	super(sourceClassPath, binaryClassPath);

	// Turn off certain inheritance checks which the compiler
	// needs but which interfere badly with javadoc.  In the future,
	// it would be good if this workaround could be removed.
	SourceClass.turnOffInheritanceChecks();
    }

    public void printPaths() {
        Res.notice("main.path.msg",
                   sourcePath.toString(), binaryPath.toString());
    }

    /*
     * The name is resolved against the class's imports.
     */
    ClassDeclaration getClassDeclaration(ClassDefinition defn,
                                                Identifier ID) {
        Environment resenv;
        if (defn instanceof SourceClass) {
            SourceClass sc = (SourceClass) defn;
            resenv = sc.getImports().newEnvironment(this);
        } else {
            resenv = this;
        }
        Identifier resID = defn.resolveName(resenv, ID);
        return getClassDeclaration(resID);
    }

    ClassFile getSourceClassFile(String filename) {
        return sourcePath.getFile(filename);
    }

    static boolean isSourceClass(ClassDefinition def) {
        return (def instanceof SourceClass);
    }
}
