/*
 * @(#)ExecutableMemberDocImpl.java	1.9 98/07/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.MemberDefinition;
import sun.tools.java.ClassDeclaration;
import sun.tools.tree.LocalMember;

import java.util.Enumeration;

import java.text.CollationKey;

import java.lang.reflect.Modifier;

/**
 * Represents a method or constructor of a java class.
 *
 * @since JDK1.2
 * @author Robert Field
 */
class ExecutableMemberDocImpl extends MemberDocImpl 
                                    implements ExecutableMemberDoc {

    final String signature;
    final String flatSignature;

    final ParameterImpl[] parameters;
    final ClassDocImpl[] thrownExceptions;

    /**
     * constructor.
     */
    ExecutableMemberDocImpl(Env env, MemberDefinition f) {
        super(env, f);
        sun.tools.java.Type type = f.getType();
	signature = type.typeString("",false,false);
	flatSignature = type.typeString("",true,false);
	parameters = parameters(env, f);
	thrownExceptions = thrownExceptions(env, f);
    }


    /**
     * Constructor assist
     *
     * Get argument information.
     *
     * @see ParameterImpl
     *
     * @return an array of ParameterImpl, one element per argument
     * in the order the arguments are present.
     */
    private static ParameterImpl[] parameters(Env env, MemberDefinition f) {
        Enumeration e;
        sun.tools.java.Type types[] = f.getType().getArgumentTypes();
        if (types.length == 0) {
            return new ParameterImpl[0];
	}
        if (f.getArguments() == null) {
            e = null;
        } else {
            e = f.getArguments().elements();
        }

        if (e != null && !f.isStatic()) {
            e.nextElement();
        }

        ParameterImpl[] params = new ParameterImpl[types.length];
        for (int i = 0; i < types.length ; i++) {
            String name;
            if(e != null) {
                name = ((LocalMember)e.nextElement()).getName().toString();
            } else {
                name = "";
            }
            params[i] = new ParameterImpl(TypeMaker.getType(env, types[i]), name);
        }
        return params;
    }

    /**
     * Constructor assist
     *
     * Return exceptions this method or constructor throws. 
     *
     * @return an array of ClassDocImpl[] representing the exceptions
     * thrown by this method.
     */
    private static ClassDocImpl[] thrownExceptions(Env env, MemberDefinition f) {
	ClassDeclaration[] exp = f.getExceptions(env);
	ClassDocImpl[] ta = new ClassDocImpl[exp.length];
	for(int i=0; i<exp.length; i++){
	    ta[i] = ClassDocImpl.getClassDocImpl(env, exp[i]);
	}
	return ta;
    }

    /**
     * if this ExecutableMemberDoc has the same name and parameters 
     *    return true 
     * else 
     *    return false. 
     * @return boolean 
     */
    boolean hasNameAndParameters(String methodName, String[] paramTypeArray) {
        if (name().equals(methodName)) {
            if (paramTypeArray == null) {
                return true;
            }
            if (paramTypeArray.length == parameters.length) {
                if (equalParameterTypes(paramTypeArray)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Return true, if the respective parameter strings from ParameterImpl
     * Array and String array are equal else return false.
     */
    boolean equalParameterTypes(String[] paramTypeArray) {
        for (int i = 0; i < paramTypeArray.length; i++) {
            com.sun.javadoc.Type paramtype = parameters[i].type();
            String dimension = paramtype.dimension();
            String qualname = paramtype.qualifiedTypeName() + dimension;
            String simplename = paramtype.typeName() + dimension;
            if (!qualname.equals(paramTypeArray[i]) &&
                    !simplename.equals(paramTypeArray[i])) {
                return false;
            }
        }
        return true;
    }

    /**
     * generate a key for sorting. <p>
     * default is toString();
     */
    CollationKey generateKey() {
        String signature = flatSignature(); 
        if (signature.indexOf(",") > 0) {
            signature = signature.replace(',', ' ');
        }
        return RootDocImpl.collator.getCollationKey(name() + signature);
    }

    /**
     * Return exceptions this method or constructor throws. 
     *
     * @return an array of ClassDoc[] representing the exceptions
     * thrown by this method.
     */
    public ClassDoc[] thrownExceptions() {
	return thrownExceptions;
    }

    /** 
     * Return true if this method is native 
     */
    public boolean isNative() {
        return Modifier.isNative(modifiers);
    }

    /** 
     * Return true if this method is synchronized 
     */
    public boolean isSynchronized() {
        return Modifier.isSynchronized(modifiers);
    }

    /**
     * Get argument information.
     *
     * @see ParameterImpl
     *
     * @return an array of ParameterImpl, one element per argument
     * in the order the arguments are present.
     */
    public Parameter[] parameters() {
        return parameters;
    }

    /**
     * Return the throws tags in this method.
     * 
     * @return an array of ThrowTagImpl containing all <code>&#64exception</code> 
     * and <code>&#64throws</code> tags.
     */
    public ThrowsTag[] throwsTags() {
        return comment().throwsTags();
    }

    /**
     * Return the param tags in this method.
     * 
     * @return an array of ParamTagImpl containing all <code>&#64param</code> tags.
     */
    public ParamTag[] paramTags() {
        return comment().paramTags();
    }

    /**
     * Get the signature. It is the parameter list, type is qualified.  
     *      For instance, for a method <code>mymethod(String x, int y)</code>, 
     *      it will return <code>(java.lang.String,int)</code>.
     */
    public String signature() {
	return signature;
    }

    /**
     * get flat signature.  all types are not qualified.
     *      return a String, which is the flat signiture of this member.
     *      It is the parameter list, type is not qualified.  
     *      For instance, for a method <code>mymethod(String x, int y)</code>, 
     *      it will return <code>(String, int)</code>. 
     */
    public String flatSignature() {
	return flatSignature;
    }
}
