/*
 * @(#)FieldDocImpl.java	1.10 98/07/07
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.MemberDefinition;

import java.util.Map;
import java.util.HashMap;

import java.lang.reflect.Modifier;

/**
 * Represents a field in a java class.
 *
 * @see MemberDocImpl
 *
 * @since JDK1.2
 * @author Robert Field
 */
class FieldDocImpl extends MemberDocImpl implements FieldDoc {

    /**
     * map MemberDefinitions to FieldDocImpl, for uniqueness.
     */
    private static Map map = new HashMap();

    final Type type;

    /**
     * constructor.
     */
    private FieldDocImpl(Env env, MemberDefinition f, String lookupKey) {
        super(env, f);
	this.type = TypeMaker.getType(env, f.getType());
        map.put(lookupKey, this);
    }

    static FieldDocImpl getFieldDocImpl(Env env, MemberDefinition f) {
        String lookupKey = lookupKey(f);
        FieldDocImpl m = (FieldDocImpl)map.get(lookupKey);
        if (m == null) {
            m = new FieldDocImpl(env, f, lookupKey);
        }
        return m;
    }

    /**
     * Get type of this field.
     */
    public Type type() {
	return type;
    }

    /**
     * Is this Doc item a field?
     *
     * @return true is it represents a field, and it does.
     */
    public boolean isField() {
        return true;
    }

    /**
     * Return true if this field is transient
     */
    public boolean isTransient() {
        return Modifier.isTransient(modifiers);
    }

    /**
     * Return true if this field is volatile
     */
    public boolean isVolatile() {
        return Modifier.isVolatile(modifiers);
    }

    /**
     * Return the serialField tags in this FieldDocImpl item.
     *
     * @return an array of <tt>SerialFieldTagImpl<\tt> containing all
     *         <code>&#64serialField<\code> tags.
     */
    public SerialFieldTag[] serialFieldTags() {
        return comment().serialFieldTags();
    }

}
