/*
 * @(#)MethodDocImpl.java	1.12 98/08/05
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.ClassDeclaration;
import sun.tools.java.MemberDefinition;
import sun.tools.java.ClassNotFound;

import java.util.Map;
import java.util.HashMap;

import java.lang.reflect.Modifier;

/**
 * Represents a method of a java class.
 *
 * @since JDK1.2
 * @author Robert Field
 */
class MethodDocImpl extends ExecutableMemberDocImpl 
                                implements MethodDoc {

    /**
     * map MemberDefinitions to MethodDocImpl, for uniqueness.
     */
    private static Map map = new HashMap();

    final Type returnType;
    final ClassDocImpl overriddenClass;
    private MethodDocImpl overriddenMethod;

    /**
     * constructor.
     */
    private MethodDocImpl(Env env, MemberDefinition f, String lookupKey) {
        super(env, f);
	this.returnType = TypeMaker.getType(env, f.getType().getReturnType());
	this.overriddenClass = overriddenClass(env, f);
        map.put(lookupKey, this);
    }

    /**
     * Constructor assist
     *
     * Return the class that overrides this method.
     *
     * @return a ClassDocImpl representing the superclass
     * that overrides this method,  null if
     * this method is not overridden. 
     */
    ClassDocImpl overriddenClass(Env env, MemberDefinition f) {
	MemberDefinition overrides = null;
	try {
            if ((!f.isStatic()) && (!f.isConstructor())) {
                ClassDeclaration decl = f.getClassDefinition().
                                                            getSuperClass(env);
                while (decl != null) {
                    overrides = decl.getClassDefinition(env)
                                       .findMethod(env, 
                                                   f.getName(), f.getType());
                    if (overrides != null) {
                        if (showAccess.checkMember(overrides)) { 
                            break;
                        } else {
                            return null;
                        } 
                    }
                    decl = decl.getClassDefinition().getSuperClass(env);
                }
            }
	} catch (ClassNotFound ex) {
	    overrides = null;
	}
	if (overrides == null) 
	    return null;
	else{
          //This is commented, bcs'e may break other things, since it is not 
          //tested throughly. 
	  //overriddenMethod = getMethodDocImpl(env, overrides);
	    return ClassDocImpl.getClassDocImpl(env, 
                                               overrides.getClassDeclaration());
	}
    }

    static MethodDocImpl getMethodDocImpl(Env env, MemberDefinition f) {
        String lookupKey = lookupKey(f);
        MethodDocImpl m = (MethodDocImpl)map.get(lookupKey);
        if (m == null) {
            m = new MethodDocImpl(env, f, lookupKey);
        }
        return m;
    }

    /** 
     * Return true if it is a method, which it is. 
     * Note: constructors are not methods.
     *
     * @return tree
     */
    public boolean isMethod() { 
        return true;
    }

    /** 
     * Return true if this method is abstract 
     */
    public boolean isAbstract() {
        return Modifier.isAbstract(modifiers);
    }

    /**
     * Get return type. 
     *
     * @return the return type of this method, null if it
     * is a constructor.
     */
    public Type returnType() {
	return returnType;
    }

    /** 
     * Return the class that overrides this method.
     *
     * @return a ClassDocImpl representing the superclass
     * that overrides this method,  null if
     * this method is not overriden. 
     */
    public ClassDoc overriddenClass() {
	return overriddenClass;
    }
}
