/*
 * @(#)ParamTagImpl.java	1.12 98/04/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import java.util.*;

import sun.tools.java.*;

/**
 * Represents an @param documentation tag.
 * The parses and stores the name and comment parts of the 
 * method/constructor parameter tag.
 * 
 * @author Robert Field
 *
 */
class ParamTagImpl extends TagImpl implements ParamTag {

    final String parameterName;
    final String parameterComment;

    ParamTagImpl(Doc holder, String name, String text) {
        super(holder, name, text);
        String[] sa = divideAtWhite();
        this.parameterName = sa[0];
        this.parameterComment = sa[1];
        if (!(holder instanceof ExecutableMemberDocImpl)) {
            Res.warning("tag.not_on_method", 
			holder.toString(), this.text);
        }
    }
    
    /**
     * Return the parameter name.
     */
    public String parameterName() {
        return parameterName;
    }

    /**
     * Return the parameter comment.
     */
    public String parameterComment() {
        return parameterComment;
    }

    /**
     * Return the kind of this tag.
     */
    public String kind() {
        return "@param";
    }

    /**
     * convert this object to a string.
     */
    public String toString() {
        return name + ":" + text;
    }

    /**
     * For the parameter comment with embedded @link tags return the array of
     * TagImpls consisting of SeeTagImpl(s) and text containing TagImpl(s).
     *
     * @return TagImpl[] Array of tags with inline SeeTagImpls.
     * @see TagImpl#inlineTagImpls()
     * @see ThrowsTagImpl#inlineTagImpls()
     */
    public Tag[] inlineTags() {
        return Comment.getInlineTags(holder, parameterComment);
    }
}
