/*
 * @(#)ProgramElementDocImpl.java	1.6 98/07/07
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import sun.tools.java.ClassDefinition;
import java.lang.reflect.Modifier;

/**
 * Represents a java program element: class, interface, field,
 * constructor, or method.
 * This is an abstract class dealing with information common to
 * these elements.
 *
 * @see MemberDocImpl
 * @see ClassDocImpl
 *
 * @author Robert Field
 */
abstract class ProgramElementDocImpl extends DocImpl
                                  implements ProgramElementDoc {

  final int modifiers;

  final ClassDocImpl containingClass;

    /**
     * constructor.
     */
    ProgramElementDocImpl(Env env, String documentation,
                      int modifiers, ClassDefinition containing) {
        super(env, documentation);
	this.modifiers = modifiers;
	this.containingClass = containing == null?
	                            null :
	                            ClassDocImpl.getClassDocImpl(env, containing);
    }

    /**
     * Get the containing class of this program element.
     *
     * @return a ClassDocImpl for this element's containing class.
     * If this is a class with no outer class, return null.
     */
    public ClassDoc containingClass() {
        return containingClass;
    }

    /**
     * Get the modifier specifier integer.
     *
     * @see java.lang.reflect.Modifier
     */
    public int modifierSpecifier() {
        return modifiers;
    }

    /**
     * Get modifiers string.
     * <pre>
     * Example, for:
     *   public abstract int foo() { ... }
     * modifiers() would return:
     *   'public abstract'
     * </pre>
     */
    public String modifiers() {
	return Modifier.toString(modifiers);
    }

    /**
     * Return true if this program element is public
     */
    public boolean isPublic() {return Modifier.isPublic(modifiers);}

    /**
     * Return true if this program element is protected
     */
    public boolean isProtected() {return Modifier.isProtected(modifiers);}

    /**
     * Return true if this program element is private
     */
    public boolean isPrivate() {return Modifier.isPrivate(modifiers);}

    /**
     * Return true if this program element is package private
     */
    public boolean isPackagePrivate() {
        return !(isPublic() || isPrivate() || isProtected());
    }

    /**
     * Return true if this program element is static
     */
    public boolean isStatic() {return Modifier.isStatic(modifiers);}

    /**
     * Return true if this program element is final
     */
    public boolean isFinal() {return Modifier.isFinal(modifiers);}
}
