/*
 * @(#)Res.java	1.6 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;
import java.io.FileOutputStream;
import java.io.File;
import java.io.IOException;

/**
 * Utility for integrating with javadoc tools and for localization.
 * Handle Resources. Access to error and warning counts.
 * Message formatting.
 *
 * @see java.util.ResourceBundle
 * @see java.text.MessageFormat
 */
class Res {

    private static ResourceBundle messageRB;

    /**
     * Initialize ResourceBundle
     */
    static void initResource() {
	try {
	    messageRB =
		ResourceBundle.getBundle("com.sun.tools.javadoc.resources.javadoc");
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for javadoc is missing");
	}
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     */
    static String getText(String key) {
	return getText(key, (String)null);
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     * @param a1 first argument
     */
    static String getText(String key, String a1) {
	return getText(key, a1, null);
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     */
    static String getText(String key, String a1, String a2) {
	return getText(key, a1, a2, null);
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     * @param a3 third argument
     */
    static String getText(String key, String a1, String a2, String a3) {
	if (messageRB == null) {
	    initResource();
	}
	try {
	    String message = messageRB.getString(key);
	    String[] args = new String[3];
	    args[0] = a1;
	    args[1] = a2;
	    args[2] = a3;
	    return MessageFormat.format(message, args);
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for javadoc is broken. There is no " + key + " key in resource.");
	}
    }

    /**
     * Print error message, increment error count.
     *
     * @param msg message to print
     */
    static void printError(String msg) {
        System.err.println(Main.program + ": " + msg);
	Main.nerrors++;
    }

    /**
     * Print warning message, increment warning count.
     *
     * @param msg message to print
     */
    static void printWarning(String msg) {
        System.err.println(Main.program +  ": warning - " + msg);
	Main.nwarnings++;
    }

    /**
     * Print a message.
     *
     * @param msg message to print
     */
    static void printNotice(String msg) {
        System.err.println(msg);
    }

    /**
     * Print error message, increment error count.
     *
     * @param key selects message from resource
     */
    static void error(String key) {
        printError(getText(key));
    }

    /**
     * Print error message, increment error count.
     *
     * @param key selects message from resource
     * @param a1 first argument
     */
    static void error(String key, String a1) {
        printError(getText(key, a1));
    }

    /**
     * Print error message, increment error count.
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     */
    static void error(String key, String a1, String a2) {
        printError(getText(key, a1, a2));
    }

    /**
     * Print error message, increment error count.
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     * @param a3 third argument
     */
    static void error(String key, String a1, String a2, String a3) {
        printError(getText(key, a1, a2, a3));
    }

    /**
     * Print warning message, increment warning count.
     *
     * @param key selects message from resource
     */
    static void warning(String key) {
        printWarning(getText(key));
    }

    /**
     * Print warning message, increment warning count.
     *
     * @param key selects message from resource
     * @param a1 first argument
     */
    static void warning(String key, String a1) {
        printWarning(getText(key, a1));
    }

    /**
     * Print warning message, increment warning count.
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     */
    static void warning(String key, String a1, String a2) {
        printWarning(getText(key, a1, a2));
    }

    /**
     * Print warning message, increment warning count.
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     * @param a3 third argument
     */
    static void warning(String key, String a1, String a2, String a3) {
        printWarning(getText(key, a1, a2, a3));
    }

    /**
     * Print a message.
     *
     * @param key selects message from resource
     */
    static void notice(String key) {
        printNotice(getText(key));
    }

    /**
     * Print a message.
     *
     * @param key selects message from resource
     * @param a1 first argument
     */
    static void notice(String key, String a1) {
        printNotice(getText(key, a1));
    }

    /**
     * Print a message.
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     */
    static void notice(String key, String a1, String a2) {
        printNotice(getText(key, a1, a2));
    }

    /**
     * Print a message.
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     * @param a3 third argument
     */
    static void notice(String key, String a1, String a2, String a3) {
        printNotice(getText(key, a1, a2, a3));
    }
}


