/*
 * @(#)ThrowsTagImpl.java	1.9 98/05/14
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package com.sun.tools.javadoc;

import com.sun.javadoc.*;

/**
 * Represents a @throws or @exception documentation tag.
 * Parses and holds the exception name and exception comment.
 * Note: @exception is a backwards compatible synonymy for @throws.
 * 
 * @author Robert Field
 * @author Atul M Dambalkar
 * @see ExecutableMemberDocImpl#throwsTags()
 *
 */
class ThrowsTagImpl extends TagImpl implements ThrowsTag {

    final String exceptionName;
    final String exceptionComment;
    final ClassDoc exceptionClass;
    
    private static ClassDocImpl runtimeException; 

    ThrowsTagImpl(Doc holder, String name, String text) {
        super(holder, name, text);
        if (runtimeException == null) {
             runtimeException = ClassDocImpl.lookup("java.lang.RuntimeException");
        }
        String[] sa = divideAtWhite();
        this.exceptionName = sa[0];
        this.exceptionComment = sa[1];
        if (!(holder instanceof ExecutableMemberDocImpl)) {
            Res.warning("tag.not_on_method", 
			holder.toString(), this.text);
            exceptionClass = null;
        } else {
            ExecutableMemberDocImpl emd = (ExecutableMemberDocImpl)holder;
            ClassDoc con = emd.containingClass();
            exceptionClass = con.findClass(exceptionName);
            if (exceptionClass == null) {
                /* may just not be in this run */
                // Res.warning("tag.throws.exception_not_found",
                //            emd.qualifiedName(), kind(), exceptionName);
            } else {
                if (!isOK(exceptionClass, emd.thrownExceptions())) {
                    Res.warning("tag.throws.does_not_declare", 
                                emd.qualifiedName(), 
                                exceptionClass.qualifiedName());
                }
            }
        }
    }

    private static boolean isOK(ClassDoc exceptionClass, ClassDoc[] thrown) {
        if (exceptionClass.isError()) {
            return true;  // don't need to be declared
        }
        if (exceptionClass.subclassOf(runtimeException)) {
            return true;  // don't need to be declared
        }
        for (int i = 0; i < thrown.length; ++i) { 
            if (exceptionClass.subclassOf(thrown[i])) {
                return true; // found
            }
        }
        return false; // not found
    }
    
    /**
     * Return the exception name.
     */
    public String exceptionName() {
        return exceptionName;
    }

    /**
     * Return the exception comment.
     */
    public String exceptionComment() {
        return exceptionComment;
    }

    /**
     * Return the exception as a ClassDocImpl.
     */
    public ClassDoc exception() {
        return exceptionClass;
    }

    /**
     * Return the kind of this tag.  Always "@throws" for instances
     * of ThrowsTagImpl.
     */
    public String kind() {
        return "@throws";
    }

    /**
     * For the exception comment with embedded @link tags return the array of
     * TagImpls consisting of SeeTagImpl(s) and text containing TagImpl(s).
     *
     * @return TagImpl[] Array of tags with inline SeeTagImpls.
     * @see TagImpl#inlineTagImpls()
     * @see ParamTagImpl#inlineTagImpls()
     */
    public Tag[] inlineTags() {
        return Comment.getInlineTags(holder, exceptionComment);
    }
}
