/*
 * @(#)AppletProps.java	1.22 98/06/10
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.applet;

import java.awt.*;
import java.io.*;
import java.util.Properties;
import sun.net.www.http.HttpClient;
import sun.net.ftp.FtpClient;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedExceptionAction;
import java.security.PrivilegedActionException;

import sun.security.action.*;

class AppletProps extends Frame {
    
    TextField proxyHost;
    TextField proxyPort;
    Choice accessMode;

    AppletProps() {
        setTitle(amh.getMessage("title"));
	Panel p = new Panel();
	p.setLayout(new GridLayout(0, 2));

        p.add(new Label(amh.getMessage("label.http.server", "Http proxy server:")));
	p.add(proxyHost = new TextField());

        p.add(new Label(amh.getMessage("label.http.proxy")));
	p.add(proxyPort = new TextField());

        p.add(new Label(amh.getMessage("label.class")));
	p.add(accessMode = new Choice());
        accessMode.addItem(amh.getMessage("choice.class.item.restricted"));
        accessMode.addItem(amh.getMessage("choice.class.item.unrestricted"));

	add("Center", p);
	p = new Panel();
        p.add(new Button(amh.getMessage("button.apply")));
        p.add(new Button(amh.getMessage("button.reset")));
        p.add(new Button(amh.getMessage("button.cancel")));
	add("South", p);
	move(200, 150);
	pack();
	reset();
    }

    void reset() {
	((AppletSecurity)System.getSecurityManager()).reset();


	Boolean tmp = (Boolean) AccessController.doPrivileged(
                new GetBooleanAction("package.restrict.access.sun"));

	boolean packageRestrict = tmp.booleanValue();

	String proxyhost = (String) AccessController.doPrivileged(
                new GetPropertyAction("http.proxyHost"));
	String proxyport = (String)  AccessController.doPrivileged(
                new GetPropertyAction("http.proxyPort"));

	if (packageRestrict) {
           accessMode.select(amh.getMessage("choice.class.item.restricted"));
	} else {
           accessMode.select(amh.getMessage("choice.class.item.unrestricted"));
	}

	if (proxyhost != null) {
	    proxyHost.setText(proxyhost);
	    proxyPort.setText(proxyport);
	    HttpClient.proxyPort = Integer.valueOf(proxyport).intValue();
	} else {
	    proxyHost.setText("");
	    proxyPort.setText("");
	}


    }

    void apply() {
	// Get properties, set version
	final Properties props = (Properties) AccessController.doPrivileged(
	 new PrivilegedAction() {
	    public Object run() {
		return System.getProperties();
	    }
	});

	props.put("appletviewer.version", AppletViewer.theVersion);
	if (proxyHost.getText().length() > 0) {
	    props.put("http.proxyHost", proxyHost.getText().trim());
	    props.put("http.proxyPort", proxyPort.getText().trim());
	} else {
	    props.remove("http.proxyHost");
	}

	if (amh.getMessage("choice.class.item.restricted").equals(accessMode.getSelectedItem())) {
	    props.put("package.restrict.access.sun", "true");
	    props.put("package.restrict.access.netscape", "true");
	} else {
	    props.put("package.restrict.access.sun", "false");
	    props.put("package.restrict.access.netscape", "false");
	}

	// Save properties
	try {
	    reset();
	    AccessController.doPrivileged(new PrivilegedExceptionAction() {
		public Object run() throws IOException { 
		    FileOutputStream out =
		      new FileOutputStream(AppletViewer.theUserPropertiesFile);
		    props.save(out, "AppletViewer");
		    out.close();
		    return null;
		}
	    });
	    hide();
	} catch (java.security.PrivilegedActionException e) {
            System.out.println(amh.getMessage("apply.exception", 
					      e.getException()));
	    e.printStackTrace();
	    reset();
	}
    }

    public boolean action(Event evt, Object obj) {
        if (amh.getMessage("button.apply").equals(obj)) {
	    apply();
	    return true;
	}
        if (amh.getMessage("button.reset").equals(obj)) {
	    reset();
	    return true;
	}
        if (amh.getMessage("button.cancel").equals(obj)) {
            reset();
	    hide();
	    return true;
	}
	return false;
    }

    private static AppletMessageHandler amh = new AppletMessageHandler("appletprops");

}
