/*
 * @(#)AppletSecurity.java	1.100 98/09/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.applet;

import java.io.File;
import java.io.FilePermission;
import java.io.IOException;
import java.io.FileDescriptor;
import java.net.URL;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.SocketPermission;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Hashtable;
import java.security.*;
import sun.awt.AppContext;

import sun.security.provider.*;

/**
 * This class defines an applet security policy
 *
 * @version 	1.100, 09/01/98
 */
public
class AppletSecurity extends SecurityManager {
    private AppContext mainAppContext;

    /**
     * Construct and initialize.
     */
    public AppletSecurity() {
	reset();
	mainAppContext = AppContext.getAppContext();
    }

    /**
     * Reset from Properties
     */
    public void reset() {

    }

    /**
     * get the current (first) instance of an AppletClassLoader on the stack.
     */
    private AppletClassLoader currentAppletClassLoader()
    {
	// try currentClassLoader first
	ClassLoader loader = currentClassLoader();
	if ((loader == null) || (loader instanceof AppletClassLoader))
	    return (AppletClassLoader)loader;

	// if that fails, get all the classes on the stack and check them.
	Class[] context = getClassContext();
	for (int i = 0; i < context.length; i++) {
	    loader = context[i].getClassLoader();
	    if (loader instanceof AppletClassLoader)
		return (AppletClassLoader)loader;
	}

	// if that fails, try the context class loader
	loader = Thread.currentThread().getContextClassLoader();
	if (loader instanceof AppletClassLoader)
	    return (AppletClassLoader)loader;

	// no AppletClassLoaders on the stack
	return (AppletClassLoader)null;
    }

    /**
     * Returns true if this threadgroup is in the applet's own thread
     * group. This will return false if there is no current class
     * loader.
     */
    protected boolean inThreadGroup(ThreadGroup g) {
	if (currentAppletClassLoader() == null)
	    return false;
	else
	    return getThreadGroup().parentOf(g);
    }

    /**
     * Returns true of the threadgroup of thread is in the applet's
     * own threadgroup.
     */
    protected boolean inThreadGroup(Thread thread) {
	return inThreadGroup(thread.getThreadGroup());
    }

    /**
     * Applets are not allowed to manipulate threads outside
     * applet thread groups.
     */
    public synchronized void checkAccess(Thread t) {
	if (!inThreadGroup(t)) {
	    if (threadPermission == null)
		  threadPermission = new RuntimePermission("modifyThread");
	    checkPermission(threadPermission);
	}
    }

    private static RuntimePermission threadPermission;
    private static RuntimePermission threadGroupPermission;

    private boolean inThreadGroupCheck = false;

    /**
     * Applets are not allowed to manipulate thread groups outside
     * applet thread groups.
     */
    public synchronized void checkAccess(ThreadGroup g) {
	if (inThreadGroupCheck) {
	    // if we are in a recursive check, it is because
	    // inThreadGroup is calling appletLoader.getThreadGroup
	    // in that case, only do the super check, as appletLoader
	    // has a begin/endPrivileged
	    if (threadGroupPermission == null)
		threadGroupPermission = 
		    new RuntimePermission("modifyThreadGroup");
	    checkPermission(threadGroupPermission);
	} else {
	    try {
		inThreadGroupCheck = true;
		if (!inThreadGroup(g)) {
		    if (threadGroupPermission == null)
			threadGroupPermission = 
			    new RuntimePermission("modifyThreadGroup");
		    checkPermission(threadGroupPermission);
		}
	    } finally {
		inThreadGroupCheck = false;
	    }
	}
    }
    
    /**
     * Throws a <code>SecurityException</code> if the 
     * calling thread is not allowed to access the package specified by 
     * the argument. 
     * <p>
     * This method is used by the <code>loadClass</code> method of class 
     * loaders. 
     * <p>
     * The <code>checkPackageAccess</code> method for class 
     * <code>SecurityManager</code>  calls
     * <code>checkPermission</code> with the
     * <code>RuntimePermission("accessClassInPackage."+pkg)</code>
     * permission.
     *
     * @param      pkg   the package name.
     * @exception  SecurityException  if the caller does not have
     *             permission to access the specified package.
     * @see        java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    public void checkPackageAccess(final String pkgname) {

	// first see if the VM-wide policy allows access to this package
	super.checkPackageAccess(pkgname);

	final boolean[] check = { false };
	AccessController.doPrivileged(new PrivilegedAction(){
	    public Object run() {
		int i;
		String pkg = pkgname;
		do {
		    String prop = "package.restrict.access." + pkg;
		    if (Boolean.getBoolean(prop)) {
			check[0] = true;
			break;
		    }
		    if ((i = pkg.lastIndexOf('.')) != -1) {
			pkg = pkg.substring(0, i);
		    }
		} while (i != -1);
		return null;
	    }
	});

    	if (check[0])
	    checkPermission(new java.lang.RuntimePermission
			("accessClassInPackage." + pkgname));
    }

    /**
     * Tests if a client can get access to the AWT event queue.
     * <p>
     * This method calls <code>checkPermission</code> with the
     * <code>AWTPermission("accessEventQueue")</code> permission.
     *
     * @since   JDK1.1
     * @exception  SecurityException  if the caller does not have 
     *             permission to accesss the AWT event queue.
     */
    public void checkAwtEventQueueAccess() {
	AppContext appContext = AppContext.getAppContext();
	AppletClassLoader appletClassLoader = currentAppletClassLoader();

	if ((appContext == mainAppContext) && (appletClassLoader != null)) {
	    // If we're about to allow access to the main EventQueue,
	    // and anything untrusted is on the class context stack,
	    // disallow access.
	    super.checkAwtEventQueueAccess();
	}
    }

    /**
     * Returns the thread group of the applet. We consult the classloader
     * if there is one.
     * 
     * @exception  SecurityException  if the caller isn't running within
     *                                a recognized applet ThreadGroup.
     */
    public ThreadGroup getThreadGroup() {
	/* First we check if any classloaded thing is on the stack. */
	ClassLoader loader = currentAppletClassLoader();
	if (loader != null && (loader instanceof AppletClassLoader)) {
	    AppletClassLoader appletLoader = (AppletClassLoader)loader;
	    ThreadGroup loaderGroup = appletLoader.getThreadGroup();
	    if (loaderGroup == null)
		return super.getThreadGroup(); // No ClassLoader ThreadGroup
	    ThreadGroup currentGroup = Thread.currentThread().getThreadGroup();
	    if (!loaderGroup.parentOf(currentGroup)
	     || (loaderGroup == currentGroup)) {
			// Not currently in a child of loaderGroup;
			// throw a SecurityException, because we can't
			// identify a reasonable ThreadGroup to return.
		throw new SecurityException(
		    "Access denied to ThreadGroup not owned by applet");
	    }
	    /* At this point, we've established that we're running in a
	     * ThreadGroup which is a child of an AppletClassLoader's
	     * ThreadGroup.  We return the ThreadGroup corresponding to
	     * the current AppContext.
	     */
	    AppContext appContext = AppContext.getAppContext();
	    ThreadGroup contextThreadGroup =
			    AppletPanel.getAppContextThreadGroup(appContext);
	    if (contextThreadGroup != null) {
		return contextThreadGroup;
	    } else {
		// AppContext somehow wasn't found; throw a SecurityException
		// because we can't identify a reasonable ThreadGroup to return
		throw new SecurityException(
		    "Access denied to ThreadGroup not owned by applet");
	    }
	} else {
	    return super.getThreadGroup();
	}
    } // getThreadGroup()

}
