/*
 * @(#)AudioData.java	1.29 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.*;


/**
 * A clip of audio data. This data can be used to construct an
 * AudioDataStream, which can be played. <p>
 *
 * @author  Arthur van Hoff
 * @author  Kara Kytle
 * @version 1.29, 98/09/21
 * @see     AudioDataStream
 * @see     AudioPlayer
 */

 /*
  * the idea here is that the AudioData object encapsulates the
  * data you need to play an audio clip based on a defined set
  * of data.  to do this, you require the audio data (a byte 
  * array rather than an arbitrary input stream) and a format
  * object.
  */


public class AudioData {

    AudioFormat format;   // carry forth the format array amusement
    byte buffer[];

    /**
     * Constructor
     */
    public AudioData(byte buffer[]) {

        // okay, we need to extract the format and the byte buffer of data

        try {

            AudioContainerInputStream acis = AudioContainerInputStream.getStream(new ByteArrayInputStream(buffer));
            acis.setData();
            Format[] format = acis.getFormat();
            this.format = (AudioFormat)format[0];
            this.buffer = acis.getBytes();

        } catch (IOException e) {

            // if we cannot extract valid format information, we resort to assuming the data will be 8k mono u-law
            // in order to provide maximal backwards compatibility....

            AudioFormat mono_8k_ulaw_format = new AudioFormat(
                                                8000,                                   // sample rate
                                                AudioFormat.JAUDIO_G711_ULAW,           // encoding 
                                                1,                                      // sample size
                                                1,                                      // samples per unit
                                                1,                                      // number of channels
                                                true,                                   // big endian (this is true for SUN and DEC ulaw files, but okay for all 8bit....
                                                true                                    // signed
                                                );
                                            
            this.format = mono_8k_ulaw_format;
            this.buffer = buffer; 
        }
    }


    /**
     * Non-public constructor; this is the one we use in ADS and CADS
     * constructors.
     */
    AudioData(AudioFormat format, byte[] buffer) {

        this.format = format;
        this.buffer = buffer;
    }
}
