/*
 * @(#)AudioMidiInputStream.java	1.13 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.audio;

/**
 * Midi file format audio streams.
 *
 * @version 1.13 98/09/21
 * @author Kara Kytle
 * @author David Rivas
 */

import java.io.InputStream;
import java.io.IOException;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.SequenceInputStream;

class AudioMidiInputStream extends AudioContainerInputStream {


    /**
     * AudioMidiInputStream
     * Constructor for java.media.audio.AudioMidiInputStream.
     * Read data from any input stream and set Audio format.
     * Any bytes which have already been read from the stream 
     * are provided in an array of bytes.
     *
     * This should get called by ACIS.getStream().
     *
     * We go ahead and call setData() in the constructor because
     * we always need to have read all the data to play a midi song.
     */
    AudioMidiInputStream(byte[] bytes, InputStream in)
	throws IOException, BadAudioHeaderException {
	    super(bytes, in);
            setData();
    }


    /**
     * readFormat()
     * read MIDI Header. 
     * @exception BadAudioHeaderException, IOException
     * @return AudioFormat instance
     */	
    protected AudioFormat readFormat(byte[] bytes)
	throws BadAudioHeaderException, IOException {
	    
        // here we assemble the bytes + the input stream into
        // one InputStream

        if (bytes != null) in = new SequenceInputStream(new ByteArrayInputStream(bytes), in);

        return(null); 
    }   

    boolean setData() throws IOException {

		return readAllBytes();
    }
}
