/*
 * @(#)AudioStream.java	1.23 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.InputStream;
import java.io.DataInputStream;
import java.io.FilterInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;

/**
 * Convert an InputStream to an AudioStream. 
 *
 * @version 	1.23 98/09/21
 */


public class AudioStream extends FilterInputStream {

    AudioContainerInputStream acis;


	/*
	* create the AudioStream; if we survive without throwing
	* an exception, we should now have some subclass of
	* ACIS with all the header info already read
	*/

    public AudioStream(InputStream in) throws IOException {
		
	super(in);
	
	// if it's already an ACIS, just hold onto it
	// otherwise we try to make one out of the stream provided.
	// note that ACIS.getStream() can throw an IOException
	// or a BadAudioHeaderException (subclass of IOException))

        acis = AudioContainerInputStream.getStream(in);

	// set the value of the underlying input stream
        this.in = acis;
    }


    /**
     * A blocking read.
     */
    public int read(byte buf[], int pos, int len) throws IOException {

	return(acis.readFully(buf, pos, len));
    }


    /**
     * Get the data.
     */
    public AudioData getData() throws IOException {

        acis.setData();
        
        if (acis.stream instanceof ByteArrayInputStream) {
            Format[] format = acis.getFormat();
            byte[] bytes = acis.getBytes();
            if (bytes == null)
                throw new IOException("could not create AudioData object: no data received");
            return new AudioData((AudioFormat)format[0], bytes);
        }

        throw new IOException("could not create AudioData object");
    }

    
    public int getLength() {

        return(acis.getTotalBytes());
    }

}
