/*
 * @(#)HaeAudioClip.java	1.8 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.IOException; 
import sun.audio.AudioContainerInputStream;
import sun.audio.AudioFormat;
import sun.audio.Format;

class HaeAudioClip extends HaeWaveMemory {

    String                    streamName;
    AudioContainerInputStream inputStream;
    AudioFormat               inputStreamFormat;
    AudioFormat               outputStreamFormat;

    boolean verbose;

    /**
     * Creates a new HaeAudioClip from a file.
     *
     * @param inputStream with audio data.
     * @param streamName name of the input stream.
     *
     * @throws HaeException can't open the stream on the mixer.
     * @throws IOException if we can't open the file or have read errors.
     */

    public 
    HaeAudioClip(AudioContainerInputStream inputStream, String streamName)
				throws HaeException, IOException {
	super(HaeDevice.getSystemMixer());

	this.inputStream = inputStream;
	this.streamName = streamName;


	// Get the format of the raw input and the output to the engine (after conversions)
	Format f[] = inputStream.getFormat();
	inputStreamFormat = (AudioFormat) f[0];
	outputStreamFormat = inputStream.getOutputFormat();

	int rate              = inputStreamFormat.getSampleRate();
	int sampleSizeInBytes = inputStreamFormat.getSampleSize();
	int channels          = inputStreamFormat.getChannels();
	int sampleFrameSize	  = sampleSizeInBytes * channels;

	int outputRate = outputStreamFormat.getSampleRate();
	int outputSampleSizeInBytes = outputStreamFormat.getSampleSize(); 
	int outputChannels = outputStreamFormat.getChannels();
	int outputSampleFrameSize = outputSampleSizeInBytes * outputChannels;

	// calculate the number of frames, then calculate the number of bytes in the OUTPUT format
	int audioSize = inputStream.getTotalBytes();
	int framesToRead = audioSize / sampleFrameSize;
	byte[] audioData = new byte[framesToRead * outputSampleFrameSize];

	// Get valid data, with any conversions, from the stream
	inputStream.getConvertedData(audioData, framesToRead);

	// ... get the WaveMemory object.
	open(audioData, audioSize, outputSampleSizeInBytes * 8, outputChannels, 
	     (double) outputRate);
    }
}
