/*
 * @(#)HaeThread.java	1.17 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;
import java.io.PrintStream;

/**
 * Thread to manage the inner loop of the audio engine.
 *
 * @version 1.17, 98/06/29.
 * @author David Rivas, Chris Schardt
 */

class HaeThread extends Thread
{
    private HaeThread() // only native code can instantiate
    {
    super(getTopmostThreadGroup(), ""); //CLS: why is this needed

        setDaemon(true);
        setPriority(Thread.MAX_PRIORITY);
        setName("HAE frame thread");

        paused = false;
    }
    private static ThreadGroup getTopmostThreadGroup()
    {
    ThreadGroup g = currentThread().getThreadGroup();

        while ((g.getParent() != null) &&
               (g.getParent().getParent() != null))
        {
            g = g.getParent();
        }
        return g;
    }

    public void run()
    {
        while (true)
        {
            runNative();
            synchronized(this)
            {
                paused = true;
                // repeatedly call wait() until a call to unPause()
                while (paused)
                {
                    try
                    {
//                      TRACE("HaeThread: calling wait()");
                        wait();     // paused gets cleared here
//                      TRACE("HaeThread: returned from wait()");
                    }
                    catch(InterruptedException e)
                    {
//                      TRACE("HaeThread: wait() interrupted");
                    }
                }
            }
        }
    }

    // Processes frames of audio data
    // Returns after HAE_ReleaseAudioCard() is called
    native private void runNative();

    // Causes the while loop above to continue by returning from its wait() call
    private synchronized void unPause()
    {
//      TRACE("HaeThread: unPause() called, notifying...");
        paused = false;
        notify();
    }

    // Returns when the thread has paused.
	// $$kk: 05.08.98: added this method as part of fix for bug #4135246.
    private void pause()
    {
		while (!paused) {
				Thread.yield();
		}
    }


    // Set after runNative() returns
    // Cleared by unPause()
    // Causes this thread to repeatedly call wait(), just after runNative()
    private boolean     paused;

    // DEBUGGING
//  static private final void TRACE(String string)
//  {
//      System.err.println(string);
//  }
}


