/*
 * @(#)HaeWaveNoise.java	1.10 98/04/16
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

/**
 * HaeWaveNoise{}
 * Plays a sound from waveform data.  Make a subclass or use
 * HaeWaveStream or HaeWaveMemory.
 */

abstract class HaeWaveNoise extends HaeNoise implements HaePlayable {
// PUBLIC OPERATIONS:
    public void play()    // start, don't return until finished
                throws HaeException {
        if (!isPlaying()) {
			start();
		}

        // $$ jdr: This needs to be reimplemented with some
        // wait/notify mechanism. Will require some hacking on stop
        // and isPlaying like code.
        while (isPlaying()) {
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
            }
        }
    }

    public void start() throws HaeException {    // start, return right away
        start (1, 0, true);
    }

	public void start(double volume,   // 1.0 is normal
                      double pan,      // left=-1.0;center=0.0;right=1.0
                      boolean reverb
					  ) throws HaeException {
        m_isPaused = false;
        m_useReverb = reverb;

        privateStart(volume, pan, reverb);
    }

    public void pause() {
        m_isPaused = true;
        privateSetSampleRate(0);
    }

    public void resume() {
        m_isPaused = false;
        privateSetSampleRate(m_sampleRate);
    }

    public void setSampleRate(double sampleRate) {
        m_sampleRate = sampleRate;
        privateSetSampleRate(m_isPaused ? 0
                                        : m_sampleRate);
    }

    public abstract void setVolume(double volume);
    public abstract void setPan(double position);// left=-1.0;center=0.0;right=1.0

    public void setReverb(boolean useReverb) {
        m_useReverb = useReverb;
        privateSetReverb(useReverb);
    }

    // Convert mono 8-bit data read from SoundDesigner2 file
    // for use with HaeWaveNoise's
    public static void prepareMonoSd8Data(byte[] data, int length) {
        int i = 0;

        while (i < length) {
            data[i] += 128;
            i++;
        }
    }

    // Convert mono 16-bit data read from SoundDesigner2 file
    // for use with HaeWaveNoise{}s
    // length is frames, not bytes.
    public static void prepareMonoSd16Data(byte[] data, int length) {
        int i;

        length *= 2;
        i = 0;
        while (i < length) {
            byte temp;

            temp = data[i];
            data[i] = data[i + 1];
            data[i + 1] = temp;
            i += 2;
        }
    }

// PUBLIC DATA:
    public abstract boolean isPlaying();
    public boolean          isPaused()
        { return m_isPaused; }
    public double           getSampleRate()
        { return m_sampleRate; }

    public int              getBitsPerSample()
        { return m_bitsPerSample; }
    public int              getChannelsPerFrame()
        { return m_channelsPerFrame; }
    public int              getBytesPerFrame()
        { return m_bytesPerFrame; }

    public abstract double  getVolume(); 
    public abstract double  getPan(); // left=-1.0;center=0.0;right=1.0
    public boolean          getReverb()
        { return m_useReverb; }

// PROTECTED CONSTRUCTION:
    protected HaeWaveNoise(HaeMixer mixer) {
        super(mixer);

        m_id = 0;
        m_useReverb = false;
        m_isPaused = false;

        m_sampleRate = 0;
        m_bitsPerSample = 0;
        m_channelsPerFrame = 0;
        m_bytesPerFrame = 0;
    }

// PROTECTED OPERATIONS:
    protected void setSampleInfo(double sampleRate,
                                 int bitsPerSample, int channelsPerFrame) {
        m_sampleRate = sampleRate;

        m_bitsPerSample = bitsPerSample;
        m_channelsPerFrame = channelsPerFrame;
        m_bytesPerFrame = ((bitsPerSample + 7) / 8) * channelsPerFrame;
    }

	// $$kk: 01.14.98: added offset param
    protected abstract void privateStart(double volume,    // 1.0 is normal
                                         double pan,    // left=-1.0;center=0.0;right=1.0
                                         boolean reverb) throws HaeException;
    protected abstract void privateSetSampleRate(double sampleRate);
    protected abstract void privateSetReverb(boolean useReverb);

// PROTECTED DATA:
    protected double  m_sampleRate;
    protected boolean m_useReverb;
    protected int     m_id;    // identifies stream to global code
	
// PRIVATE DATA:
    private int       m_bitsPerSample;
    private int       m_channelsPerFrame;
    private int       m_bytesPerFrame;
    private boolean   m_isPaused;
};


// Audio Sample Data Format. (ASDF)
// Support for 8, 16 bit data, mono and stereo. Can be extended for multi 
// channel beyond 2 channels, but not required at the moment.
//
//    DATA BLOCK
//        8 bit mono data
//            ZZZZZZZ...
//                Where Z is unsigned 8 bit data
//
//        16 bit mono data
//            WWWWW...
//                Where W is signed 16 bit data
//
//        8 bit stereo data
//            ZXZXZXZX...
//                Where Z is unsigned 8 bit data for left channel, and X is 
//                signed 8 bit data for right channel.
//
//        16 bit stereo data
//            WQWQWQ...
//                Where W is signed 16 bit data for left channel, and Q is 
//                signed 16 bit data for right channel.
//
