/*
 * @(#)HaeWaveStreamBuffer.java	1.8 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

/**
 * HaeWaveStreamBuffer:
 * Facilitates passing waveform data into HaeWaveStream.getWavedata().
 * Data can be returned in two ways:
 * 1) Call set() to return data in an array that you have kicking around
 * 2) Call allocateByte() or allocateShort(), then fill the returned array
 */

class HaeWaveStreamBuffer {
// PUBLIC OPERATIONS:
    public void set(byte[] byteData, int frameCount) {
        if (byteData != null) {
        int maxFrameCount = getFrameCount(byteData.length);

            if (frameCount > maxFrameCount) {
                frameCount = maxFrameCount;
            }
        } else {
            System.err.println("null \"byteData\" passed to HaeWaveStreamBuffer.set()");
            frameCount = 0;
        }
        m_byteData = byteData;
        m_shortData = null;
        m_frameCount = frameCount;
    }

    public void set(short[] shortData, int frameCount) {
        if (shortData != null) {
        int maxFrameCount = getFrameCount(shortData.length * 2);

            if (frameCount > maxFrameCount) {
                frameCount = maxFrameCount;
            }
        } else {
            System.err.println("null \"shortData\" passed to HaeWaveStreamBuffer.set()");
            frameCount = 0;
        }
        m_byteData = null;
        m_shortData = shortData;
        m_frameCount = frameCount;
    }

    public byte[] allocateByte(int frameCount) {
        int byteCount = getByteCount(frameCount);

        m_shortData = null;

        if ((m_byteData == null) || (m_byteData.length < byteCount)) {
            m_byteData = null;
            if (byteCount > 0) {
                m_byteData = new byte[byteCount];
            }
        }
        m_frameCount = frameCount;
        return m_byteData;
    }

    public short[] allocateShort(int frameCount) {
        int shortCount = getByteCount(frameCount) * 2;

        m_byteData = null;

        if ((m_shortData == null) || (m_shortData.length < shortCount)) {
            m_shortData = null;
            if (shortCount > 0) {
                m_shortData = new short[shortCount];
            }
        }
        m_frameCount = frameCount;
        return m_shortData;
    }

    public void reduceFrameCount(int frameCount) {
        if (frameCount <= m_frameCount) {
            m_frameCount = frameCount;
        } else {
            System.err.println("HaeWaveStreamBuffer.reduceFrameCount() " +
			       "bad frameCount: " + frameCount +
		               ", created with: " + m_frameCount);
        }
    }

// PUBLIC DATA:
    public int getFrameCount(int byteCount)
        { return byteCount / m_bytesPerFrame; }
    public int getByteCount(int frameCount)
        { return frameCount * m_bytesPerFrame; }
    public int getValidFrameCount()
        { return m_frameCount; }

// PROTECTED CONSTRUCTION:
    protected HaeWaveStreamBuffer(int bytesPerFrame) {
        m_bytesPerFrame = bytesPerFrame;
//      m_frameCount = 0;    SHOULD BE INITIALIZED THUS AUTOMATICALLY
    }

// PRIVATE DATA:
    private int     m_bytesPerFrame;
    private byte[]  m_byteData;
    private short[] m_shortData;
    private int     m_frameCount;
}
