/*
 * @(#)MediaInputStream.java	1.9 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

/**
  * MediaInputStream
  * An MediaInputStream is an abstract base class for 
  * AudioContainerInputStream, VideoContainerInputStream.
  */

  /*
   * $$kk
   * 1) can we change 'Format f[]' to 'Format f'?
   * 2) can we make setFormat throw an exception 
   *    (i.e. if it's not a valid format for the subclass)?
   */


import java.io.FilterInputStream;
import java.io.InputStream;
import java.io.IOException;

abstract class MediaInputStream extends FilterInputStream {

    protected final static int MAX_READ_LIMIT = 2048;
    protected Format f[];
    protected int fcount;

    public MediaInputStream(InputStream in) {
	super(in);
    }

    public abstract Format[] getFormat();

    protected abstract void setFormat(Format theFormat); 

    // It returns the amount of data is being read.
    public abstract long getDataPos();

    // It is used to read bytes from the input stream and
    // converts them into an integer array, buffer[]. 
    public synchronized int read(int buffer[], int len) 
        throws IOException {

        int num = 0;
        int numInt =  len / 4;
        int remain = 0;

        for (int i = 0; i < numInt; i++) {
	    buffer[i] = 0;
	    for (int k = 0; k < 4; k++) {
		buffer[i] = (buffer[i] << 8) | in.read();
	    }
            num+=4;
        }
        remain = len - ( 4 * numInt);
 
        if (remain == 0)
            return num;

        buffer[numInt] = 0;
        for (int i = 0; i < remain ; i++ ) {
	    buffer[numInt] = (buffer[numInt] << 8) | in.read();
	    num ++;
        }
        return num;
    }

    // It is used to read bytes from the input stream
    // reads the requested number of bytes from the input stream.
    public synchronized int readFully(byte buffer[], int off, int numBytes)
        throws IOException {

        // Note: The maximum numByte > 2048, the following code
        // doesn't work. [ cania 1/17/97]
        // num = in.read(buffer, 0, numBytes);
        // Note: Instead of read byte by byte, read the maximum chunks at one time.
        //       Continue read until the requested number of bytes are copied into
        //      buffer.
        //      for (num = 0, i = 0; i < numBytes; i++ ) {
        //          buffer[i] = (byte) in.read();
        //          num++;
        //      }

	// Initialize ....
        int num = off;
	int readByte = 0;
	int readCount = 0;

	// ... and bound.
	if ( numBytes >= MAX_READ_LIMIT ) {
	    readCount = MAX_READ_LIMIT;
	} else {
	    readCount = numBytes;
	}

	// Read as many bytes as was asked for,
	// unless we get EOS before then.
	try {
	    while (num < numBytes) {
		
		// Get as many bytes as the stream will
		// give us, and as we need ...
	        readByte = in.read(buffer, num, readCount);
		if (readByte == -1) {
		    if (num == off) {
		        return -1;
		    } else {
		        break;
		    }
		}
		num += readByte;

		// .... and simulate O.S. support for
		// rescheduling on a read ...
	        Thread.currentThread().yield();

		// ... recompute current stats.
		if ( (numBytes - num) >= MAX_READ_LIMIT ) {       
                    readCount = MAX_READ_LIMIT;
                } else {       
                    readCount = numBytes - num; 
                }
	    }
	} catch (IOException e) {
	    num = -1;
	    throw new IOException(e.getMessage());	
	}       
        return num;
    }
}
