/*
 * @(#)SunAudioClip.java	1.2 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.IOException;
import java.io.InputStream;
import java.applet.AudioClip;

/**
 * Applet audio clip;
 *
 * @version 1.2, 98/09/21
 * @author Arthur van Hoff, Kara Kytle
 */

public class SunAudioClip implements AudioClip {

    // data object for this audio clip
    AudioData data;

    // the stream from which the data is read
    AudioDataStream stream;

    boolean DEBUG = false /*true*/;

    public SunAudioClip(InputStream in) throws IOException {

		if (DEBUG)
			System.out.println("Instantiating sun.audio.SunAudioClip");
		
		// this creates the appropriate subclass of ACIS inside the AudioStream
        AudioStream as = new AudioStream(in);

        // now we extract the data
        data = as.getData();
    }



    public synchronized void play() {
		
	if (data != null) {
            
            stop();

            // Build a new stream from the data and createa a player
            // around it.

            stream = new AudioDataStream(data);

            AudioPlayer.player.start(stream);
		    
            if (DEBUG) {
                System.err.println("Clip should be playing");
            }
        }
    }

    
    public synchronized void loop() {
		
	if (data != null) {

	    stop();

            stream = new ContinuousAudioDataStream(data);
            AudioPlayer.player.start(stream);
		    
            if (DEBUG)
                System.err.println("Clip should be looping");
        }
    }

    public synchronized void stop() {

	if (stream != null) {
	    AudioPlayer.player.stop(stream);
	}
	if (DEBUG) System.err.println("Clip should be stopped");
    }


    public String toString() {
        return getClass().toString();
    }

    
    protected void finalize() {

        if (stream != null) {
	    AudioPlayer.player.stop(stream);
            try { stream.close(); } catch (IOException e) {  }
	}
    }
}
