/*
 * @(#)TestMidiSong.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.*;


public class TestMidiSong 
{
	public static void main(String args[])
	{
	String			patchesPath;
	HaeMixer		mixer;
	int				arg;

		if (args.length < 2)
		{
		    System.err.println("Usage: HaeMixer <patch file> <midi/rmf files>");
		    return;
		}
		patchesPath = args[0];
		System.out.println("Creating mixer using \"" + patchesPath + "\"...");
		try
		{
		    mixer = new HaeMixer(patchesPath);
		    mixer.reengageAudio();
		}
		catch (HaeException e)
		{
		    System.err.println("Couldn't create a mixer using \"" + patchesPath + "\" : " + e);
		    return;
		}

		arg = 0;
		while (++arg < args.length)
		{
		String				songPath = args[arg];
		byte[]				fileBuffer;
		HaeMidiSong			song;

			System.out.println("Reading song file \"" + songPath + "\"...");
			try
			{
			FileInputStream		file = new FileInputStream(songPath);
			int					fileBytes = file.available();

				fileBuffer = new byte[fileBytes];
				if (file.read(fileBuffer) != fileBytes)
				{
					throw new IOException("Incorrect number of bytes read, expected " + fileBytes);
				}
				file.close();
			}
			catch (FileNotFoundException e)
			{
				System.err.println("Couldn't find \"" + songPath + "\" : " + e);
				return;
			}
			catch (IOException e)
			{
				System.err.println("Couldn't read or close \"" + songPath + "\" : " + e);
				return;
			}

			if (songPath.endsWith("rmf"))
			{
			    try
			    {
				HaeRmfSong	rmfSong;

					rmfSong = new HaeRmfSong(mixer);
					rmfSong.open(fileBuffer);
					song = rmfSong;
			    }
			    catch (HaeException e)
			    {
					System.err.println("Couldn't open the RMF song, \"" + songPath + "\" : " + e);
					return;
			    }
			}
			else
			{
			    try
			    {
					song = new HaeMidiSong(mixer);
					song.open(fileBuffer);
			    }
			    catch (HaeException e)
			    {
					System.err.println("Couldn't open the MIDI song, \"" + songPath + "\" : " + e);
					return;
			    }
			}

			System.out.println("Playing song \"" + songPath + "\"...");
			try
			{
			    song.play();
			}
			catch (HaeException e)
			{
				System.err.println("Couldn't play the song, \"" + songPath + "\" : " + e);
			}

			song.close();
		}
		mixer.close();
	}
};


