/*
 * @(#)TestWaveStream1.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;


// TestWaveStream1{}
//	Tests the HaeWaveStream{} class.

public class TestWaveStream1 extends HaeWaveStream
{
// PUBLIC CONSTRUCTION:
	public	TestWaveStream1(HaeMixer mixer)
	{
		super(mixer);
	}

// PUBLIC OPERATIONS
	public void				play()	// start, don't return until finished
								throws HaeException
	{
		try
		{
			Thread.sleep(1000);
		}
		catch (InterruptedException e)
		{
		}

		if (!isPlaying()) start();

		while (isPlaying())
		{
			try
			{
				Thread.sleep(1000);
			}
			catch (InterruptedException e)
			{
			}
//			stop(false);		//!!! FOR TESTING OF STREAM CODE
		}
	}

// PUBLIC TEST PROGRAM:
	public static void main(String args[])
	{
	// Parse args, getting a patch file name and a midi/rmf file name.
	String			patchesPath;
	HaeMixer		mixer;
	TestWaveStream1	stream;
	int				bufferBytes;

		if (args.length < 1)
		{
		    System.err.println("Usage: TestWaveStream1 <patch file> [<buffer bytes>]");
		    return;
		}
		patchesPath = args[0];
		System.out.println("Creating mixer using \"" + patchesPath + "\"...");
		try
		{
		    mixer = new HaeMixer(patchesPath);
		    mixer.setAudioModes(HaeMixer.SAMPLE_RATE_44K,
		    					mixer.getInterpolation(),
		    					mixer.getModifiers());
		    mixer.reengageAudio();
			System.out.println("Mixer created.");
		}
		catch (HaeException e)
		{
		    System.err.println("Couldn't create a mixer using \"" + patchesPath + "\" : " + e);
		    return;
		}

		bufferBytes = 256 * 1024;
		if (args.length >= 2)
		{
			try
			{
				bufferBytes = Integer.parseInt(args[1]);
			}
			catch (NumberFormatException e)
			{
			}
		}
		System.out.println("Creating the stream...");
		stream = new TestWaveStream1(mixer);
		try
		{
			stream.m_cycleCount = 0;
			stream.m_waveIndex = 0;
			stream.m_pulseWidth = 0;
			stream.m_pulseWidthIncrement = 1;
			stream.open(bufferBytes, BITS_PER_SAMPLE, 1, SAMPLES_PER_CYCLE * CYCLES_PER_SECOND);
		}
		catch (HaeException e)
		{
		    System.err.println("Couldn't open the stream : " + e);
		}

		System.out.println("Playing the stream with a " + 
							stream.getBufferBytes() +
							"-byte buffer...");
		try
		{
			stream.play();	// start, don't return until done
		}
		catch (HaeException e)
		{
		    System.err.println("Couldn't play the stream : " + e);
		}

		stream.close();
		mixer.close();
	}

// PROTECTED OPERATIONS:
	protected boolean		getWavedata(HaeWaveStreamBuffer buffer,
										int maxFrameCount)
	{
		if (BITS_PER_SAMPLE == 16)
		{
		short[]			data;
		int				i;

			data = buffer.allocateShort(maxFrameCount);

			i = 0;
			while (i < maxFrameCount)
			{
				data[i] = (short)(m_waveIndex <= m_pulseWidth ? 0x7FFF
															  : 0x8000);
				i++;
				m_waveIndex++;
				if (m_waveIndex >= SAMPLES_PER_CYCLE)
				{
					m_waveIndex = 0;
					m_cycleCount++;
					if (m_cycleCount >= TOTAL_CYCLES)
					{
						buffer.reduceFrameCount(i);
						return false;
					}
					m_pulseWidth += m_pulseWidthIncrement;
					if ((m_pulseWidth < 0) ||
						(m_pulseWidth >= SAMPLES_PER_CYCLE - 1))
					{
						m_pulseWidthIncrement = -m_pulseWidthIncrement;
						m_pulseWidth += m_pulseWidthIncrement;
					}
				}
			}
			return true;
		}
		else
		{
		byte[]			data;
		int				i;

			data = buffer.allocateByte(maxFrameCount);

			i = 0;
			while (i < maxFrameCount)
			{
				data[i] = (byte)(m_waveIndex <= m_pulseWidth ? 0xFF
															 : 0);
				i++;
				m_waveIndex++;
				if (m_waveIndex >= SAMPLES_PER_CYCLE)
				{
					m_waveIndex = 0;
					m_cycleCount++;
					if (m_cycleCount >= TOTAL_CYCLES)
					{
						buffer.reduceFrameCount(i);
						return false;
					}
					m_pulseWidth += m_pulseWidthIncrement;
					if ((m_pulseWidth < 0) ||
						(m_pulseWidth >= SAMPLES_PER_CYCLE - 1))
					{
						m_pulseWidthIncrement = -m_pulseWidthIncrement;
						m_pulseWidth += m_pulseWidthIncrement;
					}
				}
			}
			return true;
		}
	}

// PRIVATE CONSTANTS:
	private static final int BITS_PER_SAMPLE = 16;
	private static final int CYCLES_PER_SECOND = 220;	// A below middle C
	private static final int SAMPLES_PER_CYCLE = 200;
	private static final int TOTAL_CYCLES = 1 * CYCLES_PER_SECOND;	// five seconds

// PRIVATE DATA:
	int			m_cycleCount;
	int			m_waveIndex;
	int			m_pulseWidth;
	int			m_pulseWidthIncrement;
};


