/*
 * @(#)TestWaveStream2.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.audio;

import java.io.*;


// HaeWaveStreamFile{}
//	Plays a Sound Designer II audio file with HaeJava's streaming audio class

public class TestWaveStream2 extends HaeWaveStream
{
// PUBLIC CONSTRUCTION:
	public	TestWaveStream2(HaeMixer mixer, FileInputStream file)
	{
		super(mixer);
		m_file = file;
	}

// PUBLIC TEST PROGRAM:
	public static void main(String args[])
	{
	String			patchesPath;
	HaeMixer		mixer;
	int				arg;

		if (args.length < 2)
		{
		    System.err.println("Usage: HaeWaveMemory <patch file> <sd2 files>");
		    return;
		}
		patchesPath = args[0];
		System.out.println("Creating mixer using \"" + patchesPath + "\"...");
		try
		{
		    mixer = new HaeMixer(patchesPath);
		    mixer.reengageAudio();
		    mixer.setAudioModes(HaeMixer.SAMPLE_RATE_22K, HaeMixer.INTERPOLATION_DROP_SAMPLE,
		    					HaeMixer.FLAGS_USE_16 | HaeMixer.FLAGS_USE_STEREO);
		}
		catch (HaeException e)
		{
		    System.err.println("Couldn't create a mixer using \"" + patchesPath + "\" : " + e);
		    return;
		}

		arg = 0;
		while (++arg < args.length)
		{
		String			wavePath = args[arg];
		int				bitsPerSample = wavePath.endsWith(".sd8") ? 8 : 16;

			System.out.println("Opening wave \"" + wavePath + "\"...");
			try
			{
			FileInputStream		file = new FileInputStream(wavePath);
			TestWaveStream2		stream;

				System.out.println("Creating the stream, " +
									bitsPerSample + " bits per sample");
				stream = new TestWaveStream2(mixer, file);
				stream.open(16384, bitsPerSample, 1, SAMPLES_PER_SECOND);
				System.out.println("Playing the stream...");
				stream.m_count = 0;
				stream.play();
/*
			    {
			    	while (stream.isPlaying())
			    	{
			    		if (stream.m_count == 2)
			    		{
							System.out.println("Stopping the stream");
			    			stream.stop(false);
			    		}
			    	}
			    	stream.throwOnLastError();
			    }
*/
				System.out.println("Sleeping...");
				Thread.sleep(2000);
				stream.close();
				file.close();
			}
			catch (FileNotFoundException e)
			{
				System.err.println("Couldn't find \"" + wavePath + "\" : " + e);
			}
			catch (IOException e)
			{
				System.err.println("Couldn't close the file : " + e);
			}
			catch (HaeException e)
			{
				System.err.println("Couldn't play the wave, \"" + wavePath + "\" : " + e);
			}
			catch (InterruptedException e)
			{
				System.err.println("Wave file \"" + wavePath + "\" interrupted");
			}
		}
		mixer.close();
	}

// PROTECTED OPERATIONS:
	protected boolean		getWavedata(HaeWaveStreamBuffer buffer,
										int maxFrameCount)
	{
	byte[]		data;

		System.out.println("Allocating buffer of " + maxFrameCount + " frames.");
		data = buffer.allocateByte(maxFrameCount);

		try
		{
		int			bytes = m_file.read(data, 0, maxFrameCount * getBytesPerFrame());
		int			frames = bytes / getBytesPerFrame();

			System.out.println(bytes + " bytes read");
			if (getBitsPerSample() <= 8)
			{
				prepareMonoSd8Data(data, frames);
			}
			else
			{
				prepareMonoSd16Data(data, frames);
			}
			System.out.println("Reducing buffer to " + frames + " frames.");
			buffer.reduceFrameCount(frames);
			m_count++;
			return (frames >= maxFrameCount);
//			return (frames != 0);
		}
		catch (IOException e)
		{
			System.err.println("Couldn't read the file : " + e);
			return false;
		}
	}

// PRIVATE CONSTANTS:
	private static final int SAMPLES_PER_SECOND = 16000;

// PRIVATE DATA:
	FileInputStream		m_file;
	int					m_count;
};


