/*
 * @(#)MAreaGeometry.java	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
 * Internal implementations of the Area Geometry class in Java
 */

public abstract class MAreaGeometry {

    // === CAG Operations ===
    public MAreaGeometry cloneAndAdd(MAreaGeometry that) {
         return new TAddGeometry(this, that);
    }
    public MAreaGeometry cloneAndIntersect(MAreaGeometry that) {
         return new TIntersectGeometry(this, that);
    }
    public MAreaGeometry cloneAndSubtract(MAreaGeometry that) {
         return new TSubtractGeometry(this, that);
    }
    public MAreaGeometry cloneAndExclusiveOr(MAreaGeometry that) {
         return new TExclusiveOrGeometry(this, that);
    }

    // === Transformation ===
    public MAreaGeometry cloneAndTransform(TGrafMatrix matrix) {
         return new TTransformedGeometry(this, matrix);
    }

    // === Extraction ===
    public abstract void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix);

    public final void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode) {
        extract(extractor, parentNode, fkIdentity);
    }

    // === Queries ===
    public abstract boolean isEmpty();

    public boolean isPolygonal() {    // defaults to false
        return false;
    }
    public boolean isRectangular() {    // defaults to false
        return false;
    }
    public boolean isSingular() {    // defaults to true
        return true;
    }

    // === Hit methods ===
    public abstract TGRect getBounds();
    public abstract boolean intersects(TGRect rect);
    public abstract boolean contains(TGPoint point);
    public abstract boolean contains(TGRect rect);

    public long getTimeStamp() {     // default implementation returns contained time stamp
        return fTimeStamp;
    }
    private synchronized static long nextTimeStamp() {
	return ++fgTimeStampBase;
    }
    final void updateTimeStamp() {
        fTimeStamp = nextTimeStamp();
    }

    // === Java compatible methods ===
    public abstract Object clone();
    public boolean equals(MAreaGeometry that) {
        return (getTimeStamp() == that.getTimeStamp());
    }

    // === Constructors & Assignment ===
    protected MAreaGeometry() {
        updateTimeStamp();
    }
    protected MAreaGeometry(MAreaGeometry source) {
        fTimeStamp = source.fTimeStamp;
    }
    protected MAreaGeometry copyFrom(MAreaGeometry source) {
        if (this != source)
            fTimeStamp = source.fTimeStamp;
        return this;
    }

    // MAreaGeometry data:
    private long fTimeStamp;
    private static long fgTimeStampBase;    // it's okay to not initialise this

    private static final TGrafMatrix fkIdentity = // Why? Look at TGStatics.java
        new TGrafMatrix();

    // Used only by binary area geometry methods...
    public static MAreaGeometry add(MAreaGeometry left,
				    MAreaGeometry right) {
        if (right != null) {
            TGRect leftBounds = left.getBounds();
            TGRect rightBounds = right.getBounds();
            if (right.isRectangular() && rightBounds.contains(leftBounds))
                return right;
            else if ((!left.isRectangular() || !leftBounds.contains(rightBounds)) && !left.equals(right)) {
                return left.cloneAndAdd(right);
            }
        }
        return left;
    }

    public static MAreaGeometry subtract(MAreaGeometry left,
					 MAreaGeometry right) {
        if (right != null) {
            TGRect leftBounds = left.getBounds();
            TGRect rightBounds = right.getBounds();
            if (leftBounds.intersects(rightBounds)) {
                if ((right.isRectangular() && rightBounds.contains(leftBounds)) || left.equals(right))
                    return null;
                else
                    return left.cloneAndSubtract(right);
            }
        }
        return left;
    }

    public static MAreaGeometry intersect(MAreaGeometry left,
					  MAreaGeometry right) {
        if (right != null) {
            TGRect leftBounds = left.getBounds();
            TGRect rightBounds = right.getBounds();
            if (!leftBounds.intersects(rightBounds))
                return null;
            else if (left.isRectangular() && leftBounds.contains(rightBounds))
                return right;
            else if ((!right.isRectangular() || !rightBounds.contains(leftBounds)) || !left.equals(right))
                return left.cloneAndIntersect(right);
        }
        return left;
    }

    public static MAreaGeometry exclusiveOr(MAreaGeometry left,
					    MAreaGeometry right) {
        if (right != null) {
            if (left.equals(right))    // if they are equal, set to empty
                return null;
            else
                return left.cloneAndExclusiveOr(right);
        }
        return left;
    }
}
