/*
 * @(#)TArrayOfTBezier.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

class TArrayOfTBezier extends TRawPrimitiveArray {

    private TBezier[] fValues = null;
    public TArrayOfTBezier() {
        super(0);
    }

    // access
    public TBezier getValue(int i) {
        if (i < 0 || i >= fNumberOfValues)
            throw new IndexOutOfBoundsException("RawArray.value(index)");
        return fValues[i];
    }
    public void append(TBezier p) {
        int numValues = fNumberOfValues;
        resize(numValues + 1); // This changes fNumberOfValues!
        fValues[numValues] = p;
    }

    void allocateArray(int size) {
        fValues = null;
        if (size > 0)
            fValues = new TBezier[size];
    }
    void reallocateArray(int newSize) {
        TBezier oldValues[] = fValues;
        allocateArray(newSize);
        if (newSize > 0)
            System.arraycopy(
                oldValues, 0,
                fValues, 0,
                fArraySize > newSize ? newSize : fArraySize);
    }

    // Checks if a given point is inside the array
    // returns the index of the point found inside the array
    // if the point was not found then returns -1
    static public int isInsideArray(TGPointArray arrayToCheck, TGPoint checkPoint) {
        int allValues = arrayToCheck.getNumberOfValues();
        for (int j=0; j<allValues; j++) {
            if (TBezier.isClose(arrayToCheck.getValue(j), checkPoint))
                return j;
        }
        return -1;
    }

    // special function for this class to search an edge with end points
    int searchBezier(TGPoint firstEndPoint, TGPoint lastEndPoint) {
        for (int i=0; i<fNumberOfValues; i++) {
            if ((TBezier.isClose(firstEndPoint, fValues[i].getPoint(0), 0.1) &&
                 TBezier.isClose(lastEndPoint,fValues[i].getPoint(3), 0.1)) ||
                (TBezier.isClose(firstEndPoint, fValues[i].getPoint(3), 0.1) &&
                 TBezier.isClose(lastEndPoint,fValues[i].getPoint(0), 0.1))) {
                return i;
            }
        }
        return -1;  // error condition
    }
}
