/*
 * @(#)TBaseCurve.java	1.5 98/04/02
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

class TBaseCurve implements Cloneable {

    public static final int kPinned = 0;    // Knot vector for curve interpolating endpoints
    public static final int kFloating = 1;  // Uniform knot vector (no endpoint interpolation)
    public static final int kBezier = 2;    // Piecewise Bezier curve

    public static final int kBreak = 0;     // Break in curve - c0 discontinuity
    public static final int kKink = 1;      // "Kink" - c1 (tangent) discontinuity
    public static final int kJerk = 2;      // "Jerk" - c2 (accelleration) discon.
                                            // [Note - above term from automotive
                                            // industry - see Blossoming book
                                            // by Ramshaw]

    // original data for TBaseCurve<...>
    protected int fNumPoints;     // CACHED COPY
    protected int fOrder;
    protected TGRPointArray fPoints;
    protected TGParametricArray fKnots;

    public TBaseCurve() {
        fPoints = new TGRPointArray(0);
        fKnots = new TGParametricArray(0);
        fOrder = 0;
        fNumPoints = 0;

        // Note: for all TBaseCurve methods fOrder==0 means 'empty curve', ie
        // implies fNumPoints==0. In no other case should fOrder be 0 (or
        // 1 or negative)
    }

    public TBaseCurve(int order, int numberPoints) {
/* this is a internal class and I promise I will never supply wrong parametres... 
        if (order < 2)
            throw new IndexOutOfBoundsException("curve order below range");
        if (numberPoints < order)
            throw new IndexOutOfBoundsException("curve number of points above range");
*/

        fPoints = new TGRPointArray(numberPoints);
        fKnots = new TGParametricArray(order + numberPoints);

        fOrder = order;
        fNumPoints = numberPoints;

        // if no knotvector given,generate default knot vector=pinned
        // endpoints (Not to be confused with the null fKnots flag used by
        // the Bezier curves!)

        setKnotScheme(kPinned);
    }

/*
    public TBaseCurve(int order, int numberPoints, TGParametricArray knots) {
        if (order < 2)
            throw new IndexOutOfBoundsException("curve order below range");
        if (numberPoints < order)
            throw new IndexOutOfBoundsException("curve number of points above range");

        fOrder = order;
        fNumPoints = numberPoints;

        if (fKnots.getNumberOfValues() != fNumPoints + fOrder)
            throw new IllegalArgumentException("curve number of knots is inconsistant");

        fPoints = new TGRPointArray(numberPoints);
        fKnots = new TGParametricArray(order + numberPoints);
    }

    public TBaseCurve(int order, TGRPointArray points) {
        if (order < 2)
            throw new IndexOutOfBoundsException("curve order below range");
        int numberPoints = points.getNumberOfPoints();
        if (numberPoints < order)
            throw new IndexOutOfBoundsException("curve number of points above range");

        fPoints = new TGRPointArray(points);
        fKnots = new TGParametricArray(numberPoints + order);

        fOrder = order;
        fNumPoints = numberPoints;

        setKnotScheme(kPinned);
    }
*/

    public TBaseCurve(int order, TGRPointArray points, TGParametricArray knots) {
        if (order < 2)
            throw new IndexOutOfBoundsException("curve order below range");
        int numberPoints = points.getNumberOfPoints();
        if (numberPoints < order)
            throw new IndexOutOfBoundsException("curve number of points above range");
        if (knots.getNumberOfValues() != order + numberPoints)
            throw new IllegalArgumentException("curve number of knots is inconsistant");

        fPoints = new TGRPointArray(points);
        fKnots = new TGParametricArray(knots);

        fOrder = order;
        fNumPoints = numberPoints;
    }

    public TBaseCurve(TGPoint p0, TGPoint p1, TGPoint p2) {
        this(3, 3);
        fPoints.setPoint(0, p0);
        fPoints.setPoint(1, p1);
        fPoints.setPoint(2, p2);
    }


    public TBaseCurve(TGPoint p0, TGPoint p1, TGPoint p2, TGPoint p3) {
        this(4, 4);
        fPoints.setPoint(0, p0);
        fPoints.setPoint(1, p1);
        fPoints.setPoint(2, p2);
        fPoints.setPoint(3, p3);
    }

    public TBaseCurve(TBaseCurve curve) {
        fPoints = new TGRPointArray(curve.fPoints);
        fKnots = new TGParametricArray(curve.fKnots);

        fOrder = curve.fOrder;
        fNumPoints = curve.fNumPoints;
    }

    public Object clone() {
        return new TBaseCurve(this);
    }

    public TBaseCurve copyFrom(TBaseCurve src) {
        if (src != this)
        {
            fOrder = src.fOrder;
            fNumPoints = src.fNumPoints;
            fKnots.copyFrom(src.fKnots);
            fPoints.copyFrom(src.fPoints);
        }
        return(this);
    }

    public boolean equals(TBaseCurve other) {
        // Two empty curves are equal
        if (fOrder == 0)
            return (fOrder == other.fOrder);
        else
            return
                (fNumPoints == other.fNumPoints) &&
                (fOrder == other.fOrder) &&
                (fPoints.equals(other.fPoints)) &&
                (fKnots.equals(other.fKnots));
    }

/*================================================================
// Evaluation
//
*/

    // Evaluation.  The parameter u has a range defined by the knot
    // vector.  The valid limits on this range are:
    // TGCurve::minParameter() >= u > TGCurve::maxParameter()
    // Evaluation outside these limits falls off towards zero as you
    // get further outside of the range.  For pinned knot vectors (the
    // usual case), there is a c0 discontinuity between the curve
    // within the valid range and zero outside of it.  Since the W
    // component also falls to zero, the result becomes a NaN when the
    // W's get divided out.
    //
    // In plain talk, don't evaluate outside the range given above.

//  evaluate (1)
    public TGPoint evaluate(double u) {
        return evaluateW(u).divW();
    }

    // Internal version used when the brkpoint is already known
    protected TGPoint evaluate(double u, int delta) {

        TGRPoint p = new TGRPoint();
/*
        double coef[] = new double[fOrder];

        double theKnots[] = fKnots.getArray();
        prStatic.ComputeSplineValues(u, delta, fOrder, theKnots, coef); // prKnots
        prStatic.BasisMul(delta, fOrder, coef, fPoints.getArray(), p); // prSplineRoutines
*/
        prStatic.NurbEval(
            fPoints.getArray(),
            delta,
            u,
            fKnots.getArray(),
            fOrder,
            true, // RatFlag - ignored
            p);
        return p.divW();
    }

    // Second version; doesn't divide W out
    public TGRPoint evaluateW(double u) {

        TGRPoint p = new TGRPoint();
/*
        double coef[] = new double[fOrder];
*/
        double theKnots[] = fKnots.getArray();
/*
        int delta = prStatic.FindInterval(u, theKnots, fNumPoints, fOrder); // prKnots
        prStatic.ComputeSplineValues(u, delta, fOrder, theKnots, coef); // prKnots
        prStatic.BasisMul(delta, fOrder, coef, fPoints.getArray(), p); // prSplineRoutines
*/
        prStatic.NurbEval(                                           // prKnots
            fPoints.getArray(),
            prStatic.FindInterval(u, theKnots, fNumPoints, fOrder),  // prKnots
            u,
            theKnots,
            fOrder,
            true, // RatFlag - ignored
            p);

        return p;
    }

/*
//  evaluate (2)
    public TGPoint evaluate(double u, TGPoint tangent) {

        double bvals[] = new double[fOrder];
        double dvals1[] = new double[fOrder];

        TGRPoint tp = new TGRPoint();
        TGRPoint.SetToZero(tp);
        TGRPoint td1 = new TGRPoint(tp);

        double denom;
        int brkPoint, i;

        double theKnots[] = fKnots.getArray();
        TGRPoint pts[] = fPoints.getArray();
        brkPoint = prStatic.FindInterval(u, theKnots, fNumPoints, fOrder); // prKnots

        prStatic.ComputeSplineValues(u, brkPoint, fOrder, theKnots, bvals, dvals1, null); // prKnots

        for (i = 0; i < fOrder; i++) {
            tp.addBy(TGRPoint.multiply(pts[brkPoint-i], bvals[i]));
            td1.addBy(TGRPoint.multiply(pts[brkPoint-i], dvals1[i]));
        }

        // Derivatives require the use of the quotient rule for deriv's
        denom = tp.fW * tp.fW;
        TGPoint tmptd1 = td1.dropW();
        TGPoint tmptp = tp.dropW();

        tangent.copyFrom(
            tmptd1.multiplyBy(tp.fW).subtractBy(tmptp.multiplyBy(td1.fW)).divideBy(denom));

        return tp.divW();
    }

//  evaluate (3)
    public TGPoint evaluate(double u, TGPoint tangent, TGPoint deriv2) {
        double bvals[] = new double[ fOrder ];
        double dvals1[] = new double[ fOrder ];
        double dvals2[] = new double[ fOrder ];

        TGRPoint tp = new TGRPoint();
        TGRPoint.SetToZero(tp);
        TGRPoint td1 = new TGRPoint(tp);
        TGRPoint td2 = new TGRPoint(tp);

        double knots[] = fKnots.getArray();
        TGRPoint pts[] = fPoints.getArray();
        int brkPoint = prStatic.FindInterval(u, knots, fNumPoints, fOrder); // prKnots

        prStatic.ComputeSplineValues(u, brkPoint, fOrder, knots, bvals, dvals1, dvals2); // prKnots
        for (int i = 0; i < fOrder; i++) {
            tp.addBy(TGRPoint.multiply(pts[brkPoint-i], bvals[i]));
            td1.addBy(TGRPoint.multiply(pts[brkPoint-i], dvals1[i]));
            td2.addBy(TGRPoint.multiply(pts[brkPoint-i], dvals2[i]));
        }

        // Derivatives require the use of the quotient rule for deriv's
        double denom = tp.fW * tp.fW;
        TGPoint tptmp = tp.dropW();
        TGPoint td1tmp = td1.dropW();
        TGPoint td2tmp = td2.dropW();

        // tangent = (td1tmp*tp.fW - tptmp*td1.fW) / denom;
        tangent.copyFrom(
            TGPoint.multiply(td1tmp, tp.fW)
                .subtractBy(TGPoint.multiply(tptmp, td1.fW))
                    .divideBy(denom));

        denom *= tp.fW;

        // deriv2 = (((td2tmp*tp.fW - tptmp*td2.fW) * tp.fW) - ((td1tmp*tp.fW - tptmp*td1.fW) * (2 * td1.fW))) / denom;
        deriv2.copyFrom(
            TGPoint.multiply(td2tmp, tp.fW)
                .subtractBy(TGPoint.multiply(tptmp, td2.fW))
                    .multiplyBy(tp.fW)
                        .subtractBy(
                            TGPoint.multiply(td1tmp, tp.fW)
                                .subtractBy(TGPoint.multiply(tptmp, td1.fW))
                                    .multiplyBy(2 * td1.fW)).divideBy(denom));

        return tp.divW();
    }
*/

/*
//================================================================
// Curve length
//
//
    // Return the length along the curve from the parameter uFrom
    // to the parameter uTo.  Note: the current implementation is likely
    // to have accuracy problems for curves with order > 5.

    public double arcLength(double uFirst, double uLast) {
        int brkPoint;
        double u0, u1;         // Current interval
        double tmp;
        double length;
        TCurveIntegrator integrator = new TCurveIntegrator(this);

        if (uFirst == uLast)        // Duh...
            return 0.0;

        if (uFirst > uLast) {       // Must be in accending order
            tmp = uLast;
            uLast = uFirst;
            uFirst = tmp;
        }

        u0 = uFirst;
        double knots[] = fKnots.getArray();
        brkPoint = prStatic.FindInterval(u0, knots, fNumPoints, fOrder); // prKnots
        u1 = knots[brkPoint + 1];

        if (u1 > uLast)
            u1 = uLast;

        TGRPoint pts[] = fPoints.getArray();
        length = 0.0;

        // Step through the curve segment by segment, accumulating the length

        while (u0 < uLast) {
            if (fOrder == 2) {      // Skip the integration for linear curves
                TGPoint p0 = pts[brkPoint-1].divW();
                TGPoint p1 = pts[brkPoint].divW();
                if ((u0 == knots[brkPoint]) && (u1 == knots[brkPoint + 1]))
                    length += TGPoint.subtract(p0,p1).vectorLength();
                else {
                    length +=
                        TGPoint.subtract(
                            TGPoint.multiply(
                                TGPoint.subtract(p0,TGPoint.add(p0,p1)),
                                u0 - knots[brkPoint]),
                            TGPoint.multiply(
                                TGPoint.subtract(p0,TGPoint.add(p0,p1)),
                                u1 - knots[brkPoint])
                       ).vectorLength();
                }
            }
            else
                length += integrator.gaussianQuadrature(u0, u1);

            u0 = u1;
            brkPoint = prStatic.FindNextInterval(brkPoint, knots, fNumPoints); // prKnots
            u1 = knots[brkPoint + 1];
            if (u1 > uLast)
                u1 = uLast;
        }

        return length;
    }

    // Works over entire parameter range
    public double arcLength() {
        return arcLength(getMinParameter(), getMaxParameter());
    }
*/

/*
    // Approximate the parameter, given a particular arc length along the
    // curve.  Tolerance is used for the accuracy of the root finder, leave it
    // alone for most "screen sized" data.
    // in ArcLength.inline:
    public double approximateParameterFromArcLength(
        double length)
    {
        return approximateParameterFromArcLength(length, (double)1.0e-5);
    }
    public double approximateParameterFromArcLength(
        double length,
        double tolerance)
    {
        if (length == 0.0)
            return getMinParameter();

        if (length >= arcLength())
            return getMaxParameter();

        if (fOrder == 2)            // Don't do root finding on linear curves
        {
            double u0 = getMinParameter();
            double knots[] = fKnots.getArray();
            int brkPoint = prStatic.FindInterval(u0, knots, fNumPoints, fOrder); // prKnots
            double u1 = knots[brkPoint + 1];
            double uMax = getMaxParameter();
            double curLength, segLength;
            TGRPoint pts[] = fPoints.getArray();

            curLength = 0.0;
            while (u0 < uMax) {
                segLength = pts[brkPoint-1].divW().subtractBy(pts[brkPoint].divW()).vectorLength();

                if ((length >= curLength) && (length < curLength + segLength)) {
                    return ((u1 - u0) * ((length - curLength) / segLength)) + u0;
                }

                curLength += segLength;

                u0 = u1;
                brkPoint = prStatic.FindNextInterval(brkPoint, knots, fNumPoints); // prKnots
                u1 = knots[brkPoint + 1];
            }
            return getMaxParameter(); // Should never get here...
        }
        else {
            TCurveLengthRoot rootFinder = new TCurveLengthRoot(this, length);
            return rootFinder.findRoot(getMinParameter(), getMaxParameter(), tolerance);
        }
    }
*/

/* // #ifdef NO_REPARAM_YET
    // PULLED FROM RELEASE - The interface and implementation are too
    // questionable and the benifit is too small to risk maintaining it
    // at this time

    // Generate a reparameterization curve that translates arc length
    // to parameter.  The curve may then be evaluated by arc length by
    // something like:
    //
    //          curve.evaluate(reparam.evaluate(lengthValue).fY);
    //
    // The tolerance default is a useful value for "screen-sized" curves,
    // those on much different scales may need adjustment.  MaxPts is the
    // total number of control points allowed in the reparameterization curve,
    // the default is 20 * the number of control points in the original curve.
    // Because the approximation is adaptive, it will generally use *much* less
    // than the total number allowed, so be generous (small numbers can cause
    // the approximation to be too conservative and inaccurate).
    //
    ///??? void approximateArcLengthReparameterizationCurve(
        // TGCurve reparamCurve,
        // int maxPts = 0,
        // double tolerance = 0.01);
// #endif
*/

/*================================================================
// Basic inquery
//
*/
    // Returns true if the curve was created with the empty constructor
    public boolean isEmpty() {
        return (fOrder == 0);
    }

    public int getOrder() {
        return fOrder;
    }

    // Returns # of control points
    public int getNumberOfPoints() {
        return fNumPoints;
    }

    // Get the i'th control point
    public TGRPoint getPoint(int i) {
        return new TGRPoint(fPoints.getValue(i));
    }

    public void getPoints(TGRPointArray pts) {
        pts.copyFrom(fPoints);
    }

/*================================================================
// Parameter / Knot vector query
//
*/

    // Start/end parameters of the curve.  Curve is undefined outside
    // these values.
    public double getMinParameter() {
        return fKnots.getValue(fOrder-1);
    }
    public double getMaxParameter() {
        return fKnots.getValue(fNumPoints);
    }

    // Order + # of control points
    public int getNumberOfKnots() {
        return fOrder+fNumPoints;
    }

    // Return the i'th knot
    public double getKnot(int i) {
        return fKnots.getValue(i);
    }

/*
    public void getKnots(TGParametricArray knots) {
        knots.copyFrom(fKnots);
    }
*/

    // Returns true if the curve is pinned (interpolates the first and
    // last ctl points)
    public boolean isPinned() {
        return prStatic.IsPinnedKnots(fOrder, fNumPoints, fKnots); // prKnots
    }

    // Returns true if the curve has a Bezier knot vector
    public boolean isBezier() {
        return prStatic.IsBezierKnots(fOrder, fNumPoints, fKnots); // prKnots
    }

/*
    // Returns the parametric value of the closest point on the curve to the
    // center of the rectangle.
    // WARNING:  The current implementation works by discretizing the curve in
    // 72 DPI device space - beware of this fact if you call GetParametric with
    // gigantic or microscopic coordinates.  In the future, we expect to have a
    // resolution independant implementation.
    public double nearestParametric(TGPoint hitPt) {
        TSpanBounds spanInfo[] = new TSpanBounds[fNumPoints];
        for (int ii = 0;  ii < fNumPoints;  ii++) {
            spanInfo[ii] = new TSpanBounds();
        }

        double knots[] = fKnots.getArray();
        double t, u = knots[fOrder - 1];

        double last_u = knots[fNumPoints];

        TGPoint lastPt, curPt;
        double nearestDist = Double.POSITIVE_INFINITY;

        int i, span, nearestSpanInd = 0, numSpans = 0;
        int numSteps = 2 * fOrder; // crude heuristic...
        int brkPoint = prStatic.FindInterval(u, knots, fNumPoints, fOrder); // prKnots

        // Step through the curve and record the points on the segment bounds
        // nearest and furthest from the hit point.
        while (u < last_u) {
            TGBounds segmentBounds = new TGBounds(brkPoint, fOrder, fPoints);
            spanInfo[numSpans].setDistances(brkPoint, hitPt, segmentBounds);
            brkPoint = prStatic.FindNextInterval(brkPoint, knots, fNumPoints); // prKnots
            u = knots[brkPoint];
            numSpans++;
            segmentBounds = null;
        }

        // Find the index of the span whose furthest distance from the hit point
        // is the smallest.  Any span whose *nearest* distance is not smaller
        // than this can be eliminated from further inspection.
        int nearestFarthestIndex = 0;
        for (i = 1; i < numSpans; i++) {
            if (spanInfo[i].fFarthest < spanInfo[nearestFarthestIndex].fFarthest)
                nearestFarthestIndex = i;
        }

        // Eliminate uninteresting cases...
        for (i = 0; i < numSpans; i++) {
            if (spanInfo[i].fNearest > spanInfo[nearestFarthestIndex].fFarthest)
                spanInfo[i].fInteresting = false;
        }

        DoubleType curDist = new DoubleType(0.);    // be used for PointSegDist calls...

        // Now test the "interesting" segments more closely by evaluating the curve
        for (span = 0; span < numSpans; span++) {
            TSpanBounds curSpan = spanInfo[span];

            // See if actual measure from a previous spans rules this one out
            if (curSpan.fNearest > nearestDist)
                curSpan.fInteresting = false;

            // Still interesting, so check the curve
            if (curSpan.fInteresting) {
                brkPoint = curSpan.fBrkPoint;
                lastPt = evaluate(knots[brkPoint], brkPoint);
                last_u = knots[brkPoint];
                curSpan.fNearest = Double.POSITIVE_INFINITY; // will set to actual dist to curve

                for (i = 1; i <= numSteps; i++) {
                    if (i == numSteps)
                        u = knots[brkPoint + 1];
                    else
                        u = knots[brkPoint] +
                            (knots[brkPoint+1] - knots[brkPoint]) *
                            (double)i / (double)numSteps;

                    curPt = evaluate(u, brkPoint);

                    // interpolate u values of evaluated points to get
                    // an estimated u value for the nearest point
                    t = prStatic.PointSegDist(lastPt, curPt, hitPt, curDist); // prSplineRoutines

                    if (curDist.fValue < curSpan.fNearest)
                    {
                        curSpan.fNearest = curDist.fValue;
                        curSpan.fNearestU = last_u + (u - last_u) * t;
                        curSpan.fStepNum = i;
                    }
                    if (curDist.fValue < nearestDist)
                    {
                        nearestSpanInd = span;
                        nearestDist = curDist.fValue;
                    }

                    last_u = u;
                    lastPt = curPt;
                }
            }
        }

        TSpanBounds nearestSpan = spanInfo[nearestSpanInd];
        brkPoint = nearestSpan.fBrkPoint;

        // If param is already at end of span, then hitpoint was beyond
        // the span (entire curve or discon segment of it), so it's good enough
        // Otherwise, further polish the result.

        if ((nearestSpan.fStepNum != i || nearestSpan.fNearestU != knots[brkPoint]) &&
            (nearestSpan.fStepNum != numSteps || nearestSpan.fNearestU != knots[brkPoint + 1]))
        {
            // Now that we've found where on the curve the nearest point is,
            // polish the result by doing the same number of steps over a small
            // region near what we found above.  A more scientific (& less
            // heuristic) way of doing this would be to use function maximization
            // techniques, like those in Numerical recipes (ch 10).  We can even
            // get the derivative of the distance function for this.  However,
            // from the description in N.R., it's not clear such an approach would
            // be much more robust (need to test it someday, though).

            // Find space around approximation found above
            double subspanstep =
                (knots[brkPoint+1] - knots[brkPoint]) / (double)(numSteps * numSteps);
            last_u = nearestSpan.fNearestU - subspanstep * (numSteps / 2);

            // Hmmm - should really only do these checks near a c0 discon!
            if (last_u < knots[brkPoint])
                last_u = knots[brkPoint];
            if (last_u + subspanstep * numSteps > knots[brkPoint + 1])
                last_u = knots[brkPoint + 1] - subspanstep * numSteps;

            // Step near the guess, repeating process above
            nearestDist = Double.POSITIVE_INFINITY;
            lastPt = evaluate(last_u, brkPoint);
            for (i = 1; i <= numSteps; i++) {
                u = last_u + subspanstep;

                curPt = evaluate(u, brkPoint);

                t = prStatic.PointSegDist(lastPt, curPt, hitPt, curDist); // prSplineRoutines

                if (curDist.fValue < nearestDist) {
                    nearestDist = curDist.fValue;
                    nearestSpan.fNearestU = last_u + (u - last_u) * t;
                }
                last_u = u;
                lastPt = curPt;
            }
        }

        u = nearestSpan.fNearestU;  // Save result before deleting array...
        return u;
    }
*/

    // Given an index into the KnotVector, return the next
    // disconinuity of the type requested (c0, C1, C2 ... Corder-1).
    // Default is c0 (a break in the curve).
    //
    // nextDiscontinuity returns the index of the first knot of a
    // multiple knot sequence.  Call getKnot on this value to find the
    // parameter of the curve at the discontinuity.  Note that for a
    // c0 break in the curve, the value returned by
    // getKnot(nextDiscontinuity(...)) is the start of the section
    // of the curve after the break.  Subtracting any amount from that
    // jumps to the section before the break).
    //
    // To find the control point associated with a C1 discontinuity,
    // subtract one from the value returned by nextDiscontinuity.
    // For a c0 discontinuity, nextDiscontinuity(...)-1 is the
    // point before the break, and the following point is after the
    // break.
    //
    // To skip from one discontinuity to the next, you can pass the value
    // returned from the last call as the "startIndex" parameter.
    // If there are no more discontinuities in the curve, numberKnots()-1 is
    // returned (the last valid knot).  Assertions fail if the discontinuity
    // parameter is not 0 <= discon < order.

    public int getNextDiscontinuity(int startIndex) {
        return getNextDiscontinuity(startIndex, 0);
    }

    public int getNextDiscontinuity(int startIndex, int discontinuity) {
        return
            prStatic.GetNextDiscon( // prKnots
                startIndex, discontinuity, fOrder, fNumPoints, fKnots.getArray());
    }

/*================================================================
// Modelling tools
//
*/

    // Return the section of a curve, in the parametric range "from" to "to".
    // Note if "to" < "from", then the direction of the subsection is reversed.
    public void getSectionOfCurve(double MinValue, double MaxValue, TBaseCurve outCurve)
    {
        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

        // This creates a new curve which has a pinned knotvector with
        // range MinValue..MaxValue by refining a copy of this. The
        // (*this) curve is unchanged. While the method checks for knot
        // equality (with MinValue or MaxValue) it does not change for
        // knot nearness, and so may leave a controlpoint too much in the
        // curve if there is already a knot value very close
        // to Min or MaxValue

        int i, j, m, s1, s2;
        int nrKnots;
        boolean reversed = false;

        // No range spanned, just return nothing.
        if (MinValue == MaxValue) {
            outCurve.copyFrom(new TBaseCurve());
            return;
        }

        // Note the fact if the result runs backwards from the original
        if (MaxValue < MinValue) {
            double tmp = MinValue;
            MinValue = MaxValue;
            MaxValue = tmp;
            reversed = true;
        }

        // clamp requested parameters to curve\'s range
        if (MinValue < getMinParameter())
            MinValue = getMinParameter();
        if (MaxValue > getMaxParameter())
            MaxValue = getMaxParameter();

        // If the range spans the entire curve, just return it.
        if (MinValue == getMinParameter() && MaxValue == getMaxParameter()) {
            outCurve.copyFrom(this);
            return;
        }

        TBaseCurve tmpCurve = new TBaseCurve(this);

        nrKnots = fOrder + fNumPoints;
        TGParametricArray newKnots = new TGParametricArray(nrKnots + 2*fOrder);
        double nknots[] = newKnots.getArray();
        double tknots[] = tmpCurve.fKnots.getArray();

        for (i = 0, j = 0;  tknots[i] < MinValue; i++, j++)
            nknots[j] = tknots[i];  // copy all knots < MinValue

        s1 = j; // save index at which fOrder-fold knot at MinValue starts
        for (m = 0; tknots[i] == MinValue; m++, i++, j++)
            nknots[j] = tknots[i];  // copy all knots < MinValue
        for (; m < fOrder; m++, j++)
            nknots[j] = MinValue;   // add fOrder-m knots at MinValue (total fOrder-fold knot)
        for (; tknots[i] < MaxValue; i++, j++)
            nknots[j] = tknots[i];  // copy all knots < MaxValue
        for (m = 0; (i < tknots.length) && tknots[i] == MaxValue; m++, i++, j++)
            nknots[j] = tknots[i];  // copy m-fold knot AT MaxValue
        for (; m < fOrder; m++, j++)
            nknots[j] = MaxValue;   // add fOrder-m knots at MaxValue (total fOrder-fold knot)

        s2 = j; // save index at which fOrder-fold knot at MaxValue ends
        for (; i < nrKnots; i++, j++)
            nknots[j] = tknots[i];  // copy rest of knots > MaxValue
        newKnots.resize(j);
        tmpCurve.refine(newKnots);

        // copycopycopycopy...
        int newNumKnots = s2 - s1;
        newKnots.resize(newNumKnots);    // reuse the newKnots array allocation...
        System.arraycopy(tknots, s1, newKnots.getArray(), 0, newNumKnots);

        int newNumPoints = newNumKnots - fOrder;
        TGRPointArray newPoints = tmpCurve.fPoints;    // reuse the tmpCurve allocation...
        if (s1 > 0) {
            TGRPoint pts[] = newPoints.getArray();
            System.arraycopy(pts, s1, pts, 0, newNumPoints);
        }
        newPoints.resize(newNumPoints);

	// TBaseCurve bcurve = new TBaseCurve(fOrder, newPoints, newKnots);
	// outCurve.copyFrom(bcurve);
	// may be this would be more efficient:
        outCurve.fNumPoints = newNumPoints;
        outCurve.fPoints = newPoints;
        outCurve.fKnots = newKnots;
        outCurve.fOrder = fOrder;

        if (reversed)
            outCurve.reverseDirection();
    }

    // CurveConcat adds given a curve at the end of the current curve
    // glue parameter determines whether joining beginning and
    // endpoints should be merged (glue==true) or left as double
    // control points.
    // add(TGRPoint) adds a linear piece between the end of the
    // curve and that point.
    public void concatenate(TBaseCurve p) {
        concatenate(p, true);
    }
    public void concatenate(TBaseCurve p, boolean glue) {
        if (fOrder == 0) {
            copyFrom(p);
            return;
        }
        TBaseCurve copy = new TBaseCurve(p);
        if (!copy.isPinned())
            copy.refineToPinned();
        prAppendCurve(copy, glue);
    }

/*
    // Change curve shape so that at parameter u, it goes through ToPoint
    public void dragPosition(
        double u,
        TGPoint ToPoint,
        double segmentMinimum,
        double segmentMaximum)
    {
        // Drag a BSpline curve around using the method described in "A Technique for
        // the Direct Manipulation of Spline Curves" by Bartels & Beatty, in Graphics
        // Interface '89 (p.33)

        double bvals[] =  new double[fOrder];
        double theKnots[] = fKnots.getArray();
        TGRPoint fromPt = new TGRPoint(evaluate(u));    // Find point actually on the curve
        TGRPoint delta = ((TGRPoint)ToPoint.clone()).subtractBy(fromPt);

        int brkPoint = prStatic.FindInterval(u, theKnots, fNumPoints, fOrder); // prKnots

                                            // Find position along the span
        double t = (u - theKnots[brkPoint]) / (theKnots[brkPoint+1] - theKnots[brkPoint]);

                                            // This specifies all points in the span move
        int first = brkPoint - (fOrder - 1);
        int last = brkPoint;

        if (fOrder != 2 || t < segmentMinimum || t > segmentMaximum) {
            first++;                        // Now only interior points moved...
            last--;
        }

        if (t < segmentMinimum)             // Example of affecting the points moved by
            first--;                        // position within the span
        if (t > segmentMaximum)
            last++;

        prStatic.ComputeSplineValues(u, brkPoint, fOrder, theKnots, bvals); // prKnots

        double denom = 0.0;
        int i, j;
        for (i = first, j = brkPoint - first; i <= last; i++, j--)
            denom += bvals[j] * bvals[j];    // Square(bvals[j]);

        TGRPoint pts[] = fPoints.getArray();
        TGRPoint temp = fromPt;    // reuse the allocation...
        for (i = first, j = brkPoint - first; i <= last; i++, j--) {
            bvals[j] /= denom;
            pts[i].addBy(temp.copyFrom(delta).multiplyBy(bvals[j]));
        }
    }

    // Like DragCurve, but instead modifies the -tangent- at the u
    // parameter given.  i.e., Adobe rocker arms you can put anyplace
    // on the curve.
    public void dragTangent(
        double u,
        TGPoint toTangent,
        double segmentMinimum,
        double segmentMaximum)
    {
        // Like dragPosition, but manipulates the tangent.  This comes from
        // Barry Fowler's extensions to Bartel's original paper above.

        TGPoint fromTangent = new TGPoint();
        double bvals[]= new double[fOrder];
        double dvals[]= new double[fOrder];

        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

        if (fOrder <= 2)
            throw new IndexOutOfBoundsException("curve order below range");

        evaluate(u, fromTangent);

        // Note - fW should wind up as (1 - 1) = zero.
        TGRPoint t0 = new TGRPoint(toTangent);
        TGRPoint t1 = new TGRPoint(fromTangent);
        TGRPoint delta = TGRPoint.subtract(t0, t1);

        double knots[] = fKnots.getArray();
        int brkPoint = prStatic.FindInterval(u, knots, fNumPoints, fOrder); // prKnots
        int first = brkPoint - (fOrder - 1); // This moves all points in span
        int last = brkPoint;

        // Find position along the span
        double t = (u - knots[brkPoint]) / (knots[brkPoint + 1] - knots[brkPoint]);

        // Determine how many control points to move
        if ((t < segmentMinimum) && ((brkPoint - (first-1)) < fOrder))
            first--;
        if ((t > segmentMaximum) && (last + 1 <= brkPoint))
            last++;

        int i, j;
        for (i = 0; i < fOrder; i++)
            dvals[i] = 0.0;

        prStatic.ComputeSplineValues(u, brkPoint, fOrder, knots, bvals, dvals, null); // prKnots

        double b00 = 0.0;
        double b01 = 0.0;
        double b11 = 0.0;
        for (i = first, j = brkPoint - first; i <= last; i++, j--)
        {
            b00 += bvals[j] * bvals[j];
            b01 += bvals[j] * dvals[j];
            b11 += dvals[j] * dvals[j];
        }

        double denom = b00 * b11 - b01 * b01;

        TGRPoint pts[] = fPoints.getArray();
        for (i = first, j = brkPoint - first; i <= last; i++, j--) {
            bvals[j] = (-bvals[j] * b01 + dvals[j] * b00) / denom;
            pts[i].addBy(t0.copyFrom(delta).multiplyBy(bvals[j]));
        }
    }
*/

    // Reverse the order of a curve's points so the parameterization
    // runs the opposite direction.  The geometric shape is unchanged.
    public void reverseDirection()
    {
        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

        // flip control points
        TGRPoint pts[] = fPoints.getArray();
        int i, j;
        for (i = 0, j = fNumPoints - 1; i < j; i++, j--) {
            TGRPoint tmp = pts[i];
            pts[i] = pts[j];
            pts[j] = tmp;
        }

        // now flip knotvector too
        double knots[] = fKnots.getArray();
        j = fNumPoints + fOrder - 1;
        double offset = knots[j] + knots[0];
        for (i = 0; i < j; i++, j--) {
            double tmpk = knots[i];
            knots[i] = offset - knots[j];
            knots[j] = offset - tmpk;
        }
    }

/*================================================================
// Curve order manipulation
//
*/
    // Change order of curve. This DOES change the shape, but leaves
    // number points intact.
/*
    public void setOrder(int Order) {
        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

        if (Order <= 2 || Order > fNumPoints)
            throw new IndexOutOfBoundsException("curve order out of range");

        fOrder = Order;
        fKnots.resize(fOrder + fNumPoints);
        setKnotScheme(kPinned);   // setKnots updates the seed
    }
*/

    // Raise order of curve without changing its shape (changes number
    // points and knots)
    public void raiseOrder(int NewOrder) {
        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

         if (NewOrder > fOrder)
            for (int i = fOrder; i < NewOrder; i++)
                incrementOrder();
    }

/*
    // Approximate a higher order curve with a lower order one.
    // Control points are added to the lower order curve until the
    // approximation falls within "tolerance".  The curve in question
    // is Bezified first.  You can't lower something to linear
    // (newOrder must be > 2) or "lower" it to an order higher than it
    // already is (doing so fails an assertion). The tolerance is
    // specified in absolute coordinates (not relative to the size of
    // the curve).
    public void approximateLowerOrder(int newOrder) {
        approximateLowerOrder(newOrder, 0.2);
    }
    public void approximateLowerOrder(int newOrder, double tolerance) {
        if (newOrder == fOrder)
            return;

        if (newOrder <= 2 || newOrder > fOrder)
            throw new IndexOutOfBoundsException("curve order out of range");

        refineToBeziers();              // This makes the seed update calls

        double knots[] = fKnots.getArray();
        int discons[] = prStatic.MakeBezierDiscons(knots, fOrder, fNumPoints); // prKnots
        int numDiscons = discons.length;

        // Find and flag c0 discontinuities in the piecewise Bezier
        tolerance /= fOrder - newOrder;

        //
        // Reduce the order by repeatedly dropping the order by one.
        //

        for (; fOrder > newOrder; fOrder--)
            fNumPoints =
                TBaseCurve.reduceOneDegree(
                    fPoints, fOrder, fNumPoints, tolerance, discons, numDiscons);

        // reset the storage to match the actual size used.
        fPoints.resize(fNumPoints);

        // Rebuild the knot vector, taking into account the updated discon's.
        fKnots.resize(fNumPoints + fOrder);

        prStatic.CreateKnotsFromDiscons(knots, fOrder, fNumPoints, discons); // prKnots
    }
*/

/*================================================================
// Point and Knot vector manipulation
//
*/
    public void setPoint(int i, TGRPoint p) {
        fPoints.setPoint(i, p);
    }

    // Copies points in from your heap block (said heap block must
    // have same number of points)
/*
    public void setPoints(TGRPointArray srcPoints) {
        int numPoints = srcPoints.getNumberOfValues();

        if (numPoints < 2)
            throw new IndexOutOfBoundsException("curve number of points below range");

        // If the point array passed in isn't consistant with the existing one,
        // the the knot vector (and possibly the order) must be blow away and
        // reset.  It's the only humane thing to do...

        if (numPoints < fOrder)
            fOrder = numPoints;

        if (numPoints != fNumPoints)
        {
            fNumPoints = numPoints;
            fKnots.resize(numPoints + fOrder);
            setKnotScheme(kPinned);
        }
        fPoints.copyFrom(srcPoints);    // Does copy
    }
*/

    // set the knot vector (does NOT do any refinement)
    public void setKnotScheme(int theType) {
        /* set the type of knot vector (floating or pinned to endpoints)
           and the spacing of internal knots (uniform or chordlength) */

        int i;

        switch (theType) {
        case kPinned:
            prStatic.SetPinnedKnots(fOrder, fNumPoints, fKnots); // prKnots
            break;
        case kFloating:
            for (i = 0; i < fNumPoints+fOrder; i++)
                fKnots.setValue(i, (double)i);
            break;
        case kBezier:
            prStatic.SetBezierKnots(fOrder, fNumPoints, fKnots); // prKnots
            break;
        default: break;
        }
    }

/*
    // Implies same number of knots
    public void setKnots(TGParametricArray knots) {
        fKnots.copyFrom(knots);
    }

    // move the knot at index i to u. May resort knots to ensure
    // knot(i+1) >= knot(i) for all i
    public void moveKnot(int index, double u) {
        // This will move knot[i] to u. If u<knot[i-1], it will
        // resort knot vector. same for u>knot[i+1]

        prStatic.MoveAndSortKnot(index, u, fOrder, fNumPoints, fKnots); // prKnots
    }

    // Make compatible forces two (or N) curves to have the same
    // order, number of points, and knot vectors, without changing the
    // geometric shape of any of the curves.  This is accomplished by
    // raising the order and refining the knot vector as necessary.

    public static void makeCompatible(TBaseCurve crv1, TBaseCurve crv2) {

        TGParametricArray resultKnots = new TGParametricArray();

        // raise both curves to the same order
        int o1 = crv1.getOrder();
        int o2 = crv2.getOrder();

        if (o1 > o2)
            crv2.raiseOrder(o1);
        else if (o2 > o1)
            crv1.raiseOrder(o2);

        double Knotvector0[] = crv1.fKnots.getArray();
        int order0 = crv1.fOrder;
        int numKnots0 = crv1.getNumberOfKnots();

        double Knotvector[] = crv2.fKnots.getArray();
        int numKnots = crv2.getNumberOfKnots();

        // scale second knotvector range to be same as first ones
        double scale =
            (Knotvector0[numKnots0-1] - Knotvector0[0]) / (Knotvector[numKnots-1] - Knotvector[0]);

        double firstKnot = Knotvector[0];
        for (int i = 0; i < numKnots; i++)
            Knotvector[i] = (Knotvector[i] - firstKnot) * scale + Knotvector0[0];

        // merge second curve's knots into the first curves knotvector
        prStatic.privateMergeKnots( // prKnots
                        // use NumberPoints below, not numKnots0 !!
            Knotvector, numKnots, order0,    // crv2 stuff (src)
            Knotvector0, numKnots0, order0,  // crv1 stuff (dst)
            resultKnots);

        // refine both curves so they have the same (or-ed) knotvector
        crv1.refine(resultKnots);
        crv2.refine(resultKnots);
    }
*/

/*================================================================
// Refinement routines (add control points w/o changing shape)
//
*/
/*
    // Inserts a single new knot/point at given location.
    public void insertKnot(double u) {
        //  Insert a single knot at u using Boehm's algorithm (see CAD, July 1980, p.199)
        //  Note the knot must be in the valid parameter range (u[k-1] < u < u[m+1])

        TGParametricArray newKnots = new TGParametricArray(fNumPoints + fOrder + 1);
        double nknots[] = newKnots.getArray();
        TGRPointArray newControlPoints = new TGRPointArray(fNumPoints + 1);

        int el = prStatic.FindInterval(u, fKnots.getArray(), fNumPoints, fOrder); // prKnots
        int i;

        // Copy new knot into the w knot vector
        double knots[] = fKnots.getArray();
        for (i = 0; i <= el; i++)
            nknots[i] = knots[i];

        newKnots.setValue(el+1, u);
        for (i = el+1; i < fNumPoints+fOrder; i++)
            nknots[i+1] = knots[i];

        TGRPoint npts[] = newControlPoints.getArray();
        TGRPoint pts[] = fPoints.getArray();
        // Update the control points
        for (i = 0; i <= fNumPoints; i++) {
            if (i <= el-fOrder+1)
            npts[i] = pts[i];
            else if ((el-fOrder+2 <= i) && (i <= el)) {
                double omega = (u - nknots[i]) / (nknots[i+fOrder] - nknots[i]);
                double omega1 = 1.0 - omega;
                newControlPoints.setValue(
                    i,
                    TGRPoint.multiply(omega, pts[i]).addBy(TGRPoint.multiply(omega1, pts[i-1])));
            } else if (el+1 <= i)
                npts[i] = pts[i-1];
        }
        fPoints.swap(newControlPoints);
        fKnots.swap(newKnots);
        fNumPoints++;
    }
*/

    // Adds a Cn discontinuity at a parametric location u.  Values for n:
    //      kBreak (n=0), adds a break to the curve
    //      kKink (n=1), adds a kink to the curve
    //      kJerk (n=2), changes the parametric speed
    // If a discontinuity already exists at u, it is either left as is
    // or raised to the desired one.  Note this method just introduces
    // the discontinuity into the knot vector, it does not change the
    // shape of the curve.

/*
    public void makeDiscontinuity(double u, int cont) {

        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

        double knots[] = fKnots.getArray();

        if ((u < knots[0]) && (u > knots[fNumPoints + fOrder-1]))
            throw new IndexOutOfBoundsException("curve parametre out of range");

           int i, mult = 0;

        if (u == knots[fNumPoints]) {
            for (i = fNumPoints; (i < fNumPoints + fOrder) && (knots[i] == u); i++);
            i--;
        }
        else
            i = prStatic.FindInterval(u, knots, fNumPoints, fOrder); // prKnots

        // check to see if u matches existing knot, and if so, find multiplicty
        // of that knot.

        while ((i >= 0) && (knots[i] == u)) {
            mult++;
            i--;
        }

        for (i = 0; i < (fOrder - cont) - mult; i++)
            insertKnot(u);
    }

    // Inserts a new point halfway between each interval.
    public void refineUniform() {
        //  Inserts new knots between each nonzero internal knot interval. The fOrder
        //  boundary knots in the beginning and end are just copied. Note that the
        //  multiplicity of knots is not changed by this procedure.
        //  Implementation comment: numberknots below is worst case (no multiple knots=
        //  maximal refinement) The output index o contains the actual number of knots
        //  in the new knotvector.

        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

        TGParametricArray newKnots = new TGParametricArray(0);   // Will be swapped with the real KV

        prStatic.RefineKVUniform(fOrder, fNumPoints, fKnots.getArray(), newKnots); // prKnots

        refine(newKnots); // This updates the seed
    }
*/

    // Turn the curve into a collection of piecewise Bezier segments,
    // a la PostScript.  Adds control points, but does not change the shape.
    public void refineToBeziers()
    {
        //
        // Make the curve a piecewise Bezier by making all interier knots
        // multiplicity order-1
        //

        if (fOrder == 0)
            throw new IllegalArgumentException("curve not initialized");

        TGParametricArray newKnots = new TGParametricArray(0);   // Will be swapped with the real KV

        if (!isPinned())
            refineToPinned();       // Do ends first

        prStatic.RefineKVToBezier(fOrder, fNumPoints, fKnots.getArray(), newKnots); // prKnots

        // Only refine if new knots added
        if (newKnots.getNumberOfValues() != fOrder + fNumPoints)
            refine(newKnots); // This updates the seed
    }

    // Inserts new points at given locations
    public void refine(TGParametricArray newKnots)
    {
        int numberKnots = newKnots.getNumberOfValues();

        double knots[] = fKnots.getArray();
        double nknots[] = newKnots.getArray();

        TGRPointArray newControlPoints = new TGRPointArray(numberKnots - fOrder);

        int i;
        {    // scope for the alpha matrix:
            double alpha[][] =
                prStatic.ComputeAlphaMatrix(knots, nknots, fNumPoints, numberKnots, fOrder); // prKnots

            // now calculate the new control points using the alpha matrix and
            // the old control points
            TGRPoint npts[] = newControlPoints.getArray();
            TGRPoint pts[] = fPoints.getArray();
            for (int j = 0; j < numberKnots - fOrder; j++) {
                TGRPoint dp = npts[j];                // sp = src ptr, dp = dest ptr
                TGRPoint sp;

                int delta = fNumPoints - 1 + fOrder;    // source = old fPoints
                while ((nknots[j] < knots[delta]) && (delta > 0))
                    delta--;
                int i1 = delta - fOrder + 1;
                if (i1 < 0)
                    i1 = 0;
                TGRPoint.SetToZero(dp);
                for (i = i1; i <= delta; i++) {
                    sp = pts[i];
                    dp.addBy(TGRPoint.multiply(sp, alpha[i - i1][j]));
                }
            }
            fNumPoints = numberKnots - fOrder;              // inform curve of new cp's

            // Free up the alpha matrix ... I hope
        }

        fKnots.resize(fOrder + fNumPoints);
        knots = fKnots.getArray();
        System.arraycopy(nknots, 0, knots, 0, fOrder + fNumPoints); // copy incoming new knots (user kills own copy)

        fPoints.swap(newControlPoints);            // swap old controlpoints with the new set of control points
    }

    // Creates a pinned curve & knotvector. Does not change shape of the curve
    public void refineToPinned()
    {
        /* only applies to floating end condition curves. Creates a new curve with the
           same shape, but pinned end conditions */

        if (isPinned())             // Nothing to do
            return;

        TGParametricArray newKnots = new TGParametricArray(fOrder + fNumPoints);

        prStatic.RefineKVToPinned(fOrder, fNumPoints, fKnots, newKnots); // prKnots
        refine(newKnots);     // This updates the seed
    }

/*================================================================
// Debugging
//
*/
/*
    // Checks for invalid knot vectors, W's, duplicate points,
    // and values grossly out of range.
    public boolean isConsistent() {
        boolean ok=true;
        int i;
        TGRPoint p;
        TGRPoint lastp = new TGRPoint();

        if (fNumPoints != fPoints.getNumberOfValues()) {
            // printf("Curve check: fNumPoints: %d != fPoints.fNumPoints %d\n",
                 // fNumPoints, fPoints.numberOfValues());
            ok = false;
        }

        if (fOrder<2 || fOrder>20) {
            // printf("Curve check: Order=%ld (a little too small or large ?)\n", fOrder);
            ok = false;
        }

        if (fNumPoints<2) {
            // printf("Curve check: numPoints=%ld (so few?)\n", fNumPoints);
            ok = false;
        }
        if (fNumPoints>3000) {
            // printf("Curve check: numPoints=%ld(really so many ?)\n", fNumPoints);
            ok = false;
        }

        if (!prStatic.IsKnotvectorConsistent(fKnots.getArray(), fNumPoints+fOrder, fOrder)) { // prKnots
            // printf("Curve: Bad knotvector\n");
            ok = false;
        }

        for (i=0; i<fNumPoints; i++) {
            p = fPoints.getValue(i);
            if (p.fW <= 0.0) {
            // printf("Curve check: point[%ld] = ", i);
            // printf("%lf, %lf, %lf bad fW!\n", p.fX, p.fY, p.fW);
            ok = false;
            }

            if (i>0) {
                if (lastp==p) {
                    // printf("Curve check: warning, points %ld and %ld are the same\n",i-1,i);
                }
            }

            lastp.copyFrom(p);
        }

        return ok;
    }
*/

    /*
     * Convert a k order spline to a k+1 order spline.  This is from the paper
     * "Algorithms for Degree-Raising of Splines" by Cohen, Lyche & Schumaker
     * in the July '85 TOG.  Note this procedure generates spurious control points
     * splines with other than "Closed" end conditions (k-fold knots at each end).
     *
     */

    protected void incrementOrder() {
        int order = fOrder;            // Prevent unsignedLong/long confusion

        /* Allocate an array to flag c0 discontinuities
         * (published alg tramples them) */

        int discon[] = new int[2 * fNumPoints + order + 1];
        TGParametricArray newknots = new TGParametricArray(2 * fNumPoints + order + 2);

        int i, j;                     // Most of these names are from the paper...
        for (j = 0; j < 2  * fNumPoints + order + 1; j++)
            discon[j] = 0;

        /*
         * extend the knot vector as per section 2 of the paper.  Although the
         * paper doesn't make it clear, what it means is "increase the
         * multipicity of each knot by one".
         */

        int mult = 0;
        j = 0;
        double nknots[] = newknots.getArray();
        double knots[] = fKnots.getArray();
        double lastKnot = knots[0];
        for (i = 0; i < fNumPoints + order; i++) {
            if (lastKnot != knots[i]) {
                mult = 1;
                nknots[j] = nknots[j - 1];  /* Increase multiplicty by one */
                j++;
            }
            else {
                mult++;
                /*
                 * If a c0 discontinuty is found, then j - k is the point in the
                 * new curve that must be set to the i - k'th point of the old
                 * curve.
                 */
                if ((mult == order) && (j > order))
                    discon[j - order] = i - order;
            }
            nknots[j] = knots[i];
            j++;
            lastKnot = knots[i];
        }
        nknots[j] = nknots[j - 1];

        /* Realloc newknots to reflect the actual number used. */
        newknots.resize(j + 1);

        TGRPointArray newpoints = new TGRPointArray(j - order);
        TGRPoint cc[] = (new TGRPointArray(order + 2)).getArray();
        TGRPoint cw[] = (new TGRPointArray(order + 2)).getArray();

        /* Endpoints remain the same */
        TGRPoint pts[] = fPoints.getArray();
        newpoints.setPoint(0, pts[0]);
        newpoints.setPoint(j - order - 1, pts[fNumPoints-1]);
        fNumPoints = j - order;

        int numPoints = fNumPoints;     // Convert to long due to Max/Min bugs

        /*
         * Note: the algorithm differs somewhat from the paper in that zero based
         * arrays are used instead of one based arrays (gotta watch them FORTRAN
         * geeks...)
         */

        double yi, yik, y1, y2, den;
        for (j = 1; j < numPoints - 1; j++) {
            /* Manually skip c0 discontinutities */

            if (discon[j] != 0) {
                newpoints.setValue(j, pts[discon[j]]);
                j++;    /* Skip to next point */
            }

            int r = 0;
            while (!(nknots[j + 1] < knots[r + 1]))
                r++;
            r++;

            for (i = 0; i < order; i++) {
                TGRPoint.SetToZero(cc[i]);
                TGRPoint.SetToZero(cw[i]);
            }

            int lim = Math.min(order - 1, numPoints + order - r - 1);
            for (i = Math.max(0, order - r); i <= lim; i++) {
                cc[i].copyFrom(pts[i + r - order]);
                cw[i].copyFrom(cc[i]);
            }

            for (int el = order-1; el >= 1; el--) {
                y1 = nknots[j + el];
                y2 = nknots[j + el + 1];

                lim = Math.max(order - el, order - r);
                for (i = Math.min(order - 1, numPoints + 2 * order - el - r - 1); i >= lim; i--){
                    yi = knots[r + i - order];
                    yik = knots[r + i + el - order];
                    den = yik - yi;       // Elain assures us this won't blow up...
                    cc[i] =
                        TGRPoint.multiply(y2 - yi, cc[i]).addBy(
                            TGRPoint.multiply(yik - y2, cc[i-1])).divideBy(den);
                    cw[i] =
                        TGRPoint.multiply(y1 - yi, cw[i]).addBy(
                            TGRPoint.multiply(yik - y1, cw[i-1])).divideBy(den).addBy(cc[i]);
                }
            }
            newpoints.setValue(j, TGRPoint.divide(cw[order-1], (double)order));
        }

        fOrder++;

        /* Switch control points and kv to new ones & clean up */
        fPoints.swap(newpoints);
        fKnots.swap(newknots);
    }

    public void prAppendCurve(TBaseCurve crv, boolean glue)
    {
        // this is a private routine used by public add() methods only
        // crv is always a pinned kv curve that may be modified.

        if (!isPinned())               // if current curve (*this) is not pinned
            refineToPinned();           // do it now

        // make sure both curves are the same order (always preserve both
        // current and added curves shape)
        int newOrder = crv.getOrder();
        if (newOrder>fOrder)
            raiseOrder(newOrder);
        else if (fOrder>newOrder)
            crv.raiseOrder(fOrder);

        TGRPoint pts[] = fPoints.getArray();
        TGRPoint cps[] = crv.fPoints.getArray();

        // now sew the two curves together if the beginning- and endpoints
        // match up, only a fOrder-1 multiplicity knot is needed (and a
        // single cp at the join), otherwise a fOrder-fold knot is
        // inserted and thus a double control point (or two distinct ones)
        TGRPoint p = new TGRPoint(pts[fNumPoints-1]);
        TGRPoint q = new TGRPoint(cps[0]);

        int off;                   // glue & nonglue code can be folded with this variable
        int n = crv.getNumberOfPoints();
        double offset;         // by how much to offset new knot values

        if (glue && p.equals(q))
            off = 1;
        else
            off = 0;

        // extend pt array by n-1 points
        fPoints.resize(fNumPoints+n-off);
        pts = fPoints.getArray();

        // extend knot array by n-1 knots
        double knots[] = fKnots.getArray();
        TGParametricArray newknots = new TGParametricArray(fNumPoints + fOrder + n - off);
        System.arraycopy(knots, 0, newknots.getArray(), 0, fNumPoints + fOrder);

        // switch in the new knots vector...
        fKnots = newknots;
        knots = fKnots.getArray();

        // append knots of new curve onto *this
        double cknots[] = crv.fKnots.getArray();
        offset = knots[fOrder+fNumPoints-1] - cknots[0];
        int dst = fOrder + fNumPoints - off;
        int src = fOrder;

        int i;
        for (i = 0; i < n; i++, src++, dst++) {
            // note that for glue case (off=1) the first src knot over-
            // writes the last dst knot, so that knot's multiplicity
            // is reduced by 1 to fOrder-1
            knots[dst] = cknots[src] + offset;
        }

        // append control points of new curve onto *this
        dst = fNumPoints;
        src = off;
        for (i = 0; i < n - off; i++, src++, dst++) {
            pts[dst].copyFrom(cps[src]);
        }
        fNumPoints += n - off;
    }

/*
    //
    // Reduce a curve by one degree.  Assumes the input is a complete
    // peicewise Bezier curve of the specified order (does no error checking!)
    //

    // returns the number of points after modifying the point array
    static int reduceOneDegree(
        TGRPointArray points,
        int order,
        int numPts,
        double tolerance,
        int discons[],
        int numDiscons)
    {
        int degree = order - 1;         // For beziers, degrees are more natural
        int numSegs = (numPts - numDiscons - 1) / degree;

        // Make storage for tmp arrays...
        TGRPoint left[] = (new TGRPointArray(order)).getArray();
        TGRPoint right[] = (new TGRPointArray(order)).getArray();
        TGRPoint tmpPts[] = (new TGRPointArray(order)).getArray();

        TGRPoint endPoints[] = new TGRPoint[2];

        int curTmpChunk = 2;
        int origPts = numPts;
        points.resize(origPts * curTmpChunk);

        TCurveReductionMatrix.MakeReductionMatrices(order);
        TCurveReductionMatrix.MakeReductionMatrices(order-1);

        TGRPoint[] pts = points.getArray();
        endPoints[0] = pts[0];      // HEY - do this for discons too!
        int seg = 0;
        int curDiscon = 0;

        int i, j, ptrAt;
        while (seg < numSegs) {
            // If we've hit a discon, slide the indexing forward.
            if (seg == discons[curDiscon])
                curDiscon++;

            endPoints[1] = pts[seg * degree + degree + curDiscon];

            // Convert segment from Bezier to Chebycheff polynomial basis
            for (i = 0; i < order; i++)
            {
                ptrAt = seg * degree + curDiscon;

                TGRPoint.SetToZero(tmpPts[i]);

                for (j = 0; j < order; j++) {
                    tmpPts[i].addBy(
                        TGRPoint.multiply(
                            pts[ptrAt], TCurveReductionMatrix.BezToCheb[order][j][i]));
                    ptrAt++;
                }
            }

            // check the highest (unused) Chebycheff term to find the error
            double err = 0.0;
            err = Math.max(err, Math.abs(tmpPts[degree].fX));
            err = Math.max(err, Math.abs(tmpPts[degree].fY));

            // InternalAssert((err <= 100), "Degree reduction went bezerk (err > 100?, tolerence too small?)...");

            double subDivs;
            if (err == 0.0)
                subDivs = 0.;
            else
                subDivs = 1.0 / degree * prStatic.Log2(err / tolerance); // prNumeric

            if (subDivs > 1.0) {
                int r;
                //
                // If the error was too large, then the segment needs to
                // be split so we can try again on the pieces.
                //

                // May need more control point storage (ugh)
                if ((numSegs + 1) * degree + numDiscons + 1 >= origPts * curTmpChunk) {
                    curTmpChunk++;
                    points.resize(curTmpChunk * origPts);
                }

                // Make room for the new segment
                for (i = (numPts-1)+degree; i >= (seg + 2) * degree; i--) {
                    pts[i + curDiscon] = pts[i - degree + curDiscon];
                }

                // Now split the segment (ooh fer lexical scoping...)
                for (i = 0; i < order; i++) {
                    TGRPoint.SetToZero(left[i]);
                    TGRPoint.SetToZero(right[i]);

                    ptrAt = degree * seg + curDiscon;
                    for (r = 0; r <= i; r++) {
                        left[i].addBy(
                            TGRPoint.multiply(
                                pts[ptrAt],
                                prStatic.fchoose(i, r) / Math.pow(2, i))); // prSplineRoutines
                        ptrAt++;
                    }

                    ptrAt = i + degree * seg + curDiscon;
                    for (r = i; r <= degree; r++) {
                        right[i].addBy(
                            TGRPoint.multiply(
                                pts[ptrAt],
                                prStatic.fchoose(degree-i, degree-r) / Math.pow(2, degree - i))); // prSplineRoutines
                        ptrAt++;
                    }
                }

                // Copy subdivided parts back to original array
                for (i = 0; i <= degree; i++)
                    pts[i + degree * seg + curDiscon] = left[i];
                for (i = 0; i <= degree; i++)
                    pts[degree + i + degree * seg + curDiscon] = right[i];

                numSegs++;
                numPts = degree * numSegs + numDiscons + 1;

                // Update the discons array to account for the new segment
                i = curDiscon;
                while (discons[i] > 0) {
                    if (discons[i] > seg) // Only downstream
                        discons[i]++;
                    i++;
                }
            }
            else {
                // Now convert segment back to Bezier, dropping a term on the way
                for (i = 0; i < degree; i++) {
                    ptrAt = seg * (degree-1) + i + curDiscon;
                    TGRPoint.SetToZero(pts[ptrAt]);

                    for (j = 0; j < degree; j++)
                        pts[ptrAt].addBy(
                            TGRPoint.multiply(
                                tmpPts[j], TCurveReductionMatrix.ChebToBez[degree][j][i]));
                    ptrAt++;
                }

                // Force endpoints back to their original locations

                pts[seg * (degree-1) + (degree-1) + curDiscon] = endPoints[1];
                seg++;
            }
        }

        pts[0] = endPoints[0];

        return (degree-1) * numSegs + numDiscons + 1;
    }
*/

    /******************************************************************************
        TGCurve constructor for uniform cubic matrix splines
     ******************************************************************************/
/*
    void setBasisSpline(double B[], int step)
    {
        //  Basis Matrix transformation into B-Spline basis explained in
        // Patch Work, Rob Cook, Technical Memo #118, LucasFilm, in SIGGRAPH 86
        // Image Rendering Tricks Tutorial #16

        double convMatrix[] = new double[16];
        boolean c0Continous = (step == 1); // Dazed and confused....
        TGRPointArray points = new TGRPointArray(fPoints); // Use local copy

        int nrSegments = (fNumPoints - (4 - step)) / step;

        // construct Bezier^-1 basis matrix.
        // This was manually constructed
        double invBezier[] = {
            0.0, 0.0, 0.0, 1.0,
            0.0, 0.0, 2.0/6.0, 1.0,
            0.0, 2.0/6.0, 4.0/6.0, 1.0,
            1.0, 1.0, 1.0, 1.0 };

        // construct MM = Mto^-1Mfrom
        CubicBasisMatrixMul(invBezier, B, convMatrix); // prKnots

        // send points through matrix
        fOrder = 4;
        if (4 * nrSegments != fNumPoints) {
            fNumPoints = c0Continous ? 3 * nrSegments + 1 : 4 * nrSegments;
            fPoints.resize(fNumPoints);
            fKnots.resize(fNumPoints + fOrder);
        }

        int segSize = c0Continous ? 3 : 4;

        int src, dst, offset;
        int i,j;
        for (i = 0, src = 0, dst = 0; i < nrSegments; i++) {
            if (i == nrSegments - 1)
                segSize = 4;                // So last point is included
            for (j = 0; j < segSize; j++) {
                offset = j << 2;
                 fPoints.getValue(dst+j).fX =
                      convMatrix[offset]   * points.getValue(src).fX
                    + convMatrix[offset+1] * points.getValue(src+1).fX
                    + convMatrix[offset+2] * points.getValue(src+2).fX
                    + convMatrix[offset+3] * points.getValue(src+3).fX;

                 fPoints.getValue(dst+j).fY =
                      convMatrix[offset]   * points.getValue(src).fY
                    + convMatrix[offset+1] * points.getValue(src+1).fY
                    + convMatrix[offset+2] * points.getValue(src+2).fY
                    + convMatrix[offset+3] * points.getValue(src+3).fY;

                fPoints.getValue(dst+j).fW = 1.0;
            }
            dst += segSize;
            src += step;
        }

        double knots[] = fKnots.getArray();
        if (c0Continous) {
            dst = 0;
            knots[dst++] = 0.0;
            for (i = 0; i <= nrSegments; i++) {
                for (j = 0; j < 3; j++) {
                    knots[dst++] = (double)i;
                }
            }
            knots[dst] = (double)(i - 1);
        }
        else {
            dst = 0;
            for (i = 0; i <= nrSegments; i++) {
                for (j = 0; j < 4; j++) {
                    knots[dst++] = (double)i;
                }
            }
        }
    }
*/
}
