/*
 * @(#)TBezier.java	1.7 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/**
 * TBezier class is used to represent a cubic curve specified with
 * four points. It contains all the gory math dealing with Bezier
 * curves.
 *
 * @version     
 * @author      Taligent, Inc.
 * @see TBezierList
 * @since       JDK1.2
 */

class TBezier {

    // public TGPoint fPoint0;
    // public TGPoint fPoint1;
    // public TGPoint fPoint2;
    // public TGPoint fPoint3;

    // The drab un-object-oriented way in which the following data has been
    // described is to highlight the fact that we want to make this class as
    // efficient as possible in terms of storage and absolutely no dependency
    // on any other class of Albert. Having TGPoints here means added
    // garbage collection since we do a new of TGPoint for every creation
    // of this class.
    public double fPoint0_fX;
    public double fPoint0_fY;
    public double fPoint1_fX;
    public double fPoint1_fY;
    public double fPoint2_fX;
    public double fPoint2_fY;
    public double fPoint3_fX;
    public double fPoint3_fY;

    public TBezier() {
        // Do nothing here.
        // Its OK because these values will subsequently be filled in.
    }

    public TBezier(
        double point0X,
        double point0Y,
        double point1X,
        double point1Y,
        double point2X,
        double point2Y,
        double point3X,
        double point3Y) {
        fPoint0_fX = point0X;
        fPoint0_fY = point0Y;
        fPoint1_fX = point1X;
        fPoint1_fY = point1Y;
        fPoint2_fX = point2X;
        fPoint2_fY = point2Y;
        fPoint3_fX = point3X;
        fPoint3_fY = point3Y;
    }

/* not used
    public TBezier(TGPoint point0, TGPoint point1,
                    TGPoint point2, TGPoint point3) {
        fPoint0_fX = point0.x;
        fPoint0_fY = point0.y;
        fPoint1_fX = point1.x;
        fPoint1_fY = point1.y;
        fPoint2_fX = point2.x;
        fPoint2_fY = point2.y;
        fPoint3_fX = point3.x;
        fPoint3_fY = point3.y;
    }

    public TBezier(TBezier bezCurve) {
        fPoint0_fX = bezCurve.fPoint0_fX;
        fPoint0_fY = bezCurve.fPoint0_fY;
        fPoint1_fX = bezCurve.fPoint1_fX;
        fPoint1_fY = bezCurve.fPoint1_fY;
        fPoint2_fX = bezCurve.fPoint2_fX;
        fPoint2_fY = bezCurve.fPoint2_fY;
        fPoint3_fX = bezCurve.fPoint3_fX;
        fPoint3_fY = bezCurve.fPoint3_fY;
    }
*/

    private static double MeanValue(double p1, double p2) {
        return (p1 + p2) / 2.0;
    }
    public TBezier[] Split() {
        TBezier[] r = new TBezier[2];
        r[0] = new TBezier();
        r[1] = new TBezier();

        r[0].fPoint0_fX = fPoint0_fX;
        r[0].fPoint0_fY = fPoint0_fY;

        r[1].fPoint3_fX = fPoint3_fX;
        r[1].fPoint3_fY = fPoint3_fY;

        r[0].fPoint1_fX = MeanValue(fPoint0_fX, fPoint1_fX);
        r[0].fPoint1_fY = MeanValue(fPoint0_fY, fPoint1_fY);

        r[1].fPoint2_fX = MeanValue(fPoint2_fX, fPoint3_fX);
        r[1].fPoint2_fY = MeanValue(fPoint2_fY, fPoint3_fY);

        r[1].fPoint1_fX = MeanValue(fPoint1_fX, fPoint2_fX); // temporary holding spot
        r[1].fPoint1_fY = MeanValue(fPoint1_fY, fPoint2_fY); // temporary holding spot

        r[0].fPoint2_fX = MeanValue(r[0].fPoint1_fX, r[1].fPoint1_fX);
        r[0].fPoint2_fY = MeanValue(r[0].fPoint1_fY, r[1].fPoint1_fY);

        r[1].fPoint1_fX = MeanValue(r[1].fPoint1_fX, r[1].fPoint2_fX); // Real value this time
        r[1].fPoint1_fY = MeanValue(r[1].fPoint1_fY, r[1].fPoint2_fY); // Real value this time

        r[1].fPoint0_fX = MeanValue(r[0].fPoint2_fX, r[1].fPoint1_fX);
        r[1].fPoint0_fY = MeanValue(r[0].fPoint2_fY, r[1].fPoint1_fY);

        r[0].fPoint3_fX = MeanValue(r[0].fPoint2_fX, r[1].fPoint1_fX);
        r[0].fPoint3_fY = MeanValue(r[0].fPoint2_fY, r[1].fPoint1_fY);

        return r;
    }

    static double evalValue(double parametric, double f0, double f1) {
        return f0 + parametric * (f1 - f0);
    }
    private void ParameterSplitLeft(double t, TBezier left) {
        left.fPoint0_fX = fPoint0_fX;
        left.fPoint0_fY = fPoint0_fY;

        left.fPoint1_fX = evalValue(t, fPoint0_fX, fPoint1_fX);
        left.fPoint1_fY = evalValue(t, fPoint0_fY, fPoint1_fY);

        left.fPoint2_fX = evalValue(t, fPoint1_fX, fPoint2_fX); // temporary holding spot
        left.fPoint2_fY = evalValue(t, fPoint1_fY, fPoint2_fY); // temporary holding spot

        fPoint2_fX = evalValue(t, fPoint2_fX, fPoint3_fX);
        fPoint2_fY = evalValue(t, fPoint2_fY, fPoint3_fY);

        fPoint1_fX = evalValue(t, left.fPoint2_fX, fPoint2_fX);
        fPoint1_fY = evalValue(t, left.fPoint2_fY, fPoint2_fY);

        left.fPoint2_fX = evalValue(t, left.fPoint1_fX, left.fPoint2_fX);
        left.fPoint2_fY = evalValue(t, left.fPoint1_fY, left.fPoint2_fY);

        fPoint0_fX = evalValue(t, left.fPoint2_fX, fPoint1_fX);
        fPoint0_fY = evalValue(t, left.fPoint2_fY, fPoint1_fY);

        left.fPoint3_fX = fPoint0_fX;
        left.fPoint3_fY = fPoint0_fY;
    }

    // IMPORTANT COPYRIGHT NOTE :
    // The algorithm below for finding the intersection of two Bezier curves
    // has been taken from the article "Intersecting Parametric Cubic Curves
    // by Midpoint Subdivision" by R Victor Klassen from the book Graphics Gems
    // Volume IV. The article appears on Page 261. The C++ implementation has
    // been changed to a Java implementation by Anupam Bagchi, Taligent Inc.

    // Intersect with another curve.  Return two 10-element arrays. Array 0
    // contains fragments of self. Array 1 contains fragments of other curve.
    // Fragments continue until one with nil pointers pointing at point data.
    /*
     * Intersect two curves, returning an array of two arrays of curves.
     * The first array of curves corresponds to `this' curve, the second
     * corresponds to curve B, passed in.
     * The intersection parameter values are computed by findIntersections,
     * and they come back in the range 0..1, using the original parameterization.
     * Once one segment has been removed, ie the curve is split at splitT, the
     * parameterization of the second half is from 0..1, so the parameter for
     * the next split point, if any, must be adjusted.
     * If we split at t[i], the split point at t[i+1] is
     * (t[i+1] - t[i]) / (t - t[i]) of the way to the end from the new
     * start point.
     */
/*
    private TBezier[][] Intersect(TBezier B) {

        TBezier[][] rvalue = new TBezier [2][10];
        int i, j;
        for (i=0; i<2; ++i) {
            for (j=0; j<10; ++j) {
                rvalue[i][j] = new TBezier();
            }
        }

        double[][] t = findIntersections(this, B);

        int index = 0;
        if (t[0][0] > -0.5) {
            ParameterSplitLeft(t[0][0], rvalue[0][0]);
            B.ParameterSplitLeft(t[1][0], rvalue[1][0]);
            index++;
            while ((index < 9) && (t[0][index] > -0.5)) {
                double splitT = (t[0][index] - t[0][index-1])/(1.0 - t[0][index-1]);
                ParameterSplitLeft(splitT, rvalue[0][index]);
                splitT = (t[1][index] - t[1][index-1])/(1.0 - t[1][index-1]);
                B.ParameterSplitLeft(splitT, rvalue[1][index]);
                index++;
            }
        }
        rvalue[0][index] = this;
        rvalue[1][index] = B;
        return rvalue;
    }
*/

    /*
    * Test the bounding boxes of two Bezier curves for interference.
    * Several observations:
    *    First, it is cheaper to compute the bounding box of the second curve
    *    and test its bounding box for interference than to use a more direct
    *    approach of comparing all control points of the second curve with
    *    the various edges of the bounding box of the first curve to test
    *     for interference.
    *    Second, after a few subdivisions it is highly probable that two corners
    *    of the bounding box of a given Bezier curve are the first and last
    *    control point.  Once this happens once, it happens for all subsequent
    *    subcurves.  It might be worth putting in a test and then short-circuit
    *    code for further subdivision levels.
    *    Third, in the final comparison (the interference test) the comparisons
    *    should both permit equality.  We want to find intersections even if they
    *    occur at the ends of segments.
    *    Finally, there are tighter bounding boxes that can be derived. It isn't
    *    clear whether the higher probability of rejection (and hence fewer
    *    subdivisions and tests) is worth the extra work.
    */
    private static boolean IntersectBB(TBezier a, TBezier b) {
        // Compute bounding box for a
        double minax, maxax, minay, maxay;
        if (a.fPoint0_fX > a.fPoint3_fX) { // These are the most likely to be extremal
            minax = a.fPoint3_fX;
            maxax = a.fPoint0_fX;
        }
        else {
            minax = a.fPoint0_fX;
            maxax = a.fPoint3_fX;
        }

        if (a.fPoint2_fX < minax)
            minax = a.fPoint2_fX;
        else if (a.fPoint2_fX > maxax)
            maxax = a.fPoint2_fX;

        if (a.fPoint1_fX < minax)
            minax = a.fPoint1_fX;
        else if (a.fPoint1_fX > maxax)
            maxax = a.fPoint1_fX;


        if (a.fPoint0_fY > a.fPoint3_fY) {
            minay = a.fPoint3_fY;
            maxay = a.fPoint0_fY;
        }
        else {
            minay = a.fPoint0_fY;
            maxay = a.fPoint3_fY;
        }

        if (a.fPoint2_fY < minay)
            minay = a.fPoint2_fY;
        else if (a.fPoint2_fY > maxay)
            maxay = a.fPoint2_fY;

        if (a.fPoint1_fY < minay)
            minay = a.fPoint1_fY;
        else if (a.fPoint1_fY > maxay)
            maxay = a.fPoint1_fY;


        // Compute bounding box for b
        double minbx, maxbx, minby, maxby;
        if (b.fPoint0_fX > b.fPoint3_fX) {
            minbx = b.fPoint3_fX;
            maxbx = b.fPoint0_fX;
        }
        else {
            minbx = b.fPoint0_fX;
            maxbx = b.fPoint3_fX;
        }
        if (b.fPoint2_fX < minbx)
            minbx = b.fPoint2_fX;
        else if (b.fPoint2_fX > maxbx)
            maxbx = b.fPoint2_fX;

        if (b.fPoint1_fX < minbx)
            minbx = b.fPoint1_fX;
        else if (b.fPoint1_fX > maxbx)
            maxbx = b.fPoint1_fX;

        if (b.fPoint0_fY > b.fPoint3_fY) {
            minby = b.fPoint3_fY;
            maxby = b.fPoint0_fY;
        }
        else {
            minby = b.fPoint0_fY;
            maxby = b.fPoint3_fY;
        }
        if (b.fPoint2_fY < minby)
            minby = b.fPoint2_fY;
        else if (b.fPoint2_fY > maxby)
            maxby = b.fPoint2_fY;
        if (b.fPoint1_fY < minby)
            minby = b.fPoint1_fY;
        else if (b.fPoint1_fY > maxby)
            maxby = b.fPoint1_fY;

        // Test bounding box of b against bounding box of a
        if ((minax > maxbx) || (minay > maxby)  // Not >= : need boundary case
           || (minbx > maxax) || (minby > maxay))
            return false; // they don't intersect
        else
            return true; // they intersect
    }


    /*
    * Recursively intersect two curves keeping track of their real parameters
    * and depths of intersection.
    * The results are returned in a 2-D array of doubles indicating the parameters
    * for which intersections are found.  The parameters are in the order the
    * intersections were found, which is probably not in sorted order.
    * When an intersection is found, the parameter value for each of the two
    * is stored in the index elements array, and the index is incremented.
    *
    * If either of the curves has subdivisions left before it is straight
    *    (depth > 0)
    * that curve (possibly both) is (are) subdivided at its (their) midpoint(s).
    * the depth(s) is (are) decremented, and the parameter value(s) corresponding
    * to the midpoints(s) is (are) computed.
    * Then each of the subcurves of one curve is intersected with each of the
    * subcurves of the other curve, first by testing the bounding boxes for
    * interference.  If there is any bounding box interference, the corresponding
    * subcurves are recursively intersected.
    *
    * If neither curve has subdivisions left, the line segments from the first
    * to last control point of each segment are intersected.  (Actually the
    * only the parameter value corresponding to the intersection point is found).
    *
    * The apriori flatness test is probably more efficient than testing at each
    * level of recursion, although a test after three or four levels would
    * probably be worthwhile, since many curves become flat faster than their
    * asymptotic rate for the first few levels of recursion.
    *
    * The bounding box test fails much more frequently than it succeeds, providing
    * substantial pruning of the search space.
    *
    * Each (sub)curve is subdivided only once, hence it is not possible that for
    * one final line intersection test the subdivision was at one level, while
    * for another final line intersection test the subdivision (of the same curve)
    * was at another.  Since the line segments share endpoints, the intersection
    * is robust: a near-tangential intersection will yield zero or two
    * intersections.
    */
    private static void RecursivelyIntersect(TBezier a, double t0, double t1, int deptha,
                                              TBezier b, double u0, double u1, int depthb,
                                              double[][] parameters, int[] index) {

        /*
        System.out.println ("TBezier A: = " + a +
                            " isDegenerate =" + a.isDegenerate() +
                            " length = " + a.lengthEstimateSquare());
        System.out.println ("TBezier B: = " + b +
                            " isDegenerate =" + b.isDegenerate() +
                            " length = " + b.lengthEstimateSquare());

        if (a.isClose(b))
            System.out.println(" A is close to B");
        else
            System.out.println(" A and B are not close");
        */

// System.out.println("t0 = " + t0 + "    t1 = " + t1 + "       deptha = " + deptha);
// System.out.println("u0 = " + u0 + "    u1 = " + u1 + "       depthb = " + depthb);
    
        // There is a special case that has to be handled before anything else.
        // If the two curves are identical, then they theoretically intersect at
        // infinite points. The reason we want to intersect the two curves (recursively)
        // is because we would like seperate non-intersecting pieces to be generated
        // for the geometric sorting algorithm to pick-up (which is used to boolean
        // operation on geometries). Thus we can arbitrarily assign parameters
        // at which the curves may be broken safely without causing further
        // degradation.
/*
        if (a.isClose(b)) {
            index[0] = 7;
            parameters[0][0] = 0.1;
            parameters[0][1] = 0.2;
            parameters[0][2] = 0.4;
            parameters[0][3] = 0.5;
            parameters[0][4] = 0.6;
            parameters[0][5] = 0.8;
            parameters[0][6] = 0.9;

            parameters[1][0] = 0.125;
            parameters[1][1] = 0.25;
            parameters[1][2] = 0.375;
            parameters[1][3] = 0.5;
            parameters[1][4] = 0.625;
            parameters[1][5] = 0.75;
            parameters[1][6] = 0.875;
            System.out.println("Manual splitting done.");
            return;
        }
*/

        double intersectionParameters[][] = new double[2][2];
        if (a.tracesOver(b, intersectionParameters)) {

            // ...also have parameters seperate so that you don't mess up with
            // the old incoming data. Remember, the incoming data may be valid too

            if (intersectionParameters[0][0] != Double.POSITIVE_INFINITY)
                parameters[0][index[0]] = intersectionParameters[0][0]; 
            if (intersectionParameters[1][0] != Double.POSITIVE_INFINITY)
                parameters[1][index[0]] = intersectionParameters[1][0]; 

            index[0]++;

            if ((intersectionParameters[0][1] != Double.POSITIVE_INFINITY) || 
                (intersectionParameters[1][1] != Double.POSITIVE_INFINITY)) {
                if (intersectionParameters[0][1] != Double.POSITIVE_INFINITY)
                    parameters[0][index[0]] = intersectionParameters[0][1]; 
                if (intersectionParameters[1][1] != Double.POSITIVE_INFINITY)
                    parameters[1][index[0]] = intersectionParameters[1][1]; 
                index[0]++;
            }

/*
            index[0] = 1;
            parameters is already set
            System.out.println("A traces over B");
            System.out.print("Parameters = ");
            for (int k=0; k<2; k++)
               for (int l=0; l<2; l++)
                   System.out.println("["+k+"]["+l+"] = "+parameters[k][l]);
*/

            return;
        }
        else {
            //  System.out.println("A DOES NOT trace over B");
        }


        if (deptha > 0) {
            TBezier[] A = a.Split();
            double tmid = (t0+t1)*0.5;
            deptha--;
            if (depthb > 0) {
                TBezier[] B = b.Split();
                double umid = (u0+u1)*0.5;
                depthb--;
                if (IntersectBB(A[0], B[0]) == true)
                    RecursivelyIntersect(A[0], t0, tmid, deptha, B[0], u0, umid, depthb, parameters, index);
                if (IntersectBB(A[1], B[0]) == true)
                    RecursivelyIntersect(A[1], tmid, t1, deptha, B[0], u0, umid, depthb, parameters, index);
                if (IntersectBB(A[0], B[1]) == true)
                    RecursivelyIntersect(A[0], t0, tmid, deptha, B[1], umid, u1, depthb, parameters, index);
                if (IntersectBB(A[1], B[1]) == true)
                    RecursivelyIntersect(A[1], tmid, t1, deptha, B[1], umid, u1, depthb, parameters, index);
            }
            else {
                if (IntersectBB(A[0], b) == true)
                    RecursivelyIntersect(A[0], t0, tmid, deptha, b, u0, u1, depthb, parameters, index);
                if (IntersectBB(A[1], b) == true)
                    RecursivelyIntersect(A[1], tmid, t1, deptha, b, u0, u1, depthb, parameters, index);
            }
        }
        else if (depthb > 0) {
            TBezier[] B = b.Split();
            double umid = (u0 + u1)*0.5;
            depthb--;
            if (IntersectBB(a, B[0]) == true)
                RecursivelyIntersect(a, t0, t1, deptha, B[0], u0, umid, depthb, parameters, index);
            if (IntersectBB(a, B[1]) == true)
                RecursivelyIntersect(a, t0, t1, deptha, B[0], umid, u1, depthb, parameters, index);
        }
        else { // Both segments are fully subdivided; now do line segments
            double xlk = a.fPoint3_fX - a.fPoint0_fX;
            double ylk = a.fPoint3_fY - a.fPoint0_fY;
            double xnm = b.fPoint3_fX - b.fPoint0_fX;
            double ynm = b.fPoint3_fY - b.fPoint0_fY;
            double xmk = b.fPoint0_fX - a.fPoint0_fX;
            double ymk = b.fPoint0_fY - a.fPoint0_fY;
            double det = xnm * ylk - ynm * xlk;

            if (index[0] > 8) {
                // System.out.println("index[0] reached 9 ... breaking out of recursion");
                return;
            }

            if (1.0 + det == 1.0)
                return;
            else {
                double detinv = 1.0 / det;
                double s = (xnm * ymk - ynm *xmk) * detinv;
                double t = (xlk * ymk - ylk * xmk) * detinv;
                if ((s < 0.0) || (s > 1.0) || (t < 0.0) || (t > 1.0))
                    return;
                parameters[0][index[0]] = t0 + s * (t1 - t0);
                parameters[1][index[0]] = u0 + t * (u1 - u0);
                index[0]++;
            }
        }
    }

    private static double log4(double x) {
        return 0.5 * Math.log(x) / Math.log(2);
    }

    public static void Sort(double[] array) {
        QSort.sort(array, array.length);
/*
   System.out.print("qS " + array.length + ">:");
   for (int ii = 0; ii<array.length; ++ii) System.out.print(array[ii] + " ");
   System.out.println();
*/
    }


    private static double[] SortAndEliminateDuplicates(double[] array, int arrayLen) {

        if ((array == null) || (arrayLen < 1)) return null;
        if (arrayLen == 1) return array;

        // NOTE : The arrayLen passed in may be shorter that the actual elements in the array.
        //        There may be some garbage values at the end of the array.

        QSort.sort(array, arrayLen);

        // Eliminate all duplicate values inside the array
        int l = 0;  // mark the last unique element
        for (int i = 1; i < arrayLen; ++i) {
            if ((array[l] != array[i]) && (++l != i)) {
                array[l] = array[i];
            } // else just bump up j
        }
        ++l;  // make it the final length of the new array
 /*
    System.out.print("S&E " + array.length + ">:");
    for (int ii = 0; ii<l; ++ii) System.out.print(array[ii] + " ");
    System.out.println();
 */
        if (l == arrayLen)
            return array;
        else {
            // resize the array!
            double[] newArray = new double[l];
            System.arraycopy(array, 0, newArray, 0, l);
            return newArray;
        }
    }

    private static final double M_SQRT2 = Math.sqrt(2.0);
    private static final double INITIAL_VALUE = 173778.56254440593; // 58389596.999516;
    private static final double INV_EPS = 65536.;
    // actually INITIAL_VALUE = M_SQRT2 * 6.0 / 8.0 * INV_EPS * l0;  but never got it to work !!

    final int subDivisionDepth() {
        double la1_fX = Math.abs((fPoint2_fX - fPoint1_fX) - (fPoint1_fX - fPoint0_fX));
        double la2_fX = Math.abs((fPoint3_fX - fPoint2_fX) - (fPoint2_fX - fPoint1_fX));
        double la1_fY = Math.abs((fPoint2_fY - fPoint1_fY) - (fPoint1_fY - fPoint0_fY));
        double la2_fY = Math.abs((fPoint3_fY - fPoint2_fY) - (fPoint2_fY - fPoint1_fY));

        double l0 = Math.max(Math.max(la1_fX, la2_fX), Math.max(la1_fY, la2_fY));

        if (l0 * 0.75 * M_SQRT2 + 1.0 == 1.0)
            return 0;
        else
            return (int)Math.ceil(TBezier.log4(INITIAL_VALUE));
    }

    /*
     * Wang's theorem is used to estimate the level of subdivision required,
     * but only if the bounding boxes interfere at the top level.
     * Assuming there is a possible intersection, RecursivelyIntersect is
     * used to find all the parameters corresponding to intersection points.
     * these are then sorted and returned in an array.
     */


    public static double[][] findIntersections(TBezier a, TBezier b) {

        // Two Bezier curves can intersect at at-most 9 places. We prepare ourselves
        // for the worst scenario and allocate enough memory to start with.
        double[][] parameters = new double[2][9];
        int[] index = new int[1];    // Just a wrapper for an integer return parameter
        index[0] = 0;

        // Determine sub-divisions by checking for bounding-box intersections
        if (TBezier.IntersectBB(a, b) == true) {
            int ra = a.subDivisionDepth();
            int rb = b.subDivisionDepth();

            TBezier.RecursivelyIntersect(a, 0., 1., ra, b, 0., 1., rb, parameters, index);
        }

        // The value of index[0] now tells us how many places the curves intersect.
        // We will resize the double arrays to eliminate redundant values.

        double[][] justRightParameters = new double[2][index[0]];

        // Sort the parameter values before you return
        justRightParameters[0] = SortAndEliminateDuplicates(parameters[0], index[0]);
        justRightParameters[1] = SortAndEliminateDuplicates(parameters[1], index[0]);

        return justRightParameters;
    }


    public static double[] findOverlaps(TBezier a, TBezier b) {

        // Two Bezier curves may not intersect, but may "overlap" each other if the
        // straight line segment between their end points intersect. This is strictly
        // our own definition, and we need it because we want to eliminate all cases
        // where intersection is detected between bezier segments considering them to
        // be straight patches with a one-to-one correspondence to straight lines.
        //
        TGPoint A = a.getPoint(0);
        TGPoint B = a.getPoint(3);
        TGPoint C = b.getPoint(0);
        TGPoint D = b.getPoint(3);

        //A.x = 400.0; A.y = 260.0; B.x = 325.0; B.y = 310.0;
        //C.x = 293.0; C.y = 223; D.x = 343; D.y = 298.048;

        // If both the ends match either way, then the curves are considered to intersect.
        // Since we have to split this curve anyway, we arbitrarily assume they need to be
        // split at the mid-point where t=0.5.
        if ((isClose(A, C) && isClose(B, D)) || (isClose(A, D) && isClose(B, C))) {
            boolean aLinear = a.isLinear();
            boolean bLinear = b.isLinear();
            if ((!aLinear) && (!bLinear)) { // if none are linear
                double[] parameters = new double[2];
                parameters[0] = 0.5;
                parameters[1] = 0.5;
                return parameters;
            }
            else if ((!aLinear) && (bLinear)) {
                double[] parameters = new double[2];
                parameters[0] = 0.5;
                parameters[1] = 0.0;
                return parameters;
            }
            else if ((aLinear) && (!bLinear)) {
                double[] parameters = new double[2];
                parameters[0] = 0.0;
                parameters[1] = 0.5;
                return parameters;
            }
            else // Both are linear
                return null;

        }

        if (A.equals(C) || A.equals(D) || B.equals(C) || B.equals(D))
            return null;

        double denom = (B.x - A.x)*(D.y - C.y) - (B.y - A.y)*(D.x - C.x);
        if (denom == 0.0)
            return null;    // Better than getting an exception
        double num1 =  (C.x - A.x)*(D.y - C.y) - (C.y - A.y)*(D.x - C.x);
        double t1 = num1 / denom;
        double num2 =  (B.x - A.x)*(A.y - C.y) - (B.y - A.y)*(A.x - C.x);
        double t2 = num2 / denom;

        if (((t1 > 0.0) && (t1 < 1.0)) && ((t2 > 0.0) && (t2 < 1.0))) {
            double[] parameters = new double[2];
            parameters[0] = t1;
            parameters[1] = t2;
            return parameters;
        }
        else
            return null;
    }

    public TGPoint evaluate(double t) {
        double xValue =       (1.0-t) * (1.0-t) * (1.0-t) * fPoint0_fX +
                        3.0 *      t  * (1.0-t) * (1.0-t) * fPoint1_fX +
                        3.0 *      t  *      t  * (1.0-t) * fPoint2_fX +
                                   t  *      t  *      t  * fPoint3_fX;
        double yValue =       (1.0-t) * (1.0-t) * (1.0-t) * fPoint0_fY +
                        3.0 *      t  * (1.0-t) * (1.0-t) * fPoint1_fY +
                        3.0 *      t  *      t  * (1.0-t) * fPoint2_fY +
                                   t  *      t  *      t  * fPoint3_fY;
        return new TGPoint(xValue, yValue);
    }

    // Returns an array of parameter values at which the tangent of the Bezier
    // curve changes direction.
    double[] findZeroDerivatives () {
        if (isLinear() == true)
            return null;

        double c2 = -fPoint0_fY + 3.0*fPoint1_fY - 3.0*fPoint2_fY + fPoint3_fY;
        double c1 = 2.0*fPoint0_fY - 4.0*fPoint1_fY + 2.0*fPoint2_fY;
        double c0 = -fPoint0_fY + fPoint1_fY;

        return SolveQuadratic(c0, c1, c2);
    }

    // Returns an array of parameter values at which the tangent of the Bezier
    // curve becomes infinity
    double[] findInfiniteDerivatives() {
        if (isLinear() == true)
            return null;

        double c2 = -fPoint0_fX + 3.0*fPoint1_fX - 3.0*fPoint2_fX + fPoint3_fX;
        double c1 = 2.0*fPoint0_fX - 4.0*fPoint1_fX + 2.0*fPoint2_fX;
        double c0 = -fPoint0_fX + fPoint1_fX;

        return SolveQuadratic(c0, c1, c2);
    }

    double[] SolveQuadratic(double c0, double c1, double c2) {
        if (Math.abs(c2) < 1.0e-9) {
            double retVal = -c0 / c1;
            if ((retVal >= 0.0) && (retVal <= 1.0)) {
                double[] parameters = new double[1];
                parameters[0] = retVal;
                return parameters;
            }
        }

        double p, q, D;
        p = c1 / (2.0*c2);
        q = c0 / c2;
        D = p * p - q;

        if (Math.abs(D) < 1.0e-9) {
            double retVal = -p;
            if ((retVal >= 0.0) && (retVal <= 1.0)) {
                double[] parameters = new double[1];
                parameters[0] = retVal;
                return parameters;
            }
        }
        else if (D < 0.0) {
            return null;
        }
        else if (D > 0.0) {
            double sqrt_D = Math.sqrt(D);
            double retVal0 = sqrt_D  - p;
            double retVal1 = - sqrt_D - p;
            if ((retVal0 >= 0.0) && (retVal0 <= 1.0)) {
                // We know that retVal0 must be passed
                if ((retVal1 >= 0.0) && (retVal1 <= 1.0)) {
                    // We know that both values must be passed
                    double[] parameters = new double[2];
                    parameters[0] = retVal0;
                    parameters[1] = retVal1;
                    Sort(parameters);
                    return parameters;
                }
                else { // We know that only retVal0 is valid
                    double[] parameters = new double[1];
                    parameters[0] = retVal0;
                    return parameters;
                }
            }
            else { // Check to see if retVal1 is valid or not
                if ((retVal1 >= 0.0) && (retVal1 <= 1.0)) {
                    // Pass only retVal1
                    double[] parameters = new double[1];
                    parameters[0] = retVal1;
                    return parameters;
                }
                else { // None of the values can be passed
                    return null;
                }
            }
            // double v1 = c2*retVal0*retVal0 + c1*retVal0 + c0;
            // double v2 = c2*retVal1*retVal1 + c1*retVal1 + c0;
            // System.out.println("Values at roots are " + v1);
            // System.out.println("Values at roots are " + v2);
        }
        return null; // required by Java compiler but never executed
    }


    // Two curves are similar or close to each other if all points are
    // within curveClosenessEpsilon of each other.

    static final double curveClosenessEpsilon = 5.0E-06;
    public boolean isClose (TBezier other) {
        return ( (Eq(fPoint0_fX, other.fPoint0_fX, curveClosenessEpsilon) &&
                  Eq(fPoint0_fY, other.fPoint0_fY, curveClosenessEpsilon) &&
                  Eq(fPoint1_fX, other.fPoint1_fX, curveClosenessEpsilon) &&
                  Eq(fPoint1_fY, other.fPoint1_fY, curveClosenessEpsilon) &&
                  Eq(fPoint2_fX, other.fPoint2_fX, curveClosenessEpsilon) &&
                  Eq(fPoint2_fY, other.fPoint2_fY, curveClosenessEpsilon) &&
                  Eq(fPoint3_fX, other.fPoint3_fX, curveClosenessEpsilon) &&
                  Eq(fPoint3_fY, other.fPoint3_fY, curveClosenessEpsilon)) ||
                 (Eq(fPoint0_fX, other.fPoint3_fX, curveClosenessEpsilon) &&
                  Eq(fPoint0_fY, other.fPoint3_fY, curveClosenessEpsilon) &&
                  Eq(fPoint1_fX, other.fPoint2_fX, curveClosenessEpsilon) &&
                  Eq(fPoint1_fY, other.fPoint2_fY, curveClosenessEpsilon) &&
                  Eq(fPoint2_fX, other.fPoint1_fX, curveClosenessEpsilon) &&
                  Eq(fPoint2_fY, other.fPoint1_fY, curveClosenessEpsilon) &&
                  Eq(fPoint3_fX, other.fPoint0_fX, curveClosenessEpsilon) &&
                  Eq(fPoint3_fY, other.fPoint0_fY, curveClosenessEpsilon))
                );
    }


    public boolean equals(TBezier src) {
        return ((fPoint0_fX == src.fPoint0_fX) &&
                (fPoint0_fY == src.fPoint0_fY) &&
                (fPoint1_fX == src.fPoint1_fX) &&
                (fPoint1_fY == src.fPoint1_fY) &&
                (fPoint2_fX == src.fPoint2_fX) &&
                (fPoint2_fY == src.fPoint2_fY) &&
                (fPoint3_fX == src.fPoint3_fX) &&
                (fPoint3_fY == src.fPoint3_fY));
    }

    public boolean isLinear() {
        // Tell whether the curve is linear by checking if the
        // first two points and last two points are the same.
        if ((fPoint0_fX == fPoint1_fX) && (fPoint0_fY == fPoint1_fY) &&
            (fPoint2_fX == fPoint3_fX) && (fPoint2_fY == fPoint3_fY))
            return true;
        // We now know that the end points do not overlap. Now the curve
        // will be linear only if the slope of the three segments between
        // the four points are same.
        // Slope = delta Y / delta X
        double slope1;
        double deltaY1 = fPoint1_fY - fPoint0_fY;
        if (deltaY1 == 0.0) slope1 = Double.POSITIVE_INFINITY;
        else slope1 = (fPoint1_fX - fPoint0_fX) / deltaY1;

        double slope2;
        double deltaY2 = fPoint2_fY - fPoint1_fY;
        if (deltaY2 == 0.0) slope2 = Double.POSITIVE_INFINITY;
        else slope2 = (fPoint2_fX - fPoint1_fX) / deltaY2;

        if (slope1 == slope2) {
            double slope3;
            double deltaY3 = fPoint3_fY - fPoint2_fY;
            if (deltaY3 == 0.0) slope3 = Double.POSITIVE_INFINITY;
            else slope3 = (fPoint3_fX - fPoint2_fX) / deltaY3;

            if (slope2 == slope3)
                return true;
        }
        return false;
    }

    private final static double pointEpsilon = 0.005;

    static final boolean Eq(double a, double b) {
        return Eq(a, b, pointEpsilon);
    }
    static final boolean Eq(double a, double b, double epsilon) {
        return (Math.abs(a - b) < epsilon);
    }

    // Tells whether the curve is degenerate by checking if
    // all the points are same or not.
    public boolean isDegenerate() {
        return
            Eq(fPoint0_fX, fPoint1_fX) &&
            Eq(fPoint0_fY, fPoint1_fY) &&
            Eq(fPoint1_fX, fPoint2_fX) &&
            Eq(fPoint1_fY, fPoint2_fY) &&
            Eq(fPoint2_fX, fPoint3_fX) &&
            Eq(fPoint2_fY, fPoint3_fY);
    }

    TGPoint getPoint(int index) {
        if ((index < 0) || (index > 3))
            throw new IndexOutOfBoundsException("TBezier.getPoint(): index");

        switch (index) {
            case 0 : return new TGPoint(fPoint0_fX, fPoint0_fY);
            case 1 : return new TGPoint(fPoint1_fX, fPoint1_fY);
            case 2 : return new TGPoint(fPoint2_fX, fPoint2_fY);
            case 3 : return new TGPoint(fPoint3_fX, fPoint3_fY);
        }
        return null;
    }

    void setPoint(int index, TGPoint point) {
        if ((index < 0) || (index > 3))
            throw new IndexOutOfBoundsException("TBezier.setPoint(): index");

        switch (index) {
            case 0 : fPoint0_fX = point.x; fPoint0_fY = point.y; break;
            case 1 : fPoint1_fX = point.x; fPoint1_fY = point.y; break;
            case 2 : fPoint2_fX = point.x; fPoint2_fY = point.y; break;
            case 3 : fPoint3_fX = point.x; fPoint3_fY = point.y; break;
        }
    }

    public String toString() {
        return (
            "TBezier : (" + fPoint0_fX + "," + fPoint0_fY + ")" +
            " (" + fPoint1_fX + "," + fPoint1_fY + ")" +
            " (" + fPoint2_fX + "," + fPoint2_fY + ")" +
            " (" + fPoint3_fX + "," + fPoint3_fY + ")\n");
    }

    void reverseDirection() {
        double tmp = fPoint0_fX; fPoint0_fX = fPoint3_fX; fPoint3_fX = tmp;
        tmp = fPoint0_fY; fPoint0_fY = fPoint3_fY; fPoint3_fY = tmp;
        tmp = fPoint1_fX; fPoint1_fX = fPoint2_fX; fPoint2_fX = tmp;
        tmp = fPoint1_fY; fPoint1_fY = fPoint2_fY; fPoint2_fY = tmp;
    }

    static final public boolean isClose(TGPoint p1, TGPoint p2) {
        return Eq(p1.x, p2.x) && Eq(p1.y, p2.y);
    }
    static final public boolean isClose(TGPoint p1, TGPoint p2, double eps) {
        return Eq(p1.x, p2.x, eps) && Eq(p1.y, p2.y, eps);
    }



    // new functions Anupam made-up for the contains function TGLoop:


    // This function tells if THIS curve traces over another Bezier curve.
    // A curve 'traces over' another if it either overlaps or portions of it
    // overlap with another curve. This is treated as a special case since it
    // is notorious to send all algorithms into infinite loops. Theoretically
    // there are infinite intersections between these curves and we possibly
    // cannot handle all this data elegantly without going nuts. This function
    // approaches the problem more heuristically. It checks for certain points
    // on this curve to lie on the other curve and vice versa.The results
    // contain the parameters where intersection was detected. If the return
    // value is false, the return parameters at least contain the nearest
    // parametric values where the curves may be broken up.
    // Heuristic is the word!  But look below, it has more maths that ever before.
    // The fundamental function that invokes all the maths is findNearestParametric().
    // 
    // On return, intersectionParameters[0] contains two values on THIS curve
    // where intersetion occured
    // intersectionParameters[1] contains two values on other curve
    // where intersetion occured
    private boolean tracesOver(TBezier otherCurve, double[][] intersectionParameters) {

// System.out.println("Inside tracesOver1");
// System.out.println("Detecting traces...");
// System.out.println("THIS curve is : " + this);
// System.out.println("OTHER curve is : " + otherCurve);

        // Fill in default values for the array. Needed later for checking intersections.
        intersectionParameters[0][0] = intersectionParameters[0][1] =
            intersectionParameters[1][0] = intersectionParameters[1][1] = Double.POSITIVE_INFINITY;

        // Check if this curve is a complete subset of the other curve
        if (this.isSegmentOf(otherCurve, intersectionParameters[1])) {
//System.out.println("Inside tracesOver2");
            intersectionParameters[0][0] = 0.0;  // the beginning of this curve
            intersectionParameters[0][1] = 1.0;  // the end of this curve
            return true;
        }

        // Check if the other curve is a complete subset of this curve
        else if (otherCurve.isSegmentOf(this, intersectionParameters[0])) {
//System.out.println("Inside tracesOver3");
            intersectionParameters[1][0] = 0.0;  // the beginning of other curve
            intersectionParameters[1][1] = 1.0;  // the end of other curve
            return true;
        }
        else {
            TGPoint thisGetPoint0 = this.getPoint(0);
            TGPoint thisGetPoint3 = this.getPoint(3);
            TGPoint otherGetPoint0 = otherCurve.getPoint(0);
            TGPoint otherGetPoint3 = otherCurve.getPoint(3);
            // Check if the curves overlap partially (with non-common edges dangling at the side)
            if (otherCurve.passesThrough(thisGetPoint0, intersectionParameters[1], 0)) {
//System.out.println("Inside tracesOver4");
                // first point of this curve passes through the other curve
                // there are two new possibilities now
                // the first point of the other curve or the last point passes through this curve
                if (this.passesThrough(otherGetPoint0, intersectionParameters[0], 0)) {
//System.out.println("Inside tracesOver5");
                    if (otherGetPoint0.equals(thisGetPoint0)) {
                        // there is only one point of contact, return false
                        return false;
                    }
                    // one end point of each curve passes through the other
                    // now check if two other points in between also pass
                    return true;    // TO DO check for in-between points
                }
                else if (this.passesThrough(otherCurve.getPoint(3), intersectionParameters[0], 0)) {
//System.out.println("Inside tracesOver6");
                    if (otherGetPoint3.equals(thisGetPoint0)) {
                        // there is only one point of contact, return false
                        return false;
                    }
                    // one end point of each curve passes through the other
                    // now check if two other points in between also pass
                    return true;    // TO DO check for in-between points
                }
                else
                    return false;   // the other end point does not pass
            }
            else if (otherCurve.passesThrough(thisGetPoint3, intersectionParameters[1], 0)) {
//System.out.println("Inside tracesOver7");
                // last point of this curve passes through the other curve
                // there are two new possibilities now
                // the first point of the other curve or the last point passes through this curve
                if (this.passesThrough(otherGetPoint0, intersectionParameters[0], 0)) {
//System.out.println("Inside tracesOver8");
                    if (otherGetPoint0.equals(thisGetPoint3)) {
                        // there is only one point of contact, return false
                        return false;
                    }
                    // one end point of each curve passes through the other
                    // now check if two other points in between also pass
                    return true;    // TO DO check for in-between points
                }
                else if (this.passesThrough(otherGetPoint3, intersectionParameters[0], 0)) {
//System.out.println("Inside tracesOver9");
                    if (otherGetPoint3.equals(thisGetPoint3)) {
                        // there is only one point of contact, return false
                        return false;
                    }
                    // one end point of each curve passes through the other
                    // now check if two other points in between also pass
                    return true;    // TO DO check for in-between points
                }
                else
                    return false;   // the other end point does not pass
            }
            else {
//System.out.println("Inside tracesOver10");
                return false;   // none of the end points pass through the other curve
            }
        }
    }



    // A function that tells if a curve is a section of this curve.
    // Returns true if this curve lies entirely within the test curve
    // It is done by taking five points from this curve at parameters
    // 0.0, 0.25, 0.5, 0.75, and 1.0 and comparing it with points
    // on the other curve.
    // On return, passParameters[0] contains the starting parameter value of the other curve 
    //            passParameters[1] contains the ending parameter value of the other curve 
    // where this curve has intersected the other
    // make sure you have passParameters[] = new double[2] called before you enter

    private boolean isSegmentOf(TBezier other, double[] passParameters) {
// System.out.println("Inside isSegmentOf1");
        boolean b1 = other.passesThrough(this.evaluate(0.25), passParameters, 0);
// System.out.println("b1 = " + b1);
        boolean b2 = other.passesThrough(this.evaluate(0.50), passParameters, 0);
// System.out.println("b2 = " + b2);
        boolean b3 = other.passesThrough(this.evaluate(0.75), passParameters, 0);
// System.out.println("b3 = " + b3);
        boolean b4 = other.passesThrough(this.getPoint(0), passParameters, 0);
// System.out.println("b4 = " + b4);
        boolean b5 = other.passesThrough(this.getPoint(3), passParameters, 1);
// System.out.println("b5 = " + b5);

        // Don't merge the booleans together, even though they are ANDs.
        // We want to return back the nearest parametric values where the
        // curve ends may possibly intersect, irrespective of whether they
        // overlap or not.
        return (b1 && b2 && b3 && b4 && b5);
    }



    // A utility function that tells if this Bezier passes through the given point
    // On return, atParameter[index] has the parametric value of the curve closest to the point
    // If the return value is true, this parametric value is obviously the point of intersection
    private boolean passesThrough(TGPoint p, double[] atParameter, int index) {
        
        TGPoint pntOnCurve = new TGPoint();
// System.out.println("Inside passesThrough");
// System.out.println("Testing for " + this + " with " + p);
        atParameter[index] = findNearestParametric(p, pntOnCurve);
// System.out.println("atParameter[index] = " + atParameter[index]);
        TGPoint pointOnCurve = evaluate(atParameter[index]); // don't need this anymore
// System.out.println("pointOnCurve = " + pointOnCurve);
// System.out.println("pntOnCurve = " + pntOnCurve);
// System.out.println("distance between the two points are " +
//                    Math.sqrt(((p.x-pointOnCurve.x) * (p.x-pointOnCurve.x)) +
//                    ((p.y-pointOnCurve.y) * (p.y-pointOnCurve.y))));

        boolean res = isClose(p, pntOnCurve);
// System.out.println("passesThrough is " + res);
        return res;
    }


    private double lengthEstimateSquare() {
         double result =
             ((fPoint0_fX - fPoint1_fX) * (fPoint0_fX - fPoint1_fX)) + 
             ((fPoint0_fY - fPoint1_fY) * (fPoint0_fY - fPoint1_fY)) + 

             ((fPoint1_fX - fPoint2_fX) * (fPoint1_fX - fPoint2_fX)) + 
             ((fPoint1_fY - fPoint2_fY) * (fPoint1_fY - fPoint2_fY)) + 

             ((fPoint2_fX - fPoint3_fX) * (fPoint2_fX - fPoint3_fX)) + 
             ((fPoint2_fY - fPoint3_fY) * (fPoint2_fY - fPoint3_fY));
         return (result);
    }


    // A Bezier curve has inflexion if its derivative changes direction
    // Inflexion on Bezier curves can be detected by checking for intersection
    // between edges of the polygon formed by its control points.

    private static boolean linesIntersect(TGPoint p1, TGPoint p2, TGPoint p3, TGPoint p4) {
        // This private function tells whether the line p1-p2 intersects line p3-p4.
        if (p1.equals(p3) || p1.equals(p4) || p2.equals(p3) || p2.equals(p4))
            return false;

        double denom = (p2.x - p1.x)*(p4.y - p3.y) - (p2.y - p1.y)*(p4.x - p3.x);
        if (denom == 0.0)
            return false;    // Better than getting an exception

        double t1 = ((p3.x - p1.x)*(p4.y - p3.y) - (p3.y - p1.y)*(p4.x - p3.x)) / denom;
        double t2 = ((p2.x - p1.x)*(p1.y - p3.y) - (p2.y - p1.y)*(p1.x - p3.x)) / denom;

        return ((t1 > 0.0) && (t1 < 1.0)) && ((t2 > 0.0) && (t2 < 1.0));
    }


    public final boolean hasInflexion() {
        TGPoint A = getPoint(0);
        TGPoint B = getPoint(1);
        TGPoint C = getPoint(2);
        TGPoint D = getPoint(3);
        if (linesIntersect(A, B, C, D) || linesIntersect(B, C, A, D))
            return true;
        return false;
    }


    public final boolean containsWithinConvexHull(TGPoint point) {
        TGPolygon convexHull = new TGPolygon(4);
        convexHull.setPoint(0, getPoint(0));
        convexHull.setPoint(1, getPoint(1));
        convexHull.setPoint(2, getPoint(2));
        convexHull.setPoint(3, getPoint(3));
        return convexHull.contains(point);
    }


    public final TGPoint normal(double t) {
        // Calculate exactly as tangent
        double xValue = - 3.0 *            (1.0-t) *           (1.0-t)  * fPoint0_fX +
                          3.0 * ((1.0-t) * (1.0-t) - 2.0 * t * (1.0-t)) * fPoint1_fX +
                          3.0 * (-    t  *      t  + 2.0 * t * (1.0-t)) * fPoint2_fX +
                          3.0 *       t  *                 t            * fPoint3_fX;
        double yValue = - 3.0 *            (1.0-t) *           (1.0-t)  * fPoint0_fY +
                          3.0 * ((1.0-t) * (1.0-t) - 2.0 * t * (1.0-t)) * fPoint1_fY +
                          3.0 * (-    t  *      t  + 2.0 * t * (1.0-t)) * fPoint2_fY +
                          3.0 *       t  *                 t            * fPoint3_fY;

        // The normal is (-dx/dy) as opposed to the tangent which is (dy/dx)
        return new TGPoint(-yValue, xValue);
    }

    /*
     *  findNearestParametric :
     *      Compute the parameter value of the point on a Bezier
     *      curve segment closest to some arbtitrary, user-input point.
     *      Returns this parametric value.
     *      Also returns the point on the curve at that parameter value.
     *
     */

    private static final int W_DEGREE = 5;  //  Degree of equation to find roots of
    private static final int DEGREE = 3;    //  Cubic Bezier curve
    private static final int MAXDEPTH = 64; //  Maximum depth for recursion

    private static int SGN(double a) { return ((a<0.0) ? -1 : 1); }

    public double findNearestParametric(TGPoint P, TGPoint nearestPoint)
    //    TGPoint   P;          // The user-supplied point
    //    TGPoint   *V;         // Control points of cubic Bezier
    {
        TGPoint[] w;            // Ctl pts for 5th-degree eqn
        double[] t_candidate = new double[W_DEGREE];    // Possible roots
        int n_solutions;        // Number of roots found
        double t;           // Parameter value of closest pt
        TGPoint[] V = new TGPoint[DEGREE+1];
        V[0] = getPoint(0);
        V[1] = getPoint(1);
        V[2] = getPoint(2);
        V[3] = getPoint(3);

        //  Convert problem to 5th-degree Bezier form
        w = convertToBezierForm(P, V);

        // Find all possible roots of 5th-degree equation
        n_solutions = findRoots(w, W_DEGREE, t_candidate, 0);
        // free((char *)w);

        // Compare distances of P to all candidates, and to t=0, and t=1
        {
            double  dist, new_dist;
            TGPoint p;
            TGPoint v;
            int     i;
        
            // Check distance to beginning of curve, where t = 0
            TGPoint diff = TGPoint.subtract(P, V[0]);
            dist = diff.vectorSquaredLength();
            // V2SquaredLength(V2Sub(&P, &V[0], &v));
            t = 0.0;

            // Find distances for candidate points
            for (i = 0; i < n_solutions; i++) {
                p = evaluateWithFill(V, DEGREE, t_candidate[i], null, null);
                new_dist = (TGPoint.subtract(P, p)).vectorSquaredLength();
                if (new_dist < dist) {
                        dist = new_dist;
                        t = t_candidate[i];
                }
            }

        // Finally, look at distance to end point, where t = 1.0
            // new_dist = V2SquaredLength(V2Sub(&P, &V[DEGREE], &v));
            new_dist = (TGPoint.subtract(P, V[DEGREE])).vectorSquaredLength();
                if (new_dist < dist) {
                    dist = new_dist;
                t = 1.0;
            }
        }

        //  Return the point on the curve at parameter value t
        // System.out.println("t = " + t);
        nearestPoint.copyFrom(evaluateWithFill(V, DEGREE, t, null, null));
        return t;
    }



    /*
     *  convertToBezierForm :
     *      Given a point and a Bezier curve, generate a 5th-degree
     *      Bezier-format equation whose solution finds the point on the
     *      curve nearest the user-defined point.
     */

    private TGPoint[] convertToBezierForm(TGPoint P, TGPoint[] V)
    //    TGPoint   P;          // The point to find t for
    //    TGPoint   *V;         // The control points
    {
        int     i, j, k, m, n, ub, lb;  
        int     row, column;        // Table indices
        TGPoint c[] = new TGPoint[DEGREE+1];        // V(i)'s - P
        TGPoint d[] = new TGPoint[DEGREE];      // V(i+1) - V(i)

        double cdTable[][] = new double[3][4];      // Dot product of c, d
        double z[][] = {    // Precomputed "z" for cubics
        {1.0, 0.6, 0.3, 0.1},
        {0.4, 0.6, 0.6, 0.4},
        {0.1, 0.3, 0.6, 1.0},
        };


//System.out.println("P1 = " + P);
        // Determine the c's -- these are vectors created by subtracting
        // point P from each of the control points
        for (i = 0; i <= DEGREE; i++) {
            // V2Sub(&V[i], &P, &c[i]);
            c[i] = TGPoint.subtract(V[i], P);
        }
        // Determine the d's -- these are vectors created by subtracting
//System.out.println("P2 = " + P);
        // each control point from the next
        for (i = 0; i <= DEGREE - 1; i++) { 
            d[i] = TGPoint.subtract(V[i+1], V[i]);
            d[i].multiplyBy(3.0);
            // d[i] = V2ScaleII(V2Sub(&V[i+1], &V[i], &d[i]), 3.0);
        }

//System.out.println("P3 = " + P);
        // Create the c,d table -- this is a table of dot products of the */
        // c's and d's
        for (row = 0; row <= DEGREE - 1; row++) {
            for (column = 0; column <= DEGREE; column++) {
                // cdTable[row][column] = V2Dot(&d[row], &c[column]);
                cdTable[row][column] = d[row].dotProduct(c[column]);
            }
        }

//System.out.println("P4 = " + P);
        // Now, apply the z's to the dot products, on the skew diagonal
        // Also, set up the x-values, making these "points"
        // TGPoint[] w;         // Ctl pts of 5th-degree curve
        TGPoint[] w = new TGPoint[W_DEGREE+1];
        for (i = 0; i <= W_DEGREE; i++) {
//System.out.println("i = " + i);
            w[i] = new TGPoint((double)(i) / W_DEGREE, 0.0);
            // w[i].y = 0.0;
            // w[i].x = (double)(i) / W_DEGREE;
        }

//System.out.println("P5 = " + P);
        n = DEGREE;
        m = DEGREE-1;
        for (k = 0; k <= n + m; k++) {
            lb = Math.max(0, k - m);
            ub = Math.min(k, n);
            for (i = lb; i <= ub; i++) {
                j = k - i;
                w[i+j].y += cdTable[j][i] * z[j][i];
            }
        }

        return (w);
    }



    /*
     *  findRoots :
     *  Given a 5th-degree equation in Bernstein-Bezier form, find
     *  all of the roots in the interval [0, 1].  Return the number
     *  of roots found.
     */
    static int findRoots(TGPoint[] w, int degree, double[] t, int depth)
        // TGPoint  *w;     // The control points
        // int  degree;     // The degree of the polynomial
        // double   *t;     // RETURN candidate t-values
        // int  depth;      // The depth of the recursion
    {  
        int     i;
        TGPoint[] Left = new TGPoint[W_DEGREE+1];   // New left and right
        TGPoint[] Right = new TGPoint[W_DEGREE+1];  // control polygons
        int left_count, right_count;    // Solution count from children
        double[] left_t = new double[W_DEGREE+1];   // Solutions from kids
        double[] right_t = new double[W_DEGREE+1];

        switch (crossingCount(w, degree)) {
            case 0 : {  // No solutions here
                return 0;   
            }
            case 1 : {  // Unique solution
                // Stop recursion when the tree is deep enough
                // if deep enough, return 1 solution at midpoint
                if (depth >= MAXDEPTH) {
                    t[0] = (w[0].x + w[W_DEGREE].x) / 2.0;
                    return 1;
                }
                if (isControlPolygonFlatEnough(w, degree)) {
                    t[0] = computeXIntercept(w, degree);
                    return 1;
                }
                break;
            }
        }

        // Otherwise, solve recursively after subdividing control polygon
        evaluateWithFill(w, degree, 0.5, Left, Right);
        left_count  = findRoots(Left,  degree, left_t, depth+1);
        right_count = findRoots(Right, degree, right_t, depth+1);


        // Gather solutions together
        for (i = 0; i < left_count; i++) {
            t[i] = left_t[i];
        }
        for (i = 0; i < right_count; i++) {
            t[i+left_count] = right_t[i];
        }

        // Send back total number of solutions
        return (left_count+right_count);
    }


    /*
     * crossingCount :
     *  Count the number of times a Bezier control polygon 
     *  crosses the 0-axis. This number is >= the number of roots.
     *
     */
    private static int crossingCount(TGPoint[] V, int degree)
    //    TGPoint   *V;         //  Control pts of Bezier curve
    //    int       degree;     //  Degreee of Bezier curve
    {
        int     i;  
        int     n_crossings = 0;    //  Number of zero-crossings
        int     sign, old_sign;     //  Sign of coefficients

        sign = old_sign = SGN(V[0].y);
        for (i = 1; i <= degree; i++) {
            sign = SGN(V[i].y);
            if (sign != old_sign) n_crossings++;
            old_sign = sign;
        }
        return n_crossings;
    }


    /*
     *  isControlPolygonFlatEnough :
     *  Check if the control polygon of a Bezier curve is flat enough
     *  for recursive subdivision to bottom out.
     *
     */
    private static boolean isControlPolygonFlatEnough(TGPoint[] V, int degree)
    //    TGPoint   *V;     // Control points
    //    int   degree;     // Degree of polynomial
    {
        int     i;          // Index variable
        // double   *distance;      // Distances from pts to line
        double  max_distance_above; // maximum of these
        double  max_distance_below;
        double  error;          // Precision of root
        double  intercept_1,
                intercept_2,
                left_intercept,
                right_intercept;
        double  a, b, c;        // Coefficients of implicit
                            // eqn for line from V[0]-V[deg]

        // Find the  perpendicular distance
        // from each interior control point to
        // line connecting V[0] and V[degree]
        // distance = (double *)malloc((unsigned)(degree + 1) * sizeof(double));
        double[] distance = new double[degree + 1];
        {
        double  abSquared;

        // Derive the implicit equation for line connecting first
        // and last control points */
        a = V[0].y - V[degree].y;
        b = V[degree].x - V[0].x;
        c = V[0].x * V[degree].y - V[degree].x * V[0].y;

        abSquared = (a * a) + (b * b);

            for (i = 1; i < degree; i++) {
            // Compute distance from each of the points to that line
                distance[i] = a * V[i].x + b * V[i].y + c;
                if (distance[i] > 0.0) {
                    distance[i] = (distance[i] * distance[i]) / abSquared;
                }
                if (distance[i] < 0.0) {
                    distance[i] = -((distance[i] * distance[i]) / abSquared);
                }
            }
        }


        // Find the largest distance
        max_distance_above = 0.0;
        max_distance_below = 0.0;
        for (i = 1; i < degree; i++) {
            if (distance[i] < 0.0) {
                max_distance_below = Math.min(max_distance_below, distance[i]);
            };
            if (distance[i] > 0.0) {
                max_distance_above = Math.max(max_distance_above, distance[i]);
            }
        }
        // free((char *)distance);

        {
        double  det, dInv;
        double  a1, b1, c1, a2, b2, c2;

        //  Implicit equation for zero line
        a1 = 0.0;
        b1 = 1.0;
        c1 = 0.0;

        //  Implicit equation for "above" line
        a2 = a;
        b2 = b;
        c2 = c + max_distance_above;

        det = a1 * b2 - a2 * b1;
        dInv = 1.0/det;
        
        intercept_1 = (b1 * c2 - b2 * c1) * dInv;

        //  Implicit equation for "below" line
        a2 = a;
        b2 = b;
        c2 = c + max_distance_below;
        
        det = a1 * b2 - a2 * b1;
        dInv = 1.0/det;
        
        intercept_2 = (b1 * c2 - b2 * c1) * dInv;
        }

        // Compute intercepts of bounding box
        left_intercept = Math.min(intercept_1, intercept_2);
        right_intercept = Math.max(intercept_1, intercept_2);

        // double flatnessEpsilon = Math.exp(1.0,-MAXDEPTH-1); // Flatness control value
        double flatnessEpsilon = Math.exp(-MAXDEPTH-1); // Flatness control value
        error = 0.5 * (right_intercept-left_intercept);    
        if (error < flatnessEpsilon) {
            return true;
        }
        else {
            return false;
        }
    }


    /*
     *  computeXIntercept :
     *  Compute intersection of chord from first control point to last
     *      with 0-axis.
     * 
     */
    /* NOTE: "T" and "Y" do not have to be computed, and there are many useless
     * operations in the following (e.g. "0.0 - 0.0").
     */
    private static double computeXIntercept(TGPoint[] V, int degree)
    //    TGPoint   *V;         //  Control points
    //    int       degree;     //  Degree of curve
    {
        double  XLK, YLK, XNM, YNM, XMK, YMK;
        double  det, detInv;
        double  S, T;
        double  X, Y;

        XLK = 1.0; // actually 1.0 - 0.0;
        YLK = 0.0; // actually 0.0 - 0.0;
        XNM = V[degree].x - V[0].x;
        YNM = V[degree].y - V[0].y;
        XMK = V[0].x; // actually V[0].x - 0.0;
        YMK = V[0].y; // actually V[0].y - 0.0;

        det = XNM*YLK - YNM*XLK;
        detInv = 1.0/det;

        S = (XNM*YMK - YNM*XMK) * detInv;
        //  T = (XLK*YMK - YLK*XMK) * detInv;

        X = 0.0 + XLK * S;
        //  Y = 0.0 + YLK * S;

        return X;
    }



    /*
     *  Bezier : 
     *  Evaluate a Bezier curve at a particular parameter value
     *  Fill in control points for resulting sub-curves if "Left" and
     *  "Right" are non-null.
     * 
     */
    private static TGPoint evaluateWithFill(TGPoint[] V, int degree, double t, TGPoint[] Left, TGPoint[] Right)
    //    TGPoint   *V;         // Control pts
    //    int   degree;         // Degree of bezier curve
    //    double    t;          // Parameter value
    //    TGPoint   *Left;      // RETURN left half ctl pts
    //    TGPoint   *Right;     // RETURN right half ctl pts
    {
        int     i, j;       // Index variables
        TGPoint Vtemp[][] = new TGPoint[W_DEGREE+1][W_DEGREE+1];

// System.out.println("Reached 1");
        // Copy control points
        for (j =0; j <= degree; j++) {
// System.out.println("j = " + j);
            Vtemp[0][j] = V[j];
        }

// System.out.println("Reached 2");
        // Triangle computation
        for (i = 1; i <= degree; i++) { 
            for (j =0 ; j <= degree - i; j++) {
                Vtemp[i][j] = new TGPoint(
                    (1.0 - t) * Vtemp[i-1][j].x + t * Vtemp[i-1][j+1].x,
                    (1.0 - t) * Vtemp[i-1][j].y + t * Vtemp[i-1][j+1].y);
                // Vtemp[i][j].x =
                    // (1.0 - t) * Vtemp[i-1][j].x + t * Vtemp[i-1][j+1].x;
                // Vtemp[i][j].y =
                    // (1.0 - t) * Vtemp[i-1][j].y + t * Vtemp[i-1][j+1].y;
            }
        }
        
// System.out.println("Reached 3");
        if (Left != null) {
            for (j = 0; j <= degree; j++) {
                Left[j]  = Vtemp[j][0];
            }
        }
// System.out.println("Reached 4");
        if (Right != null) {
            for (j = 0; j <= degree; j++) {
                Right[j] = Vtemp[degree-j][j];
            }
        }
// System.out.println("Reached 5");

        return (Vtemp[degree][0]);
    }


    /*==========================================================================*/
    /* Some functions that are not used any more, but I don't want to throw off */

/*******************************************************
    private double slopeDiff(TGPoint p, double t) {
        TGPoint q = evaluate(t);
        TGPoint n = normal(t);
        return Math.abs((p.y - q.y) / (p.x - q.x)) - Math.abs(n.y / n.x);
    }

    private double distanceSquare(TGPoint p, double t) {
        TGPoint q = evaluate(t);
        return ((q.x - p.x) * (q.x - p.x) + (q.y - p.y) * (q.y - p.y));
    }

    public final double findNearestParametric(TGPoint point) {
        double traverseEpsilon = 0.01;
        // Calculate the distance from beginning point to test point
        double d1 = (point.x - fPoint0_fX) * (point.x - fPoint0_fX) + (point.y - fPoint0_fY) * (point.y - fPoint0_fY);
        d1 = Math.sqrt(d1);

        // Calculate the distance from end point to test point
        double d2 = (point.x - fPoint3_fX) * (point.x - fPoint3_fX) + (point.y - fPoint3_fY) * (point.y - fPoint3_fY);
        d2 = Math.sqrt(d2);

        double midValue = d1 / (d1 + d2);
        double compareDistance = distanceSquare(point, midValue);

        // now attempt to go a bit to the left and a bit to the right
        double leftValue = midValue - traverseEpsilon;
        double rightValue = midValue + traverseEpsilon;

        int iterationCount = 0;

        while (iterationCount++ < 20) {
            double leftDistance = distanceSquare(point, leftValue);
            double rightDistance = distanceSquare(point, rightValue);

System.out.println("leftValue = " + leftValue + "  leftDistance = " + leftDistance);
System.out.println(" midValue = " + midValue + " compareDistance = " + compareDistance);
System.out.println("rightValue = " + rightValue + "  rightDistance = " + rightDistance);
            if (leftDistance < compareDistance) {
                // We need to go to the left
                midValue = leftValue;
                compareDistance = leftValue;
            }
            else if (rightDistance < compareDistance) {
                // We need to go right
                midValue = rightValue;
                compareDistance = rightValue;
            }
            else {
                // We are already at a minimum
                break;
            }
            leftValue = midValue - traverseEpsilon;
            rightValue = midValue + traverseEpsilon;
        }
        return midValue;
    }

    public final double findNearestParametric1(TGPoint point) {
        // Calculate the distance from beginning point to test point
        double d1 = (point.x - fPoint0_fX) * (point.x - fPoint0_fX) + (point.y - fPoint0_fY) * (point.y - fPoint0_fY);
        d1 = Math.sqrt(d1);

        // Calculate the distance from end point to test point
        double d2 = (point.x - fPoint3_fX) * (point.x - fPoint3_fX) + (point.y - fPoint3_fY) * (point.y - fPoint3_fY);
        d2 = Math.sqrt(d2);

        double midValue = d1 / (d1 + d2);

        double S_middle = slopeDiff(point, midValue);
        if (S_middle == 0.0)  return midValue;  // we have already found a result!
System.out.println("midValue = " + midValue);
        // We know that the result will be close to midValue
        // For the bisection method we need to find an interval containing the root
        // We assume that the interval between 'upper' and 'lower' contains the root

        double lower = Math.max(0.0, midValue - 0.1);
        double upper = Math.min(1.0, midValue + 0.1);

        double S_lower = slopeDiff(point, lower);
        double S_upper = slopeDiff(point, upper);

System.out.println("lower = " + lower + " and upper = " + upper);
System.out.println("S_lower = " + S_lower + " and S_upper = " + S_upper);
        // Iterate around the mid value to determine the closest value

        int iterationCount = 0;

        while ((iterationCount++ < 10) && !Eq(S_lower, S_upper, 0.001)) {
//System.out.println("iteration count = " + iterationCount);
            midValue = lower + S_lower * (upper - lower ) / (S_upper + S_lower);
            if (S_lower < S_upper) {     // we change the upper value
                upper = midValue;
            }
            else {    // we change the lower value
                lower = midValue;
            }
            S_lower = slopeDiff(point, lower);
            S_upper = slopeDiff(point, upper);
            if (S_lower == 0.0)  return lower;  // we have already found a result!
            if (S_upper == 0.0)  return upper;  // we have already found a result!
System.out.println("lower = " + lower + " and upper = " + upper);
System.out.println("S_lower = " + S_lower + " and S_upper = " + S_upper);
        }

        // return the parametric value
        return midValue;
    }

    // A function that tells whether two Beziers have a common end point
    // This function returns two sets of parametric values where common end points
    // were observed.
    // returnValue[0] indicates the end-points where this curve has commonality
    // returnValue[1] indicates the end-points where the other curve has commonality
    // Caller must ensure that returnParameter = new double[2] is called beforehand

    private boolean hasCommonEndPoint(TBezier other, double[][] returnParameter) {
        boolean returnValue = false;
        returnParameter[0][0] = returnParameter[0][1] = returnParameter[1][0] = returnParameter[1][1] = Double.POSITIVE_INFINITY;
        if ((fPoint0_fX == other.fPoint0_fX) && (fPoint0_fY == other.fPoint0_fY)) {
            returnParameter[0][0] = 0.0;
            returnParameter[1][0] = 0.0;
            returnValue = true;
        }
        if ((fPoint3_fX == other.fPoint0_fX) && (fPoint3_fY == other.fPoint0_fY)) {
            returnParameter[0][1] = 1.0;
            returnParameter[1][0] = 0.0;
            returnValue = true;
        }
        if ((fPoint0_fX == other.fPoint3_fX) && (fPoint0_fY == other.fPoint3_fY)) {
            returnParameter[0][0] = 0.0;
            returnParameter[1][1] = 1.0;
            returnValue = true;
        }
        else if ((fPoint3_fX == other.fPoint3_fX) && (fPoint3_fY == other.fPoint3_fY)) {
            returnParameter[0][1] = 1.0;
            returnParameter[1][1] = 1.0;
            returnValue = true;
        }
        return returnValue;
    }

    public TGPoint tangent(double t) {
        double xValue = - 3.0 *            (1.0-t) *           (1.0-t)  * fPoint0_fX +
                          3.0 * ((1.0-t) * (1.0-t) - 2.0 * t * (1.0-t)) * fPoint1_fX +
                          3.0 * (-    t  *      t  + 2.0 * t * (1.0-t)) * fPoint2_fX +
                          3.0 *       t  *                 t            * fPoint3_fX;
        double yValue = - 3.0 *            (1.0-t) *           (1.0-t)  * fPoint0_fY +
                          3.0 * ((1.0-t) * (1.0-t) - 2.0 * t * (1.0-t)) * fPoint1_fY +
                          3.0 * (-    t  *      t  + 2.0 * t * (1.0-t)) * fPoint2_fY +
                          3.0 *       t  *                 t            * fPoint3_fY;
        return new TGPoint(xValue, yValue);
    }

    public static double[][] findIntersections(TBezier a, TBezier b) {

        if (TBezier.IntersectBB(a, b) == true) {
            if (a.isLinear() && b.isLinear()) {

                TGPoint A = a.getPoint(0);
                TGPoint B = a.getPoint(3);
                TGPoint C = b.getPoint(0);
                TGPoint D = b.getPoint(3);

                if (A.equals(C) || A.equals(D) || B.equals(C) || B.equals(D))
                    return null;

                double denom = (B.x - A.x)*(D.y - C.y) - (B.y - A.y)*(D.x - C.x);
                if (denom == 0.0)
                    return null;    // Better than getting an exception
                double num1 =  (C.x - A.x)*(D.y - C.y) - (C.y - A.y)*(D.x - C.x);
                double t1 = num1 / denom;
                double num2 =  (B.x - A.x)*(A.y - C.y) - (B.y - A.y)*(A.x - C.x);
                double t2 = num2 / denom;

                if (((t1 > 0.0) && (t1 < 1.0)) && ((t2 > 0.0) && (t2 < 1.0))) {
                    double[][] parameters = new double[2][1];
                    //parameters[0] = new double[1];
                    //parameters[1] = new double[1];
                    parameters[0][0] = t1;
                    parameters[1][0] = t2;
                    return parameters;
                }
                else
                    return null;
            }
            else {
                // Two Bezier curves can intersect at at-most 9 places. We prepare ourselves
                // for the worst scenario and allocate enough memory to start with.
                double[][] parameters = new double[2][9];
                int[] index = new int[1];    // Just a wrapper for an integer return parameter
                index[0] = 0;

                // Determine sub-divisions by checking for bounding-box intersections
                int ra = a.subDivisionDepth();
                int rb = b.subDivisionDepth();

                TBezier.RecursivelyIntersect(a, 0., 1., ra, b, 0., 1., rb, parameters, index);

                // The value of index[0] now tells us how many places the curves intersect.
                // We will resize the double arrays to eliminate redundant values.

                double[][] justRightParameters = new double[2][index[0]];

                // Sort the parameter values before you return
                justRightParameters[0] = SortAndEliminateDuplicates(parameters[0], index[0]);
                justRightParameters[1] = SortAndEliminateDuplicates(parameters[1], index[0]);

                return justRightParameters;
            }
        }
        return null;

    }

*************************************************************/

}


