/*
 * @(#)TCAGAdd.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

// Class for storing the value of the addition of two children in the tree
class TCAGAdd extends TCAGBinary {

    public TCAGAdd(TCAGNode parent) {
        super(parent);
        
        fFillCount = 0;
        fLeftFrameCount = 0;
        fRightFrameCount = 0;
    
        fFillCountLeftSide = 0;
        fLeftFrameCountLeftSide = 0;
        fRightFrameCountLeftSide = 0;
    
        fFillCountRightSide = 0;
        fLeftFrameCountRightSide = 0;
        fRightFrameCountRightSide = 0;
    }

    public void transition(int fill) {
        int oldFillCount = fFillCount;
        fFillCountLeftSide = fill;
        fFillCount = (fFillCountLeftSide | fFillCountRightSide);
    
        if (oldFillCount != fFillCount)     // count has changed?
            getParent().transition(fFillCount);
    }
    public void transitionLeftFrame(int leftFrame) {
        fLeftFrameCountLeftSide = leftFrame;
    
        int oldLeftFrameCount = fLeftFrameCount;
        fLeftFrameCount = Sign(fLeftFrameCountLeftSide + fLeftFrameCountRightSide);
    
        if (oldLeftFrameCount != fLeftFrameCount)
            getParent().transitionLeftFrame(fLeftFrameCount);
    }
    public void transitionRightFrame(int rightFrame) {
        fRightFrameCountLeftSide = rightFrame;
    
        int oldRightFrameCount = fRightFrameCount;
        fRightFrameCount = Sign(fRightFrameCountLeftSide + fRightFrameCountRightSide);
    
        if (oldRightFrameCount != fRightFrameCount)
            getParent().transitionRightFrame(fRightFrameCount);
    }

    public void transitionRightSide(int fill) {
        int oldFillCount = fFillCount;
        fFillCountRightSide = fill;
        fFillCount = (fFillCountLeftSide | fFillCountRightSide);
    
        if (oldFillCount != fFillCount)     // state has changed?
            getParent().transition(fFillCount);
    }
    public void transitionLeftFrameRightSide(int leftFrame) {
        fLeftFrameCountRightSide = leftFrame;
    
        int oldLeftFrameCount = fLeftFrameCount;
        fLeftFrameCount = Sign(fLeftFrameCountLeftSide + fLeftFrameCountRightSide);
    
        if (oldLeftFrameCount != fLeftFrameCount)
            getParent().transitionLeftFrame(fLeftFrameCount);
    }
    public void transitionRightFrameRightSide(int rightFrame) {
        fRightFrameCountRightSide = rightFrame;
    
        int oldRightFrameCount = fRightFrameCount;
        fRightFrameCount = Sign(fRightFrameCountLeftSide + fRightFrameCountRightSide);
    
        if (oldRightFrameCount != fRightFrameCount)
            getParent().transitionRightFrame(fRightFrameCount);
    }

    // Wrap count of fill & frame for operation
    private int fFillCount;
    private int fLeftFrameCount;
    private int fRightFrameCount;

    // Wrap count of fill & frame for left child
    private int fFillCountLeftSide;
    private int fLeftFrameCountLeftSide;
    private int fRightFrameCountLeftSide;

    // Wrap count of fill & frame for right child
    private int fFillCountRightSide;
    private int fLeftFrameCountRightSide;
    private int fRightFrameCountRightSide;
}
