/*
 * @(#)TCAGEvenOdd.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

// Class for storing the boolean value of an even-odd geometry
class TCAGEvenOdd extends TCAGUnary {

    public TCAGEvenOdd(TCAGNode parent) {
        super(parent);
        fFillCount = 0;
        fLeftFrameCount = 0;
        fLeftFrameCountNorm = 0;
        fRightFrameCount = 0;
        fRightFrameCountNorm = 0;
    }

    public void transition(int fill) {
        fFillCount += fill;
        getParent().transition(fFillCount & 1);
    }
    public void transitionLeftFrame(int leftFrame) {
        fLeftFrameCount += leftFrame;
        evaluate();
    }
    public void transitionRightFrame(int rightFrame) {
        fRightFrameCount += rightFrame;
        evaluate();
    }

    public int getType() {
        return kEvenOdd;
    }

    private int fFillCount;             // wrap count for fill
    private int fLeftFrameCount;        // wrap count for left frame
    private int fLeftFrameCountNorm;    // normalized
    private int fRightFrameCount;       // wrap count for right frame
    private int fRightFrameCountNorm;   // normalized

    private void evaluate() {
        int oldLeftFrameCount = fLeftFrameCountNorm;
        int oldRightFrameCount = fRightFrameCountNorm;
    
        if (fLeftFrameCount != fRightFrameCount) {  // in the frame?
            fLeftFrameCountNorm = 1;
            fRightFrameCountNorm = 0;
        }
        else if (fLeftFrameCount != 0) {  // inside?
            fLeftFrameCountNorm = 1;
            fRightFrameCountNorm = 1;
        }
        else {   // outside
            fLeftFrameCountNorm = 0;
            fRightFrameCountNorm = 0;
        }
    
        if (oldLeftFrameCount != fLeftFrameCountNorm)
            getParent().transitionLeftFrame(fLeftFrameCountNorm);
        if (oldRightFrameCount != fRightFrameCountNorm)
            getParent().transitionRightFrame(fRightFrameCountNorm);
    }
}
