/*
 * @(#)TEllipseGeometry.java	1.7 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TEllipseGeometry
||
*/

public class TEllipseGeometry extends MAreaGeometry {

    public TGEllipse fEllipse;
    private static final TGRect fkZeroRect =  // Why?  Look at TGStatics.java
        new TGRect (TGStatics.kZeroRectTopLeftX,
                    TGStatics.kZeroRectTopLeftY,
                    TGStatics.kZeroRectBottomRightX,
                    TGStatics.kZeroRectBottomRightY);

    public TEllipseGeometry() {
        fEllipse = new TGEllipse(fkZeroRect);
    }
    public TEllipseGeometry(TGEllipse ellipse) {
        fEllipse = new TGEllipse(ellipse);
        updateTimeStamp();
    }
    public TEllipseGeometry(TEllipseGeometry source) {
        super(source);
        fEllipse = new TGEllipse (source.fEllipse);
    }

    // === MAreaGeometry overrides ===
    public MAreaGeometry cloneAndTransform(TGrafMatrix m) {
        if (m.isRectilinear()) {
            return new TEllipseGeometry(new TGEllipse(m.transformBounds(fEllipse.getBounds())));
        }
        else
            return super.cloneAndTransform(m);
    }

    public void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix) {

        TCAGEvenOdd node = new TCAGEvenOdd(parentNode);
        extractor.adoptCAGNode(node);
        extractor.extractEllipse(fEllipse, node, matrix);
    }

    public boolean isEmpty() {
        return getBounds().isEmpty();
    }

    public TGRect getBounds() {
        return fEllipse.getBounds();
    }
    public boolean intersects(TGRect rect) {
        return fEllipse.intersects(rect);
    }
    public boolean contains(TGPoint point) {
        return fEllipse.contains(point);
    }
    public boolean contains(TGRect rect) {
        return (fEllipse.contains(rect.getTopLeft()) &&
                fEllipse.contains(rect.getTopRight()) &&
                fEllipse.contains(rect.getBottomLeft()) &&
                fEllipse.contains(rect.getBottomRight()));
    }

    // === operators ===
    public final TEllipseGeometry copyFrom(TEllipseGeometry source) {
        if (this != source) {
            super.copyFrom(source);
            fEllipse.copyFrom(source.fEllipse);
        }
        return this;
    }

    public Object clone() {
        return (Object)(new TEllipseGeometry(this));
    }
    public boolean equals(MAreaGeometry that) {
        if ((getClass().equals(that.getClass()))) {
            TEllipseGeometry ellipseGeometry = (TEllipseGeometry)that;
            return (super.equals(that) && fEllipse.equals(ellipseGeometry.fEllipse));
        }
        else
            return false;
    }
}
