/*
 * @(#)TExclusiveOrGeometry.java	1.5 98/04/02
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TExclusiveOrGeometry
||
||  This class owns two geometries whose areas are logically exclusive-ored.
||
*/

class TExclusiveOrGeometry extends MAreaGeometry {

    public TExclusiveOrGeometry() {
        fBounds = null;
    }
    public TExclusiveOrGeometry(
        MAreaGeometry left,
        MAreaGeometry right) {

        fLeft = left;
        fRight = right;
        fBounds = null;
        updateTimeStamp();
    }
    public TExclusiveOrGeometry(TExclusiveOrGeometry source) {
        super(source);

        fLeft = null;
        fRight = null;
        fBounds = null;

        if (source.fLeft != null) {
            fLeft = (MAreaGeometry) ( source.fLeft.clone() );
        }
        if (source.fRight != null) {
            fRight = (MAreaGeometry) ( source.fRight.clone() );
        }
        if (source.fBounds != null) {
            fBounds = (TGRect) ( source.fBounds.clone() );
        }
    }

    // === MAreaGeometry overrides ===
    public MAreaGeometry cloneAndExclusiveOr(MAreaGeometry g) {
        TExclusiveOrGeometry newGeometry = new TExclusiveOrGeometry(this);
        newGeometry.fLeft = MAreaGeometry.exclusiveOr(newGeometry.fLeft, g);
        newGeometry.fBounds = null;
        newGeometry.updateTimeStamp();
        return newGeometry;
    }

    public void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix) {

        TCAGExclusiveOr node = new TCAGExclusiveOr(parentNode);
        extractor.adoptCAGNode(node);
        fLeft.extract(extractor, node, matrix);
        fRight.extract(extractor, node.getRightLink(), matrix);
    }

    public boolean isEmpty() {
        return fLeft.isEmpty() && fRight.isEmpty();
    }
    public boolean isPolygonal() {
        return fLeft.isPolygonal() && fRight.isPolygonal();
    }
    public boolean isSingular() {
        return false;
    }

    public TGRect getBounds() {
        if (fBounds == null) {
            fBounds = new TGRect(fLeft.getBounds());
            fBounds.extendTo(fRight.getBounds());
        }
        return fBounds;
    }
    public boolean intersects(TGRect rect) {
        return fLeft.intersects(rect) || fRight.intersects(rect);
    }
    public boolean contains(TGPoint point) {
        if (getBounds().contains(point)) {
            boolean leftContains = fLeft.contains(point);
            boolean rightContains = fRight.contains(point);
            return (leftContains ^ rightContains);
        }
        else
            return false;
    }
    public boolean contains(TGRect rect) {
        return (fLeft.contains(rect) && !fRight.intersects (rect)) || (fRight.contains(rect) && !fLeft.intersects (rect));
    }

    public final TExclusiveOrGeometry copyFrom(TExclusiveOrGeometry source) {
        if (this != source) {
            super.copyFrom(source);
            fLeft = source.fLeft;
            fRight = source.fRight;
            fBounds = source.fBounds;
        }
        return this;
    }

    public Object clone() {
        return (Object)(new TExclusiveOrGeometry(this));
    }
    public boolean equals(MAreaGeometry that) {
        if ((getClass().equals(that.getClass()))) {
            TExclusiveOrGeometry geometry = (TExclusiveOrGeometry)that;
            return (super.equals(that) || (fLeft == geometry.fLeft && fRight == geometry.fRight));
        }
        else
            return false;
    }

    private MAreaGeometry fLeft;
    private MAreaGeometry fRight;
    private TGRect fBounds;
}
