/*
 * @(#)TGEllipse.java	1.7 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TGEllipse
||
*/

/*
    A TGEllipse is an ellipse defined by an axis aligned rectangle (a TGRect).  The
    ellipse is inscribed in the rectangle so it just touches at the midpoint of each
    edge of the rectangle.  A TGEllipse is area enclosing geometry that
    may be filled and/or framed.;
*/

public class TGEllipse implements Cloneable {

    private TGRect fRectangle;

    // Creates an ellipse centered at the origin with zero radius.
/*
    public TGEllipse() {
        fRectangle = new TGRect (0., 0., 0., 0.); // Look at TGStatics.java
    }
*/
    public TGEllipse(TGRect r) {
        fRectangle = (TGRect)r.clone();
    }
/*
    public TGEllipse(TGPoint center, double radius) {
        fRectangle = new TGRect(center, center);
        fRectangle.inset(new TGPoint(-radius, -radius));
    }
    public TGEllipse(TGPoint center, double Xradius, double Yradius) {
        fRectangle = new TGRect(center, center);
        fRectangle.inset(new TGPoint(-Xradius, -Yradius));
    }
*/
    public TGEllipse(TGEllipse src) {
        fRectangle = new TGRect(src.fRectangle);
    }

    // operators
    public TGEllipse copyFrom(TGEllipse src) {
        fRectangle.copyFrom(src.fRectangle);
        return this;
    }
    public boolean equals(TGEllipse src) {
        return fRectangle.equals(src.fRectangle);
    }

    // Data access
    public TGPoint getSize() {
        return fRectangle.getSize();
    }
    public TGPoint getCenter() {
        return fRectangle.getCenter();
    }

    // Returns true of the ellipse crosses rect
    public boolean intersects(TGRect r) {

        if (!fRectangle.intersects(r))   // Trivial rejection
            return false;

        // If any point is inside, it intersects
        if (contains(r.getTopLeft()) || contains(r.getTopRight()) ||
            contains(r.getBottomLeft()) || contains(r.getBottomRight()))
            return true;

        // At this point, we know r intersects the bounding box, but
        // none of it's points are inside the ellipse.  The only way it
        // can still not intersect is if all four corners of r are in the
        // gap between one corner of the ellipse and its bounding box.

        // To see if this is the case, we subtract r's coordinates
        // from the center.  If the signs are the same, then the coordinates
        // must be on the same side (likewise for top & bottom).

        TGPoint centerpoint = getCenter();
        double hsign = (r.x - centerpoint.x) * (r.getMaxX() - centerpoint.x);
        double vsign = (r.y - centerpoint.y) * (r.getMaxY() - centerpoint.y);

        return (hsign <= 0.0) || (vsign <= 0.0);
    }

    // Returns true if the point is inside the ellipse
    public boolean contains(TGPoint point) {

        TGPoint rad = getSize();

        // Point can't be in an empty oval (also avoid /0.0)
        if ((rad.x == 0.0) || (rad.y == 0.0))
            return false;

        TGPoint cen = getCenter();
        rad.divideBy(2.0);
        TGPoint testPt = (TGPoint)point.clone();

        // scale test point to unit circle based on the oval
        testPt.subtractBy(cen);
        testPt.divideBy(rad);
        testPt.multiplyBy(testPt);   // testPt^2

        return (testPt.x + testPt.y) < 1.0;
    }

    public TGRect getBounds() {
        return (TGRect)fRectangle.clone();
    }

    // Modifiers
/*
    public void setCenter(TGPoint center) {
        fRectangle.setCenter(center);
    }
    public void setSize(TGPoint size) {
        fRectangle.setSize(size);
    }
    public void setBounds(TGRect bounds) {
        fRectangle.copyFrom(bounds);
    }
*/
}
