/*
 * @(#)TGLine.java	1.6 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TGLine
||
*/

/*
    TGLine is a line segment specified by two endpoints.;
*/

class TGLine implements Cloneable {

    private static final TGPoint fkOrigin =  // Why? Look at TGStatics.java
        new TGPoint(TGStatics.kOriginX, TGStatics.kOriginY);

    public TGLine() {
        this(fkOrigin, fkOrigin);
    }
    public TGLine(TGPoint startPt, TGPoint endPt) {
        fStartPt = new TGPoint(startPt);
        fEndPt = new TGPoint(endPt);
    }
    public TGLine(TGLine src) {
        this(src.fStartPt, src.fEndPt);
    }

    public Object clone() {
        return new TGLine(this);
    }

    // Data access
/*
    public TGPoint getStartPoint() { return new TGPoint(fStartPt); }
    public TGPoint getEndPoint() { return new TGPoint(fEndPt); }
*/
    public void getPoints(TGPoint startPt, TGPoint endPt) {
        startPt.copyFrom(fStartPt);
        endPt.copyFrom(fEndPt);
    }

    // Returns point on the line corresponding to the parameter;
    // 0==Start point, 1==Endpoint.  Values outside 0..1 are
    // extrapolated beyond the endpoints.
    public TGPoint evaluate(double u) {
        return
            TGPoint.add(
                TGPoint.multiply(fStartPt, (1.0 - u)),
                TGPoint.multiply(fEndPt, u));
    }

    // Modifiers
/*
    public void setStartPoint(TGPoint point) {
        fStartPt = new TGPoint(point);
    }
    public void setEndPoint(TGPoint point) {
        fEndPt = new TGPoint(point);
    }
    public void transformBy(TGrafMatrix mat) {
        fStartPt = mat.transformPoint(fStartPt);
        fEndPt = mat.transformPoint(fEndPt);
    }

    // Moves the line so the point at the parametric position u
    // is moved to "toPoint"
    public void dragPosition(double parametric, TGPoint toPoint) {
        if ((parametric <= 1.0) && (parametric >= 0.0))
        {
            double weight0 = 1.0 - parametric;
            TGPoint delta =
                TGPoint.subtract(
                    toPoint,
                    TGPoint.add(
                        TGPoint.multiply(fStartPt, weight0),
                        TGPoint.multiply(fEndPt, parametric)));
            double denom = weight0 * weight0 + parametric * parametric;
            fStartPt.addBy(TGPoint.multiply(delta, weight0 / denom));
            fEndPt.addBy(delta.multiplyBy(parametric / denom));
        }
    }
*/

/*
    // Returns u value for point on line nearest to "test". Values
    // returned are clamped to 0 >= u >= 1.
    public double nearestParametric(TGPoint test) {
        return prStatic.PointSegDist(fStartPt, fEndPt, test); // prSplineRoutines
    }
*/

    private static final int kLeftCode      = 0x1;
    private static final int kRightCode     = 0x2;
    private static final int kTopCode       = 0x4;
    private static final int kBottomCode    = 0x8;
    private static int ClipCode(TGPoint p, TGRect r)
    {
        int c = 0;

        if (p.x < r.x) c |= kLeftCode;
        else if (p.x > r.getMaxX()) c |= kRightCode;

        if (p.y < r.y) c |= kTopCode;
        else if (p.y > r.getMaxY()) c |= kBottomCode;

        return c;
    }

    // Returns true if the line intersects the rect
    public boolean intersects(TGRect g) {
        TGPoint clipStart = fStartPt;
        TGPoint clipEnd = fEndPt;

        int startcode = ClipCode(clipStart, g);
        int endcode = ClipCode(clipEnd, g);
        if ((startcode & endcode) != 0)    // line is completly to one side of the rect
            return false;

        if ((startcode == 0) || (endcode == 0))
            return true;            // One of the endpoints is completly inside

        // if Horiz, or vert, & codes don't match, it intersects
        if ((clipStart.x == clipEnd.x) || (clipStart.y == clipEnd.y))
            return true;

        // Do the full blown clip
        while ((startcode | endcode) != 0)
        {
            if ((startcode & endcode) != 0)
                return false;       // All clipped away

            int code = startcode;
            if (code == 0) code = endcode;

           TGPoint p = new TGPoint();
           if ((code & kLeftCode) != 0) {
                p.y = clipStart.y + (clipEnd.y - clipStart.y) * ((g.x - clipStart.x) / (clipEnd.x - clipStart.x));
                p.x = g.x;
            } else if ((code & kRightCode) != 0) {
                double right = g.width + g.x;
                p.y = clipStart.y + (clipEnd.y - clipStart.y) * ((right - clipStart.x) / (clipEnd.x - clipStart.x));
                p.x = right;
            } else if ((code & kTopCode) != 0) {
                p.x = clipStart.x + (clipEnd.x - clipStart.x) * ((g.y - clipStart.y) / (clipEnd.y - clipStart.y));
                p.y = g.y;
            } else if ((code & kBottomCode) != 0) {
                double bottom = g.height + g.y;
                p.x = clipStart.x + (clipEnd.x - clipStart.x) * ((bottom - clipStart.y) / (clipEnd.y - clipStart.y));
                p.y = bottom;
            }

            if (code == startcode) {
                clipStart = p;
                startcode = ClipCode(p, g);
            } else {
                clipEnd = p;
                endcode = ClipCode(p, g);
            }
        }
        return (true);
    }

    public TGRect getBounds() {
        return
            new TGRect(
                fStartPt.x < fEndPt.x ? fStartPt.x : fEndPt.x,
                fStartPt.y < fEndPt.y ? fStartPt.y : fEndPt.y,
                fStartPt.x > fEndPt.x ? fStartPt.x : fEndPt.x,
                fStartPt.y > fEndPt.y ? fStartPt.y : fEndPt.y);
    }

// No assignment operator:
// Should use copy constructor when clone. ie, TGPoint clone = new TGPoint(src);

    public boolean equals(TGLine Src) {
        return (fStartPt.equals(Src.fStartPt) && fEndPt.equals(Src.fEndPt));
    }

    private TGPoint fStartPt;
    private TGPoint fEndPt;
}
