/*
 * @(#)TGParametricArray.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
    TGParametricArray provides a low level substitute for traditional arrays which can grow
    and shrink when needed.  Since there is not templates in Java, only a Collectible
    type objects can be stored.

    Once instantiated, one should make certain only one (concrete) type of objects be
    stored in it.
*/

class TGParametricArray extends TRawPrimitiveArray {

    private double[] fValues = null;

    public TGParametricArray() {
        this(0);
    }
    public TGParametricArray(int size) {
        super(size);
        if (fArraySize > 0)
            allocateArray(fArraySize);
    }
    public TGParametricArray(TGParametricArray source) {
        if (source.fNumberOfValues > 0) {
            allocateArray(source.fArraySize);
            System.arraycopy(source.fValues, 0, fValues, 0, source.fNumberOfValues);
        }
        super.copyFrom(source);
    }

    public final TGParametricArray copyFrom(TGParametricArray source) {
        if (fArraySize != source.fArraySize)
            allocateArray(source.fArraySize);
        System.arraycopy(source.fValues, 0, fValues, 0, source.fNumberOfValues);
        super.copyFrom(source);
        return this;
    }

    public final boolean equals(TGParametricArray source) {
        if (source == this) return true;
        if (super.equals(source)) return false;
        for (int i = 0; i < fNumberOfValues; ++i) {
            if (fValues[i] != source.fValues[i])
                return false;
        }
        return true;
    }

    // access
    public double getValue(int i) {
        if (i < 0 || i >= fNumberOfValues)
            throw new IndexOutOfBoundsException("RawArray.value(index)");
        return fValues[i];
    }
    public void setValue(int i, double p) {
        if (i < 0 || i >= fNumberOfValues)
            throw new IndexOutOfBoundsException("RawArray.setValue(index, ...)");
        fValues[i] = p;
    }
    public void append(double p) {
        resize(fNumberOfValues + 1); // This changes fNumberOfValues!
        fValues[fNumberOfValues - 1] = p;
    }

    // just like an array in C, there is NO BOUNDS CHECKING
    public double[] getArray() {
        return fValues;
    }

    void allocateArray(int size) {
        fValues = null;
        if (size > 0)
            fValues = new double[size];
    }
    void reallocateArray(int newSize) {
        double oldValues[] = fValues;
        allocateArray(newSize);
        if (newSize > 0)
            System.arraycopy(
                oldValues, 0,
                fValues, 0,
                fArraySize > newSize ? newSize : fArraySize);
    }

    // This swaps the contents of an array with victim.
    public void swap(TGParametricArray victim) {
        super.swap(victim);

        double tmpData[] = fValues;
        fValues = victim.fValues;
        victim.fValues = tmpData;
    }

    public String toString() {
        String vals = new String();
        int nump = fNumberOfValues;
        for (int j = 0; j < nump; j++)
             vals = vals + getValue(j) + ",";
        return ("{ num values = " + nump + "; " + vals + "}");
    }
}
