/*
 * @(#)TGPoint.java	1.8 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

import java.awt.geom.Point2D;

/*================
||
||  TGPoint
||
||  TGPoint provides the basic means for communicating between transforms
||  and the rest of the 2D world.  Thus it is defined herein.
||
*/

/*
    TGPoints are basic two-dimensional points in Pink geometry.  They are not intended
    to be subclassed, and their components are deliberatly public.  Note the handy
    arithmetic operators; these perform vector arithmetic on the points. ;
*/

public final class TGPoint extends Point2D.Double {

    /*************************************************************************
    private static final TGPoint fkOrigin =  // Why? Look at TGStatics.java
        new TGPoint(TGStatics.kOriginX, TGStatics.kOriginY);

    private static final TGPoint fkInfinite =   // Why? Look at TGStatics.java
        new TGPoint(TGStatics.kPosInfinityX, TGStatics.kPosInfinityY);
    *************************************************************************/

    // Does NOT initialize data members
    public TGPoint() {
        super();
    }

    public TGPoint(Point2D point) {
        this((double)point.getX(), (double)point.getY());
    }

    public TGPoint(double x, double y){
        super(x, y);
    }

    public Object clone() {
        return new TGPoint(this);
    }

    // Assignment operator:
    public final TGPoint copyFrom(TGPoint src) {
        x = src.x;
        y = src.y;
        return this;
    }

    public boolean equals(TGPoint src) {
        return (x == src.x) && (y == src.y);
    }

/*
    // Allows XY indexed as 0 and 1
    public double element(int index) {
        switch (index) {
            case 0: return x;
            case 1: return y;
            default:
                throw new IndexOutOfBoundsException("TGPoint.element(): index");
        }
    }
*/

    public TGPoint negate() {
        return new TGPoint(-x,-y);
    }

    private static final double kHalfValue = 0.5;  // Shouldn't it be in prStatic? // prNumeric

/*
    // Operator to convert from TGPoint to Point
    public Point toPoint() {
        return new Point((int)(x + kHalfValue), (int)(y + kHalfValue));
    }
*/

    // Operators for vector arithmetic.
    public TGPoint addBy(TGPoint rhs) {
        x += rhs.x;
        y += rhs.y;
        return this;
    }
    public TGPoint subtractBy(TGPoint rhs) {
        x -= rhs.x;
        y -= rhs.y;
        return this;
    }
    public TGPoint multiplyBy(TGPoint rhs) {
        x *= rhs.x;
        y *= rhs.y;
        return this;
    }
    public TGPoint multiplyBy(double rhs) {
        x *= rhs;
        y *= rhs;
        return this;
    }
    public TGPoint divideBy(TGPoint rhs) {
        x /= rhs.x;
        y /= rhs.y;
        return this;
    }
    public TGPoint divideBy(double rhs) {
        x /= rhs;
        y /= rhs;
        return this;
    }

    public static TGPoint add(TGPoint a, TGPoint b) {
        return new TGPoint(a.x + b.x, a.y + b.y);
    }
    public static TGPoint subtract(TGPoint a, TGPoint b) {
        return new TGPoint(a.x - b.x, a.y - b.y);
    }
/*
    public static TGPoint multiply(TGPoint a, TGPoint b) {
        return new TGPoint(a.x * b.x, a.y * b.y);
    }
*/
    public static TGPoint multiply(double scale, TGPoint pt) {
        return new TGPoint(pt.x * scale, pt.y * scale);
    }
    public static TGPoint multiply(TGPoint pt, double scale) {
        return new TGPoint(pt.x * scale, pt.y * scale);
    }
/*
    public static TGPoint divide(TGPoint a, TGPoint b) {
        return new TGPoint(a.x / b.x, a.y / b.y);
    }
    public static TGPoint divide(TGPoint pt, double scale) {
        return new TGPoint(pt.x / scale, pt.y / scale);
    }
*/

    // Compute the dot product x*v1 + y*v1
    public double dotProduct(TGPoint vector) {
        return (x * vector.x) + (y * vector.y);
    }

/*
    // Compute the cross product x*vy - vx*y
    public double crossProduct(TGPoint vector) {
        return (x * vector.y) - (vector.x * y);
    }

    // Linear interpolate to dest (0=*this, 1=dest)
    public TGPoint interpolate(TGPoint dest, double u) {
        // (1.0-u) * this + u * dest
        double r = 1.0 - u;
        return new TGPoint(x * r + dest.x * u, y * r + dest.y * u);
    }
*/

    // Vector length: sqrt(x*x + y*y)
    public double vectorLength() {
        return Math.sqrt(x * x + y * y);
    }

    // Vector Squared length: (x*x + y*y)
    public double vectorSquaredLength() {
        return (x * x + y * y);
    }

/*
    // normalize the point x /= length, y /= length
    public double normalize() {
        double length = vectorLength();
        if (length == 0.0) {
            return 0.0;
        }
        x /= length;
        y /= length;
        return length;
    }
*/

    // Returns the angle Q of the vector w.r.t. the X axis (0 <= Q < 360 degrees)
    public double vectorAngle() {
        double offset = 0.0;

        if (x == 0.0) {
            if (y == 0.0)
                return 0.0;
            if (y > 0.0)
                return 90.0;
            else
                return 270.0;
        }
        else {
            if ((x > 0.0) && (y < 0.0))
                offset = 360.0;
            else
                if (x < 0.0)
                    offset = 180.0;
            return Math.atan(y / x) * prStatic.kRadiansToDegrees + offset; // prNumeric
        }
    }

/*
    // set the length w/o changing the angle
    public void setVectorLength(double newLength) {
        // setPolarCoordinates(vectorAngle(), length);  // Too slow!
        double length = vectorLength();
        if (length > 0.0)
            multiplyBy(newLength / length);
        else {
            x = newLength;
            y = 0.0;
        }
    }

    // set the angle w/o changing the length
    public void setVectorAngle(double angle) {
        setPolarCoordinates(angle, vectorLength());
    }
*/

    // Sets the point to the position specified in polar coordinates
    public void setPolarCoordinates(double angle, double length) {
        x = Math.cos(angle * prStatic.kDegreesToRadians) * length; // prNumeric
        y = Math.sin(angle * prStatic.kDegreesToRadians) * length; // prNumeric
    }
    public void setPolarCoordinates(double angle) {
        setPolarCoordinates(angle, 1.0);
    }
}
