/*
 * @(#)TGPolygon.java	1.6 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

import java.awt.Polygon;

/*================
||
||  TGPolygon
||
*/

/*
    Polygons are a collection of points connected with straight line segments.  It
    is automatically closed, you should not specify the same ending and starting point.
    A TGPolygon is area enclosing geometry that may be filled and/or framed.;
*/

public class TGPolygon extends TGPolyline implements Cloneable {

    private boolean fEOFill;

    // Sets number of points = zero
    public TGPolygon() {
        super();
        fEOFill = true;
    }
    public TGPolygon(int numPoints) {
        this(numPoints, true);
    }
    public TGPolygon(int numPoints, boolean EOFill)
    {
        super(numPoints);
        fEOFill = EOFill;
    }
    public TGPolygon(Polygon poly) {
        this(poly, true);
    }
    public TGPolygon(Polygon poly, boolean EOFill)
    {
        super(poly.npoints);
        int numPoints = poly.npoints;
        TGPoint temp = new TGPoint();
        for (int i = 0; i < numPoints; i++) {
            temp.x = (double)poly.xpoints[i];
            temp.y = (double)poly.ypoints[i];
            setPoint(i, temp);
        }
        fEOFill = EOFill;
    }
    public TGPolygon(TGRect rect) {
        this(rect, true);
    }
    public TGPolygon(TGRect rect, boolean EOFill)
    {
        super(4);
        setPoint(0, rect.getTopLeft());
        setPoint(1, rect.getTopRight());
        setPoint(2, rect.getBottomRight());
        setPoint(3, rect.getBottomLeft());
        fEOFill = EOFill;
    }
/*
    public TGPolygon(TGPointArray pts) {
        this(pts, true);
    }
    public TGPolygon(TGPointArray pts, boolean EOFill)
    {
        super(pts);
        fEOFill = EOFill;
    }
*/
    public TGPolygon(TGPolygon src) {
        super(src);
        fEOFill = src.fEOFill;
    }

    // operators
    public final TGPolygon copyFrom(TGPolygon poly){
        if (poly != this) {
            super.copyFrom(poly);
            fEOFill = poly.fEOFill;
        }
        return this;
    }
    public boolean equals(TGPolygon poly) {
        return (fEOFill == poly.fEOFill) && super.equals(poly);
    }

    // Data access
    public final boolean getEOFill() {
        return fEOFill;
    }

/*
    public final boolean isConvex() {
        int numPoints = getNumberOfPoints();
        boolean result = true;

        // has to have at least 3 points or its degenerate
        if (numPoints > 2)
        {
            TGPoint pts[] = fPoints.getArray();

            // first handle the boundary vertices that require
            // wrapping around fPoints

            int primaryOrientation =
                GetVertexOrientation(pts[numPoints-2], pts[numPoints-1], pts[0]);
            boolean isCollinear = (primaryOrientation == kDegenerate);
            int currentOrientation =
                GetVertexOrientation(pts[numPoints-1], pts[0], pts[1]);

            // if the first vertex was collinear and this one isn't
            // set up the primaryOrientation and turn the flag off

            if ((isCollinear) && (currentOrientation != kDegenerate)) {
                isCollinear = false;
                primaryOrientation = currentOrientation;
            }
            else { // if the currentOrientation is collinear then substitute
                // in the primaryOrientation since it is not
                if (currentOrientation == kDegenerate)
                    currentOrientation = primaryOrientation;
            }

            // at this point we have fully collinear points or
            // a mix of collinear, and/or orientations
            // so if they don't match that means that we have
            // a concave polygon so bypass the rest of the vertices

            if (currentOrientation == primaryOrientation) {
                numPoints -= 2;
                // loop over all the vertices up to the ones
                // covered by the above tests which required
                // wrapping around the point list
                for (int i = 0; (i < numPoints) && result; ++i) {
                    currentOrientation =
                        GetVertexOrientation(pts[i], pts[i+1], pts[i+2]);

                    if (currentOrientation != kDegenerate) {
                        // We have something other than collinear points.
                        // then turn the flag off and set the primaryOrientation
                        isCollinear = false;

                        // otherwise if the currentOrientation is not the
                        // same as the primaryOrientation we have a concave
                        // polygon so set result to false
                        if (currentOrientation != primaryOrientation)
                            result = false;

                        primaryOrientation = currentOrientation;
                    }
                }
                if (isCollinear) result = false;    // all points were collinear
            }
            else
                result = false;     // we have a concave polygon
                                    // because the first two vetices had
                                    // different directions
        }
        else
            result = false;     // we have a degenerate polygon
        return result;
    }
*/
    public final boolean isRectilinear() {
        TGPoint first = fPoints.getPoint(0);
        TGPoint last = fPoints.getPoint(getNumberOfPoints() - 1);
        return ((first.x == last.x) || (first.y == last.y)) && super.isRectilinear();
    }

    // Modifiers
    public final void setEOFill(boolean EOFill) {
        fEOFill = EOFill;
    }

    // Hit methods Implementation
    public final boolean intersects(TGRect g) {
        short fillCount = 0;
        int numPoints = getNumberOfPoints();
        double scanX;
        double x = g.x;
        double y = g.y;
        TGPoint pts[] = fPoints.getArray();
        TGPoint lastPt = pts[numPoints-1];
        TGPoint thisPt;
        for (int i = 0; i < numPoints; ++i) {
            thisPt = pts[i];
            TGLine line = new TGLine(lastPt, thisPt);
            if (line.intersects(g))
                return true;
            double dx = thisPt.x - lastPt.x;
            double dy = thisPt.y - lastPt.y;
            if (lastPt.y <= y) {
                if (thisPt.y > y) {
                    scanX = lastPt.x + (y - lastPt.y) * dx / dy;
                    if (scanX < x)
                        fillCount += 1;
                }
            }
            else { // lastPt.y > y
                if (thisPt.y <= y) {
                    scanX = lastPt.x + (y - lastPt.y) * dx / dy;
                    if (scanX < x)
                        fillCount -= 1;
                }
            }
            lastPt = thisPt;
        }
        if (fEOFill)
            return ((fillCount & 1) > 0);
        else
            return (fillCount > 0);
    }
    
    public String toString() {
        String returnValue = "TGPolygon : ";
        int  numPts = fPoints.getNumberOfPoints();
        if (numPts == 0)
            returnValue += "Zero Points";
        else {
            returnValue += "( " + numPts + " ) ";
            for (int i=0; i< numPts; i++)
                returnValue += " " + fPoints.getPoint(i);
        }

        return returnValue;
    }

    
    public final boolean contains(TGPoint p){
        if (isDegenerateToAPoint()) return false;
// System.out.println ("point is:" + p);        
// System.out.println ("polygon: " + this);

        short fillCount= 0;
        int numPoints = getNumberOfPoints();
        double scanX;
        double y = p.y;
        double x = p.x;
        TGPoint pts[] = fPoints.getArray();
        TGPoint lastPt = pts[numPoints-1];
        TGPoint thisPt;
        for (int i = 0; i < numPoints; i++) {
            thisPt = pts[i];
            if (p.equals(thisPt)) // if it is right on the vertex...
                return false;
            double dx = thisPt.x - lastPt.x;
            double dy = thisPt.y - lastPt.y;
            if (lastPt.y <= y) {
                if (thisPt.y > y) {
                    scanX = lastPt.x + (y - lastPt.y) * dx / dy;
                    if (scanX < x)
                        fillCount += 1;
                }
            }
            else { // lastPt.y > y
                if (thisPt.y <= y) {
                    scanX = lastPt.x + (y - lastPt.y) * dx / dy;
                    if (scanX < x)
                        fillCount -= 1;
                }
            }
            lastPt = thisPt;
        }


// System.out.println ("fEOFill = " + fEOFill);
// System.out.println ("fillCount = " + fillCount);

        if (fEOFill)
            return ((fillCount & 1) > 0);
        else
            return (fillCount != 0);

    }

/*
    // enum EVertexOrientation
    static final int kClockWise = 0;
    static final int kCounterClockwise = 1;
    static final int kDegenerate = 2;           // a line or a point

    static int GetVertexOrientation(            // static EVertexOrientation GetVertexOrientation(
        TGPoint a, TGPoint b, TGPoint c)
    {
        int result;  // EVertexOrientation result;
        double x1 = b.x - a.x;
        double y1 = b.y - a.y;
        double x2 = c.x - b.x;
        double y2 = c.y - b.y;
        double sign = x1 * y2 - y1 * x2;
        if (sign > 0.0)
            result = kClockWise;
        else if (sign < 0.0)
            result = kCounterClockwise;
        else result = kDegenerate;
        return result;
    }
*/
}
