/*
 * @(#)TGPolyline.java	1.8 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TGPolyline
||
*/

/*
    Polylines are a collection of points connected with straight line segments.  It
    is not explicitly closed (use TGPolygon for that; explictly closing a TGPolyline
    may give an unexpected "seam").;
*/

public class TGPolyline implements Cloneable {

    protected TGPointArray fPoints;

    // Initializes to zero points.
    public TGPolyline() {
        fPoints = new TGPointArray(0);
    }
    public TGPolyline(int numPoints) {
        fPoints = new TGPointArray(numPoints);
	// Initialize all points to (0.,0.). No need to do it again.
	// The constructor would have already done it.
    }
    public TGPolyline(TGPointArray points) {
        fPoints = new TGPointArray(points);
    }
    public TGPolyline(TGPolyline src) {
        this(src.fPoints);
    }

    // operators
    public TGPolyline copyFrom(TGPolyline src) {
        if (src != this)
            fPoints.copyFrom(src.fPoints);
        return this;
    }
    public boolean equals(TGPolyline src) {
        return (src == this) || (fPoints.equals(src.fPoints));
    }

    // Data access
    public int getNumberOfPoints() {
        return fPoints.getNumberOfPoints();
    }
    public TGPoint getPoint(int index){
        if ((index < 0.) || (index >= fPoints.getNumberOfPoints()))
            throw new IndexOutOfBoundsException("TGPolyline.getPoint(index)");
        return fPoints.getPoint(index);
    }
    public void getPoints(TGPointArray pts) {
        pts.copyFrom(fPoints);
    }
    TGPoint[] getPoints() {   // unsafe getPoints, for internal use only
        return fPoints.getArray();
    }

    // Returns a point on the polyline corresponding to u, with u
    // in the range 0 >= u >= # of Points.  The integer part of u determines
    // which segment of the polyline is referenced, and the fraction
    // determines where on that segment the point is.
    public TGPoint evaluate(double /*GParametric*/ u) {
        int maxIndex = fPoints.getNumberOfPoints() - 1;

        if ((u < 0.) || (u > (double)maxIndex))
            throw new IndexOutOfBoundsException("TGPolyline.evaluate(parametric)");

        int int_u = (int)u;

        if (int_u == maxIndex)
            return (TGPoint)fPoints.getPoint(maxIndex).clone();
        else {
            TGLine line =
                new TGLine(fPoints.getPoint(int_u), fPoints.getPoint(int_u + 1));
            return line.evaluate(u - (double)int_u);
        }
    }


    public boolean isDegenerateToAPoint() {
        boolean retval = true;
        int  numPts = fPoints.getNumberOfPoints();
        if (numPts > 0) {
            TGPoint  last = fPoints.getPoint (0);
            for (int i = 1; i < numPts; i++) {
                TGPoint  currentPt = fPoints.getPoint (i);
                if (! last.equals (currentPt)) {
                    retval = false;
                    break;
                }
            }
        }        
        return retval;
    }
    
    
    // Returns true if all the segments are horizontal or vertical.
    public boolean isRectilinear() {
        TGPoint point[] = fPoints.getArray();
        int numIntervals = fPoints.getNumberOfPoints() - 1;
        for (int i = 0; i < numIntervals; ++i)
        {
            TGPoint thisPoint = point[i+1];
            TGPoint lastPoint = point[i];
            if ((thisPoint.x != lastPoint.x) && (thisPoint.y != lastPoint.y))
                 return false;
        }
        return true;
    }

    // Modifiers
/*
    public void setPoints(TGPointArray newPoints) {
        fPoints.copyFrom(newPoints);
    }
*/
    public void setPoint(int index, TGPoint p){
        if (index < fPoints.getNumberOfPoints())
            fPoints.setPoint(index, p);
    }
    void setPoint(int index, double x, double y){
        if (index < fPoints.getNumberOfPoints())
            fPoints.setPoint(index, x, y);
    }
    public void addBefore(int index, TGPoint tobeadded) {
        int numPoints = fPoints.getNumberOfPoints();

        if ((index < 0) || (index > numPoints))
            throw new IndexOutOfBoundsException("TGPolyline.addBefore(index,...)");

        fPoints.resize(numPoints + 1);
        TGPoint point[] = fPoints.getArray();
        TGPoint temp = point[numPoints];            // save the last point added
        for (int i = numPoints; i > index; --i)
            point[i] = point[i - 1];
        temp.copyFrom(tobeadded);                    // copy the added into the saved
        point[index] = temp;
    }
    public void append(TGPoint p) {
        fPoints.append(p);
    }

    // Removes the point and resizes the data storage
    public void removePoint(int index){
        int numPoints = fPoints.getNumberOfPoints();

        if ((index < 0) || (index > numPoints))
            throw new IndexOutOfBoundsException("TGPolyline.removePoint(index)");

        --numPoints;
        TGPoint point[] = fPoints.getArray();
        for (int i = index; i < numPoints; ++i)
            point[i] = point[i + 1];
        fPoints.resize(numPoints);
    }

    public void reverseDirection() {
        int numPoints = fPoints.getNumberOfPoints();
        if (numPoints > 0) {
            TGPoint point[] = fPoints.getArray();
            for (int i = 0, j = numPoints - 1; i < j; ++i, --j)  {
                TGPoint temp = point[i];
                point[i] = point[j];
                point[j] = temp;
            }
        }
    }
    public void transformBy(TGrafMatrix transform) {
        transform.transformPoints(fPoints);
    }

/*
    // Moves the polyline so the point at the parametric position u
    // is moved to "toPoint"
    public void dragPosition(double parametric, TGPoint toPoint) {
        if ((parametric <= fPoints.getNumberOfPoints()) && (parametric >= 0.0))
        {
            int prev = (int)parametric; // lower boundary of edge
                                        // upper boundary of edge
            int next = ((prev >= fPoints.getNumberOfPoints() - 1) ? 0 : prev + 1);
            TGPoint point[] = fPoints.getArray();
            double weight1 = (double)((int)parametric - prev);
            double weight0 = (1.0 - weight1);
            // TGPoint delta = toPoint-(fPoints[prev]*weight0 + fPoints[next]*weight1);
            TGPoint prevP = TGPoint.multiply(point[prev], weight0);
            TGPoint nextP = TGPoint.multiply(point[next], weight1);
            prevP.addBy(nextP);
            nextP.copyFrom(toPoint);
            nextP.subtractBy(prevP);
            // now delta is nextP...
            double denom = weight0 * weight0 + weight1 * weight1;
            prevP.copyFrom(nextP);
            point[prev].addBy(prevP.multiplyBy(weight0/denom));
            point[next].addBy(nextP.multiplyBy(weight1/denom));
        }
    }
*/

/*
    // Find the parameter of the point on the polyline nearest p
    public double nearestParametric(TGPoint pt) {
        double nearestDist = Double.POSITIVE_INFINITY;
        double nearestParam = 0.0;

        int numPts = fPoints.getNumberOfPoints();
        TGPoint point[] = fPoints.getArray();

        for (int i = 0; i < numPts - 1; i++)
        {
            DoubleType dist = new DoubleType(0);
            double t = prStatic.PointSegDist(point[i], point[i + 1], pt, dist); // prSplineRoutines
            if (dist.fValue < nearestDist)
            {
                nearestDist = dist.fValue;
                nearestParam = t + (double)i;
            }
        }
        return nearestParam;
    }
*/

    public boolean intersects(TGRect g) {
        TGRect rectbounds = getBounds();
        // Does it intersect the bounding rectangle?
        if (g.intersects(rectbounds)) {
            int numPts = fPoints.getNumberOfPoints();
            TGPoint point[] = fPoints.getArray();
            for (int i = 1; i < numPts; i++) {
                TGLine line = new TGLine(point[i - 1], point[i]);
                if (line.intersects(g))
                    return true;
                }
            }
        return false;
    }
    public TGRect getBounds(){
        TGRect g = new TGRect();
        TGPoint point[] = fPoints.getArray();
        TGPoint ptr = point[0];
        g.setToPoint(ptr);
        int numPts = fPoints.getNumberOfPoints();
        for (int i = 1; i < numPts; i++) {
            g.extendTo(point[i]);
        }
        return g;
    }

    void resizePointArray(int newSize) {
        fPoints.resize(newSize);
    }

}
