/*
 * @(#)TGRPointArray.java	1.5 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
    TGRPointArray provides array access for collections of TGRPoints.  This
    [will be] used by TGCurve.  Eventually when we get
    templates it may become a subclass of a generic array class.;
*/

class TGRPointArray extends TRawArray {

    private TGRPoint[] fValues = null;

    private void cloneFromArray(int beginIndex, int count, TGRPoint src[]) {
        int end = beginIndex + count;
        for (int i = beginIndex; i < end; ++i)
            fValues[i] = (TGRPoint)src[i].clone();
    }

    public TGRPointArray() {
        this(0);
    }
    public TGRPointArray(int size) {
        super(size);
        if (fArraySize > 0) {
            allocateArray(fArraySize);
            newAllocation(0, fArraySize);
        }
    }
    public TGRPointArray(TGRPointArray source) {
        if (source.fNumberOfValues > 0) {
            allocateArray(source.fArraySize);
            cloneFromArray(0, source.fNumberOfValues, source.fValues);
            int remaining = source.fArraySize - source.fNumberOfValues;
            if (remaining > 0)
                newAllocation(source.fNumberOfValues, remaining);
        }
        super.copyFrom(source);
    }

    public final TGRPointArray copyFrom(TGRPointArray source) {
        if (fArraySize != source.fArraySize) {      // resize the array...
            TGRPoint oldValues[] = fValues;
            allocateArray(source.fArraySize);
            if (fArraySize > source.fArraySize)
                fArraySize = source.fArraySize;
            else    // allocate new storage if needed...
                newAllocation(fArraySize, source.fArraySize - fArraySize);
            if (fArraySize > 0)                     // reuse the old storage if any...
                System.arraycopy(oldValues, 0, fValues, 0, fArraySize);
        }
        for (int i = 0; i < source.fNumberOfValues; ++i)
            fValues[i].copyFrom(source.fValues[i]);
        super.copyFrom(source);
        return this;
    }

    public final boolean equals(TGRPointArray source) {
        if (source == this) return true;
        if (super.equals(source)) return false;
        for (int i = 0; i < fNumberOfValues; ++i) {
            if (!fValues[i].equals(source.fValues[i]))
                return false;
        }
        return true;
    }

    // access protocol...
    public TGRPoint getValue(int i) {
        if (i < 0 || i >= fNumberOfValues)
            throw new IndexOutOfBoundsException("RawArray.value(index)");

        return fValues[i];
    }
    public void setValue(int i, TGRPoint p) {
        if (i < 0 || i >= fNumberOfValues)
            throw new IndexOutOfBoundsException("RawArray.setValue(index, ...)");
        fValues[i].copyFrom(p);
    }
    public void append(TGRPoint p) {
        int newIndex = fNumberOfValues;
        resize(newIndex + 1); // This changes fNumberOfValues!
        fValues[newIndex].copyFrom(p);
    }
    public void append(TGPoint p) {
        int newIndex = fNumberOfValues;
        resize(newIndex + 1); // This changes fNumberOfValues!
        fValues[newIndex].fX = p.x;
        fValues[newIndex].fY = p.y;
        fValues[newIndex].fW = 1.;
    }

    // just like an array in C, there is NO BOUNDS CHECKING ... java throws an exception!
    public TGRPoint[] getArray() {
        return fValues;
    }

    void allocateArray(int size) {
        fValues = null;
        if (size > 0)
            fValues = new TGRPoint[size];
    }
    void reallocateArray(int newSize) {
        TGRPoint oldValues[] = fValues;
        allocateArray(newSize);
        if (newSize > 0)
            System.arraycopy(
                oldValues, 0,
                fValues, 0,
                fArraySize > newSize ? newSize : fArraySize);
    }
    void newAllocation(int beginIndex, int count) {
        int end = beginIndex + count;
        for (int i = beginIndex; i < end; ++i)
            fValues[i] = new TGRPoint();
    }

    // This swaps the contents of an array with victim.
    public void swap(TGRPointArray victim) {
        super.swap(victim);

        TGRPoint tmpData[] = fValues;
        fValues = victim.fValues;
        victim.fValues = tmpData;
    }

    // TGRPointArray member functions

    public TGRPointArray(TGPointArray source)  // Sets w's to 1.0
    {
        this(source.getNumberOfPoints());
        TGPoint src;
        TGRPoint dst;
        for (int i = 0; i < source.getNumberOfPoints(); i++) {
            src = source.getValue(i);
            dst = getValue(i);
            dst.fX = src.x;
            dst.fY = src.y;
            dst.fW = 1.0;
        }
    }

    // access
    public TGRPoint getPoint(int i) {
        return getValue(i);
    }
    public void setPoint(int i, TGRPoint p) {
        setValue(i, p);
    }
    // No more automatic cast...
    public void setPoint(int i, TGPoint p) {
        setPoint(i, p.x, p.y, 1.0);
    }

    public void setPoint(int i, double x, double y, double w) {
        if (i < 0 || i >= fNumberOfValues)
            throw new IndexOutOfBoundsException("RawArray.setValue(index, ...)");
        TGRPoint p = fValues[i];
        p.fX = x;
        p.fY = y;
        p.fW = w;
    }

    public int getNumberOfPoints() {
        return getNumberOfValues();
    }
    public TGRect getBounds() {
        int numPoints = getNumberOfValues();
        if (numPoints == 0) {
            return new TGRect(    // Why? Look at TGStatics.java
                              TGStatics.kZeroRectTopLeftX,
                              TGStatics.kZeroRectTopLeftY,
                              TGStatics.kZeroRectBottomRightX,
                              TGStatics.kZeroRectBottomRightY);
        }

        TGRect result = new TGRect();
        TGRPoint pts[] = getArray();
        result.setToPoint(pts[0].divW());
        for (int i = 1; i < numPoints; i++)
            result.extendTo(pts[i].divW());
        return result;
    }
}
