/*
 * @(#)TGRect.java	1.9 98/07/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;

public final class TGRect extends Rectangle2D.Double {

    private static final TGRect fkZeroRect =  // Why?  Look at TGStatics.java
        new TGRect(
                   TGStatics.kZeroRectTopLeftX,
                   TGStatics.kZeroRectTopLeftY,
                   TGStatics.kZeroRectBottomRightX,
                   TGStatics.kZeroRectBottomRightY);

    /************************************************************************
    private static final TGRect fkInfiniteRect =  // Why?  Look at TGStatics.java
        new TGRect(
                   TGStatics.kInfRectTopLeftX,
                   TGStatics.kInfRectTopLeftY,
                   TGStatics.kInfRectBottomRightX,
                   TGStatics.kInfRectBottomRightY);
    *************************************************************************/


    // Empty constructor may not initialization of data members
    public TGRect() {
        super();
    }

    public TGRect(double left, double top, double right, double bottom) {
        set(left, top, right, bottom);
    }

    public TGRect(TGPoint p) {
        this(p, p);
    }

    public TGRect(TGPoint p0, TGPoint p1) {
//System.out.println("TGRect(TGPoint,TGPoint):");
        set(p0, p1);
    }

    public TGRect(RectangularShape rs) {
//System.out.println("TGRect(RectangularShape):");
        set(rs.getMinX(), rs.getMinY(), rs.getMaxX(), rs.getMaxY());
    }

    public TGRect(Rectangle2D rect) {
//System.out.println("TGRect(Rectangle2D.Double):");
        set(rect.getMinX(), rect.getMinY(), rect.getMaxX(), rect.getMaxY());
    }

    public TGRect(TGRect src) {
        super.setRect(src);
    }

    public Object clone() {
        return new TGRect(this);
    }

    public String toString() {
        return ("(" + x + "," + y + "), width:" + width + ", height:" + height);
    }

    // Data access
    public TGPoint getTopLeft() {
        return new TGPoint(x, y);
    }
    public TGPoint getBottomRight() {
        return new TGPoint(x + width, y + height);
    }
    public TGPoint getTopRight() {
        return new TGPoint(x + width, y);
    }
    public TGPoint getBottomLeft() {
        return new TGPoint(x, y + height);
    }

/* removed to get rid of TGLine:

    public TGLine getTopLine() {
        return new TGLine(getTopLeft(), getTopRight());
    }
    public TGLine getBottomLine() {
        return new TGLine(getBottomLeft(), getBottomRight());
    }
    public TGLine getLeftLine() {
        return new TGLine(getTopLeft(), getBottomLeft());
    }
    public TGLine getRightLine() {
        return new TGLine(getTopRight(), getBottomRight());
    }
*/

    // u,v in the range 0..1 are bilinearly interpolated across the rectangle.
    public TGPoint interpolate(double u, double v) {
        return
            new TGPoint(
                (x * (1.0 - u)) + ((x + width) * u),
                (y * (1.0 - v)) + ((y + height) * v));
    }

    public TGPoint getSize() {
        return new TGPoint(width, height);
    }
    public TGPoint getCenter() {
        return new TGPoint(x + (width * 0.5), y + (height * 0.5));
    }
    public boolean equalSize(TGRect g) {
        return (width == g.width) && (height == g.height);
    }

    // Forces Point to lie in the rectangle, moving it to nearest edge if needed
    public void pin(TGPoint point) {
        double tmp;

        if (point.x < x)
            point.x = x;
        else {
          tmp = x + width; // right
          if (point.x > tmp)
              point.x = tmp;
        }

        if (point.y < y)
            point.y = y;
        else {
            tmp = y + height; // bottom
            if (point.y > tmp)
                point.y = tmp;
        }
    }

    // Modifiers
    public void setToPoint(TGPoint point) {
        x = point.x;
        y = point.y;
        width = height = 0.;
    }

    public void setTopLeft(TGPoint point) {
        x = point.x;
        y = point.y;
    }
    public void setBottomRight(TGPoint point) {
        width = point.x - x;
        height = point.y - y;
    }
    public void setTopRight(TGPoint point) {
        width = point.x - x;
        y = point.y;
    }
    public void setBottomLeft(TGPoint point) {
        x = point.x;
        height = point.y - y;
    }

    public void set(double left, double top, double right, double bottom) {
        // set the Rectangle2D.Double data elements ORDERED!
        if (left <= right) {
            x = left;
            width = right - left;
        }
        else {
            x = right;
            width = left - right;
        }
        if (top <= bottom) {
            y = top;
            height = bottom - top;
        }
        else {
            y = bottom;
            height = top - bottom;
        }
//System.out.println("TGRect:set(" + left + "," + top + "," + right + "," + bottom + ")");
//System.out.println(" -> " + this);
    }

    public void set(TGPoint a, TGPoint b) {
        set(a.x, a.y, b.x, b.y);
    }

    // move center of rectangle, maintaining size
    public void setCenter(TGPoint center) {
        TGPoint delta = TGPoint.subtract(center, getCenter());
        x += delta.x;
        y += delta.y;
    }

    // Change size of rectangle, maintaining center
    public void setSize(TGPoint size) {
        TGPoint newradius = TGPoint.subtract(size, getSize());
        width += newradius.x;
        height += newradius.y;

        newradius.divideBy(2.0);
        x -= newradius.x;
        y -= newradius.y;
    }

    // aka "Union"
    public void extendTo(TGRect g) {
        if (!isEmpty()) {
            boolean recalc = false;
            double tmp = x + width; // right
            if (g.x < x) {
                x = g.x;
                recalc |= true;
            }
            double other = g.x + g.width; // g.right
            if (other > tmp) {
                tmp = other;
                recalc |= true;
            }
            if (recalc)
                width = tmp - x;

            recalc = false;
            tmp = y + height; // bottom
            if (g.y < y) {
                y = g.y;
                recalc |= true;
            }
            other = g.y + g.height; // g.bottom
            if (other > tmp) {
                tmp = other;
                recalc |= true;
            }
            if (recalc)
                height = tmp - y;
        }
        else {
            x = g.x;
            y = g.y;
            width = g.width;
            height = g.height;
        }
    }
    // aka "Union"
    public void extendTo(TGPoint p) {
        extendTo(p.x, p.y);
    }
    public void extendTo(double px, double py) {
        boolean recalc = false;
        double tmp = x + width; // right
        if (px < x) {
            x = px;
            recalc |= true;
        }
        else if (px > tmp) {
            tmp = px;
            recalc |= true;
        }
        if (recalc)
            width = tmp - x;

        recalc = false;
        tmp = y + height; // bottom
        if (py < y) {
            y = py;
            recalc |= true;
        }
        else if (py > tmp) {
            tmp = py;
            recalc |= true;
        }
        if (recalc)
            height = tmp - y;
    }
    public void intersectWith(TGRect g) {
        if (intersects(g)) {
            boolean recalc = false;
            double tmp = x + width; // right
            if (g.x > x) {
                x = g.x;
                recalc |= true;
            }
            double other = g.x + g.width; // g.right
            if (other < tmp) {
                tmp = other;
                recalc |= true;
            }
            if (recalc)
                width = tmp - x;

            recalc = false;
            tmp = y + height; // bottom
            if (g.y > y) {
                y = g.y;
                recalc |= true;
            }
            other = g.y + g.height; // g.bottom
            if (other < tmp) {
                tmp = other;
                recalc |= true;
            }
            if (recalc)
                height = tmp - y;
        }
        else
            copyFrom(fkZeroRect);
    }
    public void offset(TGPoint dg) {
        x += dg.x;
        y += dg.y;
    }
    public void inset(TGPoint inset) {
        x += inset.x;
        y += inset.y;
        width -= (inset.x + inset.x);
        height -= (inset.y + inset.y);
    }

    public TGRect copyFrom(TGRect src) {
        x = src.x;
        y = src.y;
        width = src.width;
        height = src.height;
        return this;
    }

    public boolean equals(TGRect src) {
        return (x == src.x && y == src.y && width == src.width && height == src.height);
    }

}
