/*
 * @(#)TGStatics.java	1.1 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
 * Why was this class needed?
 *
 * Have to give you the background first.
 *
 * BACKGROUND DISCUSSION
 * ---------------------
 * 
 * Java lacks C++ style "const" declaration. 'final' applies only to
 * primitive data types and object references. There is no mechanism
 * other than class design which prevents one from modifying any
 * public static final object variables defined in a class.
 *
 *   public static final TGPoint  kOrigin = new TGPoint (0.0, 0.0);
 *
 * just prevents anyone from reassigning kOrigin to point to another
 * TGPoint instance. It does not prevent any one from changing the
 * kOrigin x and y values. This is possible as the containing classes
 * are mutable.
 *
 * We do not want anyone to modify the values that we have assigned to
 * these static variables.
 *
 * Variables of this kind within Albert are:
 *   TGPoint.kOrigin
 *   TGPoint.kInfinite
 *   TGrafMatrix.getIdentity() { return kgIdentity; }
 *   TGRect.kZeroRect
 *   TGRect.kInfiniteRect
 *
 * There are three ways this problem can be fixed.
 *
 *   1. Each of these variables are declared as private and getter's are
 *      provided which always return clones.
 *
 *   2. Every class using these variables define their own private scope copy.
 *
 *   3. Restrict these variables to have only 'package' scope.
 *
 *
 * The first approach has a big performance impact as every time a new
 * temporary copy is created which will eventually result in increase in
 * garbage collection time. These variables are used quite often within
 * Albert.
 *
 * The second solution results in increasing the class size, but does not
 * result in a performance hit.
 *
 * The third solution does not work because it does not prevent anyone from
 * maliciously declaring their file to be part of the Albert package and
 * getting access to these static variables and modifying their value.
 *
 *
 * Hence, We have chosen the second approach based on the following
 * assumptions:
 *
 *   - sun.awt.Albert package is for internal use only. None of the
 *     Albert classes will ever be called directly by end user.
 *
 *   - currently only java/awt/geom uses Albert package.
 *
 *
 * THE ANSWER
 * ----------
 *
 * Coming finally to the question why this file was needed.
 *
 * Based on the choice of the solution, we now have to define the constants
 * in multiple files within the Albert package. To ease the maintenance a
 * little, we define the values for individual data members here in this
 * file.
 * 
 */

public final class TGStatics {

    /*
     * Constants for defining TGPoint.fkOrigin
     */

    static final double kOriginX = 0.0;
    static final double kOriginY = 0.0;

    /*
     * Constants for defining TGPoint.fkInfinity
     */

    static final double kPosInfinityX = Double.POSITIVE_INFINITY;
    static final double kPosInfinityY = Double.POSITIVE_INFINITY;


    /*
     * Constants for defining TGRect.fkZeroRect
     */

    static final double kZeroRectTopLeftX       = 0.0;
    static final double kZeroRectTopLeftY       = 0.0;
    static final double kZeroRectBottomRightX   = 0.0;
    static final double kZeroRectBottomRightY   = 0.0;

    /*
     * Constants for defining TGRect.fkInfiniteRect
     */

    static final double kInfRectTopLeftX     = Double.NEGATIVE_INFINITY;
    static final double kInfRectTopLeftY     = Double.NEGATIVE_INFINITY;
    static final double kInfRectBottomRightX = Double.POSITIVE_INFINITY;
    static final double kInfRectBottomRightY = Double.POSITIVE_INFINITY;

} // TGStatics
