/*
 * @(#)TGrafExtractor.java	1.6 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TGrafExtractor
||
||  TGrafExtractor is the means by which geometry classes express their contents.
||  Their contents are expressed in terms of the basic set of geometries(see the
||  ExtractXXX calls below).
||
*/

abstract class TGrafExtractor {

    // === Area-defining geometries ===
    public abstract void extractRect(
                        TGRect g,
                        TCAGUnary node,
                        TGrafMatrix matrix);

    public void extractRect(
                        TGRect g,
                        TCAGUnary node) {    // default the matrix to fkIdentity
        extractRect(g, node, fkIdentity);
    }

    public abstract void extractEllipse(
                        TGEllipse g,
                        TCAGUnary node,
                        TGrafMatrix matrix);

    public void extractEllipse(
                        TGEllipse g,
                        TCAGUnary node) {
        extractEllipse(g, node, fkIdentity);
    }

    public abstract void extractPolygon(
                        TGPolygon g,
                        TCAGUnary node,
                        TGrafMatrix matrix);

    public void extractPolygon(
                        TGPolygon g,
                        TCAGUnary node) {
        extractPolygon(g, node, fkIdentity);
    }

    public abstract void extractLoop(
                        TGLoop g,
                        TCAGUnary node,
                        TGrafMatrix matrix);

    public void extractLoop(
                        TGLoop g,
                        TCAGUnary node) {
        extractLoop(g, node, fkIdentity);
    }

    // === render ===
    public abstract void render(TCAGRoot node);


    // === cAG node adoption function ===
    public final void adoptCAGNode(TCAGNode nodeToBeAdopted) {
        if (fNumCAGNodes == fNumAllocatedCAGNodes) {    // need more node space?
            int oldNumAllocatedCAGNodes = fNumAllocatedCAGNodes;
            fNumAllocatedCAGNodes += kCAGNodeIncrement;
            TCAGNode[] newCAGNodes = new TCAGNode[fNumAllocatedCAGNodes];
            if (fCAGNodes != null && oldNumAllocatedCAGNodes > 0)
                System.arraycopy((Object)fCAGNodes, 0, (Object)newCAGNodes, 0, oldNumAllocatedCAGNodes);
            fCAGNodes = newCAGNodes;
        }

        fCAGNodes[fNumCAGNodes] = nodeToBeAdopted;    // remember for later delete
        fNumCAGNodes++;
    }

    protected TGrafExtractor() {
        fCAGNodes = null;
        fNumCAGNodes = 0;
        fNumAllocatedCAGNodes = 0;
    }
    protected TGrafExtractor(TGrafExtractor that) {
        this();
    }

    protected TGrafExtractor copyFrom(TGrafExtractor that) {
        deleteCAGNodes();
        return this;
    }

    protected final void deleteCAGNodes() {
        // Delete all CAG nodes
        for (int i = 0; i < fNumCAGNodes; i++)
            fCAGNodes[i] = null;
        fNumCAGNodes = 0;
        // should we force the garbage collection here?
    }


    private static final TGrafMatrix fkIdentity = // Why? Look at TGStatics.java
        new TGrafMatrix();

    private TCAGNode[] fCAGNodes;    // list of cAG nodes
    private int fNumCAGNodes;    // number of nodes in the list
    private int fNumAllocatedCAGNodes;    // number of nodes allocated
    private static final int kCAGNodeIncrement = 16;    // resize list by this amount
}
