/*
 * @(#)TIntersectGeometry.java	1.5 98/04/02
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TIntersectGeometry
||
||  This class owns two geometries whose areas are logically intersected.
||
*/

class TIntersectGeometry extends MAreaGeometry {

    public TIntersectGeometry() {
        fBounds = null;
    }
    public TIntersectGeometry(
        MAreaGeometry left,
        MAreaGeometry right) {

        fLeft = left;
        fRight = right;
        fBounds = null;
        updateTimeStamp();
    }
    public TIntersectGeometry(TIntersectGeometry source) {
        super(source);

        fLeft = null;
        fRight = null;
        fBounds = null;

        if (source.fLeft != null) {
            fLeft = (MAreaGeometry) ( source.fLeft.clone() );
        }
        if (source.fRight != null) {
            fRight = (MAreaGeometry) ( source.fRight.clone() );
        }
        if (source.fBounds != null) {
            fBounds = (TGRect) ( source.fBounds.clone() );
        }
    }

    // === MAreaGeometry overrides ===
    public MAreaGeometry cloneAndIntersect(MAreaGeometry g) {
        TIntersectGeometry newGeometry = new TIntersectGeometry(this);
        newGeometry.fLeft = MAreaGeometry.intersect(newGeometry.fLeft, g);
        newGeometry.fBounds = null;
        newGeometry.updateTimeStamp();
        return newGeometry;
    }

    public void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix) {

        TCAGIntersect node = new TCAGIntersect(parentNode);
        extractor.adoptCAGNode(node);
        fLeft.extract(extractor, node, matrix);
        fRight.extract(extractor, node.getRightLink(), matrix);
    }

    public boolean isEmpty() {
        return (getBounds().isEmpty()) || (fLeft.isEmpty() && fRight.isEmpty());
    }
    public boolean isPolygonal() {
        return fLeft.isPolygonal() && fRight.isPolygonal();
    }
    public boolean isSingular() {
        return false;
    }

    public TGRect getBounds() {
        if (fBounds == null) {
            fBounds = new TGRect(fLeft.getBounds());
            fBounds.intersectWith(fRight.getBounds());
        }
        return fBounds;
    }
    public boolean intersects(TGRect rect) {
        return fLeft.intersects(rect) && fRight.intersects(rect); // *** Not fully implemented
    }
    public boolean contains(TGPoint point) {
        return getBounds().contains(point) && (fLeft.contains(point) && fRight.contains(point));
    }
    public boolean contains(TGRect rect) {
        return fLeft.contains(rect) && fRight.contains(rect);
    }

    public final TIntersectGeometry copyFrom(TIntersectGeometry source) {
        if (this != source) {
            super.copyFrom(source);
            fLeft = source.fLeft;
            fRight = source.fRight;
            fBounds = source.fBounds;
        }
        return this;
    }

    public Object clone() {
        return (Object)(new TIntersectGeometry(this));
    }
    public boolean equals(MAreaGeometry that) {
        if ((getClass().equals(that.getClass()))) {
            TIntersectGeometry geometry = (TIntersectGeometry)that;
            return (super.equals(that) || ((fLeft == geometry.fLeft) && (fRight == geometry.fRight)));
        }
        else
            return false;
    }

    private MAreaGeometry fLeft;
    private MAreaGeometry fRight;
    private TGRect fBounds;
}
