/*
 * @(#)TLoopGeometry.java	1.8 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
 *
 *  TLoopGeometry
 *
 */

public class TLoopGeometry extends MAreaGeometry {

    public TGLoop fLoop;

    public TLoopGeometry() {
        fBounds = null;
    }

    public TLoopGeometry(TGLoop loop) {
        fLoop = (TGLoop)loop.clone();
        fBounds = null;
        updateTimeStamp();
    }

    public TLoopGeometry(TLoopGeometry source) {
        super(source);
        fLoop = null;
        fBounds = null;

        if (source.fLoop != null) {
            fLoop = (TGLoop) ( source.fLoop.clone() );
        }
        if (source.fBounds != null) {
            fBounds = (TGRect) ( source.fBounds.clone() );
        }
    }

    // === MAreaGeometry overrides ===
    public MAreaGeometry cloneAndTransform(TGrafMatrix m) {
        if (fLoop.getNumberOfPoints() > kLargeLoop)
            return super.cloneAndTransform(m);
        else {
            // If I'm shared, return a new object; else return myself
            TLoopGeometry loop = new TLoopGeometry(this);
            loop.fLoop.transformBy(m);
            loop.fBounds = null;
            loop.updateTimeStamp();
            return loop;
        }
    }

    public void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix) {

        boolean eoFill = fLoop.getEOFill();
        TCAGUnary node;
        if (eoFill)
            node = new TCAGEvenOdd(parentNode);
        else
            node = new TCAGWindNum(parentNode);

        extractor.adoptCAGNode(node);
        extractor.extractLoop(fLoop, node, matrix);
    }

    public boolean isEmpty() {
	// should probably be more precise.
        return ((fLoop.getOrder() == 0) || (fLoop.getNumberOfPoints() < 3));
    }

    public boolean isPolygonal() {
        return ((fLoop.getOrder() == 2) && (fLoop.getNumberOfPoints() >= 3));
    }

    public TGRect getBounds() {
        if (fBounds == null)
            fBounds = fLoop.getBounds();

        return fBounds;
    }

    public boolean intersects(TGRect rect) {
	// 'intersects' means 'overlaps' in the Java 2D API.
	// 'intersects' means 'boundary intersects' in the Albert package.
	// So we implement the 2D API's definition using Albert's definition.
	//
	// A rectangle overlaps a loop (can have c0 discontinuities) if
	//   the rectangle contains any point of the loop, or
	//   the loop contains any point of the rectangle (say topLeft), or
	//   the loop boundaries intersect with the rectangle boundaries.

        if (fLoop.getNumberOfPoints() > 0) {
	    if (rect.contains(fLoop.getPoint(0).divW())) return true;
	    else if (fLoop.contains(rect.getTopLeft())) return true;
	    else if (fLoop.intersects(rect)) return true;
	}
        return false;
    }

    public boolean contains(TGPoint point) {
        return fLoop.contains(point);
    }

    public boolean contains(TGRect rect) {
        // return (fLoop.contains(rect.getTopLeft()) && !fLoop.intersects(rect));
        return (fLoop.contains (rect));
    }

    // === operators ===
    public final TLoopGeometry copyFrom(TLoopGeometry source) {
        if (this != source) {
            super.copyFrom(source);
            fLoop = source.fLoop;
            fBounds = source.fBounds;
        }
        return this;
    }

    public Object clone() {
        return (Object)(new TLoopGeometry(this));
    }

    public boolean equals(MAreaGeometry that) {
        if ((getClass().equals(that.getClass()))) {
            return (super.equals(that) &&
		    fLoop.equals(((TLoopGeometry)that).fLoop));
        }
        else
            return false;
    }

    private TGRect fBounds;    // cached bounds
    private static final int kLargeLoop = 16; // number of points in a large loop
}
