/*
 * @(#)TOutlineMakerVertexEngine.java	1.5 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||    TOutlineMakerVertexEngine
||
||    The TOutlineMakerVertexEngine finds the perimeter of the polygonal geometry
||    described by the vertices passed in.  It trims away all inconsequential
||    edges and sends the results to the output vertex engine.
||    For example, the star polygon, using winding number fill mode, would be
||    converted to the ten vertices that describe its outline.
||
||    This class performs its task only for closed geometries.  Open geometries
||    are passed through unmodified.
||
*/

public class TOutlineMakerVertexEngine { //extends TVertexEngine {

    private TPtrArrayOfCAGPointArray fVertexArrays = new TPtrArrayOfCAGPointArray(); // list of vertex arrays
    private TCAGPointArray fCurrentArray; // the array new points go into

    // always closed
    //private boolean fClosed;              // closed loop
    private TCAGUnary fCAGNode;           // the current CAG node

    //private TVertexEngine fVertexEngine;  // the output vertex engine
    private TGLoop fLoop;     // referenced
    private boolean fInitialised = false;
    private TGRPointArray fPoints = new TGRPointArray();
    private TGParametricArray fKnots = new TGParametricArray();
    private double fK = 1.;
    private int fStart = 0;

    private final void addKnot(double discontinuity) {
        int num = fKnots.getNumberOfValues();
        double k = fK;
        fK = discontinuity;
        if (k != 0. || fK != 1.)    // previous one is not a break...
        {
            if (num == 0) {
                fKnots.append(0.);
                fKnots.append(0.);
            }
            else {
                double value = fKnots.getValue(num - 1) + fK;
                fKnots.append(value);
            }
        }
    }

    //private final void outSetCAGNode(TCAGUnary node) {}
    //private final void outSetClosed(boolean closed) {}
/*
    private final void outReset() {
        fPoints.resize(0);
        fKnots.resize(0);
        fK = 1.;
        fStart = 0;
        fInitialised = false;
    }
*/

    private final void outAddVertex(TGPoint vertex) {
        addKnot(1.);
        fPoints.append(vertex);
    }

    private final void outEnd() {
        int last = fPoints.getNumberOfValues();
        if (last - fStart <= 2) {   // eliminate 2 point loops...
            fPoints.resize(last - 2);
            fKnots.resize(fKnots.getNumberOfValues() - 2);
        }
        else {
            addKnot(1.);
            TGRPoint starting = fPoints.getValue(fStart);
            fPoints.append(starting);
            fStart = fPoints.getNumberOfValues();
        }
        addKnot(0.);    // add a break ...
    }

    private final void outRender(TCAGRoot root) {
        if (fPoints.getNumberOfValues() > 0 && fKnots.getNumberOfValues() > 0) {
            TGLoop newLoop = new TGLoop(2, fPoints, fKnots);
            if (!fInitialised) {
                fLoop.copyFrom(newLoop);
                fInitialised = true;
            }
            else    // concatenate the new loop to the previous one...
                fLoop.concatenate(newLoop, false);
        }
        fPoints.resize(0);
        fKnots.resize(0);
    }


    public TOutlineMakerVertexEngine(TGLoop resultantLoop) {
        fLoop = resultantLoop;  // loop to be modified to capture the vertices.

        //fVertexEngine = outputVertexEngine;
        fCurrentArray = null;
        //fClosed = false;
    }
/*
    public TOutlineMakerVertexEngine() {
        this((TVertexEngine)null);
    }
*/

    // === TVertexEngine overrides ===
    public void SetCAGNode(TCAGUnary node) {
        fCAGNode = node;
    }
/*
    public void SetClosed(boolean closed) {
        fClosed = closed;
    }
*/
    public void AddVertex(TGPoint vertex) {
        boolean redundant = false;
        if (fCurrentArray == null) {
            fCurrentArray = new TCAGPointArray(fCAGNode);
            fVertexArrays.append(fCurrentArray);
        }
        else
            redundant = vertex.equals(fCurrentArray.fPoints.getValue(fCurrentArray.fPoints.getNumberOfPoints() - 1));

        if (!redundant)        // store it only if it is different from the previous point
            fCurrentArray.fPoints.append(vertex);
    }
    public void End() {
        if (fCurrentArray != null) {
            int lastP = fCurrentArray.fPoints.getNumberOfPoints() - 1;
            TGPoint vertex = fCurrentArray.fPoints.getValue(0);
            if (vertex.equals(fCurrentArray.fPoints.getValue(lastP)))
                 fCurrentArray.fPoints.resize(lastP);

            fCurrentArray = null;
        }
    }
    public void Reset() {
        fVertexArrays.resize(0);
        fCurrentArray = null;
        //fClosed = false;
    }

    public void Render(TCAGRoot root) {
        long numInputArrays = fVertexArrays.getNumberOfValues();
        if (numInputArrays <= 0)
            return;

/*
        if (!fClosed) {
            SendInputVerticesToOutput(root);
            Reset();
            return;
        }
*/

        // for closed geometry...

        // ======================================================
        // Count the number of edges, based on the input vertices
        // ======================================================
        int numEdges = 0;
//System.out.println("OMVertextEngine::Render(): Number of InputArrays = " + numInputArrays);
        {   // s1 - counting number of edges
        for (int i = 0; i < numInputArrays; i++) {
            int numVertices = fVertexArrays.getValue(i).fPoints.getNumberOfPoints();
//System.out.println("Array[" + i + "]: Number of Vertices = " + numVertices);for (int j=0; j<numVertices; ++j) System.out.print(" " + fVertexArrays.getValue(i).fPoints.getValue(j));System.out.println();
            if (numVertices > 1)    // zero or one vertex means zero edges; two (or more)means two (or more)edges
                numEdges += numVertices;
        }
        }   // s1

        if (numEdges <= 0) {
            Reset();
            return;
        }

        double gridUnit = 0;    // amount to grid vertices to
        boolean done = false;
        while (!done) {    // keep trying until the algorithm succeeds
            try {
                // ===================================================================
                // All edge comparisons and edge intersection calculations are done
                // using the same formula, which is encapsulated in a single object.
                // The algorithm used in this object can be optimized as needed.
                // ===================================================================
                TEdgeComparator edgeComparator = new TEdgeComparator(numEdges);

                // ===================================================================
                // Make the edges, the edge pointer list, and the initial y event list
                // ===================================================================
                TPolyEdge edges[] = new TPolyEdge[numEdges];
                for (int i = 0; i < numEdges; ++i)
                    edges[i] = new TPolyEdge();

                TPolyEdge edgePtrs[] = new TPolyEdge[numEdges];
                TArrayOfPolyYEvent yEvents = new TArrayOfPolyYEvent(numEdges);    // use TRawArray here because this list grows
                {   // s2 - making the edge pointer list and the y event list
                TPolyYEvent yEventsPtr[] = yEvents.getArray();
                int yEventIdx = 0;
                int edgeIndex = 0;

                for (int i = 0; i < numInputArrays; i++) {
                    TCAGPointArray array = fVertexArrays.getValue(i);
                    int numVertices = array.fPoints.getNumberOfPoints();
                    if (numVertices > 1) {
                        int prevVertex = numVertices - 1;
                        int vertex = 0;
                        TGPoint p[] = array.fPoints.getArray();
                        TPolyEdge firstEdgePtr = edges[edgeIndex];

                        for (; vertex < numVertices; prevVertex = vertex, vertex++, edgeIndex++) {
                            edges[edgeIndex].Set(p[prevVertex], p[vertex], array.fCAGNode, edgeIndex);
                            edgePtrs[edgeIndex] = edges[edgeIndex];
                            if (vertex != 0) {  // if not the first vertex, set the yEvents ...
                                yEventsPtr[yEventIdx].Set(edges[edgeIndex - 1], edges[edgeIndex], p[prevVertex]);
                                yEventIdx++;
                            }
                        }
                        yEventsPtr[yEventIdx].Set(edges[edgeIndex - 1], firstEdgePtr, p[prevVertex]);
                        yEventIdx++;
                    }
                }
                }   // s2
                TPolyEdge edgePtrs2[] = new TPolyEdge[numEdges];    // used later
                {   // s3 - sorting the edge list

                // Sort the edge pointer list
                TPolyEdgeQSort.YSortEdges(edgePtrs, numEdges);

                // Copy the edge pointer list to a second edge pointer list to be used later
                System.arraycopy(edgePtrs, 0, edgePtrs2, 0, numEdges);
                }   // s3

                // ===========================================================
                // Make the rest of the y event list (from edge intersections).
                //
                // This is done by traversing the edge list from top to bottom (similar to scan conversion).
                // At every known y event (i.e., the y values of every vertex), intersection calculations are performed
                // between all the active edges and the newly added active edges (thus minimizing the number of
                // intersection calculations that are done).  For each intersection that is found, a new y event object
                // is added to the y event list.
                // ===========================================================
                {   // s4 - making the rest of y event list
                int firstEdge = 0;            // first edge in the active edge list
                int lastEdge = 0;            // end of the active edge list
                int endEdge = numEdges;        // end of the entire edge list

                boolean loop = false;
                double y = edgePtrs[firstEdge].fTop.y;
                TGPoint point0 = new TGPoint();
                TGPoint point1 = new TGPoint();

                while (!loop) {
                    // Add new edges to the active edge list
                    int newEdges = lastEdge;
                    while ((lastEdge < endEdge) && (y >= edgePtrs[lastEdge].fTop.y))
                        lastEdge++;

                    int thisEdgePtrPtr = firstEdge;
                    for (; thisEdgePtrPtr < lastEdge; thisEdgePtrPtr++) {
                        // every active edge
                        // Adjust start of new edges so they do not get compared with each other twice
                        if (newEdges == thisEdgePtrPtr)
                            newEdges++;

                        int thatEdgePtrPtr = newEdges;
                        for (; thatEdgePtrPtr < lastEdge; thatEdgePtrPtr++) {
                            // every newly added active edge
                            TPolyEdge thisEdge = edgePtrs[thisEdgePtrPtr];
                            TPolyEdge thatEdge = edgePtrs[thatEdgePtrPtr];

                            // If the edges share a vertex, then they have already been added to the y event list.
                            if ((thisEdge.fTop != thatEdge.fTop) &&
                                (thisEdge.fBottom != thatEdge.fBottom) &&
                                (thisEdge.fBottom != thatEdge.fTop) &&
                                (thisEdge.fTop != thatEdge.fBottom)) {

//System.out.println("Checking for intersection between:\n" + thisEdge + " and\n" + thatEdge);
                                int numIntersections =
                                    edgeComparator.FindIntersection(thisEdge, thatEdge, point0, point1);
                                if (numIntersections > 0) {
//System.out.println("Intersection detected at " + point0 + "while adding " + thisEdge + " and\n" + thatEdge);
                                    yEvents.append(new TPolyYEvent(thisEdge, thatEdge, point0));
                                    if (numIntersections > 1) {
//System.out.println("Intersection actually colinear also at " + point1);
                                        yEvents.append(new TPolyYEvent(thisEdge, thatEdge, point1));
                                    }
                                }
                            }
                        }
                    }

                    if (lastEdge < endEdge) {
                        y = edgePtrs[lastEdge].fTop.y;

                        // Remove finished edges
                        for (thisEdgePtrPtr = firstEdge; thisEdgePtrPtr < lastEdge; thisEdgePtrPtr++) {
                            if (y >= edgePtrs[thisEdgePtrPtr].fBottom.y) {
                                // Is this edge done?
                                for (int ptr = thisEdgePtrPtr; ptr > firstEdge; ptr--)
                                    edgePtrs[ptr] = edgePtrs[ptr-1];
                                firstEdge++;
                            }
                        }
                    }
                    else
                        loop = true;
                }
                }   // s4

                // =========================================
                // Make the y event pointer list and sort it
                // =========================================
                int numYEvents = yEvents.getNumberOfValues();
                TPolyYEvent yEventPtrs[] = new TPolyYEvent[numYEvents];
                {   // s5 - making y event pointer list and sort it
                System.arraycopy(yEvents.getArray(), 0, yEventPtrs, 0, numYEvents);

                TPolyYEventQSort.SortYEvents(yEventPtrs, numYEvents);
//System.out.println("after:\n"); for (int i = 0; i < numYEvents; i++) System.out.println(yEventPtrs[i]);
                }   // s5

                // =====================================
                // Finally, generate the output vertices
                // =====================================
                TPtrArrayOfOutputList outputLists = new TPtrArrayOfOutputList();    // the output vertex lists
                {   // s6 - generating the output vertices
                // the edge/transition storage arrays
                TArrayOfEdgeTransition transitionList0 = new TArrayOfEdgeTransition();
                TArrayOfEdgeTransition transitionList1 = new TArrayOfEdgeTransition();

                int firstEdge = 0;            // first edge in the active edge list edgePtrs2
                int lastEdge = 0;            // end of the active edge list
                int endEdge = numEdges;        // end of the entire edge list

                int thisEvent = 0;
                int endEvent = numYEvents;
                while (thisEvent < endEvent) {
                    TPolyYEvent thisEventItself = yEventPtrs[thisEvent];
                    TGPoint eventPoint = yEventPtrs[thisEvent].fPoint;
                    double y = eventPoint.y;

                    int endRelevantEvent = thisEvent;     // end of list of relevant events
                    do {    // mark all edges that are relevant to all y-events at this point
                        yEventPtrs[endRelevantEvent].fEdge0.fRelevant = true;
                        yEventPtrs[endRelevantEvent].fEdge1.fRelevant = true;
                        endRelevantEvent++;
                    } while ((endRelevantEvent < endEvent) && yEventPtrs[endRelevantEvent].fPoint.equals(eventPoint));

                    int thisEdge;    // for general use in edge loops

                    // Remove finished edges from the active edge list
                    for (thisEdge = firstEdge; thisEdge < lastEdge; thisEdge++) {
                        // An edge is done if it is relevant and its bottom point is the current event point
                        if (edgePtrs2[thisEdge].fRelevant && edgePtrs2[thisEdge].fBottom.equals(eventPoint))
                            firstEdge = TPolyEdge.MoveEdge(edgePtrs2, thisEdge, firstEdge);
                    }

                    // Move the relevant edges into the active edge list
                    for (thisEdge = lastEdge; (thisEdge < endEdge) && (edgePtrs2[thisEdge].fTop.y == y); thisEdge++) {
                        if (edgePtrs2[thisEdge].fRelevant)
                            lastEdge = TPolyEdge.MoveEdge(edgePtrs2, thisEdge, lastEdge);
                    }

                    // Sort the relevant edges
                    edgeComparator.SetPoint(eventPoint);
                    for (thisEdge = firstEdge; thisEdge < lastEdge; thisEdge++) {
                        TPolyEdge thisEdgePtr = edgePtrs2[thisEdge];
                        if (thisEdgePtr.fRelevant) {
                            int ptr = thisEdge - 1;
                            // Sort horizontally:  Bubble the current edge down until it is in the right location
                            for (; (ptr >= firstEdge) && edgeComparator.LessThan(thisEdgePtr, edgePtrs2[ptr]); ptr--)
                                edgePtrs2[ptr + 1] = edgePtrs2[ptr];
                            edgePtrs2[ptr + 1] = thisEdgePtr;
                        }
                    }

                    // Generate edge/transition list of the current y-event
                    {   // s7 - generate edges and transition list of the current y-event
                    transitionList1.resize(0);    // clear it first
                    boolean state = false;
                    for (thisEdge = firstEdge; thisEdge < lastEdge; thisEdge++) {        // every active edge
                        edgePtrs2[thisEdge].Transition();
                        boolean newState = (root.getFillCount() != 0);
                        transitionList1.append(new TEdgeTransition(edgePtrs2[thisEdge], (state != newState)));
                        state = newState;
                    }
                    }   // s7
                    // Traverse the edges that are relevant to the current event point, processing the output lists accordingly.
                    // This is accomplished by circling the event point counter-clockwise from the top right to the bottom right
                    // (any direction from any starting point would work).  All edges are crossed, but only the relevant ones are
                    // examined.  For every relevant edge pair that makes an inside/outside transition, a stitching job is done
                    // to the output vertex list(s)of the edges.
                    {   // s8 - traverse the relavent edges ...
                    boolean state = false;                        // the current inside/outside state of the geometry
                    boolean firstCrossedRelevantEdge = false;    // whether the edge is the first of the crossed-relevant-edge-pair
                    TPolyEdge firstCrossedEdge = null;          // first edge in the crossed-relevant-edge-pair
                    boolean firstCrossedEdgeAbove = false;        // true if firstCrossedEdge is above the current event point

                    int indexEnd = transitionList1.getNumberOfValues();
//System.out.println("transitionList:\n"); for (int z = 0; z < indexEnd; z++) System.out.println(" (" + z + "):" + transitionList1.getValue(z));
                    for (int index = -transitionList0.getNumberOfValues(); index < indexEnd; index++) {
                        TEdgeTransition edgeTrans =
                            (index < 0) ? transitionList0.getValue(-index - 1) : transitionList1.getValue(index);
                        if (!edgeTrans.fTransition)     // is there a transition at this edge?
                            continue;

                        state = !state;

                        if (!edgeTrans.fEdge.fRelevant)  // is this edge relevant to the current y-event?
                            continue;

                        firstCrossedRelevantEdge = !firstCrossedRelevantEdge;

                        if (firstCrossedRelevantEdge) {    // first edge of a crossed-relevant-edge-pair?
                            firstCrossedEdge = edgeTrans.fEdge;    // store this edge for future use
                            firstCrossedEdgeAbove = (index < 0);
                            continue;
                        }

                        // second edge of a crossed-relevant-edge-pair; do the stitchwork
                        TPolyEdge secondCrossedEdge = edgeTrans.fEdge;

                        // Check for the three cases: both edges above, first above and second below, or both edges below
                        if (firstCrossedEdgeAbove) {
                            if (index < 0) {
                                // Both edges are above the event point, so they need to be ended and stitched together
                                if ((firstCrossedEdge.fOutputList == null) ||
                                    (secondCrossedEdge.fOutputList == null))
                                    throw new TGrafExtractorException();    // something got out of whack

                                if (state) {    // check the direction of the edges
//System.out.println("Append Vertex - 1:\n current:" + firstCrossedEdge.fOutputList + "\n " + thisEventItself + "\n input:\n" + eventPoint);
                                    firstCrossedEdge.fOutputList.fPoints.append(eventPoint);
//System.out.println(" after:" + firstCrossedEdge.fOutputList);
                                    secondCrossedEdge.fOutputList.fNextEdgeList =
                                        firstCrossedEdge.fOutputList;
                                }
                                else {
//System.out.println("Append Vertex - 2:\n current:" + secondCrossedEdge.fOutputList + "\n " + thisEventItself + "\n input:" + eventPoint);
                                    secondCrossedEdge.fOutputList.fPoints.append(eventPoint);
//System.out.println(" after:" + secondCrossedEdge.fOutputList);
                                    firstCrossedEdge.fOutputList.fNextEdgeList =
                                        secondCrossedEdge.fOutputList;
                                }

                                firstCrossedEdge.fOutputList = null;
                                secondCrossedEdge.fOutputList = null;
                            }
                            else {
                                // The first edge is above and the second edge is below the event point, so the
                                // output list of the first edge is continued by the second edge

                                if (firstCrossedEdge != secondCrossedEdge) {    // nothing to do if it is the same edge
                                    if ((firstCrossedEdge.fOutputList == null) ||
                                        (secondCrossedEdge.fOutputList != null))
                                        throw new TGrafExtractorException();    // something got out of whack

//System.out.println("Append Vertex - 3:\n current:" + firstCrossedEdge.fOutputList + "\n " + thisEventItself + "\n input:" + eventPoint);
                                    firstCrossedEdge.fOutputList.fPoints.append(eventPoint);
//System.out.println(" after:" + firstCrossedEdge.fOutputList);
                                    secondCrossedEdge.fOutputList = firstCrossedEdge.fOutputList;

                                    firstCrossedEdge.fOutputList = null;
                                }
                            }
                        }
                        else {
                            // Both edges are below the current event point, so two new output lists are created

                            if ((firstCrossedEdge.fOutputList != null) ||
                                (secondCrossedEdge.fOutputList != null))
                                throw new TGrafExtractorException();    // something got out of whack

                            TOutputVertexList edgeList0 = new TOutputVertexList(false);    // backward
                            TOutputVertexList edgeList1 = new TOutputVertexList(true);    // forward
                            outputLists.append(edgeList0);
                            outputLists.append(edgeList1);

//System.out.println("Append Vertex - 4:\n current:" + edgeList1 + "\n " + thisEventItself + "\n input:" + eventPoint);
                            edgeList1.fPoints.append(eventPoint);
//System.out.println(" after:" + edgeList1);
                            edgeList0.fNextEdgeList = edgeList1;

                            if (state) {    // check the direction of the edges
                                secondCrossedEdge.fOutputList = edgeList0;
                                firstCrossedEdge.fOutputList = edgeList1;
                            }
                            else {
                                firstCrossedEdge.fOutputList = edgeList0;
                                secondCrossedEdge.fOutputList = edgeList1;
                            }
                        }
                    }
                    }   // s8

                    do {    // unmark all edges that were marked relevant to this y-event
                        yEventPtrs[thisEvent].fEdge0.fRelevant = false;
                        yEventPtrs[thisEvent].fEdge1.fRelevant = false;
                        thisEvent++;
                    } while (thisEvent < endRelevantEvent);

                    // Swap the two edge transition list pointers
                    TArrayOfEdgeTransition tempList = transitionList0;
                    transitionList0 = transitionList1;
                    transitionList1 = tempList;
                }
                }   // s6

//System.out.println("+++ Done +++");
                // ==========================
                // Check for messed up output
                // ==========================
                {   // s9 - check for messed up output
                int numEdgeLists = outputLists.getNumberOfValues();
                TOutputVertexList list[] = outputLists.getArray();
                for (int i = 0; i < numEdgeLists; i++) {
                    if (list[i].fNextEdgeList == null)
                        throw new TGrafExtractorException();    // something got out of whack
                }
                }   // s9

                // =============================================
                // Send the output vertices to the vertex engine
                // =============================================
                {   // s10 - send output vertices to vertex engine
                //outSetClosed(true);
                TCAGRoot newRoot = new TCAGRoot();
                TCAGEvenOdd evenOdd = new TCAGEvenOdd(newRoot);
                //outSetCAGNode(evenOdd);

                int numEdgeLists = outputLists.getNumberOfValues();
//System.out.println(" Begin : num list = " + numEdgeLists);
                TOutputVertexList list[] = outputLists.getArray();
                for (int i = 0; i < numEdgeLists; i++) {
//System.out.println("output list["+i+"]:"+list[i]);
                    if (!list[i].fProcessed) {
                        TOutputVertexList outPtr = list[i];
                        for (; !outPtr.fProcessed; outPtr = outPtr.fNextEdgeList) {
                            outPtr.fProcessed = true;
                            TGPoint array[] = outPtr.fPoints.getArray();
                            int numVertices = outPtr.fPoints.getNumberOfPoints();
                            if (outPtr.fForward) {
                                for (int j = 0; j < numVertices; j++) {
                                    outAddVertex(array[j]);
                                }
                            }
                            else {    // backward
                                for (int j = numVertices; --j >= 0;) {
                                    outAddVertex(array[j]);
                                }
                            }
                        }
                        outEnd();
//System.out.println(" Break!");
                    }
                }
//System.out.println(" End");
                outRender(newRoot);
                }   // s10

                done = true;
            }
            catch (TGrafExtractorException e) {    // something went wrong in the algorithm, darn it!
                // Make the grid amount (if any)a little coarser, grid the vertices, and try again

                if (gridUnit == 0) {
                    // Find the initial grid unit based on the values of the vertices
                    double xMin = Double.POSITIVE_INFINITY;
                    double yMin = Double.POSITIVE_INFINITY;
                    double xMax = Double.NEGATIVE_INFINITY;
                    double yMax = Double.NEGATIVE_INFINITY;
                    for (int i = 0; i < numInputArrays; i++) {
                        TGPointArray vertexArray = fVertexArrays.getValue(i).fPoints;
                        int numVertices = vertexArray.getNumberOfPoints();
                        TGPoint points[] = vertexArray.getArray();
                        for (int j = 0; j < numVertices; j++) {
                            TGPoint point = points[j];
                            double x = point.x;
                            double y = point.y;
                            xMin = Math.min(xMin, x);
                            xMax = Math.max(xMax, x);
                            yMin = Math.min(yMin, y);
                            yMax = Math.max(yMax, y);
                        }
                    }

                    /* Compute the grid unit from the magnitude of the max/min values.
                       Note: If max == min, this would be an infinite loop; but if they were equal
                       then all the vertices would be the same and it would not have reached this point. */
                    double range = Math.max(xMax - xMin, yMax - yMin);
                    for (double gridMult = 1e-20; gridUnit == 0; gridMult *= 10)
                        gridUnit = range * gridMult;
                }
                else
                    gridUnit *= 10;

                // Grid the vertices to the grid unit
                for (int i = 0; i < numInputArrays; i++) {
                    TGPointArray vertexArray = fVertexArrays.getValue(i).fPoints;
                    int numVertices = vertexArray.getNumberOfPoints();
                    TGPoint points[] = vertexArray.getArray();
                    for (int j = 0; j < numVertices; j++) {
                        TGPoint point = points[j];
                        point.x = Math.floor((point.x / gridUnit) + 0.5)* gridUnit;
                        point.y = Math.floor((point.y / gridUnit) + 0.5)* gridUnit;
                    }
                }
            }
        }
        Reset();
    }

/*
    public final void SetVertexEngine(TVertexEngine vertexEngine) {
        fVertexEngine = vertexEngine;
    }

    public final TVertexEngine GetVertexEngine() {
        return fVertexEngine;
    }

    private final void SendInputVerticesToOutput(TCAGRoot root) {
        // Send the input vertices on to the output vertex engine
        //outSetClosed(fClosed);
        int numVertexArrays = fVertexArrays.getNumberOfValues();
        for (int i = 0; i < numVertexArrays; i++) {
            TCAGPointArray array = fVertexArrays.getValue(i);
            //outSetCAGNode(array.fCAGNode);

            int numVertices = array.fPoints.getNumberOfPoints();
            TGPoint pts[] = array.fPoints.getArray();
            for (int j = 0; j < numVertices; j++)
                outAddVertex(pts[j]);
            outEnd();
        }
        outRender(root);
    }
*/

}
