/*
 * @(#)TPolyEdge.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

// polygonal edge: two points and a CAG node
class TPolyEdge {
    TGPoint fTop;
    TGPoint fBottom;
    double fSlope;      // the slope
    TCAGUnary fCAGNode;    // parent CAG node
    int fIndex;        // unique index
    byte fSign;            // +1 or -1
    boolean fRelevant;    // belongs to the current y-event?

    TOutputVertexList fOutputList;    // output vertex list this edge is contributing to; null if none

    TPolyEdge() {}
    TPolyEdge(TGPoint p0, TGPoint p1, TCAGUnary node, int index) {
        Set(p0, p1, node, index);
    }

    public String toString() {
        return (
            "(" + fIndex + ") {" +
            "fTop" + fTop + ", fBottom" + fBottom + " " +
            "fSlope:" + fSlope + ", fSign:" + fSign + ", fRelevant:" + fRelevant + "}");
    }

    final void Set(TGPoint p0, TGPoint p1, TCAGUnary node, int index) {
        boolean horizontal = (p0.y == p1.y);
        if ((p0.y < p1.y) || (horizontal && (p0.x < p1.x))) {
            fTop = p0;
            fBottom = p1;
            fSign = 1;
        }
        else {
            fTop = p1;
            fBottom = p0;
            fSign = -1;
        }

        if (!horizontal)
            fSlope = (fBottom.x - fTop.x) / (fBottom.y - fTop.y);
        else if (fBottom.x > fTop.x)
            fSlope = Double.POSITIVE_INFINITY;

        fCAGNode = node;
        fIndex = index;
        fRelevant = false;
        fOutputList = null;
    }

    final void Transition() {
        fCAGNode.transition((int)fSign);
    }

    static int MoveEdge(TPolyEdge[] edges, int thisEdge, int edgeListStart) {
        // Bubble the edge down to the start of the edge list - rotate list
        TPolyEdge edgePtr = edges[thisEdge];
        while (--thisEdge >= edgeListStart)
            edges[thisEdge + 1] = edges[thisEdge];
        edges[edgeListStart] = edgePtr;
        ++edgeListStart;
        return edgeListStart;
    }
}
