/*
 * @(#)TPolygonGeometry.java	1.7 98/09/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TPolygonGeometry
||
*/

public class TPolygonGeometry extends MAreaGeometry {

    public TGPolygon fPolygon;

    public TPolygonGeometry() {
        fBounds = null;
    }
    public TPolygonGeometry(TGPolygon polygon) {
        fPolygon = new TGPolygon(polygon);
        fBounds = null;
        updateTimeStamp();
    }
    public TPolygonGeometry(TPolygonGeometry source) {
        super(source);
        fPolygon = new TGPolygon(source.fPolygon);
        fBounds = null;
        if (source.fBounds != null) {
            fBounds = (TGRect) ( source.fBounds.clone() );
        }
    }

    // === MAreaGeometry overrides ===
    public MAreaGeometry cloneAndTransform(TGrafMatrix m) {
        if (fPolygon.getNumberOfPoints() > kLargePolygon)
            return super.cloneAndTransform(m);
        else {
            TPolygonGeometry polygon = new TPolygonGeometry(this);
            polygon.fPolygon.transformBy(m);
            polygon.fBounds = null;
            polygon.updateTimeStamp();
            return polygon;
        }
    }

    public void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix) {

        boolean eoFill = fPolygon.getEOFill();
        TCAGUnary node;
        if (eoFill)
            node = new TCAGEvenOdd(parentNode);
        else
            node = new TCAGWindNum(parentNode);

        extractor.adoptCAGNode(node);
        extractor.extractPolygon(fPolygon, node, matrix);
    }

    public boolean isEmpty() {
	// should probably be more precise than this.
        return (fPolygon.getNumberOfPoints() < 3);
    }
    public boolean isPolygonal() {
        return true;
    }
    public boolean isRectangular() {
        // 2 points: true
        // 3 points: Triangle
        // 4 points: If rectilinear, MAY be a rectangle, but it may be degenerate!
        // 5 or more points: Forget it!
        return (fPolygon.getNumberOfPoints() <= 2);
    }

    public TGRect getBounds() {
        if (fBounds == null)
            fBounds = fPolygon.getBounds();

        return fBounds;
    }
    public boolean intersects(TGRect rect) {
        return fPolygon.intersects(rect);
    }
    public boolean contains(TGPoint point) {
        return fPolygon.contains(point);
    }
    public boolean contains(TGRect rect) {
        return (!fPolygon.intersects(rect)) && fPolygon.contains(rect.getTopLeft());
    }

    // === operators ===
    public final TPolygonGeometry copyFrom(TPolygonGeometry source) {
        if (this != source) {
            super.copyFrom(source);
            fPolygon = source.fPolygon;
            fBounds = source.fBounds;
        }
        return this;
    }

    public Object clone() {
        return (Object)(new TPolygonGeometry(this));
    }
    public boolean equals(MAreaGeometry that) {
        if ((getClass().equals(that.getClass()))) {
            TPolygonGeometry poly = (TPolygonGeometry)that;
            return (super.equals(that) && fPolygon.equals(poly.fPolygon));
        }
        else
            return false;
    }

    private TGRect fBounds;    // cached bounds

    private static final int kLargePolygon = 16;  // number of points in a large polygon
}
