/*
 * @(#)TRawArray.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
    TRawArray provides a low level substitute for traditional C arrays created
    with malloc & free.  Unlike other C++ objects, the storage is managed with
    the low-level heap storage routines.  ONLY PRIMITIVE TYPES AND OBJECTS
    WITH NO VIRTUAL METHODS MAY BE STORED IN TRAWARRAY!!.  Since TRawArray does
    NOT call constructors for the space it allocates, objects with vtables will
    not be properly initialized, most certainly resulting in mysterious crashes.
 */

// Since Java does not have templates, to make life a bit easier, at least the common
// things are covered. BUT this cannot be used as a base class polymorphically !!!

abstract class TRawArray {

    protected static final int kDefaultGrowAmount = 8;

    protected int fArraySize = 0;
    protected int fNumberOfValues = 0;
    protected int fGrowAmount = kDefaultGrowAmount;

    public TRawArray() {}
    public TRawArray(int size) {
        fNumberOfValues = size;
        fArraySize = size;
    }
    public final boolean equals(TRawArray source) {
        return fNumberOfValues == source.fNumberOfValues;
    }
    public final TRawArray copyFrom(TRawArray source) {
        fNumberOfValues = source.fNumberOfValues;
        fArraySize = source.fArraySize;
        return this;
    }

    // access
    public int getGrowIncrement() {
        return fGrowAmount;
    }
    public void setGrowIncrement(int increment) {
        if (increment <= 0)
            throw new IllegalArgumentException("RawArray.setGrowIncrement()");
        fGrowAmount = increment;
    }

    // access
    public int getNumberOfValues() {
        return fNumberOfValues;
    }

    // This swaps the contents of an array with victim.
    public void swap(TRawArray victim) {
        int tmpNum;

        tmpNum = fNumberOfValues;
        fNumberOfValues = victim.fNumberOfValues;
        victim.fNumberOfValues = tmpNum;

        tmpNum = fArraySize;
        fArraySize = victim.fArraySize;
        victim.fArraySize = tmpNum;

        tmpNum = fGrowAmount;
        fGrowAmount = victim.fGrowAmount;
        victim.fGrowAmount = tmpNum;
    }

    private int roundUpSize(int newsize) {
/*
        int wholePieces = newsize / fGrowAmount;
        if ((newsize % fGrowAmount) != 0)
            ++wholePieces;
        return wholePieces * fGrowAmount;
*/
        newsize += fGrowAmount - 1;
        return newsize - (newsize % fGrowAmount);
    }

    abstract void allocateArray(int size);
    abstract void reallocateArray(int newSize);
    abstract void newAllocation(int beginIndex, int count);

    // Reallocates the heap block
    public void resize(int newsize) {
        if (newsize != fNumberOfValues) {
            if (fNumberOfValues == 0) // Empty to full...
            {
                fArraySize = roundUpSize(newsize);
                allocateArray(fArraySize);
                newAllocation(0, fArraySize);
            }
            else if (newsize == 0)    // Full to empty...
            {
                fArraySize = 0;
                allocateArray(0);
            }                         // Full to full
            else {
                int newArraySize = roundUpSize(newsize);
                if (newArraySize != fArraySize) {
                    int OldArraySize = fArraySize;
                    reallocateArray(newArraySize);
                    fArraySize = newArraySize;
                    if (newArraySize > OldArraySize) {  // expand in size...
                        newAllocation(OldArraySize, newArraySize - OldArraySize);
                    }
                }
            }
            fNumberOfValues = newsize;
        }
    }

}
