/*
 * @(#)TRectGeometry.java	1.7 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TRectGeometry
||
*/

public class TRectGeometry extends MAreaGeometry {

    public TGRect fRect;

    public TRectGeometry() {
        fRect = new TGRect(    // Why? Look at TGStatics.java
                           TGStatics.kZeroRectTopLeftX,
                           TGStatics.kZeroRectTopLeftY,
                           TGStatics.kZeroRectBottomRightX,
                           TGStatics.kZeroRectBottomRightY);

        // TGRect.kZeroRect;  // old
    }

    public TRectGeometry(TGRect rect) {
        fRect = (TGRect)rect.clone();
        updateTimeStamp();
    }
    public TRectGeometry(TRectGeometry source) {
        super(source);
        fRect = null;
        if (source.fRect != null) {
            fRect = (TGRect) ( source.fRect.clone() );
        }
    }

    // === MAreaGeometry overrides ===
    public MAreaGeometry cloneAndAdd(MAreaGeometry g) {
        if (g.contains(fRect))
            return g;
        else
            return super.cloneAndAdd(g);
    }
    public MAreaGeometry cloneAndSubtract(MAreaGeometry g) {
        if (g.contains(fRect))
            return null;
        else
            return super.cloneAndSubtract(g);
    }
    public MAreaGeometry cloneAndIntersect(MAreaGeometry g) {
        if (g.contains(fRect))
            return this;
        else if (g.isRectangular()) {
            TRectGeometry newRect = (TRectGeometry)clone();
            newRect.fRect.intersectWith(g.getBounds());
            newRect.updateTimeStamp();
            return newRect;
        }
        else
            return super.cloneAndIntersect(g);
    }
    public MAreaGeometry cloneAndExclusiveOr(MAreaGeometry g) {
        return super.cloneAndExclusiveOr(g);
    }

    public MAreaGeometry cloneAndTransform(TGrafMatrix matrix) {
        if (matrix.isRectilinear()) {
            TRectGeometry newRect = (TRectGeometry)clone();
            newRect.fRect = matrix.transformBounds(fRect);
            return newRect;
        }
        else
            return super.cloneAndTransform(matrix);
    }

    public void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix) {

        TCAGEvenOdd node = new TCAGEvenOdd(parentNode);
        extractor.adoptCAGNode(node);
        extractor.extractRect(fRect, node, matrix);
    }

    public boolean isEmpty() {
        return fRect.isEmpty();
    }
    public boolean isPolygonal() {
        return true;
    }
    public boolean isRectangular() {
        return true;
    }

    public TGRect getBounds() {
        return fRect;
    }
    public boolean intersects(TGRect rect) {
        return fRect.intersects(rect);
    }
    public boolean contains(TGPoint point) {
        return fRect.contains(point);
    }
    public boolean contains(TGRect rect) {
        return fRect.contains(rect);
    }

    // === operators ===
    public final TRectGeometry copyFrom(TRectGeometry source) {
        if (this != source) {
            super.copyFrom(source);
            fRect = source.fRect;
        }
        return this;
    }

    public Object clone() {
        return (Object)(new TRectGeometry(this));
    }
    public boolean equals(MAreaGeometry that) {
        if (getClass().equals(that.getClass())) {
            TRectGeometry g = (TRectGeometry)that;
            return (super.equals(that) &&
                    (fRect == g.fRect ||
                     (fRect != null && g.fRect != null && fRect.equals(g.fRect))));
        }
        else
            return false;
    }
}
