/*
 * @(#)TSamplingExtractor.java	1.7 98/07/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||    TSamplingExtractor
||
||    TSamplingExtractor samples each of the geometries into a series of points.  The points
||    are sent to the specified vertex engine.  Also, if the bounds engine is not NIL, a bounding
||    rectangle is calculated and sent to the specified bounds engine.
||
*/

public class TSamplingExtractor extends TGrafExtractor {

    private static final TGRect fkInfiniteRect =  // Why?  Look at TGStatics.java
        new TGRect(
                   TGStatics.kInfRectTopLeftX,
                   TGStatics.kInfRectTopLeftY,
                   TGStatics.kInfRectBottomRightX,
                   TGStatics.kInfRectBottomRightY);


    public TSamplingExtractor(
        TOutlineMakerVertexEngine outputVertexEngine,
        double epsilon)
    {
        fVertexEngine = outputVertexEngine;
        fClipBounds = fkInfiniteRect;
        fEpsilon = epsilon;
    }

    // === Area-defining geometries ===
    public void extractRect(TGRect g, TCAGUnary node, TGrafMatrix matrix) {
        fVertexEngine.SetCAGNode(node);
        //fVertexEngine.SetClosed(true);
        if (matrix == null || matrix.isIdentity()) {
            fVertexEngine.AddVertex(g.getTopLeft());
            fVertexEngine.AddVertex(g.getTopRight());
            fVertexEngine.AddVertex(g.getBottomRight());
            fVertexEngine.AddVertex(g.getBottomLeft());
        }
        else {
            fVertexEngine.AddVertex(matrix.transformPoint(g.getTopLeft()));
            fVertexEngine.AddVertex(matrix.transformPoint(g.getTopRight()));
            fVertexEngine.AddVertex(matrix.transformPoint(g.getBottomRight()));
            fVertexEngine.AddVertex(matrix.transformPoint(g.getBottomLeft()));
        }
        fVertexEngine.End();
    }

    public void extractEllipse(TGEllipse g, TCAGUnary node, TGrafMatrix m) {
/* Why reinvent the wheel again here, we could just use whatever the conic support is in TGCurve...
        if (m == null || m.isAffine()) {
            TGRect bounds = g.getBounds();
            TGRect actualBounds;
            double shear = 0.;

            if (m == null || m.isRectilinear()) {
                if (m != null)
                    bounds = m.transformBounds(bounds);
                actualBounds = bounds;
            }
            else {
                double a = m.getElement(AffineTransform.kScaleX);
                double b = m.getElement(AffineTransform.kShearX);
                double d = m.getElement(AffineTransform.kShearY);
                double e = m.getElement(AffineTransform.kScaleY);

                TGPoint newRadius = new TGPoint();
                TGPoint radius = TGPoint.multiply(bounds.getSize(), 0.5);
                double hhe = radius.y * radius.y * e;
                double wwd = radius.x * radius.x * d;
                if ((hhe != 0) || (wwd != 0)) {
                    double term = hhe * e + wwd * d;
                    newRadius.y = Math.sqrt(term);
                    newRadius.x = Math.abs(radius.x * radius.y * (a * e - b * d) / newRadius.y);
                    TGPoint center = m.transformPoint(bounds.getCenter());
                    bounds.setTopLeft(TGPoint.subtract(center, newRadius));
                    bounds.setBottomRight(TGPoint.add(center, newRadius));
                    shear = -(hhe * b + wwd * a) / term;
                }
                else
                    bounds = fkZeroRect; // don't. Instead '= new TGRect (fkZeroRect);'

                actualBounds = bounds;
                actualBounds.inset(new TGPoint(-Math.abs(shear * newRadius.y) - 0.5, 0));
            }

            TGPoint size = bounds.getSize();

            // Make sure the size doesn't go to zero
            if (size.x == 0)
                size.x = 0.00001;
            if (size.y == 0)
                size.y = 0.00001;

            double theta;
            int count;
            double error = 0.4;    // Play with this value

            TGPoint actualSize = actualBounds.getSize();
            double majorAxis = actualSize.x + actualSize.y;    // approximation
            double root = (majorAxis - 2 * error) / majorAxis;
            theta = Math.acos((2 * root * root) - 1);

            // make sure there is an integer number of steps
            double kHalfPi = Math.PI * 0.5;
            count = (int)(kHalfPi / theta) + 1;
            if (count < 1)
                count = 1;
            theta = kHalfPi / count;

            // Get the values
            double sinTheta = Math.sin(theta);
            double cosTheta = Math.cos(theta);
            double a = cosTheta;
            double b = -(size.x / size.y) * sinTheta;
            double c = (size.y / size.x) * sinTheta;
            double d = cosTheta;
            TGPoint center = bounds.getCenter();
            double centerX = center.x;
            double centerY = center.y;
            double x = 0;
            double y = -size.y * 0.5;

            if (shear != 0) {
                double extra = shear * c;
                a -= extra;
                b -= shear * extra;
                d += extra;
                x = shear * size.y * 0.5;
            }

            int numPoints = count * 4;
            // the last point is repeated, but it's not used
            TGPointArray tmp = new TGPointArray(numPoints + 1);
            TGPoint points[] = tmp.getArray();
            int i1 = 0;
            int i2 = numPoints;
            int i3 = count * 2;
            int i4 = i3;
            for (int i = 0; i <= count; i++) {
                TGPoint quad1 = points[i1];
                TGPoint quad2 = points[i2];
                TGPoint quad3 = points[i3];
                TGPoint quad4 = points[i4];

                double doubleShear = (shear != 0) ? (2 * y * shear) : 0.;

                quad1.x = centerX + x;
                quad1.y = centerY + y;
                i1++;
                quad2.x = centerX - x - doubleShear;
                quad2.y = centerY + y;
                i2--;
                quad3.x = centerX - x;
                quad3.y = centerY - y;
                i3++;
                quad4.x = centerX + x + doubleShear;
                quad4.y = centerY - y;
                i4--;

                double newX = a * x + b * y;
                y = c * x + d * y;
                x = newX;
            }

            fVertexEngine.SetCAGNode(node);
            //fVertexEngine.SetClosed(true);
            for (int i = 0; i < numPoints; i++)
                fVertexEngine.AddVertex(points[i]);
            fVertexEngine.End();
            return;
        }
        // else convert it to a TGLoop...
*/
        extractLoop(new TGLoop(g), node, m);
    }
    public void extractPolygon(TGPolygon g, TCAGUnary node, TGrafMatrix matrix) {
        int numPoints = g.getNumberOfPoints();
        if (numPoints > 0) {
            TGPoint pts[] = g.getPoints();
            if (matrix != null && !matrix.isIdentity()) {
                TGPointArray points = new TGPointArray();
                g.getPoints(points);
                matrix.transformPoints(points);
                pts = points.getArray();
            }

            fVertexEngine.SetCAGNode(node);
            //fVertexEngine.SetClosed(true);
            for (int i = 0; i < numPoints; i++)
                fVertexEngine.AddVertex(pts[i]);
            fVertexEngine.End();
        }
    }
    public void extractLoop(TGLoop g, TCAGUnary node, TGrafMatrix matrix) {
        if (!g.isEmpty()) {
            int numPoints = g.getNumberOfPoints();
            if (numPoints > 0) {
                int order = g.getOrder();
                boolean closeable = g.isCloseable();

                TGRPointArray tPts = g.fPoints;
                TGRPoint rPoints[] = tPts.getArray();
                TGParametricArray tKnots = g.fKnots;
                double knots[] = tKnots.getArray();
                if (closeable || matrix != null) {
                    TGRPoint oldPts[] = rPoints;
                    tPts = new TGRPointArray(numPoints + (closeable ? (order - 1) : 0));
                    rPoints = tPts.getArray();
                    System.arraycopy(oldPts, 0, rPoints, 0, numPoints);
                    if (matrix != null && !matrix.isIdentity())
                        matrix.transformPoints(tPts);

                    if (closeable) {
                        // If this is a "closed" loop, allow the control points to wrap around
                        for (int i = numPoints; i < numPoints + order - 1; i++)
                            rPoints[i].copyFrom(rPoints[i - numPoints]);

                        if (order > 2) {    // refine knot vector
                            double k[] = knots;
                            int numKnots = 2 * order + numPoints - 1;
                            tKnots = new TGParametricArray(numKnots);
                            knots = tKnots.getArray();
                            System.arraycopy(k, 0, knots, 0, order + numPoints);
                            double delta = knots[1] - knots[0];
                            for (int i = order + numPoints; i < numKnots; i++)
                                knots[i] = knots[i - 1] + delta;
                        }
                    }
                }

                TAreaDiscretiser rick =
                    new TAreaDiscretiser(fVertexEngine, node,
					 fkInfiniteRect, fEpsilon);
                rick.DiscretizeLoop(order, tKnots, tPts, true);
            }
        }
    }

    // === Render ===
    public void render(TCAGRoot root) {
        fVertexEngine.Render(root);
        deleteCAGNodes();
    }

/*
    // === Sets/Gets ===
    public final void SetVertexEngine(TVertexEngine outputVertexEngine) {
        fVertexEngine = outputVertexEngine;
    }
    public final TVertexEngine GetVertexEngine() {
        return fVertexEngine;
    }
*/

    // This is to clue the discretizer about the area it actually
    // needs to generate samples for.
    public final void SetClipBounds(TGRect clipRect) {
        fClipBounds = clipRect;
    }

    //private TVertexEngine fVertexEngine;    // accepts the samples
    private TOutlineMakerVertexEngine fVertexEngine;    // accepts the samples
    private TGRect fClipBounds;      // Bounds of device to draw on
    private double fEpsilon;         // Bounds of device to draw on

    // static int kVertexIncrement;  // resize the vertex array by this much
}
