/*
 * @(#)TTransformedGeometry.java	1.5 98/04/02
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*================
||
||  TTransformedGeometry
||
||  This class contains a matrix and wraps up another MAreaGeometry object.
||  It is handy for doing fast transformations; just attach one of these above
||  another geometry.
||
*/

class TTransformedGeometry extends MAreaGeometry {

    public TTransformedGeometry(MAreaGeometry g, TGrafMatrix modelMatrix) {
        fGeometry = g;
        fMatrix = (TGrafMatrix)modelMatrix.clone();
        fBounds = null;
        updateTimeStamp();
    }

    public TTransformedGeometry(TTransformedGeometry source) {
        super(source);
        fGeometry = null;
        fMatrix = null;
        fBounds = null;

        if (source.fGeometry != null) {
            fGeometry = (MAreaGeometry) ( source.fGeometry.clone() );
        }

        if (source.fMatrix != null) {
            fMatrix = (TGrafMatrix) ( source.fMatrix.clone() );
        }

        if (source.fBounds != null) {
            fBounds = (TGRect) ( source.fBounds.clone() );
        }
    }

    // === MAreaGeometry overrides ===
    public MAreaGeometry cloneAndTransform(TGrafMatrix matrix) {
        if (matrix.isIdentity() || fMatrix.isIdentity())
            return fGeometry;

        TGrafMatrix concatMatrix = (TGrafMatrix)matrix.clone();
        concatMatrix.preConcatWith(fMatrix);

        if (concatMatrix.isIdentity())
            return fGeometry;
        else
            return new TTransformedGeometry(fGeometry, concatMatrix);
    }

    public void extract(
        TGrafExtractor extractor,
        TCAGNode parentNode,
        TGrafMatrix matrix)
    {
        TGrafMatrix save = fMatrix;
        fMatrix = (TGrafMatrix)matrix.clone();
        fMatrix.preConcatWith(save);
        fGeometry.extract(extractor, parentNode, fMatrix);
        fMatrix = save;
    }

    public boolean isEmpty() {
        return fGeometry.isEmpty();
    }
    public boolean isPolygonal() {
        return fGeometry.isPolygonal();
    }
    public boolean isRectangular() {
        return (fMatrix.isRectilinear() && fGeometry.isRectangular());
    }
    public boolean isSingular() {
        return fGeometry.isSingular();
    }

    public TGRect getBounds() {
        if (fBounds == null) {
            fBounds = fMatrix.transformBounds(fGeometry.getBounds());    // creates a new TGRect
        }
        return fBounds;
    }
    public boolean intersects(TGRect rect) {
        return fGeometry.intersects(fMatrix.untransformBounds(rect));
    }
    public boolean contains(TGPoint point) {
        return fGeometry.contains(fMatrix.untransformPoint(point));
    }
    public boolean contains(TGRect rect) {
        return fGeometry.contains(fMatrix.untransformBounds(rect));
    }

    // === operators ===
    public MAreaGeometry copyFrom(MAreaGeometry source) {
        if (this.getClass().equals(source.getClass())) {
            this.copyFrom((TTransformedGeometry)source);
        }
        else
            super.copyFrom(source);
        return (MAreaGeometry)this;
    }
    public final TTransformedGeometry copyFrom(TTransformedGeometry source) {
        if (this != source)
        {
            super.copyFrom(source);
            fGeometry = source.fGeometry;
            fMatrix = source.fMatrix;
            fBounds = source.fBounds;
        }
        return this;
    }

    public Object clone() {
        return (Object)(new TTransformedGeometry(this));
    }
    public boolean equals(MAreaGeometry that) {
        if (getClass().equals(that.getClass())) {
            TTransformedGeometry g = (TTransformedGeometry)that;
            return (super.equals(that) &&
                    fMatrix.equals(g.fMatrix) &&
                    (fGeometry == g.fGeometry ||
                     (fGeometry != null && g.fGeometry != null && fGeometry.equals(g))));
        }
        else
            return false;
    }

    private MAreaGeometry fGeometry;
    private TGrafMatrix fMatrix;

    private TGRect fBounds;
}
