/*
 * @(#)prStatic.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * (C) Copyright Taligent, Inc. 1996 - 1997, All Rights Reserved
 * (C) Copyright IBM Corp. 1996 - 1998, All Rights Reserved
 *
 * The original version of this source code and documentation is
 * copyrighted and owned by Taligent, Inc., a wholly-owned subsidiary
 * of IBM. These materials are provided under terms of a License
 * Agreement between Taligent and Sun. This technology is protected
 * by multiple US and International patents.
 *
 * This notice and attribution to Taligent may not be removed.
 * Taligent is a registered trademark of Taligent, Inc.
 *
 */

package sun.awt.Albert;

/*
 * prStatic has static methods of prConics, prKnots, prNumerics and prSplineRoutines
 *
 */

class prStatic {

// Support code for conic curves:

//class prConics {

    static double sgn(double x) {
        if (x == 0.0)
            return 0.0;
        if (x < 0.0)
            return -1.0;
        else
            return 1.0;
    }

    // Arc generation note:  We want to break things down into arcs of 90 degrees or less,
    // to avoid control points that fly out to infinity (e.g.: 180 degrees arc).  Doing this relys on
    // a few simple rules:
    //
    // If V0 and V1 are two vectors with a common origin, and the angle between them is Q:
    //  V0 . V1 has the same sign as the cosine Q
    //  V0 x V1 has the same sign as the sine Q
    //  (By V0 x V1 we mean the value of the "z" component, V0x*V1y - V1x*V0y)
    //  The signs of sine and cosine as we go around the quadrants (counterclockwise) are:
    //      S+ | S+
    //      C- | C+
    //      ---+---> V0
    //      C- | C+
    //      S- | S-
    //  In other words, if abs(Q) < 90 degrees, the cosine is +, and if Q < 180 degrees, the sine is +
    //  In yet more words, the cos tells you if it's > 90 degrees, and sin gives direction (cw/ccw)
    //  A vector P perpendicular to V is P=(-Vy, Vx) [ccw] or P=(Vy, -Vx) [cw]
    //  Given V and P, a 90 degrees arc is formed with V, V + P, P; w0 =1, w1 = sqrt(2)/2, w2=1
    //
    //  A line equation is Ax + By = C, A=y, B=-dx, C=Ax0 + By0 [dy = y1-y0]

    //
    // To generate the Rational quadratic Bezier representation of an arc, given the
    // two endpoints of the arc and its center.  (If the distances from the ends to the
    // center aren't equal, then the arc won't be circular).
    //
    // A Rational Bezier representation of an arc is formed by three points:
    //                  V1
    //                  /\
    //                 /  \
    //                /****\   (*=arc)
    //               /*    *\
    //              V0      V2
    //
    // The vectors V0V1 and V2V1 are formed from vectors perpendicular to p0p1 and p2p1.
    // The rational (w) coordinate for V1 is the cosine of the angle between V0,V1 and V0,V2.
    // (See Boem, Farin & Kahmann, "A Survey of Curve and Surface techniques for CAGD", CAGD, 1984).
    //


    // This particular routine only generates the points for arcs < 90 degrees.  (It works for larger
    // arcs, but produces awkward control points out towards infinity).

    // This routine sets the last two points in "pts" (assumes the first is set by caller).
    // and returns the end count.

    static int NarrowArcEndCenterEndPts(
        TGPoint p0,
        TGPoint p1,      // Input points
        TGPoint p2,
        int count,
        TGRPointArray points)    // Resulting control points
    {
        // Generate the line equations (Ax+By=C) of the
        // vectors Perpendicular to center->endpoint lines.
        double A1 = p0.x - p1.x;
        double B1 = p0.y - p1.y;
        double C1 = A1 * p0.x + B1 * p0.y;

        double A2 = p2.x - p1.x;
        double B2 = p2.y - p1.y;
        double C2 = A2 * p2.x + B2 * p2.y;

        TGRPoint pts[] = points.getArray();

        // Find the intersection of above two lines by solving:
        //  [A1 B1] [x] = [C1]
        //  [A2 B2] [y] = [C2]

        double tmp = A1 * B2 - B1 * A2;
        if (tmp == 0.0)                 // Vectors 180 degrees apart (should not call this routine)
            // Straight line == circle of kInfinity radius
            pts[count].copyFrom(TGPoint.add(p0, p2).multiplyBy(0.5));
        else {
            pts[count].fX = (B2 * C1 - B1 * C2) / tmp;
            pts[count].fY = (A1 * C2 - A2 * C1) / tmp;

            // The rational component is the cosine of the angle from p0-p2 to p0-p1
            // Cos of this angle is equal to (|p0p2|/2) / |p0p1| (See Das Paper, sec. 30).
            // This is found with triangle ratios.
            tmp = Square(p0.x - pts[count].fX) + Square(p0.y - pts[count].fY); // prNumeric
            if (tmp == 0.0)
                pts[count].copyFrom(p0);            // Collapsed to a point
            else {
                pts[count].fW =
                    Math.sqrt((Square(p0.x - p2.x) + Square(p0.y - p2.y)) / (tmp*4.0)); // prNumeric
                pts[count].fX *= pts[count].fW;
                pts[count].fY *= pts[count].fW;
            }
        }
        count++;

        pts[count].copyFrom(p2);
        count++;

        return count;
    }

    // Construct a 90 degrees arc from perpendicular vectors v0 to perp, about origin
    // and returns the end count.

    static int Arc90Pts(
        TGPoint origin,
        TGPoint v0,
        TGPoint perp,
        int count,
        TGRPointArray points)
    {
        double r2o2 = Math.sqrt(2.0) * 0.5;
        TGRPoint pts[] = points.getArray();
        TGPoint corner = TGPoint.add(v0, perp);
        pts[count].copyFrom(corner.addBy(origin).multiplyBy(r2o2));
        pts[count].fW = r2o2;

        count++;
        pts[count].copyFrom(TGPoint.add(perp, origin));

        count++;
        return count;
    }

    //
    // Given the points for an ArcEndCenterEnd, produce the "rest" of the circle.
    // If the vectors p0p1 and p2p1 aren't equal, interpolate between them as you
    // go around the circle. (1,1/2,0 for >180 degrees <270 degrees; 1, 2/3, 1/3, 0 for >270 degrees)
    //
    // ... and returns the end count.

    static int OtherArcEndCenterEndPts(
        TGPoint p0,
        TGPoint p1,
        TGPoint p2,
        int count,
        TGRPointArray points)
    {
        TGRPoint pts[] = points.getArray();

        pts[count].copyFrom(p0);
        count++;

        TGPoint v0 = TGPoint.subtract(p0, p1);
        TGPoint v2 = TGPoint.subtract(p2, p1);

        double cossgn = v0.x * v2.x + v0.y * v2.y;
        double sinsgn = v0.x * v2.y - v0.y * v2.x;

        // Attempt to deal with case of |v0| != |v1| (non-circ. arc)
        double len0 = v0.vectorLength();
        double len2 = v2.vectorLength();

        TGPoint base = (TGPoint)v0.clone();
        if (len0 != 0.0)
            base.divideBy(len0);

        double lengthRatio;
        if (cossgn > 0.0)
            lengthRatio = (2.0 / 3.0) * (len0 - len2) + len2;
        else
            lengthRatio = 0.5 * (len0 - len2) + len2;

        TGPoint perp = new TGPoint();
        if (sinsgn < 0.0) {
            perp.x = -base.y;   // Arc is clockwise, must
            perp.y =  base.x;   // go counterclockwise
        }
        else {
            perp.x =  base.y;   // Arc is counterclockwise,
            perp.y = -base.x;   // must go clockwise
        }
        perp.multiplyBy(lengthRatio);

        count = Arc90Pts(p1, v0, perp, count, points);

        if (cossgn > 0.0)
            lengthRatio = (1.0 / 3.0) * (len0 - len2) + len2;
        else
            lengthRatio = len2;

        TGPoint opposite = TGPoint.multiply(base, -lengthRatio);
        count = Arc90Pts(p1, perp, opposite, count, points);

        if (cossgn >= 0.0)      // Arc is > 270 degrees
        {
            lengthRatio = len2;

            if (sinsgn >= 0.0) {
                perp.x = -base.y;   // Arc is clockwise, must
                perp.y =  base.x;   // go counterclockwise
            }
            else {
                perp.x =  base.y;   // Arc is counterclockwise,
                perp.y = -base.x;   // must go clockwise
            }
            perp.multiplyBy(lengthRatio);

            count = Arc90Pts(p1, opposite, perp, count, points);
        }

        return NarrowArcEndCenterEndPts(pts[count-1].dropW(), p1, p2, count, points);
    }

    // Generate an arc between p0 and p2, with the circle centered at p1.
    // Arc is circular if |p0p1| = |p2p1|.   Checks for all sorts of nasty cases.
    // ... and returns the end count.

    static int ArcEndCenterEndPts(
        TGPoint p0,
        TGPoint p1,
        TGPoint p2,
        int count,
        TGRPointArray points)
    {
        int base = count;
        TGRPoint pts[] = points.getArray();

        pts[count].copyFrom(p0);

        count++;

        TGPoint v0 = TGPoint.subtract(p0, p1);
        TGPoint v2 = TGPoint.subtract(p2, p1);

        double cossgn = v0.x * v2.x + v0.y * v2.y;
        double sinsgn = v0.x * v2.y - v0.y * v2.x;

        TGPoint tmpP1;
        if ((sinsgn == 0.0) && (cossgn < 0.0))  // Arc is 180 degrees, must do 2 parts
        {
            tmpP1 = TGPoint.add(p0, p2).multiplyBy(0.5);    // Make circle w/center between ends
            v0.copyFrom(p0).subtractBy(tmpP1);

            TGPoint perp = new TGPoint(-v0.y, v0.x);
            count = Arc90Pts(p1, v0, perp, count, points);
        }
        else
            tmpP1 = (TGPoint)p1.clone();

        count = NarrowArcEndCenterEndPts(pts[count-1].dropW(), tmpP1, p2, count, points);

        if ((cossgn < 0.0) && (sinsgn != 0.0)) {
            // Arc is > 90 degrees, split the segment to keep ctl point reasonable
            TGRPoint rp0 = pts[base + 0];         // this will not change
            TGRPoint rp1 = pts[base + 1];         // save this
            TGRPoint rp2 = pts[base + 4].copyFrom(pts[base + 2]);

            pts[base + 1] =
                new TGRPoint(
                    (rp0.fX + rp1.fX) * 0.5,
                    (rp0.fY + rp1.fY) * 0.5,
                    (rp0.fW + rp1.fW) * 0.5);

            pts[base + 2].fX = (rp0.fX + 2*rp1.fX + rp2.fX) * 0.25;
            pts[base + 2].fY = (rp0.fY + 2*rp1.fY + rp2.fY) * 0.25;
            pts[base + 2].fW = (rp0.fW + 2*rp1.fW + rp2.fW) * 0.25;

            pts[base + 3].fX = (rp1.fX + rp2.fX) * 0.5;
            pts[base + 3].fY = (rp1.fY + rp2.fY) * 0.5;
            pts[base + 3].fW = (rp1.fW + rp2.fW) * 0.5;

            count += 2;
        }

        // Hideous special case:  If we are doing a < 90 degrees arc, check to see if the corner
        // point generated by NarrowArcEndCenterEndPts is outside the arc defined
        // by v0 v1.  If so, generate the other side of the circle instead, otherwise the
        // abovementioned arc will be poorly behaved.  Question:  is there a cheaper
        // way to check for this case?

        if (cossgn >= 0.0) {
            double v0dist = v0.vectorLength();
            double v2dist = v2.vectorLength();
            double tmp = v0dist * v2dist;
            if (tmp != 0.0)
            {
                sinsgn /= (v0dist * v2dist);    // Make a real sine value
                TGPoint v1 = pts[count-2].divW().subtractBy(p1);
                double v1dist = v1.vectorLength();
                double v1sin = (v0.x * v1.y - v0.y * v1.x) / (v0dist * v1dist);
                if ((Math.abs(v1sin) > Math.abs(sinsgn)) || (sgn(v1sin) != sgn(sinsgn))) {
                    count -= 3;
                    count = OtherArcEndCenterEndPts(p0, p1, p2, count, points);
                }
            }
        }
        return count;
    }

/*
    //
    // Generate a circular arc that passes through three points, in the direction of the circle.
    // Doing this requires breaking it down into six cases, depending how p0, p1, p2 match up
    //           M
    //          /|\
    //         / | \
    //        A  B  C
    //
    // with A, B, C in the diagram above (three combinations of three).  We look at the
    // sign of the sin of the angles AMB, BMC, AMC to determine which segments are > 180 degrees.
    // (Sorry, this is impossible to explain without real pictures...)
    //
    // ... and returns the end count.

    static int ArcThru3PointsPts(
        TGPoint p0,
        TGPoint p1,
        TGPoint p2,
        int count,
        TGRPointArray points)
    {
        return ArcThru3PointsPts(p0, p1, p2, count, points, null);
    }
    static int ArcThru3PointsPts(
        TGPoint p0,
        TGPoint p1,
        TGPoint p2,
        int count,
        TGRPointArray points,
        TConicInfo c)
    {
        // Find the perpendicular bisectors of p0p1 and p1p2
        TGPoint i0 = TGPoint.add(p0, p1).multiplyBy(0.5);
        TGPoint i1 = TGPoint.add(p1, p2).multiplyBy(0.5);

        // The intersection of i0 and i1 specify the center of the circle
        double A1 = (p0.x - p1.x);
        double B1 = (p0.y - p1.y);
        double C1 = A1 * i0.x + B1 * i0.y;

        double A2 = (p2.x - p1.x);
        double B2 = (p2.y - p1.y);
        double C2 = A2 * i1.x + B2 * i1.y;

        TGRPoint pts[] = points.getArray();

        // Solve  [A1 B1] [x] = [C1]
        //        [A2 B2] [y] = [C2] using Cramer's rule
        double tmp = A1 * B2 - B1 * A2;
        if (tmp == 0.0) {
            // Points are colinear, just return something obvious...
            pts[count++].copyFrom(p0);
            pts[count++].copyFrom(p1);
            pts[count++].copyFrom(p2);
            return count;
        }

        // Find the center of the circle
        TGPoint center =
            new TGPoint((B2 * C1 - B1 * C2) / tmp, (A1 * C2 - A2 * C1) / tmp);

        TGPoint v0 = TGPoint.subtract(p0, center);
        TGPoint v1 = TGPoint.subtract(p1, center);
        TGPoint v2 = TGPoint.subtract(p2, center);

        double sin01 = v0.crossProduct(v1);
        double sin12 = v1.crossProduct(v2);
        double sin02 = v0.crossProduct(v2);

        // Should be caught by colinear check above
        // InternalAssert(((sin01 != 0.0) || (sin12 != 0.0)), "Unknown case in ArcThrough3Points?  Shouldn't get here.");
        if ((sin01 == 0.0) && (sin12 == 0.0))
            throw new IllegalArgumentException("Unknown case in ArcThrough3Points.");

        // If one of the angles is exactly 180 degrees, then figure
        // out which side of the line the third point is on, and construct
        // the appropriate 180 deg. arc out of two 90s on the opposite side.

        TGPoint perp;
        if (sin01 == 0.0) {
            if (sin02 > 0.0)
                perp = new TGPoint(v0.y, -v0.x);
            else
                perp = new TGPoint(-v0.y, v0.x);

            pts[count++].copyFrom(p0);
            count = Arc90Pts(center, v0, perp, count, points);
            count = Arc90Pts(center, perp, v1, count, points);
            count--;    // Avoid doing p1 twice
            count = ArcEndCenterEndPts(p1, center, p2, count, points);
        }
        else if (sin12 == 0.0) {
            double sin10 = -sin01;
            if (sin10 > 0.0)
                perp = new TGPoint(v1.y, -v1.x);
            else
                perp = new TGPoint(-v1.y, v1.x);

            count = ArcEndCenterEndPts(p0, center, p1, count, points);
            count = Arc90Pts(center, v1, perp, count, points);
            count = Arc90Pts(center, perp, v2, count, points);
        }
        else {
            // Basic cases...
            if (((sin01 < 0.0) && (sin12 < 0.0)) || ((sin01 > 0.0) && (sin12 > 0.0))) {
                count = ArcEndCenterEndPts(p0, center, p1, count, points);
                count--;    // Arcs share start & end
                count = ArcEndCenterEndPts(p1, center, p2, count, points);
            }
            else {
                if (((sin12 < 0.0) && (sin02 < 0.0)) || ((sin12 > 0.0) && (sin02 > 0.0))) {
                    count = ArcEndCenterEndPts(p0, center, p1, count, points);
                    count--;    // Arcs share start & end
                    count = OtherArcEndCenterEndPts(p1, center, p2, count,  points);
                }
                else {
                    count = OtherArcEndCenterEndPts(p0, center, p1, count, points);
                    count--;    // Arcs share start & end
                    count = ArcEndCenterEndPts(p1, center, p2, count, points);
                }
            }
        }

        // now the conic stuff that (*&@^#! messed up...
        if (c != null) {
            double radius = v0.vectorLength();
            c.fArcBounds = new TGRect(-radius, -radius, radius, radius);
            c.fArcBounds.setCenter(center);
            c.fDirection =  // clockwise ... Albert way.
                (sin01 > 0.0 && sin12 > 0.0) ||
                (sin01 > 0.0 && sin12 < 0.0 && sin02 < 0.0) ||
                (sin01 < 0.0 && sin12 > 0.0 && sin02 < 0.0);
        }

        return count;
    }
*/

    // set curve to an elliptical arc from angle1 to angle2.  Now, the Arc
    // always goes "counterclockwise" (cuz that's how sin's and cos's go),
    // BUT, because the coordinate system in Albert is by default "upsidedown"
    // from the mathematical one, the arcs go clockwise.  Got it?

    // ... and returns the end count.

    static int ArcAngleAnglePts(
        TGEllipse e,
        double angle1,
        double angle2,
        int count,
        TGRPointArray points)
    {
        // Make sure angles are reasonable
        boolean negdiff = (angle1 - angle2) < 0.0;

        angle1 = Remainder(angle1, 360.0); // prNumeric
        angle2 = Remainder(angle2, 360.0); // prNumeric

        if (angle1 < 0)
            angle1 += 360.0;
        if (angle2 < 0)
            angle2 += 360.0;

        boolean less180 = Math.abs(angle1 - angle2) <= 180.0;

        double a1 = angle1 * kDegreesToRadians; // prNumeric
        double a2 = angle2 * kDegreesToRadians; // prNumeric

        TGPoint center = new TGPoint(0.0, 0.0);
        TGPoint v1 = new TGPoint(Math.cos(a1), Math.sin(a1));
        TGPoint v2 = new TGPoint(Math.cos(a2), Math.sin(a2));

        //
        // Construct the section of a unit circle from the angles given.
        //
        if (!(less180 ^ negdiff))
            count = ArcEndCenterEndPts(v1, center, v2, count, points);
        else
            count = OtherArcEndCenterEndPts(v1, center, v2, count, points);

        TGRect r = e.getBounds();

        // Transform the unit circle section by the ellipse

        double radx = Math.abs(r.width) * 0.5;
        double rady = Math.abs(r.height) * 0.5;
        center.x = r.x + radx;
        center.y = r.y + rady;

        TGRPoint pts[] = points.getArray();
        for (int i = 0; i < count; i++) {
            pts[i].fX = pts[i].fX * radx + pts[i].fW * center.x;
            pts[i].fY = pts[i].fY * rady + pts[i].fW * center.y;
        }
        return count;
    }

//}


// I've split this stuff out with the intention of creating an *Internal*
// knot vector class (not useful externally because knot vectors are
// meaningless without the number of points and order - keeping that
// information redundantly is a recipe for disaster...)

//class prKnots {

    // ---------- Evaluation ----------

    //
    // Compute the derivatives.  Assumes the order-1'th set of spline
    // values is passed in dvals, and returns a modified set of dvals with
    // derivatives.  From Bartel's net code, but fixed to be insensitive
    // to knot vector spacing.
    //

    static void ComputeDeriv(
        int order,
        int brkPoint,
        int deriv,
        double knots[],
        double dvals[])
    {
        double omega;
        int di, dr, s;
        double knotScale;

        if ((order == 2) && (deriv == 1))   // Special cases "bvals" loop will miss
        {
            dvals[0] = 1.0; dvals[1] = 0.0;
        }

        if ((order == 3) && (deriv == 2))
        {
            dvals[0] = 1.0; dvals[1] = 0.0; dvals[2] = 0.0;
        }

        knotScale = knots[brkPoint + 1] - knots[brkPoint];

        for (dr = order - deriv + 1; dr <= order; dr++)
        {
            di = brkPoint - dr + 1;
            for (s = dr - 2; s >= 0L; s--)
            {
                di++;
                omega = knotScale * ((double)(dr-1)) / (knots[di+dr-1] - knots[di]);
                dvals[s + 1] += -omega * dvals[s];
                dvals[s] *= omega;
            }
        }
    }

    //
    // Compute the Spline functions B[i,k](u).  For each call, this computes the
    // the k values B[i-k][k](u)...B[i][k](u) and returns them in bvals.  If dvals1
    // or dvals are not NULL, the corresponding derivative values are computed as
    // well.  This is a critical inner loop for anything that evalutes curves or
    // surfaces.  This code is loosely based on some routines posted to the net by
    // Richard Bartels.
    //
    // Note:  The order and knot vectors are passed in as arguments so this routine
    // can be used by TGCurve3D, TGSurface and the rendering code.
    //
    // Another note:  This code does NO ERROR CHECKING.  The brkPoint, knots, etc.
    // are assumed to be for a valid interval of a valid spline.
    //
    // A final note: This code has been tuned extensivly with the MPW C
    // compiler.  If you think you can speed it up, try MEASURING it
    // first.  Many attempts to do so (common sub expression removal, more
    // pointers) have only slowed it down.
    //

    public static void ComputeSplineValues(// in the system (surface, 3D, etc) can
        double u,                    // share the implementation.
        int brkPoint,                // (Maybe move this out of the class def,
        int order,                  // and just make it a global public static???)
        double knots[],
        double bvals[])
    {
        ComputeSplineValues(u, brkPoint, order, knots, bvals, null, null);
    }
    public static void ComputeSplineValues(// in the system (surface, 3D, etc) can
        double u,                    // share the implementation.
        int brkPoint,                // (Maybe move this out of the class def,
        int order,                  // and just make it a global public static???)
        double knots[],
        double bvals[],
        double dvals1[],            // Deriv's are optional, set to NULL to
        double dvals2[])            // avoid their computation
    {
        double omega;

        bvals[0] = 1.0;

        //
        // Common cases (linear, cubic and quadratic) are unrolled.
        //
        if (order == 2)
        {
            double atBrkPt = knots[brkPoint];
            omega = (u - atBrkPt) / (knots[brkPoint+1] - atBrkPt);
            bvals[0] = omega;
            bvals[1] = 1-omega;
            if (dvals1 != null)
            {
                dvals1[0] = 1.0;
                dvals1[1] = 0.0;
            }
        }
        else if ((order == 3) && (dvals1 == null) && (dvals2 == null))
        {
            double bfBrkPt = knots[brkPoint-1];
            double atBrkPt = knots[brkPoint];
            double afBrkPt = knots[brkPoint+1];
            bvals[1] = 0.0;
            omega = (u - atBrkPt) / (afBrkPt - atBrkPt);
            bvals[1] += (1-omega) * bvals[0];
            bvals[0] *= omega;
            bvals[2] = 0.0;
            omega = (u - bfBrkPt) / (afBrkPt - bfBrkPt);
            bvals[2] += (1-omega) * bvals[1];
            bvals[1] *= omega;
            omega = (u - atBrkPt) / (knots[brkPoint+2] - atBrkPt);
            bvals[1] += (1-omega) * bvals[0];
            bvals[0] *= omega;
        }
        else if ((order == 4) && (dvals1 == null) && (dvals2 == null))
        {
            double b2BrkPt = knots[brkPoint-2];
            double bfBrkPt = knots[brkPoint-1];
            double atBrkPt = knots[brkPoint];
            double afBrkPt = knots[brkPoint+1];
            double a2BrkPt = knots[brkPoint+2];
            bvals[1] = 0.0;
            omega = (u - atBrkPt) / (afBrkPt - atBrkPt);
            bvals[1] += (1-omega) * bvals[0];
            bvals[0] *= omega;
            bvals[2] = 0.0;
            omega = (u - bfBrkPt) / (afBrkPt - bfBrkPt);
            bvals[2] += (1-omega) * bvals[1];
            bvals[1] *= omega;
            omega = (u - atBrkPt) / (a2BrkPt - atBrkPt);
            bvals[1] += (1-omega) * bvals[0];
            bvals[0] *= omega;
            bvals[3] = 0.0;
            omega = (u - b2BrkPt) / (afBrkPt - b2BrkPt);
            bvals[3] += (1-omega) * bvals[2];
            bvals[2] *= omega;
            omega = (u - bfBrkPt) / (a2BrkPt - bfBrkPt);
            bvals[2] += (1-omega) * bvals[1];
            bvals[1] *= omega;
            omega = (u - atBrkPt) / (knots[brkPoint+3] - atBrkPt);
            bvals[1] += (1-omega) * bvals[0];
            bvals[0] *= omega;
        }
        else
        {
            int s;
            for (int r = 2; r <= order; r++)
            {
                int i = brkPoint - r + 1;
                bvals[r - 1] = 0.0;
                for (s = r-2; s >= 0; s--)
                {
                    i++;
                    /* Stashing the "u - *knotsptr" only slowed it down... */
                    double atBrkPt = knots[i];
                    omega = (u - atBrkPt) / (knots[i + r - 1] - atBrkPt);

                    /* Unfolding this with pointers actually slowed down with MPW C++... */
                    bvals[s + 1] += (1.0 - omega) * bvals[s];
                    bvals[s] *= omega;
                }

                if ((dvals1 != null) && (r == order-1))
                {
                    for (s = 0; s < order; s++)
                        dvals1[s] = (s < r) ? bvals[s] : 0.0;
                }
                if ((dvals2 != null) && (r == order-2))
                {
                    for (s = 0; s < order; s++)
                        dvals2[s] = (s < r) ? bvals[s] : 0.0;
                }
            }
        }

        /*
        // Compute derivatives as well.  This was formed by mashing together
        // the two routines published by Bartels on Usenet.  They could
        // perhaps be rolled together even tighter; the kv denom is an
        // obvious example.  Perhaps the loops could be folded?
        */

        if (dvals1 != null)
            ComputeDeriv(order, brkPoint, 1, knots, dvals1);
        if (dvals2 != null)
            ComputeDeriv(order, brkPoint, 2, knots, dvals2);
    }

    //
    // Find the index of the start of the interval for a given value of u.
    // This is the Killer B's
    //

    public static int FindInterval(        // Locate parametric Interval (Killer B's)
        double u,
        double knots[],
        int numPts,
        int order)
    {
        int i;

        if (u == knots[numPts])     // Allow maximum parameter value
            i = numPts - 1;
        else
        {
            for (i = numPts - 1 + order; (u < knots[i]) && (i > 0); --i)
                ;
        }
        return i ;
    }

    //
    // Given a breakpoint, find the start of the next interval
    //

    public static int FindNextInterval(
        int curBrk,
        double knots[],
        int numPts)
    {
        // Given a kv index, skip to the next interval)
        curBrk++;
        int i;
        for (i = curBrk + 1; (knots[i] == knots[curBrk]) && (i <= numPts); ++i)
            ;

        return i - 1;
    }

    // ---------- Discontinuities ----------

    public static int GetNextDiscon(
        int startIndex,
        int discon,
        int order,
        int numPoints,
        double knots[])
    {
        // InitializedAssert(order, "");
        if (order <= 0)
            throw new IllegalArgumentException("prKnots.GetNextDiscon(..., order, ...)");

        // Sanity check on discon
        // InternalAssert((discon < order), "");
        if (discon >= order)
            throw new IllegalArgumentException("prKnots.GetNextDiscon(..., discon, order, ...)");

        double curU = knots[startIndex];
        int numKnots = order + numPoints;
        int i;

        // Skip past the current index to the next interval
        for (i = startIndex; (i < numKnots) && (curU == knots[i]); ++i)
            ;

        if (i >= numKnots)
            return (numKnots - 1);    // Already hit the end

        // count how many same knots in a row, quitting if it's enough
        int same;
        do {
            curU = knots[i];
            same = 1;
            i++;
            while ((i < numKnots) && (same < order - discon) && (knots[i] == curU))
            {
                i++; same++;
            }
        }
        while ((i < numKnots) && (same < order - discon));

        if (same >= order - discon)
            return i - same;
        else
            return numKnots - 1;
    }


    // Some algorithms (e.g., degree reduction, deformation) can only deal with
    // piecewise Bezier curves.  These must keep track of the c0 discontinuties
    // in the curve's knot vector, so that the control point indexing is correct
    // and the knot vector in the new curve has c0 discon's in the same place.
    // These routines deal with an array of discons that record which Bezier
    // segments have a c0 discontinuity.

    //
    // Create an array indicating which segments in the curve start
    // with a c0 discontinuity.  Assumes curve is piecewise Bezier.
    //

/*
    // the <array>.length field will have the correct length.
    public static int[] MakeBezierDiscons(
        double knots[],
        int order,
        int numPoints)
    {
        // Allocate based on rough-guess worst case
        int approxSize = (numPoints - 1) / (order - 1);
        int temp[] = new int[approxSize];
        temp[0] = -1;                       // end flag

        int d = order - 1;                  // Use order for Beziers...
        int seg = 1;
        int numDiscons = 0;

        while (seg <= (numPoints - numDiscons - 1) / d)
        {
            // check for multiplicity of order
            if (knots[seg * d + numDiscons + 1] == knots[seg * d + numDiscons])
            {
                temp[numDiscons++] = seg - 1;
                temp[numDiscons] = -1;
            }
            seg++;
        }

        int discons[];
        if (approxSize > 0) {
            discons = new int[numDiscons];
            System.arraycopy(temp, 0, discons, 0, numDiscons);
        }
        else
            discons = temp;
        return discons;
    }

    //
    // Given an array discons, create the appropriate knot vector
    //

    public static void CreateKnotsFromDiscons(
        double knots[],
        int order,
        int numPoints,
        int discons[])
    {
        knots[0] = 0.0;             // Hmmm...maybe keep track of the orig. range?
        int numDiscons = 0;
        int d = order - 1;          // Easier to work in degrees for Beziers...
        int seg = 0;

        while ((seg * d + numDiscons) < numPoints)
        {
            for (int i = 1; i <= d; i++)
                knots[d * seg + i + numDiscons] = (double)seg;

            if (seg == discons[numDiscons])
            {
                numDiscons++;
                knots[d * seg + d + numDiscons] = (double)seg;
            }

            seg++;
        }
        knots[d + numPoints] = (double)(seg - 1);
    }

    // check whether knots are sorted in ascending order and multiplicity<=order

    public static boolean IsKnotvectorConsistent(
        double Knots[],
        int numberKnots,
        int order)
    {
        int mul = 1;
        double last = Knots[0];
        boolean ok = true;

        for (int i = 1; i < numberKnots; i++) {
            if (last > Knots[i]) {
                ok = false;
                mul = 1;
            } else if (last == Knots[i]) {
                mul++;
                if (mul > order) {
                    ok = false;
                }
            } else
                mul = 1;
            last = Knots[i];
        }
        return ok;
    }
*/

    // ---------- Refinement ----------

    //
    // Compute an alpha refinement matrix, given an original knot vector oldKnots,
    // a refinement knot vector newKnots, the number numNew of new control points
    // (over the existing numPts).  This uses the Lyche-Morken algorithm.
    // (See "Making the Oslo Algorithm More Efficient", SIAM J. of Numerical
    //  analysis, June 1986, p. 663, Algorithm 1.).
    //

    public static double[][] ComputeAlphaMatrix(
        double oldKnots[],
        double newKnots[],
        int numPts,
        int numNew,
        int order)
    {
        int odelta, delta, ih, il, iu, nu, p;
        double b1, b, d1, d2, tj;
        double z[] = new double[order + 1];
        double ah[] = new double[order + 1];
        int i, j;

        double alpha[][] = new double[order][];
        for (i = 0; i < order; i++)
            alpha[i] = new double[numNew - order];

        /* first calculate the alpha matrix */
        for (j = 0; j < numNew-order; j++)
        {
            i = j + 1;
            odelta = FindInterval(newKnots[j], oldKnots, numPts, order);
            delta = odelta;

            while ((newKnots[i] == oldKnots[delta]) && (i < j + order))
            {
                i++;
                delta--;
            }

            ih = delta + 1;
            nu = 0;

            for (p = 1; p < order; p++)
                if (newKnots[j + p] == oldKnots[ih])
                    ih++;
                else
                {
                    nu++;
                    z[nu] = newKnots[j + p];
                }

            ah[order] = 1.0;
            for (p = 1; p <= nu; p++)
            {
                b1 = 0.0;
                tj = z[p];
                if (p > delta)
                    b1 = (tj - oldKnots[0]) * ah[1 + order - delta]
                        / (oldKnots[p + order - nu] - oldKnots[0]);
                il = Math.max(1, delta - p + 1);
                iu = Math.min(delta, numPts-1 + nu - p);

                for (i = il; i <= iu; i++)
                {
                    d1 = tj - oldKnots[i];
                    d2 = oldKnots[i + p + order - nu - 1] - tj;
                    b = ah[i + order - delta] / (d1 + d2);
                    ah[i + order - delta - 1] = d2 * b + b1;
                    b1 = d1 * b;
                }
                ah[iu+order-delta] = b1;
                if (iu < delta)
                    ah[iu+order-delta] +=
                        (oldKnots[numPts-1+order] - tj) * ah[iu+order-delta+1] /
                        (oldKnots[numPts-1+order] - oldKnots[iu+1]);
            }
            il = Math.max(delta - nu, 0);
            iu = Math.min(delta, numPts - 1);
            ih = Math.max(odelta - (order - 1), 0);

            /* NOTE: The AlphaMul should be taught to know about il and iu for
             * each value of j.  This would save a number of multiplies by zero.
             * [I tried this 11-Mar-93...it ran slower! (heap overhead)   -jp]
             */
            for (i = 0; i< order; i++)
                alpha[i][j] = 0.0;
            for (i = il; i<= iu; i++)
                alpha[i-ih][j] = ah[i+order-delta];
        }

        return alpha;
    }

/*
    public static void RefineKVUniform(
        int fOrder,
        int fNumPoints,
        double fKnots[],
        TGParametricArray newKnots)
    {
        int numberKnots = 2*fNumPoints + 1;
        TGParametricArray tmpKnots = new TGParametricArray(numberKnots);
        newKnots.swap(tmpKnots);

        double dst[] = newKnots.getArray();

        // knots at beginning of knotvector stay same
        System.arraycopy(fKnots, 0, dst, 0, fOrder);

        int o = fOrder;             // index into new Knots

        // new knots halfway between existing ones
        for (int i = fOrder; i <= fNumPoints; i++)
        {
            if (fKnots[i] != fKnots[i-1])
                dst[o++] = (fKnots[i-1] + fKnots[i]) / 2.0;
            dst[o++] = fKnots[i];
        }

        // knots at end of knotvector stay same
        System.arraycopy(fKnots, fNumPoints + 1, dst, o, fOrder - 1);

        newKnots.resize(o);
    }
*/

    // Returns false if the knot vector was already in Bezier form
    public static void RefineKVToBezier(
        int fOrder,
        int fNumPoints,
        double fKnots[],
        TGParametricArray newKnots)
    {
        int n = fNumPoints + fOrder;
        int i = n + (n - 2 * fOrder) * (fOrder - 1); // Worst case new knots

        TGParametricArray tmpKnots = new TGParametricArray(n + i);
        newKnots.swap(tmpKnots);

        double dst[] = newKnots.getArray();
        dst[0] = fKnots[0];    // Always copy first

        int tmp;
        int j = 1;
        int same = 1;
        for (i = 1; i < n - 1; i++)
        {
            dst[j] = fKnots[i]; // Look for sequences of k-1 fold knots...
            if (fKnots[i] == fKnots[i + 1])
                same++;
            else
            {
                if (same < fOrder - 1)
                {
                    for (tmp = 1; tmp < fOrder - same; tmp++)
                    {
                        j++;
                        dst[j] = fKnots[i];
                    }
                }
                same = 1;
            }
            j++;
        }
        dst[j++] = fKnots[n - 1];
        newKnots.resize(j);
    }

    public static void RefineKVToPinned(
        int fOrder,
        int fNumPoints,
        TGParametricArray fKnots,
        TGParametricArray newKnots)
    {
        double src[] = fKnots.getArray();
        double dst[] = newKnots.getArray();
        int i;
        for (i = 0; i < fOrder; i++)
            dst[i] = src[fOrder - 1];
        System.arraycopy(src, fOrder, dst, fOrder, fNumPoints - fOrder);
        for (i = fNumPoints; i < (fNumPoints + fOrder); i++)
            dst[i] = src[fNumPoints];
    }

    // ---------- Sets/Tests ----------

    //
    // A couple popular knot vector schemes...
    //
    public static void SetBezierKnots(
        int fOrder,
        int fNumPoints,
        TGParametricArray fKnots)
    {
        /* only if number controlpoints is multiple of order */
        int segments = (fNumPoints-1) / (fOrder-1);
        if (segments * (fOrder-1) == fNumPoints-1)
        {
            double knots[] = fKnots.getArray();
            int ind = 0; /* index into Knots */
            int i;
            for (i = 0; i < fOrder; i++)
                knots[ind++] = 0.0;
            for (i = 0; i < segments-1; i++)
                for (int j = 0; j < fOrder-1; j++)
                    knots[ind++] = i + 1.0;
            for (i = 0; i < fOrder; i++)
                knots[ind++] = segments;
        }
    }

    public static void SetPinnedKnots(
        int fOrder,
        int fNumPoints,
        TGParametricArray fKnots)
    {
        double knots[] = fKnots.getArray();
        int i;
        for (i = 0; i < fOrder; i++)
            knots[i] = 0.0;
        for (/*i = fOrder*/; i < fNumPoints; i++)
            knots[i] = i - fOrder + 1;
        for (/*i = fNumPoints*/; i < fNumPoints + fOrder; i++)
            knots[i] = fNumPoints - fOrder + 1;

    }

    public static boolean IsBezierKnots(
        int fOrder,
        int fNumPoints,
        TGParametricArray fKnots)
    {
        if (!IsPinnedKnots(fOrder, fNumPoints, fKnots))
            return false;   // Bezier implies pinned

        boolean isbez = true;

        double knots[] = fKnots.getArray();
        int i = fOrder;
        while ((i < fNumPoints) && (isbez))
        {
            double curU = knots[i];
            int same = 1;
            i++;
            while ((i < fNumPoints) && (curU == knots[i]))
            {
                i++;
                same++;
            }
            isbez = (same >= fOrder-1); // Must have at least k-1 sequential knots
        }
        return isbez;
    }

    public static boolean IsPinnedKnots(
        int fOrder,
        int fNumPoints,
        TGParametricArray fKnots)
    {
        double knots[] = fKnots.getArray();
        double min = knots[0];
        double max = knots[fNumPoints+fOrder-1];
        boolean pinned = true;
        for (int i = 1; (i < fOrder) && pinned; i++)
            pinned = pinned && (knots[i] == min) && (knots[fNumPoints + i] == max);
        return pinned;
    }

/*
    public static void MoveAndSortKnot(
        int index,
        double u,
        int fOrder,
        int fNumPoints,
        TGParametricArray fKnots)
    {
        // InitializedAssert(fOrder, "");
        if (fOrder <= 0)
            throw new IllegalArgumentException("prKnots.MoveAndSortKnot(..., order, ...)");

        int max = fNumPoints + fOrder - 1;
        // ParameterAssert((index <= max), "");
        if (index > max)
            throw new IllegalArgumentException("prKnots.MoveAndSortKnot(index, ...)");

        double knots[] = fKnots.getArray();
        if (knots[index] == u)     // No brainer
            return;

        // Look through the knot vector to see where the knot beints

        int brkPt = FindInterval(u, knots, fNumPoints, fOrder);
        if (u > knots[max])        // Watch end case
            brkPt = max;

        int i;
        if (index < brkPt)
            for (i = index; i < brkPt; i++)
                knots[i] = knots[i + 1];

        if (index > brkPt)
        {
            for (i = index; i > brkPt; i--)
                knots[i] = knots[i - 1];
            if (brkPt > 0)
                brkPt++;
        }

        knots[brkPt] = u;

        // check to see if they hosed the kv by exceeding valid multiplicity

        while ((brkPt < fOrder + fNumPoints - 1) && (knots[brkPt + 1] == u))
            brkPt++;
        for (i = brkPt; (i>=0) && (i >= brkPt - fOrder) && (knots[i] == u); i--);
        // ParameterAssert((brkPt - i <= fOrder), "Multiplicty > order");
        if (brkPt - i > fOrder)
            throw new IllegalArgumentException("prKnots.MoveAndSortKnot(): Multiplicty > order");
    }
*/

    /*
        Knotvector merging algorithm
        ----------------------------
        Used by makeCompatible routines (in TGCurve and TGCurve3D) and sweep
        routines, when scale and twist knotvectors are merged into trajectory
        knotvector.

        Description:
        This is a merge algorithm (an OR operation) with a few special cases.
        Src and Dst knots are merged into a new knotvector, Result. This is a
        new heap block passed back by the routine.

        Assumptions:
        The two incoming knotvectors Src and Dst have the same range, i.e. their
        first and last knots respectively have the same numeric value.
        the knotvectors are legal, ie there are no breakpoints with
        multiplicity>order
        no other assumptions regarding internal knotvector structure.
        Src and Dst discontinuities have to be preserved. Only c0 and C1
        are preserved,
        as DstOrder and DstOrder-1 multiplicity breakpoints, respectively.

        NOTE: This has been rewritten from scratch to avoid problems with
        infinite loops and glitches due to round-off errors.  This does not
        strictly follow the order constraint above, but as int as the kv's
        have the same range, this shouldn't be a problem.
    */

/*
    // Copy knot vectors to ResultKnots until next breakpoint is reached.
    static void Advance(
        double kv1[],
        double kv2[],
        IntType in1,
        IntType in2,
        int num1,
        int num2,
        TGParametricArray ResultKnots,
        int ResultNumberKnots,
        double tol)
    {
        double knots[] = ResultKnots.getArray();
        int ind1 = in1.fValue;
        int ind2 = in2.fValue;
        int mult1 = 1;
use: javac [-g][-O][-debug][-depend][-nowarn][-verbose][-classpath path][-nowrite][-deprecation][-d dir][-J<runtime flag>] file.java...
            mult1++;
            knots[ResultNumberKnots++] = kv1[ind1++];
        }
        knots[ResultNumberKnots++] = kv1[ind1++];

        int mult2 = 0;
        while ((ind2 < num2) && Compare(kv2[ind2], kv1[ind1 - 1], tol)) { // prSplineRoutines
            mult2++;
            ind2++;
        }

        for (; (mult2 - mult1) > 0; mult2--)
            knots[ResultNumberKnots++] = kv2[ind2 - 1];

        // return the values...
        in1.fValue = ind1;
        in2.fValue = ind2;
    }

    static void privateMergeKnots(
        double SrcKnots[],      // incoming knots to be merged ()
        int SrcNumberKnots,     // number of source knots
        int SrcOrder,           // source order (to determine src discons)
        double DstKnots[],      // destination knotvector
        int DstNumberKnots,     // number of destination knots
        int DstOrder,           // destination order
        TGParametricArray ResultKnots // output knotvector
       )
    {
        // Remove two least significant bits from comparisons
        double maxp = Math.max(SrcKnots[SrcNumberKnots-1], DstKnots[DstNumberKnots-1]);
        double minp = Math.min(SrcKnots[0], DstKnots[0]);
        double tol = (maxp - minp) * kDoubleEpsilon * 4.0; // prNumeric

        // allocate worst case array (no knots match except boundary knots (same range!))
        int worst = (Math.abs(SrcOrder-DstOrder) + 1) * (DstNumberKnots + SrcNumberKnots);
        TGParametricArray junk = new TGParametricArray(worst);
        ResultKnots.swap(junk);
        int ResultNumberKnots = 0;
        IntType curSrc = new IntType(0);
        IntType curDst = new IntType(0);
        while ((curSrc.fValue < SrcNumberKnots) && (curDst.fValue < DstNumberKnots))
        {
            if (SrcKnots[curSrc.fValue] < DstKnots[curDst.fValue])
                Advance(
                    SrcKnots, DstKnots, curSrc, curDst, SrcNumberKnots, DstNumberKnots,
                    ResultKnots, ResultNumberKnots, tol);
            else
                Advance(
                    DstKnots, SrcKnots, curDst, curSrc, DstNumberKnots, SrcNumberKnots,
                    ResultKnots, ResultNumberKnots, tol);
        }

        double knots[] = ResultKnots.getArray();

        // Finish the remainder
        for (; curSrc.fValue < SrcNumberKnots; curSrc.fValue++)
            knots[ResultNumberKnots++] = SrcKnots[curSrc.fValue];

        for (; curDst.fValue < DstNumberKnots; curDst.fValue++)
            knots[ResultNumberKnots++] = DstKnots[curDst.fValue];

        ResultKnots.resize(ResultNumberKnots);
    }
*/
//}


//class prNumerics {

    public static final double kRadiansToDegrees = 180.0 / Math.PI;
    public static final double kDegreesToRadians = 1. / kRadiansToDegrees;

    // copied this from IBMCPPW/include/float.h, DBL_EPSILON
    public static final double kDoubleEpsilon = 2.2204460492503131e-16;

    public static final double Log2(double x) { return Math.log(x) / Math.log(2); }

    public static double Square(double x) { return (x * x); }
    public static double Remainder(double x, double y) {
        return Math.IEEEremainder(x, y);
    }
//}


/*
    Useful routines for TGCurve and other objects. These are strictly private routines
*/

//class prSplineRoutines {

    // Multiply a set of basis functions to a set of control points
    public static void BasisMul(
        int brkPoint,
        int order,
        double bvals[],
        TGRPoint points[],
        TGRPoint p)
    {
        TGRPoint.SetToZero(p);

        for (int i = order - 1; i >= 0; --i) {
            p.addBy(TGRPoint.multiply(points[brkPoint - i], bvals[i]));
        }
    }

/*
    public static void BasisMul(
        int brkPoint,
        int order,
        double bvals[],
        TGPoint points[],
        TGPoint p)
    {
        p.setLocation(0.0, 0.0);

        for (int i = order - 1; i >= 0; --i) {
            p.addBy(TGPoint.multiply(points[brkPoint - i], bvals[i]));
        }
    }
*/

    // Return the geometric bounds of the curve segment starting at knot
    // index brkPoint

    public static TGRect GetSegmentBounds(
        int brkPoint,
        int order,
        TGRPointArray pts,
        boolean ratFlag)
    {
        TGRect segmentBounds;

        // Compute the bounds of the current segment
        TGRPoint p[] = pts.getArray();

        if (ratFlag) {
            segmentBounds = new TGRect(p[brkPoint].divW(), p[brkPoint - 1].divW());
            for (int j = brkPoint - 2; j >= (int)(brkPoint - (order - 1)); j--)
                segmentBounds.extendTo(p[j].divW());
        }
        else {
            segmentBounds = new TGRect(p[brkPoint].dropW(), p[brkPoint - 1].dropW());
            for (int j = brkPoint - 2; j >= (int)(brkPoint - (order - 1)); j--)
                segmentBounds.extendTo(p[j].dropW());
        }

        return segmentBounds;
    }

    //
    // evaluate a point on a NURB curve, using a very efficient method.
    // The code (and the notation) is from the Killer B's, p. 387.  This
    // is used instead of TGCurve[3D]::evaluate when you already know what
    // the breakpoint is, and you have raw pointer access to the curve data.
    //
/*
    public static void NurbEval(
        TGPoint[] V,      // Local control points
        int delta,          // Breakpoint
        double u,           // Parameter to eval. at.
        double knots[],     // Local knot vector
        int order,          // Order of the curve
        boolean RatFlag,    // true if rational w != 1.0
        TGPoint p)        // Output
    {
        double coef[] = new double[order + 1];
        ComputeSplineValues(u, delta, order, knots, coef); // prKnots
        BasisMul(delta, order, coef, V, p);
    }
*/
    public static void NurbEval(
        TGRPoint[] V,     // Local control points
        int delta,          // Breakpoint
        double u,           // Parameter to eval. at.
        double knots[],     // Local knot vector
        int order,          // Order of the curve
        boolean RatFlag,    // true if rational w != 1.0
        TGRPoint p)        // Output
    {
        double coef[] = new double[order + 1];
        ComputeSplineValues(u, delta, order, knots, coef); // prKnots
        BasisMul(delta, order, coef, V, p);
    }


    // For "matrix" cubic spline computation

    // Duplicate a tiny bit of code here rather than dragging in all the 3D
    // transform stuff.
/*
typedef GParametric BasisMatrix[16]; // Internal type for matrix cubic spline
*/
/*
    public static void CubicBasisMatrixMul(
        double[] a,  // const BasisMatrix a,
        double[] b,  // const BasisMatrix b,
        double[] dest) // BasisMatrix dest)
    {
        int i, j, k;
        for (i = 0; i < 4; i++)
            for (j = 0; j < 4; j++)
            {
                dest[i * 4 + j] = 0.0;
                for (k = 0; k < 4; k++)
                    dest[i * 4 + j] += a[i * 4 + k] * b[k * 4 + j];
            }
    }
*/

    //
    // Find distance from point to line -segment- measuring distances
    // to endpoints if needed
    //

/*
    public static double PointSegDist(
        TGPoint p0,
        TGPoint p1,
        TGPoint test)
    {
        double t;

        // Find two vectors, p0p1 & p0test
        TGPoint v0 = TGPoint.subtract(p1, p0);
        TGPoint vm = TGPoint.subtract(test, p0);

        // Ratio of dot prodcuts gives the projection of vm onto v0
        double denom = v0.dotProduct(v0);
        if (denom == 0.0)
            t = 0.0;
        else
            t = v0.dotProduct(vm) / denom;

        // if t < 0 or t > 1, then measure distance to the nearest endpoint
        if (t < 0)
             return 0.0;
        else if (t > 1)
             return 1.0;
        else
            return t;
    }

    public static double PointSegDist(
        TGPoint p0,
        TGPoint p1,
        TGPoint test,
        DoubleType dist)
    {
        double t;

        // Find two vectors, p0p1 & p0test
        TGPoint v0 = TGPoint.subtract(p1, p0);
        TGPoint vm = TGPoint.subtract(test, p0);

        // Ratio of dot prodcuts gives the projection of vm onto v0
        double denom = v0.dotProduct(v0);
        if (denom == 0.0)
            t = 0.0;
        else
            t = v0.dotProduct(vm) / denom;

        TGPoint p;
        // if t < 0 or t > 1, then measure distance to the nearest endpoint
        if (t < 0)
        {
            p = TGPoint.subtract(p0, test);
            dist.fValue = p.vectorLength();
            return 0.0;
        }
        else if (t > 1)
        {
            p = TGPoint.subtract(p1, test);
            dist.fValue = p.vectorLength();
            return 1.0;
        }
        else
        {
            // Distance to line is point on v0 corresponding to t minus testpoint
            p = TGPoint.multiply(v0, t);
            p.addBy(p0);
            p.subtractBy(test);
            dist.fValue = p.vectorLength();
            return t;
        }
    }
*/

    //
    // This is a floating point compare that truncates the least
    // significat bits.  LSB are likely to be corrupted when knot
    // vectors are merged, due to rescaling operations.  By lopping
    // the bits off, we greatly reduce the chances of introducing
    // unintentional discontinuities.  Tolerance should be
    // set to (maxKnot - minKnot) * kDoubleEpsilon * 4
    //
    public static boolean Compare(double a, double b, double tolerance)
    {
        return (Math.abs(a - b) < tolerance);
    }

//
// These basis conversion matrices are shared by the degree reduction code
// and the Arc length reparameterization code.  They convert from Chebyshev
// polynomials to Bezier.
//

/*
extern double *** ChebToBez;
extern void MakeReductionMatrices(int order);

    // Combitorial function.  This belongs in a math library someplace
    // also include the math-orphans at the end of BaseGeometry.java

    public static double fchoose(int d, int i)
    {
        int j;

        int a = 1;
        for (j = i + 1; j <= d; j++)
            a *= j;
        for (j = 1; j <= d - i; j++)
            a /= j;
        return (double)a;
    }
*/

}



