/*
 * @(#)CustomCursor.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import java.awt.*;
import java.awt.image.*;

/**
 * A class to encapsulate a custom image-based cursor.  
 *
 * @version 	1.4, 09/21/98
 * @author 	ThomasBall
 */
public abstract class CustomCursor extends Cursor {

    protected Image image;

    public CustomCursor(Image cursor, Point hotSpot, String name) 
            throws IndexOutOfBoundsException {
        super(name);
        image = cursor;
        Toolkit toolkit = Toolkit.getDefaultToolkit();

        // Make sure image is fully loaded.
        Component c = new Canvas(); // for its imageUpdate method
        MediaTracker tracker = new MediaTracker(c);
        tracker.addImage(cursor, 0);
        try {
            tracker.waitForAll();
        } catch (InterruptedException e) {
        }
        int width = cursor.getWidth(c);
        int height = cursor.getHeight(c);

        // Scale image to nearest supported size.
        Dimension nativeSize = toolkit.getBestCursorSize(width, height);
        if (nativeSize.width != width || nativeSize.height != height) {
            cursor = cursor.getScaledInstance(nativeSize.width, 
                                              nativeSize.height,
                                              Image.SCALE_DEFAULT);
            width = nativeSize.width;
            height = nativeSize.height;
        }
        
        // Verify that the hotspot is within cursor bounds.
        if (hotSpot.x >= width || hotSpot.y >= height) {
            throw new IndexOutOfBoundsException("invalid hotSpot");
        }

        /* Extract ARGB array from image.
         *
         * A transparency mask can be created in native code by checking
         * each pixel's top byte -- a 0 value means the pixel's transparent.
         * Since each platform's format of the bitmap and mask are likely to
         * be different, their creation shouldn't be here.
         */
	int[] pixels = new int[width * height];
	ImageProducer ip = cursor.getSource();
	ip = new FilteredImageSource(ip, getImageFilter());
	PixelGrabber pg = new PixelGrabber(ip, 0, 0, width, height,
	                                   pixels, 0, width);
	try {
	    pg.grabPixels();
            createNativeCursor(pixels, width, height, hotSpot.x, hotSpot.y);
	} catch (InterruptedException e) {
	}
    }
    
    protected abstract void createNativeCursor(int[] pixels, 
                                               int width, int height,
                                               int xHotSpot, int yHotSpot);
    
    /* Get a platform-dependent FilteredImageSource, which will modify 
     * the image's colors as needed.  For example, the filter may convert
     * to a palette of three colors: black, white, and transparent (not 
     * that I have a clue how to do this properly, of course :-).
     */
    protected ImageFilter getImageFilter() {
        // default black/white/transparent implementation
        return new DefaultImageFilter();
    }
    
    // Return black, white or transparent pixels
    private static class DefaultImageFilter extends RGBImageFilter {
        DefaultImageFilter() {
            canFilterIndexColorModel = true;
        }
        public int filterRGB(int x, int y, int rgb) {
            if ((rgb & 0xff000000) == 0) {
            	return 0;  // transparent, no alpha
            }

	    double NTSCgray =  0.30 * ((rgb >> 16) & 0xff) +
                               0.59 * ((rgb >> 8) & 0xff) +
                               0.11 * (rgb & 0xff);

            return (NTSCgray > 128.0) ? 0xffffffff : 0xff000000;  // full alpha
        }
    }
}
