/*
 * @(#)EmbeddedFrame.java	1.10 98/08/16
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import java.awt.*;
import java.awt.event.WindowEvent;
import java.awt.peer.ComponentPeer;

/**
 * A generic container used for embedding Java components, usually applets.
 * An EmbeddedFrame has two related uses:
 *
 * . Within a Java-based application, an EmbeddedFrame serves as a sort of
 *   firewall, preventing the contained components or applets from using
 *   getParent() to find parent components, such as menubars.
 *
 * . Within a C-based application, an EmbeddedFrame contains a window handle
 *   which was created by the application, which serves as the top-level
 *   Java window.  EmbeddedFrames created for this purpose are passed-in a
 *   handle of an existing window created by the application.  The window
 *   handle should be of the appropriate native type for a specific
 *   platform, as stored in the pData field of the ComponentPeer.
 *
 * @version 	1.10, 08/16/98
 * @author 	Thomas Ball
 */
public abstract class EmbeddedFrame extends Frame {

    private static final String base = "embeddedFrame";
    private static int nameCounter = 0;
    private boolean isCursorAllowed = false;

    // JDK 1.1 compatibility
    private static final long serialVersionUID = 2967042741780317130L;

    protected EmbeddedFrame() {
	this(0);
    }

    protected EmbeddedFrame(int handle) {
    }

    /**
     * Block introspection of a parent window by this child.
     */
    public Container getParent() {
        return null;
    }

    /**
     * Block modifying any frame attributes, since they aren't applicable
     * for EmbeddedFrames.
     */
    public void setTitle(String title) {}
    public void setIconImage(Image image) {}
    public void setMenuBar(MenuBar mb) {}
    public void setResizable(boolean resizable) {}
    public void remove(MenuComponent m) {}

    public boolean isResizable() {
	return false;
    }

    public void addNotify() {
	int ncomponents = this.getComponentCount();
	for (int i = 0 ; i < ncomponents ; i++) {
	    getComponent(i).addNotify();
	}
    }

    // Need a native call to circumvent Component.peer not being public.
    protected native void setPeer(ComponentPeer p);

    /**
     * Sets whether or not Frame.setCursor works for the embedded frame
     * (currently only used in Win32)
     */
    public void setCursorAllowed(boolean isAllowed) {
	isCursorAllowed = isAllowed;
    }

    /**
     * Is Frame.setCursor allowed for the embedded frame
     * (currently only used in Win32)
     */
    public boolean isCursorAllowed() {
	return isCursorAllowedImpl();
    }

    /*
     * This is called by the native code, so client code can't
     * be called on the toolkit thread.
     */
    final boolean isCursorAllowedImpl() {
	return isCursorAllowed;
    }
} // class EmbeddedFrame
