/*
 * @(#)PlatformFont.java	1.34 98/09/11
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt;

import java.awt.Font;
import java.awt.peer.FontPeer;
import java.util.Properties;
import java.util.Hashtable;
import java.util.Vector;
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.PrintStream;
import sun.io.CharToByteConverter;
import sun.io.CharacterEncoding;
import sun.io.MalformedInputException;
import sun.io.ConversionBufferFullException;
import sun.io.UnknownCharacterException;

public abstract class PlatformFont implements FontPeer {

    static {
        NativeLibLoader.loadLibraries();
        initIDs();
    }

    protected FontDescriptor[] componentFonts;
    protected char defaultChar;
    protected Properties props;

    protected FontDescriptor defaultFont;

    protected static Hashtable charsetRegistry = new Hashtable(5);
    protected String aliasName;
    protected String styleString;

    private Object[] fontCache;

    // Maybe this should be a property that is set based
    // on the locale?
    protected static int FONTCACHESIZE = 256;
    protected static int FONTCACHEMASK = PlatformFont.FONTCACHESIZE - 1;
    protected static String osVersion;

    public PlatformFont(String name, int style){

	if (fprops == null){
	    // there is no default font properties
	    this.props = null;
	    return;
	}

	this.props = fprops;

	aliasName = props.getProperty("alias" + "." +
					     name.toLowerCase());
	if (aliasName == null)
	    aliasName = name.toLowerCase();

	// check this name is correct or not
	if ((props.getProperty(aliasName + ".0") == null) &&
	    (props.getProperty(aliasName + ".plain.0") == null))
	    aliasName = "sansserif";

	styleString = styleStr(style);

	Vector compFonts = new Vector(5);

	int numOfFonts = 0;
	for (;; numOfFonts++){
	    String index = String.valueOf(numOfFonts);

	    // search native font name
	    //
	    String nativeName = props.getProperty
		(aliasName + "." + styleString + "." + index);
	    if (nativeName == null){
		nativeName = props.getProperty(aliasName + "." + index);
		if (nativeName == null){
		    break;
		}
	    }

	    // search font charset
	    //
	    String fcName = props.getProperty
		("fontcharset." + aliasName + "." + styleString + "." + index);
	    if (fcName == null){
		fcName = props.getProperty
		    ("fontcharset." + aliasName + "." + index);
		if (fcName == null){
		    fcName = "default";
		}
	    }

	    CharToByteConverter
		fontCharset = getFontCharset(fcName.trim(), nativeName);

	    // search exclusion range for this font
	    //
	    String exString = props.getProperty
		("exclusion." + aliasName + "." + styleString +  "." + index);
	    if (exString == null){
		exString = props.getProperty
		    ("exclusion." + aliasName + "." + index);
		if (exString == null){
		    exString = "none";
		}
	    }

	    int[] exRange;
	    if (exString.equals("none")){
		exRange = new int[0];
	    } else {
		/*
		 * range format is xxxx-XXXX,yyyy-YYYY,.....
		 */
		int numRange = 1, idx = 0;
		for (;; numRange++){
		    idx = exString.indexOf(',', idx);
		    if (idx == -1){
			break;
		    }
		    idx++;
		}
		exRange = new int[numRange];
		for (int j = 0; j < numRange; j++){
		    String lower, upper;
		    int lo = 0, up =0;
		    try {
			lower = exString.substring(j * 10, j * 10 + 4);
			upper = exString.substring(j * 10 + 5, j * 10 + 9);
		    } catch(StringIndexOutOfBoundsException e){
			exRange = new int[0];
			break;
		    }
		    try {
			lo = Integer.parseInt(lower, 16);
			up = Integer.parseInt(upper, 16);
		    } catch(NumberFormatException e){
			exRange = new int[0];
			break;
		    }
		    exRange[j] = lo << 16 | up;
		}
	    }

	    compFonts.addElement(new FontDescriptor(nativeName,
						    fontCharset,
						    exRange));
	}

	componentFonts = new FontDescriptor[numOfFonts];
	for (int i = 0; i < numOfFonts; i++){
	    componentFonts[i] = (FontDescriptor)compFonts.elementAt(i);
	}

	// search default character
	//
	int dfChar;
	try {
	    dfChar = Integer.parseInt
		(props.getProperty("default.char", "003f"), 16);
	} catch (NumberFormatException e){
	    dfChar = 0x3f;
	}

	defaultChar = 0x3f;
	if (componentFonts.length > 0)
	    defaultFont = componentFonts[0];

	for (int i = 0; i < componentFonts.length; i++){
	    if (componentFonts[i].isExcluded((char)dfChar)){
		continue;
	    }

	    if (componentFonts[i].fontCharset.canConvert((char)dfChar)){
		defaultFont = componentFonts[i];
		defaultChar = (char)dfChar;
		break;
	    }
	}
    }

    /*
     * make default font properties.
     */
    private static Properties fprops;

    static {
	java.security.AccessController.doPrivileged(
	    new java.security.PrivilegedAction() {
	    public Object run() {
		initializeProps();
		return null;
	    }
	});
    }

    private static File tryOpeningFontProp(
            File f, String homedir,
            String language, String ext) {
        if (f != null) {
            return f;       // already validated
        }

        String filename = homedir + File.separator
                            + "lib" + File.separator
                            + "font.properties";
        if (language != null) {
            filename += "." + language;
            if (ext != null) {
                filename += "_" + ext;
            }
        }
        // Try the os-specific prop file first
        File propsFile = new File(filename+"."+osVersion);
        if ((propsFile != null) && propsFile.canRead()) {
            return propsFile;
        }

        propsFile = new File(filename);
        if ((propsFile != null) && propsFile.canRead()) {
            return propsFile;
        }
        
        return null;
    }


    private static void initializeProps() {
	// initialize fprops

	// Find property file
	String jhome = System.getProperty("java.home");
	String uhome = System.getProperty("user.home");
	if (jhome == null){
	    throw new Error("java.home property not set");
	}
	String language = System.getProperty("user.language", "en");
	String region = System.getProperty("user.region");

        // Translate the raw encoding name returned by the VM to the canonical
        // name from the alias table in CharacterEncoding. Map unlisted raw
        // encoding names to themselves. - liu 8/10/98 bug 4163038
	String rawEncoding = System.getProperty("file.encoding");
        String encoding = CharacterEncoding.aliasName(rawEncoding);
        if (encoding == null) {
            encoding = rawEncoding;
        }

        if (osVersion == null) {
            osVersion = System.getProperty("os.version");
        }

	try {
	    File f = null;

            if (region != null) {
                f = tryOpeningFontProp(
                        f, uhome, language, region + "_" + encoding);
                f = tryOpeningFontProp(
                        f, jhome, language, region + "_" + encoding);
                f = tryOpeningFontProp(f, uhome, language, region);
                f = tryOpeningFontProp(f, jhome, language, region);
            }

            f = tryOpeningFontProp(f, uhome, language, encoding);
            f = tryOpeningFontProp(f, jhome, language, encoding);
            f = tryOpeningFontProp(f, uhome, language, null);
            f = tryOpeningFontProp(f, jhome, language, null);
            f = tryOpeningFontProp(f, uhome, encoding, null);
            f = tryOpeningFontProp(f, jhome, encoding, null);
            f = tryOpeningFontProp(f, uhome, null, null);
            f = tryOpeningFontProp(f, jhome, null, null);

	    // set default props to prevent crashing
	    // with corrupted font.properties
	    Properties defaultProps = new Properties();
	    defaultProps.put("serif.0", "unknown");
	    defaultProps.put("sansserif.0", "unknown");
	    defaultProps.put("monospaced.0", "unknown");
	    defaultProps.put("dialog.0", "unknown");
	    defaultProps.put("dialoginput.0", "unknown");

	    fprops = new Properties(defaultProps);
	    // Load property file
	    InputStream in =
		new BufferedInputStream(new FileInputStream(f.getPath()));
	    fprops.load(in);
	    in.close();
	} catch (Exception e){
	}
    }


    /**
     * make a array of CharsetString with given String.
     */
    public CharsetString[] makeMultiCharsetString(String str){
	return makeMultiCharsetString(str.toCharArray(), 0, str.length());
    }

    /**
     * make a array of CharsetString with given char array.
     * @param str The char array to convert.
     * @param offset offset of first character of interest
     * @param len number of characters to convert
     */
    public CharsetString[] makeMultiCharsetString(char str[], int offset, int len) {

	if (len < 1) {
	    return new CharsetString[0];
	}
	Vector mcs = null;
	char[] tmpStr = new char[len];
	char tmpChar = defaultChar;

	FontDescriptor currentFont = defaultFont;


	for (int i = 0; i < componentFonts.length; i++) {
	    if (componentFonts[i].isExcluded(str[offset])){
		continue;
	    }

	    if (componentFonts[i].fontCharset.canConvert(str[offset])){
		currentFont = componentFonts[i];
		tmpChar = str[offset];
		break;
	    }
	}
	tmpStr[0] = tmpChar;

	int lastIndex = 0;
	for (int i = 1; i < len; i++){
	    char ch = str[offset + i];
	    FontDescriptor fd = defaultFont;
	    tmpChar = defaultChar;
	    for (int j = 0; j < componentFonts.length; j++){
		if (componentFonts[j].isExcluded(ch)){
		    continue;
		}

		if (componentFonts[j].fontCharset.canConvert(ch)){
		    fd = componentFonts[j];
		    tmpChar = ch;
		    break;
		}
	    }
	    tmpStr[i] = tmpChar;
	    if (currentFont != fd){
		if (mcs == null) {
		    mcs = new Vector(3);
		}
		mcs.addElement(new CharsetString(tmpStr, lastIndex,
						 i-lastIndex, currentFont));
		currentFont = fd;
		fd = defaultFont;
		lastIndex = i;
	    }
	}
	CharsetString[] result;
	CharsetString cs = new CharsetString(tmpStr, lastIndex,
					    len-lastIndex, currentFont);
	if (mcs == null) {
	    result = new CharsetString[1];
	    result[0] = cs;
	} else {
	    mcs.addElement(cs);
	    result = new CharsetString[mcs.size()];
	    for (int i = 0; i < mcs.size(); i++){
		result[i] = (CharsetString)mcs.elementAt(i);
	    }
	}
	return result;
    }

    /**
     * Is it possible that this font's metrics require the multi-font calls?
     * This might be true, for example, if the font supports kerning.
    **/
    public boolean mightHaveMultiFontMetrics() {
	return props != null;
    }

    /**
     * Specialized fast path string conversion for AWT.
     */
    public Object[] makeConvertedMultiFontString(String str)
        throws MalformedInputException,ConversionBufferFullException {
        return makeConvertedMultiFontChars(str.toCharArray(),0,str.length());
    }

    public Object[] makeConvertedMultiFontChars(char[] data,
						int start, int len)
    throws MalformedInputException,ConversionBufferFullException
    {
        Object[] result = new Object[2];
        Object[] workingCache;
        byte[] convertedData = null;
        int stringIndex = start;
        int convertedDataIndex = 0;
        int resultIndex = 0;
        int cacheIndex;
        FontDescriptor currentFontDescriptor = null;
        FontDescriptor lastFontDescriptor = null;
        char currentDefaultChar;
        PlatformFontCache theChar;

        // Simple bounds check
        int end = start + len;
        if (start < 0 || end > data.length) {
            throw new ArrayIndexOutOfBoundsException();
        }

        if(stringIndex >= end) {
            return null;
        }

        // coversion loop
        while(stringIndex < end)
        {
            currentDefaultChar = data[stringIndex];

            // Note that cache sizes must be a power of two!
            cacheIndex = (int)(currentDefaultChar & this.FONTCACHEMASK);

            theChar = (PlatformFontCache)getFontCache()[cacheIndex];

            // Is the unicode char we want cached?
            if(theChar == null || theChar.uniChar != currentDefaultChar)
            {
                /* find a converter that can convert the current character */
                currentFontDescriptor = defaultFont;
                currentDefaultChar = defaultChar;
                char ch = (char)data[stringIndex];
                int componentCount = componentFonts.length;

                for (int j = 0; j < componentCount; j++) {

                    FontDescriptor fontDescriptor = componentFonts[j];

                    fontDescriptor.fontCharset.reset();
                    fontDescriptor.fontCharset.setSubstitutionMode(false);

                    if (fontDescriptor.isExcluded(ch)) {
                        continue;
                    }
                    if (fontDescriptor.fontCharset.canConvert(ch)) {
                        currentFontDescriptor = fontDescriptor;
                        currentDefaultChar = ch;
                        break;
                    }
                }
                try {
                    char[] input = new char[1];
                    input[0] = currentDefaultChar;

                    theChar = new PlatformFontCache();

                    theChar.encodingSize =
                        (byte)currentFontDescriptor.fontCharset.convert(input, 0, 1, theChar.encoding, 0, 3);
                    theChar.fontDescriptor = currentFontDescriptor;
                    theChar.uniChar = data[stringIndex];
                    getFontCache()[cacheIndex] = theChar;
                } catch(Exception e){

                    // Should never happen!
                    System.err.println(e);
                    e.printStackTrace();

                    return null;
                }
            }

            // Check to see if we've changed fonts.
            if(lastFontDescriptor != theChar.fontDescriptor) {
                if(lastFontDescriptor != null) {
                    result[resultIndex++] = lastFontDescriptor;
                    result[resultIndex++] = convertedData;
		    //  Add the size to the converted data field.
		    if(convertedData != null) {
			convertedDataIndex -= 4;
			convertedData[0] = (byte)(convertedDataIndex >> 24);
			convertedData[1] = (byte)(convertedDataIndex >> 16);
			convertedData[2] = (byte)(convertedDataIndex >> 8);
			convertedData[3] = (byte)convertedDataIndex;
		    }

                    if(resultIndex >= result.length) {
                        Object[] newResult = new Object[result.length * 2];

                        System.arraycopy(result, 0, newResult, 0,
					 result.length);
                        result = newResult;
                    }
                }

                convertedData = new byte[(end - stringIndex + 1) *
                                        theChar.fontDescriptor.fontCharset.getMaxBytesPerChar()
                                        + 4];

                convertedDataIndex = 4;

                lastFontDescriptor = theChar.fontDescriptor;
            }

            if(theChar.encodingSize == 1)
                convertedData[convertedDataIndex++] = theChar.encoding[0];
            else if(theChar.encodingSize == 2)
            {
                convertedData[convertedDataIndex++] = theChar.encoding[0];
                convertedData[convertedDataIndex++] = theChar.encoding[1];
            }
            else if(theChar.encodingSize == 3)
            {
                convertedData[convertedDataIndex++] = theChar.encoding[0];
                convertedData[convertedDataIndex++] = theChar.encoding[1];
                convertedData[convertedDataIndex++] = theChar.encoding[2];
            }
            else if(theChar.encodingSize == 4)
            {
	        convertedData[convertedDataIndex++] = theChar.encoding[0];
                convertedData[convertedDataIndex++] = theChar.encoding[1];
                convertedData[convertedDataIndex++] = theChar.encoding[2];
                convertedData[convertedDataIndex++] = theChar.encoding[3];
            }
            stringIndex++;
        }

        result[resultIndex++] = lastFontDescriptor;
        result[resultIndex++] = convertedData;

        //  Add the size to the converted data field.
        if(convertedData != null) {
            convertedDataIndex -= 4;
            convertedData[0] = (byte)(convertedDataIndex >> 24);
            convertedData[1] = (byte)(convertedDataIndex >> 16);
            convertedData[2] = (byte)(convertedDataIndex >> 8);
            convertedData[3] = (byte)convertedDataIndex;
        }
        return result;
    }

    protected abstract CharToByteConverter getFontCharset(String charsetName,
							   String fontName);

    /*
     * Create fontCache on demand instead of during construction to
     * reduce overall memory consumption.
     *
     * This method is declared final so that its code can be inlined
     * by the compiler.
     */
    protected final Object[] getFontCache() {
        // This method is not MT-safe by design. Since this is just a
        // cache anyways, it's okay if we occasionally allocate the array
        // twice or return an array which will be dereferenced and gced
        // right away.
        if (fontCache == null) {
	    fontCache = new Object[this.FONTCACHESIZE];
	}

	return fontCache;
    }

    /*
     * return String representation of style
     */
    public static String styleStr(int num){
	switch(num){
	  case Font.BOLD:
	    return "bold";
	  case Font.ITALIC:
	    return "italic";
	  case Font.ITALIC+Font.BOLD:
	    return "bolditalic";
	  default:
	    return "plain";
	}
    }

    /**
     * Initialize JNI field and method IDs
     */
    private static native void initIDs();

    class PlatformFontCache
    {
        char uniChar;
        FontDescriptor fontDescriptor;
        byte encodingSize;
        byte[] encoding = new byte[4];
    }
}

