/*
 * @(#)ExtendedTextSourceLabel.java	1.6 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)ExtendedTextSourceLabel.java	1.6 98/09/21
 *
 * (C) Copyright IBM Corp. 1998 - All Rights Reserved
 */

package sun.awt.font;

import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Shape;

import java.awt.font.FontRenderContext;
import java.awt.font.GlyphMetrics;
import java.awt.font.GlyphVector;
import java.awt.font.LineMetrics;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;

import sun.awt.font.NativeFontWrapper;

/**
 * Default implementation of ExtendedTextLabel.
 */

// !!! this implementation does not perform indic reordering

public class ExtendedTextSourceLabel extends ExtendedTextLabel {
  TextSource source;

  // caches
  Rectangle2D lb;
  Rectangle2D ab;
  Rectangle2D vb;
  GlyphVector gv;
  String str;
  Segment shapedSeg;
  float[] charinfo;

  /**
   * Create from a TextSource.
   */
  public ExtendedTextSourceLabel(TextSource source) {
    this.source = source;
  }

  /**
   * Create from a TextSource, optionally using cached data from oldLabel starting at the offset.
   * If present oldLabel must have been created from a run of text that includes the text used in
   * the new label.  Start in source corresponds to logical character offset in oldLabel.
   */
  public ExtendedTextSourceLabel(TextSource source, ExtendedTextSourceLabel oldLabel, int offset) {
    // currently no optimization.
    this.source = source;
  }

  // TextLabel API

  public Rectangle2D getLogicalBounds(float x, float y) {
    if (lb == null) {
      lb = createLogicalBounds();
    }
    return new Rectangle2D.Float((float)(lb.getX() + x), 
				 (float)(lb.getY() + y), 
				 (float)lb.getWidth(), 
				 (float)lb.getHeight());
  }

  public Rectangle2D getVisualBounds(float x, float y) {
    if (vb == null) {
      vb = createVisualBounds();
    }
    return new Rectangle2D.Float((float)(vb.getX() + x), 
				 (float)(vb.getY() + y), 
				 (float)vb.getWidth(), 
				 (float)vb.getHeight());
  }

  public Rectangle2D getAlignBounds(float x, float y) {
    if (ab == null) {
      ab = createAlignBounds();
    }
    return new Rectangle2D.Float((float)(ab.getX() + x), 
				 (float)(ab.getY() + y), 
				 (float)ab.getWidth(),
				 (float)ab.getHeight());

  }

  public Shape getOutline(float x, float y) {
    return getGV().getOutline(x, y);
  }

  public void draw(Graphics2D g, float x, float y) {
    if (gv != null) {
      g.drawGlyphVector(gv, x, y);
    } else if (false && source.dataIsLTR()) {
	  // REMIND: eliminate recursion error
      // !!! Danger Will Robinson !!!
      // !!! stack overflow results from recursion if GlyphVector.cpp
      // has a different definition of what requires bidi or shaping than
      // we do.  So disable this optimization that we went to so much work to
      // enable...  Alternatively, if we just had a way to force graphics
      // to draw the string without analyzing it for bidi...
      Font oldFont = g.getFont();
      g.setFont(source.getFont());
      g.drawString(getString(), x, y);
      g.setFont(oldFont);
    } else {
      g.drawGlyphVector(getGV(), x, y);
    }
  }

  protected Rectangle2D createLogicalBounds() {
    // return getGV().getLogicalBounds();
    // !!! current gv implementation is slow and incorrect
    // this one is just slow

    LineMetrics lm = source.getLineMetrics();

    float[] info = getCharinfo();

    float ll = 0f;
    float lt = -lm.getAscent();
    float lw = 0f;
    float lh = lm.getAscent() + lm.getDescent() + lm.getLeading();
   
    int rn = info.length - numvals;
    while (rn > 0) {
      if (info[rn+advx] != 0) break;
      rn -= numvals;
    }
      
    if (rn >= 0) {
      int ln = 0;
      while (ln < rn) {
	if (info[ln+advx] != 0) break;
	ln += numvals;
      }

      ll = Math.min(0f, info[ln+posx]);
      lw = info[rn+posx] + info[rn+advx] - ll;
    }

    return new Rectangle2D.Float(ll, lt, lw, lh);
  }

  protected Rectangle2D createVisualBounds() {
    return getGV().getVisualBounds();
  }

  // like createLogicalBounds except ignore leading and logically trailing white space
  // this assumes logically trailing whitespace is also visually trailing

  protected Rectangle2D createAlignBounds() {
    LineMetrics lm = source.getLineMetrics();

    float[] info = getCharinfo();

    float al = 0f;
    float at = -lm.getAscent();
    float aw = 0f;
    float ah = lm.getAscent() + lm.getDescent();

    boolean lineIsLTR = source.lineIsLTR();
    int rn = info.length - numvals;
    while (rn > 0 && ((info[rn+advx] == 0) || (lineIsLTR && info[rn+visw] == 0))) {
      rn -= numvals;
    }
      
    if (rn >= 0) {
      int ln = 0;
      while (ln < rn && ((info[ln+advx] == 0) || (!lineIsLTR && info[ln+visw] == 0))) {
	ln += numvals;
      }

      al = Math.max(0f, info[ln+posx]);
      aw = info[rn+posx] + info[rn+advx] - al;
    }

    return new Rectangle2D.Float(al, at, aw, ah);
  }

  private final GlyphVector getGV() {
    if (gv == null) {
      gv = createGV();
    }

    return gv;
  }

  private final String getString() {
    if (str == null) {
      str = createString();
    }
    return str;
  }

  private final Segment getShapedSegment() {
    if (shapedSeg == null) {
      shapedSeg = createShapedSegment();
    }
    return shapedSeg;
  }

  protected Segment createShapedSegment() {
    Segment seg = new Segment();
    Shaping.getShapedSegment(source, seg);
    return seg;
  }

  protected String createString() {
    Segment seg = new Segment(getShapedSegment());
    Shaping.compressSegment(seg, null);

    // sigh, can't construct a glyphvector without exact array, and
    // can't draw an array at a float position in Graphics2D anyway, so can't avoid it...

    return new String(seg.text, seg.start, seg.length);
  }

  protected GlyphVector createGV() {
    Font f = source.getFont();
    FontRenderContext frc = source.getFRC();

    return f.createGlyphVector(frc, getString());
  }

  // ExtendedTextLabel API
  
  private static final int posx = 0,
    posy = 1,
    advx = 2, 
    advy = 3,
    visx = 4,
    visy = 5,
    visw = 6,
    vish = 7;
  private static final int numvals = 8;
      
  public int getNumCharacters() {
    return source.getLength();
  }

  public float getItalicAngle() {
    return source.getFont().getItalicAngle();
  }

  public LineMetrics getLineMetrics() {
    return source.getLineMetrics();
  }

  public float getCharX(int index) {
    validate(index);
    return getCharinfo()[l2v(index) * numvals + posx];
  }
    
  public float getCharY(int index) {
    validate(index);
    return getCharinfo()[l2v(index) * numvals + posy];
  }

  public float getCharAdvance(int index) {
    validate(index);
    return getCharinfo()[l2v(index) * numvals + advx];
  }

  public Rectangle2D getCharVisualBounds(int index, float x, float y) {
    validate(index);
    float[] charinfo = getCharinfo();
    index = l2v(index) * numvals;
    return new Rectangle2D.Float(
      charinfo[index + visx] + x,
      charinfo[index + visy] + y,
      charinfo[index + visw],
      charinfo[index + vish]);
  }

  private void validate(int index) {
    if (index < 0) {
      throw new IllegalArgumentException("index " + index + " < 0");
    } else if (index >= source.getLength()) {
      throw new IllegalArgumentException("index " + index + " < " + source.getLength());
    }
  }

  /*
  public int hitTestChar(float x, float y) {
    // !!! return index of char hit, for swing
    // result is negative for trailing-edge hits
    // no italics so no problem at margins.
    // for now, ignore y since we assume horizontal text

    // find non-combining char origin to right of x 
    float[] charinfo = getCharinfo();

    int n = 0;
    int e = source.getLength();
    while (n < e && charinfo[n + advx] != 0 && charinfo[n + posx] > x) {
      n += numvals;
    }
    float rightx = n < e ? charinfo[n+posx] : charinfo[e - numvals + posx] + charinfo[e - numvals + advx];

    // find non-combining char to left of that char
    n -= numvals;
    while (n >= 0 && charinfo[n+advx] == 0) {
      n -= numvals;
    }
    float leftx = n >= 0 ? charinfo[n+posx] : 0;
    float lefta = n >= 0 ? charinfo[n+advx] : 0;

    n /= numvals;

    boolean left = true;
    if (x < leftx + lefta / 2f) {
      // left of prev char
    } else if (x < (leftx + lefta + rightx) / 2f) {
      // right of prev char
      left = false;
    } else {
      // left of follow char
      n += 1;
    }

    if (!source.dataIsLTR()) {
      n = getNumCharacters() - 1 - n;
      left = !left;
    }

    return left ? n : -n;
  }
  */

  public int logicalToVisual(int logicalIndex) {
    validate(logicalIndex);
    return l2v(logicalIndex);
  }

  public int visualToLogical(int visualIndex) {
    validate(visualIndex);
    return v2l(visualIndex);
  }

  public int getLineBreakIndex(int start, float width) {
    float[] charinfo = getCharinfo();
    int length = source.getLength();
    --start;
    while (width >= 0 && ++start < length) {
      width -= charinfo[l2v(start) * numvals + advx];
    }

    return start; 
  }

  public float getCharAdvanceBetween(int start, int limit) {
    float a = 0f;

    float[] charinfo = getCharinfo();
    --start;
    while (++start < limit) {
      a += charinfo[l2v(start) * numvals + advx];
    }

    return a;
  }

  private final float[] getCharinfo() {
    if (charinfo == null) {
      charinfo = createCharinfo();
    }
    return charinfo;
  }

  protected float[] createCharinfo() {
    Segment seg = getShapedSegment();

    Font f = source.getFont();
    FontRenderContext frc = source.getFRC();
    AffineTransform xf = (AffineTransform)frc.getTransform().clone();
    xf.scale(f.getSize2D(), f.getSize2D());
    xf.preConcatenate(f.getTransform());
    double[] matrix = new double[6];
    xf.getMatrix(matrix);

    float[] result = new float[seg.length * numvals];
    result[0] = (float)matrix[4];
    result[1] = (float)matrix[5];

    NativeFontWrapper.getGlyphInfo(f, seg.text, seg.start, seg.length, 
				   matrix, frc.usesFractionalMetrics(), frc.isAntiAliased(), 
				   result);

    return result;
  }

  protected int l2v(int index) {
    return source.dataIsLTR() ? index : source.getLength() - 1 - index;
  }

  protected int v2l(int index) {
    return source.dataIsLTR() ? index : source.getLength() - 1 - index;
  }

  public String toString() {
    StringBuffer buf = new StringBuffer();
    buf.append(super.toString());
    buf.append("[source:");
    buf.append(source);
    buf.append(", lb:");
    buf.append(lb);
    buf.append(", ab:");
    buf.append(ab);
    buf.append(", vb:");
    buf.append(vb);
    buf.append(", gv:");
    buf.append(gv);
    buf.append(", str:");
    if (str == null) {
      buf.append(str);
    } else {
      buf.append("\"");
      int len = str.length();
      for (int i = 0; i < len; i++) {
	if (i>0) {
	  buf.append(" ");
	}
	buf.append(Integer.toHexString(str.charAt(i)));
      }
      buf.append("\"");
    }
    buf.append(", seg:");
    buf.append(shapedSeg);
    buf.append(", ci:");
    buf.append(charinfo == null ? "null" : Integer.toString(charinfo.length));
    buf.append("]");

    return buf.toString();
  }

  public static ExtendedTextLabel create(TextSource source) {
    return new ExtendedTextSourceLabel(source);
  }
}
