/*
 * @(#)FontNameAliases.java	1.4 98/09/04
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.font;

import java.util.Hashtable;
import java.util.Enumeration;

/* 
 * This is a class to support SunGraphicsEnvironment
 * providing font name compatability with JDK 1.x and previous versions.
 * Ulitmately this class should use the font properties files
*/
public class FontNameAliases {
    public static FontNameAliases oldNameMatcher = new FontNameAliases();

    /*
     * REMIND jk. use type safe enums
     */
    public static final int FONTFORMAT_NONE = -1;   
    public static final int FONTFORMAT_TRUETYPE = 0;   
    public static final int FONTFORMAT_TYPE1	= 1;
    public static final int FONTFORMAT_T2K	= 2;
    public static final int FONTFORMAT_TTC	= 3;
    public static final int FONTFORMAT_COMPOSITE = 4;
    public static final int FONTFORMAT_NATIVE	= 5;

    private static final int TABLESIZE = 8;
    
    class NamePair {
        public String oldName;
        public String newTTName;
        public String newT1Name;
        public boolean caseSens;
        NamePair(String o, boolean c, String tt, String t1) {
            oldName = o;
            caseSens = c;
            newTTName = tt;
            newT1Name = t1;
        }
    }
    
    /* 
     * REMIND jk. Alex needs to make the cache work for old names only
     */
    private NamePair[] builtin;
    //private Hashtable cache;
    
    
    public FontNameAliases() {
        builtin = new NamePair[TABLESIZE];
        //cache = new Hashtable();
        
        builtin[0] = new NamePair("serif", false,
                                  "Times New Roman", "Times Roman");
        builtin[1] = new NamePair("sanserif", false, "Arial", "Helvetica");
        builtin[2] = new NamePair("monospaced", false,
                                  "Courier New", "Courier New");
        builtin[3] = new NamePair("Dialog", false, "sanserif", "sanserif");
        builtin[4] = new NamePair("DialogInput", false,
                                  "monospaced", "monospaced");
        builtin[5] = new NamePair("TimesRoman", true, "serif", "serif");
        builtin[6] = new NamePair("Default", true, "sanserif", "sanserif");
        builtin[7] = new NamePair("Courier", true, "monospaced", "monospaced");
    }
    
    public String ResolveName(String fontName, int fontType) {
        String key = fontName + "$" + fontType + "$";
        String val;
        
        if ( true /*(val = (String) cache.get(key)) == null*/) {
            boolean done = false;
            boolean found = false;
            val = fontName;
            String oldName = fontName;
            int i;
            
            
            while (!done) {
                for (i = 0; i < TABLESIZE; i++) {
                    if (builtin[i].caseSens) {
                        found = builtin[i].oldName.equals(oldName);
                    } else {
                        found = builtin[i].oldName.equalsIgnoreCase(oldName);
                    }
                    if (found) {
                        break;
                    }
                }
                
                if (found) {
                    switch (fontType) {
                    case FONTFORMAT_TRUETYPE:
                    case FONTFORMAT_NONE:
                        val = builtin[i].newTTName;
                        break;
                    case FONTFORMAT_TYPE1:
                        val = builtin[i].newT1Name;
                        break;
                    }
                    oldName = val;
                } else {
                    done = true;
                }
            }
            
            //cache.put(key, val);
        }

        return val;
    }
}


