/*
 * @(#)NativeFontWrapper.java	1.35 98/09/15
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.awt.font;

import sun.java2d.loops.ImageData;
import java.awt.Color;
import java.awt.Shape;
import java.util.Vector;
import java.awt.font.GlyphMetrics;
import java.awt.font.GlyphJustificationInfo;
import java.awt.Font;
import java.awt.font.GlyphVector;

/**
 * NativeFontWrapper: a collection of native methods which perform
 * mutual exlcusion to prevent problems with non-reentrant native
 * code:  REMIND: make native code reentrant
 */

public class NativeFontWrapper {

    static {
	java.security.AccessController.doPrivileged(
		  new sun.security.action.LoadLibraryAction("awt"));
	java.security.AccessController.doPrivileged(
		  new sun.security.action.LoadLibraryAction("fontmanager"));
    }

    public synchronized static native String getFontPath();

    // methods which operate on the font collection
    public synchronized static native void registerFonts(
            Vector fontCollection,
            int size, Vector nameCollection, int fontFormat);
    public synchronized static native void registerCompositeFont(
            String compositeFontName,
            String [] componentFontNames,
            int [] exclusionRanges, int [] exclusionCounts);
    public synchronized static native int getNumFonts();
    public synchronized static native String getFamilyNameByIndex(
            int fontIndex);
    public synchronized static native String getFullNameByIndex(
            int fontIndex);
    public synchronized static native String getFullNameByFileName(
            String fileName);


// device independent font operations -- float returns are in "points"
    // REMIND: may want more state in the native instance
    public synchronized static native void initializeFont(
            Font theFont, String name, int style);
    public synchronized static native String getFamilyName(Font theFont, short lcid);
    public synchronized static native String getFullName(Font theFont, short lcid);
    public synchronized static native int getNumGlyphs(Font theFont);
    public synchronized static native int getMissingGlyphCode(Font theFont);
    public static byte getBaselineFor(Font theFont, char c) {
        return 0;
    }
    public synchronized static native boolean canDisplay(Font theFont, char c);
    public synchronized static native boolean doesGlyphShaping(Font theFont);
    public synchronized static native boolean isStyleSupported(Font theFont,int theStyle);
// end of device independent font operations

// device dependent font operations -- float returns are in "pixels"
    /**
     * returns baselines[]
     */
    public static void getBaseLineOffsetsFor(
            Font theFont, char c,
            double[] matrix,
            float [] baselines)
    {
        float metrics[] = {0, 0, 0, 0};
        getFontMetrics(theFont, matrix, false, false, metrics);
        //REMIND: replace with real implementation
        baselines[0] = 0;
        baselines[1] = (metrics[0] + metrics[1]) / 2 - metrics[0];
        baselines[2] = metrics[1] - metrics[0];
    }

    /**
     * Returns metrics constaining four floats: the ascent, descent, leading, and max advance.
     */
    // REMIND - answers depend on baseline
    public synchronized static native void getFontMetrics(
            Font font,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable,
            float [] metrics);
    public synchronized static native float getItalicAngle(
            Font font,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable);
    // REMIND - answers not entirely useful.  more context needed
    // when factors such as shaping and kerning occur.  It is more
    // meaningful to ask for these answers about an individual
    // member of a collection
    public synchronized static native float getAdvance(
            Font theFont,
            char c,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable);
    // REMIND - answers not entirely useful.  more context needed
    // when factors such as shaping and kerning occur.  It is more
    // meaningful to ask for these answers about an individual
    // member of a collection
    public synchronized static native void getGlyphMetrics(
            Font theFont,
            int glyphCode,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable,
            GlyphMetrics output);

    public synchronized static native void getGlyphInfo(
	    Font font,
            char[] glyphs,
	    int start,
	    int length,
            double[] matrix,
            boolean doAntiAlias,
	    boolean doFractEnable,
	    float[] result);

// end of device dependent font operations

// glyphvector operations
    public synchronized static native void populateGlyphVector(
            Font theFont,
            char [] unicodes,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable,
            GlyphVector target
    );
    public synchronized static native void populateAndLayoutGlyphVector(
            Font theFont,
            char [] unicodes,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable,
            float xpos,
            float ypos,
            GlyphVector target
    );
    public synchronized static native void layoutGlyphVector(
            Font theFont,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable,
            float xpos,
            float ypos,
            GlyphVector target
    );
    // REMIND: implement shaping client.
    // if it is not present, perform maximum compression
    // otherwise, ask the client to approve substitutions, and
    // then give it a bidirectional mapping from src to dst indices
    public synchronized static native void shapeGlyphVector(
            GlyphVector src,
            Font theFont,
            double[] matrix,
            boolean doAntiAlias,
            boolean doFractEnable,
            Object  shapingClient,
            GlyphVector target
    );
// end of glyphvector operations

// drawString operations
    // drawString operations convert supplied characters to glyphs
    // and draw the glyphs.  They will throw an exception if
    // any of the characters supplied is in a script system which
    // *requires* shaping.

    // convert string to shape
    public synchronized static native Shape drawStringOutline(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        boolean doAntiAlias,
        boolean doFractEnable,
        Shape dstShape
    );

    // non-antialiased (opaque) int discrete - to arbitrary surface
    public synchronized static native void drawStringIntDiscrete(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) int discrete - to raster surface
    public synchronized static native void drawStringIntDiscreteRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) short discrete - to arbitrary surface
    public synchronized static native void drawStringShortDiscrete(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) short discrete - to raster surface
    public synchronized static native void drawStringShortDiscreteRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // non-antialiased (opaque) byte discrete - to arbitrary surface
    public synchronized static native void drawStringByteDiscrete(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) byte discrete - to raster surface
    public synchronized static native void drawStringByteDiscreteRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) 3-byte discrete - to arbitrary surface
    public synchronized static native void drawString3Byte(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) int discrete - to raster surface
    public synchronized static native void drawString3ByteRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary xRGB surface
    public synchronized static native void drawStringARGB(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster xRGB surface
    public synchronized static native void drawStringARGBRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary xBGR surface
    public synchronized static native void drawStringBGR(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster xBGR surface
    public synchronized static native void drawStringBGRRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 555 surface
    public synchronized static native void drawString555(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 555 surface
    public synchronized static native void drawString555Raster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 565 surface
    public synchronized static native void drawString565(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 565 surface
    public synchronized static native void drawString565Raster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased byte discrete for arbitrary indexed surface
    public synchronized static native void drawStringIndexed(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased byte discrete for arbitrary index 8  surface
    public synchronized static native void drawStringIndex8Gray(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased byte discrete for arbitrary Byte Gray surface
    public synchronized static native void drawStringByteGray(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased int discrete for arbitrary 24BitRGB surface
    public synchronized static native void drawString24BitRGB(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster 24BitRGB surface
    public synchronized static native void drawString24BitRGBRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary 24BitBGR surface
    public synchronized static native void drawString24BitBGR(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster 24BitBGR surface
    public synchronized static native void drawString24BitBGRRaster(
        String srcData,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

// end of drawString operations

// drawChars operations
    // drawChars operations convert supplied characters to glyphs
    // and draw the glyphs.  They will throw an exception if
    // any of the characters supplied is in a script system which
    // *requires* shaping.

    // non-antialiased (opaque) int discrete - to arbitrary surface
    public synchronized static native void drawCharsIntDiscrete(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) int discrete - to raster surface
    public synchronized static native void drawCharsIntDiscreteRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) short discrete - to arbitrary surface
    public synchronized static native void drawCharsShortDiscrete(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) short discrete - to raster surface
    public synchronized static native void drawCharsShortDiscreteRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // non-antialiased (opaque) byte discrete - to arbitrary surface
    public synchronized static native void drawCharsByteDiscrete(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) byte discrete - to raster surface
    public synchronized static native void drawCharsByteDiscreteRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) 3-byte discrete - to arbitrary surface
    public synchronized static native void drawChars3Byte(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) 3-byte discrete - to raster surface
    public synchronized static native void drawChars3ByteRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary xRGB surface
    public synchronized static native void drawCharsARGB(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster xRGB surface
    public synchronized static native void drawCharsARGBRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary BGR surface
    public synchronized static native void drawCharsBGR(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster BGR surface
    public synchronized static native void drawCharsBGRRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 555 surface
    public synchronized static native void drawChars555(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 555 surface
    public synchronized static native void drawChars555Raster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 565 surface
    public synchronized static native void drawChars565(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 565 surface
    public synchronized static native void drawChars565Raster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased byte discrete for arbitrary indexed surface
    public synchronized static native void drawCharsIndexed(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased byte discrete for arbitrary index8 Gray surface
    public synchronized static native void drawCharsIndex8Gray(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased byte discrete for arbitrary byte_gray surface
    public synchronized static native void drawCharsByteGray(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased int discrete for arbitrary 24BitRGB surface
    public synchronized static native void drawChars24BitRGB(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster 24BitRGB surface
    public synchronized static native void drawChars24BitRGBRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary 24BitBGR surface
    public synchronized static native void drawChars24BitBGR(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster 24BitBGR surface
    public synchronized static native void drawChars24BitBGRRaster(
        char [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );
// end of drawChars operations

// drawBytes operations
    // drawBytes operations convert supplied characters to glyphs
    // and draw the glyphs.  No exception will be thrown since
    // we have defined drawBytes explicity to not do shaping.

    // non-antialiased (opaque) int discrete - to arbitrary surface
    public synchronized static native void drawBytesIntDiscrete(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) int discrete - to raster surface
    public synchronized static native void drawBytesIntDiscreteRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) short discrete - to arbitrary surface
    public synchronized static native void drawBytesShortDiscrete(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) short discrete - to raster surface
    public synchronized static native void drawBytesShortDiscreteRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // non-antialiased (opaque) byte discrete - to arbitrary surface
    public synchronized static native void drawBytesByteDiscrete(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) byte discrete - to raster surface
    public synchronized static native void drawBytesByteDiscreteRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) 3-byte discrete - to arbitrary surface
    public synchronized static native void drawBytes3Byte(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) 3-byte discrete - to raster surface
    public synchronized static native void drawBytes3ByteRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary xRGB surface
    public synchronized static native void drawBytesARGB(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster xRGB surface
    public synchronized static native void drawBytesARGBRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary BGR surface
    public synchronized static native void drawBytesBGR(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster BGR surface
    public synchronized static native void drawBytesBGRRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 555 surface
    public synchronized static native void drawBytes555(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 555 surface
    public synchronized static native void drawBytes555Raster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 565 surface
    public synchronized static native void drawBytes565(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 565 surface
    public synchronized static native void drawBytes565Raster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased byte discrete for arbitrary indexed surface
    public synchronized static native void drawBytesIndexed(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased byte discrete for arbitrary index8 Gray surface
    public synchronized static native void drawBytesIndex8Gray(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

     // antialiased byte discrete for arbitrary byte gray surface
    public synchronized static native void drawBytesByteGray(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased int discrete for arbitrary 24BitRGB surface
    public synchronized static native void drawBytes24BitRGB(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );


   // antialiased int discrete for raster 24BitRGB surface
    public synchronized static native void drawBytes24BitRGBRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary 24BitBGR surface
    public synchronized static native void drawBytes24BitBGR(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster 24BitBGR surface
    public synchronized static native void drawBytes24BitBGRRaster(
        byte [] srcData, int srcOffset, int length,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] matrix,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );
// end of drawBytes operations

// drawGlyphVector operations
    public synchronized static native Shape getGlyphVectorOutline(
            GlyphVector src,
            Font theFont,
            double[] glyphTX,
            boolean doAntiAlias,
            boolean doFractEnable,
            float xpos,
            float ypos,
            Shape theShape
    );
    public synchronized static native Shape getGlyphOutline(
            GlyphVector src,
            Font theFont,
            double[] glyphTX,
            boolean doAntiAlias,
            boolean doFractEnable,
            int glyphIndex,
            Shape theShape
    );
    // non-antialiased (opaque) int discrete - to arbitrary surface
    public synchronized static native void drawGlyphVectorIntDiscrete(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) int discrete - to raster surface
    public synchronized static native void drawGlyphVectorIntDiscreteRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) short discrete - to arbitrary surface
    public synchronized static native void drawGlyphVectorShortDiscrete(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) short discrete - to raster surface
    public synchronized static native void drawGlyphVectorShortDiscreteRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // non-antialiased (opaque) byte discrete - to arbitrary surface
    public synchronized static native void drawGlyphVectorByteDiscrete(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) byte discrete - to raster surface
    public synchronized static native void drawGlyphVectorByteDiscreteRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );
    // non-antialiased (opaque) 3-byte discrete - to arbitrary surface
    public synchronized static native void drawGlyphVector3Byte(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // non-antialiased (opaque) 3-byte discrete - to raster surface
    public synchronized static native void drawGlyphVector3ByteRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary xRGB surface
    public synchronized static native void drawGlyphVectorARGB(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster xRGB surface
    public synchronized static native void drawGlyphVectorARGBRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary BGR surface
    public synchronized static native void drawGlyphVectorBGR(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster BGR surface
    public synchronized static native void drawGlyphVectorBGRRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 555 surface
    public synchronized static native void drawGlyphVector555(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 555 surface
    public synchronized static native void drawGlyphVector555Raster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased short discrete for arbitrary 565 surface
    public synchronized static native void drawGlyphVector565(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased short discrete for raster 565 surface
    public synchronized static native void drawGlyphVector565Raster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr
    );

    // antialiased byte discrete for arbitrary indexed surface
    public synchronized static native void drawGlyphVectorIndexed(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );

    // antialiased byte discrete for arbitrary index8 gray  surface
    public synchronized static native void drawGlyphVectorIndex8Gray(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased byte discrete for arbitrary byte gray surface
    public synchronized static native void drawGlyphVectorByteGray(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for arbitrary 24BitRGB surface
    public synchronized static native void drawGlyphVector24BitRGB(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster 24BitRGB surface
    public synchronized static native void drawGlyphVector24BitRGBRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );

    // antialiased int discrete for arbitrary 24BitBGR surface
    public synchronized static native void drawGlyphVector24BitBGR(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        ImageData dest
    );
    // antialiased int discrete for raster 24BitBGR surface
    public synchronized static native void drawGlyphVector24BitBGRRaster(
        int [] srcGlyphs,
        float [] positionsXY,
        float fX,
        float fY,
        Font fontObject,
        boolean doFractEnable,
        double[] glyphTX, double[] devTX,
        int fgColor,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr
    );
// end of drawGlyphVector operations

// methods to remove -- REMIND - do it
    // REMIND questionable value
    // REMIND need matrix on this call
    public synchronized static native void getGlyphJustificationInfo(
                                                Font theFont,
                                                int glyphCode,
                                                GlyphJustificationInfo gji);
// end of methods to remove -- REMIND - do it
}
