/*
 * @(#)TextSource.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)TextSource.java	1.2 98/09/21
 *
 * (C) Copyright IBM Corp. 1998 - All Rights Reserved
 */

package sun.awt.font;

import java.awt.Font;
import java.awt.font.FontRenderContext;
import java.awt.font.LineMetrics;

/**
 * A text source represents text for rendering, plus context information.
 * All text in the source uses the same font, metrics, and render context,
 * and is at the same bidi level.
 */

public abstract class TextSource {
  /** Source character data. */
  public abstract char[] getChars();

  /** Start of source data in char array returned from getChars. */
  public abstract int getStart();

  /** Length of source data. */
  public abstract int getLength();

  /** Start of context data in char array returned from getChars. */
  public abstract int getContextStart();

  /** Length of context data. */
  public abstract int getContextLength();

  /** Return true if source data is uniformly LTR. */
  public abstract boolean dataIsLTR();

  /** Return true if line the text is to be displayed on is LTR. */
  public abstract boolean lineIsLTR();

  /** Bidi level of all the characters in context. */
  public abstract int getBidiLevel();

  /** Font for source data. */
  public abstract Font getFont();

  /** Font render context to use when measuring or rendering source data. */
  public abstract FontRenderContext getFRC();

  /** Line metrics for source data. */
  public abstract LineMetrics getLineMetrics();
}
