/*
 * @(#)InputMethodDescriptor.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.util.Locale;

class InputMethodDescriptor {
    static final int HOST_ADAPTER = 1;
    static final int IIIMP_ADAPTER = 2;
    static final int JAVA_INPUT_METHOD = 3;

    private int type;
    private Locale locale;
    private String className;
    private String resource;

    InputMethodDescriptor(int type) {
	this(type, Locale.getDefault(), null, null);
    }

    InputMethodDescriptor(int type,
			  Locale locale,
			  String className) {
	this(type, locale, className, null);
    }

    InputMethodDescriptor(int type,
			  Locale locale,
			  String className,
			  String resource) {
	if (!(type == HOST_ADAPTER ||
	      type == IIIMP_ADAPTER ||
	      type == JAVA_INPUT_METHOD))
	  throw new IllegalArgumentException("Invalid input method type value");

	this.type = type;
	this.locale = locale;
	this.className = className;
	this.resource = resource;
    }

    public boolean equals(Object obj) {
	InputMethodDescriptor desc = (InputMethodDescriptor) obj;

	if (this == desc)
	    return true;

	if (this.type != desc.type)
	    return false;

	if (this.className != null) {
	    if (!this.className.equals(desc.className))
		return false;
	} else if (desc.className != null)
	    return false;

	if (this.locale != null) {
	    if (!this.locale.equals(desc.locale))
		return false;
	} else if (desc.locale != null)
	    return false;

	if (this.resource != null) {
	    if (!this.resource.equals(desc.resource))
		return false;
	} else if (desc.resource != null)
	    return false;
	return true;
    }

    int getType() {
	return type;
    }

    Locale getLocale() {
	return locale;
    }

    String getClassName() {
	return className;
    }

    String getResource() {
	return resource;
    }

    public String toString() {
	String typeName;
	String str;

	switch (type) {
	case HOST_ADAPTER:
	    typeName = "HOST_ADAPTER";
	    break;
	case IIIMP_ADAPTER:
	    typeName = "IIIMP_ADAPTER";
	    break;
	case JAVA_INPUT_METHOD:
	    typeName = "JAVA_INPUT_METHOD";
	    break;
	default:
	    typeName = "UNKNOWN";
	    break;
	}

	str = typeName + ", " + locale;
	if (className != null)
	    str += ", classname=" + className;
	if (resource != null)
	    str += ", resource=" + resource;
	return str;
    }
}
