/*
 * @(#)InputMethodHighlightMapping.java	1.4 98/06/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im;

import java.awt.im.InputMethodHighlight;
import java.io.InvalidObjectException;
import java.text.AttributedCharacterIterator.Attribute;
import java.util.HashMap;
import java.util.Map;

/**
 * Interface to map abstract input method highlight attributes to
 * visual ones.
 *
 * @see java.awt.im.InputMethodHighlight
 */
public interface InputMethodHighlightMapping {

    /**
     * Defines attribute keys and attribute values used for rendering input method highlights.
     */
    public static final class InputMethodHighlightAttribute extends Attribute {

        // table of all instances in this class, used by readResolve
        private static final Map instanceMap = new HashMap(7);

        /**
         * Constructs a InputMethodHighlightAttribute with the given name.
         */
        protected InputMethodHighlightAttribute(String name) {
            super(name);
            if (this.getClass() == InputMethodHighlightAttribute.class) {
                instanceMap.put(name, this);
            }
        }

        /**
         * Resolves instances being deserialized to the predefined constants.
         */
        protected Object readResolve() throws InvalidObjectException {
            if (this.getClass() != InputMethodHighlightAttribute.class) {
                throw new InvalidObjectException("subclass didn't correctly implement readResolve");
            }

            InputMethodHighlightAttribute instance = (InputMethodHighlightAttribute) instanceMap.get(getName());
            if (instance != null) {
                return instance;
            } else {
                throw new InvalidObjectException("unknown attribute name");
            };
        }

        /**
         * Attribute key for input method underline adornments.
         */
        public static final InputMethodHighlightAttribute INPUT_METHOD_UNDERLINE
                 = new InputMethodHighlightAttribute("input method underline");

        /** Single pixel solid low underline. */
        public static final Integer UNDERLINE_LOW_ONE_PIXEL = new Integer(1);

        /** Double pixel solid low underline. */
        public static final Integer UNDERLINE_LOW_TWO_PIXEL = new Integer(2);

        /** Single pixel dotted low underline. */
        public static final Integer UNDERLINE_LOW_DOTTED = new Integer(3);

        /** Double pixel gray low underline. */
        public static final Integer UNDERLINE_LOW_GRAY = new Integer(4);

        /** Single pixel dashed low underline. */
        public static final Integer UNDERLINE_LOW_DASHED = new Integer(5);

    }

    /**

     * Returns a map of visual attributes that have been mapped from
     * the given abstract input method highlight attribute. Returns
     * null if no mapping to visual attributes is found.
     *
     * @param highlight abstract input method highlight attribute
     */
    public Map mapHighlight(InputMethodHighlight highlight);
}
