/*
 * @(#)InputMethodManagerResources.java	1.1 98/05/15
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.im.resources;

import java.util.Locale;
import sun.awt.im.InputMethodManager;

import java.util.Hashtable;
import java.util.ResourceBundle;
import java.util.Enumeration;

/**
 * <CODE>InputMethodManagerResources</CODE> provides
 * <CODE>ListResourceBundle</CODE>-style resources that can be
 * localized. The <CODE>getLocaleName</CODE> method takes a Locale
 * object as a key to look up a localized locale name for an input
 * method. Localization of this resources class can be done as
 * follows:
 * <UL>
 * <LI>
 * Define <CODE>InputMethodManagerResources_&lt;locale></CODE> as a subclass of
 * <CODE>InputMethodManagerResources</CODE>.</LI>
 * 
 * <LI>
 * Override the <CODE>getTable</CODE> method that returns Object[][] consisting
 * of key/value pairs. Key must be either Locale or String. Value must be
 * String.</LI>
 * </UL>
 * 
 * @see java.util.ListResourceBundle
 * @version 1.1 05/15/98
 */

public class InputMethodManagerResources extends ResourceBundle
{
    /**
     * Translation table to be localized. Key can be Locale or String.
     */
    private static Object[][] translationTable = {
	// InputMethodManager
	{ "InputMethodManager.hostAdapterName", "System Input Method" },
	{ "InputMethodManager.IIIMPName", "Network Input Method" },
	{ "InputMethodManager.switchIMMenu", "Select Input Method" },
	{ "InputMethodManager.error.notFound", " not found" },
	{ "InputMethodManager.error.invokationError", " invocation error" },

	// CompositionArea
	{ "CompositionArea.inputWindowTitle", "Input Window" },

	// localized locale names for input methods which may or may
	// not be the same as Locale.getDisplayLanguage() values.
	//
	// NOTE: java.util.Locale defines both languages and
	// countries. The same language and country combination is
	// sometimes used for language and country constants
	// (objects). java.util.Hashtable is used to look up the table
	// and Hashtable uses equals() to compare keys. Therefore,
	// SIMPLIFIED_CHINESE, CHINA and PRC are the same thing, for
	// example. Only languages should be defined here.
	{ Locale.CHINESE, "Chinese" },
	{ Locale.SIMPLIFIED_CHINESE, "Simplified Chinese" },
	{ Locale.TRADITIONAL_CHINESE, "Traditional Chinese" },
	{ Locale.KOREAN, "Korean" },
	{ Locale.JAPANESE, "Japanese" }
    };

    private Hashtable hashedTranslationTable;

    protected Object[][] getTable() {
	return translationTable;
    }

    private synchronized void initializeTranslationTable() {
	Object[][] messages = getTable();
	Hashtable table = new Hashtable(messages.length);
	for (int i = 0; i < messages.length; i++)
	    table.put(messages[i][0], messages[i][1]);
	hashedTranslationTable = table;
    }
	
    /**
     * @see java.util.ResourceBundle#handleGetObject
     */
    public Object handleGetObject(String key) {
	if (hashedTranslationTable == null)
	    initializeTranslationTable();
	return hashedTranslationTable.get(key);
    }

    /**
     * @see java.util.ResourceBundle#getKeys
     */
    public Enumeration getKeys() {
	if (hashedTranslationTable == null)
	    initializeTranslationTable();
	return hashedTranslationTable.keys();
    }
	
    /**
     * Returns a localized locale name for input methods with the
     * given locale. It falls back to Locale.getDisplayLanguage() and
     * then to Locale.toString() if no localized locale name is found.
     *
     * @param locale Locale for which localized locale name is obtained
     */
    public String getLocaleName(Locale locale) {
	String localeName = (String) hashedTranslationTable.get(locale);
	if (localeName == null) {
	    localeName = Locale.getDefault().getDisplayLanguage(locale);
	    if (localeName == null || localeName.length() == 0)
		localeName = locale.toString();
	}
	return localeName;
    }
}
