/*
 * @(#)BufferedImageGraphics2D.java	1.104 98/10/02
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


package sun.awt.image;

import java.awt.Graphics2D;
import sun.java2d.SunGraphics2D;
import sun.java2d.OutputManager;
import sun.java2d.SunOutputManager;
import sun.java2d.loops.RasterOutputManager;
import sun.java2d.loops.RasterLockingException;
import sun.java2d.loops.GraphicsPrimitive;
import sun.java2d.loops.GraphicsPrimitiveMgr;
import sun.java2d.loops.LockableRaster;
import sun.java2d.loops.ImageData;
import sun.java2d.loops.DrawLine;
import sun.java2d.loops.DrawLineRasterContext;
import sun.java2d.loops.DrawRect;
import sun.java2d.loops.DrawRectRasterContext;
import sun.java2d.loops.DrawPolygons;
import sun.java2d.loops.DrawPolygonsRasterContext;
import sun.java2d.loops.FillPolygons;
import sun.java2d.loops.FillPolygonsRasterContext;
import sun.java2d.loops.FillRect;
import sun.java2d.loops.FillRectRasterContext;
import sun.java2d.loops.FillSpans;
import sun.java2d.loops.DrawStringRasterContext;
import sun.java2d.loops.DrawGlyphVectorRasterContext;
import sun.java2d.loops.DrawCharsRasterContext;
import sun.java2d.loops.DrawBytesRasterContext;
import sun.java2d.loops.DrawString;
import sun.java2d.loops.DrawGlyphVector;
import sun.java2d.loops.DrawChars;
import sun.java2d.loops.DrawBytes;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import sun.awt.image.ByteComponentRaster;
import sun.awt.image.BytePackedRaster;
import sun.awt.image.IntegerComponentRaster;
import sun.awt.image.ShortComponentRaster;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.image.DataBuffer;
import java.awt.image.DirectColorModel;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.SampleModel;
import java.awt.BasicStroke;
import java.awt.Component;
import java.awt.Composite;
import java.awt.CompositeContext;
import java.awt.Paint;
import java.awt.PaintContext;
import java.awt.Transparency;
import java.awt.AlphaComposite;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Point2D;
import java.awt.Stroke;
import java.awt.font.GlyphVector;
import java.awt.Color;
import java.awt.color.ColorSpace;
import java.awt.image.ColorConvertOp;
import java.awt.Image;
import java.awt.image.ImageObserver;
import java.awt.Shape;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.GraphicsConfiguration;
import sun.awt.AWTFinalizer;
import sun.awt.AWTFinalizeable;
import sun.java2d.pipe.DuctusRenderer;
import sun.java2d.pipe.PixelDrawPipe;
import sun.java2d.pipe.PixelFillPipe;
import sun.java2d.pipe.ShapeDrawPipe;
import sun.java2d.pipe.SimpleRenderPipe;
import sun.java2d.pipe.TranslateablePipe;
import sun.java2d.pipe.TextPipe;
import sun.java2d.pipe.SpanIterator;
import sun.java2d.pipe.ShapeSpanIterator;
import sun.dc.path.PathConsumer;
import sun.dc.path.PathException;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.geom.PathIterator;
import java.awt.font.FontRenderContext;

/**
 * This is a prototype of a Graphics2D class that uses the Ductus
 * rasterizer interface to rasterize a path.  This particular class
 * implements the renderAlphaRaster method using an animated Memory
 * image.
 *
 * @version 1.2 01/15/97
 * @author Jim Graham
 */
public class BufferedImageGraphics2D
    extends SunGraphics2D
    implements SimpleRenderPipe,
        PixelDrawPipe,
        PixelFillPipe,
	ShapeDrawPipe,
	TranslateablePipe,
        TextPipe
{
    public BufferedImage bufImg;
    Component component;
    WritableRaster raster;
    private ImageData cachedID;
    private int primtype;

    // cached state for faster rendering in this BufferedImage context
    Object drawLineState;           DrawLineRasterContext       dlrc;
    Object fillRectState;           FillRectRasterContext       frrc;
    FillSpanCachedState fillSpanState;
    Object drawRectState;           DrawRectRasterContext       drrc;
    Object drawPolysState;
    Object drawPolygonState;
    Object drawPolylineState;       DrawPolygonsRasterContext   dprc;
    Object fillPolysState;
    Object fillPolygonState;        FillPolygonsRasterContext   fprc;
    Object drawStringState;         DrawStringRasterContext     dsrc;
    Object drawGlyphVectorState;    DrawGlyphVectorRasterContext gvrc;
    Object drawCharsState;          DrawCharsRasterContext      dcrc;
    Object drawBytesState;          DrawBytesRasterContext      dbrc;
    Font lastGlyphVectorFont;

    // cached state for faster rendering of clearRect in this BufferedImage
    // context.
    FillRectRasterContext       crrc;

    // cached state for faster rendering in this BufferedImage context
    private int clipX;
    private int clipY;
    private int clipW;
    private int clipH;

    // state for "implements translateable"
    private int xOrigin;
    private int yOrigin;

    // cached state for various draw[String,Char,Byte] optimizations
    private Font            cachedFont;
    private AffineTransform cachedTextTransform;
    private boolean         textIsAntiAliased;
    private boolean         textUsesFractionalMetrics;
    private boolean         useFastTextClip;
    private float           fastClipMinY;   // user coordinates
    private float           fastClipMaxY;   // user coordinates

    private final static int slowTextTransformMask =
                            AffineTransform.TYPE_GENERAL_TRANSFORM
                        |   AffineTransform.TYPE_MASK_ROTATION
                        |   AffineTransform.TYPE_FLIP;

    public BufferedImageGraphics2D(BufferedImage bufImg, Component comp) {
        this.bufImg = bufImg;
        this.raster = bufImg.getRaster();
        this.component = comp;

	setDevClip(0, 0, bufImg.getWidth(), bufImg.getHeight());

        if (component != null) {
	    setColor(component.getForeground());
	    setBackground(component.getBackground());
	    setFont(component.getFont());
        } else {
	    setColor(Color.white);
	    setBackground(Color.black);
	    setFont(new Font("Dialog", Font.PLAIN, 12));
        }

        primtype = bufImg.getType();
        if ((primtype == GraphicsPrimitive.TYPE_BYTE_INDEXED) &&
            (bufImg.getColorModel() instanceof IndexColorModel)) {
            IndexColorModel icm = (IndexColorModel) bufImg.getColorModel();
            if ((ImageData.isAllOpaqueGrayICM(icm)) &&
                (icm.getPixelSize() == 8)) {
                primtype = GraphicsPrimitive.TYPE_INDEX8_GRAY;
            }
        } else if ((primtype == GraphicsPrimitive.TYPE_CUSTOM) &&
            (bufImg.getColorModel() instanceof IndexColorModel) &&
            (this.raster.getNumBands() == 1) &&
            (this.raster instanceof ShortComponentRaster)) {
            IndexColorModel icm = (IndexColorModel) bufImg.getColorModel();
            ShortComponentRaster sraster = (ShortComponentRaster) this.raster;
            if (sraster.getPixelStride() == 1 && icm.getPixelSize() == 12 &&
                ImageData.isAllOpaqueGrayICM(icm)) {
                primtype = GraphicsPrimitive.TYPE_INDEX12_GRAY;
            }
        }

        invalidateCachedState();
    }

    protected void invalidatePipe() {
        super.invalidatePipe();
        invalidateCachedState();
    }

    private void invalidateFontState() {
        this.drawStringState = null;        this.dsrc = null;
        this.drawGlyphVectorState = null;   this.gvrc = null;
        this.drawCharsState = null;         this.dcrc = null;
        this.drawBytesState = null;         this.dbrc = null;
        this.lastGlyphVectorFont = null;

        this.cachedFont = null;
        this.useFastTextClip = false;
    }

    private void invalidateCachedState() {
        this.drawLineState = null;      this.dlrc = null;
        this.fillRectState = null;      this.frrc = null;
	this.fillSpanState = null;
        this.drawRectState = null;      this.drrc = null;

        this.drawPolysState = null;
        this.drawPolygonState = null;
        this.drawPolylineState = null;  this.dprc = null;

        this.fillPolysState = null;
        this.fillPolygonState = null;   this.fprc = null;

        this.crrc = null;

	Rectangle clipRect = getCompBounds();
	if (clipRect != null) {
	    clipX = clipRect.x;
	    clipY = clipRect.y;
	    clipW = clipRect.width;
	    clipH = clipRect.height;
        } else if (devClip != null) {
            // REMIND: SunGraphics2D constructor needs to call an
			// overridable "getDevClip()" so that devClip is 
			// initialized before invalidateCachedState is called
	    clipX = devClip.x;
	    clipY = devClip.y;
	    clipW = devClip.width;
	    clipH = devClip.height;
        }
        
        this.cachedID = null;

        invalidateFontState();
    }

    // cached state for various draw[String,Char,Byte] optimizations
    private void computeFontOptimizations(Font font) {
        if (this.cachedFont != font) {
            this.cachedFont = font;

            float ptSize = font.getSize2D();
            FontRenderContext frc = getFontRenderContext();
            cachedTextTransform = (AffineTransform)frc.getTransform();
            cachedTextTransform.scale(ptSize, ptSize);
            cachedTextTransform.preConcatenate(font.getTransform());
            int txFontType = cachedTextTransform.getType();

            cachedTextTransform.preConcatenate(transform);

            textIsAntiAliased = frc.isAntiAliased();
            textUsesFractionalMetrics = frc.usesFractionalMetrics();

            useFastTextClip = false;
            if ( 0 == (txFontType & slowTextTransformMask)) {
                FontMetrics fm = getFontMetrics(font);
                int height = fm.getHeight();
                Rectangle rect = getClipBounds();
                if (rect != null) {
                    fastClipMinY = rect.y - height;
                    fastClipMaxY = rect.y + rect.height + height;
                } else {
                    fastClipMinY = devClip.y - height;
                    fastClipMaxY = devClip.y + devClip.height + height;
                }
                useFastTextClip = true;
            }
        }
    }

    public void setOrigin(int tx, int ty) {
	xOrigin = tx;
	yOrigin = ty;
    }

    public void setFont(Font font) {
        if (this.cachedFont != font) {
            super.setFont(font);
            invalidateFontState();
        }
    }

    public ColorModel getDeviceColorModel() {
	return bufImg.getColorModel();
    }

    public GraphicsConfiguration getDeviceConfiguration() {
        return BufferedImageGraphicsConfig.getConfig(bufImg);
    }

    public OutputManager getOutputManager() {
	return BufferedImageGraphicsConfig.getOutputManager(bufImg);
    }

    public void dispose() {
	// No system resources to dispose of (yet)
    }

    public void finalize() {
	// No system resources to dispose of (yet)
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver observer) {

	if (width == 0 || height == 0) {
	    return true;
	}

        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            BufferedImage bimg = (BufferedImage) img;

            super.drawImage(bimg, x, y, width, height, null, observer);

            return true;
        }

        sun.awt.image.Image simg = (sun.awt.image.Image)img;
	if (simg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img, ImageObserver.ERROR |
                                     ImageObserver.ABORT, -1, -1, -1, -1);
	    }
	    return false;
	}

	ImageRepresentation ir = simg.getImageRep();
	return ir.drawToBufImage(this, simg, x, y, width, height,
                                 null, observer);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a solid background
     * color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, Color bg,
			     ImageObserver observer) {
        if (img instanceof BufferedImage) {
            drawImage((BufferedImage)img, null, x, y);
            return true;
        }

        sun.awt.image.Image simg = (sun.awt.image.Image)img;
	if (simg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img, ImageObserver.ERROR |
                                     ImageObserver.ABORT, -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = simg.getImageRep();
        return ir.drawToBufImage(this, simg, x, y, bg, observer);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * solid background color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
 			     Color bg, ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}

        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            super.drawImage((BufferedImage)img, x, y, width, height, bg,
                            observer);
            return true;
        }

        sun.awt.image.Image simg = (sun.awt.image.Image)img;
	if (simg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img, ImageObserver.ERROR |
                                     ImageObserver.ABORT, -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = simg.getImageRep();
	return ir.drawToBufImage(this, simg, x, y, width, height, bg,
                                 observer);

    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle
     * in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer) {
        return drawImage(img, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2,
                         null, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle in
     * nonblocking mode with a solid background color and a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor, ImageObserver observer) {
        // REMIND: Guess we don't draw a line?
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}

        // REMIND: Guess we don't draw a line?
        if (sx1 == sx2 || sy1 == sy2) {
            return true;
        }

        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            super.drawImage((BufferedImage)img, dx1, dy1, dx2, dy2,
                            sx1, sy1, sx2, sy2, bgcolor, observer);
            return true;
        }

        sun.awt.image.Image simg = (sun.awt.image.Image)img;
	if (simg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img, ImageObserver.ERROR |
                                     ImageObserver.ABORT, -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = simg.getImageRep();

	return ir.drawToBufImage(this, simg,
                                 dx1, dy1, dx2, dy2,
                                 sx1, sy1, sx2, sy2,
                                 bgcolor, observer);

    }

    /**
     * Draw an image, applying a transform from image space into user space
     * before drawing.
     * The transformation from user space into device space is done with
     * the current transform in the Graphics2D.
     * The given transformation is applied to the image before the
     * transform attribute in the Graphics2D state is applied.
     * The rendering attributes applied include the clip, transform,
     * paint or color and composite attributes. Note that the result is
     * undefined, if the given transform is non-invertible.
     * @param img The image to be drawn.
     * @param xform The transformation from image space into user space.
     * @param obs The image observer to be notified on the image producing
     * progress.
     * @see #transform
     * @see #setComposite
     * @see #setClip
     */

    public boolean drawImage(Image img,
                          AffineTransform xform,
                          ImageObserver obs) {
        if (img instanceof BufferedImage) {
            super.drawImage((BufferedImage)img, xform, 0, 0, null, obs);
            return true;
        }

        sun.awt.image.Image simg = (sun.awt.image.Image)img;
	if (simg.hasError()) {
	    if (obs != null) {
		obs.imageUpdate(img, ImageObserver.ERROR |
                                     ImageObserver.ABORT, -1, -1, -1, -1);
	    }
	    return false;
	}

	ImageRepresentation ir = simg.getImageRep();
        return ir.drawToBufImage(this, simg, xform, obs);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {

        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, x, y, null, observer);
        }

        sun.awt.image.Image simg = (sun.awt.image.Image)img;
	if (simg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img, ImageObserver.ERROR |
                                     ImageObserver.ABORT, -1, -1, -1, -1);
	    }
	    return false;
	}

	ImageRepresentation ir = simg.getImageRep();
        return ir.drawToBufImage(this, simg, x, y, null, observer);
    }

    protected void checkCustomComposite() {
	// Custom composites are always allowed on offscreen images
	return;
    }

    /**
     * Sets the Paint in the current graphics state.
     * @param paint The Paint object to be used to generate color in
     * the rendering process.
     * @see java.awt.Graphics#setColor
     * @see GradientPaint
     * @see TexturePaint
     */
    public void setPaint(Paint paint) {
        super.setPaint(paint);
        invalidateCachedState();
    }

    public void setBackground(Color color) {
        if (backgroundColor != color) {
            crrc = null;    // invalidate clearRect cached state
            super.setBackground(color);
        }
    }

    /**
    * setClip implementation to cache state needed by:
    *           SimpleRenderPipe
    *           PixelDrawPipe
    *   and     PixelFillPipe
    */
    public void setClip(int x, int y, int w, int h) {
	super.setClip(x, y, w, h);
        invalidateCachedState();
    }
     
    public void setClip(Shape s) {
        super.setClip(s);
        invalidateCachedState();
    }

    public void clipRect(int x, int y, int w, int h) {
        super.clipRect(x, y, w, h);
        invalidateCachedState();
    }

    public void clip(Shape s) {
        super.clip(s);
        invalidateCachedState();
    }
     
    /**
    * SimpleRenderPipe methods
    */
    public void drawLine(SunGraphics2D dest2D,
            int x1, int y1, int x2, int y2) {
	x1 += xOrigin;
	y1 += yOrigin;
	x2 += xOrigin;
	y2 += yOrigin;
        if (dlrc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawLine.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                DrawLine dl = (DrawLine)op;
                dlrc = dl.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (dlrc != null) {
            dlrc.invoke(x1, y1, x2, y2);
        } else {
            drawLineState = RasterOutputManager.getManager().drawLine(
                                        dest2D, x1, y1, x2, y2,
                                        drawLineState);
        }
    }

    public void fillRect(SunGraphics2D dest2D,
            int x, int y, int w, int h) {
	x += xOrigin;
	y += yOrigin;
        if (frrc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        FillRect.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                FillRect fr = (FillRect)op;
                frrc = fr.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (frrc != null) {
            frrc.invoke(x, y, w, h);
        } else {
            fillRectState = RasterOutputManager.getManager().fillRect(
                                        dest2D, x, y, w, h,
                                        fillRectState);
        }
    }

    public void draw(SunGraphics2D dest2D, Shape s) {
	ShapeSpanIterator sr = new ShapeSpanIterator();

	try {
	    sr.setOutputArea(dest2D.getCompBounds());
	    sr.setRule(PathIterator.WIND_NON_ZERO);

	    BasicStroke bs = (BasicStroke) dest2D.stroke;
	    AffineTransform transform =
		(dest2D.transformState == dest2D.TRANSFORM_GENERIC
		 ? dest2D.transform : null);

	    PathConsumer consumer =
		DuctusRenderer.createStroker(sr, bs, transform);

	    transform =
		((dest2D.transformState == TRANSFORM_ISIDENT)
		 ? null
		 : dest2D.transform);
	    PathIterator pi = s.getPathIterator(transform);

	    try {
		DuctusRenderer.feedConsumer(pi, consumer);
	    } catch (PathException e) {
		throw new InternalError("Unable to Stroke shape ("+
					e.getMessage()+")");
	    }
	    fillSpans(dest2D, sr);
	} finally {
	    sr.dispose();
	}
    }

    public void fill(SunGraphics2D dest2D, Shape s) {
	ShapeSpanIterator sr = new ShapeSpanIterator();
	try {
	    sr.setOutputArea(dest2D.getCompBounds());
	    AffineTransform at =
		((dest2D.transformState == TRANSFORM_ISIDENT)
		 ? null
		 : dest2D.transform);
	    sr.appendPath(s.getPathIterator(at));
	    fillSpans(dest2D, sr);
	} finally {
	    sr.dispose();
	}
    }

    private static class FillSpanCachedState {
	FillSpans fillSpanObj;
	LockableRaster lrDst;
    }

    public void fillSpans(SunGraphics2D dest2D, SpanIterator si) {
	FillSpans fs = null;
	LockableRaster dst = null;
	if (fillSpanState != null) {
	    fs = fillSpanState.fillSpanObj;
	    dst = fillSpanState.lrDst;
	} else {
	    dst = new LockableRaster(dest2D);
	    int [] types = {dst.type};
	    GraphicsPrimitive op =
		GraphicsPrimitiveMgr.locate(FillSpans.getMethodSignature(),
					    types, dst.devID);
	    if (op != null) {
		fillSpanState = new FillSpanCachedState();
		fillSpanState.fillSpanObj = fs = (FillSpans) op;
		fillSpanState.lrDst = dst;
	    } else if (frrc == null) {
		op = GraphicsPrimitiveMgr.locate(FillRect.getMethodSignature(),
						 types, dst.devID);
		if (op != null) {
		    FillRect fr = (FillRect)op;
		    frrc = fr.createCachedContext(  raster,
						  bufImg.getColorModel(),
						  getColor(),
						  clipX, clipY, clipW, clipH);
		}
	    }
	}
	int spanbox[] = new int[4];
	if (fs != null) {
	    si.getPathBox(spanbox);
	    try {
		dst.lock(fs.getLockFlags()[0],
			 spanbox[0], spanbox[1],
			 spanbox[2] - spanbox[0],
			 spanbox[3] - spanbox[1]);
		fs.FillSpans(dst.id, dest2D.getColor(), si);
	    } catch (RasterLockingException e) {
	    } finally {
		dst.unlock();
	    }
	    return;
	}
	RasterOutputManager rom = RasterOutputManager.getManager();
	while (si.nextSpan(spanbox)) {
	    if (frrc != null) {
		frrc.invoke(spanbox[0], spanbox[1],
			    spanbox[2] - spanbox[0],
			    spanbox[3] - spanbox[1]);
	    } else {
		fillRectState = rom.fillRect(dest2D,
					     spanbox[0], spanbox[1],
					     spanbox[2] - spanbox[0],
					     spanbox[3] - spanbox[1],
					     fillRectState);
	    }
	}
    }

    private void translatePolys(int[][] points, int nPoints[], int numpolys) {
	int i, numpoints = 0;
	int P[], newP[];
	if (nPoints == null) {
	    numpoints = numpolys;
	} else {
	    for (i = 0; i < numpolys; i++) {
		numpoints += nPoints[i];
	    }
	}
	if (xOrigin != 0) {
	    newP = new int[numpoints];
	    P = points[0];
	    for (i = 0; i < numpoints; i++) {
		newP[i] = P[i] + xOrigin;
	    }
	    points[0] = newP;
	}
	if (yOrigin != 0) {
	    newP = new int[numpoints];
	    P = points[1];
	    for (i = 0; i < numpoints; i++) {
		newP[i] = P[i] + yOrigin;
	    }
	    points[1] = newP;
	}
    }

    public void drawPolys(SunGraphics2D dest2D,
            int[] xPoints, int[] yPoints, int nPoints[],
            int numpolys, boolean close) {
	if (xOrigin != 0 || yOrigin != 0) {
	    int[][] points = {xPoints, yPoints};
	    translatePolys(points, nPoints, numpolys);
	    xPoints = points[0];
	    yPoints = points[1];
	}
        if (dprc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawPolygons.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                DrawPolygons dp = (DrawPolygons)op;
                dprc = dp.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (dprc != null) {
            dprc.invoke(xPoints, yPoints, nPoints, numpolys, close);
        } else {
            drawPolysState = RasterOutputManager.getManager().drawPolys(
                                        dest2D, xPoints, yPoints, nPoints,
                                        numpolys, close, drawPolysState);
        }
    }

    public void fillPolys(SunGraphics2D dest2D,
            int[] xPoints, int[] yPoints, int nPoints[],
            int numpolys, boolean evenodd) {
	if (xOrigin != 0 || yOrigin != 0) {
	    int[][] points = {xPoints, yPoints};
	    translatePolys(points, nPoints, numpolys);
	    xPoints = points[0];
	    yPoints = points[1];
	}
        if (fprc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        FillPolygons.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                FillPolygons fp = (FillPolygons)op;
                fprc = fp.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (fprc != null) {
            fprc.invoke(xPoints, yPoints, nPoints, numpolys, evenodd );
        } else {
            fillPolysState = RasterOutputManager.getManager().fillPolys(
                                        dest2D, xPoints, yPoints, nPoints,
                                        numpolys, evenodd, fillPolysState);
        }
    }

    /**
    * PixelDrawPipe methods
    */
    public void drawRect(SunGraphics2D dest2D,
			 int x, int y, int w, int h) {
	x += xOrigin;
	y += yOrigin;
        if (drrc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawRect.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                DrawRect dr = (DrawRect)op;
                drrc = dr.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (drrc != null) {
            drrc.invoke(x, y, w, h);
        } else {
            drawRectState = RasterOutputManager.getManager().drawRect(
                                        dest2D, x, y, w, h,
                                        drawRectState);
        }
    }

    public void drawRoundRect(SunGraphics2D sg,
			      int x, int y, int w, int h,
			      int aW, int aH) {
	sg.shapepipe.draw(sg, new RoundRectangle2D.Float(x, y, w, h, aW, aH));
    }

    public void drawOval(SunGraphics2D sg,
			 int x, int y, int w, int h) {
	sg.shapepipe.draw(sg, new Ellipse2D.Float(x, y, w, h));
    }

    public void drawArc(SunGraphics2D sg,
			int x, int y, int w, int h,
			int start, int extent) {
	sg.shapepipe.draw(sg, new Arc2D.Float(x, y, w, h,
					      start, extent, Arc2D.OPEN));
    }

    public void drawPolyline(SunGraphics2D dest2D,
			     int xPoints[], int yPoints[],
			     int nPoints) {
	if (xOrigin != 0 || yOrigin != 0) {
	    int[][] points = {xPoints, yPoints};
	    translatePolys(points, null, nPoints);
	    xPoints = points[0];
	    yPoints = points[1];
	}
        if (dprc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawPolygons.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                DrawPolygons dp = (DrawPolygons)op;
                dprc = dp.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (dprc != null) {
            int nPointsArray[] = { nPoints };
            dprc.invoke(xPoints, yPoints, nPointsArray, 1, false );
        } else {
            drawPolylineState = RasterOutputManager.getManager().drawPolyline(
                                        dest2D,
                                        xPoints, yPoints, nPoints,
                                        drawPolylineState);
        }
    }

    public void drawPolygon(SunGraphics2D dest2D,
			    int xPoints[], int yPoints[],
			    int nPoints) {
	if (xOrigin != 0 || yOrigin != 0) {
	    int[][] points = {xPoints, yPoints};
	    translatePolys(points, null, nPoints);
	    xPoints = points[0];
	    yPoints = points[1];
	}
        if (dprc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawPolygons.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                DrawPolygons dp = (DrawPolygons)op;
                dprc = dp.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (dprc != null) {
            int nPointsArray[] = { nPoints };
            dprc.invoke(xPoints, yPoints, nPointsArray, 1, true );
        } else {
            drawPolygonState = RasterOutputManager.getManager().drawPolygon(
                                    dest2D,
                                    xPoints, yPoints, nPoints,
                                    drawPolygonState);
        }
    }

    /**
    * PixelFillPipe methods
    */
    public void fillRoundRect(SunGraphics2D sg,
			      int x, int y, int w, int h,
			      int aW, int aH) {
	sg.shapepipe.fill(sg, new RoundRectangle2D.Float(x, y, w, h, aW, aH));
    }

    public void fillOval(SunGraphics2D sg,
			 int x, int y, int w, int h) {
	sg.shapepipe.fill(sg, new Ellipse2D.Float(x, y, w, h));
    }

    public void fillArc(SunGraphics2D sg,
			int x, int y, int w, int h,
			int start, int extent) {
	sg.shapepipe.fill(sg, new Arc2D.Float(x, y, w, h,
					      start, extent, Arc2D.PIE));
    }

    public void fillPolygon(SunGraphics2D dest2D,
			    int xPoints[], int yPoints[],
			    int nPoints) {
	if (xOrigin != 0 || yOrigin != 0) {
	    int[][] points = {xPoints, yPoints};
	    translatePolys(points, null, nPoints);
	    xPoints = points[0];
	    yPoints = points[1];
	}
        if (fprc == null) {
            LockableRaster dst = new LockableRaster(dest2D);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        FillPolygons.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                FillPolygons fp = (FillPolygons)op;
                fprc = fp.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (fprc != null) {
            int nPointsArray[] = { nPoints };
            fprc.invoke(xPoints, yPoints, nPointsArray, 1, true );
        } else {
            fillPolygonState = RasterOutputManager.getManager().fillPolygon(
                                            dest2D, xPoints, yPoints, nPoints,
                                            fillPolygonState);
        }
    }

    public void clearRect(int x, int y, int w, int h) {
        if (fillpipe != this) {
            super.clearRect(x, y, w, h);
	    return;
        }
        if (crrc == null && clipState != CLIP_SHAPE) {
            LockableRaster dst = new LockableRaster(this);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        FillRect.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                FillRect fr = (FillRect)op;
                crrc = fr.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getBackground(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (crrc != null) {
            x += xOrigin;
            y += yOrigin;
            crrc.invoke(x, y, w, h);
        } else {
            super.clearRect(x, y, w, h);
        }
    }

    /**
    * TextPipe methods
    */
    public Object drawString(SunGraphics2D g2d, String str,
                             float x, float y, Object cs)
    {
        if (dsrc == null) {
            LockableRaster dst = new LockableRaster(this);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawString.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                if (cachedFont == null) {
                    computeFontOptimizations(getFont());
                }
                DrawString ds = (DrawString)op;
                dsrc = ds.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                cachedFont,
                                                cachedTextTransform,
                                                textIsAntiAliased,
                                                textUsesFractionalMetrics,
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (useFastTextClip && ((y < fastClipMinY) || (y > fastClipMaxY))) {
            return null;
        }
        if (dsrc != null) {
	    if (g2d.transformState == TRANSFORM_GENERIC) {
		double origin[] = {x, y};
		g2d.transform.deltaTransform(origin, 0, origin, 0, 1);
		x = (float) origin[0];
		y = (float) origin[1];
	    }
            dsrc.invoke(str, x, y);
        } else {
            drawStringState = RasterOutputManager.getManager().drawString(
                                        g2d, str, x, y,
                                        drawStringState);
        }
        return null;
    }

    public Object drawGlyphVector(SunGraphics2D g2d, GlyphVector glyphVector,
            float x, float y, Object cs)
    {
        Font font = glyphVector.getFont();
        if (font != lastGlyphVectorFont) {
            drawGlyphVectorState = null;
            gvrc = null;
            lastGlyphVectorFont = font;
        }
        if (gvrc == null) {
            LockableRaster dst = new LockableRaster(this);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawGlyphVector.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                float ptSize = font.getSize2D();
                FontRenderContext frc = glyphVector.getFontRenderContext();
                AffineTransform tx = (AffineTransform)frc.getTransform();
                tx.scale(ptSize, ptSize);
                tx.preConcatenate(font.getTransform());
                tx.preConcatenate(transform);
                DrawGlyphVector gv = (DrawGlyphVector)op;
                gvrc = gv.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                font,
                                                tx, transform,
                                                frc.isAntiAliased(),
                                                frc.usesFractionalMetrics(),
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (gvrc != null) {
	    if (g2d.transformState == TRANSFORM_GENERIC) {
		double origin[] = {x, y};
		g2d.transform.deltaTransform(origin, 0, origin, 0, 1);
		x = (float) origin[0];
		y = (float) origin[1];
	    }
            gvrc.invoke(glyphVector, x, y);
        } else {
            drawGlyphVectorState =
                RasterOutputManager.getManager().drawGlyphVector(
                                                    g2d, glyphVector, x, y,
                                                    drawGlyphVectorState);
        }
        return null;
    }

    public Object drawChars(SunGraphics2D g2d,
            char data[], int offset, int length,
            int x, int y, Object cs)
    {
        if (dcrc == null) {
            LockableRaster dst = new LockableRaster(this);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawChars.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                if (cachedFont == null) {
                    computeFontOptimizations(getFont());
                }
                DrawChars dc = (DrawChars)op;
                dcrc = dc.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                cachedFont,
                                                cachedTextTransform,
                                                textIsAntiAliased,
                                                textUsesFractionalMetrics,
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (useFastTextClip && ((y < fastClipMinY) || (y > fastClipMaxY))) {
            return null;
        }
        if (dcrc != null) {
	    float ux, uy;
	    if (g2d.transformState == TRANSFORM_GENERIC) {
		double origin[] = {x, y};
		g2d.transform.deltaTransform(origin, 0, origin, 0, 1);
		ux = (float) origin[0];
		uy = (float) origin[1];
	    } else {
		ux = x;
		uy = y;
	    }
            dcrc.invoke(data, offset, length, ux, uy);
        } else {
            drawCharsState = RasterOutputManager.getManager().drawChars(
                                        g2d, data, offset, length, x, y,
                                        drawCharsState);
        }
        return null;
    }

    public Object drawBytes(SunGraphics2D g2d,
            byte data[], int offset, int length,
            int x, int y, Object cs)
    {
        if (dbrc == null) {
            LockableRaster dst = new LockableRaster(this);
            int [] types = {dst.type};
            GraphicsPrimitive op = GraphicsPrimitiveMgr.locate(
                                        DrawBytes.getMethodSignature(),
                                        types, dst.devID);
            if (op != null) {
                if (cachedFont == null) {
                    computeFontOptimizations(getFont());
                }
                DrawBytes db = (DrawBytes)op;
                dbrc = db.createCachedContext(  raster,
                                                bufImg.getColorModel(),
                                                getColor(),
                                                cachedFont,
                                                cachedTextTransform,
                                                textIsAntiAliased,
                                                textUsesFractionalMetrics,
                                                clipX, clipY, clipW, clipH);
            }
        }
        if (useFastTextClip && ((y < fastClipMinY) || (y > fastClipMaxY))) {
            return null;
        }
        if (dbrc != null) {
	    float ux, uy;
	    if (g2d.transformState == TRANSFORM_GENERIC) {
		double origin[] = {x, y};
		g2d.transform.deltaTransform(origin, 0, origin, 0, 1);
		ux = (float) origin[0];
		uy = (float) origin[1];
	    } else {
		ux = x;
		uy = y;
	    }
            dbrc.invoke(data, offset, length, ux, uy);
        } else {
            drawBytesState = RasterOutputManager.getManager().drawBytes(
                                        g2d, data, offset, length, x, y,
                                        drawBytesState);
        }
        return null;
    }

    /*
    * REMIND: DOCS!!!
    * REMIND: remove extra code from this file when done with
    * REMIND: FrameBuffer Architecture
    */

    /**
     * The method prepares locks for a SunGraphics2D to make it accessible
     * for direct operation on graphics content by GraphicsPrimitives
     */
    public ImageData lock(
             int xView, int yView, int dxView, int dyView, int lockFlags)
    {
        ImageData imgData = cachedID;
        if (null == imgData || imgData.amILocked()) {
            imgData = new ImageData();
        }
        imgData.setViewArea(xView, yView, dxView, dyView);
        imgData.setLockMethod(lockFlags);
        if (SunGraphics2D.LOCK_OK == this.lock(imgData)) {
            cachedID = imgData;
            return imgData;
        }
        return null;
    }

    private int lock(ImageData idData) {
        if (null != idData) {
            if (idData.amILocked()) {
                return LOCK_ALREADY_LOCKED;
            }
            // REMIND: validation needs to be more thorough
            // since it appears that the bufferedImage can
            // be told to change its raster and/or color model
            // by external objects.
            if (!idData.equals(cachedID)) {
                if (!LockableRaster.prepareImageData(
                        idData, raster,
                        bufImg.getColorModel(), primtype)) {
                    return LOCK_DATA_STRUCT_PROBLEM;
                }
            }
	    if (LockableRaster.prepareRasterBounds(idData, clipX, clipY,
						   clipW, clipH)) {
                idData.incLockCount();
                cachedID = idData;
                return LOCK_OK;
            }
        }
        return LOCK_DATA_STRUCT_PROBLEM;
    }

    public int unlock(ImageData idData) {
        int nRC = LOCK_DATA_STRUCT_PROBLEM;

        Raster tRaster = raster;
        nRC = LOCK_NOT_LOCKED;
        if ((null != idData) && (true == idData.amILocked())) {
            idData.decLockCount();
        }
        return nRC;
    }
    public String toString() {
        String str = "BufferedImageGraphics2D[";
        str +="BufferedImage["+bufImg+"]";
        str +=" Component["+component+"]";
        return str;
    }
}
