/*
 * @(#)BufferedImageGraphicsConfig.java	1.19 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.image;

import java.awt.GraphicsConfiguration;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.GraphicsDevice;
import java.awt.Transparency;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.image.DirectColorModel;
import java.awt.geom.AffineTransform;
import sun.java2d.OutputManager;
import sun.java2d.SunOutputManager;
import sun.java2d.loops.RasterOutputManager;
import sun.java2d.SunGraphics2D;

public class BufferedImageGraphicsConfig
    extends GraphicsConfiguration
{
    private static final int numconfigs = BufferedImage.TYPE_BYTE_BINARY;
    private static BufferedImageGraphicsConfig configs[] =
	new BufferedImageGraphicsConfig[numconfigs];

    public static GraphicsConfiguration getConfig(BufferedImage bImg) {
	BufferedImageGraphicsConfig ret;
	int type = bImg.getType();
	if (type > 0 && type < numconfigs) {
	    ret = configs[type];
	    if (ret != null) {
		return ret;
	    }
	}
	ret = new BufferedImageGraphicsConfig(bImg);
	if (type > 0 && type < numconfigs) {
	    configs[type] = ret;
	}
	return ret;
    }

    public static OutputManager getOutputManager(BufferedImage bImg) {
        return RasterOutputManager.getManager();
    }

    BufferedImage bImg;

    public BufferedImageGraphicsConfig(BufferedImage bufImg) {
        bImg = bufImg;
    }

    /**
     * Return the graphics device associated with this configuration.
     */
    public GraphicsDevice getDevice() {
        return null;
    }

    /**
     * Returns a BufferedImage with channel layout and color model
     * compatible with this graphics configuration.  This method
     * has nothing to do with memory-mapping
     * a device.  This BufferedImage has
     * a layout and color model
     * that is closest to this native device configuration and thus
     * can be optimally blitted to this device.
     */
    public BufferedImage createCompatibleImage(int width, int height) {
	ColorModel model = bImg.getColorModel();
	WritableRaster raster =
           bImg.getRaster().createCompatibleWritableRaster(width, height);
        return new BufferedImage(model, raster, model.isAlphaPremultiplied(),
                                 null);
    }

    /**
     * Returns a BufferedImage that supports the specified transparency
     * and has a channel layout and color model
     * compatible with this graphics configuration.  This method
     * has nothing to do with memory-mapping
     * a device. This BufferedImage has a layout and
     * color model that can be optimally blitted to a device
     * with this configuration.
     * @see Transparency#OPAQUE
     * @see Transparency#BITMASK
     * @see Transparency#TRANSLUCENT
     */
    public BufferedImage createCompatibleImage(int width, int height,
                                                        int transparency) {
        //REMIND: process transparency
        return createCompatibleImage(width, height);
    }

    /**
     * Returns the color model associated with this configuration.
     */
    public ColorModel getColorModel() {
        return bImg.getColorModel();
    }

    /**
     * Returns the color model associated with this configuration that
     * supports the specified transparency.
     */
    public ColorModel getColorModel(int transparency) {
        ColorModel cm = bImg.getColorModel();
        if (cm.getTransparency() == transparency) {
            return cm;
        }

        // REMIND: Now what?
        return null;
    }

    /**
     * Returns the default Transform for this configuration.  This
     * Transform is typically the Identity transform for most normal
     * screens.  Device coordinates for screen and printer devices will
     * have the origin in the upper left-hand corner of the target region of
     * the device, with X coordinates
     * increasing to the right and Y coordinates increasing downwards.
     * For image buffers, this Transform will be the Identity transform.
     */
    public AffineTransform getDefaultTransform() {
        return new AffineTransform();
    }

    /**
     *
     * Returns a Transform that can be composed with the default Transform
     * of a Graphics2D so that 72 units in user space will equal 1 inch
     * in device space.
     * Given a Graphics2D, g, one can reset the transformation to create
     * such a mapping by using the following pseudocode:
     * <pre>
     *      GraphicsConfiguration gc = g.getGraphicsConfiguration();
     *
     *      g.setTransform(gc.getDefaultTransform());
     *      g.transform(gc.getNormalizingTransform());
     * </pre>
     * Note that sometimes this Transform will be identity (e.g. for
     * printers or metafile output) and that this Transform is only
     * as accurate as the information supplied by the underlying system.
     * For image buffers, this Transform will be the Identity transform,
     * since there is no valid distance measurement.
     */
    public AffineTransform getNormalizingTransform() {
	return new AffineTransform();
    }
}
