/*
 * @(#)ImageConsumerQueue.java	1.6 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.image;

import java.awt.image.ImageConsumer;

class ImageConsumerQueue {
    ImageConsumerQueue next;

    ImageConsumer consumer;
    boolean interested;

    Object securityContext;
    boolean secure;

    static ImageConsumerQueue removeConsumer(ImageConsumerQueue cqbase,
					     ImageConsumer ic,
					     boolean stillinterested)
    {
	ImageConsumerQueue cqprev = null;
	for (ImageConsumerQueue cq = cqbase; cq != null; cq = cq.next) {
	    if (cq.consumer == ic) {
		if (cqprev == null) {
		    cqbase = cq.next;
		} else {
		    cqprev.next = cq.next;
		}
		cq.interested = stillinterested;
		break;
	    }
	    cqprev = cq;
	}
	return cqbase;
    }

    static boolean isConsumer(ImageConsumerQueue cqbase, ImageConsumer ic) {
	for (ImageConsumerQueue cq = cqbase; cq != null; cq = cq.next) {
	    if (cq.consumer == ic) {
		return true;
	    }
	}
	return false;
    }

    ImageConsumerQueue(InputStreamImageSource src, ImageConsumer ic) {
	consumer = ic;
	interested = true;
	// ImageReps do their own security at access time.
	if (ic instanceof ImageRepresentation) {
	    ImageRepresentation ir = (ImageRepresentation) ic;
	    if (ir.image.source != src) {
		throw new SecurityException("ImageRep added to wrong image source");
	    }
	    secure = true;
	} else {
            SecurityManager security = System.getSecurityManager();
            if (security != null) {
                securityContext = security.getSecurityContext();
            } else {
                securityContext = null;
            }
	}
    }

    public String toString() {
	return ("[" + consumer +
		", " + (interested ? "" : "not ") + "interested" +
		(securityContext != null ? ", " + securityContext : "") +
		"]");
    }
}
