/*
 * @(#)OffScreenImage.java	1.15 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.image;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.image.ImageProducer;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import sun.awt.image.BufferedImageGraphics2D;

public class OffScreenImage extends BufferedImage {
    Component c;
    OffScreenImageSource osis;

    /**
     * Constructs an OffScreenImage with the default image type,
     * TYPE_4BYTE_ABGR, for offscreen rendering to be used with a
     * given component.
     * The component is used to obtain the foreground color, background
     * color and font.
     * REMIND:  At some point, we might want to look at the component
     * and figure out the best image type to construct.
     */
    public OffScreenImage(Component c, int width, int height) {
        // REMIND:  Should get image type from the component.getGC
        super(width, height, TYPE_4BYTE_ABGR);
        this.c = c;
	initBackground(width, height);
    }

    /**
     * Constructs an OffScreenImage with the given image type,
     * for offscreen rendering to be used with a given component.
     * The component is used to obtain the foreground color, background
     * color and font.
     */
    public OffScreenImage(Component c, int width, int height,
                             int imageType)
    {
        super (width, height, imageType);
        this.c = c;
	initBackground(width, height);
    }

    /**
     * Constructs an OffScreenImage given a color model and tile,
     * for offscreen rendering to be used with a given component.
     * The component is used to obtain the foreground color, background
     * color and font.
     */
    public OffScreenImage(Component c, ColorModel cm, WritableRaster raster,
                            boolean isRasterPremultiplied)
    {
        super(cm, raster, isRasterPremultiplied, null);
        this.c = c;
	initBackground(raster.getWidth(), raster.getHeight());
    }

    public Graphics getGraphics() {
        return new BufferedImageGraphics2D(this, c);
    }

    public Graphics2D createGraphics() {
        return new BufferedImageGraphics2D(this, c);
    }

    public void initBackground(int width, int height) {
	Graphics2D g2 = createGraphics();
	g2.clearRect(0, 0, width, height);
	g2.dispose();
    }

    public ImageProducer getSource() {
	if (osis == null) {
	    osis = new OffScreenImageSource(this);
	}
        return osis;
    }

}
