/*
 * @(#)CharToByteUnicode.java	1.9 98/07/07
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.io;
import java.io.*;


/**
 * Convert arrays containing Unicode characters into arrays of bytes, using the
 * platform-default byte order.
 *
 * @version 	1.9, 98/07/07
 * @author	Mark Reinhold
 */

public class CharToByteUnicode extends CharToByteConverter {

    static final char BYTE_ORDER_MARK = (char) 0xfeff;
    protected boolean marked = false;

    static final int UNKNOWN = 0;
    static final int BIG = 1;
    static final int LITTLE = 2;
    protected int byteOrder = UNKNOWN;

    public CharToByteUnicode() {
	String enc = (String) java.security.AccessController.doPrivileged(
           new sun.security.action.GetPropertyAction("sun.io.unicode.encoding",
							  "UnicodeBig"));
	if (enc.equals("UnicodeBig"))
	    byteOrder = BIG;
	else if (enc.equals("UnicodeLittle"))
	    byteOrder = LITTLE;
	else
	    byteOrder = BIG;
    }

    public String getCharacterEncoding() {
	switch (byteOrder) {
	case BIG:     return "UnicodeBig";
	case LITTLE:  return "UnicodeLittle";
	default:      return "UnicodeUnknown";
	}
    }

    public int convert(char in[], int inOff, int inEnd,
		       byte out[], int outOff, int outEnd)
	throws ConversionBufferFullException, MalformedInputException
    {
	charOff = inOff;
	byteOff = outOff;

	if (inOff >= inEnd)
	    return 0;

	int inI = inOff,
	    outI = outOff,
	    outTop = outEnd - 2;

	if (!marked) {
	    if (outI > outTop)
		throw new ConversionBufferFullException();
	    if (byteOrder == BIG) {
		out[outI++] = (byte) (BYTE_ORDER_MARK >> 8);
		out[outI++] = (byte) (BYTE_ORDER_MARK & 0xff);
	    }
	    else {
		out[outI++] = (byte) (BYTE_ORDER_MARK & 0xff);
		out[outI++] = (byte) (BYTE_ORDER_MARK >> 8);
	    }
	    marked = true;
	}

	if (byteOrder == BIG) {
	    while (inI < inEnd) {
		if (outI > outTop) {
		    charOff = inI;
		    byteOff = outI;
		    throw new ConversionBufferFullException();
		}
		char c = in[inI++];
		out[outI++] = (byte) (c >> 8);
		out[outI++] = (byte) (c & 0xff);
	    }
	}
	else {
	    while (inI < inEnd) {
		if (outI > outTop) {
		    charOff = inI;
		    byteOff = outI;
		    throw new ConversionBufferFullException();
		}
		char c = in[inI++];
		out[outI++] = (byte) (c & 0xff);
		out[outI++] = (byte) (c >> 8);
	    }
	}

	charOff = inI;
	byteOff = outI;
	return outI - outOff;
    }

    public int flush(byte in[], int inOff, int inEnd) {
	byteOff = charOff = 0;
	return 0;
    }

    public void reset () {
	byteOff = charOff = 0;
    }

    public int getMaxBytesPerChar() {
	return 4;		/* To allow for writing the byte-order mark */
    }

}
