/*
 * @(#)OutputManager.java	1.14 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Font;
import java.awt.CompositeContext;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import sun.java2d.pipe.AlphaColorPipe;

public interface OutputManager {
    /**
     * Construct and install the necessary shapepipe, drawpipe, and
     * fillpipe objects into the indicated SunGraphics2D based upon
     * the current attributes of that object.
     */
    public void validatePipe(SunGraphics2D sg);

    /**
     * Render the indicated image with the specified modifications.
     * The current transform, composite, and clip of the SunGraphics2D
     * are also applied.  May return cached state which can accelerate
     * successive operations in the same context
     */
    public Object renderImage(SunGraphics2D sg,
			    BufferedImage bImg, Color bg,
			    int dx, int dy, int w, int h,
			    byte[] clipAlpha, int rasterScan,
                            Object cachedState);

    public Object copyArea(SunGraphics2D dest2D, byte[] clipAlpha,
                           int alphaScan, int alphaOffset,
                           int x, int y, int w, int h, int dx, int dy,
                           Object cachedState);

    public void drawString(SunGraphics2D sg, int encoding,
                           String str, float x, float y,
                           float alignment,
                           float[] matrix, Font font,
                           boolean doAntialias, boolean doFractEnable);
    /**
     * Returns the preferred ColorModel to convert to when a source
     * image of the indicated srcModel needs to be composited.
     */
    public ColorModel getPreferredModel(ColorModel srcModel,
					Raster srcRaster,
					AlphaComposite ac,
                                        Color bg,
                                        boolean needsClip);

    /**
     * Returns the AlphaColorPipe object to perform composite operations
     * for solid colors.
     */
    public AlphaColorPipe getColorPipe();

}
