/*
 * @(#)SunGraphics2D.java	1.254 98/10/02
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.RenderingHints.Key;
import java.awt.geom.Area;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.RenderableImage;
import java.awt.image.renderable.RenderContext;
import java.awt.image.RasterOp;
import java.awt.image.AffineTransformOp;
import sun.awt.image.IntegerComponentRaster;
import sun.awt.image.ShortComponentRaster;
import sun.awt.image.ByteComponentRaster;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.Image;
import java.awt.Composite;
import java.awt.CompositeContext;
import java.awt.Color;
import java.awt.color.ColorSpace;
import java.awt.image.DataBuffer;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DirectColorModel;
import java.awt.GraphicsConfiguration;
import java.awt.Paint;
import java.awt.PaintContext;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.PathIterator;
import java.awt.geom.GeneralPath;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.FontMetrics;
import java.awt.Rectangle;
import java.text.AttributedCharacterIterator;
import sun.dc.pr.Rasterizer;
import sun.dc.pr.PRException;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.font.GlyphMetrics;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.image.ImageObserver;
import sun.awt.image.ImageRepresentation;
import java.awt.Shape;
import java.awt.image.ColorConvertOp;
import java.awt.Transparency;
import java.awt.Toolkit;
import java.awt.font.GlyphVector;
import java.awt.font.TextLayout;
import sun.awt.font.FontDesignMetrics;
import sun.java2d.pipe.CompositePipe;
import sun.java2d.pipe.PixelDrawPipe;
import sun.java2d.pipe.PixelFillPipe;
import sun.java2d.pipe.ShapeDrawPipe;
import sun.java2d.pipe.ValidatePipe;
import sun.java2d.pipe.PixelToShapeConverter;
import sun.java2d.loops.ImageData;
import sun.java2d.loops.RasterOutputManager;
import sun.java2d.pipe.ShapeSpanIterator;
import sun.java2d.pipe.Region;
import sun.java2d.pipe.RegionIterator;
import sun.java2d.pipe.TextPipe;
import sun.awt.font.NativeFontWrapper;
import java.awt.font.FontRenderContext;
import sun.awt.font.ShapingException;
import sun.java2d.loops.XORComposite;
import sun.awt.ConstrainableGraphics;
import sun.awt.SunHints;
import java.util.Map;
import java.util.Iterator;

/**
 * This is a the master Graphics2D superclass for all of the Sun
 * Graphics implementations.  This class relies on subclasses to
 * manage the various device information, but provides an overall
 * general framework for performing all of the requests in the
 * Graphics and Graphics2D APIs.
 *
 * @version 1.211 05/07/98
 * @author Jim Graham
 */
public abstract class SunGraphics2D
    extends Graphics2D
    implements ConstrainableGraphics, Cloneable
{
    public static final ColorModel XRGBModel =
        new DirectColorModel(24,0x00ff0000,0x0000ff00,0x000000ff);

    public static final int LOCK_OK                         = 1;
    public static final int UNLOCK_OK                       = 2;
    public static final int LOCK_FORMAT_PROBLEM             = 3;
    public static final int LOCK_GEOMETRY_PROBLEM           = 4;
    public static final int LOCK_DATA_STRUCT_PROBLEM        = 5;
    public static final int LOCK_NOT_LOCKED                 = 6;
    public static final int LOCK_ALREADY_LOCKED             = 7;

    /*
     * Attribute States
     */
    /* Paint */
    public static final int PAINT_TILE        = 2; /* delivered as Tile */
    public static final int PAINT_SINGLECOLOR = 1; /* Single Color */
    public static final int PAINT_SOLIDCOLOR  = 0; /* Solid single Color */

    /* Composite*/
    public static final int COMP_CUSTOM = 3;/* Custom Composite */
    public static final int COMP_XOR    = 2;/* XOR Mode Composite */
    public static final int COMP_ALPHA  = 1;/* AlphaComposite */
    public static final int COMP_ISCOPY = 0;/* SRC, extraAlpha = 1.0 or
					     * SRC_OVER, extraAlpha = 1.0 */

    /* Stroke */
    public static final int STROKE_CUSTOM = 1; /* custom Stroke */
    public static final int STROKE_BASIC  = 0; /* BasicStroke */

    /* Transform */
    public static final int TRANSFORM_GENERIC = 2; /* any 3x2 */
    public static final int TRANSFORM_TRANSLATEONLY = 1;
    public static final int TRANSFORM_ISIDENT = 0; /* Identity */

    /* Clipping */
    public static final int CLIP_SHAPE       = 2; /* arbitrary clip */
    public static final int CLIP_RECTANGULAR = 1; /* rectangular clip */
    public static final int CLIP_DEVICE      = 0; /* no clipping set */

    /* XOR Mode */
    public static final int XOR_MODE   = 1;
    public static final int PAINT_MODE = 0;

    public int paintState;
    public int compositeState;
    public int strokeState;
    public int transformState;
    public int clipState;
    public int modeState;

    protected static final Color nullXORColor = new Color(0, 0, 0, 0);

    public Color foregroundColor;
    public Color backgroundColor;
    public Color xorColor;

    public AffineTransform transform;
    public int transX;
    public int transY;
    public boolean complexTransform;
    
    protected static final Stroke defaultStroke = new BasicStroke();
    protected static final Composite defaultComposite = AlphaComposite.SrcOver;

    public Stroke stroke;
    public Composite composite;

    public Paint paint;
    public int renderHint;
    public int antialiasHint;
    private int textAntialiasHint;
    private int fractionalMetricsHint;
    private int interpolationHint;
    protected Font font;
    protected FontMetrics fontMetrics;
    public RenderingHints hints;

    public Rectangle constrainClip;		// lightweight bounds
    public int constrainX;
    public int constrainY;

    public Region clipRegion;
    public Shape usrClip;
    protected Rectangle devClip;		// Actual physical drawable
    protected Rectangle compBounds;

    protected static ValidatePipe invalidpipe = new ValidatePipe();

    public PixelDrawPipe drawpipe;
    public PixelFillPipe fillpipe;
    public ShapeDrawPipe shapepipe;
    public TextPipe textpipe;
    public CompositePipe alphapipe;
    private Object copyAreaStateCache;
    private Object rpics;
    public boolean simpleText;

    // cached state for text rendering
    // REMIND: think about when to invalidate these
    private Object dscc;
    private Object dvcc;
    private Object dccc;
    private Object dbcc;

    // This constructor is here because Graphics2D has one to prevent
    // javadoc from adding a default constructor
    protected SunGraphics2D() {
        simpleText = true;

	foregroundColor = Color.white;
	backgroundColor = Color.black;
	xorColor        = nullXORColor;

	transform = new AffineTransform();
	stroke = defaultStroke;
	composite = defaultComposite;
	paint = foregroundColor;

	renderHint = SunHints.INTVAL_RENDER_DEFAULT;
	antialiasHint = SunHints.INTVAL_ANTIALIAS_OFF;
	textAntialiasHint = SunHints.INTVAL_TEXT_ANTIALIAS_DEFAULT;
	fractionalMetricsHint = SunHints.INTVAL_FRACTIONALMETRICS_OFF;
	interpolationHint = -1;

	invalidatePipe();
    }

    protected Object clone() {
	try {
	    SunGraphics2D g = (SunGraphics2D) super.clone();
	    g.transform = new AffineTransform(this.transform);
	    if (hints != null) {
		g.hints = (RenderingHints) this.hints.clone();
	    }
	    g.invalidatePipe();
	    return g;
	} catch (CloneNotSupportedException e) {
	}
	return null;
    }

    /**
     * Create a new SunGraphics2D based on this one.
     */
    public Graphics create() {
	return (Graphics) clone();
    }

    public void setDevClip(int x, int y, int w, int h) {
	Rectangle r = new Rectangle(x, y, w, h);
	Rectangle c = constrainClip;
	if (c == null) {
	    devClip = r;
	} else {
	    devClip = r.intersection(c);
	}
	invalidateClip();
    }

    /**
     * Constrain rendering for lightweight objects.
     */
    public void constrain(int x, int y, int w, int h) {
	Rectangle c = constrainClip;
	if (c != null && (x|y) == 0 &&
	    (c.x >= constrainX) && (c.y >= constrainY) &&
	    (constrainX + w) >= (c.x + c.width) &&
	    (constrainY + h) >= (c.y + c.height))
	{
	    // Common case to ignore
	    return;
	}
	if ((x|y) != 0) {
	    constrainX += x;
	    constrainY += y;
	    // Note: we *pre* concatenate this translation here...
	    if (transformState == TRANSFORM_GENERIC) {
		AffineTransform at =
		    AffineTransform.getTranslateInstance(x, y);
		transform.preConcatenate(at);
		invalidateTransform();
	    } else {
		translate(x, y);
	    }
	}
	Rectangle r = new Rectangle(constrainX, constrainY, w, h);
	if (c != null) {
	    if (r.contains(c)) {
		return;
	    }
	    r = r.intersection(c);
	}
	constrainClip = r;
	if (!r.contains(devClip)) {
	    devClip = devClip.intersection(r);
	    invalidateClip();
	}
    }

    /*
     * Invalidate the pipeline
     */
    protected void invalidatePipe() {
	drawpipe = invalidpipe;
	fillpipe = invalidpipe;
	shapepipe = invalidpipe;
	textpipe = invalidpipe;
	alphapipe = null;
        dscc = null;
        dbcc = null;
        dccc = null;
        dvcc = null;
    }

    public void validatePipe() {
	getOutputManager().validatePipe(this);
    }

    /*
     * Intersect two Shapes by the simplest method, attempting to produce
     * a simplified result.
     * The boolean arguments keep1 and keep2 specify whether or not
     * the first or second shapes can be modified during the operation
     * or whether that shape must be "kept" unmodified.
     */
    Shape intersectShapes(Shape s1, Shape s2, boolean keep1, boolean keep2) {
	if (s1 instanceof Rectangle && s2 instanceof Rectangle) {
	    return ((Rectangle) s1).intersection((Rectangle) s2);
	}
	if (s1 instanceof Rectangle2D) {
	    return intersectRectShape((Rectangle2D) s1, s2, keep1, keep2);
	} else if (s2 instanceof Rectangle2D) {
	    return intersectRectShape((Rectangle2D) s2, s1, keep2, keep1);
	}
	return intersectByArea(s1, s2, keep1, keep2);
    }

    /*
     * Intersect a Rectangle with a Shape by the simplest method,
     * attempting to produce a simplified result.
     * The boolean arguments keep1 and keep2 specify whether or not
     * the first or second shapes can be modified during the operation
     * or whether that shape must be "kept" unmodified.
     */
    Shape intersectRectShape(Rectangle2D r, Shape s,
			     boolean keep1, boolean keep2) {
	if (s instanceof Rectangle2D) {
	    Rectangle2D r2 = (Rectangle2D) s;
	    Rectangle2D outrect;
	    if (!keep1) {
		outrect = r;
	    } else if (!keep2) {
		outrect = r2;
	    } else {
		outrect = new Rectangle2D.Float();
	    }
	    double x1 = r.getX();
	    double y1 = r.getY();
	    double x2 = r2.getX();
	    double y2 = r2.getY();
	    double x = Math.max(x1, x2);
	    double y = Math.max(y1, y2);
	    x1 += r.getWidth();
	    y1 += r.getHeight();
	    x2 += r2.getWidth();
	    y2 += r2.getHeight();
	    outrect.setFrameFromDiagonal(x, y,
					 Math.min(x1, x2),
					 Math.min(y1, y2));
	    return outrect;
	}
	if (r.contains(s.getBounds2D())) {
	    if (keep2) {
		s = cloneShape(s);
	    }
	    return s;
	}
	return intersectByArea(r, s, keep1, keep2);
    }

    protected static Shape cloneShape(Shape s) {
	return new GeneralPath(s);
    }
    
    /*
     * Intersect two Shapes using the Area class.  Presumably other
     * attempts at simpler intersection methods proved fruitless.
     * The boolean arguments keep1 and keep2 specify whether or not
     * the first or second shapes can be modified during the operation
     * or whether that shape must be "kept" unmodified.
     * @see #intersectShapes
     * @see #intersectRectShape
     */
    Shape intersectByArea(Shape s1, Shape s2, boolean keep1, boolean keep2) {
	Area a1, a2;

	// First see if we can find an overwriteable source shape
	// to use as our destination area to avoid duplication.
	if (!keep1 && (s1 instanceof Area)) {
	    a1 = (Area) s1;
	} else if (!keep2 && (s2 instanceof Area)) {
	    a1 = (Area) s2;
	    s2 = s1;
	} else {
	    a1 = new Area(s1);
	}

	if (s2 instanceof Area) {
	    a2 = (Area) s2;
	} else {
	    a2 = new Area(s2);
	}

	a1.intersect(a2);
	if (a1.isRectangular()) {
	    return a1.getBounds();
	}

	return a1;
    }

    /*
     * Intersect usrClip bounds and device bounds to determine the composite
     * rendering boundaries
     */
    public Rectangle getCompBounds() {
	Rectangle bounds;

	if ((bounds = compBounds) == null) {
	    switch (clipState) {
	    case CLIP_DEVICE:
		bounds = devClip;
		break;
	    case CLIP_RECTANGULAR:
		if (usrClip instanceof Rectangle) {
		    bounds = devClip.intersection((Rectangle) usrClip);
		} else {
		    bounds = devClip.intersection(usrClip.getBounds());
		}
		break;
	    case CLIP_SHAPE:
		PathIterator cpi = usrClip.getPathIterator(null);
		int box[] = new int[4];

		ShapeSpanIterator sr = new ShapeSpanIterator();
		sr.setOutputArea(devClip);
		sr.appendPath(cpi);
		sr.getPathBox(box);
		bounds = new Rectangle(box[0], box[1],
				       box[2] - box[0],
				       box[3] - box[1]);
		bounds = bounds.intersection(devClip);
		Region r = new Region();
		r.setOutputArea(bounds);
		r.appendSpans(sr);
		clipRegion = r;
		break;
	    }
	    compBounds = bounds;
	}

	return bounds;
    }
    
    /*
     * Convert a BufferedImage into another BufferedImage with a desired
     * ColorModel.
     */
    protected BufferedImage convertCM(BufferedImage src, ColorModel cm) {
        WritableRaster wr = 
              cm.createCompatibleWritableRaster(src.getWidth(), 
                                                 src.getHeight());
        BufferedImage dst = new BufferedImage(cm, wr,
                                              src.isAlphaPremultiplied(),
                                              null);
        for (int i = 0 ; i < src.getHeight() ; i++) {
            for (int j = 0 ; j < src.getWidth() ; j++) {
                dst.setRGB(j, i, src.getRGB(j, i));
            }
        }
        return dst;
    }

    /*
     * Convert a given Raster to the desired data format.
     */
    public WritableRaster convertRaster(Raster inRaster, 
                                         ColorModel inCM, 
                                         ColorModel outCM) {
        // Use a faster conversion if this is an IndexColorModel
        if (inCM instanceof IndexColorModel &&
            ((outCM.equals(ColorModel.getRGBdefault())) ||
             (outCM.equals(XRGBModel)))) {
            IndexColorModel icm = (IndexColorModel) inCM;
            BufferedImage dbi = icm.convertToIntDiscrete(inRaster, false);
            return dbi.getRaster();
        }

        BufferedImage dbi =
            new BufferedImage(outCM,
                    outCM.createCompatibleWritableRaster(inRaster.getWidth(),
                              inRaster.getHeight()),
                              outCM.isAlphaPremultiplied(),
                              null);
        
        //     ColorSpace[] cs = {inCM.getColorSpace(), outCM.getColorSpace()};
        //     ColorConvertOp cOp = new ColorConvertOp(cs);
        //     cOp.filter(sbi, dbi);

        // use this slow method to convert untill ColorConvertOp is available.
        // Does not take in to account quality dithering if applicable.
        Object buffer = inRaster.getDataElements(0, 0, null);
        for (int i = 0 ; i < dbi.getHeight() ; i++) {
            for (int j = 0 ; j < dbi.getWidth() ; j++) {
                dbi.setRGB(j, i, inCM.getRGB(inRaster.getDataElements(j,i,buffer)));
            }
        }

        return dbi.getRaster();
    }

    private static float veclength(float x, float y) {
	return (float) Math.sqrt(x * x + y * y);
    }

    public Font getFont() {
        if (font == null) {
            font = new Font("Dialog", 10, Font.PLAIN);
        }
        return font;
    }

    public void setFont(Font font) {
        if (font != null && font != this.font) {
            this.font = font;
            this.fontMetrics = null;
            dscc = null;
            dbcc = null;
            dccc = null;
            dvcc = null;
        }
    }

    private static class CachedMetrics {
	public Font font;
	public double matrix[];
	public boolean isAA;
	public boolean isFract;
	public FontMetrics fm;
	public CachedMetrics(Font font, double matrix[],
			     boolean isAA, boolean isFract,
			     FontMetrics fm) {
	    this.font = font;
	    this.matrix = matrix;
	    this.isAA = isAA;
	    this.isFract = isFract;
	    this.fm = fm;
	}
	public final boolean matches(Font font, double matrix[],
				     boolean isAA, boolean isFract) {
	    return ((this.font == font) &&
		    (this.isAA == isAA) &&
		    (this.isFract == isFract) &&
		    (this.matrix[0] == matrix[0]) &&
		    (this.matrix[1] == matrix[1]) &&
		    (this.matrix[2] == matrix[2]) &&
		    (this.matrix[3] == matrix[3]));
	}
    }

    private static final float metrics[] = new float[4];
    public static CachedMetrics metricscache[] = new CachedMetrics[5];
    public synchronized static FontMetrics
	makeFontMetrics(Font font, double matrix[],
			boolean isAA, boolean isFract)
    {
	int i;
	for (i = 0; i < metricscache.length; i++) {
	    CachedMetrics cm = metricscache[i];
	    if (cm == null) {
		break;
	    }
	    if (cm.matches(font, matrix, isAA, isFract)) {
		if (i > 1) {
		    System.arraycopy(metricscache, 0, metricscache, 1, i);
		    metricscache[0] = cm;
		}
		return cm.fm;
	    }
	}
        NativeFontWrapper.getFontMetrics(font, matrix, isAA, isFract, metrics);
        FontMetrics fm = new FontDesignMetrics(font,
					       metrics[0], metrics[1],
					       metrics[2], metrics[3]);
	if (i == metricscache.length) {
	    System.arraycopy(metricscache, 0, metricscache, 1, i-1);
	    i = 0;
	}
	// Note - don't need to copy matrix since it is local to caller
	metricscache[i] = new CachedMetrics(font, matrix, isAA, isFract, fm);
	return fm;
    }

    public FontMetrics getFontMetrics(Font font) {
        if ((this.fontMetrics != null) && (font == this.font)) {
            return this.fontMetrics;
        }
        // REMIND: debate adding user transform here...
        // (probably not a good idea to add it)
        float ptSize = font.getSize2D();
        AffineTransform tx = font.getTransform();
        tx.scale(ptSize, ptSize);
        double matrix[] = new double[4];
	tx.getMatrix(matrix);
        int aahint =
	    (textAntialiasHint == SunHints.INTVAL_TEXT_ANTIALIAS_DEFAULT
	     ? antialiasHint : textAntialiasHint);
        FontMetrics fm =
	    makeFontMetrics(font, matrix,
			    (aahint == SunHints.INTVAL_ANTIALIAS_ON), 
			    (fractionalMetricsHint
			     == SunHints.INTVAL_FRACTIONALMETRICS_ON));
        if (this.font == font) {
            this.fontMetrics = fm;
        }
        return fm;
    }

    /**
     * Checks to see if a Path intersects the specified Rectangle in device
     * space.  The rendering attributes taken into account include the
     * clip, transform, and stroke attributes.
     * @param rect The area in device space to check for a hit.
     * @param p The path to check for a hit.
     * @param onStroke Flag to choose between testing the stroked or
     * the filled path.
     * @return True if there is a hit, false otherwise.
     * @see #setStroke
     * @see #fillPath
     * @see #drawPath
     * @see #transform
     * @see #setTransform
     * @see #clip
     * @see #setClip
     */
    public boolean hit(Rectangle rect, Shape s, boolean onStroke) {
        if (onStroke) {
            s = stroke.createStrokedShape(s);
        }

        s = transformShape(s);
	if ((constrainX|constrainY) != 0) {
	    rect = new Rectangle(rect);
	    rect.translate(constrainX, constrainY);
	}

        return s.intersects(rect);
    }

    /**
     * Return the ColorModel associated with this Graphics2D.
     */
    public abstract ColorModel getDeviceColorModel();

    /**
     * Return the device configuration associated with this Graphics2D.
     */
    public abstract GraphicsConfiguration getDeviceConfiguration();

    /**
     * Return the OutputManager object assigned to manage this Graphics2D.
     */
    public abstract OutputManager getOutputManager();


    /**
     * Sets the Composite in the current graphics state. Composite is used
     * in all drawing methods such as drawImage, drawString, drawPath,
     * and fillPath.  It specifies how new pixels are to be combined with
     * the existing pixels on the graphics device in the rendering process.
     * @param comp The Composite object to be used for drawing.
     * @see java.awt.Graphics#setXORMode
     * @see java.awt.Graphics#setPaintMode
     * @see AlphaComposite
     */
    public void setComposite(Composite comp) {
	if (comp == null) {
	    throw new IllegalArgumentException("null Composite");
	}
	int newCompState;
	int newModeState;
        if (comp instanceof AlphaComposite) {
	    newModeState = PAINT_MODE;
	    AlphaComposite alphacomp = (AlphaComposite) comp;
            if (alphacomp.getAlpha() >= 1.0f &&
                (alphacomp.getRule() == AlphaComposite.SRC ||
		 alphacomp.getRule() == AlphaComposite.SRC_OVER))
	    {
		newCompState = COMP_ISCOPY;
            } else {
                newCompState = COMP_ALPHA;
            }
        } else if (comp instanceof XORComposite) {
            newModeState = XOR_MODE;
            newCompState = COMP_XOR;
        } else {
	    checkCustomComposite();
	    newModeState = PAINT_MODE;
            newCompState = COMP_CUSTOM;
        }
        if (compositeState != newCompState || modeState != newModeState) {
	    compositeState = newCompState;
	    modeState = newModeState;
	    invalidatePipe();
        }
        else {
            // Still need to invalidate text render cache
            dscc = null;
            dbcc = null;
            dccc = null;
            dvcc = null;
        }
	composite = comp;
    }

    static java.security.Permission compPermission;

    /*
     * BufferedImageGraphics2D will override this to allow
     * custom composites in all cases.
     */
    protected void checkCustomComposite() {
	SecurityManager sm = System.getSecurityManager();
	if (sm != null) {
	    if (compPermission == null) {
		compPermission =
		    new java.awt.AWTPermission("readDisplayPixels");
	    }
	    sm.checkPermission(compPermission);
	}
    }

    /**
     * Sets the Paint in the current graphics state.
     * @param paint The Paint object to be used to generate color in
     * the rendering process.
     * @see java.awt.Graphics#setColor
     * @see GradientPaint
     * @see TexturePaint
     */
    public void setPaint(Paint paint) {
	if (paint == null) {
	    return;
	}
        int savePaintState = paintState;
        this.paint = paint;
        if (paint instanceof Color) {
	    foregroundColor = (Color) paint;
            if (foregroundColor.getAlpha() == 255) {
                paintState = PAINT_SOLIDCOLOR;
            } else {
                paintState = PAINT_SINGLECOLOR;
            }
        } else {
            paintState = PAINT_TILE;
        }
        if (paintState != savePaintState) {
	    invalidatePipe();
        }
        else {
            // Still need to invalidate text render cache
            dscc = null;
            dbcc = null;
            dccc = null;
            dvcc = null;
        }
    }

    /*
     * Sets the Stroke in the current graphics state.
     * @param s The Stroke object to be used to stroke a Path in
     * the rendering process.
     * @see BasicStroke
     */
    public void setStroke(Stroke s) {
	if (s == null) {
	    throw new IllegalArgumentException("null Stroke");
	}
        int saveStrokeState = strokeState;
        boolean fInvalidate = !stroke.equals(s);
        stroke = s;
        if (s instanceof BasicStroke) {
            strokeState = STROKE_BASIC;
        } else {
            fInvalidate = true;
            strokeState = STROKE_CUSTOM;
        }
        if (fInvalidate) {
	    invalidatePipe();
        }
    }

    /**
     * Sets the preferences for the rendering algorithms.
     * Hint categories include controls for rendering quality and
     * overall time/quality trade-off in the rendering process.
     * @param hintKey The key of hint to be set. The strings are
     * defined in the RenderingHints class.  
     * @param hintValue The value indicating preferences for the specified
     * hint category. These strings are defined in the RenderingHints
     * class.  
     * @see RenderingHints
     */
    public void setRenderingHint(Key hintKey, Object hintValue) {
	// If we recognize the key, we must recognize the value
	//     otherwise throw an IllegalArgumentException
	//     and don't change the Hints object
	// If we don't recognize the key, just pass it through
	//     to the Hints object untouched
	if (!hintKey.isCompatibleValue(hintValue)) {
	    throw new IllegalArgumentException
		(hintValue+" is not compatible with "+hintKey);
	}
	if (hintKey instanceof SunHints.Key) {
	    boolean stateChanged;
	    boolean recognized = true;
	    SunHints.Key sunKey = (SunHints.Key) hintKey;
	    int newHint = ((SunHints.Value) hintValue).getIndex();
	    switch (sunKey.getIndex()) {
	    case SunHints.INTKEY_RENDERING:
		stateChanged = (renderHint != newHint);
		renderHint = newHint;
		break;
	    case SunHints.INTKEY_ANTIALIASING:
		stateChanged = (antialiasHint != newHint);
		antialiasHint = newHint;
		break;
	    case SunHints.INTKEY_TEXT_ANTIALIASING:
		stateChanged = (textAntialiasHint != newHint);
		if (stateChanged) {
		    fontMetrics = null; 
		}
		textAntialiasHint = newHint;
		break;
	    case SunHints.INTKEY_FRACTIONALMETRICS:
		stateChanged = (fractionalMetricsHint != newHint);
		if (stateChanged) {
		    fontMetrics = null; 
		}
		fractionalMetricsHint = newHint;
		break;
	    case SunHints.INTKEY_INTERPOLATION:
		switch (newHint) {
		case SunHints.INTVAL_INTERPOLATION_BICUBIC:
		case SunHints.INTVAL_INTERPOLATION_BILINEAR:
		    newHint = AffineTransformOp.TYPE_BILINEAR;
		    break;
		default:
		case SunHints.INTVAL_INTERPOLATION_NEAREST_NEIGHBOR:
		    newHint = AffineTransformOp.TYPE_NEAREST_NEIGHBOR;
		    break;
		}
		stateChanged = (interpolationHint != newHint);
		interpolationHint = newHint;
		break;
	    default:
		recognized = false;
		stateChanged = false;
		break;
	    }
	    if (recognized) {
		if (stateChanged) {
		    invalidatePipe();
		}
		if (hints != null) {
		    hints.put(hintKey, hintValue);
		}
		return;
	    }
	}
	// Nothing we recognize so none of "our state" has changed
	if (hints == null) {
	    hints = makeHints(null);
	}
	hints.put(hintKey, hintValue);
    }


    /**
     * Returns the preferences for the rendering algorithms.
     * @param hintCategory The category of hint to be set. The strings
     * are defined in the RenderingHints class.
     * @return The preferences for rendering algorithms. The strings
     * are defined in the RenderingHints class.
     * @see RenderingHints
     */
    public Object getRenderingHint(Key hintKey) {
	if (hints != null) {
	    return hints.get(hintKey);
	}
	if (!(hintKey instanceof SunHints.Key)) {
	    return null;
	}
	int keyindex = ((SunHints.Key)hintKey).getIndex();
	switch (keyindex) {
	case SunHints.INTKEY_RENDERING:
	    return SunHints.Value.get(SunHints.INTKEY_RENDERING,
				      renderHint);
	case SunHints.INTKEY_ANTIALIASING:
	    return SunHints.Value.get(SunHints.INTKEY_ANTIALIASING,
				      antialiasHint);
	case SunHints.INTKEY_TEXT_ANTIALIASING:
	    return SunHints.Value.get(SunHints.INTKEY_TEXT_ANTIALIASING,
				      textAntialiasHint);
	case SunHints.INTKEY_FRACTIONALMETRICS:
	    return SunHints.Value.get(SunHints.INTKEY_FRACTIONALMETRICS,
				      fractionalMetricsHint);
	case SunHints.INTKEY_INTERPOLATION:
	    switch (interpolationHint) {
	    case AffineTransformOp.TYPE_NEAREST_NEIGHBOR:
		return SunHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR;
	    case AffineTransformOp.TYPE_BILINEAR:
		return SunHints.VALUE_INTERPOLATION_BILINEAR;
//	    case AffineTransformOp.TYPE_BICUBIC:
//		return SunHints.VALUE_INTERPOLATION_BICUBIC;
	    }
	    return null;
	}
	return null;
    }

    /**
     * Sets the preferences for the rendering algorithms.
     * Hint categories include controls for rendering quality and
     * overall time/quality trade-off in the rendering process.
     * @param hints The rendering hints to be set
     * @see RenderingHints
     */
    public void setRenderingHints(Map hints) {
	this.hints = null;
	renderHint = SunHints.INTVAL_RENDER_DEFAULT;
	antialiasHint = SunHints.INTVAL_ANTIALIAS_OFF;
	textAntialiasHint = SunHints.INTVAL_TEXT_ANTIALIAS_DEFAULT;
	fractionalMetricsHint = SunHints.INTVAL_FRACTIONALMETRICS_OFF;
	interpolationHint = -1;
	boolean customHintPresent = false;
	Iterator iter = hints.keySet().iterator();
	while (iter.hasNext()) {
	    Object key = iter.next();
	    if (key == SunHints.KEY_RENDERING ||
		key == SunHints.KEY_ANTIALIASING ||
		key == SunHints.KEY_TEXT_ANTIALIASING ||
		key == SunHints.KEY_FRACTIONALMETRICS ||
		key == SunHints.KEY_INTERPOLATION)
	    {
		setRenderingHint((Key) key, hints.get(key));
	    } else {
		customHintPresent = true;
	    }
	}
	if (customHintPresent) {
	    this.hints = makeHints(hints);
	}
	invalidatePipe();
    }

    /**
     * Adds a number of preferences for the rendering algorithms.
     * Hint categories include controls for rendering quality and
     * overall time/quality trade-off in the rendering process.
     * @param hints The rendering hints to be set
     * @see RenderingHints
     */
    public void addRenderingHints(Map hints) {
	boolean customHintPresent = false;
	Iterator iter = hints.keySet().iterator();
	while (iter.hasNext()) {
	    Object key = iter.next();
	    if (key == SunHints.KEY_RENDERING ||
		key == SunHints.KEY_ANTIALIASING ||
		key == SunHints.KEY_TEXT_ANTIALIASING ||
		key == SunHints.KEY_FRACTIONALMETRICS ||
		key == SunHints.KEY_INTERPOLATION)
	    {
		setRenderingHint((Key) key, hints.get(key));
	    } else {
		customHintPresent = true;
	    }
	}
	if (customHintPresent) {
	    if (this.hints == null) {
		this.hints = makeHints(hints);
	    } else {
		this.hints.putAll(hints);
	    }
	}
    }

    /**
     * Gets the preferences for the rendering algorithms.
     * Hint categories include controls for rendering quality and
     * overall time/quality trade-off in the rendering process.
     * @see RenderingHints
     */
    public RenderingHints getRenderingHints() {
	if (hints == null) {
	    return makeHints(null);
	} else {
	    return (RenderingHints) hints.clone();
	}
    }

    RenderingHints makeHints(Map hints) {
	RenderingHints model = new RenderingHints(hints);
	model.put(SunHints.KEY_RENDERING,
		  SunHints.Value.get(SunHints.INTKEY_RENDERING,
				     renderHint));
	model.put(SunHints.KEY_ANTIALIASING,
		  SunHints.Value.get(SunHints.INTKEY_ANTIALIASING,
				     antialiasHint));
	model.put(SunHints.KEY_TEXT_ANTIALIASING,
		  SunHints.Value.get(SunHints.INTKEY_TEXT_ANTIALIASING,
				     textAntialiasHint));
	model.put(SunHints.KEY_FRACTIONALMETRICS,
		  SunHints.Value.get(SunHints.INTKEY_FRACTIONALMETRICS,
				     fractionalMetricsHint));
	Object value;
	switch (interpolationHint) {
	case AffineTransformOp.TYPE_NEAREST_NEIGHBOR:
	    value = SunHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR;
	    break;
	case AffineTransformOp.TYPE_BILINEAR:
	    value = SunHints.VALUE_INTERPOLATION_BILINEAR;
	    break;
//	case AffineTransformOp.TYPE_BICUBIC:
//	    value = SunHints.VALUE_INTERPOLATION_BICUBIC;
//	    break;
	default:
	    return model;
	}
	model.put(SunHints.KEY_INTERPOLATION, value);
	return model;
    }
    
    /**
     * Concatenates the current transform of this Graphics2D with a
     * translation transformation.
     * This is equivalent to calling transform(T), where T is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   1    0    tx  ]
     *		[   0    1    ty  ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void translate(double tx, double ty) {
	transform.translate(tx, ty);
	invalidateTransform();
    }

    /**
     * Concatenates the current transform of this Graphics2D with a
     * rotation transformation.
     * This is equivalent to calling transform(R), where R is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   cos(theta)    -sin(theta)    0   ]
     *		[   sin(theta)     cos(theta)    0   ]
     *		[       0              0         1   ]
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     */
    public void rotate(double theta) {
	transform.rotate(theta);
	invalidateTransform();
    }

    /**
     * Concatenates the current transform of this Graphics2D with a
     * translated rotation transformation.
     * This is equivalent to the following sequence of calls:
     * <pre>
     *		translate(x, y);
     *		rotate(theta);
     *		translate(-x, -y);
     * </pre>
     * Rotating with a positive angle theta rotates points on the positive
     * x axis toward the positive y axis.
     * @param theta The angle of rotation in radians.
     * @param x The x coordinate of the origin of the rotation
     * @param y The x coordinate of the origin of the rotation
     */
    public void rotate(double theta, double x, double y) {
	transform.rotate(theta, x, y);
	invalidateTransform();
    }

    /**
     * Concatenates the current transform of this Graphics2D with a
     * scaling transformation.
     * This is equivalent to calling transform(S), where S is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   sx   0    0   ]
     *		[   0    sy   0   ]
     *		[   0    0    1   ]
     * </pre>
     */
    public void scale(double sx, double sy) {
	transform.scale(sx, sy);
	invalidateTransform();
    }

    /**
     * Concatenates the current transform of this Graphics2D with a
     * shearing transformation.
     * This is equivalent to calling transform(SH), where SH is an
     * AffineTransform represented by the following matrix:
     * <pre>
     *		[   1   shx   0   ]
     *		[  shy   1    0   ]
     *		[   0    0    1   ]
     * </pre>
     * @param shx The factor by which coordinates are shifted towards the
     * positive X axis direction according to their Y coordinate
     * @param shy The factor by which coordinates are shifted towards the
     * positive Y axis direction according to their X coordinate
     */
    public void shear(double shx, double shy) {
	transform.shear(shx, shy);
	invalidateTransform();
    }

    /**
     * Composes a Transform object with the transform in this
     * Graphics2D according to the rule last-specified-first-applied.
     * If the currrent transform is Cx, the result of composition
     * with Tx is a new transform Cx'.  Cx' becomes the current
     * transform for this Graphics2D.
     * Transforming a point p by the updated transform Cx' is
     * equivalent to first transforming p by Tx and then transforming
     * the result by the original transform Cx.  In other words,
     * Cx'(p) = Cx(Tx(p)).
     * A copy of the Tx is made, if necessary, so further
     * modifications to Tx do not affect rendering.
     * @param Tx The Transform object to be composed with the current
     * transform.
     * @see #setTransform
     * @see AffineTransform
     */
    public void transform(AffineTransform xform) {
        this.transform.concatenate(xform); 
	invalidateTransform();
    }
    
    /**
     * Translate
     */
    public void translate(int x, int y) {
	transform.translate(x, y);
	if (!complexTransform) {
	    transX += x;
	    transY += y;
	    transformState = ((transX == 0 && transY == 0) ?
			      TRANSFORM_ISIDENT : TRANSFORM_TRANSLATEONLY);
	    invalidatePipe();
	} else {
	    // Once invalidateTransform is ever called,
	    // complexTransform becomes true
	    invalidateTransform();
	}
    }

    /**
     * Sets the Transform in the current graphics state.
     * @param Tx The Transform object to be used in the rendering process.
     * @see #transform
     * @see TransformChain
     * @see AffineTransform
     */
    public void setTransform(AffineTransform Tx) {
	if ((constrainX|constrainY) == 0) {
	    transform.setTransform(Tx);
	} else {
	    transform.setToTranslation(constrainX, constrainY);
	    transform.concatenate(Tx);
	}
	invalidateTransform();
    }

    protected void invalidateTransform() {
	complexTransform = true;	// Do not shortcut translate(int, int)
	int type = transform.getType();
	if (type == AffineTransform.TYPE_IDENTITY) {
            transformState = TRANSFORM_ISIDENT;
	    complexTransform = false;
	    transX = transY = 0;
	} else if (type == AffineTransform.TYPE_TRANSLATION) {
	    transformState = TRANSFORM_TRANSLATEONLY;
	    // transX/Y Should not be used in this mode, but just in case...
	    transX = (int) transform.getTranslateX();
	    transY = (int) transform.getTranslateY();
	} else {
	    transformState = TRANSFORM_GENERIC;
	    transX = transY = 0;
	}
	invalidatePipe();
    }

    /**
     * Returns the current Transform in the Graphics2D state.
     * @see #transform
     * @see #setTransform
     */
    public AffineTransform getTransform() {
	if ((constrainX|constrainY) == 0) {
	    return new AffineTransform(transform);
	}
	AffineTransform tx =
	    AffineTransform.getTranslateInstance(-constrainX, -constrainY);
	tx.concatenate(transform);
        return tx;
    }

    /**
     * Returns the current Transform ignoring the "constrain"
     * rectangle.
     */
    public AffineTransform cloneTransform() {
	return new AffineTransform(transform);
    }

    /**
     * Returns the current Paint in the Graphics2D state.
     * @see #setPaint
     * @see java.awt.Graphics#setColor
     */
    public Paint getPaint() {
        return paint;
    }

    /**
     * Returns the current Composite in the Graphics2D state.
     * @see #setComposite
     */
    public Composite getComposite() {
        return composite;
    }

    public Color getColor() {
        return foregroundColor;
    }

    public void setColor(Color color) {
	if (paint != color) {
	    setPaint(color);
	}
    }

    /**
     * Sets the background color in this context used for clearing a region.
     * When Graphics2D is constructed for a component, the backgroung color is
     * inherited from the component. Setting the background color in the
     * Graphics2D context only affects the subsequent clearRect() calls and
     * not the background color of the component. To change the background
     * of the component, use appropriate methods of the component.
     * @param color The background color that should be used in
     * subsequent calls to clearRect().
     * @see getBackground
     * @see Graphics.clearRect()
     */
    public void setBackground(Color color) {
        backgroundColor = color;
    }

    /**
     * Returns the background color used for clearing a region.
     * @see setBackground
     */
    public Color getBackground() {
        return backgroundColor;
    }

    /**
     * Returns the current Stroke in the Graphics2D state.
     * @see setStroke
     */
    public Stroke getStroke() {
        return stroke;
    }
    
    public Rectangle getClipBounds() {
	Shape s = getClip();
        if (s != null) {
            return s.getBounds();
        } else {
            return null;
        }
    }

    protected void invalidateClip() {
	compBounds = null;
	clipRegion = null;
	if (usrClip == null) {
	    clipState = CLIP_DEVICE;
	} else if (usrClip instanceof Rectangle2D) {
	    clipState = CLIP_RECTANGULAR;
	} else {
	    clipState = CLIP_SHAPE;
	}
	invalidatePipe();
    }

    static final int NON_RECTILINEAR_TRANSFORM_MASK =
	(AffineTransform.TYPE_GENERAL_TRANSFORM |
	 AffineTransform.TYPE_GENERAL_ROTATION);

    protected Shape transformShape(Shape s) {
	if (s == null) {
	    return null;
	}
	if (complexTransform) {
	    return transformShape(transform, s);
	} else {
	    return transformShape(transX, transY, s);
	}
    }

    protected Shape untransformShape(Shape s) {
	if (s == null) {
	    return null;
	}
	if (complexTransform) {
	    try {
		return transformShape(transform.createInverse(), s);
	    } catch (NoninvertibleTransformException e) {
		return null;
	    }
	} else {
	    return transformShape(-transX, -transY, s);
	}
    }

    protected static Shape transformShape(int tx, int ty, Shape s) {
	if (s == null) {
	    return null;
	}

	if (s instanceof Rectangle) {
	    Rectangle r = s.getBounds();
	    r.translate(tx, ty);
	    return r;
	}
	if (s instanceof Rectangle2D) {
	    Rectangle2D rect = (Rectangle2D) s;
	    return new Rectangle2D.Double(rect.getX() + ty,
					  rect.getY() + ty,
					  rect.getWidth(),
					  rect.getHeight());
	}

	if (tx == 0 && ty == 0) {
	    return cloneShape(s);
	}

	AffineTransform mat = AffineTransform.getTranslateInstance(tx, ty);
	return mat.createTransformedShape(s);
    }

    protected static Shape transformShape(AffineTransform tx, Shape clip) {
        if (clip == null) {
            return null;
        }

	if (clip instanceof Rectangle2D &&
	    (tx.getType() & NON_RECTILINEAR_TRANSFORM_MASK) == 0)
	{
	    Rectangle2D rect = (Rectangle2D) clip;
	    double matrix[] = new double[4];
	    matrix[0] = rect.getX();
	    matrix[1] = rect.getY();
	    matrix[2] = matrix[0] + rect.getWidth();
	    matrix[3] = matrix[1] + rect.getHeight();
	    tx.transform(matrix, 0, matrix, 0, 2);
	    rect = new Rectangle2D.Float();
	    rect.setFrameFromDiagonal(matrix[0], matrix[1],
				      matrix[2], matrix[3]);
	    return rect;
	}

        if (tx.isIdentity()) {
	    return cloneShape(clip);
        }

	return tx.createTransformedShape(clip);
    }

    public void clipRect(int x, int y, int w, int h) {
	clip(new Rectangle(x, y, w, h));
    }

    public void setClip(int x, int y, int w, int h) {
	setClip(new Rectangle(x, y, w, h));
    }

    public Shape getClip() {
	return untransformShape(usrClip);
    }

    public void setClip(Shape sh) {
	usrClip = transformShape(sh);
	invalidateClip();
    }

    /**
     * Intersects the current clip with the specified Path and sets the
     * current clip to the resulting intersection. The clip is transformed
     * with the current transform in the Graphics2D state before being
     * intersected with the current clip. This method is used to make the
     * current clip smaller. To make the clip larger, use any setClip method.
     * @param p The Path to be intersected with the current clip.
     */
    public void clip(Shape s) {
	s = transformShape(s);
	if (usrClip != null) {
	    s = intersectShapes(usrClip, s, true, true);
	}
	usrClip = s;
	invalidateClip();
    }

    public void setPaintMode() {
        modeState = PAINT_MODE;
	xorColor = nullXORColor;
        setComposite(AlphaComposite.SrcOver);
    }
    
    public void setXORMode(Color c) {
	if (c == null) {
	    throw new IllegalArgumentException("null XORColor");
	}
        modeState = XOR_MODE;
        xorColor = c;
        setComposite(new XORComposite(c));
    }

    public void copyArea(int x, int y, int w, int h, int dx, int dy) {
	Rectangle rect = new Rectangle(x, y, w, h);
	rect = transformBounds(rect, transform);
        Point2D    point = new Point2D.Float(dx, dy);
        Point2D    root  = new Point2D.Float(0, 0);
        point = transform.transform(point, point);
        root  = transform.transform(root, root);
        int fdx = (int)(point.getX()-root.getX());
        int fdy = (int)(point.getY()-root.getY());

        Rectangle r = getCompBounds().intersection(rect.getBounds());

        if (r.isEmpty()) {
            return;
        }
        
        // Begin Rasterizer for Clip Shape
        boolean skipClip = true;
        byte[] clipAlpha = null;

        if (clipState == CLIP_SHAPE) {

	    int box[] = new int[4];

	    clipRegion.getBounds(box);
	    Rectangle devR = new Rectangle(box[0], box[1],
					   box[2] - box[0],
					   box[3] - box[1]);
	    if (!devR.isEmpty()) {
		OutputManager mgr = getOutputManager();
		RegionIterator ri = clipRegion.getIterator();
		while (ri.nextYRange(box)) {
		    int spany = box[0];
		    int spanh = box[1] - spany;
		    while (ri.nextXBand(box)) {
			int spanx = box[2];
			int spanw = box[3] - spanx;
			copyAreaStateCache = mgr.copyArea(this, null,
                                                          spanw, 0,
                                                          spanx, spany,
                                                          spanw, spanh,
                                                          fdx, fdy,
                                                          copyAreaStateCache);
		    }
		}
	    }
	    return;
	}
        // End Rasterizer for Clip Shape
        
        copyAreaStateCache = getOutputManager().copyArea(this, null,
                                                         r.width, 0,
                                                         r.x, r.y, r.width,
                                                         r.height, fdx, fdy,
                                                         copyAreaStateCache);
    }

    public void drawLine(int x1, int y1, int x2, int y2) {
	drawpipe.drawLine(this, x1, y1, x2, y2);
    }

    public void drawRoundRect(int x, int y, int w, int h, int arcW, int arcH) {
	drawpipe.drawRoundRect(this, x, y, w, h, arcW, arcH);
    }

    public void fillRoundRect(int x, int y, int w, int h, int arcW, int arcH) {
	fillpipe.fillRoundRect(this, x, y, w, h, arcW, arcH);
    }

    public void drawOval(int x, int y, int w, int h) {
	drawpipe.drawOval(this, x, y, w, h);
    }

    public void fillOval(int x, int y, int w, int h) {
	fillpipe.fillOval(this, x, y, w, h);
    }

    public void drawArc(int x, int y, int w, int h,
			int startAngl, int arcAngl) {
	drawpipe.drawArc(this, x, y, w, h, startAngl, arcAngl);
    }

    public void fillArc(int x, int y, int w, int h,
			int startAngl, int arcAngl) {
	fillpipe.fillArc(this, x, y, w, h, startAngl, arcAngl);
    }

    public void drawPolyline(int xPoints[], int yPoints[], int nPoints) {
	drawpipe.drawPolyline(this, xPoints, yPoints, nPoints);
    }

    public void drawPolygon(int xPoints[], int yPoints[], int nPoints) {
	drawpipe.drawPolygon(this, xPoints, yPoints, nPoints);
    }

    public void fillPolygon(int xPoints[], int yPoints[], int nPoints) {
	fillpipe.fillPolygon(this, xPoints, yPoints, nPoints);
    }

    public void drawRect (int x, int y, int w , int h) {
	drawpipe.drawRect(this, x, y, w, h);
    }
    
    public void fillRect (int x, int y, int w , int h) {
	fillpipe.fillRect(this, x, y, w, h);
    }

    public void clearRect(int x, int y, int w, int h) {
        Composite c = composite;
	Paint p = paint;
	Color xorc = xorColor;
	boolean isxor = (modeState == XOR_MODE);
        setComposite(AlphaComposite.Src);
        // REMIND: has some "interesting" consequences if threads are
        // not synchronized
        setColor(getBackground());
	if (isxor) {
	    setPaintMode();
	}
	validatePipe();
        fillRect(x, y, w, h);
	if (isxor) {
	    setXORMode(xorc);
	}
        setPaint(p);
	setComposite(c);
    }

    /**
     * Strokes the outline of a Path using the settings of the current
     * graphics state.  The rendering attributes applied include the
     * clip, transform, paint or color, composite and stroke attributes.
     * @param p The path to be drawn.
     * @see #setStroke
     * @see #setPaint
     * @see java.awt.Graphics#setColor
     * @see #transform
     * @see #setTransform
     * @see #clip
     * @see #setClip
     * @see #setComposite
     */
    public void draw(Shape s) {
	shapepipe.draw(this, s);
    }


    /**
     * Fills the interior of a Path using the settings of the current
     * graphics state. The rendering attributes applied include the
     * clip, transform, paint or color, and composite.
     * @see #setPaint
     * @see java.awt.Graphics#setColor
     * @see #transform
     * @see #setTransform
     * @see #setComposite
     * @see #clip
     * @see #setClip
     */
    public void fill(Shape s) {
	shapepipe.fill(this, s);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a solid background
     * color and a callback object.
     */
    protected boolean drawImage(BufferedImage bImg, int x, int y, Color bg,
                                ImageObserver observer)
    {
        renderingPipeImage(bImg, x, y, null, null, bg);
        return true;
    }
    
    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * solid background color and a callback object.
     */
    protected boolean drawImage(BufferedImage bImg, int x, int y,
                                int width, int height,
                                Color bg, ImageObserver observer)
    {
	if (width == 0 || height == 0) {
	    return true;
	}
        // Append the new transform
        AffineTransform atfm = null;
        if (width != bImg.getWidth() || height != bImg.getHeight()) {
            atfm = new AffineTransform();
            atfm.scale((double)width/bImg.getWidth(),
                       (double)height/bImg.getHeight());
        }

        return drawImage(bImg, atfm, x, y, bg, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle in
     * nonblocking mode with a solid background color and a callback object.
     */
    protected boolean drawImage(BufferedImage bImg,
                                int dx1, int dy1, int dx2, int dy2,
                                int sx1, int sy1, int sx2, int sy2,
                                Color bgcolor, ImageObserver observer) {

	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}

        if (sx1 == sx2 || sy1 == sy2) {
            return true;
        }
        int width;
        int height;
        int xOrig;
        int yOrig;
        int xNew;
        int yNew;

        if (sy2 > sy1) {
            height = sy2-sy1;
            yOrig = sy1;
        }
        else {
            height = sy1-sy2;
            yOrig = sy2;
        }

        if (sx2 > sx1) {
            width = sx2 - sx1;
            xOrig = sx1;
        }
        else {
            width = sx1 - sx2;
            xOrig = sx2;
        }
        if (dx2 > dx1) {
            xNew = dx1;
        }
        else {
            xNew = dx2;
        }
        if (dy2 > dy1) {
            yNew = dy1;
        }
        else {
            yNew = dy2;
        }

        if (width <= 0 || height <=0 ) {
            return true;
        }

        if (xOrig >= 0 && yOrig >=0) {
            WritableRaster raster = bImg.getRaster();
            int bw = raster.getWidth();
            int bh = raster.getHeight();

            // Make sure we aren't out of bounds
            if (xOrig + width > bw) {
                width = bw - xOrig;
            }
            if (yOrig + height > bh) {
                height = bh - yOrig;
            }

            // The actual semantics of this method is to draw the image
            // up to, but not including the width and height.  We need
            // to include the width and height in the source image so
            // that the interpolation will have enough data to fill in
            // the last pixel(s).  We subtract the last point in the
            // bounding box rectangle in renderingPipeImage.
            bImg = bImg.getSubimage(xOrig+bImg.getMinX(),
                                    yOrig+bImg.getMinY(), width, height);
        }

        // Append the new transform
        AffineTransform atfm = new AffineTransform();
        atfm.translate( dx1, dy1);
        if ((dx2-dx1 != sx2-sx1) || (dy2-dy1 != sy2-sy1)) {
            double m00 = (double)(dx2-dx1)/(sx2-sx1);
            double m11 = (double)(dy2-dy1)/(sy2-sy1);
            atfm.scale(m00, m11);
        }
        atfm.translate(xOrig-sx1, yOrig-sy1);
        
        return drawImage(bImg, atfm, 0, 0, bgcolor, observer);
    }
    
    /**
     * Draw an image, applying a transform from image space into user space
     * before drawing.
     * The transformation from user space into device space is done with
     * the current transform in the Graphics2D.
     * The given transformation is applied to the image before the
     * transform attribute in the Graphics2D state is applied.
     * The rendering attributes applied include the clip, transform,
     * paint or color and composite attributes. Note that the result is
     * undefined, if the given transform is non-invertible.
     */
    protected boolean drawImage(BufferedImage bImg,
                                AffineTransform xform,
                                int x, int y, Color bgcolor,
                                ImageObserver observer)
    {
        renderingPipeImage(bImg, x, y, null, xform, bgcolor);

        return true;
    }

    public void drawImage(BufferedImage bImg,
                          BufferedImageOp op,
                          int x,
                          int y)  {
        renderingPipeImage(bImg, x, y, op, null, null);
    }

    /**
     * Draws an image, applying a transform from image space into user space
     * before drawing.
     * The transformation from user space into device space is done with
     * the current transform in the Graphics2D.
     * The given transformation is applied to the image before the
     * transform attribute in the Graphics2D state is applied.
     * The rendering attributes applied include the clip, transform,
     * and composite attributes. Note that the result is
     * undefined, if the given transform is noninvertible.
     * @param img The image to be drawn.
     * @param xform The transformation from image space into user space.
     * @see #transform
     * @see #setTransform
     * @see #setComposite
     * @see #clip
     * @see #setClip
     */
    public void drawRenderedImage(RenderedImage img,
                                  AffineTransform xform) {

        // we're forwarding the call onto the drawImage() call which
        // only deals with images at the origin.  Make a translated
        // version of this image and encode the translation in the
        // AffineTransform

        Raster raster;
        ColorModel cm = img.getColorModel();

        if (img instanceof BufferedImage) {
            BufferedImage bufImg = (BufferedImage)img;
            drawImage(bufImg,xform,null);
            return;
        }

        Rectangle rect = getCompBounds();

        try {
	    double p[] = new double[8];
	    p[0] = p[2] = rect.x;
	    p[4] = p[6] = rect.x + rect.width;
	    p[1] = p[5] = rect.y;
	    p[3] = p[7] = rect.y + rect.height;

	    transform.inverseTransform(p, 0, p, 0, 4);
	    xform.inverseTransform(p, 0, p, 0, 4);

            double x0,x1,y0,y1;
            x0 = x1 = p[0];
            y0 = y1 = p[1];

            for (int i = 2; i < 8; ) {
                double pt = p[i++];
                if (pt < x0)  {
                    x0 = pt;
                } else if (pt > x1) {
                    x1 = pt;
                }
                pt = p[i++];
                if (pt < y0)  {
                    y0 = pt;
                } else if (pt > y1) { 
                    y1 = pt;
                }
            }

            // This is padding for anti-aliasing and such.  It may
            // be more than is needed - jbeheler 9/3/98

            int x = (int)(x0-3);
            int w = (int)(x1-x0+6);
            int y = (int)(y0-3);
            int h = (int)(y1-y0+6);

            Rectangle clipRect = new Rectangle(x,y,w,h);
            Rectangle imageRect = new Rectangle(img.getMinX(),img.getMinY(),
                                                img.getWidth(),img.getHeight());
            Rectangle iRect = clipRect.intersection(imageRect);

            if (iRect.width <= 0 || iRect.height <= 0)
                return;

            raster = img.getData(iRect);
        } catch (NoninvertibleTransformException nte) {
            // just get the whole thing
            raster = img.getData();
        }

        WritableRaster wRaster =
              Raster.createWritableRaster(raster.getSampleModel(),
                                          raster.getDataBuffer(),
                                          null);
 
        AffineTransform transXform = (AffineTransform)xform.clone();
        transXform.translate(raster.getMinX(),raster.getMinY());
      
        BufferedImage bufImg = new BufferedImage(cm, wRaster, false, null);
        drawImage(bufImg, transXform, null);
    }


    public void drawRenderableImage(RenderableImage img,
                                    AffineTransform xform) {
        AffineTransform pipeTransform = transform;
        AffineTransform concatTransform = new AffineTransform(xform);
        concatTransform.concatenate(pipeTransform);
        AffineTransform reverseTransform;
 
        RenderContext rc = new RenderContext(concatTransform);

        try {
            reverseTransform = pipeTransform.createInverse();
        } catch (NoninvertibleTransformException nte) {
            rc = new RenderContext(pipeTransform);
            reverseTransform = new AffineTransform();
        }

        RenderedImage rendering = img.createRendering(rc);
        drawRenderedImage(rendering,reverseTransform);
    }



    /*
     * Return the color model to be used with this BufferedImage and
     * transform.
     */
    private ColorModel getTransformColorModel(BufferedImage bImg,
                                              AffineTransform tx) {
        ColorModel cm = bImg.getColorModel();
        ColorModel dstCM = cm;

        if (tx.isIdentity()) {
            return dstCM;
        }
        int type = tx.getType();
        boolean needTrans =
            ((type&(tx.TYPE_MASK_ROTATION|tx.TYPE_GENERAL_TRANSFORM)) != 0);
        if (! needTrans && type != tx.TYPE_TRANSLATION && type != tx.TYPE_IDENTITY)
        {
            double[] mtx = new double[4];
            tx.getMatrix(mtx);
            // Check out the matrix.  A non-integral scale will force ARGB
            // since the edge conditions can't be guaranteed.
            needTrans = (mtx[0] != (int)mtx[0] || mtx[3] != (int)mtx[3]);
        }

        if (renderHint != SunHints.INTVAL_RENDER_QUALITY) {
            if (cm instanceof IndexColorModel) {
                Raster raster = bImg.getRaster();
                IndexColorModel icm = (IndexColorModel) cm;
                // Just need to make sure that we have a transparent pixel
                if (needTrans && cm.getTransparency() == cm.OPAQUE) {
                    double[] matrix = new double[6];
                    tx.getMatrix(matrix);
                    if (matrix[1] == 0. && matrix[2] == 0. && matrix[4] == 0.
                        && matrix[5] == 0.) {
                        // Only scaling so do not need to create
                    }
                    else {
                        int mapSize = icm.getMapSize();
                        if (mapSize < 256) {
                            int[] cmap = new int[mapSize+1];
                            icm.getRGBs(cmap);
                            cmap[mapSize] = 0x0000;
                            dstCM = new
                                IndexColorModel(icm.getPixelSize(), mapSize+1,
                                                cmap, 0, true, mapSize,
                                                DataBuffer.TYPE_BYTE);
                        }
                        else {
                            dstCM = ColorModel.getRGBdefault();
                        }
                    }  /* if (matrix[0] < 1.f ...) */
                } /* if (cm.getTransparency() == cm.OPAQUE) */
            } /* if (cm instanceof IndexColorModel) */
            else if (needTrans && cm.getTransparency() == cm.OPAQUE) {
                // Need a bitmask transparency
                // REMIND: for now, use full transparency since no loops
                // for bitmask
                dstCM = ColorModel.getRGBdefault();
            }
        } /* if (renderHint == RENDER_QUALITY) */
        else {

            if (cm instanceof IndexColorModel ||
                (needTrans && cm.getTransparency() == cm.OPAQUE))
            {
                // Need a bitmask transparency
                // REMIND: for now, use full transparency since no loops
                // for bitmask
                dstCM = ColorModel.getRGBdefault();
            }
        }

        return dstCM;
    }

    
    /*
     * Transform the bounding box of the BufferedImage
     */
    protected Rectangle transformBounds(Rectangle rect,
                                        AffineTransform tx) {
        if (tx.isIdentity()) {
            return rect;
        }

        Shape s = transformShape(tx, rect);
        return s.getBounds();
    }

    private void renderingPipeImage(BufferedImage bImg, int x, int y,
                                    BufferedImageOp op, 
                                    AffineTransform xf, Color bgcolor) {
        Rectangle clipRect = getCompBounds();
        int interpolationType = interpolationHint;
        
        // fast case most common call
        if (
                (op == null)
                &&  ((xf == null) || (xf.isIdentity()))
                &&  (clipState != CLIP_SHAPE)
                &&  (transformState != TRANSFORM_GENERIC)
        ) {
            int cx = clipRect.x;
            int cy = clipRect.y;
            int cw = clipRect.width;
            int ch = clipRect.height;
            int x2 = x;
            int y2 = y;
            if (transformState == TRANSFORM_TRANSLATEONLY) {
                x2 += transform.getTranslateX();
                y2 += transform.getTranslateY();
            }
            int maxX = x2 + bImg.getWidth();
            int maxY = y2 + bImg.getHeight();
            if (x2 > cx) {
                cw = (cw + cx) - x2;
                cx = x2;
            }
            if (y2 > cy) {
                ch = (ch + cy) - y2;
                cy = y2;
            }
            if (maxX < (cx + cw)) {
                cw = maxX - cx;
            }
            if (maxY < (cy + ch)) {
                ch = maxY - cy;
            }
            if ((cw <= 0) || (ch <=0)) {
                return;
            }
	    if ((cx > x2) || (cy > y2)) {
                bImg = bImg.getSubimage(bImg.getMinX() + (cx - x2),
                                        bImg.getMinY() + (cy - y2),
                                        cw, ch);
	    }
            rpics = RasterOutputManager.getManager().renderImage(
                                        this, bImg, bgcolor,
				        cx, cy, cw, ch,
				        null, cw, rpics);
            return;
        }

        AffineTransform opXform = null;
        if (op != null) {
            if (op instanceof AffineTransformOp) {
                AffineTransformOp atop = (AffineTransformOp) op;
                opXform = atop.getTransform();
                interpolationType = atop.getInterpolationType();
            }
            else {
                bImg = op.filter(bImg, null);
            }
        }

        int bImgWidth = bImg.getWidth();
        int bImgHeight = bImg.getHeight();

        // Begin Transform
        AffineTransform tx = new AffineTransform(transform);
	tx.translate(x, y);
        if (xf != null && !xf.isIdentity()) {
            tx.concatenate(xf);
        }
        if (opXform != null && !opXform.isIdentity()) {
            tx.concatenate(opXform);
        }

        double mat[] = new double[6];
        tx.getMatrix(mat);

        // Only a translation
        if (mat[0] == 1. && mat[1] == 0. && mat[2] == 0. && mat[3] == 1.) {
            x = (int) Math.round(mat[4]);
            y = (int) Math.round(mat[5]);
            Rectangle bounds = new Rectangle(x, y, bImgWidth, bImgHeight);
            clipRect = clipRect.intersection(bounds);
            if (clipRect.isEmpty()) {
		return;
            }
            if (!clipRect.equals(bounds)) {
		bImg = bImg.getSubimage(clipRect.x - x, clipRect.y - y,
					clipRect.width, clipRect.height);
		bImgWidth = clipRect.width;
		bImgHeight = clipRect.height;
            }
        } else {
            ColorModel dstCM = getTransformColorModel(bImg, tx);
            int hint = interpolationType;
            if (interpolationType == -1) {
                hint = (renderHint == SunHints.INTVAL_RENDER_QUALITY
			? AffineTransformOp.TYPE_BILINEAR
			: AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
            }
            AffineTransformOp affOp = new AffineTransformOp(tx, hint);
            // Transform the bounding box of the image in order to
            // extract the translation and to figure out the destination
            // width and height
            Rectangle bounds = affOp.getBounds2D(bImg).getBounds();
            clipRect = clipRect.intersection(bounds);
            if (clipRect.isEmpty()) {
		return;
            }

            x = clipRect.x;
            y = clipRect.y;
            bImgWidth  = clipRect.width;
            bImgHeight = clipRect.height;
            
            if (x != 0 || y != 0) {
                tx.preConcatenate(new AffineTransform(1.,0.,0.,1., -x, -y));
                affOp = new AffineTransformOp(tx, hint);
            }
            
            // Now transform the image by the new transform.
            BufferedImage dst =
                new BufferedImage(dstCM,
                             dstCM.createCompatibleWritableRaster(bImgWidth,
                                                                  bImgHeight),
                             false, null);

            bImg = affOp.filter(bImg, dst);
            if (bImg == null) {
                return;
            }
        }

        // End Transform

        // Begin Rasterizer for Clip Shape
        if (clipState == CLIP_SHAPE) {
	    int box[] = new int[4];

	    clipRegion.getBounds(box);
	    Rectangle devR = new Rectangle(box[0], box[1],
					   box[2] - box[0],
					   box[3] - box[1]);
	    devR = devR.intersection(clipRect);
	    if (!devR.isEmpty()) {
		OutputManager mgr = getOutputManager();
		RegionIterator ri = clipRegion.getIterator();
		BufferedImage subImg;
		int cx1 = clipRect.x;
		int cy1 = clipRect.y;
		int cx2 = cx1 + clipRect.width;
		int cy2 = cy1 + clipRect.height;
		while (ri.nextYRange(box)) {
		    int spany1, spany2;
		    if ((spany1 = box[0]) > cy2) {
			break;
		    }
		    if ((spany2 = box[1]) < cy1) {
			continue;
		    }
		    if (spany1 < cy1) {
			spany1 = cy1;
		    }
		    if (spany2 > cy2) {
			spany2 = cy2;
		    }
		    int spanh = spany2 - spany1;
		    while (ri.nextXBand(box)) {
			int spanx1, spanx2;
			if ((spanx1 = box[2]) > cx2) {
			    break;
			}
			if ((spanx2 = box[3]) < cx1) {
			    continue;
			}
			if (spanx1 < cx1) {
			    spanx1 = cx1;
			}
			if (spanx2 > cx2) {
			    spanx2 = cx2;
			}
			int spanw = spanx2 - spanx1;
			subImg = bImg.getSubimage(spanx1 - cx1,
						  spany1 - cy1,
						  spanw, spanh);
			mgr.renderImage(this, subImg, bgcolor,
					spanx1, spany1, spanw, spanh,
					null, spanw, null);
		    }
		}
	    }
	    return;
	}
        // End Rasterizer for Clip Shape
	getOutputManager().renderImage(this, bImg, bgcolor,
				       clipRect.x, clipRect.y,
                                       clipRect.width, clipRect.height,
				       null, clipRect.width, null);
    }

// text rendering methods
    public void drawString(String str, int x, int y)
    {
        try {
            dscc = textpipe.drawString(this, str, (float)x, (float)y, dscc);
        } catch (ShapingException e) {
            TextLayout tl = new TextLayout(str, getFont(), getFontRenderContext());
            tl.draw(this, (float)x, (float)y);
        }
    }

    public void drawString(String str, float x, float y)
    {
        try {
            dscc = textpipe.drawString(this, str, x, y, dscc);
        } catch (ShapingException e) {
            TextLayout tl = new TextLayout(str, getFont(), getFontRenderContext());
            tl.draw(this, x, y);
        }
    }

    public void drawString(AttributedCharacterIterator iterator, int x, int y) {
        TextLayout tl = new TextLayout(iterator, getFontRenderContext());
        tl.draw(this, (float) x, (float) y);
    }

    public void drawString(AttributedCharacterIterator iterator, float x, float y) {
        TextLayout tl = new TextLayout(iterator, getFontRenderContext());
        tl.draw(this, x, y);
    }

    public void drawGlyphVector(GlyphVector gv, float x, float y)
    {
        dvcc = textpipe.drawGlyphVector(this, gv, x, y, dvcc);
    }

    public void drawChars(
            char data[], int offset, int length, int x, int y)
    {
        try {
            dccc = textpipe.drawChars(this, data, offset, length, x, y, dccc);
        } catch (ShapingException e) {
	    TextLayout tl = new TextLayout(new String(data, offset, length),
					   getFont(), getFontRenderContext());
            tl.draw(this, (float)x, (float)y);
        }
    }

    public void drawBytes(byte data[], int offset, int length, int x, int y)
    {
        try {
            dbcc = textpipe.drawBytes(this, data, offset, length, x, y, dbcc);
        } catch (ShapingException e) {
            TextLayout tl = new TextLayout(
                                    new String(data, 0, offset, length),
                                    getFont(), getFontRenderContext());
            tl.draw(this, (float)x, (float)y);
        }
    }
// end of text rendering methods

// implementation of textpipe -- used by child classes only to delegate
// implementation.
    public Object drawString(SunGraphics2D g2d, String s,
                            float x, float y, Object cs) {
	return RasterOutputManager.getManager().
                                    drawString(g2d, s, x, y, cs);
    }

    public Object drawGlyphVector(SunGraphics2D g2d, GlyphVector g,
                            float x, float y, Object cs) {
	return RasterOutputManager.getManager().
                                    drawGlyphVector(g2d, g, x, y, cs);
    }

    public Object drawChars(SunGraphics2D g2d,
            char data[], int offset, int length, int x, int y,
            Object cs) {
	return RasterOutputManager.getManager().drawChars(g2d,
                                        data, offset, length, x, y, cs);
    }

    public Object drawBytes(SunGraphics2D g2d,
            byte data[], int offset, int length, int x, int y,
            Object cs) {
	return RasterOutputManager.getManager().drawBytes(g2d,
                                        data, offset, length, x, y, cs);
    }
// end of implementation of textpipe

    /**
    * Get the rendering context of the font
    * within this Graphics2D context.
    */
    public FontRenderContext getFontRenderContext() {
        int aahint =
	    (textAntialiasHint == SunHints.INTVAL_TEXT_ANTIALIAS_DEFAULT
	     ? antialiasHint : textAntialiasHint);
        // REMIND: should be normalizing transform and should fix
        // metrics for hotjava
        return new FontRenderContext
	    (getDeviceConfiguration().getDefaultTransform(),
	     (aahint == SunHints.INTVAL_ANTIALIAS_ON),
	     (fractionalMetricsHint == SunHints.INTVAL_FRACTIONALMETRICS_ON));
    }

    /**
    *   device ID
    *   each child class should return a unique string (prefer classname)
    *   if the child class has device specific functions to be registered
    *   with GraphicPrimitiveMgr
    */
    public String getDevID() {
        return null;
    }

    /*
    * frame buffer architecture locking protocol interfaces
    * REMIND need DOCS
    */

    /**
     * The method prepares locks for a SunGraphics2D to make it accessible
     * for direct operation on graphics content by GraphicsPrimitives
     */
    abstract public ImageData lock(
                int xView, int yView, int dxView, int dyView, int lockFlags);
    abstract public int unlock(ImageData idData);
}
