/*
 * @(#)BinaryCompositeBg.java	1.8 98/07/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitive;
import java.awt.Color;

/**
*   BinaryCompositeBg
*   1)  copies rectangle of pixels from one surface to another
*   2)  blends src and dest based on alpha rule,
*       extraAlpha, srcAlpha, and dstAlpha
*   3) must accept [sx, sy, dx, dy]
*      from within the surface description data
*   4) applies an XOR color
*   5) fills transparent result pixels with a background color
*
*   precise behavior is undefined if the source surface
*   and the destination surface are the same surface
*   with overlapping regions of pixels
*/

public abstract class BinaryCompositeBg extends GraphicsPrimitive
{
    private final static String methodSignature =
        "BinaryCompositeBg(Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;III[BIIFLjava/awt/Color;Ljava/awt/Color;)V";
                
    public static String getMethodSignature() {
        return methodSignature;
    }

    private final static int lockFlags[] = { LOCK_FOR_READ, LOCK_FOR_WRITE };

    protected BinaryCompositeBg(int[] surfaceTypes) {
	super(surfaceTypes, lockFlags, methodSignature);
    }

    /**
    *   All BinaryCompositeBg implementors must have this invoker method
    */
    public abstract void BinaryCompositeBg(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,

        Color bgColor, Color XORColor
    );
}
