/*
 * @(#)ByteDiscreteRenderer.java	1.13 98/07/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.pipe.SpanIterator;
import java.awt.image.ColorModel;
import sun.awt.image.ByteComponentRaster;
import sun.java2d.SunGraphics2D;
import java.awt.Color;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.color.ColorSpace;

/**
 * ByteDiscreteRenderer collection
 * Basically, a collection of components which permit basic
 * rendering to occur on Byte Rasters with a single
 * component and a pixstride of 1
 */

public final class ByteDiscreteRenderer implements GraphicsPrimitives {

    //
    // int arrays used for the surface types:
    //
    static int[] ST_BYTE_GRAY = { GraphicsPrimitive.TYPE_BYTE_GRAY };
    static int[] ST_BYTE_INDEXED = { GraphicsPrimitive.TYPE_BYTE_INDEXED };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "FillRectGray",
					ST_BYTE_GRAY,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectByteIndexed",
					ST_BYTE_INDEXED,
					FillRect.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "FillSpansGray",
					ST_BYTE_GRAY,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansByteIndexed",
					ST_BYTE_INDEXED,
					FillSpans.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawLineGray",
					ST_BYTE_GRAY,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineByteIndexed",
					ST_BYTE_INDEXED,
					DrawLine.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "FillPolygonsGray",
					ST_BYTE_GRAY,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsByteIndexed",
					ST_BYTE_INDEXED,
					FillPolygons.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawPolygonsGray",
					ST_BYTE_GRAY,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsByteIndexed",
					ST_BYTE_INDEXED,
					DrawPolygons.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawRectGray",
					ST_BYTE_GRAY,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectByteIndexed",
					ST_BYTE_INDEXED,
					DrawRect.getMethodSignature())
        };
        return primitives;
    }

    /**
    * convert RGB value to Grayscale
    */
    static int toGray(Color clr) {
        int color = clr.getRGB();
        int red, green, blue;
        red   = ((color >> 16) & 0xff);
        green = ((color >> 8) & 0xff);
        blue  = ((color) & 0xff);
        return (int)(red*0.299 + green*.587 + blue*.114);
        // REMIND: Might need a conversion if COLOR_QUALITY is high
        // and RGB space is non-linear?
//         ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_GRAY);
//         float rgb[] = clr.getRGBComponents(null);
//         float grayValue[] = cs.fromRGB(rgb);
//         return (int)(grayValue[0] * 255.0);
    }

    /**
    * convert RGB value to Indexed Color
    * using the color model of the destination
    */
    static int toIndexed(Color color, ColorModel cm) {
        byte [] pixels = {0};
        cm.getDataElements(color.getRGB(), pixels);;
        return pixels[0];
    }

    static native void devSetRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr);

    static native void devSetSpans(
        int color, SpanIterator si, long pIterator,
        int offset, byte [] dst, int pixStr, int scanStr);

    static native void devDrawLine(
        int color,
        int x1, int y1, int x2, int y2,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr);

    static native void devFillPolygons(
        int color,
        int xPoints[], int yPoints[],int nPoints[], int numpolys,
        boolean evenodd,
        int originx, int originy,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr);

    static native void devDrawPolygons(
        int color,
        int[] xPoints, int[] yPoints, int nPoints[],
        int numpolys, boolean close,
        int originX, int originY,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr);

    static native void devDrawRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int pixStr, int scanStr);
}

class BCRFillRectRasterContext extends FillRectRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRFillRectRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	ByteDiscreteRenderer.devSetRect(
	    this.color, x, y, w, h,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class FillRectGray extends FillRect {
    FillRectGray() {
	super(ByteDiscreteRenderer.ST_BYTE_GRAY);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devSetRect(
	    ByteDiscreteRenderer.toGray(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRFillRectRasterContext(
			(ByteComponentRaster)raster, 
			ByteDiscreteRenderer.toGray(color),
			clipX, clipY, clipW, clipH);
    }
}

class FillRectByteIndexed extends FillRect {
    FillRectByteIndexed() {
	super(ByteDiscreteRenderer.ST_BYTE_INDEXED);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devSetRect(
	    ByteDiscreteRenderer.toIndexed(color, dstImage.getColorModel()),
	    x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRFillRectRasterContext(
			(ByteComponentRaster)raster,
			ByteDiscreteRenderer.toIndexed(color, cm),
			clipX, clipY, clipW, clipH);
    }
}

class FillSpansGray extends FillSpans {
    FillSpansGray() {
	super(ByteDiscreteRenderer.ST_BYTE_GRAY);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devSetSpans(
	    ByteDiscreteRenderer.toGray(color),
	    si, si.getNativeIterator(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
}

class FillSpansByteIndexed extends FillSpans {
    FillSpansByteIndexed() {
	super(ByteDiscreteRenderer.ST_BYTE_INDEXED);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devSetSpans(
	    ByteDiscreteRenderer.toIndexed(color, dstImage.getColorModel()),
	    si, si.getNativeIterator(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
}

class BCRDrawLineRasterContext extends DrawLineRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawLineRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x1, int y1, int x2, int y2) {
	ByteDiscreteRenderer.devDrawLine(
	    this.color, x1, y1, x2, y2,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class DrawLineGray extends DrawLine {
    DrawLineGray() {
	super(ByteDiscreteRenderer.ST_BYTE_GRAY);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devDrawLine(
	    ByteDiscreteRenderer.toGray(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRDrawLineRasterContext(
			(ByteComponentRaster)raster, ByteDiscreteRenderer.toGray(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineByteIndexed extends DrawLine {
    DrawLineByteIndexed() {
	super(ByteDiscreteRenderer.ST_BYTE_INDEXED);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devDrawLine(
	    ByteDiscreteRenderer.toIndexed(color, dstImage.getColorModel()),
	    x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRDrawLineRasterContext(
			(ByteComponentRaster)raster,
			ByteDiscreteRenderer.toIndexed(color, cm),
			clipX, clipY, clipW, clipH);
    }
}
class BCRFillPolygonsRasterContext extends FillPolygonsRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRFillPolygonsRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(
	    int xPoints[], int yPoints[], int nPoints[],
	    int numPolys, boolean evenOdd) {
	ByteDiscreteRenderer.devFillPolygons(
	    this.color, xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}

class FillPolygonsGray extends FillPolygons {
    FillPolygonsGray() {
	super(ByteDiscreteRenderer.ST_BYTE_GRAY);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devFillPolygons(
	    ByteDiscreteRenderer.toGray(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRFillPolygonsRasterContext(
			(ByteComponentRaster)raster,
			ByteDiscreteRenderer.toGray(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsByteIndexed extends FillPolygons {
    FillPolygonsByteIndexed() {
	super(ByteDiscreteRenderer.ST_BYTE_INDEXED);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devFillPolygons(
	    ByteDiscreteRenderer.toIndexed(color, dstImage.getColorModel()),
	    xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRFillPolygonsRasterContext(
			(ByteComponentRaster)raster,
			ByteDiscreteRenderer.toIndexed(color, cm),
			clipX, clipY, clipW, clipH);
    }
}

class BCRDrawPolygonsRasterContext extends DrawPolygonsRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawPolygonsRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(
	    int xPoints[], int yPoints[], int nPoints[],
	    int numPolys, boolean close) {
	ByteDiscreteRenderer.devDrawPolygons(
	    this.color, xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class DrawPolygonsGray extends DrawPolygons {
    DrawPolygonsGray() {
	super(ByteDiscreteRenderer.ST_BYTE_GRAY);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devDrawPolygons(
	    ByteDiscreteRenderer.toGray(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRDrawPolygonsRasterContext(
			(ByteComponentRaster)raster,
			ByteDiscreteRenderer.toGray(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsByteIndexed extends DrawPolygons {
    DrawPolygonsByteIndexed() {
	super(ByteDiscreteRenderer.ST_BYTE_INDEXED);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devDrawPolygons(
	    ByteDiscreteRenderer.toIndexed(color, dstImage.getColorModel()),
	    xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRDrawPolygonsRasterContext(
			(ByteComponentRaster)raster,
			ByteDiscreteRenderer.toIndexed(color, cm),
			clipX, clipY, clipW, clipH);
    }
}
class BCRDrawRectRasterContext extends DrawRectRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawRectRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	ByteDiscreteRenderer.devDrawRect(
	    this.color, x, y, w, h,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class DrawRectGray extends DrawRect {
    DrawRectGray() {
	super(ByteDiscreteRenderer.ST_BYTE_GRAY);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devDrawRect(
	    ByteDiscreteRenderer.toGray(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRDrawRectRasterContext(
			(ByteComponentRaster)raster, ByteDiscreteRenderer.toGray(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectByteIndexed extends DrawRect {
    DrawRectByteIndexed() {
	super(ByteDiscreteRenderer.ST_BYTE_INDEXED);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ByteComponentRaster bcr =
		(ByteComponentRaster)dstImage.getRaster();
	ByteDiscreteRenderer.devDrawRect(
	    ByteDiscreteRenderer.toIndexed(color, dstImage.getColorModel()),
	    x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bcr.getDataOffset(0),
	    bcr.getDataStorage(),
	    bcr.getPixelStride(),
	    bcr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BCRDrawRectRasterContext(
			(ByteComponentRaster)raster,
			ByteDiscreteRenderer.toIndexed(color, cm),
			clipX, clipY, clipW, clipH);
    }
}




