/*
 * @(#)CustomComponent.java	1.13 98/08/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 *
 */

/*
 * @author Charlton Innovations, Inc.
 * @author Jim Graham
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.loops.OpaqueBlit;
import sun.java2d.loops.ImageData;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;

/**
*   CustomComponent, collection of GraphicsPrimitive
*   Basically, this collection of components performs conversion from
*   ANY to ANY via opaque copy
*/

public final class CustomComponent implements GraphicsPrimitives {

    //
    // int array used for the surface types:
    //
    static int[] ST_CUSTOM__CUSTOM = { GraphicsPrimitive.TYPE_CUSTOM, GraphicsPrimitive.TYPE_CUSTOM };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
	    new GraphicsPrimitiveProxy(this, "OpaqueCopyAny",
					ST_CUSTOM__CUSTOM,
					OpaqueBlit.getMethodSignature())
	};
        return primitives;
    }
}

/**
*   ANY format to ANY format OpaqueBlit
*/
class OpaqueCopyAny extends OpaqueBlit {
    OpaqueCopyAny() {
	super(CustomComponent.ST_CUSTOM__CUSTOM);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {

  
        int srcXOffs = src.getXViewArea();
        int srcYOffs = src.getYViewArea();
        int dstXOffs = dst.getXViewArea();
        int dstYOffs = dst.getYViewArea();

	if ((null == src) || (null == dst)) {
	    return;
	}

        
	ColorModel srcCM = src.getColorModel();
	ColorModel dstCM = dst.getColorModel();
	Raster srcRast = src.getRaster();
	WritableRaster dstRast = (WritableRaster) dst.getRaster();
	if ((null == srcRast) || (null == dstRast)) {
	    return;
	}
        
        w = Math.min(srcRast.getWidth()-srcXOffs, w);
        w = Math.min(dstRast.getWidth()-dstXOffs, w);
        h = Math.min(srcRast.getHeight()-srcYOffs, h);
        h = Math.min(dstRast.getHeight()-dstYOffs, h);
        
	Object srcPix = null;
	Object dstPix = null;

        if (dst.getTransparency() == java.awt.Transparency.OPAQUE) {
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++) {
                    srcPix = srcRast.getDataElements(x+srcXOffs, y+srcYOffs,
                                                     srcPix);
                    int rgb = srcCM.getRGB(srcPix);
                    // Only copy the pixel if there is an alpha
                    if ((rgb&0xff000000) != 0) {
                        dstPix = dstCM.getDataElements(rgb, dstPix);
                        dstRast.setDataElements(x+dstXOffs, y+dstYOffs,
                                                dstPix);
                    }
                }
            }
        }
        else {
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++) {
                    srcPix = srcRast.getDataElements(x+srcXOffs, y+srcYOffs,
                                                     srcPix);
                    int rgb = srcCM.getRGB(srcPix);
                    dstPix = dstCM.getDataElements(rgb, dstPix);
                    dstRast.setDataElements(x+dstXOffs, y+dstYOffs,
                                            dstPix);
                }
            }
        }
    }
}
