/*
 * @(#)DefaultComponent.java	1.23 98/09/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.loops.OpaqueBlit;
import sun.java2d.loops.OpaqueDitheredBlit;
import sun.java2d.loops.ColorPaint;
import sun.java2d.loops.BinaryComposite;
import sun.java2d.loops.BinaryCompositeBg;
import java.awt.Color;

/**
*   DefaultComponent, collection of GraphicsPrimitive
*   Basically, this collection of components performs conversion from
*   ANY to INT_ARGB, INT_RGB to ANY
*/

/* Added Byte_Gray cases here, but they should actually be in some other files.   Reminder: Reorg this files to have only ANY->INT_ARGB, INT_RGB to ANY
 */
public final class DefaultComponent implements GraphicsPrimitives {

    //
    // int arrays used for the surface types:
    //
    static int[] ST_INT_BINARY__INT_BINARY = { GraphicsPrimitive.TYPE_INT_BINARY, 
					       GraphicsPrimitive.TYPE_INT_BINARY};
    static int[] ST_INT_RGB__ST_INT_RGB = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_INT_ARGB__INT_RGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_INT_ARGB__INT_ARGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INT_ARGB_PRE__INT_ARGB_PRE = { GraphicsPrimitive.TYPE_INT_ARGB_PRE, 
					   GraphicsPrimitive.TYPE_INT_ARGB_PRE};
    static int[] ST_INT_BGR__INT_BGR = { GraphicsPrimitive.TYPE_INT_BGR, 
					   GraphicsPrimitive.TYPE_INT_BGR};
    static int[] ST_INT_RGBx__INT_RGBx = { GraphicsPrimitive.TYPE_INT_RGBx, 
					   GraphicsPrimitive.TYPE_INT_RGBx};
    static int[] ST_INT_BGRx__INT_BGRx = { GraphicsPrimitive.TYPE_INT_BGRx, 
					   GraphicsPrimitive.TYPE_INT_BGRx};
    static int[] ST_INT_ARGB_BM__INT_ARGB_BM = { GraphicsPrimitive.TYPE_INT_ARGB_BM, 
					   GraphicsPrimitive.TYPE_INT_ARGB_BM};
    static int[] ST_INT_ARGB_PRE_BM__INT_ARGB_PRE_BM = { GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM, 
					   GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM};
     static int[] ST_INT_RGB__BYTE_GRAY = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_BYTE_GRAY};
    static int[] ST_INT_ARGB__BYTE_GRAY = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_BYTE_GRAY};
   static int[] ST_INT_RGB__24BIT_RGB = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_24BIT_RGB};
    static int[] ST_INT_ARGB__24BIT_RGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_24BIT_RGB};
    static int[] ST_INT_RGB__3BYTE_BGR = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_3BYTE_BGR};
    static int[] ST_INT_ARGB__3BYTE_BGR = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_3BYTE_BGR};
    static int[] ST_INT_RGB__SHORT_555_RGB = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_SHORT_555_RGB};
    static int[] ST_INT_ARGB__SHORT_555_RGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_SHORT_555_RGB};
    static int[] ST_INT_RGB__SHORT_565_RGB = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_SHORT_565_RGB};
    static int[] ST_INT_ARGB__SHORT_565_RGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_SHORT_565_RGB};
    static int[] ST_INT_RGB__INT_RGBx = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_INT_RGBx};
    static int[] ST_INT_RGB__INT_BGR = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_INT_BGR};
    static int[] ST_INT_RGB__INT_BGRx = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_INT_BGRx};
    static int[] ST_INT_ARGB__INT_BGR = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_INT_BGR};
    static int[] ST_INT_ARGB__INT_BGRx = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_INT_BGRx};
    static int[] ST_INT_RGB__BINARY = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_BINARY};
    static int[] ST_INT_ARGB__BINARY = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_BINARY};
    static int[] ST_INT_RGB__DIB_xRGB = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_DIB_xRGB};
    static int[] ST_INT_ARGB__DIB_xRGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_DIB_xRGB};
    static int[] ST_INT_RGB__INT_ARGB = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INT_RGB__INT_ARGB_PRE = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_INT_ARGB_PRE};
    static int[] ST_INT_RGB__4BYTE_ABGR = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_4BYTE_ABGR};
    static int[] ST_INT_RGB__4BYTE_ABGR_PRE = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_4BYTE_ABGR_PRE};
    static int[] ST_INT_ARGB__4BYTE_ABGR = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_4BYTE_ABGR};
    static int[] ST_INT_ARGB__4BYTE_ABGR_PRE = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_4BYTE_ABGR_PRE};
    static int[] ST_DIB_xRGB__INT_ARGB = { GraphicsPrimitive.TYPE_DIB_xRGB, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_DIB_xRGB__INT_RGB = { GraphicsPrimitive.TYPE_DIB_xRGB, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_INT_ARGB_PRE__INT_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_4BYTE_ABGR__INT_ARGB = { GraphicsPrimitive.TYPE_4BYTE_ABGR, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_4BYTE_ABGR__INT_RGB = { GraphicsPrimitive.TYPE_4BYTE_ABGR, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_4BYTE_ABGR_PRE__INT_ARGB = { GraphicsPrimitive.TYPE_4BYTE_ABGR_PRE, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_4BYTE_ABGR_PRE__INT_RGB = { GraphicsPrimitive.TYPE_4BYTE_ABGR_PRE, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_BINARY__INT_ARGB = { GraphicsPrimitive.TYPE_BINARY, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_BINARY__INT_RGB = { GraphicsPrimitive.TYPE_BINARY, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_BINARY__INT_BGR = { GraphicsPrimitive.TYPE_BINARY, 
					   GraphicsPrimitive.TYPE_INT_BGR};
    static int[] ST_BYTE_INDEXED__INT_ARGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_BYTE_INDEXED__INT_RGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_SHORT_BINARY__SHORT_BINARY =
                                       { GraphicsPrimitive.TYPE_SHORT_BINARY,
                                         GraphicsPrimitive.TYPE_SHORT_BINARY};
    static int[] ST_SHORT_555_RGB__SHORT_555_RGB =
                                       { GraphicsPrimitive.TYPE_SHORT_555_RGB,
                                         GraphicsPrimitive.TYPE_SHORT_555_RGB};
    static int[] ST_SHORT_565_RGB__SHORT_565_RGB =
                                       { GraphicsPrimitive.TYPE_SHORT_565_RGB,
                                         GraphicsPrimitive.TYPE_SHORT_565_RGB};
    static int[] ST_SHORT_555_RGB__INT_ARGB = { GraphicsPrimitive.TYPE_SHORT_555_RGB, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_SHORT_555_RGB__INT_RGB = { GraphicsPrimitive.TYPE_SHORT_555_RGB, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_SHORT_565_RGB__INT_ARGB = { GraphicsPrimitive.TYPE_SHORT_565_RGB, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_SHORT_565_RGB__INT_RGB = { GraphicsPrimitive.TYPE_SHORT_565_RGB, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_24BIT_RGB__INT_ARGB = { GraphicsPrimitive.TYPE_24BIT_RGB, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_24BIT_RGB__INT_RGB = { GraphicsPrimitive.TYPE_24BIT_RGB, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_INT_RGBx__INT_ARGB = { GraphicsPrimitive.TYPE_INT_RGBx, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INT_RGBx__INT_RGB = { GraphicsPrimitive.TYPE_INT_RGBx, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_3BYTE_BGR__INT_ARGB = { GraphicsPrimitive.TYPE_3BYTE_BGR, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_3BYTE_BGR__INT_RGB = { GraphicsPrimitive.TYPE_3BYTE_BGR, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_INT_BGR__INT_ARGB = { GraphicsPrimitive.TYPE_INT_BGR, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INT_BGR__INT_RGB = { GraphicsPrimitive.TYPE_INT_BGR, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_INT_BGRx__INT_ARGB = { GraphicsPrimitive.TYPE_INT_BGRx, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INT_BGRx__INT_RGB = { GraphicsPrimitive.TYPE_INT_BGRx, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_INT_RGB__BYTE_INDEXED = { GraphicsPrimitive.TYPE_INT_RGB, 
					   GraphicsPrimitive.TYPE_BYTE_INDEXED};
    static int[] ST_INT_ARGB__BYTE_INDEXED = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_BYTE_INDEXED};
    static int[] ST_INT_RGB = { GraphicsPrimitive.TYPE_INT_RGB };
    static int[] ST_INT_ARGB = { GraphicsPrimitive.TYPE_INT_ARGB };
    static int[] ST_INT_ARGB_BM__INT_ARGB = { GraphicsPrimitive.TYPE_INT_ARGB_BM, 
					   GraphicsPrimitive.TYPE_INT_ARGB };
    static int[] ST_INT_ARGB_PRE_BM__INT_ARGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM, 
					   GraphicsPrimitive.TYPE_INT_ARGB };
    static int[] ST_INT_ARGB_BM__INT_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_BM, 
					   GraphicsPrimitive.TYPE_INT_RGB };
    static int[] ST_INT_ARGB_PRE_BM__INT_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM, 
					   GraphicsPrimitive.TYPE_INT_RGB };
    static int[] ST_BYTE_INDEXED_BM__INT_ARGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_BYTE_GRAY__INT_ARGB = { GraphicsPrimitive.TYPE_BYTE_GRAY, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_BYTE_GRAY__INT_BGR = { GraphicsPrimitive.TYPE_BYTE_GRAY, 
					   GraphicsPrimitive.TYPE_INT_BGR};
    static int[] ST_BYTE_GRAY__INT_RGB = { GraphicsPrimitive.TYPE_BYTE_GRAY, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_BYTE_GRAY__BYTE_GRAY = { GraphicsPrimitive.TYPE_BYTE_GRAY, 
					   GraphicsPrimitive.TYPE_BYTE_GRAY};
    static int[] ST_BYTE_GRAY__INDEX8_GRAY = { GraphicsPrimitive.TYPE_BYTE_GRAY, 
					   GraphicsPrimitive.TYPE_INDEX8_GRAY};
    static int[] ST_BYTE_GRAY__BYTE_INDEXED = { GraphicsPrimitive.TYPE_BYTE_GRAY, 
					   GraphicsPrimitive.TYPE_BYTE_INDEXED};
    static int[] ST_USHORT_GRAY__INT_ARGB = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
					   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INT_ARGB__USHORT_GRAY = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_USHORT_GRAY};
    static int[] ST_USHORT_GRAY__INT_BGR = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
					   GraphicsPrimitive.TYPE_INT_BGR};
    static int[] ST_USHORT_GRAY__INT_RGB = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
					   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_USHORT_GRAY__INDEX8_GRAY = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
					   GraphicsPrimitive.TYPE_INDEX8_GRAY};
    static int[] ST_USHORT_GRAY__BYTE_GRAY = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
					   GraphicsPrimitive.TYPE_BYTE_GRAY};
    static int[] ST_INT_ARGB__INT_RGBx = { GraphicsPrimitive.TYPE_INT_ARGB,
					   GraphicsPrimitive.TYPE_INT_RGBx};
    static int[] ST_INT_ARGB__INT_ARGB_PRE = { GraphicsPrimitive.TYPE_INT_ARGB,
					   GraphicsPrimitive.TYPE_INT_ARGB_PRE};
    static int[] ST_INT_ARGB_PRE__INT_ARGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE,
					   GraphicsPrimitive.TYPE_INT_ARGB};

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new GraphicsPrimitiveProxy(this, "IntIsoMorphicCopy", 
					ST_INT_BINARY__INT_BINARY,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToIntRgb", 
					ST_INT_RGB__ST_INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToIntRgb", 
					ST_INT_ARGB__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToIntArgb", 
					ST_INT_ARGB__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbbgfillToIntArgb", 
					ST_INT_ARGB__INT_ARGB,
					OpaqueBlitBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbPreToIntArgbPre",
					ST_INT_ARGB_PRE__INT_ARGB_PRE,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntBgrToIntBgr",
					ST_INT_BGR__INT_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbxToIntRgbx",
					ST_INT_RGBx__INT_RGBx,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntBgrxToIntBgrx",
					ST_INT_BGRx__INT_BGRx,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbBmToIntArgbBm",
					ST_INT_ARGB_BM__INT_ARGB_BM,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbPreBmToIntArgbPreBm",
					ST_INT_ARGB_PRE_BM__INT_ARGB_PRE_BM,
					OpaqueBlit.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "IntRgbToByteGray",
                                       ST_INT_RGB__BYTE_GRAY,
                                       OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToByteGray",
                                       ST_INT_ARGB__BYTE_GRAY,
                                       OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToRgb24Bit",
					ST_INT_RGB__24BIT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToRgb24Bit",
					ST_INT_ARGB__24BIT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToBgr24Bit",
					ST_INT_RGB__3BYTE_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToBgr24Bit",
					ST_INT_ARGB__3BYTE_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbTo555",
					ST_INT_RGB__SHORT_555_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbTo555",
					ST_INT_ARGB__SHORT_555_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbTo565",
					ST_INT_RGB__SHORT_565_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbTo565",
					ST_INT_ARGB__SHORT_565_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToRgbx",
					ST_INT_RGB__INT_RGBx,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToXbgr",
					ST_INT_RGB__INT_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToBgrx",
					ST_INT_RGB__INT_BGRx,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToRgbx",
					ST_INT_ARGB__INT_RGBx,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToXbgr",
					ST_INT_ARGB__INT_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToBgrx",
					ST_INT_ARGB__INT_BGRx,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToBinary",
					ST_INT_RGB__BINARY,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToBinary",
					ST_INT_ARGB__BINARY,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToDibXrgb",
					ST_INT_RGB__DIB_xRGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToDibXrgb",
					ST_INT_ARGB__DIB_xRGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToArgb",
					ST_INT_RGB__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbToArgbPre",
					ST_INT_RGB__INT_ARGB_PRE,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbTo4byteAbgr",
					ST_INT_RGB__4BYTE_ABGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbTo4byteAbgrPre",
					ST_INT_RGB__4BYTE_ABGR_PRE,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbTo4byteAbgr",
					ST_INT_ARGB__4BYTE_ABGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbTo4byteAbgrPremul",
					ST_INT_ARGB__4BYTE_ABGR_PRE,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbToArgbPremul",
					ST_INT_ARGB__INT_ARGB_PRE,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "DibXrgbToArgb",
					ST_DIB_xRGB__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "DibXrgbToXrgb",
					ST_DIB_xRGB__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ArgbPremulToArgb",
					ST_INT_ARGB_PRE__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ArgbPremulToXrgb",
					ST_INT_ARGB_PRE__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Abgr4byteToArgb",
					ST_4BYTE_ABGR__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Abgr4byteToXrgb",
					ST_4BYTE_ABGR__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Abgr4bytePremulToArgb",
					ST_4BYTE_ABGR_PRE__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Abgr4bytePremulToXrgb",
					ST_4BYTE_ABGR_PRE__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "BinaryToArgb",
					ST_BINARY__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "BinaryToXrgb",
					ST_BINARY__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "BinaryToXbgr",
					ST_BINARY__INT_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteIndexToArgb",
					ST_BYTE_INDEXED__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteIndexToXrgb",
					ST_BYTE_INDEXED__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ShortIsoMorphicCopy",
					ST_SHORT_BINARY__SHORT_BINARY,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Short555ToShort555",
					ST_SHORT_555_RGB__SHORT_555_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Short565ToShort565",
					ST_SHORT_565_RGB__SHORT_565_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Short555ToArgb",
					ST_SHORT_555_RGB__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Short555ToXrgb",
					ST_SHORT_555_RGB__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Short565ToArgb",
					ST_SHORT_565_RGB__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Short565ToXrgb",
					ST_SHORT_565_RGB__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Rgb24BitToArgb",
					ST_24BIT_RGB__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Rgb24BitToXrgb",
					ST_24BIT_RGB__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "RgbxToArgb",
					ST_INT_RGBx__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "RgbxToXrgb",
					ST_INT_RGBx__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Bgr24BitToArgb",
					ST_3BYTE_BGR__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Bgr24BitToXrgb",
					ST_3BYTE_BGR__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "XbgrToArgb",
					ST_INT_BGR__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "XbgrToXrgb",
					ST_INT_BGR__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "BgrxToArgb",
					ST_INT_BGRx__INT_ARGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "BgrxToXrgb",
					ST_INT_BGRx__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbTo8bitDithered",
					ST_INT_RGB__BYTE_INDEXED,
					OpaqueDitheredBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbTo8bitDithered",
					ST_INT_ARGB__BYTE_INDEXED,
					OpaqueDitheredBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntRgbTo8bit",
					ST_INT_RGB__BYTE_INDEXED,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "IntArgbTo8bit",
					ST_INT_ARGB__BYTE_INDEXED,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ColorFillAlphaToXRGB",
					ST_INT_RGB,
					ColorPaint.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ColorFillAlphaToARGB",
					ST_INT_ARGB,
					ColorPaint.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbOntoArgb",
					ST_INT_ARGB__INT_ARGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreOntoArgb",
					ST_INT_ARGB_PRE__INT_ARGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbOntoXrgb",
					ST_INT_ARGB__INT_RGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreOntoXrgb",
					ST_INT_ARGB_PRE__INT_RGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbMaskOntoArgb",
					ST_INT_ARGB_BM__INT_ARGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreMaskOntoArgb",
					ST_INT_ARGB_PRE_BM__INT_ARGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbMaskOntoXrgb",
					ST_INT_ARGB_BM__INT_RGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreMaskOntoXrgb",
					ST_INT_ARGB_PRE_BM__INT_RGB,
					BinaryComposite.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbBkOntoArgb",
					ST_INT_ARGB__INT_ARGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreBkOntoArgb",
					ST_INT_ARGB_PRE__INT_ARGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbBkOntoXrgb",
					ST_INT_ARGB__INT_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreBkOntoXrgb",
					ST_INT_ARGB_PRE__INT_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbBkMaskOntoArgb",
					ST_INT_ARGB__INT_ARGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreBkMaskOntoArgb",
					ST_INT_ARGB_PRE__INT_ARGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbBkMaskOntoXrgb",
					ST_INT_ARGB__INT_ARGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "paintArgbPreBkMaskOntoXrgb",
					ST_INT_ARGB_PRE__INT_ARGB,
					BinaryCompositeBg.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteIndexXparToArgb",
					ST_BYTE_INDEXED_BM__INT_ARGB,
					TransparentBlit.getMethodSignature()),
    	    new GraphicsPrimitiveProxy(this, "ByteGrayToArgb",
                                       ST_BYTE_GRAY__INT_ARGB,
                                       OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteGrayToXbgr",
					ST_BYTE_GRAY__INT_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteGrayToXrgb",
					ST_BYTE_GRAY__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteGrayToByteGray",
                    ST_BYTE_GRAY__BYTE_GRAY,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteGrayToIndex8Gray",
                    ST_BYTE_GRAY__INDEX8_GRAY,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ByteGrayTo8bitDithered",
					ST_BYTE_GRAY__BYTE_INDEXED,
					OpaqueBlit.getMethodSignature()),
    	    new GraphicsPrimitiveProxy(this, "ShortGrayToArgb",
    					ST_USHORT_GRAY__INT_ARGB,
    					OpaqueBlit.getMethodSignature()),
    	    new GraphicsPrimitiveProxy(this, "ArgbToShortGray",
    					ST_INT_ARGB__USHORT_GRAY,
    					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ShortGrayToXbgr",
					ST_USHORT_GRAY__INT_BGR,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ShortGrayToXrgb",
					ST_USHORT_GRAY__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ShortGrayToByteGray",
					ST_USHORT_GRAY__BYTE_GRAY,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "ShortGrayToIndex8Gray",
					ST_USHORT_GRAY__INDEX8_GRAY,
					OpaqueBlit.getMethodSignature())

//            new GraphicsPrimitiveProxy(this, "ByteIndexFillToArgb", xx, xx),
//            new GraphicsPrimitiveProxy(this, "ByteIndexXparToXrgb", xx, xx),
//            new GraphicsPrimitiveProxy(this, "ByteIndexFillToXrgb", xx, xx),
//            new GraphicsPrimitiveProxy(this, "ShortIndexToArgb", xx, xx),
        };
        return primitives;
    }

    /**
    *   INT_XXXX_XXXX isomorphic formats OpaqueBlit
    */
    static native void IntIsomorphicCopy(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    /* Int ARGB To Int Argb OpaqueBlitBg Loops */
    static native void IntArgbbgfillToIntArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );

    /**
    *   TYPE_INT_RGB to many different opaque formats
    *   TYPE_INT_ARGB to many different opaque formats
    */
    static native void IntRgbToRgb24Bit(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbToBgr24Bit(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbTo555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbTo565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbToRgbx(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbToXbgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbToBgrx(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbToBinary(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntArgbToByteGray(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );
   static native void IntRgbTo4byteBgra(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntRgbTo4byteAbgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntArgbTo4byteAbgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntArgbTo4byteAbgrPremul(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void IntArgbToArgbPremul(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    /**
    *   many different opaque formats to TYPE_INT_RGB
    *   many different opaque formats to TYPE_INT_ARGB
    */
    static native void DibXrgbToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void ArgbPremulToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void Abgr4byteToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void Abgr4bytePremulToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void BinaryToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void ByteIndexToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void ShortIsomorphicCopy(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void Short555ToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void Short565ToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void Rgb24BitToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void RgbxToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void Bgr24BitToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void XbgrToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void BgrxToArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    /**
    *   TYPE_INT_RGB  and TYPE_INT_ARGB to dithered 8-bit indexed
    */
    static native void IntRgbTo8bitDithered(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    /**
    *   solid color fill through alpha mask to XRGB
    */
    static native void ColorFillAlphaToXRGB(
        ImageData dstImage,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        int nFGColor
    );

    /**
    *   solid color fill through alpha mask to ARGB
    */
    static native void ColorFillAlphaToARGB(
        ImageData dstImage,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        int nFGColor
    );

    /**
    *   Blend routines
    */
    static native void paintArgbOntoArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintArgbOntoXrgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaScanSize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintArgbMaskOntoArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintArgbMaskOntoXrgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    /**
    *   Blend routines which fill transparent pixels with a background color
    */
    static native void paintArgbBkOntoArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    static native void paintArgbBkOntoXrgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    static native void paintArgbBkMaskOntoArgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    static native void paintArgbBkMaskOntoXrgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    static native void ByteIndexXparToArgb(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
    );

//    static native void ByteIndexFillToArgb(
//        ImageData srcImage,
//        ImageData dstImage,
//        int width, int height,
//
//        int bgColor
//    );
//
//    static native void ShortIndexToArgb(
//        ImageData srcImage,
//        ImageData dstImage,
//        int width, int height
//    );


    static native void ByteGrayToArgb(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
    );
    static native void ByteGrayToByteGray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
    );
    static native void ByteGrayToIndex8Gray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
    );
    static native void ByteGrayTo8bitDithered(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );
    static native void ShortGrayToArgb(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height, int bits
       );
    static native void ArgbToShortGray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );
    static native void ShortGrayToIndex8Gray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height, int bits
       );
    static native void ShortGrayToByteGray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height, int bits
       );
    
}
    /**
    *   INT_XXXX_XXXX isomorphic formats OpaqueBlit
    */
class IntIsoMorphicCopy extends OpaqueBlit {
    IntIsoMorphicCopy() {
	super(DefaultComponent.ST_INT_BINARY__INT_BINARY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntArgbToByteGray extends OpaqueBlit {
    IntArgbToByteGray() {
	super(DefaultComponent.ST_INT_ARGB__BYTE_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntArgbToByteGray(src, dst, w, h);
    }
}
class IntRgbToByteGray extends OpaqueBlit {
    IntRgbToByteGray() {
	super(DefaultComponent.ST_INT_RGB__BYTE_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntArgbToByteGray(src, dst, w, h);
    }
}
class IntRgbToIntRgb extends OpaqueBlit {
    IntRgbToIntRgb() {
	super(DefaultComponent.ST_INT_RGB__ST_INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntArgbToIntRgb extends OpaqueBlit {
    IntArgbToIntRgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntArgbToIntArgb extends OpaqueBlit {
    IntArgbToIntArgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntArgbbgfillToIntArgb extends OpaqueBlitBg {
    IntArgbbgfillToIntArgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_ARGB);
    }
    public void OpaqueBlitBg(ImageData src, ImageData dst, int w, int h,
                           Color bgColor) {
	DefaultComponent.IntArgbbgfillToIntArgb(src, dst, w, h,
                                            bgColor.getRGB());
    }
}
class IntArgbPreToIntArgbPre extends OpaqueBlit {
    IntArgbPreToIntArgbPre() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_ARGB_PRE);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntBgrToIntBgr extends OpaqueBlit {
    IntBgrToIntBgr() {
	super(DefaultComponent.ST_INT_BGR__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntRgbxToIntRgbx extends OpaqueBlit {
    IntRgbxToIntRgbx() {
	super(DefaultComponent.ST_INT_RGBx__INT_RGBx);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntBgrxToIntBgrx extends OpaqueBlit {
    IntBgrxToIntBgrx() {
	super(DefaultComponent.ST_INT_BGRx__INT_BGRx);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntArgbBmToIntArgbBm extends OpaqueBlit {
    IntArgbBmToIntArgbBm() {
	super(DefaultComponent.ST_INT_ARGB_BM__INT_ARGB_BM);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
class IntArgbPreBmToIntArgbPreBm extends OpaqueBlit {
    IntArgbPreBmToIntArgbPreBm() {
	super(DefaultComponent.ST_INT_ARGB_PRE_BM__INT_ARGB_PRE_BM);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntIsomorphicCopy(src, dst, w, h);
    }
}
    /**
    *   TYPE_INT_RGB to many different opaque formats
    *   TYPE_INT_ARGB to many different opaque formats
    */
class IntRgbToRgb24Bit extends OpaqueBlit {
    IntRgbToRgb24Bit() {
	super(DefaultComponent.ST_INT_RGB__24BIT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToRgb24Bit(src, dst, w, h);
    }
}
class IntArgbToRgb24Bit extends OpaqueBlit {
    IntArgbToRgb24Bit() {
	super(DefaultComponent.ST_INT_ARGB__24BIT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToRgb24Bit(src, dst, w, h);
    }
}
class IntRgbToBgr24Bit extends OpaqueBlit {
    IntRgbToBgr24Bit() {
	super(DefaultComponent.ST_INT_RGB__3BYTE_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToBgr24Bit(src, dst, w, h);
    }
}
class IntArgbToBgr24Bit extends OpaqueBlit {
    IntArgbToBgr24Bit() {
	super(DefaultComponent.ST_INT_ARGB__3BYTE_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
            DefaultComponent.IntRgbToBgr24Bit(src, dst, w, h);
    }
}
class IntRgbTo555 extends OpaqueBlit {
    IntRgbTo555() {
	super(DefaultComponent.ST_INT_RGB__SHORT_555_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo555(src, dst, w, h);
    }
}
class IntArgbTo555 extends OpaqueBlit {
    IntArgbTo555() {
	super(DefaultComponent.ST_INT_ARGB__SHORT_555_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo555(src, dst, w, h);
    }
}
class IntRgbTo565 extends OpaqueBlit {
    IntRgbTo565() {
	super(DefaultComponent.ST_INT_RGB__SHORT_565_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo565(src, dst, w, h);
    }
}
class IntArgbTo565 extends OpaqueBlit {
    IntArgbTo565() {
	super(DefaultComponent.ST_INT_ARGB__SHORT_565_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo565(src, dst, w, h);
    }
}
class IntRgbToRgbx extends OpaqueBlit {
    IntRgbToRgbx() {
	super(DefaultComponent.ST_INT_RGB__INT_RGBx);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToRgbx(src, dst, w, h);
    }
}
class IntRgbToXbgr extends OpaqueBlit {
    IntRgbToXbgr() {
	super(DefaultComponent.ST_INT_RGB__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToXbgr(src, dst, w, h);
    }
}
class IntRgbToBgrx extends OpaqueBlit {
    IntRgbToBgrx() {
	super(DefaultComponent.ST_INT_RGB__INT_BGRx);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToBgrx(src, dst, w, h);
    }
}
class IntArgbToRgbx extends OpaqueBlit {
    IntArgbToRgbx() {
	super(DefaultComponent.ST_INT_ARGB__INT_RGBx);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToRgbx(src, dst, w, h);
    }
}
class IntArgbToXbgr extends OpaqueBlit {
    IntArgbToXbgr() {
	super(DefaultComponent.ST_INT_ARGB__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToXbgr(src, dst, w, h);
    }
}
class IntArgbToBgrx extends OpaqueBlit {
    IntArgbToBgrx() {
	super(DefaultComponent.ST_INT_ARGB__INT_BGRx);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToBgrx(src, dst, w, h);
    }
}
class IntRgbToBinary extends OpaqueBlit {
    IntRgbToBinary() {
	super(DefaultComponent.ST_INT_RGB__BINARY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToBinary(src, dst, w, h);
    }
}
class IntArgbToBinary extends OpaqueBlit {
    IntArgbToBinary() {
	super(DefaultComponent.ST_INT_ARGB__BINARY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToBinary(src, dst, w, h);
    }
}
class IntRgbToDibXrgb extends OpaqueBlit {
    IntRgbToDibXrgb() {
	super(DefaultComponent.ST_INT_RGB__DIB_xRGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo4byteBgra(src, dst, w, h);
    }
}
class IntArgbToDibXrgb extends OpaqueBlit {
    IntArgbToDibXrgb() {
	super(DefaultComponent.ST_INT_ARGB__DIB_xRGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo4byteBgra(src, dst, w, h);
    }
}
class IntRgbToArgb extends OpaqueBlit {
    IntRgbToArgb() {
	super(DefaultComponent.ST_INT_RGB__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToArgb(src, dst, w, h);
    }
}
class IntRgbToArgbPre extends OpaqueBlit {
    IntRgbToArgbPre() {
	super(DefaultComponent.ST_INT_RGB__INT_ARGB_PRE);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbToArgb(src, dst, w, h);
    }
}
class IntRgbTo4byteAbgr extends OpaqueBlit {
    IntRgbTo4byteAbgr() {
	super(DefaultComponent.ST_INT_RGB__4BYTE_ABGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo4byteAbgr(src, dst, w, h);
    }
}
class IntRgbTo4byteAbgrPre extends OpaqueBlit {
    IntRgbTo4byteAbgrPre() {
	super(DefaultComponent.ST_INT_RGB__4BYTE_ABGR_PRE);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntRgbTo4byteAbgr(src, dst, w, h);
    }
}
class IntArgbTo4byteAbgr extends OpaqueBlit {
    IntArgbTo4byteAbgr() {
	super(DefaultComponent.ST_INT_ARGB__4BYTE_ABGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntArgbTo4byteAbgr(src, dst, w, h);
    }
}
class IntArgbTo4byteAbgrPremul extends OpaqueBlit {
    IntArgbTo4byteAbgrPremul() {
	super(DefaultComponent.ST_INT_ARGB__4BYTE_ABGR_PRE);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntArgbTo4byteAbgrPremul(src, dst, w, h);
    }
}
class IntArgbToArgbPremul extends OpaqueBlit {
    IntArgbToArgbPremul() {
	super(DefaultComponent.ST_INT_ARGB__INT_ARGB_PRE);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.IntArgbToArgbPremul(src, dst, w, h);
    }
}
class DibXrgbToArgb extends OpaqueBlit {
    DibXrgbToArgb() {
	super(DefaultComponent.ST_DIB_xRGB__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.DibXrgbToArgb(src, dst, w, h);
    }
}
    /**
    *   many different opaque formats to TYPE_INT_RGB
    *   many different opaque formats to TYPE_INT_ARGB
    */
class DibXrgbToXrgb extends OpaqueBlit {
    DibXrgbToXrgb() {
	super(DefaultComponent.ST_DIB_xRGB__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.DibXrgbToArgb(src, dst, w, h);
    }
}
class ArgbPremulToArgb extends OpaqueBlit {
    ArgbPremulToArgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.ArgbPremulToArgb(src, dst, w, h);
    }
}
class ArgbPremulToXrgb extends OpaqueBlit {
    ArgbPremulToXrgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.ArgbPremulToArgb(src, dst, w, h);
    }
}
class Abgr4byteToArgb extends OpaqueBlit {
    Abgr4byteToArgb() {
	super(DefaultComponent.ST_4BYTE_ABGR__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Abgr4byteToArgb(src, dst, w, h);
    }
}
class Abgr4byteToXrgb extends OpaqueBlit {
    Abgr4byteToXrgb() {
	super(DefaultComponent.ST_4BYTE_ABGR__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Abgr4byteToArgb(src, dst, w, h);
    }
}
class Abgr4bytePremulToArgb extends OpaqueBlit {
    Abgr4bytePremulToArgb() {
	super(DefaultComponent.ST_4BYTE_ABGR_PRE__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Abgr4bytePremulToArgb(src, dst, w, h);
    }
}
class Abgr4bytePremulToXrgb extends OpaqueBlit {
    Abgr4bytePremulToXrgb() {
	super(DefaultComponent.ST_4BYTE_ABGR_PRE__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Abgr4bytePremulToArgb(src, dst, w, h);
    }
}
class BinaryToArgb extends OpaqueBlit {
    BinaryToArgb() {
	super(DefaultComponent.ST_BINARY__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.BinaryToArgb(src, dst, w, h);
    }
}
class BinaryToXrgb extends OpaqueBlit {
    BinaryToXrgb() {
	super(DefaultComponent.ST_BINARY__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.BinaryToArgb(src, dst, w, h);
    }
}
class BinaryToXbgr extends OpaqueBlit {
    BinaryToXbgr() {
	super(DefaultComponent.ST_BINARY__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.BinaryToArgb(src, dst, w, h);
    }
}
class ByteIndexToArgb extends OpaqueBlit {
    ByteIndexToArgb() {
	super(DefaultComponent.ST_BYTE_INDEXED__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.ByteIndexToArgb(src, dst, w, h);
    }
}
class ByteIndexToXrgb extends OpaqueBlit {
    ByteIndexToXrgb() {
	super(DefaultComponent.ST_BYTE_INDEXED__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.ByteIndexToArgb(src, dst, w, h);
    }
}
class ShortIsoMorphicCopy extends OpaqueBlit {
    ShortIsoMorphicCopy() {
	super(DefaultComponent.ST_SHORT_BINARY__SHORT_BINARY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.ShortIsomorphicCopy(src, dst, w, h);
    }
}
class Short555ToShort555 extends OpaqueBlit {
    Short555ToShort555() {
        super(DefaultComponent.ST_SHORT_555_RGB__SHORT_555_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.ShortIsomorphicCopy(src, dst, w, h);
    }
}    
class Short565ToShort565 extends OpaqueBlit {
    Short565ToShort565() {
        super(DefaultComponent.ST_SHORT_565_RGB__SHORT_565_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.ShortIsomorphicCopy(src, dst, w, h);
    }
}    
class Short555ToArgb extends OpaqueBlit {
    Short555ToArgb() {
	super(DefaultComponent.ST_SHORT_555_RGB__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Short555ToArgb(src, dst, w, h);
    }
}
class Short555ToXrgb extends OpaqueBlit {
    Short555ToXrgb() {
	super(DefaultComponent.ST_SHORT_555_RGB__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Short555ToArgb(src, dst, w, h);
    }
}
class Short565ToArgb extends OpaqueBlit {
    Short565ToArgb() {
	super(DefaultComponent.ST_SHORT_565_RGB__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Short565ToArgb(src, dst, w, h);
    }
}
class Short565ToXrgb extends OpaqueBlit {
    Short565ToXrgb() {
	super(DefaultComponent.ST_SHORT_565_RGB__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Short565ToArgb(src, dst, w, h);
    }
}
class Rgb24BitToArgb extends OpaqueBlit {
    Rgb24BitToArgb() {
	super(DefaultComponent.ST_24BIT_RGB__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Rgb24BitToArgb(src, dst, w, h);
    }
}
class Rgb24BitToXrgb extends OpaqueBlit {
    Rgb24BitToXrgb() {
	super(DefaultComponent.ST_24BIT_RGB__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Rgb24BitToArgb(src, dst, w, h);
    }
}
class RgbxToArgb extends OpaqueBlit {
    RgbxToArgb() {
	super(DefaultComponent.ST_INT_RGBx__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.RgbxToArgb(src, dst, w, h);
    }
}
class RgbxToXrgb extends OpaqueBlit {
    RgbxToXrgb() {
	super(DefaultComponent.ST_INT_RGBx__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.RgbxToArgb(src, dst, w, h);
    }
}
class Bgr24BitToArgb extends OpaqueBlit {
    Bgr24BitToArgb() {
	super(DefaultComponent.ST_3BYTE_BGR__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Bgr24BitToArgb(src, dst, w, h);
    }
}
class Bgr24BitToXrgb extends OpaqueBlit {
    Bgr24BitToXrgb() {
	super(DefaultComponent.ST_3BYTE_BGR__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.Bgr24BitToArgb(src, dst, w, h);
    }
}
class XbgrToArgb extends OpaqueBlit {
    XbgrToArgb() {
	super(DefaultComponent.ST_INT_BGR__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.XbgrToArgb(src, dst, w, h);
    }
}
class XbgrToXrgb extends OpaqueBlit {
    XbgrToXrgb() {
	super(DefaultComponent.ST_INT_BGR__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.XbgrToArgb(src, dst, w, h);
    }
}
class BgrxToArgb extends OpaqueBlit {
    BgrxToArgb() {
	super(DefaultComponent.ST_INT_BGRx__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.BgrxToArgb(src, dst, w, h);
    }
}
class BgrxToXrgb extends OpaqueBlit {
    BgrxToXrgb() {
	super(DefaultComponent.ST_INT_BGRx__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	DefaultComponent.BgrxToArgb(src, dst, w, h);
    }
}

    /**
    *   TYPE_INT_RGB  and TYPE_INT_ARGB to dithered 8-bit indexed
    */

class IntRgbTo8bitDithered extends OpaqueDitheredBlit {
    IntRgbTo8bitDithered() {
	super(DefaultComponent.ST_INT_RGB__BYTE_INDEXED);
    }
    public void OpaqueDitheredBlit( ImageData src, ImageData dst,
				    int w, int h) {
	DefaultComponent.IntRgbTo8bitDithered(src, dst, w, h);
    }
}
class IntArgbTo8bitDithered extends OpaqueDitheredBlit {
    IntArgbTo8bitDithered() {
	super(DefaultComponent.ST_INT_ARGB__BYTE_INDEXED);
    }
    public void OpaqueDitheredBlit( ImageData src, ImageData dst,
				    int w, int h) {
	DefaultComponent.IntRgbTo8bitDithered(src, dst, w, h);
    }
}
// REMIND: the next two are using dithered loops -- need solid
class IntRgbTo8bit extends OpaqueBlit {
    IntRgbTo8bit() {
	super(DefaultComponent.ST_INT_RGB__BYTE_INDEXED);
    }
    public void OpaqueBlit( ImageData src, ImageData dst,
				    int w, int h) {
	DefaultComponent.IntRgbTo8bitDithered(src, dst, w, h);
    }
}
class IntArgbTo8bit extends OpaqueBlit {
    IntArgbTo8bit() {
	super(DefaultComponent.ST_INT_ARGB__BYTE_INDEXED);
    }
    public void OpaqueBlit( ImageData src, ImageData dst,
				    int w, int h) {
	DefaultComponent.IntRgbTo8bitDithered(src, dst, w, h);
    }
}

    /**
    *   solid color fill through alpha mask to XRGB
    */
class ColorFillAlphaToXRGB extends ColorPaint {
    ColorFillAlphaToXRGB() {
	super(DefaultComponent.ST_INT_RGB);
    }
    public void ColorPaint(
	ImageData dstImage,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color fgColor
    ) {
	DefaultComponent.ColorFillAlphaToXRGB(
	    dstImage, nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, fgColor.getRGB());
    }
}

    /**
    *   solid color fill through alpha mask to ARGB
    */
class ColorFillAlphaToARGB extends ColorPaint {
    ColorFillAlphaToARGB() {
	super(DefaultComponent.ST_INT_ARGB);
    }
    public void ColorPaint(
	ImageData dstImage,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color fgColor
    ) {
	DefaultComponent.ColorFillAlphaToARGB(
	    dstImage, nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, fgColor.getRGB());
    }
}

    /**
    *   Blend routines
    */
class paintArgbOntoArgb extends BinaryComposite {
    paintArgbOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_ARGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintArgbPreOntoArgb extends BinaryComposite {
    paintArgbPreOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_ARGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintArgbOntoXrgb extends BinaryComposite {
    paintArgbOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintArgbPreOntoXrgb extends BinaryComposite {
    paintArgbPreOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintArgbMaskOntoArgb extends BinaryComposite {
    paintArgbMaskOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB_BM__INT_ARGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbMaskOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintArgbPreMaskOntoArgb extends BinaryComposite {
    paintArgbPreMaskOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE_BM__INT_ARGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbMaskOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintArgbMaskOntoXrgb extends BinaryComposite {
    paintArgbMaskOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB_BM__INT_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbMaskOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintArgbPreMaskOntoXrgb extends BinaryComposite {
    paintArgbPreMaskOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE_BM__INT_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	DefaultComponent.paintArgbMaskOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}

    /**
    *   Blend routines which fill transparent pixels with a background color
    */
class paintArgbBkOntoArgb extends BinaryCompositeBg {
    paintArgbBkOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_ARGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintArgbPreBkOntoArgb extends BinaryCompositeBg {
    paintArgbPreBkOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_ARGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintArgbBkOntoXrgb extends BinaryCompositeBg {
    paintArgbBkOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintArgbPreBkOntoXrgb extends BinaryCompositeBg {
    paintArgbPreBkOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintArgbBkMaskOntoArgb extends BinaryCompositeBg {
    paintArgbBkMaskOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_ARGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkMaskOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintArgbPreBkMaskOntoArgb extends BinaryCompositeBg {
    paintArgbPreBkMaskOntoArgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_ARGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkMaskOntoArgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintArgbBkMaskOntoXrgb extends BinaryCompositeBg {
    paintArgbBkMaskOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB__INT_ARGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkMaskOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintArgbPreBkMaskOntoXrgb extends BinaryCompositeBg {
    paintArgbPreBkMaskOntoXrgb() {
	super(DefaultComponent.ST_INT_ARGB_PRE__INT_ARGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	DefaultComponent.paintArgbBkMaskOntoXrgb(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
class ByteIndexXparToArgb extends TransparentBlit {
    ByteIndexXparToArgb() {
	super(DefaultComponent.ST_BYTE_INDEXED_BM__INT_ARGB);
    }
    public void TransparentBlit(ImageData src, ImageData dst,
				int w, int h) {
	DefaultComponent.ByteIndexXparToArgb(src, dst, w, h);
    }
}
class ByteGrayToArgb extends OpaqueBlit {
    ByteGrayToArgb() {
	super(DefaultComponent.ST_BYTE_GRAY__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	DefaultComponent.ByteGrayToArgb(src, dst, w, h);
    }
}
class ByteGrayToXbgr extends OpaqueBlit {
    ByteGrayToXbgr() {
	super(DefaultComponent.ST_BYTE_GRAY__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	DefaultComponent.ByteGrayToArgb(src, dst, w, h);
    }
}
class ByteGrayToByteGray extends OpaqueBlit {
    ByteGrayToByteGray() {
	super(DefaultComponent.ST_BYTE_GRAY__BYTE_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        long mSecs=System.currentTimeMillis();
	DefaultComponent.ByteGrayToByteGray(src, dst, w, h);
    }
}
class ByteGrayToIndex8Gray extends OpaqueBlit {
    ByteGrayToIndex8Gray() {
	super(DefaultComponent.ST_BYTE_GRAY__INDEX8_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	DefaultComponent.ByteGrayToIndex8Gray(src, dst, w, h);
    }
}
class ByteGrayToXrgb extends OpaqueBlit {
    ByteGrayToXrgb() {
	super(DefaultComponent.ST_BYTE_GRAY__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	DefaultComponent.ByteGrayToArgb(src, dst, w, h);
    }
}
class ByteGrayTo8bitDithered extends OpaqueBlit {
    ByteGrayTo8bitDithered() {
	super(DefaultComponent.ST_BYTE_GRAY__BYTE_INDEXED);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	DefaultComponent.ByteGrayTo8bitDithered(src, dst, w, h);
    }
}
class ShortGrayToArgb extends OpaqueBlit {
    ShortGrayToArgb() {
	super(DefaultComponent.ST_USHORT_GRAY__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        int bits = (src.getColorModel()).getComponentSize(0);
        DefaultComponent.ShortGrayToArgb(src, dst, w, h, (bits-8));
    }
}
class ArgbToShortGray extends OpaqueBlit {
    ArgbToShortGray() {
	super(DefaultComponent.ST_INT_ARGB__USHORT_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        DefaultComponent.ArgbToShortGray(src, dst, w, h);
    }
}
class ShortGrayToXbgr extends OpaqueBlit {
    ShortGrayToXbgr() {
	super(DefaultComponent.ST_USHORT_GRAY__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        int bits = (src.getColorModel()).getComponentSize(0);
        DefaultComponent.ShortGrayToArgb(src, dst, w, h, (bits-8));
    }
}
class ShortGrayToXrgb extends OpaqueBlit {
    ShortGrayToXrgb() {
	super(DefaultComponent.ST_USHORT_GRAY__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        int bits = (src.getColorModel()).getComponentSize(0);
        DefaultComponent.ShortGrayToArgb(src, dst, w, h, (bits-8));
    }
}
class ShortGrayToByteGray extends OpaqueBlit {
    ShortGrayToByteGray() {
	super(DefaultComponent.ST_USHORT_GRAY__BYTE_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        int bits = (src.getColorModel()).getComponentSize(0);
        DefaultComponent.ShortGrayToByteGray(src, dst, w, h, (bits-8));
    }
}

class ShortGrayToIndex8Gray extends OpaqueBlit {
    ShortGrayToIndex8Gray() {
	super(DefaultComponent.ST_USHORT_GRAY__INDEX8_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        int bits = (src.getColorModel()).getComponentSize(0);
        DefaultComponent.ShortGrayToIndex8Gray(src, dst, w, h, (bits-8));
    }
}

