/*
 * @(#)DrawBytes.java	1.8 98/08/02
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitive;
import java.awt.Color;
import java.awt.Font;
import java.awt.geom.AffineTransform;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import sun.java2d.loops.DrawBytesRasterContext;

/**
*   DrawBytes
*   1) draw solid color text onto destination surface
*   2) must accept output area [x, y, dx, dy]
*      from within the surface description data for clip rect
*/

public abstract class DrawBytes extends GraphicsPrimitive
{
    private final static String methodSignature =
        "DrawBytes(Lsun/java2d/loops/ImageData;[BIIFFLjava/awt/Color;Ljava/lang/Font;[FZZ)V";

    public static String getMethodSignature() {
        return methodSignature;
    }

    private final static int lockFlags[] = {LOCK_FOR_WRITE | LOCK_FOR_READ};

    protected DrawBytes(int[] surfaceTypes) {
	super(surfaceTypes, lockFlags, methodSignature);
    }

    /**
    *   All DrawBytes implementors must have this invoker method
    */
    public abstract void DrawBytes(
        ImageData dest,
        byte [] srcData, int offset, int length,
        float fX,
        float fY,
        Color fgColor,
        Font fontObject,
        double[] matrix,
        boolean doAntiAlias,
        boolean doFractEnable
    );
    /*
    *   create cached context for raster drawing
    */
    public DrawBytesRasterContext createCachedContext(
            Raster raster, ColorModel cm, Color color, Font font,
            AffineTransform tx,
            boolean doAntiAlias,
            boolean doFractEnable,
            int clipX, int clipY, int clipW, int clipH) {
        return null;
    }
}
