/*
 * @(#)GIFAcceleratorLoops.java	1.18 98/09/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.loops.OpaqueBlit;
import sun.java2d.loops.OpaqueBlitBg;
import sun.java2d.loops.OpaqueDitheredBlit;
import sun.java2d.loops.OpaqueDitheredBlitBg;
import sun.java2d.loops.TransparentBlit;
import sun.java2d.loops.TransparentDitheredBlit;
import java.awt.Color;

/**
 * GIFAcceleratorLoops collection
 * Basically, a collection of components which permit display
 * of 8-bit images on various surfaces, with optional transparency
 * and background fill
 */

public final class GIFAcceleratorLoops implements GraphicsPrimitives {

    //
    // int arrays used for the surface types:
    //

    static int[] ST_BYTE_INDEXED__BYTE_INDEXED = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
						   GraphicsPrimitive.TYPE_BYTE_INDEXED };
    static int[] ST_BYTE_INDEXED__INT_RGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
						   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_BYTE_INDEXED__3BYTE_BGR = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
						   GraphicsPrimitive.TYPE_3BYTE_BGR};
    static int[] ST_BYTE_INDEXED__SHORT_565_RGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
						   GraphicsPrimitive.TYPE_SHORT_565_RGB};
    static int[] ST_BYTE_INDEXED__SHORT_555_RGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
						   GraphicsPrimitive.TYPE_SHORT_555_RGB};
    static int[] ST_BYTE_INDEXED__INT_BGR = { GraphicsPrimitive.TYPE_BYTE_INDEXED, 
						   GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_BYTE_INDEXED_BM__BYTE_INDEXED = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
						   GraphicsPrimitive.TYPE_BYTE_INDEXED};
    static int[] ST_BYTE_INDEXED_BM__INT_RGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
						   GraphicsPrimitive.TYPE_INT_RGB};
    static int[] ST_BYTE_INDEXED_BM__INT_ARGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
						   GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_BYTE_INDEXED_BM__3BYTE_BGR = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
						   GraphicsPrimitive.TYPE_3BYTE_BGR};
    static int[] ST_BYTE_INDEXED_BM__SHORT_565_RGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
						   GraphicsPrimitive.TYPE_SHORT_565_RGB};
    static int[] ST_BYTE_INDEXED_BM__SHORT_555_RGB = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
						   GraphicsPrimitive.TYPE_SHORT_555_RGB};
    static int[] ST_BYTE_INDEXED_BM__INT_BGR = { GraphicsPrimitive.TYPE_BYTE_INDEXED_BM, 
						   GraphicsPrimitive.TYPE_INT_BGR};

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "LUTcopyToIndexed",
					ST_BYTE_INDEXED__BYTE_INDEXED,
					OpaqueBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTcopyDitheredToIndexed",
					ST_BYTE_INDEXED__BYTE_INDEXED,
					OpaqueDitheredBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTcopyToIntRgb",
					ST_BYTE_INDEXED__INT_RGB,
					OpaqueBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTcopyTo24BitBgr",
					ST_BYTE_INDEXED__3BYTE_BGR,
					OpaqueBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTcopyTo16Bit565",
					ST_BYTE_INDEXED__SHORT_565_RGB,
					OpaqueBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTcopyTo16Bit555",
					ST_BYTE_INDEXED__SHORT_555_RGB,
					OpaqueBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTcopyTo32BitXBgr",
					ST_BYTE_INDEXED__INT_BGR,
					OpaqueBlit.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "LUTxparToIndexed",
					ST_BYTE_INDEXED_BM__BYTE_INDEXED,
					TransparentBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTxparDitheredToIndexed",
					ST_BYTE_INDEXED_BM__BYTE_INDEXED,
					TransparentDitheredBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTxparToIntRgb",
					ST_BYTE_INDEXED_BM__INT_RGB,
					TransparentBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTxparTo24BitBgr",
					ST_BYTE_INDEXED_BM__3BYTE_BGR,
					TransparentBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTxparTo16Bit565",
					ST_BYTE_INDEXED_BM__SHORT_565_RGB,
					TransparentBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTxparTo16Bit555",
					ST_BYTE_INDEXED_BM__SHORT_555_RGB,
					TransparentBlit.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTxparTo32BitXBgr",
					ST_BYTE_INDEXED_BM__INT_BGR,
					TransparentBlit.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "LUTbgfillToIndexed",
					ST_BYTE_INDEXED_BM__BYTE_INDEXED,
					OpaqueBlitBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTbgfillDitheredToIndexed",
					ST_BYTE_INDEXED_BM__BYTE_INDEXED,
					OpaqueDitheredBlitBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTbgfillToIntRgb",
					ST_BYTE_INDEXED_BM__INT_RGB,
					OpaqueBlitBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTbgfillToIntArgb",
                                        ST_BYTE_INDEXED_BM__INT_ARGB,
                                        OpaqueBlitBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTbgfillTo24BitBgr",
					ST_BYTE_INDEXED_BM__3BYTE_BGR,
					OpaqueBlitBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTbgfillTo16Bit565",
					ST_BYTE_INDEXED_BM__SHORT_565_RGB,
					OpaqueBlitBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTbgfillTo16Bit555",
					ST_BYTE_INDEXED_BM__SHORT_555_RGB,
					OpaqueBlitBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "LUTbgfillTo32BitXBgr",
					ST_BYTE_INDEXED_BM__INT_BGR,
					OpaqueBlitBg.getMethodSignature())
        };
        return primitives;
    }

/**
*   Opaque copy over destination
*/
    static native void LUTcopyToIndexed(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTcopyDitheredToIndexed(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTcopyToIntRgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTcopyTo24BitBgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTcopyTo16Bit565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTcopyTo16Bit555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTcopyTo32BitXBgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

/**
*   Bitmasked Transparent pixels, Opaque copy over destination
*/
    static native void LUTxparToIndexed(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTxparDitheredToIndexed(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTxparToIntRgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTxparTo24BitBgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTxparTo16Bit565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTxparTo16Bit555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

    static native void LUTxparTo32BitXBgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height
    );

/**
*   Bitmasked Background fill, Opaque copy over destination
*/
    static native void LUTbgfillToIndexed(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );

    static native void LUTbgfillDitheredToIndexed(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );

    static native void LUTbgfillToIntRgb(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );

    static native void LUTbgfillTo24BitBgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );

    static native void LUTbgfillTo16Bit565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );

    static native void LUTbgfillTo16Bit555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );

    static native void LUTbgfillTo32BitXBgr(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nBGColor
    );
}

/**
*   Opaque copy over destination
*/
class LUTcopyToIndexed extends OpaqueBlit {
    LUTcopyToIndexed() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED__BYTE_INDEXED);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	GIFAcceleratorLoops.LUTcopyToIndexed(src, dst, w, h);
    }
}
class LUTcopyDitheredToIndexed extends OpaqueDitheredBlit {
    LUTcopyDitheredToIndexed() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED__BYTE_INDEXED);
    }
    public void OpaqueDitheredBlit( ImageData src, ImageData dst,
				    int w, int h) {
	GIFAcceleratorLoops.LUTcopyDitheredToIndexed(src, dst, w, h);
    }
}
class LUTcopyToIntRgb extends OpaqueBlit {
    LUTcopyToIntRgb() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED__INT_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	GIFAcceleratorLoops.LUTcopyToIntRgb(src, dst, w, h);
    }
}
class LUTcopyTo24BitBgr extends OpaqueBlit {
    LUTcopyTo24BitBgr() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED__3BYTE_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	GIFAcceleratorLoops.LUTcopyTo24BitBgr(src, dst, w, h);
    }
}
class LUTcopyTo16Bit565 extends OpaqueBlit {
    LUTcopyTo16Bit565() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED__SHORT_565_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	GIFAcceleratorLoops.LUTcopyTo16Bit565(src, dst, w, h);
    }
}
class LUTcopyTo16Bit555 extends OpaqueBlit {
    LUTcopyTo16Bit555() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED__SHORT_555_RGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	GIFAcceleratorLoops.LUTcopyTo16Bit555(src, dst, w, h);
    }
}
class LUTcopyTo32BitXBgr extends OpaqueBlit {
    LUTcopyTo32BitXBgr() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	GIFAcceleratorLoops.LUTcopyTo32BitXBgr(src, dst, w, h);
    }
}

/**
*   Bitmasked Transparent pixels, Opaque copy over destination
*/
class LUTxparToIndexed extends TransparentBlit {
    LUTxparToIndexed() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__BYTE_INDEXED);
    }
    public void TransparentBlit(   ImageData src, ImageData dst,
				int w, int h) {
	GIFAcceleratorLoops.LUTxparToIndexed(
				    src, dst, w, h);
    }
}
class LUTxparDitheredToIndexed extends TransparentDitheredBlit {
    LUTxparDitheredToIndexed() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__BYTE_INDEXED);
    }
    public void TransparentDitheredBlit(   ImageData src, ImageData dst,
				int w, int h) {
	GIFAcceleratorLoops.LUTxparDitheredToIndexed(
				    src, dst, w, h);
    }
}
class LUTxparToIntRgb extends TransparentBlit {
    LUTxparToIntRgb() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__INT_RGB);
    }
    public void TransparentBlit(   ImageData src, ImageData dst,
				int w, int h) {
	GIFAcceleratorLoops.LUTxparToIntRgb(src, dst, w, h);
    }
}
class LUTxparTo24BitBgr extends TransparentBlit {
    LUTxparTo24BitBgr() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__3BYTE_BGR);
    }
    public void TransparentBlit(   ImageData src, ImageData dst,
				int w, int h) {
	GIFAcceleratorLoops.LUTxparTo24BitBgr(src, dst, w, h);
    }
}
class LUTxparTo16Bit565 extends TransparentBlit {
    LUTxparTo16Bit565() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__SHORT_565_RGB);
    }
    public void TransparentBlit(   ImageData src, ImageData dst,
				int w, int h) {
	GIFAcceleratorLoops.LUTxparTo16Bit565(src, dst, w, h);
    }
}
class LUTxparTo16Bit555 extends TransparentBlit {
    LUTxparTo16Bit555() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__SHORT_555_RGB);
    }
    public void TransparentBlit(   ImageData src, ImageData dst,
				int w, int h) {
	GIFAcceleratorLoops.LUTxparTo16Bit555(src, dst, w, h);
    }
}
class LUTxparTo32BitXBgr extends TransparentBlit {
    LUTxparTo32BitXBgr() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__INT_BGR);
    }
    public void TransparentBlit(   ImageData src, ImageData dst,
				int w, int h) {
	GIFAcceleratorLoops.LUTxparTo32BitXBgr(src, dst, w, h);
    }
}

/**
*   Bitmasked Background fill, Opaque copy over destination
*/
class LUTbgfillToIndexed extends OpaqueBlitBg {
    LUTbgfillToIndexed() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__BYTE_INDEXED);
    }
    public void OpaqueBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillToIndexed(src, dst, w, h,
					       bgColor.getRGB());
    }
}
class LUTbgfillDitheredToIndexed extends OpaqueDitheredBlitBg {
    LUTbgfillDitheredToIndexed() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__BYTE_INDEXED);
    }
    public void OpaqueDitheredBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillDitheredToIndexed(src, dst, w, h,
						       bgColor.getRGB());
    }
}
class LUTbgfillToIntRgb extends OpaqueBlitBg {
    LUTbgfillToIntRgb() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__INT_RGB);
    }
    public void OpaqueBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillToIntRgb(src, dst, w, h,
					      bgColor.getRGB());
    }
}
class LUTbgfillToIntArgb extends OpaqueBlitBg {
    LUTbgfillToIntArgb() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__INT_ARGB);
    }
    public void OpaqueBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillToIntRgb(src, dst, w, h,
					      bgColor.getRGB());
    }
}
class LUTbgfillTo24BitBgr extends OpaqueBlitBg {
    LUTbgfillTo24BitBgr() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__3BYTE_BGR);
    }
    public void OpaqueBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillTo24BitBgr(src, dst, w, h,
						bgColor.getRGB());
    }
}
class LUTbgfillTo16Bit565 extends OpaqueBlitBg {
    LUTbgfillTo16Bit565() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__SHORT_565_RGB);
    }
    public void OpaqueBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillTo16Bit565(src, dst, w, h,
						bgColor.getRGB());
    }
}
class LUTbgfillTo16Bit555 extends OpaqueBlitBg {
    LUTbgfillTo16Bit555() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__SHORT_555_RGB);
    }
    public void OpaqueBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillTo16Bit555(src, dst, w, h,
						bgColor.getRGB());
    }
}
class LUTbgfillTo32BitXBgr extends OpaqueBlitBg {
    LUTbgfillTo32BitXBgr() {
	super(GIFAcceleratorLoops.ST_BYTE_INDEXED_BM__INT_BGR);
    }
    public void OpaqueBlitBg(   ImageData src, ImageData dst,
				int w, int h, Color bgColor) {
	GIFAcceleratorLoops.LUTbgfillTo32BitXBgr(src, dst, w, h,
						 bgColor.getRGB());
    }
}
