/*
 * @(#)Gray16Blters.java	1.9 98/05/27
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.loops.ColorPaint;
import sun.java2d.loops.BinaryComposite;
import sun.java2d.loops.BinaryCompositeBg;
import java.awt.Color;

/**
 * IndexedCompositing collection
 * Basically, a collection of components which permit alpha paint and
 * composite to occur against 8 bit indexed pixels
 */

public final class Gray16Blters implements GraphicsPrimitives {

    //
    // int array used for the surface types:
    //
    static int[] ST_USHORT_GRAY = { GraphicsPrimitive.TYPE_USHORT_GRAY };
    static int[] ST_INT_RGB__USHORT_GRAY = { GraphicsPrimitive.TYPE_INT_RGB, 
						GraphicsPrimitive.TYPE_USHORT_GRAY };
    static int[] ST_INT_ARGB__USHORT_GRAY = { GraphicsPrimitive.TYPE_INT_ARGB, 
						GraphicsPrimitive.TYPE_USHORT_GRAY };
    static int[] ST_USHORT_GRAY__USHORT_GRAY = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
						GraphicsPrimitive.TYPE_USHORT_GRAY };
    static int[] ST_USHORT_GRAY__INT_RGB = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
						GraphicsPrimitive.TYPE_INT_RGB };
    static int[] ST_USHORT_GRAY__INT_ARGB = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
						GraphicsPrimitive.TYPE_INT_ARGB };
    static int[] ST_USHORT_GRAY__INT_BGR = { GraphicsPrimitive.TYPE_USHORT_GRAY, 
						GraphicsPrimitive.TYPE_INT_BGR };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "ColorFillAlphaToGray16",
					ST_USHORT_GRAY,
					ColorPaint.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintXRGB_ONTO_Gray16",
					ST_INT_RGB__USHORT_GRAY,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_ONTO_Gray16",
					ST_INT_ARGB__USHORT_GRAY,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintGray16_ONTO_Gray16",
					ST_USHORT_GRAY__USHORT_GRAY,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintGray16_ONTO_XRGB",
					ST_USHORT_GRAY__INT_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintGray16_ONTO_ARGB",
					ST_USHORT_GRAY__INT_ARGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintGray16_ONTO_XBGR",
					ST_USHORT_GRAY__INT_BGR,
					BinaryComposite.getMethodSignature())
        };
        return primitives;
    }

    static native void ColorFillAlphaToGray16(
        ImageData dstImage,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        int nFGColor
    );

    static native void paintXRGB_ONTO_Gray16(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_ONTO_Gray16(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintGray16_ONTO_Gray16(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintGray16_ONTO_XRGB(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintGray16_ONTO_ARGB(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintGray16_ONTO_XBGR(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );
}

class ColorFillAlphaToGray16 extends ColorPaint {
    ColorFillAlphaToGray16() {
	super(Gray16Blters.ST_USHORT_GRAY);
    }
    public void ColorPaint(
	ImageData dstImage,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color fgColor
    ) {
	Gray16Blters.ColorFillAlphaToGray16(
	    dstImage, nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, fgColor.getRGB());
    }
}
class paintXRGB_ONTO_Gray16 extends BinaryComposite {
    paintXRGB_ONTO_Gray16() {
	super(Gray16Blters.ST_INT_RGB__USHORT_GRAY);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	Gray16Blters.paintXRGB_ONTO_Gray16(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_ONTO_Gray16 extends BinaryComposite {
    paintARGB_ONTO_Gray16() {
	super(Gray16Blters.ST_INT_ARGB__USHORT_GRAY);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	Gray16Blters.paintARGB_ONTO_Gray16(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintGray16_ONTO_Gray16 extends BinaryComposite {
    paintGray16_ONTO_Gray16() {
	super(Gray16Blters.ST_USHORT_GRAY__USHORT_GRAY);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	Gray16Blters.paintGray16_ONTO_Gray16(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintGray16_ONTO_XRGB extends BinaryComposite {
    paintGray16_ONTO_XRGB() {
	super(Gray16Blters.ST_USHORT_GRAY__INT_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	Gray16Blters.paintGray16_ONTO_XRGB(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintGray16_ONTO_ARGB extends BinaryComposite {
    paintGray16_ONTO_ARGB() {
	super(Gray16Blters.ST_USHORT_GRAY__INT_ARGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	Gray16Blters.paintGray16_ONTO_ARGB(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintGray16_ONTO_XBGR extends BinaryComposite {
    paintGray16_ONTO_XBGR() {
	super(Gray16Blters.ST_USHORT_GRAY__INT_BGR);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	Gray16Blters.paintGray16_ONTO_ARGB(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
