/*
 * @(#)ImageData.java	1.21 98/08/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */


package sun.java2d.loops;

import java.util.*;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.Raster;
import java.awt.image.BufferedImage;
import sun.java2d.SunGraphics2D;
import sun.java2d.loops.GraphicsPrimitive;

public final class ImageData
{
    private static final boolean debug = false;
    private static final boolean debugTrace = false;

    static {
        initIDs();
    }

//
//  Instance members
//
    private int theDataPointer; //
    private int theLutData;
    private int theType;
    private int theLutDataLength;
    private int thePixelStride;
    private int theScanlineStride;
    private int theNumChannels;
    private int thePixelsPerDataUnit;
    private int theBytePerChannel;
    private int theXSize;
    private int theYSize;

    private int theXViewArea;
    private int theYViewArea;
    private int theDXViewArea;
    private int theDYViewArea;

    private int theXDeviceArea;
    private int theYDeviceArea;
    private int theDXDeviceArea;
    private int theDYDeviceArea;

    private int theXOutputArea;
    private int theYOutputArea;
    private int theDXOutputArea;
    private int theDYOutputArea;

    private int offsetX;
    private int offsetY;

    private int thePlatformInfo;
    private int theLockMethod;
    private int theLockFunction;
    private int[] naData;
    private byte[] baData;
    private short[] saData;
    private int[] naLutData;

    private int theTransparency;
    private ColorModel theColorModel;
    private Raster theRaster;

    private int theLockCount;

    private int theGrayInverseLutData;  // pointer set from native code
                                        // points to 256-element C byte array
                                        // assumes C pointer will fit in int
                                        // (similar to theLutData)
//
//  Constructors
//
    public ImageData() {
        this.theType          = GraphicsPrimitive.TYPE_CUSTOM;
        this.theLockMethod    = GraphicsPrimitive.LOCK_FOR_READ
                                | GraphicsPrimitive.LOCK_FOR_WRITE;
    }

    public ImageData(int aType, int lockMethod) {
        this.theType          = aType;
        this.theLockMethod    = lockMethod;
    }
//
//  Member access methods
//
    public final void setDataPointer(int aData) {
        theDataPointer       = aData;
    }
    public final void setLutData(int aLutData) {
        theLutData    = aLutData;
    }
    public final void setType(int aType) {
        theType       = aType;
    }
    public final void setLutDataLength(int aLutDataLength) {
        theLutDataLength  = aLutDataLength;
    }
    public final void setPixelStride(int aPixelStride) {
        thePixelStride = aPixelStride;
    }
    public final void setScanlineStride(int aScanlineStride) {
        theScanlineStride = aScanlineStride;
    }
    public final void setNumChannels(int aNumChannels) {
         theNumChannels = aNumChannels;
    }
    public final void setPixelsPerDataUnit(int aPixPerDat) {
         thePixelsPerDataUnit = aPixPerDat;
    }
    public final void setBytePerChannel(int aBytePerChannel) {
        theBytePerChannel = aBytePerChannel;
    }
    public final void setXSize(int aXSize) {
        theXSize = aXSize;
    }
    public final void setYSize(int aYSize) {
        theYSize = aYSize;
    }
    public void setViewArea(int aX,int aY, int aDX, int aDY)
    {
        theXViewArea = aX;
        theYViewArea = aY;
        theDXViewArea = aDX;
        theDYViewArea = aDY;
    }
    public void setXViewArea (int aXViewArea) {
        theXViewArea = aXViewArea;
    }
    public void setYViewArea(int aYViewArea) {
        theYViewArea = aYViewArea;
    }
    public void setDXViewArea(int aDXViewArea) {
        theDXViewArea = aDXViewArea;
    }
    public void setDYViewArea(int aDYViewArea) {
        theDYViewArea = aDYViewArea;
    }
    public void setDeviceArea(int aX,int aY,int aDX,int aDY) {
        theXDeviceArea = aX;
        theYDeviceArea = aY;
        theDXDeviceArea = aDX;
        theDYDeviceArea = aDY;
    }
    public void setXDeviceArea(int aXDeviceArea) {
        theXDeviceArea = aXDeviceArea;
    }
    public void setYDeviceArea(int aYDeviceArea) {
        theYDeviceArea = aYDeviceArea;
    }
    public void setDXDeviceArea(int aDXDeviceArea) {
        theDXDeviceArea = aDXDeviceArea;
    }
    public void setDYDeviceArea(int aDYDeviceArea) {
        theDYDeviceArea = aDYDeviceArea;
    }
    public void moveOutputArea(int dX,int dY) {
        theXOutputArea += dX;
        theYOutputArea += dY;
        theDXOutputArea -= dX;
        theDYOutputArea -= dY;
    }
    public void setOutputArea(int aX,int aY,int aDX,int aDY) {
        theXOutputArea = aX;
        theYOutputArea = aY;
        theDXOutputArea = aDX;
        theDYOutputArea = aDY;
    }
    public void setXOutputArea(int aXOutputArea) {
        theXOutputArea = aXOutputArea;
    }
    public void setYOutputArea(int aYOutputArea) {
        theYOutputArea = aYOutputArea;
    }
    public void setDXOutputArea(int aDXOutputArea) {
        theDXOutputArea = aDXOutputArea;
    }
    public void setDYOutputArea(int aDYOutputArea) {
        theDYOutputArea = aDYOutputArea;
    }
    public final void setTransparency(int aTransparency) {
        theTransparency = aTransparency;
    }
    public final void setColorModel(ColorModel aColorModel) {
        theColorModel = aColorModel;
    }
    public final void setRaster(Raster aRaster) {
        theRaster = aRaster;
    }

    public final void setLockMethod(int aLockMethod) {
        theLockMethod = aLockMethod;
    }
    public final void setPlatformInfo(int aPlatformInfo) {
        thePlatformInfo   = aPlatformInfo;
    }
    public final void setIntDataArray(int[] intArrData) {
        naData = intArrData;
    }
    public final void setByteDataArray(byte[] byteArrData) {
        baData = byteArrData;
    }
    public final void setShortDataArray(short[] shrtArrData) {
        saData = shrtArrData;
    }
    public final void setIntLutDataArray(int[] intArrLutData) {
        naLutData = intArrLutData;
    }
    public final void setOffsetX(int aOffset) {
        offsetX = aOffset;
    }
    public final void setOffsetY(int aOffset) {
        offsetY = aOffset;
    }
    public final void setGrayInverseLutData(int aGrayInverseLutData) {
        theGrayInverseLutData = aGrayInverseLutData;
    }

    public final int getData() {
        return theDataPointer;
        // + theXOutputArea * thePixelStride * theBytePerChannel +
        //       theYOutputArea * theScanlineStride * theBytePerChannel * thePixelStride;
    }
    public final int getDataPointer() {
        return theDataPointer;
    }
    public final int getLutData() {
        return theLutData;
    }
    public final int getType() {
        return theType;
    }
    public final int getLutDataLength() {
        return theLutDataLength;
    }
    public final int getPixelStride() {
        return thePixelStride;
    }
    public final int getScanlineStride() {
        return theScanlineStride;
    }
    public final int getNumChannels() {
        return theNumChannels;
    }
    public final int getPixelsPerDataUnit() {
        return thePixelsPerDataUnit;
    }
    public final int getBytePerChannel() {
        return theBytePerChannel;
    }
    public final int getXSize() {
        return theXSize;
    }
    public final int getYSize() {
        return theYSize;
    }

    public int getXViewArea() {
        return theXViewArea;
    }
    public int getYViewArea() {
        return theYViewArea;
    }
    public int getDXViewArea() {
        return theDXViewArea;
    }
    public int getDYViewArea() {
        return theDYViewArea;
    }
    public int getXDeviceArea() {
        return theXDeviceArea;
    }
    public int getYDeviceArea() {
        return theYDeviceArea;
    }
    public int getDXDeviceArea() {
        return theDXDeviceArea;
    }
    public int getDYDeviceArea() {
        return theDYDeviceArea;
    }
    public int getXOutputArea() {
        return theXOutputArea;
    }
    public int getYOutputArea() {
        return theYOutputArea;
    }
    public int getDXOutputArea() {
        return theDXOutputArea;
    }
    public int getDYOutputArea() {
        return theDYOutputArea;
    }

    public final int getXView() {
        return getXOutputArea();
    }
    public final int getYView() {
        return getYOutputArea();
    }
    public final int getDXView() {
        return getDXOutputArea();
    }
    public final int getDYView() {
        return getDYOutputArea();
    }


    public final int getLockMethod() {
        return theLockMethod;
    }
    public final int getPlatformInfo() {
        return thePlatformInfo;
    }
    public final int[] getIntDataArray() {
        return naData;
    }
    public final byte[] getByteDataArray() {
        return baData;
    }
    public final short[] getShortDataArray() {
        return saData;
    }
    public final int[] getIntLutDataArray() {
        return naLutData;
    }

    public final int getTransparency() {
        return theTransparency;
    }
    public final ColorModel getColorModel() {
        return theColorModel;
    }
    public final Raster getRaster() {
        return theRaster;
    }
    public final int getOffsetX() {
        return offsetX;
    }
    public final int getOffsetY() {
        return offsetY;
    }
    public final int getGrayInverseLutData() {
        return theGrayInverseLutData;
    }

    // REMIND: will want synchronization for lock count
    public final boolean amILocked() {
        return (0 < theLockCount);
    }
    public final void incLockCount() {
        theLockCount++;
    }
    public final void decLockCount() {
        if (theLockCount > 0) {
            theLockCount--;
        }
    }

   public void printContent() {
//        if (debug)  {
//            System.out.println("====================================");
//            System.out.println("ImageData "+this+" content");
//            System.out.println("====================================");
//            System.out.println(">>>>>>Main data pointer   : "+theDataPointer);
// //            System.out.println(">>>>>>Data type           : "+strType());
//            System.out.println(">>>>>>Data type           : "+theType);
//            System.out.println(">>>>>>Pixel stride        : "+thePixelStride);
//            System.out.println(">>>>>>Scan line stride    : "+
//                               theScanlineStride);
//            System.out.println(">>>>>>Number of channels  : "+theNumChannels);
//            System.out.println(">>>>>>Pixels per Data Unit: "+
//                               thePixelsPerDataUnit);
//            System.out.println(">>>>>>Bytes per channel   : "+
//                               theBytePerChannel);
//            System.out.println(">>>>>>Horizontal size     : "+theXSize);
//            System.out.println(">>>>>>Vertical size       : "+theYSize);
// //
//            System.out.println(">>>>>>Java View Area X0   : "+theXViewArea);
//            System.out.println(">>>>>>Java View Area Y0   : "+theYViewArea);
//            System.out.println(">>>>>>Java View Area DX   : "+theDXViewArea);
//            System.out.println(">>>>>>Java View Area DV   : "+theDYViewArea);
// //
//            System.out.println(">>>>>>Device View Area X0 : "+theXDeviceArea);
//            System.out.println(">>>>>>Device View Area Y0 : "+theYDeviceArea);
//            System.out.println(">>>>>>Device View Area DX : "+theDXDeviceArea);
//            System.out.println(">>>>>>Device View Area DV : "+theDYDeviceArea);
// //
//            System.out.println(">>>>>>Screen View Port X0 : "+theXOutputArea);
//            System.out.println(">>>>>>Screen View Area Y0 : "+theYOutputArea);
//            System.out.println(">>>>>>Screen View Area DX : "+theDXOutputArea);
//            System.out.println(">>>>>>Screen View Area DV : "+theDYOutputArea);
// //
// //
//            if( theLockCount > 0 )
//            {
//                System.out.println(">>>>>>Image is LOCKED");
//            } else  {
//                        System.out.println(">>>>>>Image is NOT LOCKED");
//                    }
// //
//            System.out.println(">>>>>>Platform related data "+
//                               "pointer: "+thePlatformInfo);
//            System.out.println(">>>>>>Method for locking    "+
//                               "       : "+theLockMethod);
//            System.out.println(">>>>>>Integer array of data "+
//                               "       : "+naData);
//            System.out.println(">>>>>>Byte array of data    "+
//                               "       : "+baData);
//            System.out.println(">>>>>>Short array of data   "+
//                               "       : "+saData);
//            System.out.println("====================================");
//        }
   }
//
    /**
     * Initialize JNI field IDs
     */
    private static native void initIDs();

    public static native boolean isAllOpaqueGrayICM(IndexColorModel cm);

    public static native void allocNativeICMData(IndexColorModel cm, int[] rgb,
                                           int rgbsize, boolean allgrayopaque);

    public static native void freeNativeICMData(IndexColorModel cm);

}


