/*
 * @(#)IndexGrayAccelerators.java	1.3 98/09/03
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.loops.OpaqueBlit;
import java.awt.Color;

/**
*  Index*Gray, collection of GraphicsPrimitive
*   Basically, this collection of components performs conversion from
*  Index*_GRAY to INT_ARGB, BYTE_GRAY, INDEX8_GRAY
*/

public final class IndexGrayAccelerators implements GraphicsPrimitives {

    static int[] ST_INT_ARGB__INDEX8_GRAY = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_INDEX8_GRAY};
    static int[] ST_INDEX8_GRAY__INT_ARGB = { GraphicsPrimitive.TYPE_INDEX8_GRAY, 
                                            GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INDEX8_GRAY__INDEX8_GRAY = { GraphicsPrimitive.TYPE_INDEX8_GRAY, 
                      GraphicsPrimitive.TYPE_INDEX8_GRAY};

    static int[] ST_INDEX8_GRAY__BYTE_GRAY = { GraphicsPrimitive.TYPE_INDEX8_GRAY, 
                                         GraphicsPrimitive.TYPE_BYTE_GRAY};
    static int[] ST_INDEX8_GRAY__INT_BGR = { GraphicsPrimitive.TYPE_INDEX8_GRAY, 
                                         GraphicsPrimitive.TYPE_INT_BGR};

    static int[] ST_INT_ARGB__INDEX12_GRAY = { GraphicsPrimitive.TYPE_INT_ARGB, 
					   GraphicsPrimitive.TYPE_INDEX12_GRAY};

    static int[] ST_INDEX12_GRAY__INT_ARGB = { GraphicsPrimitive.TYPE_INDEX12_GRAY, 
                        GraphicsPrimitive.TYPE_INT_ARGB};
    static int[] ST_INDEX12_GRAY__INDEX8_GRAY = { GraphicsPrimitive.TYPE_INDEX12_GRAY, 
                                       GraphicsPrimitive.TYPE_INDEX8_GRAY};
    static int[] ST_INDEX12_GRAY__BYTE_GRAY = { GraphicsPrimitive.TYPE_INDEX12_GRAY, 
                                         GraphicsPrimitive.TYPE_BYTE_GRAY};
    static int[] ST_INDEX12_GRAY__INT_BGR = { GraphicsPrimitive.TYPE_INDEX12_GRAY, 
                                         GraphicsPrimitive.TYPE_INT_BGR};


    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new GraphicsPrimitiveProxy(this, "IntArgbToIndex8Gray",
					ST_INT_ARGB__INDEX8_GRAY,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index8GrayToIndex8Gray",
                    ST_INDEX8_GRAY__INDEX8_GRAY,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index8GrayToByteGray",
                    ST_INDEX8_GRAY__BYTE_GRAY,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index8GrayToIntArgb",
                    ST_INDEX8_GRAY__INT_ARGB,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index8GrayToIntBgr",
                    ST_INDEX8_GRAY__INT_BGR,
                    OpaqueBlit.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "IntArgbToIndex12Gray",
					ST_INT_ARGB__INDEX12_GRAY,
					OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index12GrayToByteGray",
                    ST_INDEX12_GRAY__BYTE_GRAY,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index12GrayToIndex8Gray",
                    ST_INDEX12_GRAY__INDEX8_GRAY,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index12GrayToIntArgb",
                    ST_INDEX12_GRAY__INT_ARGB,
                    OpaqueBlit.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "Index12GrayToIntBgr",
                    ST_INDEX12_GRAY__INT_BGR,
                    OpaqueBlit.getMethodSignature()),
        };
        return primitives;
    }

    static native void IntArgbToIndex8Gray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );

    static native void Index8GrayToIndex8Gray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );

    static native void Index8GrayToByteGray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );
    
    static native void Index8GrayToIntArgb(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );

    static native void IntArgbToIndex12Gray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );

    static native void Index12GrayToIndex8Gray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );

    static native void Index12GrayToByteGray(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );
    
    static native void Index12GrayToIntArgb(
       ImageData srcImage,
       ImageData dstImage,
       int width, int height
       );

}
class IntArgbToIndex8Gray extends OpaqueBlit {
    IntArgbToIndex8Gray() {
	super(IndexGrayAccelerators.ST_INT_ARGB__INDEX8_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.IntArgbToIndex8Gray(src, dst, w, h);
    }
}

class Index8GrayToIndex8Gray extends OpaqueBlit {
    Index8GrayToIndex8Gray() {
	super(IndexGrayAccelerators.ST_INDEX8_GRAY__INDEX8_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.Index8GrayToIndex8Gray(src, dst, w, h);
    }
}
class Index8GrayToByteGray extends OpaqueBlit {
    Index8GrayToByteGray() {
	super(IndexGrayAccelerators.ST_INDEX8_GRAY__BYTE_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.Index8GrayToByteGray(src, dst, w, h);
    }
}
class Index8GrayToIntArgb extends OpaqueBlit {
    Index8GrayToIntArgb() {
	super(IndexGrayAccelerators.ST_INDEX8_GRAY__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.Index8GrayToIntArgb(src, dst, w, h);
    }
}
class Index8GrayToIntBgr extends OpaqueBlit {
    Index8GrayToIntBgr() {
	super(IndexGrayAccelerators.ST_INDEX8_GRAY__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        /* It is a gray component, therefore use the ARGB routine */
	IndexGrayAccelerators.Index8GrayToIntArgb(src, dst, w, h);
    }
}

class Index12GrayToIndex8Gray extends OpaqueBlit {
    Index12GrayToIndex8Gray() {
	super(IndexGrayAccelerators.ST_INDEX12_GRAY__INDEX8_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.Index12GrayToIndex8Gray(src, dst, w, h);
    }
}
class Index12GrayToByteGray extends OpaqueBlit {
    Index12GrayToByteGray() {
	super(IndexGrayAccelerators.ST_INDEX12_GRAY__BYTE_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.Index12GrayToByteGray(src, dst, w, h);
    }
}
class Index12GrayToIntArgb extends OpaqueBlit {
    Index12GrayToIntArgb() {
	super(IndexGrayAccelerators.ST_INDEX12_GRAY__INT_ARGB);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.Index12GrayToIntArgb(src, dst, w, h);
    }
}

class IntArgbToIndex12Gray extends OpaqueBlit {
    IntArgbToIndex12Gray() {
	super(IndexGrayAccelerators.ST_INT_ARGB__INDEX12_GRAY);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
	IndexGrayAccelerators.IntArgbToIndex12Gray(src, dst, w, h);
    }
}
class Index12GrayToIntBgr extends OpaqueBlit {
    Index12GrayToIntBgr() {
	super(IndexGrayAccelerators.ST_INDEX12_GRAY__INT_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst,
			   int w, int h) {
        /* It is a gray component, therefore use the ARGB routine */
	IndexGrayAccelerators.Index12GrayToIntArgb(src, dst, w, h);
    }
}
