/*
 * @(#)IntDiscreteRenderer.java	1.32 98/07/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.awt.image.IntegerComponentRaster;
import sun.java2d.SunGraphics2D;
import java.awt.Color;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import sun.java2d.pipe.SpanIterator;

/**
 * IntDiscreteRenderer collection
 * Basically, a collection of components which permit basic
 * rendering to occur on Integer Rasters with a single
 * component and a pixstride of 1
 */

public final class IntDiscreteRenderer implements GraphicsPrimitives {

    //
    // int arrays used for the surface types:
    //
    static int[] ST_INT_ARGB = { GraphicsPrimitive.TYPE_INT_ARGB };
    static int[] ST_INT_RGB = { GraphicsPrimitive.TYPE_INT_RGB };
    static int[] ST_INT_ARGB_PRE = { GraphicsPrimitive.TYPE_INT_ARGB_PRE };
    static int[] ST_INT_BGR = { GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_INT_RGBx = { GraphicsPrimitive.TYPE_INT_RGBx };
    static int[] ST_INT_BGRx = { GraphicsPrimitive.TYPE_INT_BGRx };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "FillRectARGB",
					ST_INT_ARGB,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectXRGB",
					ST_INT_RGB,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectARGB_PRE",
					ST_INT_ARGB_PRE,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectXBGR",
					ST_INT_BGR,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectRGBX",
					ST_INT_RGBx,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectBGRX",
					ST_INT_BGRx,
					FillRect.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "FillSpansARGB",
					ST_INT_ARGB,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansXRGB",
					ST_INT_RGB,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansARGB_PRE",
					ST_INT_ARGB_PRE,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansXBGR",
					ST_INT_BGR,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansRGBX",
					ST_INT_RGBx,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansBGRX",
					ST_INT_BGRx,
					FillSpans.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawLineARGB",
					ST_INT_ARGB,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineXRGB",
					ST_INT_RGB,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineARGB_PRE",
					ST_INT_ARGB_PRE,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineXBGR",
					ST_INT_BGR,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineRGBX",
					ST_INT_RGBx,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineBGRX",
					ST_INT_BGRx,
					DrawLine.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "FillPolygonsARGB",
					ST_INT_ARGB,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsXRGB",
					ST_INT_RGB,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsARGB_PRE",
					ST_INT_ARGB_PRE,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsXBGR",
					ST_INT_BGR,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsRGBX",
					ST_INT_RGBx,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsBGRX",
					ST_INT_BGRx,
					FillPolygons.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawPolygonsARGB",
					ST_INT_ARGB,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsXRGB",
					ST_INT_RGB,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsARGB_PRE",
					ST_INT_ARGB_PRE,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsXBGR",
					ST_INT_BGR,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsRGBX",
					ST_INT_RGBx,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsBGRX",
					ST_INT_BGRx,
					DrawPolygons.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawRectARGB",
					ST_INT_ARGB,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectXRGB",
					ST_INT_RGB,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectARGB_PRE",
					ST_INT_ARGB_PRE,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectXBGR",
					ST_INT_BGR,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectRGBX",
					ST_INT_RGBx,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectBGRX",
					ST_INT_BGRx,
					DrawRect.getMethodSignature())
        };
        return primitives;
    }

    static int toXBGR(Color clr) {
        int color = clr.getRGB();
        int a,r,g,b;
        a = (color >> 24) & 0xff;
        r = (color >> 16) & 0xff;
        g = (color >> 8) & 0xff;
        b = (color) & 0xff;
        return (b << 16) | (g << 8) | r;
    }
    static int toARGB_PRE(Color clr) {
        int color = clr.getRGB();
        int a,r,g,b;
        a = (color >> 24) & 0xff;
        r = (color >> 16) & 0xff;
        g = (color >> 8) & 0xff;
        b = (color) & 0xff;
        r = (r * a) / 255;
        g = (g * a) / 255;
        b = (b * a) / 255;
        return (a << 24) | (r << 16) | (g << 8) | b;
    }
    static int toRGBX(Color clr) {
        int color = clr.getRGB();
        int a,r,g,b;
        a = (color >> 24) & 0xff;
        r = (color >> 16) & 0xff;
        g = (color >> 8) & 0xff;
        b = (color) & 0xff;
        return (r << 24) | (g << 16) | (b << 8);
    }
    static int toBGRX(Color clr) {
        int color = clr.getRGB();
        int a,r,g,b;
        a = (color >> 24) & 0xff;
        r = (color >> 16) & 0xff;
        g = (color >> 8) & 0xff;
        b = (color) & 0xff;
        return (b << 24) | (g << 16) | (r << 8);
    }

    static native void devSetRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr);

    static native void devSetSpans(
        int color, SpanIterator si, long pIterator,
        int offset, int [] dst, int pixStr, int scanStr);

    static native void devDrawLine(
        int color,
        int x1, int y1, int x2, int y2,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr);

    static native void devFillPolygons(
        int color,
        int xPoints[], int yPoints[],int nPoints[], int numpolys,
        boolean evenodd,
        int originx, int originy,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr);

    static native void devDrawPolygons(
        int color,
        int[] xPoints, int[] yPoints, int nPoints[],
        int numpolys, boolean close,
        int originX, int originY,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr);

    static native void devDrawRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, int [] dst, int pixStr, int scanStr);
}
class ICRFillRectRasterContext extends FillRectRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRFillRectRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	IntDiscreteRenderer.devSetRect(
	    this.argb, x, y, w, h,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class FillRectARGB extends FillRect {
    FillRectARGB() {
	super(IntDiscreteRenderer.ST_INT_ARGB);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetRect(
	    color.getRGB(), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillRectRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class FillRectXRGB extends FillRect {
    FillRectXRGB() {
	super(IntDiscreteRenderer.ST_INT_RGB);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetRect(
	    color.getRGB(), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillRectRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class FillRectARGB_PRE extends FillRect {
    FillRectARGB_PRE() {
	super(IntDiscreteRenderer.ST_INT_ARGB_PRE);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetRect(
	    IntDiscreteRenderer.toARGB_PRE(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toARGB_PRE(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillRectXBGR extends FillRect {
    FillRectXBGR() {
	super(IntDiscreteRenderer.ST_INT_BGR);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetRect(
	    IntDiscreteRenderer.toXBGR(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toXBGR(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillRectRGBX extends FillRect {
    FillRectRGBX() {
	super(IntDiscreteRenderer.ST_INT_RGBx);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetRect(
	    IntDiscreteRenderer.toRGBX(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toRGBX(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillRectBGRX extends FillRect {
    FillRectBGRX() {
	super(IntDiscreteRenderer.ST_INT_BGRx);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetRect(
	    IntDiscreteRenderer.toBGRX(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toBGRX(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillSpansARGB extends FillSpans {
    FillSpansARGB() {
	super(IntDiscreteRenderer.ST_INT_ARGB);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetSpans(
	    color.getRGB(), si, si.getNativeIterator(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
}
class FillSpansXRGB extends FillSpans {
    FillSpansXRGB() {
	super(IntDiscreteRenderer.ST_INT_RGB);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetSpans(
	    color.getRGB(), si, si.getNativeIterator(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
}
class FillSpansARGB_PRE extends FillSpans {
    FillSpansARGB_PRE() {
	super(IntDiscreteRenderer.ST_INT_ARGB_PRE);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetSpans(
	    IntDiscreteRenderer.toARGB_PRE(color), si, si.getNativeIterator(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
}
class FillSpansXBGR extends FillSpans {
    FillSpansXBGR() {
	super(IntDiscreteRenderer.ST_INT_BGR);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetSpans(
	    IntDiscreteRenderer.toXBGR(color), si, si.getNativeIterator(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
}
class FillSpansRGBX extends FillSpans {
    FillSpansRGBX() {
	super(IntDiscreteRenderer.ST_INT_RGBx);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetSpans(
	    IntDiscreteRenderer.toRGBX(color), si, si.getNativeIterator(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
}
class FillSpansBGRX extends FillSpans {
    FillSpansBGRX() {
	super(IntDiscreteRenderer.ST_INT_BGRx);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devSetSpans(
	    IntDiscreteRenderer.toBGRX(color), si, si.getNativeIterator(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
}
class ICRDrawLineRasterContext extends DrawLineRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawLineRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x1, int y1, int x2, int y2) {
	IntDiscreteRenderer.devDrawLine(
	    this.argb, x1, y1, x2, y2,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}

class DrawLineARGB extends DrawLine {
    DrawLineARGB() {
	super(IntDiscreteRenderer.ST_INT_ARGB);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawLine(
	    color.getRGB(), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawLineRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineXRGB extends DrawLine {
    DrawLineXRGB() {
	super(IntDiscreteRenderer.ST_INT_RGB);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawLine(
	    color.getRGB(), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawLineRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineARGB_PRE extends DrawLine {
    DrawLineARGB_PRE() {
	super(IntDiscreteRenderer.ST_INT_ARGB_PRE);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawLine(
	    IntDiscreteRenderer.toARGB_PRE(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawLineRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toARGB_PRE(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineXBGR extends DrawLine {
    DrawLineXBGR() {
	super(IntDiscreteRenderer.ST_INT_BGR);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawLine(
	    IntDiscreteRenderer.toXBGR(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawLineRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toXBGR(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineRGBX extends DrawLine {
    DrawLineRGBX() {
	super(IntDiscreteRenderer.ST_INT_RGBx);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawLine(
	    IntDiscreteRenderer.toRGBX(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawLineRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toRGBX(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineBGRX extends DrawLine {
    DrawLineBGRX() {
	super(IntDiscreteRenderer.ST_INT_BGRx);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawLine(
	    IntDiscreteRenderer.toBGRX(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawLineRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toBGRX(color),
			clipX, clipY, clipW, clipH);
    }
}

class ICRFillPolygonsRasterContext extends FillPolygonsRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRFillPolygonsRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(
	    int xPoints[], int yPoints[], int nPoints[],
	    int numPolys, boolean evenOdd) {
	IntDiscreteRenderer.devFillPolygons(
	    this.argb, xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class FillPolygonsARGB extends FillPolygons {
    FillPolygonsARGB() {
	super(IntDiscreteRenderer.ST_INT_ARGB);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devFillPolygons(
	    color.getRGB(), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillPolygonsRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsXRGB extends FillPolygons {
    FillPolygonsXRGB() {
	super(IntDiscreteRenderer.ST_INT_RGB);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devFillPolygons(
	    color.getRGB(), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillPolygonsRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsARGB_PRE extends FillPolygons {
    FillPolygonsARGB_PRE() {
	super(IntDiscreteRenderer.ST_INT_ARGB_PRE);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devFillPolygons(
	    IntDiscreteRenderer.toARGB_PRE(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toARGB_PRE(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsXBGR extends FillPolygons {
    FillPolygonsXBGR() {
	super(IntDiscreteRenderer.ST_INT_BGR);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devFillPolygons(
	    IntDiscreteRenderer.toXBGR(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toXBGR(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsRGBX extends FillPolygons {
    FillPolygonsRGBX() {
	super(IntDiscreteRenderer.ST_INT_RGBx);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devFillPolygons(
	    IntDiscreteRenderer.toRGBX(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toRGBX(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsBGRX extends FillPolygons {
    FillPolygonsBGRX() {
	super(IntDiscreteRenderer.ST_INT_BGRx);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devFillPolygons(
	    IntDiscreteRenderer.toBGRX(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRFillPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toBGRX(color),
			clipX, clipY, clipW, clipH);
    }
}

class ICRDrawPolygonsRasterContext extends DrawPolygonsRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawPolygonsRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(
	    int xPoints[], int yPoints[], int nPoints[],
	    int numPolys, boolean close) {
	IntDiscreteRenderer.devDrawPolygons(
	    this.argb, xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class DrawPolygonsARGB extends DrawPolygons {
    DrawPolygonsARGB() {
	super(IntDiscreteRenderer.ST_INT_ARGB);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawPolygons(
	    color.getRGB(), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawPolygonsRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsXRGB extends DrawPolygons {
    DrawPolygonsXRGB() {
	super(IntDiscreteRenderer.ST_INT_RGB);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawPolygons(
	    color.getRGB(), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawPolygonsRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsARGB_PRE extends DrawPolygons {
    DrawPolygonsARGB_PRE() {
	super(IntDiscreteRenderer.ST_INT_ARGB_PRE);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawPolygons(
	    IntDiscreteRenderer.toARGB_PRE(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toARGB_PRE(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsXBGR extends DrawPolygons {
    DrawPolygonsXBGR() {
	super(IntDiscreteRenderer.ST_INT_BGR);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawPolygons(
	    IntDiscreteRenderer.toXBGR(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toXBGR(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsRGBX extends DrawPolygons {
    DrawPolygonsRGBX() {
	super(IntDiscreteRenderer.ST_INT_RGBx);
    }
    public int[] getSurfaceTypes() {
	int[] surfaceTypes ={ TYPE_INT_RGBx };
	return surfaceTypes;
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawPolygons(
	    IntDiscreteRenderer.toRGBX(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toRGBX(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsBGRX extends DrawPolygons {
    DrawPolygonsBGRX() {
	super(IntDiscreteRenderer.ST_INT_BGRx);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawPolygons(
	    IntDiscreteRenderer.toBGRX(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawPolygonsRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toBGRX(color),
			clipX, clipY, clipW, clipH);
    }
}

class ICRDrawRectRasterContext extends DrawRectRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawRectRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	IntDiscreteRenderer.devDrawRect(
	    this.argb, x, y, w, h,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class DrawRectARGB extends DrawRect {
    DrawRectARGB() {
	super(IntDiscreteRenderer.ST_INT_ARGB);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawRect(
	    color.getRGB(), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawRectRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectXRGB extends DrawRect {
    DrawRectXRGB() {
	super(IntDiscreteRenderer.ST_INT_RGB);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawRect(
	    color.getRGB(), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawRectRasterContext(
			(IntegerComponentRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectARGB_PRE extends DrawRect {
    DrawRectARGB_PRE() {
	super(IntDiscreteRenderer.ST_INT_ARGB_PRE);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawRect(
	    IntDiscreteRenderer.toARGB_PRE(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toARGB_PRE(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectXBGR extends DrawRect {
    DrawRectXBGR() {
	super(IntDiscreteRenderer.ST_INT_BGR);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawRect(
	    IntDiscreteRenderer.toXBGR(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toXBGR(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectRGBX extends DrawRect {
    DrawRectRGBX() {
	super(IntDiscreteRenderer.ST_INT_RGBx);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawRect(
	    IntDiscreteRenderer.toRGBX(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toRGBX(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectBGRX extends DrawRect {
    DrawRectBGRX() {
	super(IntDiscreteRenderer.ST_INT_BGRx);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	IntegerComponentRaster icr =
		(IntegerComponentRaster)dstImage.getRaster();
	IntDiscreteRenderer.devDrawRect(
	    IntDiscreteRenderer.toBGRX(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    icr.getDataOffset(0),
	    icr.getDataStorage(),
	    icr.getPixelStride(),
	    icr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new ICRDrawRectRasterContext(
			(IntegerComponentRaster)raster, IntDiscreteRenderer.toBGRX(color),
			clipX, clipY, clipW, clipH);
    }
}
