/*
 * @(#)ShortCompositing.java	1.15 98/07/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.loops.ColorPaint;
import sun.java2d.loops.BinaryComposite;
import sun.java2d.loops.BinaryCompositeBg;
import java.awt.Color;

/**
 * ShortCompositing collection
 * Basically, a collection of components which permit alpha paint and
 * composite to occur against 16 bit pixels in 565 or 555 format
 */

public final class ShortCompositing implements GraphicsPrimitives {

    //
    // int array used for the surface types:
    //
    static int[] ST_SHORT_555_RGB = { GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_INT_RGB__SHORT_555_RGB = { GraphicsPrimitive.TYPE_INT_RGB, 
						GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_INT_ARGB__SHORT_555_RGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
						GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_INT_ARGB_PRE__SHORT_555_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE, 
						GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_INT_ARGB_BM__SHORT_555_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_BM, 
						GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_INT_ARGB_PRE_BM__SHORT_555_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM, 
						GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_SHORT_565_RGB = { GraphicsPrimitive.TYPE_SHORT_565_RGB };
    static int[] ST_INT_RGB__SHORT_565_RGB = { GraphicsPrimitive.TYPE_INT_RGB, 
						GraphicsPrimitive.TYPE_SHORT_565_RGB };
    static int[] ST_INT_ARGB__SHORT_565_RGB = { GraphicsPrimitive.TYPE_INT_ARGB, 
						GraphicsPrimitive.TYPE_SHORT_565_RGB };
    static int[] ST_INT_ARGB_PRE__SHORT_565_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE, 
						GraphicsPrimitive.TYPE_SHORT_565_RGB };
    static int[] ST_INT_ARGB_BM__SHORT_565_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_BM, 
						GraphicsPrimitive.TYPE_SHORT_565_RGB };
    static int[] ST_INT_ARGB_PRE_BM__SHORT_565_RGB = { GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM, 
						GraphicsPrimitive.TYPE_SHORT_565_RGB };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "ColorFillAlphaToS555",
					ST_SHORT_555_RGB,
					ColorPaint.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintXRGB_ONTO_S555",
					ST_INT_RGB__SHORT_555_RGB,
					BinaryComposite.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "paintARGB_ONTO_S555",
					ST_INT_ARGB__SHORT_555_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_ONTO_S555",
					ST_INT_ARGB_PRE__SHORT_555_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BITMASK_ONTO_S555",
					ST_INT_ARGB_BM__SHORT_555_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BITMASK_ONTO_S555",
					ST_INT_ARGB_PRE_BM__SHORT_555_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BG_CLR_ONTO_S555",
					ST_INT_ARGB__SHORT_555_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BG_CLR_ONTO_S555",
					ST_INT_ARGB_PRE__SHORT_555_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S555",
					ST_INT_ARGB_BM__SHORT_555_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_S555",
					ST_INT_ARGB_PRE_BM__SHORT_555_RGB,
					BinaryCompositeBg.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "ColorFillAlphaToS565",
					ST_SHORT_565_RGB,
					ColorPaint.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintXRGB_ONTO_S565",
					ST_INT_RGB__SHORT_565_RGB,
					BinaryComposite.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "paintARGB_ONTO_S565",
					ST_INT_ARGB__SHORT_565_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_ONTO_S565",
					ST_INT_ARGB_PRE__SHORT_565_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BITMASK_ONTO_S565",
					ST_INT_ARGB_BM__SHORT_565_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BITMASK_ONTO_S565",
					ST_INT_ARGB_PRE_BM__SHORT_565_RGB,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BG_CLR_ONTO_S565",
					ST_INT_ARGB__SHORT_565_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BG_CLR_ONTO_S565",
					ST_INT_ARGB_PRE__SHORT_565_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S565",
					ST_INT_ARGB_BM__SHORT_565_RGB,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_S565",
					ST_INT_ARGB_PRE_BM__SHORT_565_RGB,
					BinaryCompositeBg.getMethodSignature()),
        };
        return primitives;
    }

    /**
    *   Short 555 formats
    */
    static native void ColorFillAlphaToS555(
        ImageData dstImage,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        int nFGColor
    );

    static native void paintXRGB_ONTO_S555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_ONTO_S555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_WITH_BITMASK_ONTO_S555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_WITH_BG_CLR_ONTO_S555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    static native void paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S555(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    /**
    *   Short 565 formats
    */
    static native void ColorFillAlphaToS565(
        ImageData dstImage,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        int nFGColor
    );

    static native void paintXRGB_ONTO_S565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_ONTO_S565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_WITH_BITMASK_ONTO_S565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_WITH_BG_CLR_ONTO_S565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    static native void paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S565(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );
}
class ColorFillAlphaToS555 extends ColorPaint {
    ColorFillAlphaToS555() {
	super(ShortCompositing.ST_SHORT_555_RGB);
    }
    public void ColorPaint(
	ImageData dstImage,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color fgColor
    ) {
	ShortCompositing.ColorFillAlphaToS555(
	    dstImage, nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, fgColor.getRGB());
    }
}
class paintXRGB_ONTO_S555 extends BinaryComposite {
    paintXRGB_ONTO_S555() {
	super(ShortCompositing.ST_INT_RGB__SHORT_555_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintXRGB_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_ONTO_S555 extends BinaryComposite {
    paintARGB_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB__SHORT_555_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_PRE_ONTO_S555 extends BinaryComposite {
    paintARGB_PRE_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB_PRE__SHORT_555_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintARGB_WITH_BITMASK_ONTO_S555 extends BinaryComposite {
    paintARGB_WITH_BITMASK_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB_BM__SHORT_555_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BITMASK_ONTO_S555 extends BinaryComposite {
    paintARGB_PRE_WITH_BITMASK_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB_PRE_BM__SHORT_555_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintARGB_WITH_BG_CLR_ONTO_S555 extends BinaryCompositeBg {
    paintARGB_WITH_BG_CLR_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB__SHORT_555_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BG_CLR_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BG_CLR_ONTO_S555 extends BinaryCompositeBg {
    paintARGB_PRE_WITH_BG_CLR_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB_PRE__SHORT_555_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BG_CLR_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S555 extends BinaryCompositeBg {
    paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB_BM__SHORT_555_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_S555 extends BinaryCompositeBg {
    paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_S555() {
	super(ShortCompositing.ST_INT_ARGB_PRE_BM__SHORT_555_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S555(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
    /**
    *   Short 565 formats
    */
class ColorFillAlphaToS565 extends ColorPaint {
    ColorFillAlphaToS565() {
	super(ShortCompositing.ST_SHORT_565_RGB);
    }
    public void ColorPaint(
	ImageData dstImage,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color fgColor
    ) {
	ShortCompositing.ColorFillAlphaToS565(
	    dstImage, nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, fgColor.getRGB());
    }
}
class paintXRGB_ONTO_S565 extends BinaryComposite {
    paintXRGB_ONTO_S565() {
	super(ShortCompositing.ST_INT_RGB__SHORT_565_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintXRGB_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_ONTO_S565 extends BinaryComposite {
    paintARGB_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB__SHORT_565_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_PRE_ONTO_S565 extends BinaryComposite {
    paintARGB_PRE_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB_PRE__SHORT_565_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintARGB_WITH_BITMASK_ONTO_S565 extends BinaryComposite {
    paintARGB_WITH_BITMASK_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB_BM__SHORT_565_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BITMASK_ONTO_S565 extends BinaryComposite {
    paintARGB_PRE_WITH_BITMASK_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB_PRE_BM__SHORT_565_RGB);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintARGB_WITH_BG_CLR_ONTO_S565 extends BinaryCompositeBg {
    paintARGB_WITH_BG_CLR_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB__SHORT_565_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BG_CLR_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BG_CLR_ONTO_S565 extends BinaryCompositeBg {
    paintARGB_PRE_WITH_BG_CLR_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB_PRE__SHORT_565_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BG_CLR_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S565 extends BinaryCompositeBg {
    paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB_BM__SHORT_565_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_S565 extends BinaryCompositeBg {
    paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_S565() {
	super(ShortCompositing.ST_INT_ARGB_PRE_BM__SHORT_565_RGB);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	ShortCompositing.paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_S565(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
