/*
 * @(#)ShortDiscreteRenderer.java	1.15 98/07/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.pipe.SpanIterator;
import sun.awt.image.ShortComponentRaster;
import sun.java2d.SunGraphics2D;
import java.awt.image.ColorModel;
import java.awt.Color;
import java.awt.color.ColorSpace;
import java.awt.image.Raster;

/**
 * ShortDiscreteRenderer collection
 * Basically, a collection of components which permit basic
 * rendering to occur on Short Rasters with a single
 * component and a pixstride of 1
 */

public final class ShortDiscreteRenderer implements GraphicsPrimitives {

    //
    // int arrays used for the surface types:
    //
    static int[] ST_SHORT_555_RGB = { GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_SHORT_565_RGB = { GraphicsPrimitive.TYPE_SHORT_565_RGB };
    static int[] ST_USHORT_GRAY = { GraphicsPrimitive.TYPE_USHORT_GRAY };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "FillRectS555",
					ST_SHORT_555_RGB,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectS565",
					ST_SHORT_565_RGB,
					FillRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillRectGray16",
					ST_USHORT_GRAY,
					FillRect.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "FillSpansS555",
					ST_SHORT_555_RGB,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansS565",
					ST_SHORT_565_RGB,
					FillSpans.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillSpansGray16",
					ST_USHORT_GRAY,
					FillSpans.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawLineS555",
					ST_SHORT_555_RGB,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineS565",
					ST_SHORT_565_RGB,
					DrawLine.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawLineGray16",
					ST_USHORT_GRAY,
					DrawLine.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "FillPolygonsS555",
					ST_SHORT_555_RGB,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsS565",
					ST_SHORT_565_RGB,
					FillPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "FillPolygonsGray16",
					ST_USHORT_GRAY,
					FillPolygons.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawPolygonsS555",
					ST_SHORT_555_RGB,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsS565",
					ST_SHORT_565_RGB,
					DrawPolygons.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawPolygonsGray16",
					ST_USHORT_GRAY,
					DrawPolygons.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawRectS555",
					ST_SHORT_555_RGB,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectS565",
					ST_SHORT_565_RGB,
					DrawRect.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "DrawRectGray16",
					ST_USHORT_GRAY,
					DrawRect.getMethodSignature())
        };
        return primitives;
    }

    static int toS555(Color clr) {
        int color = clr.getRGB();
        int r5 = (color >> 19) & 0x1f;
        int g5 = (color >> 11) & 0x1f;
        int b5 = (color >> 3) & 0x1f;
        return (r5 << 10) | (g5 << 5) | b5;
    }
    static int toS565(Color clr) {
        int color = clr.getRGB();
        int r5 = (color >> 19) & 0x1f;
        int g6 = (color >> 10) & 0x3f;
        int b5 = (color >> 3) & 0x1f;
        return (r5 << 11) | (g6 << 5) | b5;
    }
    static int toGray16(Color clr) {
        int color = clr.getRGB();
        int red, green, blue;
        // 255*256 is 65280 so to get the full range we do
        // comp = (comp << 8) + comp
        red   = ((color >> 16) & 0xff);
        red   += (red << 8);
        green = ((color >> 8)  & 0xff);
        green += (green << 8);
        blue  = ((color)       & 0xff);
        blue += (blue << 8);
        return (int)(red*0.299 + green*.587 + blue*.114);
        // REMIND: Might need a conversion if COLOR_QUALITY is high
        // and RGB space is non-linear?
//         ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_GRAY);
//         float rgb[] = clr.getRGBComponents(null);
//         float grayValue[] = cs.fromRGB(rgb);
//         return (int)(grayValue[0] * 65535.0);
    }

    static native void devSetRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr);

    static native void devSetSpans(
        int color, SpanIterator si, long pIterator,
        int offset, short [] dst, int pixStr, int scanStr);

    static native void devDrawLine(
        int color,
        int x1, int y1, int x2, int y2,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr);

    static native void devFillPolygons(
        int color,
        int xPoints[], int yPoints[],int nPoints[], int numpolys,
        boolean evenodd,
        int originx, int originy,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr);

    static native void devDrawPolygons(
        int color,
        int[] xPoints, int[] yPoints, int nPoints[],
        int numpolys, boolean close,
        int originX, int originY,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr);

    static native void devDrawRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, short [] dst, int pixStr, int scanStr);
}

class SCRFillRectRasterContext extends FillRectRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRFillRectRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	ShortDiscreteRenderer.devSetRect(
                this.color, x, y, w, h,
                this.clipX, this.clipY, this.clipW, this.clipH,
                this.offset, this.dst, this.pixStr,
                this.scanStr);
    }
}
class FillRectS555 extends FillRect {
    FillRectS555() {
	super(ShortDiscreteRenderer.ST_SHORT_555_RGB);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devSetRect(
	    ShortDiscreteRenderer.toS555(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRFillRectRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS555(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillRectS565 extends FillRect {
    FillRectS565() {
	super(ShortDiscreteRenderer.ST_SHORT_565_RGB);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devSetRect(
	    ShortDiscreteRenderer.toS565(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRFillRectRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS565(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillRectGray16 extends FillRect {
    FillRectGray16() {
	super(ShortDiscreteRenderer.ST_USHORT_GRAY);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devSetRect(
	    ShortDiscreteRenderer.toGray16(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRFillRectRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toGray16(color),
			clipX, clipY, clipW, clipH);
    }
}

class FillSpansS555 extends FillSpans {
    FillSpansS555() {
	super(ShortDiscreteRenderer.ST_SHORT_555_RGB);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devSetSpans(
	    ShortDiscreteRenderer.toS555(color),
	    si, si.getNativeIterator(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
}
class FillSpansS565 extends FillSpans {
    FillSpansS565() {
	super(ShortDiscreteRenderer.ST_SHORT_565_RGB);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devSetSpans(
	    ShortDiscreteRenderer.toS565(color),
	    si, si.getNativeIterator(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
}
class FillSpansGray16 extends FillSpans {
    FillSpansGray16() {
	super(ShortDiscreteRenderer.ST_USHORT_GRAY);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devSetSpans(
	    ShortDiscreteRenderer.toGray16(color),
	    si, si.getNativeIterator(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
}

class SCRDrawLineRasterContext extends DrawLineRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawLineRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x1, int y1, int x2, int y2) {
	ShortDiscreteRenderer.devDrawLine(
	    this.color, x1, y1, x2, y2,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class DrawLineS555 extends DrawLine {
    DrawLineS555() {
	super(ShortDiscreteRenderer.ST_SHORT_555_RGB);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawLine(
	    ShortDiscreteRenderer.toS555(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawLineRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS555(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineS565 extends DrawLine {
    DrawLineS565() {
	super(ShortDiscreteRenderer.ST_SHORT_565_RGB);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawLine(
	    ShortDiscreteRenderer.toS565(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawLineRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS565(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawLineGray16 extends DrawLine {
    DrawLineGray16() {
	super(ShortDiscreteRenderer.ST_USHORT_GRAY);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawLine(
	    ShortDiscreteRenderer.toGray16(color), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawLineRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toGray16(color),
			clipX, clipY, clipW, clipH);
    }
}

class SCRFillPolygonsRasterContext extends FillPolygonsRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRFillPolygonsRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(
	    int xPoints[], int yPoints[], int nPoints[],
	    int numPolys, boolean evenOdd) {
	ShortDiscreteRenderer.devFillPolygons(
	    this.color, xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class FillPolygonsS555 extends FillPolygons {
    FillPolygonsS555() {
	super(ShortDiscreteRenderer.ST_SHORT_555_RGB);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devFillPolygons(
	    ShortDiscreteRenderer.toS555(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRFillPolygonsRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS555(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsS565 extends FillPolygons {
    FillPolygonsS565() {
	super(ShortDiscreteRenderer.ST_SHORT_565_RGB);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devFillPolygons(
	    ShortDiscreteRenderer.toS565(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRFillPolygonsRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS565(color),
			clipX, clipY, clipW, clipH);
    }
}
class FillPolygonsGray16 extends FillPolygons {
    FillPolygonsGray16() {
	super(ShortDiscreteRenderer.ST_USHORT_GRAY);
    }
    public void FillPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean evenOdd
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devFillPolygons(
	    ShortDiscreteRenderer.toGray16(color), xPoints, yPoints, nPoints,
	    numPolys, evenOdd, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public FillPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRFillPolygonsRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toGray16(color),
			clipX, clipY, clipW, clipH);
    }
}

class SCRDrawPolygonsRasterContext extends DrawPolygonsRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawPolygonsRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(
	    int xPoints[], int yPoints[], int nPoints[],
	    int numPolys, boolean close) {
	ShortDiscreteRenderer.devDrawPolygons(
	    this.color, xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}
class DrawPolygonsS555 extends DrawPolygons {
    DrawPolygonsS555() {
	super(ShortDiscreteRenderer.ST_SHORT_555_RGB);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawPolygons(
	    ShortDiscreteRenderer.toS555(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawPolygonsRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS555(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsS565 extends DrawPolygons {
    DrawPolygonsS565() {
	super(ShortDiscreteRenderer.ST_SHORT_565_RGB);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawPolygons(
	    ShortDiscreteRenderer.toS565(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawPolygonsRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS565(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawPolygonsGray16 extends DrawPolygons {
    DrawPolygonsGray16() {
	super(ShortDiscreteRenderer.ST_USHORT_GRAY);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawPolygons(
	    ShortDiscreteRenderer.toGray16(color), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawPolygonsRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toGray16(color),
			clipX, clipY, clipW, clipH);
    }
}

class SCRDrawRectRasterContext extends DrawRectRasterContext {
    final int color;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawRectRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.color = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	ShortDiscreteRenderer.devDrawRect(
	    this.color, x, y, w, h,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr,
	    this.scanStr);
    }
}


class DrawRectS555 extends DrawRect {
    DrawRectS555() {
	super(ShortDiscreteRenderer.ST_SHORT_555_RGB);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawRect(
	    ShortDiscreteRenderer.toS555(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawRectRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS555(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectS565 extends DrawRect {
    DrawRectS565() {
	super(ShortDiscreteRenderer.ST_SHORT_565_RGB);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawRect(
	    ShortDiscreteRenderer.toS565(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawRectRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toS565(color),
			clipX, clipY, clipW, clipH);
    }
}
class DrawRectGray16 extends DrawRect {
    DrawRectGray16() {
	super(ShortDiscreteRenderer.ST_USHORT_GRAY);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ShortComponentRaster scr =
		(ShortComponentRaster)dstImage.getRaster();
	ShortDiscreteRenderer.devDrawRect(
	    ShortDiscreteRenderer.toGray16(color), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    scr.getDataOffset(0),
	    scr.getDataStorage(),
	    scr.getPixelStride(),
	    scr.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new SCRDrawRectRasterContext(
			(ShortComponentRaster)raster,
                        ShortDiscreteRenderer.toGray16(color),
			clipX, clipY, clipW, clipH);
    }
}
