/*
 * @(#)TextRendering.java	1.22 98/09/15
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import java.awt.Color;
import java.awt.Font;
import java.awt.geom.AffineTransform;
import java.awt.font.GlyphVector;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.color.ColorSpace;
import sun.awt.image.IntegerComponentRaster;
import sun.awt.image.ShortComponentRaster;
import sun.awt.image.ByteComponentRaster;
import sun.awt.font.NativeFontWrapper;
import sun.java2d.SunGraphics2D;
import sun.java2d.loops.DrawStringRasterContext;
import sun.java2d.loops.DrawGlyphVectorRasterContext;
import sun.java2d.loops.DrawCharsRasterContext;
import sun.java2d.loops.DrawBytesRasterContext;
import sun.java2d.loops.DrawString;
import sun.java2d.loops.DrawGlyphVector;
import sun.java2d.loops.DrawChars;
import sun.java2d.loops.DrawBytes;

/**
 * TextRendering collection
 * Basically, a collection of components which permit painting from
 * font renderer onto various surfaces
 */

public class TextRendering  implements GraphicsPrimitives {

    //
    // int array used for the surface types:
    //
    static int[] ST_INT_ARGB = { GraphicsPrimitive.TYPE_INT_ARGB };
    static int[] ST_INT_RGB = { GraphicsPrimitive.TYPE_INT_RGB };
    static int[] ST_INT_BGR = { GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_SHORT_555_RGB = { GraphicsPrimitive.TYPE_SHORT_555_RGB };
    static int[] ST_SHORT_565_RGB = { GraphicsPrimitive.TYPE_SHORT_565_RGB };
    static int[] ST_BYTE_INDEXED =  { GraphicsPrimitive.TYPE_BYTE_INDEXED };
    static int[] ST_BYTE_GRAY =  { GraphicsPrimitive.TYPE_BYTE_GRAY };
    static int[] ST_INDEX8_GRAY =  { GraphicsPrimitive.TYPE_INDEX8_GRAY };
    static int[] ST_3BYTE_BGR = { GraphicsPrimitive.TYPE_3BYTE_BGR };
    static int[] ST_24BIT_RGB = { GraphicsPrimitive.TYPE_24BIT_RGB };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new GraphicsPrimitiveProxy(this, "drawStringARGB",
					ST_INT_ARGB,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawCharsARGB",
					ST_INT_ARGB,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytesARGB",
					ST_INT_ARGB,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVectorARGB",
					ST_INT_ARGB,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawStringRGB",
					ST_INT_RGB,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawCharsRGB",
					ST_INT_RGB,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytesRGB",
					ST_INT_RGB,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVectorRGB",
					ST_INT_RGB,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawStringBGR",
					ST_INT_BGR,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawCharsBGR",
					ST_INT_BGR,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytesBGR",
					ST_INT_BGR,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVectorBGR",
					ST_INT_BGR,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawString555",
					ST_SHORT_555_RGB,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawChars555",
					ST_SHORT_555_RGB,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytes555",
					ST_SHORT_555_RGB,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVector555",
					ST_SHORT_555_RGB,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawString565",
					ST_SHORT_565_RGB,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawChars565",
					ST_SHORT_565_RGB,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytes565",
					ST_SHORT_565_RGB,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVector565",
					ST_SHORT_565_RGB,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawStringIndexed",
					ST_BYTE_INDEXED,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawCharsIndexed",
					ST_BYTE_INDEXED,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytesIndexed",
					ST_BYTE_INDEXED,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVectorIndexed",
					ST_BYTE_INDEXED,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawStringByteGray",
					ST_BYTE_GRAY,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawCharsByteGray",
					ST_BYTE_GRAY,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytesByteGray",
					ST_BYTE_GRAY,
					DrawBytes.getMethodSignature()),
                 new GraphicsPrimitiveProxy(this, "drawGlyphVectorByteGray",
					ST_BYTE_GRAY,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawStringIndex8Gray",
					ST_INDEX8_GRAY,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawCharsIndex8Gray",
					ST_INDEX8_GRAY,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytesIndex8Gray",
					ST_INDEX8_GRAY,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVectorIndex8Gray",
					ST_INDEX8_GRAY,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawString24BITBGR",
					ST_3BYTE_BGR,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawChars24BITBGR",
					ST_3BYTE_BGR,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytes24BITBGR",
					ST_3BYTE_BGR,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVector24BITBGR",
					ST_3BYTE_BGR,
					DrawGlyphVector.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "drawString24BITRGB",
					ST_24BIT_RGB,
					DrawString.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawChars24BITRGB",
					ST_24BIT_RGB,
					DrawChars.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawBytes24BITRGB",
					ST_24BIT_RGB,
					DrawBytes.getMethodSignature()),
            new GraphicsPrimitiveProxy(this, "drawGlyphVector24BITRGB",
					ST_24BIT_RGB,
					DrawGlyphVector.getMethodSignature())

        };
        return primitives;
    }

    static int toS555(Color clr) {
        int color = clr.getRGB();
        int r5 = (color >> 19) & 0x1f;
        int g5 = (color >> 11) & 0x1f;
        int b5 = (color >> 3) & 0x1f;
        return (r5 << 10) | (g5 << 5) | b5;
    }
    static int toS565(Color clr) {
        int color = clr.getRGB();
        int r5 = (color >> 19) & 0x1f;
        int g6 = (color >> 10) & 0x3f;
        int b5 = (color >> 3) & 0x1f;
        return (r5 << 11) | (g6 << 5) | b5;
    }
    static int toGray16(Color clr) {
        // REMIND: Might need a conversion if COLOR_QUALITY is high
        // and RGB space is non-linear?
        int color = clr.getRGB();
        int red, green, blue;
        red   = ((color >> 16) & 0xff);
        red   += (red << 8);
        green = ((color >> 8)  & 0xff);
        green += (green << 8);
        blue  = ((color)       & 0xff);
        blue += (blue << 8);
        return (int)(red*0.299 + green*.587 + blue*.114);
    }
    static int toGray8(Color clr) {
        // REMIND: Might need a conversion if COLOR_QUALITY is high
        // and RGB space is non-linear?
        int color = clr.getRGB();
        int red, green, blue;
        red   = ((color >> 16) & 0xff);
        green = ((color >> 8) & 0xff);
        blue  = ((color) & 0xff);
        return (int)(red*0.299 + green*.587 + blue*.114);
    }
    static int toXBGR(Color clr) {
        int color = clr.getRGB();
        int r = (color >> 16) & 0xff;
        int g = (color >> 8) & 0xff;
        int b = (color) & 0xff;
        return (b << 16) | (g << 8) | r;
    }
    static int toIndexed(Color color, ColorModel cm) {
        byte [] pixels = {0};
        cm.getDataElements(color.getRGB(), pixels);;
        return pixels[0];
    }

}

// anti-aliased string context
class ICRDrawStringRasterContextAA
	extends DrawStringRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawStringRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawStringARGBRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class ICRDrawStringRasterContext
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawStringRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawStringIntDiscreteRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}

class drawStringARGB extends DrawString {
    drawStringARGB() {
	super(TextRendering.ST_INT_ARGB);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawStringARGB(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawStringIntDiscrete(
		srcData, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawStringRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawStringRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}

class drawStringRGB extends DrawString {
    drawStringRGB() {
	super(TextRendering.ST_INT_RGB);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawStringARGB(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawStringIntDiscrete(
		srcData, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawStringRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawStringRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}

// anti-aliased string context
class ICRDrawCharsRasterContextAA
	extends DrawCharsRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawCharsRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
			int offset, int length, float x, float y) {
	NativeFontWrapper.drawCharsARGBRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class ICRDrawCharsRasterContext
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawCharsRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawCharsIntDiscreteRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawCharsARGB extends DrawChars {
    drawCharsARGB() {
	super(TextRendering.ST_INT_ARGB);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawCharsARGB(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawCharsIntDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawCharsRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawCharsRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawCharsRGB extends DrawChars {
    drawCharsRGB() {
	super(TextRendering.ST_INT_RGB);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawCharsARGB(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawCharsIntDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawCharsRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawCharsRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}

// anti-aliased string context
class ICRDrawBytesRasterContextAA
	extends DrawBytesRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawBytesRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] chars,
			int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytesARGBRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class ICRDrawBytesRasterContext
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawBytesRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytesIntDiscreteRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawBytesARGB extends DrawBytes {
    drawBytesARGB() {
	super(TextRendering.ST_INT_ARGB);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytesARGB(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytesIntDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawBytesRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawBytesRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawBytesRGB extends DrawBytes {
    drawBytesRGB() {
	super(TextRendering.ST_INT_RGB);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytesARGB(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytesIntDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawBytesRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawBytesRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
// anti-aliased string context
class ICRDrawGlyphVectorRasterContextAA
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] glyphTX;
    final double [] devTX;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawGlyphVectorRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
            boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVectorARGBRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.doFractEnable,
            this.glyphTX, this.devTX,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class ICRDrawGlyphVectorRasterContext
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawGlyphVectorRasterContext(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVectorIntDiscreteRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.glyphTX, this.devTX,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawGlyphVectorARGB extends DrawGlyphVector {
    drawGlyphVectorARGB() {
	super(TextRendering.ST_INT_ARGB);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVectorARGB(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorIntDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawGlyphVectorRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, glyphTX, devTX, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawGlyphVectorRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawGlyphVectorRGB extends DrawGlyphVector {
    drawGlyphVectorRGB() {
	super(TextRendering.ST_INT_RGB);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVectorARGB(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorIntDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawGlyphVectorRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, glyphTX, devTX, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawGlyphVectorRasterContext(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}
// anti-aliased string context
class ICRDrawStringBGRRasterContextAA
	extends DrawStringRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawStringBGRRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawStringBGRRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawStringBGR extends DrawString {
    drawStringBGR() {
	super(TextRendering.ST_INT_BGR);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawStringBGR(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawStringIntDiscrete(
		srcData, fX, fY,
		fontObject,
		matrix, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawStringBGRRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawStringRasterContext(
			    (IntegerComponentRaster)raster,
			    TextRendering.toXBGR(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
// anti-aliased string context
class ICRDrawCharsBGRRasterContextAA
	extends DrawCharsRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawCharsBGRRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
			int offset, int length, float x, float y) {
	NativeFontWrapper.drawCharsBGRRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawCharsBGR extends DrawChars {
    drawCharsBGR() {
	super(TextRendering.ST_INT_BGR);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawCharsBGR(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawCharsIntDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawCharsBGRRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawCharsRasterContext(
			    (IntegerComponentRaster)raster,
			    TextRendering.toXBGR(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
// anti-aliased string context
class ICRDrawBytesBGRRasterContextAA
	extends DrawBytesRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] matrix;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawBytesBGRRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] chars,
			int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytesBGRRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawBytesBGR extends DrawBytes {
    drawBytesBGR() {
	super(TextRendering.ST_INT_BGR);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytesBGR(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytesIntDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawBytesBGRRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawBytesRasterContext(
			    (IntegerComponentRaster)raster,
			    TextRendering.toXBGR(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
// anti-aliased string context
class ICRDrawGlyphVectorBGRRasterContextAA
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final boolean doFractEnable;
    final double [] glyphTX;
    final double [] devTX;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final int [] dst;
    ICRDrawGlyphVectorBGRRasterContextAA(
	    IntegerComponentRaster icr,
	    int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
            boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = icr.getDataOffset(0);
	this.dst = icr.getDataStorage();
	this.pixStr = icr.getPixelStride();
	this.scanStr = icr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVectorBGRRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.doFractEnable,
            this.glyphTX, this.devTX,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawGlyphVectorBGR extends DrawGlyphVector {
    drawGlyphVectorBGR() {
	super(TextRendering.ST_INT_BGR);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVectorBGR(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorIntDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new ICRDrawGlyphVectorBGRRasterContextAA(
			    (IntegerComponentRaster)raster,
			    color.getRGB(),
			    font, glyphTX, devTX, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new ICRDrawGlyphVectorRasterContext(
			    (IntegerComponentRaster)raster,
			    TextRendering.toXBGR(color),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}
// non anti-aliased string context
class SCRDrawStringRasterContext
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int shortColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawStringRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.shortColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawStringShortDiscreteRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.shortColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class SCRDrawCharsRasterContext
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int shortColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawCharsRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.shortColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawCharsShortDiscreteRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.shortColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class SCRDrawBytesRasterContext
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int shortColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawBytesRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.shortColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] bytes,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytesShortDiscreteRaster(
	    bytes, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.shortColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class SCRDrawGlyphVectorRasterContext
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final float originX, originY;
    final int shortColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawGlyphVectorRasterContext(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.shortColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVectorShortDiscreteRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.glyphTX, this.devTX,
	    this.shortColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 565 string context
class SCRDrawString565RasterContextAA
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawString565RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawString565Raster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 565 string context
class SCRDrawChars565RasterContextAA
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawChars565RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawChars565Raster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 565 string context
class SCRDrawBytes565RasterContextAA
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawBytes565RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] bytes,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytes565Raster(
	    bytes, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 565 string context
class SCRDrawGlyphVector565RasterContextAA
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawGlyphVector565RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVector565Raster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.doFractEnable,
            this.glyphTX, this.devTX,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 555 string context
class SCRDrawString555RasterContextAA
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawString555RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawString555Raster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 555 string context
class SCRDrawChars555RasterContextAA
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawChars555RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawChars555Raster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 555 string context
class SCRDrawBytes555RasterContextAA
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawBytes555RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] bytes,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytes555Raster(
	    bytes, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 555 string context
class SCRDrawGlyphVector555RasterContextAA
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final short [] dst;
    SCRDrawGlyphVector555RasterContextAA(
	    ShortComponentRaster scr,
	    int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = scr.getDataOffset(0);
	this.dst = scr.getDataStorage();
	this.pixStr = scr.getPixelStride();
	this.scanStr = scr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVector555Raster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.doFractEnable,
            this.glyphTX, this.devTX,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawString565 extends DrawString {
    drawString565() {
	super(TextRendering.ST_SHORT_565_RGB);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawString565(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawStringShortDiscrete(
		srcData, fX, fY,
		fontObject,
		matrix, TextRendering.toS565(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawString565RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawStringRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS565(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawChars565 extends DrawChars {
    drawChars565() {
	super(TextRendering.ST_SHORT_565_RGB);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawChars565(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawCharsShortDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toS565(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawChars565RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawCharsRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS565(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawBytes565 extends DrawBytes {
    drawBytes565() {
	super(TextRendering.ST_SHORT_565_RGB);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytes565(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytesShortDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toS565(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawBytes565RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawBytesRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS565(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawGlyphVector565 extends DrawGlyphVector {
    drawGlyphVector565() {
	super(TextRendering.ST_SHORT_565_RGB);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVector565(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorShortDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, TextRendering.toS565(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawGlyphVector565RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, glyphTX, devTX, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawGlyphVectorRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS565(color),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawString555 extends DrawString {
    drawString555() {
	super(TextRendering.ST_SHORT_555_RGB);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawString555(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawStringShortDiscrete(
		srcData, fX, fY,
		fontObject,
		matrix, TextRendering.toS555(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawString555RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawStringRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS555(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawChars555 extends DrawChars {
    drawChars555() {
	super(TextRendering.ST_SHORT_555_RGB);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawChars555(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawCharsShortDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toS555(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawChars555RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawCharsRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS555(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawBytes555 extends DrawBytes {
    drawBytes555() {
	super(TextRendering.ST_SHORT_555_RGB);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytes555(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytesShortDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toS555(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawBytes555RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawBytesRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS555(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawGlyphVector555 extends DrawGlyphVector {
    drawGlyphVector555() {
	super(TextRendering.ST_SHORT_555_RGB);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVector555(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorShortDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, TextRendering.toS555(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new SCRDrawGlyphVector555RasterContextAA(
			    (ShortComponentRaster)raster,
			    color.getRGB(),
			    font, glyphTX, devTX, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new SCRDrawGlyphVectorRasterContext(
			    (ShortComponentRaster)raster,
			    TextRendering.toS555(color),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}
// non anti-aliased string context
class BCRDrawStringRasterContext
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int byteColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawStringRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.byteColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawStringByteDiscreteRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.byteColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class BCRDrawCharsRasterContext
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int byteColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawCharsRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.byteColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawCharsByteDiscreteRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.byteColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class BCRDrawBytesRasterContext
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int byteColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawBytesRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.byteColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] bytes,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytesByteDiscreteRaster(
	    bytes, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.byteColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class BCRDrawGlyphVectorRasterContext
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final float originX, originY;
    final int byteColor;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawGlyphVectorRasterContext(
	    ByteComponentRaster bcr,
	    int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(0);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.byteColor = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVectorByteDiscreteRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font,
            this.glyphTX, this.devTX,
	    this.byteColor,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}

class drawStringIndexed extends DrawString {
    drawStringIndexed() {
	super(TextRendering.ST_BYTE_INDEXED);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawStringIndexed(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawStringByteDiscrete(
		srcData, fX, fY,
		fontObject,
		matrix, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return null;
	} else {
	    return new BCRDrawStringRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawCharsIndexed extends DrawChars {
    drawCharsIndexed() {
	super(TextRendering.ST_BYTE_INDEXED);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawCharsIndexed(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawCharsByteDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return null;
	} else {
	    return new BCRDrawCharsRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawBytesIndexed extends DrawBytes {
    drawBytesIndexed() {
	super(TextRendering.ST_BYTE_INDEXED);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytesIndexed(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytesByteDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return null;
	} else {
	    return new BCRDrawBytesRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawGlyphVectorIndexed extends DrawGlyphVector {
    drawGlyphVectorIndexed() {
	super(TextRendering.ST_BYTE_INDEXED);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVectorIndexed(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorByteDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return null;
	} else {
	    return new BCRDrawGlyphVectorRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}

// @ Byte Gray
class drawStringByteGray extends DrawString {
    drawStringByteGray() {
	super(TextRendering.ST_BYTE_GRAY);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawStringByteGray(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
        
	    NativeFontWrapper.drawStringByteDiscrete(
          srcData, fX, fY,
          fontObject,
          matrix, TextRendering.toGray8(fgColor),
          dest.getXOutputArea(),
          dest.getYOutputArea(),
          dest.getDXOutputArea(),
          dest.getDYOutputArea(),
          dest
          );
    }
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	  	if (doAntiAlias) {
            return null;
        } else {  
            return new BCRDrawStringRasterContext(
                                   (ByteComponentRaster)raster,
                                   TextRendering.toGray8(color),
                                   font, tx,
                                   clipX, clipY, clipW, clipH);
        }
    }
}
class drawCharsByteGray extends DrawChars {
    drawCharsByteGray() {
	super(TextRendering.ST_BYTE_GRAY);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {

	  	if (doAntiAlias) {
            NativeFontWrapper.drawCharsByteGray(
                                   srcData, offset, length, fX, fY,
                                   fontObject, doFractEnable,
                                   matrix, fgColor.getRGB(),
                                   dest.getXOutputArea(),
                                   dest.getYOutputArea(),
                                   dest.getDXOutputArea(),
                                   dest.getDYOutputArea(),
                                   dest
	    );
        } else {
        NativeFontWrapper.drawCharsByteDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toGray8(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
        }
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
       	if (doAntiAlias) {
	    return null;
        } else { 
            return new BCRDrawCharsRasterContext(
                        (ByteComponentRaster)raster,
                        TextRendering.toGray8(color),
                        font, tx,
                        clipX, clipY, clipW, clipH);
        }
    }
}
class drawBytesByteGray extends DrawBytes {
    drawBytesByteGray() {
	super(TextRendering.ST_BYTE_GRAY);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {

      	if (doAntiAlias) {
	    NativeFontWrapper.drawBytesByteGray(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {  
	    NativeFontWrapper.drawBytesByteDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toGray8(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
    }
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
        if (doAntiAlias) {
            return null;
        } else {
            return new BCRDrawBytesRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toGray8(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
        }
    }
}
class drawGlyphVectorByteGray extends DrawGlyphVector {
    drawGlyphVectorByteGray() {
	super(TextRendering.ST_BYTE_GRAY);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVectorByteGray(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorByteDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, TextRendering.toGray8(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
    }
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
        if (doAntiAlias) {
            return null;
        } else {
	    return new BCRDrawGlyphVectorRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toGray8(color),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
        }
    }
}

// @@ INDEX8 GRAY Cases 
class drawStringIndex8Gray extends DrawString {
    drawStringIndex8Gray() {
	super(TextRendering.ST_INDEX8_GRAY);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
        // REMINDER- Add Anti-Alias loops
	if (doAntiAlias) {
	    NativeFontWrapper.drawStringIndex8Gray(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawStringByteDiscrete(
		srcData, fX, fY,
		fontObject,
		matrix, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
      	if (doAntiAlias) {
	    return null;
        } else {
	    return new BCRDrawStringRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, tx,
			    clipX, clipY, clipW, clipH);
        }
    }
}
class drawCharsIndex8Gray extends DrawChars {
    drawCharsIndex8Gray() {
	super(TextRendering.ST_INDEX8_GRAY);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawCharsIndex8Gray(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawCharsByteDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
    }
	}

    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
        if (doAntiAlias) {
            return null;
        } else {
	    return new BCRDrawCharsRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, tx,
			    clipX, clipY, clipW, clipH);
        }
    }
}
class drawBytesIndex8Gray extends DrawBytes {
    drawBytesIndex8Gray() {
	super(TextRendering.ST_INDEX8_GRAY);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
        if (doAntiAlias) {
            NativeFontWrapper.drawBytesIndex8Gray(
                                  srcData, offset, length, fX, fY,
                                  fontObject, doFractEnable,
                                  matrix, fgColor.getRGB(),
                                  dest.getXOutputArea(),
                                  dest.getYOutputArea(),
                                  dest.getDXOutputArea(),
                                  dest.getDYOutputArea(),
                                  dest
	    );
	} else {
	    NativeFontWrapper.drawBytesByteDiscrete(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
    }
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return null;
	} else {
	    return new BCRDrawBytesRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, tx,
			    clipX, clipY, clipW, clipH);
    }
    }
}
class drawGlyphVectorIndex8Gray extends DrawGlyphVector {
    drawGlyphVectorIndex8Gray() {
	super(TextRendering.ST_INDEX8_GRAY);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVectorIndex8Gray(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVectorByteDiscrete(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, TextRendering.toIndexed(fgColor, dest.getColorModel()),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
        }
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
        if (doAntiAlias) {
            return null;
        } else {
            return new BCRDrawGlyphVectorRasterContext(
			    (ByteComponentRaster)raster,
			    TextRendering.toIndexed(color, cm),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
        }
    }
}

// @@ 24BIT
// non anti-aliased string context
class BCRDrawString24BitRasterContext
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int color24Bit;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawString24BitRasterContext(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.color24Bit = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawString3ByteRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.color24Bit,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class BCRDrawChars24BitRasterContext
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int color24Bit;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawChars24BitRasterContext(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.color24Bit = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawChars3ByteRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.color24Bit,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class BCRDrawBytes24BitRasterContext
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final float originX, originY;
    final int color24Bit;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawBytes24BitRasterContext(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.color24Bit = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] bytes,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytes3ByteRaster(
	    bytes, offset, length, x + this.originX, y + this.originY,
	    this.font, this.matrix,
	    this.color24Bit,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// non anti-aliased string context
class BCRDrawGlyphVector24BitRasterContext
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final float originX, originY;
    final int color24Bit;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawGlyphVector24BitRasterContext(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.color24Bit = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVector3ByteRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font,
            this.glyphTX, this.devTX,
	    this.color24Bit,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITRGB string context
class BCRDrawString24BITRGBRasterContextAA
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawString24BITRGBRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawString24BitRGBRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITRGB string context
class BCRDrawChars24BITRGBRasterContextAA
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawChars24BITRGBRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawChars24BitRGBRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITRGB string context
class BCRDrawBytes24BITRGBRasterContextAA
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawBytes24BITRGBRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] bytes,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytes24BitRGBRaster(
	    bytes, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITRGB string context
class BCRDrawGlyphVector24BITRGBRasterContextAA
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawGlyphVector24BITRGBRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVector24BitRGBRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.doFractEnable,
            this.glyphTX, this.devTX,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITBGR string context
class BCRDrawString24BITBGRRasterContextAA
	extends DrawStringRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawString24BITBGRRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(String str, float x, float y) {
	NativeFontWrapper.drawString24BitBGRRaster(
	    str, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITBGR string context
class BCRDrawChars24BITBGRRasterContextAA
	extends DrawCharsRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawChars24BITBGRRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(char [] chars,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawChars24BitBGRRaster(
	    chars, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITBGR string context
class BCRDrawBytes24BITBGRRasterContextAA
	extends DrawBytesRasterContext
{
    final Font font;
    final double [] matrix;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawBytes24BITBGRRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform tx, boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)tx.getTranslateX();
	this.originY = (float)tx.getTranslateY();
	this.matrix = new double[4];
        tx.getMatrix(this.matrix);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(byte [] bytes,
		int offset, int length, float x, float y) {
	NativeFontWrapper.drawBytes24BitBGRRaster(
	    bytes, offset, length, x + this.originX, y + this.originY,
	    this.font, this.doFractEnable, this.matrix,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
// anti-aliased 24BITBGR string context
class BCRDrawGlyphVector24BITBGRRasterContextAA
	extends DrawGlyphVectorRasterContext
{
    final Font font;
    final double [] glyphTX;
    final double [] devTX;
    final boolean doFractEnable;
    final float originX, originY;
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, pixStr, scanStr;
    final byte [] dst;
    BCRDrawGlyphVector24BITBGRRasterContextAA(
	    ByteComponentRaster bcr,
	    int rootBand, int color,
	    Font font, AffineTransform txGlyph,
            AffineTransform txDev,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH)
    {
	this.offset = bcr.getDataOffset(rootBand);
	this.dst = bcr.getDataStorage();
	this.pixStr = bcr.getPixelStride();
	this.scanStr = bcr.getScanlineStride();
	this.font = font;
	this.doFractEnable = doFractEnable;
	this.originX = (float)txGlyph.getTranslateX();
	this.originY = (float)txGlyph.getTranslateY();
	this.glyphTX= new double[4];
        txGlyph.getMatrix(this.glyphTX);
        this.devTX = new double[6];
        txDev.getMatrix(this.devTX);
	this.argb = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(GlyphVector gv, float x, float y) {
	int numGlyphs = gv.getNumGlyphs();
	NativeFontWrapper.drawGlyphVector24BitBGRRaster(
	    gv.getGlyphCodes(0, numGlyphs, null),
	    gv.getGlyphPositions(0, numGlyphs, null),
	    x + this.originX, y + this.originY,
	    this.font, this.doFractEnable,
            this.glyphTX, this.devTX,
	    this.argb,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.pixStr, this.scanStr
	);
    }
}
class drawString24BITRGB extends DrawString {
    drawString24BITRGB() {
	super(TextRendering.ST_24BIT_RGB);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawString24BitRGB(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawString3Byte(
		srcData, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawString24BITRGBRasterContextAA(
			    (ByteComponentRaster)raster,
			    0, color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawString24BitRasterContext(
			    (ByteComponentRaster)raster,
			    0, color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawChars24BITRGB extends DrawChars {
    drawChars24BITRGB() {
	super(TextRendering.ST_24BIT_RGB);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawChars24BitRGB(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawChars3Byte(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawChars24BITRGBRasterContextAA(
			    (ByteComponentRaster)raster,
			    0, color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawChars24BitRasterContext(
			    (ByteComponentRaster)raster,
			    0, color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawBytes24BITRGB extends DrawBytes {
    drawBytes24BITRGB() {
	super(TextRendering.ST_24BIT_RGB);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytes24BitRGB(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytes3Byte(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawBytes24BITRGBRasterContextAA(
			    (ByteComponentRaster)raster,
			    0, color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawBytes24BitRasterContext(
			    (ByteComponentRaster)raster,
			    0, color.getRGB(),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawGlyphVector24BITRGB extends DrawGlyphVector {
    drawGlyphVector24BITRGB() {
	super(TextRendering.ST_24BIT_RGB);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVector24BitRGB(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVector3Byte(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawGlyphVector24BITRGBRasterContextAA(
			    (ByteComponentRaster)raster,
			    0, color.getRGB(),
			    font, glyphTX, devTX, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawGlyphVector24BitRasterContext(
			    (ByteComponentRaster)raster,
		            0, color.getRGB(),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawString24BITBGR extends DrawString {
    drawString24BITBGR() {
	super(TextRendering.ST_3BYTE_BGR);
    }
    public void DrawString(
	    ImageData dest,
	    String srcData,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawString24BitBGR(
		srcData, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawString3Byte(
		srcData, fX, fY,
		fontObject,
		matrix, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawStringRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawString24BITBGRRasterContextAA(
			    (ByteComponentRaster)raster,
			    2, color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawString24BitRasterContext(
			    (ByteComponentRaster)raster,
			    2, TextRendering.toXBGR(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawChars24BITBGR extends DrawChars {
    drawChars24BITBGR() {
	super(TextRendering.ST_3BYTE_BGR);
    }
    public void DrawChars(
	    ImageData dest,
	    char [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawChars24BitBGR(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawChars3Byte(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawCharsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawChars24BITBGRRasterContextAA(
			    (ByteComponentRaster)raster,
			    2, color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawChars24BitRasterContext(
			    (ByteComponentRaster)raster,
			    2, TextRendering.toXBGR(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}
class drawBytes24BITBGR extends DrawBytes {
    drawBytes24BITBGR() {
	super(TextRendering.ST_3BYTE_BGR);
    }
    public void DrawBytes(
	    ImageData dest,
	    byte [] srcData, int offset, int length,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] matrix,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	if (doAntiAlias) {
	    NativeFontWrapper.drawBytes24BitBGR(
		srcData, offset, length, fX, fY,
		fontObject, doFractEnable,
		matrix, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawBytes3Byte(
		srcData, offset, length, fX, fY,
		fontObject,
		matrix, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawBytesRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform tx,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawBytes24BITBGRRasterContextAA(
			    (ByteComponentRaster)raster,
			    2, color.getRGB(),
			    font, tx, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawBytes24BitRasterContext(
			    (ByteComponentRaster)raster,
			    2, TextRendering.toXBGR(color),
			    font, tx,
			    clipX, clipY, clipW, clipH);
	}
    }
}

class drawGlyphVector24BITBGR extends DrawGlyphVector {
    drawGlyphVector24BITBGR() {
	super(TextRendering.ST_3BYTE_BGR);
    }
    public void DrawGlyphVector(
	    ImageData dest,
	    GlyphVector gv,
	    float fX,
	    float fY,
	    Color fgColor,
	    Font fontObject,
	    double[] glyphTX, double[] devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable
    ) {
	int numGlyphs = gv.getNumGlyphs();
	if (doAntiAlias) {
	    NativeFontWrapper.drawGlyphVector24BitBGR(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject, doFractEnable,
		glyphTX, devTX, fgColor.getRGB(),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	} else {
	    NativeFontWrapper.drawGlyphVector3Byte(
		gv.getGlyphCodes(0, numGlyphs, null),
		gv.getGlyphPositions(0, numGlyphs, null),
		fX, fY,
		fontObject,
		glyphTX, devTX, TextRendering.toXBGR(fgColor),
		dest.getXOutputArea(),
		dest.getYOutputArea(),
		dest.getDXOutputArea(),
		dest.getDYOutputArea(),
		dest
	    );
	}
    }
    public DrawGlyphVectorRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color, Font font,
	    AffineTransform glyphTX, AffineTransform devTX,
	    boolean doAntiAlias,
	    boolean doFractEnable,
	    int clipX, int clipY, int clipW, int clipH) {
	if (doAntiAlias) {
	    return new BCRDrawGlyphVector24BITBGRRasterContextAA(
			    (ByteComponentRaster)raster,
			    2, color.getRGB(),
			    font, glyphTX, devTX, doFractEnable,
			    clipX, clipY, clipW, clipH);
	} else {
	    return new BCRDrawGlyphVector24BitRasterContext(
			    (ByteComponentRaster)raster,
			    2, TextRendering.toXBGR(color),
			    font, glyphTX, devTX,
			    clipX, clipY, clipW, clipH);
	}
    }
}
