/*
 * @(#)ThreeByteRenderer.java	1.1 98/09/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.loops;

import sun.java2d.loops.GraphicsPrimitives;
import sun.awt.image.ByteInterleavedRaster;
import sun.java2d.SunGraphics2D;
import java.awt.Color;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import sun.java2d.pipe.SpanIterator;

/**
 * 3BYTE Renderer collection
 * Basically, a collection of components which permit basic
 * rendering to occur on 24-bit Rasters with B,G,R components
 * and a pixstride of 3
 */
public final class ThreeByteRenderer implements GraphicsPrimitives {

    //
    // int arrays used for the surface types:
    //
    static int[] ST_3BYTE_BGR = { GraphicsPrimitive.TYPE_3BYTE_BGR };
    static int[] ST_3BYTE_BGR__3BYTE_BGR = {
	GraphicsPrimitive.TYPE_3BYTE_BGR,
	GraphicsPrimitive.TYPE_3BYTE_BGR
	};

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "FillRectBGR3",
					ST_3BYTE_BGR,
					FillRect.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "FillSpansBGR3",
					ST_3BYTE_BGR,
					FillSpans.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawLineBGR3",
					ST_3BYTE_BGR,
					DrawLine.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawPolygonsBGR3",
					ST_3BYTE_BGR,
					DrawPolygons.getMethodSignature()),

            new  GraphicsPrimitiveProxy(this, "DrawRectBGR3",
					ST_3BYTE_BGR,
					DrawRect.getMethodSignature()),

            new GraphicsPrimitiveProxy(this, "ThreeByteIsoMorphicCopy", 
				       ST_3BYTE_BGR__3BYTE_BGR,
				       OpaqueBlit.getMethodSignature()),
        };
        return primitives;
    }

    static native void devSetRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int scanStr);

    static native void devSetSpans(
        int color, SpanIterator si, long pIterator,
        int offset, byte [] dst, int scanStr);

    static native void devDrawLine(
        int color,
        int x1, int y1, int x2, int y2,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int scanStr);

    static native void devDrawPolygons(
        int color,
        int[] xPoints, int[] yPoints, int nPoints[],
        int numpolys, boolean close,
        int originX, int originY,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int scanStr);

    static native void devDrawRect(
        int color,
        int x, int y, int w, int h,
        int cx, int cy, int cw, int ch,
        int offset, byte [] dst, int scanStr);

    static native void IsomorphicCopy(ImageData src, ImageData dst,
				      int w, int h);
}

class ThreeByteIsoMorphicCopy extends OpaqueBlit {
    ThreeByteIsoMorphicCopy() {
	super(ThreeByteRenderer.ST_3BYTE_BGR__3BYTE_BGR);
    }
    public void OpaqueBlit(ImageData src, ImageData dst, int w, int h) {
	ThreeByteRenderer.IsomorphicCopy(src, dst, w, h);
    }
}

class BGR3FillRectRasterContext extends FillRectRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, scanStr;
    final byte [] dst;
    BGR3FillRectRasterContext(
	    ByteInterleavedRaster bir,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bir.getDataOffset(2);
	this.dst = bir.getDataStorage();
	this.scanStr = bir.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	ThreeByteRenderer.devSetRect(
	    this.argb, x, y, w, h,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.scanStr);
    }
}

class FillRectBGR3 extends FillRect {
    FillRectBGR3() {
	super(ThreeByteRenderer.ST_3BYTE_BGR);
    }
    public void FillRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ByteInterleavedRaster bir =
		(ByteInterleavedRaster)dstImage.getRaster();
	ThreeByteRenderer.devSetRect(
	    color.getRGB(), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bir.getDataOffset(2),
	    bir.getDataStorage(),
	    bir.getScanlineStride());
    }
    public FillRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BGR3FillRectRasterContext(
			(ByteInterleavedRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}

class FillSpansBGR3 extends FillSpans {
    FillSpansBGR3() {
	super(ThreeByteRenderer.ST_3BYTE_BGR);
    }
    public void FillSpans(
	ImageData dstImage,
	Color color,
	SpanIterator si
    ) {
	ByteInterleavedRaster bir =
		(ByteInterleavedRaster)dstImage.getRaster();
	ThreeByteRenderer.devSetSpans(
	    color.getRGB(), si, si.getNativeIterator(),
	    bir.getDataOffset(2),
	    bir.getDataStorage(),
	    bir.getScanlineStride());
    }
}

class BGR3DrawLineRasterContext extends DrawLineRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, scanStr;
    final byte [] dst;
    BGR3DrawLineRasterContext(
	    ByteInterleavedRaster bir,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bir.getDataOffset(2);
	this.dst = bir.getDataStorage();
	this.scanStr = bir.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x1, int y1, int x2, int y2) {
	ThreeByteRenderer.devDrawLine(
	    this.argb, x1, y1, x2, y2,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.scanStr);
    }
}

class DrawLineBGR3 extends DrawLine {
    DrawLineBGR3() {
	super(ThreeByteRenderer.ST_3BYTE_BGR);
    }
    public void DrawLine(
	ImageData dstImage,
	Color color,
	int x1, int y1, int x2, int y2
    ) {
	ByteInterleavedRaster bir =
		(ByteInterleavedRaster)dstImage.getRaster();
	ThreeByteRenderer.devDrawLine(
	    color.getRGB(), x1, y1, x2, y2,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bir.getDataOffset(2),
	    bir.getDataStorage(),
	    bir.getScanlineStride());
    }
    public DrawLineRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BGR3DrawLineRasterContext(
			(ByteInterleavedRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}

class BGR3DrawPolygonsRasterContext extends DrawPolygonsRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, scanStr;
    final byte [] dst;
    BGR3DrawPolygonsRasterContext(
	    ByteInterleavedRaster bir,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bir.getDataOffset(2);
	this.dst = bir.getDataStorage();
	this.scanStr = bir.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(
	    int xPoints[], int yPoints[], int nPoints[],
	    int numPolys, boolean close) {
	ThreeByteRenderer.devDrawPolygons(
	    this.argb, xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.scanStr);
    }
}

class DrawPolygonsBGR3 extends DrawPolygons {
    DrawPolygonsBGR3() {
	super(ThreeByteRenderer.ST_3BYTE_BGR);
    }
    public void DrawPolygons(
	ImageData dstImage,
	Color color,
	int xPoints[], int yPoints[], int nPoints[],
	int numPolys, boolean close
    ) {
	ByteInterleavedRaster bir =
		(ByteInterleavedRaster)dstImage.getRaster();
	ThreeByteRenderer.devDrawPolygons(
	    color.getRGB(), xPoints, yPoints, nPoints,
	    numPolys, close, 0, 0,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bir.getDataOffset(2),
	    bir.getDataStorage(),
	    bir.getScanlineStride());
    }
    public DrawPolygonsRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BGR3DrawPolygonsRasterContext(
			(ByteInterleavedRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}

class BGR3DrawRectRasterContext extends DrawRectRasterContext {
    final int argb;
    final int clipX, clipY, clipW, clipH;
    final int offset, scanStr;
    final byte [] dst;
    BGR3DrawRectRasterContext(
	    ByteInterleavedRaster bir,
	    int color,
	    int clipX, int clipY, int clipW, int clipH) {
	this.offset = bir.getDataOffset(2);
	this.dst = bir.getDataStorage();
	this.scanStr = bir.getScanlineStride();
	this.argb  = color;
	this.clipX = clipX;
	this.clipY = clipY;
	this.clipW = clipW;
	this.clipH = clipH;
    }
    public void invoke(int x, int y, int w, int h) {
	ThreeByteRenderer.devDrawRect(
	    this.argb, x, y, w, h,
	    this.clipX, this.clipY, this.clipW, this.clipH,
	    this.offset, this.dst, this.scanStr);
    }
}

class DrawRectBGR3 extends DrawRect {
    DrawRectBGR3() {
	super(ThreeByteRenderer.ST_3BYTE_BGR);
    }
    public void DrawRect(
	ImageData dstImage,
	Color color,
	int x, int y, int w, int h
    ) {
	ByteInterleavedRaster bir =
		(ByteInterleavedRaster)dstImage.getRaster();
	ThreeByteRenderer.devDrawRect(
	    color.getRGB(), x, y, w, h,
	    dstImage.getXOutputArea(),
	    dstImage.getYOutputArea(),
	    dstImage.getDXOutputArea(),
	    dstImage.getDYOutputArea(),
	    bir.getDataOffset(2),
	    bir.getDataStorage(),
	    bir.getScanlineStride());
    }
    public DrawRectRasterContext createCachedContext(
	    Raster raster, ColorModel cm, Color color,
	    int clipX, int clipY, int clipW, int clipH) {
	return new BGR3DrawRectRasterContext(
			(ByteInterleavedRaster)raster, color.getRGB(),
			clipX, clipY, clipW, clipH);
    }
}
