/*
 * @(#)XBGRLoops.java	1.15 98/07/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

package sun.java2d.loops;


import sun.java2d.loops.GraphicsPrimitives;
import sun.java2d.loops.ColorPaint;
import sun.java2d.loops.BinaryComposite;
import sun.java2d.loops.BinaryCompositeBg;
import java.awt.Color;

/**
 * XBGRLoops collection
 * Basically, a collection of components which permit alpha paint and
 * composite to occur against INT_BGR pixels
 */

public final class XBGRLoops implements GraphicsPrimitives {

    //
    // int array used for the surface types:
    //
    static int[] ST_INT_BGR = { GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_INT_RGB__INT_BGR = { GraphicsPrimitive.TYPE_INT_RGB, 
						GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_INT_ARGB__INT_BGR = { GraphicsPrimitive.TYPE_INT_ARGB, 
						GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_INT_ARGB_PRE__INT_BGR = { GraphicsPrimitive.TYPE_INT_ARGB_PRE, 
						GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_INT_ARGB_BM__INT_BGR = { GraphicsPrimitive.TYPE_INT_ARGB_BM, 
						GraphicsPrimitive.TYPE_INT_BGR };
    static int[] ST_INT_ARGB_PRE_BM__INT_BGR = { GraphicsPrimitive.TYPE_INT_ARGB_PRE_BM, 
						GraphicsPrimitive.TYPE_INT_BGR };

    public final GraphicsPrimitive[] getPrimitives() {
        GraphicsPrimitive[] primitives = {
            new  GraphicsPrimitiveProxy(this, "ColorFillAlphaToXBGR",
					ST_INT_BGR,
					ColorPaint.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintXRGB_ONTO_XBGR",
					ST_INT_RGB__INT_BGR,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_ONTO_XBGR",
					ST_INT_ARGB__INT_BGR,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_ONTO_XBGR",
					ST_INT_ARGB_PRE__INT_BGR,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BITMASK_ONTO_XBGR",
					ST_INT_ARGB_BM__INT_BGR,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BITMASK_ONTO_XBGR",
					ST_INT_ARGB_PRE_BM__INT_BGR,
					BinaryComposite.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BG_CLR_ONTO_XBGR",
					ST_INT_ARGB__INT_BGR,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BG_CLR_ONTO_XBGR",
					ST_INT_ARGB_PRE__INT_BGR,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR",
					ST_INT_ARGB_BM__INT_BGR,
					BinaryCompositeBg.getMethodSignature()),
            new  GraphicsPrimitiveProxy(this, "paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR",
					ST_INT_ARGB_PRE_BM__INT_BGR,
					BinaryCompositeBg.getMethodSignature())
        };
        return primitives;
    }

    static native void ColorFillAlphaToXBGR(
        ImageData dstImage,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        int nFGColor
    );

    static native void paintXRGB_ONTO_XBGR(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_ONTO_XBGR(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_WITH_BITMASK_ONTO_XBGR(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int nXORColor
    );

    static native void paintARGB_WITH_BG_CLR_ONTO_XBGR(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );

    static native void paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR(
        ImageData srcImage,
        ImageData dstImage,
        int width, int height,

        int nRule,

        byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
        float fExtraAlpha,
        boolean bSourcePremultiplied,

        int bgColor, int nXORColor
    );
}
class ColorFillAlphaToXBGR extends ColorPaint {
    ColorFillAlphaToXBGR() {
	super(XBGRLoops.ST_INT_BGR);
    }
    public void ColorPaint(
	ImageData dstImage,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color fgColor
    ) {
	XBGRLoops.ColorFillAlphaToXBGR(
	    dstImage, nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, fgColor.getRGB());
    }
}
class paintXRGB_ONTO_XBGR extends BinaryComposite {
    paintXRGB_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_RGB__INT_BGR);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	XBGRLoops.paintXRGB_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_ONTO_XBGR extends BinaryComposite {
    paintARGB_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB__INT_BGR);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	XBGRLoops.paintARGB_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_PRE_ONTO_XBGR extends BinaryComposite {
    paintARGB_PRE_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB_PRE__INT_BGR);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	XBGRLoops.paintARGB_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintARGB_WITH_BITMASK_ONTO_XBGR extends BinaryComposite {
    paintARGB_WITH_BITMASK_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB_BM__INT_BGR);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	XBGRLoops.paintARGB_WITH_BITMASK_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BITMASK_ONTO_XBGR extends BinaryComposite {
    paintARGB_PRE_WITH_BITMASK_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB_PRE_BM__INT_BGR);
    }
    public void BinaryComposite(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color XORColor
    ) {
	XBGRLoops.paintARGB_WITH_BITMASK_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, XORColor.getRGB());
    }
}
class paintARGB_WITH_BG_CLR_ONTO_XBGR extends BinaryCompositeBg {
    paintARGB_WITH_BG_CLR_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB__INT_BGR);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	XBGRLoops.paintARGB_WITH_BG_CLR_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BG_CLR_ONTO_XBGR extends BinaryCompositeBg {
    paintARGB_PRE_WITH_BG_CLR_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB_PRE__INT_BGR);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	XBGRLoops.paintARGB_WITH_BG_CLR_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR extends BinaryCompositeBg {
    paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB_BM__INT_BGR);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	XBGRLoops.paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, false, bgColor.getRGB(), XORColor.getRGB());
    }
}
class paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR extends BinaryCompositeBg {
    paintARGB_PRE_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR() {
	super(XBGRLoops.ST_INT_ARGB_PRE_BM__INT_BGR);
    }
    public void BinaryCompositeBg(
	ImageData srcImage,
	ImageData dstImage,
	int width, int height,

	int nRule,

	byte [] bAlphaArray, int nAlphaArrayOffset, int nAlphaArraySize,
	float fExtraAlpha,
	Color bgColor, Color XORColor
    ) {
	XBGRLoops.paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_XBGR(
	    srcImage, dstImage, width, height,
	    nRule,
	    bAlphaArray, nAlphaArrayOffset, nAlphaArraySize,
	    fExtraAlpha, true, bgColor.getRGB(), XORColor.getRGB());
    }
}
