/*
 * @(#)AlphaColorPipe.java	1.13 98/07/06
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Shape;
import sun.java2d.SunGraphics2D;
import sun.java2d.OutputManager;
import sun.java2d.loops.RasterOutputManager;
import sun.java2d.loops.XORComposite;

/**
 * This class implements a CompositePipe that renders path alpha tiles
 * into a destination that supports direct alpha compositing of a solid
 * color, according to one of the rules in the AlphaComposite class.
 */
public class AlphaColorPipe implements CompositePipe {

    private class ColorPipeContext {
        SunGraphics2D dst2D;
        Object        loopState;

        ColorPipeContext(SunGraphics2D dst2D) {
            this.dst2D = dst2D;
            this.loopState = null;
        }
    }

    public AlphaColorPipe() {
    }

    public Object startSequence(SunGraphics2D sg, Shape s, Rectangle dev,
				int[] abox) {
	return new ColorPipeContext(sg);
    }

    public boolean needTile(Object context, int x, int y, int w, int h) {
	return true;
    }

    public void renderPathTile(Object context,
			       byte[] atile, int offset, int tilesize,
			       int x, int y, int w, int h) {
        ColorPipeContext cpc = (ColorPipeContext) context;
	SunGraphics2D sg = cpc.dst2D;
        float extraAlpha = 1.0f;
        int rule = 0;
        Color color;
        if (sg.composite instanceof AlphaComposite) {
            AlphaComposite ac = (AlphaComposite) sg.composite;
            rule = ac.getRule();
            extraAlpha = ac.getAlpha();
            color = (Color)sg.paint;
        } else if (sg.composite instanceof XORComposite) {
            XORComposite xc = (XORComposite)sg.composite;
            rule = 0;
            color = new Color(0xff000000 |
                              ((xc.getXorColor()).getRGB() ^
                               ((Color) sg.paint).getRGB()), true);
        } else {
            throw new InternalError("unrecognized Composite object");
        }

        RasterOutputManager rom = (RasterOutputManager)sg.getOutputManager();
        cpc.loopState = rom.compositeColorLoop(
            sg, rule, extraAlpha,
            atile, offset, tilesize, x, y, w, h,
            color, cpc.loopState );	
    }

    public void skipTile(Object context, int x, int y) {
	return;
    }

    public void endSequence(Object context) {
	return;
    }
}
