/*
 * @(#)AlphaPaintPipe.java	1.21 98/09/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.AlphaComposite;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.PaintContext;
import java.awt.image.ColorModel;
import java.awt.Color;
import java.awt.image.Raster;
import java.awt.geom.AffineTransform;
import sun.java2d.SunGraphics2D;
import sun.java2d.OutputManager;
import sun.java2d.loops.RasterOutputManager;
import sun.java2d.loops.XORComposite;

/**
 * This class implements a CompositePipe that renders path alpha tiles
 * into a destination that supports direct alpha compositing of a paint
 * data tile, according to one of the rules in the AlphaComposite class.
 */
public class AlphaPaintPipe implements CompositePipe {
    class TileContext {
	SunGraphics2D sunG2D;
	PaintContext paintCtxt;
        Object pipeState;

	public TileContext(SunGraphics2D sg, PaintContext pc) {
	    sunG2D = sg;
	    paintCtxt = pc;
            pipeState = null;
	}
    }

    public Object startSequence(SunGraphics2D sg, Shape s, Rectangle devR,
				int[] abox) {
        PaintContext paintContext =
	    sg.paint.createContext(sg.getDeviceColorModel(),
				   devR,
				   s.getBounds2D(),
                                   sg.cloneTransform(),
                                   sg.getRenderingHints());
	return new TileContext(sg, paintContext);
    }

    public boolean needTile(Object context, int x, int y, int w, int h) {
	return true;
    }

    public void renderPathTile(Object ctx,
			       byte[] atile, int offset, int tilesize,
			       int x, int y, int w, int h) {
	TileContext context = (TileContext) ctx;
	PaintContext paintCtxt = context.paintCtxt;
	SunGraphics2D sg = (SunGraphics2D) context.sunG2D;

	Raster srcRaster = paintCtxt.getRaster(x, y, w, h);
	ColorModel paintModel = paintCtxt.getColorModel();
	RasterOutputManager rom =(RasterOutputManager)sg.getOutputManager();

	boolean srcAlpPremult = paintModel.isAlphaPremultiplied();
	context.pipeState = rom.compositePaintLoop(sg, atile, offset,
                                                   tilesize, x, y, w, h,
                                                   paintModel, srcRaster,
                                                   context.pipeState);
    }

    public void skipTile(Object context, int x, int y) {
	return;
    }

    public void endSequence(Object ctx) {
	TileContext context = (TileContext) ctx;
        if (context.paintCtxt != null) {
            context.paintCtxt.dispose();
        }
    }
}
