/*
 * @(#)DrawToFillConverter.java	1.9 98/09/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.BasicStroke;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.PathIterator;
import sun.java2d.SunGraphics2D;
import sun.dc.path.FastPathProducer;
import sun.dc.path.PathConsumer;
import sun.dc.path.PathException;

/**
 * This interface defines the set of calls that pipeline objects
 * can use to pass on responsibility for drawing generic Shape
 * objects.
 */
public class DrawToFillConverter
    extends DuctusRenderer
    implements ShapeDrawPipe
{
    SimpleRenderPipe outpipe;
    ShapeDrawPipe fillpipe;
    boolean doTx;

    public DrawToFillConverter(SimpleRenderPipe pipe,
			       ShapeDrawPipe fillpipe,
			       boolean doTransform) {
	outpipe = pipe;
	this.fillpipe = fillpipe;
	doTx = doTransform;
    }

    public void draw(SunGraphics2D sg, Shape s) {
	ShapeSpanIterator sr = new ShapeSpanIterator();
	try {
	    draw2(sg, sr, s);
	} finally {
	    sr.dispose();
	}
    }

    public void draw2(SunGraphics2D sg, ShapeSpanIterator sr, Shape s) {
	sr.setOutputArea(sg.getCompBounds());

	if (sg.stroke instanceof BasicStroke) {
	    sr.setRule(PathIterator.WIND_NON_ZERO);

	    BasicStroke bs = (BasicStroke) sg.stroke;
	    AffineTransform transform =
		(sg.transformState == sg.TRANSFORM_GENERIC
		 ? sg.transform : null);

	    PathConsumer consumer =
		DuctusRenderer.createStroker(sr, bs, transform);

	    transform =
		((doTx && sg.transformState != sg.TRANSFORM_ISIDENT)
		 ? sg.transform : null);
	    PathIterator pi = s.getPathIterator(transform);

	    try {
		DuctusRenderer.feedConsumer(pi, consumer);
	    } catch (PathException e) {
		throw new InternalError("Unable to Stroke shape ("+
					e.getMessage()+")");
	    }
	} else {
	    s = sg.stroke.createStrokedShape(s);
	    AffineTransform at =
		((doTx && sg.transformState != sg.TRANSFORM_ISIDENT)
		 ? sg.transform : null);
	    sr.appendPath(s.getPathIterator(at));
	}
	outpipe.fillSpans(sg, sr);
    }

    public void fill(SunGraphics2D sg, Shape s) {
	fillpipe.fill(sg, s);
    }
}
