/*
 * @(#)DuctusClipRenderer.java	1.6 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.PathIterator;
import sun.java2d.SunGraphics2D;
import sun.dc.pr.Rasterizer;
import sun.dc.pr.PRException;

/**
 * This class uses the Ductus Rasterizer to modify the alpha tiles
 * created during Shape rendering based upon a non-rectangular clipping
 * path.
 * This class might also be used to apply soft-edged clipping to
 * rectangular clip areas, depending on the rendering hints.
 */
public class DuctusClipRenderer extends DuctusRenderer
				implements CompositePipe
{
    CompositePipe outpipe;

    public DuctusClipRenderer(CompositePipe pipe) {
	outpipe = pipe;
    }

    class DCRcontext {
	Rasterizer rast;
	Object outcontext;
	int state;
	byte[] alpha = new byte[Rasterizer.TILE_SIZE * Rasterizer.TILE_SIZE];

	public DCRcontext(Rasterizer r, Object outctx) {
	    rast = r;
	    outcontext = outctx;
	}
    }

    public Object startSequence(SunGraphics2D sg, Shape s, Rectangle devR,
				int[] abox) {
	PathIterator cpi = sg.usrClip.getPathIterator(null);
	Rasterizer r = createShapeRasterizer(cpi, null, null);
	try {
	    r.setOutputArea(abox[0], abox[1],
			    abox[2] - abox[0], abox[3] - abox[1]);
	} catch (PRException e) {
	    dropRasterizer(r);
	    r = null;
	}

	return new DCRcontext(r, outpipe.startSequence(sg, s, devR, abox));
    }

    public boolean needTile(Object ctx, int x, int y, int w, int h) {
	DCRcontext context = (DCRcontext) ctx;
	Rasterizer r = context.rast;
	context.state = ((r == null)
			 ? Rasterizer.TILE_IS_ALL_0
			 : context.rast.getTileState());
	return (outpipe.needTile(context.outcontext, x, y, w, h) &&
		context.state != Rasterizer.TILE_IS_ALL_0);
    }

    protected native void mergeAlphas(byte [] srcA, byte[] dstA,
				      int srcOff, int dstOff,
				      int tsize, int w, int h,
				      int bitDepth);

    public void renderPathTile(Object ctx,
			       byte[] atile, int offset, int tsize,
			       int x, int y, int w, int h) {
	byte [] ctile;
	DCRcontext context = (DCRcontext) ctx;
	Rasterizer r = context.rast;
	if (context.state == Rasterizer.TILE_IS_ALL_1) {
	    r.nextTile();
	    ctile = atile;
	} else {
	    ctile = context.alpha;
	    try {
		getAlpha(r, ctile, 1, tsize, 0);
		if (atile != null) {
		    int bitDepth = 8; // Assume 8 bit alpha for now.
		    mergeAlphas(atile, ctile, offset, 0, tsize, w, h, bitDepth);
		}
		offset = 0;
	    } catch (PRException e) {
		dropRasterizer(r);
		context.rast = null;
	    }
	}

	outpipe.renderPathTile(context.outcontext,
			       ctile, offset, tsize,
			       x, y, w, h);
    }

    public void skipTile(Object ctx, int x, int y) {
	DCRcontext context = (DCRcontext) ctx;
	context.rast.nextTile();
	outpipe.skipTile(context.outcontext, x, y);
    }

    public void endSequence(Object ctx) {
	DCRcontext context = (DCRcontext) ctx;
	dropRasterizer(context.rast);
	outpipe.endSequence(context.outcontext);
    }
}
