/*
 * @(#)GeneralCompositePipe.java	1.20 98/09/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.AlphaComposite;
import java.awt.CompositeContext;
import java.awt.PaintContext;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.image.ColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import sun.java2d.SunGraphics2D;
import sun.java2d.OutputManager;
import sun.java2d.loops.RasterOutputManager;

public class GeneralCompositePipe implements CompositePipe {
    class TileContext {
	SunGraphics2D sunG2D;
	PaintContext paintCtxt;
	CompositeContext compCtxt;
	ColorModel compModel;
        Object pipeState;

	public TileContext(SunGraphics2D sg, PaintContext pCtx,
			   CompositeContext cCtx, ColorModel cModel) {
	    sunG2D = sg;
	    paintCtxt = pCtx;
	    compCtxt = cCtx;
	    compModel = cModel;
	}
    }

    public Object startSequence(SunGraphics2D sg, Shape s, Rectangle devR,
				int[] abox) {
	RenderingHints hints = sg.getRenderingHints();
	ColorModel model = sg.getDeviceColorModel();
        PaintContext paintContext =
	    sg.paint.createContext(model, devR, s.getBounds2D(),
                                   sg.cloneTransform(),
                                   hints);
        CompositeContext compositeContext =
	    sg.composite.createContext(paintContext.getColorModel(), model,
                                       hints);
	return new TileContext(sg, paintContext, compositeContext, model);
    }

    public boolean needTile(Object ctx, int x, int y, int w, int h) {
	return true;
    }

    /**
    * GeneralCompositePipe.renderPathTile works with custom composite operator
    * provided by an application
    */
    public void renderPathTile(Object ctx,
			       byte[] atile, int offset, int tilesize,
			       int x, int y, int w, int h) {
	TileContext context = (TileContext) ctx;
	PaintContext paintCtxt = context.paintCtxt;
	CompositeContext compCtxt = context.compCtxt;
	SunGraphics2D sg = context.sunG2D;

        Raster srcRaster = paintCtxt.getRaster(x, y, w, h);
	ColorModel paintModel = paintCtxt.getColorModel();
	RasterOutputManager rom = (RasterOutputManager)sg.getOutputManager();
        context.pipeState =
            rom.compositePaintLoop(sg, atile, offset, tilesize, x, y, w, h,
                                   paintModel, srcRaster,
                                   context.pipeState);

    }

    public void skipTile(Object ctx, int x, int y) {
	return;
    }

    public void endSequence(Object ctx) {
	TileContext context = (TileContext) ctx;
        if (context.paintCtxt != null) {
            context.paintCtxt.dispose();
        }
        if (context.compCtxt != null) {
            context.compCtxt.dispose();
        }
    }

}
