/*
 * @(#)RegionIterator.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.Rectangle;

/**
 * This class defines the API for iterating through the bands
 * of a region object.
 */
public class RegionIterator {
    Region region;
    int curIndex;
    int numXbands;

    RegionIterator(Region r) {
	region = r;
    }

    /**
     * Returns a new RegionIterator object representing the same
     * iteration state as this object to allow multiple iteration
     * branches from the current position.
     */
    public RegionIterator createCopy() {
	RegionIterator r = new RegionIterator(region);
	r.curIndex = this.curIndex;
	r.numXbands = this.numXbands;
	return r;
    }

    /**
     * Copies the iteration state from this RegionIterator object
     * into another RegionIterator object to allow multiple iteration
     * branches from the current position.
     */
    public void copyStateFrom(RegionIterator ri) {
	if (this.region != ri.region) {
	    throw new InternalError("region mismatch");
	}
	this.curIndex = ri.curIndex;
	this.numXbands = ri.numXbands;
    }

    /**
     * Moves the iteration state to the beginning of the next
     * Y range in the region returning true if one is found
     * and recording the low and high Y coordinates of the
     * range in the array at locations 0 and 1 respectively.
     */
    public boolean nextYRange(int range[]) {
	curIndex += numXbands * 2;
	numXbands = 0;
	if (curIndex >= region.endIndex) {
	    return false;
	}
	range[0] = region.bands[curIndex++];
	range[1] = region.bands[curIndex++];
	numXbands = region.bands[curIndex++];
	return true;
    }

    /**
     * Moves the iteration state to the beginning of the next
     * X band in the current Y range returning true if one is
     * found and recording the low and high X coordinates of
     * the range in the array at locations 2 and 3 respectively.
     */
    public boolean nextXBand(int range[]) {
	if (numXbands <= 0) {
	    return false;
	}
	numXbands--;
	range[2] = region.bands[curIndex++];
	range[3] = region.bands[curIndex++];
	return true;
    }
}
