/*
 * @(#)ShapeSpanIterator.java	1.4 98/07/31
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.geom.PathIterator;
import java.awt.Rectangle;
import sun.dc.path.PathConsumer;
import sun.dc.path.PathException;
import sun.dc.path.FastPathProducer;

/**
 * This class can iterate individual span elements generated by scan
 * converting a Shape.
 * This particular implementation flattens the incoming path and then
 * performs simple polygon tracing to calculate the spans.
 */
public class ShapeSpanIterator implements SpanIterator, PathConsumer {
    long pData;

    static {
	initIDs();
    }

    public static native void initIDs();

    public ShapeSpanIterator() {
    }

    /*
     * Appends the geometry and winding rule from the indicated
     * path iterator.
     */
    public void appendPath(PathIterator pi) {
	float coords[] = new float[6];

	setRule(pi.getWindingRule());
	while (!pi.isDone()) {
	    addSegment(pi.currentSegment(coords), coords);
	    pi.next();
	}
	endPath();
    }

    /*
     * Sets the rectangle of interest for storing and returning
     * span segments.
     */
    public native void setOutputArea(int x, int y, int w, int h);

    /*
     * Sets the rectangle of interest for storing and returning
     * span segments.
     */
    public void setOutputArea(Rectangle r) {
	setOutputArea(r.x, r.y, r.width, r.height);
    }

    /*
     * Sets the winding rule in the native data structures.
     */
    public native void setRule(int rule);

    /*
     * Adds a single PathIterator segment to the internal list of
     * path element structures.
     */
    public native void addSegment(int type, float coords[]);

    /*
     * Gets the bbox of the available path segments, clipped to the
     * OutputArea.
     */
    public native void getPathBox(int pathbox[]);

    /*
     * Fetches the next span that needs to be operated on.
     * If the return value is false then there are no more spans.
     */
    public native boolean nextSpan(int spanbox[]);

    /**
     * This method tells the iterator that it may skip all spans
     * whose Y range is completely above the indicated Y coordinate.
     */
    public native void skipDownTo(int y);

    /**
     * This method returns a native pointer to a function block that
     * can be used by a native method to perform the same iteration
     * cycle that the above methods provide while avoiding upcalls to
     * the Java object.
     * The definition of the structure whose pointer is returned by
     * this method is defined in:
     * <pre>
     *     src/share/native/sun/java2d/pipe/SpanIterator.h
     * </pre>
     */
    public native long getNativeIterator();

    /*
     * Cleans out all internal data structures.
     */
    public native void dispose();

    protected void finalize() {
	dispose();
    }

    public void beginPath() {
    }

    public native void beginSubpath(float x0, float y0);

    public native void appendLine(float x1, float y1);

    public native void appendQuadratic(float xm, float ym,
				       float x1, float y1);

    public native void appendCubic(float xm, float ym,
				   float xn, float yn,
				   float x1, float y1);

    public void closedSubpath() {
    }

    public native void endPath();

    public void	useProxy(FastPathProducer proxy)
	throws PathException
    {
	proxy.sendTo(this);
    }

    public native long getCPathConsumer();
}
