/*
 * @(#)ShapeToPolyConverter.java	1.5 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.Polygon;
import java.awt.Shape;
import java.awt.geom.PathIterator;
import sun.java2d.SunGraphics2D;

/**
 * This class converts calls to the basic pixel rendering
 * methods into calls to line drawing or rect or polygon filling.
 */
public class ShapeToPolyConverter implements ShapeDrawPipe {
    SimpleRenderPipe outpipe;
    ShapeDrawPipe    fillpipe;
    boolean doTx;

    public ShapeToPolyConverter(SimpleRenderPipe pipe, boolean doTransform) {
	outpipe = pipe;
	doTx = doTransform;
    }

    public ShapeToPolyConverter(SimpleRenderPipe pipe,
				ShapeDrawPipe fillpipe,
				boolean doTransform) {
	outpipe = pipe;
	this.fillpipe = fillpipe;
	doTx = doTransform;
    }

    public void draw(SunGraphics2D sg, Shape s) {
	doDraw(sg, s, false);
    }

    public void fill(SunGraphics2D sg, Shape s) {
	if (fillpipe != null) {
	    fillpipe.fill(sg, s);
	} else {
	    doDraw(sg, s, true);
	}
    }

    private int[] sendPoly(SunGraphics2D sg, Polygon pg,
			   boolean doFill, boolean evenodd,
			   int nPoints[], boolean done) {
	int numpolys = nPoints[0];
	int prevpts = numpolys == 0 ? 0 : nPoints[numpolys];
	int numpts = pg.npoints - prevpts;
	if (numpts < 2) {
	    pg.npoints = prevpts;
	} else {
	    numpolys++;
	    if (numpolys >= nPoints.length) {
		int temp[] = new int[nPoints.length + 10];
		System.arraycopy(nPoints, 0, temp, 0,
				 nPoints.length);
		nPoints = temp;
	    }
	    nPoints[numpolys] = pg.npoints;
	    nPoints[0] = numpolys;
	}
	if (done) {
	    prevpts = 0;
	    for (int i = 1; i <= numpolys; i++) {
		numpts = nPoints[i];
		nPoints[i - 1] = numpts - prevpts;
		prevpts = numpts;
	    }
	    if (doFill) {
		outpipe.fillPolys(sg, pg.xpoints, pg.ypoints, nPoints,
				  numpolys, evenodd);
	    } else {
		outpipe.drawPolys(sg, pg.xpoints, pg.ypoints, nPoints,
				  numpolys, false);
	    }
	}
	return nPoints;
    }

    private void doDraw(SunGraphics2D sg, Shape s, boolean doFill) {
	PathIterator pi = s.getPathIterator(doTx ? sg.transform : null, 1.0);
	boolean evenodd = (pi.getWindingRule() == PathIterator.WIND_EVEN_ODD);
	Polygon pg = new Polygon();
	float point[] = new float[6];
	int nPoints[] = new int[10];
	int lastmove = 0;

	while (!pi.isDone()) {
	    switch (pi.currentSegment(point)) {
	    case PathIterator.SEG_MOVETO:
		nPoints = sendPoly(sg, pg, doFill, evenodd, nPoints, false);
		lastmove = pg.npoints;
		// NO BREAK
	    case PathIterator.SEG_LINETO:
		// Given that most integer-based graphics systems
		// use the convention that the coordinate is at the
		// center of the pixel, and given that our incoming
		// coordinate system puts coordinates at the edges
		// of pixels, there is an implicit translate(0.5)
		// involved in converting to integers so rounding
		// is equivalent to flooring
		pg.addPoint((int) Math.floor(point[0]),
			    (int) Math.floor(point[1]));
		break;
	    case PathIterator.SEG_QUADTO:
		// Should not happen
		break;
	    case PathIterator.SEG_CUBICTO:
		// Should not happen
		break;
	    case PathIterator.SEG_CLOSE:
		if (pg.npoints > 0) {
		    pg.addPoint(pg.xpoints[lastmove], pg.ypoints[lastmove]);
		}
		break;
	    }
	    pi.next();
	}

	nPoints = sendPoly(sg, pg, doFill, evenodd, nPoints, true);
    }
}
