/*
 * @(#)ValidatePipe.java	1.6 98/08/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.java2d.pipe;

import java.awt.Shape;
import sun.java2d.SunGraphics2D;
import java.awt.font.GlyphVector;

/**
 * This class is used to force a revalidation of the pipelines of
 * the indicated SunGraphics2D object before a draw command.
 * After calling SunGraphics2D.validatePipe() to force the pipeline
 * to be revalidated, this object redispatches the draw command to
 * the new valid pipe object.
 */
public class ValidatePipe
    implements PixelDrawPipe, PixelFillPipe, ShapeDrawPipe, TextPipe
{
    public void drawLine(SunGraphics2D sg,
			 int x1, int y1, int x2, int y2) {
	sg.validatePipe();
	sg.drawpipe.drawLine(sg, x1, y1, x2, y2);
    }

    public void drawRect(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	sg.validatePipe();
	sg.drawpipe.drawRect(sg, x, y, width, height);
    }

    public void fillRect(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	sg.validatePipe();
	sg.fillpipe.fillRect(sg, x, y, width, height);
    }

    public void drawRoundRect(SunGraphics2D sg,
			      int x, int y, int width, int height,
			      int arcWidth, int arcHeight) {
	sg.validatePipe();
	sg.drawpipe.drawRoundRect(sg, x, y, width, height,
				   arcWidth, arcHeight);
    }

    public void fillRoundRect(SunGraphics2D sg,
			      int x, int y, int width, int height,
			      int arcWidth, int arcHeight) {
	sg.validatePipe();
	sg.fillpipe.fillRoundRect(sg, x, y, width, height,
				   arcWidth, arcHeight);
    }

    public void drawOval(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	sg.validatePipe();
	sg.drawpipe.drawOval(sg, x, y, width, height);
    }

    public void fillOval(SunGraphics2D sg,
			 int x, int y, int width, int height) {
	sg.validatePipe();
	sg.fillpipe.fillOval(sg, x, y, width, height);
    }

    public void drawArc(SunGraphics2D sg,
			int x, int y, int width, int height,
			int startAngle, int arcAngle) {
	sg.validatePipe();
	sg.drawpipe.drawArc(sg, x, y, width, height, startAngle, arcAngle);
    }

    public void fillArc(SunGraphics2D sg,
			int x, int y, int width, int height,
			int startAngle, int arcAngle) {
	sg.validatePipe();
	sg.fillpipe.fillArc(sg, x, y, width, height, startAngle, arcAngle);
    }

    public void drawPolyline(SunGraphics2D sg,
			     int xPoints[], int yPoints[],
			     int nPoints) {
	sg.validatePipe();
	sg.drawpipe.drawPolyline(sg, xPoints, yPoints, nPoints);
    }

    public void drawPolygon(SunGraphics2D sg,
			    int xPoints[], int yPoints[],
			    int nPoints) {
	sg.validatePipe();
	sg.drawpipe.drawPolygon(sg, xPoints, yPoints, nPoints);
    }

    public void fillPolygon(SunGraphics2D sg,
			    int xPoints[], int yPoints[],
			    int nPoints) {
	sg.validatePipe();
	sg.fillpipe.fillPolygon(sg, xPoints, yPoints, nPoints);
    }

    public void draw(SunGraphics2D sg, Shape s) {
	sg.validatePipe();
	sg.shapepipe.draw(sg, s);
    }

    public void fill(SunGraphics2D sg, Shape s) {
	sg.validatePipe();
	sg.shapepipe.fill(sg, s);
    }
    public Object drawString(SunGraphics2D sg, String s,
                            float x, float y, Object cs) {
	sg.validatePipe();
	return sg.textpipe.drawString(sg, s, x, y, cs);
    }
    public Object drawGlyphVector(SunGraphics2D sg, GlyphVector g,
                                float x, float y, Object cs) {
	sg.validatePipe();
	return sg.textpipe.drawGlyphVector(sg, g, x, y, cs);
    }
    public Object drawChars(SunGraphics2D sg,
                                char data[], int offset, int length,
                                int x, int y, Object cs) {
	sg.validatePipe();
	return sg.textpipe.drawChars(sg, data, offset, length, x, y, cs);
    }
    public Object drawBytes(SunGraphics2D sg,
                                byte data[], int offset, int length,
                                int x, int y, Object cs) {
	sg.validatePipe();
	return sg.textpipe.drawBytes(sg, data, offset, length, x, y, cs);
    }
}
