/*
 * @(#)JdbcOdbc.c	1.25 98/11/06
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      jdbcodbc.c
//
// Description:	Implements the ODBC API routines for native java calls.
//              It is the intent to do as little processing as possible in
//              this implementation.  All of the work is being done in Java,
//              all that is being done here is to simply call the ODBC
//              API and return the results.  All buffers are provided from
//              Java, so we need not allocate any memory here (or worry about
//              freeing it)
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1997 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

#ifdef macintosh
#define sys/types.h Types.h
#endif

#include <sys/types.h>

#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#endif

#ifdef UNIX
#include <sys/stat.h>
#include <sys/uio.h>
#include <unistd.h>

// Various defines/typedefs for UNIX

typedef	int HWND;	
typedef int BOOL;
typedef unsigned int UINT;
typedef void * HANDLE;
typedef char * PU8;
typedef const char * PCU8;
typedef char * LPSTR;
#define WORD  SWORD
#define DWORD SDWORD
#define CALLBACK
#define EXPORT
#define	FAR
#define PASCAL
#define	TRUE	1
#define	FALSE	0

#endif

//#include "javaString.h"
#include <fcntl.h>

// This is an ODBC 2.1 application

#define ODBCVER 0x210

#include "sqlext.h"
#include "odbcinst.h"
#include <jni.h>
#include "stdio.h"

//----------------------------------------------------------------------------
// bufferToInt
// Converts the given buffer to a native int (long)
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bufferToInt
  (JNIEnv *env, jobject *callingObject, jbyteArray buf)


{
// Get the data structure portion of the buffer
	UCHAR*  pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, buf, 0); //pBuf=*Buf
	long	n = 0;

	memcpy (&n, pBuf, sizeof (n));
	(*env)->ReleaseByteArrayElements(env,buf,pBuf,0);
	return n;
}


//----------------------------------------------------------------------------
// allocConnect
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_allocConnect
           (JNIEnv *env, jobject *callingObject, 
		   jint hEnv, jbyteArray errorCode)


{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	HDBC    hDbc = SQL_NULL_HDBC;
	RETCODE	rc;

// Allocate a new connection handle
	
	rc = SQLAllocConnect (
		(HENV) hEnv,		// Environment handle
		&hDbc);			// Pointer to connection handle

// Set the error code
	
	errCode[0] = (char) rc;
        (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
// Return the connection handle
        return (long) hDbc;
}


//----------------------------------------------------------------------------
// allocEnv
//----------------------------------------------------------------------------

JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_allocEnv
     (JNIEnv *env, jobject *callingObject, jbyteArray errorCode)

{
// Get the data structure portion of the error code
        UCHAR*   errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
        HENV    hEnv = SQL_NULL_HENV;
        RETCODE rc;

// Allocate a new environment handle
    rc = SQLAllocEnv (
            &hEnv);                 // Pointer to environment handle

// Set the error code
	
    errCode[0] = (char) rc;


	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
// Return the environment handle
        return (long) hEnv;
}

//----------------------------------------------------------------------------
// allocStmt
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_allocStmt
     (JNIEnv *env, jobject *callingObject, jint hDbc,
	 jbyteArray errorCode)


{
 //Get the data structure portion of the error code
        UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
        HSTMT   hStmt = SQL_NULL_HSTMT;
        RETCODE rc;

// Allocate a new statement handle
	
	rc = SQLAllocStmt (
		(HDBC) hDbc,		// Connection handle
		&hStmt);		// Pointer to statement handle

// Set the error code
	errCode[0] = (char) rc;
        (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
// Return the statement handle
        return (long) hStmt;
}

//----------------------------------------------------------------------------
// bindColDefault
//----------------------------------------------------------------------------

JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindColDefault
     (JNIEnv *env, jobject *callingObject, jint hStmt, jint icol,
	 jint prgb, jint ppcb, jbyteArray rgb, jbyteArray pcb, jbyteArray errorCode)
{
        UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
        UCHAR*  rgbValue = (UCHAR*) (*env)->GetByteArrayElements(env, rgb, 0);
        SDWORD  cbValueMax = (SDWORD) (*env)->GetArrayLength(env, rgb);
        UCHAR*  pcbValue = (UCHAR*) (*env)->GetByteArrayElements(env, pcb, 0);
    
		
		RETCODE	rc;

// Bind the column
		
	rc = SQLBindCol (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) icol,		// Column number of result data
		SQL_C_DEFAULT,		// C data type of the result data
		rgbValue,		// Pointer to storage for the data
		cbValueMax,		// Maximum length of rgbValue buffer
		(SDWORD*) pcbValue);	// Number of bytes available
	
// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, rgb, rgbValue, 0);
	(*env)->ReleaseByteArrayElements(env, pcb, pcbValue, 0);
}

//----------------------------------------------------------------------------
// bindInParameterAtExec
// Used for binding
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterAtExec
      (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar,
	  jint SQLtype, jint len, jbyteArray dataBuf, jbyteArray lenBuf, 
          jbyteArray errorCode, jintArray buffers)


{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;
	SWORD   Ctype = SQL_C_CHAR;
	SDWORD	atExec = SQL_LEN_DATA_AT_EXEC (len);
	UCHAR*	pBuf = NULL;
	UCHAR*	pLen = NULL;
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	jobject gLenBuf = (*env)->NewGlobalRef(env, lenBuf);

        int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

// Get the buffer for the parameter's data.  In this case, we'll set it
// to the parameter number.  This number will be returned by SQLParamData.
// It is assumed that this buffer is large enough to fit a 4-byte integer.

	if (dataBuf != NULL) {

		pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);
		memcpy (pBuf, &ipar, sizeof (ipar));
                pBuffers[0]=(int)pBuf;  //Store the pointers
                pBuffers[1]=(int)gDataBuf;
	}
	
// Get the buffer used for the parameter's length.  It is assumed that
// this buffer is large enough to fit a 4-byte integer
	
	if (lenBuf != NULL) {

		pLen = (UCHAR*) (*env)->GetByteArrayElements(env, gLenBuf, 0);
		memcpy (pLen, &atExec, sizeof (atExec));
		//Store the pointer
                pBuffers[2]=(int)pLen;
                pBuffers[3]=(int)gLenBuf;
	}
                (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);
	
// The SQL Type was given, make the C-Type match
// Only SQL_CHAR and SQL_BINARY are currently supported.
	
	if ((SQLtype == SQL_BINARY) ||
	    (SQLtype == SQL_VARBINARY) ||
	    (SQLtype == SQL_LONGVARBINARY)) {
		Ctype = SQL_C_BINARY;
	}

// Bind the parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		Ctype,			// The C data type
		(SWORD) SQLtype,	// The SQL data type
		(UDWORD) len,		// Precision
		0,			// Scale
		pBuf,			// Pointer to parameter's data
		sizeof (ipar),		// Maximum length of parameter buffer
		(SDWORD*) pLen);	// Pointer to parameter's length

// Set the error code
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// bindInParameterBinary
//----------------------------------------------------------------------------

JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterBinary
     (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar, 
     jint SQLtype, jbyteArray value, jint cbColDef, jbyteArray dataBuf, 
		jbyteArray lenBuf, jbyteArray errorCode, jintArray buffers)

{
// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*	errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	jobject	gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	jobject	gValue = (*env)->NewGlobalRef(env, value);
	jobject	gLenBuf = (*env)->NewGlobalRef(env, lenBuf);
	UCHAR*	rgbValue = (UCHAR*) (*env)->GetByteArrayElements(env, gValue, 0);
	SDWORD	lValue = (*env)->GetArrayLength(env, value);
	UCHAR*	pBuf = NULL;
	SDWORD	lBuf = 0;
	UCHAR*	pLen = NULL;

        int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

                 pBuffers[4]=(int)rgbValue;  //Store the pointers
                 pBuffers[5]=(int)gValue;

// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.

	if (dataBuf != NULL) {
				pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);

				pBuffers[0]=(int)pBuf;  //Store the pointers
				pBuffers[1]=(int)gDataBuf;

                lBuf = (*env)->GetArrayLength(env, dataBuf);
  
		// Sanity check the length
		if (lBuf > lValue) {
			lBuf = lValue;
		}

		// Make a copy of the data
		memcpy (pBuf, rgbValue, lBuf);
	}
	
// Get the buffer used for the parameter's length.  It is assumed that
// this buffer is large enough to fit a 4-byte integer
	
	if (lenBuf != NULL) {
		pLen = (UCHAR*) (*env)->GetByteArrayElements(env, gLenBuf, 0);
		memcpy (pLen, &lBuf, sizeof(lBuf));
		//Store the pointer
                pBuffers[2]=(int)pLen;
                pBuffers[3]=(int)gLenBuf;
	}
                (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);
	
// Bind the binary parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_BINARY,		// The C data type
		(SWORD) SQLtype,	// The SQL data type
		(UDWORD) cbColDef,	// Precision
		0,			// Scale
		pBuf,			// Pointer to parameter's data
		lBuf,			// Maximum length of parameter buffer
		(SDWORD*) pLen);	// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
        (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// bindInParameterDate
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterDate
  (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar, 
  jint year, jint month, jint day, jbyteArray dataBuf, jbyteArray errorCode,
  jintArray buffers)

{
// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	UCHAR*  pBuf = NULL;
	DATE_STRUCT dt;
	SDWORD	lBuf = sizeof (dt);
	
	int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

// Construct our date structure
	memset (&dt, 0x00, sizeof (dt));
	dt.year = (short) year;
	dt.month = (short) month;
	dt.day = (short) day;
	
// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.  If a buffer is provided, it is assumed to be large enough
// to fit the date structure

	if (dataBuf != NULL) {
		pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);
                // Make a copy of the incoming value
		memcpy (pBuf, &dt, lBuf);
                pBuffers[0]=(int)pBuf;
                // pBuffers[1]=(int)dataBuf;
                pBuffers[1]=(int)gDataBuf;
	}

                (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);
	
// Bind the parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_DATE,		// The C data type
		SQL_DATE,		// The SQL data type
		10,			// Precision
		0,			// Scale
		pBuf,			// Pointer to parameter's data
		lBuf,			// Maximum length of parameter buffer
		NULL);			// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// bindInParameterDouble
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterDouble
     (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar, 
	 jint SQLtype, jint scale, jdouble value, jbyteArray dataBuf, 
         jbyteArray errorCode, jintArray buffers)

{
// Get the data structure portion of the error code
	RETCODE rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pBuf = NULL;
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	SDWORD	lBuf = sizeof (value);
	int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.  If a buffer is provided, it is assumed to be large enough
// to fit the float value

	if (dataBuf != NULL) {
		pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);
		// Make a copy of the incoming value
		memcpy (pBuf, &value, lBuf);
                pBuffers[0]=(int)pBuf;
                pBuffers[1]=(int)gDataBuf;
	}

                (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);
	
// Bind the parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_DOUBLE,		// The C data type
		(SWORD) SQLtype,	// The SQL data type
		18,			// Precision
		(SWORD) scale,		// Scale
		pBuf,			// Pointer to parameter's data
		lBuf,			// Maximum length of parameter buffer
		NULL);			// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// bindInParameterFloat
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterFloat
     (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar, 
	 jint SQLtype, jint scale, jfloat value, jbyteArray dataBuf, 
         jbyteArray errorCode, jintArray buffers)
{
// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	UCHAR*	pBuf = NULL;
	SDWORD	lBuf = sizeof (value);
	int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);


// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.  If a buffer is provided, it is assumed to be large enough
// to fit the float value

	if (dataBuf != NULL) {
        pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);
		// Make a copy of the incoming value
		memcpy (pBuf, &value, lBuf);
                pBuffers[0]=(int)pBuf;
                pBuffers[1]=(int)gDataBuf;
	}

                (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);
// Bind the parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_FLOAT,		// The C data type
		(SWORD) SQLtype,	// The SQL data type
		18,			// Precision
		(SWORD) scale,		// Scale
		pBuf,			// Pointer to parameter's data
		lBuf,			// Maximum length of parameter buffer
		NULL);			// Pointer to parameter's length

// Set the error code
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}


//----------------------------------------------------------------------------
// bindInParameterInteger
//----------------------------------------------------------------------------

JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterInteger
  (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar,
  jint SQLtype, jint value, jbyteArray dataBuf, jbyteArray errorCode, 
  jintArray buffers)
{

// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pBuf = NULL;
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	SDWORD	lBuf = sizeof (value);

	int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

		
// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.  If a buffer is provided, it is assumed to be large enough
// to fit the value

	if (dataBuf != NULL) {
		pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);

		// Make a copy of the incoming value
		memcpy (pBuf, &value, lBuf);
                pBuffers[0]=(int)pBuf;
                pBuffers[1]=(int)gDataBuf;
	}
	(*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);

// Bind the parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_SLONG,		// The C data type
		(SWORD) SQLtype,	// The SQL data type
		(UDWORD) lBuf,		// Precision
		0,			// Scale
		pBuf,			// Pointer to parameter's data
		lBuf,			// Maximum length of parameter buffer
		NULL);			// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}
//----------------------------------------------------------------------------
// bindInParameterNull
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterNull
     (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar, 
         jint SQLtype, jbyteArray lenBuf, jbyteArray errorCode, jintArray buffers)
{
// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	jobject gLenBuf = (*env)->NewGlobalRef(env, lenBuf);
	UCHAR*	pLen = (UCHAR*) (*env)->GetByteArrayElements(env, gLenBuf, 0);
	SDWORD  lenValue = SQL_NULL_DATA;
	UDWORD	prec = 0;
	
        int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

        pBuffers[0]=(int)pLen;
        pBuffers[1]=(int)gLenBuf;

        (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);

// Set the length parameter to indicate a null value
	
	memcpy (pLen, &lenValue, sizeof (lenValue));

	// If the SQL type is CHAR, set the precision.  This is a workaround
	// for Oracle

	if (SQLtype == SQL_CHAR) {
		prec = 1;
	}

// Bind the parameter to NULL
	
	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_DEFAULT,		// The C data type
		(SWORD) SQLtype,	// The SQL data type
		prec,				// Precision
		0,					// Scale
		NULL,				// Pointer to parameter's data
		0,					// Maximum length of parameter buffer
		(SDWORD*) pLen);	// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// bindInParameterString
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterString
     (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar, 
	 jint SQLtype, jbyteArray value, jint cbColDef, jint ibScale, 
         jbyteArray dataBuf, jbyteArray errorCode, jintArray buffers)
{
// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	jobject gValue = (*env)->NewGlobalRef(env, value);
	UCHAR*  rgbValue = (UCHAR*) (*env)->GetByteArrayElements(env, gValue, 0);
	UCHAR*	pBuf = NULL;
	SDWORD	lBuf = 0;

        int* pBuffers = (int*)(*env)->GetIntArrayElements(env, buffers, 0);
        pBuffers[2] = (int)rgbValue;
        pBuffers[3] = (int)gValue;
// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.

	if (dataBuf != NULL) {
		pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);
		lBuf = (*env)->GetArrayLength(env, dataBuf);
                pBuffers[0]=(int)pBuf;
                pBuffers[1]=(int)gDataBuf;
		// Sanity check the length
		if (lBuf > (SDWORD) strlen ((LPCSTR) rgbValue)) {
			lBuf = strlen ((LPCSTR) rgbValue);
		}

		// Make a copy of the data
		memcpy (pBuf, rgbValue, lBuf);
	}
		 	
         (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);

// Bind the string parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_CHAR,			// The C data type
		(SWORD) SQLtype,	// The SQL data type
		(UDWORD) cbColDef,	// Precision
		(SWORD) ibScale,	// Scale
		pBuf,				// Pointer to parameter's data
		lBuf,				// Maximum length of parameter buffer
		NULL);				// Pointer to parameter's length

	// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// bindInParameterTime
//----------------------------------------------------------------------------

JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterTime
     (JNIEnv *env, jobject *callingObject, jint hStmt, jint ipar, 
	  jint hour, jint minutes, jint seconds, jbyteArray dataBuf, 
          jbyteArray errorCode, jintArray buffers)
{
// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pBuf = NULL;
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	TIME_STRUCT tm;
	int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);
	SDWORD	lBuf = sizeof (tm);
	


// Construct our time structure
	memset (&tm, 0x00, sizeof (tm));
	tm.hour = (short) hour;
	tm.minute = (short) minutes;
	tm.second = (short) seconds;
	
// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.  If a buffer is provided, it is assumed to be large enough
// to fit the time structure

	if (dataBuf != NULL) {
		pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);

		// Make a copy of the incoming value
		memcpy (pBuf, &tm, lBuf);
                pBuffers[0]=(int)pBuf;
                pBuffers[1]=(int)gDataBuf;
	}
        (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);
	
// Bind the parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_TIME,		// The C data type
		SQL_TIME,		// The SQL data type
		8,			// Precision
		0,			// Scale
		pBuf,			// Pointer to parameter's data
		lBuf,			// Maximum length of parameter buffer
		NULL);			// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// bindInParameterTimestamp
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_bindInParameterTimestamp (
		JNIEnv *env, jobject *callingObject, 
		jint hStmt, 
		jint ipar, 
		jint year, 
		jint month, 
		jint day, 
		jint hour, 
		jint minutes, 
		jint seconds, 
		jint nanos,
		jbyteArray dataBuf, 
		jbyteArray errorCode,
		jintArray buffers)
{
// Get the data structure portion of the error code
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pBuf = NULL;
	TIMESTAMP_STRUCT ts;
	jobject gDataBuf = (*env)->NewGlobalRef(env, dataBuf);
	int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

	SDWORD	lBuf = sizeof (ts);
	
// Construct our timestamp structure
	memset (&ts, 0x00, sizeof (ts));
	ts.year = (short) year;
	ts.month = (short) month;
	ts.day = (short) day;
	ts.hour = (short) hour;
	ts.minute = (short) minutes;
	ts.second = (short) seconds;
	ts.fraction = nanos;
	
// Make a copy of the data coming in into the permanent data buffer.  This
// data is in native format.  If no buffer is provided, the data buffer
// will be NULL.  If a buffer is provided, it is assumed to be large enough
// to fit the timestamp structure

	if (dataBuf != NULL) {
		pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);

		// Make a copy of the incoming value
		memcpy (pBuf, &ts, lBuf);
        pBuffers[0]=(int)pBuf;
        pBuffers[1]=(int)gDataBuf;
	}

        (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);
	
// Bind the parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) ipar,		// Parameter number
		SQL_PARAM_INPUT,	// Type of parameter
		SQL_C_TIMESTAMP,	// The C data type
		SQL_TIMESTAMP,		// The SQL data type
		29,			// Precision
		9,			// Scale
		pBuf,			// Pointer to parameter's data
		lBuf,			// Maximum length of parameter buffer
		NULL);			// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}


//----------------------------------------------------------------------------
// bindOutParameterString
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_bindOutParameterString (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint ipar,
		jint SQLtype,
		jint scale,
		jbyteArray dataBuf,
		jbyteArray lenBuf,
		jbyteArray errorCode,
		jintArray buffers)
{
// Get the data structure portion of the error code
	jobject gDataBuf=(*env)->NewGlobalRef(env, dataBuf); 
	jobject gLenBuf=(*env)->NewGlobalRef(env, lenBuf);
	RETCODE	rc;
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*  pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, gDataBuf, 0);
	UCHAR*	pLen = (UCHAR*) (*env)->GetByteArrayElements(env, gLenBuf, 0);
	SDWORD	lBuf = (*env)->GetArrayLength(env, dataBuf);
	int* pBuffers=(int*)(*env)->GetIntArrayElements(env, buffers, 0);

	


        pBuffers[0]=(int)pBuf;
        pBuffers[1]=(int)gDataBuf;
        pBuffers[2]=(int)pLen;
        pBuffers[3]=(int)gLenBuf;


        (*env)->ReleaseIntArrayElements(env, buffers, pBuffers, 0);

// Bind the output string parameter

	rc = SQLBindParameter (
		(HSTMT) hStmt,		// Statement handle
		(SWORD) ipar,		// Parameter number
		SQL_PARAM_OUTPUT,	// Type of parameter
		SQL_C_CHAR,		// The C data type
		(SWORD) SQLtype,	// The SQL data type
		(UDWORD) lBuf - 1,	// Precision
		(SWORD) scale,		// Scale
		pBuf,			// Pointer to parameter data buffer
		lBuf,			// Maximum length of parameter
		(SDWORD*) pLen);	// Pointer to parameter's length

// Set the error code
	
	errCode[0] = (char) rc;
    	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// browseConnect
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_browseConnect (
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jbyteArray connStrIn,
		jbyteArray connStrOut,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	szConnStrOut = (UCHAR*) (*env)->GetByteArrayElements(env, connStrOut, 0);
	SWORD	cbConnStrOutMax = (SWORD) (*env)->GetArrayLength(env, connStrOut);
	UCHAR*	pConnectString = (UCHAR*) (*env)->GetByteArrayElements(env, connStrIn, 0);
	
	RETCODE	rc;
	SWORD	cbConnStrOut = 0;

    rc = SQLBrowseConnect (
		(HDBC) hDbc,		// Connection handle
		pConnectString,		// Full connection string
		SQL_NTS,		    // Length of connection string
		szConnStrOut,		// Pointer to completed connection string
		cbConnStrOutMax,	// Maximum length of completed connection string
		&cbConnStrOut);		// Pointer to completed connection string length

	errCode[0] = (char) rc;
    (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
    (*env)->ReleaseByteArrayElements(env, connStrOut, szConnStrOut, 0);
    (*env)->ReleaseByteArrayElements(env, connStrIn, pConnectString, 0);
}

//----------------------------------------------------------------------------
// cancel
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_cancel
     (JNIEnv *env, jobject *callingObject, jint hStmt, 
	  jbyteArray errorCode)

{
// Get the data structure portion of the error code
    UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;

// Cancel the statement
	
	rc = SQLCancel (
		(HSTMT) hStmt);		// Statement handle

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// colAttributes
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_colAttributes (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jint type,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*	errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	long	value = 0;
	SWORD	lValue = 0;
	
	RETCODE	rc;
	
// Get the column attribute
	
	rc = SQLColAttributes (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		(UWORD) type,		// Descriptor type
		NULL,			// Pointer to descriptor storage
		0,			// Maximum length of descriptor
		&lValue,		// Length of descriptor
		&value);		// Pointer to integer value for
					//  descriptor

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return value;
}

//----------------------------------------------------------------------------
// colAttributesString
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_colAttributesString (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jint type,
		jbyteArray rgbDesc,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pDesc = (UCHAR*) (*env)->GetByteArrayElements(env, rgbDesc, 0);
	SWORD	lDesc = (SWORD) (*env)->GetArrayLength(env, rgbDesc);
	SWORD	lValue = 0;
	
	RETCODE	rc;
	
// Get the column attribute
	
	rc = SQLColAttributes (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		(UWORD) type,		// Descriptor type
		pDesc,			// Pointer to descriptor buffer
		lDesc,			// Maximum length of descriptor
		&lValue,		// Pointer to descriptor length
		NULL);			// Pointer to integer value for
					//  descriptor
	
// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
    (*env)->ReleaseByteArrayElements(env, rgbDesc, pDesc, 0);
}

//----------------------------------------------------------------------------
// columns
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_columns (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
        jbyteArray table, jboolean tableNull,        
        jbyteArray column, jboolean columnNull,      
        jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*  pTable = NULL;
	UCHAR*  pColumn = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!tableNull) {
		pTable = (UCHAR*) (*env)->GetByteArrayElements(env, table, 0);
	}
	
	if (!columnNull) {
		pColumn = (UCHAR*) (*env)->GetByteArrayElements(env, column, 0);
	}

	rc = SQLColumns (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Table catalog (qualifier)
		SQL_NTS,		// Length of catalog
		pSchema,		// Table schema name (owner)
		SQL_NTS,		// Length of schema name
		pTable,			// Table name
		SQL_NTS,		// Length of table name
		pColumn,		// Search pattern for column names
		SQL_NTS);		// Length of search pattern

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!tableNull) 
		(*env)->ReleaseByteArrayElements(env, table, pTable, 0);
	if (!columnNull) 
		(*env)->ReleaseByteArrayElements(env, column, pColumn, 0);
}

//----------------------------------------------------------------------------
// columnPrivileges
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_columnPrivileges (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
        jbyteArray table, jboolean tableNull,        
        jbyteArray column, jboolean columnNull,      
        jbyteArray errorCode)

{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*  pTable = NULL;
	UCHAR*  pColumn = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!tableNull) {
		pTable = (UCHAR*) (*env)->GetByteArrayElements(env, table, 0);
	}
	
	if (!columnNull) {
		pColumn = (UCHAR*) (*env)->GetByteArrayElements(env, column, 0);
	}


	rc = SQLColumnPrivileges (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Table catalog (qualifier)
		SQL_NTS,		// Length of table catalog
		pSchema,		// Table schema name (owner)
		SQL_NTS,		// Length of schema name
		pTable,			// Table name
		SQL_NTS,		// Length of table name
		pColumn,		// Search pattern for column names
		SQL_NTS);		// Length of search pattern

	errCode[0] = (char) rc;
    (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements (env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements (env, schema, pSchema, 0);
	if (!tableNull) 
		(*env)->ReleaseByteArrayElements (env, table, pTable, 0);
	if (!columnNull) 
		(*env)->ReleaseByteArrayElements (env, column, pColumn, 0);

}

//----------------------------------------------------------------------------
// describeParam
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_describeParam (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint param,
		jint returnParam,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	SWORD	sqlType = 0;
	UDWORD	precision = 0;
	SWORD	scale = 0;
	SWORD	nullable = 0;
	long	returnValue = 0;
	
// SQLDescribeParam
	
	rc = SQLDescribeParam (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) param,		// Parameter marker number
		&sqlType,		// The SQL data type
		&precision,		// Precision
		&scale,			// Scale
		&nullable);		// Parameter allow NULL values?

// Set the error code
	
	errCode[0] = (char) rc;

// Determine which piece of information to return
	if (returnParam == 1) {
		returnValue = sqlType;
	}
	else if (returnParam == 2) {
		returnValue = precision;
	}
	else if (returnParam == 3) {
		returnValue = scale;
	}
	else if (returnParam == 4) {
		returnValue = nullable;
	}
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return returnValue;
}

//----------------------------------------------------------------------------
// disconnect
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_disconnect(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;
	
// Disconnect
	
	rc = SQLDisconnect (
		(HDBC) hDbc);		// Connection handle

// Set the error code
	
 errCode[0] = (char) rc;
 (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// driverConnect
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_driverConnect(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jbyteArray connectString,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*	errCode = (UCHAR*)(*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*  pConnectString = (UCHAR*)(*env)->GetByteArrayElements(env, connectString, 0);
    UCHAR	szConnStrOut[256];
	SWORD	cbConnStrOut = 0;

	rc = SQLDriverConnect (
		(HDBC) hDbc,		// Connection handle
		NULL,			// Window handle
		pConnectString,		// Full connection string
		SQL_NTS,		// Length of connection string
		szConnStrOut,		// Pointer to completed connection
					//  string
		sizeof (szConnStrOut),	// Maximum length of completed
					//  connection string
		&cbConnStrOut,		// Pointer to completed connection
					//  string length
		SQL_DRIVER_NOPROMPT);	// Prompt flag

	errCode[0] = (char) rc;

	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
    (*env)->ReleaseByteArrayElements(env,  connectString, pConnectString, 0);
}

//----------------------------------------------------------------------------
// error
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_error(
		JNIEnv *env, jobject *callingObject,
		jint hEnv,
		jint hDbc,
		jint hStmt,
		jbyteArray sqlState,
		jbyteArray errorMsg,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;
	UCHAR*	pSqlState = (UCHAR*) (*env)->GetByteArrayElements(env, sqlState, 0);
	UCHAR*	pErrorMsg = (UCHAR*) (*env)->GetByteArrayElements(env, errorMsg, 0);
	SWORD	lErrorMsg = (SWORD)  (*env)->GetArrayLength(env, errorMsg);
	SWORD	cbErrorMsg = 0;
	long	nativeError = 0;
	
	rc = SQLError (
		(HENV) hEnv,		// Environment handle
		(HDBC) hDbc,		// Connection handle
		(HSTMT) hStmt,		// Statement handle
		pSqlState,		// Pointer to SQLSTATE
		&nativeError,		// Pointer to native error code
		pErrorMsg,		// Pointer to error message buffer
		lErrorMsg,		// Maximum length of error message
		&cbErrorMsg);		// Length of error message buffer

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, sqlState, pSqlState, 0);
	(*env)->ReleaseByteArrayElements(env, errorMsg, pErrorMsg, 0);
	return nativeError;
}

//----------------------------------------------------------------------------
// execDirect
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_execDirect (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray SQLString,
		jbyteArray errorCode)
{
	// Get the data structure portion of the error code
	UCHAR*	errCode = (UCHAR*)(*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;
	UCHAR*  pSQL = (UCHAR*)(*env)->GetByteArrayElements(env, SQLString, 0);

	rc = SQLExecDirect (
    	 (HSTMT) hStmt,		// Statement handle
 		 pSQL,			// SQL Statement
		 SQL_NTS);		// Length of SQL statement

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
    (*env)->ReleaseByteArrayElements(env,  SQLString, pSQL, 0);
}

//----------------------------------------------------------------------------
// execute
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_execute (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray errorCode)
{
	// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;

	rc = SQLExecute (
		(HSTMT) hStmt);		// Statement handle

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// fetch
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_fetch (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);

	RETCODE	rc;

// Fetch the next row from the statement
	
	rc = SQLFetch (
		(HSTMT) hStmt);		// Statement handle

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// foreignKeys
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_foreignKeys (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray PKcatalog, jboolean PKcatalogNull,
		jbyteArray PKschema, jboolean PKschemaNull,
		jbyteArray PKtable, jboolean PKtableNull,
		jbyteArray FKcatalog, jboolean FKcatalogNull,
		jbyteArray FKschema, jboolean FKschemaNull,
		jbyteArray FKtable, jboolean FKtableNull,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pPKCatalog = NULL;
	UCHAR*	pPKSchema = NULL;
	UCHAR*  pPKTable = NULL;
	UCHAR*	pFKCatalog = NULL;
	UCHAR*	pFKSchema = NULL;
	UCHAR*  pFKTable = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	if (!PKcatalogNull) {
		pPKCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, PKcatalog, 0);
	}
	
	if (!PKschemaNull) {
		pPKSchema = (UCHAR*) (*env)->GetByteArrayElements(env, PKschema,0);
	}
	
	if (!PKtableNull) {
		pPKTable = (UCHAR*) (*env)->GetByteArrayElements(env, PKtable, 0);
	}
	if (!FKcatalogNull) {
		pFKCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, FKcatalog, 0);
	}
	
	if (!FKschemaNull) {
		pFKSchema = (UCHAR*) (*env)->GetByteArrayElements(env, FKschema, 0);
	}
	
	if (!FKtableNull) {
		pFKTable = (UCHAR*) (*env)->GetByteArrayElements(env, FKtable, 0);
	}
	
	rc = SQLForeignKeys (
		(HSTMT) hStmt,		// Statement handle
		pPKCatalog,		// Primary key catalog (qualifier)
		SQL_NTS,		// Length of primary key catalog
		pPKSchema,		// Primary key schema name (owner)
		SQL_NTS,		// Length of primary key schema
		pPKTable,		// Primary key table name
		SQL_NTS,		// Length of primary key table name
		pFKCatalog,		// Foreign key catalog (qualifier)
		SQL_NTS,		// Length of foreign key catalog
		pFKSchema,		// Foreign key schema name (owner)
		SQL_NTS,		// Length of foreign key schema
		pFKTable,		// Foreign key table name
		SQL_NTS);		// Length of foreign key table name

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!PKcatalogNull) 
		(*env)->ReleaseByteArrayElements(env, PKcatalog, pPKCatalog, 0);
	if (!PKschemaNull) 
		(*env)->ReleaseByteArrayElements(env, PKschema, pPKSchema, 0);
	if (!PKtableNull) 
		(*env)->ReleaseByteArrayElements(env, PKtable, pPKTable, 0);
	if (!FKcatalogNull) 
		(*env)->ReleaseByteArrayElements(env, FKcatalog, pFKCatalog, 0);
	if (!FKschemaNull) 
		(*env)->ReleaseByteArrayElements(env, FKschema, pFKSchema, 0);
	if (!FKtableNull) 
		(*env)->ReleaseByteArrayElements(env, FKtable, pFKTable, 0);
}


//----------------------------------------------------------------------------
// freeConnect
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_freeConnect(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;

// Free the connection handle
	
	rc = SQLFreeConnect (
		(HDBC) hDbc);		// Connection handle

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// freeEnv
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_freeEnv(
		JNIEnv *env, jobject *callingObject,
		jint hEnv,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;
	
// Free the environment handle
	
	rc = SQLFreeEnv (
		(HENV) hEnv);		// Environment handle

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// freeStmt
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL    Java_sun_jdbc_odbc_JdbcOdbc_freeStmt (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint fOption,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	
// Free the statement handle
	
	rc = SQLFreeStmt (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) fOption);	// Option

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// getConnectOption
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getConnectOption(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jshort fOption,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	long	vParam = 0;

// Get the connection option
	
	rc = SQLGetConnectOption (
		(HDBC) hDbc,		// Connection handle
		fOption,		// Option to retrieve
		&vParam);		// Pointer to value

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return vParam;
}

//----------------------------------------------------------------------------
// getConnectOptionString
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL  Java_sun_jdbc_odbc_JdbcOdbc_getConnectOptionString(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jshort fOption,
		jbyteArray szParam,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pParam = (UCHAR*) (*env)->GetByteArrayElements(env, szParam, 0);
	
	RETCODE	rc;

// Get the connection option
	
	rc = SQLGetConnectOption (
		(HDBC) hDbc,		// Connection handle
		fOption,		// Option to retrieve
		pParam);		// Pointer to value

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, szParam, pParam, 0);
}

//----------------------------------------------------------------------------
// getCursorName
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_getCursorName (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray szCursor,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
    UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
    UCHAR*  pCursor = (UCHAR*) (*env)->GetByteArrayElements(env, szCursor, 0);
	SWORD	maxLen = (SWORD) (*env)->GetArrayLength(env, szCursor);
	SWORD	lValue = 0;
	
	RETCODE	rc;
	
// Get the column data
	
	rc = SQLGetCursorName (
		(HSTMT) hStmt,		// Statement handle
		pCursor,		// Pointer to storage for cursor name
		maxLen,			// Maximum length of cursor name
		&lValue);		// Pointer to cursor name length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, szCursor, pCursor, 0);
}

//----------------------------------------------------------------------------
// getDataBinary
// Returns number of bytes read, -1 if eof
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getDataBinary (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jint cType,
		jbyteArray rgbValue,
		jint maxLen,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pValue = (UCHAR*) (*env)->GetByteArrayElements(env, rgbValue, 0);
	long	lValue = 0;
	
	RETCODE	rc;
	
// Get the column data
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		(SWORD) cType,		// The C data type
		pValue,			// Pointer to storage for the data
		(UWORD) maxLen,	// Maximum length of data
		&lValue);		// Pointer to data length

// If the number of bytes available is greater than our buffer length,
// return the buffer length (data was truncated)
	
	if ((lValue > (long) maxLen) ||
	    (lValue == SQL_NO_TOTAL)) {
		lValue = maxLen;
	}
	
// Set the null indicator
	
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
		lValue = -1;
	}
	
// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, rgbValue, pValue, 0);
	return lValue;
}		


//----------------------------------------------------------------------------
// getDataDouble
//----------------------------------------------------------------------------
JNIEXPORT jdouble JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getDataDouble (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	double	value = 0;
	long	lValue = 0;
	
	RETCODE	rc;
	
// Get the column data
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		SQL_C_DOUBLE,		// The C data type
		&value,			// Pointer to storage for the data
		sizeof (value),		// Maximum length of the data
		&lValue);		// Pointer to data length

// Set the error code
	
	errCode[0] = (char) rc;

// Set the null indicator
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
	}
    (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return value;
}

//----------------------------------------------------------------------------
// getDataFloat
//----------------------------------------------------------------------------
JNIEXPORT jfloat JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getDataFloat (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	float	value = (float) 0;
	long	lValue = 0;
	
	RETCODE	rc;
	
// Get the column data
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		SQL_C_FLOAT,		// The C data type
		&value,			// Pointer to storage for the data
		sizeof (value),		// Maximum length of the data
		&lValue);		// Pointer to data length

// Set the error code
	
	errCode[0] = (char) rc;

// Set the null indicator
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
	}
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return value;
}

//----------------------------------------------------------------------------
// getDataInteger
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getDataInteger (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	long	value = 0;
	long	lValue = 0;
	
	RETCODE	rc;
	
// Get the column data
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		SQL_C_LONG,		// The C data type
		&value,			// Pointer to storage for the data
		sizeof (value),		// Maximum length of the data
		&lValue);		// Pointer to data length

// Set the error code
	
	errCode[0] = (char) rc;

// Set the null indicator
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
	}
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return value;
}		

//----------------------------------------------------------------------------
// getDataString
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getDataString (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jbyteArray rgbValue,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pValue = (UCHAR*) (*env)->GetByteArrayElements(env, rgbValue, 0);
	SDWORD	maxLen = (SWORD) (*env)->GetArrayLength(env, rgbValue);
	long	lValue = 0;
	
	RETCODE	rc;

// Init the data buffer to an empty string

	memset (pValue, 0x00, maxLen);
	
// Get the column data
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		SQL_C_CHAR,		// The C data type
		pValue,			// Pointer to storage for the data
		maxLen,			// Maximum length of the data
		&lValue);		// Pointer to data length

// Set the error code
	errCode[0] = (char) rc;


// Set the null indicator
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
	}

	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, rgbValue, pValue, 0);
	return (long) lValue;
}

//----------------------------------------------------------------------------
// getDataStringDate
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_getDataStringDate (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jbyteArray rgbValue,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
    UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
    UCHAR*	pValue = (UCHAR*) (*env)->GetByteArrayElements(env, rgbValue, 0);
	long	lValue = 0;
	DATE_STRUCT dt;
	
	RETCODE	rc;

// Construct our date structure
	
	memset (&dt, 0x00, sizeof (dt));

// Init the data buffer to an empty string
	
	pValue[0] = 0x00;

// Get the column data as a date, then convert to a string
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		SQL_C_DATE,		// The C data type
		&dt,			// Pointer to storage for the data
		sizeof (dt),		// Maximum length of the data
		&lValue);		// Pointer to data length

// Set the error code
	
	errCode[0] = (char) rc;

// Set the null indicator
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
	}

	// Not a null column.  Create the string date
	
	else {
		sprintf (pValue, "%04i-%02i-%02i", dt.year, dt.month,
				dt.day);
	}
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, rgbValue, pValue, 0);
}

//----------------------------------------------------------------------------
// getDataStringTime
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_getDataStringTime (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jbyteArray rgbValue,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pValue = (UCHAR*) (*env)->GetByteArrayElements(env, rgbValue, 0);
	long	lValue = 0;
	TIME_STRUCT tm;
	
	RETCODE	rc;

// Construct our date structure
	
	memset (&tm, 0x00, sizeof (tm));

// Init the data buffer to an empty string
	
	pValue[0] = 0x00;

// Get the column data as a date, then convert to a string
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		SQL_C_TIME,		// The C data type
		&tm,			// Pointer to storage for the data
		sizeof (tm),		// Maximum length of the data
		&lValue);		// Pointer to data length

// Set the error code
	
	errCode[0] = (char) rc;

// Set the null indicator
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
	}

	// Not a null column.  Create the string time
	
	else {
		sprintf (pValue, "%02i:%02i:%02i", tm.hour, tm.minute,
				tm.second);
	}
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, rgbValue, pValue, 0);
}

//----------------------------------------------------------------------------
// getDataStringTimestamp
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_getDataStringTimestamp (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jint column,
		jbyteArray rgbValue,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
        UCHAR*  pValue = (UCHAR*) (*env)->GetByteArrayElements(env, rgbValue, 0);
	long	lValue = 0;
	TIMESTAMP_STRUCT tm;
	
	RETCODE	rc;

// Construct our date structure
	
	memset (&tm, 0x00, sizeof (tm));

// Init the data buffer to an empty string
	
	pValue[0] = 0x00;

// Get the column data as a date, then convert to a string
	
	rc = SQLGetData (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) column,		// Column number
		SQL_C_TIMESTAMP,	// The C data type
		&tm,			// Pointer to storage for the data
		sizeof (tm),		// Maximum length of the data
		&lValue);		// Pointer to data length

// Set the error code
	
	errCode[0] = (char) rc;

// Set the null indicator
	errCode[1] = 0;
	if (lValue == SQL_NULL_DATA) {
		errCode[1] = 1;
	}

	// Not a null column.  Create the string time
	
	else {
		sprintf (pValue, "%04i-%02i-%02i %02i:%02i:%02i",
				tm.year, tm.month, tm.day, tm.hour,
				tm.minute, tm.second);
		if (tm.fraction != 0) {
			sprintf (&pValue[strlen ((LPCSTR) pValue)], ".%09li",
				tm.fraction);
		}
	}
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, rgbValue, pValue, 0);
}		

//----------------------------------------------------------------------------
// getInfo
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getInfo(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jshort fInfoType,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*	errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;
	long	infoValue = 0;
	SWORD	lInfoValue = 0; 

// Get the requested info
	
	rc = SQLGetInfo (
		(HDBC) hDbc,		// Connection handle
		fInfoType,		// Type of information
		&infoValue,		// Pointer to storage for the info
		(SWORD) sizeof (infoValue),	// Maximum length of the data
		&lInfoValue);		// Pointer to the information length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return infoValue;
}

//----------------------------------------------------------------------------
// getInfoShort
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getInfoShort(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jshort fInfoType,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UWORD	infoValue = 0;
	SWORD	lInfoValue = 0; 

// Get the requested info
	
	rc = SQLGetInfo (
		(HDBC) hDbc,		// Connection handle
		fInfoType,		// Type of information
		&infoValue,		// Pointer to storage for the info
		(SWORD) sizeof (infoValue),	// Maximum length of the data
		&lInfoValue);		// Pointer to the information length

// Set the error code
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return infoValue;
}

//----------------------------------------------------------------------------
// getInfoString
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_getInfoString(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jshort fInfoType,
		jbyteArray szParam,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*	pParam = (UCHAR*) (*env)->GetByteArrayElements(env, szParam, 0);
	SWORD	lParam = (SWORD) (*env)->GetArrayLength(env, szParam);
	SWORD	lInfoValue = 0; 
	
	RETCODE	rc;

// Get the requested info
	
	rc = SQLGetInfo (
		(HDBC) hDbc,		// Connection handle
		fInfoType,		// Type of information
		pParam,			// Pointer to storage for the info
		lParam,			// Maximum length of the data
		&lInfoValue);		// Pointer to the information length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, szParam, pParam, 0);
}

//----------------------------------------------------------------------------
// getStmtOption
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_getStmtOption (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jshort fOption,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	long	vParam = 0;

// Get the connection option
	
	rc = SQLGetStmtOption (
		(HSTMT) hStmt,		// Statement handle
		fOption,		// Option to retrieve
		&vParam);		// Value

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return vParam;
}

//----------------------------------------------------------------------------
// getTypeInfo
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_getTypeInfo (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jshort fSqlType,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*)(*env)->GetByteArrayElements(env, errorCode, 0);	
	RETCODE	rc;
	long	vParam = 0;

// Get the type infos
	
	rc = SQLGetTypeInfo (
		(HSTMT) hStmt,		// Statement handle
		fSqlType);		// The SQL data type

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// moreResults
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_moreResults (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;

// Get the next result set
	
	rc = SQLMoreResults (
		(HSTMT) hStmt);		// Statement handle

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// nativeSql
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_nativeSql (
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jbyteArray query,
		jbyteArray nativeQuery,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
        UCHAR*  pQuery = (UCHAR*) (*env)->GetByteArrayElements(env, query, 0);
	UCHAR*	pNativeQuery = (UCHAR*) (*env)->GetByteArrayElements(env, nativeQuery, 0);
	SDWORD	lNativeQuery = (*env)->GetArrayLength(env, nativeQuery);
	SDWORD	lOut = 0;

	RETCODE rc;

// Call SQLNativeSql

	rc = SQLNativeSql (
		(HDBC) hDbc,		// Connection handle
		pQuery,			// SQL text string to be translated
		SQL_NTS,		// Length of text string
		pNativeQuery,		// Pointer to storage for translated
					//  string
		lNativeQuery,		// Maximum length of translated string
		&lOut);			// Pointer to translated string length

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
    (*env)->ReleaseByteArrayElements(env, query, pQuery, 0);
	(*env)->ReleaseByteArrayElements(env, nativeQuery, pNativeQuery, 0);
}


//----------------------------------------------------------------------------
// numParams
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_numParams (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	short	numParams = 0;

// Get the number of parameters in the prepared statement
	
	rc = SQLNumParams (
		(HSTMT) hStmt,		// Statement handle
		&numParams);		// Number of parameters in the
					//  statement

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return numParams;
}

//----------------------------------------------------------------------------
// numResultCols
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_numResultCols (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*)(*env)->GetByteArrayElements(env, errorCode, 0);	
	
	RETCODE	rc;
	short	numCols = 0;

// Get the number of columns in the result set
	
	rc = SQLNumResultCols (
		(HSTMT) hStmt,		// Statement handle
		&numCols);		// Number of columns in the result set

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return numCols;
}

//----------------------------------------------------------------------------
// paramData
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_paramData (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UDWORD	pBuf;
	long	param = -1;

// Get the parameter data.  This is simply the parameter number, as
// supplied to SQLBindParameter
	
	rc = SQLParamData (
		(HSTMT) hStmt,		// Statement handle
		(PTR) &pBuf);		// Pointer to storage for the value

// If we need data, get the parameter numer that needs is
	
	if (rc == SQL_NEED_DATA) {
		memcpy (&param, (PTR) pBuf, sizeof (param));
	}
	
// Set the error code
	
	errCode[0] = (char) rc;
    (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return param;
}

//----------------------------------------------------------------------------
// prepare
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_prepare (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray SQLString,
		jbyteArray errorCode)
{
	// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	RETCODE	rc;
	UCHAR*	pSQL = (UCHAR*) (*env)->GetByteArrayElements(env, SQLString, 0);

   rc = SQLPrepare (
		(HSTMT) hStmt,		// Statement handle
		pSQL,			// SQL text string
		SQL_NTS);		// Length of the SQL text string

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, SQLString, pSQL, 0);
}

//----------------------------------------------------------------------------
// primaryKeys
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_primaryKeys (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
		jbyteArray table, jboolean tableNull,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*  pTable = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!tableNull) {
		pTable = (UCHAR*) (*env)->GetByteArrayElements(env, table, 0);
	}
	
	rc = SQLPrimaryKeys (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Table catalog (qualifier)
		SQL_NTS,		// Length of catalog	
		pSchema,		// Table schema name (owner)
		SQL_NTS,		// Length of schema name
		pTable,			// Table name
		SQL_NTS);		// Length of table name

	errCode[0] = (char) rc;

    (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!tableNull) 
		(*env)->ReleaseByteArrayElements(env, table, pTable, 0);
}

//----------------------------------------------------------------------------
// procedures
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_procedures (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
		jbyteArray procedure, jboolean procedureNull,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*  pProcedure = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.

	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!procedureNull) {
		pProcedure = (UCHAR*) (*env)->GetByteArrayElements(env, procedure, 0);
	}

	rc = SQLProcedures (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Procedure catalog (qualifier)
		SQL_NTS,		// Length of catalog
		pSchema,		// Procedure schema name (owner)
		SQL_NTS,		// Length of schema name
		pProcedure,		// Search pattern for procedure names
		SQL_NTS);		// Length of search pattern

	errCode[0] = (char) rc;
    (*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!procedureNull) 
		(*env)->ReleaseByteArrayElements(env, procedure, pProcedure, 0);
}

//----------------------------------------------------------------------------
// procedureColumns
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_procedureColumns (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
		jbyteArray procedure, jboolean procedureNull,
		jbyteArray column, jboolean columnNull,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*  pProcedure = NULL;
	UCHAR*  pColumn = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}

	if (!procedureNull) {
		pProcedure = (UCHAR*) (*env)->GetByteArrayElements(env, procedure, 0);
	}
	
	if (!columnNull) {
		pColumn = (UCHAR*) (*env)->GetByteArrayElements(env, column, 0);
	}

	rc = SQLProcedureColumns (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Procedure catalog (qualifier)
		SQL_NTS,		// Length of catalog
		pSchema,		// Procedure schema name (owner)
		SQL_NTS,		// Length of schema name
		pProcedure,		// Search string for procedure names
		SQL_NTS,		// Length of search string
		pColumn,		// Search string for procedure columns
		SQL_NTS);		// Length of search string

	errCode[0] = (char) rc;

	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!procedureNull) 
		(*env)->ReleaseByteArrayElements(env, procedure, pProcedure, 0);
	if (!columnNull) 
		(*env)->ReleaseByteArrayElements(env, column, pColumn, 0);
}

//----------------------------------------------------------------------------
// putData
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_putData (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray dataBuf,
		jint dataLen,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	UCHAR*  pBuf = (UCHAR*) (*env)->GetByteArrayElements(env, dataBuf, 0);
	
	RETCODE	rc;

// Put the data

	rc = SQLPutData (
		(HSTMT) hStmt,		// Statement handle
		pBuf,			// Pointer to data
		(SDWORD) dataLen);	// Length of data

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, dataBuf, pBuf, 0);
}

//----------------------------------------------------------------------------
// rowCount
//----------------------------------------------------------------------------
JNIEXPORT jint JNICALL Java_sun_jdbc_odbc_JdbcOdbc_rowCount (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	long	numRows = 0;

// Get the number of rows affected
	
	rc = SQLRowCount (
		(HSTMT) hStmt,		// Statement handle
		&numRows);		// Number of rows affected

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	return numRows;
}

//----------------------------------------------------------------------------
// setConnectOption
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_setConnectOption(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jshort fOption,
		jint vParam,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;

// Set the connection option
	
	rc = SQLSetConnectOption (
		(HDBC) hDbc,		// Connection handle
		fOption,		// Option to set
		(UDWORD) vParam);	// Value

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// setConnectOptionString
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_setConnectOptionString(
		JNIEnv *env, jobject *callingObject,
		jint hDbc,
		jshort fOption,
		jbyteArray szParam,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pParam = (UCHAR*) (*env)->GetByteArrayElements(env, szParam, 0);

// Set the connection option
	
	rc = SQLSetConnectOption (
		(HDBC) hDbc,		// Connection handle
		fOption,		// Option to set
		(UDWORD) pParam);	// Value

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, szParam, pParam, 0);
}

//----------------------------------------------------------------------------
// setCursorName
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_setCursorName (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray szCursor,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCursor = (UCHAR*) (*env)->GetByteArrayElements(env, szCursor, 0);

// Set the cursor name
	
	rc = SQLSetCursorName (
		(HSTMT) hStmt,		// Statement handle
		pCursor,		// Cursor name
		SQL_NTS);		// Length of cursor name

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	(*env)->ReleaseByteArrayElements(env, szCursor, pCursor, 0);
}

//----------------------------------------------------------------------------
// setStmtOption
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_setStmtOption(
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jshort fOption,
		jint vParam,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;

// Set the statement option
	
	rc = SQLSetStmtOption (
		(HSTMT) hStmt,		// Statement handle
		fOption,		// Option to set
		(UDWORD) vParam);	// Value

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}

//----------------------------------------------------------------------------
// specialColumns
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_specialColumns (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jshort fOption,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
		jbyteArray table, jboolean tableNull,
		jint fScope,
		jint fNullable,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*  pTable = NULL;
	UWORD   nullable = SQL_NO_NULLS;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!tableNull) {
		pTable = (UCHAR*) (*env)->GetByteArrayElements(env, table, 0);
	}

	if (fNullable) {
		nullable = SQL_NULLABLE;
	}
	
	rc = SQLSpecialColumns (
		(HSTMT) hStmt,		// Statement handle
		(UWORD) fOption,	// Type of column to return
		pCatalog,		// Table catalog (qualifier)
		SQL_NTS,		// Length of catalog
		pSchema,		// Table schema name (owner)	
		SQL_NTS,		// Length of schema name
		pTable,			// Table name
		SQL_NTS,		// Length of table name
		(UWORD) fScope,		// Minimum scope
		nullable);		// Return column that can have nulls?

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!tableNull) 
		(*env)->ReleaseByteArrayElements(env, table, pTable, 0);
}

//----------------------------------------------------------------------------
// statistics
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_statistics (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
		jbyteArray table, jboolean tableNull,
		jint unique,
		jint approximate,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*   pTable = NULL;
	UWORD	fUnique = SQL_INDEX_ALL;
	UWORD	fAccuracy = SQL_ENSURE;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!tableNull) {
		pTable = (UCHAR*) (*env)->GetByteArrayElements(env, table, 0);
	}
	

	// If only unique indexes were specified, set the proper flag
	
	if (unique) {
		fUnique = SQL_INDEX_UNIQUE;
	}

	// If only approximate indexes are required, set the proper flag

	if (approximate) {
		fAccuracy = SQL_QUICK;
	}
	
	rc = SQLStatistics (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Table catalog (qualifer)
		SQL_NTS,		// Length of catalog
		pSchema,		// Table schema name (owner)
		SQL_NTS,		// Length of schema name
		pTable,			// Table name
		SQL_NTS,		// Length of table name
		fUnique,		// Type of index
		fAccuracy);		// Importance of cardinality and pages

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!tableNull) 
		(*env)->ReleaseByteArrayElements(env, table, pTable, 0);
}

//----------------------------------------------------------------------------
// tables
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_tables (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
		jbyteArray table, jboolean tableNull,
		jbyteArray types, jboolean typesNull,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*   pTable = NULL;
	UCHAR*   pTypes = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!tableNull) {
		pTable = (UCHAR*) (*env)->GetByteArrayElements(env, table, 0);
	}

	if (!typesNull) {
		pTypes = (UCHAR*) (*env)->GetByteArrayElements(env, types, 0);
	}

	rc = SQLTables (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Table catalog (qualifer)
		SQL_NTS,		// Length of catalog
		pSchema,		// Table schema name (owner)
		SQL_NTS,		// Length of schema name
		pTable,			// Search pattern for table names
		SQL_NTS,		// Length of search pattern
		pTypes,			// List of table types to match
		SQL_NTS);		// Length of list

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!tableNull) 
		(*env)->ReleaseByteArrayElements(env, table, pTable, 0);
	if (!typesNull) 
		(*env)->ReleaseByteArrayElements(env, types, pTypes, 0);
}

//----------------------------------------------------------------------------
// tablePrivileges
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL sun_jdbc_odbc_JdbcOdbc_tablePrivileges (
		JNIEnv *env, jobject *callingObject,
		jint hStmt,
		jbyteArray catalog, jboolean catalogNull,
		jbyteArray schema, jboolean schemaNull,
		jbyteArray table, jboolean tableNull,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);

	RETCODE	rc;
	UCHAR*	pCatalog = NULL;
	UCHAR*	pSchema = NULL;
	UCHAR*  pTable = NULL;

	// Check if the parameter is null.  If it is, we need to pass
	// a NULL to ODBC.  Unfortuneatly, Java converts a null string
	// to an empty string.
	
	if (!catalogNull) {
		pCatalog = (UCHAR*) (*env)->GetByteArrayElements(env, catalog, 0);
	}
	
	if (!schemaNull) {
		pSchema = (UCHAR*) (*env)->GetByteArrayElements(env, schema, 0);
	}
	
	if (!tableNull) {
		pTable = (UCHAR*) (*env)->GetByteArrayElements(env, table, 0);
	}
	
	rc = SQLTablePrivileges (
		(HSTMT) hStmt,		// Statement handle
		pCatalog,		// Table catalog (qualifier)
		SQL_NTS,		// Length of catalog
		pSchema,		// Table schema name (owner)
		SQL_NTS,		// Length of schema name
		pTable,			// Search pattern for table names
		SQL_NTS);		// Length of search pattern

	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
	if (!catalogNull) 
		(*env)->ReleaseByteArrayElements(env, catalog, pCatalog, 0);
	if (!schemaNull) 
		(*env)->ReleaseByteArrayElements(env, schema, pSchema, 0);
	if (!tableNull) 
		(*env)->ReleaseByteArrayElements(env, table, pTable, 0);
}

//----------------------------------------------------------------------------
// transact
//----------------------------------------------------------------------------
JNIEXPORT void JNICALL	Java_sun_jdbc_odbc_JdbcOdbc_transact (
		JNIEnv *env, jobject *callingObject,
		jint hEnv,
		jint hDbc,
		jshort fType,
		jbyteArray errorCode)
{
// Get the data structure portion of the error code
	UCHAR*  errCode = (UCHAR*) (*env)->GetByteArrayElements(env, errorCode, 0);
	
	RETCODE	rc;

// Call SQLTransact
	
	rc = SQLTransact (
		(HENV) hEnv,		// Environment handle
		(HDBC) hDbc,		// Connection handle
		fType);			// Type (commit, rollback)

// Set the error code
	
	errCode[0] = (char) rc;
	(*env)->ReleaseByteArrayElements(env, errorCode, errCode, 0);
}



// These functions release the stored buffers (Java's garbage collector
// will no longer maintain, or pin, the buffers). 
JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_ReleaseStoredBytes (
        JNIEnv *env, jobject *callingObject,
		jint x1,
		jint x2)
{
	(*env)->ReleaseByteArrayElements(env, (jbyteArray)x2, (UCHAR*)x1, 0);
	(*env)->DeleteGlobalRef(env, (jobject)x2);
}

JNIEXPORT void JNICALL Java_sun_jdbc_odbc_JdbcOdbc_ReleaseStoredChars (
        JNIEnv *env, jobject *callingObject,
		jint s1,
		jint s2)
{
	(*env)->ReleaseStringUTFChars(env, (jstring)s2, (UCHAR*)s1);
	(*env)->DeleteGlobalRef(env, (jobject)s2);
}
