/*
 * %W% %E%                    
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbc.java
//
// Description: Defines the JdbcOdbc class.  This class provides a java 
//              interface access to the ODBC API routines
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.math.BigDecimal;
import java.sql.*;
import sun.io.*;
class JdbcOdbc
        extends JdbcOdbcObject {

	//====================================================================
	// Public methods
	//====================================================================

	//--------------------------------------------------------------------
	// Constructor
	// Perform any initialization, including loading the jdbcodbc
	// bridge library
	//--------------------------------------------------------------------

    public JdbcOdbc (String prefix)
			throws SQLException
	{

		try {
             
			if (isTracing ()) {
				java.util.Date curDate = new java.util.Date ();
 
				String s = "";
				int minor = MinorVersion;

				// Format the minor version to have 4 digits,
				// with leading 0's if necessary

				if (minor < 1000) s += "0";
				if (minor < 100)  s += "0";
				if (minor < 10)   s += "0";
				s += "" + minor;

				trace ("JDBC to ODBC Bridge " + MajorVersion + "." + s);
				trace ("Current Date/Time: " +
					curDate.toString ());
				trace ("Loading " + prefix + "JdbcOdbc library");
			}
                    
			java.security.AccessController.doPrivileged(
			    new sun.security.action.LoadLibraryAction(prefix + "JdbcOdbc"));

			//System.loadLibrary (prefix + "JdbcOdbc");

		} catch (UnsatisfiedLinkError e) {

			if (isTracing ()) {
				trace ("Unable to load " + prefix + "JdbcOdbc library");
			}
                 
			throw new SQLException (
				"Unable to load " + prefix + "JdbcOdbc library");
		}

	}

	//--------------------------------------------------------------------
	// Public java methods (wrappers around native C calls)
	//--------------------------------------------------------------------

	//--------------------------------------------------------------------
	// SQLAllocConnect
	//--------------------------------------------------------------------

	public synchronized int SQLAllocConnect (
			int hEnv)
			throws SQLException
	{
		int	hDbc = 0;
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Allocating Connection handle (SQLAllocConnect)");
		}

		errorCode = new byte[1];
		hDbc = allocConnect (hEnv, errorCode);

		if (errorCode[0] != 0) {
			throwGenericSQLException ();
		}
		else {
			if (isTracing ()) {
				trace ("hDbc=" + hDbc);
			}
		}
		return hDbc;
	}

	//--------------------------------------------------------------------
	// SQLAllocEnv
	//--------------------------------------------------------------------

	public synchronized int SQLAllocEnv () throws SQLException
	{
		int	hEnv = 0;
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Allocating Environment handle (SQLAllocEnv)");
		}

		errorCode = new byte[1];
		hEnv = allocEnv (errorCode);

		if (errorCode[0] != 0) {
			throwGenericSQLException ();
		}
		else {
			if (isTracing ()) {
				trace ("hEnv=" + hEnv);
			}
		}
		return hEnv;
	}

	//--------------------------------------------------------------------
	// SQLAllocStmt
	//--------------------------------------------------------------------

	public synchronized int SQLAllocStmt (
			int hDbc)
			throws SQLException
	{
		int	hStmt = 0;
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Allocating Statement Handle (SQLAllocStmt), hDbc=" + hDbc);
		}

		errorCode = new byte[1];
		hStmt = allocStmt (hDbc, errorCode);

		if (errorCode[0] != 0) {
			throwGenericSQLException ();
		}
		else {
			if (isTracing ()) {
				trace ("hStmt=" + hStmt);
			}
		}
		return hStmt;
	}

	//--------------------------------------------------------------------
	// SQLBindColDefault
	//--------------------------------------------------------------------

	public synchronized void SQLBindColDefault (
			int hStmt,
			int ipar,
			byte rgbValue[],
			byte pcbValue[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding default column (SQLBindCol), hStmt=" + hStmt + ", ipar=" + ipar + ", 			length=" + rgbValue.length);
		}

		errorCode = new byte[1];
		bindColDefault (hStmt, ipar, rgbValue, pcbValue,
				errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterAtExec
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterAtExec (
			int hStmt,
			int ipar,
			int SQLtype,
			int len,
			byte dataBuf[],
                        byte lenBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding DATA_AT_EXEC parameter (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", SQLtype=" + SQLtype + ", len=" + len);
		}

		errorCode = new byte[1];
		bindInParameterAtExec (hStmt, ipar, SQLtype, len, dataBuf,
                                lenBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterBinary
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterBinary (
			int hStmt,
			int ipar,
			int SQLtype,
			byte value[],
			byte dataBuf[],
                        byte lenBuf[],
                        int buffers[])                        throws SQLException
	{
		byte	errorCode[];

		// Since we don't know the precision of the column that
		// we are about to bind, we will set it to this arbitrary
		// value.

		int	precision = JdbcOdbcLimits.DEFAULT_IN_PRECISION;

		if (isTracing ()) {
			trace ("Binding IN binary parameter (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", SQLtype=" + SQLtype);
			dumpByte (value, value.length);
		}

		errorCode = new byte[1];
		bindInParameterBinary (hStmt, ipar,
				OdbcDef.jdbcTypeToOdbc (SQLtype),
                                value, precision, dataBuf, lenBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterDate
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterDate (
			int hStmt,
			int ipar,
			java.sql.Date value,
                        byte dataBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN parameter date (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", rgbValue=" + value.toString ());
		}

		errorCode = new byte[1];

		bindInParameterDate (hStmt, ipar, value.getYear () + 1900,
				value.getMonth () + 1, value.getDate (),
                                dataBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterDouble
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterDouble (
			int hStmt,
			int ipar,
			int SQLtype,
			int scale,
			double value,
                        byte dataBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN parameter double (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", SQLtype=" + SQLtype + ", scale=" + scale + ", rgbValue=" + value);
		}

		errorCode = new byte[1];
		bindInParameterDouble (hStmt, ipar,
				OdbcDef.jdbcTypeToOdbc (SQLtype),
                                scale, value, dataBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterFloat
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterFloat (
			int hStmt,
			int ipar,
			int SQLtype,
			int scale,
			float value,
                        byte dataBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN parameter float (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", SQLtype=" + SQLtype + ", scale=" + scale + ", rgbValue=" + value);
		}

		errorCode = new byte[1];
		bindInParameterFloat (hStmt, ipar,
				OdbcDef.jdbcTypeToOdbc (SQLtype),
                                scale, value, dataBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterInteger
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterInteger (
			int hStmt,
			int ipar,
			int SQLtype,
			int value,
                        byte dataBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN parameter integer (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + "SQLtype=" + SQLtype + ", rgbValue=" + value);
		}

		errorCode = new byte[1];
		bindInParameterInteger (hStmt, ipar,
				OdbcDef.jdbcTypeToOdbc (SQLtype),
                                value, dataBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterNull
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterNull (
			int hStmt,
			int ipar,
			int SQLtype,
                        byte lenBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN NULL parameter (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", SQLtype=" + SQLtype);
		}

		errorCode = new byte[1];
                bindInParameterNull (hStmt, ipar, SQLtype, lenBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterString
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterString (
			int hStmt,
			int ipar,
			int SQLtype,
			int precision,
			int scale,
                        byte value[],
                        byte dataBuf[],
                        int buffers[])
                        throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN string parameter (SQLBindParameter), hStmt=" +
					hStmt + ", ipar=" + ipar + ", SQLtype=" + SQLtype +
					", precision=" + precision + ", scale=" + scale + 
					", rgbValue=" + value);
		}

		errorCode = new byte[1];
		bindInParameterString (hStmt, ipar,
				OdbcDef.jdbcTypeToOdbc (SQLtype),
                                value, precision, scale, dataBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterTime
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterTime (
			int hStmt,
			int ipar,
			java.sql.Time value,
                        byte dataBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN parameter time (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", rgbValue=" + value.toString ());
		}

		errorCode = new byte[1];

		bindInParameterTime (hStmt, ipar, value.getHours (),
				value.getMinutes (), value.getSeconds (),
                                dataBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindInParameterTimestamp
	//--------------------------------------------------------------------

	public synchronized void SQLBindInParameterTimestamp (
			int hStmt,
			int ipar,
			java.sql.Timestamp value,
                        byte dataBuf[],
                        int buffers[])
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding IN parameter timestamp (SQLBindParameter), hStmt=" + hStmt + ", ipar=" + ipar + ", rgbValue=" + value.toString ());
		}

		errorCode = new byte[1];

		bindInParameterTimestamp (hStmt, ipar, value.getYear () + 1900,
				value.getMonth () + 1, value.getDate (),
				value.getHours (), value.getMinutes (),
				value.getSeconds (), value.getNanos (), 
                                dataBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBindOutParameterString
	//--------------------------------------------------------------------

	public synchronized void SQLBindOutParameterString (
			int hStmt,
			int ipar,
			int SQLtype,
			int scale,
			byte dataBuf[],
                        byte lenBuf[],
                        int [] buffers)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Binding OUT string parameter (SQLBindParameter), hStmt=" +
					hStmt + ", ipar=" + ipar + ", SQLtype=" + SQLtype +
					", prec=" + (dataBuf.length - 1) + ", scale=" + scale);
		}

		errorCode = new byte[1];

                bindOutParameterString (hStmt, ipar, SQLtype, scale, dataBuf,
                                lenBuf, errorCode, buffers);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLBrowseConnect
	// Returns the attributes and attribute values still needed to 
	// connect.  Returns null if a successful connection has been
	// established
	//--------------------------------------------------------------------

	public synchronized String SQLBrowseConnect (
			int hDbc,
			String connectString)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];
		byte	connStrOut[];
		String	attributes = null;

		if (isTracing ()) {
			trace ("Connecting (SQLBrowseConnect), hDbc=" + hDbc + ", szConnStrIn=" + connectString);
		}

		errorCode = new byte[1];
		connStrOut = new byte[JdbcOdbcLimits.MAX_BROWSE_RESULT_LENGTH];

		byte[] bConnectString = null; 
		char[] cConnectString = null;
		if (connectString != null)
			cConnectString = connectString.toCharArray();
		try {
			if (connectString != null)
				bConnectString = CharsToBytes (charSet, cConnectString);
		} catch (java.io.UnsupportedEncodingException exx) {}

		browseConnect (hDbc, bConnectString, connStrOut, errorCode);

		// Required/optional attributes needed

		if (errorCode[0] == OdbcDef.SQL_NEED_DATA) {
                    attributes = new String (connStrOut);
			attributes = attributes.trim ();
			errorCode[0] = 0;
		}

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}

		if (isTracing ()) {
			trace ("Attributes=" + attributes);
		}

		return attributes;
	}


	//--------------------------------------------------------------------
	// SQLCancel
	//--------------------------------------------------------------------

	public synchronized void SQLCancel (
			int hStmt)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Cancelling (SQLCancel), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		cancel (hStmt, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLColAttributes
	//--------------------------------------------------------------------

	public synchronized int SQLColAttributes (
			int hStmt,
			int column,
			int type)
			throws SQLException, JdbcOdbcSQLWarning
	{
		int	value = 0;
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Column attributes (SQLColAttributes), hStmt=" + hStmt + ", icol=" + column + ", type=" + type);
		}

		errorCode = new byte[1];
		value = colAttributes (hStmt, column, type, errorCode);

		if (errorCode[0] != 0) {
			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {
				if (isTracing ()) {
					trace ("value (int)=" + value);
				}

				// If we caught a warning, we need to save
				// off the original return value

				ex.value = (Object) BigDecimal.valueOf(value);

				// Re-throw the warning, with the saved value

				throw ex;
			}

		}
		else {
			if (isTracing ()) {
				trace ("value (int)=" + value);
			}
		}

		return value;
	}

	//--------------------------------------------------------------------
	// SQLColAttributesString
	//--------------------------------------------------------------------

	public synchronized String SQLColAttributesString (
			int hStmt,
			int column,
			int type)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	rgbDesc[];
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Column attributes (SQLColAttributes), hStmt=" + hStmt + ", icol=" + column + ", type=" + type);
		}

		errorCode = new byte[1];
		rgbDesc = new byte[JdbcOdbcLimits.DEFAULT_BUFFER_LENGTH];
		colAttributesString (hStmt, column, type, rgbDesc, errorCode);

		if (errorCode[0] != 0) {
			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {

				// If we caught a warning, we need to save
				// off the original return value

				String sValue=new String();
				try {
					sValue = BytesToChars (charSet, rgbDesc);
				} catch (java.io.UnsupportedEncodingException e) {
					System.out.println (e);
				}

				if (isTracing ()) {
					trace ("value (String)=" + 
						sValue.trim ());
				}
				ex.value = sValue.trim ();

				// Re-throw the warning, with the saved value

				throw ex;
			}

		}
		String sValue=new String();
		try {
			sValue = BytesToChars (charSet, rgbDesc);
		} catch (java.io.UnsupportedEncodingException e) {
			System.out.println (e);
		}

		if (isTracing ()) {
			trace ("value (String)=" + sValue.trim ());
		}
		return sValue.trim ();
	}

	//--------------------------------------------------------------------
	// SQLColumns
	//--------------------------------------------------------------------

	public synchronized void SQLColumns (
			int hStmt,
			String catalog,
			String schema,
			String table,
                    String column)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("(SQLColumns), hStmt=" + hStmt + ", catalog=" + catalog + ", schema=" + schema + ", table=" + table + ", column=" + column);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bTable=null;
		byte[] bColumn=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cTable=null;
		char[] cColumn=null;  
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (table != null)
			cTable=table.toCharArray();
		if (column != null)
			cColumn=column.toCharArray();  
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (table != null)
				bTable = CharsToBytes (charSet, cTable);
			if (column != null)
				bColumn = CharsToBytes (charSet, cColumn);
		} catch (java.io.UnsupportedEncodingException exx) {}


		columns (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bTable, (table == null),
                        bColumn, (column == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLColumnPrivileges
	//--------------------------------------------------------------------

	public synchronized void SQLColumnPrivileges (
			int hStmt,
			String catalog,
			String schema,
			String table,
			String column)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("(SQLColumnPrivileges), hStmt=" + hStmt + ", catalog=" + catalog + ", schema=" + schema + ", table=" + table + ", column=" + column);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bTable=null;
		byte[] bColumn=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cTable=null;
		char[] cColumn=null;  
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (table != null)
			cTable=table.toCharArray();
		if (column != null)
			cColumn=column.toCharArray();  
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (table != null)
				bTable = CharsToBytes (charSet, cTable);
			if (column != null)
				bColumn = CharsToBytes (charSet, cColumn);
		} catch (java.io.UnsupportedEncodingException exx) {}

		columnPrivileges (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bTable, (table == null),
                        bColumn, (column == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLDescribeParamNullable
	//--------------------------------------------------------------------

	public synchronized boolean SQLDescribeParamNullable (
			int hStmt,
			int param)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	value;
		boolean nullable = false;

		if (isTracing ()) {
			trace ("Parameter nullable (SQLDescribeParam), hStmt=" + hStmt + ", ipar=" + param);
		}

		errorCode = new byte[1];

		// Call SQLDescribeParam and return 4th output parameter
		// (nullable)

		value = describeParam (hStmt, param, 4, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}

		if (value == OdbcDef.SQL_NULLABLE) {
			nullable = true;
		}
		
		if (isTracing ()) {
			trace ("nullable=" + nullable);
		}

		return nullable;
	}
	//--------------------------------------------------------------------
	// SQLDescribeParamPrecision
	//--------------------------------------------------------------------

	public synchronized int SQLDescribeParamPrecision (
			int hStmt,
			int param)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	precision;

		if (isTracing ()) {
			trace ("Parameter precision (SQLDescribeParam), hStmt=" + hStmt + ", ipar=" + param);
		}

		errorCode = new byte[1];

		// Call SQLDescribeParam and return 2nd output parameter
		// (precision)

		precision = describeParam (hStmt, param, 2, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}

		if (isTracing ()) {
			trace ("precision=" + precision);
		}
		
		return precision;
	}

	//--------------------------------------------------------------------
	// SQLDescribeParamScale
	//--------------------------------------------------------------------

	public synchronized int SQLDescribeParamScale (
			int hStmt,
			int param)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	scale;

		if (isTracing ()) {
			trace ("Parameter scale (SQLDescribeParam), hStmt=" + hStmt + ", ipar=" + param);
		}

		errorCode = new byte[1];

		// Call SQLDescribeParam and return 3rd output parameter
		// (scale)

		scale = describeParam (hStmt, param, 3, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}

		if (isTracing ()) {
			trace ("scale=" + scale);
		}

		return scale;
	}

	//--------------------------------------------------------------------
	// SQLDescribeParamType
	//--------------------------------------------------------------------

	public synchronized int SQLDescribeParamType (
			int hStmt,
			int param)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	type;

		if (isTracing ()) {
			trace ("Parameter type (SQLDescribeParam), hStmt=" + hStmt + ", ipar=" + param);
		}

		errorCode = new byte[1];

		// Call SQLDescribeParam and return 1st output parameter
		// (type)

		type = describeParam (hStmt, param, 1, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}

		if (isTracing ()) {
			trace ("type=" + type);
		}

		return type;
	}

	//--------------------------------------------------------------------
	// SQLDisconnect
	//--------------------------------------------------------------------

	public synchronized void SQLDisconnect (
			int hDbc)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Disconnecting (SQLDisconnect), hDbc=" + hDbc);
		}

		errorCode = new byte[1];
		disconnect (hDbc, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}
	}

	//--------------------------------------------------------------------
	// SQLDriverConnect
	//--------------------------------------------------------------------

	public synchronized void SQLDriverConnect (
			int hDbc,
			String connectString)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Connecting (SQLDriverConnect), hDbc=" + hDbc + ", szConnStrIn=" + connectString);
		}

		errorCode = new byte[1];
		byte[] bConnectString=null;
		char[] cConnectString=null;

		if (connectString != null)
			cConnectString=connectString.toCharArray();

		try {
			if (connectString != null)
				bConnectString = CharsToBytes (charSet, cConnectString);

		} catch (java.io.UnsupportedEncodingException exx){}

		driverConnect (hDbc, bConnectString, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}
	}

	//--------------------------------------------------------------------
	// SQLExecDirect
	//--------------------------------------------------------------------
	public synchronized void SQLExecDirect (
			int hStmt,
			String query)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Executing (SQLExecDirect), hStmt=" + hStmt + ", szSqlStr=" + query);
		}

		errorCode = new byte[1];

		byte[] bQuery=null;
		char[] cQuery=null;
		if (query != null)
			cQuery=query.toCharArray();
		try {
			if (query != null)
				bQuery = CharsToBytes (charSet, cQuery);
		} catch (java.io.UnsupportedEncodingException exx){}
			
		execDirect (hStmt, bQuery, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV, 
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLExecute
	// Returns true if more data is required (i.e. there were data-at-
	// execution parameters)
	//--------------------------------------------------------------------
	public synchronized boolean SQLExecute (
			int hStmt)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];
		boolean needData = false;

		if (isTracing ()) {
			trace ("Executing (SQLExecute), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		execute (hStmt, errorCode);

		// If SQL_NEED_DATA (data-at-execution parameters) is
		// returned, set the return code.

		if (errorCode[0] == OdbcDef.SQL_NEED_DATA) {
			if (isTracing ()) {
				trace ("SQL_NEED_DATA returned");
			}
			needData = true;
			errorCode[0] = 0;
		}

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
		return needData;
	}

	//--------------------------------------------------------------------
	// SQLFetch
	//--------------------------------------------------------------------
	public synchronized boolean SQLFetch (
			int hStmt)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];
		boolean rc = true;

		if (isTracing ()) {
			trace ("Fetching (SQLFetch), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		fetch (hStmt, errorCode);

		// Check for SQL_NO_DATA_FOUND, indicating that this is the
		// end of the result set.  Set the return code to false and
		// clear the error

		if (errorCode[0] == OdbcDef.SQL_NO_DATA_FOUND) {
			rc = false;
			errorCode[0] = 0;
			if (isTracing ()) {
				trace ("End of result set (SQL_NO_DATA)");
			}
		}

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
		return rc;
	}

	//--------------------------------------------------------------------
	// SQLForeignKeys
	//--------------------------------------------------------------------

	public synchronized void SQLForeignKeys (
			int hStmt,
			String PKcatalog,
			String PKschema,
			String PKtable,
			String FKcatalog,
			String FKschema,
			String FKtable)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("(SQLForeignKeys), hStmt=" + hStmt + ", Pcatalog=" + PKcatalog + ", Pschema=" + PKschema + ", Ptable=" + PKtable + ", Fcatalog=" + FKcatalog + ", Fschema=" + FKschema + ", Ftable=" + FKtable);
		}

		errorCode = new byte[1];

		byte[] bPKCatalog=null;
		byte[] bPKSchema=null;
		byte[] bPKTable=null;
		byte[] bFKCatalog=null;
		byte[] bFKSchema=null;
		byte[] bFKTable=null;
		char[] cPKCatalog=null;
		char[] cPKSchema=null;
		char[] cPKTable=null;
		char[] cFKCatalog=null;
		char[] cFKSchema=null;   
		char[] cFKTable=null;
		if (PKcatalog != null)
			cPKCatalog=PKcatalog.toCharArray();
		if (PKschema != null)
			cPKSchema=PKschema.toCharArray();   
		if (PKtable != null)
			cPKTable=PKtable.toCharArray();
		if (FKcatalog != null)
			cFKCatalog=FKcatalog.toCharArray();
		if (FKschema != null)
			cFKSchema=FKschema.toCharArray();   
		if (FKtable != null)
			cFKTable=FKtable.toCharArray();
		try {
			if (PKcatalog != null)
				bPKCatalog = CharsToBytes (charSet, cPKCatalog);
			if (PKschema != null)
				bPKSchema = CharsToBytes (charSet, cPKSchema);
			if (PKtable != null)
				bPKTable = CharsToBytes (charSet, cPKTable);
			if (FKcatalog != null)
				bFKCatalog = CharsToBytes (charSet, cFKCatalog);
			if (FKschema != null)
				bFKSchema = CharsToBytes (charSet, cFKSchema);
			if (FKtable != null)
				bFKTable = CharsToBytes (charSet, cFKTable);
		} catch (java.io.UnsupportedEncodingException exx) {}

		foreignKeys (hStmt,
                        bPKCatalog, (PKcatalog == null),
                        bPKSchema, (PKschema == null),
                        bPKTable, (PKtable == null),
                        bFKCatalog, (FKcatalog == null),
                        bFKSchema, (FKschema == null),
                        bFKTable, (FKtable == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLFreeConnect
	//--------------------------------------------------------------------

	public synchronized void SQLFreeConnect (
			int hDbc)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Closing connection (SQLFreeConnect), hDbc=" + hDbc);
		}

		errorCode = new byte[1];
		freeConnect (hDbc, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}
	}

	//--------------------------------------------------------------------
	// SQLFreeEnv
	//--------------------------------------------------------------------

	public synchronized void SQLFreeEnv (
			int hEnv)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Closing environment (SQLFreeEnv), hEnv=" + hEnv);
		}

		errorCode = new byte[1];
		freeEnv (hEnv, errorCode);

		if (errorCode[0] != 0) {
			throwGenericSQLException ();
		}
	}

	//--------------------------------------------------------------------
	// SQLFreeStmt
	//--------------------------------------------------------------------

	public synchronized void SQLFreeStmt (
			int hStmt,
			int fOption)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Free statement (SQLFreeStmt), hStmt=" + hStmt + ", fOption=" + fOption);
		}

		errorCode = new byte[1];
		freeStmt (hStmt, fOption, errorCode);

		if (errorCode[0] != 0) {
			throwGenericSQLException ();
		}
	}

	//--------------------------------------------------------------------
	// SQLGetConnectOption
	//--------------------------------------------------------------------

	public synchronized int SQLGetConnectOption (
			int   hDbc,
			short fOption)
			throws SQLException
	{
		byte	errorCode[];
		int	vParam;

		if (isTracing ()) {
			trace ("Connection Option (SQLGetConnectOption), hDbc=" + hDbc + ", fOption=" + fOption);
		}
		
		errorCode = new byte[1];
		vParam = getConnectOption (hDbc, fOption, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}

		if (isTracing ()) {
			trace ("option value (int)=" + vParam);
		}

		return vParam;
	}

	//--------------------------------------------------------------------
	// SQLGetConnectOptionString
	//--------------------------------------------------------------------

	public synchronized String SQLGetConnectOptionString (
			int   hDbc,
			short fOption)
			throws SQLException
	{
		byte	errorCode[];
		byte	szParam[];

		if (isTracing ()) {
			trace ("Connection Option (SQLGetConnectOption), hDbc=" + hDbc + ", fOption=" + fOption);
		}

		errorCode = new byte[1];
		szParam = new byte[JdbcOdbcLimits.DEFAULT_BUFFER_LENGTH];

		getConnectOptionString (hDbc, fOption, szParam, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}

		String  sValue = new String ();
		try {
			sValue = BytesToChars (charSet, szParam);
		} catch (java.io.UnsupportedEncodingException exxxx) {}

		if (isTracing ()) {
			trace ("option value (int)=" + sValue.trim ());
		}

		return sValue.trim ();
	}

	//--------------------------------------------------------------------
	// SQLGetCursorName
	//--------------------------------------------------------------------

	public synchronized String SQLGetCursorName (
			int hStmt)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	szCursor[];
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Cursor name (SQLGetCursorName), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		szCursor = new byte[JdbcOdbcLimits.DEFAULT_BUFFER_LENGTH];
		getCursorName (hStmt, szCursor, errorCode);

		if (errorCode[0] != 0) {
			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {

				// If we caught a warning, we need to save
				// off the original return value

				String  sValue = new String ();
				try {
					sValue = BytesToChars (charSet, szCursor);
				} catch (java.io.UnsupportedEncodingException exxxx) {}

				if (isTracing ()) {
					trace ("value=" + sValue.trim ());
				}
				ex.value = sValue.trim ();

				// Re-throw the warning, with the saved value

				throw ex;
			}

		}
		String  sValue = new String (szCursor);
		if (isTracing ()) {
			trace ("value=" + sValue.trim ());
		}
		return sValue.trim ();
	}

	//--------------------------------------------------------------------
	// SQLGetDataBinary
	// Returns number of bytes read, -1 for eof
	//--------------------------------------------------------------------
	public synchronized int SQLGetDataBinary (
			int hStmt,
			int column,
			byte b[])
			throws SQLException, JdbcOdbcSQLWarning
	{
		return SQLGetDataBinary (hStmt, column, OdbcDef.SQL_C_BINARY,
								b, b.length);
	}

	//--------------------------------------------------------------------
	// SQLGetDataBinary
	// Returns number of bytes read, -1 for eof
	//--------------------------------------------------------------------
	public synchronized int SQLGetDataBinary (
			int hStmt,
			int column,
			int cType,
			byte b[],
			int length)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	bytesRead = 0;

		if (isTracing ()) {
			trace ("Get binary data (SQLGetData), hStmt=" + hStmt +
					", column=" + column + ", type=" + cType + 
					", length=" + length);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
		bytesRead = getDataBinary (hStmt, column, cType, b, length, errorCode);

		// If no data was found, return a -1

		if (errorCode[0] == OdbcDef.SQL_NO_DATA) {
			bytesRead = -1;
			errorCode[0] = 0;
		}

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				if (isTracing ()) {
					if (bytesRead == -1) {
						trace ("NULL");
					}
					else {
						if (isTracing ()) {
							trace ("Bytes: " + bytesRead);
						}
//						dumpByte (b, bytesRead);
					}
				}

				// If we caught a warning, we need to save
				// off the original return value

				ex.value = (Object) new Integer (bytesRead);
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		if (isTracing ()) {
			if (bytesRead == -1) {
				trace ("NULL");
			}
			else {
				if (isTracing ()) {
					trace ("Bytes: " + bytesRead);
				}
//				dumpByte (b, bytesRead);
			}
		}
		return bytesRead;
	}

	//--------------------------------------------------------------------
	// SQLGetDataDouble
	//--------------------------------------------------------------------
	public synchronized Double SQLGetDataDouble (
			int hStmt,
			int column)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		double	value;

		if (isTracing ()) {
			trace ("Get double data (SQLGetData), hStmt=" + hStmt + ", column=" + column);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
		value = getDataDouble (hStmt, column, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				// If we caught a warning, we need to save
				// off the original return value

				if (errorCode[1] == 0) {
					if (isTracing ()) {
						trace ("value=" + value);
					}

					// The value is not null
					ex.value = (Object) new Double (
							value);
				}
				else {
					if (isTracing ()) {
						trace ("NULL");
					}

					// The value is null
					ex.value = null;
				}
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		
		// Check for a NULL value

		if (errorCode[1] == 0) {
			if (isTracing ()) {
				trace ("value=" + value);
			}
			// The value is not null
			return new Double (value);
		}
		else {
			if (isTracing ()) {
				trace ("NULL");
			}
			// The value is null
			return null;
		}
	}

	//--------------------------------------------------------------------
	// SQLGetDataFloat
	//--------------------------------------------------------------------
	public synchronized Float SQLGetDataFloat (
			int hStmt,
			int column)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		float	value;

		if (isTracing ()) {
			trace ("Get float data (SQLGetData), hStmt=" + hStmt + ", column=" + column);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
		value = getDataFloat (hStmt, column, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				// If we caught a warning, we need to save
				// off the original return value

				if (errorCode[1] == 0) {
					if (isTracing ()) {
						trace ("value=" + value);
					}
					// The value is not null
					ex.value = (Object) new Float (
							value);
				}
				else {
					if (isTracing ()) {
						trace ("NULL");
					}
					// The value is null
					ex.value = null;
				}
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		
		// Check for a NULL value

		if (errorCode[1] == 0) {
			// The value is not null
			if (isTracing ()) {
				trace ("value=" + value);
			}
			return new Float (value);
		}
		else {
			if (isTracing ()) {
				trace ("NULL");
			}
			// The value is null
			return null;
		}
	}

	//--------------------------------------------------------------------
	// SQLGetDataInteger
	//--------------------------------------------------------------------
	public synchronized Integer SQLGetDataInteger (
			int hStmt,
			int column)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	value;

		if (isTracing ()) {
			trace ("Get integer data (SQLGetData), hStmt=" + hStmt + ", column=" + column);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
		value = getDataInteger (hStmt, column, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				// If we caught a warning, we need to save
				// off the original return value

				if (errorCode[1] == 0) {
					if (isTracing ()) {
						trace ("value=" + value);
					}
					// The value is not null
					ex.value = (Object) new Integer (
							value);
				}
				else {
					if (isTracing ()) {
						trace ("NULL");
					}
					// The value is null
					ex.value = null;
				}
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		
		// Check for a NULL value

		if (errorCode[1] == 0) {
			if (isTracing ()) {
				trace ("value=" + value);
			}
			// The value is not null
			return new Integer (value);
		}
		else {
			if (isTracing ()) {
				trace ("NULL");
			}
			// The value is null
			return null;
		}
	}

	//--------------------------------------------------------------------
	// SQLGetDataString
	//--------------------------------------------------------------------
	public synchronized String SQLGetDataString (
			int hStmt,
			int column,
			int maxLen,
			boolean trimString)
                        throws SQLException, JdbcOdbcSQLWarning
                {
		byte	errorCode[];
		byte	rgbValue[];

		if (isTracing ()) {
			trace ("Get string data (SQLGetData), hStmt=" + hStmt + ", column=" + column + ", maxLen=" + maxLen);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
                rgbValue = new byte[maxLen];

		int rlen = getDataString (hStmt, column, rgbValue, errorCode);
	
		// If we read 0 bytes, treat it as a null

		if (rlen == 0) {
			errorCode[1] = 1;
		}

		// Sanity check the return length.  For some drivers, the length is
		// returned as the total length of the column when the data was
		// truncated.  Set the length to be the maximum length of the
		// buffer in this case.

		if (rlen > maxLen) {
                        rlen = maxLen;
		}

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				// If we caught a warning, we need to save
				// off the original return value

				if (errorCode[1] == 0) {

					// The value is not null.  If we know the return length,
					// create the String the proper size
					char [] csValue = new char[rlen];
					String sValue=new String();
					if (rlen > 0) {
						try { 
							sValue=BytesToChars(charSet, rgbValue);
						} catch (java.io.UnsupportedEncodingException e){
							System.out.println(e);
						}
					}
					else {
						try { 
							sValue=BytesToChars(charSet, rgbValue);
						} catch (java.io.UnsupportedEncodingException e){
							System.out.println(e);
						}
					}

					if (isTracing ()) {
						trace (sValue.trim ());
					}
					if (trimString ) {
						ex.value = sValue.trim ();
					}
					else {
						ex.value = sValue;
					}
				}
				else {
					if (isTracing ()) {
						trace ("NULL");
					}
					// The value is null
					ex.value = null;
				}
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		
		// Check for a NULL value

		if (errorCode[1] == 0) {
			// The value is not null.  If we know the return length,
			// create the String the proper size

			String sValue=new String();
			char [] csValue = new char[rlen];
			if (rlen > 0)  {
				try {
					sValue=BytesToChars(charSet, rgbValue);
				} catch (java.io.UnsupportedEncodingException e){
					System.out.println(e);
				}
			}
			else {
				try {
					sValue=BytesToChars(charSet, rgbValue);
				} catch (java.io.UnsupportedEncodingException e){
					System.out.println(e);
				}
			}

			if (isTracing ()) {
				trace (sValue.trim ());
			}

			// If the string should be trimmed, so it now

			if (trimString) {
				return sValue.trim ();
			}
			else {
				return sValue;
			}
		}
		else {
			if (isTracing ()) {
				trace ("NULL");
			}
			return null;
		}
	}

	//--------------------------------------------------------------------
	// SQLGetDataStringDate
	//--------------------------------------------------------------------
	public synchronized String SQLGetDataStringDate (
			int hStmt,
			int column)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		byte	rgbValue[];

		if (isTracing ()) {
			trace ("Get date data (SQLGetData), hStmt=" +
					hStmt + ", column=" + column);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
		rgbValue = new byte[11];

		getDataStringDate (hStmt, column, rgbValue, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				// If we caught a warning, we need to save
				// off the original return value

				if (errorCode[1] == 0) {
					// The value is not null
					String  sValue = new String ();
					try {
						sValue = BytesToChars(charSet, rgbValue);
					} catch (java.io.UnsupportedEncodingException exx) {}

					if (isTracing ()) {
						trace (sValue.trim ());
					}
					ex.value = sValue.trim ();
				}
				else {
					if (isTracing ()) {
						trace ("NULL");
					}
					// The value is null
					ex.value = null;
				}
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		
		// Check for a NULL value

		if (errorCode[1] == 0) {
			String  sValue = new String ();
			try {
				sValue = BytesToChars(charSet, rgbValue);
			} catch (java.io.UnsupportedEncodingException exx) {}

			if (isTracing ()) {
				trace (sValue.trim ());
			}

			// If the string should be trimmed, so it now

			return sValue.trim ();
		}
		else {
			if (isTracing ()) {
				trace ("NULL");
			}
			return null;
		}
	}

	//--------------------------------------------------------------------
	// SQLGetDataStringTime
	//--------------------------------------------------------------------
	public synchronized String SQLGetDataStringTime (
			int hStmt,
			int column)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		byte	rgbValue[];

		if (isTracing ()) {
			trace ("Get time data (SQLGetData), hStmt=" +
					hStmt + ", column=" + column);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
		rgbValue = new byte[9];

		getDataStringTime (hStmt, column, rgbValue, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				// If we caught a warning, we need to save
				// off the original return value

				if (errorCode[1] == 0) {
					// The value is not null
					String  sValue = new String ();
					try {
						sValue = BytesToChars(charSet, rgbValue);
					} catch (java.io.UnsupportedEncodingException exx) {}

					if (isTracing ()) {
						trace (sValue.trim ());
					}
					ex.value = sValue.trim ();
				}
				else {
					if (isTracing ()) {
						trace ("NULL");
					}
					// The value is null
					ex.value = null;
				}
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		
		// Check for a NULL value

		if (errorCode[1] == 0) {
			String  sValue = new String ();
			try {
				sValue = BytesToChars(charSet, rgbValue);
			} catch (java.io.UnsupportedEncodingException exx) {}

			if (isTracing ()) {
				trace (sValue.trim ());
			}

			// If the string should be trimmed, so it now

			return sValue.trim ();
		}
		else {
			if (isTracing ()) {
				trace ("NULL");
			}
			return null;
		}
	}

	//--------------------------------------------------------------------
	// SQLGetDataStringTimestamp
	//--------------------------------------------------------------------
	public synchronized String SQLGetDataStringTimestamp (
			int hStmt,
			int column)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		byte	rgbValue[];

		if (isTracing ()) {
			trace ("Get timestamp data (SQLGetData), hStmt=" +
					hStmt + ", column=" + column);
		}

		// Use a 2-byte error code.  The first byte is for the
		// actual SQL return code, the second byte is a null-value
		// indicator (0=not null)

		errorCode = new byte[2];
		rgbValue = new byte[30];

		getDataStringTimestamp (hStmt, column, rgbValue, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			
			catch (JdbcOdbcSQLWarning ex) {
			
				// If we caught a warning, we need to save
				// off the original return value

				if (errorCode[1] == 0) {
					// The value is not null
					String  sValue = new String ();
					try {
						sValue = BytesToChars(charSet, rgbValue);
					} catch (java.io.UnsupportedEncodingException exx) {}

					if (isTracing ()) {
						trace (sValue.trim ());
					}
					ex.value = sValue.trim ();
				}
				else {
					if (isTracing ()) {
						trace ("NULL");
					}
					// The value is null
					ex.value = null;
				}
				
				// Re-throw the warning, with the saved value
				throw ex;
			}
		}
		
		// Check for a NULL value

		if (errorCode[1] == 0) {
			String  sValue = new String ();
			try {
				sValue = BytesToChars(charSet, rgbValue);
			} catch (java.io.UnsupportedEncodingException exx) {}

			if (isTracing ()) {
				trace (sValue.trim ());
			}

			// If the string should be trimmed, so it now

			return sValue.trim ();
		}
		else {
			if (isTracing ()) {
				trace ("NULL");
			}
			return null;
		}
	}

	//--------------------------------------------------------------------
	// SQLGetInfo
	//--------------------------------------------------------------------

	public synchronized int SQLGetInfo (
			int   hDbc,
			short fInfoType)
			throws SQLException
	{
		byte	errorCode[];
		int	infoValue;

		if (isTracing ()) {
			trace ("Get connection info (SQLGetInfo), hDbc=" +
				hDbc + ", fInfoType=" + fInfoType);
		}

		errorCode = new byte[1];
		infoValue = getInfo (hDbc, fInfoType, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}
		if (isTracing ()) {
			trace (" int value=" + infoValue);
		}
		return infoValue;
	}

	//--------------------------------------------------------------------
	// SQLGetInfoShort
	//--------------------------------------------------------------------

	public synchronized int SQLGetInfoShort (
			int   hDbc,
			short fInfoType)
			throws SQLException
	{
		byte	errorCode[];
		int	infoValue;

		if (isTracing ()) {
			trace ("Get connection info (SQLGetInfo), hDbc=" +
				hDbc + ", fInfoType=" + fInfoType);
		}

		errorCode = new byte[1];
		infoValue = getInfoShort (hDbc, fInfoType, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}
		if (isTracing ()) {
			trace (" short value=" + infoValue);
		}
		return infoValue;
	}

	//--------------------------------------------------------------------
	// SQLGetInfoString
	//--------------------------------------------------------------------

	public synchronized String SQLGetInfoString (
			int   hDbc,
			short fInfoType)
			throws SQLException
	{
		// Get info with result buffer
                 return SQLGetInfoString (hDbc, fInfoType,
				JdbcOdbcLimits.DEFAULT_BUFFER_LENGTH);
         	}

	//--------------------------------------------------------------------
	// SQLGetInfoString
	// Optional interface to supply the size of the result buffer
	//--------------------------------------------------------------------

	public synchronized String SQLGetInfoString (
			int   hDbc,
			short fInfoType,
			int   buffSize)
			throws SQLException
	{
		byte	errorCode[];
		byte	szParam[];
		String	vParam;

		if (isTracing ()) {
			trace ("Get connection info string (SQLGetInfo), hDbc=" + hDbc + ", fInfoType=" + fInfoType + ", len=" + buffSize);
		}

		errorCode = new byte[1];
		szParam = new byte[buffSize];

		getInfoString (hDbc, fInfoType, szParam, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}

		String  sValue = new String ();
		try {
			sValue = BytesToChars (charSet, szParam);
		} catch (java.io.UnsupportedEncodingException exx) {}

		if (isTracing ()) {
			trace (sValue.trim ());
		}
		return sValue.trim ();
	}

	//--------------------------------------------------------------------
	// SQLGetStmtOption
	//--------------------------------------------------------------------

	public synchronized int SQLGetStmtOption (
			int hStmt,
			short fOption)
			throws SQLException, JdbcOdbcSQLWarning
	{
		int	value = 0;
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Get statement option (SQLGetStmtOption), hStmt=" + hStmt + ", fOption=" + fOption);
		}

		errorCode = new byte[1];
		value = getStmtOption (hStmt, fOption, errorCode);

		if (errorCode[0] != 0) {
			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {

				if (isTracing ()) {
					trace ("value=" + value);
				}
				// If we caught a warning, we need to save
				// off the original return value

				ex.value = (Object) BigDecimal.valueOf(value);

				// Re-throw the warning, with the saved value

				throw ex;
			}

		}
		if (isTracing ()) {
			trace ("value=" + value);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// SQLGetTypeInfo
	//--------------------------------------------------------------------

	public synchronized void SQLGetTypeInfo (
			int   hStmt,
			short fSqlType)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Get type info (SQLGetTypeInfo), hStmt=" + hStmt + ", fSqlType=" + fSqlType);
		}

		errorCode = new byte[1];
		getTypeInfo (hStmt, fSqlType, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLMoreResults
	//--------------------------------------------------------------------
	public synchronized boolean SQLMoreResults (
			int hStmt)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];
		boolean rc = true;

		if (isTracing ()) {
			trace ("Get more results (SQLMoreResults), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		moreResults (hStmt, errorCode);

		// Check for SQL_NO_DATA_FOUND, indicating that this is the
		// end of the result sets.  Set the return code to false and
		// clear the error

		if (errorCode[0] == OdbcDef.SQL_NO_DATA_FOUND) {
			rc = false;
			errorCode[0] = 0;	
		}

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
		if (isTracing ()) {
			trace ("More results: " + rc);
		}
		return rc;
	}

	//--------------------------------------------------------------------
	// SQLNativeSql
	//--------------------------------------------------------------------

	public synchronized String SQLNativeSql (
			int hDbc,
			String query)
			throws SQLException
	{
		byte	errorCode[];
		byte	nativeQuery[];

		errorCode = new byte[1];

		// Make some estimate as to how large the convert SQL string
		// might be.

		int nativeLen = JdbcOdbcLimits.DEFAULT_NATIVE_SQL_LENGTH;
                  if (query.length () * 4 > nativeLen) {
                        nativeLen = query.length () * 4;
			if (nativeLen > 32768) {
				nativeLen = 32768;
                        } 
		}

		if (isTracing ()) {
			trace ("Convert native SQL (SQLNativeSql), hDbc=" + hDbc + ", nativeLen=" + nativeLen + ", SQL=" + query);
		}

		nativeQuery = new byte[nativeLen];
		byte[] bQuery = null;
		char[] cQuery = null;
		if (query != null)
			cQuery = query.toCharArray();
		try {
			if (query != null)
				bQuery = CharsToBytes (charSet, cQuery);
		} catch (java.io.UnsupportedEncodingException exxx){}

		nativeSql (hDbc, bQuery, nativeQuery, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}

		String  sValue = new String ();
		try {
			sValue = BytesToChars (charSet, nativeQuery);
		} catch (java.io.UnsupportedEncodingException exx){}

		if (isTracing ()) {
			trace ("Native SQL=" + sValue.trim ());
		}
		return sValue.trim ();
	}

	//--------------------------------------------------------------------
	// SQLNumParams
	//--------------------------------------------------------------------
	public synchronized int SQLNumParams (
			int   hStmt)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	numParams = 0;

		if (isTracing ()) {
			trace ("Number of parameter markers (SQLNumParams), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		numParams = numParams (hStmt, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
		if (isTracing ()) {
			trace ("value=" + numParams);
		}
		return numParams;
	}

	//--------------------------------------------------------------------
	// SQLNumResultCols
	//--------------------------------------------------------------------
	public synchronized int SQLNumResultCols (
			int   hStmt)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	numCols = 0;

		if (isTracing ()) {
			trace ("Number of result columns (SQLNumResultCols), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		numCols = numResultCols (hStmt, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {
				if (isTracing ()) {
					trace ("value=" + numCols);
				}

				// If we caught a warning, we need to save
				// off the original return value

				ex.value = (Object)BigDecimal.valueOf(numCols);

				// Re-throw the warning, with the saved value

				throw ex;
			}
		}
		if (isTracing ()) {
			trace ("value=" + numCols);
		}
		return numCols;
	}

	//--------------------------------------------------------------------
	// SQLParamData
	// Returns -1 if no more data is needed
	//--------------------------------------------------------------------
	public synchronized int SQLParamData (
			int   hStmt)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	param = 0;

		if (isTracing ()) {
			trace ("Get parameter number (SQLParamData), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		param = paramData (hStmt, errorCode);

		// If we need data, return the parameter number that we got
		// This is not an error condition.

		if (errorCode[0] == OdbcDef.SQL_NEED_DATA) {
			errorCode[0] = 0;
		}
		else {

			// We don't need any more data, return -1 to signal
			// this
			param = -1;
		}

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}

		if (isTracing ()) {
			trace ("Parameter needing data=" + param);
		}

		return param;
	}

	//--------------------------------------------------------------------
	// SQLPrepare
	//--------------------------------------------------------------------
	public synchronized void SQLPrepare (
			int hStmt,
			String query)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Preparing (SQLPrepare), hStmt=" + hStmt + ", szSqlStr=" + query);
		}

		errorCode = new byte[1];

		byte[] bQuery = null;
		char[] cQuery = null;
		if (query != null)
			cQuery = query.toCharArray();
		try {
			if (query != null)
				bQuery = CharsToBytes (charSet, cQuery);
		} catch (java.io.UnsupportedEncodingException exxx){}

		prepare (hStmt, bQuery, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLPutData
	//--------------------------------------------------------------------

	public synchronized void SQLPutData (
			int hStmt,
			byte dataBuf[],
			int dataLen)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Putting data (SQLPutData), hStmt=" + hStmt + ", len=" + dataLen);
//			dumpByte (dataBuf, dataLen);
		}

		errorCode = new byte[1];
		putData (hStmt, dataBuf, dataLen, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			standardError (errorCode[0],
				OdbcDef.SQL_NULL_HENV,
				OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLPrimaryKeys
	//--------------------------------------------------------------------

	public synchronized void SQLPrimaryKeys (
			int hStmt,
			String catalog,
			String schema,
			String table)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Primary keys (SQLPrimaryKeys), hStmt=" + hStmt + ", catalog=" + catalog + ", schema=" + schema + ", table=" + table);
		}

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bTable=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cTable=null;
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (table != null)
			cTable=table.toCharArray();
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (table != null)
				bTable = CharsToBytes (charSet, cTable);
		} catch (java.io.UnsupportedEncodingException exx) {}

		errorCode = new byte[1];
		primaryKeys (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bTable, (table == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLProcedures
	//--------------------------------------------------------------------

	public synchronized void SQLProcedures (
			int hStmt,
			String catalog,
			String schema,
			String procedure)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Procedures (SQLProcedures), hStmt=" + hStmt + ", catalog=" + catalog + ", schema=" + schema + ", procedure=" + procedure);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bProcedure=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cProcedure=null;
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (procedure != null)
			cProcedure=procedure.toCharArray();
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (procedure != null)
				bProcedure = CharsToBytes (charSet, cProcedure);
		} catch (java.io.UnsupportedEncodingException exx) {}

		procedures (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bProcedure, (procedure == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLProcedureColumns
	//--------------------------------------------------------------------

	public synchronized void SQLProcedureColumns (
			int hStmt,
			String catalog,
			String schema,
			String procedure,
			String column)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Procedure columns (SQLProcedureColumns), hStmt=" + hStmt + ", catalog=" + catalog + ", schema=" + schema + ", procedure=" + procedure + ", column=" + column);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bProcedure=null;
		byte[] bColumn=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cProcedure=null;
		char[] cColumn=null;  
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (procedure != null)
			cProcedure=procedure.toCharArray();
		if (column != null)
			cColumn=column.toCharArray();  
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (procedure != null)
				bProcedure = CharsToBytes (charSet, cProcedure);
			if (column != null)
				bColumn = CharsToBytes (charSet, cColumn);
		} catch (java.io.UnsupportedEncodingException exx) {}

		procedureColumns (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bProcedure, (procedure == null),
                        bColumn, (column == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLRowCount
	//--------------------------------------------------------------------
	public synchronized int SQLRowCount (
			int   hStmt)
			throws SQLException, JdbcOdbcSQLWarning
	{
		byte	errorCode[];
		int	numRows = 0;

		if (isTracing ()) {
			trace ("Number of affected rows (SQLRowCount), hStmt=" + hStmt);
		}

		errorCode = new byte[1];
		numRows = rowCount (hStmt, errorCode);

		if (errorCode[0] != 0) {

			// Process the error
			try {
				standardError (errorCode[0],
					OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
			}
			catch (JdbcOdbcSQLWarning ex) {

				if (isTracing ()) {
					trace ("value=" + numRows);
				}

				// If we caught a warning, we need to save
				// off the original return value

				ex.value = (Object)BigDecimal.valueOf(numRows);

				// Re-throw the warning, with the saved value

				throw ex;
			}
		}
		if (isTracing ()) {
			trace ("value=" + numRows);
		}
		return numRows;
	}

	//--------------------------------------------------------------------
	// SQLSetConnectOption
	//--------------------------------------------------------------------

	public synchronized void SQLSetConnectOption (
			int   hDbc,
			short fOption,
			int   vParam)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Setting connection option (SQLSetConnectOption), hDbc=" + hDbc + ", fOption=" + fOption + ", vParam=" + vParam);
		}

		errorCode = new byte[1];
		setConnectOption (hDbc, fOption, vParam, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}
	}

	//--------------------------------------------------------------------
	// SQLSetConnectOption
	//--------------------------------------------------------------------

	public synchronized void SQLSetConnectOption (
			int   hDbc,
			short fOption,
			String vParam)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Setting connection option string (SQLSetConnectOption), hDbc=" + hDbc + ", fOption=" + fOption + ", vParam=" + vParam);
		}

		errorCode = new byte[1];
		byte[] bVparam = null;
		char[] cVparam = null;
		if (vParam != null)
			cVparam = vParam.toCharArray();
		try {
			if (vParam != null)
				bVparam = CharsToBytes (charSet, cVparam);
		} catch (java.io.UnsupportedEncodingException ex) {}

		setConnectOptionString (hDbc, fOption, bVparam, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					hDbc, OdbcDef.SQL_NULL_HSTMT);
		}
	}

	//--------------------------------------------------------------------
	// SQLSetCursorName
	//--------------------------------------------------------------------

	public synchronized void SQLSetCursorName (
			int   hStmt,
			String szCursor)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Setting cursor name (SQLSetCursorName), hStmt=" + hStmt + ", szCursor=" + szCursor);
		}

		errorCode = new byte[1];
                
		byte[] bSzcursor = null;
		char[] cSzcursor = null;
		if (szCursor != null)
			cSzcursor = szCursor.toCharArray();
		try {
			if (szCursor != null)
				bSzcursor = CharsToBytes (charSet, cSzcursor);
		} catch (java.io.UnsupportedEncodingException ex) {}

		setCursorName (hStmt, bSzcursor, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLSetStmtOption
	//--------------------------------------------------------------------

	public synchronized void SQLSetStmtOption (
			int   hStmt,
			short fOption,
			int   vParam)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Setting statement option (SQLSetStmtOption), hStmt=" + hStmt + ", fOption=" + fOption + ", vParam=" + vParam);
		}

		errorCode = new byte[1];
		setStmtOption (hStmt, fOption, vParam, errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
                }
	}

	//--------------------------------------------------------------------
	// SQLSpecialColumns
	//--------------------------------------------------------------------

	public synchronized void SQLSpecialColumns (
			int hStmt,
			short fColType,
			String catalog,
			String schema,
			String table,
			int fScope,
			boolean fNullable)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Special columns (SQLSpecialColumns), hStmt=" + hStmt + ", fColType=" + fColType + ",catalog=" + catalog + ", schema=" + schema + ", table=" + table + ", fScope=" + fScope + ", fNullable=" + fNullable);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bTable=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cTable=null;
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (table != null)
			cTable=table.toCharArray();
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (table != null)
				bTable = CharsToBytes (charSet, cTable);
		} catch (java.io.UnsupportedEncodingException exx) {}

		specialColumns (hStmt, fColType,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bTable, (table == null),
			fScope, fNullable,
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLStatistics
	//--------------------------------------------------------------------

	public synchronized void SQLStatistics (
			int hStmt,
			String catalog,
			String schema,
			String table,
			boolean unique,
			boolean approximate)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Statistics (SQLStatistics), hStmt=" + hStmt + ",catalog=" + catalog + ", schema=" + schema + ", table=" + table + ", unique=" + unique + ", approximate=" + approximate);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bTable=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cTable=null;
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (table != null)
			cTable=table.toCharArray();
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (table != null)
				bTable = CharsToBytes (charSet, cTable);
		} catch (java.io.UnsupportedEncodingException exx) {}

		statistics (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bTable, (table == null),
			unique, approximate,
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLTables
	//--------------------------------------------------------------------

	public synchronized void SQLTables (
			int hStmt,
			String catalog,
			String schema,
			String table,
			String types)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Tables (SQLTables), hStmt=" + hStmt + ",catalog=" + catalog + ", schema=" + schema + ", table=" + table + ", types=" + types);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bTable=null;
		byte[] bTypes=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cTable=null;
		char[] cTypes=null;
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (table != null)
			cTable=table.toCharArray();
		if (types != null)
			cTypes=types.toCharArray();
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (table != null)
				bTable = CharsToBytes (charSet, cTable);
			if (types != null)
				bTypes = CharsToBytes (charSet, cTypes);
		} catch (java.io.UnsupportedEncodingException exx) {}

		tables (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bTable, (table == null),
                        bTypes, (types == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLTablePrivileges
	//--------------------------------------------------------------------

	public synchronized void SQLTablePrivileges (
			int hStmt,
			String catalog,
			String schema,
			String table)
			throws SQLException, SQLWarning
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Tables (SQLTables), hStmt=" + hStmt + ",catalog=" + catalog + ", schema=" + schema + ", table=" + table);
		}

		errorCode = new byte[1];

		byte[] bCatalog=null;
		byte[] bSchema=null;
		byte[] bTable=null;
		char[] cCatalog=null; 
		char[] cSchema=null;   
		char[] cTable=null;
		if (catalog != null)
			cCatalog=catalog.toCharArray(); 
		if (schema != null)
			cSchema=schema.toCharArray();   
		if (table != null)
			cTable=table.toCharArray();
		try {
			if (catalog != null)
				bCatalog = CharsToBytes (charSet, cCatalog);
			if (schema != null)
				bSchema = CharsToBytes (charSet, cSchema);
			if (table != null)
				bTable = CharsToBytes (charSet, cTable);
		} catch (java.io.UnsupportedEncodingException exx) {}

		tablePrivileges (hStmt,
                        bCatalog, (catalog == null),
                        bSchema, (schema == null),
                        bTable, (table == null),
			errorCode);

		if (errorCode[0] != 0) {

			// Process the error

			standardError (errorCode[0], OdbcDef.SQL_NULL_HENV,
					OdbcDef.SQL_NULL_HDBC, hStmt);
		}
	}

	//--------------------------------------------------------------------
	// SQLTransact
	//--------------------------------------------------------------------

	public synchronized void SQLTransact (
			int hEnv,
			int hDbc,
			short fType)
			throws SQLException
	{
		byte	errorCode[];

		if (isTracing ()) {
			trace ("Transaction (SQLTransact), hEnv=" + hEnv + ", hDbc=" + hDbc + ", fType=" + fType);
		}

		errorCode = new byte[1];
		transact (hEnv, hDbc, fType, errorCode);

		if (errorCode[0] != 0) {
			throwGenericSQLException ();
		}
	}

	//====================================================================
	// Non-SQL methods
	//====================================================================

	//--------------------------------------------------------------------
	// bufferToInt
	// Returns an integer from the native buffer supplied.  In some
	// cases, data has to be bound within ODBC.  The data that is placed
	// in the bound buffers is in native format - we don't want to have
	// to interpret this from within Java.  So, we'll let the native
	// bridge code convert it for us.
	//--------------------------------------------------------------------

	public native int bufferToInt (
		byte b[]);


	//--------------------------------------------------------------------
	// convertWarning
	// Given a SQLWarning, look for certain SQLStates and generate the
	// proper SQLWarning sub-types (i.e. DataTruncation)
	//--------------------------------------------------------------------

	public static SQLWarning convertWarning (
		JdbcOdbcSQLWarning w)
	{
		SQLWarning convertedWarning = w;

		// If a data truncation warning was generated, convert
		// it to a DataTruncation exception type

		if ((w.getSQLState ()).equals ("01004")) {
			DataTruncation dt = new DataTruncation (-1, false, true,
						0, 0);
			convertedWarning = dt;
		}
		return convertedWarning;
	}
	
	//====================================================================
	// Protected methods
	//====================================================================

	//--------------------------------------------------------------------
	// allocConnect
	//--------------------------------------------------------------------
	protected native int  allocConnect (
			int hEnv,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// allocEnv
	//--------------------------------------------------------------------
	protected native int  allocEnv (
			byte errorCode[]);

	//--------------------------------------------------------------------
	// allocStmt
	//--------------------------------------------------------------------
	protected native int allocStmt (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// allocStmt
	//--------------------------------------------------------------------
	protected native void cancel (
			int hStmt,
			byte errorCode[]);
	//--------------------------------------------------------------------
	// bindColDefault
	//--------------------------------------------------------------------
	protected native void bindColDefault (
			int hStmt,
			int ipar,
			byte rgbValue[],
			byte pcbValue[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// bindInParameterAtExec
	//--------------------------------------------------------------------
	protected native void bindInParameterAtExec (
			int hStmt,
			int ipar,
			int type,
			int len,
			byte dataBuf[],
			byte lenBuf[],
                        byte errorCode[],
                        int buffers[]);
	//--------------------------------------------------------------------
	// bindInParameterBinary
	//--------------------------------------------------------------------
	protected native void bindInParameterBinary (
			int hStmt,
			int ipar,
			int SQLtype,
			byte value[],
			int precision,
			byte dataBuf[],
			byte lenBuf[],
                        byte errorCode[],
                        int buffers[]);
	//--------------------------------------------------------------------
	// bindInParameterDate
	//--------------------------------------------------------------------
	protected native void bindInParameterDate (
			int hStmt,
			int ipar,
			int year,
			int month,
			int day,
			byte dataBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// bindInParameterDouble
	//--------------------------------------------------------------------
	protected native void bindInParameterDouble (
			int hStmt,
			int ipar,
			int SQLtype,
			int scale,
			double value,
			byte dataBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// bindInParameterFloat
	//--------------------------------------------------------------------
	protected native void bindInParameterFloat (
			int hStmt,
			int ipar,
			int SQLtype,
			int scale,
			float value,
			byte dataBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// bindInParameterInteger
	//--------------------------------------------------------------------
	protected native void bindInParameterInteger (
			int hStmt,
			int ipar,
			int SQLtype,
			int value,
			byte dataBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// bindInParameterNull
	//--------------------------------------------------------------------
	protected native void bindInParameterNull (
			int hStmt,
			int ipar,
			int SQLtype,
			byte lenBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// bindInParameterString
	//--------------------------------------------------------------------
	protected native void bindInParameterString (
			int hStmt,
			int ipar,
			int SQLtype,
                        byte value[],
			int precision,
			int scale,
			byte dataBuf[],
                        byte errorCode[],
                        int buffers[]);
	//--------------------------------------------------------------------
	// bindInParameterTime
	//--------------------------------------------------------------------
	protected native void bindInParameterTime (
			int hStmt,
			int ipar,
			int hours,
			int minutes,
			int seconds,
			byte dataBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// bindInParameterTimestamp
	//--------------------------------------------------------------------
	protected native void bindInParameterTimestamp (
			int hStmt,
			int ipar,
			int year,
			int month,
			int day,
			int hours,
			int minutes,
			int seconds,
			int nanos,
			byte dataBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// bindOutParameterString
	//--------------------------------------------------------------------
	protected native void bindOutParameterString (
			int hStmt,
			int ipar,
			int SQLtype,
			int scale,
			byte dataBuf[],
			byte lenBuf[],
                        byte errorCode[],
                        int buffers[]);

	//--------------------------------------------------------------------
	// browseConnect
	//--------------------------------------------------------------------
	protected native void browseConnect (
			int hDbc,
                        byte bConnectString[],
			byte connStrOut[],
			byte errorCode[]);


	//--------------------------------------------------------------------
	// colAttributes
	//--------------------------------------------------------------------
	protected native int colAttributes (
			int hStmt,
			int column,
			int type,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// colAttributesString
	//--------------------------------------------------------------------
	protected native void colAttributesString (
			int hStmt,
			int column,
			int type,
			byte rgbDesc[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// columns
	//--------------------------------------------------------------------
	protected native void columns (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bTable, boolean tableNull,
                        byte[] bColumn, boolean columnNull,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// columnPrivileges
	//--------------------------------------------------------------------
	protected native void columnPrivileges (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bTable, boolean tableNull,
                        byte[] bColumn, boolean columnNull,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// describeParam
	//--------------------------------------------------------------------
	protected native int describeParam (
			int hStmt,
			int param,
			int returnParam,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// disconnect
	//--------------------------------------------------------------------
	protected native void disconnect (
			int hDbc,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// driverConnect
	//--------------------------------------------------------------------
	protected native void driverConnect (
			int hDbc,
                        byte[] bConnectString,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// error
	//--------------------------------------------------------------------
	protected native int error (
			int hEnv,
			int hDbc,
			int hStmt,
			byte sqlState[],
			byte errorMessage[],
			byte errorCode[]);
			
	//--------------------------------------------------------------------
	// execDirect
	//--------------------------------------------------------------------
	protected native void execDirect (
			int hStmt,
                        byte[] bQuery,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// execute
	//--------------------------------------------------------------------
	protected native void execute (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// fetch
	//--------------------------------------------------------------------
	protected native void fetch (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// foriegnKeys
	//--------------------------------------------------------------------
	protected native void foreignKeys (
			int hStmt,
                        byte[] bPKcatalog, boolean PKcatalogNull,
                        byte[] bPKschema, boolean PKschemaNull,
                        byte[] bPKtable, boolean PKtableNull,
                        byte[] bFKcatalog, boolean FKcatalogNull,
                        byte[] bFKschema, boolean FKschemaNull,
                        byte[] bFKtable, boolean FKtableNull,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// freeConnect
	//--------------------------------------------------------------------
	protected native void freeConnect (
			int hDbc,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// freeEnv
	//--------------------------------------------------------------------
	protected native void freeEnv (
			int hEnv,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// freeStmt
	//--------------------------------------------------------------------
	protected native void freeStmt (
			int hEnv,
			int fOption,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getConnectOption
	//--------------------------------------------------------------------
	protected native int getConnectOption (
			int hDbc,
			short fOption,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getConnectOptionString
	//--------------------------------------------------------------------
	protected native void getConnectOptionString (
			int hDbc,
			short fOption,
			byte szParam[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getCursorName
	//--------------------------------------------------------------------
	protected native void getCursorName (
			int hStmt,
			byte szCursor[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getStmtOption
	//--------------------------------------------------------------------
	protected native int getStmtOption (
			int hStmt,
			short fOption,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getDataBinary
	//--------------------------------------------------------------------
	protected native int getDataBinary (
			int hStmt,
			int column,
			int cType,
			byte b[],
			int length,
			byte errorCode[]);
			
	//--------------------------------------------------------------------
	// getDataDouble
	//--------------------------------------------------------------------
	protected native double getDataDouble (
			int hStmt,
			int column,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getDataFloat
	//--------------------------------------------------------------------
	protected native float getDataFloat (
			int hStmt,
			int column,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getDataInteger
	//--------------------------------------------------------------------
	protected native int getDataInteger (
			int hStmt,
			int column,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getDataString
	//--------------------------------------------------------------------
	protected native int getDataString (
			int hStmt,
			int column,
			byte rgbValue[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getDataStringDate
	//--------------------------------------------------------------------
	protected native void getDataStringDate (
			int hStmt,
			int column,
			byte rgbValue[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getDataStringTime
	//--------------------------------------------------------------------
	protected native void getDataStringTime (
			int hStmt,
			int column,
			byte rgbValue[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getDataStringTimestamp
	//--------------------------------------------------------------------
	protected native void getDataStringTimestamp (
			int hStmt,
			int column,
			byte rgbValue[],
			byte errorCode[]);
			
	//--------------------------------------------------------------------
	// getInfo
	//--------------------------------------------------------------------
	protected native int getInfo (
			int hDbc,
			short fInfoType,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getInfoShort
	//--------------------------------------------------------------------
	protected native int getInfoShort (
			int hDbc,
			short fInfoType,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getInfoString
	//--------------------------------------------------------------------
	protected native void getInfoString (
			int hDbc,
			short fInfoType,
			byte szParam[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// getTypeInfo
	//--------------------------------------------------------------------
	protected native void getTypeInfo (
			int hStmt,
			short fSqlType,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// moreResults
	//--------------------------------------------------------------------
	protected native void moreResults (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// nativeSql
	//--------------------------------------------------------------------
	protected native void nativeSql (
			int hDbc,
                        byte[] bQuery,
			byte nativeQuery[],
			byte errorCode[]);

	//--------------------------------------------------------------------
	// numParams
	//--------------------------------------------------------------------
	protected native int numParams (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// numResultCols
	//--------------------------------------------------------------------
	protected native int numResultCols (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// paramData
	//--------------------------------------------------------------------
	protected native int paramData (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// prepare
	//--------------------------------------------------------------------
	protected native void prepare (
			int hStmt,
                        byte[] bQuery,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// primaryKeys
	//--------------------------------------------------------------------
	protected native void primaryKeys (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bTable, boolean tableNull,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// procedures
	//--------------------------------------------------------------------
	protected native void procedures (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bProcedure, boolean procedureNull,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// procedureColumns
	//--------------------------------------------------------------------
	protected native void procedureColumns (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bProcedure, boolean procedureNull,
                        byte[] bColumn, boolean columnNull,
			byte errorCode[]);
	//--------------------------------------------------------------------
	// putData
	//--------------------------------------------------------------------
	protected native void putData (
			int hStmt,
			byte dataBuf[],
			int dataLen,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// rowCount
	//--------------------------------------------------------------------
	protected native int rowCount (
			int hStmt,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// setConnectOption
	//--------------------------------------------------------------------
	protected native void setConnectOption (
			int hDbc,
			short fOption,
			int vParam,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// setConnectOptionString
	//--------------------------------------------------------------------
	protected native void setConnectOptionString (
			int hDbc,
			short fOption,
                        byte[] bVparam,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// setCursorName
	//--------------------------------------------------------------------
	protected native void setCursorName (
			int hStmt,
                        byte[] bSzcursor,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// setStmtOption
	//--------------------------------------------------------------------
	protected native void setStmtOption (
			int hStmt,
			short fOption,
			int vParam,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// specialColumns
	//--------------------------------------------------------------------
	protected native void specialColumns (
			int hStmt, short fColType,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bTable, boolean tableNull,
			int fScope, boolean fNullable,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// statistics
	//--------------------------------------------------------------------
	protected native void statistics (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bTable, boolean tableNull,
			boolean unique, boolean approximate,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// tables
	//--------------------------------------------------------------------
	protected native void tables (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bTable, boolean tableNull,
                        byte[] bTypes, boolean typesNull,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// tablePrivileges
	//--------------------------------------------------------------------
	protected native void tablePrivileges (
			int hStmt,
                        byte[] bCatalog, boolean catalogNull,
                        byte[] bSchema, boolean schemaNull,
                        byte[] bTable, boolean tableNull,
			byte errorCode[]);

	//--------------------------------------------------------------------
	// transact
	//--------------------------------------------------------------------
	protected native void transact (
			int hEnv,
			int hDbc,
			short fType,
			byte errorCode[]);


        protected native void ReleaseStoredBytes (int x1, int x2);
        protected native void ReleaseStoredChars (int s1, int s2);

	//--------------------------------------------------------------------
	// createSQLException
	// Get state, message, and native error code from SQLError.  If
	// multiple errors exist, chain them together.  Return the first
	// error in the list.
	//--------------------------------------------------------------------

	SQLException createSQLException (
			int hEnv,
			int hDbc,
			int hStmt)
	{
		byte	sqlState[];
		byte	errorMsg[];
		byte	errorCode[];
		int	nativeError;
		boolean	done = false;
		SQLException	firstEx = null;
		SQLException	lastEx = null;

		if (isTracing ()) {
			trace ("ERROR - Generating SQLException...");
		}

		while (!done) {
			errorCode = new byte[1];
			sqlState = new byte[6];
			errorMsg = new byte[
				JdbcOdbcLimits.DEFAULT_BUFFER_LENGTH];

			nativeError = error (hEnv, hDbc, hStmt, sqlState,
				errorMsg, errorCode);

			if (errorCode[0] != OdbcDef.SQL_SUCCESS) {
				done = true;
			}
			else {
				SQLException	ex = null;

                                String msg = new String();
                                String state = new String();
                                try {
                                    msg = BytesToChars (charSet, errorMsg);
                                    state = BytesToChars (charSet, sqlState);
                                } catch (java.io.UnsupportedEncodingException exxx) {}


				// Create a new exception record
				
				ex = new SQLException (msg.trim (),
						state.trim (), nativeError);

				// If we don't have an exception yet, set the
				// first one the in chain.  Otherwise, set
				// the chain pointers.

				if (firstEx == null) {
					firstEx = ex;
				}
				else {
					lastEx.setNextException (ex);
				}

				// Save our last exception so we can chain
				lastEx = ex;
			}
		}

		// If we didn't get an error message from SQLError, make
		// one for ourselves

		if (firstEx == null) {
			String msg   = "General error";
			String state = "S1000";

			if (isTracing ()) {
				trace ("ERROR - " + state + " " + msg);
			}
			firstEx = new SQLException (msg, state);
		}
		return firstEx;
	}

	//--------------------------------------------------------------------
	// createSQLWarning
	// Get state, message, and native error code from SQLError.  If
	// multiple warnings exist, chain them together.  Return the first
	// warning in the list.
	//--------------------------------------------------------------------

	SQLWarning createSQLWarning (
			int hEnv,
			int hDbc,
			int hStmt)
	{
		byte	sqlState[];
		byte	errorMsg[];
		byte	errorCode[];
		int	nativeError;
		boolean	done = false;
		SQLWarning	firstEx = null;
		SQLWarning	lastEx = null;

		if (isTracing ()) {
			trace ("WARNING - Generating SQLWarning...");
		}

		while (!done) {
			errorCode = new byte[1];
			sqlState = new byte[6];
			errorMsg = new byte[
				JdbcOdbcLimits.DEFAULT_BUFFER_LENGTH];

			nativeError = error (hEnv, hDbc, hStmt, sqlState,
				errorMsg, errorCode);

			if (errorCode[0] != OdbcDef.SQL_SUCCESS) {
				done = true;
			}
			else {
				JdbcOdbcSQLWarning	ex = null;
                                String msg = new String();
                                String state = new String();
                                try {
                                    msg = BytesToChars (charSet, errorMsg);
                                    state = BytesToChars (charSet, sqlState);
                                } catch (java.io.UnsupportedEncodingException exx) {}

				// Create a new warning record

				ex = new JdbcOdbcSQLWarning (msg.trim (),
						state.trim (),
						nativeError);

				// If we don't have a warning yet, set the
				// first one the in chain.  Otherwise, set
				// the chain pointers.

				if (firstEx == null) {
					firstEx = ex;
				}
				else {
					lastEx.setNextWarning (ex);
				}

				// Save our last warning so we can chain
				lastEx = ex;
			}
		}

		// If we didn't get an warning message from SQLError, make
		// one for ourselves

		if (firstEx == null) {
			String msg   = "General warning";
			String state = "S1000";

			if (isTracing ()) {
				trace ("WARNING - " + state + " " + msg);
			}
			firstEx = new JdbcOdbcSQLWarning (msg, state);
		}
		return firstEx;
	}

	//--------------------------------------------------------------------
	// throwGenericSQLException
	// Throw a generic SQLException object.
	//--------------------------------------------------------------------

	void throwGenericSQLException () throws SQLException
	{
		String msg   = "General error";
		String state = "S1000";

		if (isTracing ()) {
			trace ("ERROR - " + state + " " + msg);
		}
		throw new SQLException (msg, state);
	}

	//--------------------------------------------------------------------
	// standardError
	// Checks for standard error codes, and throws either a SQLException
	// or SQLWarning
	//--------------------------------------------------------------------

	void standardError (
			short errorCode,
			int hEnv,
			int hDbc,
			int hStmt)
			throws SQLException, SQLWarning
	{
		String msg;

		if (isTracing ()) {
			trace ("RETCODE = " + errorCode);
		}

		switch (errorCode) {

		// If this is a real error (SQL_ERROR), throw a
		// SQLException

		case OdbcDef.SQL_ERROR:
			throw createSQLException (hEnv, hDbc, hStmt);

		// If this is a success with addition information
		// (SQL_SUCCESS_WITH_INFO), throw a SQLWarning

		case OdbcDef.SQL_SUCCESS_WITH_INFO:
			throw createSQLWarning (hEnv, hDbc, hStmt);

		// Check for an invalid handle (SQL_INVALID_HANDLE)

		case OdbcDef.SQL_INVALID_HANDLE:
			msg = "Invalid handle";
			if (isTracing ()) {
				trace ("ERROR - " + msg);
			}
			throw new SQLException (msg);
			
		// Check for no data found

		case OdbcDef.SQL_NO_DATA:
			msg = "No data found";
			if (isTracing ()) {
				trace ("ERROR - " + msg);
			}
			throw new SQLException (msg);

		// If this is some other type of error, throw a
		// generic SQLException

		default:
			throwGenericSQLException ();
		}
	}
	
	//====================================================================
	// Data attributes
	//====================================================================

	public final static int MajorVersion = 1;
        public final static int MinorVersion = 2001;    // .2001
        public String charSet;

}
