/*
 * @(#)JdbcOdbcBoundCol.java	1.15 98/04/28
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcBoundCol.java
//
// Description: Class for a result set column.  This class is used a place
//              holder for information needed for each result set column.
//              Currently, this includes columns that are using an InputStream
//              to return data from the result set.
//
//              An array of JdbcOdbcBoundCol objects is created for this
//              purpose.
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.sql.*;

public class JdbcOdbcBoundCol
	extends	JdbcOdbcObject {

	//--------------------------------------------------------------------
	// Constructor
	// Perform any necessary initialization.
	//--------------------------------------------------------------------

	public JdbcOdbcBoundCol ()
	{
		type = OdbcDef.SQL_TYPE_UNKNOWN;
		len = -1;
	}
	
	//--------------------------------------------------------------------
	// setInputStream
	// Sets the input stream object.
	//--------------------------------------------------------------------

	public void setInputStream (JdbcOdbcInputStream inStream)
	{
		inputStream = inStream;
	}

	//--------------------------------------------------------------------
	// closeInputStream
	// Closes (invalidates) the current input stream, and removes
	// the reference to the input stream
	//--------------------------------------------------------------------

	public void closeInputStream ()
	{
		if (inputStream != null) {
			inputStream.invalidate ();
			inputStream = null;
		}
	}

	//--------------------------------------------------------------------
	// setType
	// Sets the JDBC SQL type
	//--------------------------------------------------------------------

	public void setType (
		int javaType)
	{	
		type = javaType;
	}

	//--------------------------------------------------------------------
	// getType
	// Gets the JDBC SQL type
	//--------------------------------------------------------------------

	public int getType ()
	{	
		return type;
	}

	//--------------------------------------------------------------------
	// setLength
	// Sets the column length
	//--------------------------------------------------------------------

	public void setLength (
		int l)
	{	
		len = l;
	}

	//--------------------------------------------------------------------
	// getLength
	// Gets the column length
	//--------------------------------------------------------------------

	public int getLength ()
	{	
		return len;
	}

	//====================================================================
	// Data attributes
	//====================================================================

	protected int type;			// JDBC SQL type for the
						//  column.  Initially set
						//  to SQL_TYPE_UNKNOWN

	protected int len;			// Length of the column.
						//  Initially set to -1
						//  indicating that it is 
						//  unknown.

	protected JdbcOdbcInputStream inputStream;
						// If an input stream is used
						//  to get data from a column,
						//  a reference to that object
						//  is kept here.  Once the
						//  Statement is closed or
						//  the cursor is moved (next)
						//  all of the input streams
						//  for the statement are
						//  closed (marked as invalid)
						//  and then the reference is
						//  removed.
}
