/*
 * @(#)JdbcOdbcBoundParam.java	1.16 98/04/28
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcBoundParam.java
//
// Description: Class for a single bound parameter.  When a parameter is
//              bound for a prepared statement, a copy of the native data is
//              kept in this object.  This is so Java won't garbage collect
//              the data before we are done with it.  An array of these
//              objects are kept with the PreparedStatement object, one
//              for each parameter marker.
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.sql.*;

public class JdbcOdbcBoundParam
	extends	JdbcOdbcObject {

	//--------------------------------------------------------------------
	// initialize
	// Perform an necessary initialization
	//--------------------------------------------------------------------
	public void initialize ()
	{
		// Allocate storage for the length.  Note - the length is
		// stored in native format, and will have to be converted
		// to a Java int.  The jdbcodbc 'C' bridge provides an
		// interface to do this.

		paramLength = new byte[4];
	}

	//--------------------------------------------------------------------
	// allocBindDataBuffer
	// Allocates and returns a new bind data buffer of the specified
	// length
	//--------------------------------------------------------------------
	public byte[] allocBindDataBuffer (
		int bufLen)
	{
		binaryData = new byte[bufLen];

		// Reset the input stream, we are doing a new bind
		setInputStream (null, 0);

		return binaryData;
	}

	//--------------------------------------------------------------------
	// getBindDataBuffer
	// Returns the data buffer to be used when binding to a parameter
	//--------------------------------------------------------------------
	public byte[] getBindDataBuffer ()
	{
		return binaryData;
	}

	//--------------------------------------------------------------------
	// getBindLengthBuffer
	// Returns the length buffer to be used when binding to a parameter
	//--------------------------------------------------------------------
	public byte[] getBindLengthBuffer ()
	{
		return paramLength;
	}

	//--------------------------------------------------------------------
	// setInputStream
	// Sets the input stream for the bound parameter
	//--------------------------------------------------------------------
	public void setInputStream (
		java.io.InputStream inputStream,
		int len)
	{
		paramInputStream = inputStream;
		paramInputStreamLen = len;
	}

	//--------------------------------------------------------------------
	// getInputStream
	// Gets the input stream for the bound parameter
	//--------------------------------------------------------------------
	public java.io.InputStream getInputStream ()
	{
		return paramInputStream;
	}

	//--------------------------------------------------------------------
	// getInputStreamLen
	// Gets the input stream length for the bound parameter
	//--------------------------------------------------------------------
	public int getInputStreamLen ()
	{
		return paramInputStreamLen;
	}

	//--------------------------------------------------------------------
	// setSqlType
	// Sets the Java sql type used to register an OUT parameter
	//--------------------------------------------------------------------
	
	public void setSqlType (
		int type)
	{
		sqlType = type;
	}

	//--------------------------------------------------------------------
	// getSqlType
	// Gets the Java sql type used to register an OUT parameter
	//--------------------------------------------------------------------
	
	public int getSqlType ()
	{
		return sqlType;
	}

	//--------------------------------------------------------------------
	// setStreamType
	// Sets the input stream type used to register an OUT parameter
	//--------------------------------------------------------------------
	
	public void setStreamType (
		int type)
	{
		streamType = type;
	}

	//--------------------------------------------------------------------
	// getStreamType
	// Gets the input stream type used to register an OUT parameter
	//--------------------------------------------------------------------
	
	public int getStreamType ()
	{
		return streamType;
	}

	//--------------------------------------------------------------------
	// setOutputParameter
	// Sets the flag indicating if this is an OUTPUT parameter
	//--------------------------------------------------------------------
	
	public void setOutputParameter (
		boolean output)
	{
		outputParameter = output;
	}

	//--------------------------------------------------------------------
	// isOutputParameter
	// Gets the OUTPUT parameter flag
	//--------------------------------------------------------------------
	
	public boolean isOutputParameter ()
	{
		return outputParameter;
	}

	//====================================================================
	// Data attributes
	//====================================================================

	protected byte binaryData[];		       // Storage area to be used
								// when binding the parameter

	protected byte paramLength[];		       // Storage area to be used
								// for the bound length of the
								// parameter.  Note that this
								// data is in native format.

	protected java.io.InputStream paramInputStream;
								// When an input stream is 
								// bound to a parameter, the
								// input stream is saved
								// until needed.

	protected int paramInputStreamLen;	              // Length of input stream

	protected int sqlType;                          // Java SQL type used to
	                                                // register an OUT parameter

	protected int streamType;			       // Input stream type
								// (ASCII, BINARY, UNICODE)
		public final static short ASCII   = 1;
		public final static short UNICODE = 2;
		public final static short BINARY  = 3;

	protected boolean outputParameter;	// true for OUTPUT parameters

                                        
        protected int pA1=0;              //pointers
        protected int pA2=0;
        protected int pB1=0;
        protected int pB2=0;
        protected int pC1=0;
        protected int pC2=0;
        protected int pS1=0;
        protected int pS2=0;// reserved for strings(UTFChars)
}
