/*
 * @(#)JdbcOdbcCallableStatement.java	1.22 98/09/09
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcCallableStatement.java
//
// Description: Impementation of the CallableStatement interface class
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.sql.*;
import java.util.Calendar;
import java.math.BigDecimal;

public class JdbcOdbcCallableStatement
	extends		JdbcOdbcPreparedStatement
	implements	java.sql.CallableStatement {

	//====================================================================
	// We need to save registerOutParameter's scale parameters
	// for calls to getObject(BigDecimal);
        // We will use an array of bytes to save space (scale will never be 
	// more than 127.
           public byte scalez[]=new byte[200];
        //====================================================================
   
	//====================================================================
	// Public methods
	//====================================================================

	//--------------------------------------------------------------------
	// Constructor
	// Perform any necessary initialization.
	//--------------------------------------------------------------------

	public JdbcOdbcCallableStatement (
		JdbcOdbcConnectionInterface con)
	{
		super (con);
		lastParameterNull = false;
	}

	//--------------------------------------------------------------------
	// registerOutParameter
	// Before executing a stored procedure call you must explicitly
	// call registerOutParameter to register the java.sql.Type of each
	// out parameter.
	// Note that all parameters are bound to a SQL_C_CHAR data type,
	// so the resulting buffer is a string.  When retrieving the value
	// of the output parameter (with getXXXX), the buffer will be converted
	// to a String first, then the appropriate data type for the
	// getXXXX method.
	//--------------------------------------------------------------------

	public void registerOutParameter (
		int parameterIndex,
		int sqlType)
		throws SQLException
	{
		registerOutParameter (parameterIndex, sqlType, 0);
	}

	// You must also specify the scale for numeric/decimal types:	
	public void registerOutParameter (
		int parameterIndex,
		int sqlType,
		int scale)
		throws SQLException
	{
		// Set the sql type for this OUT parameter

		setSqlType (parameterIndex, sqlType);

		// Save the scale in scalez[]
                 if (parameterIndex <= 200)
                    scalez[parameterIndex]=(byte)scale;

		// Indicate that this is an output parameter
		setOutputParameter (parameterIndex, true);

		// First, allocate a buffer to bind the output parameter
		// to.

		int maxLen;

		switch (sqlType) {
		case Types.DATE:
			maxLen = 10;
			break;
		case Types.TIME:
			maxLen = 8;
			break;
		case Types.TIMESTAMP:
			maxLen = 19;
			if (scale > 0) {
				maxLen += (scale + 1);
			}
			break;
		case Types.BIT:
			maxLen = 1;
			break;
		case Types.TINYINT:
			maxLen = 4;
			break;
		case Types.SMALLINT:
			maxLen = 6;
			break;
		case Types.INTEGER:
			maxLen = 11;
			break;
		case Types.BIGINT:
			maxLen = 20;
			break;
		case Types.REAL:
			maxLen = 13;
			break;
		case Types.FLOAT:
		case Types.DOUBLE:
			maxLen = 22;
			break;
		case Types.DECIMAL:
		case Types.NUMERIC:
			maxLen = 38;
			break;
		default:
			// Get the precision for the SQL type.  

			maxLen = getPrecision (sqlType);
			if ((maxLen <= 0) ||
				(maxLen > JdbcOdbcLimits.DEFAULT_IN_PRECISION)) {
				maxLen = JdbcOdbcLimits.DEFAULT_IN_PRECISION;
			}
		}

		byte dataBuf[] = allocBindBuf (parameterIndex, maxLen + 1);

		// Get the buffer to be used for the output length

		byte dataLen[] = getLengthBuf (parameterIndex);

		// Convert the given Java SQL type to an ODBC SQL type

		sqlType = OdbcDef.jdbcTypeToOdbc (sqlType);

                int buffers[]=new int[4];
                buffers[0]=0;
                buffers[1]=0;
                buffers[2]=0;
                buffers[3]=0;

		// Bind the output parameter, always as string data
                
		OdbcApi.SQLBindOutParameterString (hStmt, parameterIndex,
                        sqlType, scale, dataBuf, dataLen, buffers);

                
                //save the native pointers from garbage collection
                boundParams[parameterIndex - 1].pA1=buffers[0];
                boundParams[parameterIndex - 1].pA2=buffers[1];
                boundParams[parameterIndex - 1].pB1=buffers[2];
                boundParams[parameterIndex - 1].pB2=buffers[3];
}

	//--------------------------------------------------------------------
	// wasNull
	// An OUT parameter may have the value of SQL NULL; wasNull reports 
	// whether the last value read has this special value.
	//--------------------------------------------------------------------

	public boolean wasNull ()
		throws SQLException
	{
		return lastParameterNull;
	}

	//--------------------------------------------------------------------
	// getString
	//--------------------------------------------------------------------

	public String getString (
		int parameterIndex)
		throws SQLException
	{
		// If the value is null, return a null object
		if (isNull (parameterIndex)) {
			return null;
		}
	
		// Get the SQL type of the parameter

		int sqlType = getSqlType (parameterIndex); 

		// Convert the data buffer for the output parameter into
		// a String
                String s=new String();
                try { 
                     s = BytesToChars (OdbcApi.charSet, getDataBuf (parameterIndex));
                } catch (java.io.UnsupportedEncodingException exx) {}
		// Remove trailing spaces

		s = s.trim ();

		if (isTracing ()) {
			trace ("String value for OUT parameter " +
					parameterIndex + "=" + s);
		}
		return s;
	}

	//--------------------------------------------------------------------
	// Since the output parameter was bound as a specific C type 
	// (always a SQL_C_CHAR), we will have to get the output parameter
	// as a character, then convert it to the proper return type.
	//--------------------------------------------------------------------

	//--------------------------------------------------------------------
	// getBoolean
	//--------------------------------------------------------------------

	public boolean getBoolean (
		int parameterIndex)
		throws SQLException
	{
		// If the value is null, return false
		if (isNull (parameterIndex)) {
			return false;
		}

		boolean rc = false;

		// Get the out parameter as a String, and look for a
		// '1' for a true value
		if (getString (parameterIndex).equals("1")) {
			rc = true;
		}
		return rc;
	}

	//--------------------------------------------------------------------
	// getByte
	//--------------------------------------------------------------------

	public byte getByte (
		int parameterIndex)
		throws SQLException
	{
		// Cast getInt to a byte value

		return (byte) getInt (parameterIndex);
	}

	//--------------------------------------------------------------------
	// getShort
	//--------------------------------------------------------------------

	public short getShort (
		int parameterIndex)
		throws SQLException
	{
		// Cast getInt to a short value

		return (short) getInt (parameterIndex);
	}

	//--------------------------------------------------------------------
	// getInt
	//--------------------------------------------------------------------

	public int getInt (
		int parameterIndex)
		throws SQLException
	{
		// If the value is null, return zero
		if (isNull (parameterIndex)) {
			return 0;
		}

		// Create an Integer object from the output parameter string
		Integer i = new Integer (getString (parameterIndex));

		return i.intValue ();
	}

	//--------------------------------------------------------------------
	// getLong
	//--------------------------------------------------------------------

	public long getLong (
		int parameterIndex)
		throws SQLException
	{
		// If the value is null, return zero
		if (isNull (parameterIndex)) {
			return 0;
		}
		// Create an Integer object from the output parameter string
		Integer i = new Integer (getString (parameterIndex));

		return i.longValue ();
	}

	//--------------------------------------------------------------------
	// getFloat
	//--------------------------------------------------------------------

	public float getFloat (
		int parameterIndex)
		throws SQLException
	{
		// If the value is null, return zero
		if (isNull (parameterIndex)) {
			return 0;
		}
		// Create a Float object from the output parameter string
		Float f = new Float (getString (parameterIndex));

		return f.floatValue ();
	}

	//--------------------------------------------------------------------
	// getDouble
	//--------------------------------------------------------------------

	public double getDouble (
		int parameterIndex)
		throws SQLException
	{
		// If the value is null, return zero
		if (isNull (parameterIndex)) {
			return 0;
		}
		// Create a Double object from the output parameter string
		Double d = new Double (getString (parameterIndex));

		return d.doubleValue ();
	}

	//--------------------------------------------------------------------
	// getBigDecimal
	//--------------------------------------------------------------------

	public BigDecimal getBigDecimal (
		int parameterIndex,
		int scale)
		throws SQLException
	{
		// If the value is null, return null
		if (isNull (parameterIndex)) {
			return null;
		}
		// Create a BigDecimal object from the output parameter string
		BigDecimal num = new BigDecimal(getString (parameterIndex));
		/* NOTE: In the old-style Bignum there was the notion of 
		   a default rounding mode. There is no longer this 
		   notion, and so we default to something reasonable. */
		return num.setScale(scale, BigDecimal.ROUND_HALF_EVEN);
	}

	//--------------------------------------------------------------------
	// getBytes
	// Returns a byte array of the bound parameter data.
	//--------------------------------------------------------------------

	public byte[] getBytes (
		int parameterIndex)
		throws SQLException
	{
		// If the value is null, return null
		if (isNull (parameterIndex)) {
			return null;
		}
		// Convert the bound character string (in hexadecimal)
		// into a byte array

		return hexStringToByteArray (getString (parameterIndex));
	}

	//--------------------------------------------------------------------
	// getDate
	//--------------------------------------------------------------------

	public java.sql.Date getDate (
		int parameterIndex)
		throws SQLException
	{
		// Get the string value and convert to Date object

		String s = getString (parameterIndex);
		if (s == null) {
			return null;
		}
		return java.sql.Date.valueOf (s);
	}

	//--------------------------------------------------------------------
	// getTime
	//--------------------------------------------------------------------

	public java.sql.Time getTime (
		int parameterIndex)
		throws SQLException
	{
		// Get the string value and convert to Time object

		String s = getString (parameterIndex);
		if (s == null) {
			return null;
		}
		return java.sql.Time.valueOf (s);
	}

	//--------------------------------------------------------------------
	// getTimestamp
	//--------------------------------------------------------------------

	public java.sql.Timestamp getTimestamp (
		int parameterIndex) 
		throws SQLException
	{
		// Get the string value and convert to Timestamp object

		String s = getString (parameterIndex);
		if (s == null) {
			return null;
		}

		// If all we got was a date, put on default seconds

		if (s.length () == 10) {
			s += " 00:00:00";
		}

		return java.sql.Timestamp.valueOf (s);
	}


	//--------------------------------------------------------------------
	// getObject
	// Returns a Java object for the parameter.
	// See the JDBC spec's "Dynamic Programming" chapter for details.
	//--------------------------------------------------------------------

	public Object getObject (
		int parameterIndex)
		throws SQLException
	{
		Object	value = null;

		// Get the SQL type of the parameter

		int sqlType = getSqlType (parameterIndex); 

		// If the column is null, always return a null object

		if (isNull (parameterIndex)) {
			return null;
		}

		// For each SQL type, call the appropriate routine and
		// convert to the proper object type

		switch (sqlType) {

		case Types.CHAR:
		case Types.VARCHAR:
		case Types.LONGVARCHAR:

//			// Get the data as a string, convert it to a byte
//			// array, and return an InputStream
//
//			String s = getLongVarChar (parameterIndex);
//			if (s != null) {
//				byte b[] = new byte[s.length ()];
//				s.getBytes (0, s.length (), b, 0);
//				value = (Object) new JdbcOdbcInputStream (
//					OdbcApi, hStmt, parameterIndex, b);
//			}

			value = (Object) getString (parameterIndex);
			break;

		case Types.NUMERIC:
		case Types.DECIMAL:

			// The getNumeric method requires us to provide
			// a scale.  The only way to get the scale is to
			// call DatabaseMetaData.getProcedureColumns
			// (which may not be supported for the data source)
			// We'll use the scale stored in scalez[]
			  if (parameterIndex<=200)
                             value = (Object) getBigDecimal(parameterIndex, scalez[parameterIndex]);
                          else
       			     value = (Object) getBigDecimal(parameterIndex, JdbcOdbcLimits.DEFAULT_OUT_SCALE);
			break;

		case Types.BIT:
			value = (Object) new Boolean (
					getBoolean (parameterIndex));
			break;

		case Types.TINYINT:
			value = (Object) new Integer (getByte (
					parameterIndex));
			break;

		case Types.SMALLINT:
			value = (Object) new Integer (getShort (
					parameterIndex));
			break;

		case Types.INTEGER:
			value = (Object) new Integer (getInt (
					parameterIndex));
			break;

		case Types.BIGINT:
			value = (Object) new Long (getLong (parameterIndex));
			break;

		case Types.REAL:
		case Types.FLOAT:
			value = (Object) new Float (getFloat (parameterIndex));
			break;

		case Types.DOUBLE:
			value = (Object) new Double (getDouble (
					parameterIndex));
			break;
			
		case Types.BINARY:
		case Types.VARBINARY:
		case Types.LONGVARBINARY:

//			// Get the data as a byte array and convert to an
//			// InputStream
//
//			byte b[] = getLongVarBinary (parameterIndex);
//			if (b != null) {
//				value = (Object) new JdbcOdbcInputStream (
//					OdbcApi, hStmt, parameterIndex, b);
//			}
			value = (Object) getBytes (parameterIndex);
			break;

		case Types.DATE:
			value = (Object) getDate (parameterIndex);
			break;

		case Types.TIME:
			value = (Object) getTime (parameterIndex);
			break;

		case Types.TIMESTAMP:
			value = (Object) getTimestamp (parameterIndex);
			break;

		}

		return value;
	}

	//--------------------------------------------------------------------
	// isNull
	// Helper method to determine if an output parameter is null.  Also
	// sets the lastParameterNull attribute for use with wasNull.
	//--------------------------------------------------------------------

	protected boolean isNull (
		int parameterIndex)
		throws SQLException
	{
		// Verify that the given parameter is for OUTPUT

		if (!isOutputParameter (parameterIndex)) {
			throw new SQLException ("Parameter " + parameterIndex +
					" is not an OUTPUT parameter");
		}

		boolean rc = false;
		rc = (getParamLength(parameterIndex) == OdbcDef.SQL_NULL_DATA);
		if (isTracing ()) {
			trace ("Output Parameter " + parameterIndex +
					" null: " + rc);
		}
		lastParameterNull = rc;
		return rc;
	}

	//--------------------------------------------------------------------
	// setSqlType
	// Sets the Java sql type for when registering an OUT parameter
	//--------------------------------------------------------------------

	protected void setSqlType (
		int index,
		int type)
	{
		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {
			boundParams[index - 1].setSqlType (type);
		}
	}

	//--------------------------------------------------------------------
	// getSqlType
	// Gets the Java sql type for when registering an OUT parameter
	//--------------------------------------------------------------------

	protected int getSqlType (
		int index)
	{
		int type = Types.OTHER;

		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {
			type = boundParams[index - 1].getSqlType ();
		}
		return type;
	}

	//--------------------------------------------------------------------
	// setOutputParameter
	// Sets the output parameter flag
	//--------------------------------------------------------------------

	protected void setOutputParameter (
		int index,
		boolean output)
	{
		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {
			boundParams[index - 1].setOutputParameter (output);
		}
	}

	//--------------------------------------------------------------------
	// isOutputParameter
	// Returns true if the given parameter is for OUTPUT
	//--------------------------------------------------------------------

	protected boolean isOutputParameter (
		int index)
	{
		boolean output = false;

		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {
			output = boundParams[index - 1].isOutputParameter ();
		}
		return output;
	}



	public synchronized void close ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Statement.close");
		}

		// Close/clear our result set

		clearMyResultSet ();

		// Reset last warning message

		try {
			clearWarnings ();
			if (hStmt != OdbcDef.SQL_NULL_HSTMT) {
				OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_DROP);
				hStmt = OdbcDef.SQL_NULL_HSTMT;
                                FreeParams();
			}
		}
		catch (SQLException ex) {
			// If we get an error, ignore
		}

		// Remove this Statement object from the Connection object's
		// list

		myConnection.deregisterStatement (this);
	}

public synchronized void FreeParams()
        throws NullPointerException
{
    try {
     for (int pindex=1; pindex <= boundParams.length; pindex++)
           {
           if (boundParams[pindex-1].pA1!=0) 
               OdbcApi.ReleaseStoredBytes (boundParams[pindex-1].pA1, boundParams[pindex-1].pA2);
           if (boundParams[pindex-1].pB1!=0)
               OdbcApi.ReleaseStoredBytes (boundParams[pindex-1].pB1, boundParams[pindex-1].pB2);
           if (boundParams[pindex-1].pC1!=0)
               OdbcApi.ReleaseStoredBytes (boundParams[pindex-1].pC1, boundParams[pindex-1].pC2);
           if (boundParams[pindex-1].pS1!=0)
               OdbcApi.ReleaseStoredChars (boundParams[pindex-1].pS1, boundParams[pindex-1].pS2);
            }
         }
         catch (NullPointerException npx)
          {
          //Do nothing
          }
} 


	//====================================================================
	// Data attributes
	//====================================================================

	private boolean lastParameterNull;	// true if the last parameter
						//  referenced by a getXXX
						//  function was null





    //--------------------------JDBC 2.0-----------------------------

    /**
     * JDBC 2.0
     *
     * Get the value of a NUMERIC parameter as a java.math.BigDecimal object.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @return the parameter value (full precision); if the value is SQL NULL, 
     * the result is null 
     * @exception SQLException if a database-access error occurs.
     */
    public 
    BigDecimal getBigDecimal(int parameterIndex) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Returns an object representing the value of OUT parameter @i.
     * Use the @map to determine the class from which to construct 
     * data of SQL structured and distinct types.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param map the mapping from SQL type names to Java classes
     * @return a java.lang.Object holding the OUT parameter value.
     * @exception SQLException if a database-access error occurs.
     */
    public 
     Object  getObject (int i, java.util.Map map) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a REF(&lt;structured-type&gt;) OUT parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @return an object representing data of an SQL REF Type
     * @exception SQLException if a database-access error occurs.
     */
    public 
     Ref getRef (int i) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a BLOB OUT parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @return an object representing a BLOB
     * @exception SQLException if a database-access error occurs.
     */
    public 
     Blob getBlob (int i) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get a CLOB OUT parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @return an object representing a CLOB
     * @exception SQLException if a database-access error occurs.
     */
    public 
     Clob getClob (int i) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Get an Array OUT parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @return an object representing an SQL array
     * @exception SQLException if a database-access error occurs.
     */
    public 
     Array getArray (int i) { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a SQL DATE parameter as a java.sql.Date object
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @return the parameter value; if the value is SQL NULL, the result is 
     * null
     * @exception SQLException if a database-access error occurs.
     */
    public 
    java.sql.Date getDate(int parameterIndex, Calendar cal) 
      { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a SQL TIME parameter as a java.sql.Time object.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @return the parameter value; if the value is SQL NULL, the result is 
	 * null
     * @exception SQLException if a database-access error occurs.
     */
    public 
    java.sql.Time getTime(int parameterIndex, Calendar cal) 
      { throw new UnsupportedOperationException(); }

    /**
     * Get the value of a SQL TIMESTAMP parameter as a java.sql.Timestamp 
     * object.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @return the parameter value; if the value is SQL NULL, the result is 
     * null
     * @exception SQLException if a database-access error occurs.
     */
    public 
    java.sql.Timestamp getTimestamp(int parameterIndex, Calendar cal) 
      { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Register an output parameter.  This version of registerOutParameter 
     * should be used for a user-named or REF output parameter.  Examples
     * of user-named types include: STRUCT, DISTINCT, JAVA_OBJECT, and 
     * named array types.
     *
     * Before executing a stored procedure call, you must explicitly
     * call registerOutParameter to register the java.sql.Type of each
     * out parameter.  For a user-named parameter the fully qualified SQL 
     * type name of the parameter should also be given, while a REF 
     * parameter requires that the fully qualified type name of the
     * referenced type be given.  A JDBC driver that does not need the 
     * type code and type name information may ignore it.   To be portable, 
     * however, applications should always provide these values for 
     * user-named and REF parameters.
     *
     * Although it is intended for user-named and Ref parameters,
     * this method may be used to register a parameter of any JDBC type.
     * If the parameter does not have a user-named or REF type then the
     * typeName is ignored.
     *
     * <P><B>Note:</B> When reading the value of an out parameter, you
     * must use the getXXX method whose Java type XXX corresponds to the
     * parameter's registered SQL type.
     *
     * @param parameterIndex the first parameter is 1, the second is 2,...
     * @param sqlType a value from java.sql.Types
     * @param typeName the fully qualified name of a SQL structured type
     * @exception SQLException if a database-access error occurs.
     * @see Type 
     */
    public
    void registerOutParameter (int paramIndex, int sqlType, String typeName) 
      throws SQLException
    { throw new UnsupportedOperationException(); }
}

