/*
 * @(#)JdbcOdbcPreparedStatement.java	1.21 98/09/09
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcPreparedStatement.java
//
// Description: Impementation of the PreparedStatement interface class
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.util.Hashtable;
import java.util.Calendar;
import java.math.BigDecimal;
import java.sql.*;

public class JdbcOdbcPreparedStatement
	extends		JdbcOdbcStatement
	implements	java.sql.PreparedStatement {

	//====================================================================
	// Public methods
	//====================================================================

	//--------------------------------------------------------------------
	// Constructor
	// Perform any necessary initialization.
	//--------------------------------------------------------------------

	public JdbcOdbcPreparedStatement (
		JdbcOdbcConnectionInterface con)
	{
		super (con);
	}

	//--------------------------------------------------------------------
	// initialize
	// Initialize the result set object.  Give the ODBC API interface
	// object, the connection handle, and optionally the statement
	// handle.  If no statement handle is given, one is created.
	//--------------------------------------------------------------------

	public void initialize (
		JdbcOdbc odbcApi,
		int hdbc)
		throws SQLException
	{
		super.initialize (odbcApi, hdbc);

		// Initialize the bound parameter objects

		initBoundParam ();
	}

	public void initialize (
		JdbcOdbc odbcApi,
		int hdbc,
		int hstmt,
		Hashtable info)
		throws SQLException
	{
		super.initialize (odbcApi, hdbc, hstmt, info);

		// Initialize the bound parameter objects

		initBoundParam ();
	}

	//--------------------------------------------------------------------
	// executeQuery
	// This method executes a "prepared" query statement.
	// See the definition of Statement.executeQuery
	//--------------------------------------------------------------------

	public ResultSet executeQuery ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*PreparedStatement.executeQuery");
		}
		ResultSet rs = null;

		if (execute ()) {
			rs = getResultSet (false);
		}
		else {

			// No ResultSet was produced.  Raise an exception

			throw new SQLException ("No ResultSet was produced");
		}
		return rs;
	}

	//--------------------------------------------------------------------
	// executeQuery
	// executeQuery with a sql String is invalid for prepared statements
	//--------------------------------------------------------------------

	public ResultSet executeQuery (
		String sql)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*PreparedStatement.executeQuery (" + sql + ")");
		}
		throw new SQLException ("Driver does not support this function",
					"IM001");
	}

	//--------------------------------------------------------------------
	// executeUpdate
	// This method executes a "prepared" modify statement.
	// See the definition of Statement.executeUpdate
	//--------------------------------------------------------------------

	public int executeUpdate ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*PreparedStatement.executeUpdate");
		}
		int numRows = -1;

		// Execute the statement.  If execute returns false, a
		// row count exists.

		if (!execute ()) {
			numRows = getUpdateCount ();
		}
		else {

			// No update count was produced (a ResultSet was).  Raise
			// an exception

			throw new SQLException ("No row count was produced");
		}
		return numRows;
	}

	//--------------------------------------------------------------------
	// executeUpdate
	// executeUpdate with a sql String is invalid for prepared statements
	//--------------------------------------------------------------------

	public int executeUpdate (
		String sql)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*PreparedStatement.executeUpdate (" + sql + ")");
		}
		throw new SQLException ("Driver does not support this function",
					"IM001");
	}

	//--------------------------------------------------------------------
	// execute
	// This method executed an arbitrary "prepared" statement.
	// See the definition of Statement.execute
	//--------------------------------------------------------------------

	public synchronized boolean execute ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*PreparedStatement.execute");
		}
		boolean hasResultSet = false;
		SQLWarning	warning = null;
		boolean needData = false;

		// Reset warnings

		clearWarnings ();

		// Call SQLExecute

		try {
			needData = OdbcApi.SQLExecute (hStmt);

			// Now loop while more data is needed (i.e. a data-at-
			// execution parameter was given).  For each parameter
			// that needs data, put the data from the input stream.

			while (needData) {

				// Get the parameter number that requires data

				int paramIndex = OdbcApi.SQLParamData (hStmt);
			
				// If the parameter index is -1, there is no
				// more data required

				if (paramIndex == -1) {
					needData = false;
				}
				else {
					// Now we have the proper parameter
					// index, get the data from the input
					// stream and do a SQLPutData
					putParamData (paramIndex);
				}
			}

		}
		catch (SQLWarning ex) {

			// Save pointer to warning and save with ResultSet
			// object once it is created.

			warning = ex;
		}
			
		// Now loop while more data is needed (i.e. a data-at-
		// execution parameter was given).  For each parameter
		// that needs data, put the data from the input stream.

		while (needData) {

			// Get the parameter number that requires data

			int paramIndex = OdbcApi.SQLParamData (hStmt);
			
			// If the parameter index is -1, there is no more
			// data required

			if (paramIndex == -1) {
				needData = false;
			}
			else {
				// Now we have the proper parameter index,
				// get the data from the input stream
				// and do a SQLPutData
				putParamData (paramIndex);
			}
		}

		// Now determine if there is a result set associated with
		// the SQL statement that was executed.  Get the column
		// count, and if it is not zero, there is a result set.

		if (getColumnCount () > 0) {
			hasResultSet = true;
		}

		return hasResultSet;
	}

	// Methods for setting IN parameters into this statement.
	// Parameters are numbered starting at 1.

	//--------------------------------------------------------------------
	// setNull
	// You always need to specify a SQL type when sending a NULL.
	//--------------------------------------------------------------------

	public void setNull (
		int parameterIndex,
		int sqlType)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*PreparedStatement.setNull (" +
				parameterIndex + "," + sqlType + ")");
		}
		// Get the buffer needed for the length

		byte lenBuf[] = getLengthBuf (parameterIndex);


                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		// Bind the parameter to NULL
                                
		OdbcApi.SQLBindInParameterNull (hStmt, parameterIndex, 
                                sqlType, lenBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 
        }


	//--------------------------------------------------------------------
	// The following methods allow you to set various SQLtypes as
	// parameters.
	// Note that the method include the SQL type in their name.
	//--------------------------------------------------------------------

	//--------------------------------------------------------------------
	// setBoolean
	//--------------------------------------------------------------------

	public void setBoolean (
		int parameterIndex,
		boolean x)
		throws SQLException
	{
		int value = 0;

		// If the parameter is true, set the value to 1
		if (x) {
			value = 1;
		}

		// Set the parameter as if it were an integer
		setInt (parameterIndex, value);
	}

	//--------------------------------------------------------------------
	// setByte
	//--------------------------------------------------------------------

	public void setByte (
		int parameterIndex,
		byte x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 4);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;


		OdbcApi.SQLBindInParameterInteger (hStmt, parameterIndex,
                                Types.TINYINT, x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 

	}

	//--------------------------------------------------------------------
	// setShort
	//--------------------------------------------------------------------

	public void setShort (
		int parameterIndex,
		short x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 4);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterInteger (hStmt, parameterIndex,
                                Types.SMALLINT, x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 
	}

	//--------------------------------------------------------------------
	// setInt
	//--------------------------------------------------------------------

	public void setInt (
		int parameterIndex,
		int x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 4);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterInteger (hStmt, parameterIndex,
                                Types.INTEGER, x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 

	}

	//--------------------------------------------------------------------
	// setLong
	//--------------------------------------------------------------------

	public void setLong (
		int parameterIndex,
		long x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 8);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;


		OdbcApi.SQLBindInParameterFloat (hStmt, parameterIndex,
                                Types.BIGINT, 0, (float) x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 

	}

	//--------------------------------------------------------------------
	// setReal
	//--------------------------------------------------------------------

	public void setReal (
		int parameterIndex,
		float x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 8);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterFloat (hStmt, parameterIndex,
                                Types.REAL, 0, x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 

	}

	//--------------------------------------------------------------------
	// setFloat
	//--------------------------------------------------------------------

	public void setFloat (
		int parameterIndex,
		float x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 8);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterFloat (hStmt, parameterIndex,
                                Types.FLOAT, 0, x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 
	}

	//--------------------------------------------------------------------
	// setDouble
	//--------------------------------------------------------------------

	public void setDouble (
		int parameterIndex,
		double x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 8);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterDouble (hStmt, parameterIndex,
                                Types.DOUBLE, 0, x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 
	}

	//--------------------------------------------------------------------
	// setBigDecimal
	//--------------------------------------------------------------------

	public void setBigDecimal (
		int parameterIndex,
		BigDecimal x) 
		throws SQLException
	{

		// Bind the parameter as a CHAR.  We could bind as a
		// double, but this may result in a loss of precision

		setChar (parameterIndex, Types.NUMERIC, x.scale (), x.toString ());

	}

	//--------------------------------------------------------------------
	// setDecimal
	//--------------------------------------------------------------------

	public void setDecimal (
		int parameterIndex,
		BigDecimal x) 
		throws SQLException
	{
		// Bind the parameter as a CHAR.  We could bind as a
		// double, but this may result in a loss of precision

		setChar (parameterIndex, Types.CHAR, 0, x.toString ());
	}

	//--------------------------------------------------------------------
	// setString
	//--------------------------------------------------------------------

	public void setString (
		int parameterIndex,
		String x)
		throws SQLException
	{
		setChar (parameterIndex, Types.CHAR, 0, x);
	}

	//--------------------------------------------------------------------
	// setBytes
	//--------------------------------------------------------------------

	public void setBytes (
		int parameterIndex,
		byte x[])
		throws SQLException
	{
		setBinary (parameterIndex, Types.BINARY, x);
	}

	//--------------------------------------------------------------------
	// setDate
	//--------------------------------------------------------------------

	public void setDate (
		int parameterIndex,
		java.sql.Date x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 32);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterDate (hStmt, parameterIndex,
                                x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 


	}

	//--------------------------------------------------------------------
	// setTime
	//--------------------------------------------------------------------

	public void setTime (
		int parameterIndex,
		java.sql.Time x) 
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 32);

                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterTime (hStmt, parameterIndex,
                                x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 

	}

	//--------------------------------------------------------------------
	// setTimestamp
	//--------------------------------------------------------------------

	public void setTimestamp (
		int parameterIndex,
		java.sql.Timestamp x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex, 32);


                int buffers[]=new int[2];
                buffers[0]=0;
                buffers[1]=0;

		OdbcApi.SQLBindInParameterTimestamp (hStmt, parameterIndex,
                                x, bindBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 

	}

	//--------------------------------------------------------------------
	// The normal setString and setBytes methods are suitable for passing
	// normal sized data.  However occasionally it may be necessary to send
	// extremely large values as LONGVARCHAR or LONGVARBINARY parameters.
	// In this case you can pass in a java.io.InputStream object, and the
	// JDBC runtimes will read data from that stream as needed, until they
	// reach end-of-file.  Note that these stream objects can either be
	// standard Java stream objects, or your own subclass that implements
	// the standard interface.
	// setAsciiStreamParameter and setUnicodeStreamParameter imply the use
	// of the SQL LONGVARCHAR type, and setBinaryStreamParameter implies
	// the SQL LONGVARBINARY type.
	// For each stream type you must specify the number of bytes to be
	// read from the stream and sent to the database.
	//--------------------------------------------------------------------

	public void setAsciiStream (
		int ParameterIndex,
		java.io.InputStream x,
		int length)
		throws SQLException
	{
		setStream (ParameterIndex, x, length, Types.LONGVARCHAR,
							JdbcOdbcBoundParam.ASCII);
	}

	public void setUnicodeStream (
		int ParameterIndex,
		java.io.InputStream x,
		int length)
		throws SQLException
	{
		setStream (ParameterIndex, x, length, Types.LONGVARCHAR,
							JdbcOdbcBoundParam.UNICODE);
	}

	public void setBinaryStream (
		int ParameterIndex,
		java.io.InputStream x,
		int length) 
		throws SQLException
	{
		setStream (ParameterIndex, x, length, Types.LONGVARBINARY,
							JdbcOdbcBoundParam.BINARY);
	}


	// Finally, if you need a more dynamic interface, or if you need
	// the driver to perform a supported coercion, then you can use
	// setObjectParameter.
	// You must explicitly specify the SQL type that you want.
	// The supported sub-types of Object are java.lang.Boolean,
	// java.lang.Byte, java.lang.Short, java.lang.Integer, java.lang.Long,
	// java.lang.Float, java.lang.Double, java.lang.String, 
	// Numeric, java.io.inputStream, byte[], 
	// java.util.Date, and their subtypes.  You can also use a Java null.

	//--------------------------------------------------------------------
	// clearParameters
	// Parameters remain in force for repeated use of the same statement.
	// You can use clearParameters to remove any parameters associated with
	// a statement.
	//--------------------------------------------------------------------

	public void clearParameters ()
		throws SQLException
	{
		OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_RESET_PARAMS);
	}



	//--------------------------------------------------------------------
	// setObject
	// You can set a parameter as a Java object.  See the JDBC spec's 
	// "Dynamic Programming" chapter for information on valid Java types.
	//--------------------------------------------------------------------

	public void setObject (
		int parameterIndex,
		Object x)
		throws SQLException
	{
		setObject (parameterIndex, x, getTypeFromObject (x));
	}

	public void setObject (
		int parameterIndex,
		Object x,
		int sqlType)
		throws SQLException
	{
		setObject (parameterIndex, x, sqlType, 0);
	}

	public void setObject (
		int parameterIndex,
		Object x,
		int sqlType,
		int scale)
		throws SQLException
	{
	
		// For each known SQL Type, call the appropriate
		// set routine

		switch (sqlType) {
		case Types.CHAR:
			setString (parameterIndex, (String) x);
			break;

		case Types.VARCHAR:
			setChar (parameterIndex, sqlType, 0, (String) x);
			break;

		case Types.LONGVARCHAR:
			setChar (parameterIndex, sqlType, 0, (String) x);
			break;

		case Types.NUMERIC:
			setBigDecimal (parameterIndex, (BigDecimal) x);
			break;

		case Types.DECIMAL:
			setDecimal (parameterIndex, (BigDecimal) x);
			break;

		case Types.BIT:
			setBoolean (parameterIndex,
					((Boolean) x).booleanValue ());
			break;

		case Types.TINYINT:
			setByte (parameterIndex, (byte)
					((Integer) x).intValue ());
			break;

		case Types.SMALLINT:
			setShort (parameterIndex, (short)
					((Integer) x).intValue ());
			break;

		case Types.INTEGER:
			setInt (parameterIndex,
					((Integer) x).intValue ());
			break;

		case Types.BIGINT:
			setLong (parameterIndex,
					((Integer) x).longValue ());
			break;

		case Types.REAL:
			setReal (parameterIndex, ((Float) x).floatValue ());
			break;

		case Types.FLOAT:
			setFloat (parameterIndex, ((Float) x).floatValue ());
			break;

		case Types.DOUBLE:
			setDouble (parameterIndex,
					((Double) x).doubleValue ());
			break;

		case Types.BINARY:
			setBytes (parameterIndex, (byte[]) x);
			break;

		case Types.VARBINARY:
			setBinary (parameterIndex, sqlType, (byte[]) x);
			break;

		case Types.LONGVARBINARY:
			setBinary (parameterIndex, sqlType, (byte[]) x);
			break;

		case Types.DATE:
			setDate (parameterIndex, (java.sql.Date) x);
			break;

		case Types.TIME:
			setTime (parameterIndex, (java.sql.Time) x);
			break;

		case Types.TIMESTAMP:
			setTimestamp (parameterIndex,
					(java.sql.Timestamp) x);
			break;

		default:
			throw new SQLException ("Unknown SQL Type for PreparedStatement.setObject (SQL Type=" + sqlType);
		}
	}

	//====================================================================
	// Protected methods
	//====================================================================

	//--------------------------------------------------------------------
	// initBoundParam
	// Initialize the bound parameter objects
	//--------------------------------------------------------------------

	protected void initBoundParam ()
		throws SQLException
	{
		// Get the number of parameters

		numParams = OdbcApi.SQLNumParams (hStmt);

		// There are parameter markers, allocate the bound
		// parameter objects

		if (numParams > 0) {

			// Allocate an array of bound parameter objects

			boundParams = new JdbcOdbcBoundParam[numParams];

			// Allocate and initialize each bound parameter
			
			for (int i = 0; i < numParams; i++) {
				boundParams[i] = new JdbcOdbcBoundParam ();
				boundParams[i].initialize ();
			}
		}
	}

	//--------------------------------------------------------------------
	// allocBindBuf
	// Allocate storage for the permanent data buffer for the bound
	// parameter.
	//--------------------------------------------------------------------
	
	protected byte[] allocBindBuf (
		int index,
		int bufLen)
	{
		byte b[] = null;

		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {
			b = boundParams[index - 1].allocBindDataBuffer (
						bufLen);
		}

		return b;
	}

	//--------------------------------------------------------------------
	// getDataBuf
	// Gets the data buffer for the given parameter index
	//--------------------------------------------------------------------
	
	protected byte[] getDataBuf (
		int index)
	{
		byte b[] = null;

		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {
			b = boundParams[index - 1].getBindDataBuffer ();
		}

		return b;
	}

	//--------------------------------------------------------------------
	// getLengthBuf
	// Gets the length buffer for the given parameter index
	//--------------------------------------------------------------------
	
	protected byte[] getLengthBuf (
		int index)
	{
		byte b[] = null;

		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {
			b = boundParams[index - 1].getBindLengthBuffer ();
		}

		return b;
	}

	//--------------------------------------------------------------------
	// getParamLength
	// Returns the length of the given parameter number.  When each
	// parameter was bound, a 4-byte buffer was given to hold the 
	// length (stored in native format).  Get the buffer, convert the
	// buffer from native format, and return it.  If the length is -1,
	// the column is considered to be NULL.
	//--------------------------------------------------------------------

	public int getParamLength (
		int index)
	{
		int paramLen = OdbcDef.SQL_NULL_DATA;
	
		// Sanity check the parameter number

		if ((index >= 1) &&
		    (index <= numParams)) {

			// Now get the length of the parameter from the
			// bound param array.  -1 is returned if it is null.
			paramLen = OdbcApi.bufferToInt (
				boundParams[index - 1].getBindLengthBuffer ());
		}
		return paramLen;
	}

	//--------------------------------------------------------------------
	// putParamData
	// Puts parameter data from a previously bound input stream.  The
	// input stream was bound using SQL_LEN_DATA_AT_EXEC.
	//--------------------------------------------------------------------
	
	protected void putParamData (
		int index)
		throws SQLException, JdbcOdbcSQLWarning
	{
		// We'll transfer up to maxLen at a time
		int	maxLen = JdbcOdbcLimits.MAX_PUT_DATA_LENGTH;
		int	bufLen;
		int realLen;
		byte	buf[] = new byte[maxLen];
		boolean	endOfStream = false;

		// Sanity check the parameter index
		if ((index < 1) ||
		    (index > numParams)) {
			if (isTracing ()) {
				trace ("Invalid index for putParamData()");
			}
			return;
		}

		// Get the information about the input stream

		java.io.InputStream inputStream =
					boundParams[index - 1].getInputStream ();
		int inputStreamLen = boundParams[index - 1].getInputStreamLen ();
		int inputStreamType = boundParams[index - 1].getStreamType ();

		// Loop while more data from the input stream

		while (!endOfStream) {

			// Read some data from the input stream

			try {
				if (isTracing ()) {
					trace ("Reading from input stream"); 
				}
				bufLen = inputStream.read (buf);
				if (isTracing ()) {
					trace ("Bytes read: " + bufLen);
				}
			}
			catch (java.io.IOException ex) {

				// If an I/O exception was generated, turn
				// it into a SQLException

				throw new SQLException (ex.getMessage ());
			}

			// -1 as the number of bytes read indicates that
			// there is no more data in the input stream

			if (bufLen == -1) {

				// Sanity check to ensure that all the data we said we
				// had was read.  If not, raise an exception

				if (inputStreamLen != 0) {
					throw new SQLException ("End of InputStream reached before satisfying length specified when InputStream was set");
				}
				endOfStream = true;
				break;
			}
			
			// If we got more bytes than necessary, truncate
			// the buffer by re-setting the buffer length.  Also,
			// indicate that we don't need to read any more.

			if (bufLen > inputStreamLen) {
				bufLen = inputStreamLen;
				endOfStream = true;
			}

			realLen = bufLen;
			
			// For UNICODE streams, strip off the high byte and set the
			// number of actual bytes present.  It is assumed that
			// there are 2 bytes present for every UNICODE character - if
			// not, then that's not our problem

			if (inputStreamType == JdbcOdbcBoundParam.UNICODE) {
				realLen = bufLen / 2;

				for (int ii = 0; ii < realLen; ii++) {
					buf[ii] = buf[(ii * 2) + 1];
				}
			}

			// Put the data

			OdbcApi.SQLPutData (hStmt, buf, realLen);

			// Decrement the number of bytes still needed

			inputStreamLen -= bufLen;

			if (isTracing ()) {
				trace ("" + inputStreamLen + " bytes remaining");
			}

			// If there is no more data to be read, exit loop

			if (inputStreamLen == 0) {
				endOfStream = true;
			}
		}
	}

	//--------------------------------------------------------------------
	// setStream
	// Sets an input stream as a parameter, using the given SQL type
	//--------------------------------------------------------------------

	public void setStream (
		int ParameterIndex,
		java.io.InputStream x,
		int length,
		int SQLtype,
		int streamType)
		throws SQLException
	{
		// Get the buffer needed for the length

		byte lenBuf[] = getLengthBuf (ParameterIndex);

		// Allocate a new buffer for the parameter data.  This buffer
		// will be returned by SQLParamData (it is set to the parameter
		// number, a 4-byte integer)

		byte dataBuf[] = allocBindBuf (ParameterIndex, 4);

                // Bind the parameter with SQL_LEN_DATA_AT_EXEC

                int buffers[]=new int[4];

                buffers[0]=0;
                buffers[1]=0;
                buffers[2]=0;
                buffers[3]=0;

		OdbcApi.SQLBindInParameterAtExec (hStmt, ParameterIndex,
                                SQLtype, length, dataBuf, lenBuf, buffers);

               //save the native pointers from Garbage Collection
               boundParams[ParameterIndex - 1].pA1=buffers[0];
               boundParams[ParameterIndex - 1].pA2=buffers[1]; 
               boundParams[ParameterIndex - 1].pB1=buffers[2];
               boundParams[ParameterIndex - 1].pB2=buffers[3]; 


		// Save the input stream

		boundParams[ParameterIndex - 1].setInputStream (x, length);

		// Set the stream type

		boundParams[ParameterIndex - 1].setStreamType (streamType);
	}

	//--------------------------------------------------------------------
	// setChar
	// Binds the given string to the given SQL type
	//--------------------------------------------------------------------

	protected void setChar (
		int parameterIndex,
		int SQLtype,
		int scale,
		String x)
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

                char[] xChars = x.toCharArray();
                byte xValue[] = new byte[0];
                try {
                   xValue = CharsToBytes (OdbcApi.charSet, xChars);
                } catch (java.io.UnsupportedEncodingException exx) {}


		byte bindBuf[] = allocBindBuf (parameterIndex,
                                (xValue.length + 1));

		// Get the precision for this SQL type.  If the precision
		// is out of bounds, set it to our default

		int precision = getPrecision (SQLtype);

		if ((precision < 0) ||
			(precision > JdbcOdbcLimits.DEFAULT_IN_PRECISION)) {

			precision = JdbcOdbcLimits.DEFAULT_IN_PRECISION;
		}

                int buffers[]=new int[4];

                buffers[0]=0;
                buffers[1]=0;
                buffers[2]=0;
                buffers[3]=0;

		OdbcApi.SQLBindInParameterString (hStmt, parameterIndex,
                                SQLtype, precision, scale, xValue, bindBuf, buffers);

               //Save the pointers from the trash
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 
               boundParams[parameterIndex - 1].pB1=buffers[2];
               boundParams[parameterIndex - 1].pB2=buffers[3]; 
	}

	//--------------------------------------------------------------------
	// setBinary
	// Binds the given byte array to the given SQL type
	//--------------------------------------------------------------------

	protected void setBinary (
		int parameterIndex,
		int SQLtype,
		byte x[])
		throws SQLException
	{
		// Allocate a buffer to be used in binding.  This will be
		// a 'permanent' buffer that the bridge will fill in with
		// the bound data in native format.

		byte bindBuf[] = allocBindBuf (parameterIndex,
				x.length);

		// Get the buffer needed for the length

		byte lenBuf[] = getLengthBuf (parameterIndex);

                int buffers[]=new int[6];

                buffers[0]=0;
                buffers[1]=0;
                buffers[2]=0;
                buffers[3]=0;
                buffers[4]=0;
                buffers[5]=0;


		OdbcApi.SQLBindInParameterBinary (hStmt, parameterIndex,
                                SQLtype, x, bindBuf, lenBuf, buffers);

               //Save the pointers from the trash
               boundParams[parameterIndex - 1].pA1=buffers[0];
               boundParams[parameterIndex - 1].pA2=buffers[1]; 
               boundParams[parameterIndex - 1].pB1=buffers[2];
               boundParams[parameterIndex - 1].pB2=buffers[3]; 
               boundParams[parameterIndex - 1].pC1=buffers[4];
               boundParams[parameterIndex - 1].pC2=buffers[5]; 

	}

	//--------------------------------------------------------------------
	// getTypeFromObject
	// Given an object of unknown type, return the Java SQL type that
	// it can be casted to.
	//--------------------------------------------------------------------

	public static int getTypeFromObject (
		Object x)
	{
		if (x == null) {
			return Types.NULL;
		}

		// Try a String type

		try {
			if ((String) x != null) return Types.VARCHAR;
		}
		catch (Exception e) {}

		// Try a Numeric type

		try {
			if ((BigDecimal) x != null) return Types.NUMERIC;
		}
		catch (Exception e) {}

		// Try a Boolean type

		try {
			if ((Boolean) x != null) return Types.BIT;
		}
		catch (Exception e) {}

		// Try an Integer type

		try {
			if ((Integer) x != null) return Types.INTEGER;
		}
		catch (Exception e) {}

		// Try a Long type

		try {
			if ((Long) x != null) return Types.BIGINT;
		}
		catch (Exception e) {}

		// Try a Float type

		try {
			if ((Float) x != null) return Types.FLOAT;
		}
		catch (Exception e) {}

		// Try a Double type

		try {
			if ((Double) x != null) return Types.DOUBLE;
		}
		catch (Exception e) {}

		// Try a byte[] type

		try {
			if ((byte[]) x != null) return Types.VARBINARY;
		}
		catch (Exception e) {}

		// Try a Date type

		try {
			if ((java.sql.Date) x != null) return Types.DATE;
		}
		catch (Exception e) {}

		// Try a Time type

		try {
			if ((java.sql.Time) x != null) return Types.TIME;
		}
		catch (Exception e) {}

		// Try a Timestamp type

		try {
			if ((java.sql.Timestamp) x != null)
					return Types.TIMESTAMP;
		}
		catch (Exception e) {}

		return Types.OTHER;
	}


       public synchronized void close ()
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*Statement.close");
		}

		// Close/clear our result set

		clearMyResultSet ();

		// Reset last warning message

		try {
			clearWarnings ();
			if (hStmt != OdbcDef.SQL_NULL_HSTMT) {
				OdbcApi.SQLFreeStmt (hStmt, OdbcDef.SQL_DROP);
				hStmt = OdbcDef.SQL_NULL_HSTMT;
                                FreeParams();
			}
		}
		catch (SQLException ex) {
			// If we get an error, ignore
		}

		// Remove this Statement object from the Connection object's
		// list

		myConnection.deregisterStatement (this);
	}

public synchronized void FreeParams()
        throws NullPointerException
{
    try {
     for (int pindex=1; pindex <= boundParams.length; pindex++)
           {
           if (boundParams[pindex-1].pA1!=0) 
               OdbcApi.ReleaseStoredBytes (boundParams[pindex-1].pA1, boundParams[pindex-1].pA2);
           if (boundParams[pindex-1].pB1!=0)
               OdbcApi.ReleaseStoredBytes (boundParams[pindex-1].pB1, boundParams[pindex-1].pB2);
           if (boundParams[pindex-1].pC1!=0)
               OdbcApi.ReleaseStoredBytes (boundParams[pindex-1].pC1, boundParams[pindex-1].pC2);
           if (boundParams[pindex-1].pS1!=0)
               OdbcApi.ReleaseStoredChars (boundParams[pindex-1].pS1, boundParams[pindex-1].pS2);
            }
         }
         catch (NullPointerException npx)
          {
          //Do nothing
          }
} 

	//====================================================================
	// Data attributes
	//====================================================================

	protected int numParams;		// Number of parameter markers
						//  for the prepared statement

	protected JdbcOdbcBoundParam boundParams[];
						// Array of bound parameter
						//  objects.  Each parameter
						//  marker will have a 
						//  corresponding object to
						//  hold bind information, and
						//  resulting data.


     //--------------------------JDBC 2.0-----------------------------

    /**
     * JDBC 2.0
     *
     * Add a set of parameters to the batch.
     * 
     * @exception SQLException if a database-access error occurs.
     * @see Statement#addBatch
     */
    public 
    void addBatch() { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * When a very large UNICODE value is input to a LONGVARCHAR
     * parameter, it may be more practical to send it via a
     * java.io.Reader. JDBC will read the data from the stream
     * as needed, until it reaches end-of-file.  The JDBC driver will
     * do any necessary conversion from UNICODE to the database char format.
     * 
     * <P><B>Note:</B> This stream object can either be a standard
     * Java stream object or your own subclass that implements the
     * standard interface.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the java reader which contains the UNICODE data
     * @param length the number of characters in the stream 
     * @exception SQLException if a database-access error occurs.
     */
    public 
    void setCharacterStream(int parameterIndex,
       			  java.io.Reader reader,
			  int length) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Set a REF(&lt;structured-type&gt;) parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing data of an SQL REF Type
     */
    public 
    void setRef (int i, Ref x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Set a BLOB parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing a BLOB
     */
    public 
    void setBlob (int i, Blob x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Set a CLOB parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing a CLOB
     */
    public 
    void setClob (int i, Clob x) { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Set an Array parameter.
     *
     * @param i the first parameter is 1, the second is 2, ...
     * @param x an object representing an SQL array
     */
    public 
    void setArray (int i, Array x) { throw new UnsupportedOperationException(); }

    /**
     * The number, types and properties of a ResultSet's columns
     * are provided by the getMetaData method.
     *
     * @return the description of a ResultSet's columns
     * @exception SQLException if a database-access error occurs.
     */
    public 
    ResultSetMetaData getMetaData() { throw new UnsupportedOperationException(); }

    /**
     * Set a parameter to a java.sql.Date value.  The driver converts this
     * to a SQL DATE value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database-access error occurs.
     */
    public 
    void setDate(int parameterIndex, java.sql.Date x, Calendar cal)
	    { throw new UnsupportedOperationException(); }

    /**
     * Set a parameter to a java.sql.Time value.  The driver converts this
     * to a SQL TIME value when it sends it to the database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value
     * @exception SQLException if a database-access error occurs.
     */
    public 
    void setTime(int parameterIndex, java.sql.Time x, Calendar cal) 
	    { throw new UnsupportedOperationException(); }

    /**
     * Set a parameter to a java.sql.Timestamp value.  The driver
     * converts this to a SQL TIMESTAMP value when it sends it to the
     * database.
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param x the parameter value 
     * @exception SQLException if a database-access error occurs.
     */
    public 
    void setTimestamp(int parameterIndex, java.sql.Timestamp x, Calendar cal)
	    { throw new UnsupportedOperationException(); }

    /**
     * JDBC 2.0
     *
     * Set a parameter to SQL NULL.  This version of setNull should
     * be used for user-named types and REF type parameters.  Examples
     * of user-named types include: STRUCT, DISTINCT, JAVA_OBJECT, and 
     * named array types.
     *
     * <P><B>Note:</B> To be portable, applications must give the
     * SQL type code and the fully qualified SQL type name when specifying
     * a NULL user-named or REF parameter.  In the case of a user-named type 
     * the name is the type name of the parameter itself.  For a REF 
     * parameter the name is the type name of the referenced type.  If 
     * a JDBC driver does not need the type code or type name information, 
     * it may ignore it.     
     *
     * Although it is intended for user-named and Ref parameters,
     * this method may be used to set a null parameter of any JDBC type.
     * If the parameter does not have a user-named or REF type then the
     * typeName is ignored.
     *
     *
     * @param parameterIndex the first parameter is 1, the second is 2, ...
     * @param sqlType a value from java.sql.Types
     * @param typeName the fully qualified name of a SQL user-named type,
     *  ignored if the parameter is not a user-named type or REF 
     * @exception SQLException if a database-access error occurs.
     */
    public
    void setNull (int paramIndex, int sqlType, String typeName) 
      throws SQLException
    { throw new UnsupportedOperationException(); }
}

