/*
 * @(#)JdbcOdbcResultSetInterface.java	1.15 98/04/28
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcResultSetInterface.java
//
// Description: Definition of the JdbcOdbcResultSet interface.  This is
//              necessary to 'forward define' the class to eliminate
//              circular dependencies
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        May, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.sql.*;

public interface JdbcOdbcResultSetInterface
	extends		java.sql.ResultSet
{

	//--------------------------------------------------------------------
	// getPseudoCol
	// Given a column number, return the corresponding pseudo column,
	// null if one does not exist for this column
	//--------------------------------------------------------------------

	public JdbcOdbcPseudoCol getPseudoCol (
		int column);

	//--------------------------------------------------------------------
	// mapColumn
	// Given a column number, map it to the corresponding column in
	// the result set.  If no column mappings exist, return the original
	// column number.  If the column number is out-of-range, return -1.
	//--------------------------------------------------------------------

	public int mapColumn (
		int	column);

	//--------------------------------------------------------------------
	// clearWarnings
	// Clears any warning information for the statement
	//--------------------------------------------------------------------

	public void clearWarnings()
			throws SQLException;

	//--------------------------------------------------------------------
	// getHSTMT
	// Returns the statement handle for the result set
	//--------------------------------------------------------------------
	
	public int getHSTMT ();

	//--------------------------------------------------------------------
	// getColumnCount
	// Return the number of columns in the ResultSet
	//--------------------------------------------------------------------

	public int getColumnCount ()
		throws SQLException;

	//--------------------------------------------------------------------
	// getScale
	// Number of digits to right of decimal
	//--------------------------------------------------------------------

	public int getScale (
		int column)
		throws SQLException;

	//--------------------------------------------------------------------
	// getColumnType
	// Returns the Java SQL type of the given column number.
	//--------------------------------------------------------------------

	public int getColumnType (
		int column)
		throws SQLException;

	//--------------------------------------------------------------------
	// getColAttribute
	// Given the column and attribute type, return the attribute value
	//--------------------------------------------------------------------

	public int getColAttribute (
		int column,
		int type)
		throws SQLException;

	//--------------------------------------------------------------------
	// setWarning
	// Sets the warning 
	//--------------------------------------------------------------------

	public void setWarning (
			SQLWarning ex)
			throws SQLException;

}
