/*
 * @(#)JdbcOdbcResultSetMetaData.java	1.15 98/04/28
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
//----------------------------------------------------------------------------
//
// Module:      JdbcOdbcResultSetMetaData.java
//
// Description: Impementation of the ResultSetMetaData interface class
//
// Product:     JDBCODBC (Java DataBase Connectivity using
//              Open DataBase Connectivity)
//
// Author:      Karl Moss
//
// Date:        March, 1996
//
// Copyright:   1990-1996 INTERSOLV, Inc.
//              This software contains confidential and proprietary
//              information of INTERSOLV, Inc.
//----------------------------------------------------------------------------

package sun.jdbc.odbc;

import java.sql.*;

public class JdbcOdbcResultSetMetaData
	extends		JdbcOdbcObject
	implements	java.sql.ResultSetMetaData {

	//====================================================================
	// Public methods
	//====================================================================

	//--------------------------------------------------------------------
	// Constructor
	// Perform any necessary initialization.
	//--------------------------------------------------------------------

	public JdbcOdbcResultSetMetaData (
		JdbcOdbc odbcApi,
		JdbcOdbcResultSetInterface rs)
	{
		// Save a pointer to the ODBC api and the ResultSet object

		OdbcApi = odbcApi;
		resultSet = rs;

		// Set the connection handle

		hStmt = rs.getHSTMT ();
	}

	//--------------------------------------------------------------------
	// getColumnCount
	// The number of columns in the ResultSet.
	//--------------------------------------------------------------------

	public int getColumnCount ()
		throws SQLException
	{
		return resultSet.getColumnCount ();
	}

	//--------------------------------------------------------------------
	// isAutoIncrement
	// Is column automaticallty numbered, thus read-only.
	//--------------------------------------------------------------------

	public boolean isAutoIncrement (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isAutoIncrement (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false

		if (resultSet.getPseudoCol (column) != null) {
			value = false;
		}
		else {
			value = getColAttributeBoolean (column,
				OdbcDef.SQL_COLUMN_AUTO_INCREMENT);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// isCaseSensitive
	// Does case matter?
	//--------------------------------------------------------------------

	public boolean isCaseSensitive (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isCaseSensitive (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false

		if (resultSet.getPseudoCol (column) != null) {
			value = false;
		}
		else {
			value = getColAttributeBoolean (column,
				OdbcDef.SQL_COLUMN_CASE_SENSITIVE);
		}
		return value;
	}


	//--------------------------------------------------------------------
	// isSearchable
	// Can column be used in where clause?
	//--------------------------------------------------------------------

	public boolean isSearchable (
		int column)
		throws SQLException
	{	
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isSearchable (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false

		if (resultSet.getPseudoCol (column) != null) {
			value = false;
		}
		else {
			int n = getColAttribute (column,
					OdbcDef.SQL_COLUMN_SEARCHABLE);
			value = (n != OdbcDef.SQL_UNSEARCHABLE);
		}
		return value;
	}


	//--------------------------------------------------------------------
	// isCurrency
	// Is it a cash value?
	//--------------------------------------------------------------------

	public boolean isCurrency (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isCurrency (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false

		if (resultSet.getPseudoCol (column) != null) {
			value = false;
		}
		else {
			value = getColAttributeBoolean (column,
				OdbcDef.SQL_COLUMN_MONEY);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// isNullable
	// Can you put a null in this column?		
	//--------------------------------------------------------------------

	public int isNullable (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isNullable (" +
				column + ")");
		}
		int value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false

		if (resultSet.getPseudoCol (column) != null) {
			value = columnNoNulls;
		}
		else {
			value = getColAttribute (column,
					OdbcDef.SQL_COLUMN_NULLABLE);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// isSigned
	// True if column is signed number.
	//--------------------------------------------------------------------

	public boolean isSigned (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isSigned (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false

		if (resultSet.getPseudoCol (column) != null) {
			value = false;
		}
		else {
			value = (getColAttributeBoolean (column,
				OdbcDef.SQL_COLUMN_UNSIGNED) == false);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// getColumnDisplaySize
	// normal max width as chars
	//--------------------------------------------------------------------

	public int getColumnDisplaySize (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getColumnDisplaySize (" +
				column + ")");
		}
		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		JdbcOdbcPseudoCol pc = resultSet.getPseudoCol (column);
		int value;

		// If a pseudo column exists for this column, get the
		// data from the pseudo column

		if (pc != null) {
			value = pc.getColumnDisplaySize ();
		}
		else {
			// No pseudo column, get the attribute

			value = getColAttribute (column,
				OdbcDef.SQL_COLUMN_DISPLAY_SIZE);
		}

		return value;
	}

	//--------------------------------------------------------------------
	// getColumnLabel
	// Suggested column title for use in printouts and displays
	//--------------------------------------------------------------------

	public String getColumnLabel (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getColumnLabel (" +
				column + ")");
		}
		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		JdbcOdbcPseudoCol pc = resultSet.getPseudoCol (column);
		String value;

		// If a pseudo column exists for this column, get the
		// data from the pseudo column

		if (pc != null) {
			value = pc.getColumnLabel ();
		}
		else {
			// No pseudo column, get the attribute
			value =  getColAttributeString (column,
				OdbcDef.SQL_COLUMN_LABEL);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// getColumnName
	// SQL name for column
	//--------------------------------------------------------------------

	public String getColumnName (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getColumnName (" +
				column + ")");
		}
		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		JdbcOdbcPseudoCol pc = resultSet.getPseudoCol (column);
		String value;

		// If a pseudo column exists for this column, get the
		// data from the pseudo column

		if (pc != null) {
			value = pc.getColumnLabel ();
		}
		else {
			// No pseudo column, get the attribute
			value = getColAttributeString (column,
				OdbcDef.SQL_COLUMN_NAME);
		}
		return value;
	}


	//--------------------------------------------------------------------
	// getSchemaName
	// Column's table's schema, or "" if not applicable.
	//--------------------------------------------------------------------

	public String getSchemaName (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getSchemaName (" +
				column + ")");
		}
		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return the
		// same schema name as column 1

		if (resultSet.getPseudoCol (column) != null) {
			column = 1;
		}

		return getColAttributeString (column,
			OdbcDef.SQL_COLUMN_OWNER_NAME);
	}

	//--------------------------------------------------------------------
	// getPrecision
	// Number of decimal digits
	//--------------------------------------------------------------------

	public int getPrecision (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getPrecision (" +
				column + ")");
		}
		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		JdbcOdbcPseudoCol pc = resultSet.getPseudoCol (column);
		int value;

		// If a pseudo column exists for this column, get the
		// data from the pseudo column

		if (pc != null) {
			value = pc.getColumnDisplaySize () - 1;
		}
		else {
			// No pseudo column, get the attribute
			value = getColAttribute (column,
				OdbcDef.SQL_COLUMN_PRECISION);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// getScale
	// Number of digits to right of decimal
	//--------------------------------------------------------------------

	public int getScale (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getScale (" +
				column + ")");
		}
		return resultSet.getScale (column);
	}

	//--------------------------------------------------------------------
	// getTableName
	// name of column's table, or "" if not applicable.
	//--------------------------------------------------------------------

	public String getTableName (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getTableName (" +
				column + ")");
		}
		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return the
		// same table name as column 1

		if (resultSet.getPseudoCol (column) != null) {
			column = 1;
		}

		return getColAttributeString (column,
			OdbcDef.SQL_COLUMN_TABLE_NAME);
	}


	//--------------------------------------------------------------------
	// getCatalogName
	// catalog of column's table, or "" if not applicable.
	//--------------------------------------------------------------------

	public String getCatalogName (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getCatalogName (" +
				column + ")");
		}
		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return the
		// same catalog name as column 1

		if (resultSet.getPseudoCol (column) != null) {
			column = 1;
		}

		return getColAttributeString (column,
			OdbcDef.SQL_COLUMN_QUALIFIER_NAME);
	}

	//--------------------------------------------------------------------
	// getColumnType
	// SQL type, from java.sql.Types
	//--------------------------------------------------------------------

	public int getColumnType (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getColumnType (" +
				column + ")");
		}
		JdbcOdbcPseudoCol pc = resultSet.getPseudoCol (column);
		int value;

		// If a pseudo column exists for this column, get the
		// data from the pseudo column

		if (pc != null) {
			value = pc.getColumnType () - 1;
		}
		else {

			// The result set already contains information about
			// the column.  Get the SQL type from here.

			value = resultSet.getColumnType (column);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// getColumnTypeName
	// SQL type name
	//--------------------------------------------------------------------

	public String getColumnTypeName (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.getColumnTypeName (" +
				column + ")");
		}
		String value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return ""

		if (resultSet.getPseudoCol (column) != null) {
			value = "";
		}
		else {
			value = getColAttributeString (column,
				OdbcDef.SQL_COLUMN_TYPE_NAME);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// isReadOnly
	// definitely not writable
	//--------------------------------------------------------------------

	public boolean isReadOnly (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isReadOnly (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return true

		if (resultSet.getPseudoCol (column) != null) {
			value = true;
		}
		else {
			int n = getColAttribute (column,
					OdbcDef.SQL_COLUMN_UPDATABLE);
			value = (n == OdbcDef.SQL_ATTR_READONLY);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// isWritable
	// a write *may* succeed
	//--------------------------------------------------------------------

	public boolean isWritable (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isWritable (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false 

		if (resultSet.getPseudoCol (column) != null) {
			value = false;
		}
		else {
			int n = getColAttribute (column,
					OdbcDef.SQL_COLUMN_UPDATABLE);
			value = (n == OdbcDef.SQL_ATTR_READWRITE_UNKNOWN);
		}
		return value;
	}

	//--------------------------------------------------------------------
	// isDefinitelyWritable
	// a write will succeed
	//--------------------------------------------------------------------
	
	public boolean isDefinitelyWritable (
		int column)
		throws SQLException
	{
		if (isTracing ()) {
			trace ("*ResultSetMetaData.isDefinitelyWritable (" +
				column + ")");
		}
		boolean value;

		// Re-map column if necessary

		column = resultSet.mapColumn (column);

		// If a pseudo column exists for this column, return false 

		if (resultSet.getPseudoCol (column) != null) {
			value = false;
		}
		else {
			int n = getColAttribute (column,
					OdbcDef.SQL_COLUMN_UPDATABLE);
			value = (n == OdbcDef.SQL_ATTR_WRITE);
		}
		return value;
	}

	//====================================================================
	// Protected methods
	//====================================================================

	protected int getColAttribute (
		int column,
		int type)
		throws SQLException
	{
		return resultSet.getColAttribute (column, type);
	}

	protected boolean getColAttributeBoolean (
		int column,
		int type)
		throws SQLException
	{
		int value = getColAttribute (column, type);
		boolean rc = false;
		
		// If the value is SQL_TRUE (1), return true

		if (value == OdbcDef.SQL_TRUE) {
			rc = true;
		}
		return rc;
	}

	protected String getColAttributeString (
		int column,
		int type)
		throws SQLException
	{
		String	value = "";

		// Reset last warning message

		resultSet.clearWarnings ();

		try {
			value = OdbcApi.SQLColAttributesString (hStmt,
					column, type);
		}
		catch (JdbcOdbcSQLWarning ex) {

			// If we got a warning, the return value was saved off
			// in the warning object.  Get it back and set the
			// return value.

			value = (String) ex.value;
			resultSet.setWarning (JdbcOdbc.convertWarning (ex));
		}
		return value.trim ();
	}


	//====================================================================
	// Data attributes
	//====================================================================

	protected JdbcOdbc OdbcApi;		// ODBC API interface object

	protected JdbcOdbcResultSetInterface resultSet;
						// Owning result set object
	
	protected int	hStmt;			// Statement handle

    //--------------------------JDBC 2.0-----------------------------------

    /**
     * JDBC 2.0
     *
     * <p>Return the fully qualified name of the Java class whose instances 
     * are manufactured if ResultSet.getObject() is called to retrieve a value 
     * from the column.  ResultSet.getObject() may return a subClass of the
     * class returned by this method.
     */
    public String getColumnClassName(int column) { throw new UnsupportedOperationException(); }
}
