/*
 * @(#)CRC16.java	1.10 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.misc;

/**
 * The CRC-16 class calculates a 16 bit cyclic redundancy check of a set
 * of bytes. This error detecting code is used to determine if bit rot
 * has occured in a byte stream.
 */

public class CRC16 {

    /** value contains the currently computed CRC, set it to 0 initally */
    public int value;

    public CRC16() {
	value = 0;
    }

    /** update CRC with byte b */
    public void update(byte aByte) {
	int a, b;

	a = (int) aByte;
	for (int count = 7; count >=0; count--) {
	    a = a << 1;
            b = (a >>> 8) & 1;
	    if ((value & 0x8000) != 0) {
		value = ((value << 1) + b) ^ 0x1021;
	    } else {
		value = (value << 1) + b;
	    }
	}
	value = value & 0xffff;
	return;
    }

    /** reset CRC value to 0 */
    public void reset() {
	value = 0;
    }
}
