/*
 * @(#)InetAddressCachePolicy.java	1.5 98/09/23
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net;

import java.net.InetAddress;

public final class InetAddressCachePolicy {

    private static final String cachePolicyProp = "sun.net.inetaddr.ttl";

    public static final int FOREVER = -1;
    public static final int NEVER = 0;

    /* The Java-level namelookup cache policy:
     * 
     * -1: caching forever
     * any positive value: the number of seconds to cache an address for
     *
     * default value is forever (FOREVER), as we let the platform do the
     * caching. For security reasons, this caching is made forever when
     * a security manager is set.
     */
    private static int cachePolicy;
    
    /* wether or not the cache policy was set using a property (cmd line) */
    private static boolean set = false;

    /*
     * Initialize
     */
    static {
	Integer tmp = (Integer)java.security.AccessController.doPrivileged
	    (new sun.security.action.GetIntegerAction(cachePolicyProp));
	if (tmp == null) {
	    tmp = new Integer(FOREVER);
	    set = false;
	} else {
	    set = true;
	}
	cachePolicy = tmp.intValue();
    }

    public static synchronized int get() {
	return cachePolicy;
    }

    /*
     * if the user specified a cache policy using a command-property,
     * do not override. 
     */
    public static synchronized void setIfNotSet(int cachePolicy) {
	if (!set) {
	    // if malicious code gets a hold of this method, prevent
	    // setting the cache policy to something laxer or some
	    // invalid negative value.
	    if (cachePolicy > InetAddressCachePolicy.cachePolicy || 
		(cachePolicy < FOREVER)) {
		throw new SecurityException("can't make InetAddress cache more lax");
	    }
	    // we may want to signal that the cache should be flushed
	    // at this point, though this doesn't seem stricly necessary
	    InetAddressCachePolicy.cachePolicy = cachePolicy;
	}
    }
}
	



