/*
 * @(#)MeteredStream.java	1.31 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www;

import java.net.URL;
import java.util.*;
import java.io.*;
import sun.net.ProgressData;
import sun.net.ProgressEntry;

public 
class MeteredStream extends FilterInputStream {

    // Instance variables.
    /* after we've read >= expected, we're "closed" and return -1
     * from subsequest read() 's
     */
    protected boolean closed = false;
    protected int expected;
    protected int count = 0; 
    protected ProgressEntry te;

    public MeteredStream(InputStream is, ProgressEntry te) {
	super(is);
	this.te = te;
	expected = te.need;
	ProgressData.pdata.update(te);
    }

    private final void justRead(int n) throws IOException {
	if (n == -1) {
	    close();
	    return;
	}
	count += n;
	te.update(count, expected);
	ProgressData.pdata.update(te);
	if (count >= expected) {
	    close();
	}
    }

    public synchronized int read() throws java.io.IOException {
	if (closed) {
	    return -1;
	}
	int c = in.read();
	if (c != -1) {
	    justRead(1);
	} else {
	    close();
	}
	return c;
    }

    public synchronized int read(byte b[], int off, int len) throws java.io.IOException {
	if (closed) {
	    return -1;
	}
	int n = in.read(b, off, len);
	justRead(n);
	return n;
    }
    
    public synchronized long skip(long n) throws IOException {
	// REMIND: what does skip do on EOF????
	if (closed) {
	    return 0;
	}
	// just skip min(n, num_bytes_left)
	int min = (n > expected - count) ? expected - count: (int)n;
	n = in.skip(min);
	justRead((int)n);
	return n;
    }

    public void close() throws IOException {
	if (closed) {
	    return;
	}
	ProgressData.pdata.unregister(te);
	closed = true;
	in.close();
    }

    public synchronized int available() throws IOException {
	return closed ? 0: in.available();
    }

    protected void finalize() {
	/* Do not explicitly de-register from finalizer.  THIS MEANS YOU! */
	te.what = ProgressData.DELETE;
    }	
}

