/*
 * @(#)KeepAliveStream.java	1.9 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.http;

import java.net.URL;
import java.net.HttpURLConnection;
import java.io.*;
import sun.net.www.MeteredStream;
import sun.net.ProgressData;
import sun.net.ProgressEntry;
import java.util.StringTokenizer;

/**
 * A stream that has the property of being able to be kept alive for
 * multiple downloads from the same server.
 *
 * @version 1.9, 06/29/98
 * @author Stephen R. Pietrowicz (NCSA)
 * @author Dave Brown
 */
public
class KeepAliveStream extends MeteredStream {

    // instance variables
    HttpClient hc;

    boolean hurried;

    /**
     * Constructor
     */
    public KeepAliveStream(InputStream is, ProgressEntry te, HttpClient hc) {
	super(is, te);
	this.hc = hc;
    }

    /**
     * Attempt to cache this connection
     */
    public void close() throws IOException {
	// Skip past the data that's left in the Inputstream because
	// some sort of error may have occurred.
	// REMIND:  Just doing a skip() is a very naive thing to do.
	// We could have died on the first two bytes of a huge
	// download, in which case it would probably be better to
	// just close down the KeepAliveStream, rather than spending
	// the time to skip all that data.  We need to determine
	// what the right cut-off time is.
	// NOTE: Don't close super class
	try {
	    if (expected > count) {
		skip(expected - count);
	    }
	    if (!closed && !hurried) {
		HttpClient.finished(hc);
	    }
	} finally {
	    ProgressData.pdata.unregister(te);
	    closed = true;
	    // make sure you close the parent stream
	    in.close();
	}
    }

    /* we explicitly do not support mark/reset */

    public boolean markSupported() {
	return false;
    }

    public void mark(int limit) {}

    public void reset() throws IOException {
	throw new IOException("mark/reset not supported");
    }

    public synchronized boolean hurry() {
	try {
	    /* CASE 0: we're actually already done */
	    if (closed || count >= expected) {
		return false;
	    } else if (in.available() < (expected - count)) {
		/* CASE I: can't meet the demand */
		return false;
	    } else {
		/* CASE II: fill our internal buffer
		 * Remind: possibly check memory here
		 */
		byte[] buf = new byte[expected - count];
		DataInputStream dis = new DataInputStream(in);
		dis.readFully(buf);
		in = new ByteArrayInputStream(buf);
		hurried = true;
		return true;
	    }
	} catch (IOException e) {
	    // e.printStackTrace();
	    return false;
	}
    }
}



