/*
 * @(#)ServerPlugin.java	1.7 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.httpd;

import java.net.URL;

/**
 * A class to establish a framework for code fragments that plug
 * into the Java http server.
 * @author  James Gosling
 */

public class ServerPlugin {
    /** The part of the URL path matched by the *.  When a plugin is
	defined in the server configuration file, like this:<pre>
	plugin /foo/* classname args...</pre>
	and it is subsequently invoked by a URL that looks like this:<pre>
	/foo/whatever</pre>
	<b>whatever</b> will be stored in <i>path</i>. */
    protected String path;
    /** The arguments from the plugin statement in the command line.
	args[0] will contain the class name. */
    protected Object args[];

    public ServerPlugin() {
    }
    final void init(String p, Object a[]) {
	path = p;
	args = a;
    }

    /** Override this method if you need to do some validation of the "args"
	array.  It is called when the configuration file is read.  Return
	null if everything is OK, a String error message otherwise */
    String validate() {
	return null;
    }

    /** Override this method to provide results back to a get.  It should
	print to the server's clientOutput stream.  In class
	sun.net.www.httpd.Server there are a number of helper methods to use to
	construct mime headers.  startHtml(), error() and
	generateProcessOutput() being the handiest. */
    void getRequest(Server s, URL u) {
	s.error("class ServerPlugin should not be directly used.");
    }
}

