/*
 * @(#)TimingTrace.java	1.7 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.httpd;
import java.io.*;

/**
 * A class to maintain a trace of timing information through
 * an execution path.
 * @author  James Gosling
 */

public class TimingTrace {
    private long times[];
    private String terms[];
    private long t0;
    private int ntimes;
    String title;

    public TimingTrace (String t, int size) {
	times = new int[size];
	terms = new String[size];
	title = t;
	reset();
    }

    public void reset() {
	t0 = System.currentTimeMillis();
	ntimes = 0;
    }

    public void stamp(String s) {
	int n = ntimes;
	if (n < terms.length) {
	    times[n] = System.currentTimeMillis();
	    terms[n] = s;
	    ntimes = n + 1;
	}
    }

    static private void po(PrintStream s, int n, int width) {
	if (n <= 0) {
	    while (--width >= 0)
		s.write(' ');
	} else {
	    po(s, n / 10, width - 1);
	    s.write('0' + n % 10);
	}
    }
    static synchronized void print(PrintStream s, TimingTrace t) {
	int lt = t.t0;
	s.print("\nTiming title: " + t.title + "\n");
	for (int i = 0; i < t.ntimes; i++) {
	    int T = t.times[i];
	    po(s, T - lt, 6);
	    po(s, T - t.t0, 6);
	    s.print("  ");
	    lt = T;
	    s.print(t.terms[i]);
	    s.print("\n");
	}
    }
    public void print() {
	print(System.out);
    }
    public void print(PrintStream s) {
	print(s, this);
    }
}
