/*
 * @(#)FileURLConnection.java	1.42 98/10/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/**
 * Open an file input stream given a URL.
 * @author	James Gosling
 * @author	Steven B. Byrne
 * @version 	1.42, 10/02/98
 */

package sun.net.www.protocol.file;

import java.net.URL;
import java.io.*;
import java.text.Collator;
import java.security.Permission;
import sun.net.*;
import sun.net.www.*;
import java.util.*;

import sun.security.action.GetPropertyAction;
import sun.security.action.GetIntegerAction;
import sun.security.action.GetBooleanAction;
public class FileURLConnection extends URLConnection {
    
    static String CONTENT_LENGTH = "content-length";
    static String CONTENT_TYPE = "content-type";
    static String TEXT_HTML = "text/html";

    String contentType;
    InputStream is;

    File file;
    String filename;
    boolean isDirectory = false;
    boolean exists = false;
    List files;

    long length = 0;

    FileURLConnection(URL u) {
	super(u);
    }

    /*  
     * Note: the semantics of FileURLConnection object is that the
     * results of the various URLConnection calls, such as
     * getContentType, getInputStream or getContentLength reflect
     * whatever was true when connect was called.  
     */
    public void connect() throws IOException {
	if (!connected) {
            try {
                file = new File(url.getFile());
                filename = file.toString();
                isDirectory = file.isDirectory();
                if (isDirectory) {
                    files = (List) Arrays.asList(file.list());
                } else {
                   is = new BufferedInputStream(new FileInputStream(filename));
                }
            } catch (IOException e) {
                throw e;
            }
	    connected = true;
	}
    }

    private boolean initializedHeaders = false;

    private void initializeHeaders() {
	try {
	    connect();
	    exists = file.exists();
	} catch (IOException e) {
	}
	if (!initializedHeaders || !exists) {
	    length = file.length();
	    if (!isDirectory) {
		MimeTable mt = MimeTable.getDefaultTable();
		MimeEntry entry = mt.findByFileName(filename);
		if (entry != null) {
		    contentType = entry.getType();
		    properties.add(CONTENT_TYPE, contentType);
		}
		properties.add(CONTENT_LENGTH, String.valueOf(length));
	    } else {
		properties.add(CONTENT_LENGTH, TEXT_HTML);
	    }
	    initializedHeaders = true;
	}
    }

    public String getHeaderField(String name) {
	initializeHeaders();
	return super.getHeaderField(name);
    }

    public String getHeaderField(int n) {
	initializeHeaders();
	return super.getHeaderField(n);
    }

    public int getContentLength() {
        initializeHeaders();
        return super.getContentLength();
    }

    public String getHeaderFieldKey(int n) {
	initializeHeaders();
	return super.getHeaderFieldKey(n);
    }

    public MessageHeader getProperties() {
	initializeHeaders();
	return super.getProperties();
    }

    public synchronized InputStream getInputStream()
	throws IOException {

	int iconHeight;
	int iconWidth;

	connect();

	if (is == null) {
	    iconHeight = ((Integer)java.security.AccessController.doPrivileged(
	            new GetIntegerAction("hotjava.file.iconheight",
					 32))).intValue();
	    iconWidth = ((Integer)java.security.AccessController.doPrivileged(
	            new GetIntegerAction("hotjava.file.iconwidth",
					 32))).intValue();
	    if (isDirectory) {
		MimeTable mt = MimeTable.getDefaultTable();
		MimeEntry entry = mt.findByFileName(filename);

		StringBuffer buf = new StringBuffer();

		if (files == null) {
		    throw new FileNotFoundException(filename);
		}

                Collections.sort(files, Collator.getInstance());

		buf.append("<title>");
		buf.append((String)java.security.AccessController.doPrivileged(
		        new GetPropertyAction("file.dir.title", 
					      "Directory Listing")));
		buf.append("</title>\n");
		buf.append("<base href=\"file://localhost/");
		buf.append(filename.substring((filename.charAt(0) == '/') ? 1 : 0));
		if (filename.endsWith("/")) {
		    buf.append("\">");
		} else {
		    buf.append("/\">");
		}
		buf.append("<h1>");
		buf.append(filename);
		buf.append("</h1>\n");
		buf.append("<hr>\n");

		Boolean tmp;
		tmp = (Boolean) java.security.AccessController.doPrivileged(
                        new GetBooleanAction("file.hidedotfiles"));
		boolean hideDotFiles = tmp.booleanValue();
		for (int i = 0 ; i < files.size() ; i++) {
                    String fileName = (String)files.get(i);

		    if (hideDotFiles) {
			if (fileName.indexOf('.') == 0) {
			    continue;
			}
		    }
 
		    // REMIND: file path
		    buf.append("<img align=middle src=\"");
		    if (new File(filename + "/" + fileName).isDirectory()) {
			buf.append(MimeEntry.defaultImagePath +
				   "/directory.gif\" width="+iconWidth+
				   " height="+iconHeight+">\n");
		    } else {
			// REMIND: file path
			String imageFileName = MimeEntry.defaultImagePath +
			    "/file.gif";

			entry = mt.findByFileName(fileName);
			if (entry != null) {
			    String realImageName = entry.getImageFileName();
			    if (realImageName != null) {
				imageFileName = realImageName;
			    }
			}
		    
			buf.append(imageFileName);
			buf.append("\" width="+iconWidth+" height="+iconHeight+
				   ">\n");
		    }
		    buf.append("<a href=\"");
		    buf.append(fileName);
		    buf.append("\">");
		    buf.append(fileName);
		    buf.append("</a><br>");
		}
		// Put it into a (default) locale-specific byte-stream.
		is = new ByteArrayInputStream(buf.toString().getBytes());
	    } else {
		throw new FileNotFoundException(filename);
	    }
	}
	return is;
    }

    Permission permission;

    /* since getOutputStream isn't supported, only read permission is
     * relevant 
     */
    public Permission getPermission() throws IOException {
	if (permission == null) {
	    if (File.separatorChar == '/') {
		permission = new FilePermission(url.getFile(), "read");
	    } else {
		permission = new FilePermission(
			url.getFile().replace('/',File.separatorChar), "read");
	    }
	}
	return permission;
    }
}




