/*
 * @(#)Handler.java	1.37 98/04/16
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *	FTP stream opener
 */

package sun.net.www.protocol.ftp;

import java.io.IOException;
import java.net.URL;
import java.util.Map;
import java.util.HashMap;
import sun.net.ftp.FtpClient;
import sun.net.www.protocol.http.HttpURLConnection;

/** open an ftp connection given a URL */
public final class Handler extends java.net.URLStreamHandler {

    protected java.net.URLConnection openConnection(URL u) 
    throws IOException {
	/* if set for proxy usage then go through the http code to get */
	/* the url connection. Bad things will happen if a user and
	 * password are specified in the ftp url */
	if (FtpClient.getUseFtpProxy()) {
	    String host = FtpClient.getFtpProxyHost();
	    if (host != null &&
	        host.length() > 0) {
		return new HttpURLConnection(u, host,
					     FtpClient.getFtpProxyPort());
	    }
	}
	/* make a direct ftp connection */
	return new FtpURLConnection(u);
    }

    protected void parseURL(URL url, String spec, int start, int limit) {
	String host = null;
	int port = -1;
	/* the slash after the first two */
	int nextSlash = -1; 
	/* absolute specification */
	boolean specComplete = spec.startsWith("ftp://");
	/* relative absolute path */
	boolean specAbsolute = (!specComplete && spec.startsWith("/"));
	/*  Get the host, which may include user and passwd */
	if (specComplete) {
	    /* this means that the spec is fully qualified */
	    spec = spec.substring(start + 2, limit);
	    nextSlash = spec.indexOf('/', 0);
	    if (nextSlash < 0) {
		nextSlash = spec.length();
	    }
	    host = spec.substring(0, nextSlash);
	} else if (url != null) {
	    /* we get the context URL host */
	    host = url.getHost();
	    if (host == null) {
		throw new NullPointerException("no host specified");
	    }
	}
	int at = host.indexOf('@');
	if (at < 0) {
	    at = 0;
	} else {
	    /* check for syntactic correctness */
	    int colon = host.substring(0, at).indexOf(':');
	    if (colon < 0) {
		throw new NullPointerException("no user/password specified");
	    }
	}
	/* the port delimiter is the first : after the user:pws if any */
	int portDelimiter = host.indexOf(':', at);
	if (portDelimiter > 0) {
	    try {
		port = Integer.parseInt(host.substring(portDelimiter + 1));
	    } catch (NumberFormatException e) {
		/* the port remains -1 */
	    }
	    host = host.substring(0, portDelimiter);
	}
	if (specAbsolute) {
	    /* don't do anything, that's the spec and we have context */
	} else if (specComplete) {
	    spec = spec.substring(nextSlash);
	} else {
	    String ctxSpec = url.getFile();
	    if (!ctxSpec.endsWith("/")) {
		ctxSpec = ctxSpec.substring(0, ctxSpec.lastIndexOf('/') + 1);
	    }
	    spec = ctxSpec + spec;
	}
	/* note, host includes user and password, if any */
	setURL(url, "ftp", host, port, spec, null);
    }
}








