/*
 * @(#)SystemResourceURLConnection.java	1.4 98/05/06
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www.protocol.systemresource;

import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.security.Permission;
import sun.net.www.MessageHeader;
import sun.applet.AppletClassLoader;

/**
 * This is for backward compatibility only. It turns all systemresource
 * URLs into jar URLs, and delegate the requests.
 *
 * @author Benjamin Renaud
 * @version 1.4, 05/06/98
 */
public class SystemResourceURLConnection extends URLConnection {

    URL delegateUrl;
    URLConnection delegateConnection;

    SystemResourceURLConnection (URL url) 
    throws MalformedURLException, IOException  {
	super(url);
	delegateUrl = makeDelegateUrl(url);
	delegateConnection = delegateUrl.openConnection();
    }
    
    private URL makeDelegateUrl(URL url) throws MalformedURLException {
	boolean fileURL = false;
	String file = url.getFile();
	if (file.startsWith("/FILE")) {
	    fileURL = true;
	}
	int offset = fileURL ? 5 : 4;
	int index = file.lastIndexOf("/+/");
	if (index < 0) {
	    throw new MalformedURLException("no /+/ found in URL");
	}
	if (fileURL) {
	    file = "file:" + file.substring(offset, index) + 
		File.separatorChar + file.substring(index + 3, file.length());
	} else {
	    file = "jar:file:" + file.substring(offset, index) + "!/" +
		file.substring(index + 3, file.length());
	}
	return new URL(file);
    }

    public void connect() throws IOException {
	delegateConnection.connect();
    }

    public Object getContent() throws IOException {
	return delegateConnection.getContent();
    }

    public String getContentType() {
	return delegateConnection.getContentType();
    }

    public InputStream getInputStream() throws IOException {
	return delegateConnection.getInputStream();
    }

    public String getHeaderField(String name) {
	return delegateConnection.getHeaderField(name);
    }

    public Permission getPermission() throws IOException {
	return delegateConnection.getPermission();
    }
}
