/*
 * @(#)BatchEnvironment.java	1.13 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.rmic;

import java.io.*;
import java.util.*;
import sun.tools.java.*;

/**
 * BatchEnvironment for rmic extends javac's version in two ways:
 * it overrides errorString() to handle looking for rmic-specific
 * error messages in rmic's resource bundle, and it provides a
 * mechanism for recording intermediate generated files so that
 * they can be deleted later.
 */
public class BatchEnvironment extends sun.tools.javac.BatchEnvironment {

    /**
     * Create a BatchEnvironment for rmic with the given class path.
     */
    public BatchEnvironment(ClassPath path) {
	super(path);
    }

    /**
     * Create a BatchEnvironment for rmic with the given class path
     * and stream for messages.
     */
    public BatchEnvironment(OutputStream out, ClassPath path) {
	super(out, path);
    }

    /** list of generated source files created in this environment */
    private Vector generatedFiles = new Vector();

    /**
     * Remember a generated source file generated so that it
     * can be removed later, if appropriate.
     */
    public void addGeneratedFile(File file) {
	generatedFiles.addElement(file);
    }

    /**
     * Delete all the generated source files made during the execution
     * of this environment (those that have been registered with the
     * "addGeneratedFile" method).
     */
    public void deleteGeneratedFiles() {
	synchronized(generatedFiles) {
	    Enumeration enum = generatedFiles.elements();
	    while (enum.hasMoreElements()) {
		File file = (File) enum.nextElement();
		file.delete();
	    }
	    generatedFiles.removeAllElements();
	}
    }

    /**
     * Return the formatted, localized string for a named error message
     * and supplied arguments.  For rmic error messages, with names that
     * being with "rmic.", look up the error message in rmic's resource
     * bundle; otherwise, defer to java's superclass method.
     */
    public String errorString(String err,
			      Object arg0, Object arg1, Object arg2)
    {
	if (err.startsWith("rmic.")) {
	    return Main.getText(err,
		(arg0 != null ? arg0.toString() : null),
		(arg1 != null ? arg1.toString() : null),
		(arg2 != null ? arg2.toString() : null));
	} else {
	    return super.errorString(err, arg0, arg1, arg2);
	}
    }
}
