/*
 * @(#)MarshalOutputStream.java	1.19 98/06/07
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.server;

import java.io.*;
import java.rmi.Remote;
import sun.rmi.transport.Utils;

/**
 * A MarshalOutputStream extends ObjectOutputStream to add functions
 * specific to marshaling of remote object references. If it is
 * necessary to serialize remote objects or objects that contain
 * references to remote objects a MarshalOutputStream must be used
 * instead of ObjectOutputStream. <p>
 *
 * A new MarshalOutputStream is constructed to serialize remote
 * objects or graphs containing remote objects. Objects are written to
 * the stream using the ObjectOutputStream.writeObject method. <p>
 *
 * MarshalOutputStream maps remote objects to the corresponding remote
 * stub and embeds the location from which to load the stub
 * classes. The location may be ignored by the client but is supplied.
 */
public class MarshalOutputStream extends ObjectOutputStream
{
    /** Create a marshaling stream to handle RMI marshaling
     */
    public MarshalOutputStream(OutputStream out) throws IOException {
	super(out);
	this.useProtocolVersion(ObjectStreamConstants.PROTOCOL_VERSION_1);
	java.security.AccessController.doPrivileged(
				    new java.security.PrivilegedAction() {
	    public Object run() {
		privilegedEnableReplaceObject();
		return null;
	    }
	});
    }

    private void privilegedEnableReplaceObject()
    {
	enableReplaceObject(true);
    }


    /** replaceObject is extended to check for instances of Remote
     * that need to be serialized as proxy objects.  RemoteProxy.getProxy
     * is called to check for and find the stub.
     */
    protected Object replaceObject(Object obj) throws IOException {
	if (obj instanceof Remote) {
	    obj = RemoteProxy.getProxy((Remote)obj);
	}
	return obj;
    }

    /**
     * annotateClass is extended to serialize a location from which
     * to load the the specified class.
     */
    protected void annotateClass(Class cl) throws IOException {
	// write the specified location (may be null).
	writeLocation(java.rmi.server.RMIClassLoader.getClassAnnotation(cl));
    }

    /**
     * Write the location for the class into the stream.  This method can
     * be overridden by subclasses that store this annotation somewhere
     * else than as the next object in the stream, as is done by this class.
     */
    protected void writeLocation(String location) throws IOException {
	writeObject(location);
    }
}
