/*
 * @(#)UnicastServerRef2.java	1.8 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.server;

import java.io.IOException;
import java.io.ObjectOutput;
import java.rmi.*;
import java.rmi.server.*;
import sun.rmi.transport.*;
import sun.rmi.transport.tcp.*;

/**
 * Server-side ref for a remote impl that uses a custom socket factory.
 *
 * @author Ann Wollrath
 * @author Roger Riggs
 */
public class UnicastServerRef2 extends UnicastServerRef
{
    /**
     * Create a new (empty) Unicast server remote reference.
     */
    public UnicastServerRef2()
    {}

    /**
     * Construct a Unicast server remote reference for a specified
     * liveRef.
     */
    public UnicastServerRef2(LiveRef ref)
    {
	super(ref);
    }

    /**
     * Construct a Unicast server remote reference to be exported
     * on the specified port.
     */
    public UnicastServerRef2(int port,
			     RMIClientSocketFactory csf,
			     RMIServerSocketFactory ssf)
    {
	super(new LiveRef(port, csf, ssf));
    }

    /**
     * Call to dispatch to the remote object (on the server side).
     * The up-call to the server and the marshaling of return result
     * (or exception) should be handled before returning from this
     * method.
     * @param obj the target remote object for the call
     * @param call the "remote call" from which operation and
     * method arguments can be obtained.
     * @exception IOException If unable to marshal return result or
     * release input or output streams
     */
    public void dispatch(Remote obj, RemoteCall call)
	throws IOException
    {
	// TBD: The following should be better integrated with
	// super.dispatch so that an exception is cleanly returned
	// to the client.
	Endpoint client = TCPTransport.getCurrentChannel().getEndpoint();
	RMIClientSocketFactory csf = ref.getClientSocketFactory();
	RMIServerSocketFactory ssf = ref.getServerSocketFactory();

	if (client instanceof TCPEndpoint) {
	    TCPEndpoint ep = (TCPEndpoint) client;
	    RMIClientSocketFactory csfClient = ep.getClientSocketFactory();
	    RMIServerSocketFactory ssfClient = ep.getServerSocketFactory();

 	    if (!(((csf == null) ^ (csfClient == null)) ||
		  ((ssf == null) ^ (ssfClient == null)) ||
		  ((csf != null) && !(csf.equals(csfClient))) ||
		  ((ssf != null) && !(ssf.equals(ssfClient))))) {

		super.dispatch(obj, call);
		return;
	    }
	}

	throw new RemoteException("invalid socket protocol for object");
    }

    /**
     * Returns the class of the ref type to be serialized
     */
    public String getRefClass(ObjectOutput out)
    {
	if (out instanceof MarshalOutputStream) {
	    throw new InternalError("UnicastServerRef.writeExternal: " +
		"server reference written to MarshalOutputStream");
	}
	return "UnicastServerRef2";
    }

    /**
     * Return the client remote reference for this remoteRef.
     * In the case of a client RemoteRef "this" is the answer.
     * For  a server remote reference, a client side one will have to
     * found or created.
     */
    protected RemoteRef getClientRef() {
	return new UnicastRef2(ref);
    }
}
