/*
 * @(#)DGCAckHandler.java	1.7 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.rmi.transport;

import java.util.*;
import java.rmi.server.UID;

/**
 * The DGCAckHandler keeps track of and handles acknowledgments
 * from the distributed garbage collector.
 *
 * @author Ann Wollrath
 */
public class DGCAckHandler implements Notifiable {

    /** table mapping ack ID to remote object list*/
    private static Hashtable objLists = new Hashtable();

    private Channel ch;
    private UID id;
    private Vector list;

    /**
     * Construct (and register) a handler for the expected
     * acknowledgment from the distributed collector.  Remember the
     */
    public DGCAckHandler(Channel channel, UID ackID, Vector objList) 
    {
	ch = channel;
	id = ackID;
	list = objList;
	objLists.put(id, this);
	ch.addNotifiable(channel.getEndpoint(), this);
    }

    /**
     * Informs handler that an acknowledgment is received. Forget
     * about list of remote objects.
     */
    public static void received(UID id) 
    {
	DGCAckHandler entry = (DGCAckHandler)objLists.remove(id);
	if (entry != null) {
	    entry.removeNotifiable();
	}
    }

    /**
     * Private method to remove notification request
     */
    private void removeNotifiable()
    {
	ch.removeNotifiable(ch.getEndpoint(), this);
    }

    /**
     * VM sending ack is inaccessible. Forget about it.
     */
    public void notify(Endpoint ep, int state)
    {
	objLists.remove(id);
    }
}
