/*
 * @(#)ByteArrayLexOrder.java	1.6 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


package sun.security.util;

import java.util.Comparator;

/**
 * Compare two byte arrays in lexicographical order.
 *
 * @version 1.6 98/12/01
 * @author D. N. Hoover
 */
public class ByteArrayLexOrder implements Comparator {

    /**
     * Perform lexicographical comparison of two byte arrays,
     * regarding each byte as unsigned.  That is, compare array entries 
     * in order until they differ--the array with the smaller entry 
     * is "smaller". If array entries are 
     * equal till one array ends, then the longer array is "bigger".
     *
     * @param  obj1 first byte array to compare.
     * @param  obj2 second byte array to compare.
     * @return negative number if obj1 < obj2, 0 if obj1 == obj2, 
     * positive number if obj1 > obj2.  
     *
     * @exception <code>ClassCastException</code> 
     * if either argument is not a byte array.
     */
    public final int compare(Object obj1, Object obj2) {

	byte[] bytes1 = (byte[]) obj1;
	byte[] bytes2 = (byte[]) obj2;

	int diff;
	for (int i = 0; i < bytes1.length && i < bytes2.length; i++) {
	    diff = (bytes1[i] & 0xFF) - (bytes2[i] & 0xFF);
	    if (diff != 0) {
		return diff;
	    }
	}
	// if array entries are equal till the first ends, then the
	// longer is "bigger"
	return bytes1.length - bytes2.length;
    }
	

}
