/*
 * @(#)Debug.java	1.7 98/06/06
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.util;


/**
 * A utility class for debuging.
 *
 * @version 1.7
 * @author Roland Schemers
 */
public class Debug {

    private String prefix;

    private static String args;

    static {
	args = (String) java.security.AccessController.doPrivileged(
               new sun.security.action.GetPropertyAction("java.security.debug"));
	if (args != null) {
	    args = args.toLowerCase();
	    if (args.equals("help")) {
		Help();
	    }
	}
    }

    public static void Help() 
    {
	System.err.println();
	System.err.println("all       turn on all debugging");
	System.err.println("access    print all checkPermission results");
	System.err.println("jar       jar verification");
	System.err.println("policy    loading and granting");
	System.err.println("scl       permissions SecureClassLoader assigns");
	System.err.println();
	System.err.println("The following can be used with access:");
	System.err.println();
	System.err.println("stack     include stack trace");
	System.err.println("domain    dumps all domains in context");
	System.err.println("failure   before throwing exception, dump stack");
	System.err.println("          and domain that didn't have permission");
	System.err.println();
	System.exit(0);
    }

    /**
     * Get a Debug object corresponding to whether or not the given
     * option is set. Set the prefix to be the same as option.
     */

    public static Debug getInstance(String option)
    {
	return getInstance(option, option);
    }

    /**
     * Get a Debug object corresponding to whether or not the given
     * option is set. Set the prefix to be prefix.
     */
    public static Debug getInstance(String option, String prefix)
    {
	if (isOn(option)) {
	    Debug d = new Debug();
	    d.prefix = prefix;
	    return d;
	} else {
	    return null;
	}
    }

    /**
     * True if the system property "security.debug" contains the 
     * string "option".
     */
    public static boolean isOn(String option)
    {
	if (args == null)
	    return false;
	else {
	    if (args.indexOf("all") != -1)
		return true;
	    else 
		return (args.indexOf(option) != -1);
	}
    }

    /**
     * print a message to stderr that is prefixed with the prefix
     * created from the call to getInstance.
     */

    public void println(String message)
    {
	System.err.println(prefix + ": "+message);
    }

    /**
     * print a blank line to stderr that is prefixed with the prefix.
     */

    public void println()
    {
	System.err.println(prefix + ":");
    }

    /**
     * print a message to stderr that is prefixed with the prefix.
     */

    public static void println(String prefix, String message)
    {
	System.err.println(prefix + ": "+message);
    }
}
