/*
 * @(#)PropertyExpander.java	1.2 98/03/18
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.util;

import java.security.GeneralSecurityException;

/**
 * A utility class to expand properties embedded in a string.
 * Strings of the form ${some.property.name} are expanded to
 * be the value of the property. Also, the special ${/} property
 * is expanded to be the same as file.separator. If a property
 * is not set, a GeneralSecurityException will be thrown.
 *
 * @version 1.2
 * @author Roland Schemers
 */
public class PropertyExpander {


    public static class ExpandException extends GeneralSecurityException {
	public ExpandException(String msg) {
	    super(msg);
	}
    }

    public static String expand(String value)
	throws ExpandException
    {
	if (value == null)
	    return null;

	int p = value.indexOf("${", 0);

	// no special characters
	if (p == -1) return value;

	StringBuffer sb = new StringBuffer(value.length());
	int max = value.length();
	int i = 0;  // index of last character we copied

    scanner:
	while (p < max) {
	    if (p > i) {
		// copy in anything before the special stuff
		sb.append(value.substring(i, p));
		i = p;
	    }
	    int pe = p+2;
	    while ((pe < max) && (value.charAt(pe) != '}')) {
		pe++;
	    }
	    if (pe == max) {
		// no matching '}' found, just add in as normal text
		sb.append(value.substring(p, pe));
		break scanner;
	    }
	    String prop = value.substring(p+2, pe);
	    //System.out.println("expanding ("+prop+")");
	    if (prop.equals("/")) {
		sb.append(java.io.File.separatorChar);
	    } else {
		String val = System.getProperty(prop);
		if (val != null) {
		    sb.append(val);
		} else {
		    throw new ExpandException(
					     "unable to expand property " + 
					     prop);
		}
	    }
	    i = pe+1;
	    p = value.indexOf("${", i);
	    if (p == -1) {
		// no more to expand. copy in any extra
		if (i < max) {
		    sb.append(value.substring(i, max));
		} 
		// break out of loop
		break scanner;
	    }
	}
	return sb.toString();
    }

    public static void main(String args[]) throws Exception
    {
	System.out.println(expand(args[0]));
    }
}
