/*
 * @(#)GeneralSubtrees.java	1.6 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.security.x509;

import java.io.*;
import java.util.Vector;

import sun.security.util.*;

/**
 * Represent the GeneralSubtrees ASN.1 object.
 *
 * @version 1.6
 *
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 */
public class GeneralSubtrees {
    private Vector	trees;

    /**
     * The default constructor for the class.
     *
     * @params trees the sequence of GeneralSubtree.
     */
    public GeneralSubtrees(Vector trees) {
        this.trees = trees;
    }

    /**
     * Create the object from the passed DER encoded form.
     *
     * @param val the DER encoded form of the same.
     */
    public GeneralSubtrees(DerValue val) throws IOException {
        trees = new Vector(1, 1);
        if (val.tag != DerValue.tag_Sequence) {
            throw new IOException("Invalid encoding of GeneralSubtrees.");
        }
        while (val.data.available() != 0) {
            DerValue opt = val.data.getDerValue();
            GeneralSubtree tree = new GeneralSubtree(opt);
            trees.addElement(tree);
        }
    }

    /**
     * Return a printable string of the GeneralSubtree.
     */
    public String toString() {
        String s = "   GeneralSubtrees:\n" + trees.toString()
                   + "\n";

        return (s);
    }

    /**
     * Encode the GeneralSubtrees.
     *
     * @params out the DerOutputStrean to encode this object to.
     */
    public void encode(DerOutputStream out) throws IOException {
        DerOutputStream seq = new DerOutputStream();

        for (int i = 0; i < trees.size(); i++) {
            ((GeneralSubtree)trees.elementAt(i)).encode(seq);
        }
        out.write(DerValue.tag_Sequence,seq);
    }
}
