/*
 * @(#)Agent.java	1.113 98/08/14
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

import java.io.*;
import java.net.*;

import java.lang.reflect.*;

import sun.tools.java.Package;
import sun.tools.java.ClassPath;
import sun.tools.java.ClassFile;
import sun.tools.java.Identifier;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;


class ThreadList {

    final Thread threads[];
    final int count;

    ThreadList() {
	threads = new Thread[Thread.activeCount()+10];
	count = Thread.enumerate(threads);
    }
}

class Agent implements Runnable, AgentConstants {

    /* The only Agent instance.*/
    static Agent the_Agent = null;

    /* The agent socket */
    ServerSocket socket;
    boolean useSockets;

    /* The output stream back to the remote debugger */
    DataOutputStream asyncOutputStream;

    /* Pipes for socketless sessions. */
    PipedInputStream cmdInputPipe = null;
    PipedOutputStream cmdOutputPipe = null;
    PipedOutputStream asyncOutputPipe = null;

    /* The cached objects */
    ObjectCache objectCache;

    /* The breakpoint handler process */
    private Handler handler;

    StepHandler stepHandler;   // GVH SAS # 4046670

    /* The class from which the initial main() is run */
    Class mainCallerClass;

    private AgentOutputStream agentOut;

    /* message log */
    static private PrintWriter messageWriter = null;

    /* Indicator as to whether we have begun, and a lock to */
    /* protect and mimic this indicator */
    static boolean runBegun = false;
    static Object runBegunLock = new Object();

    private ThreadGroup mainGroup = null;

    private DataOutputStream out = null;
    private ResponseStream outBuffer = null;
    private ClassPath sourcePath = null;
    private Object pipeLock = new Object();
    static boolean verbose = false;

    private static ClassLoader appClassLoader =
					ClassLoader.getSystemClassLoader();
    private ThreadList lastSuspended;

    public static ClassLoader getAppLoader() {
	return appClassLoader;
    }

		// ARB SAS 4046722
    volatile boolean debuggerTerminating = false;     

    static String toHex(int n) {
	char str[] = new char[10];
	for (int i = 0 ; i < 10 ; i++) {
	    str[i] = '0';
	}
	str[1] = 'x';

	for (int i = 9 ; n != 0 ; n >>>= 4, i--) {
	    int d = n & 0xF;
	    str[i] = (char)((d < 10) ? ('0' + d) : ('a' + d - 10));
	}
	return new String(str);
    }

    /* Send all currently known classes to the RemoteAgent. */
    private void dumpClasses() throws IOException {
	message("dumpClasses()");

	/* Create a vector of classes. */
	Enumeration enum = objectCache.elements();
	Vector v = new Vector();
	int nClasses = 0;
	while (enum.hasMoreElements()) {
	    try {
		Object o = enum.nextElement();
		if (o != null && o instanceof CachedClass) {
		    v.addElement(o);
		    nClasses++;
		}
	    } catch (Exception e) {
		exceptionError(e, "dumpClasses() failed: " + e.toString());
	    }
	}

	/* Pass the vector's contents to the RemoteAgent. */
	out.writeInt(nClasses);
	enum = v.elements();
	while (enum.hasMoreElements()) {
	    writeObject(enum.nextElement());
	}
    }

    private String cookieToString(int cookie) {
        int radix = PSWDCHARS.length();
	int digitCount = 0;
	boolean minval = false;

	StringBuffer buf = new StringBuffer(32);
	while (cookie > 0) {
	    if (cookie < radix) {
		buf.append(PSWDCHARS.charAt(cookie));
		cookie = 0;
	    } else {
		int j = cookie % radix;
		cookie /= radix;
		buf.append(PSWDCHARS.charAt(j));
	    }
	    digitCount++;
	}
	if (digitCount > 0) {
	    int j = buf.length();
	    char tmp[] = new char[j];
	    int i = 0;
	    while (j-- > 0) {
		tmp[j] = buf.charAt(i++);
	    }
	    return String.valueOf(tmp);
	} else
	    return "0";
    }

    private String makePassword(int port) {
	if (port > 0xffff) {
	    System.err.println("Invalid port number (" + port + ")???");
	    System.exit(1);
	}
	int mask = (int)(Math.round(Math.random() * 0x800) % 0x800);
        int cookie = 0;
        int slice = 3;
        for (int i = 0; i < 8; i++) {
            int portbits = ((port & slice) << (i + 1));
            int maskbits = ((mask & (1 << i)) << (i * 2));
            cookie |= portbits | maskbits;
            slice <<= 2;
        }
        cookie |= (mask & 0x700) << 16;
        return cookieToString(cookie);
    }

    /* Create an Agent. */
    Agent() {
        useSockets = false;
        the_Agent = this;
    }

    Agent(int port) {
        useSockets = true;
        the_Agent = this;
	for (int i = 0; ; i++) { 
	    try {
		socket = new ServerSocket(port, 10);
		System.out.println("Agent password=" +
			
	   makePassword(socket.getLocalPort()));
		return;
	    } catch(Exception e) {
		message("[Waiting to create port]\n");
		try {
		    Thread.sleep(5000);
		} catch (InterruptedException ex) {
		    // re-interrupt and drop through.
		    Thread.currentThread().interrupt();
	 	}
	    }
	    if (i >= 10) {
		error("**Failed to create port\n");
		return;
	    }
	}
    }

    /**
     * Notify anyone interested that we have begun running
     */
    static void beginRun() {
        synchronized (runBegunLock) {
	    runBegun = true;
	    runBegunLock.notifyAll();
	}
    }

    /* Create a socketless connection t-- called by the RemoteAgent thread. */
    static synchronized void connectToAgent(PipedOutputStream clientOutput, 
                                     PipedInputStream  clientInput,
                                     PipedInputStream  clientNotification) throws IOException {
        if (the_Agent == null || the_Agent.useSockets) {
            throw new IllegalAccessError();
        }

        the_Agent.cmdInputPipe = new PipedInputStream(clientOutput);
        the_Agent.cmdOutputPipe = new PipedOutputStream(clientInput);
        the_Agent.asyncOutputPipe = new PipedOutputStream(clientNotification);
        synchronized (the_Agent.pipeLock) {
            the_Agent.pipeLock.notifyAll();
        }
    }

    /* Report whether a socketless connection is available. */
    static boolean isDebuggable() {
        return (the_Agent != null && the_Agent.useSockets == false);
    }

    /**
     * Set up agent logging - if any
     */
    private void setUpMessageWriter() {
        FileReader fr = null;
        try {
            char[] cbuf = new char[256];
            fr = new FileReader(".agentLog");
            int cnt = fr.read(cbuf, 0, 255);
            String fn = (new String(cbuf, 0, cnt)).trim();
            try {
                messageWriter = new PrintWriter(new FileWriter(fn, true), true); 
                messageWriter.println("---- debug agent message log ----");
            } catch (FileNotFoundException exc) {
            } catch (IOException exc) {
                if (messageWriter != null) {
                    messageWriter.close();
                }
            }
        } catch (FileNotFoundException exc) {
        } catch (IOException exc) {
        } finally {
            if (fr != null) {
                try {
                    fr.close();
                } catch (IOException exc) {
                }
            }
        }
    }     
        
    /* Main agent event loop */
    public void run() {
	PrintStream stdout = null;
	PrintStream stderr = null;
	Socket cmdSocket = null;
	Socket asyncSocket = null;
	
	sourcePath = new ClassPath();

    setUpMessageWriter();

	try {
	    objectCache = new ObjectCache();
	    handler = new Handler(this);
	    handler.setPriority(Thread.MAX_PRIORITY - 1);

	    stepHandler = new StepHandler(this);  // GVH SAS # 4046670

	    stepHandler.start();                  // GVH SAS # 4046670
	    handler.start();

            /* mark system threads */
            initSystemThreadList();

            // initialize the event handling.  Pass an example
            // stack frame.
            initAgent(handler.the_event, new StackFrame());

            // Notify main() that debugger is finished initializing.
            Class classObject = getClass();
            synchronized (classObject) {
                classObject.notifyAll();
            }

	    while (true) {
		stdout = System.out;
                stderr = System.err;
                DataInputStream in;
                if (useSockets) {
                    cmdSocket = socket.accept();
                    message("cmd socket: " + cmdSocket.toString());
                    in = new DataInputStream(
                        new BufferedInputStream(cmdSocket.getInputStream()));
                    outBuffer = new ResponseStream(
                        cmdSocket.getOutputStream(), 8192);
                    out = new DataOutputStream(outBuffer);

                    /* Redirect agent stdout to client via an
		       AgentOutputStream. */

                    asyncSocket = socket.accept();
                    asyncOutputStream = new DataOutputStream(
                        new BufferedOutputStream(asyncSocket.
			
			 getOutputStream()));

		    agentOut = new AgentOutputStream(asyncOutputStream);
		    
		    BufferedOutputStream bufferOut = 
			new BufferedOutputStream(agentOut, 128);

		    PrintStream redirect = new PrintStream(bufferOut , true);
		    
		    System.setOut(redirect);
		    System.setErr(redirect);
                } else {
                    try {
                        synchronized (the_Agent.pipeLock) {
                            the_Agent.pipeLock.wait();
                        }
                    } catch (InterruptedException ie) {
                        continue; // try again
                    }
                    in = new DataInputStream(cmdInputPipe);
                    outBuffer = new ResponseStream(cmdOutputPipe, 8192);
                    out = new DataOutputStream(outBuffer);
                    asyncOutputStream = 
                        new DataOutputStream(asyncOutputPipe);

		    agentOut = new AgentOutputStream(asyncOutputStream);
		   
		    PrintStream redirect =
			new PrintStream(agentOut, true);

		    System.setOut(redirect);
		    System.setErr(redirect);
                }
		message("connection accepted");

		out.writeInt(AGENT_VERSION);

		try {
		    writeObject(Class.forName("java.lang.Object"));
		    writeObject(Class.forName("java.lang.Class"));
		    writeObject(Class.forName("java.lang.String"));
		} catch (ClassNotFoundException e) {
		    // If *these* classes can't be found, just bail.
		    System.exit(1);
		}
		dumpClasses();
		writeObject(Thread.currentThread().getThreadGroup());
		out.flush();

		try {
		    for (int cmd = in.read() ; cmd != -1 ; cmd = in.read()) {
			try {
			    handle(cmd, in, out);
			} catch (Throwable t) {
			    exceptionError(t, " during command processing");
			    outBuffer.reset();  // discard normal output.
			    out.writeInt(CMD_EXCEPTION);
			    out.writeUTF(t.toString());
			}
			out.flush();
		    }
		} catch (Exception exc) {
		    System.setOut(stdout);
		    System.setErr(stderr);
		    exceptionError(exc, " during command processing");
		}
		System.setOut(stdout);
		System.setErr(stderr);
		message("connection closed");
	    }
	} catch (ThreadDeath td) {
	    System.setOut(stdout);
	    System.setErr(stderr);
	    message("ThreadDeath caught - and rethrown.");
            throw td;
	} catch (IOException ex) {
	    System.setOut(stdout);
	    System.setErr(stderr);
	    message("IOException caught.");
	}
        if (asyncSocket != null) {
	    try {
		asyncSocket.close();
	    } catch (Exception ex) {
		// Ignore exceptions here
	    }
	}
	if (cmdSocket != null) {
	    try {
		cmdSocket.close();
	    } catch (Exception ex) {
		// Ignore exceptions here
	    }
	}
    }

    /*
     * Return true if this is a special system thread that shouldn\'t
     * be suspended.
     */
    public static synchronized native boolean systemThread(Thread t);

    private static void initSystemThread(String name) {
        ThreadGroup tg = Thread.currentThread().getThreadGroup();
        if (!tg.getName().equals("system")) {
            error("findThread called from wrong threadgroup (" + tg.getName() + ")");
            return;
        }

        Thread list[] = new Thread[40];
        int n = tg.enumerate(list, false);
        for (int i = 0; i < n; i++) {
            if (list[i].getName().equals(name)) {
                message("adding " + name + " to system thread list");
                addSystemThread(list[i]);
                return;
            }
        }
        message(name + " not found in system threadgroup");
    }

    private static void initSystemThreadList() {
        initSystemThread("Debugger agent");
        initSystemThread("Breakpoint handler");
        initSystemThread("Step handler");           // GVH SAS 4046670
        initSystemThread("Finalizer");
        initSystemThread("Reference Handler");
        initSystemThread("Signal dispatcher");
    }

    public static synchronized native void addSystemThread(Thread t);
    public static synchronized native void removeSystemThread(Thread t);
    public static native void suspendListOfThreads(int count, Thread[] list);
    public static native void suspendSpecificThread(Thread t);

    /* Used by the Handler. */
    /*
     * Now obtains all VM locks before suspending the threads.  This
     * prevents suspending a thread which holds one of these locks.
     * (This fix was suggested by Sheng Liang of JavaSoft.)  Note 
     * that problems can still occur if a suspended thread holds an
     * operating system lock.  This circumstance appears to be rare.
     *
     * Note that we must also obtain a number of locks used by the      
     * Agent itself to prevent deadlocks with other Agent methods,
     * which might be called by a thread when we suspend it.
     */
    void suspendAllThreads() {
	message("suspendAllThreads()");
	ThreadList list = new ThreadList();
	synchronized(handler.the_event) {
        // Synchronize below to make sure we can send events on the 
        // async output stream while the threads are suspended.
	    synchronized(asyncOutputStream) {    
		suspendListOfThreads(list.count, list.threads);
		lastSuspended = list;
	    }
	}
    }

    /*
     * Seize all the VM locks before suspending a single thread. 
     * This may result in some delay, since we have to wait until
     * all threads give released their locks, not just the target
     * thread.
     */
    void suspendSingleThread(Thread t) {
	if (!systemThread(t)) {
	    synchronized(handler.the_event) {
        // Synchronize below to make sure we can send events on the 
        // async output stream while the threads are suspended.
		synchronized(asyncOutputStream) {
		    suspendSpecificThread(t);
		}
	    }
	}
    }

    void resumeLastSuspendedThreads() {
	ThreadList list = lastSuspended;
	lastSuspended = null;
	if (list == null) {
	    message("no last suspended to resume");
	    return;
	}
	message("resumeLastSuspendedThreads()");
	for (int i = 0 ; i < list.count ; i++) {
	    resumeThread(list.threads[i]);
	}
    }

    private void resumeThread(Thread t) {
        if (!systemThread(t)) {
	    try {
	        t.resume();
	    } catch (IllegalThreadStateException e) {
	        exceptionError(e, " resume failed: ");
	    }
	}
    }

    native void initAgent(Event event, StackFrame exampleStackFrame);

    native int getThreadStatus(Thread t);

    native Class getClasses()[];

    native void setSingleStep(Thread t, boolean step);

    native Object getFieldObject(Object o, long fieldID);
    native boolean getFieldBoolean(Object o, long fieldID);
    native byte getFieldByte(Object o, long fieldID);
    native char getFieldChar(Object o, long fieldID);
    native short getFieldShort(Object o, long fieldID);
    native int getFieldInt(Object o, long fieldID);
    native long getFieldLong(Object o, long fieldID);
    native float getFieldFloat(Object o, long fieldID);
    native double getFieldDouble(Object o, long fieldID);

    native Object getStaticFieldObject(Class c, long fieldID);
    native boolean getStaticFieldBoolean(Class c, long fieldID);
    native byte getStaticFieldByte(Class c, long fieldID);
    native char getStaticFieldChar(Class c, long fieldID);
    native short getStaticFieldShort(Class c, long fieldID);
    native int getStaticFieldInt(Class c, long fieldID);
    native long getStaticFieldLong(Class c, long fieldID);
    native float getStaticFieldFloat(Class c, long fieldID);
    native double getStaticFieldDouble(Class c, long fieldID);

    native Object getStackObject(Thread t, int frameNum, int slot);
    native boolean getStackBoolean(Thread t, int frameNum, int slot);
    native int getStackInt(Thread t, int frameNum, int slot);
    native long getStackLong(Thread t, int frameNum, int slot);
    native float getStackFloat(Thread t, int frameNum, int slot);
    native double getStackDouble(Thread t, int frameNum, int slot);

    native void setFieldBoolean(Object o, long fieldID, boolean v);
    native void setFieldByte(Object o, long fieldID, byte v);
    native void setFieldChar(Object o, long fieldID, char v);
    native void setFieldShort(Object o, long fieldID, short v);
    native void setFieldInt(Object o, long fieldID, int v);
    native void setFieldLong(Object o, long fieldID, long v);
    native void setFieldFloat(Object o, long fieldID, float v);
    native void setFieldDouble(Object o, long fieldID, double v);

    native void setStaticFieldBoolean(Class c, long fieldID, boolean v);
    native void setStaticFieldByte(Class c, long fieldID, byte v);
    native void setStaticFieldChar(Class c, long fieldID, char v);
    native void setStaticFieldShort(Class c, long fieldID, short v);
    native void setStaticFieldInt(Class c, long fieldID, int v);
    native void setStaticFieldLong(Class c, long fieldID, long v);
    native void setStaticFieldFloat(Class c, long fieldID, float v);
    native void setStaticFieldDouble(Class c, long fieldID, double v);
 
    native void setStackBoolean(Thread t, int frameNum, int slot, boolean v);
    native void setStackInt(Thread t, int frameNum, int slot, int v);
    native void setStackLong(Thread t, int frameNum, int slot, long v);
    native void setStackDouble(Thread t, int frameNum, int slot, double v);

    static String exceptionStackTrace(Throwable exception) {
        StringWriter swriter = new StringWriter();
        PrintWriter pwriter = new PrintWriter(swriter);
        exception.printStackTrace(pwriter);
        return swriter.toString();
    }

    /* Temporarily suspend the thread for use in commands like 'where' which
     * only really work properly when the thread is suspended.
     * Returns the thread if the thread needs to be resumed again, or null if
     * it doesn't. */
    private Thread temporarySuspend(Thread t) {
      if (systemThread(t))
        return null;
      int status = getThreadStatus(t);
      if (status == AgentConstants.THR_STATUS_SUSPENDED ||
          status == AgentConstants.THR_STATUS_BREAK)
        return null;
      suspendSingleThread(t);
      return t;  // Successfully suspended, so return thread so it will be resumed again
    }

    private void temporaryResume(Thread t) {
      if (t == null) return;
      try {
        t.resume();
      }
      catch (IllegalThreadStateException e) {}
    }

    /* encapsulate the on-wire format so it can be changed easily */
    private void writeMethod(DataOutputStream out, CachedMethod method)
                                             throws IOException {
        out.writeInt((int)(method.methodID));
    }

    /* encapsulate the on-wire format so it can be changed easily */
    private void writeField(DataOutputStream out, CachedField field) 
                                             throws IOException {
        out.writeInt((int)(field.fieldID));
    }

    private CachedClass readClass(DataInputStream in) throws IOException {
        return (CachedClass)objectCache.getObject(in.readInt());
    }

    private Thread readThread(DataInputStream in) throws IOException {
        return (Thread)objectCache.getObject(in.readInt());
    }

    private ThreadGroup readThreadGroup(DataInputStream in) throws IOException {
        return (ThreadGroup)objectCache.getObject(in.readInt());
    }

    /* encapsulate the on-wire format so it can be changed easily */
    private CachedMethod readMethod(DataInputStream in) throws IOException {
        CachedClass cc = readClass(in);
        long methodID = (long)(in.readInt());
        return cc.getMethod(methodID);
    }

    /* encapsulate the on-wire format so it can be changed easily */
    private CachedField readField(DataInputStream in, Class clazz)
                                                  throws IOException {
        CachedClass cc = CachedClass.getCachedClass(clazz);
        long fieldID = (long)(in.readInt());
        CachedField field = null;

        // Find the field in the cached class or one of its superclasses
        while ((cc != null) && (field == null)) {
            try {
                field = cc.getField(fieldID);
            } catch (IllegalArgumentException e) {
                // field stays null; we try superclasses
            }
            cc = cc.getSuperclass();
        }

        return field;
    }

    private Object readObject(DataInputStream in) throws IOException {
        int id = in.readInt();
        Object obj = objectCache.getObject(id);
//      Agent.message("readObject #" + id + " = " + obj);
        if (obj instanceof CachedClass) {
            obj = ((CachedClass)obj).realClass;
        }
        return obj;
    }

    /* handle command */
    synchronized void handle(int cmd, DataInputStream in, DataOutputStream out) 
throws IOException {
	out.writeInt(cmd);

	switch (cmd) {

	  case CMD_LIST_CLASSES: {
	      Class list[] = getClasses();
	      out.writeInt(list.length);
	      for (int i = 0 ; i < list.length ; i++) {
		  writeObject(list[i]);
	      }
	      return;
	  }
	    
	  case CMD_GET_CLASS_INFO: {
              String sourceName;
	      CachedClass cc = readClass(in);
              try {
                  sourceName = cc.getSourceFileName();
              } catch (AbsentInformationException exc) {
                  sourceName = "";
              }
	      Class c = cc.realClass;
	      out.writeUTF(c.getName());
              out.writeUTF(sourceName);
	      out.writeInt(c.isInterface() ? 1 : 0);
	      writeObject(c.getSuperclass());
	      writeObject(c.getClassLoader());
	      Class interfaces[] = c.getInterfaces();
	      out.writeInt(interfaces.length);
	      for (int i = 0 ; i < interfaces.length ; i++) {
		  writeObject(interfaces[i]);
	      }
	      return;
	  }

 	  case CMD_GET_THREAD_NAME:
 	    out.writeUTF(readThread(in).getName());
 	    return;

	  case CMD_GET_CLASS_BY_NAME: {
	      String name = in.readUTF();
	      try {
		  writeObject(Class.forName(name, true, getAppLoader()));
	      } catch (ClassNotFoundException e) {
		  message("no such class: " + name);
		  writeObject(null);
	      } catch (NoClassDefFoundError e) {
		  message("no such class: " + name);
		  writeObject(null);
	      } catch (IllegalArgumentException e) {
                  message("illegal class name: " + name);
                  writeObject(null);
              }
	      return;
	  }

	  case CMD_GET_THREAD_STATUS: {
	    Thread t = readThread(in);
	    int stat = getThreadStatus(t);
	    message("thread " + t + " status " + stat);
	    out.writeInt(stat);
	    return;
	  }

	  case CMD_GET_STACK_TRACE: {
              Thread t = readThread(in);

              /* This synchronize is here to make the code robust.  Because it prevents
               * any other thread (in this case, Handler) from doing a
               * suspend (see suspendThread), it prevents the following scenario from
               * occurring:
               *   - CMD_GET_STACK_TRACE temporarily suspends thread t
               *   - Handler receives a breakpoint on another thread, which
               *     makes it suspend all threads
               *   - CMD_GET_STACK_TRACE doesn't know that thread t has been suspended
               *     once again, and it resumes the thread it temporarily suspended.
               *   - Result: application is stopped at breakpoint, but one thread is
               *     still running!
               *
               * The corresponding resume case is covered, because Agent is the only
               * thread that does any resuming.
               */
              synchronized (handler.the_event) {
                  // suspend thread t if it is running - the stack frame must not
                  // change while we are looking at it.
                  Thread suspended = temporarySuspend(t);
                  StackFrame[] frames = StackFrame.getStackFrames(t);
	          int nFrames = frames.length;
                  if (nFrames > 2  && frames[nFrames-2].getRealClass()
                                                      == mainCallerClass) {
                      nFrames -= 2;
                  }
	          out.writeInt(nFrames);
	          for (int i = 0; i < nFrames; i++) {
		      StackFrame sf = frames[i];
		      writeObject(sf);
		      out.writeUTF(sf.getClassName());
		      out.writeUTF(sf.getMethodName());
		      out.writeUTF(sf.getMethodSignature());
		      out.writeInt(sf.getLinenumber());
		      out.writeInt(sf.getBCI());
		      writeObject(sf.getRealClass());
                      List locals = sf.getLocalVariables();
		      out.writeInt(locals.size());
                      Iterator iter = locals.iterator();
                      while(iter.hasNext()) {
                          CachedLocalVariable variable = 
                                (CachedLocalVariable)iter.next();
		          message("lvar: slot=" +
			          variable.slot +
			          ", name=" + variable.name +
			          ", sig=" + variable.signature + 
                                  ", argument=" + variable.isArgument());
		          out.writeInt(variable.slot);
		          out.writeUTF(variable.name);
		          out.writeUTF(variable.signature);
		          out.writeBoolean(variable.isArgument());
		      }
	          }
                  // resume thread t if it was suspended above
                  temporaryResume(suspended);
              }
	      return;
	  }

	  case CMD_GET_STACK_VALUE: {
	      Thread t = readThread(in);
	      int frameNum = in.readInt();
	      int slot = in.readInt();
	      char type = in.readChar();
	      switch (type) {
		case 'Z':
		  write(getStackBoolean(t, frameNum, slot));
		  return;
		case 'B':
		  write((byte)getStackInt(t, frameNum, slot));
		  return;
		case 'C':
		  write((char)getStackInt(t, frameNum, slot));
		  return;
		case 'S':
		  write((short)getStackInt(t, frameNum, slot));
		  return;
		case 'I':
		  write(getStackInt(t, frameNum, slot));
		  return;
		case 'J':
		  write(getStackLong(t, frameNum, slot));
		  return;
		case 'F':
		  write(getStackFloat(t, frameNum, slot));
		  return;
		case 'D':
		  write(getStackDouble(t, frameNum, slot));
		  return;
                case 'L':
                case '[': {
                  Object obj;
                  try {
                      obj = getStackObject(t, frameNum, slot);
                  } catch (Throwable thr) {
                      message("CMD_GET_STACK_VALUE - getStackObject" +
                              " - returning null because of exception: " +
                              thr.getMessage());
                      if (messageWriter != null) {
                          thr.printStackTrace(messageWriter);
                      }
                      obj = null;
                  }
		  writeObject(obj);
		  return;
                }
		case 'V':
		  writeObject(null);
		  return;
		default:
		  String strError = new String("bogus signature(" +
			
		       type + ")");
		  error(strError);
		  writeObject(strError);
	      }
	      return;
	  }
	  
	  case CMD_SET_STACK_VALUE: {
	      Thread t = readThread(in);
	      int frameNum = in.readInt();
	      int slot = in.readInt();
	      int type = in.readInt();
	      switch (type) {
		case TC_BOOLEAN:
                  setStackBoolean(t, frameNum, slot, in.readBoolean());
                  break;
		case TC_INT:
                  setStackInt(t, frameNum, slot, in.readInt());
                  break;
		case TC_LONG:
                  setStackLong(t, frameNum, slot, in.readLong());
                  break;
		case TC_DOUBLE:
                  setStackDouble(t, frameNum, slot, in.readDouble());
                  break;
		default:
                  error("bogus type(" + type + ")");
                  break;
	      }
	      return;
	  }
	  
          case CMD_MARK_OBJECTS: {
              int nMarked = in.readInt();
              int[] marked = new int[nMarked];
              for (int i = 0; i < nMarked; i++) {
                  marked[i] = in.readInt();
              }

              out.flush();
              cmd = in.read();
              out.writeInt(cmd);
              switch (cmd) {
                case CMD_FREE_OBJECTS :
                    try {
                        // Blast away all but marked
                        objectCache.freeAllBut(marked);

                        // Return saved object ids.
                        Enumeration enum = objectCache.elements();
                        while (enum.hasMoreElements()) {
                            Object obj = enum.nextElement();
                            if (obj != null) {
                                out.writeInt(objectCache.getId(obj));
                            }
                        }
                        out.writeInt(NULL_OBJECT_ID);
                    } catch (Exception e) {
                        exceptionError(e, "CMD_MARK_OBJECTS failed: ");
                    }
                    break;
                default:
                    error("mark objects command failed");
                    break;
              }
              out.flush();
              return;
          }

	  case CMD_LIST_THREADGROUPS: {
	      /*
	       * There are synchronization problems in using 
	       * ThreadGroup.activeGroupCount() followed by 
	       * ThreadGroup.enumerate(), so just allocate an
	       * very large array for now.
	       */
	      ThreadGroup list[] = new ThreadGroup[1000];
	      ThreadGroup tg = readThreadGroup(in);
	      int n = 0;
	      if (tg == null) {
		  /* List all threadgroups, including "system". */
		  message("Getting all threadgroups");
		  tg = Thread.currentThread().getThreadGroup();
		  n = tg.enumerate(list);
		  out.writeInt(n + 1);
		  writeObject(tg);
	      } else {
		  message("Getting threadgroups for " + tg.getName());
		  n = tg.enumerate(list);
		  out.writeInt(n);
	      }
	      for (int i = 0; i < n; i++) {
		  writeObject(list[i]);
	      }
	      return;
	  }

	  case CMD_GET_THREADGROUP_INFO: {
	      ThreadGroup tg = readThreadGroup(in);
	      writeObject(tg.getParent());
	      out.writeUTF(tg.getName());
	      out.writeInt(tg.getMaxPriority());
	      out.writeBoolean(tg.isDaemon());
	      return;
	  }

	  case CMD_LIST_THREADS: {
	      /*
	       * There are synchronization problems in using 
	       * ThreadGroup.activeGroupCount() followed by 
	       * ThreadGroup.enumerate(), so just allocate an
	       * very large array for now.
	       */
	      Thread list[] = new Thread[2000];
	      ThreadGroup tg = readThreadGroup(in);
	      boolean recurse = in.readBoolean();
	      message("Getting threads for " + tg.getName());
	      int n = tg.enumerate(list, recurse);
	      out.writeInt(n);
	      for (int i = 0; i < n; i++) {
		  writeObject(list[i]);
	      }
	      return;
          }

            case CMD_RUN: {
                String argv[] = new String[in.read() - 1];
                String nm = in.readUTF();
                for (int i = 0 ; i < argv.length ; i++) {
                    argv[i] = in.readUTF();
                }
                MainThread mainThread;
                ThreadGroup oldMainGroup = mainGroup;
                mainGroup = null;

                try {
                    CachedClass cc = CachedClass.getCachedClass(nm);
                    mainGroup = new ThreadGroup(cc.getName() + ".main");
                    mainThread = new MainThread(this, mainGroup, 
                                                cc.realClass, argv);
                    writeObject(mainGroup);
                } catch (Throwable thrw) {
                    message(exceptionStackTrace(thrw));
                    writeObject(null);
                    return;
                }

                if (oldMainGroup != null) {
                    message("stopping the currently running " + mainGroup);
                    try {
                        oldMainGroup.stop();
                        message("finished stop");
                    } catch (Exception exc) {
                        message("stopping current thread group failed");
                        message(exceptionStackTrace(exc));
                    }
                }
                
                // resume the old stuff else we can hang
                resumeLastSuspendedThreads();

                mainCallerClass = mainThread.getClass();
                try {
                    mainThread.start();
                } catch (IllegalThreadStateException exc) {
                    exceptionError(exc, " main thread already started");
                }
                beginRun();
                     
                return;
            }

	  case CMD_SUSPEND_THREAD:
	      suspendSingleThread(readThread(in));
	      return;

	  case CMD_RESUME_THREAD:
	      resumeThread(readThread(in));
	      return;

	  case CMD_RESUME_LAST_SUSPENDED_THREADS:
	      resumeLastSuspendedThreads();
	      return;

          // This command courtesy: GVH SAS 4046670
	  case CMD_SUSPEND_ALL_THREADS:          
	      suspendAllThreads(); 
	      return; 

	  case CMD_GET_FIELDS: {
	      CachedClass cc = readClass(in);
	      List fieldList = new ArrayList();
              try {
                  // Collect the fields for this and super classes in a list
                  while (cc != null) {
                      CachedField[] classFields = cc.getFields();
                      for (int i = 0; i < classFields.length; i++) {
                          fieldList.add(classFields[i]);
                      }
                      cc = cc.getSuperclass();
                  }
              } catch(AbsentInformationException exc) {
                  out.writeInt(0); // %%% need better error recovery
                  return;
              }

              int fieldCount = fieldList.size();
	      out.writeInt(fieldCount);
	      for (int i = 0; i < fieldCount; i++) {

                  // Reverse field order to stay more compatible with earlier versions
                  CachedField field = 
                              (CachedField)fieldList.get(fieldCount - 1 - i);
		  writeField(out, field);
		  out.writeUTF(field.getName());
		  out.writeUTF(field.getTypeSignature());
                  try {
                      out.writeShort(field.getModifiers());
                      writeObject(field.getDefiningClass());
                  } catch(AbsentInformationException exc) {
                      out.writeShort(0);  // %%% need better error recovery
                      writeObject(cc);
                  }
	      }
	      return;
	  }

	  case CMD_GET_METHODS: {
	      CachedClass cc = readClass(in);
	      CachedMethod[] methods;
              try {
                  methods = cc.getMethods();
              } catch(AbsentInformationException exc) {
                  out.writeInt(0); // %%% need better error recovery
                  return;
              }
	      out.writeInt(methods.length);
	      for (int i = 0; i < methods.length ; i++) {
                  CachedMethod method = methods[i];
		  writeMethod(out, method);
		  out.writeUTF(method.getName());
		  out.writeUTF(method.getTypeSignature());
                  try {
                      out.writeShort(method.getModifiers());
                      writeObject(method.getDefiningClass());
                  } catch(AbsentInformationException exc) {
                      out.writeShort(0);  // %%% need better error recovery
                      writeObject(cc);
                  }
	      }
	      return;
	  }

	  case CMD_GET_ELEMENTS: {
	      Object obj = readObject(in);
	      int type = in.readInt();
	      int beginIndex = in.readInt();
	      int endIndex = in.readInt();
	      out.writeInt(endIndex - beginIndex + 1);
	      switch(type) {
		case TC_CHAR: {
		    char array[] = (char[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
		case TC_BYTE: {
		    byte array[] = (byte[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
	        case TC_CLASS:
		case TP_OBJECT:
        case TC_ARRAY: {
                    Object array[] = (Object[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			writeObject(array[i]);
		    }
		    return;
		}
		case TC_FLOAT: {
		    float array[] = (float[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
		case TC_DOUBLE: {
		    double array[] = (double[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
		case TC_INT: {
		    int array[] = (int[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
		case TC_LONG: {
		    long array[] = (long[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
		case TC_SHORT: {
		    short array[] = (short[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
		case TC_BOOLEAN: {
		    boolean array[] = (boolean[])obj;
		    for (int i = beginIndex; i <= endIndex ; i++) {
			write(array[i]);
		    }
		    return;
		}
		default:
		case TC_VOID: {
		    for (int i = beginIndex; i <= endIndex ; i++) {
			writeObject(null);
		    }
		    return;
		}
	      }
	  }

/**** not used
	  case CMD_GET_SLOT_SIGNATURE: {
	      CachedField field = readField(in);
	      int sig[] = field.getSignatureArray();
              if (sig == null) {
                  out.writeInt(0);
                  return;
              }
	      out.writeInt(sig.length);
	      for (int i = 0; i < sig.length; i++) {
		  out.writeInt(sig[i]);
	      }
	      return;
	  }
*******/

	  case CMD_GET_SLOT_VALUE: {
	      Object obj = readObject(in);

              /*
               * If obj is a Class, then a static field is being requested;
               * otherwise, an instance field is wanted
               */
              Class clazz;
              if (obj instanceof Class) {
                  clazz = (Class)obj;
              } else {
                  clazz = obj.getClass();
              }

	      CachedField field = readField(in, clazz);
	      String sig = field.getTypeSignature();
              if (sig == null) {
                  writeObject(new ArrayIndexOutOfBoundsException(  // %%% wrong
                      "invalid field " + field)); 
                  return;
              }
              long fieldID = field.fieldID;

              if (obj instanceof Class) {
                  writeStaticFieldWithSignature(clazz, fieldID, sig);
              } else {
                  writeFieldWithSignature(obj, fieldID, sig);
              }
              return;
	  }

	  case CMD_SET_SLOT_VALUE: {
	      Object obj = readObject(in);

              /*
               * If obj is a Class, then a static field is being requested;
               * otherwise, an instance field is wanted
               */
              Class clazz;
              if (obj instanceof Class) {
                  clazz = (Class)obj;
              } else {
                  clazz = obj.getClass();
              }

	      CachedField field = readField(in, clazz);
              long fieldID = field.fieldID;
	      int sig = in.readInt();

              if (obj instanceof Class) {
                  setStaticField(clazz, fieldID, sig, in);
              } else {
                  setField(obj, fieldID, sig, in);
              }
              return;
	  }

	  case CMD_SET_BRKPT_LINE: {
	      CachedClass cc = readClass(in); 
	      int lineno = in.readInt();
	      Location loc;
              try {
                  loc = cc.lineToLocation(lineno);
              } catch(AbsentInformationException exc) {
                  loc = null;
              }
	      String errorStr = "";
	      if (loc == null) {
		  errorStr = "No code at line " + lineno +
		      ", or class is optimized.";
	      } else {
                  CachedMethod meth = loc.method;
		  try {
		      Breakpoint.addBreakpoint(meth, loc.startBCI, Breakpoint.NORMAL);
		  } catch (Exception e) {
		      message(exceptionStackTrace(e));
		      errorStr = e.toString();
		  }
	      }
	      out.writeUTF(errorStr);
	      return;
	  }

	  case CMD_SET_BRKPT_METHOD: {
	      CachedMethod method = readMethod(in);
	      String errorStr = "";
              if ( false ){
		  errorStr = "Can't breakpoint on native or abstract method";
	      } else {
		  try {
		      Breakpoint.addBreakpoint(method, 0, Breakpoint.NORMAL);
		  } catch (Exception e) {   
		      message(exceptionStackTrace(e));
		      errorStr = e.toString();
		  }
	      }
	      out.writeUTF(errorStr);
	      return;
/**********
%%% add these checks as exceptions???
          if ( (pc == -1) || (pc == 0) ){    // GVH SAS # 4062582 (added check against 0)
		  errorStr = "Can't breakpoint on native or abstract method";
	      } else {
		  try {
		      handler.addBreakpoint(method, 0, Breakpoint.NORMAL);
		  } catch (Exception e) {     %%% maybe catch native or abstract here

****/
	  }

	  case CMD_LIST_BREAKPOINTS: {
	      Breakpoint bkpts[] = Breakpoint.listBreakpoints();
	      out.writeInt(bkpts.length);
	      for (int i = 0; i < bkpts.length; i++) {
                  Breakpoint bkpt = bkpts[i];
		  try {
                      out.writeUTF(bkpt.getRealClass().getName() + ":" + 
                                   bkpt.method.bciToLocation(bkpt.bci).line);
                  } catch (AbsentInformationException exc) {
                      out.writeUTF(bkpt.getRealClass().getName() + "." + 
                                   bkpt.method.getName());
                  }
                     
	      }
	      return;
	  }

	  case CMD_CLEAR_BRKPT: {
	      String errorStr = "";
/* %%%% not yet implemented - needs fix on Remote side
	      CachedClass cc = readClass(in);
	      int pc = in.readInt();
	      message("clearing bkpt at " + cc.getName() + ":" + pc);
	      if (pc == -1) {
		  errorStr = "No code at pc " + pc;
	      } else {
		  try {
		      handler.deleteBreakpoint(c, pc);
		  } catch (Exception e) {
		      message(exceptionStackTrace(e));
		      errorStr = e.toString();
		  }
	      }
%%%% */
	      out.writeUTF(errorStr);
	      return;
	  }

	  case CMD_CLEAR_BRKPT_LINE: {
	      String errorStr = "";
	      CachedClass cc = readClass(in);
	      int lineno = in.readInt();
	      message("clearing bkpt at " + cc.getName() + ":" + lineno);
	      Location loc;
              try {
                  loc = cc.lineToLocation(lineno);
              } catch(AbsentInformationException exc) {
                  loc = null;
              }
	      if (loc == null) {
		  errorStr = "No code at line " + lineno +
		      ", or class is optimized.";
	      } else {
                  CachedMethod meth = loc.method;
		  try {
		      Breakpoint.deleteBreakpoint(meth, loc.startBCI);
		  } catch (Exception e) {
		      message(exceptionStackTrace(e));
		      errorStr = e.toString();
		  }
	      }
	      out.writeUTF(errorStr);
	      return;
	  }

	  case CMD_CLEAR_BRKPT_METHOD: {
	      CachedMethod meth = readMethod(in);
	      message("clearing bkpt at " + meth.getName());
	      String errorStr = "";
	      if (false) {      /* %%%% */
		  errorStr = "Not a Java method";
	      } else {
		  try {
		      Breakpoint.deleteBreakpoint(meth, 0);
		  } catch (Exception e) {
		      message(exceptionStackTrace(e));
		      errorStr = e.toString();
		  }
	      }
	      out.writeUTF(errorStr);
	      return;
	  }

	  case CMD_CATCH_EXCEPTION_CLS: {
	      String errorStr = "";
	      try {
		  CachedClass cc = readClass(in);
		  handler.catchExceptionClass(cc.realClass); 
	      } catch (Exception e) {
		  exceptionError(e, " catchExceptionClass failed: ");
	      }
	      return;
	  }

	  case CMD_IGNORE_EXCEPTION_CLS: {
	      String errorStr = "";
	      try {
		  CachedClass cc = readClass(in);
		  handler.ignoreExceptionClass(cc.realClass); 
	      } catch (Exception e) {
		  exceptionError(e, "ignoreExceptionClass failed: ");
	      }
	      return;
	  }

	  case CMD_GET_CATCH_LIST: {
	      Class list[] = handler.getCatchList();
	      out.writeInt(list.length);
	      for (int i = 0; i < list.length; i++) {
		  writeObject(list[i]);
	      }
	      return;
	  }

	  case CMD_STOP_THREAD: {
	      Thread t = readThread(in);
	      t.stop();
              message(t.getName() + " stopped.");
	      return;
	  }

	  case CMD_STOP_THREAD_GROUP: {
	      ThreadGroup tg = readThreadGroup(in);
	      tg.stop();
              message(tg.getName() + " stopped.");
	      return;
	  }

	  case CMD_SET_VERBOSE:
	      verbose = in.readBoolean();
	      return;

	  case CMD_SYSTEM: {
	      switch (in.readInt()) {
		case SYS_MEMORY_FREE:
		  // REMIND: use writeLong instead
		  out.writeInt((int)Runtime.getRuntime().freeMemory());
		  break;

		case SYS_MEMORY_TOTAL:
		  // REMIND: use writeLong instead
		  out.writeInt((int)Runtime.getRuntime().totalMemory());
		  break;

		case SYS_TRACE_METHOD_CALLS:
		  Runtime.getRuntime().traceMethodCalls(in.readInt() != 0);
		  break;

		case SYS_TRACE_INSTRUCTIONS:
		  Runtime.getRuntime().traceInstructions(in.readInt() != 0);
		  break;
	      }
	      return;
	  }

	  case CMD_GET_SOURCE_FILE: {
	      String sourceName = in.readUTF();
	      String className = in.readUTF();

	      // Extract the package name from the class (if any).
	      int iDot = className.lastIndexOf('.');
	      String pkgName = (iDot >= 0) ?
		  className.substring(0, iDot) : "";
	      Package pkg = null;
	      try {
		  pkg = new Package(sourcePath, Identifier.lookup(pkgName));
	      } catch (Exception e) {
		  message("cannot create a Package for " + pkgName);
		  out.writeInt(-1);
		  return;
	      }	      
	      
	      ClassFile sourceFile = null;
	      InputStream input = null;
	      try {
                  sourceFile = pkg.getSourceFile(sourceName);
                  if (sourceFile == null) {
                      message("no source " + sourceName);
                      out.writeInt(-1);
                      return;
                  }
		  input = sourceFile.getInputStream();
	      } catch (Exception e) {
		  message("cannot find " + sourceName);
		  out.writeInt(-1);
		  return;
	      }
	      int length = (int)sourceFile.length();
	      out.writeInt(length);
	      byte buffer[] = new byte[length];
	      try {
		  int off = 0, len = buffer.length;
		  while (len > 0) {
		      int n = input.read(buffer, off, len);
		      if (n == -1) {
			  throw new IOException();
		      }
		      off += n;
		      len -= n;
		  }
		  out.write(buffer);
	      } catch (IOException e) {
		  exceptionError(e, " unable to read " + sourceName);
		  return;
	      }
	      input.close();
	      return;
	  }

	  case CMD_OBJECT_TO_STRING: {
	      Object obj = readObject(in);
	      String retStr = "";
	      try {
		  retStr = obj.toString();
	      } catch (Exception e) {}
	      out.writeUTF(retStr);
	      return;
	  }

	  case CMD_GET_SOURCEPATH: {
	      out.writeUTF(sourcePath.toString());
	      return;
	  }

	  case CMD_SET_SOURCEPATH: {
	      String sourcePathString = in.readUTF();
	      if (sourcePathString == null) {
		  sourcePathString = ".";
	      }
	      sourcePath = new ClassPath(sourcePathString);
	      return;
	  }

          // Step implementation courtesy Gordon Hirsch  GVH/SAS 4046670
	  case CMD_STEP_THREAD:
	      stepHandler.stepThread(readThread(in), in.readBoolean());
	      return;

	  case CMD_NEXT_STEP_THREAD:
	      stepHandler.stepNextThread(readThread(in));
	      return;

	  case CMD_STEP_OUT_THREAD:
	      stepHandler.stepOutThread(readThread(in));
	      return; 

          case CMD_OBJECT_FINALIZED:
              objectCache.free(in.readInt());
              return;

          case CMD_GET_LINENUMBERS: {
              CachedClass cc = readClass(in);
              int linenos[];
              try {
                  linenos = cc.getLinenumbers();
              } catch(AbsentInformationException exc) {
                  out.writeInt(0);   // %%% need better error handling
                  return;
              }
              int count = linenos.length;
              out.writeInt(count);
              for (int i = 0; i < count; i++)
                  out.writeInt(linenos[i]);
              return;
          }

          case CMD_GET_METHOD_LINENO: {
              CachedClass cc = readClass(in);
              int index = in.readInt();
              int line;
              try {
                  CachedMethod[] methods = cc.getMethods();
                  line = methods[index].getLinenumber();
              } catch(AbsentInformationException exc) {
                  line = -1;
              }
              out.writeInt(line);
              return;
          }
    
          // Quit command revamped - courtesy ARB SAS 4046722
	  case CMD_QUIT: { 
                debuggerTerminating = true;
                if (lastSuspended != null) {
			  resumeLastSuspendedThreads();
                }
                System.exit(0);
	  }
	}
	error("command not understood: " + cmd);
    }

    private void writeFieldWithSignature(Object obj, long fieldID, 
                                         String signature) throws IOException {
        switch (signature.charAt(0)) {
            case 'Z':  // boolean
                write(getFieldBoolean(obj, fieldID));
                return;
            case 'B':  // byte
                write(getFieldByte(obj, fieldID));
                return;
            case 'C':  // char
                write(getFieldChar(obj, fieldID));
                return;
            case 'S':  // short
                write(getFieldShort(obj, fieldID));
                return;
            case 'I':  // int
                write(getFieldInt(obj, fieldID));
                return;
            case 'J':  // long
                write(getFieldLong(obj, fieldID));
                return;
            case 'F':  // float
                write(getFieldFloat(obj, fieldID));
                return;
            case 'D':  // double
                write(getFieldDouble(obj, fieldID));
                return;
            case 'L':  // Object 
            case '[':  // array  %%% use next part of signature?
                Object objectRet = getFieldObject(obj, fieldID);
                writeObject(objectRet);
                return;
            case 'V':  // void
                writeObject(null);
                return;
            default:
                String strError = "bogus signature(" + signature + ")";
                error(strError);
                writeObject(strError);
        }
    }

    private void writeStaticFieldWithSignature(Class clazz, long fieldID, 
                                         String signature) throws IOException {
        switch (signature.charAt(0)) {
            case 'Z':  // boolean
                write(getStaticFieldBoolean(clazz, fieldID));
                return;
            case 'B':  // byte
                write(getStaticFieldByte(clazz, fieldID));
                return;
            case 'C':  // char
                write(getStaticFieldChar(clazz, fieldID));
                return;
            case 'S':  // short
                write(getStaticFieldShort(clazz, fieldID));
                return;
            case 'I':  // int
                write(getStaticFieldInt(clazz, fieldID));
                return;
            case 'J':  // long
                write(getStaticFieldLong(clazz, fieldID));
                return;
            case 'F':  // float
                write(getStaticFieldFloat(clazz, fieldID));
                return;
            case 'D':  // double
                write(getStaticFieldDouble(clazz, fieldID));
                return;
            case 'L':  // Object 
            case '[':  // array  %%% use next part of signature?
                Object objectRet = getStaticFieldObject(clazz, fieldID);
                writeObject(objectRet);
                return;
            case 'V':  // void
                writeObject(null);
                return;
            default:
                String strError = "bogus signature(" + signature + ")";
                error(strError);
                writeObject(strError);
        }
    }

    private void setField(Object obj, long fieldID, int sig,
                          DataInputStream in) throws IOException {
        switch (sig) {
	    case TC_BOOLEAN:
                setFieldBoolean(obj, fieldID, in.readBoolean());
                break;
	    case TC_BYTE:
                setFieldByte(obj, fieldID, in.readByte());
                break;
	    case TC_CHAR:
	        setFieldChar(obj, fieldID, in.readChar());
	        break;
	    case TC_SHORT:
	        setFieldShort(obj, fieldID, in.readShort());
	        break;
	    case TC_INT:
                setFieldInt(obj, fieldID, in.readInt());
                break;
	    case TC_LONG:
	        setFieldLong(obj, fieldID, in.readLong());
	        break;
	    case TC_FLOAT:
		setFieldFloat(obj, fieldID, in.readFloat());
		break;
	    case TC_DOUBLE:
		setFieldDouble(obj, fieldID, in.readDouble());
		break;
	    default:
                error("bogus type(" + sig + ")");
                break;
        }
    }

    private void setStaticField(Class clazz, long fieldID, int sig, 
                                DataInputStream in) throws IOException {
        switch (sig) {
	    case TC_BOOLEAN:
                setStaticFieldBoolean(clazz, fieldID, in.readBoolean());
                break;
	    case TC_BYTE:
                setStaticFieldByte(clazz, fieldID, in.readByte());
                break;
	    case TC_CHAR:
	        setStaticFieldChar(clazz, fieldID, in.readChar());
	        break;
	    case TC_SHORT:
	        setStaticFieldShort(clazz, fieldID, in.readShort());
	        break;
	    case TC_INT:
                setStaticFieldInt(clazz, fieldID, in.readInt());
                break;
	    case TC_LONG:
	        setStaticFieldLong(clazz, fieldID, in.readLong());
	        break;
	    case TC_FLOAT:
		setStaticFieldFloat(clazz, fieldID, in.readFloat());
		break;
	    case TC_DOUBLE:
		setStaticFieldDouble(clazz, fieldID, in.readDouble());
		break;
	    default:
                error("bogus type(" + sig + ")");
                break;
        }
    }


    void writeObject(Object obj, DataOutputStream out) throws IOException {
	int h = NULL_OBJECT_ID;
    boolean writingArray = false;

	if (obj == null) {
	    out.writeInt(TP_OBJECT);
	} else if (obj instanceof Class) {
	    out.writeInt(TP_CLASS);
	    CachedClass cc = CachedClass.getCachedClass((Class)obj);
	    h = objectCache.getId(cc);
	} else if (obj instanceof CachedClass) {
	    out.writeInt(TP_CLASS);
	    h = objectCache.getId(obj);
	} else if (obj instanceof String) {
	    out.writeInt(TP_STRING);
	    h = objectCache.getId(obj);
	} else {
	    if (obj instanceof Thread) {
		out.writeInt(TP_THREAD);
	    } else if (obj instanceof ThreadGroup) {
		out.writeInt(TP_THREADGROUP);
	    } else if (obj.getClass().isArray()) {
            out.writeInt(TC_ARRAY);
            writingArray = true;
        } else {
		out.writeInt(TP_OBJECT);
	    }
	    Class c = obj.getClass();
	    CachedClass cc = CachedClass.getCachedClass(c);
	    h = objectCache.getId(cc);
	    out.writeInt(h);
	    h = objectCache.getId(obj);
	}
	out.writeInt(h);
    if (writingArray) {
        out.writeInt(java.lang.reflect.Array.getLength(obj));
    }
    }

    /* Write an object */
    void writeObject(Object obj) throws IOException{
	writeObject(obj, out);
    }

    void write(boolean b) throws IOException {
	out.writeInt(TC_BOOLEAN);
	out.writeBoolean(b);
    }

    void write(byte b) throws IOException {
	out.writeInt(TC_BYTE);
	out.writeByte(b);
    }
    
    void write(char c) throws IOException {
	out.writeInt(TC_CHAR);
	out.writeChar(c);
    }
    
    void write(short s) throws IOException {
	out.writeInt(TC_SHORT);
	out.writeShort(s);
    }
    
    void write(int i) throws IOException {
	out.writeInt(TC_INT);
	out.writeInt(i);
    }
    
    void write(long l) throws IOException {
	out.writeInt(TC_LONG);
	out.writeLong(l);
    }
    
    void write(float f) throws IOException {
	out.writeInt(TC_FLOAT);
	out.writeFloat(f);
    }
    
    void write(double d) throws IOException {
	out.writeInt(TC_DOUBLE);
	out.writeDouble(d);
    }

    void reportAppExit() {
        /* The application has either called System.exit, or its
         * main has returned.  Called from agent code.
         */
	message("report application exit");

	/*
	 * reportAppExit may be called when the debugger is no longer connected
	 * to the Agent. In that case we get a socket exception that is 
	 * silently caught here (no one to notify of the quit). 
	 */
        try {
            synchronized (asyncOutputStream) {
                asyncOutputStream.write(CMD_QUIT_NOTIFY);
                asyncOutputStream.flush();
                /*
                 * In the past the equivalent of the sleep below was in agent.c
                 * right after the call to reportAppExit. Link problems with
                 * native sleep() have forced it to be moved here. Historically
                 * the call was necessary to allow the quit notification to 
                 * get through to the client under NT. It is not clear that 
                 * is necessary any longer, but we are keeping it to be safe.
                 */
                try { Thread.sleep(1000); } catch (InterruptedException e) {}
            }
        } catch (IOException exc) {
            message("unable to notify debugger");
            // Do nothing; see comment above
	} 
    }

    void reportAppException(Event event) throws IOException {
        String errorText;
        if ((event.catch_clazz == null) ||
            (event.catch_clazz == mainCallerClass)) {
            errorText = "Uncaught exception: ";
        } else {
            errorText = "Exception: ";
        }
        String excText = exceptionStackTrace(event.exception);
        errorText = errorText.concat(excText);
            
        synchronized (asyncOutputStream) {
            asyncOutputStream.write(CMD_EXCEPTION_NOTIFY);
            writeObject(event.thread, asyncOutputStream);
            asyncOutputStream.writeUTF(errorText);
            asyncOutputStream.flush();
        }
    }

    void reportBreakpoint(Event event) throws IOException {
        synchronized (asyncOutputStream) {
            asyncOutputStream.write(CMD_BRKPT_NOTIFY);
            writeObject(event.thread, asyncOutputStream);
            asyncOutputStream.flush();
        }
    }

    void reportThreadEnd(Event event) throws IOException {
        synchronized (asyncOutputStream) {
            asyncOutputStream.write(CMD_THREADDEATH_NOTIFY);
            writeObject(event.thread, asyncOutputStream);
            asyncOutputStream.flush();
        }
    }

    /* Error, status messages */
    static void error(String msg) {
        String txt = "[Internal debug-agent error: " + msg + "]";
	System.out.println(txt);
        message(txt);
    }

    /* Error, status messages */
    static void exceptionError(Throwable exc, String msg) {
        String txt = "[Internal debug-agent exception: " + 
                          exc + " - " + msg + "]";
	System.out.println(txt);
	if (messageWriter != null) {
	    messageWriter.println(txt);
            exc.printStackTrace(messageWriter);
        }
    }

    static void message(String msg) {
	if (messageWriter != null) {
	    messageWriter.println("[debug agent: " + msg + "]");
	}
    }

    /* Bootstrap the agent */
    public static void boot(int port) {
        String jitName = System.getProperty("java.compiler");
        if (!jitName.equals("NONE") && !jitName.equals("")) {
            throw new IllegalArgumentException(
                         "The JIT compiler must be disabled in debug mode; " + 
                         "use -Djava.compiler=NONE to disable it");
        }
	    Agent agent = new Agent(port);

        /*
         * Create the agent thread in the system thread group. 
         */
        ThreadGroup tg = Thread.currentThread().getThreadGroup();
        while (tg.getParent() != null) {
            tg = tg.getParent();
        }
	    Thread t = new Thread(tg, agent, "Debugger agent");
	    t.setDaemon(true);
        t.setPriority(Thread.MAX_PRIORITY);
	    t.start();
    }

    public static void debuggable_boot() {
	Agent agent = new Agent();
	Thread t = new Thread(agent);
	t.setDaemon(true);
	t.setName("Debugger agent");
        t.setPriority(Thread.MAX_PRIORITY);
	t.start();
    }

    static {
	System.loadLibrary("agent");
    }
}
