/*
 * @(#)Breakpoint.java	1.13 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

import java.util.*;
import java.io.*;

class Breakpoint {
    CachedMethod  method;
    int		  bci;
    int	  	  type;
    Thread	  thread;

    /* Breakpoint types */
    static final   int DUMMY            = 0;
    static final   int NORMAL		= 1;
    static final   int ONESHOT		= 2;
    static final   int CONDITIONAL	= 3;
    static final   int SCATTERSHOT 	= 4;

    private static Hashtable    the_bkptHash = new Hashtable();

    private Breakpoint(CachedMethod method, int bci, int type)  {
	this.method = method;
        this.bci = bci;
        this.type = type;
	thread = null;
    }

    public int hashCode() {
	return (int)(method.methodID + bci);
    }

    public boolean equals(Object obj) {
	if (obj instanceof Breakpoint) {
	    Breakpoint bs = (Breakpoint)obj;
	    return (bs.method == method && bs.bci == bci);
	} else {
	    return false;
	}
    }

    void setThread(Thread t) {
        thread = t;
    }

    Class getRealClass() {
        return method.getRealClass();
    }

    CachedMethod getMethod() {
        return method;
    }

    public String toString() {
        String typeStr = "bkpt(dummy)";
        switch (type) {
          case NORMAL : typeStr = "bkpt"; break;
          case ONESHOT : typeStr = "bkpt(oneshot)"; break;
        }
        return typeStr + ":" + method.getName() + "(" + bci + ")";
    }

    static private native void setBreakpoint(Class clazz, long methodID, int bci);
    static private native void clrBreakpoint(Class clazz, long methodID, int bci);

    static boolean exists(CachedMethod method, int bci) {
	Breakpoint bkpt = new Breakpoint(method, bci, DUMMY);
        return the_bkptHash.containsKey(bkpt);
    }
    
    static Breakpoint get(CachedMethod method, int bci) {
	Breakpoint bkpt = new Breakpoint(method, bci, DUMMY);
        return (Breakpoint)the_bkptHash.get(bkpt);
    }

    static Enumeration allBreakpoints() {
        return the_bkptHash.elements();
    }
    
    static void clearAllBreakpoints() {
        Enumeration enum = allBreakpoints();
        while (enum.hasMoreElements()) {
            Breakpoint bkpt = (Breakpoint)enum.nextElement();
            try {
                deleteBreakpoint(bkpt);
            } catch (Exception x) {
                // Ignore exceptions here
            }
        }
    }
    
    static synchronized private void internalAddBreakpoint(Breakpoint bkpt) 
					throws IOException {
	if ((Breakpoint)the_bkptHash.get(bkpt) == null) {
            Agent.message("Adding breakpoint " + bkpt);
	    try {
	        setBreakpoint(bkpt.method.getClass(), 
                              bkpt.method.methodID, bkpt.bci);
	    } catch (IllegalArgumentException ex) {
		throw new IOException("Invalid breakpoint: " 
                                         + bkpt + " - " + ex.getMessage());
	    }
	    the_bkptHash.put(bkpt, bkpt);
	} else {
            Agent.message("Breakpoint already exists " + bkpt);
	}
    }

    static Breakpoint addBreakpoint(CachedMethod method, int bci, int type) 
						throws IOException {
	Breakpoint bkpt = new Breakpoint(method, bci, type);
	internalAddBreakpoint(bkpt);
        return bkpt;
    }

    static synchronized void deleteBreakpoint(Breakpoint bkpt) 
                                                throws IOException {
	the_bkptHash.remove(bkpt);
	try {
	    clrBreakpoint(bkpt.method.getClass(), 
                              bkpt.method.methodID, bkpt.bci);
	} catch (IllegalArgumentException ex) {
		throw new IOException("Invalid breakpoint: " 
                                         + bkpt + " - " + ex.getMessage());
	}
    }

    static synchronized void deleteBreakpoint(CachedMethod method, int bci) 
                                                throws IOException {
	Breakpoint bkpt = new Breakpoint(method, bci, DUMMY);

	bkpt = (Breakpoint)the_bkptHash.get(bkpt);
	if (bkpt == null) {
	    return;
	}
	deleteBreakpoint(bkpt);
    }

    static synchronized Breakpoint[] listBreakpoints() {
	Breakpoint ret[] = new Breakpoint[the_bkptHash.size()];
	Enumeration e = the_bkptHash.elements();
	for (int i = 0; i < ret.length; i++) {
	    try {
		ret[i] = (Breakpoint)e.nextElement();
	    } catch (NoSuchElementException nse) {
		// Should never happen!
		Agent.error("Internal error" + Agent.exceptionStackTrace(nse));
		return null;
	    }
	}
	return ret;
    }
}
