/*
 * @(#)CachedLocalVariable.java	1.3 98/05/27
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

/**
 *   The Cached* classes cache data retrieved from JVMDI.  
 *   This class provides access to field information.
 *   @author Gordon Hirsch
 */
class CachedLocalVariable {
    final CachedMethod method;
    final String name;
    final String signature;
    final int startBCI;
    final int scopeLength;
    final int slot;

    CachedLocalVariable(CachedMethod method, String name, String signature,
                        int startBCI, int scopeLength, int slot) {
        this.method = method;
        this.name = name;
        this.signature = signature;
        this.startBCI = startBCI;
        this.scopeLength = scopeLength;
        this.slot = slot;
    }

    boolean isVisible(int bci) {
        return ((startBCI <= bci) && (bci < startBCI + scopeLength));
    }

    boolean isArgument() {
         try {
             return method.isArgument(this);
         } catch (AbsentInformationException exc) {
             throw new InternalError("Local variable missing from its method's table");
         }
    }
}

    

