/*
 * @(#)MainThread.java	1.7 98/05/27
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

class MainThread extends Thread {
    static ThreadDeath rethrownDeath = new ThreadDeath();
    Agent agent;
    Class clazz;
    String argv[];
    long methodID = 0;
    
    public MainThread(Agent agent, ThreadGroup group,
		      Class clazz, String argv[]) throws Throwable {
	super(group, "main");
	setPriority(NORM_PRIORITY);
	setDaemon(false);

	this.agent = agent;
	this.clazz = clazz;
	this.argv = argv;
        try {
	    methodID = CachedMethod.getStaticMethodID(clazz, "main", 
                                                 "([Ljava/lang/String;)V");
	} catch (Throwable thr) {
            agent.exceptionError(thr,
                "\"public static void main(String argv[])\" " +
                "method not found in " + clazz.getName());
            throw thr;
        }
    }

    native void runMain(Class mainClass, long methodID, String[] argv) 
                                                       throws Exception;

    public void run() {
        try {
            runMain(clazz, methodID, argv);
        } catch (ThreadDeath td) {
	    agent.message("ThreadDeath in class " + clazz.getName() 
                        + " calling main()");
            // must rethrow if ThreadDeath
            throw rethrownDeath;
        } catch (Throwable thr) {
	    agent.exceptionError(thr, " in class " + clazz.getName() 
                        + " calling main()");
        }
        agent.message(clazz.getName() + ".main() exited");
    }
}

