/*
 * @(#)StackFrame.java	1.20 98/05/27
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.agent;

import sun.tools.java.*;

import java.util.ArrayList;
import java.util.List;

/**
 * The StackFrame class represents a stack frame of a suspended thread.  
 *
 * @see RemoteDebugger
 * @see RemoteStackFrame
 * @see RemoteThread
 */
class StackFrame {
    //  these five set by constructor
    private long frameID;
    private Class clazz;
    private long methodID;
    private int bci;
    private boolean isNative;

    private CachedMethod method;
    private Location location;

    List localVariables;

    static native int frameCount(Thread thread);
    static native void fillInFrameData(Thread thread, StackFrame[] data);

    static StackFrame[] getStackFrames(Thread thread)
    {
        StackFrame[] frameData = new StackFrame[frameCount(thread)];

        for(int i=0; i<frameData.length; i++)
            frameData[i] = new StackFrame();

        fillInFrameData(thread,frameData);

        return frameData;
    }
        
   CachedMethod getMethod() {
        if (method == null) {
            method = CachedClass.getCachedClass(clazz).getMethod(methodID);
        }
        return method;
    }
        
    Location location() throws AbsentInformationException {
        if (location == null) {
            location = getMethod().bciToLocation(bci);
        }
        return location;
    }

    /* Return the line number this frame is at, 
     * return -1 on error.
     */
    int getLinenumber() {
        try {
            Location loc = location();
            if (loc == null) {
                return -1;
            } else {
                return loc.line;
            }
        } catch (AbsentInformationException exc) {
            return -1;
        }
    }

    Class getRealClass() {
        return clazz;
    }

    String getClassName() {
        return clazz.getName();
    }

    String getMethodName() {
        return getMethod().getName();
    }

    String getMethodSignature() {
        return getMethod().getTypeSignature();
    }

    int getBCI() {
        return bci;
    }

    boolean isNative() {
        return isNative;
    }

    /* Get local variable information. 
     */
    List getLocalVariables() {
        if (localVariables == null) {
            try {
                localVariables = getMethod().getVisibleLocalVariables(bci);
            } catch (AbsentInformationException exc) {
                localVariables = new ArrayList(0);
            }
        }
        return localVariables;
    }
        
    public String toString() { 
        String className = clazz.getName();
        String methodName = getMethod().getName();
	if (isNative) {
	    // Native method.
	    return className + "." + methodName + " (native method)";
	} 
        try {
            location();
        } catch(AbsentInformationException exc) {
            // just leave location null
        }
        if (location == null || location.line == -1) {
	    // No line number information available, use bci.
	    return className + "." + methodName + " (offset " + bci + ")";
	} else {
	    int iPkg = className.lastIndexOf('.');
	    String baseName = (iPkg >= 0) ?
		className.substring(iPkg + 1) : className;
	    return className + "." + methodName +
		" (" + baseName + ":" + getLinenumber() + ")";
	}
    }
}
