/*
 * @(#)ClassConstantData.java	1.15 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.asm;

import sun.tools.java.*;
import java.io.IOException;
import java.io.DataOutputStream;

/**
 * This is a class constant pool item.
 */
final
class ClassConstantData extends ConstantPoolData {
    String name;

    /**
     * Constructor
     */

    ClassConstantData(ConstantPool tab, ClassDeclaration clazz) {
	String sig = clazz.getType().getTypeSignature();
	// sig is like "Lfoo/bar;", name is like "foo/bar".
	// We assume SIG_CLASS and SIG_ENDCLASS are 1 char each.
	name = sig.substring(1, sig.length()-1);
	tab.put(name);
    }

    // REMIND: this case should eventually go away.
    ClassConstantData(ConstantPool tab, Type t) {
	name = t.getTypeSignature();
	tab.put(name);
    }

    /**
     * Write the constant to the output stream
     */
    void write(Environment env, DataOutputStream out, ConstantPool tab) throws IOException {
	out.writeByte(CONSTANT_CLASS);
	out.writeShort(tab.index(name));
    }

    /**
     * Return the order of the constant
     */
    int order() {
	return 1;
    }

    public String toString() { 
	return "ClassConstantData[" + name + "]";
    }
}
