/*
 * @(#)FieldConstantData.java	1.13 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.asm;

import sun.tools.java.*;
import java.io.IOException;
import java.io.DataOutputStream;

/**
 * This is a field constant pool data item
 */
final
class FieldConstantData extends ConstantPoolData {
    MemberDefinition field;
    NameAndTypeData nt;

    /**
     * Constructor
     */
    FieldConstantData(ConstantPool tab, MemberDefinition field) {
	this.field = field;
	nt = new NameAndTypeData(field);
	tab.put(field.getClassDeclaration());
	tab.put(nt);
    }

    /**
     * Write the constant to the output stream
     */
    void write(Environment env, DataOutputStream out, ConstantPool tab) throws IOException {
	if (field.isMethod()) {
	    if (field.getClassDefinition().isInterface()) {
		out.writeByte(CONSTANT_INTERFACEMETHOD);
	    } else {
		out.writeByte(CONSTANT_METHOD);
	    }
	} else {
	    out.writeByte(CONSTANT_FIELD);
	}
	out.writeShort(tab.index(field.getClassDeclaration()));
	out.writeShort(tab.index(nt));
    }

    /**
     * Return the order of the constant
     */
    int order() {
	return 2;
    }
}
