/*
 * @(#)NumberConstantData.java	1.12 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.asm;

import sun.tools.java.*;
import java.io.IOException;
import java.io.DataOutputStream;

/**
 * A numeric constant pool item. Can either be integer, float, long or double.
 */
final
class NumberConstantData extends ConstantPoolData {
    Number num;

    /**
     * Constructor
     */
    NumberConstantData(ConstantPool tab, Number num) {
	this.num = num;
    }

    /**
     * Write the constant to the output stream
     */
    void write(Environment env, DataOutputStream out, ConstantPool tab) throws IOException {
	if (num instanceof Integer) {
	    out.writeByte(CONSTANT_INTEGER);
	    out.writeInt(num.intValue());
	} else if (num instanceof Long) {
	    out.writeByte(CONSTANT_LONG);
	    out.writeLong(num.longValue());
	} else if (num instanceof Float) {
	    out.writeByte(CONSTANT_FLOAT);
	    out.writeFloat(num.floatValue());
	} else if (num instanceof Double) {
	    out.writeByte(CONSTANT_DOUBLE);
	    out.writeDouble(num.doubleValue());
	}
    }
    /**
     * Return the order of the constant
     */
    int order() {
	return (width() == 1) ? 0 : 3;
    }

    /**
     * Return the number of entries that it takes up in the constant pool
     */
    int width() {
	return ((num instanceof Double) || (num instanceof Long)) ? 2 : 1;
    }
}
