/*
 * @(#)AgentIn.java	1.30 98/03/26
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

import java.io.*;
import java.net.*;
import sun.tools.agent.AgentConstants;

class AgentIn implements Runnable, AgentConstants {
    RemoteAgent agent;
    DebuggerCallback client;
    DataInputStream in;
    boolean isQuitting = false;
    
    AgentIn(RemoteAgent agent, DataInputStream in, DebuggerCallback client,
	    boolean verbose) {
	this.agent = agent;
	this.client = client;
        this.in = in;
    }

    /* 
     * Handle user initiated quits. Called from the main application
     * thread.
     */
    void quit() {
        /* 
         * Just set a flag. Any in.read() in this thread will soon 
         * fail, and this flag will allow us to exit without printing
         * an error message.
         */
        isQuitting = true;
    }

    public void run() {
        // This makes no sense unless agent is running in the same VM
        // as the remote agent.  We don't support this.
        // Agent.addSystemThread(Thread.currentThread());
	Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
	try {
	    while (true) {
                int c = in.read();
                switch (c) {
                  case -1: {
                      if (!isQuitting) {
                          agent.printToClientConsole("\nThe communications channel closed.\n");
                          agent.handleRemoteClose();
                          client.quitEvent();
                          return;
                      }
                  }
                  case CMD_WRITE_BYTES: {
                      int len = in.readInt();
                      byte buf[] = new byte[len];
                      in.readFully(buf, 0, len);
                      String text = new String(buf);
                      agent.printToClientConsole(text);
                      break;
                  }
                  case CMD_BRKPT_NOTIFY: {
                      RemoteThread t = (RemoteThread)agent.readValue(in);
                      agent.resetStacks();        // GVH SAS # 4027990
                      client.breakpointEvent(t);
                      break;
                  }
                  case CMD_EXCEPTION_NOTIFY: {
                      agent.message("AgentIn: exceptionEvent!");
                      RemoteThread t = (RemoteThread)agent.readValue(in);
                      String text = in.readUTF();
                      agent.message("text=" + text);
                      agent.resetStacks();        // GVH SAS # 4027990
                      client.exceptionEvent(t, text);
                      break;
                  }
                  case CMD_THREADDEATH_NOTIFY: {
                      RemoteThread t = (RemoteThread)agent.readValue(in);
                      client.threadDeathEvent(t);
                      break;
                  }
                  case CMD_QUIT_NOTIFY:
                      if (!isQuitting) {
                          isQuitting = true;
                          client.quitEvent();
                          agent.closeRemoteInterpreter = false;
                          agent.close();
                      }
                      return;
                  default:
                      // ignore any garbage.
		}
	    }
	} catch (Exception e) {
	    try {
                if (!isQuitting) {
                    agent.printToClientConsole("\nFatal exception: " + e + "\n");
                    e.printStackTrace();
                    agent.handleRemoteClose();
                    client.quitEvent();
                    return;
                }
	    } catch (Exception ignore) {}
	}
    }
}
