/*
 * @(#)RemoteStackVariable.java	1.11 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

import sun.tools.java.Type;
import sun.tools.agent.AgentConstants;

/**
 * A RemoteStackVariable represents a method argument or local variable.
 * It is similar to a RemoteField, but is much more transient in nature.
 *
 * @see RemoteField
 */
public class RemoteStackVariable extends LocalVariable implements AgentConstants {
    RemoteStackFrame frame;
    RemoteValue value;

    RemoteStackVariable(RemoteStackFrame frame, int slot, String name, 
                        String signature, boolean argument, RemoteValue value) {
        this.frame = frame;
	this.slot = slot;
	this.name = name;
	this.signature = signature;
        this.methodArgument = argument;
	this.value = value;
    }

    /** Return the name of a stack variable or argument. */
    public String getName() {
	return name;
    }

    /** Return the value of a stack variable or argument. */
    public RemoteValue getValue() {
	return value;
    }

    /*
     * Sets the value of an integer variable.  
     */
    void setValue(int value) throws IllegalAccessException, Exception {
        int type = getType().getTypeCode();
        if (type == TC_LONG) {
            setValue((long)value);
            return;
        }
        if (type != TC_INT && type != TC_BYTE && type != TC_SHORT ) {
            throw new IllegalAccessException();
        }
        frame.setVariable(slot, value);
    }

    /*
     * Sets the value of a boolean variable.  
     */
    void setValue(boolean value) throws IllegalAccessException, Exception {
        if (getType().getTypeCode() != TC_BOOLEAN) {
            throw new IllegalAccessException();
        }
        frame.setVariable(slot, value);
    }

    /*
     * Sets the value of a char variable.  
     */
    void setValue(char value) throws IllegalAccessException, Exception {
        if (getType().getTypeCode() != TC_CHAR) {
            throw new IllegalAccessException();
        }
        frame.setVariable(slot, value);
    }

    /*
     * Sets the value of a long variable.  
     */
    void setValue(long value) throws IllegalAccessException, Exception {
        if (getType().getTypeCode() != TC_LONG) {
            throw new IllegalAccessException();
        }
        frame.setVariable(slot, value);
    }

    /*
     * Sets the value of a float variable.  
     */
    void setValue(float value) throws IllegalAccessException, Exception {
        if (getType().getTypeCode() != TC_FLOAT) {
            throw new IllegalAccessException();
        }
        frame.setVariable(slot, value);
    }

    /*
     * Sets the value of a double variable.  
     */
    void setValue(double value) throws IllegalAccessException, Exception {
        if (getType().getTypeCode() != TC_DOUBLE) {
            throw new IllegalAccessException();
        }
        frame.setVariable(slot, value);
    }

    /** Return whether variable is in scope. */
    public boolean inScope() {
        return (slot != -1);
    }

    /** Return whether variable is a method argument. */
    public boolean methodArgument() {
        return methodArgument;
    }

    /**
     * Returns a Type object for the field.
     */
    public Type getType() {
	return Type.tType(signature);
    }
}
