/*
 * @(#)RemoteThreadGroup.java	1.12 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.debug;

/**
 * The RemoteThreadGroup class allows access to a threadgroup in a
 * remote Java interpreter.  
 *
 * @see RemoteDebugger
 * @see RemoteThreadGroup
 * @author Thomas Ball
 */
public class RemoteThreadGroup extends RemoteObject {
    RemoteThreadGroup parent;
    String name;
    int maxPriority;
    boolean daemon;

    RemoteThreadGroup(RemoteAgent agent, int id, RemoteClass clazz) {
	super(agent, TP_THREADGROUP, id, clazz);
    }

    private void getThreadGroupInfo() throws Exception {
	if (name == null) {
	    agent.getThreadGroupInfo(this);
	}
    }

    /**
     * Return the threadgroup's name.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public String getName() throws Exception {
	getThreadGroupInfo();
	return name;
    }

    /**
     * Stop the remote threadgroup.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public void stop() throws Exception {
	agent.stopThreadGroup(id);
    }

    /**
     * List a threadgroup's threads.
     *
     * @exception java.lang.Exception if a general exception occurs.
     */
    public RemoteThread[] listThreads(boolean recurse) throws Exception {
	return agent.listThreads(this, recurse);
    }
}
