/*
 * @(#)Main.java	1.11 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 	Currently javac and load() method in java.util.Properties
	supports only Latin1 encoding input.
	But in Asian platforms programmer or message translator
	uses the editor which support othere than latin1 encoding
	to specify their native language string.
	So if programmer or message translator wants to use other than
	Latin1 character in his/her program source or properties file
	they must convert the file to ASCII plus \udddd notation.
	(javac/load() modification is not appropriate due to
	 time constraints for JDK1.1)
	This utility is for the purpose of that conversion.

    NAME
	native2ascii - convert native encoding file to ascii file
		       include \udddd Unicode noataion

    SYNOPSIS
	native2ascii [options] [inputfile [outputfile]]

    DESCRIPTION
	If outputfile is not described standard output is used as
	output file, and if inputfile is not also described
	stardard input is used as input file.

	Options

	-reverse
	   convert ascii with \udddd notation to native encoding

	-encoding encoding_name
	   Specify the encoding name which is used by conversion.
	   8859_[1 - 9], JIS, EUCJIS, SJIS is currently supported.
	   Default encoding is taken from System property "file.encoding".

*/

package sun.tools.native2ascii;

import java.io.*;
import java.util.*;
import java.text.MessageFormat;
import sun.io.ByteToCharConverter;
import sun.io.CharToByteConverter;

/**
 * Main program of the native2ascii
 */

public class Main {

    String inputFileName = "input file";
    String outputFileName = "output file";
    File tempFile = null;
    boolean reverse = false;

    /**
     * Run the converter
     */
    public synchronized boolean convert(String argv[]){

	Vector v = new Vector(2);
	String encodingString = null;

	// Parse arguments
	for (int i = 0; i < argv.length; i++){
	    if (argv[i].equals("-encoding")) {
		if ((i + 1) < argv.length){
		    encodingString = argv[++i];
		} else {
		    error(getMsg("err.bad.arg"));
		    usage();
		    return false;
		}
	    } else if (argv[i].equals("-reverse")){
		reverse = true;
	    } else {
		if (v.size() > 1){
		    usage();
		    return false;
		}
		v.addElement(argv[i]);
	    }
	}

	InputStream in;
	OutputStream out;

	try {
	    in = v.size() == 0 ?
		System.in : makeInputStream((String)(v.elementAt(0)));
	    out = v.size() == 2 ?
		makeOutputStream((String)(v.elementAt(1))) : System.out;

	    DataInputStream inData = new DataInputStream(in);

	    String lineStr;
	    char[] charLineBuffer;
	    byte[] byteLineBuffer;
	    String lineBreak = System.getProperty("line.separator");

	    if (reverse){
		// ascii to native
		CharToByteConverter charToByte;

		if (encodingString != null) {
		    charToByte =
			CharToByteConverter.getConverter(encodingString);
		} else {
		    charToByte =
			CharToByteConverter.getDefault();
		}

		// The character which is not included target encoding
		// should be remain as uXXXX notation.
		//
		charToByte.setSubstitutionMode(false);

		charLineBuffer = new char[256];
		byteLineBuffer = new byte[512];

		String saveUsequence;

		while ((lineStr = inData.readLine()) != null){
		    if (lineStr.length() > charLineBuffer.length){
			charLineBuffer = new char[lineStr.length()];
			byteLineBuffer = new byte[lineStr.length() * 2];
		    }
		    int i, j, ch;
		    for (i = 0, j = 0; i < lineStr.length(); i++, j++){
			if ((ch = lineStr.charAt(i)) == '\\'){
			    if (lineStr.length() > i + 1 &&
				lineStr.charAt(i + 1) == 'u'){
				int savedI = i;
				i += 2;
				while (lineStr.length() > i &&
				       lineStr.charAt(i) == 'u'){
				    i++;
				}
				if (lineStr.length() >= i + 4){
				    ch = Integer.parseInt
					(lineStr.substring(i, i + 4), 16);

				    if (charToByte.canConvert((char)ch)) {
					i += 3;
				    } else {
					i = savedI;
					ch = lineStr.charAt(i);
				    }
				}
			    }
			}
			charLineBuffer[j] = (char)ch;

		    }

		    int length = charToByte.convert
			(charLineBuffer, 0, j,
			 byteLineBuffer, 0, byteLineBuffer.length);

		    for (i = 0; i < length; i++){
			out.write(byteLineBuffer[i]);
		    }
		    for (i = 0; i < lineBreak.length(); i++){
			out.write(lineBreak.charAt(i));
		    }

		}
	    } else {
		// native to ascii
		ByteToCharConverter byteToChar;

		if (encodingString != null) {
		    byteToChar =
			ByteToCharConverter.getConverter(encodingString);
		} else {
		    byteToChar =
			ByteToCharConverter.getDefault();
		}

		charLineBuffer = new char[256];
		byteLineBuffer = new byte[256];

		while ((lineStr = inData.readLine()) != null){
		    if (lineStr.length() > byteLineBuffer.length){
			byteLineBuffer = new byte[lineStr.length()];
			charLineBuffer = new char[lineStr.length()];
		    }
		    for (int i = 0; i < lineStr.length(); i++){
			byteLineBuffer[i] = (byte)(lineStr.charAt(i));
		    }
		    int length = byteToChar.convert
			(byteLineBuffer, 0, lineStr.length(),
			 charLineBuffer, 0, charLineBuffer.length);
		    for (int i = 0; i < length; i++){
			if (charLineBuffer[i] > 0x007f){
			    // write \udddd
			    out.write('\\');
			    out.write('u');
			    String hex =
				Integer.toHexString(charLineBuffer[i]);
			    StringBuffer hex4 = new StringBuffer(hex);
			    hex4.reverse();
			    int len = 4 - hex4.length();
			    for (int j = 0; j < len; j++){
				hex4.append('0');
			    }
			    for (int j = 0; j < 4; j++){
				out.write(hex4.charAt(3 - j));
			    }
			} else {
			    out.write(charLineBuffer[i]);
			}
		    }
		    for (int i = 0; i < lineBreak.length(); i++){
			out.write(lineBreak.charAt(i));
		    }
		}
	    }
	    out.close();
	    if (tempFile != null){
		tempFile.renameTo(new File(outputFileName));
	    }
	} catch(Exception e){
	    error(e.toString());
	    return false;
	}

	return true;
    }

    private void error(String msg){
	System.out.println(msg);
    }

    private void usage(){
	System.out.println(getMsg("usage"));
    }

    private InputStream makeInputStream(String inFile) throws Exception {
	InputStream input;

	File f = new File(inFile);
	if (!f.canRead()){
	    throw new Exception(formatMsg("err.cannot.read", f.getName()));
	}
	try {
	    input = new BufferedInputStream(new FileInputStream(f));
	} catch (IOException e){
	    throw new Exception(formatMsg("err.cannot.read", f.getName()));
	}
	inputFileName = f.getAbsolutePath();
	return input;
    }

    private OutputStream makeOutputStream(String outFile) throws Exception {
	OutputStream output;

	File f = new File(outFile);
	outputFileName = f.getAbsolutePath();

	if (inputFileName.equals(outputFileName)){
	    // find non existing temporary file name
	    for (int i = 0;; i++){
		f = new File("_N2A" + i + ".TMP");
		if (!f.exists())
		    break;
	    }
	    tempFile = f;
	} else {
	    f = new File(outFile);
	}

	try {
	    output = new BufferedOutputStream(new FileOutputStream(f));
	} catch (IOException e){
	    throw new Exception(formatMsg("err.cannot.write", f.getName()));
	}

	return output;
    }

    private static ResourceBundle rsrc;

    static {
        try {
            rsrc = ResourceBundle.getBundle(
                     "sun.tools.native2ascii.resources.MsgNative2ascii");
        } catch (MissingResourceException e) {
            throw new Error("Missing message file.");
        }
    }

    private String getMsg(String key) {
        try {
            return (rsrc.getString(key));
        } catch (MissingResourceException e) {
            throw new Error("Error in  message file format.");
        }
    }

    private String formatMsg(String key, String arg) {
        String msg = getMsg(key);
        String[] args = new String[1];
        args[0] = arg;
        return MessageFormat.format(msg, args);
    }


    /**
     * Main program
     */
    public static void main(String argv[]){
	Main converter = new Main();
	System.exit(converter.convert(argv) ? 0 : 1);
    }
}
