/*
 * @(#)SerialVer.java	1.14 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.serialver;

import java.io.*;
import java.awt.*;
import java.applet.*;
import java.io.ObjectStreamClass;
import java.util.Properties;
import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.net.URLClassLoader;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.StringTokenizer;

public class SerialVer extends Applet {
    GridBagLayout gb;
    TextField classname_t;
    Button show_b;
    TextField serialversion_t;
    Label footer_l;

    public synchronized void init() {
	gb = new GridBagLayout();
	setLayout(gb);

	GridBagConstraints c = new GridBagConstraints();
	c.fill = GridBagConstraints.BOTH;

	Label l1 = new Label(Res.getText("FullClassName"));
	l1.setAlignment(Label.RIGHT);
	gb.setConstraints(l1, c);
	add(l1);

	classname_t = new TextField(20);
	c.gridwidth = GridBagConstraints.RELATIVE;
	c.weightx = 1.0;
	gb.setConstraints(classname_t, c);
	add(classname_t);

	show_b = new Button(Res.getText("Show"));
	c.gridwidth = GridBagConstraints.REMAINDER;
	c.weightx = 0.0;	/* Don't grow the button */
	gb.setConstraints(show_b, c);
	add(show_b);

	Label l2 = new Label(Res.getText("SerialVersion"));
	l2.setAlignment(Label.RIGHT);
	c.gridwidth = 1;
	gb.setConstraints(l2, c);
	add(l2);

	serialversion_t = new TextField(50);
	serialversion_t.setEditable(false);
	c.gridwidth = GridBagConstraints.REMAINDER;
	gb.setConstraints(serialversion_t, c);
	add(serialversion_t);

	footer_l = new Label();
	c.gridwidth = GridBagConstraints.REMAINDER;
	gb.setConstraints(footer_l, c);
	add(footer_l);

	/* Give the focus to the type-in area */
	classname_t.requestFocus();
    }

    public void start() {
	/* Give the focus to the type-in area */
	classname_t.requestFocus();
    }

    public boolean action(Event ev, Object obj) {
	if (ev.target == classname_t) {
	    show((String)ev.arg);
	    return true;
	} else if (ev.target == show_b) {
	    show(classname_t.getText());
	    return true;
	}
	return false;
    }


    public boolean handleEvent(Event ev) {
	boolean rc = super.handleEvent(ev);
	return rc;
    }

    /**
     * Lookup the specified classname and display it.
     */
    void show(String classname) {
	try {
	    footer_l.setText(""); // Clear the message
	    serialversion_t.setText(""); // clear the last value

	    if (classname.equals("")) {
		return;
	    }

	    String s = serialSyntax(classname);
	    if (s != null) {
		serialversion_t.setText(s);
	    } else {
		footer_l.setText(Res.getText("NotSerializable", classname));
	    }
	} catch (ClassNotFoundException cnf) {
	    footer_l.setText(Res.getText("ClassNotFound", classname));
	}
    }

    /*
     * A class loader that will load from the CLASSPATH environment
     * variable set by the user.
     */
    static URLClassLoader loader = null;

    /*
     * Create a URL class loader that will load classes from the
     * specified classpath.
     */
    static void initializeLoader(String cp)
				throws MalformedURLException, IOException {
	URL[] urls;
	StringTokenizer st = new StringTokenizer(cp, File.pathSeparator);
	int count = st.countTokens();
	urls = new URL[count];
	for (int i = 0; i < count; i++) {
	    urls[i] =
		new File(new File(st.nextToken()).getCanonicalPath()).toURL();
	}
	loader = new URLClassLoader(urls);
    }

    /*
     * From the classname find the serialVersionUID string formatted
     * for to be copied to a java class.
     */
    static String serialSyntax(String classname) throws ClassNotFoundException {
	Class cl = Class.forName(classname, true, loader);
	ObjectStreamClass desc = ObjectStreamClass.lookup(cl);
	if (desc != null) {
	    return "    static final long serialVersionUID = " +
		desc.getSerialVersionUID() + "L;";
	} else {
	    return null;
	}
    }

    public static void main(String[] args) {
	boolean show = false;

	int i;
	for (i = 0; i < args.length; i++) {
	    if (args[i].equals("-show")) {
		show = true;
	    }  else if (args[i].startsWith("-")) {
		System.err.println(Res.getText("invalid.flag", args[i]));
		usage();
		System.exit(1);
	    } else {
		break;		// drop into processing class names
	    }
	}


	/*
	 * Get user's CLASSPATH environment variable, and make a
	 * loader that can read from that path.
	 */
	String envcp = System.getProperty("env.class.path");
	if (envcp == null) {
	    envcp = ".";
	}
	try {
	    initializeLoader(envcp);
	} catch (MalformedURLException mue) {
	    System.err.println(Res.getText("error.parsing.classpath", envcp));
	    System.exit(2);
	} catch (IOException ioe) {
	    System.err.println(Res.getText("error.parsing.classpath", envcp));
	    System.exit(3);
	}

	if (!show) {
	    if (i == args.length) {
		usage();
		System.exit(0);
	    }

	    /*
	     * The rest of the parameters are classnames.
	     */
	    for (i = i; i < args.length; i++ ) {
		try {
		    String syntax = serialSyntax(args[i]);
		    if (syntax != null)
			System.out.println(args[i] + ":" + syntax);
		    else
			System.err.println(Res.getText("NotSerializable", args[i]));
		} catch (ClassNotFoundException cnf) {
		    System.err.println(Res.getText("ClassNotFound", args[i]));
		}
	    }
	} else {
	    if (i < args.length) {
		System.err.println(Res.getText("ignoring.classes"));
		System.exit(1);
	    }
	    Frame f =  new SerialVerFrame();
	    //	f.setLayout(new FlowLayout());
	    SerialVer sv = new SerialVer();
	    sv.init();

	    f.add("Center", sv);
	    f.pack();
	    f.show();
	}
    }


    /**
     * Usage
     */
    public static void usage() {
	System.err.println(Res.getText("usage"));
    }

}

/**
 * Top level frame so serialVer can be run as an main program
 * and have an exit menu item.
 */
class SerialVerFrame extends Frame {
    MenuBar menu_mb;
    Menu file_m;
    MenuItem exit_i;

    /*
     * Construct a new Frame with title and menu.
     */
    SerialVerFrame() {
	super(Res.getText("SerialVersionInspector"));

	/* Create the file menu */
	file_m = new Menu(Res.getText("File"));
	file_m.add(exit_i = new MenuItem(Res.getText("Exit")));

	/* Now add the file menu to the menu bar  */
	menu_mb = new MenuBar();
	menu_mb.add(file_m);

	/* Add the menubar to the frame */
	// Bug in JDK1.1	setMenuBar(menu_mb);
    }

    /*
     * Handle a window destroy event by exiting.
     */
    public boolean handleEvent(Event e) {
	if (e.id == Event.WINDOW_DESTROY) {
	    exit(0);
	}
	return super.handleEvent(e);
    }
    /*
     * Handle an Exit event by exiting.
     */
    public boolean action(Event ev, Object obj) {
	if (ev.target == exit_i) {
	    exit(0);
	}
	return false;
    }

    /*
     * Cleanup and exit.
     */
    void exit(int ret) {
	System.exit(ret);
    }

}

/**
 * Utility for integrating with serialver and for localization.
 * Handle Resources. Access to error and warning counts.
 * Message formatting.
 *
 * @see java.util.ResourceBundle
 * @see java.text.MessageFormat
 */
class Res {

    private static ResourceBundle messageRB;

    /**
     * Initialize ResourceBundle
     */
    static void initResource() {
	try {
	    messageRB =
		ResourceBundle.getBundle("sun.tools.serialver.resources.serialver");
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for serialver is missing");
	}
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     */
    static String getText(String key) {
	return getText(key, (String)null);
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     * @param a1 first argument
     */
    static String getText(String key, String a1) {
	return getText(key, a1, null);
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     */
    static String getText(String key, String a1, String a2) {
	return getText(key, a1, a2, null);
    }

    /**
     * get and format message string from resource
     *
     * @param key selects message from resource
     * @param a1 first argument
     * @param a2 second argument
     * @param a3 third argument
     */
    static String getText(String key, String a1, String a2, String a3) {
	if (messageRB == null) {
	    initResource();
	}
	try {
	    String message = messageRB.getString(key);
	    String[] args = new String[3];
	    args[0] = a1;
	    args[1] = a2;
	    args[2] = a3;
	    return MessageFormat.format(message, args);
	} catch (MissingResourceException e) {
	    throw new Error("Fatal: Resource for serialver is broken. There is no " + key + " key in resource.");
	}
    }
}

