/*
 * @(#)BinaryEqualityExpression.java	1.17 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.tree;

import sun.tools.java.*;

public
class BinaryEqualityExpression extends BinaryExpression {
    /**
     * constructor
     */
    public BinaryEqualityExpression(int op, long where, Expression left, Expression right) {
	super(op, where, Type.tBoolean, left, right);
    }

    /**
     * Select the type
     */
    void selectType(Environment env, Context ctx, int tm) {
	Type t;
	if ((tm & TM_ERROR) != 0) {
	    // who cares.  One of them is an error.
	    return;
	} else if ((tm & (TM_CLASS | TM_ARRAY | TM_NULL)) != 0) {
	    try {
		if (env.explicitCast(left.type, right.type) ||
		    env.explicitCast(right.type, left.type)) {
		    return;
		}
		env.error(where, "incompatible.type", 
			  left.type, left.type, right.type);
	    } catch (ClassNotFound e) {
		env.error(where, "class.not.found", e.name, opNames[op]);
	    }
	    return;
	} else if ((tm & TM_DOUBLE) != 0) {
	    t = Type.tDouble;
	} else if ((tm & TM_FLOAT) != 0) {
	    t = Type.tFloat;
	} else if ((tm & TM_LONG) != 0) {
	    t = Type.tLong;
	} else if ((tm & TM_BOOLEAN) != 0) {
	    t = Type.tBoolean;
	} else {
	    t = Type.tInt;
	} 
	left = convert(env, ctx, t, left);
	right = convert(env, ctx, t, right);
    }
}
