/*
 * @(#)InlineMethodExpression.java	1.13 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.tree;

import sun.tools.java.*;
import sun.tools.asm.Label;
import sun.tools.asm.Assembler;
import java.io.PrintStream;

public
class InlineMethodExpression extends Expression {
    MemberDefinition field;
    Statement body;

    /**
     * Constructor
     */
    InlineMethodExpression(long where, Type type, MemberDefinition field, Statement body) {
	super(INLINEMETHOD, where, type);
	this.field = field;
	this.body = body;
    }
    /**
     * Inline
     */
    public Expression inline(Environment env, Context ctx) {
	body = body.inline(env, new Context(ctx, this));
	if (body == null) {
	    return null;
	} else if (body.op == INLINERETURN) {
	    Expression expr = ((InlineReturnStatement)body).expr;
	    if (expr != null && type.isType(TC_VOID)) {
		throw new CompilerError("value on inline-void return");
	    }
	    return expr;
	} else {
	    return this;
	}
    }
    public Expression inlineValue(Environment env, Context ctx) {
	// When this node was constructed, "copyInline" walked the body
	// with a "valNeeded" flag which made all returns either void
	// or value-bearing.  The type of this node reflects that
	// earlier choice.  The present inline/inlineValue distinction
	// is ignored.
	return inline(env, ctx);
    }

    /**
     * Create a copy of the expression for method inlining
     */
    public Expression copyInline(Context ctx) {
	InlineMethodExpression e = (InlineMethodExpression)clone();
	if (body != null) {
	    e.body = body.copyInline(ctx, true);
	}
	return e;
    }

    /**
     * Code
     */
    public void code(Environment env, Context ctx, Assembler asm) {
	// pop the result if there is any (usually, type is already void)
	super.code(env, ctx, asm);
    }
    public void codeValue(Environment env, Context ctx, Assembler asm) {
	CodeContext newctx = new CodeContext(ctx, this);
	body.code(env, newctx, asm);
	asm.add(newctx.breakLabel);
    }

    /**
     * Print
     */
    public void print(PrintStream out) {
	out.print("(" + opNames[op] + "\n");
	body.print(out, 1);
	out.print(")");
    }
}
