/*
 * @(#)SuperExpression.java	1.25 98/09/21
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.tree;

import sun.tools.java.*;
import sun.tools.asm.Assembler;
import java.io.PrintStream;
import java.util.Hashtable;

public
class SuperExpression extends ThisExpression {

    /**
     * Constructor
     */
    public SuperExpression(long where) {
	super(SUPER, where);
    }

    /**
     * Constructor for "outer.super()"
     */
    public SuperExpression(long where, Expression outerArg) {
	super(where, outerArg);
    }

    public SuperExpression(long where, Context ctx) {
	super(where, ctx);
	op = SUPER;
    }

    /**
     * Check expression
     */
    public Vset checkValue(Environment env, Context ctx, Vset vset, Hashtable exp) {
	vset = checkCommon(env, ctx, vset, exp);
	if (type != Type.tError) {
	    // "super" is not allowed in this context:
	    env.error(where, "undef.var.super", idSuper);
	}
	return vset;
    }

    /**
     * Check if the present name is part of a scoping prefix.
     */
    public Vset checkAmbigName(Environment env, Context ctx,
			       Vset vset, Hashtable exp,
			       UnaryExpression loc) {
	return checkCommon(env, ctx, vset, exp);
    }

    /** Common code for checkValue and checkAmbigName */
    private Vset checkCommon(Environment env, Context ctx, Vset vset, Hashtable exp) {
	ClassDeclaration superClass = ctx.field.getClassDefinition().getSuperClass();
	if (superClass == null) {
	    env.error(where, "undef.var", idSuper);
	    type = Type.tError;
	    return vset;
	}
	vset = super.checkValue(env, ctx, vset, exp);
	type = superClass.getType();
	return vset;
    }

}
