/*
 * @(#)WhileStatement.java	1.23 98/03/18
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.tools.tree;

import sun.tools.java.*;
import sun.tools.asm.Assembler;
import sun.tools.asm.Label;
import java.io.PrintStream;
import java.util.Hashtable;

public
class WhileStatement extends Statement {
    Expression cond;
    Statement body;

    /**
     * Constructor
     */
    public WhileStatement(long where, Expression cond, Statement body) {
	super(WHILE, where);
	this.cond = cond;
	this.body = body;
    }

    /**
     * Check a while statement
     */
    Vset check(Environment env, Context ctx, Vset vset, Hashtable exp) {
	checkLabel(env, ctx);
	CheckContext newctx = new CheckContext(ctx, this);
	// remember what was unassigned on entry
	Vset vsEntry = vset.copy();
	// check the condition.  Determine which variables have values if 
	// it returns true or false.
	ConditionVars cvars = 
	      cond.checkCondition(env, newctx, reach(env, vset), exp);
	cond = convert(env, newctx, Type.tBoolean, cond);
	// check the body, given that the condition returned true.
	vset = body.check(env, newctx, cvars.vsTrue, exp);
	vset = vset.join(newctx.vsContinue);
	// make sure the back-branch fits the entry of the loop
	ctx.checkBackBranch(env, this, vsEntry, vset);
	// Exit the while loop by testing false or getting a break statement
	vset = newctx.vsBreak.join(cvars.vsFalse);
	return ctx.removeAdditionalVars(vset);
    }

    /**
     * Inline
     */
    public Statement inline(Environment env, Context ctx) {
	ctx = new Context(ctx, this);
	cond = cond.inlineValue(env, ctx);
	if (body != null) {
	    body = body.inline(env, ctx);
	}
	return this;
    }

    /**
     * The cost of inlining this statement
     */
    public int costInline(int thresh, Environment env, Context ctx) {
	return 1 + cond.costInline(thresh, env, ctx) 
	         + ((body != null) ? body.costInline(thresh, env, ctx) : 0);
    }

    /**
     * Create a copy of the statement for method inlining
     */
    public Statement copyInline(Context ctx, boolean valNeeded) {
	WhileStatement s = (WhileStatement)clone();
	s.cond = cond.copyInline(ctx);
	if (body != null) {
	    s.body = body.copyInline(ctx, valNeeded);
	}
	return s;
    }

    /**
     * Code
     */
    public void code(Environment env, Context ctx, Assembler asm) {
	CodeContext newctx = new CodeContext(ctx, this);

	asm.add(where, opc_goto, newctx.contLabel);

	Label l1 = new Label();
	asm.add(l1);

	if (body != null) {
	    body.code(env, newctx, asm);
	}
	
	asm.add(newctx.contLabel);
	cond.codeBranch(env, newctx, asm, l1, true);
	asm.add(newctx.breakLabel);
    }

    /**
     * Print
     */
    public void print(PrintStream out, int indent) {
	super.print(out, indent);
	out.print("while ");
	cond.print(out);
	if (body != null) {
	    out.print(" ");
	    body.print(out, indent);
	} else {
	    out.print(";");
	}
    }
}
