/*
 * @(#)AllFonts.java	1.21 98/09/13
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


package demos.Fonts;


import java.awt.*;
import java.awt.event.*;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.util.Vector;
import AnimatingContext;
import DemoSurface;
import DemoPanel;


/**
 * Scrolling text of fonts returned from GraphicsEnvironment.getAllFonts().
 */
public class AllFonts extends DemoSurface implements AnimatingContext {

    private static Vector fonts = new Vector();
    static {
        GraphicsEnvironment ge = 
            GraphicsEnvironment.getLocalGraphicsEnvironment();
        Font allfonts[] = ge.getAllFonts();
        for (int i = 0; i < allfonts.length; i++) {
            if (allfonts[i].canDisplayUpTo(allfonts[i].getName()) > 0) {
                fonts.addElement(allfonts[i]);
            }
        }
    }
    private Vector v = new Vector();
    private int nStrs;
    private int strH;
    private int fi;
    private static FontRenderContext frc = 
            new FontRenderContext(null, false, false);


    public AllFonts() {
        setBackground(Color.white);
        Font f = ((Font) fonts.get(0)).deriveFont(Font.PLAIN,14);
        v.addElement(new TextLayout(f.getName(),f,frc));
        sleepAmount = 200;
    }


    public void reset(int w, int h) {
        TextLayout tl = (TextLayout) v.firstElement();
        strH = (int) (tl.getAscent()+tl.getDescent());
        nStrs = h/strH + 1;
        fi = 0;
    }


    public void step(int w, int h) {
        if (fi < fonts.size()) {
            Font f = ((Font) fonts.get(fi)).deriveFont(Font.PLAIN,14);
            v.addElement(new TextLayout(f.getName(),f,frc));
        }
        if (v.size() == nStrs && v.size() != 0 || fi > fonts.size()) {
            v.removeElementAt(0);
        }
        fi = (v.size() == 0) ? 0 : ++fi;
    }


    public void drawDemo(int w, int h, Graphics2D g2) {

        g2.setColor(Color.black);

        int yy = (fi >= fonts.size()) ? 0 : h - v.size() * strH - strH/2;

        for (int i = 0; i < v.size(); i++) {
            int sw = (int) ((TextLayout)v.get(i)).getBounds().getWidth();
            ((TextLayout) v.get(i)).draw(g2, (int) (w/2-sw/2),yy += strH);
        }

    }


    public static void main(String argv[]) {
        final DemoPanel dp = new DemoPanel(new AllFonts());
        Frame f = new Frame("Java2D Demo - AllFonts");
        f.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {System.exit(0);}
            public void windowDeiconified(WindowEvent e) { 
                dp.surface.start(); 
            }
            public void windowIconified(WindowEvent e) { 
                dp.surface.stop(); 
            }
        });
        f.add("Center", dp);
        f.pack();
        f.setSize(new Dimension(400,300));
        f.show();
        dp.surface.start();
    }
}
