/*
 * @(#)InvokerGen.java	1.19 98/08/23
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * This is a utility program that generates C-code for
 * custom JNI native method invokers for given
 * method signatures.
 */

import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;

class InvokerGen implements Constants {

    static DataInputStream in;
    static PrintStream out;
    static PrintStream err = System.err;
    static int tabLevel = 0;
    /* whether to print out debugging info in the
     * generated code - useful for debugging this
     * program and the code it generates.
     */
    static boolean debug = false;
    static void db(String s) {
	if (debug) {
	    p(s);
	}
    }

    static void p(String s) {
	/* Let's generate readable machine-generated code, shall we?
	 * Keep track of code blocks for purty indentation.
	 */
	if (s.length() > 0 && s.charAt(0) == '#') {
	    /* don't indent preprocessor directives */
	    out.println(s);
	    return;
	}
	if (s.indexOf('}') >= 0) {
	    tabLevel -= 4;
	}
	for (int i = 0; i < tabLevel; i++) {
	    out.write(' ');
	}
	out.println(s);
	if (s.indexOf('{') >= 0) {
	    tabLevel += 4;
	}
    }

    static void pi() {
	out.print("#include ");
    }

    static void pq(String s) {
	out.write('"');
	out.print(s);
	out.write('"');
	out.println();
    }

    static void args(String[] a) throws IOException {
	for (int i = 0; i < a.length; i++) {
	    if ("-o".equals(a[i])) {
		out = new PrintStream(new BufferedOutputStream(new FileOutputStream(a[++i]), 100));
	    } else if ("-i".equals(a[i])) {
		in = new DataInputStream(new BufferedInputStream(new FileInputStream(a[++i]), 100));
	    } else if ("-debug".equals(a[i])) {
		debug = true;
	    }
	}

	if (out == null) {
	    out = System.out;
	}
	if (in == null) {
	    in = new DataInputStream((System.in instanceof BufferedInputStream) ?
				     System.in: new BufferedInputStream(System.in, 100));
	}
    }

    static Hashtable sigs = new Hashtable(20);

    public static void main(String[] a) {
	try {
	    args(a);
	    p("/* MACHINE GENERATED - DO NOT EDIT */");
	    db("#ifdef DEBUG");
	    db("#define _DEBUGGING_INVOKER_GEN_");
	    db("#endif");
	    writeIncludes();
	    p("");
	    writeCodeBlock(pushJNIFrame);
	    readSignatures();
	    writeInvokers();
	    writeStaticData();
	    writeCodeBlock(getInvokerFunctions);
	    in.close();
	    out.close();
	    if (out.checkError()) {
		throw new IOException("error on output stream");
	    }
	    System.exit(0);
	} catch (Throwable t) {
	    t.printStackTrace();
	    System.exit(1);
	}
    }

    static void writeStaticData() {
	p("typedef struct TableEntry {");
	p("char *sig;");
	p("Invoker code;");
	db("#ifdef _DEBUGGING_INVOKER_GEN_");
	db("char *name;");
	db("#endif");
	p("} TableEntry;");
	p("");

	p("static TableEntry map[] = {");
	Enumeration e = sigs.keys();
	int i  = 0;
	while (i++ < sigs.size() - 1) {
	    SigInfo s = (SigInfo) e.nextElement();
	    p(s.hex+", invoke_" + s.terse+",");
	    db("#ifdef _DEBUGGING_INVOKER_GEN_");
	    db("\""+s.terse+"\",");
	    db("#endif");
	}
	SigInfo last = (SigInfo) e.nextElement();
	p(last.hex+", invoke_" + last.terse);
	db("#ifdef _DEBUGGING_INVOKER_GEN_");
	db(", \""+last.terse+"\"");
	db("#endif");
	p("};");
	p("/* end invoker table */");
	p("");
    }

    static void writeInvokers() {
	Enumeration e = sigs.keys();
	while (e.hasMoreElements()) {
	    out.println();
	    SigInfo s = (SigInfo) e.nextElement();
	    s.printDeclaration();
	    writeCodeBlock(invokerPreamble);
	    s.printInvocation();
	    s.printStackChange();
	    writeCodeBlock(invokerEnd);
	    p("/* End " + s.terse + " invoker */");
	    p("");
	}
    }

    static void readSignatures() throws IOException {
	String line;
	while ((line = in.readLine()) != null) {
	    line = line.trim();
	    if (line.indexOf('(') != 0) {
		continue;
	    }
	    SigInfo s = new SigInfo(line);
	    SigInfo tmp;
	    if ((tmp = (SigInfo) sigs.get(s)) != null) {
		err.println("same effective signature: " + s.sig + " ->\n\t" + tmp.sig + ", merging..");
	    } else {
		sigs.put(s, s);
	    }
	}
    }

    static void writeIncludes() throws IOException {

	for (int i = 0; i < angleInc.length; i++) {
	    pi();
	    out.write('<');
	    out.print(angleInc[i]);
	    out.println(">");
	}

	for (int i = 0; i < qInc.length; i++) {
	    pi();
	    pq(qInc[i]);
	}

	out.println();
    }

    static void writeCodeBlock(String[] cb) {
	for (int i = 0; i < cb.length; i++) {
	    p(cb[i]);
	}
    }
}

class SigInfo implements Constants {

    static void p(String s) {
	InvokerGen.p(s);
    }

    static void db(String s) {
	InvokerGen.db(s);
    }

    String terse, sig, hex;
    byte[] binTerse;
    byte returnType;
    /* */ SigInfo(String s) {
	sig = s;
	parse();
	hex();
	//	p(sig + " --> " + terse);
    }

    void printDeclaration() {
	PrintStream out = InvokerGen.out;
	/* print a comment */
	p("/* JNI Invoker for method signature:");
	out.print(" * (");
	int i;
	for (i = 0; binTerse[i] != S_END; i++) {
	    out.print(binTerse[i] == S_OBJECT ? "jobject": argTypes[binTerse[i]]);
	    if (binTerse[i+1] != S_END) {
		out.print(", ");
	    }
	}
	p(") --> " + (binTerse[++i] == S_OBJECT ? "jobject": argTypes[binTerse[i]]));
	p(" */\n");

	/* print typedef for actual native method function/signature.
	 * This makes the code much more readable.
	 */
	out.print("typedef " + argTypes[returnType] + " (JNICALL *"+terse+")(JNIEnv*, void*");
	for (i = 0; binTerse[i] != S_END; i++) {
	  out.write(',');
	  out.write(' ');
	  out.print(argTypes[binTerse[i]]);
	}
	p(");");
	/* Print the invoker signature */
	p("static bool_t");
	p("invoke_"+terse+invokerSignature);
	p("");
	printTmpVars();
    }

    void printInvocation() {
	StringBuffer sb = new StringBuffer();
	//	p("fprintf(stderr,\"invoker %s name=%s ns=%d\\n\", \"" + terse +
	//	  "\", mb->fb.name, ns);");
	if (returnType == S_LONG || returnType == S_DOUBLE || returnType == S_OBJECT) {
	    p("ret = ");
	} else if (returnType == S_FLOAT) {
	    sb.append("old_frame->optop[0].f = ");
	} else if (returnType != S_VOID) {
	    /* any other return type besides void is integral */
	    sb.append("old_frame->optop[0].i = ");
	}
	sb.append("((");
	sb.append(terse);
	sb.append(")(mb->code))(EE2JNIEnv(ee), &o");
	if (binTerse[0] == S_END) {
	    sb.append(");");
	} else {
	    sb.append(',');
	}
	p(sb.toString());
	InvokerGen.tabLevel += 2; // pretty print
	int stack = 0;
	for (int i = 0; binTerse[i] != S_END; i++) {
	    // remind: implement double word stuff
	    switch (binTerse[i]) {
	    case S_OBJECT:
		p("op["+(i+stack)+"].p == NULL ? NULL: &(op["+(i+stack)+"].p)"+
		  (binTerse[i+1] == S_END ? ");": ","));
		break;
	    case S_FLOAT:
		p("op["+(i+stack)+"].f"+(binTerse[i+1] == S_END ? ");": ","));
		break;
	    case S_LONG:
		p("GET_INT64(tmp8, op+"+(i+stack)+")"+(binTerse[i+1] == S_END ? ");": ","));
		stack++;
		break;
	    case S_DOUBLE:
		p("GET_DOUBLE(tmp8, op+"+(i+stack)+")"+(binTerse[i+1] == S_END ? ");": ","));
		stack++;
		break;
	    case S_VOID:
		throw new Error("??? Illegal signature");  // shouldn't happen
	    default:
		// integral types
		p("op["+(i+stack)+"].i"+(binTerse[i+1] == S_END ? ");": ","));
	    }
	}
	InvokerGen.tabLevel -= 2;
	if (returnType == S_OBJECT) {
	    p("old_frame->optop[0].p = (ret == NULL) ? NULL: *ret;");
	}
	//	p("fprintf(stderr,\"invoker %s name=%s returned\\n\", \"" + terse +
	//	  "\", mb->fb.name);");
    }

    /* we need temporary variables at function scope for two-word
     * return types.
     */
    void printTmpVars() {
	if (returnType == S_LONG) {
	    p("int64_t ret;");
	    p("jvalue tmp8;");
	} else if (returnType == S_DOUBLE) {
	    p("double ret;");
	    p("jvalue tmp8;");
	} else if (returnType == S_OBJECT) {
	    p("void **ret;");
	}
    }

    void printStackChange() {
	if (returnType == S_LONG) {
	    p("SET_INT64(tmp8, old_frame->optop, ret);");
	    p("old_frame->optop += 2;");
	} else if (returnType == S_DOUBLE) {
	    p("SET_DOUBLE(tmp8, old_frame->optop, ret);");
	    p("old_frame->optop += 2;");
	} else if (returnType != S_VOID) {
	    p("old_frame->optop++;");
	}
    }

    public boolean equals(Object o) {
	return o != null && (o instanceof SigInfo) &&
	    ((SigInfo)o).terse.equals(terse);
    }

    void hex() {
	StringBuffer sb = new StringBuffer();
	sb.append('"');
	for (int i = 0; i < binTerse.length && binTerse[i] != 0; i++) {
	    sb.append('\\');
	    sb.append('x');
	    sb.append('0');
	    sb.append(Integer.toHexString(binTerse[i]).toUpperCase());
	}
	sb.append('"');
	hex = sb.toString();
    }

    public String toString() {
	return sig;
    }

    public int hashCode() {
	return terse.hashCode();
    }

    void parse() {
	StringBuffer sb = new StringBuffer();
	byte[] t = new byte[257]; // max args + 2
	char[] c = sig.toCharArray();
	int i = 1, j =0;
	boolean hitEnd = false;

	for (i = 1; i < c.length; i++) {
	    switch (c[i]) {
	    case ')':
		hitEnd = true;
		if (j == 0) {
		    sb.append('V');
		}
		sb.append('_');
		t[j++] = S_END;
		break;
	    case '[':
		t[j++] = S_OBJECT;
		sb.append('O');
		while (i < c.length && c[i] == '[') {
		    i++;
		}
		if (i < c.length && c[i] == 'L') {
		    while (i < c.length && c[i] != ';') {
			i++;
		    }
		}
		break;
	    case 'L':
		t[j++] = S_OBJECT;
		sb.append('O');
		while (i < c.length && c[i] != ';') {
		    i++;
		}
		break;
	    case 'Z':
		sb.append(hitEnd ? 'Z': 'I');
		t[j++] = hitEnd ? S_BOOL: S_INT;
		break;
	    case 'B':
		sb.append(hitEnd ? 'B': 'I');
		t[j++] = hitEnd ? S_BYTE: S_INT;
		break;
	    case 'S':
		sb.append(hitEnd ? 'S': 'I');
		t[j++] = hitEnd ? S_SHORT: S_INT;
		break;
	    case 'C':
		sb.append(hitEnd ? 'C': 'I');
		t[j++] = hitEnd ? S_BOOL: S_CHAR;
		break;
	    case 'I':
		t[j++] = S_INT;
		sb.append('I');
		break;
	    case 'F':
		t[j++] = S_FLOAT;
		sb.append(c[i]);
		break;
	    case 'D':
		t[j++] = S_DOUBLE;
		sb.append(c[i]);
		break;
	    case 'J':
		t[j++] = S_LONG;
		sb.append(c[i]);
		break;
	    case 'V':
		if (hitEnd) {
		    sb.append(c[i]);
		    t[j++] = S_VOID;
		    break;
		}
	    default:
		throw new Error("bad char: c["+i+"]="+c[i]+", of sig='"+sig+"'");
	    }
	}
	t[j] = 0;
	returnType = t[--j];
	binTerse = t;
	terse = sb.toString();
    }
}

interface Constants {
    static final String[] angleInc = {
	"ctype.h", "string.h", "sys/types.h", "stdlib.h"
    };

    static final String[] qInc = {
	"oobj.h", "tree.h", "interpreter.h", "threads.h",
	"monitor.h", "exceptions.h", "signature.h",
	"sys_api.h", "vmprofiler.h"
    };

    static final String[] pushJNIFrame = {
	"static JavaFrame*",
	"pushJNIFrame(ExecEnv *ee, JavaFrame *old_frame, int args_size, struct methodblock *mb) {",
	"stack_item *optop = old_frame->optop;",
	"JavaStack  *stack = old_frame->javastack;",
	"JavaFrame  *frame = (JavaFrame *)(optop + args_size);",
	"if (frame->ostack + JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY >= stack->end_data) {",
	"JavaFrame *tmp_frame = frame;",
        "JavaStack *tmp_stack = stack;",
	"if (!ExpandJavaStackForJNI(ee, &tmp_stack, &tmp_frame, JNI_REF_INFO_SIZE + JNI_DEFAULT_LOCAL_CAPACITY)) {",
	"return 0;",
	"}",
        "frame = tmp_frame;",
	"stack = tmp_stack;",
	"}",
	"frame->javastack = stack;",
	"frame->prev = old_frame;",
	"frame->optop = frame->ostack + JNI_REF_INFO_SIZE;",
	"frame->current_method = mb;",
	"JNI_REFS_FREELIST(frame) = NULL;",
	"JNI_N_REFS_IN_USE(frame) = 0;",
	"JNI_REFS_CAPACITY(frame) = JNI_DEFAULT_LOCAL_CAPACITY;",
	"ee->current_frame = frame;",
	"return frame;",
	"}"
    };

    static String[] getInvokerFunctions = {
	"static int invokersSorted = 0;",
	"static int",
	"entryCmp(const void *e1, const void *e2) {",
	"return strcmp(((TableEntry *)e1)->sig, ((TableEntry*)e2)->sig);",
	"}",
	"/* Get a custom JNI invoker if one is present for the given signature. */",
	"Invoker",
	"getCustomInvoker(char *sig) {",
	"TableEntry cmp;",
	"TableEntry *ret;",
	"cmp.sig = sig;",
	"if (!invokersSorted) {",
	"qsort(&(map[0]), sizeof(map)/sizeof(TableEntry),",
	"sizeof(TableEntry), entryCmp);",
	"invokersSorted = 1;",
	"}",
	"ret = (TableEntry *)bsearch(&cmp, ",
	"    &(map[0]),",
	"    sizeof(map)/sizeof(TableEntry),",
	"    sizeof(TableEntry),",
	"    entryCmp);",
	"#ifdef _DEBUGGING_INVOKER_GEN_",
	"if (ret) {",
	"fprintf(stderr, \"custom invoker %s\\n\", ret->name);",
	"}",
	"#endif",
	"return ret ? ret->code: NULL;",
	"}"
    };

    static final String[] invokerPreamble = {
	"JavaFrame  *old_frame = ee->current_frame; /* previous frame */",
	"stack_item *op = old_frame->optop; /* Points to just before args */",
	"JavaFrame  *frame = pushJNIFrame(ee, old_frame, args_size, mb);",
	"#ifdef DEBUG",
	"invokerValidateArgs(op, mb);",
	"#endif /* DEBUG */",
	"if ((mb->fb.access & ACC_STATIC) == 0) {",
	"op++;",
	"}",
	"if (frame == 0) {",
	"return FALSE;",
	"}",
	"ee->current_frame = frame;",
	"#ifdef HPROF",
        "if (jvmpi_event_flags & JVMPI_EVENT_METHOD_ENTRY_ON) {",
        "jvmpi_method_entry(ee, o);",
	"}",
	"#endif /* HPROF */",
    };

    static final String[] invokerEnd = {
	"#ifdef HPROF",
	"if (JVMPI_EVENT_IS_ENABLED(JVMPI_EVENT_METHOD_EXIT)) {",
        "jvmpi_method_exit(ee);",
	"}",
	"#endif /* HPROF */",
        "CHECK_CRITICAL_COUNT_ON_RETURN(ee);",
	"ee->current_frame = old_frame;",
	"TRACE_METHOD(ee, mb, args_size, TRACE_METHOD_NATIVE_RETURN);",
	"return !(exceptionOccurred(ee));",
	"}"
    };

    static final String[] argTypes = {
	"" /* null, shouldn't happen */, "void **", "int64_t", "double",
	"jboolean", "jbyte", "jshort", "jchar", "jint", "jfloat", "void",
	"*bogus*" /* shouldn't happen */
    };

    /* keep these in sync w/ signature.h ! */
    public static final byte S_OBJECT = 1;
    public static final byte S_LONG = 2;
    public static final byte S_DOUBLE = 3;
    public static final byte S_BOOL = 4;
    public static final byte S_BYTE = 5;
    public static final byte S_SHORT = 6;
    public static final byte S_CHAR = 7;
    public static final byte S_INT = 8;
    public static final byte S_FLOAT = 9;
    public static final byte S_VOID = 10;
    public static final byte S_END = 11;
    public static final String invokerSignature =
    "(JHandle *o, struct methodblock *mb,\n\t int args_size, ExecEnv *ee) {";

}









