/*
 * @(#)exceptions.c	1.44 98/08/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "exceptions.h"

/* Always have these exceptions object ready so that we don't
 * have to allocate them in horrible situations. */
/* The classes the internal exception objects belong to */
static char *internalExceptionClasses[] = {
    NULL,        /* for  IEXC_NONE, not useful, only here not to break JITs */
    JAVAPKG "NoClassDefFoundError",
    JAVAPKG "OutOfMemoryError",
    NULL,
};

/* An array of handles to the internal exception objects */
static JHandle *internalExceptions[] = {
    NULL,
    NULL,
    NULL,
    NULL,
};

/*
 * Internal routines.
 */

/*
 * exceptionInitInternalObjects() -- Initialize the exception subsystem's
 *	internal objects.
 */
void
exceptionInit(void) {
    int i;

    /* Initialize the internal objects */
    for (i = IEXC_NONE + 1; internalExceptionClasses[i]; i++) {
	/* Look up class, making it sticky */
	ClassClass *cb = 
	    FindStickySystemClass(NULL, internalExceptionClasses[i], TRUE);
	if (cb != 0) {
	    ExecEnv *ee = EE();
	    JNIEnv *env = EE2JNIEnv(ee);
	    internalExceptions[i] = allocObject(ee, cb);
	    (*env)->PushLocalFrame(env, 3);
            (*env)->NewGlobalRef(env, MkRefLocal(env, internalExceptions[i]));
	    (*env)->PopLocalFrame(env, 0);
	}
    }
}

/*
 * exceptionInternalObject() -- Return an internal, preallocated
 *	exception object.  These are shared by all threads, so they
 *	should only be used in a last ditch effort.
 */
exception_t	    
exceptionInternalObject(internal_exception_t exc)
{
    if (exc <= IEXC_NONE || exc >= IEXC_END) {
	return NULL;
    }
    return internalExceptions[exc];
}

/*
 * exceptionDescribe() -- Print out a description of a given exception
 *	object.
 */
void
exceptionDescribe(ExecEnv *ee)
{
    exception_t exc = ee->exception.exc;

    if (is_instance_of(exc, classJavaLangThreadDeath, ee)) {
	/*EMPTY*/
	/* Don't print anything if we are being killed. */
    } else {
	jio_fprintf(stderr, "Exception ");
	if (threadSelf()) {
	    jio_fprintf(stderr,"in thread ");
	    jio_fprintf(stderr,Object2CString((JHandle *)getThreadName()));
	}

	if (is_instance_of(exc, classJavaLangThrowable, ee)) {
	    jio_fprintf(stderr," ");
	    /* Object is an exception. */
	    execute_java_dynamic_method(ee, exc, "printStackTrace", "()V");
	} else {
	    char buf[MSG_BUF_LEN];
	    ClassClass *class = getClass(exc);
	    jio_fprintf(stderr, ".  Uncaught exception of type %s\n", 
		   classname2string(cbName(class), buf, sizeof(buf)));
	}
    }

}
