/*
 * @(#)javai.c	1.63 98/11/03
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <stdio.h>
#include "oobj.h"
#include "interpreter.h"
#include "log.h"
#include "sys_api.h"
#include "alloc_cache.h"
#include "threads.h"
#include "exceptions.h"
#include "gc.h"
#include "monitor_cache.h"

#ifdef HPROF
#include "vmprofiler.h"
#endif /* HPROF */

#ifdef BREAKPTS
#include "breakpoints.h"
#endif /* BREAKPTS */

static int TimeSlice = 0;

/* Use as few signals as possible when this option is on */
bool_t reduce_signal_usage = FALSE; 

static int specified_monitor_cache_size = 0;

/* port number for talking to debugger agent */
static int debugPort;

static int initHeapSize;
static int maxHeapSize;

/* temporary place for storing the dll name and options */
static char * dll_options = NULL;
static char * dll_name = NULL;

/*
 * Set to true if the _classpath VM option is used indicating that we
 * would like to use the old style launcher where application classes
 * are also loaded from the system class path. This is the default for
 * JNI 1.1.
 */
jboolean oldjava = JNI_FALSE;

#ifdef BREAKPTS
/* Retrieve special JNI for debugging 
 */
extern struct JNINativeInterface_ *jvmdi_jni_GetNativeInterface();
#endif /* BREAKPTS */

/*
 * Default values for minimum and maximum amounts of memory to devote
 * to the Java heap.  The current abstraction, which may not hold up
 * forever, is that the initial size of the heap is the minimum heap.
 *
 * The maximum memory to map: As we're mmapping the memory, it looks
 * like we should ask swapctl(2) how much swap we have, and thus what
 * the maximum we've got is.
 *
 * These values can be overridden on the java command line.
 */

#ifdef USE_MALLOC
#define MINHEAPSIZE ((int) ( 3 * (1024*1024)))
#define MAXHEAPSIZE MINHEAPSIZE
#else
#define MINHEAPSIZE ((int) ( 1 * (1024*1024)))
#define MAXHEAPSIZE ((int) (64 * (1024*1024)))
#endif /* USE_MALLOC */

/*
 * The default thread stack size.  This is machine-dependent, because it
 * is based on a platform's C stack usage.
 */
#define PROCSTACKSIZE 0     /* 0 means default size in HPI */
long ProcStackSize = PROCSTACKSIZE;

#define MIN_HEAP_SIZE (256 * 1024)

#ifdef JCOV

static void
jcov_setup(char * args)
{
    int wrk_jcov_flags = JCOV_MODE_DEFAULT;

    if (args != NULL) {
	char * comma;

	while (*args) {
	    if (strncmp(args, "file=", 5) == 0) {
		int len;
		args += 5;
		comma = strchr(args, ',');
		if (comma) {
		    len = comma - args;
		} else {
		    len = strlen(args);
		}
		cov_file = sysMalloc(len + 1);
		strncpy(cov_file, args, len);
		cov_file[len] = '\0';
		args += len;
		if (*args == ',') args++;
	    } else if (strncmp(args, "type=", 5) == 0) {
		int len;
		args += 5;
		comma = strchr(args, ',');
		if (comma) {
		    len = comma - args;
		} else {
		    len = strlen(args);
		}
		if (strncmp(args, "m", len) == 0) {
    		    wrk_jcov_flags = JCOV_MODE_METHODS;
		} else {
		    jio_fprintf(stderr, "Unknown type option: \"%s\"\n", args);
		    Exit(1);
		}
		args += len;
		if (*args == ',') args++;
	    } else {
		jio_fprintf(stderr, "Unknown args: \"%s\"\n", args);
		Exit(1);
	    }
	}
    }

    if (jcov_flags == 0)
    	AtExit(jcov_write_data);
    if (cov_file == NULL)
	cov_file = "java.jcov";
    jcov_flags |= wrk_jcov_flags;
    classgc = FALSE;
}
#endif

typedef struct prop_t {
    struct prop_t *next;
    char *name;
    char *value;
} prop_t;

static prop_t *init_properties;

bool_t addUserProperty(char *str)
{
    prop_t *prop = sysMalloc(sizeof(prop_t));
    prop_t **p;
    char *value = strchr(str, '=');
    int name_len = value ? value - str : strlen(str);
    int value_len;

    value = value ? value + 1 : "";
    value_len = strlen(value);

    if (prop == NULL) {
        return FALSE;
    }

    prop->name = sysMalloc(name_len + 1);
    prop->value = sysMalloc(value_len + 1);

    if (prop->name == NULL || prop->value == NULL) {
        return FALSE;
    }

    strncpy(prop->name, str, name_len);
    strncpy(prop->value, value, value_len);
    prop->name[name_len] = '\0';
    prop->value[value_len] = '\0';
    /* Add the new property at the end of the list */
    p = &init_properties;
    while (*p) {
        p = &((*p)->next);
    }
    prop->next = 0;
    *p = prop;
    return TRUE;
}

static jint Initialize11(JDK1_1InitArgs *args)
{
    char **props;
    debugging = args->debugging;
    debugPort = args->debugPort;
    initHeapSize = args->minHeapSize;
    maxHeapSize = args->maxHeapSize;
    verbose = args->verbose;
    classgc = args->enableClassGC;
    verbosegc = args->enableVerboseGC;
    if (args->classpath) {
        init_sysclasspath = args->classpath;
    }

    props = args->properties;
    if (props) {
        while (*props) {
	    if (!addUserProperty(*props)) {
	        return JNI_ENOMEM;
	    }
	    props++;
        }
    }

    if (args->nativeStackSize == 0 || args->nativeStackSize >= 1000) {
        ProcStackSize = args->nativeStackSize;
    } else {
        return JNI_EINVAL;
    }
    if (args->javaStackSize >= 1000) { 
        JavaStackSize = args->javaStackSize;
    } else {
        return JNI_EINVAL;
    }

    if ((args->verifyMode == VERIFY_NONE) || 
	(args->verifyMode == VERIFY_REMOTE) || 
	(args->verifyMode == VERIFY_ALL))
        verifyclasses = args->verifyMode;

    vfprintf_hook = args->vfprintf;
    exit_hook = args->exit;
    abort_hook = args->abort;

    oldjava = JNI_TRUE;
    return JNI_OK;
}

/*
 * Parses a memory size specification string.
 */
static jint atomi(char *s)
{
    long n = strtol(s, &s, 10);
    switch (*s) {
    case 'M': case 'm':
	n *= 1024 * 1024;
	break;
    case 'K': case 'k':
	n *= 1024;
	break;
    case '\0':
	break;
    default:
	return -1;
    }
    return n;
}

static jint Initialize12(JavaVMInitArgs *args, bool_t setDefault)
{
    int i;

    if (setDefault) {
        ProcStackSize = PROCSTACKSIZE;
	JavaStackSize = JAVASTACKSIZE;
	initHeapSize  = MINHEAPSIZE;
	maxHeapSize   = MAXHEAPSIZE;
    }

    for (i = 0; i < args->nOptions; i++) {
        JavaVMOption *option = args->options + i;
	if (!strncmp(option->optionString, "-D", 2)) {
	    if (!addUserProperty(option->optionString + 2)) {
	        return JNI_ENOMEM;
	    }
	} else if (!strncmp(option->optionString, "-verbose", 8)) {
	    char *s = option->optionString + 8;
	    if (*s == 0) {
	        verbose = 1;
	    } else {
	        if (*s != ':') {
		    jio_fprintf(stderr, "Illegal option: %s\n", 
				option->optionString);
		    return JNI_EINVAL;
		}
		s++;
	        while (*s) {
		    char opt[128];
		    int n;

		    for (n = 0; n < sizeof(opt) - 1; n++) {
		        if (*s == '\0') {
			    break;
			}
			if (*s == ',') {
			    s++;
			    break;
			}
			opt[n] = *s++;
		    }
		    opt[n] = '\0';

		    if (0 == strcmp(opt, "gc")) {
		        verbosegc = 1;
		    } else if (0 == strcmp(opt, "Xclassdep")) {
			verboseclassdep = TRUE;
		    } else if (0 == strcmp(opt, "class")) {
		        verbose = 1;
		    } else if (0 == strcmp(opt, "jni")) {
		        verbose_jni = TRUE;
		    } else {
			if (!args->ignoreUnrecognized) {
			    jio_fprintf(stderr,
					"Unrecognized verbose option: %s\n",
					option->optionString);
			    return JNI_EINVAL;
			}
		    }
		}
	    }
	} else if (!strcmp(option->optionString, "vfprintf")) {
	    vfprintf_hook = (jint (JNICALL *)(FILE *, const char *, va_list))
	        (option->extraInfo);
	} else if (!strcmp(option->optionString, "exit")) {
	    exit_hook = (void (JNICALL *)(jint))(option->extraInfo);
	} else if (!strcmp(option->optionString, "abort")) {
	    abort_hook = (void (JNICALL *)(void))(option->extraInfo);
	} else if (!strncmp(option->optionString, "-Xbootclasspath/a:", 18)) {
	    char *p = option->optionString + 18;
	    char *np = sysMalloc(strlen(p) + strlen(init_sysclasspath) + 3);
	    if (np == NULL) {
	        return JNI_ENOMEM;
	    }
	    sprintf(np, "%s" PATH_SEPARATOR "%s", init_sysclasspath, p);
	    init_sysclasspath = np;
	} else if (!strncmp(option->optionString, "-Xbootclasspath/p:", 18)) {
	    char *p = option->optionString + 18;
	    char *np = sysMalloc(strlen(p) + strlen(init_sysclasspath) + 3);
	    if (np == NULL) {
	        return JNI_ENOMEM;
	    }
	    sprintf(np, "%s" PATH_SEPARATOR "%s", p, init_sysclasspath);
	    init_sysclasspath = np;
	} else if (!strncmp(option->optionString, "-Xbootclasspath:", 16)) {
	    init_sysclasspath = option->optionString + 16;
	} else if (!strcmp(option->optionString, "-Xoldjava")) {
	    oldjava = JNI_TRUE;
#ifdef TRACING
	} else if (!strcmp(option->optionString, "-Xt")) {
	    trace++;
	} else if (!strcmp(option->optionString, "-Xtm")) {
	    tracem++;
#endif
	} else if (!strcmp(option->optionString, "-Xdebug")) {
	    debugging = TRUE;
	} else if (!strcmp(option->optionString, "-Xnoagent")) {
	    debugPort = -1;
	} else if (!strncmp(option->optionString, "-Xdebugport", 11)) {
	    debugPort = atoi(option->optionString + 11);
	} else if (!strcmp(option->optionString, "-Xnoclassgc")) {
	    classgc = FALSE;
	} else if (!strncmp(option->optionString, "-Xverify", 8)) {
	    char *kind = option->optionString + 8;
	    if (*kind == 0)
	        verifyclasses = VERIFY_ALL;
	    else if (!strcmp(kind, ":all"))
	        verifyclasses = VERIFY_ALL;
	    else if (!strcmp(kind, ":remote"))
	        verifyclasses = VERIFY_REMOTE;
	    else if (!strcmp(kind, ":none"))
	        verifyclasses = VERIFY_NONE;
	    else {
	        jio_fprintf(stderr, "Illegal option: %s\n", 
			    option->optionString);
		return JNI_EINVAL;
	    }
#ifdef TRACEGC
	} else if (!strcmp(option->optionString, "-Xtracegc")) {
	    tracegc++;
#endif
#ifdef JCOV
	} else if (!strncmp(option->optionString, "-Xjcov", 6)) {
	    char *p = option->optionString + 6;
	    if (*p) {
	        if (*p == ':') {
		    p++; /* skip ':' */
		} else {
		    jio_fprintf(stderr, "Illegal option: %s\n", 
				option->optionString);
		    return JNI_EINVAL;
		}
	    }
	    jcov_setup(p);
#endif
	} else if (!strncmp(option->optionString, "-Xss", 4)) {
	    ProcStackSize = atomi(option->optionString + 4);
	    if (ProcStackSize < 1000) {
	        jio_fprintf(stderr, "Bad native stack size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xoss", 5)) {
	    JavaStackSize = atomi(option->optionString + 5);
	    if (JavaStackSize < 1000) {
	        jio_fprintf(stderr, "Bad Java stack size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xms", 4)) {
	    initHeapSize = atomi(option->optionString + 4);
	    if (initHeapSize < 0) {
	        jio_fprintf(stderr, "Bad init heap size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xmx", 4)) {
	    maxHeapSize = atomi(option->optionString + 4);
	    if (maxHeapSize < 0) {
	        jio_fprintf(stderr, "Bad max heap size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xmaxf", 6)) {
	    maxHeapFreePercent = atof(option->optionString + 6);
	    if (maxHeapFreePercent < 0 || maxHeapFreePercent > 1) {
	        jio_fprintf(stderr, "Bad maxHeapFreePercent size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xminf", 6)) {
	    minHeapFreePercent = atof(option->optionString + 6);
	    if (minHeapFreePercent < 0 || minHeapFreePercent > 1) {
	        jio_fprintf(stderr, "Bad minHeapFreePercent size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xmaxe", 6)) {
	    maxHeapExpansion = atomi(option->optionString + 6);
	    if (maxHeapExpansion < 0) {
	        jio_fprintf(stderr, "Bad maxHeapExpansion size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xmine", 6)) {
	    minHeapExpansion = atomi(option->optionString + 6);
	    if (minHeapExpansion < 0) {
	        jio_fprintf(stderr, "Bad minHeapExpansion size: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	} else if (!strncmp(option->optionString, "-Xnm", 4)) {
	    specified_monitor_cache_size = atoi(option->optionString + 4);
#ifdef LOGGING
	} else if (!strncmp(option->optionString, "-Xl", 3)) {
	    logging_level = atoi(option->optionString + 3);
#endif
#ifdef STATISTICS
	} else if (!strcmp(option->optionString, "-Xstats")) {
	    AtExit(DumpStatistics);
#endif
	} else if (!strncmp(option->optionString, "-Xrun", 5)) {
	    char libname[128];
	    char *p = option->optionString + 5;
	    char *q = libname;
	    while (*p && *p != ':') {
	        *q++ = *p++;
	    }
	    *q = '\0';
	    if (*p == ':') p++;
	    dll_name = sysStrdup(libname);
	    dll_options = sysStrdup(p);
#ifdef HPROF
	} else if (!strncmp(option->optionString, "-Xhprof", 7)) {
	    char *p = option->optionString + 7;
	    if (*p == ':') p++;
	    dll_name = sysStrdup("hprof");
	    dll_options = sysStrdup(p);
#endif
#ifdef HEAPMETERS
	} else if (!strcmp(option->optionString, "-Xheapmeters")) {
	    enableHeapMeters = 1;
	    AtExit(dumpHeapMeters);
#endif
	} else if (!strncmp(option->optionString, "-Xmc", 4)) {
	    allocCacheSize = atomi(option->optionString + 4);
	} else if (!strncmp(option->optionString, "-Xml", 4)) {
	    allocLocalSize = atomi(option->optionString + 4);
	} else if (!strncmp(option->optionString, "-Xmh", 4)) {
	    allocHandleCount = atomi(option->optionString + 4);
	} else if (!strcmp(option->optionString, "-Xrs")) {
	    reduce_signal_usage = TRUE;
	} else if (!strncmp(option->optionString, "-Xts", 4) && 
		   sysAdjustTimeSlice(0) != SYS_ERR) {
	    TimeSlice = atoi(option->optionString + 4);
#ifdef CHECK_JNI
	} else if (!strncmp(option->optionString, "-Xcheck:jni", 11)) {
	    jni_NativeInterfacePtr = &checked_jni_NativeInterface;
#endif
	} else {
	    if ((strncmp(option->optionString, "-X", 2) ||
		 option->optionString[0] == '_') &&
		args->ignoreUnrecognized) {
	        continue;
	    } else {
	        jio_fprintf(stderr, "Unrecognized option: %s\n",
			    option->optionString);
		return JNI_EINVAL;
	    }
	}
    }
    return JNI_OK;
}

static vm_calls_t callbacks = {
    jio_fprintf,
    panic,
    monitorRegister,

#ifdef HPROF
    jvmpi_monitor_contended_enter,
    jvmpi_monitor_contended_entered,
    jvmpi_monitor_contended_exit,
#else
    NULL,
    NULL,
    NULL,
#endif /* HPROF */
};

HPI_MemoryInterface  *hpi_memory_interface;
HPI_LibraryInterface *hpi_library_interface;
HPI_SystemInterface  *hpi_system_interface;
HPI_ThreadInterface  *hpi_thread_interface;
HPI_FileInterface    *hpi_file_interface;
HPI_SocketInterface  *hpi_socket_interface;

GetInterfaceFunc GetHPI = NULL;

typedef jint (JNICALL *JVM_OnLoad_t)(JavaVM *, char *, void *);

static bool_t
loadJVMHelperLib(char *name, char *options)
{
    JNIEnv *env = EE2JNIEnv(EE());
    JVM_OnLoad_t JVM_OnLoad;
    const char *onLoadSymbols[] = JVM_ONLOAD_SYMBOLS;
    char fn[MAXPATHLEN];
    char fn2[MAXPATHLEN];
    char errmsg[STK_BUF_LEN*2];
    void *handle;
    int i;

    /* First look in the Java system directory, then in PATH */
    sysBuildLibName(fn, sizeof(fn), java_dll_dir, name);
    sysBuildLibName(fn2, sizeof(fn2), "", name);
    if (((handle = sysLoadLibrary(fn, errmsg, sizeof(errmsg))) == NULL)
	&& ((handle = sysLoadLibrary(fn2, errmsg, sizeof(errmsg))) == NULL))
    {
        jio_fprintf(stderr, "Can't load \"%s\", because %s\n", fn2, errmsg);
	return FALSE;
    }

    for (i = 0; i < sizeof(onLoadSymbols) / sizeof(char *); i++) {
        JVM_OnLoad = (JVM_OnLoad_t)sysFindLibraryEntry(handle, onLoadSymbols[i]);
	if (JVM_OnLoad) {
	    break;
	}
    }
    if (JVM_OnLoad) {
        jint res;
	JavaVM *jvm;
	(*env)->GetJavaVM(env, &jvm);
	res = (*JVM_OnLoad)(jvm, options, NULL);
	if (res < 0) {
	    jio_fprintf(stderr, "unable to initialize JVM helper library %s\n", name);
	    return FALSE;
	}
	return TRUE;
    } else {
        jio_fprintf(stderr, "corrupted JVM helper library %s\n", name);
        return FALSE;
    }
}

int InitializeJavaVM(void *args)
{
    ExecEnv *ee;
    sys_thread_t *sys_tid;
    HThread *self;
    jint version = *(jint*)args;
    jint result;
    props_md_t *sprops;

    /* Check if the host application provides an HPI implementation */
    if (version == JNI_VERSION_1_2) {
        JavaVMInitArgs *vm_args = args;
	int i;
	for (i = 0; i < vm_args->nOptions; i++) {
	    JavaVMOption *option = vm_args->options + i;
	    if (strcmp(option->optionString, "_hpi") == 0) {
	        GetHPI = (GetInterfaceFunc)option->extraInfo;
	    }
	}
    }

    if (GetHPI == NULL) {
        GetHPI = (GetInterfaceFunc)InitializeHPI(&callbacks);
    }

    if (GetHPI == NULL ||
	GetHPI((void **)&hpi_memory_interface, "Memory", 1) ||
	GetHPI((void **)&hpi_library_interface, "Library", 1) ||
	GetHPI((void **)&hpi_system_interface, "System", 1) ||
	GetHPI((void **)&hpi_thread_interface, "Thread", 1) ||
	GetHPI((void **)&hpi_file_interface, "File", 1)) {
        return JNI_ERR;
    }

    sprops = GetPropertiesMD();
    if (sprops == NULL) {
        jio_fprintf(stderr, "Can't obtain system-specific information\n");
        return JNI_ERR;
    }
    java_home_dir = sprops->java_home;
    java_dll_dir = sprops->dll_dir;
    init_sysclasspath = sprops->sysclasspath;

#if defined(DEBUG) && defined(CHECK_JNI)
    jni_NativeInterfacePtr = &checked_jni_NativeInterface;
#else
    jni_NativeInterfacePtr = &unchecked_jni_NativeInterface;
#endif /* DEBUG && CHECK_JNI */

    if (version == JNI_VERSION_1_1) {
        result = Initialize11((JDK1_1InitArgs *)args);
    } else if (version == JNI_VERSION_1_2) {
        result = Initialize12((JavaVMInitArgs *)args, TRUE);
    } else {
        jio_fprintf(stderr, "Unrecognized JNI version: 0x%lx\n", version);
        result = JNI_EVERSION;
    }
#ifdef BREAKPTS
    if (debugging) {
        jni_NativeInterfacePtr = jvmdi_jni_GetNativeInterface();
    }
#endif /* BREAKPTS */

    if (result < 0)
	return result;

#ifndef NO_HIDDEN_ENV
    {
        char *cmd_env = getenv("_JAVA_OPTIONS");
	if (cmd_env && *cmd_env) {
	    #define N_MAX_OPTIONS 32
	    JavaVMInitArgs vm_args;
	    JavaVMOption options[N_MAX_OPTIONS];
	    int i;
	    cmd_env = sysStrdup(cmd_env);
	    if (cmd_env == NULL) {
	        return JNI_ERR;
	    }
	    jio_fprintf(stderr, "Picked up _JAVA_OPTIONS: %s\n", cmd_env);
	    for (i = 0; i < N_MAX_OPTIONS;) {
		options[i++].optionString = cmd_env;
  	        if (*cmd_env != '-') {
		    jio_fprintf(stderr, "Bad _JAVA_OPTIONS: %s\n", cmd_env);
		    return JNI_ERR;
		}
		while (*cmd_env && *cmd_env != ' ') {
		    cmd_env++;
		}
		if (*cmd_env == 0) {
		    break;
		}
		*cmd_env++ = 0;
		while (*cmd_env == ' ') {
		    cmd_env++;
		}
	    }
	    vm_args.version = JNI_VERSION_1_2;
	    vm_args.options = options;
	    vm_args.nOptions = i;
	    vm_args.ignoreUnrecognized = FALSE;
	    if (Initialize12(&vm_args, FALSE) < 0) {
	        return JNI_ERR;
	    }
	}
    }

    {
	char *sizestr = getenv("_MIN_JAVASTACK_CHUNK_SIZE");
	if (sizestr) {
	    min_javastack_chunk_size = atoi(sizestr);
	    jio_fprintf(stderr, "Java stack chunks set to %d bytes.\n",
			min_javastack_chunk_size);
	}
    }
#endif /* NO_HIDDEN_ENV */

    /* Set the rest of callbacks slots to appropriate values */
#ifdef LOGGING
    sysSetLoggingLevel(logging_level);
#endif

    /* Initialize the monitor registry */
    monitorRegistryInit();

    if (sysThreadBootstrap(&sys_tid, &_queue_lock,
			   offsetof(ExecEnv, sys_thr))) {
        jio_fprintf(stderr, "Can't bootstrap threads\n");
        return JNI_ERR;
    }

    ee = SysThread2EE(sys_tid);

    /* The setting of the stack base pointer is only temporarly.
     * JNI_CreateJavaVM will later set stack_base back to zero.
     * (Invocation API functions are resilient to stack location.)
     */
    if (!InitializeExecEnv(ee, 0, &args)) {
	return JNI_ENOMEM;
    }

    AdjustUserThreadCount(1);
    
    /* Create the monitor cache */
    monitorCacheInit();

#ifdef DEBUG
    pExecuteJava = ExecuteJava_C;
#else
    if (debugging) {
        pExecuteJava = ExecuteJava_C;
    } else {
        pExecuteJava = ExecuteJava;
    }
#endif /* DEBUG */

    if (dll_name) {
	if (!loadJVMHelperLib(dll_name, dll_options)) {
	    return JNI_ERR;
	}

	sysFree(dll_name);
	sysFree(dll_options);
    }

    if (initHeapSize < MIN_HEAP_SIZE) {
        jio_fprintf(stderr, "The specified initial heap size is too small. "
		    "(%d bytes required.)\n", MIN_HEAP_SIZE);
	return JNI_EINVAL;
    }
    if (initHeapSize > maxHeapSize) {
	jio_fprintf(stderr,
	    "Incompatible initial and maximum heap sizes specified:\n\n");
	jio_fprintf(stderr,
	    "    initial size: %d bytes, maximum size: %d bytes\n\n",
	    initHeapSize, maxHeapSize);
	jio_fprintf(stderr,
	    "The initial heap size must be less than or equal to the maximum heap size.\n");
	jio_fprintf(stderr,
	    "The default initial and maximum heap sizes are %d and %d bytes.\n",
	    MINHEAPSIZE, MAXHEAPSIZE);
	return JNI_EINVAL;
    }

    if (InitializeAlloc(maxHeapSize, initHeapSize) != TRUE) {
        jio_fprintf(stderr, "Unable to initialize Java heap.\n");
	return JNI_ENOMEM;
    }

    UseLosslessQuickOpcodes = TRUE; 
    if (!InitializeInterpreter()) {
        return JNI_ENOMEM;
    }

    /* Check cache parameters */
    if (allocLocalSize >= allocCacheSize) {
	jio_fprintf(stderr,
	    "Allocation cache size (%d) must be greater than "
	    "local allocation size (%d)\n",
	    allocCacheSize, allocLocalSize);
	jio_fprintf(stderr,
	    "The default cache and local allocation sizes are "
	    " %d bytes and %d bytes\n",
	    ALLOC_CACHE_SIZE, ALLOC_LOCAL_SIZE);
	jio_fprintf(stderr,
	    "Note: You can turn off local allocation via '-Xml0'\n.");
	return JNI_EINVAL;
    }

    if (!LoadJavaLibrary() || !LoadZipLibrary()) {
        return JNI_ERR;
    }
    
    /* Before this point we must not load any classes. Class loading
     * requires the Java and Zip libraries to be loaded.
     */
    
    /* Preallocate the internal exception objects */
    exceptionInit();

    self = InitializeClassThread(ee);
    if (self == 0) {
        if (exceptionOccurred(ee)) {
	    goto error;
	}
	jio_fprintf(stderr, "Unable to initialize threads\n");
	return JNI_ERR;
    }

    InitializeSignals();

    if (TimeSlice) {
        sysAdjustTimeSlice(TimeSlice);
    }

    InitializeRefs();

    {
        ClassClass *cb = FindClass(ee, "java/lang/System", TRUE);
	if (exceptionOccurred(ee)) {
	    goto error;
	}
	execute_java_static_method(ee, cb, "initializeSystemClass", "()V");
	if (exceptionOccurred(ee)) {
	    goto error;
	}
    }

    if (!InitializeSystemClassLoader()) {
        goto error;
    }

    /* A temporary workaround to delay the loading of JIT to reduce
     * startup time. The JIT is loaded when the JNI FindClass function
     * is called.
     */
#ifdef DELAY_JIT_LOADING
#else
    FindClass(ee, JAVAPKG "Compiler", TRUE);
    if (!compilerInitialized) {
	/* Either the compiler couldn't be loaded, or it wasn't initialized */
	UseLosslessQuickOpcodes = FALSE;
    }
#endif

#ifdef BREAKPTS
    if (debugging) {
        /* So that field watchpoints can work */
        UseLosslessQuickOpcodes = TRUE;
    }
#endif /* BREAKPTS */

    if (debugging && debugPort >= 0) {
	/* JNI FindClass is base class loader aware. Since Agent class
	 * is in tools.jar, the Agent has to be found with JNI FindClass
	 * as opposed to the VM's FindClass */
	JNIEnv *env = EE2JNIEnv(ee);
        ClassClass *agentcb = (ClassClass *)DeRef(env,
	    (*env)->FindClass(env, "sun/tools/agent/Agent"));
	if (exceptionOccurred(ee)) {
	    goto error;
	}
	execute_java_static_method(0, agentcb, "boot", "(I)V", debugPort);
	if (exceptionOccurred(ee)) {
	    goto error;
	}
    }

#ifndef FAST_MONITOR
    /* Expand the monitor cache if the default was overridden */
    if (specified_monitor_cache_size - monCount > 0) {
        monitorCacheExpand(specified_monitor_cache_size - monCount);
    }
#endif/*FAST_MONITOR*/

#ifdef HPROF
    if (jvmpi_event_flags) {
        if (jvmpi_jvm_init_done() < 0) {
	    jio_fprintf(stderr, "profiler error\n");
	    return JNI_ERR;
	}
    }
#endif /* HPROF */

#ifdef BREAKPTS
    if (debugging) {
        notify_debugger_of_vm_init(ee);
    }
#endif /* BREAKPTS */

    return JNI_OK;

 error:
    exceptionDescribe(ee);
    return JNI_ERR;
}

int GetDefaultJavaVMInitArgs(JDK1_1InitArgs *args)
{
    props_md_t *sprops;

    sprops = GetPropertiesMD();
    if (sprops == NULL) {
        return JNI_ERR;
    }

    java_home_dir = sprops->java_home;
    java_dll_dir = sprops->dll_dir;

    args->nativeStackSize = PROCSTACKSIZE;
    args->minHeapSize     = MINHEAPSIZE;
    args->maxHeapSize     = MAXHEAPSIZE;
    args->classpath       = sprops->sysclasspath;
    return JNI_OK;
}

#include "java_util_Properties.h"

#define PUTPROP(props, key, val) \
    execute_java_dynamic_method(0, (HObject *)(props), \
	"put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;",\
	makeJavaString((key), strlen(key)),\
	makeJavaString((val), strlen(val)))

/*
 * Some C strings obtained from the OS (such as user name) are 
 * platform encoded.
 */
#define PUTPROP_ForPlatformCString(props, key, val) \
    execute_java_dynamic_method(0, (HObject *)(props), \
	"put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;", \
	makeJavaStringFromPlatformCString((key), strlen(key)), \
	makeJavaStringFromPlatformCString((val), strlen(val)))

#ifndef VENDOR /* Third party may overwrite this. */
#define VENDOR "Sun Microsystems Inc."
#endif

JNIEXPORT jobject JNICALL
JVM_InitProperties(JNIEnv *env, jobject p)
{
    struct Hjava_util_Properties *props =
		(Hjava_util_Properties *)DeRef(env, p);
    char *v;
    char buf[200];
    props_md_t *sprops = GetPropertiesMD();

    PUTPROP(props, "java.vm.specification.version", "1.0");
    PUTPROP(props, "java.vm.specification.name",
	    "Java Virtual Machine Specification");
    PUTPROP(props, "java.vm.specification.vendor", "Sun Microsystems Inc.");

    PUTPROP(props, "java.vm.version", RELEASE);
    PUTPROP(props, "java.vm.name", "Classic VM");
    PUTPROP(props, "java.vm.vendor", VENDOR);

    jio_snprintf(buf, sizeof(buf), 
		 "build %s, %s", FULL_VERSION, sysGetSysInfo()->name);
    PUTPROP(props, "java.vm.info", buf);

    PUTPROP_ForPlatformCString(props, "java.ext.dirs", sprops->ext_dirs);

    if (oldjava) {
        char tmp[1024];
	jio_snprintf(tmp, sizeof(tmp), "%s" PATH_SEPARATOR "%s",
		     sprops->dll_dir, sprops->library_path);
        PUTPROP_ForPlatformCString(props, "sun.boot.library.path", tmp);
        PUTPROP_ForPlatformCString(props, "java.library.path", "");
    } else {
        PUTPROP_ForPlatformCString(props, "sun.boot.library.path",
				   sprops->dll_dir);
	PUTPROP_ForPlatformCString(props, "java.library.path",
				   sprops->library_path);
    }

    PUTPROP_ForPlatformCString(props, "java.home", sprops->java_home);

    if (oldjava) {
	PUTPROP_ForPlatformCString(props, "java.class.path", init_sysclasspath);
    } else {
	PUTPROP_ForPlatformCString(props, "sun.boot.class.path", init_sysclasspath);
    }
    
    v = getenv("JAVA_COMPILER");
    if (v) {
        PUTPROP_ForPlatformCString(props, "java.compiler", v);
    } else {
        /* if a default jit is not defined before, no jit is turned on. */
#ifndef DEFAULT_JIT_NAME
#define DEFAULT_JIT_NAME "NONE"
#endif
        PUTPROP_ForPlatformCString(props, "java.compiler", DEFAULT_JIT_NAME);
    }
	
    /* user-defined properties */
    while (init_properties) {
        prop_t * prop = init_properties;
	init_properties = prop->next;

	PUTPROP_ForPlatformCString(props, prop->name, prop->value);
	sysFree(prop->name);
	sysFree(prop->value);
	sysFree(prop);
    }

    return MkRefLocal(env, props);
}

