/*
 * @(#)jcov.c	1.15 98/08/18
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * code for profiling on Java code
 * @author Leonid Arbouzov
 * leo@nbsp.nsk.su
 */


#ifdef JCOV

#include "oobj.h"
#include "interpreter.h"
#include "sys_api.h"
#include "tree.h"

#ifdef JCOV_DBG
#define JCOV_DEBUG(x) x;
#else
#define JCOV_DEBUG(x)
#endif

#define MAX_JCOV_LINE_WIDTH 	1024
enum {
	KLJ_NOTHING = 0,	/* */
	KLJ_EOF,		/* end of file */
	KLJ_ERROR,		/* read error  */
	KLJ_DATA,		/* line of data  */
	KLJ_COMMENT,		/* comments line */
	KLJ_MAGIC,		/* magic jcov line */
	KLJ_FIRST = KLJ_MAGIC,	/* first significant line */
	KLJ_CLASS,		/* class's line */
	KLJ_TIMESTAMP,		/* timestamp's line */
	KLJ_SRCFILE,		/* source file's line */
	KLJ_METHOD,		/* method's line */
	KLJ_FILTER,		/* filter's line */
	KLJ_FORMAT_DATA,	/* format data's line */
	KLJ_LAST		/* last */
} Kind_jcov_lines;

static struct {
	int kind;
	char * contents;
	} ArrKeyWords[] = {
		{KLJ_NOTHING, 	"	"},
		{KLJ_EOF, 		""},
		{KLJ_ERROR, 		""},
		{KLJ_DATA, 		""},
		{KLJ_COMMENT, 		""},
		{KLJ_MAGIC, 		"JCOV-DATA-FILE-VERSION: 2.0\n"},
		{KLJ_CLASS, 		"CLASS: "},
		{KLJ_TIMESTAMP, 	"TIMESTAMP: "},
		{KLJ_SRCFILE, 		"SRCFILE: "},
		{KLJ_METHOD, 		"METHOD: "},
		{KLJ_FILTER, 		"FILTER: "},
		{KLJ_FORMAT_DATA,	"DATA: "},
		{KLJ_LAST,		""}
	};
#define KEYWORDS(kind)	ArrKeyWords[kind].contents

#define	DATA_FORMAT_B_COMMENT	"#kind\tline\tposition\tcount"
#define DATA_FORMAT_B  		'B'
#define DATA_FORMAT_A  		'A'

void
coverage_native_method(struct methodblock *mb)
{
	if (mb->coverage_table && cov_file) {
		struct covtable *cv = mb->coverage_table;

		cv->count++;
	}
}

void
coverage_if(int flg, JavaFrame * frame, unsigned char *pc)
{
	if (frame->current_method && frame->current_method->coverage_table
	    && cov_file) {
		struct methodblock *mb = frame->current_method;
    		int 		*jcov_mode = (int*)(mb->coverage_table + mb->coverage_table_length);
		int 		*hash = jcov_mode + 1;
		int 		addr = pc - mb->code;
		struct covtable *cv;

		if ((*jcov_mode) & JCOV_MODE_METHODS)
			return;
		if (hash[addr] != 0) {
			cv = mb->coverage_table + hash[addr] - 2;
			if (flg)
				cv->count++;
			else
				(cv + 1)->count++;
		}
	}
}


void
coverage_lookupswitch(JavaFrame * frame, unsigned char *pc, int npairs, int nodef)
{
	if (frame->current_method && frame->current_method->coverage_table
	    && cov_file) {
		struct methodblock *mb = frame->current_method;
    		int 		*jcov_mode = (int*)(mb->coverage_table + mb->coverage_table_length);
		int 		*hash = jcov_mode + 1;
		int  		addr = pc - mb->code;

		if ((*jcov_mode) & JCOV_MODE_METHODS)
			return;
		if (hash[addr] != 0) {
			struct covtable *cv = mb->coverage_table + hash[addr] - 2;
			if (!nodef)
				(cv - npairs)->count++;
			else
				(cv + 1)->count++;
		}
	}
}

void
coverage_switch(JavaFrame * frame, unsigned char *pc, int key, int low, int high)
{
	if (frame->current_method && frame->current_method->coverage_table
	    && cov_file) {
		struct methodblock *mb = frame->current_method;
    		int 		*jcov_mode = (int*)(mb->coverage_table + mb->coverage_table_length);
		int 		*hash = jcov_mode + 1;
		int             addr = pc - mb->code;

		if ((*jcov_mode) & JCOV_MODE_METHODS)
			return;
		if (hash[addr] != 0) {
			struct covtable *cv = mb->coverage_table + hash[addr] - 2;
			if (key > high - low || key < 0) {
				(cv + 1)->count++;
			} else {
				if ((cv - high + low + key)->where_line) {
					(cv - high + low + key)->count++;
				} else {
					(cv + 1)->count++;
				}
			}
		}
	}
}

/* Return code */
#define JCOV_OK	 	 0	/* All is OK */
#define JCOV_REP_ERROR	 2	/* error has been reported */
#define JCOV_NOT_COVF	 3	/* file is not coverage file */

static int
DeleteTempCoverageFile(char *pathname)
{
	char            temp_pathname[MAXPATHLEN + 1];

	if (strlen(pathname) + 6 > MAXPATHLEN) {
		jio_fprintf(stderr, "jcov_exit: pathname too long! [%s]\n", pathname);
		return JCOV_REP_ERROR;
	}
	sprintf(temp_pathname, "%s.temp", pathname);
	if (remove(temp_pathname) != 0) {
		jio_fprintf(stderr,
			"jcov_exit: problem removing temp file [%s]: %s\n",
			temp_pathname, strerror(errno));
		return JCOV_REP_ERROR;
	}
	return JCOV_OK;
}

static int
OpenCoverageFile(char *pathname, FILE ** coverageFile)
{
	char            cookie[MAX_JCOV_LINE_WIDTH];


	if (access(pathname, F_OK ) == -1) 	/* file not exist */
		return JCOV_NOT_COVF;

	(*coverageFile) = fopen(pathname, "rb");
	if ((*coverageFile) == NULL) {
		jio_fprintf(stderr, "jcov_exit: Can't open a coverage file [%s]\n", pathname);
		return JCOV_REP_ERROR;
	}

	if ( (fgets(cookie, MAX_JCOV_LINE_WIDTH - 1, (*coverageFile)) != NULL)
	   && !(strcmp(cookie, KEYWORDS(KLJ_MAGIC))))
		return JCOV_OK;
	else {
		jio_fprintf(stderr, "jcov_warning: Not a coverage file [%s]\n", pathname);
		fclose(*coverageFile);
		return JCOV_NOT_COVF;
	}
}

static int
CreateTempCoverageFile(char *pathname, FILE ** tempFile)
{
	char            temp_pathname[MAXPATHLEN + 1];

	if (strlen(pathname) + 6 > MAXPATHLEN) {
		jio_fprintf(stderr, "jcov_exit: pathname too long! [%s]\n", pathname);
		return JCOV_REP_ERROR;
	}

	sprintf(temp_pathname, "%s.temp", pathname);
	*tempFile = fopen(temp_pathname, "w+b");
	if ((*tempFile) == NULL) {
	    jio_fprintf(stderr, "jcov_exit: problem opening temp file [%s]: %s\n",
			temp_pathname, strerror(errno));
	    return JCOV_REP_ERROR;
	}
	fputs(KEYWORDS(KLJ_MAGIC), *tempFile);
	return JCOV_OK;
}

static int
CreateNewCoverageFile(char *pathname, FILE ** coverageFile)
{
	(*coverageFile) = fopen(pathname, "wb");
	if ((*coverageFile) == NULL) {
		jio_fprintf(stderr, "jcov_exit: problem creating coverage file [%s]: %s\n",
			pathname, strerror(errno));
		return JCOV_REP_ERROR;
	}
	return JCOV_OK;
}


#define INTERRUPT_PROCESS \
  jio_fprintf(stderr, "jcov_exit: malformed coverage file (line %d). jcov aborted\n", currentLineNumber); \
  return_code = JCOV_REP_ERROR; \
  goto finishedProcess

#define SCAN_4(__format__,__slot1__,__slot2__,__slot3__,__slot4__) \
  if ( sscanf(currentLine, __format__, (__slot1__), (__slot2__), (__slot3__), (__slot4__)) != 4 ) \
	{ INTERRUPT_PROCESS; }

#define SEARCH_KEYWORD(__keyword__, __write__) \
	{ int rez = process_next_line(currentLine, __keyword__, __write__); \
	  switch (rez) { 	\
	  case __keyword__: 	\
		break;		\
	  case KLJ_EOF: 		\
		goto finishedCoverageFile; \
	  case KLJ_ERROR:		\
		INTERRUPT_PROCESS;\
	  default: 		\
		jio_fprintf(stdout, "jcov_exit: Wrong return code = %d\n", rez); \
	} }

static FILE    *coverageFile;		/* input coverage file */
static FILE    *tempFile;		/* temporary coverage file */
static int      currentLineNumber = 1;	/* number of next read line */
static int	writeError = 0;		/* sign of write error */

/**
 * Types and data structure for jcov filters
 */
typedef struct filter {
	struct filter * next;
	} filter_t, * filter_ptr;
static filter_ptr filter_head = NULL;

/**
 * Service function and macro to wrire with checking
 */
static void
fprintf_check(FILE * strm, const char *format,...)
{
	va_list         args;

	va_start(args, format);
	if (vfprintf(strm, format, args) < 0)
		writeError++;
	va_end(args);
}

#define fputs_check(string, stream) \
	if (fputs(string, stream) == EOF) writeError++

struct accessFlags {
	unsigned short flag;
	char * image;
	} arrayFlags [] = {
		{ACC_PUBLIC, 	"public"},
		{ACC_PRIVATE, 	"private"},
		{ACC_PROTECTED, "protected"},
		{ACC_STATIC, 	"static"},
		{ACC_FINAL, 	"final"},
		{ACC_NATIVE, 	"native"},
		{ACC_INTERFACE, "interface"},
		{ACC_ABSTRACT,  "abstract"},
};

static void print_access_modifiers(FILE * out, int access_mdf)
{
	int i;
	char * format = "%s";

	fprintf_check(out, " [");
	for (i = 0; i < (sizeof(arrayFlags) / sizeof(struct accessFlags)); i++) {
		if (access_mdf & arrayFlags[i].flag) {
			fprintf_check(out, format, arrayFlags[i].image);
			format = " %s";
		}
	}
	fprintf_check(out, "]");
}

static void print_methods_line(FILE * out, struct methodblock * mb)
{
	fprintf_check(out, "%s%s%s", KEYWORDS(KLJ_METHOD), fieldname(&mb->fb), fieldsig(&mb->fb));
	print_access_modifiers(out, mb->fb.access);
	fprintf_check(out, "\n");
}

static void print_class_line(FILE * out, ClassClass *entry)
{
	fprintf_check(out, "%s%s", KEYWORDS(KLJ_CLASS), cbName(entry));
	print_access_modifiers(out, cbAccess(entry));
	fprintf_check(out, "\n");
}

/******************************************************************
 *
 * Routine: OutputObjectSection
 *
 * Abstract:
 *
 *    Output a section for the given object. A section looks
 *  like:
 *
 *	CLASS: <name> [access-flags]
 *	SRCFILE: <absolute-path>
 * 	TIMESTAMP: <sec>
 * 	DATA B
 * 	#kind	line	position	count
 *	    ...
 *      METHOD: <name><signature> [access-flags]
 *	    ...
 *
 ******************************************************************/
static void
OutputObjectSection(FILE *out, ClassClass *entry, int write_new_class)
{
    struct methodblock *mb;
    struct covtable *cv;
    int header_printed = 0;
    int i, j;
    char stimestamp[32];

	for (mb = cbMethods(entry), j = cbMethodsCount(entry); --j >= 0; mb++) {
			cv = mb->coverage_table;
			if (cv == NULL)
				continue;
			if(!header_printed) {
				if(write_new_class) {
				    print_class_line(out, entry);
				    fprintf_check(out, "%s%s\n", KEYWORDS(KLJ_SRCFILE), (cbAbsoluteSourceName(entry) != NULL) ? cbAbsoluteSourceName(entry) : cbSourceName(entry));
				    int642CString(cbTimestamp(entry), stimestamp, sizeof(stimestamp));
				    fprintf_check(out, "%s%s\n", KEYWORDS(KLJ_TIMESTAMP), stimestamp);
				}
				fprintf_check(out, "%s%c\n", KEYWORDS(KLJ_FORMAT_DATA), DATA_FORMAT_B);
				fprintf_check(out, "%s\n", DATA_FORMAT_B_COMMENT);
				header_printed = 1;
			}

			for (i = 0; i < mb->coverage_table_length; cv++) {
				if( !(cv->type == CT_CASE && cv->where_line == 0) ) {
					if ((cv->type == CT_METHOD) || (cv->type == CT_FIKT_METHOD))
					    print_methods_line(out, mb);
					fprintf_check(out, "%d\t%ld\t%ld\t\t%ld\n",
					cv->type, cv->where_line, cv->where_pos, cv->count);
				}
				i++;
			}
	}
	if ((!header_printed) && (!write_new_class)) {
		fprintf_check(out, "%s%c\n", KEYWORDS(KLJ_FORMAT_DATA), DATA_FORMAT_B);
		fprintf_check(out, "%s\n", DATA_FORMAT_B_COMMENT);
	}
}

/**
 *	Add new filter in filter's list:
 *	filter_head = filter N -> filter N-1 -> ... -> fiter 1
 */
static void
add_new_filter(char *new_filter)
{
	int             size = strlen(new_filter) + sizeof(filter_t) + 1;
	filter_ptr      fptr;

	new_filter[strlen(new_filter) -1] = '\0';
	fptr = (filter_ptr) sysMalloc(size);
	if (fptr != (filter_ptr) NULL) {
		memset((char *) fptr, 0, size);

		strcpy((char *) (fptr + 1), new_filter);

		if (filter_head != (filter_ptr) NULL)
			fptr->next = filter_head;

		filter_head = fptr;
	}
}

static void
release_filters_memory(void)
{
	filter_ptr      next, fptr = filter_head;

	while (fptr != (filter_ptr) NULL) {
		next = fptr->next;
		sysFree(fptr);
		fptr = next;
	}
	filter_head = (filter_ptr) NULL;
}

static int
class_satisfy_filter(char *name)
{
	filter_ptr      fptr = filter_head;
	char           *filter;

	if (filter_head == (filter_ptr)NULL)
		return 1;
	while (fptr != (filter_ptr) NULL) {
		filter = (char *) (fptr + 1);
		if (strncmp(filter, name, strlen(filter)) == 0)
			return 1;
		fptr = fptr->next;
	}
	return 0;
}

/**
 * Reading and analyzis of next line from input covarage file
 */
static int
read_next_line(char *buff)
{
	int             i, rez = KLJ_ERROR;

	if (fgets(buff, MAX_JCOV_LINE_WIDTH - 1, coverageFile) != NULL) {
		currentLineNumber++;
		if (buff[0] == '#')
			rez =  KLJ_COMMENT;
		else {
			for (i = KLJ_FIRST; i < KLJ_LAST; i++) {
				if (strncmp(buff, KEYWORDS(i), strlen(KEYWORDS(i))) == 0) {
					rez = i;
					break;
				}
			}
			if (i == KLJ_LAST)
				rez =  KLJ_DATA;
		}
	} else {
		if (feof(coverageFile))
			rez =  KLJ_EOF;
		else
			rez =  KLJ_ERROR;
	}
	return rez;
}

/**
 * Preliminary processing of next line from input covarage file
 */
static int
process_next_line(char *buff, int keyword, int write)
{
	int             r;
	int		read = 1;

	while (read) {
		r = read_next_line(buff);
		switch (r) {
		case KLJ_FILTER:
			fputs_check(buff, tempFile);
			add_new_filter(buff + strlen(KEYWORDS(KLJ_FILTER)));
			break;
		case KLJ_COMMENT:
			if (write)
				fputs_check(buff, tempFile);
			break;
		case KLJ_EOF:
		case KLJ_ERROR:
			read = 0;
			break;
		default:
			read = 0;
			if ((keyword != KLJ_NOTHING) && (keyword != r))
				r =  KLJ_ERROR;
		}
	}
	return r;
}

/*
 *	Main function of Jcov
 */

int
jcov_write_data(void)
{
    char		currentLine[MAX_JCOV_LINE_WIDTH];
    char		pathname[MAX_JCOV_LINE_WIDTH];
    char    		*cl_done;	/* array of sign for each processing class */
    int			cur_class;
    ClassClass          *entry;
    char 		*objname;
    char                classsrcname[MAX_JCOV_LINE_WIDTH];
    char		stimestamp[32], classtimestamp[32];
    int 		rez;
    int 		write_to = 1;
    int			need_replace;
    int 		return_code = JCOV_OK;
    int			class_completed = 1;

    /*
     * The temporary coverage holds the new contents of the coverage file.
     */
    if ((rez = CreateTempCoverageFile(cov_file, &tempFile)) != JCOV_OK)
	return rez;

    cl_done = (char*)sysMalloc(sizeof(char) * (nbinclasses + 1));
    memset(cl_done, 0 , sizeof(char) * (nbinclasses + 1));

    switch (OpenCoverageFile(cov_file, &coverageFile)) {
        case JCOV_REP_ERROR:
		fclose(tempFile);
		DeleteTempCoverageFile(cov_file);
		return JCOV_REP_ERROR;
        case JCOV_NOT_COVF:
		goto finishedCoverageFile;
    }

    /*
     * This loop iterates through each object section in the
     * coverage file. The "newObjectSection" label is used
     * instead of implementing lookahead.
     */
    while ( 1 ) {
	char * ch;


	SEARCH_KEYWORD(KLJ_CLASS, write_to);

    newObjectSection:
	/*   We are supposed to be looking at the start of the next
	 * object section.
	 */

	write_to = 1;
	need_replace = 0;
	class_completed = 0;

	ch = strchr(currentLine + strlen(KEYWORDS(KLJ_CLASS)), ' ');
	if (ch == NULL) {
            strcpy(pathname, currentLine + strlen(KEYWORDS(KLJ_CLASS)));
            pathname[strlen(pathname) - 1] = '\0';
	}
	else {
	    int len = ch - currentLine - strlen(KEYWORDS(KLJ_CLASS));

            strncpy(pathname, currentLine + strlen(KEYWORDS(KLJ_CLASS)), len);
            pathname[len] = '\0';
	}

	if (!class_satisfy_filter(pathname))
		write_to = 0;
	else
		fputs_check(currentLine, tempFile);

	SEARCH_KEYWORD(KLJ_SRCFILE, write_to);
        strcpy(classsrcname, currentLine + strlen(KEYWORDS(KLJ_SRCFILE)));
	if (write_to)
		fputs_check(currentLine, tempFile);

	SEARCH_KEYWORD(KLJ_TIMESTAMP, write_to);
	currentLine[strlen(currentLine) - 1] = '\0';
	strcpy(stimestamp, currentLine + strlen(KEYWORDS(KLJ_TIMESTAMP)));
	if (write_to)
		fprintf_check(tempFile, "%s\n", currentLine);

	SEARCH_KEYWORD(KLJ_FORMAT_DATA, write_to);
	switch (currentLine[strlen(KEYWORDS(KLJ_FORMAT_DATA))]) {
	case DATA_FORMAT_A:
			need_replace = 1;
			break;
	case DATA_FORMAT_B:
			break;
	default:
		INTERRUPT_PROCESS;
	}
	class_completed = 1;

	entry = NULL;
	for (cur_class = 0; cur_class < nbinclasses; cur_class++) {
		if (cl_done[cur_class])
			continue;
		if (cbName(binclasses[cur_class])[0] == '[')
			continue;
		entry = binclasses[cur_class];
		if (entry == NULL)
			continue;
		objname = cbName(entry);
		if (!strcmp(pathname, objname)) {
			char           *absname = cbAbsoluteSourceName(entry);

			if (absname == NULL)
				absname = cbSourceName(entry);
			if ((absname != NULL) && !strncmp(classsrcname, absname, strlen(absname))) {
				int642CString(cbTimestamp(entry), classtimestamp, sizeof(classtimestamp));
				if (!strcmp(stimestamp, classtimestamp)) {
					break;		/* the same class */
				}
			}
		}
	}

	if (cur_class < nbinclasses && need_replace && write_to) {
		OutputObjectSection(tempFile, entry, 0);
		write_to = 0;	/* Not write this class one more in short form */
	}

	if (cur_class >= nbinclasses || !(write_to)) {

		/*
		 * 1. The current object in the coverage file wasn't accessed
		 *    in the current run so just copy the existing data over 2.
		 * 2. Class don't satisfy a some filter
		 * 3. Class has a short form and now was writted in full form
		 */

		if (cur_class < nbinclasses)
			cl_done[cur_class] = 1;
		if (write_to)
			fputs_check(currentLine, tempFile);
		while (1) {
			rez = process_next_line(currentLine, KLJ_NOTHING, write_to);
			switch (rez) {
			case KLJ_CLASS:
				goto newObjectSection;
			case KLJ_EOF:
				goto finishedCoverageFile;
			case KLJ_ERROR:
				INTERRUPT_PROCESS;
			default:
				if (write_to)
					fputs_check(currentLine, tempFile);
			}
		}

	} else {
		int             j, i;
		struct methodblock *mb;
		struct covtable *cv;
		int             type, ln, ps;
		int             count;
		int             rez, read;

		cl_done[cur_class] = 1;
		fprintf_check(tempFile, "%s%c\n", KEYWORDS(KLJ_FORMAT_DATA), DATA_FORMAT_B);

		/*
		 * Add counter values from file into object entry array.
		 */

		mb = cbMethods(entry);
		if (mb == NULL) {
			rez = process_next_line(currentLine, KLJ_NOTHING, 0);
			switch (rez) {
			case KLJ_CLASS:
				goto newObjectSection;
			case KLJ_EOF:
				goto finishedCoverageFile;
			default:
				INTERRUPT_PROCESS;
			}
		}
		for (mb = cbMethods(entry), j = cbMethodsCount(entry); --j >= 0; mb++) {
				cv = mb->coverage_table;
				if (cv == NULL)
					continue;

				for (i = 0; i < mb->coverage_table_length; cv++) {
					if (!(cv->type == CT_CASE && cv->where_line == 0)) {
						read = 1;
						while (read) {
							rez = process_next_line(currentLine, KLJ_NOTHING, 1);
							switch (rez) {
							default:
								INTERRUPT_PROCESS;
							case KLJ_METHOD:
								break;
							case KLJ_DATA:
								read = 0;
							}
						}
						SCAN_4("%d\t%d\t%d\t\t%d", &type, &ln, &ps, &count);
						if ((cv->type == CT_METHOD) || (cv->type == CT_FIKT_METHOD))
							print_methods_line(tempFile, mb);
						fprintf_check(tempFile, "%d\t%d\t%d\t\t%ld\n", type, ln, ps, cv->count + count);
					}
					i++;
				}
		}
		rez = process_next_line(currentLine, KLJ_NOTHING, 1);
		switch (rez) {
		case KLJ_CLASS:
			goto newObjectSection;
		case KLJ_EOF:
			goto finishedCoverageFile;
		default:
			INTERRUPT_PROCESS;
		}

	} /* else */

    } /* while */

finishedCoverageFile:
    if (! class_completed) {
	INTERRUPT_PROCESS;
    }

    { 	int i;
	for (i = 0; i < nbinclasses; i++) {
		char 	       *objname;
		ClassClass     *clb = binclasses[i];

		if (clb == NULL)
			continue;
		if (cbName(binclasses[i])[0] == '[')
			continue;
		objname = cbName(clb);
		if (!cl_done[i])
			if (class_satisfy_filter(objname))
				OutputObjectSection(tempFile, clb, 1);
	}
    }

    /*
     * Ok, we've finished building the new coverage file. Now,
     * erase the original coverage file and fill it with the
     * contents of the new file.
     */

    if (! writeError) {
    	if (CreateNewCoverageFile(cov_file, &coverageFile) != JCOV_OK)
		return JCOV_REP_ERROR;
    	rewind(tempFile);
    	while (fgets(currentLine, MAX_JCOV_LINE_WIDTH - 1, tempFile)) {
		fputs(currentLine, coverageFile);
    	}
    }
    else
	jio_fprintf(stderr, "jcov_exit: Write error. New jcov file not created\n");

finishedProcess:
    /* clean memory */
    release_filters_memory();
    fclose(coverageFile);
    fclose(tempFile);
    DeleteTempCoverageFile(cov_file);

    return return_code;
}

#else

/*
 * The following static is currently expected to be unused.  Its presence,
 * given that the rest of the file is empty, prevents an "empty translation
 * unit" warning from the C compiler.
 */
static unsigned char junk;

#endif /* JCOV */
