/*
 * @(#)jvmdi_jni.c	1.2 98/09/01
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* This file contains a version of JNI needed for JVMDI functionality.
 */
#ifdef BREAKPTS
#include "oobj.h"
#include "interpreter.h"
#include "tree.h"
#include "java_lang_String.h"
#include "breakpoints.h"

static struct JNINativeInterface_ jvmdi_jni_NativeInterface;

#define UNCHECKED() (&unchecked_jni_NativeInterface)

#define RECORD_STACK_BASE(env, base) \
    void *_stack_base = JNIEnv2EE(env)->stack_base; \
    int _x = (_stack_base == NULL) ? \
                ((JNIEnv2EE(env)->stack_base = (void *)base), 0) : 1

#define RECOVER_STACK_BASE(env) \
    JNIEnv2EE(env)->stack_base = _stack_base


#define WRAPPER_GETFIELD(ReturnType,Result,ty) \
static ReturnType JNICALL \
jvmdi_jni_Get##Result##Field(JNIEnv *env, jobject obj, jfieldID fieldID) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    ReturnType result; \
    if (watching_field_access) { \
        notify_debugger_of_field_access(JNIEnv2EE(env), \
                                        DeRef(env, obj), fb); \
    } \
    result = UNCHECKED()->Get##Result##Field(env,obj,fieldID); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_GETFIELD(jobject,Object,(fb->signature[0] == 'L' || fb->signature[0] == '['))
WRAPPER_GETFIELD(jboolean,Boolean,(fb->signature[0] == 'Z'))
WRAPPER_GETFIELD(jbyte,Byte,(fb->signature[0] == 'B'))
WRAPPER_GETFIELD(jshort,Short,(fb->signature[0] == 'S'))
WRAPPER_GETFIELD(jchar,Char,(fb->signature[0] == 'C'))
WRAPPER_GETFIELD(jint,Int,(fb->signature[0] == 'I'))
WRAPPER_GETFIELD(jlong,Long,(fb->signature[0] == 'J'))
WRAPPER_GETFIELD(jfloat,Float,(fb->signature[0] == 'F'))
WRAPPER_GETFIELD(jdouble,Double,(fb->signature[0] == 'D'))

#define WRAPPER_SETFIELD(ValueType,Result,JValueField) \
static void JNICALL \
jvmdi_jni_Set##Result##Field(JNIEnv *env, jobject obj, jfieldID fieldID, \
			  ValueType value) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    if (watching_field_modification) { \
        jvalue jval; \
        jval.JValueField = value; \
        notify_debugger_of_field_modification(JNIEnv2EE(env), \
                                              DeRef(env, obj), fb, jval); \
    } \
    UNCHECKED()->Set##Result##Field(env,obj,fieldID,value); \
    RECOVER_STACK_BASE(env); \
}

WRAPPER_SETFIELD(jobject,Object,l)
WRAPPER_SETFIELD(jboolean,Boolean,z)
WRAPPER_SETFIELD(jbyte,Byte,b)
WRAPPER_SETFIELD(jshort,Short,s)
WRAPPER_SETFIELD(jchar,Char,c)
WRAPPER_SETFIELD(jint,Int,i)
WRAPPER_SETFIELD(jlong,Long,j)
WRAPPER_SETFIELD(jfloat,Float,f)
WRAPPER_SETFIELD(jdouble,Double,d)

#define WRAPPER_GETSTATICFIELD(ReturnType,Result,ty) \
static ReturnType JNICALL \
jvmdi_jni_GetStatic##Result##Field(JNIEnv *env, jclass clazz, jfieldID fieldID) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    ReturnType result; \
    if (watching_field_access) { \
        notify_debugger_of_field_access(JNIEnv2EE(env), NULL, fb); \
    } \
    result = UNCHECKED()->GetStatic##Result##Field(env,clazz,fieldID); \
    RECOVER_STACK_BASE(env); \
    return result; \
}

WRAPPER_GETSTATICFIELD(jobject,Object,(fb->signature[0] == 'L' || fb->signature[0] == '['))
WRAPPER_GETSTATICFIELD(jboolean,Boolean,(fb->signature[0] == 'Z'))
WRAPPER_GETSTATICFIELD(jbyte,Byte,(fb->signature[0] == 'B'))
WRAPPER_GETSTATICFIELD(jshort,Short,(fb->signature[0] == 'S'))
WRAPPER_GETSTATICFIELD(jchar,Char,(fb->signature[0] == 'C'))
WRAPPER_GETSTATICFIELD(jint,Int,(fb->signature[0] == 'I'))
WRAPPER_GETSTATICFIELD(jlong,Long,(fb->signature[0] == 'J'))
WRAPPER_GETSTATICFIELD(jfloat,Float,(fb->signature[0] == 'F'))
WRAPPER_GETSTATICFIELD(jdouble,Double,(fb->signature[0] == 'D'))

#define WRAPPER_SETSTATICFIELD(ValueType,Result,JValueField) \
static void JNICALL \
jvmdi_jni_SetStatic##Result##Field(JNIEnv *env, jclass clazz, jfieldID fieldID, \
				ValueType value) \
{ \
    RECORD_STACK_BASE(env, &env); \
    struct fieldblock *fb = (struct fieldblock *)fieldID; \
    if (watching_field_modification) { \
        jvalue jval; \
        jval.JValueField = value; \
        notify_debugger_of_field_modification(JNIEnv2EE(env), NULL, \
                                              fb, jval); \
    } \
    UNCHECKED()->SetStatic##Result##Field(env,clazz,fieldID,value); \
    RECOVER_STACK_BASE(env); \
}

WRAPPER_SETSTATICFIELD(jobject,Object,l)
WRAPPER_SETSTATICFIELD(jboolean,Boolean,z)
WRAPPER_SETSTATICFIELD(jbyte,Byte,b)
WRAPPER_SETSTATICFIELD(jshort,Short,s)
WRAPPER_SETSTATICFIELD(jchar,Char,c)
WRAPPER_SETSTATICFIELD(jint,Int,i)
WRAPPER_SETSTATICFIELD(jlong,Long,j)
WRAPPER_SETSTATICFIELD(jfloat,Float,f)
WRAPPER_SETSTATICFIELD(jdouble,Double,d)

struct JNINativeInterface_ *
jvmdi_jni_GetNativeInterface() 
{
    /* Copy regular JNI */
    memcpy(&jvmdi_jni_NativeInterface, 
           &unchecked_jni_NativeInterface,
           sizeof(struct JNINativeInterface_));

    /* Override the functions we need to moniotr */
    jvmdi_jni_NativeInterface.GetObjectField = &jvmdi_jni_GetObjectField;
    jvmdi_jni_NativeInterface.GetBooleanField = &jvmdi_jni_GetBooleanField;
    jvmdi_jni_NativeInterface.GetByteField = &jvmdi_jni_GetByteField;
    jvmdi_jni_NativeInterface.GetCharField = &jvmdi_jni_GetCharField;
    jvmdi_jni_NativeInterface.GetShortField = &jvmdi_jni_GetShortField;
    jvmdi_jni_NativeInterface.GetIntField = &jvmdi_jni_GetIntField;
    jvmdi_jni_NativeInterface.GetLongField = &jvmdi_jni_GetLongField;
    jvmdi_jni_NativeInterface.GetFloatField = &jvmdi_jni_GetFloatField;
    jvmdi_jni_NativeInterface.GetDoubleField = &jvmdi_jni_GetDoubleField;

    jvmdi_jni_NativeInterface.SetObjectField = &jvmdi_jni_SetObjectField;
    jvmdi_jni_NativeInterface.SetBooleanField = &jvmdi_jni_SetBooleanField;
    jvmdi_jni_NativeInterface.SetByteField = &jvmdi_jni_SetByteField;
    jvmdi_jni_NativeInterface.SetCharField = &jvmdi_jni_SetCharField;
    jvmdi_jni_NativeInterface.SetShortField = &jvmdi_jni_SetShortField;
    jvmdi_jni_NativeInterface.SetIntField = &jvmdi_jni_SetIntField;
    jvmdi_jni_NativeInterface.SetLongField = &jvmdi_jni_SetLongField;
    jvmdi_jni_NativeInterface.SetFloatField = &jvmdi_jni_SetFloatField;
    jvmdi_jni_NativeInterface.SetDoubleField = &jvmdi_jni_SetDoubleField;

    jvmdi_jni_NativeInterface.GetStaticObjectField = &jvmdi_jni_GetStaticObjectField;
    jvmdi_jni_NativeInterface.GetStaticBooleanField = &jvmdi_jni_GetStaticBooleanField;
    jvmdi_jni_NativeInterface.GetStaticByteField = &jvmdi_jni_GetStaticByteField;
    jvmdi_jni_NativeInterface.GetStaticCharField = &jvmdi_jni_GetStaticCharField;
    jvmdi_jni_NativeInterface.GetStaticShortField = &jvmdi_jni_GetStaticShortField;
    jvmdi_jni_NativeInterface.GetStaticIntField = &jvmdi_jni_GetStaticIntField;
    jvmdi_jni_NativeInterface.GetStaticLongField = &jvmdi_jni_GetStaticLongField;
    jvmdi_jni_NativeInterface.GetStaticFloatField = &jvmdi_jni_GetStaticFloatField;
    jvmdi_jni_NativeInterface.GetStaticDoubleField = &jvmdi_jni_GetStaticDoubleField;

    jvmdi_jni_NativeInterface.SetStaticObjectField = &jvmdi_jni_SetStaticObjectField;
    jvmdi_jni_NativeInterface.SetStaticBooleanField = &jvmdi_jni_SetStaticBooleanField;
    jvmdi_jni_NativeInterface.SetStaticByteField = &jvmdi_jni_SetStaticByteField;
    jvmdi_jni_NativeInterface.SetStaticCharField = &jvmdi_jni_SetStaticCharField;
    jvmdi_jni_NativeInterface.SetStaticShortField = &jvmdi_jni_SetStaticShortField;
    jvmdi_jni_NativeInterface.SetStaticIntField = &jvmdi_jni_SetStaticIntField;
    jvmdi_jni_NativeInterface.SetStaticLongField = &jvmdi_jni_SetStaticLongField;
    jvmdi_jni_NativeInterface.SetStaticFloatField = &jvmdi_jni_SetStaticFloatField;
    jvmdi_jni_NativeInterface.SetStaticDoubleField = &jvmdi_jni_SetStaticDoubleField;

    return &jvmdi_jni_NativeInterface;
}

#else
int dummy_jvmdi_jni_var = 0; /* shut off warnings */
#endif /* BREAKPTS */
