/*
 * @(#)string_util.c	1.52 98/09/02
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "javaString.h"
#include "utf.h"
#include "typecodes.h"
#include "sys_api.h"
#include "common_exceptions.h"

Hjava_lang_String *
makeJavaString(char *str, int len) 
{
    HArrayOfChar *val;
    unicode *p;
    ExecEnv *ee = EE();

    if (!str) len = 0;
    val = (HArrayOfChar *) allocArray(ee, T_CHAR, len);
    if (val == 0) {
	ThrowOutOfMemoryError(0, 0);
	return 0;
    }
    p = unhand(val)->body;
    if (str) str2unicode(str, p, len);
    KEEP_POINTER_ALIVE(p);
    return (Hjava_lang_String *)
      execute_java_constructor(ee, 0, classJavaLangString, "([C)", val);
}

int javaStringLength(Hjava_lang_String *s) 
{
    Classjava_lang_String *str;
    if (!s) {
	return 0;
    }
    str = unhand(s);
    return str->count;
}

char *
javaString2CString(Hjava_lang_String *s, char *buf, int buflen) 
{
    Classjava_lang_String *str;
    HArrayOfChar *hac;

    if (s && (str = unhand(s)) && (hac = (HArrayOfChar *)str->value)) {
	unicode *src = unhand(hac)->body;
	int	len = javaStringLength(s);

	if (len >= buflen)
	    len = buflen - 1;
	unicode2str(src + str->offset, buf, len);
	KEEP_POINTER_ALIVE(src);
    } else if (buflen) {
	buf[0] = '\0';
    }
    return buf;
}

/*
    Make a Java array of bytes
*/
static HArrayOfByte *
MakeByteString(char *str, long len)
{
    int8_t *new_str_body;
    JHandle *ret;

    ret = allocArray(EE(), T_BYTE, len);
    if (ret == 0)
	return 0;
    new_str_body = ((ArrayOfByte *)unhand(ret))->body;
    if (str)
	memmove(new_str_body, (char *)str, len);
    else
	memset(new_str_body, 0, len);
    KEEP_POINTER_ALIVE(new_str_body);

    return (HArrayOfByte *)ret;
}

char *
makeCString(Hjava_lang_String *s) 
{
    int len = javaStringLength(s);
    char *buf;

    HArrayOfByte *hab = MakeByteString(NULL, len + 1);
    if (hab == 0) {
	ThrowOutOfMemoryError(0, 0);
	return 0;
    }

    buf = (char *)unhand(hab)->body;
    return javaString2CString(s, buf, len + 1);
}

char *
allocCString(Hjava_lang_String *s) 
{
    int len = javaStringLength(s);
    char *buf = sysMalloc(len + 1);
    if (buf == 0) {
	ThrowOutOfMemoryError(0, 0);
	return 0;
    }
    return javaString2CString(s, buf, len + 1);
}

Hjava_lang_String *
makeJavaStringFromPlatformCString(char *str, int len) 
{
    static to_java_string_fn_t jfc = 0;
    ExecEnv *ee = EE();
    JNIEnv *env = EE2JNIEnv(ee);
    jstring result;

    if (jfc == 0) {
        jfc = (to_java_string_fn_t)FindBuiltinEntry("NewStringPlatform");
        if (jfc == 0) {
            ThrowInternalError(0, "Can't find C to Java string converter");
            return NULL;
        }
    }

    result = (*jfc)(env, str);
    return (Hjava_lang_String *)DeRef(env, result);
}

char *
makePlatformCString(Hjava_lang_String *s)
{
    static to_c_string_fn_t cfj = 0;
    ExecEnv *ee = EE();
    JNIEnv *env = EE2JNIEnv(ee);

    if (cfj == 0) {
        cfj = (to_c_string_fn_t)FindBuiltinEntry("GetStringPlatformChars");
        if (cfj == 0) {
            ThrowInternalError(0, "Can't find Java to C string converter");
            return NULL;
        }
    }

    return (*cfj)(env, (jstring)&s, 0);
}

/*
 * Fill buf with unicode representation of hstr, upto buflen chars.
 * The UTF string will be NUL-terminated.
 *
 * If both buf and buflen are 0, malloc an appropriately sized
 * buffer for the result.
 *
 * Return result buffer.
 */
char *
javaString2UTF(HString *s, char *buf, int buflen)
{
    Classjava_lang_String *str;
    HArrayOfChar *hac;

    if (s && (str = unhand(s)) && (hac = (HArrayOfChar *)str->value)) {
	unicode *body = unhand(hac)->body;
	int unilength = str->count;
	char *result = unicode2utf(body + str->offset, unilength, buf, buflen);
	KEEP_POINTER_ALIVE(body);
	return result;
    } else if (buflen) {
	buf[0] = '\0';
    }
    return buf;
}

/*
 * Create a Java String whose characters are initialized from
 * the supplied NUL-terminated UTF string.
 */
HString *
makeJavaStringUTF(char *str)
{
    HArrayOfChar *val;
    unicode *unistr;
    int unilength = utfstrlen(str);
    ExecEnv *ee = EE();

    val = (HArrayOfChar *) allocArray(ee, T_CHAR, unilength);
    if (val == 0) {
        ThrowOutOfMemoryError(0, 0);
        return 0;
    }
    unistr = unhand(val)->body;
    utf2unicode(str, unistr, unilength, &unilength);
    KEEP_POINTER_ALIVE(unistr);
    return (Hjava_lang_String *)
	execute_java_constructor(ee, 0, classJavaLangString, "([C)", val);
}
