/*
 * @(#)wrap_util.c	1.27 98/08/14
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Support for wrapping and unwrapping primitive Java types
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "sys_api.h"		/* sysAssert */
#include "wrap.h"
#include "common_exceptions.h"

#include "java_lang_Boolean.h"
#include "java_lang_Character.h"
#include "java_lang_Byte.h"
#include "java_lang_Short.h"
#include "java_lang_Integer.h"
#include "java_lang_Long.h"
#include "java_lang_Float.h"
#include "java_lang_Double.h"

/*
 * All wrappers must have a member named 'value'
 */
#define	wrapper_value(h)	(unhand(h)->value)

/*
 * Standard system wrapper classes.
 * N.B. init_wrappers() installs typecodes.
 */

static ClassClass *
classJavaLangVoid(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Void", TRUE));
}

static ClassClass *
classJavaLangBoolean(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Boolean", TRUE));
}

static ClassClass *
classJavaLangCharacter(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Character", TRUE));
}

static ClassClass *
classJavaLangByte(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Byte", TRUE));
}

static ClassClass *
classJavaLangShort(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Short", TRUE));
}

static ClassClass *
classJavaLangInteger(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Integer", TRUE));
}

static ClassClass *
classJavaLangLong(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Long", TRUE));
}

static ClassClass *
classJavaLangFloat(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Float", TRUE));
}

static ClassClass *
classJavaLangDouble(void)
{
    static ClassClass *cb = NULL;
    if (cb) return (cb);
    return (cb = FindStickySystemClass(NULL, JAVAPKG "Double", TRUE));
}

static struct {
    ClassClass *	(*wrapper)(void);
    unsigned char	typecode;
} const wrappers[] = {
    { &classJavaLangVoid,	T_VOID },
    { &classJavaLangBoolean,	T_BOOLEAN },
    { &classJavaLangByte,	T_BYTE },
    { &classJavaLangCharacter,	T_CHAR },
    { &classJavaLangShort,	T_SHORT },
    { &classJavaLangInteger,	T_INT },
    { &classJavaLangLong,	T_LONG },
    { &classJavaLangFloat,	T_FLOAT },
    { &classJavaLangDouble,	T_DOUBLE }
};
static int init_wrappers_done = 0;

static int
init_wrappers(void)
{
    int i;
    ClassClass *cb;

    for (i = 0; i < sizeof (wrappers) / sizeof (wrappers[0]); i++) {
	cb = (*wrappers[i].wrapper)();
	cbTypeCode(cb) = wrappers[i].typecode;
    }
    return (init_wrappers_done = 1);
}

#define	INIT_WRAPPERS()	if (!init_wrappers_done) init_wrappers()

/*
 * Wrapping conversion.  Returns a new Java wrapper class instance.
 * May return NULL and signal an OutOfMemoryError.
 */

#define	WRAP(ee, nm, v)						\
if (TRUE)							\
{								\
    HObject *o;							\
    if ((o = allocObject(ee, classJavaLang##nm())) == NULL) {	\
	ThrowOutOfMemoryError(0, 0);				\
	return NULL;						\
    }								\
    wrapper_value((Hjava_lang_##nm *)o) = v;			\
    return o;							\
} else ((void) 0)

#define	WRAP2(ee, nm, set, v)					\
if (TRUE)							\
{								\
    HObject *o;							\
    jvalue tmp;							\
    if ((o = allocObject(ee, classJavaLang##nm())) == NULL) {	\
	ThrowOutOfMemoryError(0, 0);				\
	return NULL;						\
    }								\
    SET_##set(tmp, &wrapper_value((Hjava_lang_##nm *)o), v);	\
    return o;							\
} else ((void) 0)

HObject *
java_wrap(
    ExecEnv *ee,
    jvalue v,
    unsigned char code)
{
    INIT_WRAPPERS();

    switch (code) {
    case T_BOOLEAN:
        WRAP(ee, Boolean, v.z);
    case T_BYTE:
        WRAP(ee, Byte, v.b);
    case T_CHAR:
        WRAP(ee, Character, v.c);
    case T_SHORT:
        WRAP(ee, Short, v.s);
    case T_INT:
        WRAP(ee, Integer, v.i);
    case T_LONG:
        WRAP2(ee, Long, INT64, v.j);
    case T_FLOAT:
        WRAP(ee, Float, v.f);
    case T_DOUBLE:
        WRAP2(ee, Double, DOUBLE, v.d);
    case T_VOID:
        return NULL;
    default:
        return NULL;	/* Error */
    }
}

/*
 * Unwrapping conversion.  Returns typecode for primitive type and
 * side-effects vp.
 * May return T_BAD and signal an IllegalArgumentExcepion.
 */
unsigned char
java_unwrap(
    HObject *o,
    jvalue *vp)		/* doubleword aligned on machines that care */
{
    ClassClass	*ocb;
    jvalue tmp;
    unsigned char code;

    /* Wrapped objects cannot be null */
    if (o == NULL) {
	ThrowNullPointerException(0, 0);
	return T_BAD;
    }

    INIT_WRAPPERS();

    ocb = obj_array_classblock(o);
    sysAssert(!cbIsPrimitive(ocb));

    code = cbTypeCode(ocb);

    switch (code) {
    case T_BOOLEAN:
	vp->z = (jboolean) wrapper_value((Hjava_lang_Boolean*)o);
	break;
    case T_BYTE:
	vp->b = (jbyte) wrapper_value((Hjava_lang_Byte*)o);
	break;
    case T_CHAR:
	vp->c = (jchar) wrapper_value((Hjava_lang_Character*)o);
	break;
    case T_SHORT:
	vp->s = (jshort) wrapper_value((Hjava_lang_Short*)o);
	break;
    case T_INT:
	vp->i = wrapper_value((Hjava_lang_Integer*)o);
	break;
    case T_LONG:
	vp->j = GET_INT64(tmp, &wrapper_value((Hjava_lang_Long*)o));
	break;
    case T_FLOAT:
	vp->f = wrapper_value((Hjava_lang_Float*)o);
	break;
    case T_DOUBLE:
	vp->d = GET_DOUBLE(tmp, &wrapper_value((Hjava_lang_Double*)o));
	break;
    default:
	ThrowIllegalArgumentException(0, "unwrapping conversion failed");
	return T_BAD;
    }

    return code;
}
