/*
 * @(#)ObjectStreamClass.c	1.28 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "jni.h"
#include "jvm.h"

#include "java_io_ObjectStreamClass.h"

/*
 * Class:     java_io_ObjectStreamClass
 * Method:    hasStaticInitializer
 * Signature: (Ljava/lang/Class;)Z
 *
 * If the method <clinit> ()V is defined true is returned.
 * Otherwise, false is returned.
 */
JNIEXPORT jboolean JNICALL
Java_java_io_ObjectStreamClass_hasStaticInitializer(JNIEnv *env, jclass this,
						    jclass clazz)
{
    jclass superclazz = NULL;
    jmethodID superclinit = NULL;

    jmethodID clinit = (*env)->GetStaticMethodID(env, clazz,
						 "<clinit>", "()V");
    if (clinit == NULL || (*env)->ExceptionOccurred(env)) {
	(*env)->ExceptionClear(env);
	return 0;
    }

    /* Ask the superclass the same question
     * If the answer is the same then the constructor is from a superclass.
     * If different, it's really defined on the subclass.
     */
    superclazz = (*env)->GetSuperclass(env, clazz);
    if ((*env)->ExceptionOccurred(env)) {
	return 0;
    }

    if (superclazz == NULL)
	return 1;

    superclinit = (*env)->GetStaticMethodID(env, superclazz,
					    "<clinit>", "()V");
    if ((*env)->ExceptionOccurred(env)) {
	(*env)->ExceptionClear(env);
	superclinit = NULL;
    }

    return (superclinit != clinit);
}
