/*
 * @(#)RandomAccessFile.c	1.18 98/09/02
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "jni.h"
#include "jni_util.h"
#include "jlong.h"
#include "jvm.h"

#include "io_util.h"
#include "java_io_RandomAccessFile.h"

#include <fcntl.h>


/*
 * static method to store field ID's in initializers
 */

jfieldID raf_fd; /* id for jobject 'fd' in java.io.RandomAccessFile */

JNIEXPORT void JNICALL 
Java_java_io_RandomAccessFile_initIDs(JNIEnv *env, jclass fdClass) {
    raf_fd = (*env)->GetFieldID(env, fdClass, "fd", "Ljava/io/FileDescriptor;");
}


JNIEXPORT void JNICALL
Java_java_io_RandomAccessFile_open(JNIEnv *env,
	jobject this, jstring path, jboolean w) {
    int flags;
    if (w) {
	flags = O_RDWR|O_CREAT;
    } else {
	flags = O_RDONLY;
    }
    fileOpen(env, this, path, raf_fd, flags);
}

JNIEXPORT jint JNICALL
Java_java_io_RandomAccessFile_read(JNIEnv *env, jobject this) {
    return readSingle(env, this, raf_fd);
}

JNIEXPORT jint JNICALL
Java_java_io_RandomAccessFile_readBytes(JNIEnv *env,
    jobject this, jbyteArray bytes, jint off, jint len) {
    return readBytes(env, this, bytes, off, len, raf_fd); 
}

JNIEXPORT void JNICALL
Java_java_io_RandomAccessFile_write(JNIEnv *env, jobject this, jint byte) {
    writeSingle(env, this, byte, raf_fd);
}

JNIEXPORT void JNICALL
Java_java_io_RandomAccessFile_writeBytes(JNIEnv *env,
    jobject this, jbyteArray bytes, jint off, jint len) {
    writeBytes(env, this, bytes, off, len, raf_fd); 
}

JNIEXPORT jlong JNICALL
Java_java_io_RandomAccessFile_getFilePointer(JNIEnv *env, jobject this) {
    int fd;
    jlong ret;

    fd = GET_FD(raf_fd);
    if ((ret = JVM_Lseek(fd, 0L, SEEK_CUR)) == -1) {
      	JNU_ThrowIOExceptionWithLastError(env, "Seek failed");
    }
    return ret;
}

JNIEXPORT jlong JNICALL
Java_java_io_RandomAccessFile_length(JNIEnv *env, jobject this) {
    int fd;
    jlong cur = jlong_zero;
    jlong end = jlong_zero;

    fd = GET_FD(raf_fd);

    if ((cur = JVM_Lseek(fd, 0L, SEEK_CUR)) == -1) {
      	JNU_ThrowIOExceptionWithLastError(env, "Seek failed");
    } else if ((end = JVM_Lseek(fd, 0L, SEEK_END)) == -1) {
      	JNU_ThrowIOExceptionWithLastError(env, "Seek failed");
    } else if (JVM_Lseek(fd, cur, SEEK_SET) == -1) {
      	JNU_ThrowIOExceptionWithLastError(env, "Seek failed");
    }
    return end;
}

JNIEXPORT void JNICALL
Java_java_io_RandomAccessFile_seek(JNIEnv *env,
		    jobject this, jlong pos) {

    int fd;

    fd = GET_FD(raf_fd);
    if (pos < jlong_zero) {
        JNU_ThrowIOException(env, "Negative seek offset");
    } else if (JVM_Lseek(fd, pos, SEEK_SET) == -1) {
      	JNU_ThrowIOExceptionWithLastError(env, "Seek failed");
    }
}

JNIEXPORT void JNICALL
Java_java_io_RandomAccessFile_setLength(JNIEnv *env, jobject this,
					jlong newLength)
{
    int fd;
    jlong cur;

    fd = GET_FD(raf_fd);
    if ((cur = JVM_Lseek(fd, 0L, SEEK_CUR)) == -1) goto fail;
    if (JVM_SetLength(fd, newLength) == -1) goto fail;
    if (cur > newLength) {
	if (JVM_Lseek(fd, 0L, SEEK_END) == -1) goto fail;
    } else {
	if (JVM_Lseek(fd, cur, SEEK_SET) == -1) goto fail;
    }
    return;

 fail:
    JNU_ThrowIOExceptionWithLastError(env, "setLength failed");
}


JNIEXPORT void JNICALL
Java_java_io_RandomAccessFile_close(JNIEnv *env, jobject this) {
    int fd = GET_FD(raf_fd);
    SET_FD(-1, raf_fd);
    JVM_Close(fd);
}
