/*
 * @(#)timezone_map.h	1.5 98/09/16
 *
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road,
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 *
 * CopyrightVersion 1.2
 *
 */

enum EHasDST { kAny, kDaylight, kStandard };

/* OS- and language-independent time zone lookup table.  This is used to
 * determine the host time zone and set the user.timezone property.  This table
 * relies only on the raw offset, presence or absence of DST, and the country
 * code to determine the time zone.  Because it uses the country code, it can
 * differentiate between some zones that are otherwise identical, such as
 * Asia/Tokyo ("JST") and Asia/Seoul ("KST").
 *
 * See also <platform>/native/java/lang/java_props_md.c
 * alanl 5/8/98 */
static struct {
    int          offset;
    const char*  country;
    enum EHasDST hasDST;
    char*        zone;
} timezone_map[] = {
    /* Offset - the offset, in minutes, from GMT, in the standard Java
     * convention with zones to the east of Greenwich having positive
     * offsets.  Note that this is the reverse of the Win32
     * convention.  Entries are listed in order of offset.
     *
     * Country - the two-letter ISO 3166 country code.
     *
     * Daylight - indicates whether this is a Standard-only zone or
     * whether an alternate Daylight zone may exist. The alternate
     * zone may be historical! A setting of kAny indicates that we
     * don't care.
     *
     * Zone - the name of the Java zone to be assigned for this
     * offset, DST flag, and country. */

    /* Off. Ctry  Daylight    Zone */
    { -660, "NU", kAny,      "Pacific/Niue"              },
    { -660, "US", kAny,      "Pacific/Pago_Pago"         },
    { -660, "WS", kAny,      "Pacific/Apia"              },
    { -660, NULL, kAny,      "Pacific/Pago_Pago"         },
    { -600, "CK", kAny,      "Pacific/Rarotonga"         },
    { -600, "PF", kAny,      "Pacific/Tahiti"            },
    { -600, "TK", kAny,      "Pacific/Fakaofo"           },
    { -600, "US", kStandard, "Pacific/Honolulu"          },
    { -600, "US", kDaylight, "America/Adak"              },
    { -600, NULL, kStandard, "Pacific/Honolulu"          },
    { -600, NULL, kDaylight, "Pacific/Rarotonga"         },
    { -570, "PF", kAny,      "Pacific/Marquesas"         },
    { -570, NULL, kAny,      "Pacific/Marquesas"         },
    { -540, "PF", kAny,      "Pacific/Gambier"           },
    { -540, "US", kAny,      "America/Anchorage"         },
    { -540, NULL, kStandard, "Pacific/Gambier"           },
    { -540, NULL, kDaylight, "America/Anchorage"         },
    { -510, "PN", kAny,      "Pacific/Pitcairn"          },
    { -510, NULL, kAny,      "Pacific/Pitcairn"          },
    { -480, "CA", kAny,      "America/Vancouver"         },
    { -480, "MX", kAny,      "America/Tijuana"           },
    { -480, "US", kAny,      "America/Los_Angeles"       },
    { -480, NULL, kAny,      "America/Los_Angeles"       },
    { -420, "CA", kStandard, "America/Dawson_Creek"      },
    { -420, "CA", kDaylight, "America/Edmonton"          },
    { -420, "MX", kAny,      "America/Mazatlan"          },
    { -420, "US", kStandard, "America/Phoenix"           },
    { -420, "US", kDaylight, "America/Denver"            },
    { -420, NULL, kStandard, "America/Phoenix"           },
    { -420, NULL, kDaylight, "America/Denver"            },
    { -360, "BZ", kAny,      "America/Belize"            },
    { -360, "CA", kStandard, "America/Regina"            },
    { -360, "CA", kDaylight, "America/Winnipeg"          },
    { -360, "CL", kAny,      "Pacific/Easter"            },
    { -360, "CR", kAny,      "America/Costa_Rica"        },
    { -360, "EC", kAny,      "Pacific/Galapagos"         },
    { -360, "GT", kAny,      "America/Guatemala"         },
    { -360, "HN", kAny,      "America/Tegucigalpa"       },
    { -360, "MX", kAny,      "America/Mexico_City"       },
    { -360, "SV", kAny,      "America/El_Salvador"       },
    { -360, "US", kAny,      "America/Chicago"           },
    { -360, NULL, kStandard, "America/Costa_Rica"        },
    { -360, NULL, kDaylight, "America/Chicago"           },
    { -300, "BR", kAny,      "America/Porto_Acre"        },
    { -300, "BS", kAny,      "America/Nassau"            },
    { -300, "CA", kAny,      "America/Montreal"          },
    { -300, "CO", kAny,      "America/Bogota"            },
    { -300, "CU", kAny,      "America/Havana"            },
    { -300, "EC", kAny,      "America/Guayaquil"         },
    { -300, "HT", kAny,      "America/Port-au-Prince"    },
    { -300, "JM", kAny,      "America/Jamaica"           },
    { -300, "KY", kAny,      "America/Cayman"            },
    { -300, "NI", kAny,      "America/Managua"           },
    { -300, "PA", kAny,      "America/Panama"            },
    { -300, "PE", kAny,      "America/Lima"              },
    { -300, "TC", kAny,      "America/Grand_Turk"        },
    { -300, "US", kStandard, "America/Indianapolis"      },
    { -300, "US", kDaylight, "America/New_York"          },
    { -300, NULL, kStandard, "America/Indianapolis"      },
    { -300, NULL, kDaylight, "America/New_York"          },
    { -240, "AG", kAny,      "America/Antigua"           },
    { -240, "AI", kAny,      "America/Anguilla"          },
    { -240, "AN", kAny,      "America/Curacao"           },
    { -240, "AQ", kAny,      "Antarctica/Palmer"         },
    { -240, "AW", kAny,      "America/Aruba"             },
    { -240, "BB", kAny,      "America/Barbados"          },
    { -240, "BM", kAny,      "Atlantic/Bermuda"          },
    { -240, "BO", kAny,      "America/La_Paz"            },
    { -240, "BR", kStandard, "America/Manaus"            },
    { -240, "BR", kDaylight, "America/Cuiaba"            },
    { -240, "CA", kAny,      "America/Halifax"           },
    { -240, "CL", kAny,      "America/Santiago"          },
    { -240, "DM", kAny,      "America/Dominica"          },
    { -240, "DO", kAny,      "America/Santo_Domingo"     },
    { -240, "FK", kAny,      "Atlantic/Stanley"          },
    { -240, "GD", kAny,      "America/Grenada"           },
    { -240, "GL", kAny,      "America/Thule"             },
    { -240, "GP", kAny,      "America/Guadeloupe"        },
    { -240, "GY", kAny,      "America/Guyana"            },
    { -240, "KN", kAny,      "America/St_Kitts"          },
    { -240, "LC", kAny,      "America/St_Lucia"          },
    { -240, "MQ", kAny,      "America/Martinique"        },
    { -240, "MS", kAny,      "America/Montserrat"        },
    { -240, "PR", kAny,      "America/Puerto_Rico"       },
    { -240, "PY", kAny,      "America/Asuncion"          },
    { -240, "TT", kAny,      "America/Port_of_Spain"     },
    { -240, "VC", kAny,      "America/St_Vincent"        },
    { -240, "VE", kAny,      "America/Caracas"           },
    { -240, "VG", kAny,      "America/Tortola"           },
    { -240, "VI", kAny,      "America/St_Thomas"         },
    { -240, NULL, kStandard, "America/Caracas"           },
    { -240, NULL, kDaylight, "America/Santiago"          },
    { -210, "CA", kAny,      "America/St_Johns"          },
    { -210, NULL, kAny,      "America/St_Johns"          },
    { -180, "AR", kAny,      "America/Buenos_Aires"      },
    { -180, "BR", kStandard, "America/Fortaleza"         },
    { -180, "BR", kDaylight, "America/Sao_Paulo"         },
    { -180, "GF", kAny,      "America/Cayenne"           },
    { -180, "GL", kAny,      "America/Godthab"           },
    { -180, "PM", kAny,      "America/Miquelon"          },
    { -180, "SR", kAny,      "America/Paramaribo"        },
    { -180, "UY", kAny,      "America/Montevideo"        },
    { -180, NULL, kStandard, "America/Buenos_Aires"      },
    { -180, NULL, kDaylight, "America/Sao_Paulo"         },
    { -120, "BR", kAny,      "America/Noronha"           },
    { -120, "GS", kAny,      "Atlantic/South_Georgia"    },
    { -120, NULL, kAny,      "Atlantic/South_Georgia"    },
    {  -60, "CV", kAny,      "Atlantic/Cape_Verde"       },
    {  -60, "GL", kAny,      "America/Scoresbysund"      },
    {  -60, "NO", kAny,      "Atlantic/Jan_Mayen"        },
    {  -60, "PT", kAny,      "Atlantic/Azores"           },
    {  -60, NULL, kStandard, "Atlantic/Cape_Verde"       },
    {  -60, NULL, kDaylight, "Atlantic/Azores"           },
    {    0, "BF", kAny,      "Africa/Ouagadougou"        },
    {    0, "CI", kAny,      "Africa/Abidjan"            },
    {    0, "DK", kAny,      "Atlantic/Faeroe"           },
    {    0, "ES", kAny,      "Atlantic/Canary"           },
    {    0, "GB", kAny,      "Europe/London"             },
    {    0, "GH", kAny,      "Africa/Accra"              },
    {    0, "GM", kAny,      "Africa/Banjul"             },
    {    0, "GN", kAny,      "Africa/Conakry"            },
    {    0, "GW", kAny,      "Africa/Bissau"             },
    {    0, "IE", kAny,      "Europe/Dublin"             },
    {    0, "IS", kAny,      "Atlantic/Reykjavik"        },
    {    0, "LR", kAny,      "Africa/Monrovia"           },
    {    0, "MA", kAny,      "Africa/Casablanca"         },
    {    0, "ML", kAny,      "Africa/Timbuktu"           },
    {    0, "MR", kAny,      "Africa/Nouakchott"         },
    {    0, "PT", kAny,      "Europe/Lisbon"             },
    {    0, "SH", kAny,      "Atlantic/St_Helena"        },
    {    0, "SL", kAny,      "Africa/Freetown"           },
    {    0, "SN", kAny,      "Africa/Dakar"              },
    {    0, "ST", kAny,      "Africa/Sao_Tome"           },
    {    0, "TG", kAny,      "Africa/Lome"               },
    {   60, "AD", kAny,      "Europe/Andorra"            },
    {   60, "AL", kAny,      "Europe/Tirane"             },
    {   60, "AO", kAny,      "Africa/Luanda"             },
    {   60, "AT", kAny,      "Europe/Vienna"             },
    {   60, "BE", kAny,      "Europe/Brussels"           },
    {   60, "BJ", kAny,      "Africa/Porto-Novo"         },
    {   60, "CF", kAny,      "Africa/Bangui"             },
    {   60, "CG", kAny,      "Africa/Kinshasa"           },
    {   60, "CH", kAny,      "Europe/Zurich"             },
    {   60, "CM", kAny,      "Africa/Douala"             },
    {   60, "CZ", kAny,      "Europe/Prague"             },
    {   60, "DE", kAny,      "Europe/Berlin"             },
    {   60, "DK", kAny,      "Europe/Copenhagen"         },
    {   60, "DZ", kAny,      "Africa/Algiers"            },
    {   60, "ES", kAny,      "Europe/Madrid"             },
    {   60, "FR", kAny,      "Europe/Paris"              },
    {   60, "GA", kAny,      "Africa/Libreville"         },
    {   60, "GI", kAny,      "Europe/Gibraltar"          },
    {   60, "GQ", kAny,      "Africa/Malabo"             },
    {   60, "HU", kAny,      "Europe/Budapest"           },
    {   60, "IT", kAny,      "Europe/Rome"               },
    {   60, "LI", kAny,      "Europe/Vaduz"              },
    {   60, "LU", kAny,      "Europe/Luxembourg"         },
    {   60, "LY", kAny,      "Africa/Tripoli"            },
    {   60, "MC", kAny,      "Europe/Monaco"             },
    {   60, "MT", kAny,      "Europe/Malta"              },
    {   60, "NA", kAny,      "Africa/Windhoek"           },
    {   60, "NE", kAny,      "Africa/Niamey"             },
    {   60, "NG", kAny,      "Africa/Lagos"              },
    {   60, "NL", kAny,      "Europe/Amsterdam"          },
    {   60, "NO", kAny,      "Europe/Oslo"               },
    {   60, "PL", kAny,      "Europe/Warsaw"             },
    {   60, "SE", kAny,      "Europe/Stockholm"          },
    {   60, "TD", kAny,      "Africa/Ndjamena"           },
    {   60, "TN", kAny,      "Africa/Tunis"              },
    {   60, "YU", kAny,      "Europe/Belgrade"           },
    {   60, NULL, kStandard, "Africa/Algiers"            },
    {   60, NULL, kDaylight, "Europe/Paris"              },
    {  120, "BG", kAny,      "Europe/Sofia"              },
    {  120, "BI", kAny,      "Africa/Bujumbura"          },
    {  120, "BW", kAny,      "Africa/Gaborone"           },
    {  120, "BY", kAny,      "Europe/Minsk"              },
    {  120, "CG", kAny,      "Africa/Lubumbashi"         },
    {  120, "CY", kAny,      "Asia/Nicosia"              },
    {  120, "EE", kAny,      "Europe/Tallinn"            },
    {  120, "EG", kAny,      "Africa/Cairo"              },
    {  120, "FI", kAny,      "Europe/Helsinki"           },
    {  120, "GR", kAny,      "Europe/Athens"             },
    {  120, "IL", kAny,      "Asia/Jerusalem"            },
    {  120, "JO", kAny,      "Asia/Amman"                },
    {  120, "LB", kAny,      "Asia/Beirut"               },
    {  120, "LS", kAny,      "Africa/Maseru"             },
    {  120, "LT", kAny,      "Europe/Vilnius"            },
    {  120, "LV", kAny,      "Europe/Riga"               },
    {  120, "MD", kAny,      "Europe/Chisinau"           },
    {  120, "ML", kAny,      "Africa/Blantyre"           },
    {  120, "MZ", kAny,      "Africa/Maputo"             },
    {  120, "RO", kAny,      "Europe/Bucharest"          },
    {  120, "RU", kAny,      "Europe/Kaliningrad"        },
    {  120, "RW", kAny,      "Africa/Kigali"             },
    {  120, "SD", kAny,      "Africa/Khartoum"           },
    {  120, "SY", kAny,      "Asia/Damascus"             },
    {  120, "SZ", kAny,      "Africa/Mbabane"            },
    {  120, "TR", kAny,      "Europe/Istanbul"           },
    {  120, "UA", kAny,      "Europe/Kiev"               },
    {  120, "ZA", kAny,      "Africa/Johannesburg"       },
    {  120, "ZM", kAny,      "Africa/Lusaka"             },
    {  120, "ZW", kAny,      "Africa/Harare"             },
    {  120, NULL, kStandard, "Africa/Johannesburg"       },
    {  120, NULL, kDaylight, "Europe/Istanbul"           },
    {  180, "BH", kAny,      "Asia/Bahrain"              },
    {  180, "DJ", kAny,      "Africa/Djibouti"           },
    {  180, "ER", kAny,      "Africa/Asmera"             },
    {  180, "ET", kAny,      "Africa/Addis_Ababa"        },
    {  180, "IQ", kAny,      "Asia/Baghdad"              },
    {  180, "KE", kAny,      "Africa/Nairobi"            },
    {  180, "KM", kAny,      "Indian/Comoro"             },
    {  180, "KW", kAny,      "Asia/Kuwait"               },
    {  180, "MK", kAny,      "Indian/Antananarivo"       },
    {  180, "QA", kAny,      "Asia/Qatar"                },
    {  180, "RU", kAny,      "Europe/Moscow"             },
    {  180, "SA", kAny,      "Asia/Riyadh"               },
    {  180, "SO", kAny,      "Africa/Mogadishu"          },
    {  180, "TZ", kAny,      "Africa/Dar_es_Salaam"      },
    {  180, "UA", kAny,      "Europe/Simferopol"         },
    {  180, "UG", kAny,      "Africa/Kampala"            },
    {  180, "YE", kAny,      "Asia/Aden"                 },
    {  180, "YT", kAny,      "Indian/Mayotte"            },
    {  180, NULL, kStandard, "Asia/Riyadh"               },
    {  180, NULL, kDaylight, "Europe/Moscow"             },
    {  210, "IR", kAny,      "Asia/Tehran"               },
    {  210, NULL, kAny,      "Asia/Tehran"               },
    {  240, "AE", kAny,      "Asia/Dubai"                },
    {  240, "AM", kAny,      "Asia/Yerevan"              },
    {  240, "AZ", kAny,      "Asia/Baku"                 },
    {  240, "KZ", kAny,      "Asia/Aqtau"                },
    {  240, "MU", kAny,      "Indian/Mauritius"          },
    {  240, "OM", kAny,      "Asia/Muscat"               },
    {  240, "RE", kAny,      "Indian/Reunion"            },
    {  240, "RU", kAny,      "Europe/Samara"             },
    {  240, "SC", kAny,      "Indian/Mahe"               },
    {  240, NULL, kStandard, "Asia/Yerevan"              },
    {  240, NULL, kDaylight, "Europe/Samara"             },
    {  270, "AF", kAny,      "Asia/Kabul"                },
    {  270, NULL, kAny,      "Asia/Kabul"                },
    {  300, "FR", kAny,      "Indian/Kerguelen"          },
    {  300, "GE", kAny,      "Asia/Tbilisi"              },
    {  300, "IO", kAny,      "Indian/Chagos"             },
    {  300, "KG", kAny,      "Asia/Bishkek"              },
    {  300, "KZ", kAny,      "Asia/Aqtobe"               },
    {  300, "MV", kAny,      "Indian/Maldives"           },
    {  300, "PK", kAny,      "Asia/Karachi"              },
    {  300, "RU", kAny,      "Asia/Yekaterinburg"        },
    {  300, "TJ", kAny,      "Asia/Dushanbe"             },
    {  300, "TM", kAny,      "Asia/Ashkhabad"            },
    {  300, "UZ", kAny,      "Asia/Tashkent"             },
    {  300, NULL, kStandard, "Asia/Karachi"              },
    {  300, NULL, kDaylight, "Asia/Yekaterinburg"        },
    {  330, "IN", kAny,      "Asia/Calcutta"             },
    {  330, NULL, kAny,      "Asia/Calcutta"             },
    {  345, "NP", kAny,      "Asia/Katmandu"             },
    {  345, NULL, kAny,      "Asia/Katmandu"             },
    {  360, "AQ", kAny,      "Antarctica/Mawson"         },
    {  360, "BD", kAny,      "Asia/Dacca"                },
    {  360, "BT", kAny,      "Asia/Thimbu"               },
    {  360, "KZ", kAny,      "Asia/Alma-Ata"             },
    {  360, "LK", kAny,      "Asia/Colombo"              },
    {  360, "RU", kAny,      "Asia/Novosibirsk"          },
    {  360, NULL, kStandard, "Asia/Dacca"                },
    {  360, NULL, kDaylight, "Asia/Novosibirsk"          },
    {  390, "CC", kAny,      "Indian/Cocos"              },
    {  390, "MM", kAny,      "Asia/Rangoon"              },
    {  390, NULL, kAny,      "Asia/Rangoon"              },
    {  420, "AU", kAny,      "Indian/Christmas"          },
    {  420, "ID", kAny,      "Asia/Jakarta"              },
    {  420, "KH", kAny,      "Asia/Phnom_Penh"           },
    {  420, "LA", kAny,      "Asia/Vientiane"            },
    {  420, "RU", kAny,      "Asia/Krasnoyarsk"          },
    {  420, "TH", kAny,      "Asia/Bangkok"              },
    {  420, "VN", kAny,      "Asia/Saigon"               },
    {  420, NULL, kStandard, "Asia/Bangkok"              },
    {  420, NULL, kDaylight, "Asia/Krasnoyarsk"          },
    {  480, "AQ", kAny,      "Antarctica/Casey"          },
    {  480, "AU", kAny,      "Australia/Perth"           },
    {  480, "BN", kAny,      "Asia/Brunei"               },
    {  480, "CN", kAny,      "Asia/Shanghai"             },
    {  480, "HK", kAny,      "Asia/Hong_Kong"            },
    {  480, "ID", kAny,      "Asia/Ujung_Pandang"        },
    {  480, "JP", kAny,      "Asia/Ishigaki"             },
    {  480, "MN", kAny,      "Asia/Ulan_Bator"           },
    {  480, "MO", kAny,      "Asia/Macao"                },
    {  480, "MY", kAny,      "Asia/Kuala_Lumpur"         },
    {  480, "PH", kAny,      "Asia/Manila"               },
    {  480, "RU", kAny,      "Asia/Irkutsk"              },
    {  480, "SG", kAny,      "Asia/Singapore"            },
    {  480, "TW", kAny,      "Asia/Taipei"               },
    {  480, NULL, kAny,      "Asia/Shanghai"             },
    {  540, "ID", kAny,      "Asia/Jayapura"             },
    {  540, "JP", kAny,      "Asia/Tokyo"                },
    {  540, "KP", kAny,      "Asia/Pyongyang"            },
    {  540, "KR", kAny,      "Asia/Seoul"                },
    {  540, "PW", kAny,      "Pacific/Palau"             },
    {  540, "RU", kAny,      "Asia/Yakutsk"              },
    {  540, NULL, kAny,      "Asia/Tokyo"                },
    {  570, "AU", kStandard, "Australia/Darwin"          },
    {  570, "AU", kDaylight, "Australia/Adelaide"        },
    {  570, NULL, kStandard, "Australia/Darwin"          },
    {  570, NULL, kDaylight, "Australia/Adelaide"        },
    {  600, "AQ", kAny,      "Antarctica/DumontDUrville" },
    {  600, "AU", kStandard, "Australia/Brisbane"        },
    {  600, "AU", kDaylight, "Australia/Sydney"          },
    {  600, "FM", kAny,      "Pacific/Truk"              },
    {  600, "GU", kAny,      "Pacific/Guam"              },
    {  600, "MP", kAny,      "Pacific/Saipan"            },
    {  600, "PG", kAny,      "Pacific/Port_Moresby"      },
    {  600, "RU", kAny,      "Asia/Vladivostok"          },
    {  600, NULL, kStandard, "Australia/Brisbane"        },
    {  600, NULL, kDaylight, "Australia/Sydney"          },
    {  630, "AU", kAny,      "Australia/Lord_Howe"       },
    {  630, NULL, kAny,      "Australia/Lord_Howe"       },
    {  660, "FM", kAny,      "Pacific/Ponape"            },
    {  660, "NC", kAny,      "Pacific/Noumea"            },
    {  660, "RU", kAny,      "Asia/Magadan"              },
    {  660, "SB", kAny,      "Pacific/Guadalcanal"       },
    {  660, "VU", kAny,      "Pacific/Efate"             },
    {  660, NULL, kStandard, "Pacific/Guadalcanal"       },
    {  660, NULL, kDaylight, "Asia/Magadan"              },
    {  690, "NF", kAny,      "Pacific/Norfolk"           },
    {  690, NULL, kAny,      "Pacific/Norfolk"           },
    {  720, "AQ", kAny,      "Antarctica/McMurdo"        },
    {  720, "FJ", kAny,      "Pacific/Fiji"              },
    {  720, "FM", kAny,      "Pacific/Kosrae"            },
    {  720, "KI", kAny,      "Pacific/Tarawa"            },
    {  720, "MH", kAny,      "Pacific/Majuro"            },
    {  720, "NR", kAny,      "Pacific/Nauru"             },
    {  720, "NZ", kAny,      "Pacific/Auckland"          },
    {  720, "RU", kAny,      "Asia/Kamchatka"            },
    {  720, "TV", kAny,      "Pacific/Funafuti"          },
    {  720, "US", kAny,      "Pacific/Wake"              },
    {  720, "WF", kAny,      "Pacific/Wallis"            },
    {  720, NULL, kStandard, "Pacific/Fiji"              },
    {  720, NULL, kDaylight, "Pacific/Auckland"          },
    {  765, "NZ", kAny,      "Pacific/Chatham"           },
    {  765, NULL, kAny,      "Pacific/Chatham"           },
    {  780, "KI", kAny,      "Pacific/Enderbury"         },
    {  780, "RU", kAny,      "Asia/Anadyr"               },
    {  780, "TO", kAny,      "Pacific/Tongatapu"         },
    {  780, NULL, kStandard, "Pacific/Tongatapu"         },
    {  780, NULL, kDaylight, "Asia/Anadyr"               },
    {  840, "KI", kAny,      "Pacific/Kiritimati"        },
    {  840, NULL, kAny,      "Pacific/Kiritimati"        },
};

#define timezone_map_length (sizeof(timezone_map) / sizeof(timezone_map[0]))
