/*
 * @(#)DriverTools.c	1.12 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************/
/*
**	DriverTools.c
**
**	Instrument and Song resource tools for the SoundMusicSys driver.
**	All functions here at platform independent
**
**
**	History	-
**	1/31/93		Created
**	6/23/95		Integrated into SoundMusicSys Library
**	11/20/95	Removed pragma unused
**	12/11/95	Removed GetInstrumentEnvelopeData
**	12/15/95	Changed memory allocation to X_API
**				Pulled out most Mac specific code
**	1/1/96		Forced more defaults for NewSong
**	1/11/96		Forced more defaults for NewInstrument
**	1/18/96		Spruced up for C++ extra error checking
**	1/28/96		Added GetKeySplitFromPtr
**				Added NewSongPtr & DisposeSongPtr
**	2/3/96		Removed extra includes
**	2/12/96		Changed Boolean to SMSBoolean
**	2/17/96		Added more platform compile arounds
**	2/19/96		Changed NewSongPtr
**	4/21/96		Removed CPU edian issues by use XGetShort & XGetLong
**	4/29/96		Missed an edian issue in GetKeySplitFromPtr
**	5/15/96		Added CreateSongResource for use with BeginSongFromMemory
**	5/30/96		Added XBF_ignoreBadPatches to NewSong functions
**				Fixed bug with CreateSongResource. Bad data translation
**	6/30/96		Changed font and re tabbed
**				Removed Byte reference from GetKeySplitFromPtr
**	10/23/96	Removed all platform specific code
**				Changed GetKeySplitFromPtr to XGetKeySplitFromPtr
**	12/30/96	Changed copyright
**	1/3/97		Added XGetSongInformation
**	1/6/97		Added songType to XNewSongPtr
**	1/12/97		Added I_TITLE to XGetSongInformation for SMS type
**	1/13/97		Added XGetSongResourceInfo & XDisposeSongResourceInfo & 
**				XGetSongResourceObjectID & XGetSongPerformanceSettings &
**				XGetSongResourceObjectType
**	1/15/97		Fixed a bug with XNewSongPtr. Forgot to change song types
**	1/16/97		Added XGetSongInformationSize
**	1/17/97		Totally changed the SONG RMF resource type, much more resource based
**	1/18/97		Added XCheckAllInstruments & XCheckValidInstrument
**	1/29/97		Added XCollectSoundsFromInstrumentID
**				Added XGetMidiData
**				Added XSetSongLocked
**				Added XGetSongInstrumentList
**				Added XCompressAndEncrypt
**	1/30/97		Added XGetSoundResourceByName & XGetSoundResourceByID
**	2/13/97		Fixed platform bug when reading resourceCount out of the RMF structure
**				Added XSetSongVolumeGain & XGetSongVolumeGain
**	2/15/97		Changed volumeGain to songVolume, changed XSetSongVolumeGain & 
**				XGetSongVolumeGain to XSetSongVolume & XGetSongVolume
**	2/25/97		Changed more RMF feilds. Hopefully it'll be the last for awhile
**	3/12/97		Changed XNewSongPtr to only use a valid ReverbMode type
**	3/14/97		Added XIsSoundUsedInInstrument & XRenumberSampleInsideInstrument
**				Fixed a bug with XCollectSoundsFromInstrument that disposed of
**				the instrument passed into it, corrupting the heap
**	3/15/97		Found a bug in XSetSongLocked that would trash the heap if you
**				tried to lock a song that was already locked
**	3/17/97		Fixed bug with XAddKeySplit
**	3/29/97		Fixed bug with XRemoveKeySplit that trashed the heap
**	4/1/97		Fixed a bug with encrypted samples
**				Fixed a bug with XNewSongFromSongResourceInfo & XGetSongResourceInfo
**				that did not preserve the volume
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**	6/28/97		Fixed a potential bug with XGetSoundResourceByName and did not copy
**				the data before decrypting
**	8/7/97		Changed XGetSongInformation to strip control characters before
**				return textual information via XDuplicateAndStripStr
**	8/12/97		Fixed XRenumberSampleInsideInstrument to save changes to renumbered
**				instruments and fixed platform dependancy
**	8/16/97		Added ifdefs (USE_CREATION_API) around functions that are used in
**				the process of creation of RMF files, XFILES, etc.
**	8/19/97		Removed XDuplicateAndStripStr from XGetSongInformation. Now relying
**				on editor not to put bad characters into data stream
**	9/30/97		Added BankStatus structure, added XCreateBankStatus & XGetBankStatus
**	10/2/97		Changed the search order in XGetMidiData, also fixed in XGetMidiData
**				a bug in which encrypted data was not copied first.
*/
/*****************************************************************************/
#include "X_API.h"
#include "X_Formats.h"
#include "GenPriv.h"
#include "GenSnd.h"



// Create a new song resource.
SongResource * XNewSongPtr(	SongType songType,
							short int midiID,
							short int maxSongVoices, 
							short int mixLevel, 
							short int maxEffectVoices,
							ReverbMode reverbType)
{
	SongResource		*song;
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	song = NULL;
	switch (songType)
	{
		case SONG_TYPE_SMS:
			songSMS = (SongResource_SMS *)XNewPtr((long)sizeof(SongResource_SMS));
			if (songSMS)
			{
				songSMS->songType = SONG_TYPE_SMS;
				XPutShort(&songSMS->midiResourceID, midiID);
				songSMS->noteDecay = 5;
				songSMS->maxNotes = maxSongVoices;
				XPutShort(&songSMS->mixLevel, mixLevel);
				songSMS->maxEffects = maxEffectVoices;
				songSMS->reverbType = reverbType;
				songSMS->flags1 = XBF_enableMIDIProgram | XBF_interpolateSong | XBF_interpolateSong;
				songSMS->flags2 = XBF_masterEnableAmpScale | XBF_forceAmpScale | XBF_ampScaleLead | XBF_ignoreBadPatches;
				songSMS->defaultPercusionProgram = -1;		// 126
//				XPutShort(&songSMS->remapCount, 0);
			}
			song = (SongResource *)songSMS;
			break;
		case SONG_TYPE_RMF:
			songRMF = (SongResource_RMF *)XNewPtr((long)sizeof(SongResource_RMF) - sizeof(short));
			if (songRMF)
			{
				songRMF->songType = SONG_TYPE_RMF;
				XPutShort(&songRMF->rmfResourceID, midiID);
				XPutShort(&songRMF->maxNotes, maxSongVoices);
				XPutShort(&songRMF->mixLevel, mixLevel);
				XPutShort(&songRMF->maxEffects, maxEffectVoices);

				songRMF->reverbType = reverbType;
//				XPutShort(&songRMF->resourceCount, 0);		// no resources
			}
			song = (SongResource *)songRMF;
			break;
		case SONG_TYPE_MOD:
			break;
	}
	return song;
}

void XDisposeSongPtr(SongResource *theSong)
{
	XDisposePtr((XPTR)theSong);
}


// Get a keysplit entry. The result will be ordered for the native CPU
void XGetKeySplitFromPtr(InstrumentResource *theX, short int entry, KeySplit *keysplit)
{
	KeySplit	*pSplits;
	short int	count;

	count = XGetShort(&theX->keySplitCount);
	if ( (count) && (entry < count) )
	{
		pSplits = (KeySplit *) ( ((unsigned char *)&theX->keySplitCount) + sizeof(short int));
		*keysplit = pSplits[entry];
		keysplit->sndResourceID = XGetShort(&keysplit->sndResourceID);
		keysplit->miscParameter1 = XGetShort(&keysplit->miscParameter1);
		keysplit->miscParameter2 = XGetShort(&keysplit->miscParameter2);
	}
	else
	{
		XSetMemory((void *)keysplit, (long)sizeof(KeySplit), 0);
	}
}

// Set a keysplit entry. The result will be ordered for the 68k CPU
#if USE_CREATION_API == TRUE
void XSetKeySplitFromPtr(InstrumentResource *theX, short int entry, KeySplit *keysplit)
{
	KeySplit	*pSplits;
	short int	count;

	count = XGetShort(&theX->keySplitCount);
	if ( (count) && (entry < count) )
	{
		pSplits = (KeySplit *) ( ((unsigned char *)&theX->keySplitCount) + sizeof(short int));
		pSplits[entry] = *keysplit;

		XPutShort(&pSplits[entry].sndResourceID, keysplit->sndResourceID);
		XPutShort(&pSplits[entry].miscParameter1, keysplit->miscParameter1);
		XPutShort(&pSplits[entry].miscParameter2, keysplit->miscParameter2);
	}
	else
	{
		XSetMemory(keysplit, (long)sizeof(KeySplit), 0);
	}
}
#endif	// USE_CREATION_API == TRUE

// Increase the number of key splits. This will not clean the entry's that are added.
// This will add from the end. This will freeup the passed InstrumentResource
#if USE_CREATION_API == TRUE
InstrumentResource * XAddKeySplit(InstrumentResource *theX, short int howMany)
{
	long				size, moveSize, size2;
	short int			total;
	KeySplit 			*pSplits;
	char				*pSrc, *pDest;
	InstrumentResource	*theNewX;

	theNewX = NULL;
	if (theX)
	{
		total = XGetShort(&theX->keySplitCount);
		size = XGetPtrSize(theX);
		size2 =  (sizeof(KeySplit) * howMany);
		theNewX = (InstrumentResource *)XNewPtr(size2 + size);
		if (theNewX)
		{
			XBlockMove(theX, theNewX, size);	// make copy
			XPutShort(&theNewX->keySplitCount, total + howMany);
			// must move all zones down
			pSplits = (KeySplit *) ( ((unsigned char *)&theX->keySplitCount) + sizeof(unsigned short int));
			pSrc = (char *)&pSplits[total];

			pSplits = (KeySplit *) ( ((unsigned char *)&theNewX->keySplitCount) + sizeof(unsigned short int));
			pDest = (char *)&pSplits[total + howMany];
			moveSize =  size - (pSrc - (char *)theX);
			// this move must handle overlapping data
			XBlockMove(pSrc, pDest, moveSize);
		}
	}
	return theNewX;
}
#endif

// Decrease the number of key splits. This will delete from the end
#if USE_CREATION_API == TRUE
InstrumentResource * XRemoveKeySplit(InstrumentResource *theX, short int howMany)
{
	long				size, moveSize, size2;
	short int			total;
	KeySplit 			*pSplits;
	char				*pSrc, *pDest;
	InstrumentResource	*theNewX;

	theNewX = NULL;
	if (theX)
	{
		total = XGetShort(&theX->keySplitCount);
		if (howMany > total)
		{
			howMany = total;
		}

		size = XGetPtrSize(theX);
		size2 =  (sizeof(KeySplit) * howMany);
		theNewX = (InstrumentResource *)XNewPtr(size - size2);
		if (theNewX)
		{
			XBlockMove(theX, theNewX, size - size2);	// make copy
			XPutShort(&theNewX->keySplitCount, total - howMany);
			// must move all zones down
			pSplits = (KeySplit *) ( ((unsigned char *)&theX->keySplitCount) + sizeof(unsigned short int));
			pSrc = (char *)&pSplits[total];		// go to byte at the end

			pSplits = (KeySplit *) ( ((unsigned char *)&theNewX->keySplitCount) + sizeof(unsigned short int));
			pDest = (char *)&pSplits[total - howMany];

			moveSize =  size - (pSrc - (char *)theX);
			// this move must handle overlapping data
			XBlockMove(pSrc, pDest, moveSize);
		}
	}
	return theNewX;
}
#endif

/*
** Return the number of INST's and an array of resource IDs
*/
#if USE_CREATION_API == TRUE
short int XGetInstrumentArray(short int *instArray, short maxArraySize)
{
	short int	icount, totalCount;
	long		size;
	XPTR		theRes;

	/* Collect INST resources
	*/
	totalCount = 0;
	if (instArray)
	{
		for (icount = 0; icount < maxArraySize; icount++)
		{
			theRes = XGetAndDetachResource(ID_INST, icount, &size);
			if (theRes)
			{
				XDisposePtr(theRes);
				instArray[totalCount++] = icount;
				if (totalCount > maxArraySize)
				{
					totalCount = maxArraySize;
					break;
				}
			}
		}
		XBubbleSortArray((short *)instArray, (short)totalCount);
	}
	return totalCount;
}
#endif

#if USE_CREATION_API == TRUE
XBOOL XIsSoundUsedInInstrument(InstrumentResource *theX, short int sampleSoundID)
{
	short int	sndArray[MAX_SAMPLES];		/* Max samples per instrument */
	short int	count, total;
	XBOOL		used;

	used = FALSE;
	total = XCollectSoundsFromInstrument(theX, sndArray, MAX_SAMPLES);
	for (count = 0; count < total; count++)
	{
		if (sndArray[count] == sampleSoundID)
		{
			used = TRUE;
			break;
		}
	}
	return used;
}
#endif

// Renumber sample ID that is used in and instrument
#if USE_CREATION_API == TRUE
void XRenumberSampleInsideInstrument(InstrumentResource *theX, short int originalSampleID, 
																short int newSampleID)
{
	short int	count, total;
	KeySplit	split;

	total = theX->keySplitCount;
	if (total == 0)
	{
		XPutShort(&theX->sndResourceID, newSampleID);
	}
	else
	{
		for (count = 0; count < total; count++)
		{
			XGetKeySplitFromPtr(theX, count, &split);
			if (split.sndResourceID == originalSampleID)
			{
				split.sndResourceID = newSampleID;
				XSetKeySplitFromPtr(theX, count, &split);
			}
		}
	}
}
#endif

/*
**	Given an INST resource, this will return an array of snd resources
**	that are used for this instrument
*/
#if USE_CREATION_API == TRUE
short int XCollectSoundsFromInstrument(InstrumentResource *theX, short *sndArray, short maxArraySize)
{
	short int			sndOutCount, sndCount, splitCount, count, count2;
	KeySplit			theSplit;
	short int			countSndArray[MAX_SAMPLES];		/* Max samples per instrument */
	char				goodSound;
	short int			soundID;

	sndCount = 0;
	sndOutCount = 0;
	if (theX)
	{
		splitCount = XGetShort(&theX->keySplitCount);
		if ( splitCount == 0)
		{
			if (maxArraySize > 1)
			{
				sndCount = 1;
				countSndArray[0] = XGetShort(&theX->sndResourceID);
				if (countSndArray[0] == -1)
				{
					sndCount = 0;
				}
			}
		}
		else
		{
			if (maxArraySize > splitCount)
			{
				sndCount = 1;
				countSndArray[0] = XGetShort(&theX->sndResourceID);
				if (countSndArray[0] == -1)
				{
					sndCount = 0;
				}
				for (count = 0; count < splitCount; count++)
				{
					XGetKeySplitFromPtr(theX, count, &theSplit);
					countSndArray[sndCount] = theSplit.sndResourceID;
					if (countSndArray[sndCount] != -1)
					{
						sndCount++;
					}
				}
			}
		}
		XBubbleSortArray((short *)countSndArray, (short)sndCount);

// Remove duplicates
		sndOutCount = 0;
		for (count = 0; count < sndCount; count++)
		{
			goodSound = TRUE;
			soundID = countSndArray[count];
			for (count2 = 0; count2 < sndOutCount; count2++)
			{
				if (soundID == sndArray[count2])
				{
					goodSound = FALSE;
					break;
				}
			}
			if (goodSound)
			{
				sndArray[sndOutCount++] = soundID;
			}
		}
	}
	return sndOutCount;
}
#endif

#if USE_CREATION_API == TRUE
short int XCollectSoundsFromInstrumentID(short int theID, short *sndArray, short maxArraySize)
{
	InstrumentResource	*theX;
	short int			totalSnds;
	long				size;

	totalSnds = 0;
	theX = (InstrumentResource *)XGetAndDetachResource(ID_INST, theID, &size);
	if (theX)
	{
		totalSnds = XCollectSoundsFromInstrument(theX, sndArray, maxArraySize);
		XDisposePtr((XPTR)theX);
	}
	return totalSnds;
}
#endif

#if USE_CREATION_API == TRUE
short int XCheckValidInstrument(short int theID)
{
	InstrumentResource	*theX;
	short int			totalSnds, sndArray[MAX_SAMPLES];
	short int			count, badLoad;
	XPTR				theSnd;
	long				size;

	badLoad = 0;
	theX = (InstrumentResource *)XGetAndDetachResource(ID_INST, theID, &size);
	if (theX)
	{
		totalSnds = XCollectSoundsFromInstrument(theX, sndArray, MAX_SAMPLES);
		XDisposePtr((XPTR)theX);
		for (count = 0; count < totalSnds; count++)
		{
			theSnd = XGetAndDetachResource(ID_CSND, sndArray[count], &size);		// look for compressed version first
			if (theSnd == NULL)
			{
				theSnd = XGetAndDetachResource(ID_SND, sndArray[count], &size);
			}
			if (theSnd == NULL)
			{
				theSnd = XGetAndDetachResource(ID_ESND, sndArray[count], &size);
			}
			if (theSnd == NULL)
			{
				badLoad = sndArray[count];
			}
			XDisposePtr(theSnd);
			if (badLoad)
			{
				break;
			}
		}
	}
	return badLoad;
}
#endif

#if USE_CREATION_API == TRUE
XBOOL XIsSampleUsedInAllInstruments(short int soundSampleID, short int *pWhichInstrument)
{
	InstrumentResource	*theX;
	short int			count, totalInstruments;
	XBOOL				used;
	long				size;
	short int			instArray[MAX_INSTRUMENTS * MAX_BANKS];

	used = FALSE;
	totalInstruments = XGetInstrumentArray(instArray, MAX_INSTRUMENTS * MAX_BANKS);
	if (totalInstruments && pWhichInstrument)
	{
		for (count = 0; count < totalInstruments; count++)
		{
			theX = (InstrumentResource *)XGetAndDetachResource(ID_INST, instArray[count], &size);
			if (theX)
			{
				used = XIsSoundUsedInInstrument(theX, soundSampleID);
				XDisposePtr((XPTR)theX);
				if (used)
				{
					*pWhichInstrument = instArray[count];
					break;
				}
			}
		}
	}
	return used;
}
#endif

#if USE_CREATION_API == TRUE
XBOOL XCheckAllInstruments(short *badInstrument, short *badSnd)
{
	short int	count, bad;
	XBOOL		badLoad;
	short int	instArray[MAX_INSTRUMENTS * MAX_BANKS], totalInstruments;

	badLoad = FALSE;
	if (badInstrument && badSnd)
	{
		totalInstruments = XGetInstrumentArray(instArray, MAX_INSTRUMENTS * MAX_BANKS);
		if (totalInstruments)
		{
			for (count = 0; count < totalInstruments; count++)
			{
				bad = XCheckValidInstrument(instArray[count]);
				if (bad)
				{
					*badInstrument = instArray[count];
					*badSnd = bad;
					badLoad = TRUE;
					break;
				}
			}
		}
	}
	return badLoad;
}
#endif

#if USE_CREATION_API == TRUE
// Scan an array and return index
static short int PV_ConvertResourceID2Index(short *sndArray, short totalSnds, short theID)
{
	register short int count, indexValue;
	
	indexValue = -1;
	/* convert resourceID into instrument index */
	for (count = 0; count < totalSnds; count++)
	{
		if (theID == sndArray[count])
		{
			indexValue = count;
			break;
		}
	}
	return indexValue;
}

// Given an array of instruments, this will return an array of SND resources that are required to load
// these instruments
short int XGetTotalKeysplits(short int *instArray, short int totalInstruments, 
								short int *sndArray, short int totalSnds)
{
	register short int	count, count2, count3, count4, keyCount;
	KeySplit			theSplit;
	InstrumentResource	*theInstrument;
	char				*pLoaded;
	long				size;

	keyCount = 0;
	if (instArray && totalInstruments && sndArray && totalSnds)
	{
		pLoaded = (char *)XNewPtr((long)sizeof(char) * totalSnds);
		if (pLoaded)
		{
			for (count = 0; count < totalInstruments; count++)
			{
				theInstrument = (InstrumentResource *)XGetAndDetachResource(ID_INST, instArray[count], &size);
				if (theInstrument)
				{
					count3 = XGetShort(&theInstrument->keySplitCount);
					for (count2 = 0; count2 < count3; count2++)
					{
						XGetKeySplitFromPtr(theInstrument, count2, &theSplit);
						count4 = PV_ConvertResourceID2Index(sndArray, totalSnds, theSplit.sndResourceID);
						if (count4)
						{
							if (pLoaded[count4] == 0)
							{
								pLoaded[count4] = 1;
								keyCount++;
							}
						}
					}
				}
			}
			XDisposePtr((XPTR)pLoaded);
		}
	}
	return keyCount;
}
#endif

/*
**	This will walk through all Instrument resources and collect all snd resources
**	that are used.
*/
#if USE_CREATION_API == TRUE
short int XGatherAllSoundsFromAllInstruments(short int *pSndArray, short int maxArraySize)
{
	short int			count, icount, sndCount, jcount, soundID;
	XBOOL				goodSound;
	short int			instArray[MAX_INSTRUMENTS * MAX_BANKS], totalInstruments;
	short int			sndArray[MAX_SAMPLES], totalSnds;
	short int			*completeSndArray, completeSndCount;
	InstrumentResource	*theX;
	long				size;

	completeSndCount = 0;
	sndCount = 0;
	completeSndArray = (short *)XNewPtr(16384L * sizeof(short));		/* max 128 with 128 splits per instrument */
	if (completeSndArray)
	{
		sndCount = 0;
		totalInstruments = XGetInstrumentArray(instArray, MAX_INSTRUMENTS * MAX_BANKS);
		if (totalInstruments)
		{
			for (count = 0; count < totalInstruments; count++)
			{
				theX = (InstrumentResource *)XGetAndDetachResource(ID_INST, instArray[count], &size);
				if (theX)
				{
					totalSnds = XCollectSoundsFromInstrument(theX, sndArray, MAX_SAMPLES);
					XDisposePtr((XPTR)theX);

					for (icount = 0; icount < totalSnds; icount++)
					{
						completeSndArray[completeSndCount++] = sndArray[icount];
					}
				}
			}
			XBubbleSortArray((short *)completeSndArray, (short)completeSndCount);

// Remove duplicates
			for (jcount = 0; jcount < completeSndCount; jcount++)
			{
				goodSound = TRUE;
				soundID = completeSndArray[jcount];
				for (count = 0; count < sndCount; count++)
				{
					if (soundID == pSndArray[count])
					{
						goodSound = FALSE;
						break;
					}
				}
				if (goodSound)
				{
					if (sndCount < maxArraySize)
					{
						pSndArray[sndCount++] = soundID;
					}
				}
			}
		}
		XDisposePtr((XPTR)completeSndArray);
	}
	return sndCount;
}
#endif	// USE_CREATION_API == TRUE

// Given a list of instruments, this will return the sample ID's that are required to load
// all of these instruments
#if USE_CREATION_API == TRUE
short int XGetSamplesFromInstruments(short int *pInstArray, short int maxInstArraySize, 
										short int *pSndArray, short int maxSndArraySize)
{
	register long		count, instCount, sndCount, newCount, completeSndCount, soundID;
	short int			completeSndArray[MAX_SAMPLES];
	XBOOL				goodSound;
	InstrumentResource	*theX;
	long				size;

	sndCount = 0;
	if ( (pInstArray) && (pSndArray) )
	{
		instCount = 0;
		for (count = 0; count < maxInstArraySize; count++)
		{
			if (pInstArray[count] != -1)
			{
				instCount++;
			}
			else
			{
				break;
			}
		}
		for (count = 0; count < maxSndArraySize; count++)
		{
			pSndArray[count] = -1;
			completeSndArray[count] = -1;
		}


		completeSndCount = 0;
		for (count = 0; count < instCount; count++)
		{
			theX = (InstrumentResource *)XGetAndDetachResource(ID_INST, pInstArray[count], &size);
			if (theX)
			{
				newCount = XCollectSoundsFromInstrument(theX, &completeSndArray[completeSndCount], 128);
				XDisposePtr((XPTR)theX);
				completeSndCount += newCount;
			}
		}
		
		// remove duplicates in snds
		for (newCount = 0; newCount < completeSndCount; newCount++)
		{
			goodSound = TRUE;
			soundID = completeSndArray[newCount];
			for (count = 0; count < sndCount; count++)
			{
				if (soundID == pSndArray[count])
				{
					goodSound = FALSE;
					break;
				}
			}
			if (goodSound)
			{
				pSndArray[sndCount++] = soundID;
			}
		}
		XBubbleSortArray(pSndArray, sndCount);
	}
	return sndCount;
}
#endif	// USE_CREATION_API == TRUE

// Create a new instrument. Free with XDisposeInstrument
#if USE_CREATION_API == TRUE
InstrumentResource * XNewInstrument(short int leadSndID)
{
	InstrumentResource *theX;

	theX = (InstrumentResource *)XNewPtr((long)sizeof(InstrumentResource));
	if (theX)
	{
		theX->sndResourceID = leadSndID;
		theX->tremoloEnd = 0x8000;

		theX->flags1 |= ZBF_enableInterpolate | ZBF_enableAmpScale | ZBF_useSampleRate;
	}
	return theX;
}


void XDisposeInstrument(InstrumentResource *theX)
{
	XDisposePtr((XPTR)theX);
}
#endif	// USE_CREATION_API == TRUE

// Find a resource type and return a pointer to its memory block. If type is R_LAST_RESOURCE, then return
// a pointer to the very last memory block
#if USE_FULL_RMF_SUPPORT == TRUE
static void * PV_FindSongResourceTypePointer(SongResource_RMF *songRMF, SongResourceType findType, long *pResourceLength)
{
	char				*pUnit;
	short int			count, subCount;
	SongResourceType	type;
	unsigned long		length;
	short int			resourceCount;
	void				*pEnd;

	pEnd = NULL;
	if (songRMF)
	{
		resourceCount = XGetShort(&songRMF->resourceCount);
		pUnit = (char *)&songRMF->resourceData;
		for (count = 0; count < resourceCount; count++)
		{
			type = (SongResourceType)XGetLong(pUnit);
			pUnit += sizeof (SongResourceType);
			switch (type)
			{
				default:
					// found a bad type, just stop because we don't know how to proceed
					break;
				case R_PERFORMED_BY:
				case R_TITLE:
				case R_COMPOSER:
				case R_COPYRIGHT_DATE:
				case R_COPYRIGHT_LINE:
				case R_PUBLISHER_CONTACT:
				case R_LICENSED_TO_URL:
				case R_USE_OF_LICENSE:
				case R_LICENSE_TERM:
				case R_EXPIRATION_DATE:
				case R_COMPOSER_NOTES:
				case R_INDEX_NUMBER:
					// this is a zero terminated string, always
					if (songRMF->locked == FALSE)
					{
						length = XStrLen(pUnit) + 1;
					}
					else
					{
						length = XEncryptedStrLen(pUnit) + 1;
					}
					pUnit += length;
					if (pResourceLength)
					{
						*pResourceLength = length;
					}
					break;
				case R_VELOCITY_CURVE:
					pUnit += 256;
					if (pResourceLength)
					{
						*pResourceLength = 256;
					}
					break;
				case R_INSTRUMENT_REMAP:
					subCount = XGetShort(pUnit);
					pUnit += 2;
					// format is word count followed by x number of (word, word)
					length = sizeof(short) * subCount;
					pUnit += length;
					if (pResourceLength)
					{
						*pResourceLength = length;
					}
					break;
			}
			if (type == findType)
			{
				break;
			}
		}
		if (findType == R_LAST_RESOURCE)
		{
			pEnd = pUnit;
		}
	}
	return pEnd;
}
#endif	//USE_FULL_RMF_SUPPORT == TRUE

#if USE_FULL_RMF_SUPPORT == TRUE
// Given a song resource type, and the pointer to the data block; this will put the resource
// into the given pointer if found.
//
// pFillData or pFillSize can be NULL. If not NULL, then data will be placed there
static void PV_FillSongResource(SongResourceType resourceType, short int resourceCount, XBOOL encrypted, 
								void *pResourceData, unsigned long resourceDataLength, 
								void *pFillData, unsigned long *pFillSize)
{
	char				*pUnit;
	short int			count, subCount;
	SongResourceType	type;
	char				fill;
	unsigned long		length;

	if (pResourceData && resourceCount && resourceDataLength)
	{
		pUnit = (char *)pResourceData;
		for (count = 0; count < resourceCount; count++)
		{
			type = (SongResourceType)XGetLong(pUnit);
			pUnit += sizeof (SongResourceType);

			fill = (resourceType == type) ? TRUE : FALSE;
			switch (type)
			{
				default:
					// found a bad type, just stop because we don't know how to proceed
					break;
				case R_TITLE:
				case R_PERFORMED_BY:
				case R_COMPOSER:
				case R_COPYRIGHT_DATE:
				case R_COPYRIGHT_LINE:
				case R_PUBLISHER_CONTACT:
				case R_LICENSED_TO_URL:
				case R_USE_OF_LICENSE:
				case R_LICENSE_TERM:
				case R_EXPIRATION_DATE:
				case R_COMPOSER_NOTES:
				case R_INDEX_NUMBER:
					// this is a zero terminated string, always
					if (encrypted == FALSE)
					{
						length = XStrLen(pUnit) + 1;
					}
					else
					{
						length = XEncryptedStrLen(pUnit) + 1;
					}
					if (fill)
					{
						if (pFillData)
						{
							XBlockMove(pUnit, pFillData, length);
							if (encrypted)
							{
								XDecryptData(pFillData, length);
							}
						}
						if (pFillSize)
						{
							*pFillSize = length;
						}
					}
					pUnit += length;
					break;
				case R_VELOCITY_CURVE:
					if (fill)
					{
						XBlockMove(pUnit, pFillData, 256L);
					}
					pUnit += 256;
					break;
				case R_INSTRUMENT_REMAP:
					subCount = XGetShort(pUnit);
					pUnit += 2;
					// format is word count followed by x number of (word, word)
					pUnit += sizeof(short) * subCount;
					break;
			}
			if (fill)
			{
				break;
			}
		}
	}
}
#endif	//USE_FULL_RMF_SUPPORT == TRUE

#if USE_CREATION_API == TRUE
static XBOOL PV_ValidResourceForSongType(SongResourceType resourceType, SongType type)
{
	XBOOL	valid;

	valid = FALSE;
	switch (type)
	{
		case SONG_TYPE_SMS:
			switch (resourceType)
			{
				case R_TITLE:
				case R_COMPOSER:
				case R_COPYRIGHT_DATE:
				case R_PUBLISHER_CONTACT:
				case R_INSTRUMENT_REMAP:
					valid = TRUE;
					break;
			}
			break;
		case SONG_TYPE_RMF:
			switch (resourceType)
			{
				case R_TITLE:
				case R_PERFORMED_BY:
				case R_COMPOSER:
				case R_COPYRIGHT_DATE:
				case R_COPYRIGHT_LINE:
				case R_PUBLISHER_CONTACT:
				case R_LICENSED_TO_URL:
				case R_USE_OF_LICENSE:
				case R_LICENSE_TERM:
				case R_EXPIRATION_DATE:
				case R_COMPOSER_NOTES:
				case R_INDEX_NUMBER:
				case R_VELOCITY_CURVE:
				case R_INSTRUMENT_REMAP:
					valid = TRUE;
					break;
			}
			break;
	}
	return valid;
}
#endif	//USE_CREATION_API == TRUE

#if USE_CREATION_API == TRUE
static SongResource_RMF * PV_AddRMFSongResource(SongResource *theSong, SongResourceType resourceType, void *pResource, long resourceLength)
{
	SongResource_RMF	*newSong;
	long				size;
	char				*pBlock;

	newSong = NULL;
	if (theSong && pResource && resourceLength)
	{
		if (PV_ValidResourceForSongType(resourceType, SONG_TYPE_RMF))
		{
			size = XGetPtrSize(theSong);
			newSong = (SongResource_RMF *)XNewPtr(size + sizeof(SongResourceType) + resourceLength);
			if (newSong)
			{
				XBlockMove(theSong, newSong, size);
				pBlock = (char *)PV_FindSongResourceTypePointer(newSong, R_LAST_RESOURCE, NULL);
				if (pBlock)
				{
					size = XGetShort(&newSong->resourceCount) + 1;
					XPutShort(&newSong->resourceCount, size);
					XPutLong(pBlock, resourceType);
					pBlock += 4;
					XBlockMove(pResource, pBlock, resourceLength);
				}
				else
				{
					XDisposePtr(newSong);
					newSong = NULL;
				}
			}

		}
	}
	return newSong;
}

static SongResource_RMF * PV_DeleteRMFSongResource(SongResource *theSong, SongResourceType resourceType)
{
	SongResource_RMF	*newSong;
	long				size, resourceLength, offsetStart, offsetEnd;
	char				*pBlock;

	newSong = NULL;
	if (theSong)
	{
		if (PV_ValidResourceForSongType(resourceType, SONG_TYPE_RMF))
		{
			pBlock = (char *)PV_FindSongResourceTypePointer((SongResource_RMF *)theSong, resourceType, &resourceLength);
			if (pBlock)
			{
				size = XGetPtrSize(theSong) - sizeof(SongResourceType) - resourceLength;
				if (size > 0)
				{
					newSong = (SongResource_RMF *)XNewPtr(size);
					if (newSong)
					{
						offsetStart = pBlock - (char *)theSong;
						offsetEnd = offsetStart + resourceLength + sizeof(SongResourceType);
						XBlockMove(theSong, newSong, offsetStart);
						XBlockMove(((char *)theSong) + offsetEnd, ((char *)newSong) + offsetStart, size - offsetEnd);

						size = XGetShort(&newSong->resourceCount) - 1;
						XPutShort(&newSong->resourceCount, size);
					}
				}
			}
		}
	}
	return newSong;
}


// Given a SongResource and a new type, data and length, this will create a new SongResource
// with the modified resource and return it.
SongResource * XChangeSongResource(SongResource *theSong, long songSize, SongResourceType resourceType, void *pResource, long resourceLength)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;
	SongResource		*newSong;
	char				name1[256];
	char				name2[256];
	char				name3[256];
	char				name4[256];
	SongType			songType;
	char				*pData;

	newSong = NULL;
	if (theSong && pResource && resourceLength)
	{
		songType = (SongType)((SongResource_SMS *)theSong)->songType;
		if (PV_ValidResourceForSongType(resourceType, songType))
		{
			switch (songType)
			{
				case SONG_TYPE_SMS:
					XGetSongInformation(theSong, songSize, I_COPYRIGHT, name1);
					XGetSongInformation(theSong, songSize, I_COMPOSER, name2);
					XGetSongInformation(theSong, songSize, I_TITLE, name3);
					XGetSongInformation(theSong, songSize, I_PUBLISHER_CONTACT, name4);
					switch (resourceType)
					{
						case R_COPYRIGHT_DATE:
							XBlockMove(pResource, name1, resourceLength);
							goto changeSMSresource;
						case R_COMPOSER:
							XBlockMove(pResource, name2, resourceLength);
							goto changeSMSresource;
						case R_TITLE:
							XBlockMove(pResource, name3, resourceLength);
							goto changeSMSresource;
						case R_PUBLISHER_CONTACT:
							XBlockMove(pResource, name4, resourceLength);
			changeSMSresource:
							newSong = (SongResource *)XNewPtr(songSize + resourceLength);
							songSMS = (SongResource_SMS *)newSong;
							if (newSong)
							{
								XBlockMove(theSong, newSong, songSize);
								pData = (char *)&songSMS->remaps;
								pData += sizeof(Remap) * songSMS->remapCount;
								// store copyright first
								XCtoPstr(name1);
								XBlockMove(name1, pData, name1[0] + 1);
								pData += name1[0] + 1;
								// store author second
								XCtoPstr(name2);
								XBlockMove(name2, pData, name2[0] + 1);
								pData += name2[0] + 1;
								// store title third
								XCtoPstr(name3);
								XBlockMove(name3, pData, name3[0] + 1);
								pData += name3[0] + 1;
								// store contact fourth
								XCtoPstr(name4);
								XBlockMove(name4, pData, name4[0] + 1);
								pData += name4[0] + 1;
							}
							break;
					}
					break;
				case SONG_TYPE_RMF:
					songRMF = (SongResource_RMF *)theSong;
					newSong = PV_DeleteRMFSongResource(theSong, resourceType);
					if (newSong)
					{
						XDisposeSongPtr(theSong);
						theSong = newSong;
					}
					newSong = PV_AddRMFSongResource(theSong, resourceType, pResource, resourceLength);
					if (newSong)
					{
						XDisposeSongPtr(theSong);
					}
					break;
			}
		}
	}
	return newSong;
}
#endif	// USE_CREATION_API == TRUE

#if USE_FULL_RMF_SUPPORT == TRUE
unsigned long XGetSongInformationSize(SongResource *theSong, long songSize, SongInfo type)
{
	char				*pData, *pName;
	long				length, offset;
	unsigned long		temp;
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;
	unsigned long		size;

	size = 0;
	if (theSong && songSize)
	{
		switch (((SongResource_SMS *)theSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)theSong;
				switch (type)
				{
					case I_COMPOSER:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							temp = (*pData) + 1;		// skip copyright text
							pData = (char *)songSMS;
							offset += temp;
							pName = (char *)pData + offset;
							size = pName[0];
						}
						break;
					case I_COPYRIGHT:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							pData = (char *)songSMS;
							pName = (char *)pData + offset;
							size = pName[0];
						}
						break;
					case I_TITLE:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							temp = (*pData) + 1;		// skip copyright text
							offset += temp;
							pData = (char *)songSMS + offset;

							temp = (*pData) + 1;		// skip author text
							offset += temp;
							pData = (char *)songSMS;

							if (offset < length)
							{
								pName = (char *)pData + offset;
								size = pName[0];
							}
						}
						break;
					case I_PUBLISHER_CONTACT:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							temp = (*pData) + 1;		// skip copyright text
							offset += temp;
							pData = (char *)songSMS + offset;

							temp = (*pData) + 1;		// skip author text
							offset += temp;
							pData = (char *)songSMS + offset;

							temp = (*pData) + 1;		// skip title text
							offset += temp;
							pData = (char *)songSMS;

							if (offset < length)
							{
								pName = (char *)pData + offset;
								size = pName[0];
							}
						}
						break;
					case I_LICENSED_TO_URL:
					case I_USE_OF_LICENSE:
					case I_LICENSE_TERM:
					case I_EXPIRATION_DATE:
					case I_COMPOSER_NOTES:
					case I_INDEX_NUMBER:
						size = 0;
						break;
				}
				break;
			
			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)theSong;
				temp = 0;
				switch (type)
				{
					case I_TITLE:
						temp = R_TITLE;
						break;
					case I_COMPOSER:
						temp = R_COMPOSER;
						break;
					case I_COPYRIGHT:
						temp = R_COPYRIGHT_DATE;
						break;
					case I_PERFORMED_BY:
						temp = R_PERFORMED_BY;
						break;
					case I_PUBLISHER_CONTACT:
						temp = R_PUBLISHER_CONTACT;
						break;
					case I_USE_OF_LICENSE:
						temp = R_USE_OF_LICENSE;
						break;
					case I_LICENSE_TERM:
						temp = R_LICENSE_TERM;
						break;
					case I_LICENSED_TO_URL:
						temp = R_LICENSED_TO_URL;
						break;
					case I_EXPIRATION_DATE:
						temp = R_EXPIRATION_DATE;
						break;
					case I_COMPOSER_NOTES:
						temp = R_COMPOSER_NOTES;
						break;
					case I_INDEX_NUMBER:
						temp = R_INDEX_NUMBER;
						break;
				}
				if (temp)
				{
					pData = (char *)&songRMF->resourceData;
					length = pData - (char *)songRMF;
					PV_FillSongResource((SongResourceType)temp, XGetShort(&songRMF->resourceCount), songRMF->locked,
											pData, length, NULL, &size);
				}
				break;
			
			case SONG_TYPE_MOD:
				break;
		}
	}
	return size;
}
#endif	// USE_FULL_RMF_SUPPORT == TRUE

#if USE_FULL_RMF_SUPPORT == TRUE
void XGetSongInformation(SongResource *theSong, long songSize, SongInfo type, char *cName)
{
	char				*pData, *pName;
	long				length, offset;
	unsigned long		temp;
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	if (theSong && cName)
	{
		cName[0] = 0;
		switch (((SongResource_SMS *)theSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)theSong;
				switch (type)
				{
					case I_COMPOSER:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							temp = (*pData) + 1;		// skip copyright text
							pData = (char *)songSMS;
							offset += temp;
							pName = (char *)pData + offset;
							XBlockMove(pName + 1, cName, (long)pName[0]);
							cName[pName[0]] = 0;
						}
						break;
					case I_COPYRIGHT:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							pData = (char *)songSMS;
							pName = (char *)pData + offset;
							XBlockMove(pName + 1, cName, (long)pName[0]);
							cName[pName[0]] = 0;
						}
						break;
					case I_TITLE:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							temp = (*pData) + 1;		// skip copyright text
							offset += temp;
							pData = (char *)songSMS + offset;

							temp = (*pData) + 1;		// skip author text
							offset += temp;
							pData = (char *)songSMS;

							if (offset < length)
							{
								pName = (char *)pData + offset;
								XBlockMove(pName + 1, cName, (long)pName[0]);
								cName[pName[0]] = 0;
							}
						}
						break;
					case I_PUBLISHER_CONTACT:
						// name is a pascal string
						length = songSize;
		
						pData = (char *)&songSMS->remaps;
						pData += sizeof(Remap) * songSMS->remapCount;
						offset = pData - (char *)songSMS;

						// pointing at author text
						if (offset < length)
						{
							temp = (*pData) + 1;		// skip copyright text
							offset += temp;
							pData = (char *)songSMS + offset;

							temp = (*pData) + 1;		// skip author text
							offset += temp;
							pData = (char *)songSMS + offset;

							temp = (*pData) + 1;		// skip title text
							offset += temp;
							pData = (char *)songSMS;

							if (offset < length)
							{
								pName = (char *)pData + offset;
								XBlockMove(pName + 1, cName, (long)pName[0]);
								cName[pName[0]] = 0;
							}
						}
						break;

					case I_LICENSED_TO_URL:
					case I_USE_OF_LICENSE:
					case I_LICENSE_TERM:
					case I_EXPIRATION_DATE:
					case I_COMPOSER_NOTES:
					case I_INDEX_NUMBER:
					case I_PERFORMED_BY:
						XStrCpy(cName, "");
						break;
				}
				break;
			
			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)theSong;
				temp = 0;
				switch (type)
				{
					case I_TITLE:
						temp = R_TITLE;
						break;
					case I_PERFORMED_BY:
						temp = R_PERFORMED_BY;
						break;
					case I_COMPOSER:
						temp = R_COMPOSER;
						break;
					case I_COPYRIGHT:
						temp = R_COPYRIGHT_DATE;
						break;
					case I_PUBLISHER_CONTACT:
						temp = R_PUBLISHER_CONTACT;
						break;
					case I_USE_OF_LICENSE:
						temp = R_USE_OF_LICENSE;
						break;
					case I_LICENSE_TERM:
						temp = R_LICENSE_TERM;
						break;
					case I_LICENSED_TO_URL:
						temp = R_LICENSED_TO_URL;
						break;
					case I_EXPIRATION_DATE:
						temp = R_EXPIRATION_DATE;
						break;
					case I_COMPOSER_NOTES:
						temp = R_COMPOSER_NOTES;
						break;
					case I_INDEX_NUMBER:
						temp = R_INDEX_NUMBER;
						break;
				}
				offset = XGetShort(&songRMF->resourceCount);
				if ( (temp) && (offset) )
				{
					pData = (char *)&songRMF->resourceData;
					length = pData - (char *)songRMF;
					PV_FillSongResource((SongResourceType)temp, offset, songRMF->locked,
											pData, length, cName, NULL);
				}
				break;
			
			case SONG_TYPE_MOD:
				break;
		}

// there should be no bad characters
/*
		// strip out undesirable characters, give them some walking money
		pName = XDuplicateAndStripStr(cName);
		if (pName)
		{
			XStrCpy(cName, pName);	// copy stripped string
			XDisposePtr(pName);
		}
*/
	}
}
#endif	//USE_FULL_RMF_SUPPORT == TRUE


short int XGetSongResourceObjectID(SongResource *pSong)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;
	short int			id;

	id = -1;
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				id = XGetShort(&songSMS->midiResourceID);
				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				id = XGetShort(&songRMF->rmfResourceID);
				break;
		}
	}
	return id;
}

#if USE_CREATION_API == TRUE
void XSetSongResourceObjectID(SongResource *pSong, short int id)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				XPutShort(&songSMS->midiResourceID, id);
				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				XPutShort(&songRMF->rmfResourceID, id);
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE


XBOOL XIsSongLocked(SongResource *pSong)
{
	XBOOL				locked;
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	locked = FALSE;
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				locked = TEST_FLAG_VALUE(songSMS->flags1, XBF_locked);
				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				locked = songRMF->locked;
				break;
		}
	}
	return locked;
}

#if USE_CREATION_API == TRUE
void XSetSongLocked(SongResource *pSong, XBOOL locked)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;
	char				*pData;
	unsigned long		length;
	char				*pUnit;
	short int			count, resourceCount;
	SongResourceType	type;


	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				songSMS->flags1 = SET_FLAG_VALUE(songSMS->flags1, XBF_locked, locked);
				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				
				resourceCount = XGetShort(&songRMF->resourceCount);
				if (resourceCount)
				{
					pData = (char *)&songRMF->resourceData;
					length = pData - (char *)songRMF;
					pUnit = pData;
					for (count = 0; count < resourceCount; count++)
					{
						type = (SongResourceType)XGetLong(pUnit);
						pUnit += sizeof (SongResourceType);

						switch (type)
						{
							default:
								// found a bad type, just stop because we don't know how to proceed
								break;
							case R_TITLE:
							case R_PERFORMED_BY:
							case R_COMPOSER:
							case R_COPYRIGHT_DATE:
							case R_COPYRIGHT_LINE:
							case R_PUBLISHER_CONTACT:
							case R_LICENSED_TO_URL:
							case R_USE_OF_LICENSE:
							case R_LICENSE_TERM:
							case R_EXPIRATION_DATE:
							case R_COMPOSER_NOTES:
							case R_INDEX_NUMBER:
								// this is a zero terminated string, always
								if ( (songRMF->locked == FALSE) && locked)
								{
									length = XStrLen(pUnit) + 1;
									XEncryptData(pUnit, length);
								}
								if (songRMF->locked && (locked == FALSE))
								{
									length = XEncryptedStrLen(pUnit) + 1;
									XDecryptData(pUnit, length);
								}
								if (songRMF->locked && (locked))
								{
									length = XEncryptedStrLen(pUnit) + 1;
								}
								if ((songRMF->locked == FALSE) && (locked == FALSE))
								{
									length = XStrLen(pUnit) + 1;
								}
								pUnit += length;
								break;
							case R_VELOCITY_CURVE:
								pUnit += 256;
								break;
							case R_INSTRUMENT_REMAP:
								length = XGetShort(pUnit);
								pUnit += 2;
								// format is word count followed by x number of (word, word)
								pUnit += sizeof(short) * length;
								break;
						}
					}
				}
				songRMF->locked = locked;
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE

#if USE_CREATION_API == TRUE
XBOOL XGetSongEmbeddedStatus(SongResource *pSong)
{
//	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;
	XBOOL				embedded;

	embedded = FALSE;
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
//			case SONG_TYPE_SMS:
//				songSMS = (SongResource_SMS *)pSong;
//				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				embedded = songRMF->embeddedSong;
				break;
		}
	}
	return embedded;
}
#endif	// USE_CREATION_API == TRUE

#if USE_CREATION_API == TRUE
void XSetSongEmbeddedStatus(SongResource *pSong, XBOOL embedded)
{
//	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
//			case SONG_TYPE_SMS:
//				songSMS = (SongResource_SMS *)pSong;
//				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				songRMF->embeddedSong = embedded;
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE


#if USE_CREATION_API == TRUE
/*
SongResource * XConvertToSongType(SongResource *pSong, SongType newSongType)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;
	short int			id;
	SongResource		*newSong;

	newSong = NULL;
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				id = XGetShort(&songSMS->midiResourceID);
				
				switch (newSongType)
				{
					case SONG_TYPE_SMS:
						newSong = XNewSongPtr(
						break;
					case SONG_TYPE_RMF:
						break;
				}
				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				id = XGetShort(&songRMF->rmfResourceID);
				break;
		}
	}
	return newSong;
}
*/
#endif	// USE_CREATION_API == TRUE



SongType XGetSongResourceObjectType(SongResource *pSong)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;
	SongType			id;

	id = SONG_TYPE_BAD;
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				id = (SongType)songSMS->songType;
				break;

			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				id = (SongType)songRMF->songType;
				break;
		}
	}
	return id;
}

#if USE_FULL_RMF_SUPPORT == TRUE
SongResource_Info * XGetSongResourceInfo(SongResource *pSong, long songSize)
{
	SongResource_Info	*pInfo;
	char				*text;
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	pInfo = NULL;
	text = (char *)XNewPtr(4096L);
	if (pSong && songSize && text)
	{
		pInfo = (SongResource_Info *)XNewPtr((long)sizeof(SongResource_Info));
		if (pInfo)
		{
			switch (((SongResource_SMS *)pSong)->songType)
			{
				case SONG_TYPE_SMS:
					songSMS = (SongResource_SMS *)pSong;
					pInfo->maxMidiNotes = songSMS->maxNotes;
					pInfo->maxEffects = songSMS->maxEffects;
					pInfo->mixLevel = XGetShort(&songSMS->mixLevel);
					pInfo->reverbType = songSMS->reverbType;
					pInfo->objectResourceID = XGetShort(&songSMS->midiResourceID);
					pInfo->songType = (SongType)songSMS->songType;
					pInfo->songTempo = XGetShort(&songSMS->songTempo);
					pInfo->songPitchShift = songSMS->songPitchShift;

					XGetSongInformation(pSong, songSize, I_COMPOSER, text);
					pInfo->composer = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_COPYRIGHT, text);
					pInfo->copyright = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_TITLE, text);
					pInfo->title = XDuplicateStr(text);					
					XGetSongInformation(pSong, songSize, I_PUBLISHER_CONTACT, text);
					pInfo->publisher_contact_info = XDuplicateStr(text);
					pInfo->songLocked = FALSE;
					pInfo->songVolume = MAX_SONG_VOLUME;
					break;

				case SONG_TYPE_RMF:
					songRMF = (SongResource_RMF *)pSong;
					pInfo->maxMidiNotes = XGetShort(&songRMF->maxNotes);
					pInfo->maxEffects = XGetShort(&songRMF->maxEffects);
					pInfo->mixLevel = XGetShort(&songRMF->mixLevel);
					pInfo->reverbType = songRMF->reverbType;
					pInfo->objectResourceID = XGetShort(&songRMF->rmfResourceID);
					pInfo->songType = (SongType)songRMF->songType;
					pInfo->songTempo = XGetShort(&songRMF->songTempo);
					pInfo->songPitchShift = XGetShort(&songRMF->songPitchShift);
					pInfo->songLocked = songRMF->locked;	
					pInfo->songVolume = XGetSongVolume(pSong);

					XGetSongInformation(pSong, songSize, I_TITLE, text);
					pInfo->title = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_PERFORMED_BY, text);
					pInfo->performed = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_COMPOSER, text);
					pInfo->composer = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_COPYRIGHT, text);
					pInfo->copyright = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_PUBLISHER_CONTACT, text);
					pInfo->publisher_contact_info = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_USE_OF_LICENSE, text);
					pInfo->use_license = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_LICENSE_TERM, text);
					pInfo->license_term = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_LICENSED_TO_URL, text);
					pInfo->licensed_to_URL = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_EXPIRATION_DATE, text);
					pInfo->expire_date = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_COMPOSER_NOTES, text);
					pInfo->compser_notes = XDuplicateStr(text);
					XGetSongInformation(pSong, songSize, I_INDEX_NUMBER, text);
					pInfo->index_number = XDuplicateStr(text);
					break;
			}
		}
		XDisposePtr(text);
	}
	return pInfo;
}
#endif

#if USE_CREATION_API == TRUE
// given a valid SongResource_info structure and a type, this will return the string pointer
// from that structure
static char * PV_GetStringItemFromResource_info(SongResource_Info *pSongInfo, SongResourceType type)
{
	char * resourceName;

	resourceName = NULL;
	switch (type)
	{
		case R_TITLE:
			resourceName = pSongInfo->title;
			break;
		case R_COMPOSER:
			resourceName = pSongInfo->composer;
			break;
		case R_COPYRIGHT_DATE:
			resourceName = pSongInfo->copyright;
			break;
		case R_COPYRIGHT_LINE:
			break;
		case R_PUBLISHER_CONTACT:
			resourceName = pSongInfo->publisher_contact_info;
			break;
		case R_USE_OF_LICENSE:
			resourceName = pSongInfo->use_license;
			break;
		case R_LICENSED_TO_URL:
			resourceName = pSongInfo->licensed_to_URL;
			break;
		case R_LICENSE_TERM:
			resourceName = pSongInfo->license_term;
			break;
		case R_EXPIRATION_DATE:
			resourceName = pSongInfo->expire_date;
			break;
		case R_COMPOSER_NOTES:
			resourceName = pSongInfo->compser_notes;
			break;
		case R_INDEX_NUMBER:
			resourceName = pSongInfo->index_number;
			break;
		case R_PERFORMED_BY:
			resourceName = pSongInfo->performed;
			break;
	}
	return resourceName;
}
#endif

// This will create a new song resource from the more accessable SongResource_Info
#if USE_CREATION_API == TRUE
SongResource * XNewSongFromSongResourceInfo(SongResource_Info *pSongInfo)
{
	SongResource 					*newSong, *otherNew;
	SongResource_SMS				*songSMS;
	SongResource_RMF				*songRMF;
	short int						count, max;
	char							*resourceName;
	SongResourceType				type;
	SongResourceType const			*pProcess;
	long							resourceLength;
	static const SongResourceType	sms_processTypes[] = 
				{R_TITLE, R_COMPOSER, R_COPYRIGHT_DATE, R_PUBLISHER_CONTACT};
	static const SongResourceType	rmf_processTypes[] = 
				{R_TITLE, R_COMPOSER, R_COPYRIGHT_DATE, R_COPYRIGHT_LINE, 
				R_PUBLISHER_CONTACT, R_USE_OF_LICENSE, R_LICENSED_TO_URL, 
				R_LICENSE_TERM, R_EXPIRATION_DATE, R_COMPOSER_NOTES, 
				R_INDEX_NUMBER, R_PERFORMED_BY};

	newSong = NULL;
	max = 0;
	if (pSongInfo)
	{
		newSong = XNewSongPtr(pSongInfo->songType,
							pSongInfo->objectResourceID,
							pSongInfo->maxMidiNotes, 
							pSongInfo->mixLevel, 
							pSongInfo->maxEffects,
							pSongInfo->reverbType);
		if (newSong)
		{
			pProcess = NULL;
			switch (pSongInfo->songType)
			{
				case SONG_TYPE_SMS:
					pProcess = sms_processTypes;
					max = sizeof(sms_processTypes) / sizeof(SongResourceType);
					songSMS = (SongResource_SMS *)newSong;
					XPutShort(&songSMS->songTempo, pSongInfo->songTempo);
					songSMS->songPitchShift = pSongInfo->songPitchShift;
					break;
				case SONG_TYPE_RMF:
					songRMF = (SongResource_RMF *)newSong;
					XPutShort(&songRMF->songTempo, pSongInfo->songTempo);
					XPutShort(&songRMF->songPitchShift, pSongInfo->songPitchShift);
					pProcess = rmf_processTypes;
					max = sizeof(rmf_processTypes) / sizeof(SongResourceType);
					XSetSongVolume(newSong, pSongInfo->songVolume);
					break;
			}

			if (pProcess)
			{
				for (count = 0; count < max; count++)
				{
					type = pProcess[count];
					resourceName = PV_GetStringItemFromResource_info(pSongInfo, type);
					resourceLength = XStrLen(resourceName);
					if (resourceLength)
					{
						otherNew = XChangeSongResource(newSong, XGetPtrSize(newSong), 
										type, resourceName, resourceLength + 1);
						if (otherNew)
						{
							XDisposeSongPtr(newSong);
							newSong = otherNew;
						}
					}
				}
			}
		}
	}
	return newSong;
}

void XDisposeSongResourceInfo(SongResource_Info *pSongInfo)
{
	if (pSongInfo)
	{
		XDisposePtr(pSongInfo->remaps);
		XDisposePtr(pSongInfo->velocityCurve);
		XDisposePtr(pSongInfo->title);
		XDisposePtr(pSongInfo->performed);
		XDisposePtr(pSongInfo->composer);
		XDisposePtr(pSongInfo->copyright);
		XDisposePtr(pSongInfo->licensed_to_URL);
		XDisposePtr(pSongInfo->use_license);
		XDisposePtr(pSongInfo->license_term);
		XDisposePtr(pSongInfo->expire_date);
		XDisposePtr(pSongInfo->compser_notes);
		XDisposePtr(pSongInfo->index_number);
		XDisposePtr(pSongInfo->publisher_contact_info);
		XDisposePtr(pSongInfo);
	}
}
#endif	// USE_CREATION_API == TRUE

// Given a song ID and two arrays, this will return the INST resources ID and the 'snd ' resource ID
// that are needed to load the song terminated with a -1.
// Will return 0 for success or 1 for failure
#if USE_CREATION_API == TRUE
OPErr XGetSongInstrumentList(short int theSongID, short int *pInstArray, short int maxInstArraySize, 
										short int *pSndArray, short int maxSndArraySize)
{
	long			count, instCount, sndCount, newCount, completeSndCount, soundID;
	SongResource	*theSong;
	short int		completeSndArray[MAX_SAMPLES];
	short int		completeInstArray[MAX_INSTRUMENTS * MAX_BANKS];
	XBOOL			goodSound;
	OPErr			theErr;
	long			size;

	theErr = NO_ERR;
	if ( (pInstArray) && (pSndArray) )
	{
		for (count = 0; count < maxInstArraySize; count++)
		{
			pInstArray[count] = -1;
			completeInstArray[count] = -1;
		}
		for (count = 0; count < maxSndArraySize; count++)
		{
			pSndArray[count] = -1;
			completeSndArray[count] = -1;
		}
		theSong = (SongResource *)XGetAndDetachResource(ID_SONG, theSongID, &size);
		if (theSong)
		{
			instCount = GM_GetUsedPatchlist(theSong, NULL, 0L, completeInstArray, &theErr);
			if (instCount && (theErr == 0) )
			{
				// remove duplicates in inst
				sndCount = 0;
				for (newCount = 0; newCount < instCount; newCount++)
				{
					goodSound = TRUE;
					soundID = completeInstArray[newCount];
					for (count = 0; count < sndCount; count++)
					{
						if (soundID == pInstArray[count])
						{
							goodSound = FALSE;
							break;
						}
					}
					if (goodSound)
					{
						pInstArray[sndCount++] = soundID;
					}
				}
				instCount = sndCount;
				XBubbleSortArray(pInstArray, instCount);

				completeSndCount = 0;
				for (count = 0; count < instCount; count++)
				{
					newCount = XCollectSoundsFromInstrumentID(pInstArray[count], 
												&completeSndArray[completeSndCount], 128);
					completeSndCount += newCount;
				}
				
				// remove duplicates in snds
				sndCount = 0;
				for (newCount = 0; newCount < completeSndCount; newCount++)
				{
					goodSound = TRUE;
					soundID = completeSndArray[newCount];
					for (count = 0; count < sndCount; count++)
					{
						if (soundID == pSndArray[count])
						{
							goodSound = FALSE;
							break;
						}
					}
					if (goodSound)
					{
						pSndArray[sndCount++] = soundID;
					}
				}
				XBubbleSortArray(pSndArray, sndCount);
			}
			else
			{
				theErr = BAD_MIDI_DATA;
			}
		}
		XDisposePtr((XPTR)theSong);
	}
	return theErr;
}
#endif	// USE_CREATION_API == TRUE


// Global function to get a Midi resource via its various forms:
// ID_ECMI		- compressed and encrypted midi file
// ID_EMID		- encrypted midi file
// ID_CMID		- compressed midi file
// ID_MIDI		- standard midi file
// ID_MIDI_OLD	- standard midi file
XPTR XGetMidiData(long theID, long *pSize)
{
	XPTR	theData, pData;
	long	midiSize;

// try encrypted and compressed midi files
	theData = XGetAndDetachResource(ID_ECMI, theID, &midiSize);
	if (theData)
	{
		// since this is encrypted, make a new copy and decrypt
		pData = XNewPtr(midiSize);
		if (pData)
		{
			XBlockMove(theData, pData, midiSize);
			XDecryptData(pData, midiSize);				// decrypt first
		}
		XDisposePtr(theData);
		theData = pData;
		if (theData)
		{
			pData = XDecompressPtr(theData, midiSize);		// uncompress second
			if (pData)
			{
				midiSize = XGetPtrSize(pData);	// get new size
				XDisposePtr(theData);
				theData = pData;
			}
			else
			{	// something failed
				XDisposePtr(theData);
				theData = NULL;
			}
		}
	}
	else
	{
		theData = XGetAndDetachResource(ID_EMID, theID, &midiSize);
		if (theData)
		{
			// since this is encrypted, make a new copy and decrypt
			pData = XNewPtr(midiSize);
			if (pData)
			{
				XBlockMove(theData, pData, midiSize);
				XDecryptData(pData, midiSize);				// decrypt first
			}
			XDisposePtr(theData);
			theData = pData;
		}
	}

// nothing there, so try compressed midi files
	if (theData == NULL)
	{
		theData = XGetAndDetachResource(ID_CMID, theID, &midiSize);
		if (theData)
		{
			pData = XDecompressPtr(theData, midiSize);
			if (pData)
			{
				XDisposePtr(theData);
				theData = pData;
			}
		}
	}

// still nothing try standard midi files
	if (theData == NULL)
	{
		theData = XGetAndDetachResource(ID_MIDI, theID, &midiSize);
		if (theData == NULL)
		{
			theData = XGetAndDetachResource(ID_MIDI_OLD, theID, &midiSize);
		}
	}
	if (theData)
	{
		*pSize = midiSize;
	}
	return theData;
}

// Get sound resource and detach from resource manager or decompress
// This function can be replaced for a custom sound retriver
XPTR XGetSoundResourceByID(long theID, long *pSize)
{
	XPTR	thePreSound, theData;

	// look for compressed version first
	theData = XGetAndDetachResource(ID_CSND, theID, pSize);
	if (theData == NULL)
	{
		// look for encrypted version
		theData = XGetAndDetachResource(ID_ESND, theID, pSize);
		if (theData)
		{
			// since this is encrypted, make a new copy and decrypt
			thePreSound = theData;
			theData = XNewPtr(*pSize);
			if (theData)
			{
				XBlockMove(thePreSound, theData, *pSize);
			}
			XDisposePtr(thePreSound);
			XDecryptData(theData, *pSize);
		}
		// look for standard version
		if (theData == NULL)
		{
			theData = XGetAndDetachResource(ID_SND, theID, pSize);
		}
	}
	else
	{
		thePreSound = theData;
		theData = XDecompressSampleFormatPtr(thePreSound, *pSize);
		XDisposePtr(thePreSound);
		*pSize = XGetPtrSize(theData);
	}
	return theData;
}

// Get sound resource and detach from resource manager or decompress
// This function can be replaced for a custom sound retriver
XPTR XGetSoundResourceByName(void *cName, long *pSize)
{
	XPTR	thePreSound, theData;

	// look for compressed version first
	theData = XGetNamedResource(ID_CSND, cName, pSize);
	if (theData == NULL)
	{
		// look for standard version
		theData = XGetNamedResource(ID_SND, cName, pSize);

		if (theData == NULL)
		{
			// look for encrypted version
			theData = XGetNamedResource(ID_ESND, cName, pSize);
			if (theData)
			{
				// since this is encrypted, make a new copy and decrypt
				thePreSound = theData;
				theData = XNewPtr(*pSize);
				if (theData)
				{
					XBlockMove(thePreSound, theData, *pSize);
				}
				XDisposePtr(thePreSound);
				XDecryptData(theData, *pSize);
			}
		}
	}
	else
	{
		thePreSound = theData;
		theData = XDecompressSampleFormatPtr(thePreSound, *pSize);
		XDisposePtr(thePreSound);
		*pSize = XGetPtrSize(theData);
	}
	return theData;
}


#if USE_CREATION_API == TRUE
XPTR XCompressAndEncrypt(XPTR pData, long size, long *pNewSize)
{
	XPTR	newData;

	newData = NULL;
	if (pData && pNewSize)
	{
		newData = XCompressPtr(pData, size, pNewSize, FALSE);
		if (newData)
		{
			XEncryptData(newData, *pNewSize);
		}
	}
	return newData;
}
#endif	// USE_CREATION_API == TRUE


#if USE_CREATION_API == TRUE
void XSetSongPerformanceSettings(SongResource *pSong, short int maxMidiVoices, short int maxEffectsVoices,
										short int mixLevel)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				songSMS->maxNotes = maxMidiVoices;
				XPutShort(&songSMS->mixLevel, mixLevel);
				songSMS->maxEffects = maxEffectsVoices;
				break;
			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				XPutShort(&songRMF->maxNotes, maxMidiVoices);
				XPutShort(&songRMF->mixLevel, mixLevel);
				XPutShort(&songRMF->maxEffects, maxEffectsVoices);
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE


#if USE_CREATION_API == TRUE
void XGetSongPerformanceSettings(SongResource * pSong, short int *maxMidiVoices, short int *maxEffectsVoices, short int *mixLevel)
{
	SongResource_SMS	*songSMS;
	SongResource_RMF	*songRMF;

	if (pSong && maxMidiVoices && maxEffectsVoices && mixLevel)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				songSMS = (SongResource_SMS *)pSong;
				*maxMidiVoices = songSMS->maxNotes;
				*mixLevel = XGetShort(&songSMS->mixLevel);
				*maxEffectsVoices = songSMS->maxEffects;
				break;
			case SONG_TYPE_RMF:
				songRMF = (SongResource_RMF *)pSong;
				*maxMidiVoices = XGetShort(&songRMF->maxNotes);
				*mixLevel = XGetShort(&songRMF->mixLevel);
				*maxEffectsVoices = XGetShort(&songRMF->maxEffects);
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE


#if USE_CREATION_API == TRUE
long XGetSongTempoFactor(SongResource *pSong)
{
	long			tempo;

	tempo = 16667;	// 1.0
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				tempo = XGetShort(&((SongResource_SMS *)pSong)->songTempo);
				break;
			case SONG_TYPE_RMF:
				tempo = XGetShort(&((SongResource_RMF *)pSong)->songTempo);
				break;
		}
		if (tempo == 0)
		{
			tempo = 16667;
		}
	}
	return tempo;
}
#endif	// USE_CREATION_API == TRUE


#if USE_CREATION_API == TRUE
void XSetSongTempoFactor(SongResource *pSong, long newTempo)
{
	if (pSong)
	{
		if (newTempo == 16667L)
		{
			newTempo = 0;
		}
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				XPutShort(&((SongResource_SMS *)pSong)->songTempo, newTempo);
				break;
			case SONG_TYPE_RMF:
				XPutShort(&((SongResource_RMF *)pSong)->songTempo, newTempo);
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE


short int XGetSongVolume(SongResource *pSong)
{
	short int	volume;

	volume = MAX_SONG_VOLUME;
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_RMF:
				volume = XGetShort(&((SongResource_RMF *)pSong)->songVolume);
				if (volume == 0)
				{
					volume = MAX_SONG_VOLUME;
				}
				break;
		}
	}
	return volume;
}

#if USE_CREATION_API == TRUE
void XSetSongVolume(SongResource *pSong, short int volume)
{
	if (pSong)
	{
		if (volume == MAX_SONG_VOLUME)
		{
			volume = 0;
		}
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_RMF:
				XPutShort(&((SongResource_RMF *)pSong)->songVolume, volume);
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE


short int XGetSongReverbType(SongResource *pSong)
{
	short int		reverbType;

	reverbType = 0;
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				reverbType = ((SongResource_SMS *)pSong)->reverbType;
				break;
			case SONG_TYPE_RMF:
				reverbType = ((SongResource_RMF *)pSong)->reverbType;
				break;
		}
	}
	return reverbType;
}

#if USE_CREATION_API == TRUE
void XSetSongReverbType(SongResource *pSong, short int reverbType)
{
	if (pSong)
	{
		switch (((SongResource_SMS *)pSong)->songType)
		{
			case SONG_TYPE_SMS:
				((SongResource_SMS *)pSong)->reverbType = reverbType;
				break;
			case SONG_TYPE_RMF:
				((SongResource_RMF *)pSong)->reverbType = reverbType;
				break;
		}
	}
}
#endif	// USE_CREATION_API == TRUE

#if USE_FULL_RMF_SUPPORT == TRUE
#if USE_CREATION_API == TRUE
XPTR XCreateBankStatus(BankStatus *pStatus)
{
	char			*pBank;
	long			size;

	pBank = NULL;
	if (pStatus)
	{
		size = XStrLen(pStatus->bankURL) + 1;
		size += XStrLen(pStatus->bankName) + 1;
		size += sizeof(long);
		pBank = (char *)XNewPtr(size);
		if (pBank)
		{
			XPutLong(pBank, pStatus->version);
			pBank += sizeof(long);
			size = XStrLen(pStatus->bankURL) + 1;
			if (size > 1)
			{
				XBlockMove(pStatus->bankURL, pBank, size);
			}
			pBank += size;
			size = XStrLen(pStatus->bankName);
			if (size > 1)
			{
				XBlockMove(pStatus->bankName, pBank, size);
			}
		}
	}
	return pBank;
}
#endif	// USE_CREATION_API == TRUE

void XGetBankStatus(BankStatus *pStatus)
{
	char			*pBank;
	long			size;

	if (pStatus)
	{
		XSetMemory((XPTR)pStatus, (long)sizeof(BankStatus), 0);
		pBank = (char *)XGetAndDetachResource(ID_BANK, DEFAULT_RESOURCE_BANK_ID, &size);
		if (pBank)
		{
			pStatus->version = XGetLong(pBank);		// get version

			pBank += sizeof(long);					// get bank URL
			size = XStrLen(pBank) + 1;
			if (size > 1023)
			{
				size = 1023;
			}
			XBlockMove(pBank, pStatus->bankURL, size);

			pBank += size;							// get bank name
			size = XStrLen(pBank) + 1;
			if (size > 1023)
			{
				size = 1023;
			}
			XBlockMove(pBank, pStatus->bankName, size);
		}
		XDisposePtr((XPTR)pBank);
	}
}
#endif	// USE_FULL_RMF_SUPPORT == TRUE

/* EOF of DriverTools.c
*/
