/*
 * @(#)GenOutput.c	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
** "GenOutput.c"
**
**	Generalized Music Synthesis package. Part of SoundMusicSys.
**	Confidential-- Internal use only
**
**
** Modification History:
**
**	4/21/96		Removed register usage in parameters
**	11/21/96	Added PHASE_OFFSET macro to allow for zero based 8 bit DACs
**	12/30/96	Changed copyright
**	1/22/97		Added support for MOD code in all output loops
**	6/10/97		Moved MOD support code to GenSynth.c prior to output stage
**	7/28/97		Merged Sun timing test code
**	11/10/97	Flipped phase of 8 bit output for Solaris
*/
/*****************************************************************************/

/* header for gethrvtime()					Liang */
#ifdef VIS_TIMING
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#define NITER 1000000
#endif

#ifdef USE_VIS
#include "VIS_JavaSound.h"
#endif

#include "GenSnd.h"
#include "GenPriv.h"

// This is 8 bit phase. Some hardware wants silence to be 0, some wants it
// to be 128.
#if ( (X_PLATFORM == X_BE) 				||	\
	  (X_PLATFORM == X_SOLARIS) )
	#define PHASE_OFFSET	0		// silence is 0
#else
	#define PHASE_OFFSET	0x80	// silence is 128
#endif

#if (USE_8_BIT_OUTPUT == TRUE) && (USE_STEREO_OUTPUT == TRUE)
void PV_Generate8outputStereo(OUTSAMPLE8 FAR * dest8)
{
	register INT32			b, c;
	register INT32 *sourceL;
	register LOOPCOUNT		a;

#ifdef VIS_TIMING 
{
  INT32      *sourceLsave = sourceL;
  OUTSAMPLE8 *dest8save   = dest8;
  hrtime_t start, end;
  double   vtime;
  int      k;
  start = gethrvtime();
  for (k = 0; k < NITER; k ++) {
    sourceL = sourceLsave;
    dest8   = dest8save;
#endif

	/* Convert intermediate 16-bit sample format to 16 bit output samples:
	*/
	sourceL = &MusicGlobals->songBufferLeftMono[0];

	if ( (MusicGlobals->outputQuality == Q_22K) || (MusicGlobals->outputQuality == Q_44K))
	{
	// 22k & 44k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_ShiftOffsetDuplicate_U8_S32_Sat(dest8, (int*)sourceL, MusicGlobals->Four_Loop<<3, OUTPUT_SCALAR+8, PHASE_OFFSET);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			dest8[0] = (sourceL[0] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[1] = (sourceL[1] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[2] = (sourceL[2] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[3] = (sourceL[3] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[4] = (sourceL[4] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[5] = (sourceL[5] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[6] = (sourceL[6] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[7] = (sourceL[7] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8 += 8;
			sourceL += 8;
		}
#endif
	}
	else
	{                   
		// 11k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_ShiftOffsetDuplicate_U8_S32_Sat(dest8, (int*)sourceL, MusicGlobals->Four_Loop<<3, OUTPUT_SCALAR+8, PHASE_OFFSET);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			b =  (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			c =  (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[0] = b;
			dest8[1] = c;
			dest8[2] = b;
			dest8[3] = c;

			b = (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			c = (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[4] = b;
			dest8[5] = c;
			dest8[6] = b;
			dest8[7] = c;

			b = (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			c = (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[8] = b;
			dest8[9] = c;
			dest8[10] = b;
			dest8[11] = c;

			b = (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			c = (*sourceL++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[12] = b;
			dest8[13] = c;
			dest8[14] = b;
			dest8[15] = c;
			dest8 += 16;
		}
#endif
	}

#ifdef VIS_TIMING 
  }
  end = gethrvtime();
  vtime = (double) (end - start);
  printf("\n");
  printf("PV_Generate8outputStereo\n");
  printf("   Samples = %3d\n", MusicGlobals->Four_Loop<<2);
  switch (MusicGlobals->outputQuality) {
    case Q_44K: printf("   Quality =  44 KHz\n"); break;
    case Q_22K: printf("   Quality =  22 KHz\n"); break;
    case Q_11K: printf("   Quality =  11 KHz\n"); break;
  }
  printf("     Loops =  %d\n", NITER);
#ifdef USE_VIS
  printf("  VIS Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("  VIS Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#else
  printf("    C Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("    C Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#endif
  printf("\n");
  exit(2);
}
#endif
}
#endif	// (USE_8_BIT_OUTPUT == TRUE) && (USE_STEREO_OUTPUT == TRUE)

#if (USE_8_BIT_OUTPUT == TRUE) && (USE_MONO_OUTPUT == TRUE)
void PV_Generate8outputMono(OUTSAMPLE8 FAR * dest8)
{
	register INT32	b;
	register INT32 *source;
	register LOOPCOUNT		a;

	/* Convert intermediate 16-bit sample format to 16 bit output samples:
	*/
	source = &MusicGlobals->songBufferLeftMono[0];

#ifdef VIS_TIMING 
{
  INT32      *sourcesave = source;
  OUTSAMPLE8 *dest8save  = dest8;
  hrtime_t start, end;
  double   vtime;
  int      k;
  start = gethrvtime();
  for (k = 0; k < NITER; k ++) {
    source = sourcesave;
    dest8  = dest8save;
#endif
	/* Here's how to add a buzz if we want to make sure the Sound Manager's alive.
	*/
//		source[0] += 0x60;
	if ( (MusicGlobals->outputQuality == Q_22K) || (MusicGlobals->outputQuality == Q_44K))
	{
		// 22k & 44k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_ShiftOffset_U8_S32_Sat(dest8, (int*)source, MusicGlobals->Four_Loop<<2, OUTPUT_SCALAR+8, PHASE_OFFSET);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			dest8[0] = (source[0] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[1] = (source[1] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[2] = (source[2] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[3] = (source[3] >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			source += 4;
			dest8 += 4;
		}
#endif
	}
	else
	{                   
		// 11k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_ShiftOffsetDuplicate_U8_S32_Sat(dest8, (int*)source, MusicGlobals->Four_Loop<<2, OUTPUT_SCALAR+8, PHASE_OFFSET);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			b = (*source++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[0] = b;
			dest8[1] = b;

			b = (*source++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[2] = b;
			dest8[3] = b;

			b = (*source++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[4] = b;
			dest8[5] = b;

			b = (*source++ >> (OUTPUT_SCALAR + 8)) + PHASE_OFFSET;
			dest8[6] = b;
			dest8[7] = b;
			dest8 += 8;
		}
#endif
	}

#ifdef VIS_TIMING 
  }
  end = gethrvtime();
  vtime = (double) (end - start);
  printf("\n");
  printf("PV_Generate8outputMono\n");
  printf("   Samples = %3d\n", MusicGlobals->Four_Loop<<2);
  switch (MusicGlobals->outputQuality) {
    case Q_44K: printf("   Quality =  44 KHz\n"); break;
    case Q_22K: printf("   Quality =  22 KHz\n"); break;
    case Q_11K: printf("   Quality =  11 KHz\n"); break;
  }
  printf("     Loops =  %d\n", NITER);
#ifdef USE_VIS
  printf("  VIS Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("  VIS Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#else
  printf("    C Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("    C Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#endif
  printf("\n");
  exit(2);
}
#endif
}
#endif	// (USE_8_BIT_OUTPUT == TRUE) && (USE_MONO_OUTPUT == TRUE)

#if (USE_16_BIT_OUTPUT == TRUE) && (USE_STEREO_OUTPUT == TRUE)
void PV_Generate16outputStereo(OUTSAMPLE16 FAR * dest16)
{
	register INT32		b, c;
	register INT32 *sourceL;
	register LOOPCOUNT		a;
	register INT32 i, overflow_test, k8000;

	/* Convert intermediate 16-bit sample format to 16 bit output samples:
	*/
#ifdef VIS_TIMING 
{
  INT32       *sourceLsave = sourceL;
  OUTSAMPLE16 *dest16save  = dest16;
  hrtime_t start, end;
  double   vtime;
  int      k;
  start = gethrvtime();
  for (k = 0; k < NITER; k ++) {
    sourceL = sourceLsave;
    dest16  = dest16save;
#endif
	sourceL = &MusicGlobals->songBufferLeftMono[0];

	k8000 = 0x8000;
	if ( (MusicGlobals->outputQuality == Q_22K) || (MusicGlobals->outputQuality == Q_44K))
	{
		// 22k & 44k
		// 22k & 44k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_Shift_S16_S32_Sat(dest16, (int*)sourceL, MusicGlobals->Four_Loop<<3, OUTPUT_SCALAR);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			i = sourceL[0] >> OUTPUT_SCALAR;
			overflow_test = i + k8000;
			dest16[0] = i;
			i = sourceL[1] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[1] = i;

			i = sourceL[2] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[2] = i;
			i = sourceL[3] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[3] = i;

			i = sourceL[4] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[4] = i;
			i = sourceL[5] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[5] = i;

			i = sourceL[6] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[6] = i;
			i = sourceL[7] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[7] = i;

			if (overflow_test & 0xFFFF0000)
			{
				i = sourceL[0] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[0] = i - k8000;
				i = sourceL[1] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[1] = i - k8000;
	
				i = sourceL[2] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[2] = i - k8000;
				i = sourceL[3] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[3] = i - k8000;
	
				i = sourceL[4] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[4] = i - k8000;
				i = sourceL[5] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[5] = i - k8000;
	
				i = sourceL[6] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[6] = i - k8000;
				i = sourceL[7] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[7] = i - k8000;
			}
			sourceL += 8; dest16 += 8;
		}
#endif
	}
	else
	{                   
		// 11k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_ShiftDuplicate_S16_S32_Sat(dest16, (int*)sourceL, MusicGlobals->Four_Loop<<3, OUTPUT_SCALAR);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			b = *sourceL++ >> OUTPUT_SCALAR;
			b += k8000;
			if (b & 0xFFFF0000)
				{ if (b > 0) b = 0xFFFF; else b = 0;}
			b -= k8000;
			c = *sourceL++ >> OUTPUT_SCALAR;
			c += k8000;
			if (c & 0xFFFF0000)
				{ if (c > 0) c = 0xFFFF; else c = 0;}
			c -= k8000;
			dest16[0] = b;
			dest16[1] = c; 
			dest16[2] = b;
			dest16[3] = c; 

			b = *sourceL++ >> OUTPUT_SCALAR;
			b += k8000;
			if (b & 0xFFFF0000)
				{ if (b > 0) b = 0xFFFF; else b = 0;}
			b -= k8000;
			c = *sourceL++ >> OUTPUT_SCALAR;
			c += k8000;
			if (c & 0xFFFF0000)
				{ if (c > 0) c = 0xFFFF; else c = 0;}
			c -= k8000;
			dest16[4] = b;
			dest16[5] = c; 
			dest16[6] = b;
			dest16[7] = c; 

			b = *sourceL++ >> OUTPUT_SCALAR;
			b += k8000;
			if (b & 0xFFFF0000)
				{ if (b > 0) b = 0xFFFF; else b = 0;}
			b -= k8000;
			c = *sourceL++ >> OUTPUT_SCALAR;
			c += k8000;
			if (c & 0xFFFF0000)
				{ if (c > 0) c = 0xFFFF; else c = 0;}
			c -= k8000;
			dest16[8] = b;
			dest16[9] = c; 
			dest16[10] = b;
			dest16[11] = c; 

			b = *sourceL++ >> OUTPUT_SCALAR;
			b += k8000;
			if (b & 0xFFFF0000)
				{ if (b > 0) b = 0xFFFF; else b = 0;}
			b -= k8000;

			c = *sourceL++ >> OUTPUT_SCALAR;
			c += k8000;
			if (c & 0xFFFF0000)
				{ if (c > 0) c = 0xFFFF; else c = 0;}
			c -= k8000;
			dest16[12] = b;
			dest16[13] = c; 
			dest16[14] = b;
			dest16[15] = c; 

			dest16 += 16;
		}
#endif
	}

#ifdef VIS_TIMING 
  }
  end = gethrvtime();
  vtime = (double) (end - start);
  printf("\n");
  printf("PV_Generate16outputStereo\n");
  printf("   Samples = %3d\n", MusicGlobals->Four_Loop<<2);
  switch (MusicGlobals->outputQuality) {
    case Q_44K: printf("   Quality =  44 KHz\n"); break;
    case Q_22K: printf("   Quality =  22 KHz\n"); break;
    case Q_11K: printf("   Quality =  11 KHz\n"); break;
  }
  printf("     Loops =  %d\n", NITER);
#ifdef USE_VIS
  printf("  VIS Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("  VIS Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#else
  printf("    C Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("    C Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#endif
  printf("\n");
  exit(2);
}
#endif
}
#endif	// (USE_16_BIT_OUTPUT == TRUE) && (USE_STEREO_OUTPUT == TRUE)

#if (USE_16_BIT_OUTPUT == TRUE) && (USE_MONO_OUTPUT == TRUE)
void PV_Generate16outputMono(OUTSAMPLE16 FAR * dest16)
{
	register INT32 i;
	register INT32 *source;
	register LOOPCOUNT a;
	register INT32 overflow_test;
	register INT32 k8000 = 0x8000;

	/* Convert intermediate 16-bit sample format to 16 bit output samples:
	*/
#ifdef VIS_TIMING 
{
  INT32       *sourcesave = source;
  OUTSAMPLE16 *dest16save = dest16;
  hrtime_t start, end;
  double   vtime;
  int      k;
  start = gethrvtime();
  for (k = 0; k < NITER; k ++) {
    source = sourcesave;
    dest16 = dest16save;
#endif
	source = &MusicGlobals->songBufferLeftMono[0];

	/* Here's how to add a buzz if we want to make sure the Sound Manager's alive.
	*/
//		source[0] += 0x60;
	if ( (MusicGlobals->outputQuality == Q_22K) || (MusicGlobals->outputQuality == Q_44K))
	{
		// 22k & 44k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_Shift_S16_S32_Sat(dest16, (int*)source, MusicGlobals->Four_Loop<<2, OUTPUT_SCALAR);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			i = source[0] >> OUTPUT_SCALAR;
			overflow_test = i + k8000;
			dest16[0] = i;

			i = source[1] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[1] = i;

			i = source[2] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[2] = i;

			i = source[3] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[3] = i;

			if (overflow_test & 0xFFFF0000)
			{
				i = source[0] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[0] = i - k8000;
	
				i = source[1] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[1] = i - k8000;
	
				i = source[2] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[2] = i - k8000;
	
				i = source[3] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[3] = i - k8000;
			}
			source += 4; dest16 += 4;
		}
#endif
	}
	else
	{                   
		// 11k
#ifdef USE_VIS
		// use VIS version.				-Liang
		VIS_ShiftDuplicate_S16_S32_Sat(dest16, (int*)source, MusicGlobals->Four_Loop<<2, OUTPUT_SCALAR);
#else
		for (a = MusicGlobals->Four_Loop; a > 0; --a)
		{
			i = source[0] >> OUTPUT_SCALAR;
			overflow_test = i + k8000;
			dest16[0] = i;
			dest16[1] = i;

			i = source[1] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[2] = i;
			dest16[3] = i;

			i = source[2] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[4] = i;
			dest16[5] = i;

			i = source[3] >> OUTPUT_SCALAR;
			overflow_test |= i + k8000;
			dest16[6] = i;
			dest16[7] = i;

			if (overflow_test & 0xFFFF0000)
			{
				i = source[0] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[0] = i - k8000;
				dest16[1] = i - k8000;
	
				i = source[1] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[2] = i - k8000;
				dest16[3] = i - k8000;
	
				i = source[2] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[4] = i - k8000;
				dest16[5] = i - k8000;
	
				i = source[3] >> OUTPUT_SCALAR;
				i += k8000;
				if (i & 0xFFFF0000)
					{ if (i > 0) i = 0xFFFF; else i = 0;}
				dest16[6] = i - k8000;
				dest16[7] = i - k8000;
			}
			source += 4; dest16 += 8;
		}
#endif
	}

#ifdef VIS_TIMING 
  }
  end = gethrvtime();
  vtime = (double) (end - start);
  printf("\n");
  printf("PV_Generate16outputMono\n");
  printf("   Samples = %3d\n", MusicGlobals->Four_Loop<<2);
  switch (MusicGlobals->outputQuality) {
    case Q_44K: printf("   Quality =  44 KHz\n"); break;
    case Q_22K: printf("   Quality =  22 KHz\n"); break;
    case Q_11K: printf("   Quality =  11 KHz\n"); break;
  }
  printf("     Loops =  %d\n", NITER);
#ifdef USE_VIS
  printf("  VIS Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("  VIS Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("  VIS Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#else
  printf("    C Time = %10.6f msec\n", 
          vtime/((double)NITER*1000000.0));
  printf("    C Time = %10.6f nsec/sample\n", 
          vtime/((double)NITER*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f cycles/sample (on Ultra 1/170)\n",
          vtime/((double)NITER*6*(MusicGlobals->Four_Loop<<2)));
  printf("    C Time = %10.6f sec (total)\n", vtime/(1000000000.0));
#endif
  printf("\n");
  exit(2);
}
#endif
}
#endif	// (USE_16_BIT_OUTPUT == TRUE) && (USE_MONO_OUTPUT == TRUE)

// EOF of GenOutput.c

