/*
 * @(#)GenReverb.c	1.6 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
** "GenReverb.c"
**
**	Generalized Music Synthesis package. Part of SoundMusicSys.
**
**
**
** Modification History:
**
**	1/18/96		Spruced up for C++ extra error checking
**				Changed the macro 'THE_CHECK' to accept a type for typecasting the source pointer
**	3/1/96		Removed extra PV_DoCallBack, and PV_GetWavePitch
**	4/21/96		Put test around ReverbBuffer being NULL
**	12/30/96	Changed copyright
**	1/23/97		Added WebTV small footprint reverb unit, and support for PV_PostFilterStereo
**				code
**				Fixed REVERB_TYPE_6 to match be the same amount of spring for all output
**				rates
**	6/4/97		Added USE_SMALL_MEMORY_REVERB tests around code to disable when this
**				flag is used
*/
/*****************************************************************************/
#include "GenSnd.h"
#include "GenPriv.h"

#define K_REVERBMONO 0
#define K_REVERBSTEREO 1

#if ((USE_SMALL_MEMORY_REVERB == FALSE) && (USE_VARIABLE_REVERB != FALSE))
static void PV_ProcessVariableReverb (int stereo, int tap_offset)
{
	register INT32 *sourceL, *sourceReverb, src;
	register INT32 *reverbBuf, *reverbBuf2, *reverbBuf3;
	register LOOPCOUNT a;
	register long reverbPtr1, reverbPtr2, reverbPtr3, reverbPtr4, reverbPtr5, reverbPtr6, reverbPtr7;
	register long scale_a, scale_b, scale_c, scale_d, scale_e, scale_f;
	INT32 b, c, cz1, cz2, reflection, d, dz1, dz2, e, ez1, ez2;

	scale_a = MusicGlobals->reverbTaps[9+tap_offset*12];
	scale_b = MusicGlobals->reverbTaps[10+tap_offset*12];
	scale_c = MusicGlobals->reverbTaps[11+tap_offset*12];
	scale_d = MusicGlobals->reverbTaps[6+tap_offset*12];
	scale_e = MusicGlobals->reverbTaps[7+tap_offset*12];
	scale_f = MusicGlobals->reverbTaps[8+tap_offset*12];

	reverbBuf = &MusicGlobals->reverbBuffer[0];
	if (!reverbBuf) return;
	reverbBuf2 = &MusicGlobals->reverbBuffer2[0];
	if (!reverbBuf2) return;
	reverbBuf3 = &MusicGlobals->reverbBuffer3[0];
	if (!reverbBuf3) return;
	sourceL = &MusicGlobals->songBufferLeftMono[0];
	sourceReverb = &MusicGlobals->songBufferReverb[0];

	c = MusicGlobals->LPfilter_c0[tap_offset+1];
	cz1 = MusicGlobals->LPfilter_c1[tap_offset+1];
	cz2 = MusicGlobals->LPfilter_c2[tap_offset+1];
	d = MusicGlobals->LPfilter_d0[tap_offset+1];
	dz1 = MusicGlobals->LPfilter_d1[tap_offset+1];
	dz2 = MusicGlobals->LPfilter_d2[tap_offset+1];
	e = MusicGlobals->LPfilter_e0[tap_offset+1];
	ez1 = MusicGlobals->LPfilter_e1[tap_offset+1];
	ez2 = MusicGlobals->LPfilter_e2[tap_offset+1];
	reverbPtr1 = MusicGlobals->reverbPtr;


#define ZKRVB2_A (MusicGlobals->reverbTaps[0+tap_offset*12])
#define ZKRVB2_B (MusicGlobals->reverbTaps[1+tap_offset*12])
#define ZKRVB2_C (MusicGlobals->reverbTaps[2+tap_offset*12])
#define ZKRVB2_D (MusicGlobals->reverbTaps[3+tap_offset*12])
#define ZKRVB2_E (MusicGlobals->reverbTaps[4+tap_offset*12])
#define ZKRVB2_F (MusicGlobals->reverbTaps[5+tap_offset*12])



	sourceReverb = MusicGlobals->songBufferReverb;
	switch (MusicGlobals->outputQuality)
	{
		case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - ZKRVB2_A); 
					reverbPtr3 = (MusicGlobals->reverbPtr - ZKRVB2_B);
					reverbPtr4 = (MusicGlobals->reverbPtr - ZKRVB2_C);
					reverbPtr5 = (MusicGlobals->reverbPtr - ZKRVB2_D);
					reverbPtr6 = (MusicGlobals->reverbPtr - ZKRVB2_E); 
					reverbPtr7 = (MusicGlobals->reverbPtr - ZKRVB2_F);
					break;
		case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - ZKRVB2_A*2); 
					reverbPtr3 = (MusicGlobals->reverbPtr - ZKRVB2_B*2); 
					reverbPtr4 = (MusicGlobals->reverbPtr - ZKRVB2_C*2);
					reverbPtr5 = (MusicGlobals->reverbPtr - ZKRVB2_D*2);
					reverbPtr6 = (MusicGlobals->reverbPtr - ZKRVB2_E*2); 
					reverbPtr7 = (MusicGlobals->reverbPtr - ZKRVB2_F*2);
					break;
		case Q_44K:
					reverbPtr2 = (MusicGlobals->reverbPtr - ZKRVB2_A*4); 
					reverbPtr3 = (MusicGlobals->reverbPtr - ZKRVB2_B*4);
					reverbPtr4 = (MusicGlobals->reverbPtr - ZKRVB2_C*4);
					reverbPtr5 = (MusicGlobals->reverbPtr - ZKRVB2_D*4);
					reverbPtr6 = (MusicGlobals->reverbPtr - ZKRVB2_E*4); 
					reverbPtr7 = (MusicGlobals->reverbPtr - ZKRVB2_F*4);
					break;
	}

	for (a = 0; a < MusicGlobals->One_Loop; a++)
	{
		reverbPtr1 = (reverbPtr1 + 1) & REVERB_BUFFER_MASK_SHORT;
		reverbPtr2 = (reverbPtr2 + 1) & REVERB_BUFFER_MASK_SHORT;
		reverbPtr3 = (reverbPtr3 + 1) & REVERB_BUFFER_MASK_SHORT;
		reverbPtr4 = (reverbPtr4 + 1) & REVERB_BUFFER_MASK_SHORT;
		reverbPtr5 = (reverbPtr5 + 1) & REVERB_BUFFER_MASK_SHORT;
		reverbPtr6 = (reverbPtr6 + 1) & REVERB_BUFFER_MASK_SHORT;
		reverbPtr7 = (reverbPtr7 + 1) & REVERB_BUFFER_MASK_SHORT;
		src = *sourceReverb;
#define K_REVERB_SCALAR 12
		if (tap_offset)
		{
			cz2 = cz1; cz1 = c;
			c = ((reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3]) >> K_REVERB_SCALAR);
			reflection = c + cz1 + cz1 + cz2;
			b = reflection * scale_a;
			reverbBuf[reverbPtr1] += reflection * scale_d + src;
			dz2 = dz1; dz1 = d;
			d = ((reverbBuf2[reverbPtr4] + reverbBuf2[reverbPtr5]) >> K_REVERB_SCALAR);
			reflection = d + dz1 + dz1 + dz2;
			b += reflection * scale_b;
			reverbBuf2[reverbPtr1] += reflection * scale_e + src;
			ez2 = ez1; ez1 = e;
			e = ((reverbBuf3[reverbPtr6] + reverbBuf3[reverbPtr7]) >> K_REVERB_SCALAR);
			reflection = e + ez1 + ez1 + ez2;
			b += reflection * scale_c;
			reverbBuf3[reverbPtr1] += reflection * scale_f + src;
		}
		else
		{
			cz2 = cz1; cz1 = c;
			c = ((reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3]) >> K_REVERB_SCALAR);
			reflection = c + cz1 + cz1 + cz2;
			b = reflection * scale_a;
			reverbBuf[reverbPtr1] = reflection * scale_d + src;
			dz2 = dz1; dz1 = d;
			d = ((reverbBuf2[reverbPtr4] + reverbBuf2[reverbPtr5]) >> K_REVERB_SCALAR);
			reflection = d + dz1 + dz1 + dz2;
			b += reflection * scale_b;
			reverbBuf2[reverbPtr1] = reflection * scale_e + src;
			ez2 = ez1; ez1 = e;
			e = ((reverbBuf3[reverbPtr6] + reverbBuf3[reverbPtr7]) >> K_REVERB_SCALAR);
			reflection = e + ez1 + ez1 + ez2;
			b += reflection * scale_c;
			reverbBuf3[reverbPtr1] = reflection * scale_f + src;
		}
		
		if (!stereo)
		{
			*sourceL += b;
			sourceL++;
		}
		else
		{
			*sourceL += b;
			sourceL++;
			*sourceL += b;
			sourceL++;
		}
		sourceReverb++;
	}
	MusicGlobals->LPfilter_c0[tap_offset+1] = c;
	MusicGlobals->LPfilter_c1[tap_offset+1] = cz1;
	MusicGlobals->LPfilter_c2[tap_offset+1] = cz2;
	MusicGlobals->LPfilter_d0[tap_offset+1] = d;
	MusicGlobals->LPfilter_d1[tap_offset+1] = dz1;
	MusicGlobals->LPfilter_d2[tap_offset+1] = dz2;
	MusicGlobals->LPfilter_e0[tap_offset+1] = e;
	MusicGlobals->LPfilter_e1[tap_offset+1] = ez1;
	MusicGlobals->LPfilter_e2[tap_offset+1] = ez2;
}
#endif

void PV_ProcessReverbMono(void)
{
	register INT32 b, c, bz, cz;
	register INT32 *sourceL, *sourceReverb;
	register INT32 *reverbBuf;
	register LOOPCOUNT a;
	register long reverbPtr1, reverbPtr2, reverbPtr3, reverbPtr4;

	reverbBuf = &MusicGlobals->reverbBuffer[0];
	if (reverbBuf)
	{
		sourceL = &MusicGlobals->songBufferLeftMono[0];
		sourceReverb = &MusicGlobals->songBufferReverb[0];
	
		b = MusicGlobals->LPfilterL[0];
		c = MusicGlobals->LPfilterR[0];
		bz = MusicGlobals->LPfilterLz[0];
		cz = MusicGlobals->LPfilterRz[0];
		reverbPtr1 = MusicGlobals->reverbPtr;

	// NOTE: do not exceed a tap distance of 2730.  2730*6 is almost the full buffer size.
		switch (MusicGlobals->reverbUnitType)
		{
			case REVERB_TYPE_1:
				break;
#if USE_SMALL_MEMORY_REVERB == FALSE
	// ----------
#if USE_VARIABLE_REVERB != FALSE
			case REVERB_TYPE_8:
//				if (*((short *) 0x17a))
//					for (a = 0; a < MusicGlobals->One_Loop*2; a++)
//						sourceL[a] = 0;
				PV_ProcessVariableReverb(K_REVERBMONO, 0);
				PV_ProcessVariableReverb(K_REVERBMONO, 1);
				reverbPtr1 += MusicGlobals->One_Loop;		
				break;
#endif
	// ----------
			case REVERB_TYPE_3:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 632*1) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 450*1) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 798*1) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 632*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 450*2) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 798*2) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 632*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 450*4) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 798*4) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= b >> 2;
					b += (reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 3;
					reverbBuf[reverbPtr1] = *sourceL + (b >> 1);
					*sourceL += b >> 1;
					sourceL++;
					
					reverbPtr1 = (reverbPtr1 + 1) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 1) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 1) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 1) & REVERB_BUFFER_MASK;
				}
				break;
	// ----------
			case REVERB_TYPE_4:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 1100*1) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 1473*1) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1711*1) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 1100*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 1473*2) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 1711*2) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 1100*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 1473*4) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1711*4) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= (bz + b) >> 2;
					bz = b;
					b += (reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 3;
					reverbBuf[reverbPtr1] = *sourceL + b - (b >> 2);
					*sourceL = *sourceL +  (b >> 1);
					sourceL++;
					
					
					reverbPtr1 = (reverbPtr1 + 1) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 1) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 1) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 1) & REVERB_BUFFER_MASK;
				}
				break;
	// ----------
			case REVERB_TYPE_5:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 500*1) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 674*1) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1174*1) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 1200*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 500*2) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 1174*2) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 500*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 674*4) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1174*4) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= b >> 1;
					b = (reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 2;
					reverbBuf[reverbPtr1] = *sourceL + b - (b >> 2);
					*sourceL = (*sourceL + (b << 2)) >> 1;
					sourceL++;
	
					reverbPtr1 = (reverbPtr1 + 1) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 1) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 1) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 1) & REVERB_BUFFER_MASK;
				}
				break;
	// ----------
			case REVERB_TYPE_6:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 2700*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 3250*2) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 4095*2) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 2700*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 3250*4) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 4095*4) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 2700*8) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 3250*8) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 4095*8) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b += ((reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 4) - (b >> 2);
					reverbBuf[reverbPtr1] = *sourceL + b - (b >> 8);
					*sourceL = *sourceL + b;
					sourceL++;
	
					reverbPtr1 = (reverbPtr1 + 1) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 1) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 1) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 1) & REVERB_BUFFER_MASK;
				}
				break;
#else
			default:
#endif
	// ----------
			case REVERB_TYPE_7:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K:
						a = 1;
						break;
					case Q_22K:
						a = 2;
						break;
					case Q_44K:
						a = 4;
						break;
				}
				reverbPtr1 = MusicGlobals->reverbPtr & REVERB_BUFFER_MASK;
				reverbPtr2 = (MusicGlobals->reverbPtr - 1100*a) & REVERB_BUFFER_MASK; 
				reverbPtr3 = (MusicGlobals->reverbPtr - 1473*a) & REVERB_BUFFER_MASK; 
				reverbPtr4 = (MusicGlobals->reverbPtr - 1711*a) & REVERB_BUFFER_MASK;
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= (c + b) >> 2;
					c = b;
					b += reverbBuf[reverbPtr2] >> 3;
					b += reverbBuf[reverbPtr3] >> 3;
					b += reverbBuf[reverbPtr4] >> 3;
					reverbBuf[reverbPtr1] = *sourceL + b - (b >> 2);
					*sourceL = *sourceL +  (b >> 1) - (b >> 3);
					sourceL++;
							
					reverbPtr1 = (reverbPtr1 + 1) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 1) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 1) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 1) & REVERB_BUFFER_MASK;
				}
				break;
		}
		MusicGlobals->LPfilterL[0] = b;
		MusicGlobals->LPfilterLz[0] = bz;
		MusicGlobals->LPfilterR[0] = c;
		MusicGlobals->LPfilterRz[0] = cz;
		MusicGlobals->reverbPtr = reverbPtr1;
	}
}	

void PV_ProcessReverbStereo(void)
{
	register INT32 b, c, bz, cz;
	register INT32 *sourceL;
	register INT32 *reverbBuf;
	register LOOPCOUNT a;
	register long reverbPtr1, reverbPtr2, reverbPtr3, reverbPtr4;

	reverbBuf = &MusicGlobals->reverbBuffer[0];
	if (reverbBuf)
	{
		sourceL = &MusicGlobals->songBufferLeftMono[0];
	
		b = MusicGlobals->LPfilterL[0];
		c = MusicGlobals->LPfilterR[0];
		bz = MusicGlobals->LPfilterLz[0];
		cz = MusicGlobals->LPfilterRz[0];
		reverbPtr1 = MusicGlobals->reverbPtr;

	// NOTE: do not exceed a tap distance of 2047.  2047*8 is almost the full buffer size.
		switch (MusicGlobals->reverbUnitType)
		{
			case REVERB_TYPE_1:
				break;
#if USE_SMALL_MEMORY_REVERB == FALSE
	// ----------
#if USE_VARIABLE_REVERB != FALSE
			case REVERB_TYPE_8:
//				if (*((short *) 0x17a))
//					for (a = 0; a < MusicGlobals->One_Loop*2; a++)
//						sourceL[a] = 0;
				PV_ProcessVariableReverb(K_REVERBSTEREO, 0);
				PV_ProcessVariableReverb(K_REVERBSTEREO, 1);
				reverbPtr1 += MusicGlobals->One_Loop;
				break;
#endif
	// ----------
			case REVERB_TYPE_3:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 632*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 430*2) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 798*2) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 632*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 430*4) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 798*4) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 632*8) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 430*8) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 798*8) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= b >> 2;
					b += (reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 3;
					reverbBuf[reverbPtr1] = *sourceL + (b >> 1);
					*sourceL += (b + c) >> 1;
					sourceL++;
					
					c -= c >> 2;
					c += (reverbBuf[reverbPtr2+1] + reverbBuf[reverbPtr3+1] + reverbBuf[reverbPtr4+1]) >> 3;
					reverbBuf[reverbPtr1+1] = *sourceL + (c >> 1);
					*sourceL += (c + b) >> 1;
					sourceL++;
					
					reverbPtr1 = (reverbPtr1 + 2) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 2) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 2) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 2) & REVERB_BUFFER_MASK;
				}
				break;
	// ----------
			case REVERB_TYPE_4:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 1100*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 1473*2) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1711*2) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 1100*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 1473*4) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 1711*4) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 1100*8) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 1473*8) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1711*8) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= (bz + b) >> 2;
					bz = b;
					b += (reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 3;
					reverbBuf[reverbPtr1] = *sourceL + b - (b >> 2);
					*sourceL = *sourceL +  ((c + b) >> 2);
					sourceL++;
					
					c -= (cz + c) >> 2;
					cz = c;
					c += (reverbBuf[reverbPtr2+1] + reverbBuf[reverbPtr3+1] + reverbBuf[reverbPtr4+1]) >> 3;
					reverbBuf[reverbPtr1+1] = *sourceL + c - (c >> 2);
					*sourceL = *sourceL + ((c + b) >> 2);
					sourceL++;
					
					reverbPtr1 = (reverbPtr1 + 2) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 2) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 2) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 2) & REVERB_BUFFER_MASK;
				}
				break;
	// ----------
			case REVERB_TYPE_5:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 500*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 674*2) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1174*2) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 500*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 674*4) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 1174*4) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 500*8) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 674*8) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 1174*8) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= b >> 1;
					b = (reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 2;
					reverbBuf[reverbPtr1] = *sourceL + b - (b >> 2);
					*sourceL = *sourceL + b;
					sourceL++;
					
					c -= c >> 1;
					c = (reverbBuf[reverbPtr2+1] + reverbBuf[reverbPtr3+1] + reverbBuf[reverbPtr4+1]) >> 2;
					reverbBuf[reverbPtr1+1] = *sourceL + c - (c >> 2);
					*sourceL = *sourceL + c;
					sourceL++;
					
					reverbPtr1 = (reverbPtr1 + 2) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 2) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 2) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 2) & REVERB_BUFFER_MASK;
				}
				break;
	// ---------- 
			case REVERB_TYPE_6:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K: reverbPtr2 = (MusicGlobals->reverbPtr - 2700*2) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 3250*2) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 4095*2) & REVERB_BUFFER_MASK; break;
					case Q_22K: reverbPtr2 = (MusicGlobals->reverbPtr - 2700*4) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 3250*4) & REVERB_BUFFER_MASK; 
								reverbPtr4 = (MusicGlobals->reverbPtr - 4095*4) & REVERB_BUFFER_MASK; break;
					case Q_44K: reverbPtr2 = (MusicGlobals->reverbPtr - 2700*8) & REVERB_BUFFER_MASK; 
								reverbPtr3 = (MusicGlobals->reverbPtr - 3250*8) & REVERB_BUFFER_MASK;
								reverbPtr4 = (MusicGlobals->reverbPtr - 4095*8) & REVERB_BUFFER_MASK; break;
				}
				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b += ((reverbBuf[reverbPtr2] + reverbBuf[reverbPtr3] + reverbBuf[reverbPtr4]) >> 4) - (b >> 2);
					reverbBuf[reverbPtr1] = *sourceL + b - (b >> 8);
					*sourceL = *sourceL + b + c;
					sourceL++;
					
					c += ((reverbBuf[reverbPtr2+1] + reverbBuf[reverbPtr3+1] + reverbBuf[reverbPtr4+1]) >> 4) - (c >> 2);
					reverbBuf[reverbPtr1+1] = *sourceL + c - (c >> 8);
					*sourceL = *sourceL + c + b;
					sourceL++;
					
					reverbPtr1 = (reverbPtr1 + 2) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 2) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 2) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 2) & REVERB_BUFFER_MASK;
				}
				break;
#else
			default:
#endif
	// ----------
			case REVERB_TYPE_7:
				switch (MusicGlobals->outputQuality)
				{
					case Q_11K:
						a = 1;
						break;
					case Q_22K:
						a = 2;
						break;
					case Q_44K:
						a = 4;
						break;
				}
				reverbPtr1 = MusicGlobals->reverbPtr & REVERB_BUFFER_MASK;
				reverbPtr2 = (MusicGlobals->reverbPtr - 1100*a) & REVERB_BUFFER_MASK; 
				reverbPtr3 = (MusicGlobals->reverbPtr - 1473*a) & REVERB_BUFFER_MASK; 
				reverbPtr4 = (MusicGlobals->reverbPtr - 1711*a) & REVERB_BUFFER_MASK;

				for (a = MusicGlobals->One_Loop; a > 0; --a)
				{
					b -= (c + b) >> 2;
					c = b;
					b += reverbBuf[reverbPtr2] >> 3;
					b += reverbBuf[reverbPtr3] >> 3;
					b += reverbBuf[reverbPtr4] >> 3;
					reverbBuf[reverbPtr1] = *sourceL + sourceL[1] + b - (b >> 2);
					*sourceL = *sourceL +  (b >> 1) - (b >> 3);
					sourceL++;
					*sourceL = *sourceL +  (b >> 1) - (b >> 3);
					sourceL++;
							
					reverbPtr1 = (reverbPtr1 + 1) & REVERB_BUFFER_MASK;
					reverbPtr2 = (reverbPtr2 + 1) & REVERB_BUFFER_MASK;
					reverbPtr3 = (reverbPtr3 + 1) & REVERB_BUFFER_MASK;
					reverbPtr4 = (reverbPtr4 + 1) & REVERB_BUFFER_MASK;
				}
				break;
		}
		MusicGlobals->LPfilterL[0] = b;
		MusicGlobals->LPfilterLz[0] = bz;
		MusicGlobals->LPfilterR[0] = c;
		MusicGlobals->LPfilterRz[0] = cz;
		MusicGlobals->reverbPtr = reverbPtr1;
	}	
}

#if USE_SMALL_MEMORY_REVERB == FALSE
void PV_PostFilterStereo(void)
{
	register INT32 *source;
	register INT32 s1, s2, s3, s4, s5, s6, s7, s8;
	register LOOPCOUNT a;

	source = &MusicGlobals->songBufferLeftMono[0];
	s1 = MusicGlobals->s1Left;
	s2 = MusicGlobals->s2Left;
	s3 = MusicGlobals->s3Left;
	s4 = MusicGlobals->s4Left;
	s5 = MusicGlobals->s5Left;
	s6 = MusicGlobals->s6Left;
	s7 = MusicGlobals->s7Left;
	s8 = MusicGlobals->s8Left;
	for (a = MusicGlobals->One_Loop; a > 0; --a)
	{
		s1 = s2;
		s2 = s3;
		s3 = s4;
		s4 = s5;
		s5 = s6;
		s6 = s7;
		s7 = s8;
		s8 = *source;
		*source = ((s4 + s5) >> 1) + ((s3 + s6) >> 3) - ((s2 + s7) >> 3) - ((s1 + s8) >> 4);
		source += 2;
	}
	MusicGlobals->s1Left = s1;
	MusicGlobals->s2Left = s2;
	MusicGlobals->s3Left = s3;
	MusicGlobals->s4Left = s4;
	MusicGlobals->s5Left = s5;
	MusicGlobals->s6Left = s6;
	MusicGlobals->s7Left = s7;
	MusicGlobals->s8Left = s8;


	source = &MusicGlobals->songBufferLeftMono[0];
	source++;
	s1 = MusicGlobals->s1Right;
	s2 = MusicGlobals->s2Right;
	s3 = MusicGlobals->s3Right;
	s4 = MusicGlobals->s4Right;
	s5 = MusicGlobals->s5Right;
	s6 = MusicGlobals->s6Right;
	s7 = MusicGlobals->s7Right;
	s8 = MusicGlobals->s8Right;
	for (a = MusicGlobals->One_Loop; a > 0; --a)
	{
		s1 = s2;
		s2 = s3;
		s3 = s4;
		s4 = s5;
		s5 = s6;
		s6 = s7;
		s7 = s8;
		s8 = *source;
		*source = ((s4 + s5) >> 1) + ((s3 + s6) >> 3) - ((s2 + s7) >> 3) - ((s1 + s8) >> 4);
		source += 2;
	}
	MusicGlobals->s1Right = s1;
	MusicGlobals->s2Right = s2;
	MusicGlobals->s3Right = s3;
	MusicGlobals->s4Right = s4;
	MusicGlobals->s5Right = s5;
	MusicGlobals->s6Right = s6;
	MusicGlobals->s7Right = s7;
	MusicGlobals->s8Right = s8;
}	
#endif

void GM_GetReverbTaps(short int *pTapsArray, short int howMany)
{
	short int	count;

	if (pTapsArray && howMany)
	{
		if (howMany > MAX_REVERB_TAPS)
		{
			howMany = MAX_REVERB_TAPS;
		}
		for (count = 0; count < howMany; count++)
		{
			pTapsArray[count] = MusicGlobals->reverbTaps[count];
		}
	}
}

static const short int tap_inits[MAX_REVERB_TAPS] = {
3688, 2749, 27822, 1669, 1276, 924,
74, 46, 33, 14, 19, 14,

4095, 3338, 2580, 2146, 1669, 1094,
58, 72, 65, 16, 16, 9};

void GM_InitReverbTaps(void)
{
	short int count;

	for (count = 0; count < MAX_REVERB_TAPS; count++)
	{
		MusicGlobals->reverbTaps[count] = tap_inits[count];
	}
}

void GM_SetReverbTaps(short int *pTapsArray, short int howMany)
{
	short int	count;

	if (pTapsArray && howMany)
	{
		if (howMany > MAX_REVERB_TAPS)
		{
			howMany = MAX_REVERB_TAPS;
		}
		for (count = 0; count < howMany; count++)
		{
			MusicGlobals->reverbTaps[count] = pTapsArray[count];
		}
	}
}



// EOF of GenReverb.c
