/*
 * @(#)GenSample.c	1.9 98/04/16
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
** "GenSample.c"
**
**	Generalized Music Synthesis package. Part of SoundMusicSys.
**
**
**
** Overview
**	General purpose Music Synthesis software, C-only implementation 
**	(no assembly language optimizations made)
**
** Modification History:
**
**	4/15/97		Moved sample API functions to GenSample.c
**				Added code to handle sample callbacks
**				Added GM_AddSampleOffsetCallback & GM_RemoveSampleOffsetCallback
**	5/1/97		Added startOffsetFrame to GM_BeginSampleFromInfo
**	5/7/97		Changed various functions that wanted stereo pan to be in the range
**				of -255 to 255, to now accecpt the natural range of -63 to 63
**	5/8/97		Fixed GM_ChangeSampleVolume & GM_GetSampleVolume to scale the volume
**				correctly based upon the current effectsVolume
**	5/21/97		Added GM_GetSamplePlaybackPosition
**	6/17/97		Modified GM_BeginSample & GM_BeginDoubleBuffer to save the volume
**				level unscaled in the voice allocated for later use
**				Moved GM_GetEffectsVolume & GM_SetEffectsVolume from GenSetup.c and
**				modified GM_SetEffectsVolume to scale volume levels based upon
**				masterVolume and effectsVolume using the unscaled sample volume
**	6/25/97		Added GM_SetSampleLoopPoints
**	7/22/97		Added GM_GetSampleVolumeUnscaled
**	9/15/97		Added GM_GetSampleReverb
**	10/26/97	Fixed a bug with GM_BeginDoubleBuffer that forgot to look for
**				MusicGlobals being deallocated
**	11/19/97	Removed zero volume check in GM_BeginDoubleBuffer & GM_BeginSample
**
**  JAVASOFT
**	01.14.98	kk: added number of loops to GM_BeginSample calls
**
*/
/*****************************************************************************/
#include "GenSnd.h"
#include "GenPriv.h"


// Stop just sound effects
void GM_EndAllSoundEffects(void)
{           
	register LOOPCOUNT count;
	register NoteRecord *this_one;

	if (MusicGlobals)
	{
		for (count = MusicGlobals->MaxNotes; count < MusicGlobals->MaxNotes + MusicGlobals->MaxEffects; count++)
		{
			this_one = &MusicGlobals->NoteEntry[count];
			if (this_one->NoteDur >= 0)
			{
				PV_DoCallBack(this_one);
				this_one->NoteDur = -1;
			}
		}
	}
}


// Process any fading effects voices
void PV_ServeEffectsFades(void)
{
	long		count, minValue, maxValue;
	NoteRecord	*pVoice;
	long		value;

	minValue = MusicGlobals->MaxNotes;
	maxValue =  minValue + (MusicGlobals->MaxEffects-1);
	for (count = maxValue; count >= minValue; count--)
	{
		pVoice = &MusicGlobals->NoteEntry[count];
		if (pVoice->NoteDur >= 0)
		{
			if (pVoice->soundFadeRate)
			{
				pVoice->soundFixedVolume -= pVoice->soundFadeRate;
				value = FIXED_TO_LONG(pVoice->soundFixedVolume);
				if (value > pVoice->soundFadeMaxVolume)
				{
					value = pVoice->soundFadeMaxVolume;
					pVoice->soundFadeRate = 0;
				}
				if (value < pVoice->soundFadeMinVolume)
				{
					value = pVoice->soundFadeMinVolume;
					pVoice->soundFadeRate = 0;
				}
				pVoice->NoteVolume = value;
				pVoice->NoteMIDIVolume = value;
				if ((pVoice->soundFadeRate == 0) && pVoice->soundEndAtFade)
				{
					GM_EndSample(count);
				}
			}
		}
	}
}

void PV_ServeEffectCallbacks(void)
{
	long					count, minValue, maxValue;
	unsigned long			offsetStart, offsetEnd;
	NoteRecord				*pVoice;
	GM_SampleCallbackEntry	*pCallbackEntry;

	minValue = MusicGlobals->MaxNotes;
	maxValue =  minValue + (MusicGlobals->MaxEffects-1);
	for (count = maxValue; count >= minValue; count--)
	{
		pVoice = &MusicGlobals->NoteEntry[count];
		if (pVoice->NoteDur >= 0)
		{
			pCallbackEntry = pVoice->pSampleMarkList;
			if (pCallbackEntry)
			{
				// get current position of sample
				offsetStart = (pVoice->NoteWave >> STEP_BIT_RANGE);
				// get size of 1 time slice for this sample (11ms) and compute end point
				offsetEnd = offsetStart + pVoice->NoteNextSize;
				while (pCallbackEntry)
				{
					if ((pCallbackEntry->frameOffset >= offsetStart) && (pCallbackEntry->frameOffset <= offsetEnd) )
					{
						if (pCallbackEntry->pCallback)
						{
							(*pCallbackEntry->pCallback)(pCallbackEntry->reference, pVoice->NoteWave >> STEP_BIT_RANGE);
							break;
						}
					}
					pCallbackEntry = pCallbackEntry->pNext;
				}
			}
		}
	}
}

// Begin a double buffer sound effect
// stereoPosition is in the range 63 to -63
INT32 GM_BeginDoubleBuffer(G_PTR pBuffer1, G_PTR pBuffer2, INT32 theSize, INT32 theRate,
							INT16 bitSize, INT16 channels,
							INT32 sampleVolume, INT16 stereoPosition,
							INT32 refData,
							GM_DoubleBufferCallbackPtr bufferCallback)
{
	register MusicVars	*myMusicGlobals;
	register NoteRecord	*pVoice;
	register INT16		count, max, min;

	myMusicGlobals = MusicGlobals;
	if (myMusicGlobals && myMusicGlobals->MaxEffects > 0)
	{
		min = myMusicGlobals->MaxNotes;
		max =  min + (myMusicGlobals->MaxEffects-1);
		for (count = max; count >= min; count--)
		{
			pVoice = &myMusicGlobals->NoteEntry[count];
			if (pVoice->NoteDur < 0)
			{
				PV_CleanNoteEntry(pVoice);	// fill with all zero's except NoteDur field.
				pVoice->noteSamplePitchAdjust = 0x10000;	// 1.0
				pVoice->doubleBufferProc = bufferCallback;
				pVoice->NotePtr = (UBYTE *) pBuffer1;
				pVoice->NotePtrEnd = (UBYTE *) pBuffer1 + theSize;

				pVoice->doubleBufferPtr1 = (UBYTE *) pBuffer1;
				pVoice->doubleBufferPtr2 = (UBYTE *) pBuffer2;

				pVoice->NoteLoopPtr = pVoice->NotePtr;
				pVoice->NoteLoopEnd = pVoice->NotePtrEnd;

				pVoice->NotePitch = (UINT32)theRate / 22050;	// was 22254;
				pVoice->NoteLoopProc = NULL;

				pVoice->NoteEndCallback = NULL;
				pVoice->NoteProgram = -1;      
				pVoice->stereoPosition = stereoPosition;
				pVoice->bitSize = bitSize;
				pVoice->channels = channels;
				pVoice->avoidReverb = TRUE;
				pVoice->soundFadeRate = 0;

				pVoice->NoteMIDIVolume = sampleVolume;	// save unscaled
				sampleVolume = (sampleVolume * myMusicGlobals->effectsVolume) / MAX_MASTER_VOLUME;
				sampleVolume = (sampleVolume * myMusicGlobals->MasterVolume) / MAX_MASTER_VOLUME;

				pVoice->NoteVolume = sampleVolume;
				pVoice->NoteVolumeEnvelope = VOLUME_RANGE;
				pVoice->volumeADSRRecord.ADSRLevel[0] = VOLUME_RANGE;
				pVoice->volumeADSRRecord.ADSRFlags[0] = ADSR_TERMINATE;
				pVoice->volumeADSRRecord.currentLevel = VOLUME_RANGE;
				pVoice->volumeADSRRecord.currentPosition = 0;
				pVoice->volumeADSRRecord.sustainingDecayLevel = 0x10000;

				pVoice->NoteChannel = SOUND_EFFECT_CHANNEL;
				pVoice->NoteDecay = 0x7FFF;		// never release
				pVoice->NoteRefNum = refData;
				pVoice->sustainMode = SUS_NORMAL;
// is there an initial volume level in the ADSR record that starts at time=0?  If so, don't interpolate the
// note's volume up from 0 to the first target level.  Otherwise, it's a traditional ramp-up from 0.
				if (MusicGlobals->generateStereoOutput)
				{
					PV_CalculateStereoVolume(pVoice, &pVoice->lastAmplitudeL, &pVoice->lastAmplitudeR);
				}
				else
				{
					pVoice->lastAmplitudeL = (pVoice->NoteVolume * pVoice->volumeADSRRecord.ADSRLevel[0]) >> VOLUME_PRECISION_SCALAR;
				}
				pVoice->NoteDur = 0x7FFF;
				return count;
			}
		}
	}
	return -1L;
}

// stereoPosition is in the range 63 to -63
INT32 GM_BeginSampleFromInfo(GM_Waveform *pSample, INT32 refData,
								INT32 sampleVolume, INT32 stereoPosition,
								GM_LoopDoneCallbackPtr theLoopContinueProc,
								GM_SoundDoneCallbackPtr theCallbackProc,
								INT32 startOffsetFrame)
{
	unsigned long	bytes;

	if (pSample)
	{
		if (startOffsetFrame < 0)
		{
			startOffsetFrame = 0;
		}
		if (startOffsetFrame > pSample->waveFrames)
		{
			startOffsetFrame = 0;
		}
		bytes = startOffsetFrame * pSample->channels * (pSample->bitSize / 8);
		return GM_BeginSample(	(G_PTR)pSample->theWaveform + bytes,
								pSample->waveFrames - startOffsetFrame,
								pSample->sampledRate,
								pSample->startLoop,
								pSample->endLoop,
								pSample->numLoops,
								sampleVolume, stereoPosition, refData,
								pSample->bitSize,
								pSample->channels,
								theLoopContinueProc, theCallbackProc);
	}							
	return -1L;
}

// stereoPosition is in the range 63 to -63
INT32 GM_BeginSample(G_PTR theData, INT32 frames, INT32 theRate, 
				INT32 theStartLoop, INT32 theEndLoop, INT32 theLoopTarget,
				INT32 sampleVolume, INT32 stereoPosition,
				INT32 refData, INT16 bitSize, INT16 channels, 
				GM_LoopDoneCallbackPtr theLoopContinueProc,
				GM_SoundDoneCallbackPtr theCallbackProc)
{
	register MusicVars	*myMusicGlobals;
	register NoteRecord	*pVoice;
	register INT16		count, max, min;

	myMusicGlobals = MusicGlobals;
	if (myMusicGlobals && myMusicGlobals->MaxEffects > 0)
	{
		min = myMusicGlobals->MaxNotes;
		max =  min + (myMusicGlobals->MaxEffects-1);
		for (count = max; count >= min; count--)
		{
			pVoice = &myMusicGlobals->NoteEntry[count];
			if (pVoice->NoteDur < 0)
			{
				PV_CleanNoteEntry(pVoice);		// zeroes ALL entries
				pVoice->noteSamplePitchAdjust = 0x10000;

				pVoice->NotePtr = (UBYTE *) theData;
				pVoice->NotePtrEnd = (UBYTE *) theData + frames;
				pVoice->NotePitch = (UINT32)theRate / 22050;	// was 22254;

				// reset the NoteLoopCount
				pVoice->NoteLoopCount = 0;

				pVoice->NoteLoopProc = theLoopContinueProc;
				if ( (theStartLoop < theEndLoop) && (theEndLoop - theStartLoop > MIN_LOOP_SIZE) )
				{
					pVoice->NoteLoopPtr = theData + theStartLoop;
					pVoice->NoteLoopEnd = theData + theEndLoop;
				
					// set the target number of loops
					pVoice->NoteLoopTarget = theLoopTarget;
				}
				pVoice->NoteEndCallback = theCallbackProc;
				pVoice->NoteProgram = -1;      
				pVoice->stereoPosition = stereoPosition;
				pVoice->bitSize = bitSize;
				pVoice->channels = channels;
				pVoice->avoidReverb = TRUE;
				pVoice->LFORecordCount = 0;
				pVoice->inst_struct = NULL;
				pVoice->soundFadeRate = 0;

				pVoice->NoteMIDIVolume = sampleVolume;	// save unscaled
				sampleVolume = (sampleVolume * myMusicGlobals->effectsVolume) / MAX_MASTER_VOLUME;
				sampleVolume = (sampleVolume * myMusicGlobals->MasterVolume) / MAX_MASTER_VOLUME;

				pVoice->NoteVolume = sampleVolume;
				pVoice->NoteVolumeEnvelope = VOLUME_RANGE;
				pVoice->volumeADSRRecord.ADSRLevel[0] = VOLUME_RANGE;
				pVoice->volumeADSRRecord.currentLevel = VOLUME_RANGE;
				pVoice->volumeADSRRecord.currentPosition = 0;
				pVoice->volumeADSRRecord.ADSRFlags[0] = ADSR_TERMINATE;
				pVoice->volumeADSRRecord.mode = ADSR_TERMINATE;
				pVoice->volumeADSRRecord.sustainingDecayLevel = 0x10000;
				pVoice->NoteChannel = SOUND_EFFECT_CHANNEL;
				pVoice->NoteRefNum = refData;
				pVoice->sustainMode = SUS_NORMAL;
				pVoice->sampleAndHold = 1;
//					pVoice->NoteDecay = 32000;		// never release
				pVoice->NoteDecay = 8;
// is there an initial volume level in the ADSR record that starts at time=0?  If so, don't interpolate the
// note's volume up from 0 to the first target level.  Otherwise, it's a traditional ramp-up from 0.
				if (MusicGlobals->generateStereoOutput)
				{
					PV_CalculateStereoVolume(pVoice, &pVoice->lastAmplitudeL, &pVoice->lastAmplitudeR);
				}
				else
				{
					pVoice->lastAmplitudeL = (pVoice->NoteVolume * pVoice->volumeADSRRecord.ADSRLevel[0]) >> VOLUME_PRECISION_SCALAR;
				}

				pVoice->NoteDur = 32000;
				return count;
			}
		}
	}
	return -1L;
}

void GM_EndSample(INT32 reference)
{
	if (MusicGlobals)
	{
		if ( (reference >= MusicGlobals->MaxNotes) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			PV_DoCallBack(&MusicGlobals->NoteEntry[reference]);
			MusicGlobals->NoteEntry[reference].NoteDur = -1;
		}
	}
}

unsigned long GM_GetSamplePlaybackPosition(INT32 reference)
{
	unsigned long	position;

	position = 0L;
	if (MusicGlobals)
	{
		if ( (reference >= MusicGlobals->MaxNotes) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			// get current position of sample
			position = (MusicGlobals->NoteEntry[reference].NoteWave >> STEP_BIT_RANGE);
		}
	}
	return position;
}


void GM_SetSampleDoneCallback(INT32 reference, GM_SoundDoneCallbackPtr theCallbackProc, INT32 referenceData)
{
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			MusicGlobals->NoteEntry[reference].NoteRefNum = referenceData;
			MusicGlobals->NoteEntry[reference].NoteEndCallback = theCallbackProc;
		}
	}
}

BOOL_FLAG GM_IsSoundDone(INT32 reference)
{
	register INT32 count;

	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			return (MusicGlobals->NoteEntry[reference].NoteDur < 0);
		}
	
		if (reference < 0)
		{
			return TRUE;
		}
		for (count = MusicGlobals->MaxNotes; count < MusicGlobals->MaxNotes+MusicGlobals->MaxEffects; count++)
		{
			if (MusicGlobals->NoteEntry[count].NoteDur >= 0)
			{
				return FALSE;
			}
		}
	}
	return TRUE;
}                

FIXED_VALUE GM_GetSamplePitch(INT32 reference)
{
	FIXED_VALUE	rate;

	rate = 22050L * 65536L;
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			rate = MusicGlobals->NoteEntry[reference].NotePitch * 22050L;
		}
	}
	return rate;
}

void GM_ChangeSamplePitch(INT32 reference, FIXED_VALUE theNewRate)
{
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			MusicGlobals->NoteEntry[reference].NotePitch  = ( (FIXED_VALUE)theNewRate / 22050L);
		}
	}
}

INT16 GM_GetSampleReverb(INT32 reference)
{
	INT16 enable;

	enable = FALSE;
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			if (MusicGlobals->NoteEntry[reference].avoidReverb == FALSE)
			{
				enable = TRUE;
			}
		}
	}
	return enable;
}

void GM_ChangeSampleReverb(INT32 reference, INT16 enable)
{
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			MusicGlobals->NoteEntry[reference].avoidReverb = ! enable;
		}
	}
}

INT16 GM_GetSampleVolume(INT32 reference)
{
	INT16		volume;
	NoteRecord	*pVoice;

	volume = 0;
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			if (pVoice->NoteDur >= 0)
			{
				volume = (pVoice->NoteVolume * MAX_MASTER_VOLUME) / MusicGlobals->effectsVolume;
				volume = (volume * MAX_MASTER_VOLUME) / MusicGlobals->MasterVolume;
			}
		}
	}
	return volume;
}

// Return volume from a sample that is not scaled
INT16 GM_GetSampleVolumeUnscaled(INT32 reference)
{
	INT16		volume;
	NoteRecord	*pVoice;

	volume = 0;
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			if (pVoice->NoteDur >= 0)
			{
				volume = pVoice->NoteMIDIVolume;
			}
		}
	}
	return volume;
}

// Volume range is from 0 to MAX_NOTE_VOLUME
void GM_ChangeSampleVolume(INT32 reference, INT16 sampleVolume)
{
	register NoteRecord	*pVoice;

	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			pVoice->NoteMIDIVolume = sampleVolume;	// save unscaled
			sampleVolume = (sampleVolume * MusicGlobals->effectsVolume) / MAX_MASTER_VOLUME;
			sampleVolume = (sampleVolume * MusicGlobals->MasterVolume) / MAX_MASTER_VOLUME;
			pVoice->NoteVolume = sampleVolume;
		}
	}
}

void GM_SetSampleLoopPoints(INT32 reference, unsigned long start, unsigned long end)
{
	register NoteRecord	*pVoice;

	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			if ( (start < end) && (end - start > MIN_LOOP_SIZE) )
			{
				pVoice->NoteLoopPtr = (UBYTE *)pVoice->NotePtr + start;
				pVoice->NoteLoopEnd = (UBYTE *)pVoice->NotePtr + end;
			}
		}
	}
}

// range is 0 to MAX_MASTER_VOLUME (256). Note volume is from 0 to MAX_NOTE_VOLUME (127)
void GM_SetEffectsVolume(INT16 newVolume)
{
	register MusicVars		*myMusicGlobals;
	register LOOPCOUNT		count;
	register NoteRecord *	pVoice;
	short int				minValue, maxValue;

	if (newVolume < 0)
	{
		newVolume = 0;
	}
	if (newVolume > MAX_MASTER_VOLUME * 5)
	{
		newVolume = MAX_MASTER_VOLUME * 5;
	}
	myMusicGlobals = MusicGlobals;
	if (myMusicGlobals)
	{
		minValue = myMusicGlobals->MaxNotes;
		maxValue =  minValue + myMusicGlobals->MaxEffects;

		myMusicGlobals->effectsVolume = newVolume;
		newVolume = (newVolume * MAX_NOTE_VOLUME) / MAX_MASTER_VOLUME; // scale

		// update the current notes playing to the new volume
		for (count = minValue; count < maxValue; count++)
		{
			pVoice = &myMusicGlobals->NoteEntry[count];
			if (pVoice->NoteDur >= 0)
			{
				if (pVoice->NoteChannel == SOUND_EFFECT_CHANNEL)
				{
					// make sure and set the channel volume not scaled, because its scaled later
					if (newVolume == 0)
					{
						pVoice->NoteDur = 0;
						pVoice->NoteDecay = 0;
						pVoice->volumeADSRRecord.ADSRTime[0] = 1;
						pVoice->volumeADSRRecord.ADSRFlags[0] = ADSR_TERMINATE;
						pVoice->volumeADSRRecord.ADSRLevel[0] = 0;	// just in case
					}
					// now calculate the new volume based upon the current channel volume and
					// the unscaled note volume
					newVolume = (pVoice->NoteMIDIVolume * myMusicGlobals->MasterVolume) / MAX_MASTER_VOLUME;
					newVolume = (newVolume * myMusicGlobals->effectsVolume) / MAX_MASTER_VOLUME;
					pVoice->NoteVolume = newVolume;
				}
			}
		}
	}
}

INT16 GM_GetEffectsVolume(void)
{
	INT16	volume;

	volume = MAX_MASTER_VOLUME;
	if (MusicGlobals)
	{
		volume = MusicGlobals->effectsVolume;
	}
	return volume;
}



// Set sample fade rate. Its a 16.16 fixed value
// Input:	reference		sound to affect
//			fadeRate	amount to change every 11 ms
//						example:	FLOAT_TO_FIXED(2.2) will decrease volume
//									FLOAT_TO_FIXED(2.2) * -1 will increase volume
//			minVolume	lowest volume level fade will go
//			maxVolume	highest volume level fade will go
void GM_SetSampleFadeRate(INT32 reference, FIXED_VALUE fadeRate, 
							INT16 minVolume, INT16 maxVolume, BOOL_FLAG endSample)
{
	register NoteRecord	*pVoice;

	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			pVoice->soundFadeMaxVolume = maxVolume * 4;
			pVoice->soundFadeMinVolume = minVolume * 4;
			pVoice->soundFixedVolume = LONG_TO_FIXED(pVoice->NoteVolume);
			pVoice->soundEndAtFade = endSample;
			pVoice->soundFadeRate = fadeRate;
		}
	}
}

// range from -63 to 63
INT16 GM_GetSampleStereoPosition(INT32 reference)
{
	INT16	pos;

	pos = 0;
	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pos = MusicGlobals->NoteEntry[reference].stereoPosition;
		}
	}
	return pos;
}

// range from -63 to 63
void GM_ChangeSampleStereoPosition(INT32 reference, INT16 newStereoPosition)
{
	register NoteRecord	*pVoice;

	if (MusicGlobals)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			pVoice->stereoPosition = newStereoPosition;
		}
	}
}

void GM_SetSampleOffsetCallbackLinks(INT32 reference, GM_SampleCallbackEntry *pTopEntry)
{
	NoteRecord				*pVoice;

	if (MusicGlobals && pTopEntry)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			pVoice->pSampleMarkList = pTopEntry;
		}
	}
}

void GM_AddSampleOffsetCallback(INT32 reference, GM_SampleCallbackEntry *pEntry)
{
	NoteRecord				*pVoice;
	GM_SampleCallbackEntry	*pNext;

	if (MusicGlobals && pEntry)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];
			pNext = pVoice->pSampleMarkList;
			while (pNext)
			{
				if (pNext->pNext == NULL)
				{
					break;
				}
				else
				{
					pNext = pNext->pNext;
				}
			}
			if (pNext == NULL)
			{
				pVoice->pSampleMarkList = pEntry;
			}
			else
			{
				pNext->pNext = pEntry;
			}
		}
	}
}

void GM_RemoveSampleOffsetCallback(INT32 reference, GM_SampleCallbackEntry *pEntry)
{
	NoteRecord				*pVoice;
	GM_SampleCallbackEntry	*pNext, *pLast;

	if (MusicGlobals && pEntry)
	{
		if ( (reference > 0) && (reference < (MusicGlobals->MaxNotes+MusicGlobals->MaxEffects)) )
		{
			pVoice = &MusicGlobals->NoteEntry[reference];

			pLast = pNext = pVoice->pSampleMarkList;
			while (pNext)
			{
				if (pNext == pEntry)								// found object in list?
				{
					if (pNext == pVoice->pSampleMarkList)			// is object the top object
					{
						pVoice->pSampleMarkList = pNext->pNext;		// yes, change to next object
					}
					else
					{
						if (pLast)									// no, change last to point beyond next
						{
							pLast->pNext = pNext->pNext;
						}
					}
					break;
				}
				pLast = pNext;
				pNext = pNext->pNext;
			}
		}
	}
}

void GM_FreeWaveform(GM_Waveform *pWaveform)
{
	if (pWaveform)
	{
		XDisposePtr((XPTR)pWaveform->theWaveform);
		XDisposePtr((XPTR)pWaveform);
	}
}



/* EOF of GenSample.c
*/

