/*
 * @(#)GenSetup.c	1.14 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
** "GenSetup.c"
**
**	Generalized Music Synthesis package. Part of SoundMusicSys.
**
**
**
** Overview
**	General purpose Music Synthesis software, C-only implementation
**	(no assembly language optimizations made)
**
** Modification History:
**
**	1/10/96		Split from GenSynth.c
**	1/12/96		Changed GM_ChangeSystemVoices to support 32 voices instead of 31
**	1/18/96		Spruced up for C++ extra error checking
**				Fixed bug with GM_FinisGeneralSound referencing MusicVars before its allocated
**	1/19/96		Changed GM_BeginSample to support bitsize and channels
**	2/5/96		Working towards multiple songs
**	2/12/96		Moved cleaning of external midi queue to Init code
**	3/6/96		Eliminated the global songVolume
**	4/10/96		Changed min loop size from 100 to use define 'MIN_LOOP_SIZE'
**				Eliminated the 'can't use 0 as loop start' in GM_BeginSample
**	4/21/96		Removed register usage in parameters
**				Put test around ReverbBuffer being NULL
**	4/25/96		Removed SwapLong & SwapShort. Use XGetLong & XGetShort instead
**	7/4/96		Changed font and re tabbed
**	7/31/96		Fixed PV_ChangeSustainedNotes to do pedal changes as 127=on,
**				anything else as off
**	10/11/96	Added GM_BeginSampleFromInfo
**	10/23/96	Removed reference to BYTE and changed them all to UBYTE or SBYTE
**	11/3/96		Fixed a bug in GM_BeginDoubleBuffer. Changed currentPosition to
**				currentLevel
**				Changed the way the expression controller calculates extra volume,
**				now done in PV_ScaleVolumeFromChannelAndSong
**	11/8/96		Added GM_GetSampleVolume & GM_GetSamplePitch & GM_GetSampleStereoPosition
**	11/21/96	Reconfigured GM_InitGeneralSound to check parmeters in a clearer way.
**				Also prevents possible compiler problems
**	12/17/96	Fixed bug in GM_InitGeneralSound that would return an error for 32 voices
**	12/30/96	Changed copyright
**	1/12/97		Changed maxNormalizedVoices to mixLevel
**	1/22/97		Added GM_SetSampleDoneCallback
**	1/23/97		Fixed a bug with PV_ScaleVolumeFromChannelAndSong that prevent direct midi
**				control from changing the master volume
**				Changed GM_GetSampleVolume to return zero if sample reference is bad
**	1/27/97		Added fade sample engine via PV_ServeEffectsFades
**	1/28/97		Added the ability to end a sample when fade is reached
**	2/12/97		Added support for X_WIN_HARDWARE
**	2/13/97		Added GM_GetSystemVoices
**	2/20/97		Added support for TYPE 7 and 8 reverb in GM_SetReverbType
**				Fixed a bug with GM_SetMasterVolume that doesn't kill audio when
**				volume is 0
**	3/3/97		Reversed stereo placement in PV_CalculateStereoVolume
**	3/24/97		Put code in to prevent changing the voices unless they are
**				different. See GM_ChangeSystemVoices
**	4/1/97		Fixed a memory leak. Forgot to free reverb memory in GM_FinisGeneralSound
**	4/9/97		Added sampleExpansion factor
**	4/15/97		Moved sample API functions to GenSample.c
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**	5/5/97		Changed PV_CalculateStereoVolume to handle the fractional volume case
**				at 1.
**	6/4/97		Renamed InitSoundManager to GM_StartHardwareSoundManager, and
**				renamed FinsSoundManager to GM_StopHardwareSoundManager, and
**				now pass in a thread context
**				Changed GM_InitGeneralSound & GM_FinisGeneralSound & GM_ChangeAudioModes &
**				GM_PauseGeneralSound & GM_ResumeGeneralSound to pass in a thread context
**				that is passed to the low level hardware API
**	6/17/97		Changed GM_SetMasterVolume to reset all songs and samples to rescale
**				current volumes
**	7/22/97		Moved GM_SetAudioTask & GM_SetAudioOutput & GM_GetAudioTask &
**				GM_GetAudioOutput from various platform specific files to GenSetup.c
**				Changed GM_SetMasterVolume to include streams calls
**	7/28/97		In GM_SetMasterVolume, wrapped a define around the stream call
**	8/15/97		Removed reference of myMusicGlobals->channelCallbackProc. Now using new
**				memory efficant controler callback
**	8/18/97		Changed X_WIN_HAE to USE_HAE_EXTERNAL_API
**	8/19/97		Seperated GM_StartHardwareSoundManager from being called inside of
**				GM_InitGeneralSound. Now it has to be called by the caller. Likewise
**				GM_StopHardwareSoundManager is no longer being called within GM_FinisGeneralSound.
**	10/12/97	Fixed SetChannelVolume & PV_ChangeSustainedNotes & SetChannelStereoPosition &
**				SetChannelModWheel & SetChannelPitchBend to check for a particular song before
**				changing playing voices. This fixed a bug in which a pitch bend from one song
**				would affect another
**	10/16/97	Removed lame instrument cache support from GM_InitGeneralSound
**	11/12/97	Added GM_MaxDevices & GM_SetDeviceID & GM_GetDeviceID & GM_GetDeviceName
**	12/4/97		Added GM_GetDeviceTimeStamp & GM_UpdateSamplesPlayed
**				Added calculation of sampleFrameSize to GM_InitGeneralSound
**				Renamed MusicVars to GM_Mixer
*/
/*****************************************************************************/
#include "X_API.h"
#include "GenSnd.h"
#include "GenPriv.h"

#if USE_HAE_EXTERNAL_API
	#include "HAE_API.h"
#endif


OPErr GM_PauseGeneralSound(void *context)
{
	OPErr	theErr;

	theErr = NO_ERR;
	if (MusicGlobals)
	{
		if (MusicGlobals->systemPaused == FALSE)
		{
#if USE_SEQUENCER
			GM_PauseSequencer();
#endif
			GM_EndAllSoundEffects();
			MusicGlobals->systemPaused = TRUE;
			GM_StopHardwareSoundManager(context);		// disengage from hardware
		}
		else
		{
			theErr = ALREADY_PAUSED;
		}
	}
	return theErr;
}

OPErr GM_ResumeGeneralSound(void *context)
{
	OPErr	theErr;

	theErr = NO_ERR;
	if (MusicGlobals)
	{
		if (MusicGlobals->systemPaused)
		{
			if (GM_StartHardwareSoundManager(context))			// reconnect to hardware
            {
			    MusicGlobals->systemPaused = FALSE;
#if USE_SEQUENCER
			    GM_ResumeSequencer();
#endif
            }
            else
            {
                theErr = DEVICE_UNAVAILABLE;
            }
		}
		else
		{
			theErr = ALREADY_RESUMED;
		}
	}
	return theErr;
}


void GM_GetSystemVoices(INT16 *pMaxSongVoices, INT16 *pMixLevel, INT16 *pMaxEffectVoices)
{
	if (MusicGlobals && pMaxSongVoices && pMixLevel && pMaxEffectVoices)
	{
		*pMaxSongVoices = MusicGlobals->MaxNotes;
		*pMixLevel = MusicGlobals->mixLevel;
		*pMaxEffectVoices = MusicGlobals->MaxEffects;
	}
}

OPErr GM_ChangeSystemVoices(INT16 maxSongVoices, INT16 mixLevel, INT16 maxEffectVoices)
{
	OPErr		theErr;
	BOOL_FLAG	change;

	theErr = NO_ERR;
	if ( (maxSongVoices >= 0) &&
		(mixLevel > 0) &&
		(maxEffectVoices >= 0) &&
		((maxEffectVoices+maxSongVoices) > 0) &&
		((maxEffectVoices+maxSongVoices) <= MAX_VOICES) )
	{
		change = FALSE;
		if (MusicGlobals->MaxNotes != maxSongVoices)
		{
			change = TRUE;
		}
		if (MusicGlobals->mixLevel != mixLevel)
		{
			change = TRUE;
		}
		if (MusicGlobals->MaxEffects != maxEffectVoices)
		{
			change = TRUE;
		}
		if (change)
		{
			MusicGlobals->MaxNotes = maxSongVoices;
			MusicGlobals->mixLevel = mixLevel;
			MusicGlobals->MaxEffects = maxEffectVoices;

			PV_CalcScaleBack();
		}
	}
	else
	{
		theErr = PARAM_ERR;
	}
	return theErr;
}

// Set the master volume, and recalculate all volumes. Scale is 0 to MAX_MASTER_VOLUME
void GM_SetMasterVolume(INT32 theVolume)
{
	if (MusicGlobals)
	{
		MusicGlobals->MasterVolume = (INT16)theVolume;
		PV_CalcScaleBack();

#if USE_STREAM_API
		GM_AudioStreamSetVolumeAll(-1);	// recalculate stream volumes
#endif
		{
			short int	count;
			GM_Song		*pSong;

			// reset volumes for sound effects
			GM_SetEffectsVolume(GM_GetEffectsVolume());

			// walk through songs and reset volumes
			for (count = 0; count < MAX_SONGS; count++)
			{
				pSong = MusicGlobals->pSongsToPlay[count];
				if (pSong)
				{
					GM_SetSongVolume(pSong, GM_GetSongVolume(pSong));
				}
			}
		}
	}
}

INT32 GM_GetMasterVolume(void)
{
	if (MusicGlobals)
	{
		return MusicGlobals->MasterVolume;
	}
	else
	{
		return MAX_MASTER_VOLUME;
	}
}


OPErr GM_InitGeneralSound(void *context, Quality theQuality, TerpMode theTerp, AudioModifiers theMods,
				INT16 maxVoices, INT16 normVoices, INT16 maxEffects)
{
	register GM_Mixer	*myMusicGlobals;
	register INT32		count;
	OPErr				theErr;

	context = context;
	theErr = NO_ERR;
	count = maxVoices + maxEffects;
	if (count <= MAX_VOICES)
	{
		if (normVoices > MAX_VOICES)
		{
			if ((normVoices/100) > count)
			{
				theErr = PARAM_ERR;
			}
		}
		else if (normVoices > count)
		{
			theErr = PARAM_ERR;
		}
	}
	else
	{
		theErr = PARAM_ERR;
	}
// Check terp mode
	switch (theTerp)
	{
		case E_AMP_SCALED_DROP_SAMPLE:
		case E_2_POINT_INTERPOLATION:
		case E_LINEAR_INTERPOLATION:
			break;
		default:
			theErr = PARAM_ERR;
			break;
	}

#if USE_HAE_EXTERNAL_API
	if (theErr == NO_ERR)
	{
		// call setup before any memory allocation happens
		if (HAE_Setup())
		{
			theErr = MEMORY_ERR;
		}
	}
#else
	#if X_PLATFORM == X_WIN_HARDWARE
	if (theErr == NO_ERR)
	{
		// Link to VxD. We must do this before using the memory manager
		if (OpenVxD())
		{
			// Pagelock all static code and data belonging to this executable
			// Dynamically allocated data used at interrupt time must also be pagelocked
			// Interrupt callback should be in a DLL if possible to avoid pagelocking
			//		tons of unnecessary stuff
			LockMe();
		}
		else
		{
			theErr = MEMORY_ERR;
		}
	}
	#endif
#endif


	if (theErr == NO_ERR)
	{
// Allocate MusicGlobals
		MusicGlobals = (GM_Mixer *)XNewPtr( (long)sizeof(GM_Mixer) );
		myMusicGlobals = MusicGlobals;
		if (MusicGlobals)
		{
			for (count = 0; count < MAX_VOICES; count++)
			{
				myMusicGlobals->NoteEntry[count].NoteDur = -1;
			}
			myMusicGlobals->interpolationMode = theTerp;

			myMusicGlobals->MasterVolume = MAX_MASTER_VOLUME;
			myMusicGlobals->effectsVolume = MAX_MASTER_VOLUME * 3;	// samples 3 times normal volume
			myMusicGlobals->maxChunkSize = MAX_CHUNK_SIZE;
			myMusicGlobals->outputQuality = theQuality;
			switch (theQuality)
			{
				case Q_44K:
					myMusicGlobals->One_Slice = MAX_CHUNK_SIZE;		// 512 in current version
					myMusicGlobals->One_Loop = MAX_CHUNK_SIZE;
					myMusicGlobals->Two_Loop = MAX_CHUNK_SIZE/2;
					myMusicGlobals->Four_Loop = MAX_CHUNK_SIZE/4;
					myMusicGlobals->Sixteen_Loop = MAX_CHUNK_SIZE/16;
					break;
				case Q_22K:
					myMusicGlobals->One_Slice = MAX_CHUNK_SIZE/2;
					myMusicGlobals->One_Loop = MAX_CHUNK_SIZE/2;
					myMusicGlobals->Two_Loop = MAX_CHUNK_SIZE/4;
					myMusicGlobals->Four_Loop = MAX_CHUNK_SIZE/8;
					myMusicGlobals->Sixteen_Loop = MAX_CHUNK_SIZE/32;
					break;
				case Q_11K:
					myMusicGlobals->One_Slice = MAX_CHUNK_SIZE/4;
					myMusicGlobals->One_Loop = MAX_CHUNK_SIZE/4;
					myMusicGlobals->Two_Loop = MAX_CHUNK_SIZE/8;
					myMusicGlobals->Four_Loop = MAX_CHUNK_SIZE/16;
					myMusicGlobals->Sixteen_Loop = MAX_CHUNK_SIZE/64;
					break;
			}

			myMusicGlobals->sampleExpansion = 1;
			// Don't generate 16 bit output, unless hardware can do it.
			if ( (theMods & M_USE_16) == M_USE_16)
			{
				myMusicGlobals->generate16output = XIs16BitSupported();
			}
			else
			{
				// $$kk: 03.17.98: some solaris audio drivers do not support linear 8 bit output
				// so we need to check for mono support
				myMusicGlobals->generate16output = !XIs8BitSupported();
			}

			// $$kk: 03.17.98: now if whatever is set is not supported, we are hosed
			if (myMusicGlobals->generate16output)
			{
				if (!XIs16BitSupported())
				{
					theErr = UNSUPPORTED_HARDWARE;
				}
			}
			else
			{
				if (!XIs8BitSupported())
				{
					theErr = UNSUPPORTED_HARDWARE;
				}
			}


			// double check users request for Stereo output. Make sure the hardware can play it
			if ( (theMods & M_USE_STEREO) == M_USE_STEREO)
			{
				myMusicGlobals->generateStereoOutput = XIsStereoSupported();
			}
			else
			{
				myMusicGlobals->generateStereoOutput = FALSE;
			}

			myMusicGlobals->stereoFilter = ( (myMusicGlobals->generateStereoOutput) &&
												 ((theMods & M_STEREO_FILTER) == M_STEREO_FILTER) ) ? TRUE : FALSE;
			myMusicGlobals->MaxNotes = maxVoices;
			myMusicGlobals->mixLevel = normVoices;
			myMusicGlobals->MaxEffects = maxEffects;
			myMusicGlobals->reverbPtr = 0;
			if ( (theMods & M_DISABLE_REVERB) == M_DISABLE_REVERB)
			{
				myMusicGlobals->reverbBuffer = NULL;
				myMusicGlobals->reverbBuffer2 = NULL;
				myMusicGlobals->reverbBuffer3 = NULL;
			}
			else
			{
				myMusicGlobals->reverbBuffer = (INT32 *)XNewPtr(REVERB_BUFFER_SIZE * 2L * sizeof(long));
				myMusicGlobals->reverbBuffer2 = (INT32 *)XNewPtr(REVERB_BUFFER_SIZE * sizeof(long));
				myMusicGlobals->reverbBuffer3 = (INT32 *)XNewPtr(REVERB_BUFFER_SIZE * sizeof(long));
				if (myMusicGlobals->reverbBuffer)
				{
					for (count = 0; count < REVERB_BUFFER_SIZE * 2; count++)
					{
						myMusicGlobals->reverbBuffer[count] = 0;
					}
				}
				if (myMusicGlobals->reverbBuffer2)
				{
					for (count = 0; count < REVERB_BUFFER_SIZE; count++)
					{
						myMusicGlobals->reverbBuffer2[count] = 0;
					}
				}
				if (myMusicGlobals->reverbBuffer3)
				{
					for (count = 0; count < REVERB_BUFFER_SIZE; count++)
					{
						myMusicGlobals->reverbBuffer3[count] = 0;
					}
				}
			}
			GM_InitReverbTaps();
			for (count = 0; count < MAX_REVERB_GROUPS; count++)
			{
				myMusicGlobals->LPfilterL[count] = 0;
				myMusicGlobals->LPfilterR[count] = 0;
				myMusicGlobals->LPfilterLz[count] = 0;
				myMusicGlobals->LPfilterRz[count] = 0;
			}
			myMusicGlobals->reverbUnitType = DEFAULT_REVERB_TYPE;
#if USE_SEQUENCER
			GM_EndAllNotes();
#endif

// Compute volume multiplier for mix-level
			PV_CalcScaleBack();
// Turn off all notes!
			for (count = 0; count < myMusicGlobals->MaxNotes+myMusicGlobals->MaxEffects; count++)
			{
				myMusicGlobals->NoteEntry[count].NoteDur = -1;
			}

			for (count = 0; count < MAX_SAMPLES; count++)
			{
				myMusicGlobals->sampleCaches[count] = NULL;
			}
			myMusicGlobals->anoymousCacheID = -30000;
			myMusicGlobals->cacheSamples = FALSE;		// don't cache samples
			myMusicGlobals->cacheInstruments = FALSE;	// not used
		}
		else
		{
			theErr = MEMORY_ERR;
		}
	}
	if (theErr == NO_ERR && MusicGlobals)
	{
		MusicGlobals->insideAudioInterrupt = 0;
		MusicGlobals->enableDriftFixer = FALSE;
		MusicGlobals->syncCount = XMicroseconds();
        MusicGlobals->samplesPlayed = 0;
        MusicGlobals->samplesWritten = 0;
        MusicGlobals->lastSamplePosition = 0;
		MusicGlobals->sequencerPaused = TRUE;
		MusicGlobals->systemPaused = TRUE;
		PV_CleanExternalQueue();

		// calculate sample size for conversion of bytes to sample frames
		MusicGlobals->sampleFrameSize = 1;

		if (MusicGlobals->generate16output)
		{
			MusicGlobals->sampleFrameSize *= 2;
		}
		if (MusicGlobals->generateStereoOutput)
		{
			MusicGlobals->sampleFrameSize *= 2;
		}


		// since we don't call GM_StartHardwareSoundManager, we start the engine
		// up paused. You'll need to call GM_ResumeGeneralSound to start the
		// engine
//		MusicGlobals->sequencerPaused = FALSE;
//		MusicGlobals->systemPaused = FALSE;
//		if (GM_StartHardwareSoundManager(context) == FALSE)
//		{
//			theErr = DEVICE_UNAVAILABLE;
//		}
	}
	return theErr;
}


// $$kk: 01.26.98: changed this to only start and stop the device if it is opened,

OPErr GM_ChangeAudioModes(void *context,
							Quality theQuality, TerpMode theTerp, AudioModifiers theMods)
{
	OPErr				theErr;
	register GM_Mixer	*myMusicGlobals;
	INT16				maxChunkSize;

	BOOL_FLAG			reacquireDevice = FALSE;

	theErr = NO_ERR;
// Check terp mode
	switch (theTerp)
	{
		case E_AMP_SCALED_DROP_SAMPLE:
		case E_2_POINT_INTERPOLATION:
		case E_LINEAR_INTERPOLATION:
			break;
		default:
			theErr = PARAM_ERR;
			break;
	}
	switch (theQuality)
	{
		case Q_11K:
		case Q_22K:
		case Q_44K:
			break;
		default:
			theErr = PARAM_ERR;
			break;
	}
	if (theErr == NO_ERR)
	{

		myMusicGlobals = MusicGlobals;

		if (myMusicGlobals->systemPaused == FALSE)
		{
			GM_StopHardwareSoundManager(context);
			reacquireDevice = TRUE; // reopen the device when we're done configuring
		}

		if ( (theMods & M_USE_16) == M_USE_16)
		{
			myMusicGlobals->generate16output = XIs16BitSupported();
		}
		else
		{
			// $$kk: 03.17.98: some solaris audio drivers do not support linear 8 bit output
			// so we need to check for mono support
			myMusicGlobals->generate16output = !XIs8BitSupported();
		}

		// $$kk: 03.17.98: now if whatever is set is not supported, we are hosed
		if (myMusicGlobals->generate16output)
		{
			if (!XIs16BitSupported())
			{
				theErr = UNSUPPORTED_HARDWARE;
			}
		}
		else
		{
			if (!XIs8BitSupported())
			{
				theErr = UNSUPPORTED_HARDWARE;
			}
		}


		// double check users request for Stereo output. Make sure the hardware can play it
		if ( (theMods & M_USE_STEREO) == M_USE_STEREO)
		{
			myMusicGlobals->generateStereoOutput = XIsStereoSupported();
		}
		else
		{
			myMusicGlobals->generateStereoOutput = FALSE;
		}
		myMusicGlobals->stereoFilter = ( (myMusicGlobals->generateStereoOutput) &&
											 ((theMods & M_STEREO_FILTER) == M_STEREO_FILTER) ) ? TRUE : FALSE;
		if ( (theMods & M_DISABLE_REVERB) == M_DISABLE_REVERB)
		{
			if (myMusicGlobals->reverbBuffer3)
			{
				XDisposePtr((XPTR)myMusicGlobals->reverbBuffer3);
				myMusicGlobals->reverbBuffer3 = NULL;
			}
			if (myMusicGlobals->reverbBuffer2)
			{
				XDisposePtr((XPTR)myMusicGlobals->reverbBuffer2);
				myMusicGlobals->reverbBuffer2 = NULL;
			}
			if (myMusicGlobals->reverbBuffer)
			{
				XDisposePtr((XPTR)myMusicGlobals->reverbBuffer);
				myMusicGlobals->reverbBuffer = NULL;
			}
			GM_SetReverbType(REVERB_TYPE_1);		// no reverb
		}
		maxChunkSize = myMusicGlobals->maxChunkSize;
		myMusicGlobals->outputQuality = theQuality;
		switch (theQuality)
		{
			case Q_44K:
				myMusicGlobals->One_Slice = maxChunkSize;
				myMusicGlobals->One_Loop = maxChunkSize;
				myMusicGlobals->Two_Loop = maxChunkSize/2;
				myMusicGlobals->Four_Loop = maxChunkSize/4;
				myMusicGlobals->Sixteen_Loop = MAX_CHUNK_SIZE/16;
				break;
			case Q_22K:
				myMusicGlobals->One_Slice = maxChunkSize/2;
				myMusicGlobals->One_Loop = maxChunkSize/2;
				myMusicGlobals->Two_Loop = maxChunkSize/4;
				myMusicGlobals->Four_Loop = maxChunkSize/8;
				myMusicGlobals->Sixteen_Loop = MAX_CHUNK_SIZE/32;
				break;
			case Q_11K:
				myMusicGlobals->One_Slice = maxChunkSize/4;
				myMusicGlobals->One_Loop = maxChunkSize/4;
				myMusicGlobals->Two_Loop = maxChunkSize/8;
				myMusicGlobals->Four_Loop = maxChunkSize/16;
				myMusicGlobals->Sixteen_Loop = MAX_CHUNK_SIZE/64;
				break;
		}
		myMusicGlobals->interpolationMode = theTerp;
// Recompute mix level
		PV_CalcScaleBack();

		if (reacquireDevice == TRUE)
		{
			if (GM_StartHardwareSoundManager(context) == FALSE)
			{
			theErr = MEMORY_ERR;
			}
		}
	}
	return theErr;
}


void GM_FinisGeneralSound(void *context)
{
	context = context;
	if (MusicGlobals)
	{
		MusicGlobals->systemPaused = TRUE;
#if USE_SEQUENCER
		GM_FreeSong(NULL);		// free all songs
#endif
		// Close up sound manager BEFORE releasing memory!
//		GM_StopHardwareSoundManager(context);

		if (MusicGlobals->reverbBuffer3)
		{
			XDisposePtr((XPTR)MusicGlobals->reverbBuffer3);
			MusicGlobals->reverbBuffer3 = NULL;
		}
		if (MusicGlobals->reverbBuffer2)
		{
			XDisposePtr((XPTR)MusicGlobals->reverbBuffer2);
			MusicGlobals->reverbBuffer2 = NULL;
		}
		if (MusicGlobals->reverbBuffer)
		{
			XDisposePtr((XPTR)MusicGlobals->reverbBuffer);
			MusicGlobals->reverbBuffer = NULL;
		}

		XDisposePtr((XPTR)MusicGlobals);
		MusicGlobals = NULL;
	}


#if USE_HAE_EXTERNAL_API == TRUE
	HAE_Cleanup();
#endif

#if X_PLATFORM == X_WIN_HARDWARE
	// Un-pagelock all static code and data of this executable
	UnlockMe();

	// Unlink from VxD
	CloseVxD();
#endif
}

UINT32 PV_ScaleVolumeFromChannelAndSong(GM_Song *pSong, INT16 channel, UINT32 volume)
{
	register UINT32		newVolume;

	// scale song volume based upon master song volume, only if a song channel
	if (channel != SOUND_EFFECT_CHANNEL)
	{
		if (pSong)
		{
			if (pSong->channelExpression[channel])
			{
				// for now, let's just scale up the volume level of the channel
				// Say 127 is 25.5% higher
#if USE_DLS
				volume += (UINT32)pSong->channelExpression[channel];
#else
				volume += (UINT32)pSong->channelExpression[channel] / 5;
#endif
			}

			// scale note velocity via current channel volume
			newVolume = (volume * (UINT32)pSong->channelVolume[channel]) / MAX_NOTE_VOLUME;

			// scale note velocity via current song volume
			newVolume = (newVolume * (UINT32)pSong->songVolume) / MAX_NOTE_VOLUME;
		}
		else
		{
			newVolume = volume;
		}
	}
	else
	{
		// scale note velocity via current master effects volume
		newVolume = (volume * (UINT32)MusicGlobals->effectsVolume) / MAX_MASTER_VOLUME;
	}
	return newVolume;
}


// ------------------------------------------------------------------------------------------------------//

/*
static const UBYTE stereoPanRamp[] =
{
0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13,
14, 14, 15, 15, 16, 16, 17, 17, 18, 19, 19, 20, 20, 21, 22, 22, 23, 23, 24, 25, 25, 26, 27, 27,
28, 29, 30, 30, 31, 32, 33, 33, 34, 35, 36, 37, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
49, 50, 51, 52, 53, 54, 55, 56, 58, 59, 60, 62, 63, 64, 66, 67, 69, 70, 72, 73, 75, 77, 79, 81,
83, 85, 87, 89, 91, 94, 96, 99, 102, 105, 108, 111, 115, 119, 123, 127, 132, 138, 144, 151, 159,
169, 180, 195, 216, 252, 253
};
*/
#define USE_GS_RAMP		0

#if USE_GS_RAMP
// new GS ramp * 32. Divide by 32 after using this factor
static const short newStereoPanRamp[] =
{
0, 0, 32, 63, 95, 135, 167, 198, 230, 270, 302, 341, 373, 413, 445, 484, 516,
556, 595, 635, 667, 706, 746, 786, 825, 857, 897, 937, 976, 1016, 1056, 1095,
1135, 1175, 1214, 1254, 1294, 1333, 1373, 1421, 1461, 1500, 1540, 1580, 1619,
1659, 1699, 1738, 1778, 1826, 1865, 1905, 1945, 1984, 2024, 2064, 2104, 2143,
2183, 2223, 2262, 2302, 2342, 2373, 2413, 2453, 2492, 2532, 2564, 2603, 2643,
2675, 2715, 2746, 2786, 2818, 2857, 2889, 2929, 2961, 2992, 3024, 3064, 3096,
3127, 3159, 3191, 3223, 3254, 3278, 3310, 3342, 3365, 3397, 3429, 3453, 3477,
3508, 3532, 3556, 3588, 3612, 3635, 3659, 3683, 3699, 3723, 3747, 3770, 3786,
3810, 3826, 3842, 3866, 3882, 3897, 3913, 3929, 3945, 3961, 3977, 3992, 4008,
4016, 4032, 4040, 4056, 4064,
};
static void PV_RemapMidiPan(INT32 stereoPosition, UINT32 *pLeft, UINT32 *pRight)
{
	UINT32	left, right;

	left = newStereoPanRamp[63 - stereoPosition] / 16;
	right = newStereoPanRamp[stereoPosition + 63] / 16;
	*pLeft = left;
	*pRight = right;
}
#endif

#if USE_DLS
static const UBYTE stereoPanRamp[] =
{
0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6,
6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12,
12, 13, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 17, 17, 18, 18, 18, 19, 19,
20, 20, 21, 21, 22, 22, 22, 23, 23, 24, 24, 25, 26, 26, 27, 27, 28, 28, 29,
30, 30, 31, 32, 32, 33, 34, 35, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45,
46, 47, 49, 50, 51, 53, 54, 56, 58, 60, 62, 64, 67, 70, 73, 76, 80, 85, 91,
98, 109, 127
};
static void PV_RemapMidiPan(INT32 stereoPosition, UINT32 *pLeft, UINT32 *pRight)
{
	UINT32	left, right;

	left = stereoPanRamp[63 - stereoPosition] * 5;
	right = stereoPanRamp[stereoPosition + 63] * 5;
	*pLeft = left;
	*pRight = right;
}
#endif
#if USE_DLS == 0 && USE_GS_RAMP == 0
static void PV_RemapMidiPan(INT32 stereoPosition, UINT32 *pLeft, UINT32 *pRight)
{
	UINT32	left, right;

	if (stereoPosition)
	{
		stereoPosition *= 2;
		if (stereoPosition < 0)	// left changes
		{
			right = MAX_NOTE_VOLUME + stereoPosition;
//			left = MAX_NOTE_VOLUME - stereoPosition;		// new
			left = MAX_NOTE_VOLUME;
		}
		else
		{					// right changes
			right = MAX_NOTE_VOLUME;
//			right = MAX_NOTE_VOLUME + stereoPosition;		// new
			left = MAX_NOTE_VOLUME - stereoPosition;
		}
	}
	else
	{
		left = MAX_NOTE_VOLUME;
		right = MAX_NOTE_VOLUME;
	}
	*pLeft = left;
	*pRight = right;
}
#endif

// Given a stereo position from -63 to 63, return a volume level from 0 to 127
void PV_CalculateStereoVolume(NoteRecord *this_voice, INT32 *pLeft, INT32 *pRight)
{
	INT32	stereoPosition;
	UINT32	left, right;
	UINT32	noteVolume;

	stereoPosition = this_voice->stereoPosition + this_voice->stereoPanBend;
	stereoPosition *= -1;	// reverse left and right (THIS IS RIGHT)

#if 0
	if (XIsShiftOn())
	{
		stereoPosition = ((XGetMouse() & 0xFFFFL) / 6) - 63;
	}
#endif

	if (stereoPosition > 63)
	{
		stereoPosition = 63;
	}
	if (stereoPosition < -63)
	{
		stereoPosition = -63;
	}

	if (this_voice->NoteChannel != SOUND_EFFECT_CHANNEL)
	{
		// map pan for music channels
		PV_RemapMidiPan(stereoPosition, &left, &right);
	}
	else
	{
		if (stereoPosition)
		{
			// map pan for effects channels
			stereoPosition *= 2;
			if (stereoPosition < 0)	// left changes
			{
				right = MAX_NOTE_VOLUME + stereoPosition;
				left = MAX_NOTE_VOLUME - stereoPosition;		// new
			}
			else
			{					// right changes
				right = MAX_NOTE_VOLUME + stereoPosition;		// new
				left = MAX_NOTE_VOLUME - stereoPosition;
			}
		}
		else
		{
			left = MAX_NOTE_VOLUME;
			right = MAX_NOTE_VOLUME;
		}
	}
	// min out values to handle the fractional case
	if (left <= 1)
	{
		left = 0;
	}
	if (right <= 1)
	{
		right = 0;
	}
	// scale new volume based up channel volume, song volume, and current note volume
	noteVolume = PV_ScaleVolumeFromChannelAndSong(this_voice->pSong, this_voice->NoteChannel, this_voice->NoteVolume);
	noteVolume = (noteVolume * (UINT32)this_voice->NoteVolumeEnvelope) >> VOLUME_PRECISION_SCALAR;

	*pLeft = (left * noteVolume) / MAX_NOTE_VOLUME;
	*pRight = (right * noteVolume) / MAX_NOTE_VOLUME;
}


void SetChannelVolume(GM_Song *pSong, INT16 the_channel, INT16 newVolume)
{
	register GM_Mixer		*myMusicGlobals;
	register LOOPCOUNT		count;
	register NoteRecord *	theNote;

	myMusicGlobals = MusicGlobals;
	// update the current notes playing to the new volume
	for (count = 0; count < myMusicGlobals->MaxNotes; count++)
	{
		theNote = &myMusicGlobals->NoteEntry[count];
		if ( (theNote->NoteDur >= 0) && (theNote->pSong == pSong) )
		{
			if (theNote->NoteChannel == the_channel)
			{
				if (newVolume == 0)
				{
					theNote->NoteDur = 0;
					theNote->NoteDecay = 0;
					theNote->volumeADSRRecord.ADSRTime[0] = 1;
					theNote->volumeADSRRecord.ADSRFlags[0] = ADSR_TERMINATE;
					theNote->volumeADSRRecord.ADSRLevel[0] = 0;	// just in case
				}
				// now calculate the new volume based upon the current channel volume and
				// the unscaled note volume
				newVolume = PV_ScaleVolumeFromChannelAndSong(theNote->pSong, the_channel, theNote->NoteMIDIVolume);
				newVolume = (newVolume * MusicGlobals->scaleBackAmount) >> 8;
				theNote->NoteVolume = newVolume;
			}
		}
	}
}


// Put all notes that have been in 'SUS_HOLD' mode into their normal decay release mode
void PV_ChangeSustainedNotes(GM_Song *pSong, INT16 the_channel, INT16 data)
{
	register GM_Mixer		*myMusicGlobals;
	register LOOPCOUNT		count;
	register NoteRecord *	theNote;

	myMusicGlobals = MusicGlobals;
	for (count = 0; count < myMusicGlobals->MaxNotes; count++)
	{
		theNote = &myMusicGlobals->NoteEntry[count];
		if ( (theNote->NoteDur >= 0) && (theNote->pSong == pSong) )
		{
			if (theNote->NoteChannel == the_channel)
			{
				if (data < 64)	// release. ( 0-63 off, 64-127 on)
				{
					// the note has been released by the fingers, so release note
					if (theNote->sustainMode == SUS_ON_NOTE_OFF)
					{
						theNote->NoteDur = 0;	// decay note out to prevent clicks
					}
					theNote->sustainMode = SUS_NORMAL;
				}
				else
				{	// change status
					theNote->sustainMode = SUS_ON_NOTE_ON;
				}
			}
		}
	}
}

// Set stereo position from control values of 0-127. This will translate into values of 63 to -63
INT16 SetChannelStereoPosition(GM_Song *pSong, INT16 the_channel, UINT16 newPosition)
{
	register GM_Mixer		*myMusicGlobals;
	register LOOPCOUNT		count;
	register NoteRecord *	theNote;
	register INT16			newLogPosition;
	static char stereoScale[] =
	{
		63, 58, 55, 52, 50, 47, 45, 43,		41, 39, 37, 35, 33, 32, 30, 29,
		27, 26, 25, 23, 22, 21, 20, 19, 	18, 17, 17, 16, 15, 14, 14, 13,
		12, 12, 11, 11, 10, 10,  9,  9,		 8,  8,  7,  7,  7,  6,  6,  6,
		 6,  5,  5,  5,  5,  4,  4,  4,		 4,  4,  3,  3,  3,  2,  1,  0,
		 0,
		-1, -2, -3, -3, -3, -4, -4, -4, 	-4, -4, -5, -5, -5, -5, -6, -6,
		-6, -6, -7, -7, -7, -8, -8, -9, 	-9,-10,-10,-11,-11,-12,-12,-13,
	   -14,-14,-15,-16,-17,-17,-18,-19,    -20,-21,-22,-23,-25,-26,-27,-29,
	   -30,-32,-33,-35,-37,-39,-41,-43,    -45,-47,-50,-52,-55,-58,-63
	};

	myMusicGlobals = MusicGlobals;
	// make sure and set the channel stereo position
	newLogPosition = stereoScale[newPosition];
	// update the current notes playing to the new stereo position. It will get incorporated into the mix at the
	// next audio frame
	for (count = 0; count < myMusicGlobals->MaxNotes; count++)
	{
		theNote = &myMusicGlobals->NoteEntry[count];
		if ( (theNote->NoteDur >= 0) && (theNote->pSong == pSong) )
		{
			if (theNote->NoteChannel == the_channel)
			{
				theNote->stereoPosition = newLogPosition;
			}
		}
	}
	return newLogPosition;
}

// Set mod wheel position from control values of 0-127.
void SetChannelModWheel(GM_Song *pSong, INT16 the_channel, UINT16 value)
{
	register GM_Mixer		*myMusicGlobals;
	register LOOPCOUNT		count;
	register NoteRecord *	theNote;

	myMusicGlobals = MusicGlobals;

	// update the current notes playing to the new MOD wheel setting
	for (count = 0; count < myMusicGlobals->MaxNotes; count++)
	{
		theNote = &myMusicGlobals->NoteEntry[count];
		if ( (theNote->NoteDur >= 0) && (theNote->pSong == pSong) )
		{
			if (theNote->NoteChannel == the_channel)
				theNote->ModWheelValue = value;
		}
	}
}


// Change pitch all notes playing on this channel, and for new notes on this channel
INT16 SetChannelPitchBend(GM_Song *pSong, INT16 the_channel, UBYTE bendRange, UBYTE bendMSB, UBYTE bendLSB)
{
	register LOOPCOUNT		count;
	register GM_Mixer		*myMusicGlobals;
	register long			bendAmount, the_pitch_bend;
	register NoteRecord		*pNote;

	myMusicGlobals = MusicGlobals;
	// Convert LSB & MSB into values from -8192 to 8191
	the_pitch_bend = (((bendMSB * 128) + bendLSB) - 8192);

	// Scale values from -8192 to 8192 to -bend to bend in 8.8 fixed point
	bendAmount = bendRange * 256;
	the_pitch_bend = (the_pitch_bend * bendAmount) / 8192;

	// update the current note playing to the new bend value
	for (count = 0; count < myMusicGlobals->MaxNotes; count++)
	{
		pNote = &myMusicGlobals->NoteEntry[count];
		if ( (pNote->NoteDur >= 0) && (pNote->pSong == pSong) )
		{
			if (pNote->NoteChannel == the_channel)
			{
				pNote->NotePitchBend = the_pitch_bend;
			}
		}
	}
	return the_pitch_bend;
}



// Set the global reverb type
void GM_SetReverbType(ReverbMode reverbMode)
{
	if (MusicGlobals)
	{
		if (MusicGlobals->reverbBuffer)
		{
			switch (reverbMode)
			{
				case REVERB_TYPE_1:
				case REVERB_TYPE_2:
				case REVERB_TYPE_3:
				case REVERB_TYPE_4:
				case REVERB_TYPE_5:
				case REVERB_TYPE_6:
				case REVERB_TYPE_7:
				case REVERB_TYPE_8:
					break;
				case REVERB_NO_CHANGE:	// no change
				default:
					reverbMode = -1;
					break;
			}
		}
		else
		{
			reverbMode = REVERB_TYPE_1;		// none
		}
		if (reverbMode != -1)
		{
			MusicGlobals->reverbUnitType = reverbMode;
		}
	}
}

ReverbMode GM_GetReverbType(void)
{
	ReverbMode reverbMode;

	reverbMode = REVERB_TYPE_1;
	if (MusicGlobals)
	{
		reverbMode = MusicGlobals->reverbUnitType;
	}
	return reverbMode;
}

UINT32 GM_GetSamplesPlayed(void)
{
    return MusicGlobals->samplesPlayed;
}

void GM_SetAudioTask(GM_AudioTaskCallbackPtr theTask)
{
	if (MusicGlobals)
	{
		MusicGlobals->pTaskProc = theTask;
	}
}

void GM_SetAudioOutput(GM_AudioOutputCallbackPtr theHook)
{
	if (MusicGlobals)
	{
		MusicGlobals->pOutputProc = theHook;
	}
}

GM_AudioTaskCallbackPtr GM_GetAudioTask(void)
{
	if (MusicGlobals)
	{
		return MusicGlobals->pTaskProc;
	}
	return NULL;
}

GM_AudioOutputCallbackPtr GM_GetAudioOutput(void)
{
	if (MusicGlobals)
	{
		return MusicGlobals->pOutputProc;
	}
	return NULL;
}

#if USE_HAE_EXTERNAL_API
// Connect to hardware.
//
// Global variables must be set before this can be called.
//
//	MusicGlobals->generate16output
//	MusicGlobals->generateStereoOutput
//	MusicGlobals->outputQuality
//
// Return FALSE if failure, otherwise TRUE
XBOOL GM_StartHardwareSoundManager(void *context)
{
	long	sampleRate;
	int		ok;

	if (MusicGlobals)
	{
		switch (MusicGlobals->outputQuality)
		{
			case Q_44K:
				sampleRate = 44100;
				break;
			case Q_22K:
			case Q_11K:
				sampleRate = 22050;
				break;
		}

		ok = HAE_AquireAudioCard(context, sampleRate,
									(MusicGlobals->generateStereoOutput) ? 2 : 1,
									(MusicGlobals->generate16output) ? 16 : 8);
		return (ok == 0) ? TRUE : FALSE;
	}
	return FALSE;
}

// Stop generating samples and shutdown
void GM_StopHardwareSoundManager(void *context)
{
	// need this so that we can properly update samplesPlayed to the total written
	// to the *device* when we close the audio device.  UpdateSamplesPlayed() needs
	// to take the current device playback position as its argument.  when we shut
	// down the device, we want to move it to represent everything submitted.  if the
	// device has been opened and closed before, the device-end-position is different
	// than MusicGlobals->samplesWritten.
	static UINT32 lastSamplesWritten = 0;

	// everything that's going to play has been played; samples played by mixer
	// should equal samples submitted.  if we don't synch these here, our count
	// of samples played and samples submitted to device diverge after closing
	// and reopening the device.

	HAE_ReleaseAudioCard(context);
	if (MusicGlobals)
	{
		GM_UpdateSamplesPlayed((MusicGlobals->samplesWritten - lastSamplesWritten));
		lastSamplesWritten = MusicGlobals->samplesWritten;
	}
}

// Get current audio time stamp in microseconds; this is the
// microseconds' worth of samples that have passed through the
// audio device.  it never decreases.
// $$kk: this and all the time stamp methods should move into a common file
// CLS:  copied this function in from Kara's
UINT32 GM_GetDeviceTimeStamp(void)
{
	UINT16	sampleRate;

	if (MusicGlobals)
	{
		// convert from samples into microseconds
		sampleRate = MusicGlobals->outputQuality == Q_44K ? 44100 : 22050;
#if USE_FLOAT == FALSE
		return (MusicGlobals->samplesPlayed * 1000000) / sampleRate;
#else
		return (UINT32)(((float) MusicGlobals->samplesPlayed / sampleRate) * 1000000);
#endif
	}
	return 0L;
}

// Update count of samples played.  This function caluculates from number of bytes,
// given the sample frame size from the mixer variables
// $$kk: no, we're getting the currentPos in SAMPLES, not bytes, from HAE_GetDeviceSamplesPlayedPosition().
void GM_UpdateSamplesPlayed(unsigned long currentPos)
{
	long delta = 0;

	if (currentPos >= MusicGlobals->lastSamplePosition)
	{
//		delta = ((currentPos - MusicGlobals->lastSamplePosition) / MusicGlobals->sampleFrameSize);
		delta = (currentPos - MusicGlobals->lastSamplePosition);
	}
	else
	{
//		delta = (currentPos / MusicGlobals->sampleFrameSize);
		delta = currentPos;
	}

	MusicGlobals->lastSamplePosition = currentPos;

	// update mixer samples played
	MusicGlobals->samplesPlayed += delta;
//	fprintf(stderr, "GM_UpdateSamplesPlayed: updating MusicGlobals->samplesPlayed by %d to %d \n", delta, MusicGlobals->samplesPlayed);

	// update samples played for each stream
	GM_AudioStreamUpdateSamplesPlayed(delta);
}

// number of devices. ie different versions of the HAE connection. DirectSound and waveOut
// return number of devices. ie 1 is one device, 2 is two devices.
// NOTE: This function needs to function before any other calls may have happened.
long GM_MaxDevices(void)
{
	return HAE_MaxDevices();
}

// set the current device. device is from 0 to GM_MaxDevices()
// NOTE:	This function needs to function before any other calls may have happened.
//			Also you will need to call HAE_ReleaseAudioCard then HAE_AquireAudioCard
//			in order for the change to take place.
void GM_SetDeviceID(long deviceID)
{
	HAE_SetDeviceID(deviceID);
}

// return current device ID
// NOTE: This function needs to function before any other calls may have happened.
long GM_GetDeviceID(void)
{
	return HAE_GetDeviceID();
}

// get deviceID name
// NOTE:	This function needs to function before any other calls may have happened.
//			Format of string is a zero terminated comma delinated C string.
//			"platform,method,misc"
//	example	"MacOS,Sound Manager 3.0,SndPlayDoubleBuffer"
//			"WinOS,DirectSound,multi threaded"
//			"WinOS,waveOut,multi threaded"
//			"WinOS,VxD,low level hardware"
//			"WinOS,plugin,Director"
void GM_GetDeviceName(long deviceID, char *cName, unsigned long cNameLength)
{
	HAE_GetDeviceName(deviceID, cName, cNameLength);
}

// Get current audio time stamp based upon the audio built interrupt
UINT32 GM_GetSyncTimeStamp(void)
{
	if (MusicGlobals)
	{
		return MusicGlobals->syncCount;
	}
	return 0L;
}

long GM_GetAudioBufferOutputSize(void)
{
	return HAE_GetAudioByteBufferSize();
}

// Get current audio time stamp based upon the audio built interrupt, but ahead in time and quantized for
// the particular OS
UINT32 GM_GetSyncTimeStampQuantizedAhead(void)
{
	return GM_GetSyncTimeStamp() + (BUFFER_SLICE_TIME * HAE_GetAudioBufferCount());
}



#endif

/* EOF of GenSetup.c
*/

