/*
 * @(#)GenSnd.h	1.25 98/04/16
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
** "GenSnd.h"
**
**	Generalized Music Synthesis package. Part of SoundMusicSys.
**
**
** Overview
**	The purpose of this layer of code is to remove Macintosh Specific code.
**	No file, or memory access. All functions are passed pointers to data
**	that needs to be passed into the mixer, and MIDI sequencer
**
** Modification History
**
**	4/6/93		Created
**	4/12/93		First draft ready
**	4/14/93		Added Waveform structure
**	7/7/95		Added Instrument API
**	11/7/95		Major changes, revised just about everything.
**	11/11/95	Added external queued midi links
**	11/20/95	Removed the BF_ flags, now you must walk through the union structure
**				Remove bit fields. BIT FIELDS DON'T WORK WITH MPW!!!!
**	12/5/95		Added avoidReverb into instrument field; removed drop sample case
**	12/6/95		Move REVERB_TYPE from GENPRIV.H
**				Added GM_SetReverbType; removed extern references
**				Added ReverbType to GM_Song structure
**				Removed defaultPlaybackRate & defaultInterpolationMode from the GM_Song
**				structure
**	12/7/95		Moved DEFAULT_REVERB_TYPE from GenSnd.c
**				Added GM_GetReverbType
**	1/4/96		Added GM_ChangeSampleReverb for sound effects
**	1/7/96		Changed GM_BeginDoubleBuffer to use a 32 bit value for volume
**				Added GM_SetEffectsVolume & GM_GetEffectsVolume
**	1/13/96		Added extendedFormat bit to internal instrument format
**	1/18/96		Spruced up for C++ extra error checking
**	1/19/96		Changed GM_BeginSample to support bitsize and channels
**	1/29/96		Changed WaveformInfo to support FIXED_VALUE for sample rate
**				Added useSampleRate factor for playback of instruments & sampleAndHold bits
**	2/4/96		Added songMidiTickLength to GM_Song
**	2/5/96		Moved lots of variables from the MusicVars structure into
**				the GM_Song structure.
**				Changed GM_EndSong & GM_SongTicks & GM_IsSongDone & 
**				GM_SetMasterSongTempo to pass in a GM_Song pointer
**				Added GM_GetSongTickLength
**	2/12/96		Added GM_SetSongTickPosition
**				Added songMicrosecondLength to GM_Song structure
**	2/14/96		Added GM_StartLiveSong
**	2/18/96		Added panPlacement to the GM_Instrument structure
**	2/29/96		Added trackMuted to GM_Song structure
**	3/5/96		Added MAX_SONG_VOLUME
**	4/15/96		Added support to interpret SONG resource via GM_MergeExternalSong
**	4/20/96		Added defines for max lfos, and max curves, MAX_LFOS & MAX_CURVES
**	4/21/96		Added GM_GetRealtimeAudioInformation
**	6/7/96		Increased MAX_TRACKS to 65 to include tempo track and 64 tracks
**	6/9/96		Added GM_GetUserReference & GM_SetUserReference
**	6/30/96		Changed font and retabbed
**	7/2/96		Added packing pragmas
**				Removed usage of Machine.h. Now merged into X_API.h
**	7/3/96		Added support for one level of loop save and playback
**				Required for Headspace support
**	7/5/96		Added GM_KillAllNotes
**	7/6/96		Added GM_GetSyncTimeStamp
**	7/14/96		Fixed structure alignment issue for PowerPC
**	7/23/96		Changed QGM_ functions to use an unsigned long for time stamp
**	7/25/96		Added GM_GetSyncTimeStampQuantizedAhead
**	8/15/96		Added constant for LOW_PASS_AMOUNT & LPF_DEPTH
**	8/19/96		Added GM_SetAudioTask
**	9/10/96		Added GM_NoteOn & GM_NoteOff & GM_ProgramChange & GM_PitchBend & 
**				GM_Controller & GM_AllNotesOff for direct control to bypass the queue
**	9/17/96		Added GM_LoadSongInstrument & GM_UnloadSongInstrument
**				Added Q_GET_TICK
**	9/18/96		Changed GM_SongCallbackProcPtr to pass a GM_Song structure 
**				rather than an ID
**	9/19/96		Added GM_GetSongTempo
**	9/20/96		Added GM_SetSongTempo & GM_ResumeSong & GM_PauseSong
**	9/23/96		Added GM_MuteTrack & GM_UnmuteTrack & GM_MuteChannel & GM_UnmuteChannel
**	9/24/96		Added GM_SetSongTempInBeatsPerMinute & GM_GetSongTempoInBeatsPerMinute
**				Added GM_SoloTrack & GM_SoloChannel
**				Added GM_GetSongPitchOffset & GM_SetSongPitchOffset
**	9/25/96		Added GM_GetChannelMuteStatus & GM_GetTrackMuteStatus
**				Added controller 90 to change the global reverb type
**				Added GM_EndSongNotes
**	10/8/96		Removed pascal from GM_SongCallbackProcPtr
**	10/11/96	Added GM_BeginSampleFromInfo
**	10/13/96	Changed QGM_AllNotesOff to work with the queue and post an event
**				Added GM_ReadIntoMemoryWaveFile & GM_ReadIntoMemoryAIFFFile
**	10/18/96	Made WaveformInfo smaller
**	10/23/96	Removed reference to BYTE and changed them all to UBYTE or SBYTE
**				Added defines for MOD_WHEEL_CONTROL = 'MODW & SAMPLE_NUMBER = 'SAMP'
**				Added more defines for instrument types
**	10/28/96	Modified QGM_NoteOn & QGM_NoteOff & QGM_ProgramChange & 
**				QGM_PitchBend & QGM_Controller & QGM_AllNotesOff to accept a GM_Song *
**	10/31/96	Changed trackMuted and channelMuted to be bit flag based
**				Added soloMuted bit array for solo control
**	10/31/96	Added GM_IsInstrumentLoaded
**	11/3/96		Added midiNote to GM_AudioInfo structure
**	11/5/96		Changed WaveformInfo to GM_Waveform
**				Added GM_ReadFileInformation and GM_FreeWaveform
**	11/6/96		Added GM_UnsoloTrack, GM_GetTrackSoloStatus
**	11/7/96		Added GM_SongTimeCallbackProcPtr & GM_SetSongTimeCallback
**	11/8/96		Added GM_GetSampleVolume & GM_GetSamplePitch & GM_GetSampleStereoPosition
**	11/9/96		Added GM_KillSongNotes
**	11/11/96	Added more error codes and removed extra ones
**	11/14/96	Added pSong reference in GM_GetRealtimeAudioInformation
**	11/19/96	Changed MAX_CHANNELS to 17, 16 for Midi, 1 for sfx
**				Added a GM_Song structure to GM_ConvertPatchBank and removed bank
**	11/21/96	Removed GM_ConvertPatchBank
**	11/26/96	Changed MAX_BANKS to 6
**				Added GM_GetControllerValue
**	12/2/96		Added MOD file code API
**	12/9/96		Added GM_LoadInstrumentFromExternal
**	12/15/96	Added controls for DEFAULT_VELOCITY_CURVE
**	12/19/96	Added Sparc pragmas
**	1/2/97		Moved USE_MOD_API and USE_STREAM_API into X_API.h
**	1/12/97		Changed maxNormalizedVoices to mixLevel
**	1/16/97		Changed LFORecord to LFORecords
**	1/22/97		Added GM_SetSampleDoneCallback
**	1/23/97		Added M_STEREO_FILTER
**				Added GM_SetAudioOutput
**	1/24/97		Added GM_SetSongFadeRate
**				Added GM_SetModLoop & GM_GetModTempoBPM & GM_SetModTempoBPM
**				Changed disposeMidiDataWhenDone to disposeSongDataWhenDone
**				Added GM_SetAudioStreamFadeRate
**	1/28/97		Added more parmeters to GM_SetSongFadeRate to support async
**				ending of song
**	1/28/97		Eliminated terminateDecay flag. Not used anymore
**	1/30/97		Changed SYMPHONY_SIZE to MAX_VOICES
**	2/1/97		Added support for pitch offset control on a per channel basis
**				Added GM_DoesChannelAllowPitchOffset & GM_AllowChannelPitchOffset
**	2/2/97		Tightened up GM_Song and GM_Instrument a bit
**	2/13/97		Added GM_GetSystemVoices	
**	2/20/97		Added support for TYPE 7 and 8 reverb in GM_SetReverbType
**	3/12/97		Added a REVERB_NO_CHANGE reverb mode. This means don't change the
**				current mixer state
**	3/20/97		Added GM_SetSongMicrosecondPosition & GM_GetSongMicrosecondLength
**	3/27/97		Added channelChorus
**	3/27/97		Changed all 4 character constants to use the FOUR_CHAR macro
**	4/14/97		Changed KeymapSplit to GM_KeymapSplit
**	4/20/97		Added GM_SetSongMetaEventCallback
**	4/21/97		Removed enableInterpolate from GM_Instrument structure
**	5/1/97		Added startOffsetFrame to GM_BeginSampleFromInfo
**	5/19/97		Added GM_ReadAndDecodeFileStream
**	6/10/97		Added GM_SetModReverbStatus to allow for verb enabled Mod files
**	6/16/97		Added GM_AudioStreamSetVolumeAll
**	6/25/97		Added GM_SetSampleLoopPoints
**	7/8/97		Added or changed GM_GetSongLoopFlag & GM_SetSongLoopMax & 
**				GM_SetSongLoopFlag & GM_GetSongLoopMax
**	7/15/97		Added GM_GetAudioOutput & GM_GetAudioTask & GM_GetAudioBufferOutputSize
**	7/17/97		Aligned GM_Song and GM_Instrument structures to 8 bytes
**	7/21/97		Put mins and maxs on GM_SetSongTempInBeatsPerMinute. 0 and 500
**				Changed GM_SetSongTempInBeatsPerMinute & GM_GetSongTempoInBeatsPerMinute
**				GM_SetSongTempo & GM_GetSongTempo & GM_SetMasterSongTempo to all be unsigned longs
**				Changed key calculations to floating point to test for persicion loss. Use
**				USE_FLOAT in GenSnd.h to control this change.
**	7/22/97		Removed MicroJif from the MusicVars structure. Now using constant BUFFER_SLICE_TIME
**				Added GM_GetSampleVolumeUnscaled
**	7/28/97		Moved GM_FreeWaveform so it is compiled in all platform cases.
**				Removed define wrapper from errors codes in defining OPErr.
**				Fixed a praga pack problem for Sun
**	8/6/97		Moved USE_FLOAT to X_API.h
**  8/06/97		(ddz) Added PgmEntry structure def. Added midiSize, TSNumerator, TSDenominator,
**				pgmChangeInfo, and trackcumuticks[] fields to GM_Song structure for
**				implementing the Song Program List capability
**				Changed GM_SongMetaCallbackProcPtr and added currenTrack
**	8/13/97		Renamed GM_GetSongProgramChanges to GM_GetSongInstrumentChanges and changed
**				Byte reference to XBYTE
**	8/27/97		Moved GM_StartHardwareSoundManager & GM_StopHardwareSoundManager from
**				GenPriv.h
**	9/15/97		Added GM_GetSampleReverb && GM_AudioStreamGetReverb
**  		    Added PatchInfo structure to combine some of instrument scan variables into one ptr
**              Changed PgmEntry to InstrumentEntry, added bank select fields
**	9/25/97		Added TrackStatus to represent track status rather than 'R' and 'F'. Symbolic use.
**	10/15/97	Added processingSlice to GM_Song and GM_Instrument to handle threading release issues.
**	10/16/97	Changed GM_LoadSong parmeters to include an option to ignore bad instruments
**				when loading.
**				Changed GM_LoadSongInstruments to use a XBOOL for a flag rather than an int
**				Added GM_AnyStereoInstrumentsLoaded
**				Added GM_CacheSamples
**	11/12/97	Added GM_MaxDevices & GM_SetDeviceID & GM_GetDeviceID & GM_GetDeviceName
**	11/17/97	Added GM_AudioStreamGetSamplesPlayed & GM_AudioStreamUpdateSamplesPlayed from Kara
**	11/18/97	Added GM_AudioStreamResumeAll & GM_AudioStreamPauseAll
**
**  JAVASOFT
**	01.14.98	kk: added numLoops field to GM_Waveform struct.
**				added theLoopTarget to GM_BeginSample
**	03.17.98:	kk: added UNSUPPORTED_HARDWARE to OPErr
*/
/*****************************************************************************/

#ifndef G_SOUND
#define G_SOUND

#ifndef __X_API__
	#include "X_API.h"
#endif

#ifdef __cplusplus
	extern "C" {
#endif

/* System defines */


/* Used in InitGeneralSound */

// Quality types
enum
{
	Q_11K = 0,
	Q_22K,
	Q_44K
};
typedef long Quality;

// Modifier types
#define M_NONE				0L
#define M_USE_16			(1<<0L)
#define M_USE_STEREO		(1<<1L)
#define M_DISABLE_REVERB	(1<<2L)
#define M_STEREO_FILTER		(1<<3L)
typedef long AudioModifiers;

// Interpolation types
enum
{
	E_AMP_SCALED_DROP_SAMPLE = 0,
	E_2_POINT_INTERPOLATION,
	E_LINEAR_INTERPOLATION
};
typedef long TerpMode;

enum 
{
	REVERB_NO_CHANGE = 0,	// Don't change current mixer settings
	REVERB_TYPE_1 = 1,		// None
	REVERB_TYPE_2,			// Igor's Closet
	REVERB_TYPE_3,			// Igor's Garage
	REVERB_TYPE_4,			// Igor's Acoustic Lab
	REVERB_TYPE_5,			// Igor's Cavern
	REVERB_TYPE_6,			// Igor's Dungeon
	REVERB_TYPE_7,			// WebTV Reverb
	REVERB_TYPE_8			// variable reverb
};
typedef char ReverbMode;
#define DEFAULT_REVERB_TYPE	REVERB_TYPE_4
#define MAX_REVERB_TYPES	9
#define MAX_REVERB_TAPS		24

typedef enum 
{
	SCAN_NORMAL = 0,		// normal Midi scan
	SCAN_SAVE_PATCHES,		// save patches during Midi scan. Fast
	SCAN_DETERMINE_LENGTH,	// calculate tick length. Slow
	SCAN_FIND_PATCHES,		// mode for scanning to find the program changes
	SCAN_COUNT_PATCHES      // mode for counting the number of program changes
} ScanMode;

typedef enum 
{
	SAFE_TO_ACCESS = 0,		// memory structure is safe to access
	ASK_FOR_ACCESS,			// ask for access
	ACKNOWLEGE_ACCESS		// acknowlege access
} AccessSemaphore;

// used in GM_Song->trackon to track progress of a midi file
enum 
{
	TRACK_OFF = 0,
	TRACK_FREE,
	TRACK_RUNNING
};
typedef unsigned char TrackStatus;

enum
{
	VELOCITY_CURVE_1 = 0,	// default S curve
	VELOCITY_CURVE_2,		// more peaky S curve
	VELOCITY_CURVE_3,		// inward curve centered around 95 (used for WebTV)
	VELOCITY_CURVE_4,		// two time exponential
	VELOCITY_CURVE_5		// two times linear
};
#define DEFAULT_VELOCITY_CURVE	VELOCITY_CURVE_1

#define MAX_VOICES				64		// max voices at once
#define MAX_INSTRUMENTS			128		// MIDI number of programs per patch bank
#define MAX_BANKS				6		// three GM banks; three user banks
#define MAX_TRACKS				65		// max MIDI file tracks to process (64 + tempo track)
#define MAX_CHANNELS			17		// max MIDI channels + one extra for sound effects
#define MAX_CONTROLLERS			128		// max MIDI controllers
#define MAX_SONG_VOLUME			127
#define MAX_NOTE_VOLUME			127		// max note volume
#define MAX_CURVES				4		// max curve entries in instruments
#define MAX_LFOS				6		// max LFO's, make sure to add one extra for MOD wheel support
#define MAX_MASTER_VOLUME		256		// max volume level for master volume level
#define MAX_SAMPLES				768		// max number of samples that can be loaded
#define MAX_SONGS				8		// max number of songs that can play at one time
#define DEFAULT_PITCH_RANGE		2		// number of semi-tones that change with pitch bend wheel
#define PERCUSSION_CHANNEL		9		// which channel (zero based) is the default percussion channel

/* Common errors returned from the system */
typedef enum
{
	NO_ERR = 0,
	PARAM_ERR,
	MEMORY_ERR,
	BAD_SAMPLE,
	BAD_INSTRUMENT,
	BAD_MIDI_DATA,
	ALREADY_PAUSED,
	ALREADY_RESUMED,
    DEVICE_UNAVAILABLE,
	NO_SONG_PLAYING,
	STILL_PLAYING,
	NO_VOLUME,
	TOO_MANY_SONGS_PLAYING,
	BAD_FILE,
	NOT_REENTERANT,
	NOT_SETUP,
	BUFFER_TO_SMALL,
	NO_FREE_VOICES,
// stream API errors
	STREAM_STOP_PLAY,
	BAD_FILE_TYPE,
	GENERAL_BAD,
	UNSUPPORTED_HARDWARE
} OPErr;

// Need a forward reference to the GM_Song struct to keep
// our compiler from complaining.
struct GM_Song;

typedef BOOL_FLAG	(*GM_LoopDoneCallbackPtr)(INT32 refnum);
typedef void		(*GM_DoubleBufferCallbackPtr)(INT32 refnum, G_PTR pWhichBufferFinished, INT32 *pBufferSize);
typedef void		(*GM_SoundDoneCallbackPtr)(INT32 refnum);
typedef void		(*GM_ControlerCallbackPtr)(struct GM_Song *pSong, void * reference, short int channel, short int track, short int controler, short int value);
typedef	void		(*GM_SongCallbackProcPtr)(struct GM_Song *pSong);
typedef	void		(*GM_SongTimeCallbackProcPtr)(struct GM_Song *pSong, UINT32 currentMicroseconds, UINT32 currentMidiClock);
typedef	void		(*GM_SongMetaCallbackProcPtr)(struct GM_Song *pSong, char markerType, void *pMetaText, long metaTextLength, INT16 currentTrack);
typedef void		(*GM_AudioTaskCallbackPtr)(void *context, unsigned long ticks);
typedef void		(*GM_AudioOutputCallbackPtr)(void *context, void *samples, long sampleSize, long channels, unsigned long lengthInFrames);
typedef void		(*GM_SampleFrameCallbackPtr)(INT32 reference, INT32 sampleFrame);



#if CPU_TYPE == kRISC
	#pragma options align=power
#endif
/* Change pack(8) to pack(4) for SPARC.				-Liang */
#if (CPU_TYPE == kSPARC)
	#pragma pack (4)
#endif
/* $$kk: pack(4) for solaris x86 */
#if (CPU_TYPE == k80X86)
#if (X_PLATFORM == X_SOLARIS)
	#pragma pack (4)
#else
	#pragma pack (8)
#endif
#endif

struct GM_SampleCallbackEntry
{
	unsigned long					frameOffset;
	GM_SampleFrameCallbackPtr		pCallback;
	long							reference;
	struct GM_SampleCallbackEntry	*pNext;
};
typedef struct GM_SampleCallbackEntry GM_SampleCallbackEntry;

struct GM_KeymapSplit
{
	UBYTE					lowMidi;
	UBYTE					highMidi;
	INT16					miscParameter1;		// can be smodParmeter1 if enableSoundModifier
												// enabled, otherwise its a replacement
												// rootKey for sample
	INT16					miscParameter2;
	struct GM_Instrument	*pSplitInstrument;
};
typedef struct GM_KeymapSplit GM_KeymapSplit;

// Flags for embedded midi objects
#define EM_FLUSH_ID				FOUR_CHAR('F','L','U','S')	//	'FLUS'		// immediate command
#define EM_CACHE_ID				FOUR_CHAR('C','A','C','H')	//	'CACH'		// immediate command
#define EM_DATA_ID				FOUR_CHAR('D','A','T','A')	//	'DATA'		// block resources

// Flags for modules
#define ADSR_ENVELOPE			FOUR_CHAR('A','D','S','R')	//	'ADSR'
#define LOW_PASS_FILTER			FOUR_CHAR('L','P','G','F')	//	'LPGF'
#define LOW_PASS_AMOUNT			FOUR_CHAR('L','P','A','M')	//	'LPAM'
#define DEFAULT_MOD				FOUR_CHAR('D','M','O','D')	//	'DMOD'
#define EXPONENTIAL_CURVE		FOUR_CHAR('C','U','R','V') 	//	'CURV'


// Flags for ADSR module
#define ADSR_LINEAR_RAMP 		FOUR_CHAR('L','I','N','E')	//	'LINE'
#define ADSR_SUSTAIN 			FOUR_CHAR('S','U','S','T')	//	'SUST'
#define ADSR_TERMINATE			FOUR_CHAR('L','A','S','T')	//	'LAST'
#define ADSR_GOTO				FOUR_CHAR('G','O','T','O')	//	'GOTO'
#define ADSR_GOTO_CONDITIONAL	FOUR_CHAR('G','O','S','T')	//	'GOST'
#define ADSR_RELEASE			FOUR_CHAR('R','E','L','S')	//	'RELS'
#define ADSR_STAGES				8

struct ADSRRecord
{
	INT32			currentTime;
	INT32			currentPosition;
	INT32			currentLevel;
	INT32			mode;
	INT32			previousTarget;
	INT32			sustainingDecayLevel;
	INT32			ADSRLevel[ADSR_STAGES];
	INT32			ADSRTime[ADSR_STAGES];
	INT32			ADSRFlags[ADSR_STAGES];
};
typedef struct ADSRRecord ADSRRecord;

// kinds of LFO modules
#define VOLUME_LFO				FOUR_CHAR('V','O','L','U')	// 'VOLU'
#define PITCH_LFO				FOUR_CHAR('P','I','T','C')	// 'PITC'
#define STEREO_PAN_LFO			FOUR_CHAR('S','P','A','N')	// 'SPAN'
#define STEREO_PAN_NAME2		FOUR_CHAR('S','A','N',' ')	// 'PAN '
#define LPF_FREQUENCY			FOUR_CHAR('L','P','F','R')	// 'LPFR'
#define LPF_DEPTH				FOUR_CHAR('L','P','R','E')	// 'LPRE'
#define LOW_PASS_AMOUNT			FOUR_CHAR('L','P','A','M')	// 'LPAM'

// kinds of LFO wave shapes
#define SINE_WAVE 				FOUR_CHAR('S','I','N','E')	// 'SINE'
#define TRIANGLE_WAVE 			FOUR_CHAR('T','R','I','A')	// 'TRIA'
#define SQUARE_WAVE 			FOUR_CHAR('S','Q','U','A')	// 'SQUA'
#define SQUARE_WAVE2 			FOUR_CHAR('S','Q','U','2')	// 'SQU2'
#define SAWTOOTH_WAVE 			FOUR_CHAR('S','A','W','T')	// 'SAWT'
#define SAWTOOTH_WAVE2 			FOUR_CHAR('S','A','W','2')	// 'SAW2'

// Additional elements used by Curve functions
#define LOWPASS_AMOUNT			FOUR_CHAR('L','P','A','M')	// 'LPAM'
#define VOLUME_LFO_FREQUENCY 	FOUR_CHAR('V','O','L','F')	// 'VOLF'
#define PITCH_LFO_FREQUENCY 	FOUR_CHAR('P','I','T','F')	// 'PITF'
#define NOTE_VOLUME				FOUR_CHAR('N','V','O','L')	// 'NVOL'
#define VOLUME_ATTACK_TIME 		FOUR_CHAR('A','T','I','M')	// 'ATIM'
#define VOLUME_ATTACK_LEVEL 	FOUR_CHAR('A','L','E','V')	// 'ALEV'
#define SUSTAIN_RELEASE_TIME 	FOUR_CHAR('S','U','S','T')	// 'SUST'
#define SUSTAIN_LEVEL			FOUR_CHAR('S','L','E','V')	// 'SLEV'
#define RELEASE_TIME			FOUR_CHAR('R','E','L','S')	// 'RELS'
#define WAVEFORM_OFFSET			FOUR_CHAR('W','A','V','E')	// 'WAVE'
#define SAMPLE_NUMBER			FOUR_CHAR('S','A','M','P')	// 'SAMP'
#define MOD_WHEEL_CONTROL		FOUR_CHAR('M','O','D','W')	// 'MODW'

// Controls for registered parameter status
enum
{
	USE_NO_RP = 0,		// no registered parameters selected
	USE_NRPN,			// use non registered parameters
	USE_RPN				// use registered parameters
};

// Controls for channelBankMode
enum
{
	USE_GM_DEFAULT = 0,			// this is default behavior
								// normal bank for channels 1-9 and 11-16
								// percussion for channel 10
	USE_NON_GM_PERC_BANK,		// will force the use of the percussion
								// bank reguardless of channel and allow program
								// changes to reflect the percussion instrments and
								// allow you to change them with normal notes
	USE_GM_PERC_BANK,			// will force the use of the percussion
								// bank reguardless of channel and act just like a GM
								// percussion channel. ie. midi note represents the instrument
								// to play
	USE_NORM_BANK				// will force the use of the normal
								// bank reguardless of channel
	
};

struct LFORecord
{
	INT32	period;
	INT32	level;
	INT32	mode;
	INT32	currentTime, LFOcurrentTime;
	INT32	currentWaveValue;
	INT32	where_to_feed;
	INT32	DC_feed;
	INT32	waveShape;
	ADSRRecord a;
};
typedef struct LFORecord LFORecord;

struct CurveEntry
{
	INT16	from_Value;
	INT16	to_Scalar;
};
typedef struct CurveEntry CurveEntry;
struct CurveRecord
{
	INT32		tieFrom;
	INT32		tieTo;
	INT16		curveCount;
	CurveEntry	curveData[MAX_CURVES];	// set larger if need be.
};
typedef struct CurveRecord CurveRecord;

struct InstrumentEntry
{
	struct InstrumentEntry *next; // link
	UINT32		ticks;
	UBYTE		*ptr;	// location in MIDI data
	UBYTE		*bsPtr; // location of bank select pointer
	INT16		track;
	UBYTE		channel;
	UBYTE		program;
	UBYTE		bank;
	UBYTE		dirty;
	UBYTE		bankMode; // bank mode at time of pgm change
};
typedef struct InstrumentEntry InstrumentEntry;

// structure for maintaining info about patch changes in a GM_Song

struct PatchInfo
{
	InstrumentEntry		*instrChangeInfo;
	UBYTE				*instrPtrLoc; // used to communicate location of a program change (move dec. elsewhere?)
	UBYTE				*bankPtrLoc; // location of a bank change msg
	INT32				instrCount; // count of program change messages found
	INT32				rsCount;   // count of PCs with running status (unused right now)
	INT32				streamIncrement; // how much to add to midi_stream ptr after
							// PV_ProcessProgramChange has modified the data
};
typedef struct PatchInfo PatchInfo;

struct GM_KeymapSplitInfo
{
	UINT16 			defaultInstrumentID;
	UINT16			KeymapSplitCount;
	GM_KeymapSplit	keySplits[1];
};
typedef struct GM_KeymapSplitInfo GM_KeymapSplitInfo;

struct GM_Waveform
{
	INT32		waveformID;			// extra specific data, or position for files
	UBYTE		bitSize;			// number of bits per sample
	UBYTE		channels;			// number of channels (1 or 2)
	UBYTE		baseMidiPitch;		// base Midi pitch of recorded sample ie. 60 is middle 'C'
	UBYTE		noteDecayPref;		// Note Decay in 1/60th of release
	INT32		waveSize;			// total waveform size in bytes
	INT32		waveFrames;			// number of frames
	INT32		startLoop;			// start loop point offset
	INT32		endLoop;			// end loop point offset
	
	// $$kk: 01.14.98: added numLoops field
	INT32		numLoops;			// number of loops between loop points before continuing to end of sample

	FIXED_VALUE	sampledRate;		// FIXED_VALUE 16.16 value for recording
	SBYTE 		*theWaveform;		// array data that morphs into what ever you need
};
typedef struct GM_Waveform GM_Waveform;


// Internal Instrument structure
// aligned structure to 8 bytes
struct GM_Instrument
{
	AccessSemaphore		accessStatus;
	INT16				masterRootKey;
	INT16				smodResourceID;
	INT16				miscParameter1;
	INT16				miscParameter2;

	INT16				panPlacement;			// inital stereo pan placement of this instrument
	BOOL_FLAG	/*0*/	disableSndLooping;		// Disable waveform looping
	BOOL_FLAG	/*1*/	neverInterpolate;		// seems redundant: Never interpolate
	BOOL_FLAG	/*2*/	playAtSampledFreq;		// Play instrument at sampledRate only
	BOOL_FLAG	/*3*/	doKeymapSplit;			// If TRUE, then this instrument is a keysplit defination
	BOOL_FLAG	/*4*/	notPolyphonic;			// if FALSE, then instrument is a mono instrument
	BOOL_FLAG	/*5*/	enablePitchRandomness;	// if TRUE, then enable pitch variance feature

	BOOL_FLAG	/*6*/	avoidReverb;			// if TRUE, this instrument is not mixed into reverb unit
	BOOL_FLAG	/*7*/	enableSoundModifier;
	BOOL_FLAG	/*0*/	extendedFormat;			// extended format instrument
	BOOL_FLAG	/*1*/	sampleAndHold;
	BOOL_FLAG	/*2*/	useSampleRate;			// factor in sample rate into pitch calculation
	BOOL_FLAG	/*3*/	processingSlice;
	BOOL_FLAG	/*4*/	useSoundModifierAsRootKey;
	UBYTE		/*5*/	usageReferenceCount;	// number of references this instrument is associated to

	UBYTE		/*6*/	LFORecordCount;
	UBYTE		/*7*/	curveRecordCount;
	INT16				alignement_0;
	INT32				LPF_frequency;

	INT32				LPF_resonance;
	INT32				LPF_lowpassAmount;

	LFORecord			LFORecords[MAX_LFOS];
	ADSRRecord			volumeADSRRecord;
	CurveRecord			curve[MAX_CURVES];
	union
	{
		GM_KeymapSplitInfo	k;
		GM_Waveform			w;
	} u;
};
typedef struct GM_Instrument GM_Instrument;

struct GM_ControlCallback
{
	// these pointers are NULL until used, then they are allocated
	GM_ControlerCallbackPtr	callbackProc[MAX_CONTROLLERS];		// current controller callback functions
	void					*callbackReference[MAX_CONTROLLERS];
};
typedef struct GM_ControlCallback GM_ControlCallback;
typedef struct GM_ControlCallback * GM_ControlCallbackPtr;


// Internal Song structure
// aligned structure to 8 bytes
struct GM_Song
{
	INT16				songID;
	INT16				maxSongVoices;
	INT16				mixLevel;
	INT16				maxEffectVoices;

	FIXED_VALUE			MasterTempo;			// master midi tempo (fixed point)
	UINT16				songTempo;				// tempo (16667 = 1.0)
	INT16				songPitchShift;			// master pitch shift

	UINT16				allowPitchShift[(MAX_CHANNELS / 16) + 1];		// allow pitch shift

	long				userReference;			// not used by anything in the synth

	GM_SongCallbackProcPtr		songEndCallbackPtr;
	long						songEndCallbackReference1;
	long						songEndCallbackReference2;

	GM_SongTimeCallbackProcPtr	songTimeCallbackPtr;
	long						songTimeCallbackReference;

	GM_SongMetaCallbackProcPtr	metaEventCallbackPtr;
	long						metaEventCallbackReference1;
	long						metaEventCallbackReference2;

	// these pointers are NULL until used, then they are allocated
	GM_ControlCallbackPtr		controllerCallback;

	ReverbMode	/* 0 */	defaultReverbType;

	UBYTE		/* 1 */	velocityCurveType;			// which curve to use. (Range is 0 to 2)

	ScanMode	/* 2 */	AnalyzeMode;				// analyze mode (Byte)
	BOOL_FLAG	/* 3 */	ignoreBadInstruments;		// allow bad patches. Don't fail because it can't load

	BOOL_FLAG	/* 0 */	allowProgramChanges;
	BOOL_FLAG	/* 1 */	loopSong;					// loop song when done
	BOOL_FLAG	/* 2 */	disableClickRemoval;
	BOOL_FLAG	/* 3 */	enablePitchRandomness;
	BOOL_FLAG	/* 4 */	disposeSongDataWhenDone;	// if TRUE, then free midi data
	BOOL_FLAG	/* 5 */	SomeTrackIsAlive;			// song still alive
	BOOL_FLAG	/* 6 */	songFinished;				// TRUE at start of song, FALSE and end
	UBYTE		/* 7 */	processingSlice;			// TRUE if processing slice of this song

	BOOL_FLAG	/* 0 */	cacheSamples;				// if TRUE, then samples will be cached

	FIXED_VALUE			songFadeRate;				// when non-zero fading is enabled
	FIXED_VALUE			songFixedVolume;			// inital volume level that will be changed by songFadeRate
	INT16				songFadeMaxVolume;			// max volume
	INT16				songFadeMinVolume;			// min volume
	BOOL_FLAG			songEndAtFade;				// when true, stop song at end of fade

	INT16				songVolume;

	INT16				defaultPercusionProgram;	// default percussion program for percussion channel. 
													// -1 means GM style bank select, -2 means allow program changes on percussion

	INT16				songLoopCount;				// current loop counter. Starts at 0
	INT16				songMaxLoopCount;			// when songLoopCount reaches songMaxLoopCount it will be set to 0

	UINT32				songMidiTickLength;			// song midi tick length. -1 not calculated yet.
	UINT32				songMicrosecondLength;		// song microsecond length. -1 not calculated yet.

	void				*midiData;					// pointer to midi data for this song
	UINT32				midiSize;					// size of midi data

	//	instrument array. These are the instruments that are used by just this song
	GM_Instrument		*instrumentData[MAX_INSTRUMENTS*MAX_BANKS];

	INT32				instrumentRemap[MAX_INSTRUMENTS*MAX_BANKS];
	INT32				remapArray[MAX_INSTRUMENTS*MAX_BANKS];

	void				*pUsedPatchList;				// This is NULL most of the time, only
														// GM_LoadSongInstruments sets it
														// instruments by notes
														// This should be about 4096 bytes
														// during preprocess of the midi file, this will be
														// the instruments that need to be loaded
														// total divided by 8 bits. each bit represents an instrument

	SBYTE				firstChannelBank[MAX_CHANNELS];		// set during preprocess. this is the program
	INT16				firstChannelProgram[MAX_CHANNELS];	// to be set at the start of a song

// channel based controler values
	SBYTE				channelWhichParameter[MAX_CHANNELS];			// 0 for none, 1 for RPN, 2 for NRPN
	SBYTE				channelRegisteredParameterLSB[MAX_CHANNELS];	// Registered Parameter least signifcant byte
	SBYTE				channelRegisteredParameterMSB[MAX_CHANNELS];	// Registered Parameter most signifcant byte
	SBYTE				channelNonRegisteredParameterLSB[MAX_CHANNELS];	// Non-Registered Parameter least signifcant byte
	SBYTE				channelNonRegisteredParameterMSB[MAX_CHANNELS];	// Non-Registered Parameter most signifcant byte
	UBYTE				channelBankMode[MAX_CHANNELS];					// channel bank mode
	UBYTE				channelSustain[MAX_CHANNELS];					// sustain pedal on/off
	UBYTE				channelVolume[MAX_CHANNELS];					// current channel volume
	UBYTE				channelChorus[MAX_CHANNELS];					// current channel chorus
	UBYTE				channelExpression[MAX_CHANNELS];				// current channel expression
	UBYTE				channelPitchBendRange[MAX_CHANNELS];			// current bend range in half steps
	UBYTE				channelReverb[MAX_CHANNELS];					// current channel reverb
	UBYTE				channelModWheel[MAX_CHANNELS];					// Mod wheel (primarily affects pitch bend)
	INT16				channelBend[MAX_CHANNELS];						// MUST BE AN INT16!! current amount to bend new notes
	INT16				channelProgram[MAX_CHANNELS];					// current channel program
	SBYTE				channelBank[MAX_CHANNELS];						// current bank
	INT16				channelStereoPosition[MAX_CHANNELS];			// current channel stereo position

// mute controls for tracks, channels, and solos
// NOTE: Do not access these directly. Use XSetBit & XClearBit & XTestBit
	UINT32				trackMuted[(MAX_TRACKS / 32) + 1];			// track mute control bits 
	UINT32				soloTrackMuted[(MAX_TRACKS / 32) + 1];		// solo track mute control bits 
	UINT16				channelMuted[(MAX_CHANNELS / 16) + 1];		// current channel muted status
	UINT16				soloChannelMuted[(MAX_CHANNELS / 16) + 1];	// current channel muted status

// internal timing variables for sequencer
	UFLOAT				UnscaledMIDITempo;
	UFLOAT				MIDITempo;
	UFLOAT				MIDIDivision;
	UFLOAT				UnscaledMIDIDivision;
	UFLOAT				CurrentMidiClock;

	UFLOAT				songMicrosecondIncrement;
	UFLOAT				songMicroseconds;

	BOOL_FLAG			songPaused;

// storage for loop playback
	BOOL_FLAG			loopbackSaved;
	UBYTE				*pTrackPositionSave[MAX_TRACKS];
	IFLOAT				trackTicksSave[MAX_TRACKS];		// must be signed
	UFLOAT				currentMidiClockSave;
	UFLOAT				songMicrosecondsSave;
	SBYTE				loopbackCount;

// internal position variables for sequencer. Set after inital preprocess
	TrackStatus			trackon[MAX_TRACKS];			// track playing? TRACK_FREE is free, TRACK_RUNNING is playing
	UINT32				tracklen[MAX_TRACKS];			// length of track in bytes
	UBYTE				*ptrack[MAX_TRACKS];			// current position in track
	UBYTE				*trackstart[MAX_TRACKS];		// start of track
	UBYTE				runningStatus[MAX_TRACKS];		// midi running status
	IFLOAT				trackticks[MAX_TRACKS];			// current position of track in ticks. must be signed

	INT32				trackcumuticks[MAX_TRACKS];		// current number of beat ticks into track
	PatchInfo			*pPatchInfo;
// these things are currently only used by the PgmChange search but are generally useful
// time signature info
	UBYTE				TSNumerator;
	UBYTE				TSDenominator;
};
typedef struct GM_Song GM_Song;

#if CPU_TYPE == kRISC
	#pragma options align=reset
#endif

// $$jdr: This should be pragma pack()
// but for a bug in the compilers were using.
#if CPU_TYPE == kSPARC
        #pragma pack(4)
#endif

/* $$kk: pack(4) for solaris x86 */
#if (CPU_TYPE == k80X86)
#if (X_PLATFORM == X_SOLARIS)
	#pragma pack (4)
#else
	#pragma pack ()
#endif
#endif

// Functions


/**************************************************/
/*
** FUNCTION InitGeneralSound(Quality theQuality, 
**								TerpMode theTerp, INT16 maxVoices,
**								INT16 normVoices, INT16 maxEffects,
**								INT16 maxChunkSize)
**
** Overvue --
**	This will setup the sound system, allocate memory, and such,
**  so that any calls to play effects or songs can happen right
**  away.
**
** Private notes:
**  This code will preinitialize the MIDI sequencer, allocate
**  amplitude scaling buffer & init it, and init the
**  General Magic sound system.
**
**	INPUT	--	theQuality
**					Q_11K	Sound ouput is FIXED_VALUE at 11127
**					Q_22K	Sound output is FIXED_VALUE at 22254
**			--	theTerp
**					Interpolation type
**			--	use16bit
**					If true, then hardware will be setup for 16 bit output
**			--	maxVoices
**					maximum voices
**			--	normVoices
**					number of voices normally. ie a gain
**			--	maxEffects
**					number of voices to be used as effects
**
**	OUTPUT	--	
**
** NOTE:	
**	Only call this once.
*/
/**************************************************/
OPErr GM_InitGeneralSound(void *context, Quality theQuality, TerpMode theTerp, AudioModifiers theMods,
				INT16 maxVoices, INT16 normVoices, INT16 maxEffects);

/**************************************************/
/*
** FUNCTION FinisGeneralSound;
**
** Overvue --
**	This will release any memory allocated by InitGeneralSound, clean up.
**
**	INPUT	--	
**	OUTPUT	--	
**
** NOTE:	
**	Only call this once.
*/
/**************************************************/
void GM_FinisGeneralSound(void *context);

/* Sound hardware specific
*/
XBOOL		GM_StartHardwareSoundManager(void *context);
void		GM_StopHardwareSoundManager(void *context);

// number of devices. ie different versions of the HAE connection. DirectSound and waveOut
// return number of devices. ie 1 is one device, 2 is two devices.
// NOTE: This function needs to function before any other calls may have happened.
long GM_MaxDevices(void);

// set the current device. device is from 0 to HAE_MaxDevices()
// NOTE:	This function needs to function before any other calls may have happened.
//			Also you will need to call HAE_ReleaseAudioCard then HAE_AquireAudioCard
//			in order for the change to take place.
void GM_SetDeviceID(long deviceID);

// return current device ID
// NOTE: This function needs to function before any other calls may have happened.
long GM_GetDeviceID(void);

// get deviceID name 
// NOTE:	This function needs to function before any other calls may have happened.
//			Format of string is a zero terminated comma delinated C string.
//			"platform,method,misc"
//	example	"MacOS,Sound Manager 3.0,SndPlayDoubleBuffer"
//			"WinOS,DirectSound,multi threaded"
//			"WinOS,waveOut,multi threaded"
//			"WinOS,VxD,low level hardware"
//			"WinOS,plugin,Director"
void GM_GetDeviceName(long deviceID, char *cName, unsigned long cNameLength);

void GM_GetSystemVoices(INT16 *pMaxSongVoices, INT16 *pMixLevel, INT16 *pMaxEffectVoices);

OPErr GM_ChangeSystemVoices(INT16 maxVoices, INT16 mixLevel, INT16 maxEffects);

OPErr GM_ChangeAudioModes(void *context, Quality theQuality, TerpMode theTerp, AudioModifiers theMods);

/**************************************************/
/*
** FUNCTION PauseGeneralSound;
**
** Overvue --
**	This is used to pause the system, and release hardware for other tasks to
**	play around. Call ResumeGeneralSound when ready to resume working with
**	the system.
**
**	INPUT	--	
**	OUTPUT	--	OPErr	Errors in pausing system.
**
** NOTE:	
*/
/**************************************************/
OPErr GM_PauseGeneralSound(void *context);

// Pause all songs
void GM_PauseSequencer(void);
// resume all songs
void GM_ResumeSequencer(void);

// Pause just this song
void GM_PauseSong(GM_Song *pSong);
// Resume just this song
void GM_ResumeSong(GM_Song *pSong);

char GM_GetControllerValue(GM_Song *pSong, INT16 channel, INT16 controller);

/**************************************************/
/*
** FUNCTION ResumeGeneralSound;
**
** Overvue --
**	This is used to resume the system, and take over the sound hardware. Call this
**	after calling PauseGeneralSound.
**
**	INPUT	--	
**	OUTPUT	--	OPErr	Errors in resuming. Continue to call until errors have
**						stops, or alert user that something is still holding
**						the sound hardware.
**
** NOTE:	
*/
/**************************************************/
OPErr GM_ResumeGeneralSound(void *context);

/**************************************************/
/*
** FUNCTION BeginSong(GM_Song *theSong, GM_SongCallbackProcPtr theCallbackProc);
**
** Overvue --
**	This will start a song, given the data structure that contains the midi data,
**	instrument data, and specifics about playback of this song.
**
**	INPUT	--	theSong,			contains pointers to the song data, midi data,
**									all instruments used in this song.
**				theCallbackProc,	when the song is done, even in the looped
**									case, this procedure will be called.
**				useEmbeddedMixerSettings
**									when TRUE the mixer will be reconfigured
**									to use the embedded song settings. If FALSE
**									then the song will use the current settings
**	OUTPUT	--	
**
** NOTE:	
*/
/**************************************************/
OPErr GM_BeginSong(GM_Song *theSong, GM_SongCallbackProcPtr theCallbackProc, XBOOL useEmbeddedMixerSettings);

GM_Song * GM_LoadSong(short int songID, void *theExternalSong, void *theExternalMidiData, long midiSize, 
					short int *pInstrumentArray, 
					XBOOL loadInstruments, XBOOL ignoreBadInstruments,
					OPErr *pErr);
GM_Song * GM_CreateLiveSong(short int songID);
OPErr GM_StartLiveSong(GM_Song *pSong, BOOL_FLAG loadPatches);

// return note offset in semi tones	(12 is down an octave, -12 is up an octave)
long GM_GetSongPitchOffset(GM_Song *pSong);
// set note offset in semi tones	(12 is down an octave, -12 is up an octave)
void GM_SetSongPitchOffset(GM_Song *pSong, long offset);
// If allowPitch is FALSE, then "GM_SetSongPitchOffset" will have no effect on passed 
// channel (0 to 15)
void GM_AllowChannelPitchOffset(GM_Song *pSong, unsigned short int channel, BOOL_FLAG allowPitch);
// Return if the passed channel will allow pitch offset
BOOL_FLAG GM_DoesChannelAllowPitchOffset(GM_Song *pSong, unsigned short int channel);

/**************************************************/
/*
** FUNCTION EndSong;
**
** Overvue --
**	This will end the current song playing.
**  All song resources not be will be disposed or released.
**
**	INPUT	--	
**	OUTPUT	--	
**
** NOTE:	
*/
/*************************************************/
void GM_EndSong(GM_Song *pSong);

OPErr GM_FreeSong(GM_Song *theSong);

void GM_MergeExternalSong(void *theExternalSong, long theSongID, GM_Song *theSong);

/**************************************************/
/*
** FUNCTION SongTicks;
**
** Overvue --
**	This will return in 1/60th of a second, the count since the start of the song
**	currently playing.
**
**	INPUT	--	
**	OUTPUT	--	INT32,		returns ticks since BeginSong.
**						returns 0 if no song is playing.
**
** NOTE:	
*/
/**************************************************/
UINT32 GM_SongTicks(GM_Song *pSong);

// Return the length in MIDI ticks of the song passed
//	pSong	GM_Song structure. Data will be cloned for this function.
//	pErr		OPErr error type
UINT32 GM_GetSongTickLength(GM_Song *pSong, OPErr *pErr);
OPErr GM_SetSongTickPosition(GM_Song *pSong, UINT32 songTickPosition);

// scan through song and build data structure with all program changes
OPErr GM_GetSongInstrumentChanges(void *theSongResource, GM_Song **outSong, XBYTE **outTrackNames);

UINT32 GM_SongMicroseconds(GM_Song *pSong);
UINT32 GM_GetSongMicrosecondLength(GM_Song *pSong, OPErr *pErr);
// Set the song position in microseconds
OPErr GM_SetSongMicrosecondPosition(GM_Song *pSong, UINT32 songMicrosecondPosition);


// Get current audio time stamp based upon the audio built interrupt
UINT32 GM_GetSyncTimeStamp(void);

// Get current audio time stamp in microseconds; this is the
// microseconds' worth of samples that have passed through the
// audio device.  it never decreases.
UINT32 GM_GetDeviceTimeStamp(void);

// Update count of samples played.  This function caluculates from number of bytes,
// given the sample frame size from the mixer variables, and the bytes of data written
void GM_UpdateSamplesPlayed(unsigned long currentPos);

// Get current audio time stamp based upon the audio built interrupt, but ahead in time 
// and quantized for the particular OS
UINT32 GM_GetSyncTimeStampQuantizedAhead(void);

// Get current number of samples played; this is the
// number of samples that have passed through the
// audio device.  it never decreases.
UINT32 GM_GetSamplesPlayed(void);

// Return the used patch array of instruments used in the song passed.
//	theExternalSong	standard SONG resource structure
//	theExternalMidiData	if not NULL, then will use this midi data rather than what is found in external SONG resource
//	midiSize			size of midi data if theExternalMidiData is not NULL
//	pInstrumentArray	array, if not NULL will be filled with the instruments that need to be loaded.
//	pErr				pointer to an OPErr
INT32 GM_GetUsedPatchlist(void *theExternalSong, void *theExternalMidiData, long midiSize, 
					short int *pInstrumentArray, OPErr *pErr);

/**************************************************/
/*
** FUNCTION IsSongDone;
**
** Overvue --
**	This will return a BOOL_FLAG if a song is done playing or not.
**
**	INPUT	--	
**	OUTPUT	--	BOOL_FLAG,	returns TRUE if song is done playing,
**							or FALSE if not playing.
**
** NOTE:	
*/
/**************************************************/
BOOL_FLAG GM_IsSongDone(GM_Song *pSong);

// Mute and unmute tracks (0 to 64)
void GM_MuteTrack(GM_Song *pSong, short int track);
void GM_UnmuteTrack(GM_Song *pSong, short int track);
// Get mute status of all tracks. pStatus should be an array of 65 bytes
void GM_GetTrackMuteStatus(GM_Song *pSong, char *pStatus);

void GM_SoloTrack(GM_Song *pSong, short int track);
void GM_UnsoloTrack(GM_Song *pSong, short int track);
// will write only MAX_TRACKS bytes for MAX_TRACKS Midi tracks
void GM_GetTrackSoloStatus(GM_Song *pSong, char *pStatus);

// Mute and unmute channels (0 to 15)
void GM_MuteChannel(GM_Song *pSong, short int channel);
void GM_UnmuteChannel(GM_Song *pSong, short int channel);
// Get mute status of all channels. pStatus should be an array of 16 bytes
void GM_GetChannelMuteStatus(GM_Song *pSong, char *pStatus);

void GM_SoloChannel(GM_Song *pSong, short int channel);
void GM_UnsoloChannel(GM_Song *pSong, short int channel);
void GM_GetChannelSoloStatus(GM_Song *pSong, char *pStatus);


/**************************************************/
/*
** FUNCTION BeginSample(G_PTR theData, INT32 theSize, INT32 theRate, 
**						INT32 playTime,
**						INT32 theStartLoop, INT32 theEndLoop, 
**						BOOL_FLAG (*theLoopContinueProc)(void), 
**						void *theCallbackProc);
**
** Overvue --
**	This will play a sampled sound effect mixed into the current song being played.
**
**	INPUT	--	theData,		8 bit unsigned data
**				theSize,		length of sampled sound
**				theRate,		FIXED_VALUE value for sample playback, in 16.16 format
**				theStartLoop,	the starting sample of the loop point
**				theEndLoop,	the ending sample of the loop point
**				theLoopContinueProc,
**							if not NULL, then the sample will continualy play
**							the sample until this function returns FALSE, then
**							proceed to play the rest of the sample
**				theCallBackProc,
**							when the sample is finished being played, this
**							procedure will be called
**
**	OUTPUT	--	ref,		a sound reference number that is unique to 
**							every sample that is currently being played
**
** NOTE:	
*/
/**************************************************/

// $$kk: 01.14.98: added theLoopTarget to GM_BeginSample
INT32 GM_BeginSample(G_PTR theData, INT32 theSize, INT32 theRate, 
						INT32 theStartLoop, INT32 theEndLoop, INT32 theLoopTarget,
						INT32 sampleVolume, INT32 stereoPosition,
						INT32 refData, INT16 bitSize, INT16 channels, 
						GM_LoopDoneCallbackPtr theLoopContinueProc,
						GM_SoundDoneCallbackPtr theCallbackProc);

INT32 GM_BeginDoubleBuffer(G_PTR pBuffer1, G_PTR pBuffer2, INT32 size, INT32 rate,
							INT16 bitSize, INT16 channels,
							INT32 sampleVolume, INT16 stereoPosition,
							INT32 refData,
							GM_DoubleBufferCallbackPtr bufferCallback);

INT32 GM_BeginSampleFromInfo(GM_Waveform *pSample, INT32 refData,
								INT32 sampleVolume, INT32 stereoPosition,
								GM_LoopDoneCallbackPtr theLoopContinueProc,
								GM_SoundDoneCallbackPtr theCallbackProc,
								INT32 startOffsetFrame);

void GM_FreeWaveform(GM_Waveform *pWaveform);

void GM_SetSampleDoneCallback(INT32 reference, GM_SoundDoneCallbackPtr theCallbackProc, INT32 userRef);

/**************************************************/
/*
** FUNCTION EndSample;
**
** Overvue --
**	This will end the sample by the reference number that is passed.
**
**	INPUT	--	reference,	a reference number that was returned
**						by BeginSound
**	OUTPUT	--	
**
** NOTE:	
*/
/**************************************************/
void GM_EndSample(INT32 reference);

/**************************************************/
/*
** FUNCTION IsSoundDone;
**
** Overvue --
**	This will return status of a sound that is being played.
**
**	INPUT	--	reference,		a reference number that was returned
**							by BeginSound
**	OUTPUT	--	BOOL_FLAG	TRUE if sound is done playing
**							FALSE if sound is still playing
**
** NOTE:	
*/
/***********************************************/
BOOL_FLAG GM_IsSoundDone(INT32 reference);

/*************************************************/
/*
** FUNCTION SetMasterVolume;
**
** Overvue --
**	This will set the master output volume of the mixer.
**
**	INPUT	--	theVolume,	0-256 which is the master volume.
**	OUTPUT	--	
**
** NOTE:	
**	This is different that the hardware volume. This will scale the output by
**	theVolume factor.
**	There is somewhat of a CPU hit, while calulating the new scale buffer.
*/
/*************************************************/
void GM_SetMasterVolume(INT32 theVolume);

INT32 GM_GetMasterVolume(void);

/**************************************************/
/*
** FUNCTION ChangeSamplePitch(INT16 reference, INT32 theNewRate);
**
** Overvue --
**	This will change the current sample referenced by reference, to the new sample
**	rate of theNewRate.
**
**	INPUT	--		reference,		a reference number returned from BeginSound
**					theNewRate,	FIXED_VALUE value for sample playback, in 16.16 format
**	OUTPUT	--	
**
** NOTE:	
*/
/**************************************************/
void GM_ChangeSamplePitch(INT32 reference, FIXED_VALUE theNewRate);
FIXED_VALUE GM_GetSamplePitch(INT32 reference);

void GM_ChangeSampleVolume(INT32 reference, INT16 newVolume);
INT16 GM_GetSampleVolumeUnscaled(INT32 reference);
INT16 GM_GetSampleVolume(INT32 reference);
void GM_SetSampleFadeRate(INT32 reference, FIXED_VALUE fadeRate,
							INT16 minVolume, INT16 maxVolume, BOOL_FLAG endSample);

void GM_SetSampleLoopPoints(INT32 reference, unsigned long start, unsigned long end);

unsigned long GM_GetSamplePlaybackPosition(INT32 reference);

void GM_ChangeSampleStereoPosition(INT32 reference, INT16 newStereoPosition);
INT16 GM_GetSampleStereoPosition(INT32 reference);

void GM_ChangeSampleReverb(INT32 reference, INT16 enable);
INT16 GM_GetSampleReverb(INT32 reference);

void GM_SetSampleOffsetCallbackLinks(INT32 reference, GM_SampleCallbackEntry *pTopEntry);
void GM_AddSampleOffsetCallback(INT32 reference, GM_SampleCallbackEntry *pEntry);
void GM_RemoveSampleOffsetCallback(INT32 reference, GM_SampleCallbackEntry *pEntry);

void GM_InitReverbTaps(void);
void GM_SetReverbTaps(short int *pTapsArray, short int howMany);
void GM_GetReverbTaps(short int *pTapsArray, short int howMany);

/**************************************************/
/*
** FUNCTION GM_EndAllSoundEffects;
**
** Overvue --
**	This will end all sound effects from the system.  It
**	does not shut down sound hardware or deallocate
** 	memory used by the music system.
**
**	INPUT	--	
**	OUTPUT	--	
**
** NOTE:	
*/
/**************************************************/
void GM_EndAllSoundEffects(void);
void GM_EndAllNotes(void);
void GM_KillAllNotes(void);

void GM_EndSongNotes(GM_Song *pSong);
void GM_KillSongNotes(GM_Song *pSong);

/**************************************************/
/*
** FUNCTION SetMasterSongTempo(INT32 newTempo);
**
** Overvue --
**	This will set the master tempo for the currently playing song.
**
**	INPUT	--	newTempo is in microseconds per MIDI quater-note.
**				Another way of looking at it is, 24ths of a microsecond per
**				MIDI clock.
**	OUTPUT	--	NO_SONG_PLAYING is returned if there is no song playing
**
** NOTE:	
*/
/**************************************************/
OPErr GM_SetMasterSongTempo(GM_Song *pSong, FIXED_VALUE newTempo);
FIXED_VALUE GM_GetMasterSongTempo(GM_Song *pSong);

// Sets tempo in microsecond per quarter note
void GM_SetSongTempo(GM_Song *pSong, UINT32 newTempo);
// returns tempo in microsecond per quarter note
UINT32 GM_GetSongTempo(GM_Song *pSong);

// sets tempo in beats per minute
void GM_SetSongTempInBeatsPerMinute(GM_Song *pSong, UINT32 newTempoBPM);
// returns tempo in beats per minute
UINT32 GM_GetSongTempoInBeatsPerMinute(GM_Song *pSong);

// Instrument API

OPErr	GM_LoadInstrument(GM_Song *pSong, INT32 instrument);
OPErr	GM_LoadInstrumentFromExternal(GM_Song *pSong, INT32 instrument, void *theX, INT32 patchSize);
XBOOL	GM_AnyStereoInstrumentsLoaded(GM_Song *pSong);


//	Can return STILL_PLAYING if instrument fails to unload
OPErr	GM_UnloadInstrument(GM_Song *pSong, INT32 instrument);
OPErr	GM_RemapInstrument(GM_Song *pSong, INT32 from, INT32 to);

// Pass TRUE to cache samples, and share them. FALSE to create new copy for each sample
void GM_SetCacheSamples(GM_Song *pSong, XBOOL cacheSamples);
XBOOL GM_GetCacheSamples(GM_Song *pSong);

// returns TRUE if instrument is loaded, FALSE if otherwise
XBOOL	GM_IsInstrumentLoaded(GM_Song *pSong, INT32 instrument);

/**************************************************/
/*
** FUNCTION GM_LoadSongInstruments(GM_Song *theSong)
**
** Overvue --
**	This will load the instruments required for this song to play
**
**	INPUT	--	theSong,	a pointer to the GM_Song data containing the MIDI Data
**			--	pArray,	an array that will be filled with the instruments that are
**						loaded, if not NULL.
**			--	loadInstruments, if TRUE will load instruments and samples
**
**	OUTPUT	--	OPErr,		an error will be returned if the song
**							cannot be loaded
**
** NOTE:	
*/
/**************************************************/
OPErr	GM_LoadSongInstruments(GM_Song *theSong, short int *pArray, XBOOL loadInstruments);

// can return STILL_PLAYING if instruments are still in process. Call again to clear
OPErr	GM_UnloadSongInstruments(GM_Song *theSong);

OPErr	GM_LoadSongInstrument(GM_Song *pSong, INT32 instrument);
// can return STILL_PLAYING if instruments are still in process. Call again to clear
OPErr	GM_UnloadSongInstrument(GM_Song *pSong, INT32 instrument);

void	GM_SetSongLoopFlag(GM_Song *theSong, BOOL_FLAG loopSong);
BOOL_FLAG GM_GetSongLoopFlag(GM_Song *theSong);

short int GM_GetSongLoopMax(GM_Song *theSong);
void GM_SetSongLoopMax(GM_Song *theSong, short int maxLoopCount);

void	GM_SetSongVolume(GM_Song *theSong, INT16 newVolume);
INT16	GM_GetSongVolume(GM_Song *theSong);

void	GM_SetSongFadeRate(GM_Song *pSong, FIXED_VALUE fadeRate,
							INT16 minVolume, INT16 maxVolume, BOOL_FLAG endSong);


// range is 0 to MAX_MASTER_VOLUME (256)
void	GM_SetEffectsVolume(INT16 newVolume);
INT16	GM_GetEffectsVolume(void);

BOOL_FLAG GM_IsInstrumentRangeUsed(GM_Song *pSong, long thePatch, INT16 theLowKey, INT16 theHighKey);
BOOL_FLAG GM_IsInstrumentUsed(GM_Song *pSong, long thePatch, INT16 theKey);
void GM_SetUsedInstrument(GM_Song *pSong, long thePatch, INT16 theKey, BOOL_FLAG used);
void GM_SetInstrumentUsedRange(GM_Song *pSong, long thePatch, SBYTE *pUsedArray);
void GM_GetInstrumentUsedRange(GM_Song *pSong, long thePatch, SBYTE *pUsedArray);

void GM_SetSongCallback(GM_Song *theSong, GM_SongCallbackProcPtr songEndCallbackPtr);
void GM_SetSongTimeCallback(GM_Song *theSong, GM_SongTimeCallbackProcPtr songTimeCallbackPtr, long reference);
void GM_SetSongMetaEventCallback(GM_Song *theSong, GM_SongMetaCallbackProcPtr theCallback, long reference1, long reference2);

void GM_SetControllerCallback(GM_Song *theSong, void * reference, GM_ControlerCallbackPtr controllerCallback, short int controller);

// Display
INT16 GM_GetAudioSampleFrame(INT16 *pLeft, INT16 *pRight);

struct GM_AudioInfo
{
	INT16	maxNotesAllocated;
	INT16	maxEffectsAllocated;
	INT16	mixLevelAllocated;
	INT16	voicesActive;				// number of voices active
	INT16	patch[MAX_VOICES];			// current patches
	INT16	volume[MAX_VOICES];			// current volumes
	INT16	scaledVolume[MAX_VOICES];	// current scaled volumes
	INT16	channel[MAX_VOICES];		// current channel
	INT16	midiNote[MAX_VOICES];		// current midi note
	INT16	voice[MAX_VOICES];			// voice index
	GM_Song	*pSong[MAX_VOICES];			// song associated with voice
};
typedef struct GM_AudioInfo GM_AudioInfo;

void GM_GetRealtimeAudioInformation(GM_AudioInfo *pInfo);

// Set the global reverb type
void GM_SetReverbType(ReverbMode theReverbMode);

// Get the global reverb type
ReverbMode GM_GetReverbType(void);

// External MIDI links
#define Q_GET_TICK	0L		// if you pass this constant for timeStamp it will get the current
							// tick
void QGM_NoteOn(GM_Song *pSong, UINT32 timeStamp, INT16 channel, INT16 note, INT16 velocity);
void QGM_NoteOff(GM_Song *pSong, UINT32 timeStamp, INT16 channel, INT16 note, INT16 velocity);
void QGM_ProgramChange(GM_Song *pSong, UINT32 timeStamp, INT16 channel, INT16 program);
void QGM_PitchBend(GM_Song *pSong, UINT32 timeStamp, INT16 channel, UBYTE valueMSB, UBYTE valueLSB);
void QGM_Controller(GM_Song *pSong, UINT32 timeStamp, INT16 channel, INT16 controller, INT16 value);
void QGM_AllNotesOff(GM_Song *pSong, UINT32 timeStamp);
void QGM_LockExternalMidiQueue(void);
void QGM_UnlockExternalMidiQueue(void);

// External MIDI Links using a GM_Song. Will be unstable unless called via GM_AudioTaskCallbackPtr
void GM_NoteOn(GM_Song *pSong, INT16 channel, INT16 note, INT16 velocity);
void GM_NoteOff(GM_Song *pSong, INT16 channel, INT16 note, INT16 velocity);
void GM_ProgramChange(GM_Song *pSong, INT16 channel, INT16 program);
void GM_PitchBend(GM_Song *pSong, INT16 channel, UBYTE valueMSB, UBYTE valueLSB);
void GM_Controller(GM_Song *pSong, INT16 channel, INT16 controller, INT16 value);
void GM_AllNotesOff(GM_Song *pSong);

long GM_GetUserReference(void);
void GM_SetUserReference(long reference);

void GM_SetAudioTask(GM_AudioTaskCallbackPtr theTask);
void GM_SetAudioOutput(GM_AudioOutputCallbackPtr theHook);
GM_AudioOutputCallbackPtr GM_GetAudioOutput(void);
GM_AudioTaskCallbackPtr GM_GetAudioTask(void);
long GM_GetAudioBufferOutputSize(void);

#if USE_STREAM_API

enum
{
	FILE_AIFF_TYPE = 1,
	FILE_WAVE_TYPE,
	FILE_AU_TYPE
};
typedef long AudioFileType;


// This will read into memory the entire file and return a GM_Waveform structure.
// To dispose of a GM_Waveform structure, call GM_FreeWaveform
GM_Waveform * GM_ReadFileIntoMemory(XFILENAME *file, AudioFileType fileType);

// This will read from a block of memory an entire file and return a GM_Waveform structure.
// Assumes that the block of memory is formatted as a fileType
// To dispose of a GM_Waveform structure, call GM_FreeWaveform
GM_Waveform * GM_ReadFileIntoMemoryFromMemory(void *pFileBlock, unsigned long fileBlockSize, 
																	AudioFileType fileType);

// This will read into memory just the information about the file and return a 
// GM_Waveform structure.
// Read file information from file, which is a fileType file. If pFormat is not NULL, then
// store format specific format type
// To dispose of a GM_Waveform structure, call GM_FreeWaveform
GM_Waveform * GM_ReadFileInformation(XFILENAME *file, AudioFileType fileType, long *pFormat, long *pBlockSize);

// Read a block of data, based apon file type and format, decode and store into a buffer.
// Return length of buffer stored or 0 if error.
long GM_ReadAndDecodeFileStream(XFILE fileReference, 
										AudioFileType fileType, long format, 
										XPTR pBlockBuffer, unsigned long blockSize,
										XPTR pBuffer, unsigned long bufferLength,
										short int channels, short int bitSize,
										unsigned long *pStoredBufferLength,
										unsigned long *pReadBufferLength);

// Audio Sample Data Format. (ASDF)
// Support for 8, 16 bit data, mono and stereo. Can be extended for multi channel beyond 2 channels, but
// not required at the moment.
//
//	DATA BLOCK
//		8 bit mono data
//			ZZZZZZZ...
//				Where Z is signed 8 bit data
//
//		16 bit mono data
//			WWWWW...
//				Where W is signed 16 bit data
//
//		8 bit stereo data
//			ZXZXZXZX...
//				Where Z is signed 8 bit data for left channel, and X is signed 8 bit data for right channel.
//
//		16 bit stereo data
//			WQWQWQ...
//				Where W is signed 16 bit data for left channel, and Q is signed 16 bit data for right channel.
//



typedef enum
{
	STREAM_CREATE				=	1,
	STREAM_DESTROY,
	STREAM_GET_DATA,
	STREAM_GET_SPECIFIC_DATA
} GM_StreamMessage;

// The GM_StreamObjectProc callback is called to allocate buffer memory, get the next block of data to stream and
// mix into the final audio output, and finally dispose of the memory block. All messages will happen at 
// non-interrupt time. The structure GM_StreamData will be passed into all messages.
//
// INPUT:
// Message
//	STREAM_CREATE
//		Use this message to create a block a data with a length of (dataLength). Keep in mind that dataLength
//		is always total number of samples,  not bytes allocated. Allocate the block of data into the Audio Sample 
//		Data Format based upon (dataBitSize) and (channelSize). Store the pointer into (pData).
//
//	STREAM_DESTROY
//		Use this message to dispose of the memory allocated. (pData) will contain the pointer allocated.
//		(dataLength) will be the sample size not the buffer size. ie. for 8 bit data use (dataLength), 
//		for 16 bit mono data double (dataLength).
//
//	STREAM_GET_DATA
//		This message is called whenever the streaming object needs a new block of data. Right after STREAM_CREATE
//		is called, STREAM_GET_DATA will be called twice. Fill (pData) with the new data to be streamed.
//		Set (dataLength) to the amount of data put into (pData). This message is called in a linear fashion.
//
//	STREAM_GET_SPECIFIC_DATA
//		This message is optional. It will be called when a specific sample frame and length needs to be captured.
//		The function GM_AudioStreamGetData will call this message. If you do not want to implement this message
//		return an error of NOT_SETUP. Fill (pData) with the new sample data betweem sample frames (startSample)
//		and (endSample). Set (dataLength) to the amount of data put into (pData).
//		Note: this message will should not be called to retrive sample data for streaming. Its only used to capture
//		a range of data inside of a stream for display or other processes.
//
//
// OUTPUT:
// returns
//	NO_ERR
//		Everythings ok
//
//	STREAM_STOP_PLAY
//		Everything is fine, but stop playing stream
//
//	MEMORY_ERR
//		Couldn't allocate memory for buffers.
//
//	PARAM_ERR
//		General purpose error. Something wrong with parameters passed.
//
//	NOT_SETUP
//		If message STREAM_GET_SPECIFIC_DATA is called and it is not implemented you should return this error.
//

struct GM_StreamData
{
	long			streamReference;	// IN for all messages
	long			userReference;		// IN for all messages. userReference is passed in at AudioStreamStart
	void			*pData;				// OUT for STREAM_CREATE, IN for STREAM_DESTROY and STREAM_GET_DATA and STREAM_GET_SPECIFIC_DATA
	unsigned long	dataLength;			// OUT for STREAM_CREATE, IN for STREAM_DESTROY. IN and OUT for STREAM_GET_DATA and STREAM_GET_SPECIFIC_DATA
	FIXED_VALUE		sampleRate;			// IN for all messages. Fixed 16.16 value
	char			dataBitSize;		// IN for STREAM_CREATE only.
	char			channelSize;		// IN for STREAM_CREATE only.
	unsigned long	startSample;		// IN for STREAM_GET_SPECIFIC_DATA only.
	unsigned long	endSample;			// IN for STREAM_GET_SPECIFIC_DATA only
};
typedef struct GM_StreamData	GM_StreamData;


// CALLBACK FUNCTIONS TYPES

// AudioStream object callback
typedef OPErr (*GM_StreamObjectProc)(void *context, GM_StreamMessage message, GM_StreamData *pAS);



// Multi source user config based streaming
// This will setup a streaming audio object.
//
// INPUT:
//	userReference	This is a reference value that will be returned and should be passed along to all AudioStream
//					functions.
//
//	pProc			is a GM_StreamObjectProc proc pointer. At startup of the streaming the proc will be called
//					with STREAM_CREATE, then followed by two STREAM_GET_DATA calls to get two buffers of data,
//					and finally STREAM_DESTROY when finished.
//
// OUTPUT:
//	long			This is an audio stream reference number. Will be 0 if error
long		GM_AudioStreamSetup(	void *context,					// platform context
									long userReference, 			// user reference
									GM_StreamObjectProc pProc, 		// control callback
									unsigned long bufferSize, 		// buffer size 
									FIXED_VALUE sampleRate,			// Fixed 16.16
									char dataBitSize,				// 8 or 16 bit data
									char channelSize);				// 1 or 2 channels of date

// setup a streaming file
// OUTPUT:
//	long			This is an audio stream reference number. Will be 0 if error
long		GM_AudioStreamFileSetup(	void *context,					// platform context
										XFILENAME *file,			// file name
										AudioFileType fileType, 	// type of file
										unsigned long bufferSize,	// temp buffer to read file
										GM_Waveform *pFileInfo,
										XBOOL loopFile);			// TRUE will loop file

// return position of playback in samples. Only works for file based streaming
unsigned long GM_AudioStreamGetFileSamplePosition(long reference);

// once a stream has been setup, call this function
OPErr		GM_AudioStreamStart(long reference);

long		GM_AudioStreamGetReference(long reference);

OPErr		GM_AudioStreamGetData(	void *context,					// platform context
									long reference, 
									unsigned long startFrame, 
									unsigned long stopFrame,
									XPTR pBuffer, unsigned long bufferLength);

// This will stop a streaming audio object and free any memory.
//
// INPUT:
//	reference	This is the reference number returned from AudioStreamStart.
//
OPErr		GM_AudioStreamStop(void *context, long reference);

// This will return the last error of this stream
OPErr		GM_AudioStreamError(long reference);

// This will stop all streams that are current playing and free any memory.
void		GM_AudioStreamStopAll(void *context);

// This is the streaming audio service routine. Call this as much as possible, but not during an
// interrupt. This is a very quick routine. A good place to call this is in your main event loop.
void		GM_AudioStreamService(void *context);

// Returns TRUE or FALSE if a given AudioStream is still active
BOOL_FLAG	GM_IsAudioStreamPlaying(long reference);

// Returns TRUE if a given AudioStream is valid
BOOL_FLAG	GM_IsAudioStreamValid(long reference);

// Set the volume level of a audio stream
void		GM_AudioStreamSetVolume(long reference, short int newVolume);

// set the volume level of all open streams
void		GM_AudioStreamSetVolumeAll(short int newVolume);

// Get the volume level of a audio stream
short int	GM_AudioStreamGetVolume(long reference);

// start a stream fading
void GM_SetAudioStreamFadeRate(long reference, FIXED_VALUE fadeRate, 
							INT16 minVolume, INT16 maxVolume, BOOL_FLAG endStream);

// Set the sample rate of a audio stream
void		GM_AudioStreamSetRate(long reference, unsigned long newRate);

// Get the sample rate of a audio stream
unsigned long	GM_AudioStreamGetRate(long reference);

// Set the stereo position of a audio stream
void		GM_AudioStreamSetStereoPosition(long reference, short int stereoPosition);

// Get the stereo position of a audio stream
short int	GM_AudioStreamGetStereoPosition(long reference);

// Get the offset, in mixer-format samples, between the mixer play 
// position and the stream play position.
unsigned long GM_AudioStreamGetSampleOffset(long reference);

// Update the number of samples played from each stream, in the
// format of the stream.  delta is given in engine-format samples.
void        GM_AudioStreamUpdateSamplesPlayed(long delta);

// Get the number of samples actually played through the device
// from this stream, in stream-format samples.
unsigned long GM_AudioStreamGetSamplesPlayed(long reference);

// Drain this stream
void GM_AudioStreamDrain(long reference);

// Flush this stream
void GM_AudioStreamFlush(long reference);

// Enable/Disable reverb on this particular audio stream
void		GM_AudioStreamReverb(long reference, BOOL_FLAG useReverb);
XBOOL		GM_AudioStreamGetReverb(long reference);

// Pause or resume this particular audio stream
void		GM_AudioStreamResume(long reference);
void		GM_AudioStreamPause(long reference);

// Pause or resume all audio streams
void		GM_AudioStreamResumeAll(void);
void		GM_AudioStreamPauseAll(void);

#endif	// USE_STREAM_API

#if USE_MOD_API

typedef void		(*GM_ModDoneCallbackPtr)(struct GM_ModData *pMod);

struct GM_ModData
{
	void					*modControl;
	INT16					modVolume;

	FIXED_VALUE				modFadeRate;		// when non-zero fading is enabled
	FIXED_VALUE				modFixedVolume;		// inital volume level that will be changed by modFadeRate
	INT16					modFadeMaxVolume;	// max volume
	INT16					modFadeMinVolume;	// min volume
	XBOOL					modEndAtFade;
	XBOOL					enableReverb;

	long					reference;
	long					reference2;
	GM_ModDoneCallbackPtr	callback;
};
typedef struct GM_ModData GM_ModData;


GM_ModData	*GM_LoadModFile(void *pModFile, long fileSize);
void		GM_FreeModFile(GM_ModData *modReference);
void		GM_BeginModFile(GM_ModData *modReference, GM_ModDoneCallbackPtr callback, long reference);
void		GM_StopModFile(GM_ModData *modReference);

// volue max is MAX_SONG_VOLUME. You can overdrive
void		GM_SetModVolume(GM_ModData *modReference, INT16 volume);
short int	GM_GetModVolume(GM_ModData *modReference);
// start volume fade. For every 1.0 of fadeRate, volume will change by 11 ms
void		GM_SetModFadeRate(GM_ModData *modReference, FIXED_VALUE fadeRate, 
							INT16 minVolume, INT16 maxVolume, BOOL_FLAG endSong);

void		GM_SetModReverbStatus(GM_ModData *mod, XBOOL enableReverb);

BOOL_FLAG	GM_IsModPlaying(GM_ModData *modReference);
void		GM_ResumeMod(GM_ModData *modReference);
void		GM_PauseMod(GM_ModData *modReference);
void		GM_SetModTempoFactor(GM_ModData *modReference, unsigned long fixedFactor);
void		GM_SetModLoop(GM_ModData *modReference, BOOL_FLAG loop);
BOOL_FLAG	GM_GetModLoop(GM_ModData *mod);
void		GM_SetModTempoBPM(GM_ModData *modReference, unsigned long newTempoBPM);
unsigned long	GM_GetModTempoBPM(GM_ModData *modReference);
void			GM_GetModSongName(GM_ModData *mod, char *cName);
unsigned long	GM_GetModSongNameLength(GM_ModData *mod);
void			GM_GetModSongComments(GM_ModData *mod, char *cName);
unsigned long	GM_GetModSongCommentsLength(GM_ModData *mod);


#endif	// USE_MOD_API

#ifdef __cplusplus
	}
#endif

#endif /* G_SOUND */



