/*
 * @(#)HAE.cpp	1.3 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
** Modification History:
**
**	7/8/96		Created
**	11/1/96		Added HAEStreamSound class
**	11/15/96	Fixed StopMidiSong to end the sequencer
**				Fixed FadeMidiSong to restore the volume after the song ends
**	11/18/96	Added LoadResourceSample
**	11/19/96	Changed char to HAE_BOOL in cases of boolean use
**	11/21/96	Changed LoadResourceSample to pass in size of data
**	11/26/96	Changed ServiceStreams to ServiceIdle
**				Added FadeTo and async for all fades to the HAEMidiFile
**				and HAESoundStream, and HAESound
**	12/1/96		Changed GetCurrentProgramBank to GetProgramBank
**	12/2/96		Added Mod code
**	12/3/96		Fixed bug with HAESound::Start
**	12/10/96	Added HAEAudioMixer::ChangeAudioFileToMemory
**	12/16/96	Fixed bug in destructors that hung in a loop while trying to remove
**				objects from fade links
**	12/16/96	Fixed bug in HAEMidiFile::SetTempoInBeatsPerMinute.Used the wrong API
**	12/30/96	Changed copyright
**	1/2/97		Added GetSongNameFromAudioFile & GetSampleNameFromAudioFile &
**				GetInstrumentNameFromAudioFile
**	1/7/97		Added HAEMidiFile::LoadFromMemory
**	1/10/97		Added some const
**	1/15/97		Changed LoadFromBank and LoadBankSample to support native MacOS resources
**	1/16/97		Added GetInfoSize
**	1/22/97		Added HAEAudioMixer::GetTick
**				Fixed callbacks and added SetDoneCallback in HAESound and HAEMod
**	1/24/97		Fixed a bug with HAEAudioMixer::Close that could make my counters
**				get out of sync
**				Added HAEMod::SetLoopFlag & HAEMod::SetTempoInBeatsPerMinute &
**				HAEMod::GetTempoInBeatsPerMinute(void)
**				Rebuilt fade system. Completely intergrated with low level mixer
**	1/27/97		Changed Fade to not stop the audio object once it reaches zero
**	1/28/97		Added a HAERMFFile class. Its pretty much the same as a HAEMidiFile
**				object, except it knows how to read an RMF file
**	2/1/97		Added HAEMidiDirect::AllowChannelPitchOffset & 
**				HAEMidiDirect::DoesChannelAllowPitchOffset
**	2/5/97		Added HAESound::LoadMemorySample
**				Changed MOD handling not to keep MOD loaded into memory once parsed
**	2/8/97		Modified HAEMixer to change modifiers for stereo support
**	2/19/97		Added support for platform specific data when creating an HAEAudioMixer
**	2/28/97		Added NoteOnWithLoad & IsInstrumentLoaded & TranslateBankProgramToInstrument & 
**				TranslateCurrentProgramToInstrument
**	3/3/97		Changed NoteOnWithLoad to get the current time after loading instrument
**	3/10/97		Removed extra reverbmode get, and typed the SetReverb correctly
**	3/18/97		Fixed the pause/resume methods. They were reversed
**	3/20/97		Changed ChangeAudioModes to remap modes to C API
**				Added GetMicrosecondLength & SetMicrosecondPosition & GetMicrosecondPosition
**				in the HAEMidiFile class
**				Fixed a bug with HAEAudioMixer::SetReverbType. Forgot to set reverb type
**				for later retreival
**	4/1/97		Fixed a bug with ProgramBankChange. Forgot to include the controller change
**				for the non queued version
**				Changed GetTick & GetAudioLatency to pull the tick count from the new mixer
**				syncCount rather than XMicroseconds
**	4/18/97		Removed extra linkage and unused variables
**	4/20/97		Added HAEMidiFile::SetMetaEventCallback
**	5/1/97		Changed HAESound::Start to accecpt an frame offset when starting the sample
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**				Added HAESound::StartCustomSample
**	5/5/97		Fixed a problem with HAESound::StartCustomSample and HAESound::Start
**				in which the stereoPosition was scaled wrong. Now the value is -63 to 63.
**	5/7/97		Added to HAESound::StartCustomSample & HAESound::Start new error
**				messages for a voice that is still playing, a way to stop the sound
**				and an error when the volume level is zero
**	5/12/97		Fixed memory leak when failing with HAEMidiFile::LoadFromFile & 
**				HAEMidiFile::LoadFromMemory
**	5/13/97		Added HAEAudioMixer::GetVersionFromAudioFile
**	5/21/97		Added HAESound::GetPlaybackPosition
**				Changed method names for HAESoundStream. To stream a normal file,
**				you'll now call SetupFileStream then Start. To do a custom stream
**				call SetupCustomStream, then start.
**				Added HAESoundStream::GetInfo to return information about the
**				file once it has been setup.
**				Added HAESound::GetInfo
**	5/23/97		Added HAESound::GetSamplePointer & HAESound::GetSampleLoopPoints &
**				HAESound::SetSampleLoopPoints
**	6/4/97		Restricted 68k codebase to 11k, mono, drop sample
**	6/25/97		Changed an unsigned to an unsigned long in HAESound::SetSampleLoopPoints
**				Fixed HAESound::SetSampleLoopPoints to actually set the playing samples
**				loop points
**	6/27/97		Changed HAEMidiDirect::Open to set the song to use the current
**				mixer settings
**				Changed HAEAudioMixer::GetMixLevel & HAEAudioMixer::GetSoundVoices & 
**				HAEAudioMixer::GetMidiVoice & HAEAudioMixer::GetModifiers update to
**				the real values
**	7/9/97		Added HAEMidiDirect::ParseMidiData to parse midi data and disburse it
**				to the various functions
**	7/15/97		Added HAEAudioNoise base class to all HAEAudio objects
**				Added HAEAudioMixer::StartOutputToFile & HAEAudioMixer::StopOutputToFile & 
**				HAEAudioMixer::ServiceAudioOutputToFile
**	7/21/97		Changed HAEMod::GetTempoInBeatsPerMinute and HAEMidi::GetTempoInBeatsPerMinute to
**				to unsigned long
**	7/22/97		Changed SYNC_BUFFER_TIME to BUFFER_SLICE_TIME
**	7/28/97		Put USE_STREAM_API around the stream class
**	8/8/97		Modified PV_ProcessSongMetaEventCallbacks to support extra parameter
**	8/18/97		Seperated GM_StartHardwareSoundManager from being called inside of 
**				GM_InitGeneralSound. Now its called in HAEAudioMixer::Open. Likewise
**				GM_StopHardwareSoundManager is no longer being called within GM_FinisGeneralSound,
**				it nows is called at HAEAudioMixer::Close
**	8/20/97		Changed HAESoundStream::SetupCustomStream & HAESoundStream::SetupFileStream
**				to use HAE_MIN_STREAM_BUFFER_SIZE to control minimumn size of buffer
**	8/25/97		Changed parameter order in PV_ProcessSongControllerCallbacks
**	9/9/97		Fixed bug with PV_ProcessSongTimeCallbacks that used the wrong callback !
**	9/11/97		Added HAEMidiFile::IsPaused && HAEMod::IsDone && HAEMidiFile::IsPlaying &&
**				HAESoundStream::IsDone && HAESound::IsDone
**	9/15/97		Changed HAEMidiFile::Unload to call HAEMidiDirect::Close instead of duplicating
**				code.
**	9/30/97		Changed references to GM_AudioStreamError to pass in the stream reference
**				Changed HAEAudioMixer::GetInstrumentNameFromAudioFileFromID to allow for ID
**				being 0
**				Fixed a bug in which HAEAudioMixer::ChangeAudioFileToMemory && 
**				HAEAudioMixer::ChangeAudioFile would leave a bad patch file open if it
**				failed to open the passed file. Now it reverts back to the previously
**				open patch file if failure occurs.
**	9/30/97		Added HAEMidiFile::GetEmbeddedMidiVoices & HAEMidiFile::GetEmbeddedMixLevel & 
**				HAEMidiFile::GetEmbeddedSoundVoices
**	10/3/97		Added HAEMidiFile::SetEmbeddedMidiVoices & HAEMidiFile::SetEmbeddedMixLevel & 
**				HAEMidiFile::SetEmbeddedSoundVoices
**				Added HAEMidiFile::GetEmbeddedReverbType & HAEMidiFile::SetEmbeddedReverbType
**				Added HAEAudioMixer::GetQuality & HAEAudioMixer::GetTerpMode
**	10/12/97	Added HAERMFFile::LoadFromBank
**	10/15/97	Changed the HAERMFFile class to always copy the RMF data
**	10/16/97	Changed HAEMidiFile::Start to allow for an optional reconfigure of
**				the mixer when starting a song
**				Modified HAEMidiFile::LoadFromID & HAEMidiFile::LoadFromBank & HAEMidiFile::LoadFromFile &
**				HAEMidiFile::LoadFromMemory & HAERMFFile::LoadFromFile & HAERMFFile::LoadFromMemory &
**				HAERMFFile::LoadFromBank to allow for optional reporting of failure to load instruments
**				Removed HAEMidiDirect::FlushInstrumentCache. Not required or used.
**				Renamed GetPatches to GetInstruments and changed the array passed to
**				be of type HAE_INSTRUMENT
**	11/6/97		Added HAERMFFile::LoadFromID
**	11/10/97	Wrapped some conditional code around HAEMidiFile::GetInstruments
**	11/11/97	Cleaned up some garbage code from HAEAudioMixer::HAEAudioMixer
**				Added GetMaxDeviceCount & SetCurrentDevice & GetCurrentDevice & GetDeviceName
**	11/24/97	Added HAESoundStream::Flush
*/

#include "HAE.h"

#include "X_API.h"
#include "GenSnd.h"
#include "GenPriv.h"
#include "X_Formats.h"

#define USE_LINKED_OBJECTS			0		// if 1, then all objects will be linked

// variables
static char			audioSetup = 0;			// audio will only be setup once
static XFILE		thePatchFile = NULL;	// current audio patch library
static short int	midiSongCount = 0;		// everytime a new song is loaded, this is increments
											// this is used as an ID for song callbacks and such

// private functions

#if 0
	#pragma mark ### Support functions ###
#endif

// Read a file into memory and return an allocated pointer
static XPTR PV_GetFileAsData(XFILENAME *pName, long *pSize)
{
	XFILE	ref;
	long	size;
	XPTR	pData;

	*pSize = 0;
	pData = NULL;
	ref = XFileOpenForRead(pName);
	if (ref)
	{
		size = XFileGetLength(ref);
		pData = XNewPtr(size);
		if (pData)
		{
			XFileRead(ref, pData, size);
		}
		XFileClose(ref);
		*pSize = size;
	}
	return pData;
}

// translate reverb types from HAEReverbMode to ReverbMode
static ReverbMode PV_TranslateFromHAEReverb(HAEReverbMode igorVerb)
{
	ReverbMode	r;

	r = REVERB_TYPE_1;
	switch (igorVerb)
	{
		case HAE_REVERB_NO_CHANGE:
			r = REVERB_NO_CHANGE;
			break;
		case HAE_REVERB_TYPE_1:
			r = REVERB_TYPE_1;
			break;
		case HAE_REVERB_TYPE_2:
			r = REVERB_TYPE_2;
			break;
		case HAE_REVERB_TYPE_3:
			r = REVERB_TYPE_3;
			break;
		case HAE_REVERB_TYPE_4:
			r = REVERB_TYPE_4;
			break;
		case HAE_REVERB_TYPE_5:
			r = REVERB_TYPE_5;
			break;
		case HAE_REVERB_TYPE_6:
			r = REVERB_TYPE_6;
			break;
		case HAE_REVERB_TYPE_7:
			r = REVERB_TYPE_7;
			break;
		case HAE_REVERB_TYPE_8:
			r = REVERB_TYPE_8;
			break;
	}
	return r;
}

// translate reverb types to HAEReverbMode from ReverbMode
static HAEReverbMode PV_TranslateToHAEReverb(ReverbMode r)
{
	HAEReverbMode	igorVerb;

	igorVerb = HAE_REVERB_TYPE_1;
	switch (r)
	{
		case REVERB_NO_CHANGE:
			igorVerb = HAE_REVERB_NO_CHANGE;
			break;
		case REVERB_TYPE_1:
			igorVerb = HAE_REVERB_TYPE_1;
			break;
		case REVERB_TYPE_2:
			igorVerb = HAE_REVERB_TYPE_2;
			break;
		case REVERB_TYPE_3:
			igorVerb = HAE_REVERB_TYPE_3;
			break;
		case REVERB_TYPE_4:
			igorVerb = HAE_REVERB_TYPE_4;
			break;
		case REVERB_TYPE_5:
			igorVerb = HAE_REVERB_TYPE_5;
			break;
		case REVERB_TYPE_6:
			igorVerb = HAE_REVERB_TYPE_6;
			break;
		case REVERB_TYPE_7:
			igorVerb = HAE_REVERB_TYPE_7;
			break;
		case REVERB_TYPE_8:
			igorVerb = HAE_REVERB_TYPE_8;
			break;
	}
	return igorVerb;
}

// Translate from OPErr to HAEErr
static HAEErr PV_TranslateOPErr(OPErr theErr)
{
	HAEErr igorErr;

	switch (theErr)
	{
		default:
			igorErr = HAE_GENERAL_ERR;
			break;
		case NO_ERR:
			igorErr = HAE_NO_ERROR;
			break;
		case BUFFER_TO_SMALL:
			igorErr = HAE_BUFFER_TO_SMALL;
			break;
		case NOT_SETUP:
			igorErr = HAE_NOT_SETUP;
			break;
		case PARAM_ERR:
			igorErr = HAE_PARAM_ERR;
			break;
		case MEMORY_ERR:
			igorErr = HAE_MEMORY_ERR;
			break;
		case BAD_INSTRUMENT:
			igorErr = HAE_BAD_INSTRUMENT;
			break;
		case BAD_MIDI_DATA:
			igorErr = HAE_BAD_MIDI_DATA;
			break;
		case ALREADY_PAUSED:
			igorErr = HAE_ALREADY_PAUSED;
			break;
		case ALREADY_RESUMED:
			igorErr = HAE_ALREADY_RESUMED;
			break;
		case DEVICE_UNAVAILABLE:
			igorErr = HAE_DEVICE_UNAVAILABLE;
			break;
		case STILL_PLAYING:
			igorErr = HAE_STILL_PLAYING;
			break;
		case NO_SONG_PLAYING:
			igorErr = HAE_NO_SONG_PLAYING;
			break;
		case TOO_MANY_SONGS_PLAYING:
			igorErr = HAE_TOO_MANY_SONGS_PLAYING;
			break;
		case NO_VOLUME:
			igorErr = HAE_NO_VOLUME;
			break;
		case NO_FREE_VOICES:
			igorErr = HAE_NO_FREE_VOICES;
			break;
		case STREAM_STOP_PLAY:
			igorErr = HAE_STREAM_STOP_PLAY;
			break;
		case BAD_FILE_TYPE:
			igorErr = HAE_BAD_FILE_TYPE;
			break;
		case GENERAL_BAD:
			igorErr = HAE_GENERAL_BAD;
			break;
		case BAD_SAMPLE:
			igorErr = HAE_BAD_SAMPLE;
			break;
		case BAD_FILE:
			igorErr = HAE_BAD_FILE;
			break;
		case NOT_REENTERANT:
			igorErr = HAE_NOT_REENTERANT;
			break;
	}
	return igorErr;
}

#if X_PLATFORM == X_WIN95
/*

For Windows, you may need to pass in the
	window name of the application window. If you do, you can pass in a partial 'C' string
	name. If you pass NULL, it will search for an open window, but Open may fail. For all
	other platforms, you should pass NULL.
	
static char *pWindowName;
static HAE_BOOL CALLBACK PV_EnumWndProc( HWND hWnd, LPARAM lParam )
{
    char szWindowName[256];

    // Get the window name and class name
    //...................................
    GetWindowText( hWnd, szWindowName, 255 );    // This should be 11
	// Scan for the Netscape Window
	if (XStrStr(szWindowName, pWindowName))
	{
		HWND	*pFoundWindow	=	(HWND *)lParam;
		
		if (pFoundWindow)
		{
			// Got it
			*pFoundWindow	=	hWnd;
		}
	}
	// Haven't found it yet.
    return( TRUE);
}
		if (pPlatformData)
		{
			pWindowName = (char *)pPlatformData;
			// Enumerate all top-level windows currently running 
			// and see if it matches our 'C' name passed in.
			EnumWindows((WNDENUMPROC)PV_EnumWndProc, (LPARAM)&w95Init.hwndOwner );
			if (w95Init.hwndOwner)
	        {
				// Make Window the Active Window
				EnableWindow(w95Init.hwndOwner, TRUE);
				SetForegroundWindow(w95Init.hwndOwner);
				SetActiveWindow(w95Init.hwndOwner);
			}
			else
			{
				pPlatformData = NULL;	// didn't find a named window, so look for current
			}
		}
		if (pPlatformData == NULL)
*/

#endif

// Class implemention for HAEAudio

#if 0
	#pragma mark ### HAEAudioMixer class ###
#endif

HAEAudioMixer::HAEAudioMixer()
{
	mWritingToFile = FALSE;
	mWritingToFileReference = NULL;
	iMidiVoices = 0;
	iSoundVoices = 0;
	iMixLevel = 0;
	iQuality = HAE_22K;
	iTerpMode = HAE_LINEAR_INTERPOLATION;
	iReverbMode = HAE_REVERB_TYPE_4;
	iModifiers = HAE_USE_STEREO | HAE_USE_16;
	pTop = NULL;
	pTask = NULL;
	songNameCount = 0;
	songNameType = HAE_GROOVOID;
	sampleNameCount = 0;
	instrumentNameCount = 0;
}

HAEAudioMixer::~HAEAudioMixer()
{
	Close();
}

HAE_BOOL HAEAudioMixer::IsOpen(void) const
{
	return audioSetup ? TRUE : FALSE;
}

// return number of devices available to this mixer hardware
// will return a number from 1 to max number of devices.
// ie. a value of 2 means two devices
long HAEAudioMixer::GetMaxDeviceCount(void)
{
#if USE_DEVICE_ENUM_SUPPORT == TRUE
	return GM_MaxDevices();
#else
	return 0;
#endif
}

// set current device. should be a number from 0 to HAEAudioMixer::GetDeviceCount()
void HAEAudioMixer::SetCurrentDevice(long deviceID)
{
#if USE_DEVICE_ENUM_SUPPORT == TRUE
	if (deviceID < GetMaxDeviceCount())
	{
		if (IsOpen())
		{
			DisengageAudio();		// shutdown from hardware
		}
		GM_SetDeviceID(deviceID);	// change to new device
		if (IsOpen())
		{
			ReengageAudio();		// connect back to audio with new device
		}
	}
#endif
}

// get current device.
long HAEAudioMixer::GetCurrentDevice(void)
{
	return GM_GetDeviceID();
}

// get device name
// NOTE:	This function needs to function before any other calls may have happened.
//			Format of string is a zero terminated comma delinated C string.
//			"platform,method,misc"
//	example	"MacOS,Sound Manager 3.0,SndPlayDoubleBuffer"
//			"WinOS,DirectSound,multi threaded"
//			"WinOS,waveOut,multi threaded"
//			"WinOS,VxD,low level hardware"
//			"WinOS,plugin,Director"
void HAEAudioMixer::GetDeviceName(long deviceID, char *cName, unsigned long cNameLength)
{
#if USE_DEVICE_ENUM_SUPPORT == TRUE
	GM_GetDeviceName(deviceID, cName, cNameLength);
#endif
}

HAEErr HAEAudioMixer::Open(HAEPathName pAudioPathName,
						HAEQuality q, HAETerpMode t, 
						HAEReverbMode r, HAEAudioModifiers am,
						short int maxSongVoices,
						short int maxSoundVoices,
						short int mixLevel)
{
	OPErr			theErr;
	XFILENAME		theFile;
	Quality			theQuality;
	TerpMode		theTerp;
	AudioModifiers	theMods;
	ReverbMode		theReverb;


	theErr = NO_ERR;
	// if we've never setup the audio engine, do that now
	if (audioSetup == 0)
	{		
#if (X_PLATFORM == X_MACINTOSH) && (CPU_TYPE == k68000)
		// we're running on a MacOS 68k, so we've got to restrict the features in order to get decent playback
		q = HAE_11K;
		am &= ~HAE_USE_STEREO;			// mono only
		am &= ~HAE_STEREO_FILTER;		// don't allow
		am |= HAE_DISABLE_REVERB;		// don't allow
//		am &= ~HAE_USE_16;
		switch (q)
		{
			case HAE_44K:		// no way
				q = HAE_22K;
				break;
		}
		t = HAE_DROP_SAMPLE;

		switch (t)
		{
			case HAE_LINEAR_INTERPOLATION:
				t = HAE_2_POINT_INTERPOLATION;
				break;
		}
		// reverb doesn't matter because we're only using the shallow memory version anyways
		// and we force a disable
#endif

		switch (q)
		{
			case HAE_11K:
				theQuality = Q_11K;
				break;
			case HAE_22K:
				theQuality = Q_22K;
				break;
			case HAE_44K:
				theQuality = Q_44K;
				break;
			default:
				theErr = PARAM_ERR;
				break;
		}
		
		switch (t)
		{
			case HAE_DROP_SAMPLE:
				theTerp = E_AMP_SCALED_DROP_SAMPLE;
				break;
			case HAE_2_POINT_INTERPOLATION:
				theTerp = E_2_POINT_INTERPOLATION;
				break;
			case HAE_LINEAR_INTERPOLATION:
				theTerp = E_LINEAR_INTERPOLATION;
				break;
			default:
				theErr = PARAM_ERR;
				break;
		}

		switch (r)
		{
			case HAE_REVERB_TYPE_1:
				theReverb = REVERB_TYPE_1;
				break;
			case HAE_REVERB_TYPE_2:
				theReverb = REVERB_TYPE_2;
				break;
			case HAE_REVERB_TYPE_3:
				theReverb = REVERB_TYPE_3;
				break;
			case HAE_REVERB_TYPE_4:
				theReverb = REVERB_TYPE_4;
				break;
			case HAE_REVERB_TYPE_5:
				theReverb = REVERB_TYPE_5;
				break;
			case HAE_REVERB_TYPE_6:
				theReverb = REVERB_TYPE_6;
				break;
			case HAE_REVERB_TYPE_7:
				theReverb = REVERB_TYPE_7;
				break;
			case HAE_REVERB_TYPE_8:
				theReverb = REVERB_TYPE_8;
				break;
			default:
				theErr = PARAM_ERR;
				break;
		}

		theMods = M_NONE;
		if (am & HAE_USE_16)
		{
			theMods |= M_USE_16;
		}
		if ( (am & HAE_USE_STEREO) && XIsStereoSupported())
		{
			theMods |= M_USE_STEREO;
			if (am & HAE_STEREO_FILTER)
			{
				theMods |= M_STEREO_FILTER;
			}
		}
		if (am & HAE_DISABLE_REVERB)
		{
			theMods |= M_DISABLE_REVERB;
		}

		if (theErr == NO_ERR)
		{
			theErr = GM_InitGeneralSound(NULL, theQuality, theTerp, theMods,
											maxSongVoices,
											mixLevel,
											maxSoundVoices);
			if (theErr == NO_ERR)
			{
				audioSetup++;
				if (GM_ResumeGeneralSound(NULL) == NO_ERR)
				{
					iMidiVoices = maxSongVoices;
					iSoundVoices = maxSoundVoices;
					iMixLevel = mixLevel;

					iQuality = q;
					iTerpMode = t;
					iReverbMode = r;
					iModifiers = am;

					GM_SetReverbType(theReverb);

					if (pAudioPathName)
					{
						// everythings ok, so open the file
						XConvertNativeFileToXFILENAME(pAudioPathName, &theFile);
						thePatchFile = XFileOpenResource(&theFile, TRUE);
						if (thePatchFile)
						{
							XFileUseThisResourceFile(thePatchFile);
						}
						else
						{
							theErr = BAD_FILE;
						}
					}
					audioSetup++;
				}
				else
				{
					GM_FinisGeneralSound(NULL);
					theErr = NOT_SETUP;
				}
			}
		}
	}
	else
	{
		theErr = NOT_REENTERANT;		// can't be reentrant
	}
	return PV_TranslateOPErr(theErr);
}

void HAEAudioMixer::Close(void)
{
	if (audioSetup > 0)
	{
		StopOutputToFile();		// just in case

		if (thePatchFile)
		{
			XFileClose(thePatchFile);
		}
		thePatchFile = NULL;

		GM_SetAudioTask(NULL);

		if (audioSetup > 0)
		{
			audioSetup--;
			// Close up sound manager BEFORE releasing memory!
			GM_StopHardwareSoundManager(NULL);
		}
		if (audioSetup > 0)
		{
			audioSetup--;
			GM_FinisGeneralSound(NULL);
		}
	}
}

// Get version numbers from bank
void HAEAudioMixer::GetVersionFromAudioFile(short int *pVersionMajor, short int *pVersionMinor, short int *pVersionSubMinor)
{
	long		size;
	XVersion	*pData;

	if (pVersionMajor && pVersionMinor && pVersionSubMinor)
	{
		*pVersionMajor = 0;
		*pVersionMinor = 0;
		*pVersionSubMinor = 0;
		if (audioSetup)
		{
			pData = (XVersion *)XGetAndDetachResource(ID_VERS, 0, &size);
			if (pData)
			{
				*pVersionMajor = XGetShort(&pData->versionMajor);
				*pVersionMinor = XGetShort(&pData->versionMinor);
				*pVersionSubMinor = XGetShort(&pData->versionSubMinor);
				XDisposePtr(pData);
			}
		}
	}
}

// Get names of songs that are included in the audio file. Call successively until
// the name is zero
void HAEAudioMixer::GetSongNameFromAudioFile(char *cSongName, 
												long *pID, HAEFileType *pSongType)
{
	long		size;
	XPTR		pData;
	long		id;

	if (audioSetup && cSongName)
	{
		cSongName[0] = 0;
		pData = NULL;
		if (songNameType == HAE_GROOVOID)
		{
			pData = XGetIndexedResource(ID_SONG, &id, songNameCount, cSongName, &size);
			if (pData == NULL)
			{
				songNameType = HAE_RMF;
				songNameCount = 0;
			}
		}
		if (songNameType == HAE_RMF)
		{
			pData = XGetIndexedResource(ID_RMF, &id, songNameCount, cSongName, &size);
			if (pData == NULL)
			{
				songNameType = HAE_GROOVOID;
				songNameCount = 0;
			}
		}
		if (pData)
		{
			XPtoCstr(cSongName);
			XDisposePtr(pData);
			songNameCount++;
			if (pID)
			{
				*pID = id;
			}
			if (pSongType)
			{
				*pSongType = songNameType;
			}
		}
	}
}

// Get names of samples that are included in the audio file. Call successively until
// the name is zero
void HAEAudioMixer::GetSampleNameFromAudioFile(char *cSampleName, long *pID)
{
	long		size;
	XPTR		pData;

	if (cSampleName && pID)
	{
		cSampleName[0] = 0;
		if (audioSetup)
		{
			// look for compressed version first
			pData = XGetIndexedResource(ID_CSND, pID, sampleNameCount, cSampleName, &size);
			if (pData == NULL)
			{
				// look for standard version
				pData = XGetIndexedResource(ID_SND, pID, sampleNameCount, cSampleName, &size);

				if (pData == NULL)
				{
					// look for encrypted version
					pData = XGetIndexedResource(ID_ESND, pID, sampleNameCount, cSampleName, &size);
				}
			}
			if (pData)
			{
				XPtoCstr(cSampleName);
				XDisposePtr(pData);
				sampleNameCount++;
			}
			else
			{
				sampleNameCount = 0;
				cSampleName[0] = 0;
			}
		}
	}
}

// Get names of instruments that are included in the audio file. Call successively until
// the name is zero
void HAEAudioMixer::GetInstrumentNameFromAudioFile(char *cInstrumentName, long *pID)
{
	long		size;
	XPTR		pData;

	if (cInstrumentName && pID)
	{
		cInstrumentName[0] = 0;
		if (audioSetup)
		{
			pData = XGetIndexedResource(ID_INST, pID, instrumentNameCount, cInstrumentName, &size);
			if (pData)
			{
				XPtoCstr(cInstrumentName);
				XDisposePtr(pData);
				instrumentNameCount++;
			}
			else
			{
				instrumentNameCount = 0;
				cInstrumentName[0] = 0;
			}
		}
	}
}

// Get names of instruments that are included in the audio file, referenced by ID only. Will return
// and error if instrument not found.
HAEErr HAEAudioMixer::GetInstrumentNameFromAudioFileFromID(char *cInstrumentName, long theID)
{
	HAEErr		theErr;

	theErr = HAE_NO_ERROR;
	if (cInstrumentName)
	{
		cInstrumentName[0] = 0;
		if (audioSetup)
		{
			XGetResourceName(ID_INST, theID, cInstrumentName);
			if (cInstrumentName[0] == 0)
			{
				theErr = HAE_BAD_INSTRUMENT;
			}
		}
	}
	return theErr;
}

// change audio file
HAEErr HAEAudioMixer::ChangeAudioFile(HAEPathName pAudioPathName)
{
	OPErr			theErr;
	XFILENAME		theFile;
	XFILE			oldPatchFile;

	theErr = NO_ERR;
	if (audioSetup)
	{
		oldPatchFile = thePatchFile;
		XConvertNativeFileToXFILENAME(pAudioPathName, &theFile);
		thePatchFile = XFileOpenResource(&theFile, TRUE);
		if (thePatchFile)
		{
			XFileUseThisResourceFile(thePatchFile);
			// close old one
			if (oldPatchFile)
			{
				XFileClose(oldPatchFile);
			}
		}
		else
		{
			thePatchFile = oldPatchFile;	// restore old file
			theErr = BAD_FILE;
		}
	}
	else
	{
		theErr = MEMORY_ERR;		// can't be reentrant
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAEAudioMixer::ChangeAudioFileToMemory(void * pAudioFile, unsigned long fileSize)
{
	OPErr			theErr;
	XFILE			oldPatchFile;

	theErr = NO_ERR;
	if (audioSetup)
	{
		oldPatchFile = thePatchFile;
		thePatchFile = XFileOpenResourceFromMemory(pAudioFile, fileSize, FALSE);
		if (thePatchFile)
		{
			XFileUseThisResourceFile(thePatchFile);
			// close old one
			if (oldPatchFile)
			{
				XFileClose(oldPatchFile);
			}
		}
		else
		{
			thePatchFile = oldPatchFile;	// restore old file
			theErr = BAD_FILE;
		}
	}
	else
	{
		theErr = MEMORY_ERR;		// can't be reentrant
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAEAudioMixer::ChangeAudioModes(HAEQuality q, HAETerpMode t, HAEAudioModifiers am)
{
	OPErr			theErr;
	Quality			theQuality;
	TerpMode		theTerp;
	AudioModifiers	theMods;

	theErr = NO_ERR;
	switch (q)
	{
		case HAE_11K:
			theQuality = Q_11K;
			break;
		case HAE_22K:
			theQuality = Q_22K;
			break;
		case HAE_44K:
			theQuality = Q_44K;
			break;
		default:
			theErr = PARAM_ERR;
			break;
	}
		
	switch (t)
	{
		case HAE_DROP_SAMPLE:
			theTerp = E_AMP_SCALED_DROP_SAMPLE;
			break;
		case HAE_2_POINT_INTERPOLATION:
			theTerp = E_2_POINT_INTERPOLATION;
			break;
		case HAE_LINEAR_INTERPOLATION:
			theTerp = E_LINEAR_INTERPOLATION;
			break;
		default:
			theErr = PARAM_ERR;
			break;
	}

	theMods = M_NONE;
	if (am & HAE_USE_16)
	{
		theMods |= M_USE_16;
	}
	if ( (am & HAE_USE_STEREO) && XIsStereoSupported())
	{
		theMods |= M_USE_STEREO;
		if (am & HAE_STEREO_FILTER)
		{
			theMods |= M_STEREO_FILTER;
		}
	}
	if (am & HAE_DISABLE_REVERB)
	{
		theMods |= M_DISABLE_REVERB;
	}
	if (theErr == NO_ERR)
	{
		theErr = GM_ChangeAudioModes(NULL, theQuality, theTerp, theMods);
		if (theErr == NO_ERR)
		{
			iQuality = q;
			iTerpMode = t;
			iModifiers = am;
		}
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAEAudioMixer::ChangeSystemVoices(	short int maxSongVoices,
										short int maxSoundVoices,
										short int mixLevel)
{
	OPErr	theErr;

	theErr = GM_ChangeSystemVoices(maxSongVoices, mixLevel, maxSoundVoices);
	
	if (theErr == NO_ERR)
	{
		iMidiVoices = maxSongVoices;
		iSoundVoices = maxSoundVoices;
		iMixLevel = mixLevel;
	}
	return PV_TranslateOPErr(theErr);
}

// return time in 60ths of a second
static unsigned long PV_GetTick(void)
{
#if USE_FLOAT != FALSE
	double time;
	
	time = (double)XMicroseconds() / 16666.7;
	return (unsigned long)time;
#else
	return (XMicroseconds() * 10) / 166667;
#endif
}

// Performance testing code
unsigned long HAEAudioMixer::MeasureCPUPerformance(void)
{
	unsigned long	count, loops;
	unsigned long	postTicks;

//	Wait to top of tick count
	postTicks = PV_GetTick();
	while (postTicks == PV_GetTick()) {};

	loops = 60L * 2;	// 2 seconds
//	count as fast as possible for counts per ticks
	count = 0;
	postTicks = PV_GetTick() + loops;
	while (PV_GetTick() <= postTicks)
	{
		count++;
	}
	return count / loops;
}

// reconfigure, by changing sample rates, interpolation modes, bit depth, etc based
// upon performance of host cpu
void HAEAudioMixer::PerformanceConfigure(void)
{
	unsigned long			raw_ticks, engaged_ticks, percentage;
	long					count;
	HAEQuality				q;
	HAETerpMode				t;
	HAEAudioModifiers		am;
	// under 10 is a very fast CPU
	static unsigned long	cpuGauge[] = {10, 15, 20, 25, 30, 35, 40, 45, 50, 55};

	if (audioSetup)
	{
		DisengageAudio();		// disenguage from hardware
		raw_ticks = MeasureCPUPerformance();
		ReengageAudio();		// enguage from hardware
		engaged_ticks = MeasureCPUPerformance();

		percentage = 100L - ((engaged_ticks * 100L) / raw_ticks);

		for (count = 0; count < 10; count++)
		{
			if (percentage < cpuGauge[count])
			{
				break;
			}
		}
		switch (10-count)
		{
			default:
				count = -1;
				break;
			case 10:			// 10 %
				q = HAE_44K;
				t = HAE_LINEAR_INTERPOLATION;
				am = HAE_USE_16 | HAE_USE_STEREO | HAE_STEREO_FILTER;
				break;
			case 9:				// 15 %
				q = HAE_22K;
				t = HAE_LINEAR_INTERPOLATION;
				am = HAE_USE_16 | HAE_USE_STEREO;
				break;			
			case 8:				// 20 %
				q = HAE_22K;
				t = HAE_LINEAR_INTERPOLATION;
				am = HAE_USE_16;
				break;
			case 7:				// 25 %
				q = HAE_22K;
				t = HAE_LINEAR_INTERPOLATION;
				am = HAE_USE_16 | HAE_DISABLE_REVERB;
				break;
			case 6:				// 30 %
				q = HAE_22K;
				t = HAE_LINEAR_INTERPOLATION;
				am = HAE_USE_16 | HAE_DISABLE_REVERB;
				break;
			case 5:				// 35 %
				q = HAE_22K;
				t = HAE_2_POINT_INTERPOLATION;
				am = HAE_USE_16 | HAE_DISABLE_REVERB;
				break;
			case 4:				// 40 %
				q = HAE_22K;
				t = HAE_DROP_SAMPLE;
				am = HAE_USE_16 | HAE_DISABLE_REVERB;
				break;
			case 3:				// 45 %
				q = HAE_22K;
				t = HAE_DROP_SAMPLE;
				am = HAE_DISABLE_REVERB;
				break;
			case 2:				// 50 %
				q = HAE_22K;
				t = HAE_DROP_SAMPLE;
				am = HAE_DISABLE_REVERB;
				break;
			case 1:				// 55 %
				q = HAE_11K;
				t = HAE_DROP_SAMPLE;
				am = HAE_DISABLE_REVERB;
				break;
		}
		if (count != -1)
		{
			ChangeAudioModes(q, t, am);
		}
	}
}

HAEReverbMode HAEAudioMixer::GetReverbType(void)
{
	ReverbMode		r;
	HAEReverbMode	verb;

	r = GM_GetReverbType();
	verb = PV_TranslateToHAEReverb(r);
	SetReverbType(verb);
	return iReverbMode;
}


void HAEAudioMixer::SetReverbType(HAEReverbMode verb)
{
	ReverbMode	r;

	r = PV_TranslateFromHAEReverb(verb);
	iReverbMode = PV_TranslateToHAEReverb(r);

	GM_SetReverbType(r);
}

unsigned long HAEAudioMixer::GetTick(void)
{
	return GM_GetSyncTimeStamp();
}

unsigned long HAEAudioMixer::GetAudioLatency(void)
{
	return GM_GetSyncTimeStampQuantizedAhead() - GM_GetSyncTimeStamp();
}

void HAEAudioMixer::SetMasterVolume(HAE_FIXED theVolume)
{
	GM_SetMasterVolume((theVolume * MAX_MASTER_VOLUME) / 65536L);
}

HAE_FIXED HAEAudioMixer::GetMasterVolume(void) const
{
	HAE_FIXED	value;

	value = (GM_GetMasterVolume() * 65536L) / MAX_MASTER_VOLUME;
	return value;
}

void HAEAudioMixer::SetHardwareVolume(HAE_FIXED theVolume)
{
	XSetHardwareVolume((theVolume * X_FULL_VOLUME) / 65536L);
}

HAE_FIXED HAEAudioMixer::GetHardwareVolume(void) const
{
	HAE_FIXED	value;

	value = (XGetHardwareVolume() * 65536L) / X_FULL_VOLUME;
	return value;
}


void HAEAudioMixer::SetMasterSoundEffectsVolume(HAE_FIXED theVolume)
{
	GM_SetEffectsVolume((short)((theVolume * MAX_MASTER_VOLUME) / 65536L));
}

HAE_FIXED HAEAudioMixer::GetMasterSoundEffectsVolume(void) const
{
	HAE_FIXED	value;

	value = (GM_GetEffectsVolume() * 65536L) / MAX_MASTER_VOLUME;
	return value;
}

// display feedback information
// This will return the number of samples stored into the pLeft and pRight
// arrays. Usually 1024. This returns the current data points being sent
// to the hardware.
short int HAEAudioMixer::GetAudioSampleFrame(short int *pLeft, short int *pRight) const
{
	return GM_GetAudioSampleFrame(pLeft, pRight);
}

// Get realtime information about the current synthisizer state
void HAEAudioMixer::GetRealtimeStatus(HAEAudioInfo *pStatus) const
{
	GM_AudioInfo	status;
	short int		count;

	if (pStatus)
	{
		GM_GetRealtimeAudioInformation(&status);
		XSetMemory(pStatus, (long)sizeof(HAEAudioInfo), 0);
		pStatus->voicesActive = status.voicesActive;
		for (count = 0; count < status.voicesActive; count++)
		{
			pStatus->voice[count] = status.voice[count];

			pStatus->instrument[count] = status.patch[count];
			pStatus->scaledVolume[count] = status.scaledVolume[count];
			pStatus->midiVolume[count] = status.volume[count];
			pStatus->channel[count] = status.channel[count];
			pStatus->midiNote[count] = status.midiNote[count];
			if (status.pSong[count])
			{
				pStatus->userReference[count] = status.pSong[count]->userReference;
			}
		}
	}
}

// disengage from audio hardware
void HAEAudioMixer::DisengageAudio(void)
{
	GM_PauseGeneralSound(NULL);
}

// reengage to audio hardware
void HAEAudioMixer::ReengageAudio(void)
{
	GM_ResumeGeneralSound(NULL);
}

// Call this during idle time to service audio streams, fades, and other idle
// time processes
void HAEAudioMixer::ServiceIdle(void)
{
#if USE_LINKED_OBJECTS
	HAEAudioNoise	*pHAETop;

	// time to check fading objects
	pHAETop = (HAEAudioNoise *)pTop;
	while (pHAETop)
	{
		// do something interesting
		pHAETop = pHAETop->GetNext();
	}
#endif
#if USE_STREAM_API
	GM_AudioStreamService(NULL);
#endif
}


HAEAudioModifiers HAEAudioMixer::GetModifiers(void)
{
	return iModifiers;
}

HAETerpMode HAEAudioMixer::GetTerpMode(void)
{
	return iTerpMode;
}

HAEQuality HAEAudioMixer::GetQuality(void)
{
	return iQuality;
}

short int HAEAudioMixer::GetMidiVoices(void)
{
	INT16 song, mix, sound;

	GM_GetSystemVoices(&song, &mix, &sound);
	iMidiVoices = song;
	return iMidiVoices;
}

short int HAEAudioMixer::GetSoundVoices(void)
{
	INT16 song, mix, sound;

	GM_GetSystemVoices(&song, &mix, &sound);
	iSoundVoices = sound;
	return iSoundVoices;
}

short int HAEAudioMixer::GetMixLevel(void)
{
	INT16 song, mix, sound;

	GM_GetSystemVoices(&song, &mix, &sound);
	iMixLevel = mix;
	return iMixLevel;
}



// Set a interrupt level task callback
void HAEAudioMixer::SetTaskCallback(HAETaskCallbackPtr pCallback)
{
	pTask = pCallback;
}

// start the task callback
void HAEAudioMixer::StartTaskCallback(void)
{
	GM_SetAudioTask((GM_AudioTaskCallbackPtr)pTask);
}

// Stop the task callback
void HAEAudioMixer::StopTaskCallback(void)
{
	GM_SetAudioTask(NULL);
}

// Set a output callback. This will call your output code. This is used to modify the
// sample output before its sent to the hardware. Be very careful here. Don't use to
// much time or the audio will skip
void HAEAudioMixer::SetOutputCallback(HAEOutputCallbackPtr pCallback)
{
	pOutput = pCallback;
}

// start the output callback
void HAEAudioMixer::StartOutputCallback(void)
{
	GM_SetAudioOutput((GM_AudioOutputCallbackPtr)pOutput);
}

// Stop the output callback
void HAEAudioMixer::StopOutputCallback(void)
{
	GM_SetAudioOutput(NULL);
}

// start saving audio output to a file
HAEErr HAEAudioMixer::StartOutputToFile(HAEPathName pAudioOutputFile)
{
#if USE_CREATION_API == TRUE
	OPErr			theErr;
	XFILENAME		theFile;

	theErr = NO_ERR;
	// close old one first
	if (mWritingToFile)
	{
		StopOutputToFile();
	}
	XConvertNativeFileToXFILENAME(pAudioOutputFile, &theFile);
	mWritingToFileReference = (void *)XFileOpenForWrite(&theFile, TRUE);
	if (mWritingToFileReference)
	{
		GM_StopHardwareSoundManager(NULL);		// disengage from hardware
		mWritingToFile = TRUE;
	}
	else
	{
		theErr = BAD_FILE;
	}
	return PV_TranslateOPErr(theErr);
#else
	pAudioOutputFile = pAudioOutputFile;
	return HAE_NOT_SETUP;
#endif
}

// Stop saving audio output to a file
void HAEAudioMixer::StopOutputToFile(void)
{
	if (mWritingToFile && mWritingToFileReference)
	{
		XFileClose((XFILE)mWritingToFileReference);
		mWritingToFileReference = NULL;
		GM_StartHardwareSoundManager(NULL);			// reconnect to hardware
	}
	mWritingToFile = FALSE;
}

void HAEAudioMixer::ServiceAudioOutputToFile(void)
{
	GM_AudioTaskCallbackPtr		pTaskProc;
	GM_AudioOutputCallbackPtr	pOutputProc;
	long						outputSize;
	long						sampleSize, channels;
	char						bufferData[8192];

	if (mWritingToFile && mWritingToFileReference)
	{
		pTaskProc = GM_GetAudioTask();
		pOutputProc = GM_GetAudioOutput();

		outputSize = GM_GetAudioBufferOutputSize();
		if (outputSize)
		{
			if (outputSize < 8192)
			{
				MusicGlobals->insideAudioInterrupt++;

				MusicGlobals->syncCount += BUFFER_SLICE_TIME;			// 11 milliseconds
				MusicGlobals->syncBufferCount++;

				// Generate one frame audio
				ProcessSampleFrame(&bufferData[0]);		// process chunk
				if (pTaskProc)
				{
					(*pTaskProc)(NULL, MusicGlobals->syncCount);
				}
				if (pOutputProc)
				{
					sampleSize = (MusicGlobals->generate16output) ? 2 : 1;
					channels = (MusicGlobals->generateStereoOutput) ? 2 : 1;
					(*pOutputProc)(NULL, &bufferData[0], 
										sampleSize,
										channels,
										outputSize / channels / sampleSize);
										
				}
				XFileWrite((XFILE)mWritingToFileReference, &bufferData[0], outputSize);
				MusicGlobals->insideAudioInterrupt--;
			}
		}
	}
}


// Class implemention for HAEAudioNoise

#if 0
	#pragma mark ### HAEAudioNoise class ###
#endif

HAEAudioNoise::HAEAudioNoise(HAEAudioMixer *pHAEAudioMixer)
{
	mAudioMixer = pHAEAudioMixer;
#if USE_LINKED_OBJECTS
	{
		HAEAudioNoise	*pHAETop, *pHAENext;

		pNext = NULL;
		// add to link
		pHAETop = (HAEAudioNoise *)pHAEAudioMixer->pTop;
		if (pHAETop)
		{
			pHAENext = NULL;
			while (pHAETop)
			{
				pHAENext = pHAETop;
				pHAETop = (HAEAudioNoise *)pHAETop->pNext;
			}
			if (pHAENext)
			{
				pHAENext->pNext = this;
			}
		}
		else
		{
			mAudioMixer->pTop = this;
		}
	}
#endif
}

HAEAudioNoise::~HAEAudioNoise()
{
#if USE_LINKED_OBJECTS
	HAEAudioNoise	*pHAETop;

	// remove link
	pHAETop = (HAEAudioNoise *)mAudioMixer->pTop;
	if (pHAETop != this)
	{
		while (pHAETop)
		{
			if (pHAETop->pNext == this)
			{
				pHAETop->pNext = this->pNext;
				break;
			}

			pHAETop = (HAEMidiDirect *)pHAETop->pNext;
		}
	}
	else
	{
		mAudioMixer->pTop = NULL;
	}
#endif
}

HAEAudioNoise * HAEAudioNoise::GetNext(void)
{
	return pNext;
}

HAEAudioMixer * HAEAudioNoise::GetMixer(void)
{
	return mAudioMixer;
}

	
// Class implemention for HAEMidiFile

#if 0
	#pragma mark ### HAEMidiFile class ###
#endif

HAEMidiFile::HAEMidiFile(HAEAudioMixer *pHAEAudioMixer, 
						   char *pName, 
						   void * userReference):
									HAEMidiDirect(pHAEAudioMixer, 
													pName, userReference)
{
	m_pControllerCallbackReference = NULL;
	m_pControllerCallbackProc = NULL;
	// most of the real variables get setup when the HAEMidiDirect
	// gets created
}

HAEMidiFile::~HAEMidiFile()
{
	Unload();
}

// Given a pointer to a file, load it into a HAEMidiFile object. 
//
// If duplicateObject is TRUE, then the pointer passed in will be duplicated. 
// You can free the memory pointer passed after success.
// If FALSE the user pointer will be used, but
// not copied. Don't delete the object until after you have deleted this object.
HAEErr HAEMidiFile::LoadFromMemory(void *pMidiData, unsigned long midiSize, 
						HAE_BOOL duplicateObject, HAE_BOOL ignoreBadInstruments)
{
	SongResource		*pXSong;
	OPErr				theErr;
	short int			theID;
	GM_Song				*pSong;
	XPTR				newData;
	HAEAudioMixer		*pMixer;

	if (pSongVariables)
	{
		Unload();
	}
	theErr = NO_ERR;
	if (duplicateObject)
	{
		newData = XNewPtr(midiSize);
		if (newData)
		{
			XBlockMove(pMidiData, newData, midiSize);
			pMidiData = newData;
		}
		else
		{
			duplicateObject = FALSE;
		}
	}
	if (pMidiData && midiSize)
	{
		theID = midiSongCount++;	// runtime midi ID
		pMixer = HAEAudioNoise::GetMixer();
		pXSong = XNewSongPtr(SONG_TYPE_SMS,
							theID,
							pMixer->GetMidiVoices(),
							pMixer->GetMixLevel(),
							pMixer->GetSoundVoices(),
							PV_TranslateFromHAEReverb(pMixer->GetReverbType()));
	
		if (pXSong)
		{
			pPerformanceVariables = (void *)pXSong;	// preserve for use later

			pSong = GM_LoadSong(theID, (void *)pXSong,
										pMidiData,
										midiSize,
										NULL,		// no callback
										TRUE,		// load instruments
										ignoreBadInstruments,
										&theErr);
			if (pSong)
			{
				// things are cool
				pSong->disposeSongDataWhenDone = duplicateObject;	// dispose of midi data
				GM_SetSongLoopFlag(pSong, FALSE);					// don't loop song

				pSong->userReference = (long)reference;
				pSongVariables = pSong;				// preserve for use later
			}
			else
			{
				if (duplicateObject)
				{
					XDisposePtr(pMidiData);
				}
			}
		}
		else
		{
			theErr = MEMORY_ERR;
			if (duplicateObject)
			{
				XDisposePtr(pMidiData);
			}
		}
	}
	else
	{
		theErr = BAD_FILE;
	}
	return PV_TranslateOPErr(theErr);
}

// read a song from a file into memory
HAEErr HAEMidiFile::LoadFromFile(HAEPathName pMidiFilePath, HAE_BOOL ignoreBadInstruments)
{
	XFILENAME			name;
	XPTR				pMidiData;
	SongResource		*pXSong;
	long				midiSize;
	OPErr				theErr;
	short int			theID;
	GM_Song				*pSong;
	HAEAudioMixer		*pMixer;

	if (pSongVariables)
	{
		Unload();
	}
	theErr = NO_ERR;
	XConvertNativeFileToXFILENAME(pMidiFilePath, &name);
	pMidiData = PV_GetFileAsData(&name, &midiSize);
	if (pMidiData)
	{
		theID = midiSongCount++;	// runtime midi ID
		pMixer = HAEAudioNoise::GetMixer();
		pXSong = XNewSongPtr(SONG_TYPE_SMS,
							theID,
							pMixer->GetMidiVoices(),
							pMixer->GetMixLevel(),
							pMixer->GetSoundVoices(),
							PV_TranslateFromHAEReverb(pMixer->GetReverbType()));
	
		if (pXSong)
		{
			pPerformanceVariables = (void *)pXSong;	// preserve for use later

			pSong = GM_LoadSong(theID, (void *)pXSong,
										pMidiData,
										midiSize,
										NULL,		// no callback
										TRUE,		// load instruments
										ignoreBadInstruments,
										&theErr);
			if (pSong)
			{
				// things are cool
				pSong->disposeSongDataWhenDone = TRUE;	// dispose of midi data
				GM_SetSongLoopFlag(pSong, FALSE);		// don't loop song

				pSong->userReference = (long)reference;
				pSongVariables = pSong;				// preserve for use later
			}
			else
			{
				XDisposePtr(pMidiData);
			}
		}
		else
		{
			XDisposePtr(pMidiData);
			theErr = MEMORY_ERR;
		}
	}
	else
	{
		theErr = BAD_FILE;
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAEMidiFile::LoadFromBank(char *cName, HAE_BOOL ignoreBadInstruments)
{
	SongResource		*pXSong;
	long				size;
	OPErr				theErr;
	short int			theID;
	GM_Song				*pSong;

	if (pSongVariables)
	{
		Unload();
	}
	theErr = BAD_FILE;
#if X_PLATFORM != X_MACINTOSH
	if (thePatchFile)
#endif
	{
		pXSong = (SongResource *)XGetNamedResource(ID_SONG, cName, &size);		// look for song
		if (pXSong)
		{
			pPerformanceVariables = (void *)pXSong;	// preserve for use later

			theID = midiSongCount++;	// runtime midi ID
			pSong = GM_LoadSong(theID, (void *)pXSong,
										NULL,
										0L,
										NULL,		// no callback
										TRUE,		// load instruments
										ignoreBadInstruments,
										&theErr);
			if (pSong)
			{
				// things are cool
				pSong->disposeSongDataWhenDone = TRUE;	// dispose of midi data
				GM_SetSongLoopFlag(pSong, FALSE);		// don't loop song

				pSong->userReference = (long)reference;
				pSongVariables = pSong;				// preserve for use later
				theErr = NO_ERR;
			}
		}
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAEMidiFile::LoadFromID(unsigned long id, HAE_BOOL ignoreBadInstruments)
{
	SongResource		*pXSong;
	long				size;
	OPErr				theErr;
	GM_Song				*pSong;

	if (pSongVariables)
	{
		Unload();
	}
	theErr = BAD_FILE;
#if X_PLATFORM != X_MACINTOSH
	if (thePatchFile)
#endif
	{
		pXSong = (SongResource *)XGetAndDetachResource(ID_SONG, id, &size);		// look for song
		if (pXSong)
		{
			pPerformanceVariables = (void *)pXSong;	// preserve for use later

			pSong = GM_LoadSong(id, (void *)pXSong,
										NULL,
										0L,
										NULL,		// no callback
										TRUE,		// load instruments
										ignoreBadInstruments,
										&theErr);
			if (pSong)
			{
				// things are cool
				pSong->disposeSongDataWhenDone = TRUE;	// dispose of midi data
				GM_SetSongLoopFlag(pSong, FALSE);		// don't loop song

				pSong->userReference = (long)reference;
				pSongVariables = pSong;				// preserve for use later
				theErr = NO_ERR;
			}
		}
	}
	return PV_TranslateOPErr(theErr);
}

// start song. If useEmbeddedMixerSettings is TRUE then the mixer will be reconfigured
// to the embedded song settings. If false, then song will attempt to start with the
// current mixer configuration.
HAEErr HAEMidiFile::Start(HAE_BOOL useEmbeddedMixerSettings)
{
	OPErr theErr;

	theErr = BAD_FILE;
	if (pSongVariables)
	{
		theErr = GM_BeginSong((GM_Song *)pSongVariables, NULL, (XBOOL)useEmbeddedMixerSettings);
	}
	return PV_TranslateOPErr(theErr);
}

void HAEMidiFile::Unload(void)
{
	Stop();
	Close();
}

// end song now
void HAEMidiFile::Stop(HAE_BOOL startFade)
{
	short int	songVolume;

	if (pSongVariables)
	{
		if (startFade)
		{
			songVolume = GM_GetSongVolume((GM_Song *)pSongVariables);
			GM_SetSongFadeRate((GM_Song *)pSongVariables, FLOAT_TO_FIXED(2.2),
															0, songVolume, TRUE);
		}
		else
		{
			GM_KillSongNotes((GM_Song *)pSongVariables);
			GM_EndSong((GM_Song *)pSongVariables);
		}
	}
}

// fade song
void HAEMidiFile::Fade(HAE_BOOL doAsync)
{
	short int	songVolume;

	if (pSongVariables)
	{
		songVolume = GM_GetSongVolume((GM_Song *)pSongVariables);
		if (doAsync == FALSE)
		{
			// We're going to fade the song out and don't stop it	
			GM_SetSongFadeRate((GM_Song *)pSongVariables, FLOAT_TO_FIXED(2.2),
													0, songVolume, FALSE);
			while (GM_GetSongVolume((GM_Song *)pSongVariables) && 
					(GM_IsSongDone((GM_Song *)pSongVariables) == FALSE)) 
			{
				GetMixer()->ServiceAudioOutputToFile();
				GetMixer()->ServiceIdle();
				XWaitMicroseocnds(1000);
			}
		}
		else
		{
			GM_SetSongFadeRate((GM_Song *)pSongVariables, FLOAT_TO_FIXED(2.2),
															0, songVolume, FALSE);
		}
	}
}


// fade song
void HAEMidiFile::FadeTo(HAE_FIXED destVolume, HAE_BOOL doAsync)
{
	short int	songVolume, saveVolume, newSongVolume, saveNewSongVolume;
	FIXED_VALUE	delta;

	if (pSongVariables)
	{
		newSongVolume = (destVolume * MAX_SONG_VOLUME) / 65536L;
		saveNewSongVolume = newSongVolume;
		// We're going to fade the song out before we stop it.		
		songVolume = GM_GetSongVolume((GM_Song *)pSongVariables);
		saveVolume = songVolume;

		if (newSongVolume < songVolume)
		{	// fade out
			songVolume = newSongVolume;
			newSongVolume = saveVolume;
			delta = FLOAT_TO_FIXED(2.2);
		}
		else
		{	// fade in
			delta = FLOAT_TO_FIXED(-2.2);
		}
		if (doAsync == FALSE)
		{
			GM_SetSongFadeRate((GM_Song *)pSongVariables, delta,
													songVolume, newSongVolume, FALSE);
			while (GM_GetSongVolume((GM_Song *)pSongVariables) != saveNewSongVolume)
			{
				GetMixer()->ServiceAudioOutputToFile();
				GetMixer()->ServiceIdle();
				XWaitMicroseocnds(1000);
			}
		}
		else
		{
			GM_SetSongFadeRate((GM_Song *)pSongVariables, delta,
													songVolume, newSongVolume, FALSE);
		}
	}
}


// pause and resume song playback
void HAEMidiFile::Pause(void)
{
	if (pSongVariables)
	{
		GM_PauseSong((GM_Song *)pSongVariables);
	}
}

void HAEMidiFile::Resume(void)
{
	if (pSongVariables)
	{
		GM_ResumeSong((GM_Song *)pSongVariables);
	}
}

// currently paused
HAE_BOOL HAEMidiFile::IsPaused(void)
{
	HAE_BOOL	done;

	done = TRUE;
	if (pSongVariables)
	{
		done = ((GM_Song *)pSongVariables)->songPaused;
	}
	return done;
}



// get ticks in midi ticks of length of song
unsigned long HAEMidiFile::GetTickLength(void)
{
	OPErr			theErr;
	unsigned long	ticks;

	ticks = 0;
	if (pSongVariables)
	{
		ticks = GM_GetSongTickLength((GM_Song *)pSongVariables, &theErr);
	}
	return ticks;
}

// set the current playback position of song in midi ticks
HAEErr HAEMidiFile::SetTickPosition(unsigned long ticks)
{
	OPErr	theErr;

	theErr = NO_ERR;
	if (pSongVariables)
	{
		theErr = GM_SetSongTickPosition((GM_Song *)pSongVariables, ticks);
	}
	return PV_TranslateOPErr(theErr);
}

// get the current playback position of a song in midi ticks
unsigned long HAEMidiFile::GetTickPosition(void)
{
	if (pSongVariables)
	{
		return GM_SongTicks((GM_Song *)pSongVariables);
	}
	return 0;
}

// get ticks in microseconds of length of song
unsigned long HAEMidiFile::GetMicrosecondLength(void)
{
	OPErr			theErr;
	unsigned long	ticks;

	ticks = 0;
	if (pSongVariables)
	{
		ticks = GM_GetSongMicrosecondLength((GM_Song *)pSongVariables, &theErr);
	}
	return ticks;
}

// set the current playback position of song in microseconds
HAEErr HAEMidiFile::SetMicrosecondPosition(unsigned long ticks)
{
	OPErr	theErr;

	theErr = NO_ERR;
	if (pSongVariables)
	{
		theErr = GM_SetSongMicrosecondPosition((GM_Song *)pSongVariables, ticks);
	}
	return PV_TranslateOPErr(theErr);
}

// get the current playback position of a song in microseconds
unsigned long HAEMidiFile::GetMicrosecondPosition(void)
{
	if (pSongVariables)
	{
		return GM_SongMicroseconds((GM_Song *)pSongVariables);
	}
	return 0;
}

// return the patches required to play this song
HAEErr HAEMidiFile::GetInstruments(HAE_INSTRUMENT *pArray768, short int *pReturnedCount)
{
#if USE_CREATION_API == TRUE
	OPErr		theErr;
	void		*pMidiData;
	short int	instruments[MAX_INSTRUMENTS * MAX_BANKS];
	short int	count;

	theErr = NO_ERR;
	if (pArray768 && pReturnedCount && pSongVariables && pPerformanceVariables)
	{
		*pReturnedCount = 0;	// total number of patches loaded
		pMidiData = ((GM_Song *)pSongVariables)->midiData;
		if (pMidiData)
		{
			*pReturnedCount = (short int)GM_GetUsedPatchlist(pPerformanceVariables, 
															pMidiData, 
															XGetPtrSize(pMidiData), 
															instruments, 
															&theErr);
			for (count = 0; count < *pReturnedCount; count++)
			{
				pArray768[count] = (HAE_INSTRUMENT)instruments[count];
			}
		}
	}
	return PV_TranslateOPErr(theErr);
#else
	return HAE_NOT_SETUP;
#endif
}

static void PV_ProcessSongEndCallbacks(GM_Song *pSong)
{
	HAEDoneCallbackPtr	theCallback;

	if (pSong)
	{	
		theCallback = (HAEDoneCallbackPtr)pSong->songEndCallbackReference1;
		if (theCallback)
		{
			(*theCallback)((void *)pSong->songEndCallbackReference2);
		}
	}
}

// Set a call back when song is done
void HAEMidiFile::SetDoneCallback(HAEDoneCallbackPtr pSongCallback, void * pReference)
{
	if (pSongVariables)
	{
		((GM_Song *)pSongVariables)->songEndCallbackReference1 = (long)pSongCallback;
		((GM_Song *)pSongVariables)->songEndCallbackReference2 = (long)pReference;
		GM_SetSongCallback((GM_Song *)pSongVariables, (GM_SongCallbackProcPtr)PV_ProcessSongEndCallbacks);
	}
}

static void PV_ProcessSongTimeCallbacks(GM_Song *pSong, unsigned long currentMicroseconds, unsigned long currentMidiClock)
{
	HAETimeCallbackPtr	theCallback;

	if (pSong)
	{	
		theCallback = (HAETimeCallbackPtr)pSong->songTimeCallbackPtr;
		if (theCallback)
		{
			(*theCallback)(currentMicroseconds, currentMidiClock);
		}
	}
}

// Set a call back during song processing
void HAEMidiFile::SetTimeCallback(HAETimeCallbackPtr pSongCallback)
{
	if (pSongVariables)
	{
		GM_SetSongTimeCallback((GM_Song *)pSongVariables, 
								(GM_SongTimeCallbackProcPtr)PV_ProcessSongTimeCallbacks,
								(long)pSongCallback);
	}
}

static void PV_ProcessSongMetaEventCallbacks(GM_Song *pSong, 
									char type, void *pMetaText, long metaTextLength, INT16 currentTrack)
{
	HAEMetaEventCallbackPtr	theCallback;

	currentTrack = currentTrack;
	if (pSong)
	{
		theCallback = (HAEMetaEventCallbackPtr)pSong->metaEventCallbackReference1;
		if (theCallback)
		{
			(*theCallback)((void *)pSong->metaEventCallbackReference2, (HAEMetaTypes)type, pMetaText, metaTextLength);
		}
	}
}

// Set a call back during song playback for meta events. Pass NULL to clear callback.
void HAEMidiFile::SetMetaEventCallback(HAEMetaEventCallbackPtr pSongCallback, void * pReference)
{
	if (pSongVariables)
	{
		GM_SetSongMetaEventCallback((GM_Song *)pSongVariables, 
								PV_ProcessSongMetaEventCallbacks,
								(long)pSongCallback, (long)pReference);
	}
}

static void PV_ProcessSongControllerCallbacks(GM_Song *pSong, void * reference, short int channel, short int track, short int controler, 
														short int value)
{
	HAEMidiFile				*pHAEMidi;
	HAEControlerCallbackPtr	callback;
	void						*pReference;

	pSong = pSong;
	pHAEMidi = (HAEMidiFile *)reference;
	if (pHAEMidi)
	{
		callback = pHAEMidi->GetControlCallback();
		pReference = pHAEMidi->GetControlCallbackReference();
		if (callback)
		{
			(*callback)(pReference, channel, track, controler, value);
		}
	}
}
												
void HAEMidiFile::SetControlCallback(short int controller, HAEControlerCallbackPtr pControllerCallback, 
																							void * pReference)
{
	if (pSongVariables)
	{
		m_pControllerCallbackReference = pReference;
		m_pControllerCallbackProc = pControllerCallback;

		GM_SetControllerCallback((GM_Song *)pSongVariables, this, 
									PV_ProcessSongControllerCallbacks, controller);
	}
}


// poll to see if song is done
HAE_BOOL HAEMidiFile::IsDone(void)
{
	HAE_BOOL	done;

	done = TRUE;
	if (pSongVariables)
	{
		done = (HAE_BOOL)GM_IsSongDone((GM_Song *)pSongVariables);
	}
	return done;
}

HAE_BOOL HAEMidiFile::IsPlaying(void)
{
	return ! IsDone();
}
	

// pass TRUE to loop song, FALSE to not loop
void HAEMidiFile::SetLoopFlag(HAE_BOOL loop)
{
	if (pSongVariables)
	{
		GM_SetSongLoopFlag((GM_Song *)pSongVariables, loop);
	}
}

HAE_BOOL HAEMidiFile::GetLoopFlag(void)
{
	if (pSongVariables)
	{
		return (HAE_BOOL)GM_GetSongLoopFlag((GM_Song *)pSongVariables);
	}
	return FALSE;
}

void HAEMidiFile::SetLoopMax(short int maxLoopCount)
{
	if (pSongVariables)
	{
		GM_SetSongLoopMax((GM_Song *)pSongVariables, maxLoopCount - 1);
	}
}

short int HAEMidiFile::GetLoopMax(void)
{
	short int	loop;

	loop = 0;
	if (pSongVariables)
	{
		loop = GM_GetSongLoopMax((GM_Song *)pSongVariables);
	}
	return loop;
}

// set song master tempo
void HAEMidiFile::SetMasterTempo(HAE_FIXED newTempo)
{
	if (pSongVariables)
	{
		GM_SetMasterSongTempo((GM_Song *)pSongVariables, newTempo);
	}
}

// get song master tempo
HAE_FIXED HAEMidiFile::GetMasterTempo(void)
{
	HAE_FIXED tempo;

	tempo = 0;
	if (pSongVariables)
	{
		tempo = GM_GetMasterSongTempo((GM_Song *)pSongVariables);
	}
	return tempo;
}


// Sets tempo in microsecond per quarter note
void HAEMidiFile::SetTempo(unsigned long newTempo)
{
	if (pSongVariables)
	{
		GM_SetSongTempo((GM_Song *)pSongVariables, newTempo);
	}
}

// returns tempo in microsecond per quarter note
unsigned long HAEMidiFile::GetTempo(void)
{
	long	tempo;

	tempo = 0;
	if (pSongVariables)
	{
		tempo = GM_GetSongTempo((GM_Song *)pSongVariables);
	}
	return tempo;
}

// sets tempo in beats per minute
void HAEMidiFile::SetTempoInBeatsPerMinute(unsigned long newTempo)
{
	if (pSongVariables)
	{
		GM_SetSongTempInBeatsPerMinute((GM_Song *)pSongVariables, newTempo);
	}
}

// returns tempo in beats per minute
unsigned long HAEMidiFile::GetTempoInBeatsPerMinute(void)
{
	unsigned long	tempo;

	tempo = 0;
	if (pSongVariables)
	{
		tempo = GM_GetSongTempoInBeatsPerMinute((GM_Song *)pSongVariables);
	}
	return tempo;
}

// Get embedded data types
short int HAEMidiFile::GetEmbeddedMidiVoices(void)
{
	short int song;

	song = 0;
	if (pSongVariables)
	{
		song = ((GM_Song *)pSongVariables)->maxSongVoices;
	}
	return song;
}

short int HAEMidiFile::GetEmbeddedSoundVoices(void)
{
	short int sound;

	sound = 0;
	if (pSongVariables)
	{
		sound = ((GM_Song *)pSongVariables)->maxEffectVoices;
	}
	return sound;
}

short int HAEMidiFile::GetEmbeddedMixLevel(void)
{
	short int mix;

	mix = 0;
	if (pSongVariables)
	{
		mix = ((GM_Song *)pSongVariables)->mixLevel;
	}
	return mix;
}

HAEReverbMode HAEMidiFile::GetEmbeddedReverbType(void)
{
	ReverbMode		r;
	HAEReverbMode	verb;

	verb = HAE_REVERB_NO_CHANGE;
	if (pSongVariables)
	{
		r = ((GM_Song *)pSongVariables)->defaultReverbType;
		verb = PV_TranslateToHAEReverb(r);
	}
	return verb;
}


void HAEMidiFile::SetEmbeddedReverbType(HAEReverbMode verb)
{
	ReverbMode	r;

	if (pSongVariables)
	{
		r = PV_TranslateFromHAEReverb(verb);
		((GM_Song *)pSongVariables)->defaultReverbType = r;
	}
}

// set embedded data types
void HAEMidiFile::SetEmbeddedMidiVoices(short int midiVoices)
{
	if (pSongVariables)
	{
		((GM_Song *)pSongVariables)->maxSongVoices = midiVoices;
	}
}

void HAEMidiFile::SetEmbeddedSoundVoices(short int soundVoices)
{
	if (pSongVariables)
	{
		((GM_Song *)pSongVariables)->maxEffectVoices = soundVoices;
	}
}

void HAEMidiFile::SetEmbeddedMixLevel(short int mixLevel)
{
	if (pSongVariables)
	{
		((GM_Song *)pSongVariables)->mixLevel = mixLevel;
	}
}


void HAEMidiFile::MuteTrack(unsigned short int track)
{
	if (pSongVariables)
	{
		GM_MuteTrack((GM_Song *)pSongVariables, track);
	}
}

void HAEMidiFile::UnmuteTrack(unsigned short int track)
{
	if (pSongVariables)
	{
		GM_UnmuteTrack((GM_Song *)pSongVariables, track);
	}
}

void HAEMidiFile::GetTrackMuteStatus(HAE_BOOL *pTracks)
{
	if (pSongVariables)
	{
		GM_GetTrackMuteStatus((GM_Song *)pSongVariables, pTracks);
	}
}

void HAEMidiFile::SoloTrack(unsigned short int track)
{
	if (pSongVariables)
	{
		GM_SoloTrack((GM_Song *)pSongVariables, track);
	}
}

void HAEMidiFile::UnSoloTrack(unsigned short int track)
{
	if (pSongVariables)
	{
		GM_UnsoloTrack((GM_Song *)pSongVariables, track);
	}
}

void HAEMidiFile::GetSoloTrackStatus(HAE_BOOL *pTracks)
{
	if (pSongVariables)
	{
		GM_GetTrackSoloStatus((GM_Song *)pSongVariables, pTracks);
	}
}

static SongInfo PV_TranslateInfoTypes(HAEInfoTypes infoType)
{
	SongInfo	info;

	switch (infoType)
	{
		default:
			info = I_INVALID;
			break;
		case TITLE_INFO:
			info = I_TITLE;
			break;
		case PERFORMED_BY_INFO:
			info = I_PERFORMED_BY;
			break;
		case COMPOSER_INFO:
			info = I_COMPOSER;
			break;
		case COPYRIGHT_INFO:
			info = I_COPYRIGHT;
			break;
		case PUBLISHER_CONTACT_INFO:
			info = I_PUBLISHER_CONTACT;
			break;
		case USE_OF_LICENSE_INFO:
			info = I_USE_OF_LICENSE;
			break;
		case LICENSE_TERM_INFO:
			info = I_LICENSE_TERM;
			break;
		case LICENSED_TO_URL_INFO:
			info = I_LICENSED_TO_URL;
			break;
		case EXPIRATION_DATE_INFO:
			info = I_EXPIRATION_DATE;
			break;
		case COMPOSER_NOTES_INFO:
			info = I_COMPOSER_NOTES;
			break;
		case INDEX_NUMBER_INFO:
			info = I_INDEX_NUMBER;
			break;
	}
	return info;
}

// get size of info about this song file. Will an unsigned long
unsigned long HAEMidiFile::GetInfoSize(HAEInfoTypes infoType)
{
#if USE_FULL_RMF_SUPPORT == TRUE
	SongInfo		info;
	unsigned long	size;

	size = 0;
	info = PV_TranslateInfoTypes(infoType);
	if ((info != I_INVALID) && pSongVariables)
	{
		size = XGetSongInformationSize((SongResource *)pPerformanceVariables, XGetPtrSize((XPTR)pPerformanceVariables), 
									info);
	}
	return size;
#else
	infoType;
	return 0;
#endif
}

// get info about this song file. Will return a 'C' string
HAEErr HAEMidiFile::GetInfo(HAEInfoTypes infoType, char *cInfo)
{
#if USE_FULL_RMF_SUPPORT == TRUE
	SongInfo	info;
	HAEErr		theErr;

	theErr = HAE_NO_ERROR;
	cInfo[0] = 0;
	info = PV_TranslateInfoTypes(infoType);

	if ((info != I_INVALID) && pSongVariables)
	{
		XGetSongInformation((SongResource *)pPerformanceVariables, XGetPtrSize((XPTR)pPerformanceVariables), 
									info, cInfo);

#if ( (X_PLATFORM == X_WIN95) 			||	\
	  (X_PLATFORM == X_WIN_HARDWARE)	||	\
	  (X_PLATFORM == X_SOLARIS)	)
		while (*cInfo)
		{
			*cInfo = XTranslateMacToWin(*cInfo);
			cInfo++;
		}
#endif
	}
	else
	{
		theErr = HAE_PARAM_ERR;
	}
	return theErr;
#else
	infoType;
	cInfo[0] = 0;
	return HAE_NOT_SETUP;
#endif
}

// Class implemention for HAERMFFile

#if 0
	#pragma mark ### HAERMFFile class ###
#endif

HAERMFFile::HAERMFFile(HAEAudioMixer *pHAEAudioMixer, 
						   char *pName, 
						   void * userReference):
									HAEMidiFile(pHAEAudioMixer, 
													pName, userReference)
{
	// most of the real variables get setup when the HAEMidiDirect and HAEMidiFile
	// gets created
	pRMFDataBlock = NULL;
}

HAERMFFile::~HAERMFFile()
{
	Unload();
	XDisposePtr((XPTR)pRMFDataBlock);
	pRMFDataBlock = NULL;
}

HAEErr HAERMFFile::LoadFromFile(HAEPathName pRMFFilePath, HAE_BOOL ignoreBadInstruments)
{
	XFILE				fileRef;
	XFILENAME			name;
	SongResource		*pXSong;
	GM_Song				*pSong;
	OPErr				theErr;
	long				theID;
	long				size;

	theErr = BAD_FILE;

	if (pSongVariables)
	{
		Unload();
	}
	XConvertNativeFileToXFILENAME(pRMFFilePath, &name);
	fileRef = XFileOpenResource(&name, TRUE);
	if (fileRef)
	{
		// look for first song. RMF files only contain one SONG resource
		pXSong = (SongResource *)XGetIndexedResource(ID_SONG, &theID, 0, NULL, &size);
		if (pXSong)
		{
			pPerformanceVariables = (void *)pXSong;	// preserve for use later

			pSong = GM_LoadSong(theID, (void *)pXSong,
										NULL,
										0L,
										NULL,		// no callback
										TRUE,		// load instruments
										ignoreBadInstruments,
										&theErr);
			if (pSong)
			{
				// things are cool
				pSong->disposeSongDataWhenDone = TRUE;	// dispose of midi data
				GM_SetSongLoopFlag(pSong, FALSE);		// don't loop song

				pSong->userReference = (long)reference;
				pSongVariables = pSong;				// preserve for use later
				theErr = NO_ERR;
			}
		}
		XFileClose(fileRef);
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAERMFFile::LoadFromMemory(void *pRMFData, unsigned long rmfSize, 
									HAE_BOOL duplicateObject,
									HAE_BOOL ignoreBadInstruments)
{
	XFILE				fileRef;
	SongResource		*pXSong;
	GM_Song				*pSong;
	OPErr				theErr;
	long				theID;
	long				size;
	void				*newData;

	theErr = BAD_FILE;

	if (pSongVariables)
	{
		Unload();
	}
	if (pRMFData && rmfSize)
	{
		if (duplicateObject)
		{
			newData = XNewPtr(rmfSize);
			if (newData)
			{
				XBlockMove(pRMFData, newData, rmfSize);
				pRMFDataBlock = newData;
				pRMFData = newData;
			}
			else
			{
				duplicateObject = FALSE;
			}
		}
//		fileRef = XFileOpenResourceFromMemory(pRMFData, rmfSize, ! duplicateObject);

		// right now, always copy
		fileRef = XFileOpenResourceFromMemory(pRMFData, rmfSize, TRUE);
		if (fileRef)
		{
			// look for first song. RMF files only contain one SONG resource
			pXSong = (SongResource *)XGetIndexedResource(ID_SONG, &theID, 0, NULL, &size);
			if (pXSong)
			{
				pPerformanceVariables = (void *)pXSong;	// preserve for use later

				pSong = GM_LoadSong(theID, (void *)pXSong,
											NULL,
											0L,
											NULL,		// no callback
											TRUE,		// load instruments
											ignoreBadInstruments,
											&theErr);
				if (pSong)
				{
					// things are cool
					pSong->disposeSongDataWhenDone = TRUE;	// dispose of midi data
					GM_SetSongLoopFlag(pSong, FALSE);		// don't loop song

					pSong->userReference = (long)reference;
					pSongVariables = pSong;				// preserve for use later
					theErr = NO_ERR;
				}
			}
			XFileClose(fileRef);
		}
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAERMFFile::LoadFromBank(char *cName, HAE_BOOL ignoreBadInstruments)
{
	XPTR		pRMF;
	long		size;
	HAEErr		theErr;

	theErr = HAE_BAD_FILE;

	if (pSongVariables)
	{
		Unload();
	}
	pRMF = XGetNamedResource(ID_RMF, cName, &size);		// look for embedded RMF song
	if (pRMF)
	{
		theErr = LoadFromMemory(pRMF, size, TRUE, ignoreBadInstruments);
		XDisposePtr(pRMF);
	}
	return theErr;
}

HAEErr HAERMFFile::LoadFromID(unsigned long id, HAE_BOOL ignoreBadInstruments)
{
	XPTR		pRMF;
	long		size;
	HAEErr		theErr;

	theErr = HAE_BAD_FILE;

	if (pSongVariables)
	{
		Unload();
	}
	pRMF = XGetAndDetachResource(ID_RMF, id, &size);		// look for embedded RMF song
	if (pRMF)
	{
		theErr = LoadFromMemory(pRMF, size, TRUE, ignoreBadInstruments);
		XDisposePtr(pRMF);
	}
	return theErr;
}

// Class implemention for HAEMidiDirect

#if 0
	#pragma mark ### HAEMidiDirect class ###
#endif

HAEMidiDirect::HAEMidiDirect(HAEAudioMixer *pHAEAudioMixer, char *pName, 
							   void * userReference) :
									HAEAudioNoise(pHAEAudioMixer)
{
	pPerformanceVariables = NULL;
	queueMidi = TRUE;
	reference = userReference;
	pSongVariables = NULL;
	name[0] = 0;
	if (pName)
	{
		XStrCpy(name, pName);
	}
}

HAEMidiDirect::~HAEMidiDirect()
{
	Close();
}

HAEErr HAEMidiDirect::Open(HAE_BOOL loadInstruments)
{
	GM_Song			*pSong;
	OPErr			theErr;
	HAEAudioMixer	*pMixer;

	theErr = NO_ERR;
	if (pSongVariables)
	{
		Close();
	}
	pSong = GM_CreateLiveSong(midiSongCount++);
	if (pSong)
	{
		pMixer = GetMixer();
		pSong->maxSongVoices = pMixer->GetMidiVoices();
		pSong->maxEffectVoices = pMixer->GetSoundVoices();
		pSong->mixLevel = pMixer->GetMixLevel();

		pSong->userReference = (long)reference;
		theErr = GM_StartLiveSong(pSong, loadInstruments);
		if (theErr)
		{
			while (GM_FreeSong(pSong) == STILL_PLAYING)
			{
				XWaitMicroseocnds(BUFFER_SLICE_TIME);
			}
			pSong = NULL;
		}
	}
	pSongVariables = (void *)pSong;
	return PV_TranslateOPErr(theErr);
}

void HAEMidiDirect::Close(void)
{
	if (pSongVariables)
	{
		GM_KillSongNotes((GM_Song *)pSongVariables);
		((GM_Song *)pSongVariables)->disposeSongDataWhenDone = TRUE;	// free our midi pointer
		while (GM_FreeSong((GM_Song *)pSongVariables) == STILL_PLAYING)
		{
			XWaitMicroseocnds(BUFFER_SLICE_TIME);
		}
		pSongVariables = NULL;
	}
	if (pPerformanceVariables)
	{
		XDisposeSongPtr((SongResource *)pPerformanceVariables);
		pPerformanceVariables = NULL;
	}
}

// return controller value
char HAEMidiDirect::GetControlValue(unsigned char channel,  unsigned char controller)
{
	char	value;

	value = 0;
	if (pSongVariables)
	{
		value = GM_GetControllerValue((GM_Song *)pSongVariables, channel, controller);
	}
	return value;
}

void HAEMidiDirect::SetCacheSample(HAE_BOOL cacheSamples)
{
	if (pSongVariables)
	{
		GM_SetCacheSamples((GM_Song *)pSongVariables, cacheSamples);
	}
}
		
HAE_BOOL HAEMidiDirect::GetCacheSample(void)
{
	HAE_BOOL	value;

	value = 0;
	if (pSongVariables)
	{
		value = GM_GetCacheSamples((GM_Song *)pSongVariables);
	}
	return value;
}
		
// Get the current Midi program and bank values
void HAEMidiDirect::GetProgramBank(unsigned char channel, 
										   unsigned char *pProgram,
										   unsigned char *pBank)
{
	if (pSongVariables && pBank && pProgram)
	{
		*pProgram = ((GM_Song *)pSongVariables)->channelProgram[channel];
		*pBank = ((GM_Song *)pSongVariables)->channelBank[channel];
	}
}

void HAEMidiDirect::MuteChannel(unsigned short int channel)
{
	if (pSongVariables)
	{
		GM_MuteChannel((GM_Song *)pSongVariables, channel);
	}
}

void HAEMidiDirect::UnmuteChannel(unsigned short int channel)
{
	if (pSongVariables)
	{
		GM_UnmuteChannel((GM_Song *)pSongVariables, channel);
	}
}

void HAEMidiDirect::GetChannelMuteStatus(HAE_BOOL *pChannels)
{
	if (pSongVariables)
	{
		GM_GetChannelMuteStatus((GM_Song *)pSongVariables, pChannels);
	}
}

void HAEMidiDirect::SoloChannel(unsigned short int channel)
{
	if (pSongVariables)
	{
		GM_SoloChannel((GM_Song *)pSongVariables, channel);
	}
}

void HAEMidiDirect::UnSoloChannel(unsigned short int channel)
{
	if (pSongVariables)
	{
		GM_UnsoloChannel((GM_Song *)pSongVariables, channel);
	}
}

void HAEMidiDirect::GetChannelSoloStatus(HAE_BOOL *pChannels)
{
	if (pSongVariables)
	{
		GM_GetChannelSoloStatus((GM_Song *)pSongVariables, pChannels);
	}
}


// set song volume. You can overdrive by passing values larger than 1.0
void HAEMidiDirect::SetVolume(HAE_FIXED volume)
{
	if (pSongVariables)
	{
		GM_SetSongVolume((GM_Song *)pSongVariables, (short)((volume * MAX_SONG_VOLUME) / 65536L));
	}
}

// get the song volume
HAE_FIXED HAEMidiDirect::GetVolume(void)
{
	HAE_FIXED	volume;

	volume = 0;
	if (pSongVariables)
	{
		volume = (GM_GetSongVolume((GM_Song *)pSongVariables) * 65536L) / MAX_SONG_VOLUME;
	}
	return volume;
}

// If allowPitch is FALSE, then "SetPitchOffset" will have no effect on passed 
// channel (0 to 15)
void HAEMidiDirect::AllowChannelPitchOffset(unsigned short int channel, HAE_BOOL allowPitch)
{
	if (pSongVariables)
	{
		GM_AllowChannelPitchOffset((GM_Song *)pSongVariables, channel, allowPitch);
	}
}

// Return if the passed channel will allow pitch offset
HAE_BOOL HAEMidiDirect::DoesChannelAllowPitchOffset(unsigned short int channel)
{
	HAE_BOOL	flag;

	flag = FALSE;
	if (pSongVariables)
	{
		flag = GM_DoesChannelAllowPitchOffset((GM_Song *)pSongVariables, channel);
	}
	return flag;
}

// set note offset in semi tones	(-12 is down an octave, 12 is up an octave)
void HAEMidiDirect::SetPitchOffset(long offset)
{
	if (pSongVariables)
	{
		GM_SetSongPitchOffset((GM_Song *)pSongVariables, offset);
	}
}

// return note offset in semi tones	(-12 is down an octave, 12 is up an octave)
long HAEMidiDirect::GetPitchOffset(void)
{
	long	offset;

	offset = 0;
	if (pSongVariables)
	{
		offset = GM_GetSongPitchOffset((GM_Song *)pSongVariables);
	}
	return offset;
}

HAE_BOOL HAEMidiDirect::IsInstrumentLoaded(HAE_INSTRUMENT instrument)
{
	HAE_BOOL	loaded;

	loaded = FALSE;
	if (pSongVariables && (instrument >= 0) && (instrument < (MAX_INSTRUMENTS*MAX_BANKS)))
	{
//		if (MusicGlobals->InstrumentData[MusicGlobals->remapArray[instrument]])
		if (((GM_Song *)pSongVariables)->instrumentData[instrument])
		{
			loaded = TRUE;
		}
	}
	return loaded;
}

HAE_INSTRUMENT HAEMidiDirect::TranslateBankProgramToInstrument(unsigned short bank, 
											unsigned short program, unsigned short channel, unsigned short note)
{
	unsigned long	instrument;

	instrument = program;
	if (channel == PERCUSSION_CHANNEL)
	{
		bank = (bank * 2) + 1;		// odd banks are percussion
	}
	else
	{
		bank = bank * 2 + 0;		// even banks are for instruments
		note = 0;
	}

	if (bank < MAX_BANKS)
	{
		instrument = (bank * 128) + program + note;
	}

	return instrument;
}


HAEErr HAEMidiDirect::LoadInstrument(HAE_INSTRUMENT instrument)
{
	OPErr	theErr;

	theErr = NO_ERR;
	if (pSongVariables)
	{
		theErr = GM_LoadSongInstrument((GM_Song *)pSongVariables, instrument);
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAEMidiDirect::UnloadInstrument(HAE_INSTRUMENT instrument)
{
	OPErr	theErr;

	theErr = NO_ERR;
	if (pSongVariables)
	{
		theErr = GM_UnloadSongInstrument((GM_Song *)pSongVariables, instrument);
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAEMidiDirect::RemapInstrument(HAE_INSTRUMENT from, HAE_INSTRUMENT to)
{
	OPErr	theErr;

	theErr = GM_RemapInstrument((GM_Song *)pSongVariables, from, to);
	return PV_TranslateOPErr(theErr);
}


// get current midi tick in microseconds
unsigned long HAEMidiDirect::GetTick(void)
{
	return GM_GetSyncTimeStamp();
}

// set queue control of midi commmands. Use TRUE to queue commands, FALSE to
// send directly to engine. Default is TRUE
void HAEMidiDirect::SetQueue(HAE_BOOL useQueue)
{
	queueMidi = useQueue;
}

// given a midi stream, parse it out to the various midi functions
// for example:
// 0x92			0x50		0x7F		0x00
// comandByte	data1Byte	data2Byte	data3Byte
// Note 80 on with a velocity of 127 on channel 2
void HAEMidiDirect::ParseMidiData(unsigned char commandByte, unsigned char data1Byte, 
									unsigned char data2Byte, unsigned char data3Byte,
									unsigned long time)
{
	unsigned char	channel;

	channel = commandByte & 0x0F;
	data3Byte = data3Byte;
	switch(commandByte & 0xF0)
	{
		case 0x80:	// Note off
			NoteOff(channel, data1Byte, data2Byte, time);
			break;
		case 0x90:	// Note on
			NoteOn(channel, data1Byte, data2Byte, time);
			break;
		case 0xA0:	// key pressure (aftertouch)
			KeyPressure(channel, data1Byte, data2Byte, time);
			break;
		case 0xB0:	// controllers
			ControlChange(channel, data1Byte, data2Byte, time);
			break;
		case 0xC0:	// Program change
			ProgramChange(channel, data1Byte, time);
			break;
		case 0xD0:	// channel pressure (aftertouch)
			ChannelPressure(channel, data1Byte, time);
			break;
		case 0xE0:	// SetPitchBend
			PitchBend(channel, data1Byte, data2Byte, time);
			break;
	}

}

// normal note on
void HAEMidiDirect::NoteOn(unsigned char channel, 
					   unsigned char note, 
					   unsigned char velocity,
					   unsigned long time)
{
	if (time == 0)
	{
		time = GM_GetSyncTimeStamp();
	}

	if (queueMidi)
	{
		QGM_LockExternalMidiQueue();
		QGM_NoteOn((GM_Song *)pSongVariables, time, channel, note, velocity);
		QGM_UnlockExternalMidiQueue();
	}
	else
	{
		GM_NoteOn((GM_Song *)pSongVariables, channel, note, velocity);
	}
}

// normal note on with load
void HAEMidiDirect::NoteOnWithLoad(unsigned char channel, 
					   unsigned char note, 
					   unsigned char velocity,
					   unsigned long time)
{
	HAE_INSTRUMENT	inst;
	unsigned char	program, bank;

	if (queueMidi)
	{
		GetProgramBank(channel, &program, &bank);
		inst = TranslateBankProgramToInstrument(bank, program, channel, note);
		if (IsInstrumentLoaded(inst) == FALSE)
		{
			LoadInstrument(inst);
		}
		if (time == 0)
		{
			time = GM_GetSyncTimeStamp();
		}

		QGM_LockExternalMidiQueue();
		QGM_NoteOn((GM_Song *)pSongVariables, time, channel, note, velocity);
		QGM_UnlockExternalMidiQueue();
	}
}

void HAEMidiDirect::NoteOff(unsigned char channel, 
						unsigned char note, 
						unsigned char velocity,
						unsigned long time)
{
	if (time == 0)
	{
		time = GM_GetSyncTimeStamp();
	}

	if (queueMidi)
	{
		QGM_LockExternalMidiQueue();
		QGM_NoteOff((GM_Song *)pSongVariables, time, channel, note, velocity);
		QGM_UnlockExternalMidiQueue();
	}
	else
	{
		GM_NoteOff((GM_Song *)pSongVariables, channel, note, velocity);
	}
}


void HAEMidiDirect::ControlChange(unsigned char channel, 
							  unsigned char controlNumber,
							  unsigned char controlValue, 
							  unsigned long time)
{
	if (time == 0)
	{
		time = GM_GetSyncTimeStamp();
	}

	if (queueMidi)
	{
		QGM_LockExternalMidiQueue();
		QGM_Controller((GM_Song *)pSongVariables, time, channel, controlNumber, controlValue);
		QGM_UnlockExternalMidiQueue();
	}
	else
	{
		GM_Controller((GM_Song *)pSongVariables, channel, controlNumber, controlValue);
	}
}



void HAEMidiDirect::ProgramBankChange(unsigned char channel, 
							  unsigned char programNumber,
							  unsigned char bankNumber,
							  unsigned long time)
{
	if (time == 0)
	{
		time = GM_GetSyncTimeStamp();
	}

	if (queueMidi)
	{
		QGM_LockExternalMidiQueue();
		QGM_Controller((GM_Song *)pSongVariables, time, channel, 0, bankNumber);
		QGM_ProgramChange((GM_Song *)pSongVariables, time, channel, programNumber);
		QGM_UnlockExternalMidiQueue();
	}
	else
	{
		GM_Controller((GM_Song *)pSongVariables, channel, 0, bankNumber);
		GM_ProgramChange((GM_Song *)pSongVariables, channel, programNumber);
	}
}

void HAEMidiDirect::ProgramChange(unsigned char channel, 
							  unsigned char programNumber,
							  unsigned long time)
{
	if (time == 0)
	{
		time = GM_GetSyncTimeStamp();
	}

	if (queueMidi)
	{
		QGM_LockExternalMidiQueue();
		QGM_ProgramChange((GM_Song *)pSongVariables, time, channel, programNumber);
		QGM_UnlockExternalMidiQueue();
	}
	else
	{
		GM_ProgramChange((GM_Song *)pSongVariables, channel, programNumber);
	}
}

void HAEMidiDirect::ChannelPressure(unsigned char channel, unsigned char pressure, unsigned long time)
{
	time = time;
	channel = channel;
	pressure = pressure;
}

void HAEMidiDirect::KeyPressure(unsigned char channel, unsigned char note, unsigned char pressure, unsigned long time)
{
	time = time;
	pressure = pressure;
	channel = channel;
	note = note;
}

void HAEMidiDirect::PitchBend(unsigned char channel, 
						  unsigned char lsb, 
						  unsigned char msb,
						  unsigned long time)
{
	if (time == 0)
	{
		time = GM_GetSyncTimeStamp();
	}

	if (queueMidi)
	{
		QGM_LockExternalMidiQueue();
		QGM_PitchBend((GM_Song *)pSongVariables, time, channel, msb, lsb);
		QGM_UnlockExternalMidiQueue();
	}
	else
	{
		GM_PitchBend((GM_Song *)pSongVariables, channel, msb, lsb);
	}
}


void HAEMidiDirect::AllNotesOff(unsigned long time)
{
	if (time == 0)
	{
		time = GM_GetSyncTimeStamp();
	}

	if (queueMidi)
	{
		QGM_LockExternalMidiQueue();
		QGM_AllNotesOff((GM_Song *)pSongVariables, time);
		QGM_UnlockExternalMidiQueue();
	}
	else
	{
		GM_AllNotesOff((GM_Song *)pSongVariables);
	}
}


#if 0
	#pragma mark ### HAESoundStream class ###
#endif

#if USE_STREAM_API

HAESoundStream::HAESoundStream(HAEAudioMixer *pHAEAudio,
								 char *pName, void * userReference) :
									HAEAudioNoise(pHAEAudio)
{
	mReference = 0;
	mUserReference = userReference;
	mPauseVariable = 0;
	mCallbackProc = NULL;
	pName = pName;
}


HAESoundStream::~HAESoundStream()
{
	Stop();
}

// streaming file callback. Used to decode typed files.
static OPErr PV_CustomStreamCallback(void *context, GM_StreamMessage message, GM_StreamData *pAS)
{
	HAESoundStream			*pHAEStream;
	HAEStreamObjectProc	callback;
	HAEStreamData			iData;
	OPErr					theErr;
	HAEErr					igorErr;
	HAEStreamMessage		igorMessage;

	context = context;
	theErr = NO_ERR;
	igorErr = HAE_NO_ERROR;
	pHAEStream = (HAESoundStream *)pAS->userReference;
	if (pHAEStream)
	{
		iData.userReference = (long)pHAEStream->GetReference();
		iData.pData = pAS->pData;
		iData.dataLength = pAS->dataLength;
		iData.sampleRate = pAS->sampleRate;
		iData.dataBitSize = pAS->dataBitSize;
		iData.channelSize = pAS->channelSize;
		iData.startSample = pAS->startSample;
		iData.endSample = pAS->endSample;
		callback = (HAEStreamObjectProc)pHAEStream->GetCallbackProc();
		if (callback)
		{
			switch (message)
			{
				default:
					igorMessage = HAE_STREAM_NULL;
					break;
				case STREAM_CREATE:
					igorMessage = HAE_STREAM_CREATE;
					break;
				case STREAM_DESTROY:
					igorMessage = HAE_STREAM_DESTROY;
					break;
				case STREAM_GET_DATA:
					igorMessage = HAE_STREAM_GET_DATA;
					break;
				case STREAM_GET_SPECIFIC_DATA:
					igorMessage = HAE_STREAM_GET_SPECIFIC_DATA;
					break;
			}
			if (igorMessage != HAE_STREAM_NULL)
			{
				igorErr = (*callback)(igorMessage, &iData);
			}
			else
			{
				igorErr = HAE_NOT_SETUP;
			}
			pAS->pData = iData.pData;
			pAS->dataLength = iData.dataLength;
			pAS->sampleRate = iData.sampleRate;
			pAS->dataBitSize = iData.dataBitSize;
			pAS->channelSize = iData.channelSize;
			pAS->startSample = iData.startSample;
			pAS->endSample = iData.endSample;
		}
	}
	switch (igorErr)
	{
		default:
			theErr = GENERAL_BAD;
			break;
		case HAE_NO_ERROR:
			theErr = NO_ERR;
			break;
		case HAE_BUFFER_TO_SMALL:
			theErr = BUFFER_TO_SMALL;
			break;
		case HAE_NOT_SETUP:
			theErr = NOT_SETUP;
			break;
		case HAE_PARAM_ERR:
			theErr = PARAM_ERR;
			break;
		case HAE_MEMORY_ERR:
			theErr = MEMORY_ERR;
			break;
		case HAE_STREAM_STOP_PLAY:
			theErr = STREAM_STOP_PLAY;
			break;
	}
	return theErr;
}

// start streaming a file
HAEErr HAESoundStream::SetupFileStream(HAEPathName pWaveFilePath, 
								HAEFileType fileType,
								unsigned long bufferSize,			// temp buffer to read file
								HAE_BOOL loopFile)		// TRUE will loop file
{
#if USE_HIGHLEVEL_FILE_API
	XFILENAME	theFile;
	GM_Waveform	fileInfo;

	XConvertNativeFileToXFILENAME(pWaveFilePath, &theFile);

	if (bufferSize >= HAE_MIN_STREAM_BUFFER_SIZE)
	{
		mReference = GM_AudioStreamFileSetup(NULL, &theFile, fileType, bufferSize, &fileInfo, loopFile);
		mStreamSampleInfo.bitSize = fileInfo.bitSize;
		mStreamSampleInfo.channels = fileInfo.channels;
		mStreamSampleInfo.sampledRate = fileInfo.sampledRate;
		mStreamSampleInfo.baseMidiPitch = fileInfo.baseMidiPitch;
		mStreamSampleInfo.waveSize = fileInfo.waveSize;
		mStreamSampleInfo.waveFrames = fileInfo.waveFrames;
		mStreamSampleInfo.startLoop = 0;
		mStreamSampleInfo.endLoop = 0;
		return PV_TranslateOPErr(GM_AudioStreamError(mReference));
	}
	return HAE_BUFFER_TO_SMALL;
#else
	loopFile;
	pWaveFilePath;
	fileType;
	bufferSize;
	return HAE_NOT_SETUP;
#endif
}

HAEErr HAESoundStream::SetupCustomStream(HAEStreamObjectProc pProc, 	// control callback
									unsigned long bufferSize, 			// buffer size 
									HAE_FIXED sampleRate,				// Fixed 16.16
									char dataBitSize,					// 8 or 16 bit data
									char channelSize)					// 1 or 2 channels of date
{
	mCallbackProc = pProc;
	if (bufferSize >= HAE_MIN_STREAM_BUFFER_SIZE)
	{
		mReference = GM_AudioStreamSetup(NULL, (long)this, PV_CustomStreamCallback, bufferSize,
										sampleRate, dataBitSize, channelSize);					

		mStreamSampleInfo.bitSize = dataBitSize;
		mStreamSampleInfo.channels = channelSize;
		mStreamSampleInfo.baseMidiPitch = 60;
		mStreamSampleInfo.waveSize = 0;
		mStreamSampleInfo.waveFrames = 0;
		mStreamSampleInfo.startLoop = 0;
		mStreamSampleInfo.endLoop = 0;
		mStreamSampleInfo.sampledRate = sampleRate;
		return PV_TranslateOPErr(GM_AudioStreamError(mReference));
	}
	return HAE_BUFFER_TO_SMALL;
}

// Get the position of a audio stream in samples
unsigned long HAESoundStream::GetPlaybackPosition(void)
{
	return GM_AudioStreamGetFileSamplePosition(mReference);
}

HAEErr HAESoundStream::GetInfo(HAESampleInfo *pInfo)
{
	if (pInfo)
	{
		*pInfo = mStreamSampleInfo;
	}
	return HAE_NO_ERROR;
}

// This will return the last AudioStream Error
HAEErr HAESoundStream::LastError(void)
{
	return PV_TranslateOPErr(GM_AudioStreamError(mReference));
}

	// This will start a stream once data has been loaded
HAEErr HAESoundStream::Start(HAE_FIXED sampleVolume, 				// sample volume	(1.0)
									short int stereoPosition,			// stereo placement -63 to 63
									unsigned long startOffsetFrame)		// starting offset in frames
{
	OPErr	theErr;

	theErr = NO_ERR;
	if (mReference)
	{
		startOffsetFrame = startOffsetFrame;
		GM_AudioStreamSetVolume(mReference, (short)((sampleVolume * MAX_NOTE_VOLUME) / 65536L));
		GM_AudioStreamSetStereoPosition(mReference, stereoPosition);
		GM_AudioStreamStart(mReference);
	}
	else
	{
		theErr = NOT_SETUP;
	}
	return PV_TranslateOPErr(theErr);
}

// This will stop a streaming audio object and free any memory.
void HAESoundStream::Stop(HAE_BOOL startFade)
{
	short int	streamVolume;

	if (startFade)
	{
		streamVolume = GM_AudioStreamGetVolume(mReference);
		GM_SetAudioStreamFadeRate(mReference, FLOAT_TO_FIXED(2.2), 0, streamVolume, TRUE);
	}
	else
	{
		GM_AudioStreamStop(mReference);
	}
}

// This will stop and flush the current stream and force a read of data. This
// will cause gaps in the audio.
void HAESoundStream::Flush(void)
{
	GM_AudioStreamFlush(mReference);
}

// Returns TRUE or FALSE if a given AudioStream is still active
HAE_BOOL HAESoundStream::IsPlaying(void)
{
	return (HAE_BOOL)GM_IsAudioStreamPlaying(mReference);
}

HAE_BOOL HAESoundStream::IsDone(void)
{
	return ! IsPlaying();
}

// Returns TRUE if a given AudioStream is valid
HAE_BOOL HAESoundStream::IsValid(void)
{
	return (HAE_BOOL)GM_IsAudioStreamValid(mReference);
}

// Set the volume level of a audio stream
void HAESoundStream::SetVolume(HAE_FIXED newVolume)
{
	GM_AudioStreamSetVolume(mReference, (short)((newVolume * MAX_NOTE_VOLUME) / 65536L));
}

// Get the volume level of a audio stream
HAE_FIXED HAESoundStream::GetVolume(void)
{
	return (GM_AudioStreamGetVolume(mReference) * 65536L) / MAX_NOTE_VOLUME;
}

void HAESoundStream::Fade(HAE_BOOL doAsync)
{
	short int	sampleVolume;

	sampleVolume = GM_AudioStreamGetVolume(mReference);
	if (doAsync == FALSE)
	{
		// We're going to fade the stream, but don't stop

		GM_SetAudioStreamFadeRate(mReference, FLOAT_TO_FIXED(2.2),
												0, sampleVolume, FALSE);
		while (GM_AudioStreamGetVolume(mReference) && GM_IsAudioStreamPlaying(mReference)) 
		{
			GetMixer()->ServiceAudioOutputToFile();
			GetMixer()->ServiceIdle();
			XWaitMicroseocnds(1000);
		}
	}
	else
	{
		GM_SetAudioStreamFadeRate(mReference, FLOAT_TO_FIXED(2.2), 0, sampleVolume, FALSE);
	}
}


void HAESoundStream::FadeTo(HAE_FIXED destVolume, HAE_BOOL doAsync)
{
	short int	songVolume, saveVolume, newSongVolume, saveNewSongVolume;
	FIXED_VALUE	delta;

	if (mReference)
	{
		newSongVolume = (destVolume * MAX_SONG_VOLUME) / 65536L;
		saveNewSongVolume = newSongVolume;
		// We're going to fade the song out before we stop it.		
		songVolume = GM_AudioStreamGetVolume(mReference);
		saveVolume = songVolume;

		if (newSongVolume < songVolume)
		{	// fade out
			songVolume = newSongVolume;
			newSongVolume = saveVolume;
			delta = FLOAT_TO_FIXED(2.2);
		}
		else
		{	// fade in
			delta = FLOAT_TO_FIXED(-2.2);
		}
		if (doAsync == FALSE)
		{
			GM_SetAudioStreamFadeRate(mReference, delta, songVolume, newSongVolume, FALSE);
			while (GM_AudioStreamGetVolume(mReference) != saveNewSongVolume)
			{
				GetMixer()->ServiceAudioOutputToFile();
				GetMixer()->ServiceIdle();
				XWaitMicroseocnds(1000);
			}
		}
		else
		{
			GM_SetAudioStreamFadeRate(mReference, delta, songVolume, newSongVolume, FALSE);
		}
	}
}

// Set the sample rate of a audio stream
void HAESoundStream::SetRate(HAE_FIXED newRate)
{
	GM_AudioStreamSetRate(mReference, newRate);
}

// Get the sample rate of a audio stream
HAE_FIXED HAESoundStream::GetRate(void)
{
	return GM_AudioStreamGetRate(mReference);
}

// Set the stereo position of a audio stream
void HAESoundStream::SetStereoPosition(short int stereoPosition)
{
	GM_AudioStreamSetStereoPosition(mReference, stereoPosition);
}

// Get the stereo position of a audio stream
short int HAESoundStream::GetStereoPosition(void)
{
	return GM_AudioStreamGetStereoPosition(mReference);
}

// Enable/Disable reverb on this particular audio stream
void HAESoundStream::SetReverb(HAE_BOOL useReverb)
{
	GM_AudioStreamReverb(mReference, useReverb);
}

HAE_BOOL HAESoundStream::GetReverb(void)
{
	return GM_AudioStreamGetReverb(mReference);
}

// currently paused
HAE_BOOL HAESoundStream::IsPaused(void)
{
	return (mPauseVariable) ? TRUE : FALSE;
}

// Pause the stream
void HAESoundStream::Pause(void)
{
	if (mPauseVariable == 0)
	{
		mPauseVariable = (unsigned long)GetRate();
		SetRate(0L);	// pause samples in their tracks
	}
}

// Resume the stream
void HAESoundStream::Resume(void)
{
	if (mPauseVariable)
	{
		SetRate((HAE_FIXED)mPauseVariable);
		mPauseVariable = 0;
	}
}
#endif	// USE_STREAM_API

#if 0
	#pragma mark ### HAESound class ###
#endif

HAESound::HAESound(HAEAudioMixer *pHAEAudio, 
					 char *pName, void * userReference) :
									HAEAudioNoise(pHAEAudio)
{
	userReference = userReference;
	reference = 0;
	pauseVariable = 0;
	pFileVariables = NULL;
	pSoundVariables = NULL;
	pSampleFrameVariable = NULL;
	pName = pName;
}


HAESound::~HAESound()
{
	GM_SampleCallbackEntry	*pNext, *pLast;

	Stop();

	// remove callback links
	pNext = (GM_SampleCallbackEntry *)pSampleFrameVariable;
	while (pNext)
	{
		pLast = pNext;
		pNext = pNext->pNext;
		XDisposePtr((XPTR)pLast);
	}

	if (pFileVariables)
	{
		GM_FreeWaveform((GM_Waveform *)pFileVariables);
		pFileVariables = NULL;
	}
	if (pSoundVariables)
	{
		XDisposePtr(pSoundVariables);
		pSoundVariables = NULL;
	}
}

// currently paused
HAE_BOOL HAESound::IsPaused(void)
{
	return (pauseVariable) ? TRUE : FALSE;
}

void HAESound::Resume(void)
{
	if (pauseVariable)
	{
		SetRate((HAE_FIXED)pauseVariable);
		pauseVariable = 0;
	}
}

void HAESound::Pause(void)
{
	if (pauseVariable == 0)
	{
		pauseVariable = (unsigned long)GetRate();
		SetRate(0L);	// pause samples in their tracks
	}
}

// load a sample playing from a formatted block of memory. The memory will be deallocated 
// when this object is destroyed. Call start once loaded to start the playback.
HAEErr HAESound::LoadMemorySample(void *pMemoryFile, unsigned long memoryFileSize, HAEFileType fileType)
{
#if USE_HIGHLEVEL_FILE_API
	OPErr			theErr;

	theErr = NO_ERR;
	pFileVariables = NULL;
	switch (fileType)
	{
		case HAE_AIFF_TYPE:
			pFileVariables = GM_ReadFileIntoMemoryFromMemory(pMemoryFile, memoryFileSize, FILE_AIFF_TYPE);
			break;
		case HAE_WAVE_TYPE:
			pFileVariables = GM_ReadFileIntoMemoryFromMemory(pMemoryFile, memoryFileSize, FILE_WAVE_TYPE);
			break;
		case HAE_AU_TYPE:
			pFileVariables = GM_ReadFileIntoMemoryFromMemory(pMemoryFile, memoryFileSize, FILE_AU_TYPE);
			break;
		default:
			theErr = BAD_FILE;
			break;
	}
	if (pFileVariables == NULL)
	{
		theErr = BAD_FILE;
	}
	return PV_TranslateOPErr(theErr);
#else
	pMemoryFile = pMemoryFile;
	memoryFileSize = memoryFileSize;
	fileType = fileType;
	return HAE_NOT_SETUP;
#endif
}

// Load file into sound object. This will copy the file directly into memory. It
// will get disposed once you destroy this object.
HAEErr HAESound::LoadFileSample(HAEPathName pWaveFilePath, HAEFileType fileType)
{
#if USE_HIGHLEVEL_FILE_API
	XFILENAME		theFile;
	OPErr			theErr;

	theErr = NO_ERR;
	XConvertNativeFileToXFILENAME(pWaveFilePath, &theFile);
	pFileVariables = NULL;
	switch (fileType)
	{
		case HAE_AIFF_TYPE:
			pFileVariables = GM_ReadFileIntoMemory(&theFile, FILE_AIFF_TYPE);
			break;
		case HAE_WAVE_TYPE:
			pFileVariables = GM_ReadFileIntoMemory(&theFile, FILE_WAVE_TYPE);
			break;
		case HAE_AU_TYPE:
			pFileVariables = GM_ReadFileIntoMemory(&theFile, FILE_AU_TYPE);
			break;
		default:
			theErr = BAD_FILE;
			break;
	}
	if (pFileVariables == NULL)
	{
		theErr = BAD_FILE;
	}
	return PV_TranslateOPErr(theErr);
#else
	pWaveFilePath = pWaveFilePath;
	fileType = fileType;
	return HAE_NOT_SETUP;
#endif
}


HAEErr HAESound::AddSampleFrameCallback(unsigned long frame, HAESampleFrameCallbackPtr pCallback, void * pReference)
{
	GM_SampleCallbackEntry	*pEntry;
	GM_SampleCallbackEntry	*pNext;
	OPErr					theErr;

	theErr = MEMORY_ERR;
	pEntry = (GM_SampleCallbackEntry *)XNewPtr((long)sizeof(GM_SampleCallbackEntry));
	if (pEntry)
	{
		pEntry->frameOffset = frame;
		pEntry->pCallback = (GM_SampleFrameCallbackPtr)pCallback;
		pEntry->reference = (long)pReference;

		// add to linked list
		pNext = (GM_SampleCallbackEntry *)pSampleFrameVariable;
		while (pNext)
		{
			if (pNext->pNext == NULL)
			{
				break;
			}
			else
			{
				pNext = pNext->pNext;
			}
		}
		if (pNext == NULL)
		{
			pSampleFrameVariable = (void *)pEntry;
		}
		else
		{
			pNext->pNext = pEntry;
		}
		theErr = NO_ERR;
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAESound::RemoveSampleFrameCallback(unsigned long frame, HAESampleFrameCallbackPtr pCallback)
{
	GM_SampleCallbackEntry	*pEntry;
	GM_SampleCallbackEntry	*pNext, *pLast;
	OPErr					theErr;

	// find link
	pNext = (GM_SampleCallbackEntry *)pSampleFrameVariable;
	pEntry = NULL;
	while (pNext)
	{
		if (pNext->frameOffset == frame)
		{
			if (pNext->pCallback == (GM_SampleFrameCallbackPtr)pCallback)
			{
				pEntry = pNext;	// found
				break;
			}
		}
		pNext = pNext->pNext;
	}
	theErr = PARAM_ERR;
	if (pEntry)
	{
		// remove from linked list
		pLast = pNext = (GM_SampleCallbackEntry *)pSampleFrameVariable;
		while (pNext)
		{
			if (pNext == pEntry)								// found object in list?
			{
				if (pNext == (GM_SampleCallbackEntry *)pSampleFrameVariable)			// is object the top object
				{
					pSampleFrameVariable = (void *)pNext->pNext;		// yes, change to next object
				}
				else
				{
					if (pLast)									// no, change last to point beyond next
					{
						pLast->pNext = pNext->pNext;
					}
				}
				break;
			}
			pLast = pNext;
			pNext = pNext->pNext;
		}
		theErr = NO_ERR;
	}
	return PV_TranslateOPErr(theErr);
}



// load a custom sample. This will copy sample data into memory.
HAEErr HAESound::LoadCustomSample(void * sampleData,
								unsigned long frames,
								unsigned short int bitSize,
								unsigned short int channels,
								HAE_FIXED rate,
								unsigned long loopStart,
								unsigned long loopEnd)
{
	GM_Waveform 	*pWave;
	OPErr			theErr;
	long			size;

	theErr = NO_ERR;
	size = frames * channels * (bitSize / 8);
	pSoundVariables = XNewPtr(size);
	if (pSoundVariables)
	{
		XBlockMove(sampleData, pSoundVariables, size);
		pWave = (GM_Waveform *)XNewPtr((long)sizeof(GM_Waveform));
		if (pWave)
		{
			pWave->waveSize = size;
			pWave->waveFrames = frames;
			pWave->startLoop = loopStart;
			pWave->endLoop = loopEnd;
			pWave->baseMidiPitch = 60;
			pWave->bitSize = (unsigned char)bitSize;
			pWave->channels = (unsigned char)channels;
			pWave->sampledRate = rate;
	
			pWave->theWaveform = (SBYTE *)pSoundVariables;
		}
		else
		{
			XDisposePtr(pSoundVariables);
			pSoundVariables = NULL;
		}
		pFileVariables = pWave;
	}
	else
	{
		theErr = MEMORY_ERR;
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAESound::LoadResourceSample(void *pResource, unsigned long resourceSize)
{
	SampleDataInfo	newSoundInfo;
	GM_Waveform 	*pWave;
	XPTR			theData;
	OPErr			theErr;

	theErr = NO_ERR;
	if (pResource)
	{
		theData = XNewPtr(resourceSize);
		if (theData)
		{
			XBlockMove(pResource, theData, resourceSize);
			pResource = theData;
			theData = XGetSamplePtrFromSnd(pResource, &newSoundInfo);
			if (theData)
			{
				pSoundVariables = pResource;
				pWave = (GM_Waveform *)XNewPtr((long)sizeof(GM_Waveform));
				if (pWave)
				{
					pWave->waveSize = newSoundInfo.size;
					pWave->waveFrames = newSoundInfo.frames;
					pWave->startLoop = newSoundInfo.loopStart;
					pWave->endLoop = newSoundInfo.loopEnd;
					pWave->baseMidiPitch = (unsigned char)newSoundInfo.baseKey;
					pWave->bitSize = (unsigned char)newSoundInfo.bitSize;
					pWave->channels = (unsigned char)newSoundInfo.channels;
					pWave->sampledRate = newSoundInfo.rate;
					pWave->theWaveform = (SBYTE *)theData;

					pFileVariables = pWave;
				}
			}
			else
			{
				XDisposePtr(pResource);
				theErr = MEMORY_ERR;
			}
		}
		else
		{
			theErr = MEMORY_ERR;
		}
	}
	return PV_TranslateOPErr(theErr);
}

HAEErr HAESound::LoadBankSample(char *cName)
{
	OPErr			theErr;
	XPTR			thePreSound, theData;
	GM_Waveform 	*pWave;
	SampleDataInfo	newSoundInfo;
	long			size;

	theErr = BAD_SAMPLE;

#if X_PLATFORM != X_MACINTOSH
	if (thePatchFile)
#endif
	{
		theData = XGetSoundResourceByName(cName, &size);
		pSoundVariables = theData;
		if (theData)
		{
			// convert snd resource into a simple pointer of data with information

			thePreSound = XGetSamplePtrFromSnd(theData, &newSoundInfo);

			if (newSoundInfo.pMasterPtr != theData)
			{	// this means that XGetSamplePtrFromSnd created a new sample
				XDisposePtr(theData);
			}
			if (thePreSound)
			{
				pSoundVariables = newSoundInfo.pMasterPtr;
				pWave = (GM_Waveform *)XNewPtr((long)sizeof(GM_Waveform));
				if (pWave)
				{
					pWave->waveSize = newSoundInfo.size;
					pWave->waveFrames = newSoundInfo.frames;
					pWave->startLoop = newSoundInfo.loopStart;
					pWave->endLoop = newSoundInfo.loopEnd;
					pWave->baseMidiPitch = (unsigned char)newSoundInfo.baseKey;
					pWave->bitSize = (unsigned char)newSoundInfo.bitSize;
					pWave->channels = (unsigned char)newSoundInfo.channels;
					pWave->sampledRate = newSoundInfo.rate;
					pWave->theWaveform = (SBYTE *)thePreSound;

					pFileVariables = pWave;
					theErr = NO_ERR;
				}
			}
			else
			{
				theErr = MEMORY_ERR;
			}
		}
		else
		{
			theErr = BAD_SAMPLE;
		}
	}
	return PV_TranslateOPErr(theErr);
}

// set the loop points in sample frames
HAEErr HAESound::SetSampleLoopPoints(unsigned long start, unsigned long end)
{
	GM_Waveform 	*pWave;
	HAEErr			theErr;

	theErr = HAE_NO_ERROR;
	if (pFileVariables)
	{
		if ((start == 0) && end)
		{
			theErr = HAE_PARAM_ERR;
		}
		else
		{
			pWave = (GM_Waveform *)pFileVariables;
			pWave->startLoop = start;
			pWave->endLoop = end;
			GM_SetSampleLoopPoints(reference, start, end);
		}
	}
	else
	{
		theErr = HAE_NOT_SETUP;
	}
	return theErr;
}

// Get the current loop points in sample frames
HAEErr HAESound::GetSampleLoopPoints(unsigned long *pStart, unsigned long *pEnd)
{
	GM_Waveform 	*pWave;
	HAEErr			theErr;

	theErr = HAE_NO_ERROR;
	if (pFileVariables && pStart && pEnd)
	{
		pWave = (GM_Waveform *)pFileVariables;
		*pStart = pWave->startLoop;
		*pEnd = pWave->endLoop;
	}
	else
	{
		theErr = HAE_NOT_SETUP;
	}
	return theErr;
}

// Get the sample point offset by a sample frame count
void * HAESound::GetSamplePointer(unsigned long sampleFrame)
{
	char			*pSample;
	GM_Waveform 	*pWave;

	pSample = NULL;
	if (pFileVariables)
	{
		pWave = (GM_Waveform *)pFileVariables;
		pSample = (char *)pWave->theWaveform;
		if (pSample)
		{
			pSample += sampleFrame * (pWave->bitSize / 8) * pWave->channels;
		}
	}
	return (void *)pSample;
}

HAEErr HAESound::GetInfo(HAESampleInfo *pInfo)
{
	GM_Waveform 	*pWave;
	HAEErr			theErr;

	theErr = HAE_NO_ERROR;
	if (pFileVariables)
	{
		pWave = (GM_Waveform *)pFileVariables;
		pInfo->waveSize = pWave->waveSize;
		pInfo->waveFrames = pWave->waveFrames;
		pInfo->startLoop = pWave->startLoop;
		pInfo->endLoop = pWave->endLoop;
		pInfo->baseMidiPitch = pWave->baseMidiPitch;
		pInfo->bitSize = pWave->bitSize;
		pInfo->channels = pWave->channels;
		pInfo->sampledRate = pWave->sampledRate;
	}
	else
	{
		theErr = HAE_NOT_SETUP;
	}
	return theErr;
}

HAEErr HAESound::Start(HAE_FIXED sampleVolume, 			// sample volume
						short int stereoPosition,			// stereo placement
						long refData, 						// callback reference
						HAELoopDoneCallbackPtr pLoopContinueProc,
						HAEDoneCallbackPtr pDoneProc,
						unsigned long startOffsetFrame,
						HAE_BOOL stopIfPlaying)
{
	OPErr	theErr;
	long	volume;

	theErr = NO_ERR;
	if (pFileVariables)
	{
		if (IsPlaying())		// sample playing?
		{
			if (stopIfPlaying)	// can we stop it
			{
				Stop();
			}
			else
			{
				theErr = STILL_PLAYING;
			}
		}
		if (theErr == NO_ERR)
		{
			reference = 0;
			volume = (sampleVolume * MAX_NOTE_VOLUME) / 65536L;
			if (volume)
			{
				reference = GM_BeginSampleFromInfo((GM_Waveform *)pFileVariables, refData, 
													volume,
													stereoPosition,
													(GM_LoopDoneCallbackPtr)pLoopContinueProc, 
													(GM_SoundDoneCallbackPtr)pDoneProc,
													(INT32)startOffsetFrame);
				if (reference == -1L)
				{
					theErr = NO_FREE_VOICES;
				}
				else
				{
					GM_SetSampleOffsetCallbackLinks(reference, 
											(GM_SampleCallbackEntry *)pSampleFrameVariable);
				}
			}
			else
			{
				theErr = NO_VOLUME;
			}
		}
	}
	return PV_TranslateOPErr(theErr);
}

void HAESound::Stop(HAE_BOOL startFade)
{
	short int	sampleVolume;

	if (startFade)
	{
		sampleVolume = GM_GetSampleVolume(reference);
		GM_SetSampleFadeRate(reference, FLOAT_TO_FIXED(2.2),
												0, sampleVolume, TRUE);
	}
	else
	{
		GM_EndSample(reference);
		GM_SetSampleOffsetCallbackLinks(reference, NULL);
	}
}

// Get the position of a audio playback in samples
unsigned long HAESound::GetPlaybackPosition(void)
{
	return GM_GetSamplePlaybackPosition(reference);
}

HAEErr HAESound::StartCustomSample(void * sampleData,					// pointer to audio data
									unsigned long frames, 				// number of frames of audio
									unsigned short int bitSize, 		// bits per sample 8 or 16
									unsigned short int channels, 		// mono or stereo 1 or 2
									HAE_FIXED rate, 					// 16.16 fixed sample rate
									unsigned long loopStart, 			// loop start in frames
									unsigned long loopEnd,				// loop end in frames
									HAE_FIXED sampleVolume, 			// sample volume	(1.0)
									short int stereoPosition,			// stereo placement
									long refData, 						// callback reference
									HAELoopDoneCallbackPtr pLoopContinueProc,
									HAEDoneCallbackPtr pDoneProc,
									HAE_BOOL stopIfPlaying)
{
	OPErr	theErr;
	long	volume;

	theErr = NO_ERR;
	if (IsPlaying())		// sample playing?
	{
		if (stopIfPlaying)	// can we stop it
		{
			Stop();
		}
		else
		{
			theErr = STILL_PLAYING;
		}
	}
	if (theErr == NO_ERR)
	{
		reference = 0;
		volume = (sampleVolume * MAX_NOTE_VOLUME) / 65536L;
		if (volume)
		{
			reference = GM_BeginSample((G_PTR)sampleData, frames, rate, 
										loopStart, loopEnd, 
										volume, stereoPosition,
										refData, bitSize, channels, 
										(GM_LoopDoneCallbackPtr)pLoopContinueProc, 
										(GM_SoundDoneCallbackPtr)pDoneProc);
			if (reference == -1L)
			{
				theErr = NO_FREE_VOICES;
			}
			else
			{
				GM_SetSampleOffsetCallbackLinks(reference, 
										(GM_SampleCallbackEntry *)pSampleFrameVariable);
			}
		}
		else
		{
			theErr = NO_ERR;
		}
	}
	else
	{
		theErr = NO_VOLUME;
	}
	return PV_TranslateOPErr(theErr);
}

// Set a call back when song is done
void HAESound::SetDoneCallback(HAEDoneCallbackPtr pDoneProc, void * pReference)
{
	GM_SetSampleDoneCallback((long)reference, (GM_SoundDoneCallbackPtr)pDoneProc, (long)pReference);
}


HAE_BOOL HAESound::IsPlaying(void)
{
	return (HAE_BOOL) ! GM_IsSoundDone(reference);
}

HAE_BOOL HAESound::IsDone(void)
{
	return ! IsPlaying();
}

void HAESound::SetRate(HAE_FIXED newRate)
{
	GM_ChangeSamplePitch(reference, newRate);
}

HAE_FIXED HAESound::GetRate(void)
{
	return GM_GetSamplePitch(reference);
}

void HAESound::SetVolume(HAE_FIXED volume)
{
	GM_ChangeSampleVolume(reference, (short)((volume * MAX_NOTE_VOLUME) / 65536L));
}

HAE_FIXED HAESound::GetVolume(void)
{
	return (GM_GetSampleVolume(reference) * 65536L) / MAX_NOTE_VOLUME;
}

void HAESound::Fade(HAE_BOOL doAsync)
{
	short int	sampleVolume;

	sampleVolume = GM_GetSampleVolume(reference);
	if (doAsync == FALSE)
	{
		// We're going to fade the sample out and don't stop it		

		GM_SetSampleFadeRate(reference, FLOAT_TO_FIXED(2.2),
												0, sampleVolume, FALSE);
		while (GM_GetSampleVolume(reference) && (GM_IsSoundDone(reference) == FALSE)) 
		{
			GetMixer()->ServiceAudioOutputToFile();
			GetMixer()->ServiceIdle();
			XWaitMicroseocnds(1000);
		}
	}
	else
	{
		GM_SetSampleFadeRate(reference, FLOAT_TO_FIXED(2.2),
												0, sampleVolume, FALSE);
	}
}


void HAESound::FadeTo(HAE_FIXED destVolume, HAE_BOOL doAsync)
{
	short int	soundVolume, saveVolume, newSoundVolume, saveNewSoundVolume;
	FIXED_VALUE	delta;

	newSoundVolume = (destVolume * MAX_NOTE_VOLUME) / 65536L;
	saveNewSoundVolume = newSoundVolume;
	// We're going to fade the Sound out and don't stop it		
	soundVolume = GM_GetSampleVolume(reference);
	saveVolume = soundVolume;

	if (newSoundVolume < soundVolume)
	{	// fade out
		soundVolume = newSoundVolume;
		newSoundVolume = saveVolume;
		delta = FLOAT_TO_FIXED(2.2);
	}
	else
	{	// fade in
		delta = FLOAT_TO_FIXED(-2.2);
	}
	if (doAsync == FALSE)
	{
		GM_SetSampleFadeRate(reference, delta, soundVolume, newSoundVolume, FALSE);
		while (GM_GetSampleVolume(reference) != saveNewSoundVolume)
		{
			GetMixer()->ServiceAudioOutputToFile();
			GetMixer()->ServiceIdle();
			XWaitMicroseocnds(1000);
		}
	}
	else
	{
		GM_SetSampleFadeRate(reference, delta, soundVolume, newSoundVolume, FALSE);
	}
}

void HAESound::SetStereoPosition(short int stereoPosition)
{
	GM_ChangeSampleStereoPosition(reference, stereoPosition);
}

short int HAESound::GetStereoPosition(void)
{
	return GM_GetSampleStereoPosition(reference);
}

void HAESound::SetReverb(HAE_BOOL useReverb)
{
	GM_ChangeSampleReverb(reference, useReverb);
}

HAE_BOOL HAESound::GetReverb(void)
{
	return GM_GetSampleReverb(reference);
}

#if 0
	#pragma mark ### HAEMod class ###
#endif
#if USE_MOD_API
HAEMod::HAEMod(HAEAudioMixer *pHAEAudio, 
					 char *pName, void * userReference) :
									HAEAudioNoise(pHAEAudio)
{
	userReference = userReference;
	pauseVariable = 0;
	pSoundVariables = NULL;
	pName = pName;
}

HAEMod::~HAEMod()
{
	Stop();

	if (pSoundVariables)
	{
		GM_FreeModFile((GM_ModData *)pSoundVariables);
		pSoundVariables = NULL;
	}
}

// currently paused
HAE_BOOL HAEMod::IsPaused(void)
{
	return (pauseVariable) ? TRUE : FALSE;
}

void HAEMod::Pause(void)
{
	if (pauseVariable == 0)
	{
		pauseVariable = 1;
		GM_PauseMod((GM_ModData *)pSoundVariables);
	}
}

void HAEMod::Resume(void)
{
	if (pauseVariable)
	{
		pauseVariable = 0;
		GM_ResumeMod((GM_ModData *)pSoundVariables);
	}
}

// Load file into sound object. This will copy the file directly into memory. It
// will get disposed once you destroy this object.
HAEErr HAEMod::LoadFromFile(HAEPathName pModFilePath)
{
	XFILENAME		theFile;
	OPErr			theErr;
	long			fileSize;
	XPTR			pFileVariables;

	theErr = NO_ERR;
	XConvertNativeFileToXFILENAME(pModFilePath, &theFile);
	pFileVariables = PV_GetFileAsData(&theFile, &fileSize);
	if (pFileVariables)
	{
		pSoundVariables = GM_LoadModFile(pFileVariables, fileSize);
		if (pSoundVariables == NULL)
		{
			theErr = BAD_FILE;
		}
		XDisposePtr(pFileVariables);	// throw away file, we've parsed the mod file
	}
	else
	{
		theErr = MEMORY_ERR;
	}
	return PV_TranslateOPErr(theErr);
}

// Load memory mapped MOD pointer into HAEMod object. This will parse the MOD file and get
// it ready for playing. You can dispose of the data passed once this method returns
HAEErr HAEMod::LoadFromMemory(void *pModData, unsigned long modSize)
{
	OPErr			theErr;

	theErr = PARAM_ERR;
	if (pModData && modSize)
	{
		pSoundVariables = GM_LoadModFile(pModData, modSize);
		if (pSoundVariables == NULL)
		{
			theErr = BAD_FILE;
		}
		else
		{
			theErr = NO_ERR;
		}
	}
	return PV_TranslateOPErr(theErr);
}

static void PV_ModDoneCallback(GM_ModData *pMod)
{
	HAEDoneCallbackPtr pDoneProc;

	pDoneProc = (HAEDoneCallbackPtr)pMod->reference2;
	if (pDoneProc)
	{
		(*pDoneProc)((void *)pMod->reference);
	}		
}

HAEErr HAEMod::Start(HAEDoneCallbackPtr pDoneProc)
{
	OPErr	theErr;

	theErr = BAD_FILE;
	if (pSoundVariables)
	{
		if (IsPlaying())
		{
			Stop();
		}
		((GM_ModData *)pSoundVariables)->reference2 = (long)pDoneProc;
		GM_BeginModFile((GM_ModData *)pSoundVariables, 
						(GM_ModDoneCallbackPtr)PV_ModDoneCallback, (long)GetReference());
		theErr = NO_ERR;
	}
	return PV_TranslateOPErr(theErr);
}

void HAEMod::Stop(HAE_BOOL startFade)
{
	short int	songVolume;

	if (pSoundVariables)
	{
		if (startFade)
		{
			songVolume = GM_GetModVolume((GM_ModData *)pSoundVariables);
			GM_SetModFadeRate((GM_ModData *)pSoundVariables, FLOAT_TO_FIXED(2.2),
													0, songVolume, TRUE);
		}
		else
		{
			GM_StopModFile((GM_ModData *)pSoundVariables);
		}
	}
}


void HAEMod::Fade(HAE_BOOL doAsync)
{
	short int	songVolume;

	if (pSoundVariables)
	{
		songVolume = GM_GetModVolume((GM_ModData *)pSoundVariables);
		if (doAsync == FALSE)
		{
			// We're going to fade the song out and don't stop it
			GM_SetModFadeRate((GM_ModData *)pSoundVariables, FLOAT_TO_FIXED(2.2),
													0, songVolume, FALSE);
			while (	GM_GetModVolume((GM_ModData *)pSoundVariables) && 
					GM_IsModPlaying((GM_ModData *)pSoundVariables)) 
			{
				GetMixer()->ServiceAudioOutputToFile();
				GetMixer()->ServiceIdle();
				XWaitMicroseocnds(1000);
			}
		}
		else
		{
			GM_SetModFadeRate((GM_ModData *)pSoundVariables, FLOAT_TO_FIXED(2.2),
															0, songVolume, FALSE);
		}
	}
}

// FadeTo a volume level
void HAEMod::FadeTo(HAE_FIXED destVolume, HAE_BOOL doAsync)
{
	short int	songVolume, saveVolume, newSongVolume, saveNewSongVolume;
	FIXED_VALUE	delta;

	if (pSoundVariables)
	{
		newSongVolume = (destVolume * MAX_SONG_VOLUME) / 65536L;
		saveNewSongVolume = newSongVolume;
		// We're going to fade the song out before we stop it.		
		songVolume = GM_GetModVolume((GM_ModData *)pSoundVariables);
		saveVolume = songVolume;

		if (newSongVolume < songVolume)
		{	// fade out
			songVolume = newSongVolume;
			newSongVolume = saveVolume;
			delta = FLOAT_TO_FIXED(2.2);
		}
		else
		{	// fade in
			delta = FLOAT_TO_FIXED(-2.2);
		}
		if (doAsync == FALSE)
		{
			GM_SetModFadeRate((GM_ModData *)pSoundVariables, delta,
													songVolume, newSongVolume, FALSE);
			while (GM_GetModVolume((GM_ModData *)pSoundVariables) != saveNewSongVolume)
			{
				GetMixer()->ServiceAudioOutputToFile();
				GetMixer()->ServiceIdle();
				XWaitMicroseocnds(1000);
			}
		}
		else
		{
			GM_SetModFadeRate((GM_ModData *)pSoundVariables, delta,
													songVolume, newSongVolume, FALSE);
		}
	}
}

// Set a call back when song is done
void HAEMod::SetDoneCallback(HAEDoneCallbackPtr pDoneProc, void * pReference)
{
	if (pSoundVariables)
	{
		((GM_ModData *)pSoundVariables)->reference = (long)pReference;
		((GM_ModData *)pSoundVariables)->reference2 = (long)pDoneProc;
	}
}


// set song master tempo. (1.0 uses songs encoded tempo, 2.0 will play
// song twice as fast, and 0.5 will play song half as fast
void HAEMod::SetMasterTempo(HAE_FIXED tempoFactor)
{
	if (pSoundVariables)
	{
		GM_SetModTempoFactor((GM_ModData *)pSoundVariables, tempoFactor);
	}
}


HAE_BOOL HAEMod::IsPlaying(void)
{
	HAE_BOOL	play;

	play = FALSE;
	if (pSoundVariables)
	{
		play = GM_IsModPlaying((GM_ModData *)pSoundVariables);
	}
	return play;
}

HAE_BOOL HAEMod::IsDone(void)
{
	return ! IsPlaying();
}

void HAEMod::SetVolume(HAE_FIXED volume)
{
	if (pSoundVariables)
	{
		GM_SetModVolume((GM_ModData *)pSoundVariables, (volume * MAX_SONG_VOLUME) / 65536L);
	}
}

HAE_FIXED HAEMod::GetVolume(void)
{
	HAE_FIXED	value;

	value = 0;
	if (pSoundVariables)
	{
		value = (GM_GetModVolume((GM_ModData *)pSoundVariables) * 65536L) / MAX_SONG_VOLUME;
	}
	return value;
}

// sets tempo in beats per minute
void HAEMod::SetTempoInBeatsPerMinute(unsigned long newTempoBPM)
{
	if (pSoundVariables)
	{
		GM_SetModTempoBPM((GM_ModData *)pSoundVariables, newTempoBPM);
	}
}

// returns tempo in beats per minute
unsigned long HAEMod::GetTempoInBeatsPerMinute(void)
{
	unsigned long tempo;
	
	tempo = 0;
	if (pSoundVariables)
	{
		tempo = (unsigned long)GM_GetModTempoBPM((GM_ModData *)pSoundVariables);
	}
	return tempo;
}

// pass TRUE to entire loop song, FALSE to not loop
void HAEMod::SetLoopFlag(HAE_BOOL loop)
{
	if (pSoundVariables)
	{
		GM_SetModLoop((GM_ModData *)pSoundVariables, loop);
	}
}

HAE_BOOL HAEMod::GetLoopFlag(void)
{
	if (pSoundVariables)
	{
		return (HAE_BOOL)GM_GetModLoop((GM_ModData *)pSoundVariables);
	}
	return FALSE;
}

unsigned long HAEMod::GetInfoSize(HAEInfoTypes infoType)
{
	unsigned long	size;

	size = 0;

	if (pSoundVariables)
	{
		switch (infoType)
		{
			case TITLE_INFO:
				size = GM_GetModSongNameLength((GM_ModData *)pSoundVariables);
				break;
			case COMPOSER_NOTES_INFO:
				size = GM_GetModSongCommentsLength((GM_ModData *)pSoundVariables);
				break;
		}
	}
	return size;
}

HAEErr HAEMod::GetInfo(HAEInfoTypes infoType, char *cInfo)
{
	if (pSoundVariables && cInfo)
	{
		cInfo[0] = 0;
		switch (infoType)
		{
			case TITLE_INFO:
				GM_GetModSongName((GM_ModData *)pSoundVariables, cInfo);
				break;
			case COMPOSER_NOTES_INFO:
				GM_GetModSongComments((GM_ModData *)pSoundVariables, cInfo);
				break;
		}
	}
	return HAE_NO_ERROR;
}

#endif	// USE_MOD_API

// EOF of HAE.cpp



