/*
 * @(#)HaeEnv.cpp	1.6 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


#include "Hae_Impl.h"
#include "HaeEnv.h"
#include "HaeMixer_constants.h"


////////////////////////////////////////////////// OPERATIONS:

// Throws a Java exception.
// You'd think this wouldn't return on exception but it does.
jboolean HaeEnv::ThrowOnError(OPErr opErr)
{
	switch (opErr)
	{

	// This really isn't an error, more of a result.
	case STREAM_STOP_PLAY:
	case NO_ERR:
		return JNI_FALSE;

	case PARAM_ERR:
		return Throw("Bad Parameters");
	case MEMORY_ERR:
		return Throw("Out of Memory");
	case GENERAL_BAD:
		return Throw("Callback Failure");
	case NOT_REENTERANT:
		return Throw("Only one HaeMixer may be created.");
	case EXTRA_ERROR_PLATFORM:
		return Throw("Failure to initialize platform's audio system.");
	//DAVID: add the next 2 lines
	case EXTRA_ERROR_NOT_IMPL:
		return Throw("Feature not implemented.");

	case BAD_MIDI_DATA:
		return Throw("Bad MIDI Data");
	case BAD_SAMPLE:
		return Throw("Bad Sample Data");
	case BAD_FILE:
		return Throw("Bad File Data");

	case TOO_MANY_SONGS_PLAYING:
		return Throw("No More Songs Available");
	case NO_FREE_VOICES:
		return Throw("No More Voices Available");

// $$kk: start changes
	case DEVICE_UNAVAILABLE:
		return Throw("Audio Device Unavailable");
// $$kk: end changes

// $$kk: 03.17.98: start changes
	case UNSUPPORTED_HARDWARE:
		return Throw("Unsupported audio hardware.  System must support linear 8 bit or 16 bit rendering");
// $$kk: 03.17.98: end changes

	// errors that shouldn't happen
	case BAD_INSTRUMENT:
		return Throw("Bad Instrument (shouldn't have happened)");
	case ALREADY_PAUSED:
		return Throw("Already Paused (shouldn't have happened)");
	case ALREADY_RESUMED:
		return Throw("Already Resumed (shouldn't have happened)");
	case NO_SONG_PLAYING:
		return Throw("No Song Playing (shouldn't have happened)");
	default:
		return Throw("Unknown Error (shouldn't have happened)");
	}
}
jboolean HaeEnv::Throw(ConStr message)
{
	JncEnv::Throw(JAVA_PACKAGE_NAME"/HaeException", message);	//KARA4: changed this line
	return JNI_TRUE;
}


////////////////////////////////////////////////// DATA:

jint HaeEnv::TranslateQuality(jint haeQuality)
{
	switch (haeQuality)
	{
	default:
		ThrowOnError(PARAM_ERR);		// returns!
	case HaeJava_HaeMixer_SAMPLE_RATE_11K:
		return Q_11K;
	case HaeJava_HaeMixer_SAMPLE_RATE_22K:
		return Q_22K;
	case HaeJava_HaeMixer_SAMPLE_RATE_44K:
		return Q_44K;
	}
}
jint HaeEnv::TranslateInterpolation(jint haeInterpolation)
{
	switch (haeInterpolation)
	{
	default:
		ThrowOnError(PARAM_ERR);		// returns!
	case HaeJava_HaeMixer_INTERPOLATION_DROP_SAMPLE:
		return E_AMP_SCALED_DROP_SAMPLE;
	case HaeJava_HaeMixer_INTERPOLATION_2_POINT:
		return E_2_POINT_INTERPOLATION;
	case HaeJava_HaeMixer_INTERPOLATION_LINEAR:
		return E_LINEAR_INTERPOLATION;
	}
}
jint HaeEnv::TranslateModifiers(jint haeModifiers)
{
jint			modifiers;

	modifiers = 0;

	if (haeModifiers & HaeJava_HaeMixer_FLAGS_USE_16)
	{
		modifiers |= M_USE_16;
	}
	if ((haeModifiers & HaeJava_HaeMixer_FLAGS_USE_STEREO) && ::XIsStereoSupported())
	{
		modifiers |= M_USE_STEREO;
		if (haeModifiers & HaeJava_HaeMixer_FLAGS_STEREO_FILTER)
		{
			modifiers |= M_STEREO_FILTER;
		}
	}
	if (!(haeModifiers & HaeJava_HaeMixer_FLAGS_ENABLE_REVERB))
	{
		modifiers |= M_DISABLE_REVERB;
	}

	return modifiers;
}
ReverbMode HaeEnv::TranslateReverbType(jint haeReverb)
{
	switch (haeReverb)
	{
	default:
		ThrowOnError(PARAM_ERR);		// returns!
	case HaeJava_HaeMixer_REVERB_TYPE_1:
		return REVERB_TYPE_1;
	case HaeJava_HaeMixer_REVERB_TYPE_2:
		return REVERB_TYPE_2;
	case HaeJava_HaeMixer_REVERB_TYPE_3:
		return REVERB_TYPE_3;
	case HaeJava_HaeMixer_REVERB_TYPE_4:
		return REVERB_TYPE_4;
	case HaeJava_HaeMixer_REVERB_TYPE_5:
		return REVERB_TYPE_5;
	case HaeJava_HaeMixer_REVERB_TYPE_6:
		return REVERB_TYPE_6;
	case HaeJava_HaeMixer_REVERB_TYPE_7:
		return REVERB_TYPE_7;
	case HaeJava_HaeMixer_REVERB_TYPE_8:
		return REVERB_TYPE_8;
	}
}

jint HaeEnv::GetUniqueId()
{
STA jint		id = 0;

	return id++;
}


