/*
 * @(#)HaeMidiNoise_Impl.cpp	1.9 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


#include "Hae_Impl.h"
#include "sun_audio_HaeMidiNoise.h"
#include "HaeMidiNoise.h"
#include "JncByteArray.h"


// $$jdr: This is TOTALLY BOGUS and must be fixed.
// This shouldn't have to be here, but something
// is referencing delete in this code, though we haven't
// seen it called.

//#if 0	// $$cls: The problem seems to have gone away, this definition is no longer needed
#if 1   // $$kk:  The problem is still there for me, at least on solaris.
/*static*/ void operator delete(void* p)
{
	fprintf(stderr, "AUDIO ENGINE ERROR: "
		         "Delete shouldn't be called in HaeMidiNoise.");
}
#endif


////////////////////////////////////////////////// VARIABLES:

//KARA7: deleted static class cache allocation
static char     const songDataName[] = "m_songData";


////////////////////////////////////////////////// FORWARDS:

static jboolean IsPercussionChannel(jbyte channel);
static jboolean LoadInstrument(HaeEnv& env, GM_Song *song, long instrument);
static long GetInstrumentNumber(jbyte bank, jbyte program, jbyte channel, jbyte note);
static long GetPercussionInstrumentNumber(jbyte bank, jbyte note);
static long GetTonalInstrumentNumber(jbyte bank, jbyte program);

static jint VerifyTick(jint tick);


////////////////////////////////////////////////// MEMBERS:

void HaeMidiNoise::SetSongData(GM_Song* data)
{
    SetPointerField(songDataName, data);
}
GM_Song* HaeMidiNoise::GetSongData()
{
    return (GM_Song*)GetPointerField(songDataName);
}


////////////////////////////////////////////////// OPERATIONS:

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_close(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_KillSongNotes(song);
        song->disposeSongDataWhenDone = TRUE;   // free our midi pointer
        GM_FreeSong(song);
        thiz.SetSongData(NULL);
    }
}


JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_preloadInstrument(JNIEnv* e, jobject thisObj,
                                            jbyte bank, jbyte program)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        ::LoadInstrument(env, song, ::GetTonalInstrumentNumber(bank, program));
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_preloadPercussionInstrument(JNIEnv* e, jobject thisObj,
                                                        jbyte bank, jbyte program, jbyte note)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        ::LoadInstrument(env, song, ::GetPercussionInstrumentNumber(bank, note));
    }
}



JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_noteOn(JNIEnv* e, jobject thisObj,
                                    jint tick, jbyte channel,
                                    jbyte note,
                                    jbyte velocity)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        if (IsPercussionChannel(channel))   //KARA2: add these 9 lines
        {
        jbyte   const bank = ::GM_GetControllerValue(song, channel, 0);

            if (!::LoadInstrument(env, song, ::GetPercussionInstrumentNumber(bank, note)))
            {
                return; // exception thrown
            }
        }

        QGM_LockExternalMidiQueue();
        QGM_NoteOn(song, VerifyTick(tick), channel, note, velocity);
        QGM_UnlockExternalMidiQueue();
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_noteOff(JNIEnv* e, jobject thisObj,
                                    jint tick, jbyte channel,
                                    jbyte note,
                                    jbyte velocity)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        QGM_LockExternalMidiQueue();
        QGM_NoteOff(song, VerifyTick(tick), channel, note, velocity);
        QGM_UnlockExternalMidiQueue();
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_bankPatch(JNIEnv* e, jobject thisObj,
                                        jint tick, jbyte channel,
                                        jbyte bank,
                                        jbyte patch)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        if (!IsPercussionChannel(channel))  //KARA2: add these 7 lines
        {
            if (!::LoadInstrument(env, song, ::GetTonalInstrumentNumber(bank, patch)))
            {
                return; // exception thrown
            }
        }

        QGM_LockExternalMidiQueue();
        tick = VerifyTick(tick);
        QGM_Controller(song, tick, channel, 0, bank);
        QGM_ProgramChange(song, tick, channel, patch);
        QGM_UnlockExternalMidiQueue();
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_patch(JNIEnv* e, jobject thisObj,
                                    jint tick, jbyte channel,
                                    jbyte patch)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        if (!IsPercussionChannel(channel))  //KARA2: add these 9 lines
        {
        jbyte   const bank = GM_GetControllerValue(song, channel, 0);

            if (!::LoadInstrument(env, song, ::GetTonalInstrumentNumber(bank, patch)))
            {
                return; // exception thrown
            }
        }
        QGM_LockExternalMidiQueue();
        QGM_ProgramChange(song, VerifyTick(tick), channel, patch);
        QGM_UnlockExternalMidiQueue();
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_controller(JNIEnv* e, jobject thisObj,
                                        jint tick, jbyte channel,
                                        jbyte controller,
                                        jbyte value)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        QGM_LockExternalMidiQueue();
        QGM_Controller(song, VerifyTick(tick), channel, controller, value);
        QGM_UnlockExternalMidiQueue();
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_pitchBend(JNIEnv* e, jobject thisObj,
                                        jint tick, jbyte channel,
                                        jint value)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        QGM_LockExternalMidiQueue();
        QGM_PitchBend(song, VerifyTick(tick), channel,
                        (unsigned char)(value >> 7 & 0x007F), (unsigned char)(value & 0x007F));
        QGM_UnlockExternalMidiQueue();
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_allNotesOff(JNIEnv* e, jobject thisObj,
                                        jint tick)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        QGM_LockExternalMidiQueue();
        QGM_AllNotesOff(song, VerifyTick(tick));
        QGM_UnlockExternalMidiQueue();
    }
}


JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_setVolume(JNIEnv* e, jobject thisObj, jdouble volume)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_SetSongVolume(song, DOUBLE_TO_VOLUME(volume));
    }
}
// set note offset in semi tones    (12 is down an octave, -12 is up an octave)
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_setPitchOffset(JNIEnv* e, jobject thisObj, jint offset)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_SetSongPitchOffset(song, offset);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_muteChannel(JNIEnv* e, jobject thisObj, jint channel)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_MuteChannel(song, (short)channel);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_unmuteChannel(JNIEnv* e, jobject thisObj, jint channel)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_UnmuteChannel(song, (short)channel);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_soloChannel(JNIEnv* e, jobject thisObj, jint channel)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_SoloChannel(song, (short)channel);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiNoise_unSoloChannel(JNIEnv* e, jobject thisObj, jint channel)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_UnsoloChannel(song, (short)channel);
    }
}


////////////////////////////////////////////////// DATA:

JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeMidiNoise_getVolume(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        return VOLUME_TO_DOUBLE(GM_GetSongVolume(song));
    }
    return 0.0;
}
// return note offset in semi tones (12 is down an octave, -12 is up an octave)
JNIEXPORT jint JNICALL
Java_sun_audio_HaeMidiNoise_getPitchOffset(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        return GM_GetSongPitchOffset(song);
    }
    return 0;
}

JNIEXPORT jbyteArray JNICALL
Java_sun_audio_HaeMidiNoise_getChannelMuteStatus(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
    char        channels[16];

        GM_GetChannelMuteStatus(song, channels);
        return JncByteArray::New(env, 16, (jbyte*)channels);    //KARA7: changed this line
    }
    return NULL;
}
JNIEXPORT jbyteArray JNICALL
Java_sun_audio_HaeMidiNoise_getChannelSoloStatus(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiNoise    thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
    char        channels[16];

        GM_GetChannelSoloStatus(song, channels);
        return JncByteArray::New(env, 16, (jbyte*)channels);    //KARA7: changed this line
    }
    return NULL;
}

// get current midi tick in microseconds
JNIEXPORT jint JNICALL
Java_sun_audio_HaeMidiNoise_getTick(JNIEnv* e, jobject thisObj)
{
    return GM_GetSyncTimeStampQuantizedAhead();
}


////////////////////////////////////////////////// LOCAL FUNCTIONS:

static jint VerifyTick(jint tick)
{
    if (tick == 0)
    {
        tick = ::GM_GetSyncTimeStampQuantizedAhead();
    }
    return tick;
}

static jboolean IsPercussionChannel(jbyte channel)
{
    //??STEVE: need better test, right?
    return (channel == PERCUSSION_CHANNEL) ? TRUE
                                           : FALSE;
}
static jboolean LoadInstrument(HaeEnv& env, GM_Song *song, long instrument)
{
    if (song->instrumentData[instrument])
    {   // already loaded
        return TRUE;
    }
    else
    {
        return !env.ThrowOnError(::GM_LoadSongInstrument(song, instrument));
    }
}
static long GetInstrumentNumber(jbyte bank, jbyte program, jbyte channel, jbyte note)
{
    if (IsPercussionChannel(channel))
    {
        return ::GetPercussionInstrumentNumber(bank, note);
    }
    else
    {
        return ::GetTonalInstrumentNumber(bank, program);
    }
}
static long GetPercussionInstrumentNumber(jbyte bank, jbyte note)
{
long        instrument;

    instrument = ::GetTonalInstrumentNumber(bank, 0);
    instrument += 128;
    instrument += note;
    return instrument;
}
static long GetTonalInstrumentNumber(jbyte bank, jbyte program)
{
    if (bank >= MAX_BANKS / 2) bank = 0;
    return (long)bank * 256 + program;
}






