/*
 * @(#)HaeMidiSong_Impl.cpp	1.12 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


#include "Hae_Impl.h"
#include "sun_audio_HaeMidiNoise.h"
#include "sun_audio_HaeMidiSong.h"
#include "HaePath.h"
#include "HaeMixer.h"
#include "HaeMidiSong.h"
#include "JncByteArray.h"


////////////////////////////////////////////////// VARIABLES:

//KARA7: deleted static class cache allocation
static char     const stringDataName[] = "m_stringData";


////////////////////////////////////////////////// FORWARDS:

static jstring GetInfoString(JNIEnv* e, jobject thisObj, SongInfo infoCode);

static void HaeMidiSong_open(HaeEnv& env, jobject thisObj,
                                XPTR midiData, long midiBytes);


////////////////////////////////////////////////// MEMBERS:

void HaeMidiSong::SetStringData(SongResource* data)
{
    SetPointerField(stringDataName, data);
}
SongResource* HaeMidiSong::GetStringData()
{
    return (SongResource*)GetPointerField(stringDataName);
}


////////////////////////////////////////////////// OPERATIONS:

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_open(JNIEnv* e, jobject thisObj, jbyteArray dataObj)
{
    HaeEnv         env(e);
    JncByteArray    data(env, dataObj);
    jsize           dataBytes;

    Java_sun_audio_HaeMidiSong_close(e, thisObj);

    dataBytes = data.GetLength();
    if (dataBytes > 0)
    {
    XPTR            const dataCopy = ::XNewPtr(dataBytes);

        if (dataCopy)
        {
            data.CopyOut((jbyte*)dataCopy, 0, dataBytes);
            ::HaeMidiSong_open(env, thisObj, dataCopy, dataBytes);
        }
        else
        {
            env.ThrowOnError(MEMORY_ERR);
        }
    }
    else
    {
        ERROR0("HaeMidiSong.open(byte[]) called with zero-length array\n");
    }
}

#if HAE_FILES
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_openFile(JNIEnv* e, jobject thisObj, jstring pathObj)
{
    HaeEnv             env(e);
    HaePath             path(env, pathObj);
    long                midiBytes;
    XPTR                midiData;

    Java_sun_audio_HaeMidiSong_close(env, thisObj);

    midiData = path.ReadFile(&midiBytes);
    if (midiData)
    {
        ::HaeMidiSong_open(env, thisObj, midiData, midiBytes);
    }
    else
    {
        env.ThrowOnError(BAD_FILE);
    }
}
#endif // HAE_FILES

void HaeMidiSong_open(HaeEnv& env, jobject thisObj,
                        XPTR midiData, long midiBytes)
{
    HaeMidiSong         thiz(env, thisObj);
    HaeMixer            mixer(env, thiz.GetMixer());
    jint                id;
    SongResource*       xSong;
    OPErr               opErr;

    id = HaeEnv::GetUniqueId();
    xSong = ::XNewSongPtr(SONG_TYPE_SMS, (short)id,
                            (short)mixer.GetMidiVoiceCount(),
                            (short)mixer.GetVolumeDivisor(),
                            (short)mixer.GetSoundVoiceCount(),
                            env.TranslateReverbType(mixer.GetReverbType()));
    if (xSong)
    {
    GM_Song*            song;

        thiz.SetStringData(xSong);

        song = ::GM_LoadSong((short)id, (void*)xSong,
                                midiData, midiBytes,
                                NULL,       // no callback
                                TRUE,       // load instruments
                                TRUE,		// ignore bad instruments $$kk: 02.17.98: changed this to TRUE.
											// this is a hack to get around the JMF problem where erratically a MIDI
											// song tries and fails to load an invalid instrument.
//                                FALSE,      // ignore bad instruments
                                &opErr);
        if (song)
        {
            ::GM_SetSongLoopFlag(song, FALSE);      // don't loop song
            song->disposeSongDataWhenDone = TRUE;   // dispose of midi data
            song->userReference = id;
            thiz.SetSongData(song);
            return;
        }
    }
    else
    {
        opErr = MEMORY_ERR;
    }

    // failure!
    ::XDisposePtr(midiData);
    env.ThrowOnError(opErr);
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_close(JNIEnv* e, jobject thisObj)
{
HaeEnv              env(e);
HaeMidiSong         thiz(env, thisObj);

    Java_sun_audio_HaeMidiNoise_close(e, thisObj);

    {
    SongResource*   const stringData = thiz.GetStringData();

        if (stringData)
        {
            XDisposeSongPtr(stringData);
            thiz.SetStringData(NULL);
        }
    }
}


JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_start(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
		// $$kk: 03.19.98: JMF player bug #4114584
		// third param is "useEmbeddedMixerSettings."
		// if this is true, the song will  call GM_ChangeSystemVoices and
		// reallocate the MusicGlobal SYSTEM voices among a total of *32*
		//  midi voices and effect voices.  this can cause a stream voice to
		// become invalid and stop playing.  JMF player bug #4114584
        // env.ThrowOnError(GM_BeginSong(song, NULL, TRUE));
        env.ThrowOnError(GM_BeginSong(song, NULL, FALSE));
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_stop(JNIEnv* e, jobject thisObj, jboolean withFade)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        if (withFade)
        {
        short int   songVolume;

            songVolume = GM_GetSongVolume(song);
            GM_SetSongFadeRate(song, DOUBLE_TO_FIXED(2.2), 0, songVolume, TRUE);
        }
        else
        {
            GM_KillSongNotes(song);
            GM_EndSong(song);
        }
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_pause(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_PauseSong(song);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_resume(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_ResumeSong(song);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_adjustTempo(JNIEnv* e, jobject thisObj, jdouble newTempo)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_SetMasterSongTempo(song, DOUBLE_TO_FIXED(newTempo));
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_setLoopFlag(JNIEnv* e, jobject thisObj,
                                        jboolean doLoop)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_SetSongLoopFlag(song, (BOOL_FLAG)doLoop);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_muteTrack(JNIEnv* e, jobject thisObj, jint track)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_MuteTrack(song, (short)track);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_unmuteTrack(JNIEnv* e, jobject thisObj, jint track)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_UnmuteTrack(song, (short)track);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_soloTrack(JNIEnv* e, jobject thisObj, jint track)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_SoloTrack(song, (short)track);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_unSoloTrack(JNIEnv* e, jobject thisObj,
                            jint track)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        GM_UnsoloTrack(song, (short)track);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_setTickPosition(JNIEnv* e, jobject thisObj, jint ticks)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        env.ThrowOnError(GM_SetSongTickPosition(song, ticks));
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMidiSong_setMicrosecondPosition(JNIEnv* e, jobject thisObj, jint usec)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        env.ThrowOnError(GM_SetSongMicrosecondPosition(song, usec));
    }
}


////////////////////////////////////////////////// DATA:

JNIEXPORT jboolean JNICALL
Java_sun_audio_HaeMidiSong_isPlaying(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        return GM_IsSongDone(song) ? JNI_FALSE
                                   : JNI_TRUE;
    }
    return JNI_FALSE;
}
JNIEXPORT jboolean JNICALL
Java_sun_audio_HaeMidiSong_isPaused(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
//        return !GM_IsSongDone(song);
		// $$kk: 02.10.98
		// this paused thing doesn't work.  GM_IsSongDone returns false from when
		// you say "start" until you say "stop" regardless of "pause" and "resume."
		// i think a song is paused if 1) it is not done and 2) it is paused
		return ( (!GM_IsSongDone(song)) && (song->songPaused) ? JNI_TRUE : JNI_FALSE);

    }
    return JNI_FALSE;
}

JNIEXPORT jbyteArray JNICALL
Java_sun_audio_HaeMidiSong_getTrackMuteStatus(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
    char        tracks[64];

        GM_GetTrackMuteStatus(song, tracks);
        return JncByteArray::New(env, 64, (jbyte*)tracks);  //KARA7: changed this line
    }
    return NULL;
}
JNIEXPORT jbyteArray JNICALL
Java_sun_audio_HaeMidiSong_getSoloTrackStatus(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
    char        tracks[64];

        GM_GetTrackSoloStatus(song, tracks);
        return JncByteArray::New(env, 64, (jbyte*)tracks);  //KARA7: changed this line
    }
    return NULL;
}

// get ticks in microseconds of length of song
JNIEXPORT jint JNICALL
Java_sun_audio_HaeMidiSong_getTickLength(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
    OPErr           opErr;
    unsigned long   ticks;

        ticks = GM_GetSongTickLength(song, &opErr);
        env.ThrowOnError(opErr);
        return ticks;
    }
    return 0;
}


JNIEXPORT jint JNICALL
Java_sun_audio_HaeMidiSong_getMicrosecondLength(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        OPErr opErr;
        unsigned long ticks;
        ticks = GM_GetSongMicrosecondLength(song, &opErr);
        env.ThrowOnError(opErr);
        return ticks;
    }
    return 0;
}


// get the current playback position of a song in microseconds
JNIEXPORT jint JNICALL
Java_sun_audio_HaeMidiSong_getTickPosition(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
GM_Song*        const song = thiz.GetSongData();

    if (song)
    {
        return GM_SongTicks(song);
    }
    return 0;
}

#if HAE_STRINGS

JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoTitle(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_TITLE);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoPerformer(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_PERFORMED_BY);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoComposer(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_COMPOSER);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoCopyright(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_COPYRIGHT);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoPublisher(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_PUBLISHER_CONTACT);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoLicenseUse(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_USE_OF_LICENSE);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoLicenseTerm(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_LICENSE_TERM);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoLicensedTo(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_LICENSED_TO_URL);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoExpirationDate(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_EXPIRATION_DATE);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoComposerNotes(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_COMPOSER_NOTES);
}
JNIEXPORT jstring JNICALL
Java_sun_audio_HaeMidiSong_getInfoIndexNumber(JNIEnv* e, jobject thisObj)
{
    return GetInfoString(env, thisObj, I_INDEX_NUMBER);
}


////////////////////////////////////////////////// LOCAL FUNCTIONS:

static jstring GetInfoString(JNIEnv* e, jobject thisObj, SongInfo infoCode)
{
HaeEnv          env(e);
HaeMidiSong     thiz(env, thisObj);
SongResource*   const data = thiz.GetStringData();
unsigned long   size;
char            buffer[1024];

    if (data)
    {
        size = XGetSongInformationSize(data, XGetPtrSize((XPTR)data),
                                       infoCode);
        if (size >= 1024)
        {
            env.ThrowOnError(BAD_FILE);
            return NULL;
        }
        else
        {
            XGetSongInformation(data, XGetPtrSize((XPTR)data),
                                infoCode, buffer);
        }
    }
    else
    {
        size = 0;
    }
    buffer[size] = '\0';

    return JncString::New(env, buffer);
}

#endif // HAE_STRINGS

