/*
 * @(#)HaeMixer_Impl.cpp	1.24 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


#include "Hae_Impl.h"
#include "sun_audio_HaeMixer.h"
#include "HaePath.h"
#include "HaeMixer.h"
#if HAE_MIXER_FEEDBACK
#include "HaeMixerInfo.h"
#include "HaeMixerFrame.h"
#endif
#include "HaeMixer_constants.h"
#include "JncShortArray.h"


////////////////////////////////////////////////// VARIABLES:

//KARA7: deleted static class cache allocations


////////////////////////////////////////////////// LOCAL STATIC VARIABLES:

static XFILE        thePatchFile = NULL;    // current audio patch library


////////////////////////////////////////////////// OPERATIONS:

static void     AudioTaskCallbackProc(void* environment, unsigned long ticks);

JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_open(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeMixer        thiz(env, thisObj);
OPErr           error;

#if X_PLATFORM == X_WIN95
/*
    if (!thePatchFile)
    {
    static WIN95INIT    w95Init;
    int                 count;

        // This has to be done for Windows only, because the message system needs
        // a window handle to process messages.

        w95Init.cbSize = sizeof(WIN95INIT);

        // we're going to count to 100, and wait for a window to become active by
        // sleeping in this thread then checking to see if anything has changed.
        count = 100;
        while (--count >= 0)
        {
            w95Init.hwndOwner = GetForegroundWindow();
            if (w95Init.hwndOwner) break;

            w95Init.hwndOwner = GetActiveWindow();
            if (w95Init.hwndOwner) break;

            w95Init.hwndOwner = GetFocus();
            if (w95Init.hwndOwner) break;

            Sleep(10);  // sleep 10 milliseconds
        }

        ::GM_SetUserReference((long)&w95Init);
    }
*/
#endif

    TRACE0("Mixer_Impl: Starting up the engine.\n");
    error = ::GM_InitGeneralSound(e,
                                    env.TranslateQuality(thiz.GetQuality()),
                                    env.TranslateInterpolation(thiz.GetInterpolation()),
                                    env.TranslateModifiers(thiz.GetModifiers()),
                                    (short)thiz.GetMidiVoiceCount(),
                                    (short)thiz.GetVolumeDivisor(),
                                    (short)thiz.GetSoundVoiceCount());
#if X_PLATFORM == X_WIN95
    if (error == MEMORY_ERR) error = EXTRA_ERROR_PLATFORM;
#endif
    if (!env.ThrowOnError(error))
    {
        Java_sun_audio_HaeMixer_setReverbType(e, thisObj, thiz.GetReverbType());

#if X_PLATFORM == X_MACINTOSH
        #error MUST SPAWN JAVA THREAD TO RUN STREAMS
#else
        TRACE0("Mixer_Impl: Setting up audio task.\n");
        ::GM_SetAudioTask(AudioTaskCallbackProc);
#endif
    }
    else
    {
        TRACE0("Mixer_Impl: Error on startup.\n");
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_close(JNIEnv* e, jobject thisObj)
{
JncEnv          env(e);

    if (thePatchFile)
    {
        ::XFileClose(thePatchFile);
        thePatchFile = NULL;
    }

    ::GM_AudioStreamStopAll(e);
#if X_PLATFORM == X_MACINTOSH
    #error MUST STOP JAVA THREAD THAT RUN STREAMS
#else
    ::GM_SetAudioTask(NULL);
#endif
    ::GM_StopHardwareSoundManager(e);
    ::GM_FinisGeneralSound(e);
    JncClassCache::UncacheAll(env); //KARA7: changed this line
}
#if X_PLATFORM != X_MACINTOSH
static void AudioTaskCallbackProc(void* environment, unsigned long ticks)
{
//  TRACE0("HaeMixerImpl:AudioTaskCallbackProc - service streams.\n");
    ::GM_AudioStreamService(environment);
//  TRACE0("HaeMixerImpl:AudioTaskCallbackProc - service streams done.\n");
}
#endif

JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_setSoundbank(JNIEnv* e, jobject thisObj,
                                    jstring soundbankPathObj)
{
HaeEnv          env(e);
HaePath         path(env, soundbankPathObj);

    if (thePatchFile)
    {
        ::XFileClose(thePatchFile);
    }

    thePatchFile = path.OpenResourceFile();
    if (thePatchFile)
    {
        ::XFileUseThisResourceFile(thePatchFile);
    }
    else
    {
        env.ThrowOnError(BAD_FILE);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_setAudioModes(JNIEnv* e, jobject thisObj,
                                    jint q, jint t, jint am)
{
HaeEnv          env(e);
HaeMixer        thiz(env, thisObj);

    if (!env.ThrowOnError(::GM_ChangeAudioModes(e,
                                                env.TranslateQuality(q),
                                                env.TranslateInterpolation(t),
                                                env.TranslateModifiers(am))))
    {
        thiz.SetQuality(q);
        thiz.SetInterpolation(t);
        thiz.SetModifiers(am);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_setReverbType(JNIEnv* e, jobject thisObj, jint r)
{
HaeEnv          env(e);
HaeMixer        thiz(env, thisObj);

    thiz.SetReverbType(r);
    ::GM_SetReverbType((BOOL_FLAG)env.TranslateReverbType(r));
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_setVoiceCounts(JNIEnv* e, jobject thisObj,
                                     jint maxSongVoiceCount,
                                     jint maxSoundVoiceCount,
                                     jint volumeDivisor)
{
HaeEnv          env(e);
HaeMixer        thiz(env, thisObj);

    if (!env.ThrowOnError(GM_ChangeSystemVoices((short)maxSongVoiceCount,
                                                    (short)volumeDivisor,
                                                    (short)maxSoundVoiceCount)))
    {
        thiz.SetMidiVoiceCount(maxSongVoiceCount);
        thiz.SetSoundVoiceCount(maxSoundVoiceCount);
        thiz.SetVolumeDivisor(volumeDivisor);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_setMasterVolume(JNIEnv* e, jobject thisObj, jdouble volume)
{
    ::GM_SetMasterVolume((unsigned long)(volume * MAX_MASTER_VOLUME));
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_disengageAudio(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);

    TRACE0("MixerImpl: pausing the engine.\n");

// $$kk: start changes

    // the behaviour we want is just to disgengage from the device if we have it;
    // if we don't have it, we just return quietly.  the only time this call returns
    // an error code is when the audio is already disengaged.   so we're just going
    // to swollow the error here....

    if (::GM_PauseGeneralSound(e) == ALREADY_PAUSED)
    {
        TRACE0("MixerImpl: engine was already paused.\n");
    }
    else
    {
        TRACE0("MixerImpl: engine paused.\n");
    }
// $$kk: end changes
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_reengageAudio0(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
OPErr           opErr;

    TRACE0("MixerImpl: restarting the engine.\n");

    opErr = ::GM_ResumeGeneralSound(e);
    if (opErr == ALREADY_RESUMED)
    {
        TRACE0("MixerImpl: engine was already resumed.\n");
    }
    else
    {
        env.ThrowOnError(opErr);
    }

    TRACE0("MixerImpl: engine restarted.\n");
}


////////////////////////////////////////////////// DATA:

JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeMixer_getMasterVolume(JNIEnv* e, jobject thisObj)
{
    return (jdouble)::GM_GetMasterVolume() / MAX_MASTER_VOLUME;
}
JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeMixer_getTick(JNIEnv* e, jobject thisObj)
{
// $$kk: start changes
    //return ::GM_GetSyncTimeStamp();
    return ::GM_GetDeviceTimeStamp();
// $$kk: end changes

}

JNIEXPORT jlong JNICALL
Java_sun_audio_HaeMixer_getSamplesPlayed(JNIEnv* e, jobject thisObj)
{
    return (jlong)::GM_GetSamplesPlayed();
}



JNIEXPORT jint JNICALL
Java_sun_audio_HaeMixer_getAudioLatency(JNIEnv* e, jobject thisObj)
{
    return ::GM_GetSyncTimeStampQuantizedAhead() - ::XMicroseconds();
}

#if HAE_MIXER_FEEDBACK
JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_getInfo(JNIEnv* e, jobject thisObj, jobject targetObj)
{
HaeEnv          env(e);
HaeMixerInfo    target(env, targetObj);
GM_AudioInfo    status;

    ::GM_GetRealtimeAudioInformation(&status);
    target.SetActiveVoiceCount(status.voicesActive);

    {
    JncShortArray   array(env, target.GetInternalIndex());

        array.CopyIn(0, status.voice, MAX_VOICES);
    }

    {
    JncShortArray   array(env, target.GetInstrument());

        array.CopyIn(0, status.patch, MAX_VOICES);
    }

    {
    JncShortArray   array(env, target.GetMidiVolume());

        array.CopyIn(0, status.volume, MAX_VOICES);
    }

    {
    JncShortArray   array(env, target.GetScaledVolume());

        array.CopyIn(0, status.scaledVolume, MAX_VOICES);
    }

    {
    JncShortArray   array(env, target.GetChannel());

        array.CopyIn(0, status.channel, MAX_VOICES);
    }

    {
    JncShortArray   array(env, target.GetMidiNote());

        array.CopyIn(0, status.midiNote, MAX_VOICES);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeMixer_getFrame(JNIEnv* e, jobject thisObj, jobject targetObj)
{
HaeEnv          env(e);
HaeMixerFrame   target(env, targetObj);
JncShortArray   leftBuffer(env, target.GetLeftBuffer());
JncShortArray   rightBuffer(env, target.GetRightBuffer());
short*          const left = leftBuffer.Access();
short*          const right = rightBuffer.Access();

    if (left && right)
    {
        target.SetValidLength(GM_GetAudioSampleFrame(left, right));
    }
}
#endif


