/*
 * @(#)HaeRmfSong_Impl.cpp	1.7 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


#include "Hae_Impl.h"
#include "sun_audio_HaeMidiSong.h"
#include "sun_audio_HaeRmfSong.h"
#include "HaePath.h"
#include "HaeRmfSong.h"
#include "JncByteArray.h"


////////////////////////////////////////////////// VARIABLES:

//KARA7: deleted static class cache allocation


////////////////////////////////////////////////// FORWARDS:

static void HaeRmfSong_open(HaeEnv& env, jobject thisObj, XFILE fileRef);


////////////////////////////////////////////////// HaeRmfSong FUNCTIONS:

JNIEXPORT void JNICALL
Java_sun_audio_HaeRmfSong_open(JNIEnv* e, jobject thisObj, jbyteArray dataObj)
{
HaeEnv          env(e);
JncByteArray    data(env, dataObj);
jsize           dataBytes;
XFILE           fileRef;

    ::Java_sun_audio_HaeMidiSong_close(e, thisObj);

    dataBytes = data.GetLength();
    if (dataBytes > 0)
    {
    XPTR                    const dataCopy = ::XNewPtr(dataBytes);

        if (dataCopy)
        {
            data.CopyOut((jbyte*)dataCopy, 0, dataBytes);
            fileRef = ::XFileOpenResourceFromMemory(dataCopy, dataBytes, JNI_TRUE);
            if (fileRef)
            {
                ::HaeRmfSong_open(env, thisObj, fileRef);
                // fileRef closed inside HaeRmfSong_open()
            }
            else
            {
                env.ThrowOnError(BAD_FILE);    //?? Is this right?
            }
        }
        else
        {
            env.ThrowOnError(MEMORY_ERR);
        }
    }
    else
    {
        ERROR0("HaeRmfSong.open(byte[]) called with zero-length array\n");
    }
}

#if HAE_FILES
JNIEXPORT void JNICALL
Java_sun_audio_HaeRmfSong_openFile(JNIEnv* e, jobject thisObj, jstring pathObj)
{
HaeEnv          env(e);
HaePath         path(env, pathObj);
XFILE           fileRef;

    ::Java_sun_audio_HaeMidiSong_close(env, thisObj);

    fileRef = path.OpenResourceFile();
    if (fileRef)
    {
        ::HaeRmfSong_open(env, thisObj, fileRef);
        // fileRef closed inside HaeRmfSong_open()
    }
    else
    {
        env.ThrowOnError(BAD_FILE);
    }
}
#endif // HAE_FILES


////////////////////////////////////////////////// LOCAL FUNCTIONS:

void HaeRmfSong_open(HaeEnv& env, jobject thisObj, XFILE fileRef)
{
long            id;
long            size;
SongResource*   xSong;
OPErr           opErr;

    // look for first song. RMF files only contain one SONG resource
    xSong = (SongResource*)::XGetIndexedResource(ID_SONG, &id, 0, NULL, &size);
    if (xSong)
    {
    HaeMidiSong     thiz(env, thisObj);
    GM_Song*        song;

        thiz.SetStringData(xSong);

        song = ::GM_LoadSong((short)id, xSong, NULL, 0,
                             NULL,// no callback
                             TRUE,// load instruments
                             TRUE,// ignore bad instruments
                             &opErr);
        if (song)
        {
            song->ignoreBadInstruments = FALSE;
            song->disposeSongDataWhenDone = TRUE;   // dispose of midi data
            ::GM_SetSongLoopFlag(song, FALSE);      // don't loop song

            song->userReference = id;
            thiz.SetSongData(song);
            opErr = NO_ERR;             //??STEVE: Why is this necessary?
        }
    }
    else
    {
        opErr = BAD_FILE;
    }

    ::XFileClose(fileRef);
    env.ThrowOnError(opErr);
}

