
/*
 * @(#)HaeThread_Impl.cpp	1.9 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#if JAVA_THREAD

#include "Hae_Impl.h"
#include "sun_audio_HaeThread.h"
#include "JncObject.h"
#include "JncGlobalRef.h"
#include "HAE_API.h"




////////////////////////////////////////////////// CLASSES:

class HaeThread : public JncObject
{
PUBLIC_EXISTENCE:
    HaeThread(JncEnv& env, jobject obj)
    : JncObject(env, obj)
    {}

PUBLIC_OPERATIONS:
    void        Start()
    { CallVoidMethod("start"); }
    void        UnPause()
    { CallVoidMethod("unPause"); }

	// $$kk: 05.08.98: added this method as part of fix for bug #4135246.
    void        Pause()
    { CallVoidMethod("pause"); }

PUBLIC_DATA:

};


////////////////////////////////////////////////// STATIC VARIABLES:

STA char        const theClassName[] = JAVA_PACKAGE_NAME"/HaeThread";

STA JncGlobalRef        theFrameThread;
STA HAE_FrameThreadProc theFrameProc = NULL;


////////////////////////////////////////////////// NATIVE METHODS:

// Called by HaeThread.run()
// Calls the proc passed into HAE_CreateFrameThread()
JNIEXPORT void JNICALL Java_sun_audio_HaeThread_runNative(JNIEnv* e, jobject thisObj)
{
#if JNC_DEBUG
JncEnv          env(e);
HaeThread       thread(env, thisObj);

    ASSERT(thread.IsSameObject(theFrameThread.GetJObject()));
#endif

    TRACE0("HaeThread_Impl: runNative() called\n");
    if (theFrameProc)
    {
        (*theFrameProc)(e);
    }
    else
    {
        ASSERT(FALSE);
    }
    TRACE0("HaeThread_Impl: runNative() returning\n");
}


////////////////////////////////////////////////// THREAD-API FUNCTIONS:
extern "C"
{

// The frame thread is created and started only once.
// If it already exists, unPause() is called which causes run() to exit
// its wait() loop.
int HAE_CreateFrameThread(void* context, HAE_FrameThreadProc proc)
{
    if (proc)
    {
    JncEnv          env((JNIEnv*)context);

        if (theFrameThread.IsSet())
        {   // thread has already been created, simply un-pause it
        HaeThread       thread(env, theFrameThread.GetJObject());

            theFrameProc = proc;
            TRACE0("HAE_CreateFrameThread(): calling unPause()\n");
            thread.UnPause();
            return 0;
        }
        else
        {   // create and start thread
        JncClassBase    clazz(env, theClassName);
        HaeThread       thread(env, clazz.NewObject());

            thread.SetDeleteLocal(JNI_TRUE);
            if (thread.IsValid())
            {
                if (theFrameThread.Set(env, thread.GetJObject()))
                {
                    theFrameProc = proc;
                    TRACE0("HAE_CreateFrameThread(): calling start()\n");
                    thread.Start();
                    return 0;
                }
            }
        }
    }
    else
    {
        ASSERT(FALSE);
    }
    return -1;
}

// Called by HAE_ReleaseAudioCard() where global variables are set to cause
// (*theFrameProc)() and runNative() to return.
// run() will then call wait() in a loop, to be interrupted by
// a call to unPause().
// Don't actually destroy the frame thread, just prevent
// runNative() doing anything until HAE_CreateFrameThread() is called.
int HAE_DestroyFrameThread(void* context)
{
    TRACE0("HAE_DestroyFrameThread() called\n");
    if (!theFrameProc)
    {
        ERROR0("HAE_DestroyFrameThread() when thread not started\n");
    }

	// $$kk: 05.08.98: added this block as part of fix for bug #4135246.
	{
		JncEnv          env((JNIEnv*)context);
		HaeThread       thread(env, theFrameThread.GetJObject());
		thread.Pause();
	}

	theFrameProc = NULL;
    return 0;
}

// Make the frame thread sleep for the given number of milliseconds
int HAE_SleepFrameThread(void* context, long msec)
{
JncEnv      env((JNIEnv*)context);
JncStatic   clazz(env, theClassName);

    clazz.CallStaticVoidMethod("sleep", "(J)V", (jlong)msec);
    env.ClearPendingException();
    return 0;
}

}


#endif  // JAVA_THREAD


