/*
 * @(#)HaeWaveDirect_Impl.cpp	1.8 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


#include "Hae_Impl.h"
#include "sun_audio_HaeWaveDirect.h"
#include "HaeWaveDirect.h"

#if HAE_DIRECT_WAVES


////////////////////////////////////////////////// VARIABLES:

//KARA7: deleted static class cache allocation
static char     const gmWaveformName[] = "m_gmWaveform";


////////////////////////////////////////////////// HaeWaveDirect MEMBERS:

void HaeWaveDirect::SetGmWaveform(GM_Waveform* data)
{
    SetPointerField(gmWaveformName, data);
}
GM_Waveform* HaeWaveDirect::GetGmWaveform()
{
    return (GM_Waveform*)GetPointerField(gmWaveformName);
}


////////////////////////////////////////////////// HaeWaveDirect OPERATIONS:

// $$kk: 01.14.98: added numLoops param
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_setLoop(JNIEnv* e, jobject thisObj,
                                    jint startFrame,
                                    jint endFrame,
									jint numLoops)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);
GM_Waveform*    const data = thiz.GetGmWaveform();

    if (data)
    {
        data->startLoop = startFrame;
        data->endLoop = endFrame;
		data->numLoops = numLoops;
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_close(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);
GM_Waveform*    const data = thiz.GetGmWaveform();

    if (data)
    {
        Java_sun_audio_HaeWaveDirect_stop(env, thisObj, FALSE);

        if (data->theWaveform)
        {
            XDisposePtr(data->theWaveform);
        }
        GM_FreeWaveform(data);
        thiz.SetGmWaveform(NULL);
    }
    thiz.SetWaveId(0);
}

// $$kk: 01.14.98: added offset param
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_privateStart(JNIEnv* e, jobject thisObj,
                                        jdouble volume,
                                        jdouble pan,
                                        jboolean reverb,
										jlong offset)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);
GM_Waveform*    const data = thiz.GetGmWaveform();
OPErr           opErr;

    if (data)
    {
        if (Java_sun_audio_HaeWaveDirect_isPlaying(env, thisObj))
        {
            Java_sun_audio_HaeWaveDirect_stop(env, thisObj, FALSE);
        }

        if (volume)
        {
        long            id;

            id = ::GM_BeginSampleFromInfo(data, 0,
                                            DOUBLE_TO_VOLUME(volume), DOUBLE_TO_PAN(pan),
                                            NULL, NULL, (INT32)offset);
            if (id != -1L)
            {
                thiz.SetWaveId(id);
                Java_sun_audio_HaeWaveDirect_privateSetReverb(env, thisObj, reverb);
                opErr = NO_ERR;
            }
            else
            {
                opErr = NO_FREE_VOICES;
            }
        }
        else
        {
            opErr = GENERAL_BAD;
        }
    }
    else
    {
        opErr = BAD_FILE;
    }
    env.ThrowOnError(opErr);
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_stop(JNIEnv* e, jobject thisObj,
                                jboolean withFade)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);
long            const id = thiz.GetWaveId();

    if (withFade)
    {
        GM_SetSampleFadeRate(id, DOUBLE_TO_FIXED(2.2),
                             0, GM_GetSampleVolume(id), TRUE);
    }
    else
    {
        GM_EndSample(id);
    }
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_setVolume(JNIEnv* e, jobject thisObj,
                                        jdouble volume)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    GM_ChangeSampleVolume(thiz.GetWaveId(), DOUBLE_TO_VOLUME(volume));
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_setPan(JNIEnv* e, jobject thisObj,
                                    jdouble pan)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    GM_ChangeSampleStereoPosition(thiz.GetWaveId(), DOUBLE_TO_PAN(pan));
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_privateSetSampleRate(JNIEnv* e, jobject thisObj,
                                                jdouble sampleRate)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    GM_ChangeSamplePitch(thiz.GetWaveId(), DOUBLE_TO_FIXED(sampleRate));
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveDirect_privateSetReverb(JNIEnv* e, jobject thisObj,
                                            jboolean useReverb)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    GM_ChangeSampleReverb(thiz.GetWaveId(), (BOOL_FLAG)useReverb);
}


////////////////////////////////////////////////// HaeWaveDirect DATA:

JNIEXPORT jboolean JNICALL
Java_sun_audio_HaeWaveDirect_isPlaying(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    return GM_IsSoundDone(thiz.GetWaveId()) ? JNI_FALSE
                                            : JNI_TRUE;
}

JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeWaveDirect_getSampleRate(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    return FIXED_TO_DOUBLE(GM_GetSamplePitch(thiz.GetWaveId()));
}
JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeWaveDirect_getVolume(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    return VOLUME_TO_DOUBLE(GM_GetSampleVolume(thiz.GetWaveId()));
}
JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeWaveDirect_getPan(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveDirect   thiz(env, thisObj);

    return PAN_TO_DOUBLE(GM_GetSampleStereoPosition(thiz.GetWaveId()));
}


#endif // HAE_DIRECT_WAVES



