/*
 * @(#)HaeWaveStream_Impl.cpp	1.33 98/09/10
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// #define DEBUG

#include <malloc.h>
#include <memory.h> // for ::memset??
#include "Hae_Impl.h"
#include "sun_audio_HaeWaveStream.h"
#include "HaeWaveStream.h"
#include "HaeWaveStreamBuffer.h"
#include "JncByteArray.h"
#include "JncShortArray.h"



////////////////////////////////////////////////// SWITCHES:


////////////////////////////////////////////////// PRIVATE CLASSES:

//KARA7: This class no longer stores an HaeEnv that had to be updated all
// the time for the current thread.  m_env and m_this have been replaced
// with m_globalThis.

class CallbackStuff
{
PUBLIC_EXISTENCE:
    CallbackStuff(HaeEnv& env, jobject thisObj)
    : m_globalThis(env, thisObj)
    {
        m_bytesPerFrame = 0;
        m_opened = JNI_FALSE;
    }
    ~CallbackStuff()
    {
        if (m_globalThis.IsSet())
        {
            ERROR0("In ~CallbackStuff()...\n");
            // more error message will be displayed by ~m_globalThis
        }
    }

PUBLIC_OPERATIONS:
STA CallbackStuff*  New(HaeEnv& env, jobject thisObj)
    {
        return new CallbackStuff(env, thisObj);
    }
    void            Delete(HaeEnv& env) //KARA7: added this function
    {
        m_globalThis.Unset(env);
        delete this;
    }

PUBLIC_DATA:
    JncGlobalRef    m_globalThis;   //KARA7: changed fields
    jint            m_bytesPerFrame;
    jboolean        m_opened;

PRIVATE_OPERATIONS:
STA void*           operator new(size_t nSize)
    { return ::malloc(nSize); }
STA void            operator delete(void* p)
    { ::free(p); }

PRIVATE_DATA:
};


////////////////////////////////////////////////// VARIABLES:

//KARA7: deleted static class cache allocations


////////////////////////////////////////////////// HaeWaveStream OPERATIONS:

static OPErr    HaeWaveStreamObjectProc(void* environment,
                                        GM_StreamMessage message,
                                        GM_StreamData* gmData);
static OPErr    HaeWaveStreamGetData(GM_StreamData* gmData,
                                        HaeWaveStream& thiz,    //KARA7: added
                                        CallbackStuff* stuff);

JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_nativeOpen(JNIEnv* e, jobject thisObj,
                                        jdouble sampleRate,
                                        jint bitsPerSample,
                                        jint channelsPerFrame,
                                        jint bytesPerFrame,
                                        jint bufferBytes)
{
HaeEnv              env(e);
HaeWaveStream       thiz(env, thisObj); //KARA7: added this line
CallbackStuff*      const stuff = CallbackStuff::New(env, thisObj);

    if (!stuff)
    {
        ERROR0("Failure to allocate thiz in HaeWaveStream.nativeOpen()\n");
        env.ThrowOnError(MEMORY_ERR);
    }
    else if (!stuff->m_globalThis.IsSet())
    {
        stuff->Delete(env);
        ERROR0("Failure to allocate thisObj global ref in HaeWaveStream.nativeOpen()\n");
        env.ThrowOnError(MEMORY_ERR);
    }
    else
    {
    int         id;

        stuff->m_bytesPerFrame = bytesPerFrame;
//        TRACE0("Stream_Imp: Setup stream.\n");
        id = ::GM_AudioStreamSetup(e, (long)stuff, HaeWaveStreamObjectProc,
                                    bufferBytes,
                                    DOUBLE_TO_FIXED(sampleRate),
                                    (char)bitsPerSample, (char)channelsPerFrame);
// $$kk: 10.20.97
// now need id to throw exception (changed in GenAudioStreams.c)
        if (id)
        {   //KARA7: Why is the above test necessary?
            thiz.SetWaveId(id); //KARA7: changed this line
        }

//        TRACE0("Stream_Imp: Setupstream done.\n");
        Java_sun_audio_HaeWaveStream_throwOnLastError(e, thisObj);
        if (!env.IsPendingException())
        {
            stuff->m_opened = JNI_TRUE;
        }
        else
        {
            ERROR0("Stream_Imp: Setup failed.\n");  //KARA8: changed this to error
            // stuff has been deleted!
        }
//      TRACE0("Stream_Imp: stream open.\n");
    }
}
static OPErr HaeWaveStreamObjectProc(void* environment,
                                     GM_StreamMessage message, GM_StreamData* gmData)
{
CallbackStuff*  const stuff = (CallbackStuff*)gmData->userReference;
HaeEnv              env((JNIEnv*)environment);  //KARA7: added this line
HaeWaveStream       thiz(env, stuff->m_globalThis.GetJObject());    //KARA7: add this line


    if (!stuff)
    {
        return GENERAL_BAD;
    }

    if (message == STREAM_CREATE)
    {
    jint    const byteCount = gmData->dataLength * stuff->m_bytesPerFrame;

//      TRACE0("Stream_Impl: Stream Create callback.\n");
        gmData->pData = ::XNewPtr(byteCount);
        if (!gmData->pData)
        {
            ERROR1("Failure to allocate %d bytes in HaeWaveStreamObjectProc()\n", byteCount);
            return MEMORY_ERR;
        }
    }
    else if (message == STREAM_DESTROY)
    {
//      TRACE0("HaeWaveStreamImpl: Callback destroy.\n");
        if (gmData->pData)
        {
            ::XDisposePtr(gmData->pData);
        }
        if (stuff->m_opened)
        {
            thiz.CallStreamingFinished();  // STREAM_DESTROY only called once!
			
			// fix for bug #4172570: Animator applet loses background sound.
			// after this callback, the stream can be freed.  the HaeWaveStream must not
			// retain the old WaveId.  it can be assigned to another stream, and then this
			// HaeWaveStream may erroneously control it.
			thiz.SetWaveId(NULL);
        
		}
//      TRACE0("HaeWaveStreamImpl: Destroy - deleting state.\n");
        stuff->Delete(env); //KARA7: changed this line
//      TRACE0("HaeWaveStreamImpl: Destroy " "- done delete.\n");
    }
    else if (message == STREAM_GET_DATA)
    {
        return HaeWaveStreamGetData(gmData, thiz, stuff);   //KARA7: changed this line
    }

    else
    {
        ERROR1("Bad message code in HaeWaveStreamObjectProc() : %d\n", message);
        return GENERAL_BAD;
    }
    return NO_ERR;
}
static OPErr HaeWaveStreamGetData(GM_StreamData* gmData,
                                    HaeWaveStream& thiz,    //KARA7: added this line
                                    CallbackStuff* stuff)
{
jobject         bufferObj;
jboolean        doContinue;

//  TRACE0("StreamImpl: Calling the getwavedata callback.\n");
    doContinue = thiz.CallPrivateGetWavedata(gmData->dataLength);
//  TRACE1("StreamImpl: done with getwavedata callback: return==%d\n", doContinue);
    if (thiz.GetEnv().IsPendingException())
    {
        ERROR0("Exception thrown in HaeWaveStream.getWavedata()\n");
        thiz.GetEnv().ClearPendingException();
        return GENERAL_BAD;
    }

    bufferObj = thiz.GetBuffer();
    if (!bufferObj)
    {
        ERROR0("m_buffer==NULL in HaeWaveStreamObjectProc()");
        return GENERAL_BAD;
    }
    else
    {
    HaeWaveStreamBuffer buffer(thiz.GetEnv(), bufferObj);   //KARA7: changed this line
    unsigned int        const frameCount  = buffer.GetFrameCount();
    jint                const byteCount = frameCount * stuff->m_bytesPerFrame;

        buffer.SetDeleteLocal();    //KARA6: changed this line
        if (frameCount > gmData->dataLength)
        {
            ERROR2("%d frames returned from HaeWaveStream.getWaveData(), maximum = %d\n",
                    frameCount, gmData->dataLength);
            return STREAM_STOP_PLAY;
        }

        gmData->dataLength = frameCount;
        if (frameCount != 0)
        {
            jbyteArray          byteArray;
            jshortArray         shortArray;

            if (byteArray = buffer.GetByteData())
            {
            JncByteArray        bufferData(thiz.GetEnv(), byteArray);   //KARA7: changed this line

                bufferData.SetDeleteLocal();    //KARA7: changed this line
                bufferData.CopyOut((jbyte*)gmData->pData, 0, byteCount);
            }
            else if (shortArray = buffer.GetShortData())
            {
            JncShortArray       bufferData(thiz.GetEnv(), shortArray);  //KARA7: changed this line

                bufferData.SetDeleteLocal();    //KARA7: changed this line
                bufferData.CopyOut((jshort*)gmData->pData, 0, byteCount / 2);
            }
            else
            {
                ERROR0("HaeWaveStreamBuffer returned with no array");
            }
        }

        if (!doContinue)
        {
            return STREAM_STOP_PLAY;
        }
        return NO_ERR;
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_throwOnLastError(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);
jint            const id = thiz.GetWaveId();

//  TRACE0("Stream_Impl: Maybe throwing an exception.\n");
    env.ThrowOnError(GM_AudioStreamError(id));
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_privateStart(JNIEnv* e, jobject thisObj,
                                        jdouble volume, jdouble pan, jboolean useReverb)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);
jint            const id = thiz.GetWaveId();

    if (id != NULL)
    {
        ::GM_AudioStreamSetVolume(id, DOUBLE_TO_VOLUME(volume));
        ::GM_AudioStreamSetStereoPosition(id, DOUBLE_TO_PAN(pan));
        ::GM_AudioStreamReverb(id, (BOOL_FLAG)useReverb);
        ::GM_AudioStreamStart(id);
    }
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_stop(JNIEnv* e, jobject thisObj, jboolean withFade)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);
jint            const id = thiz.GetWaveId();

    if (id != NULL)
    {
        if (withFade && ::GM_IsAudioStreamPlaying(id))
        {
        short       streamVolume;

            streamVolume = ::GM_AudioStreamGetVolume(id);
            ::GM_SetAudioStreamFadeRate(id, DOUBLE_TO_FIXED(2.2), 0, streamVolume, TRUE);
            // We must not call this if the stream is out of data.
            // It sets a flag which will only be seen by the engine if the stream is playing.
        }
        else
        {
            ::GM_AudioStreamStop(e, id);
        }
        thiz.SetWaveId(NULL);   // this will let us avoid calling stop on the
                                // stream twice, which causes a problem if the
                                // memory has been freed and then given to another
                                // stream, which has the same id and can get this stop
                                // message....
    }
}


JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_setVolume(JNIEnv* e, jobject thisObj, jdouble volume)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    ::GM_AudioStreamSetVolume(thiz.GetWaveId(), DOUBLE_TO_VOLUME(volume));
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_setPan(JNIEnv* e, jobject thisObj, jdouble pan)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    ::GM_AudioStreamSetStereoPosition(thiz.GetWaveId(), DOUBLE_TO_PAN(pan));
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_privateSetSampleRate(JNIEnv* e, jobject thisObj, jdouble sampleRate)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    ::GM_AudioStreamSetRate(thiz.GetWaveId(), DOUBLE_TO_FIXED(sampleRate));
}
JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_privateSetReverb(JNIEnv* e, jobject thisObj, jboolean useReverb)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    ::GM_AudioStreamReverb(thiz.GetWaveId(), (BOOL_FLAG)useReverb);
}

JNIEXPORT jlong JNICALL
Java_sun_audio_HaeWaveStream_privateGetSamplesPlayed(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    return (long)::GM_AudioStreamGetSamplesPlayed(thiz.GetWaveId());
}




////////////////////////////////////////////////// HaeWaveStream DATA:

JNIEXPORT jboolean JNICALL
Java_sun_audio_HaeWaveStream_isPlaying(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    return ::GM_IsAudioStreamPlaying(thiz.GetWaveId()) ? JNI_TRUE
                                                       : JNI_FALSE;
}

JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeWaveStream_getSampleRate(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    return FIXED_TO_DOUBLE(GM_AudioStreamGetRate(thiz.GetWaveId()));
}
JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeWaveStream_getVolume(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    return VOLUME_TO_DOUBLE(GM_AudioStreamGetVolume(thiz.GetWaveId()));
}
JNIEXPORT jdouble JNICALL
Java_sun_audio_HaeWaveStream_getPan(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    return PAN_TO_DOUBLE(GM_AudioStreamGetStereoPosition(thiz.GetWaveId()));
}


JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_drain(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    ::GM_AudioStreamDrain(thiz.GetWaveId());
}


JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_flush(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    GM_AudioStreamFlush(thiz.GetWaveId());
}

JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_pause(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    GM_AudioStreamPause(thiz.GetWaveId());
}


JNIEXPORT void JNICALL
Java_sun_audio_HaeWaveStream_resume(JNIEnv* e, jobject thisObj)
{
HaeEnv          env(e);
HaeWaveStream   thiz(env, thisObj);

    GM_AudioStreamResume(thiz.GetWaveId());
}





