/*
 * @(#)Jnc.h	1.8 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// Jnc.h

#ifndef Jnc_H
#define Jnc_H   1

#include <jni.h>


////////////////////////////////////////////////// DEBUGGING MACROS

#ifndef JNC_DEBUG
#define JNC_DEBUG   0
#endif

#ifndef JNC_TRACE
#define JNC_TRACE   0
#endif

#ifndef JNC_CACHE_IDS
#define JNC_CACHE_IDS   0
#endif

#ifndef JNC_CACHE_WITH_POINTERS
#define JNC_CACHE_WITH_POINTERS 0
#endif

#if JNC_DEBUG
#define ERROR0(string)              fprintf(stderr, (string))
#define ERROR1(string, p1)          fprintf(stderr, (string), (p1))
#define ERROR2(string, p1, p2)      fprintf(stderr, (string), (p1), (p2))
#define ERROR3(string, p1, p2, p3)  fprintf(stderr, (string), (p1), (p2), (p3))
#else
#define ERROR0(string)
#define ERROR1(string, p1)
#define ERROR2(string, p1, p2)
#define ERROR3(string, p1, p2, p3)
#endif

#if JNC_TRACE
static int jnc_trace = 1;
#define TRACE0(string)              if (jnc_trace) fprintf(stdout, (string))
#define TRACE1(string, p1)          if (jnc_trace) fprintf(stdout, (string), (p1))
#define TRACE2(string, p1, p2)      if (jnc_trace) fprintf(stdout, (string), (p1), (p2))
#define TRACE3(string, p1, p2, p3)  if (jnc_trace) fprintf(stdout, (string), (p1), (p2), (p3))
#else
#define TRACE0(string)
#define TRACE1(string, p1)
#define TRACE2(string, p1, p2)
#define TRACE3(string, p1, p2, p3)
#endif

#ifdef _DEBUG   //KARA8:  added this groovy stuff

#if X_PLATFORM == X_WIN95
#pragma optimize("gle", off)
inline void DEBUGGER()
{
    _asm int 3;
}
#pragma optimize("", on)
#else
inline void DEBUGGER()
{
    fprintf(stderr, "ASSERTION FAILED\n");
}
#endif

#define ASSERT(exp) if (!(exp)) DEBUGGER();
#define VERIFY(exp) if (!(exp)) DEBUGGER();

#else

inline void DEBUGGER() {}
#define ASSERT(exp) ((void)0)
#define VERIFY(exp) ((void)(exp))

#endif


////////////////////////////////////////////////// CODE-NEATNESS MACROS

// substitutes less than 3 letters so that they fit inside
// the left tab in class definitions
#define STA     static
#define STA_I               // implementation of a static field or method
#define VIR     virtual
#define VIR_I               // implementation of a virtual method
#define INL     inline

// big fat names for sections in class definitions:
#define PUBLIC_TYPES            public
#define PUBLIC_EXISTENCE        public
#define PUBLIC_OPERATIONS       public
#define PUBLIC_NOTIFICATIONS    public
#define PUBLIC_DATA             public
#define PROTECTED_TYPES         protected
#define PROTECTED_EXISTENCE     protected
#define PROTECTED_OPERATIONS    protected
#define PROTECTED_NOTIFICATIONS protected
#define PROTECTED_DATA          protected
#define PRIVATE_TYPES           private
#define PRIVATE_EXISTENCE       private
#define PRIVATE_OPERATIONS      private
#define PRIVATE_NOTIFICATIONS   private
#define PRIVATE_DATA            private


////////////////////////////////////////////////// HANDY TYPES:

typedef char const*     ConStr;


////////////////////////////////////////////////// HANDY FUNCTIONS:

#define MIN(n1, n2)     ((n1) <= (n2) ? (n1) : (n2))
#define MAX(n1, n2)     ((n1) >= (n2) ? (n1) : (n2))

INL int Min(int n1, int n2)
{
    return MIN(n1, n2);
}
INL int Max(int n1, int n2)
{
    return MAX(n1, n2);
}

INL unsigned short Abs(short n)
{
    return (unsigned short)((n < 0) ? -n : n);
}
INL unsigned int Abs(int n)
{
    return (n < 0) ? -n : n;
}


#endif
