/*
 * @(#)JncClass.cpp	1.2 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncClass.c


#include <string.h>
#include "JncEnv.h"
#include "JncClass.h"
#include "JncString.h"


////////////////////////////////////////////////// EXISTENCE:


////////////////////////////////////////////////// OPERATIONS:

jobject JncClass::Call_newInstance()
{
char        signature[512];

    signature[0] = '(';
    signature[1] = ')';
    GetSignature(signature + 2, sizeof(signature) / sizeof(char) - 3);

    return CallObjectMethod("newInstance", signature);
}


////////////////////////////////////////////////// DATA:

jsize JncClass::GetSignature(char* buffer, jsize maxLength)
{
char        className[512];
char const* src;
char*       dst;
int         bracketCount;
int         letterCount;

    GetName(className, ::Min(sizeof(className) / sizeof(char) - 1, maxLength - 2));
    src = className;
    dst = buffer;

    bracketCount = 0;
    letterCount = 0;

    while (JNI_TRUE)
    {
        switch (*src)
        {
        case '[' :
            bracketCount++;
            *dst++ = *src;
            break;
        case 'Z' :
        case 'B' :
        case 'C' :
        case 'S' :
        case 'I' :
        case 'J' :
        case 'F' :
        case 'D' :
            if ((*(src + 1) == '\0') && (letterCount == 0) && (bracketCount > 0))
            {
            // have to check whether we have a primitive type or a single-letter class name
            jclass          superclass;

                superclass = GetJObject();
                while (superclass && (--bracketCount >= 0))
                {
                    superclass = GetSuperclass(GetEnv(), superclass);
                }
                if (superclass)
                {
                char            superclassName[512];

                    GetClassName(GetEnv(), superclass,
                                    superclassName, sizeof(superclassName) / sizeof(char) - 1);
                    if (::strcmp(src, superclassName) != 0)
                    {
                    // the class is an n-dimensional array of primitive elements
                        *dst++ = *src;
                        *dst = '\0';
                        return dst - buffer;
                    }
                }
            }
        default :
            if (letterCount == 0)
            {
            // the class is an n-dimensional array of objects
                *dst++ = 'L';
            }
            letterCount++;
            *dst++ = *src;
            break;
        case '.' :
            // replace package-delimiting dots with slashes
            *dst++ = '/';
            break;
        case '\0' :
            *dst++ = ';';
            *dst = '\0';
            return dst - buffer;
        }
        src++;
    }
}

jsize JncClass::GetName(char* buffer, jsize maxLength)
{
JncString   string(m_env, Call_getName());

    if (string.IsValid())
    {
        string.SetDeleteLocal();
#if JNC_DEBUG
        if (string.GetLength() > maxLength)
        {
            ERROR2("ERROR: JncClass{}'s name (\"%s\") too long for %d-char buffer\n",
                    string.AccessUTF(), maxLength);
        }
#endif
        return string.CopyOut(buffer, 0, maxLength);
    }
    *buffer = '\0';
    return 0;
}


