/*
 * @(#)JncClassBase.h	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


/****************************************************************************
 *  JncClassBase.h
 *
 * JncClassBase{}
 * It provides methods for getting JNI field and method IDs.
 *
 ****************************************************************************/

#ifndef JncClassBase_H
#define JncClassBase_H  1

#include "JncClassCache.h"


////////////////////////////////////////////////// FORWARDS:

class JncEnv;


////////////////////////////////////////////////// CLASS:

class JncClassBase
{
PUBLIC_EXISTENCE:
    JncClassBase(JncEnv& env, jclass clazz);
    JncClassBase(JncEnv& env, ConStr className);
VIR ~JncClassBase();

PUBLIC_OPERATIONS:
    jobject         NewObject(ConStr signature, ...);
    jobject         NewObject()
    { return NewObject(m_sigReturnsVoid); }

PUBLIC_DATA:
    JncEnv&         GetEnv() const
    { return m_env; }
    jclass          GetClass();
    jclass          GetSuperclass()
    { return GetSuperclass(m_env, GetClass()); }
    jsize           GetClassName(char* buffer, jsize maxLength);
    jsize           GetClassSignature(char* buffer, jsize maxLength);

STA jclass          GetClass(JncEnv& env, ConStr className);
STA jclass          GetClass(JncEnv& env, jobject obj);
STA jclass          GetSuperclass(JncEnv& env, jclass clazz);
STA jsize           GetClassName(JncEnv& env, jclass clazz, char* buffer, jsize maxLength);
STA jsize           GetClassSignature(JncEnv& env, jclass clazz, char* buffer, jsize maxLength);

    jfieldID        GetFieldID(ConStr name, ConStr signature);
    jfieldID        GetStaticFieldID(ConStr name, ConStr signature);
    jmethodID       GetMethodID(ConStr name, ConStr signature);
    jmethodID       GetStaticMethodID(ConStr name, ConStr signature);
    jmethodID       GetConstructorID(ConStr signature);

STA const char      m_sigBoolean[];
STA const char      m_sigChar[];
STA const char      m_sigByte[];
STA const char      m_sigShort[];
STA const char      m_sigInt[];
STA const char      m_sigLong[];
STA const char      m_sigFloat[];
STA const char      m_sigDouble[];
STA const char      m_sigString[];
STA const char      m_sigBooleanArray[];
STA const char      m_sigCharArray[];
STA const char      m_sigByteArray[];
STA const char      m_sigShortArray[];
STA const char      m_sigIntArray[];
STA const char      m_sigLongArray[];
STA const char      m_sigFloatArray[];
STA const char      m_sigDoubleArray[];

STA const char      m_sigReturnsVoid[];
STA const char      m_sigReturnsBoolean[];
STA const char      m_sigReturnsChar[];
STA const char      m_sigReturnsByte[];
STA const char      m_sigReturnsShort[];
STA const char      m_sigReturnsInt[];
STA const char      m_sigReturnsLong[];
STA const char      m_sigReturnsFloat[];
STA const char      m_sigReturnsDouble[];
STA const char      m_sigReturnsString[];
STA const char      m_sigReturnsBooleanArray[];
STA const char      m_sigReturnsCharArray[];
STA const char      m_sigReturnsByteArray[];
STA const char      m_sigReturnsShortArray[];
STA const char      m_sigReturnsIntArray[];
STA const char      m_sigReturnsLongArray[];
STA const char      m_sigReturnsFloatArray[];
STA const char      m_sigReturnsDoubleArray[];

PROTECTED_EXISTENCE:
    JncClassBase(JncEnv& env, jobject obj); // for JncObject{}

PROTECTED_DATA:
    JncEnv&         m_env;
    jobject         const m_jobj;   // It's kind of a hack for this to be here, but it's for a good cause

PRIVATE_DATA:
#if JNC_CACHE_IDS
    JncClassCache*  GetCache();
    JncClassCache*  m_cache;
#else
    jclass          m_class;
    jboolean        m_deleteClass;
#endif
};


////////////////////////////////////////////////// INLINE DATA:


#endif
