/*
 * @(#)JncClassCache.cpp	1.8 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncClassCache.c

#include "Jnc.h"
#if JNC_CACHE_IDS

#include <string.h>     // for ::strcmp()
#include "JncClassCache.h"
#include "JncClassBase.h"
#include "JncID.h"


////////////////////////////////////////////////// PRIVATE CLASS:

class JncClassCacheList
{
PUBLIC_EXISTENCE:
    JncClassCacheList()
    { m_head = NULL; }
    ~JncClassCacheList()
    {
        if (m_head)
        {
            ERROR0("ERROR: JncClassCache::UncacheAll() must be called.\n");
        }
    }

PUBLIC_DATA:
    JncClassCache*      m_head;
};


////////////////////////////////////////////////// VARIABLES:

JncClassCacheList   theCacheList;

STA char            theCacheSemaphore = 0;


////////////////////////////////////////////////// EXISTENCE:

JncClassCache::JncClassCache(JncEnv& env, jclass clazz)
: m_globalClass(env, clazz)
{
    m_IDList = NULL;

    m_next = theCacheList.m_head;
    theCacheList.m_head = this;
}
JncClassCache::~JncClassCache()
//, ~m_globalClass()
{
    if (m_globalClass.IsSet())
    {
        ERROR0("ERROR: In ~JncClassCache()...\n");
        // more error message will follow from ~JncGlobalRef{}
    }
    DeleteIDList();
}


////////////////////////////////////////////////// OPERATIONS:

STA_I JncClassCache* JncClassCache::Get(JncEnv& env, jclass clazz)
{
    if (clazz)
    {
    JncClassCache** link;
    JncClassCache*  cache;

        link = &theCacheList.m_head;
        while (cache = *link)
        {
            if (cache->Matches(env, clazz))
            {
                *link = cache->m_next;
                cache->m_next = theCacheList.m_head;
                theCacheList.m_head = cache;
                return cache;
            }
            link = &cache->m_next;
        }

        cache = new JncClassCache(env, clazz);
        if (!cache)
        {
            ERROR0("ERROR: Failure to create new JncClassCache{}\n");
        }
        return cache;
    }
    else
    {
        ERROR0("ERROR: JncClassCache::Get() called with NULL jclass\n");
        return NULL;
    }
}

STA_I void JncClassCache::UncacheAll(JncEnv& env)
{
JncClassCache*  cache;

    cache = theCacheList.m_head;
    while (cache)
    {
    JncClassCache*  const next = cache->m_next;

        cache->m_globalClass.Unset(env);
        delete cache;
        cache = next;
    }
    theCacheList.m_head = NULL;
}

void JncClassCache::SaveID(ConStr name, ConStr signature, jfieldID fieldID)
{
    while (theCacheSemaphore > 0) TRACE0("TRACE: busy-waiting in JncClassCache::SaveID(field)\n");
    theCacheSemaphore++;
    if (theCacheSemaphore > 1) ERROR1("ERROR: theCacheSemaphore == %d\n", theCacheSemaphore);

    TRACE2("TRACE: Saving fieldID: %s=%s\n", name, signature);
    new JncID(name, signature, fieldID, &m_IDList);

    theCacheSemaphore--;
}
void JncClassCache::SaveID(ConStr name, ConStr signature, jmethodID methodID)
{
    while (theCacheSemaphore > 0) TRACE0("TRACE: busy-waiting in JncClassCache::SaveID(method)\n");
    theCacheSemaphore++;
    if (theCacheSemaphore > 1) ERROR1("ERROR: theCacheSemaphore == %d\n", theCacheSemaphore);

    TRACE2("TRACE: Saving methodID: %s=%s\n", name, signature);
    new JncID(name, signature, methodID, &m_IDList);

    theCacheSemaphore--;
}

void JncClassCache::DeleteIDList()
{
JncID*          next;
JncID*          id;

    while (theCacheSemaphore > 0) TRACE0("TRACE: busy-waiting in JncClassCache::DeleteIDList()\n");
    theCacheSemaphore++;
    if (theCacheSemaphore > 1) ERROR1("ERROR: theCacheSemaphore == %d\n", theCacheSemaphore);

    next = m_IDList;
    while (id = next)
    {
        next = id->m_next;
        delete id;
    }
    m_IDList = NULL;

    theCacheSemaphore--;
}


////////////////////////////////////////////////// DATA:

JncID const* JncClassCache::FindID(ConStr name, ConStr signature)
{
JncID**     link;
JncID*      id;

    while (theCacheSemaphore > 0) TRACE0("TRACE: busy-waiting in JncClassCache::FindID()\n");
    theCacheSemaphore++;
    if (theCacheSemaphore > 1) ERROR1("ERROR: theCacheSemaphore == %d\n", theCacheSemaphore);

#if JNC_CACHE_WITH_POINTERS
    link = &m_IDList;
    while (id = *link)
    {
        // quick pointer-comparison
        if ((id->m_name == name) && (id->m_signature == signature))
        {
            *link = id->m_next;
            id->m_next = m_IDList;
            m_IDList = id;
            TRACE2("TRACE: Found ID by pointer: %s=%s\n", name, signature);
            theCacheSemaphore--;
            return id;
        }
        link = &id->m_next;
    }
#endif

    link = &m_IDList;
    while (id = *link)
    {
        if (id->IsValid())
        {
            if (!::strcmp(id->m_name, name) && !::strcmp(id->m_signature, signature))
            {
                *link = id->m_next;
                id->m_next = m_IDList;
                m_IDList = id;
                TRACE2("TRACE: Found ID by string: %s=%s\n", name, signature);
                theCacheSemaphore--;
                return id;
            }
            link = &id->m_next;
        }
        else
        {
            ERROR0("ERROR: Bad JncID{} found and deleted.\n");
            *link = id->m_next;
            delete id;
        }
    }

    theCacheSemaphore--;
    return NULL;
}

STA_I jboolean JncClassCache::IsAnythingCached()
{
    return theCacheList.m_head ? JNI_TRUE
                               : JNI_FALSE;
}


#endif
