/*
 * @(#)JncClassCache.h	1.9 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncClassCache.h

#ifndef JncClassCache_H
#define JncClassCache_H 1

#include <malloc.h>
#include "Jnc.h"
#include "JncGlobalRef.h"


////////////////////////////////////////////////// FORWARDS:

class JncID;

////////////////////////////////////////////////// CLASSES:

#if JNC_CACHE_IDS
class JncClassCache
{
PUBLIC_OPERATIONS:
STA void                UncacheAll(JncEnv& env);

PUBLIC_DATA:
STA jboolean            IsAnythingCached();

friend class JncClassBase;

PROTECTED_OPERATIONS:
    void                SaveID(ConStr name, ConStr signature, jfieldID fieldID);
    void                SaveID(ConStr name, ConStr signature, jmethodID methodID);

STA JncClassCache*      Get(JncEnv& env, jclass clazz);

PROTECTED_DATA:
    jboolean            Matches(JncEnv& env, jclass clazz) const
    { return env->IsSameObject(GetClass(), clazz); }
    JncID const*        FindID(ConStr name, ConStr signature);
    jclass              GetClass() const
    { return (jclass)m_globalClass.GetJObject(); }

PRIVATE_EXISTENCE:
    // must use static operations above
    JncClassCache(JncEnv& env, jclass clazz);
    ~JncClassCache();

PRIVATE_OPERATIONS:
STA void*               operator new(size_t nSize)
    { return ::malloc(nSize); }
STA void                operator delete(void* p)
    { ::free(p); }

    void                DeleteIDList();

PRIVATE_DATA:
    JncGlobalRef        m_globalClass;
    JncID*              m_IDList;

    JncClassCache*      m_next;
};
#else
class JncClassCache
{
PUBLIC_OPERATIONS:
STA void                UncacheAll(JncEnv& env)
    {};

PUBLIC_DATA:
STA jboolean            IsAnythingCached()
    { return JNI_FALSE; }
};
#endif


#endif

