/*
 * @(#)JncEnv.cpp	1.4 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncEnv.c


#include "JncEnv.h"


////////////////////////////////////////////////// EXISTENCE:

JncEnv::JncEnv(JNIEnv* env)
: m_env(env)
{
    if (!env)
    {
        ERROR0("ERROR: JncEnv{} instantiated with NULL JNIEnv*\n");
    }
    m_isPendingException = JNI_FALSE;
}


////////////////////////////////////////////////// OPERATIONS:

jboolean JncEnv::CheckForPendingException()
{
    if (m_isPendingException)
    {
        return JNI_TRUE;
    }
    else
    {
    jthrowable      const exception = GetPendingException();

        if (exception)
        {
            DeleteLocalRef(exception);
            return JNI_TRUE;
        }
        return JNI_FALSE;
    }
}
void JncEnv::ClearPendingException()
{
    m_env->ExceptionClear();
    m_isPendingException = JNI_FALSE;
}

void JncEnv::Throw(ConStr exceptionClassName, ConStr message)
{
    if (!IsPendingException())
    {
    jclass      clazz;

        TRACE2("TRACE: Throwing Exception: %s \"%s\"\n", exceptionClassName, message);

        clazz = m_env->FindClass(exceptionClassName);
        if (!clazz)
        {
            ERROR1("ERROR: Can't find exception class name, \"%s\"\n", exceptionClassName);
            clazz = m_env->FindClass("java/lang/Exception");
        }
        Throw(clazz, message);
    }
}
void JncEnv::Throw(jclass exceptionClass, ConStr message)
{
    if (!IsPendingException())
    {
        if (exceptionClass)
        {
            m_env->ThrowNew(exceptionClass, message);
            m_isPendingException = JNI_TRUE;
        }
        else
        {
            ERROR1("ERROR: Attempt to throw exception of NULL jclass, \"%s\"\n", message);
        }
    }
}
void JncEnv::Throw(jthrowable exception)
{
    if (!IsPendingException())
    {
        if (exception)
        {
            m_env->Throw(exception);
            m_isPendingException = JNI_TRUE;
        }
        else
        {
            ERROR0("ERROR: Attempt to throw NULL jthrowable\n");
        }
    }
}


////////////////////////////////////////////////// DATA:

jthrowable JncEnv::GetPendingException()
{
jthrowable      const exception = m_env->ExceptionOccurred();

    if (exception)
    {
#if JNC_DEBUG
        m_env->ExceptionDescribe();
#endif
        m_isPendingException = JNI_TRUE;
    }
    else
    {
        m_isPendingException = JNI_FALSE;   // may be useful
    }
    return exception;
}


