/*
 * @(#)JncEnv.h	1.5 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


/****************************************************************************
 *  JncEnv.h
 *
 * Encapsulates a JNIEnv pointer.
 * Used by all other JNC classes to make JNI calls.
 * Keeps track of whether an exception has been thrown.
 *  If so, all JNC calls except those used to handle exceptions will fail
 *  gracefully without calling JNI (as the JNI spec demands).
 * Wraps miscellaneous JNI functionality :
 *  checking for/clearing/throwing exceptions
 *  creating/deleting local/global references
 *  aborting the VM
 *
 * Existence:
 *
 *
 * Operations:
 *
 *
 * Data:
 *
 *
 ****************************************************************************/

#ifndef JncEnv_H
#define JncEnv_H    1

#include "Jnc.h"


////////////////////////////////////////////////// CLASS:

class JncEnv
{
PUBLIC_EXISTENCE:
    JncEnv(JNIEnv* env);

PUBLIC_OPERATIONS:
    jboolean        CheckForPendingException();
    void            ClearPendingException();
    void            SetPendingException()
    { m_isPendingException = JNI_TRUE; }

    void            Throw(ConStr exceptionClassName, ConStr message);
    void            Throw(jclass exceptionClass, ConStr message);
    void            Throw(jthrowable exception);
    void            Punt(ConStr message) const
    { m_env->FatalError(message); }

    jobject         NewGlobalRef(jobject obj)
    { return m_env->NewGlobalRef(obj); /* THROWS: nothing   ??DAVID: really? */}
    void            DeleteGlobalRef(jobject globalRef)
    { m_env->DeleteGlobalRef(globalRef); }
    void            DeleteLocalRef(jobject localRef)
    { m_env->DeleteLocalRef(localRef); }

PUBLIC_DATA:
    jboolean        IsPendingException()
    { return m_isPendingException; }
    jthrowable      GetPendingException();

    operator JNIEnv*() const
    { return m_env; }
    JNIEnv* operator->() const
    { return m_env; }

PROTECTED_DATA:
    JNIEnv*         const m_env;
    jboolean        m_isPendingException;
};


#endif
