/*
 * @(#)JncID.h	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncID.h

#ifndef JncID_H
#define JncID_H 1

#include <malloc.h>
#include <string.h>
#include "Jnc.h"


////////////////////////////////////////////////// CLASSES:

class JncID
{
friend class JncClassCache;

PUBLIC_OPERATIONS:
STA void*           operator new(size_t nSize)
    { return ::malloc(nSize); }
STA void            operator delete(void* p)
    { ::free(p); }

PUBLIC_DATA:
    jboolean        IsValid() const;
    jfieldID        GetFieldID() const;
    jmethodID       GetMethodID() const;

PRIVATE_EXISTENCE:  // only meant to be created by JncClassCache{}
    JncID(ConStr name, ConStr signature, jfieldID fieldID, JncID** list);
    JncID(ConStr name, ConStr signature, jmethodID methodID, JncID** list);
    ~JncID();

PRIVATE_OPERATIONS:
STA ConStr          PrepareToSave(ConStr string);

PRIVATE_TYPES:
    union jID
    {
        jfieldID        field;
        jmethodID       method;
    };

PRIVATE_DATA:
    const ConStr    m_name;
    const ConStr    m_signature;
    jID             m_id;
#if JNC_DEBUG
    jboolean        m_isMethod;
#endif
    JncID*          m_next;
};


#if JNC_CACHE_IDS


////////////////////////////////////////////////// INLINE EXISTENCE:

INL JncID::JncID(ConStr name, ConStr signature, jfieldID fieldID, JncID** list)
: m_name(PrepareToSave(name))
, m_signature(PrepareToSave(signature))
{
#if !JNC_CACHE_WITH_POINTERS
    if (!m_name || !m_signature)
    {
        delete this;
        return;
    }
#endif

    m_id.field = fieldID;
#if JNC_DEBUG
    m_isMethod = JNI_FALSE;
#endif
    m_next = *list;
    *list = this;
}
INL JncID::JncID(ConStr name, ConStr signature, jmethodID methodID, JncID** list)
: m_name(PrepareToSave(name))
, m_signature(PrepareToSave(signature))
{
    m_id.method = methodID;
#if JNC_DEBUG
    m_isMethod = JNI_TRUE;
#endif
    m_next = *list;
    *list = this;
}

INL JncID::~JncID()
{
#if !JNC_CACHE_WITH_POINTERS
    if (m_name)
    {
        ::free((void*)m_name);
    }
    if (m_signature)
    {
        ::free((void*)m_signature);
    }
#endif
}


////////////////////////////////////////////////// INLINE OPERATIONS:

INL STA_I ConStr JncID::PrepareToSave(ConStr string)
{
    if (!string)
    {
        ERROR0("ERROR: JncID{} instantiated with NULL string\n");
    }

#if JNC_CACHE_WITH_POINTERS
    return string;
#else
    {
    char*       const copy = (char*)::malloc((::strlen(string) + 1) * sizeof(char));

        if (copy)
        {
            ::strcpy(copy, string);
        }
        else
        {
            ERROR1("ERROR: In JncID::PrepareToSave(), failed to allocate for \"%s\"\n", string);
        }
        return copy;
    }
#endif
}

////////////////////////////////////////////////// INLINE DATA:

INL jboolean JncID::IsValid() const
{
#if JNC_CACHE_WITH_POINTERS
    return JNI_TRUE;
#else
    return m_name && m_signature ? JNI_TRUE
                                 : JNI_FALSE;
#endif
}

INL jfieldID JncID::GetFieldID() const
{
#if JNC_DEBUG
    if (m_isMethod)
    {
        ERROR2("ERROR: JncID::GetFieldID() called for method: \"%s=%s\"\n", m_name, m_signature);
    }
#endif
    return m_id.field;
}
INL jmethodID JncID::GetMethodID() const
{
#if JNC_DEBUG
    if (!m_isMethod)
    {
        ERROR2("ERROR: JncID::GetMethodID() called for field: \"%s=%s\"\n", m_name, m_signature);
    }
#endif
    return m_id.method;
}


#endif


#endif
