/*
 * @(#)JncObject.cpp	1.10 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncObject.c


#include "Jnc.h"
#include "JncObject.h"
#include "JncEnv.h"
#include "JncClassCache.h"
#include "JncClass.h"


////////////////////////////////////////////////// EXISTENCE:

JncObject::JncObject(JncEnv& env, jobject obj)
: JncStatic(env, obj)
{
    m_synchronizeCount = 0;
    m_deleteLocalOnDestroy = JNI_FALSE;

#if 0   //KARA8: killed this trace message
    {
    static jboolean inAlready = JNI_FALSE;

        if (!inAlready)
        {
        char        className[512];

            inAlready = JNI_TRUE;
            GetClassSignature(className, sizeof(className) / sizeof(char) - 1);
            TRACE1("TRACE: JncObject{} of class \"%s\" instantiated\n", className);
            inAlready = JNI_FALSE;
        }
    }
#endif
}

JncObject::~JncObject()
//  : ~JncStatic()
{
    if (m_synchronizeCount != 0)
    {
        ERROR1("ERROR: JncObject{} destroyed with synchronize count of %d\n",
                m_synchronizeCount);
    }
    if (m_deleteLocalOnDestroy && m_jobj)
    {
        m_env->DeleteLocalRef(m_jobj);
    }
}


////////////////////////////////////////////////// METHOD-CALLING OPERATIONS:

void JncObject::CallVoidMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
        m_env->functions->CallVoidMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
    }

va_end(args);
}
jboolean JncObject::CallBooleanMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jboolean    result;

        result = m_env->functions->CallBooleanMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return JNI_FALSE;

va_end(args);
}
jchar JncObject::CallCharMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jchar   result;

        result = m_env->functions->CallCharMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return '\0';

va_end(args);
}
jbyte JncObject::CallByteMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jbyte   result;

        result = m_env->functions->CallByteMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jshort JncObject::CallShortMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jshort  result;

        result = m_env->functions->CallShortMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jint JncObject::CallIntMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jint    result;

        result = m_env->functions->CallIntMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jlong JncObject::CallLongMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jlong   result;

        result = m_env->functions->CallLongMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jfloat JncObject::CallFloatMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jfloat  result;

        result = m_env->functions->CallFloatMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0.0;

va_end(args);
}
jdouble JncObject::CallDoubleMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jdouble result;

        result = m_env->functions->CallDoubleMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0.0;

va_end(args);
}
jobject JncObject::CallObjectMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetMethodID(name, signature);

    if (methodID)
    {
    jobject result;

        result = m_env->functions->CallObjectMethodV(m_env, m_jobj, methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return NULL;

va_end(args);
}

void JncObject::Call_wait(jlong milliseconds)
{
    CallVoidMethod("wait", "(J)V", milliseconds);
}
void JncObject::Call_wait()
{
    CallVoidMethod("wait");
}
void JncObject::Call_notify()
{
    CallVoidMethod("notify");
}
void JncObject::Call_notifyAll()
{
    CallVoidMethod("notifyAll");
}
jstring JncObject::Call_toString()
{
    return CallStringMethod("toString");
}
jint JncObject::Call_hashCode()
{
    return CallIntMethod("hashCode");
}

jboolean JncObject::Call_equals(jobject obj)
{
JncClass    jncClass(GetEnv(), GetClass());
char        signature[512];
char*       p;

    signature[0] = '(';
    p = signature + 1 +
        jncClass.GetSignature(signature + 1, sizeof(signature) / sizeof(char) - 4);
    *p++ = ')';
    *p++ = 'Z';
    *p = '\0';

    return CallBooleanMethod("equals", signature, obj);
}
jobject JncObject::Call_clone()
{
JncClass    jncClass(GetEnv(), GetClass());
char        signature[512];

    signature[0] = '(';
    signature[1] = ')';
    jncClass.GetSignature(signature + 2, sizeof(signature) / sizeof(char) - 3);

    return CallObjectMethod("clone", signature);
}


////////////////////////////////////////////////// FIELD-SETTING OPERATIONS:

void JncObject::SetBooleanField(ConStr name, jboolean value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigBoolean);

    if (fieldID)
    {
        m_env->SetBooleanField(m_jobj, fieldID, value);
    }
}
void JncObject::SetCharField(ConStr name, jchar value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigChar);

    if (fieldID)
    {
        m_env->SetCharField(m_jobj, fieldID, value);
    }
}
void JncObject::SetByteField(ConStr name, jbyte value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigByte);

    if (fieldID)
    {
        m_env->SetByteField(m_jobj, fieldID, value);
    }
}
void JncObject::SetShortField(ConStr name, jshort value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigShort);

    if (fieldID)
    {
        m_env->SetShortField(m_jobj, fieldID, value);
    }
}
void JncObject::SetIntField(ConStr name, jint value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigInt);

    if (fieldID)
    {
        m_env->SetIntField(m_jobj, fieldID, value);
    }
}
void JncObject::SetLongField(ConStr name, jlong value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigLong);

    if (fieldID)
    {
        m_env->SetLongField(m_jobj, fieldID, value);
    }
}
void JncObject::SetFloatField(ConStr name, jfloat value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigFloat);

    if (fieldID)
    {
        m_env->SetFloatField(m_jobj, fieldID, value);
    }
}
void JncObject::SetDoubleField(ConStr name, jdouble value)
{
jfieldID        const fieldID = GetFieldID(name, m_sigDouble);

    if (fieldID)
    {
        m_env->SetDoubleField(m_jobj, fieldID, value);
    }
}
void JncObject::SetObjectField(ConStr name, ConStr signature, jobject value)
{
jfieldID        const fieldID = GetFieldID(name, signature);

    if (fieldID)
    {
        m_env->SetObjectField(m_jobj, fieldID, value);
    }
}

////////////////////////////////////////////////// FIELD-GETTING OPERATIONS:

jboolean JncObject::GetBooleanField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigBoolean);

    if (fieldID)
    {
        return m_env->GetBooleanField(m_jobj, fieldID);
    }
    return JNI_FALSE;
}
jchar JncObject::GetCharField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigChar);

    if (fieldID)
    {
        return m_env->GetCharField(m_jobj, fieldID);
    }
    return '\0';
}
jbyte JncObject::GetByteField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigByte);

    if (fieldID)
    {
        return m_env->GetByteField(m_jobj, fieldID);
    }
    return 0;
}
jshort JncObject::GetShortField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigShort);

    if (fieldID)
    {
        return m_env->GetShortField(m_jobj, fieldID);
    }
    return 0;
}
jint JncObject::GetIntField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigInt);

    if (fieldID)
    {
        return m_env->GetIntField(m_jobj, fieldID);
    }
    return 0;
}
jlong JncObject::GetLongField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigLong);

    if (fieldID)
    {
        return m_env->GetLongField(m_jobj, fieldID);
    }
    return 0;
}
jfloat JncObject::GetFloatField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigFloat);

    if (fieldID)
    {
        return m_env->GetFloatField(m_jobj, fieldID);
    }
    return 0.0;
}
jdouble JncObject::GetDoubleField(ConStr name)
{
jfieldID        const fieldID = GetFieldID(name, m_sigDouble);

    if (fieldID)
    {
        return m_env->GetDoubleField(m_jobj, fieldID);
    }
    return 0.0;
}
jobject JncObject::GetObjectField(ConStr name, ConStr signature)
{
jfieldID        const fieldID = GetFieldID(name, signature);

    if (fieldID)
    {
        return m_env->GetObjectField(m_jobj, fieldID);
    }
    return NULL;
}


////////////////////////////////////////////////// OPERATIONS:

void JncObject::EnterSynchronized()
{
    if (!m_env.IsPendingException())
    {
        if (m_env->MonitorEnter(m_jobj) == 0)   // THROWS: nothing
        {
            m_synchronizeCount++;
        }
        else
        {
            //DAVID?? Does failure mean monitor not available, does MonitorEnter() block?
            ERROR0("ERROR: Failure to enter an object's monitor.\n");
        }
    }
}
void JncObject::ExitSynchronized()
{
    if (!m_env.IsPendingException())
    {
        if (m_env->MonitorExit(m_jobj) == 0)
        // THROWS: nothing
        {
            m_synchronizeCount--;
        }
        else
        {
            ERROR0("ERROR: Failure to exit an object's monitor.\n");
        }
    }
}


////////////////////////////////////////////////// DATA:

jboolean JncObject::IsInstanceOf(jclass clazz) const
{
    if (!m_env.IsPendingException())
    {
        return m_env->IsInstanceOf(m_jobj, clazz);
        // THROWS: nothing
    }
    return JNI_FALSE;
}
jboolean JncObject::IsSameObject(jobject obj) const
{
    if (!m_env.IsPendingException())
    {
        return m_env->IsSameObject(m_jobj, obj);
        // THROWS: nothing
    }
    return JNI_FALSE;
}


