/*
 * @(#)JncObject.h	1.10 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncObject.h

#ifndef JncObject_H
#define JncObject_H 1

#include "Jnc.h"
#include "JncStatic.h"
#include "JncClassCache.h"


////////////////////////////////////////////////// CLASS:

class JncEnv;

class JncObject : public JncStatic
{
PUBLIC_EXISTENCE:
    JncObject(JncEnv& env, jobject obj);
VIR ~JncObject();

PUBLIC_OPERATIONS:
    void            CallVoidMethod(ConStr name, ConStr signature, ...);
    jboolean        CallBooleanMethod(ConStr name, ConStr signature, ...);
    jchar           CallCharMethod(ConStr name, ConStr signature, ...);
    jbyte           CallByteMethod(ConStr name, ConStr signature, ...);
    jshort          CallShortMethod(ConStr name, ConStr signature, ...);
    jint            CallIntMethod(ConStr name, ConStr signature, ...);
    jlong           CallLongMethod(ConStr name, ConStr signature, ...);
    jfloat          CallFloatMethod(ConStr name, ConStr signature, ...);
    jdouble         CallDoubleMethod(ConStr name, ConStr signature, ...);
    jobject         CallObjectMethod(ConStr name, ConStr signature, ...);

    void            CallVoidMethod(ConStr name)
    { CallVoidMethod(name, m_sigReturnsVoid); }
    jboolean        CallBooleanMethod(ConStr name)
    { return CallBooleanMethod(name, m_sigReturnsBoolean); }
    jchar           CallCharMethod(ConStr name)
    { return CallCharMethod(name, m_sigReturnsChar); }
    jbyte           CallByteMethod(ConStr name)
    { return CallByteMethod(name, m_sigReturnsByte); }
    jshort          CallShortMethod(ConStr name)
    { return CallShortMethod(name, m_sigReturnsShort); }
    jint            CallIntMethod(ConStr name)
    { return CallIntMethod(name, m_sigReturnsInt); }
    jlong           CallLongMethod(ConStr name)
    { return CallLongMethod(name, m_sigReturnsLong); }
    jfloat          CallFloatMethod(ConStr name)
    { return CallFloatMethod(name, m_sigReturnsFloat); }
    jdouble         CallDoubleMethod(ConStr name)
    { return CallDoubleMethod(name, m_sigReturnsDouble); }
    jstring         CallStringMethod(ConStr name)
    { return (jstring)CallObjectMethod(name, m_sigReturnsString); }
    jbooleanArray   CallBooleanArrayMethod(ConStr name)
    { return (jbooleanArray)CallObjectMethod(name, m_sigReturnsBooleanArray); }
    jcharArray      CallCharArrayMethod(ConStr name)
    { return (jcharArray)CallObjectMethod(name, m_sigReturnsCharArray); }
    jbyteArray      CallByteArrayMethod(ConStr name)
    { return (jbyteArray)CallObjectMethod(name, m_sigReturnsByteArray); }
    jshortArray     CallShortArrayMethod(ConStr name)
    { return (jshortArray)CallObjectMethod(name, m_sigReturnsShortArray); }
    jintArray       CallIntArrayMethod(ConStr name)
    { return (jintArray)CallObjectMethod(name, m_sigReturnsIntArray); }
    jlongArray      CallLongArrayMethod(ConStr name)
    { return (jlongArray)CallObjectMethod(name, m_sigReturnsLongArray); }
    jfloatArray     CallFloatArrayMethod(ConStr name)
    { return (jfloatArray)CallObjectMethod(name, m_sigReturnsFloatArray); }
    jdoubleArray    CallDoubleArrayMethod(ConStr name)
    { return (jdoubleArray)CallObjectMethod(name, m_sigReturnsDoubleArray); }

    void            Call_wait(jlong milliseconds);
    void            Call_wait();
    void            Call_notify();
    void            Call_notifyAll();
    jstring         Call_toString();
    jint            Call_hashCode();
    jboolean        Call_equals(jobject obj);
    jobject         Call_clone();

    void            SetBooleanField(ConStr name, jboolean value);
    void            SetCharField(ConStr name, jchar value);
    void            SetByteField(ConStr name, jbyte value);
    void            SetShortField(ConStr name, jshort value);
    void            SetIntField(ConStr name, jint value);
    void            SetLongField(ConStr name, jlong value);
    void            SetFloatField(ConStr name, jfloat value);
    void            SetDoubleField(ConStr name, jdouble value);
    void            SetObjectField(ConStr name, ConStr signature,
                                    jobject value);
    void            SetPointerField(ConStr name, void* value)
    { SetIntField(name, (jint)value); }
    void            SetStringField(ConStr name, jstring value)
    { SetObjectField(name, m_sigString, value); }
    void            SetBooleanArrayField(ConStr name, jbooleanArray value)
    { SetObjectField(name, m_sigBooleanArray, value); }
    void            SetCharArrayField(ConStr name, jcharArray value)
    { SetObjectField(name, m_sigCharArray, value); }
    void            SetByteArrayField(ConStr name, jbyteArray value)
    { SetObjectField(name, m_sigByteArray, value); }
    void            SetShortArrayField(ConStr name, jshortArray value)
    { SetObjectField(name, m_sigShortArray, value); }
    void            SetIntArrayField(ConStr name, jintArray value)
    { SetObjectField(name, m_sigIntArray, value); }
    void            SetLongArrayField(ConStr name, jlongArray value)
    { SetObjectField(name, m_sigLongArray, value); }
    void            SetFloatArrayField(ConStr name, jfloatArray value)
    { SetObjectField(name, m_sigFloatArray, value); }
    void            SetDoubleArrayField(ConStr name, jdoubleArray value)
    { SetObjectField(name, m_sigDoubleArray, value); }

    jboolean        GetBooleanField(ConStr name);
    jchar           GetCharField(ConStr name);
    jbyte           GetByteField(ConStr name);
    jshort          GetShortField(ConStr name);
    jint            GetIntField(ConStr name);
    jlong           GetLongField(ConStr name);
    jfloat          GetFloatField(ConStr name);
    jdouble         GetDoubleField(ConStr name);
    jobject         GetObjectField(ConStr name, ConStr signature);
    void*           GetPointerField(ConStr name)
    { return (void*)GetIntField(name); }
    jstring         GetStringField(ConStr name)
    { return (jstring)GetObjectField(name, m_sigString); }
    jbooleanArray   GetBooleanArrayField(ConStr name)
    { return (jbooleanArray)GetObjectField(name, m_sigBooleanArray); }
    jcharArray      GetCharArrayField(ConStr name)
    { return (jcharArray)GetObjectField(name, m_sigCharArray); }
    jbyteArray      GetByteArrayField(ConStr name)
    { return (jbyteArray)GetObjectField(name, m_sigByteArray); }
    jshortArray     GetShortArrayField(ConStr name)
    { return (jshortArray)GetObjectField(name, m_sigShortArray); }
    jintArray       GetIntArrayField(ConStr name)
    { return (jintArray)GetObjectField(name, m_sigIntArray); }
    jlongArray      GetLongArrayField(ConStr name)
    { return (jlongArray)GetObjectField(name, m_sigLongArray); }
    jfloatArray     GetFloatArrayField(ConStr name)
    { return (jfloatArray)GetObjectField(name, m_sigFloatArray); }
    jdoubleArray    GetDoubleArrayField(ConStr name)
    { return (jdoubleArray)GetObjectField(name, m_sigDoubleArray); }

    void            SetDeleteLocal(jboolean deleteOnDestroy = JNI_TRUE)
    { m_deleteLocalOnDestroy = deleteOnDestroy; }

    void            EnterSynchronized();
    void            ExitSynchronized();

PUBLIC_DATA:
    jboolean        IsValid() const
    { return m_jobj ? JNI_TRUE : JNI_FALSE; }
    jobject         GetJObject() const
    { return m_jobj; }
    jboolean        IsSameObject(jobject obj) const;
    jboolean        IsInstanceOf(jclass clazz) const;

PROTECTED_DATA:

PRIVATE_DATA:
    int             m_synchronizeCount;
    jboolean        m_deleteLocalOnDestroy;
};


#endif
