/*
 * @(#)JncObjectArray.cpp	1.6 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncObjectArray.cpp


#include "JncObjectArray.h"


////////////////////////////////////////////////// OPERATIONS:

jsize JncObjectArray::CopyOut(jobject* dst, jsize srcIndex, jsize maxLength) const
{
jsize       length;
jsize       i;
jsize       iEnd;

    length = ::Min(maxLength, GetLength() - srcIndex);
    i = srcIndex;
    iEnd = i + length;
    while (i < iEnd)
    {
        *dst++ = GetElement(i++);
        if (m_env.IsPendingException()) break;
    }
    return length;
}
jsize JncObjectArray::CopyIn(jsize dstIndex, jobject const* src, jsize maxLength) const
{
jsize       length;
jsize       i;
jsize       iEnd;

    length = ::Min(maxLength, GetLength() - dstIndex);
    i = dstIndex;
    iEnd = i + length;
    while (i < iEnd)
    {
        SetElement(i++, *src++);
        if (m_env.IsPendingException()) break;
    }
    return length;
}

void JncObjectArray::SetElement(jsize index, jobject value) const
{
    if (!m_env.IsPendingException())
    {
        m_env->SetObjectArrayElement(GetJObject(), index, value);
        // THROWS: ArrayIndexOutOfBoundsException, ArrayStoreException
        m_env.CheckForPendingException();
    }
}

STA_I jobjectArray JncObjectArray::New(JncEnv& env,
                                        jclass elementClass, jsize length,
                                        jobject initialElementValue /*= NULL*/)
{
    if (!env.IsPendingException())
    {
    jobjectArray    array;

        //??DAVID:  Does NewObjectArray() accept NULL for an initial element value?
        array = env->NewObjectArray(length, elementClass, initialElementValue);
        // THROWS: OutOfMemoryError
        if (!array)
        {
            ERROR0("ERROR: Failure to create object array\n");
        }
        return array;
    }
    return NULL;
}


////////////////////////////////////////////////// DATA:

jobject JncObjectArray::GetElement(jsize index) const
{
    if (!m_env.IsPendingException())
    {
    jobject     element;

        element = m_env->GetObjectArrayElement(GetJObject(), index);
        // THROWS: ArrayIndexOutOfBoundsException, ArrayStoreException
        if (!element)
        {
            ERROR1("ERROR: Failure to get element %d of object array\n", (int)index);
            m_env.SetPendingException();
        }
        return element;
    }
    return NULL;
}


