/*
 * @(#)JncPrimitiveArray.cpp	1.2 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncPrimitiveArray.cpp


#include "JncPrimitiveArray.h"


////////////////////////////////////////////////// EXISTENCE:

JncPrimitiveArray::JncPrimitiveArray(JncEnv& env, jarray array)
: JncArray(env, array)
{
    m_openedElements = NULL;
    m_openedElementsIsCopy = JNI_FALSE;
}
JncPrimitiveArray::~JncPrimitiveArray()
{
    EndAccess();
}


////////////////////////////////////////////////// OPERATIONS:

jsize JncPrimitiveArray::CopyOut(void* dst, jsize srcIndex, jsize maxLength)
{
    if (!m_env.IsPendingException())
    {
    jsize           length;

        CopyBackAccess();   // make sure all data is flushed back

        length = ::Min(GetLength() - srcIndex, maxLength);
        GetElements(dst, srcIndex, length);
        //THROWS: ArrayIndexOutOfBoundsException

        m_env.CheckForPendingException();
        return length;
    }
    return 0;
}
jsize JncPrimitiveArray::CopyIn(jsize dstIndex, void const* src, jsize maxLength)
{
    if (!m_env.IsPendingException())
    {
    jsize           length;

        EndAccess();    // let's not have access and CopyIn conflict

        length = ::Min(GetLength() - dstIndex, maxLength);
        SetElements(dstIndex, src, length);
        //THROWS: ArrayIndexOutOfBoundsException

        m_env.CheckForPendingException();
        return length;
    }
    return 0;
}

void JncPrimitiveArray::CopyBackAccess()
{
    if (m_openedElementsIsCopy)
    {
        //??DAVID:  SHOULD WE CHECK m_env.IsPendingException()?
        CloseElements(m_openedElements, JNI_COMMIT);
    }
}
void JncPrimitiveArray::EndAccess()
{
    if (m_openedElements)
    {
        //??DAVID:  SHOULD WE CHECK m_env.IsPendingException()?
        CloseElements(m_openedElements, 0);
        m_openedElements = NULL;
        m_openedElementsIsCopy = JNI_FALSE;
    }
}


////////////////////////////////////////////////// DATA:

void* JncPrimitiveArray::AccessElements()
{
    if (!m_env.IsPendingException())
    {
        if (!m_openedElements)
        {
            m_openedElements = OpenElements(&m_openedElementsIsCopy);
            if (!m_openedElements)
            {
                ERROR0("ERROR: Failure in JncPrimitiveArray::AccessElements()\n");
                m_env.CheckForPendingException();
            }
        }
        return m_openedElements;
    }
    return NULL;
}
jsize JncPrimitiveArray::CheckRange(jsize index)
{
    if (!m_env.IsPendingException())
    {
    jsize       const length = GetLength();

        if ((unsigned)index >= (unsigned)length)
        {
            m_env.Throw("java/lang/ArrayIndexOutOfBoundsException",
                            "in JncPrimitiveArray::CheckRange()");
            index = 0;
        }
    }
    return index;
}


