/*
 * @(#)JncStatic.cpp	1.5 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncStatic.c

#include "JncEnv.h"
#include "JncStatic.h"


////////////////////////////////////////////////// METHOD-CALLING OPERATIONS:

void JncStatic::CallStaticVoidMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
        m_env->functions->
            CallStaticVoidMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
    }

va_end(args);
}
jboolean JncStatic::CallStaticBooleanMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jboolean    result;

        result = m_env->functions->
            CallStaticBooleanMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return JNI_FALSE;

va_end(args);
}
jchar JncStatic::CallStaticCharMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jchar   result;

        result = m_env->functions->
            CallStaticCharMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return '\0';

va_end(args);
}
jbyte JncStatic::CallStaticByteMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jbyte   result;

        result = m_env->functions->
            CallStaticByteMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jshort JncStatic::CallStaticShortMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jshort  result;

        result = m_env->functions->
            CallStaticShortMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jint JncStatic::CallStaticIntMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jint    result;

        result = m_env->functions->
            CallStaticIntMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jlong JncStatic::CallStaticLongMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jlong   result;

        result = m_env->functions->
            CallStaticLongMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0;

va_end(args);
}
jfloat JncStatic::CallStaticFloatMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jfloat  result;

        result = m_env->functions->
            CallStaticFloatMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0.0;

va_end(args);
}
jdouble JncStatic::CallStaticDoubleMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jdouble result;

        result = m_env->functions->
            CallStaticDoubleMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return 0.0;

va_end(args);
}
jobject JncStatic::CallStaticObjectMethod(ConStr name, ConStr signature, ...)
{
va_list         args;
va_start(args, signature);

jmethodID       const methodID = GetStaticMethodID(name, signature);

    if (methodID)
    {
    jobject result;

        result = m_env->functions->
            CallStaticObjectMethodV(m_env, GetClass(), methodID, args);
        m_env.CheckForPendingException();
        return result;
    }
    return NULL;

va_end(args);
}

////////////////////////////////////////////////// FIELD-SETTING OPERATIONS:

void JncStatic::SetStaticBooleanField(ConStr name, jboolean value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigBoolean);

    if (fieldID)
    {
        m_env->SetStaticBooleanField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticCharField(ConStr name, jchar value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigChar);

    if (fieldID)
    {
        m_env->SetStaticCharField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticByteField(ConStr name, jbyte value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigByte);

    if (fieldID)
    {
        m_env->SetStaticByteField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticShortField(ConStr name, jshort value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigShort);

    if (fieldID)
    {
        m_env->SetStaticShortField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticIntField(ConStr name, jint value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigInt);

    if (fieldID)
    {
        m_env->SetStaticIntField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticLongField(ConStr name, jlong value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigLong);

    if (fieldID)
    {
        m_env->SetStaticLongField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticFloatField(ConStr name, jfloat value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigFloat);

    if (fieldID)
    {
        m_env->SetStaticFloatField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticDoubleField(ConStr name, jdouble value)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigDouble);

    if (fieldID)
    {
        m_env->SetStaticDoubleField(GetClass(), fieldID, value);
    }
}
void JncStatic::SetStaticObjectField(ConStr name, ConStr signature, jobject value)
{
jfieldID        const fieldID = GetStaticFieldID(name, signature);

    if (fieldID)
    {
        m_env->SetStaticObjectField(GetClass(), fieldID, value);
    }
}

////////////////////////////////////////////////// FIELD-GETTING OPERATIONS:

jboolean JncStatic::GetStaticBooleanField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigBoolean);

    if (fieldID)
    {
        return m_env->GetStaticBooleanField(GetClass(), fieldID);
    }
    return JNI_FALSE;
}
jchar JncStatic::GetStaticCharField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigChar);

    if (fieldID)
    {
        return m_env->GetStaticCharField(GetClass(), fieldID);
    }
    return '\0';
}
jbyte JncStatic::GetStaticByteField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigByte);

    if (fieldID)
    {
        return m_env->GetStaticByteField(GetClass(), fieldID);
    }
    return 0;
}
jshort JncStatic::GetStaticShortField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigShort);

    if (fieldID)
    {
        return m_env->GetStaticShortField(GetClass(), fieldID);
    }
    return 0;
}
jint JncStatic::GetStaticIntField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigInt);

    if (fieldID)
    {
        return m_env->GetStaticIntField(GetClass(), fieldID);
    }
    return 0;
}
jlong JncStatic::GetStaticLongField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigLong);

    if (fieldID)
    {
        return m_env->GetStaticLongField(GetClass(), fieldID);
    }
    return 0;
}
jfloat JncStatic::GetStaticFloatField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigFloat);

    if (fieldID)
    {
        return m_env->GetStaticFloatField(GetClass(), fieldID);
    }
    return 0.0;
}
jdouble JncStatic::GetStaticDoubleField(ConStr name)
{
jfieldID        const fieldID = GetStaticFieldID(name, m_sigDouble);

    if (fieldID)
    {
        return m_env->GetStaticDoubleField(GetClass(), fieldID);
    }
    return 0.0;
}
jobject JncStatic::GetStaticObjectField(ConStr name, ConStr signature)
{
jfieldID        const fieldID = GetStaticFieldID(name, signature);

    if (fieldID)
    {
        return m_env->GetStaticObjectField(GetClass(), fieldID);
    }
    return NULL;
}


////////////////////////////////////////////////// DATA:


