/*
 * @(#)JncStatic.h	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncStatic.h

#ifndef JncStatic_H
#define JncStatic_H 1

#include "Jnc.h"
#include "JncClassBase.h"


////////////////////////////////////////////////// CLASS:

class JncStatic : public JncClassBase
{
PUBLIC_EXISTENCE:
    JncStatic(JncEnv& env, ConStr className)
    : JncClassBase(env, className)
    {}
    JncStatic(JncEnv& env, jclass clazz)
    : JncClassBase(env, clazz)
    {}

PUBLIC_OPERATIONS:
    void            CallStaticVoidMethod(ConStr name, ConStr signature, ...);
    jboolean        CallStaticBooleanMethod(ConStr name, ConStr signature, ...);
    jchar           CallStaticCharMethod(ConStr name, ConStr signature, ...);
    jbyte           CallStaticByteMethod(ConStr name, ConStr signature, ...);
    jshort          CallStaticShortMethod(ConStr name, ConStr signature, ...);
    jint            CallStaticIntMethod(ConStr name, ConStr signature, ...);
    jlong           CallStaticLongMethod(ConStr name, ConStr signature, ...);
    jfloat          CallStaticFloatMethod(ConStr name, ConStr signature, ...);
    jdouble         CallStaticDoubleMethod(ConStr name, ConStr signature, ...);
    jobject         CallStaticObjectMethod(ConStr name, ConStr signature, ...);

    void            CallStaticVoidMethod(ConStr name)
    { CallStaticVoidMethod(name, m_sigReturnsVoid); }
    jboolean        CallStaticBooleanMethod(ConStr name)
    { return CallStaticBooleanMethod(name, m_sigReturnsBoolean); }
    jchar           CallStaticCharMethod(ConStr name)
    { return CallStaticCharMethod(name, m_sigReturnsChar); }
    jbyte           CallStaticByteMethod(ConStr name)
    { return CallStaticByteMethod(name, m_sigReturnsByte); }
    jshort          CallStaticShortMethod(ConStr name)
    { return CallStaticShortMethod(name, m_sigReturnsShort); }
    jint            CallStaticIntMethod(ConStr name)
    { return CallStaticIntMethod(name, m_sigReturnsInt); }
    jlong           CallStaticLongMethod(ConStr name)
    { return CallStaticLongMethod(name, m_sigReturnsLong); }
    jfloat          CallStaticFloatMethod(ConStr name)
    { return CallStaticFloatMethod(name, m_sigReturnsFloat); }
    jdouble         CallStaticDoubleMethod(ConStr name)
    { return CallStaticDoubleMethod(name, m_sigReturnsDouble); }
    jstring         CallStaticStringMethod(ConStr name)
    { return (jstring)CallStaticObjectMethod(name, m_sigReturnsString); }
    jbooleanArray   CallStaticBooleanArrayMethod(ConStr name)
    { return (jbooleanArray)CallStaticObjectMethod(name, m_sigReturnsBooleanArray); }
    jcharArray      CallStaticCharArrayMethod(ConStr name)
    { return (jcharArray)CallStaticObjectMethod(name, m_sigReturnsCharArray); }
    jbyteArray      CallStaticByteArrayMethod(ConStr name)
    { return (jbyteArray)CallStaticObjectMethod(name, m_sigReturnsByteArray); }
    jshortArray     CallStaticShortArrayMethod(ConStr name)
    { return (jshortArray)CallStaticObjectMethod(name, m_sigReturnsShortArray); }
    jintArray       CallStaticIntArrayMethod(ConStr name)
    { return (jintArray)CallStaticObjectMethod(name, m_sigReturnsIntArray); }
    jlongArray      CallStaticLongArrayMethod(ConStr name)
    { return (jlongArray)CallStaticObjectMethod(name, m_sigReturnsLongArray); }
    jfloatArray     CallStaticFloatArrayMethod(ConStr name)
    { return (jfloatArray)CallStaticObjectMethod(name, m_sigReturnsFloatArray); }
    jdoubleArray    CallStaticDoubleArrayMethod(ConStr name)
    { return (jdoubleArray)CallStaticObjectMethod(name, m_sigReturnsDoubleArray); }

    void            SetStaticBooleanField(ConStr name, jboolean value);
    void            SetStaticCharField(ConStr name, jchar value);
    void            SetStaticByteField(ConStr name, jbyte value);
    void            SetStaticShortField(ConStr name, jshort value);
    void            SetStaticIntField(ConStr name, jint value);
    void            SetStaticLongField(ConStr name, jlong value);
    void            SetStaticFloatField(ConStr name, jfloat value);
    void            SetStaticDoubleField(ConStr name, jdouble value);
    void            SetStaticObjectField(ConStr name, ConStr signature, jobject value);
    void            SetStaticPointerField(ConStr name, void* value)
    { SetStaticIntField(name, (jint)value); }
    void            SetStaticStringField(ConStr name, jstring value)
    { SetStaticObjectField(name, m_sigString, value); }
    void            SetStaticBooleanArrayField(ConStr name, jbooleanArray value)
    { SetStaticObjectField(name, m_sigBooleanArray, value); }
    void            SetStaticCharArrayField(ConStr name, jcharArray value)
    { SetStaticObjectField(name, m_sigCharArray, value); }
    void            SetStaticByteArrayField(ConStr name, jbyteArray value)
    { SetStaticObjectField(name, m_sigByteArray, value); }
    void            SetStaticShortArrayField(ConStr name, jshortArray value)
    { SetStaticObjectField(name, m_sigShortArray, value); }
    void            SetStaticIntArrayField(ConStr name, jintArray value)
    { SetStaticObjectField(name, m_sigIntArray, value); }
    void            SetStaticLongArrayField(ConStr name, jlongArray value)
    { SetStaticObjectField(name, m_sigLongArray, value); }
    void            SetStaticFloatArrayField(ConStr name, jfloatArray value)
    { SetStaticObjectField(name, m_sigFloatArray, value); }
    void            SetStaticDoubleArrayField(ConStr name, jdoubleArray value)
    { SetStaticObjectField(name, m_sigDoubleArray, value); }

    jboolean        GetStaticBooleanField(ConStr name);
    jchar           GetStaticCharField(ConStr name);
    jbyte           GetStaticByteField(ConStr name);
    jshort          GetStaticShortField(ConStr name);
    jint            GetStaticIntField(ConStr name);
    jlong           GetStaticLongField(ConStr name);
    jfloat          GetStaticFloatField(ConStr name);
    jdouble         GetStaticDoubleField(ConStr name);
    jobject         GetStaticObjectField(ConStr name, ConStr signature);
    void*           GetStaticPointerField(ConStr name)
    { return (void*)GetStaticIntField(name); }
    jstring         GetStaticStringField(ConStr name)
    { return (jstring)GetStaticObjectField(name, m_sigString); }
    jbooleanArray   GetStaticBooleanArrayField(ConStr name)
    { return (jbooleanArray)GetStaticObjectField(name, m_sigBooleanArray); }
    jcharArray      GetStaticCharArrayField(ConStr name)
    { return (jcharArray)GetStaticObjectField(name, m_sigCharArray); }
    jbyteArray      GetStaticByteArrayField(ConStr name)
    { return (jbyteArray)GetStaticObjectField(name, m_sigByteArray); }
    jshortArray     GetStaticShortArrayField(ConStr name)
    { return (jshortArray)GetStaticObjectField(name, m_sigShortArray); }
    jintArray       GetStaticIntArrayField(ConStr name)
    { return (jintArray)GetStaticObjectField(name, m_sigIntArray); }
    jlongArray      GetStaticLongArrayField(ConStr name)
    { return (jlongArray)GetStaticObjectField(name, m_sigLongArray); }
    jfloatArray     GetStaticFloatArrayField(ConStr name)
    { return (jfloatArray)GetStaticObjectField(name, m_sigFloatArray); }
    jdoubleArray    GetStaticDoubleArrayField(ConStr name)
    { return (jdoubleArray)GetStaticObjectField(name, m_sigDoubleArray); }

PUBLIC_DATA:

PROTECTED_EXISTENCE:
    JncStatic(JncEnv& env, jobject obj) // for JncObject{}
    : JncClassBase(env, obj)
    {}
};


#endif
