/*
 * @(#)JncString.cpp	1.6 98/03/18
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


// JncString.c


#include <memory.h>     // for memcpy()
#include "JncString.h"


////////////////////////////////////////////////// EXISTENCE:

JncString::JncString(JncEnv& env, jstring string)
: JncObject(env, string)
{
    m_utfAccess = NULL;
    m_unicodeAccess = NULL;
}

JncString::~JncString()     // virtual
//  : ~JncObject()
{
    EndAccess();
}


////////////////////////////////////////////////// OPERATIONS:

jchar const* JncString::AccessUnicode()
{
    if (!m_unicodeAccess)
    {
        if (!m_env.IsPendingException())
        {
            m_unicodeAccess = m_env->GetStringChars(GetJObject(), NULL);
            // THROWS: nothing
            if (!m_unicodeAccess)
            {
                ERROR0("ERROR: Failure to access string's Unicode characters\n");
                m_env.CheckForPendingException();
            }
            //??DAVID:  Is the string returned by GetStringChars() NULL-terminated?
            //??DAVID:  Are Unicode strings ever NULL-terminated?
        }
    }
    return m_unicodeAccess;
}
ConStr JncString::AccessUTF()
{
    if (!m_utfAccess)
    {
        if (!m_env.IsPendingException())
        {
            m_utfAccess = m_env->GetStringUTFChars(GetJObject(), NULL);
            // THROWS: nothing
            if (!m_utfAccess)
            {
                ERROR0("ERROR: Failure to access string's UTF characters\n");
                m_env.CheckForPendingException();
            }
        }
    }
    return m_utfAccess;
}
void JncString::EndAccess()
{
    //DAVID?? SHOULD I CHECK m_env.IsPendingException() OR NOT?

    if (m_unicodeAccess)
    {
        m_env->ReleaseStringChars(GetJObject(), m_unicodeAccess);
        // THROWS: nothing
        m_unicodeAccess = NULL;
    }
    if (m_utfAccess)
    {
        m_env->ReleaseStringUTFChars(GetJObject(), m_utfAccess);
        // THROWS: nothing
        m_utfAccess = NULL;
    }
}

jsize JncString::CopyOut(jchar* dst, jsize srcIndex, jsize maxLength)
{
jchar const*    const src = AccessUnicode();

    if (src)
    {
    jsize       length;

        length = GetLength() - srcIndex;
        if (length >= maxLength)
        {
            length = maxLength - 1;
        }
        ::memcpy(dst, src + srcIndex, length * sizeof(jchar));
        dst[length] = 0;
        return length;
    }
    return 0;
}
jsize JncString::CopyOut(char* dst, jsize srcIndex, jsize maxLength)
{
ConStr  const src = AccessUTF();

    if (src)
    {
    jsize       length;

        length = GetLength() - srcIndex;
        if (length >= maxLength)
        {
            length = maxLength - 1;
        }
        ::memcpy(dst, src + srcIndex, length * sizeof(char));
        dst[length] = 0;
        return length;
    }
    return 0;
}

STA_I jstring JncString::New(JncEnv& env, jsize length, jchar const* unicodeString)
{
    if (!env.IsPendingException())
    {
    jstring     const string = env->NewString(unicodeString, length);
    // THROWS: OutOfMemoryError

        if (!string)
        {
            ERROR0("ERROR: Failure to create new unicode jstring\n");
            env.SetPendingException();
        }
        return string;
    }
    return NULL;
}
STA_I jstring JncString::New(JncEnv& env, ConStr utfString)
{
    if (!env.IsPendingException())
    {
    jstring     const string = env->NewStringUTF(utfString);
    // THROWS: OutOfMemoryError

        if (!string)
        {
            ERROR1("ERROR: Failure to create new jstring: \"%s\"\n", utfString);
            env.SetPendingException();
        }
        return string;
    }
    return NULL;
}


////////////////////////////////////////////////// DATA:

jsize JncString::GetLength() const
{
    if (!m_env.IsPendingException())
    {
        return m_env->GetStringLength(GetJObject());
        // THROWS: nothing, never fails
    }
    return 0;
}

jchar JncString::operator[](jsize index)
{
    if ((unsigned)index >= (unsigned)GetLength())
    {
        m_env.Throw("java/lang/ArrayIndexOutOfBoundsException",
                        "in JncString::operator[]");
        return 0;
    }
    if (!AccessUnicode())
    {
        return 0;
    }
    return m_unicodeAccess[index];
}


