/*
 * @(#)X_API.c	1.18 98/07/13
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	X_API.c
**
**		This provides platform specfic functions
**

**
**	History	-
**	9/25/95		Created
**	12/14/95	Added XGetAndDetachResource
**	12/19/95	Added XSetBit & XTestBit
**	1/18/96		Spruced up for C++ extra error checking
**	1/28/96		Changed XGetAndDetachResource
**	2/3/96		Removed extra includes
**	2/11/96		Added XIsStereoSupported & XIs16BitSupported
**				Added XFixedDivide & XFixedMultiply
**				Started putting in platform defines around code functions
**	2/21/96		Changed XGetAndDetachResource to return a size
**	3/25/96		Modifed XExpandMace1to6 & XExpandMace1to3 to create silence if
**				not supported
**	3/29/96		Added XPutLong & XPutShort
**	4/20/96		Moved myFixedMultiply from GenSynth.c and moved all references
**				to XFixedMultiply
**				Moved myFixedDivide from GenSynth.c and moved all references
**				to XFixedDivide
**	4/21/96		Removed register usage in parameters
**	5/14/96		Fixed Odd address error in XIsOurMemoryPtr
**				Replaced XFixedMultiply and XFixedDivide with integer versions rather
**				than floating point
**	6/1/96		Added some resource management code
**	6/5/96		Added some Win95 stuff.
**	6/28/96		Added BeBox stuff
**	6/30/96		Changed font and re tabbed
**				Moved DecompressSampleFormatPtr from MacSpecificSMS
**				Moved XSetVolume & XGetVolume from MacSpecificSMS
**				Moved DecompressPtr from MacSpecificSMS
**	7/1/96		Added XCreateAccessCache
**				Changed XFILERESOURCEID to XFILERESOURCE_ID
**	7/2/96		Changed XFileOpenResource to support read only access
**	7/3/96		Added XStrCmp
**	7/4/96		Added error checking for opening read only resources
**	7/7/96		Added XStrnCmp
**	7/18/96		Modifed XFileOpenResource to read Cache resource
**				Fixed a NULL reference in PV_XFileValid
**				Fixed problem with cached resources and byte ordering
**	7/21/96		Fixed dumb bug in XGetFileResource that caused extra file scanning
**	7/23/96		Added XMicroseconds
**	8/6/96		Added XGetIndexedType & XCountTypes
**	8/12/96		Moved XExpandMace1to3 & XExpandMace1to6 to X_Compress.c
**	9/22/96		Added XRandom & XSeedRandom & XRandomRange
**	10/8/96		Added XStrCpy
**	10/11/96	Added XConvertNativeFileToXFILENAME
**				Added XFileSetPositionRelative
**				Added XDetermineByteOrder
**				Added XSwapLong & XSwapShort
**	10/15/96	Added XFileOpenResourceFromMemory
**	10/23/96	Changed XGetAndDetachResource to support our resource manager
**				if a file has been opened
**				Changed XFileClose to NULL out currentResourceFile when closing
**				file
**	11/7/96		Fixed a bug in XMicroseconds
**	11/8/96		Fixed bug in XStrCpy. Forgot to copy zero terminator
**	11/14/96	Added XGetResourceName & XGetNamedResource
**				Removed dependancy on MacSpecificSMS.h
**	11/26/96	Changed file read status native to MacOS from read/write
**				to read only
**	12/3/96		Removed C++ warnings on void * to char * conversion
**	12/4/96		Added XStrLen
**	12/15/96	Added X_SOLARIS
**	12/17/96	Fixed a bug in XBlockMove for WinOS.
**	12/18/96	Added all Solaris APIs
**	12/19/96	Added create flag in XFileOpenForWrite
**	12/30/96	Changed copyright
**	1/2/97		Added XCtoPstr & XPtoCstr
**	1/12/97		Renamed DecompressPtr to XDecompressPtr
**	1/13/97		Added XDuplicateStr
**	1/16/97		Fixed bug with XStrLen that added extra byte
**	1/20/97		Added XLStrCmp
**				Put in more support for MacOS native resource management
**	1/21/97		Added XLStrnCmp
**	1/23/97		Special cased XFixedMultiply to handle zero faster
**	1/24/97		Fixed XWaitMicroseocnds on WinOS to handle threads better
**	1/28/97		Changed XGetFileResource to not duplicate a resource if its
**				being loaded from memory
**				Added some Navio Code
**				Fixed a bug with XIsOurMemoryPtr (Thanks Jeff!). Forgot to
**				get the data blocks in a platform way.
**				Added XGetIndexedResource
**	1/29/97		Fixed some platform bugs with XGetIndexedFileResource
**				Added XCompressPtr
**	2/4/97		Fixed XBlockMove for Navio case. It was backwards
**	2/5/97		Added XFileOpenForReadFromMemory
**	2/6/97		Fixed a bug with XFileOpenResourceFromMemory that creates a cache,
**				its not needed for memory based resources
**	2/11/97		Modified XGetIndexedFileResource to support MacOS native resource manager
**	2/18/97		Fixed XFileOpenResource & XFileOpenResourceFromMemory to fail if file
**				is not a XFile resource
**	2/19/97		Added XStrStr
**	4/22/97		Fixed a bug with XGetPtrSize that referenced a NULL pointer
******************************************************************************
**
**	6/5/97		(Beatnik Editor Windows)
**				bvk - Changed the _creat call (in XFileOpenResource) to use the right
**				flags based on this goofy Microsoft C++ 5.0 (_S_IREAD | _S_IWRITE
**				instead of _O_RDWR).  Added an include (sys/stat.h).
**
**	6/12/97		bvk - Added XGetResourceNameOnly().  Copied/modified from other code.
				Doesn't load the resource data, just checks resource headers.
**	6/13/97		bvk	- Added XCountFileResourcesOfType(XFILE fileRef, long int theType);
**
**	6/19/97		bvk - Fixed bug in XAddFileResource.  Wasn't flipping the 'next' value,
**					Fixed bug getting name in XGetResourceNameOnly().
**
**	6/22/97		bvk - Changed XGetPtrSize(XPTR data).  It used to innocently assume that
					it could get the handle for any data pointer to get the size.  That
					actually leaves a lot of room for error...
**
**	6/27/97		bvk - Added XDeleteFileResource.  This functions in two ways, either to delete
**					the resource immediately, or to just mark it as a 'TRSH' resource.
**					Then, trash collection occurs when XCleanResourceFile() is called.
**					For either case, resource deletion fails if the file has been opened as
**					read only.
**
**	7/02/97		bvk - Fixed bug in XGetResourceNameOnly().
**
**	7/03/97		bvk - Fixed bug in XCleanResourceFile() that was writing the # of resources wrong.
**
**	7/05/97		bvk - Modified the resource info caching system.  Read carefully:
					Originally, the resource cache was only created if the res
					file was opened ReadOnly.  Ostensibly, this was because the
					resource map cache would be out of date should a resource add
					or delete take place.

					Now, the cache is used for all types of files.  It is updated
					correctly should the res file have added or deleted resources.

					I also rewrote some frequently called routines to use the cache.

					I also disabled loading the pre-generated cache from the resource file.
					Too many of my files didn't have 'em (and none of these routines write it out!)
**
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**	5/13/97		Fixed extra byte swap with XGetIndexedFileResource. Thanks Moe!
**	6/20/97		Changed XGetHardwareVolume & XSetHardwareVolume for MacOS
**				to support news names for CW 2.0
**	6/30/97		Fixed a bug in XCompressPtr that forgot the extra four bytes! Thanks Moe!
**	7/14/97		Put in special case to XFixedDivide to handle zero values better and faster.
**				Put in support for new X_PLATFORM type, called X_WIN_HAE
**	7/17/97		Removed XIsVirtualMemoryAvailable & XLockMemory & XUnlockMemory. Because
**				its assumed that all memory is locked.
**				Removed XSetHardwareSampleRate & XGetHardwareSampleRate
**	7/18/97		Added XFileSetLength
**	7/21/97		Changed XGetResourceName to call XGetResourceNameOnly.
**	7/30/97		Removed references to creat because its not ANSI, now using the ANSI way which
**				is to call open with O_CREAT flag.
**	8/13/97		Fixed bug in XDuplicateAndStripStr that didn't increment counter the right way
**	8/18/97		Changed X_WIN_HAE to USE_HAE_EXTERNAL_API
**	9/3/97		Wired XGetHardwareVolume & XSetHardwareVolume to HAE_EXTERNAL_API
**				Wrapped a few more functions around USE_CREATION_API
**	9/29/97		Changed XSetBit & XClearBit & XTestBit to be unsigned long rather than long
**	10/2/97		Fixed a bug with XGetFileResource when trying to load a memory based resource
**				it still continued searching after it found what it was looking for.
**				Fixed XGetIndexedFileResource to support memory based resource files without
**				duplicating memory
**	10/14/97	Fixed bug with XFileOpenForWrite that didn't allow creation of files.
**	10/18/97	Modified XDisposePtr to mark blocks deallocated
**	11/19/97	Modified XNewPtr to allocate memory for performance
*/
/*****************************************************************************/

#include "X_API.h"
#include "X_Formats.h"

#if X_PLATFORM == X_WEBTV
	#include "ObjectStore.h"
	#include "MemoryManager.h"
#endif

#if X_PLATFORM == X_WEBTV
	#include "MacintoshUndefines.h"
	#include <stdio.h>
	#include <string.h>
#endif

#if X_PLATFORM == X_MACINTOSH
	#include <Resources.h>
	#include <Memory.h>
	#include <Gestalt.h>
	#include <ToolUtils.h>
	#include <Sound.h>
	#include <Timer.h>
#endif

#if X_PLATFORM == X_WEBTV
	#include "MacintoshRedefines.h"
#endif

#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
	#include <windows.h>
	#include <windowsx.h>
	#include <commdlg.h>
	#include <stdio.h>
	#include <io.h>
	#include <fcntl.h>
	#include <sys/types.h>
	#include <sys/stat.h>
#endif

#if USE_HAE_EXTERNAL_API == TRUE
	#include "HAE_API.h"
#endif

#if X_PLATFORM == X_BE
	#include <KernelKit.h>
	#include <OS.h>
	#include <malloc.h>
	#include <stdio.h>
	#include <unistd.h>
	#include <fcntl.h>
#endif

#if X_PLATFORM == X_SOLARIS
	#include <stdio.h>
	#include <memory.h>
	#include <stdlib.h>
	#include <unistd.h>
	#include <sys/time.h>
	#include <fcntl.h>
#endif

#if X_PLATFORM == X_NAVIO
        #include <stdlib.h>
		#include <kernel/Thread.h>
#endif



#define DEBUG_PRINT_RESOURCE		0
#define USE_WIN32_FILE_IO			1


// Structures

// Every block of data allocated with XNewPtr will contain this structure before the pointer that is
// passed to the user.
struct XPI_Memblock
{
	long	blockID_one;		// ID that this is our block. part 1
	long	blockSize;			// block size
	long	blockID_two;		// ID that this is our block. part 2
};
typedef struct XPI_Memblock XPI_Memblock;

#define XPI_BLOCK_1_ID		FOUR_CHAR('I','G','O','R')	//	'IGOR'
#define XPI_BLOCK_2_ID		FOUR_CHAR('G','S','N','D')	//	'GSND'
#define XPI_BLOCK_3_ID		FOUR_CHAR('F','L','A','T')	//	'FLAT'

// $$kk: 03.05.98: changed as per anand palaniswami to fix compliation warning
// (bug #4115315)
#define XPI_DEAD_ID			FOUR_CHAR((unsigned char)0xDE, (unsigned char)0xAD, (unsigned char)0xFF,(unsigned char)0xFF)	//	Dead block

#define MAX_OPEN_XFILES		5	// max number of open resource files

// Variables
static short int	resourceFileCount = 0;				// number of open resource files
static XFILE		openResourceFiles[MAX_OPEN_XFILES];

// Private functions

// Check for a valid file reference
static XBOOL PV_XFileValid(XFILE fileRef)
{
	XFILENAME	*pReference;
	XBOOL		valid;

	valid = FALSE;
	pReference = (XFILENAME *)fileRef;
	if (pReference)
	{
		if (pReference->fileValidID == XPI_BLOCK_3_ID)
		{
			valid = TRUE;
		}
	}
	return valid;
}

// Given a valid XFILE, this will return the open index. Will return -1 if not valid
static short int PV_FindResourceFileReferenceIndex(XFILE fileRef)
{
	short int	count;

	for (count = 0; count < resourceFileCount; count++)
	{
		if (openResourceFiles[count] == fileRef)
		{
			return count;
		}
	}
	return -1;
}

static XBOOL PV_AddResourceFileToOpenFiles(XFILE fileRef)
{
	XBOOL		full;
	short int	count;

	full = TRUE;
	if (resourceFileCount < MAX_OPEN_XFILES)
	{
		for (count = MAX_OPEN_XFILES-2; count >= 0; count--)
		{
			openResourceFiles[count+1] = openResourceFiles[count];
		}
		openResourceFiles[0] = fileRef;
		resourceFileCount++;
		full = FALSE;
	}
	return full;
}

static void PV_RemoveResourceFileFromOpenFiles(XFILE fileRef)
{
	short int	count;
	short int	found;

	found = -1;
	for (count = 0; count < resourceFileCount; count++)
	{
		if (openResourceFiles[count] == fileRef)
		{
			found = count;
			break;
		}
	}
	if (found != -1)
	{
		for (count = found; count < resourceFileCount-1; count++)
		{
			openResourceFiles[count] = openResourceFiles[count+1];
		}
		openResourceFiles[count] = 0;
		resourceFileCount--;
	}
}

static INLINE XBOOL PV_IsAnyOpenResourceFiles(void)
{
	return (resourceFileCount) ? TRUE : FALSE;
}

// Functions

// convert a c string to a pascal string
void * XCtoPstr(register void *cstr)
{
	register char *	source;
	register char *	dest;
	register long	length;
	char			data[256];

	source = (char *)cstr;
	dest = data+1;
	length = 0;
	while (*source)
	{
		length++;
		*dest++ = *source++;
	}
	data[0] = (unsigned char)length;
	XBlockMove(data, cstr, length + 1);
	return cstr;
}

// convert a pascal string to a c string
void * XPtoCstr(register void *pstr)
{
	register char *	source;
	register char *	dest;
	register long	length;
	char			data[256];

	length = *((unsigned char *)pstr);
	source = (char *)pstr;
	source++;
	dest = data;

	while (length--)
	{
		*dest++ = *source++;
	}
	*dest = 0;
	length = *((unsigned char *)pstr) + 1;
	XBlockMove(data, pstr, length);
	return pstr;
}


#if X_PLATFORM == X_MACINTOSH
static XBOOL PV_IsVirtualMemoryAvailable(void)
{
	static long		cachedResult = -1;
	long			feature;

	if (cachedResult == -1)
	{
		feature = 0;
		if (Gestalt(gestaltVMAttr, &feature) == noErr)
		{
			if (feature & (1<<gestaltVMPresent))
			{
				cachedResult = TRUE;
			}
			else
			{
				cachedResult = FALSE;
			}
		}
	}
	return (XBOOL)cachedResult;
}
#endif

// Determine if a data block was allocated with our memory allocation API
static XPI_Memblock * XIsOurMemoryPtr(XPTR data)
{
	char			*pData;
	XPI_Memblock	*pBlock, *pBlockReturn;
#if USE_HAE_EXTERNAL_API == TRUE
		short int		code;
#endif
	pBlockReturn = NULL;
	if (data)
	{
		pData = (char *)data;
		pData -= sizeof(XPI_Memblock);
		pBlock = (XPI_Memblock *)pData;
#if USE_HAE_EXTERNAL_API == TRUE
		code = HAE_IsBadReadPointer(pBlock, sizeof(XPI_Memblock));
		if (code == 0 || code == 2)
#else
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
		if (!IsBadReadPtr(pBlock, sizeof(XPI_Memblock)))
	#endif
#endif
		{
			if ( (XGetLong(&pBlock->blockID_one) == XPI_BLOCK_1_ID) &&
				(XGetLong(&pBlock->blockID_two) == XPI_BLOCK_2_ID) )
			{
				pBlockReturn = pBlock;
			}
		}
	}
	return pBlockReturn;
}


// Allocates a block of ZEROED!!!! memory and locks it down
XPTR XNewPtr(long size)
{
	char			*data;
	XPI_Memblock	*pBlock;

	size += sizeof(XPI_Memblock);
#if USE_HAE_EXTERNAL_API == TRUE
	data = (char *)HAE_Allocate(size);
#else
	#if X_PLATFORM == X_MACINTOSH
	data = (Ptr)NewPtrClear(size);
	if (data)
	{
		if (PV_IsVirtualMemoryAvailable())
		{
			HoldMemory(data, size);
			LockMemory(data, size);
		}
	}
	#endif
	#if X_PLATFORM == X_WIN_HARDWARE
	data = (char *)BobAllocate(size);
	if (data)
	{
//		LockMemory(data, size);
		XSetMemory(data, size, 0);
	}
	#endif
	#if X_PLATFORM == X_WIN95
	data = (char *)GlobalAllocPtr(GHND, size);
	// the GHND flag includes the GMEM_ZEROINIT flag
//	VirtualLock(data, size);
	#endif
	#if X_PLATFORM == X_WEBTV
	data = (char *)AllocateZero(size);
	#endif
	#if X_PLATFORM == X_BE
	data = (char *)malloc(size);
	if (data)
	{
		XSetMemory(data, size, 0);
	}
	#endif
	#if X_PLATFORM == X_SOLARIS
    // $$kk: 10.14.97
    // changed this line as per Liang He's performance recommendations
    //	data = (char *)malloc(size);

    // $$kk: 10.23.97
    // changed th is line as per Liang He's performance revommendations.
    // buffer should be 8-byte aligned, not 4-byte aligned.
	//data = (char *)memalign(8, size);

	data = (char *)memalign(8, size + 8);
	if (data)
	{
		XSetMemory(data, size, 0);
	}
	#endif
	#if X_PLATFORM == X_NAVIO
	data = (char *)malloc(size);
	if (data)
	{
		XSetMemory(data, size, 0);
	}
	#endif
#endif
// $$kk: 07.07.98: removed this change to fix solaris memory leak bug #4149136 (ok with liang he)
// #if X_PLATFORM == X_SOLARIS
#if 0
    // make sure the buffer is 8-byte aligned.  -Liang
	if (data)
	{
		pBlock = (XPI_Memblock *)data;
		XPutLong(&pBlock->blockID_one, XPI_BLOCK_1_ID);
		XPutLong(&pBlock->blockID_two, XPI_BLOCK_2_ID);
		pBlock->blockSize = size + 8 - sizeof(XPI_Memblock);
		data += sizeof(XPI_Memblock);
		data += (8 - ((UINT32)data & 7));
	}
#else
	if (data)
	{
		pBlock = (XPI_Memblock *)data;
		XPutLong(&pBlock->blockID_one, XPI_BLOCK_1_ID);			// set our ID for this block
		XPutLong(&pBlock->blockID_two, XPI_BLOCK_2_ID);
		pBlock->blockSize = size - sizeof(XPI_Memblock);
		data += sizeof(XPI_Memblock);
	}
#endif
	return (XPTR)data;
}

void XDisposePtr(XPTR data)
{
	long			size;
	XPTR			osAllocatedData;
	XPI_Memblock	*pBlock;

	osAllocatedData = (XPTR)XIsOurMemoryPtr(data);
	if (osAllocatedData)
	{
		size = XGetPtrSize(data);	// need to get the size before we translate the pointer

		pBlock = (XPI_Memblock *)osAllocatedData;
		XPutLong(&pBlock->blockID_one, (unsigned long)XPI_DEAD_ID);			// set our ID for this block
		XPutLong(&pBlock->blockID_two, (unsigned long)XPI_DEAD_ID);			// to be dead. Used for tracking
#if USE_HAE_EXTERNAL_API == TRUE
		HAE_Deallocate(osAllocatedData);
#else
	#if X_PLATFORM == X_MACINTOSH
		if (PV_IsVirtualMemoryAvailable())
		{
			size = GetPtrSize((char *)osAllocatedData);
			UnholdMemory(osAllocatedData, size);
			UnlockMemory(osAllocatedData, size);
		}
		DisposePtr((Ptr)osAllocatedData);
	#endif
	#if X_PLATFORM == X_WIN_HARDWARE
		BobFree(osAllocatedData, size);
	#endif
	#if X_PLATFORM == X_WIN95
//		VirtualUnlock(osAllocatedData, XGetPtrSize(data));
		GlobalFreePtr(osAllocatedData);
	#endif
	#if X_PLATFORM == X_WEBTV
		FreeMemory(osAllocatedData);
	#endif
	#if X_PLATFORM == X_BE
		free(osAllocatedData);
	#endif
	#if X_PLATFORM == X_SOLARIS
		free(osAllocatedData);
	#endif
	#if X_PLATFORM == X_NAVIO
		free(osAllocatedData);
	#endif
#endif
	}
}

long XGetPtrSize(XPTR data)
{
	long		size;
	XPTR		odata;

	size = 0;
	if (data)
	{
		odata = (XPTR)XIsOurMemoryPtr(data);
		if (odata)
		{
			size = ((XPI_Memblock *)odata)->blockSize;
		}
		else
		{
			// then this block is not ours, so use the system to determine its real size
#if USE_HAE_EXTERNAL_API == TRUE
			size = HAE_SizeOfPointer(data);
#else
	#if X_PLATFORM == X_MACINTOSH
			size = GetPtrSize((Ptr)data);
	#endif
	#if X_PLATFORM == X_WEBTV
			size = MemorySize(data);
	#endif
	#if X_PLATFORM == X_WIN_HARDWARE
			size = 0;
	#endif
	#if X_PLATFORM == X_WIN95
			HANDLE	hData;	//bvk

			hData = GlobalPtrHandle(data);
			if (hData)
			{
				size = GlobalSize(hData);
			}
	#endif
	#if X_PLATFORM == X_BE
			size = 0;
	#endif
	#if X_PLATFORM == X_SOLARIS
			size = 0;
	#endif
	#if X_PLATFORM == X_NAVIO
			size = 0;
	#endif
#endif
		}
	}
	return size;
}

// Typical block move, expect it must be able to handle overlapping source and
// destination pointers
void XBlockMove(XPTR source, XPTR dest, long size)
{
	if (source && dest && size)
	{
#if USE_HAE_EXTERNAL_API == TRUE
		HAE_BlockMove(source, dest, size);
#else
	#if X_PLATFORM == X_MACINTOSH
		BlockMoveData(source, dest, size);
	#endif
	#if X_PLATFORM == X_WEBTV
		CopyMemory(source, dest, size);
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
		memcpy(dest, source, size);
	#endif
	#if X_PLATFORM == X_BE
		memcpy(dest, source, size);
	#endif
	#if X_PLATFORM == X_SOLARIS
		memcpy(dest, source, size);
	#endif
	#if X_PLATFORM == X_NAVIO
		memcpy(dest, source, size);
	#endif
#endif
	}
}

// set memory range with value
void XSetMemory(void *pAdr, long len, char value)
{
	register char *pData;

	if (pAdr && (len > 0))
	{
		pData = (char *)pAdr;
		do
		{
			*pData++ = value;
			len--;
		}
		while (len);
	}
}

// Given a pointer, and a bit number; this will set that bit to 1
void XSetBit(void *pBitArray, unsigned long whichbit)
{
	unsigned long	byteindex, bitindex;
	unsigned char *byte;

	if (pBitArray)
	{
		byteindex = whichbit / 8;
		bitindex = whichbit % 8;
		byte = &((unsigned char *)pBitArray)[byteindex];
		*byte |= (1L << bitindex);
	}
}

// Given a pointer, and a bit number; this will set that bit to 0
void XClearBit(void *pBitArray, unsigned long whichbit)
{
	unsigned long	byteindex, bitindex;
	unsigned char *byte;

	if (pBitArray)
	{
		byteindex = whichbit / 8;
		bitindex = whichbit % 8;
		byte = &((unsigned char *)pBitArray)[byteindex];
		*byte &= ~(1L << bitindex);
	}
}

// Given a pointer, and a bit number; this return the value of that bit
XBOOL XTestBit(void *pBitArray, unsigned long whichbit)
{
	register unsigned long	byteindex, byte, bitindex;

	if (pBitArray)
	{
		byteindex = whichbit / 8;
		bitindex = whichbit % 8;
		byte = ((unsigned char *)pBitArray)[byteindex];
		return (byte & (1L << bitindex)) ? TRUE : FALSE;
	}
	else
	{
		return FALSE;
	}
}

/* Sort an integer array from the lowest to the highest
*/
#if USE_CREATION_API == TRUE
void XBubbleSortArray(short int theArray[], short int theCount)
{
	register short int i, j, swapValue;

	for (i = 0; i < (theCount - 1); i++)
	{
		for (j = i + 1; j < theCount; j++)
		{
			if (theArray[i] > theArray[j])
			{
				swapValue = theArray[i];
				theArray[i] = theArray[j];
				theArray[j] = swapValue;
			}
		}
	}
}
#endif	// USE_CREATION_API


// Does sound hardware support stereo output
XBOOL XIsStereoSupported(void)
{
#if USE_HAE_EXTERNAL_API == TRUE
	return HAE_IsStereoSupported();
#else
	#if X_PLATFORM == X_MACINTOSH
	static long		cachedResult = -1;
	long			feature;

	if (cachedResult == -1)
	{
		feature = 0;
		cachedResult = FALSE;
		if (Gestalt(gestaltSoundAttr, &feature) == noErr)
		{
			if (feature & (1L<<gestaltStereoCapability))
			{
				cachedResult = TRUE;
			}
		}
	}
	return (XBOOL)cachedResult;
	#endif
	#if ( (X_PLATFORM == X_WIN95) 			||	\
		  (X_PLATFORM == X_WIN_HARDWARE)	||	\
		  (X_PLATFORM == X_WEBTV)			||	\
		  (X_PLATFORM == X_BE)				||	\
		  (X_PLATFORM == X_SOLARIS)			||	\
		  (X_PLATFORM == X_NAVIO) )
	return TRUE;
	#endif
#endif
}

// wait for a waitAmount of microseconds to pass
// CLS??: If this function is called from within the frame thread and
// JAVA_THREAD is non-zero, we'll probably crash.
void XWaitMicroseocnds(unsigned long waitAmount)
{
#if USE_HAE_EXTERNAL_API == TRUE
	HAE_WaitMicroseocnds(waitAmount);
#else
	#if X_PLATFORM == X_BE
	snooze((double)waitAmount);		// wait microseconds
	#endif
	#if X_PLATFORM == X_NAVIO
	ThreadSleep( waitAmount/1000 );
	#endif
	#if X_PLATFORM == X_WIN95
	unsigned long	ticks;

	ticks = XMicroseconds() + waitAmount;
	while (XMicroseconds() < ticks)
	{
		Sleep(0);	// Give up the rest of this time slice to other threads
	}
	#else
	unsigned long	ticks;

	ticks = XMicroseconds() + waitAmount;
	while (XMicroseconds() < ticks) {}
	#endif
#endif
}

#if X_PLATFORM == X_WIN95
static unsigned long PV_GetMicrosecondTick(void)
{

	return timeGetTime() * 1000;
#if 0

	LARGE_INTEGER	freq, tick;
	unsigned long	ms;
	double			freq_d, tick_d;
	ms = 0;
	if (QueryPerformanceFrequency(&freq) == FALSE)
	{	// does not support this type of timer
		ms = timeGetTime() * 1000;	// timeGetTime only returns milliseconds and we want microseconds
	}
	else
	{

		freq_d = (double)freq.LowPart;
		QueryPerformanceCounter(&tick);
		tick_d = (double)tick.LowPart;
		ms = (tick_d / freq_d * 1000.0);
	}

	return ms;
#endif

}
#endif

#if X_PLATFORM == X_WIN_HARDWARE
extern long hardwareTicks;
#endif


// Returns microseconds since boot
// 1 second = 10 milliseconds = 1000 microseconds
unsigned long XMicroseconds(void)
{
#if USE_HAE_EXTERNAL_API == TRUE
	return HAE_Microseconds();
#else
	#if X_PLATFORM == X_MACINTOSH
	static unsigned long	starttick = 0;
	UnsignedWide			tick;

	if (starttick == 0)
	{
		Microseconds(&tick);
		starttick = tick.lo;
	}
	Microseconds(&tick);
	return tick.lo - starttick;
	#endif
	#if X_PLATFORM == X_WIN95
	static unsigned long starttick = 0;

	if (starttick == 0)
	{
		starttick = PV_GetMicrosecondTick();
	}
	return (PV_GetMicrosecondTick() - starttick);
	#endif
	#if X_PLATFORM == X_BE
	static double betick = 0;

	if (betick == 0)
	{
		betick = system_time();
	}
	return (unsigned long) (system_time() - betick);
	#endif
	#if X_PLATFORM == X_WIN_HARDWARE
	static unsigned long starttick = 0;

	if (starttick == 0)
	{
		starttick = hardwareTicks;
	}
	return (hardwareTicks - starttick);
	#endif
	#if X_PLATFORM == X_SOLARIS
	static hrtime_t solaristick = 0;

	if (solaristick == 0)
	{
		solaristick = gethrtime();
	}
	/* ghrtime is in nanoseconds, we want microseconds. */
	return (unsigned long) ((gethrtime() - solaristick) / 1000);
	#endif
#endif
}

// Does sound hardware support 16 bit output
XBOOL XIs16BitSupported(void)
{
#if USE_HAE_EXTERNAL_API == TRUE
	return HAE_Is16BitSupported();
#else
	#if X_PLATFORM == X_MACINTOSH
	static long		cachedResult = -1;
	long			feature;

	if (cachedResult == -1)
	{
		feature = 0;
		cachedResult = FALSE;
		if (Gestalt(gestaltSoundAttr, &feature) == noErr)
		{
			if (feature & (1L<<gestalt16BitAudioSupport))
			{
				cachedResult = TRUE;
			}
		}
	}
	return (XBOOL)cachedResult;
	#endif
	#if ( (X_PLATFORM == X_WIN95) 			||	\
		  (X_PLATFORM == X_WIN_HARDWARE)	||	\
		  (X_PLATFORM == X_BE)				||	\
		  (X_PLATFORM == X_SOLARIS)			||	\
		  (X_PLATFORM == X_WEBTV)			||	\
		  (X_PLATFORM == X_NAVIO) )
	return TRUE;
	#endif
#endif
}

// Does sound hardware support 8 bit *linear* output
XBOOL XIs8BitSupported(void)
{
#if USE_HAE_EXTERNAL_API == TRUE
	return HAE_Is8BitSupported();
#else
	return TRUE;
#endif
}

#define PV_USE_FLOAT		0

// Used for XFixedMultiply & XFixedDivide
#if PV_USE_FLOAT
static const float FixedScalar = 1/65536.0;
#endif

// do a 16.16 fixed point multiply
XDWORD XFixedMultiply(XDWORD prodA, XDWORD prodB)
{
#if PV_USE_FLOAT
	XDWORD	result;
	float	prodAF, prodBF;

	if (prodA && prodB)
	{
		prodAF = prodA;
		prodBF = prodB * FixedScalar;
		result = prodAF * prodBF;
	}
	else
	{
		result = 0;
	}
	return result;
#else
	long result;

	if (prodA && prodB)
	{
		result = (((prodA >> 16L) & 0xFFFFL) * ((prodB >> 16L) & 0xFFFFL)) << 16L;
		result += ((prodA >> 16L) & 0xFFFFL) * ( prodB & 0xFFFFL);
		result += ((prodB >> 16L) & 0xFFFFL) * ( prodA & 0xFFFFL);
		result += (((prodA & 0xFFFFL) * (prodB & 0xFFFFL)) >> 16L) & 0xFFFFL;
	}
	else
	{
		result = 0;
	}
	return result;
#endif
}

// do a 16.16 fixed point divide
XDWORD XFixedDivide(XDWORD divisor, XDWORD dividend)
{
#if PV_USE_FLOAT
	XDWORD	result;
	float	prodAF, prodBF;

	if (divisor && dividend)
	{
		prodAF = divisor;
		prodBF = dividend * FixedScalar;
		result = prodAF / prodBF;
	}
	else
	{
		result = 0;
	}
	return result;
#else
	long result, temp, rfactor, i;
	long f2;

	if (divisor && dividend)
	{
		f2 = dividend;
		temp = divisor;
		result = 0;
		rfactor = 0x10000L;
		for (i = 0; i < 16; i++)
		{
			while ((temp >= f2) && (rfactor != 0) && (temp != 0))
			{
				temp -= f2;
				result += rfactor;
			}
			f2 = f2 >> 1L;
			rfactor = rfactor >> 1L;
		}
	}
	else
	{
		result = 0;
	}
	return result;
#endif
}

// given a pointer and a value, this with put a short in a ordered 68k way
void XPutShort(void *pData, unsigned long data)
{
	register unsigned char	*pByte;

	pByte = (unsigned char *)pData;
	pByte[0] = (data >> 8L) & 0xFFL;
	pByte[1] = data & 0xFFL;
}

// given a pointer and a value, this with put a long in a ordered 68k way
void XPutLong(void *pData, unsigned long data)
{
	register unsigned char	*pByte;

	pByte = (unsigned char *)pData;
	pByte[0] = (data >> 24L) & 0xFFL;
	pByte[1] = (data >> 16L) & 0xFFL;
	pByte[2] = (data >> 8L) & 0xFFL;
	pByte[3] = data & 0xFFL;
}


// given a pointer, get a long ordered in a Motorola way
unsigned long XGetLong(void *pData)
{
	register unsigned char	*pByte;
	register unsigned long	data;

	pByte = (unsigned char *)pData;
	data = ((unsigned long)pByte[0] << 24L) | ((unsigned long)pByte[1] << 16L) |
			((unsigned long)pByte[2] << 8L) | ((unsigned long)pByte[3]);
	return data;
}


// given a pointer, get a short int ordered in a Motorola way
unsigned short int XGetShort(void *pData)
{
	register unsigned char	*pByte;
	register unsigned short	data;

	pByte = (unsigned char *)pData;
	data = ((unsigned short int)pByte[0] << 8) | (unsigned short int)pByte[1];
	return data;
}


unsigned long XSwapLong(unsigned long value)
{
	long			newValue;
	unsigned char	*pOld;
	unsigned char	*pNew;
	unsigned char 	temp;

	pOld = (unsigned char *)&value;
	pNew = (unsigned char *)&newValue;

	temp = pOld[0];
	pNew[3] = temp;
	temp = pOld[1];
	pNew[2] = temp;
	temp = pOld[2];
	pNew[1] = temp;
	temp = pOld[3];
	pNew[0] = temp;

	return newValue;
}

unsigned short XSwapShort(unsigned short value)
{
	unsigned short	newValue;
	unsigned char	*pOld;
	unsigned char	*pNew;
	unsigned char 	temp;

	pOld = (unsigned char *)&value;
	pNew = (unsigned char *)&newValue;

	temp = pOld[0];
	pNew[1] = temp;
	temp = pOld[1];
	pNew[0] = temp;

	return newValue;
}

// if TRUE, then motorola; if FALSE then intel
XBOOL XDetermineByteOrder(void)
{
	static long	value = 0x12345678;
	XBOOL		order;

	order = FALSE;
	if (XGetLong(&value) == 0x12345678)
	{
		order = TRUE;
	}
	return order;
}

void XConvertNativeFileToXFILENAME(void *file, XFILENAME *xfile)
{
	if (xfile && file)
	{
		XSetMemory(xfile, (long)sizeof(XFILENAME), 0);
#if USE_HAE_EXTERNAL_API == TRUE
		{
			void	*dest;

			dest = &xfile->theFile;
			HAE_CopyFileNameNative(file, dest);
		}
#else
	#if X_PLATFORM == X_MACINTOSH
		xfile->theFile = *((FSSpec *)file);
	#endif
	#if ( (X_PLATFORM == X_WIN95) 			||	\
		  (X_PLATFORM == X_WIN_HARDWARE)	||	\
		  (X_PLATFORM == X_BE)				||	\
		  (X_PLATFORM == X_SOLARIS)			||	\
		  (X_PLATFORM == X_NAVIO) )
		XStrCpy((char *)xfile->theFile, (char *)file);
	#endif
#endif
	}
}

XFILE XFileOpenResourceFromMemory(XPTR pResource, unsigned long resourceLength, XBOOL allowCopy)
{
	XFILENAME			*pReference;
	XFILERESOURCEMAP	map;
	short int			err;

	err = 0;
	pReference = (XFILENAME *)XNewPtr((long)sizeof(XFILENAME));
	if (pReference)
	{
		pReference->pResourceData = pResource;
		pReference->resMemLength = resourceLength;
		pReference->resMemOffset = 0;
		pReference->resourceFile = TRUE;
		pReference->allowMemCopy = allowCopy;
		pReference->fileValidID = XPI_BLOCK_3_ID;
		if (pReference)
		{
			if (PV_AddResourceFileToOpenFiles((XFILE)pReference))
			{	// can't open any more files
				err = 1;
			}
			else
			{
				pReference->pCache = NULL;
				// since we are pointer based, we don't care about caching.
				// if we can handle the extra memory, we can build a cache.
//				pReference->pCache = XCreateAccessCache((XFILE)pReference);

				// validate resource file
				XFileSetPosition((XFILE)pReference, 0L);		// at start
				if (XFileRead((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
				{
					if (XGetLong(&map.mapID) != XFILERESOURCE_ID)
					{
						err = 2;
					}
				}
				else
				{
					err = 3;
				}
			}
		}
		if (err)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	}
	return (XFILE)pReference;
}

XFILE XFileOpenResource(XFILENAME *file, XBOOL readOnly)
{
	XFILENAME			*pReference;
	XFILERESOURCEMAP	map;

	pReference = (XFILENAME *)XNewPtr((long)sizeof(XFILENAME));
	if (pReference)
	{
		*pReference = *file;
		pReference->resourceFile = TRUE;
		pReference->fileValidID = XPI_BLOCK_3_ID;
		pReference->pResourceData = NULL;
		pReference->allowMemCopy = TRUE;
		pReference->readOnly = readOnly;

#if USE_HAE_EXTERNAL_API == TRUE
		if (readOnly)
		{
			pReference->fileReference = HAE_FileOpenForRead((void *)&pReference->theFile);
			if (pReference->fileReference == -1)
			{
				XDisposePtr(pReference);
				pReference = NULL;
			}
		}
		else
		{
			pReference->fileReference = HAE_FileOpenForReadWrite((void *)&pReference->theFile);
			if (pReference->fileReference == -1)
			{
				// must not be there, so create it and prepare it as a resource file
				HAE_FileCreate((void *)&pReference->theFile);
				pReference->fileReference = HAE_FileOpenForReadWrite((void *)&pReference->theFile);
				if (pReference->fileReference == -1)
				{
					XDisposePtr(pReference);
					pReference = NULL;
				}
				else
				{
					// prepare it as a resource file
					XFileSetPosition((XFILE)pReference, 0L);		// at start
					XPutLong(&map.mapID, XFILERESOURCE_ID);
					XPutLong(&map.version, 1);
					XPutLong(&map.totalResources, 0);
					XFileWrite((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP));
				}
			}
		}
#else

	#if X_PLATFORM == X_MACINTOSH
		{
			short int			err;

			if (readOnly)
			{
				err = FSpOpenDF(&pReference->theFile, fsRdPerm, &pReference->fileReference);
				if (err)
				{
					XDisposePtr(pReference);
					pReference = NULL;
				}
			}
			else
			{
				err = FSpOpenDF(&pReference->theFile, fsRdWrPerm, &pReference->fileReference);
				if (err)
				{
					// must not be there, so create it and prepare it as a resource file
					FSpDelete(&pReference->theFile);
					err = FSpCreate(&pReference->theFile, 'MPLR', 'FLAT', 0);
					if (err == noErr)
					{
						err = FSpOpenDF(&pReference->theFile, fsRdWrPerm, &pReference->fileReference);
						if (err)
						{
							XDisposePtr(pReference);
							pReference = NULL;
						}
						else
						{
							// prepare it as a resource file
							XFileSetPosition((XFILE)pReference, 0L);		// at start
							XPutLong(&map.mapID, XFILERESOURCE_ID);
							XPutLong(&map.version, 1);
							XPutLong(&map.totalResources, 0);
							XFileWrite((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP));
						}
					}
					else
					{
						XDisposePtr(pReference);
						pReference = NULL;
					}
				}
			}
		}
	#endif

	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
		if (readOnly)
		{
			pReference->fileReference = _open(pReference->theFile, _O_RDONLY | _O_BINARY);
			if (pReference->fileReference == -1)
			{
				XDisposePtr(pReference);
				pReference = NULL;
			}
		}
		else
		{
			pReference->fileReference = _open(pReference->theFile, _O_RDWR | _O_BINARY);
			if (pReference->fileReference == -1)
			{
				// must not be there, so create it and prepare it as a resource file
				pReference->fileReference = _open(pReference->theFile, _O_CREAT | _O_RDWR | _O_BINARY | _O_TRUNC);
				if (pReference->fileReference == -1)
				{
					XDisposePtr(pReference);
					pReference = NULL;
				}
				else
				{
					// prepare it as a resource file
					XFileSetPosition((XFILE)pReference, 0L);		// at start
					XPutLong(&map.mapID, XFILERESOURCE_ID);
					XPutLong(&map.version, 1);
					XPutLong(&map.totalResources, 0);
					XFileWrite((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP));
				}
			}
		}
	#endif

	#if X_PLATFORM == X_BE
		if (readOnly)
		{
			pReference->fileReference = open(pReference->theFile, O_RDONLY | O_BINARY);
			if (pReference->fileReference == -1)
			{
				XDisposePtr(pReference);
				pReference = NULL;
			}
		}
		else
		{
			pReference->fileReference = open(pReference->theFile, O_RDWR | O_BINARY);
			if (pReference->fileReference == -1)
			{
				// must not be there, so create it and prepare it as a resource file
				pReference->fileReference = open(pReference->theFile, O_CREAT | O_RDWR | O_BINARY | O_TRUNC);
				if (pReference->fileReference == -1)
				{
					XDisposePtr(pReference);
					pReference = NULL;
				}
				else
				{
					// prepare it as a resource file
					XFileSetPosition((XFILE)pReference, 0L);		// at start
					XPutLong(&map.mapID, XFILERESOURCE_ID);
					XPutLong(&map.version, 1);
					XPutLong(&map.totalResources, 0);
					XFileWrite((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP));
				}
			}
		}
	#endif

	#if X_PLATFORM == X_SOLARIS
		if (readOnly)
		{
			pReference->fileReference = open(pReference->theFile, O_RDONLY);	// O_BINARY
			if (pReference->fileReference == -1)
			{
				XDisposePtr(pReference);
				pReference = NULL;
			}
		}
		else
		{
			pReference->fileReference = open(pReference->theFile, O_RDWR);	// O_BINARY
			if (pReference->fileReference == -1)
			{
				// must not be there, so create it and prepare it as a resource file
				pReference->fileReference = open(pReference->theFile, O_CREAT | O_RDWR | O_TRUNC);	// O_BINARY
				if (pReference->fileReference == -1)
				{
					XDisposePtr(pReference);
					pReference = NULL;
				}
				else
				{
					// prepare it as a resource file
					XFileSetPosition((XFILE)pReference, 0L);		// at start
					XPutLong(&map.mapID, XFILERESOURCE_ID);
					XPutLong(&map.version, 1);
					XPutLong(&map.totalResources, 0);
					XFileWrite((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP));
				}
			}
		}
	#endif
	#if X_PLATFORM == X_NAVIO
		if (readOnly)
		{
			pReference->fileReference = FileOpen(pReference->theFile, FILE_O_RDONLY );
			if (pReference->fileReference == NULL)
			{
				XDisposePtr(pReference);
				pReference = NULL;
			}
		}
		else
		{
			pReference->fileReference = FileOpen(pReference->theFile, FILE_O_RDWR );
			if (pReference->fileReference == NULL)
			{
				// must not be there, so create it and prepare it as a resource file
				pReference->fileReference = FileCreate(pReference->theFile);
				if (pReference->fileReference == NULL)
				{
					XDisposePtr(pReference);
					pReference = NULL;
				}
				else
				{
					// prepare it as a resource file
					XFileSetPosition((XFILE)pReference, 0L);		// at start
					XPutLong(&map.mapID, XFILERESOURCE_ID);
					XPutLong(&map.version, 1);
					XPutLong(&map.totalResources, 0);
					XFileWrite((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP));
				}
			}
		}
	#endif
#endif
		if (pReference)
		{
			// success
			if (PV_AddResourceFileToOpenFiles((XFILE)pReference))
			{	// can't open any more files
				XDisposePtr(pReference);
				pReference = NULL;
			}
			else
			{
				pReference->pCache = NULL;
//				if (readOnly)
				{
					// Try to open XFILECACHE_ID cache block.
					pReference->pCache = (XFILERESOURCECACHE *)XGetFileResource((XFILE)pReference,
																XFILECACHE_ID, 0, NULL, NULL);
					if (pReference->pCache == NULL)
					{
						pReference->pCache = XCreateAccessCache((XFILE)pReference);
					}
#if X_WORD_ORDER != FALSE
					else
					{
						XFILE_CACHED_ITEM	*pItem;
						long				count;

						// walk through CACH resource and fix up any platform byte ordering issues
						pReference->pCache->totalResources = XGetLong(&pReference->pCache->totalResources);
						for (count = 0; count < pReference->pCache->totalResources; count++)
						{
							pItem = &pReference->pCache->cached[count];
							pItem->resourceType = XGetLong(&pItem->resourceType);
							pItem->resourceID = XGetLong(&pItem->resourceID);
							pItem->resourceLength = XGetLong(&pItem->resourceLength);
							pItem->fileOffsetName = XGetLong(&pItem->fileOffsetName);
							pItem->fileOffsetData = XGetLong(&pItem->fileOffsetData);
						}
					}
#endif
				}
				// validate resource file
				XFileSetPosition((XFILE)pReference, 0L);		// at start
				if (XFileRead((XFILE)pReference, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
				{
					if (XGetLong(&map.mapID) != XFILERESOURCE_ID)
					{
						XDisposePtr(pReference);
						pReference = NULL;
					}
				}
				else
				{
					XDisposePtr(pReference);
					pReference = NULL;
				}
			}
		}
	}
	return (XFILE)pReference;
}

XFILE XFileOpenForReadFromMemory(XPTR pMemoryBlock, unsigned long memoryBlockSize)
{
	XFILENAME	*pReference;

	pReference = (XFILENAME *)XNewPtr((long)sizeof(XFILENAME));
	if (pReference)
	{
		pReference->pResourceData = pMemoryBlock;
		pReference->resMemLength = memoryBlockSize;
		pReference->resMemOffset = 0;
		pReference->resourceFile = FALSE;
		pReference->allowMemCopy = TRUE;
		pReference->fileValidID = XPI_BLOCK_3_ID;
		pReference->pCache = NULL;
		pReference->fileReference = 0;
	}
	return (XFILE)pReference;
}

XFILE XFileOpenForRead(XFILENAME *file)
{
	XFILENAME	*pReference;

	pReference = (XFILENAME *)XNewPtr((long)sizeof(XFILENAME));
	if (pReference)
	{
		*pReference = *file;
		pReference->resourceFile = FALSE;
		pReference->fileValidID = XPI_BLOCK_3_ID;
		pReference->pResourceData = NULL;
		pReference->allowMemCopy = TRUE;
		pReference->pCache = NULL;
#if USE_HAE_EXTERNAL_API == TRUE
		pReference->fileReference = HAE_FileOpenForRead((void *)&pReference->theFile);
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
#else

	#if X_PLATFORM == X_MACINTOSH
		if (FSpOpenDF(&pReference->theFile, fsRdPerm, &pReference->fileReference))
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
		pReference->fileReference = _open(pReference->theFile, _O_RDONLY | _O_BINARY);
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if X_PLATFORM == X_BE
		pReference->fileReference = open(pReference->theFile, O_RDONLY | O_BINARY);
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if X_PLATFORM == X_SOLARIS
		pReference->fileReference = open(pReference->theFile, O_RDONLY);	// O_BINARY
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if X_PLATFORM == X_NAVIO
		pReference->fileReference = FileOpen(pReference->theFile, FILE_O_RDONLY);
		if (pReference->fileReference == NULL)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
#endif
	}
	return (XFILE)pReference;
}

#if USE_CREATION_API == TRUE
XFILE XFileOpenForWrite(XFILENAME *file, XBOOL create)
{
	XFILENAME	*pReference;

	pReference = (XFILENAME *)XNewPtr((long)sizeof(XFILENAME));
	if (pReference)
	{
		*pReference = *file;
		pReference->resourceFile = FALSE;
		pReference->fileValidID = XPI_BLOCK_3_ID;
		pReference->pResourceData = NULL;
		pReference->allowMemCopy = TRUE;
		pReference->pCache = NULL;
#if USE_HAE_EXTERNAL_API == TRUE
		if (create)
		{
			pReference->fileReference = HAE_FileCreate((void *)&pReference->theFile);
		}
		pReference->fileReference = HAE_FileOpenForReadWrite((void *)&pReference->theFile);
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
#else
	#if X_PLATFORM == X_MACINTOSH
		if (create)
		{
			FSpDelete(&pReference->theFile);
			FSpCreate(&pReference->theFile, 'IGOR', 'IGOR', 0);
			FSpCreateResFile(&pReference->theFile, 'IGOR', 'IGOR', 0);
		}
		if (FSpOpenDF(&pReference->theFile, fsRdWrPerm, &pReference->fileReference))
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
		if (create)
		{
			pReference->fileReference = _open(pReference->theFile, _O_WRONLY | _O_CREAT | _O_TRUNC);
		}
		else
		{
			pReference->fileReference = _open(pReference->theFile, O_RDWR);
		}
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if X_PLATFORM == X_BE
		if (create)
		{
			pReference->fileReference = open(pReference->theFile, O_WRONLY | O_CREAT | O_TRUNC);
		}
		else
		{
			pReference->fileReference = open(pReference->theFile, O_RDWR);
		}
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if X_PLATFORM == X_SOLARIS
		if (create)
		{
			pReference->fileReference = open(pReference->theFile, O_WRONLY | O_CREAT | O_TRUNC);
		}
		else
		{
			pReference->fileReference = open(pReference->theFile, O_RDWR);
		}
		if (pReference->fileReference == -1)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
	#if X_PLATFORM == X_NAVIO
		pReference->fileReference = FileOpen(pReference->theFile, FILE_O_RDWR);
		if (pReference->fileReference == NULL)
		{
			XDisposePtr(pReference);
			pReference = NULL;
		}
	#endif
#endif
	}
	return (XFILE)pReference;
}
#endif

void XFileClose(XFILE fileRef)
{
	XFILENAME	*pReference;

	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pCache)
		{
			XDisposePtr((XPTR)pReference->pCache);
			pReference->pCache = NULL;
		}
		if (pReference->pResourceData)
		{
			XDisposePtr((XPTR)pReference->pResourceData);
			pReference->pResourceData = NULL;
		}
		else
		{
#if USE_HAE_EXTERNAL_API == TRUE
			HAE_FileClose(pReference->fileReference);
#else
	#if X_PLATFORM == X_MACINTOSH
			FSClose(pReference->fileReference);
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
			_close(pReference->fileReference);
	#endif
	#if X_PLATFORM == X_BE
			close(pReference->fileReference);
	#endif
	#if X_PLATFORM == X_SOLARIS
			close(pReference->fileReference);
	#endif
	#if X_PLATFORM == X_NAVIO
			FileClose(pReference->fileReference);
	#endif
#endif
		}
		PV_RemoveResourceFileFromOpenFiles(fileRef);
		XDisposePtr(pReference);
	}
}

long XFileRead(XFILE fileRef, void * buffer, long bufferLength)
{
	XFILENAME	*pReference;
	long		newLength;
	long		err;

	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pResourceData)
		{
			err = 0;
			newLength = pReference->resMemOffset + bufferLength;
			if (newLength > pReference->resMemLength)
			{
				bufferLength -= (newLength - pReference->resMemLength);
				err = -1;	// eof
			}
			XBlockMove((char *)pReference->pResourceData + pReference->resMemOffset,
						buffer, bufferLength);
			pReference->resMemOffset += bufferLength;
			return err;
		}
		else
		{
#if USE_HAE_EXTERNAL_API == TRUE
			return (HAE_ReadFile(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
#else
	#if X_PLATFORM == X_MACINTOSH
			return (FSRead(pReference->fileReference, &bufferLength, buffer) == noErr) ? 0 : -1;
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
			return (_read(pReference->fileReference, (char *)buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
	#if X_PLATFORM == X_BE
			return (read(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
	#if X_PLATFORM == X_SOLARIS
			return (read(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
	#if X_PLATFORM == X_NAVIO
			return (FileRead(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
#endif
		}
	}
	return -1;
}

long XFileWrite(XFILE fileRef, void *buffer, long bufferLength)
{
	XFILENAME	*pReference;

	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pResourceData)
		{
			return -1;	// can't write a memory resource
		}
		else
		{
#if USE_HAE_EXTERNAL_API == TRUE
			return (HAE_WriteFile(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
#else
	#if X_PLATFORM == X_MACINTOSH
			return (FSWrite(pReference->fileReference, &bufferLength, buffer) == noErr) ? 0 : -1;
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
			return (_write(pReference->fileReference, (char *)buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
	#if X_PLATFORM == X_BE
			return (write(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
	#if X_PLATFORM == X_SOLARIS
			return (write(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
	#if X_PLATFORM == X_NAVIO
			return (FileWrite(pReference->fileReference, buffer, bufferLength) == bufferLength) ? 0 : -1;
	#endif
#endif
		}
	}
	return -1;
}

long XFileSetPosition(XFILE fileRef, long filePosition)
{
	XFILENAME	*pReference;
	long		err;

	err = -1;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pResourceData)
		{
			if ( (filePosition >= 0) && (filePosition < pReference->resMemLength) )
			{
				pReference->resMemOffset = filePosition;
				err = 0;
			}
		}
		else
		{
#if USE_HAE_EXTERNAL_API == TRUE
			err = HAE_SetFilePosition(pReference->fileReference, filePosition);
#else
	#if X_PLATFORM == X_MACINTOSH
			err = SetFPos(pReference->fileReference, fsFromStart, filePosition);
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
			err = (_lseek(pReference->fileReference, filePosition, SEEK_SET) == -1) ? -1 : 0;
	#endif
	#if X_PLATFORM == X_BE
			err = (lseek(pReference->fileReference, filePosition, SEEK_SET) == -1) ? -1 : 0;
	#endif
	#if X_PLATFORM == X_SOLARIS
			err = (lseek(pReference->fileReference, filePosition, SEEK_SET) == -1) ? -1 : 0;
	#endif
	#if X_PLATFORM == X_NAVIO
			err = (FileSeek(pReference->fileReference, FILE_SEEK_SET, filePosition) == -1) ? -1 : 0;
	#endif
#endif
		}
	}
	return err;
}

long XFileSetPositionRelative(XFILE fileRef, long relativeOffset)
{
	long		pos;

	pos = XFileGetPosition(fileRef);
	if (pos != -1)
	{
		pos = XFileSetPosition(fileRef, pos + relativeOffset);
	}
	return pos;
}

static XBYTE * PV_GetFilePositionFromMemoryResource(XFILE fileRef)
{
	XFILENAME	*pReference;
	XBYTE		*pos;

	pos = NULL;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pResourceData)
		{
			pos = ((XBYTE *)pReference->pResourceData) + pReference->resMemOffset;
		}
	}
	return pos;
}

long XFileGetPosition(XFILE fileRef)
{
	XFILENAME	*pReference;
	long		pos;

	pos = -1;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pResourceData)
		{
			pos = pReference->resMemOffset;
		}
		else
		{
#if USE_HAE_EXTERNAL_API == TRUE
			pos = HAE_GetFilePosition(pReference->fileReference);
#else
	#if X_PLATFORM == X_MACINTOSH
			GetFPos(pReference->fileReference, &pos);
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
			pos = _lseek(pReference->fileReference, 0, SEEK_CUR);
	#endif
	#if X_PLATFORM == X_BE
			pos = lseek(pReference->fileReference, 0, SEEK_CUR);
	#endif
	#if X_PLATFORM == X_SOLARIS
			pos = lseek(pReference->fileReference, 0, SEEK_CUR);
	#endif
	#if X_PLATFORM == X_NAVIO
			pos = FileSeek(pReference->fileReference, FILE_SEEK_CUR, 0);
	#endif
#endif
		}
	}
	return pos;
}

long XFileSetLength(XFILE fileRef, unsigned long newSize)
{
	XFILENAME	*pReference;
	long		error;

	error = 0;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pResourceData == NULL)
		{	// not a memory file
#if USE_HAE_EXTERNAL_API == TRUE
			error = HAE_SetFileLength(pReference->fileReference, newSize);
#else
	#if X_PLATFORM == X_MACINTOSH
			error = SetEOF(pReference->fileReference, newSize);
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
			_chsize(pReference->fileReference, newSize);
	#endif
	#if X_PLATFORM == X_BE
			error = -1;
			// deleting of resources will fail until this function is fixed
//			chsize(pReference->fileReference, newSize);
	#endif
	#if X_PLATFORM == X_SOLARIS
			error = -1;
			// deleting of resources will fail until this function is fixed
//			chsize(pReference->fileReference, newSize);
	#endif
	#if X_PLATFORM == X_NAVIO
		#error "Need to define this function"
	#endif
#endif
		}
	}
	return (error) ? -1 : 0;
}

long XFileGetLength(XFILE fileRef)
{
	XFILENAME	*pReference;
	long		pos;

	pos = -1;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pResourceData)
		{
			pos = pReference->resMemLength;
		}
		else
		{
#if USE_HAE_EXTERNAL_API == TRUE
			pos = HAE_GetFileLength(pReference->fileReference);
#else
	#if X_PLATFORM == X_MACINTOSH
			GetEOF(pReference->fileReference, &pos);
	#endif
	#if (X_PLATFORM == X_WIN95) || (X_PLATFORM == X_WIN_HARDWARE)
			pos = _lseek(pReference->fileReference, 0, SEEK_END);
			_lseek(pReference->fileReference, 0, SEEK_SET);
	#endif
	#if X_PLATFORM == X_BE
			pos = lseek(pReference->fileReference, 0, SEEK_END);
			lseek(pReference->fileReference, 0, SEEK_SET);
	#endif
	#if X_PLATFORM == X_SOLARIS
			pos = lseek(pReference->fileReference, 0, SEEK_END);
			lseek(pReference->fileReference, 0, SEEK_SET);
	#endif
	#if X_PLATFORM == X_NAVIO
			pos = FileSeek(pReference->fileReference, FILE_SEEK_END, 0);
			FileSeek(pReference->fileReference, FILE_SEEK_SET, 0);
	#endif
#endif
		}
	}
	return pos;
}

// search the cache for a particular item
static XFILE_CACHED_ITEM * PV_XGetCacheEntry(XFILE fileRef, long resourceType, long resourceID)
{
	XFILENAME			*pReference;
	long				count, total;
	XFILERESOURCECACHE	*pCache;
	XFILE_CACHED_ITEM	*pItem;

	pItem = NULL;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		pCache = pReference->pCache;
		if (pCache)
		{
			total = pCache->totalResources;
			for (count = 0; count < total; count++)
			{
				if (pCache->cached[count].resourceType == resourceType)
				{
					if (pCache->cached[count].resourceID == resourceID)
					{
						pItem = &pCache->cached[count];
						break;
					}
				}
			}
		}
	}
	return pItem;
}

// search the cache for a particular item with a name
static XFILE_CACHED_ITEM * PV_XGetNamedCacheEntry(XFILE fileRef, long resourceType, void *cName)
{
	XFILENAME			*pReference;
	long				count, total;
	XFILERESOURCECACHE	*pCache;
	XFILE_CACHED_ITEM	*pItem;
	long				err;
	char				PName[256];
	long				savePos;

	pItem = NULL;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		savePos = XFileGetPosition(fileRef);
		pCache = pReference->pCache;
		if (pCache)
		{
			total = pCache->totalResources;
			for (count = 0; count < total; count++)
			{
				if (pCache->cached[count].resourceType == resourceType)
				{
					XFileSetPosition(fileRef, pCache->cached[count].fileOffsetName);
					err = XFileRead(fileRef, &PName[0], 1L);		// get name
					if (PName[0])
					{
						err = XFileRead(fileRef, &PName[1], (long)PName[0]);
						if (XStrCmp((char *)cName, (char *)XPtoCstr(PName)) == 0)
						{
							pItem = &pCache->cached[count];
							break;
						}
					}
				}
			}
		}
		XFileSetPosition(fileRef, savePos);
	}
	return pItem;
}


// bvk 6/12/97
//	This scans the resource file for the header data and returns the name associated
//		with a type and id.  Taken pretty much verbatim from XGetFileResource (just
//		without the data loading).

char *	XGetResourceNameOnly( XFILE fileRef, long resourceType, long resourceID, char *pName )
{
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total;
	char				PName[256];
	XFILE_CACHED_ITEM	*pCacheItem;

	err = 0;
	PName[0] = 0;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		// first do we have a cache?
		if (pReference->pCache)
		{
			// second check to see if its in the cache
			pCacheItem = PV_XGetCacheEntry(fileRef, resourceType, resourceID);
			if (pCacheItem)
			{
				// get name
				if (pName)
				{
					XFileSetPosition(fileRef, pCacheItem->fileOffsetName);
					err = XFileRead(fileRef, &PName[0], 1L);		// get name
					if (PName[0])
					{
						err = XFileRead(fileRef, &PName[1], (long)PName[0]);
						if (pName)
						{
							XBlockMove(PName, pName, (long)PName[0] + 1);
						}
					}
				}
			}
			else
			{
				err = -1;	// can't find it
			}
		}
		else
		{
			XFileSetPosition(fileRef, 0L);		// at start
			if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
			{
				if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
				{
					next = sizeof(XFILERESOURCEMAP);
					total = XGetLong(&map.totalResources);
					for (count = 0; (count < total) && (err == 0); count++)
					{
						err = XFileSetPosition(fileRef, next);		// at start
						if (err == 0)
						{
							err = XFileRead(fileRef, &next, (long)sizeof(long));		// get next pointer
							next = XGetLong(&next);
							if (next != -1L)
							{
								err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
								if (XGetLong(&data) == resourceType)
								{
									err = XFileRead(fileRef, &data, (long)sizeof(long));		// get ID
									if (XGetLong(&data) == resourceID)
									{
										err = XFileRead(fileRef, &PName[0], 1L);		// get name
										if (PName[0])
										{
											err = XFileRead(fileRef, &PName[1], (long)PName[0]);
											if (pName)
											{
												XBlockMove(PName, pName, (long)PName[0] + 1 );
												break;
											}
										}
										err = XFileRead(fileRef, &data, (long)sizeof(long));		// get length
										data = XGetLong(&data);		// get resource size

									}
								}
							}
							else
							{
								err = -4;
								//DebugStr("\pNext offset is bad");
								break;
							}
						}
						else
						{
							err = -3;
							//DebugStr("\pCan't set next position");
							break;
						}
					}
				}
			}
		}
	}
	return (err ? NULL : pName);
}


XPTR XGetFileResource(XFILE fileRef, long resourceType, long resourceID, void *pName, long *pReturnedResourceSize)
{
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total;
	XPTR				pData;
	char				PName[256];
	XFILE_CACHED_ITEM	*pCacheItem;

	err = 0;
	pData = NULL;
	if (pReturnedResourceSize)
	{
		*pReturnedResourceSize = 0;
	}
#if DEBUG_PRINT_RESOURCE
	XPutLong(PName, resourceType);
	PName[4] = 0;
	printf("GetResource %s %ld is ", PName, resourceID);
#endif
	PName[0] = 0;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		// first do we have a cache?
		if (pReference->pCache)
		{
			// second check to see if its in the cache
			pCacheItem = PV_XGetCacheEntry(fileRef, resourceType, resourceID);
			if (pCacheItem)
			{
				// get name
				if (pName)
				{
					XFileSetPosition(fileRef, pCacheItem->fileOffsetName);
					err = XFileRead(fileRef, &PName[0], 1L);		// get name
					if (PName[0])
					{
						err = XFileRead(fileRef, &PName[1], (long)PName[0]);
						if (pName)
						{
							XBlockMove(PName, pName, (long)PName[0] + 1);
						}
					}
				}
				// get data
				XFileSetPosition(fileRef, pCacheItem->fileOffsetData);

				// is data memory based?
				if (pReference->pResourceData && (pReference->allowMemCopy == FALSE) )
				{	// don't bother coping data again, since its already in memory
					pData = PV_GetFilePositionFromMemoryResource(fileRef);
					if (pData)
					{
						if (pReturnedResourceSize)
						{
							*pReturnedResourceSize = pCacheItem->resourceLength;
						}
					}
					else
					{
						err = -2;
						//DebugStr("\pOut of memory; can't allocate resource");
					}
				}
				else
				{
					pData = XNewPtr(pCacheItem->resourceLength);
					if (pData)
					{
						err = XFileRead(fileRef, pData, pCacheItem->resourceLength);
						if (pReturnedResourceSize)
						{
							*pReturnedResourceSize = pCacheItem->resourceLength;
						}
					}
					else
					{
						err = -2;
						//DebugStr("\pOut of memory; can't allocate resource");
					}
				}
			}
			else
			{
				err = -1;	// can't find it
			}
		}
		else
		{
			XFileSetPosition(fileRef, 0L);		// at start
			if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
			{
				if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
				{
					next = sizeof(XFILERESOURCEMAP);
					total = XGetLong(&map.totalResources);
					for (count = 0; (count < total) && (err == 0); count++)
					{
						err = XFileSetPosition(fileRef, next);		// at start
						if (err == 0)
						{
							err = XFileRead(fileRef, &next, (long)sizeof(long));		// get next pointer
							next = XGetLong(&next);
							if (next != -1L)
							{
								err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
								if (XGetLong(&data) == resourceType)
								{
									err = XFileRead(fileRef, &data, (long)sizeof(long));		// get ID
									if (XGetLong(&data) == resourceID)
									{
										err = XFileRead(fileRef, &PName[0], 1L);		// get name
										if (PName[0])
										{
											err = XFileRead(fileRef, &PName[1], (long)PName[0]);
											if (pName)
											{
												XBlockMove(PName, pName, (long)PName[0] + 1);
											}
										}
										err = XFileRead(fileRef, &data, (long)sizeof(long));		// get length
										data = XGetLong(&data);		// get resource size

										// get data
										// is data memory based?
										if (pReference->pResourceData && (pReference->allowMemCopy == FALSE) )
										{	// don't bother coping data again, since its already in memory
											pData = PV_GetFilePositionFromMemoryResource(fileRef);
											if (pData)
											{
												if (pReturnedResourceSize)
												{
													*pReturnedResourceSize = data;
												}
												err = 0;
												break;
											}
											else
											{
												err = -2;
												//DebugStr("\pOut of memory; can't allocate resource");
											}
										}
										else
										{
											pData = XNewPtr(data);
											if (pData)
											{
												err = XFileRead(fileRef, pData, data);
												if (pReturnedResourceSize)
												{
													*pReturnedResourceSize = data;
												}
												break;
											}
											else
											{
												err = -2;
												//DebugStr("\pOut of memory; can't allocate resource");
												break;
											}
										}
									}
								}
							}
							else
							{
								err = -4;
								//DebugStr("\pNext offset is bad");
								break;
							}
						}
						else
						{
							err = -3;
							//DebugStr("\pCan't set next position");
							break;
						}
					}
				}
			}
		}
	}
#if DEBUG_PRINT_RESOURCE
	printf((pData) ? "OK\n" : "BAD\n");
#endif
	return pData;
}


//bvk New!
//	Adds another cache entry to end of cache.

#if USE_CREATION_API == TRUE
static XBOOL PV_AddToAccessCache(XFILE fileRef, XFILE_CACHED_ITEM *cacheItemPtr )
{
	XFILENAME			*pReference;
	XFILERESOURCECACHE	*pCache,*newCache;
	long int			resCount;

	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		pCache = pReference->pCache;
		if (pCache)
		{
			resCount = pCache->totalResources + 1;
			newCache = (XFILERESOURCECACHE *)XNewPtr((long)sizeof(XFILERESOURCECACHE) +
													((long)sizeof(XFILE_CACHED_ITEM) * resCount));
			if (newCache)
			{
				XBlockMove(pCache, newCache, (long)sizeof(XFILERESOURCECACHE) +
												((long)sizeof(XFILE_CACHED_ITEM) * (resCount - 1)));

				XDisposePtr(pCache);
				pReference->pCache = newCache;
				newCache->totalResources = resCount;

				newCache->cached[resCount - 1].resourceType = cacheItemPtr->resourceType;
				newCache->cached[resCount - 1].resourceID = cacheItemPtr->resourceID;			// resource ID
				newCache->cached[resCount - 1].resourceLength = cacheItemPtr->resourceLength;		// resource ID
				newCache->cached[resCount - 1].fileOffsetName = cacheItemPtr->fileOffsetName;		// file offset from 0 to resource name
				newCache->cached[resCount - 1].fileOffsetData = cacheItemPtr->fileOffsetData;		// file offset from 0 to resource data

				return TRUE;
			}
		}
	}
	return FALSE;
}
#endif	// USE_CREATION_API == TRUE

// This will scan through the resource file and return an pointer to an array of
XFILERESOURCECACHE * XCreateAccessCache(XFILE fileRef)
{
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total;
	XFILERESOURCECACHE	*pCache;
	char				pPName[256];

	err = 0;
	pCache = NULL;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		XFileSetPosition(fileRef, 0L);		// at start
		if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
		{
			if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
			{
				next = sizeof(XFILERESOURCEMAP);
				total = XGetLong(&map.totalResources);
				pCache = (XFILERESOURCECACHE *)XNewPtr((long)sizeof(XFILERESOURCECACHE) +
														((long)sizeof(XFILE_CACHED_ITEM) * total));
				if (pCache)
				{
					pCache->totalResources = total;
					for (count = 0; (count < total) && (err == 0); count++)
					{
						err = XFileSetPosition(fileRef, next);		// at start
						if (err == 0)
						{
							err = XFileRead(fileRef, &next, (long)sizeof(long));				// get next pointer
							next = XGetLong(&next);
							if (next != -1L)
							{
								err = XFileRead(fileRef, &data, (long)sizeof(long));			// get type
								pCache->cached[count].resourceType = XGetLong(&data);

								err = XFileRead(fileRef, &data, (long)sizeof(long));			// get ID
								pCache->cached[count].resourceID = XGetLong(&data);

								pCache->cached[count].fileOffsetName = XFileGetPosition(fileRef);	// get name
								err = XFileRead(fileRef, &pPName[0], 1L);
								if (pPName[0])
								{
									err = XFileRead(fileRef, &pPName[1], (long)pPName[0]);
								}

								err = XFileRead(fileRef, &data, (long)sizeof(long));			// get length
								pCache->cached[count].resourceLength = XGetLong(&data);
								pCache->cached[count].fileOffsetData = XFileGetPosition(fileRef);	// save data offset
							}
							else
							{
								err = -4;
								//DebugStr("\pNext offset is bad");
								break;
							}
						}
						else
						{
							err = -3;
							//DebugStr("\pCan't set next position");
							break;
						}
					}
				}
			}
		}
	}
	if (err)
	{
		XDisposePtr((XPTR)pCache);
		pCache = NULL;
	}
	return pCache;
}

static XBOOL PV_CheckForTypes(long *pTypes, long total, long typeCheck)
{
	long	count;
	XBOOL	found;

	found = FALSE;
	for (count = 0; count < total; count++)
	{
		if (pTypes[count] == typeCheck)
		{
			found = TRUE;
			break;
		}
	}
	return found;
}

#define MAX_XFILE_SCAN_TYPES		5120L

// Return the number of resource types included in file

//	bvk:  why doesn't this use the cache?

long XCountTypes(XFILE fileRef)
{
	long				typeCount;
	long				lastResourceType;
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total;
	long				*pTypes;

	err = 0;
	typeCount = 0;
	lastResourceType = 0;
#if X_PLATFORM == X_MACINTOSH
	if (PV_IsAnyOpenResourceFiles() == FALSE)
	{
		// use native resource manager
		return Count1Types() - 1;
	}
#endif
	if (PV_IsAnyOpenResourceFiles())
	{
		pTypes = (long *)XNewPtr((sizeof(long) * MAX_XFILE_SCAN_TYPES));
		if (pTypes)
		{
			pReference = (XFILENAME *)fileRef;
			if (PV_XFileValid(fileRef))
			{
				XFileSetPosition(fileRef, 0L);		// at start
				if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
				{
					if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
					{
						next = sizeof(XFILERESOURCEMAP);
						total = XGetLong(&map.totalResources);
						for (count = 0; (count < total) && (err == 0); count++)
						{
							err = XFileSetPosition(fileRef, next);		// at start
							if (err == 0)
							{
								err = XFileRead(fileRef, &next, (long)sizeof(long));		// get next pointer
								next = XGetLong(&next);
								if (next != -1L)
								{
									err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
									lastResourceType = XGetLong(&data);
									if (typeCount < MAX_XFILE_SCAN_TYPES)
									{
										if (PV_CheckForTypes(pTypes, typeCount, lastResourceType) == FALSE)
										{
											pTypes[typeCount] = lastResourceType;
											typeCount++;
										}
									}
									else
									{
										err = -5;
										break;
									}
								}
								else
								{
									err = -4;
									//DebugStr("\pNext offset is bad");
									break;
								}
							}
							else
							{
								err = -3;
								//DebugStr("\pCan't set next position");
								break;
							}
						}
					}
				}
			}
			XDisposePtr((XPTR)pTypes);
		}
	}
	return typeCount;
}

// Return the type from the file based upon an index of 0 to XCountTypes

//	bvk:  why doesn't this use the cache?

long XGetIndexedType(XFILE fileRef, long resourceIndex)
{

	long				typeCount;
	long				lastResourceType;
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total;
	long				*pTypes;

#if X_PLATFORM == X_MACINTOSH
	if (PV_IsAnyOpenResourceFiles() == FALSE)
	{
		// use native resource manager
		ResType	type;

		Get1IndType(&type, resourceIndex + 1);
		return type;
	}
#endif
	err = 0;
	typeCount = 0;
	lastResourceType = 0;
	pTypes = (long *)XNewPtr((sizeof(long) * MAX_XFILE_SCAN_TYPES));
	if (pTypes)
	{
		pReference = (XFILENAME *)fileRef;
		if (PV_XFileValid(fileRef))
		{
			XFileSetPosition(fileRef, 0L);		// at start
			if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
			{
				if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
				{
					next = sizeof(XFILERESOURCEMAP);
					total = XGetLong(&map.totalResources);
					for (count = 0; (count < total) && (err == 0); count++)
					{
						err = XFileSetPosition(fileRef, next);		// at start
						if (err == 0)
						{
							err = XFileRead(fileRef, &next, (long)sizeof(long));		// get next pointer
							next = XGetLong(&next);
							if (next != -1L)
							{
								err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
								lastResourceType = XGetLong(&data);
								if (typeCount < MAX_XFILE_SCAN_TYPES)
								{
									if (PV_CheckForTypes(pTypes, typeCount, lastResourceType) == FALSE)
									{
										pTypes[typeCount] = lastResourceType;
										if (typeCount == resourceIndex)
										{
											break;
										}
										else
										{
											typeCount++;
										}
									}
								}
								else
								{
									err = -5;
									break;
								}
							}
							else
							{
								err = -4;
								//DebugStr("\pNext offset is bad");
								break;
							}
						}
						else
						{
							err = -3;
							//DebugStr("\pCan't set next position");
							break;
						}
					}
				}
			}
		}
		XDisposePtr((XPTR)pTypes);
	}
	return lastResourceType;
}

// Returns TRUE if the file is read only

static XBOOL PV_IsXFileLocked(XFILE fileRef)
{
	XFILENAME			*pReference;

	if (PV_XFileValid(fileRef))
	{
		pReference = (XFILENAME *)fileRef;
		return pReference->readOnly;
	}
	return FALSE;
}


#if USE_CREATION_API == TRUE
static int PV_CopyWithinFile( XFILE fileRef, long offsetIn, long offsetOut, long size,
							void *buffer, long bufferSize)
{
	long	xferSize, bytesTransferred = 0;
	long	offset=0, err=0;

	if (buffer == NULL)
	{
		//GACK!
		return -1;
	}

	while (bytesTransferred < size)
	{
		xferSize = XMIN(bufferSize, (size - bytesTransferred));

		//Read data
		err = XFileSetPosition(fileRef, offsetIn + offset);
		if (err == 0)
		{
			err = XFileRead(fileRef, buffer, xferSize );
			if (err == 0)
			{
				//Write data
				err = XFileSetPosition(fileRef, offsetOut + offset);
				if (err == 0)
				{
					err = XFileWrite(fileRef, buffer, xferSize );
					if (err != 0)
					{
						err = -5;
						break;
					}
				}
				else
				{
					err = -4;
					break;
				}
			}
			else
			{
				err = -3;
				break;
			}

			bytesTransferred += xferSize;
			offset += xferSize;
		}
		else
		{
			err = -2;
			break;
		}
	}
	return err;
}
#endif	// USE_CREATION_API == TRUE



// bvk - NEW for Beatnik Windows
//
//	Scans the file for all resources of 'TRSH' type, and rewrites the file.
//	It does this by scanning through the file, maintaining two sets of offsets.
//	When a 'TRSH' resource is found, the function skips the resource and starts
//	copying over the trashed resource in place.  At the end, the file size is
//	reset and the appropriate header data changed.
//
//	In essence, it is a combination of XFileAddResource and XFileGetResource.
//

#define XFER_BUFFER_SIZE	1024

#if USE_CREATION_API == TRUE
XBOOL XCleanResourceFile( XFILE fileRef )
{
	XFILENAME			*pReference;
	long				err=0;
	XFILERESOURCEMAP	map;
	long				data, nextIn, nextOut, resType, resID, totalResSize;
	long				count, inResTotal, outResTotal;
	long				resDataSize, fileSize, resInStart, resDataIn, resDataOut;
	char				pName[256];
	XBOOL				isCompacting;
	void				*fileBuffer = NULL;

#if X_PLATFORM == X_MACINTOSH
	if (PV_IsAnyOpenResourceFiles() == FALSE)
	{
		// use native resource manager
		UpdateResFile(CurResFile());
		return TRUE;
	}
#endif
	isCompacting = FALSE;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		// first, do we have memory data?
		if (pReference->pResourceData)
		{
			return FALSE;
		}
		// second, are we a read only file?
		else if (PV_IsXFileLocked(fileRef))
		{
			return FALSE;
		}
		else
		{
			fileSize = XFileGetLength(fileRef);
			XFileSetPosition(fileRef, 0L);		// at start
			if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
			{
				if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
				{
					nextOut = nextIn = sizeof(XFILERESOURCEMAP);
					inResTotal = XGetLong(&map.totalResources);
					outResTotal = inResTotal;
					for (count = 0; (count < inResTotal) && (err == 0); count++)
					{
						err = XFileSetPosition(fileRef, nextIn);		// at start of input resource
						if (err != 0)
						{
							err = -4;
							break;
						}
						resInStart = nextIn;
						err = XFileRead(fileRef, &nextIn, (long)sizeof(long));		// get next input pointer
						if (err != 0)
						{
							err = -5;
							break;
						}
						nextIn = XGetLong(&nextIn);
						totalResSize = nextIn - resInStart;
						if (nextIn != -1L)
						{
							err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
							if (err != 0)
							{
								err = -6;
								break;
							}
							resType = XGetLong(&data);
							if (resType == XFILETRASH_ID)
							{
								//Since we have skipped a trash resource,
								//now we must copy/compact every resource that follows!
								if (isCompacting == FALSE)
								{
									isCompacting = TRUE;
									// do we have a cache?
									if (pReference->pCache)
									{
										//if we do, REMOVE IT.  It will be invalid
										XDisposePtr((XPTR)pReference->pCache);
										pReference->pCache = NULL;
									}

									fileBuffer = XNewPtr(XFER_BUFFER_SIZE);
									if (fileBuffer == NULL)
									{
										err = -7;
										break;
									}
								}
								//Note that the outputBuffer pointer (nextOut) is NOT being updated!
								fileSize -= totalResSize;
								outResTotal--;
							}
							else
							{
								//This wasn't a trash resource, so move the output pointer
								//and/or copy the resource!
								if (isCompacting)
								{
									//Get remainder of input info
									err = XFileRead(fileRef, &data, (long)sizeof(long));		// get ID
									if (err != 0)
									{
										err = -8;
										break;
									}
									resID = XGetLong(&data);
									err = XFileRead(fileRef, &pName[0], 1L);		// get name
									if (pName[0])
									{
										err = XFileRead(fileRef, &pName[1], (long)pName[0]);
									}
									err = XFileRead(fileRef, &data, (long)sizeof(long));		// get length
									if (err != 0)
									{
										err = -9;
										break;
									}
									resDataSize = XGetLong(&data);		// get resource size
									resDataIn = XFileGetPosition(fileRef);	// get current pos

									//reposition to start of output block
									err = XFileSetPosition(fileRef, nextOut);		// at start
									if (err != 0)
									{
										err = -10;
										break;
									}

									nextOut += totalResSize;	//move output pointer

									//write out information data
									XPutLong(&data, nextOut);
									err = XFileWrite(fileRef, &data, (long)sizeof(long));
									if (err != 0)
									{
										err = -16;
										break;
									}
									XPutLong(&data, resType);
									err = XFileWrite(fileRef, &data, (long)sizeof(long));		// put type
									if (err != 0)
									{
										err = -11;
										break;
									}
									XPutLong(&data, resID);
									err = XFileWrite(fileRef, &data, (long)sizeof(long));		// put ID
									if (err != 0)
									{
										err = -12;
										break;
									}
									err = XFileWrite(fileRef, pName, (long)(((char *)pName)[0])+1L);		// put name
									if (err != 0)
									{
										err = -13;
										break;
									}
									XPutLong(&data, resDataSize);
									err = XFileWrite(fileRef, &data, (long)sizeof(long));				// put length
									if (err != 0)
									{
										err = -14;
										break;
									}
									resDataOut = XFileGetPosition(fileRef);	// get current pos

									//Since we are filling for deleted resources, copy this
									//res data back into this file over old data!
									err = PV_CopyWithinFile( fileRef,
														resDataIn, resDataOut, resDataSize,
														fileBuffer, XFER_BUFFER_SIZE );
									if (err != 0)
									{
										err = -15;
										break;
									}
								}
								else
								{
									nextOut += totalResSize;	//move output pointer
								}
							}
						}
						else
						{
							err = -3;
							//DebugStr("\pCan't set next position");
							break;
						}
					}
				}
				else
				{
					err = -3;
				}
			}
			else
			{
				err = -2;
			}
		}
	}
	else
	{
		err = -1;
	}

	if (isCompacting && err == 0)
	{
		XFileSetLength(fileRef, fileSize);
		XDisposePtr(fileBuffer);

		//Update the resource map
		XFileSetPosition(fileRef, 0L);		// at start
		XPutLong(&map.totalResources, outResTotal);
		err = XFileWrite(fileRef, &map, (long)sizeof(XFILERESOURCEMAP));

		pReference->pCache = XCreateAccessCache(fileRef);
	}

	if (fileBuffer)
	{
		XDisposePtr(fileBuffer);
		fileBuffer = NULL;
	}
	return (err == 0) ? TRUE : FALSE;
}
#endif	// USE_CREATION_API == TRUE


// bvk - NEW for Beatnik Windows
//
//	Marks the selected resource as a 'TRSH' type w/ID of 0.  If collectTrash is TRUE,
//	writes out the file IN PLACE to remove dead space.  Does not work on read only or
//	memory mapped files.
//
//	Not the fastest thing in the world, but none of the resource functions are!
//
#if USE_CREATION_API == TRUE
XBOOL XDeleteFileResource(XFILE fileRef, long resourceType, long resourceID, XBOOL collectTrash )
{
	XFILENAME			*pReference;
	long				err=0;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total;
	long				whereType, whereID;
	XFILE_CACHED_ITEM	*pCachedItem;

#if X_PLATFORM == X_MACINTOSH
	if (PV_IsAnyOpenResourceFiles() == FALSE)
	{
		Handle	theData;

		// use native resource manager
		theData = Get1Resource(resourceType, resourceID);
		if (theData)
		{
			RemoveResource(theData);
		}
		return TRUE;
	}
#endif

	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
		// first, do we have memory data?
		if (pReference->pResourceData)
		{
			return FALSE;
		}
		// second, are we a read only file?
		else if (PV_IsXFileLocked(fileRef))
		{
			return FALSE;
		}

		// do we have a cache?
		if (pReference->pCache)
		{
			pCachedItem = PV_XGetCacheEntry(fileRef, resourceType, resourceID);
			if (pCachedItem)
			{
				pCachedItem->resourceType = XFILETRASH_ID;
				pCachedItem->resourceID = 0;
				whereType = pCachedItem->fileOffsetName;
				whereType -= ( sizeof(resourceType) + sizeof(resourceID) );
				err = XFileSetPosition(fileRef, whereType );
				if (err != -1)
				{
					XPutLong(&data, XFILETRASH_ID);
					err = XFileWrite(fileRef, &data, (long)sizeof(long));						// put type

					if (err == 0)
					{
						XPutLong(&data, 0);
						err = XFileWrite(fileRef, &data, (long)sizeof(long));					// put ID
					}
				}
				else
				{
					err = -7;
				}
			}
			else
			{
				err = -6;
			}
		}
		else
		{
			XFileSetPosition(fileRef, 0L);		// at start
			if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
			{
				if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
				{
					next = sizeof(XFILERESOURCEMAP);
					total = XGetLong(&map.totalResources);
					for (count = 0; (count < total) && (err == 0); count++)
					{
						err = XFileSetPosition(fileRef, next);		// at start
						if (err == 0)
						{
							err = XFileRead(fileRef, &next, (long)sizeof(long));		// get next pointer
							next = XGetLong(&next);
							if (next != -1L)
							{
								whereType = XFileGetPosition(fileRef);	// get current pos

								err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
								if (XGetLong(&data) == resourceType)
								{
									whereID = XFileGetPosition(fileRef);	// get current pos

									err = XFileRead(fileRef, &data, (long)sizeof(long));		// get ID
									if (XGetLong(&data) == resourceID)
									{
										//We found it!
										//Now 'TRASH' everything
										//td - added to properly overwrite existing id
										err = XFileSetPosition(fileRef, whereType);

										XPutLong(&data, XFILETRASH_ID);
										err = XFileWrite(fileRef, &data, (long)sizeof(long));						// put type

										if (err == 0)
										{
											XPutLong(&data, 0);
											err = XFileWrite(fileRef, &data, (long)sizeof(long));					// put ID
											break;
										}
										else
										{
											err = -5;
											break;
										}
									}
								}
							}
							else
							{
								err = -4;
								//DebugStr("\pNext offset is bad");
								break;
							}
						}
						else
						{
							err = -3;
							//DebugStr("\pCan't set next position");
							break;
						}
					}
				}
			}
		}
	}

	if (collectTrash == TRUE)
	{
		XCleanResourceFile( fileRef );
	}
	return (err == 0);
}
#endif	// USE_CREATION_API == TRUE


// bvk - NEW for Beatnik Windows
//	For a given resource type, counts the number of resources of a type.
//	Does not check for duplicate IDs.
#if USE_CREATION_API == TRUE
long XCountFileResourcesOfType(XFILE fileRef, long int theType)
{
	long				resCount;
	long				resourceType;
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total;
	XFILERESOURCECACHE	*pCache;
	XFILE_CACHED_ITEM	*pCacheItem;

	err = 0;
	resCount = 0;
	resourceType = 0;
#if X_PLATFORM == X_MACINTOSH
	if (PV_IsAnyOpenResourceFiles() == FALSE)
	{
		return Count1Resources(theType);
	}
#endif
	pReference = (XFILENAME *)fileRef;
	if (PV_IsAnyOpenResourceFiles())
	{
		if (PV_XFileValid(fileRef))
		{
			if (pReference->pCache)
			{
				pCache = pReference->pCache;

				for (count = 0; count < pCache->totalResources; count++)
				{
					pCacheItem = &pCache->cached[count];

					if (theType == pCacheItem->resourceType)
					{
						resCount++;
					}
				}
			}
			else
			{
				XFileSetPosition(fileRef, 0L);		// at start
				if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
				{
					if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
					{
						next = sizeof(XFILERESOURCEMAP);
						total = XGetLong(&map.totalResources);
						for (count = 0; (count < total) && (err == 0); count++)
						{
							err = XFileSetPosition(fileRef, next);		// at start
							if (err == 0)
							{
								err = XFileRead(fileRef, &next, (long)sizeof(long));		// get next pointer
								next = XGetLong(&next);
								if (next != -1L)
								{
									err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
									resourceType = XGetLong(&data);
									if (resourceType == theType )
									{
										resCount++;
									}
								}
								else
								{
									err = -4;
									//DebugStr("\pNext offset is bad");
									break;
								}
							}
							else
							{
								err = -3;
								//DebugStr("\pCan't set next position");
								break;
							}
						}
					}
				}
			}
		}
	}
	return resCount;
}
#endif	// USE_CREATION_API == TRUE


// search through open resource files until the resourceType is found based upon the resource index
XPTR XGetIndexedResource(long resourceType, long *pReturnedID, long resourceIndex,
								void *pName, long *pReturnedResourceSize)
{
	long	count;
	XPTR	pData;

	pData = NULL;
	if (PV_IsAnyOpenResourceFiles())
	{
		for (count = 0; count < resourceFileCount; count++)
		{
			pData = XGetIndexedFileResource(openResourceFiles[count], resourceType,
									pReturnedID, resourceIndex, pName, pReturnedResourceSize);
			if (pData)
			{
				break;
			}
		}
	}
#if X_PLATFORM == X_MACINTOSH
	if ((pData == NULL) && (PV_IsAnyOpenResourceFiles() == FALSE))
	{
		// use native resource manager
		ResType		type;
		Handle		data;
		short		shortID;
		char		pPName[256];
		long		total;

		data = Get1IndResource(resourceType, resourceIndex + 1);
		if (data)
		{
			total = GetHandleSize(data);
			if (pReturnedResourceSize)
			{
				*pReturnedResourceSize = total;
			}
			pData = XNewPtr(total);
			if (pData)
			{
				HLock(data);
				XBlockMove(*data, pData, total);
				HUnlock(data);

				GetResInfo(data, &shortID, &type, (unsigned char *)pPName);
				if (pReturnedID)
				{
					*pReturnedID = shortID;
				}
				if (pName)
				{
					XBlockMove(pPName, pName, pPName[0]+1);
				}
				ReleaseResource(data);
			}
		}
	}
#endif
	return pData;
}

// Get a resource based upon its entry count into the resource file. Return NULL if resourceIndex
// is out of range
XPTR XGetIndexedFileResource(XFILE fileRef, long resourceType, long *pReturnedID,
								long resourceIndex, void *pName, long *pReturnedResourceSize)
{
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data, next;
	long				count, total, typeCount;
	XPTR				pData;
	char				pPName[256];
	XFILERESOURCECACHE	*pCache;
	XFILE_CACHED_ITEM	*pCacheItem;

	pData = NULL;
	err = 0;
	if (pReturnedResourceSize)
	{
		*pReturnedResourceSize = 0;
	}
	pPName[0] = 0;
	pReference = (XFILENAME *)fileRef;
	typeCount = 0;
	if (PV_XFileValid(fileRef))
	{
		if (pReference->pCache)
		{
			pCache = pReference->pCache;

			for (count = 0; count < pCache->totalResources; count++)
			{
				pCacheItem = &pCache->cached[count];

				if (pCacheItem->resourceType == resourceType)
				{
					if (resourceIndex == typeCount)
					{
						*pReturnedID = pCacheItem->resourceID;

						XFileSetPosition(fileRef, pCacheItem->fileOffsetName);
						err = XFileRead(fileRef, &pPName[0], 1L);		// get name length
						if (pPName[0])
						{
							err = XFileRead(fileRef, &pPName[1], (long)pPName[0]);
						}
						// get data
						XFileSetPosition(fileRef, pCacheItem->fileOffsetData);

						// is data memory based?
						if (pReference->pResourceData && (pReference->allowMemCopy == FALSE) )
						{	// don't bother coping data again, since its already in memory
							pData = PV_GetFilePositionFromMemoryResource(fileRef);
							if (pData)
							{
								if (pReturnedResourceSize)
								{
									*pReturnedResourceSize = pCacheItem->resourceLength;
								}
							}
							else
							{
								err = -2;
								//DebugStr("\pOut of memory; can't allocate resource");
							}
						}
						else
						{
							pData = XNewPtr(pCacheItem->resourceLength);
							if (pData)
							{
								err = XFileRead(fileRef, pData, pCacheItem->resourceLength);
								if (pReturnedResourceSize)
								{
									*pReturnedResourceSize = XGetLong(&data);
								}
								break;
							}
							else
							{
								err = -2;
								//DebugStr("\pOut of memory; can't allocate resource");
								break;
							}
						}
					}
					typeCount++;
				}
			}
		}
		else
		{
			XFileSetPosition(fileRef, 0L);		// at start
			if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
			{
				if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
				{
					next = sizeof(XFILERESOURCEMAP);
					total = XGetLong(&map.totalResources);
					for (count = 0; (count < total) && (err == 0); count++)
					{
						err = XFileSetPosition(fileRef, next);		// at start
						if (err == 0)
						{
							err = XFileRead(fileRef, &next, (long)sizeof(long));		// get next pointer
							next = XGetLong(&next);
							if (next != -1L)
							{
								err = XFileRead(fileRef, &data, (long)sizeof(long));		// get type
								if (XGetLong(&data) == resourceType)
								{
									if (resourceIndex == typeCount)
									{
										err = XFileRead(fileRef, pReturnedID, (long)sizeof(long));		// get ID
										*pReturnedID = XGetLong(pReturnedID);
										err = XFileRead(fileRef, &pPName[0], 1L);		// get name length
										if (pPName[0])
										{
											err = XFileRead(fileRef, &pPName[1], (long)pPName[0]);
										}
										err = XFileRead(fileRef, &data, (long)sizeof(long));		// get length
										data = XGetLong(&data);		// get resource size
										// get data
										// is data memory based?
										if (pReference->pResourceData && (pReference->allowMemCopy == FALSE) )
										{	// don't bother coping data again, since its already in memory
											pData = PV_GetFilePositionFromMemoryResource(fileRef);
											if (pData)
											{
												if (pReturnedResourceSize)
												{
													*pReturnedResourceSize = data;
												}
												err = 0;
												break;
											}
											else
											{
												err = -2;
												//DebugStr("\pOut of memory; can't allocate resource");
											}
										}
										else
										{
											pData = XNewPtr(data);
											if (pData)
											{
												err = XFileRead(fileRef, pData, data);
												if (pReturnedResourceSize)
												{
													*pReturnedResourceSize = data;
												}
												break;
											}
											else
											{
												err = -2;
												//DebugStr("\pOut of memory; can't allocate resource");
												break;
											}
										}
									}
									typeCount++;
								}
							}
							else
							{
								err = -4;
								//DebugStr("\pNext offset is bad");
								break;
							}
						}
						else
						{
							err = -3;
							//DebugStr("\pCan't set next position");
							break;
						}
					}
				}
			}
		}
	}
	if (pName)
	{
		XBlockMove(pPName, pName, (long)pPName[0] + 1);
	}
	return pData;
}

#if USE_CREATION_API == TRUE
long XAddFileResource(XFILE fileRef, long resourceType, long resourceID, void *pName, void *pData, long length)
{
	XFILENAME			*pReference;
	long				err;
	XFILERESOURCEMAP	map;
	long				data;
	char				fakeName[2];
	long				next, nextsave;
	XFILE_CACHED_ITEM	cacheItem;

#if X_PLATFORM == X_MACINTOSH
	if (PV_IsAnyOpenResourceFiles() == FALSE)
	{
		// use native resource manager
		Handle	theData;

		theData = NewHandle(length);
		if (theData)
		{
			HLock(theData);
			XBlockMove(pData, *theData, length);
			HUnlock(theData);
			AddResource(theData, resourceType, resourceID, (unsigned char *)pName);
		}
		return 0;
	}
#endif
	err = -1;
	pReference = (XFILENAME *)fileRef;
	if (PV_XFileValid(fileRef))
	{
//		// the cache will be blown if we add a new file
//		if (pReference->pCache)
//		{
//			XDisposePtr((XPTR)pReference->pCache);
//			pReference->pCache = NULL;
//		}
		if (pData && length)
		{
			XFileSetPosition(fileRef, 0L);		// at start
			if (XFileRead(fileRef, &map, (long)sizeof(XFILERESOURCEMAP)) == 0)
			{
				if (XGetLong(&map.mapID) == XFILERESOURCE_ID)
				{
					XFileSetPosition(fileRef, XFileGetLength(fileRef));						// at end

					nextsave = XFileGetPosition(fileRef);		// save for later change
					next = -1;
					err = XFileWrite(fileRef, &next, (long)sizeof(long));

					XPutLong(&data, resourceType);
					err = XFileWrite(fileRef, &data, (long)sizeof(long));						// put type
					XPutLong(&cacheItem.resourceType, resourceType);

					if (err == 0)
					{
						XPutLong(&data, resourceID);
						err = XFileWrite(fileRef, &data, (long)sizeof(long));					// put ID
						XPutLong(&cacheItem.resourceID, resourceID);

						if (err == 0)
						{
							data = XFileGetPosition(fileRef);	// get name pos
							XPutLong(&cacheItem.fileOffsetName, data);

							if (pName)
							{
								err = XFileWrite(fileRef, pName, (long)(((char *)pName)[0])+1L);		// put name
							}
							else
							{
								fakeName[0] = 0;
								err = XFileWrite(fileRef, fakeName, 1L);		// put name
							}
							XPutLong(&data, length);
							err = XFileWrite(fileRef, &data, (long)sizeof(long));				// put length
							XPutLong(&cacheItem.resourceLength, length);

							if (err == 0)
							{
								err = XFileWrite(fileRef, pData, length);					// put data block

								next = XFileGetPosition(fileRef);	// get current pos

								if (err == 0)
								{
									XFileSetPosition(fileRef, 0L);		// at start
									data = XGetLong(&map.totalResources) + 1;
									XPutLong(&map.totalResources, data);
									err = XFileWrite(fileRef, &map, (long)sizeof(XFILERESOURCEMAP));

									data = XFileGetPosition(fileRef);	// get data pos
									XPutLong(&cacheItem.fileOffsetData, data);

									nextsave = XFileSetPosition(fileRef, nextsave);		// save for later change
									//bvk 06/19/97
									//You flip all of the other things, you should flip this!
									XPutLong(&data, next);	//bvk

									err = XFileWrite(fileRef, &data, (long)sizeof(long)); //bvk
									//bvk err = XFileWrite(fileRef, &next, (long)sizeof(long));

									// Now we add this to the cache!
									if (pReference->pCache)
									{
										PV_AddToAccessCache(fileRef, &cacheItem );
									}
								}
							}
						}
					}
				}
			}
		}
	}
	return err;
}
#endif	// USE_CREATION_API == TRUE


// get a resource type from the name.
XPTR XGetNamedResource(long resourceType, void *cName, long *pReturnedResourceSize)
{
	XPTR				pData;
	XFILE_CACHED_ITEM	*pCacheItem;
	char				pName[256];
	short int			count;

	pData = NULL;
	if (pReturnedResourceSize)
	{
		*pReturnedResourceSize = 0;
	}
#if X_PLATFORM == X_MACINTOSH
	{
		Handle	theData;
		long	size;

		// first look inside any open resource files
		if (PV_IsAnyOpenResourceFiles())
		{
			for (count = 0; count < resourceFileCount; count++)
			{
				pCacheItem = PV_XGetNamedCacheEntry(openResourceFiles[count], resourceType, cName);
				if (pCacheItem)
				{
					pData = XGetFileResource(openResourceFiles[count], pCacheItem->resourceType,
																	pCacheItem->resourceID,
																	pName, pReturnedResourceSize);
					// we found our resource
					break;
				}
			}
		}
		if (pData == NULL)
//		if ((pData == NULL) && (PV_IsAnyOpenResourceFiles() == FALSE))
		{
			// use native resource manager
			XStrCpy(pName, (char *)cName);
			XCtoPstr(pName);
			theData = Get1NamedResource(resourceType, (unsigned char *)pName);
			if (theData)
			{
				size = GetHandleSize(theData);
				pData = XNewPtr(size);
				if (pData)
				{
					HLock(theData);
					XBlockMove(*theData, pData, size);
					HUnlock(theData);
					ReleaseResource(theData);
					if (pReturnedResourceSize)
					{
						*pReturnedResourceSize = size;
					}
				}
			}
		}
	}
#else
	if (PV_IsAnyOpenResourceFiles())
	{
		for (count = 0; count < resourceFileCount; count++)
		{
			pCacheItem = PV_XGetNamedCacheEntry(openResourceFiles[count], resourceType, cName);
			if (pCacheItem)
			{
				pData = XGetFileResource(openResourceFiles[count], pCacheItem->resourceType,
																pCacheItem->resourceID,
																pName, pReturnedResourceSize);
				// we found our resource
				break;
			}
			else
			{
				// search through without cache. NEED TO BUILT THIS!!
			}
		}
	}
#endif
	return pData;
}

void XGetResourceName(long resourceType, long resourceID, void *cName)
{
	char		szPName[256];
	long		size = 0;
	XPTR		pData = NULL;
	short int	count;

	if (cName)
	{
		((char *)cName)[0] = 0;
#if X_PLATFORM == X_MACINTOSH
		// first look in any open resource files
		if (PV_IsAnyOpenResourceFiles())
		{
			for (count = 0; count < resourceFileCount; count++)
			{
				XGetResourceNameOnly(openResourceFiles[count], resourceType, resourceID, szPName);
//				pData = XGetFileResource(openResourceFiles[count], resourceType, resourceID, szPName, &size);
//				if (pData)
				{
//					XDisposePtr(pData);
					XStrCpy((char *)cName, szPName);
					// we found data
					break;
				}
			}
		}
		if (((char *)cName)[0] == 0)
		{	// use native resource manager
			short int		theID;
			unsigned long	theType;
			Handle			theData;

			SetResLoad(FALSE);
			theData = Get1Resource(resourceType, (short int)resourceID);
			SetResLoad(TRUE);
			if (theData)
			{
				GetResInfo(theData, &theID, &theType, (unsigned char *)cName);
				XPtoCstr(cName);
				DetachResource(theData);
				DisposeHandle(theData);
			}
		}
#else
		for (count = 0; count < resourceFileCount; count++)
		{
			XGetResourceNameOnly(openResourceFiles[count], resourceType, resourceID, szPName);
//			pData = XGetFileResource(openResourceFiles[count], resourceType, resourceID, szPName, &size);
//			if (pData)
			{
//				XDisposePtr(pData);
				XStrCpy((char *)cName, szPName);
				// we found data
//				break;
			}
		}
#endif
	}
}

// Get a resource and detach it from the resource manager. Which means that you'll need to call XDisposePtr
// to free the memory
XPTR XGetAndDetachResource(long resourceType, long resourceID, long *pReturnedResourceSize)
{
#if X_PLATFORM == X_MACINTOSH
	Handle		theData;
	XPTR		pData;
	long		size;
	char		szPName[256];
	short int	count;

	pData = NULL;
	if (pReturnedResourceSize)
	{
		*pReturnedResourceSize = 0;
	}
	// first look in any open resource files
	if (PV_IsAnyOpenResourceFiles())
	{
		for (count = 0; count < resourceFileCount; count++)
		{
			pData = XGetFileResource(openResourceFiles[count], resourceType, resourceID, szPName, &size);
			if (pData)
			{
				if (pReturnedResourceSize)
				{
					*pReturnedResourceSize = size;
				}
				// we found data
				break;
			}
		}
	}
	if (pData == NULL)
//	if ((pData == NULL) && (PV_IsAnyOpenResourceFiles() == FALSE))
	{	// use native resource manager
//		if (resourceType == ID_RMF)
//		{
//			Debugger();
//		}
		theData = Get1Resource(resourceType, (short int)resourceID);
		if (theData)
		{
			size = GetHandleSize(theData);
			pData = XNewPtr(size);
			if (pData)
			{
				HLock(theData);
				XBlockMove(*theData, pData, size);
				HUnlock(theData);
				ReleaseResource(theData);
				if (pReturnedResourceSize)
				{
					*pReturnedResourceSize = size;
				}
			}
		}
	}
	return pData;
#else	// X_PLATFORM == X_MACINTOSH
	#if X_PLATFORM == X_WEBTV
	char		fileName[32];
	char		type[10];
	FSNode		*romFile;
	XPTR		pData;

	switch (resourceType)
	{
		case ID_INST:
			XStrCpy(type, "INST");
			break;
		case ID_SND:
			XStrCpy(type, "SND");
			break;
		case ID_CSND:
			XStrCpy(type, "CSND");
			break;
		case ID_ESND:
			XStrCpy(type, "ESND");
			break;
		default:
			XBlockMove((char *)&resourceType, type, 4L);
			type[4] = 0;
			break;
	}
	// First we look for a RAM copy, if not there then we use the ROM versions
	sprintf(fileName, "/RAM/GM_Patches/%s_%ld", type, resourceID);
	romFile = Resolve(fileName, false);
	if (romFile == NULL)
	{
		sprintf(fileName, "/ROM/GM_Patches/%s_%ld", type, resourceID);
		romFile = Resolve(fileName, false);
	}
	pData = NULL;
	if (pReturnedResourceSize)
	{
		*pReturnedResourceSize = 0;
	}
	if (romFile)
	{
		// We just return a pointer and the size. Our dispose system tags our real memory blocks, so
		// nothing should happen if we try to dispose of a rom based pointer.
		if (pReturnedResourceSize)
		{
			*pReturnedResourceSize = romFile->dataLength;
		}
		pData = romFile->data;
	}
	return pData;
	#else
	char		szPName[256];
	long		lSize = 0;
	XPTR		pData = NULL;
	short int	count;

	for (count = 0; count < resourceFileCount; count++)
	{
		pData = XGetFileResource(openResourceFiles[count], resourceType, resourceID, szPName, &lSize);
		if (pData)
		{
			if (pReturnedResourceSize)
			{
				*pReturnedResourceSize = lSize;
			}
			// we found data
			break;
		}
	}
	return pData;
	#endif	// X_PLATFORM == X_WEBTV
#endif	//	X_PLATFORM == X_MACINTOSH
}

// make sure this resource file is first in the scan list
void XFileUseThisResourceFile(XFILE fileRef)
{
	short int	fileCount;
	XFILE		currentFirst;

	if (PV_XFileValid(fileRef))
	{
		fileCount = PV_FindResourceFileReferenceIndex(fileRef);
		if (fileCount != -1)
		{
			currentFirst = openResourceFiles[0];
			openResourceFiles[0] = fileRef;
			openResourceFiles[fileCount] = currentFirst;
		}
	}
}

// theVolume is in the range of 0 to X_FULL_VOLUME
void XSetHardwareVolume(short int theVolume)
{
#if USE_HAE_EXTERNAL_API == TRUE
	HAE_SetHardwareVolume(theVolume);
#else
	#if X_PLATFORM == X_MACINTOSH
	long	newVolume;

	if (theVolume < 0)
	{
		theVolume = 0;
	}
	if (theVolume > X_FULL_VOLUME)
	{
		theVolume = X_FULL_VOLUME;
	}

/* Scale volume to match the Macintosh hardware
*/
	newVolume = theVolume;
	newVolume |= (newVolume << 16L);
	SetDefaultOutputVolume(newVolume);
	#else
	theVolume;
	#endif
#endif
}

// returned volume is in the range of 0 to X_FULL_VOLUME
short int XGetHardwareVolume(void)
{
#if USE_HAE_EXTERNAL_API == TRUE
	return HAE_GetHardwareVolume();
#else
	#if X_PLATFORM == X_MACINTOSH
	short	theRealVolume;
	long	theLongRealVolume;

	GetDefaultOutputVolume(&theLongRealVolume);
	theRealVolume = theLongRealVolume & 0xFFFFL;
	return theRealVolume;
	#else
	return 0;
	#endif
#endif
}

// Decompress a 'csnd' format sound.
// First byte is a type.
// Next three bytes are a length.
// Type 0 is Delta LZSS compression
void * XDecompressSampleFormatPtr(void * pData, long dataSize)
{
	long	theTotalSize;
	char	theType;
	XPTR	theNewData;

	theNewData = NULL;
	if (pData && dataSize)
	{
		theTotalSize = XGetLong(pData);
		theType = (char)(theTotalSize >> 24L);		// get type
		theTotalSize &= 0x0FFFFFFFL;
		theNewData = XNewPtr(theTotalSize);
		if (theNewData)
		{
			switch (theType)
			{
				case 0:
					LZSSDeltaUncompress((unsigned char *)(char *)(pData) + sizeof(long), dataSize - sizeof(long),
									(unsigned char *)theNewData, &theTotalSize);
					break;
				default:
					XDisposePtr(theNewData);
					theNewData = NULL;
					break;
			}
		}
	}
	return theNewData;
}

void * XDecompressPtr(void * pData, long dataSize)
{
	long theTotalSize;
	XPTR theNewData;

	theNewData = NULL;
	if (pData && dataSize)
	{
		theTotalSize = XGetLong(pData);
		theNewData = XNewPtr(theTotalSize);
		if (theNewData)
		{
			LZSSUncompress((unsigned char *)pData + sizeof(long), dataSize - sizeof(long),
							(unsigned char *)theNewData, &theTotalSize);
		}
	}
	return theNewData;
}

// Given a block of data and a size, this will compress it and return a new pointer. The
// original pointer is not deallocated, and the new pointer must be deallocated when finished.
// Will return NULL if cannot compress. First 4 bytes is uncompressed length
#if USE_CREATION_API == TRUE
void * XCompressPtr(void *pData, long dataSize, long *pNewSize, XBOOL useDelta)
{
	long	newSize;
	XPTR	compressedData;
	char	*realData;

	realData = NULL;
	newSize = -1;
	if (pData && dataSize)
	{
		compressedData = XNewPtr(dataSize);
		if (compressedData)
		{
			if (useDelta)
			{
				LZSSDeltaCompress((unsigned char *)pData, dataSize, (unsigned char *)compressedData, &newSize);
			}
			else
			{
				LZSSCompress((unsigned char *)pData, dataSize, (unsigned char *)compressedData, &newSize);
			}
			if (newSize == -1L)
			{	// failed to compress
				XDisposePtr(compressedData);
			}
			else
			{
				newSize += sizeof(long);
				realData = (char *)XNewPtr(newSize);
				if (realData)
				{
					XPutLong(realData, dataSize);
					XBlockMove(compressedData, realData + sizeof(long), newSize - sizeof(long));
				}
			}
		}
	}
	if (pNewSize)
	{
		*pNewSize = newSize;
	}
	return realData;
}
#endif	// USE_CREATION_API == TRUE


char * XDuplicateStr(char *src)
{
	char *dup;

	dup = NULL;
	if (src)
	{
		dup = (char *)XNewPtr(XStrLen(src)+1);
		if (dup)
		{
			XStrCpy(dup, src);
		}
	}
	return dup;
}

// Duplicate and string characters below 32
char * XDuplicateAndStripStr(char *src)
{
	long	length;									// must be signed
	char	*cStrippedString, *cSource, *cDest;

	cStrippedString = NULL;
	// strip out undesirable characters, give them some walking money
	length = XStrLen(src);
	if (length)
	{
		cStrippedString = (char *)XNewPtr(length+1);
		if (cStrippedString)
		{
			cDest = cStrippedString;
			cSource = src;
			while (*cSource)
			{
				if (*cSource >= 32)
				{
					*cDest++ = *cSource;
				}
				cSource++;
			}
			*cDest = 0;
		}
	}
	return cStrippedString;
}

// standard strcpy
// Copies 'C' string src into dest
char * XStrCpy(char *dest, char *src)
{
	char *sav;

	sav = dest;
	if (src == NULL)
	{
		src = "";
	}
	if (dest)
	{
		while (*src)
		{
			*dest++ = *src++;
		}
		*dest = 0;
	}
	return sav;
}



char * XStrStr(char *source, char *pattern)
{
	unsigned char * s1;
	unsigned char * p1;
	unsigned char firstc, c1, c2;

	if (source == NULL)
	{
		source = "";
	}
	if (pattern == NULL)
	{
		pattern = "";
	}
	s1 = (unsigned char *)source;
	p1 = (unsigned char *)pattern;

	if (!(firstc = *p1++))
	{
		return((char *) s1);
	}
	while((c1 = *s1) != 0)
	{
		s1++;
		if (c1 == firstc)
		{
			const unsigned char * s2 = s1;
			const unsigned char * p2 = p1;

			while ((c1 = *s2++) == (c2 = *p2++) && c1) {};

			if (!c2)
			{
				return((char *) s1 - 1);
			}
		}
	}
	return(NULL);
}

short int XStrLen(char *src)
{
	short int len;

	len = -1;
	if (src == NULL)
	{
		src = "";
	}
	do
	{
		len++;
	} while  (*src++);

	return len;
}

// standard strcmp
short int XStrCmp(const char *s1, const char *s2)
{
	if (s1 == NULL)
	{
		s1 = "";
	}
	if (s2 == NULL)
	{
		s2 = "";
	}

	while (1)
	{
		if (*s1 == *s2)
		{
			if (*s1 == 0)
			{
				return 0;
			}
			else
			{
				s1++;
				s2++;
			}
		}
		else if (*s1 > *s2)
		{
			return 1;
		}
		else
		{
			return -1;
		}
	}
}

static short int PV_LowerCase(short int c)
{
	return( ( ((c >= 'A') && (c <= 'Z')) ? c | 0x20 : c) );
}

// standard strcmp, but ignore case
short int XLStrCmp(const char *s1, const char *s2)
{
	if (s1 == NULL)
	{
		s1 = "";
	}
	if (s2 == NULL)
	{
		s2 = "";
	}

	while (1)
	{
		if (PV_LowerCase(*s1) == PV_LowerCase(*s2))
		{
			if (*s1 == 0)
			{
				return 0;
			}
			else
			{
				s1++;
				s2++;
			}
		}
		else if (PV_LowerCase(*s1) > PV_LowerCase(*s2))
		{
			return 1;
		}
		else
		{
			return -1;
		}
	}
}

// Standard strncmp, but ignore case. Compares zero terminated s1 with non zero terminated s2 with s2 having
// a length of n
short int XLStrnCmp(register const char *s1, register const char *s2, register long n)
{
	if (s1 == NULL)
	{
		s1 = "";
	}
	if (s2 == NULL)
	{
		s2 = "";
	}

	if (n)
	{
		do
		{
			if (PV_LowerCase(*s1) != PV_LowerCase(*s2++))
			{
				return (*(unsigned char *)s1 - *(unsigned char *)--s2);
			}
			if (*s1++ == 0)
			{
				break;
			}
		} while (--n != 0);
	}
	return 0;
}
// Standard strncmp. Compares zero terminated s1 with non zero terminated s2 with s2 having
// a length of n
short int XStrnCmp(register const char *s1, register const char *s2, register long n)
{
	if (s1 == NULL)
	{
		s1 = "";
	}
	if (s2 == NULL)
	{
		s2 = "";
	}

	if (n)
	{
		do
		{
			if (*s1 != *s2++)
			{
				return (*(unsigned char *)s1 - *(unsigned char *)--s2);
			}
			if (*s1++ == 0)
			{
				break;
			}
		} while (--n != 0);
	}
	return 0;
}

// This will convert a string to a base 10 long value
long XStrnToLong(char *pData, long length)
{
	long	result, num, count;
	char	data[12];

	result = 0;
	num = 0;
	for (count = 0; count < length; count++)
	{
		if (*pData != 0x20)
		{
			if ( ((*pData) >= '0') && ((*pData) <= '9') )
			{
				data[num++] = *pData;
				if (num > 11)
				{
					break;
				}
			}
			else
			{
				break;
			}
		}
		pData++;
	}
	if (num)
	{
		for (count = 0; count < num; count++)
		{
			result *= 10;
			result += data[count] - '0';
		}
	}
	return result;
}



short int XMemCmp(const void * src1, const void * src2, long n)
{
	const	unsigned char * p1;
	const	unsigned char * p2;


	for (p1 = (const unsigned char *) src1, p2 = (const unsigned char *) src2, n++; --n;)
	{
		if (*p1++ != *p2++)
		{
			return((*--p1 < *--p2) ? -1 : +1);
		}
	}
	return 0;
}



#if USE_CREATION_API == TRUE
/*
 *	pseudo-random number generator
 *
 */
static unsigned long seed = 1;

// return a pseudo-random number in the range of 0-32767
short int XRandom(void)
{
	seed = seed * 1103515245 + 12345;

	return (short int)(seed & 0x7FFFL);		// high word of long, remove high bit
}


/*
 *  seed pseudo-random number generator
 *
 */

void XSeedRandom(unsigned long n)
{
	seed = n;
}

short int XRandomRange(short int max)
{
	static char	setup = 0;

	if (setup == 0)
	{
		XSeedRandom(XMicroseconds());
		setup = 1;
	}
	return XRandom() % max;
}
#endif	// USE_CREATION_API

#if USE_FULL_RMF_SUPPORT == TRUE
/****************************************************************************
**
** XIsWinInMac() determines whether a Macintosh character has an
** equivalent in the Windows character set.
**
** XIsMacInWin() determines whether a Windows character has an
** equivalent in the Macintosh character set.
**
** XTranslateWinToMac() provides the Macintosh-equivalent of a Windows
** character code.
**
** XTranslateMacToWin() provides the Windows-equivalent of a Macintosh
** character code.
**
*****************************************************************************/
////////////////////////////////////////////////// DATA:

static unsigned char macToWinTable[128] =
{
	0xC4,	// A-umlaut
	0xC5,	// A-circle
	0xC7,	// C-cedire
	0xC9,	// E-acute
	0xD1,	// N-tilde
	0xD6,	// O-umlaut
	0xDC,	// U-umlaut
	0xE1,	// a-acute
	0xE0,	// a-grave
	0xE2,	// a-circumflex
	0xE4,	// a-umlaut
	0xE3,	// a-tilde
	0xE5,	// a-circle
	0xE7,	// c-cedire
	0xE9,	// e-acute
	0xE8,	// e-grave

	0xEA,	// e-circumflex
	0xEB,	// e-umlaut
	0xED,	// i-acute
	0xEC,	// i-grave
	0xEE,	// i-circumflex
	0xEF,	// i-umlaut
	0xF1,	// n-tilde
	0xF3,	// o-acute
	0xF2,	// o-grave
	0xF4,	// o-circumflex
	0xF6,	// o-umlaut
	0xF5,	// o-tilde
	0xFA,	// u-acute
	0xF9,	// u-grave
	0xFB,	// u-circumflex
	0xFC,	// u-umlaut

	0x86,	// little up arrow (dagger substituted)
	0xB0,	// degrees
	0xA2,	// cents
	0xA3,	// pounds
	0xA7,	// section
	0x95,	// bullet
	0xB6,	// paragraph
	0xDF,	// german double-s
	0xAE,	// registered
	0xA9,	// copyright (same in both sets)
	0x99,	// tm
	0xB4,	// acute
	0xA8,	// umlaut
	0xB1,	// not equals (plus-minus substituted)
	0xC6,	// AE
	0xD8,	// O-slash

	0xBF,	// infinity (upsidedown-? substituted)
	0xB1,	// plus-minus
	'<',	// less-than or equal-to
	'>',	// greater-than or equal-to
	0xA5,	// yen
	0xB5,	// mu
	'd',	// delta
	'S',	// SIGMA
	'P',	// PI
	'p',	// pi
	0x83,	// f without bar (fi substituted)
	0xAA,	// small a with bar
	0xBA,	// small o with bar
	'O',	// omega
	0xE6,	// ae
	0xF8,	// o-slash

	0xBF,	// upsidedown-?
	0xA1,	// upsidedown-!
	0xAC,	// rho(?)
	'/',	// integral
	0x83,	// fi(?)
	'~',	// approx=
	'D',	// DELTA
	0xAB,	// euro double-open-quote
	0xBB,	// euro double-close-quote
	0x85,	// ellipsis
	0xA0,	// non-breaking space
	0xC0,	// A-grave
	0xC3,	// A-tilde
	0xD5,	// O-tilde
	0x8C,	// OE
	0x9C,	// oe

	0x96,	// dash
	0x97,	// long dash
	0x93,	// open "
	0x94,	// close "
	0x91,	// open '
	0x92,	// close '
	0xF7,	// divide by
	'$',	// wordstar diamond
	0xFF,	// y-umlaut
	0x9F,	// Y-umlaut
	'/',	// some other kinda slash?
	0xA4,	// can't remember what this is called
	0x8B,	// euro single-open-quote
	0x9B,	// euro single-close-quote
	'f',	// fi
	'f',	// fl

	0x87,	// little up/down arrow (double dagger substituted)
	0xB7,	// center dot
	0x82,	// baseline single-quote
	0x84,	// baseline double-quote
	0x89,	// per-thousand
	0xC2,	// A-circumflex
	0xCA,	// E-circumflex
	0xC1,	// A-acute
	0xCB,	// E-umlaut
	0xC8,	// E-grave
	0xCD,	// I-acute
	0xCE,	// I-circumflex
	0xCF,	// I-umlaut
	0xCC,	// I-grave
	0xD3,	// O-acute
	0xD4,	// O-circumflex

	'@',	// apple
	0xD2,	// O-grave
	0xDA,	// U-acute
	0xDB,	// U-circumflex
	0xD9,	// U-grave
	'i',	// dotless i
	0x88,	// circumflex
	0x98,	// tilde
	0x8F,	// bar
	'\'',	// scoop accent
	0xB0,	// single dot accent (circle accent substituted)
	0xB0,	// circle accent
	0xB8,	// cedire
	0x98,	// another tilde(?)
	'.',	// backwards cedire
	'\'',	// another scoop accent
};


////////////////////////////////////////////////// FUNCTIONS:

XBOOL XIsWinInMac(char ansiChar)
{
	return (XTranslateWinToMac(ansiChar) == 0xF0) ? FALSE
												 : TRUE;
}

XBOOL XIsMacInWin(char macChar)
{
char		ansiChar = XTranslateMacToWin(macChar);

	return (XTranslateWinToMac(ansiChar) != macChar) ? FALSE
													: TRUE;
}

char XTranslateWinToMac(char ansiChar)
{
	if (ansiChar < 0x80)
	{
		return ansiChar;
	}
	else
	{
	int			macChar;

		macChar = 0x80;
		while (--macChar >= 0)
		{
			if ((char)macToWinTable[macChar] == ansiChar)
			{
				return (char)(macChar + 0x80);
			}
		}
		return 0xF0;	// apple character
	}
}

char XTranslateMacToWin(char macChar)
{
	if (macChar < 0x80)
	{
		return macChar;
	}
	else
	{
		return (char)macToWinTable[macChar - 0x80];
	}
}
#endif	// USE_FULL_RMF_SUPPORT

// EOF of X_API.c
