/*
 * @(#)X_Decompress.c	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	X_Decompress.c
**
**	Tools for decompressing/encrypting/decrypting audio data
**
**
**	History	-
**	8/18/96		Created
**				Moved XExpandMace1to3 & XExpandMace1to6 from X_API.c
**	9/22/96		Changed XExpandMacADPCMto16BitLinear to XExpandMacADPCMtoXBitLinear
**				to support 8 bit expansion
**	11/14/96	Removed dependancy on MacSpecificSMS.h
**	12/19/96	Fixed an unassgined variable in IMA decoder
**	12/30/96	Changed copyright
**	1/12/97		Added XDecryptData and XEncryptData
**	1/20/97		Added XEncryptedStrCpy && XEncryptedStrLen
**	1/22/97		Changed strcpy functions to XStrCpy
**	1/23/97		Removed platform include files
**	2/21/97		Fixed stereo IMA decoding bug
**	5/3/97		Fixed a few potential problems that the MOT compiler found
**	5/15/97		Put in code for ADPCM decoding for WAVE files & AIFF files
**	9/30/97		Wrapped XEncryptData with USE_CREATION_API == TRUE
*/
/*****************************************************************************/
#include "X_API.h"
#include "X_Formats.h"


void XGetCompressionName(long compressionType, void *cName)
{
	switch (compressionType)
	{
		case C_NONE:
			XStrCpy((char *)cName, "no compression");
			break;
		case C_IMA4:
//			XStrCpy((char *)cName, "CCITT G.721 ADPCM compression (IMA 4 to 1)");
			XStrCpy((char *)cName, "ADPCM compression (IMA 4 to 1)");
			break;
		case C_MACE3:
			XStrCpy((char *)cName, "Apple MACE 3 to 1");
			break;
		case C_MACE6:
			XStrCpy((char *)cName, "Apple MACE 6 to 1");
			break;
		case C_ULAW:
			XStrCpy((char *)cName, "ulaw 2 to 1");
			break;
	}
}



// expand data 1 to 3 use MacOS MACE compression
void XExpandMace1to3(void *inBuffer, void *outBuffer, unsigned long length, 
					void * inState, void * outState, 
					unsigned long numChannels, unsigned long whichChannel)
{
#if X_PLATFORM == X_MACINTOSH
	Exp1to3(inBuffer, outBuffer, length, (StateBlockPtr)inState, (StateBlockPtr)outState, numChannels, whichChannel);
#else
	inBuffer = inBuffer;
	outBuffer = outBuffer;
	inState = inState;
	outState = outState;
	whichChannel = whichChannel;

	// since this is a specifc Mac 8-bit format, we'll just do silence for now
	XSetMemory(outBuffer, (length * 6) * numChannels, 0x80);
#endif
}

// expand data 1 to 6 use MacOS MACE compression
void XExpandMace1to6(void *inBuffer, void *outBuffer, unsigned long length, 
					void * inState, void * outState, 
					unsigned long numChannels, unsigned long whichChannel)
{
#if X_PLATFORM == X_MACINTOSH
	Exp1to6(inBuffer, outBuffer, length, (StateBlockPtr)inState, (StateBlockPtr)outState, numChannels, whichChannel);
#else
	inBuffer = inBuffer;
	outBuffer = outBuffer;
	inState = inState;
	outState = outState;
	whichChannel = whichChannel;

	// since this is a specifc Mac 8-bit format, we'll just do silence for now
	XSetMemory(outBuffer, (length * 6) * numChannels, 0x80);
#endif
}


/*
** This routine converts from ulaw to 16 bit linear.
**
** Craig Reese: IDA/Supercomputing Research Center
** 29 September 1989
**
** References:
** 1) CCITT Recommendation G.711  (very difficult to follow)
** 2) MIL-STD-188-113,"Interoperability and Performance Standards
**     for Analog-to_Digital Conversion Techniques,"
**     17 February 1987
**
** Input: 8 bit ulaw sample
** Output: signed 16 bit linear sample
*/
#if 0
static short int st_ulaw_to_linear(unsigned char ulawbyte)
{
	static short int exp_lut[8] = { 0, 132, 396, 924, 1980, 4092, 8316, 16764 };
	short int sign, exponent, mantissa, sample;

	ulawbyte = ~ ulawbyte;
	sign = ( ulawbyte & 0x80 );
	exponent = ( ulawbyte >> 4 ) & 0x07;
	mantissa = ulawbyte & 0x0F;
	sample = exp_lut[exponent] + ( mantissa << ( exponent + 3 ) );
	if ( sign != 0 )
	{
		sample = -sample;
	}
	return sample;
}
#else
static short int st_ulaw_to_linear(unsigned char ulawbyte)
{
	static const short int ulaw_exp_table[256] = {
			 -32124,-31100,-30076,-29052,-28028,-27004,-25980,-24956,
			 -23932,-22908,-21884,-20860,-19836,-18812,-17788,-16764,
			 -15996,-15484,-14972,-14460,-13948,-13436,-12924,-12412,
			 -11900,-11388,-10876,-10364, -9852, -9340, -8828, -8316,
			  -7932, -7676, -7420, -7164, -6908, -6652, -6396, -6140,
			  -5884, -5628, -5372, -5116, -4860, -4604, -4348, -4092,
			  -3900, -3772, -3644, -3516, -3388, -3260, -3132, -3004,
			  -2876, -2748, -2620, -2492, -2364, -2236, -2108, -1980,
			  -1884, -1820, -1756, -1692, -1628, -1564, -1500, -1436,
			  -1372, -1308, -1244, -1180, -1116, -1052,  -988,  -924,
			   -876,  -844,  -812,  -780,  -748,  -716,  -684,  -652,
			   -620,  -588,  -556,  -524,  -492,  -460,  -428,  -396,
			   -372,  -356,  -340,  -324,  -308,  -292,  -276,  -260,
			   -244,  -228,  -212,  -196,  -180,  -164,  -148,  -132,
			   -120,  -112,  -104,   -96,   -88,   -80,   -72,   -64,
			    -56,   -48,   -40,   -32,   -24,   -16,    -8,     0,
			  32124, 31100, 30076, 29052, 28028, 27004, 25980, 24956,
			  23932, 22908, 21884, 20860, 19836, 18812, 17788, 16764,
			  15996, 15484, 14972, 14460, 13948, 13436, 12924, 12412,
			  11900, 11388, 10876, 10364,  9852,  9340,  8828,  8316,
			   7932,  7676,  7420,  7164,  6908,  6652,  6396,  6140,
			   5884,  5628,  5372,  5116,  4860,  4604,  4348,  4092,
			   3900,  3772,  3644,  3516,  3388,  3260,  3132,  3004,
			   2876,  2748,  2620,  2492,  2364,  2236,  2108,  1980,
			   1884,  1820,  1756,  1692,  1628,  1564,  1500,  1436,
			   1372,  1308,  1244,  1180,  1116,  1052,   988,   924,
			    876,   844,   812,   780,   748,   716,   684,   652,
			    620,   588,   556,   524,   492,   460,   428,   396,
			    372,   356,   340,   324,   308,   292,   276,   260,
			    244,   228,   212,   196,   180,   164,   148,   132,
			    120,   112,   104,    96,    88,    80,    72,    64,
			     56,    48,    40,    32,    24,    16,     8,     0};

	return ulaw_exp_table[ulawbyte];
}


#endif

void XExpandULawto16BitLinear(unsigned char *pSource, short int *pDest, long frames, long channels)
{
	long	count;

	frames *= channels;
	for (count = 0; count < frames; count++)
	{
		*pDest = st_ulaw_to_linear(*pSource);
		pSource++;
		pDest++;
	}
}



// Mac ADPCM decompression (IMA 4 to 1)

static const short		ima_indextab[16] =	{
									-1,-1,-1,-1, 2, 4, 6, 8,
									-1,-1,-1,-1, 2, 4, 6, 8
								};
static const short		ima_steptab[89] =		{	
									7, 8, 9, 10, 11, 12, 13, 14,
									16,  17,  19,  21,  23,  25,  28,
									31,  34,  37,  41,  45,  50,  55,
									60,  66,  73,  80,  88,  97, 107,
									118, 130, 143, 157, 173, 190, 209,
									230, 253, 279, 307, 337, 371, 408,
									449, 494, 544, 598, 658, 724, 796,
									876, 963,1060,1166,1282,1411,1552,
							   		1707,1878,2066,2272,2499,2749,3024,3327,3660,4026,
									4428,4871,5358,5894,6484,7132,7845,8630,
									9493,10442,11487,12635,13899,15289,16818,
									18500,20350,22385,24623,27086,29794,32767
								};

// We store IMA data in a stream of 64 byte blocks, with a 2 byte predictor. 
// This allows us random access of the data. The predictor is the same value 
// computed for the block, if it were compress in a streaming fashion. But 
// saving this value in each block allows random starting points, so long as 
// you start at a block boundary.

// In out IMA data, you get the following 
// compression info when calling GetCompressionInfo().

// 64 samplesPerPacket
// 34 bytesPerPacket, 68 bytesPerPacket for stereo
// 2 bytesPerSample

// In the CmdSoundHeader, you'll find the field numFrames, which is the 
// number of blocks/frames/packets, each containing 64 decompressed samples.

void XExpandMacADPCMtoXBitLinear(char *pSource, long frames, short int channels, 
									void *pDest, short int bitSize)
{
	char			*pSourceR;
	short int		*pDest16;
	char			*pDest8;
	long			stepL, indexL;
	long			diffL, predsampleL;
	char			codebufL, codeL;
	long			stepR, indexR;
	long			diffR, predsampleR;
	char			codebufR, codeR;
	long			count;
	char			macBlockOffset;

	if ( (bitSize == 8) || (bitSize == 16) )
	{
		// setup inital predictor and index
		stepL = 7;
		predsampleL = 0;
		indexL = 0;
		codebufL = 0;
		pDest16 = (short int *)pDest;
		pDest8 = (char *)pDest;
		if (channels == 1)
		{
			macBlockOffset = 32;	// skip first two predictor bytes
			for (count = 0; count < frames; count++)
			{
				if (macBlockOffset == 32)
				{
					macBlockOffset = 0;
					pSource += 2;	// skip next predictor bytes
				}

				if	(count & 1)							/* two samples per inbuf char */
				{
					codeL = (codebufL >> 4) & 0x0F;
				}
				else	
				{
					codebufL = *pSource++;			/* buffer two ADPCM nibbles */
					codeL = codebufL & 0x0F;
					macBlockOffset++;
				}
				/* compute new sample estimate predsample */
				 diffL = 0;
				 if (codeL & 4) 
				 	diffL += stepL;
				 if (codeL & 2) 
				 	diffL += stepL >> 1;
				 if (codeL & 1) 
				 	diffL += stepL >> 2;
				 diffL += stepL >> 3;
				 if (codeL & 8)
				 	diffL = -diffL;
				predsampleL += diffL;

				 if (predsampleL > 32767)				/* check for overflow */
				 	predsampleL = 32767;
				 else if (predsampleL < -32768)
				 	predsampleL = -32768;

				/* store estimate to output buffer */
				if (bitSize == 16)
				{
					*pDest16++ = predsampleL;		
				}
				else
				{
					*pDest8++ = (predsampleL >> 8L) - 128;
				}
				/* compute new stepsize step */
				indexL += ima_indextab[codeL];
				if(indexL < 0) indexL = 0;
				else if(indexL > 88) indexL = 88;
				stepL = ima_steptab[indexL];
			}
		}
		else
		{	// STEREO DECODER
			pSourceR = pSource;		// handle the starting case
			pSource -= 34;
			stepR = 7;
			predsampleR = 0;
			indexR = 0;
			codebufR = 0;
			macBlockOffset = 32;	// skip first two predictor bytes

			for (count = 0; count < frames; count++)
			{
				if (macBlockOffset == 32)
				{
					macBlockOffset = 0;
					pSource += 2 + 34;	// skip next predictor bytes
					pSourceR += 2 + 34;	// skip next predictor bytes
				}

				if	(count & 1)							/* two samples per inbuf char */
				{
					codeL = (codebufL >> 4) & 0x0F;
					codeR = (codebufR >> 4) & 0x0F;
				}
				else	
				{
					codebufL = *pSource++;			/* buffer two ADPCM nibbles */
					codeL = codebufL & 0x0F;
					codebufR = *pSourceR++;			/* buffer two ADPCM nibbles */
					codeR = codebufR & 0x0F;
					macBlockOffset++;
				}
				/* compute new sample estimate predsample */
				 diffL = 0;
				 if (codeL & 4) 
				 	diffL += stepL;
				 if (codeL & 2) 
				 	diffL += stepL >> 1;
				 if (codeL & 1) 
				 	diffL += stepL >> 2;
				 diffL += stepL >> 3;
				 if (codeL & 8)
				 	diffL = -diffL;
				predsampleL += diffL;

				 diffR = 0;
				 if (codeR & 4) 
				 	diffR += stepR;
				 if (codeR & 2) 
				 	diffR += stepR >> 1;
				 if (codeR & 1) 
				 	diffR += stepR >> 2;
				 diffR += stepR >> 3;
				 if (codeR & 8)
				 	diffR = -diffR;
				predsampleR += diffR;

				 if (predsampleL > 32767)				/* check for overflow */
				 	predsampleL = 32767;
				 else if (predsampleL < -32768)
				 	predsampleL = -32768;

				 if (predsampleR > 32767)				/* check for overflow */
				 	predsampleR = 32767;
				 else if (predsampleR < -32768)
				 	predsampleR = -32768;

				/* store estimate to output buffer */
				if (bitSize == 16)
				{
					*pDest16++ = predsampleL;
					*pDest16++ = predsampleR;
				}
				else
				{
					*pDest8++ = (predsampleL >> 8L) - 128;
					*pDest8++ = (predsampleR >> 8L) - 128;
				}
				/* compute new stepsize step */
				indexL += ima_indextab[codeL];
				if(indexL < 0) indexL = 0;
				else if(indexL > 88) indexL = 88;
				stepL = ima_steptab[indexL];

				indexR += ima_indextab[codeR];
				if(indexR < 0) indexR = 0;
				else if(indexR > 88) indexR = 88;
				stepR = ima_steptab[indexR];
			}
		}
	}
}

#if USE_HIGHLEVEL_FILE_API
//==========================================================================;
//
//	  DECODE ROUTINES
//
//==========================================================================;
#define IMAADPCM_HEADER_LENGTH		4	// In bytes, per channel.
#define	MAC_IMAADPCM_HEADER_LENGTH	2


//--------------------------------------------------------------------------;
//  
//  int imaadpcmSampleDecode
//  
//  Description:
//	  This routine decodes a single ADPCM sample.  For efficiency, it is
//	  declared INLINE.  Note that, depending on the optimization flags,
//	  it may not actually be implemented as INLINE.  Optimizing for speed
//	  (-Oxwt) will generally obey the INLINE specification.
//  
//  Arguments:
//	  int nEncodedSample:  The sample to be decoded.
//	  int nPredictedSample:  The predicted value of the sample (in PCM).
//	  int nStepSize:  The quantization step size used to encode the sample.
//  
//  Return (int):  The decoded PCM sample.
//  
//--------------------------------------------------------------------------;

static INLINE int imaadpcmSampleDecode
(
	int				 nEncodedSample,
	int				 nPredictedSample,
	int				 nStepSize
)
{
	long			lDifference;
	long			lNewSample;

	//
	//  calculate difference:
	//
	//	  lDifference = (nEncodedSample + 1/2) * nStepSize / 4
	//
	lDifference = nStepSize>>3;

	if (nEncodedSample & 4) 
		lDifference += nStepSize;

	if (nEncodedSample & 2) 
		lDifference += nStepSize>>1;

	if (nEncodedSample & 1) 
		lDifference += nStepSize>>2;

	//
	//  If the 'sign bit' of the encoded nibble is set, then the
	//  difference is negative...
	//
	if (nEncodedSample & 8)
		lDifference = -lDifference;

	//
	//  adjust predicted sample based on calculated difference
	//
	lNewSample = nPredictedSample + lDifference;

	//
	//  check for overflow and clamp if necessary to a 16 signed sample.
	//  Note that this is optimized for the most common case, when we
	//  don't have to clamp.
	//
	if( (long)(short)lNewSample == lNewSample )
	{
		return (int)lNewSample;
	}

	//
	//  Clamp.
	//
	if( lNewSample < -32768 )
		return (int)-32768;
	else
		return (int)32767;
}


//--------------------------------------------------------------------------;
//  
//  int imaadpcmNextStepIndex
//  
//  Description:
//	  This routine calculates the step index value to use for the next
//	  encode, based on the current value of the step index and the current
//	  encoded sample.  For efficiency, it is declared INLINE.  Note that,
//	  depending on the optimization flags, it may not actually be 
//	  implemented as INLINE.  Optimizing for speed (-Oxwt) will generally 
//	  obey the INLINE specification.
//  
//  Arguments:
//	  int nEncodedSample:  The current encoded ADPCM sample.
//	  int nStepIndex:  The step index value used to encode nEncodedSample.
//  
//  Return (int):  The step index to use for the next sample.
//  
//--------------------------------------------------------------------------;

static INLINE int imaadpcmNextStepIndex
(
	int					 nEncodedSample,
	int					 nStepIndex
)
{
	//
	//  compute new stepsize step
	//
	nStepIndex += ima_indextab[nEncodedSample];

	if (nStepIndex < 0)
		nStepIndex = 0;
	else if (nStepIndex > 88)
		nStepIndex = 88;

	return (nStepIndex);
}



//--------------------------------------------------------------------------;
//  
//  BOOL imaadpcmValidStepIndex
//  
//  Description:
//	  This routine checks the step index value to make sure that it is
//	  within the legal range.
//  
//  Arguments:
//	  
//	  int nStepIndex:  The step index value.
//  
//  Return (BOOL):  TRUE if the step index is valid; FALSE otherwise.
//  
//--------------------------------------------------------------------------;

static INLINE XBOOL imaadpcmValidStepIndex
(
	int					 nStepIndex
)
{

	if( nStepIndex >= 0 && nStepIndex <= 88 )
		return TRUE;
	else
		return FALSE;
}


//--------------------------------------------------------------------------;
//  
//  XDWORD imaadpcmDecode4Bit_M08
//  XDWORD imaadpcmDecode4Bit_M16
//  XDWORD imaadpcmDecode4Bit_S08
//  XDWORD imaadpcmDecode4Bit_S16
//  
//  Description:
//	  These functions decode a buffer of data from ADPCM to PCM in the
//	  specified format.  The appropriate function is called once for each
//	  ACMDM_STREAM_CONVERT message received.  Note that since these
//	  functions must share the same prototype as the encoding functions
//	  (see acmdStreamOpen() and acmdStreamConvert() in codec.c for more
//	  details), not all the parameters are used by these routines.
//  
//  Arguments:
//	  unsigned char * pbSrc:  Pointer to the source buffer (ADPCM data).
//	  XDWORD cbSrcLength:  The length of the source buffer (in bytes).
//	  unsigned char * pbDst:  Pointer to the destination buffer (PCM data).  Note
//					  that it is assumed that the destination buffer is
//					  large enough to hold all the encoded data; see
//					  acmdStreamSize() in codec.c for more details.
//	  unsigned int nBlockAlignment:  The block alignment of the ADPCM data (in
//					  bytes).
//  
//  Return (XDWORD):  The number of bytes used in the destination buffer.
//  
//--------------------------------------------------------------------------;

static INLINE XDWORD imaadpcmDecode4Bit_M08
(
	unsigned char *			pbSrc,
	XDWORD					cbSrcLength,
	unsigned char *			pbDst,
	unsigned int			nBlockAlignment
)
{
	unsigned char *			pbDstStart;
	unsigned int			cbHeader;
	unsigned int			cbBlockLength;
	XBYTE					bSample;
	int						nStepSize;

	int						nEncSample;
	int						nPredSample;
	int						nStepIndex;

	
	pbDstStart = pbDst;
	cbHeader = IMAADPCM_HEADER_LENGTH * 1;  //  1 = number of channels.


//	DPF(3,"Starting imaadpcmDecode4Bit_M08().");


	//
	//
	//
	while (cbSrcLength >= cbHeader)
	{
		XDWORD	   dwHeader;

		cbBlockLength  = (unsigned int)XMIN(cbSrcLength, nBlockAlignment);
		cbSrcLength   -= cbBlockLength;
		cbBlockLength -= cbHeader;

		//
		//  block header
		//
		dwHeader = *(XDWORD *)pbSrc;
		#if X_WORD_ORDER == FALSE	// motorola?
			dwHeader = XSwapLong(dwHeader);
		#endif
		pbSrc   += sizeof(XDWORD);
		nPredSample = (int)(short)XLOWORD(dwHeader);
		nStepIndex  = (int)(XBYTE)XHIWORD(dwHeader);

		if( !imaadpcmValidStepIndex(nStepIndex) ) {
			//
			//  The step index is out of range - this is considered a fatal
			//  error as the input stream is corrupted.  We fail by returning
			//  zero bytes converted.
			//
//			DPF(1,"imaadpcmDecode4Bit_M08: invalid step index.");
			return 0;
		}
		

		//
		//  write out first sample
		//
		*pbDst++ = (XBYTE)((nPredSample >> 8) + 128);


		//
		//
		//
		while (cbBlockLength--)
		{
			bSample = *pbSrc++;

			//
			//  sample 1
			//
			nEncSample  = (bSample & (XBYTE)0x0F);
			nStepSize   = ima_steptab[nStepIndex];
			nPredSample = imaadpcmSampleDecode(nEncSample, nPredSample, nStepSize);
			nStepIndex  = imaadpcmNextStepIndex(nEncSample, nStepIndex);

			//
			//  write out sample
			//
			*pbDst++ = (XBYTE)((nPredSample >> 8) + 128);

			//
			//  sample 2
			//
			nEncSample  = (bSample >> 4);
			nStepSize   = ima_steptab[nStepIndex];
			nPredSample = imaadpcmSampleDecode(nEncSample, nPredSample, nStepSize);
			nStepIndex  = imaadpcmNextStepIndex(nEncSample, nStepIndex);

			//
			//  write out sample
			//
			*pbDst++ = (XBYTE)((nPredSample >> 8) + 128);
		}
	}

	//
	//  We return the number of bytes used in the destination.  This is
	//  simply the difference in bytes from where we started.
	//
	return (XDWORD)(pbDst - pbDstStart);

} // imaadpcmDecode4Bit_M08()



//--------------------------------------------------------------------------;
//--------------------------------------------------------------------------;

static INLINE XDWORD imaadpcmDecode4Bit_M16
(
	unsigned char *			pbSrc,
	XDWORD					cbSrcLength,
	unsigned char *			pbDst,
	unsigned int			nBlockAlignment
)
{
	unsigned char *				  pbDstStart;
	unsigned int					cbHeader;
	unsigned int					cbBlockLength;
	XBYTE					bSample;
	int					 nStepSize;

	int					 nEncSample;
	int					 nPredSample;
	int					 nStepIndex;

	
	pbDstStart = pbDst;
	cbHeader = IMAADPCM_HEADER_LENGTH * 1;  //  1 = number of channels.


//	DPF(3,"Starting imaadpcmDecode4Bit_M16().");


	//
	//
	//
	while (cbSrcLength >= cbHeader)
	{
		XDWORD	   dwHeader;

		cbBlockLength  = (unsigned int)XMIN(cbSrcLength, nBlockAlignment);
		cbSrcLength   -= cbBlockLength;
		cbBlockLength -= cbHeader;

		//
		//  block header
		//
		dwHeader = *(XDWORD *)pbSrc;
		#if X_WORD_ORDER == FALSE	// motorola?
			dwHeader = XSwapLong(dwHeader);
		#endif
		pbSrc   += sizeof(XDWORD);
		nPredSample = (int)(short)XLOWORD(dwHeader);
		nStepIndex  = (int)(XBYTE)XHIWORD(dwHeader);

		if( !imaadpcmValidStepIndex(nStepIndex) ) {
			//
			//  The step index is out of range - this is considered a fatal
			//  error as the input stream is corrupted.  We fail by returning
			//  zero bytes converted.
			//
//			DPF(1,"imaadpcmDecode4Bit_M16: invalid step index.");
			return 0;
		}
		

		//
		//  write out first sample
		//
		*(short *)pbDst = (short)nPredSample;
		pbDst += sizeof(short);


		//
		//
		//
		while (cbBlockLength--)
		{
			bSample = *pbSrc++;

			//
			//  sample 1
			//
			nEncSample  = (bSample & (XBYTE)0x0F);
			nStepSize   = ima_steptab[nStepIndex];
			nPredSample = imaadpcmSampleDecode(nEncSample, nPredSample, nStepSize);
			nStepIndex  = imaadpcmNextStepIndex(nEncSample, nStepIndex);

			//
			//  write out sample
			//
			*(short *)pbDst = (short)nPredSample;
			pbDst += sizeof(short);

			//
			//  sample 2
			//
			nEncSample  = (bSample >> 4);
			nStepSize   = ima_steptab[nStepIndex];
			nPredSample = imaadpcmSampleDecode(nEncSample, nPredSample, nStepSize);
			nStepIndex  = imaadpcmNextStepIndex(nEncSample, nStepIndex);

			//
			//  write out sample
			//
			*(short *)pbDst = (short)nPredSample;
			pbDst += sizeof(short);
		}
	}

	//
	//  We return the number of bytes used in the destination.  This is
	//  simply the difference in bytes from where we started.
	//
	return (XDWORD)(pbDst - pbDstStart);

} // imaadpcmDecode4Bit_M16()



//--------------------------------------------------------------------------;
//--------------------------------------------------------------------------;

static INLINE XDWORD imaadpcmDecode4Bit_S08
(
	unsigned char *			pbSrc,
	XDWORD					cbSrcLength,
	unsigned char *			pbDst,
	unsigned int			nBlockAlignment
)
{
	unsigned char *				  pbDstStart;
	unsigned int					cbHeader;
	unsigned int					cbBlockLength;
	int					 nStepSize;
	XDWORD				   dwHeader;
	XDWORD				   dwLeft;
	XDWORD				   dwRight;
	int					 i;

	int					 nEncSampleL;
	int					 nPredSampleL;
	int					 nStepIndexL;

	int					 nEncSampleR;
	int					 nPredSampleR;
	int					 nStepIndexR;

	
	pbDstStart = pbDst;
	cbHeader = IMAADPCM_HEADER_LENGTH * 2;  //  2 = number of channels.


//	DPF(3,"Starting imaadpcmDecode4Bit_S08().");


	//
	//
	//
	while( 0 != cbSrcLength )
	{
		//
		//  The data should always be block aligned.
		//
		cbBlockLength  = nBlockAlignment;
		cbSrcLength   -= cbBlockLength;
		cbBlockLength -= cbHeader;


		//
		//  LEFT channel header
		//
		dwHeader = *(XDWORD *)pbSrc;
		#if X_WORD_ORDER == FALSE	// motorola?
			dwHeader = XSwapLong(dwHeader);
		#endif
		pbSrc   += sizeof(XDWORD);
		nPredSampleL = (int)(short)XLOWORD(dwHeader);
		nStepIndexL  = (int)(XBYTE)XHIWORD(dwHeader);

		if( !imaadpcmValidStepIndex(nStepIndexL) ) {
			//
			//  The step index is out of range - this is considered a fatal
			//  error as the input stream is corrupted.  We fail by returning
			//  zero bytes converted.
			//
//			DPF(1,"imaadpcmDecode4Bit_S08: invalid step index (L).");
			return 0;
		}
		
		//
		//  RIGHT channel header
		//
		dwHeader = *(XDWORD *)pbSrc;
		#if X_WORD_ORDER == FALSE	// motorola?
			dwHeader = XSwapLong(dwHeader);
		#endif
		pbSrc   += sizeof(XDWORD);
		nPredSampleR = (int)(short)XLOWORD(dwHeader);
		nStepIndexR  = (int)(XBYTE)XHIWORD(dwHeader);

		if( !imaadpcmValidStepIndex(nStepIndexR) ) {
			//
			//  The step index is out of range - this is considered a fatal
			//  error as the input stream is corrupted.  We fail by returning
			//  zero bytes converted.
			//
//			DPF(1,"imaadpcmDecode4Bit_S08: invalid step index (R).");
			return 0;
		}
		

		//
		//  write out first sample
		//
		*pbDst++ = (XBYTE)((nPredSampleL >> 8) + 128);
		*pbDst++ = (XBYTE)((nPredSampleR >> 8) + 128);


		//
		//  The first XDWORD contains 4 left samples, the second XDWORD
		//  contains 4 right samples.  We process the source in 8-byte
		//  chunks to make it easy to interleave the output correctly.
		//
		while( 0 != cbBlockLength )
		{
			cbBlockLength -= 8;

			dwLeft   = *(XDWORD *)pbSrc;
			pbSrc   += sizeof(XDWORD);
			dwRight  = *(XDWORD *)pbSrc;
			pbSrc   += sizeof(XDWORD);
			#if X_WORD_ORDER == FALSE	// motorola?
				dwLeft = XSwapLong(dwLeft);
				dwRight = XSwapLong(dwRight);
			#endif

			for( i=8; i>0; i-- )
			{
				//
				//  LEFT channel
				//
				nEncSampleL  = (int)(dwLeft & 0x0F);
				nStepSize	= ima_steptab[nStepIndexL];
				nPredSampleL = imaadpcmSampleDecode(nEncSampleL, nPredSampleL, nStepSize);
				nStepIndexL  = imaadpcmNextStepIndex(nEncSampleL, nStepIndexL);

				//
				//  RIGHT channel
				//
				nEncSampleR  = (int)(dwRight & 0x0F);
				nStepSize	= ima_steptab[nStepIndexR];
				nPredSampleR = imaadpcmSampleDecode(nEncSampleR, nPredSampleR, nStepSize);
				nStepIndexR  = imaadpcmNextStepIndex(nEncSampleR, nStepIndexR);

				//
				//  write out sample
				//
				*pbDst++ = (XBYTE)((nPredSampleL >> 8) + 128);
				*pbDst++ = (XBYTE)((nPredSampleR >> 8) + 128);

				//
				//  Shift the next input sample into the low-order 4 bits.
				//
				dwLeft  >>= 4;
				dwRight >>= 4;
			}
		}
	}

	//
	//  We return the number of bytes used in the destination.  This is
	//  simply the difference in bytes from where we started.
	//
	return (XDWORD)(pbDst - pbDstStart);

} // imaadpcmDecode4Bit_S08()



//--------------------------------------------------------------------------;
//--------------------------------------------------------------------------;

static INLINE XDWORD imaadpcmDecode4Bit_S16
(
	unsigned char *			pbSrc,
	XDWORD					cbSrcLength,
	unsigned char *			pbDst,
	unsigned int			nBlockAlignment
)
{
	unsigned char *		pbDstStart;
	unsigned int		cbHeader;
	unsigned int		cbBlockLength;
	int					nStepSize;
	XDWORD				dwHeader;
	XDWORD				dwLeft;
	XDWORD				dwRight;
	int					i;

	int					nEncSampleL;
	int					nPredSampleL;
	int					nStepIndexL;

	int					nEncSampleR;
	int					nPredSampleR;
	int					nStepIndexR;

	
	pbDstStart = pbDst;
	cbHeader = IMAADPCM_HEADER_LENGTH * 2;  //  2 = number of channels.


//	DPF(3,"Starting imaadpcmDecode4Bit_S16().");


	//
	//
	//
	while( 0 != cbSrcLength )
	{
		//
		//  The data should always be block aligned.
		//
		cbBlockLength  = nBlockAlignment;
		cbSrcLength   -= cbBlockLength;
		cbBlockLength -= cbHeader;


		//
		//  LEFT channel header
		//
		dwHeader = *(XDWORD *)pbSrc;
		#if X_WORD_ORDER == FALSE	// motorola?
			dwHeader = XSwapLong(dwHeader);
		#endif
		pbSrc   += sizeof(XDWORD);
		nPredSampleL = (int)(short)XLOWORD(dwHeader);
		nStepIndexL  = (int)(XBYTE)XHIWORD(dwHeader);

		if( !imaadpcmValidStepIndex(nStepIndexL) ) {
			//
			//  The step index is out of range - this is considered a fatal
			//  error as the input stream is corrupted.  We fail by returning
			//  zero bytes converted.
			//
//			DPF(1,"imaadpcmDecode4Bit_S16: invalid step index %u (L).", nStepIndexL);
			return 0;
		}
		
		//
		//  RIGHT channel header
		//
		dwHeader = *(XDWORD *)pbSrc;
		#if X_WORD_ORDER == FALSE	// motorola?
			dwHeader = XSwapLong(dwHeader);
		#endif
		pbSrc   += sizeof(XDWORD);
		nPredSampleR = (int)(short)XLOWORD(dwHeader);
		nStepIndexR  = (int)(XBYTE)XHIWORD(dwHeader);

		if( !imaadpcmValidStepIndex(nStepIndexR) ) {
			//
			//  The step index is out of range - this is considered a fatal
			//  error as the input stream is corrupted.  We fail by returning
			//  zero bytes converted.
			//
//			DPF(1,"imaadpcmDecode4Bit_S16: invalid step index %u (R).",nStepIndexR);
			return 0;
		}
		

		//
		//  write out first sample
		//
		*(XDWORD *)pbDst = XMAKELONG(nPredSampleL, nPredSampleR);
		pbDst += sizeof(XDWORD);


		//
		//  The first XDWORD contains 4 left samples, the second XDWORD
		//  contains 4 right samples.  We process the source in 8-byte
		//  chunks to make it easy to interleave the output correctly.
		//
		while( 0 != cbBlockLength )
		{
			cbBlockLength -= 8;

			dwLeft   = *(XDWORD *)pbSrc;
			pbSrc   += sizeof(XDWORD);
			dwRight  = *(XDWORD *)pbSrc;
			pbSrc   += sizeof(XDWORD);
			#if X_WORD_ORDER == FALSE	// motorola?
				dwLeft = XSwapLong(dwLeft);
				dwRight = XSwapLong(dwRight);
			#endif

			for( i=8; i>0; i-- )
			{
				//
				//  LEFT channel
				//
				nEncSampleL  = (int)(dwLeft & 0x0F);
				nStepSize	= ima_steptab[nStepIndexL];
				nPredSampleL = imaadpcmSampleDecode(nEncSampleL, nPredSampleL, nStepSize);
				nStepIndexL  = imaadpcmNextStepIndex(nEncSampleL, nStepIndexL);

				//
				//  RIGHT channel
				//
				nEncSampleR  = (int)(dwRight & 0x0F);
				nStepSize	= ima_steptab[nStepIndexR];
				nPredSampleR = imaadpcmSampleDecode(nEncSampleR, nPredSampleR, nStepSize);
				nStepIndexR  = imaadpcmNextStepIndex(nEncSampleR, nStepIndexR);

				//
				//  write out sample
				//
				*(XDWORD *)pbDst = XMAKELONG(nPredSampleL, nPredSampleR);
				pbDst += sizeof(XDWORD);

				//
				//  Shift the next input sample into the low-order 4 bits.
				//
				dwLeft  >>= 4;
				dwRight >>= 4;
			}
		}
	}

	//
	//  We return the number of bytes used in the destination.  This is
	//  simply the difference in bytes from where we started.
	//
	return (XDWORD)(pbDst - pbDstStart);

} // imaadpcmDecode4Bit_S16()

//  XDWORD imaadpcmDecode4Bit_M08
//  XDWORD imaadpcmDecode4Bit_M16
//  XDWORD imaadpcmDecode4Bit_S08
//  XDWORD imaadpcmDecode4Bit_S16

XDWORD XDecodeIMA4Bit(	unsigned char *pbSrc, XDWORD cbSrcLength,
						unsigned char	*pbDst,
						unsigned int	nBlockAlignment,
						unsigned short	channels,
						unsigned short 	destinationBitSize)
{
	if ( (channels == 1) && (destinationBitSize == 8) )
	{
		return imaadpcmDecode4Bit_M08(pbSrc, cbSrcLength, pbDst, nBlockAlignment);
	}
	if ( (channels == 1) && (destinationBitSize == 16) )
	{
		return imaadpcmDecode4Bit_M16(pbSrc, cbSrcLength, pbDst, nBlockAlignment);
	}
	if ( (channels == 2) && (destinationBitSize == 8) )
	{
		return imaadpcmDecode4Bit_S08(pbSrc, cbSrcLength, pbDst, nBlockAlignment);
	}
	if ( (channels == 2) && (destinationBitSize == 16) )
	{
		return imaadpcmDecode4Bit_S16(pbSrc, cbSrcLength, pbDst, nBlockAlignment);
	}
	return 0L;
}
#endif	// USE_HIGHLEVEL_FILE_API

#if USE_HIGHLEVEL_FILE_API

// We store IMA data in a stream of 64 byte blocks, with a 2 byte predictor. 
// This allows us random access of the data. The predictor is the same value 
// computed for the block, if it were compress in a streaming fashion. But 
// saving this value in each block allows random starting points, so long as 
// you start at a block boundary.

// In out IMA data, you get the following 
// compression info when calling GetCompressionInfo().

// 64 samplesPerPacket
// 34 bytesPerPacket, 68 bytesPerPacket for stereo
// 2 bytesPerSample

// In the CmdSoundHeader, you'll find the field numFrames, which is the 
// number of blocks/frames/packets, each containing 64 decompressed samples.
// nBlockAlignment should be 34 in most cases

unsigned long XExpandIMAtoXBitLinearStreamed(char *pSourceData, unsigned long sourceLength, 
											unsigned char *pDest, 
											unsigned short int nBlockAlignment,
											unsigned short int channels, 
											unsigned short int destinationBitSize,
											long *pIndexL, long *pIndexR)
{
	char			*pSourceL;
	char			*pSourceR;
	short int		*pDest16;
	char			*pDest8;
	long			stepL, indexL;
	long			predsampleL;
	char			codebufL, codeL;
	long			stepR, indexR;
//	long			diffL;
//	long			diffR;
	long			predsampleR;
	char			codebufR, codeR;
	unsigned long	count, frame;

	nBlockAlignment = nBlockAlignment;
	pDest16 = (short int *)pDest;
	pDest8 = (char *)pDest;
	if ( (destinationBitSize == 8) || (destinationBitSize == 16) )
	{
		if (channels == 1)
		{
			pSourceL = pSourceData - nBlockAlignment;	// handle the starting case

			// setup inital predictor and index
			indexL = *pIndexL;
			codebufL = 0;
			stepL = 7;
			predsampleL = 0;

			for (frame = 0; frame < sourceLength;)
			{
				pSourceL += nBlockAlignment;

				predsampleL = (signed short)XGetShort(pSourceL);		// get next predictor bytes
				pSourceL += 2;
				frame += 2;

				for (count = 0; count < 64; count++)
				{
					if	(count & 1)							/* two samples per inbuf char */
					{
						codeL = (codebufL >> 4) & 0x0F;
					}
					else	
					{
						codebufL = *pSourceL++;			/* buffer two ADPCM nibbles */
						codeL = codebufL & 0x0F;
					}
					frame++;
#if 1
					stepL = ima_steptab[indexL];
					predsampleL = imaadpcmSampleDecode(codeL, predsampleL, stepL);
					indexL  = imaadpcmNextStepIndex(codeL, indexL);
#else
					/* compute new sample estimate predsample */
					 diffL = 0;
					 if (codeL & 4) 
					 	diffL += stepL;
					 if (codeL & 2) 
					 	diffL += stepL >> 1;
					 if (codeL & 1) 
					 	diffL += stepL >> 2;
					 diffL += stepL >> 3;
					 if (codeL & 8)
					 	diffL = -diffL;
					predsampleL += diffL;

					 if (predsampleL > 32767)				// check for overflow
					 	predsampleL = 32767;
					 else if (predsampleL < -32768)
					 	predsampleL = -32768;

					// compute new stepsize step
					indexL += ima_indextab[codeL];
					if(indexL < 0) indexL = 0;
					else if(indexL > 88) indexL = 88;
					stepL = ima_steptab[indexL];
#endif
					// store estimate to output buffer
					if (destinationBitSize == 16)
					{
						*pDest16++ = predsampleL;		
					}
					else
					{
						*pDest8++ = (predsampleL >> 8L) - 128;
					}
				}
			}
		}
		else
		{	// STEREO DECODER
			pSourceL = pSourceData - nBlockAlignment;	// handle the starting case
			pSourceR = pSourceData;

			// setup inital predictor and index
			indexL = *pIndexL;
			codebufL = 0;
			stepL = 7;
			predsampleL = 0;

			indexR = *pIndexR;
			codebufR = 0;
			stepR = 7;
			predsampleR = 0;

			for (frame = 0; frame < sourceLength;)
			{
				pSourceL += nBlockAlignment;
				pSourceR += nBlockAlignment;

				predsampleL = (signed short)XGetShort(pSourceL);		// get next predictor bytes
				pSourceL += 2;
				predsampleR = (signed short)XGetShort(pSourceR);
				pSourceR += 2;

				frame += 4;

				for (count = 0; count < 64; count++)
				{
					if	(count & 1)							/* two samples per inbuf char */
					{
						codeL = (codebufL >> 4) & 0x0F;
						codeR = (codebufR >> 4) & 0x0F;
					}
					else	
					{
						codebufL = *pSourceL++;			/* buffer two ADPCM nibbles */
						codeL = codebufL & 0x0F;
						codebufR = *pSourceR++;			/* buffer two ADPCM nibbles */
						codeR = codebufR & 0x0F;
					}
					frame++;
#if 1
					stepL = ima_steptab[indexL];
					stepR = ima_steptab[indexR];
					predsampleL = imaadpcmSampleDecode(codeL, predsampleL, stepL);
					predsampleR = imaadpcmSampleDecode(codeR, predsampleR, stepR);

					indexL  = imaadpcmNextStepIndex(codeL, indexL);
					indexR  = imaadpcmNextStepIndex(codeR, indexR);
#else
					/* compute new sample estimate predsample */

					 diffL = 0;
					 if (codeL & 4) 
					 	diffL += stepL;
					 if (codeL & 2) 
					 	diffL += stepL >> 1;
					 if (codeL & 1) 
					 	diffL += stepL >> 2;
					 diffL += stepL >> 3;
					 if (codeL & 8)
					 	diffL = -diffL;
					predsampleL += diffL;

					 diffR = 0;
					 if (codeR & 4) 
					 	diffR += stepR;
					 if (codeR & 2) 
					 	diffR += stepR >> 1;
					 if (codeR & 1) 
					 	diffR += stepR >> 2;
					 diffR += stepR >> 3;
					 if (codeR & 8)
					 	diffR = -diffR;
					predsampleR += diffR;

					 if (predsampleL > 32767)				// check for overflow
					 	predsampleL = 32767;
					 else if (predsampleL < -32768)
					 	predsampleL = -32768;

					 if (predsampleR > 32767)				// check for overflow
					 	predsampleR = 32767;
					 else if (predsampleR < -32768)
					 	predsampleR = -32768;

					/* compute new stepsize step */
					indexL += ima_indextab[codeL];
					if(indexL < 0) indexL = 0;
					else if(indexL > 88) indexL = 88;
					stepL = ima_steptab[indexL];

					indexR += ima_indextab[codeR];
					if(indexR < 0) indexR = 0;
					else if(indexR > 88) indexR = 88;
					stepR = ima_steptab[indexR];
#endif

					/* store estimate to output buffer */
					if (destinationBitSize == 16)
					{
						*pDest16++ = predsampleL;
						*pDest16++ = predsampleR;
					}
					else
					{
						*pDest8++ = (predsampleL >> 8L) - 128;
						*pDest8++ = (predsampleR >> 8L) - 128;
					}
				}
			}
		}
	}
	*pIndexL = indexL;
	*pIndexR = indexR;
	if (destinationBitSize == 16)
	{
		count = (char *)pDest16 - (char *)pDest;
	}
	else
	{
		count = pDest8 - (char *)pDest;
	}
	return count;
}
#endif	// USE_HIGHLEVEL_FILE_API


//==========================================================================;





#define INIT_R		(unsigned short int)56549
#define C1			(unsigned short int)52845
#define C2			(unsigned short int)22719

static unsigned int R;

static INLINE unsigned char PV_Decrypt(register unsigned char cipher)
{
	unsigned char plain;

	plain = (cipher ^ (R >> 8));
	R = (cipher + R) * C1 + C2;
	return plain;
}

static INLINE unsigned char PV_Encrypt(register unsigned char plain)
{
	unsigned char cipher;

	cipher = (plain ^ (R >> 8));
	R = (cipher + R) * C1 + C2;
	return cipher;
}

short int XEncryptedStrLen(char *src)
{
	short int len;

	len = -1;
	if (src == NULL)
	{
		src = "";
	}

	// decrypt
	R = INIT_R;
	do
	{
		len++;
	} while  (PV_Decrypt(*src++));

	return len;
}


// standard strcpy, but with crypto controls
// Copies 'C' string s2 into s1
char * XEncryptedStrCpy(char *dest, char *src, short int copy)
{
	char *sav;
	char data;

	if (src == NULL)
	{
		src = "";
	}

	// decrypt
	R = INIT_R;
	sav = dest;
	if (dest)
	{
		while (1)
		{
			data = PV_Decrypt(*src);
			if (data)
			{
				switch (copy)
				{
					case X_SOURCE_ENCRYPTED:
						*dest++ = data;
						break;
					case X_SOURCE_DEST_ENCRYPTED:
						*dest++ = *src;
						break;
				}
			}
			else
			{
				break;
			}
			src++;
		}
		*dest = 0;
	}
	return sav;
}

// duplicate encrypted string, but decrypt clone
char * XDecryptAndDuplicateStr(char *src)
{
	char *dup;

	dup = NULL;
	if (src)
	{
		dup = (char *)XNewPtr(XEncryptedStrLen(src)+1);
		if (dup)
		{
			// decrypt and copy
			XEncryptedStrCpy(dup, src, X_SOURCE_ENCRYPTED);
		}
	}
	return dup;
}

#if USE_CREATION_API == TRUE
// Encrypt a block of data. This should be U.S. munitions safe. ie below 40 bit
void XEncryptData(void *pData, unsigned long size)
{
	register unsigned char *pByte, *pEnd;

	if (pData && size)
	{
		// encrypt
		R = INIT_R;
		pByte = (unsigned char *)pData;
		pEnd = pByte + size;
		while (pByte < pEnd)
		{
			*pByte = PV_Encrypt(*pByte);
			pByte++;
		}
	}
}
#endif	// USE_CREATION_API == TRUE

// Decrypt a block of data. This should be U.S. munitions safe. ie below 40 bit
void XDecryptData(void *pData, unsigned long size)
{
	register unsigned char *pByte, *pEnd;

	if (pData && size)
	{
		// decrypt
		R = INIT_R;
		pByte = (unsigned char *)pData;
		pEnd = pByte + size;
		while (pByte < pEnd)
		{
			*pByte = PV_Decrypt(*pByte);
			pByte++;
		}
	}
}


// EOF of X_Decompress.c

