/*
 * @(#)X_Formats.h	1.12 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	X_Formats.h
**
**		This is platform independent file and data formats for SoundMusicSys
**
**
**	History	-
**	6/30/96		Created
**	7/3/96		Added packing pragmas
**	7/14/96		Removed PRAGMA_ALIGN_SUPPORTED
**	8/19/96		Added compressionType to SampleDataInfo
**	10/23/96	Changed GetKeySplitFromPtr to XGetKeySplitFromPtr
**	12/5/96		Added locked flags for songs and instruments
**	12/10/96	Added ID_RMF type
**	12/19/96	Added Sparc pragmas
**	12/30/96	Changed copyrights
**	1/2/97		Added ID_MOD type
**	1/3/97		Added XGetSongInformation
**	1/6/97		Added songType to XNewSongPtr
**	1/7/97		Changed structures typedef forms
**	1/12/97		Broke SongResource into two types: SongResource_SMS and SongResource_RMF
**	1/13/97		Added XGetSongResourceInfo & XDisposeSongResourceInfo & 
**				XGetSongResourceObjectID & XGetSongPerformanceSettings &
**				XGetSongResourceObjectType
**	1/18/97		Added XCheckAllInstruments & XCheckValidInstrument
**	1/24/97		Added SongResource_MOD
**	1/29/97		Added XGetSongInstrumentList
**				Added XGetMidiData
**				Added ID_ESND and ID_EMID types
**	1/30/97		Added XGetSoundResourceByName & XGetSoundResourceByID
**				Added XGetSongVoices & XSetSongVoices
**	2/14/97		Added volumeGain in SongResource_RMF
**				Added XSetSongVolumeGain & XGetSongVolumeGain
**	2/15/97		Changed volumeGain to songVolume, changed XSetSongVolumeGain & 
**				XGetSongVolumeGain to XSetSongVolume & XGetSongVolume
**	3/13/97		Added embeddedSong to SongResource_RMF structure
**				Added XGetSongEmbeddedStatus & XSetSongEmbeddedStatus
**	3/14/97		Added XIsSoundUsedInInstrument & XRenumberSampleInsideInstrument
**	3/16/97		Changed KeySplit to support a replacement for sample root key
**				on a per split basis
**	3/27/97		Changed all 4 character constants to use the FOUR_CHAR macro
**	5/13/97		Added ID_VERS & XVersion
**	6/20/97		Added XGetSampleNameFromID
**	7/24/97		Added the structure XExtSndHeader2
**				Added XCreateSoundObjectFromData & XGetSoundEmbeddedStatus & 
**				XSetSoundEmbeddedStatus
**	7/28/97		Changed a pragma for Sparc compilers (jdr)
**	8/7/97		Added XTranslateFromWaveformToSampleData
**	8/25/97		Fixed some Sun compiler warnings
**	9/30/97		Added BankStatus structure, added XCreateBankStatus & XGetBankStatus
**	10/9/97		Added ID_TEXT
*/
/*****************************************************************************/

#ifndef X_FORMATS
#define X_FORMATS

#ifndef __X_API__
	#include "X_API.h"
#endif

#ifndef G_SOUND
	#include "GenSnd.h"
#endif

#ifdef __cplusplus
	extern "C" {
#endif

#if CPU_TYPE == kRISC
	#pragma options align=mac68k
#endif
#if ((CPU_TYPE == k80X86) || (CPU_TYPE == kSPARC))
	#pragma pack (1)
#endif

/* Instrument and Song structures
*/
typedef struct 
{
	short int		instrumentNumber;
	short int		ResourceINSTID;
} Remap;
	
enum
{
	ID_NULL		=	0L,
	ID_SONG		=	FOUR_CHAR('S','O','N','G'),	//	'SONG'		// song
	ID_INST		=	FOUR_CHAR('I','N','S','T'),	//	'INST'		// instrument format
	ID_MIDI		=	FOUR_CHAR('M','i','d','i'),	//	'Midi'		// standard midi file
	ID_MIDI_OLD	=	FOUR_CHAR('M','I','D','I'),	//	'MIDI'		// standard midi file
	ID_CMID		=	FOUR_CHAR('c','m','i','d'),	//	'cmid'		// compressed midi file
	ID_CMIDI	=	ID_CMID,
	ID_EMID		=	FOUR_CHAR('e','m','i','d'),	//	'emid'		// encrypted midi file
	ID_ECMI		=	FOUR_CHAR('e','c','m','i'),	//	'ecmi'		// encrypted and compressed midi file
	ID_SND		=	FOUR_CHAR('s','n','d',' '),	//	'snd '		// sample
	ID_CSND		=	FOUR_CHAR('c','s','n','d'),	//	'csnd'		// compressed sample
	ID_ESND		=	FOUR_CHAR('e','s','n','d'),	//	'esnd'		// encrypted sample
	ID_RMF		=	FOUR_CHAR('R','M','F','!'),	//	'RMF!'		// rmf object
	ID_MOD		=	FOUR_CHAR('M','O','D','!'),	//	'MOD!'		// mod object

	ID_BANK		=	FOUR_CHAR('B','A','N','K'),	//	'BANK'		// bank ID
 
	ID_VERS		=	FOUR_CHAR('V','E','R','S'),	//	'VERS'		// version ID
	ID_TEXT		=	FOUR_CHAR('T','E','X','T'),	//	'TEXT'		// text
 
	ID_MTHD		=	FOUR_CHAR('M','T','h','d'),	//	'MThd'		// midi header ID
	ID_MTRK		=	FOUR_CHAR('M','T','r','k')	//	'MTrk'		// midi track ID
};


// bits for SongResource_SMS flags1
#define XBF_locked						0x80
#define XBF_terminateDecay				0x40
#define XBF_interpolateSong				0x20
#define XBF_interpolateLead				0x10
#define XBF_fileTrackFlag				0x08
#define XBF_enableMIDIProgram			0x04
#define XBF_disableClickRemoval			0x02
#define XBF_useLeadForAllVoices			0x01

// bits for SongResource_SMS flags2
#define XBF_ignoreBadPatches			0x80
#define XBF_reserved4					0x40
#define XBF_reserved5					0x20
#define XBF_masterEnablePitchRandomness	0x10
#define XBF_ampScaleLead				0x08
#define XBF_forceAmpScale				0x04
#define XBF_masterEnableAmpScale		0x02
#define XBF_reserved6					0x01

typedef enum
{
	SONG_TYPE_BAD = -1,
	SONG_TYPE_SMS = 0,
	SONG_TYPE_RMF = 1,
	SONG_TYPE_MOD = 2
} SongType;

// Song resource (SMS type)
typedef struct
{
	short int		midiResourceID;
	char			reserved_0;
	char			reverbType;
	unsigned short	songTempo;
	char			songType;						//	0 - SMS, 1 - RMF, 2 - MOD

	char			songPitchShift;
	char			maxEffects;
	char			maxNotes;
	short int		mixLevel;
	unsigned char	flags1;							// see XBF for flags1
	char			noteDecay;
	char			defaultPercusionProgram;		// yes, I wanted signed!
	unsigned char	flags2;							// see XBF for flags2
	short int		remapCount;
	char 			remaps;							// Remap variable
//	unsigned char	copyright;						// variable pascal string
	unsigned char	author;							// variable pascal string
	unsigned char	title;							// variable pascal string
	unsigned char	licensor_contact;
} SongResource_SMS;


typedef enum
{
	R_LAST_RESOURCE		= 0,							// empty. Not used
	R_TITLE				= FOUR_CHAR('T','I','T','L'),	// TITL Byte zero terminated string
	R_PERFORMED_BY		= FOUR_CHAR('P','E','R','F'),	// PERF Byte zero terminated string
	R_COMPOSER			= FOUR_CHAR('C','O','M','P'),	// COMP Byte zero terminated string
	R_COPYRIGHT_DATE	= FOUR_CHAR('C','O','P','D'),	// COPD Byte zero terminated string
	R_COPYRIGHT_LINE	= FOUR_CHAR('C','O','P','L'),	// COPL Byte zero terminated string
	R_PUBLISHER_CONTACT	= FOUR_CHAR('L','I','C','C'),	// LICC Byte zero terminated string
	R_USE_OF_LICENSE	= FOUR_CHAR('L','U','S','E'),	// LUSE Byte zero terminated string
	R_LICENSED_TO_URL	= FOUR_CHAR('L','D','O','M'),	// LDOM Byte zero terminated string
	R_LICENSE_TERM		= FOUR_CHAR('L','T','R','M'),	// LTRM Byte zero terminated string
	R_EXPIRATION_DATE	= FOUR_CHAR('E','X','P','D'),	// EXPD Byte zero terminated string
	R_COMPOSER_NOTES	= FOUR_CHAR('N','O','T','E'),	// NOTE Byte zero terminated string
	R_INDEX_NUMBER		= FOUR_CHAR('I','N','D','X'),	// INDX Byte zero terminated string

	R_INSTRUMENT_REMAP	= FOUR_CHAR('R','M','A','P'),	// RMAP variable amount
	R_VELOCITY_CURVE	= FOUR_CHAR('V','E','L','C')	// VELC 128 words
} SongResourceType;


// Song resource (RMF type)
typedef struct
{
	short int		rmfResourceID;
	char			reserved_0;
	char			reverbType;
	unsigned short	songTempo;
	char			songType;						//	0 - SMS, 1 - RMF, 2 - MOD

	char			locked;							// are resource encrypted
	short int		songPitchShift;
	short int		maxEffects;
	short int		maxNotes;
	short int		mixLevel;
	short int		songVolume;						// 127 is 100%, 256 is 200% etc.
	char			embeddedSong;					// TRUE if embedded in a bank
	char			reserved_1;
	long			unused[7];
	
	short int		resourceCount;
	short int		resourceData;					// subtract this when calculating empty structure
	//
	// from this point on, the data is based upon types and data blocks
//	char			title[1];						// variable C string
//	char			composer[1];					// variable C string
//	char			copyright_date[1];				// variable C string
//	char			copyright_line[1];				// variable C string
//	char			contact_info[1];				// variable C string
//	char			use_license[1];					// variable C string
//	char			license_term[1];				// variable C string
//	char			territory[1];					// variable C string
//	char			expire_date[1];					// variable C string
//	char			foreign_rights[1];				// variable C string
//	char			compser_notes[1];				// variable C string
//	char			index_number[1];				// variable C string
//	short int		velocityCurve[128];
} SongResource_RMF;

// Song resource (MOD type)
typedef struct
{
	short int		modResourceID;
	char			reserved_0;
	char			reverbType;
	unsigned short	songTempo;
	char			songType;						//	0 - SMS, 1 - RMF, 2 - MOD

	char			locked;
	short int		maxEffects;
	short int		maxNotes;
	short int		mixLevel;
	short int		songVolume;						// 100 is 100%, 200 is 200% etc.
	short int		reserved_1;
	long			unused[7];
} SongResource_MOD;

// version structure
typedef struct
{
	short int	versionMajor;
	short int	versionMinor;
	short int	versionSubMinor;
} XVersion;

typedef void SongResource;

// SongResource structure expanded. These values are always in native word order
// Use XGetSongResourceInfo, and XDisposeSongResourceInfo
typedef struct
{
	short int		maxMidiNotes;
	short int		maxEffects;
	short int		mixLevel;
	short int		reverbType;
	short int		objectResourceID;
	short int		songVolume;
	SongType		songType;
	long			songTempo;
	short int		songPitchShift;
	XBOOL			songLocked;

	char			*title;							// 0
	char			*performed;						// 1
	char			*composer;						// 2
	char			*copyright;						// 3
	char			*publisher_contact_info;		// 4
	char			*use_license;					// 5
	char			*licensed_to_URL;				// 6
	char			*license_term;					// 7
	char			*expire_date;					// 8
	char			*compser_notes;					// 9
	char			*index_number;					// 10

	short int		remapCount;
	short int		*remaps;

	short int		*velocityCurve;
} SongResource_Info;


typedef struct
{
	char			lowMidi;
	char			highMidi;
	short int		sndResourceID;
	short int		miscParameter1;		// can be smodParmeter1 if ZBF_enableSoundModifier
										// enabled, otherwise its a replacement
										// rootKey for sample
	short int		miscParameter2;		// if ZBF_enableSoundModifier is enabled then its
										// used as smodParmeter2, otherwise its a volume level
} KeySplit;

// bits for Instrument flags1
#define ZBF_enableInterpolate			0x80
#define ZBF_enableAmpScale				0x40
#define ZBF_disableSndLooping			0x20
#define ZBF_reserved_1					0x10
#define ZBF_useSampleRate				0x08
#define ZBF_sampleAndHold				0x04
#define ZBF_extendedFormat				0x02
#define ZBF_avoidReverb					0x01
// bits for Instrument flags2
#define ZBF_neverInterpolate			0x80
#define ZBF_playAtSampledFreq			0x40
#define ZBF_fitKeySplits				0x20
#define ZBF_enableSoundModifier			0x10
#define ZBF_useSoundModifierAsRootKey	0x08
#define ZBF_notPolyphonic				0x04
#define ZBF_enablePitchRandomness		0x02
#define ZBF_playFromSplit				0x01

#define SET_FLAG_VALUE(oldflag, newflag, value)		(value) ? ((oldflag) | (newflag)) : ((oldflag) & ~(newflag))
#define TEST_FLAG_VALUE(flags, flagbit)				((flags) & (flagbit)) ? TRUE : FALSE

// Special Instrument resource. This can only be used when there is no tremolo data, or key splits
typedef struct
{
	short int		sndResourceID;
	short int		midiRootKey;
	char			panPlacement;
	unsigned char	flags1;				// see ZBF bits for values
	unsigned char	flags2;				// see ZBF bits for values
	char			smodResourceID;
	short int		miscParameter1;		// can be smodParmeter1 if ZBF_enableSoundModifier
										// enabled, otherwise its a replacement
										// rootKey for sample
	short int		miscParameter2;		// if ZBF_enableSoundModifier is enabled then its
										// used as smodParmeter2, otherwise its a volume level
	short int		keySplitCount;		// if this is non-zero, then KeySplit structure is inserted
	// to go beyond this point, if keySplitCount is non-zero, you must use function calls.
	short int		tremoloCount;		// if this is non-zero, then a Word is inserted.
	short int		tremoloEnd;			// Always 0x8000
	short int		reserved_3;
	short int		descriptorName;		// Always 0
	short int		descriptorFlags;	// Always 0
} InstrumentResource;

#define DEFAULT_RESOURCE_BANK_ID	0	// ID used inside of RMF file
typedef struct
{
	unsigned long	version;
	char			bankURL[1024];
	char			bankName[1024];
} BankStatus;

// audioType for the AudioResource structure
enum
{
	AUDIO_NAME_TYPE	=	FOUR_CHAR('A','T','X','T'),	//	'ATXT'		// non-terminated string. Length of string is
													//				// length of resource
	AUDIO_SND		=	FOUR_CHAR('s','n','d',' '),	//	'snd '		// MacOS 'snd' format
	AUDIO_CSND		=	FOUR_CHAR('c','s','n','d'),	//	'csnd'		// MacOS 'snd' format compressed
	AUDIO_ESND		=	FOUR_CHAR('e','s','n','d'),	//	'esnd'		// MacOS 'snd' format encrypted
	AUDIO_MPEG		=	FOUR_CHAR('M','P','E','G'),	//	'MPEG'		// MPEG
	AUDIO_IMA2_PCM	=	FOUR_CHAR('I','M','A','2'),	//	'IMA2'		// IMA PCM 2 to 1
	AUDIO_IMA4_PCM	=	FOUR_CHAR('I','M','A','4'),	//	'IMA4'		// IMA PCM 4 to 1
	AUDIO_RAW_PCM	=	FOUR_CHAR('R','P','C','M')	//	'RPCM'		// raw pcm data
};

#define AUDIO_OBJECT_VERSION			0x0001
typedef struct
{
	unsigned long	version;			// structure version 1
	unsigned long	dataLength;			// length of sample data in bytes
	unsigned long	dataOffset;			// offset from begining of structure to data
	unsigned long	audioType;			// audio type of audioType
	unsigned long	usageType;
	unsigned long	sampleRate;			// sample rate in 16.16 fixed point
	unsigned long	sampleFrames;		// number of sample frames
	unsigned long	loopStart;			// first loop start
	unsigned long	loopEnd;			// first loop end
	short int		baseMidiKey;		// base root midi key
	short int		bitSize;			// 8 or 16 bits per sample
	short int 		channels;			// 1 or 2 channels
	unsigned long	nameResourceType;	// Resource name type. ie (AUDIO_NAME_TYPE)
										// if ID_NULL, then no name
	unsigned long	nameResourceID;		// Resource name id. ie AUDIO_NAME_TYPE ID 2000
	char			usedInBank;			// if true, then sample is embedded in a bank
	char			unusedFlag2;
	char			unusedFlag3;
	char			unusedFlag4;
	unsigned long	filler[16];
	unsigned long	firstSampleFiller;
//	data
} AudioResource;

// Sun compiler warns about this as an enum
#define rate48khz			0xBB800000L	/*48000.00000 in fixed-point*/

// These are included here, because we want to be independent of MacOS, but use this standard format
#ifndef __SOUND__

// Sun compiler warns about this as an enum
#define rate44khz			0xAC440000L,	/*44100.00000 in fixed-point*/

enum 
{
	notCompressed			= 0,			/*compression ID's*/
	fixedCompression		= -1,			/*compression ID for fixed-sized compression*/
	variableCompression		= -2,			/*compression ID for variable-sized compression*/
	twoToOne				= 1,
	eightToThree			= 2,
	threeToOne				= 3,
	sixToOne				= 4,

	stdSH					= 0x00,			/*Standard sound header encode value*/
	extSH					= 0xFF,			/*Extended sound header encode value*/
	cmpSH					= 0xFE,			/*Compressed sound header encode value*/

	rate22050hz				= 0x56220000L,	/*22050.00000 in fixed-point*/
	rate22khz				= 0x56EE8BA3L,	/*22254.54545 in fixed-point*/
	rate11khz				= 0x2B7745D1L,	/*11127.27273 in fixed-point*/
	rate11025hz				= 0x2B110000,	/*11025.00000 in fixed-point*/

	kMiddleC				= 60,			/*MIDI note value for middle C*/

	soundCmd				= 80,
	bufferCmd				= 81,
	firstSoundFormat		= 0x0001,		/*general sound format*/
	secondSoundFormat		= 0x0002		/*special sampled sound format (HyperCard)*/
};
#endif

typedef struct
{
	char					*samplePtr;		/*if NIL then samples are in sampleArea*/
	unsigned long			length;			/*length of sound in bytes*/
	unsigned long			sampleRate;		/*sample rate for this sound*/
	unsigned long			loopStart;		/*start of looping portion*/
	unsigned long			loopEnd;		/*end of looping portion*/
	unsigned char			encode;			/*header encoding*/
	unsigned char			baseFrequency;	/*baseFrequency value*/
	unsigned char			sampleArea[1];	/*space for when samples follow directly*/
} XSoundHeader;
typedef XSoundHeader *XSoundHeaderPtr;

typedef struct
{
	char					*samplePtr;		/*if nil then samples are in sample area*/
	unsigned long			numChannels;	/*number of channels i.e. mono = 1*/
	unsigned long			sampleRate;		/*sample rate in Apples Fixed point representation*/
	unsigned long			loopStart;		/*loopStart of sound before compression*/
	unsigned long			loopEnd;		/*loopEnd of sound before compression*/
	unsigned char			encode;			/*data structure used , stdSH, extSH, or cmpSH*/
	unsigned char			baseFrequency;	/*same meaning as regular SoundHeader*/
	unsigned long			numFrames;		/*length in frames ( packetFrames or sampleFrames )*/
	char					AIFFSampleRate[10];	/*IEEE sample rate*/
	char					*markerChunk;	/*sync track*/
	long					format;			/*data format type, was futureUse1*/
	char					futureUse2_0;	/*reserved by Apple, Igor will use as IMA encoder to 8 or 16 bit output. Set to 0x80 */
											// to encode as 8 bit output
	char					soundIsEmbedded;	/*reserved by Apple. Igor uses it as a flag */
	char					futureUse2_2;	/*reserved by Apple*/
	char					futureUse2_3;	/*reserved by Apple*/
	void					*stateVars;		/*pointer to State Block*/
	void					*leftOverSamples;	/*used to save truncated samples between compression calls*/
	short					compressionID;	/*0 means no compression, non zero means compressionID*/
	unsigned short			packetSize;		/*number of bits in compressed sample packet*/
	unsigned short			snthID;			/*resource ID of Sound Manager snth that contains NRT C/E*/
	unsigned short			sampleSize;		/*number of bits in non-compressed sample*/
	unsigned char			sampleArea[1];	/*space for when samples follow directly*/
} XCmpSoundHeader;
typedef XCmpSoundHeader * XCmpSoundHeaderPtr;

typedef struct
{
	char					*samplePtr;		/*if nil then samples are in sample area*/
	unsigned long			numChannels;	/*number of channels,  ie mono = 1*/
	unsigned long			sampleRate;		/*sample rate in Apples Fixed point representation*/
	unsigned long			loopStart;		/*same meaning as regular SoundHeader*/
	unsigned long			loopEnd;		/*same meaning as regular SoundHeader*/
	unsigned char			encode;			/*data structure used , stdSH, extSH, or cmpSH*/
	unsigned char			baseFrequency;	/*same meaning as regular SoundHeader*/
	unsigned long			numFrames;		/*length in total number of frames*/
	char					AIFFSampleRate[10];	/*IEEE sample rate*/
	char					*markerChunk;	/*sync track*/
	char					*instrumentChunks;	/*AIFF instrument chunks*/
	char					*AESRecording;
	unsigned short			sampleSize;		/*number of bits in sample*/
	char					soundIsEmbedded;	/*reserved by Apple. Igor uses it as a flag */
	char					futureUse1_1;	/*reserved by Apple*/
	unsigned long			futureUse2;		/*reserved by Apple*/
	unsigned long			futureUse3;		/*reserved by Apple*/
	unsigned long			futureUse4;		/*reserved by Apple*/
	unsigned char			sampleArea[1];	/*space for when samples follow directly*/
} XExtSoundHeader;
typedef XExtSoundHeader *XExtSoundHeaderPtr;

typedef struct
{
	short int		type;
	short int		numModifiers;
	unsigned short	modNumber;
	long			modInit;
	short int		numCommands;
// first command
	unsigned short	cmd;
	short int		param1;
	long			param2;
} XSoundFormat1;

typedef struct
{
	XSoundFormat1	sndHeader;
	XSoundHeader	sndBuffer;
} XSndHeader1;

typedef struct
{
	XSoundFormat1	sndHeader;
	XExtSoundHeader	sndBuffer;
} XExtSndHeader1;


typedef struct
{
	short int		type;
	short int		refCount;
	short int		numCmds;
// first command
	unsigned short	cmd;
	short int		param1;
	long			param2;
} XSoundFormat2;

typedef struct
{
	XSoundFormat2	sndHeader;
	XSoundHeader	sndBuffer;
} XSndHeader2;

typedef struct
{
	XSoundFormat2	sndHeader;
	XExtSoundHeader	sndBuffer;
} XExtSndHeader2;

#if CPU_TYPE == kRISC
	#pragma options align=reset
#endif

// $$jdr: This should be pack(), but for a
// bug in the compilers we're using.
#if (CPU_TYPE == kSPARC)
    #pragma pack(4)
#endif

/* $$kk: pack(4) for solaris x86 */
#if (CPU_TYPE == k80X86)
#if (X_PLATFORM == X_SOLARIS)
	#pragma pack (4)
#else
	#pragma pack ()
#endif
#endif

enum 
{
// Compression Types
	// these are used in the snd formatted resource
	C_NONE				= FOUR_CHAR('n','o','n','e'),	// 'none'
	C_IMA4				= FOUR_CHAR('i','m','a','4'),	// 'ima4'	CCITT G.721 ADPCM compression (IMA 4 to 1)
	C_MACE3				= FOUR_CHAR('m','a','c','3'),	// 'mac3'	Apple MACE type 3 to 1
	C_MACE6				= FOUR_CHAR('m','a','c','6'),	// 'mac6'	Apple MACE type 6 to 1
	C_ULAW				= FOUR_CHAR('u','l','a','w'),	// 'ulaw'	ulaw; 2 to 1

	// these are used in AIFF/AIFC formatted files
	X_NONE				= FOUR_CHAR('N','O','N','E'),	//	'NONE'
	X_ACE2				= FOUR_CHAR('A','C','E','2'),	//	'ACE2'
	X_ACE8				= FOUR_CHAR('A','C','E','8'),	//	'ACE8'
	X_MACE3				= FOUR_CHAR('M','A','C','3'),	//	'MAC3'
	X_MACE6				= FOUR_CHAR('M','A','C','6'),	//	'MAC6'
	X_IMA4				= FOUR_CHAR('i','m','a','4')	//	'ima4'
};


typedef struct
{
	unsigned long	rate;				// sample rate
	unsigned long	frames;				// number of audio frames
	unsigned long	size;				// size in bytes
	unsigned long	loopStart;			// loop start frame
	unsigned long	loopEnd;			// loop end frame
	short int		bitSize;			// sample bit size; 8 or 16
	short int		channels;			// mono or stereo; 1 or 2
	short int		baseKey;			// base sample key
	short int		theID;				// sample ID if required
	long			compressionType;	// compression type
	void			*pMasterPtr;		// master pointer if required
} SampleDataInfo;


typedef enum
{
	I_INVALID = 0,				// invalid type
	I_TITLE,					// Title
	I_PERFORMED_BY,				// Performed by
	I_COMPOSER,					// Composer(s)
	I_COPYRIGHT,				// Copyright Date
	I_PUBLISHER_CONTACT,		// Publisher Contact Info
	I_USE_OF_LICENSE,			// Use of License
	I_LICENSED_TO_URL,			// License to URL
	I_LICENSE_TERM,				// License term
	I_EXPIRATION_DATE,			// Expiration Date
	I_COMPOSER_NOTES,			// Composer Notes
	I_INDEX_NUMBER				// Index Number
} SongInfo;

void XGetSongInformation(SongResource *theSong, long songSize, SongInfo type, char *cName);

unsigned long XGetSongInformationSize(SongResource *theSong, long songSize, SongInfo type);

// Utillities for instruments
XBOOL XIsSoundUsedInInstrument(InstrumentResource *theX, short int sampleSoundID);
void XRenumberSampleInsideInstrument(InstrumentResource *theX, short int originalSampleID, 
																short int newSampleID);
short int XCollectSoundsFromInstrument(InstrumentResource *theX, short *sndArray, short maxArraySize);
short int XCollectSoundsFromInstrumentID(short int theID, short *sndArray, short maxArraySize);
XBOOL XCheckAllInstruments(short *badInstrument, short *badSnd);
short int XCheckValidInstrument(short int theID);
short int XGetInstrumentArray(short int *instArray, short maxArraySize);

short int XGatherAllSoundsFromAllInstruments(short int *pSndArray, short int maxArraySize);

XBOOL XIsSampleUsedInAllInstruments(short int soundSampleID, short int *pWhichInstrument);

// Given a song ID and two arrays, this will return the INST resources ID and the 'snd ' resource ID
// that are needed to load the song terminated with a -1.
// Will return 0 for success or 1 for failure
OPErr XGetSongInstrumentList(short int theSongID, short int *pInstArray, short int maxInstArraySize, 
										short int *pSndArray, short int maxSndArraySize);

short int XGetSamplesFromInstruments(short int *pInstArray, short int maxInstArraySize, 
										short int *pSndArray, short int maxSndArraySize);

void XGetKeySplitFromPtr(InstrumentResource *theX, short int entry, KeySplit *keysplit);
void XSetKeySplitFromPtr(InstrumentResource *theX, short int entry, KeySplit *keysplit);

InstrumentResource * XAddKeySplit(InstrumentResource *theX, short int howMany);
InstrumentResource * XRemoveKeySplit(InstrumentResource *theX, short int howMany);

short int XGetTotalKeysplits(short int *instArray, short int totalInstruments, 
								short int *sndArray, short int totalSnds);


// Create a new song resource.
SongResource * XNewSongPtr(	SongType songType, 
							short int midiID,
							short int maxSongVoices, 
							short int mixLevel, 
							short int maxEffectVoices,
							ReverbMode reverbType);

void XDisposeSongPtr(SongResource *theSong);

XPTR XGetSoundResourceByID(long theID, long *pSize);
XPTR XGetSoundResourceByName(void *cName, long *pSize);

XPTR XGetMidiData(long theID, long *pSize);
XPTR XCompressAndEncrypt(XPTR pData, long size, long *pNewSize);

XPTR XGetSamplePtrFromSnd(XPTR pRes, SampleDataInfo *pInfo);

void XSetSoundLoopPoints(XPTR pRes, long loopStart, long loopEnd);
void XSetSoundSampleRate(XPTR pRes, unsigned long sampleRate);
void XSetSoundBaseKey(XPTR pRes, short int baseKey);
short int XGetSoundBaseKey(XPTR pRes);

XBOOL XGetSoundEmbeddedStatus(XPTR pRes);
void XSetSoundEmbeddedStatus(XPTR pRes, XBOOL soundEmbedded);

// Given a structure filled with no compression, create a type 1 sound object.
// Will return NULL if out of memory, failed to get the right info from the
// SampleDataInfo structure, and if pSampleInfo->pMasterPtr is NULL.
XPTR XCreateSoundObjectFromData(XPTR pPCMData, SampleDataInfo *pSampleInfo);

XBOOL XGetSampleNameFromID(short int sampleSoundID, char *cName);

InstrumentResource * XNewInstrument(short int leadSndID);
void XDisposeInstrument(InstrumentResource *theX);

SongResource_Info * XGetSongResourceInfo(SongResource *pSong, long songSize);
void XDisposeSongResourceInfo(SongResource_Info *pSongInfo);

SongResource * XNewSongFromSongResourceInfo(SongResource_Info *pSongInfo);

short int XGetSongResourceObjectID(SongResource *pSong);
void XSetSongResourceObjectID(SongResource *pSong, short int id);
XBOOL XIsSongLocked(SongResource *pSong);
void XSetSongLocked(SongResource *pSong, XBOOL locked);
void XGetSongPerformanceSettings(SongResource * theSong, short int *maxMidiVoices, 
									short int *maxEffectsVoices, short int *mixLevel);
void XSetSongPerformanceSettings(SongResource *pSong, short int maxMidiVoices, short int maxEffectsVoices,
										short int mixLevel);

short int XGetSongReverbType(SongResource *pSong);
void XSetSongReverbType(SongResource *pSong, short int reverbType);

short int XGetSongVolume(SongResource *pSong);
void XSetSongVolume(SongResource *pSong, short int volume);

long XGetSongTempoFactor(SongResource *pSong);
void XSetSongTempoFactor(SongResource *pSong, long newTempo);

SongType XGetSongResourceObjectType(SongResource *pSong);

SongResource * XChangeSongResource(SongResource *theSong, long songSize, 
					SongResourceType resourceType, void *pResource, long resourceLength);

XBOOL XGetSongEmbeddedStatus(SongResource *pSong);
void XSetSongEmbeddedStatus(SongResource *pSong, XBOOL embedded);

// Translate a GM_Waveform structure into a SampleDataInfo structure
void XTranslateFromWaveformToSampleData(GM_Waveform *pSource, SampleDataInfo *pDest);

// Create a bank resource from a BankStatus structure. This data can be written out
// as a BANK_ID resource
XPTR XCreateBankStatus(BankStatus *pStatus);
// Get bank resource from currently open resource file
void XGetBankStatus(BankStatus *pStatus);

#ifdef __cplusplus
	}
#endif

#endif	// X_FORMATS


