/*
 * @(#)alphaCompst_paintloop.h	1.15 98/07/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <alphaCompst_XOR_MODE.h>
#include <alphaCompst_SRC.h>
#include <alphaCompst_SRC_OVER.h>
#include <alphaCompst_DST_OVER.h>
#include <alphaCompst_SRC_IN.h>
#include <alphaCompst_DST_IN.h>
#include <alphaCompst_SRC_OUT.h>
#include <alphaCompst_DST_OUT.h>
#include <alphaCompst_loop_uti.h>


extern PaintLoopFunc    TILE_BLENDFUNC;

void
TILE_BLENDFUNC(JNIEnv *env,
	       void *dstBase, jint dstScanStr, jint dstPixStr,
               void *srcBase, jint srcScanStr, jint srcPixStr,
	       unsigned char *pathData, jint offset,
	       jint alphaTileScan, jint rule, jfloat fXtraAlpha,
               jint width, jint height, jboolean srcPre, jint xorColor,
	       ColorData *cData)
{
    int extraAlpha = (int)(fXtraAlpha*(float)NORM_SCALE);
    int row, column;
    int rastAlpha = 256;
    int compAlpha = 0;
    int srcAlpha;
    int srcRed;
    int srcGreen;
    int srcBlue;
    int dstAlpha;
    int dstRed;
    int dstGreen;
    int dstBlue;
    int pathDataInc = 1;
    int xorRed   = ((xorColor >> 16) & 0xff);
    int xorGreen = ((xorColor >>  8) & 0xff);
    int xorBlue  = ((xorColor >>  0) & 0xff);
    int skipDst;
    int skipDither = 0;


    DeclareSrc
    DeclareDst

    InitSrc;
    InitRastAlpha;

    InitDst;

    switch (rule) {

    case XOR_MODE:
        skipDither = 1;
#ifdef FastXORLoop
        FastXORLoop;
#else
        COMPOSITE_TILE(EMPTY_STATEMENT,RASTEROP_XOR(rastAlpha, compAlpha,
                                                    extraAlpha,
                                                    srcAlpha,
                                                    (srcRed ^ xorRed),
                                                    (srcGreen ^ xorGreen),
                                                    (srcBlue ^ xorBlue),
                                                    dstAlpha, dstRed,
                                                    dstGreen, dstBlue));
#endif
        skipDither = 0;
        break;

    case CLEAR:
        CLEAR_COMP;
        break;

    case SRC:
#define SKIP_READ_DST_PIXEL
#ifdef FastCopyLoop
    if ((pathData  == 0) && (extraAlpha == 256)) {
        FastCopyLoop;
    } else {
        COMPOSITE_TILE(EMPTY_STATEMENT, SRC_RULE(rastAlpha, compAlpha,
                                                 extraAlpha,
                                                 srcAlpha, srcRed,
                                                 srcGreen, srcBlue,
                                                 dstAlpha, dstRed,
                                                 dstGreen, dstBlue));
    }
#else
    COMPOSITE_TILE(EMPTY_STATEMENT, SRC_RULE(rastAlpha, compAlpha, extraAlpha,
                                             srcAlpha, srcRed,
                                             srcGreen, srcBlue,
                                             dstAlpha, dstRed,
                                             dstGreen, dstBlue));
#endif
#undef SKIP_READ_DST_PIXEL
        break;

    case SRC_OVER:
        COMPOSITE_TILE(EMPTY_STATEMENT, SRC_OVER_RULE(rastAlpha, compAlpha,
                                                      extraAlpha,
                                                      srcAlpha, srcRed,
                                                      srcGreen, srcBlue,
                                                      dstAlpha, dstRed,
                                                      dstGreen, dstBlue));
        break;

    case DST_OVER:
        COMPOSITE_TILE(EMPTY_STATEMENT, DST_OVER_RULE(rastAlpha, compAlpha,
                                                      extraAlpha,
                                                      srcAlpha, srcRed,
                                                      srcGreen, srcBlue,
                                                      dstAlpha, dstRed,
                                                      dstGreen, dstBlue));
        break;

    case SRC_IN:
        COMPOSITE_TILE(EMPTY_STATEMENT, SRC_IN_RULE(rastAlpha, compAlpha,
                                                    extraAlpha,
                                                    srcAlpha, srcRed,
                                                    srcGreen, srcBlue,
                                                    dstAlpha, dstRed,
                                                    dstGreen, dstBlue));
        break;

    case DST_IN:
        COMPOSITE_TILE(EMPTY_STATEMENT, DST_IN_RULE(rastAlpha, compAlpha,
                                                    extraAlpha,
                                                    srcAlpha, srcRed,
                                                    srcGreen, srcBlue,
                                                    dstAlpha, dstRed,
                                                    dstGreen, dstBlue));
        break;

    case SRC_OUT:
        COMPOSITE_TILE(EMPTY_STATEMENT, SRC_OUT_RULE(rastAlpha, compAlpha,
                                                     extraAlpha,
                                                     srcAlpha, srcRed,
                                                     srcGreen, srcBlue,
                                                     dstAlpha, dstRed,
                                                     dstGreen, dstBlue));
        break;

    case DST_OUT:
        COMPOSITE_TILE(EMPTY_STATEMENT, DST_OUT_RULE(rastAlpha, compAlpha,
                                                     extraAlpha,
                                                     srcAlpha, srcRed,
                                                     srcGreen, srcBlue,
                                                     dstAlpha, dstRed,
                                                     dstGreen, dstBlue));
        break;

    }
}
