/*
 * @(#)CMM.c	1.9 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/* File:  CMM.c  @(#)CMM.c	1.7  11/03/97
 * Native functions for the Java CMM
 */

/**********************************************************************
 **********************************************************************
 **********************************************************************
 *** COPYRIGHT (c) Eastman Kodak Company, 1997                      ***
 *** As  an unpublished  work pursuant to Title 17 of the United    ***
 *** States Code.  All rights reserved.                             ***
 **********************************************************************
 **********************************************************************
 **********************************************************************/


#include <stdio.h>
#include "sun_awt_color_CMM.h"
#include "sprofile.h"
#include "sprof-pr.h"


#define JAVAsearchCMM (1)
#define JAVAsearchVersion (2)
#define JAVAsearchClass (3)
#define JAVAsearchColorSpace (4)
#define JAVAsearchPCS (5)
#define JAVAsearchBeforeDate (6)
#define JAVAsearchOnDate (7)
#define JAVAsearchAfterDate (8)
#define JAVAsearchPlatform (9)
#define JAVAsearchFlags (10)
#define JAVAsearchManufacturer (11)
#define JAVAsearchModel (12)
#define JAVAsearchAttributes (13)
#define JAVAsearchRenderingIntent (14)
#define JAVAsearchIlluminant (15)
#define JAVAsearchCreator (16)

#define JAVAsearchSize (16)
#define JAVAsearchTrue (1)
#define JAVAsearchFalse (0)

#define JAVA_ICC_TransformAny (-1)

#define JAVA_TYPE_INT_RGB (1)
#define JAVA_TYPE_INT_ARGB (2)
#define JAVA_TYPE_INT_ARGB_PRE (3)
#define JAVA_TYPE_3BYTE_BGR (5)
#define JAVA_TYPE_4BYTE_ABGR (6)
#define JAVA_TYPE_4BYTE_ABGR_PRE (7)

#define typeCompUByte (256)
#define typeCompUShort12 (257)
#define typeCompUShort (258)
#define typePelUByte (259)
#define typePelUShort12 (260)
#define typePelUShort (261)

#define SpTagHeader    SpTagIdConst('h', 'e', 'a', 'd')

/* Rendering Intents, used in the profile header */
typedef enum {
    icPerceptual            = 0,
    icRelativeColorimetric        = 1,
    icSaturation            = 2,
    icAbsoluteColorimetric        = 3,
    icMaxEnumIntent            = 0xFFFFFFFF    /* enum = 4 bytes max */
} icRenderingIntent;

typedef    jbyte* jbyte_p;
typedef    jshort* jshort_p;
typedef    jint* jint_p;
typedef    jlong* jlong_p;
typedef    SpXform_t* SpXform_p;
typedef    SpProfile_t* SpProfile_p;
typedef    SpPixelLayout_t* SpPixelLayout_p;

typedef union storeID_s {    /* store SProfile stuff in a Java Long */
    SpCallerId_t    cid;
    SpProfile_t        pf;
    SpXform_t        xf;
    jlong            j;
} storeID_t, *storeID_p;

static SpCallerId_t getCallerID (JNIEnv*);
static SpStatus_t initImageLayout (JNIEnv*, jobject, SpPixelLayout_p);
static void releaseArrayData (JNIEnv*, jobject, SpPixelLayout_p);
static void pfToID (JNIEnv*, jlongArray, SpProfile_p, KpInt32_t);
static SpStatus_t criteriaFromHeader (JNIEnv*, jbyteArray, jbyteArray, SpSearch_t*, SpSearchCriterion_t*);
static void returnInt (JNIEnv *env, jintArray theReturn, jint theInt);
static jint checkStatus (SpStatus_t theStatus);
static void setObjectID (JNIEnv *env, jobject theTransform, storeID_t theXform);
static SpStatus_t setClassID (JNIEnv *env, jclass theClass, SpCallerId_t theID);
static SpCallerId_t getClassID (JNIEnv *env, jclass theClass);

#define DIAG_PRINTF1(x) ((void)0)
#define DIAG_PRINTF2(x,y) ((void)0)
#define DIAG_PRINTF3(x,y,z) ((void)0)
#define DIAG_PRINTF4(x,y,z,a) ((void)0)
/* #define DIAG_PRINTF1 printf */
/* #define DIAG_PRINTF2 printf */
/* #define DIAG_PRINTF3 printf*/
/* #define DIAG_PRINTF4 printf*/
/* #define DIAG_PRINTF printf */
/* #define DIAG_PRINTF dummyPrintf */ /* */
/* static void dummyPrintf (const char * control, ...) {if (control) {}} */ /* */



/* Class:     sun_awt_color_CMM
 * Method:    cmmInit
 * Signature: ()I
 * static native int cmmInit();
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmInit (JNIEnv *env, jclass class)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmInit\n");

    theStatus = SpInitialize (&theCallerId, NULL, NULL);

    if (theStatus == SpStatSuccess) {
        theStatus = setClassID (env, class, theCallerId);
    }

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmTerminate
 * Signature: ()I
 * static native int cmmTerminate ();
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmTerminate (JNIEnv *env, jclass class)
{
SpStatus_t    theStatus = SpStatSuccess, aStatus;
SpCallerId_t    theCallerId;

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmTerminate\n");

    theCallerId = getClassID (env, class);
    if (theCallerId != NULL) {
        theStatus = SpTerminate (&theCallerId);
    }

    theCallerId = NULL;
    aStatus = setClassID (env, class, theCallerId);

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmLoadProfile
 * Signature: ([B[J)I
 * static native int cmmLoadProfile (byte[] data, long[] profileID);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmLoadProfile (JNIEnv *env, jobject obj, jbyteArray data, jlongArray profileID)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;
jbyte_p        dataP;
jlong_p        longP;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmLoadProfile\n");

    theSProfile.pf = NULL;

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        dataP = (*env)->GetByteArrayElements (env, data, 0);    /* get the data address */

        theStatus = SpProfileLoadFromBuffer (theCallerId, dataP, &theSProfile.pf);

        (*env)->ReleaseByteArrayElements (env, data, dataP, 0);    /* don't need to hold data */

        if (theStatus == SpStatSuccess) {
            DIAG_PRINTF2("theSProfile %x\n", theSProfile.pf);
        }
    }

    longP = (*env)->GetLongArrayElements (env, profileID, 0);

    longP[0] = theSProfile.j;                    /* return the profile ID in the first element */

    (*env)->ReleaseLongArrayElements (env, profileID, longP, 0);

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmFreeProfile
 * Signature: (J)I
 * static native int cmmFreeProfile (long profileID);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmFreeProfile (JNIEnv *env, jobject obj, jlong profileID)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmFreeProfile\n");

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theSProfile.j = profileID;

        theStatus = SpProfileFree (&theSProfile.pf);
    }

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmGetProfileSize
 * Signature: (J[I)I
 * static native int cmmGetProfileSize (long profileID, int[] size);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmGetProfileSize (JNIEnv *env, jobject obj, jlong profileID, jintArray size)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;
KpUInt32_t    profileSize = 0;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmGetProfileSize\n");

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theSProfile.j = profileID;

        theStatus = SpProfileGetProfileSize (theSProfile.pf, &profileSize);    /* get size of this profile */

        DIAG_PRINTF3("theSProfile %x; size %d\n", theSProfile.pf, profileSize);
    }

    returnInt (env, size, profileSize);

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmGetProfileData
 * Signature: (J[B)I
 * static native int cmmGetProfileData (long profileID, byte[] data);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmGetProfileData (JNIEnv *env, jobject obj, jlong profileID, jbyteArray data)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;
jbyte_p        dataP;
KpUInt32_t    bufferSize;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmGetProfileData\n");

    theCallerId = getCallerID (env);    /* not actually needed, but must be there */
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theSProfile.j = profileID;
        DIAG_PRINTF2("theSProfile = %x\n", theSProfile.pf);

        dataP = (*env)->GetByteArrayElements(env, data, 0);                /* get the data address */
        bufferSize = (KpUInt32_t) (*env)->GetArrayLength (env, data);    /* and size */

        theStatus = SpProfileSaveToBuffer (theSProfile.pf, (KpChar_h)&dataP, &bufferSize);

        (*env)->ReleaseByteArrayElements (env, data, dataP, 0);            /* don't need to hold data */
    }

    return checkStatus (theStatus);
}

/*
 * Class:     sun_awt_color_CMM
 * Method:    cmmCullICC_Profiles
 * Signature: ([B[B[J[J[I)I
 * static native int cmmCullICC_Profiles (byte[] template, byte[] options, long[] sourceIDs, long[] matchIDs, int[] nMatch);
 */
JNIEXPORT jint JNICALL Java_sun_awt_color_CMM_cmmCullICC_1Profiles
  (JNIEnv *env, jobject obj, jbyteArray template, jbyteArray options, jlongArray sourceIDs, jlongArray matchIDs, jintArray nMatch)
{
SpStatus_t        theStatus;
SpCallerId_t    theCallerId;
storeID_t        theSProfile;
KpInt32_t        i1, srcSize;
jlong_p            javaIDlist;
SpProfile_p        srcProfiles;
KpInt32_t        foundCount = 0;
SpSearch_t        theSearch;
SpSearchCriterion_t    theCriteria [SPSEARCH_LIST];

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmCullICC_1Profiles\n");

    theCallerId = getCallerID (env);    /* not actually needed, but must be there */
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        srcSize = (KpUInt32_t) (*env)->GetArrayLength (env, sourceIDs);                    /* # of source profiles */
        srcProfiles = (SpProfile_p) allocBufferPtr (srcSize * sizeof (SpProfile_t));    /* get memory for it */

        if (srcProfiles == NULL) {
            theStatus = SpStatMemory;
        }
        else {
            javaIDlist = (*env)->GetLongArrayElements (env, sourceIDs, 0);

            DIAG_PRINTF1("srcProfiles =");
            for (i1 = 0; i1 < srcSize; i1++) {
                theSProfile.j = javaIDlist [i1];
                srcProfiles [i1] = theSProfile.pf;        /* get profile IDs from java IDs list */
                DIAG_PRINTF2(" %x", srcProfiles [i1]);
            }
            DIAG_PRINTF1("\n");
                
            (*env)->ReleaseLongArrayElements (env, sourceIDs, javaIDlist, 0);

            theStatus = criteriaFromHeader (env, template, options, &theSearch, theCriteria);

            if (theStatus == SpStatSuccess) {
                theStatus = SpProfileSearchRefine (&theSearch, srcProfiles, srcSize, &foundCount);

                pfToID (env, matchIDs, srcProfiles, foundCount);
            }

            freeBufferPtr ((KpGenericPtr_t) srcProfiles);            /* free profile list memory */
        }
    }

    returnInt (env, nMatch, foundCount);

    return checkStatus (theStatus);
}



/*
 * Class:     sun_awt_color_CMM
 * Method:    cmmFindICC_Profiles
 * Signature: (JLjava/lang/String;[J[I)I
 * static native int cmmFindICC_Profiles (byte[] template, byte[] options, String profileDir, long[] matchIDs, int[] nMatch);
 */
JNIEXPORT jint JNICALL Java_sun_awt_color_CMM_cmmFindICC_1Profiles
  (JNIEnv *env, jobject obj, jbyteArray template, jbyteArray options, jstring profileDir, jlongArray matchIDs, jintArray nMatch)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
KpUInt32_t    listSize;
SpProfile_p    profileList;
KpInt32_t    foundCount = 0;
SpDataBase_t        theDB;
SpDataBaseEntry_t    theDirectory;
SpSearch_t        theSearch;
SpSearchCriterion_t    theCriteria [SPSEARCH_LIST];

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmFindICC_1Profiles\n");

    theCallerId = getCallerID (env);        /* not actually needed, but must be there */
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        listSize = (KpUInt32_t) (*env)->GetArrayLength (env, matchIDs);                    /* get the max # of matched profiles */
        profileList = (SpProfile_p) allocBufferPtr (listSize * sizeof (SpProfile_t));    /* get memory for it */

        if (profileList == NULL) {
            theStatus = SpStatMemory;
        }
        else {
            /* make the DataBaseList */
            theDB.numEntries = 1;
            theDB.Entries = &theDirectory;
#if defined(KPMAC) || defined (KPMSMAC)
            theDirectory.fileProps.vRefNum = 0;
#endif
            theDirectory.dirName = (KpChar_p) (*env)->GetStringUTFChars(env, profileDir, 0);

            DIAG_PRINTF2("theDirectory.dirName = %s\n", &theDirectory.dirName);

            theStatus = criteriaFromHeader (env, template, options, &theSearch, theCriteria);

            if (theStatus == SpStatSuccess) {
                theStatus = SpProfileSearch (theCallerId, &theDB, &theSearch, profileList, listSize, &foundCount);

                pfToID (env, matchIDs, profileList, foundCount);
            }

            (*env)->ReleaseStringUTFChars (env, profileDir, theDirectory.dirName);    /* release directory name string */

            freeBufferPtr ((KpGenericPtr_t) profileList);                    /* free profile list memory */
        }
    }

    returnInt (env, nMatch, foundCount);

    return checkStatus (theStatus);
}



/* convert a profile's header into a list of criteria */
static SpStatus_t
    criteriaFromHeader (    JNIEnv*            env,
                jbyteArray        template,
                jbyteArray        options,
                SpSearch_t*        theSearch,
                SpSearchCriterion_t*    theCriteria)
{
SpStatus_t    theStatus;
SpHeader_t    theHeader;
KpInt32_t    critCount = 0;
jbyte_p        dataP, control;

    dataP = (*env)->GetByteArrayElements (env, template, 0);            /* get the template data */

    theStatus = SpHeaderToPublic ((KpInt8_p)dataP, HEADER_SIZE, &theHeader);    /* turn it into a header */

    (*env)->ReleaseByteArrayElements (env, template, dataP, 0);

    control = (*env)->GetByteArrayElements (env, options, 0);            /* get the search control data */

    if (theStatus == SpStatSuccess) {
        if (control[JAVAsearchCMM] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_PREFERREDCMM;
            theCriteria [critCount].SearchValue.Signature = theHeader.CMMType;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_PREFERREDCMM; %.4s\n", &theHeader.CMMType);
        }

        if (control[JAVAsearchVersion] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_VERSION;
            theCriteria [critCount].SearchValue.Value = theHeader.ProfileVersion;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_VERSION; %x\n", theHeader.ProfileVersion);
        }

        if (control[JAVAsearchClass] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_PROFILECLASS;
            theCriteria [critCount].SearchValue.Signature = theHeader.DeviceClass;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_PROFILECLASS; %.4s\n", &theHeader.DeviceClass);
        }

        if (control[JAVAsearchColorSpace] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_DEVICECOLORSPACE;
            theCriteria [critCount].SearchValue.Signature = theHeader.DataColorSpace;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_DEVICECOLORSPACE; %.4s\n", &theHeader.DataColorSpace);
        }

        if (control[JAVAsearchPCS] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_CONNECTIONSPACE;
            theCriteria [critCount].SearchValue.Signature = theHeader.InterchangeColorSpace;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_CONNECTIONSPACE; %.4s\n", &theHeader.InterchangeColorSpace);
        }

        if (control[JAVAsearchBeforeDate] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_BEFOREDATE;
            theCriteria [critCount].SearchValue.Date = theHeader.DateTime;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_BEFOREDATE; %lx\n", theHeader.DateTime);
        }

        if (control[JAVAsearchOnDate] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_ONDATE;
            theCriteria [critCount].SearchValue.Date = theHeader.DateTime;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_ONDATE; %lx\n", theHeader.DateTime);
        }

        if (control[JAVAsearchAfterDate] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_AFTERDATE;
            theCriteria [critCount].SearchValue.Date = theHeader.DateTime;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_AFTERDATE; %lx\n", theHeader.DateTime);
        }

        if (control[JAVAsearchPlatform] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_PLATFORM;
            theCriteria [critCount].SearchValue.Signature = theHeader.Platform;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_PLATFORM; %.4s\n", &theHeader.Platform);
        }

        if (control[JAVAsearchFlags] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_PROFILEFLAGS;
            theCriteria [critCount].SearchValue.Value = theHeader.Flags;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_PROFILEFLAGS; %x\n", theHeader.Flags);
        }

        if (control[JAVAsearchManufacturer] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_DEVICEMFG;
            theCriteria [critCount].SearchValue.Signature = theHeader.DeviceManufacturer;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_DEVICEMFG; %.4s\n", &theHeader.DeviceManufacturer);
        }

        if (control[JAVAsearchModel] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_DEVICEMODEL;
            theCriteria [critCount].SearchValue.Signature = theHeader.DeviceModel;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_DEVICEMODEL; %.4s\n", &theHeader.DeviceModel);
        }

        if (control[JAVAsearchAttributes] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_DEVICEATTRIBUTESHI;
            theCriteria [critCount].SearchValue.Value = theHeader.DeviceAttributes.hi;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_DEVICEATTRIBUTESHI; %x\n", theHeader.DeviceAttributes.hi);

            theCriteria [critCount].SearchElement = SPSEARCH_DEVICEATTRIBUTESLO;
            theCriteria [critCount].SearchValue.Value = theHeader.DeviceAttributes.lo;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_DEVICEATTRIBUTESLO; %x\n", theHeader.DeviceAttributes.lo);
        }

        if (control[JAVAsearchRenderingIntent] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_RENDERINGINTENT;
            theCriteria [critCount].SearchValue.Signature = theHeader.RenderingIntent;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_RENDERINGINTENT; %.4s\n", &theHeader.RenderingIntent);
        }

        if (control[JAVAsearchIlluminant] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_ILLUMINANT;
            theCriteria [critCount].SearchValue.XYZ.X = theHeader.Illuminant.X;
            theCriteria [critCount].SearchValue.XYZ.Y = theHeader.Illuminant.Y;
            theCriteria [critCount].SearchValue.XYZ.Z = theHeader.Illuminant.Z;
            critCount++;
            DIAG_PRINTF4("SPSEARCH_ILLUMINANT; %x %x %x\n", theHeader.Illuminant.X, theHeader.Illuminant.Y, theHeader.Illuminant.Z);
        }

        if (control[JAVAsearchCreator] == JAVAsearchTrue) {
            theCriteria [critCount].SearchElement = SPSEARCH_ORIGINATOR;
            theCriteria [critCount].SearchValue.Signature = theHeader.Originator;
            critCount++;
            DIAG_PRINTF2("SPSEARCH_ORIGINATOR; %.4s\n", &theHeader.Originator);
        }
    }

    (*env)->ReleaseByteArrayElements (env, options, control, 0);

    theSearch->critCount = critCount;
    theSearch->criterion = theCriteria;
    theSearch->critSize = SPSEARCH_LIST;

    return theStatus;
}


/* copy a list of profile IDs into a Java list of IDs */
static void pfToID (JNIEnv *env, jlongArray javaIDs, SpProfile_p profileList, KpInt32_t nIDs)
{
KpInt32_t    i1;
storeID_t    theSProfile;
jlong_p        javaIDlist;

    javaIDlist = (*env)->GetLongArrayElements (env, javaIDs, 0);

    DIAG_PRINTF1("match profiles =");
    for (i1 = 0; i1 < nIDs; i1++) {
        theSProfile.pf = profileList [i1];        /* copy profile IDs into java IDs list */
        javaIDlist [i1] = theSProfile.j;
        DIAG_PRINTF2(" %x", theSProfile.j);
    }
    DIAG_PRINTF1("\n");
            
    (*env)->ReleaseLongArrayElements (env, javaIDs, javaIDlist, 0);
}




/* Class:     sun_awt_color_CMM
 * Method:    cmmGetTagData
 * Signature: (JI[B)I
 * static native int cmmGetTagData (long profileID, int tagSignature, byte[] data);
 */
JNIEXPORT jint JNICALL Java_sun_awt_color_CMM_cmmGetTagData
  (JNIEnv *env, jobject obj, jlong profileID, jint tagSignature, jbyteArray data)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;
KpHandle_t    sDataH;
jbyte_p        jDataP;
KpInt8_p    sDataP;
KpInt32_t    i1, sBufSize, jBufSize;
SpTagId_t    TagId;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmGetTagData\n");

    theCallerId = getCallerID (env);        /* not actually needed, but must be there */
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theSProfile.j = profileID;
        DIAG_PRINTF3("theSProfile = %x; tag = %.4s\n", theSProfile.pf, &tagSignature);

        TagId = (SpTagId_t) tagSignature;

        jBufSize = (KpUInt32_t) (*env)->GetArrayLength(env, data);    /* get dest size */
        jDataP = (*env)->GetByteArrayElements (env, data, 0);        /* get the data address */

        if (TagId == SpTagHeader) {
            theStatus = SpRawHeaderGet (theSProfile.pf, jBufSize, jDataP);
        }
        else {
            theStatus = SpRawTagDataGet (theSProfile.pf, TagId, (KpUInt32_p)&sBufSize, &sDataH);
            if (theStatus == SpStatSuccess) {
                if (jBufSize < sBufSize) {
                    theStatus = SpStatBufferTooSmall;
                }
                else {
                    sDataP = (KpInt8_p) lockBuffer (sDataH);

                    for (i1 = 0; i1 < sBufSize; i1++) {
                        jDataP [i1] = sDataP [i1];
                    }

                    unlockBuffer (sDataH);
                }
            }
        }

        (*env)->ReleaseByteArrayElements (env, data, jDataP, 0);
    }

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmGetTagSize
 * Signature: (JI[I)I
 * static native int cmmGetTagSize (long profileID, int tagSignature, int[] size);
 */
JNIEXPORT jint JNICALL Java_sun_awt_color_CMM_cmmGetTagSize
  (JNIEnv *env, jobject obj, jlong profileID, jint tagSignature, jintArray size)
{
SpStatus_t    theStatus = SpStatBadCallerId;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;
KpUInt32_t    TagDataSize = 0;
SpTagId_t    TagId;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmGetTagSize\n");

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theSProfile.j = profileID;
        DIAG_PRINTF3("theSProfile = %x; tag = %.4s\n", theSProfile.pf, &tagSignature);

        TagId = (SpTagId_t) tagSignature;

        if (TagId == SpTagHeader) {
            TagDataSize = HEADER_SIZE;
            theStatus = SpStatSuccess;
        }
        else {
            theStatus = SpRawTagDataGetSize (theSProfile.pf, TagId, &TagDataSize);    /* get size of this tag */
        }
    }

    returnInt (env, size, TagDataSize);

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmSetTagData
 * Signature: (JI[B)I
 * static native int cmmSetTagData (long profileID, int tagSignature, byte[] data);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmSetTagData (JNIEnv *env, jobject obj, jlong profileID, jint tagSignature, jbyteArray data)
{
SpStatus_t    theStatus = SpStatBadCallerId;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;
jbyte_p        dataP;
KpUInt32_t    TagDataSize;
SpTagId_t    TagId;
SpHeader_t    thePublicHeader;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmSetTagData\n");

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theSProfile.j = profileID;
        DIAG_PRINTF3("theSProfile = %x; tag = %.4s\n", theSProfile.pf, &tagSignature);

        TagId = (SpTagId_t) tagSignature;

        dataP = (*env)->GetByteArrayElements (env, data, 0);            /* get the data address */
        TagDataSize = (KpUInt32_t) (*env)->GetArrayLength(env, data);    /* and size */

        if (TagId == SpTagHeader) {
            theStatus = SpHeaderToPublic ((KpInt8_p)dataP, TagDataSize, &thePublicHeader);
            if (theStatus == SpStatSuccess) {
                theStatus = SpProfileSetHeader (theSProfile.pf, &thePublicHeader);
            }
        }
        else {
            theStatus = SpRawTagDataSet (theSProfile.pf, TagId, TagDataSize, (KpInt8_p)dataP);    /* set tag data */
        }
        
        (*env)->ReleaseByteArrayElements (env, data, dataP, 0);            /* don't hold data */
    }

    return checkStatus (theStatus);
}


/*
 * Class:     sun_awt_color_CMM
 * Method:    cmmGetTransform
 * Signature: (Ljava/awt/color/ICC_Profile;IILjava/awt/color/ICC_Transform;)I
 * static native int cmmGetTransform
        (ICC_Profile profile, int renderType, int transformType, ICC_Transform result);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmGetTransform (JNIEnv *env, jobject obj, jobject profile, jint renderType, jint transformType, jobject result)
{
SpStatus_t    theStatus = SpStatSuccess;
SpCallerId_t    theCallerId;
storeID_t    theSProfile;
storeID_t    theXform;
KpInt32_t    whichRender;
jclass        cls;
jfieldID    fid;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmGetTransform\n");
    theXform.xf = NULL;

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        cls = (*env)->GetObjectClass (env, profile);    /* get the profile ID */

        fid = (*env)->GetFieldID (env, cls, "ID", "J");
        if (fid != 0) {
            theSProfile.j = (*env)->GetLongField (env, profile, fid);
        }

        switch (renderType) {
        case JAVA_ICC_TransformAny:
            whichRender = SpTransRenderAny;
            break;
        
        case icPerceptual:
            whichRender = SpTransRenderPerceptual;
            break;
        
        case icRelativeColorimetric:
            whichRender = SpTransRenderColormetric;
            break;

        case icSaturation:
            whichRender = SpTransRenderSaturation;
            break;

        case icAbsoluteColorimetric:
            whichRender = SpTransRenderAbsColormetric;
            break;
            
        default:
            theStatus = SpStatOutOfRange;
            whichRender = -1;    /* avoid warning */
        }

        if (theStatus == SpStatSuccess) {
            theStatus = SpXformGet (theSProfile.pf, whichRender, transformType, &theXform.xf);

            DIAG_PRINTF3("theSProfile = %x; xForm = %x\n", theSProfile.pf, theXform.xf);
        }
    }

    setObjectID (env, result, theXform);        /* set ID in the transform object */

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmFreeTransform
 * Signature: (J)I
 * static native int cmmFreeTransform (long transformID);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmFreeTransform (JNIEnv *env, jobject obj, jlong transformID)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
storeID_t    theXform;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmFreeTransform\n");

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theXform.j = transformID;

        theStatus = SpXformFree (&theXform.xf);
    }

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmGetNumComponents
 * Signature: (J[I)I
 * static native int cmmGetNumComponents (long transformID, int[] nComps);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmGetNumComponents (JNIEnv *env, jobject obj, jlong transformID, jintArray nComps)
{
SpStatus_t    theStatus = SpStatBadCallerId;
SpCallerId_t    theCallerId;
storeID_t    theXform;
jint_p        dataP;
KpInt32_t    nIn = 0, nOut = 0;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmGetNumComponents\n");

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theXform.j = transformID;

        theStatus = SpXformGetChannels (theXform.xf, &nIn, &nOut);    /* get number of components of this transform */

        DIAG_PRINTF4("theXform = %x; nIn = %d; nOut = %d\n", theXform.xf, nIn, nOut);
    }

    dataP = (*env)->GetIntArrayElements (env, nComps, 0);            /* get the data address */

    dataP[0] = nIn;        /* return the number of input components */
    dataP[1] = nOut;    /* and the number of output components */

    (*env)->ReleaseIntArrayElements (env, nComps, dataP, 0);        /* don't hold data */

    return checkStatus (theStatus);
}



/* Class:     sun_awt_color_CMM
 * Method:    cmmCombineTransforms
 * Signature: ([JLjava/awt/color/ICC_Transform;)I
 * static native int cmmCombineTransforms (long[] transforms, ICC_Transform result);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmCombineTransforms (JNIEnv *env, jobject obj, jlongArray transforms, jobject result)
{
SpStatus_t    theStatus;
SpCallerId_t    theCallerId;
storeID_t    theXform;
SpXform_p    xformIDs;
KpInt32_t    nXforms, FailingXform, i1;
storeID_p    dataP;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmCombineTransforms\n");

    theXform.xf = NULL;

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        nXforms = (*env)->GetArrayLength (env, transforms);                        /* number of transforms */

        xformIDs = (SpXform_p) allocBufferPtr (nXforms * sizeof (SpXform_t));
        
        dataP = (storeID_p) (*env)->GetLongArrayElements (env, transforms, 0);    /* get the array address */

        DIAG_PRINTF1("xformIDs =");
        for (i1 = 0; i1 < nXforms; i1++) {
            xformIDs[i1] = dataP[i1].xf;        /* get tranform IDs */
            DIAG_PRINTF2(" %x", xformIDs[i1]);
        }

        (*env)->ReleaseLongArrayElements(env, transforms, (jlong_p) dataP, 0);    /* don't need to hold data */

        theStatus = SpConnectSequence (nXforms, xformIDs, &theXform.xf, &FailingXform, NULL, NULL);

        DIAG_PRINTF2("\ntheXform = %x\n", theXform.xf);

        freeBufferPtr ((KpGenericPtr_t) xformIDs);
    }

    setObjectID (env, result, theXform);            /* set ID in the transform object */

    return checkStatus (theStatus);
}


/* Class:     sun_awt_color_CMM
 * Method:    cmmColorConvert
 * Signature: (JLjava/awt/color/imageLayout;Ljava/awt/color/imageLayout;)I
 * static native int cmmColorConvert (long transformID, imageLayout src, imageLayout dest);
 */
JNIEXPORT jint JNICALL
    Java_sun_awt_color_CMM_cmmColorConvert (JNIEnv *env, jobject obj, jlong transformID, jobject src, jobject dest)
{
SpCallerId_t    theCallerId;
storeID_t    theXform;
SpStatus_t    theStatus;
SpPixelLayout_t    SrcLayout;
SpPixelLayout_t    DestLayout;

    if (obj) {}        /* avoid compiler warnings */

    DIAG_PRINTF1("Java_sun_awt_color_CMM_cmmColorConvert\n");

    theXform.xf = NULL;

    theCallerId = getCallerID (env);
    if (theCallerId == NULL) {
        theStatus = SpStatBadCallerId;
    }
    else {
        theXform.j = transformID;                        /* get the transform ID */
        DIAG_PRINTF2("theXform = %x\n", theXform.xf);

        theStatus = initImageLayout (env, src, &SrcLayout);
        
        if (theStatus == SpStatSuccess) {
            theStatus = initImageLayout (env, dest, &DestLayout);

            if (theStatus == SpStatSuccess) {
                theStatus = SpEvaluate (theXform.xf, &SrcLayout, &DestLayout, NULL, NULL);
            }
        }

        releaseArrayData (env, src, &SrcLayout);
        releaseArrayData (env, dest, &DestLayout);
    }

    return checkStatus (theStatus);
}


    /* get an image layout */
static SpStatus_t initImageLayout (JNIEnv *env, jobject src, SpPixelLayout_t *theLayout)

{
KpInt32_t    theType = -1, i1, dataInc;
jclass       cls;
jfieldID     fid;
jobject      chanData = NULL;
jbyteArray   byteArray;
jbyte_p      byteDataP;
jshortArray  shortArray;
jshort_p     shortDataP;
jintArray    intArray;
jint_p       intDataP;
jintArray    DataOffsets;
jint_p       offsetp;
int          offset;
KpUInt8_p    dataP;
SpStatus_t   theStatus = SpStatSuccess;

    DIAG_PRINTF1("initImageLayout\n");

    cls = (*env)->GetObjectClass (env, src);

    fid = (*env)->GetFieldID (env, cls, "Type", "I");
    if (fid != 0) {
        theType = (KpInt32_t) (*env)->GetIntField (env, src, fid);
        DIAG_PRINTF2("Type = %x\n", theType);
    }

    fid = (*env)->GetFieldID (env, cls, "NumCols", "I");
    if (fid != 0) {
        theLayout->NumCols = (KpInt32_t) (*env)->GetIntField (env, src, fid);
        DIAG_PRINTF2("NumCols = %x\n", theLayout->NumCols);
    }

    fid = (*env)->GetFieldID (env, cls, "NumRows", "I");
    if (fid != 0) {
        theLayout->NumRows = (KpInt32_t) (*env)->GetIntField (env, src, fid);
        DIAG_PRINTF2("NumRows = %x\n", theLayout->NumRows);
    }

    fid = (*env)->GetFieldID (env, cls, "OffsetColumn", "I");
    if (fid != 0) {
        theLayout->OffsetColumn = (KpInt32_t) (*env)->GetIntField (env, src, fid);
        DIAG_PRINTF2("OffsetColumn = %x\n", theLayout->OffsetColumn);
    }

    fid = (*env)->GetFieldID (env, cls, "OffsetRow", "I");
    if (fid != 0) {
        theLayout->OffsetRow = (KpInt32_t) (*env)->GetIntField (env, src, fid);
        DIAG_PRINTF2("OffsetRow = %x\n", theLayout->OffsetRow);
    }

    fid = (*env)->GetFieldID (env, cls, "NumChannels", "I");
    if (fid != 0) {
        theLayout->NumChannels = (KpInt32_t) (*env)->GetIntField (env, src, fid);
        DIAG_PRINTF2("NumChannels = %x\n", theLayout->NumChannels);
    }

    fid = (*env)->GetFieldID (env, cls, "chanData", "[Ljava/lang/Object;");
    if (fid != 0) {
        chanData = (jobject) (*env)->GetObjectField (env, src, fid);
        DIAG_PRINTF2("chanData = %x\n", chanData);
    }

    fid = (*env)->GetFieldID (env, cls, "DataOffsets", "[I");
    if (fid != 0) {
        DataOffsets = (jintArray) (*env)->GetObjectField (env, src, fid);
        DIAG_PRINTF2("DataOffsets = %x\n", DataOffsets);
    }

    switch (theType) {
        case JAVA_TYPE_INT_RGB:
        case JAVA_TYPE_INT_ARGB:
        case JAVA_TYPE_INT_ARGB_PRE:
            theLayout->SampleType = SpSampleType_UByte;

            intArray = (jintArray) (*env)->GetObjectArrayElement (env, chanData, 0);
            DIAG_PRINTF2("intArray = %x", intArray);

            offsetp = (*env)->GetIntArrayElements(env, DataOffsets, 0);
            offset = *offsetp;
            (*env)->ReleaseIntArrayElements(env, DataOffsets,
                                            (jint_p) offsetp, 0);

            intDataP = (*env)->GetIntArrayElements (env, intArray, 0);    /* get the data address */
            DIAG_PRINTF2("; intDataP = %x\n", intDataP);
            intDataP += offset;

            dataP = ((KpUInt8_p) intDataP) + 1;
            dataInc = 1;

    #if defined (KPLSBFIRST)
            dataP = ((KpUInt8_p) intDataP) + 2;
            dataInc = -1;
    #endif

            theLayout->NumChannels = 3;            /* 3 components in a single int */

            DIAG_PRINTF1("theLayout->BaseAddrs =");
            for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
                theLayout->BaseAddrs[i1] = (SpHugeBuffer_t) (dataP + (dataInc * i1));
                DIAG_PRINTF2(" %x", theLayout->BaseAddrs[i1]);
            }
            DIAG_PRINTF1("\n");

            break;

        case JAVA_TYPE_3BYTE_BGR:
        case JAVA_TYPE_4BYTE_ABGR:
        case JAVA_TYPE_4BYTE_ABGR_PRE:
            theLayout->SampleType = SpSampleType_UByte;

            byteArray = (jbyteArray) (*env)->GetObjectArrayElement (env, chanData, 0);
            DIAG_PRINTF2("byteArray = %x", byteArray);

            offsetp = (*env)->GetIntArrayElements(env, DataOffsets, 0);
            offset = *offsetp;
            (*env)->ReleaseIntArrayElements(env, DataOffsets,
                                            (jint_p) offsetp, 0);

            byteDataP = (*env)->GetByteArrayElements (env, byteArray, 0);    /* get the data address */
            DIAG_PRINTF2("; byteDataP = %x\n", byteDataP);
            byteDataP += offset;

            if (theType == JAVA_TYPE_3BYTE_BGR) {
                dataP = ((KpUInt8_p) byteDataP) + 2;  /* no alpha channel */
            }
            else {
                dataP = ((KpUInt8_p) byteDataP) + 3; /* skip over alpha chan */
            }

            theLayout->NumChannels = 3;                /* 3 color components */

            DIAG_PRINTF1("theLayout->BaseAddrs =");
            for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
                theLayout->BaseAddrs[i1] = (SpHugeBuffer_t) (dataP - i1);
                DIAG_PRINTF2(" %x", theLayout->BaseAddrs[i1]);
            }
            DIAG_PRINTF1("\n");

            break;

        case typeCompUByte:
            theLayout->SampleType = SpSampleType_UByte;

            DIAG_PRINTF1("byteArray/theLayout->BaseAddrs =");
            for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
                byteArray = (jbyteArray) (*env)->GetObjectArrayElement (env, chanData, i1);
                DIAG_PRINTF2(" %x/", byteArray);
            
                byteDataP = (*env)->GetByteArrayElements (env, byteArray, 0);    /* get the data address */

                theLayout->BaseAddrs[i1] = (SpHugeBuffer_t) byteDataP;
                DIAG_PRINTF2("%x", theLayout->BaseAddrs[i1]);
            }
            DIAG_PRINTF1("\n");

            break;

        case typeCompUShort12:
        case typeCompUShort:
            if (theType == typeCompUShort12) {
                theLayout->SampleType = SpSampleType_UShort12;
            }
            else {
                theLayout->SampleType = SpSampleType_UShort;
            }

            DIAG_PRINTF1("shortArray/theLayout->BaseAddrs =");
            for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
                shortArray = (jshortArray) (*env)->GetObjectArrayElement (env, chanData, i1);
                DIAG_PRINTF2(" %x/", shortArray);
                shortDataP = (*env)->GetShortArrayElements (env, shortArray, 0);    /* get the data address */

                theLayout->BaseAddrs[i1] = (SpHugeBuffer_t) shortDataP;
                DIAG_PRINTF2(" %x", theLayout->BaseAddrs[i1]);
            }
            DIAG_PRINTF1("\n");

            break;

        case typePelUByte:
            theLayout->SampleType = SpSampleType_UByte;

            byteArray = (jbyteArray) (*env)->GetObjectArrayElement (env, chanData, 0);
            DIAG_PRINTF2("byteArray = %x", byteArray);

            dataP = (KpUInt8_p) (*env)->GetByteArrayElements (env, byteArray, 0);    /* get the data address */
            DIAG_PRINTF2("; byteDataP = %x\n", byteDataP);

            DIAG_PRINTF1("theLayout->BaseAddrs =");
            for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
                theLayout->BaseAddrs[i1] = (SpHugeBuffer_t) (dataP + (sizeof (KpUInt8_t) * i1));
                DIAG_PRINTF2(" %x", theLayout->BaseAddrs[i1]);
            }
            DIAG_PRINTF1("\n");

            break;

        case typePelUShort12:
        case typePelUShort:
            if (theType == typePelUShort12) {
                theLayout->SampleType = SpSampleType_UShort12;
            }
            else {
                theLayout->SampleType = SpSampleType_UShort;
            }

            shortArray = (jshortArray) (*env)->GetObjectArrayElement (env, chanData, 0);
            DIAG_PRINTF2("shortArray = %x", shortArray);

            dataP = (KpUInt8_p) (*env)->GetShortArrayElements (env, shortArray, 0);    /* get the data address */
            DIAG_PRINTF2("; shortDataP = %x\n", shortDataP);

            DIAG_PRINTF1("theLayout->BaseAddrs =");
            for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
                theLayout->BaseAddrs[i1] = (SpHugeBuffer_t) (dataP + (sizeof (KpUInt16_t) * i1));
                DIAG_PRINTF2(" %x", theLayout->BaseAddrs[i1]);
            }
            DIAG_PRINTF1("\n");

            break;

        default:
            theStatus = SpStatFailure;
            theLayout->SampleType = 0;
    }
    
    return theStatus;
}


    /* release data used for image processing */
static void
    releaseArrayData (    JNIEnv            *env,
                jobject            src,
                SpPixelLayout_t*    theLayout)

{
KpInt32_t    i1;
jclass        cls;
jfieldID    fid;
jobject        chanData = NULL;
jbyteArray    byteArray;
jshortArray    shortArray;
jintArray    intArray;
jbyte_p        dataP;
jintArray    DataOffsets;
jint_p       offsetp;
int          offset;
jint_p       intDataP;
KpInt32_t    theType;

    cls = (*env)->GetObjectClass (env, src);

    fid = (*env)->GetFieldID (env, cls, "Type", "I");
    if (fid != 0) {
        theType = (KpInt32_t) (*env)->GetIntField (env, src, fid);
        DIAG_PRINTF2("Type = %x\n", theType);
    }

    fid = (*env)->GetFieldID (env, cls, "chanData", "[Ljava/lang/Object;");
    if (fid != 0) {
        chanData = (jobject) (*env)->GetObjectField (env, src, fid);
        DIAG_PRINTF2("chanData = %x\n", chanData);
    }

    fid = (*env)->GetFieldID (env, cls, "DataOffsets", "[I");
    if (fid != 0) {
        DataOffsets = (jintArray) (*env)->GetObjectField (env, src, fid);
        DIAG_PRINTF2("DataOffsets = %x\n", DataOffsets);
    }

    switch (theType) {
        case JAVA_TYPE_INT_RGB:
        case JAVA_TYPE_INT_ARGB:
        case JAVA_TYPE_INT_ARGB_PRE:
        intArray = (jintArray) (*env)->GetObjectArrayElement (env, chanData, 0);
        DIAG_PRINTF2("intArray = %x\n", intArray);

        dataP = (jbyte_p) theLayout->BaseAddrs[0];

        dataP -= 1;
#if defined (KPLSBFIRST)
        /* dataP -= 2; */
        dataP -= 1;
#endif

        offsetp = (*env)->GetIntArrayElements(env, DataOffsets, 0);
        offset = *offsetp;
        (*env)->ReleaseIntArrayElements(env, DataOffsets,
                                        (jint_p) offsetp, 0);
        intDataP = (jint_p) dataP;
        intDataP -= offset;

        (*env)->ReleaseIntArrayElements (env, intArray, intDataP, 0); /* release the data address */

        break;

        case JAVA_TYPE_3BYTE_BGR:
        case JAVA_TYPE_4BYTE_ABGR:
        case JAVA_TYPE_4BYTE_ABGR_PRE:
        byteArray = (jbyteArray) (*env)->GetObjectArrayElement (env, chanData, 0);
        DIAG_PRINTF2("byteArray = %x\n", byteArray);

        dataP = theLayout->BaseAddrs[0];

        if (theLayout->SampleType == JAVA_TYPE_3BYTE_BGR) {
            dataP -= 2;  /* no alpha channel */
        }
        else {
            dataP -= 3; /* skip over alpha chan */
        }

        offsetp = (*env)->GetIntArrayElements(env, DataOffsets, 0);
        offset = *offsetp;
        (*env)->ReleaseIntArrayElements(env, DataOffsets,
                                        (jint_p) offsetp, 0);
        dataP -= offset;

        (*env)->ReleaseByteArrayElements (env, byteArray, dataP, 0); /* release the data address */

        break;
        case typePelUByte:
        byteArray = (jbyteArray) (*env)->GetObjectArrayElement (env, chanData, 0);
        dataP = theLayout->BaseAddrs[0];

        (*env)->ReleaseByteArrayElements (env, byteArray, dataP, 0); /* release the data address */

        break;

        case typeCompUByte:
        DIAG_PRINTF1("byteArray =");
        for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
            byteArray = (jbyteArray) (*env)->GetObjectArrayElement (env, chanData, i1);
            DIAG_PRINTF2(" %x", byteArray);

            (*env)->ReleaseByteArrayElements (env, byteArray, theLayout->BaseAddrs[i1], 0);    /* release the data address */
        }
        DIAG_PRINTF1("\n");

        break;

        case typeCompUShort12:
        case typeCompUShort:
        DIAG_PRINTF1("shortArray =");
        for (i1 = 0; i1 < theLayout->NumChannels; i1++) {
            shortArray = (jshortArray) (*env)->GetObjectArrayElement (env, chanData, i1);
            DIAG_PRINTF2(" %x", shortArray);

            (*env)->ReleaseShortArrayElements (env, shortArray, theLayout->BaseAddrs[i1], 0);    /* release the data address */
        }
        DIAG_PRINTF1("\n");

        break;

        case typePelUShort12:
        case typePelUShort:
        shortArray = (jshortArray) (*env)->GetObjectArrayElement (env, chanData, 0);
        DIAG_PRINTF2("shortArray = %x\n", shortArray);

        (*env)->ReleaseShortArrayElements (env, shortArray, theLayout->BaseAddrs[0], 0);    /* release the data address */

        break;

    default:
        break;

    }
}


/* get the caller ID */

static SpCallerId_t getCallerID (JNIEnv *env)
{
jclass    cls;

    cls = (*env)->FindClass (env, "sun/awt/color/CMM");

    return getClassID (env, cls);
}


static SpCallerId_t getClassID (JNIEnv *env, jclass theClass)
{
jfieldID    fid;
storeID_t    theCallerId;

    fid = (*env)->GetStaticFieldID (env, theClass, "ID", "J");
    if (fid != 0) {
        theCallerId.j = (*env)->GetStaticLongField (env, theClass, fid);
    }
    else {
        DIAG_PRINTF1("SpStatBadCallerId\n");
        theCallerId.cid = NULL;
    }

    return theCallerId.cid;
}



/* set the ID of a class */
static SpStatus_t setClassID (JNIEnv *env, jclass theClass, SpCallerId_t theID)
{
jfieldID    fid;
SpStatus_t    theStatus;
storeID_t    theJId;

    theJId.cid = theID;
    
    fid = (*env)->GetStaticFieldID (env, theClass, "ID", "J");
    if (fid == 0) {
        theStatus = SpStatFailure;
    }
    else {
        theJId.cid = theID;
    
        (*env)->SetStaticLongField (env, theClass, fid, theJId.j);
        theStatus = SpStatSuccess;
    }
    
    return theStatus;
}



/* set the ID of an  object */
static void setObjectID (JNIEnv *env, jobject theObject, storeID_t theID)
{
jclass        cls;
jfieldID    fid;

    cls = (*env)->GetObjectClass (env, theObject);

    fid = (*env)->GetFieldID (env, cls, "ID", "J");
    if (fid != 0) {
        (*env)->SetLongField (env, theObject, fid, theID.j);
    }
}


/* store an int into the first element of an int array */
static void returnInt (JNIEnv *env, jintArray theReturn, jint theInt)
{ 
jint_p        dataP;

    dataP = (*env)->GetIntArrayElements (env, theReturn, 0);

    dataP[0] = theInt;                    /* return the number of matches in the first element */

    (*env)->ReleaseIntArrayElements (env, theReturn, dataP, 0);
}


/* check for success, print error if not */
static jint checkStatus (SpStatus_t theStatus)
{
    if (theStatus != SpStatSuccess) {
        DIAG_PRINTF2("theStatus = %d\n", theStatus);
    }

    return (jint) theStatus;
}

