/*
 * @(#)cmpsattf.c	1.4 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: cmpsattf.c	2.10 98/02/17
 *
 *	Contains:	FuT specific attribute propagation when composing PTs
 *
 **********************************************************************
 *    COPYRIGHT (c) Eastman Kodak Company, 1991-1998
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 **********************************************************************
 */

#include "kcmsos.h"
#include "kcmptlib.h"
#include "attrib.h"
#include "attrcipg.h"
#include "kcptmgr.h"
#include "kcpfchn.h"

#define ATTR_LIST_END 0

/* attributes which are propagated from the first("Input") PT */
static KpInt32_t propIAttrF[] = {
	KCM_IN_CHAIN_CLASS,
	KCM_IN_CHAIN_CLASS_2,
	ATTR_LIST_END};

/* attributes which are propagated from the second("Output") PT */
static KpInt32_t propOAttrF[] = {
	KCM_OUT_CHAIN_CLASS,
	KCM_OUT_CHAIN_CLASS_2,
	ATTR_LIST_END};


/* use PTRefNum1 and PTRefNum2 to propagate attributes to PTRefNumR */
PTErr_t ComposeAttrFut(	PTRefNum_t PTRefNum1,
						PTRefNum_t PTRefNum2,
						PTRefNum_t PTRefNumR)
{
PTErr_t	errnum = KCP_SUCCESS;
KpInt8_t	strInSpace[KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpInt8_t	strOutSpace[KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpInt32_t	inspace, outspace, attrSize;

	/* get 1st PT */
	attrSize = KCM_MAX_ATTRIB_VALUE_LENGTH;
	errnum = PTGetAttribute(PTRefNum1, KCM_SPACE_OUT, &attrSize, strOutSpace);
	if (errnum == KCP_SUCCESS) {
		outspace = KpAtoi(strOutSpace);

		/* get 2nd PT */
		attrSize = KCM_MAX_ATTRIB_VALUE_LENGTH;
		errnum = PTGetAttribute(PTRefNum2, KCM_SPACE_IN, &attrSize, strInSpace);

		if (errnum == KCP_SUCCESS) {
			inspace = KpAtoi(strInSpace);

			if ((outspace == KCM_UNKNOWN) && (inspace != KCM_UNKNOWN)) {
				errnum = moveAttrList (PTRefNum2, 0, propIAttrF, PTRefNumR);
				if (errnum == KCP_SUCCESS) {
					errnum = moveAttrList (PTRefNum2, 0, propOAttrF, PTRefNumR);
				}
				return (errnum);
			}
			else {
				if ((outspace != KCM_UNKNOWN) && (inspace == KCM_UNKNOWN)) {
					errnum = moveAttrList (PTRefNum1, 0, propIAttrF, PTRefNumR);
					if (errnum == KCP_SUCCESS) {
						errnum = moveAttrList (PTRefNum1, 0, propOAttrF, PTRefNumR);
					}
					return (errnum);
				}
			}
		}
	}

/* propagate "input" attributes */
	errnum = moveAttrList (PTRefNum1, 0, propIAttrF, PTRefNumR);
	if (errnum == KCP_SUCCESS) {

	/* propagate "output" attributes */
		errnum = moveAttrList (PTRefNum2, 0, propOAttrF, PTRefNumR);
	}
	return (errnum);
}

