/*
 * @(#)combine.c	1.7 98/09/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: combine.c	2.76 98/02/25
 *
 *	Contains:	execute the KCMS compose function, which is
 *			called combine
 *
 *********************************************************************
 *    COPYRIGHT (c) 1991-1998 Eastman Kodak Company
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *********************************************************************
 */


#include <string.h>

#include "kcmptlib.h"
#include "kcptmgrd.h"
#include "attrib.h"
#include "attrcipg.h"
#include "kcptmgr.h"
#include "kcpfchn.h"
#include "makefuts.h"

/* prototypes */
static PTErr_t checkFuT (PTRefNum_t PTRefNum);
static fut_ptr_t do_compose (fut_t FAR*, KpInt32_t, fut_t FAR*, KpInt32_t, KpInt32_t);
static fut_ptr_t do_pre_compose (PTRefNum_t, fut_ptr_t,  PTRefNum_t, fut_ptr_t, KpInt32_t, KpInt32_t, KpInt32_t);
static fut_ptr_t do_post_compose (fut_ptr_t fut1, KpInt32_t iomask);


/*----------------------------------------------------------------------
 *  get_fut
 *----------------------------------------------------------------------
 */

static fut_ptr_t
	get_fut (	KpChar_p	fname,
				fut_ptr_t	(*makeAFuT) (KpInt32_t),
				KpInt32_t	size)
{
fut_ptr_t			futp = NULL;
#if defined KCP_STORE_AUX_FILES
initializedGlobals_p	iGP;
KpChar_t			Fullname[256];
KpFileProps_t		KCPDataDirProps;	/* file properties for CP data files */

	iGP = getInitializedGlobals ();
	if (iGP == NULL) {
		return NULL;
	}

	/* make fully qualified file name */
	strcpy (Fullname, iGP->KCPDataDir);
	strcat (Fullname, fname);
	SetKCPDataDirProps (&KCPDataDirProps);

	/* try to load it from the file */
	futp = fut_load_fp (Fullname, KCPDataDirProps);
#else
	if (fname) {}
#endif

	if (futp == NULL) {	/* The creation of the aux pt file can only be performed by one
						process at a time.  To enforce this initialize every open for
						write locks the file.
						*/

		/* create the file data and write it to disk */
		futp = makeAFuT (size);

		#if defined KCP_STORE_AUX_FILES
		if (futp != NULL) {
			KpUInt32_t		retVal;

			/* If write fails, use the fut anyway.  It may be slower
			   since it is recreated each time but it works */
			retVal = fut_store_fp (futp, Fullname, KCPDataDirProps);
		}
		#endif
	}

	return futp;
}


/*----------------------------------------------------------------------
 *  get_linlab
 *----------------------------------------------------------------------
 */

fut_ptr_t
	get_linlab (KpInt32_t size)
{
PTErr_t		errnum;
KpInt32_t	PTCubeSize;					/* size of ICC profiles (pts) */

	errnum = getPTCubeSize (&PTCubeSize);
	if (errnum != KCP_SUCCESS) {
		return NULL;
	}

	if (size == PTCubeSize) {
		return get_fut ("CP23", get_linlab_fut, size);

	} else if (size == SIXTEEN_CUBE) {
		return get_fut ("CP24", get_linlab_fut, size);

	} else {
		return (get_linlab_fut(size));
	}
}

/*----------------------------------------------------------------------
 *  get_lin3d
 *----------------------------------------------------------------------
 */

fut_ptr_t
	get_lin3d (KpInt32_t size)
{
PTErr_t		errnum;
KpInt32_t	PTCubeSize;					/* size of ICC profiles (pts) */

	errnum = getPTCubeSize (&PTCubeSize);
	if (errnum != KCP_SUCCESS) {
		return NULL;
	}

	if (size == PTCubeSize) {
		return get_fut ("CP12", get_lin3d_fut, size);

	} else if (size == SIXTEEN_CUBE) {
		return get_fut ("CP16", get_lin3d_fut, size);

	} else {
		return (get_lin3d_fut(size));
	}
}

/*----------------------------------------------------------------------
 *  get_lin4d
 *----------------------------------------------------------------------
 */

fut_ptr_t
	get_lin4d (KpInt32_t size)
{
PTErr_t		errnum;
KpInt32_t	PTCubeSize;					/* size of ICC profiles (pts) */

	errnum = getPTCubeSize (&PTCubeSize);
	if (errnum != KCP_SUCCESS) {
		return NULL;
	}

	if (size == PTCubeSize) {
		return get_fut ("CP13", get_lin4d_fut, size);

	} else if (size == SIXTEEN_CUBE) {
		return get_fut ("CP17", get_lin4d_fut, size);

	} else {
		return (get_lin4d_fut(size));
	}
}

/*----------------------------------------------------------------------
 *  get_lab2xyz
 *----------------------------------------------------------------------
 */

fut_ptr_t
	get_lab2xyz (KpInt32_t size)
{
PTErr_t		errnum;
KpInt32_t	PTCubeSize;					/* size of ICC profiles (pts) */

	errnum = getPTCubeSize (&PTCubeSize);
	if (errnum != KCP_SUCCESS) {
		return NULL;
	}

	if (size == PTCubeSize) {
		return get_fut ("CP14", get_lab2xyz_fut, size);

	} else if (size == SIXTEEN_CUBE) {
		return get_fut ("CP18", get_lab2xyz_fut, size);

	} else {
		return (NULL);
	}
}

/*----------------------------------------------------------------------
 *  get_xyz2lab
 *----------------------------------------------------------------------
 */

fut_ptr_t
	get_xyz2lab (KpInt32_t size)
{
PTErr_t		errnum;
KpInt32_t	PTCubeSize;					/* size of ICC profiles (pts) */

	errnum = getPTCubeSize (&PTCubeSize);
	if (errnum != KCP_SUCCESS) {
		return NULL;
	}

	if (size == PTCubeSize) {
		return get_fut ("CP20", get_xyz2lab_fut, size);

	} else if (size == SIXTEEN_CUBE) {
		return get_fut ("CP21", get_xyz2lab_fut, size);

	} else {
		return (NULL);
	}
}

/*----------------------------------------------------------------------
 *  kcpGetColorSpace
 *----------------------------------------------------------------------
 */

/* get color space */
KpInt32_t
	kcpGetColorSpace (	PTRefNum_t	PTRefNum,
					KpInt32_t		AttrId)
{
PTErr_t	errnum;
KpInt32_t	attrSize;
KpChar_t	attribute [KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpChar_t	cie_attrib [KCM_MAX_ATTRIB_VALUE_LENGTH+1];

	attrSize = sizeof (attribute) - 1;
	errnum = PTGetAttribute (PTRefNum, AttrId, &attrSize, attribute);
	if (errnum != KCP_SUCCESS)
		return KCM_UNKNOWN;

	KpItoa(KCM_CIE_LAB, cie_attrib);
	if (0 == strcmp (attribute, cie_attrib))
		return KCM_CIE_LAB;

	KpItoa(KCM_CIE_XYZ, cie_attrib);
	if (0 == strcmp (attribute, cie_attrib))
		return KCM_CIE_XYZ;

	return KCM_UNKNOWN;
}


/* do compose, taking into account possibly differant colorspaces */
static fut_ptr_t
	do_compose (fut_ptr_t		fut2,
					KpInt32_t			in_of_fut2,
					fut_ptr_t	fut1,
					KpInt32_t			out_of_fut1,
					KpInt32_t			iomask)
{
fut_ptr_t	fut1new = FUT_NULL;
fut_ptr_t	futmatch = FUT_NULL;
fut_ptr_t	futR = FUT_NULL;

/* create color space matching fut, if needed */
	if ((out_of_fut1 == KCM_CIE_XYZ) && (in_of_fut2 == KCM_CIE_LAB)) {
		futmatch = get_xyz2lab ((KpInt32_t) SIXTEEN_CUBE);
	} else if ((out_of_fut1 == KCM_CIE_LAB) && (in_of_fut2 == KCM_CIE_XYZ)) {
		futmatch = get_lab2xyz ((KpInt32_t) SIXTEEN_CUBE);
	} else {
		return fut_comp (fut2, fut1, iomask);
	}

	if (futmatch != FUT_NULL) {
/* change fut1's color space to match fut2's color space */
		fut1new = fut_comp (futmatch, fut1, iomask);

		fut_free (futmatch);

		if (fut1new != FUT_NULL) {
/* now combine the two futs */
			futR = fut_comp (fut2, fut1new, iomask);
			fut_free (fut1new);
		}
	}
/* return the result */
	return futR;
}

/**************************************************/
/* pre compose to set the grid size appropriately */
/**************************************************/
static fut_ptr_t
	do_pre_compose (PTRefNum_t	PTRefNum2,
						 fut_ptr_t	fut2,
						 PTRefNum_t	PTRefNum1,
						 fut_ptr_t	fut1,
						 KpInt32_t		iomask,
						 KpInt32_t		newLUTDims,
						 KpInt32_t		mode)
{
KpInt32_t	OutSpace, InSpace1, InSpace2;
KpInt32_t	LUTDims1, NumInVars1, NumOutVars1, status;
KpInt32_t	LUTDims2, NumInVars2, NumOutVars2;
fut_ptr_t	fut1new;
fut_ptr_t	fut2new;
fut_ptr_t	futR;

	InSpace1 = kcpGetColorSpace (PTRefNum1, KCM_SPACE_IN);
	OutSpace = kcpGetColorSpace (PTRefNum1, KCM_SPACE_OUT);
	InSpace2 = kcpGetColorSpace (PTRefNum2, KCM_SPACE_IN);

	status = fut_mfutInfo (fut1, &LUTDims1, &NumInVars1, &NumOutVars1);
	if ((mode & PT_COMBINE_LARGEST) != 0) {
		/* pre compose with the largest grid dimension */
		fut_mfutInfo (fut2, &LUTDims2, &NumInVars2, &NumOutVars2);
		if (NumInVars1 == FOUR_CLR) {
			/* if cmyk input, limit the results to the size of the first or... */
			if (LUTDims2 > SIXTEEN_CUBE) {	/* SIXTEEN_CUBE which ever is greater */
				LUTDims2 = SIXTEEN_CUBE;
			}
		}
		if (newLUTDims < LUTDims1) {
			newLUTDims = LUTDims1;
		}
		if (newLUTDims < LUTDims2) {
			newLUTDims = LUTDims2;
		}
	}

	if (((status != 1) || (LUTDims1 != newLUTDims)) && ((NumInVars1 == THREE_CLR) || (NumInVars1 == FOUR_CLR))) {
		if (NumInVars1 == THREE_CLR) {
			if (InSpace1 == KCM_CIE_LAB) {
				fut2new = get_linlab ((KpInt32_t) newLUTDims);
			}
			else {
				fut2new = get_lin3d ((KpInt32_t) newLUTDims);
			}
		}
		else {
			fut2new = get_lin4d ((KpInt32_t) newLUTDims);
		}

		if (NULL == fut2new) {
			return NULL;
		}

		fut1new = fut_comp (fut1, fut2new, iomask);
		fut_free (fut2new);

		if (NULL == fut1new) {
			return (NULL);
		}

		/* now that the fut is the right size, do the actual composition */
		futR = do_compose (fut2, InSpace2, fut1new, OutSpace, iomask);
		fut_free (fut1new);
	} /* if status */

	else {
		futR = do_compose (fut2, InSpace2, fut1, OutSpace, iomask);
	}

	return futR;

} /* do_pre_compose */


PTErr_t
	PTCombine (	KpInt32_t		mode,
				PTRefNum_t	PTRefNum1,
				PTRefNum_t	PTRefNum2,
				PTRefNum_p	PTRefNumR)
{
KpGenericPtr_t	PTHdr;
KpHandle_t		PTHdr1, PTHdr2, PTHdrR, PTAttr, PTData1, PTData2, PTDataR;
fut_ptr_t		fut1, fut2, futR;
PTErr_t			errnum = KCP_FAILURE, errnum1 = KCP_FAILURE, errnum2 = KCP_FAILURE;
KpInt32_t		iomask, compType;
KpInt32_t		OutSpace, InSpace;
PTRefNum_t		labxyzPTRefNum, PTRefNumSerialR;
KpInt32_t		attrSize;
KpChar_t		attribute1 [KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpChar_t		attribute2 [KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpInt32_t		PTCubeSize;	/* size of ICC profiles (pts) */

	errnum = getPTCubeSize (&PTCubeSize);
	if (errnum != KCP_SUCCESS) {
		return errnum;
	}

/* Check for valid PTRefNumP */
	if (Kp_IsBadWritePtr(PTRefNumR, sizeof(*PTRefNumR))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut3;
	}

	*PTRefNumR = 0;
	PTHdr = NULL;
	PTDataR = PTHdrR = NULL;
	fut1 = fut2 = futR = FUT_NULL;
	compType = mode & PT_COMBINE_TYPE;	/* and composition type */

/* must currently be active */
	errnum1 = PTGetPTInfo (PTRefNum1, (PTAddr_t**)&PTHdr1, (PTAddr_t**)&PTAttr, (PTAddr_t**)&PTData1);
	if (errnum1 == KCP_NOT_CHECKED_IN) {
		errnum = errnum1;
		goto ErrOut3;
	}
	if (errnum1 == KCP_PT_ACTIVE) {
		errnum = checkFuT(PTRefNum1);
		if (errnum != KCP_SUCCESS)
			goto ErrOut3;
	}
	if (errnum1 != KCP_PT_INACTIVE) {
		/* must also currently be active */
		errnum2 = PTGetPTInfo (PTRefNum2, (PTAddr_t**)&PTHdr2, (PTAddr_t**)&PTAttr, (PTAddr_t**)&PTData2);
		if (errnum2 == KCP_NOT_CHECKED_IN) {
			errnum = errnum2;
			goto ErrOut3;
		}
		if (errnum2 == KCP_PT_ACTIVE) {
			errnum = checkFuT(PTRefNum2);
			if (errnum != KCP_SUCCESS) {
				goto ErrOut3;
			}
		}

		if ((errnum1 != KCP_PT_INACTIVE) && (errnum2 != KCP_PT_INACTIVE)) {

			if ((errnum1 == KCP_SERIAL_PT) || (errnum2 == KCP_SERIAL_PT)) {		/* must be for serial evaluation */
				attrSize = sizeof (attribute1) - 1;
				errnum =  PTGetAttribute (PTRefNum1, KCM_CLASS, &attrSize, attribute1);
				if (errnum != KCP_SUCCESS) {
					attribute1[0] = 0;
				}

				attrSize = sizeof (attribute2) - 1;
				errnum =  PTGetAttribute (PTRefNum2, KCM_CLASS, &attrSize, attribute2);
				if (errnum != KCP_SUCCESS) {
					attribute2[0] = 0;
				}

				/* Don't add chainning rules to serial pt list! */
				if ((strcmp(attribute1, KCM_AUX_RULE_CLASS_S) != 0) && (strcmp(attribute2, KCM_AUX_RULE_CLASS_S) != 0)) {

					/* create color space matching fut, if needed */
					OutSpace = kcpGetColorSpace (PTRefNum1, KCM_SPACE_OUT);
					InSpace = kcpGetColorSpace (PTRefNum2, KCM_SPACE_IN);

					if ((OutSpace == KCM_CIE_XYZ) && (InSpace == KCM_CIE_LAB)) {
						futR = get_xyz2lab (SIXTEEN_CUBE);
					}
					else {
						if ((OutSpace == KCM_CIE_LAB) && (InSpace == KCM_CIE_XYZ)) {
							futR = get_lab2xyz (SIXTEEN_CUBE);
						}
					}

					if (futR == NULL) {
						errnum = addSerialPT (PTRefNum1, PTRefNum2, PTRefNumR);
					}
					else {
						PTDataR = (KpHandle_t)fut_unlock_fut(futR);	/* unlock the fut */
						errnum = registerPT(NULL, NULL, &labxyzPTRefNum);
						if (errnum != KCP_SUCCESS) {
							goto ErrOut3;
						}

						makeActive (labxyzPTRefNum, PTDataR);
						makeSerial (labxyzPTRefNum);

						errnum = addSerialPT (PTRefNum1, labxyzPTRefNum, &PTRefNumSerialR);
						if (errnum != KCP_SUCCESS) {
							goto ErrOut3;
						}

						errnum = PTCheckOut (labxyzPTRefNum);		/* leave just enough for serial evaluation */
						if (errnum != KCP_SUCCESS) {
							goto ErrOut3;
						}

						errnum = addSerialPT (PTRefNumSerialR, PTRefNum2, PTRefNumR);

						if (errnum == KCP_SUCCESS) {
							errnum = PTCheckOut (PTRefNumSerialR);
						}
					}
				}
				else {
					errnum = registerPT (NULL, NULL, &PTRefNumSerialR);
					if (errnum != KCP_SUCCESS) {
						goto ErrOut3;
					}

					makeSerial (PTRefNumSerialR);

					/* Don't add chaining rules to serial pt list! */
					if (strcmp(attribute1, KCM_AUX_RULE_CLASS_S) != 0) {
						errnum = addSerialPT (PTRefNumSerialR, PTRefNum1, PTRefNumR);
					}
					else {
						errnum = addSerialPT (PTRefNumSerialR, PTRefNum2, PTRefNumR);
					}

					if (errnum == KCP_SUCCESS) {
						errnum = PTCheckOut (PTRefNumSerialR);
					}
				}

				if (errnum != KCP_SUCCESS) {
					goto ErrOut3;
				}
			}

			if ((PTData1 != NULL) && (PTData2 != NULL)) {	/* compose PT data */
				/* define iomask */
				iomask = FUT_PASS(FUT_XYZT);
				if ((mode & PT_COMBINE_CUBIC) != 0) {
					iomask |= FUT_ORDER (FUT_CUBIC);		/* use cubic interpolation for composition evaluations */
				}

				/* get buffer for resultant info header */
				PTHdr = allocBufferPtr (sizeof(fut_hdr_t));

				if (PTHdr == NULL) {
					errnum = KCP_NO_CHECKIN_MEM;
					goto ErrOut3;
				}

				fut1 = fut_lock_fut((KpHandle_t)PTData1);	/* get the fut addresses */
				if (fut1 == FUT_NULL) {
					errnum = KCP_FAILURE;
					goto ErrOut;
				}
				fut2 = fut_lock_fut((KpHandle_t)PTData2);
				if (fut2 == FUT_NULL) {
					errnum = KCP_FAILURE;
					goto ErrOut1;
				}

				switch (compType) {
				case PT_COMBINE_STD:
					futR = fut_comp(fut2, fut1, iomask);
					if (futR == FUT_NULL) {
						errnum = KCP_FAILURE;
						goto ErrOut2;
					}
					break;

#if !defined (KP_NO_IOTBL_COMBINE)
				case PT_COMBINE_ITBL:
					futR = fut_comp_itbl(fut2, fut1, iomask);
					if (futR == FUT_NULL) {
						errnum = KCP_FAILURE;
						goto ErrOut2;
					}
					break;

				case PT_COMBINE_OTBL:
					futR = fut_comp_otbl(fut2, fut1, iomask);
					if (futR == FUT_NULL) {
						errnum = KCP_FAILURE;
						goto ErrOut2;
					}
					break;
#endif

				case PT_COMBINE_PF_8:
do_combine_pf8:
					futR = do_pre_compose (PTRefNum2, fut2, PTRefNum1, fut1, iomask, PTCubeSize, mode);
					if (futR == FUT_NULL) {
						errnum = KCP_FAILURE;
						goto ErrOut2;
					}
					break;

				case PT_COMBINE_PF_16:
					if (PTCubeSize != EIGHT_CUBE) {
						goto do_combine_pf8;
					}
					futR = do_pre_compose (PTRefNum2, fut2, PTRefNum1, fut1, iomask, SIXTEEN_CUBE, mode);
					if (futR == FUT_NULL) {
						errnum = KCP_FAILURE;
						goto ErrOut2;
					}
					break;

				case PT_COMBINE_PF:
					if (PTCubeSize != EIGHT_CUBE) {
						goto do_combine_pf8;
					}
					if ((mode & PT_COMBINE_LARGEST) != 0) {
						futR = do_pre_compose (PTRefNum2, fut2, PTRefNum1, fut1, iomask, TWO_CUBE, mode);
					}
					else {
						OutSpace = kcpGetColorSpace (PTRefNum1, KCM_SPACE_OUT);
						InSpace = kcpGetColorSpace (PTRefNum2, KCM_SPACE_IN);
						futR = do_compose (fut2, InSpace, fut1, OutSpace, iomask);
					}
					if (futR == FUT_NULL) {
						errnum = KCP_FAILURE;
						goto ErrOut2;
					}
					break;

				default:
					errnum = KCP_BAD_COMP_MODE;
					goto ErrOut2;
				}

				if (!fut_io_encode (futR, PTHdr)) {	/* make the info header */
					errnum = KCP_FAILURE;
					goto ErrOut2;
				}

				PTDataR = (KpHandle_t)fut_unlock_fut(fut2);	/* unlock the futs */
				if (PTDataR == NULL) {
					errnum = KCP_FAILURE;
					fut_unlock_fut(futR);
					goto ErrOut1;
				}
				PTDataR = (KpHandle_t)fut_unlock_fut(fut1);
				if (PTDataR == NULL) {
					errnum = KCP_FAILURE;
					fut_unlock_fut(futR);
					goto ErrOut;
				}
				if ( (futR != FUT_NULL) && (futR->idstr != NULL) ) {
					fut_free_idstr (futR->idstr);
					futR->idstr = NULL;
				}
				PTDataR = (KpHandle_t)fut_unlock_fut(futR);
				if (PTDataR == NULL) {
					errnum = KCP_FAILURE;
					goto ErrOut;
				}

				PTHdrR = unlockBufferPtr(PTHdr);	/* unlock the header buffer */
				if (PTHdrR == NULL) {
					errnum = KCP_FAILURE;
					goto ErrOut3;
				}
				PTHdr = NULL;						/* indicate that this is unlocked */

			/* checkin and activate */
				if (*PTRefNumR == 0) {
					errnum = PTNewPT(PTHdrR, PTDataR, PTRefNumR);
					if (errnum != KCP_SUCCESS) {
						goto ErrOut;
					}
				}
				else {
					makeActive (*PTRefNumR, PTDataR);
					makeSerial (*PTRefNumR);
					errnum = TpGenerateAttr (*PTRefNumR); /* generate constant attributes */
					if (errnum != KCP_SUCCESS) {
							(void) PTCheckOut(*PTRefNumR);	/* clean up if error */
					}
					errnum = setPTHdr (*PTRefNumR, PTHdrR);
					if (errnum != KCP_SUCCESS) {
						goto ErrOut;
					}
				}
			}	/* end of NOT serial evaluation */

			/* propagate attributes */
			errnum = ComposeAttr (PTRefNum1, PTRefNum2, compType, *PTRefNumR);
			if (errnum != KCP_SUCCESS) {
				goto ErrOut;
			}

			/* propagate FuT attributes */
			errnum = ComposeAttrFut (PTRefNum1, PTRefNum2, *PTRefNumR);
			if (errnum != KCP_SUCCESS) {
				goto ErrOut;
			}
		}
	}

ErrOut3:
	return (errnum);

ErrOut2:
	fut_unlock_fut(fut2);

ErrOut1:
	fut_unlock_fut(fut1);

ErrOut:
	if (PTHdr != NULL) {
		freeBufferPtr(PTHdr);
	}
	else if (PTHdrR != NULL) {
		freeBuffer(PTHdrR);
	}
	if (futR == FUT_NULL) {
		fut_free(futR);
	}
	goto ErrOut3;
}


/* see if a PT is a FuT */
static PTErr_t
	checkFuT (	PTRefNum_t	PTRefNum)
{
PTErr_t		errnum;
KpInt32_t	attrSize;
KpChar_t	attribute[KCM_MAX_ATTRIB_VALUE_LENGTH+1];

	attrSize = KCM_MAX_ATTRIB_VALUE_LENGTH;
	errnum = PTGetAttribute(PTRefNum, KCM_TECH_TYPE, &attrSize, attribute);
	if (errnum == KCP_SUCCESS) {
		if ((strcmp(attribute, KCM_FUT_S) != 0) && (strcmp(attribute, KCM_SERIAL_FUT_S) != 0)) {
			errnum = KCP_NOT_FUT;			/* PT is not FuT technology */
		}
	}

	return (errnum);
}


/* get size of ICC profiles (pts) */
PTErr_t
	getPTCubeSize (	KpInt32_p	size)
{
initializedGlobals_p	iGP;

	iGP = getInitializedGlobals ();
	if (iGP == NULL) {
		return KCP_NO_PROCESS_GLOBAL_MEM;
	}

	*size = iGP->PTCubeSize;

	return KCP_SUCCESS;
}


