/*
 * @(#)evalth13.c	1.4 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: evalth13.c	1.6 98/02/25
 *
 *
 *	tetrahedral evaluation functions
 *
 *	Author:			George Pawle
 *
 *	Create Date:	12/20/96
 *
 *	Note: This implementation supports number of inputs equal to 3 only.
 *	It supports number of outputs equal to 1 through 8.
 *	Please refer to function evalGrid3() for an example.
 *
 ****************************************************************
 *    COPYRIGHT (c) 1996-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */

#include "kcpcache.h"

/* Macros used by 3-input functions */

#define TH1_STORE_DATA8(chan) \
	*outp##chan## = prevRes##chan##;	/* write to buffer */ \
	outp##chan## = (KpUInt8_p)outp##chan## + outStride##chan##;	/* next location */

#define TH1_STORE_DATA16(chan) \
	*outp##chan## = prevRes##chan##;	/* write to buffer */ \
	outp##chan## = (KpUInt16_p)((KpUInt8_p)outp##chan## + outStride##chan##);	/* next location */

#define TH1_EVAL_INIT \
esItbl_p	inLut0 = PTTableP->esILuts.P; \
KpUInt8_p	baseP, gridBase = (KpUInt8_p)PTTableP->esGLuts.P; \
KpUInt32_t	i, data0, data1, data2; \
KpInt32_t	tResult, tvert1, tvert2, tvert3; \
esInterp_t	Xf, Yf, Zf; \
KpInt32_t	inStride0 = inStride[0], inStride1 = inStride[1], inStride2 = inStride[2];

#if defined (TH1_FP_GRID_DATA)
#define TH1_3D_TMP_DATA esGridData_t tvert0Data, tvert1Data, tvert2Data;
#else
#define TH1_3D_TMP_DATA KpInt32_t mullacc, tvertData;
#endif

#define TH1_3D_VARIABLES esInterp_t	xf, yf, zf; \
KpUInt32_t	a001, a010, a011, a100, a101, a110;

#define TH1_3D_GRID_OFFSETS_INIT \
	a001 = PTTableP->esGOffsets[1];	/* copy grid offsets into locals */ \
	a010 = PTTableP->esGOffsets[2]; \
	a011 = PTTableP->esGOffsets[3]; \
	a100 = PTTableP->esGOffsets[4]; \
	a101 = PTTableP->esGOffsets[5]; \
	a110 = PTTableP->esGOffsets[6]; \
	tvert3 = PTTableP->esGOffsets[7];

#define TH1_3D_INIT \
TH1_EVAL_INIT \
TH1_3D_TMP_DATA \
TH1_3D_VARIABLES \
TH1_3D_GRID_OFFSETS_INIT

#define TH1_3D_INITD8 \
KpUInt8_p	inp0 = inp[0].p8, inp1 = inp[1].p8, inp2 = inp[2].p8; \
KpUInt8_p	outLut0 = (KpUInt8_p)PTTableP->esOLuts.P; \
KpUInt32_t	thisColor, prevColor = 0xffffffff; \
 \
	TH1_3D_INIT

#define TH1_3D_INITD16 \
KpUInt16_p	inp0 = inp[0].p16, inp1 = inp[1].p16, inp2 = inp[2].p16; \
esItbl_p	inLut1, inLut2; \
KpUInt16_p	outLut0 = (KpUInt16_p)PTTableP->esOLuts.P; \
KpUInt32_t	ColorPart1, prevPart1 = 0, prevPart2 = 0xffffffff, dataMask;  \
 \
	TH1_3D_INIT \
 \
	inLut1 = inLut0 + PTTableP->esILutEntries; \
	inLut2 = inLut1 + PTTableP->esILutEntries; \
 \
	if (PTTableP->esDataTypeI == KCM_USHORT_12) {	/* set up data mask to prevent input table memory access violations */ \
		dataMask = 0xfff; \
	} \
	else { \
		dataMask = 0xffff; \
	}


#define TH1_3D_GETDATAD8 \
	data0 = *inp0; 								/* get channel 0 input data */ \
	inp0 += inStride0; \
	data1 = *inp1; 								/* get channel 1 input data */ \
	inp1 += inStride1; \
	data2 = *inp2; 								/* get channel 2 input data */ \
	inp2 += inStride2; \
 \
	thisColor = (data0 << 16) | (data1 << 8) | (data2);	/* calc this color   */


#define TH1_3D_GETDATAD16 \
	data0 = *inp0; 								/* get channel 0 input data */ \
	data0 &= dataMask; \
	inp0 = (KpUInt16_p)((KpUInt8_p)inp0 + inStride0); \
	data1 = *inp1; 								/* get channel 1 input data */ \
	data1 &= dataMask; \
	inp1 = (KpUInt16_p)((KpUInt8_p)inp1 + inStride1); \
	data2 = *inp2; 								/* get channel 2 input data */ \
	data2 &= dataMask; \
	inp2 = (KpUInt16_p)((KpUInt8_p)inp2 + inStride2); \
\
	ColorPart1 = (data0 << 16) | (data1);	/* calc this color */


#define TH1_3D_FINDTETRAD8    \
\
	prevColor = thisColor;    \
\
	baseP = gridBase; \
	baseP += inLut0[(0*FUT_INPTBL_ENT) + data0].index; 	/* pass input data through input tables */ \
	Xf = inLut0[(0*FUT_INPTBL_ENT) + data0].frac; \
	baseP += inLut0[(1*FUT_INPTBL_ENT) + data1].index; \
	Yf = inLut0[(1*FUT_INPTBL_ENT) + data1].frac; \
	baseP += inLut0[(2*FUT_INPTBL_ENT) + data2].index; \
	Zf = inLut0[(2*FUT_INPTBL_ENT) + data2].frac; \
\
	TH1_3D_CALCTETRA


#define TH1_3D_FINDTETRAD16 \
\
	prevPart1 = ColorPart1;  \
	prevPart2 = data2;  \
\
	baseP = gridBase; \
	baseP += inLut0[data0].index;		/* pass input data through input tables */ \
	Xf = inLut0[data0].frac; \
	baseP += inLut1[data1].index; \
	Yf = inLut1[data1].frac; \
	baseP += inLut2[data2].index; \
	Zf = inLut2[data2].frac; \
\
	TH1_3D_CALCTETRA


#define	TH1_3D_CALCTETRA   \
	/* find the tetrahedron in which the point is located */ \
	if (Xf > Yf) { \
		if (Yf > Zf) {	/* AHEG */ \
			xf = Xf;		/* unchanged order */ \
			yf = Yf; \
			zf = Zf; \
\
			tvert1 = a110; \
			tvert2 = a100; \
		} \
		else { \
			zf = Yf;		/* y into z */ \
			tvert1 = a101; \
			if (Xf > Zf) {	/* AHEF */ \
				xf = Xf;		/* x does not change */ \
				yf = Zf;		/* z into y */ \
\
				tvert2 = a100; \
			} \
			else {			/* AHBF */ \
				xf = Zf;		/* z into x */ \
				yf = Xf;		/* x into y */ \
\
				tvert2 = a001; \
			} \
		} \
	} \
	else { \
		if (Yf > Zf) { \
			xf = Yf;		/* y into x */ \
			tvert2 = a010; \
			if (Xf > Zf) {	/* AHCG */ \
				yf = Xf;		/* x into y */ \
				zf = Zf;		/* z into z */ \
\
				tvert1 = a110; \
			} \
			else {			/* AHCD */ \
				yf = Zf;		/* z into y */ \
				zf = Xf;		/* x into z */ \
\
				tvert1 = a011; \
			} \
		} \
		else {			/* AHDB */ \
			xf = Zf;		/* z into x */ \
			yf = Yf;		/* y into y */ \
			zf = Xf;		/* x into z */ \
\
			tvert1 = a011; \
			tvert2 = a001; \
		} \
	}

#if defined (TH1_FP_GRID_DATA)

#define TH1_3D_TETRAINTERP \
	tvert2Data = *(esGridData_p)(baseP + tvert1); \
	tvert1Data = zf * (*(esGridData_p)(baseP + tvert3) - tvert2Data);	/* (H - tvert1) * z */ \
 \
	tvert0Data = *(esGridData_p)(baseP + tvert2); \
	tvert1Data += (yf * (tvert2Data - tvert0Data));					/* (tvert1 - tvert2) * y */ \
 \
	tvert2Data = *(esGridData_p)(baseP + 0); \
	tvert1Data += (xf * (tvert0Data - tvert2Data));					/* (tvert2 - A) * x */ \
 \
	tResult = tvert2Data + tvert1Data + 0.49999;					/* A + (mults) */

#else
#if defined (KPALPHA)
#define COMP_TRESULT  \
		if ((mullacc + (1<<15)) < 0)  \
			tResult -= (((-mullacc + (1 << 15) -1) ) >> 16) ; \
		else  tResult += ((mullacc + (1 << 15)) >> 16);
#else
#define COMP_TRESULT tResult += ((mullacc + (1 << 15)) >> 16);					/* A + (mults) */

#endif

#define TH1_3D_TETRAINTERP \
	tResult = (KpInt32_t)*(esGridData_p)(baseP + tvert1); \
	mullacc = zf * ((KpInt32_t)*(esGridData_p)(baseP + tvert3) - tResult);	/* (tvert3 - tvert1) * z */ \
 \
	tvertData = (KpInt32_t)*(esGridData_p)(baseP + tvert2); \
	mullacc += (yf * (tResult - tvertData));					/* (tvert1 - tvert2) * y */ \
 \
	tResult = (KpInt32_t)*(esGridData_p)(baseP + 0); \
	mullacc += (xf * (tvertData - tResult));					/* (tvert2 - A) * x */ \
 \
	COMP_TRESULT			/* A + (mults) */

#endif

#define TH1_3D_TETRAINTERP_AND_OLUT(chan) \
	baseP += sizeof (esGridData_t); \
	TH1_3D_TETRAINTERP		/* tetrahedral interpolation for this channel */ \
	prevRes##chan## = outLut0[(##chan##*4096)+tResult];



/******************************************************************************/

void
	evalTh1i3o1d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8;
KpInt32_t	outStride0 = outStride[0];
KpUInt8_t	prevRes0 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];
		}

		TH1_STORE_DATA8(0);		/* store the data */
	}
}


/******************************************************************************/

void
	evalTh1i3o1d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt16_p	outp0 = outp[0].p16;
KpInt32_t	outStride0 = outStride[0];
KpUInt16_t	prevRes0 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];
		}

		TH1_STORE_DATA16(0);		/* store the data */
	}
}


/******************************************************************************/

void
	evalTh1i3o2d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
	}
}


/******************************************************************************/

void
	evalTh1i3o2d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)

{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
	}
}


/******************************************************************************/

void
	evalTh1i3o3d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
	}
}


/******************************************************************************/

void
	evalTh1iB24oB24 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8;
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;


	TH1_3D_INITD8

	if (outStride) {}

	for (i = n; i > 0; i--) {
		data0 = *inp0++; 					/* get channel 0 input data */
		data1 = *inp0++; 					/* get channel 1 input data */
		data2 = *inp0++; 					/* get channel 2 input data */

		thisColor = (data0 << 16) | (data1 << 8) | (data2);	/* calc this color */

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
		}

		*outp0++ = prevRes0;	/* use result from previous color evaluation */
		*outp0++ = prevRes1;
		*outp0++ = prevRes2;
	}
}


/******************************************************************************/

void
	evalTh1iL24oL24 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp2 = outp[2].p8;
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;

	TH1_3D_INITD8

	if (outStride) {}

	for (i = n; i > 0; i--) {
		data2 = inp2[0]; 					/* get channel 2 input data */
		data1 = inp2[1]; 					/* get channel 1 input data */
		data0 = inp2[2]; 					/* get channel 0 input data */

		inp2 += 3;

		thisColor = (data0 << 16) | (data1 << 8) | (data2);	/* calc this color */

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
		}

		outp2[0] = (KpUInt8_t)(prevRes2);		/* use result from previous color evaluation */
		outp2[1] = (KpUInt8_t)(prevRes1);
		outp2[2] = (KpUInt8_t)(prevRes0);

		outp2 += 3;
	}
}


#if defined (KPMAC)
/**************************** Quick Draw Formats **********************************/

void
	evalTh1iQDoQD (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt32_p	inpL, outpL;
KpUInt32_t	alphaData, prevResult = 0;

	TH1_3D_INITD8

	if (inStride == outStride) {}

	inpL = (KpUInt32_p)(inp[0].p8 -1);
	outpL = (KpUInt32_p)(outp[0].p8 -1);

	for (i = n; i > 0; i--) {
		thisColor = *inpL++;
		alphaData = thisColor & 0xff000000;		/* get alpha channel input data */
		data0 = (thisColor >> 16) & 0xff;		/* get channel 0 input data */
		data1 = (thisColor >> 8) & 0xff;		/* get channel 1 input data */
		data2 = (thisColor) & 0xff;				/* get channel 2 input data */

		thisColor &= 0xffffff;					/* clear off alpha channel */

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP					/* channel 0 */

			prevResult = (KpUInt32_t)outLut0[(0*4096)+tResult];

			baseP += sizeof (esGridData_t);
			TH1_3D_TETRAINTERP					/* channel 1 */

			prevResult <<= 8;
			prevResult |= (KpUInt32_t)outLut0[(1*4096)+tResult];

			baseP += sizeof (esGridData_t);
			TH1_3D_TETRAINTERP					/* channel 2 */

			prevResult <<= 8;
			prevResult |= (KpUInt32_t)outLut0[(2*4096)+tResult];
		}

		*outpL++ = prevResult | alphaData;
	}
}


/******************************************************************************/

void
	evalTh1iQDo3 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt32_p	inpL;
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8;
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2];

	TH1_3D_INITD8

	if (inStride) {}

	inpL = (KpUInt32_p)(inp[0].p8 -1);

	for (i = n; i > 0; i--) {
		thisColor = *inpL++;
		data0 = (thisColor >> 16) & 0xff;		/* get channel 0 input data */
		data1 = (thisColor >> 8) & 0xff;		/* get channel 1 input data */
		data2 = (thisColor) & 0xff;				/* get channel 2 input data */

		thisColor &= 0xffffff;					/* clear off alpha channel */

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
	}
}


/******************************************************************************/

void
	evalTh1i3oQD (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt32_p	outpL;
KpUInt32_t	prevResult = 0;

	TH1_3D_INITD8

	if (outStride) {}

	outpL = (KpUInt32_p)(outp[0].p8 -1);

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP					/* channel 0 */

			prevResult = (KpUInt32_t)outLut0[(0*4096)+tResult];
			prevResult |= 0xff00;				/* insert white into alpha channel */

			baseP += sizeof (esGridData_t);
			TH1_3D_TETRAINTERP					/* channel 1 */

			prevResult <<= 8;
			prevResult |= (KpUInt32_t)outLut0[(1*4096)+tResult];

			baseP += sizeof (esGridData_t);
			TH1_3D_TETRAINTERP					/* channel 2 */

			prevResult <<= 8;
			prevResult |= (KpUInt32_t)outLut0[(2*4096)+tResult];
		}

		*outpL++ = prevResult;
	}
}

#endif 	/* KPMAC */


/*********************************************************************************************/

void
	evalTh1i3o3d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)

{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
	}
}


/******************************************************************************/

void
	evalTh1i3o4d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8, outp3 = outp[3].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
		TH1_STORE_DATA8(3);
	}
}


/******************************************************************************/

void
	evalTh1i3o4d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)

{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16, outp3 = outp[3].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
		TH1_STORE_DATA16(3);
	}
}


/******************************************************************************/

void
	evalTh1i3o5d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8, outp3 = outp[3].p8;
KpUInt8_p	outp4 = outp[4].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
		TH1_STORE_DATA8(3);
		TH1_STORE_DATA8(4);
	}
}


/******************************************************************************/

void
	evalTh1i3o5d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16, outp3 = outp[3].p16;
KpUInt16_p	outp4 = outp[4].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
		TH1_STORE_DATA16(3);
		TH1_STORE_DATA16(4);
	}
}


/******************************************************************************/

void
	evalTh1i3o6d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8, outp3 = outp[3].p8;
KpUInt8_p	outp4 = outp[4].p8, outp5 = outp[5].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4], outStride5 = outStride[5];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0, prevRes5 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
			TH1_3D_TETRAINTERP_AND_OLUT(5)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
		TH1_STORE_DATA8(3);
		TH1_STORE_DATA8(4);
		TH1_STORE_DATA8(5);
	}
}


/******************************************************************************/

void
	evalTh1i3o6d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)

{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16, outp3 = outp[3].p16;
KpUInt16_p	outp4 = outp[4].p16, outp5 = outp[5].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4], outStride5 = outStride[5];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0, prevRes5 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
			TH1_3D_TETRAINTERP_AND_OLUT(5)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
		TH1_STORE_DATA16(3);
		TH1_STORE_DATA16(4);
		TH1_STORE_DATA16(5);
	}
}

/******************************************************************************/

void
	evalTh1i3o7d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8, outp3 = outp[3].p8;
KpUInt8_p	outp4 = outp[4].p8, outp5 = outp[5].p8, outp6 = outp[6].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4], outStride5 = outStride[5], outStride6 = outStride[6];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0, prevRes5 = 0, prevRes6 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
			TH1_3D_TETRAINTERP_AND_OLUT(5)
			TH1_3D_TETRAINTERP_AND_OLUT(6)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
		TH1_STORE_DATA8(3);
		TH1_STORE_DATA8(4);
		TH1_STORE_DATA8(5);
		TH1_STORE_DATA8(6);
	}
}


/******************************************************************************/

void
	evalTh1i3o7d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)

{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16, outp3 = outp[3].p16;
KpUInt16_p	outp4 = outp[4].p16, outp5 = outp[5].p16, outp6 = outp[6].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4], outStride5 = outStride[5], outStride6 = outStride[6];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0, prevRes5 = 0, prevRes6 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
			TH1_3D_TETRAINTERP_AND_OLUT(5)
			TH1_3D_TETRAINTERP_AND_OLUT(6)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
		TH1_STORE_DATA16(3);
		TH1_STORE_DATA16(4);
		TH1_STORE_DATA16(5);
		TH1_STORE_DATA16(6);
	}
}


/******************************************************************************/

void
	evalTh1i3o8d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8, outp3 = outp[3].p8;
KpUInt8_p	outp4 = outp[4].p8, outp5 = outp[5].p8, outp6 = outp[6].p8, outp7 = outp[7].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4], outStride5 = outStride[5], outStride6 = outStride[6], outStride7 = outStride[7];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0, prevRes5 = 0, prevRes6 = 0, prevRes7 = 0;

	TH1_3D_INITD8

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD8

		if (thisColor != prevColor) {

			TH1_3D_FINDTETRAD8

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
			TH1_3D_TETRAINTERP_AND_OLUT(5)
			TH1_3D_TETRAINTERP_AND_OLUT(6)
			TH1_3D_TETRAINTERP_AND_OLUT(7)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
		TH1_STORE_DATA8(3);
		TH1_STORE_DATA8(4);
		TH1_STORE_DATA8(5);
		TH1_STORE_DATA8(6);
		TH1_STORE_DATA8(7);
	}
}


/******************************************************************************/

void
	evalTh1i3o8d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)

{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16, outp3 = outp[3].p16;
KpUInt16_p	outp4 = outp[4].p16, outp5 = outp[5].p16, outp6 = outp[6].p16, outp7 = outp[7].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpInt32_t	outStride4 = outStride[4], outStride5 = outStride[5], outStride6 = outStride[6], outStride7 = outStride[7];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0, prevRes4 = 0, prevRes5 = 0, prevRes6 = 0, prevRes7 = 0;

	TH1_3D_INITD16

	for (i = n; i > 0; i--) {
		TH1_3D_GETDATAD16

		if ((ColorPart1 != prevPart1) || (data2 != prevPart2)) {

			TH1_3D_FINDTETRAD16

			TH1_3D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_3D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_3D_TETRAINTERP_AND_OLUT(2)
			TH1_3D_TETRAINTERP_AND_OLUT(3)
			TH1_3D_TETRAINTERP_AND_OLUT(4)
			TH1_3D_TETRAINTERP_AND_OLUT(5)
			TH1_3D_TETRAINTERP_AND_OLUT(6)
			TH1_3D_TETRAINTERP_AND_OLUT(7)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
		TH1_STORE_DATA16(3);
		TH1_STORE_DATA16(4);
		TH1_STORE_DATA16(5);
		TH1_STORE_DATA16(6);
		TH1_STORE_DATA16(7);
	}
}
