/*
 * @(#)evalth14.c	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: evalth14.c	1.12 98/02/25
 *
 *	4 input evaluation functions using tetrahedral interpolation
 *
 *
 *	Author:			George Pawle
 *
 *	Create Date:	12/22/96
 *
 *	Note: This implementation supports number of inputs equal to 4 only.
 *	It supports number of outputs equal to 1, 2, 3, 4.
 *
 ****************************************************************
 *    COPYRIGHT (c) 1996-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */

#include "kcpcache.h"

/* Macros used by 4-input functions */

#define TH1_STORE_DATA8(chan) \
	*outp##chan## = prevRes##chan##;	/* write to buffer */ \
	outp##chan## = (KpUInt8_p)outp##chan## + outStride##chan##;	/* next location */

#define TH1_STORE_DATA16(chan) \
	*outp##chan## = prevRes##chan##;	/* write to buffer */ \
	outp##chan## = (KpUInt16_p)((KpUInt8_p)outp##chan## + outStride##chan##);	/* next location */

#define TH1_EVAL_INIT \
esItbl_p	inLut0 = PTTableP->esILuts.P; \
KpUInt8_p	baseP, gridBase = (KpUInt8_p)PTTableP->esGLuts.P; \
KpUInt32_t	i, data0, data1, data2; \
KpInt32_t	tResult, tvert1, tvert2, tvert3; \
esInterp_t	Xf, Yf, Zf; \
KpInt32_t	inStride0 = inStride[0], inStride1 = inStride[1], inStride2 = inStride[2];

#if defined (TH1_FP_GRID_DATA)
#define TH1_4D_TMP_DATA esGridData_t tvert0Data, tvert1Data, tvert2Data, tvert3Data, tvert4Data;
#else
#define TH1_4D_TMP_DATA KpInt32_t	mullacc, tvertData;
#endif

#define TH1_4D_INIT \
TH1_EVAL_INIT \
TH1_4D_TMP_DATA \
KpUInt32_t	data3; \
KpInt32_t	tvert4; \
esInterp_t	interp[4], Tf; \
KpInt32_t	inStride3 = inStride[3]; \
KpUInt32_t	index; \
th1_4dControl_p	pentaInfo, finderP; \
 \
	finderP = &PTTableP->esFinder[-1];	/* -1 due to Matlab indexing starting at 1 */

#define TH1_4D_INITD8 \
KpUInt8_p	inp0 = inp[0].p8, inp1 = inp[1].p8, inp2 = inp[2].p8, inp3 = inp[3].p8; \
KpUInt8_p	outLut0 = (KpUInt8_p)PTTableP->esOLuts.P; \
KpUInt32_t	thisColor, prevColor; \
 \
	TH1_4D_INIT \
	prevColor = (~(*inp0)) << 24; 			/* make sure cache is not valid */

#define TH1_4D_INITD16 \
KpUInt16_p	inp0 = inp[0].p16, inp1 = inp[1].p16, inp2 = inp[2].p16, inp3 = inp[3].p16; \
esItbl_p	inLut1, inLut2, inLut3; \
KpUInt16_p	outLut0 = (KpUInt16_p)PTTableP->esOLuts.P; \
KpUInt32_t	thisColor1, thisColor2, prevColor1, prevColor2 = 0, dataMask; \
 \
	TH1_4D_INIT \
 \
	inLut1 = inLut0 + PTTableP->esILutEntries; \
	inLut2 = inLut1 + PTTableP->esILutEntries; \
	inLut3 = inLut2 + PTTableP->esILutEntries; \
 \
	if (PTTableP->esDataTypeI == KCM_USHORT_12) {	/* set up data mask to prevent input table memory access violations */ \
		dataMask = 0xfff; \
	} \
	else { \
		dataMask = 0xffff; \
	} \
 \
	prevColor1 = (~(*inp0)) << 16; 			/* make sure cache is not valid */

#define TH1_STORE_DATA8(chan) \
	*outp##chan## = prevRes##chan##;	/* write to buffer */ \
	outp##chan## = (KpUInt8_p)outp##chan## + outStride##chan##;	/* next location */

#define TH1_STORE_DATA16(chan) \
	*outp##chan## = prevRes##chan##;	/* write to buffer */ \
	outp##chan## = (KpUInt16_p)((KpUInt8_p)outp##chan## + outStride##chan##);	/* next location */

#define TH1_4D_GETDATA8 \
	data0 = *inp0; 					/* get channel 0 input data */ \
	inp0 += inStride0; \
	data1 = *inp1; 					/* get channel 1 input data */ \
	inp1 += inStride1; \
	data2 = *inp2; 					/* get channel 2 input data */ \
	inp2 += inStride2; \
	data3 = *inp3; 					/* get channel 2 input data */ \
	inp3 += inStride3; \
\
	thisColor = (data0 << 24) | (data1 << 16) | (data2 << 8) | (data3);	/* calc this color */

#define TH1_4D_GETDATA16 \
	data0 = *inp0; 					/* get channel 0 input data */ \
	data0 &= dataMask; \
	inp0 = (KpUInt16_p)((KpUInt8_p)inp0 + inStride0); \
	data1 = *inp1; 					/* get channel 1 input data */ \
	data1 &= dataMask; \
	inp1 = (KpUInt16_p)((KpUInt8_p)inp1 + inStride1); \
	data2 = *inp2; 					/* get channel 2 input data */ \
	data2 &= dataMask; \
	inp2 = (KpUInt16_p)((KpUInt8_p)inp2 + inStride2); \
	data3 = *inp3; 					/* get channel 2 input data */ \
	data3 &= dataMask; \
	inp3 = (KpUInt16_p)((KpUInt8_p)inp3 + inStride3); \
\
	thisColor1 = (data0 << 16) | (data1);	/* calc this color */  \
    thisColor2 = (data2 << 16) | (data3);

#define TH1_4D_FINDTETRAD8 \
	prevColor = thisColor;  \
\
	baseP = gridBase; \
	baseP += inLut0[(0*FUT_INPTBL_ENT) + data0].index; 	/* pass input data through input tables */ \
	Xf = inLut0[(0*FUT_INPTBL_ENT) + data0].frac; \
	baseP += inLut0[(1*FUT_INPTBL_ENT) + data1].index; \
	Yf = inLut0[(1*FUT_INPTBL_ENT) + data1].frac; \
	baseP += inLut0[(2*FUT_INPTBL_ENT) + data2].index; \
	Zf = inLut0[(2*FUT_INPTBL_ENT) + data2].frac; \
	baseP += inLut0[(3*FUT_INPTBL_ENT) + data3].index; \
	Tf = inLut0[(3*FUT_INPTBL_ENT) + data3].frac; \
\
	TH1_4D_CALCTETRA

#define TH1_4D_FINDTETRAD16 \
	prevColor1 = thisColor1;  \
	prevColor2 = thisColor2;  \
\
	baseP = gridBase; \
	baseP += inLut0[data0].index;		/* pass input data through input tables */ \
	Xf = inLut0[data0].frac; \
	baseP += inLut1[data1].index; \
	Yf = inLut1[data1].frac; \
	baseP += inLut2[data2].index; \
	Zf = inLut2[data2].frac; \
	baseP += inLut3[data3].index; \
	Tf = inLut3[data3].frac; \
\
	TH1_4D_CALCTETRA

#define TH1_4D_CALCTETRA \
	/* find the pentahedron in which the point is located */ \
	/* this builds a binary number based of the possible comparisons of the 4 interpolants */ \
	index = 0; \
	if (Xf > Yf) { \
		index += (1<<5); \
	} \
	if (Zf > Tf) { \
		index += (1<<4); \
	} \
	if (Xf > Zf) { \
		index += (1<<3); \
	} \
	if (Yf > Tf) { \
		index += (1<<2); \
	} \
	if (Yf > Zf) { \
		index += (1<<1); \
	} \
	if (Xf > Tf) { \
		index += (1<<0); \
	} \
\
	pentaInfo = finderP + pentahedron[index];	/* get pentahedron info */ \
\
	tvert1 = pentaInfo->tvert1;	/* offsets from the base grid point to the pentahedral corners */ \
	tvert2 = pentaInfo->tvert2; \
	tvert3 = pentaInfo->tvert3; \
	tvert4 = pentaInfo->tvert4; \
\
	interp[pentaInfo->dx] = Xf;	/* re-dorder the interpolants */ \
	interp[pentaInfo->dy] = Yf; \
	interp[pentaInfo->dz] = Zf; \
	interp[pentaInfo->dt] = Tf; \
	Tf = interp[0]; \
	Zf = interp[1]; \
	Yf = interp[2]; \
	Xf = interp[3];

#if defined (TH1_FP_GRID_DATA)

#define TH1_4D_TETRAINTERP \
	tvert0Data = *(esGridData_p)(baseP + 0); \
	tvert1Data = *(esGridData_p)(baseP + tvert1); \
	tvert2Data = *(esGridData_p)(baseP + tvert2); \
	tvert3Data = *(esGridData_p)(baseP + tvert3); \
	tvert4Data = *(esGridData_p)(baseP + tvert4); \
\
	tResult = 0.49999 + tvert0Data \
			+ ((tvert1Data - tvert0Data) * Xf) \
			+ ((tvert2Data - tvert1Data) * Yf) \
			+ ((tvert3Data - tvert2Data) * Zf) \
			+ ((tvert4Data - tvert3Data) * Tf);

#else
#if defined (KPALPHA)
#define COMP_TRESULT  \
		if ((mullacc + (1<<15)) < 0)  \
			tResult -= (((-mullacc + (1 << 15) -1) ) >> 16) ; \
		else  tResult += ((mullacc + (1 << 15)) >> 16);
#else
#define COMP_TRESULT tResult += ((mullacc + (1 << 15)) >> 16);					/* A + (mults) */

#endif
#define TH1_4D_TETRAINTERP \
	tvertData = (KpInt32_t)*(esGridData_p)(baseP + tvert3); \
	mullacc = Tf * ((KpInt32_t)*(esGridData_p)(baseP + tvert4) - tvertData);	/* (tvert4 - tvert3) * t */ \
 \
	tResult = (KpInt32_t)*(esGridData_p)(baseP + tvert2); \
	mullacc += (Zf * (tvertData - tResult));					/* (tvert3 - tvert2) * z */ \
 \
	tvertData = (KpInt32_t)*(esGridData_p)(baseP + tvert1); \
	mullacc += (Yf * (tResult - tvertData));					/* (tvert2 - tvert1) * y */ \
 \
	tResult = (KpInt32_t)*(esGridData_p)(baseP + 0); \
	mullacc += (Xf * (tvertData - tResult));					/* (tvert1 - tvert0) * x */ \
 \
	COMP_TRESULT					/* tvert0 + (mullacc) */
#endif

#define TH1_4D_TETRAINTERP_AND_OLUT(chan) \
	baseP += sizeof (esGridData_t); \
	TH1_4D_TETRAINTERP		/* tetrahedral interpolation for this channel */ \
	prevRes##chan## = outLut0[(##chan##*4096)+tResult];


/* constant tables for finding pentahedral volumes and the corresponding coefficients and multipliers */

#define TH1_4D_COMBINATIONS 64

static KpUInt32_t	pentahedron[TH1_4D_COMBINATIONS] = { 24,  0, 23,  0,  0,  0, 12,  0,  0,  0, 22,  0,
															  0,  0,  9,  8, 18,  0,  0,  0, 17, 16, 11, 10,
															  0,  0,  0,  0,  0,  0,  0,  7, 21,  0,  0,  0,
															  0,  0,  0,  0, 20,  6, 19,  3,  0,  0,  0,  2,
															 15, 14,  0,  0,  0, 13,  0,  0,  0,  5,  0,  0,
															  0,  4,  0,  1};

/**************************************************************
 * evalTh1i4o1 ---- 8 BIT
 *  Evaluation routine for evaluating 4 channel to 1 channel.
 */

void
	evalTh1i4o1d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8;
KpInt32_t	outStride0 = outStride[0];
KpUInt8_t	prevRes0 = 0;

	TH1_4D_INITD8

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA8

		if (thisColor != prevColor) {

			TH1_4D_FINDTETRAD8

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];
		}

		TH1_STORE_DATA8(0);		/* store the data */
	}
}

/**************************************************************
 * evalTh1i4o1 ---- 16 bit
 *  Evaluation routine for evaluating 4 channel to 1 channel.
 */

void
	evalTh1i4o1d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt16_p	outp0 = outp[0].p16;
KpInt32_t	outStride0 = outStride[0];
KpUInt16_t	prevRes0 = 0;

	TH1_4D_INITD16

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA16

		if ((thisColor1 != prevColor1) || (thisColor2 != prevColor2)) {

			TH1_4D_FINDTETRAD16

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];
		}

		TH1_STORE_DATA16(0);		/* store the data */
	}
}

/**************************************************************
 * evalTh1i4o2 --- 8 bit
 **************************************************************/

void
	evalTh1i4o2d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0;

	TH1_4D_INITD8

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA8

		if (thisColor != prevColor) {

			TH1_4D_FINDTETRAD8

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
	}
}

/**************************************************************
 * evalTh1i4o2  ----  16 bit
 **************************************************************/

void
	evalTh1i4o2d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0;

	TH1_4D_INITD16

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA16

		if ((thisColor1 != prevColor1) || (thisColor2 != prevColor2)) {

			TH1_4D_FINDTETRAD16

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
	}
}

/**************************************************************
 * evalTh1i4o3   -----   8 bit
 **************************************************************/

void
	evalTh1i4o3d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;

	TH1_4D_INITD8

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA8

		if (thisColor != prevColor) {

			TH1_4D_FINDTETRAD8

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_4D_TETRAINTERP_AND_OLUT(2)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
	}
}


#if defined (KPMAC)

typedef union QDBuff_s {
	KpUInt8_t	cbuf[4];
	KpUInt32_t	lword;
} QDBuff_t;

/**************************************************************
 * evalTh1i4o3QD
 **************************************************************/
void
	evalTh1i4o3QD (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt32_p	outpL;
QDBuff_t	QDoBuf;
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;

	TH1_4D_INITD8

	if (outStride) {}

	outpL = (KpUInt32_p)(outp[0].p8 -1);

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA8

		if (thisColor != prevColor) {

			TH1_4D_FINDTETRAD8

			TH1_4D_TETRAINTERP				/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_4D_TETRAINTERP_AND_OLUT(2)
		}

		QDoBuf.lword = *outpL;				/* preserve alpha channel */

		QDoBuf.cbuf[1] = prevRes0;			/* use result from previous color evaluation */
		QDoBuf.cbuf[2] = prevRes1;
		QDoBuf.cbuf[3] = prevRes2;

		*outpL++ = QDoBuf.lword;
	}
}

#endif 	/* if defined KPMAC */


/**************************************************************
 * evalTh1i4o3   ---- 16 bit
 **************************************************************/

void
	evalTh1i4o3d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0;

	TH1_4D_INITD16

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA16

		if ((thisColor1 != prevColor1) || (thisColor2 != prevColor2)) {

			TH1_4D_FINDTETRAD16

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_4D_TETRAINTERP_AND_OLUT(2)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
	}
}


/**************************************************************
 * evalTh1i4o4   ----  8 bit
 *  Evaluation routine for evaluating 4 channel to 4 channels.
 **************************************************************/
void
	evalTh1i4o4d8 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8, outp1 = outp[1].p8, outp2 = outp[2].p8, outp3 = outp[3].p8;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0;

	TH1_4D_INITD8

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA8

		if (thisColor != prevColor) {

			TH1_4D_FINDTETRAD8

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_4D_TETRAINTERP_AND_OLUT(2)
			TH1_4D_TETRAINTERP_AND_OLUT(3)
		}

		TH1_STORE_DATA8(0);		/* store the data */
		TH1_STORE_DATA8(1);
		TH1_STORE_DATA8(2);
		TH1_STORE_DATA8(3);
	}
}


/**************************************************************
 * evalTh1i4oB32
 **************************************************************/

void
	evalTh1iB32oB32 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp0 = outp[0].p8;
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0;

	TH1_4D_INITD8

	if (inStride || outStride) {}

	thisColor = ~inp0[3];		/* make sure cache is not valid */

	for (i = n; i > 0; i--) {
		data0 = *inp0++; 					/* get channel 0 input data */
		data1 = *inp0++; 					/* get channel 1 input data */
		data2 = *inp0++; 					/* get channel 2 input data */
		data3 = *inp0++; 					/* get channel 3 input data */

		thisColor = (data0 << 24) | (data1 << 16) | (data2 << 8) | (data3);	/* calc this color */

		if (thisColor != prevColor) {

			TH1_4D_FINDTETRAD8

			TH1_4D_TETRAINTERP				/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_4D_TETRAINTERP_AND_OLUT(2)
			TH1_4D_TETRAINTERP_AND_OLUT(3)
		}

		*outp0++ = prevRes0;				/* use result from previous color evaluation */
		*outp0++ = prevRes1;
		*outp0++ = prevRes2;
		*outp0++ = prevRes3;
	}
}

/**************************************************************
 * evalTh1i4oL32
 *  Evaluation routine for evaluating 4 channel to 4 channels,
 **************************************************************/
void
	evalTh1iL32oL32 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt8_p	outp3 = outp[3].p8;
KpUInt8_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0;

	TH1_4D_INITD8

	if (inStride || outStride) {}

	thisColor = ~inp0[0];		/* make sure cache is not valid */

	for (i = n; i > 0; i--) {
		data3 = inp3[0]; 					/* get channel 3 input data */
		data2 = inp3[1]; 					/* get channel 2 input data */
		data1 = inp3[2]; 					/* get channel 1 input data */
		data0 = inp3[3]; 					/* get channel 0 input data */

		inp3 += 4;

		thisColor = (data0 << 24) | (data1 << 16) | (data2 << 8) | (data3);	/* calc this color */

		if (thisColor != prevColor) {

			TH1_4D_FINDTETRAD8

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_4D_TETRAINTERP_AND_OLUT(2)
			TH1_4D_TETRAINTERP_AND_OLUT(3)
		}

		outp3[0] = prevRes3;		/* use result from previous color evaluation */
		outp3[1] = prevRes2;
		outp3[2] = prevRes1;
		outp3[3] = prevRes0;

		outp3 += 4;
	}

}

/**************************************************************
 * evalTh1i4o4   ---- 16 bit
 *  Evaluation routine for evaluating 4 channel to 4 channels.
 **************************************************************/
void
	evalTh1i4o4d16 (imagePtr_p	inp,
					KpInt32_p	inStride,
					imagePtr_p	outp,
					KpInt32_p	outStride,
					KpInt32_t	n,
					PTTable_p	PTTableP)
{
KpUInt16_p	outp0 = outp[0].p16, outp1 = outp[1].p16, outp2 = outp[2].p16, outp3 = outp[3].p16;
KpInt32_t	outStride0 = outStride[0], outStride1 = outStride[1], outStride2 = outStride[2], outStride3 = outStride[3];
KpUInt16_t	prevRes0 = 0, prevRes1 = 0, prevRes2 = 0, prevRes3 = 0;

	TH1_4D_INITD16

	for (i = n; i > 0; i--) {
		TH1_4D_GETDATA16

		if ((thisColor1 != prevColor1) || (thisColor2 != prevColor2)) {

			TH1_4D_FINDTETRAD16

			TH1_4D_TETRAINTERP		/* tetrahedral interpolation for channel 0 */
			prevRes0 = outLut0[(0*4096)+tResult];

			TH1_4D_TETRAINTERP_AND_OLUT(1)	/* and the remaining channels */
			TH1_4D_TETRAINTERP_AND_OLUT(2)
			TH1_4D_TETRAINTERP_AND_OLUT(3)
		}

		TH1_STORE_DATA16(0);		/* store the data */
		TH1_STORE_DATA16(1);
		TH1_STORE_DATA16(2);
		TH1_STORE_DATA16(3);
	}
}
