/*
 * @(#)evalth1c.c	1.6 98/09/15
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: evalth1c.c	1.21 98/02/25
 *
 *
 *	tetrahedral interpolation with optimized tables
 *
 *	Author:				 G. Pawle
 *
 ****************************************************************
 *    COPYRIGHT (c) 1996-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */

#include "fut_util.h"
#include "attrib.h"
#include "kcpcache.h"

static KpUInt32_t th1CheckFut (KpHandle_t, KpInt32_t, KpInt32_p, KpInt32_p);
static void th1Itbl2InLut (fut_itbldat_ptr_t, KpInt32_t, KpInt32_t, esItbl_p);
static void th1Itbl2In12tbl (fut_itbl_ptr_t, fut_itbldat_ptr_t);
static void th1Itbl2In16tbl (fut_itbl_ptr_t, fut_itbldat_ptr_t);
static void th1Otbl2Out8Lut (fut_otbldat_ptr_t, KpUInt8_p);
static void th1Otbl2Out12Lut (fut_otbldat_ptr_t, KpUInt16_p);
static void th1Otbl2Out16Lut (fut_otbldat_ptr_t, KpUInt16_p);
static void lockEvalState (PTTable_p);
static void freeEsMem (esMem_p);
static void nullEsMem (esMem_p);
static void lockEsMem (esMem_p);
static void unlockEsMem (esMem_p);
static void allocEsMem (esMem_p, KpInt32_t);
static evalTh1Proc_t getTh1EvalFuncOpt (KpInt32_t, KpInt32_t, KpUInt32_t, KpUInt32_t, KpUInt32_t);
static PTErr_t kcpInitTh1Luts (KpInt32_t, KpInt32_t, fut_gtbl_ptr_t, fut_itbl_ptr_t FAR*,
								fut_gtbldat_ptr_t FAR*, fut_otbldat_ptr_t FAR*, PTTable_p);
static void expandITblMEP (fut_itbldat_ptr_t, KpUInt32_t, fut_itbldat_ptr_t, KpUInt32_t);
#if defined KCP_SINGLE_EVAL_CACHE
static PTErr_t	getEvalStatePT (PTRefNum_p);
static PTErr_t	putEvalStatePT (PTRefNum_t);
#endif

/* constant tables for finding pentahedral volumes and
 * the corresponding coefficients and multipliers
 */
static KpUInt32_t	bceiBase[TH1_4D_PENTAHEDROA] =
	{ 9,  9,  9,  9,  9,  9,  5,  5,  5,  5, 5, 5,  3,  3,  3,  3, 3, 3,  2,  2,  2,  2, 2, 2};
static KpUInt32_t	dfgjkmBase[TH1_4D_PENTAHEDROA] =
	{13, 13, 10, 11, 11, 10, 13, 13,  6,  7, 7, 6, 11, 11,  4,  7, 7, 4, 10, 10,  4,  6, 6, 4};
static KpUInt32_t	hlnoBase[TH1_4D_PENTAHEDROA] =
	{15, 14, 14, 15, 12, 12, 15, 14, 14, 15, 8, 8, 15, 12, 12, 15, 8, 8, 14, 12, 12, 14, 8, 8};
static KpUInt32_t	dxBase[TH1_4D_PENTAHEDROA] =
	{ 3,  3,  3,  3,  3,  3,  2,  2,  1,  1, 0, 0,  2,  2,  1,  1, 0, 0,  2,  2,  1,  1, 0, 0};
static KpUInt32_t	dyBase[TH1_4D_PENTAHEDROA] =
	{ 2,  2,  1,  1,  0,  0,  3,  3,  3,  3, 3, 3,  1,  0,  0,  2, 2, 1,  1,  0,  0,  2, 2, 1};
static KpUInt32_t	dzBase[TH1_4D_PENTAHEDROA] =
	{ 1,  0,  0,  2,  2,  1,  1,  0,  0,  2, 2, 1,  3,  3,  3,  3, 3, 3,  0,  1,  2,  0, 1, 2};
static KpUInt32_t	dtBase[TH1_4D_PENTAHEDROA] =
	{ 0,  1,  2,  0,  1,  2,  0,  1,  2,  0, 1, 2,  0,  1,  2,  0, 1, 2,  3,  3,  3,  3, 3, 3};


/********************* getTh1EvalFuncOpt ***************************/
/* See if this evaluation has been optimized.
   Returns function address if it has.
*/


evalTh1Proc_t
	getTh1EvalFuncOpt (	KpInt32_t	numInputs,
						KpInt32_t	numOutputs,
						KpUInt32_t	ifmt,
						KpUInt32_t	ofmt,
						KpUInt32_t	evalDataType)
{
evalTh1Proc_t	func;

/*	return NULL; */	/* return null to force non-optimized evaluation */

	switch (evalDataType) {
	case KCM_UBYTE:
		switch (numInputs) {
		case 3:
			switch (numOutputs) {
				case 1:
					func = evalTh1i3o1d8;
					break;

				case 2:
					func = evalTh1i3o2d8;
					break;

				case 3:
					func = evalTh1i3o3d8;

		#if defined (KPMAC)
					if (ifmt == FMT_QD) {
						if (ofmt == FMT_QD) {
							func = evalTh1iQDoQD;
						}
						else {
							func = evalTh1iQDo3;
						}
					}
					else {
						if (ofmt == FMT_QD) {
							func = evalTh1i3oQD;
						}
						else {
		#endif
							if ((ifmt == FMT_BIGENDIAN24) && (ofmt == FMT_BIGENDIAN24)) {
								func = evalTh1iB24oB24;
							}
							else {
								if ((ifmt == FMT_LITTLEENDIAN24) && (ofmt == FMT_LITTLEENDIAN24)) {
									func = evalTh1iL24oL24;
								}
							}
		#if defined (KPMAC)
						}
					}
		#endif
					break;

				case 4:
					func = evalTh1i3o4d8;
					break;

				case 5:
					func = evalTh1i3o5d8;
					break;

				case 6:
					func = evalTh1i3o6d8;
					break;

				case 7:
					func = evalTh1i3o7d8;
					break;

				case 8:
					func = evalTh1i3o8d8;
					break;

				default:
					func = NULL;
					break;
			}

			break;

		case 4:
			switch (numOutputs) {
				case 1:
					func = evalTh1i4o1d8;
					break;

				case 2:
					func = evalTh1i4o2d8;
					break;

				case 3:
					func = evalTh1i4o3d8;

		#if defined (KPMAC)
					if (ofmt == FMT_QD) {
						func = evalTh1i4o3QD;
					}
		#endif
					break;

				case 4:
					func = evalTh1i4o4d8;

					if ((ifmt == FMT_BIGENDIAN32) && (ofmt == FMT_BIGENDIAN32)) {
						func = evalTh1iB32oB32;
					}
					else {
						if ((ifmt == FMT_LITTLEENDIAN32) && (ofmt == FMT_LITTLEENDIAN32)) {
							func = evalTh1iL32oL32;
						}
					}

					break;

				default:
					func = NULL;
					break;
			}
			break;

			default:
				func = NULL;
				break;
		}
		break;

	case KCM_USHORT_12:
	case KCM_USHORT:
		switch (numInputs) {
		case 3:
			switch (numOutputs) {
				case 1:
					func = evalTh1i3o1d16;
					break;

				case 2:
					func = evalTh1i3o2d16;
					break;

				case 3:
					func = evalTh1i3o3d16;
					break;

				case 4:
					func = evalTh1i3o4d16;
					break;

				case 5:
					func = evalTh1i3o5d16;
					break;

				case 6:
					func = evalTh1i3o6d16;
					break;

				case 7:
					func = evalTh1i3o7d16;
					break;

				case 8:
					func = evalTh1i3o8d16;
					break;

				default:
					func = NULL;
					break;
			}
			break;

		case 4:
			switch (numOutputs) {
				case 1:
					func = evalTh1i4o1d16;
					break;

				case 2:
					func = evalTh1i4o2d16;
					break;

				case 3:
					func = evalTh1i4o3d16;
					break;

				case 4:
					func = evalTh1i4o4d16;
					break;

				default:
					func = NULL;
					break;
			}
			break;

		default:
			func = NULL;
			break;
		}
		break;

	default:
		func = NULL;
		break;
	}

	return (func);
}


/******************************************************************************/
/* Can this fut have its tables optimized?
   Returns 1 if it can, also returns # of inputs and outputs.
   Check for the following:
	1) Trilinear or default interpolation order
	2) Shared itbls
*/

static KpUInt32_t
	th1CheckFut (	KpHandle_t	futH,
					KpInt32_t	iomask,
					KpInt32_p	numInputsP,
					KpInt32_p	numOutputsP)
{
KpInt32_t 	imask, omask, i, nIn = 0, o, nOut;
fut_chan_ptr_t	futChan;

	imask = (KpInt32_t)FUT_IMASK(iomask);
	omask = (KpInt32_t)FUT_OMASK(iomask);

	/* check the interpolation order */
	if ((FUT_ORDMASK(iomask) != FUT_LINEAR) && (FUT_ORDMASK(iomask) != FUT_DEFAULT)) {
		goto NotThis;		/* Wrong interpolation order */
	}

	/* find number of output channels */
	for (nOut = 0, o = 0; o < FUT_NOCHAN; o++) {
		if (omask & FUT_BIT(o)) {
			for (nIn = 0, i = 0; i < FUT_NICHAN; i++) {	/* make sure input tables are shared */
				if (imask & FUT_BIT(i)) {

					futChan = FCHANP(FFUTP(futH)->chanHandle[o]);	/* get pointer to chan */

					if (futChan->itblHandle[i] != FFUTP(futH)->itblHandle[i]) {
						goto NotThis;		/* Input tables not shared */
					}
					nIn++;
				}
			}
			nOut++;
		}
	}

	*numInputsP = nIn;
	*numOutputsP = nOut;

	return 1;		/* can optimize tables */


NotThis:
	return 0;		/* can not optimize tables */
}


/******************************************************************************/
/* Is the current evaluation state valid? */
PTErr_t
	checkEvalState (	PTTable_p		PTTableP,
						KpInt32_t		nEvals,
						KpUInt32_t		ifmt,
						KpUInt32_t		ofmt,
						evalControl_p	evalControlP)
{
PTErr_t		PTErr;
KpInt32_t	numInputs, numOutputs, totalEvals;
KpUInt32_t	standardFut;
evalTh1Proc_t	func;

#if defined KCP_SINGLE_EVAL_CACHE
PTRefNum_t	evalStatePT;

	PTErr = getEvalStatePT (&evalStatePT);
	if (PTErr != KCP_SUCCESS) {
		return	PTErr;
	}
#endif

	PTErr = KCP_FAILURE;						/* assume eval state is not valid */
	evalControlP->optimizedEval = 0;			/* and that optimized tables can not be used */

	if (evalControlP->nFuts == 1) {
		standardFut = th1CheckFut (PTTableP->data, PTTableP->evalIOMask, &numInputs, &numOutputs);

		if (standardFut == 1) {

			/* get the function to use */
			func = getTh1EvalFuncOpt (numInputs, numOutputs, ifmt, ofmt, evalControlP->evalDataTypeI);

			if (func != NULL) {
				evalControlP->optimizedEval = 1;		/* can use optimized method */
				evalControlP->evalFunc = func;

				/* Do the current tables match this evaluation? */
				if ((PTTableP->esIOMask == PTTableP->evalIOMask)
				 && (PTTableP->esDataTypeI == evalControlP->evalDataTypeI)
				 && (PTTableP->esDataTypeO == evalControlP->evalDataTypeO)
				 && (PTTableP->esILuts.H != NULL)) {

					lockEvalState (PTTableP);	/* lock it down */

					PTErr = KCP_SUCCESS;		/* eval state is valid */
				}
				else {
					totalEvals = numOutputs * nEvals;

					if ((totalEvals < TH1_MIN_EVALS) &&				/* verify sufficient # of evaluations */
						(ifmt != FMT_QD) && (ofmt != FMT_QD)) {		/* still need to use iQDoQD for QD */
						evalControlP->optimizedEval = 0;			/* do not use optimized method */
					}
				}
			}
		}
	}

	if (evalControlP->optimizedEval == 0) {
		evalControlP->evalFunc = evalTh1gen;
	}

	return PTErr;
}


/******************************************************************************/
/* initialize the evaluation state */
PTErr_t
	initEvalState (	PTTable_p		PTTableP,
					evalControl_p	evalControlP)
{
KpInt32_t			theSizef, i, dimx, dimy, dimz, numInputs, nOutputChans, interleaveFactor;
KpUInt32_t			imask, omask, index;
fut_ptr_t			fut = PTTableP->futP;
fut_itbl_ptr_t		iTblsP [TH1_MAX_INPUT_VARS];
fut_chan_ptr_t		aChan;
fut_gtbl_ptr_t		aGTbl = NULL;
fut_otbldat_ptr_t	oTblsP [TH1_MAX_OUTPUT_CHANS];
fut_gtbldat_ptr_t	gTblsP [TH1_MAX_OUTPUT_CHANS];
PTErr_t				PTErr = KCP_SUCCESS;
#if defined KCP_SINGLE_EVAL_CACHE
PTRefNum_t		evalStatePT;
#endif

	imask = (KpInt32_t)FUT_IMASK(PTTableP->evalIOMask);
	omask = (KpInt32_t)FUT_OMASK(PTTableP->evalIOMask);

	/* Find number of input variables */
	for (i = 0, numInputs = 0; i < TH1_MAX_INPUT_VARS ; i++) {
		if (imask & FUT_BIT(i)) {
			iTblsP[numInputs] = fut->itbl[i];			/* make a list of input tables */
			numInputs++;
		}
	}

	/* Find number of output channels */
	for (i = 0, nOutputChans = 0; i < TH1_MAX_OUTPUT_CHANS; i++) {
		if (omask & FUT_BIT(i)) {
			aChan = fut->chan[i];						/* this output cahnnel is being evaluated */

			gTblsP[nOutputChans] = aChan->gtbl->tbl;	/* make a list of gtbls */
			oTblsP[nOutputChans] = aChan->otbl->tbl;	/* and otbls */

			aGTbl = aChan->gtbl;						/* shared input tables, so any gtbl will do */

			nOutputChans++;
		}
	}

	interleaveFactor = 1;		/* assume using fut gtbls */

	if (evalControlP->optimizedEval != 0) {	/* optimized evaluation, set up input, grid, and output tables */

	#if defined KCP_SINGLE_EVAL_CACHE
		PTErr = getEvalStatePT (&evalStatePT);
		if (PTErr != KCP_SUCCESS) {
			return	PTErr;
		}

		if (evalStatePT != PTTableP->refNum) {
			freeEvalState (evalStatePT);		/* retain a single set of optimized tables */
		}
	#endif

		PTTableP->esIOMask = PTTableP->evalIOMask;
		PTTableP->esDataTypeI = evalControlP->evalDataTypeI;
		PTTableP->esDataTypeO = evalControlP->evalDataTypeO;

		PTErr = kcpInitTh1Luts (numInputs, nOutputChans, aGTbl, iTblsP, gTblsP, oTblsP, PTTableP);

		if ((PTErr == KCP_NO_MEMORY) || (PTMemTest () == 0)) {	/* switch to general eval function */
			evalControlP->optimizedEval = 0;	/* not optimized */
			evalControlP->evalFunc = evalTh1gen;
			PTErr = KCP_SUCCESS;
		}
		else {									/* Update evaluation state info */
			interleaveFactor = nOutputChans;	/* this many output channels are interleaved */

		#if defined KCP_SINGLE_EVAL_CACHE
			PTErr = putEvalStatePT (PTTableP->refNum);	/* remember which PT has the optimized tables */
			if (PTErr != KCP_SUCCESS) {
				return	PTErr;
			}
		#endif
		}
	}

	/* set up offsets in grid */
	switch (numInputs) {
	case 1:
		dimx = 0;	/* not used */
		dimy = 0;
		dimz = 0;
		break;

	case 2:
		dimx = 0;	/* not used */
		dimy = 0;
		dimz = aGTbl->size[1];
		break;

	case 3:
		dimx = 0;	/* not used */
		dimy = aGTbl->size[1];
		dimz = aGTbl->size[2];
		break;

	case 4:
		dimx = aGTbl->size[1];
		dimy = aGTbl->size[2];
		dimz = aGTbl->size[3];
		break;

	default:
		return (KCP_NOT_COMPLETE);
	}

	PTTableP->esGOffsets[0] = 0;														/* offset to 0000 */
	PTTableP->esGOffsets[1] = 1;														/* offset to 0001 */
	PTTableP->esGOffsets[2] = dimz;													/* offset to 0010 */
	PTTableP->esGOffsets[3] = PTTableP->esGOffsets[2] + 1;							/* offset to 0011 */
	PTTableP->esGOffsets[4] = dimy * PTTableP->esGOffsets[2];							/* offset to 0100 */
	PTTableP->esGOffsets[5] = PTTableP->esGOffsets[4] + 1;							/* offset to 0101 */
	PTTableP->esGOffsets[6] = PTTableP->esGOffsets[4] + PTTableP->esGOffsets[2];		/* offset to 0110 */
	PTTableP->esGOffsets[7] = PTTableP->esGOffsets[6] + 1;							/* offset to 0111 */
	PTTableP->esGOffsets[8] = dimx * PTTableP->esGOffsets[4];							/* offset to 1000 */
	PTTableP->esGOffsets[9] = PTTableP->esGOffsets[8] + 1;							/* offset to 1001 */
	PTTableP->esGOffsets[10] = PTTableP->esGOffsets[8] + PTTableP->esGOffsets[2];	/* offset to 1010 */
	PTTableP->esGOffsets[11] = PTTableP->esGOffsets[10] + 1;							/* offset to 1011 */
	PTTableP->esGOffsets[12] = PTTableP->esGOffsets[8] + PTTableP->esGOffsets[4];	/* offset to 1100 */
	PTTableP->esGOffsets[13] = PTTableP->esGOffsets[12] + 1;							/* offset to 1101 */
	PTTableP->esGOffsets[14] = PTTableP->esGOffsets[8] + PTTableP->esGOffsets[4] + PTTableP->esGOffsets[2];	/* offset to 1110 */
	PTTableP->esGOffsets[15] = PTTableP->esGOffsets[14] + 1;							/* offset to 1111 */

	theSizef = interleaveFactor * sizeof (esGridData_t);
	for (i = 0; i < TH1_NUM_OFFSETS; i++) {
		PTTableP->esGOffsets[i] *= theSizef;	/* adjust for # and size of channels in interleaved grid tables */
	}

	for (i = 0; i < TH1_4D_PENTAHEDROA; i++) {
		index = bceiBase[i] -1;
		PTTableP->esFinder[i].tvert1 = PTTableP->esGOffsets[index];	/* grid offset to bcei corner */
		index = dfgjkmBase[i] -1;
		PTTableP->esFinder[i].tvert2 = PTTableP->esGOffsets[index];	/* grid offset to dfgjkm corner */
		index = hlnoBase[i] -1;
		PTTableP->esFinder[i].tvert3 = PTTableP->esGOffsets[index];	/* grid offset to hlno corner */
		PTTableP->esFinder[i].tvert4 = PTTableP->esGOffsets[15];			/* grid offset to bcei corner */
		PTTableP->esFinder[i].dx = dxBase[i];							/* dx multiplier */
		PTTableP->esFinder[i].dy = dyBase[i];							/* dy multiplier */
		PTTableP->esFinder[i].dz = dzBase[i];							/* dz multiplier */
		PTTableP->esFinder[i].dt = dtBase[i];							/* dt multiplier */
	}

	return PTErr;
}


/******************************************************************************/
/* set up input, grid, and output lookup tables for optimized evaluation */
static PTErr_t
	kcpInitTh1Luts (	KpInt32_t				numInputs,
						KpInt32_t				nOutputChans,
						fut_gtbl_ptr_t			aGTbl,
						fut_itbl_ptr_t FAR*		iTblsP,
						fut_gtbldat_ptr_t FAR*	gTblsP,
						fut_otbldat_ptr_t FAR*	oTblsP,
						PTTable_p				PTTableP)
{
KpInt32_t			theSizef, i, j, inputTableEntries, srcITblSize, gridTblSize;
KpUInt32_t			gridTblBytes, totalInputLutBytes, totalGridTableBytes, totalOutputLutBytes;
fut_itbl_ptr_t		theITbl;
fut_itbldat_ptr_t	iTblDat, srcITblDat;
esItbl_p			iLut;
esGridData_p		interleavedGridP;
esGridData_t		tmp;
evalOLutPtr_t		oLut0;
PTErr_t				PTErr;

	/* Input Tables */
	switch (PTTableP->esDataTypeI) {
	case KCM_UBYTE:
		inputTableEntries = FUT_INPTBL_ENT;
		break;

	case KCM_USHORT_12:
		inputTableEntries = FUT_INPTBL_ENT2;
		break;

	case KCM_USHORT:
		inputTableEntries = FUT_INPTBL_ENT3;
		break;

	default:
		goto ErrOut1;
	}
	PTTableP->esILutEntries = inputTableEntries;

	totalInputLutBytes = inputTableEntries * sizeof (esItbl_t) * numInputs;	/* size needed for input tables */

	allocEsMem (&PTTableP->esILuts, totalInputLutBytes);	/* allocate necessary memory for the input tables */
	if (PTTableP->esILuts.H == NULL) {
		goto ErrOut2;
	}

	/* set up the input table for each variable
	   this is done in 2 steps:
	   expand (if necessary) the current input table to the size needed
           for the input data
	   convert that table into a lut for evaluation
	 */

	theSizef = nOutputChans * sizeof (esGridData_t);
	for (i = numInputs-1; i >= 0 ; i--) {

		theITbl = iTblsP[i];						/* input table to convert */
		iLut = (esItbl_p)PTTableP->esILuts.P + (i * PTTableP->esILutEntries);	/* converted table */

		if (inputTableEntries == FUT_INPTBL_ENT) {	/* convert 256 entry table  */
			iTblDat = theITbl->tbl;
		}
		else {		/* need 4096 or 65536 entry table */
			if (theITbl->tblFlag == 0) {			/* use largest table available */
				srcITblDat = theITbl->tbl;
				srcITblSize = FUT_INPTBL_ENT;
			}
			else {
				srcITblDat = theITbl->tbl2;
				srcITblSize = FUT_INPTBL_ENT2;
			}

			if (srcITblSize == inputTableEntries) {
				iTblDat = srcITblDat;				/* already the right size */
			}
			else {	/* expand to needed size */
					/* because an entry of the eval lut is larger than an entry of the input table data, */
					/* the end of the eval lut can be used as temporary memory to expand the input table data. */

				iTblDat = (fut_itbldat_ptr_t)(iLut + inputTableEntries);	/* end of lut */
				iTblDat -= inputTableEntries;								/* back up to start of expanded table */

				expandITblMEP (srcITblDat, srcITblSize, iTblDat, inputTableEntries);
			}
		}

		th1Itbl2InLut (iTblDat, inputTableEntries, theSizef, iLut);		/* convert the input table to an eval lut */

		theSizef *= theITbl->size;
	}

	/* Grid Tables */
	gridTblBytes = aGTbl->tbl_size;

	/* is memory already allocated and the right size for the interleaved grid tables? */
	gridTblSize = gridTblBytes / sizeof (fut_gtbldat_t);
	totalGridTableBytes = gridTblSize * nOutputChans * sizeof (esGridData_t);	/* size needed for this fut's grid tables */

	allocEsMem (&PTTableP->esGLuts, totalGridTableBytes);	/* allocate necessary memory for the input tables */
	if (PTTableP->esGLuts.H == NULL) {
		goto ErrOut2;
	}

	/* interleave the grid tables */
	interleavedGridP = PTTableP->esGLuts.P;
	for (i = 0; i < gridTblSize; i++) {
		for (j = 0; j < nOutputChans; j++) {
			tmp = *gTblsP[j]++;
			*interleavedGridP++ = tmp;
		}
	}

	/* Output Tables */
	if (PTTableP->esDataTypeO == KCM_UBYTE)  {		/* 8 bit data  */
		totalOutputLutBytes = sizeof(KpUInt8_t);
	}
	else {
		totalOutputLutBytes = sizeof(KpUInt16_t);
	}

	totalOutputLutBytes *= FUT_OUTTBL_ENT * nOutputChans;		/* size needed for this fut's output luts */

	allocEsMem (&PTTableP->esOLuts, totalOutputLutBytes);	/* allocate necessary memory for the output tables */
	if (PTTableP->esOLuts.H == NULL) {
		goto ErrOut2;
	}

	/* set up the output table for each channel */
	oLut0.p8 = (KpUInt8_p)PTTableP->esOLuts.P;
	for (i = 0; i < nOutputChans; i++) {
		switch (PTTableP->esDataTypeO) {
		case KCM_UBYTE:
			th1Otbl2Out8Lut (oTblsP[i], &oLut0.p8[i * FUT_OUTTBL_ENT]);
			break;

		case KCM_USHORT_12:
			th1Otbl2Out12Lut (oTblsP[i], &oLut0.p16[i * FUT_OUTTBL_ENT]);
			break;

		case KCM_USHORT:
			th1Otbl2Out16Lut (oTblsP[i], &oLut0.p16[i * FUT_OUTTBL_ENT]);
			break;

		default:
			goto ErrOut1;
		}
	}

	return KCP_SUCCESS;


ErrOut1:
	PTErr = KCP_INVAL_EVAL;
	goto ErrOut;

ErrOut2:
	PTErr = KCP_NO_MEMORY;

ErrOut:
	freeEvalState (PTTableP->refNum);	/* free whatever may have been allocated */
	return PTErr;
}


/******************************************************************************/
/* Expand an input table using the Map End Points method */
/* source and destination table sizes when multiplied MUST fit a KpUInt32_t */
static void
	expandITblMEP (	fut_itbldat_ptr_t	srcITblDat,
					KpUInt32_t			srcITblSize,
					fut_itbldat_ptr_t	dstITblDat,
					KpUInt32_t			dstITblSize)
{
KpInt32_t			interp, fracHalf, temp1, temp2;
KpUInt32_t			sIMax, dIMax, fracBits, fracOne, fracMask;
KpUInt32_t			i, multMax, index, sPosition;
fut_itbldat_ptr_t	dstPtr;
fut_itbldat_t		base, next, interpData;
#if defined (KPALPHA)
fut_itbldat_t		diff;
#endif

	/* use linear interpolation for table expansion */
	dstPtr = dstITblDat;
	sIMax = srcITblSize -1;	/* maximum index for source table */
	dIMax = dstITblSize -1;	/* maximum index for destination table */

	multMax = sIMax * dIMax;	/* calc maximum # of fractional bits, using KpUInt32_t */

	for (fracBits = 0; (multMax & 0x80000000) == 0; multMax <<= 1, fracBits++) {}

	fracOne = 1 << fracBits;	/* 1.0 with this # of fractional bits */
	fracHalf = fracOne >> 1;	/* value for rounding */
	fracMask = fracOne -1;		/* fractional bits mask */

	sIMax <<= fracBits;			/* scale up by # of fractional bits */

	for (i = 0; i < dstITblSize; i++) {
		sPosition = (i * sIMax) / dIMax;
		index = sPosition >> fracBits;
		interp = sPosition & fracMask;

		base = srcITblDat[index];		/* base table data */
		next = srcITblDat[index +1];	/* next table data */

		temp1 = (next - base) * interp;
		temp2 =  temp1 + fracHalf;
		if (temp2 < 0) {
			interpData = base - ((-temp1 + fracHalf -1) >> fracBits);
		}
		else {
			interpData = base + (temp2 >> fracBits);
		}

		dstPtr[i] = interpData;
	}
}


/******************************************************************************/
/* Convert fut itbl to input table in special format */
static void
	th1Itbl2InLut (	fut_itbldat_ptr_t	futLut,
					KpInt32_t			nEntries,
					KpInt32_t			sizef,
					esItbl_p			evalLut)
{
KpInt32_t	i, index, v;

	/* Convert the itbl value to convenient values. */
	for (i = 0; i < nEntries; i++) {
		v = futLut[i];						/* get fut lut value */

		index = v >> FUT_INP_FRACBITS;		/* get fut integer index value */
		evalLut[i].index = index * sizef;	/* position index for offset into fut grid */

#if defined (TH1_FP_GRID_DATA)
		evalLut[i].frac = (esInterp_t)(v & 0xffff)/65535.0;	 /* get fut fractional value */
#else
		{
		KpUInt32_t	frac;
			frac = v & 0xffff;	 				/* get fut fractional value */
			evalLut[i].frac = frac;					/* in 2 steps to avoid sign extension */
		}
#endif
	}
}


/******************************************************************************/
/* convert a 12 bit lut to an 8 bit lut */
static void
	th1Otbl2Out8Lut (	fut_otbldat_ptr_t	oTbl,
						KpUInt8_p			outLut)
{
KpInt32_t	i;

	if (oTbl != FUT_NULL_OTBLDAT) {
		for (i = 0; i < FUT_OUTTBL_ENT; i++) {
			outLut[i] = (u_int8) FUT_OTBL_NINT (oTbl[i]);
		}
	}
	else {
		for (i = 0; i < FUT_OUTTBL_ENT; i++) {
			outLut[i] = (u_int8) FUT_OTBL_NINT (i);
		}
	}
}


/******************************************************************************/
/* convert a 12 bit lut to an 12 bit lut */
static void
	th1Otbl2Out12Lut (	fut_otbldat_ptr_t	oTbl,
						KpUInt16_p			outLut)
{
KpInt32_t	i, oTblData;

	for (i = 0; i < FUT_OUTTBL_ENT; i++) {
		if (oTbl != FUT_NULL_OTBLDAT) {
			oTblData = oTbl[i];				/* just copy */
		}
		else {
			oTblData = i;
		}

		KCP_SCALE_OTBL_DATA(oTblData, KCP_MAX_12BIT)

		outLut[i] = (KpUInt16_t) oTblData;
	}
}


/******************************************************************************/
/*    convert a 12 bit output lut to a 16 bit output (evaluation) lut         */
static void
	th1Otbl2Out16Lut (	fut_otbldat_ptr_t	oTbl,
						KpUInt16_p			outLut)
{
KpInt32_t	i, oTblData;

	for (i = 0; i < FUT_OUTTBL_ENT; i++) {
		if (oTbl != FUT_NULL_OTBLDAT) {
			oTblData = oTbl[i];
		}
		else {
			oTblData = i;
		}

		KCP_SCALE_OTBL_DATA(oTblData, KCP_MAX_16BIT)

		outLut[i] = (KpUInt16_t) oTblData;
	}
}


/******************************************************************************/
/* Deallocate all allocated memory and null out handles. */
/* Note: will deallocate properly if only some of the memory got allocated. */
void
	freeEvalState (	PTRefNum_t	PTRefNum)
{
PTTable_p PTTableP;

#if defined KCP_SINGLE_EVAL_CACHE
	putEvalStatePT (NULL);
#endif

	PTTableP = lockPTTable (PTRefNum);

	if (PTTableP == NULL) {
		return;
	}

	freeEsMem (&PTTableP->esILuts);		/* free current input */
	freeEsMem (&PTTableP->esGLuts);		/* grid */
	freeEsMem (&PTTableP->esOLuts);		/* and output lookup table memory */

	nullEvalState (PTTableP);			/* set evaluation state to null */

	unlockPTTable (PTRefNum);
}


/******************************************************************************/
/* Lock the evaluation state */
static void
	lockEvalState (	PTTable_p	PTTableP)
{
	if (PTTableP != NULL) {
		lockEsMem (&PTTableP->esILuts);
		lockEsMem (&PTTableP->esGLuts);
		lockEsMem (&PTTableP->esOLuts);
	}
}


/******************************************************************************/
/* Unlock the evaluation state */

void
	unlockEvalState (	PTTable_p	PTTableP)
{
	if (PTTableP != NULL) {
		unlockEsMem (&PTTableP->esILuts);		/* unlock input */
		unlockEsMem (&PTTableP->esGLuts);		/* grid */
		unlockEsMem (&PTTableP->esOLuts);	/* and output lookup table memory */
	}
}


/******************************************************************************/
/* Set the evaluation state to unused */
void
	nullEvalState (	PTTable_p	PTTableP)
{
KpInt32_t	i;

	if (PTTableP == NULL) {
		return;
	}

	PTTableP->esIOMask = 0;				/* set evaluation state to null */
	PTTableP->esDataTypeI = 0;
	PTTableP->esDataTypeO = 0;
	PTTableP->esILutEntries = 0;
	nullEsMem (&PTTableP->esILuts);
	nullEsMem (&PTTableP->esGLuts);
	nullEsMem (&PTTableP->esOLuts);

	for (i = 0; i < TH1_NUM_OFFSETS; i++) {
		PTTableP->esGOffsets[i] = 0;
	}

	return;
}


/******************************************************************************/
/* allocate evaluation state memory */
static void
	allocEsMem (esMem_p		theEsMemP,
				KpInt32_t	bytesNeeded)
{
	/* is memory already allocated and the correct size? */
	if (theEsMemP->bytes != bytesNeeded) {
		freeEsMem (theEsMemP);							/* free current allocation */

		theEsMemP->H = allocBufferHandle (bytesNeeded);	/* get new memory */
		if (theEsMemP->H != NULL) {
			theEsMemP->bytes = bytesNeeded;
		}
	}

	lockEsMem (theEsMemP);	/* lock it down */
}


/******************************************************************************/
/* free  evaluation state memory */
static void
	freeEsMem (esMem_p	theEsMemP)
{
	if (theEsMemP->H != NULL) {
		freeBuffer (theEsMemP->H);
	}
	nullEsMem (theEsMemP);
}


/******************************************************************************/
/*  set evaluation state memory to null */
static void
	nullEsMem (esMem_p	theEsMemP)
{
	theEsMemP->H = NULL;
	theEsMemP->P = NULL;
	theEsMemP->bytes = 0;
}


/******************************************************************************/
/* lock evaluation state memory */
static void
	lockEsMem (esMem_p	theEsMemP)
{
	if (theEsMemP != NULL) {
		if (theEsMemP->P == NULL) {
			theEsMemP->P = (esMem_p) lockBuffer (theEsMemP->H);
		}
	}
}


/******************************************************************************/
/* unlock evaluation state memory */
static void
	unlockEsMem (esMem_p	theEsMemP)
{
	if (theEsMemP != NULL) {
		(void) unlockBuffer (theEsMemP->H);
		theEsMemP->P = NULL;
	}
}


#if defined KCP_SINGLE_EVAL_CACHE

/* get the current chaining state from thread globals */
static PTErr_t
	getEvalStatePT (	PTRefNum_p	evalStatePT)
{
processGlobals_p	pGP;

	pGP = loadProcessGlobals();
	if (pGP == NULL) {
		return	KCP_NO_PROCESS_GLOBAL_MEM;
	}

	*evalStatePT = pGP->evalStatePT;

	unloadProcessGlobals ();

	return KCP_SUCCESS;
}


/* put the current chaining state in the thread globals */
static PTErr_t
	putEvalStatePT (	PTRefNum_t	evalStatePT)
{
processGlobals_p	pGP;

	pGP = loadProcessGlobals();
	if (pGP == NULL) {
		return	KCP_NO_PROCESS_GLOBAL_MEM;
	}

	pGP->evalStatePT = evalStatePT;

	unloadProcessGlobals ();

	return KCP_SUCCESS;
}

#endif
