/*
 * @(#)fwdmono.c	1.6 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: fwdmono.c	1.5 98/02/17
 *
 *	Contains:	makeForwardXformMono
 *
 **********************************************************************
 *    COPYRIGHT (c) 1997-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 **********************************************************************
 */

#include <math.h>
#include <string.h>
#include "csmatrix.h"
#include "monopt.h"
#include "makefuts.h"

PTErr_t makeForwardXformMono (ResponseRecord FAR*	grayTRC,
									KpUInt32_t		gridsize,
									fut_ptr_t *		theFut)
{
PTErr_t			PTErr = KCP_FAILURE;
KpInt32_t		dim;
fut_ptr_t		futp;
fut_otbldat_ptr_t	otblDat;
ResponseRecord_t rrt;
KpInt32_t		nchan;
int				futReturn;
double			gamma;
KpUInt16_t		rrpData[2] = { 0, RRECORD_DATA_SIZE -1 };

 	/* Check for valid ptrs */
	if (Kp_IsBadReadPtr(grayTRC, sizeof(*grayTRC)))
		return KCP_BAD_PTR;

	if (gridsize < 2) {
	   return KCP_INCON_PT;
	}

	/* assume all dimensions are the same */
	dim = (KpInt32_t)gridsize;

	/* Create (1D -> 3D) FuT */
	futp = fut_new_empty (1, &dim, 3);
	if (futp == FUT_NULL) {
	   return KCP_NO_MEMORY;
	}

	/* compute new table entries */
	for (nchan = 0; nchan < 3; nchan++) {
		if (!fut_calc_gtblEx (futp->chan[nchan]->gtbl, lin_gFunc_x, NULL)) {
			fut_free (futp);
			return KCP_SYSERR_0;
		}

		if (!fut_calc_otblEx (futp->chan[nchan]->otbl, otblFunc, NULL)) {
			fut_free (futp);
			return KCP_SYSERR_0;
		}
	}

	if ( futp->idstr != NULL ) {
		fut_free_idstr (futp->idstr);
		futp->idstr = NULL;
	}

	/* get address of the first output table */
	futReturn = fut_get_otbl (futp, 0, &otblDat);
	if ((futReturn != 1) || (otblDat == (fut_otbldat_ptr_t)NULL)) {
		fut_free (futp);
		return KCP_SYSERR_0;
	}

	/* setup the output table */
	switch (grayTRC->count)
	{
		case 0:
				/* setup the responseRecord struct */
				rrt.count = 2;
				rrt.data = rrpData;

				/* make the output table */
				PTErr = calcOtblLSN (otblDat, &rrt);
				if (PTErr != KCP_SUCCESS) {
					fut_free (futp);
					return (PTErr);
				}
				break;
		case 1:
				gamma = (double)grayTRC->data[0] / SCALEDOT8;
				if (gamma <= 0.0) {
					fut_free (futp);
					return KCP_INCON_PT;
				}

				/* make the output table */
				PTErr = calcOtblLS1 (otblDat, gamma);
				if (PTErr != KCP_SUCCESS) {
					fut_free (futp);
					return (PTErr);
				}
				break;

		default:
				/* make the output table */
				makeMonotonic (grayTRC->count, grayTRC->data);
				PTErr = calcOtblLSN (otblDat, grayTRC);
				if (PTErr != KCP_SUCCESS) {
					fut_free (futp);
					return (PTErr);
				}
	}

     /* Set fut pointer and return successfully:  */
	*theFut = futp;
	return (PTErr);
}
