/*
 * @(#)kcpcache.h	1.7 98/09/15
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: kcpcache.h	1.60	98/02/17
 *
 *	Defines interface to color processors' special evaluations
 *
 ****************************************************************
 *    COPYRIGHT (c) 1995-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */

#ifndef _KCPCACHE_H_
#define _KCPCACHE_H_ 1

#include "kcmptdef.h"
#include "kcptmgr.h"
#include "attrib.h"


#define KCP_MAX_12BIT (0xfff)
#define KCP_MAX_16BIT (0xffff)

#define KCP_SCALE_OTBL_DATA(data, maxData) \
		data = ((data * maxData) + (FUT_MAX_PEL12/2)) / FUT_MAX_PEL12; \
		if (data > maxData) { \
			data = maxData; \
		}

#define KCP_EXTRACT_COMPONENT(data, dataBits, position) \
	((data >> position) & ((1 << dataBits) -1))

#define KCP_CONVERT_DOWN(data, startBits, endBits) \
	((data + ((1 << (startBits - endBits -1)) - (data >> endBits))) >> (startBits - endBits))

/* the following macro will only work if (startBits - (endBits - startBits)) >= 0
 * which means endBits can be no more than twice startBits
 * Use the macro successively if this condition is not met */
#define KCP_CONVERT_UP(data, startBits, endBits) \
((data << (endBits-startBits)) + ((data >> (startBits - (endBits - startBits)))))

/* image pointers definition */
typedef union imagePtr_u {
	PTImgAddr_t pI;
	KpUInt8_p	p8;
	KpUInt16_p	p16;
	KpUInt32_p	p32;
} imagePtr_t, FAR* imagePtr_p;

/* Evaluation lut definition  */
typedef union evalOLutPtr_u {
	KpUInt8_p	p8;
	KpUInt16_p	p16;
} evalOLutPtr_t, FAR* evalOLutPtr_p;

/* Number of input variables for each output channel */
#define TH1_MAX_INPUT_VARS (4)

/* Maximum number of output channels */
#define TH1_MAX_OUTPUT_CHANS (8)

/* The overhead associated with optimizing the tables means that below some
	threshold evaluation will be faster using non-optimized evaluation */
#define TH1_MIN_EVALS (1500)

/* #define TH1_FP_GRID_DATA 1 */	/* undefine to use integer */

#if defined (TH1_FP_GRID_DATA)
	typedef KpFloat32_t esGridData_t, FAR* esGridData_p;	/* floating point grid table data */
	typedef KpFloat32_t esInterp_t, FAR* esInterp_p;		/* floating point interpolant */
#else
	typedef KpInt16_t esGridData_t, FAR* esGridData_p;	/* integer grid table data */
	typedef KpInt32_t esInterp_t, FAR* esInterp_p;		/* integer interpolant */
#endif

typedef struct esItbl_s {
	KpInt32_t	index;	/* index into grid table */
	esInterp_t	frac;	/* the interpolant */
} esItbl_t, FAR* esItbl_p;

/*  flags for input lut size  */
#define EVAL_LUT_8BIT  0
#define EVAL_LUT_16BIT 1     /* using 16 bit input data */

typedef void (*evalTh1Proc_t) (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
typedef evalTh1Proc_t FAR* evalTh1Proc_p;
typedef void (*formatFunc_t) (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);

typedef struct evalControl_s {
	callBack_p		callBack;
	evalTh1Proc_t	evalFunc;
	KpInt32_t		nFuts;
	PTTable_p*		evalList;
	KpUInt32_t		compatibleDataType;
	KpUInt32_t		optimizedEval;
	KpInt32_t		imageLines, imagePels;
	imagePtr_t		inputData[FUT_NICHAN], outputData[FUT_NOCHAN];
	KpInt32_t		inPelStride[FUT_NICHAN], inLineStride[FUT_NICHAN], outPelStride[FUT_NOCHAN], outLineStride[FUT_NOCHAN];

	/* needed only if image formats do not match an optimized function */
	formatFunc_t	formatFuncI, formatFuncO;
	KpInt32_t		nInputs, nOutputs;
	KpUInt32_t		evalDataTypeI, evalDataTypeO;
	KpInt32_t		tempPelStride[FUT_NOCHAN];
	PTErr_t			errnum;
} evalControl_t, FAR* evalControl_p, FAR* FAR* evalControl_h;


#if !defined(KPNONANSIC)

PTErr_t		allocEvalState (PTTable_p);
void		unlockEvalState (PTTable_p);
PTErr_t		checkEvalState (PTTable_p, KpInt32_t, KpUInt32_t, KpUInt32_t, evalControl_p);
PTErr_t		initEvalState (PTTable_p, evalControl_p);

void evalTh1gen (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o1d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o1d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o2d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o2d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o3d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o3d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1iB24oB24 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1iL24oL24 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1iQDoQD (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1iQDo3 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3oQD (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o4d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o4d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o5d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o5d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o6d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o6d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o7d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o7d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o8d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i3o8d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o1d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o1d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o2d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o2d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o3d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o3d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o3QD (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1iB32oB32 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1iL32oL32 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o4d8 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);
void evalTh1i4o4d16 (imagePtr_p, KpInt32_p, imagePtr_p, KpInt32_p, KpInt32_t, PTTable_p);

void pass8in (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format555to8 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format565to8 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format8to12 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void pass16in (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format16to12 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format555to12 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format565to12 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format10to12 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void pass8out (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format8to555 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format8to565 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format12to8 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void pass16out (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format12to16 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format12to555 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format12to565 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);
void format12to10 (KpInt32_t, KpInt32_t, imagePtr_p, KpInt32_p, imagePtr_p);

#endif

#endif	/* _KCPCACHE_H_ */
