/*
 * @(#)kcpmgr.c	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: kcpmgr.c	2.94 98/02/26
 *
 *	Contains:       execute a KCMS function which was initially
 *			called through KCMS_proc_send
 *
 ****************************************************************
 *    COPYRIGHT (c) 1991-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */


#define KCP_GLOBAL

#include "kcms_sys.h"
#include "bytes.h"

#include <string.h>

#include "kcmptlib.h"
#include "kcmptdef.h"
#include "kcptmgr.h"
#include "kcpmgru.h"
#include "kcpfchn.h"
#if defined (KCP_ACCEL)
#include "ctelib.h"
#endif


static	KpThreadMemHdl_t	theRootID;
static initializedGlobals_t	initializedGlobals;	/* initialized globals */


/*--------------------------------------------------------------------
 * FUNCTION NAME
 * PTGetFlavor
 *
 * DESCRIPTION
 * This function returns the Color Processor flavor
 *
 *--------------------------------------------------------------------*/

PTErr_t PTGetFlavor (KpInt32_t * kcpFlavor)
{
KpInt32_t		ptFlavor = 0;

#if defined (KCP_COMP_2)
	ptFlavor |= PT_COMPOSITION2;
#endif

#if defined (KCP_ACCEL)
	ptFlavor |= PT_ACCELERATOR;
#endif

	ptFlavor |= PT_TLI_EVALUATION1;	/* determine the evaluation type */

	*kcpFlavor = ptFlavor;

	return (KCP_SUCCESS);
}


/* set up the main driver */
KpInt32_t
	KCMDsetup (KpGenericPtr_t FAR* theIGPtr)
{
initializedGlobals_p	iGP;			/* initialized globals */
#if defined (KCP_ACCEL)
PTErr_t 	PTErr;
KpInt16_t	num;
#endif

	iGP = getInitializedGlobals ();
	if (iGP == NULL) {
		return (KCMS_FAIL);
	}

	bzero ((KpGenericPtr_t) iGP, sizeof (initializedGlobals_t));	/* set them all to 0 */

	KCPInitIGblP (theIGPtr, iGP);		/* do platform specific initialization */

#if defined (KCP_COMP_2)
	KCPChainSetup (iGP);				/* initialize chaining rules */
#endif

	iGP->PTCubeSize = EIGHT_CUBE;		/* setup default pt size */

#if defined (KCP_ACCEL)
	if (iGP->SWalways == 0) {
	/* Either we're not a mac, or we are but there was no command/s. */
		PTErr = PTProcessorReset_cte ();		/* initialize the CTE driver */
		if (PTErr == KCP_SUCCESS) {
			PTErr = PTEvaluators_cte (&num);	/* check for the CTE driver */
			if ((PTErr == KCP_SUCCESS) && (num > 0)) {
				iGP->haveCTE = 1;				/* CTE is present */
			}
		}
	}
#endif

	return (KCMS_SUCCESS);
}


PTErr_t
	KCMDTerminate (void)
{
	return (KCP_SUCCESS);
}


#if (defined KCP_SINGLE_EVAL_CACHE) || (defined KCP_MACPPC_MP)

/* set up the globals for each application */
KpInt32_t
	KCPappSetup (KpGenericPtr_t theIGPtr)
{
processGlobals_p	pGP;

	if (theIGPtr) {}

	/* set up the process globals */
	pGP = KpThreadMemCreate (&theRootID, KPPROCMEM, sizeof(processGlobals_t));
	if (pGP == NULL) {
		return (KCMS_FAIL);
	}

	bzero ((KpGenericPtr_t) pGP, sizeof (processGlobals_t));	/* set them all to 0 */

	unloadProcessGlobals ();	/* Unlock this app's Globals */

	return KCMS_SUCCESS;
}


PTErr_t
	KpTermProcess (void)
{
processGlobals_p	pGP;

	pGP = loadProcessGlobals();
	if (pGP == NULL) {
		return KCP_FAILURE;
	}

	KpThreadMemUnlock (&theRootID, KPPROCMEM);
	KpThreadMemDestroy (&theRootID, KPPROCMEM);

	return (KCP_SUCCESS);
}


/* load the process globals */
/* these must be unloaded with unloadProcessGlobals when done */
processGlobals_p
	loadProcessGlobals (void)
{
processGlobals_p	pGP;

	pGP = (processGlobals_p) KpThreadMemFind (&theRootID, KPPROCMEM);

	return pGP;
}


/* unload the process globals */
void
	unloadProcessGlobals (void)
{
	KpThreadMemUnlock (&theRootID, KPPROCMEM);
}

#endif


#if !defined KPMAC
/* thread dummies, only for backward compatibility */
PTErr_t
	PTInitThread (void)
{
	return KCP_SUCCESS;
}


PTErr_t
	PTTermThread (void)
{
	return KCP_SUCCESS;
}

#endif


/* reset the color processor to post-initialization state */
PTErr_t
	PTProcessorReset (void)
{
PTErr_t PTErr = KCP_SUCCESS;

#if defined (KCP_ACCEL)
	PTErr = PTProcessorReset_cte(); 	/* reset CTE */
#endif

	return (PTErr);
}


PTErr_t
	PTTerminate (void)
{
PTErr_t	PTErr;

	PTErr = PTTerminatePlatform ();

	return (PTErr);
}


initializedGlobals_p
	getInitializedGlobals (void)
{
	return &initializedGlobals;
}


#if defined (KCP_COMP_2)

/* get the current chaining state */
PTErr_t
	getChainState (chainState_p	FAR * cS)
{

	*cS = (chainState_p) KpThreadMemFind (&theRootID, KPTHREADMEM);
	if (*cS == NULL) {
		*cS = KpThreadMemCreate (&theRootID, KPTHREADMEM, sizeof(chainState_t));
		if (*cS == NULL) {
			return KCP_NO_THREAD_GLOBAL_MEM;
		}

		bzero (*cS, sizeof (chainState_t));
	}

	return KCP_SUCCESS;
}


/* put the current chaining state */
PTErr_t
	putChainState (	chainState_p	cS)
{
	if (cS) {}

	KpThreadMemUnlock (&theRootID, KPTHREADMEM);

	return KCP_SUCCESS;
}


/* abort the current PT chaining function */
void
	clearChain (	chainState_p	cS)
{
	PTCheckOut (cS->currentPT);			/* free the current PT */

	bzero (cS, sizeof (chainState_t));	/* no chaining active */

	KpThreadMemDestroy (&theRootID, KPTHREADMEM);
}

#endif
