/*
 * @(#)kcptmgr.h	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: kcptmgr.h	2.138 98/03/03
 *
 *	Contains:	header file for KCMS system processor
 *
 *	Written by:	Drivin' Team
 *
 *********************************************************************
 *    COPYRIGHT (c) 1992-1998 Eastman Kodak Company
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *********************************************************************
 */

#ifndef _KCMSYS_H_
#define _KCMSYS_H_ 1

/* !!!! change 'vers' in KPCPversion.r !!!! */
/* THIS MUST NOT GO TO 6 (count em 6) chars or the precision api will barf !!! */
#define KPCP_VERSION "3.50b"		/* color processor version number */


/* #define DIAGWIND 1 */		/* do this for the alpha release only!!! */

#include "kcms_sys.h"
#include "kcmptlib.h"
#include "fut.h"


#if defined (KPMAC)
#include <Components.h>
#endif

#if defined (KCP_MACPPC_MP)
#include <MP.h>
#endif

/* chaining constants */
#define MAX_PT_CHAIN_SIZE 20	/* maximum # of PTs in a chain */
#define NOT_CHECKED_IN 0		/* show this pt is not checked in */
#define IS_CHECKED_IN 1			/* show this pt is checked in */
#define NOT_SERIAL_PT 2			/* show this is not a serial pt */
#define IS_SERIAL_PT 3			/* show this is a serial pt */

/* the dimensions of the futs */
#define EIGHT_CUBE	8

/* the maximum number of processors active at one time (including main processor) */
#define KCP_MAX_PROCESSORS	4

/* handle dereferencing */
#if defined KPMAC
#define FFUTP(handle) (*(fut_ptr_t FAR*)handle)
#define FCHANP(handle) (*(fut_chan_ptr_t FAR*)handle)
#else
#define FFUTP(handle) ((fut_ptr_t)handle)
#define FCHANP(handle) ((fut_chan_ptr_t)handle)
#endif

/* PT format */
typedef struct PT_s {
	KpInt32_t	magic;			/* for technology id and byte swapping */
	KpInt32_t	version;		/* allow for future modifications */
	KpInt32_t	attrSize;		/* # bytes of attribute information */
	KpChar_t	hdr[488];		/* private header info */
							/* followed by attribute data */
							/* followed by technology data */
} PT_t;

#if defined KCP_COMP_2
#define MAX_RULE_COMPONENT 3	/* max # of rule components */
#define MAX_ICOMP 20			/* maximum # of input PT composition groups */
#define MAX_OCOMP 20			/* maximum # of output PT composition groups */

/* composition rules */
typedef struct composeRule_s {
	KpInt32_t	start[MAX_RULE_COMPONENT+1];	/* start rules */
	KpInt32_t	finish[MAX_RULE_COMPONENT+1];	/* finish rules */
} composeRule_t, FAR* composeRule_p;

typedef struct chainState_s {
	KpInt32_t	compMode;					/* composition mode for entire chain */
	KpInt32_t	rulesKey;					/* key to apply chaining rules */
	KpInt32_t	chainLength;				/* # of PTs in the chain */
	KpInt32_t	chainIndex;					/* current position in chain */
	PTRefNum_t	currentPT;					/* current result of PT chain */
	KpInt32_t 	iComp;						/* input chain rule of first PT in chain */
	KpInt32_t 	oComp;						/* output chain rule of last PT in chain */
	KpInt32_t 	inSense;					/* input sense of first PT in chain */
	KpInt32_t 	outSense;					/* output sense of last PT in chain */
	PTRefNum_t	chainDef[MAX_PT_CHAIN_SIZE];	/* list of PTs to chain */
} chainState_t, FAR* chainState_p;

#endif

#define TH1_NUM_OFFSETS (2*2*2*2)	/* number of offsets in the interpolation volume */
#define TH1_4D_PENTAHEDROA	24		/* number of pentahedra in a 4D volume */

typedef struct esMem_s {
	KpInt32_t		bytes;
	KpGenericPtr_t	P;
	KpHandle_t		H;
} esMem_t, FAR* esMem_p;

typedef struct th1_4dControl_s {
	KpUInt32_t	tvert1, tvert2, tvert3, tvert4;	/* offsets from the base grid point to the pentahedral corners */
	KpUInt32_t	dx, dy, dz, dt;
} th1_4dControl_t, FAR* th1_4dControl_p, FAR* FAR* th1_4dControl_h;

/* PT table definition */
typedef struct PTTable_s {
	KpUInt32_t	uid;					/* structure identifier */
	PTRefNum_t	refNum;					/* the reference number for the PT */
	KpHandle_t	hdr;					/* address of the PT header */
	KpHandle_t	attrBase;				/* address of the PT attributes */
	KpHandle_t	data;					/* address of the PT data */
	KpUInt32_t	checkInFlag;			/* is this PT checked in? */
	KpInt32_t	inUseCount;				/* number of links to this PT */
	KpUInt32_t	serialPTflag;			/* is this a serial PT? */
	KpInt32_t	serialCount;			/* # of PTs in the serial chain */
	PTRefNum_t	serialDef[MAX_PT_CHAIN_SIZE];	/* list of PTs for serial eval */

	/* this is what is needed for the pending evaluation */
	/* It may not match the current evaluation state */
	fut_ptr_t	futP;				/* locked fut */
	KpUInt32_t	evalIOMask;			/* I/O mask for the evaluation */
	KpUInt32_t	evalDataTypeI;		/* input and output data types */
	KpUInt32_t	evalDataTypeO;

	/* the following holds the current evaluation state */
	KpUInt32_t	esIOMask;			/* I/O mask for the evaluation */
	KpUInt32_t	esDataTypeI;		/* input and output data types */
	KpUInt32_t	esDataTypeO;

	KpInt32_t	esILutEntries;		/* # of entries in each input table */
	esMem_t		esILuts;			/* Input lookup tables for the current evaluation state */
	esMem_t		esGLuts;			/* grid lookup tables for for the current evaluation state */
	esMem_t		esOLuts;			/* Output lookup tables for the current evaluation state */

	KpInt32_t	esGOffsets[TH1_NUM_OFFSETS];	/* Offset in grid from base of a cell to next entry in each dimension */

	th1_4dControl_t	esFinder[TH1_4D_PENTAHEDROA];	/* finder tables for pentahedron data */

} PTTable_t, FAR* PTTable_p;

#define PTT_UID (0x70747462) /* PTTable_s identifier, is 'pttb' */

#if defined (KPMAC)
typedef PTErr_t (FAR PASCAL *PTRelay_t) (long savedA5, long savedA4, PTProgress_t progressFunc, KpInt32_t perCent);
#endif

#if defined (KCP_MACPPC_MP)
#define	kEvaluate (1)
#define	kTerminate (2)
#define	kComplete (3)
#define	kErrMP (4)

typedef struct taskControl_s {
	MPTaskID	ID;				/* ID of this task */
	MPQueueID	fromMain;		/* messages from main task to others */
	MPQueueID	toMain;			/* messages from others to main task */
	MPQueueID	termination;	/* termination queue for this task */
} taskControl_t, FAR* taskControl_p;

#endif

/* structures for explicit global handling */

/* globals which are common to all components and which are set up at system startup time */
typedef struct initializedGlobals_s {
	KpInt32_t	SWalways;  			/* init = 0: default is to try to load the CTE driver. */
	KpChar_t	KCPDataDir [256];	/* fully qualified path to CP data directory */
	KpInt32_t	PTCubeSize;			/* size of ICC profiles (pts) */
	KpInt32_t	numProcessors;		/* number of available processors */

#if defined (KCP_ACCEL)
	KpUInt32_t	haveCTE;			/* 1 if CTE present */
#endif

#if defined (KPMAC)
	PTRelay_t	callBackRelay;		/* progress call back relay for PPC->68K */
#endif

#if defined KCP_COMP_2
	composeRule_t	composeRule[MAX_ICOMP][MAX_OCOMP];
#endif

#if defined (KPWIN)
	KpModuleId	moduleId;
	KpInstance	instance;
	KpModuleId	appModuleId;
#endif

} initializedGlobals_t, FAR* initializedGlobals_p;


#if (defined KCP_SINGLE_EVAL_CACHE) || (defined KCP_MACPPC_MP)
typedef struct processGlobals_s {
#if defined KCP_SINGLE_EVAL_CACHE
	PTRefNum_t			evalStatePT;			/* the PT which has evaluation tables */
#endif

#if defined (KCP_MACPPC_MP)
	KpUInt32_t			isMPinitialized;		/* Only want to initialize the MP once */
	taskControl_p		taskListP;
#endif

} processGlobals_t, FAR* processGlobals_p;
#endif


typedef struct callBack_s {
	KpInt32_t 			loopStart;
	KpInt32_t			loopCount;
	PTProgress_t		progressFunc;
	KpInt32_t			currPasses;
	KpInt32_t			totalPasses;
	KpInt32_t			lastProg100;
	PTProgress_t		apiProgressFunc;
#if defined (KPMAC)
	long 				gHostA4;
	long 				gHostA5;
#endif
} callBack_t, FAR* callBack_p;


/* color processor globals */
#ifndef KCP_GLOBAL
#define KCP_GLOBAL extern
#endif


/* function prototypes */
PTErr_t		doActivate ARGS((PTRefNum_t, KpInt32_t, PTAddr_t, KpInt32_t));
PTErr_t		doGetSizeF ARGS((PTRefNum_t, PTType_t, KpInt32_p, KpInt32_t));

PTErr_t		registerPT ARGS((KpHandle_t, KpHandle_t, PTRefNum_p));
PTErr_t		getPTStatus ARGS((PTRefNum_t));
KpHandle_t	getPTHdr ARGS((PTRefNum_t));
PTErr_t		setPTHdr ARGS((PTRefNum_t, KpHandle_t));
KpHandle_t	getPTAttr ARGS((PTRefNum_t));
void		setPTAttr ARGS((PTRefNum_t, KpHandle_t));
KpHandle_t	getPTData ARGS((PTRefNum_t));
PTErr_t 	resolvePTData (PTRefNum_t, KpInt32_p, PTRefNum_p);
KpInt32_t	getCheckInSize ARGS((PTRefNum_t));
void		makeActive ARGS((PTRefNum_t, KpHandle_t));
void		makeSerial ARGS((PTRefNum_t));
PTErr_t		makeInActive ARGS((PTRefNum_t));
PTErr_t	 	makeCheckedOut (PTRefNum_t);
PTErr_t 	addSerialPT (PTRefNum_t, PTRefNum_t, PTRefNum_p);
PTTable_p	lockPTTable ARGS((PTRefNum_t));
void		unlockPTTable ARGS((PTRefNum_t));

KpInt32_t getAttrSize ARGS((KpHandle_t));
PTErr_t GetAttribute ARGS((KpHandle_t, KpInt32_t, KpInt32_p, KpChar_p));
PTErr_t readAttributes ARGS((KpFd_p, KpInt32_t, KpHandle_t FAR*));
PTErr_t writeAttributes ARGS((KpFd_p, KpHandle_t));
PTErr_t freeAttributes ARGS((KpHandle_t));
PTErr_t copyAllAttr ARGS((PTRefNum_t, PTRefNum_t));

PTErr_t GetEval ARGS((PTEvalTypes_t, PTEvalTypes_p));
PTErr_t	getResizeAuxPT (PTRefNum_t, PTRefNum_p);

PTErr_t	TpReadHdr ARGS((KpFd_p, KpHandle_t FAR*, PTType_p));
PTErr_t	TpWriteHdr ARGS((KpFd_p, PTType_t, KpHandle_t, KpInt32_t));
PTErr_t	TpCompareHdr ARGS((KpHandle_t, KpHandle_t));
PTErr_t	TpFreeHdr ARGS((KpHandle_t));
PTErr_t	TpReadData ARGS((KpFd_p, PTType_t, PTRefNum_t, KpHandle_t, KpHandle_t FAR*));
PTErr_t	TpWriteData ARGS((KpFd_p, PTType_t, KpHandle_t, KpHandle_t));
PTErr_t	TpFreeData ARGS((KpHandle_t));
KpInt32_t TpGetDataSize ARGS((KpHandle_t, KpHandle_t, PTType_t));
PTErr_t TpGenerateAttr ARGS((PTRefNum_t));
PTErr_t TpSetImplicitAttr ARGS((PTRefNum_t));
#if !defined KCMS_NO_CRC
PTErr_t TpCalCrc ARGS((KpHandle_t, KpHandle_t, KpInt32_p));
#endif

PTErr_t PTEvalP (PTRefNum_t, PTEvalPB_p, PTEvalTypes_t, KpInt32_t, KpInt32_t, opRefNum_p, callBack_p);
PTErr_t PTEvaluate (PTRefNum_t, PTEvalDTPB_p, PTEvalTypes_t, KpInt32_t, KpInt32_t, opRefNum_p, callBack_p);
PTErr_t PTEvalSeq (KpInt32_t, PTTable_p*, PTEvalDTPB_p, callBack_p);
void	nullEvalState (PTTable_p);
void	freeEvalState (PTRefNum_t);

KpInt32_t PTMemTest ARGS((void));
PTErr_t unlockPT ARGS((KpHandle_t, fut_ptr_t));
PTErr_t	checkPT (PTRefNum_t);
PTErr_t	getPTCubeSize (KpInt32_p);

PTErr_t PTNewEmpty ARGS((KpInt32_t, KpInt32_p, KpInt32_t, PTRefNum_p));
PTErr_t PTNewEmptySep ARGS((KpInt32_t, KpInt32_p, PTRefNum_p));
PTErr_t fut2PT ARGS((fut_ptr_t, PTRefNum_p));
PTErr_t PTGetItbl ARGS((PTRefNum_t, KpInt32_t, KpInt32_t, KpHandle_t FAR*));
PTErr_t PTGetGtbl ARGS((PTRefNum_t, KpInt32_t, KpInt32_p, KpInt32_p, KpHandle_t FAR*));
PTErr_t PTGetOtbl ARGS((PTRefNum_t, KpInt32_t, KpHandle_t FAR*));

fut_ptr_t get_lab2xyz ARGS((KpInt32_t));

#if defined KCP_COMP_2
KpChar_p getComposeRule (KpInt32_t);
void KCPChainSetup ARGS((initializedGlobals_p));
void clearChain ARGS((chainState_p));
PTErr_t getChainState (chainState_p FAR*);
PTErr_t putChainState (chainState_p	cS);
#endif

KpInt32_t	CollisionCheck ARGS((void));
KpInt32_t	kcpGetColorSpace (PTRefNum_t, KpInt32_t);


#if defined DIAGWIND
void ClrGErr ARGS((void));
void diagWindow ARGS((KpChar_p errStr, KpInt32_t errnum));
#else
#define ClrGErr()
#define diagWindow(errStr, errnum)
#endif

initializedGlobals_p	getInitializedGlobals (void);

PTErr_t		KCMDTerminate (void);
KpInt32_t	KCMDsetup (KpGenericPtr_t FAR*);

#if (defined KCP_SINGLE_EVAL_CACHE) || (defined KCP_MACPPC_MP)
KpInt32_t			KCPappSetup (KpGenericPtr_t);
PTErr_t				KpTermProcess (void);
processGlobals_p	loadProcessGlobals (void);
void				unloadProcessGlobals (void);

	#if defined (KCP_MACPPC_MP)
	OSStatus	evalTaskMac (void*);
	PTErr_t		KCPInitializeMP (void);
	#endif
#endif

void		KCPInitIGblP (KpGenericPtr_t FAR*, initializedGlobals_p IGblP);
PTErr_t		PTTerminatePlatform (void);

PTErr_t		SetKCPDataDirProps (KpFileProps_p);

#if defined (KPMAC)
KpInt32_t	KCMDtakeDown (void);
KpInt32_t	KCPappTakeDown (void);
PTErr_t		KCMDcommand (KpInt16_t, KpGenericPtr_t);
KpInt32_t	CanDoSelector (KpInt16_t);
KpInt32_t	GetVersion (void);
void		kcpGetFPU ();
#endif

#endif
