/*
 * @(#)lab2xyz.c	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: lab2xyz.c	1.12 98/02/17
 *	Contains:	This module contains functions to create
 *			an lab->XYZ FuT.
 *
 *			Created by lsh, November 11, 1993
 *
 *********************************************************************
 *    COPYRIGHT (c) 1993-1998 Eastman Kodak Company
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *********************************************************************
 */

#include "kcptmgr.h"
#include "makefuts.h"

/******
#define	D65	1
******/
#undef	D65

#ifdef	D65				/* white point for CIELAB */
#define	XWHITE	0.950
#define	YWHITE	1.000
#define ZWHITE	1.089
#else	/* D50 */
#define	XWHITE	0.9642
#define	YWHITE	1.0000
#define	ZWHITE	0.8249
#endif

typedef struct fData_s {
	KpInt32_t	dimSize, ySize, zSize, chan;
} fData_t, FAR* fData_p;


/*---------------------------------------------------------------------
 *  ubyte, Lifun, abifun -- input mappings:  Convert a* and b* to unsigned
 *---------------------------------------------------------------------
 */
static double
	Lifun (double x, KpGenericPtr_t dataP)
{
	if (dataP) {}

	return x;
}

static double
	abifun (double ab, KpGenericPtr_t dataP)
{
KpInt32_t	dimSize;
double		dimSizeM1, dimSizeD2;

	dimSize = ((fData_p) dataP)->dimSize;

	dimSizeM1 = (double)(dimSize - 1);
	dimSizeD2 = (double)(dimSize / 2);

	ab = 255.0 * ab;

	if (ab <= 128.0) {
		ab = (dimSizeD2 / dimSizeM1) * (ab / 128.0);
	}
	else {
		ab = 1.0 - ((dimSizeM1 - dimSizeD2) / dimSizeM1)
			* ((255.0 - ab) / 127.0);
	}

	return ab;
}


/*---------------------------------------------------------------------
 *  trifun -- grid-table functions:  Lab -> XYZ
 *
 * representing L* in [0, 100], a* and b* in [-128, 127] */

static fut_gtbldat_t
	gfun (double_p	args, KpGenericPtr_t dataP)
{
double	l, a, b, tristim, g;
KpInt32_t	ySize, zSize;

	l = args[0];
	a = args[1];
	b = args[2];

	ySize = ((fData_p) dataP)->ySize;
	zSize = ((fData_p) dataP)->zSize;

/* Undo grid mapping of a* and b*:  */
	if (a <= (double)(ySize / 2) / (double)(ySize - 1)) {
		a = (128.0 / 255.0) * ((double)(ySize - 1) / (double)(ySize / 2)) * a;
	}
	else {
		a = 1.0 - (127.0 / 255.0) * ((double)(ySize - 1) / (double)(ySize - 1 - (ySize / 2)))
			* (1.0 - a);
	}

	if (b <= (double)(zSize / 2) / (double)(zSize - 1)) {
		b = (128.0 / 255.0) * ((double)(zSize - 1) / (double)(zSize / 2)) * b;
	}
	else {
		b = 1.0 - (127.0 / 255.0) * ((double)(zSize - 1) / (double)(zSize - 1 - (zSize / 2)))
			* (1.0 - b);
	}

/* Shift and rescale a* and b*:  */
	a = 255.0 * a - 128.0;		/* CIE 1976 a* */
	a = 0.00232 * a;		/* H(X/X_n) - H(Y/Y_n), in [-0.297, 0.295] */
	b = 255.0 * b - 128.0;		/* CIE 1976 b* */
	b = 0.00580 * b;		/* H(Y/Y_n) - H(Z/Z_n), in [-0.742, 0.737] */

/* Separate XYZ channels:  */
	switch (((fData_p) dataP)->chan) {
	case 1:
		tristim = a + l;	/* H(X/X_n), in [-0.297, 1.295 */
		break;

	case 2:
		tristim = l;		/* H(Y/Y_n), in [0, 1] */
		break;

	case 3:
		tristim = l - b;	/* H(Z/Z_n), in [-0.742, 1.737] */
		break;
	}

/* Rescale & return:  */
	tristim = (tristim + 1.0) / 3.0;	/* in [0.086, 0.9123] */

	g = RESTRICT (tristim, 0.0, 1.0);

	return QUANT (g, FUT_GRD_MAXVAL);
}


/*------------------------------------------------------------------
 *  ofun -- quantization call-back functions for FuT library
 *------------------------------------------------------------------
 */
static fut_otbldat_t
	ofun (fut_gtbldat_t q, KpGenericPtr_t dataP)
{
double		s, p;

	p = DEQUANT (q, FUT_GRD_MAXVAL);

	p = (3.0 * p) - 1.0;		/* [0, 1] -> [-1, 2] */
	p = RESTRICT (p, 0.0, 2.0);	/* leave headroom */
	p = H_inverse (p);			/* X/X_n, Y/Y_n, Z/Z_n */

	switch (((fData_p) dataP)->chan) {
	case 1:
		p *= XWHITE;	/* X */
		break;

	case 2:
		p *= YWHITE;	/* Y */
		break;

	case 3:
		p *= ZWHITE;	/* Z */
		break;
	}

     /* Adjust tristimulus values for encoding:  */
	p *= XYZSCALE;				/* rescale for InterColor encoding */
	s = RESTRICT (p, 0.0, 1.0);	/* clip to valid range [0, 1] */

	return QUANT (s, FUT_MAX_PEL12);
}

/*----------------------------------------------------------------------
 *  get_lab2xyz_fut --	construct an Lab to XYZ FuT FuT
 *						to establish a grid for composition

 		MUST be called from within a critical section!!!!!
 *----------------------------------------------------------------------
 */
fut_ptr_t
	get_lab2xyz_fut (KpInt32_t size)
{
fut_ptr_t		futp;
fut_itbl_ptr_t	itblx, itbly, itblz;
fut_gtbl_ptr_t	gtbl [3];
fut_otbl_ptr_t	otbl [3];
KpInt32_t		ichan, sizeArray[3], iomask;
fData_t			fData;

	iomask = FUT_IN(FUT_XYZ) | FUT_OUT(FUT_XYZ);

/* Compute shared input tables:  */
	fData.dimSize = size;
	fData.ySize = size;
	fData.zSize = size;

	itblx = fut_new_itblEx (size, Lifun, NULL);
	itbly = fut_new_itblEx (size, abifun, &fData);
	itblz = fut_new_itblEx (size, abifun, &fData);

/* Compute grid tables and output tables:  */
	sizeArray[0] = size;
	sizeArray[1] = fData.ySize;
	sizeArray[2] = fData.zSize;

	for (ichan = 0; ichan < 3; ichan++) {
	   fData.chan = ichan + 1;
	   gtbl [ichan] = fut_new_gtblEx (iomask, gfun, &fData, sizeArray);
	   otbl [ichan] = fut_new_otblEx (ofun, &fData);
	}

/* Assemble FuT:  */
	futp = fut_new (iomask, itblx, itbly, itblz,
							gtbl [0], otbl [0],
							gtbl [1], otbl [1],
							gtbl [2], otbl [2]);

	if ( futp->idstr != NULL ) {
		fut_free_idstr (futp->idstr);
		futp->idstr = NULL;
	}

	fut_free_tbls ((KpInt32_t FAR *)itblx, itbly, itblz,
							gtbl [0], gtbl [1], gtbl [2],
							otbl [0], otbl [1], otbl [2], NULL);

	return (futp);
}
