/*
 * @(#)lin4d.c	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: lin4d.c	1.18 98/02/17
 *
 *	Contains:	This module contains functions to create the Lin4d FuT.
 *
 *			Created by lsh, November 11, 1993
 *
 *********************************************************************
 *    COPYRIGHT (c) 1993-1998 Eastman Kodak Company
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *********************************************************************
 */


#include "kcptmgr.h"
#include "makefuts.h"


/*----------------------------------------------------------------------
 *  get_lin4d_fut --	construct an identity FuT
 *						to establish a grid for composition
 *----------------------------------------------------------------------
 */
fut_ptr_t
	get_lin4d_fut (KpInt32_t size)
{
fut_ptr_t		futp;
fut_itbl_ptr_t	itbl[4];
fut_gtbl_ptr_t	gtbl[4];
fut_otbl_ptr_t	otbl[4];
KpInt32_t		chan, iomask, sizeArray[4];

	sizeArray[0] = size;
	sizeArray[1] = size;
	sizeArray[2] = size;
	sizeArray[3] = size;

	/* Compute grid tables */
	iomask = FUT_IN(FUT_XYZT);

	gtbl[0] = fut_new_gtblEx (iomask, lin_gFunc_x, NULL, sizeArray);
	gtbl[1] = fut_new_gtblEx (iomask, lin_gFunc_y, NULL, sizeArray);
	gtbl[2] = fut_new_gtblEx (iomask, lin_gFunc_z, NULL, sizeArray);
	gtbl[3] = fut_new_gtblEx (iomask, lin_gFunc_t, NULL, sizeArray);

	/* Compute shared input and output tables:  */
	for (chan = 0; chan < 4; chan++) {
		itbl [chan] = fut_new_itblEx (sizeArray[chan], lin_iFunc, NULL);
		otbl [chan] = fut_new_otblEx (lin_oFunc, NULL);
	}

	/* Assemble FuT:  */
	iomask = FUT_IN(FUT_XYZT) | FUT_OUT(FUT_XYZT);
	futp = fut_new (iomask, itbl [0], itbl [1], itbl [2], itbl [3],
						gtbl [0], otbl [0],
						gtbl [1], otbl [1],
						gtbl [2], otbl [2],
						gtbl [3], otbl [3]);

	if ( futp->idstr != NULL ) {
		fut_free_idstr (futp->idstr);
		futp->idstr = NULL;
	}

	fut_free_tbls ((KpInt32_t FAR *) itbl [0], itbl [1], itbl [2], itbl [3],
						gtbl [0], gtbl [1], gtbl [2], gtbl [3],
						otbl [0], otbl [1], otbl [2], otbl [3], NULL);

	return (futp);
}
