/*
 * @(#)makefuts.h	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: makefuts.h	1.14 98/02/17
 *
 *	Contains:	header file for making ICM specific futs
 *
 *	Written by:	Poe
 *
 *	Change History (most recent first):
 *
 ****************************************************************
 *    COPYRIGHT (c) 1991-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */


#ifndef _MAKEFUTS_H_
#define _MAKEFUTS_H_ 1

#include <math.h>
#include "fut.h"
#include "fut_util.h"

/* These constants are used in logrgb.c and loguvl.c to shape shadows by
	adding `flare' */

#define	FLARE_YMIN		0.00392156862745	/* Y_min = 1/255, a la DM8000 */
#define  FLARE_YFLARE	0.00393700787402	/* Y_min/(1 + Y_min) */
#define	FLARE_DMAX		2.40654018043395	/* D_max = -(log Y_min) */
#define	FLARE_TANPT		0.01065992873906	/* eY_min */
#define	FLARE_MAPPT		0.18046425546277	/* (log e)/D_max */
#define	FLARE_SLOPE		16.9292178100213	/* MAPPT/TANPT */

/* This constant was define in the non-ansi c compilers math.h.  It does not
	seem to be defined by any of the compilers that we are using   					*/

#define M_LN10				2.30258509299404568402

	/*
		This scale factor is applied to the X, Y, and Z values prior to
		the 12-bit quantization so that, when shifted left 4 bits,
		it will produce an unsigned 16-bit number with one integer bit,
		as per the ColorSync open profile format.  It is *approximately*
		1/2:  the exact number is 2^15/(4080 << 4), since 1.0 is represented
		as 2^15 and FUT_MAX_PEL12 (= 4080) is the quantization scale factor.
	*/
#define	XYZSCALE	0.50196078431373	/* = 2^15 / (FUT_MAX_PEL12 << 4) */

	/* restricts t to interval [low, high] */
#define RESTRICT(t, low, high)	\
			(MAX ((low), MIN ((high), (t))))

	/* quantizes t to nearest (short) integer in given scale */
#define QUANT(t, scale)	\
			((short)((double)(scale) * RESTRICT((t), 0.0, 1.0) + 0.5))

	/* quantizes t to nearest integer in given scale */
	/* seperate the lines to fix floating point precision error on 68k */
#define QUANT1(t, scale)	\
			t = (double)RESTRICT((t), 0.0, 1.0); \
			t *= (double)scale; \
			t += (double)0.5

	/* (approximate) inverse of QUANT() */
#define DEQUANT(t, scale)	\
			((double)(t)/(double)(scale))


#define	POW(x, power)	( ((x) > 0.0) ? exp ((power) * log ((x))) \
			 	      : pow ((x), (power)) )

#define	ANTILOG(x)	( exp ((x) * M_LN10) )

	/* CIE visual-response function */
#define H(t)			\
	( ((t) <= 0.008856) ? 9.033 * (t) \
			    : 1.16 * POW((t), 1.0/3.0) - 0.16 )

	/* inverse of H() */
#define H_inverse(t)		\
	( ((t) <= 0.08) ? (t)/9.033 \
			      : POW(((t) + 0.16)/1.16, 3.0) )


/* function prototypes */
KpInt32_t make_lin3d ARGS((KpChar_p fname));
KpInt32_t make_lin4d ARGS((KpChar_p fname));
KpInt32_t make_lab2xyz ARGS((KpChar_p fname));
KpInt32_t make_xyz2lab ARGS((KpChar_p fname));
fut_ptr_t get_lin3d_fut ARGS((KpInt32_t));
fut_ptr_t get_lin4d_fut ARGS((KpInt32_t));
fut_ptr_t get_linlab_fut ARGS((KpInt32_t));
fut_ptr_t get_lab2xyz_fut ARGS((KpInt32_t));
fut_ptr_t get_xyz2lab_fut ARGS((KpInt32_t));
double lin_iFunc (double, KpGenericPtr_t);
fut_gtbldat_t lin_gFunc_x (double_p, KpGenericPtr_t);
fut_gtbldat_t lin_gFunc_y (double_p, KpGenericPtr_t);
fut_gtbldat_t lin_gFunc_z (double_p, KpGenericPtr_t);
fut_gtbldat_t lin_gFunc_t (double_p, KpGenericPtr_t);
fut_otbldat_t lin_oFunc (fut_gtbldat_t, KpGenericPtr_t);

#endif	/* _MAKEFUTS_H_ */

