/*
 * @(#)profilem.c	1.4 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: profilem.c	1.5 98/02/17
 *
 *	Contains:	forward and inverse monochrome PT via PTNewMonoPT
 *
 *	Written by:	Color Proccessor group
 *
 ****************************************************************
 *    COPYRIGHT (c) 1995-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */

#include "csmatrix.h"
#include "monopt.h"
#include "attrib.h"
#include "kcptmgr.h"

/*------------------------------------------------------------------------------
 *  PTNewMonoPT
 *------------------------------------------------------------------------------
 */

PTErr_t
	 PTNewMonoPT (	ResponseRecord_p	grayTRC,
					KpUInt32_t			gridsize,
					bool				invert,
					PTRefNum_p			thePTRefNumP)
{
PTErr_t	PTErr;
fut_ptr_t theFut;
KpChar_t inColorSpaceAttr[5], outColorSpaceAttr[5], compositionAttr[5];

/* Check for valid ptrs */
	if ((grayTRC == NULL) || (thePTRefNumP == NULL)) {
		PTErr = KCP_BAD_PTR;
		goto GetOut;
	}

	if (Kp_IsBadReadPtr(grayTRC, sizeof(*grayTRC)))	{
		PTErr = KCP_BAD_PTR;
		goto GetOut;
	}

	if (Kp_IsBadWritePtr(thePTRefNumP, sizeof(*thePTRefNumP))) {
		PTErr = KCP_BAD_PTR;
		goto GetOut;
	}

	/* pass the input arguments along to the fut maker */
	if (invert == False) {
		PTErr = makeForwardXformMono
						(grayTRC, gridsize, &theFut);

		/* setup the foward color space */
		KpItoa(KCM_MONO, inColorSpaceAttr);
		KpItoa(KCM_CIE_LAB, outColorSpaceAttr);
	}
	else {
		PTErr = makeInverseXformMono
						(grayTRC, gridsize, &theFut);

		/* setup the inverse color space */
		KpItoa(KCM_CIE_LAB, inColorSpaceAttr);
		KpItoa(KCM_MONO, outColorSpaceAttr);
	}

	if (PTErr != KCP_SUCCESS) {
		goto GetOut;
	}

	/* convert fut to a PT */
	PTErr = fut2PT (theFut, thePTRefNumP);
	if (PTErr != KCP_SUCCESS) {
		goto GetOut;
	}

	/* set the color space attributes */
	PTErr = PTSetAttribute (*thePTRefNumP, KCM_IN_SPACE, inColorSpaceAttr);
	if (PTErr != KCP_SUCCESS) {
		goto GetOut;
	}

	PTErr = PTSetAttribute (*thePTRefNumP, KCM_OUT_SPACE, outColorSpaceAttr);
	if (PTErr != KCP_SUCCESS) {
		goto GetOut;
	}

	/* set the composition state attribute */
	KpItoa(KCM_SINGLE, compositionAttr);
	PTErr = PTSetAttribute (*thePTRefNumP, KCM_COMPOSITION_STATE, compositionAttr);

GetOut:
	return (PTErr);
}
