/*
 * @(#)pteval.c	1.4 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: pteval.c	2.47 98/02/17
 *
 *	Contains:	Handle function evaluations using PTs
 *
 ****************************************************************
 *    COPYRIGHT (c) 1991-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */

#include "attrib.h"
#include "kcmptlib.h"
#include "kcptmgrd.h"
#include "kcptmgr.h"
#include "kcpmgru.h"
#include "fut_util.h"

#if defined (KCP_ACCEL)
#include "ctelib.h"

#define	FASTER_IN_SW 8000
#endif

static KpUInt32_t calcChanMask (KpInt32_t, KpInt32_t, PTCompDef_ptr_t);
static KpInt32_t calcChans (KpUInt32_t);
static KpUInt32_t getImageSizeAndStart (KpInt32_t, KpInt32_t, KpInt32_t, KpInt32_t, PTImgAddr_t, PTImgAddr_t*);
static PTErr_t setupEvalList (KpUInt32_t, PTTable_p*, PTEvalDTPB_p, KpUInt32_p);
static PTErr_t getDataBytes (KpInt32_t,	KpInt32_p);


PTErr_t
	 PTEval (	PTRefNum_t		PTRefNum,
				PTEvalPB_p		evalDef,
				PTEvalTypes_t	evalID,
				KpInt32_t		devNum,
				KpInt32_t		aSync,
				opRefNum_p		opRefNum,
				PTProgress_t	progress)
{
PTErr_t			PTErr;
callBack_t		callBack;

	callBack.progressFunc = progress;

	PTErr = PTEvalP (PTRefNum, evalDef, evalID, devNum, aSync, opRefNum, &callBack);

	return (PTErr);
}

PTErr_t
	 PTEvalP (	PTRefNum_t		PTRefNum,
				PTEvalPB_p		evalDef,
				PTEvalTypes_t	evalID,
				KpInt32_t		devNum,
				KpInt32_t		aSync,
				opRefNum_p		opRefNum,
				callBack_p		callBack)
{
PTEvalDTPB_t	evalDTDef;
PTCompDef_t 	imageInput[FUT_NICHAN];
PTCompDef_t 	imageOutput[FUT_NOCHAN];
KpInt32_t		i;
PTErr_t			PTErr;

	/* Check for valid PTEvalPB_p */
	if (Kp_IsBadReadPtr(evalDef, (KpUInt32_t)sizeof(*evalDef))) {
		return (KCP_BAD_PTR);
	}

	if (Kp_IsBadReadPtr(evalDef->input,
						evalDef->nInputs * (KpUInt32_t)sizeof(*evalDef->input))) {
		return (KCP_BAD_PTR);
	}

	if (Kp_IsBadReadPtr(evalDef->output,
					evalDef->nOutputs * (KpUInt32_t)sizeof(*evalDef->output))) {
		return (KCP_BAD_PTR);
	}

	evalDTDef.nPels = evalDef->nPels;
	evalDTDef.nLines = evalDef->nLines;
	evalDTDef.nInputs = evalDef->nInputs;
	evalDTDef.input = imageInput;
	evalDTDef.dataTypeI = KCM_UBYTE;

	if (evalDTDef.nInputs > FUT_NICHAN) {
		return (KCP_INVAL_IN_VAR);
	}

	for (i = 0; i < evalDTDef.nInputs; i++) {
			evalDTDef.input[i] = evalDef->input[i];
	}
	evalDTDef.nOutputs = evalDef->nOutputs;
	evalDTDef.output = imageOutput;
	evalDTDef.dataTypeO = KCM_UBYTE;
	for (i = 0; i < evalDTDef.nOutputs; i++) {
			evalDTDef.output[i] = evalDef->output[i];
	}

	PTErr = PTEvaluate (PTRefNum, &evalDTDef, evalID, devNum, aSync, opRefNum, callBack);

	return (PTErr);
}


PTErr_t
	PTEvalDT (	PTRefNum_t		PTRefNum,
				PTEvalDTPB_p	evalDef,
				PTEvalTypes_t	evalID,
				KpInt32_t		devNum,
				KpInt32_t		aSync,
				opRefNum_p		opRefNum,
				PTProgress_t	progress)
{
PTErr_t			PTErr;
callBack_t		callBack;

	callBack.progressFunc = progress;

	PTErr = PTEvaluate (PTRefNum, evalDef, evalID, devNum, aSync, opRefNum, &callBack);

	return (PTErr);
}


PTErr_t
	PTEvaluate (	PTRefNum_t		PTRefNum,
					PTEvalDTPB_p	evalDef,
					PTEvalTypes_t	evalID,
					KpInt32_t		devNum,
					KpInt32_t		aSync,
					opRefNum_p		opRefNum,
					callBack_p		callBack)
{
PTErr_t			errnum;
PTEvalDTPB_t	lEvalDef;
PTCompDef_t		thisInput[FUT_NICHAN], thisOutput[FUT_NOCHAN];
PTRefNum_t		PTList[MAX_PT_CHAIN_SIZE];
PTTable_p		evalList[MAX_PT_CHAIN_SIZE], PTTableP;
PTTable_p*		listStart;
KpInt32_t		theSerialCount, i1, i2, i3, i4, nOutputs, nFuts, PTcount, size;
KpUInt32_t		tempMemNeeded, oMask;
#if defined (KCP_ACCEL)
PTEvalTypes_t	evaluator;
KpInt32_t		numEvals;
#if defined (KCP_NEVER_DEFINED)
KpInt8_t		strInSpace[KCM_MAX_ATTRIB_VALUE_LENGTH+1];
KpInt32_t		attrSize, inspace;
#endif
#endif
PTImgAddr_t		addr, startAddr;

#if defined (KCP_MACPPC_MP)
	KCPInitializeMP ();
#endif

	errnum = getPTStatus (PTRefNum);			/* must be an active or serial PT */
	if ((errnum != KCP_PT_ACTIVE) && (errnum != KCP_SERIAL_PT)) {
		goto ErrOut1;
	}

#if defined (KCP_ACCEL)
	errnum = GetEval (evalID, &evaluator);		/* get an evaluator */
	if (errnum != KCP_SUCCESS) {
		goto ErrOut1;
	}
#endif

/* Check for valid callback address... */
	if ((callBack->progressFunc != NULL)
		&& (Kp_IsBadCodePtr((OsiProcPtr_t)callBack->progressFunc))) {

		errnum = KCP_BAD_CALLBACK_ADDR;
		goto ErrOut1;
	}

/*  and valid opRefNum address */
	if (Kp_IsBadWritePtr (opRefNum, (KpUInt32_t)sizeof(*opRefNum))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut1;
	}

/*  and valid PTEvalDTPB_p */
	if (Kp_IsBadReadPtr(evalDef, (KpUInt32_t)sizeof(*evalDef))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut1;
	}

	if (Kp_IsBadReadPtr(evalDef->input, evalDef->nInputs * (KpUInt32_t)sizeof(*evalDef->input))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut1;
	}

	if (Kp_IsBadReadPtr(evalDef->output, evalDef->nOutputs * (KpUInt32_t)sizeof(*evalDef->output))) {
		errnum = KCP_BAD_PTR;
		goto ErrOut1;
	}

	/* set up the local evaluation structures */
	/* set input and output to NULL */
	/* this preserves the channel position while allowing */
	/* the number of channels to indicate the number of valid addresses */
	/* this is needed to keep both the CTE and SW evaluations happy */
	for (i1 = 0; i1 < FUT_NICHAN; i1++) {
		thisInput[i1].pelStride = 0;
		thisInput[i1].lineStride = 0;
		thisInput[i1].addr = NULL;
	}

	lEvalDef.nPels = evalDef->nPels;
	lEvalDef.nLines = evalDef->nLines;

	lEvalDef.nInputs = FUT_NICHAN;
	lEvalDef.dataTypeI = evalDef->dataTypeI;
	lEvalDef.input = thisInput;
	for (i1 = 0; i1 < evalDef->nInputs; i1++) {
		lEvalDef.input[i1].pelStride = evalDef->input[i1].pelStride;
		lEvalDef.input[i1].lineStride = evalDef->input[i1].lineStride;
		addr = evalDef->input[i1].addr;
		lEvalDef.input[i1].addr = addr;
		if (addr != NULL) {	/* Check all of the data input addresses for validity */
			size = getImageSizeAndStart (lEvalDef.nPels, lEvalDef.input[i1].pelStride,
										lEvalDef.nLines, lEvalDef.input[i1].lineStride,
										addr, &startAddr);

			if (Kp_IsBadHugeReadPtr (startAddr, size)) {
				errnum = KCP_BAD_PTR;
				goto ErrOut1;
			}
		}
	}

	/* output addresses are loaded in the evaluation loop */
	/* so just check addresses here */
	for (i1 = 0; i1 < evalDef->nOutputs; i1++) {
		addr = evalDef->output[i1].addr;
		if (addr != NULL) {				/* check addresses for validity */
			size = getImageSizeAndStart (lEvalDef.nPels, evalDef->output[i1].pelStride,
										lEvalDef.nLines, evalDef->output[i1].lineStride,
										addr, &startAddr);

			if (Kp_IsBadHugeWritePtr (startAddr, size)) {
				errnum = KCP_BAD_PTR;
				goto ErrOut1;
			}
		}
	}

	*opRefNum = 1;	/* use constant, since everything is synchronous now */

	/* clear the lists, just for clarity */
	for (i1 = 0; i1 < MAX_PT_CHAIN_SIZE; i1++) {
		PTList[i1] = NULL;
		evalList[i1] = NULL;
	}

	/* get the list of PTs which we must actually evaluate */
	errnum = resolvePTData (PTRefNum, &theSerialCount, PTList);

	/* set up list of futs through which the image is evaluated */
	for (i1 = 0; i1 < theSerialCount; i1++) {
		PTTableP = lockPTTable (PTList[i1]);	/* lock tables while evaluating */
		evalList[i1] = PTTableP;
		PTTableP->futP = NULL;					/* clear evaluation info */
		PTTableP->evalIOMask = 0;
		PTTableP->evalDataTypeI = 0;
		PTTableP->evalDataTypeO = 0;
	}

	/* initialize the evaluation list */
	errnum = setupEvalList (theSerialCount, evalList, evalDef, &tempMemNeeded);
	if (errnum != KCP_SUCCESS) {
		goto ErrOut2;
	}

	/* if temporary memory is not needed, */
	/* then evaluate a full image at a time until */
	/* the image has been processed through all futs */
	/* if temporary memory is needed, */
	/* then this level processes the image just once */
	/* and a lower level processes the image through all futs */
	if (tempMemNeeded == 0) {
		PTcount = theSerialCount;
	}
	else {
		PTcount = 1;
	}

	initProgressPasses (PTcount, callBack);

	/* process the image through each fut in the list */
	for (i1 = 0; i1 < PTcount; i1++) {

		/* set up the output data addresses */
		/* use io mask to order output channels properly */
		if (tempMemNeeded == 1) {
			nFuts = theSerialCount;								/* this many futs to evaluate */
			listStart = &evalList[0];
			oMask = FUT_OMASK(evalList[nFuts-1]->evalIOMask);	/* use mask of last fut */
		}
		else {
			nFuts = 1;											/* evaluate one fut */
			listStart = &evalList[i1];
			oMask = FUT_OMASK(evalList[i1]->evalIOMask);		/* use mask of this fut */
		}

		/* initialize the output data structures */
		lEvalDef.nOutputs = FUT_NOCHAN;
		lEvalDef.dataTypeO = evalDef->dataTypeO;
		lEvalDef.output = thisOutput;

		for (i2 = 0; i2 < FUT_NOCHAN; i2++) {
			thisOutput[i2].pelStride = 0;
			thisOutput[i2].lineStride = 0;
			thisOutput[i2].addr = NULL;
		}

		/* set the output channel addresses */
		if (i1 == (PTcount -1)) {		/* last, just use supplied stuff */
			for (i2 = 0, nOutputs = 0; i2 < evalDef->nOutputs; i2++) {
				lEvalDef.output[i2].pelStride = evalDef->output[i2].pelStride;
				lEvalDef.output[i2].lineStride = evalDef->output[i2].lineStride;
				lEvalDef.output[i2].addr = evalDef->output[i2].addr;

				nOutputs++;		/* count actual outputs */
			}

			getDataBytes (evalDef->dataTypeO, &i3);	/* get output data size */
			if (i3 == 0) {
				nOutputs = 3;
			}
		}
		else {
			for (i2 = oMask, i3 = 0, i4 = 0, nOutputs = 0; i2 != 0; i2 >>= 1, i3++) {
				if ((i2 & 1) == 1) {	/* this output channel is needed */
					while ((addr = evalDef->output[i4].addr) == NULL) {
						i4++;		/* get next available output channel */
					}

					if (i4 > evalDef->nOutputs) {
						errnum = KCP_PTERR_4;				/* programming error */
						goto ErrOut2;
					}

					lEvalDef.output[i3].pelStride = evalDef->output[i4].pelStride;
					lEvalDef.output[i3].lineStride = evalDef->output[i4].lineStride;
					lEvalDef.output[i3].addr = addr;

					i4++;			/* next output address */
					nOutputs++;		/* count actual outputs */
				}
			}
		}

#if defined (KCP_ACCEL)
		/* if there are less than FASTER_IN_SW evaluations, it's faster to do it in software. */
		numEvals = lEvalDef.nPels * lEvalDef.nLines * nOutputs;

		if ((numEvals < FASTER_IN_SW) || (tempMemNeeded == 1)) {
			evaluator = KCP_EVAL_SW;
		}

		switch (evaluator) {
		case KCP_EVAL_SW:		/* evaluate in software */

software_evaluation:
#endif

			errnum = PTEvalSeq (nFuts, listStart, &lEvalDef, callBack);
			if (errnum != KCP_SUCCESS) {
				goto ErrOut2;
			}

#if defined (KCP_ACCEL)
			break;

		case KCP_EVAL_CTE:	/* evaluate using the NFE */
		{
		PTRefNum_t		thePTRefNum;
		KpHandle_t		PTData;

			thePTRefNum = listStart[0]->refNum;			/* get the PT reference number */

			PTData = getPTData (thePTRefNum);	/* get the transform data */

			errnum = PT_eval_cteDT (PTData, thePTRefNum, &lEvalDef, devNum, aSync, callBack);

			if ((errnum == KCP_CTE_GRID_TOO_BIG) || (errnum == KCP_CTE_NOT_ATTEMPTED)) {
				goto software_evaluation;
			}

			break;
		}

		default:
			errnum = KCP_INVAL_EVAL;

			break;
		}
#else
		aSync += 0;  /* this is to humor the PC 32-bit compiler  */
		devNum += 0; /* these are unreferenced formal parameters */
		evalID += 0;
#endif

		/* output for this PT is input for next PT */
		lEvalDef.nInputs = lEvalDef.nOutputs;
		lEvalDef.dataTypeI = lEvalDef.dataTypeO;

		for (i2 = 0; i2 < lEvalDef.nInputs; i2++) {
			lEvalDef.input[i2].addr = lEvalDef.output[i2].addr;
			lEvalDef.input[i2].pelStride = lEvalDef.output[i2].pelStride;
			lEvalDef.input[i2].lineStride = lEvalDef.output[i2].lineStride;
		}
	}


ErrOut2:
	/* unlock the PTs used for evaluation */
	for (i1 = 0; i1 < theSerialCount; i1++) {
		unlockPTTable (PTList[i1]);
	}

ErrOut1:
	return (errnum);
}


/* set up the fut evaluation list */
/* Determine whether or not temporary memory is needed */
static PTErr_t
	setupEvalList (	KpUInt32_t		numPTs,
					PTTable_p*		evalList,
					PTEvalDTPB_p	evalDef,
					KpUInt32_p		tempMemNeeded)
{
PTErr_t		errnum = KCP_SUCCESS;
KpInt32_t	i1, i2;
KpInt32_t	finalOutputs, nOutputs, maxOutputs;
KpUInt32_t	thisOmask, thisImask;
KpInt32_t	sizeInData, sizeOutData;
fut_chan_ptr_t	futChan;

	thisOmask = calcChanMask (evalDef->dataTypeO, evalDef->nOutputs, evalDef->output);	/* calculate final output mask */
	finalOutputs = calcChans (thisOmask);		/* and number of output channels */

	/* does the fut have those outputs? */
	if ((FFUTP(evalList[numPTs -1]->data)->iomask.out & thisOmask) != thisOmask) {
		errnum = KCP_INVAL_EVAL;
		goto ErrOut;
	}

	nOutputs = finalOutputs;	/* # outputs of fut at the end of the list */
	maxOutputs = 0;				/* none yet */

	for (i1 = (numPTs -1); i1 >= 0; i1--) {
		if (nOutputs > maxOutputs) {		/* remember max # outputs needed */
			maxOutputs = nOutputs;
		}

		/* get inputs required for this fut as determined by outputs required from this fut */
		thisImask = 0;

		for (i2 = 0; i2 < FUT_NOCHAN; i2++) {
			if (thisOmask & FUT_BIT(i2)) {				/* is this output channel needed? */

				futChan = FCHANP(FFUTP(evalList[i1]->data)->chanHandle[i2]);	/* get pointer to chan */

				thisImask |= (KpUInt32_t)FUT_CHAN_IMASK(futChan);	/* include inputs required for this chan */
			}
		}

		evalList[i1]->evalIOMask = FUT_OUT(thisOmask) | FUT_IN(thisImask);	/* store the evaluation I/O mask */

		/* set up outputs for preceeding fut in the list */
		thisOmask = thisImask;				/* which are inputs to this fut */
		nOutputs = calcChans (thisOmask);	/* get number of output channels */
	}

	thisImask = calcChanMask (evalDef->dataTypeI, evalDef->nInputs, evalDef->input);	/* calculate input mask for the first fut */

	/* does the first fut have the required inputs? */
	if ((FFUTP(evalList[0]->data)->iomask.in & thisImask) != thisImask) {
		errnum = KCP_INVAL_EVAL;
		goto ErrOut;
	}

	/* Is temporary memory required? */
	/* If just one PT, then temporary data memory is not needed */
	/* if more than one PT (serial evaluation), then the number and size of the outputs */
	/* determines whether or not temporary memory is needed */

	errnum = getDataBytes (evalDef->dataTypeI, &sizeInData);	/* get input data size */
	if (errnum != KCP_SUCCESS) {
		goto ErrOut;
	}

	errnum = getDataBytes (evalDef->dataTypeO, &sizeOutData);	/* get output data size */
	if (errnum != KCP_SUCCESS) {
		goto ErrOut;
	}

	if (numPTs == 1) {
		(*tempMemNeeded) = 0;		/* not serial evaluation, no temporary data memory needed */
	}
	else {						/* serial evaluation */
								/* does this require temporary memory? */
		if ((maxOutputs > finalOutputs) || (sizeInData > sizeOutData)) {
			(*tempMemNeeded) = 1;				/* yes */
		}
		else {
			(*tempMemNeeded) = 0;				/* no */
		}
	}

ErrOut:
	return errnum;
}


/* calculate channel usage mask */
KpUInt32_t
	calcChanMask (	KpInt32_t dataType,
					KpInt32_t nAddr,
					PTCompDef_ptr_t	addrList)
{
KpUInt32_t	chanMask;
KpInt32_t	i1;

	chanMask = 0;

	if (   (dataType == KCM_USHORT_555)
		|| (dataType == KCM_USHORT_565)
		|| (dataType == KCM_R10G10B10) ) {	/* single address, multiple channels */
		if (addrList[0].addr != NULL) {
			chanMask = FUT_BIT(0) | FUT_BIT(1) | FUT_BIT(2);
		}
	}
	else {									/* discrete channels */
		for (i1 = 0; i1 < nAddr; i1++) {
			if (addrList[i1].addr != NULL) {
				chanMask |= FUT_BIT(i1);
			}
		}
	}

	return chanMask;
}


/* calculate number of channels active in a channel usage mask */
KpInt32_t
	calcChans (	KpUInt32_t	mask)
{
KpInt32_t	nChans, i1;

	for (i1 = mask, nChans = 0; i1 != 0; i1 >>= 1) {	/* count total addresses needed */
		if ((i1 & 1) == 1) {
			nChans++;
		}
	}

	return nChans;
}


/* returns the # of bytes per datum for a given data type */
static PTErr_t
	getDataBytes (	KpInt32_t	dataType,
					KpInt32_p	dataTypeSize)
{
	switch (dataType) {
	case KCM_USHORT_555:
	case KCM_USHORT_565:
	case KCM_R10G10B10:
		(*dataTypeSize) = 0;	/* less than 1 byte */
		break;

	case KCM_UBYTE:
		(*dataTypeSize) = 1;	/* 1 byte per datum */
		break;

	case KCM_USHORT_12:
	case KCM_USHORT:
		(*dataTypeSize) = 2;	/* 2 bytes per datum */
		break;

	default:
		return KCP_INVAL_DATA_TYPE;
	}

	return (KCP_SUCCESS);
}


/* calculate the exact bytes required for a given image */
static KpUInt32_t
	getImageSizeAndStart (	KpInt32_t		nPels,
							KpInt32_t		pelStride,
							KpInt32_t		nLines,
							KpInt32_t		lineStride,
							PTImgAddr_t		addr,
							PTImgAddr_t*	startAddr)
{
KpInt32_t size;

	if (nPels == 1) {
		pelStride = 1;
	}

	if (nLines == 1) {
		lineStride = 1;
	}

	size = ((nLines -1) * lineStride) + ((nPels -1) * pelStride) +1;

	if (size < 0) {
		size = -size;
		*startAddr = (KpUInt8_p)addr - size;
	}
	else {
		*startAddr = addr;
	}

	return size;
}


#if defined(KCP_ACCEL)
PTErr_t
	PTEvalRdy (	opRefNum_t	opRefNum,
				int32_p		progress)
{
	if (opRefNum || progress) {}

	return KCP_SUCCESS;
}


PTErr_t
	PTEvalCancel (	opRefNum_t opRefNum)
{
	if (opRefNum) {}

	return KCP_INVAL_OPREFNUM;
}
#endif

