/*
 * @(#)pttmgr.c	1.4 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: pttmgr.c	2.26 98/02/25
 *
 *	Contains:	PT table management
 *
 ****************************************************************
 *    COPYRIGHT (c) 1991-1998 Eastman Kodak Company.
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *****************************************************************
 */


#if defined (KPMAC)
	#include <Packages.h>
	#include <Memory.h>
#endif

#include "kcms_sys.h"
#include "kcmptlib.h"
#include "kcptmgrd.h"
#include "kcptmgr.h"
#include "bytes.h"

#if defined KPMAC
#define PTTP(handle) (*(PTTable_p FAR*)handle)
#else
#define PTTP(handle) ((PTTable_p)handle)
#endif


/* prototypes */
static void deletePTTable ARGS((PTRefNum_t));
static PTErr_t freeSerialData (PTRefNum_t);


/***********************************************************************/
/* get the pointer to a PT table */
PTTable_p
	lockPTTable (PTRefNum_t PTRefNum)
{
PTTable_p	PTTableP = NULL;

	if (checkPT (PTRefNum) == KCP_SUCCESS) {

		PTTableP = (PTTable_p)lockBuffer ((KpHandle_t)PTRefNum);	/* get the table pointer */

		if (PTTableP != NULL) {
			if ((PTTableP->uid != PTT_UID) || (PTTableP->refNum != PTRefNum)) {
				unlockBuffer ((KpHandle_t)PTRefNum);
				PTTableP = NULL;
			}
		}
	}

	return PTTableP;
}


/***********************************************************************/
/* unlock a PT table */
void
	unlockPTTable (PTRefNum_t PTRefNum)
{
	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return;
	}

	unlockBuffer ((KpHandle_t)PTRefNum);
}


/***********************************************************************/
/* delete the given PT table */
static void
	deletePTTable (	PTRefNum_t	PTRefNum)
{
PTTable_p	PTTableP;
int32_p		attrTagCountP;

	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return;
	}

	PTTableP = (PTTable_p)lockBuffer ((KpHandle_t)PTRefNum);	/* get the table pointer */

	if (PTTableP->attrBase != NULL) {	/* free space allocated to this PTs' attributes */
		attrTagCountP = lockBuffer (PTTableP->attrBase);
		if (*attrTagCountP != 0) {
			freeAttributes(PTTableP->attrBase);
		}
		freeBuffer (PTTableP->attrBase);
	}

	bzero (PTTableP, sizeof (PTTable_t));	/* just being careful */

	unlockPTTable (PTRefNum);

	freeBuffer ((KpHandle_t)PTRefNum);	/* free space allocated to this table */
}


/**************************************************************************/
/* register a PT by setting its implicit attributes and entering its info */
/* into the PT table */
PTErr_t
	registerPT (	KpHandle_t	hdr,
					KpHandle_t	attrBase,
					PTRefNum_p	PTRefNumP)
{
PTRefNum_t	PTRefNum;
PTTable_p	PTTableP;
PTErr_t		PTErr;
KpInt32_t	i;

	PTRefNum = (PTRefNum_t)allocBufferHandle (sizeof (PTTable_t));	/* allocate new PT and get a handle to it */

	if ((PTRefNum == NULL) || (PTMemTest () == 0)) {
		freeBuffer (PTRefNum);		/* in case PTMemTest() failed */
		PTErr = KCP_NO_CHECKIN_MEM;
	}
	else {
		PTTableP = (PTTable_p)lockBuffer ((KpHandle_t)PTRefNum);	/* get the table pointer */

		bzero (PTTableP, sizeof (PTTable_t));	/* just being careful */

		PTTableP->uid = PTT_UID;				/* insert unique id */
		PTTableP->refNum = PTRefNum;			/* save reference number */
		PTTableP->hdr = hdr;					/* save header address */
		PTTableP->attrBase = attrBase;			/* save attribute address */
		PTTableP->data = NULL;					/* show there's no PT data */
		PTTableP->checkInFlag = IS_CHECKED_IN;	/* show this pt is checked in */
		PTTableP->serialPTflag = NOT_SERIAL_PT;	/* show this pt not a serial PT */
		PTTableP->inUseCount = 0;				/* set in use count */
		PTTableP->serialCount = 0;				/* no serial PTs yet */

		for (i = 0; i < MAX_PT_CHAIN_SIZE; i++) {
			PTTableP->serialDef[i] = NULL;
		}

		nullEvalState (PTTableP);				/* set evaluation state to null */

		unlockPTTable (PTRefNum);

		PTErr = KCP_SUCCESS;

		if (hdr != NULL) {
			PTErr = TpSetImplicitAttr (PTRefNum);	/* set attributes using header info */
			if (PTErr != KCP_SUCCESS) {
				deletePTTable (PTRefNum);
				PTRefNum = NULL;
			}
		}
	}

	*PTRefNumP = PTRefNum;

	return (PTErr);
}


/***********************************************************************/
/* get a PT table entry and load it with given serial PT info */
PTErr_t
	addSerialPT (	PTRefNum_t		CurrentPTRefNum,
					PTRefNum_t		InputPTRefNum,
					PTRefNum_p		ResultPTRefNum)
{
PTErr_t		PTErr;
KpInt32_t	newSerialCount, inputSerialCount, i1;

	if ((PTErr = checkPT (CurrentPTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	if ((PTErr = checkPT (InputPTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	PTErr = registerPT (NULL, NULL, ResultPTRefNum);
	if (PTErr != KCP_SUCCESS) {
		return (PTErr);
	}

	makeSerial (*ResultPTRefNum);

	newSerialCount = 0;

	/* move in the "current" PT */
	if (PTTP(CurrentPTRefNum)->serialCount == 0) {
		if (PTTP(CurrentPTRefNum)->data != NULL) {
			PTTP(*ResultPTRefNum)->serialDef[0] = CurrentPTRefNum;
			makeActive (CurrentPTRefNum, PTTP(CurrentPTRefNum)->data);
			newSerialCount = 1;
		}
	}
	else {
		newSerialCount = PTTP(CurrentPTRefNum)->serialCount;
		if (newSerialCount >= MAX_PT_CHAIN_SIZE) {
			PTErr = KCP_EXCESS_PTCHAIN;
			goto ErrOut;
		}
		for (i1 = 0; i1 < newSerialCount; i1++) {
			PTTP(*ResultPTRefNum)->serialDef[i1] = PTTP(CurrentPTRefNum)->serialDef[i1];
			makeActive (PTTP(*ResultPTRefNum)->serialDef[i1], NULL);
		}
	}

	/* now add in the "input" PT */
	if (PTTP(InputPTRefNum)->serialCount == 0) {
		if (PTTP(InputPTRefNum)->data != NULL) {
			if (newSerialCount >= MAX_PT_CHAIN_SIZE) {
				PTErr = KCP_EXCESS_PTCHAIN;
				goto ErrOut;
			}
			PTTP(*ResultPTRefNum)->serialDef[newSerialCount] = InputPTRefNum;
			makeActive (InputPTRefNum, PTTP(InputPTRefNum)->data);
			newSerialCount++;
		}
	}
	else {
		if ((newSerialCount+PTTP(CurrentPTRefNum)->serialCount) > MAX_PT_CHAIN_SIZE) {
			PTErr = KCP_EXCESS_PTCHAIN;
			goto ErrOut;
		}
		inputSerialCount = PTTP(InputPTRefNum)->serialCount;
		for (i1 = 0; i1 < inputSerialCount; i1++) {
			PTTP(*ResultPTRefNum)->serialDef[i1] = PTTP(CurrentPTRefNum)->serialDef[newSerialCount];
			makeActive (PTTP(*ResultPTRefNum)->serialDef[newSerialCount], NULL);
			newSerialCount++;
		}
	}

	PTTP(*ResultPTRefNum)->serialCount = newSerialCount;

ErrOut:
	return (PTErr);
}


/***********************************************************************/
/* make the given PT active */
void
	makeActive (PTRefNum_t PTRefNum,
				KpHandle_t PTData)
{
	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return;
	}

	if (PTTP(PTRefNum)->data == NULL) {
		PTTP(PTRefNum)->data = PTData;		/* define PT address if not already */
		PTTP(PTRefNum)->inUseCount = 1;		/* used once */
	}
	else {
		PTTP(PTRefNum)->inUseCount++;			/* count the PT */
	}
}


/***********************************************************************/
/* make the given PT active */
void
	makeSerial (PTRefNum_t		PTRefNum)
{
	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return;
	}

	PTTP(PTRefNum)->serialPTflag = IS_SERIAL_PT;	/* show this is a serial PT */
}


/***********************************************************************/
/* make the given PT inactive */
PTErr_t makeInActive (PTRefNum_t PTRefNum)
{
KpHandle_t	PTData;
PTErr_t		PTErr;

	if ((PTErr = checkPT (PTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	PTErr = freeSerialData ( PTRefNum);	/* free any serial data */

	if (PTTP(PTRefNum)->data != NULL) {
		PTTP(PTRefNum)->inUseCount--;
		if (PTTP(PTRefNum)->inUseCount == 0) {

			PTData = PTTP(PTRefNum)->data;
			PTTP(PTRefNum)->data = NULL;
			PTErr = TpFreeData (PTData);

			freeEvalState (PTRefNum);			/* free the evaluation state memory */

			if (PTTP(PTRefNum)->checkInFlag == NOT_CHECKED_IN) {
	 			deletePTTable (PTRefNum);
			}
		}
	}

	return (PTErr);
}


/***********************************************************************/
/* release the serial data of the given PT */
static PTErr_t
	freeSerialData (	PTRefNum_t  PTRefNum)
{
PTErr_t		PTErr = KCP_SUCCESS;
KpInt32_t	i, theCount;
PTRefNum_t	theRefNum;

	if ((PTErr = checkPT (PTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	theCount = PTTP(PTRefNum)->serialCount;
	PTTP(PTRefNum)->serialCount = 0;
	PTTP(PTRefNum)->serialPTflag = NOT_SERIAL_PT;		/* show this pt not a serial PT */

	for (i = 0; i < theCount; i++) {
		theRefNum = PTTP(PTRefNum)->serialDef[i];
		PTTP(PTRefNum)->serialDef[i] = NULL;
		PTErr = makeInActive (theRefNum);		/* deactivate the PT */
	}

	return (PTErr);
}


/***********************************************************************/
/* make the given PT inactive */
PTErr_t makeCheckedOut (PTRefNum_t PTRefNum)
{
KpHandle_t	PTHdr, PTAttrBase;
KpUInt32_t	PTinUseCount;
PTErr_t		PTErr = KCP_SUCCESS, errnum1 = KCP_SUCCESS;

	if ((PTErr = checkPT (PTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	if (PTTP(PTRefNum)->checkInFlag == NOT_CHECKED_IN) {
		PTErr = KCP_NOT_CHECKED_IN;
	}
	else {
		PTHdr = PTTP(PTRefNum)->hdr;
		PTTP(PTRefNum)->hdr = NULL;
		PTAttrBase = PTTP(PTRefNum)->attrBase;
		PTinUseCount = PTTP(PTRefNum)->inUseCount;
		PTTP(PTRefNum)->checkInFlag = NOT_CHECKED_IN;

		PTErr = TpFreeHdr (PTHdr);		/* free the header */
		errnum1 = freeAttributes(PTAttrBase);			/* free the attributes */

		if (PTinUseCount == 0) {
 			deletePTTable (PTRefNum);
		}
	}

	if (PTErr == KCP_SUCCESS) {
		PTErr = errnum1;
	}

	return (PTErr);
}


/***********************************************************************/
/* get the status for a given PT reference number */
PTErr_t getPTStatus (PTRefNum_t PTRefNum)
{
PTErr_t 	PTErr;

	if ((PTErr = checkPT (PTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	if (PTTP(PTRefNum)->serialPTflag == IS_SERIAL_PT) {
		PTErr = KCP_SERIAL_PT;			/* must be a serial pt */
	}
	else {
		if (PTTP(PTRefNum)->checkInFlag == NOT_CHECKED_IN) {
			PTErr = KCP_PT_INVISIBLE;		/* must be a serial component */
		}
		else {
			if (PTTP(PTRefNum)->inUseCount == 0) {
				PTErr = KCP_PT_INACTIVE;		/* not active */
			}
			else {
				PTErr = KCP_PT_ACTIVE;			/* active */
			}
		}
	}

	return (PTErr);
}


/***********************************************************************/
/* get the PT header address for a given PT reference number */
KpHandle_t getPTHdr (PTRefNum_t PTRefNum)
{
KpHandle_t hdrHandle;

	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return NULL;
	}

	hdrHandle = PTTP(PTRefNum)->hdr;

	return (hdrHandle);
}

/***********************************************************************/
/* get the PT header address for a given PT reference number */
PTErr_t setPTHdr (PTRefNum_t PTRefNum, KpHandle_t hdrHandle)
{
PTErr_t		PTErr = KCP_BAD_PTR;

	if ((PTErr = checkPT (PTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	PTTP(PTRefNum)->hdr = hdrHandle;				/* set header */
	PTErr = TpSetImplicitAttr (PTRefNum);	/* set attributes using header info */

	return (PTErr);
}


/***********************************************************************/
/* get the attribute base for a given PT reference number */
KpHandle_t getPTAttr (PTRefNum_t PTRefNum)
{
KpHandle_t attrHandle;

	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return NULL;
	}

	attrHandle = PTTP(PTRefNum)->attrBase;

	return (attrHandle);
}


/***********************************************************************/
/* set the attribute base for a given PT reference number */
void setPTAttr (PTRefNum_t PTRefNum, KpHandle_t attrBase)
{
	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return;
	}

	PTTP(PTRefNum)->attrBase = attrBase;
}


/***********************************************************************/
/* get the PT data location for a given PT reference number */
KpHandle_t getPTData (PTRefNum_t PTRefNum)
{
KpHandle_t dataHandle;

	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return NULL;
	}

	dataHandle = PTTP(PTRefNum)->data;

	return (dataHandle);
}


/***********************************************************************/
/* get the PT data location for a given PT reference number */
PTErr_t resolvePTData (		PTRefNum_t PTRefNum,
							KpInt32_p SerialCount,
							PTRefNum_p SerialRefNum)
{
KpInt32_t	i;
PTErr_t		PTErr;

	if ((PTErr = checkPT (PTRefNum)) != KCP_SUCCESS) {
		return PTErr;
	}

	if ((PTTP(PTRefNum)->serialPTflag == IS_SERIAL_PT) && (PTTP(PTRefNum)->serialCount > 0)) {
		*SerialCount = PTTP(PTRefNum)->serialCount;
		for (i = 0; i < PTTP(PTRefNum)->serialCount; i++) {
			SerialRefNum[i] = PTTP(PTRefNum)->serialDef[i];
		}
	}
	else {
		*SerialCount = 1;
		SerialRefNum[0] = PTRefNum;
	}

	PTErr = KCP_SUCCESS;

	return (PTErr);
}


/***********************************************************************/
/* get the size in bytes of the header and attribute area */
KpInt32_t getCheckInSize (PTRefNum_t PTRefNum)
{
KpInt32_t size;

	if (checkPT (PTRefNum) != KCP_SUCCESS) {
		return 0;
	}

	size = ((KpInt32_t)sizeof (PT_t) + getAttrSize (PTTP(PTRefNum)->attrBase));

	return (size);
}


PTErr_t
	checkPT (PTRefNum_t PTRefNum)
{
	if (PTRefNum == NULL) {
		return KCP_NOT_CHECKED_IN;
	}

	if (Kp_IsBadWritePtr (PTTP(PTRefNum), (KpUInt32_t)sizeof(PTTable_t))) {
		return KCP_NOT_CHECKED_IN;
	}


	if ((PTTP(PTRefNum)->uid == PTT_UID) && (PTTP(PTRefNum)->refNum == PTRefNum)){
		return KCP_SUCCESS;
	}

	return KCP_INVAL_REFNUM;
}
