/*
 * @(#)xyz2lab.c	1.5 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * File: xyz2lab.c	1.13 98/02/17
 *
 *	Contains:	This module contains functions to create
 * 			an XYZ->lab FuT.
 *
 *				Created by lsh, November 11, 1993
 *
 *********************************************************************
 *    COPYRIGHT (c) 1992-1998 Eastman Kodak Company
 *    As an unpublished work pursuant to Title 17 of the United
 *    States Code.  All rights reserved.
 *********************************************************************
 */

#include "kcptmgr.h"
#include "makefuts.h"

/******
#define	D65	1
******/
#undef	D65

#ifdef	D65				/* white point for CIELAB */
#define	XWHITE	0.950
#define	YWHITE	1.000
#define ZWHITE	1.089
#else	/* D50 */
#define	XWHITE	0.9642
#define	YWHITE	1.0000
#define	ZWHITE	0.8249
#endif

#define	AB_UNSIGNED	1		/* use unsigned rep for a* and b* */


typedef struct fData_s {
	KpInt32_t	channel, xSize, ySize, zSize;
} fData_t, FAR* fData_p;


/*---------------------------------------------------------------------
 *  xfun, yfun, zfun -- input mappings
 *---------------------------------------------------------------------
 */
static double xfun (double x, KpGenericPtr_t dataP)
{
KpInt32_t	xSize;

	xSize = ((fData_p) dataP)->xSize;

	x /= (XYZSCALE * XWHITE);
	x = H (x);
	x *= (double)(xSize - 2) / (double)(xSize - 1);
	return RESTRICT (x, 0.0, 1.0);
}

static double yfun (double y, KpGenericPtr_t dataP)
{
KpInt32_t	ySize;

	ySize = ((fData_p) dataP)->ySize;

	y /= (XYZSCALE * YWHITE);
	y = H (y);
	y *= (double)(ySize - 2) / (double)(ySize - 1);
	return RESTRICT (y, 0.0, 1.0);
}

static double zfun (double z, KpGenericPtr_t dataP)
{
KpInt32_t	zSize;

	zSize = ((fData_p) dataP)->zSize;

	z /= (XYZSCALE * ZWHITE);
	z = H (z);
	z *= (double)(zSize - 2) / (double)(zSize - 1);
	return RESTRICT (z, 0.0, 1.0);
}


/*---------------------------------------------------------------------
 *  gfun -- grid-table functions:
 *				(H(X/X_n), H(Y/Y_n), H(Z/Z_n)) --> (L*, a*, b*)
 *---------------------------------------------------------------------
 */
static fut_gtbldat_t gfun (double *args, KpGenericPtr_t dataP)
{
double	p = 0;
double	tempd[3];
double	x, y, z;
KpInt32_t	xSize, ySize, zSize;

	x = args[0];
	y = args[1];
	z = args[2];

	xSize = ((fData_p) dataP)->xSize;
	ySize = ((fData_p) dataP)->ySize;
	zSize = ((fData_p) dataP)->zSize;

/* Convert to CIELAB, with L* in [0, 100], a* & b* in [-200, 200]:  */
	switch (((fData_p) dataP)->channel) {
    case 1:	/* (L*)/100 */
		tempd[0] = (double)(ySize - 1);
		tempd[1] = (double)(ySize - 2);
		tempd[2] = tempd[0] / tempd[1];
		tempd[2] *= y ;
		p = tempd[2];
		break;

   case 2:	/* (2048/4095)[1 + (a*)/200] */
		x *= (double)(xSize - 1) / (double)(xSize - 2);
		y *= (double)(ySize - 1) / (double)(ySize - 2);
		p = 0.50012210012210 * (1.0 + 2.15517241379310 * (x - y));
		break;

   case 3:	/* (2048/4095)[1 + (b*)/200] */
		y *= (double)(ySize - 1) / (double)(ySize - 2);
		z *= (double)(zSize - 1) / (double)(zSize - 2);
		p = 0.50012210012210 * (1.0 + 0.86206896551724 * (y - z));
		break;

   default:
   		p = 6.023e+23;	/* Avogadro's number */
		break;
	}
	p = RESTRICT (p, 0.0, 1.0);

	p = QUANT1 (p, FUT_GRD_MAXVAL);

	return ((fut_gtbldat_t)p);
}


/*----------------------------------------------------------------------------
 *  outfun -- rescale and clip a* and b* and convert to signed representation
 *----------------------------------------------------------------------------
 */
static fut_otbldat_t ofun (fut_gtbldat_t q, KpGenericPtr_t dataP)
{
double	p;

	p = DEQUANT (q, FUT_GRD_MAXVAL);

	switch (((fData_p) dataP)->channel) {
	case 1:	break;
	case 2:
	case 3:
		p *= 1.99951171875;
		p = 200.0 * (p - 1.0);	/* CIE 1976 a*, b*, in [-200, 200] */
		p = RESTRICT (p, -128.0, 127.0);	/* clip to [-128, 127] */
#ifdef AB_UNSIGNED
		p = p + 128.0;		/* -> [0, 255] */
#else
		if (p <= -1.0)		/* in [-128, -1] */
		   p += 256.0;		/* -> [128, 255] */
		else if (p < -0.5)	/* round to -1 */
		   p = 255.0;
		else if (p < 0.0)	/* round to 0 */
		   p = 0.0;
		else			/* in [0, 127] */
		   ;			/* leave as is */
#endif
		p /= 255.0;		/* from [0, 255] to [0, 1] */
		break;

   default:
		p = 6.023e+23;
		break;
	}
	p = RESTRICT (p, 0.0, 1.0);	/* superfluous, but whatthehell */
			/* L* in [0, 100]; a* & b* in [-128, 127] */

	p =  QUANT1 (p, FUT_MAX_PEL12);

	return ((fut_otbldat_t)p);
}

/*----------------------------------------------------------------------
 *  get_xyz2lab_fut --	construct an XYZ to Lab FuT
 *						to establish a grid for composition

 		MUST be called from within a critical section!!!!!
 *----------------------------------------------------------------------
 */
fut_ptr_t
	get_xyz2lab_fut (KpInt32_t size)
{
fut_ptr_t		futp;
fut_itbl_ptr_t	itblx, itbly, itblz;
fut_gtbl_ptr_t	gtbl [3];
fut_otbl_ptr_t	otbl [3];
KpInt32_t		ichan;
KpInt32_t		iomask, sizeArray[3];
fData_t			fData;

	/* assume all dimensions are the same */
	fData.xSize = size;
	fData.ySize = size;
	fData.zSize = size;

/* Compute shared input tables:  */
	itblx = fut_new_itblEx (fData.xSize, xfun, &fData);
	itbly = fut_new_itblEx (fData.ySize, yfun, &fData);
	itblz = fut_new_itblEx (fData.zSize, zfun, &fData);

/* Compute grid tables and output tables:  */
	iomask = FUT_IN(FUT_XYZ);

	sizeArray[0] = fData.xSize;
	sizeArray[1] = fData.ySize;
	sizeArray[2] = fData.zSize;

	for (ichan = 0; ichan < 3; ichan++) {
	   fData.channel = ichan + 1;
	   gtbl [(KpInt32_t) ichan] = fut_new_gtblEx (iomask, gfun, &fData, sizeArray);
	   otbl [(KpInt32_t) ichan] = fut_new_otblEx (ofun, &fData);
	}

/* Assemble FuT:  */
	iomask = FUT_IN(FUT_XYZ) | FUT_OUT(FUT_XYZ);
	futp = fut_new (iomask, itblx, itbly, itblz,
							gtbl [0], otbl [0],
							gtbl [1], otbl [1],
							gtbl [2], otbl [2]);

	if ( futp->idstr != NULL ) {
		fut_free_idstr (futp->idstr);
		futp->idstr = NULL;
	}

	fut_free_tbls ((KpInt32_t fut_far *)itblx, itbly, itblz,
							gtbl [0], gtbl [1], gtbl [2],
							otbl [0], otbl [1], otbl [2], NULL);

	return (futp);
}
