/*
 * @(#)CompositeFont.cpp	1.14 98/10/08
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "FontGlue.h"
#include "CompositeFont.h"
#include "Strike.h"         // REMIND: solely for counting glyphs
#include <stdlib.h>
#include <ctype.h>

const void*
CompositeFont::ReadChunk(UInt32 offset, UInt32 length, void* chunk) {
    return NULL;
}

// REMIND: need failure status
void CompositeFont::readBlock(size_t offs, size_t size, char *dest) {
    return;
}

void CompositeFont::ReleaseChunk(const void* chunk) {
}

Boolean CompositeFont::MatchName(int nameID, const UInt16 *name, int nameLen) {
    Boolean retval = (nameLen == fNameLen) && fCompositeName;
    if (retval) {
	UInt16 val;
        for (int i=0; i<nameLen && retval ; i++) {
	    if (name [i] > 0xff) {
		val = name [i];
	    } else {
		val = tolower (name[i]);
	    }
            retval &= (val == fCompositeName[i]);
        }
    }
    return retval;
}

// REMIND: move to common location
#define GET16(P)    (UInt16)(   (((UInt8 *)&(P))[0] << 8) \
                                | (((UInt8 *)&(P))[1]) )

// REMIND: need max name length
int CompositeFont::GetName(UInt16& platformID, UInt16& scriptID,
                           UInt16& languageID, UInt16& nameID, UInt16 *name) {
    if (name) {
        for (int i=0; i<fNameLen; i++) {
            name[i] = GET16(fCompositeName[i]);
        }
    }
    platformID = 3;
    scriptID = 0;
    return fNameLen * sizeof(*fCompositeName);
}

const Unicode *
CompositeFont::GetFontName(int& nameLen) {
    nameLen = fNameLen;
    return fCompositeName;
}

UInt32 CompositeFont::GetDataLength() {
    return 0;
}

CompositeFont::CompositeFont(JNIEnv *env, jstring compositeNameArg)
    : fInitialized(JNI_FALSE), fCompositeName(NULL),
      fRanges(NULL), fMaxIndices(NULL), fNumComponents(0),
      fComponentNames(NULL), fComponentFonts(NULL), fComponentChunks(NULL)
{
    JStringBuffer compositeName(env, compositeNameArg);
    fNameLen = compositeName.getLength();
    fCompositeName = new jchar[fNameLen];
    if (fCompositeName) {
        memcpy(fCompositeName, compositeName.buffer(),
                                                fNameLen * sizeof(jchar));
    }
    fFormat = kCompositeFontFormat;
}

void
CompositeFont::addRanges(JNIEnv *env, jintArray exclusionRanges) {
    UInt32Buffer ranges(env, exclusionRanges);
    fRanges = new jint[ranges.getNumElements()];
    if (fRanges) {
        memcpy(fRanges, ranges.buffer(),
            sizeof(jint) * ranges.getNumElements());
    }
}

int
CompositeFont::isExcludedChar(int slot, Unicode unicode) {
    if (!fRanges || !fMaxIndices) {
        return false;
    }
    int minIndex = 0;
    int maxIndex = fMaxIndices[slot];
    if (slot > 0) {
        minIndex = fMaxIndices[slot - 1];
    }
    int curIndex = minIndex;
    while (maxIndex > curIndex) {
        if (    (unicode >= fRanges[curIndex])
                && (unicode <= fRanges[curIndex+1]))
        {
            // excluded
            return true;
        }
        curIndex += 2;
    }
    return false;
}

void
CompositeFont::addMaxIndices(JNIEnv *env, jintArray maxIndices) {
    UInt32Buffer maxIndex(env, maxIndices);
    fNumComponents = maxIndex.getNumElements();
    fMaxIndices = new jint[fNumComponents];
    if (fMaxIndices) {
        memcpy(fMaxIndices, maxIndex.buffer(), sizeof(jint) * fNumComponents);
    }
    fComponentNames = new jchar *[fNumComponents];
    if (fComponentNames) {
        memset(fComponentNames, 0, sizeof(jchar *) * fNumComponents);
    }
    fComponentNameLengths = new int [fNumComponents];
    if (fComponentNameLengths) {
        memset(fComponentNameLengths, 0, sizeof(int) * fNumComponents);
    }
    fComponentFonts = new fontObject *[fNumComponents];
    if (fComponentFonts) {
        memset(fComponentFonts, 0, sizeof(fontObject *) * fNumComponents);
    }
    fComponentChunks = new const void *[fNumComponents];
    if (fComponentChunks) {
        memset(fComponentChunks, 0, sizeof(const void *) * fNumComponents);
    }
}

void
CompositeFont::registerComponentName(int slot, JStringBuffer& componentName) {
    if (fComponentNames
            && fComponentNameLengths
            && (slot >= 0)
            && (slot < fNumComponents)
            && fComponentFonts) {
        if (fComponentNames[slot]) {
            delete [] fComponentNames[slot];
            fComponentNames[slot] = NULL;
        }
        int nameLen = componentName.getLength();
        fComponentNameLengths[slot] = nameLen;
        fComponentNames[slot] = new jchar[nameLen];
        if (fComponentNames[slot]) {
            memcpy(fComponentNames[slot], componentName.buffer(),
                                                    nameLen * sizeof(jchar));
        }
    }
    fInitialized = JNI_TRUE;
    fInitialized = fInitialized && fCompositeName;
    fInitialized = fInitialized && fRanges;
    fInitialized = fInitialized && fMaxIndices;
    fInitialized = fInitialized && fComponentNames;
    if (fInitialized) {
        for (int i = 0; i < fNumComponents; i++) {
            fInitialized = fInitialized && fComponentNames[i];
        }
    }
}

int
CompositeFont::isValid() {
    return fInitialized;
}

int
CompositeFont::countSlots(void) {
    return fNumComponents;
}

fontObject *
CompositeFont::getSlotFont(int slot) {
    fontObject *retval = NULL;
    if ((slot >= 0) && (slot < fNumComponents)) {
        if (fComponentFonts) {
            retval = fComponentFonts[slot];
            if (retval == NULL) {
                fComponentFonts[slot] =
                    ::FindFontObject(   fComponentNames[slot],
                                        fComponentNameLengths[slot], 0);
// REMIND: later speedup to use font file name for quick locator
//                fComponentFonts[slot] =
//                    ::GetFontObject(fComponentNames[slot], nameLen);
                retval = fComponentFonts[slot];
                if (retval == NULL) {
                    // font not located.
                    fComponentFonts[slot] = (fontObject *)-1;
                }
                else if (fComponentChunks) {
                    // Leave this file open
                    if (fComponentChunks[slot] == NULL) {
                        fComponentChunks[slot] = retval->ReadChunk(0, 1, NULL);
                    }
                }
            }
        }
    }
    if (retval == (fontObject *)-1) {
        retval = 0;
    }
    return retval;
}

CompositeFont::~CompositeFont() {
    if (fCompositeName) {
        delete [] fCompositeName;
        fCompositeName = NULL;
    }
    if (fRanges) {
        delete [] fRanges;
        fRanges = NULL;
    }
    if (fMaxIndices) {
        delete [] fMaxIndices;
        fMaxIndices = NULL;
    }
    if (fComponentNames) {
        for (int i=0; i < fNumComponents; i++) {
            delete [] fComponentNames[i];
            fComponentNames[i] = NULL;
        }
        delete [] fComponentNames;
        fComponentNames = NULL;
    }
    if (fComponentNameLengths) {
        delete [] fComponentNameLengths;
        fComponentNameLengths = NULL;
    }
    if (fComponentChunks) {
        if (fComponentFonts) {
            for (int i=0; i < fNumComponents; i++) {
                if (fComponentChunks[i]) {
                    fComponentFonts[i]->ReleaseChunk(fComponentChunks[i]);
                }
            }
        }
        delete [] fComponentChunks;
        fComponentChunks = NULL;
    }
    if (fComponentFonts) {
        delete [] fComponentFonts;
        fComponentFonts = NULL;
    }
}
