/*
 * @(#)CompositeFont.h	1.8 98/10/07
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#ifndef CompositeFont_header
#define CompositeFont_header

#include "FontGlue.h"

class CompositeFont : public fontObject {
private:
    virtual const void* ReadChunk(UInt32 offset,
                                  UInt32 length, void* chunk = nil);
    virtual void readBlock(size_t offs, size_t size, char *dest);
    virtual void ReleaseChunk(const void* chunk);
    virtual Boolean MatchName(int nameID, const UInt16 *name, int nameLen);
    virtual int GetName(UInt16& platformID, UInt16& scriptID,
                        UInt16& languageID, UInt16& nameID, UInt16 *name);
    virtual const Unicode *GetFontName(int& nameLen);
    virtual UInt32 GetDataLength();

    CompositeFont();                // no implementation
    CompositeFont(CompositeFont&);  // no implementation
    CompositeFont& operator =(const CompositeFont&);

private:
    jchar *fCompositeName;
    int   fNameLen;
    jint  *fRanges;
    jint  *fMaxIndices;
    jchar **fComponentNames;
    int   *fComponentNameLengths;
    fontObject **fComponentFonts;
    const void **fComponentChunks;  // Used to keep open the component fonts
    jint  fNumComponents;
    int   fInitialized;

public:
    virtual ~CompositeFont();
    CompositeFont(JNIEnv *env, jstring compositeName);
    virtual void addRanges(JNIEnv *env, jintArray exclusionRanges);
    virtual void addMaxIndices(JNIEnv *env, jintArray maxIndices);
    virtual void registerComponentName(int slot, JStringBuffer& componentName);
    virtual int countSlots(void);
    virtual fontObject *getSlotFont(int slot);
    virtual int isValid(void);
    virtual int isExcludedChar(int slot, Unicode unicode);
};


#endif // CompositeFont_header
