/*
 * @(#)FontGlue.h	1.18 98/09/15
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */
#ifndef fontGlue_header
#define fontGlue_header

#include <jni.h>
#include <jni_util.h>
#include <jlong_md.h>
#include <sun_awt_font_NativeFontWrapper.h>

#include <memory.h>

#include "fontObject.h"
#include "hsGGlyphCache.h"
#include "hsGFontScaler.h"
#include "hsFixedTypes.h"
#include "ImageData.h"

extern fontObject *getFontPtr(JNIEnv *env, jobject font);

// FontTransform --- constructed from primitiveArray of float
class FontTransform {
private:
    jdouble fMatrix[4];
private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
//    FontTransform(FontTransform&);  // no implementation
    FontTransform& operator =(const FontTransform&);
//    static void *operator new(size_t size);
//    static void operator delete(void *ptr);
public:
    FontTransform() {
        fMatrix[0] = 1;
        fMatrix[1] = 0;
        fMatrix[2] = 0;
        fMatrix[3] = 1;
    }
    FontTransform(float *matrix) {
        if (matrix) {
            fMatrix[0] = matrix[0];
            fMatrix[1] = matrix[1];
            fMatrix[2] = matrix[2];
            fMatrix[3] = matrix[3];
        }
    }
    FontTransform(JNIEnv *env, jdoubleArray matRef) {
        jint matSize = env->GetArrayLength(matRef);
        if (matSize >= 4) {
            jdouble matrix[4];
            env->GetDoubleArrayRegion(matRef, 0, 4, matrix);
            fMatrix[0] = matrix[0];
            fMatrix[1] = matrix[1];
            fMatrix[2] = matrix[2];
            fMatrix[3] = matrix[3];
        }
    }
    ~FontTransform() {
    }
    void getMatrixInto(float *dest, unsigned int destSize) {
        if (fMatrix && (destSize >= 4)) {
            dest[0] = fMatrix[0];
            dest[1] = fMatrix[1];
            dest[2] = fMatrix[2];
            dest[3] = fMatrix[3];
        } else {
            memset(dest, 0, destSize * sizeof(*dest));
        }
    }
    int equals(const FontTransform& cTX) const {
        return fMatrix
            && cTX.fMatrix
            && fMatrix[0] == cTX.fMatrix[0]
            && fMatrix[1] == cTX.fMatrix[1]
            && fMatrix[2] == cTX.fMatrix[2]
            && fMatrix[3] == cTX.fMatrix[3];
    }
    FontTransform(FontTransform& cTX) {
        fMatrix[0] = cTX.fMatrix[0];
        fMatrix[1] = cTX.fMatrix[1];
        fMatrix[2] = cTX.fMatrix[2];
        fMatrix[3] = cTX.fMatrix[3];
    }
};

// UInt32Buffer --- constructed from primitiveArray of float
class UInt32Buffer {
private:
    const jintArray arrayRef;
    JNIEnv * const fEnv;
    UInt32 *fStorage;
    UInt32 *fBuffer;
    UInt32 fArraySize;
private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    UInt32Buffer();               // no implementation
    UInt32Buffer(UInt32Buffer&);    // no implementation
    UInt32Buffer& operator =(const UInt32Buffer&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);
public:
    UInt32Buffer(JNIEnv *env, jintArray intArray, UInt32 offset = 0)
            : fEnv(env), arrayRef(intArray),
                fStorage(NULL), fBuffer(NULL), fArraySize(0){
        if (!JNU_IsNull (env, arrayRef)) {
            fArraySize = env->GetArrayLength(arrayRef);
            fStorage = (UInt32 *)env->GetPrimitiveArrayCritical(arrayRef, NULL);
            if (fStorage) {
                fBuffer = fStorage + offset;
            }
        }
    }
    ~UInt32Buffer() {
        if (fStorage) {
            fEnv->ReleasePrimitiveArrayCritical(
                                            arrayRef, fStorage, JNI_ABORT);
        }
    }
    UInt32 getNumElements(void) const { return fArraySize;}
    UInt32 *buffer(void) const { return fBuffer; }
    Boolean invalid(void) const {   return (fBuffer == NULL); }
};

// UInt16Buffer --- constructed from primitiveArray of float
class UInt16Buffer {
private:
    const jshortArray arrayRef;
    JNIEnv * const fEnv;
    UInt16 *fStorage;
    UInt16 *fBuffer;
    UInt32 fArraySize;
private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    UInt16Buffer();               // no implementation
    UInt16Buffer(UInt16Buffer&);    // no implementation
    UInt16Buffer& operator =(const UInt16Buffer&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);
public:
    UInt16Buffer(JNIEnv *env, jshortArray shortArray, jint offset = 0)
            : fEnv(env), arrayRef(shortArray),
                fStorage(NULL), fBuffer(NULL), fArraySize(0){
        if (!JNU_IsNull (env, arrayRef)) {
            fArraySize = env->GetArrayLength(arrayRef);
            fStorage = (UInt16 *)env->GetPrimitiveArrayCritical(arrayRef, NULL);
            if (fStorage) {
                fBuffer = fStorage + offset;
            }
        }
    }
    ~UInt16Buffer() {
        if (fStorage) {
            fEnv->ReleasePrimitiveArrayCritical(
                                            arrayRef, fStorage, JNI_ABORT);
        }
    }
    UInt32 getNumElements(void) const { return fArraySize;}
    UInt16 *buffer(void) const { return fBuffer; }
    Boolean invalid(void) const {   return (fBuffer == NULL); }
};

// UInt8Buffer --- constructed from primitiveArray of float
class UInt8Buffer {
private:
    const jbyteArray arrayRef;
    JNIEnv * const fEnv;
    UInt8 *fStorage;
    UInt8 *fBuffer;
    UInt32 fArraySize;
private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    UInt8Buffer();               // no implementation
    UInt8Buffer(UInt8Buffer&);    // no implementation
    UInt8Buffer& operator =(const UInt8Buffer&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);
public:
    UInt8Buffer(JNIEnv *env, jbyteArray byteArray, jint offset = 0)
            : fEnv(env), arrayRef(byteArray),
                fStorage(NULL), fBuffer(NULL), fArraySize(0){
        if (!JNU_IsNull (env, arrayRef)) {
            fArraySize = env->GetArrayLength(arrayRef);
            fStorage = (UInt8 *)env->GetPrimitiveArrayCritical(arrayRef, NULL);
            if (fStorage) {
                fBuffer = fStorage + offset;
            }
        }
    }
    ~UInt8Buffer() {
        if (fStorage) {
            fEnv->ReleasePrimitiveArrayCritical(
                                            arrayRef, fStorage, JNI_ABORT);
        }
    }
    UInt32 getNumElements(void) const { return fArraySize;}
    UInt8 *buffer(void) const { return fBuffer; }
    Boolean invalid(void) const {   return (fBuffer == NULL); }
};

// JFloatBuffer --- constructed from primitiveArray of float
class JFloatBuffer {
private:
    const jfloatArray arrayRef;
    JNIEnv * const fEnv;
    jfloat *fStorage;
    jfloat *fBuffer;
    UInt32 fArraySize;
private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    JFloatBuffer();               // no implementation
    JFloatBuffer(JFloatBuffer&);    // no implementation
    JFloatBuffer& operator =(const JFloatBuffer&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);
public:
    JFloatBuffer(JNIEnv *env, jfloatArray floatArray, jint offset = 0)
            : fEnv(env), arrayRef(floatArray),
                fStorage(NULL), fBuffer(NULL), fArraySize(0){
        if (!JNU_IsNull (env, arrayRef)) {
            fArraySize = env->GetArrayLength(arrayRef);
            fStorage = (jfloat *)env->GetPrimitiveArrayCritical(arrayRef, NULL);
            if (fStorage) {
                fBuffer = fStorage + offset;
            }
        }
    }
    ~JFloatBuffer() {
        if (fStorage) {
            fEnv->ReleasePrimitiveArrayCritical(
                                            arrayRef, fStorage, JNI_ABORT);
        }
    }
    UInt32 getNumElements(void) const { return fArraySize;}
    jfloat *buffer(void) const { return fBuffer; }
    Boolean invalid(void) const {   return (fBuffer == NULL); }
};

// JStringBuffer --- constructed from primitiveArray of String
class JStringBuffer {
private:
    const jstring stringRef;
    JNIEnv * const fEnv;
    const jchar *fStorage;
    jsize fLength;
private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    JStringBuffer();               // no implementation
    JStringBuffer(JStringBuffer&);    // no implementation
    JStringBuffer& operator =(const JStringBuffer&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);
public:
    JStringBuffer(JNIEnv *env, jstring theString)
            : fEnv(env), stringRef(theString),
                fStorage(NULL) {
        if (!JNU_IsNull (env, theString)) {
            fLength = env->GetStringLength(stringRef);
            fStorage = env->GetStringCritical(stringRef, JNI_FALSE);
        }
    }
    ~JStringBuffer() {
        if (fStorage) {
            fEnv->ReleaseStringCritical(stringRef, fStorage);
        }
    }
    UInt32 getLength(void) const { return fLength;}
    const jchar *buffer(void) const { return fStorage; }
    Boolean invalid(void) const {   return (fStorage == NULL); }
};

class TempBuffer {
    char fTempSpace[100];
    char *fStorage;
private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    TempBuffer();               // no implementation
    TempBuffer(TempBuffer&);    // no implementation
    TempBuffer& operator =(const TempBuffer&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);
public:
    TempBuffer(UInt32 count) : fStorage(NULL) {
        if (count > sizeof(fTempSpace)) {
            fStorage = new char[count];
        } else {
            fStorage = fTempSpace;
        }
    }
    ~TempBuffer() {
        if (fStorage && (fStorage != fTempSpace)) {
            delete [] fStorage;
            fStorage = NULL;
        }
    }
    void *storage(void) const { return fStorage; }
};

// ImageDataWrapper --- constructed from ImageData
class ImageDataWrapper {
private:
    const jobject fImageData;
    JNIEnv * const fEnv;
    unsigned int fElementSize;
    void *fOriginPtr;
    unsigned int fPixelStride;
    unsigned int fScanStride;
    Boolean fAAHint;
    Boolean fIndexed;
    Boolean fInvalid;
    int fType;

    ImageDataIntLockInfo    intLockInfo;
    ImageDataShortLockInfo  shortLockInfo;
    ImageDataByteLockInfo   byteLockInfo;
    ImageDataByteIndexedLockInfo byteIndexedLockInfo;
    ImageDataIndex8GrayLockInfo index8GrayLockInfo;

private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    ImageDataWrapper();               // no implementation
    ImageDataWrapper(ImageDataWrapper&);    // no implementation
    ImageDataWrapper& operator =(const ImageDataWrapper&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);

public:
    ImageDataWrapper(JNIEnv *env, jobject imageData, Boolean aaHint,
                    unsigned int elementSize, jint& cx, jint& cy);
    ~ImageDataWrapper();
    Boolean invalid(void) const      { return fInvalid; }
    void *topLeftPtr(void) const     { return fOriginPtr; }
    unsigned int pixStr(void) const  { return fPixelStride; }
    unsigned int scanStr(void) const { return fScanStride; }
    const unsigned int *getDstLut(void) const;
    const char *getInvCmap(void) const;
    const ColorData *getCData(void) const;
    const unsigned int *getinverseGrayLut(void) const;
};

#endif // fontGlue_header
