/*
 * @(#)FontWrapper.cpp	1.23 98/09/15
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "FontGlue.h"
#include "CompositeFont.h"
#include "Strike.h"
#include "GlyphVector.h"
#include "cpu.h"

// private field inside of Font object to reference native instance
static jfieldID gFont_pNativeFont = 0;
static jfieldID gFont_style = 0;

// get native instance from Font object
fontObject *getFontPtr(JNIEnv *env, jobject font)
{
    fontObject *fo = NULL;
    if (JNU_IsNull (env, font)) {
        JNU_ThrowIllegalArgumentException(env, "Font");
        return NULL;
    }
    if (!gFont_pNativeFont || !gFont_style) {
        jclass fontClass = env->GetObjectClass(font);
        if (!fontClass) {
            JNU_ThrowClassNotFoundException(env, "No Font class");
            return NULL;
        }
        if (!gFont_pNativeFont) {
            gFont_pNativeFont = env->GetFieldID(fontClass, "pNativeFont","J");
            if (!gFont_pNativeFont) {
                JNU_ThrowNoSuchFieldException(env, "in Font class");
                return NULL;
            }
        }
        if (!gFont_style) {
            gFont_style = env->GetFieldID(fontClass, "style", "I");
            if (!gFont_style) {
                JNU_ThrowNoSuchFieldException(env, "in Font class");
                return NULL;
            }
        }
    }
    jlong pNativeFont = env->GetLongField(font, gFont_pNativeFont);
    fo = (fontObject *)jlong_to_ptr(pNativeFont);
    jint style = env->GetIntField(font, gFont_style);
    // REMIND: this is yet another thing which makes this code not reentrant.
    fo->m_currentStyle = style;
    return fo;
}

// set native instance into Font object
static void setFontPtr(JNIEnv *env, jobject font, fontObject *foPtr)
{
    if (!gFont_pNativeFont) {
        jclass fontClass = env->GetObjectClass(font);
        if (!fontClass) {
            JNU_ThrowClassNotFoundException(env, "No Font class");
            return;
        }
        gFont_pNativeFont = env->GetFieldID(fontClass, "pNativeFont", "J");
        if (!gFont_pNativeFont) {
            JNU_ThrowNoSuchFieldException(env, "in Font class");
            return;
        }
    }
    env->SetLongField(font, gFont_pNativeFont, ptr_to_jlong(foPtr));
}

static jclass g_gmClass             = 0;
static jfieldID g_gmAdvanceField    = 0;
static jfieldID g_gmBoundsRect      = 0;
static jclass g_rect2DClass         = 0;
static jfieldID g_rect2DX           = 0;
static jfieldID g_rect2DY           = 0;
static jfieldID g_rect2DW           = 0;
static jfieldID g_rect2DH           = 0;

static Boolean initGMIDs(JNIEnv *env, jobject gm) {
    if (!g_gmClass) {
        g_gmClass = env->GetObjectClass(gm);
        if (!g_gmClass) {
            JNU_ThrowClassNotFoundException(env, "No GlyphMetrics class");
            return false;
        }
    }
    if (!g_gmAdvanceField || !g_gmBoundsRect) {
        g_gmAdvanceField = env->GetFieldID(g_gmClass, "advance", "F");
        g_gmBoundsRect =   env->GetFieldID(g_gmClass, "bounds",
                                        "Ljava/awt/geom/Rectangle2D$Float;");
        if (!g_gmAdvanceField || !g_gmBoundsRect) {
            JNU_ThrowNoSuchFieldException(env,
                                            "GlyphMetrics advance or bounds");
            return false;
        }
    }
    if (!g_rect2DClass) {
        jobject bounds = env->GetObjectField(gm, g_gmBoundsRect);
        if (bounds) {
            g_rect2DClass = env->GetObjectClass(bounds);
        }
        if (!g_rect2DClass) {
            JNU_ThrowClassNotFoundException(env,
                    "No Rectangle2D.Float class");
            return false;
        }
    }
    if (!g_rect2DX || !g_rect2DY || !g_rect2DW || !g_rect2DH) {
        g_rect2DX = env->GetFieldID(g_rect2DClass, "x", "F");
        g_rect2DY = env->GetFieldID(g_rect2DClass, "y", "F");
        g_rect2DW = env->GetFieldID(g_rect2DClass, "width", "F");
        g_rect2DH = env->GetFieldID(g_rect2DClass, "height", "F");
        if (!g_rect2DX || !g_rect2DY || !g_rect2DW || !g_rect2DH) {
            JNU_ThrowNoSuchFieldException(env,
                                        "Rectangle2D x, y, width, height");
            return false;
        }
    }
    return true;
}

static void setGlyphMetrics(JNIEnv *env, jobject gm, jfloat advance,
                            jfloat left, jfloat top,
                            jfloat width, jfloat height)
{
    if (initGMIDs(env, gm)) {
        jobject bounds = env->GetObjectField(gm, g_gmBoundsRect);
        env->SetFloatField(gm, g_gmAdvanceField, advance);
        env->SetFloatField(bounds, g_rect2DX, left);
        env->SetFloatField(bounds, g_rect2DY, top);
        env->SetFloatField(bounds, g_rect2DW, width);
        env->SetFloatField(bounds, g_rect2DH, height);
    }
}

// Native implementations of JAVA routines...
extern "C" {
/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    canDisplay
 * Signature: (Ljava/awt/Font;C)Z
 */
JNIEXPORT jboolean JNICALL
Java_sun_awt_font_NativeFontWrapper_canDisplay
    (JNIEnv *env, jclass clsNFW, jobject theFont, jchar c)
{
    NativeLock (env);
    jboolean retval = JNI_FALSE;
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        // REMIND: pass parameters from JAVA to here
        FontTransform tx;
        jboolean isAntiAliased = false;
        jboolean usesFractionalMetrics = false;
        Strike& theStrike = fo->getStrike(
                                    tx,
                                    isAntiAliased, usesFractionalMetrics);
        retval = theStrike.canDisplay(c)? JNI_TRUE: JNI_FALSE;
    }
    NativeUnlock (env);
    return retval;
}

/*   
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    doesGlyphShaping
 * Signature: (Ljava/awt/Font;)Z
 */
JNIEXPORT jboolean JNICALL
Java_sun_awt_font_NativeFontWrapper_doesGlyphShaping
    (JNIEnv *env, jclass clsNFW, jobject theFont)
{
    NativeLock (env);
    jboolean retval = JNI_FALSE;
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        // REMIND: pass parameters from JAVA to here
        FontTransform tx;
        jboolean isAntiAliased = false;
        jboolean usesFractionalMetrics = false;
        Strike& theStrike = fo->getStrike(
                                    tx,
                                    isAntiAliased, usesFractionalMetrics);
        retval = theStrike.doesGlyphShaping()? JNI_TRUE: JNI_FALSE;
    }
    NativeUnlock (env);
    return retval;
}

/*  
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    isStyleSupported
 * Signature: (Ljava/awt/Font;)Z
 */
JNIEXPORT jboolean JNICALL
Java_sun_awt_font_NativeFontWrapper_isStyleSupported
    (JNIEnv *env, jclass clsNFW, jobject theFont,jint theStyle)
{
    NativeLock (env);
    jboolean retval = JNI_FALSE;
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        // REMIND: pass parameters from JAVA to here
        FontTransform tx;
        jboolean isAntiAliased = false;
        jboolean usesFractionalMetrics = false;
        Strike& theStrike = fo->getStrike(
                                    tx,
                                    isAntiAliased, usesFractionalMetrics);
        retval = theStrike.isStyleSupported((int)theStyle);
    }
    NativeUnlock (env);
    return retval;
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getAdvance
 * Signature: (Ljava/awt/Font;C[FZZ)F
 */
JNIEXPORT jfloat JNICALL
Java_sun_awt_font_NativeFontWrapper_getAdvance
    (JNIEnv *env, jclass clsNFW, jobject theFont, jchar theChar,
     jdoubleArray matrix,
     jboolean isAntiAliased, jboolean usesFractionalMetrics)
{
    NativeLock (env);
    jfloat retval = 0;
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        FontTransform tx(env, matrix);
        Strike& theStrike = fo->getStrike(tx, isAntiAliased, usesFractionalMetrics);
        retval = theStrike.getCharAdvanceX(theChar);
    }
    NativeUnlock (env);
    return retval;
}

// look up nameID in theFont and return the result as a Java string.
static jstring lookupName
  (JNIEnv *env, jobject theFont, jshort lcid, UInt16 nameID)
{
    UInt16 name[1024];
    UInt16 platformID = 3;
    UInt16 scriptID = 1;
    UInt16 languageID =lcid;

    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
	// length is the number of bytes in the name, NOT the number of chars
        int length = fo->GetName(platformID, scriptID, languageID, nameID, name);
        if (length) { 
            if ((platformID == 3) || (platformID == 0)) {
#ifdef littleEndian
                for (int j = 0; j < length/2; j++) {
                    name[j] = SWAPW(name[j]);
                }
#endif
                name[length /2] = 0;
                return env->NewString(name, length / 2);
            } else {
	      // Note:  this assumes that if the encoding is not
	      // Unicode it is UTF.  This may not be true, particularly
	      // for some Apple encodings.
                return env->NewStringUTF((char *)name);
            }
        }
    }
    return env->NewStringUTF("");
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getFamilyName
 * Signature: (Ljava/awt/Font;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL
Java_sun_awt_font_NativeFontWrapper_getFamilyName
  (JNIEnv *env, jclass clsNFW, jobject theFont, jshort lcid)
{
  return lookupName(env, theFont, lcid, kFamilyName);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getFamilyNameByIndex
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL
Java_sun_awt_font_NativeFontWrapper_getFamilyNameByIndex
    (JNIEnv *env, jclass clsNFW, jint fontIndex)
{
    UInt16 name[256];
    UInt16 platformID = 3;
    UInt16 scriptID = 1;
    UInt16 languageID = 0xFFFF;
    UInt16 nameID = kFamilyName;

    fontObject* fo = GetFontObject(fontIndex);
    if (fo) {
        int length = fo->GetName(platformID, scriptID, languageID, nameID, name);
        if (length) {
            if ((platformID == 3) || (platformID == 0)) {
#ifdef littleEndian
                for (int j = 0; j < length/2; j++) {
                    name[j] = SWAPW(name[j]);
                }
#endif
                name[length /2] = 0;
                return env->NewString(name, length / 2);
            } else {
                return env->NewStringUTF((char *)name);
            }
        }
    }
    return env->NewStringUTF("");
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getFontMetrics
 * Signature: (Ljava/awt/Font;[FZZ[F)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_getFontMetrics
        (JNIEnv *env, jclass clsNFW, jobject theFont, jdoubleArray transform,
        jboolean isAntiAliased, jboolean usesFractionalMetrics,
        jfloatArray metrics)
{
    NativeLock (env);
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        FontTransform tx(env, transform);
        Strike& theStrike = fo->getStrike(tx, isAntiAliased, usesFractionalMetrics);
        hsFixedPoint2 ascent, descent, baseline, leading, maxAdvance;
        jfloat* resultArray;
        theStrike.GetLineHeight(ascent, descent,
                             baseline, leading, maxAdvance);
        resultArray = (jfloat *)env->GetPrimitiveArrayCritical(metrics, 0);
        if (resultArray) {
            resultArray[0] = -hsFixedToFloat(ascent.fY);
            resultArray[1] = hsFixedToFloat(descent.fY);
            resultArray[2] = -hsFixedToFloat(leading.fY);
            resultArray[3] = hsFixedToFloat(maxAdvance.fX);
            env->ReleasePrimitiveArrayCritical(metrics,
                                               resultArray, JNI_ABORT);
        }
    }
    NativeUnlock (env);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getFullName
 * Signature: (Ljava/awt/Font;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL
Java_sun_awt_font_NativeFontWrapper_getFullName
        (JNIEnv *env, jclass clsNFW, jobject theFont, jshort lcid)
{
  return lookupName(env, theFont, lcid, kFullName);
}


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getFullNameByIndex
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL
Java_sun_awt_font_NativeFontWrapper_getFullNameByIndex
    (JNIEnv *env, jclass clsNFW, jint fontIndex)
{
    UInt16 name[1024];
    UInt16 platformID = 3;
    UInt16 scriptID = 1;
    UInt16 languageID = 0xFFFF;
    UInt16 nameID = kFullName;

    fontObject* fo = GetFontObject(fontIndex);
    if (fo) {
        int length = fo->GetName(platformID, scriptID, languageID, nameID, name);
        if (length) {
            if ((platformID == 3) || (platformID == 0)) {
#ifdef littleEndian
                for (int j = 0; j < length/2; j++) {
                    name[j] = SWAPW(name[j]);
                }
#endif
                name[length /2] = 0;
                return env->NewString(name, length / 2);
            } else {
                return env->NewStringUTF((char *)name);
            }
        }
    }
    return env->NewStringUTF("");
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getFullNameByFileName
 * Signature: (I)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL
Java_sun_awt_font_NativeFontWrapper_getFullNameByFileName
    (JNIEnv *env, jclass clsNFW, jstring fontFileName)
{
    UInt16 name[1024];
    UInt16 platformID = 3;
    UInt16 scriptID = 1;
    UInt16 languageID = 0xFFFF;
    UInt16 nameID = kFullName;
    fontObject *fo = NULL;

    {
        JStringBuffer ffName(env, fontFileName);
        fo = GetFontObject(ffName.buffer(), ffName.getLength());
    }
    if (fo) {
        int length = fo->GetName(platformID, scriptID, languageID, nameID, name);
        if (length) {
            if ((platformID == 3) || (platformID == 0)) {
#ifdef littleEndian
                for (int j = 0; j < length/2; j++) {
                    name[j] = SWAPW(name[j]);
                }
#endif
                name[length /2] = 0;
                return env->NewString(name, length / 2);
            } else {
                return env->NewStringUTF((char *)name);
            }
        }
    }
    return NULL;
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getGlyphJustificationInfo
 * Signature: (Ljava/awt/Font;ILjava/awt/font/GlyphJustificationInfo;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_getGlyphJustificationInfo
  (JNIEnv *, jclass, jobject, jint, jobject);

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getGlyphMetrics
 * Signature: (Ljava/awt/Font;I[FZZLjava/awt/font/GlyphMetrics;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_getGlyphMetrics
        (JNIEnv *env, jclass clsNFW, jobject theFont,
        jint glyphCode, jdoubleArray matrix,
        jboolean doAntiAlias, jboolean doFractEnable, jobject glyphMetrics)
{
// REMIND: this is not a good implementation, since
// it should be context dependent, which means
// that we can only do meaurements once layout has been performed
// on the glyphset itself.  tracking and kerning affect glyphs
// differently depending on their context
    NativeLock (env);
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        hsGGlyph glyph;
        hsFixedPoint2 glyphAdvance;
        {
            FontTransform tx(env, matrix);
            Strike& theStrike = fo->getStrike(tx, doAntiAlias, doFractEnable);
            //ask strike for advance & bounds
            theStrike.getMetrics(glyphCode, glyph, glyphAdvance);
        }
        jfloat advance = hsFixedToFloat(glyphAdvance.fX);
        jfloat boundsWidth = (jfloat) glyph.fWidth;
        jfloat boundsHeight = (jfloat) glyph.fHeight;
        jfloat boundsLeft = hsFixedToFloat(glyph.fTopLeft.fX);
        jfloat boundsTop = hsFixedToFloat(glyph.fTopLeft.fY);

        setGlyphMetrics(env, glyphMetrics, advance,
                        boundsLeft, boundsTop, boundsWidth, boundsHeight);
    }
    NativeUnlock (env);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getGlyphInfo
 * Signature: (Ljava/awt/Font;[CII[FZZ[F)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_getGlyphInfo
        (JNIEnv *env, jclass clsNFW, jobject theFont,
        jcharArray glyphs, jint start, jint length, jdoubleArray matrix,
        jboolean doAntiAlias, jboolean doFractEnable, jfloatArray results)
{
    NativeLock (env);
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
      FontTransform tx(env, matrix);
      Strike& theStrike = fo->getStrike(tx, doAntiAlias, doFractEnable);

      // Can't use this since it calls GetArrayLength().  We should
      // consider having stack-based 'locker' classes which just
      // lock and unlock arrays.
      //UInt16Buffer gb(env, (jshortArray)glyphs, start);
      //UInt16* gp = gb.buffer();
      UInt16* gpStorage = (UInt16*) 
		    env->GetPrimitiveArrayCritical(glyphs, NULL);

      // We modify this later; gpStorage preserves original pointer
      if (gpStorage) {
	UInt16* gp = gpStorage + start;
	UInt16* ep = gp + length;
	--gp;

	//JFloatBuffer rb(env, results);
	//jfloat* rp = rb.buffer();
	jfloat* rpStorage = (jfloat*) env->GetPrimitiveArrayCritical(results, NULL);
	jfloat* rp = rpStorage;
	if (rp) {
	  --rp;

	  jfloat x = *(rp+1);
	  jfloat y = *(rp+2);

	  while (++gp != ep) {
	    hsGGlyph glyph;
	    hsFixedPoint2 advanceXY;

	    *++rp = x;
	    *++rp = y;
	    if (*gp == 0xffff) {
	      *++rp = 0;
	      *++rp = 0;
	      *++rp = x;
	      *++rp = y;
	      *++rp = 0;
	      *++rp = 0;
	    } else {
	      theStrike.getMetrics(theStrike.CharToGlyph(*gp), glyph, advanceXY);

	      //	    fprintf(stderr, "glyph w:%d h:%d rb:%d tlx:%f tly:%f\n", (int)glyph.fWidth, (int)glyph.fHeight, (int)glyph.fRowBytes, (double)hsFixedToFloat(glyph.fTopLeft.fX), (double)hsFixedToFloat(glyph.fTopLeft.fY));
	      //	    fprintf(stderr, "advxy x:%x y:%x\n", (unsigned int)advanceXY.fX, (unsigned int)advanceXY.fY);

	      jfloat ax = hsFixedToFloat(advanceXY.fX);
	      jfloat ay = hsFixedToFloat(advanceXY.fY);

	      //	    fprintf(stderr, "dump %x -> %x %f %f %f %f\n", *gp, fo->CharToGlyph(*gp), (double)x, (double)y, (double)ax, (double)ay);

	      *++rp = ax;
	      *++rp = ay;
	      *++rp = x + hsFixedToFloat(glyph.fTopLeft.fX);
	      *++rp = y + hsFixedToFloat(glyph.fTopLeft.fY);
	      *++rp = (jfloat)glyph.fWidth;
	      *++rp = (jfloat)glyph.fHeight;
	      x += ax;
	      y += ay;
	    }
	  }
	  env->ReleasePrimitiveArrayCritical(results, rpStorage, JNI_COMMIT);
	}
	env->ReleasePrimitiveArrayCritical(glyphs, gpStorage, JNI_ABORT);
      }
    }
    NativeUnlock (env);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getItalicAngle
 * Signature: (Ljava/awt/Font;[FZZ)F
 */
JNIEXPORT jfloat JNICALL
Java_sun_awt_font_NativeFontWrapper_getItalicAngle
    (JNIEnv *env, jclass clsNFW, jobject theFont,
    jdoubleArray matrix, jboolean isAntiAliased, jboolean usesFractionalMetrics)
{
    NativeLock (env);
    jfloat retVal = 0.0;
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        FontTransform tx(env, matrix);
        Strike& theStrike = fo->getStrike(tx, isAntiAliased, usesFractionalMetrics);
        hsFixedPoint2 italicAngle;
        theStrike.GetItalicAngle(italicAngle);
        if (italicAngle.fX != 0) {
            // REMIND: caller code is expecting something a little different
            retVal = hsFixedToFloat(italicAngle.fX)
                        / hsFixedToFloat(italicAngle.fY);
            //return (    hsFixedToFloat(italicAngle.fX)
             //           / hsFixedToFloat(italicAngle.fY));
        }
    }
    NativeUnlock (env);
    //return 0;
    return retVal;
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getMissingGlyphCode
 * Signature: (Ljava/awt/Font;)I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_font_NativeFontWrapper_getMissingGlyphCode
        (JNIEnv *env, jclass clsNFW, jobject theFont)
{
    NativeLock (env);
    jint retval = 0;
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        // REMIND: pass parameters from JAVA to here
        FontTransform tx;
        jboolean isAntiAliased = false;
        jboolean usesFractionalMetrics = false;
        Strike& theStrike = fo->getStrike(
                                    tx,
                                    isAntiAliased, usesFractionalMetrics);
        retval = theStrike.getMissingGlyphCode();
    }
    NativeUnlock (env);
    return retval;
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getNumFonts
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_font_NativeFontWrapper_getNumFonts(JNIEnv *env, jclass clsNFW)
{
    return ::CountFontObjects();
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getNumGlyphs
 * Signature: (Ljava/awt/Font;)I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_font_NativeFontWrapper_getNumGlyphs
        (JNIEnv *env, jclass clsNFW, jobject theFont)
{
    NativeLock (env);
    jint retVal = 0;
    fontObject *fo = ::getFontPtr(env, theFont);

    if (fo) {
        // REMIND: pass parameters from JAVA to here
        FontTransform tx;
        jboolean isAntiAliased = false;
        jboolean usesFractionalMetrics = false;
        Strike& theStrike = fo->getStrike(
                                    tx,
                                    isAntiAliased, usesFractionalMetrics);
        //return theStrike.GetNumGlyphs();
        retVal = theStrike.GetNumGlyphs();
    }
    NativeUnlock (env);
    //return 0;
    return retVal;
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    initializeFont
 * Signature: (Ljava/awt/Font;Ljava/lang/String;I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_initializeFont
        (JNIEnv *env, jclass clsNFW, jobject theFont, jstring name, jint style)
{
    fontObject* fo = NULL;

    if (!JNU_IsNull (env, name)) {
        const UInt16 *fontName = NULL;
        unsigned int nameLen = 0;
        fontName = (UInt16 *)env->GetStringChars(name, NULL);
        if (!fontName) {
            JNU_ThrowIllegalArgumentException(env, "Name");
            return;
        }
        nameLen = env->GetStringLength(name);
        fo = ::FindFontObject(fontName, nameLen, style);
        env->ReleaseStringChars(name, fontName);
    } else {
        fo = ::FindFontObject(NULL, 0, style);
    }
    ::setFontPtr(env, theFont, fo);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    layoutGlyphVector
 * Signature: (Ljava/awt/Font;[FZZFFLjava/awt/font/GlyphVector;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_font_NativeFontWrapper_layoutGlyphVector
    (JNIEnv *env, jclass clsNFW, jobject theFont,
            jdoubleArray matrix,
            jboolean isAntiAliased, jboolean usesFractionalMetrics,
            jfloat x, jfloat y, jobject target)
{
    NativeLock (env);
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        GlyphVector gv(env, matrix, isAntiAliased, usesFractionalMetrics, fo);
        gv.getGlyphCodes(target);
        gv.positionGlyphs(x, y, matrix, isAntiAliased, usesFractionalMetrics);
        gv.writePositions(target);
    }
    NativeUnlock (env);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    populateAndLayoutGlyphVector
 * Signature: (Ljava/awt/Font;[C[FZZFFLjava/awt/font/GlyphVector;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_populateAndLayoutGlyphVector
    (JNIEnv *env, jclass clsNFW, jobject theFont, jcharArray unicodes,
            jdoubleArray matrix,
            jboolean isAntiAliased, jboolean usesFractionalMetrics,
            jfloat x, jfloat y, jobject target)
{
    NativeLock (env);
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        GlyphVector gv(env, unicodes, 0, env->GetArrayLength(unicodes),
                       matrix, isAntiAliased, usesFractionalMetrics, fo);
        gv.positionGlyphs(x, y, matrix, isAntiAliased, usesFractionalMetrics);
        gv.writeGlyphCodes(target);
        gv.writePositions(target);
    }
    NativeUnlock (env);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    populateGlyphVector
 * Signature: (Ljava/awt/Font;[CLjava/awt/font/GlyphVector;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_populateGlyphVector
    (JNIEnv *env, jclass clsNFW,
        jobject theFont, jcharArray unicodes,
        jdoubleArray matrix,
        jboolean isAntiAliased, jboolean usesFractionalMetrics,
        jobject target)
{
    NativeLock (env);
    fontObject *fo = ::getFontPtr(env, theFont);
    if (fo) {
        GlyphVector gv(env, unicodes, 0, env->GetArrayLength(unicodes),
                       matrix, isAntiAliased, usesFractionalMetrics, fo);
        gv.writeGlyphCodes(target);
    }
    NativeUnlock (env);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    shapeGlyphVector
 * Signature: (Ljava/awt/font/GlyphVector;Ljava/awt/Font;[FZZLjava/lang/Object;Ljava/awt/font/GlyphVector;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_font_NativeFontWrapper_shapeGlyphVector
    (JNIEnv *, jclass, jobject, jobject, jdoubleArray, jboolean, jboolean, jobject, jobject)
{
    // REMIND finish me
    // we don't do glyph shaping yet.
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    registerCompositeFont
 * Signature: (Ljava/lang/String;[Ljava/lang/String;[I[I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_registerCompositeFont
  (JNIEnv *env, jclass clsNFW,
    jstring compositeFontName,
    jobjectArray componentFontNames,
    jintArray exclusionRanges, jintArray exclusionMaxIndex)
{
    if (JNU_IsNull (env, compositeFontName)
            || JNU_IsNull (env, componentFontNames)
            || JNU_IsNull (env, exclusionRanges)
            || JNU_IsNull (env, exclusionMaxIndex)) {
        return;
    }
    jsize numComponents = env->GetArrayLength(componentFontNames);
    if (numComponents <= 0) {
        return;
    }
    CompositeFont *cf = new CompositeFont(env, compositeFontName);
    if (!cf) {
        return;
    }
    cf->addRanges(env, exclusionRanges);
    cf->addMaxIndices(env, exclusionMaxIndex);
    for (int i = 0; i < numComponents; i++) {
        jobject temp = env->GetObjectArrayElement(componentFontNames, i);
        jstring theName = (jstring)temp;
        JStringBuffer componentName(env, theName);

        cf->registerComponentName(i, componentName);
    }
    if (cf->isValid()) {
        registerFont(kCompositeFontFormat, cf);
    }
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    registerFonts
 * Signature: (Ljava/util/Vector;II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_font_NativeFontWrapper_registerFonts
        (JNIEnv *env, jclass obj, jobject fonts, jint size, 
		jobject names, jint format);
// REMIND finish me

};  // end of extern "C"

// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation (Solaris)
void UInt32Buffer::operator delete(void *ptr)
{
}
// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation (Solaris)
void TempBuffer::operator delete(void *ptr)
{
}
// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation (Solaris)
void JFloatBuffer::operator delete(void *ptr)
{
}
// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation (Solaris)
void UInt16Buffer::operator delete(void *ptr)
{
}
// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation (Solaris)
void UInt8Buffer::operator delete(void *ptr)
{
}
// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation (Solaris)
void JStringBuffer::operator delete(void *ptr)
{
}

