/*
 * @(#)GlyphVector.h	1.11 98/08/31
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "FontGlue.h"

// REMIND: speedup if we write a "hsDesc" which can live on the stack

class GlyphVector {
private:
    struct ImageRef {
        const void *pixels;
        unsigned int rowBytes;
        unsigned int width;
        unsigned int height;
    };
    fontObject * const fFont;
    JNIEnv * const fEnv;

    Boolean fNeedShaping;

    UInt32  fNumGlyphs;

    // default storage, to avoid heap allocations
    enum {
        eDefaultStorage = 100
    };
    UInt32  fBaseGlyphs[eDefaultStorage];
    float   fBasePositions[eDefaultStorage*2];
    ImageRef fBaseImageRefs[eDefaultStorage];

    UInt32  *fGlyphs;
    float   *fPositions;
    ImageRef *fImageRefs;

    jdoubleArray fMatrix;
    jboolean fIsAntiAliased;
    jboolean fUsesFractionalMetrics;

    float fDevTX[4];

    void allocateGlyphs(void);
    void allocatePositions(void);
    void allocateImageRefs(void);

    void initGlyphs(const jchar *theChars, FontTransform& tx);

private:
// make sure that caller must place these objects on the stack
// since they have internal state which is stack dependent.
// (actually this is a gentle hint to other programmers, since
// they can get around this if they want to)
    GlyphVector();              // no implementation
    GlyphVector(GlyphVector&);  // no implementation
    GlyphVector& operator =(const GlyphVector&);
    static void *operator new(size_t size);
    static void operator delete(void *ptr);

public:
    GlyphVector(JNIEnv *env, jstring theString,
                jdoubleArray matrix,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo);
    GlyphVector(JNIEnv *env,
                jcharArray theCharArray, jint offset, jint count,
                jdoubleArray matrix,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo);
    GlyphVector(JNIEnv *env,
                jbyteArray theByteArray, jint offset, jint count,
                jdoubleArray matrix,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo);
    GlyphVector(JNIEnv *env, jintArray theGlyphArray,
                jdoubleArray matrix,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo);
    GlyphVector(JNIEnv *env,
                jdoubleArray matrix,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo);
    ~GlyphVector();

    // return true if any glyphs can be shaped in the font
    Boolean needShaping();

    // set position of each glyph from specified array
    // this will perform no work if they array reference is NULL
    void setPositions(jfloatArray newPos);

    // layout glyphs and retrieve a ref to their bitmaps
    // with respect to specified transform and rendering hints
    // this method causes state changes which should not be
    // returned to the caller's glyphvector.  We could enforce this
    // rule programmatically, but have choosen not to bother for the
    // time being.
    void positionGlyphsWithImages(
            jfloat x, jfloat y,
            jdoubleArray matrix,
            Boolean isAntiAliased,
            Boolean usesFractionalMetrics);

    // layout glyphs with respect to specified transform and rendering hints
    void positionGlyphs(
            jfloat x, jfloat y,
            jdoubleArray matrix,
            Boolean isAntiAliased,
            Boolean usesFractionalMetrics);

    void setDevTX(jdoubleArray devTX);

    inline unsigned int getNumGlyphs(void) const
    {
        return fNumGlyphs;
    }

    inline void getGlyphPosition(int index, float& x, float& y) const
    {
        x = fPositions[index*2];
        y = fPositions[index*2+1];
    }
    inline void getGlyphBlitInfo(
        int index,
        const void * &pixels,
        unsigned int &rowBytes,
        int &width,
        int &height) const
    {
        pixels      = fImageRefs[index].pixels;
        rowBytes    = fImageRefs[index].rowBytes;
        width       = fImageRefs[index].width;
        height      = fImageRefs[index].height;
    }
    void getGlyphCodes(jobject gvTarget);
    // getPositions will perform basic layout if positions are not
    // provided by the caller
    void getPositions(jobject gvTarget, float x, float y,
                        jdoubleArray matrix,
                        Boolean isAntiAliased, Boolean usesFractionalMetrics);

    void writeGlyphCodes(jobject gvTarget);
    void writePositions(jobject gvTarget);
    void scanShapeInto(jobject shapeTarget, jdoubleArray matrix,
                    jboolean isAntiAliased, jboolean usesFractionalMetrics);
    void scanShapeInto(jobject shapeTarget, jint glyphIndex,
                    jdoubleArray matrix,
                    jboolean isAntiAliased, jboolean usesFractionalMetrics);
    void scanGlyph(jobject shapeTarget,
                    Strike& theStrike,
                    hsPathSpline& path, CurveWalker& walker,
                    float xPos, float yPos,
                    jint glyphIndex);
};
