/*
 * @(#)ImageDataWrapper.cpp	1.13 98/09/21
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "FontGlue.h"
#include <stdio.h>
#include "sun_java2d_loops_GraphicsPrimitive.h"

ImageDataWrapper::ImageDataWrapper(
                    JNIEnv *env, jobject imageData, Boolean aaHint,
                    unsigned int elementSize, jint& cx, jint &cy)
    :   fEnv(env), fImageData(imageData),
        fElementSize(elementSize), fAAHint(aaHint), fInvalid(true),
        fIndexed(false), fType(0)
{
    int x1, y1, x2, y2, x3, y3;
    getViewOriginFromImageData(fEnv, fImageData, &x1, &y1);
    getDeviceOriginFromImageData(fEnv, fImageData, &x2, &y2);
    getOutputOriginFromImageData(fEnv, fImageData, &x3, &y3);
    getTypeFromImageData(fEnv, fImageData, &fType);

    switch (fElementSize) {
    case 4:
        getIntImageLockInfo(fEnv, fImageData, &intLockInfo);
        fOriginPtr = lockIntImageData(fEnv, &intLockInfo);
        fInvalid = (fOriginPtr == NULL);
        fPixelStride = intLockInfo.pixelStride;
        fScanStride = intLockInfo.scanStride;
        fOriginPtr = (int *)fOriginPtr
                    + (x2-x1-x3) * fPixelStride + (y2-y1-y3) * fScanStride;
        cx = cx - x2 + x1;
        cy = cy - y2 + y1;
        break;

    case 2:
        getShortImageLockInfo(fEnv, fImageData, &shortLockInfo);
        fOriginPtr = lockShortImageData(fEnv, &shortLockInfo);
        fInvalid = (fOriginPtr == NULL);
        fPixelStride = shortLockInfo.pixelStride;
        fScanStride = shortLockInfo.scanStride;
        fOriginPtr = (short *)fOriginPtr
                    + (x2-x1-x3) * fPixelStride + (y2-y1-y3) * fScanStride;
        cx = cx - x2 + x1;
        cy = cy - y2 + y1;
        break;

    case 1:
         /* TYPE_BINARY is not being registered in TextRendering.java, and 
            therefore not adding it here, please not, in the future will 
            need to add it. */
        if ((fType == sun_java2d_loops_GraphicsPrimitive_TYPE_BYTE_INDEXED)) {
           getByteIndexedImageLockInfo(fEnv, fImageData, &byteIndexedLockInfo);
           fOriginPtr = lockByteIndexedImageData(fEnv, &byteIndexedLockInfo);
           fInvalid = (fOriginPtr == NULL);
           fPixelStride = byteIndexedLockInfo.pixelStride;
           fScanStride = byteIndexedLockInfo.scanStride;
           fOriginPtr = (char *)fOriginPtr
               + (x2-x1-x3) * fPixelStride + (y2-y1-y3) * fScanStride;
           cx = cx - x2 + x1;
           cy = cy - y2 + y1;
           fIndexed = true;
        }
        else if (fType == sun_java2d_loops_GraphicsPrimitive_TYPE_INDEX8_GRAY){
            getIndex8GrayImageLockInfo(fEnv, fImageData, &index8GrayLockInfo);
            fOriginPtr = lockIndex8GrayImageData(fEnv, &index8GrayLockInfo);
            fInvalid = (fOriginPtr == NULL);
            fPixelStride = index8GrayLockInfo.pixelStride;
            fScanStride = index8GrayLockInfo.scanStride;
            fOriginPtr = (char *)fOriginPtr
                        + (x2-x1-x3) * fPixelStride + (y2-y1-y3) * fScanStride;
            cx = cx - x2 + x1;
            cy = cy - y2 + y1;
            fIndexed = true;
        }
        else {
            /* This else case is left as default, as the cases that use it are
               TYPE_3BYTE_BGR, TYPE_4BYTE_ABGR, TYPE_4BYTE_ABGR_PRE,
               TYPE_BYTE_GRAY and maybe others */
            getByteImageLockInfo(fEnv, fImageData, &byteLockInfo);
            fOriginPtr = lockByteImageData(fEnv, &byteLockInfo);
            fInvalid = (fOriginPtr == NULL);
            fPixelStride = byteLockInfo.pixelStride;
            fScanStride = byteLockInfo.scanStride;
            fOriginPtr = (char *)fOriginPtr
                        + (x2-x1-x3) * fPixelStride + (y2-y1-y3) * fScanStride;
            cx = cx - x2 + x1;
            cy = cy - y2 + y1;
        } 
        break;

    default:
        // REMIND: need cases with LUT, so that we can do alpha blend
        break;
    }
}

ImageDataWrapper::~ImageDataWrapper()
{
    switch (fElementSize) {
    case 4:
        unlockIntImageData(fEnv, &intLockInfo);
        break;
    case 2:
        unlockShortImageData(fEnv, &shortLockInfo);
        break;
    case 1:
        if ((fType == sun_java2d_loops_GraphicsPrimitive_TYPE_BYTE_INDEXED)) {
            unlockByteIndexedImageData(fEnv, &byteIndexedLockInfo);
        }
        else if (fType == sun_java2d_loops_GraphicsPrimitive_TYPE_INDEX8_GRAY){
            unlockIndex8GrayImageData(fEnv, &index8GrayLockInfo);
        }
        else {
            unlockByteImageData(fEnv, &byteLockInfo);
        } 
        break;
    default:
        // REMIND: need cases with LUT, so that we can do alpha blend
        break;
    }
}

const unsigned int *
ImageDataWrapper::getDstLut(void) const
{
    if (fInvalid || !fIndexed) {
        return NULL;
    }

    if ((fType == sun_java2d_loops_GraphicsPrimitive_TYPE_BYTE_INDEXED)) {
        return byteIndexedLockInfo.lockedLut;
    }
    else if (fType == sun_java2d_loops_GraphicsPrimitive_TYPE_INDEX8_GRAY) {
        return index8GrayLockInfo.lockedLut;
    }
    return NULL;
}
const unsigned int *
ImageDataWrapper::getinverseGrayLut(void) const
{
    if (fInvalid || !fIndexed) {
        return NULL;
    }
    if (fType == sun_java2d_loops_GraphicsPrimitive_TYPE_INDEX8_GRAY) {
        return index8GrayLockInfo.lockedInverseGrayLut;
    }
    return NULL;
}

const char *
ImageDataWrapper::getInvCmap(void) const
{
    if (fInvalid || !fIndexed) {
        return NULL;
    }
     if (fType == sun_java2d_loops_GraphicsPrimitive_TYPE_INDEX8_GRAY) {
        return index8GrayLockInfo.inv_cmap;
    }
     else if ((fType == sun_java2d_loops_GraphicsPrimitive_TYPE_BYTE_INDEXED)){
         return byteIndexedLockInfo.inv_cmap;
     }
     return NULL;
}

const ColorData *
ImageDataWrapper::getCData(void) const
{
    if (fInvalid || !fIndexed) {
        return NULL;
    }
    if ((fType == sun_java2d_loops_GraphicsPrimitive_TYPE_BYTE_INDEXED)){
        return byteIndexedLockInfo.colorData;
    }
    else if (fType == sun_java2d_loops_GraphicsPrimitive_TYPE_INDEX8_GRAY) {
        return index8GrayLockInfo.colorData;
    }
    return NULL;
}

// REMIND: lame.  the C++ compiler/linker complain if there
// is no implementation (Solaris)
void ImageDataWrapper::operator delete(void *ptr)
{
}
