/*
 * @(#)Strike.cpp	1.25 98/09/15
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "Strike.h"
#include "CompositeFont.h"
#include <string.h>
#include <java_awt_Font.h>

//////////////////////////////////////////////////////////////////////////////

const unsigned int SLOTMASK     = 0xff000000;
const unsigned int GLYPHMASK    = 0x00ffffff;
#define GLYPHSLOT(g)            ((g >> 24) & 0xff)
#define GLYPHCODE(s, g) \
        (( (unsigned int)s << 24) | ((unsigned int)g & GLYPHMASK))

class CompositeGlyphMapper : public CharToGlyphMapper {
private:
    class GlyphMapEntry {
    public:
        int     fSlot;
        UInt32  fGlyphCode;
    };

private:
    CompositeFont&  fCompositeFont;
    const Strike&   fStrike;
    UInt32          fMissingGlyph;
    int             fNumSlots;
    GlyphMapEntry   *fGlyphMaps[256];

private:
    CompositeGlyphMapper();                         // no implementation
    CompositeGlyphMapper(CompositeGlyphMapper&);    // no implementation
    CompositeGlyphMapper& operator =(const CompositeGlyphMapper&);
    UInt32 getCachedGlyphCode(const Unicode unicode) const;
    void setCachedGlyphCode(const Unicode unicode,
                            const UInt32 glyphCode,
                            const int slot) const;

public:
    virtual ~CompositeGlyphMapper() {
        for (int i=0; i < 256; i++) {
            if (fGlyphMaps[i]) {
                delete [] fGlyphMaps[i];
                fGlyphMaps[i] = NULL;
            }
        }
    };

    CompositeGlyphMapper(CompositeFont& cf, const Strike& strike)
            : fCompositeFont(cf), fStrike(strike) {
        fMissingGlyph = -1;
        fNumSlots = fCompositeFont.countSlots();
        memset(fGlyphMaps, 0, sizeof(fGlyphMaps));
    };

    virtual void CharsToGlyphs(
        int count, const Unicode unicodes[], UInt32 glyphs[]) const;
    virtual Boolean canDisplay(Unicode ch) const;
    virtual int getMissingGlyphCode() const;
};
//////////////////////////////////////////////////////////////////////////////

Strike::Strike( fontObject& fo, FontTransform& tx,
        jboolean isAntiAliased, jboolean usesFractionalMetrics)
    :   fFont(fo), strike(NULL), fCachedLineMetrics(JNI_FALSE),
        fMapper(NULL), compositeStrike(NULL), fStrikeDescs(NULL),
        totalCompositeGlyphs(0), baseGlyphs(0), isComposite(JNI_FALSE)
{
    enum FontFormats format = fFont.GetFormat();
    enum FontFormats scalerID = ((fontObject&)fFont).getScalerID(
                                            tx,
                                            isAntiAliased,
                                            usesFractionalMetrics);
    if (format == kCompositeFontFormat) {
        isComposite = JNI_TRUE;
        templateDesc.fFontObj = NULL;
        tx.getMatrixInto(templateDesc.fMatrix, 4);
        templateDesc.fDoAntiAlias = isAntiAliased;
        templateDesc.fDoFracEnable = usesFractionalMetrics;
        templateDesc.fStyle = fFont.m_currentStyle;
        return;
    }

    UInt32  sizes[2] = { sizeof(UInt32), sizeof(FontStrikeDesc) };

    hsDescriptor desc = hsDescriptor_New(2, sizes);
    FontStrikeDesc* javaDesc;

    hsDescriptor_Add32(desc, kScalerID_FontScalerDesc, scalerID);
    javaDesc = (FontStrikeDesc *)hsDescriptor_Add(desc,
                                                kFontStrikeDescTag,
                                                sizeof(FontStrikeDesc));
    javaDesc->fFontObj = (fontObject *)&fFont;
    tx.getMatrixInto(javaDesc->fMatrix, 4);
    javaDesc->fDoAntiAlias = isAntiAliased;
    javaDesc->fDoFracEnable = usesFractionalMetrics;
    javaDesc->fStyle = fFont.m_currentStyle;

    hsDescriptor_UpdateCheckSum(desc);

    strike = hsGGlyphStrike::RefStrike(desc);

    hsDescriptor_Delete(desc);
}

Strike::Strike( fontObject& fo, enum FontFormats scalerID, FontTransform& tx,
        jboolean isAntiAliased, jboolean usesFractionalMetrics)
    :   fFont(fo), strike(NULL), fCachedLineMetrics(JNI_FALSE),
        fMapper(NULL), compositeStrike(NULL), fStrikeDescs(NULL),
        totalCompositeGlyphs(0), baseGlyphs(0), isComposite(JNI_FALSE)
{
    enum FontFormats format = fFont.GetFormat();
    if (format == kCompositeFontFormat) {
        isComposite = JNI_TRUE;
        templateDesc.fFontObj = NULL;
        tx.getMatrixInto(templateDesc.fMatrix, 4);
        templateDesc.fDoAntiAlias = isAntiAliased;
        templateDesc.fDoFracEnable = usesFractionalMetrics;
        templateDesc.fStyle = fFont.m_currentStyle;
        return;
    }

    UInt32  sizes[2] = { sizeof(UInt32), sizeof(FontStrikeDesc) };

    hsDescriptor desc = hsDescriptor_New(2, sizes);
    FontStrikeDesc* javaDesc;

    hsDescriptor_Add32(desc, kScalerID_FontScalerDesc, scalerID);
    javaDesc = (FontStrikeDesc *)hsDescriptor_Add(desc,
                                                kFontStrikeDescTag,
                                                sizeof(FontStrikeDesc));
    javaDesc->fFontObj = (fontObject *)&fFont;
    tx.getMatrixInto(javaDesc->fMatrix, 4);
    javaDesc->fDoAntiAlias = isAntiAliased;
    javaDesc->fDoFracEnable = usesFractionalMetrics;
    javaDesc->fStyle = fFont.m_currentStyle;

    hsDescriptor_UpdateCheckSum(desc);

    strike = hsGGlyphStrike::RefStrike(desc);

    hsDescriptor_Delete(desc);
}

Strike::~Strike() {
    if (strike) {
        hsRefCnt_SafeUnRef(strike);
        strike = NULL;
    } else {
        if (fMapper) {
            delete fMapper;
        }
    }
    if (compositeStrike) {
        CompositeFont *cf = (CompositeFont *)&fFont;
        int numSlots = cf->countSlots();
        for (int i=0; i < numSlots; i++) {
            if (compositeStrike[i]) {
                delete compositeStrike[i];
                compositeStrike[i] = NULL;
            }
        }
        if (compositeStrike != defaultStrikes) {
            delete [] compositeStrike;
        }
        compositeStrike = NULL;
    }
    if (fStrikeDescs) {
        if (fStrikeDescs != defaultStrikeDescs) {
            delete [] fStrikeDescs;
        }
        fStrikeDescs = NULL;
    }
}

jfloat Strike::getCharAdvanceX(Unicode ch) {
    jfloat retval = 0;
    // REMIND: removing const --- fix method declaration and implementation
    return getGlyphAdvanceX(CharToGlyph(ch));
}

jfloat Strike::getGlyphAdvanceX(int glyphCode) {
    jfloat retval = 0;
    hsGGlyphStrike *theStrike = strike;
    if (!theStrike) {
        theStrike = compositeStrikeForGlyph(glyphCode);
    }
    if (theStrike) {
        hsFixedPoint2 glyphAdvance;
        theStrike->GetMetricsWithImage(glyphCode, nil, &glyphAdvance);
        retval = hsFixedToFloat(glyphAdvance.fX);
    }
    return retval;
}

Boolean
Strike::getMetrics(int glyphCode, hsGGlyph& glyphRef, hsFixedPoint2& advXY)
{
    Boolean retval = false;
    hsGGlyphStrike *theStrike = strike;
    if (!theStrike) {
        theStrike = compositeStrikeForGlyph(glyphCode);
    }
    if (theStrike) {
        theStrike->GetMetricsWithImage(glyphCode, &glyphRef, &advXY);
        retval = true;
    }
    return retval;
}

const void *
Strike::getImage(int glyphCode)
{
    const void *retval = NULL;
    hsGGlyphStrike *theStrike = strike;
    if (!theStrike) {
        theStrike = compositeStrikeForGlyph(glyphCode);
    }
    if (theStrike) {
        retval = theStrike->GetImage(glyphCode);
    }
    return retval;
}

static void useMax(hsFixedPoint2& target, hsFixedPoint2& src) {
    if (src.fX > target.fX) {
        target.fX = src.fX;
    }
    if (src.fY > target.fY) {
        target.fY = src.fY;
    }
}

static void useMin(hsFixedPoint2& target, hsFixedPoint2& src) {
    if (src.fX < target.fX) {
        target.fX = src.fX;
    }
    if (src.fY < target.fY) {
        target.fY = src.fY;
    }
}

void
Strike::GetLineHeight(hsFixedPoint2& ascent,
                      hsFixedPoint2& descent,
                      hsFixedPoint2& baseline,
                      hsFixedPoint2& leading,
                      hsFixedPoint2& maxAdvance)
{
    if (fCachedLineMetrics) {
        ascent      = fAscent;
        descent     = fDescent;
        baseline    = fBaseline;
        leading     = fLeading;
        maxAdvance  = fMaxAdvance;
        return;
    }

    if (strike) {
        strike->GetLineHeight(&ascent, &descent,
                              &baseline, &leading, &maxAdvance);
    } else {
        ascent.fX       = 0;
        ascent.fY       = 0;
        descent.fX      = 0;
        descent.fY      = 0;
        baseline.fX     = 0;
        baseline.fY     = 0;
        leading.fX      = 0;
        leading.fY      = 0;
        maxAdvance.fX   = 0;
        maxAdvance.fY   = 0;
        if (isComposite) {
            CompositeFont *cf = (CompositeFont *)&fFont;
            int numSlots = cf->countSlots();
            for (int i=0; i < numSlots; i++) {
                hsFixedPoint2 lAscent = {0, 0};
                hsFixedPoint2 lDescent = {0, 0};
                hsFixedPoint2 lBaseline = {0, 0};
                hsFixedPoint2 lLeading = {0, 0};
                hsFixedPoint2 lMaxAdvance = {0, 0};

                hsGGlyphStrike *currentStrike = compositeStrikeForSlot(i);
                if (currentStrike) {
                    currentStrike->GetLineHeight(&lAscent, &lDescent,
                                                &lBaseline,
                                                &lLeading, &lMaxAdvance);
                }
                useMin(ascent,      lAscent);
                useMax(descent,     lDescent);
                useMax(baseline,    lBaseline);
                useMin(leading,     lLeading);
                useMax(maxAdvance,  lMaxAdvance);
            }
        }
    }
    fCachedLineMetrics = JNI_TRUE;
    fAscent      = ascent;
    fDescent     = descent;
    fBaseline    = baseline;
    fLeading     = leading;
    fMaxAdvance  = maxAdvance;
}

void
Strike::GetItalicAngle(hsFixedPoint2& italicAngle)
{
    hsGGlyphStrike *theStrike = strike;
    if (!theStrike) {
		int glyphCode = 0;
        theStrike = compositeStrikeForGlyph(glyphCode);
    }
    if (theStrike) {
        theStrike->GetItalicAngle(&italicAngle);
    } else {
        italicAngle.fX = 0;
        italicAngle.fY = 0;
    }
}

void
Strike::GetGlyphOutline(int glyphCode, hsPathSpline& path)
{
    hsGGlyphStrike *theStrike = strike;
    if (!theStrike) {
        theStrike = compositeStrikeForGlyph(glyphCode);
    }
    if (theStrike) {
        theStrike->GetPath(glyphCode, &path, nil);
    }
}

int Strike::GetNumGlyphs()
{
    int retval = 0;
    if (strike) {
        retval = strike->CountGlyphs();
    } else if (isComposite) {
        if (!totalCompositeGlyphs) {
            CompositeFont *cf = (CompositeFont *)&fFont;
            int numSlots = cf->countSlots();
            for (int i=0; i < numSlots; i++) {
                hsGGlyphStrike *slotStrike = compositeStrikeForSlot(i);
                if (slotStrike) {
                    totalCompositeGlyphs += slotStrike->CountGlyphs();
                }
            }
        }
        retval = totalCompositeGlyphs;
    }
    return retval;
}

void
Strike::allocateStrikes()
{
    if (!compositeStrike) {
        CompositeFont *cf = (CompositeFont *)&fFont;
        int numSlots = cf->countSlots();
        if (numSlots <= DEFAULT_STRIKE_SLOTS) {
            compositeStrike = defaultStrikes;
        } else {
            compositeStrike = new Strike *[numSlots];
        }
        if (compositeStrike) {
            for (int i=0; i < numSlots; i++) {
                compositeStrike[i] = NULL;
            }
        }
    }
}

void
Strike::allocateDescs()
{
    if (!fStrikeDescs) {
        CompositeFont *cf = (CompositeFont *)&fFont;
        int numSlots = cf->countSlots();
        if (numSlots <= DEFAULT_STRIKE_SLOTS) {
            fStrikeDescs = defaultStrikeDescs;
        } else {
            fStrikeDescs = new FontStrikeDesc[numSlots];
        }
        if (fStrikeDescs) {
            memset(fStrikeDescs, 0, numSlots * sizeof(FontStrikeDesc));
        }
    }
}

static int
CompAsciiUnicode(const char *aStr, const Unicode *uStr, int len) {
    if (!len || !aStr || !uStr) {
        return 0;
    }
    while (*aStr && len--) {
        if (*aStr++ != *uStr++) {
            return 0;
        }
    }
    return (*aStr == 0);
}

int
Strike::algorithmicStyle(
        const CompositeFont& cf, const fontObject& fo, int slot) {
    int cfNameLen;
    const Unicode *cfName = ((fontObject&)cf).GetFontName(cfNameLen);
    if (!cfName) {
        return java_awt_Font_PLAIN;
    }
    for (int index=0; index < cfNameLen; index++) {
        if (cfName[index] == '.') {
            // it is a styled composite font
            fontObject *foPlain =
                ::FindFontObject(cfName, index, java_awt_Font_PLAIN);
            if (foPlain && (foPlain->GetFormat() == kCompositeFontFormat)) {
                // and root is also composite font
                CompositeFont *cfPlain = (CompositeFont *)foPlain;
                fontObject *foPlainSlot = cfPlain->getSlotFont(slot);
                if (foPlainSlot == &fo) {
                    // and root slot == styled slot
                    if (CompAsciiUnicode(".bolditalic",
                                    &cfName[index], cfNameLen-index)) {
                        return java_awt_Font_BOLD | java_awt_Font_ITALIC;
                    }
                    if (CompAsciiUnicode(".bold",
                                    &cfName[index], cfNameLen-index)) {
                        return java_awt_Font_BOLD;
                    }
                    if (CompAsciiUnicode(".italic",
                                    &cfName[index], cfNameLen-index)) {
                        return java_awt_Font_ITALIC;
                    }
                }
            }
            return java_awt_Font_PLAIN;
        }
    }
    return java_awt_Font_PLAIN;
}

hsGGlyphStrike *
Strike::compositeStrikeForSlot(int slot)
{
    hsGGlyphStrike *retval = NULL;
    CompositeFont *cf = (CompositeFont *)&fFont;
    int numSlots = cf->countSlots();
    if ((slot < 0) || (slot >= numSlots)) {
        return retval;
    }
    if (!compositeStrike) {
        allocateStrikes();
        if (!compositeStrike) {
            return retval;
        }
    }
    Strike *slotStrike = compositeStrike[slot];
    if (!slotStrike) {
        if (!fStrikeDescs) {
            allocateDescs();
        }
        if (!fStrikeDescs) {
            return retval;
        }
        FontStrikeDesc& desc = fStrikeDescs[slot];
        if (!desc.fFontObj) {
            desc = templateDesc;
            desc.fFontObj = cf->getSlotFont(slot);
            if (desc.fFontObj) {
                // if this is a styled composite font and this
                // slot's fontObject is the same as the "plain"
                // equivalent composite font, apply algorithmic
                // styles with a transform.
                FontTransform tx(desc.fMatrix);
                desc.fFontObj->m_currentStyle = cf->m_currentStyle;
                if (desc.fFontObj->m_currentStyle == java_awt_Font_PLAIN) {
                    desc.fFontObj->m_currentStyle = 
                                algorithmicStyle(*cf, *desc.fFontObj, slot);
                }
                compositeStrike[slot] = new Strike(*desc.fFontObj, tx,
                                                      desc.fDoAntiAlias,
                                                      desc.fDoFracEnable);
            }
        }
        slotStrike = compositeStrike[slot];
    }
    if (slotStrike) {
        retval = slotStrike->strike;
    }
    return retval;
}

hsGGlyphStrike *
Strike::compositeStrikeForGlyph(int& glyphCode)
{
    hsGGlyphStrike *retval = NULL;
    if (glyphCode >= 0) {
        if (glyphCode < baseGlyphs) {
            // performance optimization
            retval = compositeStrike[0]->strike;
        } else if (glyphCode & SLOTMASK) {
            // assumes glyph code contains slot number
            retval = compositeStrikeForSlot(GLYPHSLOT(glyphCode));
            glyphCode &= GLYPHMASK;
        } else {
            // assumes glyph ranges for component fonts have been concatenated
            CompositeFont *cf = (CompositeFont *)&fFont;
            int numSlots = cf->countSlots();
            int slotGlyphs = 0;
            int slot = 0;
            for (; slot < numSlots; slot++) {
                hsGGlyphStrike *slotStrike = compositeStrikeForSlot(slot);
                int slotGlyphs = slotStrike->CountGlyphs();
                if (glyphCode < slotGlyphs) {
                    break;
                }
                glyphCode -= slotGlyphs;
            }
            if (!baseGlyphs && !slot) {
                baseGlyphs = slotGlyphs;
            }
            retval = compositeStrikeForSlot(slot);
        }
    }
    return retval;
}

Boolean Strike::isStyleSupported(int aStyle)
{
    Boolean retCode = false;
    if ( NULL != strike ) {
        retCode = strike->isStyleSupported(aStyle);
    }
    return retCode;
}

UInt32 Strike::CharToGlyph(Unicode unicode)
{
    if (getMapper()) {
        UInt32 glyph;
        fMapper->CharsToGlyphs(1, &unicode, &glyph);
        return glyph;
    }
    return getMissingGlyphCode();
}

int Strike::getMissingGlyphCode() const
{
    if (getMapper()) {
        return fMapper->getMissingGlyphCode();
    }
    return 0;
}

Boolean Strike::canDisplay(Unicode ch)
{
    if (getMapper()) {
        return fMapper->canDisplay(ch);
    }
    return 0;
}

CharToGlyphMapper *Strike::getMapper() const
{
    if (fMapper) {
        return fMapper;
    }
    // override const for private cached readonly value
    Strike *myPtr = (Strike *)this;
    hsGGlyphStrike *theStrike = strike;
    if (!theStrike) {
        myPtr->fMapper = new CompositeGlyphMapper(
                                    (CompositeFont&)fFont, *this);
    } else {
        myPtr->fMapper = theStrike->getMapper();
    }
    return fMapper;
}

UInt32
CompositeGlyphMapper::getCachedGlyphCode(const Unicode unicode) const
{
    UInt32 hiByte = (unicode >> 8) & 0xff;
    if (fGlyphMaps[hiByte]) {
        UInt32 loByte = unicode & 0xff;
        if (fGlyphMaps[hiByte][loByte].fGlyphCode) {
            return fGlyphMaps[hiByte][loByte].fGlyphCode;
        }
    }
    return getMissingGlyphCode();
}

void
CompositeGlyphMapper::setCachedGlyphCode(const Unicode unicode,
                                         const UInt32 glyphCode,
                                         const int slot) const
{
    // not really const ... but changes no external behavior
    // except to boost performance...we override const to maintain
    // read-only cached state
    CompositeGlyphMapper *myPtr = (CompositeGlyphMapper *)this;
    UInt32 hiByte = (unicode >> 8) & 0xff;
    if (!fGlyphMaps[hiByte]) {
        myPtr->fGlyphMaps[hiByte] = new GlyphMapEntry[256];
        if (fGlyphMaps[hiByte]) {
            memset(myPtr->fGlyphMaps[hiByte], 0, 256 * sizeof(GlyphMapEntry));
        }
    }
    if (fGlyphMaps[hiByte]) {
        UInt32 loByte = unicode & 0xff;
        myPtr->fGlyphMaps[hiByte][loByte].fSlot      = slot;
        myPtr->fGlyphMaps[hiByte][loByte].fGlyphCode =
                                                GLYPHCODE(slot, glyphCode);
    }
}

void
CompositeGlyphMapper::CharsToGlyphs(
        int count, const Unicode unicodes[],
        UInt32 glyphs[]) const
{
    getMissingGlyphCode();
    if (count) {
        Strike& myStrike = (Strike&)fStrike;
        for (int i=0; i<count; i++) {

            glyphs[i] = getCachedGlyphCode(unicodes[i]);
            if (glyphs[i] != fMissingGlyph) {
                continue;
            }

            for (int slot = 0; slot < fNumSlots; slot++) {
                hsGGlyphStrike *slotStrike =
                                    myStrike.compositeStrikeForSlot(slot);
                if (    slotStrike
                        && !fCompositeFont.isExcludedChar(slot, unicodes[i]))
                {
                    int glyphCode = fMissingGlyph;
                    CharToGlyphMapper *strikeMapper =
                                            slotStrike->getMapper();
                    if (strikeMapper) {
                        glyphCode = strikeMapper->CharToGlyph(unicodes[i]);
                    }
                    glyphs[i] = GLYPHCODE(slot, glyphCode);
                    if (glyphCode != fMissingGlyph) {
                        setCachedGlyphCode(unicodes[i], glyphCode, slot);
                        break;
                    }
                }
            }
        }
    }
}

Boolean
CompositeGlyphMapper::canDisplay(Unicode ch) const
{
    UInt32 glyph;
    CharsToGlyphs(1, &ch, &glyph);
    return (glyph != getMissingGlyphCode());
}

int
CompositeGlyphMapper::getMissingGlyphCode() const
{
    if (-1 == fMissingGlyph) {
        // remove const for cached read only init
        CompositeGlyphMapper *myPtr = (CompositeGlyphMapper *)this;
        myPtr->fMissingGlyph = 0;
        Strike& myStrike = (Strike&)fStrike;
        hsGGlyphStrike *strike = myStrike.compositeStrikeForSlot(0);
        if (strike) {
            CharToGlyphMapper *strikeMapper = strike->getMapper();
            if (strikeMapper) {
                myPtr->fMissingGlyph = strikeMapper->getMissingGlyphCode();
            }
        }
    }
    return fMissingGlyph;
}
