/*
 * @(#)TextRendering.cpp	1.26 98/10/01
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include <img_util_md.h>
#include <stdio.h>
#include "FontGlue.h"
#include "GlyphVector.h"
#include "BlitMacros.h"


// REMIND: speedup if we write a "hsDesc" which can live on the stack

typedef UInt8 ThreeByte;
typedef UInt8Buffer ThreeByteBuffer;

struct IndexedColorData {
    jint    fgColor;
    const unsigned int *dstLut;
    const char *inv_cmap;
	const ColorData *cData;
    const unsigned int *inverseGrayLut;
};


#define OPAQUE_RASTER_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, jfloat x, jfloat y, \
        jobject theFont, jdoubleArray matrix, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jint offset, DST_TYPE dst, jint pixStr, jint scanStr)

#define OPAQUE_IMAGEDATA_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, jfloat x, jfloat y, \
        jobject theFont, jdoubleArray matrix, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jobject imageData)

#define OPAQUE_RASTER_ARRAY_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, \
        SRC_TYPE theArray, jint offsetSrc, jint count, jfloat x, jfloat y, \
        jobject theFont, jdoubleArray matrix, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jint offset, DST_TYPE dst, jint pixStr, jint scanStr)

#define OPAQUE_IMAGEDATA_ARRAY_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, \
        SRC_TYPE theArray, jint offsetSrc, jint count, jfloat x, jfloat y, \
        jobject theFont, jdoubleArray matrix, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jobject imageData)

#define OPAQUE_GLYPH_RASTER_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, \
        jfloatArray positions, \
        jfloat x, jfloat y, \
        jobject theFont, jdoubleArray matrix, \
        jdoubleArray devTX, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jint offset, DST_TYPE dst, jint pixStr, jint scanStr)

#define OPAQUE_GLYPH_IMAGEDATA_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, \
        jfloatArray positions, \
        jfloat x, jfloat y, \
        jobject theFont, jdoubleArray matrix, \
        jdoubleArray devTX, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jobject imageData)

#define ALPHA_RASTER_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, jfloat x, jfloat y, \
        jobject theFont, jboolean usesFractionalMetrics, jdoubleArray matrix, \
        jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jint offset, DST_TYPE dst, jint pixStr, jint scanStr)

#define ALPHA_IMAGEDATA_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, jfloat x, jfloat y, \
        jobject theFont, jboolean usesFractionalMetrics, jdoubleArray matrix, \
        jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jobject imageData)

#define ALPHA_RASTER_ARRAY_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, \
        SRC_TYPE theArray, jint offsetSrc, jint count, jfloat x, jfloat y, \
        jobject theFont, jboolean usesFractionalMetrics, jdoubleArray matrix, \
        jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jint offset, DST_TYPE dst, jint pixStr, jint scanStr)

#define ALPHA_IMAGEDATA_ARRAY_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, \
        SRC_TYPE theArray, jint offsetSrc, jint count, jfloat x, jfloat y, \
        jobject theFont, jboolean usesFractionalMetrics, jdoubleArray matrix, \
        jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jobject imageData)

#define ALPHA_GLYPH_RASTER_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, \
        jfloatArray positions, \
        jfloat x, jfloat y, \
        jobject theFont, jboolean usesFractionalMetrics, jdoubleArray matrix, \
        jdoubleArray devTX, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jint offset, DST_TYPE dst, jint pixStr, jint scanStr)

#define ALPHA_GLYPH_IMAGEDATA_ARGS(SRC_TYPE, DST_TYPE) \
        (JNIEnv *env, jclass clsNFW, SRC_TYPE theString, \
        jfloatArray positions, \
        jfloat x, jfloat y, \
        jobject theFont, jboolean usesFractionalMetrics, jdoubleArray matrix, \
        jdoubleArray devTX, jint fgColor, \
        jint cx, jint cy, jint cw, jint ch, \
        jobject imageData)

#define IMAGEDATA_REF(PTR_TYPE, X, Y, MATRIX, AAHINT, FMHINT) \
        gv.positionGlyphsWithImages(X, Y, MATRIX, AAHINT, FMHINT); \
        /* REMIND: make cx, cy macro args */ \
        ImageDataWrapper idw( env, imageData, \
                                AAHINT, sizeof(PTR_TYPE), cx, cy); \
        if (idw.invalid()) { \
            /* warning only */ \
            /* JNU_ThrowIllegalArgumentException(env, "Image Data"); */ \
            return; \
        } \
        PTR_TYPE *topLeftPtr = (PTR_TYPE *)idw.topLeftPtr(); \
        unsigned int pixStr = idw.pixStr(); \
        unsigned int scanStr = idw.scanStr();

#define RASTER_DATA_REF(PTR_TYPE, X, Y, MATRIX, AAHINT, FMHINT) \
        gv.positionGlyphsWithImages(X, Y, MATRIX, AAHINT, FMHINT); \
        PTR_TYPE ## Buffer dstSurface(env, dst, offset); \
        PTR_TYPE *topLeftPtr = dstSurface.buffer();

#define GLYPH_IMAGEDATA_REF(PTR_TYPE, X, Y, MATRIX, AAHINT, FMHINT) \
        gv.setDevTX(devTX); \
        gv.setPositions(positions); \
        IMAGEDATA_REF(PTR_TYPE, X, Y, MATRIX, AAHINT, FMHINT);

#define GLYPH_RASTER_DATA_REF(PTR_TYPE, X, Y, MATRIX, AAHINT, FMHINT) \
        gv.setDevTX(devTX); \
        gv.setPositions(positions); \
        RASTER_DATA_REF(PTR_TYPE, X, Y, MATRIX, AAHINT, FMHINT);

#define OPAQUE_TEXT_RENDER(SRC_TYPE, ARGS, IMAGE_REF, DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theString, matrix, false, false, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, false, false); \
            OpaqueTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, fgColor); \
        } \
	NativeUnlock (env); \
    }

#define OPAQUE_TEXT_ARRAY_RENDER(SRC_TYPE, ARGS, \
                                IMAGE_REF, DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theArray, offsetSrc, count, \
                            matrix, false, false, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, false, false); \
            OpaqueTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, fgColor); \
        } \
	NativeUnlock (env); \
    }

#define ALPHA_TEXT_RENDER(SRC_TYPE, ARGS, IMAGE_REF, \
                            DST_FORMAT, DST_TYPE, DST_NAME) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theString, \
                            matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            DST_NAME ## TextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, fgColor); \
        } \
	NativeUnlock (env); \
    }

#define ALPHA_TEXT_ARRAY_RENDER(SRC_TYPE, ARGS, IMAGE_REF, \
                            DST_FORMAT, DST_TYPE, DST_NAME) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theArray, offsetSrc, count, \
                matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            DST_NAME ## TextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, fgColor); \
        } \
	NativeUnlock (env); \
    }

#define ALPHA_TEXT_RENDER_INDEXED(SRC_TYPE, ARGS, IMAGE_REF, \
                                  DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theString, \
                            matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            IndexedColorData icd; \
            icd.fgColor = fgColor; \
            icd.dstLut  = idw.getDstLut(); \
            icd.inv_cmap = idw.getInvCmap(); \
            icd.cData = idw.getCData(); \
            AlphaIndexedTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, icd); \
        } \
	NativeUnlock (env); \
    }

#define ALPHA_TEXT_ARRAY_RENDER_INDEXED(SRC_TYPE, ARGS, IMAGE_REF, \
                                        DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theArray, offsetSrc, count, \
                matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            IndexedColorData icd; \
            icd.fgColor = fgColor; \
            icd.dstLut  = idw.getDstLut(); \
            icd.inv_cmap = idw.getInvCmap(); \
            icd.cData = idw.getCData(); \
            AlphaIndexedTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, icd); \
        } \
	NativeUnlock (env); \
    }

#define ALPHA_TEXT_RENDER_INDEX8GRAY(SRC_TYPE, ARGS, IMAGE_REF, \
                                     DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theString, \
                            matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            IndexedColorData icd; \
            icd.fgColor = fgColor; \
            icd.dstLut  = idw.getDstLut(); \
            icd.inv_cmap = idw.getInvCmap(); \
            icd.cData = idw.getCData(); \
            icd.inverseGrayLut = idw.getinverseGrayLut(); \
            AlphaIndex8GrayTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, icd); \
        } \
	NativeUnlock (env); \
    }

#define ALPHA_TEXT_ARRAY_RENDER_INDEX8GRAY(SRC_TYPE, ARGS, IMAGE_REF, \
                                           DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theArray, offsetSrc, count, \
                matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            IndexedColorData icd; \
            icd.fgColor = fgColor; \
            icd.dstLut  = idw.getDstLut(); \
            icd.inv_cmap = idw.getInvCmap(); \
            icd.cData = idw.getCData(); \
            icd.inverseGrayLut = idw.getinverseGrayLut(); \
            AlphaIndex8GrayTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, icd); \
        } \
	NativeUnlock (env); \
    }
#define ALPHA_TEXT_RENDER_BYTEGRAY(SRC_TYPE, ARGS, IMAGE_REF, \
                                  DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theString, \
                            matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            AlphaByteGrayTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, fgColor); \
        } \
	NativeUnlock (env); \
    }

#define ALPHA_TEXT_ARRAY_RENDER_BYTEGRAY(SRC_TYPE, ARGS, IMAGE_REF, \
                                        DST_FORMAT, DST_TYPE) \
    ARGS(SRC_TYPE, DST_TYPE) { \
	NativeLock (env); \
        fontObject *fo = getFontPtr(env, theFont); \
        if (fo) { \
            GlyphVector gv(env, theArray, offsetSrc, count, \
                matrix, true, usesFractionalMetrics, fo); \
            if (gv.needShaping()) { \
                JNU_ThrowByName(env, "sun/awt/font/ShapingException", \
                                            "Shaping may be needed"); \
		NativeUnlock (env); \
                return; \
            } \
            IMAGE_REF(DST_FORMAT, x, y, matrix, true, usesFractionalMetrics);\
            AlphaByteGrayTextRendering ## DST_FORMAT(gv, topLeftPtr, \
                                            cx, cy, cw, ch, \
                                            scanStr, pixStr, fgColor); \
        } \
	NativeUnlock (env); \
    }

#define OPAQUE_TEXT_TO_IMAGEDATA(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        OPAQUE_TEXT_RENDER(SRC_TYPE, OPAQUE_IMAGEDATA_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define OPAQUE_TEXT_TO_RASTER(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        OPAQUE_TEXT_RENDER(SRC_TYPE, OPAQUE_RASTER_ARGS, \
                            RASTER_DATA_REF, DST_FORMAT, DST_TYPE)

#define OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        OPAQUE_TEXT_ARRAY_RENDER(SRC_TYPE, OPAQUE_IMAGEDATA_ARRAY_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define OPAQUE_TEXT_ARRAY_TO_RASTER(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        OPAQUE_TEXT_ARRAY_RENDER(SRC_TYPE, OPAQUE_RASTER_ARRAY_ARGS, \
                            RASTER_DATA_REF, DST_FORMAT, DST_TYPE)

#define OPAQUE_GLYPHS_TO_RASTER(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        OPAQUE_TEXT_RENDER(SRC_TYPE, OPAQUE_GLYPH_RASTER_ARGS, \
                            GLYPH_RASTER_DATA_REF, DST_FORMAT, DST_TYPE)

#define OPAQUE_GLYPHS_TO_IMAGEDATA(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        OPAQUE_TEXT_RENDER(SRC_TYPE, OPAQUE_GLYPH_IMAGEDATA_ARGS, \
                            GLYPH_IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_TEXT_TO_IMAGEDATA(SRC_TYPE, DST_FORMAT, DST_TYPE, DST_NAME) \
        ALPHA_TEXT_RENDER(SRC_TYPE, ALPHA_IMAGEDATA_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE, DST_NAME)

#define ALPHA_TEXT_TO_IMAGEDATA_INDEXED(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_INDEXED(SRC_TYPE, ALPHA_IMAGEDATA_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_TEXT_TO_IMAGEDATA_INDEX8GRAY(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_INDEX8GRAY(SRC_TYPE, ALPHA_IMAGEDATA_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_TEXT_TO_IMAGEDATA_BYTEGRAY(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_BYTEGRAY(SRC_TYPE, ALPHA_IMAGEDATA_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_TEXT_TO_RASTER(SRC_TYPE, DST_FORMAT, DST_TYPE, DST_NAME) \
        ALPHA_TEXT_RENDER(SRC_TYPE, ALPHA_RASTER_ARGS, \
                        RASTER_DATA_REF, DST_FORMAT, DST_TYPE, DST_NAME)

#define ALPHA_TEXT_TO_RASTER_INDEXED(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_INDEXED(SRC_TYPE, ALPHA_RASTER_ARGS, \
                        RASTER_DATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_TEXT_ARRAY_TO_IMAGEDATA(SRC_TYPE, DST_FORMAT, \
                                                    DST_TYPE, DST_NAME) \
        ALPHA_TEXT_ARRAY_RENDER(SRC_TYPE, ALPHA_IMAGEDATA_ARRAY_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE, DST_NAME)

#define ALPHA_TEXT_ARRAY_TO_IMAGEDATA_INDEXED(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_ARRAY_RENDER_INDEXED(SRC_TYPE, ALPHA_IMAGEDATA_ARRAY_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_TEXT_ARRAY_TO_IMAGEDATA_INDEX8GRAY(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_ARRAY_RENDER_INDEX8GRAY(SRC_TYPE, ALPHA_IMAGEDATA_ARRAY_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)
#define ALPHA_TEXT_ARRAY_TO_IMAGEDATA_BYTEGRAY(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_ARRAY_RENDER_BYTEGRAY(SRC_TYPE, ALPHA_IMAGEDATA_ARRAY_ARGS, \
                            IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_TEXT_ARRAY_TO_RASTER(SRC_TYPE, DST_FORMAT, DST_TYPE, DST_NAME) \
        ALPHA_TEXT_ARRAY_RENDER(SRC_TYPE, ALPHA_RASTER_ARRAY_ARGS, \
                        RASTER_DATA_REF, DST_FORMAT, DST_TYPE, DST_NAME)

#define ALPHA_TEXT_ARRAY_TO_RASTER_INDEXED(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_ARRAY_RENDER_INDEXED(SRC_TYPE, ALPHA_RASTER_ARRAY_ARGS, \
                        RASTER_DATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_GLYPHS_TO_IMAGEDATA(SRC_TYPE, DST_FORMAT, DST_TYPE, DST_NAME) \
        ALPHA_TEXT_RENDER(SRC_TYPE, ALPHA_GLYPH_IMAGEDATA_ARGS, \
                        GLYPH_IMAGEDATA_REF, DST_FORMAT, DST_TYPE, DST_NAME)

#define ALPHA_GLYPHS_TO_IMAGEDATA_INDEXED(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_INDEXED(SRC_TYPE, ALPHA_GLYPH_IMAGEDATA_ARGS, \
                        GLYPH_IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_GLYPHS_TO_IMAGEDATA_INDEX8GRAY(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_INDEX8GRAY(SRC_TYPE, ALPHA_GLYPH_IMAGEDATA_ARGS, \
                        GLYPH_IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_GLYPHS_TO_IMAGEDATA_BYTEGRAY(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_BYTEGRAY(SRC_TYPE, ALPHA_GLYPH_IMAGEDATA_ARGS, \
                        GLYPH_IMAGEDATA_REF, DST_FORMAT, DST_TYPE)

#define ALPHA_GLYPHS_TO_RASTER(SRC_TYPE, DST_FORMAT, DST_TYPE, DST_NAME) \
        ALPHA_TEXT_RENDER(SRC_TYPE, ALPHA_GLYPH_RASTER_ARGS, \
                        GLYPH_RASTER_DATA_REF, DST_FORMAT, DST_TYPE, DST_NAME)

#define ALPHA_GLYPHS_TO_RASTER_INDEXED(SRC_TYPE, DST_FORMAT, DST_TYPE) \
        ALPHA_TEXT_RENDER_INDEXED(SRC_TYPE, ALPHA_GLYPH_RASTER_ARGS, \
                        GLYPH_RASTER_DATA_REF, DST_FORMAT, DST_TYPE)

#define RGB24_TO_THREE_COMPONENTS(s,r,g,b) \
    (r) = (s)[0]; (g) = (s)[1]; (b) = (s)[2];

#define BGR24_TO_THREE_COMPONENTS(s,r,g,b) \
    (r) = (s)[2]; (g) = (s)[1]; (b) = (s)[0];

// all alpha blends have ARGB pen color.
#define ALPHA_ARGB_PIXELS_INIT(PEN_COLOR) \
        int a, r, g, b; \
        ARGB_TO_FOUR_COMPONENTS(&PEN_COLOR,r,g,b,a);
#define ALPHA_3BYTE_RGB_PIXELS_INIT     ALPHA_ARGB_PIXELS_INIT
#define ALPHA_BGR_PIXELS_INIT           ALPHA_ARGB_PIXELS_INIT
#define ALPHA_3BYTE_BGR_PIXELS_INIT     ALPHA_ARGB_PIXELS_INIT
#define ALPHA_565_PIXELS_INIT           ALPHA_ARGB_PIXELS_INIT
#define ALPHA_555_PIXELS_INIT           ALPHA_ARGB_PIXELS_INIT

#define ALPHA_GRAY_PIXELS_INIT(PEN_COLOR) \
        int a, r, g, b; \
        ARGB_TO_FOUR_COMPONENTS(&p,r,g,b,a); \

#define ALPHA_INDEXED8_PIXELS_INIT(COLORDATA)       \
    jint fgColor = COLORDATA.fgColor;               \
    const unsigned int *srcLut = COLORDATA.dstLut;  \
    IndexedColorData& dstLockInfo = COLORDATA;      \
    const ColorData *cData = COLORDATA.cData;              \
    ALPHA_ARGB_PIXELS_INIT(fgColor);

#define ALPHA_INDEX8GRAY_PIXELS_INIT(COLORDATA)       \
    jint fgColor = COLORDATA.fgColor;               \
    const unsigned int *srcLut = COLORDATA.dstLut;  \
    IndexedColorData& dstLockInfo = COLORDATA;      \
    const ColorData *cData = COLORDATA.cData;              \
    const unsigned int *inverseLut = COLORDATA.inverseGrayLut;  \
    ALPHA_ARGB_PIXELS_INIT(fgColor);

#define ALPHA_ARGB_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstA, dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        ARGB_TO_FOUR_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB, dstA); \
        dstA = (dstA * mixValDst + a * mixValSrc) / 255; \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        DST_PIXELS[PIXEL_INDEX] = \
            (dstA << 24) | (dstR << 16) | (dstG << 8) | dstB; \
    } while (0);

#define ALPHA_3BYTE_RGB_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        RGB24_TO_THREE_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX*3, dstR, dstG, dstB); \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        THREE_COMPONENTS_TO_RGB_24BIT( \
            DST_PIXELS + PIXEL_INDEX*3, dstR, dstG, dstB); \
    } while (0);


#define BGR_TO_THREE_COMPONENTS(s,r,g,b) \
    do { \
        unsigned int p = *(s); \
        (r) = p & 0xff; (g) = (p >> 8)  & 0xff; (b) = (p >> 16) & 0xff; \
    } while (0);

#define ALPHA_BGR_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        BGR_TO_THREE_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB); \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        THREE_COMPONENTS_TO_XBGR( \
            DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB); \
    } while (0);

#define ALPHA_3BYTE_BGR_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        BGR24_TO_THREE_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX*3, dstR, dstG, dstB); \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        THREE_COMPONENTS_TO_BGR_24BIT( \
            DST_PIXELS + PIXEL_INDEX*3, dstR, dstG, dstB); \
    } while (0);

#define ALPHA_565_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstA, dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        SHORT_565_TO_FOUR_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB, dstA); \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        THREE_COMPONENTS_TO_565(DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB) \
    } while (0);

#define ALPHA_555_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstA, dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        SHORT_555_TO_FOUR_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB, dstA); \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        THREE_COMPONENTS_TO_555(DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB) \
    } while (0);

#define ALPHA_INDEXED8_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        BIDX_TO_THREE_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB); \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        *(DST_PIXELS + PIXEL_INDEX) = CUBEMAP(dstR, dstG, dstB); \
    } while (0);
#define ALPHA_INDEX8GRAY_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstR, dstG, dstB; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        BIDX_TO_THREE_COMPONENTS( \
            DST_PIXELS + PIXEL_INDEX, dstR, dstG, dstB); \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        unsigned char gray_val = (unsigned  char ) (dstR*0.299 + dstG*.587 + dstB * .114 ); \
        * (DST_PIXELS + PIXEL_INDEX ) = inverseLut[gray_val]; \
    } while (0);

#define ALPHA_GRAY_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    do { \
        unsigned int dstR, dstG, dstB, dstA; \
        unsigned int mixValSrc = GLYPH_PIXELS[PIXEL_INDEX]; \
        unsigned int mixValDst = 255 - mixValSrc; \
        unsigned char p1 = *(DST_PIXELS + PIXEL_INDEX); \
        dstR = dstG = dstB = p1; \
        dstR = (dstR * mixValDst + r * mixValSrc) / 255; \
        dstG = (dstG * mixValDst + g * mixValSrc) / 255; \
        dstB = (dstB * mixValDst + b * mixValSrc) / 255; \
        *(DST_PIXELS + PIXEL_INDEX) = (unsigned char)(dstR*0.299 + dstG*.587 + dstB*.114); \
    } while (0);

#define OPAQUE_PIXELS_INIT(PEN_COLOR)

#define OPAQUE_PIXELS(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    if (GLYPH_PIXELS[PIXEL_INDEX]) { \
        DST_PIXELS[PIXEL_INDEX] = FG_COLOR; \
    }

#define OPAQUE_PIXELS_3BYTE_INIT(PEN_COLOR) \
        UInt8 fg0 = (PEN_COLOR >> 16 ) & 0xff; \
        UInt8 fg1 = (PEN_COLOR >>  8 ) & 0xff; \
        UInt8 fg2 = (PEN_COLOR       ) & 0xff;

#define OPAQUE_PIXELS_3BYTE(GLYPH_PIXELS, PIXEL_INDEX, DST_PIXELS, FG_COLOR) \
    if (GLYPH_PIXELS[PIXEL_INDEX]) { \
        DST_PIXELS[PIXEL_INDEX * 3 + 0] = fg0; \
        DST_PIXELS[PIXEL_INDEX * 3 + 1] = fg1; \
        DST_PIXELS[PIXEL_INDEX * 3 + 2] = fg2; \
    }

#define TEXT_RENDERING_LOOP(DST_FORMAT, PIXEL_LOGIC, LOGIC_NAME, COLOR_TYPE) \
void LOGIC_NAME ## TextRendering ## DST_FORMAT(GlyphVector& gv, \
                            DST_FORMAT *dstSurface, \
                            jfloat cx, jfloat cy, jfloat cw, jfloat ch, \
                            jint scanStr,  \
                            jint pixStr,   \
                            COLOR_TYPE colorData) \
{ \
    if (!dstSurface) { \
        return; \
    } \
    int clipLeft    = (cx + 0.5); \
    int clipRight   = (cx + cw + 0.5); \
    int clipTop     = (cy + 0.5); \
    int clipBottom  = (cy + ch + 0.5); \
    int totalGlyphs = gv.getNumGlyphs(); \
    PIXEL_LOGIC ## _INIT(colorData); \
    for (int glyphCounter = 0; glyphCounter < totalGlyphs; glyphCounter++) { \
        float xPos, yPos; \
        const void *pixelPtr; \
        const UInt8 * pixels; \
        unsigned int rowBytes; \
        int width, height; \
        gv.getGlyphPosition(glyphCounter, xPos, yPos); \
        gv.getGlyphBlitInfo(glyphCounter, pixelPtr, rowBytes, width, height); \
        pixels = (const UInt8 *)pixelPtr; \
        if (!pixels) { \
            continue; \
        } \
        /* if any clipping required, modify parameters now */ \
        int left   = (int)xPos; \
        int right  = left + width; \
        int top    = (int)yPos; \
        int bottom = top + height; \
        if (left < clipLeft) { \
            pixels += clipLeft - left; \
            left = clipLeft; \
        } \
        if (top < clipTop) { \
            pixels += (clipTop - top) * rowBytes; \
            top = clipTop; \
        } \
        if (right > clipRight) { \
            right = clipRight; \
        } \
        if (bottom > clipBottom) { \
            bottom = clipBottom; \
        } \
        width = right - left; \
        height = bottom - top; \
        if ((width <= 0) || (height <= 0)) { \
            continue; \
        } \
        DST_FORMAT *dstRow = dstSurface + left * pixStr + top * scanStr; \
        /* REMIND can speedup inner loop by using B&W bits */ \
        while (height--) { \
            for (int i=0; i < width; i++) { \
                PIXEL_LOGIC(pixels, i, dstRow, colorData); \
            } \
            dstRow += scanStr; \
            pixels += rowBytes; \
        } \
    } \
}

TEXT_RENDERING_LOOP(UInt32, OPAQUE_PIXELS, Opaque, jint);
TEXT_RENDERING_LOOP(UInt16, OPAQUE_PIXELS, Opaque, jint);
TEXT_RENDERING_LOOP(UInt8,  OPAQUE_PIXELS, Opaque, jint);
TEXT_RENDERING_LOOP(ThreeByte,  OPAQUE_PIXELS_3BYTE, Opaque, jint);

TEXT_RENDERING_LOOP(UInt8, ALPHA_GRAY_PIXELS, AlphaByteGray, jint);

TEXT_RENDERING_LOOP(UInt32, ALPHA_ARGB_PIXELS, AlphaArgb, jint);
TEXT_RENDERING_LOOP(UInt32, ALPHA_BGR_PIXELS, AlphaBgr, jint);
TEXT_RENDERING_LOOP(UInt16, ALPHA_555_PIXELS, Alpha555, jint);
TEXT_RENDERING_LOOP(UInt16, ALPHA_565_PIXELS, Alpha565, jint);
TEXT_RENDERING_LOOP(ThreeByte, ALPHA_3BYTE_BGR_PIXELS, AlphaBgr24, jint);
TEXT_RENDERING_LOOP(ThreeByte, ALPHA_3BYTE_RGB_PIXELS, AlphaRgb24, jint);
TEXT_RENDERING_LOOP(UInt8, ALPHA_INDEXED8_PIXELS, AlphaIndexed, IndexedColorData);
TEXT_RENDERING_LOOP(UInt8, ALPHA_INDEX8GRAY_PIXELS, AlphaIndex8Gray, IndexedColorData);

extern "C" {
/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringIntDiscreteRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringIntDiscreteRaster
    OPAQUE_TEXT_TO_RASTER(jstring, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsIntDiscreteRaster
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsIntDiscreteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jcharArray, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesIntDiscreteRaster
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesIntDiscreteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jbyteArray, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorIntDiscreteRaster
 * Signature: ([II[FIFFLjava/awt/Font;[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorIntDiscreteRaster
    OPAQUE_GLYPHS_TO_RASTER(jintArray, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringShortDiscreteRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringShortDiscreteRaster
    OPAQUE_TEXT_TO_RASTER(jstring, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsShortDiscreteRaster
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsShortDiscreteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jcharArray, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesShortDiscreteRaster
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesShortDiscreteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jbyteArray, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorShortDiscreteRaster
 * Signature: ([II[FIFFLjava/awt/Font;[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorShortDiscreteRaster
    OPAQUE_GLYPHS_TO_RASTER(jintArray, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringByteDiscreteRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringByteDiscreteRaster
    OPAQUE_TEXT_TO_RASTER(jstring, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsByteDiscreteRaster
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsByteDiscreteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jcharArray, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesByteDiscreteRaster
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIII[BII)V
  */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesByteDiscreteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jbyteArray, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorByteDiscreteRaster
 * Signature: ([II[FIFFLjava/awt/Font;[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorByteDiscreteRaster
    OPAQUE_GLYPHS_TO_RASTER(jintArray, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringIntDiscrete
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringIntDiscrete
    OPAQUE_TEXT_TO_IMAGEDATA(jstring, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsIntDiscrete
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsIntDiscrete
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesIntDiscrete
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesIntDiscrete
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorIntDiscrete
 * Signature: ([II[FIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorIntDiscrete
    OPAQUE_GLYPHS_TO_IMAGEDATA(jintArray, UInt32, jintArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringShortDiscrete
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringShortDiscrete
    OPAQUE_TEXT_TO_IMAGEDATA(jstring, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsShortDiscrete
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsShortDiscrete
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesShortDiscrete
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesShortDiscrete
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorShortDiscrete
 * Signature: ([II[FIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorShortDiscrete
    OPAQUE_GLYPHS_TO_IMAGEDATA(jintArray, UInt16, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringByteDiscrete
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringByteDiscrete
    OPAQUE_TEXT_TO_IMAGEDATA(jstring, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsByteDiscrete
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsByteDiscrete
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesByteDiscrete
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesByteDiscrete
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorByteDiscrete
 * Signature: ([II[FIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorByteDiscrete
    OPAQUE_GLYPHS_TO_IMAGEDATA(jintArray, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringARGBRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringARGBRaster
    ALPHA_TEXT_TO_RASTER(jstring, UInt32, jintArray, AlphaArgb)


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsARGBRaster
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsARGBRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jcharArray, UInt32, jintArray, AlphaArgb)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesARGBRaster
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesARGBRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jbyteArray, UInt32, jintArray, AlphaArgb)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorARGBRaster
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorARGBRaster
    ALPHA_GLYPHS_TO_RASTER(jintArray, UInt32, jintArray, AlphaArgb)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringARGB
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringARGB
    ALPHA_TEXT_TO_IMAGEDATA(jstring, UInt32, jintArray, AlphaArgb)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsARGB
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsARGB
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, UInt32, jintArray, AlphaArgb)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesARGB
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesARGB
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, UInt32, jintArray, AlphaArgb)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorARGB
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorARGB
      ALPHA_GLYPHS_TO_IMAGEDATA(jintArray, UInt32, jintArray, AlphaArgb)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringBGRRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringBGRRaster
    ALPHA_TEXT_TO_RASTER(jstring, UInt32, jintArray, AlphaBgr)


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsBGRRaster
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsBGRRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jcharArray, UInt32, jintArray, AlphaBgr)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesBGRRaster
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesBGRRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jbyteArray, UInt32, jintArray, AlphaBgr)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorBGRRaster
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIII[III)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorBGRRaster
    ALPHA_GLYPHS_TO_RASTER(jintArray, UInt32, jintArray, AlphaBgr)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringBGR
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringBGR
    ALPHA_TEXT_TO_IMAGEDATA(jstring, UInt32, jintArray, AlphaBgr)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsBGR
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsBGR
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, UInt32, jintArray, AlphaBgr)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesBGR
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesBGR
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, UInt32, jintArray, AlphaBgr)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorBGR
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorBGR
      ALPHA_GLYPHS_TO_IMAGEDATA(jintArray, UInt32, jintArray, AlphaBgr)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString555Raster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString555Raster
    ALPHA_TEXT_TO_RASTER(jstring, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars555Raster
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars555Raster
    ALPHA_TEXT_ARRAY_TO_RASTER(jcharArray, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes555Raster
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes555Raster
    ALPHA_TEXT_ARRAY_TO_RASTER(jbyteArray, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector555Raster
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector555Raster
    ALPHA_GLYPHS_TO_RASTER(jintArray, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString555
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString555
    ALPHA_TEXT_TO_IMAGEDATA(jstring, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars555
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars555
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes555
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes555
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector555
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector555
    ALPHA_GLYPHS_TO_IMAGEDATA(jintArray, UInt16, jshortArray, Alpha555)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString565Raster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString565Raster
    ALPHA_TEXT_TO_RASTER(jstring, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars565Raster
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars565Raster
    ALPHA_TEXT_ARRAY_TO_RASTER(jcharArray, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes565Raster
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes565Raster
    ALPHA_TEXT_ARRAY_TO_RASTER(jbyteArray, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector565Raster
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIII[SII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector565Raster
    ALPHA_GLYPHS_TO_RASTER(jintArray, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString565
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString565
    ALPHA_TEXT_TO_IMAGEDATA(jstring, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars565
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars565
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes565
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes565
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector565
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector565
    ALPHA_GLYPHS_TO_IMAGEDATA(jintArray, UInt16, jshortArray, Alpha565)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringIndexed
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringIndexed
    ALPHA_TEXT_TO_IMAGEDATA_INDEXED(jstring, UInt8, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsIndexed
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawCharsIndexed
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA_INDEXED(jcharArray, UInt8, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesIndexed
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytesIndexed
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA_INDEXED(jbyteArray, UInt8, jshortArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorIndexed
 * Signature: ([II[FIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorIndexed
    ALPHA_GLYPHS_TO_IMAGEDATA_INDEXED(jintArray, UInt8, jbyteArray)


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringIndex8Gray
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[DIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_font_NativeFontWrapper_drawStringIndex8Gray
  ALPHA_TEXT_TO_IMAGEDATA_INDEX8GRAY(jstring, UInt8, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsIndex8Gray
 * Signature: ([CIIFFLjava/awt/Font;Z[DIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_font_NativeFontWrapper_drawCharsIndex8Gray
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA_INDEX8GRAY(jcharArray, UInt8, jbyteArray)


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytesIndex8Gray
 * Signature: ([BIIFFLjava/awt/Font;Z[DIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_font_NativeFontWrapper_drawBytesIndex8Gray
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA_INDEX8GRAY(jbyteArray, UInt8, jbyteArray)


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVectorIndex8Gray
 * Signature: ([I[FFFLjava/awt/Font;Z[D[DIIIIILsun/java2d/loops/ImageData;)V
 */ 
JNIEXPORT void JNICALL 
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorIndex8Gray
    ALPHA_GLYPHS_TO_IMAGEDATA_INDEX8GRAY(jintArray, UInt8, jbyteArray)


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringByteGray
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[DIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_font_NativeFontWrapper_drawStringByteGray
    ALPHA_TEXT_TO_IMAGEDATA_BYTEGRAY(jstring, UInt8, jbyteArray)


/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawCharsByteGray
 * Signature: ([CIIFFLjava/awt/Font;Z[DIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_font_NativeFontWrapper_drawCharsByteGray
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA_BYTEGRAY(jcharArray, UInt8, jbyteArray)

/* 
 * Class: sun_awt_font_NativeFontWrapper 
 * Method: drawBytesByteGray 
 * Signature: ([BIIFFLjava/awt/Font;Z[DIIIIILsun/java2d/loops/ImageData;)V */
JNIEXPORT void JNICALL Java_sun_awt_font_NativeFontWrapper_drawBytesByteGray 
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA_BYTEGRAY(jbyteArray, UInt8, jbyteArray)

/* 
 * Class: sun_awt_font_NativeFontWrapper 
 * Method: drawGlyphVectorByteGray 
 * Signature: ([I[FFFLjava/awt/Font;Z[D[DIIIIILsun/java2d/loops/ImageData;)V */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVectorByteGray 
    ALPHA_GLYPHS_TO_IMAGEDATA_BYTEGRAY(jintArray, UInt8, jbyteArray)
 

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes24BitBGR
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes24BitBGR
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes24BitBGRRaster
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes24BitBGRRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jbyteArray, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes24BitRGB
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes24BitRGB
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes24BitRGBRaster
 * Signature: ([BIIFFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes24BitRGBRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jbyteArray, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes3Byte
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes3Byte
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jbyteArray, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawBytes3ByteRaster
 * Signature: ([BIIFFLjava/awt/Font;[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawBytes3ByteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jbyteArray, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars24BitBGR
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars24BitBGR
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars24BitBGRRaster
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars24BitBGRRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jcharArray, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars24BitRGB
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars24BitRGB
    ALPHA_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars24BitRGBRaster
 * Signature: ([CIIFFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars24BitRGBRaster
    ALPHA_TEXT_ARRAY_TO_RASTER(jcharArray, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars3Byte
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars3Byte
    OPAQUE_TEXT_ARRAY_TO_IMAGEDATA(jcharArray, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawChars3ByteRaster
 * Signature: ([CIIFFLjava/awt/Font;[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawChars3ByteRaster
    OPAQUE_TEXT_ARRAY_TO_RASTER(jcharArray, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector24BitBGR
 * Signature: ([I[FFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector24BitBGR
    ALPHA_GLYPHS_TO_IMAGEDATA(jintArray, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector24BitBGRRaster
 * Signature: ([I[FFFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector24BitBGRRaster
    ALPHA_GLYPHS_TO_RASTER(jintArray, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector24BitRGB
 * Signature: ([I[FFFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector24BitRGB
    ALPHA_GLYPHS_TO_IMAGEDATA(jintArray, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector24BitRGBRaster
 * Signature: ([I[FFFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector24BitRGBRaster
    ALPHA_GLYPHS_TO_RASTER(jintArray, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector3Byte
 * Signature: ([I[FFFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector3Byte
    OPAQUE_GLYPHS_TO_IMAGEDATA(jintArray, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawGlyphVector3ByteRaster
 * Signature: ([I[FFFLjava/awt/Font;[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawGlyphVector3ByteRaster
    OPAQUE_GLYPHS_TO_RASTER(jintArray, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString24BitBGR
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString24BitBGR
    ALPHA_TEXT_TO_IMAGEDATA(jstring, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString24BitBGRRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString24BitBGRRaster
    ALPHA_TEXT_TO_RASTER(jstring, ThreeByte, jbyteArray, AlphaBgr24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString24BitRGB
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString24BitRGB
    ALPHA_TEXT_TO_IMAGEDATA(jstring, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString24BitRGBRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;Z[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString24BitRGBRaster
    ALPHA_TEXT_TO_RASTER(jstring, ThreeByte, jbyteArray, AlphaRgb24)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString3Byte
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIILsun/java2d/loops/ImageData;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString3Byte
    OPAQUE_TEXT_TO_IMAGEDATA(jstring, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawString3ByteRaster
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FIIIIII[BII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_font_NativeFontWrapper_drawString3ByteRaster
    OPAQUE_TEXT_TO_RASTER(jstring, ThreeByte, jbyteArray)

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    drawStringOutline
 * Signature: (Ljava/lang/String;FFLjava/awt/Font;[FZZLjava/awt/Shape;)Ljava/awt/Shape
 */
JNIEXPORT jobject JNICALL
Java_sun_awt_font_NativeFontWrapper_drawStringOutline
        (JNIEnv *env, jclass clsNFW,
        jstring theString, jfloat x, jfloat y, jobject theFont,
        jdoubleArray matrix,
        jboolean isAntiAliased, jboolean usesFractionalMetrics,
        jobject theShape)
{
    NativeLock (env);
    fontObject *fo = getFontPtr(env, theFont);
    if (fo) {
        GlyphVector gv(env, theString,
                        matrix, isAntiAliased, usesFractionalMetrics, fo);
        if (gv.needShaping()) {
            JNU_ThrowByName(env, "sun/awt/font/ShapingException",
                                                    "Shaping may be needed");
	    NativeUnlock (env);
            return theShape;
        }
        gv.positionGlyphs(x, y, matrix, isAntiAliased, usesFractionalMetrics);
        gv.scanShapeInto(theShape, matrix, isAntiAliased, usesFractionalMetrics);
    }
    NativeUnlock (env);
    return theShape;
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getGlyphVectorOutline
 * Signature: (Ljava/awt/font/GlyphVector;Ljava/awt/Font;[FZZFFLjava/awt/Shape;)Ljava/awt/Shape;
 */
JNIEXPORT jobject JNICALL
Java_sun_awt_font_NativeFontWrapper_getGlyphVectorOutline
        (JNIEnv *env, jclass clsNFW,
        jobject glyphVector, jobject theFont,
        jdoubleArray matrix,
        jboolean isAntiAliased, jboolean usesFractionalMetrics,
        jfloat x, jfloat y,
        jobject theShape)
{
    NativeLock (env);
    fontObject *fo = getFontPtr(env, theFont);
    if (fo) {
        GlyphVector gv(env,
                        matrix, isAntiAliased, usesFractionalMetrics, fo);
        if (gv.needShaping()) {
            JNU_ThrowByName(env, "sun/awt/font/ShapingException",
                                                    "Shaping may be needed");
	    NativeUnlock (env);
            return theShape;
        }
        gv.getGlyphCodes(glyphVector);
        gv.getPositions(glyphVector, x, y,
                        matrix, isAntiAliased, usesFractionalMetrics);
        gv.scanShapeInto(theShape, matrix, isAntiAliased, usesFractionalMetrics);
    }
    NativeUnlock (env);
    return theShape;
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    getGlyphOutline
 * Signature: (Ljava/awt/font/GlyphVector;Ljava/awt/Font;[FZZFFLjava/awt/Shape;)Ljava/awt/Shape;
 */
JNIEXPORT jobject JNICALL
Java_sun_awt_font_NativeFontWrapper_getGlyphOutline
        (JNIEnv *env, jclass clsNFW,
        jobject glyphVector, jobject theFont,
        jdoubleArray matrix,
        jboolean isAntiAliased, jboolean usesFractionalMetrics,
        jint glyphIndex,
        jobject theShape)
{
    NativeLock (env);
    fontObject *fo = getFontPtr(env, theFont);
    if (fo) {
        GlyphVector gv(env,
                        matrix, isAntiAliased, usesFractionalMetrics, fo);
        if (gv.needShaping()) {
            JNU_ThrowByName(env, "sun/awt/font/ShapingException",
                                                    "Shaping may be needed");
	    NativeUnlock (env);
            return theShape;
        }
        gv.getGlyphCodes(glyphVector);
        gv.getPositions(glyphVector, 0, 0,
                        matrix, isAntiAliased, usesFractionalMetrics);
        gv.scanShapeInto(theShape, glyphIndex,
                        matrix, isAntiAliased, usesFractionalMetrics);
    }
    NativeUnlock (env);
    return theShape;
}


}; // end of extern "C"

// REMIND: remove each of these when testing of the method is complete
void *functions[] = {
    0
//    Java_sun_awt_font_NativeFontWrapper_getGlyphJustificationInfo,
//    Java_sun_awt_font_NativeFontWrapper_registerFonts,
//    Java_sun_awt_font_NativeFontWrapper_shapeGlyphVector,
};
