/*
 * @(#)cmaps.cpp	1.19 98/09/29
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Alex D. Gelfenbain
 * @author Xueming Shen
 */

#include "fontObject.h"
#include "cmaps.h"
#include "swaps.h"

#include <stdlib.h>
#include <ctype.h>
#include <jni.h>
#include <jni_util.h>
#include <jvm.h>

#include <stdio.h>

extern "C" {
    static uint16 *GetConverter(uint16 );

    JavaVM *jvm;

    JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *vm, void *reserved)
    {
        jvm = vm;
        return JNI_VERSION_1_2;
    }

#define HANDLE_EXCEPTION()            \
    if (env->ExceptionOccurred()) {   \
        env->ExceptionDescribe();     \
        env->ExceptionClear();        \
        if (retArray)                 \
            free(retArray);           \
        if (tmpBytes)                 \
            free(tmpBytes);           \
        return NULL;                  \
    }

    UInt16 *makeNativeToUnicodeArray(int encodingID) {
        jclass      cls = NULL;
        jobject     obj;
        jmethodID   mid;
        jbyteArray  jInputBytes;
        jcharArray  jOutputChars;
        int         dBegin = 0x8000;
        int         dEnd = 0xffff;
        UInt8       *tmpBytes = NULL;
        UInt16      *retArray = NULL;
        jint        ret;
        JNIEnv      *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
        int         i,j;

        //speed up the converting by setting the range for double
        //byte characters;
        switch (encodingID) {
            case 0:     // Symbol
                break;
            case 2:      //sjis
                dBegin = 0x8140;
                dEnd   = 0xfcfc;
                cls = env->FindClass("sun/io/ByteToCharSJIS");
                break;
            case 3:      //big5
                dBegin = 0xa140;
                dEnd   = 0xfefe;
                //should use MS950 here?
                //cls = env->FindClass("sun/io/ByteToCharMS950");
                cls = env->FindClass("sun/io/ByteToCharBig5");
                break;
            case 4:     //prc
                dBegin = 0x8140;
                dEnd   = 0xfea0;
                cls = env->FindClass("sun/io/ByteToCharGBK");
                break;
            case 5:    //Wansung
                dBegin = 0xa1a1;
                dEnd   = 0xfefe;
                cls = env->FindClass("sun/io/ByteToCharEUC_KR");
                break;

            case 6:    //Johab
                dBegin = 0x8141;
                dEnd   = 0xfdfe;
                cls = env->FindClass("sun/io/ByteToCharJohab");
                break;

            default:
                printf("missing encoder [%d]\n", encodingID);
                return NULL;
        }

        if (cls == 0
            || (mid = env->GetMethodID(cls, "<init>", "()V")) == NULL
            || (obj = env->NewObject(cls, mid)) == NULL)
            return NULL;
        HANDLE_EXCEPTION();

        //substitute the uncoverted default char to "0x0"
        mid = env->GetMethodID(cls, "setSubstitutionChars", "([C)V");
        HANDLE_EXCEPTION();
        if (mid != NULL) {
            jcharArray jtmp = env->NewCharArray(1);
            HANDLE_EXCEPTION();
            UInt16 cc[1] = {0};
            env->SetCharArrayRegion(jtmp, 0, 1, (jchar*)cc);
            HANDLE_EXCEPTION();
            env->CallVoidMethod(obj, mid, jtmp);
            HANDLE_EXCEPTION();
        }

        mid = env->GetMethodID(cls, "convert", "([BII[CII)I");
        HANDLE_EXCEPTION();

        if ((retArray = (UInt16*) malloc(sizeof (UInt16) * 65536)) == NULL) {
            JNU_ThrowOutOfMemoryError(env, NULL);
            return NULL;
        }

        for (i=0; i<65536; i++) {
            retArray[i] = 0xFFFD;
        }

        //convert the double byte characters
        jInputBytes = env->NewByteArray((dEnd - dBegin + 1) * 2);
        HANDLE_EXCEPTION();
        jOutputChars = env->NewCharArray(dEnd - dBegin + 1);
        HANDLE_EXCEPTION()
            if (jInputBytes == NULL || jOutputChars == NULL){
                free(retArray);
                return NULL;
            }

        if ((tmpBytes = (UInt8*) malloc(sizeof (UInt8) * 2 * (dEnd - dBegin + 1))) == NULL){
            JNU_ThrowOutOfMemoryError(env, NULL);
            free(retArray);
            return NULL;
        }

        j = 0;
        int firstByte;
        if (encodingID == 2) {
            for (i = dBegin; i <= dEnd; i++){
                firstByte = i >> 8 & 0xff;
                if (firstByte >= 0xa1 && firstByte <= 0xdf){
                    //sjis halfwidth katakana
                    *(tmpBytes + j++) = 0xff;
                    *(tmpBytes + j++) = 0xff;
                } else {
                    *(tmpBytes + j++) = firstByte;
                    *(tmpBytes + j++) = i & 0xff;
                }
            }
        } else
            for (i = dBegin; i <= dEnd; i++) {
                *(tmpBytes + j++) = i >> 8 & 0xff;
                *(tmpBytes + j++) = i & 0xff;
            }
        env->SetByteArrayRegion(jInputBytes,
                                0,
                                env->GetArrayLength(jInputBytes),
                                (jbyte*)tmpBytes);
        HANDLE_EXCEPTION();
        ret = env->CallIntMethod(obj,
                                 mid,
                                 jInputBytes,
                                 0,
                                 env->GetArrayLength(jInputBytes),
                                 jOutputChars,
                                 0,
                                 env->GetArrayLength(jOutputChars));
        HANDLE_EXCEPTION();

        env->GetCharArrayRegion(jOutputChars,
                                0,
                                env->GetArrayLength(jOutputChars),
                                (jchar*)(retArray + dBegin));
        HANDLE_EXCEPTION();

        //single byte ascii
        for (i = 0x20; i <= 0x7e; i++)
            retArray[i] = i;

        //sjis halfwidth katakana
        if (encodingID == 2) {
            for (i = 0xa1; i <= 0xdf; i++){
                retArray[i] = i - 0xa1 + 0xff61;
            }
        }

        free(tmpBytes);
        return retArray;
    }
};

const uint32 cmapFontTableTag =  0x636D6170; // 'cmap'

// cmap converters for tables 0, 2, 4 and 6 are

// Format 0: Byte Encoding table
static uint16 getGlyph0(byte *cmap, uint16 charCode)
{
    uint16 index = 0;
    byte *charP = cmap;
    if (charCode < 256) {
		charP += 6; 										/* skip format, length, and version */

		index = charP[charCode];
	}
	return index; /*****/
}

// Format 2: High-byte mapping through table
static uint16 getGlyph2(byte *cmap, uint16 charCode)
{
    byte				*charP = cmap;
	uint16				index, mapMe;
	uint16				highByte;
	uint16				lowByte;
	uint16				key;
	//sfnt_mappingTable2	*Table2;
    //sfnt_subHeader		*subHeader;
    byte               *subHeaderKeys;
    byte               *subHeaders;
    byte               *subHeader;

	charP += 6; 										/* jump over format, length, and version */

    index = 0;		/* Assume the missing glyph */

    /*
	if (wStuff && wStuff->glyph->longCharCode != 0xffff)
    charCode = wStuff->glyph->longCharCode;
    */

	highByte = charCode >> 8;
	lowByte = charCode & 0x00ff;

    //Table2 = (sfnt_mappingTable2 *) charP;
    subHeaderKeys = charP;
    subHeaders    = charP + 512;

    if ((key = GetUInt16(subHeaderKeys, highByte * 2)) != 0) {
        mapMe = lowByte;
    } else {
        mapMe = highByte;
    }

    if (mapMe == 0) {
        mapMe = lowByte;
    }


#if 0
	if ((key = Table2->subHeadersKeys[highByte]) != 0) {
		mapMe = lowByte;	/* We also need the low byte */
		if (wStuff)
			wStuff->glyph->numBytesUsed = 2;
	} else {
		mapMe = highByte;
		if (wStuff)
			wStuff->glyph->numBytesUsed = 1;
	}
	if ((wStuff == 0L) && (mapMe == 0)) {		/* JTS 7-2-93 for Inc-J AFM Files */
		mapMe = lowByte;
	}
#endif


    //  	subHeader = (sfnt_subHeader *) ((char  *) &Table2->subHeaders + Table2->subHeadersKeys [highByte]);

    subHeader = subHeaders + key;

    uint16 firstCode  = GetUInt16(subHeader,0);
    uint16 entryCount = GetUInt16(subHeader,2);


	if (mapMe < firstCode)
		return 0;			/***** Missing *****/

#if 0
	if ((Table2->subHeadersKeys[highByte] == 0) && (wStuff == 0L) && (charCode > 0xff))
		return 0;						/* single byte character (AFM fix) JTS 7-2-93 */
#endif

    if ((key == 0) && (charCode > 0xFF)) {
        return 0;
    }

	mapMe -= firstCode;					/* Subtract first code */


	if ( mapMe < entryCount ) {
		uint16 glyph;
        uint16 offset;
        uint16 idRangeOffset = GetUInt16(subHeader, 6);
        uint16 idDelta       = GetUInt16(subHeader, 4);

        //- calculate the offset from the beginning of subHeader !!!
        offset = idRangeOffset + 6 + mapMe + mapMe;
        index = GetUInt16(subHeader, offset);

        /*
		shortP = (uint16 *)&(subHeader->idRangeOffset);
        shortP = (uint16 *)((long)shortP + *shortP + mapMe + mapMe);

		if (*shortP)
        index = *shortP + subHeader->idDelta;
        */

        if (index) {
            index += idDelta;
        }
	}


#ifdef MS_CODE      // ??? Was in Sampo's code
  if (mapMe < subHeader->entryCount ) {
    uint16 glyph;

    if (glyph = * ((uint16 *) ((char *) &subHeader + subHeader->idRangeOffset) + mapMe))
      index = glyph + (uint16) (subHeader->idDelta);
  }
#endif /* MS_CODE */

    return index;
}


#if 0
// Format 4: Segment mapping to delta values
static uint16 getGlyph4(byte *cmap, uint16 charCode)
{
	uint16 segCountX2, offset, idDelta, index, uTmp;
	byte *charP = cmap;

	charP += 6; 										/* skip format, length, and version */

	segCountX2 = *charP++; segCountX2 <<= 8; segCountX2 |= *charP++;
	charP += 6;											/* skip searchRange, entrySelector, and rangeShift */

	/* go one too far, so we skip the reserved word */
	/* Do a linear search, could be optimized by a binary search, later... */
	do {
		uTmp = *charP++; uTmp <<= 8; uTmp |= *charP++;
	} while ( charCode > uTmp );
	/* charP now points at startCount[] */

	charP += segCountX2; /* point at startCount[n] */
	uTmp = *charP; uTmp <<= 8; uTmp |= *(charP+1);
	index = 0;
	if ( charCode >= uTmp ) {
		offset  = (unsigned short)(charCode - uTmp);
		charP += segCountX2; 	/* point at idDelta[n] */
		idDelta = *charP; idDelta <<= 8; idDelta |= *(charP+1);
		charP += segCountX2; 	/* point at idRangeOffset[n] */
		uTmp = *charP; uTmp <<= 8; uTmp |= *(charP+1);
		if ( uTmp == 0 ) {
			index	= (unsigned short)(charCode + idDelta);
		} else {
			charP += uTmp + offset + offset; /* point at glyphIdArray[n] */
			uTmp = *charP; uTmp <<= 8; uTmp |= *(charP+1);
			index = (unsigned short)(uTmp + idDelta);
		}
	}
    printf("getGlyph4/S: %04X (%d) -> %d.\n", charCode, charCode, index);
	return index; /*****/
}
#else

#define maxLinearX2 16
#define BinaryIteration \
		newP = (unsigned short *) ((char *)tableP + (range >>= 1)); \
		if ( charCode > SWAPW(*newP) ) tableP = newP;

/*
 * Segment gxMapping to delta values, Yack.. !
 *
 * In memory of Peter Edberg. Initial code taken from code example supplied by Peter.
 */
static uint16 getGlyph4(byte *cmap, uint16 charCode)
{
	unsigned short idDelta;
    unsigned short offset, segCountX2, index;
    unsigned short *tableP = (unsigned short *) (cmap + 6);

	index = 0; /* assume missing initially */
	segCountX2 = SWAPWINC(tableP);

	if (segCountX2 < maxLinearX2) {
		tableP += 3; /* skip binary search parameters */
	} else {
		/* start with unrolled binary search */
		unsigned short *newP;
		short  range; 		                      /* size of current search range */
		unsigned short selector; 	              /* where to jump into unrolled binary search */
		unsigned short shift; 		              /* for shifting of range */

                /* == 2*(2**floor(log2(segCount))) == 2*largest power of two
                 * <= segCount
                 */
        range 		= SWAPWINC(tableP);
		selector 	= SWAPWINC(tableP);	          /* == log2(range/2) */
		shift 		= SWAPWINC(tableP);           /* == 2*segCount-range */
		/* tableP points at endCount[] */

		if ( charCode >= SWAPW(*((unsigned short *)((char *)tableP + range))))
			tableP = (unsigned short *) ((char *)tableP + shift); /* range to low shift it up */
		switch (selector) {
            case 15: BinaryIteration;
            case 14: BinaryIteration;
            case 13: BinaryIteration;
            case 12: BinaryIteration;
            case 11: BinaryIteration;
            case 10: BinaryIteration;
            case  9: BinaryIteration;
            case  8: BinaryIteration;
            case  7: BinaryIteration;
            case  6: BinaryIteration;
            case  5: BinaryIteration;
            case  4: BinaryIteration;
            case  3:
            case  2:  /* drop through */
            case  1:
            case  0:
                break;
        }
    }

    /* Now do linear search */
	while (charCode > SWAPWINC(tableP)) /* goes one to far, so we can skip the reservedPad */
		;

	/* End of search, now do gxMapping */

	tableP = (unsigned short *) ((char *)tableP + segCountX2); /* gxPoint at startCount[] */
	if (charCode >= SWAPW(*tableP)) {
		offset = charCode - SWAPW(*tableP);
		tableP = (unsigned short *) ((char *)tableP + segCountX2); /* gxPoint to idDelta[] */
		idDelta = SWAPW(*tableP);
		tableP = (unsigned short *) ((char *)tableP + segCountX2); /* gxPoint to idRangeOffset[] */
		if ( *tableP == 0 ) {
			index	= charCode + idDelta;
		} else {
			offset += offset; /* make word offset */
			tableP 	= (unsigned short *) ((char *)tableP +
                                                      SWAPW(*tableP) +
                                                      offset ); /* gxPoint to glyphIndexArray[] */
			if (index = SWAPW(*tableP))
				index += idDelta;
		}
    }

    //printf("getGlyph4/A: %04X (%d) -> %d.\n", charCode, charCode, index);

	return index;
}

#endif


// Format 6: Trimmed table mapping
static uint16 getGlyph6(byte *cmap, uint16 charCode)
{
	uint16 entryCount, index;
	byte *charP = cmap;

	charP += 6; 										/* skip format, length, and version */

	index = *charP++;
	index <<= 8;
	index |= *charP++;
	charCode = (unsigned short)(charCode - index); /* -firstCode */
	entryCount = *charP++;
	entryCount <<= 8;
	entryCount |= *charP++; /* entryCount */
	index = 0;
	if ( charCode < entryCount ) {
		charP += charCode; /* *2, since word Array */
		charP += charCode; /* *2, since word Array */
		index = *charP++;
		index <<= 8;
		index |= *charP++;
	}
	return index; /*****/
}

// Format 1: Plain Table (65536 uint16 glyphIDs)

static uint16 getGlyph103(byte *cmap, uint16 charCode)
{
    return GetUInt16(cmap + 6, charCode *2);
}




static inline uint16 getGlyph(byte *cmap, uint16 charCode)
{
    switch(GetUInt16(cmap, 0)) {
        case 0: return getGlyph0(cmap, charCode);
        case 2: return getGlyph2(cmap, charCode);
        case 4: return getGlyph4(cmap, charCode);
        case 6: return getGlyph6(cmap, charCode);
        case 103: return getGlyph103(cmap, charCode);
#ifdef DEBUG
        default:
            printf("ERROR: Unsupported cmap format.\n");
#endif
    }
    return 0;
}


// Returns an array that converts Microsoft encoding to Unicode.
static uint16 *GetConverter(uint16 encodingID)
{
    //  0 - Symbol
    //  2 - ShiftJIS
    //  3 - BIG5
    //  4 - PRC
    //  5 - Wansung
    //  6 - Johab
    static uint16 *carray[7] = {0, 0, 0, 0, 0, 0, 0};
    if (encodingID > 6) {
        return 0;
    }

    if (carray[encodingID] == 0) {
        carray[encodingID] = makeNativeToUnicodeArray(encodingID);
        if (carray[encodingID] == 0) {
            carray[encodingID] = (uint16 *) -1;
        }
    }

    if (carray[encodingID] == (uint16 *) -1) {
        return 0;
    }

    return carray[encodingID];
}



/*+----------------------------------------------------------------------
 |
 | Name:            CreateCMAP()
 |
 | Usage:           This functions generates a Unicode cmap
 |
 | Arguments:       cmap - points to a cmap subtable
 |                  xlat - translation array that translates a character code
 |                         used to index the given cmap to a Unicode number.
 |
 | Description:
 |
 | Return value:    pointer to a new cmap subtable. Memory for this table
 |                  is allocated dynamically and should be freed when the
 |                  cmap is no longer needed
 |
 +---------------------------------------------------------------------- */

byte *CreateCMAP(byte *cmap, uint16 *xlat)
{
    byte *res=0, *type4;
    byte *buffer      = new byte [65536 * 2 + 6];
    byte *glyphArray  = buffer + 6;

    if (buffer == 0) return 0;

    memset(glyphArray, 0, 65536 * 2);


    int i, j;
    for (i=0; i<65536; i++) {
        if (xlat[i] != 0xFFFD) {
            PutInt16(getGlyph(cmap, i), glyphArray, xlat[i] * 2);
        }
    }

    PutInt16(103, buffer, 0);                                 //- cmap format 103

    return buffer;
}

static byte *CreateSymbolCMAP(byte *cmap, UInt16* Symbols)
{
    byte *res = 0, *type4;
    byte   *buffer      = new byte [65536 * 2 + 6];
    byte   *glyphArray  = buffer + 6;
    if (buffer == 0) {
        return 0;
    }
    memset(glyphArray, 0, 65536 * 2);
  
    int i, j;
    for (i=0; i<256; i++) {
        if (Symbols[i] != 0xFFFD) {
            PutInt16(getGlyph(cmap, i + 0xf000), glyphArray, Symbols[i] * 2);
        }
    }
    PutInt16(103, buffer, 0);  
    return buffer;
}

static char *upcasestr(char *str) 
{
  char *res=str;
  while (*str) {
    *str = toupper(*str);
    str++;
  }
  return res;
}

// searches for the second string in the first
static char *j_strcasestr(const char *s1, const char *s2)
{
  char *ts1=upcasestr(strdup(s1));
  char *ts2=upcasestr(strdup(s2));

  char *res = strstr(ts1, ts2);

  free(ts1);
  free(ts2);

  return res;
}

  
void InitializeCMAP(sfntFileFontObject& font, byte * &cmap)
{
    int i;

    uint32 cmapOffset, cmapLength;
    cmapLength = font.FindTableSize(cmapFontTableTag, &cmapOffset);
    byte *buffer = new byte[cmapLength];
    assert(buffer != 0);
    font.readBlock(cmapOffset, cmapLength, (char *) buffer);
    uint16 numberSubtables = GetUInt16(buffer, 2);
    uint32  ThreeZero  = 0;                   // MS Symbol
    uint32  ThreeOne   = 0;                   // MS Unicode
    uint32  ThreeTwo   = 0;                   // MS ShiftJIS
    uint32  ThreeThree = 0;                   // MS Big5
    uint32  ThreeFour  = 0;                   // MS PRC
    uint32  ThreeFive  = 0;                   // MS Wansung
    uint32  ThreeSix   = 0;                   // MS Johab
    Boolean ThreeStar  = false;               // Any MS encoding

    uint16 platformID, encodingID;

    for (i=0; i<numberSubtables; i++) {
        uint32 offset;
        platformID = GetUInt16(buffer, 8 * i + 4);
        encodingID = GetUInt16(buffer, 8 * i + 6);
        offset     = GetUInt32(buffer, 8 * i + 8);

        if (platformID == 3) {
            ThreeStar = true;
            switch (encodingID) {
                case 0: ThreeZero  = offset; break;
                case 1: ThreeOne   = offset; break;
                case 2: ThreeTwo   = offset; break;
                case 3: ThreeThree = offset; break;
                case 4: ThreeFour  = offset; break;
                case 5: ThreeFive  = offset; break;
                case 6: ThreeSix   = offset; break;
            }
        }
    }

    if (ThreeStar) {
#ifdef DEBUG
        if (ThreeZero)  printf("%s - CMAP 3,0 found at offset %d.\n", font.GetFontLocalPlatformName(), ThreeZero);
        if (ThreeOne)   printf("%s - CMAP 3,1 found at offset %d.\n", font.GetFontLocalPlatformName(), ThreeOne);
        if (ThreeTwo)   printf("%s - CMAP 3,2 found at offset %d.\n", font.GetFontLocalPlatformName(), ThreeTwo);
        if (ThreeThree) printf("%s - CMAP 3,3 found at offset %d.\n", font.GetFontLocalPlatformName(), ThreeThree);
        if (ThreeFour)  printf("%s - CMAP 3,4 found at offset %d.\n", font.GetFontLocalPlatformName(), ThreeFour);
        if (ThreeFive)  printf("%s - CMAP 3,5 found at offset %d.\n", font.GetFontLocalPlatformName(), ThreeFive);
        if (ThreeSix)   printf("%s - CMAP 3,6 found at offset %d.\n", font.GetFontLocalPlatformName(), ThreeSix);
#endif
        if (ThreeOne) {                 //- MS Unicode cmap found, just use it
            uint16 length = GetUInt16(buffer, ThreeOne+2); // length of the subtable
            cmap = new byte [length];
            assert(cmap != 0);
            memcpy(cmap, buffer+ThreeOne, length);
        } else if (ThreeZero) {         //- The font has a Symbol encoding. Use it.
//            cmap = CreateCMAP(buffer+ThreeZero, GetConverter(0));
            // REMIND: need platform independent determination.
            const char*fontname = font.GetFontLocalPlatformName();
            if (j_strcasestr(fontname, "SYMBOL.TTF") != NULL) {
                cmap = CreateSymbolCMAP(buffer+ThreeZero, Symbols_b2c);
            } else if (j_strcasestr(fontname, "WINGDING.TTF") != NULL) {
                cmap = CreateSymbolCMAP(buffer+ThreeZero, WingDings_b2c);
            } else {
                uint16 length = GetUInt16(buffer, ThreeZero+2); // length of the subtable
                cmap = new byte [length];
                assert(cmap != 0);
                memcpy(cmap, buffer+ThreeZero, length);
            }
        } else if (ThreeTwo) {          //- The font has a ShiftJIS cmap. Convert it to Unicode.
            cmap = CreateCMAP(buffer+ThreeTwo, GetConverter(2));
        } else if ( ThreeThree ) {                                    //- Big 5
            cmap = CreateCMAP ( buffer + ThreeThree, GetConverter(3));
        } else if (ThreeFour) {                                       //- GB
            cmap = CreateCMAP(buffer+ThreeFour, GetConverter(4));
        } else if ( ThreeFive ) { // Wansung
	   cmap = CreateCMAP ( buffer + ThreeFive, GetConverter(5));
	} else if ( ThreeSix ) { // Johab
	   cmap = CreateCMAP ( buffer + ThreeSix, GetConverter(6));
	}
#ifdef DEBUG
        else {
            printf("No known MS cmap subtables were found.\n");
        }
#endif


    } else {
#ifdef DEBUG
        printf("No usable cmap tables found.\n");
#endif
    }

    // Initialize the character to glyph mapping function
    // Unicode cmap can be either format 4 or format 6, anything else is an error
    if (cmap != 0) {
        uint16 format = GetUInt16(cmap, 0);
        switch (format) {
            case 103:                     // XXX cmap format 1
            case 4:
            case 6: break;
            default:
                // Can't handle this cmap
                delete [] cmap;
                cmap = 0;
#ifdef DEBUG
                printf("Unicode cmap table is not usable (format: %d.)\n", format);
#endif
        }
    }

    delete [] buffer;
}



//- Public APIs

void ConvertUnicodeToGlyphs(sfntFileFontObject& font, byte * &cmap, int numChars, const Unicode chars[], UInt32 glyphs[])
{
    int i;

    //- Initialize the glyph array with "missing glyph" IDs
    memset(glyphs, 0, numChars * 4);

#if 0
    if (cmap == 0) {
        if (!font.isValid) {
            return;
        } else {
            InitializeCMAP(font, cmap);
        }
    }

    if (cmap == 0) {
        font.isValid = false;
        return;
    }
#endif

    for (i=0; i<numChars; i++) {
        glyphs[i] = getGlyph(cmap, chars[i]);
    }
    return;
}

uint32 ConvertUnicodeToGlyph(sfntFileFontObject& font, byte * &cmap, Unicode theChar)
{
    uint32 res = 0;

#if 0
    if (cmap == 0) {
        if (!font.isValid) {
            return res;
        } else {
            InitializeCMAP(font, cmap);
        }
    }

    if (cmap != 0) {
        res = getGlyph(cmap, theChar);
    } else {
        font.isValid = false;
    }
#else
    res = getGlyph(cmap, theChar);
#endif

    return res;
}

