/*
 * @(#)fontObject.cpp	1.83 98/10/13
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <assert.h>
#include <string.h> 
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "fontObject.h"
#include "hsConfig.h"
#include "hsFixedTypes.h"
#include "hsScalar.h"
#include "hsGFontScaler.h"
#include "hsGGlyphCache.h"
#include "t2kScalerContext.h"
#include "nrScalerContext.h"
#include "FontGlue.h"
#include "Strike.h"
#include "cmaps.h"

#include <jni.h>
#include <jni_util.h>
#include <jlong_md.h>
#include <sun_awt_font_NativeFontWrapper.h>

inline jfloat fixedToFloat(hsFixed fixed) {
    return fixed / (float)(65536);
}
fontObject *DefaultFontObject();

#undef PRINT_INFO

#include "sfnt_types.h"

    

typedef struct {
    UInt32 tag;
    UInt32 version;
    UInt32 count;
} TTCHeader;

    
#define CHUNK_LEN 4096

#define GET16(P)    (UInt16)(   (((UInt8 *)&(P))[0] << 8) \
                                | (((UInt8 *)&(P))[1]) )

#define GET32(P)    (UInt32)(   (((UInt8 *)&(P))[0] << 24) \
                                | (((UInt8 *)&(P))[1] << 16) \
                                | (((UInt8 *)&(P))[2] << 8) \
                                | (((UInt8 *)&(P))[3]) )

#define PUT32(P, V)    do {                                 \
                        UInt8 *val = (UInt8 *)&(P);         \
                        val[0] = ((V) >>  24) & 0xff;       \
                        val[1] = ((V) >>  16) & 0xff;       \
                        val[2] = ((V) >>  8) & 0xff;        \
                        val[3] = (V) & 0xff;                \
                    } while (0)

fontObject::~fontObject() {
    if (fFontname) {
        delete [] fFontname;
    }
    if (fStrike) {
        delete fStrike;
    }
    if (fStrikeTx) {
        delete fStrikeTx;
    }
}

Strike& fontObject::getStrike(
                    FontTransform& tx,
                    jboolean isAntiAliased,
                    jboolean usesFractionalMetrics) {
    if (        (fStrike == NULL)
            ||  (!fStrikeTx)
            ||  (isAntiAliased != fStrikeIsAntiAliased)
            ||  (usesFractionalMetrics != fStrikeUsesFractionalMetrics)
            ||  (!tx.equals(*fStrikeTx))) {
        if (fStrike) {
            delete fStrike;
            fStrike = NULL;
        }
        fStrike = new Strike(*this, tx, isAntiAliased, usesFractionalMetrics);
        if (fStrikeTx) {
            delete fStrikeTx;
            fStrikeTx = NULL;
        }
        fStrikeTx = new FontTransform(tx);
        fStrikeIsAntiAliased = isAntiAliased;
        fStrikeUsesFractionalMetrics = usesFractionalMetrics;
    }
    return *fStrike;
}

void fontObject::setStrike(
                    Strike *theStrike, FontTransform& tx,
                    jboolean isAntiAliased,
                    jboolean usesFractionalMetrics)
{
    if(NULL != fStrike){
        delete fStrike;
    }
    fStrike = theStrike;
    if(NULL != fStrikeTx){
        delete fStrikeTx;
    }
    fStrikeTx = new FontTransform(tx);
    fStrikeIsAntiAliased = isAntiAliased;
    fStrikeUsesFractionalMetrics = usesFractionalMetrics;
}


UInt32 fontObject::FindTableSize(UInt32 tableTag, UInt32* offset)
{
	// baseclass does not implement this, 'cause it may not have tables
	return 0;
}

const void* fontObject::ReadTable(UInt32 tableTag, UInt32* tableLength)
{
    UInt32 offset;
    UInt32 length = this->FindTableSize(tableTag, &offset);
    const void* table = nil;

    if (length)
       table = this->ReadChunk(offset, length);
	if (tableLength)
		*tableLength = length;
    return table;
}

class FOTempChunk {
    fontObject* fFontObj;
    const void* fChunk;
public:
    FOTempChunk(fontObject* fo, UInt32 offset, UInt32 length) {
        fFontObj = fo;
        fChunk = fo->ReadChunk(offset, length);
    }

    ~FOTempChunk() {
        fFontObj->ReleaseChunk(fChunk);
    }

    const void* GetData() const { return fChunk; }
};



class FOTempTable {
    sfntFileFontObject* fFontObj;
    UInt32 fOffset, fLength;
    const void* fTable;
public:
    FOTempTable(sfntFileFontObject* fo, UInt32 tableTag) : fFontObj(fo) {
#ifdef PRINT_INFO
        fprintf(stderr, "FOTempTable - CP1\n");
#endif
        fLength = fo->FindTableSize(tableTag, &fOffset);
        assert(fLength > 0);

#ifdef PRINT_INFO
        fprintf(stderr, "FOTempTable() fOffset: %d, fLength: %d.\n", fOffset, fLength);
#endif
        
        fTable = fo->ReadChunk(fOffset, fLength);
    }

    ~FOTempTable() {
        fFontObj->ReleaseChunk(fTable);
        fTable = nil;		// debug only
    }

    UInt32 GetOffset() const { return fOffset; }
    const void* GetTable() const { return fTable; }
};


class FOTempFontInfo {    
    type1FileFontObject* fFontObj;
    t1FontInfo *fFontInfo;
public:
    FOTempFontInfo(type1FileFontObject* fo) : fFontObj(fo) {
        fFontInfo = fFontObj->GetFontInfo();
    }

    ~FOTempFontInfo() {
        fFontObj->DeleteFontInfo(fFontInfo);
        fFontInfo = nil;
    }

    t1FontInfo* GetFontInfo() const { return fFontInfo; }
};

const void* fileFontObject::ReadChunk(UInt32 offset, UInt32 length,
                                      void* chunk)
{
#ifdef PRINT_INFO
    fprintf(stderr, "ReadChunk(%d, %d, %X)\n", offset, length, chunk);
#endif
    char* data = (char*)chunk;
    
    if (fUseCount == 0) {
        fFile = fopen(fFontLocalPlatformName, "rb");
        if (fFile == NULL || ferror(fFile) != 0) {
//         assert(fFile);
//         assert(ferror(fFile) == 0);
            return NULL;
        }
        if (fFileSize == 0) {
            fseek(fFile, 0, SEEK_END);
            fFileSize = ftell(fFile);
        }
    }
    fUseCount += 1;
    
    if (length > 0 && fFileSize > 0) {
        assert(offset < fFileSize);
        if (offset + length > fFileSize)
            length = fFileSize - offset;
        
        if (data == nil) {
            data = new char[length];
        }

        if( nil == data ) {
#ifdef DEBUG
	  fprintf(
		  stderr,"new returns nil, fontObject.cpp,%s\n",
		  "fileFontObject::ReadChunk(UInt32,UInt32,void*)\n"
		  );
#endif
	  return NULL;
	}                    

        int err = fseek(fFile, offset, SEEK_SET);
        if (err) {
            fprintf(stderr, "fseek(%d) returned %d\n", offset, err);
        }
        
        long bytesRead = fread(data, sizeof(char),length, fFile);
        if (bytesRead != length) {
            err = ferror(fFile);
            int tellvalue = ftell(fFile);
#ifdef DEBUG
            fprintf(stderr,
                    "<%s> of %d, ln %d, rd %d, sz %d, tell %d, err %d\n",
                    fFileName, offset, length, bytesRead, fFileSize, tellvalue,
                    err);
#endif
        }
    }
    
    // If the caller provided the chunk, call release with nil
    // to decrement the usecount, since the caller won't call
    // release.
    if (chunk) {
        this->ReleaseChunk(nil);
    }
    return data;
}

//- Light-weight ReadChunk
void fileFontObject::readBlock(size_t offset, size_t size, char *dest)
{
    if (fUseCount == 0) {
        fFile = fopen(fFontLocalPlatformName, "rb");
        assert(fFile);
        assert(ferror(fFile) == 0);
        if (fFileSize == 0) {
            fseek(fFile, 0, SEEK_END);
            fFileSize = ftell(fFile);
        }
    }
    
    assert(size > 0);

    if (ftell(fFile) != offset) {
        int error = fseek(fFile, offset, SEEK_SET);
        assert (error == 0);
    }
    size_t count = fread(dest, sizeof(char), size, fFile);
    assert(ferror(fFile) == 0 && count == (size_t) size);

    if (fUseCount == 0) {
        fclose(fFile);
        fFile = NULL;
    }
    else {
        fUseCount++;
    }
}




void fileFontObject::ReleaseChunk(const void* chunk)
{

    assert(fUseCount > 0);

    if (chunk != nil) {
        delete[] (char*)chunk;
    }
        
    fUseCount -= 1;
    if (fUseCount == 0) {
        int err = fclose(fFile);
        if (err) {
#ifdef DEBUG
            fprintf(stderr, "<%s> fclosed = %d\n", fFileName, err);
#endif
        }
    }
}

UInt32 fileFontObject::GetDataLength()
{
	if (fFileSize == 0) {
		(void)this->ReadChunk(0, 0);
		this->ReleaseChunk(nil);
	}
	return fFileSize;
}

/* returns:
 *        0 got next char
 *        -1 end of file (or ASCII block) found
 */

int
type1FileFontObject:: NextChar(sDataRecord *sData) {
  if (sData->eof)
    return -1;
  
  sData->ch = sData->buffer[sData->boffs++];

  if (sData->ch == 0x80) {
    sData->eof = 1;
    return -1;
  }
  
  if (sData->boffs >= sData->buflen) {
    sData->buflen = fread(sData->buffer, sizeof(char), CHUNK_LEN, sData->fd);
    sData->boffs = 0;
    if (sData->buflen <= 0) {
      sData->eof = 1;
      return -1;
    }
  }
  return 0;
}

/* returns -1 if eof found, 0 otherwise.
 * Skips over white-space characters and Postscript comments
 */
int
type1FileFontObject:: SkipSpaces(sDataRecord *sData) {
  int res;

  while (sData->ch == '%') {
    do {
      res = NextChar(sData);
      if (sData->ch == 13 || sData->ch == 10)
        break;
    } while (res != -1);
    if (res == -1)
      break;
  }

  if (res == -1)
    return -1;

  while (sData->ch <= 32) {
    res = NextChar(sData);
    if (res == -1)
      return -1;
  }
  return 0;
}

/* returns -1 on eof, token class otherwise.
 * Token classes:
 * 1 - /
 * 2 - string
 */
int 
type1FileFontObject::NextToken(sDataRecord *sData, char *token, int toklen) {
  int res = SkipSpaces(sData);
  int i = 0;
  token[0] = 0;

  if (!res) {
    sData->prevToken = sData->curToken;

    if (sData->ch == '/') {
      NextChar(sData);
       
      return sData->curToken = 1;
    }

    if (sData->ch == '(') {             /* quoted string is found */
      res = NextChar(sData);
      while ((res != -1) && (sData->ch != ')')) {
        if (i < toklen-1)
          token[i++] = sData->ch;
        NextChar(sData);
      }
      if (i < toklen)
        token[i] = 0;
       
      return sData->curToken = 2;
    }

    while ((res != -1) && (sData->ch > 32)) {
      if (i < toklen-1)
        token[i++] = sData->ch;
      NextChar(sData);
    }
    if (i < toklen)
      token[i] = 0;
    if (!strcmp(token,"eexec"))                   /* hack! don't read data after eexec */
      sData->eof = 1;
    return sData->curToken = 2;
  }
  return -1;
}
    
/* Closes a font */
void 
type1FileFontObject::CloseFont(sDataRecord *sData) {
  if (sData->buffer)
    free(sData->buffer);
  if (sData->fd != NULL)
    fclose(sData->fd);
  sData->eof = 1;
}


/* Opens a PFA or a PFB Postscript Type 1 font */

int 
type1FileFontObject::OpenFont(const char *fileName, sDataRecord *sData) {
  int errflag = 0;

  assert(sData != NULL);
  sData->buffer = NULL;
  sData->eof = 0;
  sData->curToken = 0;
  sData->prevToken = 0;

#ifdef PRINT_INFO  
  printf ( "The font file path is :%s\n", fileName );
#endif
  if ((sData->fd = fopen(fileName, "r")) != NULL) {
    if ((sData->buffer = (char*)malloc(CHUNK_LEN)) != NULL) {
      sData->buflen = fread(sData->buffer, sizeof(char),CHUNK_LEN,sData->fd);
      if (sData->buflen > 0) {
        if (sData->buffer[0] == 0x80)   /* font is PFB */
          sData->boffs = 6;
        else
          sData->boffs = 0;             /* font is PFA */
      } else
        errflag = 1;
    } else
      errflag = 1;
  } else
    errflag = 1;

  if (sData->boffs >= sData->buflen)
    errflag = 1;

  if (errflag) {
    CloseFont(sData);
    return -1;
  }

  NextChar(sData);

  return 0;
}
      
/* Public APIs */
        
t1FontInfo*
type1FileFontObject::GetFontInfo() {
  t1FontInfo *fi = NULL;
  sDataRecord sData;
  int res;
  char tokbuf[1024];
  int f1 = 0, f2 = 0, f3 = 0;

  if (OpenFont(GetFontLocalPlatformName(), &sData) != -1 &&
     (fi = (t1FontInfo*) malloc(sizeof(t1FontInfo))) != NULL) {
    memset(fi, 0, sizeof(*fi));
    do {
      res = NextToken(&sData, tokbuf, 1024);
      if (res == 2) {
        if (!strcmp(tokbuf, "FullName") && sData.prevToken == 1) {
          res = NextToken(&sData, tokbuf, 1024);
          fi->fullName = strdup(tokbuf);
          f1 = 1;
        }

        if (!strcmp(tokbuf, "FamilyName") && sData.prevToken == 1) {
          res = NextToken(&sData, tokbuf, 1024);
          fi->familyName = strdup(tokbuf);
          f2 = 1;
        }
        
        if (!strcmp(tokbuf, "FontName") && sData.prevToken == 1) {
          res = NextToken(&sData, tokbuf, 1024);
          if (res == 1) {
            res = NextToken(&sData, tokbuf, 1024);
            fi->fontName = strdup(tokbuf);
            f3 = 1;
          }
        }
      }
      if (f1 && f2 && f3)
        break;
    } while (res > 0);
  }
  CloseFont(&sData);
  return fi;
}

void type1FileFontObject::DeleteFontInfo(t1FontInfo *fi) {
  if (fi != NULL) {
    if (fi->fontName) free(fi->fontName);
    if (fi->fullName) free(fi->fullName);
    if (fi->familyName) free(fi->familyName);
    free(fi);
  }
}

// unicode char to lowercase
inline UInt16 toLowerU(UInt16 unicodeChar) {
    if ((unicodeChar > 0x40) && (unicodeChar <= 0x40 + 26)) {
        return unicodeChar + 0x20;
    }
    return unicodeChar;
}

// case insensitive
static int equalUnicodeToAsciiNC(const UInt16 *name, int unicodeNameLen,
                               const char *ascii, int asciiBytes = 0)

{
    int retval = 0;
    if (!asciiBytes && ascii) {
        asciiBytes = strlen(ascii);
    }
    if (name && unicodeNameLen && asciiBytes && ascii && *ascii) {
        if (asciiBytes == unicodeNameLen) {
            int i;
            retval = 1;
            for (i=0; i<unicodeNameLen; i++) {
                if ((name[i] > 255) || (name[i] != (0xff & ascii[i]))) {
                    if (toLowerU(name[i]) == toLowerU(0xff & ascii[i])) {
                        continue;
                    }
                    retval = 0;
                    break;
                }
            }
        }
    }
    return retval;
}

// REMIND: need case insensitive match
Boolean type1FileFontObject::MatchName(int nameID, const UInt16 *name,
                                       int nameLen)
{
    char *matchName = NULL;
    FOTempFontInfo fInfo(this);
    t1FontInfo *fi = fInfo.GetFontInfo();

    switch(nameID) {
    case kFullName:
        matchName = (fi && fi->fullName) ? fi->fullName : "";
        break;
    case kFamilyName:
        matchName = (fi && fi->familyName) ? fi->familyName : "";
        break;
    default:
        return false;
    }
    return equalUnicodeToAsciiNC(name, nameLen, matchName, strlen(matchName));
}

static void copyAsciiToUnicode(const char *src, UInt16 *dest, int copySize = 0)
{
    if (copySize == 0 ) {
        // REMIND add some boundary overflow checking
        while (*dest++ = (UInt16)(*src++)) {
        }
    } else {
        for (int i=0; i < copySize; i++) {
            dest[i] = src[i] & 0xff;
        }
    }
}

static Unicode *strDupASCIItoUNICODE(const char *src, int& retNameLen, int maxInput = 0) {
    if (src) {
        if (!maxInput) {
            retNameLen = strlen(src);
        } else {
            retNameLen = maxInput;
        }
        Unicode *retval = new Unicode[retNameLen];

	if( nil != retval ) {
	  copyAsciiToUnicode(src, retval, retNameLen);
	} else {
#ifdef DEBUG
	  fprintf(
		  stderr,
		  "new returns nil. fontObject.cpp, strDupASCIItoUNICODE(const char*,int&,int)\n"
		  );
#endif
	  retNameLen = 0;
	}
        return retval;
    } else {
        retNameLen = 0;
        return NULL;
    }
}


int type1FileFontObject::GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name)
{
    int retval = 0;
    FOTempFontInfo fInfo(this);
    t1FontInfo *fi = fInfo.GetFontInfo();
    if (name) {
        switch(nameID) {
        case kFullName:
            if (fi && fi->fullName) {
	      strcpy((char *)name, fi->fullName);
              retval = strlen(fi->fullName);
            }
            break;
        case kFamilyName:
            if (fi && fi->familyName) {
	      strcpy((char *)name, fi->familyName);
              retval = strlen(fi->familyName);
            }
            break;
        default:
            break;
        }
        platformID = 1;                 //-
        scriptID   = 0;
    }
    return retval;
}

const Unicode *type1FileFontObject::GetFontName(int& nameLen)
{
    if (!fFontname) {
        FOTempFontInfo fInfo(this);
        t1FontInfo *fi = fInfo.GetFontInfo();
        if (fi && fi->fullName) {
            fFontname = strDupASCIItoUNICODE(fi->fullName, fFontNameLen);
        }
    }
    nameLen = fFontNameLen;
    return fFontname;
}

///////////////////////////////////////////////////////////////////////

#if 0
static void fprintUnicode(const UInt16 *name, int nameLen) {
    for (int i=0; i < nameLen; i++) {
        fprintf(stderr,"%c", (char)name[i]);
    }
}
static void fprintUnicodeSW(const UInt16 *name, int nameLen) {
    for (int i=0; i < nameLen; i++) {
        fprintf(stderr,"%c", (char)GET16(name[i]));
    }
}
#endif

// REMIND: need case insensitive match
Boolean sfntFileFontObject::MatchName(int meaningID, const UInt16 *name,
                                      int nameLen)
{
    FOTempTable table(this, nameFontTableTag);
    const sfntNameHeader* nameTable = (sfntNameHeader*)table.GetTable();

    const sfntNameRecord* nameRecord = (sfntNameRecord*) ((UInt8*) nameTable + 
                                                 sizeof_sfntNameHeader);
    short count = GET16(nameTable->count);
    short nameID;
    nameID = meaningID;

#if 0
    fprintf(stderr, "Locating ["); fprintUnicode(name, nameLen); fprintf(stderr, "]\n");
#endif
    for (--count; count >= 0; --count) {
        if (nameID == -1 || GET16(nameRecord->nameID) == nameID) {
            short platformID = GET16(nameRecord->platformID);
            UInt8 *namePtr = (UInt8*) nameTable;
            namePtr += GET16(nameTable->stringOffset);
            namePtr += GET16(nameRecord->offset);
            short nameBytes = GET16(nameRecord->length);

            if ((platformID == 0) || (platformID == 3)) {
                UInt16 *unicodeName = (UInt16 *)namePtr;
#if 0
                fprintf(stderr, "   Compare to U["); fprintUnicodeSW(unicodeName, nameBytes/2); fprintf(stderr, "]\n");
#endif
                if ((nameLen * 2) == nameBytes) {
                    Boolean matched = true;
                    int j;
                    for (j=0; j<nameLen; j++) {
                        if (toLowerU(name[j]) != toLowerU(GET16(unicodeName[j]))) {
                            matched = false;
                            break;
                        }
                    }
                    if (matched) {
                        return true;
                    }
                }
            } else {
#if 0
                fprintf(stderr, "   Compare to A[%.*s]\n", nameBytes, (char *)namePtr);
#endif
                if (equalUnicodeToAsciiNC(name, nameLen, (char *)namePtr, nameBytes)) {
                    return true;
                }
            }
        }
	nameRecord++;
    }
    return false;
}

sfntFileFontObject::~sfntFileFontObject()
{
    if (offsets) {
        delete [] offsets;
    }

}

// This function verifies that the TrueType file is not corrupted
// by comparing the number of glyphs declared in 'maxp' table
// and the capacity of the 'loca' table. This is faster then
// calculating checksums (especially for Asian fonts).

#define tag_maxp 0x6D617870
#define tag_head 0x68656164
#define tag_loca 0x6C6F6361

Boolean VerifyTrueTypeFile(sfntFileFontObject *fo)
{ 
  // remind : put a better verificatiion test below. PARRY
#if 0

  // This test below fails for many fonts in the CJK locales
  // the fonts are off by atleast one in the loca table.
  // if we enforce the check then the fonts are rejected.
  // Obviously we have to register the fonts - which otherwise
  // behave properly.
  // To check for totally bogus font data a test to check for the
  // required tables like the cmap, loca, glyf etc. could be instituted.

    unsigned char *maxp = (unsigned char *) fo->ReadTable(tag_maxp, 0);
    unsigned char *head = (unsigned char *) fo->ReadTable(tag_head, 0);

    if (maxp == 0 || head == 0) {
        return false;
    }

    UInt32 locaOffs, locaSize;
    locaSize = fo->FindTableSize(tag_loca, &locaOffs);

    UInt16 numGlyphs = GetUInt16(maxp, 4);
    UInt16 indexfmt  = GetUInt16(head, 50);
    UInt32 locaCapacity = locaSize / (indexfmt ? 4 : 2);

    fo->ReleaseChunk(maxp);
    fo->ReleaseChunk(head);

    if (locaCapacity - 1 != numGlyphs) {
        return false;
    }
#endif
    return true;
}


Boolean sfntFileFontObject::Init(
        const Unicode& fileName, int nameLen, const char& localPlatformName,
        const char &nativeName, enum FontFormats format, int logicalNumber)
{
    fileFontObject::Init(fileName, nameLen, localPlatformName, nativeName,
			 format);
    Boolean ret(true);

    TTCHeader theHeader = { 0, 0, 0 };
    // Can't read header
    if (this->ReadChunk(0, sizeof theHeader, &theHeader) == NULL) {
        return false;
    }

    switch (GET32(theHeader.tag)) {
        case 0x74746366: {                        //- 'ttcf'
            // format = kTTCFontFormat;
            // fprintf(stderr,"%s/%d -> TTC file.\n", fileName, logicalNumber);
            fFontCount = GET32(theHeader.count);

            offsets = new UInt32 [fFontCount];

	    if( nil == offsets ) {
#ifdef DEBUG
	      fprintf(
		      stderr,"new returns nil. fontObject.cpp. %s %s\n",
		      "sfntFileFontObject::Init(const Unicode&,int,const char&,enum FontFormats,int)",
		      "switch(GET32(theHeader.tag)); case 0x74746366" 
		      );
#endif
	      return false;
	    }

            this->ReadChunk(sizeof theHeader, fFontCount * sizeof *offsets, offsets);
            for (int i=0; i<fFontCount; i++) {
                offsets[i] = GET32(offsets[i]);
            }
            break;
        }
        case 0x00010000:                          //- TrueType version 1
        case 0x74727565: {                        //- 'true' - TrueType version 2
            //format = kTrueTypeFontFormat;
            //fprintf(stderr,"%s -> TrueType file.\n", fileName);
            fFontCount = 1;
            offsets = new UInt32 [1];
	    if( nil == offsets ) {
#ifdef DEBUG
	      fprintf(
		      stderr,"new returns nil. fontObject.cpp. %s %s\n",
		      "sfntFileFontObject::Init(const Unicode&,int,const char&,enum FontFormats,int)",
		      "switch(GET32(theHeader.tag)); case 0x74727565" 
		      );
#endif
	      return false;
	    }


            offsets[0] = 0;
            if (!VerifyTrueTypeFile(this)) {
#ifdef DEBUG
                fprintf(stderr, "Warning: TrueType font file `%s` is corrupted.\n", &localPlatformName);
#endif
                ret = false;
            }
            break;
        }
        default:
            ret = false;
#ifdef DEBUG
            fprintf(stderr, "Warning: Unrecognized font file format: `%s`.\n", &localPlatformName);
#endif
    }
        
            
    if (logicalNumber < fFontCount) {
        fCurFont = logicalNumber;
    } else {
#ifdef DEBUG
        fprintf(stderr,"Incorrect logical font number: %d, number of logical fonts: %d.\n", logicalNumber, fFontCount);
#endif
        fCurFont = fFontCount-1;
    }
    return ret;
}


int sfntFileFontObject::GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name)
{
  // Changed search a little: If an exact match is not found, try to at 
  // least match the platform.
  // Old behavior was to pick the first record with the right nameID, but this
  // may be for another platform.

#ifdef PRINT_INFO
//    fprintf(stderr, "sfntFileFontObject::GetName(%d)\n", meaningID);
#endif
  
    FOTempTable table(this, nameFontTableTag);
    const sfntNameHeader* nameTable = (sfntNameHeader*)table.GetTable();

    const sfntNameRecord* nameRecord = (sfntNameRecord*) ((UInt8*) nameTable + 
                                                          sizeof_sfntNameHeader);
    const sfntNameRecord* recordToUse = NULL;
    const sfntNameRecord* firstNameMatch = NULL;
    const sfntNameRecord* firstPlatformMatch = NULL;
        
    short count = GET16(nameTable->count);
    
#ifdef PRINT_INFO
    fprintf(stderr,"count: %d.\n", count);
#endif

    // fprintf(stderr,"GetName: CP 1\n");

    for (--count; count >= 0; --count) {
#ifdef PRINT_INFO
      fprintf(stderr, "platformID: %d, scriptID: %d, languageID: %d, nameID: %d.\n",
                nameRecord->platformID, nameRecord->scriptID, nameRecord->languageID, nameRecord->nameID);
#endif
      // Do wildcard names really make sense?  I mean, can you really use
      // a result which may be the family name, copyright notice, manufacturer
      // name, or one of several URL's?
      if (nameID == 0xFFFF || nameID == GET16(nameRecord->nameID)) {
        if (firstNameMatch==NULL) {
          firstNameMatch = nameRecord;
        }

        if (platformID == 0xFFFF || GET16(nameRecord->platformID) == platformID) {

          if (firstPlatformMatch==NULL) {
            firstPlatformMatch = nameRecord;
          }

          if ((scriptID   == 0xFFFF || GET16(nameRecord->scriptID) == scriptID) &&
	      (languageID == 0xFFFF || GET16(nameRecord->languageID) == languageID)) {

	    recordToUse = nameRecord;
	    break;
	  }
	}
      }
      nameRecord++;
    }

    if (recordToUse == NULL) {
      if (firstPlatformMatch != NULL) {
	recordToUse = firstPlatformMatch;
      }
      else if (firstNameMatch != NULL) {
	recordToUse = firstNameMatch;
      }
    }

    if (recordToUse != NULL) {

	// write back the arguments
	platformID = GET16(recordToUse->platformID);
	scriptID   = GET16(recordToUse->scriptID);
	languageID = GET16(recordToUse->languageID);

        if (name) {
            char *buffer = (char *)name;
            memcpy(name, ((UInt8*)nameTable + GET16(nameTable->stringOffset))
                   + GET16(recordToUse->offset), GET16(recordToUse->length));
            buffer[GET16(recordToUse->length)] = 0;
            buffer[GET16(recordToUse->length)+1] = 0;
        }
        return GET16(recordToUse->length);
    } else {
        return 0;
    }
}

// REMIND: need locale preference in request.
const Unicode *sfntFileFontObject::GetFontName(int& nameLen) {
    if (!fFontname) {
        FOTempTable table(this, nameFontTableTag);
        const sfntNameHeader* nameTable = (sfntNameHeader*)table.GetTable();
        const sfntNameRecord* nameRecord = (sfntNameRecord*) ((UInt8*) nameTable + sizeof_sfntNameHeader);
        const sfntNameRecord* savedPtr = nameRecord;
        short count = GET16(nameTable->count);
        Boolean found(false);
        int bpc(0), i;

        for (i=0; i<count; i++) {
            short pID = GET16(nameRecord->platformID);
            short sID = GET16(nameRecord->scriptID);
            short nID = GET16(nameRecord->nameID);
            
            if (pID == 1 && sID == 0 && nID == 6) {         //- Macintosh/Undefined/PS Name
                found = true;
                bpc = 1;                                    //- 1 byte/character
                break;
            } else if (pID == 3 && sID == 0 && nID == 6) {  //- Microsoft/Symbol/PS Name
                found = true;
                bpc = 2;                                    //- 2 bytes/character
                break;
            }
            nameRecord++;
        }

        if (!found) {
            fFontname = strDupASCIItoUNICODE("unknown font", fFontNameLen);
        } else {
            int len = GET16(nameRecord->length);
            void *namePtr = (char *) nameTable
                                + GET16(nameTable->stringOffset)
                                + GET16(nameRecord->offset);

            switch (bpc) {
                case 1:
                    fFontname = strDupASCIItoUNICODE((char *)namePtr, fFontNameLen, len);
                    break;
                case 2:
                    fFontNameLen = len / 2;
                    fFontname = new Unicode[fFontNameLen];
		    if( nil == fFontname ) {
#ifdef DEBUG
		      fprintf(
			      stderr,"new returns nil. fontObject.cpp, %s\n",
			      "sfntFileFontObject::GetFontName(int&)"
			      );
#endif
		      return NULL;
		    }
                    memcpy(fFontname, namePtr, fFontNameLen * sizeof(Unicode));
                    break;
            }

        }
    }
    nameLen = fFontNameLen;
    return fFontname;
}
            
        

UInt32 sfntFileFontObject::FindTableSize(UInt32 tableTag, UInt32* offset)
{
#ifdef PRINT_INFO
    fprintf(stderr, "sfntFileFontObject::FindTableSize()\n");
#endif

    sfntDirectory	dir;
    (void)this->ReadChunk(offsets[fCurFont], sizeof_sfntDirectory, &dir);

    int numTables = GET16(dir.numOffsets);

    UInt32 tableOffset = 0;
    UInt32 tableLength = 0;
        
    if (numTables > 0) {
        const sfntDirectoryEntry* dirEntry = 
            (sfntDirectoryEntry*)this->ReadChunk(offsets[fCurFont] + sizeof_sfntDirectory,
                                                 (UInt32)sizeof(sfntDirectoryEntry) *
                                                 numTables);
            
        const sfntDirectoryEntry *entry = dirEntry;
        const sfntDirectoryEntry *stop = entry + numTables;

	if( nil == entry ) {
#ifdef DEBUG
	  fprintf(
		  stderr,"ReadChunk returns nil. fontObject.cpp, %s\n",
		  "sfntFileFontObject::FindTableSize(UInt32, UInt32*)"
		  );
#endif
	  return tableLength;
	}

        do {
            if (GET32(entry->tableTag) == tableTag) {
                tableLength = GET32(entry->length);
                tableOffset = GET32(entry->offset);
                break;
            }
        } while (++entry < stop);
        
        this->ReleaseChunk(dirEntry);
    }
    // don't return something if we didn't find the table
    if (offset && tableOffset) {
        *offset = tableOffset;
    }
    return tableLength;
}

/////////////////////////////
static jboolean ttContainsHintsTable(const sfntFileFontObject *ttFO) {

    return true;
#if 0
// REMIND: more thorough algorithm needed
#define tag_gasp 0x67617370

    unsigned char *gasp = (unsigned char *) ttFO->ReadTable(tag_gasp, 0);
    return (gasp != NULL);

#undef tag_gasp
#endif
}

enum FontFormats
sfntFileFontObject::getScalerID(
                        FontTransform& tx,
                        jboolean isAntiAliased,
                        jboolean usesFractionalMetrics)
{
    // for native scaler/rasterizer ...
    // require opaque, non-antialiased, non-fractional metrics
    // require simple scaling transform (pointSize)
    // require hints in the font file
    // require recognition by native system
    if (!isAntiAliased && !usesFractionalMetrics) {
        if (ttContainsHintsTable(this)) {
            Strike *theStrike = 
                new Strike(*this, kNRFontFormat,
                            tx, isAntiAliased, usesFractionalMetrics);
            if (theStrike->GetNumGlyphs()) {
                // finally ... all conditions passed
                setStrike(  theStrike, tx, 
                            isAntiAliased,
                            usesFractionalMetrics);
                return kNRFontFormat;
            } else {
                delete theStrike;
            }
        }
    }
    return fontObject::getScalerID(tx, isAntiAliased, usesFractionalMetrics);
}

///////////////////////////////////////////////////////////////////////

static int fontListCount = 0;
static struct fontListNode {
    int           type;
    fontObject   *font;
    fontListNode *next;
} *fontList = nil, *fontListTail = nil;


static void addToTheListHead(int type, fontObject *fo)
{
    fontListNode *node = new fontListNode;
    if (node) {
        node->type = type;
        node->font = fo;
        node->next = fontList;
        fontList = node;
        if (!fontListTail) {
            fontListTail = node;
        }
        fontListCount++;
    }
}

static void addToTheListTail(int type, fontObject *fo)
{
    fontListNode *node = new fontListNode;
    if (node) {
        node->type = type;
        node->font = fo;
        node->next = nil;
        if (!fontList) {
            fontList = node;
            fontListTail = node;
        } else {
            fontListTail->next = node;
            fontListTail = node;
        }
        fontListCount++;
    }
}


static fontListNode* getNodeByNumber(int n)
{
    fontListNode* node = fontList;

    while (n && node && node->next) {
        node = node->next;
        n--;
    }

    return node;
}

#ifdef PRINT_INFO
static void dumpList()
{
    fontListNode *node = fontList;
    int i(0);

    while (node) {
        printf("%d: %s `%s`\n", i++, ((fileFontObject *) node->font)->GetFileName(),
               node->font->GetFontName());
        node = node->next;
    }
}
#endif

extern "C" {

void FillBytes(void *destination, long length, long pattern) {
	memset(destination, pattern, length);
}

/*
 * Class:     sun_awt_font_NativeFontWrapper
 * Method:    registerFonts
 * Signature: (Ljava/util/Vector;II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_font_NativeFontWrapper_registerFonts
        (JNIEnv *env, jclass obj, jobject fonts, jint size, 
	 jobject names, jint jFormat)
{
    // REMIND this is an evil cast -- use types from javah headers
    enum FontFormats format = (enum FontFormats)jFormat;
    jclass vectorClass;
    jmethodID elementAtID;
    jstring fileName;
    jstring nativeString;
    const Unicode *fontFileName;
    int fontFileNameLen;
    const char *localPlatformName;
    const char *nativeName;
    int i;

    vectorClass = env->GetObjectClass(fonts);
    elementAtID = env->GetMethodID(vectorClass, "elementAt",
                                   "(I)Ljava/lang/Object;");
    
    if ((elementAtID == NULL)) {
        JNU_ThrowInternalError(env, "elementAtID failed\n");
        return;
    }
    for (i=0; i<size; i++) {
        fileName = (jstring) env->CallObjectMethod(fonts, elementAtID, i);
        if (JNU_IsNull(env, fileName)) {
            JNU_ThrowInternalError(env, "Empty file name in registerFonts().\n");
            return;
        }
        localPlatformName = JNU_GetStringPlatformChars (env, fileName, NULL);
        //printf("format: %d, file: %s\n", format, localPlatformName);
        nativeString = (jstring) env->CallObjectMethod(names, elementAtID, i);
        if (JNU_IsNull(env, nativeString)) {
            JNU_ThrowInternalError(env, "Empty Native name in registerFonts().\n");
            return;
        }
        nativeName = JNU_GetStringPlatformChars (env, nativeString, NULL);

        switch (format) {
            case kTrueTypeFontFormat: {
                sfntFileFontObject* ffo;
                int j(0);
                do {          //- adg: there is at least one logical font in a TrueType or TTC file
                    ffo = new sfntFileFontObject;
                    Boolean valid(false);
                    {
                        JStringBuffer theFileName(env, fileName);
                        if (theFileName.invalid()) {
                            JNU_ReleaseStringPlatformChars(env, fileName, localPlatformName);
                            JNU_ReleaseStringPlatformChars(env, nativeString, nativeName);
                            JNU_ThrowInternalError(env, "Empty file name in registerFonts().\n");
                            return;
                        }
                        fontFileName = theFileName.buffer();
                        fontFileNameLen = theFileName.getLength();
                        valid = ffo->Init(*fontFileName, fontFileNameLen, *localPlatformName, *nativeName, kTrueTypeFontFormat, j++);
                    }
                    if (!valid) {
                        delete ffo;
                    } else {
                        byte *cmap; // dummy entry
                        //addToTheListTail(format, ffo);

                        InitializeCMAP(*ffo, cmap);
                        addToTheListTail(format, ffo);
		    	MakeNativeMapper (nativeName);
                    }
                } while (j < ffo->getLogicalFontCount());
                break;
            }
            case kType1FontFormat: {
                type1FileFontObject* ffo = new type1FileFontObject;

                JStringBuffer theFileName(env, fileName);
                if (theFileName.invalid()) {
                    JNU_ReleaseStringPlatformChars(env, fileName, localPlatformName);
                    JNU_ReleaseStringPlatformChars(env, nativeString, nativeName);
                    JNU_ThrowInternalError(env, "Empty file name in registerFonts().\n");
                    return;
                }
                fontFileName = theFileName.buffer();
                fontFileNameLen = theFileName.getLength();

                if( !ffo->Init(*fontFileName, fontFileNameLen, *localPlatformName, *nativeName, kType1FontFormat) ) {
                    delete ffo;
                } else {
                    addToTheListTail(format, ffo);
                }
                break;
            }
            case kNRFontFormat: {
                fileFontObject * ffo = GetNativeFontObject ();
                Boolean valid(false);
		{
                    JStringBuffer theFileName(env, fileName);
                    if (theFileName.invalid()) {
                        JNU_ReleaseStringPlatformChars(env, fileName, localPlatformName);
                        JNU_ReleaseStringPlatformChars(env, nativeString, nativeName);
                        JNU_ThrowInternalError(env, "Empty file name in registerFonts().\n");
                        return;
                    }
                    fontFileName = theFileName.buffer();
                    fontFileNameLen = theFileName.getLength();
                    valid = ffo->Init (*fontFileName, fontFileNameLen, 
				       *localPlatformName, *nativeName, 
				       kNRFontFormat);
                }
                if (!valid) {
                    delete ffo;
                } else {
                    addToTheListTail(format, ffo);
		    MakeNativeMapper (nativeName);
                }
                break;
	    }
        }
        JNU_ReleaseStringPlatformChars(env, fileName, localPlatformName);
        JNU_ReleaseStringPlatformChars(env, nativeString, nativeName);
    }

    
#ifdef PRINT_INFO
    dumpList();
#endif

    static jboolean scalerRegistered = false;
    if (!scalerRegistered) {
        hsGFontScaler::Register(kT2KFontFormat, new t2kFontScaler);
        hsGFontScaler::Register(kNRFontFormat, new nrFontScaler);
        scalerRegistered = true;
    }
}

} // extern "C"

//////////////// Name cache /////////////////
#define USE_FONTNAME_CACHE

#ifdef USE_FONTNAME_CACHE
class fontNameCache {
    struct Entry {
        Entry           *fNext;

        int             fLength;
        const Unicode   *fName;
        fontObject      *fFO;
    };
    Entry*	fEntry;
public:
    fontNameCache();
    ~fontNameCache();

    fontObject *Find(const Unicode name[], int nameLen) const;
    void Add(const Unicode name[], int nameLen, fontObject *fo);
};

fontNameCache::fontNameCache() : fEntry(nil)
{
}

fontNameCache::~fontNameCache()
{
    Entry   *entry = fEntry;

    while (entry) {
        Entry* next = entry->fNext;
        delete [] (Unicode *)entry->fName;
        delete entry;
        entry = next;
    }
}

fontObject *fontNameCache::Find(const Unicode name[], int nameLen) const
{
    const Entry* entry = fEntry;

    while (entry) {
        if (nameLen == entry->fLength &&
                memcmp(name, entry->fName, sizeof(Unicode) * nameLen) == 0) {
            return entry->fFO;
        }
        entry = entry->fNext;
    }
    return nil;
}

inline Unicode *DupName(const Unicode name[], int nameLen)
{
    Unicode *copy = new Unicode [nameLen];

    if (nil != copy ) {
        for (int i = 0; i < nameLen; i++) {
            copy[i] = name[i];
        }
    } else {
        #ifdef DEBUG
            fprintf(stderr,"new returns nil, fontObject.cpp, fontNameCache:DupName(const Unicode[], int)\n");
        #endif
    }
    return copy;
}

void fontNameCache::Add(const Unicode name[], int nameLen, fontObject *fo)
{
    Entry   *entry = new Entry;

    if (nil != entry ) {
        entry->fNext = fEntry;
        fEntry = entry;

        entry->fLength = nameLen;
        entry->fName = DupName(name, nameLen);
        entry->fFO = fo;
    } else {
        #ifdef DEBUG
            fprintf(stderr,"new returns nil, fontObject.cpp,fontNameCache: Add(const Unicode[], int,fontObject*)\n");
        #endif	  
    }
}
#endif

//////////////// Public functions /////////////////

fontObject *findFO(const UInt16 *name, int nameLen) {
    fontObject *fo = 0;
    
#ifdef USE_FONTNAME_CACHE
    static fontNameCache* gFontNameCache = nil;
    if (gFontNameCache == nil)
        gFontNameCache = new fontNameCache;
    
    fo = gFontNameCache->Find(name, nameLen);
    if (fo)
        return fo;
#endif

    fontListNode *node = fontList;

    while (node) {
        if (node->font->MatchName(kFullName, name, nameLen)) {
            fo = node->font;
            goto FOUND_IT;
        }
        node = node->next;
    }

    node = fontList;
            
    while (node) {
        if (node->font->MatchName(kFamilyName, name, nameLen)) {
            fo = node->font;
            goto FOUND_IT;
        }
        node = node->next;
    }

    fo = DefaultFontObject();

FOUND_IT:

#ifdef USE_FONTNAME_CACHE
    gFontNameCache->Add(name, nameLen, fo);
#endif
    return fo;
}

fontObject *DefaultFontObject() {
    // adg: should check for the best font for a given locale
    static fontObject *defaultFont = 0;

    if (defaultFont == 0) {
        fontListNode *node = fontList;
        while (node) {
            int nameLen = 0;
            const Unicode *nodeFontName = node->font->GetFontName(nameLen);
            if (nodeFontName && nameLen) {
	      #if 1
                // adg: don't use Wingdings-Regular as a default font
                if ((!equalUnicodeToAsciiNC(nodeFontName, nameLen, "Wingdings-Regular")) &&
		    (!equalUnicodeToAsciiNC(nodeFontName, nameLen, "ItcZapfdingbats")) &&
		    (!equalUnicodeToAsciiNC(nodeFontName, nameLen, "Symbol"))) {

                    defaultFont = node->font;
                }
              #endif
		// fix by PARRY - the default font in Java is a Serif font
		#if 0
		if ( equalUnicodeToAsciiNC(nodeFontName, nameLen, "Serif")) {
		  
                    defaultFont = node->font;
                    break;
                }
                #endif

		#if 1
                // adg: if ArialMT is found - use it as a default font
                if (equalUnicodeToAsciiNC(nodeFontName, nameLen, "ArialMT")) {
                    defaultFont = node->font;
                    break;
                }
		#endif

		/* Use LucidaSans as default (for Solaris). */
                if (equalUnicodeToAsciiNC(nodeFontName, nameLen, "LucidaSans")) {
                    defaultFont = node->font;
#ifdef DEBUG
fprintf (stderr, "\n******************   WARNING   *************\n");
fprintf (stderr, "The TrueType fonts specified in the font.properties file\n");
fprintf (stderr, "were not found on this system. Package SUNWi1of contains\n");
fprintf (stderr, "TrueType fonts that are used by Java2D. You can find this\n");
fprintf (stderr, "package on your Solaris 2.6 or 2.7 CD. You might want\n");
fprintf (stderr, "to install this package so that Java2D will have access\n");
fprintf (stderr, "to its default set of fonts. For now, using LucidaSans as\n");
fprintf (stderr, "the default font for Java2D.\n\n");
#endif
		   
                    break;
                }
            }
            node = node->next;
        }
        if (defaultFont == 0) {
            // XXX throw a Java exception istead
            fprintf(stderr,"\nFatal error: No usable fonts found.\n\n");
            exit(1);
        }
    }

    return defaultFont;
}


void registerFont(enum FontFormats format, fontObject *cf) {
    addToTheListHead(format, cf);
}

fontObject *GetFontObject(const Unicode *fileName, int fileNameLen) {
    if (!fileName || !fileNameLen) {
        return NULL;
    }
    fontListNode *node = fontList;
    while (node) {
        if (node->font) {
            int nodeNameLen = 0;
            const Unicode *nodeFileName = node->font->GetFileName(nodeNameLen);
            if (nodeFileName && nodeNameLen) {
                if (nodeNameLen >= fileNameLen) {
                    int nodeOffset = nodeNameLen - fileNameLen;
                    if (!memcmp(fileName, nodeFileName + nodeOffset, fileNameLen * sizeof(Unicode))) {
                        return node->font;
                    }
                }
            }
        }
        node = node->next;
    }
    return NULL;
}

fontObject* GetFontObject(unsigned index)
{
    if (fontListCount == 0) {
        fprintf(stderr,"\nFatal error: no fonts were loaded.\n\n");
        exit(1);
    }
    assert(index < fontListCount);
    return getNodeByNumber(index)->font;
}


///////////////////////////////////////////////////////////////////////////////


extern "C" {
int CountFontObjects()
{
    return fontListCount;
}

fontObject *FindFontObject(const UInt16 * name, unsigned int nameLen, int style) {
    int i, count = CountFontObjects();
    fontObject *newFO;
    UInt16 platformID  = 1;
    UInt16 scriptID = 0;
    UInt16 languageID = 0xFFFF;
    UInt16 nameID = kFamilyName;

    if (!name || (nameLen <= 0)) {
        return DefaultFontObject();
    }
    if (equalUnicodeToAsciiNC(name, nameLen, "Default", strlen("Default"))) {
        return DefaultFontObject();
    }

    newFO = findFO(name, nameLen);

#if 0
    //REMIND finish me
    if (style != 0) {
        char newName[1024];
        newFO->GetName(&platformID, &scriptID, &languageID, &nameID, newName);
        nameLen = strlen(newName);
        // REMIND: get manifest constants from extracted header file
        if ((style == 1) || (style == 3)) {
            strcat(newName, " ");
            strcat(newName, "Bold");
            nameLen += 5;
        }
        if ((style == 2) || (style == 3)) {
            strcat(newName, " ");
            strcat(newName, "Italic");
            nameLen += 7;
        }
        newFO = findFO(newName, nameLen);
    }
#endif
    return newFO;
}

const void* FOReadChunk(fontObject *fo, UInt32 offset, UInt32 length)
{
    assert(fo != nil);

    return ((fontObject*)fo)->ReadChunk(offset, length);
}

void FOReleaseChunk(fontObject *fo, const void* chunk)
{
    assert(fo != nil);

    ((fontObject*)fo)->ReleaseChunk(chunk);
}

long FOFindTableSize(fontObject *fo, UInt32 tableTag, UInt32* offset)
{
    assert(fo != nil);
    
    return ((fontObject*)fo)->FindTableSize(tableTag, offset);
}

const void* FOGetData(fontObject *fo)
{
    assert(fo != nil);

	return ((fontObject*)fo)->GetData();
}

UInt32 FOGetDataLength(fontObject *fo)
{
    assert(fo != nil);

	return ((fontObject*)fo)->GetDataLength();
}
} // end of extern "C"
