/*
 * @(#)HeadSpin.h	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc.
 * Chapel Hill, NC USA.
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. HEADSPIN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
 */

#ifndef HeadSpinHDefined
#define HeadSpinHDefined

#ifndef hsConfigDefined
	#include "hsConfig.h"
#endif

#if !(HS_BUILD_FOR_REFERENCE)
	#if HS_BUILD_FOR_MAC
		#include <Types.h>
	#endif
		#include <stdio.h>
#endif
#if HS_CAN_USE_FLOAT
	#include <math.h>
#endif

#ifdef HS_DEBUGGING
#if HS_BUILD_FOR_WIN32

#ifdef HS_FIND_MEM_LEAKS
#define VC_EXTRALEAN	     // Exclude rarely-used stuff from Windows headers
#include <afx.h>
#define new DEBUG_NEW
#else // HS_FIND_LEAKS
#include <crtdbg.h>		// for hsAssert _RPT
#endif // HS_FIND_LEAKS

#endif // HS_BUILD_FOR_WIN32
#endif // HS_DEBUGGING

//////////////////////////////////////////////////////////////////////////////

#define kPosInfinity16		(32767)
#define kNegInfinity16		(-32768)

#define kPosInfinity32		(0x7fffffff)
#define kNegInfinity32		(0x80000000)

typedef unsigned long	UInt32;
typedef long			Int32;
typedef unsigned short	UInt16;
typedef short			Int16;
typedef unsigned char	UInt8;

typedef UInt8			Byte;
typedef Int32 			hsFixed;
typedef Int32 			hsFract;
typedef char			Boolean;
typedef UInt32			Bool32;

#ifndef false
	#define false		0
#endif

#ifndef true
	#define true		1
#endif

#if HS_CAN_USE_FLOAT
	#define HS_PI		3.1415927f
#endif

#ifndef nil
#define nil (0)
#endif

typedef Int32	hsError;

#define hsLongAlign(n)		(((n) + 3) & ~3L)

#define hsMaximum(a, b)		((a) > (b) ? (a) : (b))
#define hsMinimum(a, b)		((a) < (b) ? (a) : (b))
#define hsABS(x)			((x) < 0 ? -(x) : (x))
#define hsSGN(x) 			(((x) < 0) ? -1 : ( ((x) > 0) ? 1 : 0 ))



#include "hsScalar.h"

#ifdef __cplusplus
inline Boolean hsBitTst2Bool(UInt32 a, UInt32 b)
{
	return (a & b) != 0;
}

//////////////////////////////////////////////////////////////////////////////

int		hsStrlen(const char *name);
char*	hsStrcpy(char *destOrNil, const char *src);
void		hsStrcat(char *dest, const char *src);
Boolean	hsStrEQ(const char *s1, const char *s2);
char*	hsScalarToStr(hsScalar);

/////////////////////////////////////////////////////////////////////////////

typedef void (*hsDebugMessageProc)(const char message[]);
extern hsDebugMessageProc gHSDebugProc;
#define HSDebugProc(m)	{ if (gHSDebugProc)	gHSDebugProc(m); }
hsDebugMessageProc hsSetDebugMessageProc(hsDebugMessageProc newProc);

#ifdef HS_DEBUGGING
	void	hsDebugMessage(const char message[], long refcon);
	void	hsAssertFunc(int line, const char *file, const char *message);
	#define hsDebugCode(code)		code
#if HS_BUILD_FOR_WIN32
	#define hsAssert(cond, m) do { if(!(cond)) { hsDebugMessage(m, 0); _RPTF0(_CRT_ERROR, m); } } while (0)
#else
	#define hsAssert(cond, message) if(!(cond)) hsAssertFunc(__LINE__,__FILE__, message); else 1
#endif
	#define hsIfDebugMessage(clause, message, refcon)	if (clause) hsDebugMessage(message, refcon); else 1
#else	// Not debugging
	#define hsDebugMessage(message, refcon)
	#define hsAssertFunc(line, file, message)

	#define hsDebugCode(code)		
	#define hsIfDebugMessage(clause, message, refcon)
	#define hsAssert(cond, message)
#endif

//////////////////////////////////////////////////////////////////////////////

inline void hsSwap(Int32& a, Int32& b)
{
	UInt32	c = a;
	a = b;
	b = c;
}

inline void hsSwap(UInt32& a, UInt32& b)
{
	UInt32	c = a;
	a = b;
	b = c;
}

inline void hsSwap(hsScalar& a, hsScalar& b)
{
	hsScalar	c = a;
	a = b;
	b = c;
}

//////////////////////////////////////////////////////////////////////////////

enum {
	kModifierOption 	= 0x0001,
	kModifierCtrl		= 0x0002,
	kModifierCmd 		= 0x0004,
	kModifierShift 		= 0x0008,
	kModifierSpace 		= 0x0010,
#if HS_BUILD_FOR_WIN32
	kModifierMenuAccess	= kModifierCtrl,
#endif
#if HS_BUILD_FOR_MAC
	kModifierMenuAccess	= kModifierCmd,
#endif
	kModifierDoubleClick	= 0x0080
};
typedef UInt32 hsModifierKeys;

hsModifierKeys hsGetModifierKeys();

#if HS_BUILD_FOR_MAC
	extern hsModifierKeys MacEventToModifiers(const struct EventRecord* event);
#endif

//////////////////////////////////////////////////////////////////////////////

struct hsRGBAColor32 {
#if HS_BUILD_FOR_WIN32
	Byte	b, g, r, a;
#else
	Byte	a, r, g, b;
#endif
	void		Set(Byte rr, Byte gg, Byte bb)
                        {
                            this->r = rr; this->g = gg; this->b = bb;
                        }
	void		Set(Byte aa, Byte rr, Byte gg, Byte bb)
                        {
                            this->a = aa; this->r = rr; this->g = gg; this->b = bb;
                        }

	friend int	operator==(const hsRGBAColor32& a, const hsRGBAColor32& b)
			{
				return *(UInt32*)&a == *(UInt32*)&b;
			}
	friend int	operator!=(const hsRGBAColor32& a, const hsRGBAColor32& b) { return !(a == b); }
};

struct hsColorRGBA {
	hsScalar		r,g,b,a;
	
	hsRGBAColor32 ToRGBA32() const;
};

struct hsVector3;

struct hsIntPoint2 {
	Int32	fX, fY;
	
	hsIntPoint2* Set(Int32 x, Int32 y) { fX = x; fY = y; return this; }
};

struct hsIntRect {
	Int32		fLeft, fTop, fRight, fBottom;

	Int32		Width() const { return fRight - fLeft; }
	Int32		Height() const { return fBottom - fTop; }
	Int32		CenterX() const { return fLeft + fRight >> 1; }
	Int32		CenterY() const { return fTop + fBottom >> 1; }

	Boolean		Empty() const { return fLeft >= fRight || fTop >= fBottom; }
	void			SetEmpty() { fLeft = fTop = fRight = fBottom = 0; }

	hsIntRect*	Set(Int32 left, Int32 top, Int32 right, Int32 bottom)
				{
					fLeft = left; fTop = top; fRight = right; fBottom = bottom;
					return this;
				}
	hsIntRect*	Set(const struct hsRect* r);
	hsIntRect*	Union(const hsIntRect *);
	Boolean		Intersect(const hsIntRect*);
	Boolean		Intersect(Int32 left, Int32 top, Int32 right, Int32 bottom);
	Boolean		Contains(Int32 x, Int32 y) const
				{
					return x >= fLeft && x < fRight && y >= fTop && y < fBottom;
				}
	Boolean		Contains(const hsIntRect* r) const
				{
					return fLeft <= r->fLeft && fTop <= r->fTop && fRight >= r->fRight && fBottom >= r->fBottom;
				}
	hsIntRect*	Offset(Int32 dx, Int32 dy)
				{
					fLeft += dx; fTop += dy; fRight += dx; fBottom += dy;
					return this;
				}
	hsIntRect*	Inset(Int32 dx, Int32 dy);

	friend int		operator==(const hsIntRect& a, const hsIntRect& b)
				{
					return	a.fLeft == b.fLeft && a.fTop == b.fTop &&
							a.fRight == b.fRight && a.fBottom == b.fBottom;
				}
	friend int		operator&&(const hsIntRect& a, const hsIntRect& b)
				{
					return	a.fLeft < b.fRight && a.fRight > b.fLeft &&
							a.fTop < b.fBottom && a.fBottom > b.fTop;
				}

#if HS_BUILD_FOR_MAC
	Rect*		ToRect(Rect* r)
				{
					r->left = fLeft; r->top = fTop; r->right = fRight; r->bottom = fBottom;
					return r;
				}
#endif
};

typedef UInt32	hsGSeedValue;

enum {
	kIdentityMatrixType		= 0,
	kTranslateMatrixType	= 0x01,
	kScaleMatrixType		= 0x02,
	kRotateMatrixType		= 0x04,
	kPerspectiveMatrixType	= 0x08,
	kUnknownMatrixType	= 0x80
};
typedef UInt32 hsMatrixType;

///////////////////////////////////////////////////////////////////////////

#if HS_BUILD_FOR_WIN32
	inline UInt16 hsSWAP16(UInt16 swap)
	{
		return (swap >> 8) | (swap << 8);
	}
	UInt32 hsSWAP32(UInt32 swap);
#else
	#define hsSWAP16(n)		(n)
	#define hsSWAP32(n)		(n)
#endif

class hsStream {
protected:
	UInt32		fBytesRead;
public:
				hsStream() : fBytesRead(0) {}
	virtual		~hsStream() {}

	virtual Boolean AtEnd();
	virtual UInt32	Read(UInt32 byteCount, void * buffer) = 0;
	virtual UInt32	Write(UInt32 byteCount, const void* buffer) = 0;
	virtual void	Skip(UInt32 deltaByteCount) = 0;

	void			WriteString(const char cstring[]);

	Boolean		GetToken(char *, int maxLen  = 0);
	UInt8		ReadByte();
	UInt16		ReadSwap16();
	void			ReadSwap16(int count, UInt16 values[]);
	UInt32		ReadSwap32();
	void 			ReadSwap32(int count, UInt32 values[]);
	float			ReadSwapFloat();
	void			ReadSwapFloat(int count, float values[]);
	hsScalar		ReadSwapScalar();

	void			WriteSwapScalar(hsScalar value);
	void			WriteByte(UInt8 value);
	void			WriteSwap16(UInt16 value);
	void			WriteSwap16(int count, const UInt16 values[]);
	void			WriteSwap32(UInt32 value);
	void 			WriteSwap32(int count, const  UInt32 values[]);
	void			WriteSwapFloat(float value);
	void			WriteSwapFloat(int count, const float values[]);
};

class hsStreamable {
public:
	virtual void	Read(hsStream* stream) = 0;
	virtual void	Write(hsStream* stream) = 0;
	virtual UInt32	GetStreamSize() = 0;
};

class hsFileStream: public hsStream {	
	UInt32		fRef;
public:
				hsFileStream();
	virtual		~hsFileStream();

	virtual Boolean AtEnd();
	virtual UInt32	Read(UInt32 byteCount, void* buffer);
	virtual UInt32	Write(UInt32 byteCount, const void* buffer);
	virtual void	Skip(UInt32 deltaByteCount);

	virtual UInt32	GetFileRef();
	virtual void	SetFileRef(UInt32 refNum);
};

#if !(HS_BUILD_FOR_REFERENCE)
class hsUNIXStream: public hsStream
{	
	FILE*		fRef;
public:
	virtual Boolean AtEnd();
	virtual UInt32	Read(UInt32 byteCount, void* buffer);
	virtual UInt32	Write(UInt32 byteCount, const void* buffer);
	virtual void	Skip(UInt32 deltaByteCount);

	FILE*		GetFILE() { return fRef; }
	void			SetFILE(FILE* file) { fRef = file; }
};
#endif

class hsRAMStream : public hsStream {
	UInt32				fTotalWritten;
	UInt32				fTotalOffset;
	struct hsRAMChunk*		fChunkBase;

	UInt32				fLocalOffset;
	struct hsRAMChunk*		fCurrChunk;
public:
				hsRAMStream();
	virtual		~hsRAMStream();
	virtual Boolean AtEnd();
	virtual UInt32	Read(UInt32 byteCount, void * buffer);
	virtual UInt32	Write(UInt32 byteCount, const void* buffer);
	virtual void	Skip(UInt32 deltaByteCount);
	
	UInt32		GetEOF() { return fTotalOffset; }
	void			Rewind();
};

///////////////////////////////////////////////////////////////////

class hsRefCnt {
private:
	Int32		fRefCnt;
public:
				hsRefCnt() : fRefCnt(1) {}
	virtual		~hsRefCnt();

	Int32		RefCnt() const { return fRefCnt; }
	virtual void	UnRef();
	virtual void	Ref();
};

#define hsRefCnt_SafeRef(obj)		do { if (obj) (obj)->Ref(); } while (0)
#define hsRefCnt_SafeUnRef(obj)	do { if (obj) (obj)->UnRef(); } while (0)

#define hsRefCnt_SafeAssign(dst, src)		\
		do {							\
			hsRefCnt_SafeRef(src);		\
			hsRefCnt_SafeUnRef(dst);		\
			dst = src;					\
		} while (0)

//////////////////////////////////////////////////////

struct hsPolar {
	hsScalar	fRadius;
	hsScalar	fAngle;
};

struct hsPoint2 {
	hsScalar	fX, fY;

	hsPoint2*	Set(hsScalar x, hsScalar y) { fX = x; fY = y; return this; }
	hsPoint2*	Grid(hsScalar period);
	Boolean	CloseEnough(const hsPoint2* p, hsScalar tolerance) const;

	hsScalar	Magnitude() const { return hsPoint2::Magnitude(fX, fY); }
	hsScalar	MagnitudeSquared() const { return fX*fX + fY * fY; }
	hsPolar*	ToPolar(hsPolar* polar) const;

	friend hsPoint2	operator-(const hsPoint2& s, const hsPoint2& t)
		{
			hsPoint2	result;
			return *result.Set(s.fX - t.fX, s.fY - t.fY);
		}
	friend Boolean operator==(const hsPoint2& s, const hsPoint2& t)
		{
			return (s.fX == t.fX && s.fY == t.fY);
		}
	friend Boolean operator!=(const hsPoint2& s, const hsPoint2& t)
		{
			return (s.fX != t.fX || s.fY != t.fY);
		}
	friend hsPoint2 operator+(const hsPoint2& s, const hsPoint2& t)
		{
			hsPoint2	result;
			return *result.Set(s.fX + t.fX, s.fY + t.fY);
		}
	friend hsPoint2 operator*(const hsPoint2& s, hsScalar t)
		{
			hsPoint2	result;
			return *result.Set(s.fX * t, s.fY * t);
		}
	friend hsPoint2 operator*( hsScalar t, const hsPoint2& s)
		{
			hsPoint2	result;
			return *result.Set(s.fX * t, s.fY * t);
		}

	static hsScalar	Magnitude(hsScalar x, hsScalar y) { return hsSquareRoot(x * x + y * y); }
	static hsScalar	Distance(const hsPoint2& p1, const hsPoint2& p2);
	static hsPoint2	Average(const hsPoint2& p1, const hsPoint2& p2);
	static hsScalar	ComputeAngle(const hsPoint2& a, const hsPoint2& b, const hsPoint2& c);
};

struct hsRect {
	hsScalar		fLeft, fTop, fRight, fBottom;

	hsRect*		Set(hsScalar left, hsScalar top, hsScalar right, hsScalar bottom)
				{
					fLeft		= left;
					fTop		= top;
					fRight	= right;
					fBottom	= bottom;
					return this;
				}
	hsRect*		Set(const hsPoint2* p1, const hsPoint2* p2);
	hsRect*		Set(UInt32 count, const hsPoint2 p[]);
	hsRect*		Set(const hsIntRect* r)
				{
					return this->Set(hsScalar(r->fLeft), hsScalar(r->fTop),
								hsScalar(r->fRight), hsScalar(r->fBottom));
				}

	hsRect*		Move(hsScalar dx, hsScalar dy);
	hsRect*		Inset(hsScalar dx, hsScalar dy);
	hsRect*		Union(const hsRect *r);
	hsRect*		Union(const hsPoint2* p);
	hsRect*		Union(UInt32 count, const hsPoint2 p[]);
	hsRect*		UnionX(hsScalar x)
				{
					if (x < fLeft) fLeft = x; else
					if (x > fRight) fRight = x;
					return this;
				}
	hsRect*		UnionY(hsScalar y)
				{
					if (y < fTop) fTop = y; else
					if (y > fBottom) fBottom = y;
					return this;
				}
	Boolean		Intersect(const hsRect*);

	friend Boolean	operator==(const hsRect& r1, const hsRect& r2);
	friend Boolean	operator!=(const hsRect& r1, const hsRect& r2);

	Boolean		Contains(hsScalar x, hsScalar y) const
				{
					return	fLeft <= x && fRight >= x &&
							fTop <= y && fBottom >= y;
				}
	Boolean		Contains(const hsPoint2* p) const
				{
					return	fLeft <= p->fX && fRight >= p->fX &&
							fTop <= p->fY && fBottom >= p->fY;
				}
	Boolean		Contains(const hsRect* r) const
				{
					return	fLeft <= r->fLeft && fTop <= r->fTop &&
							fRight >= r->fRight && fBottom >= r->fBottom;
				}
	Boolean		IsEmpty(void) const { return fLeft >= fRight || fTop >= fBottom; }

	hsScalar		Width(void) const { return fRight - fLeft; }
	hsScalar		Height(void) const { return fBottom - fTop; }
	hsScalar		CenterX(void) const { return (fLeft + fRight) / hsScalar(2); }
	hsScalar		CenterY(void) const { return (fTop + fBottom) / hsScalar(2); }
	hsPoint2*		Center(hsPoint2* center) const;
	hsPoint2*		ToQuad(hsPoint2 quad[4]) const;
	hsScalar		Area() const { return this->Width() * this->Height(); }

	hsIntRect*	Round(hsIntRect* r) const;
	hsIntRect* 	RoundOut(hsIntRect* r) const;
	hsIntRect*	Truncate(hsIntRect* r) const;
};

/*@
@page hsstream.html
@class hsStream
This class defines the base class for stream objects. You can create your own stream object
by simply overridding the virtual methods in this base class.
@method Read read bytes from the stream
@method Write write bytes to the stream
@method Skip skip bytes in the stream
@endclass
@class hsFileStream
This class implements a stream which reads or writes from a file. The file is created or opened by the
application using the standard OS file routines.  
@parent hsStream
@method Read read bytes from the stream
@method Write write bytes to the stream
@method Skip skip bytes in the stream
@method GetFileRef get the file reference number for this stream
@method SetFileRef set the file reference number for this stream
@methoddef hsFileStream SetFileRef
This method is used to set the file reference number for a hsFileStream. This number is
returned by the file system open routines such as FSpOpenDF() on the Mac or OpenFile on
windows.
@param UInt32 refnum  the number from the system open routine
@endclass
@page hssimpletypes.html
@struct hsPoint2
This is the standard struct used for 2-D points.
@field hsScalar  fX the x coordinate
@field hsScalar fY the y coordinate
@endstruct
@struct hsRect
This is the standard struct for axis aligned rectangles. Note that fRight should be >= fLeft 
and fBottom should be >= fTop;
@field hsScalar fLeft the left edge
@field hsScalar fTop the top edge
@field hsScalar fRight the right edge
@field hsScalar fBottom the bottom edge
@endstruct
@endpage
*/
#endif // __CPLUSPLUS
#endif
