/*
 * @(#)cmaps.h	1.10 98/09/08
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Alex D. Gelfenbain
 *
 */

#ifndef __CMAPS_H
#define __CMAPS_H

#include <assert.h>


#include "fontObject.h"
#include "dtypes.h"


//- public methods

//- initialize the font.cmap
void InitializeCMAP(sfntFileFontObject& font, byte * &cmap);

//- convert an array of Unicode characters to an array of font-specific glyphs
void ConvertUnicodeToGlyphs(sfntFileFontObject& font, byte * &cmap, int numChars, const Unicode chars[], UInt32 glyphs[]);

//- convert a Unicode character to a font-specific glyph
uint32 ConvertUnicodeToGlyph(sfntFileFontObject& font, byte * &cmap, Unicode theChar);


//- Data access macros for data stored in big-endian format
inline Int16  GetInt16(const byte *ptr, size_t offset=0) /*FOLD00*/
{
    assert(ptr != 0);

    Int16 t = (ptr+offset)[0] << 8 | (ptr+offset)[1];
    return t;
}

inline UInt16 GetUInt16(const byte *ptr, size_t offset=0) /*FOLD00*/
{
    assert(ptr != 0);

    UInt16 t = (ptr+offset)[0] << 8 | (ptr+offset)[1];
    return t;
}

inline Int32  GetInt32(const byte *ptr, size_t offset=0) /*FOLD00*/
{
    assert(ptr != 0);

    Int32 t = (ptr+offset)[0] << 24 | (ptr+offset)[1] << 16 |
        (ptr+offset)[2] << 8  | (ptr+offset)[3];
    return t;
}

inline UInt32 GetUInt32(const byte *ptr, size_t offset=0) /*FOLD00*/
{
    assert(ptr != 0);

    UInt32 t = (ptr+offset)[0] << 24 | (ptr+offset)[1] << 16 |
        (ptr+offset)[2] << 8  | (ptr+offset)[3];
    return t;
}

inline void PutInt16(Int16 val, byte *ptr, size_t offset=0) /*FOLD00*/
{
    assert(ptr != 0);

    ptr[offset] = (val >> 8) & 0xFF;
    ptr[offset+1] = val & 0xFF;

}

static UInt16 WingDings_b2c[] = {
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x2702,    (UInt16)0x2701, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x2706,    (UInt16)0x2709,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x2707,    (UInt16)0x270d, 
   (UInt16)0xfffd,    (UInt16)0x270c,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x2708,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x2744,    (UInt16)0xfffd,    (UInt16)0x271e,    (UInt16)0xfffd, 
   (UInt16)0x2720,    (UInt16)0x2721,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x2751,    (UInt16)0x2752,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x2756,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x2740, 
   (UInt16)0x273f,    (UInt16)0x275d,    (UInt16)0x275e,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x2780,    (UInt16)0x2781,    (UInt16)0x2782, 
   (UInt16)0x2783,    (UInt16)0x2784,    (UInt16)0x2785,    (UInt16)0x2786, 
   (UInt16)0x2787,    (UInt16)0x2788,    (UInt16)0x2789,    (UInt16)0xfffd, 
   (UInt16)0x278a,    (UInt16)0x278b,    (UInt16)0x278c,    (UInt16)0x278d, 
   (UInt16)0x278e,    (UInt16)0x278f,    (UInt16)0x2790,    (UInt16)0x2791, 
   (UInt16)0x2792,    (UInt16)0x2793,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x274d,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x2736,    (UInt16)0x2734,    (UInt16)0xfffd,    (UInt16)0x2735, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x272a,    (UInt16)0x2730,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x27a5,    (UInt16)0xfffd,    (UInt16)0x27a6,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x27a2,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x27b3,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x27a1,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x27a9,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x2717, 
   (UInt16)0x2713,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd
};

static UInt16 Symbols_b2c[] = {
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x2200,    (UInt16)0xfffd, 
   (UInt16)0x2203,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x220d, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x2217,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x2212,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x2245,    (UInt16)0x391,    (UInt16)0x392,    (UInt16)0x3a7, 
   (UInt16)0x394,    (UInt16)0x395,    (UInt16)0x3a6,    (UInt16)0x393, 
   (UInt16)0x397,    (UInt16)0x399,    (UInt16)0x3d1,    (UInt16)0x39a, 
   (UInt16)0x39b,    (UInt16)0x39c,    (UInt16)0x39d,    (UInt16)0x39f, 
   (UInt16)0x3a0,    (UInt16)0x398,    (UInt16)0x3a1,    (UInt16)0x3a3, 
   (UInt16)0x3a4,    (UInt16)0x3a5,    (UInt16)0x3c2,    (UInt16)0x3a9, 
   (UInt16)0x39e,    (UInt16)0x3a8,    (UInt16)0x396,    (UInt16)0xfffd, 
   (UInt16)0x2234,    (UInt16)0xfffd,    (UInt16)0x22a5,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x3b1,    (UInt16)0x3b2,    (UInt16)0x3c7, 
   (UInt16)0x3b4,    (UInt16)0x3b5,    (UInt16)0x3c6,    (UInt16)0x3b3, 
   (UInt16)0x3b7,    (UInt16)0x3b9,    (UInt16)0x3d5,    (UInt16)0x3ba, 
   (UInt16)0x3bb,    (UInt16)0x3bc,    (UInt16)0x3bd,    (UInt16)0x3bf, 
   (UInt16)0x3c0,    (UInt16)0x3b8,    (UInt16)0x3c1,    (UInt16)0x3c3, 
   (UInt16)0x3c4,    (UInt16)0x3c5,    (UInt16)0x3d6,    (UInt16)0x3c9, 
   (UInt16)0x3be,    (UInt16)0x3c8,    (UInt16)0x3b6,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x223c,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x3d2,    (UInt16)0xfffd,    (UInt16)0x2264, 
   (UInt16)0x2215,    (UInt16)0x221e,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x2218,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x2265, 
   (UInt16)0xfffd,    (UInt16)0x221d,    (UInt16)0xfffd,    (UInt16)0x2219, 
   (UInt16)0xfffd,    (UInt16)0x2260,    (UInt16)0x2261,    (UInt16)0x2248, 
   (UInt16)0x22ef,    (UInt16)0x2223,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x2297,    (UInt16)0x2295,    (UInt16)0x2205,    (UInt16)0x2229, 
   (UInt16)0x222a,    (UInt16)0x2283,    (UInt16)0x2287,    (UInt16)0x2284, 
   (UInt16)0x2282,    (UInt16)0x2286,    (UInt16)0x2208,    (UInt16)0x2209, 
   (UInt16)0xfffd,    (UInt16)0x2207,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x221a,    (UInt16)0x22c5, 
   (UInt16)0xfffd,    (UInt16)0x2227,    (UInt16)0x2228,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0x22c4,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0x2211,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0x222b,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd, 
   (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd,    (UInt16)0xfffd
};
#endif //__CMAPS_H
