/*
 * @(#)hsExceptions.h	1.3 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsExceptionDefined
#define hsExceptionDefined

#include "HeadSpin.h"

enum hsErrorEnum {
	kNo_hsError,
	kBadAlloc_hsError,
	kNilParam_hsError,
	kBadParam_hsError,
	kInternal_hsError,
	kOS_hsError
};

//////////////////////////////////////////////////////////////////////////////

class hsException {
public:
	hsErrorEnum	fError;
	long			fParam;
	
	hsException(hsErrorEnum error, long param = 0) : fError(error), fParam(param) {}
};

class hsBadAllocException : public hsException {
public:
	hsBadAllocException() : hsException(kBadAlloc_hsError) {}
};

class hsNilParamException : public hsException {
public:
	hsNilParamException() : hsException(kNilParam_hsError) {}
};

class hsBadParamException : public hsException {
public:
	hsBadParamException() : hsException(kBadParam_hsError) {}
};

class hsInternalException : public hsException {
public:
	hsInternalException() : hsException(kInternal_hsError) {}
};

class hsOSException : public hsException {
public:
	hsOSException(long error) : hsException(kOS_hsError, error) {}
};

/////////////////////////////////////////////////////////////////////////////////

inline void hsThrowIfNilParam(const void* p)
{
#ifndef CC_NOEX
	if (p == nil)
		throw hsNilParamException();
#endif // !CC_NOEX
}

inline void hsThrowIfBadParam(int trueIfBadParam)
{
#ifndef CC_NOEX
	if (trueIfBadParam)
		throw hsBadParamException();
#endif // !CC_NOEX
}

inline void hsThrowIfOSErr(long osErr)
{
#ifndef CC_NOEX
	if (osErr != 0)
		throw hsOSException(osErr);
#endif // !CC_NOEX
}

#endif
