/*
 * @(#)hsFixedTypes.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsFixedTypesDefined
#define hsFixedTypesDefined

#include "HeadSpin.h"
#if HS_BUILD_FOR_MAC
	#include <ToolUtils.h>
	#include <FixMath.h>
#endif

#define hsIntToFixed(x)		((hsFixed)(x) << 16)
#define hsFixedToInt(x)		((x) >> 16)
#define hsFixedRound(x)		((x) + 0x8000 >> 16)
#define hsFixed1			hsIntToFixed(1)
#define hsFixedPiOver2		(0x1921f)

#define hsFixedToFract(x)	((hsFract)(x) << 14)
#define hsFractToFixed(x)	((hsFixed)(x) >> 14)
#define hsFract1			hsFixedToFract(hsFixed1)

#if HS_CAN_USE_FLOAT
	#define hsFixedToFloat(x)		((x) / float(65536))
	#define hsFloatToFixed(x)		hsFixed((x) * hsFixed1)
#endif

#if HS_BUILD_FOR_MAC68K && !(HS_PIN_MATH_OVERFLOW)
	#define hsFixMul(a, b)	FixMul(a, b)
#else
	hsFixed hsFixMul(hsFixed a, hsFixed b);
#endif

#if HS_BUILD_FOR_MAC && !(HS_PIN_MATH_OVERFLOW) && !(HS_MP_SAFE)
	#define hsFixDiv(a, b)	FixDiv(a, b)
	#define hsFracMul(a, b)	FracMul(a, b)
	#define hsFracDiv(a, b)	FracDiv(a, b)
#else
	hsFract hsFixDiv(hsFixed a, hsFixed b);
	hsFract hsFracMul(hsFract a, hsFract b);
	hsFract hsFracDiv(hsFract a, hsFract b);
#endif

hsFract	hsFracSqrt(hsFract value);
#define	hsFixSqrt(value)	(hsFracSqrt(value) >> 7)
hsFract	hsFracCubeRoot(hsFract value);
hsFixed	hsFixedCos(hsFixed s);
hsFixed	hsFixedSin(hsFixed s);

UInt16	hsSqrt32(UInt32 value);
UInt16	hsCubeRoot(UInt32 value);
Int32	hsMulDiv(Int32 numer1, Int32 numer2, Int32 denom);
Int32	hsMagnitude(Int32 x, Int32 y);

#define kHighWordMask	0xFFFF0000

inline hsFixed hsFixFloor(hsFixed x)
{
	if (x < 0)
		return - hsFixed((-x + 0xFFFF) & kHighWordMask);
	else
		return x & kHighWordMask;
}

inline hsFixed hsFixCeiling(hsFixed x)
{
	if (x < 0)
		return - hsFixed(-x & kHighWordMask);
	else
		return (x + 0xFFFF) & kHighWordMask;
}

////////////////////////////////////////////////////////////////////////////////////

struct hsFixedPoint2 {
	hsFixed		fX, fY;

	hsFixedPoint2*	Set(hsFixed x, hsFixed y) { fX = x; fY = y; return this; }
};

////////////////////////////////////////////////////////////////////////////////////

struct hsFixedMatrix33 {
	hsFixed			fMap[3][3];

	hsMatrixType		GetType() const;

	hsFixedMatrix33*	Reset();
	hsFixedMatrix33*	SetScale(hsFixed sx, hsFixed sy, hsFixed px, hsFixed py);
	hsFixedMatrix33*	SetRotate(hsFixed degrees, hsFixed px, hsFixed py);
	hsFixedMatrix33*	SetTranslate(hsFixed dx, hsFixed dy);

	hsFixedMatrix33*	Scale(hsFixed sx, hsFixed sy, hsFixed px, hsFixed py);
	hsFixedMatrix33*	Rotate(hsFixed degrees, hsFixed px, hsFixed py);
	hsFixedMatrix33*	Translate(hsFixed dx, hsFixed dy);

	hsFixedMatrix33*	Invert(hsFixedMatrix33* inverse) const;
	hsFixedPoint2*		MapPoints(long count, const hsFixedPoint2 src[], hsFixedPoint2 dst[], hsMatrixType matType = kUnknownMatrixType) const;
	hsFixedPoint2*		MapPoints(long count, hsFixedPoint2 points[], hsMatrixType matType = kUnknownMatrixType) const;

	friend hsFixedMatrix33	operator*(const hsFixedMatrix33& a, const hsFixedMatrix33& b);
	friend int			operator==(const hsFixedMatrix33& a, const hsFixedMatrix33& b)
					{
						return	a.fMap[0][0] == b.fMap[0][0] && a.fMap[0][1] == b.fMap[0][1] && a.fMap[0][2] == b.fMap[0][2] &&
								a.fMap[1][0] == b.fMap[1][0] && a.fMap[1][1] == b.fMap[1][1] && a.fMap[1][2] == b.fMap[1][2] &&
								a.fMap[2][0] == b.fMap[2][0] && a.fMap[2][1] == b.fMap[2][1] && a.fMap[2][2] == b.fMap[2][2];
					}
	friend int			operator!=(const hsFixedMatrix33& a, const hsFixedMatrix33& b) { return !(a == b); }

	hsFixedMatrix33*	QuadToQuad(const hsFixedPoint2 source[], const hsFixedPoint2 destination[]);
};

////////////////////////////////////////////////////////////////////////////////////

struct hsFixedRect {
	hsFixed		fLeft, fTop, fRight, fBottom;

	hsFixed		Width() const { return fRight - fLeft; }
	hsFixed		Height() const { return fBottom - fTop; }
	hsFixed		CenterX() const { return fLeft + fRight >> 1; }
	hsFixed		CenterY() const { return fTop + fBottom >> 1; }
	hsFixedPoint2*	ToQuad(hsFixedPoint2 quad[4]) const;

	hsFixedRect*	Set(hsFixed left, hsFixed top, hsFixed right, hsFixed bottom)
				{
					fLeft = left; fTop = top; fRight = right; fBottom = bottom;
					return this;
				}
	hsFixedRect*	Set(const hsIntRect* r)
				{
					return Set(hsIntToFixed(r->fLeft), hsIntToFixed(r->fTop), hsIntToFixed(r->fRight), hsIntToFixed(r->fBottom));
				}
	hsFixedRect*	Transform(const hsFixedMatrix33* matrix);
	hsFixedRect*	Union(const hsFixedRect* r);
	hsFixedRect*	Union(Int32 count, const hsFixedPoint2 p[]);
	hsFixedRect*	UnionX(hsFixed x) { if (x < fLeft) fLeft = x; else if (x > fRight) fRight = x; return this; }
	hsFixedRect*	UnionY(hsFixed y) { if (y < fTop) fTop = y; else if (y > fBottom) fBottom = y; return this; }
	Boolean		Contains(hsFixed x, hsFixed y) const
				{
					return x >= fLeft && x < fRight && y >= fTop && y < fBottom;
				}
	hsIntRect*	Truncate(hsIntRect* dst)
				{
					return dst->Set(hsFixedToInt(fLeft), hsFixedToInt(fTop), hsFixedToInt(fRight), hsFixedToInt(fBottom));
				}
	hsIntRect*	Round(hsIntRect* dst)
				{
					return dst->Set(hsFixedRound(fLeft), hsFixedRound(fTop), hsFixedRound(fRight), hsFixedRound(fBottom));
				}
	hsIntRect*	RoundOut(hsIntRect* dst)
				{
					return dst->Set(hsFixFloor(fLeft), hsFixFloor(fTop),
								hsFixCeiling(fRight), hsFixCeiling(fBottom));
				}

	friend int		operator==(const hsFixedRect& a, const hsFixedRect& b)
				{
					return	a.fLeft == b.fLeft && a.fTop == b.fTop &&
							a.fRight == b.fRight && a.fBottom == b.fBottom;
				}
};

struct hsFixedPlane {
	hsFixed	fA, fB, fC;

	void		Set(hsFixed a, hsFixed b, hsFixed c) { fA = a; fB = b; fC = c; }

	hsFixed	FixEval(hsFixed x, hsFixed y) const { return hsFixMul(fA, x) + hsFixMul(fB, y) + fC; }
	Int32	IntEval(Int32 x, Int32 y) const { return fA * x + fB * y + fC; }
	void 		ShiftDown(UInt32 i) { fA >>= i; fB >>= i; fC >>= i;}
};

////////////////////////////////////////////////////////////////////////////////////

struct hsFixedPolyContour {
	long			fPointCount;
	hsFixedPoint2	fPoints[1];
};

struct hsFixedPolygon {
	long				fContourCount;
	hsFixedPolyContour	fContours[1];

	UInt32			ComputeSize() const;
	hsFixedRect*		ComputeBounds(hsFixedRect* bounds) const;
	void				Transform(const struct hsFixedMatrix33* matrix);

	UInt32			Read(hsStream* stream);
	UInt32			Write(hsStream* stream) const;
};

struct hsFixedPathContour {
	long			fPointCount;
	unsigned long	fControlBits[1];
//	hsFixedPoint2	fPoints[pointCount];
};

struct hsFixedPath {
	long				fContourCount;
	hsFixedPathContour	fContours[1];

	UInt32			ComputeSize() const;
	hsFixedRect*		ComputeBounds(hsFixedRect* bounds, Boolean exact) const;
	void				Transform(const struct hsFixedMatrix33* matrix);

	UInt32			Read(hsStream* stream);
	UInt32			Write(hsStream* stream) const;
};

typedef hsFixedPathContour	hsFixedCubicContour;

struct hsFixedCubic : hsFixedPath {
	hsFixedRect*	ComputeBounds(hsFixedRect* bounds, Boolean exact) const;
};


#define PolyContour_Next(c)		(hsFixedPolyContour*)((c)->fPoints + (c)->fPointCount)

#define ControlBits_Longs(count)	((count) + 31 >> 5)
#define ControlBits_Bytes(count)	(ControlBits_Longs(count) << 2)
#define PathContour_Points(c)		((hsFixedPoint2*)((char*)(c)->fControlBits + ControlBits_Bytes((c)->fPointCount)))
#define PathContour_Next(c)		((hsFixedPathContour*)(PathContour_Points(c) + (c)->fPointCount))

#endif
