/*
 * @(#)hsGBitmap.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGBitmapDefined
#define hsGBitmapDefined

#include "hsGColor.h"

#ifdef HS_DEBUGGING
	#define ASSERT_PIXELSIZE(bitmap, pixelsize)		hsAssert((bitmap)->fPixelSize == (pixelsize), "pixelSize mismatch")
#else
	#define ASSERT_PIXELSIZE(bitmap, pixelsize)
#endif

struct hsGBitmap {
	enum hsGBitmapFlags {
		kNone			= 0x0,
		kAlphaChannel	= 0x1,
		kAlphaBit		= 0x2
	};
	UInt32			fFlags;
	UInt32			fWidth;
	UInt32			fHeight;
	UInt32			fRowBytes;
	void*			fImage;
	UInt32			fPixelSize;	// 16, 32
	
	UInt32	ImageSize() const
			{
				return (UInt32)fHeight * (UInt32)fRowBytes;
			}
	void		Erase(const hsGColor* color);

	Boolean	EqualStructure(const hsGBitmap* bm) const
			{
				return	this->fFlags		== bm->fFlags &&
						this->fWidth		== bm->fWidth &&
						this->fHeight		== bm->fHeight &&
						this->fRowBytes	== bm->fRowBytes &&
						this->fPixelSize	== bm->fPixelSize;
			}

	virtual UInt32	Read(hsStream* stream);
	virtual UInt32	Write(hsStream* stream) const;

	//	The following methods are for accessing pixels in an hsGBitmap
	//	The 16, 32 refers to the bitdepth of the hsGBitmap
	//	They are meant to be fast, therefore they are inlined and do not check
	//	the fPixelSize field at runtime (except when debugging)

	//	These methods return the address of the pixel specified by x and y

	UInt8*	GetAddr8(unsigned x, unsigned y) const
			{
				ASSERT_PIXELSIZE(this, 8);
				return (UInt8*)((char*)fImage + y * fRowBytes + x);
			}
	UInt16*	GetAddr16(unsigned x, unsigned y) const
			{
				ASSERT_PIXELSIZE(this, 16);
				return (UInt16*)((char*)fImage + y * fRowBytes + (x << 1));
			}
	UInt32*	GetAddr32(unsigned x, unsigned y) const
			{
				ASSERT_PIXELSIZE(this, 32);
				return (UInt32*)((char*)fImage + y * fRowBytes + (x << 2));
			}
};
/*@
@page hsgbitmap.html

@struct hsGBitmap
This struct is used to define bitmaps both for rendering primitives and the device pixel buffer. Currently only 32 bit
bitmaps are supported as drawing primitives. The 32 bit images are in (A:8,R:8, G:8, B:8) format.  Both 32 and 16
bit (A:1, R:5, G:5, B:5) images can be used for the device pixel buffer.
be used as the device pixel buffer. 
@field UInt32 fWidth	the width of the bitmap in pixels
@field UInt32 fHeight the height of the bitmap in pixels
@field UInt32 fRowBytes the number of byte per row of the bitmap
@field void* fImage the address of the start of the bitmap pixels
@field UInt32 fPixelSize Number of bits per pixel (must be 16 or 32)
@method Erase clear the bitmap to a color
@method ImageSize return the number of bytes in this bitmap
@method Read read the bitmap from a stream object
@method Write write the bitmap to a stream object
@method GetAddr16 return the address of a pixel in a 16 bit bitmap
@method GetAddr32 return the address of a pixel in a 32 bit bitmap
@endstruct

@methoddef hsGBitmap Read
This function reads a bitmap in from the specified stream. If the bitmap's fImage pointer is currently pointing to an image,
that memory is deleteed before the new image is read in. hsGBitmap::Read allocates the memory for the newly read in bitmap. 
The method returns the number of bytes read
@param hsStream* stream the stream object to read from
@return UInt32 
@endmethod

@methoddef hsGBitmap Write
This function writes a bitmap to the specified stream. The method returns the number of bytes written
@param hsStream* stream the stream object to write to
@return UInt32 
@endmethod

@endpage
*/
#endif
