/*
 * @(#)hsGColor.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGColorDefined
#define hsGColorDefined

#include "HeadSpin.h"

typedef UInt16	hsGColorValue;

#define kHalf_hsGColorValue	0x8000
#define kFull_hsGColorValue	0xFFFF

struct hsGColor {
	hsGColorValue	fAlpha;
	union {
		struct {
			hsGColorValue	fRed;
			hsGColorValue	fGreen;
			hsGColorValue	fBlue;
		} fRGB;
	};

	void	SetRGB(hsGColorValue red, hsGColorValue green, hsGColorValue blue)
		{
			fRGB.fRed = red; fRGB.fGreen = green; fRGB.fBlue = blue;
		}
	void	SetARGB(hsGColorValue alpha, hsGColorValue red, hsGColorValue green, hsGColorValue blue)
		{
			fAlpha = alpha;
			fRGB.fRed = red; fRGB.fGreen = green; fRGB.fBlue = blue;
		}

	static int	EqualARGB(const hsGColor* a, const hsGColor* b)
		{
			return	a->fAlpha == b->fAlpha && a->fRGB.fRed == b->fRGB.fRed
					&& a->fRGB.fGreen == b->fRGB.fGreen && a->fRGB.fBlue == b->fRGB.fBlue;
		}

	static Byte ColorValueToByte(hsGColorValue value)
		{
			return value >> 8;
		}
	static hsGColorValue ByteToColorValue(Byte byte)
		{
			return (hsGColorValue(byte) << 8) | byte;
		}

	static unsigned ColorValueTo5(hsGColorValue value)
		{
			return value >> 11;
		}
	static unsigned FiveToColorValue(unsigned value)
		{
			return (value << 11) | (value << 6) | (value << 1) | (value >> 4);
		}
};
/*@
@page hsgcolor.html

@typedef hsGColorValue	UInt16
hsGColorValue specifies a component of a color. It ranges from 0 to 65535 (0xFFFF). In RGB, color values (0,0,0) represent black,
and (0xFFFF,0xFFFF,0xFFFF) represent white.

@struct hsGColor
This struct is used to define colors. Currently only RGB colors are supported. Each color is 
a 16 bit unsigned integer value. The alpha component is the opacity for the color. When alpha 
is 1 (0xffff) the color is completely opaque.
@field hsGColorValue fAlpha the alpha component
@field hsGColorValue fRGB.fRed the red component
@field hsGColorValue fRGB.fBlue the blue component
@field hsGColorValue fRGB.fGreen the green component
@method SetRGB sets the RGB components of the color. Alpha is untouched.
@method SetARGB sets the RGB and alpha components
@method EqualARGB returns true if two color are equal
@endclass

@endpage
*/
#endif

