/*
 * @(#)hsGFont.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGFontListDefined
#define hsGFontListDefined

#include "HeadSpin.h"

typedef UInt32 hsGFontID;

class hsGFontList {
public:
	enum NameID {
		kFamilyName,
		kStyleName,
		kFullName,
		kPostScriptName
	};

	static UInt32		Count();
	static hsGFontID	Get(UInt32 index);
	static hsGFontID	Find(NameID nameID, const char name[]);

	static UInt32		GetName(hsGFontID fontID, NameID nameID, char name[]);
	static Boolean		MatchName(hsGFontID fontID, NameID nameID, const char name[]);

	// This need only be called when an app terminates, and then only if it wants to worry
	// about delete all of its objects when looking for memory leaks.
	static void		KillFontList();
};

/*@
@page hsgfont.html
@typedef hsGFontID	UInt32
hsGFontID is a unique 32-bit identifier for a font. Font IDs are returned by methods on hsGFontList, and are passed to styles
via SetFontID().
@class hsGFontList
hsGFontList is a class with only static methods. It manages a global list of fonts. Fonts are identified by a 32-bit value called a
hsGFontID. The fontID is what you give to a style to select a font. The fontID uniquely identifies a font, meaning a particular style
within a font family such as Times Italic or Courier Bold. Fonts are not created or deleted by applications, the global font list is
responsible for managing them. To search for a font by name, use Find(), specifying the type of name you are providing. Possible
types are , kStyleName (Roman, Bold, Italic, Demi, etc.), kFullName (Times Italic, Courier Bold,
Helvetica, etc.), or kPostScriptName (Times-Roman, etc.).

@enum NameID
These specify the type of name for a font. They are a required parameter to the Find() method for finding a font from within the
global list, GetName() for retrieving a name from a font, and MatchName() for querying a font if it has a given name.
@token 	kFamilyName		Times, Helvetica, Courier, Symbol, etc.
@token 	kStyleName		Regular, Roman, Bold, Italic, Demi, Narrow, etc.
@token 	kFullName			Times Italic, Helvetica, Courier Bold, etc.
@token 	kPostScriptName	Times-Roman, Helvetica-Bold, etc.
@endflags

@method Count			Returns the number of available fonts. Use this to iterate throught the list with Get(index)
@method Get			Returns the fontID for a given index. Valid indices are 0..Count() - 1
@method Find			Returns the fontID for the font that matches the given name
@method GetName		Given a fontID, it returns the length and the text (if not nil) for the requested name
@method MatchName		Returns true if the given fontID has the specified name.
@method KillFontList		Optional method for deleting all references to fonts. Useful when checking for leaks.
@endclass

@methoddef hsGFontList	Count
	Returns the number of available fonts. Use this to iterate throught the list with Get(index)
	@return	UInt32
@endmethod
@methoddef hsGFontList	Get
	Returns the fontID for a given index. Valid indices are 0..Count() - 1
	@return	hsGFontID
@endmethod
@methoddef hsGFontList	Find
	Returns the fontID for the font that matches the given name
	@return	hsGFontID
@endmethod
@methoddef hsGFontList	GetName
	Given a fontID, it returns the length and the text (if not nil) for the requested name. If the name is not found, it returns 0.
	@return	UInt32
@endmethod
@methoddef hsGFontList	MatchName
	Returns true if the given fontID has the specified name.
	@return	Boolean
@endmethod
@methoddef hsGFontList	KillFontList
	Optional method for deleting all references to fonts. Useful when checking for leaks.
	@return	void
@endmethod

@endpage
*/

#endif
