/*
 * @(#)hsGFont_Internal.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGFontInternalDefined
#define hsGFontInternalDefined

#include "hsGFont.h"
#include "hsTemplates.h"

#if HS_BUILD_FOR_MAC
	#include <Files.h>
	#include <Folders.h>
	#include <Resources.h>

	typedef Handle	hsGSfntType;		// resource
	typedef FSSpec	hsGType1Type;		// rsrc ID
#elif HS_BUILD_FOR_WIN32
	#include "hsDirExplore.h"

	typedef FILE*	hsGSfntType;
	typedef FILE*	hsGType1Type;
#endif

#define kTrueType_hsGFontFormat		0x74727565	// 'true'
#define kType1_hsGFontFormat			0x74797031	// 'typ1'

class hsGFont {
private:
	hsGFontID			fFontID;
	UInt32			fFormat;
public:
					hsGFont(UInt32 format);
	virtual			~hsGFont();

	hsGFontID			GetID() const { return fFontID; }
	UInt32			GetFormat() const { return fFormat; }

	virtual UInt32		GetName(hsGFontList::NameID nameID, char name[]) = 0;
	virtual Boolean		MatchName(hsGFontList::NameID nameID, const char name[]) = 0;

	// These are low-level members for reading font data
	virtual const void*	GetData() = 0;		// sfnt for TrueType, filespec for Type1
	virtual UInt32		GetDataLength() = 0;
	virtual void		ReleaseData() = 0;

	static hsGFont* 	Find(hsGFontID fontID);
};

class hsGSfntFont : public hsGFont {
private:
	hsGSfntType		fSfnt;
	Int32			fLockCount;
	void*			fData;
public:
					hsGSfntFont(hsGSfntType sfnt, UInt32 format);
	virtual			~hsGSfntFont();

	virtual UInt32		GetName(hsGFontList::NameID nameID, char name[]);
	virtual Boolean		MatchName(hsGFontList::NameID nameID, const char name[]);

	virtual const void*	GetData();			// sfntDirectory*
	virtual UInt32		GetDataLength();	// size of sfnt
	virtual void		ReleaseData();		// decrement lockcount
};

#if HS_BUILD_FOR_MAC
class hsGMacType1Font : public hsGFont {
	FSSpec			fSpec;
public:
					hsGMacType1Font(const FSSpec* spec);

	virtual UInt32		GetName(hsGFontList::NameID nameID, char name[]);
	virtual Boolean		MatchName(hsGFontList::NameID nameID, const char name[]);

	virtual const void*	GetData();			// FSSpec*
	virtual UInt32		GetDataLength();	// post error
	virtual void		ReleaseData();		// post error
	
	static Boolean		SupportedFormat(const FSSpec* spec);
};
#endif

////////////////////////////////////////////////////////////////////////////

class hsGGlobalFontList : public hsDynamicArray<hsGFont*> {
#if HS_BUILD_FOR_WIN32
	hsDirExplore	fExplore;
#endif
	hsGFont*		fDefaultFont;
public:
				hsGGlobalFontList() : fDefaultFont(nil) {}
				~hsGGlobalFontList();

	hsGFont*		GetDefaultFont();
	hsGFont*		FindByID(hsGFontID fontID);
	hsGFont*		FindByName(hsGFontList::NameID nameID, const char name[]);
	void			LoadFonts();
};

#endif
