/*
 * @(#)hsGMatrixHandler.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGMatrixHandlerDefined
#define hsGMatrixHandlerDefined

#include "hsMatrix33.h"

class hsGView;
class hsGDevice;

class hsGMatrixHandler {
	hsGView*		fLastView;		// the last view passed to GetTotalMatrix()
	hsGSeedValue	fMatrixSeed;		// the matrixseed the last time fTotalMatrix was built
	hsGDevice*	fLastDevice;		// the last device passed to GetTotalMatrix()
	hsGSeedValue	fDeviceSeed;		// the deviceseed the last time fTotalMatrix was built

	hsMatrix33	fTotalMatrix;		// the cached answer
protected:
	hsMatrix33	fMatrix;			// the client's matrix
public:
				hsGMatrixHandler();

	hsMatrix33*	GetMatrix(hsMatrix33* matrix) const;
	void			SetMatrix(const hsMatrix33* matrix);

	void			Read(hsStream* stream);
	void			Write(hsStream* stream);

	const hsMatrix33*	GetTotalMatrix(hsGView* view, hsGDevice* device);
};

/*@
@page hsgmatrixhandler.html
@class hsGMatrixHandler
This class is used by the view heirarchy to keep track of the concatention of the transform matricies.
@method GetMatrix returns the current local matrix
@method SetMatrix updates the current local matrix
@methoddef hsGMatrixHandler SetMatrix
This function sets the current local matrix for the matrix handler object.
@param const hsMatrix33* matrix the new local matrix.
@endmethod
@endpage
*/
#endif
