/*
 * @(#)hsGeometry.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGeometryDefined
#define hsGeometryDefined

#include "hsMatrix33.h"

////////////////////////////////////////////////////////////////////////

struct hsPolyContour {
	UInt32		fPointCount;
	hsPoint2*	fPoints;
};

struct hsPolygon {
	UInt32		fContourCount;
	hsPolyContour*	fContours;

	hsRect*	        ComputeBounds(hsRect* bounds) const;
	void		Transform(const hsMatrix33* matrix);
	hsPolygon*      Copy(hsPolygon* dstOrNil) const;
	
	void		Read(hsStream* stream);
	void		Write(hsStream* stream) const;

	// This assumes that the poly was allocated with new for each array, like the Copy method
	static void Delete(hsPolygon* poly);
	static void Init(hsPolygon* poly)
	{
		poly->fContourCount	= 0;
		poly->fContours	= nil;
	}
};

struct hsPathContour {
	UInt32		fPointCount;
	hsPoint2*	fPoints;
	UInt32*		fControlBits;	// set to nil for a poly contour

	enum {
		kQuadOvalPointCount		= 8,
		kCubicOvalPointCount	= 12
	};
	void		SetQuadOval(const hsRect* r);
	void		SetCubicOval(const hsRect* r);

	static UInt32 ControlBitLongs(UInt32 pointCount) { return pointCount + 31 >> 5; }
	static UInt32 ControlBitBytes(UInt32 pointCount) { return ControlBitLongs(pointCount) << 2; }
};

struct hsPathSpline {
	enum Type {
		kQuadType	= 1,
		kCubicType
	};

	UInt32		fContourCount;
	hsPathContour*	fContours;

	void		Transform(const hsMatrix33* matrix);
	hsPathSpline*	Copy(hsPathSpline* dstOrNil) const;
	//	SoftCopy doesn't allocate anything
	hsPathSpline*	SoftCopy(hsPathSpline* dst) const;

	void		Read(hsStream* stream);
	void		Write(hsStream* stream) const;

	// This assumes that the poly was allocated with new for each array, like the Copy method
	static void Delete(hsPathSpline* path);
	static void Init(hsPathSpline* path)
	{
		path->fContourCount	= 0;
		path->fContours	= nil;
	}
};

typedef hsPathContour	hsQuadContour;

struct hsQuadraticSpline : hsPathSpline {
	hsRect*	ComputeBounds(hsRect* bounds, Boolean exact) const;
};

Boolean hsDivideQuadAtMax(hsScalar a, hsScalar b, hsScalar c, hsScalar* t);

typedef hsPathContour	hsCubicContour;

struct hsCubicSpline : hsPathSpline {
	hsRect*	ComputeBounds(hsRect* bounds, Boolean exact) const;
};

class CurveWalker {
	UInt32			count;
	const UInt32*		controlBits;
	const hsPoint2*	points;
	Bool32			timeToStop;
	Bool32			closed;
	UInt32			index_0, index_1;

	inline void		NextIndex();
	inline int		OnCurve(int index) const;
	inline void		SetAverage(hsPoint2* ave);

public:
	hsPoint2			a, b, c, d;			// OUTPUT - this is set by Next()
	Bool32			drawLine;			// OUTPUT - this is set by Next()

	void		Init(const hsPathContour* contour, Boolean closed);
	Boolean	NextQuad();
	Boolean	NextCubic();
};

#endif
