/*
 * @(#)hsGeometryUtils.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsGeometryUtilsDefined
#define hsGeometryUtilsDefined

#include "hsGeometry.h"

class hsPolygonCopy : public hsPolygon {
public:
	hsPolygonCopy()
	{
		fContourCount = 0;
	}
	hsPolygonCopy(const hsPolygon* src)
	{
		fContourCount = 0;
		this->CopyFrom(src);
	}
	~hsPolygonCopy()
	{
		hsPolygon::Delete(this);
	}
	void	CopyFrom(const hsPolygon* src)	
	{
		hsPolygon::Delete(this);
		(void)src->Copy(this);
	}
};

class hsPathCopy : public hsPathSpline {
public:
	hsPathCopy()
	{
		fContourCount = 0;
	}
	hsPathCopy(const hsPathSpline* src)
	{
		fContourCount = 0;
		this->CopyFrom(src);
	}
	~hsPathCopy()
	{
		hsPathSpline::Delete(this);
	}
	void	CopyFrom(const hsPathSpline* src)	
	{
		hsPathSpline::Delete(this);
		(void)src->Copy(this);
	}
};

class hsQuadraticCopy : public hsPathCopy {
public:
	hsQuadraticCopy() {}
	hsQuadraticCopy(const hsQuadraticSpline* src) : hsPathCopy(src) {}
	
	operator hsQuadraticSpline*() { return (hsQuadraticSpline*)this; }
	hsQuadraticSpline*	operator->() { return (hsQuadraticSpline*)this; }
};

class hsCubicCopy : public hsPathCopy {
public:
	hsCubicCopy() {}
	hsCubicCopy(const hsCubicSpline* src) : hsPathCopy(src) {}

	operator hsCubicSpline*() { return (hsCubicSpline*)this; }
	hsCubicSpline*	operator->() { return (hsCubicSpline*)this; }
};

template <class T> class hsTempGeo : public T {
public:
	hsTempGeo() { T::Init(this); }
	~hsTempGeo() { T::Delete(this); }
};

//////////////////////////////////////////////////////////////////////////////////

class hsQuadraticOval : public hsQuadraticSpline {
	hsQuadContour		fContourStorage;
	hsPoint2			fPointStorage[hsPathContour::kQuadOvalPointCount];
	UInt32			fControlBitStorage;
public:
					hsQuadraticOval() { fContourCount = 0; }
					hsQuadraticOval(const hsRect* r);
					
	void				SetOval(const hsRect* r);
};

class hsCubicOval : public hsCubicSpline {
	hsCubicContour		fContourStorage;
	hsPoint2			fPointStorage[hsPathContour::kCubicOvalPointCount];
	UInt32			fControlBitStorage;
public:
					hsCubicOval() { fContourCount = 0; }
					hsCubicOval(const hsRect* r);
					
	void				SetOval(const hsRect* r);
};

class hsQuadraticRRect : public hsQuadraticSpline {
	enum {
		kMaxPoints	= 16
	};
	hsQuadContour		fContourStorage;
	hsPoint2			fPointStorage[hsQuadraticRRect::kMaxPoints];
	UInt32			fControlBitStorage;
public:
	hsQuadraticRRect() { fContourCount = 0; }
	hsQuadraticRRect(const hsRect* r, hsScalar ovalWidth, hsScalar ovalHeight);

	void	SetRRect(const hsRect* r, hsScalar ovalWidth, hsScalar ovalHeight);
};

class hsQuadraticArc : public hsQuadraticSpline {
	enum {
		kMaxPoints	= 11
	};
	hsQuadContour		fContourStorage;
	hsPoint2			fPointStorage[hsQuadraticArc::kMaxPoints];
	UInt32			fControlBitStorage;
public:
	hsQuadraticArc() { fContourCount = 0; }
	hsQuadraticArc(const hsRect* r, hsScalar startAngle, hsScalar sweepAngle, Boolean wedge);

	void	SetArc(const hsRect* r, hsScalar startAngle, hsScalar sweepAngle, Boolean wedge);
};

#endif
