/*
 * @(#)hsMatrix33.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsMatrix33Defined
#define hsMatrix33Defined

#include "HeadSpin.h"

struct hsFixedMatrix33;

struct hsMatrix33 {
	hsScalar			fMap[3][3];

	hsScalar			TranslateX() const { return fMap[0][2]; }
	hsScalar			TranslateY() const { return fMap[1][2]; }

	hsMatrixType		GetType() const;

	hsMatrix33*		Reset();
	hsMatrix33*		Normalize();

	hsMatrix33*		SetScale(hsScalar sx, hsScalar sy, hsScalar px, hsScalar py);
	hsMatrix33*		SetRotate(hsScalar degrees, hsScalar px, hsScalar py);
	hsMatrix33*		SetTranslate(hsScalar x, hsScalar y);
	hsMatrix33*		SetConcat(const hsMatrix33* a, const hsMatrix33* b);
	hsMatrix33*		SetQuadToQuad(const hsPoint2 src[4], const hsPoint2 dst[4]);

	hsMatrix33*		Scale(hsScalar sx, hsScalar sy, hsScalar px, hsScalar py);
	hsMatrix33*		Rotate(hsScalar degrees, hsScalar px, hsScalar py);
	hsMatrix33*		Translate(hsScalar dx, hsScalar dy);

	hsMatrix33*		Invert(hsMatrix33* inverse) const;
	hsRect*			MapRect(const hsRect* src, hsRect* dst, hsMatrixType matType = kUnknownMatrixType) const;
	hsRect*			MapRect(hsRect* rect, hsMatrixType matType = kUnknownMatrixType) const;
	hsPoint2*			MapPoints(UInt32 count, const hsPoint2 src[], hsPoint2 dst[], hsMatrixType matType = kUnknownMatrixType) const;
	hsPoint2*			MapPoints(UInt32 count, hsPoint2 points[], hsMatrixType matType = kUnknownMatrixType) const;
	hsPoint2*			MapVectors(UInt32 count, const hsPoint2 src[], hsPoint2 dst[], hsMatrixType matType = kUnknownMatrixType) const;
	hsPoint2*			MapVectors(UInt32 count, hsPoint2 vectors[], hsMatrixType matType = kUnknownMatrixType) const;

	hsFixedMatrix33*	ToFixed33(hsFixedMatrix33* matrix) const;

	friend hsMatrix33	operator*(const hsMatrix33& a, const hsMatrix33& b);
	friend int			operator==(const hsMatrix33& a, const hsMatrix33& b)
					{
						return	a.fMap[0][0] == b.fMap[0][0] && a.fMap[0][1] == b.fMap[0][1] && a.fMap[0][2] == b.fMap[0][2] &&
								a.fMap[1][0] == b.fMap[1][0] && a.fMap[1][1] == b.fMap[1][1] && a.fMap[1][2] == b.fMap[1][2] &&
								a.fMap[2][0] == b.fMap[2][0] && a.fMap[2][1] == b.fMap[2][1] && a.fMap[2][2] == b.fMap[2][2];
					}
	friend int			operator!=(const hsMatrix33& a, const hsMatrix33& b) { return !(a == b); }
};
/*@
@page hsmatrix33.html

@struct hsMatrix33
This class is used to define transformation in 2D with perspective.
NewPoint = Matrix * OriginalPoint
@field hsScalar fMap[3][3] The values in the matrix
@method Reset			Replace the matrix with the identiy matrix
@method SetScale		Replace the matrix with a scaling matrix
@method SetRotate		Replace the matrix with a rotation matrix
@method SetTranslate	Replace the matrix with a translation matrix
@method QuadToQuad	Replace the matrix with a perspective matrix defined by 2 quadrilaterals
@method Scale			Apply scaling to the matrix
@method Rotate		Apply a rotation to the matrix
@method Translate		Apply a translation to the matrix
@method Invert			Compute the inverse of the matrix
@method MapRect		Apply the matrix to a rectangle
@method MapPoints		Apply the matrix to an array of points
@method MapVectors	Apply the matrix to an array of vectors
@endstruct

@methoddef hsMatrix33	SetScale
SetScale replaces the matrix with a scale matrix specified by the parameters. The "pivot" parameters
specify a coordinate that is left untouched by the matrix.
@return hsMatrix33*	The matrix
@param hsScalar scaleX	The x-scale factor
@param hsScalar scaleY	The y-scale factor
@param hsScalar pivotX	The x-coordinate of the pivot
@param hsScalar pivotY	The y-coordinate of the pivot
@endmethod

@methoddef hsMatrix33	SetRotate
SetRotate replaces the matrix with a rotation matrix specified by the parameters. The "pivot" parameters
specify a coordinate that is left untouched by the matrix.
@return hsMatrix33*	The matrix
@param hsScalar degrees	The rotation angle, in degrees, for clock-wise rotation
@param hsScalar pivotX	The x-coordinate of the pivot
@param hsScalar pivotY	The y-coordinate of the pivot
@endmethod

@methoddef hsMatrix33	SetTranslate
SetScale replaces the matrix with a translation matrix specified by the parameters.
@return hsMatrix33*	The matrix
@param hsScalar transX	The amout to translate in X
@param hsScalar transY	The amout to translate in Y
@endmethod

@methoddef hsMatrix33	QuadToQuad
The first array of 4 points specifies the source, and the second array of points specifies the destination.
The matrix is constructed to map the interior of the source quad to the interior of the destination quad.
@return hsMatrix33*	The matrix
@param const hsPoint2* src	The source array of 4 points
@param const hsPoint2* dst	The destination array of 4 points
@endmethod

@methoddef hsMatrix33	Scale
Scale modifies an existing matrix by concatenating a scaling matrix on its left. Result = Scale * Original
@return hsMatrix33*	The matrix
@param hsScalar scaleX	The x-scale factor
@param hsScalar scaleY	The y-scale factor
@param hsScalar pivotX	The x-coordinate of the pivot
@param hsScalar pivotY	The y-coordinate of the pivot
@endmethod

@methoddef hsMatrix33	Rotate
Rotate modifies an existing matrix by concatenating a rotation matrix on its left. Result = Rotate * Original
@return hsMatrix33*	The matrix
@param hsScalar degrees	The rotation angle, in degrees, for clock-wise rotation
@param hsScalar pivotX	The x-coordinate of the pivot
@param hsScalar pivotY	The y-coordinate of the pivot
@endmethod

@methoddef hsMatrix33	Translate
Translate modifies an existing matrix by concatenating a translation matrix on its left. Result = Translate * Original
@return hsMatrix33*	The matrix
@param hsScalar transX	The amout to translate in X
@param hsScalar transY	The amout to translate in Y
@endmethod

@methoddef hsMatrix33	const Invert
Invert computes the inverse of itself, and returns the result in the output parameter. If the matrix is non-invertible,
the function returns nil.
@return hsMatrix33*		The inverse matrix, or nil if the matrix is non-invertible
@param hsmatrix33* inverse	OUTPUT - the inverse of the matrix
@endmethod

@methoddef hsMatrix33	const MapRect
MapRect applies the matrix to a rectangle, returning the result in another rectangle. If the matrix contains
rotation or perspective, then the 4 corners of the source rectangle are mapped by the matrix, and the result
rectangle is the bounds of those 4 points.
@return hsRect*			The result rectangle
@param const hsRect* src	The source rectangle
@param hsRect* dst			OUTPUT - the result rectangle
@param hsMatrixType mType	OPTIONAL - the type of matrix
@endmethod

@methoddef hsMatrix33	const MapPoints
MapPoints applies the matrix to an array of source points, and returns the results in another array of points.
The source and result arrays may be the same.
@return hsPoint2*			The result points
@param UInt32 count		The number of points in the source and destination arrays
@param const hsPoint2* src	The array of source points
@param hsPoint2* dst		OUTPUT - the array of result points
@param hsMatrixType mType	OPTIONAL - the type of matrix
@endmethod

@methoddef hsMatrix33	const MapVectors
MapVectors applies the matrix to an array of source points, and returns the results in another array of points.
The source and result arrays may be the same. MapVectors differs from MapPoints in that any translation in
the matrix is ignored.
@return hsPoint2*			The result points
@param UInt32 count		The number of points in the source and destination arrays
@param const hsPoint2* src	The array of source points
@param hsPoint2* dst		OUTPUT - the array of result points
@param hsMatrixType mType	OPTIONAL - the type of matrix
@endmethod

@endpage
*/
#endif
