/*
 * @(#)hsMemory.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsMemoryDefined
#define hsMemoryDefined

#include "HeadSpin.h"
#if HS_BUILD_FOR_MAC
	#include <Memory.h>
#endif

#if HS_BUILD_FOR_WIN32
	#include <string.h>
#endif
#include <string.h>

class HSMemory {
public:
#if HS_BUILD_FOR_MAC
	static void	BlockMove(const void* src, void* dst, UInt32 length)
				{
					::BlockMoveData(src, dst, length);
				}
#elif HS_BUILD_FOR_WIN32
	static void	BlockMove(const void* src, void* dst, UInt32 length)
				{
					memmove(dst, src, length);
				}
#else
	static void	BlockMove(const void* src, void* dst, UInt32 length)
        {
            memmove(dst, src, length);
        }
#endif
	static void	Clear(void *m, UInt32 byteLen);
	static void	ClearMemory(void *m, UInt32 byteLen) { HSMemory::Clear(m, byteLen); }

	static Boolean	EqualBlocks(const void* block1, const void* block2, UInt32 length);

	static void*	New(UInt32 size);
	static void	Delete(void* block);
	static void*	Copy(UInt32 length, const void* source);
	
	static void*	SoftNew(UInt32 size);	// returns nil if can't allocate
};

class hsScratchMem {
	enum {
		kBufferSize = 32
	};
	UInt8*	fMem;
	UInt8	fMemBuffer[kBufferSize];
	UInt32	fLength;
public:
	hsScratchMem() : fLength(kBufferSize)
	{
		fMem = fMemBuffer;
	}
	~hsScratchMem()
	{
		if (fMem != fMemBuffer)
			delete[] fMem;
	}
	UInt8* GetMem(UInt32 length)
	{
		if (length > fLength)
		{	if (fMem != fMemBuffer)
				delete[] fMem;
			fMem = new UInt8[length];
			fLength = length;
		}
		return fMem;
	}
};

#endif

