/*
 * @(#)hsQuat.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef HSQUAT_inc
#define HSQUAT_inc

#include "hsVector3.h"

class hsQuat {

public:
	float fX,fY,fZ,fW;

	// Constructors
	hsQuat(){}
	hsQuat(float X, float Y, float Z, float W)  { fX = X; fY = Y; fZ = Z; fW = W; }
#if 0
	hsQuat(double X, double Y, double Z, double W)  { 
		fX = (float)X; fY = (float)Y; fZ = (float)Z; fW = (float)W; 
		}
#endif
	hsQuat(const hsQuat& a) { fX = a.fX; fY = a.fY; fZ = a.fZ; fW = a.fW; } 
	hsQuat(float af[4]) { fX = af[0]; fY = af[1]; fZ = af[2]; fW = af[3]; }
	hsQuat(float rad, hsVector3 axis);

	// Access methods
	hsScalar &x(void) { return fX; }
	hsScalar &y(void) { return fY; }
	hsScalar &z(void) { return fZ; }
	hsScalar &w(void) { return fW; }

	void Set(float X, float Y, float Z, float W)  { fX = X; fY = Y; fZ = Z; fW = W; }
	void GetAngleAxis(float *rad, hsVector3 *axis);
	
	// Access operators
	float& operator[](int i) { return (&fX)[i]; }     
	const float& operator[](int i) const { return (&fX)[i]; }  

	// Conversion function
	operator float*() { return(&fX); }

	// Unary operators
	hsQuat operator-() const { return(hsQuat(-fX,-fY,-fZ,-fW)); } 
	hsQuat operator+() const { return *this; } 

#if 0
	// Assignment operators
	hsQuat& operator-=(const hsQuat&);
	hsQuat& operator+=(const hsQuat&);
	hsQuat& operator*=(const hsQuat&);
	hsQuat& operator*=(float);
	hsQuat& operator/=(float);
#endif

	// Comparison
	int operator==(const hsQuat& a) const
	{ return (fX==a.fX && fY==a.fY && fZ==a.fZ && fW==a.fW); }

	void Identity() { fX = fY = fZ = (float)0.0; fW = (float) 1.0; }
	int IsIdentity() const
	{ return (fX==0.0 && fY==0.0 && fZ==0.0 && fW==1.0); }
	void Normalize();  
	void NormalizeIfNeeded();  
	void MakeMatrix(hsMatrix44 *mat) const;
	float Magnitude();
	float MagnitudeSquared();
	hsQuat Conjugate(const hsQuat &q)
	{ return hsQuat(-q.fX,-q.fY,-q.fZ,q.fW); }


	// Binary operators
	hsQuat operator-(const hsQuat&) const;
	hsQuat operator+(const hsQuat&) const;
	hsQuat operator*(const hsQuat&) const;
	hsQuat operator*(float f) const
		{ return hsQuat(fX*f,fY*f,fZ*f,fW*f); }
	hsQuat operator/(const hsQuat&) const;

	//
	void SetFromSlerp(hsQuat *q1, hsQuat *q2, float t, int spin=0);
	float Dot(const hsQuat &q2)
		{	return (fX*q2.fX + fY*q2.fY + fZ*q2.fZ + fW*q2.fW); }

	// I/O
	void Read(hsStream *stream);
	void Write(hsStream *stream);

	};
#endif
