/*
 * @(#)hsScalar.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsScalarMacrosDefined
#define hsScalarMacrosDefined

#include <math.h>

#ifndef WIN32
#define MATH_H_NEEDS_FLOAT_VERSIONS
#endif

#ifndef HS_SCALAR_IS_FLOAT
	#define HS_SCALAR_IS_FIXED		0
	#define HS_SCALAR_IS_FLOAT		1
	#define HS_NEVER_USE_FLOAT		0
#endif

#if HS_SCALAR_IS_FLOAT && HS_NEVER_USE_FLOAT
	#error	cant define HS_SCALAR_IS_FLOAT and HS_NEVER_USE_FLOAT
#endif

#if HS_SCALAR_IS_FIXED
	typedef hsFixed		hsScalar;

	#define hsScalar1			hsFixed1
	#define hsScalarHalf			(hsFixed1 >> 1)
	#define hsScalarPI			(0x3243F)

	#if HS_CAN_USE_FLOAT
		#define hsFloatToScalar(x)	hsFixed((x) * float(hsFixed1))
		#define hsScalarToFloat(x)	((x) / float(hsFixed1))
	#endif

	#define hsIntToScalar(x)		hsIntToFixed(x)
	#define hsScalarToInt(x)		hsFixedToInt(x)
	#define hsScalarRound(x)		hsFixedRound(x)

	#define hsFixedToScalar(x)	(x)
	#define hsScalarToFixed(x)	(x)

	#define hsFractToScalar(x)	hsFractToFixed(x)
	#define hsScalarToFract(x)	hsFixedToFract(x)

	#define hsScalarMul(a, b)		hsFixMul(a, b)
	#define hsScalarMul2(a)		((a) << 1)
	#define hsScalarDiv(a, b)		hsFixDiv(a, b)
	#define hsScalarDiv2(a)		((a) >> 1)
	#define hsScalarInvert(a)		hsFixDiv(hsFixed1, a)

	#define hsScalarMulDiv(n1, n2, d)	hsMulDiv(n1, n2, d)

	#define hsSquareRoot(scalar)	hsSqrt32(scalar)
	#define hsSine(angle)		hsFixedSin(angle)
	#define hsCosine(angle)		hsFixedCos(angle)
#ifdef __cplusplus
	#if HS_CAN_USE_FLOAT
		inline hsScalar hsACosine(hsScalar value)
					{
						return hsFloatToScalar(acosf(hsScalarToFloat(value)));
					}
		inline hsScalar hsPow(hsScalar base, hsScalar exponent)
					{
						return hsFloatToScalar(powf(hsScalarToFloat(base), hsScalarToFloat(exponent)));
					}
		inline hsScalar hsATan2(hsScalar y, hsScalar x)
					{
						return hsFloatToScalar(atan2f(hsScalarToFloat(y), hsScalarToFloat(x)));
					}
	#endif
	inline hsScalar hsCeil(hsScalar x) { return (x + 0xFFFF) & 0xFFFF0000; }
	inline hsScalar hsFloor(hsScalar x) { return x + 0xFFFF0000; }
#endif 
#endif
#if HS_SCALAR_IS_FLOAT
	typedef float		hsScalar;

	#define hsScalar1			float(1)
	#define hsScalarHalf			float(0.5)
	#define hsScalarPI			float(3.1415927)

	#define hsFloatToScalar(x)	float(x)
	#define hsScalarToFloat(x)	float(x)

	#define hsIntToScalar(x)		float(x)
	#define hsScalarToInt(x)		Int32(x)


	#define hsFixedToScalar(x)	((hsScalar)(x) / float(hsFixed1))
	#define hsScalarToFixed(x)	hsFixed((x) * float(hsFixed1))

	#define hsFractToScalar(x)	((x) / float(hsFract1))
	#define hsScalarToFract(x)	hsFract((x) * float(hsFract1))
#ifdef __cplusplus
	inline Int32 hsScalarRound(float x)
	{
		float half = hsScalarHalf;
		if (x < 0)
			half = -half;
		return Int32(x + half);
	}

	inline hsScalar hsScalarMul(hsScalar a , hsScalar b) { return a*b; }
	inline hsScalar hsScalarMul2(hsScalar a ) { return a*2.0f; }
	inline hsScalar hsScalarDiv(hsScalar a , hsScalar b) { return a/b; }
	inline hsScalar hsScalarDiv2(hsScalar a ) { return a*0.5f; }
	inline hsScalar hsScalarInvert(hsScalar a ) { return 1.0f/a; }

	inline hsScalar hsScalarMulDiv(hsScalar n1, hsScalar n2, hsScalar d ) { return (n1*n2)/d; }

	inline hsScalar ScalarAverage(hsScalar a, hsScalar b) { return (a + b) / hsScalar(2); }
	inline hsScalar ScalarAverage(hsScalar a, hsScalar b, hsScalar t) { return a + t * (b - a); }

#if HS_BUILD_FOR_WIN32
	// Note: turn this off for VC++ 4.2 or higher
#undef MATH_H_NEEDS_FLOAT_VERSIONS
#endif

#define hsScalarDegToRad(deg)		hsScalarMulDiv(deg, hsScalarPI, hsIntToScalar(180))
#define hsScalarRadToDeg(rad)		hsScalarMulDiv(rad, hsIntToScalar(180), hsScalarPI)

#ifdef MATH_H_NEEDS_FLOAT_VERSIONS	
inline float acosf(float _X)
	{return ((float)acos((double)_X)); }
inline float asinf(float _X)
	{return ((float)asin((double)_X)); }
inline float atanf(float _X)
	{return ((float)atan((double)_X)); }
inline float atan2f(float _X, float _Y)
	{return ((float)atan2((double)_X, (double)_Y)); }
inline float ceilf(float _X)
	{return ((float)ceil((double)_X)); }
inline float cosf(float _X)
	{return ((float)cos((double)_X)); }
inline float coshf(float _X)
	{return ((float)cosh((double)_X)); }
inline float expf(float _X)
	{return ((float)exp((double)_X)); }
inline float fabsf(float _X)
	{return ((float)fabs((double)_X)); }
inline float floorf(float _X)
	{return ((float)floor((double)_X)); }
inline float fmodf(float _X, float _Y)
	{return ((float)fmod((double)_X, (double)_Y)); }
inline float logf(float _X)
	{return ((float)log((double)_X)); }
inline float log10f(float _X)
	{return ((float)log10((double)_X)); }
/*inline float modff(float _X, float *_Y)
	{ double _Di, _Df = modf((double)_X, &_Di);
	*_Y = (float)_Di;
	return ((float)_Df); }*/
inline float powf(float _X, float _Y)
	{return ((float)pow((double)_X, (double)_Y)); }
inline float sinf(float _X)
	{return ((float)sin((double)_X)); }
inline float sinhf(float _X)
	{return ((float)sinh((double)_X)); }
inline float sqrtf(float _X)
	{return ((float)sqrt((double)_X)); }
inline float tanf(float _X)
	{return ((float)tan((double)_X)); }
inline float tanhf(float _X)
	{return ((float)tanh((double)_X)); }
#endif

	inline hsScalar hsSquareRoot(hsScalar scalar) { return sqrtf(scalar); }
	inline hsScalar hsSine(hsScalar angle) { return sinf(angle); }
	inline hsScalar hsCosine(hsScalar angle) { return cosf(angle); }
	inline hsScalar hsACosine(hsScalar value) { return acosf(value); }
	inline hsScalar hsPow(hsScalar base, hsScalar exponent) { return powf(base, exponent); }
	inline hsScalar hsATan2(hsScalar y, hsScalar x) { return atan2f(y, x); }
	inline hsScalar hsCeil(hsScalar x) { return ceilf(x); }
	inline hsScalar hsFloor(hsScalar x) { return floorf(x); }
#endif
#endif // __CPLUSPLUS
#endif
