/*
 * @(#)hsSfnt.h	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#ifndef hsSfntDefined
#define hsSfntDefined

#include "HeadSpin.h"

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

#if PRAGMA_IMPORT_SUPPORTED
#pragma import on
#endif
 
typedef UInt32 sfntTableTag;

struct sfntDirectoryEntry {
	sfntTableTag	tableTag;
	UInt32		checkSum;
	UInt32		offset;
	UInt32		length;
};
typedef struct sfntDirectory sfntDirectory;

/* The search fields limits numOffsets to 4096. */
struct sfntDirectory {
	sfntTableTag	format;
	UInt16		numOffsets;			/* number of tables */
	UInt16		searchRange;			/* (max2 <= numOffsets)*16 */
	UInt16		entrySelector;			/* log2(max2 <= numOffsets) */
	UInt16		rangeShift;			/* numOffsets*16-searchRange*/
	struct sfntDirectoryEntry table[1];		/* table[numOffsets] */
};
typedef struct sfntDirectoryEntry sfntDirectoryEntry;

enum {
	sizeof_sfntDirectory		= 12
};

/* Cmap - character id to glyph id gxMapping */
#define cmapFontTableTag			0x636D6170		// 'cmap'

struct sfntCMapSubHeader {
	UInt16		format;
	UInt16		length;
	UInt16		languageID;		/* base-1 */
};
typedef struct sfntCMapSubHeader sfntCMapSubHeader;

enum {
	sizeof_sfntCMapSubHeader	= 6
};

struct sfntCMapEncoding {
	UInt16		platformID;		/* base-0 */
	UInt16		scriptID;			/* base-0 */
	UInt32		offset;
};
typedef struct sfntCMapEncoding sfntCMapEncoding;

enum {
	sizeof_sfntCMapEncoding		= 8
};

struct sfntCMapHeader {
	UInt16				version;
	UInt16				numTables;
	struct sfntCMapEncoding	encoding[1];
};
typedef struct sfntCMapHeader sfntCMapHeader;

enum {
	sizeof_sfntCMapHeader		= 4
};

/* Name table */
#define nameFontTableTag			0x6E616D65		// 'name'

struct sfntNameRecord {
	UInt16	platformID;		/* base-0 */
	UInt16	scriptID;			/* base-0 */
	UInt16	languageID;		/* base-0 */
	UInt16	nameID;			/* base-0 */
	UInt16	length;
	UInt16	offset;
};
typedef struct sfntNameRecord sfntNameRecord;

enum {
	sizeof_sfntNameRecord	= 12
};

struct sfntNameHeader {
	UInt16			format;
	UInt16			count;
	UInt16			stringOffset;
	sfntNameRecord	rec[1];
};
typedef struct sfntNameHeader sfntNameHeader;

enum {
	sizeof_sfntNameHeader	= 6
};

#define kSfnt_WildCard	(0xFFFF)		// 16-bits explicitly, so that if we swap it, we get the same thing

enum {
	kNoFontName = kSfnt_WildCard,
	kCopyrightFontName = 0,
	kFamilyFontName,
	kStyleFontName,
	kUniqueFontName,
	kFullFontName,
	kVersionFontName,
	kPostscriptFontName,
	kTrademarkFontName,
	kManufacturerFontName,

	kNoPlatform = kSfnt_WildCard,
	kUnicodePlatform = 0,
	kMacintoshPlatform,

	kNoScript = kSfnt_WildCard,
	kRomanScript = 0,

	kNoLanguage = kSfnt_WildCard,
	kEnglishLanguage = 0
};

#if 0		// do these later

/* Fvar table - gxFont variations */
enum {
	variationFontTableTag		= 'fvar'
};

/* These define each gxFont variation */
struct sfntVariationAxis {
	gxFontVariationTag				axisTag;
	Fixed							minValue;
	Fixed							defaultValue;
	Fixed							maxValue;
	short							flags;
	short							nameID;
};
typedef struct sfntVariationAxis sfntVariationAxis;


enum {
	sizeof_sfntVariationAxis	= 20
};

/* These are named locations in gxStyle-space for the user */
struct sfntInstance {
	short							nameID;
	short							flags;
	Fixed							coord[1];					/* [axisCount] */
/* room to grow since the header carries a tupleSize field */
};
typedef struct sfntInstance sfntInstance;


enum {
	sizeof_sfntInstance			= 4
};

struct sfntVariationHeader {
	Fixed							version;					/* 1.0 Fixed */
	unsigned short					offsetToData;				/* to first axis = 16*/
	unsigned short					countSizePairs;				/* axis+inst = 2 */
	unsigned short					axisCount;
	unsigned short					axisSize;
	unsigned short					instanceCount;
	unsigned short					instanceSize;
/* other <count,size> pairs */
	struct sfntVariationAxis		axis[1];					/* [axisCount] */
	struct sfntInstance				instance[1];				/* [instanceCount]  other arrays of data */
};
typedef struct sfntVariationHeader sfntVariationHeader;


enum {
	sizeof_sfntVariationHeader	= 16
};

/* Fdsc table - gxFont descriptor */
enum {
	descriptorFontTableTag		= 'fdsc'
};

struct sfntDescriptorHeader {
	Fixed							version;					/* 1.0 in Fixed */
	long							descriptorCount;
	gxFontDescriptor				descriptor[1];
};
typedef struct sfntDescriptorHeader sfntDescriptorHeader;


enum {
	sizeof_sfntDescriptorHeader	= 8
};

/* Feat Table - layout feature table */
enum {
	featureFontTableTag			= 'feat'
};

struct sfntFeatureName {
	unsigned short					featureType;
	unsigned short					settingCount;
	long							offsetToSettings;
	unsigned short					featureFlags;
	unsigned short					nameID;
};
typedef struct sfntFeatureName sfntFeatureName;

struct sfntFontRunFeature {
	unsigned short					featureType;
	unsigned short					setting;
};
struct sfntFeatureHeader {
	long							version;					/* 1.0 */
	unsigned short					featureNameCount;
	unsigned short					featureSetCount;
	long							reserved;					/* set to 0 */
	struct sfntFeatureName			names[1];
	struct gxFontFeatureSetting		settings[1];
	struct sfntFontRunFeature		runs[1];
};
typedef struct sfntFeatureHeader sfntFeatureHeader;

/* OS/2 Table */

enum {
	os2FontTableTag				= 'OS/2'
};

/*  Special invalid glyph ID value, useful as a sentinel value, for example */
enum {
	nonGlyphID					= 65535
};

#endif	// do these later

class hsSfnt {
public:
	// This allocates its return result using hsMemory::New
	static sfntDirectory*	ReadDirectory(FILE* sfntFile);

	// These take a pointer to the font data

	static UInt32		ComputeSize(const sfntDirectory* dir);
	static const char*	FindTable(const sfntDirectory* sfnt, sfntTableTag tableTag);

	static UInt32		FindName(const sfntNameHeader* table, UInt16 nameID, UInt16 plat, UInt16 script, UInt16 lang, UInt32* index);	// return length
	static UInt32		GetName(const sfntNameHeader* table, UInt32 index, Byte name[]);	// return length
	static Boolean		MatchName(const sfntNameHeader* table, UInt16 nameID, UInt16 plat,
								UInt16 script, UInt16 lang, int length, const Byte name[]);
};

#if PRAGMA_IMPORT_SUPPORTED
#pragma import off
#endif

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=reset
#endif

#endif
