/*
 * @(#)hsGGlyphCache.cpp	1.13 98/10/07
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Copyright (C) 1996-1997 all rights reserved by HeadSpin Technology Inc. Chapel Hill, NC USA
 *
 * This software is the property of HeadSpin Technology Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * HEADSPIN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * This information in this software is subject to change without notice
*/

#if !defined(HS_BUILD_FOR_MAC)
#include <new.h>	// For _set_new_handler
#endif
#include <assert.h>
#include "hsGGlyphCache.h"
#include "hsTemplates.h"
#include "hsGFontScaler.h"

///////////////////////////////////////////////////////////////////////////////////
#define HS_BUILD_FOR_WIN32 1
typedef void (*NewHandlerProc)();
typedef void (*new_handler)(); 
// new_handler set_new_handler(new_handler); 

class hsGGlyphCache {
	static void		New_Handler();
	static NewHandlerProc	gPrevHandler;
	static hsGGlyphCache*	gCache;

	hsDynamicArray<hsGGlyphStrike*>	fStrikeList;

	Boolean			Purge();

					hsGGlyphCache() {}
	virtual			~hsGGlyphCache();
public:
	hsGGlyphStrike*	FindStrike(hsConstDescriptor desc);
	void				AddStrike(hsGGlyphStrike* strike);

	static void		Kill();
	static hsGGlyphCache* GetGlobalCache()
	{
		if (gCache == nil)
		{	gCache = new hsGGlyphCache();
	#if HS_BUILD_FOR_WIN32
		//	gPrevHandler = _set_new_handler(hsGGlyphCache::New_Handler);
	#else
//                		gPrevHandler = set_new_handler(hsGGlyphCache::New_Handler);
	#endif
		}
		return gCache;
	}
};

NewHandlerProc hsGGlyphCache::gPrevHandler;
hsGGlyphCache* hsGGlyphCache::gCache = nil;

///////////////////////////////////////////////////////////////////////////////////

struct hsGGlyphStrikeEntry {
	hsGGlyph		fGlyph;
	hsFixedPoint2	fAdvance;
};

#define kNoMetrics_Width		0xFFFF	// marks the fWidth field to indicate that there are no valid metrics yet
#define kNoImage_Image		nil		// marks the fImage field to indicate that there is no valid image yet

hsGGlyphStrike::hsGGlyphStrike(hsConstDescriptor desc)
{
	fDesc = hsDescriptor_Copy(desc);

        UInt32 scalerID = hsDescriptor_Find32(desc, kScalerID_FontScalerDesc);
        hsGFontScaler*	scaler = hsGFontScaler::Find(scalerID);
	hsThrowIfNilParam(scaler);

	fScalerContext	= scaler->CreateContext(desc);
	fGlyphCount	= fScalerContext->CountGlyphs();
	fEntries	= new hsGGlyphStrikeEntry[fGlyphCount];

	for (unsigned i = 0; i < fGlyphCount; i++)
	{	fEntries[i].fGlyph.fWidth = kNoMetrics_Width;
		fEntries[i].fGlyph.fImage = kNoImage_Image;
	}
        if (fGlyphCount) {
            fScalerContext->GetLineHeight(&fAscent, &fDescent, &fBaseline,
                                          &fLeading, &fMaxAdvance);
        }
}

hsGGlyphStrike::~hsGGlyphStrike()
{
	for (unsigned i = 0; i < fGlyphCount; i++)
		if (fEntries[i].fGlyph.fImage != kNoImage_Image)
			HSMemory::Delete((void*)fEntries[i].fGlyph.fImage);
	delete[] fEntries;
	delete fScalerContext;
	
	hsDescriptor_Delete(fDesc);
}

hsPathSpline::Type hsGGlyphStrike::GetPathType() const
{
	return fScalerContext->GetPathType();
}

CharToGlyphMapper *
hsGGlyphStrike::getMapper()
{
    return fScalerContext->getMapper();
}

void hsGGlyphStrike::GetMetrics(UInt16 index, hsGGlyph* glyph, hsFixedPoint2* advance)
{
	hsAssert(index < fGlyphCount, "bad index");
        if (index >= fGlyphCount) {
            if (glyph) {
                memset(glyph, 0, sizeof(*glyph));
            }
            if (advance) {
                memset(advance, 0, sizeof(*advance));
            }
            return;
        }

	hsGGlyphStrikeEntry* entry = &fEntries[index];
	
	if (entry->fGlyph.fWidth == kNoMetrics_Width) {
            fScalerContext->GenerateMetrics(index, &entry->fGlyph, &entry->fAdvance);
            hsAssert(entry->fGlyph.fWidth != kNoMetrics_Width, "bad width flag");
	}
	if (glyph)
		*glyph = entry->fGlyph;
	if (advance)
		*advance = entry->fAdvance;
}

void hsGGlyphStrike::GetMetricsWithImage(UInt16 index, hsGGlyph* glyph, hsFixedPoint2* advance)
{
	hsAssert(index < fGlyphCount, "bad index");
        if (index >= fGlyphCount) {
            if (glyph) {
                memset(glyph, 0, sizeof(*glyph));
            }
            if (advance) {
                memset(advance, 0, sizeof(*advance));
            }
            return;
        }

	hsGGlyphStrikeEntry* entry = &fEntries[index];
	
	if (        (entry->fGlyph.fWidth == kNoMetrics_Width)
                ||  (entry->fGlyph.fImage == kNoImage_Image)) {
            fScalerContext->GenerateMetricsWithImage(index, &entry->fGlyph, &entry->fAdvance);
            hsAssert(entry->fGlyph.fWidth != kNoMetrics_Width, "bad width flag");
            hsAssert(entry->fGlyph.fImage != kNoImage_Image, "bad image");
	}
	if (glyph)
		*glyph = entry->fGlyph;
	if (advance)
		*advance = entry->fAdvance;
}

void hsGGlyphStrike::GetItalicAngle(hsFixedPoint2* italicAngle) {
    fScalerContext->GetCaretAngle(italicAngle);
}

const void* hsGGlyphStrike::GetImage(UInt16 index)
{
	hsAssert(index < fGlyphCount, "bad index");
        if (index >= fGlyphCount) {
            return nil;
        }

	hsGGlyphStrikeEntry* entry = &fEntries[index];

	if (entry->fGlyph.fImage == kNoImage_Image) {
            UInt32 imageSize = entry->fGlyph.fHeight * entry->fGlyph.fRowBytes;
            void* image = HSMemory::SoftNew(imageSize);

            if (image) {
#ifndef CC_NOEX
                try {
#endif // !CC_NOEX
                    fScalerContext->GenerateImage(index, &entry->fGlyph, image);
#ifndef CC_NOEX
                }
                catch (...) {
                    HSMemory::Delete(image);
                    image = kNoImage_Image;
                }
#endif // !CC_NOEX
            }
            entry->fGlyph.fImage = image;
	}
	return entry->fGlyph.fImage;
}

void hsGGlyphStrike::GetPath(UInt16 index, hsPathSpline* path, hsFixedPoint2* advance)
{
    hsAssert(index < fGlyphCount, "bad index");
    if (index >= fGlyphCount) {
        return;
    }
    fScalerContext->GeneratePath(index, path, advance);
}

void hsGGlyphStrike::GetLineHeight(hsFixedPoint2* ascent, hsFixedPoint2* descent,
                                   hsFixedPoint2* baseline, hsFixedPoint2 *leading,
                                   hsFixedPoint2* maxAdvance)
{
	if (ascent)
		*ascent = fAscent;
	if (descent)
		*descent = fDescent;
	if (baseline)
		*baseline = fBaseline;
        if (leading)
                *leading = fLeading;
        if (maxAdvance)
                *maxAdvance = fMaxAdvance;
}

hsGGlyphStrike* hsGGlyphStrike::RefStrike(hsConstDescriptor desc)
{
	hsGGlyphCache* cache = hsGGlyphCache::GetGlobalCache();
	hsGGlyphStrike* strike = cache->FindStrike(desc);

	if (strike == nil) {
            strike = new hsGGlyphStrike(desc);
            cache->AddStrike(strike);
	}
        // this is balanced by UnRef() the client when they're done using it
	strike->Ref();	
	
        return strike;
}

void hsGGlyphStrike::KillGlyphCache()
{
	hsGGlyphCache::Kill();
}

Boolean hsGGlyphStrike::isStyleSupported(int aStyle)
{
    if (fScalerContext) {
        return fScalerContext->isStyleSupported(aStyle);
    }
    return false;
}

///////////////////////////////////////////////////////////////////////////////////

static Boolean unrefStrike(hsGGlyphStrike*& strike)
{
	hsAssert(strike->RefCnt() == 1, "too many refs on strike");
	strike->UnRef();
	return false;
}

hsGGlyphCache::~hsGGlyphCache()
{
	(void)fStrikeList.ForEach(unrefStrike);
}

hsGGlyphStrike* hsGGlyphCache::FindStrike(hsConstDescriptor desc)
{
    hsGGlyphStrike**	strikeList = fStrikeList.AcquireArray();
    hsGGlyphStrike**	strike = strikeList;
    hsGGlyphStrike**	stop = strike + fStrikeList.GetCount();
    
    while (strike != stop) {
        if (hsDescriptor_Equal((*strike)->GetDesc(), desc)) {
            fStrikeList.ReleaseArray(strikeList);
            return *strike;
        }
        strike += 1;
    }
    fStrikeList.ReleaseArray(strikeList);
    return nil;
}

#define kStrikeLimit		256

void hsGGlyphCache::AddStrike(hsGGlyphStrike* strike)
{
	if (fStrikeList.GetCount() > kStrikeLimit)
	{	Boolean	result = this->Purge();
		hsAssert(result, "cache is locked");
	}
	fStrikeList.Append(strike);
}

Boolean hsGGlyphCache::Purge()
{
	Boolean didPurge = false;

	for (int i = fStrikeList.GetCount() - 1; i >= 0; i--)
	{	hsGGlyphStrike* strike = fStrikeList.Get(i);
	
		if (strike->RefCnt() == 1)
		{	delete strike;
			fStrikeList.Remove(i);
			didPurge = true;
		}
	}
	return didPurge;
}

void hsGGlyphCache::Kill()
{
	if (gCache)
	{	delete gCache;
		gCache = nil;
#if HS_BUILD_FOR_WIN32
//		set_new_handler(gPrevHandler);
#else
//                	set_new_handler(gPrevHandler);
#endif
	}
}

void hsGGlyphCache::New_Handler()
{
	if (gCache && gCache->Purge())
		return;

	if (gPrevHandler)
		gPrevHandler();
#ifndef CC_NOEX
	else
		throw "new failed";
#endif // !CC_NOEX
}

#undef HS_BUILD_FOR_WIN32
