/*
 * @(#)shapet.c	1.7 98/07/21
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * SHAPET.c
 * Copyright (C) 1989-1998 all rights reserved by Type Solutions, Inc. Plaistow, NH, USA.
 * Author: Sampo Kaasila
 *
 * This software is the property of Type Solutions, Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
#include "syshead.h"

#include "config.h"
#include "dtypes.h"
#include "tsimem.h"
#include "t2kstrm.h"
#include "glyph.h"
#include "autogrid.h"
#include "shapet.h"
#include "util.h"

#ifdef ALGORITHMIC_STYLES
/*
 *
 *
 * (1) x2 + dx2 * t2 = x1 + dx1 * t1
 * (2) y2 + dy2 * t2 = y1 + dy1 * t1
 *
 *  1  =>  t1 = ( x2 - x1 + dx2 * t2 ) / dx1
 *  +2 =>  y2 + dy2 * t2 = y1 + dy1/dx1 * [ x2 - x1 + dx2 * t2 ]
 *
 *     => t2 * [dy1/dx1 * dx2 - dy2] = y2 - y1 - dy1/dx1*(x2-x1)
 *     => t2(dy1*dx2 - dy2*dx1) = dx1(y2 - y1) + dy1(x1-x2)
 *     => t2 = [dx1(y2-y1) + dy1(x1-x2)] / [dy1*dx2 - dy2*dx1]
 *     => t2 = [dx1(y2-y1) - dy1(x2-x1)] / [dx2*dy1 - dy2*dx1]
 *     t2 = Num/Denom
 *     =>
 *	    Num   = (y2 - y1) * dx1 - (x2 - x1) * dy1;
 *		Denom = dx2 * dy1 - dy2 * dx1;
 *
 */
static void ComputeIntersection( short line1_pt1_x, short line1_pt1_y, short line1_pt2_x, short line1_pt2_y,
								 short line2_pt1_x, short line2_pt1_y, short line2_pt2_x, short line2_pt2_y,
								 short *x, short *y )
{
 	long dx1, dy1, dx2, dy2, x1, y1, x2, y2;
	long Num,Denom, t;
	
	
	dx1 = line1_pt2_x - line1_pt1_x;	dy1 = line1_pt2_y - line1_pt1_y;
	dx2 = line2_pt2_x - line2_pt1_x;	dy2 = line2_pt2_y - line2_pt1_y;
 
 	x1 = line1_pt1_x; y1 = line1_pt1_y;
 	x2 = line2_pt1_x; y2 = line2_pt1_y;

	Num   = (y2 - y1) * dx1 - (x2 - x1) * dy1;
	Denom = dx2 * dy1 - dy2 * dx1;

	if ( Denom != 0 ) { /* 3/19/98 changed 0.0 to 0 ---Sampo */
		t = util_FixDiv( Num, Denom );
		*x = (short)( x2 + util_FixMul( dx2, t ) );
		*y = (short)( y2 + util_FixMul( dy2, t ) );
	} else {
		*x = (short)((line1_pt2_x+line2_pt1_x)/2);
		*y = (short)((line1_pt2_y+line2_pt1_y)/2);
	}
}

void tsi_SHAPET_BOLD_METRICS( register hmtxClass *hmtx, tsiMemObject *mem, short UPEM, F16Dot16 params[] )
{
	register int32 i, limit;
	short moveX, delta;
	F16Dot16 multiplier = params[0];
	
	UNUSED(mem);
	
	
	moveX = (short)(UPEM/12);
	moveX = (short)(util_FixMul( multiplier, moveX ) - moveX);
	delta = (short)(moveX + moveX);
	limit = hmtx->numGlyphs;

	for ( i = 0; i < limit; i++ ) {
		hmtx->aw[i] = (uint16)(hmtx->aw[i] + delta);
	}
}


/*
 * multiplier == 1.0 means do nothing
 */
void tsi_SHAPET_BOLD_GLYPH( GlyphClass *glyph, tsiMemObject *mem, short UPEM, F16Dot16 params[] )
{
	register long ctr, A, B, C, n;
	short xA, xB, xC, yA, yB, yC, orig_xB, orig_yB;
	short line1_pt1_x, line1_pt1_y;
	short line1_pt2_x, line1_pt2_y;
	short line2_pt1_x, line2_pt1_y;
	short line2_pt2_x, line2_pt2_y;
	short rot1Dx, rot1Dy;
	short rot2Dx, rot2Dy;
	short *x, *y;
	
	short clockWiseWinding = 1;
	short maxMove, moveX = 20;
	
	F16Dot16 multiplier = params[0];
	
	moveX = (short)(UPEM/12);
	moveX = (short)(util_FixMul( multiplier, moveX ) - moveX);
	maxMove = (short)(moveX * 2);
	if ( maxMove < 0 ) maxMove = (short)-maxMove;
	
	if ( glyph->contourCount > 0 && moveX != 0 ) {
		n = glyph->ep[glyph->contourCount-1] + 1;
		/* Save the original coordinates. */
		x = (short *)tsi_AllocMem( mem, (n+n) * sizeof(short) );
		y = &x[n];
		for ( A = 0; A < n; A++ ) {
			x[A] = glyph->oox[A];
			y[A] = glyph->ooy[A];
		}
		
		for ( ctr = 0; ctr < glyph->contourCount; ctr++ ) {
			n = glyph->ep[ctr];
			A = n;
			xB = glyph->oox[A]; yB = glyph->ooy[A];
			B = glyph->sp[ctr];
			orig_xB = glyph->oox[B]; orig_yB = glyph->ooy[B];
			xC = orig_xB; yC = orig_yB;
			for ( ; B <= n; B++ ) {
				xA = xB; yA = yB;
				xB = xC; yB = yC;
			
				C = B + 1; 
				if ( C > n ) {
				 	/* C = glyph->sp[ctr]; */
					xC = orig_xB; yC = orig_yB; /* can't read oox[B] since it has been reset */
				} else {
					xC = glyph->oox[C]; yC = glyph->ooy[C];
				}
				
				line1_pt1_x = xA;	line1_pt1_y = yA;
				line1_pt2_x = xB;	line1_pt2_y = yB;

				line2_pt1_x = xB;	line2_pt1_y = yB;
				line2_pt2_x = xC;	line2_pt2_y = yC;
				
				/* rotate anti-clockwise 90 degrees */
				rot1Dx = (short)-(line1_pt2_y - line1_pt1_y);	/* -dy */
				rot1Dy = (short)(line1_pt2_x - line1_pt1_x);		/* +dx */
				
				rot2Dx = (short)-(line2_pt2_y - line2_pt1_y);	/* -dy */
				rot2Dy = (short)(line2_pt2_x - line2_pt1_x);		/* +dx */

#ifdef SOMEDAY
				if ( !clockWiseWinding ) {
					rot1Dx = (short)-rot1Dx;
					rot1Dy = (short)-rot1Dy;
					rot2Dx = (short)-rot2Dx;
					rot2Dy = (short)-rot2Dy;
				}

#endif
				if ( rot1Dx > 0 ) {
					line1_pt1_x = (short)(line1_pt1_x + moveX);
					line1_pt2_x = (short)(line1_pt2_x + moveX);
				} else if ( rot1Dx < 0 ) {
					line1_pt1_x = (short)(line1_pt1_x - moveX);
					line1_pt2_x = (short)(line1_pt2_x - moveX);
				}

				if ( rot2Dx > 0 ) {
					line2_pt1_x = (short)(line2_pt1_x + moveX);
					line2_pt2_x = (short)(line2_pt2_x + moveX);
				} else if ( rot2Dx < 0 ) {
					line2_pt1_x = (short)(line2_pt1_x - moveX);
					line2_pt2_x = (short)(line2_pt2_x - moveX);
				}
				
				if ( line1_pt2_x == line2_pt1_x ) {
					glyph->oox[B] = line1_pt2_x;
				} else {
					/* compute the intersection of line1 and line2 */
					ComputeIntersection( line1_pt1_x, line1_pt1_y, line1_pt2_x, line1_pt2_y,
								 		 line2_pt1_x, line2_pt1_y, line2_pt2_x, line2_pt2_y, &glyph->oox[B], &glyph->ooy[B] );
					{
						long dx, dy, abx, aby, dist;
						
						dx = glyph->oox[B] - xB;
						dy = glyph->ooy[B] - yB;
						
						abx = dx > 0 ? dx : -dx;
						aby = dy > 0 ? dy : -dy;
						dist = abx > aby ? abx + (aby>>1) : aby + (abx>>1);
						if ( dist > maxMove ) {
							dx = dx * maxMove / dist;
							dy = dy * maxMove / dist;
							glyph->oox[B] = (short)(xB + dx);
							glyph->ooy[B] = (short)(yB + dy);
						}
						
					}
				}
				A = B;
			}
		}
		/* Repair the outlines */
		for ( ctr = 0; ctr < glyph->contourCount; ctr++ ) {
			n = glyph->ep[ctr];
			A = n;
			for ( B = glyph->sp[ctr]; B <= n; B++ ) {
				if ( (x[B] - x[A]) * (glyph->oox[B] - glyph->oox[A])  < 0 ) {
					/* sign flip ! */
					glyph->oox[B] = glyph->oox[A];
				}
				if ( (y[B] - y[A]) * (glyph->ooy[B] - glyph->ooy[A])  < 0 ) {
					/* sign flip ! */
					glyph->ooy[B] = glyph->ooy[A];
				}
				A = B;
			}
		}
		for ( ctr = 0; ctr < glyph->contourCount; ctr++ ) {
			long sum, count;
			n = glyph->ep[ctr];
			A = n;
			sum = glyph->oox[A]; count = 1;
			for ( B = glyph->sp[ctr]; B <= n; B++ ) {
				if ( x[B] == x[A] ) {
					sum += glyph->oox[B]; count++;
					continue; /*****/
				}
				if ( count > 1 ) {
					sum /= count;
					glyph->oox[A] 	= (short)sum;
					glyph->oox[--B] = (short)sum;
					while( count-- > 2 ) {
						glyph->oox[--B] = (short)sum;
					}
				}
				A = B;
				sum = glyph->oox[A]; count = 1;
			}
		}

		tsi_DeAllocMem( mem,  x );
	}
}

#endif /* ALGORITHMIC_STYLES */


#ifdef SOMEDAY

SHAPETClass *tsi_NewSHAPETClass( tsiMemObject *mem )
{
	register SHAPETClass *t = (SHAPETClass *)tsi_AllocMem( mem, sizeof( SHAPETClass ) );
	t->mem = mem;
	return t; /*****/
}


void tsi_DeleteSHAPETClass( SHAPETClass *t )
{
	if ( t != NULL ) {
		/* tsi_DeAllocMem( t->mem, t->baseAddr ); */
		tsi_DeAllocMem( t->mem, (char *)t );
	}
}
#endif
