/*
 * @(#)t2k.c	1.8 98/09/08
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * T2K.C
 * Copyright (C) 1989-1998 all rights reserved by Type Solutions, Inc. Plaistow, NH, USA.
 * http://www.typesolutions.com/
 * Author: Sampo Kaasila
 *
 * This software is the property of Type Solutions, Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
#include "syshead.h"

#include "t2k.h"
#include "util.h"
#include "autogrid.h"
#include "t2ksc.h"
#ifdef ENABLE_T2KE
#include "t2kclrsc.h"
#endif
#include "ghints.h"


#define T2K_MAGIC1 0x5a1234a5
#define T2K_MAGIC2 0xa5fedc5a

#ifdef ENABLE_AUTO_GRIDDING
/*
 *
 */
T2K *NewT2K( tsiMemObject *mem, sfntClass *font, int *errCode  )
{
	assert( errCode != NULL );
	if ( mem == NULL ) {
		*errCode = T2K_ERR_MEM_IS_NULL;
	} else if ( (*errCode = setjmp(mem->env)) == 0 ) {
		/* try */
		register T2K *t = (T2K *)tsi_AllocMem( mem, sizeof( T2K ) );
		t->mem = mem;
		
		t->stamp1 = T2K_MAGIC1;
		t->font	  = font;
		t->stamp2 = T2K_MAGIC2;
		
		
		t->glyph = NULL;
		t->hintHandle = NULL;
		t->baseAddr = NULL;
#ifdef ENABLE_T2KE
		t->baseARGB = NULL;
#endif /* ENABLE_T2KE */
		
		t->ag_xPixelsPerEm = t->ag_yPixelsPerEm = -1;
		/* t->globalHintsCache = NULL; */
		
		
#ifdef LAYOUT_CACHE_SIZE
		{
			int i;
			
			for ( i = 0; i < LAYOUT_CACHE_SIZE; i++ ) {
				t->tag[i] = 0xffffffff;
			}
		}
#endif
		t->font->preferedPlatformID 			= 0xffff;
		t->font->preferedPlatformSpecificID 	= 0xffff;
		return t; /*****/
	} else {
		/* catch */
		tsi_EmergencyShutDown( mem );
	}

	return NULL; /*****/
}

static void T2KDoGriddingSetUp( T2K *t )
{
	int err;
	
	if ( t->hintHandle == NULL ) {
		int maxPointCount;
		short unitsPerEm;
		/* ag_FontCategory fontCat; */

/* longjmp( t->mem->env, 9999 ); */

		maxPointCount = GetMaxPoints( t->font );
		unitsPerEm	  = GetUPEM( t->font );
	
		err = ag_HintInit( t->mem, maxPointCount, unitsPerEm, &t->hintHandle );
		tsi_Assert( t->mem, err == 0, err );
		t->fontCategory = GetNumGlyphs_sfntClass( t->font ) < 1000 ? ag_ROMAN : ag_KANJI; /* ag_ROMAN/ag_KANJI guess */
		if ( t->font->globalHintsCache == NULL ) {
			InputStream *in = NULL;
			ag_GlobalDataType globalHints;
			
			#ifdef ENABLE_T1
			if ( t->font->T1 != NULL ) {
				; /* Always do recomputation for T1 */
			} else 
			#endif
			#ifdef ENABLE_CFF
			if ( t->font->T2 != NULL ) {
				; /* Always do recomputation for T2 */
			} else 
			#endif
			/* Force recomputation if algorithmic styling is on */
			if ( t->font->StyleFunc == NULL ) {
				in = GetStreamForTable( t->font, tag_T2KG );
			}
			if ( in != NULL ) {
				ReadGHints( &globalHints, in);
				Delete_InputStream( in, NULL );
#ifdef OLD_TEST
				if ( true ) {
					ag_GlobalDataType globalHints2;
					int i;
					
					ComputeGlobalHints( t->font, t->hintHandle, &globalHints2 );
					for ( i = 0; i < ag_MAX_HEIGHTS_IN; i++ ) {
						assert( globalHints.heights[i].flat == globalHints2.heights[i].flat );
						assert( globalHints.heights[i].round == globalHints2.heights[i].round );
						assert( globalHints.heights[i].overLap == globalHints2.heights[i].overLap );
					}
					for ( i = 0; i < ag_MAXWEIGHTS; i++ ) {
						assert( globalHints.xWeight[i] == globalHints2.xWeight[i] );
						assert( globalHints.yWeight[i] == globalHints2.yWeight[i] );
					}
					printf("OK\n");
				}
#endif /* OLD_TEST */
			} else {
				ComputeGlobalHints( t->font, t->hintHandle, &globalHints, t->fontCategory == ag_KANJI  );
			}
			t->font->globalHintsCache = tsi_AllocMem( t->mem, sizeof(ag_GlobalDataType) );
			memcpy(t->font->globalHintsCache, &globalHints, sizeof(ag_GlobalDataType));
		}
		err = ag_SetHintInfo( t->hintHandle, (ag_GlobalDataType *)t->font->globalHintsCache, t->fontCategory ); 
		tsi_Assert( t->mem, err == 0, err );
	}
	if ( ((t->ag_xPixelsPerEm != t->xPixelsPerEm) || (t->ag_yPixelsPerEm != t->yPixelsPerEm))  ) {
		err = ag_SetScale( t->hintHandle, t->xPixelsPerEm, t->yPixelsPerEm, &t->xWeightIsOne );
		t->ag_xPixelsPerEm = t->xPixelsPerEm;
		t->ag_yPixelsPerEm = t->yPixelsPerEm;
		tsi_Assert( t->mem, err == 0, err );
	}	
}

/*
 *
 */
static void T2K_NewTransformationInternal( T2K *t, int doSetUpNow, long xPixelsPerEm, long yPixelsPerEm, T2K_TRANS_MATRIX *trans )
{
	t->t00 = trans->t00;
	t->t01 = trans->t01;
	t->t10 = trans->t10;
	t->t11 = trans->t11;
	t->is_Identity = 	t->t00 == ONE16Dot16	&& t->t01 == 0 &&
		 				t->t10 == 0				&& t->t11 == ONE16Dot16;
	t->xPixelsPerEm = xPixelsPerEm;
	t->yPixelsPerEm = yPixelsPerEm;
	if (doSetUpNow) {
		T2KDoGriddingSetUp(t);
	}
}

void T2K_TransformXFunits( T2K *t, short xValueInFUnits, F16Dot16 *x, F16Dot16 *y )
{
	F16Dot16 x16Dot16, y16Dot16;
	F16Dot16 tmpX;

	x16Dot16 = xValueInFUnits; x16Dot16 <<= 16;
	y16Dot16 = 0;
	
	x16Dot16 = util_FixMul( x16Dot16, t->xMul );
	if ( !t->is_Identity ) {
		tmpX = x16Dot16; /* tmpY = 0; */
		x16Dot16 = util_FixMul( t->t00, tmpX ) /* + util_FixMul( t->t01, tmpY ) */;
		y16Dot16 = util_FixMul( t->t10, tmpX ) /* + util_FixMul( t->t11, tmpY ) */;
	}
	*x = x16Dot16;
	*y = y16Dot16;
}

void T2K_TransformYFunits( T2K *t, short yValueInFUnits, F16Dot16 *x, F16Dot16 *y )
{
	F16Dot16 x16Dot16, y16Dot16;
	F16Dot16 tmpY;

	x16Dot16 = 0;
	y16Dot16 = yValueInFUnits; y16Dot16 <<= 16;
	
	y16Dot16 = util_FixMul( y16Dot16, t->yMul );
	if ( !t->is_Identity ) {
		tmpY = y16Dot16; /* tmpX = 0; */
		x16Dot16 = /* util_FixMul( t->t00, tmpX ) */ + util_FixMul( t->t01, tmpY );
		y16Dot16 = /* util_FixMul( t->t10, tmpX ) */ + util_FixMul( t->t11, tmpY );
	}
	*x = x16Dot16;
	*y = y16Dot16;
}

/*
 *
 */
void T2K_NewTransformation( T2K *t, int doSetUpNow, long xRes, long yRes, T2K_TRANS_MATRIX *trans, int *errCode )
{
	F16Dot16 xPointSize, yPointSize;
	F16Dot16 xResRatio, yResRatio;
	long xPixelsPerEm, yPixelsPerEm;
	register short UPEM;
	
	assert( errCode != NULL );
	if ( (*errCode = setjmp(t->mem->env)) == 0 ) {
		/* try */
		tsi_Assert( t->mem, trans != NULL, T2K_ERR_TRANS_IS_NULL );
		tsi_Assert( t->mem, xRes > 0 && yRes > 0, T2K_ERR_RES_IS_NOT_POS );
		UPEM = GetUPEM( t->font );
		
		/*
		xPointSize = trans->t00 >= 0 ? trans->t00 : -trans->t00;
		yPointSize = trans->t11 >= 0 ? trans->t11 : -trans->t11;
		*/
		xPointSize = util_EuclidianDistance( trans->t00, trans->t10 );
		yPointSize = util_EuclidianDistance( trans->t01, trans->t11 );
		
		xResRatio = (xRes << 16) / 72;
		yResRatio = (yRes << 16) / 72;
		
		xPixelsPerEm = util_FixMul( xPointSize, xResRatio );
		yPixelsPerEm = util_FixMul( yPointSize, yResRatio );
		t->xPixelsPerEm16Dot16 = xPixelsPerEm;
		t->yPixelsPerEm16Dot16 = yPixelsPerEm;
		t->xMul = util_FixDiv( t->xPixelsPerEm16Dot16, ((long)(UPEM)) << 16);
		t->yMul = util_FixDiv( t->yPixelsPerEm16Dot16, ((long)(UPEM)) << 16);
		/* Round to integer ppem */
		xPixelsPerEm += 0x00008000; xPixelsPerEm >>= 16;
		yPixelsPerEm += 0x00008000; yPixelsPerEm >>= 16;

		/* Now remove the point size from the matrix */
		/*
		trans->t00 = trans->t00 >= 0 ? (ONE16Dot16) : -(ONE16Dot16);
		trans->t11 = trans->t11 >= 0 ? (ONE16Dot16) : -(ONE16Dot16);
		*/
		if ( xPixelsPerEm > 0 && yPixelsPerEm > 0 ) {
			trans->t00 = util_FixDiv( trans->t00, xPointSize );
			trans->t10 = util_FixDiv( trans->t10, xPointSize );
			trans->t11 = util_FixDiv( trans->t11, yPointSize );
			trans->t01 = util_FixDiv( trans->t01, yPointSize );
		} else {
			trans->t00 = 0;
			trans->t10 = 0;
			trans->t11 = 0;
			trans->t01 = 0;
		}

	/*
		xPtr[i] = util_FixMul( t00, tmpX ) + util_FixMul( t01, tmpY );
		yPtr[i] = util_FixMul( t10, tmpX ) + util_FixMul( t11, tmpY );
	*/	

		T2K_NewTransformationInternal( t, doSetUpNow, xPixelsPerEm, yPixelsPerEm, trans );
		assert( t != NULL && t->font != NULL );
		{
			int16 Ascender, Descender, LineGap;
			uint16 maxAW;
			
			GetLineSpacing( t->font, &Ascender, &Descender, &LineGap, &maxAW, &t->caretDx, &t->caretDy );
			
			if ( !t->is_Identity ) {
				F16Dot16 tmpX = t->caretDx;
				F16Dot16 tmpY = t->caretDy;
				t->caretDx = util_FixMul( t->t00, tmpX ) + util_FixMul( t->t01, tmpY );
				t->caretDy = util_FixMul( t->t10, tmpX ) + util_FixMul( t->t11, tmpY );
			}
			T2K_TransformYFunits( t, Ascender,	&t->xAscender,	&t->yAscender);
			T2K_TransformYFunits( t, Descender,	&t->xDescender,	&t->yDescender);
			T2K_TransformYFunits( t, LineGap, 	&t->xLineGap,	&t->yLineGap);
			T2K_TransformXFunits( t, (int16)maxAW, 	&t->xMaxLinearAdvanceWidth,	&t->yMaxLinearAdvanceWidth );
			t->numGlyphs = GetNumGlyphs_sfntClass( t->font );
			/*
			T2K_TransformXFunits( t, UPEM,	&t->em0Dx,	&t->em0Dy);
			T2K_TransformYFunits( t, UPEM,	&t->em1Dx,	&t->em1Dy);
			*/

		}
	} else {
		/* catch */
		tsi_EmergencyShutDown( t->mem );
	}
}

	
static F26Dot6 scaleValue( T2K *t, F26Dot6 value, long ppem )
{
	register short UPEM = GetUPEM( t->font );
	
	value *= ppem*64;
	value += UPEM>>1;
	value /= UPEM;
	return value;
}

/* Not for external use */
static void T2K_PurgeMemoryInternal( register T2K *t, int level )
{
	int err;

	Delete_GlyphClass( t->glyph ); t->glyph = NULL;
	/* tsi_DeAllocMem( t->mem, t->x ); t->x = NULL; t->y = NULL; */
	
	if ( level > 0 ) {
		tsi_DeAllocMem( t->mem, t->baseAddr ); t->baseAddr = NULL;
		#ifdef ENABLE_T2KE
			tsi_DeAllocMem( t->mem, t->baseARGB ); t->baseARGB = NULL;
		#endif /* ENABLE_T2KE */

		if ( level > 1 ) {
			err = ag_HintEnd( t->hintHandle ); t->hintHandle = NULL;
			t->ag_xPixelsPerEm = t->ag_yPixelsPerEm = -1;
			tsi_Assert( t->mem, err == 0, err  );
		}
	}
}

void T2K_PurgeMemory( register T2K *t, int level, int *errCode )
{
	assert( errCode != NULL );
	if ( (*errCode = setjmp(t->mem->env)) == 0 ) {
		/* try */
		T2K_PurgeMemoryInternal( t, level );
	} else {
		/* catch */
		tsi_EmergencyShutDown( t->mem );
	}
}





/*
 *
 */
static void T2K_RenderGlyphInternal( T2K *t, long code, int8 xFracPenDelta, int8 yFracPenDelta, uint8 greyScaleLevel, uint8 cmd )
{
	F26Dot6 *xPtr, *yPtr;
	int err, pointCount;
	long i, n;
	register GlyphClass *glyph;
	int isFigure = false;
	short kern;
	uint16 aw;
	register short UPEM;
	
	tsi_Assert( t->mem, /* greyScaleLevel >= BLACK_AND_WHITE_BITMAP && */ greyScaleLevel <= GREY_SCALE_BITMAP_EXTREME_QUALITY, T2K_ERR_BAD_GRAY_CMD  );
	tsi_Assert( t->mem, xFracPenDelta >= 0 && xFracPenDelta < 64 && yFracPenDelta >= 0 && yFracPenDelta < 64, T2K_ERR_BAD_FRAC_PEN  );

	
	UPEM = GetUPEM( t->font );
	T2K_PurgeMemoryInternal( t, 1 );
	
	kern = 0;
	if ( cmd & T2K_CODE_IS_GINDEX ) {
		glyph = GetGlyphByIndex( t->font, code, false, &aw );
		isFigure = IsFigure( t->font, (unsigned short)code );
	} else {
		glyph = GetGlyphByCharCode( t->font, code, false, &aw );
		if ( code >= '0' && code <= '9' ) isFigure = true;
	}
	tsi_Assert( t->mem, glyph != NULL, T2K_ERR_GOT_NULL_GLYPH  );
	t->glyph = glyph;
	if ( glyph->contourCount < 0 ) {
		GlyphClass *base = NULL;
		uint8 comp_cmd;
		uint16 flags, oredFlags = 0;
		T2K_TRANS_MATRIX save, newbie;
		long save_XPPEM, save_YPPEM;
		int newT;
		short *componentData = glyph->componentData; /* Grab it! */
		glyph->componentData = NULL;
		/*
		Delete_GlyphClass( glyph );
		t->glyph = NULL;
		return;
		*/
		
		/* t->hmtx->lsb[t->rootGIndex], t->hmtx->aw[t->rootGIndex] */
		/* base = New_EmptyGlyph( t->mem, t->font->hmtx->lsb[t->font->rootGIndex], t->font->hmtx->aw[t->font->rootGIndex] ); */
		
		save.t00 = t->t00;
		save.t01 = t->t01;
		save.t10 = t->t10;
		save.t11 = t->t11;
		save_XPPEM = t->xPixelsPerEm;
		save_YPPEM = t->yPixelsPerEm;
		comp_cmd = (unsigned char)(cmd & ( T2K_GRID_FIT | T2K_USE_FRAC_PEN));
		comp_cmd |= T2K_RETURN_OUTLINES | T2K_CODE_IS_GINDEX;
/* #define KOREA_EXPERIMENT */
#define KOREA_EXPERIMENT
#ifdef KOREA_EXPERIMENT
	if ( t->fontCategory == ag_KANJI ) {
		comp_cmd &= ~T2K_GRID_FIT;
	}
#endif
		i = 0;
		do {
			int gIndex;
			long arg1, arg2;
			
			flags  = componentData[i++]; oredFlags |= flags;
			gIndex = componentData[i++];
 			if ( (flags & ARG_1_AND_2_ARE_WORDS) != 0 ) {
				arg1 = componentData[i++];
				arg2 = componentData[i++];
 			} else {
				arg1 = componentData[i++];
 				arg2 = arg1 & 0xff;
 				arg1 >>= 8;
 			}

 			newT = false;
 			if ( flags & ARGS_ARE_XY_VALUES ) {
 				arg1 = scaleValue( t, arg1, t->xPixelsPerEm );
 				arg2 = scaleValue( t, arg2, t->yPixelsPerEm );
 			}
 			if ( (flags & WE_HAVE_A_SCALE) != 0 ) {
				newbie.t00 = componentData[i] << 2;
				newbie.t01 = 0;
				newbie.t10 = 0;
				newbie.t11 = componentData[i] << 2;
				i++;
				newT = true;
 			} else if ( (flags & WE_HAVE_AN_X_AND_Y_SCALE) != 0 ) {
				newbie.t00 = componentData[i++] << 2;
				newbie.t01 = 0;
				newbie.t10 = 0;
				newbie.t11 = componentData[i++] << 2;
				newT = true;
 			} else if ( (flags & WE_HAVE_A_TWO_BY_TWO) != 0 ) {
				newbie.t00 = componentData[i++] << 2;
				newbie.t01 = componentData[i++] << 2;
				newbie.t10 = componentData[i++] << 2;
				newbie.t11 = componentData[i++] << 2;
				newT = true;
 			}
 			if ( newT ) {
 				T2K_NewTransformationInternal( t, false, t->xPixelsPerEm, t->yPixelsPerEm, &newbie );
 			}
			/* T2K_RenderGlyph( t, gIndex, xFracPenDelta, yFracPenDelta, 0, comp_cmd, NULL ); */
			T2K_RenderGlyphInternal( t, gIndex, xFracPenDelta, yFracPenDelta, 0, comp_cmd );

				
 			if ( newT ) {
 				T2K_NewTransformationInternal( t, false, save_XPPEM, save_YPPEM, &save );
 			}
 			
			Add_GlyphClass( &base, t->glyph, flags, arg1, arg2 );
			if ( base != t->glyph ) {
				Delete_GlyphClass( t->glyph );
			}
			t->glyph = NULL;
		} while ( flags & MORE_COMPONENTS );
		t->glyph = glyph = base;
		tsi_DeAllocMem( t->mem, componentData );
		xPtr = glyph->x;
		yPtr = glyph->y;
		pointCount = glyph->pointCount; n = pointCount + 2;

		if ( !(oredFlags & USE_MY_METRICS) ) {
			glyph->oox[glyph->pointCount + 0] = 0;
			glyph->oox[glyph->pointCount + 1] = (short)(glyph->oox[t->glyph->pointCount + 0] + aw);
			glyph->x[glyph->pointCount + 1]   = scaleValue( t, glyph->oox[glyph->pointCount + 1], t->xPixelsPerEm );
			if (cmd & T2K_GRID_FIT) {
				glyph->x[glyph->pointCount + 1] += 32;
				glyph->x[glyph->pointCount + 1] &= ~63;
			}
			glyph->x[glyph->pointCount + 0]   = 0; /* == scaleValue(0) */
		}
		
		
#ifdef KOREA_EXPERIMENT
		if ( t->fontCategory == ag_KANJI && (cmd & T2K_GRID_FIT) ) {
			/* inverse-scale glyph */
			register long xMultiplier = t->xPixelsPerEm*64;
			register long yMultiplier = t->yPixelsPerEm*64;
			register F26Dot6 tmpX, tmpY;
			register F16Dot16 xMul = util_FixDiv( UPEM, t->xPixelsPerEm*64 );
			register F16Dot16 yMul = util_FixDiv( UPEM, t->yPixelsPerEm*64 );
			
			pointCount = glyph->pointCount;
			n    = pointCount + 2;
			for ( i = 0; i < n; i++ ) {
				tmpX = xPtr[i];
				tmpY = yPtr[i];
				tmpX = util_FixMul( tmpX, xMul );
				tmpY = util_FixMul( tmpY, yMul );
#ifdef OLD
				tmpX *= UPEM;
				tmpY *= UPEM;
				tmpX /= xMultiplier;
				tmpY /= yMultiplier;
#endif
				glyph->oox[i] = (short)tmpX;
				glyph->ooy[i] = (short)tmpY;
			}
			{
				ag_ElementType elem;
				short curveType = 2;
				
				T2KDoGriddingSetUp( t );
				elem.contourCount   = glyph->contourCount;
				tsi_Assert( t->mem, pointCount <= 32000, T2K_ERR_TOO_MANY_POINTS  );
				
				elem.pointCount 	= (short)pointCount;
				elem.sp 			= glyph->sp;
				elem.ep 			= glyph->ep;
				elem.oox 			= glyph->oox;
				elem.ooy 			= glyph->ooy;
				elem.onCurve 		= glyph->onCurve;
				elem.x 				= xPtr;
				elem.y 				= yPtr;
				err = ag_AutoGridOutline( t->hintHandle, &elem, (short)isFigure, curveType, (short)(greyScaleLevel > BLACK_AND_WHITE_BITMAP) );
				tsi_Assert( t->mem, err == 0, err  );
			}
		}
#endif
	} else {
		pointCount = glyph->pointCount;
		n    = pointCount + 2;
		xPtr = (long*) tsi_AllocMem( t->mem, (n) * 2 * sizeof( long ) );
		yPtr = &xPtr[ n ];
		glyph->x = xPtr;
		glyph->y = yPtr;
		
		
#ifdef ENABLE_T2KE
		if ( t->font != NULL && t->font->T2KE != NULL && t->font->T2KE->properties[T2KE_PROP_HINTS_DISABLED] != 0 ) {
			cmd &= ~T2K_GRID_FIT;
		}
#endif		
		if ( (cmd & T2K_GRID_FIT) && !(cmd & T2K_USE_FRAC_PEN) && pointCount > 0 ) {
			ag_ElementType elem;
			short curveType = 2;
			
			T2KDoGriddingSetUp( t );
			elem.contourCount   = glyph->contourCount;
			tsi_Assert( t->mem, pointCount <= 32000, T2K_ERR_TOO_MANY_POINTS  );

			elem.pointCount 	= (short)pointCount;
			elem.sp 			= glyph->sp;
			elem.ep 			= glyph->ep;
			elem.oox 			= glyph->oox;
			elem.ooy 			= glyph->ooy;
			elem.onCurve 		= glyph->onCurve;
			elem.x 				= xPtr;
			elem.y 				= yPtr;
			err = ag_AutoGridOutline( t->hintHandle, &elem, (short)isFigure, curveType, (short)(greyScaleLevel > BLACK_AND_WHITE_BITMAP) );
			
			
			tsi_Assert( t->mem, err == 0, err  );
			/*
			t->xAdvanceWidth26Dot6	= elem.advanceWidth26Dot6;
			t->xAdvanceWidthInt		= elem.advanceWidthInt;
			*/
		} else  {
			/* scale glyph */
			register long xMultiplier = t->xPixelsPerEm*64;
			register long yMultiplier = t->yPixelsPerEm*64;
			register F26Dot6 tmpX, tmpY;
			
			for ( i = 0; i < n; i++ ) {
				tmpX = glyph->oox[i];
				tmpY = glyph->ooy[i];
				tmpX *= xMultiplier;
				tmpY *= yMultiplier;
				tmpX += UPEM>>1;
				tmpY += UPEM>>1;
				tmpX /= UPEM;
				tmpY /= UPEM;
				xPtr[i] = tmpX + xFracPenDelta;
				yPtr[i] = tmpY - yFracPenDelta;
			}
			/* Undo for the side bearing points */ 
			xPtr[--i] -= xFracPenDelta;
			yPtr[  i] += yFracPenDelta;
			xPtr[--i] -= xFracPenDelta;
			yPtr[  i] += yFracPenDelta;
		}
		if ( !(cmd & T2K_USE_FRAC_PEN) ) {
			/* To ensure integer spacing for no gridfitting, and something like the space character with zero points */
			F26Dot6 A,W;
			
			A = xPtr[pointCount];
			W = xPtr[pointCount+1] - A;
			A +=  32; A &= ~63; 		/* integralize */
			W +=  32; W &= ~63;
			xPtr[pointCount] 	= A; 	/* write out */
			xPtr[pointCount+1] 	= A+W; 	/* write out */
		}
	}
	t->xLinearAdvanceWidth16Dot16 = aw; t->xLinearAdvanceWidth16Dot16 <<= 16;
	t->yLinearAdvanceWidth16Dot16 = 0;
	
	#ifdef ENABLE_T1
	if ( t->font->T1 != NULL && t->font->T1->m01 != 0 ) {
		F16Dot16 skew = t->font->T1->m01;
		for ( i = 0; i < n; i++ ) {
			xPtr[i] += util_FixMul( skew, yPtr[i] );
		}
	} 
	#endif
	#ifdef ENABLE_CFF
	if ( t->font->T2 != NULL && t->font->T2->topDictData.m01 != 0 ) {
		F16Dot16 skew = t->font->T2->topDictData.m01;
		for ( i = 0; i < n; i++ ) {
			xPtr[i] += util_FixMul( skew, yPtr[i] );
		}
	} 
	#endif
	{
		/*
		F16Dot16 mul;
		
		mul = util_FixDiv( t->xPixelsPerEm16Dot16, ((long)(UPEM)) << 16);
		*/
		t->xLinearAdvanceWidth16Dot16	= util_FixMul( t->xLinearAdvanceWidth16Dot16, t->xMul );
		
		/* go to 16.16 for the side bearing points */
		xPtr[pointCount] <<= 10;
		yPtr[pointCount] <<= 10;
		xPtr[pointCount+1] <<= 10;
		yPtr[pointCount+1] <<= 10;
	}
	
	if ( !t->is_Identity ) {
		F16Dot16 t00 = t->t00;
		F16Dot16 t01 = t->t01;
		F16Dot16 t10 = t->t10;
		F16Dot16 t11 = t->t11;
		register F26Dot6 tmpX, tmpY;
		
		if ( t01 == 0 && t10 == 0 ) {
			for ( i = 0; i < n; i++ ) {
				tmpX = xPtr[i]; tmpY = yPtr[i];
				xPtr[i] = util_FixMul( t00, tmpX ) + 0;
				yPtr[i] = 0 + 						 util_FixMul( t11, tmpY );
			}
		} else {
			for ( i = 0; i < n; i++ ) {
				tmpX = xPtr[i]; tmpY = yPtr[i];
				xPtr[i] = util_FixMul( t00, tmpX ) + util_FixMul( t01, tmpY );
				yPtr[i] = util_FixMul( t10, tmpX ) + util_FixMul( t11, tmpY );
			}
		}
		tmpX = t->xLinearAdvanceWidth16Dot16; tmpY = 0;
		t->xLinearAdvanceWidth16Dot16 = util_FixMul( t00, tmpX ) /* + util_FixMul( t01, tmpY ) */;
		t->yLinearAdvanceWidth16Dot16 = util_FixMul( t10, tmpX ) /* + util_FixMul( t11, tmpY ) */;
	}
	t->xAdvanceWidth16Dot16	= xPtr[pointCount+1] - xPtr[pointCount];
	t->yAdvanceWidth16Dot16	= yPtr[pointCount+1] - yPtr[pointCount];
	
	/* Translate back to 26.6 from 16.16 */
	xPtr[pointCount] >>= 10;
	yPtr[pointCount] >>= 10;
	xPtr[pointCount+1] >>= 10;
	yPtr[pointCount+1] >>= 10;
	
	/* Initialize */
#ifdef ENABLE_T2KE
	t->baseARGB = NULL;
#endif /* ENABLE_T2KE */
	
	t->baseAddr = NULL;	t->rowBytes = 0;
	t->width = 0; 		t->height = 0;
	t->fTop26Dot6 = 0;	t->fLeft26Dot6 = 0;
	if ( cmd & T2K_SCAN_CONVERT && glyph->pointCount > 0 ) {
		tsiScanConv *sc = NULL;
#ifdef ENABLE_T2KE
		tsiColorScanConv *scc = NULL;
#endif

		char xWeightIsOne = (char)(t->xWeightIsOne && (cmd & T2K_GRID_FIT) && t->ag_xPixelsPerEm <= 24);
	
		if ( glyph->colorPlaneCount > 0 ) {
#ifdef ENABLE_T2KE
			assert( t->font->T2KE != NULL );
			scc = tsi_NewColorScanConv( t->mem, (void *)t->font->T2KE, (void*)t, glyph->contourCount, glyph->sp, glyph->ep,
								 	glyph->colors, glyph->colorPlaneCount,
				                  	xPtr, yPtr, (char *)glyph->onCurve, greyScaleLevel, glyph->oox, glyph->ooy, GetUPEM( t->font ) );
			MakeColorBits( scc, greyScaleLevel, xWeightIsOne, (char)(cmd & T2K_SKIP_SCAN_BM) );
			t->width  		= scc->right  - scc->left;
			t->height 		= scc->bottom - scc->top;
			t->fTop26Dot6	= scc->fTop26Dot6;
			t->fLeft26Dot6  = scc->fLeft26Dot6;
			t->rowBytes 	= scc->rowBytes;
#else
			assert( false );
#endif
		} else {
			sc = tsi_NewScanConv( t->mem, glyph->contourCount, glyph->sp, glyph->ep,
				                  xPtr, yPtr, (char *)glyph->onCurve, greyScaleLevel, (char)glyph->curveType );
			MakeBits( sc, greyScaleLevel, xWeightIsOne, (char)(cmd & T2K_SKIP_SCAN_BM), true );
			t->width  		= sc->right  - sc->left;
			t->height 		= sc->bottom - sc->top;
			t->fTop26Dot6	= sc->fTop26Dot6;
			t->fLeft26Dot6  = sc->fLeft26Dot6;
			t->rowBytes 	= sc->rowBytes;
		}


		
		if ( glyph->colorPlaneCount > 0 ) {
#ifdef ENABLE_T2KE
			t->baseARGB = scc->baseARGB; scc->baseARGB = NULL; /* We take over and deallocate t->baseARGB */
			t->baseAddr = NULL;
#else
			assert( false );
#endif
		} else {
			t->baseAddr = sc->baseAddr; sc->baseAddr = NULL; /* We take over and deallocate t->baseAddr */
			#ifdef ENABLE_T2KE
				t->baseARGB = NULL;
			#endif
		}
		
		/* Compensate if the lsb_point != 0 */
		if ( (i=xPtr[pointCount]) != 0 ) {
			t->fLeft26Dot6 -= i;
		}
		if ( (i=yPtr[pointCount]) != 0 ) {
			t->fTop26Dot6 -= i;
		}
		
		t->fLeft26Dot6 &= ~63;
		t->fTop26Dot6  &= ~63;
		
#ifdef ENABLE_T2KE
		tsi_DeleteColorScanConv( scc );
#endif
		tsi_DeleteScanConv( sc );
	}
	
	if ( cmd & T2K_RETURN_OUTLINES ) {
		;
	} else {
		T2K_PurgeMemoryInternal( t, 0 );
	}
}


int32 T2K_GetNumAxes(T2K *t)
{
	sfntClass *font;
	int32 numAxes = 0;
	
	font = t->font;
	assert( font != NULL );
	
#ifdef ENABLE_T1
	if ( font->T1 != NULL ) {
		numAxes = font->T1->numAxes;
	}
#endif
#ifdef ENABLE_CFF
	if ( font->T2 != NULL ) {
		numAxes = font->T2->topDictData.numAxes;
	}
#endif
#ifdef ENABLE_T2KE
	if ( font->T2KE != NULL ) {
		numAxes = font->T2KE->numAxes;
	}
#endif
	/* granularity;  location; */
	return numAxes; /*****/
}

F16Dot16 T2K_GetAxisGranularity(T2K *t, int32 n)
{
	F16Dot16 granularity = ONE16Dot16;
	sfntClass *font = t->font;

	assert( font != NULL );
	assert( n >= 0 && n < T2K_GetNumAxes( t ) );
	
#ifdef ENABLE_T2KE
	if ( font->T2KE != NULL ) {
		granularity = font->T2KE->granularity[n];
	}
#else
	UNUSED( n );
#endif
	return granularity; /*****/
}

F16Dot16 T2K_GetCoordinate(T2K *t, int32 n )
{
	F16Dot16 coordinate = 0;
	sfntClass *font = t->font;

	assert( font != NULL );
	assert( n >= 0 && n < T2K_GetNumAxes( t ) );
	
#ifdef ENABLE_T2KE
	if ( font->T2KE != NULL ) {
		coordinate = font->T2KE->currentCoordinate[n];
	}
#else
	UNUSED( n );
#endif
	return coordinate; /*****/
}

void T2K_SetCoordinate(T2K *t, int32 n, F16Dot16 value )
{
	sfntClass *font = t->font;

	assert( font != NULL );
	assert( n >= 0 && n < T2K_GetNumAxes( t ) );
	
#ifdef ENABLE_T2KE
	if ( font->T2KE != NULL ) {
		font->T2KE->currentCoordinate[n] = value;
	}
#else
	UNUSED( n );
	UNUSED( value );
#endif
}




/*
 *
 */
void T2K_RenderGlyph( T2K *t, long code, int8 xFracPenDelta, int8 yFracPenDelta, uint8 greyScaleLevel, uint8 cmd, int *errCode )
{
	assert( errCode != NULL );
	if ( (*errCode = setjmp(t->mem->env)) == 0 ) {
		/* try */
		tsi_Assert( t->mem, t->mem->state == T2K_STATE_ALIVE, T2K_ERR_USE_PAST_DEATH );
		T2K_RenderGlyphInternal( t, code, xFracPenDelta, yFracPenDelta, greyScaleLevel, cmd );
	} else {
		/* catch */
		tsi_EmergencyShutDown( t->mem );
	}
}


/*
 *
 */
void DeleteT2K( T2K *t, int *errCode )
{
	assert( errCode != NULL );
	if ( (*errCode = setjmp(t->mem->env)) == 0 ) {
		/* try */
		tsi_Assert( t->mem, t->stamp1 == (signed) T2K_MAGIC1 && t->stamp2 == (signed) T2K_MAGIC2, T2K_ERR_BAD_T2K_STAMP  );

		T2K_PurgeMemoryInternal( t, 2 );

		tsi_DeAllocMem( t->mem, t->font->globalHintsCache );
		tsi_DeAllocMem( t->mem, t );
	} else {
		/* catch */
		tsi_EmergencyShutDown( t->mem );
	}
}

#ifdef ENABLE_LINE_LAYOUT

uint16 T2K_GetGlyphIndex( T2K *t, uint16 charCode )
{
	return GetSfntClassGlyphIndex( t->font, charCode ); /*****/
}



uint32 T2K_MeasureTextInX(T2K *t, const uint16 *text, int16 *xKernValuesInFUnits, uint32 numChars )
{
	uint32 i, totalWidth, thisWidth;
	uint16 glyphIndex, prevGlyphIndex;
	uint16 charCode, prevCharCode;
	uint16 *awArr;
#ifdef LAYOUT_CACHE_SIZE
	uint32 cachePos, cacheTag;
#endif
#ifdef ENABLE_KERNING
	int16 xKern, yKern;
#endif
	
	assert( t != NULL );
	assert( t->font != NULL );
	assert( t->font->hmtx != NULL );
	assert( xKernValuesInFUnits != NULL );
	totalWidth		= 0;
	awArr			= t->font->hmtx->aw;
	glyphIndex 		= 0xffff;
	prevCharCode	= 32; /* space character */
	prevGlyphIndex	= 0xffff;
	for ( i = 0; i < numChars; i++ ) {
		charCode = text[i];
		
#ifdef LAYOUT_CACHE_SIZE
		cachePos   = prevCharCode;
		cachePos <<= 4;
		cachePos  ^= charCode;
		cachePos  %= LAYOUT_CACHE_SIZE;
		
		cacheTag   = prevCharCode;
		cacheTag <<= 16;
		cacheTag  |= charCode;
		if ( t->tag[cachePos] == cacheTag ) {
			thisWidth	= t->kernAndAdvanceWidth[ cachePos ];
#ifdef ENABLE_KERNING
			xKern		= t->kern[ cachePos ];
#endif
			glyphIndex	= 0xffff;
		} else {
#endif	/* LAYOUT_CACHE_SIZE */	
			glyphIndex	= GetSfntClassGlyphIndex( t->font, charCode );
			thisWidth	= awArr[ glyphIndex ];
			#ifdef ENABLE_KERNING
			{
				if ( prevGlyphIndex == 0xffff ) {
					prevGlyphIndex = GetSfntClassGlyphIndex( t->font, prevCharCode );
				}
				GetSfntClassKernValue( t->font, prevGlyphIndex, glyphIndex, &xKern, &yKern );
				thisWidth += xKern;
			}
			#endif /* ENABLE_KERNING */	
#ifdef LAYOUT_CACHE_SIZE
			/* cache the results */
			t->tag[cachePos]					= cacheTag;
			t->kernAndAdvanceWidth[ cachePos ]	= (short)thisWidth;
#ifdef ENABLE_KERNING
			t->kern[ cachePos ]					= xKern;
#endif
		}
#endif	/* LAYOUT_CACHE_SIZE */
#ifdef ENABLE_KERNING
		xKernValuesInFUnits[i]	= xKern; /* in fUnits */
#else
		xKernValuesInFUnits[i]	= 0; 
#endif
		totalWidth		+= thisWidth;
		prevGlyphIndex	 = glyphIndex;
		prevCharCode	 = charCode;
	}
	/* now scale to pixels */
	totalWidth = util_FixMul( totalWidth, t->xMul );
	return totalWidth; /*****/
}


#ifdef ENABLE_KERNING
void T2K_GetIdealLineWidth( T2K *t,         const T2KCharInfo cArr[], long lineWidth[], T2KLayout out[] )
#else
void T2K_GetIdealLineWidth( T2K *UNUSED(t), const T2KCharInfo cArr[], long lineWidth[], T2KLayout out[] )
#endif
{
	int i;
	const T2KCharInfo *cd;
	long totalIntWidthX, totalIntWidthY;
	F16Dot16 totSumX, totSumY;
	uint16 prevIndex = 0;

	totalIntWidthX = totalIntWidthY = 0;
	totSumX = totSumY = 0;
	for ( i = 0; (cd = &cArr[ i ])->charCode != 0 ; i++ ) {
		totSumX += cd->LinearAdvanceWidth16Dot16[ T2K_X_INDEX ];
		totSumY += cd->LinearAdvanceWidth16Dot16[ T2K_Y_INDEX ];
		/* Initialize to the non-linear metrics */
		out[i].BestAdvanceWidth16Dot16[ T2K_X_INDEX ] = cd->AdvanceWidth16Dot16[ T2K_X_INDEX ];
		out[i].BestAdvanceWidth16Dot16[ T2K_Y_INDEX ] = cd->AdvanceWidth16Dot16[ T2K_Y_INDEX ];
#ifdef ENABLE_KERNING
		if ( i != 0 ) {
			int16 xKern, yKern;
			F16Dot16 delta;
			GetSfntClassKernValue( t->font, prevIndex, cd->glyphIndex, &xKern, &yKern );
			if ( xKern != 0 ) {
				delta = xKern; delta <<= 16;
				delta = util_FixMul( delta, t->xMul );
				totSumX += delta;
				/* truncate */
				delta &= 0xffff0000;
				out[i].BestAdvanceWidth16Dot16[ T2K_X_INDEX ] += delta;
			}
			if ( yKern != 0 ) {
				delta = yKern; delta <<= 16;
				delta = util_FixMul( delta, t->yMul );
				totSumY += delta;
				/* truncate */
				delta &= 0xffff0000;
				out[i].BestAdvanceWidth16Dot16[ T2K_Y_INDEX ] += delta;
			}
		}
#endif /* ENABLE_KERNING */		
		totalIntWidthX += totSumX >> 16; totSumX &= 0x0000ffff;
		totalIntWidthY += totSumY >> 16; totSumY &= 0x0000ffff;
		prevIndex = cd->glyphIndex;
	}
	lineWidth[ T2K_X_INDEX ] = totalIntWidthX;
	lineWidth[ T2K_Y_INDEX ] = totalIntWidthY;
}

void T2K_LayoutString( const T2KCharInfo cArr[], const long LineWidthGoal[], T2KLayout out[] )
{
	int i, j, MY_INDEX;
	const T2KCharInfo *cd;
	long totalIntWidth;
	F16Dot16 fracSum, spaceAdvance = 0;
	long error, goal, delta, deltaI, spaceCount, strLen;

	
	
	/* Set to the dimension that we will tweak, the other we will just scale */
	if ( LineWidthGoal[ T2K_X_INDEX ] >= LineWidthGoal[ T2K_Y_INDEX ] ) {
		goal = LineWidthGoal[ T2K_X_INDEX ];
		MY_INDEX = T2K_X_INDEX;
	} else {
		goal = LineWidthGoal[ T2K_Y_INDEX ];
		MY_INDEX = T2K_Y_INDEX;
	}
	totalIntWidth = 0; fracSum = 0;
	for ( spaceCount = 0, strLen = 0, i = 0; (cd = &cArr[ i ])->charCode != 0; i++ ) {
		/*
		 already done
		out[i].BestAdvanceWidth16Dot16[ T2K_X_INDEX ] = cd->AdvanceWidth16Dot16[ T2K_X_INDEX ];
		out[i].BestAdvanceWidth16Dot16[ T2K_Y_INDEX ] = cd->AdvanceWidth16Dot16[ T2K_Y_INDEX ];
		*/
		fracSum += out[i].BestAdvanceWidth16Dot16[ MY_INDEX ];
		totalIntWidth += fracSum >> 16; fracSum &= 0x0000ffff;
		if ( cd->charCode == 32 ) {
			spaceCount++;
			spaceAdvance = cd->LinearAdvanceWidth16Dot16[ MY_INDEX ];
		}
		strLen++;
	}
	error = totalIntWidth - goal;
	if ( strLen == 0 ) return; /*****/
	
	
	if ( error > 0 ) {
		deltaI = -1;
		delta  = -ONE16Dot16;
	} else {
		deltaI = 1;
		delta  = ONE16Dot16;
	}
	if ( spaceCount > 0 ) {
		long tmp 		= spaceAdvance;
		long minSpace	= spaceAdvance/2 + 1;
		long maxSpace	= spaceAdvance*4;
		while ( error != 0 && tmp >= minSpace && tmp <= maxSpace) {
			for ( i = 0; i < strLen; i++ ) {
				cd = &cArr[ i ];
				if ( cd->charCode == 32 ) {
					tmp = out[i].BestAdvanceWidth16Dot16[ MY_INDEX ] + delta;
					if ( tmp < minSpace || tmp > maxSpace ) break; /*****/
					out[i].BestAdvanceWidth16Dot16[ MY_INDEX ] = tmp;
					error += deltaI;
					if ( error == 0 ) break; /*****/
				}
			}
		}
	}
	if ( error >= strLen || error <= -strLen ) {
		int mul = error/strLen;
		if ( mul < 0 ) mul = -mul;
		for ( i = 0; i < strLen; i++ ) {
			cd = &cArr[ i ];
			out[i].BestAdvanceWidth16Dot16[ MY_INDEX ] += delta*mul;
			error += deltaI*mul;
		}
	}
	/* assert( error < strLen && error > -strLen ); */

	if ( error != 0 ) {
		int absError = error > 0 ? error : -error;
		int inc = strLen / ( absError + 1 );
		
		
		for ( i = inc>>1; error != 0; i = i % strLen ) {
			cd = &cArr[ i ];
			if ( out[i].BestAdvanceWidth16Dot16[ MY_INDEX ] > 0 ) {
				out[i].BestAdvanceWidth16Dot16[ MY_INDEX ] += delta;
				error += deltaI;
				i += inc;
			} else {
				i++;
			}
		}
	}
	
	/* Scale the other dimension(s) */
	for ( j = 0; j < T2K_NUM_INDECES; j++ ) {
		if ( j == MY_INDEX ) continue; /*****/
		for ( i = 0; i < strLen; i++ ) {
			F16Dot16 scaleFactor;
			scaleFactor = cd->AdvanceWidth16Dot16[ j ] > 0 ? util_FixDiv( out[i].BestAdvanceWidth16Dot16[ MY_INDEX ], cd->AdvanceWidth16Dot16[ MY_INDEX ] ) : 0;
			out[i].BestAdvanceWidth16Dot16[ j ] = util_FixMul( out[i].BestAdvanceWidth16Dot16[ j ], scaleFactor );
		}
	}
}
#endif /* ENABLE_LINE_LAYOUT */




#endif /* ENABLE_AUTO_GRIDDING */
