/*
 * @(#)t2kstrm.c	1.7 98/07/21
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * T2KSTRM.C
 * Copyright (C) 1989-1998 all rights reserved by Type Solutions, Inc. Plaistow, NH, USA.
 * Author: Sampo Kaasila
 *
 * This software is the property of Type Solutions, Inc. and it is furnished
 * under a license and may be used and copied only in accordance with the
 * terms of such license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or otherwise
 * made available to any other person or entity except as allowed under license.
 * No title to and ownership of the software or intellectual property
 * therewithin is hereby transferred.
 *
 * This information in this software is subject to change without notice
 */
#include "syshead.h"

#include "dtypes.h"
#include "config.h"
#include "tsimem.h"
#include "t2kstrm.h"



int32 ReadInt32( InputStream *t )
{
	unsigned long pos = t->pos;
	register uint32 lword;
	register unsigned char *ptr;
#ifdef ENABLE_NON_RAM_STREAM
	unsigned char base[4];
	if ( t->privateBase == NULL ) {
		ptr = base;
		t->ReadToRamFunc( t->nonRamID, ptr, pos, 4 );
	} else {
		ptr = &t->privateBase[pos];
	}
#else
	ptr = &t->privateBase[pos];
#endif
	
	pos = pos + 4;
#if 0
// Parry as per Sampo's email
if ( pos > t->maxPos ) {
	pos = 0;
}
#endif
	assert( pos <= t->maxPos ); 
	t->pos = pos;
	
	lword = *ptr++;
	lword <<= 8;
	lword |= *ptr++;
	lword <<= 8;
	lword |= *ptr++;
	lword <<= 8;
	lword |= *ptr;
	
	return (int32)lword; /*****/
}

int16 ReadInt16( InputStream *t )
{
	unsigned long pos = t->pos;
	register uint16 word;
	register unsigned char *ptr;
#ifdef ENABLE_NON_RAM_STREAM
	unsigned char base[2];
	if ( t->privateBase == NULL ) {
		ptr = base;
		t->ReadToRamFunc( t->nonRamID, ptr, pos, 2 );
	} else {
		ptr = &t->privateBase[pos];
	}
#else
	ptr = &t->privateBase[pos];
#endif
	
	pos = pos + 2;
	assert( pos <= t->maxPos ); 
	t->pos = pos;
	
	word = *ptr++;
	word <<= 8;
	word |= *ptr;
	
	return (int16)word; /*****/
}

uint8 ReadUnsignedByte( InputStream *t )
{
	unsigned long pos = t->pos;
	uint8 byte;
	register unsigned char *ptr;
#ifdef ENABLE_NON_RAM_STREAM
	unsigned char base[1];
	if ( t->privateBase == NULL ) {
		ptr = base;
		t->ReadToRamFunc( t->nonRamID, ptr, pos, 1 );
	} else {
		ptr = &t->privateBase[pos];
	}
#else
	ptr = &t->privateBase[pos];
#endif
	
	assert( pos < t->maxPos ); 
	pos = pos + 1;
	byte = *ptr;
	t->pos = pos;
	return byte; /*****/
}


void Read( InputStream *t, uint8 *dest, long numBytes )
{
	if ( numBytes > 0 ) {
		unsigned long pos = t->pos;
		unsigned char *ptr;
		
#ifdef ENABLE_NON_RAM_STREAM
		if ( t->privateBase == NULL ) {
			t->ReadToRamFunc( t->nonRamID, dest, pos, numBytes );
		} else {
			ptr = &t->privateBase[pos];
			memcpy( dest, ptr, numBytes );
		}
#else
		ptr = &t->privateBase[pos];	
		memcpy( dest, ptr, numBytes );
#endif
		pos = pos + numBytes;
		if ( pos > t->maxPos ) {
			assert( pos <= t->maxPos );
		}
		t->pos = pos;
	}
}


long SizeInStream( InputStream *stream )
{
	return stream->maxPos - stream->posZero; /*****/
}

unsigned char *GetEntireStreamIntoMemory( InputStream *stream  )
{
	assert( stream->privateBase != NULL ); /* Only used for Type 1, does not work for non-RAM fonts */
	return &stream->privateBase[stream->posZero]; /*****/
}

/*
 *
 */
InputStream *New_InputStream( tsiMemObject *mem, unsigned char *data, unsigned long length, int *errCode )
{
	InputStream *t;
	
	if ( errCode == NULL || (*errCode = setjmp( mem->env)) == 0 ) {
		/* try */
		t  = (InputStream*) tsi_AllocMem( mem, sizeof( InputStream ) );
		
		t->mem				= mem;
		t->privateBase		= data;
#ifdef ENABLE_NON_RAM_STREAM
		t->ReadToRamFunc	= NULL;
		t->nonRamID			= 0;
#endif
		t->pos				= 0;
		t->posZero			= 0;
		t->maxPos			= length;
		t->bitBufferIn  	= 0;
		t->bitCountIn		= 0;
		t->constructorType 	= 1;
	} else {
		/* catch */
		t = NULL;
		tsi_EmergencyShutDown( mem );
	}
	
	return t; /*****/
}

/*
 *
 */
InputStream *New_InputStream2( tsiMemObject *mem, InputStream *in, unsigned long offset, unsigned long length, int *errCode )
{
	InputStream *t;
	
	if ( errCode == NULL || (*errCode = setjmp( mem->env)) == 0 ) {
		/* try */
		t = (InputStream*) tsi_AllocMem( mem, sizeof( InputStream ) );
		
		t->mem				= mem;
		t->privateBase		= in->privateBase;
#ifdef ENABLE_NON_RAM_STREAM
		t->ReadToRamFunc	= in->ReadToRamFunc;
		t->nonRamID			= in->nonRamID;
#endif
		t->pos				= offset;
		t->posZero			= offset;
		t->maxPos			= offset + length;
		t->bitBufferIn  	= 0;
		t->bitCountIn		= 0;
		t->constructorType 	= 2;
	} else {
		/* catch */
		t = NULL;
		tsi_EmergencyShutDown( mem );
	}
	
	return t; /*****/
}

InputStream *New_InputStream3( tsiMemObject *mem, unsigned char *data, unsigned long length, int *errCode )
{
	InputStream *t;
	
	if ( errCode == NULL || (*errCode = setjmp( mem->env)) == 0 ) {
		/* try */
		t = (InputStream*) tsi_AllocMem( mem, sizeof( InputStream ) );
		
		t->mem				= mem;
		t->privateBase		= data;
#ifdef ENABLE_NON_RAM_STREAM
		t->ReadToRamFunc	= NULL;
		t->nonRamID			= 0;
#endif
		t->pos				= 0;
		t->posZero			= 0;
		t->maxPos			= length;
		t->bitBufferIn  	= 0;
		t->bitCountIn		= 0;
		t->constructorType 	= 3;
	} else {
		/* catch */
		t = NULL;
		tsi_EmergencyShutDown( mem );
	}
	
	return t; /*****/
}

#ifdef ENABLE_NON_RAM_STREAM
InputStream *New_NonRamInputStream( tsiMemObject *mem, void *nonRamID, PF_READ_TO_RAM readFunc, unsigned long length, int *errCode )
{
	InputStream *t;
	
	if ( errCode == NULL || (*errCode = setjmp( mem->env)) == 0 ) {
		/* try */
		t  = (InputStream*) tsi_AllocMem( mem, sizeof( InputStream ) );
		
		t->mem				= mem;
		t->privateBase		= NULL;
		t->ReadToRamFunc	= readFunc;
		t->nonRamID			= nonRamID;
		t->pos				= 0;
		t->posZero			= 0;
		t->maxPos			= length;
		t->bitBufferIn  	= 0;
		t->bitCountIn		= 0;
		t->constructorType 	= 4;
	} else {
		/* catch */
		t = NULL;
		tsi_EmergencyShutDown( mem );
	}
	
	return t; /*****/
}
#endif /* ENABLE_NON_RAM_STREAM */


/*
 *
 */
void Rewind_InputStream( InputStream *t )
{
	t->pos = t->posZero;
}

/*
 *
 */
void Seek_InputStream( InputStream *t, uint32 offset )
{
	t->pos = t->posZero + offset;
}
/*
 *
 */
uint32 Tell_InputStream( InputStream *t )
{
	return t->pos - t->posZero; /*****/
}



/*
 *
 */
void Delete_InputStream( InputStream *t, int *errCode )
{
	if ( t != NULL ) {
		if ( errCode == NULL || (*errCode = setjmp( t->mem->env)) == 0 ) {
			/* try */
			if ( t->constructorType == 1 ) {
				tsi_DeAllocMem( t->mem, t->privateBase );
			}
			tsi_DeAllocMem( t->mem, t );
		} else {
			/* catch */
			tsi_EmergencyShutDown( t->mem );
		}
	}
}


#ifdef ENABLE_WRITE
void WriteInt32( OutputStream *stream, int32 value )
{
	register unsigned long pos = stream->pos;
	register unsigned char *ptr;
	register uint32 lword = (uint32)value;
	
	pos += 4;
	if ( pos > stream->maxLength ) {
		stream->maxLength = pos + (pos>>1);
		stream->base = (uint8 *)tsi_ReAllocMem( stream->mem, stream->base, stream->maxLength );
		assert( stream->base != NULL );
	}
	ptr = &stream->base[stream->pos];
	stream->pos = pos;
	
	*ptr++ = (unsigned char)(lword >> 24);
	*ptr++ = (unsigned char)(lword >> 16);
	*ptr++ = (unsigned char)(lword >> 8);
	*ptr   = (unsigned char)(lword);
}	


void WriteInt16( OutputStream *stream, int16 value )
{
	register unsigned long pos = stream->pos;
	register unsigned char *ptr;
	register uint16 word = (uint16)value;
	
	pos += 2;
	if ( pos > stream->maxLength ) {
		stream->maxLength = pos + (pos>>1);
		stream->base = (uint8 *)tsi_ReAllocMem( stream->mem, stream->base, stream->maxLength );
		assert( stream->base != NULL );
	}
	ptr = &stream->base[stream->pos];
	stream->pos = pos;
	
	*ptr++ = (unsigned char)(word >> 8);
	*ptr   = (unsigned char)word;
}	


void WriteUnsignedByte( OutputStream *stream, uint8 value )
{
	register unsigned long pos = stream->pos;
	register unsigned char *ptr;
	
	pos++;
	if ( pos > stream->maxLength ) {
		stream->maxLength = pos + (pos>>1);
		stream->base = (uint8 *)tsi_ReAllocMem( stream->mem, stream->base, stream->maxLength );
		assert( stream->base != NULL );
	}
	ptr = &stream->base[stream->pos];
	stream->pos = pos;
	
	*ptr   = value;
}

void Write( OutputStream *stream, uint8 *src, long numBytes )
{
	if ( numBytes > 0 ) {
		register unsigned long pos = stream->pos;
		register unsigned char *ptr;
		
		pos += numBytes;
		if ( pos > stream->maxLength ) {
			stream->maxLength = pos + (pos>>1);
			stream->base = (uint8 *)tsi_ReAllocMem( stream->mem, stream->base, stream->maxLength );
			assert( stream->base != NULL );
		}
		ptr = &stream->base[stream->pos];
		stream->pos = pos;
		memcpy( ptr, src, numBytes );
	}
}

long SizeOutStream( OutputStream *stream )
{
	if ( stream->pos > stream->maxPos ) {
		stream->maxPos = stream->pos;
	}
	return stream->maxPos; /*****/
}

long OutStreamPos( OutputStream *stream )
{
	if ( stream->pos > stream->maxPos ) {
		stream->maxPos = stream->pos;
	}
	return stream->pos; /*****/
}

/*
 *
 */
OutputStream *New_OutputStream( tsiMemObject *mem, long initialSize )
{
	OutputStream *t = (OutputStream*) tsi_AllocMem( mem, sizeof( OutputStream ) );
	
	t->mem			= mem;
	t->pos			= 0;
	t->maxPos		= 0;
	if ( initialSize <= 0 ) initialSize = 1024;
	t->maxLength	= initialSize;
	
	t->base 		= (uint8 *)tsi_AllocMem( t->mem, t->maxLength );
	assert( t->base != NULL );
	
	t->bitBufferOut  	= 0;
	t->bitCountOut		= 0;
	return t; /*****/
}

void Rewind_OutputStream( OutputStream *t )
{
	assert( t->pos <= t->maxLength );
	if ( t->pos > t->maxPos ) {
		t->maxPos = t->pos;
	}
	t->pos = 0;
}

/*
 *
 */
void Delete_OutputStream( OutputStream *t )
{
	if ( t != NULL ) {
		assert( t->pos <= t->maxLength );
		tsi_DeAllocMem( t->mem, t->base );
		tsi_DeAllocMem( t->mem, t );
	}
}

#endif /* ENABLE_WRITE */


#ifdef ENABLE_WRITE

void WriteBitsToStream( OutputStream *out, unsigned long bits, unsigned long count )
{
	/* First we have t->bitCountOut bits, followed by count bits */
	/* We always keep the bits slammed up againts the "left edge" */
	bits       <<= 32 - count - out->bitCountOut;
	out->bitCountOut += count;
	out->bitBufferOut |= bits;

	while ( out->bitCountOut >= 8 ) {
		WriteUnsignedByte( out, (unsigned char)(out->bitBufferOut >> 24) );
		out->bitBufferOut <<= 8;
		out->bitCountOut -= 8;
	}

}

void FlushOutStream( OutputStream *out )
{
	assert( out->bitCountOut < 8 );
	if ( out->bitCountOut > 0 ) {
		WriteUnsignedByte( out, (unsigned char)(out->bitBufferOut >> 24) );
		out->bitCountOut = 0;
	}
	out->bitBufferOut = 0;
}


void WriteUnsignedNumber( OutputStream *out, unsigned long n )
{
	unsigned char value;
	
	do {
		value = (unsigned char)(n & 0x7f);
		if ( n > 0x7f ) {
			value |= 0x80;
			n >>= 7;
		}
		WriteUnsignedByte( out, value );
	} while (value & 0x80);
}
#endif /* ENABLE_WRITE */

unsigned long ReadUnsignedNumber( InputStream *in )
{
	unsigned char value;
	unsigned long n = 0;
	unsigned long shift = 0;
	
	do {
		value = ReadUnsignedByteMacro( in );
		n |= ((value & 0x7f) << shift );
		shift += 7;
	} while (value & 0x80);
	return n; /*****/
}




