/*
 * @(#)mlib_c_ImageAffine.c	1.18 98/09/05
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifdef _MSC_VER
#pragma warning(disable:4068 4244 4101)
#endif /*_MSC_VER*/

#ifdef __SUNPRO_C
#pragma ident  "@(#)mlib_c_ImageAffine.c	1.34  98/08/04 SMI"
#endif /* __SUNPRO_C */

/*
 * FUNCTION
 *      mlib_ImageAffine - image affine transformation with edge condition
 *
 * SYNOPSIS
 *      mlib_status mlib_ImageAffine(mlib_image  *dst,
 *				     mlib_image  *src,
 *				     mlib_f32    *mtx,
 *				     mlib_filter filter,
 *				     mlib_edge   edge)
 *
 * ARGUMENTS
 *      dst       Pointer to destination image
 *      src       Pointer to source image
 *      mtx       Transformation matrix, where
 *			mtx[0] holds a;  mtx[1] holds b;
 *			mtx[2] holds tx; mtx[3] holds c;
 *			mtx[4] holds d;  mtx[5] holds ty.
 *      filter    Type of resampling filter.
 *      edge	  Type of edge condition.
 *
 * DESCRIPTION
 *                      xd = a*xs + b*ys + tx
 *                      yd = c*xs + d*ys + ty
 *
 *	The upper-left corner pixel of an image is located at (0.5, 0.5).
 *
 *	The resampling filter can be one of the following:
 *		MLIB_NEAREST
 *		MLIB_BILINEAR
 *		MLIB_BICUBUIC
 *
 *	The edge condition can be one of the following:
 *		MLIB_EDGE_DST_NO_WRITE	(default)
 *		MLIB_EDGE_DST_FILL_ZERO
 *		MLIB_EDGE_OP_NEAREST
 *		MLIB_EDGE_SRC_EXTEND
 */

#include <stdlib.h>
#include "mlib_image.h"

/***************************************************************/

#define LIMIT 600

#define MLIB_AFFINEDEFPARAM \
  mlib_s32 *leftEdges,      \
  mlib_s32 *rightEdges,     \
  mlib_s32 *xStarts,        \
  mlib_s32 *yStarts,        \
  mlib_s32 *sides,          \
  mlib_u8  *dstData,        \
  mlib_u8  **lineAddr,      \
  mlib_s32 dstYStride       \

#define MLIB_AFFINEDEFPARAM2 MLIB_AFFINEDEFPARAM, mlib_s32 srcYStride

/***************************************************************/

#define MLIB_AFFINECALLPARAM               \
  leftEdges, rightEdges, xStarts, yStarts, \
  sides, dstData, lineAddr, dstYStride

#define MLIB_AFFINECALLPARAM2 MLIB_AFFINECALLPARAM, srcYStride

void mlib_c_ImageAffine_u8_1ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_u8_2ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_u8_3ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_u8_4ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_u8_1ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_u8_2ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_u8_3ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_u8_4ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_u8_1ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_u8_2ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_u8_3ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_u8_4ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_1ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s16_2ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s16_3ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s16_4ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s16_1ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_2ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_3ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_4ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_1ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_2ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_3ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s16_4ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_1ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s32_2ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s32_3ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s32_4ch_nn(MLIB_AFFINEDEFPARAM);
void mlib_c_ImageAffine_s32_1ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_2ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_3ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_4ch_bl(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_1ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_2ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_3ch_bc(MLIB_AFFINEDEFPARAM2);
void mlib_c_ImageAffine_s32_4ch_bc(MLIB_AFFINEDEFPARAM2);

/***************************************************************/

mlib_status mlib_AffineEdges(mlib_s32 *leftEdges,
                             mlib_s32 *rightEdges,
                             mlib_s32 *xStarts,
                             mlib_s32 *yStarts,
                             mlib_s32 *sides,
                             mlib_d64 xClip,
                             mlib_d64 yClip,
                             mlib_d64 wClip,
                             mlib_d64 hClip,
                             mlib_d64 dstW,
                             mlib_d64 dstH,
                             mlib_f32 *mtx,
                             mlib_d64 delta);

/***************************************************************/

void mlib_ImageAffineEdgeZero(mlib_image *dst,
                              mlib_s32 *leftEdges,
                              mlib_s32 *rightEdges,
                              mlib_s32 *sides,
                              mlib_s32 *leftEdgesE,
                              mlib_s32 *rightEdgesE,
                              mlib_s32 *sidesE);

/***************************************************************/

void mlib_ImageAffineEdgeNearest(mlib_image *dst,
                                mlib_s32 *leftEdges,
                                mlib_s32 *rightEdges,
                                mlib_s32 *sides,
                                mlib_s32 *leftEdgesE,
                                mlib_s32 *rightEdgesE,
                                mlib_s32 *xStartsE,
                                mlib_s32 *yStartsE,
                                mlib_s32 *sidesE,
                                mlib_u8  **lineAddr);

/***************************************************************/

void mlib_ImageAffineEdgeExtend_BL(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32 *leftEdges,
                                  mlib_s32 *rightEdges,
                                  mlib_s32 *sides,
                                  mlib_s32 *leftEdgesE,
                                  mlib_s32 *rightEdgesE,
                                  mlib_s32 *xStartsE,
                                  mlib_s32 *yStartsE,
                                  mlib_s32 *sidesE,
                                  mlib_u8  **lineAddr);

/***************************************************************/

void mlib_ImageAffineEdgeExtend_BC(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32 *leftEdges,
                                  mlib_s32 *rightEdges,
                                  mlib_s32 *sides,
                                  mlib_s32 *leftEdgesE,
                                  mlib_s32 *rightEdgesE,
                                  mlib_s32 *xStartsE,
                                  mlib_s32 *yStartsE,
                                  mlib_s32 *sidesE,
                                  mlib_u8  **lineAddr);

/***************************************************************/

#define PREPARE(X, Y, W, H, DELTA) \
  status = mlib_AffineEdges(leftEdges, rightEdges, xStarts, yStarts, \
    sides, X, Y, W, H, dstWidth, dstHeight, mtx, DELTA);

/***************************************************************/

#ifdef MLIB_TEST

mlib_status  mlib_c_ImageAffine(mlib_image  *dst,
                                mlib_image  *src,
                                mlib_f32    *mtx,
                                mlib_filter filter,
                                mlib_edge   edge)

#else

mlib_status  mlib_ImageAffine(mlib_image  *dst,
                              mlib_image  *src,
                              mlib_f32    *mtx,
                              mlib_filter filter,
                              mlib_edge   edge)

#endif

{
  mlib_type  srcType, dstType;
  mlib_s32   nSrcChan, nDstChan;
  mlib_s32   srcWidth,  dstWidth, srcHeight, dstHeight;
  mlib_s32   srcYStride, dstYStride;
  mlib_s32   sides[4];
  mlib_s32   *leftEdges, *rightEdges, *xStarts, *yStarts;
  mlib_s32   leArr[LIMIT], reArr[LIMIT], xsArr[LIMIT], ysArr[LIMIT];
  mlib_u8    *laArr[LIMIT], **lineAddr, *memBuffer;
  mlib_u8    *srcLinePtr;
  mlib_u8    *srcData, *dstData;
  mlib_status status;
  mlib_s32   i;

  /* check for obvious errors */
  if (src == NULL || (srcData = mlib_ImageGetData(src)) == NULL ||
    dst == NULL || (dstData = mlib_ImageGetData(dst)) == NULL)
    return MLIB_NULLPOINTER;

  srcType      = mlib_ImageGetType(src);
  dstType      = mlib_ImageGetType(dst);
  nSrcChan     = mlib_ImageGetChannels(src);
  nDstChan     = mlib_ImageGetChannels(dst);
  srcWidth     = mlib_ImageGetWidth(src);
  srcHeight    = mlib_ImageGetHeight(src);
  dstWidth     = mlib_ImageGetWidth(dst);
  dstHeight    = mlib_ImageGetHeight(dst);
  srcYStride   = mlib_ImageGetStride(src);
  dstYStride   = mlib_ImageGetStride(dst);

  if (nSrcChan <= 0 || nSrcChan > 4 ||
    nDstChan <= 0 || nDstChan > 4 ||
    srcWidth <= 0 || dstWidth <= 0 ||
    srcHeight <= 0 || dstHeight <= 0 ||
    srcYStride <= 0 || dstYStride <= 0)
    return MLIB_OUTOFRANGE;

  if (srcType != dstType || nSrcChan != nDstChan)
    return MLIB_FAILURE;

  if (srcType != MLIB_BYTE && srcType != MLIB_SHORT && srcType != MLIB_INT)
    return MLIB_FAILURE;

  if (srcHeight < LIMIT && dstHeight < LIMIT)
  {
     lineAddr    = laArr;
     leftEdges   = leArr;
     rightEdges  = reArr;
     xStarts     = xsArr;
     yStarts     = ysArr;
  }
  else
  {
    memBuffer = mlib_malloc(4*dstHeight*sizeof(mlib_s32) +
      srcHeight*sizeof(mlib_u8 *));
    if (memBuffer == NULL)  return MLIB_FAILURE;
    leftEdges   = (mlib_s32 *)(memBuffer);
    rightEdges  = (mlib_s32 *)(memBuffer + dstHeight * sizeof(mlib_s32));
    xStarts     = (mlib_s32 *)(memBuffer + 2 * dstHeight * sizeof(mlib_s32));
    yStarts     = (mlib_s32 *)(memBuffer + 3 * dstHeight * sizeof(mlib_s32));
    lineAddr    = (mlib_u8 **)(memBuffer + 4 * dstHeight * sizeof(mlib_s32));
  }

  switch(filter)
  {
    case MLIB_NEAREST:
      PREPARE(0., 0., srcWidth, srcHeight, 0.)
      break;

    case MLIB_BILINEAR:
      PREPARE(.5, .5, srcWidth-.5, srcHeight-.5, -.5)
      break;

    case MLIB_BICUBIC:
      PREPARE(1.5, 1.5, srcWidth-1.5, srcHeight-1.5, -.5)
      break;
  }

  if (status != MLIB_SUCCESS) return MLIB_SUCCESS;

  srcLinePtr = (mlib_u8 *)srcData;

  for (i = 0; i < srcHeight; i++)
  {
    lineAddr[i] = srcLinePtr;
    srcLinePtr += srcYStride;
  }

  dstData += (sides[0]-1)*dstYStride;

  switch (srcType) {
    case MLIB_BYTE: switch (filter) {
      case MLIB_NEAREST: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_u8_1ch_nn(MLIB_AFFINECALLPARAM); break;
        case 2: mlib_c_ImageAffine_u8_2ch_nn(MLIB_AFFINECALLPARAM); break;
        case 3: mlib_c_ImageAffine_u8_3ch_nn(MLIB_AFFINECALLPARAM); break;
        case 4: mlib_c_ImageAffine_u8_4ch_nn(MLIB_AFFINECALLPARAM); break;
      } break;
      case MLIB_BILINEAR: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_u8_1ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 2: mlib_c_ImageAffine_u8_2ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 3: mlib_c_ImageAffine_u8_3ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 4: mlib_c_ImageAffine_u8_4ch_bl(MLIB_AFFINECALLPARAM2); break;
      } break;
      case MLIB_BICUBIC: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_u8_1ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 2: mlib_c_ImageAffine_u8_2ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 3: mlib_c_ImageAffine_u8_3ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 4: mlib_c_ImageAffine_u8_4ch_bc(MLIB_AFFINECALLPARAM2); break;
      } break;
    } break;
    case MLIB_SHORT: switch (filter) {
      case MLIB_NEAREST: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_s16_1ch_nn(MLIB_AFFINECALLPARAM); break;
        case 2: mlib_c_ImageAffine_s16_2ch_nn(MLIB_AFFINECALLPARAM); break;
        case 3: mlib_c_ImageAffine_s16_3ch_nn(MLIB_AFFINECALLPARAM); break;
        case 4: mlib_c_ImageAffine_s16_4ch_nn(MLIB_AFFINECALLPARAM); break;
      } break;
      case MLIB_BILINEAR: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_s16_1ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 2: mlib_c_ImageAffine_s16_2ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 3: mlib_c_ImageAffine_s16_3ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 4: mlib_c_ImageAffine_s16_4ch_bl(MLIB_AFFINECALLPARAM2); break;
      } break;
      case MLIB_BICUBIC: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_s16_1ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 2: mlib_c_ImageAffine_s16_2ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 3: mlib_c_ImageAffine_s16_3ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 4: mlib_c_ImageAffine_s16_4ch_bc(MLIB_AFFINECALLPARAM2); break;
      } break;
    } break;
    case MLIB_INT: switch (filter) {
      case MLIB_NEAREST: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_s32_1ch_nn(MLIB_AFFINECALLPARAM); break;
        case 2: mlib_c_ImageAffine_s32_2ch_nn(MLIB_AFFINECALLPARAM); break;
        case 3: mlib_c_ImageAffine_s32_3ch_nn(MLIB_AFFINECALLPARAM); break;
        case 4: mlib_c_ImageAffine_s32_4ch_nn(MLIB_AFFINECALLPARAM); break;
      } break;
      case MLIB_BILINEAR: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_s32_1ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 2: mlib_c_ImageAffine_s32_2ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 3: mlib_c_ImageAffine_s32_3ch_bl(MLIB_AFFINECALLPARAM2); break;
        case 4: mlib_c_ImageAffine_s32_4ch_bl(MLIB_AFFINECALLPARAM2); break;
      } break;
      case MLIB_BICUBIC: switch(nSrcChan) {
        case 1: mlib_c_ImageAffine_s32_1ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 2: mlib_c_ImageAffine_s32_2ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 3: mlib_c_ImageAffine_s32_3ch_bc(MLIB_AFFINECALLPARAM2); break;
        case 4: mlib_c_ImageAffine_s32_4ch_bc(MLIB_AFFINECALLPARAM2); break;
      } break;
    } break;
  }

  if ((filter != MLIB_NEAREST) && (edge != MLIB_EDGE_DST_NO_WRITE))
  {
    mlib_s32  *leftEdgesE, *rightEdgesE, *xStartsE, *yStartsE;
    mlib_s32  leArrE[LIMIT], reArrE[LIMIT], xsArrE[LIMIT], ysArrE[LIMIT];
    mlib_u8   *laArrE[LIMIT], **lineAddrE, *memBufferE;
    mlib_s32  sidesE[4];

    if (srcHeight < LIMIT && dstHeight < LIMIT)
    {
       lineAddrE    = laArrE;
       leftEdgesE   = leArrE;
       rightEdgesE  = reArrE;
       xStartsE     = xsArrE;
       yStartsE     = ysArrE;
    }
    else
    {
      memBufferE = mlib_malloc(4*dstHeight*sizeof(mlib_s32) +
                              srcHeight*sizeof(mlib_u8 *));

      if (memBufferE == NULL)
        return MLIB_FAILURE;

      leftEdgesE   = (mlib_s32 *)(memBufferE);
      rightEdgesE  = (mlib_s32 *)(memBufferE + dstHeight * sizeof(mlib_s32));
      xStartsE     = (mlib_s32 *)(memBufferE + 2 * dstHeight * sizeof(mlib_s32));
      yStartsE     = (mlib_s32 *)(memBufferE + 3 * dstHeight * sizeof(mlib_s32));
      lineAddrE    = (mlib_u8 **)(memBufferE + 4 * dstHeight * sizeof(mlib_s32));
    }

    mlib_AffineEdges(leftEdgesE, rightEdgesE, xStartsE, yStartsE, sidesE,
                     0.0, 0.0, srcWidth, srcHeight, dstWidth, dstHeight, mtx, 0);

    switch (edge)
    {
      case MLIB_EDGE_DST_FILL_ZERO:
        mlib_ImageAffineEdgeZero(dst, leftEdges, rightEdges, sides,
                                leftEdgesE, rightEdgesE, sidesE);
      break;

      case MLIB_EDGE_OP_NEAREST:
        mlib_ImageAffineEdgeNearest(dst, leftEdges, rightEdges, sides,
                                    leftEdgesE, rightEdgesE,
                                    xStartsE, yStartsE, sidesE, lineAddr);
      break;

      case MLIB_EDGE_SRC_EXTEND:
        if (filter == MLIB_BILINEAR)
          mlib_ImageAffineEdgeExtend_BL(dst, src, leftEdges, rightEdges, sides,
                                  leftEdgesE, rightEdgesE, xStartsE, yStartsE,
                                  sidesE, lineAddr);
        else if (filter == MLIB_BICUBIC)
          mlib_ImageAffineEdgeExtend_BC(dst, src, leftEdges, rightEdges, sides,
                                  leftEdgesE, rightEdgesE, xStartsE, yStartsE,
                                  sidesE, lineAddr);
      break;
    }

    if (lineAddrE != laArrE)
      mlib_free(memBufferE);
  }

  if (lineAddr != laArr)
    mlib_free(memBuffer);

  return MLIB_SUCCESS;
}

