/*
 * @(#)mlib_c_ImageAffineIndex_BL.c	1.3 98/09/05
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifdef _MSC_VER
#pragma warning(disable:4068 4244 4101)
#endif /*_MSC_VER*/

#ifdef __SUNPRO_C
#pragma ident   "@(#)mlib_c_ImageAffineIndex_BL.c	1.9    98/06/26 SMI"
#endif /* __SUNPRO_C */

#include <stdlib.h>
#include "mlib_image.h"
#include "mlib_ImageColormap.h"
#include "mlib_c_ImageFilters.h"

/*****************************************************************************/

#define MLIB_LIMIT 1024
#define MLIB_SHIFT 16
#define MLIB_PREC  (1 << MLIB_SHIFT)
#define MLIB_MASK  (MLIB_PREC - 1)

#ifdef MLIB_OS64BIT
#define MLIB_POINTER_SHIFT(P) (P >> (MLIB_SHIFT-3)) &~ 7
#define MLIB_POINTER_GET(A,P) (*(MLIB_TYPE **)((mlib_u8 *)A + P))
#else
#define MLIB_POINTER_SHIFT(P) (P >> (MLIB_SHIFT-2)) &~ 3
#define MLIB_POINTER_GET(A,P) (*(MLIB_TYPE **)((mlib_addr)A + P))
#endif

/*****************************************************************************/
#define DECLAREVAR()                                                          \
  mlib_s32  xLeft, xRight, X, Y;                                              \
  mlib_s32  xSrc, ySrc;                                                       \
  mlib_s32  yStart = sides[0];                                                \
  mlib_s32  yFinish = sides[1];                                               \
  mlib_s32  dX = sides[2];                                                    \
  mlib_s32  dY = sides[3];                                                    \
  MLIB_TYPE *srcIndexPtr0, *srcIndexPtr1;                                     \
  MLIB_TYPE *dstIndexPtr;                                                     \
  mlib_s32  i, j;                                                             \
  mlib_d64  scale = 1.0 / 65536.0;                                             
/*****************************************************************************/
#define DECLARE_INTERNAL_VAR_3CH()                                            \
  mlib_d64  fdx, fdy;                                                         \
  mlib_d64  a00_0, a01_0, a10_0, a11_0;                                       \
  mlib_d64  a00_1, a01_1, a10_1, a11_1;                                       \
  mlib_d64  a00_2, a01_2, a10_2, a11_2;                                       \
  mlib_d64  pix0_0, pix1_0, res0;                                             \
  mlib_d64  pix0_1, pix1_1, res1;                                             \
  mlib_d64  pix0_2, pix1_2, res2;
/*****************************************************************************/  
#define DECLARE_INTERNAL_VAR_4CH()                                            \
  mlib_d64  fdx, fdy;                                                         \
  mlib_d64  a00_0, a01_0, a10_0, a11_0;                                       \
  mlib_d64  a00_1, a01_1, a10_1, a11_1;                                       \
  mlib_d64  a00_2, a01_2, a10_2, a11_2;                                       \
  mlib_d64  a00_3, a01_3, a10_3, a11_3;                                       \
  mlib_d64  pix0_0, pix1_0, res0;                                             \
  mlib_d64  pix0_1, pix1_1, res1;                                             \
  mlib_d64  pix0_2, pix1_2, res2;                                             \
  mlib_d64  pix0_3, pix1_3, res3;  
/*****************************************************************************/
#define CLIP()                                                                \
  dstData += dstYStride;                                                      \
  xLeft = leftEdges[j];                                                       \
  xRight = rightEdges[j];                                                     \
  X = xStarts[j];                                                             \
  Y = yStarts[j];                                                             \
  if (xLeft > xRight)                                                         \
      continue;                                                               \
  dstIndexPtr = (MLIB_TYPE *)dstData + xLeft;
/*****************************************************************************/
#define GET_PIXELS_POINTERS()                                                 \
  fdx = (X & MLIB_MASK) * scale;                                              \
  fdy = (Y & MLIB_MASK) * scale;                                              \
  ySrc = MLIB_POINTER_SHIFT(Y);  Y += dY;                                     \
  xSrc = X>>MLIB_SHIFT;  X += dX;                                             \
  srcIndexPtr0 = MLIB_POINTER_GET(lineAddr, ySrc) + xSrc;                     \
  srcIndexPtr1 = (MLIB_TYPE *)((mlib_u8 *)srcIndexPtr0 + srcYStride);
/*****************************************************************************/  
#define GET_COLOR_POINTERS(ind)                                               \
  pcolor00 = (lut + srcIndexPtr0[0]*ind);                                     \
  pcolor10 = (lut + srcIndexPtr1[0]*ind);                                     \
  pcolor01 = (lut + srcIndexPtr0[1]*ind);                                     \
  pcolor11 = (lut + srcIndexPtr1[1]*ind); 
/*****************************************************************************/
#define COUNT_BL(ind)                                                         \
  pix0_##ind = a00_##ind + fdy * (a10_##ind - a00_##ind);                     \
  pix1_##ind = a01_##ind + fdy * (a11_##ind - a01_##ind);                     \
  res##ind = pix0_##ind + fdx * (pix1_##ind - pix0_##ind);
/*****************************************************************************/
#define COUNT_BL_3CH()                                                        \
  COUNT_BL(0)                                                                 \
  COUNT_BL(1)                                                                 \
  COUNT_BL(2)
/*****************************************************************************/
#define COUNT_BL_4CH()                                                        \
  COUNT_BL(0)                                                                 \
  COUNT_BL(1)                                                                 \
  COUNT_BL(2)                                                                 \
  COUNT_BL(3)  
/*****************************************************************************/  
#define LOAD(ind)                                                             \
  a00_##ind = pcolor00[ind];                                                  \
  a01_##ind = pcolor01[ind];                                                  \
  a10_##ind = pcolor10[ind];                                                  \
  a11_##ind = pcolor11[ind];   
/*****************************************************************************/
#define LOAD_3CH()                                                            \
  LOAD(0)                                                                     \
  LOAD(1)                                                                     \
  LOAD(2)                                                               
/*****************************************************************************/
#define LOAD_4CH()                                                            \
  LOAD(0)                                                                     \
  LOAD(1)                                                                     \
  LOAD(2)                                                                     \
  LOAD(3)  
/*****************************************************************************/
#define STORE_INTO_INTERM_BUF_3CH()                                           \
  dstPixelPtr[0] = res0;                                                      \
  dstPixelPtr[1] = res1;                                                      \
  dstPixelPtr[2] = res2;
/*****************************************************************************/
#define STORE_INTO_INTERM_BUF_4CH()                                           \
  dstPixelPtr[0] = res0;                                                      \
  dstPixelPtr[1] = res1;                                                      \
  dstPixelPtr[2] = res2;                                                      \
  dstPixelPtr[3] = res3;  
/*****************************************************************************/

#undef  MLIB_TYPE
#define MLIB_TYPE mlib_u8

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_U8_U8_3CH_BL(mlib_s32   *leftEdges, 
                                                 mlib_s32   *rightEdges,
                                                 mlib_s32   *xStarts,
                                                 mlib_s32   *yStarts,
                                                 mlib_s32   *sides,
                                                 mlib_u8    *dstData,
                                                 mlib_u8    **lineAddr,
                                                 mlib_s32   dstYStride,
                                                 mlib_s32   srcYStride,
                                                 void       *colormap)
{
  DECLAREVAR()
  mlib_u8  *dstPixelPtr;  
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 3;    
  mlib_d64  dstRowData[MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;
  
  if (dstYStride > 2*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(3 * dstYStride);

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_3CH()

    CLIP()    
    dstPixelPtr = (mlib_u8 *)dstRowPtr;
    
    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(3)
    LOAD_3CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 3) {
      COUNT_BL_3CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(3)
      LOAD_3CH()

      STORE_INTO_INTERM_BUF_3CH()
    }
    
    COUNT_BL_3CH()
    STORE_INTO_INTERM_BUF_3CH()

    mlib_ImageColorTrue2IndexLine_U8_U8_3((mlib_u8 *)dstRowPtr,
                                          dstIndexPtr,
                                          xRight - xLeft + 1,
                                          colormap);
  }
  
  if (dstYStride > 2*MLIB_LIMIT)
    mlib_free(dstRowPtr);

  return MLIB_SUCCESS;
}

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_U8_S16_3CH_BL(mlib_s32  *leftEdges, 
                                                  mlib_s32   *rightEdges,
                                                  mlib_s32   *xStarts,
                                                  mlib_s32   *yStarts,
                                                  mlib_s32   *sides,
                                                  mlib_u8    *dstData,
                                                  mlib_u8    **lineAddr,
                                                  mlib_s32   dstYStride,
                                                  mlib_s32   srcYStride,
                                                  void       *colormap)
{
  DECLAREVAR()
  mlib_s16  *dstPixelPtr;
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 3; 
  mlib_d64  dstRowData[2 * MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;

  if (dstYStride > 2*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(6 * dstYStride);

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_3CH()
    
    CLIP()
    dstPixelPtr = (mlib_s16 *)dstRowPtr;
    
    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(3)
    LOAD_3CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 3) {
      COUNT_BL_3CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(3)
      LOAD_3CH()

      STORE_INTO_INTERM_BUF_3CH()
    }
    
    COUNT_BL_3CH()
    STORE_INTO_INTERM_BUF_3CH()

    mlib_ImageColorTrue2IndexLine_S16_U8_3((mlib_s16 *)dstRowPtr,
                                          dstIndexPtr,
                                          xRight - xLeft + 1,
                                          colormap);
  }

  if (dstYStride > 2*MLIB_LIMIT)
    mlib_free(dstRowPtr);

  return MLIB_SUCCESS;
}

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_U8_U8_4CH_BL(mlib_s32  *leftEdges, 
                                                 mlib_s32   *rightEdges,
                                                 mlib_s32   *xStarts,
                                                 mlib_s32   *yStarts,
                                                 mlib_s32   *sides,
                                                 mlib_u8    *dstData,
                                                 mlib_u8    **lineAddr,
                                                 mlib_s32   dstYStride,
                                                 mlib_s32   srcYStride,
                                                 void       *colormap)
{
  DECLAREVAR()
  mlib_u8   *dstPixelPtr;
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 4;  
  mlib_d64  dstRowData[MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;

  if (dstYStride > 2*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(4 * dstYStride);

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_4CH()
    
    CLIP()
    dstPixelPtr = (mlib_u8 *)dstRowPtr;
    
    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(4)
    LOAD_4CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 4) {
      COUNT_BL_4CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(4)
      LOAD_4CH()

      STORE_INTO_INTERM_BUF_4CH()
    }
    
    COUNT_BL_4CH()
    STORE_INTO_INTERM_BUF_4CH()

    mlib_ImageColorTrue2IndexLine_U8_U8_4((mlib_u8 *)dstRowPtr,
                                          dstIndexPtr,
                                          xRight - xLeft + 1,
                                          colormap);
  }
  
  if (dstYStride > 2*MLIB_LIMIT)
    mlib_free(dstRowPtr);

  return MLIB_SUCCESS;
}

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_U8_S16_4CH_BL(mlib_s32  *leftEdges, 
                                                  mlib_s32   *rightEdges,
                                                  mlib_s32   *xStarts,
                                                  mlib_s32   *yStarts,
                                                  mlib_s32   *sides,
                                                  mlib_u8    *dstData,
                                                  mlib_u8    **lineAddr,
                                                  mlib_s32   dstYStride,
                                                  mlib_s32   srcYStride,
                                                  void       *colormap)
{
  DECLAREVAR()
  mlib_s16  *dstPixelPtr;
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 4;   
  mlib_d64  dstRowData[2 * MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;

  if (dstYStride > 2*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(8 * dstYStride);

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_4CH()
    
    CLIP()
    dstPixelPtr = (mlib_s16 *)dstRowPtr;

    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(4)
    LOAD_4CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 4) {
      COUNT_BL_4CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(4)
      LOAD_4CH()

      STORE_INTO_INTERM_BUF_4CH()
    }
    
    COUNT_BL_4CH()
    STORE_INTO_INTERM_BUF_4CH()

    mlib_ImageColorTrue2IndexLine_S16_U8_4((mlib_s16 *)dstRowPtr,
                                          dstIndexPtr,
                                          xRight - xLeft + 1,
                                          colormap);
  }
  
  if (dstYStride > 2*MLIB_LIMIT)
    mlib_free(dstRowPtr);
  
  return MLIB_SUCCESS;
}

/*****************************************************************************/

#undef  MLIB_TYPE
#define MLIB_TYPE mlib_s16

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_S16_U8_3CH_BL(mlib_s32  *leftEdges, 
                                                  mlib_s32   *rightEdges,
                                                  mlib_s32   *xStarts,
                                                  mlib_s32   *yStarts,
                                                  mlib_s32   *sides,
                                                  mlib_u8    *dstData,
                                                  mlib_u8    **lineAddr,
                                                  mlib_s32   dstYStride,
                                                  mlib_s32   srcYStride,
                                                  void       *colormap)
{
  DECLAREVAR()
  mlib_u8   *dstPixelPtr;
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 3;
  
  mlib_d64  dstRowData[MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;

  if (dstYStride > 4*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(3 * (dstYStride >> 1));

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_3CH()

    CLIP()    
    dstPixelPtr = (mlib_u8 *)dstRowPtr;
    
    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(3)
    LOAD_3CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 3) {
      COUNT_BL_3CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(3)
      LOAD_3CH()

      STORE_INTO_INTERM_BUF_3CH()
    }
    
    COUNT_BL_3CH()
    STORE_INTO_INTERM_BUF_3CH()

    mlib_ImageColorTrue2IndexLine_U8_S16_3((mlib_u8 *)dstRowPtr,
                                          dstIndexPtr,
                                          xRight - xLeft + 1,
                                          colormap);
  }
  
  if (dstYStride > 4*MLIB_LIMIT)
    mlib_free(dstRowPtr);
  
  return MLIB_SUCCESS;
}

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_S16_S16_3CH_BL(mlib_s32  *leftEdges, 
                                                   mlib_s32   *rightEdges,
                                                   mlib_s32   *xStarts,
                                                   mlib_s32   *yStarts,
                                                   mlib_s32   *sides,
                                                   mlib_u8    *dstData,
                                                   mlib_u8    **lineAddr,
                                                   mlib_s32   dstYStride,
                                                   mlib_s32   srcYStride,
                                                   void       *colormap)
{
  DECLAREVAR()
  mlib_s16  *dstPixelPtr;
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 3;
  mlib_d64  dstRowData[2 * MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;

  if (dstYStride > 4*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(3 * dstYStride);

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_3CH()
    
    CLIP()
    dstPixelPtr = (mlib_s16 *)dstRowPtr;
    
    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(3)
    LOAD_3CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 3) {
      COUNT_BL_3CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(3)
      LOAD_3CH()

      STORE_INTO_INTERM_BUF_3CH()
    }
    
    COUNT_BL_3CH()
    STORE_INTO_INTERM_BUF_3CH()

    mlib_ImageColorTrue2IndexLine_S16_S16_3((mlib_s16 *)dstRowPtr,
                                          dstIndexPtr,
                                          xRight - xLeft + 1,
                                          colormap);
  }
  if (dstYStride > 4*MLIB_LIMIT)
    mlib_free(dstRowPtr);

  return MLIB_SUCCESS;
}

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_S16_U8_4CH_BL(mlib_s32  *leftEdges, 
                                                  mlib_s32   *rightEdges,
                                                  mlib_s32   *xStarts,
                                                  mlib_s32   *yStarts,
                                                  mlib_s32   *sides,
                                                  mlib_u8    *dstData,
                                                  mlib_u8    **lineAddr,
                                                  mlib_s32   dstYStride,
                                                  mlib_s32   srcYStride,
                                                  void       *colormap)
{
  DECLAREVAR()
  mlib_u8   *dstPixelPtr;
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 4;   
  mlib_d64  dstRowData[MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;

  if (dstYStride > 4*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(2 * dstYStride);

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_4CH()
    
    CLIP()
    dstPixelPtr = (mlib_u8 *)dstRowPtr;
    
    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(4)
    LOAD_4CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 4) {
      COUNT_BL_4CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(4)
      LOAD_4CH()

      STORE_INTO_INTERM_BUF_4CH()
    }
    
    COUNT_BL_4CH()
    STORE_INTO_INTERM_BUF_4CH()

    mlib_ImageColorTrue2IndexLine_U8_S16_4((mlib_u8 *)dstRowPtr,
                                          dstIndexPtr,
                                          xRight - xLeft + 1,
                                          colormap);
  }
  
  if (dstYStride > 4*MLIB_LIMIT)
    mlib_free(dstRowPtr);
  
  return MLIB_SUCCESS;
}

/*****************************************************************************/

mlib_status mlib_c_ImageAffineIndex_S16_S16_4CH_BL(mlib_s32  *leftEdges, 
                                                   mlib_s32   *rightEdges,
                                                   mlib_s32   *xStarts,
                                                   mlib_s32   *yStarts,
                                                   mlib_s32   *sides,
                                                   mlib_u8    *dstData,
                                                   mlib_u8    **lineAddr,
                                                   mlib_s32   dstYStride,
                                                   mlib_s32   srcYStride,
                                                   void       *colormap)
{
  DECLAREVAR()
  mlib_s16  *dstPixelPtr;
  mlib_d64 *pcolor00, *pcolor10, *pcolor01, *pcolor11;
  mlib_d64 *lut = (mlib_d64*)mlib_ImageGetLutDoubleData(colormap) -   
                             mlib_ImageGetLutOffset(colormap) * 4;   
  mlib_d64  dstRowData[2 * MLIB_LIMIT];
  mlib_d64  *dstRowPtr = dstRowData;

  if (dstYStride > 4*MLIB_LIMIT)
    dstRowPtr = (mlib_d64 *)mlib_malloc(4 * dstYStride);

  if (dstRowPtr == NULL)
    return MLIB_FAILURE;

  for (j = yStart; j <= yFinish; j++) {
    DECLARE_INTERNAL_VAR_4CH()
    
    CLIP()
    dstPixelPtr = (mlib_s16 *)dstRowPtr;

    GET_PIXELS_POINTERS()
    GET_COLOR_POINTERS(4)
    LOAD_4CH()
    
#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (i = 0; i < (xRight - xLeft); i++, dstPixelPtr += 4) {
      COUNT_BL_4CH()
      
      GET_PIXELS_POINTERS()
      GET_COLOR_POINTERS(4)
      LOAD_4CH()

      STORE_INTO_INTERM_BUF_4CH()
    }
    
    COUNT_BL_4CH()
    STORE_INTO_INTERM_BUF_4CH()

    mlib_ImageColorTrue2IndexLine_S16_S16_4((mlib_s16 *)dstRowPtr,
                                           dstIndexPtr,
                                           xRight - xLeft + 1,
                                           colormap);
  }
  if (dstYStride > 4*MLIB_LIMIT)
    mlib_free(dstRowPtr);
  
  return MLIB_SUCCESS;
}

/*****************************************************************************/

