/*
 * @(#)mlib_c_ImageAffine_BC.c	1.7 98/09/05
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


#if _MSC_VER
#pragma warning(disable: 4068 4244 4101)
#endif

#ifdef __SUNPRO_C
#pragma ident  "@(#)mlib_c_ImageAffine_BC.c	1.6  98/08/04 SMI"
#endif /* __SUNPRO_C */

/*
 * FUNCTION
 *      mlib_c_ImageAffine_u8_1ch_bc
 *      mlib_c_ImageAffine_u8_2ch_bc
 *      mlib_c_ImageAffine_u8_3ch_bc
 *      mlib_c_ImageAffine_u8_4ch_bc
 *      mlib_c_ImageAffine_s16_1ch_bc
 *      mlib_c_ImageAffine_s16_2ch_bc
 *      mlib_c_ImageAffine_s16_3ch_bc
 *      mlib_c_ImageAffine_s16_4ch_bc
 *      mlib_c_ImageAffine_s32_1ch_bc
 *      mlib_c_ImageAffine_s32_2ch_bc
 *      mlib_c_ImageAffine_s32_3ch_bc
 *      mlib_c_ImageAffine_s32_4ch_bc
 *        - image affine transformation with Bicubic filtering
 * SYNOPSIS
 *      void mlib_c_ImageAffine_[u8|s16|s32]_?ch_bc
 *                                  (mlib_s32 *leftEdges,
 *                                   mlib_s32 *rightEdges,
 *                                   mlib_s32 *xStarts,
 *                                   mlib_s32 *yStarts,
 *                                   mlib_s32 *sides,
 *                                   mlib_u8  *dstData,
 *                                   mlib_u8  **lineAddr,
 *                                   mlib_s32 dstYStride,
 *                                   mlib_s32 srcYStride)
 *
 * ARGUMENTS
 *      leftEdges  array[dstHeight] of xLeft coordinates
 *      RightEdges array[dstHeight] of xRight coordinates
 *      xStarts    array[dstHeight] of xStart * 65536 coordinates
 *      yStarts    array[dstHeight] of yStart * 65536 coordinates
 *      sides      output array[4]. sides[0] is yStart, sides[1] is yFinish,
 *                 sides[2] is dx * 65536, sides[3] is dy * 65536
 *      dstData    pointer to the first pixel on (yStart - 1) line
 *      lineAddr   array[srcHeight] of pointers to the first pixel on
 *                 the corresponding lines
 *      dstYStride stride of destination image
 *      srcYStride stride of source image
 *
 * DESCRIPTION
 *      The functions step along the lines from xLeft to xRight and apply
 *      the bicubic filtering.
 *
 */

#include "mlib_image.h"
#include "mlib_c_ImageFilters.h"

extern mlib_d64 mlib_U82D64[];

/***************************************************************/

#define MLIB_SHIFT       16
#define MLIB_PREC        (1 << MLIB_SHIFT)
#define MLIB_MASK        (MLIB_PREC - 1)

/***************************************************************/

#define MLIB_AFFINEDEFPARAM \
  mlib_s32 *leftEdges,      \
  mlib_s32 *rightEdges,     \
  mlib_s32 *xStarts,        \
  mlib_s32 *yStarts,        \
  mlib_s32 *sides,          \
  mlib_u8  *dstData,        \
  mlib_u8  **lineAddr,      \
  mlib_s32 dstYStride       \

#define MLIB_AFFINEDEFPARAM2 MLIB_AFFINEDEFPARAM, mlib_s32 srcYStride

/***************************************************************/

#define MLIB_AFFINECALLPARAM               \
  leftEdges, rightEdges, xStarts, yStarts, \
  sides, dstData, lineAddr, dstYStride

#define MLIB_AFFINECALLPARAM2 MLIB_AFFINECALLPARAM, srcYStride

/***************************************************************/
#define DECLAREVAR               \
  mlib_s32  xLeft, xRight, X, Y; \
  mlib_s32  xSrc, ySrc;          \
  mlib_s32  yStart = sides[0];   \
  mlib_s32  yFinish = sides[1];  \
  mlib_s32  dX = sides[2];       \
  mlib_s32  dY = sides[3];       \
  MLIB_TYPE *srcPixelPtr;        \
  MLIB_TYPE *dstPixelPtr;        \
  MLIB_TYPE *dstLineEnd;         \
  mlib_s32  j;
/***************************************************************/
#define CLIP(N)                                   \
  dstData += dstYStride;                          \
  xLeft = leftEdges[j];                           \
  xRight = rightEdges[j];                         \
  X = xStarts[j];                                 \
  Y = yStarts[j];                                 \
  if (xLeft > xRight) continue;                   \
  dstPixelPtr  = (MLIB_TYPE*)dstData + N * xLeft; \
  dstLineEnd   = (MLIB_TYPE*)dstData + N * xRight;
/***************************************************************/

#undef  MLIB_TYPE
#define MLIB_TYPE mlib_u8
#define FILTER_SHIFT  4
#define FILTER_MASK   (((1 << 8) - 1) << 4)
#define TABLE mlib_filters_u8f_bc

/***************************************************************/

void  mlib_c_ImageAffine_u8_1ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR
  mlib_d64 sat = (mlib_d64)0x7F800000;

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos;
    mlib_f32  *fptr;
    mlib_u8 s0, s1, s2, s3;

    CLIP(1)

    filterpos = (X >> FILTER_SHIFT) & FILTER_MASK;
    fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

    xf0 = fptr[0];
    xf1 = fptr[1];
    xf2 = fptr[2];
    xf3 = fptr[3];

    filterpos = (Y >> FILTER_SHIFT) & FILTER_MASK;
    fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

    yf0 = fptr[0];
    yf1 = fptr[1];
    yf2 = fptr[2];
    yf3 = fptr[3];

    xSrc = (X>>MLIB_SHIFT)-1;
    ySrc = (Y>>MLIB_SHIFT)-1;

    srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + xSrc;
    s0 = srcPixelPtr[0]; s1 = srcPixelPtr[1];
    s2 = srcPixelPtr[2]; s3 = srcPixelPtr[3];

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
    for (; dstPixelPtr <= (dstLineEnd-1); dstPixelPtr++)
    {
      X += dX;
      Y += dY;

      c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
            mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[1]]*xf1 +
            mlib_U82D64[srcPixelPtr[2]]*xf2 + mlib_U82D64[srcPixelPtr[3]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[1]]*xf1 +
            mlib_U82D64[srcPixelPtr[2]]*xf2 + mlib_U82D64[srcPixelPtr[3]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[1]]*xf1 +
            mlib_U82D64[srcPixelPtr[2]]*xf2 + mlib_U82D64[srcPixelPtr[3]]*xf3);

      filterpos = (X >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

      filterpos = (Y >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      dstPixelPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;

      xSrc = (X>>MLIB_SHIFT)-1;
      ySrc = (Y>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + xSrc;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[1];
      s2 = srcPixelPtr[2]; s3 = srcPixelPtr[3];

    }
    c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
          mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[1]]*xf1 +
          mlib_U82D64[srcPixelPtr[2]]*xf2 + mlib_U82D64[srcPixelPtr[3]]*xf3);
    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[1]]*xf1 +
          mlib_U82D64[srcPixelPtr[2]]*xf2 + mlib_U82D64[srcPixelPtr[3]]*xf3);
    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[1]]*xf1 +
          mlib_U82D64[srcPixelPtr[2]]*xf2 + mlib_U82D64[srcPixelPtr[3]]*xf3);

    val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

    dstPixelPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;
  }
}

/***************************************************************/

void  mlib_c_ImageAffine_u8_2ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR
  mlib_d64 sat = (mlib_d64)0x7F800000;

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos, k;
    mlib_f32  *fptr;
    mlib_u8 s0, s1, s2, s3;

    CLIP(2)

    for (k = 0; k < 2; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 2*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[2];
      s2 = srcPixelPtr[4]; s3 = srcPixelPtr[6];

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
      for (; dPtr <= (dstLineEnd-1); dPtr+=2)
      {
        X1 += dX;
        Y1 += dY;

        c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
              mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[2]]*xf1 +
              mlib_U82D64[srcPixelPtr[4]]*xf2 + mlib_U82D64[srcPixelPtr[6]]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[2]]*xf1 +
              mlib_U82D64[srcPixelPtr[4]]*xf2 + mlib_U82D64[srcPixelPtr[6]]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[2]]*xf1 +
              mlib_U82D64[srcPixelPtr[4]]*xf2 + mlib_U82D64[srcPixelPtr[6]]*xf3);

        filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        xf0 = fptr[0];
        xf1 = fptr[1];
        xf2 = fptr[2];
        xf3 = fptr[3];

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        yf0 = fptr[0];
        yf1 = fptr[1];
        yf2 = fptr[2];
        yf3 = fptr[3];

        dPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 2*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[2];
        s2 = srcPixelPtr[4]; s3 = srcPixelPtr[6];

      }
      c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
            mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[2]]*xf1 +
            mlib_U82D64[srcPixelPtr[4]]*xf2 + mlib_U82D64[srcPixelPtr[6]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[2]]*xf1 +
            mlib_U82D64[srcPixelPtr[4]]*xf2 + mlib_U82D64[srcPixelPtr[6]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[2]]*xf1 +
            mlib_U82D64[srcPixelPtr[4]]*xf2 + mlib_U82D64[srcPixelPtr[6]]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

      dPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;
    }
  }
}

/***************************************************************/
void  mlib_c_ImageAffine_u8_3ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR
  mlib_d64 sat = (mlib_d64)0x7F800000;

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos, k;
    mlib_f32  *fptr;
    mlib_u8 s0, s1, s2, s3;

    CLIP(3)

    for (k = 0; k < 3; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 3*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[3];
      s2 = srcPixelPtr[6]; s3 = srcPixelPtr[9];

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
      for (; dPtr <= (dstLineEnd-1); dPtr+=3)
      {
        X1 += dX;
        Y1 += dY;

        c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
              mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[3]]*xf1 +
              mlib_U82D64[srcPixelPtr[6]]*xf2 + mlib_U82D64[srcPixelPtr[9]]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[3]]*xf1 +
              mlib_U82D64[srcPixelPtr[6]]*xf2 + mlib_U82D64[srcPixelPtr[9]]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[3]]*xf1 +
              mlib_U82D64[srcPixelPtr[6]]*xf2 + mlib_U82D64[srcPixelPtr[9]]*xf3);

        filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        xf0 = fptr[0];
        xf1 = fptr[1];
        xf2 = fptr[2];
        xf3 = fptr[3];

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        yf0 = fptr[0];
        yf1 = fptr[1];
        yf2 = fptr[2];
        yf3 = fptr[3];

        dPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 3*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[3];
        s2 = srcPixelPtr[6]; s3 = srcPixelPtr[9];

      }
      c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
            mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[3]]*xf1 +
            mlib_U82D64[srcPixelPtr[6]]*xf2 + mlib_U82D64[srcPixelPtr[9]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[3]]*xf1 +
            mlib_U82D64[srcPixelPtr[6]]*xf2 + mlib_U82D64[srcPixelPtr[9]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[3]]*xf1 +
            mlib_U82D64[srcPixelPtr[6]]*xf2 + mlib_U82D64[srcPixelPtr[9]]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

      dPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;
    }
  }
}

/***************************************************************/

void  mlib_c_ImageAffine_u8_4ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR
  mlib_d64 sat = (mlib_d64)0x7F800000;

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos, k;
    mlib_f32  *fptr;
    mlib_u8 s0, s1, s2, s3;

    CLIP(4)

    for (k = 0; k < 4; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 4*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[4];
      s2 = srcPixelPtr[8]; s3 = srcPixelPtr[12];

#ifdef __SUNPRO_C
#pragma pipeloop(0)
#endif /* __SUNPRO_C */
      for (; dPtr <= (dstLineEnd-1); dPtr+=4)
      {
        X1 += dX;
        Y1 += dY;

        c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
              mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[4]]*xf1 +
              mlib_U82D64[srcPixelPtr[8]]*xf2 + mlib_U82D64[srcPixelPtr[12]]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[4]]*xf1 +
              mlib_U82D64[srcPixelPtr[8]]*xf2 + mlib_U82D64[srcPixelPtr[12]]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[4]]*xf1 +
              mlib_U82D64[srcPixelPtr[8]]*xf2 + mlib_U82D64[srcPixelPtr[12]]*xf3);

        filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        xf0 = fptr[0];
        xf1 = fptr[1];
        xf2 = fptr[2];
        xf3 = fptr[3];

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        yf0 = fptr[0];
        yf1 = fptr[1];
        yf2 = fptr[2];
        yf3 = fptr[3];

        dPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 4*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[4];
        s2 = srcPixelPtr[8]; s3 = srcPixelPtr[12];

      }
      c0 = (mlib_U82D64[s0]*xf0 + mlib_U82D64[s1]*xf1 +
            mlib_U82D64[s2]*xf2 + mlib_U82D64[s3]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c1 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[4]]*xf1 +
            mlib_U82D64[srcPixelPtr[8]]*xf2 + mlib_U82D64[srcPixelPtr[12]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[4]]*xf1 +
            mlib_U82D64[srcPixelPtr[8]]*xf2 + mlib_U82D64[srcPixelPtr[12]]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (mlib_U82D64[srcPixelPtr[0]]*xf0 + mlib_U82D64[srcPixelPtr[4]]*xf1 +
            mlib_U82D64[srcPixelPtr[8]]*xf2 + mlib_U82D64[srcPixelPtr[12]]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

      dPtr[0] = ((int)(val0 - sat) >> 24) ^ 0x80;
    }
  }
}

/***************************************************************/

#undef  MLIB_TYPE
#define MLIB_TYPE mlib_s16
#undef  FILTER_SHIFT
#define FILTER_SHIFT 3
#undef  FILTER_MASK
#define FILTER_MASK  (((1 << 9) - 1) << 4)
#undef  TABLE
#define TABLE mlib_filters_s16f_bc

/***************************************************************/

void  mlib_c_ImageAffine_s16_1ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos;
    mlib_f32  *fptr;
    mlib_s32  s0, s1, s2, s3;
    mlib_s32  s4, s5, s6, s7;

    CLIP(1)

    filterpos = (X >> FILTER_SHIFT) & FILTER_MASK;
    fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

    xf0 = fptr[0];
    xf1 = fptr[1];
    xf2 = fptr[2];
    xf3 = fptr[3];

    filterpos = (Y >> FILTER_SHIFT) & FILTER_MASK;
    fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

    yf0 = fptr[0];
    yf1 = fptr[1];
    yf2 = fptr[2];
    yf3 = fptr[3];

    xSrc = (X>>MLIB_SHIFT)-1;
    ySrc = (Y>>MLIB_SHIFT)-1;

    srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + xSrc;
    s0 = srcPixelPtr[0]; s1 = srcPixelPtr[1];
    s2 = srcPixelPtr[2]; s3 = srcPixelPtr[3];

    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    s4 = srcPixelPtr[0]; s5 = srcPixelPtr[1];
    s6 = srcPixelPtr[2]; s7 = srcPixelPtr[3];

    for (; dstPixelPtr <= (dstLineEnd - 1); dstPixelPtr++)
    {

      X += dX;
      Y += dY;

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
            srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
            srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);

      filterpos = (X >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

      filterpos = (Y >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      dstPixelPtr[0] = ((int)val0) >> 16;

      xSrc = (X>>MLIB_SHIFT)-1;
      ySrc = (Y>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + xSrc;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[1];
      s2 = srcPixelPtr[2]; s3 = srcPixelPtr[3];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[1];
      s6 = srcPixelPtr[2]; s7 = srcPixelPtr[3];
    }

    c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
    c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
          srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);
    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
          srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);

    val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
    dstPixelPtr[0] = ((int)val0) >> 16;
  }
}

/***************************************************************/
void  mlib_c_ImageAffine_s16_2ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos, k;
    mlib_f32  *fptr;
    mlib_s32  s0, s1, s2, s3;
    mlib_s32  s4, s5, s6, s7;

    CLIP(2)

    for (k = 0; k < 2; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 2*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[2];
      s2 = srcPixelPtr[4]; s3 = srcPixelPtr[6];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[2];
      s6 = srcPixelPtr[4]; s7 = srcPixelPtr[6];

      for (; dPtr <= (dstLineEnd - 1); dPtr+=2)
      {

        X1 += dX;
        Y1 += dY;

        c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
        c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
              srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
              srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);

        filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        xf0 = fptr[0];
        xf1 = fptr[1];
        xf2 = fptr[2];
        xf3 = fptr[3];

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        yf0 = fptr[0];
        yf1 = fptr[1];
        yf2 = fptr[2];
        yf3 = fptr[3];

        dPtr[0] = ((int)val0) >> 16;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 2*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[2];
        s2 = srcPixelPtr[4]; s3 = srcPixelPtr[6];

        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        s4 = srcPixelPtr[0]; s5 = srcPixelPtr[2];
        s6 = srcPixelPtr[4]; s7 = srcPixelPtr[6];
      }

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
            srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
            srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
      dPtr[0] = ((int)val0) >> 16;
    }
  }
}

/***************************************************************/

void  mlib_c_ImageAffine_s16_3ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos, k;
    mlib_f32  *fptr;
    mlib_s32  s0, s1, s2, s3;
    mlib_s32  s4, s5, s6, s7;

    CLIP(3)

    for (k = 0; k < 3; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 3*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[3];
      s2 = srcPixelPtr[6]; s3 = srcPixelPtr[9];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[3];
      s6 = srcPixelPtr[6]; s7 = srcPixelPtr[9];

      for (; dPtr <= (dstLineEnd - 1); dPtr+=3)
      {

        X1 += dX;
        Y1 += dY;

        c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
        c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
              srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
              srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);

        filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        xf0 = fptr[0];
        xf1 = fptr[1];
        xf2 = fptr[2];
        xf3 = fptr[3];

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        yf0 = fptr[0];
        yf1 = fptr[1];
        yf2 = fptr[2];
        yf3 = fptr[3];

        dPtr[0] = ((int)val0) >> 16;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 3*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[3];
        s2 = srcPixelPtr[6]; s3 = srcPixelPtr[9];

        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        s4 = srcPixelPtr[0]; s5 = srcPixelPtr[3];
        s6 = srcPixelPtr[6]; s7 = srcPixelPtr[9];
      }

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
            srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
            srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
      dPtr[0] = ((int)val0) >> 16;
    }
  }
}

/***************************************************************/

void  mlib_c_ImageAffine_s16_4ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_s32  filterpos, k;
    mlib_f32  *fptr;
    mlib_s32  s0, s1, s2, s3;
    mlib_s32  s4, s5, s6, s7;

    CLIP(4)

    for (k = 0; k < 4; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      xf0 = fptr[0];
      xf1 = fptr[1];
      xf2 = fptr[2];
      xf3 = fptr[3];

      filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
      fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

      yf0 = fptr[0];
      yf1 = fptr[1];
      yf2 = fptr[2];
      yf3 = fptr[3];

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 4*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[4];
      s2 = srcPixelPtr[8]; s3 = srcPixelPtr[12];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[4];
      s6 = srcPixelPtr[8]; s7 = srcPixelPtr[12];

      for (; dPtr <= (dstLineEnd - 1); dPtr+=4)
      {

        X1 += dX;
        Y1 += dY;

        c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
        c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
              srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
              srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);

        filterpos = (X1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        xf0 = fptr[0];
        xf1 = fptr[1];
        xf2 = fptr[2];
        xf3 = fptr[3];

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        filterpos = (Y1 >> FILTER_SHIFT) & FILTER_MASK;
        fptr = (mlib_f32 *) ((mlib_u8 *)TABLE + filterpos);

        yf0 = fptr[0];
        yf1 = fptr[1];
        yf2 = fptr[2];
        yf3 = fptr[3];

        dPtr[0] = ((int)val0) >> 16;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 4*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[4];
        s2 = srcPixelPtr[8]; s3 = srcPixelPtr[12];

        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        s4 = srcPixelPtr[0]; s5 = srcPixelPtr[4];
        s6 = srcPixelPtr[8]; s7 = srcPixelPtr[12];
      }

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
            srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
            srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
      dPtr[0] = ((int)val0) >> 16;
    }
  }
}

/***************************************************************/

#undef  MLIB_TYPE
#define MLIB_TYPE mlib_s32

/***************************************************************/

void  mlib_c_ImageAffine_s32_1ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  dx, dx_2, dx2, dx3_2, dx3_3;
    mlib_d64  dy, dy_2, dy2, dy3_2, dy3_3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_d64  scale = 1. / 65536.;
    mlib_s32  s0, s1, s2, s3;
    mlib_s32  s4, s5, s6, s7;

    CLIP(1)

    dx =  (X & MLIB_MASK) * scale;  dy = (Y & MLIB_MASK) * scale;
    dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
    dx2 = dx * dx;      dy2 = dy * dy;
    dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
    dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

    xf0 = dx2 - dx3_2 - dx_2;
    xf1 = dx3_3 - 2.5 * dx2 + 1.0;
    xf2 = 2.0 * dx2 - dx3_3 + dx_2;
    xf3 = dx3_2 - 0.5 * dx2;

    yf0 = dy2 - dy3_2 - dy_2;
    yf1 = dy3_3 - 2.5 * dy2 + 1.0;
    yf2 = 2.0 * dy2 - dy3_3 + dy_2;
    yf3 = dy3_2 - 0.5 * dy2;

    xSrc = (X>>MLIB_SHIFT)-1;
    ySrc = (Y>>MLIB_SHIFT)-1;

    srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + xSrc;
    s0 = srcPixelPtr[0]; s1 = srcPixelPtr[1];
    s2 = srcPixelPtr[2]; s3 = srcPixelPtr[3];

    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    s4 = srcPixelPtr[0]; s5 = srcPixelPtr[1];
    s6 = srcPixelPtr[2]; s7 = srcPixelPtr[3];

    for (; dstPixelPtr <= (dstLineEnd - 1); dstPixelPtr++)
    {

      X += dX;
      Y += dY;

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
            srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
            srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);

      dx =  (X & MLIB_MASK) * scale;  dy = (Y & MLIB_MASK) * scale;
      dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
      dx2 = dx * dx;      dy2 = dy * dy;
      dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
      dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

      xf0 = dx2 - dx3_2 - dx_2;
      xf1 = dx3_3 - 2.5 * dx2 + 1.0;
      xf2 = 2.0 * dx2 - dx3_3 + dx_2;
      xf3 = dx3_2 - 0.5 * dx2;

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

      yf0 = dy2 - dy3_2 - dy_2;
      yf1 = dy3_3 - 2.5 * dy2 + 1.0;
      yf2 = 2.0 * dy2 - dy3_3 + dy_2;
      yf3 = dy3_2 - 0.5 * dy2;

      dstPixelPtr[0] = (mlib_s32)val0;

      xSrc = (X>>MLIB_SHIFT)-1;
      ySrc = (Y>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + xSrc;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[1];
      s2 = srcPixelPtr[2]; s3 = srcPixelPtr[3];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[1];
      s6 = srcPixelPtr[2]; s7 = srcPixelPtr[3];
    }

    c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
    c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
          srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);
    srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
    c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[1]*xf1 +
          srcPixelPtr[2]*xf2 + srcPixelPtr[3]*xf3);

    val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
    dstPixelPtr[0] = (mlib_s32)val0;
  }
}

/***************************************************************/
void  mlib_c_ImageAffine_s32_2ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_d64  dx, dx_2, dx2, dx3_2, dx3_3;
    mlib_d64  dy, dy_2, dy2, dy3_2, dy3_3;
    mlib_d64  scale = 1. / 65536.;
    mlib_s32  s0, s1, s2, s3, k;
    mlib_s32  s4, s5, s6, s7;

    CLIP(2)

    for (k = 0; k < 2; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      dx =  (X1 & MLIB_MASK) * scale;  dy = (Y1 & MLIB_MASK) * scale;
      dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
      dx2 = dx * dx;      dy2 = dy * dy;
      dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
      dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

      xf0 = dx2 - dx3_2 - dx_2;
      xf1 = dx3_3 - 2.5 * dx2 + 1.0;
      xf2 = 2.0 * dx2 - dx3_3 + dx_2;
      xf3 = dx3_2 - 0.5 * dx2;

      yf0 = dy2 - dy3_2 - dy_2;
      yf1 = dy3_3 - 2.5 * dy2 + 1.0;
      yf2 = 2.0 * dy2 - dy3_3 + dy_2;
      yf3 = dy3_2 - 0.5 * dy2;

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 2*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[2];
      s2 = srcPixelPtr[4]; s3 = srcPixelPtr[6];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[2];
      s6 = srcPixelPtr[4]; s7 = srcPixelPtr[6];

      for (; dPtr <= (dstLineEnd - 1); dPtr+=2)
      {

        X1 += dX;
        Y1 += dY;

        c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
        c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
              srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
              srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);

        dx =  (X1 & MLIB_MASK) * scale;  dy = (Y1 & MLIB_MASK) * scale;
        dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
        dx2 = dx * dx;      dy2 = dy * dy;
        dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
        dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

        xf0 = dx2 - dx3_2 - dx_2;
        xf1 = dx3_3 - 2.5 * dx2 + 1.0;
        xf2 = 2.0 * dx2 - dx3_3 + dx_2;
        xf3 = dx3_2 - 0.5 * dx2;

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        yf0 = dy2 - dy3_2 - dy_2;
        yf1 = dy3_3 - 2.5 * dy2 + 1.0;
        yf2 = 2.0 * dy2 - dy3_3 + dy_2;
        yf3 = dy3_2 - 0.5 * dy2;

        dPtr[0] = (mlib_s32)val0;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 2*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[2];
        s2 = srcPixelPtr[4]; s3 = srcPixelPtr[6];

        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        s4 = srcPixelPtr[0]; s5 = srcPixelPtr[2];
        s6 = srcPixelPtr[4]; s7 = srcPixelPtr[6];
      }

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
            srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[2]*xf1 +
            srcPixelPtr[4]*xf2 + srcPixelPtr[6]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
      dPtr[0] = (mlib_s32)val0;
    }
  }
}

/***************************************************************/

void  mlib_c_ImageAffine_s32_3ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_d64  dx, dx_2, dx2, dx3_2, dx3_3;
    mlib_d64  dy, dy_2, dy2, dy3_2, dy3_3;
    mlib_d64  scale = 1. / 65536.;
    mlib_s32  s0, s1, s2, s3, k;
    mlib_s32  s4, s5, s6, s7;

    CLIP(3)

    for (k = 0; k < 3; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      dx =  (X1 & MLIB_MASK) * scale;  dy = (Y1 & MLIB_MASK) * scale;
      dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
      dx2 = dx * dx;      dy2 = dy * dy;
      dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
      dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

      xf0 = dx2 - dx3_2 - dx_2;
      xf1 = dx3_3 - 2.5 * dx2 + 1.0;
      xf2 = 2.0 * dx2 - dx3_3 + dx_2;
      xf3 = dx3_2 - 0.5 * dx2;

      yf0 = dy2 - dy3_2 - dy_2;
      yf1 = dy3_3 - 2.5 * dy2 + 1.0;
      yf2 = 2.0 * dy2 - dy3_3 + dy_2;
      yf3 = dy3_2 - 0.5 * dy2;

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 3*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[3];
      s2 = srcPixelPtr[6]; s3 = srcPixelPtr[9];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[3];
      s6 = srcPixelPtr[6]; s7 = srcPixelPtr[9];

      for (; dPtr <= (dstLineEnd - 1); dPtr+=3)
      {

        X1 += dX;
        Y1 += dY;

        c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
        c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
              srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
              srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);

        dx =  (X1 & MLIB_MASK) * scale;  dy = (Y1 & MLIB_MASK) * scale;
        dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
        dx2 = dx * dx;      dy2 = dy * dy;
        dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
        dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

        xf0 = dx2 - dx3_2 - dx_2;
        xf1 = dx3_3 - 2.5 * dx2 + 1.0;
        xf2 = 2.0 * dx2 - dx3_3 + dx_2;
        xf3 = dx3_2 - 0.5 * dx2;

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        yf0 = dy2 - dy3_2 - dy_2;
        yf1 = dy3_3 - 2.5 * dy2 + 1.0;
        yf2 = 2.0 * dy2 - dy3_3 + dy_2;
        yf3 = dy3_2 - 0.5 * dy2;

        dPtr[0] = (mlib_s32)val0;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 3*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[3];
        s2 = srcPixelPtr[6]; s3 = srcPixelPtr[9];

        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        s4 = srcPixelPtr[0]; s5 = srcPixelPtr[3];
        s6 = srcPixelPtr[6]; s7 = srcPixelPtr[9];
      }

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
            srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[3]*xf1 +
            srcPixelPtr[6]*xf2 + srcPixelPtr[9]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
      dPtr[0] = (mlib_s32)val0;
    }
  }
}

/***************************************************************/

void  mlib_c_ImageAffine_s32_4ch_bc (MLIB_AFFINEDEFPARAM2)
{
  DECLAREVAR

  for (j = yStart; j <= yFinish; j++)
  {
    mlib_d64  xf0, xf1, xf2, xf3;
    mlib_d64  yf0, yf1, yf2, yf3;
    mlib_d64  c0, c1, c2, c3, val0;
    mlib_d64  dx, dx_2, dx2, dx3_2, dx3_3;
    mlib_d64  dy, dy_2, dy2, dy3_2, dy3_3;
    mlib_d64  scale = 1. / 65536.;
    mlib_s32  s0, s1, s2, s3, k;
    mlib_s32  s4, s5, s6, s7;

    CLIP(4)

    for (k = 0; k < 4; k++) {
      mlib_s32 X1 = X;
      mlib_s32 Y1 = Y;
      MLIB_TYPE *dPtr = dstPixelPtr + k;

      dx =  (X1 & MLIB_MASK) * scale;  dy = (Y1 & MLIB_MASK) * scale;
      dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
      dx2 = dx * dx;      dy2 = dy * dy;
      dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
      dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

      xf0 = dx2 - dx3_2 - dx_2;
      xf1 = dx3_3 - 2.5 * dx2 + 1.0;
      xf2 = 2.0 * dx2 - dx3_3 + dx_2;
      xf3 = dx3_2 - 0.5 * dx2;

      yf0 = dy2 - dy3_2 - dy_2;
      yf1 = dy3_3 - 2.5 * dy2 + 1.0;
      yf2 = 2.0 * dy2 - dy3_3 + dy_2;
      yf3 = dy3_2 - 0.5 * dy2;

      xSrc = (X1>>MLIB_SHIFT)-1;
      ySrc = (Y1>>MLIB_SHIFT)-1;

      srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 4*xSrc + k;
      s0 = srcPixelPtr[0]; s1 = srcPixelPtr[4];
      s2 = srcPixelPtr[8]; s3 = srcPixelPtr[12];

      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      s4 = srcPixelPtr[0]; s5 = srcPixelPtr[4];
      s6 = srcPixelPtr[8]; s7 = srcPixelPtr[12];

      for (; dPtr <= (dstLineEnd - 1); dPtr+=4)
      {

        X1 += dX;
        Y1 += dY;

        c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
        c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
              srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);
        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
              srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);

        dx =  (X1 & MLIB_MASK) * scale;  dy = (Y1 & MLIB_MASK) * scale;
        dx_2 = 0.5 * dx;    dy_2 = 0.5 * dy;
        dx2 = dx * dx;      dy2 = dy * dy;
        dx3_2 = dx_2 * dx2; dy3_2 = dy_2 * dy2;
        dx3_3 = 3.0 * dx3_2;  dy3_3 = 3.0 * dy3_2;

        xf0 = dx2 - dx3_2 - dx_2;
        xf1 = dx3_3 - 2.5 * dx2 + 1.0;
        xf2 = 2.0 * dx2 - dx3_3 + dx_2;
        xf3 = dx3_2 - 0.5 * dx2;

        val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);

        yf0 = dy2 - dy3_2 - dy_2;
        yf1 = dy3_3 - 2.5 * dy2 + 1.0;
        yf2 = 2.0 * dy2 - dy3_3 + dy_2;
        yf3 = dy3_2 - 0.5 * dy2;

        dPtr[0] = (mlib_s32)val0;

        xSrc = (X1>>MLIB_SHIFT)-1;
        ySrc = (Y1>>MLIB_SHIFT)-1;

        srcPixelPtr = ((MLIB_TYPE **)lineAddr)[ySrc] + 4*xSrc + k;
        s0 = srcPixelPtr[0]; s1 = srcPixelPtr[4];
        s2 = srcPixelPtr[8]; s3 = srcPixelPtr[12];

        srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
        s4 = srcPixelPtr[0]; s5 = srcPixelPtr[4];
        s6 = srcPixelPtr[8]; s7 = srcPixelPtr[12];
      }

      c0 = (s0*xf0 + s1*xf1 + s2*xf2 + s3*xf3);
      c1 = (s4*xf0 + s5*xf1 + s6*xf2 + s7*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c2 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
            srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);
      srcPixelPtr = (MLIB_TYPE*)((mlib_addr)srcPixelPtr + srcYStride);
      c3 = (srcPixelPtr[0]*xf0 + srcPixelPtr[4]*xf1 +
            srcPixelPtr[8]*xf2 + srcPixelPtr[12]*xf3);

      val0 = (c0*yf0 + c1*yf1 + c2*yf2 + c3*yf3);
      dPtr[0] = (mlib_s32)val0;
    }
  }
}


