/*
 * @(#)mlib_c_ImageLookUp.c	1.7 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifdef _MSC_VER
#pragma warning(disable:4068 4244 4101)
#endif /*_MSC_VER*/


#pragma ident	"@(#)mlib_c_ImageLookUp.c	1.5	98/03/18 SMI"

/*
 * $RCSfile: mlib_c_ImageLookUp.c,v $
 * $Revision: 4.2 $
 * $Date: 1997/02/18 16:31:09 $
 * $Author: liang $
 */

/*
 * FUNCTION
 *      mlib_ImageLookUp - table lookup
 *
 * SYNOPSIS
 *      mlib_status mlib_ImageLookUp(mlib_image *dst,
 *                                   mlib_image *src,
 *                                   void **table);
 *
 * ARGUMENT
 *      dst     printer to output image
 *      src     printer to input image
 *      table   lookup table
 *
 * DESCRIPTION
 *      dst = table[src]
 */




#include <stdlib.h>
#include "mlib_image.h"

/***************************************************************/
#define MLIB_C_IMAGELOOKUP_U8(STYPE, TABLE)                             \
{                                                                       \
  mlib_s32 i, j, k;                                                     \
                                                                        \
  if (xsize < 2) {                                                      \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb){                 \
      for(k = 0; k < csize; k++) {                                      \
        mlib_u8 *da = dst + k;                                          \
        STYPE   *sa = src + k;                                          \
        mlib_u8 *tab = (mlib_u8*) TABLE[k];                             \
                                                                        \
        for(i = 0; i < xsize; i++, da += csize, sa += csize)            \
        *da=tab[*sa];                                                   \
      }                                                                 \
    }                                                                   \
  } else {                                                              \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb) {                \
      for(k = 0; k < csize; k++) {                                      \
        mlib_u8 *da = dst + k;                                          \
        STYPE   *sa = src + k;                                          \
        mlib_u8 *tab = (mlib_u8*) TABLE[k];                             \
        mlib_s32 s0, t0, s1, t1;                                        \
                                                                        \
        s0 = sa[0];                                                     \
        s1 = sa[csize];                                                 \
        sa += 2*csize;                                                  \
                                                                        \
        for(i = 0; i < xsize - 3; i+=2, da += 2*csize, sa += 2*csize) { \
          t0 = tab[s0];                                                 \
          t1 = tab[s1];                                                 \
          s0 = sa[0];                                                   \
          s1 = sa[csize];                                               \
          da[0] = t0;                                                   \
          da[csize] = t1;                                               \
        }                                                               \
        t0 = tab[s0];                                                   \
        t1 = tab[s1];                                                   \
        da[0] = t0;                                                     \
        da[csize] = t1;                                                 \
        if (xsize & 1) da[2*csize] = tab[sa[0]];                        \
      }                                                                 \
    }                                                                   \
  }                                                                     \
}

/***************************************************************/
#define MLIB_C_IMAGELOOKUP_S16(STYPE, TABLE)                            \
{                                                                       \
  mlib_s32 i, j, k;                                                     \
                                                                        \
  if (xsize < 2) {                                                      \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb){                 \
      for(k = 0; k < csize; k++) {                                      \
        mlib_s16 *da = dst + k;                                         \
        STYPE    *sa = src + k;                                         \
        mlib_s16 *tab = (mlib_s16*) TABLE[k];                           \
                                                                        \
        for(i = 0; i < xsize; i++, da += csize, sa += csize)            \
        *da=tab[*sa];                                                   \
      }                                                                 \
    }                                                                   \
  } else {                                                              \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb) {                \
      for(k = 0; k < csize; k++) {                                      \
        mlib_s16 *da = dst + k;                                         \
        STYPE    *sa = src + k;                                         \
        mlib_s16 *tab = (mlib_s16*) TABLE[k];                           \
        mlib_s32 s0, t0, s1, t1;                                        \
                                                                        \
        s0 = sa[0] << 1;                                                \
        s1 = sa[csize] << 1;                                            \
        sa += 2*csize;                                                  \
                                                                        \
        for(i = 0; i < xsize - 3; i+=2, da += 2*csize, sa += 2*csize) { \
          t0 = *(mlib_s16*)((mlib_addr)tab + s0);                       \
          t1 = *(mlib_s16*)((mlib_addr)tab + s1);                       \
          s0 = sa[0] << 1;                                              \
          s1 = sa[csize] << 1;                                          \
          da[0] = t0;                                                   \
          da[csize] = t1;                                               \
        }                                                               \
        t0 = *(mlib_s16*)((mlib_addr)tab + s0);                         \
        t1 = *(mlib_s16*)((mlib_addr)tab + s1);                         \
        da[0] = t0;                                                     \
        da[csize] = t1;                                                 \
        if (xsize & 1) da[2*csize] = tab[sa[0]];                        \
      }                                                                 \
    }                                                                   \
  }                                                                     \
}

/***************************************************************/
#define MLIB_C_IMAGELOOKUPSI_U8(STYPE, TABLE)                           \
{                                                                       \
  mlib_s32 i, j, k;                                                     \
                                                                        \
  if (xsize < 2) {                                                      \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb){                 \
      for(k = 0; k < csize; k++) {                                      \
        mlib_u8 *da = dst + k;                                          \
        STYPE   *sa = src;                                              \
        mlib_u8 *tab = (mlib_u8*) TABLE[k];                             \
                                                                        \
        for(i = 0; i < xsize; i++, da += csize, sa ++)                  \
        *da=tab[*sa];                                                   \
      }                                                                 \
    }                                                                   \
  } else {                                                              \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb) {                \
      for(k = 0; k < csize; k++) {                                      \
        mlib_u8 *da = dst + k;                                          \
        STYPE   *sa = src;                                              \
        mlib_u8 *tab = (mlib_u8*) TABLE[k];                             \
        mlib_s32 s0, t0, s1, t1;                                        \
                                                                        \
        s0 = sa[0];                                                     \
        s1 = sa[1];                                                     \
        sa += 2;                                                        \
                                                                        \
        for(i = 0; i < xsize - 3; i+=2, da += 2*csize, sa += 2) {       \
          t0 = tab[s0];                                                 \
          t1 = tab[s1];                                                 \
          s0 = sa[0];                                                   \
          s1 = sa[1];                                                   \
          da[0] = t0;                                                   \
          da[csize] = t1;                                               \
        }                                                               \
        t0 = tab[s0];                                                   \
        t1 = tab[s1];                                                   \
        da[0] = t0;                                                     \
        da[csize] = t1;                                                 \
        if (xsize & 1) da[2*csize] = tab[sa[0]];                        \
      }                                                                 \
    }                                                                   \
  }                                                                     \
}

/***************************************************************/
#define MLIB_C_IMAGELOOKUPSI_S16(STYPE, TABLE)                          \
{                                                                       \
  mlib_s32 i, j, k;                                                     \
                                                                        \
  if (xsize < 2) {                                                      \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb){                 \
      for(k = 0; k < csize; k++) {                                      \
        mlib_s16 *da = dst + k;                                         \
        STYPE    *sa = src;                                             \
        mlib_s16 *tab = (mlib_s16*) TABLE[k];                           \
                                                                        \
        for(i = 0; i < xsize; i++, da += csize, sa ++)                  \
        *da=tab[*sa];                                                   \
      }                                                                 \
    }                                                                   \
  } else {                                                              \
    for(j = 0; j < ysize; j++, dst += dlb, src += slb) {                \
      for(k = 0; k < csize; k++) {                                      \
        mlib_s16 *da = dst + k;                                         \
        STYPE    *sa = src;                                             \
        mlib_s16 *tab = (mlib_s16*) TABLE[k];                           \
        mlib_s32 s0, t0, s1, t1;                                        \
                                                                        \
        s0 = sa[0] << 1;                                                \
        s1 = sa[1] << 1;                                                \
        sa += 2;                                                        \
                                                                        \
        for(i = 0; i < xsize - 3; i+=2, da += 2*csize, sa += 2) {       \
          t0 = *(mlib_s16*)((mlib_addr)tab + s0);                       \
          t1 = *(mlib_s16*)((mlib_addr)tab + s1);                       \
          s0 = sa[0] << 1;                                              \
          s1 = sa[1] << 1;                                              \
          da[0] = t0;                                                   \
          da[csize] = t1;                                               \
        }                                                               \
        t0 = *(mlib_s16*)((mlib_addr)tab + s0);                         \
        t1 = *(mlib_s16*)((mlib_addr)tab + s1);                         \
        da[0] = t0;                                                     \
        da[csize] = t1;                                                 \
        if (xsize & 1) da[2*csize] = tab[sa[0]];                        \
      }                                                                 \
    }                                                                   \
  }                                                                     \
}

/***************************************************************/

void mlib_c_ImageLookUp_U8_U8(mlib_u8  *src, mlib_s32 slb,
                              mlib_u8  *dst, mlib_s32 dlb,
                              mlib_s32 xsize, mlib_s32 ysize,
                              mlib_s32 csize,
                              mlib_u8  **table)
{
  MLIB_C_IMAGELOOKUP_U8(mlib_u8, table)

}

/***************************************************************/

void mlib_c_ImageLookUp_S16_U8(mlib_s16 *src, mlib_s32 slb,
                               mlib_u8  *dst, mlib_s32 dlb,
                               mlib_s32 xsize, mlib_s32 ysize,
                               mlib_s32 csize,
                               mlib_u8  **table)
{
  mlib_u8 *table_base[4];
  int c;

  for (c = 0; c < csize; c++) {
    table_base[c] = &table[c][32768];
  }

  MLIB_C_IMAGELOOKUP_U8(mlib_s16, table_base)
}

/***************************************************************/

void mlib_c_ImageLookUp_U8_S16(mlib_u8  *src, mlib_s32 slb,
                               mlib_s16 *dst, mlib_s32 dlb,
                               mlib_s32 xsize, mlib_s32 ysize,
                               mlib_s32 csize,
                               mlib_s16 **table)
{
  MLIB_C_IMAGELOOKUP_S16(mlib_u8, table)
}

/***************************************************************/

void mlib_c_ImageLookUp_S16_S16(mlib_s16 *src, mlib_s32 slb,
                                mlib_s16 *dst, mlib_s32 dlb,
                                mlib_s32 xsize, mlib_s32 ysize,
                                mlib_s32 csize,
                                mlib_s16 **table)
{
  mlib_s16 *table_base[4];
  int c;

  for (c = 0; c < csize; c++) {
    table_base[c] = &table[c][32768];
  }

  MLIB_C_IMAGELOOKUP_S16(mlib_s16, table_base)
}

/***************************************************************/

void mlib_c_ImageLookUpSI_U8_U8(mlib_u8  *src, mlib_s32 slb,
                                mlib_u8  *dst, mlib_s32 dlb,
                                mlib_s32 xsize, mlib_s32 ysize,
                                mlib_s32 csize,
                                mlib_u8  **table)
{
  MLIB_C_IMAGELOOKUPSI_U8(mlib_u8, table)
}

/***************************************************************/

void mlib_c_ImageLookUpSI_S16_U8(mlib_s16 *src, mlib_s32 slb,
                                 mlib_u8  *dst, mlib_s32 dlb,
                                 mlib_s32 xsize, mlib_s32 ysize,
                                 mlib_s32 csize,
                                 mlib_u8  **table)
{
  mlib_u8 *table_base[4];
  int c;

  for (c = 0; c < csize; c++) {
    table_base[c] = &table[c][32768];
  }

  MLIB_C_IMAGELOOKUPSI_U8(mlib_s16, table_base)

}

/***************************************************************/

void mlib_c_ImageLookUpSI_U8_S16(mlib_u8  *src, mlib_s32 slb,
                                 mlib_s16 *dst, mlib_s32 dlb,
                                 mlib_s32 xsize, mlib_s32 ysize,
                                 mlib_s32 csize,
                                 mlib_s16 **table)
{
  MLIB_C_IMAGELOOKUPSI_S16(mlib_u8, table)
}

/***************************************************************/

void mlib_c_ImageLookUpSI_S16_S16(mlib_s16 *src, mlib_s32 slb,
                                  mlib_s16 *dst, mlib_s32 dlb,
                                  mlib_s32 xsize, mlib_s32 ysize,
                                  mlib_s32 csize,
                                  mlib_s16 **table)
{
  mlib_s16 *table_base[4];
  int c;

  for (c = 0; c < csize; c++) {
    table_base[c] = &table[c][32768];
  }

  MLIB_C_IMAGELOOKUPSI_S16(mlib_s16, table_base)
}

/***************************************************************/

#ifdef MLIB_TEST

mlib_status mlib_c_ImageLookUp(mlib_image *dst,
                               mlib_image *src,
                               void **table)
#else

mlib_status mlib_ImageLookUp(mlib_image *dst,
                             mlib_image *src,
                             void **table)
#endif

{
  mlib_s32   slb, dlb, xsize, ysize, nchan, ichan;
  mlib_type  stype, dtype;
  void       *sa, *da;

#ifdef MLIB_DEBUG
  if (mlib_ImageGetType(src) == MLIB_BIT) {
    return MLIB_FAILURE;
  }

  if (mlib_ImageGetType(dst) == MLIB_BIT) {
    return MLIB_FAILURE;
  }

  if (mlib_ImageGetChannels(src) != 1) {
    if (mlib_ImageGetChannels(src) != mlib_ImageGetChannels(dst)) {
      return MLIB_FAILURE;
    }
  }

  if (mlib_ImageGetWidth(src) != mlib_ImageGetWidth(dst)) {
    return MLIB_FAILURE;
  }

  if (mlib_ImageGetHeight(src) != mlib_ImageGetHeight(dst)) {
    return MLIB_FAILURE;
  }
#endif

  stype = mlib_ImageGetType(src);
  dtype = mlib_ImageGetType(dst);
  ichan = mlib_ImageGetChannels(src);
  nchan = mlib_ImageGetChannels(dst);
  xsize = mlib_ImageGetWidth(src);
  ysize = mlib_ImageGetHeight(src);
  slb   = mlib_ImageGetStride(src);
  dlb   = mlib_ImageGetStride(dst);
  sa    = mlib_ImageGetData(src);
  da    = mlib_ImageGetData(dst);

  if (ichan == nchan) {

    if (dtype == MLIB_BYTE) {

      if (stype == MLIB_BYTE) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUp_U8_U8()\n");
#endif
        mlib_c_ImageLookUp_U8_U8(sa, slb,
                                 da, dlb,
                                 xsize, ysize, nchan,
                                 (mlib_u8 **) table);

        return MLIB_SUCCESS;

      } else if (stype == MLIB_SHORT) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUp_S16_U8()\n");
#endif
        mlib_c_ImageLookUp_S16_U8(sa, slb/2,
                                  da, dlb,
                                  xsize, ysize, nchan,
                                  (mlib_u8 **) table);
        return MLIB_SUCCESS;
      }

    } else if (dtype == MLIB_SHORT) {

      if (stype == MLIB_BYTE) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUp_U8_S16()\n");
#endif
        mlib_c_ImageLookUp_U8_S16(sa, slb,
                                  da, dlb/2,
                                  xsize, ysize, nchan,
                                  (mlib_s16 **) table);

        return MLIB_SUCCESS;

      } else if (stype == MLIB_SHORT) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUp_S16_S16()\n");
#endif
        mlib_c_ImageLookUp_S16_S16(sa, slb/2,
                                   da, dlb/2,
                                   xsize, ysize, nchan,
                                   (mlib_s16 **) table);
        return MLIB_SUCCESS;
      }
    }

  } else if (ichan == 1) {

    if (dtype == MLIB_BYTE) {

      if (stype == MLIB_BYTE) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUpSI_U8_U8()\n");
#endif
        mlib_c_ImageLookUpSI_U8_U8(sa, slb,
                                   da, dlb,
                                   xsize, ysize, nchan,
                                   (mlib_u8 **) table);

        return MLIB_SUCCESS;

      } else if (stype == MLIB_SHORT) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUpSI_S16_U8()\n");
#endif
        mlib_c_ImageLookUpSI_S16_U8(sa, slb/2,
                                    da, dlb,
                                    xsize, ysize, nchan,
                                    (mlib_u8 **) table);
        return MLIB_SUCCESS;
      }

    } else if (dtype == MLIB_SHORT) {

      if (stype == MLIB_BYTE) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUpSI_U8_S16()\n");
#endif
        mlib_c_ImageLookUpSI_U8_S16(sa, slb,
                                    da, dlb/2,
                                    xsize, ysize, nchan,
                                    (mlib_s16 **) table);

        return MLIB_SUCCESS;

      } else if (stype == MLIB_SHORT) {

#ifdef MLIB_DIAG
        printf("DIAG    mlib_c_ImageLookUpSI_S16_S16()\n");
#endif
        mlib_c_ImageLookUpSI_S16_S16(sa, slb/2,
                                     da, dlb/2,
                                     xsize, ysize, nchan,
                                     (mlib_s16 **) table);
        return MLIB_SUCCESS;
      }
    }
  }

  return MLIB_FAILURE;
}

/***************************************************************/
