/*
 * @(#)mlib_image_proto.h	1.11 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef MLIB_IMAGE_PROTO_H
#define MLIB_IMAGE_PROTO_H

#ifdef __SUNPRO_C
#pragma ident    "@(#)mlib_image_proto.h	1.53    98/09/02 SMI"
#endif /* __SUNPRO_C */

#include <mlib_types.h>
#include <mlib_status.h>
#include <mlib_image_types.h>

#ifdef __cplusplus
extern "C" {
#endif

mlib_image *mlib_ImageCreate(mlib_type type,
                             mlib_s32  channels,
                             mlib_s32  width,
                             mlib_s32  height);

mlib_image *mlib_ImageCreateStruct(mlib_type type,
                                   mlib_s32  channels,
                                   mlib_s32  width,
                                   mlib_s32  height,
                                   mlib_s32  stride,
                                   void      *data);

mlib_image *mlib_ImageCreateSubimage(mlib_image *img,
                                     mlib_s32   x,
                                     mlib_s32   y,
                                     mlib_s32   w,
                                     mlib_s32   h);

void mlib_ImageDelete(mlib_image *img);

static mlib_type  mlib_ImageGetType(mlib_image *img);
static mlib_s32   mlib_ImageGetChannels(mlib_image *img);
static mlib_s32   mlib_ImageGetWidth(mlib_image *img);
static mlib_s32   mlib_ImageGetHeight(mlib_image *img);
static mlib_s32   mlib_ImageGetStride(mlib_image *img);
static void      *mlib_ImageGetData(mlib_image *img);
static void      *mlib_ImageGetRowTable(mlib_image *img);
static mlib_s32   mlib_ImageGetFlags(mlib_image *img);

static int mlib_ImageTestFlags(mlib_image *img,
                               mlib_s32   flags);

static int mlib_ImageIsNotAligned64(mlib_image *img);
static int mlib_ImageIsNotAligned8(mlib_image *img);
static int mlib_ImageIsNotAligned4(mlib_image *img);
static int mlib_ImageIsNotAligned2(mlib_image *img);
static int mlib_ImageIsNotWidth8X(mlib_image *img);
static int mlib_ImageIsNotWidth4X(mlib_image *img);
static int mlib_ImageIsNotWidth2X(mlib_image *img);
static int mlib_ImageIsNotHeight8X(mlib_image *img);
static int mlib_ImageIsNotHeight4X(mlib_image *img);
static int mlib_ImageIsNotHeight2X(mlib_image *img);
static int mlib_ImageIsNotStride8X(mlib_image *img);
static int mlib_ImageIsNotOneDvector(mlib_image *img);
static int mlib_ImageIsUserAllocated(mlib_image *img);

void *mlib_ImageCreateRowTable(mlib_image *image);
void mlib_ImageDeleteRowTable(mlib_image *image);

mlib_status mlib_ImageDataTypeConvert(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageChannelExtract(mlib_image *dst,
                                     mlib_image *src,
                                     mlib_s32   cmask);

mlib_status mlib_ImageChannelInsert(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_s32   cmask);

mlib_status mlib_ImageChannelCopy(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32   cmask);

mlib_status mlib_ImageCopy(mlib_image *dst,
                           mlib_image *src);

mlib_status mlib_ImageCopyArea(mlib_image *img,
                               mlib_s32   x,
                               mlib_s32   y,
                               mlib_s32   w,
                               mlib_s32   h,
                               mlib_s32   dx,
                               mlib_s32   dy);

mlib_status mlib_ImageCopySubimage(mlib_image *dst,
                                   mlib_image *src,
                                   mlib_s32   xd,
                                   mlib_s32   yd,
                                   mlib_s32   xs,
                                   mlib_s32   ys,
                                   mlib_s32   w,
                                   mlib_s32   h);

mlib_status mlib_ImageCopyMask(mlib_image *dst,
                               mlib_image *src,
                               mlib_image *mask,
                               mlib_s32   *thresh);

mlib_status mlib_ImageCopyMask_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_image *mask,
                                  mlib_d64   *thresh);

mlib_status mlib_ImageClear(mlib_image *img,
                            mlib_s32   *color);

mlib_status mlib_ImageClear_Fp(mlib_image *img,
                               mlib_d64   *color);

mlib_status mlib_ImageClearEdge(mlib_image *img,
                                mlib_s32   dx,
                                mlib_s32   dy,
                                mlib_s32   *color);

mlib_status mlib_ImageClearEdge_Fp(mlib_image *img,
                                   mlib_s32   dx,
                                   mlib_s32   dy,
                                   mlib_d64   *color);

mlib_status mlib_ImageConv2x2(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *kernel,
                              mlib_s32   scale,
                              mlib_s32   cmask,
                              mlib_edge  edge);

mlib_status mlib_ImageConv3x3(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *kernel,
                              mlib_s32   scale,
                              mlib_s32   cmask,
                              mlib_edge  edge);

mlib_status mlib_ImageConv4x4(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *kernel,
                              mlib_s32   scale,
                              mlib_s32   cmask,
                              mlib_edge  edge);

mlib_status mlib_ImageConv5x5(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *kernel,
                              mlib_s32   scale,
                              mlib_s32   cmask,
                              mlib_edge  edge);

mlib_status mlib_ImageConv7x7(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *kernel,
                              mlib_s32   scale,
                              mlib_s32   cmask,
                              mlib_edge  edge);

mlib_status mlib_ImageConvMxN(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *kernel,
                              mlib_s32   m,
                              mlib_s32   n,
                              mlib_s32   dm,
                              mlib_s32   dn,
                              mlib_s32   scale,
                              mlib_s32   cmask,
                              mlib_edge  edge);

mlib_status mlib_ImageConv3x3Index(mlib_image *dst,
                                   mlib_image *src,
                                   mlib_s16   *kernel,
                                   mlib_s32   scale,
                                   mlib_edge  edge,
                                   void       *colormap);

mlib_status mlib_ImageConv5x5Index(mlib_image *dst,
                                   mlib_image *src,
                                   mlib_s16   *kernel,
                                   mlib_s32   scale,
                                   mlib_edge  edge,
                                   void       *colormap);

mlib_status mlib_ImageConv7x7Index(mlib_image *dst,
                                   mlib_image *src,
                                   mlib_s16   *kernel,
                                   mlib_s32   scale,
                                   mlib_edge  edge,
                                   void       *colormap);

mlib_status mlib_ImageConvMxNIndex(mlib_image *dst,
                                   mlib_image *src,
                                   mlib_s32   m,
                                   mlib_s32   n,
                                   mlib_s16   *kernel,
                                   mlib_s32   scale,
                                   mlib_edge  edge,
                                   void       *colormap);

mlib_status mlib_ImageConv3x3_Fp(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_d64   *kernel,
                                 mlib_s32   cmask,
                                 mlib_edge  edge);

mlib_status mlib_ImageConv5x5_Fp(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_d64   *kernel,
                                 mlib_s32   cmask,
                                 mlib_edge  edge);

mlib_status mlib_ImageConv7x7_Fp(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_d64   *kernel,
                                 mlib_s32   cmask,
                                 mlib_edge  edge);

mlib_status mlib_ImageConvMxN_Fp(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_d64   *kernel,
                                 mlib_s32   m,
                                 mlib_s32   n,
                                 mlib_s32   cmask,
                                 mlib_s32   dm,
                                 mlib_s32   dn,
                                 mlib_edge  edge);

mlib_status mlib_ImageConvKernelConvert(mlib_s32  *ikernel, 
                                        mlib_s32  *iscale,
                                        mlib_d64  *fkernel,
                                        mlib_s32  m,
                                        mlib_s32  n,
                                        mlib_type type);

mlib_status mlib_ImageConvolveMxN(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_d64   *kernel,
                                  mlib_s32   m,
                                  mlib_s32   n,
                                  mlib_s32   dm,
                                  mlib_s32   dn,
                                  mlib_s32   cmask,
                                  mlib_edge  edge);

mlib_status mlib_ImageConvolveMxN_Fp(mlib_image *dst,
                                     mlib_image *src,
                                     mlib_d64   *kernel,
                                     mlib_s32   m,
                                     mlib_s32   n,
                                     mlib_s32   dm,
                                     mlib_s32   dn,
                                     mlib_s32   cmask,
                                     mlib_edge  edge);

mlib_status mlib_ImageSConv3x3(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *hkernel,
                               mlib_s32   *vkernel,
                               mlib_s32   scale,
                               mlib_s32   cmask,
                               mlib_edge  edge);

mlib_status mlib_ImageSConv5x5(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *hkernel,
                               mlib_s32   *vkernel,
                               mlib_s32   scale,
                               mlib_s32   cmask,
                               mlib_edge  edge);

mlib_status mlib_ImageSConv7x7(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *hkernel,
                               mlib_s32   *vkernel,
                               mlib_s32   scale,
                               mlib_s32   cmask,
                               mlib_edge  edge);

mlib_status mlib_ImageSConv3x3_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_d64   *hkernel,
                                  mlib_d64   *vkernel,
                                  mlib_s32   cmask,
                                  mlib_edge  edge);

mlib_status mlib_ImageSConv5x5_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_d64   *hkernel,
                                  mlib_d64   *vkernel,
                                  mlib_s32   cmask,
                                  mlib_edge  edge);

mlib_status mlib_ImageSConv7x7_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_d64   *hkernel,
                                  mlib_d64   *vkernel,
                                  mlib_s32   cmask,
                                  mlib_edge  edge);

mlib_status mlib_ImageSConvKernelConvert(mlib_s32  *ihkernel, 
                                         mlib_s32  *ivkernel, 
                                         mlib_s32  *iscale,
                                         mlib_d64  *fhkernel,
                                         mlib_d64  *fvkernel,
                                         mlib_s32  m,
                                         mlib_s32  n,
                                         mlib_type type);

mlib_status mlib_ImageMaxFilter3x3(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageMaxFilter5x5(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageMaxFilter7x7(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageMaxFilter3x3_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMaxFilter5x5_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMaxFilter7x7_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMinFilter3x3(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageMinFilter5x5(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageMinFilter7x7(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageMinFilter3x3_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMinFilter5x5_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMinFilter7x7_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageRankFilter3x3(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_s32   rank);

mlib_status mlib_ImageRankFilter5x5(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_s32   rank);

mlib_status mlib_ImageRankFilter7x7(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_s32   rank);

mlib_status mlib_ImageRankFilterMxN(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_s32   m,
                                    mlib_s32   n,
                                    mlib_s32   rank);

mlib_status mlib_ImageRankFilter3x3_Fp(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   rank);

mlib_status mlib_ImageRankFilter5x5_Fp(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   rank);

mlib_status mlib_ImageRankFilter7x7_Fp(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   rank);

mlib_status mlib_ImageRankFilterMxN_Fp(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   m,
                                       mlib_s32   n,
                                       mlib_s32   rank);

mlib_status mlib_ImageRankFilter3x3_US(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   rank,
                                       mlib_s32   bits);

mlib_status mlib_ImageRankFilter5x5_US(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   rank,
                                       mlib_s32   bits);

mlib_status mlib_ImageRankFilter7x7_US(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   rank,
                                       mlib_s32   bits);

mlib_status mlib_ImageRankFilterMxN_US(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_s32   m,
                                       mlib_s32   n,
                                       mlib_s32   rank,
                                       mlib_s32   bits);

mlib_status mlib_ImageMedianFilter3x3(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMedianFilter5x5(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMedianFilter7x7(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageMedianFilterMxN(mlib_image *dst,
                                      mlib_image *src,
                                      mlib_s32   m,
                                      mlib_s32   n);

mlib_status mlib_ImageMedianFilter3x3_Fp(mlib_image *dst,
                                         mlib_image *src);

mlib_status mlib_ImageMedianFilter5x5_Fp(mlib_image *dst,
                                         mlib_image *src);

mlib_status mlib_ImageMedianFilter7x7_Fp(mlib_image *dst,
                                         mlib_image *src);

mlib_status mlib_ImageMedianFilterMxN_Fp(mlib_image *dst,
                                         mlib_image *src,
                                         mlib_s32   m,
                                         mlib_s32   n);

mlib_status mlib_ImageMedianFilter3x3_US(mlib_image *dst,
                                         mlib_image *src,
                                         mlib_s32   bits);

mlib_status mlib_ImageMedianFilter5x5_US(mlib_image *dst,
                                         mlib_image *src,
                                         mlib_s32   bits);

mlib_status mlib_ImageMedianFilter7x7_US(mlib_image *dst,
                                         mlib_image *src,
                                         mlib_s32   bits);

mlib_status mlib_ImageMedianFilterMxN_US(mlib_image *dst,
                                         mlib_image *src,
                                         mlib_s32   m,
                                         mlib_s32   n,
                                         mlib_s32   bits);

mlib_status mlib_ImageDilate4(mlib_image *dst,
                              mlib_image *src);

mlib_status mlib_ImageDilate8(mlib_image *dst,
                              mlib_image *src);

mlib_status mlib_ImageDilate4_Fp(mlib_image *dst,
                                 mlib_image *src);

mlib_status mlib_ImageDilate8_Fp(mlib_image *dst,
                                 mlib_image *src);

mlib_status mlib_ImageErode4(mlib_image *dst,
                             mlib_image *src);

mlib_status mlib_ImageErode8(mlib_image *dst,
                             mlib_image *src);

mlib_status mlib_ImageErode4_Fp(mlib_image *dst,
                                mlib_image *src);

mlib_status mlib_ImageErode8_Fp(mlib_image *dst,
                                mlib_image *src);

mlib_status mlib_ImageAbs(mlib_image *dst,
                          mlib_image *src);

mlib_status mlib_ImageAbs_Inp(mlib_image *srcdst);

mlib_status mlib_ImageAbs_Fp(mlib_image *dst,
                             mlib_image *src);

mlib_status mlib_ImageAbs_Fp_Inp(mlib_image *srcdst);

mlib_status mlib_ImageSqrShift(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   shift);

mlib_status mlib_ImageSqrShift_Inp(mlib_image *srcdst,
                                   mlib_s32   shift);

mlib_status mlib_ImageSqr_Fp(mlib_image *dst,
                             mlib_image *src);

mlib_status mlib_ImageSqr_Fp_Inp(mlib_image *srcdst);

mlib_status mlib_ImageInvert(mlib_image *dst,
                             mlib_image *src);

mlib_status mlib_ImageInvert_Inp(mlib_image *srcdst);

mlib_status mlib_ImageInvert_Fp(mlib_image *dst,
                                mlib_image *src);

mlib_status mlib_ImageInvert_Fp_Inp(mlib_image *srcdst);

mlib_status mlib_ImageScale(mlib_image *dst,
                            mlib_image *src,
                            mlib_s16   *alpha,
                            mlib_s16   *beta,
                            mlib_s32   shift);

mlib_status mlib_ImageScale_Inp(mlib_image *srcdst,
                                mlib_s16   *alpha,
                                mlib_s16   *beta,
                                mlib_s32   shift);

mlib_status mlib_ImageScale_Fp(mlib_image *dst,
                               mlib_image *src,
                               mlib_d64   *alpha,
                               mlib_d64   *beta);

mlib_status mlib_ImageScale_Fp_Inp(mlib_image *srcdst,
                                   mlib_d64   *alpha,
                                   mlib_d64   *beta);

mlib_status mlib_ImageConstAdd(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *c);

mlib_status mlib_ImageConstAdd_Inp(mlib_image *srcdst,
                                   mlib_s32   *c);

mlib_status mlib_ImageConstAdd_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_d64   *c);

mlib_status mlib_ImageConstAdd_Fp_Inp(mlib_image *srcdst,
                                      mlib_d64   *c);

mlib_status mlib_ImageConstSub(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *c);

mlib_status mlib_ImageConstSub_Inp(mlib_image *srcdst,
                                   mlib_s32   *c);

mlib_status mlib_ImageConstSub_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_d64   *c);

mlib_status mlib_ImageConstSub_Fp_Inp(mlib_image *srcdst,
                                      mlib_d64   *c);

mlib_status mlib_ImageAve(mlib_image *dst,
                          mlib_image *src1,
                          mlib_image *src2);

mlib_status mlib_ImageAve_Inp(mlib_image *src1dst,
                              mlib_image *src2);

mlib_status mlib_ImageAve_Fp(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

mlib_status mlib_ImageAve_Fp_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

mlib_status mlib_ImageAdd(mlib_image *dst,
                          mlib_image *src1,
                          mlib_image *src2);

mlib_status mlib_ImageAdd_Inp(mlib_image *src1dst,
                              mlib_image *src2);

mlib_status mlib_ImageAdd_Fp(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

mlib_status mlib_ImageAdd_Fp_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

mlib_status mlib_ImageSub(mlib_image *dst,
                          mlib_image *src1,
                          mlib_image *src2);

/* src1dst = src1dst - src2 */
mlib_status mlib_ImageSub1_Inp(mlib_image *src1dst,
                               mlib_image *src2);

/* src2dst = src1 - src2dst */
mlib_status mlib_ImageSub2_Inp(mlib_image *src2dst,
                               mlib_image *src1);

mlib_status mlib_ImageSub_Fp(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

/* src1dst = src1dst - src2 */
mlib_status mlib_ImageSub1_Fp_Inp(mlib_image *src1dst,
                                  mlib_image *src2);

/* src2dst = src1 - src2dst */
mlib_status mlib_ImageSub2_Fp_Inp(mlib_image *src2dst,
                                  mlib_image *src1);

mlib_status mlib_ImageMulShift(mlib_image *dst,
                               mlib_image *src1,
                               mlib_image *src2,
                               mlib_s32   shift);

mlib_status mlib_ImageMulShift_Inp(mlib_image *src1dst,
                                   mlib_image *src2,
                                   mlib_s32   shift);

mlib_status mlib_ImageMul_Fp(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

mlib_status mlib_ImageMul_Fp_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

mlib_status mlib_ImageScalarBlend(mlib_image *dst,
                                  mlib_image *src1,
                                  mlib_image *src2,
                                  mlib_s32   *alpha);

mlib_status mlib_ImageScalarBlend_Inp(mlib_image *src1dst,
                                      mlib_image *src2,
                                      mlib_s32   *alpha);

mlib_status mlib_ImageScalarBlend_Fp(mlib_image *dst,
                                     mlib_image *src1,
                                     mlib_image *src2,
                                     mlib_d64   *alpha);

mlib_status mlib_ImageScalarBlend_Fp_Inp(mlib_image *src1dst,
                                         mlib_image *src2,
                                         mlib_d64   *alpha);

mlib_status mlib_ImageBlend(mlib_image *dst,
                            mlib_image *src1,
                            mlib_image *src2,
                            mlib_image *alpha);

/* src1dst = src1dst * alpha + src2 * (1 - alpha) */
mlib_status mlib_ImageBlend1_Inp(mlib_image *src1dst,
                                 mlib_image *src2,
                                 mlib_image *alpha);

/* src2dst = src1 * alpha + src2dst * (1 - alpha) */
mlib_status mlib_ImageBlend2_Inp(mlib_image *src2dst,
                                 mlib_image *src1,
                                 mlib_image *alpha);

mlib_status mlib_ImageBlend_Fp(mlib_image *dst,
                               mlib_image *src1,
                               mlib_image *src2,
                               mlib_image *alpha);

/* src1dst = src1dst * alpha + src2 * (1 - alpha) */
mlib_status mlib_ImageBlend1_Fp_Inp(mlib_image *src1dst,
                                    mlib_image *src2,
                                    mlib_image *alpha);

/* src2dst = src1 * alpha + src2dst * (1 - alpha) */
mlib_status mlib_ImageBlend2_Fp_Inp(mlib_image *src2dst,
                                    mlib_image *src1,
                                    mlib_image *alpha);

mlib_status mlib_ImageMax(mlib_image *dst,
                          mlib_image *src1,
                          mlib_image *src2);

mlib_status mlib_ImageMax_Inp(mlib_image *src1dst,
                              mlib_image *src2);

mlib_status mlib_ImageMax_Fp(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

mlib_status mlib_ImageMax_Fp_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

mlib_status mlib_ImageMin(mlib_image *dst,
                          mlib_image *src1,
                          mlib_image *src2);

mlib_status mlib_ImageMin_Inp(mlib_image *src1dst,
                              mlib_image *src2);

mlib_status mlib_ImageMin_Fp(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

mlib_status mlib_ImageMin_Fp_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

mlib_status mlib_ImageMulAlpha(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   cmask);

mlib_status mlib_ImageDivAlpha(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   cmask);

mlib_status mlib_ImageMulAlpha_Inp(mlib_image *img,
                                   mlib_s32 cmask);

mlib_status mlib_ImageDivAlpha_Inp(mlib_image *img,
                                   mlib_s32 cmask);

mlib_status mlib_ImageMulAlpha_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32   cmask);

mlib_status mlib_ImageDivAlpha_Fp(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32   cmask);

mlib_status mlib_ImageMulAlpha_Fp_Inp(mlib_image *img,
                                      mlib_s32 cmask);

mlib_status mlib_ImageDivAlpha_Fp_Inp(mlib_image *img,
                                      mlib_s32 cmask);

mlib_status mlib_ImageNot(mlib_image *dst,
                          mlib_image *src);

mlib_status mlib_ImageNot_Inp(mlib_image *srcdst);

mlib_status mlib_ImageAnd(mlib_image *dst,
                          mlib_image *src1,
                          mlib_image *src2);

mlib_status mlib_ImageAnd_Inp(mlib_image *src1dst,
                              mlib_image *src2);

mlib_status mlib_ImageOr(mlib_image *dst,
                         mlib_image *src1,
                         mlib_image *src2);

mlib_status mlib_ImageOr_Inp(mlib_image *src1dst,
                             mlib_image *src2);

mlib_status mlib_ImageXor(mlib_image *dst,
                          mlib_image *src1,
                          mlib_image *src2);

mlib_status mlib_ImageXor_Inp(mlib_image *src1dst,
                              mlib_image *src2);

mlib_status mlib_ImageNotAnd(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

mlib_status mlib_ImageNotAnd_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

mlib_status mlib_ImageNotOr(mlib_image *dst,
                            mlib_image *src1,
                            mlib_image *src2);

mlib_status mlib_ImageNotOr_Inp(mlib_image *src1dst,
                                mlib_image *src2);

mlib_status mlib_ImageNotXor(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

mlib_status mlib_ImageNotXor_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

mlib_status mlib_ImageAndNot(mlib_image *dst,
                             mlib_image *src1,
                             mlib_image *src2);

/* src1dst = src1dst & (~src2) */
mlib_status mlib_ImageAndNot1_Inp(mlib_image *src1dst,
                                  mlib_image *src2);

/* src2dst = src1 & (~src2dst) */
mlib_status mlib_ImageAndNot2_Inp(mlib_image *src2dst,
                                  mlib_image *src1);

mlib_status mlib_ImageOrNot(mlib_image *dst,
                            mlib_image *src1,
                            mlib_image *src2);

/* src1dst = src1dst | (~src2) */
mlib_status mlib_ImageOrNot1_Inp(mlib_image *src1dst,
                                 mlib_image *src2);

/* src2dst = src1 | (~src2dst) */
mlib_status mlib_ImageOrNot2_Inp(mlib_image *src2dst,
                                 mlib_image *src1);

mlib_status mlib_ImageConstAnd(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *c);

mlib_status mlib_ImageConstAnd_Inp(mlib_image *srcdst,
                                   mlib_s32   *c);

mlib_status mlib_ImageConstOr(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *c);

mlib_status mlib_ImageConstOr_Inp(mlib_image *srcdst,
                                  mlib_s32   *c);

mlib_status mlib_ImageConstXor(mlib_image *dst,
                               mlib_image *src,
                               mlib_s32   *c);

mlib_status mlib_ImageConstXor_Inp(mlib_image *srcdst,
                                   mlib_s32   *c);

mlib_status mlib_ImageConstNotAnd(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32   *c);

mlib_status mlib_ImageConstNotAnd_Inp(mlib_image *srcdst,
                                      mlib_s32   *c);

mlib_status mlib_ImageConstNotOr(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_s32   *c);

mlib_status mlib_ImageConstNotOr_Inp(mlib_image *srcdst,
                                     mlib_s32   *c);

mlib_status mlib_ImageConstNotXor(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32   *c);

mlib_status mlib_ImageConstNotXor_Inp(mlib_image *srcdst,
                                      mlib_s32   *c);

mlib_status mlib_ImageConstAndNot(mlib_image *dst,
                                  mlib_image *src,
                                  mlib_s32   *c);

/* srcdst = (~srcdst) & c */
mlib_status mlib_ImageConstAndNot_Inp(mlib_image *srcdst,
                                      mlib_s32   *c);

mlib_status mlib_ImageConstOrNot(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_s32   *c);

/* srcdst = (~srcdst) | c */
mlib_status mlib_ImageConstOrNot_Inp(mlib_image *srcdst,
                                     mlib_s32   *c);

mlib_status mlib_ImageColorConvert1(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_f32   *cmat);

mlib_status mlib_ImageColorConvert2(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_f32   *cmat,
                                    mlib_f32   *offset);

mlib_status mlib_ImageColorConvert1_Fp(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_d64   *cmat);

mlib_status mlib_ImageColorConvert2_Fp(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_d64   *cmat,
                                       mlib_d64   *offset);

mlib_status mlib_ImageColorRGB2XYZ(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorXYZ2RGB(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorRGB2XYZ_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageColorXYZ2RGB_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageColorRGB2YCC(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorYCC2RGB(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorRGB2HSV(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorHSV2RGB(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorRGB2HSV_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageColorHSV2RGB_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageColorRGB2HSL(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorHSL2RGB(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageColorRGB2HSL_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageColorHSL2RGB_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageColorRGB2Mono(mlib_image *dst,
                                    mlib_image *src,
                                    mlib_f32   *weight);

mlib_status mlib_ImageColorRGB2CIEMono(mlib_image *dst,
                                       mlib_image *src);

mlib_status mlib_ImageColorRGB2Mono_Fp(mlib_image *dst,
                                       mlib_image *src,
                                       mlib_d64   *weight);

mlib_status mlib_ImageColorRGB2CIEMono_Fp(mlib_image *dst,
                                          mlib_image *src);

mlib_status mlib_ImageColorTrue2IndexInit(void      **colormap,
                                          mlib_s32  bits,
                                          mlib_type intype,
                                          mlib_type outtype,
                                          mlib_s32  channels,
                                          mlib_s32  entries,
                                          mlib_s32  offset,
                                          void      **table);

mlib_status mlib_ImageColorTrue2Index(mlib_image *dst,
                                      mlib_image *src,
                                      void       *colormap);

void mlib_ImageColorTrue2IndexFree(void *colormap);

mlib_status mlib_ImageColorErrorDiffusion3x3(mlib_image *dst,
                                             mlib_image *src,
                                             mlib_s16   *kernel,
                                             mlib_s32   scale,
                                             void       *colormap);

mlib_status mlib_ImageColorOrderedDither8x8(mlib_image *dst,
                                            mlib_image *src,
                                            mlib_s16   *dmask,
                                            mlib_s32   scale,
                                            void       *colormap);

mlib_status mlib_ImageZoomIn2X(mlib_image  *dst,
                               mlib_image  *src,
                               mlib_filter filter,
                               mlib_edge   edge);

mlib_status mlib_ImageZoomOut2X(mlib_image  *dst,
                                mlib_image  *src,
                                mlib_filter filter,
                                mlib_edge   edge);

mlib_status mlib_ImageZoom(mlib_image  *dst,
                           mlib_image  *src,
                           mlib_d64    zoomx,
                           mlib_d64    zoomy,
                           mlib_filter filter,
                           mlib_edge   edge);

mlib_status mlib_ImageZoomIn2XIndex(mlib_image  *dst,
                                    mlib_image  *src,
                                    mlib_filter filter,
                                    mlib_edge   edge,
                                    void        *colormap);

mlib_status mlib_ImageZoomOut2XIndex(mlib_image  *dst,
                                     mlib_image  *src,
                                     mlib_filter filter,
                                     mlib_edge   edge,
                                     void        *colormap);

mlib_status mlib_ImageZoomIndex(mlib_image  *dst,
                                mlib_image  *src,
                                mlib_d64    zoomx,
                                mlib_d64    zoomy,
                                mlib_filter filter,
                                mlib_edge   edge,
                                void        *colormap);

mlib_status mlib_ImageZoomIn2X_Fp(mlib_image  *dst,
                                  mlib_image  *src,
                                  mlib_filter filter,
                                  mlib_edge   edge);

mlib_status mlib_ImageZoomOut2X_Fp(mlib_image  *dst,
                                   mlib_image  *src,
                                   mlib_filter filter,
                                   mlib_edge   edge);

mlib_status mlib_ImageZoom_Fp(mlib_image  *dst,
                              mlib_image  *src,
                              mlib_d64    zoomx,
                              mlib_d64    zoomy,
                              mlib_filter filter,
                              mlib_edge   edge);

mlib_status mlib_ImageRotate90(mlib_image *dst,
                               mlib_image *src);

mlib_status mlib_ImageRotate180(mlib_image *dst,
                                mlib_image *src);

mlib_status mlib_ImageRotate270(mlib_image *dst,
                                mlib_image *src);

mlib_status mlib_ImageFlipX(mlib_image *dst,
                            mlib_image *src);

mlib_status mlib_ImageFlipY(mlib_image *dst,
                            mlib_image *src);

mlib_status mlib_ImageFlipMainDiag(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageFlipAntiDiag(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageRotate90_Fp(mlib_image *dst,
                                  mlib_image *src);

mlib_status mlib_ImageRotate180_Fp(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageRotate270_Fp(mlib_image *dst,
                                   mlib_image *src);

mlib_status mlib_ImageFlipX_Fp(mlib_image *dst,
                               mlib_image *src);

mlib_status mlib_ImageFlipY_Fp(mlib_image *dst,
                               mlib_image *src);

mlib_status mlib_ImageFlipMainDiag_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageFlipAntiDiag_Fp(mlib_image *dst,
                                      mlib_image *src);

mlib_status mlib_ImageRotate(mlib_image  *dst,
                             mlib_image  *src,
                             mlib_f32    angle,
                             mlib_f32    xcenter,
                             mlib_f32    ycenter,
                             mlib_filter filter,
                             mlib_edge   edge);

mlib_status mlib_ImageRotateIndex(mlib_image  *dst,
                                  mlib_image  *src,
                                  mlib_f32    angle,
                                  mlib_f32    xcenter,
                                  mlib_f32    ycenter,
                                  mlib_filter filter,
                                  mlib_edge   edge,
                                  void        *colormap);

mlib_status mlib_ImageRotate_Fp(mlib_image  *dst,
                                mlib_image  *src,
                                mlib_d64    angle,
                                mlib_d64    xcenter,
                                mlib_d64    ycenter,
                                mlib_filter filter,
                                mlib_edge   edge);

mlib_status mlib_ImageAffine(mlib_image  *dst,
                             mlib_image  *src,
                             mlib_f32    *mtx,
                             mlib_filter filter,
                             mlib_edge   edge);

mlib_status mlib_ImageAffineIndex(mlib_image  *dst,
                                  mlib_image  *src,
                                  mlib_f32    *mtx,
                                  mlib_filter filter,
                                  mlib_edge   edge,
                                  void        *colormap);

mlib_status mlib_ImageAffine_Fp(mlib_image  *dst,
                                mlib_image  *src,
                                mlib_d64    *mtx,
                                mlib_filter filter,
                                mlib_edge   edge);

mlib_status mlib_ImageAffineTransform(mlib_image  *dst,
                                      mlib_image  *src,
                                      mlib_f32    *mtx,
                                      mlib_filter filter,
                                      mlib_edge   edge);

mlib_status mlib_ImageAffineTransformIndex(mlib_image  *dst,
                                           mlib_image  *src,
                                           mlib_f32    *mtx,
                                           mlib_filter filter,
                                           mlib_edge   edge,
                                           void        *colormap);

mlib_status mlib_ImageAffineTransform_Fp(mlib_image  *dst,
                                         mlib_image  *src,
                                         mlib_d64    *mtx,
                                         mlib_filter filter,
                                         mlib_edge   edge);

mlib_status mlib_ImageHistogram(mlib_s32   **histo,
                                mlib_image *img);

mlib_status mlib_ImageLookUp(mlib_image *dst,
                             mlib_image *src,
                             void       **table);

mlib_status mlib_ImageLookUp_Inp(mlib_image *srcdst,
                                 void       **table);

mlib_status mlib_ImageLookUpMask(mlib_image *dst,
                                 mlib_image *src,
                                 void       **table,
                                 mlib_s32   channels,
                                 mlib_s32   cmask);

mlib_status mlib_ImageThresh1(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *thresh,
                              mlib_s32   *ghigh,
                              mlib_s32   *glow);

mlib_status mlib_ImageThresh2(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *thresh,
                              mlib_s32   *glow);

mlib_status mlib_ImageThresh3(mlib_image *dst,
                              mlib_image *src,
                              mlib_s32   *thresh,
                              mlib_s32   *ghigh);

mlib_status mlib_ImageThresh1_Inp(mlib_image *srcdst,
                                  mlib_s32   *thresh,
                                  mlib_s32   *ghigh,
                                  mlib_s32   *glow);

mlib_status mlib_ImageThresh2_Inp(mlib_image *srcdst,
                                  mlib_s32   *thresh,
                                  mlib_s32   *glow);

mlib_status mlib_ImageThresh3_Inp(mlib_image *srcdst,
                                  mlib_s32   *thresh,
                                  mlib_s32   *ghigh);

mlib_status mlib_ImageThresh1_Fp(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_d64   *thresh,
                                 mlib_d64   *ghigh,
                                 mlib_d64   *glow);

mlib_status mlib_ImageThresh2_Fp(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_d64   *thresh,
                                 mlib_d64   *glow);

mlib_status mlib_ImageThresh3_Fp(mlib_image *dst,
                                 mlib_image *src,
                                 mlib_d64   *thresh,
                                 mlib_d64   *ghigh);

mlib_status mlib_ImageThresh1_Fp_Inp(mlib_image *srcdst,
                                     mlib_d64   *thresh,
                                     mlib_d64   *ghigh,
                                     mlib_d64   *glow);

mlib_status mlib_ImageThresh2_Fp_Inp(mlib_image *srcdst,
                                     mlib_d64   *thresh,
                                     mlib_d64   *glow);

mlib_status mlib_ImageThresh3_Fp_Inp(mlib_image *srcdst,
                                     mlib_d64   *thresh,
                                     mlib_d64   *ghigh);

mlib_status mlib_ImageMean(mlib_d64   *mean,
                           mlib_image *img);

mlib_status mlib_ImageMean_Fp(mlib_d64   *mean,
                              mlib_image *img);

mlib_status mlib_ImageStdDev(mlib_d64   *sdev,
                             mlib_image *img,
                             mlib_d64   *mean);

mlib_status mlib_ImageStdDev_Fp(mlib_d64   *sdev,
                                mlib_image *img,
                                mlib_d64   *mean);

mlib_status mlib_ImageMoment2(mlib_d64   *moment,
                              mlib_image *img);

mlib_status mlib_ImageMoment2_Fp(mlib_d64   *moment,
                                 mlib_image *img);

mlib_status mlib_ImageAutoCorrel(mlib_d64   *correl,
                                 mlib_image *img,
                                 mlib_s32   dx,
                                 mlib_s32   dy);

mlib_status mlib_ImageAutoCorrel_Fp(mlib_d64   *correl,
                                    mlib_image *img,
                                    mlib_s32   dx,
                                    mlib_s32   dy);

mlib_status mlib_ImageCrossCorrel(mlib_d64   *correl,
                                  mlib_image *img1,
                                  mlib_image *img2);

mlib_status mlib_ImageCrossCorrel_Fp(mlib_d64   *correl,
                                     mlib_image *img1,
                                     mlib_image *img2);

mlib_status mlib_ImageMaximum(mlib_s32   *max,
                              mlib_image *img);

mlib_status mlib_ImageMinimum(mlib_s32   *min,
                              mlib_image *img);

mlib_status mlib_ImageMaximum_Fp(mlib_d64   *max,
                                 mlib_image *img);

mlib_status mlib_ImageMinimum_Fp(mlib_d64   *min,
                                 mlib_image *img);

mlib_status mlib_ImageXProj(mlib_d64   *xproj,
                            mlib_image *img);

mlib_status mlib_ImageYProj(mlib_d64   *yproj,
                            mlib_image *img);

mlib_status mlib_ImageXProj_Fp(mlib_d64   *xproj,
                               mlib_image *img);

mlib_status mlib_ImageYProj_Fp(mlib_d64   *yproj,
                               mlib_image *img);

#ifdef __cplusplus
}
#endif

#endif    /* MLIB_IMAGE_PROTO_H */
