/*
 * @(#)mlib_image_types.h	1.6 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifdef _MSC_VER
#pragma warning(disable:4068 4244 4101)
#endif /*_MSC_VER*/

#ifndef MLIB_IMAGE_TYPES_H
#define MLIB_IMAGE_TYPES_H

#ifdef __SUNPRO_C
#pragma ident   "@(#)mlib_image_types.h	1.22    98/08/03 SMI"
#endif /* __SUNPRO_C */

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
  MLIB_BIT,	   /* 1-bit data                   */
  MLIB_BYTE,	   /* 8-bit unsigned integer data  */
  MLIB_SHORT,      /* 16-bit signed integer data   */
  MLIB_INT,        /* 32-bit signed integer data   */
  MLIB_FLOAT,      /* 32-bit floating-point data   */
  MLIB_DOUBLE      /* 64-bit floating-point data   */
} mlib_type;

typedef enum {
  MLIB_NEAREST,		/* nearest neighbor filter	*/
  MLIB_BILINEAR,	/* bilinear filter 		*/
  MLIB_BICUBIC		/* bicubic filter 		*/
} mlib_filter;

typedef enum {
  MLIB_EDGE_DST_NO_WRITE,	/* no write to dst edge */
  MLIB_EDGE_DST_FILL_ZERO,	/* set dst edge to zero */
  MLIB_EDGE_DST_COPY_SRC,	/* copy src edge to dst edge */
  MLIB_EDGE_OP_NEAREST,		/* use nearest neighbor interpolation
				   for edge pixels */
  MLIB_EDGE_OP_DEGRADED,	/* use degraded interpolation for
				   edge pixels, i.e., bicubic -> 
				   bilinear -> nearest neighbor */
  MLIB_EDGE_SRC_EXTEND,		/* extend src edge by replication */
  MLIB_EDGE_SRC_EXTEND_ZERO,	/* extend src edge with zeros */
  MLIB_EDGE_SRC_EXTEND_MIRROR	/* extend src edge with mirrored data */
} mlib_edge;

typedef enum {
  MLIB_BLEND_ZERO,
  MLIB_BLEND_ONE,
  MLIB_BLEND_DST_COLOR,
  MLIB_BLEND_SRC_COLOR,
  MLIB_BLEND_ONE_MINUS_DST_COLOR,
  MLIB_BLEND_ONE_MINUS_SRC_COLOR,
  MLIB_BLEND_SRC_ALPHA,
  MLIB_BLEND_ONE_MINUS_SRC_ALPHA,
  MLIB_BLEND_DST_ALPHA,
  MLIB_BLEND_ONE_MINUS_DST_ALPHA,
  MLIB_BLEND_SRC_ALPHA_SATURATE
} mlib_blend;

typedef struct {
  mlib_type type;        /* data type of image                       */
  mlib_s32  channels;    /* number of channels                       */
  mlib_s32  width;       /* width of image in pixels, x dimension    */
  mlib_s32  height;      /* height of image in pixels, y dimension   */
  mlib_s32  stride;      /* linestride = bytes to next row           */
  mlib_s32  flags;       /* collection of helpful hints              */
  void      *data;       /* pointer to first data pixel              */
  void      *state;      /* internal state structure                 */
  mlib_s32  reserved[10 - sizeof(void*)/2];
                         /* Reserved for future use. Also makes      */
                         /* size of this structure = 64 bytes, which */
                         /* is the size of the cache line.           */
} mlib_image;

/*
* Flags or hints are contained in a 32-bit integer. The bit structure is 
* shown below:
* 
*      3                   2                   1
*    1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
*   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ 
*   |S|                 |U|V| shint | hhint | whint |     dhint     |
*   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
* 
*   	S = 0	- attributes have been set (attribute field >= 0)
* 	    1	- attributes have not been set (attribute field < 0)
* 
* 	U = 0	- mediaLib allocated data space
* 	    1	- user allocated data space
* 
* 	V = 0	- stride == width => 1-D vector
* 	    1	- stride != width
* 
* 	shint	- last 4 bits of stride
* 
* 	hhint	- last 4 bits of height
* 
* 	whint	- last 4 bits of width
* 
* 	dhint	- last 8 bits of data address
*/

enum {
  MLIB_IMAGE_ALIGNED64     = 0x3f,
  MLIB_IMAGE_ALIGNED8      = 0x7,
  MLIB_IMAGE_ALIGNED4      = 0x2,
  MLIB_IMAGE_ALIGNED2      = 0x1,
  MLIB_IMAGE_WIDTH8X       = 0x700,
  MLIB_IMAGE_WIDTH4X       = 0x300,
  MLIB_IMAGE_WIDTH2X       = 0x100,
  MLIB_IMAGE_HEIGHT8X      = 0x7000,
  MLIB_IMAGE_HEIGHT4X      = 0x3000,
  MLIB_IMAGE_HEIGHT2X      = 0x1000,
  MLIB_IMAGE_STRIDE8X      = 0x70000,
  MLIB_IMAGE_ONEDVECTOR    = 0x100000,
  MLIB_IMAGE_USERALLOCATED = 0x200000,
  MLIB_IMAGE_ATTRIBUTESET  = 0x7fffffff 
}; 

#ifdef __cplusplus
}
#endif

#endif	/* MLIB_IMAGE_TYPES_H */
