/*
 * %W% %E%
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#ifndef _Included_BlitMacros
#define BlitMacros

#ifndef _Included_ImageData
#include <ImageData.h>
#endif

/* REMIND: fix portability issues */
#define MAX_INT (0x7fffffff)

/* guts of BLIT loops exposed here in macros */

#define DEFAULT_SRC_LUT int __srcLut = 0
#define DEFAULT_DST_LUT int __dstLut = 0
#define DEFAULT_SAME_LUT int __sameLut = 0
#define DEFAULT_COLORDATA ColorData *cData = 0
#define DEFAULT_INVERSE_LUT unsigned int __inverseGrayLut = 0

#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
#define DECLARE_DST_LUT DEFAULT_DST_LUT
#define DECLARE_SAME_LUT DEFAULT_SAME_LUT
#define DECLARE_COLORDATA DEFAULT_COLORDATA
#define DECLARE_INVERSE_LUT DEFAULT_INVERSE_LUT

#define DEFAULT_SRC_INFO(varName) void *varName = (void *)srcBase
#define DECLARE_SRC_INFO(varName)   DEFAULT_SRC_INFO(varName)
#define INIT_SRC_INFO

#define DEFAULT_DST_INFO(varName) void *varName = (void *)dstBase
#define DECLARE_DST_INFO(varName)   DEFAULT_DST_INFO(varName)
#define INIT_DST_INFO

#define DEFAULT_CHECK_SAME_LUT		do {} while (0)
#define DEFAULT_IS_SAME_LUT		(0)

#define IDX_CHECK_SAME_LUT \
    do { \
	if (srcLockInfo.arrayLutSize <= dstLockInfo.arrayLutSize) { \
	    unsigned int iLUT; \
	    __sameLut = 1; \
	    for (iLUT = 0; iLUT < srcLockInfo.arrayLutSize; iLUT++) { \
		if (srcLockInfo.lockedLut[iLUT] != \
		    dstLockInfo.lockedLut[iLUT]) \
		{ \
		    __sameLut = 0; \
		    break; \
		} \
	    } \
	} \
    } while (0)

#define IDX_IS_SAME_LUT	(__sameLut != 0)

#define CHECK_SAME_LUT	DEFAULT_CHECK_SAME_LUT
#define IS_SAME_LUT	DEFAULT_IS_SAME_LUT

#define INVOKE_BLIT( SRC_NAME, SRC_TYPE, DST_NAME, DST_TYPE, \
                        INNER_SCALE, SRC_BUMP, DST_BUMP) \
    do { \
        int pixelCount = minImageWidths(env, width, srcImage, dstImage); \
        int bltHeight   = minImageRows(env, height, srcImage, dstImage); \
        int rowCount = bltHeight; \
        if (pixelCount && rowCount) { \
            unsigned SRC_TYPE *srcBase; \
            unsigned DST_TYPE *dstBase; \
            unsigned int srcOffset; \
            int srcScanStride, dstScanStride; \
            ImageData ## SRC_NAME ## LockInfo srcLockInfo; \
            ImageData ## DST_NAME ## LockInfo dstLockInfo; \
            get ## SRC_NAME ## ImageLockInfo(env, srcImage, &srcLockInfo); \
            srcScanStride = srcLockInfo.scanStride; \
            offsetOfSrcData(env, dstImage, srcScanStride, srcOffset); \
            get ## DST_NAME ## ImageLockInfo(env, dstImage, &dstLockInfo); \
            dstScanStride = dstLockInfo.scanStride; \
            srcBase = lock ## SRC_NAME ## ImageData(env, &srcLockInfo); \
            dstBase = lock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            if (srcBase && dstBase) { \
                unsigned SRC_TYPE *srcRow = srcBase; \
                unsigned DST_TYPE *dstRow = dstBase; \
                unsigned int innerCount = pixelCount * INNER_SCALE; \
                DECLARE_INVERSE_LUT; \
                DECLARE_COLORDATA; \
                DECLARE_SRC_LUT; \
                DECLARE_DST_LUT; \
                DECLARE_SAME_LUT; \
                CHECK_SAME_LUT; \
                srcRow += srcOffset; \
                while (rowCount--) { \
                    INNER_LOOP( \
                        innerCount, \
                        SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP); \
                    srcRow += srcScanStride; \
                    dstRow += dstScanStride; \
                } \
            } \
            if (dstBase) { \
                unlock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            } \
            if (srcBase) { \
                unlock ## SRC_NAME ## ImageData(env, &srcLockInfo); \
            } \
        } \
    } while (0);

#define INNER_LOOP_4COMP_COPY(    INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            int alpha, red, green, blue; \
            SRC_TO_COMPONENTS(srcPixel, red, green, blue, alpha); \
            COMPONENTS_TO_DEST(dstPixel, red, green, blue, alpha); \
            srcPixel += SRC_BUMP; \
            dstPixel += DST_BUMP; \
        } \
    } while (0);

#define INNER_LOOP_3COMP_COPY(    INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            int red, green, blue; \
            SRC_TO_COMPONENTS(srcPixel, red, green, blue, 0); \
            COMPONENTS_TO_DEST(dstPixel, red, green, blue, 0); \
            srcPixel += SRC_BUMP; \
            dstPixel += DST_BUMP; \
        } \
    } while (0);

#define INNER_LOOP_XBGR_SWAP(   INNER_COUNT, \
                                SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned int *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned int p = *srcPixel++; \
            unsigned int g = p & 0x0000ff00; \
            p &= 0x00ff00ff; \
            p = (p >> 16) | (p << 16); \
            p |= g; \
            p |= 0xff000000; \
            *dstPixel++ = p; \
        } \
    } while (0);


#define INNER_LOOP_SET_INT_ALPHA(   INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned int *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        while (pixels_left--) { \
            *dstPixel++ = 0xff000000 | *srcPixel++; \
        } \
    } while (0);

#define INNER_LOOP_IDX_LUT8(   INNER_COUNT, \
                              SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            *dstPixel++ = srcLut[(*srcPixel++ & 0xff)]; \
        } \
    } while (0);

#define INNER_LOOP_IDX_LUT12(   INNER_COUNT, \
                              SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            *dstPixel++ = srcLut[(*srcPixel++ & 0xfff)]; \
        } \
    } while (0);

#define INNER_LOOP_BIDX_TO_3COMP(   INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            int red, green, blue; \
            BIDX_TO_THREE_COMPONENTS(srcPixel, red, green, blue); \
            COMPONENTS_TO_DEST(dstPixel, red, green, blue, alpha); \
            srcPixel += SRC_BUMP; \
            dstPixel += DST_BUMP; \
        } \
    } while (0);

#define INNER_LOOP_BIDX_FILL_TO_INT_ALPHA(   INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned int p = srcLut[*srcPixel++]; \
            if ( !(p & 0xff000000)) { \
                p = bgColor; \
            } \
            *dstPixel++ = 0xff000000 | p; \
        } \
    } while (0);

#define INNER_LOOP_BIDX_FILL_TO_3COMP(  INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            int red, green, blue; \
            unsigned int p = srcLut[*srcPixel++]; \
            if ( !(p & 0xff000000)) { \
                p = bgColor; \
            } \
            red = (p >> 16) & 0xff; \
            green = (p >> 8) & 0xff; \
            blue = p & 0xff; \
            COMPONENTS_TO_DEST(dstPixel, red, green, blue, alpha); \
            dstPixel += DST_BUMP; \
        } \
    } while (0);

#define INNER_LOOP_BIDX_XPAR_TO_INT_ALPHA(   INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned int p = srcLut[*srcPixel++]; \
            if (p & 0xff000000) { \
                *dstPixel = 0xff000000 | p;\
            } \
            dstPixel++; \
        } \
    } while (0);

#define INNER_LOOP_BIDX_XPAR_TO_3COMP(  INNER_COUNT, \
                                    SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned int p = srcLut[*srcPixel++]; \
            if (p & 0xff000000) { \
                int red, green, blue; \
                red = (p >> 16) & 0xff; \
                green = (p >> 8) & 0xff; \
                blue = p & 0xff; \
                COMPONENTS_TO_DEST(dstPixel, red, green, blue, alpha); \
            } \
            dstPixel += DST_BUMP; \
        } \
    } while (0);

#define INNER_LOOP_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
      if (IS_SAME_LUT) { \
	memcpy(dstRow, srcRow, INNER_COUNT); \
      } else { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned char *dstPixel = dstRow; \
        char *rerr = cData->img_oda_red + ((rowCount & 7)<<3); \
        char *gerr = cData->img_oda_green + ((rowCount & 7)<<3); \
        char *berr = cData->img_oda_blue + ((rowCount & 7)<<3); \
        while (pixels_left--) { \
            int red, green, blue; \
            SRC_TO_COMPONENTS(srcPixel, red, green, blue, 0); \
            srcPixel++; \
            red     += rerr[pixels_left & 7]; \
            green   += gerr[pixels_left & 7]; \
            blue    += berr[pixels_left & 7]; \
            if (((red | green | blue) & (~255)) != 0) { \
                if (red < 0) red = 0; \
                if (red > 255) red = 255; \
                if (green < 0) green = 0; \
                if (green > 255) green = 255; \
                if (blue < 0) blue = 0; \
                if (blue > 255) blue = 255; \
            } \
            *dstPixel++ = CUBEMAP(red,green,blue); \
        } \
      } \
    } while (0);

#define INNER_LOOP_FILL_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned char *dstPixel = dstRow; \
        char *rerr = cData->img_oda_red + ((rowCount & 7)<<3); \
        char *gerr = cData->img_oda_green + ((rowCount & 7)<<3); \
        char *berr = cData->img_oda_blue + ((rowCount & 7)<<3); \
        while (pixels_left--) { \
            unsigned int p = srcLut[*srcPixel++]; \
            if (p & 0xff000000) { \
                int red, green, blue; \
                red     = (p >> 16) & 0xff; \
                green   = (p >> 8)  & 0xff; \
                blue    = p & 0xff; \
                red     += rerr[pixels_left & 7]; \
                green   += gerr[pixels_left & 7]; \
                blue    += berr[pixels_left & 7]; \
                if (((red | green | blue) & (~255)) != 0) { \
                    if (red < 0) red = 0; \
                    if (red > 255) red = 255; \
                    if (green < 0) green = 0; \
                    if (green > 255) green = 255; \
                    if (blue < 0) blue = 0; \
                    if (blue > 255) blue = 255; \
                } \
                *dstPixel++ = CUBEMAP(red,green,blue); \
            } else { \
                *dstPixel++ = bgColorIndexed; \
            } \
        } \
    } while (0);

#define INNER_LOOP_XPAR_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned char *dstPixel = dstRow; \
        char *rerr = cData->img_oda_red + ((rowCount & 7)<<3); \
        char *gerr = cData->img_oda_green + ((rowCount & 7)<<3); \
        char *berr = cData->img_oda_blue + ((rowCount & 7)<<3); \
        while (pixels_left--) { \
            unsigned int p = srcLut[*srcPixel++]; \
            if (p & 0xff000000) { \
                int red, green, blue; \
                red     = (p >> 16) & 0xff; \
                green   = (p >> 8)  & 0xff; \
                blue    = p & 0xff; \
                red     += rerr[pixels_left & 7]; \
                green   += gerr[pixels_left & 7]; \
                blue    += berr[pixels_left & 7]; \
                if (((red | green | blue) & (~255)) != 0) { \
                    if (red < 0) red = 0; \
                    if (red > 255) red = 255; \
                    if (green < 0) green = 0; \
                    if (green > 255) green = 255; \
                    if (blue < 0) blue = 0; \
                    if (blue > 255) blue = 255; \
                } \
                *dstPixel++ = CUBEMAP(red,green,blue); \
            } else { \
                dstPixel++; \
            } \
        } \
    } while (0);

#define ARGB_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned int p = *(s); \
        (a) = (p >> 24) & 0xff; (r) = (p >> 16) & 0xff; \
        (g) = (p >> 8)  & 0xff; (b) = p & 0xff; \
    } while (0);

#define XRGB_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned int p = *(s); \
        (a) = 0xff; (r) = (p >> 16) & 0xff; \
        (g) = (p >> 8)  & 0xff; (b) = p & 0xff; \
    } while (0);

#define XBGR_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned int p = *(s); \
        (a) = 0xff; (b) = (p >> 16) & 0xff; \
        (g) = (p >> 8)  & 0xff; (r) = p & 0xff; \
    } while (0);

#define BGRX_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned int p = *(s); \
        (a) = 0xff; (r) = (p >> 8) & 0xff; \
        (g) = (p >> 16)  & 0xff; (b) = p >> 24; \
    } while (0);

#define RGBX_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned int p = *(s); \
        (a) = 0xff; (b) = (p >> 8) & 0xff; \
        (g) = (p >> 16)  & 0xff; (r) = p >> 24; \
    } while (0);

#define FOUR_COMPONENTS_TO_ABGR_4BYTE(d,r,g,b,a) \
    (d)[0] = (a); (d)[1] = (b); (d)[2] = (g); (d)[3] = (r);

#define FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a) \
    *(d) = ((a) << 24) | ((r) << 16) | ((g) << 8) | (b);

#define FOUR_COMPONENTS_TO_BGRA_4BYTE(d,r,g,b,a) \
    (d)[0] = (b); (d)[1] = (g); (d)[2] = (r); (d)[3] = (a);

#define BLEND(alpha, fg) ((((alpha) * (fg)) >> 8) + (alpha >> 7))

#define FOUR_COMPONENTS_TO_ABGR_4BYTE_PREMUL(d,r,g,b,a) \
    (d)[0] = (a); \
    if (a) { \
        if ((a) < 255) { \
            (d)[1] = BLEND((a),(b)); \
            (d)[2] = BLEND((a),(g)); \
            (d)[3] = BLEND((a),(r)); \
        } else { \
            (d)[1] = (b); \
            (d)[2] = (g); \
            (d)[3] = (r); \
        } \
    } else { \
        (d)[1] = 0; \
        (d)[2] = 0; \
        (d)[3] = 0; \
    }

#define FOUR_COMPONENTS_TO_ARGB_PREMUL(d,r,g,b,a) \
    if (a) { \
        if ((a) < 255) { \
            *(d) =  ((a) << 24) \
                | (BLEND((a),(r)) << 16) \
                | (BLEND((a),(g)) << 8) \
                | BLEND((a),(b)); \
        } else { \
            *(d) = ((a) << 24) | ((r) << 16) | ((g) << 8) | (b); \
        } \
    } else { \
        *(d) = 0; \
    }

#define UNBLEND(alpha,color) ((((color) - ((alpha) >> 7))<<8) / (alpha))

#define ARGB_PREMUL_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned int p = *(s); \
        (a) = (p >> 24) & 0xff; (r) = (p >> 16) & 0xff; \
        (g) = (p >> 8)  & 0xff; (b) = p & 0xff; \
        if (a) { \
            if ((a) < 255) { \
                (r) = UNBLEND((a),(r)); \
                (g) = UNBLEND((a),(g)); \
                (b) = UNBLEND((a),(b)); \
            } \
        } else { \
            (r) = 0x0ff; \
            (g) = 0x0ff; \
            (b) = 0x0ff; \
        } \
    } while (0);

#define ABGR_4BYTE_PREMUL_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        (a) = (s)[0]; (b) = (s)[1]; (g) = (s)[2]; (r) = (s)[3]; \
        if (a) { \
            if ((a) < 255) { \
                (r) = UNBLEND((a),(r)); \
                (g) = UNBLEND((a),(g)); \
                (b) = UNBLEND((a),(b)); \
            } \
        } else { \
            (r) = 0x0ff; \
            (g) = 0x0ff; \
            (b) = 0x0ff; \
        } \
    } while (0);

#define DIB_XRGB_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    (a) = 0xff; (r) = (s)[1]; (g) = (s)[2]; (b) = (s)[3];

#define SHORT_555_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned short p = *(s); \
        (a) = 0xff; \
        (r) = (p >> 10) & 0x1f; \
        (r) = ((r) << 3) | ((r) >> 2); \
        (g) = (p >> 5) & 0x1f; \
        (g) = ((g) << 3) | ((g) >> 2); \
        (b) = p & 0x1f; \
        (b) = ((b) << 3) | ((b) >> 2); \
    } while (0);

#define SHORT_565_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned short p = *(s); \
        (a) = 0xff; \
        (r) = p >> 11; \
        (r) = ((r) << 3) | ((r) >> 2); \
        (g) = (p >> 5) & 0x3f; \
        (g) = ((g) << 2) | ((g) >> 4); \
        (b) = p & 0x1f; \
        (b) = ((b) << 3) | ((b) >> 2); \
    } while (0);

#define RGB_24BIT_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    (a) = 0xff; (r) = (s)[0]; (g) = (s)[1]; (b) = (s)[2];

#define BGR_24BIT_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    (a) = 0xff; (r) = (s)[2]; (g) = (s)[1]; (b) = (s)[0];

#define ABGR_4BYTE_TO_FOUR_COMPONENTS(s,r,g,b,a) \
    (a) = (s)[0]; (r) = (s)[3]; (g) = (s)[2]; (b) = (s)[1];

#define XRGB_TO_THREE_COMPONENTS(s,r,g,b) \
    do { \
        unsigned int p = *(s); \
        (r) = (p >> 16) & 0xff; \
        (g) = (p >> 8)  & 0xff; \
        (b) = p & 0xff; \
    } while (0);

/* REMIND: need to deal with translucent colors in LUT */
#define BIDX_TO_THREE_COMPONENTS(s,r,g,b) \
    do { \
        unsigned int p = srcLut[*(s)]; \
        (r) = (p >> 16) & 0xff; \
        (g) = (p >> 8)  & 0xff; \
        (b) = p & 0xff; \
    } while (0);

#define BIDX_XPAR_TO_THREE_COMPONENTS(s,r,g,b,a) \
    do { \
        unsigned int p = srcLut[*(s)]; \
        (r) = (p >> 16) & 0xff; \
        (g) = (p >> 8)  & 0xff; \
        (b) = p & 0xff; \
        (a) = p & 0xff000000; \
    } while (0);

#define THREE_COMPONENTS_TO_DITHER_BIDX(d,r,g,b) \

#define THREE_COMPONENTS_TO_555(d,r,g,b) \
    *(d) = (((r) & 0xf8) << 7) | (((g) & 0xf8) << 2) | ((b) >> 3);

#define THREE_COMPONENTS_TO_565(d,r,g,b) \
    *(d) = (((r) & 0xf8) << 8) | (((g) & 0xfc) << 3) | ((b) >> 3);

#define THREE_COMPONENTS_TO_BGRX(d,r,g,b) \
    *(d) = ((b) << 24) | ((g) << 16) | ((r) << 8);

#define THREE_COMPONENTS_TO_RGBX(d,r,g,b) \
    *(d) = ((r) << 24) | ((g) << 16) | ((b) << 8);

#define THREE_COMPONENTS_TO_XBGR(d,r,g,b) \
    *(d) = ((b) << 16) | ((g) << 8) | (r);

#define THREE_COMPONENTS_TO_RGB_24BIT(d,r,g,b) \
    (d)[0] = (r); (d)[1] = (g); (d)[2] = (b);

#define THREE_COMPONENTS_TO_BGR_24BIT(d,r,g,b) \
    (d)[0] = (b); (d)[1] = (g); (d)[2] = (r);

#define INVOKE_COLOR_LOOP(CALLED_LOOP,  DST_NAME, DST_TYPE) \
    do { \
        int pixelCount = minImageWidths(env, alphaScanSize, \
                                        dstImage, dstImage); \
        int rowCount   = minImageRows(env, MAX_INT, \
                                        dstImage, dstImage); \
        if (pixelCount && rowCount) { \
            unsigned DST_TYPE *dstBase; \
            int dstScanStride; \
            ImageData ## DST_NAME ## LockInfo dstLockInfo; \
            get ## DST_NAME ## ImageLockInfo(env, dstImage, &dstLockInfo); \
            dstScanStride = dstLockInfo.scanStride; \
            alphaOffset += offsetOfAlphaData(env, dstImage, \
                                alphaScanSize); \
            dstBase = lock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            if (dstBase) { \
                unsigned char *pathBase; \
		DECLARE_COLORDATA; \
                DECLARE_DST_INFO(dstInfo);\
                INIT_DST_INFO;\
                pathBase = getAlphaInfo(env, alpha); \
                (CALLED_LOOP)(env, \
                    dstInfo, dstScanStride, 1, \
                    pathBase, alphaOffset, alphaScanSize, \
                    rule, \
                    extraAlpha, \
                    pixelCount, rowCount, \
                    fgColor, cData \
                ); \
                dropBufs(env, alpha, pathBase, 0, 0); \
            } \
            if (dstBase) { \
                unlock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            } \
        } \
    } while (0);

#define INVOKE_PAINT_LOOP(CALLED_LOOP,  SRC_NAME, SRC_TYPE, \
                                        DST_NAME, DST_TYPE) \
    do { \
        int pixelCount = minImageWidths(env, width, \
                                        srcImage, dstImage); \
        int rowCount   = minImageRows(env, height, \
                                        srcImage, dstImage); \
        if (pixelCount && rowCount) { \
            unsigned SRC_TYPE *srcBase; \
            unsigned DST_TYPE *dstBase; \
            unsigned int srcOffset; \
            int srcScanStride ,dstScanStride; \
            ImageData ## SRC_NAME ## LockInfo srcLockInfo; \
            ImageData ## DST_NAME ## LockInfo dstLockInfo; \
            get ## SRC_NAME ## ImageLockInfo(env, srcImage, &srcLockInfo); \
            srcScanStride = srcLockInfo.scanStride; \
            offsetOfSrcData(env, dstImage, srcScanStride, srcOffset); \
            get ## DST_NAME ## ImageLockInfo(env, dstImage, &dstLockInfo); \
            dstScanStride = dstLockInfo.scanStride; \
            alphaOffset += offsetOfAlphaData(env, srcImage, \
                                alphaScanSize); \
            srcBase = lock ## SRC_NAME ## ImageData(env, &srcLockInfo); \
            dstBase = lock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            if (srcBase && dstBase) { \
                unsigned char *pathBase; \
		DECLARE_COLORDATA; \
                DECLARE_SRC_INFO(srcInfo);\
                DECLARE_DST_INFO(dstInfo);\
                INIT_SRC_INFO; \
                INIT_DST_INFO; \
                srcBase += srcOffset; \
                pathBase = getAlphaInfo(env, alpha); \
                (CALLED_LOOP)(env, \
                    dstInfo, dstScanStride, 1, \
                    srcInfo, srcScanStride, 1, \
                    pathBase, alphaOffset, alphaScanSize, \
                    rule, \
                    extraAlpha, \
                    pixelCount, rowCount, \
                    srcPreMultiplied, \
                    xorColor, cData \
                ); \
                dropBufs(env, alpha, pathBase, 0, 0); \
            } \
            if (dstBase) { \
                unlock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            } \
            if (srcBase) { \
                unlock ## SRC_NAME ## ImageData(env, &srcLockInfo); \
            } \
        } \
    } while (0);

#define INVOKE_BG_PAINT_LOOP(CALLED_LOOP,   SRC_NAME, SRC_TYPE, \
                                            DST_NAME, DST_TYPE) \
    do { \
        int pixelCount = minImageWidths(env, width, \
                                        srcImage, dstImage); \
        int rowCount   = minImageRows(env, height, \
                                        srcImage, dstImage); \
        if (pixelCount && rowCount) { \
            unsigned SRC_TYPE *srcBase; \
            unsigned DST_TYPE *dstBase; \
            unsigned int srcOffset; \
            int srcScanStride ,dstScanStride; \
            ImageData ## SRC_NAME ## LockInfo srcLockInfo; \
            ImageData ## DST_NAME ## LockInfo dstLockInfo; \
            get ## SRC_NAME ## ImageLockInfo(env, srcImage, &srcLockInfo); \
            srcScanStride = srcLockInfo.scanStride; \
            offsetOfSrcData(env, dstImage, srcScanStride, srcOffset); \
            get ## DST_NAME ## ImageLockInfo(env, dstImage, &dstLockInfo); \
            dstScanStride = dstLockInfo.scanStride; \
            alphaOffset += offsetOfAlphaData(env, srcImage, \
                                alphaScanSize); \
            srcBase = lock ## SRC_NAME ## ImageData(env, &srcLockInfo); \
            dstBase = lock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            if (srcBase && dstBase) { \
                unsigned char *pathBase; \
		DECLARE_COLORDATA; \
                DECLARE_SRC_INFO(srcInfo);\
                DECLARE_DST_INFO(dstInfo);\
                INIT_SRC_INFO; \
                INIT_DST_INFO; \
                srcBase += srcOffset; \
                pathBase = getAlphaInfo(env, alpha); \
                (CALLED_LOOP)(env, \
                    dstInfo, dstScanStride, 1, \
                    srcInfo, srcScanStride, 1, \
                    pathBase, alphaOffset, alphaScanSize, \
                    rule, \
                    extraAlpha, \
                    pixelCount, rowCount, \
                    srcPreMultiplied, \
                    bgColor, xorColor, cData \
                ); \
                dropBufs(env, alpha, pathBase, 0, 0); \
            } \
            if (dstBase) { \
                unlock ## DST_NAME ## ImageData(env, &dstLockInfo); \
            } \
            if (srcBase) { \
                unlock ## SRC_NAME ## ImageData(env, &srcLockInfo); \
            } \
        } \
    } while (0);

#endif
