/*
 * @(#)DefaultComponent.c	1.31 98/09/09
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include <stdlib.h>
#include <jni.h>
#include <jni_util.h>
#include <sun_java2d_loops_DefaultComponent.h>
#include <memory.h>
#include <math.h>

#include <alphaCompst.h>
#include <img_globals.h>

#ifndef _Included_ImageData
#include <ImageData.h>
#endif

#ifndef _Included_BlitMacros
#include <BlitMacros.h>
#endif

/* REMIND: many function signatures in comments are wrong */

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbToRgb24Bit
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbToRgb24Bit(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    XRGB_TO_THREE_COMPONENTS(s,r,g,b)
#define COMPONENTS_TO_DEST(d,r,g,b,a)   THREE_COMPONENTS_TO_RGB_24BIT(d,r,g,b)
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    INNER_LOOP_3COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Byte, char, 1, 1, 3);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbToBgr24Bit
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbToBgr24Bit(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    XRGB_TO_THREE_COMPONENTS(s,r,g,b)
#define COMPONENTS_TO_DEST(d,r,g,b,a)   THREE_COMPONENTS_TO_BGR_24BIT(d,r,g,b)
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    INNER_LOOP_3COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Byte, char, 1, 1, 3);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}


/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntIsomorphicCopy
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntIsomorphicCopy(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
/* REMIND -- may want memmove if */
/* behavior is defined somewhere */
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    memcpy(dstRow, srcRow, INNER_COUNT);

    INVOKE_BLIT(Int, int, Int, int, sizeof(unsigned int), 1, 1);
#undef INNER_LOOP
}
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntArgbbgfillToIntArgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntArgbbgfillToIntArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned int p = (unsigned int) *srcPixel++; \
            if ( !(p & 0xff000000)) { \
                p = bgColor; \
            } \
            *dstPixel++ =  p; \
        } \
    } while (0);

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

    #undef INNER_LOOP
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbTo555
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbTo555(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    XRGB_TO_THREE_COMPONENTS(s,r,g,b)
#define COMPONENTS_TO_DEST(d,r,g,b,a)   THREE_COMPONENTS_TO_555(d,r,g,b)
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    INNER_LOOP_3COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Short, short, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbTo565
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbTo565(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    XRGB_TO_THREE_COMPONENTS(s,r,g,b)
#define COMPONENTS_TO_DEST(d,r,g,b,a)   THREE_COMPONENTS_TO_565(d,r,g,b)
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    INNER_LOOP_3COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Short, short, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbToBgrx
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbToBgrx(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    XRGB_TO_THREE_COMPONENTS(s,r,g,b)
#define COMPONENTS_TO_DEST(d,r,g,b,a)   THREE_COMPONENTS_TO_BGRX(d,r,g,b)
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    INNER_LOOP_3COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbToRgbx
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbToRgbx(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    XRGB_TO_THREE_COMPONENTS(s,r,g,b)
#define COMPONENTS_TO_DEST(d,r,g,b,a)   THREE_COMPONENTS_TO_RGBX(d,r,g,b)
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    INNER_LOOP_3COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbToXbgr
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbToXbgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    INNER_LOOP_XBGR_SWAP(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntArgbTo4byteAbgr
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntArgbTo4byteAbgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #define SRC_TO_COMPONENTS(s,r,g,b,a)    \
        ARGB_TO_FOUR_COMPONENTS(s,r,g,b,a)
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        FOUR_COMPONENTS_TO_ABGR_4BYTE(d,r,g,b,a)
    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
        INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                              DST_TYPE, DST_BUMP)

        INVOKE_BLIT(Int, int, Byte, char, 1, 1, 4);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbTo4byteAbgr
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbTo4byteAbgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
        #define SRC_TO_COMPONENTS(s,r,g,b,a)    \
            XRGB_TO_FOUR_COMPONENTS(s,r,g,b,a)
        #define COMPONENTS_TO_DEST(d,r,g,b,a) \
            FOUR_COMPONENTS_TO_ABGR_4BYTE(d,r,g,b,a)
        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                            DST_TYPE, DST_BUMP)

            INVOKE_BLIT(Int, int, Byte, char, 1, 1, 4);

        #undef INNER_LOOP
        #undef COMPONENTS_TO_DEST
        #undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbTo4byteBgra
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbTo4byteBgra(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #define SRC_TO_COMPONENTS(s,r,g,b,a)    \
        XRGB_TO_FOUR_COMPONENTS(s,r,g,b,a)
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        FOUR_COMPONENTS_TO_BGRA_4BYTE(d,r,g,b,a)
    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                        DST_TYPE, DST_BUMP)

        INVOKE_BLIT(Int, int, Byte, char, 1, 1, 4);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_SET_INT_ALPHA(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntArgbTo4byteAbgrPremul
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntArgbTo4byteAbgrPremul(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    \
    ARGB_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR_4BYTE_PREMUL(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Byte, char, 1, 1, 4);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntArgbToArgbPremul
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntArgbToArgbPremul(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a)    \
    ARGB_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ARGB_PREMUL(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int , int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

ColorLoopFunc XRGBcolorloop;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ColorFillAlphaToXRGB
 * Signature: (Lsun/awt/image/ImageData;I[BIIFI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ColorFillAlphaToXRGB(
    JNIEnv *env, jobject self,
    jobject dstImage,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jint fgColor)
{
    INVOKE_COLOR_LOOP(XRGBcolorloop,Int,int);
}

ColorLoopFunc ARGBcolorloop;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ColorFillAlphaToARGB
 * Signature: (Lsun/awt/image/ImageData;I[BIIFI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ColorFillAlphaToARGB(
    JNIEnv *env, jobject self,
    jobject dstImage,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jint fgColor)
{
    INVOKE_COLOR_LOOP(ARGBcolorloop,Int,int);
}

PaintLoopFunc ARGBpaintARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbOntoArgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbOntoArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint xorColor
  )
{
    INVOKE_PAINT_LOOP(ARGBpaintARGB,Int,int,Int,int);
}

PaintLoopFunc XRGBpaintARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbOntoXrgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbOntoXrgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint xorColor
  )
{
    INVOKE_PAINT_LOOP(XRGBpaintARGB,Int,int,Int,int);
}

BgPaintLoopFunc XRGBbgpaintARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbBkOntoXrgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbBkOntoXrgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint bgColor, jint xorColor
  )
{
    INVOKE_BG_PAINT_LOOP(XRGBbgpaintARGB,Int,int,Int,int);
}

BgPaintLoopFunc   ARGBbgpaintARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbBkOntoArgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbBkOntoArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint bgColor, jint xorColor
  )
{
    INVOKE_BG_PAINT_LOOP(ARGBbgpaintARGB,Int,int,Int,int);
}

BgPaintLoopFunc   ARGBbgpaintBitmaskARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbBkMaskOntoArgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbBkMaskOntoArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint bgColor, jint xorColor
  )
{
    INVOKE_BG_PAINT_LOOP(ARGBbgpaintBitmaskARGB,Int,int,Int,int);
}

BgPaintLoopFunc   XRGBbgpaintBitmaskARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbBkMaskOntoXrgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbBkMaskOntoXrgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint bgColor, jint xorColor
  )
{
    INVOKE_BG_PAINT_LOOP(XRGBbgpaintBitmaskARGB,Int,int,Int,int);
}

PaintLoopFunc   ARGBpaintBitmaskARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbMaskOntoArgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbMaskOntoArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint xorColor
  )
{
    INVOKE_PAINT_LOOP(ARGBpaintBitmaskARGB,Int,int,Int,int);
}

PaintLoopFunc   XRGBpaintBitmaskARGB;
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    paintArgbMaskOntoXrgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_paintArgbMaskOntoXrgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint xorColor
  )
{
    INVOKE_PAINT_LOOP(XRGBpaintBitmaskARGB,Int,int,Int,int);
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    Abgr4bytePremulToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_Abgr4bytePremulToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    ABGR_4BYTE_PREMUL_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Byte, char, Int, int, 1, 4, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ArgbPremulToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ArgbPremulToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    ARGB_PREMUL_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    DibXrgbToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_DibXrgbToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    DIB_XRGB_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Byte, char, Int, int, 1, 4, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ShortIsomorphicCopy
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ShortIsomorphicCopy(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
/* REMIND -- may want memmove if */
/* behavior is defined somewhere */
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    memcpy(dstRow, srcRow, INNER_COUNT);

    INVOKE_BLIT(Short, short, Short, short, sizeof(unsigned short), 1, 1);
#undef INNER_LOOP
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    Short555ToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_Short555ToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    SHORT_555_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Short, short, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    Short565ToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_Short565ToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    SHORT_565_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Short, short, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    Rgb24BitToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_Rgb24BitToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    RGB_24BIT_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Byte, char, Int, int, 1, 3, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    Bgr24BitToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_Bgr24BitToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    BGR_24BIT_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Byte, char, Int, int, 1, 3, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    Abgr4byteToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_Abgr4byteToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    ABGR_4BYTE_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Byte, char, Int, int, 1, 4, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    RgbxToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_RgbxToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    RGBX_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    XbgrToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_XbgrToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_XBGR_SWAP(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    BgrxToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_BgrxToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define SRC_TO_COMPONENTS(s,r,g,b,a) \
    BGRX_TO_FOUR_COMPONENTS(s,r,g,b,a)
#define COMPONENTS_TO_DEST(d,r,g,b,a) \
    FOUR_COMPONENTS_TO_ABGR(d,r,g,b,a)
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_4COMP_COPY(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(Int, int, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef COMPONENTS_TO_DEST
#undef SRC_TO_COMPONENTS
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ByteIndexToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ByteIndexToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_IDX_LUT8(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ByteIndexFillToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ByteIndexFillToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_BIDX_FILL_TO_INT_ALPHA(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ByteIndexXparToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ByteIndexXparToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    INNER_LOOP_BIDX_XPAR_TO_INT_ALPHA(INNER_COUNT, SRC_TYPE, SRC_BUMP, \
                    DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ShortIndexToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ShortIndexToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int lutLimit = srcLockInfo.arrayLutSize; \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned short *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned short p = *srcPixel++; \
            p = (p < lutLimit)? p: 0; \
            *dstPixel++ = 0xff000000 | srcLut[p]; \
        } \
    } while (0);

    INVOKE_BLIT(ShortIndexed, short, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

#define bitOffsetOfBytePacked(env, img, pixPerDat, startOffs) \
    do { \
        int x; \
        pixPerDat = (*env)->GetIntField(env, img, \
                                        gImageData.pixelsPerDataUnitID); \
        x = (*env)->GetIntField(env, img, gImageData.xOutputAreaID); \
        startOffs = x % pixPerDat; \
    } while (0);

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    BinaryToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_BinaryToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int lutLimit = srcLockInfo.arrayLutSize; \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        unsigned int bitSize = 8/srcPixPerDat; \
        unsigned int bitMask = 0xff >> (8-bitSize); \
        int bits; \
        unsigned char c = *srcPixel++; \
        unsigned int p; \
        for (bits = srcPixPerDat-srcXStartBitOffset-1 ; bits >= 0 ; bits--) { \
            p = (c >> bits*bitSize) & bitMask; \
            *dstPixel++ = srcLut[p]; \
            if (--pixels_left == 0) { \
                break; \
            } \
        } \
        for (;;) { \
            if (pixels_left == 0) { \
                break; \
            } \
            c = *srcPixel++; \
            for (bits = srcPixPerDat-1 ; bits >= 0 ; bits--) { \
                p = (c >> bits*bitSize) & bitMask; \
                *dstPixel++ = srcLut[p]; \
                if (--pixels_left == 0) { \
                    break; \
                } \
            } \
        } \
    } while (0);

    unsigned int srcPixPerDat;
    unsigned int srcXStartBitOffset;

    bitOffsetOfBytePacked(env, srcImage, srcPixPerDat, srcXStartBitOffset);
    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

#undef INNER_LOOP
#undef DECLARE_SRC_LUT
#undef GET_STARTBIT
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

#define DITHERCOLS	8
#define DITHERROWS	8

static unsigned char dither1bit[] = {
       0, 127,  31, 159,   7, 135,  39, 167,
     191,  63, 223,  95, 199,  71, 231, 103,
      47, 175,  15, 143,  55, 183,  23, 151,
     239, 111, 207,  79, 247, 119, 215,  87,
      11, 139,  43, 171,   3, 131,  35, 163,
     203,  75, 235, 107, 195,  67, 227,  99,
      59, 187,  27, 155,  51, 179,  19, 147,
     251, 123, 219,  91, 243, 115, 211,  83,
};
/* max dither1bit value = 251 */
static unsigned char lookup1bit[] = {
    /* 0x000 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x010 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x020 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x030 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x040 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x050 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x060 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x070 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x080 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x090 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x0A0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x0B0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x0C0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x0D0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x0E0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x0F0 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
    /* 0x100 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x110 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x120 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x130 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x140 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x150 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x160 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x170 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x180 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x190 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x1A0 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x1B0 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x1C0 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x1D0 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x1E0 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x1F0 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
};
static unsigned char dither2bit[] = {
       0,  42,  10,  53,   2,  45,  13,  55,
      63,  21,  74,  31,  66,  23,  77,  34,
      15,  58,   5,  47,  18,  61,   7,  50,
      79,  37,  69,  26,  82,  39,  71,  29,
       3,  46,  14,  57,   1,  43,  11,  54,
      67,  25,  78,  35,  65,  22,  75,  33,
      19,  62,   9,  51,  17,  59,   6,  49,
      83,  41,  73,  30,  81,  38,  70,  27,
};
/* max dither2bit value = 83 */
static unsigned char lookup2bit[] = {
    /* 0x000 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x010 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x020 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x030 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x040 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x050 */  0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x060 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x070 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x080 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x090 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x0A0 */  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2,
    /* 0x0B0 */  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    /* 0x0C0 */  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    /* 0x0D0 */  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    /* 0x0E0 */  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    /* 0x0F0 */  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3,
    /* 0x100 */  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    /* 0x110 */  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    /* 0x120 */  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    /* 0x130 */  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    /* 0x140 */  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    /* 0x150 */  3, 3, 3,
};
static unsigned char dither4bit[] = {
       0,   8,   2,  10,   0,   9,   2,  11,
      12,   4,  14,   6,  13,   4,  15,   6,
       3,  11,   1,   9,   3,  12,   1,  10,
      15,   7,  13,   5,  16,   7,  14,   5,
       0,   9,   2,  11,   0,   8,   2,  10,
      13,   5,  15,   7,  13,   4,  15,   6,
       3,  12,   1,  10,   3,  11,   1,   9,
      16,   8,  14,   6,  16,   7,  14,   5,
};
/* max dither value = 16 */
static unsigned char lookup4bit[] = {
    /* 0x000 */  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x010 */  0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x020 */  1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
    /* 0x030 */  2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
    /* 0x040 */  3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    /* 0x050 */  4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    /* 0x060 */  5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    /* 0x070 */  6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    /* 0x080 */  7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8,
    /* 0x090 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9,
    /* 0x0A0 */  9, 9, 9, 9, 9, 9, 9, 9, 9, 9,10,10,10,10,10,10,
    /* 0x0B0 */ 10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,
    /* 0x0C0 */ 11,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,
    /* 0x0D0 */ 12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,
    /* 0x0E0 */ 13,13,13,13,13,13,13,13,13,13,13,13,13,13,14,14,
    /* 0x0F0 */ 14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,15,
    /* 0x100 */ 15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,
};
/* 8-bit gray should not go through these tables... */
static unsigned char dither8bit[] = {
       0,   0,   0,   0,   0,   0,   0,   0,
       0,   0,   0,   0,   0,   0,   0,   0,
       0,   0,   0,   0,   0,   0,   0,   0,
       0,   0,   0,   0,   0,   0,   0,   0,
       0,   0,   0,   0,   0,   0,   0,   0,
       0,   0,   0,   0,   0,   0,   0,   0,
       0,   0,   0,   0,   0,   0,   0,   0,
       0,   0,   0,   0,   0,   0,   0,   0,
};
/* max dither value = 0 */
static unsigned char lookup8bit[] = {
    /* 0x000 */  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,
    /* 0x010 */ 16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
    /* 0x020 */ 32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,
    /* 0x030 */ 48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
    /* 0x040 */ 64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,
    /* 0x050 */ 80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
    /* 0x060 */ 96,97,98,99,100,101,102,103,
		 104,105,106,107,108,109,110,111,
    /* 0x070 */ 112,113,114,115,116,117,118,119,
		 120,121,122,123,124,125,126,127,
    /* 0x080 */ 128,129,130,131,132,133,134,135,
		 136,137,138,139,140,141,142,143,
    /* 0x090 */ 144,145,146,147,148,149,150,151,
		 152,153,154,155,156,157,158,159,
    /* 0x0A0 */ 160,161,162,163,164,165,166,167,
		 168,169,170,171,172,173,174,175,
    /* 0x0B0 */ 176,177,178,179,180,181,182,183,
		 184,185,186,187,188,189,190,191,
    /* 0x0C0 */ 192,193,194,195,196,197,198,199,
		 200,201,202,203,204,205,206,207,
    /* 0x0D0 */ 208,209,210,211,212,213,214,215,
		 216,217,218,219,220,221,222,223,
    /* 0x0E0 */ 224,225,226,227,228,229,230,231,
		 232,233,234,235,236,237,238,239,
    /* 0x0F0 */ 240,241,242,243,244,245,246,247,
		 248,249,250,251,252,253,254,255,
};

static struct {
    unsigned char *errors;
    unsigned char *lookup;
} dithermatrices[] = {
    { dither8bit, lookup8bit }, /* illegal, 0 pixperdat */
    { dither8bit, lookup8bit }, /* 8-bit, 1 pixperdat */
    { dither4bit, lookup4bit }, /* 4-bit, 2 pixperdat */
    { dither4bit, lookup4bit }, /* illegal, 3 pixperdat */
    { dither2bit, lookup2bit }, /* 2-bit, 4 pixperdat */
    { dither2bit, lookup2bit }, /* illegal, 5 pixperdat */
    { dither2bit, lookup2bit }, /* illegal, 6 pixperdat */
    { dither2bit, lookup2bit }, /* illegal, 7 pixperdat */
    { dither1bit, lookup1bit }, /* 1-bit, 8 pixperdat */
};

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbToBinary
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbToBinary(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned int *srcPixel = srcRow; \
        unsigned char *dstPixel = dstRow; \
        unsigned int bitSize = 8/dstPixPerDat; \
        unsigned int y; \
        int bits; \
        unsigned char dr; \
        unsigned char d = 0; \
        unsigned int s; \
        int rowIndex = ((((bltHeight-rowCount)+yDev) % DITHERROWS) * \
                        DITHERCOLS); \
        int columnIndex = xDev; \
        for (bits = dstPixPerDat-dstXStartBitOffset ; \
             bits && pixels_left ; \
             bits--, pixels_left--) { \
            s = *srcPixel++; \
            y = ((((s >> 16) & 0xff) *  76) + \
                 (((s >>  8) & 0xff) * 151) + \
                 (((s >>  0) & 0xff) *  29))/256; \
            y = lookup[y + errors[rowIndex + \
                                 ((columnIndex++) % DITHERCOLS)]]; \
            d |= (y << ((bits-1)*bitSize)); \
        } \
        dr = *dstPixel & ~((0xff >> (dstXStartBitOffset+bits)*bitSize) << \
                            bits*bitSize); \
        *dstPixel++ = d | dr; \
        while (pixels_left) { \
            for (bits = dstPixPerDat, d = 0 ; bits && pixels_left ; \
                 bits--, pixels_left--) { \
                s = *srcPixel++; \
                y = ((((s >> 16) & 0xff) *  76) + \
                     (((s >>  8) & 0xff) * 151) + \
                     (((s >>  0) & 0xff) *  29))/256; \
                y = lookup[y + errors[rowIndex + \
                                 ((columnIndex++) % DITHERCOLS)]]; \
                d |= (y << ((bits-1)*bitSize)); \
            } \
            if (bits) { \
                dr = *dstPixel & ~(0xff << bits*bitSize); \
                *dstPixel++ = (d | dr); \
            } else { \
                *dstPixel++ = d; \
            } \
        } \
    } while (0);

    unsigned int dstPixPerDat;
    unsigned int dstXStartBitOffset;
    unsigned char *errors;
    unsigned char *lookup;
    int xDev=0;
    int yDev=0;
    
    bitOffsetOfBytePacked(env, dstImage, dstPixPerDat, dstXStartBitOffset);
    
    /* dither setup */
    errors = dithermatrices[dstPixPerDat].errors;
    lookup = dithermatrices[dstPixPerDat].lookup;
    
    xDev = (*env)->GetIntField(env, dstImage, gImageData.xDeviceAreaID);
    yDev = (*env)->GetIntField(env, dstImage, gImageData.yDeviceAreaID);
    
    INVOKE_BLIT(Int, int, ByteIndexed, char, 1, 1, 1);

#undef INNER_LOOP
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntRgbTo8bitDithered
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntRgbTo8bitDithered(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    color_init();

    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #define SRC_TO_COMPONENTS(s,r,g,b,a) \
            XRGB_TO_THREE_COMPONENTS(s,r,g,b)
        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(Int, int, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef SRC_TO_COMPONENTS
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ByteGrayToArgb
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ByteGrayToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned char *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned char p = *srcPixel++; \
            *dstPixel++ = 0xff000000|(p<<16)|(p<<8)|p; \
        } \
    } while (0);

    INVOKE_BLIT(Byte, char, Int, int, 1, 1, 1);

#undef INNER_LOOP
}
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ByteGrayToByteGray
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ByteGrayToByteGray(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        memcpy(dstRow, srcRow, INNER_COUNT); \
    } while (0);

    INVOKE_BLIT(Byte, char, Byte, char, 1, 1, 1);

#undef INNER_LOOP
}
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ByteGrayToIndex8Gray
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ByteGrayToIndex8Gray(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#undef DECLARE_INVERSE_LUT
#define DECLARE_INVERSE_LUT unsigned int *inverseGrayLut = dstLockInfo.lockedInverseGrayLut 
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            *dstPixel++ = inverseGrayLut[*srcPixel++]; \
        } \
    } while(0);
    
    INVOKE_BLIT(Byte, char, Index8Gray, char, 1, 1, 1);

#undef INNER_LOOP
#undef DECLARE_INVERSE_LUT   
#define DECLARE_INVERSE_LUT DEFAULT_INVERSE_LUT 

}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ByteGrayTo8bitDithered
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ByteGrayTo8bitDithered(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    color_init();

    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #define SRC_TO_COMPONENTS(s,r,g,b,a) \
            unsigned char p = *(s); \
            (r) = p; (b) = p; (g) = p;

        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(Byte, char, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef SRC_TO_COMPONENTS
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ShortGrayToArgb
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ShortGrayToArgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint shift_bits)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned short *srcPixel = srcRow; \
        unsigned int *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned char gray_val = (unsigned char ) ((shift_bits > 0)? ((*srcPixel++)>>shift_bits):(*srcPixel++)<<(-shift_bits)); \
            *dstPixel++ = 0xff000000|(gray_val<<16)|(gray_val<<8)|gray_val; \
        } \
    } while (0);

    INVOKE_BLIT(Short, short, Int, int, 1, 1, 1);

#undef INNER_LOOP
    
}
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ArgbToShortGray
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ArgbToShortGray(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned int *srcPixel = srcRow; \
        unsigned short *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned int s = *srcPixel++; \
            unsigned int d = ((((s >>  8) & 0xff00) *  19532) >> 16) + \
                             ((((s >>  0) & 0xff00) *  38807) >> 16) + \
                             ((((s <<  8) & 0xff00) *   7196) >> 16); \
            if (d > 0xffff) d = 0xffff; \
            *dstPixel++ = d; \
        } \
    } while (0);

    INVOKE_BLIT(Int, int, Short, short, 1, 1, 1);

#undef INNER_LOOP
}

/*
 * Class:     sun_java2d_loops_RasterOutputManager
 * Method:    getPlatformFontVar
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL
Java_sun_java2d_loops_RasterOutputManager_getPlatformFontVar (JNIEnv *env, jclass cl) {

    /*
     * Please do not print out the result of the environmental
     * value.
     */
    char *c = getenv("JAVA2D_USEPLATFORMFONT");
    if (c) {
        /* printf("JAVA2D_USEPLATFORMFONT set\n"); */
        return 0xff;
    } else {
        /* printf("JAVA2D_USEPLATFORMFONT not set\n"); */
        return 0;
    }
}
/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    IntArgbToByteGray
 * Signature: (Lsun/awt/image/ImageData { ;};Lsun/awt/image/ImageData { ;};II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_IntArgbToByteGray(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
   do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned int *srcPixel = srcRow; \
        unsigned char *dstPixel = dstRow; \
        while (pixels_left--) { \
               int tmp = *srcPixel++; \
               int r = (tmp >>16) & 0xff;\
               int g = (tmp >>8) & 0xff;\
               int b = (tmp) & 0xff;\
               *dstPixel++ = (int) (((76*r) + (150*g) + (29*b) + 128) / 256); \
            } \
    } while (0);

        INVOKE_BLIT(Int, int, Byte, char, 1, 1, 1);

#undef INNER_LOOP
}


/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ShortGrayToByteGray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ShortGrayToByteGray(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint shift_bits)
{
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            *dstPixel++ = (unsigned DST_TYPE ) ((shift_bits > 0)? ((*srcPixel++)>>shift_bits):(*srcPixel++)<<(-shift_bits)); \
        } \
    } while (0);

    INVOKE_BLIT(Short, short, Byte, char, 1, 1, 1);

#undef INNER_LOOP
    
}

/*
 * Class:     sun_java2d_loops_DefaultComponent
 * Method:    ShortGrayToIndex8Gray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_DefaultComponent_ShortGrayToIndex8Gray(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint shift_bits)
{
#undef DECLARE_INVERSE_LUT
#define DECLARE_INVERSE_LUT unsigned int *inverseGrayLut = dstLockInfo.lockedInverseGrayLut 
#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                    DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned char gray_val = (unsigned char ) ((shift_bits > 0)? ((*srcPixel++)>>shift_bits):(*srcPixel++)<<(-shift_bits)); \
            *dstPixel++ = inverseGrayLut[gray_val]; \
        } \
    } while (0);

    INVOKE_BLIT(Short, short, Index8Gray, char, 1, 1, 1);

#undef INNER_LOOP
#undef DECLARE_INVERSE_LUT   
#define DECLARE_INVERSE_LUT DEFAULT_INVERSE_LUT  
}

