/*
 * @(#)GIFAcceleratorLoops.c	1.22 98/08/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include <jni.h>
#include <jni_util.h>
#include <sun_java2d_loops_GIFAcceleratorLoops.h>
#include <memory.h>
#include <math.h>
#include <img_globals.h>
#include <img_util_md.h>

#ifndef _Included_ImageData
#include <ImageData.h>
#endif

#ifndef _Included_BlitMacros
#include <BlitMacros.h>
#endif

/* REMIND duplicated cube map code belongs in a shared location */

/* 8BIT_INDEXED family START */
/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTcopyToIndexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTcopyToIndexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    color_init();

    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
	#undef CHECK_SAME_LUT
	#define CHECK_SAME_LUT IDX_CHECK_SAME_LUT
	#undef IS_SAME_LUT
	#define IS_SAME_LUT IDX_IS_SAME_LUT

        #define SRC_TO_COMPONENTS(s,r,g,b,a) \
            BIDX_TO_THREE_COMPONENTS(s,r,g,b)
        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(ByteIndexed, char, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef SRC_TO_COMPONENTS
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
	#undef CHECK_SAME_LUT
	#define CHECK_SAME_LUT DEFAULT_CHECK_SAME_LUT
	#undef IS_SAME_LUT
	#define IS_SAME_LUT DEFAULT_IS_SAME_LUT
    }
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTbgfillToIndexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTbgfillToIndexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{
    color_init();

    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT \
            unsigned int *srcLut = srcLockInfo.lockedLut; \
            unsigned char bgColorIndexed = CUBEMAP( \
                ((bgColor >> 16) & 0xff), \
                ((bgColor >> 8) & 0xff), \
                (bgColor & 0xff) \
            )

        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_FILL_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(ByteIndexed, char, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTxparToIndexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTxparToIndexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    color_init();

    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut

        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_XPAR_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(ByteIndexed, char, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
        #undef DECLARE_COLORDATA
        #define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}
/* 8BIT_INDEXED family END */

/* INT_RGB family START */
/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTcopyToIntRgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTcopyToIntRgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_IDX_LUT8( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

    #undef INNER_LOOP
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTbgfillToIntRgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTbgfillToIntRgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_FILL_TO_INT_ALPHA( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

    #undef INNER_LOOP
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTxparToIntRgb
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTxparToIntRgb(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_XPAR_TO_INT_ALPHA( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

    #undef INNER_LOOP
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}
/* INT_RGB family END */

/* 24BIT_BGR family START */
/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTcopyTo24BitBgr
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTcopyTo24BitBgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_BGR_24BIT(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Byte, char, 1, 1, 3);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTbgfillTo24BitBgr
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTbgfillTo24BitBgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,jint bgColor)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_BGR_24BIT(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_FILL_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Byte, char, 1, 1, 3);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTxparTo24BitBgr
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTxparTo24BitBgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_BGR_24BIT(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_XPAR_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Byte, char, 1, 1, 3);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}
/* 24BIT_BGR family END */

/* 16BIT_565 family START */
/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTcopyTo16Bit565
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTcopyTo16Bit565(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_565(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Short, short, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTbgfillTo16Bit565
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTbgfillTo16Bit565(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_565(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_FILL_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Short, short, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTxparTo16Bit565
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTxparTo16Bit565(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_565(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_XPAR_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Short, short, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}
/* 16BIT_565 family END */

/* 16BIT_555 family START */
/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTcopyTo16Bit555
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTcopyTo16Bit555(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_555(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Short, short, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTbgfillTo16Bit555
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTbgfillTo16Bit555(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_555(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_FILL_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Short, short, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTxparTo16Bit555
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTxparTo16Bit555(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_555(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_XPAR_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Short, short, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}
/* 16BIT_555 family END */

/* 32BIT_xBGR family START */
/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTcopyTo32BitXBgr
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTcopyTo32BitXBgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_XBGR(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTbgfillTo32BitXBgr
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTbgfillTo32BitXBgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_XBGR(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_FILL_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTxparTo32BitXBgr
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTxparTo32BitXBgr(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define COMPONENTS_TO_DEST(d,r,g,b,a) \
        THREE_COMPONENTS_TO_XBGR(d,r,g,b)

    #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP) \
        INNER_LOOP_BIDX_XPAR_TO_3COMP( INNER_COUNT, \
                        SRC_TYPE, SRC_BUMP,\
                        DST_TYPE, DST_BUMP)

    INVOKE_BLIT(ByteIndexed, char, Int, int, 1, 1, 1);

    #undef INNER_LOOP
    #undef COMPONENTS_TO_DEST
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}
/* 32BIT_xBGR family END */    

/* DITHERED 8BIT_INDEXED family START */
/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTcopyDitheredToIndexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTcopyDitheredToIndexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    color_init();

    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut

        #define SRC_TO_COMPONENTS(s,r,g,b,a) \
            BIDX_TO_THREE_COMPONENTS(s,r,g,b)
        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(ByteIndexed, char, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef SRC_TO_COMPONENTS
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
        #undef DECLARE_COLORDATA
        #define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTbgfillDitheredToIndexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTbgfillDitheredToIndexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height, jint bgColor)
{
    color_init();

    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT \
            unsigned int *srcLut = srcLockInfo.lockedLut; \
            unsigned char bgColorIndexed = CUBEMAP( \
                ((bgColor >> 16) & 0xff), \
                ((bgColor >> 8) & 0xff), \
                (bgColor & 0xff) \
            )

        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_FILL_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(ByteIndexed, char, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
        #undef DECLARE_COLORDATA
        #define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

/*
 * Class:     sun_java2d_loops_GIFAcceleratorLoops
 * Method:    LUTxparDitheredToIndexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_GIFAcceleratorLoops_LUTxparDitheredToIndexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
    color_init();

    if (cubemapArray) {
        #undef DECLARE_COLORDATA
        #define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut

        #define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP) \
            INNER_LOOP_XPAR_DITHER_BIDX( INNER_COUNT, SRC_TYPE, SRC_BUMP,\
                            DST_TYPE, DST_BUMP)

        INVOKE_BLIT(ByteIndexed, char, ByteIndexed, char, 1, 1, 1);

        #undef INNER_LOOP
        #undef DECLARE_SRC_LUT
        #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
        #undef DECLARE_COLORDATA
        #define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

/* DITHERED 8BIT_INDEXED family END */
