/*
 * @(#)IndexGrayAccelerators.c	1.5 98/08/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Remind: to add alpha-pre-multiplied cases 
 */

#include <stdlib.h>
#include <jni.h>
#include <jni_util.h>
#include <sun_java2d_loops_IndexGrayAccelerators.h>
#include <memory.h>
#include <math.h>

#include <alphaCompst.h>
#include <img_globals.h>

#ifndef _Included_ImageData
#include <ImageData.h>
#endif

#ifndef _Included_BlitMacros
#include <BlitMacros.h>
#endif

/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    Index12GrayToByteGray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */
JNIEXPORT void JNICALL 
Java_sun_java2d_loops_IndexGrayAccelerators_Index12GrayToByteGray(
    JNIEnv *env, jobject joSelf, 
    jobject srcImage, jobject dstImage, jint width, jint height) 
{
    #undef DECLARE_SRC_LUT
    #define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
    #define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
        INNER_LOOP_IDX_LUT12(INNER_COUNT, SRC_TYPE, \
                           SRC_BUMP, DST_TYPE, DST_BUMP);

    INVOKE_BLIT(Index12Gray, short, Byte, char, 1, 1, 1);

    #undef INNER_LOOP
    #undef DECLARE_SRC_LUT   
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    Index12GrayToIndex8Gray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */
JNIEXPORT void JNICALL 
Java_sun_java2d_loops_IndexGrayAccelerators_Index12GrayToIndex8Gray(
    JNIEnv *env, jobject joSelf, 
    jobject srcImage, jobject dstImage, jint width, jint height) 
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#undef DECLARE_INVERSE_LUT
#define DECLARE_INVERSE_LUT unsigned int *inverseGrayLut = dstLockInfo.lockedInverseGrayLut 
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned char gray_val = (unsigned char)srcLut[(*srcPixel++ & 0xfff)]; \
            *dstPixel++ = inverseGrayLut[gray_val]; \
        } \
    } while(0);


    INVOKE_BLIT(Index12Gray, short, Index8Gray, char, 1, 1, 1);
#undef INNER_LOOP   
#undef DECLARE_SRC_LUT   
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
#undef DECLARE_SRC_LUT   
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT   
#undef DECLARE_INVERSE_LUT   
#define DECLARE_INVERSE_LUT DEFAULT_INVERSE_LUT 
}


/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    Index12GrayToIntArgb
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */ 
JNIEXPORT void JNICALL 
Java_sun_java2d_loops_IndexGrayAccelerators_Index12GrayToIntArgb(
    JNIEnv *env, jobject joSelf, 
    jobject srcImage, jobject dstImage, jint width, jint height) 
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut

#define INNER_LOOP( INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
            INNER_LOOP_IDX_LUT12(INNER_COUNT, SRC_TYPE, \
                           SRC_BUMP, DST_TYPE, DST_BUMP);

      INVOKE_BLIT(Index12Gray, short, Int, int, 1, 1, 1);
#undef INNER_LOOP   
#undef DECLARE_SRC_LUT   
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}


/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    Index8GrayToByteGray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */
JNIEXPORT void JNICALL 
Java_sun_java2d_loops_IndexGrayAccelerators_Index8GrayToByteGray(
    JNIEnv *env, jobject joSelf, 
    jobject srcImage, jobject dstImage, jint width, jint height)
{

#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
          INNER_LOOP_IDX_LUT8(INNER_COUNT, SRC_TYPE, \
                           SRC_BUMP, DST_TYPE, DST_BUMP);

    INVOKE_BLIT(Index8Gray, char, Byte, char, 1, 1, 1);
#undef INNER_LOOP   
#undef DECLARE_SRC_LUT   
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    Index8GrayToIndex8Gray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */
JNIEXPORT void JNICALL 
Java_sun_java2d_loops_IndexGrayAccelerators_Index8GrayToIndex8Gray(
    JNIEnv *env, jobject joSelf, 
    jobject srcImage, jobject dstImage, jint width, jint height)
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut 
#undef DECLARE_INVERSE_LUT
#define DECLARE_INVERSE_LUT unsigned int *inverseGrayLut = dstLockInfo.lockedInverseGrayLut 
  
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
            unsigned char gray_val = srcLut[(*srcPixel++ & 0xff)]; \
            *dstPixel++ = inverseGrayLut[gray_val]; \
        } \
    } while(0);

    INVOKE_BLIT(Index8Gray, char, Index8Gray, char, 1, 1, 1);
#undef INNER_LOOP   
#undef DECLARE_SRC_LUT   
#define DECLARE_SRC_LUT DEFAULT_SRC_LUT   
#undef DECLARE_INVERSE_LUT   
#define DECLARE_INVERSE_LUT DEFAULT_INVERSE_LUT   
}

/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    Index8GrayToIntArgb
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */
JNIEXPORT void JNICALL 
Java_sun_java2d_loops_IndexGrayAccelerators_Index8GrayToIntArgb(
    JNIEnv *env, jobject joSelf, 
    jobject srcImage, jobject dstImage, jint width, jint height) 
{
#undef DECLARE_SRC_LUT
#define DECLARE_SRC_LUT unsigned int *srcLut = srcLockInfo.lockedLut
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
     INNER_LOOP_IDX_LUT8(INNER_COUNT, SRC_TYPE, \
                        SRC_BUMP, DST_TYPE, DST_BUMP);

        
     INVOKE_BLIT(Index8Gray, char, Int, int, 1, 1, 1);
#undef INNER_LOOP   
#undef DECLARE_SRC_LUT   
    #define DECLARE_SRC_LUT DEFAULT_SRC_LUT
}

/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    IntArgbToIndex12Gray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */
JNIEXPORT void JNICALL Java_sun_java2d_loops_IndexGrayAccelerators_IntArgbToIndex12Gray
    (JNIEnv *env, jobject joSelf, 
   jobject srcImage, jobject dstImage, jint width, jint height) {
#undef DECLARE_INVERSE_LUT
#define DECLARE_INVERSE_LUT unsigned int *inverseGrayLut = dstLockInfo.lockedInverseGrayLut 

#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
     do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
           unsigned int tmp = *srcPixel++; \
           unsigned int r = (tmp >>16) & 0xff;\
           unsigned int g = (tmp >>8) & 0xff;\
           unsigned int b = (tmp) & 0xff;\
           unsigned DST_TYPE gray_val = (unsigned DST_TYPE)(((76*r)+(150*g)+(29*b)+128)/256); \
           *dstPixel++ =  inverseGrayLut[gray_val]; \
        } \
    } while (0);
        
        INVOKE_BLIT(Int, int, Index12Gray, short, 1, 1, 1);
#undef INNER_LOOP   
#undef DECLARE_INVERSE_LUT   
#define DECLARE_INVERSE_LUT DEFAULT_INVERSE_LUT   

}


/*
 * Class:     sun_java2d_loops_IndexGrayAccelerators
 * Method:    IntArgbToIndex8Gray
 * Signature: (Lsun/java2d/loops/ImageData;Lsun/java2d/loops/ImageData;II)V
 */
JNIEXPORT void JNICALL 
Java_sun_java2d_loops_IndexGrayAccelerators_IntArgbToIndex8Gray
    (JNIEnv *env, jobject joSelf, 
   jobject srcImage, jobject dstImage, jint width, jint height) {

#undef DECLARE_INVERSE_LUT
#define DECLARE_INVERSE_LUT unsigned int *inverseGrayLut = dstLockInfo.lockedInverseGrayLut 

#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
     do { \
        unsigned int pixels_left = INNER_COUNT; \
        unsigned SRC_TYPE *srcPixel = srcRow; \
        unsigned DST_TYPE *dstPixel = dstRow; \
        while (pixels_left--) { \
           unsigned int tmp = *srcPixel++; \
           unsigned int r = (tmp >>16) & 0xff;\
           unsigned int g = (tmp >>8) & 0xff;\
           unsigned int b = (tmp) & 0xff;\
           unsigned DST_TYPE gray_val = (unsigned DST_TYPE)(((76*r)+(150*g)+(29*b)+128)/256); \
           *dstPixel++ =  inverseGrayLut[gray_val]; \
        } \
    } while (0);

      INVOKE_BLIT(Int, int, Index8Gray, char, 1, 1, 1);
#undef INNER_LOOP   
#undef DECLARE_INVERSE_LUT   
#define DECLARE_INVERSE_LUT DEFAULT_INVERSE_LUT   
}

