/*
 * @(#)IndexedCompositing.c	1.7 98/09/08
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include <jni.h>
#include <jni_util.h>
#include <sun_java2d_loops_IndexedCompositing.h>
#include <memory.h>
#include <math.h>
#include <img_globals.h>

#include <stdlib.h>
#include <alpha_uti.h>
#include <alphaCompst.h>
#include <alphaCompst_Ductus.h>


#ifndef _Included_ImageData
#include <ImageData.h>
#endif

#ifndef _Included_BlitMacros
#include <BlitMacros.h>
#endif

#undef  DECLARE_DST_INFO
#define DECLARE_DST_INFO(varName)   IndexedInfo dstII;                      \
                                    void *varName = (void *)&dstII;         \
                                    dstII.pixelPtr = dstBase;               \
                                    dstII.dstLockInfo = &dstLockInfo;       \
                                    dstII.colormap = dstLockInfo.lockedLut;


#undef DST_XCCC

#undef DeclareDst
#undef InitDst
#undef ReadDstPixel
#undef PrepColorDst
#undef CopyColorToDst
#undef WriteDstPixel
#undef NextDstPixel
#undef NextDstRow
#include <Indexed_DstData.h>

/* REMIND CUBEMAP does not work with translucent LUTs */
/* REMIND duplicated cube map code belongs in a shared location */

ColorLoopFunc IndexedColorloop;
/*
 * Class:     sun_java2d_loops_IndexedCompositing
 * Method:    ColorFillAlphaToIndexed
 * Signature: (Lsun/awt/image/ImageData;I[BIIFI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_IndexedCompositing_ColorFillAlphaToIndexed(
    JNIEnv *env, jobject self,
    jobject dstImage,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jint fgColor)
{
    color_init();
    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        INVOKE_COLOR_LOOP(IndexedColorloop, ByteIndexed, char);
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

PaintLoopFunc   IndexedPaintARGB;
/*
 * Class:     sun_java2d_loops_IndexedCompositing
 * Method:    paintARGB_ONTO_Indexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_IndexedCompositing_paintARGB_1ONTO_1Indexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint xorColor
  )
{
    color_init();
    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        INVOKE_PAINT_LOOP(IndexedPaintARGB, Int, int, ByteIndexed, char);
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

PaintLoopFunc   IndexedPaintXRGB;
/*
 * Class:     sun_java2d_loops_IndexedCompositing
 * Method:    paintXRGB_ONTO_Indexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_IndexedCompositing_paintXRGB_1ONTO_1Indexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint xorColor
  )
{
    color_init();
    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        INVOKE_PAINT_LOOP(IndexedPaintXRGB, Int, int, ByteIndexed, char);
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

BgPaintLoopFunc   IndexedBgpaintARGB;
/*
 * Class:     sun_java2d_loops_IndexedCompositing
 * Method:    paintARGB_WITH_BG_CLR_ONTO_Indexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZII)V
 */
JNIEXPORT void JNICALL Java_sun_java2d_loops_IndexedCompositing_paintARGB_1WITH_1BG_1CLR_1ONTO_1Indexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint bgColor, jint xorColor
  )
{
    color_init();
    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        INVOKE_BG_PAINT_LOOP(IndexedBgpaintARGB, Int, int, ByteIndexed, char);
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

BgPaintLoopFunc   IndexedBgpaintBitmaskARGB;
/*
 * Class:     sun_java2d_loops_IndexedCompositing
 * Method:    paintARGB_WITH_BITMASK_AND_BG_CLR_ONTO_Indexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZII)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_IndexedCompositing_paintARGB_1WITH_1BITMASK_1AND_1BG_1CLR_1ONTO_1Indexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint bgColor, jint xorColor
  )
{
    color_init();
    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        INVOKE_BG_PAINT_LOOP(IndexedBgpaintBitmaskARGB, Int, int, ByteIndexed, char);
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

PaintLoopFunc   IndexedPaintBitmaskARGB;
/*
 * Class:     sun_java2d_loops_IndexedCompositing
 * Method:    paintARGB_WITH_BITMASK_ONTO_Indexed
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;III[BIIFZI)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_IndexedCompositing_paintARGB_1WITH_1BITMASK_1ONTO_1Indexed(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height,

    jint rule,

    jbyteArray alpha, jint alphaOffset, jint alphaScanSize,
    jfloat extraAlpha,
    jboolean srcPreMultiplied,
    jint xorColor
  )
{
    color_init();
    if (cubemapArray) {
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA ColorData *cData = dstLockInfo.colorData
        INVOKE_PAINT_LOOP(IndexedPaintBitmaskARGB, Int, int, ByteIndexed, char);
	#undef DECLARE_COLORDATA
	#define DECLARE_COLORDATA DEFAULT_COLORDATA
    }
}

/************************************* Indexed composite loops *********************************/

#undef DST_XCCC

#undef DeclareDst
#undef InitDst
#undef ReadDstPixel
#undef PrepColorDst
#undef CopyColorToDst
#undef WriteDstPixel
#undef NextDstPixel
#undef NextDstRow
#include <Indexed_DstData.h>

#undef COLOR_BLENDFUNC
#define COLOR_BLENDFUNC IndexedColorloop
#include <alphaCompst_colorloop.h>

#undef DeclareSrc
#undef InitSrc
#undef ReadSrcPixel
#undef NextSrcPixel
#undef NextSrcRow
#undef TILE_BG_BLENDFUNC
#include <ARGB_SrcData.h>
#define TILE_BG_BLENDFUNC IndexedBgpaintARGB
#include <alphaCompst_bgpaintloop.h>

#undef DeclareSrc
#undef InitSrc
#undef ReadSrcPixel
#undef NextSrcPixel
#undef NextSrcRow
#undef TILE_BG_BLENDFUNC
#include <BitmaskARGB_SrcData.h>
#define TILE_BG_BLENDFUNC IndexedBgpaintBitmaskARGB
#include <alphaCompst_bgpaintloop.h>


#undef DeclareSrc
#undef InitSrc
#undef ReadSrcPixel
#undef NextSrcPixel
#undef NextSrcRow
#undef TILE_BLENDFUNC
#include <BitmaskARGB_SrcData.h>
#define TILE_BLENDFUNC IndexedPaintBitmaskARGB
#include <alphaCompst_paintloop.h>

#undef DeclareSrc
#undef InitSrc
#undef ReadSrcPixel
#undef NextSrcPixel
#undef NextSrcRow
#undef TILE_BLENDFUNC
#include <ARGB_SrcData.h>
#define TILE_BLENDFUNC IndexedPaintARGB
#include <alphaCompst_paintloop.h>

#undef DeclareSrc
#undef InitSrc
#undef ReadSrcPixel
#undef NextSrcPixel
#undef NextSrcRow
#undef TILE_BLENDFUNC
#include <XRGB_SrcData.h>
#define TILE_BLENDFUNC IndexedPaintXRGB
#include <alphaCompst_paintloop.h>

