/*
 * @(#)IntDiscreteRenderer.c	1.44 98/09/14
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include <string.h>
#include <jni.h>
#include <jni_util.h>
#include <jlong.h>
#include <alphaCompst.h>
#include <stdlib.h>

#include "SpanIterator.h"

#include "sun_java2d_loops_IntDiscreteRenderer.h"

static void
doDrawLine(jint *dstBase, jint dstScanStr,
	   jint color, jint x1, jint y1, jint x2, jint y2,
	   jint cx, jint cy, jint cw, jint ch);

JNIEXPORT void JNICALL
Java_sun_java2d_loops_IntDiscreteRenderer_devDrawRect(
     JNIEnv *env, jclass clazz, jint color,
     jint x, jint y, jint w, jint h,
     jint cx, jint cy, jint cw, jint ch,
     jint offset, jintArray dst, jint dstPixStr, jint dstScanStr)
{
    jint *dstBase;
    jint count;
    unsigned int *cursor;
    jint x1, y1, x2, y2;

    x1 = x;
    y1 = y;
    x2 = x + w;
    y2 = y + h;
    if (x1 < cx) x1 = cx;
    if (y1 < cy) y1 = cy;
    if (x2 >= cx + cw) x2 = cx + cw - 1;
    if (y2 >= cy + ch) y2 = cy + ch - 1;
    if (x1 > x2 || y1 > y2) {
	return;
    }

    dstBase = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (dstBase == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
        return;
    }

    if (w == 0 || h == 0) {
	doDrawLine(dstBase + offset, dstScanStr, color,
		   x1, y1, x2, y2, cx, cy, cw, ch);

	(*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
	return;
    }

    cursor = (unsigned int *)dstBase + offset + x1 + y1*dstScanStr;

    /* line 1 */
    if (y1 == y) {
	for (count = x1; count < x2; count++) {
	    *cursor++ = color;
	}
    } else {
	if (x1 == x) {
	    *cursor = color;	/* shared pixel with line 4 */
	}
	cursor += x2 - x1;
    }
    /* line 2 */
    if (x2 == x + w) {
	for (count = y1; count < y2; count++) {
	    *cursor = color;
	    cursor += dstScanStr;
	}
    } else {
	if (y1 == y) {
	    *cursor = color;	/* shared pixel with line 1 */
	}
	cursor += dstScanStr * (y2 - y1);
    }
    /* line 3 */
    if (y2 == y + h) {
	for (count = x2; count > x1; count--) {
	    *cursor-- = color;
	}
    } else {
	if (x2 == x + w) {
	    *cursor = color;	/* shared pixel with line 2 */
	}
	cursor -= x2 - x1;
    }
    /* line 4 */
    if (x1 == x) {
	for (count = y2; count >= y1; count--) {
	    *cursor = color;
	    cursor -= dstScanStr;
	}
    } else {
	if (y2 == y + h) {
	    *cursor = color;	/* shared pixel with line 3 */
	}
	/* We don't need cursor any more anyway... */
	/* cursor -= dstScanStr * (y2 - y1 + 1); */
    }

    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
}

JNIEXPORT void JNICALL
Java_sun_java2d_loops_IntDiscreteRenderer_devDrawLine(
     JNIEnv *env, jclass clazz, jint color,
     jint x1, jint y1, jint x2, jint y2,
     jint cx, jint cy, jint cw, jint ch,
     jint offset, jintArray dst, jint dstPixStr, jint dstScanStr)
{
    jint *dstBase;

    dstBase = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (dstBase == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
        return;
    }

    if (y1 >= y2) {
        doDrawLine(dstBase + offset, dstScanStr, color,
		   x2, y2, x1, y1, cx, cy, cw, ch);
    } else {
        doDrawLine(dstBase + offset, dstScanStr, color,
		   x1, y1, x2, y2, cx, cy, cw, ch);
    }

    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
}

#define OUTCODE_TOP	1
#define OUTCODE_BOTTOM	2
#define OUTCODE_LEFT	4
#define OUTCODE_RIGHT	8

static int
outcode(jint x, jint y, jint xmin, jint ymin, jint xmax, jint ymax)
{
    int code;
    if (y < ymin) {
	code = OUTCODE_TOP;
    } else if (y > ymax) {
	code = OUTCODE_BOTTOM;
    } else {
	code = 0;
    }
    if (x < xmin) {
	code |= OUTCODE_LEFT;
    } else if (x > xmax) {
	code |= OUTCODE_RIGHT;
    }
    return code;
}

static jboolean
adjustLine(jint *px1, jint *py1, jint *px2, jint *py2,
	jint cxmin, jint cymin, jint cxmax, jint cymax)
{
    jint x1 = *px1;
    jint y1 = *py1;
    jint x2 = *px2;
    jint y2 = *py2;
    jboolean clipped;

    clipped = (x1 < cxmin || y1 < cymin || x1 > cxmax || y1 > cymax ||
	       x2 < cxmin || y2 < cymin || x2 > cxmax || y2 > cymax);

    if (x1 == x2) {
	if (x1 < cxmin || x1 > cxmax) {
	    return JNI_FALSE;
	}
	if (y1 > y2) {
	    int t = y1;
	    y1 = y2;
	    y2 = t;
	}
	if (y1 < cymin) {
	    y1 = cymin;
	}
	if (y2 > cymax) {
	    y2 = cymax;
	}
	if (y1 > y2) {
	    return JNI_FALSE;
	}
	*py1 = y1;
	*py2 = y2;
    } else if (y1 == y2) {
	if (y1 < cymin || y1 > cymax) {
	    return JNI_FALSE;
	}
	if (x1 > x2) {
	    int t = x1;
	    x1 = x2;
	    x2 = t;
	}
	if (x1 < cxmin) {
	    x1 = cxmin;
	}
	if (x2 > cxmax) {
	    x2 = cxmax;
	}
	if (x1 > x2) {
	    return JNI_FALSE;
	}
	*px1 = x1;
	*px2 = x2;
    } else {
	/* REMIND: This could overflow... */
	int outcode1, outcode2;
	int dx = x2 - x1;
	int dy = y2 - y1;
	int ax = (dx < 0) ? -dx : dx;
	int ay = (dy < 0) ? -dy : dy;
	int xmajor = (ax >= ay);

	outcode1 = outcode(x1, y1, cxmin, cymin, cxmax, cymax);
	outcode2 = outcode(x2, y2, cxmin, cymin, cxmax, cymax);
	while ((outcode1 | outcode2) != 0) {
	    int xsteps, ysteps;
	    if ((outcode1 & outcode2) != 0) {
		return JNI_FALSE;
	    }
	    if (outcode1 != 0) {
		if (outcode1 & (OUTCODE_TOP | OUTCODE_BOTTOM)) {
		    if (outcode1 & OUTCODE_TOP) {
			y1 = cymin;
		    } else {
			y1 = cymax;
		    }
		    ysteps = y1 - *py1;
		    if (ysteps < 0) {
			ysteps = -ysteps;
		    }
		    xsteps = 2 * ysteps * ax + ay;
		    if (xmajor) {
			xsteps += ay - ax - 1;
		    }
		    xsteps = xsteps / (2 * ay);
		    if (dx < 0) {
			xsteps = -xsteps;
		    }
		    x1 = *px1 + xsteps;
		} else if (outcode1 & (OUTCODE_LEFT | OUTCODE_RIGHT)) {
		    if (outcode1 & OUTCODE_LEFT) {
			x1 = cxmin;
		    } else {
			x1 = cxmax;
		    }
		    xsteps = x1 - *px1;
		    if (xsteps < 0) {
			xsteps = -xsteps;
		    }
		    ysteps = 2 * xsteps * ay + ax;
		    if (!xmajor) {
			ysteps += ax - ay - 1;
		    }
		    ysteps = ysteps / (2 * ax);
		    if (dy < 0) {
			ysteps = -ysteps;
		    }
		    y1 = *py1 + ysteps;
		}
		outcode1 = outcode(x1, y1, cxmin, cymin, cxmax, cymax);
	    } else {
		if (outcode2 & (OUTCODE_TOP | OUTCODE_BOTTOM)) {
		    if (outcode2 & OUTCODE_TOP) {
			y2 = cymin;
		    } else {
			y2 = cymax;
		    }
		    ysteps = y2 - *py2;
		    if (ysteps < 0) {
			ysteps = -ysteps;
		    }
		    xsteps = 2 * ysteps * ax + ay;
		    if (xmajor) {
			xsteps += ay - ax;
		    } else {
			xsteps -= 1;
		    }
		    xsteps = xsteps / (2 * ay);
		    if (dx > 0) {
			xsteps = -xsteps;
		    }
		    x2 = *px2 + xsteps;
		} else if (outcode2 & (OUTCODE_LEFT | OUTCODE_RIGHT)) {
		    if (outcode2 & OUTCODE_LEFT) {
			x2 = cxmin;
		    } else {
			x2 = cxmax;
		    }
		    xsteps = x2 - *px2;
		    if (xsteps < 0) {
			xsteps = -xsteps;
		    }
		    ysteps = 2 * xsteps * ay + ax;
		    if (xmajor) {
			ysteps -= 1;
		    } else {
			ysteps += ax - ay;
		    }
		    ysteps = ysteps / (2 * ax);
		    if (dy > 0) {
			ysteps = -ysteps;
		    }
		    y2 = *py2 + ysteps;
		}
		outcode2 = outcode(x2, y2, cxmin, cymin, cxmax, cymax);
	    }
	}
	*px1 = x1;
	*py1 = y1;
	*px2 = x2;
	*py2 = y2;
    }

    return JNI_TRUE;
}

static void
doDrawLine(jint *dstBase, jint dstScanStr, jint color,
		jint origx1, jint origy1, jint origx2, jint origy2,
		jint cx, jint cy, jint cw, jint ch)
{
    jint x1 = origx1;
    jint y1 = origy1;
    jint x2 = origx2;
    jint y2 = origy2;
    unsigned int *dstPtr;

    if (cw <= 0 || ch <= 0) {
	return;
    }

    if (!adjustLine(&x1, &y1, &x2, &y2, cx, cy, cx + cw - 1, cy + ch - 1)) {
	return;
    }

    dstPtr = (unsigned int *) dstBase + y1 * dstScanStr + x1;

    /* this could be made smaller, more elegant, more traditional. */
    if (x1 == x2) {
        if (y1 > y2) {
            do {
                *dstPtr = color;
                y1--;
                dstPtr -= dstScanStr;
            } while (y1 >= y2);
        } else {
            do {
                *dstPtr = color;
                y1++;
                dstPtr += dstScanStr;
            } while (y2 >= y1);
        }
    } else if (y1 == y2) {
        if (x1 > x2) {
            unsigned int *endPtr = dstPtr - (x1 - x2);
            do {
                *dstPtr-- = color;
            } while (dstPtr >= endPtr);
        } else {
            unsigned int *endPtr = dstPtr + (x2 - x1);
            do {
                *dstPtr++ = color;
            } while (dstPtr <= endPtr);
        }
    } else {
	int dx = origx2 - origx1;
	int dy = origy2 - origy1;
	int ax = (dx<0)? -dx: dx;
	int ay = (dy<0)? -dy: dy;
	int steps;
	int bumpmajor;
	int bumpminor;
	int errminor;
	int errmajor;
	int error;

	/* REMIND could be about 4x faster with multi-pixel algorithms */
	if (ax >= ay) {
	    /* x is dominant */
	    errmajor = ay * 2;
	    errminor = ax * 2;
	    bumpmajor = (dx < 0) ? -1 : 1;
	    bumpminor = (dy < 0) ? -dstScanStr : dstScanStr;
	    error = - (errminor / 2);
	    ax = -ax; /* For clipping adjustment below */
	    steps = x2 - x1;
	} else {
	    /* y is dominant */
	    errmajor = ax * 2;
	    errminor = ay * 2;
	    bumpmajor = (dy < 0) ? -dstScanStr : dstScanStr;
	    bumpminor = (dx < 0) ? -1 : 1;
	    ay = -ay; /* For clipping adjustment below */
	    steps = y2 - y1;
	}
	error = - (errminor / 2);
	if (y1 != origy1) {
	    int ysteps = y1 - origy1;
	    if (ysteps < 0) {
		ysteps = -ysteps;
	    }
	    error += ysteps * ax * 2;
	}
	if (x1 != origx1) {
	    int xsteps = x1 - origx1;
	    if (xsteps < 0) {
		xsteps = -xsteps;
	    }
	    error += xsteps * ay * 2;
	}
	if (steps < 0) {
	    steps = -steps;
	}
	do {
	    *dstPtr = color;
	    dstPtr += bumpmajor;
	    error += errmajor;
	    if (error >= 0) {
		dstPtr += bumpminor;
		error -= errminor;
	    }
	} while (--steps >= 0);
    }
}

JNIEXPORT void JNICALL
Java_sun_java2d_loops_IntDiscreteRenderer_devSetRect(
     JNIEnv *env, jclass clazz, jint color,
     jint x, jint y, jint w, jint h,
     jint cx, jint cy, jint cw, jint ch,
     jint offset, jintArray dst, jint pixStr, jint scanStr)
{
    jint *base;
    unsigned int *dataPtr;

    w += x;
    if (x < cx) {
	x = cx;
    }
    if (w > cx + cw) {
	w = cx + cw;
    }
    w -= x;
    h += y;
    if (y < cy) {
	y = cy;
    }
    if (h > cy + ch) {
	h = cy + ch;
    }
    h -= y;
    if (w <= 0 || h <= 0) {
	return;
    }

    base = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (base == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }
    offset += x + y*scanStr;
    dataPtr = (unsigned int *)base + offset;
    for (y = 0 ; y < h ; y++) {
	for (x = 0 ; x < w ; x++) {
	    dataPtr[x] = color;
	}
	dataPtr += scanStr;
    }

    (*env)->ReleasePrimitiveArrayCritical(env, dst, base, 0);
}

JNIEXPORT void JNICALL
Java_sun_java2d_loops_IntDiscreteRenderer_devSetSpans(
     JNIEnv *env, jclass clazz, jint color,
     jobject si, jlong pIterator,
     jint offset, jintArray dst, jint pixStr, jint scanStr)
{
    SpanIteratorFuncs *pFuncs = (SpanIteratorFuncs *) jlong_to_ptr(pIterator);
    void *srData;
    jint *base;
    jint spanbox[4];

    if (pFuncs == NULL) {
	JNU_ThrowNullPointerException(env, "native iterator not supplied");
	return;
    }

    srData = (*pFuncs->open)(env, si);

    base = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (base == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }
    base += offset;
    while ((*pFuncs->nextSpan)(env, srData, spanbox)) {
	jint x = spanbox[0];
	jint y = spanbox[1];
	jint w = spanbox[2] - x;
	jint h = spanbox[3] - y;
	jint *dataPtr = base + y*scanStr + x;
	do {
	    x = 0;
	    do {
		dataPtr[x++] = color;
	    } while (x < w);
	    dataPtr += scanStr;
	} while (--h > 0);
    }
    base -= offset;

    (*env)->ReleasePrimitiveArrayCritical(env, dst, base, 0);
    (*pFuncs->close)(env, srData);
}

typedef struct edgeEntry {
    int x1,y1;
    int x2,y2;
    /* bresenham coefficients for edge tracing */
    int error;
    int errx;
    int erry;
    char bumpx;
    char flipped;
    char pad2,pad3;
} EdgeEntry;

/* returns JNI_FALSE if edge is to be ignored */
static int initEdgeEntry(EdgeEntry *theEdge, int x1, int y1, int x2, int y2)
{
    int retval = JNI_FALSE;
    if ((y1 != y2) ) {
        theEdge->flipped = JNI_FALSE;
        /* edge is not horizontal, or is a point */
        if (y1 > y2) {
            /* 2nd point is higher, make it first */
            theEdge->flipped = JNI_TRUE;
            theEdge->x1 = x2;
            theEdge->y1 = y2;
            theEdge->x2 = x1;
            theEdge->y2 = y1;
        } else {
            /* 1st point is higher, keep it first */
            theEdge->x1 = x1;
            theEdge->y1 = y1;
            theEdge->x2 = x2;
            theEdge->y2 = y2;
        }
        {            
            int dx = theEdge->x2 - theEdge->x1;
            int dy = theEdge->y2 - theEdge->y1;
            int ax = (dx<0)? -dx: dx;
            int ay = (dy<0)? -dy: dy;
            if (dx < 0) {
                theEdge->bumpx = -1;
		theEdge->error = -ay;
            } else {
                theEdge->bumpx = 1;
		theEdge->error = 0;
            }
            theEdge->errx = ax;
            theEdge->erry = ay;
        }
        retval = JNI_TRUE;
    }
    return retval;
}


/* return number of edges actually retained for scanning algorithm */
static int fillEdgeTable(EdgeEntry ***edgeTable, EdgeEntry **edgeEntries,
			 jint *xpointsPtr, jint *ypointsPtr,
                         jint *npointsPtr, int npolygons,
			 int originX, int originY)
{
    int usedEdges = 0;
    jint *xpointsCursor = xpointsPtr;
    jint *ypointsCursor = ypointsPtr;
    int i;

    int totalPoints = npolygons;
    for (i=0; i<npolygons; i++) {
        totalPoints += npointsPtr[i];
    }
    *edgeEntries = (EdgeEntry *) calloc(totalPoints + npolygons,
                                        sizeof(EdgeEntry));
    *edgeTable = (EdgeEntry **) calloc(totalPoints + npolygons,
				        sizeof(EdgeEntry *));
    if (*edgeEntries && *edgeTable) {
        EdgeEntry **edgeTableCursor = *edgeTable;
        EdgeEntry *edgeEntriesCursor = *edgeEntries;
        for (i=0; i<npolygons; i++) {
            int x1, y1, x2, y2;
            int npoints = npointsPtr[i];
            x1 = xpointsCursor[0] + originX;
            y1 = ypointsCursor[0] + originY;
            x2 = xpointsCursor[npoints-1] + originX;
            y2 = ypointsCursor[npoints-1] + originY;
            if ((x1 != x2) || (y1 != y2)) {
                /* add closing edge */
                if (initEdgeEntry(edgeEntriesCursor,x2, y2,x1, y1)) {
                    *edgeTableCursor++ = edgeEntriesCursor;
                    edgeEntriesCursor++;
                    usedEdges++;
                }
            }
            while (npoints-- > 1) {
	        x2 = xpointsCursor[1] + originX;
	        y2 = ypointsCursor[1] + originY;
                if (initEdgeEntry(edgeEntriesCursor, x1, y1, x2, y2)) {
                    *edgeTableCursor++ = edgeEntriesCursor;
                    edgeEntriesCursor++;
                    usedEdges++;
                }
                xpointsCursor++;
                ypointsCursor++;
	        x1 = x2;
	        y1 = y2;
            }
            xpointsCursor++;
            ypointsCursor++;
        }
    }
    return usedEdges;
}

/*
* REMIND: CDECL needed for WIN32 "qsort"
*/

#ifdef _WIN32
#if !defined(CDECL)
	#define CDECL __cdecl
#endif
#else
#define CDECL
#endif

static int CDECL sortByHighestVertex(const void *elem1, const void *elem2 )
{
    const EdgeEntry *edge1 = *((const EdgeEntry **)elem1);
    const EdgeEntry *edge2 = *((const EdgeEntry **)elem2);

    if (edge1->y1 < edge2->y1) {
	return -1;
    }
    if (edge1->y1 > edge2->y1) {
	return 1;
    }
    if (edge1->x1 < edge2->x1) {
	return -1;
    }
    if (edge1->x1 > edge2->x1) {
	return 1;
    }
    if (edge1->y2 < edge2->y2) {
	return -1;
    }
    if (edge1->y2 > edge2->y2) {
	return 1;
    }
    if (edge1->x2 < edge2->x2) {
	return -1;
    }
    if (edge1->x2 > edge2->x2) {
	return 1;
    }
    return 0;
}

static int CDECL sortByLeftMostVertex(const void *elem1, const void *elem2 )
{
    const EdgeEntry *edge1 = *((const EdgeEntry **)elem1);
    const EdgeEntry *edge2 = *((const EdgeEntry **)elem2);

    if (edge1->x1 < edge2->x1) {
	return -1;
    }
    if (edge1->x1 > edge2->x1) {
	return 1;
    }
    return 0;
}

static void XSortActiveEdges(int activeEdges,EdgeEntry **edgeTable)
{
    if (activeEdges) {
        qsort(edgeTable, activeEdges, sizeof(EdgeEntry *),sortByLeftMostVertex);
    }
}

static void DrawEvenOddFromActiveEdges
    (jint *dstBase, jint dstScanStr, jint color,
     int activeEdges,EdgeEntry **edgeTable,
     int cx,int cy,int cw,int ch)
{
    while (activeEdges >= 2) {
        EdgeEntry *edge = *edgeTable;
        int y = edge->y1;
        int x1 = edge->x1;
        edgeTable++;
        edge = *edgeTable;
        {
            int x2 = edge->x1;
            edgeTable++;
            if (x2 > x1) {
                /* do not draw rightmost pixel of any segment */
                doDrawLine(dstBase, dstScanStr, color,
			   x1, y, x2-1, y,
			   cx, cy, cw, ch);
            }
        }
        activeEdges -= 2;
    }
}

static void DrawWindingFromActiveEdges
    (jint *dstBase, jint dstScanStr, jint color,
     int activeEdges,EdgeEntry **edgeTable,
     int cx,int cy,int cw,int ch)
{
    int wound = 0;
    int x1 = -1;
    int y1 = -1;
    while (activeEdges) {
        EdgeEntry *edge = *edgeTable;
        if (0 == wound) {
            x1 = edge->x1;
            y1 = edge->y1;
        }
        wound += edge->flipped? -1: 1;
        if (0 == wound) {
            int x2 = edge->x1;
            if (x2 > x1) {
                /* do not draw rightmost pixel of any segment */
                doDrawLine(dstBase, dstScanStr, color,
			   x1, y1, x2-1, y1,
			   cx, cy, cw, ch);
            }
        }
        activeEdges--;
        edgeTable++;
    }
}

static void DrawFromActiveEdges
    (jint *dstBase, jint dstScanStr, jint color,
     int activeEdges,EdgeEntry **edgeTable,
     int cx,int cy,int cw,int ch,
     int fEvenOdd)
{
    if (fEvenOdd) {
        DrawEvenOddFromActiveEdges(dstBase, dstScanStr, color,
				   activeEdges, edgeTable,
				   cx, cy, cw, ch);
    } else {
        DrawWindingFromActiveEdges(dstBase, dstScanStr, color,
				   activeEdges, edgeTable,
				   cx, cy, cw, ch);
    }
}

static void PurgeUsedEdges(int *countEdges, int *activeEdges,
			   EdgeEntry ***edgeTable)
{
    if (*activeEdges) {
        int edgesToCheck = *activeEdges;
        EdgeEntry **currentTableEntry = *edgeTable;
        while (edgesToCheck--) {
            if ((*currentTableEntry)->y1 >= (*currentTableEntry)->y2) {
                /* edge is entirely used up - nothing more to draw */
                if (currentTableEntry != *edgeTable) {
                    *currentTableEntry = **edgeTable;
                }
                (*edgeTable)++;
                (*countEdges)--;
                (*activeEdges)--;
            }
            currentTableEntry++;
        }
    }
}

static void ActivateNewEdges(int countEdges, int *activeEdges,
			     EdgeEntry **edgeTable)
{
    if (countEdges) {
        int y = edgeTable[0]->y1;
        while (*activeEdges < countEdges) {
            if (edgeTable[*activeEdges]->y1 != y) break;
            (*activeEdges)++;
        }
    }
}

static void AdvanceActiveEdges(int activeEdges,EdgeEntry **edgeTable) {
    EdgeEntry **edgeCursor = edgeTable;
    while (activeEdges--) {
        EdgeEntry *edge = *edgeCursor;
        edge->y1++;
        if (edge->erry) {
	    edge->error += edge->errx;
	    while (edge->error > 0) {
		edge->x1 += edge->bumpx;
		edge->error -= edge->erry;
	    }
        }
#if 0
        /* X-sorting on the fly */
        {
            EdgeEntry **edgeCursor2 = edgeCursor;
            EdgeEntry *myEdge = *edgeCursor2;
            int myX = myEdge->x1;
            while (--edgeCursor2 >= edgeTable) {
                if (myX >= (*edgeCursor2)->x1) {
                    break;
                }
                edgeCursor2[1] = edgeCursor2[0];
                edgeCursor2[0] = myEdge;
            }
        }
#endif
        edgeCursor++;
    }
}

static void RemoveDuplicateEdges(EdgeEntry **edgeTablePtr,int *countEdges)
{
    EdgeEntry **currentTableEntry;
    EdgeEntry **lastTableEntry;
    EdgeEntry **dstTableEntry;

    if (!*countEdges) {
        return;
    }
    currentTableEntry = edgeTablePtr;
    lastTableEntry = edgeTablePtr + (*countEdges);
    dstTableEntry = edgeTablePtr;
    while (currentTableEntry < lastTableEntry) {
        if (currentTableEntry + 1 < lastTableEntry
		 && (currentTableEntry[0]->y1 == currentTableEntry[1]->y1)
                 && (currentTableEntry[0]->x1 == currentTableEntry[1]->x1)
                 && (currentTableEntry[0]->y2 == currentTableEntry[1]->y2)
                 && (currentTableEntry[0]->x2 == currentTableEntry[1]->x2)
                 && (currentTableEntry[0]->flipped
                    != currentTableEntry[1]->flipped)) {
            /* skip these two reversed edges */
            currentTableEntry += 2;
            *countEdges -= 2;
            continue;
        }
        if (dstTableEntry != currentTableEntry) {
	    *dstTableEntry = *currentTableEntry;
        }
        dstTableEntry++;
        currentTableEntry++;
    }
}

static int doPolyFillMany(
     JNIEnv *env, jint *dstBase, jint dstPixStr, jint dstScanStr, jint color,
     jint *xpointsPtr, jint *ypointsPtr, jint *npointsPtr, int npolygons,
     int originX, int originY, int cx, int cy, int cw, int ch,
     int fEvenOdd)
{
    int retval = JNI_FALSE;
    EdgeEntry *edgeEntries = NULL;
    EdgeEntry **edgeTable = NULL;
    int countEdges;

    countEdges = fillEdgeTable(&edgeTable, &edgeEntries,
			       xpointsPtr, ypointsPtr,
			       npointsPtr, npolygons,
			       originX, originY);
    if (edgeTable && edgeEntries) {
	retval = JNI_TRUE;
	if (countEdges) {
	    EdgeEntry **edgeTablePtr = edgeTable;
	    int activeEdges = 0;
	    /*
	     * sort all edges in the edgeTable into acsending order
	     * of the highest vertices
	     */
	    qsort(edgeTablePtr, countEdges, sizeof(EdgeEntry *),
		  sortByHighestVertex);
	    RemoveDuplicateEdges(edgeTablePtr,&countEdges);
	    while (countEdges) {
		XSortActiveEdges(activeEdges, edgeTablePtr);
		DrawFromActiveEdges(dstBase, dstScanStr, color,
				    activeEdges, edgeTablePtr,
				    cx, cy, cw, ch, fEvenOdd);
		AdvanceActiveEdges(activeEdges, edgeTablePtr);
		ActivateNewEdges(countEdges, &activeEdges, edgeTablePtr);
		/* eliminates bottom-most pixel-row of each edge */
		PurgeUsedEdges(&countEdges, &activeEdges, &edgeTablePtr);
	    }
	}
    }

    if (edgeEntries) {
	free(edgeEntries);
    }
    if (edgeTable) {
	free(edgeTable);
    }

    return retval;
}

/*
 * Class:     sun_java2d_loops_IntDiscreteRenderer
 * Method:    devFillPolygons
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_IntDiscreteRenderer_devFillPolygons
    (JNIEnv *env, jobject clazz, jint color,
    jintArray xpoints, jintArray ypoints, jintArray npoints, jint numpolys,
    jboolean fEvenOdd, jint originX, jint originY,
    jint cx, jint cy, jint cw, jint ch,
    jint offset, jintArray dst, jint dstPixStr, jint dstScanStr)
{
    int npointsLen, xpointsLen, ypointsLen;
    jint *npointsPtr, *xpointsPtr, *ypointsPtr;
    jint *dstBase;
    int i, pointsNeeded, result;

    if (dst == NULL) {
	JNU_ThrowNullPointerException(env, "destination array");
	return;
    }
    if (xpoints == NULL || ypoints == NULL) {
	JNU_ThrowNullPointerException(env, "coordinate array");
	return;
    }
    if (npoints == NULL) {
	JNU_ThrowNullPointerException(env, "polygon length array");
	return;
    }

    npointsLen = (*env)->GetArrayLength(env, npoints);
    xpointsLen = (*env)->GetArrayLength(env, xpoints);
    ypointsLen = (*env)->GetArrayLength(env, ypoints);

    if (npointsLen < numpolys) {
	JNU_ThrowArrayIndexOutOfBoundsException(env,
						"polygon length array size");
	return;
    }

    npointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, npoints, 0);
    if (!npointsPtr) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    pointsNeeded = 0;
    for (i=0; i<numpolys; i++) {
	pointsNeeded += npointsPtr[i];
    }

    if (ypointsLen < pointsNeeded || xpointsLen < pointsNeeded) {
	(*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);
	JNU_ThrowArrayIndexOutOfBoundsException(env,
						"coordinate array length");
	return;
    }

    xpointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, xpoints, 0);
    ypointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, ypoints, 0);
    dstBase = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (!xpointsPtr || !ypointsPtr || !dstBase) {
	if (dstBase) {
	    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
	}
	if (ypointsPtr) {
	    (*env)->ReleasePrimitiveArrayCritical(env, ypoints, ypointsPtr, 0);
	}
	if (xpointsPtr) {
	    (*env)->ReleasePrimitiveArrayCritical(env, xpoints, xpointsPtr, 0);
	}
	/* We know that npointsPtr succeeded - it was tested above... */
	(*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    result = doPolyFillMany(env, dstBase + offset, dstPixStr, dstScanStr, color,
			    xpointsPtr, ypointsPtr, npointsPtr, numpolys,
			    originX, originY, cx, cy, cw, ch, fEvenOdd);

    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, ypoints, ypointsPtr, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, xpoints, xpointsPtr, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);

    if (!result) {
	JNU_ThrowOutOfMemoryError(env, "fill poly");
	return;
    }
}

static int doPolyDrawMany(
     JNIEnv *env, jint *dstBase, jint dstPixStr, jint dstScanStr, jint color,
     jint *xpointsPtr, jint *ypointsPtr, jint *npointsPtr, int npolygons,
     int originX, int originY, int cx, int cy, int cw, int ch,
     int fClose)
{
    int retval = JNI_TRUE;
    while (npolygons-- > 0) {
        jint numPts = *npointsPtr++;
        if (numPts) {
            jint xStart = *xpointsPtr;
            jint yStart = *ypointsPtr;
            while (--numPts > 0) {
                doDrawLine(dstBase, dstScanStr,
	                   color,
                           *xpointsPtr, *ypointsPtr,
                           xpointsPtr[1], ypointsPtr[1],
	                   cx, cy, cw, ch);
                xpointsPtr++;
                ypointsPtr++;
            }
            if (fClose) {
                doDrawLine(dstBase, dstScanStr,
	                   color,
                           *xpointsPtr, *ypointsPtr,
                           xStart, yStart,
	                   cx, cy, cw, ch);
            }
	    xpointsPtr++;
	    ypointsPtr++;
        }
    }
    return retval;
}

/*
 * Class:     sun_java2d_loops_IntDiscreteRenderer
 * Method:    devDrawPolygons
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_IntDiscreteRenderer_devDrawPolygons
    (JNIEnv *env, jobject clazz, jint color,
    jintArray xpoints, jintArray ypoints, jintArray npoints, jint numpolys,
    jboolean fClose, jint originX, jint originY,
    jint cx, jint cy, jint cw, jint ch,
    jint offset, jintArray dst, jint dstPixStr, jint dstScanStr)
{
    int npointsLen, xpointsLen, ypointsLen;
    jint *npointsPtr, *xpointsPtr, *ypointsPtr;
    jint *dstBase;
    int i, pointsNeeded, result;

    if (dst == NULL) {
	JNU_ThrowNullPointerException(env, "destination array");
	return;
    }
    if (xpoints == NULL || ypoints == NULL) {
	JNU_ThrowNullPointerException(env, "coordinate array");
	return;
    }
    if (npoints == NULL) {
	JNU_ThrowNullPointerException(env, "polygon length array");
	return;
    }

    npointsLen = (*env)->GetArrayLength(env, npoints);
    xpointsLen = (*env)->GetArrayLength(env, xpoints);
    ypointsLen = (*env)->GetArrayLength(env, ypoints);

    if (npointsLen < numpolys) {
	JNU_ThrowArrayIndexOutOfBoundsException(env,
						"polygon length array size");
	return;
    }

    npointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, npoints, 0);
    if (!npointsPtr) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    pointsNeeded = 0;
    for (i=0; i<numpolys; i++) {
	pointsNeeded += npointsPtr[i];
    }

    if (ypointsLen < pointsNeeded || xpointsLen < pointsNeeded) {
	(*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);
	JNU_ThrowArrayIndexOutOfBoundsException(env,
						"coordinate array length");
	return;
    }

    xpointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, xpoints, 0);
    ypointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, ypoints, 0);
    dstBase = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (!xpointsPtr || !ypointsPtr || !dstBase) {
	if (dstBase) {
	    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
	}
	if (ypointsPtr) {
	    (*env)->ReleasePrimitiveArrayCritical(env, ypoints, ypointsPtr, 0);
	}
	if (xpointsPtr) {
	    (*env)->ReleasePrimitiveArrayCritical(env, xpoints, xpointsPtr, 0);
	}
	/* We know that npointsPtr succeeded - it was tested above... */
	(*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    result = doPolyDrawMany(env, dstBase + offset, dstPixStr, dstScanStr, color,
			    xpointsPtr, ypointsPtr, npointsPtr, numpolys,
			    originX, originY, cx, cy, cw, ch, fClose);

    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, ypoints, ypointsPtr, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, xpoints, xpointsPtr, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);

    if (!result) {
	JNU_ThrowOutOfMemoryError(env, "draw poly");
	return;
    }
}
