/*
 * @(#)ThreeByteRenderer.c	1.1 98/09/07
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <string.h>
#include <jni.h>
#include <jni_util.h>
#include <jlong.h>
#include <stdlib.h>

#include "SpanIterator.h"

#include "sun_java2d_loops_ThreeByteRenderer.h"

#include "BlitMacros.h"

static void
doDrawLine(jbyte *dstBase, jint dstScanStr,
	   jint color, jint x1, jint y1, jint x2, jint y2,
	   jint cx, jint cy, jint cw, jint ch);

JNIEXPORT void JNICALL
Java_sun_java2d_loops_ThreeByteRenderer_devDrawRect(
     JNIEnv *env, jclass clazz, jint color,
     jint x, jint y, jint w, jint h,
     jint cx, jint cy, jint cw, jint ch,
     jint offset, jbyteArray dst, jint dstScanStr)
{
    jbyte *dstBase;
    jint count;
    unsigned char *cursor;
    jint x1, y1, x2, y2;
    jbyte c1, c2, c3;

    x1 = x;
    y1 = y;
    x2 = x + w;
    y2 = y + h;
    if (x1 < cx) x1 = cx;
    if (y1 < cy) y1 = cy;
    if (x2 >= cx + cw) x2 = cx + cw - 1;
    if (y2 >= cy + ch) y2 = cy + ch - 1;
    if (x1 > x2 || y1 > y2) {
	return;
    }

    dstBase = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (dstBase == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
        return;
    }

    if (w == 0 || h == 0) {
	doDrawLine(dstBase + offset, dstScanStr, color,
		   x1, y1, x2, y2, cx, cy, cw, ch);

	(*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
	return;
    }

    c1 = (color >>  0) & 0xff;
    c2 = (color >>  8) & 0xff;
    c3 = (color >> 16) & 0xff;
    cursor = (unsigned char *)dstBase + offset + x1*3 + y1*dstScanStr;

    /* line 1 */
    if (y1 == y) {
	for (count = x1; count < x2; count++) {
	    *cursor++ = c1;
	    *cursor++ = c2;
	    *cursor++ = c3;
	}
    } else {
	if (x1 == x) {
	    /* shared pixel with line 4 */
	    cursor[0] = c1;
	    cursor[1] = c2;
	    cursor[2] = c3;
	}
	cursor += (x2 - x1) * 3;
    }
    /* line 2 */
    if (x2 == x + w) {
	for (count = y1; count < y2; count++) {
	    cursor[0] = c1;
	    cursor[1] = c2;
	    cursor[2] = c3;
	    cursor += dstScanStr;
	}
    } else {
	if (y1 == y) {
	    /* shared pixel with line 1 */
	    cursor[0] = c1;
	    cursor[1] = c2;
	    cursor[2] = c3;
	}
	cursor += dstScanStr * (y2 - y1);
    }
    /* line 3 */
    if (y2 == y + h) {
	cursor += 2;
	for (count = x2; count > x1; count--) {
	    *cursor-- = c3;
	    *cursor-- = c2;
	    *cursor-- = c1;
	}
	cursor -= 2;
    } else {
	if (x2 == x + w) {
	    /* shared pixel with line 2 */
	    cursor[0] = c1;
	    cursor[1] = c2;
	    cursor[2] = c3;
	}
	cursor -= (x2 - x1) * 3;
    }
    /* line 4 */
    if (x1 == x) {
	for (count = y2; count >= y1; count--) {
	    cursor[0] = c1;
	    cursor[1] = c2;
	    cursor[2] = c3;
	    cursor -= dstScanStr;
	}
    } else {
	if (y2 == y + h) {
	    /* shared pixel with line 3 */
	    cursor[0] = c1;
	    cursor[1] = c2;
	    cursor[2] = c3;
	}
	/* We don't need cursor any more anyway... */
	/* cursor -= dstScanStr * (y2 - y1 + 1); */
    }

    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
}

JNIEXPORT void JNICALL
Java_sun_java2d_loops_ThreeByteRenderer_devDrawLine(
     JNIEnv *env, jclass clazz, jint color,
     jint x1, jint y1, jint x2, jint y2,
     jint cx, jint cy, jint cw, jint ch,
     jint offset, jbyteArray dst, jint dstScanStr)
{
    jbyte *dstBase;

    dstBase = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (dstBase == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
        return;
    }

    if (y1 >= y2) {
        doDrawLine(dstBase + offset, dstScanStr, color,
		   x2, y2, x1, y1, cx, cy, cw, ch);
    } else {
        doDrawLine(dstBase + offset, dstScanStr, color,
		   x1, y1, x2, y2, cx, cy, cw, ch);
    }

    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
}

#define OUTCODE_TOP	1
#define OUTCODE_BOTTOM	2
#define OUTCODE_LEFT	4
#define OUTCODE_RIGHT	8

static int
outcode(jint x, jint y, jint xmin, jint ymin, jint xmax, jint ymax)
{
    int code;
    if (y < ymin) {
	code = OUTCODE_TOP;
    } else if (y > ymax) {
	code = OUTCODE_BOTTOM;
    } else {
	code = 0;
    }
    if (x < xmin) {
	code |= OUTCODE_LEFT;
    } else if (x > xmax) {
	code |= OUTCODE_RIGHT;
    }
    return code;
}

static jboolean
adjustLine(jint *px1, jint *py1, jint *px2, jint *py2,
	jint cxmin, jint cymin, jint cxmax, jint cymax)
{
    jint x1 = *px1;
    jint y1 = *py1;
    jint x2 = *px2;
    jint y2 = *py2;
    jboolean clipped;

    clipped = (x1 < cxmin || y1 < cymin || x1 > cxmax || y1 > cymax ||
	       x2 < cxmin || y2 < cymin || x2 > cxmax || y2 > cymax);

    if (x1 == x2) {
	if (x1 < cxmin || x1 > cxmax) {
	    return JNI_FALSE;
	}
	if (y1 > y2) {
	    int t = y1;
	    y1 = y2;
	    y2 = t;
	}
	if (y1 < cymin) {
	    y1 = cymin;
	}
	if (y2 > cymax) {
	    y2 = cymax;
	}
	if (y1 > y2) {
	    return JNI_FALSE;
	}
	*py1 = y1;
	*py2 = y2;
    } else if (y1 == y2) {
	if (y1 < cymin || y1 > cymax) {
	    return JNI_FALSE;
	}
	if (x1 > x2) {
	    int t = x1;
	    x1 = x2;
	    x2 = t;
	}
	if (x1 < cxmin) {
	    x1 = cxmin;
	}
	if (x2 > cxmax) {
	    x2 = cxmax;
	}
	if (x1 > x2) {
	    return JNI_FALSE;
	}
	*px1 = x1;
	*px2 = x2;
    } else {
	/* REMIND: This could overflow... */
	int outcode1, outcode2;
	int dx = x2 - x1;
	int dy = y2 - y1;
	int ax = (dx < 0) ? -dx : dx;
	int ay = (dy < 0) ? -dy : dy;
	int xmajor = (ax >= ay);

	outcode1 = outcode(x1, y1, cxmin, cymin, cxmax, cymax);
	outcode2 = outcode(x2, y2, cxmin, cymin, cxmax, cymax);
	while ((outcode1 | outcode2) != 0) {
	    int xsteps, ysteps;
	    if ((outcode1 & outcode2) != 0) {
		return JNI_FALSE;
	    }
	    if (outcode1 != 0) {
		if (outcode1 & (OUTCODE_TOP | OUTCODE_BOTTOM)) {
		    if (outcode1 & OUTCODE_TOP) {
			y1 = cymin;
		    } else {
			y1 = cymax;
		    }
		    ysteps = y1 - *py1;
		    if (ysteps < 0) {
			ysteps = -ysteps;
		    }
		    xsteps = 2 * ysteps * ax + ay;
		    if (xmajor) {
			xsteps += ay - ax - 1;
		    }
		    xsteps = xsteps / (2 * ay);
		    if (dx < 0) {
			xsteps = -xsteps;
		    }
		    x1 = *px1 + xsteps;
		} else if (outcode1 & (OUTCODE_LEFT | OUTCODE_RIGHT)) {
		    if (outcode1 & OUTCODE_LEFT) {
			x1 = cxmin;
		    } else {
			x1 = cxmax;
		    }
		    xsteps = x1 - *px1;
		    if (xsteps < 0) {
			xsteps = -xsteps;
		    }
		    ysteps = 2 * xsteps * ay + ax;
		    if (!xmajor) {
			ysteps += ax - ay - 1;
		    }
		    ysteps = ysteps / (2 * ax);
		    if (dy < 0) {
			ysteps = -ysteps;
		    }
		    y1 = *py1 + ysteps;
		}
		outcode1 = outcode(x1, y1, cxmin, cymin, cxmax, cymax);
	    } else {
		if (outcode2 & (OUTCODE_TOP | OUTCODE_BOTTOM)) {
		    if (outcode2 & OUTCODE_TOP) {
			y2 = cymin;
		    } else {
			y2 = cymax;
		    }
		    ysteps = y2 - *py2;
		    if (ysteps < 0) {
			ysteps = -ysteps;
		    }
		    xsteps = 2 * ysteps * ax + ay;
		    if (xmajor) {
			xsteps += ay - ax;
		    } else {
			xsteps -= 1;
		    }
		    xsteps = xsteps / (2 * ay);
		    if (dx > 0) {
			xsteps = -xsteps;
		    }
		    x2 = *px2 + xsteps;
		} else if (outcode2 & (OUTCODE_LEFT | OUTCODE_RIGHT)) {
		    if (outcode2 & OUTCODE_LEFT) {
			x2 = cxmin;
		    } else {
			x2 = cxmax;
		    }
		    xsteps = x2 - *px2;
		    if (xsteps < 0) {
			xsteps = -xsteps;
		    }
		    ysteps = 2 * xsteps * ay + ax;
		    if (xmajor) {
			ysteps -= 1;
		    } else {
			ysteps += ax - ay;
		    }
		    ysteps = ysteps / (2 * ax);
		    if (dy > 0) {
			ysteps = -ysteps;
		    }
		    y2 = *py2 + ysteps;
		}
		outcode2 = outcode(x2, y2, cxmin, cymin, cxmax, cymax);
	    }
	}
	*px1 = x1;
	*py1 = y1;
	*px2 = x2;
	*py2 = y2;
    }

    return JNI_TRUE;
}

static void
doDrawLine(jbyte *dstBase, jint dstScanStr, jint color,
	   jint origx1, jint origy1, jint origx2, jint origy2,
	   jint cx, jint cy, jint cw, jint ch)
{
    jint x1 = origx1;
    jint y1 = origy1;
    jint x2 = origx2;
    jint y2 = origy2;
    unsigned char *dstPtr;
    jbyte c1, c2, c3;

    if (cw <= 0 || ch <= 0) {
	return;
    }

    if (!adjustLine(&x1, &y1, &x2, &y2, cx, cy, cx + cw - 1, cy + ch - 1)) {
	return;
    }

    c1 = (color >>  0) & 0xff;
    c2 = (color >>  8) & 0xff;
    c3 = (color >> 16) & 0xff;
    dstPtr = (unsigned char *) dstBase + y1 * dstScanStr + x1 * 3;

    /* this could be made smaller, more elegant, more traditional. */
    if (x1 == x2) {
        if (y1 > y2) {
            do {
		dstPtr[0] = c1;
		dstPtr[1] = c2;
		dstPtr[2] = c3;
                y1--;
                dstPtr -= dstScanStr;
            } while (y1 >= y2);
        } else {
            do {
		dstPtr[0] = c1;
		dstPtr[1] = c2;
		dstPtr[2] = c3;
                y1++;
                dstPtr += dstScanStr;
            } while (y2 >= y1);
        }
    } else if (y1 == y2) {
        if (x1 > x2) {
            unsigned char *endPtr = dstPtr - (x1 - x2) * 3;
	    dstPtr += 2;
            do {
		*dstPtr-- = c3;
		*dstPtr-- = c2;
		*dstPtr-- = c1;
            } while (dstPtr >= endPtr);
        } else {
            unsigned char *endPtr = dstPtr + (x2 - x1) * 3;
            do {
		*dstPtr++ = c1;
		*dstPtr++ = c2;
		*dstPtr++ = c3;
            } while (dstPtr <= endPtr);
        }
    } else {
	int dx = origx2 - origx1;
	int dy = origy2 - origy1;
	int ax = (dx<0)? -dx: dx;
	int ay = (dy<0)? -dy: dy;
	int steps;
	int bumpmajor;
	int bumpminor;
	int errminor;
	int errmajor;
	int error;

	/* REMIND could be about 4x faster with multi-pixel algorithms */
	if (ax >= ay) {
	    /* x is dominant */
	    errmajor = ay * 2;
	    errminor = ax * 2;
	    bumpmajor = (dx < 0) ? -3 : 3;
	    bumpminor = (dy < 0) ? -dstScanStr : dstScanStr;
	    error = - (errminor / 2);
	    ax = -ax; /* For clipping adjustment below */
	    steps = x2 - x1;
	} else {
	    /* y is dominant */
	    errmajor = ax * 2;
	    errminor = ay * 2;
	    bumpmajor = (dy < 0) ? -dstScanStr : dstScanStr;
	    bumpminor = (dx < 0) ? -3 : 3;
	    ay = -ay; /* For clipping adjustment below */
	    steps = y2 - y1;
	}
	error = - (errminor / 2);
	if (y1 != origy1) {
	    int ysteps = y1 - origy1;
	    if (ysteps < 0) {
		ysteps = -ysteps;
	    }
	    error += ysteps * ax * 2;
	}
	if (x1 != origx1) {
	    int xsteps = x1 - origx1;
	    if (xsteps < 0) {
		xsteps = -xsteps;
	    }
	    error += xsteps * ay * 2;
	}
	if (steps < 0) {
	    steps = -steps;
	}
	do {
	    dstPtr[0] = c1;
	    dstPtr[1] = c2;
	    dstPtr[2] = c3;
	    dstPtr += bumpmajor;
	    error += errmajor;
	    if (error >= 0) {
		dstPtr += bumpminor;
		error -= errminor;
	    }
	} while (--steps >= 0);
    }
}

JNIEXPORT void JNICALL
Java_sun_java2d_loops_ThreeByteRenderer_devSetRect(
     JNIEnv *env, jclass clazz, jint color,
     jint x, jint y, jint w, jint h,
     jint cx, jint cy, jint cw, jint ch,
     jint offset, jbyteArray dst, jint scanStr)
{
    jbyte *base;
    unsigned char *dataPtr;
    jbyte c1, c2, c3;

    w += x;
    if (x < cx) {
	x = cx;
    }
    if (w > cx + cw) {
	w = cx + cw;
    }
    w -= x;
    h += y;
    if (y < cy) {
	y = cy;
    }
    if (h > cy + ch) {
	h = cy + ch;
    }
    h -= y;
    if (w <= 0 || h <= 0) {
	return;
    }

    base = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (base == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    c1 = (color >>  0) & 0xff;
    c2 = (color >>  8) & 0xff;
    c3 = (color >> 16) & 0xff;
    w *= 3;
    dataPtr = (unsigned char *)base + offset + y*scanStr + x*3;
    for (y = 0 ; y < h ; y++) {
	x = 0;
	do {
	    dataPtr[x++] = c1;
	    dataPtr[x++] = c2;
	    dataPtr[x++] = c3;
	} while (x < w);
	dataPtr += scanStr;
    }

    (*env)->ReleasePrimitiveArrayCritical(env, dst, base, 0);
}

JNIEXPORT void JNICALL
Java_sun_java2d_loops_ThreeByteRenderer_devSetSpans(
     JNIEnv *env, jclass clazz, jint color,
     jobject si, jlong pIterator,
     jint offset, jbyteArray dst, jint scanStr)
{
    SpanIteratorFuncs *pFuncs = (SpanIteratorFuncs *) jlong_to_ptr(pIterator);
    void *srData;
    jbyte *base;
    jint spanbox[4];
    jbyte c1, c2, c3;

    if (pFuncs == NULL) {
	JNU_ThrowNullPointerException(env, "native iterator not supplied");
	return;
    }

    srData = (*pFuncs->open)(env, si);

    base = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (base == NULL) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    c1 = (color >>  0) & 0xff;
    c2 = (color >>  8) & 0xff;
    c3 = (color >> 16) & 0xff;
    base += offset;
    while ((*pFuncs->nextSpan)(env, srData, spanbox)) {
	jint x = spanbox[0];
	jint y = spanbox[1];
	jint w = spanbox[2] - x;
	jint h = spanbox[3] - y;
	jbyte *dataPtr = base + y*scanStr + x * 3;
	w *= 3;
	do {
	    x = 0;
	    do {
		dataPtr[x++] = c1;
		dataPtr[x++] = c2;
		dataPtr[x++] = c3;
	    } while (x < w);
	    dataPtr += scanStr;
	} while (--h > 0);
    }
    base -= offset;

    (*env)->ReleasePrimitiveArrayCritical(env, dst, base, 0);
    (*pFuncs->close)(env, srData);
}

static int doPolyDrawMany(
     JNIEnv *env, jbyte *dstBase, jint dstScanStr, jint color,
     jint *xpointsPtr, jint *ypointsPtr, jint *npointsPtr, int npolygons,
     int originX, int originY, int cx, int cy, int cw, int ch,
     int fClose)
{
    int retval = JNI_TRUE;
    while (npolygons-- > 0) {
        jint numPts = *npointsPtr++;
        if (numPts) {
            jint xStart = *xpointsPtr;
            jint yStart = *ypointsPtr;
            while (--numPts > 0) {
                doDrawLine(dstBase, dstScanStr,
	                   color,
                           *xpointsPtr, *ypointsPtr,
                           xpointsPtr[1], ypointsPtr[1],
	                   cx, cy, cw, ch);
                xpointsPtr++;
                ypointsPtr++;
            }
            if (fClose) {
                doDrawLine(dstBase, dstScanStr,
	                   color,
                           *xpointsPtr, *ypointsPtr,
                           xStart, yStart,
	                   cx, cy, cw, ch);
            }
	    xpointsPtr++;
	    ypointsPtr++;
        }
    }
    return retval;
}

/*
 * Class:     sun_java2d_loops_ThreeByteRenderer
 * Method:    devDrawPolygons
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_ThreeByteRenderer_devDrawPolygons
    (JNIEnv *env, jobject clazz, jint color,
    jintArray xpoints, jintArray ypoints, jintArray npoints, jint numpolys,
    jboolean fClose, jint originX, jint originY,
    jint cx, jint cy, jint cw, jint ch,
    jint offset, jbyteArray dst, jint dstScanStr)
{
    int npointsLen, xpointsLen, ypointsLen;
    jint *npointsPtr, *xpointsPtr, *ypointsPtr;
    jbyte *dstBase;
    int i, pointsNeeded, result;

    if (dst == NULL) {
	JNU_ThrowNullPointerException(env, "destination array");
	return;
    }
    if (xpoints == NULL || ypoints == NULL) {
	JNU_ThrowNullPointerException(env, "coordinate array");
	return;
    }
    if (npoints == NULL) {
	JNU_ThrowNullPointerException(env, "polygon length array");
	return;
    }

    npointsLen = (*env)->GetArrayLength(env, npoints);
    xpointsLen = (*env)->GetArrayLength(env, xpoints);
    ypointsLen = (*env)->GetArrayLength(env, ypoints);

    if (npointsLen < numpolys) {
	JNU_ThrowArrayIndexOutOfBoundsException(env,
						"polygon length array size");
	return;
    }

    npointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, npoints, 0);
    if (!npointsPtr) {
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    pointsNeeded = 0;
    for (i=0; i<numpolys; i++) {
	pointsNeeded += npointsPtr[i];
    }

    if (ypointsLen < pointsNeeded || xpointsLen < pointsNeeded) {
	(*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);
	JNU_ThrowArrayIndexOutOfBoundsException(env,
						"coordinate array length");
	return;
    }

    xpointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, xpoints, 0);
    ypointsPtr = (jint *)(*env)->GetPrimitiveArrayCritical(env, ypoints, 0);
    dstBase = (*env)->GetPrimitiveArrayCritical(env, dst, NULL);
    if (!xpointsPtr || !ypointsPtr || !dstBase) {
	if (dstBase) {
	    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
	}
	if (ypointsPtr) {
	    (*env)->ReleasePrimitiveArrayCritical(env, ypoints, ypointsPtr, 0);
	}
	if (xpointsPtr) {
	    (*env)->ReleasePrimitiveArrayCritical(env, xpoints, xpointsPtr, 0);
	}
	/* We know that npointsPtr succeeded - it was tested above... */
	(*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);
	JNU_ThrowNullPointerException(env, "native array lock failed");
	return;
    }

    result = doPolyDrawMany(env, dstBase + offset, dstScanStr, color,
			    xpointsPtr, ypointsPtr, npointsPtr, numpolys,
			    originX, originY, cx, cy, cw, ch, fClose);

    (*env)->ReleasePrimitiveArrayCritical(env, dst, dstBase, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, ypoints, ypointsPtr, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, xpoints, xpointsPtr, 0);
    (*env)->ReleasePrimitiveArrayCritical(env, npoints, npointsPtr, 0);

    if (!result) {
	JNU_ThrowOutOfMemoryError(env, "draw poly");
	return;
    }
}

/*
 * Class:     sun_java2d_loops_ThreeByteRenderer
 * Method:    IsomorphicCopy
 * Signature: (Lsun/awt/image/ImageData;Lsun/awt/image/ImageData;II)V
 */
JNIEXPORT void JNICALL
Java_sun_java2d_loops_ThreeByteRenderer_IsomorphicCopy(
    JNIEnv *env, jobject joSelf,
    jobject srcImage, jobject dstImage, jint width, jint height)
{
/* REMIND -- may want memmove if */
/* behavior is defined somewhere */
#define INNER_LOOP(INNER_COUNT, SRC_TYPE, SRC_BUMP, DST_TYPE, DST_BUMP) \
    memcpy(dstRow, srcRow, INNER_COUNT);

    INVOKE_BLIT(Byte, char, Byte, char, 3 * sizeof(unsigned char), 3, 3);
#undef INNER_LOOP
}
