/*
 * @(#)agent.c	1.92 98/10/01
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 *	Routines to support the debugging agent.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "jni.h"
#include "jni_util.h"
#include "jvm.h"
#include "jvmdi.h"
#include "jlong.h"

/* For sanity check purposes */
#include "sun_tools_agent_Agent.h"
#include "sun_tools_agent_AgentConstants.h"
#include "sun_tools_agent_Breakpoint.h"
#include "sun_tools_agent_CachedClass.h"
#include "sun_tools_agent_CachedField.h"
#include "sun_tools_agent_CachedMethod.h"
#include "sun_tools_agent_MainThread.h"
#include "sun_tools_agent_StackFrame.h"
#include "sun_tools_java_Constants.h"

#define FOREVER 0L

static jfieldID eventKindID;
static jfieldID eventThreadID;
static jfieldID eventClassID;
static jfieldID eventMethodID;
static jfieldID eventBciID;
static jfieldID eventExceptionID;
static jfieldID eventCatchClassID;
static jfieldID eventCatchMethodID;
static jfieldID eventCatchBciID;
static jfieldID eventWaitingID;

static jfieldID sfFrameID;
static jfieldID sfClazzID;
static jfieldID sfMethodID;
static jfieldID sfBciID;
static jfieldID sfIsNativeID;

static jclass agentClass;
static jobject agentObject;
static jobject eventObject;

#ifdef DEBUG
int be_verbose = 0;

int agentdprintf(char *fmt, ...)
{
    int ret = 0;

    if (be_verbose) {
	va_list args;
	va_start(args, fmt);
	ret = fprintf(stderr, "agent.c: ");
	vfprintf(stderr, fmt, args);
	va_end(args);
    }
    return ret;
}
#else
void agentdprintf(char *fmt, ...) {}
#endif


typedef struct ThreadListItem {
    jobject thread;
    struct ThreadListItem* next;
} ThreadListItem;

static ThreadListItem* systemThreadList = NULL;

static void throwJVMDIError(JNIEnv *env,jint err)
{
    switch(err)
    {
        case JVMDI_ERROR_NONE:
            /* don't throw this */
            break;

        case JVMDI_ERROR_INVALID_THREAD:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_THREAD");
            break;

        case JVMDI_ERROR_INVALID_FIELDID:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_FIELDID");
            break;

        case JVMDI_ERROR_INVALID_METHODID:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_METHODID");
            break;

        case JVMDI_ERROR_INVALID_LOCATION:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_LOCATION");
            break;

        case JVMDI_ERROR_INVALID_FRAMEID:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_FRAMEID");
            break;

        case JVMDI_ERROR_NO_MORE_FRAMES:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_NO_MORE_FRAMES");
            break;

        case JVMDI_ERROR_OPAQUE_FRAME:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_OPAQUE_FRAME");
            break;

        case JVMDI_ERROR_NOT_CURRENT_FRAME:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_NOT_CURRENT_FRAME");
            break;

        case JVMDI_ERROR_TYPE_MISMATCH:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_TYPE_MISMATCH");
            break;

        case JVMDI_ERROR_INVALID_SLOT:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_SLOT");
            break;

        case JVMDI_ERROR_DUPLICATE_BREAKPOINT:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_DUPLICATE_BREAKPOINT");
            break;

        case JVMDI_ERROR_NO_SUCH_BREAKPOINT:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_NO_SUCH_BREAKPOINT");
            break;

        case JVMDI_ERROR_THREAD_NOT_SUSPENDED:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_THREAD_NOT_SUSPENDED");
            break;

        case JVMDI_ERROR_THREAD_SUSPENDED:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_THREAD_SUSPENDED");
            break;

        case JVMDI_ERROR_INVALID_OBJECT:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_OBJECT");
            break;

        case JVMDI_ERROR_INVALID_CLASS:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_INVALID_CLASS");
            break;

        case JVMDI_ERROR_NULL_POINTER:
            JNU_ThrowNullPointerException(env,"JVMDI_ERROR_NULL_POINTER");
            break;

        case JVMDI_ERROR_ABSENT_INFORMATION:
            JNU_ThrowByName(env,"sun/tools/agent/AbsentInformationException"
                            ,"JVMDI_ERROR_ABSENT_INFORMATION");
            break;

        case JVMDI_ERROR_OUT_OF_MEMORY:
            JNU_ThrowOutOfMemoryError(env,"JVMDI_ERROR_OUT_OF_MEMORY");
            break;

        case JVMDI_ERROR_VM_DEAD:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_VM_DEAD");
            break;

        case JVMDI_ERROR_NOT_IMPLEMENTED:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"JVMDI_ERROR_NOT_IMPLEMENTED");
            break;

       default:
            JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"unknown JVMDI error");
            break;
    }
}

static jboolean checkForError(JNIEnv *env, jint err)
{
    if(err != JVMDI_ERROR_NONE) {
        throwJVMDIError(env, err);
        return JNI_TRUE;
    }
    return JNI_FALSE;
}


static jframeID getFrameID(JNIEnv *env,jobject thread,jint frameNumber)
{
    jframeID frameID;
    int i;

    if (checkForError(env, JVMDI_GetCurrentFrame(env,thread,&frameID) )) {
        return NULL;
    }

    for(i=0; i<frameNumber; i++) {
        if (checkForError(env, JVMDI_GetCallerFrame(env, frameID, &frameID) )) {
            return NULL;
        }
    }

    return frameID;
}

static jboolean isSystemThread(JNIEnv *env, jobject thread)
{
    ThreadListItem* tli = systemThreadList;
    while(tli != NULL) {
        if ((*env)->IsSameObject(env, tli->thread, thread)) {
            return JNI_TRUE;
        }
        tli = tli->next;
    }
    return JNI_FALSE;
}

static void debugger_jvmdi_event(JNIEnv *env, JVMDI_Event *event)
{
    /*
     * Filter out the events we don't process in sun.tools.debug
     */
    if ((event->kind == JVMDI_EVENT_SINGLE_STEP) ||
        (event->kind == JVMDI_EVENT_BREAKPOINT) || 
        (event->kind == JVMDI_EVENT_EXCEPTION) ||
        (event->kind == JVMDI_EVENT_THREAD_END) || 
        (event->kind == JVMDI_EVENT_VM_DEATH)) {

        if (event->kind == JVMDI_EVENT_VM_DEATH) {
            JNU_CallMethodByName(env, NULL,agentObject,"reportAppExit", "()V");
            return;
        }
    
        /* ignore non-debuggable threads */
        if (isSystemThread(env, event->u.exception.thread)) {
            return;
        }

        /* Exceptions in system code are a frequent cause of deadlock. */
        /* Do not report exceptions caught in java.* or sun.*.         */
        /* Exceptions caught by the agent's MainThread must be         */
        /* reported so the "uncaught" exceptions in user code show up. */
        /* This is not an elegant solution (the elegant solution is    */
        /* to deprecate sun.tools.debug).                              */
        if ((event->kind == JVMDI_EVENT_EXCEPTION)) {
            jclass ecc = event->u.exception.catch_clazz;
            if (ecc != NULL) {
                char *eccn;
                if (JVMDI_GetClassSignature(env, ecc, &eccn) 
                                          == JVMDI_ERROR_NONE) {
                    jboolean inSysCode = (strncmp(eccn, "Ljava/", 6) == 0) || 
                       ((strncmp(eccn, "Lsun/", 5) == 0) &&
                        (strcmp(eccn, "Lsun/tools/agent/MainThread;") != 0));
                    JVMDI_Deallocate(env, (jbyte *)eccn);
                    if (inSysCode) {
                        return;
                    }
                }
            }
        }
    
        (*env)->MonitorEnter(env, eventObject);
    
        /* Wait until the breakpoint handler is ready to accept an event. */
        while (! (*env)->GetBooleanField(env, eventObject, eventWaitingID)) {
            /* Wait here if another thread is busy notifying.  
             * Note that this thread could be suspended at this 
             * point if a event happens on another
             * thread while this event is pending. */
            JVM_MonitorWait(env, eventObject, FOREVER);
        }
    
        /* Communicate the information about this event.  
         * Setting waiting_for_event to false tells handler to 
         * start processing, and tells other threads
         * to wait because it is our turn. */
        (*env)->SetIntField(env, eventObject, eventKindID, event->kind);
        switch (event->kind) {
            case JVMDI_EVENT_EXCEPTION:
                (*env)->SetObjectField(env, eventObject, eventExceptionID,  
                               event->u.exception.exception);
                (*env)->SetObjectField(env, eventObject, eventCatchClassID,  
                               event->u.exception.catch_clazz);
                (*env)->SetLongField(env, eventObject, eventCatchMethodID,  
                             ptr_to_jlong(event->u.exception.catch_method));
                (*env)->SetIntField(env, eventObject, eventCatchBciID,  
                            (jint)event->u.exception.catch_location);
                /* fall through */
            case JVMDI_EVENT_BREAKPOINT:
            case JVMDI_EVENT_SINGLE_STEP:
                (*env)->SetObjectField(env, eventObject, eventClassID,  
                                       event->u.exception.clazz);
                (*env)->SetLongField(env, eventObject, eventMethodID,  
                                     ptr_to_jlong(event->u.exception.method));
                (*env)->SetIntField(env, eventObject, eventBciID, 
                                    (jint)event->u.exception.location);
                /* fall through */
            case JVMDI_EVENT_FRAME_POP:
            case JVMDI_EVENT_THREAD_START:
            case JVMDI_EVENT_THREAD_END:
                (*env)->SetObjectField(env, eventObject, eventThreadID,
                                       event->u.exception.thread);
        }
    
        (*env)->SetBooleanField(env, eventObject, eventWaitingID, JNI_FALSE);
    
        JVM_MonitorNotifyAll(env, eventObject);
    
        /* Wait until handler thread has finished processing this event.
        ** This gives the handler the chance to suspend this thread if it
        ** needs to. */
        JVM_MonitorWait(env, eventObject, FOREVER);
        (*env)->MonitorExit(env, eventObject);
    }
}

/*
 * Class:     sun_tools_agent_MainThread
 * Method:    runMain
 * Signature: (Ljava/lang/Class;J[Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_MainThread_runMain(JNIEnv *env, 
           jobject this, jclass mainClass, jlong methodID, jobjectArray args)
{
    (*env)->CallStaticVoidMethod(env, mainClass, (jmethodID)jlong_to_ptr(methodID), args);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    addSystemThread
 * Signature: (Ljava/lang/Thread;)V
 */
JNIEXPORT void JNICALL
Java_sun_tools_agent_Agent_addSystemThread(JNIEnv *env, jobject this, jobject  
thread)
{
    ThreadListItem* tli;

    /* check if thread is already registered */
    if (isSystemThread(env, thread)) {
        return;
    }

    tli = (ThreadListItem*)malloc(sizeof(ThreadListItem));
    if (tli == NULL) {
        JNU_ThrowNullPointerException(env,"malloc failed");
        return;
    }

    tli->thread = (*env)->NewGlobalRef(env, thread);
    tli->next = systemThreadList;
    systemThreadList = tli;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    removeSystemThread
 * Signature: (Ljava/lang/Thread;)V
 */
JNIEXPORT void JNICALL
Java_sun_tools_agent_Agent_removeSystemThread(JNIEnv *env, jobject this,  
jobject thread)
{
    ThreadListItem* tli = systemThreadList;
    ThreadListItem* prev = NULL;
    while(tli != NULL) {
        if (tli->thread == thread) {
            if (prev == NULL) {
                systemThreadList = tli->next;
            } else {
                prev->next = tli->next;
            }
            free(tli);
            return;
        }
        prev = tli;
        tli = tli->next;
    }

    /* thread not in list */
    return;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    systemThread
 * Signature: (Ljava/lang/Thread;)Z
 */
JNIEXPORT jboolean JNICALL
Java_sun_tools_agent_Agent_systemThread(JNIEnv *env, jobject this, jobject 
thread)
{
#if 0
    Classjava_lang_Thread *t = THREAD(thread);

    /* check if thread is alive */
    if (t->stillborn) {
        /* the thread isn't alive, so tell debugger it isn't debuggable. */
        return JNI_TRUE;
    }
#endif

    return isSystemThread(env, thread);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    initAgent
 * Signature: (Lsun/tools/agent/Event;Lsun/tools/agent/StackFrame;)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_initAgent
   (JNIEnv *env, jobject this, jobject event, jobject exampleStackFrame)
{
    jclass eventClass;
    jclass stackFrameClass;

    if(event == NULL) {
        JNU_ThrowInternalError(env,"Event is null!");
    }

    if(exampleStackFrame == NULL) {
        JNU_ThrowInternalError(env,"exampleStackFrame is null!");
    }

    agentClass = (*env)->GetObjectClass(env, this);
    eventClass = (*env)->GetObjectClass(env, event);
    stackFrameClass = (*env)->GetObjectClass(env, exampleStackFrame);
    agentObject = (*env)->NewGlobalRef(env, this);
    eventObject = (*env)->NewGlobalRef(env, event);

    eventKindID = (*env)->GetFieldID(env, eventClass, "kind", "I");
    eventThreadID = (*env)->GetFieldID(env, eventClass, "thread", "Ljava/lang/Thread;");
    eventClassID = (*env)->GetFieldID(env, eventClass, "clazz" , "Ljava/lang/Class;");
    eventMethodID = (*env)->GetFieldID(env, eventClass, "methodID" , "J");
    eventBciID = (*env)->GetFieldID(env, eventClass, "bci" , "I");
    eventExceptionID = (*env)->GetFieldID(env, eventClass, "exception" , "Ljava/lang/Throwable;");
    eventCatchClassID = (*env)->GetFieldID(env, eventClass, "catch_clazz" , "Ljava/lang/Class;");
    eventCatchMethodID = (*env)->GetFieldID(env, eventClass, "catch_methodID" , "J");
    eventCatchBciID = (*env)->GetFieldID(env, eventClass, "catch_bci" , "I");
    eventWaitingID = (*env)->GetFieldID(env, eventClass, "waiting_for_event", "Z");

    sfFrameID = (*env)->GetFieldID(env, stackFrameClass, "frameID" , "J");
    sfClazzID = (*env)->GetFieldID(env, stackFrameClass, "clazz" , "Ljava/lang/Class;");
    sfMethodID = (*env)->GetFieldID(env, stackFrameClass, "methodID" , "J");
    sfBciID = (*env)->GetFieldID(env, stackFrameClass, "bci" , "I");
    sfIsNativeID = (*env)->GetFieldID(env, stackFrameClass, "isNative" , "Z");

    JVMDI_SetEventHook(env, debugger_jvmdi_event);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    suspendListOfThreads
 * Signature: (I[Ljava/lang/Thread;)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_suspendListOfThreads
  (JNIEnv *env, jclass clazz, jint count, jobjectArray threadList)
{
    int i;
    int shouldThrowOutOfBoundsException = JNI_FALSE;
    jobject localRef;
    jvmdiError error;

    if(count > (*env)->GetArrayLength(env,threadList))
    {
        count = (*env)->GetArrayLength(env,threadList);
        shouldThrowOutOfBoundsException = JNI_TRUE;
    }

    for(i=0; i<count; i++)
    {
        /* This could probably be more efficient */
        localRef = (*env)->GetObjectArrayElement(env,threadList,i);

        if((*env)->ExceptionOccurred(env)) /* BOOM! */
            return;

        if(isSystemThread(env,localRef))
            continue;

        error = JVMDI_SuspendThread(env,localRef);
        if ((error != JVMDI_ERROR_NONE) &&
            (error != JVMDI_ERROR_THREAD_SUSPENDED)) {
            throwJVMDIError(env, error);
            return;
        }

        (*env)->DeleteLocalRef(env,localRef);
    }

    if(shouldThrowOutOfBoundsException)
        JNU_ThrowArrayIndexOutOfBoundsException(env," ");

}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    suspendSpecificThread
 * Signature: (Ljava/lang/Thread;)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_suspendSpecificThread
  (JNIEnv *env, jclass clazz, jobject thread)
{
    jvmdiError error = JVMDI_SuspendThread(env,thread);
    if ((error != JVMDI_ERROR_NONE) &&
        (error != JVMDI_ERROR_THREAD_SUSPENDED)) {
        throwJVMDIError(env, error);
    }
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getThreadStatus
 * Signature: (Ljava/lang/Thread;)I
 */

/*
 * These should match constants in AgentConstants.java
 */
#define THR_STATUS_SUSPENDED 5
#define THR_STATUS_BREAK 6

JNIEXPORT jint JNICALL Java_sun_tools_agent_Agent_getThreadStatus
  (JNIEnv *env, jobject this, jobject thread)
{
    jint threadStatus, suspendStatus;

    checkForError(env, JVMDI_GetThreadStatus(env, thread, &threadStatus, &suspendStatus));

    /*
     * Collapse two statuses into a single word, giving precedence to
     * suspend info.
     */
    if (suspendStatus & JVMDI_SUSPEND_STATUS_BREAK) {
        threadStatus = THR_STATUS_BREAK;
    } else if (suspendStatus & JVMDI_SUSPEND_STATUS_SUSPENDED) {
        threadStatus = THR_STATUS_SUSPENDED;
    }
    return threadStatus;
}

static void disposeClasses(JNIEnv *env, jint count, jclass *classes) {
    int i;

    for (i = 0; i < count; i++) {
        (*env)->DeleteGlobalRef(env, classes[i]);
    }
    JVMDI_Deallocate(env, (jbyte *)classes);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getClasses
 * Signature: ()[Ljava/lang/Class;
 */
JNIEXPORT jobjectArray JNICALL Java_sun_tools_agent_Agent_getClasses
  (JNIEnv *env, jobject this)
{
    jobjectArray retClasses = NULL;
    jclass *classes;
    jint count;
    jclass javaLangClass;
    int i;
    jint preparedCount = 0;
    jint status;

    if (checkForError(env, JVMDI_GetLoadedClasses(env, &count, &classes))) {
        return NULL;
    }

    for (i = 0; i < count; i++) {
        if (checkForError(env, JVMDI_GetClassStatus(env, classes[i], &status))) {
            disposeClasses(env, count, classes);
            return NULL;
        }
        if (status & JVMDI_CLASS_STATUS_PREPARED) {
            preparedCount++;
        }

    }

    if (count <= 0) {
        JNU_ThrowByName(env,"java/lang/InternalError"
                            ,"Invalid Loaded Class Count");
        return NULL;
    }

    /* java.lang.Class is final + checks above make this OK */
    javaLangClass = (*env)->GetObjectClass(env, classes[0]);
    retClasses = (*env)->NewObjectArray(env, preparedCount, javaLangClass, NULL);
    if (retClasses != NULL) {
        jint preparedIndex = 0;
        for (i = 0; (i < count) && !((*env)->ExceptionOccurred(env)); i++) {
            if (checkForError(env, JVMDI_GetClassStatus(env, classes[i], &status))) {
                disposeClasses(env, count, classes);
                return NULL;
            }
            if (status & JVMDI_CLASS_STATUS_PREPARED) {
                (*env)->SetObjectArrayElement(env, retClasses,
                                              preparedIndex++, classes[i]);
            }
        }
    }
    disposeClasses(env, count, classes);
    
    return retClasses;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldObject
 * Signature: (Ljava/lang/Object;J)Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_sun_tools_agent_Agent_getFieldObject
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetObjectField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldBoolean
 * Signature: (Ljava/lang/Object;J)Z
 */
JNIEXPORT jboolean JNICALL Java_sun_tools_agent_Agent_getFieldBoolean
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetBooleanField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldByte
 * Signature: (Ljava/lang/Object;J)B
 */
JNIEXPORT jbyte JNICALL Java_sun_tools_agent_Agent_getFieldByte
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetByteField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldChar
 * Signature: (Ljava/lang/Object;J)C
 */
JNIEXPORT jchar JNICALL Java_sun_tools_agent_Agent_getFieldChar
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetCharField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldShort
 * Signature: (Ljava/lang/Object;J)S
 */
JNIEXPORT jshort JNICALL Java_sun_tools_agent_Agent_getFieldShort
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetShortField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldInt
 * Signature: (Ljava/lang/Object;J)I
 */
JNIEXPORT jint JNICALL Java_sun_tools_agent_Agent_getFieldInt
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetIntField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldLong
 * Signature: (Ljava/lang/Object;J)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_Agent_getFieldLong
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetLongField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldFloat
 * Signature: (Ljava/lang/Object;J)F
 */
JNIEXPORT jfloat JNICALL Java_sun_tools_agent_Agent_getFieldFloat
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetFloatField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getFieldDouble
 * Signature: (Ljava/lang/Object;J)D
 */
JNIEXPORT jdouble JNICALL Java_sun_tools_agent_Agent_getFieldDouble
  (JNIEnv *env, jobject this, jobject target, jlong id)
{
    return (*env)->GetDoubleField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldObject
 * Signature: (Ljava/lang/Class;J)Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_sun_tools_agent_Agent_getStaticFieldObject
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticObjectField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldBoolean
 * Signature: (Ljava/lang/Class;J)Z
 */
JNIEXPORT jboolean JNICALL Java_sun_tools_agent_Agent_getStaticFieldBoolean
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticBooleanField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldByte
 * Signature: (Ljava/lang/Class;J)B
 */
JNIEXPORT jbyte JNICALL Java_sun_tools_agent_Agent_getStaticFieldByte
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticByteField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldChar
 * Signature: (Ljava/lang/Class;J)C
 */
JNIEXPORT jchar JNICALL Java_sun_tools_agent_Agent_getStaticFieldChar
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticCharField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldShort
 * Signature: (Ljava/lang/Class;J)S
 */
JNIEXPORT jshort JNICALL Java_sun_tools_agent_Agent_getStaticFieldShort
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticShortField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldInt
 * Signature: (Ljava/lang/Class;J)I
 */
JNIEXPORT jint JNICALL Java_sun_tools_agent_Agent_getStaticFieldInt
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticIntField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldLong
 * Signature: (Ljava/lang/Class;J)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_Agent_getStaticFieldLong
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticLongField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldFloat
 * Signature: (Ljava/lang/Class;J)F
 */
JNIEXPORT jfloat JNICALL Java_sun_tools_agent_Agent_getStaticFieldFloat
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticFloatField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStaticFieldDouble
 * Signature: (Ljava/lang/Class;J)D
 */
JNIEXPORT jdouble JNICALL Java_sun_tools_agent_Agent_getStaticFieldDouble
  (JNIEnv *env, jobject this, jclass target, jlong id)
{
    return (*env)->GetStaticDoubleField(env,target,(jfieldID)jlong_to_ptr(id));
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStackObject
 * Signature: (Ljava/lang/Thread;II)Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_sun_tools_agent_Agent_getStackObject
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot)
{
    jobject value;
    jobject ref;
    jframeID frameID = getFrameID(env,thread,frame);
    if((*env)->ExceptionOccurred(env))
        return NULL;

    checkForError(env, JVMDI_GetLocalObject(env, frameID, slot, &value) );
    
    ref = (*env)->NewLocalRef(env,value);
    (*env)->DeleteGlobalRef(env,value);

    return ref;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStackBoolean
 * Signature: (Ljava/lang/Thread;II)Z
 */
JNIEXPORT jboolean JNICALL Java_sun_tools_agent_Agent_getStackBoolean
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot)
{
    jint value;
    jframeID frameID = getFrameID(env,thread,frame);
    if((*env)->ExceptionOccurred(env))
        return (jboolean)0;

    checkForError(env, JVMDI_GetLocalInt(env, frameID, slot, &value));

    return (jboolean)value;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStackInt
 * Signature: (Ljava/lang/Thread;II)I
 */
JNIEXPORT jint JNICALL Java_sun_tools_agent_Agent_getStackInt
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot)
{
    jint value;
    jframeID frameID = getFrameID(env,thread,frame);

    if((*env)->ExceptionOccurred(env))
        return (jint)0;

    checkForError(env, JVMDI_GetLocalInt(env, frameID, slot, &value) );
    return value;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStackLong
 * Signature: (Ljava/lang/Thread;II)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_Agent_getStackLong
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot)
{
    jframeID frameID = getFrameID(env,thread,frame);
    jlong value;

    if((*env)->ExceptionOccurred(env))
        return (jlong)0;

    checkForError(env, JVMDI_GetLocalLong(env, frameID, slot, &value) );
    return value;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStackFloat
 * Signature: (Ljava/lang/Thread;II)F
 */
JNIEXPORT jfloat JNICALL Java_sun_tools_agent_Agent_getStackFloat
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot)
{
    jframeID frameID;
    jfloat value;

    frameID = getFrameID(env,thread,frame);

    if((*env)->ExceptionOccurred(env))
        return (jfloat)0;

    checkForError(env, JVMDI_GetLocalFloat(env, frameID, slot, &value) );
    return value;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    getStackDouble
 * Signature: (Ljava/lang/Thread;II)D
 */
JNIEXPORT jdouble JNICALL Java_sun_tools_agent_Agent_getStackDouble
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot)
{
    jframeID frameID;
    jdouble value;

    frameID = getFrameID(env,thread,frame);

    if((*env)->ExceptionOccurred(env))
        return (jdouble)0;

    checkForError(env, JVMDI_GetLocalDouble(env, frameID, slot, &value) );
    return value;
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setSingleStep
 * Signature: (Ljava/lang/Thread;Z)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setSingleStep
  (JNIEnv *env, jobject this, jobject thread, jboolean step)
{
    checkForError(env, JVMDI_SetSingleStep(env,thread,step) );
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldBoolean
 * Signature: (Ljava/lang/Object;JZ)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldBoolean
  (JNIEnv *env, jobject this, jobject target, jlong id, jboolean value)
{
    (*env)->SetBooleanField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldByte
 * Signature: (Ljava/lang/Object;JB)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldByte
  (JNIEnv *env, jobject this, jobject target, jlong id, jbyte value)
{
    (*env)->SetByteField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldChar
 * Signature: (Ljava/lang/Object;JC)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldChar
  (JNIEnv *env, jobject this, jobject target, jlong id, jchar value)
{
    (*env)->SetCharField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldShort
 * Signature: (Ljava/lang/Object;JS)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldShort
  (JNIEnv *env, jobject this, jobject target, jlong id, jshort value)
{
    (*env)->SetShortField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldInt
 * Signature: (Ljava/lang/Object;JI)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldInt
  (JNIEnv *env, jobject this, jobject target, jlong id, jint value)
{
    (*env)->SetIntField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldLong
 * Signature: (Ljava/lang/Object;JJ)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldLong
  (JNIEnv *env, jobject this, jobject target, jlong id, jlong value)
{
    (*env)->SetLongField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldFloat
 * Signature: (Ljava/lang/Object;JF)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldFloat
  (JNIEnv *env, jobject this, jobject target, jlong id, jfloat value)
{
    (*env)->SetFloatField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setFieldDouble
 * Signature: (Ljava/lang/Object;JD)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setFieldDouble
  (JNIEnv *env, jobject this, jobject target, jlong id, jdouble value)
{
    (*env)->SetDoubleField(env,target,(jfieldID)jlong_to_ptr(id),value);
}


/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldBoolean
 * Signature: (Ljava/lang/Class;JZ)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldBoolean
  (JNIEnv *env, jobject this, jclass target, jlong id, jboolean value)
{
    (*env)->SetStaticBooleanField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldByte
 * Signature: (Ljava/lang/Class;JB)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldByte
  (JNIEnv *env, jobject this, jclass target, jlong id, jbyte value)
{
    (*env)->SetStaticByteField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldChar
 * Signature: (Ljava/lang/Class;JC)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldChar
  (JNIEnv *env, jobject this, jclass target, jlong id, jchar value)
{
    (*env)->SetStaticCharField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldShort
 * Signature: (Ljava/lang/Class;JS)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldShort
  (JNIEnv *env, jobject this, jclass target, jlong id, jshort value)
{
    (*env)->SetStaticShortField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldInt
 * Signature: (Ljava/lang/Class;JI)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldInt
  (JNIEnv *env, jobject this, jclass target, jlong id, jint value)
{
    (*env)->SetStaticIntField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldLong
 * Signature: (Ljava/lang/Class;JJ)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldLong
  (JNIEnv *env, jobject this, jclass target, jlong id, jlong value)
{
    (*env)->SetStaticLongField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldFloat
 * Signature: (Ljava/lang/Class;JF)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldFloat
  (JNIEnv *env, jobject this, jclass target, jlong id, jfloat value)
{
    (*env)->SetStaticFloatField(env,target,(jfieldID)jlong_to_ptr(id),value);
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStaticFieldDouble
 * Signature: (Ljava/lang/Class;JD)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStaticFieldDouble
  (JNIEnv *env, jobject this, jclass target, jlong id, jdouble value)
{
    (*env)->SetStaticDoubleField(env,target,(jfieldID)jlong_to_ptr(id),value);
}



/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStackBoolean
 * Signature: (Ljava/lang/Thread;IIZ)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStackBoolean
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot, jboolean 
value)
{
    jframeID frameID = getFrameID(env,thread,frame);

    if((*env)->ExceptionOccurred(env))
        return;

    checkForError(env, JVMDI_SetLocalInt(env, frameID, slot, (int)value) );
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStackInt
 * Signature: (Ljava/lang/Thread;III)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStackInt
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot, jint value)
{
    jframeID frameID = getFrameID(env,thread,frame);

    if((*env)->ExceptionOccurred(env))
        return;

    checkForError(env, JVMDI_SetLocalInt(env, frameID, slot, value) );
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStackLong
 * Signature: (Ljava/lang/Thread;IIJ)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStackLong
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot, jlong 
value)
{
    jframeID frameID = getFrameID(env,thread,frame);

    if((*env)->ExceptionOccurred(env))
        return;

    checkForError(env, JVMDI_SetLocalLong(env, frameID, slot, value) );
}

/*
 * Class:     sun_tools_agent_Agent
 * Method:    setStackDouble
 * Signature: (Ljava/lang/Thread;IID)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Agent_setStackDouble
  (JNIEnv *env, jobject this, jobject thread, jint frame, jint slot, jdouble 
value)
{
    jframeID frameID = getFrameID(env,thread,frame);

    if((*env)->ExceptionOccurred(env))
        return;

    checkForError(env, JVMDI_SetLocalDouble(env, frameID, slot, value) );
}

/*
 * Breakpoint.java native methods
 */

/*
 * Class:     sun_tools_agent_Breakpoint
 * Method:    setBreakpoint
 * Signature: (Ljava/lang/Class;JI)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Breakpoint_setBreakpoint
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID, jint bci)
{
    checkForError(env, 
        JVMDI_SetBreakpoint(env, targetClass,(jmethodID)jlong_to_ptr(methodID), (jlocation)bci) );
}

/*
 * Class:     sun_tools_agent_Breakpoint
 * Method:    clrBreakpoint
 * Signature: (Ljava/lang/Class;JI)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_Breakpoint_clrBreakpoint
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID, jint bci)
{
    checkForError(env, 
        JVMDI_ClearBreakpoint(env, targetClass,(jmethodID)jlong_to_ptr(methodID), (jlocation)bci) );
}

/*
 * CachedClass.java native methods
 */


/*
 * Class:     sun_tools_agent_CachedClass
 * Method:    getFieldIDs
 * Signature: (Ljava/lang/Class;)[J
 */
JNIEXPORT jlongArray JNICALL Java_sun_tools_agent_CachedClass_getFieldIDs
  (JNIEnv *env, jclass clazz, jclass targetClass) {
    jint count;
    jfieldID *fieldIDs;
    jboolean error;
    jlongArray retArray = NULL;
    jlong *elements;
    jboolean isCopy;
    int i;

    error = checkForError(env,
         JVMDI_GetClassFields(env, targetClass, &count, &fieldIDs));
    if (!error) {
        retArray = (*env)->NewLongArray(env, count);
        if (retArray != NULL) {
            elements = (*env)->GetLongArrayElements(env, retArray, &isCopy);
            if (elements != NULL) {
                for (i = 0; i < count; i++) {
                    elements[i] = ptr_to_jlong(fieldIDs[i]);
                }
                (*env)->ReleaseLongArrayElements(env, retArray, elements, 0);
            }
            checkForError(env, JVMDI_Deallocate(env, (jbyte *)fieldIDs));
        }
    }
     
    return retArray;
}

/*
 * Class:     sun_tools_agent_CachedClass
 * Method:    getMethodIDs
 * Signature: (Ljava/lang/Class;)[J
 */
JNIEXPORT jlongArray JNICALL Java_sun_tools_agent_CachedClass_getMethodIDs
  (JNIEnv *env, jclass clazz, jclass targetClass) {
    jint count;
    jmethodID *methodIDs;
    jboolean error;
    jlongArray retArray = NULL;
    jlong *elements;
    jboolean isCopy;
    int i;

    error = checkForError(env,
         JVMDI_GetClassMethods(env, targetClass, &count, &methodIDs));
    if (!error) {
        retArray = (*env)->NewLongArray(env, count);
        if (retArray != NULL) {
            elements = (*env)->GetLongArrayElements(env, retArray, &isCopy);
            if (elements != NULL) {
                for (i = 0; i < count; i++) {
                    elements[i] = ptr_to_jlong(methodIDs[i]);
                }
                (*env)->ReleaseLongArrayElements(env, retArray, elements, 0);
            }
            checkForError(env, JVMDI_Deallocate(env, (jbyte *)methodIDs));
        }
    }
     
    return retArray;
}    

/*
 * Class:     sun_tools_agent_CachedClass
 * Method:    getSourceFileName
 * Signature: (Ljava/lang/Class;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_sun_tools_agent_CachedClass_getSourceFileName
  (JNIEnv *env, jclass clazz, jclass targetClass) {
    char *fileNameUTF;
    jstring fileName = NULL;
    if (!checkForError(env, JVMDI_GetSourceFileName(env, targetClass, &fileNameUTF))) {
        fileName = (*env)->NewStringUTF(env, fileNameUTF);
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)fileNameUTF));
    }
    return fileName;
}


/*
 * CachedField.java native methods
 */

/*
 * Class:     sun_tools_agent_CachedField
 * Method:    getDeclaringClass
 * Signature: (Ljava/lang/Class;J)Ljava/lang/Class;
 */
JNIEXPORT jclass JNICALL Java_sun_tools_agent_CachedField_getDeclaringClass
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong fieldID) {
    jobject ref;
    jclass  declaringClass;

    if (checkForError(env, JVMDI_GetFieldDeclaringClass(env, targetClass, 
                                                   jlong_to_ptr(fieldID),
                                                   &declaringClass))) {
        return NULL;
    }

    ref = (*env)->NewLocalRef(env,declaringClass);
    (*env)->DeleteGlobalRef(env,declaringClass);

    return ref;
}

/*
 * Class:     sun_tools_agent_CachedField
 * Method:    getFieldID
 * Signature: (Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_CachedField_getFieldID
  (JNIEnv *env, jclass clazz, jclass targetClass, jstring name, jstring 
signature)
{
    jfieldID id;
    char *n,*s;

    n = (char *)JNU_GetStringPlatformChars(env,name,NULL);
    if(n == NULL)
    {
        JNU_ThrowNoSuchFieldError(env,"name is null");
        return 0;
    }

    s = (char *)JNU_GetStringPlatformChars(env,signature,NULL);
    if(s == NULL)
    {
        JNU_ReleaseStringPlatformChars(env, name, (const char *)n);
        JNU_ThrowNoSuchFieldError(env,"signature is null");
        return 0;
    }

    id = (*env)->GetFieldID(env,targetClass,n,s);

    JNU_ReleaseStringPlatformChars(env, name, (const char *)n);
    JNU_ReleaseStringPlatformChars(env, signature, (const char *)s);

    return ptr_to_jlong(id);
}

/*
 * Class:     sun_tools_agent_CachedField
 * Method:    getModifiers
 * Signature: (Ljava/lang/Class;J)I
 */
JNIEXPORT jint JNICALL Java_sun_tools_agent_CachedField_getModifiers
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong fieldID) {
    jint modifiers = 0;

    checkForError(env, JVMDI_GetFieldModifiers(env, targetClass, 
                                          jlong_to_ptr(fieldID), &modifiers));
    return modifiers;
}



/*
 * Class:     sun_tools_agent_CachedField
 * Method:    getName
 * Signature: (Ljava/lang/Class;J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_sun_tools_agent_CachedField_getName
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong fieldID) {
    char *fieldNameUTF;
    char *ignore;
    jstring fieldName = NULL;
    if (!checkForError(env, JVMDI_GetFieldName(env, targetClass, 
                                              jlong_to_ptr(fieldID),
                                              &fieldNameUTF, &ignore))) {
        fieldName = (*env)->NewStringUTF(env, fieldNameUTF);
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)fieldNameUTF));
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)ignore));
    }
    return fieldName;
}

/*
 * Class:     sun_tools_agent_CachedField
 * Method:    getSignature
 * Signature: (Ljava/lang/Class;J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_sun_tools_agent_CachedField_getSignature
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong fieldID) {
    char *ignore;
    char *fieldSigUTF;
    jstring signature = NULL;
    if (!checkForError(env, JVMDI_GetFieldName(env, targetClass, 
                                              jlong_to_ptr(fieldID),
                                              &ignore, &fieldSigUTF))) {
        signature = (*env)->NewStringUTF(env, fieldSigUTF);
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)ignore));
    }
    return signature;
}

/*
 * Class:     sun_tools_agent_CachedField
 * Method:    getStaticFieldID
 * Signature: (Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_CachedField_getStaticFieldID
  (JNIEnv *env, jclass clazz, jclass targetClass, jstring name, jstring 
signature)
{
    jfieldID id;
    const char *n;
    const char *s;

    n =(*env)->GetStringUTFChars(env,name,NULL);
    if(n == NULL)
    {
        JNU_ThrowNoSuchFieldError(env,"name is null");
        return 0;
    }

    s = (*env)->GetStringUTFChars(env,signature,NULL);
    if(s == NULL)
    {
        (*env)->ReleaseStringUTFChars(env, name, n);
        JNU_ThrowNoSuchFieldError(env,"signature is null");
        return 0;
    }

    id = (*env)->GetStaticFieldID(env,targetClass,n,s);

    (*env)->ReleaseStringUTFChars(env, name, n);
    (*env)->ReleaseStringUTFChars(env, signature, s);

    return ptr_to_jlong(id);
}

/*
 * CacheMethod.java native methods
 */

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getMethodName
 * Signature: (Ljava/lang/Class;J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_sun_tools_agent_CachedMethod_getMethodName
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    char *methodNameUTF;
    char *ignore;
    jstring methodName = NULL;
    if (!checkForError(env, JVMDI_GetMethodName(env, targetClass, 
                                              jlong_to_ptr(methodID),
                                              &methodNameUTF, &ignore))) {
        methodName = (*env)->NewStringUTF(env, methodNameUTF);
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)methodNameUTF));
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)ignore));
    }
    return methodName;

}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getMethodSignature
 * Signature: (Ljava/lang/Class;J)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_sun_tools_agent_CachedMethod_getMethodSignature
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    char *ignore;
    char *methodSigUTF;
    jstring signature = NULL;
    if (!checkForError(env, JVMDI_GetMethodName(env, targetClass, 
                                              jlong_to_ptr(methodID),
                                              &ignore, &methodSigUTF))) {
        signature = (*env)->NewStringUTF(env, methodSigUTF);
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)ignore));
    }
    return signature;
}

static jlong getMethodID(JNIEnv *env, jclass targetClass, jstring name, 
                         jstring signature, jboolean isStatic)
{
    jmethodID id;
    char *nam, *sig;

    nam = (char *)JNU_GetStringPlatformChars(env, name, NULL);
    if(nam == NULL){
        JNU_ThrowNoSuchFieldError(env, "name is null");
        return 0;
    }

    sig = (char *)JNU_GetStringPlatformChars(env, signature, NULL);
    if(sig == NULL) {
        JNU_ReleaseStringPlatformChars(env, name, (const char *)nam);
        JNU_ThrowNoSuchFieldError(env, "signature is null");
        return 0;
    }

    if (isStatic) {
        id = (*env)->GetStaticMethodID(env, targetClass, nam, sig);
    } else {
        id = (*env)->GetMethodID(env, targetClass, nam, sig);
    }

    JNU_ReleaseStringPlatformChars(env, name, (const char *)nam);
    JNU_ReleaseStringPlatformChars(env, signature, (const char *)sig);

    return ptr_to_jlong(id);
}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getMethodID
 * Signature: (Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_CachedMethod_getMethodID(
                           JNIEnv *env, jclass clazz, 
                           jclass targetClass, 
                           jstring name, jstring signature)
{
    return getMethodID(env, targetClass,  name, signature, JNI_FALSE);
}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getStaticMethodID
 * Signature: (Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_CachedMethod_getStaticMethodID(
                           JNIEnv *env, jclass clazz, 
                           jclass targetClass, 
                           jstring name, jstring signature)
{
    return getMethodID(env, targetClass,  name, signature, JNI_TRUE);
}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getArgumentsSize
 * Signature: (Ljava/lang/Class;J)I
 */
JNIEXPORT jint JNICALL Java_sun_tools_agent_CachedMethod_getArgumentsSize
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    jint size = 0;

    checkForError(env, JVMDI_GetArgumentsSize(env, targetClass, 
                                          jlong_to_ptr(methodID), &size));
    return size;
}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getCode
 * Signature: (Ljava/lang/Class;J)[B
 */
JNIEXPORT jbyteArray JNICALL Java_sun_tools_agent_CachedMethod_getCode
  (JNIEnv * env, jclass clazz, jclass targetClass, jlong methodID) {
    jint count;
    jbyte *bytecodes;
    jboolean error;
    jlongArray retArray = NULL;

    error = checkForError(env,
         JVMDI_GetBytecodes(env, targetClass, jlong_to_ptr(methodID),
                            &count, &bytecodes));
    if (!error) {
        retArray = (*env)->NewByteArray(env, count);
        if (retArray != NULL) {
            (*env)->SetByteArrayRegion(env, retArray, 0, count, bytecodes);
            checkForError(env, JVMDI_Deallocate(env, bytecodes));
        }
    }
     
    return retArray;

}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getDeclaringClass
 * Signature: (Ljava/lang/Class;J)Ljava/lang/Class;
 */
JNIEXPORT jclass JNICALL Java_sun_tools_agent_CachedMethod_getDeclaringClass
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    jobject ref;
    jclass  declaringClass;

    if (checkForError(env, JVMDI_GetMethodDeclaringClass(env, targetClass, 
                                                   jlong_to_ptr(methodID),
                                                   &declaringClass))) {
        return NULL;
    }

    ref = (*env)->NewLocalRef(env,declaringClass);
    (*env)->DeleteGlobalRef(env,declaringClass);

    return ref;
}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getEndLocation
 * Signature: (Ljava/lang/Class;J)J
 */
JNIEXPORT jlong JNICALL Java_sun_tools_agent_CachedMethod_getEndLocation
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    jlocation end = 0;
    jlocation ignore = 0;

    checkForError(env, JVMDI_GetMethodLocation(env, targetClass, 
                                               jlong_to_ptr(methodID), 
                                               &ignore, &end));
    return end + 1;  /* return method length instead of last bci */
}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getLineTable
 * Signature: (Ljava/lang/Class;J)[Lsun/tools/agent/CachedMethod$LineEntry;
 */
JNIEXPORT jobjectArray JNICALL Java_sun_tools_agent_CachedMethod_getLineTable
    (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    jobjectArray retLines = NULL;
    JVMDI_line_number_entry *lines;
    jint count = 0;
    jclass entryClass = NULL;
    jmethodID entryConstructor = NULL;
    int i;

    if (checkForError(env, JVMDI_GetLineNumberTable(env, targetClass,
                                                    jlong_to_ptr(methodID),
                                                    &count, &lines))) {
        return NULL;
    }

    entryClass = (*env)->FindClass(env, 
                                   "sun/tools/agent/CachedMethod$LineEntry");
    if (!((*env)->ExceptionOccurred(env))) {
        entryConstructor = (*env)->GetMethodID(env, entryClass, 
                                               "<init>", "(JI)V");
    }
    if (!((*env)->ExceptionOccurred(env))) {
        retLines = (*env)->NewObjectArray(env, count, entryClass, NULL);
    }
    for (i = 0; (i < count) && !((*env)->ExceptionOccurred(env)); i++) {
        jobject entry = (*env)->NewObject(env, entryClass, entryConstructor,
                                          lines[i].start_location, 
                                          lines[i].line_number);

        if (!((*env)->ExceptionOccurred(env))) {
            (*env)->SetObjectArrayElement(env, retLines, i, entry);
        }
    }

    checkForError(env, JVMDI_Deallocate(env, (jbyte *)lines));
    
    return retLines;
}

/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getModifiers
 * Signature: (Ljava/lang/Class;J)I
 */
JNIEXPORT jint JNICALL Java_sun_tools_agent_CachedMethod_getModifiers
  (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    jint modifiers = 0;

    checkForError(env, JVMDI_GetMethodModifiers(env, targetClass, 
                                          jlong_to_ptr(methodID), &modifiers));
    return modifiers;
}



/*
 * Class:     sun_tools_agent_CachedMethod
 * Method:    getVariableTable
 * Signature: (Ljava/lang/Class;J)[Lsun/tools/agent/CachedMethod$VariableEntry;
 */
JNIEXPORT jobjectArray JNICALL Java_sun_tools_agent_CachedMethod_getVariableTable
    (JNIEnv *env, jclass clazz, jclass targetClass, jlong methodID) {
    jobjectArray retVariables = NULL;
    JVMDI_local_variable_entry *variables;
    jint count = 0;
    jclass entryClass = NULL;
    jmethodID entryConstructor = NULL;
    int i;

    if (checkForError(env, JVMDI_GetLocalVariableTable(env, targetClass,
                                                       jlong_to_ptr(methodID),
                                                       &count, &variables))) {
        return NULL;
    }

    entryClass = (*env)->FindClass(env, "sun/tools/agent/CachedMethod$VariableEntry");
    if (!((*env)->ExceptionOccurred(env))) {
        entryConstructor = (*env)->GetMethodID(env, entryClass, "<init>",
                                      "(JILjava/lang/String;Ljava/lang/String;I)V");
    }
    if (!((*env)->ExceptionOccurred(env))) {
        retVariables = (*env)->NewObjectArray(env, count, entryClass, NULL);
    }
    for (i = 0; (i < count) && !((*env)->ExceptionOccurred(env)); i++) {
        jstring name;
        jstring signature;
        jobject entry;
        JVMDI_local_variable_entry *var = &variables[i];

        name = (*env)->NewStringUTF(env, var->name);

        if (!((*env)->ExceptionOccurred(env))) {        
            signature = (*env)->NewStringUTF(env, var->signature);
        }
        
        if (!((*env)->ExceptionOccurred(env))) {
            entry = (*env)->NewObject(env, entryClass, entryConstructor,
                                      var->start_location,
                                      var->length, name, signature, var->slot);
        }

        checkForError(env, JVMDI_Deallocate(env, (jbyte *)var->name));
        checkForError(env, JVMDI_Deallocate(env, (jbyte *)var->signature));

        if (!((*env)->ExceptionOccurred(env))) {
            (*env)->SetObjectArrayElement(env, retVariables, i, entry);
        }
    }

    checkForError(env, JVMDI_Deallocate(env, (jbyte *)variables));
    
    return retVariables;
}

/*
 * StackFrame.java native methods
 */

/*
 * Class:     sun_tools_agent_StackFrame
 * Method:    frameCount
 * Signature: (Ljava/lang/Thread;)I
 */
JNIEXPORT jint JNICALL Java_sun_tools_agent_StackFrame_frameCount
  (JNIEnv *env, jclass clazz, jobject thread)
{
    jint err;
    jframeID frameID;
    int i;
    jint frameCount = 0;
    jclass frameClass;
    jmethodID frameMethod;

    if (checkForError(env, JVMDI_GetCurrentFrame(env,thread,&frameID) )) {
        return 0;
    }

    frameCount++;

    while((err = JVMDI_GetCallerFrame(env, frameID, &frameID)) 
                                               == JVMDI_ERROR_NONE) {
        frameCount++;
    }

    if(err != JVMDI_ERROR_NO_MORE_FRAMES) {
        throwJVMDIError(env,err);
        return 0;
    }

#ifdef I_WILL_FIGURE_THIS_OUT_LATER
    if (checkForError(env, JVMDI_GetFrameMethod(env, frameID, 
                                                &frameClass, 
                                                &frameMethod) )) {
        return 0;
    }
    if ((*env)->IsSameObject(env, frameClass, agentClass)) {
        --frameCount;
    }
#endif

    return frameCount;
}

/*
 * Class:     sun_tools_agent_StackFrame
 * Method:    fillInFrameData
 * Signature: (Ljava/lang/Thread;[Lsun/tools/agent/StackFrame;)V
 */
JNIEXPORT void JNICALL Java_sun_tools_agent_StackFrame_fillInFrameData
  (JNIEnv *env, jclass clazz, jobject thread, jobjectArray data)
{
    jframeID frameID;
    int i,count;

    if (checkForError(env, JVMDI_GetCurrentFrame(env,thread, &frameID) )) {
        return;
    }

    count = (*env)->GetArrayLength(env,data);
    for(i=0; i<count; i++) {
        jobject frame;
        jclass frameClass;
        jmethodID frameMethod;
        jlocation frameLocation;
        jboolean frameIsNative;

        frame = (*env)->GetObjectArrayElement(env,data,i);
        if((*env)->ExceptionOccurred(env)) {
            /* BOOM! */
            return;
        }

        /* Get frame class and method */
        if (checkForError(env, JVMDI_GetFrameLocation(env, frameID, 
                                                      &frameClass, 
                                                      &frameMethod, 
                                                      &frameLocation) )) {
            return;
        }
        
        /* Determine if it is native */
        if (checkForError(env, JVMDI_IsMethodNative(env, frameClass, 
                                    frameMethod, &frameIsNative) )) {
            (*env)->DeleteGlobalRef(env,frameClass);
            return;
        }

        (*env)->SetLongField(env,frame,sfFrameID,ptr_to_jlong(frameID));
        (*env)->SetObjectField(env,frame,sfClazzID,frameClass);
        (*env)->DeleteGlobalRef(env,frameClass);
        (*env)->SetLongField(env,frame,sfMethodID,ptr_to_jlong(frameMethod));
        (*env)->SetIntField(env,frame,sfBciID, (jint)frameLocation);
        (*env)->SetBooleanField(env,frame,sfIsNativeID,frameIsNative);

        /* Get the next frame unless this is the last iteration */
        if (i < count - 1) {
            if (checkForError(env, JVMDI_GetCallerFrame(env, frameID, &frameID) )) {
                return;
            }
        }
    }
}
