/*
 * @(#)CursorTest.java	1.10 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.applet.Applet;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

public class CursorTest extends Applet {
    public void init() {
	Container c;
        Font font = new Font("Serif", Font.PLAIN, 12);
	setLayout(new BorderLayout());
	add("North", c = new Panel());
	c.setLayout(new GridLayout(1, 3));
	c.add(new CursorCanvas(Cursor.NW_RESIZE_CURSOR, Color.white, font));
	c.add(new CursorCanvas(Cursor.N_RESIZE_CURSOR, Color.lightGray, font));
	c.add(new CursorCanvas(Cursor.NE_RESIZE_CURSOR, Color.white, font));
	add("South", c = new Panel());
	c.setLayout(new GridLayout(1, 3));
	c.add(new CursorCanvas(Cursor.SW_RESIZE_CURSOR, Color.white, font));
	c.add(new CursorCanvas(Cursor.S_RESIZE_CURSOR, Color.lightGray, font));
	c.add(new CursorCanvas(Cursor.SE_RESIZE_CURSOR, Color.white, font));
	add("East", c = new Panel());
	c.setLayout(new GridLayout(1, 1));
	c.add(new CursorCanvas(Cursor.E_RESIZE_CURSOR, Color.lightGray, font));
	add("West", c = new Panel());
	c.setLayout(new GridLayout(1, 1));
	c.add(new CursorCanvas(Cursor.W_RESIZE_CURSOR, Color.lightGray, font));
	add("Center", c = new Panel());
	c.setLayout(new GridLayout(3, 3));
	c.add(new CursorCanvas(Cursor.DEFAULT_CURSOR, Color.red, font));
	c.add(new CursorCanvas(Cursor.CROSSHAIR_CURSOR, Color.magenta, font));
	c.add(new CursorCanvas(Cursor.TEXT_CURSOR, Color.blue, font));
        c.add(new CursorCanvas("images/NoDrop.gif", new Point(15, 15), 
                               "Custom", Color.white, font));
        c.add(new CursorCanvas("images/NoDrop.gif", new Point(15, 15), 
                               "Custom", Color.lightGray, font));
        c.add(new CursorCanvas("images/NoDrop.gif", new Point(15, 15), 
                               "Custom", Color.darkGray, font));
	c.add(new CursorCanvas(Cursor.WAIT_CURSOR, Color.yellow, font));
	c.add(new CursorCanvas(Cursor.HAND_CURSOR, Color.green, font));
	c.add(new CursorCanvas(Cursor.MOVE_CURSOR, Color.cyan, font));
	Dimension psize = getPreferredSize();
	Dimension size = getSize();
	if (psize.width < size.width || psize.height < size.height) {
	    resize(Math.max(psize.width, size.width),
		   Math.max(psize.height, size.height));
	}
    }

    static class MyAdapter extends WindowAdapter {
        public void windowClosing(WindowEvent e) {
            System.exit(0);
        }
    }

    public static void main(String args[]) {
	Frame f = new Frame();
        f.setTitle("CursorTest");
        f.addWindowListener(new MyAdapter());

	CursorTest cursortest = new CursorTest();
	cursortest.init();
	cursortest.start();

	f.add("Center", cursortest);
	f.setSize(500, 400);
	f.show();
    }
}

class CursorCanvas extends Canvas {
    Cursor cursor;

    public CursorCanvas(int curnum, Color bg, Font font) {
        cursor = Cursor.getPredefinedCursor(curnum);
	setCursor(cursor);
	setBackground(bg);
        setFont(font);
    }

    public CursorCanvas(String gif, Point hotspot, String name, 
                        Color bg, Font font) {
        Image image = getToolkit().getImage(gif);	
        cursor = Toolkit.getDefaultToolkit().createCustomCursor(
            image, hotspot, name);
	setCursor(cursor);
	setBackground(bg);
        setFont(font);
    }

    public void paint(Graphics g) {
	Dimension d = getSize();
	FontMetrics fm = g.getFontMetrics();
	int x = (d.width - fm.stringWidth(cursor.getName())) / 2;
	int y = ((d.height - fm.getHeight()) / 2) + fm.getAscent();
	g.drawString(cursor.getName(), x, y);
    }

    public Dimension getPreferredSize() {
	return getMinimumSize();
    }

    public Dimension getMinimumSize() {
	FontMetrics fm = getFontMetrics(getFont());
	return new Dimension(fm.stringWidth(cursor.getName() + "  "),
			     fm.getHeight() * 2);
    }
}
