/*
 * @(#)AffineTransformTest.java	1.13 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;

public class AffineTransformTest {
    public static final int STACKSIZE = 5;

    public static void main(String argv[]) {
	checkconstructors();
	AffineTransform t = new AffineTransform();
	System.out.println(matstring(t));
	t.setToTranslation(4, 5);
	System.out.println(matstring(t));
	t.setToScale(4, 5);
	System.out.println(matstring(t));
	t.setToRotation(Math.PI / 2);
	System.out.println(matstring(t));
	tryConcat(0);
	tryConcat(1);
	tryConcat(2);
	tryConcat(-1);
    }

    static void checkconstructors() {
	double matd3[] = {2.0, 2.0, 2.0};
	double matd4[] = {2.0, 2.0, 2.0, 2.0};
	double matd5[] = {2.0, 2.0, 2.0, 2.0, 2.0};
	double matd6[] = {2.0, 2.0, 2.0, 2.0, 2.0, 2.0};
	float matf3[] = {2.0f, 2.0f, 2.0f};
	float matf4[] = {2.0f, 2.0f, 2.0f, 2.0f};
	float matf5[] = {2.0f, 2.0f, 2.0f, 2.0f, 2.0f};
	float matf6[] = {2.0f, 2.0f, 2.0f, 2.0f, 2.0f, 2.0f};
	System.out.println("Constructor tests:");
	System.out.println("double 1x3   "+construct(matd3, true, true));
	System.out.println("double 1x4   "+construct(matd4, true, false));
	System.out.println("double 1x5   "+construct(matd5, true, false));
	System.out.println("double 1x6   "+construct(matd6, false, false));
	System.out.println("float 1x3    "+construct(matf3, true, true));
	System.out.println("float 1x4    "+construct(matf4, true, false));
	System.out.println("float 1x5    "+construct(matf5, true, false));
	System.out.println("float 1x6    "+construct(matf6, false, false));
	System.out.println();
    }

    static String passfail(boolean pass) {
	return (pass ? "passed: " : "FAILED: ");
    }

    static boolean check(AffineTransform at, boolean expectshort) {
	double matrix[] = new double[6];
	at.getMatrix(matrix);
	for (int i = 0; i < 6; i++) {
	    boolean wasset = (matrix[i] == 2.0);
	    if (i < 4) {
		if (!wasset) {
		    return false;
		}
	    } else {
		if (wasset == expectshort) {
		    return false;
		}
	    }
	}
	return true;
    }

    static String construct(double[] mat,
			    boolean expectshort, boolean expectfail) {
	try {
	    AffineTransform at = new AffineTransform(mat);
	    return (passfail(!expectfail && check(at, expectshort)) + at);
	} catch (Exception e) {
	    return (passfail(expectfail) + e);
	}
    }

    static String construct(float[] mat,
			    boolean expectshort, boolean expectfail) {
	try {
	    AffineTransform at = new AffineTransform(mat);
	    return (passfail(!expectfail && check(at, expectshort)) + at);
	} catch (Exception e) {
	    return (passfail(expectfail) + e);
	}
    }

    public static void tryConcat(int type) {
	AffineTransform stack[] = new AffineTransform[STACKSIZE];
	String txmult, txcoll, itxmult;
	for (int i = 0; i < STACKSIZE; i++) {
	    stack[i] = makeRandomTransform(type);
	    System.out.println("stack["+i+"]: "+matstring(stack[i]));
	}
	AffineTransform collapsed = new AffineTransform();
	for (int i = STACKSIZE - 1; i >= 0; i--) {
	    collapsed.concatenate(stack[i]);
	}
	System.out.println("concat: "+matstring(collapsed));
	Point2D p1 = new Point2D.Double();
	Point2D p2 = new Point2D.Double();
	Point2D p3 = new Point2D.Double();
	for (int i = 0; i < 4; i++) {
	    float x = (float) (Math.random() * 100);
	    float y = (float) (Math.random() * 100);
	    p1.setLocation(x, y);
	    p2.setLocation(x, y);
	    p3.setLocation(x, y);
	    System.out.println(ptstring(p1));
	    for (int j = 0; j < STACKSIZE; j++) {
		stack[j].transform(p2, p2);
	    }
	    report(p2, null, "=>", "=>", false);
	    collapsed.transform(p1, p3);
	    report(p3, p2, "==", "=>", false);
	    for (int j = STACKSIZE-1; j >= 0; j--) {
                try {
                    stack[j].inverseTransform(p2, p2);
                } catch (NoninvertibleTransformException e) {
                    System.out.println ("Got noninvertible transform");
                    e.printStackTrace();
                }
	    }
	    report(p2, p1, "<=", "<=", true);
            try {
                collapsed.inverseTransform(p3, p3);
            } catch (NoninvertibleTransformException e) {
                System.out.println ("Got noninvertible transform");
                e.printStackTrace();
            }

	    report(p3, p1, "==", "<=", true);
	}
    }

    public static void report(Point2D p1, Point2D p2,
			      String rpt, String last, boolean trail) {
	String s = "\t";
	if (p2 != null) {
	    s = diffComment(p1, p2) + s;
	}
	if (!trail) {
	    s += dup(rpt, STACKSIZE-1) + last;
	}
	s += ptstring(p1);
	if (trail) {
	    s += last + dup(rpt, STACKSIZE-1);
	}
	System.out.println(s);
    }

    public static String dup(String s, int copies) {
	String t = "";
	while (--copies >= 0) {
	    t += s;
	}
	return t;
    }

    public static double pround(double val) {
	return Math.rint(val * 1E6) / 1E6;
    }

    public static String matstring(AffineTransform t) {
	double mat[] = new double[6];
	t.getMatrix(mat);
	return ("["
		+pround(mat[0])+", "+pround(mat[1])+", "
		+pround(mat[2])+", "+pround(mat[3])+", "
		+pround(mat[4])+", "+pround(mat[5])+"]");
    }

    public static String ptstring(Point2D p) {
	return "["+pround(p.getX())+", "+pround(p.getY())+"]";
    }

    public static double scaledDiff(double v1, double v2) {
	double avg = (Math.abs(v1) + Math.abs(v2)) / 2;
	if (avg < 1.0) {
	    avg = 1.0;
	}
	double diff = Math.abs(v1 - v2);
	return diff / avg;
    }

    public static String diffComment(Point2D p1, Point2D p2) {
	double sdiff = scaledDiff(p1.getX(), p2.getX())
	    + scaledDiff(p1.getY(), p2.getY());
	if (sdiff == 0.0) {
	    return "Exact";
	} else if (sdiff < 1E-14) {
	    return "Good";
	} else if (sdiff < .0001) {
	    return "Fair";
	} else {
	    return "Wrong";
	}
    }

    public static AffineTransform makeRandomTransform(int type) {
	AffineTransform t = new AffineTransform();
	if (type < 0) {
	    type = (int) (Math.random() * 3);
	}
	double v1 = Math.round(Math.random() * 800) / 8;
	double v2 = Math.round(Math.random() * 800) / 8;
	switch (type) {
	case 0:
	    t.setToTranslation(v1 - 50, v2 - 50);
	    break;
	case 1:
	    t.setToRotation(v1 / 100 * Math.PI);
	    break;
	case 2:
	    v1 = Math.max(v1, 1.0);	/* 1.0 => 100.0 */
	    v2 = Math.max(v2, 1.0);	/* 1.0 => 100.0 */
	    if (Math.random() < .5) {
		v1 = (v1 / 2.0 + 50.0) / 100.0;		/* .5 => 1.0 */
	    }
	    if (Math.random() < .5) {
		v2 = (v2 / 2.0 + 50.0) / 100.0;		/* .5 => 1.0 */
	    }
	    t.setToScale(v1, v2);
            break;
	default:
	    System.err.println("????? type = "+type+" ?????");
	}
	return t;
    }
}
