/*
 * @(#)AlphaCompstRules.java	1.13 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.applet.Applet;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.font.*;

public class AlphaCompstRules extends Applet
{
    static final int TESTWIDTH    = 60;
    static final int TESTHEIGHT   = 30;
    static final int TESTPADLEFT  = 15;
    static final int TESTPADRIGHT = 15;
    static final int TESTHPAD     = (TESTPADLEFT + TESTPADRIGHT);
    static final int TESTPADABOVE = 20;
    static final int TESTPADBELOW = 10;
    static final int TESTVPAD     = (TESTPADABOVE + TESTPADBELOW);

    static final int TESTPADDEDWIDTH  = (TESTWIDTH  + TESTHPAD);
    static final int TESTPADDEDHEIGHT = (TESTHEIGHT + TESTVPAD);

    BufferedImage s = new BufferedImage(TESTWIDTH, TESTHEIGHT,
					BufferedImage.TYPE_INT_ARGB);
    BufferedImage d = new BufferedImage(TESTWIDTH, TESTHEIGHT,
					BufferedImage.TYPE_INT_ARGB);
    Graphics2D gS = s.createGraphics();
    Graphics2D gD = d.createGraphics();

    GeneralPath srcpath = new GeneralPath();
    GeneralPath dstpath = new GeneralPath();

    static Color srcfull = new Color(255, 0, 0, 255);
    static Color srcfade = new Color(255, 0, 0, 64);
    static Paint srcgrad = new GradientPaint(0, 0, srcfull,
					     TESTWIDTH, 0, srcfade,
					     false);

    static Color dstfull = new Color(0, 0, 255, 255);
    static Color dstfade = new Color(0, 0, 255, 64);
    static Paint dstgrad = new GradientPaint(0, 0, dstfade,
					     0, TESTHEIGHT, dstfull,
					     false);

    static Paint[] srcpaints = {
	srcfull, srcfade, srcfull, srcfade,
	srcfull, srcgrad, srcfull, srcgrad,
    };
    static Paint[] dstpaints = {
	dstfull, dstfull, dstfade, dstfade,
	dstfull, dstfull, dstgrad, dstgrad,
    };

    static String compNames[] = {
	"(Source)",
	"(Dest)",
	"Clear",
	"Src",
	"SrcOver",
	"DstOver",
	"SrcIn",
	"DstIn",
	"SrcOut",
	"DstOut",
    };
    static AlphaComposite compObjs[] = {
	AlphaComposite.Src,
	AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.0f),
	AlphaComposite.Clear,
	AlphaComposite.Src,
	AlphaComposite.SrcOver,
	AlphaComposite.DstOver,
	AlphaComposite.SrcIn,
	AlphaComposite.DstIn,
	AlphaComposite.SrcOut,
	AlphaComposite.DstOut,
    };

    static final int ROWS    = srcpaints.length * 2;
    static final int COLUMNS = compObjs.length;

    public AlphaCompstRules() {
	gS.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			    RenderingHints.VALUE_ANTIALIAS_ON);

	gD.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			    RenderingHints.VALUE_ANTIALIAS_ON);

        srcpath.moveTo(0,0);
        srcpath.lineTo(TESTWIDTH,0);
	srcpath.lineTo(TESTWIDTH,TESTHEIGHT/2);
        srcpath.lineTo(0,TESTHEIGHT);
        srcpath.closePath();

        dstpath.moveTo(0,0);
	dstpath.lineTo(0,TESTHEIGHT/2);
        dstpath.lineTo(TESTWIDTH,TESTHEIGHT);
        dstpath.lineTo(TESTWIDTH,0);
        dstpath.closePath();

        setBackground(Color.white);
    }

    public Dimension getPreferredSize() {
	return new Dimension(TESTPADDEDWIDTH  * COLUMNS,
			     TESTPADDEDHEIGHT * ROWS);
    }

    public void paint(Graphics g) {
	Graphics2D g2d = (Graphics2D) g;
	g2d.translate(TESTPADLEFT, TESTPADABOVE);

	Font f = getFont();
	FontRenderContext frc = g2d.getFontRenderContext();
	LineMetrics lm = f.getLineMetrics(compNames[0], frc);
	float descent = lm.getDescent();

	for (int k = 0; k < 2; k++) {
	    for (int j = 0; j < 8; j++) {
		AffineTransform at = g2d.getTransform();

		for (int i = 0; i < compObjs.length; i++) {
		    gD.setComposite(AlphaComposite.Clear);
		    gD.fillRect(0, 0, TESTWIDTH, TESTHEIGHT);

		    if (i != 0) {
			gD.setComposite(AlphaComposite.Src);
			gD.setPaint(dstpaints[j]);
			gD.fill(dstpath);
		    }

		    if (k == 0) {
			gD.setComposite(compObjs[i]);
			gD.setPaint(srcpaints[j]);
			gD.fill(srcpath);
		    } else {
			gS.setComposite(AlphaComposite.Clear);
			gS.fillRect(0, 0, TESTWIDTH, TESTHEIGHT);

			gS.setComposite(AlphaComposite.Src);
			gS.setPaint(srcpaints[j]);
			gS.fill(srcpath);

			gD.setComposite(compObjs[i]);
			gD.drawImage(s, 0, 0, null);
		    }

		    // display the composite result
		    g2d.setColor(Color.white);
		    g2d.fillRect(- TESTPADLEFT, - TESTPADABOVE,
				 TESTPADDEDWIDTH, TESTPADDEDHEIGHT);
		    g2d.drawImage(d, 0, 0, null);
		    g2d.setColor(Color.black);
		    g2d.drawString(compNames[i], 0, - descent);
		    if (k == 0 && i == 0) {
			g2d.draw(srcpath);
		    } else {
			g2d.drawRect(0, 0, TESTWIDTH, TESTHEIGHT);
		    }

		    // move to the next position
		    g2d.translate(TESTPADDEDWIDTH, 0);
		}

		g2d.setTransform(at);
		g2d.translate(0, TESTPADDEDHEIGHT);
	    }
	}
    }
	
    public static void main(String s[]) {
        Frame f = new Frame("Test for AlphaComposite rules");
        f.add("Center", new AlphaCompstRules());
        f.pack();
        f.show();
    }
}


