/*
 * @(#)BezierAnim.java	1.27 98/07/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
 
import java.applet.Applet;
import java.awt.Frame;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Checkbox;
import java.awt.CheckboxGroup;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.BasicStroke;
import java.awt.AlphaComposite;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.Shape;
import java.awt.RenderingHints;
import java.awt.font.GlyphVector;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

public class BezierAnim extends Applet {
    static boolean verbose = false;

    BezierAnimPanel animpanel;

    public void init() {
	setLayout(new BorderLayout());
	animpanel = new BezierAnimPanel();
	add("Center", animpanel);
    }

    public void start() {
	animpanel.start();
    }

    public void stop() {
	animpanel.stop();
    }

    public Dimension getPreferredSize() {
	return new Dimension(445, 270);
    }

    public static void main(String argv[]) {
	BezierAnim.verbose = true;
	Frame f = new Frame();
	final BezierAnim ba = new BezierAnim();
	f.add(ba);
	f.addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {System.exit(0);}
	    public void windowIconified(WindowEvent e) {ba.stop();}
	    public void windowDeiconified(WindowEvent e) {ba.start();}
	});
	f.pack();
	ba.init();
	f.show();
	ba.start();
    }
}

class BezierAnimPanel extends Panel implements ItemListener {
    BezierAnimCanvas animcv;

    Checkbox animtoggle;
    Checkbox drawtoggle;
    Checkbox dashtoggle;
    Checkbox filltoggle;
    Checkbox blendtoggle;

    Checkbox cliptoggle;
    Checkbox cleartoggle;
    Checkbox antialiastoggle;
    Checkbox dblbuftoggle;
    Checkbox blittoggle;

    public BezierAnimPanel() {
	setLayout(new BorderLayout());
	animcv = new BezierAnimCanvas();
	add("Center", animcv);
	Panel p = new Panel();
	add("South", p);
	animtoggle = new Checkbox("Animate", false);
	animtoggle.addItemListener(this);
	p.add(animtoggle);
	drawtoggle = new Checkbox("Stroke", true);
	drawtoggle.addItemListener(this);
	p.add(drawtoggle);
	dashtoggle = new Checkbox("(with dashes)", true);
	dashtoggle.addItemListener(this);
	p.add(dashtoggle);
	filltoggle = new Checkbox("Fill", false);
	filltoggle.addItemListener(this);
	p.add(filltoggle);
	blendtoggle = new Checkbox("(with alpha)", true);
	blendtoggle.addItemListener(this);
	p.add(blendtoggle);
	p = new Panel();
	add("North", p);
	cliptoggle = new Checkbox("Clip", false);
	cliptoggle.addItemListener(this);
	p.add(cliptoggle);
	cleartoggle = new Checkbox("Clear", true);
	cleartoggle.addItemListener(this);
	p.add(cleartoggle);
	antialiastoggle = new Checkbox("Antialias", true);
	antialiastoggle.addItemListener(this);
	p.add(antialiastoggle);
	dblbuftoggle = new Checkbox("Offscreen", true);
	dblbuftoggle.addItemListener(this);
	p.add(dblbuftoggle);
	blittoggle = new Checkbox("(show offscr)", true);
	blittoggle.addItemListener(this);
	p.add(blittoggle);

	dashtoggle.setEnabled(animcv.doDraw);
	blendtoggle.setEnabled(animcv.doFill);
	blittoggle.setEnabled(animcv.dblBuffer);
    }

    public void start() {
	animcv.start();
    }

    public void stop() {
	animcv.stop();
    }

    public void itemStateChanged(ItemEvent e) {
	if (e.getSource() == drawtoggle) {
	    animcv.doDraw = (e.getStateChange() == ItemEvent.SELECTED);
	    dashtoggle.setEnabled(animcv.doDraw);
	} else if (e.getSource() == filltoggle) {
	    animcv.doFill = (e.getStateChange() == ItemEvent.SELECTED);
	    blendtoggle.setEnabled(animcv.doFill);
	} else if (e.getSource() == dashtoggle) {
	    animcv.doDash = (e.getStateChange() == ItemEvent.SELECTED);
	} else if (e.getSource() == animtoggle) {
	    animcv.doAnim = (e.getStateChange() == ItemEvent.SELECTED);
	} else if (e.getSource() == cliptoggle) {
	    animcv.doClip = (e.getStateChange() == ItemEvent.SELECTED);
	} else if (e.getSource() == cleartoggle) {
	    animcv.doClear = (e.getStateChange() == ItemEvent.SELECTED);
	} else if (e.getSource() == blittoggle) {
	    animcv.doBlit = (e.getStateChange() == ItemEvent.SELECTED);
	} else if (e.getSource() == blendtoggle) {
	    animcv.doBlend = (e.getStateChange() == ItemEvent.SELECTED);
	} else if (e.getSource() == dblbuftoggle) {
	    animcv.dblBuffer = (e.getStateChange() == ItemEvent.SELECTED);
	    blittoggle.setEnabled(animcv.dblBuffer);
	} else if (e.getSource() == antialiastoggle) {
	    animcv.antiAlias = (e.getStateChange() == ItemEvent.SELECTED);
	} else {
	    System.out.println("unknown item in event: "+e.getSource());
	}
    }
}

class BezierAnimCanvas extends Canvas implements Runnable {
    public static final int REPORTFRAMES = 30;
    public static final int NUMPTS = 6;
    float animpts[] = new float[NUMPTS * 2];
    float deltas[] = new float[NUMPTS * 2];
    float staticpts[] = {
	 50.0f,   0.0f,
	150.0f,   0.0f,
	200.0f,  75.0f,
	150.0f, 150.0f,
	 50.0f, 150.0f,
	  0.0f,  75.0f,
    };
    float movepts[] = new float[staticpts.length];

    boolean doAnim = false;
    boolean doClear = true;
    boolean doFill = false;
    boolean doDraw = true;
    boolean doDash = true;
    boolean doBlit = true;
    boolean doBlend = true;
    boolean dblBuffer = true;
    boolean antiAlias = true;
    boolean doClip = false;

    BufferedImage img;

    Thread anim;

    public void start() {
	Dimension size = getSize();
	for (int i = 0; i < animpts.length; i += 2) {
	    animpts[i + 0] = (float) (Math.random() * size.width);
	    animpts[i + 1] = (float) (Math.random() * size.height);
	    deltas[i + 0] = (float) (Math.random() * 4.0 + 2.0);
	    deltas[i + 1] = (float) (Math.random() * 4.0 + 2.0);
	    if (animpts[i + 0] > size.width / 2.0f) {
		deltas[i + 0] = -deltas[i + 0];
	    }
	    if (animpts[i + 1] > size.height / 2.0f) {
		deltas[i + 1] = -deltas[i + 1];
	    }
	}
	anim = new Thread(this);
	anim.setPriority(Thread.MIN_PRIORITY);
	anim.start();
    }

    public synchronized void stop() {
	anim = null;
	notify();
    }

    public void animate(float[] pts, float[] deltas, int index, int limit) {
	float newpt = pts[index] + deltas[index];
	if (newpt <= 0) {
	    newpt = -newpt;
	    deltas[index] = (float) (Math.random() * 4.0 + 2.0);
	} else if (newpt >= (float) limit) {
	    newpt = 2.0f * limit - newpt;
	    deltas[index] = - (float) (Math.random() * 4.0 + 2.0);
	}
	pts[index] = newpt;
    }

    public void run() {
	Thread me = Thread.currentThread();
	Graphics2D g2d = null;
	Graphics2D BufferG2D = null;
	Graphics2D ScreenG2D = null;
	BasicStroke solid = new BasicStroke(10.0f, BasicStroke.CAP_BUTT,
					    BasicStroke.JOIN_ROUND, 10.0f);
	float dash[] = { 10.0f };
	BasicStroke dashed = new BasicStroke(10.0f, BasicStroke.CAP_BUTT,
					     BasicStroke.JOIN_ROUND, 10.0f,
					     dash, 0.0f);
	GeneralPath gp = new GeneralPath(GeneralPath.WIND_NON_ZERO);
	int rule = AlphaComposite.SRC_OVER;
	AlphaComposite opaque = AlphaComposite.SrcOver;
	AlphaComposite blend = AlphaComposite.getInstance(rule, 0.5f);
	AlphaComposite set = AlphaComposite.Src;
	int frame = 0;
	int frametmp = 0;
	int frametarget = REPORTFRAMES;
	long orig = System.currentTimeMillis();
	long start = orig;
	Dimension oldSize = getSize();
	boolean prevClip = false;
	Shape clippath = null;
	while (anim == me) {
	    Dimension size = getSize();
	    if (size.width != oldSize.width || size.height != oldSize.height) {
		img = null;
		clippath = null;
		if (BufferG2D != null) {
		    BufferG2D.dispose();
		    BufferG2D = null;
		}
		if (ScreenG2D != null) {
		    ScreenG2D.dispose();
		    ScreenG2D = null;
		}
	    }
	    oldSize = size;
	    if (dblBuffer) {
		if (img == null) {
		    img = (BufferedImage) createImage(size.width, size.height);
		}
		if (BufferG2D == null) {
		    BufferG2D = img.createGraphics();
		    BufferG2D.setRenderingHint(RenderingHints.KEY_RENDERING,
					       RenderingHints.VALUE_RENDER_DEFAULT);
		    BufferG2D.setClip(prevClip ? clippath : null);
		}
		g2d = BufferG2D;
	    } else {
		if (ScreenG2D == null) {
		    ScreenG2D = (Graphics2D) getGraphics();
		    ScreenG2D.setRenderingHint(RenderingHints.KEY_RENDERING,
					       RenderingHints.VALUE_RENDER_DEFAULT);
		    ScreenG2D.setClip(prevClip ? clippath : null);
		}
		g2d = ScreenG2D;
	    }
	    if (clippath == null) {
		Font font = getFont();
                String str = "Java";
                GlyphVector gv =
		    font.createGlyphVector(g2d.getFontRenderContext(), str);
		Shape strpath = gv.getOutline(0, 0);
		Rectangle2D strbounds = strpath.getBounds2D();
		AffineTransform at = new AffineTransform();
		at.translate(size.width / 8.0, size.height / 8.0);
		at.scale(size.width / strbounds.getWidth() * 0.75,
			 size.height / strbounds.getHeight() * 0.75);
		at.translate(-strbounds.getX(), -strbounds.getY());
		clippath = at.createTransformedShape(strpath);
	    }
	    float[] ctrlpts;
	    if (doAnim) {
		for (int i = 0; i < animpts.length; i += 2) {
		    animate(animpts, deltas, i + 0, size.width);
		    animate(animpts, deltas, i + 1, size.height);
		}
		ctrlpts = animpts;
	    } else {
		ctrlpts = movepts;
		animate(animpts, deltas, 0, size.width-200);
		animate(animpts, deltas, 1, size.height-150);
		for (int i = 0; i < ctrlpts.length; i += 2) {
		    ctrlpts[i + 0] = staticpts[i + 0] + animpts[0];
		    ctrlpts[i + 1] = staticpts[i + 1] + animpts[1];
		}
	    }
	    int len = ctrlpts.length;
	    gp.reset();
	    int dir = 0;
	    float prevx = ctrlpts[len - 2];
	    float prevy = ctrlpts[len - 1];
	    float curx = ctrlpts[0];
	    float cury = ctrlpts[1];
	    float midx = (curx + prevx) / 2.0f;
	    float midy = (cury + prevy) / 2.0f;
	    gp.moveTo(midx, midy);
	    for (int i = 2; i <= ctrlpts.length; i += 2) {
		float x1 = (midx + curx) / 2.0f;
		float y1 = (midy + cury) / 2.0f;
		prevx = curx;
		prevy = cury;
		if (i < ctrlpts.length) {
		    curx = ctrlpts[i + 0];
		    cury = ctrlpts[i + 1];
		} else {
		    curx = ctrlpts[0];
		    cury = ctrlpts[1];
		}
		midx = (curx + prevx) / 2.0f;
		midy = (cury + prevy) / 2.0f;
		float x2 = (prevx + midx) / 2.0f;
		float y2 = (prevy + midy) / 2.0f;
		gp.curveTo(x1, y1, x2, y2, midx, midy);
	    }
	    gp.closePath();

	    // Snapshot doClip to avoid race with GUI thread.
	    boolean newClip = doClip;
	    // If the clip state is transitioning then we want to do the
	    // clear without any clip to avoid garbage outside the clip
	    // and then set the clip if needed after the clear.
	    if (prevClip && !newClip) {
		if (BufferG2D != null) {
		    BufferG2D.setClip(null);
		}
		if (ScreenG2D != null) {
		    ScreenG2D.setClip(null);
		}
	    }
	    if (doClear) {
		g2d.setComposite(set);
		g2d.setBackground(java.awt.Color.white);
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				     RenderingHints.VALUE_ANTIALIAS_OFF);
		g2d.clearRect(0, 0, size.width, size.height);
	    }
	    if (newClip && !prevClip) {
		if (BufferG2D != null) {
		    BufferG2D.setStroke(solid);
		    BufferG2D.setComposite(opaque);
		    BufferG2D.setColor(java.awt.Color.red);
		    BufferG2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					       RenderingHints.VALUE_ANTIALIAS_ON);
		    BufferG2D.draw(clippath);
		    BufferG2D.setClip(clippath);
		}
		if (ScreenG2D != null) {
		    ScreenG2D.setStroke(solid);
		    ScreenG2D.setComposite(opaque);
		    ScreenG2D.setColor(java.awt.Color.red);
		    ScreenG2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					       RenderingHints.VALUE_ANTIALIAS_ON);
		    ScreenG2D.draw(clippath);
		    ScreenG2D.setClip(clippath);
		}
	    }
	    prevClip = newClip;

	    g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				 (antiAlias
				  ? RenderingHints.VALUE_ANTIALIAS_ON
				  : RenderingHints.VALUE_ANTIALIAS_OFF));
	    if (doDraw) {
		g2d.setColor(java.awt.Color.blue);
		g2d.setComposite(opaque);
		g2d.setStroke(doDash ? dashed : solid);
		g2d.draw(gp);
	    }
	    if (doFill) {
		g2d.setColor(java.awt.Color.green);
		g2d.setComposite(doBlend ? blend : opaque);
		g2d.fill(gp);
	    }
	    if (g2d == BufferG2D) {
		if (doBlit) {
		    //repaint();
		    Graphics g = getGraphics();
		    update(g);
		    g.dispose();
		}
	    }
	    ++frame;
	    if (++frametmp >= frametarget) {
		long end = System.currentTimeMillis();
		long rel = (end - start);
		if (rel < 500) {
		    if (rel < 100) {
			frametarget *= 2;
		    } else {
			frametarget = (int) (501 * frametarget / rel);
		    }
		} else {
		    getToolkit().sync();
		    end = System.currentTimeMillis();
		    rel = (end - start);
		    long tot = (end - orig);
		    if (BezierAnim.verbose) {
			System.out.println("frame: "+frame
					   +", time: "+rel+" ms"
					   +", rate: "+frametmp / (rel/1000.0f)
					   +", all: "+ frame / (tot/1000.0f));
		    }
		    start = System.currentTimeMillis();
		    frametmp = 0;
		    frametarget = REPORTFRAMES;
		}
	    }
	    Thread.yield();
	}
	if (g2d != null) {	
	    g2d.dispose();
	}
    }

    public void update(Graphics g) {
	paint(g);
    }

    public void paint(Graphics g) {
	Graphics2D g2d = (Graphics2D) g;
	if (img != null) {
	    int imgw = img.getWidth();
	    int imgh = img.getHeight();
	    g2d.setComposite(AlphaComposite.Src);
	    g2d.drawImage(img, null, 0, 0);
	}
    }
}
