/*
 * @(#)Blitters2D.java	1.8 98/08/24
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Frame;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.Shape;
import java.awt.Point;

import java.awt.image.ComponentColorModel;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.color.ColorSpace;
import java.awt.Transparency;
import java.awt.image.DataBuffer;


import java.applet.Applet;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Event;
import java.lang.Math;



public class Blitters2D extends Applet implements Runnable
{
    private static int DRAWSIZE = 150;
    private static int INTERMEDIA = 5;


    Thread myThread;

    private void ReportHeader() {
        System.out.println(">=================================================");
        System.out.println("SEQUENCE STARTED");
        System.out.println(">=================================================");
    }

    private void ReportFooter() {
        System.out.println(">=================================================");
        System.out.println("SEQUENCE FINISHED");
        System.out.println(">=================================================");
    }

    private void PerformTest(Graphics g,TestObject testObject) {
        boolean ok = true;

        System.out.println(">=================================================");

        try
            {
                System.out.println(testObject.getDescription());
                testObject.initialize(g,DRAWSIZE,INTERMEDIA);
                testObject.performTest();
            } catch (Exception e)   {
                String msg = e.getMessage();
                e.printStackTrace();
                ok = false;
            } finally   {
                if( true == ok )
                    {
                        System.out.println("passed");   
                    } else  {
                        System.out.println("failed");   
                    }
            }
            System.out.println(">=================================================");
    }

    private void BasicTests(Graphics g) {
        ReportHeader();

	PerformTest(g,new TestCurrent_INT_RGB_Current());
	PerformTest(g,new TestCurrent_BYTE_INDEXED_Current());
	PerformTest(g,new TestCurrent_INT_RGB_Current());        
	PerformTest(g,new TestCurrent_INT_ARGB_Current());  
	PerformTest(g,new TestCurrent_INT_ARGB_PRE_Current());    
        
	PerformTest(g,new TestCurrent_BYTE_BINARY_Current());                            
        PerformTest(g,new TestCurrent_BYTE_GRAY_Current());                         
        PerformTest(g,new TestCurrent_USHORT_GRAY_Current());                         
        PerformTest(g,new TestCurrent_USHORT_565_RGB_Current());                     
        PerformTest(g,new TestCurrent_USHORT_555_RGB_Current());                     
        PerformTest(g,new TestCurrent_4BYTE_ABGR_Current());      
        PerformTest(g,new TestCurrent_4BYTE_ABGR_PRE_Current());                    
        PerformTest(g,new TestCurrent_3BYTE_BGR_Current()); 
        
        PerformTest(g,new TestCurrent_INT_BGR_Current());     
        ReportFooter();                                
    }
	
  
    private void TestBlitters(Graphics g) {
        BasicTests(g);
    }

    public void start() {
        if (myThread == null) {
            myThread = new Thread(this);
        }
	myThread.setPriority(Thread.MAX_PRIORITY);
        myThread.start();
    }

    public void stop() {
        if (myThread != null) {
            myThread.stop();
            myThread = null;
        }
    }

    public void run() {
        Graphics graphics = getGraphics();
 
        TestBlitters(graphics);
    }

    public Dimension getPreferredSize() {
        return new Dimension(461, 220);
    }

    public void paint(Graphics g) {
        BasicTests(g);
    }
    
    public static void main(String[] argv) {
        Frame f = new Frame();
        f.add(new Blitters2D());
        f.pack();
        f.show();
    }
}



abstract class TestObject 
{

    Graphics		graphics;
    BufferedImage	offscreen;
    Rectangle		originalArea;
    Rectangle		controlCopyArea;
    Rectangle		offscreenCopyArea;
    Rectangle		offscreenStretchCopyArea;

    Color			colorRed;
    Color			colorGreen;
    Color			colorBlue;
    Color			colorBlack;
    Color			colorWhite;

    Color			savedColor;
    Color			colorYellow;
    //------------- begin group
    //  Initializes internal structure
    public void initialize(Graphics givenGraphics,int drawSize,int gap)
	{
            originalArea		= new Rectangle(0,0,drawSize,drawSize);
            controlCopyArea		= new Rectangle(drawSize + gap,0,drawSize,drawSize);
            offscreenCopyArea	= new Rectangle(2*(drawSize+gap),0,drawSize,drawSize);
            offscreenStretchCopyArea = new Rectangle(3*(drawSize+gap),0,drawSize,drawSize);


            graphics		= givenGraphics;
            offscreen		= makeOffscreen();

            colorRed		= new Color(0xFF,0x00,0x00);
            colorGreen		= new Color(0x00,0xFF,0x00);
            colorBlue		= new Color(0x00,0x00,0xFF);
            colorWhite		= new Color(0xFF,0xFF,0xFF);
            colorBlack		= new Color(0x00,0x00,0x00);

            colorYellow     = new Color(0xFF,0xFF,0x00);
	}
    //------------- end group

    public void waitLoop(int delay)
        {
            long    start = System.currentTimeMillis();

            while( System.currentTimeMillis() - start < delay)
                {
                } 
        }

    //------------- begin group
    //  Draw primitives
    void drawBar(Graphics g,int x,int y,int h, int v,Color c)
	{
            savedColor = g.getColor();

            g.setColor(c);
		
            g.fillRect(x,y,h,v);

            g.setColor(savedColor);
	}
    void drawRect(Graphics g,int x,int y,int h, int v,Color c)
	{
            savedColor = g.getColor();

            g.setColor(c);
		
            g.drawRect(x,y,h,v);

            g.setColor(savedColor);
	}
    void drawRing(Graphics g,int x,int y,int h, int v,Color c)
	{
            savedColor = g.getColor();

            g.setColor(c);
		
            g.drawOval(x,y,h,v);

            g.setColor(savedColor);
	}
    void drawGrid(Graphics g,int x,int y,int h, int v,int d,Color c)
	{
            savedColor = g.getColor();

            g.setColor(c);
		
            {
                int		i;

                for(i=y;i<y+v;i+=d)
                    {
                        g.drawLine(x,i,x+h,i);
                    }
                for(i=x;i<x+v;i+=d)
                    {
                        g.drawLine(i,y,i,y+v);
                    }
            }
		

            g.setColor(savedColor);
	}
    void typeLine(Graphics g,int x,int y,String s,Color c)
	{
            savedColor = g.getColor();

            g.setColor(c);
		

            g.drawString(s,x,y);

            g.setColor(savedColor);
	}
    //------------- end group

    //------------- begin group
    //  Make patterns on abstract locations
    void makeOriginal(Graphics g,int x,int y,int width,int height)
	{

            x+=30;
            y+=30;


            drawBar(g,x,y,40,40,colorRed);
            drawBar(g,x+50,y,40,40,colorGreen);
            drawBar(g,x+25,y+50,40,40,colorBlue);

            drawGrid(g,x + 20,y + 20,51,51,5,colorBlack);

            drawRing(g,x-30,y-30,width,height,colorBlack);   

	}

    void makeControlCopy(Graphics g,int x,int y,int width,int height,int dx,int dy)
	{
            g.copyArea(x,y,width,height,dx,dy);
	}
    //------------- end group

    //------------- begin group
    //  Draw area content
    void drawOriginalOnItsArea()
        {
            makeOriginal(graphics,originalArea.x,originalArea.y,originalArea.width,originalArea.height);
        }

    void drawOriginalOnOffscreen()
        {
            Graphics gOff = offscreen.getGraphics();
            drawBar(gOff,1,1,originalArea.width,originalArea.height,colorWhite);
            makeOriginal(gOff,0,0,originalArea.width,originalArea.height); 
        }
    //------------- end group

    //------------- begin group
    //  Prepare areas
    public void prepareOriginalArea()
	{

            graphics.clearRect( originalArea.x,originalArea.y,
                                originalArea.width,originalArea.height);

            drawRect(   graphics,originalArea.x,originalArea.y,
                        originalArea.width,originalArea.height,colorBlack);

            typeLine( graphics,originalArea.x,originalArea.y + originalArea.height + 15,"Image",colorBlack);
            drawBar(graphics,originalArea.x,originalArea.y + originalArea.height + 16,200,35,colorWhite);
            typeLine( graphics,originalArea.x,originalArea.y + originalArea.height + 35,getDescription(),colorBlack);
	}


    public void prepareControlCopyArea()
	{
            graphics.clearRect(	controlCopyArea.x,controlCopyArea.y,
                                controlCopyArea.width,controlCopyArea.height);


            drawRect(   graphics,controlCopyArea.x,controlCopyArea.y,
                        controlCopyArea.width,controlCopyArea.height,colorBlack);

            typeLine(   graphics,controlCopyArea.x,controlCopyArea.y + controlCopyArea.height + 15,
                        "Image's control copy",colorBlack);
	}
    public void prepareOffscreenCopyArea()
	{
            graphics.clearRect(	offscreenCopyArea.x,offscreenCopyArea.y,
                                offscreenCopyArea.width,offscreenCopyArea.height);


            drawRect(   graphics,offscreenCopyArea.x,offscreenCopyArea.y,
                        offscreenCopyArea.width,offscreenCopyArea.height,colorBlack);

            typeLine(   graphics,offscreenCopyArea.x,offscreenCopyArea.y + offscreenCopyArea.height + 15,
                        "Offscreen copy",colorBlack);
	}
    /*
      public void prepareOffscreenStretchCopyArea()
      {
      graphics.clearRect(	offscreenStretchCopyArea.x,offscreenStretchCopyArea.y,
      offscreenStretchCopyArea.width,offscreenStretchCopyArea.height);


      drawRect(   graphics,offscreenStretchCopyArea.x,offscreenStretchCopyArea.y,
      offscreenStretchCopyArea.width,offscreenStretchCopyArea.height,colorBlack);

      typeLine(   graphics,offscreenStretchCopyArea.x,
      offscreenStretchCopyArea.y + offscreenStretchCopyArea.height + 15,
      "Stretched offscreen copy",colorBlack);
      }
      */
    //------------- end group

    //------------- begin group
    //  Copy fragments
    public void copyOriginalToControlCopyArea()
        {
            makeControlCopy(    graphics,originalArea.x,originalArea.y,
                                originalArea.width,originalArea.height,
                                controlCopyArea.x - originalArea.x,0);
        }   
    public void copyOffscreenToOffscreenCopyArea()
        {
            graphics.drawImage(offscreen,offscreenCopyArea.x,offscreenCopyArea.y,null);

        }
    /*
      public void copyOffscreenToOffscreenStretchCopyArea()
      {
      int     x = offscreenStretchCopyArea.x+1;
      int     y = offscreenStretchCopyArea.y+1;
      int     width = offscreenStretchCopyArea.width;
      int     height = offscreenStretchCopyArea.height;

      graphics.drawImage( offscreen,x,y,width / 2,height / 2,null);

      graphics.drawImage( offscreen,x + (width / 2),y + (height / 2),
      width / 3,height / 3,null);

      graphics.drawImage( offscreen,x + (width / 2) + 1,y,
      width / 4,height / 4,null);

      }
      */

    //------------- end group

    
    public void performTest()
	{
            //		prepareOffscreenStretchCopyArea(); 

            prepareOffscreenCopyArea();

            prepareControlCopyArea();

            prepareOriginalArea();

            drawOriginalOnItsArea();

            drawOriginalOnOffscreen();

            copyOriginalToControlCopyArea();

            copyOffscreenToOffscreenCopyArea();

            //        copyOffscreenToOffscreenStretchCopyArea();

            waitLoop(2000);

	}

    abstract public BufferedImage makeOffscreen();
    abstract public String getDescription();
}

class TestCurrent_INT_RGB_Current extends TestObject 
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_RGB);	
	}
    public String getDescription()
        {
            return "TYPE_INT_RGB";
        }
}
class TestCurrent_INT_ARGB_Current extends TestObject 
{                                                                               

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_ARGB);	
	}
    public String getDescription()
        {
            return "TYPE_INT_ARGB";
        }
}
class TestCurrent_INT_ARGB_PRE_Current extends TestObject 
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_ARGB_PRE);	
	}
    public String getDescription()
        {
            return "TYPE_INT_ARGB_PRE";
        }
}
class TestCurrent_INT_BGR_Current extends TestObject 
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_BGR);	
	}
    public String getDescription()
        {
            return "TYPE_INT_BGR";
        }
}
class TestCurrent_3BYTE_BGR_Current extends TestObject 
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_3BYTE_BGR);	
	}
    public String getDescription()
        {
            return "TYPE_3BYTE_BGR";
        }
}
class TestCurrent_4BYTE_ABGR_Current extends TestObject 
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_4BYTE_ABGR);	
	}
    public String getDescription()
        {
            return "TYPE_4BYTE_ABGR";
        }
}
class TestCurrent_4BYTE_ABGR_PRE_Current extends TestObject                        
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_4BYTE_ABGR_PRE);	
	}
    public String getDescription()
        {
            return "TYPE_4BYTE_ABGR_PRE";
        }
}
class TestCurrent_USHORT_565_RGB_Current extends TestObject                        
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_USHORT_565_RGB);	
	}
    public String getDescription()
        {
            return "TYPE_USHORT_565_RGB";
        }
}
class TestCurrent_USHORT_555_RGB_Current extends TestObject                        
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_USHORT_555_RGB);	
	}
    public String getDescription()
        {
            return "TYPE_USHORT_555_RGB";
        }
}
class TestCurrent_BYTE_GRAY_Current extends TestObject                        
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_GRAY);	
	}
    public String getDescription()
        {
            return "TYPE_BYTE_GRAY";
        }
}
class TestCurrent_USHORT_GRAY_Current extends TestObject                        
{

    public BufferedImage makeOffscreen()
        {
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_USHORT_GRAY);	
	}
    
    public String getDescription()
        {
            return "TYPE_USHORT_GRAY";
        }
}
class TestCurrent_BYTE_BINARY_Current extends TestObject                        
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_BINARY);	
	}
    public String getDescription()
        {
            return "TYPE_BINARY";
        }
}
class TestCurrent_BYTE_INDEXED_Current extends TestObject                        
{

    public BufferedImage makeOffscreen()
	{
            return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_INDEXED);	
	}
    public String getDescription()
        {
            return "TYPE_BYTE_INDEXED";
        }
}
