/*
 * @(#)Composite.java	1.26 98/05/03
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


import java.applet.Applet;
import java.awt.Frame;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Button;
import java.awt.Color;
import java.awt.Component;
import java.awt.TextField;
import java.awt.Label;
import java.awt.Image;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.BorderLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Scrollbar;
import java.awt.Toolkit;
import java.awt.AWTEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.AdjustmentEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;
import java.awt.image.ImageObserver;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.AlphaComposite;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.BasicStroke;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import sun.awt.image.URLImageSource;
import sun.awt.image.FileImageSource;


import java.net.URL;
import java.net.MalformedURLException;

public class Composite extends Applet implements AdjustmentListener,
                                                ActionListener,
                                                MouseListener,
                                                MouseMotionListener,
                                                Runnable
{
    static final int WIDTH      = 0;
    static final int SKEW       = 1;
    static final int SCALE      = 2;
    static final int ROTATE     = 3;
    static final int ALPHA      = 4;

    static final int MIN = 0;
    static final int DEF = 1;
    static final int MAX = 2;

    static final int ORGWIDTH = 430;
    static final int ORGHEIGHT = 256;

    static double limits[][] = {
	{0.0,		20.0,	80.0},
	{-2.0,		0.0,	2.0},
	{0.01,		1.0,	3.0},
	{-Math.PI,	0.0,	Math.PI},
	{0.0,	        0.5,	1.0},
    };

    static float mx[] = new float[3];
    static float my[] = new float[3];
    static float ax, ay;
    static int count;
    static float delta = 10.0f;


    String sliderLabel[] = {"Width", "Skew", "Scale", "Rotate", "Alpha"};
    Scrollbar widthSlider = new Scrollbar(Scrollbar.HORIZONTAL, 0, 20, 0, 100);
    Scrollbar alphaSlider = new Scrollbar(Scrollbar.HORIZONTAL, 0, 20, 0, 100);
    Scrollbar scaleSlider = new Scrollbar(Scrollbar.HORIZONTAL, 0, 20, 0, 100);
    Scrollbar rotateSlider =
			new Scrollbar(Scrollbar.HORIZONTAL, 0, 20, 0, 100);
    Scrollbar skewSlider = new Scrollbar(Scrollbar.HORIZONTAL, 0, 20, 0, 100);

    Scrollbar slider[] = {widthSlider, skewSlider, scaleSlider, rotateSlider,
			  alphaSlider};


    TextField fields[] = new TextField[sliderLabel.length];
    Panel p = new Panel();
    TextField fileNameTxtFld = new TextField("writeonce.gif", 20);
    Label statusLabel = new Label();
    Button clearButton = new Button("Clear");
    Button resetButton = new Button("Reset");
    GridBagLayout gb = new GridBagLayout();
    GridBagConstraints gbConstraints = new GridBagConstraints();
    ImageCanvas imgCanvas = new ImageCanvas(0, 0, ORGWIDTH, ORGHEIGHT);
    BufferedImage img = null;
    GeneralPath mpath = null;
    Thread myThread;
    int width = ORGWIDTH;
    int height = ORGHEIGHT;

    public Composite()
    {
        
        imgCanvas.addMouseListener(this);
        imgCanvas.addMouseMotionListener(this);
        
	setLayout(new BorderLayout());
	add("North", imgCanvas);
        p.setLayout(gb);

	addComponent(new Label("Background Image", Label.RIGHT), 0, 0, 1, 1, 0.0, 0.0);
	addComponent(fileNameTxtFld, 1, 0, 1, 1, 0.0, 0.0);
	fileNameTxtFld.addActionListener(this);

        addComponent(clearButton, 2, 0, 1, 1, 0.0, 0.0);
        clearButton.addActionListener(this);

        addComponent(resetButton, 3, 0, 1, 1, 0.0, 0.0);
        resetButton.addActionListener(this);
        
	for (int i = 0 ; i < sliderLabel.length ; i++) {
	    addComponent(new Label(sliderLabel[i], Label.RIGHT), 0, 2+i, 1, 1, 0.0, 0.0);
	    addComponent(slider[i], 1, 2+i, 1, 1, 0.0, 0.0);
	    fields[i] = new TextField("", 6);
	    addComponent(fields[i], 2, 2+i, 1, 1, 0.0, 0.0);
	    slider[i].addAdjustmentListener(this);
	}
        
        resetSliders();

	addComponent(statusLabel, 0, 7, 3, 1, 1.0, 0.0);

	add("South", p);
// 	pack();
	setVisible(true);

	showStatus("Ready...");

        img = goGetImage(fileNameTxtFld.getText());
//         this.start();
     }

    void resetSliders() {
        for (int i = 0; i < limits.length; i++) {
            double def = limits[i][DEF];
            fields[i].setText(roundstr(def));
            slider[i].setValue(sliderset(i, def));
            handleValue(i, def);
        }
    }

    public void start() {
        mx[0] = (float)width*.35f; my[0] = (float)height*.50f;
        mx[1] = (float)width*.50f; my[1] = (float)height*.50f;
        mx[2] = (float)width*.75f; my[2] = (float)height*.50f;
        
        if (myThread == null) {
            myThread = new Thread(this);
        }
	myThread.setPriority(Thread.MIN_PRIORITY);
        myThread.start();
    }

    public void stop() {
        if (myThread != null) {
            myThread.stop();
            myThread = null;
            mpath = null;
        }
    }
    
    public void run() {
       	Dimension size = imgCanvas.getPreferredSize();
        int height = size.height;
        
        while(true) {
            my[1] += delta;
            mpath = new GeneralPath();
            mpath.moveTo(mx[0], my[0]);
            mpath.curveTo(mx[0] ,my[0], mx[1] ,my[1], mx[2] ,my[2]);
            if ((int)my[1] > height || (int)my[1] < 0) {
                delta = -delta;
            }
            imgCanvas.repaint();
            try {
                myThread.sleep(10);
            } catch (InterruptedException e) {
            }
        }
    }

    public void addComponent(Component comp, int x, int y, int w, int h,
					double weightx, double weighty) {

	GridBagConstraints gbc = new GridBagConstraints();
	gbc.fill = GridBagConstraints.BOTH;
	gbc.gridx = x;
	gbc.gridy = y;
	gbc.gridwidth = w;
	gbc.gridheight = h;
	gbc.weightx = weightx;
	gbc.weighty = weighty;
	((GridBagLayout)p.getLayout()).setConstraints(comp, gbc);
	p.add(comp);
    }

    BufferedImage goGetImage(String s) {
            
        BufferedImageConsumer pg;

        try {
            URL url = new URL(s);
            pg = new BufferedImageConsumer(new URLImageSource(url), false);
        } catch (MalformedURLException e) {
            pg = new BufferedImageConsumer(new FileImageSource(s), false);
        }
        try {
            pg.grabPixels();
            // Check for errors
            if ((pg.status() & ImageObserver.ABORT) != 0) {
                showStatus ("Error while fetching image");
                return null;
            }
        } catch(Exception e) {
            return null;
        }
        
        return pg.getImage();
              
    }

    public void actionPerformed(ActionEvent e) {

	if (e.getSource() == fileNameTxtFld) {
            showStatus("Loading...");
            img = goGetImage(fileNameTxtFld.getText());
	    if (img != null) {
		if (img.getWidth(imgCanvas) < 0 ||
		    img.getHeight(imgCanvas) < 0) {
		    showStatus("Loading...");
		} else {
                    imgCanvas.changeSize(img.getWidth(), img.getHeight());
                    showStatus("Done loading. Ready...");
		}
	    } else {
		showStatus("Failed to load image " + fileNameTxtFld.getText());
	    }
  	} else if (e.getSource() == clearButton) {
            mpath = null;
            imgCanvas.doDrawStar = false;
            imgCanvas.repaint();
            this.stop();
        } else if (e.getSource() == resetButton) {
            resetSliders();
            mpath = null;
            imgCanvas.doDrawStar = false;
            imgCanvas.repaint();
            this.stop();
            this.start();
        } else {
	    int i;
	    for (i = 0; i < fields.length; i++) {
		if (e.getSource() == fields[i]) {
		    handleValue(i, fieldVal(i));
		    imgCanvas.repaint();
		    imgCanvas.doDrawStar = true;
		    break;
		}
	    }
	}
    }

    public void handleValue(int index, double val) {
	switch (index) {
	case WIDTH:
	    imgCanvas.penWidth = val;
	    break;
	case SKEW:
	    imgCanvas.skew = val;
	    break;
	case SCALE:
	    imgCanvas.scale = val;
	    break;
	case ROTATE:
	    imgCanvas.rotate = val;
	    break;
	case ALPHA:
	    imgCanvas.alpha = val;
	    break;
	}
    }

    public String roundstr(double value) {
	return Double.toString(Math.round(value * 100.0) / 100.0);
    }

    public int sliderset(int fieldIndex, double val) {
	return (int) (100.0 * ((val - limits[fieldIndex][MIN])
			       / (limits[fieldIndex][MAX]
				  - limits[fieldIndex][MIN])));
    }

    public double fieldVal(int fieldIndex) {
	TextField t = fields[fieldIndex];
	String val = t.getText();
	double ret;
	try {
	    ret = Double.valueOf(val).doubleValue();
	} catch (NumberFormatException e) {
	    ret = limits[fieldIndex][DEF];
	}
	ret = Math.max(limits[fieldIndex][MIN],
		       Math.min(limits[fieldIndex][MAX], ret));
	t.setText(roundstr(ret));
	slider[fieldIndex].setValue(sliderset(fieldIndex, ret));
	return ret;
    }

    public double sliderVal(int fieldIndex) {
	Scrollbar s = slider[fieldIndex];
	int val = s.getValue();
	double ret = (limits[fieldIndex][MIN]
		      + val / 100.0 * (limits[fieldIndex][MAX]
				       - limits[fieldIndex][MIN]));
	fields[fieldIndex].setText(roundstr(ret));
	return ret;
    }

    public void adjustmentValueChanged(AdjustmentEvent e) {
	Object target = e.getSource();
	for (int i = 0; i < slider.length; i++) {
	    if (target == slider[i]) {
		handleValue(i, sliderVal(i));
		imgCanvas.repaint();
		imgCanvas.doDrawStar = true;
		break;
	    }
	}
    }
    
    public void mouseDragged(MouseEvent e) {

        if (mpath != null) {
            mx[count] = (float)e.getX();
            my[count] = (float)e.getY();
            if (count == 2) {
                mpath.curveTo(mx[0] ,my[0], mx[1] ,my[1], mx[2] ,my[2]);
                imgCanvas.repaint();
                
            }
            
            count = ++count % 3;
        } else {
            mpath = new GeneralPath();
            ax = (float)e.getX();
            ay = (float)e.getY();
            mpath.moveTo(ax, ay);
            count = 0;
        }
        imgCanvas.doDrawStar = false;
        this.stop();
    }

    public void mouseMoved(MouseEvent e) {
    }

    public void mouseClicked(MouseEvent e) {
        if (mpath != null) {
            mpath.lineTo((float)e.getX(), (float)e.getY());
            imgCanvas.repaint();
        } else {
            mpath = new GeneralPath();
            ax = (float)e.getX();
            ay = (float)e.getY();
            mpath.moveTo(ax, ay);
        }
        this.stop();
    }


    public void mouseEntered(MouseEvent e) {
    }
    
    public void mouseExited(MouseEvent e) {
    }
    
    public void mousePressed(MouseEvent e) {
    }
    
    public void mouseReleased(MouseEvent e) {
    }
    

    public void showStatus(String s) {
	statusLabel.setText(s);
    }



class ImageCanvas extends Canvas {

    int x;
    int y;
    Dimension size = new Dimension();
    boolean doDrawStar = false;
    double penWidth = 40.0;
    double scale = 1.0;
    double rotate = 0.0;
    double skew = 0.0;
    double alpha = 0.3;
    float  x1, y1;
    float  x2, y2;
    float  x3, y3;
    float  x4, y4;
    float  x5, y5;
    float  x6, y6;
    BufferedImage bufImg;
    Graphics2D gi;
    
    void drawStar(Graphics g) {
	AlphaComposite composite = AlphaComposite.getInstance(
				    AlphaComposite.SRC_OVER, (float)alpha);
    	GeneralPath p = new GeneralPath();
	Dimension size = imgCanvas.getPreferredSize();

	int width = size.width;
	int height = size.height;
 
	Graphics2D g2 = (Graphics2D) g;
        AffineTransform savedTx = g2.getTransform();

	g2.translate(width / 2.0, height / 2.0);
	g2.scale(scale, scale);
	g2.rotate(rotate);
	g2.shear(-skew, 0.0);

	p.moveTo(- width / 2.0f, - height / 8.0f);
	p.lineTo(+ width / 2.0f, - height / 8.0f);
	p.lineTo(- width / 4.0f, + height / 2.0f);
	p.lineTo(+         0.0f, - height / 2.0f);
	p.lineTo(+ width / 4.0f, + height / 2.0f);
	p.closePath();

	g2.setPaint(new Color(255, 0, 0));
	g2.setComposite(composite);
	g2.setStroke(new BasicStroke((float)penWidth, BasicStroke.CAP_ROUND,
			 BasicStroke.JOIN_ROUND));
	g2.draw(p);

        g2.setTransform(savedTx);
	doDrawStar = false;
    }

    void drawMouse(Graphics g) {
	AlphaComposite composite = AlphaComposite.getInstance(
				    AlphaComposite.SRC_OVER, (float)alpha);
	Dimension size = imgCanvas.getPreferredSize();

	int width = size.width;
	int height = size.height;

       	Graphics2D g2 = (Graphics2D) g;
        AffineTransform savedTx = g2.getTransform();

	g2.translate(width / 2.0, height / 2.0);
	g2.scale(scale, scale);
	g2.rotate(rotate);
	g2.shear(-skew, 0.0);
	g2.translate(-width / 2.0, -height / 2.0);
        
        g2.setPaint(new Color(0, 0, 255));
	g2.setComposite(composite);
       	g2.setStroke(new BasicStroke((float)penWidth, BasicStroke.CAP_ROUND,
			 BasicStroke.JOIN_ROUND));
        g2.draw(mpath);
        g2.setTransform(savedTx);
        
        //mpath = new GeneralPath();
        //mpath.moveTo(ax, ay);
    }

        
    public void changeSize(int w, int h) {

        this.setSize(w, h);
	size.setSize(w, h);
//         pack();

        bufImg = createBufferedImageDefault(w, h);
        gi = bufImg.createGraphics();
        Rectangle2D cl = new Rectangle2D.Float(200f, 150f, 100f, 100f);
        //        gi.setClip(cl);

        gi.setBackground(Color.white);
    }

    public ImageCanvas(int x, int y, int w, int h) {

	this.x = x;
	this.y = y;

	changeSize(w, h);

    }

    public BufferedImage createBufferedImageDefault(int w, int h) {
	bufImg = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);

        return bufImg;
    }


    public Dimension getPreferredSize()
    {
	return size;
    }

    public void paint(Graphics g)
    {
        Graphics2D g2 = (Graphics2D)g;
        g2.setComposite(AlphaComposite.Src);
        
        
	if (img != null) {
            gi.setComposite(AlphaComposite.Src);
            gi.drawImage(img, null, 0, 0);
	    showStatus("Done loading. Ready...");
	} else {
            gi.clearRect(0, 0, bufImg.getWidth(), bufImg.getHeight());
        }

        if (mpath != null) {
            drawMouse(gi);
        } else if (doDrawStar) {
	    drawStar(gi);
	}
        g2.drawImage(bufImg, null, 0, 0);
    }

    public void update(Graphics g) {
        paint(g);
    }

    public boolean imageUpdate(Image img, int info, int x, int y, int w, int h) {
	super.imageUpdate(img, info, x, y, w, h);
	if ((info & ImageObserver.HEIGHT) != 0 && (info & ImageObserver.WIDTH) != 0) {
	    changeSize(w, h);
	}
	return true;
    }

}

}
