/*
 * @(#)ConvolvePipe.java	1.14 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.applet.Applet;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ConvolveOp;
import java.awt.image.Kernel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;


public class ConvolvePipe extends Pipe implements ActionListener {
    int numRows = 3;
    int numCols = 3;
    TextField[] tf;
    Kernel kernel;
    float[] data = {-1.f, -1.f, -1.f,
                    -1.f, 8.f, -1.f,
                    -1.f, -1.f, -1.f};



    public ConvolvePipe(Pipeline pipeline, Applet applet,
                        BufferedImage bimage) {
        super(pipeline, applet, bimage, true);

        String kstring = applet.getParameter("kernel");
        if (kstring != null) {
            int curidx = 0;
            int nextidx;
            String space = " ";
            for (int i=0; i < 9; i++) {
                nextidx = kstring.indexOf(space, curidx);
                if (nextidx == -1) nextidx=kstring.length()-1;
                data[i] = Float.valueOf(
                            kstring.substring(curidx,nextidx)).floatValue();
                curidx = nextidx+1;
            }
        }

        kernel = new Kernel(3,3, data);
    }

    public ConvolvePipe(Pipeline pipeline, Applet applet,
                        BufferedImage bimage, Kernel initial) {
        super(pipeline, applet, bimage, false);
        kernel = initial;
        process();
    }

    public Container addGui() {
        GridBagLayout gb = new GridBagLayout();
        Panel p = new Panel();
        p.setLayout(gb);
        int idx=0;
        tf = new TextField[numRows*numCols];

        for (int row=0; row < numRows; row++) {
            for (int col=0; col < numCols; col++) {
                tf[idx] = new TextField(Float.toString(data[idx]));
                addComponent(p, gb, tf[idx], col, row, 1, 1, 0., 0.);
                tf[idx].addActionListener(this);
                idx++;
            }
        }

        return p;
    }

    public void actionPerformed(ActionEvent e) {
        String text;
        int idx = 0;
        for (int row=0; row < numRows; row++) {
            for (int col=0; col < numCols; col++, idx++) {
                text = tf[idx].getText();
                data[idx] = Float.valueOf(text).floatValue();
            }
        }
        kernel = new Kernel(numRows, numCols, data);

        process();
    }

    public void process() {
        long first, last;
        ConvolveOp convolveOp = new ConvolveOp(kernel,
                                               ConvolveOp.EDGE_ZERO_FILL,
                                               null);

        // Need to create a new data buffer
        if (usedRef) {
            createDestRaster();
        }

        first = System.currentTimeMillis();
        convolveOp.filter(inImage, outImage);
        last = System.currentTimeMillis();
        System.out.println("elapsed: "+(last-first+1));
        pipeline.trigger(this);
    }

    public String toString() {
        return "ConvolvePipe";
    }
}
