/*
 * @(#)CurveDraw.java	1.8 98/08/12
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
    
import java.applet.Applet;
import java.awt.Shape;
import java.awt.Choice;
import java.awt.Label;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Frame;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.QuadCurve2D;
import java.awt.geom.CubicCurve2D;
import java.awt.image.BufferedImage;
import java.awt.event.ItemListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.ItemEvent;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

public class CurveDraw extends Applet {
    public CurveDraw() {
	setLayout(new BorderLayout());
	CurveDrawCanvas sdcv = new CurveDrawCanvas();
	add("Center", sdcv);
	CurveDrawControls sdctl = new CurveDrawControls(sdcv);
	add("South", sdctl);
    }

    public static void main(String argv[]) {
	Frame f = new Frame();
	f.add(new CurveDraw());
	f.addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {System.exit(0);}
	});
	f.pack();
	f.show();
    }
}

class CurveDrawControls extends Panel implements ItemListener {
    Choice curvechoice;
    CurveDrawCanvas canvas;

    public CurveDrawControls(CurveDrawCanvas cv) {
	canvas = cv;
	setLayout(new FlowLayout());
	add(new Label("Curve type:"));
	curvechoice = new Choice();
	curvechoice.add("Quadratic");
	curvechoice.add("Cubic");
	curvechoice.addItemListener(this);
	add(curvechoice);
	chooseEditor("Quadratic");
    }

    void chooseEditor(String s) {
	try {
	    Class curveEditClass = Class.forName(s+"Editor");
	    canvas.setEditor((CurveEditor) curveEditClass.newInstance());
	} catch (Exception exc) {
	    exc.printStackTrace();
	}
    }

    /**
     * Invoked when the state of an item has been changed.
     */
    public void itemStateChanged(ItemEvent e) {
	if (e.getStateChange() == ItemEvent.SELECTED) {
	    if (e.getSource() == curvechoice) {
		chooseEditor(e.getItem().toString());
	    }
	}
    }
}

abstract class CurveEditor {
    public abstract void setCurve(Shape sh);

    public abstract Shape getCurve();

    public abstract Point2D getControlPoint(int index);

    public abstract int getNumControlPoints();

    public double getDistanceSq(Point2D p1, Point2D p2) {
	double dx = p1.getX() - p2.getX();
	double dy = p1.getY() - p2.getY();
	return (dx * dx + dy * dy);
    }

    public int getControlIndex(Point2D p) {
	int index = 0;
	double dist = getDistanceSq(p, getControlPoint(0));
	for (int i = getNumControlPoints(); --i > 0;) {
	    double newdist = getDistanceSq(p, getControlPoint(i));
	    if (newdist < dist) {
		dist = newdist;
		index = i;
	    }
	}
	return index;
    }

    public abstract void moveControlPoint(int index, double dx, double dy);
}

class QuadraticEditor extends CurveEditor {
    QuadCurve2D curve;

    public Shape getCurve() {
	return curve;
    }

    public void setCurve(Shape sh) {
	if (sh == null) {
	    curve = new QuadCurve2D.Float(50.0f, 300.0f,
					  200.0f, 100.0f,
					  350.0f, 300.0f);
	} else if (sh instanceof CubicCurve2D) {
	    CubicCurve2D cubic = (CubicCurve2D) sh;
	    double x1 = cubic.getX1();
	    double y1 = cubic.getY1();
	    double cx1 = cubic.getCtrlX1();
	    double cy1 = cubic.getCtrlY1();
	    double cx2 = cubic.getCtrlX2();
	    double cy2 = cubic.getCtrlY2();
	    double x2 = cubic.getX2();
	    double y2 = cubic.getY2();
	    double cx = ((3.0 * cx1 - x1) + (3.0 * cx2 - x2)) / 4.0;
	    double cy = ((3.0 * cy1 - y1) + (3.0 * cy2 - y2)) / 4.0;
	    curve = new QuadCurve2D.Float((float) x1, (float) y1,
					  (float) cx, (float) cy,
					  (float) x2, (float) y2);
	} else {
	    setCurve(null);
	}
    }

    public int getNumControlPoints() {
	return 3;
    }

    public Point2D getControlPoint(int index) {
	switch (index) {
	case 0:
	    return new Point2D.Double(curve.getX1(), curve.getY1());
	case 1:
	    return new Point2D.Double(curve.getCtrlX(), curve.getCtrlY());
	case 2:
	    return new Point2D.Double(curve.getX2(), curve.getY2());
	default:
	    throw new IndexOutOfBoundsException("no such quadratic control point: "+index);
	}
    }

    public void moveControlPoint(int index, double dx, double dy) {
	double x1 = curve.getX1();
	double y1 = curve.getY1();
	double ctrlx = curve.getCtrlX();
	double ctrly = curve.getCtrlY();
	double x2 = curve.getX2();
	double y2 = curve.getY2();
	switch (index) {
	case 0:
	    x1 += dx;
	    y1 += dy;
	    break;
	case 1:
	    ctrlx += dx;
	    ctrly += dy;
	    break;
	case 2:
	    x2 += dx;
	    y2 += dy;
	    break;
	default:
	    throw new IndexOutOfBoundsException("no such quadratic control point: "+index);
	}
	curve.setCurve(x1, y1, ctrlx, ctrly, x2, y2);
    }
}

class CubicEditor extends CurveEditor {
    CubicCurve2D curve;

    public Shape getCurve() {
	return curve;
    }

    public void setCurve(Shape sh) {
	if (sh == null) {
	    curve = new CubicCurve2D.Float(50.0f, 300.0f,
					   150.0f, 100.0f,
					   250.0f, 100.0f,
					   350.0f, 300.0f);
	} else if (sh instanceof QuadCurve2D) {
	    QuadCurve2D quad = (QuadCurve2D) sh;
	    double x1 = quad.getX1();
	    double y1 = quad.getY1();
	    double ctrlx = quad.getCtrlX();
	    double ctrly = quad.getCtrlY();
	    double x2 = quad.getX2();
	    double y2 = quad.getY2();
	    double cx1 = (x1 + 2.0 * ctrlx) / 3.0;
	    double cy1 = (y1 + 2.0 * ctrly) / 3.0;
	    double cx2 = (x2 + 2.0 * ctrlx) / 3.0;
	    double cy2 = (y2 + 2.0 * ctrly) / 3.0;
	    curve = new CubicCurve2D.Float((float) x1, (float) y1,
					   (float) cx1, (float) cy1,
					   (float) cx2, (float) cy2,
					   (float) x2, (float) y2);
	} else {
	    setCurve(null);
	}
    }

    public int getNumControlPoints() {
	return 4;
    }

    public Point2D getControlPoint(int index) {
	switch (index) {
	case 0:
	    return new Point2D.Double(curve.getX1(), curve.getY1());
	case 1:
	    return new Point2D.Double(curve.getCtrlX1(), curve.getCtrlY1());
	case 2:
	    return new Point2D.Double(curve.getCtrlX2(), curve.getCtrlY2());
	case 3:
	    return new Point2D.Double(curve.getX2(), curve.getY2());
	default:
	    throw new IndexOutOfBoundsException("no such cubic control point: "+index);
	}
    }

    public void moveControlPoint(int index, double dx, double dy) {
	double x1 = curve.getX1();
	double y1 = curve.getY1();
	double ctrlx1 = curve.getCtrlX1();
	double ctrly1 = curve.getCtrlY1();
	double ctrlx2 = curve.getCtrlX2();
	double ctrly2 = curve.getCtrlY2();
	double x2 = curve.getX2();
	double y2 = curve.getY2();
	switch (index) {
	case 0:
	    x1 += dx;
	    y1 += dy;
	    break;
	case 1:
	    ctrlx1 += dx;
	    ctrly1 += dy;
	    break;
	case 2:
	    ctrlx2 += dx;
	    ctrly2 += dy;
	    break;
	case 3:
	    x2 += dx;
	    y2 += dy;
	    break;
	default:
	    throw new IndexOutOfBoundsException("no such cubic control point: "+index);
	}
	curve.setCurve(x1, y1, ctrlx1, ctrly1, ctrlx2, ctrly2, x2, y2);
    }
}

class CurveDrawCanvas extends Canvas
    implements MouseListener, MouseMotionListener
{
    CurveEditor editor;
    Point2D origin;
    Rectangle2D shadow = new Rectangle2D.Float();
    boolean mouseinside;
    boolean shadowinside;
    boolean shadowtouches;
    int dragindex = -1;
    int nearindex = -1;

    public CurveDrawCanvas() {
	addMouseListener(this);
	addMouseMotionListener(this);
    }

    public void setEditor(CurveEditor newEditor) {
	newEditor.setCurve((editor != null) ? editor.getCurve() : null);
	editor = newEditor;
	repaint();
    }

    public Dimension getPreferredSize() {
	return new Dimension(500, 500);
    }

    public void mouseClicked(MouseEvent e) {}

    public void mousePressed(MouseEvent e) {
	Point2D p = e.getPoint();
	origin = p;
	dragindex = editor.getControlIndex(p);
	repaint();
    }

    public void mouseReleased(MouseEvent e) {
	if (dragindex >= 0) {
	    mouseDragged(e);
	    dragindex = -1;
	    origin = null;
	    repaint();
	}
	mouseMoved(e);
    }

    public void mouseEntered(MouseEvent e) {
	mouseMoved(e);
    }

    public void mouseExited(MouseEvent e) {
	if (dragindex >= 0) {
	    mouseReleased(e);
	} else {
	    mouseMoved(e);
	}
    }

    public void mouseDragged(MouseEvent e) {
	if (dragindex >= 0) {
	    Point2D p = e.getPoint();
	    editor.moveControlPoint(dragindex,
				    p.getX() - origin.getX(),
				    p.getY() - origin.getY());
	    origin = p;
	    repaint();
	}
	mouseMoved(e);
    }

    public void mouseMoved(MouseEvent e) {
	Point2D p = e.getPoint();
	Shape curve = editor.getCurve();
	nearindex = editor.getControlIndex(p);
	shadow.setRect((float) p.getX() - 5.0f,
		       (float) p.getY() - 5.0f,
		       10.0f, 10.0f);
	mouseinside = curve.contains(p);
	shadowtouches = curve.intersects(shadow);
	shadowinside = curve.contains(shadow);
	repaint();
    }

    BufferedImage buf;

    public void paint(Graphics g) {
	if (buf == null) {
	    update(g);
	} else {
	    Graphics2D g2 = (Graphics2D) g;
	    g2.setComposite(AlphaComposite.Src);
	    g2.drawImage(buf, null, 0, 0);
	}
    }

    static BasicStroke solid =
	new BasicStroke(4.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
    static AlphaComposite highlight =
	AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.75f);

    public void update(Graphics g) {
	Dimension size = getSize();
	if (buf == null ||
	    buf.getWidth() != size.width ||
	    buf.getHeight() != size.height)
	{
	    buf = (BufferedImage) createImage(size.width, size.height);
	}
	Graphics2D g2d = buf.createGraphics();
	g2d.setColor(Color.gray);
	g2d.fillRect(0, 0, size.width, size.height);
	g2d.setStroke(solid);
	Shape curve = editor.getCurve();
	g2d.setColor(Color.black);
	g2d.draw(curve);
	g2d.setColor(shadowtouches
		     ? (mouseinside
			? (shadowinside ? Color.green : Color.yellow)
			: (shadowinside ? Color.red : Color.cyan))
		     : (mouseinside
			? (shadowinside ? Color.red : Color.red)
			: (shadowinside ? Color.red : Color.blue)));
	g2d.drawString("point "+(mouseinside ? "in" : "out"), 10, 10);
	g2d.drawString("shadow "+(shadowtouches
				  ? (shadowinside ? "in" : "touch")
				  : (shadowinside ? "bad" : "out")), 10, 20);
	g2d.fill(curve);
	g2d.setComposite(highlight);
	g2d.setColor(Color.darkGray);
	g2d.fill(shadow);
	Rectangle2D cprect = new Rectangle2D.Float();
	for (int i = editor.getNumControlPoints(); --i >= 0;) {
	    g2d.setColor(i == dragindex
			 ? Color.red
			 : (dragindex < 0 && i == nearindex
			    ? Color.white
			    : Color.darkGray));
	    Point2D p = editor.getControlPoint(i);
	    cprect.setRect((float) p.getX() - 5.0f,
			   (float) p.getY() - 5.0f,
			   10.0f, 10.0f);
	    g2d.fill(cprect);
	}
	g2d.dispose();
	Graphics2D g2 = (Graphics2D) g;
	g2.setComposite(AlphaComposite.Src);
	g2.drawImage(buf, null, 0, 0);
    }
}
