/*
 * @(#)GlyphSetTest.java	1.8 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.applet.Applet;
import java.awt.Frame;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.Font;
import java.awt.geom.AffineTransform;
import java.awt.font.FontMetrics2D;
import java.awt.font.GlyphSet;
import java.awt.font.TextAttribute;

public class GlyphSetTest extends Applet {
    public static void main(String argv[]) {
        Frame f = new Frame();
        f.add(new Tree());
        f.pack();
        f.show();
    }
    public Dimension getPreferredSize() {
        return new Dimension(600, 600);
    }

    public void paint(Graphics g) {
        Dimension size = getSize();
        int mindim = Math.min(size.width, size.height);
        Graphics2D g2d = (Graphics2D) g;

        /*AffineTransform at = new AffineTransform();
        at.translate((size.width - mindim) / 2.0,
                     (size.height - mindim) / 2.0);
        at.scale(mindim, mindim);
        at.translate(0.5, 0.5);
        at.scale(0.3, 0.3);
        at.translate(-(Twidth + Rwidth), FontHeight / 4.0);
        g2d.transform(at);
        */

        Font theFont = new Font("TimesRoman",Font.PLAIN,30);
        runTest(g,theFont);
    }

    void success(Graphics g) {
        Font testFont = new Font("Serif",Font.PLAIN,10);
        Dimension size = getSize();
        setFont(testFont);
        g.drawString(new String("All GlyphSet Tests have passed !!!"),(int)0,(int)(size.height - 12));
    }

    void failure(Graphics g,String msg) {
        Font testFont = new Font("Serif",Font.PLAIN,10);
        Dimension size = getSize();
        setFont(testFont);
        g.drawString(msg,(int)0,(int)(size.height - 12));
    }

    static int x_pos = 25;
    static int y_pos = 25;

    public void runTest(Graphics g,Font currentFont) {
        Graphics2D g2d = (Graphics2D) g;

        String theString = new String("abcdEFGHijkl");

        // a  b  c  d  H  G  F  E  i  j   k   l
        int[] baseVis =	{ 0, 1, 2, 3, 7, 6, 5, 4, 8, 9, 10, 11 };

        // 0   1   2   3   4   5   6   7   8   9  10  11
        float baseXadv[] = {  10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21 };

        // 0  1  2  3  4  5  6  7  8  9  10  11
        byte baseDir[] = { 0, 0, 0, 0, 1, 1, 1, 1, 0, 0,  0,  0 };

        byte baseline = TextAttribute.BASELINE_ROMAN.byteValue();

        GlyphSet set = g2d.getGlyphSet(theString.toCharArray(),0,
                                               theString.length(),baseline,
                                               baseVis,baseDir);
        FontMetrics2D fm2d = g2d.getFontMetrics2D();
        set.draw(g2d,x_pos,y_pos + fm2d.getAscent());

       {
            //subset test 1	 - take the first part of a glyph set
           String subString = "abcdEFGH";
           int vis[] = {0,1,2,3,7,6,5,4};
           float xadv[] = {10,11,12,13,14,15,16,17};
           byte dir[] = {0,0,0,0,1,1,1,1};
           GlyphSet expectedResult = g2d.getGlyphSet(subString.toCharArray(),
                                                              0,subString.length(),
                                                              baseline,vis,dir);

            GlyphSet actualResult = set.subset(0, 8);
            if (!actualResult.equals(expectedResult)) {
                failure(g,"GlyphSet: Test 1 failed");
                return;
            } else {
                y_pos += fm2d.getAscent() +  fm2d.getAscent() + fm2d.getDescent();
                actualResult.draw(g2d,x_pos,y_pos);
            }
        }

        {
            //subset test 2	 - take the end of the glyph set
            String subString = "GHijkl";
            int vis[] = { 1, 0, 2, 3, 4, 5};
            float xadv[] = {16, 17, 18, 19, 20, 21};
            byte dir[] = {1, 1, 0, 0, 0, 0};
            GlyphSet expectedResult = g2d.getGlyphSet(subString.toCharArray(),
                                                              0,subString.length(),
                                                              baseline,vis,dir);

            GlyphSet actualResult = set.subset(6, 12);
            if (!actualResult.equals(expectedResult)) {
                failure(g,"GlyphSet: Test 2 failed");
                return;
            } else {
                y_pos += fm2d.getAscent() + fm2d.getDescent();
                actualResult.draw(g2d,x_pos,y_pos);
            }
        }

        {
            //subset test 3 - take a section out of the center that starts at a multichar glyph
            String subString = "cdEFGHi";
            float xadv[] = 	{12, 13, 14, 15, 16, 17, 18};
            int vis[] = {0, 1, 5, 4, 3, 2, 6};
            byte[] dir = {0, 0, 1, 1, 1, 1, 0};
            GlyphSet expectedResult = g2d.getGlyphSet(subString.toCharArray(),
                                                              0,subString.length(),
                                                              baseline,vis,dir);

            GlyphSet actualResult = set.subset(2,9);
            if (!actualResult.equals(expectedResult)) {
                failure(g,"GlyphSet: Test 3 failed");
                return;
            } else {
                y_pos += fm2d.getAscent() + fm2d.getDescent();
                actualResult.draw(g2d,x_pos,y_pos);
            }
        }
        success(g);
    }
}
