/*
 * @(#)GradAnim.java	1.8 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
 
import java.applet.Applet;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.GradientPaint;
import java.awt.image.BufferedImage;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

public class GradAnim extends Applet implements Runnable {
    public Dimension getPreferredSize() {
	return new Dimension(300, 300);
    }

    Thread animator;

    public void start() {
	animator = new Thread(this);
	animator.start();
    }

    BufferedImage img;
    double framerate;

    class animval {
	float curval;
	float lowval;
	float highval;
	float currate;
	float lowrate;
	float highrate;

	public animval(int lowval, int highval,
		       int lowrate, int highrate) {
	    this.lowval = lowval;
	    this.highval = highval;
	    this.lowrate = lowrate;
	    this.highrate = highrate;
	    this.curval = randval(lowval, highval);
	    this.currate = randval(lowrate, highrate);
	}

	public float randval(float low, float high) {
	    return (float) (low + Math.random() * (high - low));
	}

	public float getFlt() {
	    return curval;
	}

	public int getInt() {
	    return (int) curval;
	}

	public void anim() {
	    curval += currate;
	    clip();
	}

	public void clip() {
	    if (curval > highval) {
		curval = highval - (curval - highval);
		if (curval < lowval) {
		    curval = highval;
		}
		currate = - randval(lowrate, highrate);
	    } else if (curval < lowval) {
		curval = lowval + (lowval - curval);
		if (curval > highval) {
		    curval = lowval;
		}
		currate = randval(lowrate, highrate);
	    }
	}

	public void newlimits(int lowval, int highval) {
	    this.lowval = lowval;
	    this.highval = highval;
	    clip();
	}
    }

    static final int MAX_HUE = 256 * 6;

    public static Color getColor(int hue) {
	int leg = (hue / 256) % 6;
	int step = (hue % 256) * 2;
	int falling = (step < 256) ? 255 : 511 - step;
	int rising = (step < 256) ? step : 255;
	int r, g, b;
	r = g = b = 0;
	switch (leg) {
	case 0:
	    r = 255;
	    break;
	case 1:
	    r = falling;
	    g = rising;
	    break;
	case 2:
	    g = 255;
	    break;
	case 3:
	    g = falling;
	    b = rising;
	    break;
	case 4:
	    b = 255;
	    break;
	case 5:
	    b = falling;
	    r = rising;
	    break;
	}
	return new Color(r, g, b);
    }

    public void run() {
	Thread me = Thread.currentThread();
	me.setPriority(Thread.MIN_PRIORITY);
	Graphics2D gi = img == null ? null : img.createGraphics();
	Dimension dims = getSize();
	animval x1 = new animval(0, dims.width, 2, 10);
	animval y1 = new animval(0, dims.height, 2, 10);
	animval x2 = new animval(0, dims.width, 2, 10);
	animval y2 = new animval(0, dims.height, 2, 10);
	int hue = (int) (Math.random() * MAX_HUE);
	long start = System.currentTimeMillis();
	int numframes = 0;
	while (animator == me) {
	    dims = getSize();
	    if (img == null ||
		dims.width != img.getWidth() ||
		dims.height != img.getHeight())
	    {
		if (gi != null) {
		    gi.dispose();
		    gi = null;
		}
		img = (BufferedImage) createImage(dims.width, dims.height);
		gi = img.createGraphics();
		x1.newlimits(0, dims.width);
		y1.newlimits(0, dims.height);
		x2.newlimits(0, dims.width);
		y2.newlimits(0, dims.height);
	    }
	    Color c1 = getColor(hue);
	    Color c2 = getColor(hue + 256 * 3);
	    GradientPaint gp = new GradientPaint(x1.getFlt(), y1.getFlt(), c1,
						 x2.getFlt(), y2.getFlt(), c2,
						 true);
	    gi.setPaint(gp);
	    gi.fillRect(0, 0, dims.width, dims.height);
	    gi.setColor(Color.yellow);
	    gi.drawLine(x1.getInt(), y1.getInt(), x2.getInt(), y2.getInt());
	    long now = System.currentTimeMillis();
	    numframes++;
	    if (now - start > 1000) {
		framerate = numframes / ((now - start) / 1000.0);
		numframes = 0;
		start = now;
	    }
	    Graphics g = getGraphics();
	    paint(g);
	    g.dispose();
	    x1.anim(); y1.anim();
	    x2.anim(); y2.anim();
	    hue = (hue + (int) (Math.random() * 10)) % MAX_HUE;
	}
    }

    public void stop() {
	animator = null;
    }

    public void paint(Graphics g) {
	if (img != null) {
	    g.drawImage(img, 0, 0, this);
	    Dimension size = getSize();
	    g.setColor(Color.white);
	    g.drawString((Math.rint(framerate * 10.0) / 10.0) + " fps",
			 5, size.height - 5);
	    getToolkit().sync();
	}
    }

    public static void main(String argv[]) {
	Frame f = new Frame();
	final GradAnim c = new GradAnim();
	f.add(c);
	f.addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {System.exit(0);}
	    public void windowIconified(WindowEvent e) {c.stop();}
	    public void windowDeiconified(WindowEvent e) {c.start();}
	});
	f.pack();
	f.show();
	c.start();
    }
}
