/*
 * @(#)ImageSpeed.java	1.9 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.applet.Applet;
import java.awt.*;
import java.awt.image.*;
import sun.awt.image.*;

public class ImageSpeed extends Applet {

    Animation animation;
    VideoCanvas canvas;
    Label frameRate;
    
    public ImageSpeed() {
	canvas = new VideoCanvas(320, 240);
	setLayout(new BorderLayout());
	add("Center", canvas); 
	frameRate = new Label("0", Label.CENTER);
	add("South", frameRate);
    }

    public void init() {
	resize(400, 300);
	setVisible(true);
	canvas.setDoClear("true".equals(getParameter("doclear")));
	canvas.setDoFill("true".equals(getParameter("dofill")));
    }

    public void start() {
	if (animation == null) {
	    animation = new Animation(canvas);
	    animation.frameRate = frameRate;
	}
	animation.start();
    }
}


class Animation extends Thread {

    VideoCanvas canvas;
    public Label frameRate;
    
    public Animation(VideoCanvas cvs) {
	canvas = cvs;
    }

    public void run() {
	int c = 0;
	int[] color = {0xffff0000, 0xff00ff00, 0xffffffff};
	int f = 0;
	int w = 0, h = 0;
	long t1 = System.currentTimeMillis();

	while (true) {
	    int [] pixels = canvas.getPixels();
	    int width = canvas.getWidth();
	    int height = canvas.getHeight();
	    int rx = (int)((double)width*Math.random());
	    int ry = (int)((double)height*Math.random());
                
	    pixels[(ry*width)+rx] = color[f%3];
    
	    if (++f == 30) {
		long t2 = System.currentTimeMillis();
		frameRate.setText("30 frames elapsed time = " + (t2 - t1) + " millisecs");
		t1 = t2;
		f = 0;
	    }

	    canvas.render();
	}
    }
}


class VideoCanvas extends Canvas {

    protected int width, height;
    protected int [] pixels;
    protected BufferedImage bufImage;

    boolean doClear, doFill;
    protected Graphics2D buf2D;

    public VideoCanvas(int w, int h) {
	width = w;
	height = h;
	bufImage = new BufferedImage(width, height,
                                     BufferedImage.TYPE_INT_ARGB);
        pixels = ((IntegerComponentRaster)bufImage.getRaster()).getDataStorage();
    }

    public void setDoClear(boolean flag) {
	if (flag && buf2D == null) {
	    buf2D = bufImage.createGraphics();
	}
	doClear = flag;
    }

    public void setDoFill(boolean flag) {
	if (flag && buf2D == null) {
	    buf2D = bufImage.createGraphics();
	}
	doFill = flag;
    }

    /**
     * Return the width of the canvas.
     */
    public int getWidth() {
	return width;
    }

    /**
     * Return the height of the canvas.
     */
    public int getHeight() {
	return height;
    }

    public int [] getPixels() {
	return pixels;
    }

    /**
     * Repaint the image when it is damaged.
     */
    public void paint(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
        g2.setComposite(AlphaComposite.Src);
	g2.drawImage(bufImage, null, 0, 0);
    }

    /**
     * Given the VidData, render it on screen.
     */
    public void render() {
	Graphics2D g = (Graphics2D) getGraphics();
        g.setComposite(AlphaComposite.Src);
	g.drawImage(bufImage, null, 0, 0);
	if (doClear) {
	    buf2D.clearRect (0,0,320,240);
	}
	if (doFill) {
	    buf2D.fillRect (0,0,320,240);
	}
    }
}

