/*
 * @(#)ImagingTest.java	1.17 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.applet.Applet;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Canvas;
import java.awt.Choice;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Label;
import java.awt.MediaTracker;
import java.awt.Panel;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Scrollbar;
import java.awt.TextField;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.MemoryImageSource;
import java.awt.image.PixelGrabber;
import java.awt.image.BufferedImage;
import java.awt.image.ByteLookupTable;
import java.awt.image.ColorModel;
import java.awt.image.ConvolveOp;
import java.awt.Graphics2D;
import java.awt.image.Kernel;
import java.awt.image.LookupOp;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Vector;

public class ImagingTest extends Applet implements ActionListener
{
    ImagingCanvas canvas;
    Choice opChoice = new Choice();
    Button button = new Button("Add");
    Frame frame   = new Frame("Image Ops");
    boolean frameLocation = false;
    GridBagLayout frameGB = new GridBagLayout();
    int frameY = 0;
    String[] opNames;
    
    public void init() {
        URL url;
        GridBagLayout gbag = new GridBagLayout();
        int win = 100;
        int lev = 46;
        int y = 0;
        
        setLayout (gbag);
        String pimage = getParameter("image");
        Image image = getImage(getCodeBase(), pimage);
        
        canvas = new ImagingCanvas(this, image.getSource());
        opNames = canvas.getPipeNames();
        
        addComponent(this, gbag, canvas, 0, y++, 0, 0, 0., 0.);
        
        Panel p = new Panel();
        addComponent(this, gbag, p, 0, y++, 0, 0, 0., 0.);

        GridBagLayout gridbag = new GridBagLayout();
        p.setLayout (gridbag);

        // Add the choice label
        addComponent(p, gridbag, new Label("Imaging Ops"), 0, 0, 1, 1, 0., 0.);
        
        // Add the choices
        for (int i=0; i < opNames.length; i++) {
            opChoice.addItem(opNames[i]);
        }
        addComponent(p, gridbag, opChoice, 1, 0, 1, 1, 0., 0.);

        button.addActionListener(this);
        addComponent(p, gridbag, button, 0, 1, 1, 1, 0., 0.);

        Dimension lookSize = canvas.getPreferredSize();
            
        if (getParameter("hide") != null) {
            p.setVisible(false);
            resize(new Dimension(lookSize.width, lookSize.height));
        }

        frame.setLayout(frameGB);
        frame.setVisible(false);
        frame.setResizable(true);
        validate();
    }
    
    void addComponent(Container p, GridBagLayout gbl,
                      Component c, int x, int y, int w, int h,
                      double weightX, double weightY) {
        GridBagConstraints gbc = new GridBagConstraints();

        gbc.fill  = GridBagConstraints.BOTH;
        gbc.gridx = x;
        gbc.gridy = y;
        gbc.gridwidth = w;
        gbc.gridwidth = h;
        gbc.weightx = weightX;
        gbc.weighty = weightY;
        gbl.setConstraints(c, gbc);
        p.add(c);
    }

    public void actionPerformed(ActionEvent e) {
        Object target = e.getSource();

        if (target == button) {
            Container c = canvas.add(opChoice.getSelectedIndex());
            
            if (c != null) {
                Cursor cursor = frame.getCursor();
                frame.setCursor(cursor);
                // Add the GUI
                addComponent(frame, frameGB, c, 0, frameY++, 1, 1,
                             1., 0.);
                frame.pack();
                if (frameLocation == false) {
                    Rectangle r = this.getBounds();
                    frame.setLocation (new Point(r.x+r.width, r.y));
                    frameLocation = true;
                    frame.setVisible(true);
                }
                canvas.repaint();
                frame.setCursor(cursor);
            }
        }
    }

}

class ImagingCanvas extends Canvas {
    BufferedImage inImage;
    BufferedImage outImage;
    int width;
    int height;
    int[] outpixels;
    Pipeline pipeline;
    
    ImagingCanvas (Applet applet, ImageProducer producer) {
        BufferedImageConsumer pg = null;
        try {
            pg = new BufferedImageConsumer(producer, true);
            pg.grabPixels();

            // Check for errors
            if ((pg.status() & ImageObserver.ABORT) != 0) {
                System.err.println ("Error while fetching image");
                System.exit(1);
            }
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }

        // Grab the buffered image
        inImage = pg.getImage();
        width = inImage.getWidth();
        height = inImage.getHeight();

        pipeline = new Pipeline(applet, this, inImage);
    }
    
    ImagingCanvas (Applet applet, BufferedImage bimage) {
        this.inImage = bimage;
        width = inImage.getWidth();
        height = inImage.getHeight();

        pipeline = new Pipeline(applet, this, inImage);
    }
    

    public BufferedImage getImage() {
        return pipeline.getImage();
    }

    public String[] getPipeNames() {
        return Pipeline.getPipeNames();
    }
    
    public Container add(int i) {
        return pipeline.add(i);
    }
    
    public Dimension getPreferredSize() {
	return new Dimension (width, height);
    }

    public Dimension getMinimumSize() {
	return getPreferredSize();
    }

    public void paint(Graphics g) {
        Graphics2D g2 = (Graphics2D) g;
        BufferedImage image = pipeline.getImage();
        g2.drawImage(image, 0, 0, null);
    }
}
