/*
 * @(#)LookupPipe.java	1.14 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.applet.Applet;
import java.awt.event.AdjustmentListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.Container;
import java.awt.GridBagLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.Scrollbar;
import java.awt.TextField;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.LookupOp;
import java.awt.image.ByteLookupTable;
import java.awt.image.LookupTable;
import java.awt.image.ShortLookupTable;
import java.awt.image.Raster;

public class LookupPipe extends Pipe
    implements AdjustmentListener, ActionListener
{
    LookupTable table = null;
    LookupOp    lookupOp = null;
    byte[] byteLUT = new byte[256];
    Scrollbar sbarLev;
    Scrollbar sbarWin;
    TextField window;
    TextField level;
    int win = 150;
    int lev = 100;

    public LookupPipe(Pipeline pipeline, Applet applet, BufferedImage bimage){
        super(pipeline, applet, bimage, true);

        String wstring = applet.getParameter("window");
        String lstring = applet.getParameter("level");
        if (wstring != null) {
            win = Integer.parseInt(wstring);
        }
        if (lstring != null) {
            lev = Integer.parseInt(lstring);
        }

        // Create a linear ramp
        updateLUT (lev, win);

        // Create a new panel
        Panel p = new Panel();
        GridBagLayout gb = new GridBagLayout();

    }

    public Container addGui() {
        GridBagLayout gb = new GridBagLayout();
        Panel p = new Panel();
        p.setLayout(gb);
        int idx=0;

        addComponent(p, gb, new Label("Window"), 0, 0, 1, 1, 0., 0.);

        sbarWin = new Scrollbar (Scrollbar.HORIZONTAL, win, 32, 1, 256);
        sbarWin.addAdjustmentListener(this);
        addComponent(p, gb, sbarWin, 1, 0, 3, 1, 1., 0.);

        window = new TextField(Integer.toString(win));
        window.addActionListener(this);
        addComponent(p, gb, window, 2, 0, 1, 1, 0., 0.);

        addComponent(p, gb, new Label("Level"), 0, 1, 1, 1, 0., 0.);

        sbarLev = new Scrollbar (Scrollbar.HORIZONTAL, lev, 32, 0, 255);
        sbarLev.addAdjustmentListener(this);
        addComponent(p, gb, sbarLev, 1, 1, 3, 1, 1., 0.);

        level = new TextField(Integer.toString(lev));
        level.addActionListener(this);
        addComponent(p, gb, level, 2, 1, 1, 1, 0., 0.);

        return p;
    }

    public void process() {
        long first, last;

        // Need to create a new data buffer
        if (usedRef) {
            createDestRaster();
        }

        first = System.currentTimeMillis();
        lookupOp.filter(inImage, outImage);
        last = System.currentTimeMillis();
        System.out.println("elapsed: "+(last-first+1));
        pipeline.trigger(this);
    }

    final void updateLUT(int level, int window) {
        int start = level - (window/2);
        int end   = start + window;
        double incr = 256./window;
        double sval = 0;
        int idx   = 0;
        int i = 0;

        System.out.println ("w/l "+window+" "+level+
                            " start "+start+ " end "+end);
        // Set the LUT
        if (start > 0) {
            for (idx=0; idx < start; idx++) {
                byteLUT[idx] = 0;
            }
        }
        else {
            sval = -start*incr;
            for (idx=0; idx < (int)sval; idx++) {
                byteLUT[idx] = 0;
            }
            start = 0;
        }

        for (i = start; i < end && idx < 256; i++, idx++) {
            byteLUT[idx] = (byte) Math.round(sval);

            sval += incr;
        }
        for (i = end; i < 256 && idx < 256; i++, idx++) {
            byteLUT[idx] = (byte) 255;
        }

        if (table == null) {
            // Create the lookup table
            table = new ByteLookupTable(0, byteLUT);
            lookupOp = new LookupOp(table, null);
        }
    }

    public String toString() {
        return ("LookupPipe");
    }
    public void adjustmentValueChanged (AdjustmentEvent e) {
        Object target = e.getSource();
        int lvalue = sbarLev.getValue();
        int wvalue = sbarWin.getValue();
        boolean update = false;

        if (target == sbarLev) {
            // Need to set the window/level
            level.setText(Integer.toString(lvalue));
            update = true;
        }
        else if (target == sbarWin) {
            window.setText(Integer.toString(wvalue));
            update = true;
        }

        if (update) {
            updateLUT(lvalue, wvalue);
            process();
        }
    }

    public void actionPerformed(ActionEvent e) {
        Object target = e.getSource();
        int lvalue = atoi(level);
        int wvalue = atoi(window);
        boolean update = false;

        if (target == level) {
            update = true;
        }
        else if (target == window) {
            update = true;
        }
        if (update) {
            sbarLev.setValue(lvalue);
            sbarWin.setValue(wvalue);
            updateLUT(lvalue, wvalue);
            process();
        }
    }

    int atoi(TextField w) {
        String s = w.getText();
        return (Integer.parseInt(w.getText()));
    }
}
