/*
 * @(#)ColorModels.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)AffineAccuracy.java	1.1 97/10/20
 * 
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road, 
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * CopyrightVersion 1.2
 * 
 */
import java.lang.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.applet.*;

abstract class AppletProto extends Applet implements Runnable {
  Thread              theThread;
  Graphics2D          theScreen;
  private boolean     theActivityStatus;

  public void start() {
    if (this.theThread == null) {
      this.theThread = new Thread(this);
      if( null != this.theThread ) {
	this.theThread.setPriority(Thread.MAX_PRIORITY);
	this.theThread.start();
	this.theActivityStatus = true;
      } else {
	debugReport("Thread initialization problem");
      } 
    }
  }

  public void stop() {
    if (this.theThread != null) {
      this.theThread.stop();
      this.theThread = null;
      this.theActivityStatus = false;
    }
  }

  public void run() {
    privateStartup();
    if( true == setupInternalData() ) {
      runApplet();
      cleanupInternalData();
    } else {
      debugReport("Error in initialization!");
    }
    privateCleanup();
  }

  private void privateStartup(){
    this.theScreen = (Graphics2D)getGraphics();
  }
  private void privateCleanup(){
    this.theScreen = null;
  }

  public boolean isActive() {
    return this.theActivityStatus;
  }

  public void debugReport(String reportLine){
    System.out.println(">>> "+reportLine);
  }

  abstract boolean setupInternalData();
  abstract public void cleanupInternalData();
  abstract public void runApplet();
  
}

public class ColorModels extends AppletProto {
  IndexColorModel     cfgColorModel;
  IndexColorModel     imgColorModel;

  public boolean setupInternalData(){
    BufferedImage               theImage = new BufferedImage(10,10,BufferedImage.TYPE_BYTE_INDEXED);
    GraphicsConfiguration       theConfig = this.theScreen.getDeviceConfiguration();
   
    if( (null != theImage) && (null != theConfig) ) {

      imgColorModel = (IndexColorModel)theImage.getColorModel();

      cfgColorModel = (IndexColorModel)theConfig.getColorModel(Transparency.OPAQUE);


      if( null != cfgColorModel ) {
	System.out.println("---------------------------------------------");
	System.out.println("Config Color Model:\n"+cfgColorModel);
	System.out.println("Image ColorModel:\n"+imgColorModel);
	System.out.println("---------------------------------------------");
	System.out.println("\t\tConfig CM pixels\tImage CM pixels  ");
	System.out.println("---------------------------------------------");
	{
	  int         idx;
	  String      imdStr,cfgRGB,imgRGB;
	  for(idx = 0; idx < 256; idx++) {
	    cfgRGB = Integer.toHexString(cfgColorModel.getRGB(idx));
	    imgRGB = Integer.toHexString(imgColorModel.getRGB(idx));
	    imdStr = "\t\t";
	    if(cfgRGB.length() < 4) {
	      imdStr = "\t\t\t";
	    }
	    System.out.println(">>> "+idx+"\t\t"+cfgRGB+imdStr+imgRGB);
	  }
	}
	System.out.println("---------------------------------------------");

      }
    }
  
    

      return true;
  }
  public void cleanupInternalData() {
  }
  public void runApplet() {
  }
}
