/*
 * @(#)FontNamesAppletOnscreen.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)AffineAccuracy.java	1.1 97/10/20
 * 
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road, 
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * CopyrightVersion 1.2
 * 
 */
import java.lang.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.applet.*;

abstract class AppletProto extends Applet implements Runnable {
  Thread              theThread;
  Graphics2D          theScreen;
  private boolean     theActivityStatus;
  private boolean     theInitializedStatus;

  public AppletProto() {
    this.theInitializedStatus = false;    
  }

  public void start() {
    if (this.theThread == null) {
      this.theThread = new Thread(this);
      if( null != this.theThread ) {
	this.theThread.setPriority(Thread.MAX_PRIORITY);
	this.theThread.start();
	this.theActivityStatus = true;
      } else {
	debugReport("Thread initialization problem");
      } 
    }
  }

  public void stop() {
    if (this.theThread != null) {
      this.theThread.stop();
      this.theThread = null;
      this.theActivityStatus = false;
    }
  }

  public void run() {
    privateStartup();
    if( true == setupInternalData() ) {
      runApplet();
      cleanupInternalData();
    } else {
      debugReport("Error in initialization!");
    }
    privateCleanup();
  }

  private void privateStartup(){
    this.theScreen = null;//(Graphics2D)getGraphics();
  }
  private void privateCleanup(){
    this.theScreen = null;
  }

  public boolean isActive() {
    return this.theActivityStatus;
  }

  public void setInitialized() {
    this.theInitializedStatus = true;  
  }
  public void cleanInitialized() {
    this.theInitializedStatus = false;  
  }
  public boolean isInitialized() {
    return this.theInitializedStatus;  
  }


  public void debugReport(String reportLine){
    System.out.println(">>> "+reportLine);
  }

  abstract boolean setupInternalData();
  abstract public void cleanupInternalData();
  abstract public void runApplet();
  
}

public class FontNamesAppletOnscreen extends AppletProto {
  Frame[]             testFrames;
  TestCanvas[]        testCanvases;
  static int[]        factors = { Font.PLAIN, Font.BOLD, Font.BOLD | Font.ITALIC, Font.ITALIC};
  static String[]     names   = { "Onscreen: PLAIN","Onscreen: BOLD","Onscreen: BOLD | ITALIC","Onscreen: ITALIC" };

  public boolean setupInternalData(){
    boolean           retCode      = false;
    int               count        = 0;
    int               successCount = 0;

    if( true == isInitialized() ) {
      return true;
    }

    if( null != (this.testCanvases = new TestCanvas[4]) ) {
      if( null != (this.testFrames = new Frame[4]) ) {
	for( count = 0; count < 4; count ++ ) {
	  if( null != (this.testFrames[count] = new Frame(names[count])) ) {
	    if( null != (this.testCanvases[count] = new TestCanvas()) ) {
	      this.testCanvases[count].init((factors[count]));
	      this.testFrames[count].setSize(300,300);
	      this.testFrames[count].add(this.testCanvases[count]);
	      this.testFrames[count].pack();
	      this.testFrames[count].validate();
	      this.testFrames[count].setVisible(true);
	      successCount ++;
	    }
	  }
	}
	if( 4 == successCount ) {
	  setInitialized();
	  retCode = true;
	}
      }
    }
    return retCode;
  }
  public void cleanupInternalData() {
  }
  public void runApplet() {
  }
}

class TestCanvas extends Canvas {
  Font[]   fontArray;
  static String[] namesArray = {
    "Serif",
    "SansSerif",
    "Dialog",
    "DialogInput",
    "Symbol",
    "Monospaced",
    "Default"};
  public boolean init(int givenFactor) {
    int     count        = 0;
    boolean retCode      = false;
    int     successCount = 0;
    
    setSize(300, 300);
    setBackground(Color.white);
    setVisible(true);
    setEnabled(true);
    
    if( null != (fontArray = new Font[7]) ) {
      for( count = 0; count < 7; count++ ) {
	if( null != (fontArray[count] = new Font(namesArray[count],givenFactor,12)) ) {
	  successCount ++;
	}
      }
      if( 7 == successCount ) {
	retCode = true;
      }
    }
    return retCode;
  }

  public void paint(Graphics g) {
    Font      savedFont = null;
    int       count     = 0;
    int       startPos  = 20;
    if( null != g ) {
      if( null != (savedFont = g.getFont()) ) {
	for( count=0; count < 7; count++ ) {
	  if( null != this.fontArray[count] ) {
	    g.setFont(this.fontArray[count]);
	    g.drawString(this.namesArray[count]+" Font: QWER1234",5,startPos);
	    startPos += 20;
	  }
	}	   
	g.setFont(savedFont);
      }
    }
  }
}

class TestPlainCanvas extends TestCanvas {
  public boolean init() {
    return super.init(Font.PLAIN);
  }
}
class TestBoldCanvas extends TestCanvas {
  public boolean init() {
    return super.init(Font.BOLD);
  }
}
