/*
 * @(#)GifOutput.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)AffineAccuracy.java	1.1 97/10/20
 * 
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road, 
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * CopyrightVersion 1.2
 * 
 */
import java.lang.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.applet.*;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;



abstract class AppletProto extends Applet implements Runnable {
  Thread              theThread;
  Graphics2D          theScreen;
  private boolean     theActivityStatus;
  private boolean     theInitializedStatus;

  public AppletProto() {
    this.theInitializedStatus = false;    
  }

  public void start() {
    if (this.theThread == null) {
      this.theThread = new Thread(this);
      if( null != this.theThread ) {
	this.theThread.setPriority(Thread.MAX_PRIORITY);
	this.theThread.start();
	this.theActivityStatus = true;
      } else {
	debugReport("Thread initialization problem");
      } 
    }
  }

  public void stop() {
    if (this.theThread != null) {
      this.theThread.stop();
      this.theThread = null;
      this.theActivityStatus = false;
    }
  }

  public void run() {
    privateStartup();
    if( true == setupInternalData() ) {
      runApplet();
      cleanupInternalData();
    } else {
      debugReport("Error in initialization!");
    }
    privateCleanup();
  }

  private void privateStartup(){
    this.theScreen = (Graphics2D)getGraphics();
  }
  private void privateCleanup(){
    this.theScreen = null;
  }

  public boolean isActive() {
    return this.theActivityStatus;
  }

  public void setInitialized() {
    this.theInitializedStatus = true;  
  }
  public void cleanInitialized() {
    this.theInitializedStatus = false;  
  }
  public boolean isInitialized() {
    return this.theInitializedStatus;  
  }


  public static void debugReport(String reportLine){
    System.out.println(">>> "+reportLine);
  }

  abstract boolean setupInternalData();
  abstract public void cleanupInternalData();
  abstract public void runApplet();
  
}

public class GifOutput extends AppletProto {
  Image               originalImage;
  BufferedImage       newOffscreen;
  BufferedImage       compatibleOffscreen;
  ImagePanorama       newPanorama;
  ImagePanorama       compatiblePanorama;
  VideoTest           videoTest;
  int                 theWidth;
  int                 theHeight;


 
  public boolean setupInternalData(){

    if( true == isInitialized() ) {
      return true;
    }

    this.originalImage       = null;
    this.newOffscreen        = null;
    this.compatibleOffscreen = null;
    this.newPanorama         = null;
    this.compatiblePanorama  = null;
    this.videoTest           = null;
    this.theWidth            = 0;
    this.theHeight           = 0;
    if( loadImage() ) {
      if( createNewOffscreen() ) {
	if( createCompatibleOffscreen() ) { 
	  if( createNewPanorama() ) {
	    if( createCompatiblePanorama() ) {
	      if( createVideoTest() ) {
		//		printLUTs();
		setInitialized();
		return true; 
	      }
	    }
	  }
	}
      }
    }
    return false;
  }
  public void cleanupInternalData() {
  }
  public void runApplet() {
  }
  public void paint(Graphics givenGraphics) {
    if( null != this.originalImage ) {
      givenGraphics.drawImage(this.originalImage,0,0,this);
      if( null != this.videoTest ) {
	this.videoTest.applyToScreen(givenGraphics,10,10);
      }
    }
  }
  void printLUTs() {
    if( (null != this.theScreen) && (null != this.newOffscreen) && (null != this.compatibleOffscreen) ) {
      GraphicsConfiguration    theConfig = this.theScreen.getDeviceConfiguration();
      if( null != theConfig ) {
	IndexColorModel     scrCM = (IndexColorModel)(theConfig.getColorModel(Transparency.OPAQUE));
	IndexColorModel     newCM = (IndexColorModel)(this.newOffscreen.getColorModel());
	IndexColorModel     cmpCM = (IndexColorModel)(this.compatibleOffscreen.getColorModel());
	int                 count = 0;
	String              scrS;
        String              newS;
	String              cmpS;
	if( (null != scrCM) && (null != newCM) && (null != cmpCM) ) {
	  System.out.println("=====================================================================");
	  System.out.println(" #\tScreen\t\tnew image\tcompatible one"); 
	  System.out.println("=====================================================================");
	  for( count = 0; count < 256; count ++ ) {
	    scrS = Integer.toHexString(scrCM.getRGB(count));
	    newS = Integer.toHexString(newCM.getRGB(count));
	    cmpS = Integer.toHexString(cmpCM.getRGB(count));

	    if( scrS.length() < 4 ) {
	      scrS += "\t\t";
	    } else {
	      scrS += "\t";
	    }

	    if( newS.length() < 4 ) {
	      newS += "\t\t";
	    } else {
	      newS += "\t";
	    }
	    
	    System.out.println(" "+count+"\t"+scrS+newS+cmpS);

	  } 
	  System.out.println("=====================================================================");
	}
      }
    }
  }
 
  boolean loadImage() {
    boolean       retCode = false;
    Toolkit       theToolkit = null;
    MediaTracker  theTracker = null;
    boolean       waitFactor = true;
    
    if( null != (theToolkit = Toolkit.getDefaultToolkit()) ) {
      if( null != (this.originalImage = theToolkit.getImage("Image.gif")) ) {
	if( null != (theTracker = new MediaTracker(this)) ) {
	  debugReport("Loading image");
	  theTracker.addImage(this.originalImage,0xFEBA);
	  
	  while(true == waitFactor) {
	    try{
	      theTracker.waitForID(0xFEBA);
	      waitFactor = false;
	    } catch(InterruptedException e){
	      waitFactor = true;
	    }
	  }

	  debugReport("Image "+this.originalImage+" loaded");
	  this.theWidth = this.originalImage.getWidth(this);
	  this.theHeight = this.originalImage.getHeight(this);
	  setSize(this.theWidth,this.theHeight);
	  retCode = true;
	}	 
      }
    }
    return retCode;
  }
  boolean createNewOffscreen() {
    boolean   retCode = false;
    Graphics  theGraphics = null;
    if( null != (this.newOffscreen = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_BYTE_INDEXED)) ) {
      debugReport("Image "+this.newOffscreen+" created");
      if( null != (theGraphics = this.newOffscreen.getGraphics()) ) {
	theGraphics.drawImage(this.originalImage,0,0,this);
	retCode = true;
      }
    }
    return retCode;
  }
  boolean createCompatibleOffscreen() {
    boolean                  retCode = false;
    GraphicsConfiguration    theConfig = null;
    Graphics                 theGraphics = null;

    if( null != this.theScreen ) {
      if( null != (theConfig = this.theScreen.getDeviceConfiguration()) ) {
	if( null != (this.compatibleOffscreen = theConfig.createCompatibleImage(getWidth(),getHeight())) ) {
	  debugReport("Image "+this.compatibleOffscreen+" created");
	  if( null != (theGraphics = this.compatibleOffscreen.getGraphics()) ) {
	    //	    ((Graphics2D)theGraphics).setRenderingHints(Graphics2D.ANTIALIASING,Graphics2D.ANTIALIAS_OFF);
	    //	    if( Graphics2D.ANTIALIAS_ON == ((Graphics2D)theGraphics).getRenderingHints(Graphics2D.ANTIALIASING) ) {
	    //	      debugReport("ANTIALIASING WAS ON");
	    //	    }
	    theGraphics.drawImage(this.originalImage,0,0,this);
	    retCode = true;
	  }
	}
      }
    }
    return retCode;
  }

  boolean createNewPanorama() {
    boolean retCode = false;
    if( null != (this.newPanorama = new ImagePanorama()) ) {
      if( true == this.newPanorama.init(this.newOffscreen,getWidth(),getHeight(),"NEW IMAGE") ) {
	retCode = true;
      }
    }
    return retCode;
  }
  boolean createCompatiblePanorama() {
    boolean retCode = false;
    if( null != (this.compatiblePanorama = new ImagePanorama()) ) {
      if( true == this.compatiblePanorama.init(this.compatibleOffscreen,getWidth(),getHeight(),"COMPATIBLE IMAGE") ) {
	retCode = true;
      }
    }   
    return retCode;
  }
  boolean createVideoTest() {
    boolean  retCode = false;
    Graphics theGraphics = null;
    if( null != (this.videoTest = new VideoTest()) ) {
      if( true == this.videoTest.init() ) {
	if( null != (theGraphics = this.newOffscreen.getGraphics()) ) {
	  this.videoTest.applyToScreen(theGraphics,10,10);
	  theGraphics = null;
	  if( null != (theGraphics = this.compatibleOffscreen.getGraphics()) ) {
	    this.videoTest.applyToScreen(theGraphics,10,10);
	    retCode = true;
	  }
	}
      }
    }
    return retCode;
  }
  

  public int getWidth() {
    return this.theWidth;
  }
  public int getHeight() {
    return this.theHeight;
  }

  public static void main(String argv[]) {
    Frame appletFrame = new Frame();
    final GifOutput applet = new GifOutput();
    appletFrame.add(applet);
    appletFrame.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {System.exit(0);}
      public void windowIconified(WindowEvent e) {applet.stop();}
      public void windowDeiconified(WindowEvent e) {applet.start();}
    });
    appletFrame.setSize(300,300);
    applet.init();
    appletFrame.show();
    applet.start();
  }
}

class VideoTest {
  Color      clRed;
  Color      clGreen;
  Color      clBlue;

  Color      clYellow;
  Color      clGray;
  Color      clWhite;

  Color      clBlack;
  Color      clSaved;

  boolean init() {
    boolean retCode = false;
    if( null != (clRed = new Color(0xFF,0x00,0x00)) ) {
      if( null != (clGreen = new Color(0x00,0xFF,0x00)) ) {
	if( null != (clBlue = new Color(0x00,0x00,0xFF)) ) {
	  if( null != (clYellow = new Color(0xFF,0xFF,0x00)) ) {
	    if( null != (clGray = new Color(0x77,0x77,0x77)) ) {
	      if( null != (clWhite = new Color(0xFF,0xFF,0xFF)) ) {
		if( null != (clBlack = new Color(0x00,0x00,0x00)) ) {
		  retCode = true;
		}
	      }
	    }
	  }
	}
      }
    }      
    return retCode;
  }

  public void applyToScreen(Graphics givenScreen, int nX,int nY) {
    if( null != givenScreen ) {

      clSaved = givenScreen.getColor();

      givenScreen.setColor(clWhite);
      givenScreen.fillRect(nX,nY,70,70);
      givenScreen.setColor(clYellow);
      givenScreen.fillRect(nX+5,nY+5,60,60);
      
      givenScreen.setColor(clRed);
      givenScreen.fillRect(nX+10,nY+10,50,50);
      givenScreen.setColor(clGreen);
      givenScreen.fillRect(nX+15,nY+15,40,40);
      givenScreen.setColor(clBlue);
      givenScreen.fillRect(nX+20,nY+20,30,30);

      givenScreen.setColor(clSaved);
    }
  }

}

class ImagePanorama extends Canvas {
  Image      myImage;
  Frame      myFrame;
  int        width,heigth;

  public boolean init(Image givenImage,int givenWidth,int givenHeigth,String givenName) {
    boolean   retCode = false;
    this.myImage = givenImage;
    this.width = givenWidth;
    this.heigth = givenHeigth;
    setSize(givenWidth,givenHeigth);
    setBackground(new Color(0,0,0));
    
    if( null != (this.myFrame = new Frame(givenName)) ) {
      this.myFrame.setSize(givenWidth,givenHeigth);
      this.myFrame.add(this);
      this.myFrame.pack();
      this.myFrame.validate();
      this.myFrame.setVisible(true);
      retCode = true;
    }    
    return retCode;
  }
  
  public void paint(Graphics givenGraphics) {
    if( (null != this.myImage) && (null != givenGraphics) ) {
      {
	Color cc = givenGraphics.getColor();
	givenGraphics.setColor(new Color(0,0,0));
	givenGraphics.fillRect(0,0,this.width,this.heigth);
	givenGraphics.setColor(cc);
      }
      //      ((Graphics2D)givenGraphics).setRenderingHints(Graphics2D.ANTIALIASING,Graphics2D.ANTIALIAS_OFF);
      givenGraphics.drawImage(this.myImage,0,0,this);
    }
  }
}

