/*
 * @(#)GlyphOutlineApplet.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)AffineAccuracy.java	1.1 97/10/20
 * 
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road, 
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * CopyrightVersion 1.2
 * 
 */
import java.lang.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.font.*;
import java.applet.*;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.ActionListener;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;


abstract class AppletProto extends Applet implements Runnable {
  Thread              theThread;
  Graphics2D          theScreen;
  private boolean     theActivityStatus;
  private boolean     theInitializedStatus;

  public AppletProto() {
    this.theInitializedStatus = false;
  }

  public void start() {
    if (this.theThread == null) {
      this.theThread = new Thread(this);
      if( null != this.theThread ) {
	this.theThread.setPriority(Thread.MAX_PRIORITY);
	this.theThread.start();
	this.theActivityStatus = true;
      } else {
	debugReport("Thread initialization problem");
      } 
    }
  }

  public void stop() {
    if (this.theThread != null) {
      this.theThread.stop();
      this.theThread = null;
      this.theActivityStatus = false;
    }
  }

  public void run() {
    privateStartup();
    if( true == setupInternalData() ) {
      runApplet();
      cleanupInternalData();
    } else {
      debugReport("Error in initialization!");
    }
    privateCleanup();
  }

  private void privateStartup(){
    //    this.theScreen = (Graphics2D)getGraphics();
  }
  private void privateCleanup(){
    this.theScreen = null;
  }

  public boolean isActive() {
    return this.theActivityStatus;
  }

  public void debugReport(String reportLine){
    System.out.println(">>> "+reportLine);
  }

  public void setInitialized() {
    this.theInitializedStatus = true;  
  }
  public void cleanInitialized() {
    this.theInitializedStatus = false;  
  }
  public boolean isInitialized() {
    return this.theInitializedStatus;  
  }

  abstract boolean setupInternalData();
  abstract public void cleanupInternalData();
  abstract public void runApplet();
}

interface Dispatcher {
  public void dispatchCommand(int commandCode);
  public Object dispatchRequest(int requestCode);
}


public class GlyphOutlineApplet extends AppletProto {
  ImageShapeMagnifier    theMagnifier;
  MagnifierController    theControlSet;
  GraphicsTest           theTest;

  RelationDispatcher     theDispatcher;

  public boolean setupInternalData(){
    boolean        retCode = false;
    BorderLayout   layout  = null;
    BufferedImage  theImage;
    AttachedShape  theOutlineShape;
    AttachedShape  theHiliteShape;

    if( isInitialized() ) {
      return true;
    }

    if( null != (layout = new BorderLayout()) ) {
      if( null != (this.theDispatcher = new RelationDispatcher()) ) {
	if( null != (this.theTest = new GraphicsTest()) ) {
	  if( null != (this.theMagnifier = new ImageShapeMagnifier()) ) {
	    if( null != (this.theControlSet = new MagnifierController()) ) {
	      if( true == this.theDispatcher.init(this,this.theMagnifier,this.theTest) ) {
		if( true == this.theControlSet.init(this.theDispatcher) ) {
		  if( true == this.theTest.init(this.theDispatcher,100,100) ) {
		    if( null != (theImage = this.theTest.getTestImage()) ) {
		      if( null != (theOutlineShape = this.theTest.getTestOutlineShape()) ) {            //   0     Outline Shape to observe
			if( null != (theHiliteShape = this.theTest.getTestHiliteShape()) ) {            //   1     Hilite Shape to observe
		          //                                 0               1              2        3   4   5 6 7  8         //   2     Image to observe
			  if( true == this.theMagnifier.init(theOutlineShape,theHiliteShape,theImage,400,400,4,4,40,4) ) {    //   3 - 4 Physical size
			    setLayout(layout);                                                          //   5     Initial PPU Factor (Physical elts)
			    add("Center",this.theMagnifier);                                            //   6     Min PPUF           ---""----------
			    add("North",this.theControlSet);                                            //   7     Max PPUF           ---""----------
			    validate();                                                                 //   8     Increment of PPU   ---""----------
			    retCode = true;
			    
			    setInitialized();
			  }
			}
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    return retCode;
  }

  public void cleanupInternalData() {
  }
  public void runApplet() {
  }

  public static void main(String argv[]) {
    Frame appletFrame = new Frame();
    final GlyphOutlineApplet applet = new GlyphOutlineApplet();
    appletFrame.add(applet);
    appletFrame.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {System.exit(0);}
      public void windowIconified(WindowEvent e) {applet.stop();}
      public void windowDeiconified(WindowEvent e) {applet.start();}
    });
    appletFrame.setSize(300,300);
    applet.init();
    appletFrame.show();
    applet.start();
  }

}

class RelationDispatcher implements Dispatcher {

  public static final int COMMAND_SCROLL_LEFT            = 0x0001;
  public static final int COMMAND_SCROLL_RIGHT           = 0x0002;
  public static final int COMMAND_SCROLL_UP              = 0x0003;
  public static final int COMMAND_SCROLL_DOWN            = 0x0004;
  public static final int COMMAND_ZOOM_IN                = 0x0005;
  public static final int COMMAND_ZOOM_OUT               = 0x0006;
  public static final int COMMAND_EXIT_TEST              = 0x0007;
  public static final int COMMAND_REQUEST_TEST_HINTS     = 0x0008;
  public static final int COMMAND_SET_TEST_HINTS         = 0x0009;
  public static final int COMMAND_SKIP_TEST_HINTS        = 0x000A;
  public static final int COMMAND_REQUEST_TEST_TRANSFORM = 0x000B;
  public static final int COMMAND_SET_TEST_TRANSFORM     = 0x000C;
  public static final int COMMAND_SKIP_TEST_TRANSFORM    = 0x000D;

  public static final int COMMAND_REQUEST_TEST_FONT_AND_TEXT = 0x000E;
  public static final int COMMAND_SET_TEST_FONT_AND_TEXT     = 0x000F;
  public static final int COMMAND_SKIP_TEST_FONT_AND_TEXT    = 0x0010;

  public static final int COMMAND_REFRESH    = 0x0011;



  ImageMagnifier         theMagnifier;
  GraphicsTest           theTest;
  Applet                 theApplet;

  HintsDialog            theHintsDialog;
  FontAndTextDialog      theFontAndTextDialog;

  public boolean init(Applet givenApplet,ImageMagnifier givenMagnifier,GraphicsTest givenTest) {
    boolean retCode = false;
    if( null != givenMagnifier ) {
      if( null != givenTest ) {
	this.theMagnifier   = givenMagnifier;
	this.theTest        = givenTest;
	this.theApplet      = givenApplet;
	this.theHintsDialog = null;
	this.theFontAndTextDialog = null;

	retCode = true;
      }
    }
    return retCode;
  }
  
  public void dispatchCommand(int commandCode) {
    switch( commandCode ) {
    case COMMAND_SCROLL_LEFT:
      this.theMagnifier.scrollLeft();
      break;
    case COMMAND_SCROLL_RIGHT:
      this.theMagnifier.scrollRight();
      break;
    case COMMAND_SCROLL_UP:
      this.theMagnifier.scrollUp();
      break;
    case COMMAND_SCROLL_DOWN:
      this.theMagnifier.scrollDown();
      break;
    case COMMAND_ZOOM_IN:
      this.theMagnifier.zoomIn();
      break;
    case COMMAND_ZOOM_OUT:
      this.theMagnifier.zoomOut();
      break;
    case COMMAND_REQUEST_TEST_HINTS:
      _requestTestHints();
      break;
    case COMMAND_EXIT_TEST:
      _exit_all();
      break;
    case COMMAND_SET_TEST_HINTS:
      _setTestHints();
      break;
    case COMMAND_SKIP_TEST_HINTS:
      _skipTestHints();
      break;
    case COMMAND_REFRESH:
      _refresh();
      break;
    case COMMAND_REQUEST_TEST_FONT_AND_TEXT:
      _requestFontAndText();
      break;
    case COMMAND_SET_TEST_FONT_AND_TEXT:
      _setFontAndText();
      break;
    case COMMAND_SKIP_TEST_FONT_AND_TEXT:
      _skipFontAndText();
      break;
    default:
      break;
    }

  }
  public Object dispatchRequest(int requestCode) {
    return null;
  }

  private void _exit_all() {
    if( null != this.theApplet ) {
      this.theApplet.destroy();
    }
  }

  private void _requestTestHints() {
    if( null != this.theTest ) {    
      if( null == this.theHintsDialog ) {
	if( null != (this.theHintsDialog = new HintsDialog((Frame)this.theTest)) ) {
	  Object theAntialias          = this.theTest.getAntialias();
	  Object theRendering          = this.theTest.getRendering();
	  Object theTextAntialias      = this.theTest.getTextAntialias();
	  Object theFractionalMetrics  = this.theTest.getFractionalMetrics();
	  boolean theStroke            = this.theTest.getStroke();
	  
	  this.theHintsDialog.init(this,theAntialias,theRendering,theTextAntialias,theFractionalMetrics,theStroke);
	}
      }
    }
  }  
  private void _setTestHints() {
    if( null != this.theTest ) {    
      if( null != this.theHintsDialog ) {
	Object theAntialias          = this.theHintsDialog.getAntialias();
	Object theRendering          = this.theHintsDialog.getRendering();
	Object theTextAntialias      = this.theHintsDialog.getTextAntialias();
	Object theFractionalMetrics  = this.theHintsDialog.getFractionalMetrics();
	boolean theStroke            = this.theHintsDialog.getStroke();

	this.theTest.setAntialias(theAntialias);
	this.theTest.setRendering(theRendering);
	this.theTest.setTextAntialias(theTextAntialias);
	this.theTest.setFractionalMetrics(theFractionalMetrics);
	this.theTest.setStroke(theStroke);
	
	this.theTest.forceUpdate();
	this.theMagnifier.forceUpdate();
	
      }
    }
  }
  private void _skipTestHints() {
    if( null != this.theTest ) {    
      if( null != this.theHintsDialog ) {
	
	this.theHintsDialog.dispose();
	this.theHintsDialog = null;

      }
    }
  }
  private void _requestFontAndText(){
    if( null != this.theTest ) {
      if( null == this.theFontAndTextDialog ) {
	if( null != (this.theFontAndTextDialog = new FontAndTextDialog(this.theTest)) ) {
	  boolean drawString        = this.theTest.getDrawString();	 
	  boolean drawGlyphVector   = this.theTest.getDrawGlyphVector();	 
	  boolean drawOutline       = this.theTest.getDrawOutline();
	  boolean drawBounds        = this.theTest.getDrawBounds();
	  boolean drawLines         = this.theTest.getDrawLines();
	  Font    aFont             = this.theTest.getFont();
	  String  aText             = this.theTest.getText();
	  int     hiliteIndex       = this.theTest.getHiliteIndex();

	  this.theFontAndTextDialog.init(this,aFont,aText,drawString,drawGlyphVector,drawOutline,drawBounds,drawLines,hiliteIndex);

	}
      }
    }
  }
  private void _setFontAndText(){
    if( null != this.theTest ) {
      if( null != this.theFontAndTextDialog ) {
	boolean drawString        = this.theFontAndTextDialog.getDrawString();	 
	boolean drawGlyphVector   = this.theFontAndTextDialog.getDrawGlyphVector();	 
	boolean drawOutline       = this.theFontAndTextDialog.getDrawOutline();
	boolean drawBounds        = this.theFontAndTextDialog.getDrawBounds();
	boolean drawLines         = this.theFontAndTextDialog.getDrawLines();
	Font    aFont             = this.theFontAndTextDialog.getTestFont();
	String  aText             = this.theFontAndTextDialog.getText();
	int     hiliteIndex       = this.theFontAndTextDialog.getHiliteIndex();
	
	this.theTest.setDrawString(drawString);	 
	this.theTest.setDrawGlyphVector(drawGlyphVector);	 
	this.theTest.setDrawOutline(drawOutline);
	this.theTest.setDrawBounds(drawBounds);
	this.theTest.setDrawLines(drawLines);
	this.theTest.setFont(aFont);
	this.theTest.setText(aText);
	this.theTest.setHiliteIndex(hiliteIndex);

	this.theTest.forceUpdate();
	this.theMagnifier.forceUpdate();
      }      
    }
  }
  private void _skipFontAndText(){
    if( null != this.theTest ) {    
      if( null != this.theFontAndTextDialog ) {	
	this.theFontAndTextDialog.dispose();
	this.theFontAndTextDialog = null;
      }
    }
  }

  private void _refresh() {
    this.theTest.forceUpdate();
    this.theMagnifier.forceUpdate();
  }
}

//====================================================================================================
//  DIALOGS
//====================================================================================================
class FontAndTextDialog extends Dialog {
  boolean             drawString;
  boolean             drawGlyphVector;
  boolean             drawOutline;
  boolean             drawBounds;
  boolean             drawLines;
  String              theFontName;
  int                 theFontStyle;
  int                 theFontSize;
  String              theText;
  int                 theHiliteIndex;

  Panel               theDRW0Panel;
  Panel               theDRW1Panel;
  Panel               theFNTPanel;
  Panel               theBTNPanel;
  Checkbox            theDrawSTR,theDrawGV,theDrawOTL,theDrawBND,theDrawLN;
  TextField           theTextField;

  TextField           theFSize;
  TextField           theHIndex;

  Choice              theFStyle;
  Choice              theFName;
  Button              theBTNOk,theBTNCnl;

  GridLayout          theDlgLayout;
  GridLayout          thePnlLayout;

  Dispatcher          theDispatcher;


 

  FontAndTextDialog(Frame givenParent) {
    super(givenParent,false);
    setFont(new Font("SansSerif",Font.PLAIN,12));
  }
  public boolean init(Dispatcher givenDispatcher,Font givenFont,String givenText,boolean givenString,boolean givenGlyphVector,boolean givenDrawOutline,boolean givenDrawBounds,boolean givenDrawLines,int givenHiliteIDX) {
    boolean   retCode = false;
    if( null != givenDispatcher ) {
      if( null != givenFont ) {
	this.theDispatcher = givenDispatcher;
	this.theFontName   = givenFont.getName();
	this.theFontStyle  = givenFont.getStyle();
	this.theFontSize   = givenFont.getSize();
	this.theText       = givenText;
	this.drawString    = givenString;
	this.drawGlyphVector = givenGlyphVector;
	this.drawOutline     = givenDrawOutline;
	this.drawBounds      = givenDrawBounds;
	this.drawLines       = givenDrawLines;
	this.theHiliteIndex  = givenHiliteIDX;
	
	if( null != (this.thePnlLayout = new GridLayout(0,4,0,4)) ) {
	  if( true == _makeFNTPanel() ) {
	    if( true == _makeBTNPanel() ) {
	      if( true == _makeDRW0Panel() ) {
		if( true == _makeDRW1Panel() ) {
		  if( null != (this.theDlgLayout = new GridLayout(4,0,4,0)) ) {
		    setLayout(this.theDlgLayout);
		    add(this.theFNTPanel);
		    add(this.theDRW0Panel);
		    add(this.theDRW1Panel);
		    add(this.theBTNPanel);
		    setTitle("Font and String");
		    pack();
		    show();
		    retCode = true;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    return retCode;
  }
  boolean getDrawString() {
    return this.drawString;
  }
  boolean getDrawGlyphVector() {
    return this.drawGlyphVector;
  }
  boolean getDrawOutline() {
    return this.drawOutline;
  }
  boolean getDrawBounds() {
    return this.drawBounds;
  }
  boolean getDrawLines() {
    return this.drawLines;
  }
  Font getTestFont() {
    return new Font(this.theFontName,this.theFontStyle,this.theFontSize);
  }
  String getText() {
    return this.theText;
  }
  int getHiliteIndex() {
    return this.theHiliteIndex;
  }


  boolean _makeFNTPanel() {
    boolean retCode = false;
    Label   label   = null;
    
    if( null != (this.theFNTPanel = new Panel()) ) {
      if( null != (label = new Label("Font Attributes")) ) {
	if( null != (this.theFName = new Choice()) ) {
	  if( null != (this.theFSize = new TextField(String.valueOf(this.theFontSize))) ) {
	    if( null != (this.theFStyle = new Choice()) ) {
	      this.theFStyle.addItem("Plain");
	      this.theFStyle.addItem("Bold");
	      this.theFStyle.addItem("Italic");
	      this.theFStyle.addItem("Bold-Italic");
	      
	      for( int i = 0; i<getToolkit().getFontList().length; i++ ) {
		this.theFName.addItem(getToolkit().getFontList()[i]);
	      }
	      
	      this.theFNTPanel.setLayout(this.thePnlLayout);
	      this.theFNTPanel.add(label);
	      this.theFNTPanel.add(this.theFName);
	      this.theFNTPanel.add(this.theFStyle);
	      this.theFNTPanel.add(this.theFSize);
	      
	      retCode = true;
	      
	    }
	  }
	}
      }	
    }
    return retCode;
  }
  boolean _makeDRW0Panel() {
    boolean retCode = false;
    Label   label   = null;
    if( null != (this.theDRW0Panel = new Panel() ) ) {
      if( null != (label = new Label("Draw Variant")) ) {
	if( null != (this.theDrawSTR = new Checkbox("Draw String",this.drawString)) ) {
	  if( null != (this.theDrawGV = new Checkbox("Draw Glyph Veftor",this.drawGlyphVector)) ) {
	    if( null != (this.theDrawOTL = new Checkbox("Draw Outline",this.drawOutline)) ){
	      this.theDRW0Panel.setLayout(this.thePnlLayout);
	      this.theDRW0Panel.add(label);
	      this.theDRW0Panel.add(this.theDrawSTR);
	      this.theDRW0Panel.add(this.theDrawGV);
	      this.theDRW0Panel.add(this.theDrawOTL);
	      retCode = true;
	    }
	  }
	}
      }
    }
    return retCode;
  }
  boolean _makeDRW1Panel() {
    boolean retCode = false;
    if( null != (this.theDRW1Panel = new Panel()) ) {
      if( null != (this.theDrawBND = new Checkbox("Draw Boundary",this.drawBounds)) ) {
	if( null != (this.theDrawLN = new Checkbox("Draw Lines",this.drawLines)) ) {
	  if( null != (this.theHIndex = new TextField(String.valueOf(this.theHiliteIndex))) ) {
	    this.theDRW1Panel.setLayout(this.thePnlLayout);
	    this.theDRW1Panel.add(new Label("Hilite"));
	    this.theDRW1Panel.add(this.theHIndex);
	    this.theDRW1Panel.add(this.theDrawBND);
	    this.theDRW1Panel.add(this.theDrawLN);
	    retCode = true;
	  }
	}
      }      
    }
    return retCode;
  }
  boolean _makeBTNPanel() {
    boolean retCode = false;
    if( null != (this.theBTNPanel = new Panel()) ) {
      if( null != (this.theTextField = new TextField(this.theText)) ) {
	if( null != (this.theBTNOk = new Button("Accept")) ) {
	  if( null != (this.theBTNCnl = new Button("Cancel")) ) {
	    this.theBTNPanel.setLayout(new GridLayout(0,3,0,5));
	    this.theBTNPanel.add(this.theTextField);
	    this.theBTNPanel.add(this.theBTNOk);
	    this.theBTNPanel.add(this.theBTNCnl);
	    retCode = true;
	  }
	}
      }
    }
    return retCode;
  }
  public boolean action(Event givenEvent, Object givenObject) {
    boolean         itemStatus;
    if( (null != givenEvent) && (null != givenObject) && (null != this.theDispatcher) ) {
 
      if( ! (givenEvent.target instanceof Checkbox) ) {
	if( givenEvent.target == this.theBTNCnl ) {
	  this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SKIP_TEST_FONT_AND_TEXT);
	  return true;
	}
	if( givenEvent.target == this.theBTNOk ) {
	  String textLine = theTextField.getText();
	  String sizeLine = this.theFSize.getText();
	  String indexLine = this.theHIndex.getText();
	  int    size     = 10;
	  int    index    = 0;
	  
	  if( null != textLine ) {
	    this.theText = textLine;
	  }
	  if( null != sizeLine ) {
	    size = Integer.parseInt(sizeLine,10);
	  }
	  if( 0 != size ) {
	    this.theFontSize = size;
	  }

	  if( null != indexLine ) {
	    index = Integer.parseInt(indexLine,10);
	  }
	  if( index != this.theHiliteIndex ) {
	    this.theHiliteIndex = index;
	  }

	  if( this.theText.length() - 1 < this.theHiliteIndex  ) {
	    this.theHiliteIndex = this.theText.length() - 1;
	    if( null != this.theHIndex ) {
	      this.theHIndex.setText(String.valueOf(this.theHiliteIndex));
	    }
	  }

	  
	  this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SET_TEST_FONT_AND_TEXT);
	  return true;
	}
	if( givenEvent.target == this.theFName ) {
	  this.theFontName = new String((String)givenObject);
	  return true;
	}
	if( givenEvent.target == this.theFStyle ) {
	  if( "Plain".equals(givenObject) ) {
	    this.theFontStyle = Font.PLAIN;
	  }
	  if( "Bold".equals(givenObject) ) {
	    this.theFontStyle = Font.BOLD;
	  }
	  if( "Italic".equals(givenObject) ) {
	    this.theFontStyle = Font.ITALIC;
	  }
	  if( "Bold-Italic".equals(givenObject) ) {
	    this.theFontStyle = Font.ITALIC | Font.BOLD;
	  }
	  return true;
	}
      }

      itemStatus = ((Boolean)givenObject).booleanValue();
      
      
      if( givenEvent.target == this.theDrawSTR ) {
	if( true == itemStatus ) {
	  this.drawString = true;
	} else { 
	  this.drawString = false;
	}
	return true;
      }
      if( givenEvent.target == this.theDrawGV ) {
	if( true == itemStatus ) {
	  this.drawGlyphVector = true;
	} else { 
	  this.drawGlyphVector = false;
	}
	return true;
      }
      if( givenEvent.target == this.theDrawOTL ) {
	if( true == itemStatus ) {
	  this.drawOutline = true;
	} else { 
	  this.drawOutline = false;
	}
	return true;
      }
      if( givenEvent.target == this.theDrawBND ) {
	if( true == itemStatus ) {
	  this.drawBounds = true;
	} else { 
	  this.drawBounds = false;
	}
	return true;
      }
      if( givenEvent.target == this.theDrawLN ) {
	if( true == itemStatus ) {
	  this.drawLines = true;
	} else { 
	  this.drawLines = false;
	}
	return true;
      }

    }
    return false;
  }
}

class HintsDialog extends Dialog {
  Dispatcher             theDispatcher;
  Object                 theAntialias;
  Object                 theRendering;
  Object                 theTextAntialias;
  Object                 theFractionMetrics;
  boolean                theStroke;

  GridLayout             theDlgLayout;
  GridLayout             thePnlLayout;

  Panel                  theAAPanel;
  Panel                  theRDPanel;
  Panel                  theTAPanel;
  Panel                  theFMPanel;
  Panel                  theBTNPanel;

  Checkbox               theAAOn,theAAOff,theAADef;
  Checkbox               theRDSpd,theRDQty,theRDDef;
  Checkbox               theTAOn,theTAOff,theTADef;
  Checkbox               theFMOn,theFMOff,theFMDef;

  Checkbox               theStrk;
  Button                 theBTNOk,theBTNCnl;

  HintsDialog(Frame givenParent) {
    super(givenParent,false);
    setFont(new Font("SansSerif",Font.PLAIN,12));
  }

  public boolean init(Dispatcher givenDispatcher,Object givenAntialias,Object givenRendering,Object givenTextAntialias,Object givenFractionMetrics,boolean givenStroke) {
    boolean              retCode = false;
    if( null != givenDispatcher ) {
      this.theStroke     = givenStroke;
      this.theDispatcher = givenDispatcher;
      this.theAntialias  = givenAntialias;
      this.theRendering  = givenRendering;
      this.theTextAntialias     = givenTextAntialias;
      this.theFractionMetrics   = givenFractionMetrics; 
      if( null != (this.thePnlLayout = new GridLayout(0,4,0,5)) ) {
	if( true == _makeAAPanel(this.theAntialias) ) {
	  if( true == _makeRDPanel(this.theRendering) ) {
	    if( true == _makeTAPanel(this.theTextAntialias) ) {
	      if( true == _makeFMPanel(this.theFractionMetrics) ) {
		if( true == _makeBTNPanel() ) {
		  if( null != (this.theDlgLayout = new GridLayout(5,0,5,0)) ) {
		    setLayout(this.theDlgLayout);
		    add(this.theAAPanel);
		    add(this.theRDPanel);
		    add(this.theTAPanel);
		    add(this.theFMPanel);
		    add(this.theBTNPanel);
		    setTitle("Test Rendering Hints");
		    pack();
		    show();
		    retCode = true;
		  }
		}
	      }
	    }	    
	  }
	}
      }
    }
    return retCode;
  }

  Object getAntialias() {
    return this.theAntialias;
  }
  Object getRendering() {
    return this.theRendering;
  }
  Object getTextAntialias() {
    return this.theTextAntialias;
  }
  Object getFractionalMetrics() {
    return this.theFractionMetrics;
  }
  boolean getStroke() {
    return this.theStroke;
  }

  private boolean _makeAAPanel(Object givenStatus) {
    boolean        retCode = false;
    CheckboxGroup  boxGroup;

    if( null != this.thePnlLayout ) {
      if( null != (this.theAAPanel = new Panel()) ) {
	if( null != (boxGroup = new CheckboxGroup()) ) {
	  if( null != (this.theAAOn = new Checkbox("On",boxGroup,givenStatus == RenderingHints.VALUE_ANTIALIAS_ON)) ) {
	    if( null != (this.theAAOff = new Checkbox("Off",boxGroup,givenStatus == RenderingHints.VALUE_ANTIALIAS_OFF)) ) {
	      if( null != (this.theAADef = new Checkbox("Default",boxGroup,givenStatus == RenderingHints.VALUE_ANTIALIAS_DEFAULT)) ) {
		this.theAAPanel.setLayout(this.thePnlLayout);
		this.theAAPanel.add(new Label("Antialiasing:"));
		this.theAAPanel.add(this.theAAOn);
		this.theAAPanel.add(this.theAAOff);
		this.theAAPanel.add(this.theAADef);
		retCode = true;
	      }
	    }
	  }
	}
      }
    } 
    return retCode;
  }  
  private boolean _makeRDPanel(Object givenStatus) {
    boolean        retCode = false;
    CheckboxGroup  boxGroup;

    if( null != this.thePnlLayout ) {
      if( null != (this.theRDPanel = new Panel()) ) {
	if( null != (boxGroup = new CheckboxGroup()) ) {
	  if( null != (this.theRDSpd = new Checkbox("Speed",boxGroup,givenStatus == RenderingHints.VALUE_RENDER_SPEED)) ) {
	    if( null != (this.theRDQty = new Checkbox("Quality",boxGroup,givenStatus == RenderingHints.VALUE_RENDER_QUALITY)) ) {
	      if( null != (this.theRDDef = new Checkbox("Default",boxGroup,givenStatus == RenderingHints.VALUE_RENDER_DEFAULT)) ) {
		this.theRDPanel.setLayout(this.thePnlLayout);
		this.theRDPanel.add(new Label("Rendering:"));
		this.theRDPanel.add(this.theRDSpd);
		this.theRDPanel.add(this.theRDQty);
		this.theRDPanel.add(this.theRDDef);
		retCode = true;
	      }
	    }
	  }
	}
      }
    } 
    return retCode;
  }
  private boolean _makeTAPanel(Object givenStatus) {
    boolean        retCode = false;
    CheckboxGroup  boxGroup;

    if( null != this.thePnlLayout ) {
      if( null != (this.theTAPanel = new Panel()) ) {
	if( null != (boxGroup = new CheckboxGroup()) ) {
	  if( null != (this.theTAOn = new Checkbox("On",boxGroup,givenStatus == RenderingHints.VALUE_TEXT_ANTIALIAS_ON)) ) {
	    if( null != (this.theTAOff = new Checkbox("Off",boxGroup,givenStatus == RenderingHints.VALUE_TEXT_ANTIALIAS_OFF)) ) {
	      if( null != (this.theTADef = new Checkbox("Default",boxGroup,givenStatus == RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT)) ) {
		this.theTAPanel.setLayout(this.thePnlLayout);
		this.theTAPanel.add(new Label("Text Antialiasing:"));
		this.theTAPanel.add(this.theTAOn);
		this.theTAPanel.add(this.theTAOff);
		this.theTAPanel.add(this.theTADef);
		retCode = true;
	      }
	    }
	  }
	}
      }
    } 
    return retCode;
  }
  private boolean _makeFMPanel(Object givenStatus) {
    boolean        retCode = false;
    CheckboxGroup  boxGroup;

    if( null != this.thePnlLayout ) {
      if( null != (this.theFMPanel = new Panel()) ) {
	if( null != (boxGroup = new CheckboxGroup()) ) {
	  if( null != (this.theFMOn = new Checkbox("On",boxGroup,givenStatus == RenderingHints.VALUE_FRACTIONALMETRICS_ON)) ) {
	    if( null != (this.theFMOff = new Checkbox("Off",boxGroup,givenStatus == RenderingHints.VALUE_FRACTIONALMETRICS_OFF)) ) {
	      if( null != (this.theFMDef = new Checkbox("Default",boxGroup,givenStatus == RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT)) ) {
		this.theFMPanel.setLayout(this.thePnlLayout);
		this.theFMPanel.add(new Label("Fractional Metrics:"));
		this.theFMPanel.add(this.theFMOn);
		this.theFMPanel.add(this.theFMOff);
		this.theFMPanel.add(this.theFMDef);
		retCode = true;
	      }
	    }
	  }
	}
      }
    } 
    return retCode;
  }
  private boolean _makeBTNPanel() {
    boolean        retCode = false;
    CheckboxGroup  boxGroup;

    if( null != this.thePnlLayout ) {
      if( null != (this.theBTNPanel = new Panel()) ) {
	if( null != (this.theBTNOk = new Button("Accept")) ) {
	  if( null != (this.theBTNCnl = new Button("Cancel")) ) {
	    if( null != (this.theStrk = new Checkbox("Stroke On/Off",this.theStroke)) ) {
	      this.theBTNPanel.setLayout(new GridLayout(0,3));
	      this.theBTNPanel.add(this.theStrk);
	      this.theBTNPanel.add(this.theBTNOk);
	      this.theBTNPanel.add(this.theBTNCnl);
	      retCode = true;	    	    
	    }
	  }
	}
      }
    } 
    return retCode;
  }  
  // Behaviour
  public boolean action(Event givenEvent, Object givenObject) {
    boolean         itemStatus;
    if( (null != givenEvent) && (null != givenObject) && (null != this.theDispatcher) ) {
 
      if( ! (givenEvent.target instanceof Checkbox) ) {
	if( givenEvent.target == this.theBTNCnl ) {
	  this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SKIP_TEST_HINTS);
	  return true;
	}
	if( givenEvent.target == this.theBTNOk ) {
	  this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SET_TEST_HINTS);
	  return true;
	}
      }

      itemStatus = ((Boolean)givenObject).booleanValue();
      
      
      if( givenEvent.target == this.theAAOn ) {
	if( true == itemStatus ) {
	  this.theAntialias = RenderingHints.VALUE_ANTIALIAS_ON;
	}
	return true;
      }
      if( givenEvent.target == this.theAAOff ) {
	if( true == itemStatus ) {
	  this.theAntialias = RenderingHints.VALUE_ANTIALIAS_OFF;
	}
	return true;
      }
      if( givenEvent.target == this.theAADef ) {
	if( true == itemStatus ) {
	  this.theAntialias = RenderingHints.VALUE_ANTIALIAS_DEFAULT;
	}
	return true;
      }
      
      if( givenEvent.target == this.theRDSpd ) {
	if( true == itemStatus ) {
	  this.theRendering = RenderingHints.VALUE_RENDER_SPEED;
	}
	return true;
      }
      if( givenEvent.target == this.theRDQty ) {
	if( true == itemStatus ) {
	  this.theRendering = RenderingHints.VALUE_RENDER_QUALITY;
	}
	return true;
      }
      if( givenEvent.target == this.theRDDef ) {
	if( true == itemStatus ) {
	  this.theRendering = RenderingHints.VALUE_RENDER_DEFAULT;
	}
	return true;
      }
      
      if( givenEvent.target == this.theTAOn ) {
	if( true == itemStatus ) {
	  this.theTextAntialias = RenderingHints.VALUE_TEXT_ANTIALIAS_ON;
	}
	return true;
      }
      if( givenEvent.target == this.theTAOff ) {
	if( true == itemStatus ) {
	  this.theTextAntialias = RenderingHints.VALUE_TEXT_ANTIALIAS_OFF;
	}
	return true;
      }
      if( givenEvent.target == this.theTADef ) {
	if( true == itemStatus ) {
	  this.theTextAntialias = RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT;
	}
	return true;
      }
      
      if( givenEvent.target == this.theFMOn ) {
	if( true == itemStatus ) {
	  this.theFractionMetrics = RenderingHints.VALUE_FRACTIONALMETRICS_ON;
	}
	return true;
      }
      if( givenEvent.target == this.theFMOff ) {
	if( true == itemStatus ) {
	  this.theFractionMetrics = RenderingHints.VALUE_FRACTIONALMETRICS_OFF;
	}
	return true;
      }
      if( givenEvent.target == this.theFMDef ) {
	if( true == itemStatus ) {
	  this.theFractionMetrics = RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT;
	}
	return true;
      }
      if( givenEvent.target == this.theStrk ) {
	if( true == itemStatus ) {
	  this.theStroke = true;
	} else {
	  this.theStroke = false;
	}
      }
    }
    return false;
  }
}

//====================================================================================================
//  GRAPHICS TEST AREA
//====================================================================================================
class GraphicsTest extends Frame {
  FontTestArea        theTestArea;
  Dispatcher          theDispatcher;

  public boolean init(Dispatcher givenDispatcher,int givenSizeX, int givenSizeY) {     
    boolean         retCode = false;
    BorderLayout    layout  = null;

    if( true == _makeMenu() ) {
      if( null != (layout = new BorderLayout()) ) {
	if( null != (this.theTestArea = new FontTestArea()) ) {
	  //	  if( true == this.theTestArea.init(new Font("Serif",Font.PLAIN,12),new String("QWERTYqwerty"),givenSizeX,givenSizeY,BufferedImage.TYPE_INT_ARGB) ) {
	  if( true == this.theTestArea.init(new Font("Serif",Font.PLAIN,12),new String("\u3041\u3042\u3043\u3044\u3045"),givenSizeX,givenSizeY,BufferedImage.TYPE_INT_ARGB) ) {

	    this.theDispatcher = givenDispatcher;
	    
	    setLayout(layout);
	    add("Center",this.theTestArea);
	    
	    pack();
	    validate();
	    setTitle("TEST AREA");
	    setResizable(false);
	    show();
	    retCode = true;
	    
	  }	  
	}
      }
    }
    return retCode;
  }

  // Get - set methods
  public void setAntialias(Object givenAntialias) {
    this.theTestArea.setAntialias(givenAntialias);
  }
  public Object getAntialias() {
    return this.theTestArea.getAntialias();
  }
  public void setTextAntialias(Object givenTextAntialias) {
    this.theTestArea.setTextAntialias(givenTextAntialias);
  }
  public Object getTextAntialias() {
    return this.theTestArea.getTextAntialias();
  }
  public void setRendering(Object givenRendering) {
    this.theTestArea.setRendering(givenRendering);
  }
  public Object getRendering() {
    return this.theTestArea.getRendering();
  }
  public void setFractionalMetrics(Object givenFractionalMetrics) {
    this.theTestArea.setFractionalMetrics(givenFractionalMetrics);
  }
  public Object getFractionalMetrics() {
    return this.theTestArea.getFractionalMetrics();
  }
  public void setStroke(boolean givenStroke) {
    this.theTestArea.setStroke(givenStroke);
  }
  public boolean getStroke() {
    return this.theTestArea.getStroke();
  }

  public boolean getDrawString() {
    return this.theTestArea.getDrawString();
  }
  public void setDrawString(boolean givenDrawString) {
    this.theTestArea.setDrawString(givenDrawString);
  }
  public boolean getDrawGlyphVector() {
    return this.theTestArea.getDrawGlyphVector();
  }
  public void setDrawGlyphVector(boolean givenGlyphVector) {
    this.theTestArea.setDrawGlyphVector(givenGlyphVector);
  }
  public boolean getDrawOutline() {
    return this.theTestArea.getDrawOutline();
  }
  public void setDrawOutline(boolean givenDrawOutline) {
    this.theTestArea.setDrawOutline(givenDrawOutline);
  }
  public boolean getDrawBounds() {
    return this.theTestArea.getDrawBounds();
  }
  public void setDrawBounds(boolean givenDrawBounds) {
    this.theTestArea.setDrawBounds(givenDrawBounds);
  }
  public boolean getDrawLines() {
    return this.theTestArea.getDrawLines();
  }
  public void setDrawLines(boolean givenDrawLines) {
    this.theTestArea.setDrawLines(givenDrawLines);
  }
  public Font getFont() {
    return this.theTestArea.getFont();
  }
  public void setFont(Font givenFont) {
    this.theTestArea.setFont(givenFont);
  }
  public String getText() {
    return this.theTestArea.getText();
  }
  public void setText(String givenText) {
    this.theTestArea.setText(givenText);
  }
  public int getHiliteIndex() {
    return this.theTestArea.getHiliteIndex();
  }
  public void setHiliteIndex(int givenIndex) {
    this.theTestArea.setHiliteIndex(givenIndex);
  }

  public BufferedImage getTestImage() {
    return this.theTestArea.getTestImage();
  }
  public AttachedShape getTestOutlineShape() {
    return this.theTestArea.getGlyphOutlineShape();
  }
  public AttachedShape getTestHiliteShape() {
    return this.theTestArea.getGlyphHiliteShape();
  }


  public void forceUpdate() {
    this.theTestArea.buildTestArea();
    this.theTestArea.forceUpdate();
  }
  // Behaviour
  public boolean action(Event givenEvent, Object givenObject) {
    if( (null != givenEvent) && (null != givenObject) && (null != this.theDispatcher) ) {
      if( "Exit".equals(givenObject) ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_EXIT_TEST);
	return true;
      }
      if( "Context Hints...".equals(givenObject) ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_REQUEST_TEST_HINTS);
	return true;
      }
      if( "Affine Transforms...".equals(givenObject) ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_REQUEST_TEST_TRANSFORM);
	return true;
      }
      if( "Font and Text...".equals(givenObject) ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_REQUEST_TEST_FONT_AND_TEXT);
	return true;
      }

      if( "Refresh".equals(givenObject) ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_REFRESH);
	return true;
      }
    }
    return false;
  }
  // Private methods
  private boolean _makeMenu() {
    MenuBar menuBar = null;
    boolean retCode = false;

    if( null != (menuBar = new MenuBar()) ) {
      Menu      menu;
      if( null != (menu = new Menu("File")) ) {
	menu.add("Exit");
	menuBar.add(menu);
	if( null != (menu = new Menu("Edit")) ) {
	  menu.add("Context Hints...");
	  menu.add("Affine Transforms...");
	  menu.add("-");
	  menu.add("Font and Text...");
	  menu.add("-");
	  menu.add("Refresh");
	  menuBar.add(menu);
	  setMenuBar(menuBar);
	  retCode = true;
	}
      }
    }
    return retCode;
  }
}


class GraphicsTestArea extends Canvas {

  BufferedImage       theTestImage;
  Graphics2D          theTestImageContext;

  Dimension           thePhysicalDimension;

  Object              theAntialias;
  Object              theRendering;
  Object              theTextAntialias;
  Object              theFractionalMetrics;
  boolean             theStroke;


  int                 borderSize;
  // Methods:
  // Public methods:
  // Constructor implies
  // Initializer
  public boolean init(int givenSizeX,int givenSizeY,int givenGraphicsFormat) {
    boolean    retCode = false;

    if( null != (this.thePhysicalDimension = new Dimension(givenSizeX,givenSizeY)) ) {
      if( null != (this.theTestImage = new BufferedImage(givenSizeX,givenSizeY,givenGraphicsFormat)) ) {
	if( null != (this.theTestImageContext = (Graphics2D)(this.theTestImage.getGraphics())) ) {
	  
	  if( givenSizeX < 200 ) {
	    borderSize = (200 - givenSizeX) / 2;
	  }
	  if( givenSizeY < 200 ) {
	    borderSize = (200 - givenSizeY) / 2;
	  }
	  
	  this.theAntialias         = RenderingHints.VALUE_ANTIALIAS_DEFAULT;
	  this.theRendering         = RenderingHints.VALUE_RENDER_DEFAULT;
	  this.theTextAntialias     = RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT;
	  this.theFractionalMetrics = RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT;
	  this.theStroke            = false;
     
	  buildTestArea();
	  retCode = true;
	}
      }
    }
    return retCode;
  }
  // Get - set methods
  public void setAntialias(Object givenAntialias) {
    this.theAntialias = givenAntialias;
  }
  public Object getAntialias() {
    return this.theAntialias;
  }
  public void setTextAntialias(Object givenTextAntialias) {
    this.theTextAntialias = givenTextAntialias;
  }
  public Object getTextAntialias() {
    return this.theTextAntialias;
  }
  public void setRendering(Object givenRendering) {
    this.theRendering = givenRendering;
  }
  public Object getRendering() {
    return this.theRendering;
  }
  public void setFractionalMetrics(Object givenFractionalMetrics) {
    this.theFractionalMetrics = givenFractionalMetrics;
  }
  public Object getFractionalMetrics() {
    return this.theFractionalMetrics;
  }
  public void setStroke(boolean givenStroke) {
    this.theStroke = givenStroke;
  }
  public boolean getStroke() {
    return this.theStroke;
  }
  public boolean getDrawString() {
    return false;
  }
  public void setDrawString(boolean givenDrawString) {
  }
  public boolean getDrawGlyphVector() {
    return false;
  }
  public void setDrawGlyphVector(boolean givenGlyphVector) {
  }
  public boolean getDrawOutline() {
    return false;
  }
  public void setDrawOutline(boolean givenDrawOutline) {
  }
  public boolean getDrawBounds() {
    return false;
  }
  public void setDrawBounds(boolean givenDrawBounds) {
  }
  public boolean getDrawLines() {
    return false;
  }
  public void setDrawLines(boolean givenDrawLines) {
  }
  public Font getFont() {
    return this.theTestImageContext.getFont();
  }
  public void setFont(Font givenFont) {
    this.theTestImageContext.setFont(givenFont);
  }
  public String getText() {
    return null;
  }
  public void setText(String givenText) {
  }
  
  public BufferedImage getTestImage() {
    return this.theTestImage;
  }

  // Size calculation
  public Dimension getPreferredSize() {
    Dimension retDim = new Dimension(this.thePhysicalDimension.width + this.borderSize * 2,this.thePhysicalDimension.height + this.borderSize * 2);
    return retDim;
  }
  // Paint 
  public void paint(Graphics givenContext) {
    _paintContent(givenContext);
  }
  // Update
  public void update(Graphics givenContext) {
    _paintContent(givenContext);
  }  
  // Update all the area notwithstanding clip regions
  public void forceUpdate() {
    invalidate();
    repaint();
  }
  // Test area preparation end action
  public void buildTestArea() {
    if( null != this.theTestImageContext ) {
      this.theTestImageContext.setRenderingHints(new RenderingHints(RenderingHints.KEY_ANTIALIASING,(Object)this.theAntialias));
      this.theTestImageContext.setRenderingHints(new RenderingHints(RenderingHints.KEY_RENDERING,(Object)this.theRendering));
      this.theTestImageContext.setRenderingHints(new RenderingHints(RenderingHints.KEY_TEXT_ANTIALIASING,(Object)this.theTextAntialias));
      this.theTestImageContext.setRenderingHints(new RenderingHints(RenderingHints.KEY_FRACTIONALMETRICS,(Object)this.theFractionalMetrics));
      paintTestArea(this.theTestImageContext);
    }
  }  
  // Set one pixel on the BufferedImage
  protected void drawPhysicalPixel(Graphics2D givenGraphics,int givenX,int givenY) {
    if( null != givenGraphics ) {
      givenGraphics.drawLine(givenX,givenY,givenX,givenY);
    }
  }
  // Paints test area
  // Must be abstract. jsut for test
  protected void paintTestArea(Graphics2D givenTestContext) {
    Color cC = givenTestContext.getColor();
    givenTestContext.setColor(new Color(0xFF,0xFF,0xFF));
    givenTestContext.fillRect(0,0,this.thePhysicalDimension.width,this.thePhysicalDimension.height);
    givenTestContext.setColor(new Color(0xAA,0x11,0x33));

    givenTestContext.drawLine(0,0,thePhysicalDimension.width,this.thePhysicalDimension.height);
    givenTestContext.drawLine(0,this.thePhysicalDimension.height,thePhysicalDimension.width,0);

    givenTestContext.setColor(cC);
  }
  // 

  // Protected group:
  // Update common end action
  protected void _paintContent(Graphics givenContext) {
    if( (null != givenContext) && (null != this.theTestImage) ) {
      Dimension resultDimenstion = getSize();
      if( null != resultDimenstion ) {
	int  sX = resultDimenstion.width - this.theTestImage.getWidth();
	int  sY = resultDimenstion.height - this.theTestImage.getHeight();
	givenContext.drawImage(this.theTestImage,sX/2,sY/2,null);
      }
    }
  }
}

class AttachedShape extends Object {
  Shape      theShape;
  Point      theOrigin;
  public boolean init(int givenX,int givenY,Shape givenShape) {
    boolean  retCode = false;
    if( null != givenShape ){
      if( null != (theOrigin = new Point(givenX,givenY)) ){
	this.theShape = givenShape;
	retCode = true;
      }
    }
    return retCode;
  }

  Shape getShape() {
    return this.theShape;
  }
  Point getOrigin() {
    return this.theOrigin;
  }
}

class FontTestArea extends GraphicsTestArea {
  Font                theFont;
  String              theText;
  GlyphVector         theGlyphVector;
  FontRenderContext   theFontRenderContext;
  FontMetrics         theMetrics;
  Dimension           theTextDimension;
  Shape               theGlyphOutline;
  Shape               theGlyphHilite;
  AttachedShape       theGlyphOutlineExporter;
  AttachedShape       theGlyphHiliteExporter;
  int                 theHiliteIndex;

  boolean             drawString;
  boolean             drawGlyphVector;
  boolean             drawBounds;
  boolean             drawLines;
  boolean             drawOutline;


  public boolean init(Font givenFont, String givenText,int givenSizeX,int givenSizeY, int givenImageFormat) {
    boolean     retCode = false;

    if( null != givenFont ) {
      if( null != givenText ) { 
	if( true == super.init(givenSizeX,givenSizeY,givenImageFormat)) {
	  if( null != (this.theGlyphOutlineExporter = new AttachedShape()) ) {
	    if( null != (this.theGlyphHiliteExporter = new AttachedShape()) ) {
	      this.theHiliteIndex = 8;
	      if( true == _setupFontAndTextRelatedData(givenFont,givenText) ) {
		this.drawString      = true;
		this.drawGlyphVector = true;
		this.drawBounds      = true;
		this.drawLines       = true;
		this.drawOutline     = true;	
		buildTestArea();
		retCode = true;	
	      }	  
	    }
	  }
	}	  
      }
    }
    return retCode;
  }

  protected void paintTestArea(Graphics2D givenTestContext) {
    
    Color    savedColor = givenTestContext.getColor();
    givenTestContext.setColor(new Color(0xFF,0xFF,0xFF));
    givenTestContext.fillRect(0,0,this.theTestImage.getWidth(),this.theTestImage.getHeight());
    givenTestContext.setColor(savedColor);
    
    if( true == this.drawOutline ) {
      _drawOutline(givenTestContext);
    }
    if( true == this.drawString ) {
      _drawString(givenTestContext);
    }
    if( true == this.drawGlyphVector ) {
      _drawGlyphVector(givenTestContext);
    }
    if( true == this.drawBounds ) {
      _drawBounds(givenTestContext);
    }
    if( true == this.drawLines ) {
      _drawLines(givenTestContext);
    }

    //    givenTestContext.setColor(new Color(0xFF,0x00,0x44));
    //    givenTestContext.drawLine(0,0,this.theTestImage.getWidth(),this.theTestImage.getHeight());
    //    givenTestContext.setColor(savedColor);


  }

  protected void _drawString(Graphics2D givenContext) {
    Color   savedColor = givenContext.getColor();
    float   X = (this.theTestImage.getWidth() - this.theTextDimension.width) / 2;
    float   Y = (this.theTestImage.getHeight() - this.theTextDimension.height) / 2; 

    givenContext.setColor(new Color(0x22,0x00,0x77));

    givenContext.drawString(this.theText,X,Y);

    givenContext.setColor(savedColor);
  }
  protected void _drawGlyphVector(Graphics2D givenContext) {
    Color   savedColor = givenContext.getColor();
    float   X = (this.theTestImage.getWidth() - this.theTextDimension.width) / 2;
    float   Y = (this.theTestImage.getHeight() - this.theTextDimension.height) / 2; 

    givenContext.setColor(new Color(0x77,0x00,0x22));
    
    givenContext.drawGlyphVector(this.theGlyphVector,X,Y); 

    givenContext.setColor(savedColor);

  }
  protected void _drawOutline(Graphics2D givenContext) {
    Color   savedColor = givenContext.getColor();
    float   X = (this.theTestImage.getWidth() - this.theTextDimension.width) / 2;
    float   Y = (this.theTestImage.getHeight() - this.theTextDimension.height) / 2; 
    AffineTransform savedTransform = givenContext.getTransform();

    //    givenContext.setXORMode(new Color(0x11,0x88,0x22));
    givenContext.setColor(new Color(0x11,0x88,0x22));

    givenContext.translate((int)X - 1,(int)Y - 1);

    if( true == this.theStroke ) {
      givenContext.fill(this.theGlyphOutline);
    } else {
      givenContext.draw(this.theGlyphOutline);
    } 

    givenContext.setTransform(savedTransform);
    givenContext.setColor(savedColor);
    //    givenContext.setPaintMode();
 
  }
  protected void _drawLines(Graphics2D givenContext) {

    Color         savedColor = givenContext.getColor();
    float   X = (this.theTestImage.getWidth() - this.theTextDimension.width) / 2;
    float   Y = (this.theTestImage.getHeight() - this.theTextDimension.height) / 2; 

    int     ascent = this.theMetrics.getAscent();
    int     descent = this.theMetrics.getDescent();

    Rectangle2D   bounds = this.theGlyphVector.getVisualBounds();

    givenContext.setColor(new Color(0x00,0xEE,0x22));

    {
      int X0        = (int)(X);
      int X1        = (int)bounds.getWidth() + X0;

      int YBaseline = (int)(Y);
      int YLowerline= (int)(Y + bounds.getY() - 1) + (int)bounds.getHeight();
      int YUpperline= (int)(Y + bounds.getY() - 1);

      givenContext.drawLine(X0,YUpperline,X1,YUpperline);
      givenContext.drawLine(X0,YBaseline,X1,YBaseline);
      givenContext.drawLine(X0,YLowerline,X1,YLowerline);


    }
    givenContext.setColor(savedColor);
    
  }
  protected void _drawBounds(Graphics2D givenContext) {
    Color   savedColor = givenContext.getColor();
    float   X          = (this.theTestImage.getWidth() - this.theTextDimension.width) / 2;
    float   Y          = (this.theTestImage.getHeight() - this.theTextDimension.height) / 2; 

    Rectangle2D bounds = this.theGlyphVector.getVisualBounds();

    givenContext.setColor(new Color(0xBB,0x00,0x22)); 

    {
      int X0 = (int)(X + bounds.getX() - 1);
      int Y0 = (int)(Y + bounds.getY() - 1);
      int X1 = X0 + (int)bounds.getWidth();
      int Y1 = Y0 + (int)bounds.getHeight();

      givenContext.drawLine(X0,Y0,X1,Y0);
      givenContext.drawLine(X0,Y1,X1,Y1);
      givenContext.drawLine(X0,Y0,X0,Y1);
      givenContext.drawLine(X1,Y0,X1,Y1);
     
    }

    givenContext.setColor(savedColor);

  }
 
  private boolean _setupFontAndTextRelatedData(Font givenFont,String givenText){
    boolean    retCode = false;
    if( (null != givenFont) && (null != givenText) ) {
      this.theFont = givenFont;
      this.theText = givenText;

      if( this.theText.length() < this.theHiliteIndex - 1 ) {
	this.theHiliteIndex = this.theText.length() - 1;
      }

      if( null != this.theTestImageContext ) {
	if( null != (this.theFontRenderContext = this.theTestImageContext.getFontRenderContext()) ) {
	  if( null != (this.theGlyphVector = this.theFont.createGlyphVector(this.theFontRenderContext,this.theText)) ) {
	    if( null != (this.theMetrics = this.theTestImageContext.getFontMetrics(this.theFont)) ) {
	      if( null != (this.theTextDimension = new Dimension(this.theMetrics.stringWidth(this.theText),theMetrics.getHeight())) ) {
		if( null != (this.theGlyphOutline = this.theGlyphVector.getOutline()) ) {
		  if( null != (this.theGlyphHilite = this.theGlyphVector.getGlyphOutline(this.theHiliteIndex)) ) {
		    float   X = (this.theTestImage.getWidth() - this.theTextDimension.width) / 2;
		    float   Y = (this.theTestImage.getHeight() - this.theTextDimension.height) / 2; 
		    this.theTestImageContext.setFont(this.theFont);
		    if( false != this.theGlyphOutlineExporter.init((int)X-1,(int)Y-1,this.theGlyphOutline) ){

		      if( 0 != this.theHiliteIndex ) { 
			String activePart = this.theText.substring(0,this.theHiliteIndex);
			float  dX         = this.theMetrics.stringWidth(activePart);
			X += dX;
		      }
		      if( false != this.theGlyphHiliteExporter.init((int)X-1,(int)Y-1,this.theGlyphHilite) ){
	
			retCode = true;
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }   
    return retCode;
  }

  public boolean getDrawString() {
    return this.drawString;
  }
  public void setDrawString(boolean givenDrawString) {
    this.drawString = givenDrawString;
  }
  public boolean getDrawGlyphVector() {
    return this.drawGlyphVector;
  }
  public void setDrawGlyphVector(boolean givenGlyphVector) {
    this.drawGlyphVector = givenGlyphVector;
  }
  public boolean getDrawOutline() {
    return this.drawOutline;
  }
  public void setDrawOutline(boolean givenDrawOutline) {
    this.drawOutline = givenDrawOutline;
  }
  public boolean getDrawBounds() {
    return this.drawBounds;
  }
  public void setDrawBounds(boolean givenDrawBounds) {
    this.drawBounds = givenDrawBounds;
  }
  public boolean getDrawLines() {
    return this.drawLines;
  }
  public void setDrawLines(boolean givenDrawLines) {
    this.drawLines = givenDrawLines;
  }
  public Font getFont() {
    return this.theFont;
  }
  public void setFont(Font givenFont) {
    if( null != givenFont ) {    
      _setupFontAndTextRelatedData(givenFont,this.theText);  
    }

  }
  public String getText() {
    return this.theText;
  }
  public void setText(String givenText) {
    if( null != givenText ) {
      _setupFontAndTextRelatedData(this.theFont,givenText);
    }
  }
  public void setHiliteIndex(int givenIndex){
    this.theHiliteIndex = givenIndex;
    _setupFontAndTextRelatedData(this.theFont,this.theText);
  }
  public int getHiliteIndex(){
    return this.theHiliteIndex;
  }

  public AttachedShape getGlyphOutlineShape() {
    return this.theGlyphOutlineExporter;
  }
  public AttachedShape getGlyphHiliteShape() {
    return this.theGlyphHiliteExporter;
  }

}

//====================================================================================================
//  GRAPHICS TEST AREA ends here
//====================================================================================================



//====================================================================================================
//  MAGNIFIER (Prototype. To be worked for efficient Scrolling and better design)
//====================================================================================================
//  Magnifier controller
class MagnifierController extends Panel implements ItemListener{ 
  Button        theZoomIn;
  Button        theZoomOut;
  Button        theUp;
  Button        theDown;
  Button        theLeft;
  Button        theRight;

  Dispatcher    theDispatcher;

  public boolean init(Dispatcher givenDispatcher) {
    boolean  retCode = false;

    if( null != givenDispatcher ) {
      if( null != (theZoomIn = new Button("+")) ) {
	if( null != (theZoomOut = new Button("-")) ) {
	  if( null != (theLeft = new Button("<")) ) {
	    if( null != (theRight = new Button(">")) ) {
	      if( null != (theUp = new Button("^")) ) {
		if( null != (theDown = new Button("v")) ) {
		  add(this.theZoomIn);
		  add(this.theZoomOut);
		  
		  add(this.theUp);
		  add(this.theDown);
		  add(this.theLeft);
		  add(this.theRight);
		  
		  this.theDispatcher = givenDispatcher;
		  
		  retCode = true;
		}
	      }
	    }
	  }
	}
      }
    }
    return retCode;
  }
  
  public void itemStateChanged(ItemEvent e){
    repaint();
  }
  public boolean action(Event givenEvent, Object givenObject) {
    if( (null != givenEvent) && (null != givenObject) ) {
      if( givenEvent.target == this.theZoomIn ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_ZOOM_IN);
      }
      if( givenEvent.target == this.theZoomOut ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_ZOOM_OUT);
      }
      if( givenEvent.target == this.theLeft ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SCROLL_LEFT);
      }
      if( givenEvent.target == this.theRight ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SCROLL_RIGHT);
      }
      if( givenEvent.target == this.theUp ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SCROLL_UP);
      }
      if( givenEvent.target == this.theDown ) {
	this.theDispatcher.dispatchCommand(RelationDispatcher.COMMAND_SCROLL_DOWN);
      }
    }
    return false;
  }
}
//  End of magnifier controller
//  Magnifier Workfield Field
abstract class Field extends Canvas {

  Dimension         theLogicalDimension;
  Dimension         thePhysicalDimension;

  Point             theOffset;  // Positive offset, shows the position of the start of logical dimension in physical dimension. Physical elements.
  int               theScrollStep;
  
  int               thePPUFactor; 

  //  Image             theOffscreen;

  public boolean init(int givenPhysicalWidth,int givenPhysicalHeight,int givenLogicalWidth,int givenLogicalHeight,int givenPPUFactor) {
    boolean retCode = false;
    if( null != (this.thePhysicalDimension  = new Dimension(givenPhysicalWidth,givenPhysicalHeight)) ) {
      if( null != (this.theLogicalDimension = new Dimension(givenLogicalWidth,givenLogicalHeight)) ) {
	if( null != (this.theOffset = new Point(0,0)) ) {
	  this.thePPUFactor  = givenPPUFactor; 
	  this.theScrollStep = givenPPUFactor / 2;
	  //	this.theOffscreen = null;
	  retCode = true;
	}
      }
    }
    return retCode;
  }
  public Dimension getPreferredSize() {
    return this.thePhysicalDimension;
  }

  public void setPPUFactor(int givenPPUFactor) {
    this.thePPUFactor = givenPPUFactor;
    repaint();
  }

  public void paint(Graphics givenContext) {
    _paintContent(givenContext);
  }

  public void update(Graphics givenContext) {
    _paintContent(givenContext);
  }
  
  public void forceUpdate() {
    invalidate();
    repaint();
  }

  public void scrollLeft() {
    if( this.theOffset.x > 0 ) {
      this.theOffset.x -= _getScrollStep();
      if( this.theOffset.x < 0 ) {
	this.theOffset.x = 0;
      }
      forceUpdate();
    }
  }
  public void scrollRight() {
    int    nN;
    if( this.theOffset.x <= (nN = this.theLogicalDimension.width * this.thePPUFactor - this.thePhysicalDimension.width) ) {
      this.theOffset.x += _getScrollStep();
      if( this.theOffset.x > nN ) {
	this.theOffset.x = nN;
      }
      forceUpdate();
    }
  }
  public void scrollUp() {
    if( this.theOffset.y > 0 ) {
      this.theOffset.y -= _getScrollStep();
      if( this.theOffset.y < 0 ) {
	this.theOffset.y = 0;
      }
      forceUpdate();
    }
  }
  public void scrollDown() {
    int    nN;
    if( this.theOffset.y <= (nN = this.theLogicalDimension.height * this.thePPUFactor - this.thePhysicalDimension.height) ) {
      this.theOffset.y += _getScrollStep();
      if( this.theOffset.y > nN ) {
	this.theOffset.y = nN;
      }
      forceUpdate();
    }
  }


  public void normalizeOffset() {
    int   nN;
    if( this.theOffset.x > (nN = this.theLogicalDimension.width * this.thePPUFactor - this.thePhysicalDimension.width) ) {
      this.theOffset.x = nN;
    }
    if( this.theOffset.y > (nN = this.theLogicalDimension.height * this.thePPUFactor - this.thePhysicalDimension.height) ) {
      this.theOffset.y = nN;
    }

    if( this.theOffset.x < 0 ) {
      this.theOffset.x = 0;
    }
    if( this.theOffset.y < 0 ) {
      this.theOffset.y = 0;
    }    
  }

  
  protected synchronized void _paintContent(Graphics givenContext) {
    Graphics          offscreenGraphics;
    if( null != givenContext ) {
      if( null != (this.thePhysicalDimension = getSize()) ) {
	if( (0!=thePhysicalDimension.width) && (0!=thePhysicalDimension.height) ) {
	  //	  if( null != (this.theOffscreen = createImage(thePhysicalDimension.width,thePhysicalDimension.height)) ) {
	  //	    if( null != (offscreenGraphics = this.theOffscreen.getGraphics()) ) {
	      givenContext.clearRect(0,0,thePhysicalDimension.width,thePhysicalDimension.height);
	      normalizeOffset();
	      _do_paint(givenContext);
	  //	      givenContext.drawImage(this.theOffscreen,0,0,null);
	  //	    }
	  //	  }
	}
      }
    }
  }

  protected synchronized void _do_paint(Graphics givenContext) {
    paintContent(givenContext);
    _paintGrid(givenContext);
  }

  public Point l2P(Point pP) {
    pP.x = (pP.x * this.thePPUFactor) - this.theOffset.x;
    pP.y = (pP.y * this.thePPUFactor) - this.theOffset.y;
    return pP;
  }

  public Point p2L(Point pP) {
    pP.x = (pP.x + this.theOffset.x) / this.thePPUFactor;
    pP.y = (pP.y + this.theOffset.y) / this.thePPUFactor;

    return pP;
  }

  public void drawLogicalPixel(Graphics givenContext,int givenX,int givenY) {
    if( null != givenContext ) {
      Point  theOrigin = new Point(givenX,givenY);
      if( null != theOrigin ) {
	int     X,Y;
	int     nDX = this.thePPUFactor;
	int     nDY = this.thePPUFactor;

	if( false == _visibleLogicalPixel(givenX,givenY) ) {
	  return;
	}

	l2P(theOrigin);
	X = theOrigin.x;
	Y = theOrigin.y;

	if( X < 0 ) {
	  nDX += X;
	  X = 0;
	}
	if( Y < 0 ) {
	  nDY += Y;
	  Y = 0;
	}
	
	if( nDX < 0 ) {
	  return;
	}
	if( nDY < 0 ) {
	  return;
	}
	givenContext.fillRect(X,Y,nDX,nDY);      
      }
    }
  }
  public void markLogicalPixel(Graphics givenContext,int givenX,int givenY) {
    if( null != givenContext ) {
      Point  theOrigin = new Point(givenX,givenY);
      if( null != theOrigin ) {
	int     X,Y,XX,YY;
	int     nDDX,nDX;
	int     nDDY,nDY;

	nDDX = nDX = this.thePPUFactor;
	nDDY = nDY = this.thePPUFactor;

	if( false == _visibleLogicalPixel(givenX,givenY) ) {
	  return;
	}

	l2P(theOrigin);
	XX = X = theOrigin.x;
	YY = Y = theOrigin.y;

	XX += this.thePPUFactor / 4;
	XX += this.thePPUFactor / 4;
	nDDX -= this.thePPUFactor / 4;
	nDDY -= this.thePPUFactor / 4;

	if( X < 0 ) {
	  nDX += X;
	  X = 0;
	}
	if( Y < 0 ) {
	  nDY += Y;
	  Y = 0;
	}
	
	if( nDX < 0 ) {
	  return;
	}
	if( nDY < 0 ) {
	  return;
	}
	givenContext.drawRect(X,Y,nDX,nDY);  

	if( XX < 0 ) {
	  nDDX += XX;
	  XX = 0;
	}
	if( YY < 0 ) {
	  nDDY += YY;
	  YY = 0;
	}
	
	if( nDDX < 0 ) {
	  return;
	}
	if( nDDY < 0 ) {
	  return;
	}


	givenContext.fillRect(XX,YY,nDDX,nDDY);
	
      }
    }
  }

  public void drawLogicalPixel(Graphics givenContext,int givenX,int givenY, Color givenColor) {
    if( null != givenContext ) {
      Color savedColor = givenContext.getColor();
      givenContext.setColor(givenColor);
      drawLogicalPixel(givenContext,givenX,givenY);
      givenContext.setColor(savedColor);
    }
  }
  public void markLogicalPixel(Graphics givenContext,int givenX,int givenY, Color givenColor) {
    if( null != givenContext ) {
      givenContext.setXORMode(givenColor);
      markLogicalPixel(givenContext,givenX,givenY);
      givenContext.setPaintMode();
    }
  }

  public void drawLogicalPixel(Graphics givenContext,int givenX,int givenY, int givenRGB) {
      drawLogicalPixel(givenContext,givenX,givenY,new Color(givenRGB));    
  }
  public void markLogicalPixel(Graphics givenContext,int givenX,int givenY, int givenRGB) {
      markLogicalPixel(givenContext,givenX,givenY,new Color(givenRGB));    
  }

  private boolean _visibleLogicalPixel(int givenX, int givenY) {
    if( givenX < 0 ) {
      return false;
    }
    if( givenY < 0 ) {
      return false;
    }
    if( givenX > this.theLogicalDimension.width ) {
      return false;
    }
    if( givenY > this.theLogicalDimension.height ) {
      return false;
    }
    return true;
  }

  private void _paintGrid(Graphics givenContext) {
    Dimension theArea = getSize();
    if( (null != theArea) && (null != givenContext) ) {  
      givenContext.setXORMode(new Color(0x77,0x77,0x77));
      
      {
	int   X,Y;
	
	for( X = 0; X < theArea.width; X += this.thePPUFactor) {
	  givenContext.drawLine(X,0,X,theArea.height);
	}
	for( Y = 0; Y < theArea.height; Y += this.thePPUFactor) {
	  givenContext.drawLine(0,Y,theArea.width,Y);
	}
      }      
      givenContext.setPaintMode();
    }
  }

  protected int _getScrollStep() {
    return this.theScrollStep;
  }

  abstract public void paintContent(Graphics givenContext);
}
//  End of Magnifier Field
//  Magnifier itself
class Magnifier extends Field {
  int               theMinPPU;
  int               theMaxPPU;
  int               thePPUStep;



 public boolean init(int givenPhysicalWidth,int givenPhysicalHeight,int givenLogicalWidth,int givenLogicalHeight,int givenPPUFactor,int givenMinPPU,int givenMaxPPU, int givenPPUStep) {
   boolean retCode = false;
   
   if( true == super.init(givenPhysicalWidth,givenPhysicalHeight,givenLogicalWidth,givenLogicalHeight,givenPPUFactor) ) {
     this.theMinPPU    = givenMinPPU;
     this.theMaxPPU    = givenMaxPPU;
     this.thePPUStep   = givenPPUStep;
     retCode = true;
   }
   return retCode;
  }

  public Object getValue(int valueCode) {
    return null;
  }

  public void zoomIn() {
    if( this.thePPUFactor < this.theMaxPPU ) {
      this.thePPUFactor += this.thePPUStep;
      normalizeOffset();
      forceUpdate();
    }
  }
  
  public void zoomOut() {
    if( this.thePPUFactor > this.theMinPPU ) {
      this.thePPUFactor -= this.thePPUStep;
      normalizeOffset();
      forceUpdate();
    }
  }

  public void paintContent(Graphics givenContext) {
    if( null != givenContext ) {
      int    X,Y;
      Color sc = givenContext.getColor();
      givenContext.setColor(new Color(0xAA,0x00,0x11));
      for( X = 0, Y = this.theLogicalDimension.height / 2; X <= this.theLogicalDimension.width - 1; X ++ ) {
	drawLogicalPixel(givenContext,X,Y);
      } 
      for( Y = 0, X = this.theLogicalDimension.width / 2; Y <= this.theLogicalDimension.height - 1; Y ++ ) {
	drawLogicalPixel(givenContext,X,Y);
      } 
      
      for( X = 0; X < this.theLogicalDimension.width; X ++ ) { 
		drawLogicalPixel(givenContext,X,0);
		drawLogicalPixel(givenContext,X,this.theLogicalDimension.height - 1);
      } 
      for( Y = 0; Y < this.theLogicalDimension.height; Y ++ ) {
	drawLogicalPixel(givenContext,0,Y);
	drawLogicalPixel(givenContext,this.theLogicalDimension.width - 1,Y);
      } 
      givenContext.setColor(sc);
    }
  }
  public void paintContent1(Graphics givenContext) {
    if( null != givenContext ) {
      for( int X = 0; X < this.theLogicalDimension.width; X ++ ) {
	for( int Y = 0; Y < this.theLogicalDimension.height; Y ++ ) {
	  drawLogicalPixel(givenContext,X,Y);
	} 
      }     
    }
  }
}
//  End of magnifier itself
//  Image magnifier
class ImageMagnifier extends Magnifier {

  BufferedImage        theImage;

  public boolean init(BufferedImage givenImage,int givenPhysicalWidth,int givenPhysicalHeight,int givenPPUFactor,int givenMinPPU,int givenMaxPPU,int givenPPUStep) {
    boolean retCode = false;
    if( null != givenImage ) {
      if( true == super.init(givenPhysicalWidth,givenPhysicalHeight,givenImage.getWidth(),givenImage.getHeight(),givenPPUFactor,givenMinPPU,givenMaxPPU,givenPPUStep) ) {
	this.theImage = givenImage;
	retCode = true;
      }
    }
    return retCode;
  }

  public void paintContent(Graphics givenContext) {
    if( null != givenContext ) {
      //      int X,Y;
      //      int LX = this.theImage.getWidth();
      //      int LY = this.theImage.getHeight();

      int theRGB;

      int X = this.theOffset.x / this.thePPUFactor;
      int Y = this.theOffset.y / this.thePPUFactor;
      int LX = this.theImage.getWidth() - X;
      int LY = this.theImage.getHeight() - Y;

      if( LX > this.thePhysicalDimension.width ) {
	LX = this.thePhysicalDimension.width;
      }
      if( LY > this.thePhysicalDimension.height ) {
	LY = this.thePhysicalDimension.height;
      }

      for( Y = 0; Y < LY; Y ++ ) {
	for( X = 0; X < LX; X ++ ) {
	  theRGB = this.theImage.getRGB(X,Y);

	  drawLogicalPixel(givenContext,X,Y,theRGB);
	}
      }      
    }
  }
}

class ImageShapeMagnifier extends ImageMagnifier {
  AttachedShape theOutlineShape;
  AttachedShape theHiliteShape;

  public boolean init(AttachedShape givenOutlineShape,AttachedShape givenHiliteShape,BufferedImage givenImage,int givenPhysicalWidth,int givenPhysicalHeight,int givenPPUFactor,int givenMinPPU,int givenMaxPPU,int givenPPUStep) {
    boolean retCode = false;
    if( (null != givenOutlineShape) && (null != givenHiliteShape) ) {
      if( true == super.init(givenImage,givenPhysicalWidth,givenPhysicalHeight,givenPPUFactor,givenMinPPU,givenMaxPPU,givenPPUStep) ) {
	this.theOutlineShape = givenOutlineShape;
	this.theHiliteShape = givenHiliteShape;
	retCode = true;
      }
    }
    return retCode;
  }

  public void paintContent(Graphics givenContext) {
    if( null != givenContext ) {
      super.paintContent(givenContext);
      drawLogicalShape((Graphics2D)givenContext,this.theOutlineShape.getShape(),theOutlineShape.getOrigin().x,theOutlineShape.getOrigin().y,new Color(0xFF,0x00,0x00));  // 't is silly but for now...
      drawLogicalShape((Graphics2D)givenContext,this.theHiliteShape.getShape(),theHiliteShape.getOrigin().x,theHiliteShape.getOrigin().y,new Color(0x77,0x77,0xFF));  // 't is silly but for now...
    }      
  }

  public void drawLogicalShape(Graphics2D givenContext,Shape givenShape,int givenX,int givenY,Color givenColor) {
    Color   toSave = givenContext.getColor();
    
    givenContext.setColor(givenColor);

    drawLogicalShape(givenContext,givenShape,givenX,givenY);
    
    givenContext.setColor(toSave);
  }
  public void drawLogicalShape(Graphics2D givenContext,Shape givenShape,int givenX,int givenY) {
    if( null != givenContext ) {
      Point  theOrigin = new Point(givenX,givenY);
      if( null != theOrigin ) {
	int     X,Y;
	int     nDT = this.thePPUFactor;

	l2P(theOrigin);
	X = theOrigin.x;
	Y = theOrigin.y;

	{
	  GeneralPath toDraw = new GeneralPath(givenShape);
	  if( null != toDraw ) {
	    AffineTransform toTrans = new AffineTransform();
	    if( null != toTrans ) {
	      toTrans.scale(nDT,nDT);
	      toDraw.transform(toTrans);
	      {
		AffineTransform toSave = givenContext.getTransform();


		givenContext.translate((int)X,(int)Y);

		givenContext.draw(toDraw);


		givenContext.setTransform(toSave);

 
	      }
	    }
	  }
	}      
      }
    }
  }


}



//====================================================================================================
//  MAGNIFIER ends here
//====================================================================================================









