/*
 * @(#)ConcurrentThreads.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.awt.*;

class ThreadShell extends Thread 
{
  public ThreadShell(String name) {
    super(name);
  }
  public ThreadShell() {
    super();
  }
  public void run() {
    try {
      runThread();
    } catch(Exception e) {

    }
  }

  void runThread() {
  }
}

class PaintThread extends ThreadShell
{
  static final int   PROC_PAINT       = 0x0001;
  static final int   PROC_WAIT_PAINT  = 0x0002;
  static final int   PROC_WAIT_ERASE  = 0x0003;
  static final int   PROC_ERASE       = 0x0004;

  private Color     paintColor;
  private int       procStatus;
  private Graphics  context;
  private Rectangle targetArea;
  private boolean   errStatus;

  public boolean initialize(String name,Color cC,Graphics gG,Rectangle tA,boolean erase) {
    boolean retCode = true;
    if( null != cC ) {
      this.paintColor = cC;
      if( null != gG ) {
	this.context = gG;
	if( null != tA ) {
	  this.targetArea = tA;
	  this.errStatus = erase;
	  this.procStatus = PROC_ERASE;
	  this.setPriority(MIN_PRIORITY);
	  if( null != name ) {
	    this.setName(name);
	  }
	  this.start();
	  retCode = true;
	}
      }
    }
   return retCode;
  }
  public void setErase() {
    this.errStatus = true;
  }
  public void cleanErase() {
    this.errStatus = false;
  }
  public boolean terminate() {
    this.stop();
    this.context.clearRect(targetX(),targetY(),targetW(),targetH());
    return true;
  }

  void runThread() {
    Color    saved;
    long     curTime;
    long     lastTime = System.currentTimeMillis();
    Graphics g        = this.context;
    while(true) {
      curTime  = System.currentTimeMillis();
      switch(this.procStatus) {
      case PROC_PAINT:
	saved = g.getColor();
	g.setColor(this.paintColor);
	g.fillRect(targetX(),targetY(),targetW(),targetH());
	g.setColor(saved);
	this.procStatus = PROC_WAIT_PAINT;
	System.out.println(">>> "+getName()+" paint");
	break;
      case PROC_WAIT_PAINT:
	curTime = System.currentTimeMillis();
	if( curTime - lastTime > 10 ) {
	  if( true == this.errStatus ) {
	    this.procStatus = PROC_ERASE;
	  } else {
	    this.procStatus = PROC_PAINT;
	  }
	}
	System.out.println(">>> "+getName()+" wait after paint");
	break;
      case PROC_WAIT_ERASE:
	curTime = System.currentTimeMillis();
	if( curTime - lastTime > 10 ) {
	  this.procStatus = PROC_PAINT;
	}
	System.out.println(">>> "+getName()+" wait after erase");
	break;
      case PROC_ERASE:
	g.clearRect(targetX(),targetY(),targetW(),targetH());
	this.procStatus = PROC_WAIT_ERASE;
	System.out.println(">>> "+getName()+" erase");
	break;
      }
      lastTime = curTime;
    }
  }


  int targetX() {
    int retCode = 0;
    if( null != this.targetArea ) {
      retCode = this.targetArea.x;
    }
    return retCode;
  }
  int targetY() {
    int retCode = 0;
    if( null != this.targetArea ) {
      retCode = this.targetArea.y;
    }
    return retCode;
  }
  int targetW() {
    int retCode = 0;
    if( null != this.targetArea ) {
      retCode = this.targetArea.width;
    }
    return retCode;
  }
  int targetH() {
    int retCode = 0;
    if( null != this.targetArea ) {
      retCode = this.targetArea.height;
    }
    return retCode;
  }
  
};




public class ConcurrentThreads extends Canvas 
{
  boolean errStatus;

  Color savedColor;

  PaintThread  th1,th2,th3;

  
  public ConcurrentThreads() {    
    setBackground(Color.white);
    th1 = null;
    th2 = null;
    th3 = null;
    errStatus = false;
  }
  
  public void paint(Graphics g) {  
    savedColor = g.getColor();
    g.setColor(savedColor);
  }
  
  public static void main(String s[]) {
    AppletFrame          f = new AppletFrame();
    ConcurrentThreads    a = new ConcurrentThreads();
    f.init(a);
    
    f.add("Center", a);  
    f.pack();  
    f.setSize(new Dimension(435, 300));  
    f.show(); 
    
  }

  void doExit() {
    stopT1();
    stopT2();
    stopT3();
    System.exit(0);
  }
  void setErase() {
    if( null != th1 ) {
      th1.setErase();
    }
    if( null != th2 ) {
      th2.setErase();
    }
    if( null != th3 ) {
      th3.setErase();
    }
    errStatus = true;
  }
  void cleanErase() {
    if( null != th1 ) {
      th1.cleanErase();
    }
    if( null != th2 ) {
      th2.cleanErase();
    }
    if( null != th3 ) {
      th3.cleanErase();
    }
    errStatus = false;
  }
  void startT1() {
    if( null == th1 ) {
      System.out.println("Starting thread");
      th1 = new PaintThread();
      if( false == th1.initialize("T0",new Color(0x00,0x00,0xFF),getGraphics(),new Rectangle(5,5,100,100),errStatus) ) {
	System.out.println("Failed");
      } 
    }
  }
  void startT2() {
    if( null == th2 ) {
      System.out.println("Starting thread");
      th2 = new PaintThread();
      if( false == th2.initialize("T1",new Color(0x00,0xFF,0x00),getGraphics(),new Rectangle(110,5,100,100),errStatus) ) {
	System.out.println("Failed");
      }
    }
  }
  void startT3() {
    if( null == th3 ) {
      System.out.println("Starting thread");
      th3 = new PaintThread();
      if( false == th3.initialize("T1",new Color(0xFF,0x00,0x00),getGraphics(),new Rectangle(50,50,100,100),errStatus) ) {
	System.out.println("Failed");
      }
    }
  }
  void stopT1() {
    if( null != th1 ) {
      th1.terminate();
      th1 = null;
    }
  }
  void stopT2() {
    if( null != th2 ) {
      th2.terminate();
      th2 = null;
    }
  }
  void stopT3() {
    if( null != th3 ) {
      th3.terminate();
      th3 = null;
    }
  }


  void drawCross(Graphics g,int x,int y, int dx, int dy) {
    int x1 = x + dx;
    int y1 = y + dy;
    
    g.drawLine(x,y,x1,y1);
    g.drawLine(x,y1,x1,y);
  }
  
  public Dimension getPreferredSize() {
    return new Dimension(435,300);
  }
  
} 

class AppletFrame extends Frame
{
  ConcurrentThreads canvas;
  MenuItem          movItem;
 

  public boolean init(ConcurrentThreads ca) {
    canvas = ca;
    return _makeMenu();
  }

  private boolean _makeMenu() {
    MenuBar menuBar = null;
    boolean retCode = false;
    
    if( null != (menuBar = new MenuBar()) ) {
      Menu      menu;
      if( null != (menu = new Menu("File")) ) {
	menu.add("Exit");
	menuBar.add(menu);
	if( null != (menu = new Menu("Action")) ) {
	  menu.add("Start T1");
	  menu.add("Start T2");
	  menu.add("Start T3");
	  menu.add("Stop T1");
	  menu.add("Stop T2");
	  menu.add("Stop T3");
	  menu.addSeparator();
	  menu.add("- Erase");
	  movItem = menu.getItem(7);
	  menuBar.add(menu);
	  setMenuBar(menuBar);
	  retCode = true;
	}
      }
    }
    return retCode;
  }

  public boolean action(Event givenEvent, Object givenObject) {
    if( (null != givenEvent) && (null != givenObject) && (null != this.canvas) ) {
      if( "Exit".equals(givenObject) ) {
	canvas.doExit();
	return true;
      }
      if( "- Erase".equals(givenObject) ) {
	canvas.setErase();
	movItem.setLabel("+ Erase");
	return true;
      }
      if( "+ Erase".equals(givenObject) ) {
	canvas.cleanErase();
	movItem.setLabel("- Erase");
	return true;
      }
      if( "Start T1".equals(givenObject) ) {
	canvas.startT1();
	return true;
      }
      if( "Start T2".equals(givenObject) ) {
	canvas.startT2();
	return true;
      }
      if( "Start T3".equals(givenObject) ) {
	canvas.startT3();
	return true;
      }
      if( "Stop T1".equals(givenObject) ) {
	canvas.stopT1();
	return true;
      }
      if( "Stop T2".equals(givenObject) ) {
	canvas.stopT2();
	return true;
      }
      if( "Stop T3".equals(givenObject) ) {
	canvas.stopT3();
	return true;
      }
    }
    return false;
  }

}





