/*
 * @(#)CopyArea.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.awt.*;


public class CopyArea extends Canvas 
{
  static final short          CLEAN                 = 0x0001;
  static final short          COPY                  = 0x0002;
  static final short          COPY_OVERLAP          = 0x0003;
  static final short          COPY_CLIP             = 0x0004;
  static final short          COPY_CLIP_OVERLAP     = 0x0005;

  short status;

  Rectangle  copySource;
  Rectangle  copyDestination;
  Rectangle  copyOverlapSource;
  Rectangle  copyOverlapDestination;
  Rectangle  copyClipSource;
  Rectangle  copyClipDestination;
  Rectangle  copyClipOverlapSource;
  Rectangle  copyClipOverlapDestination;

  Color copySrcColor;
  Color copyDstColor;
  Color copyOverlapSrcColor;
  Color copyOverlapDstColor;
  Color copyClipSrcColor;
  Color copyClipDstColor;
  Color copyClipOverlapSrcColor;
  Color copyClipOverlapDstColor;

  Color colorTestCross;

 
  Color savedColor;
  
  public CopyArea() {    
    setBackground(Color.white);

    copySource                 = new Rectangle(5,5,100,100);
    copyDestination            = new Rectangle(5,105,100,100);

    copyOverlapSource          = new Rectangle(110,5,100,100);
    copyOverlapDestination     = new Rectangle(110,55,100,100);

    copyClipSource             = new Rectangle(215,5,100,100);
    copyClipDestination        = new Rectangle(215,105,80,80);

    copyClipOverlapSource      = new Rectangle(320,5,100,100);
    copyClipOverlapDestination = new Rectangle(320,75,80,80);


    copySrcColor               = new Color(0xFF,0x00,0x00);
    copyDstColor               = new Color(0x77,0x00,0x00);

    copyOverlapSrcColor        = new Color(0xFF,0x00,0x00);
    copyOverlapDstColor        = new Color(0x77,0x00,0x00);
    
    copyClipSrcColor           = new Color(0xFF,0x00,0x00);
    copyClipDstColor           = new Color(0x77,0x00,0x00);
    
    copyClipOverlapSrcColor    = new Color(0xFF,0x00,0x00);
    copyClipOverlapDstColor    = new Color(0x77,0x00,0x00);

    colorTestCross             = new Color(0xF0,0xF0,0xF0);
    

    status = CLEAN;
    

  }
  
  public void paint(Graphics g) {  
    savedColor = g.getColor();
    switch(this.status) {
    case CLEAN:
      performClean(g);
      break;
    case COPY:
      performCopy(g);
      break;
    case COPY_OVERLAP:
      performCopyOverlap(g);
      break;
    case COPY_CLIP:
      performCopyClip(g);
      break;
    case COPY_CLIP_OVERLAP:
      performCopyClipOverlap(g);
      break;
    }
    g.setColor(savedColor);
  }
  
  public static void main(String s[]) {
    AppletFrame f = new AppletFrame();
    CopyArea    a = new CopyArea();
    f.init(a);
    
    f.add("Center", a);  
    f.pack();  
    f.setSize(new Dimension(435, 300));  
    f.show(); 
    
  }

  public void doClean() {
    this.status = CLEAN;
    invalidate();
    repaint();
  }
  public void doCopy() {
    this.status = COPY;
    invalidate();
    repaint();
  }
  public void doCopyOverlap() {
    this.status = COPY_OVERLAP;
    invalidate();  
    repaint(); 
  }
  public void doCopyClip() {
    this.status = COPY_CLIP;
    invalidate();
    repaint();
  }
  public void doCopyClipOverlap() {
    this.status = COPY_CLIP_OVERLAP;
    invalidate();
    repaint();
  }

  
  public void  performClean(Graphics g) {
    if( null != g ) {
      Dimension   dim   = getPreferredSize();
      Color       saved = g.getColor();
      g.clearRect(0,0,dim.width,dim.height);

      //      g.setColor(copyDstColor);
      //      g.fillRect(copyDestination.x,copyDestination.y,copyDestination.width,copyDestination.height);
      g.setColor(copySrcColor);
      g.fillRect(copySource.x,copySource.y,copySource.width,copySource.height);
      g.setColor(copyDstColor);
      g.drawRect(copyDestination.x,copyDestination.y,copyDestination.width,copyDestination.height);
      g.setColor(colorTestCross);
      drawCross(g,copySource.x,copySource.y,copySource.width,copySource.height);

      //     g.setColor(copyOverlapDstColor);
      //      g.fillRect(copyOverlapDestination.x,copyOverlapDestination.y,copyOverlapDestination.width,copyOverlapDestination.height);
      g.setColor(copyOverlapSrcColor);
      g.fillRect(copyOverlapSource.x,copyOverlapSource.y,copyOverlapSource.width,copyOverlapSource.height);
      g.setColor(copyOverlapDstColor);
      g.drawRect(copyOverlapDestination.x,copyOverlapDestination.y,copyOverlapDestination.width,copyOverlapDestination.height);
      g.setColor(colorTestCross);
      drawCross(g,copyOverlapSource.x,copyOverlapSource.y,copyOverlapSource.width,copyOverlapSource.height);

      //      g.setColor(copyClipDstColor);
      //      g.fillRect(copyClipDestination.x,copyClipDestination.y,copyClipDestination.width,copyClipDestination.height);
      g.setColor(copyClipSrcColor);
      g.fillRect(copyClipSource.x,copyClipSource.y,copyClipSource.width,copyClipSource.height);
      g.setColor(copyClipDstColor);
      g.drawRect(copyClipDestination.x,copyClipDestination.y,copyClipDestination.width,copyClipDestination.height);
      g.setColor(colorTestCross);
      drawCross(g,copyClipSource.x,copyClipSource.y,copyClipSource.width,copyClipSource.height);

      //      g.setColor(copyClipOverlapDstColor);
      //      g.fillRect(copyClipOverlapDestination.x,copyClipOverlapDestination.y,copyClipOverlapDestination.width,copyClipOverlapDestination.height);
      g.setColor(copyClipOverlapSrcColor);
      g.fillRect(copyClipOverlapSource.x,copyClipOverlapSource.y,copyClipOverlapSource.width,copyClipOverlapSource.height);
      g.setColor(copyClipOverlapDstColor);
      g.drawRect(copyClipOverlapDestination.x,copyClipOverlapDestination.y,copyClipOverlapDestination.width,copyClipOverlapDestination.height);
      g.setColor(colorTestCross);
      drawCross(g,copyClipOverlapSource.x,copyClipOverlapSource.y,copyClipOverlapSource.width,copyClipOverlapSource.height);

    }
  }
  public void performCopy(Graphics g) {
    performClean(g);

    int dx = copyDestination.x - copySource.x;
    int dy = copyDestination.y - copySource.y;
    g.copyArea(copySource.x,copySource.y,copySource.width,copySource.height,dx,dy);
  }
  public void  performCopyOverlap(Graphics g) {
    performClean(g);
    int dx = copyOverlapDestination.x - copyOverlapSource.x;
    int dy = copyOverlapDestination.y - copyOverlapSource.y;
    g.copyArea(copyOverlapSource.x,copyOverlapSource.y,copyOverlapSource.width,copyOverlapSource.height,dx,dy);
  }
  public void  performCopyClip(Graphics g) {
    performClean(g);
    int dx = copyClipDestination.x - copyClipSource.x;
    int dy = copyClipDestination.y - copyClipSource.y;
    g.copyArea(copyClipSource.x,copyClipSource.y,copyClipDestination.width,copyClipDestination.height,dx,dy);
  }
  public void performCopyClipOverlap(Graphics g) {
    performClean(g);
    int dx = copyClipOverlapDestination.x - copyClipOverlapSource.x;
    int dy = copyClipOverlapDestination.y - copyClipOverlapSource.y;
    g.copyArea(copyClipOverlapSource.x,copyClipOverlapSource.y,copyClipOverlapDestination.width,copyClipOverlapDestination.height,dx,dy);
  }

void drawCross(Graphics g,int x,int y, int dx, int dy) {
  int x1 = x + dx;
  int y1 = y + dy;

  g.drawLine(x,y,x1,y1);
  g.drawLine(x,y1,x1,y);
}

  public Dimension getPreferredSize() {
    return new Dimension(435,300);
  }
  
} 

class AppletFrame extends Frame
{
  CopyArea canvas;

 

  public boolean init(CopyArea ca) {
    canvas = ca;
    return _makeMenu();
  }

  private boolean _makeMenu() {
    MenuBar menuBar = null;
    boolean retCode = false;
    
    if( null != (menuBar = new MenuBar()) ) {
      Menu      menu;
      if( null != (menu = new Menu("File")) ) {
	menu.add("Exit");
	menuBar.add(menu);
	if( null != (menu = new Menu("Action")) ) {
	  menu.add("Clean");
	  menu.add("Copy");
	  menu.add("Copy Overlap");
	  menu.add("Copy Clip");
	  menu.add("Copy Clip Overlap");
	  menuBar.add(menu);
	  setMenuBar(menuBar);
	  retCode = true;
	}
      }
    }
    return retCode;
  }

  public boolean action(Event givenEvent, Object givenObject) {
    if( (null != givenEvent) && (null != givenObject) && (null != this.canvas) ) {
      if( "Exit".equals(givenObject) ) {
	System.exit(0);
	return true;
      }
      if( "Clean".equals(givenObject) ) {
	canvas.doClean();
	return true;
      }
      if( "Copy".equals(givenObject) ) {
	canvas.doCopy();
	return true;
      }
      if( "Copy Overlap".equals(givenObject) ) {
	canvas.doCopyOverlap();
	return true;
      }

      if( "Copy Clip".equals(givenObject) ) {
	canvas.doCopyClip();
	return true;
      }

      if( "Copy Clip Overlap".equals(givenObject) ) {
	canvas.doCopyClipOverlap();
	return true;
      }
    }
    return false;
  }

}





