/*
 * @(#)Test4163881.java	1.2 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.lang.Math;


public class Test4163881 extends Canvas {


    public Test4163881() {
	setBackground(Color.black);
    }


    public void paint(Graphics g) {
	Graphics2D g2 = (Graphics2D) g;
	int w = getSize().width;
	int h = getSize().height;

	g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
    		RenderingHints.VALUE_ANTIALIAS_ON);

        Color colors[] = { Color.red, Color.green, Color.white };
	for (int i = 0; i < 300; i++) {
	    g2.setColor(colors[i%colors.length]);
	    g2.setComposite(AlphaComposite.getInstance(
                         AlphaComposite.SRC_OVER, (float) Math.random()));
	    g2.fill(new Rectangle((int)(w*Math.random()),(int)(h*Math.random()),2,2));
	}

	g2.setComposite(AlphaComposite.SrcOver);

	g2.setColor(Color.red);
	QuadCurve2D curve = new QuadCurve2D.Float(w/8f,h*0.5f,w*0.5f,
					w*0.2f,w-w/8f,h*0.5f);

        String curvedText = "Java 2D Test Product";
	Font font = new Font("Times New Roman", Font.PLAIN, 42);
        FontMetrics fm = g2.getFontMetrics(font);
        g2.setFont(font);

        ComputeLinePts clpOps = new ComputeLinePts();
        double curvedTextLength = 0.0;
        Point2D lengthPoint = new Point2D.Float();
        Point2D theTangent = new Point2D.Float();
	AffineTransform theRotation;

        
        lengthPoint.setLocation(curve.getX1(), curve.getY1());
        for (int i = 0; i < curvedText.length(); i++) 
	{
              theTangent = clpOps.
                  CurveLengthToPointTangent(curvedTextLength, curve);
          
              theRotation = new AffineTransform(theTangent.getX(),
                                                theTangent.getY(),
                                                (-1) * theTangent.getY(),
                                                theTangent.getX(),
                                                0, 0);

              try {
                  theRotation.inverseTransform(lengthPoint, lengthPoint);
              } catch(NoninvertibleTransformException e) {
                  System.out.println("Can't be inverted");
              }

              g2.setTransform(theRotation);
              g2.drawChars(curvedText.toCharArray(), i, 1, 
                        (int)lengthPoint.getX(), (int)lengthPoint.getY());
 
              curvedTextLength += fm.charWidth(curvedText.charAt(i));
              lengthPoint = clpOps.CurveLengthToPoint(curvedTextLength, curve);  
        }
    }

    public static void main(String argv[]) {
	WindowListener l = new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {System.exit(0);}
	};
	Frame f = new Frame("TestProduct - Test4163881");
	f.addWindowListener(l);
	f.add("Center", new Test4163881());
	f.pack();
	f.setSize(new Dimension(500,300));
	f.show();
    }



public class ComputeLinePts {

    double a0, a1, a2; // x(t) = a0 + t * (a1 + t * a2);
    double b0, b1, b2; // y(t) = b0 + t * (b1 + t * b2);
    boolean isLinear;  // true is all three points are colinear
    

    public ComputeLinePts() {
    }
    
    /**

      Function takes a curve that is really a line and makes a line structure
      out of it.  It checks the 3 combinations of the three points to find
      the two furthest apart because it is not necessarily the first and last
      points.

      We assume that the curve was all colinear points
      */
    public Line2D CurveToLine(QuadCurve2D theCurve) {
       
        double x1, y1, x2, y2, x3, y3;
        double d1, d2, d3;
        Line2D theLine = new Line2D.Float();

        x1 = theCurve.getX1();
        y1 = theCurve.getY1();
        x2 = theCurve.getCtrlX();
        y2 = theCurve.getCtrlY();
        x3 = theCurve.getX2();
        y3 = theCurve.getY2();

        d1 = (x2-x1) * (x2-x1) + (y2-y1) * (y2-y1); //first - control
        d2 = (x3-x1) * (x3-x1) + (y3-y1) * (y3-y1); //first - last
        d3 = (x3-x2) * (x3-x2) + (y3-y2) * (y3-y2); //control - last

        if ((d1 > d2) && (d1 > d3)) { //d1 longest
            theLine.setLine(x1, y1, x2, y2);
        } else if ((d2 > d1) && (d2 > d3)) { //d2 longest
            theLine.setLine(x1, y1, x3, y3);
        } else {
            theLine.setLine(x2, y2, x3, y3);
        }

        return theLine;

    }

    /**
      We get the length of the line
      */
    public double GetLineLength(Line2D theLine) {
    
        double a = theLine.getX2() - theLine.getX1();
        double b = theLine.getY2() - theLine.getX1();
        return Math.sqrt(a * a + b * b);
    }

    /**
     */
    public Point2D LineLengthToPoint(double theLength, Line2D theLine) {

        double totalLength = GetLineLength(theLine);
        double t;
        double dx, dy;
        Point2D thePoint = new Point2D.Float();

        if (totalLength != 0.0) {
            t = theLength / totalLength;
        } else {
            t = 0.0;
        }

        dx = theLine.getX2() - theLine.getX1();
        dy = theLine.getY2() - theLine.getY1();

        thePoint.setLocation(theLine.getX1() + (t * dx),
                             theLine.getY1() + (t * dy));
        
        return thePoint;
    }

    /**
      */
    public Point2D LineLengthToPointTangent(double theLength, Line2D theLine) {

        double totalLength = GetLineLength(theLine);
        double t;
        double dx, dy;
        Point2D thePoint = new Point2D.Float();
        Point2D theTangent = new Point2D.Float();

        if (totalLength != 0.0) {
            t = theLength / totalLength;
        } else {
            t = 0.0;
        }

        dx = theLine.getX2() - theLine.getX1();
        dy = theLine.getY2() - theLine.getY1();

        thePoint.setLocation(theLine.getX1() + (t * dx),
                             theLine.getY1() + (t * dy));


        //now normalize the tangent

        if (totalLength != 0.0) {
            theTangent.setLocation(dx/totalLength, dy/totalLength);
        }

        return theTangent;
    }

    /**
      The conditions:

      (x(0), y(0)) = theCurve.getX1[Y1]() = theCurve.getX2[Y2]()
      (x'(0), y'(0)) = theCurve.getCtrlX[Y]() - theCurve.getX1[Y1]()
      (x'(1), y'(1)) = theCurve.getX2[Y2]() - theCurve.getCtrlX[Y]()

      are sufficient to determine the coefficients of the quadratic
      parametrization
      */

    public void ComputeCurveCoefficients(QuadCurve2D theCurve) {
        

        a0 = theCurve.getX1();
        a1 = (theCurve.getCtrlX() - theCurve.getX1()) * 2;
        a2 = theCurve.getX2() - (theCurve.getCtrlX() * 2) + theCurve.getX1();

        b0 = theCurve.getY1();
        b1 = (theCurve.getCtrlY() - theCurve.getY1()) * 2;
        b2 = theCurve.getY2() - (theCurve.getCtrlY() * 2) + theCurve.getY1();

        if (((a2 == 0.0) && (b2 == 0.0)) ||
            ((theCurve.getX1() - theCurve.getX2() + theCurve.getY1()
              - theCurve.getY2()) == 0.0)) {
            isLinear = true;
        } else {
            isLinear = false;
        }

    }


    /**
      That's the integrand sqrt(x'(t)^2 + y'(t)^2) (if you look closely)
      */
    public double LengthFunction(double t) {
        
        double A = 4 * ((a2 * a2) + (b2 * b2));
        double B = 4 * ((a1 * a2) + (b1 * b2));
        double C = (a1 * a1) + (b1 * b1);

        return Math.sqrt(C + t * (B + t * A));
    }

    /**
      After all this is the most appropriate solution:
      Used the closed formula for the length integral!
      (Mathematica did it for me ...)
      */

      public double EvaluateClosedFormula(double x) {

          double A, B, C, SC, SCBX, result;

          A = 4 * ((a2 * a2) + (b2 * b2)); //should always be > 0

          if (A == 0.0)
              {
                  //throw error message
              }
          
          B = 4 * ((a1 * a2) + (b1 * b2)) / A;
          C = ((a1 * a1) + (b1 * b1)) / A;
          
          SC = Math.sqrt(C);
          SCBX = Math.sqrt(C + x * (B + x));
          
          result = 4.0 * x * SCBX - 2.0 * B * (SC - SCBX);
          result = result + (B * B - 4.0 * C) * 
              (Math.log(0.5 * B + SC) - Math.log(0.5 * B + x + SCBX));
          result = Math.sqrt(A) * result / 8.0;
         
          return result;
         
      }

    public double NewLengthToParameterFunction(double s) {
        
        double a = 0.0;
        double b = 1.0;
        double c, d;
        
        // could set error if s is "out of range"

        if (s <= 0.0) {
            return 0.0;
        }

        if (s >= EvaluateClosedFormula(1.0)) {
            return 1.0;
        }
        
        do {
            c = 0.5 * (a +b);
            d = s - EvaluateClosedFormula(c);
            if (d < 0) { // c too far right
                b = c;
            } else {     // c too far left
                a = c; 
            }
        } while (Math.abs(d) > 0.1);

        return c;
    }

    public double GetCurveLength(QuadCurve2D theCurve) {
        
        double result;

        ComputeCurveCoefficients(theCurve);

        if (isLinear) {
           
            Line2D theLine = new Line2D.Float();

            theLine = CurveToLine(theCurve);
            result = GetLineLength(theLine);

        } else {

            result = EvaluateClosedFormula(1.0);

        }

        return result;
    }

    public Point2D EvaluateCurveParametrization(double t) {
        
        Point2D thePoint = new Point2D.Float();

        thePoint.setLocation(a0 + t * (a1 + t * a2),
                             b0 + t * (b1 + t * b2));
        return thePoint;
    }

    public Point2D CurveLengthToPoint(double theLength, QuadCurve2D theCurve) {
        
        double tx, ty, norm;
        double t;
        Point2D thePoint = new Point2D.Float();

        ComputeCurveCoefficients(theCurve);
        
        if (isLinear) {
            
            Line2D theLine = new Line2D.Float();
            theLine = CurveToLine(theCurve);

            thePoint = LineLengthToPoint(theLength, theLine);
        } else {
            
            t = NewLengthToParameterFunction(theLength);
            thePoint = EvaluateCurveParametrization(t);

        }

        return thePoint;
    }

    public Point2D CurveLengthToPointTangent(double theLength,
                                             QuadCurve2D theCurve) {
        
        double tx, ty, norm;
        double t;
        Point2D theTangent = new Point2D.Float();
        Point2D thePoint = new Point2D.Float();

        ComputeCurveCoefficients(theCurve);
        
        if (isLinear) {
            
            Line2D theLine = new Line2D.Float();
            theLine = CurveToLine(theCurve);

            thePoint = LineLengthToPoint(theLength, theLine);
            theTangent = LineLengthToPointTangent(theLength, theLine);
        } else {
            
            t = NewLengthToParameterFunction(theLength);
            EvaluateCurveParametrization(t);

            tx = a1 + t * 2.0 * a2; //that's the derivative
            ty = b1 + t * 2.0 * b2;

            norm = Math.sqrt(tx * tx + ty * ty);

            if (norm > 0) {
                theTangent.setLocation(tx/norm, ty/norm);
            }

        }
        return theTangent;
    }

    public Line2D DetermineSubLine(Line2D theLine, double a, double b) {
        
        Point2D temp1, temp2;
        Line2D subLine = new Line2D.Float();
        
        temp1 = LineLengthToPoint(a, theLine);
        temp2 = LineLengthToPoint(b, theLine);

        subLine.setLine(temp1, temp2);
        return subLine;
    }
	
    public QuadCurve2D DetermineSubCurve(QuadCurve2D theCurve, double a, 
                                         double b) {
        double t0, t1;
        Point2D midPoint;
        QuadCurve2D subCurve = new QuadCurve2D.Float();

        ComputeCurveCoefficients(theCurve);

        if (isLinear) {
            
            //treat the curve as a line if it was linear
            Line2D theLine = new Line2D.Float();
            Line2D subLine = new Line2D.Float();

            theLine = CurveToLine(theCurve);

            subLine = DetermineSubLine(theLine, a, b);

            subCurve.setCurve(subLine.getX1(), subLine.getY1(),
                              (subLine.getX1() + subLine.getX2())/2,
                              (subLine.getY1() + subLine.getY2())/2,
                              subLine.getX2(), subLine.getY2());
        } else {
            
            //compute the sub curve

            Point2D temp1 = new Point2D.Float();
            Point2D temp2 = new Point2D.Float();

            t0 = NewLengthToParameterFunction(a);
            t1 = NewLengthToParameterFunction(b);

            temp1 = EvaluateCurveParametrization(t0);
            temp2 = EvaluateCurveParametrization(t1);
            midPoint = EvaluateCurveParametrization(0.5 * (t0 + t1));

            subCurve.setCurve(temp1.getX(), temp1.getY(),
                     (((midPoint.getX()*2) - (temp1.getX() + temp2.getX())/2)),
                     (((midPoint.getY()*2) - (temp1.getY() + temp2.getY())/2)),
                              temp2.getX(), temp2.getY());
        }
     
        return subCurve;
    }

} // ComputeLinePts

} // Test4163881
