/*
 * @(#)Offscr2Offscr2D.java	1.7 98/06/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Color;
import java.awt.Shape;

import java.awt.image.BufferedImage;


import java.applet.Applet;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Event;
import java.lang.Math;



public class Offscr2Offscr2D extends Applet implements Runnable
{
    private static int DRAWSIZE = 150;
	private static int INTERMEDIA = 5;


    Thread myThread;

    private void ReportHeader() {
        System.out.println(">=================================================");
        System.out.println("SEQUENCE STARTED");
        System.out.println(">=================================================");
    }

    private void ReportFooter() {
        System.out.println(">=================================================");
        System.out.println("SEQUENCE FINISHED");
        System.out.println(">=================================================");
    }

    private void PerformTest(Graphics g,TestObject testObject) {
        boolean ok = true;

        System.out.println(">=================================================");

        try
        {
            System.out.println(testObject.getDescription());
		    testObject.initialize(g,DRAWSIZE,INTERMEDIA);
		    testObject.performTest();
        } catch (Exception e)   {
                                    String msg = e.getMessage();
                                    e.printStackTrace();
                                    ok = false;
                                } finally   {
                                                if( true == ok )
                                                {
                                                    System.out.println("passed");
                                                } else  {
                                                            System.out.println("failed");
                                                        }
                                            }
        System.out.println(">=================================================");
    }

    private void BasicTests(Graphics g) {
        ReportHeader();

        PerformTest(g,new TestCurrent_INT_RGB_Current());
        PerformTest(g,new TestCurrent_INT_ARGB_Current());
        PerformTest(g,new TestCurrent_INT_ARGB_PRE_Current());
        PerformTest(g,new TestCurrent_BINARY_Current());
        PerformTest(g,new TestCurrent_BYTE_GRAY_Current());
        PerformTest(g,new TestCurrent_SHORT_565_RGB_Current());
        PerformTest(g,new TestCurrent_SHORT_555_RGB_Current());
        PerformTest(g,new TestCurrent_4BYTE_ABGR_Current());
        PerformTest(g,new TestCurrent_4BYTE_ABGR_PRE_Current());
        PerformTest(g,new TestCurrent_3BYTE_BGR_Current());
        PerformTest(g,new TestCurrent_INT_BGR_Current());

        ReportFooter();
 }


    private void TestBlitters(Graphics g) {
        BasicTests(g);
    }

    public void start() {
        if (myThread == null) {
            myThread = new Thread(this);
        }
	myThread.setPriority(Thread.MAX_PRIORITY);
        myThread.start();
    }

    public void stop() {
        if (myThread != null) {
            myThread.stop();
            myThread = null;
        }
    }

    public void run() {
		Graphics graphics = getGraphics();

		TestBlitters(graphics);
    }
}



abstract class TestObject
{

	Graphics		graphics;
	BufferedImage	sourceOffscreen;
	BufferedImage	destinationOffscreen;
	Rectangle		originalArea;
	Rectangle		sourceOffscreenArea;
	Rectangle		destinationOffscreenArea;
	Rectangle		offscreenStretchCopyArea;

	Color			colorRed;
	Color			colorGreen;
	Color			colorBlue;
	Color			colorBlack;
	Color			colorWhite;

	Color			savedColor;
	Color			colorYellow;
//------------- begin group
//  Initializes internal structure
	public void initialize(Graphics givenGraphics,int drawSize,int gap)
	{
		originalArea		        = new Rectangle(0,0,drawSize,drawSize);
		sourceOffscreenArea		    = new Rectangle(drawSize + gap,0,drawSize,drawSize);
		destinationOffscreenArea    = new Rectangle(2*(drawSize+gap),0,drawSize,drawSize);
		offscreenStretchCopyArea    = new Rectangle(3*(drawSize+gap),0,drawSize,drawSize);


		graphics		= givenGraphics;
		sourceOffscreen	= makeSourceOffscreen();
        destinationOffscreen = null;

		colorRed		= new Color(0xFF,0x00,0x00);
		colorGreen		= new Color(0x00,0xFF,0x00);
		colorBlue		= new Color(0x00,0x00,0xFF);
		colorWhite		= new Color(0xFF,0xFF,0xFF);
		colorBlack		= new Color(0x00,0x00,0x00);

        colorYellow     = new Color(0xFF,0xFF,0x00);
	}
//------------- end group

    public void waitLoop(int delay)
    {
        long    start = System.currentTimeMillis();

        while( System.currentTimeMillis() - start < delay)
        {
        }
    }

//------------- begin group
//  Draw primitives
	void drawBar(Graphics g,int x,int y,int h, int v,Color c)
	{
		savedColor = g.getColor();

		g.setColor(c);

		g.fillRect(x,y,h,v);

		g.setColor(savedColor);
	}
	void drawRect(Graphics g,int x,int y,int h, int v,Color c)
	{
		savedColor = g.getColor();

		g.setColor(c);

		g.drawRect(x,y,h,v);

		g.setColor(savedColor);
	}
	void drawRing(Graphics g,int x,int y,int h, int v,Color c)
	{
		savedColor = g.getColor();

		g.setColor(c);

		g.drawOval(x,y,h,v);

		g.setColor(savedColor);
	}
	void drawGrid(Graphics g,int x,int y,int h, int v,int d,Color c)
	{
		savedColor = g.getColor();

		g.setColor(c);

		{
			int		i;

			for(i=y;i<y+v;i+=d)
			{
				g.drawLine(x,i,x+h,i);
			}
			for(i=x;i<x+v;i+=d)
			{
				g.drawLine(i,y,i,y+v);
			}
		}


		g.setColor(savedColor);
	}
	void typeLine(Graphics g,int x,int y,String s,Color c)
	{

        drawBar(g,x,y - 10,150,15,colorWhite);

		savedColor = g.getColor();

		g.setColor(c);


        g.drawString(s,x,y);

		g.setColor(savedColor);
	}
//------------- end group

//------------- begin group
//  Make patterns on abstract locations
	void makeOriginal(Graphics g,int x,int y,int width,int height)
	{

		x+=30;
		y+=30;


		drawBar(g,x,y,40,40,colorRed);
		drawBar(g,x+50,y,40,40,colorGreen);
		drawBar(g,x+25,y+50,40,40,colorBlue);

		drawGrid(g,x + 20,y + 20,51,51,5,colorBlack);

		drawRing(g,x-30,y-30,width,height,colorBlack);

	}

//------------- end group

//------------- begin group
//  Draw area content
    void drawOriginalOnItsArea()
    {
		makeOriginal(graphics,originalArea.x,originalArea.y,originalArea.width,originalArea.height);
    }

    void drawOriginalOnSourceOffscreen()
    {
        Graphics gOff = sourceOffscreen.getGraphics();
        drawBar(gOff,1,1,originalArea.width,originalArea.height,colorWhite);
        makeOriginal(gOff,0,0,originalArea.width,originalArea.height);
    }
//------------- end group

//------------- begin group
//  Prepare areas
    public void prepareOriginalArea()
	{

		graphics.clearRect( originalArea.x,originalArea.y,
                            originalArea.width,originalArea.height);

		drawRect(   graphics,originalArea.x,originalArea.y,
                    originalArea.width,originalArea.height,colorBlack);

        typeLine( graphics,originalArea.x,originalArea.y + originalArea.height + 15,"Image",colorBlack);
        typeLine( graphics,originalArea.x,originalArea.y + originalArea.height + 35,getDescription(),colorBlack);
	}


	public void prepareSourceOffscreenArea()
	{
		graphics.clearRect(	sourceOffscreenArea.x,sourceOffscreenArea.y,
							sourceOffscreenArea.width,sourceOffscreenArea.height);


		drawRect(   graphics,sourceOffscreenArea.x,sourceOffscreenArea.y,
                    sourceOffscreenArea.width,sourceOffscreenArea.height,colorBlack);

        typeLine(   graphics,sourceOffscreenArea.x,sourceOffscreenArea.y + sourceOffscreenArea.height + 15,
                    "Source copy",colorBlack);
	}
	public void prepareDestinationOffscreenArea()
	{
		graphics.clearRect(	destinationOffscreenArea.x,destinationOffscreenArea.y,
							destinationOffscreenArea.width,destinationOffscreenArea.height);


		drawRect(   graphics,destinationOffscreenArea.x,destinationOffscreenArea.y,
                    destinationOffscreenArea.width,destinationOffscreenArea.height,colorBlack);

        typeLine(   graphics,destinationOffscreenArea.x,destinationOffscreenArea.y + destinationOffscreenArea.height + 15,
                    "Destination copy",colorBlack);
	}
//------------- end group

//------------- begin group
//  Copy fragments
    public void copySourceOffscreenToItsCopyArea()
    {
        graphics.drawImage(sourceOffscreen,sourceOffscreenArea.x,sourceOffscreenArea.y,null);
    }
    public void copySourceOffscreenToDestinationOne()
    {
        Graphics    g = destinationOffscreen.getGraphics();
        g.drawImage(sourceOffscreen,0,0,null);
    }
    public void copyDestinationOffscreenToItsCopyArea()
    {
        graphics.drawImage(destinationOffscreen,destinationOffscreenArea.x,destinationOffscreenArea.y,null);
    }

//------------- end group


    public void performTest()
	{

		prepareDestinationOffscreenArea();

		prepareSourceOffscreenArea();

		prepareOriginalArea();

        drawOriginalOnItsArea();

        drawOriginalOnSourceOffscreen();

		copySourceOffscreenToItsCopyArea();


        performLoop();


        waitLoop(2000);

	}

    void performLoop()
    {
        int     count;
        boolean ok, supported = true;
        String  name;
        for(count = 0;count < getDestNum(); count++ )
        {
            name = getNumType(count);
            ok = true;

            System.out.println(">   ---------------------");
            System.out.println(">   TO: "+name);


            try{
                supported = testGivenFormat(count,name);
            } catch (Exception e) {
                ok = false;
            } finally {
                if( true == supported )
                {
                    if( true == ok ) {
                        System.out.println("    passed");
                    } else {
                        System.out.println("    failed");
                    }
                } else {
                    System.out.println("    unsupported");
                }
            }
            System.out.println(">   ---------------------");
            waitLoop(5000);
        }
    }
    boolean testGivenFormat(int num,String name)
    {
        boolean          ret = false;
        destinationOffscreen = makeDestinationOffscreen(num);

        if( null != destinationOffscreen )
        {
            copySourceOffscreenToDestinationOne();
            typeLine(   graphics,destinationOffscreenArea.x + destinationOffscreenArea.width - 150,
                        destinationOffscreenArea.y + destinationOffscreenArea.height + 35,
                        name,colorBlack);
            copyDestinationOffscreenToItsCopyArea();
            ret = true;
        } else {
            typeLine(   graphics,destinationOffscreenArea.x + destinationOffscreenArea.width - 150,
                        destinationOffscreenArea.y + destinationOffscreenArea.height + 35,
                        name+" U/S",colorBlack);
        }

        return ret;
    }
    public int getDestNum()
    {
        return 12;
    }
    public BufferedImage makeDestinationOffscreen(int num)
    {
        switch(num)
        {
            case 0:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_4BYTE_ABGR);
            case 1:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_4BYTE_ABGR_PRE);
            case 2:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_USHORT_565_RGB);
            case 3:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_USHORT_555_RGB);
            case 4:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_GRAY);
            case 5:
//		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BINARY);
		    return null;    //  UNSUPPORTED
            case 6:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_INDEXED);
            case 7:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_3BYTE_BGR);
            case 8:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_BGR);
            case 9:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_ARGB_PRE);
            case 10:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_ARGB);
            case 11:
		    return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_RGB);
        }
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_RGB);
    }

    public String getNumType(int num)
    {
        switch(num)
        {
            case 0:
		    return " TYPE_4BYTE_ABGR";
            case 1:
		    return " TYPE_4BYTE_ABGR_PRE";
            case 2:
		    return " TYPE_USHORT_565_RGB";
            case 3:
		    return " TYPE_USHORT_555_RGB";
            case 4:
		    return " TYPE_BYTE_GRAY";
            case 5:
		    return " TYPE_BYTE_BINARY";
            case 6:
		    return " TYPE_BYTE_INDEXED";
            case 7:
		    return " TYPE_3BYTE_BGR";
            case 8:
		    return " TYPE_INT_BGR";
            case 9:
		    return " TYPE_INT_ARGB_PRE";
            case 10:
		    return " TYPE_INT_ARGB";
            case 11:
		    return " TYPE_INT_RGB";
        }
		return " TYPE_INT_RGB";
    }
	abstract public BufferedImage makeSourceOffscreen();
	abstract public String getDescription();
}

class TestCurrent_INT_RGB_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_RGB);
	}
    public String getDescription()
    {
        return "TYPE_INT_RGB";
    }
}
class TestCurrent_INT_ARGB_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_ARGB);
	}
    public String getDescription()
    {
        return "TYPE_INT_ARGB";
    }
}
class TestCurrent_INT_ARGB_PRE_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_ARGB_PRE);
	}
    public String getDescription()
    {
        return "TYPE_INT_ARGB_PRE";
    }
}
class TestCurrent_INT_BGR_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_INT_BGR);
	}
    public String getDescription()
    {
        return "TYPE_INT_BGR";
    }
}
class TestCurrent_3BYTE_BGR_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_3BYTE_BGR);
	}
    public String getDescription()
    {
        return "TYPE_3BYTE_BGR";
    }
}
class TestCurrent_4BYTE_ABGR_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_4BYTE_ABGR);
	}
    public String getDescription()
    {
        return "TYPE_4BYTE_ABGR";
    }
}
class TestCurrent_4BYTE_ABGR_PRE_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_4BYTE_ABGR_PRE);
	}
    public String getDescription()
    {
        return "TYPE_4BYTE_ABGR_PRE";
    }
}
class TestCurrent_SHORT_565_RGB_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_USHORT_565_RGB);
	}
    public String getDescription()
    {
        return "TYPE_USHORT_565_RGB";
    }
}
class TestCurrent_SHORT_555_RGB_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_USHORT_555_RGB);
	}
    public String getDescription()
    {
        return "TYPE_USHORT_555_RGB";
    }
}
class TestCurrent_BYTE_GRAY_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_GRAY);
	}
    public String getDescription()
    {
        return "TYPE_BYTE_GRAY";
    }
}
class TestCurrent_BINARY_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_BINARY);
	}
    public String getDescription()
    {
        return "TYPE_BINARY";
    }
}
class TestCurrent_BYTE_INDEXED_Current extends TestObject
{

	public BufferedImage makeSourceOffscreen()
	{
		return new BufferedImage(originalArea.width,originalArea.height,BufferedImage.TYPE_BYTE_INDEXED);
	}
    public String getDescription()
    {
        return "TYPE_BYTE_INDEXED";
    }
}

