/*
 * @(#)Pipeline.java	1.5 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.applet.Applet;
import java.awt.*;
import java.awt.image.*;
import java.util.Vector;


public class Pipeline {
    Vector pipeline;
    int psize = 0;
    Applet applet;
    Canvas canvas;
    BufferedImage firstImage;
    static String[] opNames = {"Convolve", "Window/Level", "Affine Transform"};
//                               "Rescale", "Threshold"};

    public Pipeline(Applet applet, Canvas canvas, BufferedImage inImage) {
        this.applet = applet;
        this.canvas = canvas;
        firstImage  = inImage;
        pipeline = new Vector(10);
    }

    /**
     * Return the list of available pipes
     */
    static String[] getPipeNames() {
        return opNames;
    }

    public Container add(int i) {
        Pipe pipe = null;
        
        switch (i) {
        case 0:  // Convolve
            pipe = new ConvolvePipe(this, applet, getImage());
            break;
        case 1:  // Lookup
            pipe = new LookupPipe(this, applet, getImage());
            break;
        case 2:  // Affine Transform
            pipe = new AffineTransformPipe(this, applet, getImage());
            break;
//        case 3:  // Rescale
//            pipe = new RescalePipe(this, applet, getImage());
//            break;
//        case 4:  // Threshold
//            pipe = new ThresholdPipe(this, applet, getImage());
//            break;
        }
        if (pipe == null) {
            throw new IllegalArgumentException("No such pipe index: "+i);
        }

        pipeline.addElement(pipe);
        psize++;
        Container c = pipe.addGui();
        pipe.process();
        
        return c;
    }
    
    public Container add(String pipeName) {
        Pipe pipe = null;
        for (int i=0; i < opNames.length; i++) {
            if (opNames[i].equalsIgnoreCase(pipeName)) {
                return add(i);
            }
        }

        throw new IllegalArgumentException("No such pipe: "+pipeName);
    }

    /**
     * Returns the last image in the pipeline
     */
    public BufferedImage getImage() {
        if (psize > 0) {
            Pipe p = (Pipe) pipeline.lastElement();
            return p.getImage();
        }
        else {
            return firstImage;
        }
    }

    public void trigger(Pipe p) {
        int index = pipeline.indexOf(p);
        if (index < 0) {
            System.out.println("No such pipe "+p+" in pipeline");
        }
        else {
            // Process the rest of the pipes
            for (index++; index < pipeline.size(); index++) {
                p = (Pipe) pipeline.elementAt(index);
                p.process();
            }
        }
        canvas.repaint();
    }
        
}
