/*
 * @(#)PolyPixelTest.java	1.6 98/09/15
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.applet.Applet;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Canvas;
import java.awt.Checkbox;
import java.awt.Panel;
import java.awt.Frame;
import java.awt.BorderLayout;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

public class PolyPixelTest extends Applet
    implements ItemListener
{
    PolyPixelCanvas canvas;
    Checkbox antialias;
    Checkbox stroke;

    public PolyPixelTest() {
	setLayout(new BorderLayout());

	canvas = new PolyPixelCanvas();
	add("Center", canvas);

	Panel p = new Panel();
	antialias = new Checkbox("Antialias", false);
	antialias.addItemListener(this);
	p.add(antialias);
	stroke = new Checkbox("Stroke", false);
	stroke.addItemListener(this);
	p.add(stroke);
	add("North", p);
    }

    public void itemStateChanged(ItemEvent e) {
	canvas.antialias = antialias.getState();
	canvas.stroke = stroke.getState();
	canvas.repaint();
    }

    public static void main(String argv[]) {
	Frame f = new Frame();
	f.add(new PolyPixelTest());
	f.addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {System.exit(0);}
	    public void windowClosed(WindowEvent e) {System.exit(0);}
	});
	f.pack();
	f.show();
    }
}

class PolyPixelCanvas extends Canvas
    implements MouseListener, MouseMotionListener
{
    Point vertex1 = new Point(15, 15);
    Point vertex2 = new Point(245, 55);
    Point vertex3 = new Point(155, 205);
    Point dragpoint;
    boolean antialias;
    boolean stroke;

    int factor = 10;

    public PolyPixelCanvas() {
	addMouseListener(this);
    }

    public int distSq(Point p1, Point p2) {
	int x = p1.x - p2.x;
	int y = p1.y - p2.y;
	return x * x + y * y;
    }

    public void mouseClicked(MouseEvent e) {}

    public void mousePressed(MouseEvent e) {
	Point p = e.getPoint();
	int dist1 = distSq(vertex1, p);
	int dist2 = distSq(vertex2, p);
	int dist3 = distSq(vertex3, p);
	if (dist1 < dist2) {
	    if (dist1 < dist3) {
		dragpoint = vertex1;
	    } else {
		dragpoint = vertex3;
	    }
	} else {
	    if (dist2 < dist3) {
		dragpoint = vertex2;
	    } else {
		dragpoint = vertex3;
	    }
	}
	addMouseMotionListener(this);
    }

    public void mouseDragged(MouseEvent e) {
	dragpoint.setLocation(e.getPoint());
	repaint();
    }

    public void mouseMoved(MouseEvent e) {}

    public void mouseReleased(MouseEvent e) {
	mouseDragged(e);
	removeMouseMotionListener(this);
	dragpoint = null;
    }

    public void mouseEntered(MouseEvent e) {}

    public void mouseExited(MouseEvent e) {}

    Image img;

    public Dimension getPreferredSize() {
	return new Dimension(300, 300);
    }

    public void paint(Graphics g) {
	Dimension d = getSize();
	int imgw = (d.width + (factor - 1)) / factor;
	int imgh = (d.height + (factor - 1)) / factor;
	if (img == null ||
	    imgw != img.getWidth(this) ||
	    imgh != img.getHeight(this))
	{
	    img = createImage(imgw, imgh);
	}
	Graphics g2 = img.getGraphics();
	g2.clearRect(0, 0, imgw, imgh);

	Polygon pg = new Polygon();
	pg.addPoint(vertex1.x, vertex1.y);
	pg.addPoint(vertex2.x, vertex2.y);
	pg.addPoint(vertex3.x, vertex3.y);

	g2.setColor(Color.blue);
	Polygon pg2;
	try {
	    Graphics2D g2d = (Graphics2D) g2;
	    g2d.scale(0.1, 0.1);
	    g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				  (antialias
				   ? RenderingHints.VALUE_ANTIALIAS_ON
				   : RenderingHints.VALUE_ANTIALIAS_OFF));
	    g2d.setRenderingHint(RenderingHints.KEY_RENDERING,
				  (antialias
				   ? RenderingHints.VALUE_RENDER_QUALITY
				   : RenderingHints.VALUE_RENDER_SPEED));
	    if (stroke && antialias) {
		g2d.setStroke(new BasicStroke(10.0f,
					      BasicStroke.CAP_ROUND,
					      BasicStroke.JOIN_ROUND));
	    }
	    pg2 = pg;
	} catch (NoClassDefFoundError e) {
	    pg2 = new Polygon();
	    pg2.addPoint((int) Math.floor(vertex1.x / 10.0),
			 (int) Math.floor(vertex1.y / 10.0));
	    pg2.addPoint((int) Math.floor(vertex2.x / 10.0),
			 (int) Math.floor(vertex2.y / 10.0));
	    pg2.addPoint((int) Math.floor(vertex3.x / 10.0),
			 (int) Math.floor(vertex3.y / 10.0));
	}
	if (stroke) {
	    g2.drawPolygon(pg2);
	} else {
	    g2.fillPolygon(pg2);
	}
	g2.dispose();

	g.drawImage(img,
		    0, 0, imgw*factor, imgh*factor,
		    0, 0, imgw, imgh,
		    null);
	g.setColor(Color.red);
	for (int i = 0; i <= d.width; i += factor) {
	    g.drawLine(i, 0, i, d.height);
	}
	for (int i = 0; i <= d.height; i += factor) {
	    g.drawLine(0, i, d.width, i);
	}
	g.setColor(Color.yellow);
	g.drawPolygon(pg);
    }
}
