/*
 * @(#)ShapeDraw.java	1.8 98/05/03
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
    
import java.applet.Applet;
import java.awt.Choice;
import java.awt.Label;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Frame;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.geom.Arc2D;
import java.awt.geom.RectangularShape;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;
import java.util.Vector;
import java.util.Enumeration;

public class ShapeDraw extends Applet {
    public ShapeDraw() {
	setLayout(new BorderLayout());
	ShapeDrawControls sdctl = new ShapeDrawControls();
	add("South", sdctl);
	ShapeDrawCanvas sdcv = new ShapeDrawCanvas(sdctl);
	add("Center", sdcv);
    }

    public static void main(String argv[]) {
	Frame f = new Frame();
	f.add(new ShapeDraw());
	f.pack();
	f.show();
    }
}

class ShapeDrawControls extends Panel implements ItemListener {
    Choice createchoice;
    Choice reshapechoice;

    public ShapeDrawControls() {
	setLayout(new FlowLayout());
	add(new Label("New Shape type:"));
	createchoice = new Choice();
	createchoice.add("Rectangle");
	createchoice.add("RoundRectangle");
	createchoice.add("Ellipse");
	createchoice.add("Arc");
	createchoice.addItemListener(this);
	add(createchoice);
	setCreator("Rectangle");
	add(new Label("Reshape method:"));
	reshapechoice = new Choice();
	reshapechoice.add("Diagonal");
	reshapechoice.add("Center");
	reshapechoice.addItemListener(this);
	add(reshapechoice);
	setEditor("Diagonal");
    }

    Class shapeCreator;
    Class shapeEditor;

    void setCreator(String s) {
	try {
	    shapeCreator = Class.forName("java.awt.geom."+s+"2D$Float");
	} catch (ClassNotFoundException exc) {
	    exc.printStackTrace();
	}
    }

    void setEditor(String s) {
	try {
	    shapeEditor = Class.forName(s+"Reshaper");
	} catch (ClassNotFoundException exc) {
	    exc.printStackTrace();
	}
    }

    /**
     * Invoked when the state of an item has been changed.
     */    
    public void itemStateChanged(ItemEvent e) {
	if (e.getStateChange() == ItemEvent.SELECTED) {
	    if (e.getSource() == createchoice) {
		setCreator(e.getItem().toString());
	    } else if (e.getSource() == reshapechoice) {
		setEditor(e.getItem().toString());
	    }
	}
    }

    public ShapeReshaper create() {
	RectangularShape rs;
	try {
	    rs = (RectangularShape) shapeCreator.newInstance();
	} catch (Exception e) {
	    e.printStackTrace();
	    return null;
	}
	return create(rs);
    }

    public ShapeReshaper create(RectangularShape rs) {
	ShapeReshaper sr;
	try {
	    sr = (ShapeReshaper) shapeEditor.newInstance();
	} catch (Exception e) {
	    e.printStackTrace();
	    return null;
	}
	sr.setShape(rs);
	return sr;
    }
}

abstract class ShapeReshaper {
    RectangularShape shape;

    public void setShape(RectangularShape sh) {
	this.shape = sh;
    }

    public Shape getShape() {
	return shape;
    }

    public void checkArcs() {
	if (shape instanceof RoundRectangle2D) {
	    RoundRectangle2D rrect = (RoundRectangle2D) shape;
	    double x = rrect.getX();
	    double y = rrect.getY();
	    double w = rrect.getWidth();
	    double h = rrect.getHeight();
	    rrect.setRoundRect(x, y, w, h, w / 4.0, h / 4.0);
	} else if (shape instanceof Arc2D) {
	    Arc2D arc = (Arc2D) shape;
	    arc.setAngleStart(45);
	    arc.setAngleExtent(270);
	    arc.setArcType(Arc2D.PIE);
	}
    }

    public abstract void reshape(Point2D pt1, Point2D pt2);
}

class DiagonalReshaper extends ShapeReshaper {
    public void reshape(Point2D pt1, Point2D pt2) {
	shape.setFrameFromDiagonal(pt1, pt2);
	checkArcs();
    }
}

class CenterReshaper extends ShapeReshaper {
    public void reshape(Point2D pt1, Point2D pt2) {
	shape.setFrameFromCenter(pt1, pt2);
	checkArcs();
    }
}

class ShapeDrawCanvas extends Canvas
    implements MouseListener, MouseMotionListener
{
    Vector shapes = new Vector();
    ShapeDrawControls controls;
    ShapeReshaper reshaper;
    Point2D origin;
    RectangularShape selected;
    boolean entirely;
    RectangularShape dragged;

    public ShapeDrawCanvas(ShapeDrawControls sdc) {
	this.controls = sdc;
	addMouseListener(this);
	addMouseMotionListener(this);
    }

    public Dimension getPreferredSize() {
	return new Dimension(500, 500);
    }

    public void mouseClicked(MouseEvent e) {}

    public void mousePressed(MouseEvent e) {
	Point2D p = e.getPoint();
	origin = p;
	if (selected != null) {
	    dragged = selected;
	    selected = null;
	    mouseDragged(e);
	} else {
	    reshaper = controls.create();
	    shapes.addElement(reshaper.getShape());
	    reshaper.reshape(p, p);
	}
	repaint();
    }

    public void mouseReleased(MouseEvent e) {
	mouseDragged(e);
	reshaper = null;
	dragged = null;
	origin = null;
	mouseMoved(e);
    }

    public void mouseEntered(MouseEvent e) {
	mouseMoved(e);
    }

    public void mouseExited(MouseEvent e) {
	if (reshaper != null || dragged != null) {
	    mouseDragged(e);
	}
	reshaper = null;
	dragged = null;
	origin = null;
	repaint();
    }

    public void mouseDragged(MouseEvent e) {
	if (dragged != null) {
	    Point2D p = e.getPoint();
	    double x = dragged.getX() + p.getX() - origin.getX();
	    double y = dragged.getY() + p.getY() - origin.getY();
	    dragged.setFrame(x, y, dragged.getWidth(), dragged.getHeight());
	    origin = p;
	    Enumeration enum = shapes.elements();
	    Rectangle2D bounds = dragged.getBounds2D();
	    selected = null;
	    while (enum.hasMoreElements()) {
		RectangularShape s = (RectangularShape) enum.nextElement();
		if (s != dragged && s.intersects(bounds)) {
		    selected = s;
		}
	    }
	    if (selected != null) {
		entirely = selected.contains(bounds);
	    }
	    repaint();
	} else if (reshaper != null) {
	    reshaper.reshape(origin, e.getPoint());
	    repaint();
	}
    }

    public void mouseMoved(MouseEvent e) {
	Enumeration enum = shapes.elements();
	Point2D p = e.getPoint();
	RectangularShape sel = null;
	while (enum.hasMoreElements()) {
	    RectangularShape s = (RectangularShape) enum.nextElement();
	    if (s.contains(p)) {
		sel = s;
	    }
	}
	if (selected != sel) {
	    repaint();
	}
	selected = sel;
	if (sel == null) {
	    return;
	}
	Rectangle2D rect = new Rectangle2D.Float((float) p.getX(),
						 (float) p.getY(),
						 1.0f, 1.0f);
	boolean all = sel.contains(rect);
	if (all != entirely) {
	    repaint();
	}
	entirely = all;
    }

    BufferedImage buf;

    public void paint(Graphics g) {
	if (buf != null) {
	    Graphics2D g2 = (Graphics2D) g;
	    g2.setComposite(AlphaComposite.Src);
	    g2.drawImage(buf, null, 0, 0);
	}
    }

    static BasicStroke solid =
	new BasicStroke(4.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
    static AlphaComposite highlight =
	AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.75f);

    public void update(Graphics g) {
	Dimension size = getSize();
	if (buf == null ||
	    buf.getWidth() != size.width ||
	    buf.getHeight() != size.height)
	{
	    buf = (BufferedImage) createImage(size.width, size.height);
	}
	Graphics2D g2d = buf.createGraphics();
	g2d.setColor(getBackground());
	g2d.fillRect(0, 0, size.width, size.height);
	g2d.setStroke(solid);
	Enumeration enum = shapes.elements();
	while (enum.hasMoreElements()) {
	    Shape s = (Shape) enum.nextElement();
	    g2d.setColor(Color.black);
	    g2d.draw(s);
	    g2d.setColor(s == dragged
			 ? Color.red
			 : (s == selected
			    ? (entirely ? Color.green : Color.cyan)
			    : Color.blue));
	    g2d.fill(s);
	}
	if (dragged != null) {
	    g2d.setColor(Color.gray);
	    g2d.setComposite(highlight);
	    g2d.fill(dragged.getBounds2D());
	}
	g2d.dispose();
	Graphics2D g2 = (Graphics2D) g;
	g2.setComposite(AlphaComposite.Src);
	g2.drawImage(buf, null, 0, 0);
    }
}
