/*
 * @(#)TextureAnim.java	1.2 98/08/20
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
    
import java.awt.Label;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Frame;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.TexturePaint;
import java.awt.GradientPaint;
import java.awt.RenderingHints;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;

public class TextureAnim extends Canvas implements Runnable {
    public static final int TEXTURE = 0;
    public static final int GRADIENT = 1;
    public static final int SOLIDFILL = 2;

    BufferedImage img;

    BufferedImage bimg;
    Rectangle tilerect;

    GradientPaint gradient;
    TexturePaint texture;
    boolean newtexture;

    Thread animator;

    boolean bouncesize = false;
    boolean bouncerect = true;
    boolean rotate = false;
    boolean shearx = false;
    boolean sheary = false;
    boolean drawoval = false;
    boolean showanchor = true;
    boolean quality = false;
    boolean dogray = false;
    int texturetype = TEXTURE;

    public TextureAnim() {
    }

    public Dimension getPreferredSize() {
	return new Dimension(300, 300);
    }

    public void start() {
	Thread t = new Thread(this);
	t.setPriority(Thread.MIN_PRIORITY);
	animator = t;
	t.start();
    }

    public void stop() {
	animator = null;
    }

    public void setSize(int size) {
	if (bimg == null || size != bimg.getWidth()) {
	    makeImage(size);
	}
    }

    public void doGray(boolean dogray) {
	if (this.dogray == dogray) {
	    return;
	}
	this.dogray = dogray;
	if (bimg != null) {
	    makeImage(bimg.getWidth());
	}
    }

    public void makeImage(int size) {
	if (dogray) {
	    bimg = new BufferedImage(size, size, bimg.TYPE_BYTE_GRAY);
	} else {
	    bimg = (BufferedImage) createImage(size, size);
	    int type = bimg.getType();
	    if (type != bimg.TYPE_INT_RGB &&
		type != bimg.TYPE_INT_BGR)
	    {
		bimg = new BufferedImage(size, size, bimg.TYPE_INT_BGR);
	    }
	}

	Graphics2D g2d = bimg.createGraphics();
	g2d.setColor(Color.white);
	g2d.fillRect(0, 0, size, size);
	for (int j = 0; j < size; j++) {
	    float red = j / (float) size;
	    for (int i = 0; i < size; i++) {
		float green = i / (float) size;
		g2d.setColor(new Color(1.0f - red, 1.0f - green, 0.0f, 1.0f));
		g2d.drawLine(i, j, i, j);
	    }
	}
	newtexture = true;
    }

    public void run() {
	Thread me = Thread.currentThread();
	if (bimg == null) {
	    makeImage(100);
	}
	int tilesize = bimg.getWidth();
	Dimension dims = getSize();
	AnimVal w = new AnimVal(0, 200, 3, 10, tilesize);
	AnimVal h = new AnimVal(0, 200, 3, 10, tilesize);
	AnimVal x = new AnimVal(-dims.width/4, dims.width/4 - w.getInt(),
				3, 10, 0);
	AnimVal y = new AnimVal(-dims.height/4, dims.height/4 - h.getInt(),
				3, 10, 0);
	AnimVal rot = new AnimVal(-360, 360, 5, 15, 0);
	AnimVal shx = new AnimVal(-50, 50, 3, 10, 0);
	AnimVal shy = new AnimVal(-50, 50, 3, 10, 0);
	tilerect = new Rectangle(x.getInt(), y.getInt(),
				 w.getInt(), h.getInt());
	texture = new TexturePaint(bimg, tilerect);
	Dimension odims = dims;
	Graphics2D gi = img == null ? null : img.createGraphics();
	AffineTransform at = null;
	if (gi != null) {
	    gi.translate(dims.width / 2, dims.height / 2);
	    at = gi.getTransform();
	}
	double framerate = 0;
	long start = System.currentTimeMillis();
	int numframes = 0;
	while (animator == me) {
	    dims = getSize();
	    if (img == null ||
		dims.width != img.getWidth() ||
		dims.height != img.getHeight())
	    {
		if (gi != null) {
		    gi.dispose();
		    gi = null;
		}
		img = (BufferedImage) createImage(dims.width, dims.height);
		x.newlimits(-dims.width/4, dims.width/4 - w.getInt());
		y.newlimits(-dims.height/4, dims.height/4 - h.getInt());
		gi = img.createGraphics();
		gi.translate(dims.width / 2, dims.height / 2);
		at = gi.getTransform();
	    }
	    if (tilesize != bimg.getWidth()) {
		tilesize = bimg.getWidth();
	    }
	    if (bouncesize) {
		w.anim();
		h.anim();
		x.newlimits(-dims.width/4, dims.width/4 - w.getInt());
		y.newlimits(-dims.height/4, dims.height/4 - h.getInt());
	    } else {
		if (w.getInt() != tilesize) {
		    w.set(tilesize);
		    x.newlimits(-dims.width/4, dims.width/4 - w.getInt());
		}
		if (h.getInt() != tilesize) {
		    h.set(tilesize);
		    y.newlimits(-dims.height/4, dims.height/4 - h.getInt());
		}
	    }
	    if (bouncerect) {
		x.anim();
		y.anim();
	    }
	    if (newtexture ||
		x.getInt() != tilerect.x || y.getInt() != tilerect.y ||
		w.getInt() != tilerect.width || h.getInt() != tilerect.height)
	    {
		newtexture = false;
		int X = x.getInt();
		int Y = y.getInt();
		int W = w.getInt();
		int H = h.getInt();
		tilerect = new Rectangle(X, Y, W, H);
		texture = new TexturePaint(bimg, tilerect);
		Color c1 = (dogray ? Color.white : Color.red);
		Color c2 = (dogray ? Color.black : Color.green);
		gradient = new GradientPaint(X, Y+H/2, c1,
					     X+W, Y+H/2, c2,
					     true);
		// Testing whether we can create an accurate copy from
		// the new accessor methods
		texture = new TexturePaint(texture.getImage(),
					   texture.getAnchorRect());
		gradient = new GradientPaint(gradient.getPoint1(),
					     gradient.getColor1(),
					     gradient.getPoint2(),
					     gradient.getColor2(),
					     gradient.isCyclic());
	    }
	    if (drawoval) {
		gi.setTransform(new AffineTransform());
		gi.setColor(Color.black);
		gi.setColor(Color.gray);
		gi.fillRect(0, 0, dims.width, dims.height);
	    }
	    gi.setTransform(at);
	    if (rotate) {
		rot.anim();
		gi.rotate(Math.toRadians(rot.getFlt()));
	    } else {
		rot.set(0);
	    }
	    if (shearx) {
		shx.anim();
		gi.shear(shx.getFlt()/100, 0.0f);
	    } else {
		shx.set(0);
	    }
	    if (sheary) {
		shy.anim();
		gi.shear(0.0f, shy.getFlt()/100);
	    } else {
		shy.set(0);
	    }
	    switch (texturetype) {
	    case TEXTURE:
		gi.setPaint(texture);
		break;
	    case GRADIENT:
		gi.setPaint(gradient);
		break;
	    case SOLIDFILL:
	    default:
		gi.setPaint(Color.black);
		break;
	    }
	    gi.setRenderingHint(RenderingHints.KEY_RENDERING,
				(quality
				 ? RenderingHints.VALUE_RENDER_QUALITY
				 : RenderingHints.VALUE_RENDER_SPEED));
	    if (drawoval) {
		gi.fillOval(-dims.width/2, -dims.height/2,
			    dims.width, dims.height);
	    } else {
		gi.fillRect(-1000, -1000, 2000, 2000);
	    }
	    if (showanchor) {
		gi.setColor(Color.black);
		gi.setColor(new Color(0f, 0f, 1f, .5f));
		gi.fill(tilerect);
	    }
	    long now = System.currentTimeMillis();
	    numframes++;
	    if (now - start > 1000) {
		framerate = numframes / ((now - start) / 1000.0);
		numframes = 0;
		start = now;
	    }
	    Graphics g = getGraphics();
	    g.drawImage(img, 0, 0, null);
	    g.setColor(Color.white);
	    g.drawString((Math.rint(framerate * 10.0) / 10.0) + " fps",
			 5, dims.height - 5);
	    g.dispose();
	}
    }

    public void paint(Graphics g) {
	if (img != null) {
	    g.drawImage(img, 0, 0, null);
	}
    }

    public static void main(String argv[]) {
	Frame f = new Frame();
	f.setLayout(new BorderLayout());
	final TextureAnim ta = new TextureAnim();
	f.add("Center", ta);
	f.addWindowListener(new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {System.exit(0);}
	    public void windowIconified(WindowEvent e) {ta.stop();}
	    public void windowDeiconified(WindowEvent e) {ta.start();}
	});
	f.add("South", new AnimControls(ta));
	f.pack();
	f.show();
	ta.start();
    }

    static class AnimControls extends Panel {
	public static final int BOUNCERECT = 0;
	public static final int BOUNCESIZE = 1;
	public static final int ROTATE = 2;
	public static final int SHEARX = 3;
	public static final int SHEARY = 4;
	public static final int QUALITY = 6;
	public static final int OVAL = 7;
	public static final int ANCHOR = 8;
	public static final int GRAY = 9;

	static class AnimListener implements ItemListener {
	    TextureAnim textureanim;
	    int which;

	    public AnimListener(TextureAnim ta, int which) {
		this.textureanim = ta;
		this.which = which;
	    }

	    public void itemStateChanged(ItemEvent e) {
		boolean val = (e.getStateChange() == ItemEvent.SELECTED);
		switch (which) {
		case BOUNCERECT: textureanim.bouncerect = val; break;
		case BOUNCESIZE: textureanim.bouncesize = val; break;
		case ROTATE: textureanim.rotate = val; break;
		case SHEARX: textureanim.shearx = val; break;
		case SHEARY: textureanim.sheary = val; break;
		case QUALITY: textureanim.quality = val; break;
		case OVAL: textureanim.drawoval = val; break;
		case ANCHOR: textureanim.showanchor = val; break;
		case GRAY: textureanim.doGray(val); break;
		}
	    }
	}

	static class AnimChooser implements ItemListener {
	    TextureAnim textureanim;

	    public AnimChooser(TextureAnim ta) {
		textureanim = ta;
	    }

	    public void itemStateChanged(ItemEvent e) {
		if (e.getStateChange() == ItemEvent.SELECTED) {
		    int size = Integer.parseInt((String) e.getItem());
		    textureanim.setSize(size);
		}
	    }
	}

	static class FillChooser implements ItemListener {
	    Choice chooser;
	    TextureAnim textureanim;

	    public FillChooser(Choice chooser, TextureAnim ta) {
		this.chooser = chooser;
		textureanim = ta;
	    }

	    public void itemStateChanged(ItemEvent e) {
		if (e.getStateChange() == ItemEvent.SELECTED) {
		    textureanim.texturetype = chooser.getSelectedIndex();
		}
	    }
	}

	public AnimControls(TextureAnim ta) {
	    Checkbox toggle;

	    setLayout(new GridLayout(0, 3));

	    toggle = new Checkbox("bounce", true);
	    toggle.addItemListener(new AnimListener(ta, BOUNCERECT));
	    add(toggle);
	    toggle = new Checkbox("resize", false);
	    toggle.addItemListener(new AnimListener(ta, BOUNCESIZE));
	    add(toggle);
	    toggle = new Checkbox("show anchor", true);
	    toggle.addItemListener(new AnimListener(ta, ANCHOR));
	    add(toggle);
	    toggle = new Checkbox("rotate", false);
	    toggle.addItemListener(new AnimListener(ta, ROTATE));
	    add(toggle);
	    toggle = new Checkbox("shearx", false);
	    toggle.addItemListener(new AnimListener(ta, SHEARX));
	    add(toggle);
	    toggle = new Checkbox("sheary", false);
	    toggle.addItemListener(new AnimListener(ta, SHEARY));
	    add(toggle);
	    toggle = new Checkbox("oval", false);
	    toggle.addItemListener(new AnimListener(ta, OVAL));
	    add(toggle);
	    toggle = new Checkbox("gray", false);
	    toggle.addItemListener(new AnimListener(ta, GRAY));
	    add(toggle);

	    add(new Label(""));

	    Choice chooser = new Choice();
	    chooser.add("4");
	    chooser.add("8");
	    chooser.add("16");
	    chooser.add("32");
	    chooser.add("100");
	    chooser.addItemListener(new AnimChooser(ta));
	    chooser.select(4);
	    add(chooser);

	    add(new Label(""));

	    chooser = new Choice();
	    chooser.add("TexturePaint");
	    chooser.add("GradientPaint");
	    chooser.add("SolidFill");
	    chooser.addItemListener(new FillChooser(chooser, ta));
	    chooser.select(0);
	    add(chooser);
	}
    }

    static class AnimVal {
	float curval;
	float lowval;
	float highval;
	float currate;
	float lowrate;
	float highrate;

	public AnimVal(int lowval, int highval,
		       int lowrate, int highrate) {
	    this.lowval = lowval;
	    this.highval = highval;
	    this.lowrate = lowrate;
	    this.highrate = highrate;
	    this.curval = randval(lowval, highval);
	    this.currate = randval(lowrate, highrate);
	}

	public AnimVal(int lowval, int highval,
		       int lowrate, int highrate,
		       int pos) {
	    this(lowval, highval, lowrate, highrate);
	    set(pos);
	}

	public float randval(float low, float high) {
	    return (float) (low + Math.random() * (high - low));
	}

	public float getFlt() {
	    return curval;
	}

	public int getInt() {
	    return (int) curval;
	}

	public void anim() {
	    curval += currate;
	    clip();
	}

	public void set(float val) {
	    curval = val;
	    clip();
	}

	public void clip() {
	    if (curval > highval) {
		curval = highval - (curval - highval);
		if (curval < lowval) {
		    curval = highval;
		}
		currate = - randval(lowrate, highrate);
	    } else if (curval < lowval) {
		curval = lowval + (lowval - curval);
		if (curval > highval) {
		    curval = lowval;
		}
		currate = randval(lowrate, highrate);
	    }
	}

	public void newlimits(int lowval, int highval) {
	    this.lowval = lowval;
	    this.highval = highval;
	    clip();
	}
    }
}
