/*
 * @(#)alphatest.h	1.1 98/04/10
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Performance and correctness testing for various Alpha composition
 * algorithms.
 */

#define False 0
#define True 1

/*
 * The typedefs for the basic composition functions.
 *
 * srcData - length pairs of source alpha, color
 * srcPremul - non-zero if source colors are alpha-premultiplied
 * dstData - length pairs of destination alpha, color
 * dstPremul - non-zero if destination colors are alpha-premultiplied
 * pathData - length samples of path coverage alpha
 * extraAlpha256 - "extra source alpha" value in the range [0..256]
 * resData - length pairs of result alpha, color (honors dstPremul)
 * length - number of samples
 */
typedef void CompositeFunc(unsigned char *srcData, int srcScan, int srcPremul,
			   unsigned char *dstData, int dstScan, int dstPremul,
			   unsigned char *pathData, int pathScan,
			   int extraAlpha256,
			   unsigned char *resData, int width, int height);

/*
 * The structure for holding a set of composition functions to handle
 * all 8 supported Porter-Duff compositing rules.
 */
typedef struct {
    CompositeFunc	*ClearFunc;
    CompositeFunc	*SrcFunc;
    CompositeFunc	*SrcOverFunc;
    CompositeFunc	*DstOverFunc;
    CompositeFunc	*SrcInFunc;
    CompositeFunc	*DstInFunc;
    CompositeFunc	*SrcOutFunc;
    CompositeFunc	*DstOutFunc;
} CompositeSet;

/*
 * The typedefs for the arrays of test values to be run through
 * the compositing functions.
 */

typedef struct {
    int numvals255;
    int *vals255;
    int numvals256;
    int *vals256;
} TestValues;

/*
 * The typedefs for the list of all Composite Sets and the list
 * of all Test Value sets for selection on the command line by name.
 */
typedef struct {
    char		*name;		/* command line selection name */
    char		*description;	/* verbose usage description */
    TestValues		*vals;		/* the test values themselves */
} TestValueEntry;

typedef struct {
    char		*name;		/* command line selection name */
    char		*description;	/* verbose usage description */
    CompositeSet	*set;		/* the composite set itself */
} CompositeSetEntry;

/*
 * The typedefs for the reference composition functions.
 */
typedef void ReferenceFunc(int srcAlpha, int srcColor, int srcPremul,
			   int dstAlpha, int dstColor, int dstPremul,
			   int pathAlpha, int extraAlpha256,
			   float *resAlpha, float *resColor);

/*
 * Various global data and utility functions.
 */
extern int verbose;
extern char *progname;

extern void usage();

extern int NumTestVals;
extern TestValueEntry TestValList[];
extern int NumCompSets;
extern CompositeSetEntry CompSetList[];

extern ReferenceFunc ClearReference;
extern ReferenceFunc SrcReference;
extern ReferenceFunc SrcOverReference;
extern ReferenceFunc DstOverReference;
extern ReferenceFunc SrcInReference;
extern ReferenceFunc DstInReference;
extern ReferenceFunc SrcOutReference;
extern ReferenceFunc DstOutReference;
