/*
 * @(#)beta3set.c	1.3 98/05/07
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * A CompositeSet that calculates the results using the JDK 1.2 beta3
 * AlphaComposite loop macros.
 */

#include "alphatest.h"

#define COMPST_TWO_COMPONENTS

#define NORM_SCALE	256
#define NORM_SCALE_1	255

#define ReadRastAlpha(rA)		\
    do {				\
	rA = *pathData;			\
        if (rA > 127) rA++;         	\
    } while (0)

#define ReadSrcPixel(sA, sR, sG, sB)	\
    do {				\
	sA = srcData[0];		\
	sR = srcData[1];		\
	if (!srcPremul &&		\
	    (sA < NORM_SCALE_1))	\
	{				\
	    if (sA) {			\
		int a = sA;		\
		if (a > NORM_SCALE_1/2) \
		    a++;		\
		sR = (a*sR)/NORM_SCALE;	\
	    } else {			\
		sR = 0;			\
	    }				\
	}				\
    } while (0)

#define ReadDstPixel(dA, dR, dG, dB)	\
    do {				\
	dA = dstData[0];		\
	dR = dstData[1];		\
	if (!dstPremul &&		\
	    dA < NORM_SCALE_1) {	\
	    if (dA) {			\
		int a = dA;		\
		if (a > NORM_SCALE_1/2)	\
		    a++;		\
		dR = (a*dR)/NORM_SCALE;	\
	    } else {			\
		dR = 0;			\
	    }				\
	}				\
    } while (0)

#define WriteDstPixel(dA, dR, dG, dB)	\
    do {				\
	int dRt;			\
        if (!dstPremul &&		\
	    dA < NORM_SCALE_1 &&	\
	    dA > 0)			\
	{				\
            dRt = (dR*NORM_SCALE_1)/dA;	\
	} else {			\
	    dRt = dR;			\
        }				\
	resData[0] = dA;		\
	resData[1] = dRt;		\
    } while (0)

#define NextRastAlpha			\
    do {				\
	pathData++;			\
    } while (0)

#define NextSrcPixel			\
    do {				\
	srcData += 2;			\
    } while (0)

#define NextDstPixel			\
    do {				\
	dstData += 2;			\
	resData += 2;			\
    } while (0)

#define NextRastAlphaRow		\
    do {				\
	pathData += pathScan;		\
    } while (0)

#define NextSrcRow			\
    do {				\
	srcData += srcScan;		\
    } while (0)

#define NextDstRow			\
    do {				\
	dstData += dstScan;		\
	resData += dstScan;		\
    } while (0)

#define extraAlpha extraAlpha256

#include "alphaCompst_loop_uti.h"
#include "alphaCompst_SRC.h"
#include "alphaCompst_SRC_OVER.h"
#include "alphaCompst_DST_OVER.h"
#include "alphaCompst_SRC_IN.h"
#include "alphaCompst_DST_IN.h"
#include "alphaCompst_SRC_OUT.h"
#include "alphaCompst_DST_OUT.h"

#define COMPOSELOOP(MACRO) 					\
    do {							\
	int row, column;					\
	int rastAlpha;						\
	int compAlpha;						\
	int srcAlpha;						\
	int srcRed;						\
	int dstAlpha;						\
	int dstRed;						\
	srcScan -= width * 2;					\
	dstScan -= width * 2;					\
	pathScan -= width;					\
	COMPOSITE_TILE(0, MACRO(rastAlpha, compAlpha, extraAlpha256,\
				srcAlpha, srcRed, 0, 0,		\
				dstAlpha, dstRed, 0, 0));	\
    } while (0)

static void
ClearBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    int row, column;
    int rastAlpha;
    int dstAlpha;
    int dstRed;
    int dstGreen;
    int dstBlue;
    dstScan -= width * 2;
    pathScan -= width;
    CLEAR_COMP;
}

static void
SrcBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	 unsigned char *dstData, int dstScan, int dstPremul,
	 unsigned char *pathData, int pathScan, int extraAlpha256,
	 unsigned char *resData, int width, int height)
{
    COMPOSELOOP(SRC_RULE);
}

static void
SrcOverBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	     unsigned char *dstData, int dstScan, int dstPremul,
	     unsigned char *pathData, int pathScan, int extraAlpha256,
	     unsigned char *resData, int width, int height)
{
    COMPOSELOOP(SRC_OVER_RULE);
}

static void
DstOverBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	     unsigned char *dstData, int dstScan, int dstPremul,
	     unsigned char *pathData, int pathScan, int extraAlpha256,
	     unsigned char *resData, int width, int height)
{
    COMPOSELOOP(DST_OVER_RULE);
}

static void
SrcInBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPOSELOOP(SRC_IN_RULE);
}

static void
DstInBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPOSELOOP(DST_IN_RULE);
}

static void
SrcOutBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	    unsigned char *dstData, int dstScan, int dstPremul,
	    unsigned char *pathData, int pathScan, int extraAlpha256,
	    unsigned char *resData, int width, int height)
{
    COMPOSELOOP(SRC_OUT_RULE);
}

static void
DstOutBeta3(unsigned char *srcData, int srcScan, int srcPremul,
	    unsigned char *dstData, int dstScan, int dstPremul,
	    unsigned char *pathData, int pathScan, int extraAlpha256,
	    unsigned char *resData, int width, int height)
{
    COMPOSELOOP(DST_OUT_RULE);
}

CompositeSet Beta3Set = {
    ClearBeta3,
    SrcBeta3,
    SrcOverBeta3,
    DstOverBeta3,
    SrcInBeta3,
    DstInBeta3,
    SrcOutBeta3,
    DstOutBeta3,
};

