/*
 * @(#)intset.c	1.1 98/04/10
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * A basic CompositeSet that calculates the results using integer math.
 * No intensive optimizations are made here, the answer is calculated
 * simply and straightforwardly.
 */

#include "alphatest.h"

#define MULDIV(a, b, c)			(((a)*(b) + ((c) / 2)) / (c))
#define MULDIV2(a, b, A, B, c)		(((a)*(b) + (A)*(B) + ((c) / 2)) / (c))

#define COMPOSELOOP(FA, FB) 						\
    do {								\
	pathScan -= width;						\
	srcScan -= width * 2;						\
	dstScan -= width * 2;						\
	while (--height >= 0) {						\
	    int w = width;						\
	    while (--w >= 0) {						\
		int pathAlpha = *pathData++;				\
		if (pathAlpha) {					\
		    int srcAlpha;					\
		    int srcColor;					\
		    int dstAlpha;					\
		    int dstColor;					\
		    int Fa, Fb;						\
		    if (extraAlpha256 == 0) {				\
			srcAlpha = 0;					\
			srcColor = 0;					\
		    } else {						\
			srcAlpha = srcData[0];				\
			if (srcAlpha == 0) {				\
			    srcColor = 0;				\
			} else {					\
			    srcAlpha += srcAlpha >> 7;			\
			    srcColor = srcData[1];			\
			    if (!srcPremul && srcAlpha < 256) {		\
				srcColor = MULDIV(srcColor,		\
						  srcAlpha, 256);	\
			    }						\
			    if (extraAlpha256 < 256) {			\
				srcAlpha = MULDIV(srcAlpha,		\
						  extraAlpha256,	\
						  256);			\
				srcColor = MULDIV(srcColor,		\
						  extraAlpha256,	\
						  256);			\
			    }						\
			}						\
		    }							\
		    dstAlpha = dstData[0];				\
		    if (dstAlpha == 0) {				\
			dstColor = 0;					\
		    } else {						\
			dstAlpha += dstAlpha >> 7;			\
			dstColor = dstData[1];				\
			if (!dstPremul && dstAlpha < 256) {		\
			    dstColor = MULDIV(dstColor,			\
					      dstAlpha,			\
					      256);			\
			}						\
		    }							\
		    Fa = (FA);						\
		    Fb = (FB);						\
		    srcAlpha = MULDIV2(srcAlpha, Fa,			\
				       dstAlpha, Fb, 256);		\
		    srcColor = MULDIV2(srcColor, Fa,			\
				       dstColor, Fb, 256);		\
		    if (pathAlpha < 255) {				\
			int c;						\
			pathAlpha += pathAlpha >> 7;			\
			c = 256 - pathAlpha;				\
			srcAlpha = MULDIV2(srcAlpha, pathAlpha,		\
					   dstAlpha, c, 256);		\
			srcColor = MULDIV2(srcColor, pathAlpha,		\
					   dstColor, c, 256);		\
		    }							\
		    if (!dstPremul && srcAlpha > 0) {			\
			srcColor = MULDIV(srcColor, 256, srcAlpha);	\
		    }							\
		    srcAlpha = MULDIV(srcAlpha, 255, 256);		\
		    /* REMIND: Better math wouldn't need this clip! */	\
		    if (srcColor > 255) srcColor = 255;			\
		    resData[0] = srcAlpha;				\
		    resData[1] = srcColor;				\
		}							\
		srcData += 2;						\
		dstData += 2;						\
		resData += 2;						\
	    }								\
	    pathData += pathScan;					\
	    srcData += srcScan;						\
	    dstData += dstScan;						\
	    resData += dstScan;						\
	}								\
    } while (0)

static void
ClearInt(unsigned char *srcData, int srcScan, int srcPremul,
	 unsigned char *dstData, int dstScan, int dstPremul,
	 unsigned char *pathData, int pathScan, int extraAlpha256,
	 unsigned char *resData, int width, int height)
{
    COMPOSELOOP(0, 0);
}

static void
SrcInt(unsigned char *srcData, int srcScan, int srcPremul,
       unsigned char *dstData, int dstScan, int dstPremul,
       unsigned char *pathData, int pathScan, int extraAlpha256,
       unsigned char *resData, int width, int height)
{
    COMPOSELOOP(256, 0);
}

static void
SrcOverInt(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPOSELOOP(256, 256 - srcAlpha);
}

static void
DstOverInt(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPOSELOOP(256 - dstAlpha, 256);
}

static void
SrcInInt(unsigned char *srcData, int srcScan, int srcPremul,
	 unsigned char *dstData, int dstScan, int dstPremul,
	 unsigned char *pathData, int pathScan, int extraAlpha256,
	 unsigned char *resData, int width, int height)
{
    COMPOSELOOP(dstAlpha, 0);
}

static void
DstInInt(unsigned char *srcData, int srcScan, int srcPremul,
	 unsigned char *dstData, int dstScan, int dstPremul,
	 unsigned char *pathData, int pathScan, int extraAlpha256,
	 unsigned char *resData, int width, int height)
{
    COMPOSELOOP(0, srcAlpha);
}

static void
SrcOutInt(unsigned char *srcData, int srcScan, int srcPremul,
	  unsigned char *dstData, int dstScan, int dstPremul,
	  unsigned char *pathData, int pathScan, int extraAlpha256,
	  unsigned char *resData, int width, int height)
{
    COMPOSELOOP(256 - dstAlpha, 0);
}

static void
DstOutInt(unsigned char *srcData, int srcScan, int srcPremul,
	  unsigned char *dstData, int dstScan, int dstPremul,
	  unsigned char *pathData, int pathScan, int extraAlpha256,
	  unsigned char *resData, int width, int height)
{
    COMPOSELOOP(0, 256 - srcAlpha);
}

CompositeSet IntSet = {
    ClearInt,
    SrcInt,
    SrcOverInt,
    DstOverInt,
    SrcInInt,
    DstInInt,
    SrcOutInt,
    DstOutInt,
};
