/*
 * @(#)proxyset.c	1.1 98/04/10
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * A default CompositeSet for testing the operation of the program.
 * These functions simply call the reference functions for an answer
 * and then cast that answer to the appropriate type.
 */

#include "alphatest.h"

static int
round255(float val)
{
    return (int) ((val * 255.0f) + 0.5);
}

#define COMPLOOP(REFFUNC)					\
    do {							\
	int j;							\
	srcScan -= width * 2;					\
	dstScan -= width * 2;					\
	pathScan -= width;					\
	for (j = 0; j < height; j++) {				\
	    int i;						\
	    for (i = 0; i < width; i++) {			\
		int srcAlpha = *srcData++;			\
		int srcColor = *srcData++;			\
		int dstAlpha = *dstData++;			\
		int dstColor = *dstData++;			\
		int pathAlpha = *pathData++;			\
		float resA, resC;				\
								\
		REFFUNC(srcAlpha, srcColor, srcPremul,		\
			dstAlpha, dstColor, dstPremul,		\
			pathAlpha, extraAlpha256,		\
			&resA, &resC);				\
								\
		*resData++ = round255(resA);			\
		*resData++ = round255(resC);			\
	    }							\
	    srcData += srcScan;					\
	    dstData += dstScan;					\
	    resData += dstScan;					\
	    pathData += pathScan;				\
	}							\
    } while (0)

static void
ClearProxy(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPLOOP(ClearReference);
}

static void
SrcProxy(unsigned char *srcData, int srcScan, int srcPremul,
	 unsigned char *dstData, int dstScan, int dstPremul,
	 unsigned char *pathData, int pathScan, int extraAlpha256,
	 unsigned char *resData, int width, int height)
{
    COMPLOOP(SrcReference);
}

static void
SrcOverProxy(unsigned char *srcData, int srcScan, int srcPremul,
	     unsigned char *dstData, int dstScan, int dstPremul,
	     unsigned char *pathData, int pathScan, int extraAlpha256,
	     unsigned char *resData, int width, int height)
{
    COMPLOOP(SrcOverReference);
}

static void
DstOverProxy(unsigned char *srcData, int srcScan, int srcPremul,
	     unsigned char *dstData, int dstScan, int dstPremul,
	     unsigned char *pathData, int pathScan, int extraAlpha256,
	     unsigned char *resData, int width, int height)
{
    COMPLOOP(DstOverReference);
}

static void
SrcInProxy(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPLOOP(SrcInReference);
}

static void
DstInProxy(unsigned char *srcData, int srcScan, int srcPremul,
	   unsigned char *dstData, int dstScan, int dstPremul,
	   unsigned char *pathData, int pathScan, int extraAlpha256,
	   unsigned char *resData, int width, int height)
{
    COMPLOOP(DstInReference);
}

static void
SrcOutProxy(unsigned char *srcData, int srcScan, int srcPremul,
	    unsigned char *dstData, int dstScan, int dstPremul,
	    unsigned char *pathData, int pathScan, int extraAlpha256,
	    unsigned char *resData, int width, int height)
{
    COMPLOOP(SrcOutReference);
}

static void
DstOutProxy(unsigned char *srcData, int srcScan, int srcPremul,
	    unsigned char *dstData, int dstScan, int dstPremul,
	    unsigned char *pathData, int pathScan, int extraAlpha256,
	    unsigned char *resData, int width, int height)
{
    COMPLOOP(DstOutReference);
}

CompositeSet RefProxySet = {
    ClearProxy,
    SrcProxy,
    SrcOverProxy,
    DstOverProxy,
    SrcInProxy,
    DstInProxy,
    SrcOutProxy,
    DstOutProxy,
};

