/*
 * @(#)AreaDemo.java	1.5 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


/* Revision: AreaDemo.java 1.5 @(#)AreaDemo.java	1.5 
 *
 * @package
 * @copyright
 *  (C) Copyright Taligent, Inc.,  1996 - AllRights Reserved
 *  (C) Copyright IBM Corp.,  1996
 *  This source code and documentation are Taligent Confidential materials. These materials are
 *  provided for internal evaluation purposes only. This notice may not be removed. This
 *  technology is protected by multiple US and international paterns.
 *
 *  Taligent is a registered trademark of Taligent, Inc.
 *
 */

import java.awt.*;
import java.awt.geom.*;
import java.applet.*;

import java.util.Vector;

public class AreaDemo extends Applet {
    public void init() {
        setLayout(new BorderLayout());
        DrawPanel dp = new DrawPanel();
        add("Center", dp);
        add("South", new DrawControls(dp));
    }

    public boolean handleEvent(Event e) {
        switch (e.id) {
          case Event.WINDOW_DESTROY:
            System.exit(0);
            return true;
          default:
            return false;
        }
    }

    public static void main(String args[]) {
        Frame f = new Frame("AreaDemo");
        AreaDemo demo = new AreaDemo();
        demo.init();
        demo.start();

        f.add("Center", demo);
        f.resize(400, 300);
        f.show();
    }
}

class DrawPanel extends Panel {
    public static final int RECTS = 0;
    public static final int OVALS = 1;
    int       mode = RECTS;

    Area theArea = new Area();

    public static final int ADD = 0;
    public static final int SUBTRACT = 1;
    public static final int INTERSECT = 2;
    public static final int XOR = 3;
    
    boolean updateArea = false;

    int action;   // default to add
    int x1,y1;
    int x2,y2;
    int xl,yl;

    public DrawPanel() {
        setBackground(Color.white);
        action = ADD;
    }

    public void clearArea() {
        theArea.reset();
        x2 = xl = -1;
        repaint();
    }
    public void updateArea(Area rhs) {
        switch (action) {
          case ADD:
            theArea.add(rhs);
            break;
          case SUBTRACT:
            theArea.subtract(rhs);
            break;
          case INTERSECT:
            theArea.intersect(rhs);
            break;
          case XOR:
            theArea.exclusiveOr(rhs);
            break;
          default:
            throw new IllegalArgumentException();
        }
        updateArea = true;
    }
    public void setAreaOperation(int opn) {
        switch (opn) {
          case ADD:
          case SUBTRACT:
          case INTERSECT:
          case XOR:
            action = opn;
            break;
          default:
            throw new IllegalArgumentException();
        }
    }
    public void setDrawMode(int theMode) {
        switch (theMode) {
          case RECTS:
          case OVALS:
            mode = theMode;
            break;
          default:
            throw new IllegalArgumentException();
        }
    }

    public boolean handleEvent(Event e) {
        switch (e.id) {
          case Event.MOUSE_DOWN:
            switch (mode) {
              case RECTS:
              case OVALS:
                x1 = e.x;
                y1 = e.y;
                x2 = -1;
                break;
              default:  // points?
                x1 = e.x;
                y1 = e.y;
                repaint();
                break;
            }
            return true;
          case Event.MOUSE_UP:
            Area target;
            switch (mode) {
              case RECTS:
                target = new Area(new Rectangle(x1, y1, e.x - x1, e.y - y1));
                x2 = xl = -1;
                updateArea(target);
                break;
              case OVALS:
                target = new Area(new Ellipse2D.Float(x1, y1, e.x - x1, e.y - y1));
                x2 = xl = -1;
                updateArea(target);
                break;
              default:
                break;
            }
            repaint();
            return true;
          case Event.MOUSE_DRAG:
            switch (mode) {
              case RECTS:
              case OVALS:
                xl = x2;
                yl = y2;
                x2 = e.x;
                y2 = e.y;
                break;
              default:  // points?
                x1 = e.x;
                y1 = e.y;
                break;
            }
            repaint();
            return true;
          case Event.WINDOW_DESTROY:
            System.exit(0);
            return true;
          default:
            return false;
        }
    }

    public void paint(Graphics g) {
//    int np = lines.size();

    /* draw the current lines */
//    g.setColor(getForeground());
//    g.setPaintMode();
//    for (int i=0; i < np; i++) {
//        Rectangle p = (Rectangle)lines.elementAt(i);
//        g.setColor((Color)colors.elementAt(i));
//        if (p.width != -1) {
//        g.drawLine(p.x, p.y, p.width, p.height);
//        } else {
//        g.drawLine(p.x, p.y, p.x, p.y);
//        }
//    }
        //IRootGrafPort rootPort = new IRootGrafPort(g);

	Graphics2D g2 = (Graphics2D)g;

        g2.setColor(Color.yellow);
	g2.fill(theArea);
        g2.setColor(Color.red);
        g2.draw(theArea);
        //IGrafBundle areaBundle =
            //new IGrafBundle(IBasicColor.kYellow, IBasicColor.kRed);
            //new IGrafBundle(IBasicColor.kGreen, IAttributeState.kFill);
        //rootPort.draw(theArea, areaBundle);

        if (x2 != -1 && xl != -1) {
            //IGrafBundle xorBundle = new IGrafBundle(IBasicColor.kBlack);
            //xorBundle.setFrameTransferMode(new IColorTransferMode(IColorTransferMode.kXOR));
            //ILinkedGrafPort xorPort = new ILinkedGrafPort(rootPort, xorBundle);
            g2.setColor(Color.black);
            g2.setXORMode(Color.green);
            switch (mode) {
              case RECTS:
                //xorPort.draw(new IGRect2D(new Rectangle(x1, y1, xl - x1, yl - y1)));
                g2.drawRect(x1, y1, xl - x1, yl - y1);
                break;
              case OVALS:
                //xorPort.draw(new IGEllipse2D(new IGRect2D(new Rectangle(x1, y1, xl - x1, yl - y1))));
                g2.drawOval(x1, y1, xl - x1, yl - y1);
                break;
              default:
                break;
            }
            g2.setPaintMode();
            //  end of scope of the XOR port...
        }
    }
}


class DrawControls extends Panel {
    DrawPanel target;

    public DrawControls(DrawPanel target) {
        this.target = target;
        setLayout(new FlowLayout());
        setBackground(Color.lightGray);
        target.setForeground(Color.red);
        CheckboxGroup group = new CheckboxGroup();
        Checkbox b;
        add(new Button("Clear"));
        add(b = new Checkbox("Add", group, false));
        add(b = new Checkbox("Sub", group, false));
        add(b = new Checkbox("Int", group, false));
        add(b = new Checkbox("XOR", group, false));
        Choice shapes = new Choice();
        shapes.addItem("Rectangle");
        shapes.addItem("Ellipse");
        shapes.setBackground(Color.lightGray);
        add(shapes);
    }

    public void paint(Graphics g) {
        Rectangle r = bounds();

        g.setColor(Color.lightGray);
        g.draw3DRect(0, 0, r.width, r.height, false);
    }

    public boolean action(Event e, Object arg) {
        if (e.target instanceof Checkbox) {
            if ("XOR".equals(((Checkbox)e.target).getLabel())) {
                target.setAreaOperation(DrawPanel.XOR);
            } else if ("Int".equals(((Checkbox)e.target).getLabel())) {
                target.setAreaOperation(DrawPanel.INTERSECT);
            } else if ("Sub".equals(((Checkbox)e.target).getLabel())) {
                target.setAreaOperation(DrawPanel.SUBTRACT);
            } else if ("Add".equals(((Checkbox)e.target).getLabel())) {
                target.setAreaOperation(DrawPanel.ADD);
            }
        } else if (e.target instanceof Choice) {
            if ("Rectangle".equals(arg)) {
                target.setDrawMode(DrawPanel.RECTS);
            } else if ("Ellipse".equals(arg)) {
                target.setDrawMode(DrawPanel.OVALS);
            }
        } else {
            if ("Clear".equals(arg)) {
                target.clearArea();
            }
        }
        return true;
    }
}




