/*
 * @(#)AreaTest.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
 
import java.io.PrintStream;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.font.*;
import java.awt.image.*;

public class AreaTest {
    public static boolean doIterNoTransform = true;
    public static boolean doIterTransform = true;
    public static boolean doIterNoFlat = true;
    public static boolean doIterFlat = true;
    public static boolean doRoundVals = true;
    public static long    minWorst = 10;

    public static void main(String argv[]) {
	if (argv.length == 0) {
	    new AreaTestHarness().start();
	} else {
	    Frame f = new Frame();
	    final AreaTestPanel t = new AreaTestPanel();
	    f.addWindowListener(new WindowAdapter() {
		public void windowClosed(WindowEvent e) {System.exit(0);}
		public void windowClosing(WindowEvent e) {System.exit(0);}
		public void windowIconified(WindowEvent e) {t.stop();}
		public void windowDeiconified(WindowEvent e) {t.start();}
	    });
	    f.add(t);
	    f.pack();
	    f.show();
	    t.start();
	}
    }
}

class AreaTestHarness extends Thread {
    public static ShapeFactory factories[] = {
	ShapeFactory.Rect,
	ShapeFactory.Oval,
	ShapeFactory.Arc,
	//ShapeFactory.Text,
    };

    public static AreaOperator operators[] = {
	AreaOperator.Union,
	AreaOperator.Intersect,
	AreaOperator.Subtract,
	AreaOperator.Xor,
    };

    public static Dimension areaDim = new Dimension(200, 200);

    AnimatingShape Left = new AnimatingShape();
    AnimatingShape Right = new AnimatingShape();
    AreaOperator Comp;
    PathExtractor extractor;
    long iterstart;
    long worst = AreaTest.minWorst-1;

    public void run() {
	Thread watchdog = new WatchDog(this);

	try {
	    watchdog.start();
	    runHarness();
	} finally {
	    watchdog.interrupt();
	}
    }

    public void runHarness() {
	long runstart = 0;
	int loops = 0;
	int ops = 0;
	while (true) {
	    for (int l = 0; l < factories.length; l++) {
		Left.setFactory(factories[l]);
		Area L = Left.getArea(areaDim);
		for (int r = 0; r < factories.length; r++) {
		    Right.setFactory(factories[r]);
		    Area R = Right.getArea(areaDim);
		    for (int c = 0; c < operators.length; c++) {
			Comp = operators[c];
			ops += readOut(L, R);
		    }
		}
	    }
	    if (worst < AreaTest.minWorst) {
		worst = AreaTest.minWorst;
		ops = 0;
		runstart = System.currentTimeMillis();
		System.out.println("// Priming pass completed - testing started");
	    } else {
		if (++loops >= 1) {
		    loops = 0;
		    long now = System.currentTimeMillis();
		    long total = now - runstart;
		    System.out.println("// "+ops+" operations executed in "+
				       total+" ms (avg = "+
				       roundiv(total, ops)+" ms, worst = "+
				       worst+" ms)");
		}
		Left.animStep();
		Right.animStep();
	    }
	}
    }

    public int readOut(Area L, Area R) {
	int ops = 0;
	if (AreaTest.doIterNoFlat) {
	    if (AreaTest.doIterNoTransform) {
		readOut(PathExtractor.None, L, R);
		ops++;
	    }
	    if (AreaTest.doIterTransform) {
		readOut(PathExtractor.TxOnly, L, R);
		ops++;
	    }
	}
	if (AreaTest.doIterFlat) {
	    if (AreaTest.doIterNoTransform) {
		readOut(PathExtractor.FlatOnly, L, R);
		ops++;
	    }
	    if (AreaTest.doIterTransform) {
		readOut(PathExtractor.TxAndFlat, L, R);
		ops++;
	    }
	}
	return ops;
    }

    private static double coords[] = new double[6];

    public void readOut(PathExtractor extractor, Area L, Area R) {
	try {
	    readOutUnprotected(extractor, L, R);
	} catch (Throwable t) {
	    report(System.out, extractor, "Exception ("+t+") during");
	}
    }

    public void readOutUnprotected(PathExtractor extractor, Area L, Area R) {
	startIter(extractor);
	PathIterator pi = extractor.getIterator(Comp.getResult(L, R));
	pi.getWindingRule();
	while (!pi.isDone()) {
	    pi.currentSegment(coords);
	    pi.next();
	}
	endIter();
	long iterend = System.currentTimeMillis();
	long itertotal = iterend - iterstart;
	if (itertotal > worst && worst >= AreaTest.minWorst) {
	    worst = itertotal;
	    report(System.out, extractor,
		   "New worst case - "+itertotal+" ms for");
	}
    }

    public synchronized void startIter(PathExtractor extractor) {
	System.gc();
	this.iterstart = System.currentTimeMillis();
	this.extractor = extractor;
	notify();
    }

    public synchronized void endIter() {
	System.gc();
	this.extractor = null;
    }

    public static double roundiv(long num, int denom) {
	return Math.round(num * 100.0 / denom) / 100.0;
    }

    public void printOut(PrintStream out, String varname, Shape s) {
	double coords[] = new double[6];
	PathIterator pi = s.getPathIterator(null);
	out.println("        DoublePath "+varname+
		    " = new DoublePath(DoublePath."+
		    (pi.getWindingRule() == pi.WIND_EVEN_ODD
		     ? "WIND_EVEN_ODD" : "WIND_NON_ZERO")+");");
	while (!pi.isDone()) {
	    String name;
	    int numcoords;
	    switch (pi.currentSegment(coords)) {
	    case pi.SEG_MOVETO:
		name = "moveTo";
		numcoords = 2;
		break;
	    case pi.SEG_LINETO:
		name = "lineTo";
		numcoords = 2;
		break;
	    case pi.SEG_QUADTO:
		name = "quadTo";
		numcoords = 4;
		break;
	    case pi.SEG_CUBICTO:
		name = "curveTo";
		numcoords = 6;
		break;
	    case pi.SEG_CLOSE:
		name = "closePath";
		numcoords = 0;
		break;
	    default:
		name = "Unknown";
		numcoords = 6;
		break;
	    }
	    out.print("        "+varname+"."+name+"(");
	    for (int i = 0; i < numcoords; i++) {
		if (i != 0) {
		    out.print(", ");
		    if ((i & 1) == 0) {
			out.println();
			int j = 10 + varname.length() + name.length();
			while (--j >= 0) {
			    out.print(" ");
			}
		    }
		}
		out.print(coords[i]);
	    }
	    out.println(");");
	    pi.next();
	}
    }

    private static int testN = 1;

    public synchronized void report(PrintStream out,
				    PathExtractor extractor, String label) {
	out.println("// "+label+" "+
		    Comp.getLabel(Left.getName(), Right.getName())+":");
	out.println();
	if (testN == 1) {
	    out.println("import java.awt.geom.*;");
	}
	out.println("class Test"+testN+" {");
	testN++;
	out.println("    public static AffineTransform Tx =");
	out.println("        "+PathExtractor.getTxInitCode());
	out.println("    public static void main(String argv[]) {");
	out.println("        double coords[] = new double[6];");
	printOut(out, "left", Left.getShape(areaDim));
	printOut(out, "right", Right.getShape(areaDim));
	out.println("        for (int i = 0; i < 4; i++) {");
	out.println("            System.gc();");
	out.println("            long start = System.currentTimeMillis();");
	out.println("            Area result = new Area(left);");
	out.println("            "+Comp.getLabel("result.",
						 "(new Area(right))")+";");
	out.println("            PathIterator pi = result.getPathIterator("+
		    extractor.getArgs()+");");
	out.println("            pi.getWindingRule();");
	out.println("            while (!pi.isDone()) {");
	out.println("                pi.currentSegment(coords);");
	out.println("                pi.next();");
	out.println("            }");
	out.println("            System.gc();");
	out.println("            long end = System.currentTimeMillis();");
	out.println("            System.out.println(\"operation took \"+");
	out.println("                               (end-start)+\" ms\");");
	out.println("        }");
	out.println("    }");
	out.println("}");
	out.println();
    }

    public static class WatchDog extends Thread {
	public static final long WATCHDOG_TIMEOUT = 2000;

	AreaTestHarness harness;

	public WatchDog(AreaTestHarness harness) {
	    setPriority(MAX_PRIORITY - 1);
	    setDaemon(true);
	    this.harness = harness;
	}

	public void run() {
	    synchronized (harness) {
		try {
		    runWatchDog();
		} catch (InterruptedException e) {
		}
	    }
	}

	public void runWatchDog() throws InterruptedException {
	    while (true) {
		long now = System.currentTimeMillis();
		long delay;
		if (harness.extractor == null) {
		    delay = 0;
		} else if (now > harness.iterstart + WATCHDOG_TIMEOUT) {
		    harness.report(System.out, harness.extractor,
				   "Watchdog timeout during");
		    delay = 0;
		} else {
		    delay = harness.iterstart + WATCHDOG_TIMEOUT + 1 - now;
		}
		harness.wait(delay);
	    }
	}
    }
}

class AreaTestPanel extends Panel implements Runnable {
    Panel showPanel;
    ShapePanel shapePanelA;
    ShapePanel shapePanelB;
    boolean flatten = false;
    boolean antialias = false;

    public AreaTestPanel() {
	setLayout(new BorderLayout());
	showPanel = new Panel();
	add("Center", showPanel);
	showPanel.setLayout(new GridLayout(2, 3));
	showPanel.add(new ShowArea(this, AreaOperator.Left));
	showPanel.add(new ShowArea(this, AreaOperator.Union));
	showPanel.add(new ShowArea(this, AreaOperator.Intersect));
	showPanel.add(new ShowArea(this, AreaOperator.Right));
	showPanel.add(new ShowArea(this, AreaOperator.Subtract));
	showPanel.add(new ShowArea(this, AreaOperator.Xor));
	shapePanelA = new ShapePanel(this, "Left", false, true);
	shapePanelA.chooseRect();
	add("North", shapePanelA);
	shapePanelB = new ShapePanel(this, "Right", true, false);
	shapePanelB.chooseOval();
	add("South", shapePanelB);
    }

    public Area getAreaA(Dimension d) {
	return shapePanelA.getArea(d);
    }

    public Area getAreaB(Dimension d) {
	return shapePanelB.getArea(d);
    }

    public Dimension getPreferredSize() {
	return new Dimension(600, 400);
    }

    public void repaint() {
	super.repaint();
	for (int i = 0; i < showPanel.getComponentCount(); i++) {
	    Component c = showPanel.getComponent(i);
	    c.repaint();
	}
    }

    public void paintShowPanels() {
	for (int i = 0; i < showPanel.getComponentCount(); i++) {
	    Component c = showPanel.getComponent(i);
	    Graphics g = c.getGraphics();
	    c.update(g);
	    g.dispose();
	}
    }

    Thread animator;

    public void start() {
	animator = new Thread(this);
	animator.setPriority(Thread.MIN_PRIORITY);
	animator.start();
    }

    public void stop() {
	animator = null;
    }

    public void run() {
	Thread me = Thread.currentThread();
	while (animator == me) {
	    shapePanelA.animStep();
	    shapePanelB.animStep();
	    paintShowPanels();
	}
    }
}

abstract class ShapeFactory {
    public abstract String getName();
    public abstract Shape getShape(float vals[]);

    public static ShapeFactory Rect = new ShapeFactory() {
	public String getName() {
	    return "Rect";
	}

	public Shape getShape(float vals[]) {
	    return new Rectangle2D.Float(vals[0], vals[1], vals[2], vals[3]);
	}
    };

    public static ShapeFactory Oval = new ShapeFactory() {
	public String getName() {
	    return "Oval";
	}

	public Shape getShape(float vals[]) {
	    return new Ellipse2D.Float(vals[0], vals[1], vals[2], vals[3]);
	}
    };

    public static ShapeFactory Arc = new ShapeFactory() {
	public String getName() {
	    return "Arc";
	}

	public Shape getShape(float vals[]) {
	    return new Arc2D.Float(vals[0], vals[1], vals[2], vals[3],
				   0.0f, vals[4] * 360.0f, Arc2D.PIE);
	}
    };

    public static ShapeFactory Text = new TextImpl();

    private static class TextImpl extends ShapeFactory {
	static Shape rawGlyphs;
	static AffineTransform lastTx;
	static Shape lastGlyphs;

	static {
	    Font theFont = new Font("Times Roman", Font.BOLD, 100);
	    FontRenderContext frc =
		new FontRenderContext(new AffineTransform(), false, true);
	    GlyphVector glyphs = theFont.createGlyphVector(frc, "Java");
	    Shape outline = glyphs.getOutline();

	    Rectangle2D bounds = outline.getBounds2D();
	    AffineTransform txnorm = new AffineTransform();
	    txnorm.scale(1.0 / bounds.getWidth(), 1.0 / bounds.getHeight());
	    txnorm.translate(-bounds.getX(), -bounds.getY());

	    rawGlyphs = txnorm.createTransformedShape(outline);
	    lastGlyphs = rawGlyphs;
	}

	static synchronized Shape getCachedShape(AffineTransform tx) {
	    if (!tx.equals(lastTx)) {
		lastTx = tx;
		lastGlyphs = tx.createTransformedShape(rawGlyphs);
	    }
	    return lastGlyphs;
	}

	public String getName() {
	    return "Text";
	}

	public Shape getShape(float vals[]) {
	    AffineTransform tx = new AffineTransform();
	    tx.translate(vals[0], vals[1]);
	    tx.scale(vals[2], vals[3]);
	    return getCachedShape(tx);
	}
    }
}

abstract class AreaOperator {
    public abstract String getLabel(String l, String r);
    public abstract Area getResult(Area left, Area right);

    public static AreaOperator Left = new AreaOperator() {
	public String getLabel(String l, String r) {
	    return l;
	}

	public Area getResult(Area left, Area right) {
	    return left;
	}
    };

    public static AreaOperator Right = new AreaOperator() {
	public String getLabel(String l, String r) {
	    return r;
	}

	public Area getResult(Area left, Area right) {
	    return right;
	}
    };

    public static AreaOperator Union = new AreaOperator() {
	public String getLabel(String l, String r) {
	    return l + " add " + r;
	}

	public Area getResult(Area left, Area right) {
	    Area ret = new Area(left);
	    ret.add(right);
	    return ret;
	}
    };

    public static AreaOperator Intersect = new AreaOperator() {
	public String getLabel(String l, String r) {
	    return l + " intersect " + r;
	}

	public Area getResult(Area left, Area right) {
	    Area ret = new Area(left);
	    ret.intersect(right);
	    return ret;
	}
    };

    public static AreaOperator Subtract = new AreaOperator() {
	public String getLabel(String l, String r) {
	    return l + " subtract " + r;
	}

	public Area getResult(Area left, Area right) {
	    Area ret = new Area(left);
	    ret.subtract(right);
	    return ret;
	}
    };

    public static AreaOperator Xor = new AreaOperator() {
	public String getLabel(String l, String r) {
	    return l + " exclusiveOr " + r;
	}

	public Area getResult(Area left, Area right) {
	    Area ret = new Area(left);
	    ret.exclusiveOr(right);
	    return ret;
	}
    };
}

abstract class PathExtractor {
    public abstract PathIterator getIterator(Shape s);
    public abstract String getArgs();

    private static AffineTransform Tx =
	AffineTransform.getScaleInstance(3.0, 4.0);

    public static String getTxInitCode() {
	return "AffineTransform.getScaleInstance(3.0, 4.0);";
    }

    public static PathExtractor None = new PathExtractor() {
	public PathIterator getIterator(Shape s) {
	    return s.getPathIterator(null);
	}
	public String getArgs() {
	    return "null";
	}
    };

    public static PathExtractor TxOnly = new PathExtractor() {
	public PathIterator getIterator(Shape s) {
	    return s.getPathIterator(Tx);
	}
	public String getArgs() {
	    return "Tx";
	}
    };

    public static PathExtractor FlatOnly = new PathExtractor() {
	public PathIterator getIterator(Shape s) {
	    return s.getPathIterator(null, 1.0f);
	}
	public String getArgs() {
	    return "null, 1.0f";
	}
    };

    public static PathExtractor TxAndFlat = new PathExtractor() {
	public PathIterator getIterator(Shape s) {
	    return s.getPathIterator(Tx, 1.0f);
	}
	public String getArgs() {
	    return "Tx, 1.0f";
	}
    };
}

class AnimatingShape {
    public static final int NUM_VALS = 5;	// XYWH + aux

    float vals[] = new float[NUM_VALS + 1];	// extra 0 value for "link"
    static float initlo[] = { 0.1f, 0.1f, 0.25f, 0.25f, 0.0f };
    static float inithi[] = { 0.4f, 0.4f, 0.5f, 0.5f, 1.0f };
    static float min[] = { 0.0f, 0.0f, 0.20f, 0.20f, 0.0f };
    static float max[] = { 1.0f, 1.0f, 0.6f, 0.6f, 1.0f };
    static int animorder[] = {2, 3, 0, 1, 4};
    static int animlink[] = {2, 3, 5, 5, 5};

    float speeds[] = new float[NUM_VALS];
    static float slow[] = { 0.02f, 0.02f, 0.02f, 0.02f, 3.0f / 360.0f };
    static float fast[] = { 0.10f, 0.10f, 0.10f, 0.10f, 10.0f / 360.0f };

    ShapeFactory factory;

    public AnimatingShape() {
	for (int i = 0; i < NUM_VALS; i++) {
	    vals[i] = rand(initlo[i], inithi[i]);
	    speeds[i] = rand(slow[i], fast[i]);
	}
    }

    public AnimatingShape(ShapeFactory factory) {
	this();
	setFactory(factory);
    }

    public void setFactory(ShapeFactory factory) {
	this.factory = factory;
    }

    public String getName() {
	return factory.getName();
    }

    public static float rand(float low, float hi) {
	return (float) (Math.random() * (hi - low) + low);
    }

    public void animStep() {
	for (int j = 0; j < NUM_VALS; j++) {
	    int i = animorder[j];
	    float val = vals[i] + speeds[i];
	    float link = vals[animlink[i]];
	    if (val + link > max[i]) {
		val = (max[i] - link) * 2.0f - val;
		speeds[i] = -rand(slow[i], fast[i]);
	    } else if (val < min[i]) {
		val = min[i] * 2.0f - val;
		speeds[i] = rand(slow[i], fast[i]);
	    }
	    vals[i] = val;
	}
    }

    public Area getArea(Dimension d) {
	return new Area(getShape(d));
    }

    public Shape getShape(Dimension d) {
	float scaledvals[] = { vals[0],
			       vals[1],
			       vals[2],
			       vals[3],
			       vals[4] };
	if (AreaTest.doRoundVals) {
	    scaledvals[0] = roundit(scaledvals[0]);
	    scaledvals[1] = roundit(scaledvals[1]);
	    scaledvals[2] = roundit(scaledvals[2]);
	    scaledvals[3] = roundit(scaledvals[3]);
	}
	scaledvals[0] *= d.width;
	scaledvals[1] *= d.height;
	scaledvals[2] *= d.width;
	scaledvals[3] *= d.height;
	return factory.getShape(scaledvals);
    }

    public static float roundit(float v) {
	return Math.round(v * 10.0f) / 10.0f;
    }
}

class ShowArea extends Canvas {
    AreaTestPanel test;
    AreaOperator operator;
    BufferedImage img;

    public ShowArea(AreaTestPanel test, AreaOperator operator) {
	this.test = test;
	this.operator = operator;
    }

    public void paint(Graphics g) {
	if (img == null) {
	    update(g);
	} else {
	    g.drawImage(img, 0, 0, this);
	}
    }

    public void update(Graphics g) {
	Dimension compDim = getSize();
	if (img == null ||
	    compDim.width != img.getWidth() ||
	    compDim.height != img.getHeight())
	{
	    img = (BufferedImage) createImage(compDim.width, compDim.height);
	}
	Dimension areaDim = new Dimension(compDim.width - 2,
					  compDim.height - 21);
	Area a = test.getAreaA(areaDim);
	Area b = test.getAreaB(areaDim);
	Shape s = operator.getResult(a, b);
	if (test.flatten) {
	    s = new FlatShape(s);
	}
	//Graphics2D g2 = (Graphics2D) g;
	Graphics2D g2 = img.createGraphics();
	if (test.antialias) {
	    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
	}
	g2.clearRect(0, 0, compDim.width, compDim.height);
	g2.setColor(Color.black);
	g2.drawRect(0, 0, compDim.width - 1, compDim.height - 1);
	((Graphics)g2).drawString(operator.getLabel("Left", "Right"), 0, 20);
	g2.translate(1, 20);
	g2.setColor(Color.blue);
	//g2.setRenderingHints(g2.ANTIALIASING, g2.ANTIALIAS_ON);
	//g2.setRenderingHints(g2.RENDERING, g2.RENDER_QUALITY);
	g2.fill(s);
	g2.setColor(Color.red);
	g2.draw(a);
	g2.setColor(Color.green);
	g2.draw(b);
	g2.setColor(Color.white);
	g2.draw(s);
	if (g2 != g) {
	    g2.dispose();
	}
	g.drawImage(img, 0, 0, this);
    }
}

class FlatShape implements Shape {
    Shape delegate;

    public FlatShape(Shape s) {
	delegate = s;
    }

    public Rectangle getBounds() {
	return delegate.getBounds();
    }

    public Rectangle2D getBounds2D() {
	return delegate.getBounds2D();
    }

    public boolean contains(double x, double y) {
	return delegate.contains(x, y);
    }

    public boolean contains(Point2D p) {
	return delegate.contains(p);
    }

    public boolean intersects(double x, double y, double w, double h) {
	return delegate.intersects(x, y, w, h);
    }

    public boolean intersects(Rectangle2D r) {
	return delegate.intersects(r);
    }

    public boolean contains(double x, double y, double w, double h) {
	return delegate.contains(x, y, w, h);
    }

    public boolean contains(Rectangle2D r) {
	return delegate.contains(r);
    }

    public PathIterator getPathIterator(AffineTransform at) {
	return delegate.getPathIterator(at, .001);
    }

    public PathIterator getPathIterator(AffineTransform at, double flatness) {
	return delegate.getPathIterator(at, flatness);
    }
}

class ShapePanel extends Panel implements ItemListener {
    AreaTestPanel test;
    Checkbox rectToggle;
    Checkbox ovalToggle;
    Checkbox arcToggle;
    Checkbox textToggle;
    Checkbox flatToggle;
    Checkbox aaToggle;
    AnimatingShape theShape = new AnimatingShape();

    public ShapePanel(AreaTestPanel test, String label,
		      boolean flat, boolean AA) {
	this.test = test;
	add(new Label(label+" shape type: "));

	CheckboxGroup g = new CheckboxGroup();
	rectToggle = new Checkbox("Rectangle", true, g);
	rectToggle.addItemListener(this);
	add(rectToggle);

	ovalToggle = new Checkbox("Ellipse", false, g);
	ovalToggle.addItemListener(this);
	add(ovalToggle);

	arcToggle = new Checkbox("Arc", false, g);
	arcToggle.addItemListener(this);
	add(arcToggle);

	textToggle = new Checkbox("Text", false, g);
	textToggle.addItemListener(this);
	add(textToggle);

	if (flat) {
	    flatToggle = new Checkbox("Flatten", false);
	    flatToggle.addItemListener(this);
	    add(flatToggle);
	}

	if (AA) {
	    aaToggle = new Checkbox("Antialias", false);
	    aaToggle.addItemListener(this);
	    add(aaToggle);
	}
    }

    public void chooseRect() {
	rectToggle.setState(true);
	theShape.setFactory(ShapeFactory.Rect);
    }

    public void chooseOval() {
	ovalToggle.setState(true);
	theShape.setFactory(ShapeFactory.Oval);
    }

    public void chooseArc() {
	arcToggle.setState(true);
	theShape.setFactory(ShapeFactory.Arc);
    }

    public void chooseText() {
	textToggle.setState(true);
	theShape.setFactory(ShapeFactory.Text);
    }

    public void itemStateChanged(ItemEvent e) {
	ItemSelectable item = e.getItemSelectable();
	if (item == rectToggle) {
	    theShape.setFactory(ShapeFactory.Rect);
	} else if (item == ovalToggle) {
	    theShape.setFactory(ShapeFactory.Oval);
	} else if (item == arcToggle) {
	    theShape.setFactory(ShapeFactory.Arc);
	} else if (item == textToggle) {
	    theShape.setFactory(ShapeFactory.Text);
	} else if (item == flatToggle) {
	    test.flatten = !test.flatten;
	    test.paintShowPanels();
	} else if (item == aaToggle) {
	    test.antialias = !test.antialias;
	    test.paintShowPanels();
	}
	test.repaint();
    }

    public Area getArea(Dimension d) {
	return theShape.getArea(d);
    }

    public void animStep() {
	theShape.animStep();
    }
}

class DoublePath implements Shape {
    public static final int WIND_EVEN_ODD = PathIterator.WIND_EVEN_ODD;

    public static final int WIND_NON_ZERO = PathIterator.WIND_NON_ZERO;
    
    // For code simplicity, copy these constants to our namespace
    // and cast them to byte constants for easy storage.
    private static final byte SEG_MOVETO  = (byte) PathIterator.SEG_MOVETO;
    private static final byte SEG_LINETO  = (byte) PathIterator.SEG_LINETO;
    private static final byte SEG_QUADTO  = (byte) PathIterator.SEG_QUADTO;
    private static final byte SEG_CUBICTO = (byte) PathIterator.SEG_CUBICTO;
    private static final byte SEG_CLOSE   = (byte) PathIterator.SEG_CLOSE;

    byte[] pointTypes;
    double[] pointCoords;
    int numTypes;
    int numCoords;
    int windingRule;

    static final int INIT_SIZE = 20;
    static final int EXPAND_MAX = 500;

    public DoublePath(int rule) {
	windingRule = rule;
	pointTypes = new byte[INIT_SIZE];
	pointCoords = new double[INIT_SIZE * 2];
    }

    private void needRoom(int newTypes, int newCoords, boolean needMove) {
	if (needMove && numTypes == 0) {
	    throw new IllegalPathStateException("missing initial moveto "+
						"in path definition");
	}
	int size = pointCoords.length;
	if (numCoords + newCoords > size) {
	    int grow = size;
	    if (grow > EXPAND_MAX * 2) {
		grow = EXPAND_MAX * 2;
	    }
	    if (grow < newCoords) {
		grow = newCoords;
	    }
	    double[] arr = new double[size + grow];
	    System.arraycopy(pointCoords, 0, arr, 0, numCoords);
	    pointCoords = arr;
	}
	size = pointTypes.length;
	if (numTypes + newTypes > size) {
	    int grow = size;
	    if (grow > EXPAND_MAX) {
		grow = EXPAND_MAX;
	    }
	    if (grow < newTypes) {
		grow = newTypes;
	    }
	    byte[] arr = new byte[size + grow];
	    System.arraycopy(pointTypes, 0, arr, 0, numTypes);
	    pointTypes = arr;
	}
    }

    public synchronized void moveTo(double x, double y) {
	if (numTypes > 0 && pointTypes[numTypes - 1] == SEG_MOVETO) {
	    pointCoords[numCoords - 2] = x;
	    pointCoords[numCoords - 1] = y;
	} else {
	    needRoom(1, 2, false);
	    pointTypes[numTypes++] = SEG_MOVETO;
	    pointCoords[numCoords++] = x;
	    pointCoords[numCoords++] = y;
	}
    }

    public synchronized void lineTo(double x, double y) {
	needRoom(1, 2, true);
	pointTypes[numTypes++] = SEG_LINETO;
	pointCoords[numCoords++] = x;
	pointCoords[numCoords++] = y;
    }

    public synchronized void quadTo(double x1, double y1, double x2, double y2) {
	needRoom(1, 4, true);
	pointTypes[numTypes++] = SEG_QUADTO;
	pointCoords[numCoords++] = x1;
	pointCoords[numCoords++] = y1;
	pointCoords[numCoords++] = x2;
	pointCoords[numCoords++] = y2;
    }

    public synchronized void curveTo(double x1, double y1,
				     double x2, double y2,
				     double x3, double y3) {
	needRoom(1, 6, true);
	pointTypes[numTypes++] = SEG_CUBICTO;
	pointCoords[numCoords++] = x1;
	pointCoords[numCoords++] = y1;
	pointCoords[numCoords++] = x2;
	pointCoords[numCoords++] = y2;
	pointCoords[numCoords++] = x3;
	pointCoords[numCoords++] = y3;
    }

    public synchronized void closePath() {
	if (numTypes == 0 || pointTypes[numTypes - 1] != SEG_CLOSE) {
	    needRoom(1, 0, true);
	    pointTypes[numTypes++] = SEG_CLOSE;
	}
    }

    public java.awt.Rectangle getBounds() {
	return new Area(this).getBounds();
    }

    public synchronized Rectangle2D getBounds2D() {
	return new Area(this).getBounds2D();
    }

    public boolean contains(double x, double y) {
	return new Area(this).contains(x, y);
    }

    public boolean contains(Point2D p) {
	return contains(p.getX(), p.getY());
    }

    public boolean contains(double x, double y, double w, double h) {
	return new Area(this).contains(x, y, w, h);
    }

    public boolean contains(Rectangle2D r) {
	return contains(r.getX(), r.getY(), r.getWidth(), r.getHeight());
    }

    public boolean intersects(double x, double y, double w, double h) {
	return new Area(this).intersects(x, y, w, h);
    }

    public boolean intersects(Rectangle2D r) {
	return intersects(r.getX(), r.getY(), r.getWidth(), r.getHeight());
    }

    /**
     * Return an iterator object that iterates along the boundary of
     * the shape and provides access to the geometry of the outline
     * of the shape.
     */
    public PathIterator getPathIterator(AffineTransform at) {
	return new Iterator(this, at);
    }

    /**
     * Return an iterator object that iterates along the boundary of
     * the flattened shape and provides access to the geometry of the
     * outline of the shape.
     */
    public PathIterator getPathIterator(AffineTransform at, double flatness) {
	return new FlatteningPathIterator(getPathIterator(at), flatness);
    }

    static class Iterator implements PathIterator {
	int typeIdx = 0;
	int pointIdx   = 0;
	DoublePath path;
	AffineTransform affine;

	private static final int curvesize[] = {2, 2, 4, 6, 0};

	Iterator(DoublePath path, AffineTransform at) {
	    this.path = path;
	    this.affine = at;
	}

	public int getWindingRule() {
	    return path.windingRule;
	}

	public boolean isDone() {
	    return (typeIdx >= path.numTypes);
	}

	public void next() {
	    int type = path.pointTypes[typeIdx++];
	    pointIdx += curvesize[type];
	}

	public int currentSegment(float[] coords) {
	    int type = path.pointTypes[typeIdx];
	    int numCoords = curvesize[type];
	    if (numCoords > 0 && affine != null) {
		affine.transform(path.pointCoords, pointIdx,
				 coords, 0,
				 numCoords / 2);
	    } else {
		for (int i=0; i < numCoords; i++) {
		    coords[i] = (float) path.pointCoords[pointIdx + i];
		}
	    }
	    return type;
	}

	public int currentSegment(double[] coords) {
	    int type = path.pointTypes[typeIdx];
	    int numCoords = curvesize[type];
	    if (numCoords > 0 && affine != null) {
		affine.transform(path.pointCoords, pointIdx,
				 coords, 0,
				 numCoords / 2);
	    } else {
		System.arraycopy(path.pointCoords, pointIdx,
				 coords, 0,
				 numCoords);
	    }
	    return type;
	}
    }
}
