/*
 * @(#)TestManager2D.java	1.12 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * @(#)TestManager2D.java	1.12 98/09/21
 * 
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road, 
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * CopyrightVersion 1.2
 * 
 */

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Color;
import java.awt.Image;
import java.awt.Font;
import java.awt.image.BufferedImage;

import java.applet.Applet;
import java.awt.Canvas;
import java.awt.Panel;
import java.awt.Event;
import java.lang.Math;

import java.awt.geom.AffineTransform;

class Statistics 
{
    double      desiredAccuracy;
    double      experiments;
    double		accumulatorX;
    double      accumulatorX2;
    double		accumulatorY;
    double		accumulatorY2;
    double		accumulatorXY;

    String      context;
    String      strX;

    public Statistics(String ctx,String sX)
    {
        desiredAccuracy = 0.9999;
        //        desiredAccuracy = 0.9999986;
        experiments     = 0;
        accumulatorX    = 0;
        accumulatorX2   = 0;
        accumulatorY	= 0;
        accumulatorY2	= 0;	
        accumulatorXY	= 0;

        context     = ctx;
        strX        = sX;
    }

    public void newExperimentUpdate(double x,double y)
    {
        experiments  ++;
        accumulatorX += x;
        accumulatorX2 += (x*x);
        accumulatorY  += y;
        accumulatorY2 += (y*y);
        accumulatorXY += (x*y);
        
    } 
    
    public boolean needMoreExperiments()
    {
        double correlation = getCorrelationCoefficient();
        boolean retval = (correlation > -desiredAccuracy);
        retval &= (correlation < desiredAccuracy);
        return retval;
    }

    public double getVarianceX()
    {
        double mean = getMeanX();
        return accumulatorX2/experiments - mean*mean;
    }

    public double getMeanX()
    {
        return accumulatorX / experiments;
    }

    public double getMeanY()
    {
        return accumulatorY / experiments;
    }

    public double getVarianceY()
    {
        double mean = getMeanY();
        return accumulatorY2/experiments - mean*mean;
    }
	
    public double getSlope()
    {
        if (experiments > 0) {
            return (accumulatorXY - accumulatorX*accumulatorY/experiments) / (accumulatorX2 - accumulatorX*accumulatorX/experiments); 
        } else {
            return 0;
        }
    }

    public double getCorrelationCoefficient()
    {
        //		return getSlope()*Math.sqrt(getVarianceX())/Math.sqrt(getVarianceY()) ;
        /*
         * not mathematically the right thing, but precisely related
         * and easier to calculate
         */
        double slope = getSlope();
        return slope * slope * getVarianceX()/getVarianceY() ;
    }
    public void dump()
    {
        System.out.println("    Ex: "+ experiments);

        System.out.println("    X : "+ accumulatorX);
        System.out.println("    X2: "+ accumulatorX2);
        System.out.println("    Y : "+ accumulatorY);
        System.out.println("    Y2: "+ accumulatorY2);
        System.out.println("    XY: "+ accumulatorXY);
        System.out.println("   y/x: "+ getMeanY()/getMeanX());
        System.out.println("     m: "+ getSlope());
        System.out.println("    vx: "+ getVarianceX());
        System.out.println("    vy: "+ getVarianceY());
    }	 
}


public class TestManager2D extends Applet implements Runnable
{
    private static int DRAWSIZE = 300;

    Thread myThread;

    private void ReportHeader(TestPortion2D testObject) {
        System.out.println("name, size, " + testObject.getUnits());
    }

    private void ReportResults(TestPortion2D testObject, int testNum, long speed,String sContextName) {
        System.out.println(
                           sContextName + "-" +
                           testObject.getTestName(testNum) + "," +
                           testObject.getSize() + "," +
                           speed
                           );

    }

    private void ReportStatHeader(TestPortion2D testObject) {
        System.out.println("Name"+ ",Test"+ ",Size"+ ",Slope"+ ",Correlation Coefficient");
    }

    private void ReportStatResults(TestPortion2D testObject,int testNum,String sContextName,double baseLine, Statistics stat) {
        if ( null != stat) {
            System.out.println(
                               testObject.getTestName(testNum) + "," +
                               sContextName + "," +
                               testObject.getSize() + "," +
                               1000.0 * (stat.getSlope()-baseLine) + "," + // convert to microseconds
                               stat.getCorrelationCoefficient() 
                               );
            //            stat.dump();
        }
    }

    private void PerformTests(Graphics g,TestPortion2D testObject,String sContextName) {
        g.clearRect(0,0,DRAWSIZE,DRAWSIZE);
        int size;
        int nTests = testObject.getNumTests();
        int testNum;

        for (testNum = 0; testNum < nTests; testNum++) {
            ReportHeader(testObject);
            for (size = 0; size < 10; size++) {
                testObject.setSize(size);
                Runtime rt = Runtime.getRuntime();
                rt.gc();
                long speed = testObject.MeasureSpeed(testNum);
                ReportResults(testObject,testNum,speed,sContextName);
            }
            for (size = 10; size < DRAWSIZE; size += 20) {
                testObject.setSize(size);
                Runtime rt = Runtime.getRuntime();
                rt.gc();
                long speed = testObject.MeasureSpeed(testNum);
                ReportResults(testObject,testNum,speed,sContextName);
            }
        }
    }

    private void PerformStatTests(Graphics g,TestPortion2D testObject,String sContextName) {
        g.clearRect(0,0,DRAWSIZE,DRAWSIZE);
        int size;
        int nTests = testObject.getNumTests();
        int testNum;

        Statistics  stat;
        ReportStatHeader(testObject);
        for (testNum = 0; testNum < nTests; testNum++) {
            double baseLine = 0;

            Runtime rt = Runtime.getRuntime();
            // prime the system, load all class dependencies.
            rt.gc();
            testObject.setSize(1);
            stat = testObject.MeasureStatistics(testNum);
            // System.err.println("baseline[" + baseLine + "] time["+testObject.elapsedTime()+"]");

            rt = Runtime.getRuntime();
            rt.gc();
            testObject.setSize(0);
            stat = testObject.MeasureStatistics(testNum);
            baseLine = stat.getSlope();
            System.err.println("baseline[" + baseLine + "] time["+testObject.elapsedTime()+"]");
            
            for (size = 1; size < 10; size++) {
                testObject.setSize(size);
                rt = Runtime.getRuntime();
                rt.gc();
                stat = testObject.MeasureStatistics(testNum);
                //System.err.println("baseline[" + stat.getSlope() + "] time["+testObject.elapsedTime()+"]");
                ReportStatResults(testObject,testNum,sContextName,baseLine,stat);
				
            } 
            for (size = 20; size < DRAWSIZE; size += 20) {
                testObject.setSize(size);
                rt = Runtime.getRuntime();
                rt.gc();
                stat = testObject.MeasureStatistics(testNum);
                //System.out.println("baseline[" + stat.getSlope() + "] time["+testObject.elapsedTime()+"]");
                ReportStatResults(testObject,testNum,sContextName,baseLine,stat);
            } 
        }
    }

    private void BasicTests(Graphics g,String sContextName) {
        //        PerformTests(g,new Test2DfillRect(g),sContextName);
        //        PerformTests(g,new Test2DdrawRect(g),sContextName);
        //        PerformTests(g,new Test2DclearRect(g),sContextName);
        //        PerformTests(g,new Test2DdrawLine(g),sContextName);
        //        PerformTests(g,new Test2Ddraw3DRect(g),sContextName);
        //        PerformTests(g,new Test2Dfill3DRect(g),sContextName);
        //        PerformTests(g,new Test2DdrawOval(g),sContextName);
        //        PerformTests(g,new Test2DfillOval(g),sContextName);
        //        PerformTests(g,new Test2DdrawRoundRect(g),sContextName);
        //        PerformTests(g,new Test2DfillRoundRect(g),sContextName);
        //        PerformTests(g,new Test2DdrawArc(g),sContextName);
        //        PerformTests(g,new Test2DfillArc(g),sContextName);
        //        PerformTests(g,new Test2DdrawPolyline(g),sContextName);
        //        PerformTests(g,new Test2DdrawPolygon(g),sContextName);
        //        PerformTests(g,new Test2DfillPolygon(g),sContextName);
                PerformTests(g,new Test2DsetFont(g),sContextName);
                PerformTests(g,new Test2DnewFont(g),sContextName);
                PerformTests(g,new Test2DdrawString(g),sContextName);
                PerformTests(g,new Test2DdrawChars(g),sContextName);
                PerformTests(g,new Test2DdrawBytes(g),sContextName);
        //        PerformTests(g,new Test2DcopyArea(g),sContextName);
        //    public abstract boolean drawImage(Image img, int x, int y, ImageObserver observer);
        //    public abstract boolean drawImage(Image img, int x, int y, int width, int height, ImageObserver observer);
        //    public abstract boolean drawImage(Image img, int x, int y, Color bgcolor, ImageObserver observer);
        //    public abstract boolean drawImage(Image img, int x, int y, int width, int height, Color bgcolor, ImageObserver observer);
        //    public abstract boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, ImageObserver observer);
        //    public abstract boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, Color bgcolor, ImageObserver observer);

    }

    private void BasicStatTests(Graphics g,String sContextName) {
        /*
          PerformStatTests(g,new Test2DcopyOffscreenToOnscreen(g),sContextName);
          PerformStatTests(g,new Test2DfillRect(g),sContextName);
          PerformStatTests(g,new Test2DdrawRect(g),sContextName);
          PerformStatTests(g,new Test2DclearRect(g),sContextName);
          PerformStatTests(g,new Test2DdrawLine(g),sContextName);
          PerformStatTests(g,new Test2Ddraw3DRect(g),sContextName);
          PerformStatTests(g,new Test2Dfill3DRect(g),sContextName);
          PerformStatTests(g,new Test2DdrawOval(g),sContextName);
          PerformStatTests(g,new Test2DfillOval(g),sContextName);
          PerformStatTests(g,new Test2DdrawRoundRect(g),sContextName);
          PerformStatTests(g,new Test2DfillRoundRect(g),sContextName);
          PerformStatTests(g,new Test2DdrawArc(g),sContextName);
          PerformStatTests(g,new Test2DfillArc(g),sContextName);
          PerformStatTests(g,new Test2DdrawPolyline(g),sContextName);
          PerformStatTests(g,new Test2DdrawPolygon(g),sContextName);
          PerformStatTests(g,new Test2DfillPolygon(g),sContextName);
          */
//          PerformStatTests(g,new Test2DdrawString(g),sContextName);
            PerformStatTests(g,new Test2DsetFont(g),sContextName);
            PerformStatTests(g,new Test2DnewFont(g),sContextName);
            PerformStatTests(g,new Test2DdrawString(g),sContextName);
            PerformStatTests(g,new Test2DdrawChars(g),sContextName);
            PerformStatTests(g,new Test2DdrawBytes(g),sContextName);

//        PerformStatTests(g, new Test2DRotatingDrawImage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DRotatingDrawImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DShearingDrawImage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DShearingDrawImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DScalingDrawImage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DScalingDrawImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//
//        PerformStatTests(g, new Test2DRotatingCopyImage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DRotatingCopyImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DShearingCopyImage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DShearingCopyImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DScalingCopyImage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DScalingCopyImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
//        PerformStatTests(g, new Test2DDrawImage(g,DRAWSIZE,DRAWSIZE),sContextName); 
        /*
          PerformStatTests(g, new Test2DDrawImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
          PerformStatTests(g, new Test2DDrawImageRotate(g,DRAWSIZE,DRAWSIZE),sContextName); 
          PerformStatTests(g, new Test2DDrawImageRotateViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
          PerformStatTests(g, new Test2DDrawImageShear(g,DRAWSIZE,DRAWSIZE),sContextName); 
          PerformStatTests(g, new Test2DDrawImageShearViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 
          PerformStatTests(g, new Test2DDrawImageScale(g,DRAWSIZE,DRAWSIZE),sContextName); 
          PerformStatTests(g, new Test2DDrawImageScaleViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName); 


          PerformStatTests(g, new Test2DCopyImage(g,DRAWSIZE,DRAWSIZE),sContextName);
          PerformStatTests(g, new Test2DCopyImageViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName);
          PerformStatTests(g, new Test2DCopyImageRotate(g,DRAWSIZE,DRAWSIZE),sContextName);
          PerformStatTests(g, new Test2DCopyImageRotateViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName);
          PerformStatTests(g, new Test2DCopyImageShear(g,DRAWSIZE,DRAWSIZE),sContextName);
          PerformStatTests(g, new Test2DCopyImageShearViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName);
          PerformStatTests(g, new Test2DCopyImageScale(g,DRAWSIZE,DRAWSIZE),sContextName);
          PerformStatTests(g, new Test2DCopyImageScaleViaSubimage(g,DRAWSIZE,DRAWSIZE),sContextName);
          */

        //        PerformStatTests(g,new Test2DdrawChars(g),sContextName);
        //        PerformStatTests(g,new Test2DdrawBytes(g),sContextName);
    }
	
    private void AdvancedTests(Graphics2D g2D,String sContextName) {
    }
    private void AdvancedStatTests(Graphics2D g2D,String sContextName) {
    }
  
    private void TestGraphics(Graphics g,String sContextName) {
        BasicTests(g,sContextName);
        try {
            if (g instanceof Graphics2D) {
                AdvancedTests((Graphics2D)g,sContextName);
            }
        } catch (Exception e) {
            //
        } finally {
            //
        } 
    }
    
    private void StatTestGraphics(Graphics g,String sContextName) {
        BasicStatTests(g,sContextName);
        try {
            if (g instanceof Graphics2D) {
                AdvancedStatTests((Graphics2D)g,sContextName);
            }
        } catch (Exception e) {
            //
        } finally {
            //
        }     
    }

    public void start() {
        if (myThread == null) {
            myThread = new Thread(this);
        }
	myThread.setPriority(Thread.MAX_PRIORITY);
        myThread.start();
    }

    public void stop() {
        if (myThread != null) {
            myThread.stop();
            myThread = null;
        }
    }

    public void run() {
	Graphics g = getGraphics();
        System.err.println("start of test");
//        g.setClip(0,0,DRAWSIZE,DRAWSIZE);
        g.setFont(new Font("serif",Font.PLAIN, 12));
        // test onscreen (size of graphics canvas from appletviewer context)
        StatTestGraphics(g,"Onscreen");

        // test offscreen
        Image img = createImage(DRAWSIZE,DRAWSIZE);
        Graphics gOff = img.getGraphics();
        gOff.setFont(new Font("serif",Font.PLAIN, 12));
        StatTestGraphics(gOff,"Offscreen"); 

        System.err.println("end of test");
    }
}


abstract class TestPortion2D {

    protected int theSize;
    private long beginTime,lastTime;

    private int    iterationCounter;

    int getSize() {
        return theSize;
    }
    void setSize(int theSize) {
        this.theSize = theSize;
    }

    protected long timeSync() {
        beginTime = System.currentTimeMillis();
        lastTime = beginTime;
        do {
            lastTime = System.currentTimeMillis();
        } while (lastTime == beginTime);
        beginTime = lastTime;
        return beginTime;
    }

    /*
     *  minimum elapsed time required - prevent intervals which are
     *  too short, and produce erroneous results
     */  

    public int deltaTime() {
        long timeMoment = System.currentTimeMillis();
        return (int)(timeMoment - beginTime);
    }

    protected boolean timedEnough() {
        boolean retval = false;
        long currentTime = System.currentTimeMillis();
        if (currentTime != lastTime) {
            lastTime = currentTime;
            return ((currentTime - beginTime) > 75); 
        }
        return retval;
    }

    protected boolean timedTooMuch() {
        boolean retval = false;
        long currentTime = System.currentTimeMillis();
        if ((currentTime - beginTime) > 60000) {
            retval = true;
        }
        return retval;
    }

    protected long elapsedTime() {
        long retval = lastTime - beginTime;
        if (retval == 0) {
            retval = 1;
        }
        return retval;
    }

    abstract int getNumTests();
    abstract String getTestName(int testNum);
    abstract long MeasureSpeed(int testNum);
    abstract Statistics MeasureStatistics(int testNum);
    abstract String getUnits();
}

class Test2DfillRect extends TestPortion2D {
    private Graphics g;

    Test2DfillRect(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "fillRect";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                if (this.theSize != 0) {
                    g.fillRect(0,0,theSize,theSize);
                }
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics stat = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;
        timeSync();
        boolean exitFlag = false;
        int minIter = 20;
        int i;
        //        long loopstarttime = System.currentTimeMillis();
        //        System.err.println("start time is " + loopstarttime);
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.fillRect(0,0,theSize,theSize);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            //            count++;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        //        long loopendtime = System.currentTimeMillis();
        //        long looptime = loopendtime - loopstarttime ;
        //        System.err.println("time taken for the loop is [" + looptime + "]");

        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.fillRect(0,0,1,1);    // force context allocation -- deferred otherwise
    }
}

class Test2DdrawRect extends TestPortion2D {
    private Graphics g;

    Test2DdrawRect(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "drawRect";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += 4 * theSize;
                g.drawRect(0,0,theSize,theSize);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }
    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawRect(0,0,theSize,theSize);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.drawRect(0,0,1,1);    // force context allocation -- deferred otherwise
    }
}

class Test2DclearRect extends TestPortion2D {
    private Graphics g;

    Test2DclearRect(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "clearRect";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * theSize;
                g.clearRect(0,0,theSize,theSize);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.clearRect(0,0,theSize,theSize);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.clearRect(0,0,1,1);    // force context allocation -- deferred otherwise
    }
}

class Test2DdrawLine extends TestPortion2D {
    private Graphics g;

    Test2DdrawLine(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 5;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        String retval = null;
        switch (testNum) {
        case 0: retval = "drawLine-horizontal"; break;
        case 1: retval = "drawLine-22_degrees"; break;
        case 2: retval = "drawLine-45_degrees"; break;
        case 3: retval = "drawLine-68_degrees"; break;
        case 4: retval = "drawLine-vertical"; break;
        default: retval = "Invalid!"; break;
        }
        return retval;
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize;
                g.drawLine(0,0,theSize,0);
                break;
            case 1:
                pixels += theSize;
                g.drawLine(0,0,theSize,theSize/2);
                break;
            case 2:
                pixels += theSize;
                g.drawLine(0,0,theSize,theSize);
                break;
            case 3:
                pixels += theSize;
                g.drawLine(0,0,theSize/2,theSize);
                break;
            case 4:
                pixels += theSize;
                g.drawLine(0,0,0,theSize);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        switch (testNum) {
        case 1:
        case 3:
            pixels = (long)((double)pixels*Math.sqrt(5)/2);
            break;
        case 2:
            pixels = (long)((double)pixels*Math.sqrt(2)/2);
            break;
        case 0:
        case 4:
        default:
            break;
        }
        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawLine(0,0,theSize,0);
                    }
                    break;
                case 1:
                    if (this.theSize != 0) {
                        g.drawLine(0,0,theSize,theSize/2);
                    }
                    break;
                case 2:
                    if (this.theSize != 0) {
                        g.drawLine(0,0,theSize,theSize);
                    }	
                    break;
                case 3:
                    if (this.theSize != 0) {
                        g.drawLine(0,0,theSize/2,theSize);
                    }
                    break;
                case 4:
                    if (this.theSize != 0) {
                        g.drawLine(0,0,0,theSize);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.drawLine(0,0,1,1);    // force context allocation -- deferred otherwise
    }
}

class Test2Ddraw3DRect extends TestPortion2D {
    private Graphics g;

    Test2Ddraw3DRect(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 2;
    }

    String getUnits() {
        return "edge_pixels_per_second";
    }

    String getTestName(int testNum) {
        String retval = null;
        switch (testNum) {
        case 0: retval = "draw3DRect-nonRaised"; break;
        case 1: retval = "draw3DRect-raised"; break;
        default: retval = "Invalid!"; break;
        }
        return retval;
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * 4;
                g.draw3DRect(0,0,theSize,theSize,false);
                break;
            case 1:
                pixels += theSize * 4;
                g.draw3DRect(0,0,theSize,theSize,true);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.draw3DRect(0,0,theSize,theSize,false);
                    }
                    break;
                case 1:
                    if (this.theSize != 0) {
                        g.draw3DRect(0,0,theSize,theSize,true);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.draw3DRect(0,0,1,1,true);    // force context allocation -- deferred otherwise
    }
}

class Test2Dfill3DRect extends TestPortion2D {
    private Graphics g;

    Test2Dfill3DRect(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 2;
    }

    String getUnits() {
        return "interior_pixels_per_second";
    }

    String getTestName(int testNum) {
        String retval = null;
        switch (testNum) {
        case 0: retval = "fill3DRect-nonRaised"; break;
        case 1: retval = "fill3DRect-raised"; break;
        default: retval = "Invalid!"; break;
        }
        return retval;
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                if (this.theSize != 0) {
                    pixels += theSize * theSize;
                    g.fill3DRect(0,0,theSize,theSize,false);
                }
                break;
            case 1:
                if (this.theSize != 0) {
                    pixels += theSize * theSize;
                    g.fill3DRect(0,0,theSize,theSize,true);
                }
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.fill3DRect(0,0,theSize,theSize,false);
                    }
                    break;
                case 1:
                    if (this.theSize != 0) {
                        g.fill3DRect(0,0,theSize,theSize,true);
                    }	
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.fill3DRect(0,0,1,1,true);    // force context allocation -- deferred otherwise
    }
}

class Test2DfillOval extends TestPortion2D {
    private Graphics g;

    Test2DfillOval(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "fillOval";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * theSize;
                g.fillOval(0,0,theSize,theSize);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)((Math.PI / 4) * ((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.fillOval(0,0,theSize,theSize);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.fillOval(0,0,1,1);    // force context allocation -- deferred otherwise
    }
}

class Test2DdrawOval extends TestPortion2D {
    private Graphics g;

    Test2DdrawOval(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "drawOval";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize;
                g.drawOval(0,0,theSize,theSize);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(Math.PI * ((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawOval(0,0,theSize,theSize);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.drawOval(0,0,1,1);    // force context allocation -- deferred otherwise
    }
}

class Test2DdrawRoundRect extends TestPortion2D {
    private Graphics g;

    Test2DdrawRoundRect(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "drawRoundRect";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * 4;
                g.drawRoundRect(0,0,theSize,theSize,5,5);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawRoundRect(0,0,theSize,theSize,5,5);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.drawRoundRect(0,0,1,1,5,5);    // force context allocation -- deferred otherwise
    }
}

class Test2DfillRoundRect extends TestPortion2D {
    private Graphics g;

    Test2DfillRoundRect(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "fillRoundRect";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * theSize;
                g.fillRoundRect(0,0,theSize,theSize,5,5);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.fillRoundRect(0,0,theSize,theSize,5,5);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.fillRoundRect(0,0,1,1,5,5);    // force context allocation -- deferred otherwise
    }
}

class Test2DdrawArc extends TestPortion2D {
    private Graphics g;

    Test2DdrawArc(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "drawArc";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize;
                g.drawArc(0,0,theSize,theSize,0,120);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        pixels = pixels * 2 + (long)((double)pixels * (2 * Math.PI) / 3);
        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawArc(0,0,theSize,theSize,0,120);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.drawArc(0,0,1,1,0,120);    // force context allocation -- deferred otherwise
    }
}

class Test2DfillArc extends TestPortion2D {
    private Graphics g;

    Test2DfillArc(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "fillArc";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * theSize;
                g.fillArc(0,0,theSize,theSize,0,120);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        pixels = (long)((double)pixels * Math.PI / 3);
        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.fillArc(0,0,theSize,theSize,0,120);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.fillArc(0,0,1,1,0,120);    // force context allocation -- deferred otherwise
    }
}

class Test2DdrawPolyline extends TestPortion2D {
    private Graphics g;

    Test2DdrawPolyline(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "drawPolyline";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize;
                g.drawPolyline(xPoints,yPoints,4);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)((double)pixels * 3 * Math.sqrt(2) / 2);
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawPolyline(xPoints,yPoints,4);
                    }
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};

        g.drawPolyline(xPoints,yPoints,4);  // force context allocation -- deferred otherwise
    }
}


class Test2DdrawPolygon extends TestPortion2D {
    private Graphics g;

    Test2DdrawPolygon(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "drawPolygon";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize;
                g.drawPolygon(xPoints,yPoints,4);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)((double)pixels * 2 * Math.sqrt(2));
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawPolygon(xPoints,yPoints,4);
                    }
                    break;
                default:
                    break;
                }
            } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};

        g.drawPolygon(xPoints,yPoints,4);  // force context allocation -- deferred otherwise
    }
}

class Test2DfillPolygon extends TestPortion2D {
    private Graphics g;

    Test2DfillPolygon(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        return "fillPolygon";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * theSize;
                g.fillPolygon(xPoints,yPoints,4);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        pixels = pixels / 2;
        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.fillPolygon(xPoints,yPoints,4);
                    }
                    break;
                default:
                    break;
                }
            } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};

        g.fillPolygon(xPoints,yPoints,4);  // force context allocation -- deferred otherwise
    }
}

class Test2DsetFont extends TestPortion2D {
    private Graphics g;

    Test2DsetFont(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "fonts_per_second";
    }

    String getTestName(int testNum) {
        return "setFont";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long chars = 0;
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyx";
        String theString = srcStr.substring(0,(theSize!=0)? theSize: 1);

        long beginTime = timeSync();
        do {
            switch (testNum) {
            case 0:
                chars += theSize;
                Font f = new Font("Dialog", Font.BOLD, 10);
                g.setFont(f);
                f = new Font("Dialog", Font.PLAIN, 10);
                g.setFont(f);
                // Exception e = new Exception("trace");
                //e.printStackTrace();
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return chars;
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
        String theString = srcStr.substring(0,(theSize!=0)? theSize: 1);
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 10;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        Font f = new Font("Dialog", Font.BOLD, 10);
                        g.setFont(f);
                        f = new Font("Dialog", Font.PLAIN, 10);
                        g.setFont(f);
                    }
                    break;
                default:
                    break;
                }
            } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        if (theSize > 50) {
            theSize = 50;
        }
        super.setSize(theSize);
        Font f = new Font("Dialog", Font.BOLD, 10);
        g.setFont(f);
        f = new Font("Dialog", Font.PLAIN, 10);
        g.setFont(f);
    }
}

class Test2DnewFont extends TestPortion2D {
    private Graphics g;

    Test2DnewFont(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "fonts_per_second";
    }

    String getTestName(int testNum) {
        return "newFont";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long chars = 0;
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyx";
        String theString = srcStr.substring(0,(theSize!=0)? theSize: 1);

        long beginTime = timeSync();
        do {
            switch (testNum) {
            case 0:
                chars += theSize;
                Font f = new Font("Dialog", Font.BOLD, 10);
                // Exception e = new Exception("trace");
                //e.printStackTrace();
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return chars;
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
        String theString = srcStr.substring(0,(theSize!=0)? theSize: 1);
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 10;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        Font f = new Font("Dialog", Font.BOLD, 10);
                    }
                    break;
                default:
                    break;
                }
            } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        if (theSize > 50) {
            theSize = 50;
        }
        super.setSize(theSize);
        Font f = new Font("Dialog", Font.BOLD, 10);
    }
}

class Test2DdrawString extends TestPortion2D {
    private Graphics g;

    Test2DdrawString(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "chars_per_second";
    }

    String getTestName(int testNum) {
        return "drawString";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long chars = 0;
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyx";
        String theString = srcStr.substring(0,(theSize!=0)? theSize: 1);

        long beginTime = timeSync();
        do {
            switch (testNum) {
            case 0:
                chars += theSize;
                g.drawString(theString,0,50);
                // Exception e = new Exception("trace");
                //e.printStackTrace();
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return chars;
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
        String theString = srcStr.substring(0,(theSize!=0)? theSize: 1);
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 10;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    if (this.theSize != 0) {
                        g.drawString(theString,0,50);
                    }
                    break;
                default:
                    break;
                }
            } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        if (theSize > 50) {
            theSize = 50;
        }
        super.setSize(theSize);
        g.drawString("",0,0);  // force context allocation -- deferred otherwise
    }
}

class Test2DdrawChars extends TestPortion2D {
    private Graphics g;

    Test2DdrawChars(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "chars_per_second";
    }

    String getTestName(int testNum) {
        return "drawChars";
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long chars = 0;
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyx";
        String theString = srcStr.substring(0,theSize);
        char theChars[] = theString.toCharArray();

        long beginTime = timeSync();
        do {
            switch (testNum) {
            case 0:
                chars += theSize;
                g.drawChars(theChars,0,theSize,0,50);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return chars;
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyx";
        String theString = srcStr.substring(0,theSize);
        char theChars[] = theString.toCharArray();
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 1;
        int i;
        do {
            //          for (i=0;i<minIter;i++) {
            switch (testNum) {
            case 0:
                if (this.theSize != 0) {
                    g.drawChars(theChars,0,theSize,0,50);
                }
                break;
            default:
                break;
            }
            //            } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        if (theSize > 50) {
            theSize = 50;
        }
        super.setSize(theSize);
        g.drawChars("".toCharArray(),0,0,0,0);  // force context allocation -- deferred otherwise
    }
}

class Test2DdrawBytes extends TestPortion2D {
    private Graphics g;

    Test2DdrawBytes(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "chars_per_second";
    }

    String getTestName(int testNum) {
        return "drawBytes";
    }

    private byte[] toByteArray(String theString) {
	int i, max = theString.length();
	byte result[] = new byte[max];
	theString.getBytes(0, max, result, 0);
	return result;
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long chars = 0;
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyx";
        String theString = srcStr.substring(0,theSize);
        byte theBytes[] = toByteArray(theString);

        long beginTime = timeSync();
        do {
            switch (testNum) {
            case 0:
                chars += theSize;
                g.drawBytes(theBytes,0,theSize,0,50);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return chars;
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        // REMIND: will want string which represents distribution of pairs in english
        String srcStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyx";
        String theString = srcStr.substring(0,theSize);
        byte theBytes[] = toByteArray(theString);
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 1;
        int i;
        do {
            //         for (i=0;i<minIter;i++) {
            switch (testNum) {
            case 0:
                if (this.theSize != 0) {
                    g.drawBytes(theBytes,0,theSize,0,50);
                }
                break;
            default:
                break;
            }
            //          } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        byte theBytes[] = {65};
        if (theSize > 50) {
            theSize = 50;
        }
        super.setSize(theSize);
        g.drawBytes(theBytes,0,0,0,0);  // force context allocation -- deferred otherwise
    }
}

class Test2DcopyArea extends TestPortion2D {
    private Graphics g;

    Test2DcopyArea(Graphics g) {
        this.theSize = 10;
        this.g = g;
    }

    int getNumTests() {
        return 3;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    String getTestName(int testNum) {
        String retval = null;
        switch (testNum) {
        case 0: retval = "copyArea-noOverlap"; break;
        case 1: retval = "copyArea-leftOverlap"; break;
        case 2: retval = "copyArea-rightOverlap"; break;
        default: retval = "Invalid!"; break;
        }
        return retval;
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;
        timeSync();

        do {
            switch (testNum) {
            case 0:
                pixels += theSize * theSize;
                // remind - degenerates to clipped when area too big
                g.copyArea(0,0,theSize,theSize,theSize,theSize);
                break;
            case 1:
                pixels += theSize * theSize;
                g.copyArea(0,0,theSize,theSize,10,10);
                break;
            case 2:
                pixels += theSize * theSize;
                g.copyArea(10,10,theSize,theSize,0,0);
                break;
            default:
                break;
            }
        } while (!timedEnough());

        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics  stat    = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        timeSync();
        boolean exitFlag = false;
        int minIter = 30;
        int i;
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    // remind - degenerates to clipped when area too big
                    if (this.theSize != 0) {
                        g.copyArea(0,0,theSize,theSize,theSize,theSize);
                    }
                    break;
                case 1:
                    if (this.theSize != 0) {
                        g.copyArea(0,0,theSize,theSize,10,10);
                    }
                    break;
                case 2:
                    if (this.theSize != 0) {
                        g.copyArea(10,10,theSize,theSize,0,0);
                    }
                    break;
                default:
                    break;
                }
            } 
            count+=minIter;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        return stat;
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        g.copyArea(0,0,1,1,1,1);    // force context allocation -- deferred otherwise
        int xPoints[] = {theSize/2, theSize, theSize/2, 0};
        int yPoints[] = {0, theSize/2, theSize, theSize/2};

        g.drawPolygon(xPoints,yPoints,4);  // force context allocation -- deferred otherwise
    }
}

//======================
//
//======================
abstract class TestPortion2DDraw extends TestPortion2D {
    Color colorRed;
    Color colorGreen;
    Color colorBlue;
    Color colorWhite;
    Color colorBlack;
    Color savedColor;

    //------------- begin group
    //  Draw primitives
    void drawBar(Graphics grpx,int x,int y,int h, int v,Color c)
    {
        savedColor = grpx.getColor();

        grpx.setColor(c);
		
        grpx.fillRect(x,y,h,v);

        grpx.setColor(savedColor);
    }
    void drawRect(Graphics grpx,int x,int y,int h, int v,Color c)
    {
        savedColor = grpx.getColor();

        grpx.setColor(c);
		
        grpx.drawRect(x,y,h,v);

        grpx.setColor(savedColor);
    }
    void drawGrid(Graphics grpx,int x,int y,int h, int v,int d,Color c)
    {
        savedColor = grpx.getColor();

        grpx.setColor(c);

        if( 0 ==d ){
            d = 1;
        }
		
        {
            int		i;

            for(i=y;i<y+v;i+=d)
                {
                    grpx.drawLine(x,i,x+h,i);
                }
            for(i=x;i<x+v;i+=d)
                {
                    grpx.drawLine(i,y,i,y+v);
                }
        }
        grpx.setColor(savedColor);
    }
    void cleanupArea(Graphics grpx,int x,int y,int dx, int dy){
        drawBar(grpx,x,y,dx,dy,colorWhite);
    }
    void drawArea(Graphics grpx,int x,int y,int dx, int dy){
        if( null != grpx )
            {
                int         minDX = dx / 30;
                int         minDY = dy / 30;
                int         maxDX = dx / 3;
                int         maxDY = dy / 3;
                int         midDX = dx / 2;
                int         midDY = dy / 2;

                drawBar(grpx,x,y,dx,dy,colorWhite);
                drawBar(grpx,x+minDX,y+minDY,maxDX,maxDY,colorRed);
                drawBar(grpx,x+midDX + minDX,y+minDY,maxDX,maxDY,colorGreen);
                drawBar(grpx,x+midDX - 6*minDX,y+midDY,maxDX,maxDY,colorBlue);
                drawGrid(grpx,x,y,dx,dy,minDX,colorBlack);
            }
    }

    //------------- end group
    void Initialize() {
        colorRed		= new Color(0xFF,0x00,0x00);
        colorGreen		= new Color(0x00,0xFF,0x00);
        colorBlue		= new Color(0x00,0x00,0xFF);
        colorWhite		= new Color(0xFF,0xFF,0xFF);
        colorBlack		= new Color(0x00,0x00,0x00);
    }
    
}

abstract class TestRasterOperations extends TestPortion2DDraw {
    private AffineTransform savedTransform;
    public  int             sizeX,sizeY;
    private Graphics2D      theScreen;
    
    abstract void testAct();

    final Graphics2D getScreen(){
        return this.theScreen;
    }

    void beforeLoop(){
        cleanupArea(getScreen(),0,0,this.sizeX,this.sizeY);
        savedTransform = getScreen().getTransform();
    }

    void afterLoop(){
        getScreen().setTransform(savedTransform);
        cleanupArea(getScreen(),0,0,this.sizeX,this.sizeY);
    }

    void Initialize(Graphics g,int dx,int dy){
        super.Initialize();
        this.theScreen = (Graphics2D)g;
        sizeX = dx;
        sizeY = dy;
    }

    // returns pixels/second
    long MeasureSpeed(int testNum) {
        long pixels = 0;

        beforeLoop();
        timeSync();

        do {
            switch (testNum) {
            case 0:
                testAct();
                break;
            default:
                break;
            }
        } while (!timedEnough());
        afterLoop();
        return (long)(((double)pixels*1000)/(double)elapsedTime());
    }

    // returns statistics
    Statistics MeasureStatistics(int testNum) {
        Statistics stat = new Statistics(getTestName(testNum),"time, msec");
        double count = 0;

        boolean exitFlag = false;
        int minIter = 20;
        int i;

        beforeLoop();
        timeSync();
        //        long loopstarttime = System.currentTimeMillis();
        //        System.err.println("start time is " + loopstarttime);
        do {
            for (i=0;i<minIter;i++) {
                switch (testNum) {
                case 0:
                    testAct();
                    //                            exitFlag = true;
                    break;
                default:
                    break;
                }
            }
            count+=minIter;
            //            count++;
            stat.newExperimentUpdate(count,deltaTime());
            exitFlag = timedEnough();
            exitFlag &= !stat.needMoreExperiments();
            exitFlag |= timedTooMuch();
            //            System.out.println("size["+this.theSize+"] corr["+stat.getCorrelationCoefficient() + "] slope ["+stat.getSlope()+"] time["+elapsedTime()+"]");
        } while(!exitFlag);
        //        long loopendtime = System.currentTimeMillis();
        //        long looptime = loopendtime - loopstarttime ;
        //        System.err.println("time taken for the loop is [" + looptime + "]");

        afterLoop();
        return stat;
    }

    int getNumTests() {
        return 1;
    }

    String getUnits() {
        return "pixels_per_second";
    }

    void setSize(int theSize) {
        super.setSize(theSize);
        getScreen().fillRect(0,0,1,1);    // force context allocation -- deferred otherwise
    }
}

class Test2DCopyImage extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }
    public Test2DCopyImage(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }
    public Test2DCopyImage(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }

    String getTestName(int testNum) {
        return "CopyImage";
    }

    void testAct()
    {
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        if (this.theSize != 0) {
            getScreen().drawImage(srcOffscreen,x,y,this.theSize,this.theSize,null);
        }
    }
}

class Test2DCopyImageViaSubimage extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }


    public Test2DCopyImageViaSubimage(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
    }
    public Test2DCopyImageViaSubimage(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "CopyImage, via subimage";
    }

    void testAct()
    {
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        prepareSrcOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.srcOffscreen.getSubimage(0,0,this.theSize,this.theSize);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,x,y,this.theSize,this.theSize,null);
            }
        }
    }
}

class Test2DCopyImageRotate extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;
    private AffineTransform theTransform;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180); 
        ((Graphics2D)this.getScreen()).transform(this.theTransform);
    }

    public Test2DCopyImageRotate(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }
    public Test2DCopyImageRotate(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }

    String getTestName(int testNum) {
        return "CopyImage, rotating";
    }

    void testAct()
    {
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        if (this.theSize != 0) {
            getScreen().drawImage(srcOffscreen,x,y,this.theSize,this.theSize,null);
        }
    }
}
class Test2DCopyImageRotateViaSubimage extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;
    private AffineTransform theTransform;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180); 
        ((Graphics2D)this.getScreen()).transform(this.theTransform);
    }

    public Test2DCopyImageRotateViaSubimage(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
    }
    public Test2DCopyImageRotateViaSubimage(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "CopyImage, rotating, subimage";
    }

    void testAct()
    {
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        prepareSrcOffscreen(0,0,this.sizeX,this.sizeY);
        {
            BufferedImage   theSection = srcOffscreen.getSubimage(0,0,this.sizeX,this.sizeY);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,x,y,this.theSize,this.theSize,null);
            }
        }
    }
}
class Test2DCopyImageShear extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;
    private AffineTransform theTransform;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1); 
        ((Graphics2D)this.getScreen()).transform(this.theTransform);
    }
    public Test2DCopyImageShear(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }
    public Test2DCopyImageShear(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }

    String getTestName(int testNum) {
        return "CopyImage, shearing";
    }

    void testAct()
    {
        if (this.theSize != 0) {
            getScreen().drawImage(srcOffscreen,0,0,this.theSize,this.theSize,null);
        }
    }
}
class Test2DCopyImageShearViaSubimage extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;
    private AffineTransform theTransform;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1); 
        ((Graphics2D)this.getScreen()).transform(this.theTransform);
    }
    public Test2DCopyImageShearViaSubimage(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
    }
    public Test2DCopyImageShearViaSubimage(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "CopyImage, shearing, subimage";
    }

    void testAct()
    {
        prepareSrcOffscreen(0,0,this.sizeX,this.sizeY);
        {
            BufferedImage   theSection = srcOffscreen.getSubimage(0,0,this.sizeX,this.sizeY);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
}
class Test2DCopyImageScale extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;
    private AffineTransform theTransform;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.5,0.5); 
        getScreen().transform(this.theTransform);
    }
    public Test2DCopyImageScale(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }
    public Test2DCopyImageScale(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;

        prepareSrcOffscreen(0,0,dx,dy);
    }

    String getTestName(int testNum) {
        return "CopyImage, scaling";
    }

    void testAct()
    {
        int     x = (this.sizeX-this.theSize/2)/2;
        int     y = (this.sizeY-this.theSize/2)/2;
        if (this.theSize != 0) {
            getScreen().drawImage(srcOffscreen,x,y,this.theSize,this.theSize,null);
        }
    }
}
class Test2DCopyImageScaleViaSubimage extends TestRasterOperations {
    private BufferedImage   srcOffscreen;
    private BufferedImage   dstOffscreen;
    private AffineTransform theTransform;

    private void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.5,0.5); 
        getScreen().transform(this.theTransform);
    }

    public Test2DCopyImageScaleViaSubimage(Graphics g,int srcType,int dstType,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
    }
    public Test2DCopyImageScaleViaSubimage(Graphics g,int dx,int dy) {

        this.dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        this.srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "CopyImage, scaling, subimage";
    }

    void testAct()
    {
        int     x = (this.sizeX-this.theSize/2)/2;
        int     y = (this.sizeY-this.theSize/2)/2;
        prepareSrcOffscreen(0,0,this.sizeX,this.sizeY);
        {
            BufferedImage   theSection = srcOffscreen.getSubimage(0,0,this.sizeX,this.sizeY);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,x,y,this.theSize,this.theSize,null);
            }
        }
    }
}


class Test2DDrawImage extends TestRasterOperations {
    private BufferedImage   offscreen;

    private void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    Test2DDrawImage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(20,20,BufferedImage.TYPE_INT_RGB);
        this.theSize = 20;
        prepareOffscreen(0,0,dx,dy);
    }
    Test2DDrawImage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(20,20,imageType);
        this.theSize = 20;
        prepareOffscreen(0,0,dx,dy);
    }

    String getTestName(int testNum) {
        return "drawImage";
    }

    void testAct(){
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        if (this.theSize != 0) {
            getScreen().drawImage(this.offscreen,x,y,null);
        }
    }
}
class Test2DDrawImageViaSubimage extends TestRasterOperations {
    private BufferedImage   offscreen;

    private void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    Test2DDrawImageViaSubimage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }
    Test2DDrawImageViaSubimage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "drawImage, subimage";
    }

    void testAct(){
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        prepareOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage    theSection = this.offscreen.getSubimage(0,0,this.theSize,this.theSize);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,x,y,this.theSize,this.theSize,null);
            }
        }
    }
}

class Test2DDrawImageRotate extends TestRasterOperations {
    private BufferedImage   offscreen;
    private AffineTransform theTransform;

    private void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180); // radians=degree * pie / 180
        getScreen().transform(theTransform);
    }
    Test2DDrawImageRotate(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;

        prepareOffscreen(0,0,dx,dy);
    }
    Test2DDrawImageRotate(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;

        prepareOffscreen(0,0,dx,dy);
    }

    String getTestName(int testNum) {
        return "drawImage, rotating";
    }

    void testAct(){
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        if (this.theSize != 0) {
            getScreen().drawImage(this.offscreen,x,y,this.theSize,this.theSize,null);
        }
    }
}
class Test2DDrawImageRotateViaSubimage extends TestRasterOperations {
    private BufferedImage   offscreen;
    private AffineTransform theTransform;

    private void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }
    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180); // radians=degree * pie / 180
        getScreen().transform(theTransform);
    }
    Test2DDrawImageRotateViaSubimage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }
    Test2DDrawImageRotateViaSubimage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "drawImage, rotating, subimage";
    }

    void testAct(){
        int     x = (this.sizeX-this.theSize)/2;
        int     y = (this.sizeY-this.theSize)/2;
        prepareOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.offscreen.getSubimage(0,0,this.theSize,this.theSize);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,x,y,this.theSize,this.theSize,null);
            }
        }
    }
}
class Test2DDrawImageShear extends TestRasterOperations {
    private BufferedImage   offscreen;
    private AffineTransform theTransform;

    private void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }
    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1);
        getScreen().transform(theTransform);
    }
    Test2DDrawImageShear(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;

        prepareOffscreen(0,0,dx,dy);
    }
    Test2DDrawImageShear(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;

        prepareOffscreen(0,0,dx,dy);

    }

    String getTestName(int testNum) {
        return "drawImage, shearing";
    }

    void testAct(){
        if (this.theSize != 0) {
            this.getScreen().drawImage(this.offscreen,0,0,this.theSize,this.theSize,null);
        }
    }
}
class Test2DDrawImageShearViaSubimage extends TestRasterOperations {
    private BufferedImage   offscreen;
    private AffineTransform theTransform;

    private void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1);
        getScreen().transform(theTransform);
    }
    Test2DDrawImageShearViaSubimage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }
    Test2DDrawImageShearViaSubimage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "drawImage, shearing, subimage";
    }

    void testAct(){
        prepareOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.offscreen.getSubimage(0,0,this.theSize,this.theSize);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
}

class Test2DDrawImageScale extends TestRasterOperations {
    private BufferedImage   offscreen;
    private AffineTransform theTransform;

    private void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }
    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.5,0.5);
        getScreen().transform(theTransform);
    }
    Test2DDrawImageScale(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        this.offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;

        prepareOffscreen(0,0,dx,dy);
    }
    Test2DDrawImageScale(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;

        prepareOffscreen(0,0,dx,dy);

    }

    String getTestName(int testNum) {
        return "drawImage, scaling";
    }

    void testAct(){
        int     x = (this.sizeX-this.theSize/2)/2;
        int     y = (this.sizeY-this.theSize/2)/2;
        if (this.theSize != 0) {
            getScreen().drawImage(this.offscreen,x,y,this.theSize,this.theSize,null);
        }
    }
}
class Test2DDrawImageScaleViaSubimage extends TestRasterOperations {
    private BufferedImage   offscreen;
    private AffineTransform theTransform;

    void prepareOffscreen(int x,int y,int dx,int dy){
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }
    void beforeLoop(){
        super.beforeLoop();
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.5,0.5);
        getScreen().transform(theTransform);
    }
    Test2DDrawImageScaleViaSubimage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
    }
    Test2DDrawImageScaleViaSubimage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
    }

    String getTestName(int testNum) {
        return "drawImage, scaling, subimage";
    }

    void testAct(){
        int     x = (this.sizeX-this.theSize/2)/2;
        int     y = (this.sizeY-this.theSize/2)/2;
        prepareOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.offscreen.getSubimage(0,0,this.theSize,this.theSize);
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,x,y,this.theSize,this.theSize,null);
            }
        }
    }
}


abstract class Test2DTransformingRasterOperations extends TestRasterOperations {
    BufferedImage               offscreen;
    AffineTransform             theTransform;
    AffineTransform             theReverseTransform;
    private int                 theStepCount;
    private int                 theStepLimit;
    private int                 theStepSwitch;

    abstract void               generateTransform           ();
    abstract void               generateReverseTransform    ();
    void setTransform() {
        if( 0 == this.theStepSwitch ) {
            getScreen().transform(this.theTransform);
        } else {
            getScreen().transform(this.theReverseTransform);
        }
        if( this.theStepCount < this.theStepLimit / 2 ) {
            this.theStepSwitch = 0;    
        } else {
            this.theStepSwitch = 1;    
        }

        if( this.theStepCount > this.theStepLimit ) {
            this.theStepCount = 0;
        } else {
            this.theStepCount ++;
        }
    }
    
    void setLimit(int limit) {
        this.theStepCount = 0;
        this.theStepSwitch = 0;
        this.theStepLimit = limit;
    }

    void prepareOffscreen(int x,int y,int dx,int dy) {
        if( null != offscreen )
            {
                Graphics    grpx = offscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }

    void beforeLoop(){
        super.beforeLoop();
        generateTransform();
        generateReverseTransform();
    }
    
}

class Test2DRotatingDrawImage extends Test2DTransformingRasterOperations {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180); // radians=degree * pie / 180
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.rotate((-10*java.lang.Math.PI)/180); // radians=degree * pie / 180
    }
    Test2DRotatingDrawImage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    Test2DRotatingDrawImage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    void testAct(){
        {
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(this.offscreen,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "drawImage, in constant rotating";
    }
}
class Test2DRotatingDrawImageViaSubimage extends Test2DTransformingRasterOperations {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180);   
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.rotate((-10*java.lang.Math.PI)/180);   
    }
    Test2DRotatingDrawImageViaSubimage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        setLimit(10);
    }
    Test2DRotatingDrawImageViaSubimage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
        setLimit(10);
    }
    void testAct(){
        prepareOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.offscreen.getSubimage(0,0,this.theSize,this.theSize);
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "drawImage, subimage, in constant rotating";
    }
}

class Test2DShearingDrawImage extends Test2DTransformingRasterOperations {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1);
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.shear(-0.1,-0.1); 
    }
    Test2DShearingDrawImage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    Test2DShearingDrawImage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    void testAct(){
        {
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(this.offscreen,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "drawImage, in constant shearing";
    }
}
class Test2DShearingDrawImageViaSubimage extends Test2DTransformingRasterOperations {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1);   
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.shear(-0.1,-0.1);   
    }
    Test2DShearingDrawImageViaSubimage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        setLimit(10);
    }
    Test2DShearingDrawImageViaSubimage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
        setLimit(10);
    }
    void testAct(){
        prepareOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.offscreen.getSubimage(0,0,this.theSize,this.theSize);
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "drawImage, subimage, in constant shearing";
    }
}

class Test2DScalingDrawImage extends Test2DTransformingRasterOperations {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.75,0.75);
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.scale(-0.75,-0.75); 
    }
    Test2DScalingDrawImage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    Test2DScalingDrawImage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    void testAct(){
        {
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(this.offscreen,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "drawImage, in constant scaling";
    }
}

class Test2DScalingDrawImageViaSubimage extends Test2DTransformingRasterOperations {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.75,0.75);   
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.scale(-0.75,-0.75);   
    }
    Test2DScalingDrawImageViaSubimage(Graphics g,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        setLimit(10);
    }
    Test2DScalingDrawImageViaSubimage(Graphics g,int imageType,int dx,int dy) {
        super.Initialize(g,dx,dy);
        offscreen = new BufferedImage(dx,dy,imageType);
        this.theSize = 10;
        setLimit(10);
    }
    void testAct(){
        prepareOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.offscreen.getSubimage(0,0,this.theSize,this.theSize);
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "drawImage, subimage, in constant scaling";
    }
}

abstract class Test2DTransformingRasterOperationsOffscreen extends Test2DTransformingRasterOperations {
    BufferedImage   srcOffscreen;
    BufferedImage   dstOffscreen;

    void prepareSrcOffscreen(int x,int y,int dx,int dy){
        if( null != this.srcOffscreen )
            {
                Graphics    grpx = srcOffscreen.getGraphics();
                drawArea(grpx,x,y,dx,dy);
            }
    }
}

class Test2DRotatingCopyImage extends Test2DTransformingRasterOperationsOffscreen {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180); // radians=degree * pie / 180
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.rotate((-10*java.lang.Math.PI)/180); // radians=degree * pie / 180
    }
    Test2DRotatingCopyImage(Graphics g,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    Test2DRotatingCopyImage(Graphics g,int srcType,int dstType,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
        prepareOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    void testAct(){
        {
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(this.srcOffscreen,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "Copy Image, in constant rotating";
    }
}
class Test2DRotatingCopyImageViaSubimage extends Test2DTransformingRasterOperationsOffscreen {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.rotate((10*java.lang.Math.PI)/180);   
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.rotate((-10*java.lang.Math.PI)/180);   
    }
    Test2DRotatingCopyImageViaSubimage(Graphics g,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        setLimit(10);
    }
    Test2DRotatingCopyImageViaSubimage(Graphics g,int srcType,int dstType,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
        setLimit(10);
    }
    void testAct(){
        prepareSrcOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.srcOffscreen.getSubimage(0,0,this.theSize,this.theSize);
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "Copy Image, subimage, in constant rotating";
    }
}

class Test2DShearingCopyImage extends Test2DTransformingRasterOperationsOffscreen {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1);
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.shear(-0.1,-0.1); 
    }
    Test2DShearingCopyImage(Graphics g,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        prepareSrcOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    Test2DShearingCopyImage(Graphics g,int srcType,int dstType,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
        prepareSrcOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    void testAct(){
        {
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(this.srcOffscreen,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "Copy Image, in constant shearing";
    }
}
class Test2DShearingCopyImageViaSubimage extends Test2DTransformingRasterOperationsOffscreen {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.shear(0.1,0.1);   
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.shear(-0.1,-0.1);   
    }
    Test2DShearingCopyImageViaSubimage(Graphics g,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        setLimit(10);
    }
    Test2DShearingCopyImageViaSubimage(Graphics g,int srcType,int dstType,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
        setLimit(10);
    }
    void testAct(){
        prepareSrcOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.srcOffscreen.getSubimage(0,0,this.theSize,this.theSize);
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "Copy Image, subimage, in constant shearing";
    }
}

class Test2DScalingCopyImage extends Test2DTransformingRasterOperationsOffscreen {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.75,0.75);
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.scale(-0.75,-0.75); 
    }
    Test2DScalingCopyImage(Graphics g,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        prepareSrcOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    Test2DScalingCopyImage(Graphics g,int srcType,int dstType,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
        prepareSrcOffscreen(0,0,this.sizeX,this.sizeY);
        setLimit(10);
    }
    void testAct(){
        {
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(this.srcOffscreen,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "Copy Image, in constant scaling";
    }
}

class Test2DScalingCopyImageViaSubimage extends Test2DTransformingRasterOperationsOffscreen {
    void generateTransform() {
        this.theTransform = new AffineTransform();
        this.theTransform.scale(0.75,0.75);   
    }
    void generateReverseTransform() {
        this.theReverseTransform = new AffineTransform();
        this.theReverseTransform.scale(-0.75,-0.75);   
    }
    Test2DScalingCopyImageViaSubimage(Graphics g,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,BufferedImage.TYPE_INT_RGB);
        this.theSize = 10;
        setLimit(10);
    }
    Test2DScalingCopyImageViaSubimage(Graphics g,int srcType,int dstType,int dx,int dy) {
        dstOffscreen = new BufferedImage(dx,dy,dstType);
        super.Initialize(dstOffscreen.getGraphics(),dx,dy);
        srcOffscreen = new BufferedImage(dx,dy,srcType);
        this.theSize = 10;
        setLimit(10);
    }
    void testAct(){
        prepareSrcOffscreen(0,0,this.theSize,this.theSize);
        {
            BufferedImage   theSection = this.srcOffscreen.getSubimage(0,0,this.theSize,this.theSize);
            setTransform();
            if (this.theSize != 0) {
                getScreen().drawImage(theSection,0,0,this.theSize,this.theSize,null);
            }
        }
    }
    String getTestName(int testNum) {
        return "Copy Image, subimage, in constant scaling";
    }
}



//    public abstract boolean drawImage(Image img, int x, int y, ImageObserver observer);
//    public abstract boolean drawImage(Image img, int x, int y, int width, int height, ImageObserver observer);
//    public abstract boolean drawImage(Image img, int x, int y, Color bgcolor, ImageObserver observer);
//    public abstract boolean drawImage(Image img, int x, int y, int width, int height, Color bgcolor, ImageObserver observer);
//    public abstract boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, ImageObserver observer);
//    public abstract boolean drawImage(Image img, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2, Color bgcolor, ImageObserver observer);


//    public abstract void translate(int x, int y);
//    public abstract Color getColor();
//    public abstract void setColor(Color c);
//    public abstract void setPaintMode();
//    public abstract void setXORMode(Color c1);
//    public abstract Font getFont();
//    public abstract void setFont(Font font);
//    public abstract FontMetrics getFontMetrics(Font f);
//    public abstract Rectangle getClipBounds();
//    public abstract void clipRect(int x, int y, int width, int height);
//    public abstract void setClip(int x, int y, int width, int height);
//    public abstract Shape getClip();
//    public abstract void setClip(Shape clip);
//    public Rectangle getClipRect();
