/*
 * @(#)PrintShapes.java	1.1 98/09/14
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.awt.*;
import java.awt.geom.*;
import java.awt.print.*;
import java.util.*;

public class PrintShapes implements Printable {

    private static final int INCH = 72;
    private static final float LINE_HEIGHT = (float) (INCH * 0.5);

    Vector mShapes = new Vector();
    Stroke mDash;
    Polygon mPoly;

    public static void main(String[] args) {
	PrinterJob pj = PrinterJob.getPrinterJob();
	
	pj.setPrintable(new PrintShapes());
	long startTime = System.currentTimeMillis();

	if (pj.printDialog()) {

	    try {
		pj.print();
	    } catch (PrinterException e) {
		System.out.println("Printer error: " + e);
	    }
	    long endTime = System.currentTimeMillis();
	    System.out.println("Print Time: " + ((endTime - startTime) / 1000.0));
	}
    }

    public PrintShapes() {
	mPoly = new Polygon();
	mPoly.addPoint(0, 0);
	mPoly.addPoint(20, 5);
	mPoly.addPoint(30, 15);
	mPoly.addPoint(20, 25);
	mPoly.addPoint(0, 30);
	mPoly.addPoint(0, 0);
	mShapes.add(mPoly);

	mShapes.add(new Rectangle2D.Double(0, 0, 30, 30));

	mShapes.add(new RoundRectangle2D.Double(0, 0, 30, 30, 10, 10));

	mShapes.add(new Arc2D.Double(0, 0, 30, 30, 0, 60, Arc2D.PIE));

	mShapes.add(new Ellipse2D.Double(0, 0, 30, 20));

	mShapes.add(new CubicCurve2D.Double(0, 0, 10, 30, 20, 30, 30, 0));

	mShapes.add(new QuadCurve2D.Double(0, 0, 15, 30, 30, 0));
	
	float[] dash = new float[4];
	dash[0] = 7;
	dash[1] = 5;
	dash[2] = 5;
	dash[3] = 7;

	mDash = new BasicStroke(5, BasicStroke.CAP_ROUND,
				BasicStroke.JOIN_ROUND, 10, dash, 0);
    }


    public int print(Graphics g, PageFormat pageFormat, int pageIndex) {
	int result;

	if (pageIndex == 0) {
	    Graphics2D g2 = (Graphics2D) g;
	    
	    Rectangle2D.Double imageRect = new Rectangle2D.Double(pageFormat.getImageableX(),
						    pageFormat.getImageableY(),
						    pageFormat.getImageableWidth(),
						    pageFormat.getImageableHeight());

	    g2.translate(imageRect.getX(), imageRect.getY());
	    g2.setPaint(Color.black);

	    Iterator i = mShapes.iterator();
	    while (i.hasNext()) {
		Graphics2D tempGraphics;
		Shape shape = (Shape) i.next();

		tempGraphics = (Graphics2D) g2.create();
		tempGraphics.draw(shape);
		tempGraphics.translate(INCH, 0);
		tempGraphics.fill(shape);
		tempGraphics.translate(INCH, 0);
		tempGraphics.setPaint(Color.red);
		tempGraphics.fill(shape);
		tempGraphics.translate(INCH, 0);
		tempGraphics.setPaint(Color.green);
		tempGraphics.setStroke(mDash);
		tempGraphics.draw(shape);
		tempGraphics.translate(INCH, 0);

		tempGraphics.dispose();

		g2.translate(0, LINE_HEIGHT);

	    }	    	    

	    Color saved = g2.getColor();
	    g2.setColor(Color.gray);
	    g2.draw3DRect(0, 0, 30, 30, true);
	    g2.fill3DRect(INCH, 0, 30, 30, true);
	    g2.setColor(saved);

	    g2.drawArc(INCH * 2, 0, 30, 30, 120, 260);
	    g2.fillArc(INCH * 3, 0, 30, 30, 120, 260);

	    g2.drawLine(INCH * 4, 0, INCH * 4 + 30, 30);

	    g2.translate(0, LINE_HEIGHT);

	    g2.drawOval(0, 0, 30, 30);
	    g2.fillOval(INCH, 0, 30, 30);

	    g2.drawRect(INCH * 2, 0, 30, 30);
	    g2.fillRect(INCH * 3, 0, 30, 30);

	    g2.drawRoundRect(INCH * 4, 0, 30, 30, 5, 5);
	    g2.fillRoundRect(INCH * 5, 0, 30, 30, 5, 5);

	    //-------------------------------------
	    g2.translate(0, LINE_HEIGHT);

	    {
		int[] xPoints = new int[5];
		int[] yPoints = new int[5];
		xPoints[0] = 0;
		xPoints[1] = 20;
		xPoints[2] = 30;
		xPoints[3] = 20;
		xPoints[4] = 0;

		yPoints[0] = 0;
		yPoints[1] = 5;
		yPoints[2] = 15;
		yPoints[3] = 25;
		yPoints[4] = 30;

		g2.drawPolyline(xPoints, yPoints, 5);
	    }

	    {
		int[] xPoints = new int[5];
		int[] yPoints = new int[5];
		xPoints[0] = INCH + 0;
		xPoints[1] = INCH + 20;
		xPoints[2] = INCH + 30;
		xPoints[3] = INCH + 20;
		xPoints[4] = INCH + 0;

		yPoints[0] = 0;
		yPoints[1] = 5;
		yPoints[2] = 15;
		yPoints[3] = 25;
		yPoints[4] = 30;
		g2.drawPolygon(xPoints, yPoints, 5);
	    }

	    {
		Graphics2D tempGraphics = (Graphics2D) g2.create();
		tempGraphics.translate(2 * INCH, 0);
		tempGraphics.fillPolygon(mPoly);
		tempGraphics.translate(INCH, 0);
		tempGraphics.drawPolygon(mPoly);
		tempGraphics.dispose();
	    }

	    result = Printable.PAGE_EXISTS;

	} else {
	    result = Printable.NO_SUCH_PAGE;
	}

	return result;
    }

}
