/*
 * @(#)OldNMI-impl.c	1.5 98/05/04
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Sanity check old style NMI.
 */

#include "monitor.h"
#include "exceptions.h"
#include "OldNMI.h"
#include "Helper.h"
#include <string.h>

/*
 * Help routines.
 */

static long *
getclassvariable(ClassClass *cb, char *fname)
{
    int    n = cbFieldsCount(cb);
    struct fieldblock *fb;

    for (fb = cbFields(cb); --n >= 0; fb++)
    if ((fb->access & ACC_STATIC) && strcmp(fname, fieldname(fb)) == 0) {
        char *signature = fieldsig(fb);
        if (signature[0] == SIGNATURE_LONG ||
            signature[0] == SIGNATURE_DOUBLE)
            return (long *)twoword_static_address(fb);
        else
            return (long *)normal_static_address(fb);
    }
    return (long *)0;
}

static ClassClass*
getHelperClass()
{
    ClassClass* helperClass
	= FindClass(EE(), "Helper", TRUE);
    
    if (helperClass == NULL || exceptionOccurred(EE())) {
	/* Exception will be propogated upwards. */
	return NULL;
    }
    return helperClass;
}

static HHelper*
getHelperInstance()
{
    ExecEnv* ee = EE();
    JHandle* helper;
    ClassClass* helperClass = getHelperClass();
    
    if (helperClass == NULL) {
	return NULL;
    }
    helper = *(JHandle**)getclassvariable(helperClass, "aHelper");
    if (helper == NULL || exceptionOccurred(ee)) {
	/* Exception will be propogated upwards. */
	return NULL;
    }
    return (HHelper*)helper;
}

static HString*
getStringJavaSoft()
{
    HHelper* helper = getHelperInstance();
    if (helper == NULL) {
	return NULL;
    }
    return ((ClassHelper*)unhand(helper))->stringOfLength8;
}

/*
 * Tests.
 */
long
OldNMI_javaStringLength(struct HOldNMI* thiz)
{
    HString* jstr = getStringJavaSoft();
    if (jstr == NULL) {
	return FALSE;
    }
    if (javaStringLength(jstr) == 8) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_makeJavaString(struct HOldNMI* thiz) 
{
    char* cstr = "JavaSoft";
    HString* jstr = makeJavaString(cstr, strlen(cstr));
    /* Note: we assume String object layout. Perhaps not a good thing. */
    if ((unhand(jstr)->count - unhand(jstr)->offset) == strlen(cstr)) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_makeCString(struct HOldNMI* thiz) 
{
    HString* jstr = getStringJavaSoft();
    if (jstr == NULL) {
	return FALSE;
    }
    if (strcmp("JavaSoft", makeCString(jstr)) == 0) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_allocCString(struct HOldNMI* thiz) 
{
    HString* jstr = getStringJavaSoft();
    char* cstr;
    if (jstr == NULL) {
	return FALSE;
    }
    cstr = allocCString(jstr);
    if (strcmp("JavaSoft", cstr) == 0) {
	return TRUE;
    }
    free(cstr);
    return FALSE;
}

long
OldNMI_javaString2unicode(struct HOldNMI* thiz)
{
    unicode buf[10];
    char* cstr = "JavaSoft";
    int cstrlen = strlen(cstr);
    int i = 0;
    HString* jstr = getStringJavaSoft();

    if (jstr == NULL) {
	return FALSE;
    }
    javaString2unicode(jstr, buf, sizeof(buf)/sizeof(unicode));
    for (i = 0; i < cstrlen; i++) {
	if (cstr[i] != (char)buf[i]) {
	    return FALSE;
	}
    }
    return TRUE;
}

long
OldNMI_javaString2CString(struct HOldNMI* thiz)
{
    char buf[10];
    char* cstr = "JavaSoft";
    HString* jstr = getStringJavaSoft();

    if (jstr == NULL) {
	return FALSE;
    }
    javaString2CString(jstr, buf, sizeof(buf)/sizeof(char));
    if (strcmp(cstr, buf) == 0) {
	return TRUE;
    }
    return FALSE;
}


long
OldNMI_FindClass(struct HOldNMI* thiz)
{
    /*
     * Currently the implementation of javaStringLength test in this
     * file uses FindClass, so if that succeeds, we should be alright.
     */
    return OldNMI_javaStringLength(thiz);
}

long
OldNMI_FindStickySystemClass(struct HOldNMI* thiz)
{
    ClassClass *stringClass = FindStickySystemClass(EE(), "java/lang/System",
						    TRUE);
    if (CCIs(stringClass, Sticky)) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_FindClassFromClass(struct HOldNMI* thiz) 
{
    /* 
     * This is a compile time only test because I am too lazy to write
     * a classloader.
     */
    ClassClass *helperClass = FindClassFromClass(EE(), "Helper", TRUE,
						 thiz->methods->classdescriptor);
    if (helperClass == NULL || exceptionOccurred(EE())) {
	return FALSE;
    }
    return TRUE;
}

long
OldNMI_SignalError(struct HOldNMI* thiz)
{
    SignalError(EE(), "java/lang/RuntimeException", "foobar");
    if (!exceptionOccurred(EE())) {
	return FALSE;
    }
    exceptionClear(EE());
    if (exceptionOccurred(EE())) {
	return FALSE;
    }
    return TRUE;
}

long
OldNMI_EE(struct HOldNMI* thiz)
{
    /* SignalError is a good check of EE usage. */
    return OldNMI_SignalError(thiz);
}

long
OldNMI_ArrayAlloc(struct HOldNMI* thiz)
{
    HArrayOfObject *newarray = (HArrayOfObject *)ArrayAlloc(T_CLASS, 10);
    ClassClass *helperClass = getHelperClass();
    int ret;
    JHandle *helper;
    JHandle **ptrToHandleToArrayField;

    if (helperClass == NULL) {
	return FALSE;
    }
    helper = *(JHandle**)getclassvariable(helperClass, "aHelper");
    if (helper == NULL) {
	return FALSE;
    }
    ptrToHandleToArrayField = (JHandle **)getclassvariable(helperClass, 
						   "helperArray");
    *ptrToHandleToArrayField = (JHandle *)newarray;
    ret = execute_java_dynamic_method(EE(), helper, "arraySizeCheck",
				      "(I)Z", 10);
    if (exceptionOccurred(EE())) {
	/* overly defensive, this is very very bogus. */
	return FALSE;
    }
    return ret;
}

long
OldNMI_ObjAlloc(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    long beforeCount;
    long afterCount;
    HHelper* newHelper;
    
    if (helperClass == NULL) {
	return FALSE;
    }
    beforeCount = *(getclassvariable(helperClass, "allocCount"));
    newHelper = (HHelper *)ObjAlloc(helperClass, 0);
    execute_java_dynamic_method(EE(), (JHandle *)newHelper, "<init>", "()V");
    afterCount = *(getclassvariable(helperClass, "allocCount"));
    if (afterCount == (beforeCount + 1)) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_newobject(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    long beforeCount;
    long afterCount;
    HHelper* newHelper;
    
    if (helperClass == NULL) {
	return FALSE;
    }
    beforeCount = *(getclassvariable(helperClass, "allocCount"));
    newHelper = (HHelper *)newobject(helperClass, 0, EE());
    execute_java_dynamic_method(EE(), (JHandle *)newHelper, "<init>", "()V");
    afterCount = *(getclassvariable(helperClass, "allocCount"));
    if (afterCount == (beforeCount + 1)) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_do_execute_java_method(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    long result;
    
    if (helperClass == NULL || exceptionOccurred(EE())) {
	return FALSE;
    }
    
    result = do_execute_java_method(EE(), helperClass, "isArgument10Static",
				    "(I)Z", NULL, TRUE, 10);
    return result;
}

static long
doVarArgTest(ExecEnv *ee, void *obj, char *mname, char *msig, struct
	     methodblock *mb, bool_t isStatic, ...) 
{
    va_list args;
    long result;
    va_start(args, isStatic);
    result = do_execute_java_method_vararg(ee, obj, mname, msig, NULL,
					   isStatic, args, 0, FALSE);
    va_end(args);
    return result;
}

long
OldNMI_do_execute_java_method_vararg(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    long result;
    
    if (helperClass == NULL || exceptionOccurred(EE())) {
	return FALSE;
    }
    
    result = doVarArgTest(EE(), helperClass, "isArgument10Static",
			  "(I)Z", NULL, TRUE, 10);
    return result;
}


long
OldNMI_execute_java_dynamic_method(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    JHandle* helper = (JHandle *)getHelperInstance();
    long result;
    
    if (helperClass == NULL || helper == NULL || exceptionOccurred(EE())) {
	return FALSE;
    }
    
    result = execute_java_dynamic_method(EE(), helper, "isArgument10",
					 "(I)Z", 10);
    return result;
}

long
OldNMI_execute_java_static_method(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    JHandle* helper = (JHandle *)getHelperInstance();
    long result;
    
    if (helperClass == NULL || helper == NULL || exceptionOccurred(EE())) {
	return FALSE;
    }
    
    result = execute_java_static_method(EE(), helperClass, "isArgument10Static",
					 "(I)Z", 10);
    return result;
}

long
OldNMI_execute_java_constructor(struct HOldNMI* thiz)
{
    /* This is already tested by ObjAlloc test, so elide it. */
    return OldNMI_ObjAlloc(thiz);
}

long
OldNMI_is_instance_of(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    JHandle* helper = (JHandle *)getHelperInstance();
    return is_instance_of(helper, helperClass, EE());
}

long
OldNMI_ImplementsInterface(struct HOldNMI* thiz)
{
    ClassClass* helperClass = getHelperClass();
    ClassClass* interfaceClass = FindClass(EE(), "HelperInterface",
					   TRUE);
    if (helperClass == NULL || interfaceClass == NULL ||
	exceptionOccurred(EE())) {
	return FALSE;
    }
    return ImplementsInterface(helperClass, interfaceClass, EE());
}

#ifdef WIN32
__declspec(dllimport) ClassClass *Thread_classblock;
#endif

long
OldNMI_Thread_classblock(struct HOldNMI* thiz) 
{
    /* For some reason this is exported on Win32, so test for it. */
    extern ClassClass *Thread_classblock;
    if (strcmp(cbName(Thread_classblock), "java/lang/Thread") == 0) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_classname2string(struct HOldNMI* thiz) 
{
    char *src = "java/soft";
    char dst[10];
    char *gold = "java.soft";
    
    classname2string(src, dst, 10);
    if (strcmp(gold, dst) == 0) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_jio_snprintf(struct HOldNMI* thiz)
{
    char str[10];
    int result;
#define HEHE "HEHE"
    result = jio_snprintf(str, 10, HEHE);
    if (result == strlen(HEHE) && strcmp(str, HEHE) == 0) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_jio_printf(struct HOldNMI* thiz)
{
    if (FALSE) {
	jio_printf("ha, this will never print");
    }
    return TRUE;
}

long
OldNMI_jio_fprintf(struct HOldNMI* thiz)
{
    if (FALSE) {
	jio_fprintf(stderr, "ha, this will never print");
    }
    return TRUE;
}

long
OldNMI_jio_vfprintf(struct HOldNMI* thiz)
{
    if (FALSE) {
	jio_vfprintf(stderr, "ha, this will never print", (va_list)thiz);
    }
    return TRUE;
}

long
OldNMI_exceptionDescribe(struct HOldNMI* thiz)
{
    if (exceptionOccurred(EE())) {
	exceptionDescribe(EE());
    }
    return TRUE;
}

long
OldNMI_sizearray(struct HOldNMI* thiz)
{
    /* 
     * Assumption that a java int will be a put into a platform int
     * should be valid most of the time...
     */
    if (sizearray(T_INT, 10) == (sizeof(int32_t) * 10)) {
	return TRUE;
    }
    return FALSE;
}

long
OldNMI_threadSelf(struct HOldNMI* thiz)
{
    if (threadSelf() == NULL) {
	return FALSE;
    }
    return TRUE;
}

/* 
 * Perhaps these monitor tests should be run-time tested. No time to
 * do that now.
 */
static long
montest(struct HOldNMI* obj) 
{
    monitor_t mon;
    unsigned int dummy = (unsigned int)obj;
    
    if (threadSelf() == NULL) { /* never happens. */
	monitorInit(&mon);
	monitorEnter(dummy);
	monitorExit(dummy);
	monitorWait(dummy, 10);
	monitorNotify(dummy);
	monitorNotifyAll(dummy);
    }
}

long
OldNMI_monitorInit(struct HOldNMI* thiz){
    return montest(thiz);
}

long
OldNMI_monitorEnter(struct HOldNMI* thiz){
    return montest(thiz);
}

long
OldNMI_monitorExit(struct HOldNMI* thiz){
    return montest(thiz);
}

long
OldNMI_monitorWait(struct HOldNMI* thiz){
    return montest(thiz);
}

long
OldNMI_monitorNotify(struct HOldNMI* thiz){
    return montest(thiz);
}

long
OldNMI_monitorNotifyAll(struct HOldNMI* thiz){
    return montest(thiz);
}

