#!/bin/ksh

#
#Shell script to run the Juicer as an application
#

set -e

PATH=/bin:/usr/bin:${PATH}
export PATH

if [ $# -ne 1 ]
	then
	echo "Usage: run.sh <path to JDK>"
	exit 1
fi

# redirection blindly copied from JDK1.1 scripts although they look worng
PRG=`whence $0` >/dev/null 2>&1
case "$PRG" in
'') PRG="$0"
esac
SRCDIR=`dirname $PRG`
#echo "XXX PRG=$PRG SRCDIR=$SRCDIR 0=$0"

JAVA_HOME=$1
PATH=$JAVA_HOME/bin:$PATH
export PATH
#LD_LIBRARY_PATH=$JAVA_HOME/lib/sparc"
#export LD_LIBRARY_PATH



#
# show some string (if -v, only in verbose)
#
show()
{
    if [ x"$1" = x"-v" ]; then
	    shift
	    echo "..." "$@"
    elif [ x"$1" = x"-c" ]; then
	shift
	echo ""
	echo "$@"
    else
	echo "$@"
    fi
}

#
# (There is no need to run rmiregistry for this example since
# the AppleUserImpl class bundles its own registry.)
#

CLASSPATH=.:$JAVA_HOME/lib/rmi.zip:$CLASSPATH
export CLASSPATH

show -v "CLASSPATH=$CLASSPATH"

#
# Compile the source
#
show -c "Compiling Java sources"
javac -d . -deprecation ${SRCDIR}/*.java

#
# Run rmic
#
show -c "Running rmic on the implementations"
rmic -d . AppleUserImpl AppleImpl AppletServer OrangeImpl OrangeEchoImpl

#
# Running the server
#
codebase="file://localhost${SRCDIR}/"
scb="-Djava.rmi.server.codebase=$codebase"

show -c "Starting AppleUserImpl server"
java -Djava.security.policy="${SRCDIR}/policy" $scb AppleUserImpl &
pid1=$!
echo "AppleUserImpl.pid = $pid1"
trap "kill $pid1" 0
trap 'exit 1' 2

# sleep to give the server time to start up and print its message
sleep 6

show -c "Starting the AppletServer as an application"
java -Djava.security.policy="${SRCDIR}/policy" $scb AppletServer
res=$?

kill $pid1
trap '' 0

wait
exit $res
