/*
 * @(#)AttribData.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.util.*;

/**
 * The Attribute data for a particular test. 
 */
public class AttribData implements java.io.Serializable{

	private boolean arraysUsed;

	private Vector figures;

	private int[] sizeList;

	private String measuredAttrib;

	/**
	 * Constructs an AttribData object for a particular test <em>not</em>
	 * involving arrays, vectors, btree, etc.
	 * @param name    a unique identifier for the test.
	 * @param figs    a vector of <i>PerfFigures</i> for the particular
	 *                attribute this <i>AttribData</i> refers to. The
 	 *                vector size is equal to the number of loops executed
	 *                for this test
	 *
	 * @see PerfFigures
	 */
	
	public AttribData(String name, Vector figs){

		measuredAttrib = name;
		figures = figs;
		arraysUsed = false;
		sizeList = null; 

	}

	/**
	 * Constructs an AttribData object for a particular test
	 * <em>involving</em> arrays, vectors, btrees or any data structure
	 * for which a parameter other than the range of iterations can be
	 * specified.
	 *
	 * @param name    a unique identifier for the test.
	 * @param figs    a vector of <i>PerfFigures</i> for the particular
	 *                attribute this <i>AttribData</i> refers to. The
 	 *                vector size is equal to the number of loops executed
	 *                for this test
	 * @param listOfArraySizes An integer array giving the range of array
	 *		  sizes tested for this test.
	 *
	 * @see PerfFigures
	 */
	
	public AttribData(String name, Vector figs, 
			int[] listOfArraySizes){

		measuredAttrib = name;
		figures = figs;
		arraysUsed = true;
		sizeList = new int[listOfArraySizes.length];
		sizeList = listOfArraySizes;

	}

	public String getMeasuredAttrib() {
		
		return measuredAttrib;

	}

	public Vector getFigures() {

		return figures;

	}

	public int[] getSizeList() {
		
		return sizeList;

	}

	/**
   	 * Returns <i>true</i> if the test involves a data structure for which
	 * a set of parameters can be specified.      
	 *
	 * @return a true/false value indicating whether the test uses any 
	 *	   such data structures.
	 */
	public boolean areArraysUsed(){

		return arraysUsed;
	}

}

			







