/*
 * @(#)ClientData.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.util.*;


/**
 * Contains all description and performance figures for each client.
 *
 * @see AttribData
 * @see LogData
 * @see PerfFigures
 */

public class ClientData implements java.io.Serializable{

	/*
	 * Client Identification number assigned by the DataCollector to each 
	 * client to help identify the client at the time of logging in 
	 * performance figures.
	 */
	private int ClientNum;

	/*
	 * A set of properties for each client including the system properties
	 * for the client site.
	 */
	private Properties ClientDesc; 

	//The design for perfData was used with easy extensibility in mind.
	//We can easily make more than one set of iterations as well as add
	//new tests without modifying much of the code. The only place that
	//needs modification is the TestClient code for the actual test itself 
	//and that isn't asking too much :)

	/*
	 * A hashtable containing performance figures for each measured attrib. 
	 * key: the Measured attribute 
         * data: An attribData Object containing the performance figures for  
	 * 	 a particular test.
	 * 
	 */ 

 
	private Hashtable perfData ;

	/*
	 * The RMI performance mark for this client. It is the sum of the times
	 * taken for each test carried out.
	 */
	private float RMIPerfMark ;



	/**
	 * Constructs a ClientData object with the given client number and
	 * records the client properties in this object.
	 *
	 * @param cNum  the clientNum used to identify the client to the 
	 *              the DataCollector.
	 * @param cDesc the client Properties including the system properties
	 *              for the client site.
	 */

	public ClientData(int cNum, Properties cDesc) {

		ClientNum = cNum;
		ClientDesc = cDesc;
		perfData = null;

	}

	/**
	 * Insert the performance data for each client . Contains a set of 
	 * AttribData objects keyed on the test name.
	 *
	 * @param data The hashtable containing the <i>AttribData</i> objects
	 * 	       keyed on the test name.
	 * @see AttribData 
	 *
	 */

	public void addPerfFigures (Hashtable data) {

		perfData = data;
		return;
	} 

	public Properties getDesc () {

		return ClientDesc;

	}

	public Hashtable getPerfFigures () {

		return perfData;

	}

	public void setPerfMark (float RMIPerformanceMark) {
		
		RMIPerfMark = RMIPerformanceMark;
		return;
	
	}

	public float getPerfMark () {
	
		return RMIPerfMark;

	}
}		
