/*
 * @(#)ControlPanel.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.awt.*;
import java.io.*;
import java.awt.event.*;
import java.net.*;
import java.util.*;
import visualBenchmarks.QueryDialog;
import visualBenchmarks.MessageBox;

/**
 * The Control Panel provides a user-friendly interface to edit the properties
 * file for the Data Collector.
 **/

public class ControlPanel extends Frame
	implements ActionListener, ItemListener,FilenameFilter, TextListener {

	private TextField tfVersion = new TextField(30);
	private TextField tfReportDest = new TextField(30);
	private Button bSelectReport = new Button("Select");
	private TextField tfCentralStore = new TextField(30);
	private TextField tfLogDir = new TextField(30);
	private Button bSelectLogDir = new Button("Select");
	private Checkbox cbLocalHost = new Checkbox("Local Host");
	private TextField tfLogDest = new TextField(30);
	private Button bUpdate = new Button("Update");
	private Button bCancel = new Button("Cancel");
        private Button bExit = new Button("Exit");

	private boolean NEW_FILE = false;

	//The properties file name
	private String propsFilename = "benchmarks.server.properties";

	// a Hashtable to store the properties and record changes as they
	// are edited.
	private Hashtable propsTable;

	/**
	 * Reads the properties file and parses it into <i> &lt key, value &gt
	 * </i> pairs - the keys are the Property names in the properties file.
	 * These are then stored in a hashtable
	 *
	 * @return the hashtable with all the properties and their
	 * 	   values recorded.
	 */
	protected Hashtable readProps () {

		String line, key, value;
		Hashtable props = new Hashtable();
		int start, end;
		boolean EOF = false;
		
		try {

		   //The LineNumberReader is chosen, since it allows you to
		   //read complete lines at a time. 

		   LineNumberReader in = new LineNumberReader(new FileReader(
						propsFilename));

	           do {
			start = 0;
			try {
				line = new String(in.readLine());
				//Skip Comment lines
				if (!(line.startsWith("#"))) {
					end = line.indexOf('=');

					//Skip if no "=" - indicating not a
					//property line.
					if (end != -1) {
						//Extract key and it's value. 
						//Record it in the hashtable

						key = line.substring(start,end);
						value = line.substring(end + 1);
						props.put(key,value);
					}		
				}
			}catch (NullPointerException e){
				EOF = true;
			}
		   }while (!EOF);
		}catch (Exception e) {
			System.out.println("ControlPanel err: (readProps()): " +
				 e.getMessage());
			e.printStackTrace();
			System.exit(-1);
		}

		return props;
	}

	/**
	 * Display a FileDialog to select the report file path and name. 
	 * 
	 * @param  title the title for the FileDialog Box
	 * @return the absolute path for the report file
	 */
	protected String getFileName(String title) {
	  
		FilenameFilter filter;
		String currPath, dirPath=new String(), fileName = new String();
		String path = new String();
		int index;

		//Open a new FileDialog box in the save mode 
		//(allows you to create new files)

	  	FileDialog fd = new FileDialog(this, title, FileDialog.SAVE);
			
		char separator = File.separatorChar;
		currPath = tfReportDest.getText();

		index = currPath.lastIndexOf(separator);
	
		if (index != -1) {
			//If not just the filename, then parse 
			//the string to extract the Directory path 
			//and the Filename

			dirPath = currPath.substring(0,index);
			fileName = currPath.substring(index +1);
		}else {
			//else get the directory path *from* the filename
			File logFile = new File(currPath);
			if (logFile.exists()){
				dirPath = logFile.getParent();
				fileName = currPath;
			}
		}			

		fd.setDirectory(dirPath);
		fd.setFile(fileName);
		fd.setFilenameFilter(this);
		
		fd.show();

		dirPath = fd.getDirectory();
		fileName = fd.getFile();

		//If the user did not select a file name,
		//keep the original one, else return the new path

		if (fileName == null) {	
			path = new String(currPath);
		}else {
			path = new String(dirPath +  fileName);
		}

		return path;
	}	
	
	/**
	 * Read the user input and update the properties file
	 */
	protected void update ( ) {
		
		//This method creates an intermediate file, and copies 
		//the original properties file to it, modifying the property 
		//values with those entered by the user.

		boolean EOF = false;
		String line = new String();
		String key = new String(), value = new String();		
		boolean isPropLine;
		boolean isComment;
		int pos;
		String msg = new String();
		MessageBox msgBox = null;


		Hashtable propsCopy = (Hashtable)propsTable.clone();

		try {
		   BufferedReader in = new BufferedReader(
					new FileReader(propsFilename));

		   File intermediate = new File("_tmpFile");
		   BufferedWriter out = new BufferedWriter(
					new FileWriter(intermediate));

		   do {
			line = in.readLine();
			try {
				//Is this line a property line (yes if it has a
				// "=" and it does not begin with a hash)?

				isPropLine=(line.indexOf('=') != -1)?true:false;
				isComment = line.startsWith("#")?true:false;

				//If it is a property line, extract the key .
				//Determine the new value for this key from 
				//the properties table and insert this new 
				//value in place of the old one.

				if (isPropLine && (!isComment)) {
					pos = line.indexOf('=');
					key = line.substring(0,pos);
					value = 
					 new String((String)propsCopy.get(key));
					if (value.equals("")) {	
				          if (!(key.equals("LogDir"))) {
					    msg =new String("No value entered "+
						"for property " + key);
					    msgBox = new MessageBox(this,
						"ERROR!",msg);
					    msgBox.show();
					    return;
				          }else if (cbLocalHost.getState()){
					    msg =new String("No value entered "+
						"for property " + key);
					    msgBox = new MessageBox(this,
						"ERROR!",msg);
					    msgBox.show();
					    return;
				          }
				        }
					line = new String(key + "=" +value);
				}
				out.write(line,0,line.length());
				out.newLine();
			}catch (NullPointerException e) {
				out.flush();
				out.close();
				in.close();
				EOF = true;
			}
		   }while (!EOF);
		
		   //Overwrite the old file with the contents of the 
		   //intermediate file by renaming the intermediate file to 
		   //the old filename.

		   if (!intermediate.renameTo(new File(propsFilename))) {
			System.out.println("Could not overwrite properties " +
				"file");
		   }
		}catch (Exception e) {
			System.out.println("ControlPanel err: (updata()): " +
				e.getMessage());
			e.printStackTrace();
			System.exit(-1);
		}
	}
	
	/**
	 * Read the user input and create a new properties file
	 */
	protected void create ( ) {

		boolean EOF = false;
		String line = new String();
		String key = new String(), value = new String();		
		boolean isPropLine;
		boolean isComment;
		int pos;
		String msg = new String();
		MessageBox msgBox = null;
 
		if (propsTable.isEmpty()) {
			msg = new String("Enter values for the " +
				"properties");
			msgBox = new MessageBox(this,"ERROR!", msg);
			msgBox.show();
			return;
		}
			
		Hashtable propsCopy = (Hashtable)propsTable.clone();
		try {

		   BufferedWriter out = new BufferedWriter(
					new FileWriter(propsFilename));

		   Enumeration e = propsCopy.keys();
		   while (e.hasMoreElements()) {
			key = (String)e.nextElement();
			value = (String)propsTable.get(key);
			if (value.equals("") ) {	
				if (!(key.equals("LogDir"))){
					msg = new String("No value entered " +
						"for property " + key);
					msgBox = new MessageBox(this,
						"ERROR!",msg);
					msgBox.show();
					return;
				}else if (cbLocalHost.getState()){
				    
					msg = new String("No value entered " +
						"for property " + key);
					msgBox = new MessageBox(this,
						"ERROR!",msg);
					msgBox.show();
					return;
				}
			}else {
				line = new String(key + "=" + value);
				out.write(line,0,line.length());
				out.newLine();
			}
		   }
		   NEW_FILE = false;

		}catch (Exception e) {
			System.out.println("ControlPanel err: (updata()): " +
				e.getMessage());
			e.printStackTrace();
			System.exit(-1);
		}
	}
	
	/**
	 * Initializes the layout ,reads the properties file, etc.
	 */	
	public void init() {

		Label label;
		String centralStore, hostName = new String();;

		File propsFile = new File(propsFilename);

		//If the properties file already exists, read it and populate 
		//the properties table
		//Else check if the specified filename does not represent a 
		//directory. 
		if (propsFile.exists() && !(propsFile.isDirectory())) {
		 	propsTable = readProps();
		}else {
			if (propsFile.isDirectory()) {
			   System.out.println("The specified filename for the "+
				"properties file is a directory. Specify " +
				"another name...");
			   System.exit(-1);
			}
			NEW_FILE = true;
			propsTable = new Hashtable();
		}

		//Layout the various components on the frame.
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
	
		setLayout(gbl);

		label = new Label("Version: ", Label.RIGHT);
		gbc.gridx = 0;
		gbc.gridy = 0;
		gbc.fill = GridBagConstraints.BOTH;
		gbl.setConstraints(label, gbc);
		add(label);
	
		gbc.gridx = 1;
		gbc.gridy = 0;
		gbc.gridwidth = 1;
		gbc.weightx = 1.0;
		gbl.setConstraints(tfVersion, gbc); 
		if (NEW_FILE) {
			tfVersion.setText("");
		}else {
			tfVersion.setText((String)propsTable.get("Version"));
		}
		tfVersion.addActionListener(this);
		tfVersion.addTextListener(this);
		add(tfVersion);
	
		label = new Label("Report File path: ", Label.RIGHT);
		gbc.gridx = 0;
		gbc.gridy = 1;
		gbc.weightx = 0.0;
		gbc.gridwidth = 1;
		gbl.setConstraints(label, gbc);
		add(label);
			
		
		gbc.gridx = 1;
		gbc.gridy = 1;
		gbc.weightx = 1.0;
		gbc.gridwidth = 1;
		gbl.setConstraints(tfReportDest, gbc);
		if (NEW_FILE) {
		   tfReportDest.setText("");
		}else {
		   tfReportDest.setText((String)propsTable.get("ReportDest"));
		}
		tfReportDest.addActionListener(this);
		tfReportDest.addTextListener(this);
		add(tfReportDest);
		
		
		gbc.gridx = 2;
		gbc.gridy = 1;
		gbc.weightx = 0.0;
		gbc.gridwidth = 1;
		gbc.fill = GridBagConstraints.NONE;
		gbc.anchor = GridBagConstraints.WEST;
		gbl.setConstraints(bSelectReport, gbc);
		bSelectReport.addActionListener(this);
		add(bSelectReport);

		label = new Label("Central Store: ", Label.RIGHT);
		gbc.gridx = 0;
		gbc.gridy = 2;
		gbc.weightx = 0.0;
		gbc.gridwidth = 1;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.anchor = GridBagConstraints.CENTER;
		gbl.setConstraints(label, gbc);
		add(label);
		
		gbc.gridx = 1;
		gbc.gridy = 2;
		gbc.fill = GridBagConstraints.NONE;
		gbc.anchor = GridBagConstraints.EAST;
		gbl.setConstraints(cbLocalHost, gbc);
		cbLocalHost.addItemListener(this);
		add(cbLocalHost);

		gbc.gridx = 2;
		gbc.gridy = 2;
		gbc.weightx = 1.0;
		gbc.gridwidth = 1;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.anchor = GridBagConstraints.CENTER;
		gbl.setConstraints(tfCentralStore, gbc);
		tfCentralStore.addActionListener(this);
		tfCentralStore.addTextListener(this);

		//If the properties file is merely being edited, show the old
		//values in the text box, else display an empty string.
		if (NEW_FILE) {
			centralStore = new String("");
		}else {
			centralStore = (String)propsTable.get("CentralStore");
		}

		//Compare centralStore with the local host name. If they match, 
		//do not do a Naming.lookup() , instead save things to the local
		//file , in the directory indicated by "LogDir"

		try {
			hostName = (InetAddress.getLocalHost()).getHostName();
		}catch (UnknownHostException e) {
			System.out.println("ControlPanel err: (init()): " +
				e.getMessage());
			System.exit(-1);
		}

		//If it's the local host, enable the checkbox and clear the 
		//CentralStore location textbox

		if (centralStore.equals(hostName) ||
			 centralStore.equals("localHost")){
			cbLocalHost.setState(true);
			tfCentralStore.setText("");
			tfCentralStore.setEnabled(false);

		}else {
			tfCentralStore.setText(centralStore);
		}
		add(tfCentralStore);

		label = new Label("Log File path: ", Label.RIGHT);
		gbc.gridx = 0;
		gbc.gridy = 3;
		gbc.weightx = 0.0;
		gbc.gridwidth = 1;
		gbl.setConstraints(label, gbc);
		add(label);

		gbc.gridx = 1;
		gbc.gridy = 3;
		gbc.gridwidth = 1;
		gbc.weightx = 1.0;
		gbl.setConstraints(tfLogDest, gbc);
		tfLogDest.addActionListener(this);
		tfLogDest.addTextListener(this);
		if (NEW_FILE) {
			tfLogDest.setText("");
		}else {
			tfLogDest.setText((String)propsTable.get("LogDest"));
		}
		add(tfLogDest);

		gbc.gridx = 0;
		gbc.gridy = 4;
		gbc.gridwidth = 1;
		gbc.weightx = 0.0;
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.fill = GridBagConstraints.NONE;
		gbc.insets.top = gbc.insets.bottom = 5;
		gbl.setConstraints(bUpdate, gbc);
		bUpdate.addActionListener(this);
		add(bUpdate);
	
		gbc.gridx = 1;
		gbc.gridy = 4;
		gbl.setConstraints(bExit, gbc);
		bExit.addActionListener(this);
		add(bExit);

		gbc.gridx = 2;
		gbc.gridy = 4;
		gbl.setConstraints(bCancel, gbc);
		bCancel.addActionListener(this);
		add(bCancel);
	}

	/**
	 * Constructs a frame for this application, adds a Window Listener for
	 * window events fired by the frame window, and lays out the application
	 * components on the frame.
	 */
	public ControlPanel () {
		super("DataCollector Control Panel");

		//Add a window adapter as a listener to any window events fired
		//by this frame. Used to execute some cleanup code when the
		//frame window is destroyed.

		addWindowListener(new WinAdapter());
		init();
		pack();
		show();
		//Bring the user focus onto the Version text area.
		tfVersion.requestFocus();
	}

	public void actionPerformed(ActionEvent evt) {
		if (evt.getSource() instanceof Button){
			Button bTarget = (Button)evt.getSource();
			if (bTarget == bSelectReport) {
				//Start up a FileDialog and get the new filename
				tfReportDest.setText(getFileName(
					"Select Report File"));
			}else if (bTarget == bUpdate) {
				//If the properties file does not already exist
				//create a new one, else update the existing 
				//one
				if (NEW_FILE) {
					create();
				}else {
					update();
				}
			}else if (bTarget == bCancel) {
				String msg = new String("Any unsaved changes " +
				    "will be lost. Do you wish to proceed?");
				QueryDialog qd = new QueryDialog(this,
				    "Confirm",msg);
				qd.show();
				if (qd.response == QueryDialog.ACCEPT){
					System.exit(0);
				}
			}else if (bTarget == bExit) {
				setVisible(false);
				dispose();
				System.exit(0);
			}

		}
	}

	public void textValueChanged(TextEvent evt) {

	//Record any changes in the properties table. Later, when the 
	//user wishes to update the properties file, these changes are saved 
	//onto the file

		 if (evt.getSource() instanceof TextField) {
			TextField tfTarget = (TextField)evt.getSource();
			if (tfTarget == tfReportDest) {
				propsTable.remove("ReportDest");
				propsTable.put("ReportDest",
				  tfReportDest.getText());
			}else if (tfTarget == tfLogDest) {
				propsTable.remove("LogDest");
				propsTable.put("LogDest",tfLogDest.getText());
			}else if (tfTarget == tfCentralStore){
				if (!(cbLocalHost.getState())){
					propsTable.remove("CentralStore");
					propsTable.put("CentralStore",
					  tfCentralStore.getText());
				}
			}else if (tfTarget == tfVersion) {
				propsTable.remove("Version");
				propsTable.put("Version", tfVersion.getText());
			}
		}
	}
		
	public void itemStateChanged(ItemEvent evt) {

		if (evt.getSource() instanceof Checkbox) {
			Checkbox cbTarget = (Checkbox)evt.getSource();
			if (cbTarget == cbLocalHost) {

				//If the localHost checkbox is enabled, set the
				//"CentralStore" property to local host, else
				//set it to the text contents of the Central
				//Store text field.

				if (cbLocalHost.getState()) {
					propsTable.remove("CentralStore");
					propsTable.put("CentralStore",
					  "localHost");
				}else {
				  	propsTable.remove("CentralStore");
					propsTable.put("CentralStore", 
					  tfCentralStore.getText());
				}
			
				//Toggle the state for the Central Store text
				//field whenever you select or deselect the 
				//Local Host checkbox
				tfCentralStore.setEnabled(
					!(tfCentralStore.isEnabled()));
			}
		}		
	}

	/**
	 * Accept filenames ending with .html
	 * 
	 * @param dir  the Directory
	 * @param name the filename
	 * @return  true if the filename ends in .html, false if not.
	 */
	public boolean accept(File dir, String name) {
	  
	  
		return (name.endsWith(".html")?true:false);		
	}

	public static void main (String[] args) {

		ControlPanel ctrl = new ControlPanel();
	}
		
}

/**
 * A window adapter, for the control panel. Executes cleanup code when the user
 * closes the frame window.
 */

class WinAdapter extends WindowAdapter {

	public void windowClosing(WindowEvent evt) {
		
		Frame src = (Frame)evt.getSource();
		src.setVisible(false);
		src.dispose();
		System.exit(0);
	}
}
