/*
 * @(#)DataServer.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package visualBenchmarks;

import java.rmi.*;
import benchmarks.*;
import java.util.*;

/**
 * The functionality provided by the DataServer to the LogViewer clients
 */

public interface DataServer extends Remote {
  
  	/**
	 * Get a list of JDK versions tested in a particular test suite version
	 * The test set version is selected by setting the <i>Version</i> 
	 * property in the properties file
	 */
	public String[] getVersions() throws RemoteException;

	/**
	 * Get the list of tests performed in a particular test suite version
	 * This method actually returns a vector of TestInfo objects which
	 * include information whether the tests use a second parameter other
	 * than the range of iterations.
	 * 
	 * @see TestInfo
	 */
	public Vector getListOfTests() throws RemoteException;
 	
	/**
	 * Get information specific to a particular log, identified by the 
	 * JDK version and the test date). Information includes server 
	 * side system properties and the client side properties, etc.
	 *
	 * @exception DataServerException if the file cannot be read
	 */
	public String getLogInfo (String key) 
		throws RemoteException, DataServerException;

	/** 
	 * Get the performance figures for a set of logs identified by the 
	 * JDK version and the test date
	 *
	 * @exception DataServerException if the files cannot be read
	 */
	public LogData[] getPerfLogs (String[] keys) 
		throws RemoteException, DataServerException;

  	/**
	 * Delete a logfile 
	 *
	 * @exception DataServerException if the file cannot be deleted
	 */
	public void deleteLog (String key)
		throws RemoteException, DataServerException;
}
