/*
 * @(#)LogViewer.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package visualBenchmarks;

import java.applet.*;
import java.awt.*;
import java.awt.List;
import NFBarchartApp;
import java.awt.event.*;
import java.rmi.*;
import java.rmi.registry.*;
import benchmarks.*;
import java.util.*;

/**
 * The LogViewer provides a GUI for the data stored by the DataCollector.
 * It displays data in the form of graphs, also allowing you to flexibly 
 * select what data is to be displayed. The data to be displayed is fetched
 * from the DataServer.
 *
 * @see DataServerImpl
 * @see benchmarks.DataCollector
 */

public class LogViewer extends Applet 
	implements ItemListener, ActionListener{

	private DataServer ds = null;

	private String errMsg;
	private MessageBox err;

	//The various components on the GUI
  
	private List JavaVersions  = new List();
	private List SelectedVersions = new List();
	private Choice Tests = new Choice();
	private TextArea SysInfo = new TextArea();
	private Button ItnsInfo, VectorInfo ;	  
	private NFBarchartApp vectorBar, itnsBar;
       	private Checkbox normalize = new Checkbox("Normalize",true);
	private CardLayout cl = new CardLayout();
	private Panel graphPanel = new Panel(cl);
 
	//Indicates whether the graph is to be displayed with the timings
	//normalized to a 100 iterations.
	private boolean normalizeTo100 = true;

	//List of tests performed in a particular version of the test-suite
	private Vector testList;

	//Normailizes the times for "itns" iterations to the time taken for
	//100 iterations.

	private float normalizeTo100(float time, int itns) {
		
		return time*100.0f/itns;
	}

	//Normalize an array of times (their sum) for "itns" iterations, to 
	//the time taken for a 100 iterations.

	private float normalizeTo100(float[] times, int itns) {

		float sum = 0.0f;

		for (int i = 0; i < times.length; i++) {
			sum += times[i];
		}

		return normalizeTo100(sum, itns);
	}	

	//Display the test details (the system properties on both the server
	//and the client side) for a particular test <key> in the test suite

	private void showTestInfo (String key) {

		String str;

		try {
			str = ds.getLogInfo(key).toString();
		       	SysInfo.setText(str);
		}catch (DataServerException e){
		  	showStatus("LogViewer Err: (showTestInfo()): " +
				e.getMessage());
			System.out.println("LogViewer Err: (showTestInfo()): "+
			  	e.getMessage());
			e.printStackTrace();
		}catch (RemoteException e) {
			showStatus("LogViewer Err: (showTestInfo()): " +
				e.getMessage());
			System.out.println("LogViewer Err: (showTestInfo()): "+
				e.getMessage());
			e.printStackTrace();
		}

	}


  	private void showVectorInfo() {
	
		String testName = Tests.getSelectedItem();
	
		String leftTitle;

		if (normalizeTo100) {
			leftTitle =new String("seconds/100 iterations");
		}else {
			leftTitle = new String("seconds");
		}

		StringBuffer cmdString = 
		  new StringBuffer("Background = (white,NONE,1,null,TILE); " +
			     "DwellLabel = (\"ON\", black); " +
			     "DwellBox = (yellow, RAISED, 2); " +
			     "Header = (\" " + testName + " \"); " + 
			     "HeaderBox = (gray, RAISED, 2); " +
			     "LeftTitle = (\" " + leftTitle +" \", " +
				",,,90 ); " +
			     "LeftTitleBox = (gray,RAISED,2); " +
			     "LeftFormat = (FLOAT, \"%g s\"); " + 
			     "Bar3DDepth = 10; " +
			     "BarWidth = 40; " +
			     "Legend = (\"VectorSize/Tree Depth\",white," +
				"\"TimesRoman\",12); "+
			     "LegendBox = (gray,RAISED,2); " +
			     "LegendLayout = (HORIZONTAL, BOTTOM); " +
			     "Grid = (lightgray, gray); " +
			     "GridLine = (HORIZONTAL, BAR); " +
			     "GraphType = STACK ;");
	    
		String DataSets = "DataSets = ";
		String[] DataSetValues ;
		String[] ActiveLabels ;
		String BarLabels = "BarLabels = ";

		LogData[] perfLogs;
		Hashtable currClientLog, perfData;
		ClientData currClientData = null; 
		AttribData attribData = null;
		Vector figures;
		PerfFigures perfFigure;
		int i, j, itns; 
		float time, overhead;
		float[] totalTime;
		int[] sizeList;
	
		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		showStatus("Loading data for the graph...");
	
		 
		//Since the list is in multiple selection mode, we need to 
		//use getSelectedIndex(es) even though we expect only a single
		//item

		int[] indexes = SelectedVersions.getSelectedIndexes();
		String[] keys = SelectedVersions.getSelectedItems();

		//Error (Allow viewing only one log at a time)
		if (indexes.length !=1) {
			errMsg = new String("Select ONE JDK version " +
				" to view!");
			err = new MessageBox(getFrame(),"ERROR!",errMsg);
			err.show();
			showStatus("");
			setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
			return ;
		}

		try {
			//We are using a LogData[] array only for compatibility
			//with the getPerfLogs() method. There is only a single
			//logData

			perfLogs = ds.getPerfLogs(keys);

			currClientLog = perfLogs[0].getClientLog();

			//NOTE: We assume here that each test is
			//conducted with only one client and a server.
			//Hence, we get the data for only the first 
			//client.
		
			currClientData = (ClientData)currClientLog.get(
					new Integer(1));

			if (currClientData != null) {

				//Get the performance figures for 
 				//this client
				perfData = currClientData.getPerfFigures();	

				//Extract the performance data for the
				//test chosen.
				attribData = (AttribData)perfData.get(testName);
				sizeList = attribData.getSizeList();

				figures = attribData.getFigures();

				//Specify the datasets to be displayed. In 
				//this case the datasets are the vector or
				//array sizes (also, for btrees, it is the
				//tree depth)

				for (j = 0; j < sizeList.length; j++) {
					if ( j != 0) 
						DataSets += ",";
					DataSets += " (\"" + sizeList[j] +
						"\")";
				}
				DataSets += ";";
				cmdString.append(DataSets);

				DataSetValues = new String[sizeList.length];
				ActiveLabels = new String[sizeList.length];

				//Get the data to be displayed
				for (j = 0; j <figures.size(); j++) {
			 		perfFigure = 
					  (PerfFigures)figures.elementAt(j);
					itns = perfFigure.getIterations();
					totalTime = perfFigure.getTotalTime();
					overhead = 
					  (float)perfFigure.getOverhead();
					if (normalizeTo100) {
					  overhead = 
					   normalizeTo100(overhead, itns);
					}

					//Set the DataSet Values
					for (i = 0; i < sizeList.length; i++){

					   if (j == 0){
					      DataSetValues[i] = 
					         new String("DataSet" +
						    (i+1) + " = " );
					      ActiveLabels[i] = 
						 new String("ActiveLabels" + 
						    (i+1) + " = " );
					   }
					   
					   if (normalizeTo100){
					      time = 
						normalizeTo100(totalTime[i],
							itns);
				 	   }else{
					      time = totalTime[i];
					   }

					   //Add the Dataset values to the
					   //appropriate Data set
					   //For the first value in a data
					   //set, skip the comma separator.

					   if (j ==0 ) {
					      DataSetValues[i] += time; 
					      ActiveLabels[i] += 
					        ("(\""+time+ "(" + overhead +
					          ")\")");
					
					   }else {
					      DataSetValues[i] += (","+time);
					      ActiveLabels[i] += 
					        (",(\""+time+ "("+ overhead + 
					           ")\")");
					   }
					}
				}

				//Append the semicolons at the end of the
				//the DataSetValues and ActiveLabels 
				//parameter strings

				for (i = 0; i < sizeList.length; i++) {
				    DataSetValues[i] += "; ";
				    cmdString.append(DataSetValues[i]);
				}

				for (i = 0; i < sizeList.length; i++) {
				    ActiveLabels[i] += "; ";
				    cmdString.append(ActiveLabels[i]);
				}


				//Generate the bar Labels 
				for (i = 0; i < figures.size(); i++) {
					perfFigure = 
					  (PerfFigures)figures.elementAt(i);
					itns = perfFigure.getIterations();
					if (i != 0) {
						BarLabels += ", ";
					}
					BarLabels += "\"(" + itns + " its.)\"";
				}
				BarLabels += ";";
				cmdString.append(BarLabels);

				//Add a footer.Note the \\ *are REQUIRED* 
				//so that the generated string contains \n

				String note=new String("Move mouse pointer " +
					"over the bars to get more details.");

				cmdString.append("Footer = (\"" + note + 
					"\", black, \"TimesRoman\",10) ;");

				vectorBar.loadParams(cmdString.toString());

				//Show the Vector Info panel on the card 
				//layout.
				cl.show(graphPanel,"Vector");
				vectorBar.loadParams("Update");
			}
		}catch (Exception e) {
			showStatus("LogViewer Err: (showItnsInfo()): " + 
				e.getMessage());
			System.out.println("LogViewer Err: (showItnsInfo()): "+
				 e.getMessage());
			e.printStackTrace();
		}
		showStatus("");
		setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		return;		

	}

	//Display the timings for a particular test and a particular log.
	//*NOTE*:At a time data for only one log can be displayed. The log
	//selected is the one selected in the selected versions list.
	//The data is displayed on the display panel in the form of a stacked
	//bar graph. The graph uses "active" labels to display exact values
	//(the values in the brackets represent loop overheads in measuring
	//the data).

	private void showItnsInfo() {

		String[] keys;
		LogData[] perfLogs = null;
		StringBuffer cmdString ;
		Hashtable currClientLog, perfData;
		ClientData currClientData = null;
		AttribData attribData = null;
		Vector figures = null;
		int itns, i, j ;
		PerfFigures perfFigure = null;
		float time = 0.0f,overhead = 0.0f;
		float[] totalTime;

		String testName = Tests.getSelectedItem();
		String leftTitle;

		if (normalizeTo100){
			leftTitle = new String("seconds/100 iterations");
		}else {
			leftTitle = new String("seconds");
		}

		String cmdHeader = 
		  new String("Background = (white,NONE,1,null,TILE); " +
			     "DwellLabel = (\"ON\", black); " +
			     "DwellBox = (yellow, RAISED, 2); " +
			     "Header = (\" " + testName + " \"); " + 
			     "HeaderBox = (gray, RAISED, 2); " +
			     "LeftTitle = (\" " + leftTitle + " \", " +
				",,,90 ); " +
			     "LeftTitleBox = (gray,RAISED,2); " +
			     "LeftFormat = (FLOAT, \"%g s\"); " + 
			     "Bar3DDepth = 10; " +
			     "BarWidth = 40; " +
			     "Legend = (\"\",white,\"TimesRoman\",12); "+
			     "LegendBox = (gray,RAISED,2); " +
			     "LegendLayout = (HORIZONTAL, BOTTOM); " +
			     "Grid = (lightgray, gray); " +
			     "GridLine = (HORIZONTAL, BAR); ");

		String DataSets = "DataSets = ";
		String[] DataSetValues = null;
		String[] ActiveLabels= null;
		String BarLabels, BarActiveLabels;


		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		showStatus("Loading data for the graph...");
		boolean firstDataSet = true;	
		 
		int numItems = SelectedVersions.getItemCount();

		//Error
		if (numItems == 0) {
			errMsg = new String("No JDK Versions selected!");
			err = new MessageBox(getFrame(),"ERROR!",errMsg);
			err.show();
			showStatus("");
			setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
			return ;
		}


		keys = new String[numItems];
		for (i = 0; i < numItems; i++) {
			keys[i] = new String(SelectedVersions.getItem(i));
		} 
		cmdString = new StringBuffer(cmdHeader);
		try {
			perfLogs = ds.getPerfLogs(keys);

			for (i = 0; i < perfLogs.length; i++) {

				currClientLog = perfLogs[i].getClientLog();

				//NOTE: We assume here that each test is
				//conducted with only one client and a server.
				//Hence, we get the data for only the first 
				//client.
		
				currClientData = (ClientData)currClientLog.get(
					new Integer(1));

				if (currClientData != null) {

					//Get the performance figures for 
 					//this client
					perfData =
					   currClientData.getPerfFigures();

					//Extract the performance data for the
					//test chosen.
					attribData = 
					   (AttribData)perfData.get(testName);

				 	figures = 
					   attribData.getFigures();

					//Initialize the DataSetValues string
					//array to the appropriate size.

					if (firstDataSet) {
					    DataSetValues = 
						new String[figures.size()];
					    ActiveLabels = 
						new String[figures.size()];
					}
				
					for (j = 0; j <figures.size(); j++) {

					   perfFigure = 
					     (PerfFigures)figures.elementAt(j);
					   itns = perfFigure.getIterations();

					   totalTime =perfFigure.getTotalTime();

					   if (normalizeTo100) {
					      time = normalizeTo100(
					        totalTime,itns);
					   }else {
					      for (int k = 0; 
						   k < totalTime.length ; k++){
						time += totalTime[k];
					      }
					   }

					   overhead = 
					      (float)perfFigure.getOverhead();
					   if (normalizeTo100) {
					      overhead = 
						normalizeTo100(overhead, itns);
					   }
					   
					   if (firstDataSet) {
						if (j != 0) {
							DataSets += ",";
						}
						DataSets +=  "(\"" + itns + 
							" Itns.\") ";
						DataSetValues[j] = 
							"DataSet" +(j+1) +"= ";
						ActiveLabels[j] = "Active" +
							"Labels" + (j+1) +"= ";
					   }

					   if (firstDataSet) {
					 	DataSetValues[j] += time;
					   }else { 
					   	DataSetValues[j] += (","+time);
					   }

					   if (firstDataSet) {
						ActiveLabels[j] += ("(\"" +
						 time+ "(" + overhead + ")\")");
					   }else {
					   	ActiveLabels[j] += (",(\"" + 
						 time+ "(" + overhead + ")\")");
					   }
					}

					if (firstDataSet) {
						DataSets += ";";
					}
					firstDataSet = false;
				}
			}

			cmdString.append(DataSets);

			for (i = 0; i < DataSetValues.length; i++){
				DataSetValues[i] += "; ";
				cmdString.append(DataSetValues[i]);
			}

			for (i = 0; i < ActiveLabels.length; i++){
				ActiveLabels[i] += "; ";
				cmdString.append(ActiveLabels[i]);
			}

			//Generate the bar Labels 
			BarLabels = "BarLabels = ";
			for (i = 0; i < keys.length; i++) {
				if (i != 0) {
					BarLabels += ", ";
				}
				BarLabels += "\"(" + (i+1) + ")\"";
			}
			BarLabels += ";";
			cmdString.append(BarLabels);

			//Generate the bar active Labels
			BarActiveLabels = "BarActiveLabels = ";
			for (i = 0; i < keys.length; i++) {
				if (i != 0) {
					BarActiveLabels += ", ";
				}
				BarActiveLabels += 
					 "(\"" + keys[i] + "\")";
			}
			BarActiveLabels += ";";
			cmdString.append(BarActiveLabels);

			//Add a footer.Note the \\ *are REQUIRED" so that the
			//generated string contains \n
			String note=new String("Move mouse pointer over the " +
			  "bars or\\n the bottom labels to get more details.");

			cmdString.append("Footer = (\"" + note + "\", black, "+
				"\"TimesRoman\",10) ;");

			itnsBar.loadParams(cmdString.toString());
			cl.show(graphPanel,"Itns");
			itnsBar.loadParams("Update");
		}catch (Exception e) {
			showStatus("LogViewer Err: (showItnsInfo()): " + 
				e.getMessage());
			System.out.println("LogViewer Err: (showItnsInfo()): "+
				 e.getMessage());
			e.printStackTrace();
		}
		showStatus("");
		setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		return;
	}

	/**
	 * Copy elements from the source list to the destination list or 
	 * remove elements from the source list.
	 *
	 * @param src    the source list
	 * @param dest   the destination list
	 * @param all    a boolean parameter indicating wheter <em>all</em>
	 *	         are to be copied/moved
	 * @param remove if set to true, the elements are <em>removed</em> 
	 *		 from the source list, rather than being 
	 * 		 <em>copied</em> to the destination list.
	 */
  
	protected void move (List src,List dest,boolean all,boolean remove) {
		
		int i;
		
		if (all) {
			if (remove) {
				src.removeAll();
			}else {
			   for (i = 0; i < src.getItemCount(); i++) {
				dest.add(src.getItem(i));
			   }				
			}
		}else {
			String[] items = src.getSelectedItems();
			int[] indexes = src.getSelectedIndexes();

			for (i = 0; i < indexes.length; i++) {

				if (remove) {
					src.remove(items[i]);
				} else {
					dest.add(items[i]);
				
					//Select it  
					dest.select(dest.getItemCount() - 1);

					//Make the added items visible 
					if (i==0) {
						dest.makeVisible(
							dest.getItemCount()-1);
					}
				}
			}
		}
		deselectAll(src);
	}

	/**
	 * Deselect all the items in a list.
	 *
	 * @param l the list whose items are to be deselected.
 	 */

	protected void deselectAll(List l) {

		int i;

		for ( i = 0; i < l.getItemCount(); i++) {
			l.deselect(i);
		}

	}

	/**
	 * Gets the frame container for the applet (usually this is either
	 * the appletviewer or the browser window
	 *
	 * @return the Frame container.
	 */

	protected Frame getFrame(){

		//To get the frame container for the applet, we recursively
		//get the parent component, till the retreived parent is an 
	 	//instance of the Frame class.

		Object anchorPoint = getParent();
		while (!(anchorPoint instanceof Frame)){
			anchorPoint = ((Component)anchorPoint).getParent();
		}
		
		return (Frame)anchorPoint;
	}

	/**
	 * Determines if a particular test in a testsuite uses arrays/vectors
	 * or some such structure ,for which the test involves parameters
	 * other than the range of iterations.
	 *
	 * @return <i>true</i> if the test involves the use of such data
	 * 		       structures, <i>false</i> if not.
	 */
	 
	protected boolean areArraysUsed(String testName) {

		TestInfo currTestInfo;
		String currTest;
		boolean arraysUsed = false;

		for (int i = 0; i < testList.size(); i++){
			currTestInfo = (TestInfo)testList.elementAt(i);
			currTest = currTestInfo.getTestName();
			if (currTest.equals(testName)){
				arraysUsed = 
				  currTestInfo.areArraysUsed()?true:false;
				break;
			}
		}
		return arraysUsed;
	}
	
	/**
	 * Return the applet information such as creator, purpose, etc.
	 *
	 * @return the applet information
	 */
	public String getAppletInfo() {

		String info = new String("LogViewer for the benchmark logs\n" +
				"Created by : Nandkumar K. Kesavan\n");
		
		return info;
	}

	public void init() {

		int i;

		Button button = null;
		Box box = null;

		GridBagLayout gbl = null;
		GridBagConstraints gbc = null;


		setCursor(new Cursor(Cursor.WAIT_CURSOR));
		showStatus("Loading data from log file...");

		//Set initial properties

		//Make the system props text area "not editable"
		SysInfo.setEditable(false); 

		try {

			Panel versions = new Panel();
			gbl = new GridBagLayout();
			gbc = new GridBagConstraints();

			//Layout the lists and the buttons on the versions
			//Panel. 
			//NOTE: Each time a new component is added , it's
			//gridbagconstraints are set by making only the changes
			//needed from those for the *previous* component.
			//Hence, any new components to be added must be 
			//added at the end

			versions.setLayout(gbl);
		
			//Place the Available Versions label in the panel
			gbc.gridx = 0;
			gbc.gridy = 0;
			gbc.gridwidth = gbc.gridheight = 1;
			gbc.fill = GridBagConstraints.NONE;
			Label lbl = new Label("Available Versions");
			gbl.setConstraints(lbl, gbc);
			versions.add(lbl);

			//Place the JavaVersions list in the versions panel

			gbc.gridx = 0;
			gbc.gridy = 1;
			gbc.gridwidth = 1;
			gbc.gridheight = GridBagConstraints.REMAINDER;
			gbc.fill = GridBagConstraints.BOTH;
			gbc.weightx = 2.0;
			gbl.setConstraints(JavaVersions, gbc);
			versions.add(JavaVersions);

			//Place the SelectedVersions label
			gbc.gridx = 2;
			gbc.gridy = 0;
			gbc.gridwidth = gbc.gridheight = 1;
			gbc.fill = GridBagConstraints.NONE;
			lbl = new Label("Selected Versions");
			gbl.setConstraints(lbl, gbc);
			versions.add(lbl);

			//Place the SelectedVersions list in the versions panel

			gbc.gridx = 2;   //Other params remain as for JavaVers.
			gbc.gridy = 1;
			gbc.gridwidth = 1;
			gbc.gridheight = GridBagConstraints.REMAINDER;
			gbc.fill = GridBagConstraints.BOTH;
			gbc.weightx = 2.0;
			gbl.setConstraints(SelectedVersions, gbc);
			versions.add(SelectedVersions);

			//Place the Buttons in the panel

			gbc.gridx = 1;
			gbc.gridy = 1;
			gbc.gridwidth = gbc.gridheight = 1;

			//Setting weightx to 1.0 for the col.of buttons and to
			//2.0 for the lists ensures that the extra column space
			//is distributed in the ratio 2:1:2 when resized.
 
		 	gbc.weightx = 1.0;
			button = new Button("Add >");
			button.addActionListener(this); 
			gbl.setConstraints(button,gbc);
			versions.add(button);
			
			gbc.gridy = 2;
			button = new Button("Remove <");
			button.addActionListener(this);
			gbl.setConstraints(button,gbc);
			versions.add(button);

			gbc.gridy = 3;
			button = new Button("Clear <<");
			button.addActionListener(this);
			gbl.setConstraints(button,gbc);
			versions.add(button);
			
			gbc.gridy = 4;
			button = new Button("Show Details");
			button.addActionListener(this);
			gbl.setConstraints(button,gbc);
			versions.add(button);

			gbc.gridy = 5;
			button = new Button("Delete Log");
			button.addActionListener(this);
			gbl.setConstraints(button,gbc);
			versions.add(button);

			//Populate the two lists.

			JavaVersions.setMultipleMode(true);
			SelectedVersions.setMultipleMode(true);
			ds = (DataServer)Naming.lookup("//" +
				getCodeBase().getHost() + ":1234/DataServer");
			String[] versionList = ds.getVersions();
			if (versionList == null) {
			   showStatus("No log files available... Exiting");
			   System.out.println("No log files available...");
			   System.exit(-1);
			}
			for (i=0; i< versionList.length; i++) {
			   JavaVersions.add(versionList[i]);
			}
			JavaVersions.addItemListener(this);
			JavaVersions.addActionListener(this);

			testList = ds.getListOfTests();
			for (i=0; i < testList.size(); i++) {
			   Tests.add(
			     ((TestInfo)testList.elementAt(i)).getTestName());
			}
			Tests.addItemListener(this);

			showStatus("");


			//Arrange the componenents on the applet
			//As noted for the versions panel, any new components
			//must be added at the end.
			Panel left = new Panel();

			gbl = new GridBagLayout();
			gbc = new GridBagConstraints();
			left.setLayout(gbl);

			box = new Box(Tests,"Select test to view");
			gbc.gridx = 0;
			gbc.gridy = 0;
			gbc.gridwidth = 3;
			gbc.fill = GridBagConstraints.BOTH;
			gbl.setConstraints(box,gbc);
			left.add(box);

			vectorBar = new NFBarchartApp(this);
			itnsBar = new NFBarchartApp(this);
			vectorBar.init();
			itnsBar.init();
			vectorBar.start();
			itnsBar.start();
			graphPanel.add(vectorBar,"Vector");
			graphPanel.add(itnsBar,"Itns");

			Panel display = new Panel();
			GridBagLayout displayLayout = new GridBagLayout();
			GridBagConstraints displayConstraints = 
				new GridBagConstraints();
			display.setLayout(displayLayout);
			
			//Add graph components to the display panel
			displayConstraints.gridwidth = 3;
			displayConstraints.gridheight = 3;
			displayConstraints.weightx = 1.0;
			displayConstraints.weighty = 1.0;
			displayConstraints.fill = GridBagConstraints.BOTH;
			displayLayout.setConstraints(
				graphPanel, displayConstraints);
			display.add(graphPanel);
			
			displayConstraints.gridx = 0;
			displayConstraints.gridy = 3;
			displayConstraints.gridheight = 1;
			displayConstraints.gridwidth = 1;
			displayConstraints.fill = GridBagConstraints.NONE;
			displayConstraints.weightx = 0.0;
			displayConstraints.weighty = 0.0;
			displayConstraints.insets.top = 5;
			displayConstraints.insets.bottom = 5;
			ItnsInfo = new Button("ItnsInfo");
			ItnsInfo.addActionListener(this);
			displayLayout.setConstraints(
				ItnsInfo,displayConstraints);
			display.add(ItnsInfo);

			displayConstraints.gridx = 2;
			displayConstraints.anchor = GridBagConstraints.EAST;
			VectorInfo = new Button("VectorInfo");
			VectorInfo.addActionListener(this);
			VectorInfo.setEnabled(false);
			displayLayout.setConstraints(
				VectorInfo, displayConstraints);
			display.add(VectorInfo);
	
			gbc.fill = GridBagConstraints.BOTH;
			box = new Box(display,"Graph");	
			gbc.gridx = 0;
			gbc.gridy = 1;
			gbc.gridwidth = 3;
			gbc.gridheight = 4;
			gbc.weightx = gbc.weighty = 1.0;

			//Restore default
			gbc.anchor = GridBagConstraints.CENTER; 
			gbl.setConstraints(box, gbc);
			left.add(box);
			
			Panel right = new Panel();

			gbl = new GridBagLayout();
			gbc = new GridBagConstraints();
			right.setLayout(gbl);

			box = new Box(versions,"Select JDK versions");
			gbc.gridx = 0;
			gbc.gridy = 0;
			gbc.gridwidth = 1;
			gbc.gridheight = 3;
			gbc.weightx = 1.0;
			gbc.weighty = 0.0;
			gbc.fill = GridBagConstraints.BOTH;
			gbl.setConstraints(box, gbc);
			right.add(box);
			
			box = new Box(normalize,"Graph format");
			normalize.addItemListener(this);
			gbc.gridx = 0;
			gbc.gridy = 3;
			gbc.gridheight = 1;
			gbc.gridwidth = 3;
			gbc.weightx = gbc.weighty = 1.0;
			gbc.fill = GridBagConstraints.NONE;
			gbl.setConstraints(box,gbc);
			right.add(box);

			gbc.gridx = 0;
			gbc.gridy = 4;
			gbc.gridheight = 1;
			gbc.weightx = gbc.weighty = 0.0;  
			gbc.fill = GridBagConstraints.NONE;
			gbl.setConstraints(SysInfo, gbc);
			right.add(SysInfo);

			GridLayout gl = new GridLayout();
			gl.setHgap(5);
			setLayout(gl);
			add(left);
			add(right);


		}catch (Exception e) {
			showStatus("LogViewer Err: (init()): " +
				e.getMessage());
			setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
			System.out.println("LogViewer Err: (init()): " + 
				e.getMessage());
			e.printStackTrace();
		}
		showStatus("");
		setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
	}

	public void actionPerformed(ActionEvent evt) {

		String arg = evt.getActionCommand();
	
		if (arg.equals("Add >")){
			move(JavaVersions, SelectedVersions, false,false);
		} else if (arg.equals("Remove <")){
			move(SelectedVersions, JavaVersions, false, true);
		} else if (arg.equals("Clear <<")) {
			move(SelectedVersions, JavaVersions, true, true);
		} else if (arg.equals("Show Details")) {
		 	String[] items = SelectedVersions.getSelectedItems();
			
			if (items.length == 1) {	
				showTestInfo (items[0]);
			}else {
			    errMsg = new String ("Can show details for only "+
					"one test at a time!");
			    err = new MessageBox(getFrame(),"ERROR!",errMsg);
			    err.show();
			} 
		} else if (arg.equals("Delete Log")) {
			String[] items = SelectedVersions.getSelectedItems();
	
			if (items.length == 0) {
				errMsg = new String ("No files selected!");
				err = new MessageBox(getFrame(),"ERROR!",
					errMsg);
				err.show();
			}else {
				String msg = new String("Do you wish to "+
					"delete these files?"); 
				QueryDialog qd =
				  new QueryDialog(getFrame(),"Confirm",msg);
				
				qd.show();
				if (qd.response == QueryDialog.ACCEPT){
				  try {
					for (int i = 0; i < items.length; i++){
					   ds.deleteLog(items[i]);
					   SelectedVersions.remove(items[i]);
					   JavaVersions.remove(items[i]);
					}
				  }catch (Exception e){
				        showStatus("LogViewer Err: " +
					     e.getMessage());
					System.out.println("LogViewer Err: "+
					     "actionPerformed(): " +
					     e.getMessage());
					e.printStackTrace();
				  }
				}
			}
		} else if (arg.equals("ItnsInfo")) {
			showItnsInfo();
		} else if (arg.equals("VectorInfo")) {
			showVectorInfo();
		}
	}

	public void itemStateChanged(ItemEvent evt) {

		if (evt.getSource() instanceof List) {
			List targetList = (List)evt.getSource();

			if (targetList == JavaVersions) {
				deselectAll (SelectedVersions);
			}else if (targetList == SelectedVersions) {
				deselectAll (JavaVersions);
			}
		}else if (evt.getSource() instanceof Choice) {
			Choice targetChoice = (Choice)evt.getSource();
			
			if (targetChoice == Tests) {
				String testName = Tests.getSelectedItem();
				VectorInfo.setEnabled(
					areArraysUsed(testName)?true:false);
			}
		}else if (evt.getSource() instanceof Checkbox) {
			Checkbox targetCheckbox = (Checkbox)evt.getSource();
			
			if (targetCheckbox == normalize) {
				normalizeTo100 = !(normalizeTo100);
				if (vectorBar.isVisible()) {
					showVectorInfo();
				}else {
					showItnsInfo();
				}
			}
		}
	}

	public void destroy() {

		//Explicitly release the connetion to the server.

		ds = null;

	}

}
