/*
 * @(#)MessageBox.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package visualBenchmarks;

import java.awt.*;
import java.awt.event.*;
import java.awt.TextArea.*;
import java.lang.*;

/**
 * An extension of the java.awt.Dialog class.
 * Constructs a <em>modal</em> dialog box wiht an  message and a title
 */

public class MessageBox extends Dialog implements ActionListener
{
	/**
	 * An OK button displayed towards the bottom of the dialog box. Only
	 * when the button is pressed, can the user resume normal execution.
	 */
	protected Button button;
  
  	/**
	 * Region where the message is displayed
	 */
	protected TextArea text;

	/**
	 * Constructs a Message box with a title and an error message
	 *
	 * @param parent  the Frame to which the dialog box is attached
	 * @param title   the title for the Message box
	 * @param message the message
	 */
	public MessageBox(Frame parent, String title, String message) 
	{
		super(parent, title, true);

		this.setLayout(new BorderLayout(15, 15));   

		text = new TextArea(message,3 ,50,TextArea.SCROLLBARS_NONE );

		//Disable editing of the error message
		text.setEditable(false);

		//Adding the message box to the center causes it to occupy
	 	//all the space left after the other component has been layed 
		//out
		this.add("Center", text);

		// Create and add and OK button.
		button = new Button("OK");
		button.addActionListener(this);
		Panel p = new Panel();
		p.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));
		p.add(button);

		//Center the Dialog box in the parent's window space (approx.)
		setLocation(parent.getSize().width/2,
			 parent.getSize().height/2);

		this.add("South", p);

		// Resize the window to the preferred size of its components.
		this.pack();
	}

	public void actionPerformed(ActionEvent evt)
	{
		String str = evt.getActionCommand();

		if(str.equals("OK"))
		{
			this.setVisible(false);
			this.dispose();
		}
	}
}

