/*
 * @(#)PerfFigures.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.io.*;

/**
 * Stores the performance timings for a particular number of 
 * iterations , a specific attribute and (possibly) a range of 
 * vector sizes (<i> or tree depths, or some such paramter which can be 
 * specified for the test in addition to the range of iterations.</i>) 
 * These are stored as elements in a vector (of size <i>n</i> where <i>n</i> 
 * is the number of loops tested for that attribute. 
 *
 * @see AttribData
 * @see ClientData
 * @see LogData
 */
public class PerfFigures implements Serializable {

        /** 
	 * The number of iterations carried out for this particular test
	 */
	protected int iterations;

	/**
	 * A set of timings for a particular number of iterations and over 
	 * a range of values for a second parameter such as the vector size
	 * for tests involving arrays or the tree depth in tests passing
	 * a bTree object.
	 */

 	protected float[] totalTime;
 	protected float[] stotalTime;


	/**
	 * The loop overhead incurred in taking the measurements. Also includes
	 * the overhead in taking the timestamps.
	 * This time is in milliseconds.
	 */
	protected long Overhead;

      	/**
	 * Constructs a PerfFigures object for data <em>not</em> involving
	 * arrays/vectors
	 * 
	 * @param itns      the number of iterations      
	 * @param totTime   the total Time for the operation, excluding  
	 *                  overheads, in seconds.
	 * @param stotTime  the total Time for the serialization operation, 
         * 		    excluding overheads, in seconds.
	 * @param loopOvrhd the overheads incurred in taking the measurement,
	 *                  in milliseconds.
	 */

	public PerfFigures (int itns,float totTime, float stotTime, long loopOvrhd){

		//For measurements not involving arrays/vectors, the arraySize
		//list is set to null. When using PerfFigures for any purpose
		//like generating reports, etc. arraySize should be tested for
		//null to determine whether the test involved arrays/vectors

		iterations = itns;
		totalTime= new float[1];
		stotalTime= new float[1];
		totalTime[0] = totTime;
		stotalTime[0] = stotTime;
		Overhead = loopOvrhd;

	}


	/**
	 * Constructs a PerfFigures object for tests <em>involving</em> arrays
	 * or vectors.
	 * 
	 * @param itns      the number of iterations      
	 * @param totTime   a vector of the time for the operation for a range
	 *                  of array sizes, excluding  overheads, in seconds. 
	 *                  The vector size is equal to the no. of vector  
	 *                  sizes used in the test
	 * @param stotTime  a vector of the time for the serialization 
         *		    operation for a range of array sizes
	 * @param loopOvrhd the overheads incurred in taking the measurement,
	 *                  in milliseconds.
	 * 
	 * @see AttribData
	 */

	public PerfFigures (int itns ,float[] totTime, float[] stotTime, long loopOvrhd){
		
		iterations = itns;
		totalTime = new float[totTime.length];
		stotalTime = new float[stotTime.length];

		System.arraycopy(totTime,0, totalTime, 0, totTime.length);
		System.arraycopy(stotTime,0, stotalTime, 0, stotTime.length);

		Overhead = loopOvrhd;

	}

	public int getIterations() {

		return iterations;

	}

	public float[] getTotalTime () {

		return totalTime;

	}

	public float[] getserTotalTime () {

		return stotalTime;

	}
	public long getOverhead () {

		return Overhead;

	}
	
}




