/*
 * @(#)Repository.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.io.*;
import java.util.*;

/**
 * The Repository server which stores all the log files organized into 
 * appropriate directories and subdirectories.
 */

public class Repository extends UnicastRemoteObject
	implements RepositoryInterface {

	private String name;

	private static File logDir;

	private static Properties props = new Properties();

	public Repository (String name)
		throws RemoteException {
		
		this.name = name;

	}

	/**
         * Adds a logFile to the appropriate directory. 
	 * 
	 * @param data    the LogData to be stored in the log file
	 * @param name    the log filename
	 * @param version the test suite version
	 * @exception RemoteException     if the remote operation fails
	 * @exception RepositoryException if the log file cannot be stored
	 */
	public void addLog (LogData data,String name, String version) 
		throws RemoteException, RepositoryException {

		ObjectOutputStream os = null;
		
		String subDirName = new String ("logs" +
			 version.replace('.','_'));
			
		File subDir = new File(logDir, subDirName);
		
		if (!subDir.isDirectory()){
			if (!(subDir.mkdir())) {
				System.out.println("Could not create " +
				   "subdirectory " + subDir);
			}
		}		
		File logFile = new File(subDir,name);

		if (!logFile.exists()) {
		   try {
			os = new ObjectOutputStream(
			   new FileOutputStream(logFile));
			os.writeObject(data);
			os.flush();
			os.close();
		   }catch (IOException e) {
			System.out.println("Repository err: addLog(): " +
				e.getMessage());
			e.printStackTrace();
			throw new RepositoryException(e.getMessage());
		   }		
		}else {
		   System.out.println("A log with this name already exists. "+
			"Delete it or specify a new name...");
		   throw new RepositoryException("Log with the name " +
				logFile.getPath() + " already "+
				"exists. Delete or specify new name...");
		}
	}	
		

	public static void main(String[] args){
		
		Registry registry;
		
		System.setSecurityManager(new RMISecurityManager());

		try {
		  	props.load(new FileInputStream(
				"repository.properties"));

			Repository repository = new Repository("Repository");
			LocateRegistry.createRegistry(Registry.REGISTRY_PORT);
			registry = LocateRegistry.getRegistry(
				Registry.REGISTRY_PORT);
			
			registry.rebind("Repository",repository);
			System.out.println("Server \"Repository\" bound in " +
				"registry...");

			String dirName = props.getProperty("LogDir",
					"benchmarkLogs");
			logDir = new File(dirName);
			if (logDir.exists()) {
				if (!logDir.isDirectory()){
				  System.out.println(dirName + " is not a " +
					"directory...");
				  System.exit(-1);
				}
			}else {
			   	if (!(logDir.mkdirs())) {
				   System.out.println("Could not create " +
					"directory " + dirName);
				   System.exit(-1);
				}
			}	
		}catch (Exception e){
			System.out.println("Repository err: main(): " +
				e.getMessage());
			e.printStackTrace();
		}
	}
}

