/*
 * @(#)TestClient.java	1.4 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.io.*;
import java.io.FileInputStream;
import java.rmi.*;
import java.util.*;

// Test RMI performance .Currently meters null calls, RMI calls with integer 
// arguments, object arguments, remote parameters and returning remote objects,
// both new (newly created by the server ) and old (already seen by the client).

public class TestClient 
	extends java.rmi.server.UnicastRemoteObject 
	implements ClientInterface, Runnable  {

	//The path to the server
	private static String host ;


	//Controls Verbose or Silent Mode for the client.(def. Silent)
	//Set by the command line argument -showProg
	private static boolean showProg = false;

        //If a version string is provided in the commandline args. this string
        //is set to that, else null;
        private static String version = null;

	// The list of tests in the benchmark
	private static String[][] tlist= {
	   { "1", "Null call"},
	   { "2", "Remote method with int arg"},
	   { "3", "Remote method with intArray arg"},
	   { "4", "Remote method with byteArray arg"},
	   { "5", "Remote method with doubleArray arg"},
	   { "6", "Remote method with obj arg"},
	   { "7", "Remote method with int & obj arg"},
	   { "8", "Remote method with objArray arg"},
	   { "9", "Remote method with remote arg"},
	   { "10", "Remote method returning remote obj "},
	   { "11", "Remote method returning vector of new and oldremote objs "},
	   { "12", "Remote method throwing exceptions"},
	   { "13", "Remote method with obj Graph (Btree)   arg"}
	};
	static int[] seltests;
	static int[] deftests;
	static int[] testsrun;
	static boolean deft=true;

	//Variables holding the number of calls to be made, and arrays
	//indicating the number of iterations to be made for each loop

	private static int nullCalls, nullCallsItns[];

	private static int intArg, intArgItns[];

	private static int intArrayArg, intArrayItns[], 
		intArraySizes, intArraySizeList[];

	private static int byteArrayArg, byteArrayItns[], 
		byteArraySizes, byteArraySizeList[];

	private static int doubleArrayArg, doubleArrayItns[], 
		doubleArraySizes, doubleArraySizeList[];

	private static int objArg, objArgItns[];

	private static int twoArgCalls, twoArgItns[];

	private static int objArrayArg, objArrayItns[], 
		objArraySizes, objArraySizeList[];

	private static int remoteArg, remoteArgItns[];

	private static int returnRemote, returnRemoteItns[];

	private static int returnRemoteVector, remoteVectorItns[];

	//the size of the vector of remote objects
	private static int vectorSizes, vectorSizeList[];

	private static int exceptionCalls, exceptionCallsItns[];

	private static int bTreeArg, bTreeArgItns[],
		bTreeDepths, bTreeDepthList[];

	//The client system properties

	private static Properties clientProps;

	//The Client Identification number assigned by the data Collector
	//when a client logs in it's system properties.
	private int clientID  = NOT_ASSIGNED;	

	/**
	 * The maximum number of arguments that can be specified in the 
	 * string of integers in the properties file. 
	 *
	 * @see #fillArray
	 */
	
	//The name is somewhat of a misnomer. Although initially, it stood 
 	//for the number of loop sizes you could specify for a test in the
	//properties file, it now stands for the maximum number of values
	//you can specify for a particular parameter in the properties file
	//I was just too lazy to change it everywhere it is used. Maybe, I
	//should quote the classic "It remains for historical reasons" :)

	public static int MAX_LOOPS = 5;

	/**
	 * Indicates that the dataCollector could not log in the client's 
	 * system properties. On getting this value as the return value in
	 * the logClientDesc() method, the client periodically attempts to 
	 * log it's system properties. If the system properties cannot be
	 * logged even after MAX_ATTEMPTS the client exits with an error msg.
	 * This must be the same as it's counter part in the DataCollector
	 *
	 * @see DataCollector#NOT_ASSIGNED
	 */
	public static int NOT_ASSIGNED = -1;

	/**
	 * Maximum number of connection attempts made by the client to register
	 * with the data Collector,  before exiting with an error message.
	 */
	public static int MAX_ATTEMPTS = 10;

	//Calculate the loop overhead for a particular number of iterations.
  	private long getLoopOvrhd (int numItns) {
		
		timestamp startStamp, endStamp;
		long loopOvrhd;

	      	int dummy = 0, j;

		//Measure loop overhead
                startStamp = new timestamp(); 
                for (j = 0; j < numItns; j++){

			//A dummy statement to prevent the 
			//compiler from bypassing a null loop
			//If it still optimizes, I give up :(

               		dummy ++;
                }
                endStamp = new timestamp();
                loopOvrhd = endStamp.sub(startStamp).getTime();

		return loopOvrhd;

	}

	/*
	 * Trim the integer array to the right size, called after fillArray()
	 * since fillArray() uses the arrays of size MAX_LOOPS. If the arrays
	 * are not trimmed to the right size, you cannot use the array size
	 * to figure out the number of loops tested for a particular test.
	 */

	private static int[] trimToSize(int size, int[]array) {
		
		int i;
		int[] tmp = new int[size];

		for (i = 0; i < size; i++ ) {
			tmp[i] = array[i];
		}

		return tmp;
	}

	
	public TestClient() 
		throws RemoteException {
	}


	/**
	 * Parses a string of integers separated by the "separator" character
	 * and fills an integer array with these.
	 *
	 * @param  str       the string of integers to be parsed
	 * @param  separator the character separator for the list
	 * @param  Itns      the integer array filled in by this function 
	 *                   with integers parsed from the string str.
	 * @return           Number of integers in the list.     
	 */ 

	public static int fillArray(String str, char separator, int[] Itns){

		String substring;
		int i,offset, count, where;
		int[] tmp;

		for (offset = 0,count = 0; 
			(where = str.indexOf(separator,offset))>=0;
			offset = where + 1) {			
			if (where > 0){ 

				//Extract the integer

				substring = str.substring(offset,where);
				Itns[count] = Integer.parseInt(substring);
				count++;
			}
		}		

		//If the last character is not a separator, extract the 
		//trailing substring as an integer.

		if (str.charAt(str.length() - 1)!= separator) {
			substring = str.substring(str.lastIndexOf(separator)+1 ,
							str.length());
			Itns[count] = Integer.parseInt(substring);
			count++;
		}

		return count;
	}


	/**
	 * A null method called by the data Collector to test the connection
	 *
	 * @see DataCollector#run
	 * @exception RemoteException if the remote call fails
	 */ 

	public void ping() throws RemoteException{

		return;
	}

	/**
	 * Implement the various tests. If any new tests are incorporated, they
	 * need to be called from here. The Client is flexible enough to 
	 * incorporate the new changes without making changes all over the place
	 * The results simply need to be put in the clientdata which is passed
	 * to the server. 
	 */
	public void run () {

		// Start and End Stamps measure the various calls 
		// wherease loopStart and loopEnd are used to calculate the
		// the loop overhead. 

		timestamp startStamp, endStamp, loopStart, loopEnd;
		long totalTime, loopOvrhd;
		float[] totalTimeList;
		ObjClass[] ObjClassArray;

// Added by Vella 
	    timestamp serstartStamp, serendStamp;
	    long stotalTime;
	    float[] stotalTimeList;
	    int m, n;
	
	    ByteArrayOutputStream baos, baos1;
	    ByteArrayInputStream bais, bais1;
	    ObjectOutputStream oos, oos1;
	    ObjectInputStream ois, ois1;
//
		//Vector of remote objects received from the server on a 
		//getVector() call.

		Vector v;

		//Hashtable used to store the Performance figures 
		//Key: Attribute Measured.
		//Data: a Vector of PerfFigures objects showing 
		//      1. the number of iterations.
		//      2. the list of vector sizes, if applicable
		//      3. the total time taken in the operation for the vector
		//	   sizes. 
		//      4. the overhead.

		Hashtable data = new Hashtable();

		//Vector of PerfFigures for each measured attribute.
		Vector figures = new Vector();	

		//Number of attempts made by the client to connect to the 
		//server.
		int connectionAttempts = 0;

		//A unique identifier for each test
		String testName;	
		Integer testNum;

		//Data for a particular test for each client
		AttribData attribData, serattribData;


		int iterations,  i, j, k, count;
		



		try {

			String tmp;

			if (!(host.equals(""))){
				tmp = new String("//");
				tmp += host;
			}else
				tmp = new String(host); 

			TestInterface serverRef = (TestInterface)Naming.lookup(
				tmp + "/DataCollector");
	       		CollInterface collRef = (CollInterface)Naming.lookup(
				tmp + "/DataCollector");




		    	//Log the Client properties with the data Collector.
		    	//Also returns a clientID which is used to identify
		    	//the client to the data Collector when logging in
		    	//the performance figures.
		
			do {
		    		clientID = 
					collRef.logClientDesc(clientProps,this);
				if (clientID == NOT_ASSIGNED ) {
					if (showProg) {
						System.out.println("client ID" +
						" not assigned. Will try " +
						"again....");
					}
					connectionAttempts++;
				} else if (showProg){
					System.out.println("client ID: " +
					  clientID);
				}
				Thread.sleep(2000);
			} while ((clientID == NOT_ASSIGNED) &&
					(connectionAttempts <= MAX_ATTEMPTS));
			
			if (connectionAttempts > MAX_ATTEMPTS) {
				System.out.println("Could not connect to " +
					"server. Logging out...");
				System.exit(1);
			}

// Added loop to allow tests to be run individually  - Vella
	for ( int tn=0;  tn < testsrun.length; tn++ ) {
	    testNum=new Integer(testsrun[tn]);
	    switch (testsrun[tn]) {
		case 1:
			//Test 1: Make a specified number of Null Calls - RMI
			//        with no arguments and returning nothing.

			if (showProg) 
				System.out.println("Timing Null Calls....");
			totalTime = loopOvrhd = 0;


			for (i = 0; i < nullCalls; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						nullCallsItns[i]);
				}

// In all cases, 2 pairs of ObjectOutputStream and ObjectOutputStream are 
// created. 
// Vella's serialization addition 1 - just create 2 pairs of Streams.

            serstartStamp = new timestamp();
	    for (m = 0; m < nullCallsItns[i]; m++) {
               baos= new ByteArrayOutputStream(512);
               oos = new ObjectOutputStream(baos);
	       oos.close();
               bais= new ByteArrayInputStream(baos.toByteArray());
               ois = new ObjectInputStream(bais);

               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
	       oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);
	    }
            serendStamp = new timestamp();
	    stotalTime = serendStamp.sub(serstartStamp).getTime();
	    
	    if (showProg) {
		System.out.println("Serialization time for Null Calls: Total"+
		" time taken" + "for " + nullCallsItns[i] +  
		" itns. : " + stotalTime/1000.0f+ " seconds.");
	    }

//

				//Meter the calls
				//Preferably a large number of iterations should
				//be specified so that the times are in the 
				//range of milliseconds measurable with accuracy
				//without the use of native methods.


                       	 	startStamp = new timestamp();
				for (j = 0; j < nullCallsItns[i]; j++) {
					serverRef.nullCall();
				} 
                        	endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();

				//loopOvrhd = getLoopOvrhd(nullCallsItns[i]);
				totalTime -= loopOvrhd;

				if (showProg) {
					System.out.println("Null Calls: Total"+
						" time taken" +
						"for " + nullCallsItns[i] +  
						" itns. : " + totalTime/1000.0f+
						" seconds. \n");
				}

				figures.addElement(new PerfFigures(
				  nullCallsItns[i],totalTime/1000.0f, 
				  stotalTime/1000.0f, loopOvrhd));
			}


			//Chop to the right size.
			figures.trimToSize();
			testName = "Null Calls ";
			
			//Create an AttribData object for this test. The 
			//last parameter indicates that this test did not
			//involve the use of vectors.

			attribData = new AttribData(testName, figures);
			data.put(testNum,attribData);

			if (showProg)
				System.out.println("Timing Null Calls done..." + "\n");

			
			break; // case 1

		case 2:


			//Test 2: Measure RMI calls with a single integer 
			//        argument 

			if (showProg) 
				System.out.println("Timing int. arg. calls...");
			totalTime = loopOvrhd = 0;
			figures = new Vector();


                        for (i = 0; i < intArg; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						intArgItns[i]);
				}

// Vella's serialization addition 2

	    bais= null;
	    bais1= null;


            serstartStamp = new timestamp();
            for (m = 0; m < intArgItns[i]; m++) {
                baos= new ByteArrayOutputStream(512);
                oos = new ObjectOutputStream(baos);
                oos.writeInt(m);
                oos.close();

	        bais= new ByteArrayInputStream(baos.toByteArray());
	    	ois = new ObjectInputStream(bais);
		int ro = ois.readInt();

               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);

            }
            serendStamp = new timestamp();
            stotalTime = serendStamp.sub(serstartStamp).getTime();
	    loopOvrhd = getLoopOvrhd(intArgItns[i]);
	    stotalTime -= loopOvrhd;

            System.out.println("Serialization of Int arg : Total"
              + " time taken" + "for " + intArgItns[i] + " itns. : "
                + stotalTime/1000.0f+ " seconds.");


//
				//Meter the calls
                                startStamp = new timestamp();
                                for (j = 0; j < intArgItns[i]; j++) {
                                        serverRef.intCall(j);
                                }
                                endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();


				loopOvrhd = getLoopOvrhd(intArgItns[i]);

				totalTime -= loopOvrhd;

				if (showProg) {
					System.out.println("Int arg.Calls: " +
						"Total time taken" +
						"for " + intArgItns[i] +  
						" itns. : " + totalTime/1000.0f+
						" seconds.\n");
				}

                                figures.addElement(new PerfFigures(
				  intArgItns[i],totalTime/1000.0f,
				  stotalTime/1000.0f,  loopOvrhd));
                        }

                        figures.trimToSize();

			testName = "Integer arg. Calls ";
			attribData = new AttribData(testName, figures);
                        data.put(testNum,attribData);
			if (showProg) 
				System.out.println(
					"Timing int. arg. calls done..." + "\n");

			break; // case 2

			
		case 3:

			//Test 3:Measure byte Array argument calls

			if (showProg) 
				System.out.println(
					"Timing calls with byteArray args....");
			totalTimeList = new float[byteArraySizes];
			// Added by Vella
			stotalTimeList = new float[byteArraySizes];
			loopOvrhd = 0;
			figures = new Vector();

                        byte[] byteArray = null;

                        for (i = 0; i < byteArrayArg; i++) {

				if (showProg) {
					System.out.println("Loop: " + (i + 1) );
					System.out.println("Iterations: " +
						byteArrayItns[i]);
					
				}

				loopOvrhd = getLoopOvrhd(byteArrayItns[i]);

				for (j = 0; j < byteArraySizes; j++) {

					if (showProg) {
				 	   System.out.println("\t" +
						"Array Size: " + 
						byteArraySizeList[j]);
					}

					byteArray =
					   new byte[byteArraySizeList[j]];

// Vella's serialization addition 3

            bais= null;
            bais1= null;


            serstartStamp = new timestamp();
            for (m = 0;m < byteArrayItns[i];m++) {
                baos= new ByteArrayOutputStream(byteArraySizeList[j]*2);
                oos = new ObjectOutputStream(baos);
                oos.writeObject(byteArray);
                oos.flush();
                oos.close();
                bais= new ByteArrayInputStream(baos.toByteArray());
                ois = new ObjectInputStream(bais);
                byte[] bao = (byte[]) ois.readObject();
 
               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);

            }
            serendStamp = new timestamp();
            stotalTimeList[j] = (float)serendStamp.sub(serstartStamp).getTime();
            stotalTimeList[j] -= (float)loopOvrhd;
	    //Convert to seconds
	    stotalTimeList[j] /= 1000.0f;

	    if (showProg) {
		System.out.println("Serialization of byteArray args"  + 
                   "Total time taken for " +
		byteArrayItns[i] + " itns., &"+ " an array size of " +
		byteArraySizeList[j] + ": " + stotalTimeList[j] + " seconds.");
	    }


//

					//Meter the calls
                                	startStamp = new timestamp();
                                	for (k = 0;k < byteArrayItns[i];k++) {
	                                   serverRef.objCall(byteArray);
        	                        }
                                	endStamp = new timestamp();
					totalTimeList[j] = 
					   (float)(
					   endStamp.sub(startStamp).getTime());

					totalTimeList[j] -= (float)loopOvrhd;

					//Convert to seconds
					totalTimeList[j] /= 1000.0f;

					if (showProg) {
						System.out.println("Total "  +
						"time taken for " +
						byteArrayItns[i] + " itns., &"+
						" an array size of " +
					 	byteArraySizeList[j] + ": " +
						totalTimeList[j] + " seconds.\n");
					}	
				}

				

                                figures.addElement(new PerfFigures(
				 byteArrayItns[i],totalTimeList, stotalTimeList, 
			         loopOvrhd)); 
				 
                        }
//

                        figures.trimToSize();
			byteArray = null;

			testName= "RMI with a byte array arg. ";
			attribData = new AttribData(testName, figures,
				     byteArraySizeList);
                        data.put(testNum,attribData);
			if (showProg)  
				System.out.println(
				   "Timing calls with byte Array args. done.." + "\n");


			break; // case 3

		case 4:
			//Test 4:Measure integer Array argument calls

			if (showProg) 
				System.out.println(
					"Timing calls with IntArray args. ...");
			totalTimeList = new float[intArraySizes];
			stotalTimeList = new float[intArraySizes];
			loopOvrhd = 0;
			figures = new Vector();

                        int[] IntArray = null;

                        for (i = 0; i < intArrayArg; i++) {

				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						intArrayItns[i]);
					
				}

				loopOvrhd = getLoopOvrhd(intArrayItns[i]);

				for (j = 0; j < intArraySizes; j++) {

					if (showProg) {
				 	   System.out.println("\t" +
						"Array Size: " + 
						intArraySizeList[j]);
					}

					IntArray =new int[intArraySizeList[j]];

// Vella's serialization addition 4
 
            bais= null;

            serstartStamp = new timestamp();

	    for (m = 0; m < intArrayItns[i]; m++) {
                baos= new ByteArrayOutputStream(intArraySizeList[j]*2);
                oos = new ObjectOutputStream(baos);
                oos.writeObject(IntArray);
                oos.flush();
                oos.close();
                bais= new ByteArrayInputStream(baos.toByteArray());
                ois = new ObjectInputStream(bais);
                int[] iao = (int[]) ois.readObject();
     
               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);

            }
            serendStamp = new timestamp();
            stotalTimeList[j] = (float)serendStamp.sub(serstartStamp).getTime();
	    stotalTimeList[j] -= (float)loopOvrhd; 
	    //Convert to seconds
	    stotalTimeList[j] /= 1000.0f;

	    if (showProg) {
		System.out.println("Serialization of IntArray args: Total "  + 
                 " time taken for " + intArrayItns[i] + " itns., &"+ 
		 " an array size of " + intArraySizeList[j] + ": " + 
		 stotalTimeList[j] + " seconds.");
	    }	

//

					//Meter the calls
                                	startStamp = new timestamp();
                                	for (k = 0; k < intArrayItns[i]; k++) {
	                                   serverRef.objCall(IntArray);
        	                        }
                                	endStamp = new timestamp();
					totalTimeList[j] = 
					   (float)(
					   endStamp.sub(startStamp).getTime());

					totalTimeList[j] -= (float)loopOvrhd; 

					//Convert to seconds
					totalTimeList[j] /= 1000.0f;

					if (showProg) {
						System.out.println("Total "  +
						"time taken for " +
						intArrayItns[i] + " itns., &"+
						" an array size of " +
					 	intArraySizeList[j] + ": " +
						totalTimeList[j] + " seconds.\n");
					}	
				}
				

                                figures.addElement(new PerfFigures(
				 intArrayItns[i],totalTimeList, stotalTimeList,
				 loopOvrhd));
				 
                        }
                        figures.trimToSize();
			IntArray = null;
			testName = "RMI with an Integer array arg. ";
			attribData = new AttribData(testName, figures, 
				     intArraySizeList);
                        data.put(testNum,attribData);
			if (showProg)  
				System.out.println(
				   "Timing calls with IntArray args. done..." + "\n");

			break; // case 4

		case 5:

			//Test 5:Measure Double Array argument calls

			if (showProg) 
				System.out.println(
				   "Timing calls with double Array args. ...");

			totalTimeList = new float[intArraySizes];
			stotalTimeList = new float[intArraySizes];
			loopOvrhd = 0;
			figures = new Vector();

                        double[] doubleArray = null;

                        for (i = 0; i < doubleArrayArg; i++) {

				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						doubleArrayItns[i]);
					
				}

				loopOvrhd = getLoopOvrhd(doubleArrayItns[i]);

				for (j = 0; j < doubleArraySizes; j++) {

					if (showProg) {
				 	   System.out.println("\t" +
						"Array Size: " + 
						doubleArraySizeList[j]);
					}

					doubleArray =
					   new double[doubleArraySizeList[j]];


// Vella's serialization addition 5
 
            bais= null;
            bais1= null;

            serstartStamp = new timestamp();

            for (m =0;m < doubleArrayItns[i];m++) {
                baos= new ByteArrayOutputStream(doubleArraySizeList[j]*2);
                oos = new ObjectOutputStream(baos);
                oos.writeObject(doubleArray);
                oos.flush();
                oos.close();
                bais= new ByteArrayInputStream(baos.toByteArray());
                ois = new ObjectInputStream(bais);
                double[] dao = (double[]) ois.readObject();
 
               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);

            }
            serendStamp = new timestamp();
            stotalTimeList[j] = (float)serendStamp.sub(serstartStamp).getTime();
	    stotalTimeList[j] -= (float)loopOvrhd;
	    //Convert to seconds
    	    stotalTimeList[j] /= 1000.0f;
	    if (showProg) {
		System.out.println("Serialization of DoubleArray Args" + 
	        "Total time taken for "  + doubleArrayItns[i]+" itns., &"+ 
		" an array size of " + doubleArraySizeList[j] + ": " + 
		stotalTimeList[j] + " seconds.");
	    }	


//


					//Meter the calls
                                	startStamp = new timestamp();
                                	for (k =0;k < doubleArrayItns[i];k++) {
	                                   serverRef.objCall(doubleArray);
        	                        }
                                	endStamp = new timestamp();

					totalTimeList[j] = 
					   (float)(
					   endStamp.sub(startStamp).getTime());

					totalTimeList[j] -= (float)loopOvrhd;

					//Convert to seconds
					totalTimeList[j] /= 1000.0f;

					if (showProg) {
						System.out.println("Total "  +
						"time taken for " +
						doubleArrayItns[i]+" itns., &"+
						" an array size of " +
					 	doubleArraySizeList[j] + ": " +
						totalTimeList[j] + " seconds.\n");
					}	
				}
				

                                figures.addElement(new PerfFigures(
				 doubleArrayItns[i],totalTimeList, stotalTimeList,
				 loopOvrhd)); 
				 
                        }
                        figures.trimToSize();
			doubleArray = null;
			testName = "RMI with a double array arg. ";
			attribData = new AttribData(testName, figures,
					doubleArraySizeList);
                        data.put(testNum, attribData);
			if (showProg)  
				System.out.println(
				   "Timing calls with double Array " +
				   "args. done..." + "\n");

			break; // case 5

		case 6:


			//Test 6: Measure RMI Calls with Object arguments 
			//The objects here are the ObjClass Objects 
		
			if (showProg) 
				System.out.println(
					"Timing calls with Obj. args. ...");
			totalTime = loopOvrhd = 0;
			figures = new Vector();
			

                        for (i = 0; i < objArg; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						objArgItns[i]);
				}

				ObjClassArray = new ObjClass[objArgItns[i]];

				for (j = 0; j < objArgItns[i]; j++) {
					ObjClassArray[j] = new ObjClass(j,j);
				}
				  
// Vella's serialization addition 6
 
            bais= null;
            bais1= null;

            serstartStamp = new timestamp();

            for (m = 0; m < objArgItns[i]; m++) {
                baos= new ByteArrayOutputStream(512);
                oos = new ObjectOutputStream(baos);
                oos.writeObject(ObjClassArray[m]);
                oos.flush();
                oos.close();
                bais= new ByteArrayInputStream(baos.toByteArray());
                ois = new ObjectInputStream(bais);
                Object ro = ois.readObject();
 
               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);

            }
            serendStamp = new timestamp();
            stotalTime = serendStamp.sub(serstartStamp).getTime();
            loopOvrhd = getLoopOvrhd(objArgItns[i]);
	    stotalTime -= loopOvrhd;
	    if (showProg) {
		System.out.println("Serialization of Object Arg : Total "  
		+ " time taken for " + objArgItns[i] + " itns. : " + 
		stotalTime/1000.0f+ " seconds.");
	    }	

//

				//Meter the calls
                                startStamp = new timestamp();
                                for (j = 0; j < objArgItns[i]; j++) {
                                        serverRef.objCall(ObjClassArray[j]);
                                }
                                endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();

                                loopOvrhd = getLoopOvrhd(objArgItns[i]);
				
				totalTime -= loopOvrhd;
				if (showProg) {
					System.out.println("ObjArgCalls:Total"+
						" time taken" +
						"for " + objArgItns[i] +  
						" itns. : " + totalTime/1000.0f+
						" seconds.\n");
				}
				
                                figures.addElement(new PerfFigures(
				  objArgItns[i],totalTime/1000.0f, 
				  stotalTime/1000.0f, loopOvrhd));
                        }

                        figures.trimToSize();
			ObjClassArray = null;
			testName = "RMI with an Object arg. ";
			attribData = new AttribData(testName, figures);
                        data.put(testNum,attribData);
			
			if (showProg) 
				System.out.println(
				    "Timing calls with Obj. arg. done..." + "\n");

			break; // case 6

		case 7:

			//Test 7: Measures RMI calls with an integer and an 
			//        object argument.

			if (showProg)
				System.out.println("Timing two arg. calls...");
			totalTime = loopOvrhd = 0;
			figures = new Vector();

			 
                        for (i = 0; i < twoArgCalls; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						twoArgItns[i]);
				}

				ObjClassArray = new ObjClass[twoArgItns[i]];

				for (j = 0; j < twoArgItns[i]; j++) {
					ObjClassArray[j] = new ObjClass(j,j);
				}
			
// Vella's serialization addition 7
 
            bais= null;
            bais1= null;

            serstartStamp = new timestamp();

            for (m = 0; m < twoArgItns[i]; m++) {
                baos= new ByteArrayOutputStream(2048);
                oos = new ObjectOutputStream(baos);
                oos.writeObject(ObjClassArray[m]);
                oos.writeInt(m);
                oos.flush();
                oos.close();
                bais= new ByteArrayInputStream(baos.toByteArray());
                ois = new ObjectInputStream(bais);
                Object ro = ois.readObject();
	        int prim_i = ois.readInt();
 
               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);

            }
            serendStamp = new timestamp();
            stotalTime = serendStamp.sub(serstartStamp).getTime();
            loopOvrhd = getLoopOvrhd(twoArgItns[i]);
	    stotalTime -= loopOvrhd;
	    if (showProg) {
		System.out.println("Serialization of object and int args: Total"
	        + " time taken for  " + twoArgItns[i] + 
	        " itns. : " + stotalTime/1000.0f+ " seconds.");
	    }	


//

				//Meter the calls
                                startStamp = new timestamp();
                                for (j = 0; j < twoArgItns[i]; j++) {
                                       serverRef.twoArgCall(ObjClassArray[j],j);
                                }
                                endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();

                                loopOvrhd = getLoopOvrhd(twoArgItns[i]);
				totalTime -= loopOvrhd;
				
				if (showProg) {
					System.out.println("2ArgCalls:Total"+
						" time taken" +
						"for " + twoArgItns[i] +  
						" itns. : " + totalTime/1000.0f+
						" seconds.\n");
				}
                                figures.addElement(new PerfFigures(
				  twoArgItns[i],totalTime/1000.0f, 
				  stotalTime/1000.0f, loopOvrhd));
                        }
                        figures.trimToSize();
			ObjClassArray = null;
			testName = "RMI with two args. ";
			attribData = new AttribData(testName, figures);
                        data.put(testNum,attribData);
			if (showProg)
				System.out.println(
					"Timing two arg. calls done..." + "\n");
			break; // case 7

			
		case 8:
			//Test 8: Measure RMI calls passing Object Arrays
			
                        if (showProg)
				System.out.println(
				   "Timing calls with Obj.Array arg. ...");
			totalTimeList = new float[objArraySizes];
			stotalTimeList = new float[objArraySizes];
			loopOvrhd = 0;
			figures = new Vector();


			ObjClass[] ObjArray = null;


                        for (i = 0; i < objArrayArg; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						objArrayItns[i]);
				}

				loopOvrhd = getLoopOvrhd(objArrayItns[i]);

				for (j = 0; j <objArraySizes; j++) {
					if (showProg) {
						System.out.println("\t" + 
						"Array Size: " +
						objArraySizeList[j]);
					}

					ObjArray=
					  new ObjClass[objArraySizeList[j]];
					for (k =0;k < objArraySizeList[j];k++){
					  ObjArray[k] = new ObjClass(k,k);
					}

					
// Vella's serialization addition 8
 
            bais= null;
            bais1= null;

            serstartStamp = new timestamp();

            for (m = 0; m < objArrayItns[i]; m++) {
                baos= new ByteArrayOutputStream(objArraySizeList[j]*2);
                oos = new ObjectOutputStream(baos);
                oos.writeObject(ObjArray);
                oos.flush();
                oos.close();
                bais= new ByteArrayInputStream(baos.toByteArray());
                ois = new ObjectInputStream(bais);
                Object[] robj = (Object[])ois.readObject();
 
               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);

            }
            serendStamp = new timestamp();
            stotalTimeList[j] = (float)serendStamp.sub(serstartStamp).getTime();
	    stotalTimeList[j] -=(float)loopOvrhd;
	    stotalTimeList[j] /= 1000.0f;
	
	    if (showProg) {
		System.out.println("Object Arrays Serialization: Total"
	        + " time taken for " + objArrayItns[i] + 
	        " itns. : " + stotalTimeList[j] + " seconds.");
	    }	


//

					//Meter the calls
                                	startStamp = new timestamp();
                                	for (k = 0; k < objArrayItns[i]; k++) {
                                          serverRef.objCall(ObjArray);
                                	}
                                	endStamp = new timestamp();
					totalTimeList[j] = 
					  (float)(
					   endStamp.sub(startStamp).getTime());

					totalTimeList[j] -=(float)loopOvrhd;

					totalTimeList[j] /= 1000.0f;

					if (showProg) {
						System.out.println("Total "  +
						"time taken for " +
						objArrayItns[i]+" itns., &"+
						" an array size of " +
					 	objArraySizeList[j] + ": " +
						totalTimeList[j] + " seconds.\n");
					}	
				}
				

                                figures.addElement(new PerfFigures(
				 objArrayItns[i], totalTimeList, 
		                 stotalTimeList, loopOvrhd));
				
                        }
                        figures.trimToSize();
			ObjArray = null;
			testName = "RMI with an array of obj. ";
			attribData = new AttribData(testName, figures,
					objArraySizeList);
                        data.put(testNum,attribData);
			if (showProg)
				System.out.println( 
				  "Timing Obj. Array calls done..." + "\n");

			break; // case 8

		case 9:
			//Test 9: Measure RMI calls with a remote argument

			if (showProg)
				System.out.println(
				  "Timing calls with remote arg. ...");
			totalTime = loopOvrhd = 0;
			figures = new Vector();



                        for (i = 0; i < remoteArg; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						remoteArgItns[i]);
				}


// Vella's serialization addition 9
 
            bais= null;
            bais1= null;

            serstartStamp = new timestamp();

            for (m = 0; m < remoteArgItns[i]; m++) {
		RemoteObj remRef = new RemoteObj();
                baos= new ByteArrayOutputStream(512);
                oos = new ObjectOutputStream(baos);
                oos.writeObject(remRef);
                oos.flush();
                oos.close();
                bais= new ByteArrayInputStream(baos.toByteArray());
                ois = new ObjectInputStream(bais);
                RemoteObj roj = (RemoteObj)ois.readObject();
 
               baos1= new ByteArrayOutputStream(512);
               oos1 = new ObjectOutputStream(baos1);
               oos1.close();
               bais1= new ByteArrayInputStream(baos1.toByteArray());
               ois1 = new ObjectInputStream(bais1);
            }
            serendStamp = new timestamp();
            stotalTime = serendStamp.sub(serstartStamp).getTime();
            loopOvrhd = getLoopOvrhd(remoteArgItns[i]);
	    stotalTime -= loopOvrhd; 
	    if (showProg) {
		System.out.println("Remote Object Arg Serialization: Total"
	        + " time taken for " + remoteArgItns[i] + 
	        " itns. : " + stotalTime/1000.0f+ " seconds.");
	    }	

//


				//Meter the calls
				//Note that this time also includes the time
				//in creating the remote objects at the client
				//end 

                                startStamp = new timestamp();
                                for (j = 0; j < remoteArgItns[i]; j++) {
					RemoteObj remRef = new RemoteObj();
					serverRef.objCall(remRef);
                                }
                                endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();

                                loopOvrhd = getLoopOvrhd(remoteArgItns[i]);

				totalTime -= loopOvrhd; 
				
				if (showProg) {
					System.out.println("Total"+
						" time taken" +
						"for " + remoteArgItns[i] +  
						" itns. : " + totalTime/1000.0f+
						" seconds.\n");
				}
                                figures.addElement(new PerfFigures(
				 remoteArgItns[i],totalTime/1000.0f,
				 stotalTime/1000.0f, loopOvrhd));
                        }

                        figures.trimToSize();
			testName = "RMI with a remote Object arg. ";
			attribData = new AttribData(testName, figures);
                        data.put(testNum,attribData);
			if (showProg)
				System.out.println(
				  "Timing calls with remote arg. done..." + "\n");
			break;
		
		case 10:
			//Test 10: Measure RMI calls returning  a remote object
			//	  from the server.

			if (showProg)
				System.out.println(
				  "Timing RMI calls returning remote obj. ...");
			totalTime = loopOvrhd = 0;
			figures = new Vector();



                        for (i = 0; i < returnRemote; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						returnRemoteItns[i]);
				}




// Vella's serialization addition 10
 
            bais= null;
            bais1= null;

            Remote obj = (Remote)serverRef.getRemObj();
            serstartStamp = new timestamp();

            for (m = 0; m < returnRemoteItns[i]; m++) {
               baos= new ByteArrayOutputStream(512);
               oos = new ObjectOutputStream(baos);
               oos.close();
               bais= new ByteArrayInputStream(baos.toByteArray());
               ois = new ObjectInputStream(bais);

                baos1= new ByteArrayOutputStream(512);
                oos1= new ObjectOutputStream(baos1);
                oos1.writeObject(obj);
                oos1.flush();
                oos1.close();
                bais1= new ByteArrayInputStream(baos1.toByteArray());
                ois1 = new ObjectInputStream(bais1);
                Remote roj = (Remote)ois1.readObject();
 
            }
            serendStamp = new timestamp();
            stotalTime = serendStamp.sub(serstartStamp).getTime();
            loopOvrhd = getLoopOvrhd(returnRemoteItns[i]);
	    stotalTime -= loopOvrhd;
	    if (showProg) {
		System.out.println("Serialization of return Remote Object: Total"
	        + " time taken for " + returnRemoteItns[i] + 
	        " itns. : " + stotalTime/1000.0f+ " seconds.");
	    }	

//
				
				//Meter the calls
				//Note that this includes the time in creating
				//the remote objects at the server end.

                                startStamp = new timestamp();
                                for (j = 0; j < returnRemoteItns[i]; j++) {
					Remote remobj =
				        (Remote)serverRef.getRemObj();
                                }
                                endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();

                                loopOvrhd = getLoopOvrhd(returnRemoteItns[i]);
				totalTime -= loopOvrhd;
				
				if (showProg) {
					System.out.println("Total"+
						" time taken" +
						"for " + returnRemoteItns[i] +  
						" itns. : " + totalTime/1000.0f+
						" seconds.\n");
				}
                                figures.addElement(new PerfFigures(
				 	returnRemoteItns[i],totalTime/1000.0f,
				  	stotalTime/1000.0f, loopOvrhd));
			}
                        figures.trimToSize();
			testName = "RMI returning a remote obj. ";
			attribData = new AttribData(testName, figures);
                        data.put(testNum, attribData);
			if (showProg)
			     System.out.println( 
			      "Timing RMI calls returning remote obj. done..." + "\n");

			break; // case 10

		case 11:
             
			//Test 11: Measuring RMI calls returning a vector of 
			//	  remote objects. 

			if (showProg)
				System.out.println("Timing RMI calls " + 
				  "returning a vector of NEW remote objs. ...");
			totalTimeList = new float[vectorSizes];
			stotalTimeList = new float[vectorSizes];
			loopOvrhd = 0;
			figures = new Vector();


                        for (i = 0; i < returnRemoteVector; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1) );
					System.out.println("Iterations: " +
						remoteVectorItns[i]);
				}


				float dummy;

				loopStart = new timestamp();
				for (k=0; k < remoteVectorItns[i]; k++){
					startStamp = new timestamp();
					endStamp = new timestamp();
				 	dummy = (float)(
					  endStamp.sub(startStamp).getTime());
				}
				loopEnd = new timestamp();
				loopOvrhd = loopEnd.sub(loopStart).getTime();

				for (j =0; j < vectorSizes; j++) {

					if (showProg) {
						System.out.println("\t" +
						"Vector Size: " +
						vectorSizeList[j]);
					}

// Vella's serialization addition 11
 
            bais= null;
            bais1= null;

	    serverRef.createVector( vectorSizeList[j]);
            Vector vobj =  serverRef.getVector();
            serstartStamp = new timestamp();

            for (m = 0; m < remoteVectorItns[i]; m++) {
               baos= new ByteArrayOutputStream(512);
               oos = new ObjectOutputStream(baos);
               oos.close();
               bais= new ByteArrayInputStream(baos.toByteArray());
               ois = new ObjectInputStream(bais);

                baos1= new ByteArrayOutputStream(512);
                oos1= new ObjectOutputStream(baos1);
                oos1.writeObject(vobj);
                oos1.flush();
                oos1.close();
                bais1= new ByteArrayInputStream(baos1.toByteArray());
                ois1 = new ObjectInputStream(bais1);
                Vector roj = (Vector)ois1.readObject();

            }
            serendStamp = new timestamp();
            stotalTimeList[j] = serendStamp.sub(serstartStamp).getTime();
            loopOvrhd = getLoopOvrhd(returnRemoteItns[i]);
            stotalTimeList[j] -= (float)loopOvrhd;
            stotalTimeList[j] /= 1000.0f;

            if (showProg) {
                System.out.println("Serialization of vector of Remote Objects: Total"
                + " time taken for " + remoteVectorItns[i] +
                " itns. : " + stotalTimeList[j]+ " seconds.");
            }



//


					//Meter the calls
					//Note :These are new objects 
					//that the server 
					//first creates. The measured time 
					//*DOES NOT* include the time taken to
					// create these objects at the server.

		
                	                for (k =0;k < remoteVectorItns[i];k++){

					  //Get the next slice of size 
					  //"vectorSize" starting from 
					  //offset "k*vectorSize"

					  serverRef.createVector(
						vectorSizeList[j]);
	                                  startStamp = new timestamp();
				          v =  serverRef.getVector();
                               		  endStamp = new timestamp();
                                	  totalTimeList[j] +=
                                     	  (float)(
					  endStamp.sub(startStamp).getTime());
                                	}
					
					totalTimeList[j] -= (float)loopOvrhd;
					totalTimeList[j] /= 1000.0f;

					if (showProg) {
						System.out.println("Total "  +
						"time taken for " +
						remoteVectorItns[i]+" itns., &"+
						" an array size of " +
					 	vectorSizeList[j] + ": " +
						totalTimeList[j] + " seconds.\n");
					}	
				}
				

                                figures.addElement(new PerfFigures(
				 remoteVectorItns[i],totalTimeList,
			         stotalTimeList, loopOvrhd));
				
                        }
//
			v=null;
                        figures.trimToSize();
			testName = "RMI returning a vector of NEW remote " + 
					"objects ";
			attribData = new AttribData(testName, figures,
					vectorSizeList);
                        data.put(testNum,attribData);

			if (showProg) {
			     	System.out.println("Timing RMI calls " +  
			       		"returning a vector of NEW " +
					"remote objs. done..." + "\n");
			}


			//Test 12:  Timing RMI calls returning a vector of 
			//	    *old* remote objects. The server *DOES NOT* 
			//	    create a new vector of remote objects on 
			//          this call, but simply returns the existing
			//	    vector. 

			if (showProg){
				System.out.println("Timing RMI calls " + 
				  "returning a vector of OLD remote objs. ...");
			}
			totalTime = loopOvrhd = 0;
			figures = new Vector();

 

     for (i = 0; i < returnRemoteVector; i++) {
	if (showProg) {
		System.out.println("Loop: " + (i+1));
			System.out.println("Iterations: " +
				remoteVectorItns[i]);
	}

// Vella's serialization addition 12
 
            bais= null;
            bais1= null;


	    Vector vob=serverRef.getVector();
            serstartStamp = new timestamp();

            for (m = 0; m < remoteVectorItns[i]; m++) {
               baos= new ByteArrayOutputStream(512);
               oos = new ObjectOutputStream(baos);
               oos.close();
               bais= new ByteArrayInputStream(baos.toByteArray());
               ois = new ObjectInputStream(bais);

                baos1= new ByteArrayOutputStream(512);
                oos1= new ObjectOutputStream(baos1);
                oos1.writeObject(vob);
                oos1.flush();
                oos1.close();
                bais1= new ByteArrayInputStream(baos1.toByteArray());
                ois1 = new ObjectInputStream(bais1);
                Vector roj = (Vector)ois1.readObject();

            }
                serendStamp = new timestamp();
                stotalTime = serendStamp.sub(serstartStamp).getTime();
                loopOvrhd = getLoopOvrhd(remoteVectorItns[i]);
                stotalTime -= (float)loopOvrhd;

            if (showProg) {
                System.out.println("Serialization of vector of remote Objects: Total"
                + " time taken for " + remoteVectorItns[i] +
                " itns. : " + stotalTime/1000.0f + " seconds.");
            }


//


			//Meter the calls. The number of loops is the same
			//as before.

                                startStamp = new timestamp();
                                for (j = 0; j < remoteVectorItns[i]; j++) {
				        v = serverRef.getVector();
                                }
                                endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();

                                loopOvrhd = getLoopOvrhd(remoteVectorItns[i]);
				totalTime -= loopOvrhd;
				
				if (showProg) {
					System.out.println("Total"+
						" time taken" +
						"for " + remoteVectorItns[i] +  
						" itns. : " + totalTime/1000.0f+
						" seconds.\n");
				}
                                figures.addElement(new PerfFigures(
				 	remoteVectorItns[i],totalTime/1000.0f,
                                 	stotalTime/1000.0f, loopOvrhd));
		}
                        figures.trimToSize();
			v = null;
			testName = "RMI returning a vector of OLD remote " +
					"objects ";
			attribData = new AttribData(testName, figures);
                        data.put(testNum,attribData);

			if (showProg) {
			   System.out.println("Timing RMI calls " +  
			      "returning a vector of OLD remote objs. done" + "\n");
			}
			break; // case 11

		case 12:
			//Test 13:  Timing RMI calls throwing exceptions

			if (showProg){
				System.out.println("Timing RMI calls " + 
				  "throwing exceptions ...");
			}
			totalTime = loopOvrhd = 0;
			figures = new Vector();


                       for (i = 0; i < exceptionCalls; i++) {
                                if (showProg) {
                                        System.out.println("Loop: " + (i+1));
                                        System.out.println("Iterations: " +
                                                exceptionCallsItns[i]);
                                }

// Vella's serialization addition 13
 
            bais= null;
            bais1= null;

            serstartStamp = new timestamp();

            for (m = 0; m < exceptionCallsItns[i]; m++) {
               baos= new ByteArrayOutputStream(512);
               oos = new ObjectOutputStream(baos);
               oos.close();
               bais= new ByteArrayInputStream(baos.toByteArray());
               ois = new ObjectInputStream(bais);

                baos1= new ByteArrayOutputStream(512);
                oos1= new ObjectOutputStream(baos1);
                oos1.close();
                bais1= new ByteArrayInputStream(baos1.toByteArray());
                ois1 = new ObjectInputStream(bais1);

            }
                serendStamp = new timestamp();
                stotalTime = serendStamp.sub(serstartStamp).getTime();
                loopOvrhd = getLoopOvrhd(exceptionCallsItns[i]);
                stotalTime -= (float)loopOvrhd;

            if (showProg) {
                System.out.println("Serialization of no args: "
                + " time taken for " + exceptionCallsItns[i] +
                " itns. : " + stotalTime/1000.0f + " seconds.");
            }

//


			//Meter the calls. 

                                startStamp = new timestamp();
                                for (j = 0; j < exceptionCallsItns[i]; j++) {
				  try {
				        serverRef.exceptionCall();
				  }catch (CollectorException e){}
                                }
                                endStamp = new timestamp();
				totalTime = endStamp.sub(startStamp).getTime();

                                loopOvrhd =getLoopOvrhd(exceptionCallsItns[i]);
				totalTime -= loopOvrhd;
				
				if (showProg) {
					System.out.println("Total"+
						" time taken" +
						"for " +exceptionCallsItns[i]+  
						" itns. : " + totalTime/1000.0f+
						" seconds.\n");
				}

                                figures.addElement(new PerfFigures(
				 	exceptionCallsItns[i],
					totalTime/1000.0f,stotalTime/1000.0f,
					loopOvrhd));
                                 	
			}

                        figures.trimToSize();
			testName = "RMI calls throwing exceptions";
			attribData = new AttribData(testName, figures);
                        data.put(testNum, attribData); 

			if (showProg) {
			   System.out.println("Timing RMI calls " +  
			      "throwing exceptions done" + "\n");
			}
			break; // case 13
		
		case 13:
			//Test 14: Timing RMI calls sending objects 
			//   	   referencing other objects. In this case,
			//         these objects are B-Trees, and the number
			//	   of references is controlled by the depth
			//	   of the bTree. In each case we pass the 
			//	   the root object. NOTE: This tests the time
			//	   taken in serializing the entire graph of 
			//	   objects.

                        if (showProg)
				System.out.println(
				   "Timing calls with Obj. graphs ...");
			totalTimeList = new float[bTreeDepths];
			stotalTimeList = new float[bTreeDepths];
			loopOvrhd = 0;
			figures = new Vector();


			bTree BTree = null;
			node rootNode = null;

                        for (i = 0; i < bTreeArg; i++) {
				if (showProg) {
					System.out.println("Loop: " + (i+1));
					System.out.println("Iterations: " +
						bTreeArgItns[i]);
				}

				loopOvrhd = getLoopOvrhd(bTreeArgItns[i]);

				for (j = 0; j <bTreeDepths; j++) {
					if (showProg) {
						System.out.println("\t" + 
						"Tree Depth: " +
						bTreeDepthList[j]);
					}

					BTree = new bTree(bTreeDepthList[j]);
					rootNode = BTree.getRoot();

// Vella's serialization addition 14
 
            bais= null;
            bais1= null;

            serstartStamp = new timestamp();

            for (m = 0; m < bTreeArgItns[i]; m++) {
                baos1= new ByteArrayOutputStream(512);
                oos1= new ObjectOutputStream(baos1);
                oos1.writeObject(rootNode);
                oos1.flush();
                oos1.close();
                bais1= new ByteArrayInputStream(baos1.toByteArray());
                ois1 = new ObjectInputStream(bais1);
                node roj = (node)ois1.readObject();

               baos= new ByteArrayOutputStream(512);
               oos = new ObjectOutputStream(baos);
               oos.close();
               bais= new ByteArrayInputStream(baos.toByteArray());
               ois = new ObjectInputStream(bais);

            }
            serendStamp = new timestamp();
            stotalTimeList[j] = serendStamp.sub(serstartStamp).getTime();
            stotalTimeList[j] -= (float)loopOvrhd;
            stotalTimeList[j] /= 1000.0f;

            if (showProg) {
                System.out.println("Serialization of a root node of Btree: "
                + " time taken for " + bTreeArgItns[i] + " itns., &" + 
                " a tree depth of " + bTreeDepthList[j] + ": " +
                + stotalTimeList[j]+ " seconds.");
            }



//


					//Meter the calls
                                	startStamp = new timestamp();
                                	for (k = 0; k < bTreeArgItns[i]; k++) {
                                          serverRef.objCall(rootNode);
                                	}
                                	endStamp = new timestamp();
					totalTimeList[j] = 
					  (float)(
					   endStamp.sub(startStamp).getTime());

					totalTimeList[j] -=(float)loopOvrhd;

					totalTimeList[j] /= 1000.0f;

					if (showProg) {
						System.out.println("Total "  +
						"time taken for " +
						bTreeArgItns[i]+" itns., &"+
						" a tree depth of " +
					 	bTreeDepthList[j] + ": " +
						totalTimeList[j] + " seconds.\n");
					}	
				}
				

                                figures.addElement(new PerfFigures(
				 bTreeArgItns[i], totalTimeList, 
				 stotalTimeList, loopOvrhd));
				
                        }

                        figures.trimToSize();
			testName = "RMI with an obj. graph ";
			attribData = new AttribData(testName, figures,
					bTreeDepthList);
                        data.put(testNum,attribData);
			if (showProg)
				System.out.println( 
				  "Timing Calls with obj. graphs done..." + "\n");
			break; // case 14

	   default:
			System.out.println( " invalid test number\n");

	  } // end switch
        } // end for loop

			
			//Log the performance figures
		 	if (showProg) {	
				System.out.println("Logging Perf. figs. # " +
					 clientID);
			}
			collRef.logPerfData(data, clientID);

			//Whew!! Thought this would never end!!

		}catch (Exception e) {

			//HELLLLP!!!

			System.out.println("TestClient exception: " +
				e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		
	}

	/**
	 * Print the usage for this program 
	 */

	public static void usage() {
		
		System.out.println("usage: java benchmarks.TestClient " +
				"[ -help] "+
				"[-numClients=n][-showProg] "+
				"[-vers=\"java_version\"]" +
				"[-tlist=\"1,2,...13\"]");
		System.out.println("  where :----");
		System.out.println("       -numClients=n          :Number of " +
				"clients to be spawned.(default 1)");
		System.out.println("       -showProg              :verbose " +
				"execution.(default false)" );
		System.out.println("       -vers=\"java_version\" :the java " +
				"version.");
		System.out.println("       -tlist=\"1,2,...13\" :the list of" + 
				" tests to be run which are as follows\n ( the default is all 13 tests ):-\n"  ); 
		System.out.println("  TestNum  " + "  TestName");
		System.out.println("-----------" + "  --------");
		for ( int x=0; x<tlist.length; x++) {
			System.out.println("    " + tlist[x][0] + "	     " + tlist[x][1]);
		}
		System.out.println("\nNOTE: Specify the version as a command " +
			"line arg. only if you wish to override the version "+
			"string generated by System.getProperty(). E.g. for "+
			"internal builds, where the version string is not " +
			"detailed enough.");

		return;
	}

	public static void main(String args[]) {

		String str, subStr;
		int numClients = 1, i, start, end;

		TestClient[] client ;
		Thread[] clientThread;
	

		int p=0, q=0;
		seltests=new int[tlist.length];
		deftests=new int[tlist.length];

		//Check usage
		if (args.length > 5) {
			usage();
			System.exit(1); //error
		}
		
		// All tests are run by default;
		for ( int r=0;  r< tlist.length ; r++)  {
		    deftests[r]=Integer.parseInt(tlist[r][0]);
		}
		for (i = 0; i < args.length; i++) {
			if (args[i].startsWith("-numClients")) {
				start=args[i].indexOf('=');
				subStr = args[i].substring(start+1);
				try {
					numClients = Integer.parseInt(subStr);
				}catch (NumberFormatException e) {
					//Not an integer or some such problem
					usage(); 
					System.exit(1);
				}
			}else if (args[i].equals("-showProg")){
				showProg = true;
			}else if (args[i].startsWith("-vers")){

				//Extract the version string enclosed in
				//quotes

				start=args[i].indexOf('\"');
				end = args[i].indexOf('\"',start+1);
				version = args[i].substring(start+1,end);
			}else if (args[i].startsWith("-tlist")){
				deft=false;
				start=args[i].indexOf('\"');
				end = args[i].indexOf('\"',start+1);
				String tstr=args[i].substring(start+1,end);
			        while (p<tstr.length()) {
				     int cpos=tstr.indexOf(',',p); 
				     if ( cpos < 0 ) {
					seltests[q]=Integer.parseInt(tstr.substring(p, tstr.length()));
					break;
				     } else {
				       try {
				     	  seltests[q]=Integer.parseInt(tstr.substring(p,cpos));
					} catch (NumberFormatException e) {
                                        //Not an integer or some such problem
                                        usage();
                                        System.exit(1);
					}
			            }
				    p=cpos+1; 
				    q++;
				}
			}else if (args[i].equals("-help")){
				usage();
				System.exit(0); 
			}else {
				usage();
				System.exit(1); //error
			}
		}

		if ( deft ) {
		  testsrun=new int[deftests.length];	
		  System.arraycopy(deftests,0,testsrun,0,deftests.length);
		} else {
		  testsrun=new int[q+1];	
		  System.arraycopy(seltests,0,testsrun,0,q+1);
		}
		  
		nullCallsItns = new int[MAX_LOOPS];

		intArgItns = new int[MAX_LOOPS];

		intArrayItns = new int[MAX_LOOPS];
		intArraySizeList = new int[MAX_LOOPS];

		byteArrayItns = new int[MAX_LOOPS];
		byteArraySizeList = new int[MAX_LOOPS];

		doubleArrayItns = new int[MAX_LOOPS];
		doubleArraySizeList = new int[MAX_LOOPS];

		objArgItns = new int[MAX_LOOPS];

		twoArgItns = new int[MAX_LOOPS];

		objArrayItns = new int[MAX_LOOPS];
		objArraySizeList = new int[MAX_LOOPS];

		remoteArgItns = new int[MAX_LOOPS];

		returnRemoteItns = new int[MAX_LOOPS];

		remoteVectorItns = new int[MAX_LOOPS];

		vectorSizeList = new int[MAX_LOOPS];

		exceptionCallsItns = new int[MAX_LOOPS];

		bTreeArgItns = new int[MAX_LOOPS];
		bTreeDepthList = new int[MAX_LOOPS];
		
		//Create properties list for the client with the 
		//system properties as the default property list.
		//Load the main properties list from the properties
		//file

		try {
	 		clientProps = new Properties(System.getProperties());
			if (version != null) {
				clientProps.put("java.version",version);
			}
			clientProps.load (new FileInputStream(
			"benchmarks.client.properties"));
			
			//Get the server hostname (defaults to current host)
			host = clientProps.getProperty("Server-host","");

		} catch (Exception e) {
			System.out.println("Couldn't retreive client " +
				"properties. Error: " +e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}


		
		//Parse the string and get the number of loops (count)
		//as well as the number of iterations (Itns[])
		//within each loop

		str = clientProps.getProperty("NullCalls",
						"1000,5000,25000");
	   	nullCalls = fillArray(str,',',nullCallsItns);
		nullCallsItns = trimToSize(nullCalls,nullCallsItns);


		//Parse the property string.
		str =clientProps.getProperty("IntArg",
				"1000,5000,10000,25000");
                intArg = fillArray(str,',',intArgItns);
		intArgItns = trimToSize(intArg, intArgItns);

		//Parse the property string
                str =clientProps.getProperty("IntArrayArg",
                                "1000,5000,10000,25000");
                intArrayArg = fillArray(str,',',intArrayItns);
		intArrayItns = trimToSize(intArrayArg, intArrayItns);
                str =clientProps.getProperty("IntArraySizeList",
                                "100,1000,10000");
		intArraySizes = fillArray(str,',',intArraySizeList); 
		intArraySizeList = trimToSize(intArraySizes,intArraySizeList);

		//Parse the property string
                str =clientProps.getProperty("ByteArrayArg",
                                "1000,5000,10000,25000");
                byteArrayArg = fillArray(str,',',byteArrayItns);
		byteArrayItns = trimToSize(byteArrayArg, byteArrayItns);
                str =clientProps.getProperty("ByteArraySizeList",
                                "100,1000,10000");
		byteArraySizes = fillArray(str,',',byteArraySizeList); 
		byteArraySizeList =trimToSize(byteArraySizes,byteArraySizeList);


		//Parse the property string
                str =clientProps.getProperty("DoubleArrayArg",
                                "1000,5000,10000,25000");
                doubleArrayArg = fillArray(str,',',doubleArrayItns);
		doubleArrayItns = trimToSize(doubleArrayArg, doubleArrayItns);
                str =clientProps.getProperty("DoubleArraySizeList",
                                "100,1000,10000");
		doubleArraySizes = fillArray(str,',',doubleArraySizeList); 
		doubleArraySizeList = trimToSize(doubleArraySizes,
				doubleArraySizeList);

		//Parse the property string

                str =clientProps.getProperty("ObjArg",
                                "100,1000,5000,10000,25000");
                objArg = fillArray(str,',',objArgItns);
		objArgItns = trimToSize(objArg, objArgItns);



		//Parse the property string 
                str =clientProps.getProperty("TwoArgCalls",
                                "1000,5000,10000,25000");
               	twoArgCalls = fillArray(str,',',twoArgItns);
		twoArgItns = trimToSize(twoArgCalls, twoArgItns);



		//Parse the property string 

                str =clientProps.getProperty("ObjArrayArg",
                                "100,1000,5000,10000,25000");
                objArrayArg = fillArray(str,',',objArrayItns);
		objArrayItns = trimToSize(objArrayArg, objArrayItns);
		str =clientProps.getProperty("ObjArraySizeList",
                                "100,1000,5000,10000,25000");
		objArraySizes = fillArray(str,',',objArraySizeList);
		objArraySizeList = trimToSize(objArraySizes,objArraySizeList);


		//Parse the property string 
                str =clientProps.getProperty("RemoteArg",
                                "1000,5000,10000,25000");
                remoteArg = fillArray(str,',',remoteArgItns);
		remoteArgItns = trimToSize(remoteArg, remoteArgItns);


		//Parse the property string

                str =clientProps.getProperty("ReturnRemote",
                                "1000,5000,10000,25000");
                returnRemote = fillArray(str,',',returnRemoteItns);
		returnRemoteItns = trimToSize(returnRemote, returnRemoteItns);


		//Parse the property string
                str =clientProps.getProperty("ReturnRemoteVector",
                                "1000,5000,10000,25000");
                returnRemoteVector = fillArray(str,',',remoteVectorItns);
		remoteVectorItns =trimToSize(returnRemoteVector,
				remoteVectorItns);
		str = clientProps.getProperty("VectorSizeList","30");
		vectorSizes =  fillArray(str,',',vectorSizeList);
		vectorSizeList = trimToSize(vectorSizes, vectorSizeList);

		//Parse the property string
		str = clientProps.getProperty("ExceptionCalls",
	  			"1000,5000,10000,25000");
		exceptionCalls = fillArray(str,',',exceptionCallsItns);
		exceptionCallsItns = trimToSize(exceptionCalls,
				exceptionCallsItns);
 
		str = clientProps.getProperty("B-TreeArg",
				"1000,5000,10000,25000");
		bTreeArg = fillArray(str,',',bTreeArgItns);
		bTreeArgItns = trimToSize(bTreeArg, bTreeArgItns);
		str = clientProps.getProperty("B-TreeDepthList","5,10,15");
		bTreeDepths = fillArray(str,',',bTreeDepthList);
		bTreeDepthList = trimToSize(bTreeDepths,bTreeDepthList);

		client = new TestClient[numClients];
		clientThread = new Thread[numClients];


		for (i = 0; i < numClients; i++){

			//Spawn new Clients 
			try {
				client [i] = new TestClient();
			}catch (RemoteException e) {};
			clientThread[i] = new Thread(client[i]);

			clientThread[i].start();
		} 				

		for (i = 0; i < numClients; i++) {
			try {	
				clientThread[i].join();
			}catch (InterruptedException e) {}
		}

		System.exit(0);
	}
		
}
