/*
 * @(#)timestamp.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package benchmarks;

import java.util.Date;
 
/**
 * This class currently gets timestamps using the System.currentTimeMillis()
 * call.
 */

//The idea behind implementing a separate class for timestamps, is that it 
//can later be easily replaced with a method giving a better level of 
//granularity (possibly in nanoseconds) without having to wade throught tons
//of code

public class timestamp implements java.io.Serializable{

	/**
         * The current time to which all later timestamps are normalised to
	 * to make measurements more tractable. Set once when the first  
	 * timestamp is  created. 
	 */ 
	static long normalizer = -1;
	
	/**
	 *The current time normalized w.r.t. the normalizer
	 */
	private long clockVal;

	static {
		if (normalizer == -1){	
			normalizer = System.currentTimeMillis();
		}
	}
	
	/**
	 * Gets the current time and normalizes it w.r.t the normalizer
	 */
	public timestamp() {

		clockVal = System.currentTimeMillis();
		clockVal -= normalizer;
	}

	/**
	 * Constructs a timestamp with the normalized time
	 * 
	 * @param normalizedTime  the time already normalized w.r.t. the 
	 *                        normalizer. This is usually the result
	 *			  of an addition or subtraction on two 
	 *			  (already) normalized times.
	 */ 
	public timestamp(long normalizedTime) {
		clockVal = normalizedTime;
	}


	/**
	 * Measures the difference between the specified timestamp and itself
	 * Assumes both timestamps to be <i>normalized</i>
	 * 
	 * @param t timestamp from which the diff. is to be measured
	 * @return  a timestamp indicating the time difference.
	 */
	public timestamp sub(timestamp t) {
		return new timestamp(clockVal - t.clockVal);
	} 	 		

	public long getTime() {
		return clockVal;
		
	}

	public String toString () {
		return (clockVal + " milliseconds ");
	}
}
