/*
 * @(#)SignedObjectTest.java	1.9 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package crypto;

import java.security.*;
import java.io.*;
import java.util.*;

import test.*;

/**
 * <p>
 *
 * @version 	1.9, 09/21/98
 * @author Li Gong */

public class SignedObjectTest extends Test {
	
	public int keySize = 512;
	public int dataSize = 1000;

	boolean verify = false;
	boolean verify1 = false;

	public void run() throws TestException {
		try {
			// generate a key
			SecureRandom seed = new SecureRandom();
			KeyPairGenerator y = KeyPairGenerator.getInstance("DSA", "SUN");
			y.initialize(keySize, seed);
			KeyPair keypair = y.generateKeyPair();
			PublicKey pubkey = keypair.getPublic();
			PrivateKey prikey = keypair.getPrivate();
			Signature engine = Signature.getInstance("DSA", "SUN");

			// test 1: sign a Byte[] object, verify signature
			byte[] testx = new byte[dataSize];
			SignedObject z = new SignedObject(testx);
			z.sign(prikey, engine);
			verify = z.verify(pubkey, engine);

			// test 2: sign a String object, verify getContent() also
			String testy = new String("I am fine.");
			SignedObject z1 = new SignedObject(testy);
			z1.sign(prikey, engine);
			verify1 = z1.verify(pubkey, engine);
			if (!testy.equals(z1.getContent())) {
				verify1 = false;
			}
		} catch (Exception e) {
			System.err.println("unexpected error in signed object test.");
			throw new TestException("unexpected exception: " + e.getMessage());
		}
		if ((!verify) || (!verify1)) {
			System.err.println("Signed object test failed.");
			throw new TestFailureException("test failed: signed object " +
										   "did not verify.");
		} else {
			System.out.println("Signed object test passed.");
		}
	}
}

