/*
 * @(#)CertAttrTest.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
 
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.util.Enumeration;
import java.util.Vector;
import sun.security.x509.*;
import sun.security.util.ObjectIdentifier;

/**
 * Usage: CertAttrTest <fileName>
 */
public class CertAttrTest {
    public static void print_attrs(Object obj)
    throws CertificateException, IOException {
        String DEPTH = ".........,,,,";
        String DOT = ".";
        Object ins = null;

        if (obj instanceof CertificateVersion) {
            CertificateVersion vers = (CertificateVersion)obj;
            for (Enumeration j = vers.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = vers.get(attr);
            }
        } else if (obj instanceof CertificateSerialNumber) {
            CertificateSerialNumber sn = (CertificateSerialNumber)obj;
            for (Enumeration j = sn.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = sn.get(attr);
            }
        } else if (obj instanceof CertificateAlgorithmId) {
            CertificateAlgorithmId algId = (CertificateAlgorithmId)obj;
            for (Enumeration j = algId.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = algId.get(attr);
            }
        } else if (obj instanceof CertificateIssuerName) {
            CertificateIssuerName issuer = (CertificateIssuerName)obj;
            for (Enumeration j = issuer.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = issuer.get(attr);
            }
        } else if (obj instanceof CertificateValidity) {
            CertificateValidity val = (CertificateValidity)obj;
            for (Enumeration j = val.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = val.get(attr);
            }
        } else if (obj instanceof CertificateSubjectName) {
            CertificateSubjectName subj = (CertificateSubjectName)obj;
            for (Enumeration j = subj.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = subj.get(attr);
            }
        } else if (obj instanceof CertificateX509Key) {
            CertificateX509Key key = (CertificateX509Key)obj;
            for (Enumeration j = key.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = key.get(attr);
            }
        } else if (obj instanceof CertificateIssuerUniqueIdentity) {
            CertificateIssuerUniqueIdentity id =
                                     (CertificateIssuerUniqueIdentity)obj;
            for (Enumeration j = id.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = id.get(attr);
            }
        } else if (obj instanceof CertificateSubjectUniqueIdentity) {
            CertificateSubjectUniqueIdentity id =
                                     (CertificateSubjectUniqueIdentity)obj;
            for (Enumeration j = id.getElements(); j.hasMoreElements();) {
                String attr = (String)j.nextElement();
                System.out.println(DEPTH + attr);
                ins = id.get(attr);
            }
        } else if (obj instanceof CertificateExtensions) {
            if (obj instanceof Vector) {
                Vector ext = (Vector)obj;
                for (Enumeration j = ext.elements(); j.hasMoreElements();) {
                    Extension exts = (Extension)j.nextElement();
                    System.out.println("                " +
                              ((Object)exts).getClass().getName());
                }
            }
        }
        if (ins != null)
           System.out.println("                " +
                              ins.getClass().getName());
    }

    public static void main(String[] args) {
            if (args.length != 1) {
                System.out.println("Usage: CertAttrTest <fileName>");
                System.exit(1);
            }
            try {
                InputStream inStrm = new FileInputStream(args[0]);
                X509CertImpl crt = new X509CertImpl(inStrm);
                inStrm.close();
                for (Enumeration i = crt.getElements(); i.hasMoreElements();) {
                    String attr = (String)i.nextElement();
                    System.out.println(attr);
                    Object obj = crt.get(attr);

                    if (obj instanceof X509CertInfo) {
                      X509CertInfo info = (X509CertInfo)obj;
                      for (Enumeration j = info.getElements();
                           j.hasMoreElements();) {
                          String attrInside = (String)j.nextElement();
                          Object ins = info.get(attrInside);
                          System.out.println("........." + attrInside);
                          if (ins != null)
                              System.out.println("         " +
                                                 ins.getClass().getName());
                          print_attrs(ins);
                      }
                    }
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
    } 
}
