/*
 * @(#)CertGen.java	1.16 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.io.*;
import sun.security.x509.*;

/**
 * This is a class used to test the X.509 cert generation code.
 */
public class CertGen {

    // Usage:  CertGen key-alg sig-algorithm
    // ... key-alg = "DSA", "RSA"
    //	... sig-algorithm = "SHA/DSA", "MD5/RSA"
    public static void main(String argv[])
    {
	X500Name	subject;
	CertAndKeyGen	keypair;
	X509CertImpl	x1, x2;
	byte		cert[];
        String scert_attr = X509CertImpl.NAME + "." +
                            X509CertImpl.SIGNED_CERT;

	if (argv.length != 2) {
	    System.out.println("usage:  CertGen <key-alg> <sig-alg>");
	    System.exit(1);
	}

	try {
	    subject = new X500Name("Zaphod Beeblebrox",
		    "Transgalactic Inc.", "Ursa Major", "US");

	    keypair = new CertAndKeyGen(argv[0], argv[1]);
	    keypair.generate(1024);
	    x1 = (X509CertImpl)keypair.getSelfCertificate(subject,
		60	// seconds per minute
		* 60	// minutes per hour
		* 24	// hours per day
		* 30	// days per month (more or less)
		* 6	// months
		);
	    cert = (byte[])x1.get(scert_attr);
            
	    x2 = new X509CertImpl(cert);

	    System.out.println("Generated Cert:  " + x2);

	    x1.verify(keypair.getPublicKey());
	    x2.verify(keypair.getPublicKey());

	    if (! x1.equals(x2)) {
		System.out.println("?? x1 != x2 ??");
		System.out.println(" !!  TEST FAILED !! ");
		System.exit(1);
	    }

	} catch (Exception e) {
	    System.out.println("Exception:  " + e);
	    e.printStackTrace();
	    System.out.println(" !!  TEST FAILED !! ");
	    System.exit(1);
	}

	System.out.println("PASS:  Constructed and verified a "
	    + "self-signed X.509 Cert, using '" + argv[1] + "'");
	System.out.println();
    }
}
