/*
 * @(#)ExtensionsTest.java	1.1 98/07/15
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.util.Calendar;
import java.util.Date;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;
import java.security.PrivateKey;
import java.security.KeyPair;
import sun.security.x509.*;
import sun.security.util.ObjectIdentifier;

public class ExtensionsTest {
    public static void main(String[] args) {
        X500Name	subject, issuer;
        KeyPairGenerator keyGen;
        String		sigAlg;
        SecureRandom	prng;
        X509Key		publicKey;
        PrivateKey	privateKey;
        KeyPair		pair;

        try {
            CertificateExtensions exts = new CertificateExtensions();

            RFC822Name mail = new RFC822Name("duke@java.sun.com");
            DNSName dns = new DNSName("java.sun.com");
            URIName uri = new URIName("http://java.sun.com");

            byte[] address = new byte[4];
            address[0] = 127; address[1] = 0; address[2] = 0; address[3] = 1;
            IPAddressName ip = new IPAddressName(address);
 
            int[] oidData = new int[4];
            oidData[0] = 1; oidData[1] = 2; oidData[2] = 3; oidData[3] = 4;
            OIDName oid = new OIDName(new ObjectIdentifier(oidData));

            SubjectAlternativeNameExtension subjectName =
                                new SubjectAlternativeNameExtension();
            IssuerAlternativeNameExtension issuerName =
                                new IssuerAlternativeNameExtension();

            GeneralNames subjectNames = new GeneralNames();
            //subjectName.get(SubjectAlternativeNameExtension.SUBJECT_NAME);

            GeneralNames issuerNames = new GeneralNames();
            //issuerName.get(IssuerAlternativeNameExtension.ISSUER_NAME);

            subjectNames.addElement(mail);
            subjectNames.addElement(dns);
            subjectNames.addElement(uri);
            subjectName.set(subjectName.SUBJECT_NAME, subjectNames);

            issuerNames.addElement(ip);
            issuerNames.addElement(oid);
            issuerName.set(issuerName.ISSUER_NAME, issuerNames);

            Date firstDate = new Date();
            Calendar cal = Calendar.getInstance();
            cal.set(2000, 11, 15, 12, 30, 30);
            Date lastDate = cal.getTime();
            PrivateKeyUsageExtension pkusage =
                new PrivateKeyUsageExtension(firstDate, lastDate);

            KeyUsageExtension usage = new KeyUsageExtension();
            usage.set(usage.CRL_SIGN, new Boolean(true));
            usage.set(usage.DIGITAL_SIGNATURE, new Boolean(true));
            usage.set(usage.NON_REPUDIATION, new Boolean(true));

            byte[] keyId = {1, 2, 3, 4, 5};
            KeyIdentifier kid = new KeyIdentifier(keyId);
            SerialNumber sn = new SerialNumber(42);
            AuthorityKeyIdentifierExtension aki = new
                AuthorityKeyIdentifierExtension(kid, subjectNames, sn);

            SubjectKeyIdentifierExtension ski = new
                SubjectKeyIdentifierExtension(keyId);

            BasicConstraintsExtension cons = new
                             BasicConstraintsExtension(true, 10);

            PolicyConstraintsExtension pce = new
                             PolicyConstraintsExtension(2, 4);

            exts.set(subjectName.NAME, subjectName);
            exts.set(issuerName.NAME, issuerName);
            exts.set(pkusage.NAME, pkusage);
            exts.set(usage.NAME, usage);
            exts.set(ski.NAME, ski);
            exts.set(aki.NAME, aki);
            exts.set(cons.NAME, cons);
            exts.set(pce.NAME, pce);

            ByteArrayOutputStream boas = new ByteArrayOutputStream();
            exts.encode(boas);
            CertificateExtensions testExt = new CertificateExtensions();
            ByteArrayInputStream bais = new ByteArrayInputStream(
                                        boas.toByteArray());
            testExt.decode(bais);

            if (! exts.equals(testExt))
                throw new Exception("Decoded extension does not match\nOrig: "
                    + exts.toString() + "\nDecoded: " + testExt.toString());

            int hash1 = exts.hashCode();
            int hash2 = testExt.hashCode();
            if (hash1 != hash2)
                throw new Exception("HashCode broken");
            System.out.println("hashCodes okay, "+ hash1);

            // change and re-test
            issuerNames.addElement(uri);
            issuerName.set(issuerName.ISSUER_NAME, issuerNames);
            exts.set(issuerName.NAME, issuerName);
            if (exts.equals(testExt))
                throw new Exception("Extension should not match\nOrig: "
                    + exts.toString() + "\nDecoded: " + testExt.toString());
            System.out.println("modified ext and equals is okay");

            // basically delete issuerAltName
            issuerName.delete(issuerName.ISSUER_NAME);
            exts.set(issuerName.NAME, issuerName);
            if (exts.equals(testExt))
                throw new Exception("Extension matches\nOrig: "
                    + exts.toString() + "\nDecoded: " + testExt.toString());
            System.out.println("deleted ext and equals is okay");

            testExt.delete(issuerName.NAME);
            pkusage.delete(pkusage.NOT_BEFORE);
            exts.set(pkusage.NAME, pkusage);
            if (exts.equals(testExt))
                throw new Exception("Extension matches\nOrig: "
                    + exts.toString() + "\nDecoded: " + testExt.toString());
            System.out.println("changed pkusage ext and equals is okay");

        } catch (Exception e) {
            e.printStackTrace();
        }
    } 
}
