/*
 * @(#)PKCS10Test1.java	1.6 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.io.*;
import java.util.*;
import java.math.*;
import java.security.*;

import sun.security.x509.*;
import sun.security.util.*;
import sun.security.pkcs.*;
import sun.misc.*;

/*
 * test old pkcs10
 */
public class PKCS10Test1 {
public static void main (String [] args) {
  X500Name			subject;
  KeyPairGenerator		keyGen;
  String			sigAlg;
  SecureRandom		prng;
  X509Key			publicKey;
  PrivateKey			privateKey;
  KeyPair			pair;
  
  if (args.length != 2) {
     System.out.println("Usage: PKCS10Test1 <keyAlg> <signatureAlg>");
     System.exit(1);
  }
  try {
    subject = new X500Name("Dr Duke","Acme Corp.","Palo Alto","US");
    keyGen = KeyPairGenerator.getInstance(args[0]);
    sigAlg = args[1];
    
    prng = new SecureRandom ();
    keyGen.initialize (1024, prng);
    pair = keyGen.generateKeyPair ();      
    publicKey = (X509Key)pair.getPublic();      
    privateKey = pair.getPrivate ();
    
    Signature	signature = Signature.getInstance(sigAlg);
    signature.initSign (privateKey);
    X500Signer issuer = new X500Signer (signature, subject);

    // Create and encode the PKCS10 request
    ByteArrayOutputStream encStream = new ByteArrayOutputStream();
    PrintStream pout = new PrintStream(encStream);
    PKCS10 req = new PKCS10(publicKey);
    req.encodeAndSign(issuer);
    req.print(pout);
    System.out.println(req.toString());

    // Decode the same
    ByteArrayInputStream fin = new ByteArrayInputStream(encStream.toByteArray());
    ByteArrayOutputStream decStream = new ByteArrayOutputStream();
    BASE64Decoder decoder = new BASE64Decoder();
    String line;
    BufferedReader dis = new BufferedReader(new InputStreamReader(fin));
    while ((line = dis.readLine()) != null) {
      if (line.equals("-----BEGIN NEW CERTIFICATE REQUEST-----")) {
	continue;
      }
      if (line.equals("-----END NEW CERTIFICATE REQUEST-----")) {
	continue;
      }
      decStream.write(decoder.decodeBuffer(line));
    }
    PKCS10 resp = new PKCS10(decStream.toByteArray());
    System.out.println(resp.toString());
  } catch (Exception e) {
    e.printStackTrace();
  }
}
}
