/*
 * @(#)PKCS7SignerOrder.java	1.5 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

import java.io.*;
import java.util.*;
import java.security.cert.*;
import java.security.*;

import sun.security.util.*;
import sun.security.x509.*;
import sun.security.pkcs.*;

import sun.misc.HexDumpEncoder;

/**
 *  test  -- signed data
 *   - test verification
 *   - success also demonstrates misconception bug that DSA signing satisfies 
 *     PKCS7 signing spec.  (Doesn't tag with digest algorithm in the middle. 
 *     test should fail when signing is correctly implemented, since
 *     simple signer does it incorrectly with DSA. 
 */
public class PKCS7SignerOrder {

    static HexDumpEncoder hexDump = new HexDumpEncoder();

    static String derString1 = 
        "30 0F 06 09 2A 86 48 86   F7 0D 01 07 02 A0 02 30 00";

    static String derString2 = 
        "30 0F 06 09 2A 86 48 86   F7 0D 01 07 02 A0 02 30 00 11 22 33";


    static byte[] derBytes1 = parseHexBytes(derString1);
    static byte[] derBytes2 = parseHexBytes(derString2);

    static DerInputStream derIn;
    static DerOutputStream derOut = new DerOutputStream();

    public static void main(String[] argv) {


	try {
	    SignerInfo[] signerInfos = new SignerInfo[9];
	    SimpleSigner signer1 = new SimpleSigner(null, null, null, null);
	    signerInfos[8] = signer1.genSignerInfo(derBytes1);
	    signerInfos[7] = signer1.genSignerInfo(new byte[] {} );
	    signerInfos[6] = signer1.genSignerInfo(derBytes2);

	    SimpleSigner signer2 = new SimpleSigner(null, null, null, null);
	    signerInfos[5] = signer2.genSignerInfo(derBytes1);
	    signerInfos[4] = signer2.genSignerInfo(new byte[] {} );
	    signerInfos[3] = signer2.genSignerInfo(derBytes2);

	    SimpleSigner signer3 = new SimpleSigner(null, null, null, null);
	    signerInfos[2] = signer3.genSignerInfo(derBytes1);
	    signerInfos[1] = signer3.genSignerInfo(new byte[] {} );
	    signerInfos[0] = signer3.genSignerInfo(derBytes2);

	    ContentInfo contentInfo = new ContentInfo(derBytes1);

	    // these are never used anyway
	    AlgorithmId[] algIds = {new AlgorithmId(AlgorithmId.MD5_oid),
				    new AlgorithmId(AlgorithmId.SHA_oid)};

	    
	    X509Certificate[] certs = { signer2.getCert(), signer1.getCert() };


	    PKCS7 pkcs71 = new PKCS7(algIds, contentInfo,
				    certs, 
				    signerInfos);

	    System.out.println("SignerInfos in original.");
	    printSignerInfos(pkcs71.getSignerInfos());

	    DerOutputStream out = new DerOutputStream();
	    pkcs71.encodeSignedData(out);

	    PKCS7 pkcs72 = new PKCS7(out.toByteArray());
	    System.out.println("\nSignerInfos read back in:");
	    printSignerInfos(pkcs72.getSignerInfos());

	    System.out.println("Verified signers of original:");
	    SignerInfo[] verifs1 = pkcs71.verify();
	    for (int i = 0; i < verifs1.length; i++) {
		printArrayOccurrences(verifs1[i], signerInfos);
	    }
	    System.out.println();

	    System.out.println("Verified signers of after read-in:");
	    SignerInfo[] verifs2 = pkcs72.verify();
	    for (int i = 0; i < verifs2.length; i++) {
		printArrayOccurrences(verifs2[i], pkcs72.getSignerInfos());
	    }
	    System.out.println();

	    System.out.println("The second list of signerInfos should have " +
			       "the same elements as \n" +
			       "the first, but in order, and the lists of verified \n" +
			       "signers should have the same number of elements.\n");
	} catch (Exception e) {
	    e.printStackTrace(System.out);
	}

    }
	
    static byte[] parseHexBytes(String hexBytes) {
	StringTokenizer tokenizer = new StringTokenizer(hexBytes);

	byte[] bytes = new byte[tokenizer.countTokens()];

	for(int i=0; i < bytes.length; i++) {
	    bytes[i] = (byte) Integer.parseInt(tokenizer.nextToken(), 16);
	}
	return bytes;
    }

    static void printArrayOccurrences(Object x, Object[] a) {
	if (a == null) return;
	for (int i = 0; i < a.length; i++) {
	    if (x == a[i]) {
		System.out.print(" " + i);
	    }
	}
    }

    static void printSignerInfos(SignerInfo[] signerInfos) throws IOException {
	    ByteArrayOutputStream strm = new ByteArrayOutputStream();
	    for (int i = 0; i < signerInfos.length; i++) {
		signerInfos[i].derEncode(strm);
		System.out.println("SignerInfo[" + i + "], length: " + 
				   strm.toByteArray().length);
		System.out.println(hexDump.encode(strm.toByteArray()));
		System.out.println("\n");
		strm.reset();
	    }
    }

} 

