/*
 * @(#)V3CertInfo.java	1.5 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.io.*;
import java.util.*;
import java.security.*;
import sun.security.x509.*;
import sun.security.util.*;

/**
 * Usage: V3CertInfo <key_alg> <sig_alg>
 * Generates a V3 certificate with no extensions.
 */
public class V3CertInfo {
    public static void main (String[] args) {
    byte[]			issuerId = { 1, 2, 3, 4, 5 };
    byte[]			subjectId = { 6, 7, 8, 9, 10 };
    byte[]			subjectKey = { 12, 13, 14, 15, 16, 17 };
    X500Name			subject;
    KeyPairGenerator		keyGen;
    String			sigAlg;
    SecureRandom		prng;
    X509Key			publicKey;
    PrivateKey			privateKey;
    KeyPair			pair;
      
    try {
      subject = new X500Name("Dr Duke","ACME Corp.","Palo Alto","US");
      keyGen = KeyPairGenerator.getInstance(args[0]);
      sigAlg = args[1];
      
      prng = new SecureRandom ();
      keyGen.initialize (1024, prng);
      pair = keyGen.generateKeyPair ();      
      publicKey = (X509Key)pair.getPublic();      
      privateKey = pair.getPrivate ();

      Signature	signature = Signature.getInstance(sigAlg);
      signature.initSign (privateKey);
      X500Signer issuer = new X500Signer (signature, subject);

      Date firstDate = new Date();
      Date lastDate = new Date();
      lastDate.setTime(lastDate.getTime() + 1000 * 1000);
      CertificateValidity interval = new CertificateValidity(firstDate,lastDate);

      X509CertInfo cert = new X509CertInfo();

      cert.set(X509CertInfo.VERSION,
                        new CertificateVersion(CertificateVersion.V3));
      cert.set(X509CertInfo.SERIAL_NUMBER,
                 new CertificateSerialNumber((int) (firstDate.getTime()/1000)));
      cert.set(X509CertInfo.ALGORITHM_ID,
                        new CertificateAlgorithmId(issuer.getAlgorithmId()));
      cert.set(X509CertInfo.ISSUER,
                        new CertificateIssuerName(issuer.getSigner()));
      cert.set(X509CertInfo.SUBJECT,
                        new CertificateSubjectName(subject));
      cert.set(X509CertInfo.KEY,
                        new CertificateX509Key(publicKey));
      cert.set(X509CertInfo.VALIDITY, interval);

      //v2
      cert.set(X509CertInfo.ISSUER_ID,
       new CertificateIssuerUniqueIdentity(new UniqueIdentity(issuerId)));
      cert.set(X509CertInfo.SUBJECT_ID,
       new CertificateSubjectUniqueIdentity(new UniqueIdentity(subjectId)));

      System.out.println("V3 CertificateInfo:NO extensions: \n" + cert);

      X509CertInfo crt = new X509CertInfo(cert.getEncodedInfo());
      System.out.println("V3 CertificateInfo(new): \n" + crt);
    } catch (Exception e) {
      e.printStackTrace();
    }
  } 
}
