/*
 * @(#)EchoObj.java	1.2 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.net.*;
import java.io.*;
import java.util.Date;

/*
 * Simple class to echo objects.  It creates a thread that accepts
 * new connections and echos all objects that it receives.
 * Command line arguments are parsed as classes to be echoed.
 */
class EchoObj implements Runnable {
    String host = null;
    int iterations = 1;
    static boolean verbose = false;
    static ServerSocket serversocket = null;


    Date start;
    Date end;

    public static void main(String argv[]) {
	if (argv.length == 0) {
	    String[] def = {"java.util.Hashtable"};
	    argv = def;
	}

	EchoObj echo = new EchoObj();
	echo.start(argv);
    }

    private void start(String[] argv) {
	int nextarg;
	for (nextarg = 0; nextarg < argv.length; nextarg++) {
	    if (argv[nextarg].equals("-r")) {
		if ((nextarg + 1) < argv.length) {
		    iterations = Integer.parseInt(argv[++nextarg]);
		} else {
		    System.out.println("-r requires a repeat count");
		    usage();
		    return;
		}
	    } else if (argv[nextarg].equals("-h")) {
		if ((nextarg + 1) < argv.length) {
		    host = argv[++nextarg];
		} else {
		    System.out.println("-h requires a hostname");
		    usage();
		    return;
		}
	    } else if (argv[nextarg].equals("-v"))
		verbose = true;
	    else if (argv[nextarg].startsWith("-")) {
		System.out.println("invalid flag: " + argv[nextarg]);
		usage();
		return;
	    } else 
		break;		// rest of args are classes to test
	}

	
	try {
	    if (host == null) {
		serversocket = new ServerSocket(5066);
		new Thread(new EchoObj()).start();
	    }

	    InetAddress local = InetAddress.getLocalHost();

	    InetAddress locals[] = InetAddress.getAllByName(local.getHostName());

	    Socket client = null;
	    if (host != null) {
		client = new Socket(host, 5066);
	    } else {
		client = new Socket("", 5066);
	    }
		
	    ObjectOutputStream os = new ObjectOutputStream(client.getOutputStream());
	    os.flush();
	    ObjectInputStream is = new ObjectInputStream(client.getInputStream());
	    // For each classname on the command line
	    // create an instance, send it , receive it and print each.
	    for (;nextarg < argv.length; nextarg++) {
		Object o = makeobj(argv[nextarg]);

		start = new Date();
		for (int i = 0; i < iterations; i++) {
		    if (verbose) System.out.println("Writing " + o);
		    os.writeObject(o);
		    os.flush();
		    Object robj = is.readObject();
		    if (verbose) System.out.println("Reading " + o);
		}
		end = new Date();
		System.out.println(argv[nextarg] + ": " + this);
	    }

	} catch (Exception e) {
	    e.printStackTrace();
	}

    }


    private Object makeobj(String arg) throws Exception {
	Object o = null;
	if (arg.startsWith("+")) {
	    o = new int[2400];
	} else {
	    Class cl = Class.forName(arg);
	    o = cl.newInstance();
	}
	return o;
    }

    /**
     * Wait for a connection, when one is received create a new
     * EchoObj to handle the next accept and start echoing objects.
     */
    public void run() {
	try {
	    Socket s = serversocket.accept();
	    if (verbose) System.out.println("Starting new accept thread");
	    new Thread(new EchoObj()).start();

	    ObjectOutputStream out = new ObjectOutputStream(s.getOutputStream());
	    out.flush();
	    ObjectInputStream in = new ObjectInputStream(s.getInputStream());
	    int count = 1;
	    while (true) {
		Object o = in.readObject();
		if (verbose) System.out.println("Echo " + count++ + ": " + o);
		out.writeObject(o);
		out.flush();
	    }
	} catch (EOFException e) {
	    return;
	} catch (IOException e) {
	    e.printStackTrace();
	} catch (ClassNotFoundException e2) {
	    e2.printStackTrace();
	}
    }

    private static void usage() {
	System.out.println("arguments are: [-r <iterations>] [-h <hostname>] [-v] classnames");
	System.out.println("\t-r n\trepeat for n iterations");
	System.out.println("\t-h\thostname to echo classes on");
	System.out.println("\t-v\tverbose");
    }

    public String toString() {
	long duration = (start == null || end == null ) ?
	    0 :
	    end.getTime() - start.getTime();
	return duration + " ms, (" + (iterations * 1000.0)/duration + " per sec)";
    }

}
