/*
 * @(#)IPerf.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.io.*;
import java.util.Date;

class IPerf extends ObjectInputStream {

    private int classes = 0;	// classes unpickled
    private int objects = 0;	// objects unpickled
    private Date start = null;
    private  Date end = null;
    double itsPerSec = 0;          // How many iterations per second.



    static int WARMUP_ITERATIONS = 10; // Warm up the VM with this many iterations.
    static   int test_iterations = 10; // In the warm case run with this many iterations.
    int executions; // How many times did we execute last
    long duration; // How long did this test take anyway ??
    static int verbosity = 0;
    static FeedBack reporter= null;
    ByteArrayInputStream iStream;



    public static void main (String argv[]) {

	int nextarg;

	String fileName; // output results here.

	reporter = new screenRecorder( ); // By default results go to the screen.


	for (nextarg = 0; nextarg < argv.length; nextarg++) {
	    if (argv[nextarg].equals("-r")) {
		if ((nextarg + 1) < argv.length) {
		    test_iterations = Integer.parseInt(argv[++nextarg]);
		} else {
		    System.out.println("-r requires argument");
		    usage();
		    return;
		}

		    // Store results in a file ?

	    } else if (argv[nextarg].equals("-v")) {
		verbosity = 1;
		System.out.println("verbosity = " + verbosity);
	    } else if (argv[nextarg].equals("-vv")) {
		verbosity = 2;


		System.out.println("verbosity = " + verbosity);

	    }
	    else if ( argv[nextarg].equals("-f")) {
            ++nextarg;
            if ( nextarg == argv.length || argv[nextarg].startsWith("-")){
                System.out.println("invalid flag:" + argv[nextarg]);
                System.out.println("-f requires a file name");
                usage();
            }
            else {
                   reporter = new fileRecorder(argv[nextarg]);
            }
        }

	    else if (argv[nextarg].startsWith("-")) {
		System.out.println("invalid flag: " + argv[nextarg]);
		usage();
		return;
	    } else
		break;		// rest of args are classes to test
	}

	// For each class on the command line, test it
	for (nextarg = nextarg; nextarg < argv.length; nextarg++) {
	    String cname = argv[nextarg];

	    try {
		FileInputStream ifs = new FileInputStream(cname + ".tmp");
		byte[] data = new byte[ifs.available()];
		ifs.read(data);
		ifs.close();
		ByteArrayInputStream istreamTmp = new ByteArrayInputStream(data);


		// Load the class to minimize the effect of class
		// loading on perceived performance.
		Class.forName(cname);

		// First measure the "cold" performance
		IPerf perf = null;
		perf = new IPerf(istreamTmp);
		perf.IPerf(cname, false,1);
		if (verbosity <= 0)
		    reporter.writeResults(perf.toString());
		else
		    reporter.writeResults("Cold:: " + cname + ":\t" + perf.toString());

		// Warm up the VM
		System.gc();
		for (int j = 0; j < WARMUP_ITERATIONS; j++ ) {
		    istreamTmp.reset();
		    perf = new IPerf(istreamTmp);
		    perf.IPerf(cname, false,1);
		}
		System.gc();

        istreamTmp.reset();
		ObjectInputStream objis = new ObjectInputStream(istreamTmp);
		Object tmp = (Object)objis.readObject();
		Class cl = tmp.getClass();

		ByteArrayOutputStream ostream = new ByteArrayOutputStream();


		ObjectOutputStream objos = new ObjectOutputStream(ostream);

		for (int x = 0; x < test_iterations; x++)
		{
		    tmp = cl.newInstance();
		    objos.writeObject(tmp);
		}

		byte[] data2 = ostream.toByteArray();

		ByteArrayInputStream istream = new ByteArrayInputStream(data2);




		// Now do the real timing test..

    		istream.reset();
	    	perf = new IPerf(istream);
		    perf.IPerf(cname, false,test_iterations);
		if (verbosity > 1)
		    gcstats();
		if (verbosity <= 0)
		    reporter.writeResults(perf.toString());
		else
		    reporter.writeResults("Warm:: " + cname + ":\t" + perf);
	    } catch (IOException ex) {
		ex.printStackTrace();
	    } catch (ClassNotFoundException ex) {
		System.out.println("Class " + cname + " not found.");
	    } catch (IllegalAccessException ex) {
		System.out.println("Class " + cname + " does not have a public default constructor.");
	    } catch (InstantiationException ex) {
		System.out.println("Class " + cname + " Could not make an instance.");
	    }


	  System.out.println("Finished " + cname);
	}
	  reporter.end();
    }

    /*
     * Create a new stream and enable replaceObject
     */
    IPerf(ByteArrayInputStream out) throws IOException {
	super(out);
	iStream = out;
	try {
	    enableResolveObject(true);
	} catch (Throwable e) {
	    e.printStackTrace();
	}
    }

    /**
     * Test that all the primitive write* methods work.
     */
    public void IPerf(String classname, boolean trace, int execs )
	throws IOException, ClassNotFoundException, InstantiationException,
	    IllegalAccessException
    {


    executions = execs;


	start = new Date();

	Runtime.getRuntime().traceMethodCalls(trace);



	 for (int i = 0; i < execs; i++) {

	        Object  obj = this.readObject();
	  }



	Runtime.getRuntime().traceMethodCalls(false);
	end = new Date();


     duration = (start == null || end == null ) ?
	    0 : end.getTime() - start.getTime();

	if ((duration > 0) && (executions > 1))
        itsPerSec = ((double) executions / (double) duration ) * 1000.0;
     else
        itsPerSec =  1000.0 / duration ;


	}

    protected Class resolveClass(ObjectStreamClass classdesc)
	throws IOException, ClassNotFoundException
    {
	classes++;
	return super.resolveClass(classdesc);
    }

    protected Object resolveObject(Object obj) {
	objects++;
	return obj;
    }

    public String toString() {

	if (verbosity <= 1)
	    return String.valueOf(duration);
	else
	   //return duration + " ms, (" + 1000.0/duration + " per sec), " +
		// classes + " Classes, " + objects + " Objects";
		 return duration + " ms / " + executions + " its, (" + itsPerSec + " per sec), " +
		classes + " Classes, " + objects + " Objects";
    }

    private static void usage() {
	System.out.println("arguments are: [-r n] classnames");
	System.out.println("\t-r n\trepeat for n iterations");
	System.out.println("\t-v\tverbose output");
	System.out.println("\t-vv\tvery verbose output");
	System.out.println("\t-f filename v\t log output to specified file");

    }

    private static void gcstats() {
	Runtime r = Runtime.getRuntime();
	try {
	    reporter.writeResults("Memory free = " + r.freeMemory()/1024 + "K.");
	}
	catch (IOException e)
	{
	       System.out.println("Could not write free memory results");
	}
    }
}

