/*
 * @(#)OPerf.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.io.*;
import java.util.*;
import java.lang.*;

class OPerf extends ObjectOutputStream {

    private int classes = 0;	// classes pickled
    private int objects = 0;	// objects pickled
    double itsPerSec = 0;          // How many iterations per second.
    private Date start = null;
    private Date end = null;
    private Object obj;
    private Object interations_vector[];  // We need to write a unique object for
                                        // for each iteration.
    static private OutputStream reportStream = System.out;

    static int WARMUP_ITERATIONS = 10; // Warm up the VM with this many iterations.
    static   int test_iterations = 10; // In the warm case run with this many iterations.
            int executions; // How many times did we execute last
    long duration; // How long did this test take anyway ??
    static boolean writefile = false;
    static int verbosity = 0;
    static FeedBack reporter= null;

    public static void main (String argv[]) {
	// Warm up the mechanism but nothing specific to the class to be tested
	ByteArrayOutputStream ostream = null;
	OPerf perf = null;
	int nextarg;
	String fileName; // output results here.

	reporter = new screenRecorder( ); // By default results go to the screen.

	for (nextarg = 0; nextarg < argv.length; nextarg++) {
	    if (argv[nextarg].equals ("-w"))
		    writefile = true;

	    else if (argv[nextarg].equals("-r")) {
		    if ((nextarg + 1) < argv.length) {
		     test_iterations = Integer.parseInt(argv[++nextarg]);
		    } else {
		        System.out.println("-r requires argument");
		     usage();
		     return;
		    }


	    } else if (argv[nextarg].equals("-v"))
		verbosity = 1;

	    else if (argv[nextarg].equals("-vv"))
		verbosity = 2;

        // Store results in a file ?
        else if ( argv[nextarg].equals("-f")) {
            ++nextarg;
            if ( nextarg == argv.length || argv[nextarg].startsWith("-")){
                System.out.println("invalid flag:" + argv[nextarg]);
                System.out.println("-f requires a file name");
                usage();
            }
            else {
                   reporter = new fileRecorder(argv[nextarg]);
            }
        }



	    else if (argv[nextarg].startsWith("-")) {
		System.out.println("invalid flag: " + argv[nextarg]);
		usage();
		return;
	    } else
		    break;
	// rest of args are classes to test
	}

	// For each class on the command line, test it
	for (nextarg = nextarg; nextarg < argv.length; nextarg++) {
	    String cname = argv[nextarg];
	    try {
		// First measure the "cold" performance, no classes loaded
		ostream = new ByteArrayOutputStream();
		ostream.reset();
		perf = new OPerf(cname, ostream);
		perf.operf(1);
		if (verbosity <= 0)
		    reporter.writeResults(perf.toString());
		else if (verbosity == 1)
		    reporter.writeResults("Cold:" + cname + ":\t" + perf);
		else
		    reporter.writeResults("Cold:" + cname + ":\t" + perf + ", " +
				       ostream.size() + " bytes written");
		System.gc();
		if (verbosity > 1)
		    gcstats();
		// Warm up the VM with some iterations,

		System.gc();
		for (int j = 0; j < WARMUP_ITERATIONS; j++ ) {
		    ostream.reset();
		    perf = new OPerf(cname, ostream);
		    Runtime.getRuntime().traceMethodCalls(true);
		    perf.operf(1);
		    Runtime.getRuntime().traceMethodCalls(false);
		}
		perf.operf(test_iterations);
		System.gc();
		if (verbosity > 1)
		    gcstats();
		if (verbosity <= 0)
		    reporter.writeResults(perf.toString());
		else if (verbosity == 1)
		    reporter.writeResults("Warm: " + cname + ":\t" + perf);
		else
		    reporter.writeResults("Warm:" + cname + ":\t" + perf + ", " +
				       ostream.size() + " bytes written");

		if (writefile) {
		    FileOutputStream of = new FileOutputStream(cname + ".tmp");
		    byte[] data = ostream.toByteArray();
		    of.write(data, 0, data.length);
		    of.flush();
		    of.close();
	System.out.println("Finished: " + cname);
		}

	    } catch (IOException ex) {
		ex.printStackTrace();
	    } catch (ClassNotFoundException ex) {
		System.out.println("Class " + cname + " not found.");
	    } catch (IllegalAccessException ex) {
		System.out.println("Class " + cname + " does not have a public default constructor.");
	    } catch (InstantiationException ex) {
		System.out.println("Class " + cname + " Could not make an instance.");
	    }

	System.out.println("Finished: " + cname);
	}
	reporter.end();
	ostream = null;
	System.gc();
	perf = null;
    }

    OPerf(String cname, OutputStream out) throws IOException,
	ClassNotFoundException, InstantiationException, IllegalAccessException
    {
	super(out);
	enableReplaceObject(true);
	Class cl = Class.forName(cname);
	obj = cl.newInstance();


    }
    /**
     * Test that all the primitive write* methods work.
     */
    public void operf(int execs ) throws IOException {

    executions = execs;
    Object objects[];
    Class cl;

    objects = new Object[execs];
    try {
         cl = obj.getClass();



        objects[0] = obj;

        //If we are testing for multiple iterations we
        // need to have multiple objects. We can not
        // write the same class over and over again.


        for (int i = 1; i < execs;  i++)
        {
                objects[i] = cl.newInstance();
        }

    }
     catch (Exception e)
     {
        System.out.println( e.toString());
        System.exit(1);
     }


	start = new Date();



    for (int i = 0; i < execs; i++) {
	    this.writeObject(objects[i]);
	   // System.out.println(objects[i]);
	    // this.flush();
    }


	end = new Date();

	duration = (start == null || end == null ) ?
	    0 :
	    end.getTime() - start.getTime();

     if ((duration > 0) && (executions > 1))
        itsPerSec = ((double) executions / (double) duration ) * 1000;
     else
        itsPerSec =  1000.0 / duration ;
    }

                                                    // of executions.

    protected void annotateClass(Class cl) {
	classes++;
    }

    protected Object replaceObject(Object obj) {
	objects++;
	return obj;
    }

    public String toString() {

	if (verbosity <= 1)
	    return String.valueOf(duration);
	else
	    return duration + " ms / " + executions + " its, (" + itsPerSec + " per sec), " +
		classes + " Classes, " + objects + " Objects";
    }

    private static void usage() {
	System.out.println("arguments are: [-r n] [-w] classnames");
	System.out.println("\t-r n\trepeat for n iterations");
	System.out.println("\t-w\twriting the result to a file <classname.tmp>");
	System.out.println("\t-v\tverbose output");
	System.out.println("\t-vv\tvery verbose output");
	System.out.println("\t-f filename v\t log output to specified file");

    }

    private static void gcstats() {
	Runtime r = Runtime.getRuntime();

	try {
	    reporter.writeResults("Memory free = " + r.freeMemory()/1024 + "K.");
	}
	catch (Exception e){

    }
    }
}



