/*
 * @(#)fileRecorder.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

// Write the results to a file.

import java.lang.*;
import java.io.*;
import java.util.*;


class fileRecorder implements FeedBack {

   private RandomAccessFile file; // send all results to this file.
   private String resultsBuffer;   // keep results here until all results are in.
   private String iFileName;        // keep file name around just in case.

   public fileRecorder(String fileName)
   {

        iFileName = fileName;
        resultsBuffer = new String();





        try {
            file = new RandomAccessFile(fileName,"rw");
		    System.out.println("Length = " + file.length());
	        file.skipBytes((int)file.length());


            // At this point we should print out some key versioning
            // and platform info so we know where the results came from.
            // We only do this on the first write to the file.

            if (file.length() == 0)
                writeVersionInfo();

        }
        catch (IOException e) {
               System.out.println("Could create file" + fileName);
               System.exit(1);
        }

   }

   // This is a function to jot down some version
   private void writeVersionInfo()
   {
        Properties props = System.getProperties();

        try {
              writeResults(  "java.version = " + (String)(props.get("java.version")));
              writeResults(  "java.home = " + (String)(props.get("java.home")));
              writeResults(  "os.name = " + (String)(props.get("os.name")));
         }
         catch (IOException e)
         {
                    e.printStackTrace();
         }
   }

   // Keep everything in a temporary buffer. That we don't
   // have to worry about performance results being polluted
   // by async. File I/O.

   public void writeResults(String results) throws IOException
   {

        resultsBuffer = resultsBuffer + results + '\n';


   }

   public void  end()
   {

        try {
            file.writeBytes(resultsBuffer);
            resultsBuffer = new String();
        } catch (IOException e)
        {
            System.out.println("Could not write out results to" + iFileName);
        }
   }

}
