/*
 * @(#)ListTest.java	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/** 
 * A grab bag of tests for JList.
 * 
 * To test against JDK1.1:
 * 
 * setenv JAVA_HOME /usr/local/java/jdk1.1.5/solaris
 * setenv SWING_HOME /home/hmuller/ws/jdk12
 * setenv CLASSPATH ${SWING_HOME}/build/solaris/1.1_classes:${JAVA_HOME}/lib/classes.zip
 * ${JAVA_HOME}/bin/javac ListTest.java
 * ${JAVA_HOME}/bin/java -classpath .:${CLASSPATH} ListTest
 */

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.basic.*;
import javax.swing.plaf.*;

import java.util.Vector;
import java.util.Hashtable;

import java.io.*;



/** 
 * Display a 25 element list with constant sized elements.  
 * The container size should be exactly right.
 */
class SimpleTest extends Container
{
    SimpleTest() {

	DefaultListModel model = new DefaultListModel();
	for(int i = 0; i < 25; i++) {
	    model.addElement("DefaultListModel row " + i);
	}
	JList list = new JList(model);
	
	setLayout(new BorderLayout());
	add(list, BorderLayout.CENTER);
    }
}


/** 
 * Display a 25 element list with constant sized elements.  
 * The container size should be exactly right.
 */
class SerializationTest extends Container
{
    SerializationTest() {

	DefaultListModel model = new DefaultListModel();
	for(int i = 0; i < 25; i++) {
	    model.addElement("DefaultListModel row " + i);
	}
	JList list = new JList(model);
	
	/* Try reading/writing the JList before adding it to 
	 * the container.
	 */

	try {
	    String file = "ListTest.ser";
	    ObjectOutputStream os = new ObjectOutputStream(new FileOutputStream(file));
	    System.out.println("Writing to " + os);
	    os.writeObject(list);
	    os.close();
 
	    ObjectInputStream is = new ObjectInputStream(new FileInputStream(file));
	    System.out.println("Reading from " + is);
	    list = (JList)(is.readObject());
	}
	catch (Throwable e) {
	    e.printStackTrace();
	}

	setLayout(new BorderLayout());
	add(list, BorderLayout.CENTER);
    }
}


/** 
 * Display a 25 element list with variable sized elements.  
 * The container size should be exactly right.
 */
class VariableSizeCellTest extends Container
{
    public class CellRenderer extends BasicListCellRenderer
    {
	Font smallFont = new Font("Dialog", Font.PLAIN, 12);
	Font bigFont = new Font("Dialog", Font.PLAIN, 24);

	public Component getListCellRendererComponent(
            JList list, 
	    Object value,
	    int index, 
	    boolean isSelected, 
	    boolean cellHasFocus)
	{
	    setFont(((index % 2) == 0) ? bigFont : smallFont);
	    return super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
	}
    }

    VariableSizeCellTest() {
	DefaultListModel model = new DefaultListModel();
	for(int i = 0; i < 25; i++) {
	    model.addElement("DefaultListModel row " + i);
	}
	JList list = new JList(model);
	list.setCellRenderer(new CellRenderer());

	setLayout(new BorderLayout());
	add(list, BorderLayout.CENTER);
    }
}


/** 
 * Display a 25 element list with variable sized elements.  
 * The container size should be exactly right.
 */
class ScrollableVariableSizeCellTest extends Container
{
    public class CellRenderer extends BasicListCellRenderer
    {
	Font smallFont = new Font("Dialog", Font.PLAIN, 12);
	Font bigFont = new Font("Dialog", Font.PLAIN, 24);

	public Component getListCellRendererComponent(
            JList list, 
	    Object value,
	    int index, 
	    boolean isSelected, 
	    boolean cellHasFocus)
	{
	    setFont(((index % 2) == 0) ? bigFont : smallFont);
	    return super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
	}
    }

    ScrollableVariableSizeCellTest() {
	DefaultListModel model = new DefaultListModel();
	for(int i = 0; i < 25; i++) {
	    model.addElement("DefaultListModel row " + i);
	}
	JList list = new JList(model);
	list.setCellRenderer(new CellRenderer());
	setLayout(new BorderLayout());
	add(new JScrollPane(list), BorderLayout.CENTER);
    }
}



/** 
 * Display four 16 element lists:
 * 1 - cell width 256, and cell height 32
 * 2 - cell width 256, and cell height should be exact
 * 3 - cell width should be exact, and cell height should be 32
 * 4 - cell width and height should be exact
 */

class FixedCellSizeTest extends Container
{
    FixedCellSizeTest() {

	DefaultListModel model = new DefaultListModel();
	for(int i = 0; i < 16; i++) {
	    model.addElement("FixedCellSize row " + i);
	}

	JList list1 = new JList(model);
	list1.setFixedCellWidth(256);
	list1.setFixedCellHeight(32);

	JList list2 = new JList(model);
	list2.setFixedCellWidth(256);

	JList list3 = new JList(model);
	list3.setFixedCellHeight(32);

	JList list4 = new JList(model);

	setLayout(new FlowLayout(FlowLayout.LEFT, 8, 8)); 
	add(list1);
	add(list2);
	add(list3);
	add(list4);
    }
}


class DynamicModelTest extends Container
{
    /* This class comes from the September 1997 version of the 
     * Swing file chooser.  It exposed many JListBox problems
     * when the model was changed dynamically.
     */

    static class StringSet extends AbstractListModel 
    {
	private int size;
	private int lastnsize;
	String name = "xx";
	private String a[] = new String[10];

	public int getSize() {
	    return size;
	}

	public Object getElementAt(int i) {
	    return i < 0 || i >= size ? "Missing" : a[i];
	}

	private void notifyNew() {
	    if (size > lastnsize) {
		int oldlastnsize = lastnsize;
		lastnsize = size;
		fireIntervalAdded(this, oldlastnsize, size);
	    }
	}

	public void done() {
	    notifyNew();
	}

	public void removeAll() {
	    int oldSize = size;
	    size = lastnsize = 0;
	    fireIntervalRemoved(this, 0, oldSize);
	}

	public void add(String s) {
	    if (size >= a.length) {
		String na[] = new String[size * 2];
		System.arraycopy(a, 0, na, 0, size);
		a = na;
	    }
	    a[size++] = s;
	    if (size == 10)
		notifyNew();
	}
    }


    DynamicModelTest() 
    {
	final StringSet model = new StringSet();
	final JList list = new JList(model);
	final JScrollPane scrollPane = new JScrollPane(list);

	setLayout(new FlowLayout(FlowLayout.LEFT, 6, 6)); 
	add(scrollPane);

	list.setPrototypeCellValue("I am a new Cell 1234");

	JButton addTenCells = new JButton("Add 5 Cells");
	addTenCells.addActionListener(
	    new ActionListener() {
	        public void actionPerformed(ActionEvent e) {
		    int n = model.getSize();
		    for(int i = 0; i < 5; i++) {
			model.add("I am a new Cell " + (int)(i + n));
		    }
		    model.done();
                }
	    }
	);

	JButton removeAllCells = new JButton("Remove All Cells");
	removeAllCells.addActionListener(
	    new ActionListener() {
	        public void actionPerformed(ActionEvent e) {
		    model.removeAll();
                }
	    }
	);

	Box box = Box.createHorizontalBox();
	box.add(addTenCells);
	box.add(removeAllCells);
	add(box);
    }
}


class GiantListTest extends Container
{
    GiantListTest() {
	ListModel bigData = new AbstractListModel() {
	    public int getSize() { return Short.MAX_VALUE; }
	    public Object getElementAt(int index) { return "Index " + index; }
	};

	JList list = new JList(bigData);
	list.setPrototypeCellValue("Index 1234567");
	JScrollPane scrollPane = new JScrollPane(list);
	scrollPane.getViewport().setBackingStoreEnabled(true);
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
    }
}


class ConfigureTest extends Container
{
    ConfigureTest() {
	ListModel bigData = new AbstractListModel() {
	    public int getSize() { return Short.MAX_VALUE; }
	    public Object getElementAt(int index) { return "Index " + index; }
	};

	JList list = new JList(bigData);
	list.setForeground(Color.red);
        list.setBackground(Color.lightGray);
	UIManager.put("List.selectedCellForeground", Color.yellow);
	list.setPrototypeCellValue("Index 1234567");
	JScrollPane scrollPane = new JScrollPane(list);
	scrollPane.getViewport().setBackingStoreEnabled(true);
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
    }
}


class ClassExampleTest extends Container
{
    ClassExampleTest() {
	Vector superClasses = new Vector();
	Class rootClass = javax.swing.JList.class;
	for(Class cls = rootClass; cls != null; cls = cls.getSuperclass()) {
	    superClasses.addElement(cls);
	}
	JList list = new JList(superClasses);
	setLayout(new BorderLayout());
	add(new JScrollPane(list), BorderLayout.CENTER);
    }
}


class DoubleClickTest extends Container
{
    DoubleClickTest() {
	ListModel dataModel = new AbstractListModel() {
	    public int getSize() { return 1024; }
	    public Object getElementAt(int index) { return "Index " + index; }
	};

	final JList list = new JList(dataModel);
	list.setPrototypeCellValue("Index 1234567");

	MouseListener mouseListener = new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		if (e.getClickCount() == 2) {
		    int index = list.locationToIndex(e.getPoint());
		    System.out.println("Double clicked on Item " + index);
		}
	    }
	};
	list.addMouseListener(mouseListener);

	JScrollPane scrollPane = new JScrollPane(list);
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
    }
}


class SelectionListenerTest extends Container
{
    SelectionListenerTest() {
	ListModel dataModel = new AbstractListModel() {
	    public int getSize() { return 1024; }
	    public Object getElementAt(int index) { return "Index " + index; }
	};

	final JList list = new JList(dataModel);
	list.setPrototypeCellValue("Index 1234567");

	ListSelectionListener selectionListener = new ListSelectionListener() {
	    public void valueChanged(ListSelectionEvent e) {
		System.out.println(e);
	    }
	};
	list.addListSelectionListener(selectionListener);

	JScrollPane scrollPane = new JScrollPane(list);
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
    }
}


class TestListUI extends BasicListUI
{
    protected void configureList() 
    {
	list.setLayout(null);

	if ( ! (list.getCellRenderer() instanceof TestCellRenderer) ) {
	    list.setCellRenderer( new TestCellRenderer() );
	}

	list.setForeground(Color.red);
	list.setBackground(Color.black);
	rendererPane = new CellRendererPane();
	list.add(rendererPane);
    }

    public static ComponentUI createUI(JComponent list) {
	return new TestListUI();
    }
}


class TestCellRenderer extends BasicListCellRenderer
{
    public Component getListCellRendererComponent(
        JList list, 
	Object value,
        int index, 
        boolean isSelected, 
        boolean cellHasFocus)
    {
	if (isSelected) {
	    setBackground(Color.red);
	    setForeground(Color.white);
	}
	else {
	    setBackground(Color.white );
	    setForeground(Color.red);
	}
	
	if (value instanceof Icon) {
	    setIcon((Icon)value);
	}
	else {
	    setText((value == null) ? "" : value.toString());
	}

	return this;
    }
}


class ListUITest extends Container
{
    ListUITest() {
	ListModel dataModel = new AbstractListModel() {
	    public int getSize() { return 1024; }
	    public Object getElementAt(int index) { return "Index " + index; }
	};

	JList list = new JList(dataModel);
	list.setPrototypeCellValue("Index 1234567");
	list.setUI(new TestListUI());

	JScrollPane scrollPane = new JScrollPane(list);
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
    }
}


class CompoundCellRenderer extends Box implements ListCellRenderer
{
    JLabel topLabel = new JLabel();
    JLabel bottomLabel = new JLabel();

    public CompoundCellRenderer() {
	super(BoxLayout.Y_AXIS);
	topLabel.setOpaque(true);
	bottomLabel.setOpaque(true);
	add(topLabel);
    }

    public boolean isShowing() { 
	return true;
    }

    public Component getListCellRendererComponent(
        JList list,
        Object value,
        int index,
        boolean isSelected,
        boolean cellHasFocus)
    {
	Color fg = (isSelected) ? Color.white : list.getForeground();
	Color bg = (isSelected) ? Color.blue : list.getBackground();

	topLabel.setForeground(fg);
	topLabel.setBackground(bg);
	topLabel.setText("TOP, index is " + index);

	/* It's essntial to invalidate the Box when the list of
	 * children changes - it caches information about the children
	 * and the cache isn't automatically invalidated when this
	 * happens.
	 */

	if ((index % 3) == 0) {
	    bottomLabel.setForeground(fg);
	    bottomLabel.setBackground(bg);
	    bottomLabel.setText("Bottom, index is " + index);
	    add(bottomLabel);
	    invalidate();
	}
	else if (bottomLabel.getParent() == this) {
	    remove(bottomLabel);
	    invalidate();
	}

	return this;
    }

}


class CompoundCellTest extends Container
{
    CompoundCellTest() {
	JList list = new JList(new Object[128]);
	list.setCellRenderer(new CompoundCellRenderer());
	setLayout(new BorderLayout());
	add(new JScrollPane(list), BorderLayout.CENTER);
    }
}



/*

Test case for a CompoundCell renderering bug reported by 
Dave Brown.

From: brown@weblogic.com (Dave Brown)
To: advanced-java@xcf.berkeley.edu
Subject: JList missing cells in Swing 0.6.1
Date: Sun, 07 Dec 1997 23:39:46 -0800

I had been using a custom ListCellRenderer with JList
to display arbitrary Components of fixed height in the
JList (here, the components are JPanel with three JLabel's
in a grid layout).  This worked perfectly
in swing 0.5.1.  In 0.6.1, only the first cell is
visible, and the color-change/selection mechanism
doesn't work.  The printf's indicate the other 
components are "there" but just not getting painted.
I've fiddled with it to no avail.  So either I'm
missing something fundamental with the JList API
or something regressed between 0.5->0.6. In case
of the former, can anyone see anything wrong?  Simple
test case attached: the list should show three
rows, but only shows one.  This is NT 4.0, Solaris 2.5.1,
and JDK 1.1.4, and of course swing 0.6.1.

*/

class BrownListCell extends JPanel 
{
    String name;

    public String toString() {
	return "name="+name+super.toString();
    }

    public BrownListCell(String name) {
	super();
	setLayout(new GridLayout(1,3,5,5));
	this.name = name;
	JLabel l = new JLabel(name+",column1");
	l.setOpaque(true);
	add(l);
	l = new JLabel(name+",column2");
	l.setOpaque(true);
	add(l);
	l = new JLabel(name+",column3");
	l.setOpaque(true);
	add(l);
    }
}


class BrownCellRenderer implements ListCellRenderer 
{
    Hashtable comps = new Hashtable(10);

    void addCell(Object key, Component val) {
	comps.put(key, val);
    }
    
    public Component getListCellRendererComponent(JList l, Object k, int index,
						  boolean isSelected, boolean hasFocus)
    {
	Container c;
	if ((c = (Container)comps.get(k)) == null) {
	    return null;
	}
	for(int i = 0; i < c.getComponentCount(); i++) {
	    Component child = c.getComponent(i);
	    if (isSelected) {
		child.setForeground(Color.red);
		child.setBackground(Color.blue);
	    } else {
		child.setForeground(Color.black);
		child.setBackground(Color.white);
	    }
	}
	return c;
    }
}


class BrownCompoundCellTest extends Container
{
    static String[] data = {
	"Elem A", "Elem B", "Elem C"
    };

    BrownCompoundCellTest() 
    {
	JList list = new JList();
	BrownCellRenderer rend = new BrownCellRenderer();

	for (int i = 0; i < data.length; i++) {
	    /*JPanel*/Container lab = new BrownListCell(data[i]);
	    // lab.setOpaque(false);
	    rend.addCell(data[i], lab);
	}
	list.setBackground(Color.lightGray);
	list.setCellRenderer(rend);
	list.setListData(data);
	// -- Not needed, plus it interferes with layout
	// list.setFixedCellHeight(BrownListCell.H); 
	setLayout(new BorderLayout());
	add(list, BorderLayout.CENTER);
    }
}


/**
 * Try out the selectionForeground, selectionBackground, font,
 * foreground, and background properties.
 */
class TechnicolorListTest extends Container
{
    TechnicolorListTest() {
	ListModel bigData = new AbstractListModel() {
	    public int getSize() { return Short.MAX_VALUE; }
	    public Object getElementAt(int index) { return "Index " + index; }
	};

	JList list = new JList(bigData);
	list.setForeground(Color.yellow);
	list.setBackground(Color.black);
	list.setSelectionForeground(Color.black);
	list.setSelectionBackground(Color.magenta);
	list.setFont(new Font("Helvetica", Font.PLAIN, 18));
	list.setPrototypeCellValue("Index 1234567");
	JScrollPane scrollPane = new JScrollPane(list);
	scrollPane.getViewport().setBackingStoreEnabled(true);
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
    }
}


/**
 * Verify that list.setBorder works.  Big red border.
 */
class ListBorderTest extends Container
{
    ListBorderTest() {
	ListModel bigData = new AbstractListModel() {
	    public int getSize() { return 250; }
	    public Object getElementAt(int index) { return "Index " + index; }
	};

	JList list = new JList(bigData);
	list.setBorder(new LineBorder(Color.red, 3));
	JScrollPane scrollPane = new JScrollPane(list);
	scrollPane.getViewport().setBackingStoreEnabled(true);
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
    }
}




public class ListTest
{
    public static void main(String[] args)
    {
	WindowListener l = new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {System.exit(0);}
	};

	JFrame f = new JFrame("JList");
	f.addWindowListener(l); 
	f.getContentPane().add(new BrownCompoundCellTest());
	f.pack();
	f.show(); 
    }
}

