/*
 * @(#)SerializationTest.java	1.4 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/** 
 * A weak test of Serialization.
 * 
 * To test against JDK1.1:
 * 
 * setenv JAVA_HOME /usr/local/java/jdk1.1.5/solaris
 * setenv SWING_HOME /home/hmuller/ws/jdk12
 * setenv CLASSPATH ${SWING_HOME}/build/solaris/1.1_classes:${JAVA_HOME}/lib/classes.zip
 * ${JAVA_HOME}/bin/javac SerializationTest.java
 * ${JAVA_HOME}/bin/java -classpath .:${CLASSPATH} SerializationTest
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;

public class SerializationTest extends JPanel
{
    public SerializationTest() { 	
	setLayout(new BorderLayout());
	// JPanel
	JPanel p = new JPanel();
	p.setLayout(new BorderLayout());
	p = (JPanel)testSerialization(p);
	add(p, BorderLayout.CENTER);

	// JToolBar
	JToolBar tb = new JToolBar();
	tb = (JToolBar)testSerialization(tb);
	p.add(tb, BorderLayout.NORTH);
	
	Container p1 = new JPanel();
	// p1.add(testSerialization(new JTextField("TextField")));
	// p1.add(testSerialization(new JPasswordField("Password")));
	// p1.add(testSerialization(new JTextArea("TextArea")));
	p1.add(testSerialization(new JSlider()));
	p1.add(testSerialization(new JProgressBar()));
	p1.add(testSerialization(new JButton("I am Button")));
	p1.add(testSerialization(new JToggleButton("I am Toggle")));
	p1.add(testSerialization(new JCheckBox()));
	p1.add(testSerialization(new JRadioButton()));
	p.add(p1, BorderLayout.CENTER);
	
	
	// JList

	DefaultListModel model = new DefaultListModel();
	for(int i = 0; i < 25; i++) {
	    model.addElement("DefaultListModel row " + i);
	}
	JList list = new JList(model);
	JScrollPane scrollPane = new JScrollPane(list);

	p.add(testSerialization(scrollPane), BorderLayout.SOUTH);

	// JComboBox

        JComboBox comboBox = new JComboBox();
        comboBox.addItem("One");
        comboBox.addItem("Two");
        comboBox.addItem("Free");
        comboBox.addItem("Four");
        comboBox.addItem("Five");
        comboBox.addItem("Six");
        comboBox.addItem("Seven");
        comboBox.addItem("Height");
	p1.add(testSerialization(comboBox));


	// JButton
	JButton b = new JButton(new ImageIcon("images/new.gif"));
	b.setFocusPainted(false);
	b.setMargin(new Insets(0,0,0,0));
	b = (JButton)testSerialization(b);
	tb.add(b);

	tb.addSeparator();

	b = new JButton(new ImageIcon("images/save.gif"));
	b.setFocusPainted(false);
	b.setMargin(new Insets(0,0,0,0));
	b = (JButton)testSerialization(b);
	tb.add(b);

	// JMenuBar
	JMenuBar mb = new JMenuBar();
	mb = (JMenuBar)testSerialization(mb);
	add("North", mb);

	// JMenu
	JMenu m = new JMenu("File");
	m = (JMenu)testSerialization(m);
	mb.add(m);

	// JMenuItem
	JMenuItem mi = new JMenuItem("Open");
	mi = (JMenuItem)testSerialization(mi);
	m.add(mi);

	// JCheckBoxMenuItem
	JCheckBoxMenuItem cmi = new JCheckBoxMenuItem("Save");
	cmi = (JCheckBoxMenuItem)testSerialization(cmi);
	m.add(cmi);	

	// JRadioButtonMenuItem
	JRadioButtonMenuItem rmi = new JRadioButtonMenuItem("One");
	rmi = (JRadioButtonMenuItem)testSerialization(rmi);
	m.add(rmi);

	// JColorChooser
        JColorChooser cc = new JColorChooser();
        cc = (JColorChooser)testSerialization(cc);
        JDialog d = JColorChooser.createDialog(m,
                                                "Color Chooser", false,
                                                cc, null, null);
        d.show();
    }

    public Component testSerialization(Object o) {
	try {
	    String file = "Test.ser";
	    ObjectOutputStream os = new ObjectOutputStream(new FileOutputStream(file));
	    System.out.println("Writing " + o.getClass().getName() + " to " + os);
	    os.writeObject(o);
	    os.close();
	    
	    ObjectInputStream is = new ObjectInputStream(new FileInputStream(file));
	    System.out.println("Reading from " + is);
	    o = is.readObject();
	    is.close();
	}
	catch (Throwable e) {
	    e.printStackTrace();
	}
	return (Component)o;
    }

    public static void setMotif(JFrame f) {
	try {
	    UIManager.setLookAndFeel("javax.swing.motif.MotifLookAndFeel");
	    updateComponentTree(f);
	} 
	catch (Exception e) {
	    System.out.println(e);
	}
    }

    public static void setBasic(JFrame f) {
	try {
	    UIManager.setLookAndFeel("javax.swing.basic.BasicLookAndFeel");
	    updateComponentTree(f);
	} 
	catch (Exception e) {
	    System.out.println(e);
	}
    }

    /** 
     * A simple minded look and feel change: ask each node in the tree
     * to updateUI(), i.e. to initialize its UI property with the
     * current look and feel.
     */
    static void updateComponentTree(Component c) {
	if (c instanceof JComponent) {
	    ((JComponent) c).updateUI();
	}
	if (c instanceof Container) {
	    Component[] children = ((Container)c).getComponents();
	    for(int i = 0; i < children.length; i++) {
		updateComponentTree(children[i]);
	    }
	}
    }

    public static void main(String s[]) {
         WindowListener l = new WindowAdapter() {
             public void windowClosing(WindowEvent e) {System.exit(0);}
             public void windowClosed(WindowEvent e) {System.exit(0);}
         };
 
         JFrame f = new JFrame();
	 // setMotif(f);
         f.addWindowListener(l);
	 f.getContentPane().add("Center", new SerializationTest());
         f.pack();
         f.show();
    }

}

