/*
 * @(#)SerializationTest2.java	1.3 98/09/21
 *
 * Copyright 1998-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;

/** 
 * A non-interactive test of Swing's serialization, based on Hans'
 * excellent (if weak) SerializationTest.  The component list should
 * match the list of classes marked as Java Beans in the swing.jar
 * file.
 *
 * Run with 1.1.5 or later JDK with all the Swing jar files in your
 * classpath.
 *
 * @author Thomas Ball
 * @version 1.3 09/21/98
 */
public class SerializationTest2
{
    static int errors = 0;

    public static void main (String[] args) {
        UIManager.LookAndFeelInfo[] lnfs = 
            UIManager.getInstalledLookAndFeels();
        for (int i = 0; i < lnfs.length; i++) {
            System.out.println("will test " + lnfs[i].getName());
        }
        System.out.println();
        for (int i = 0; i < lnfs.length; i++) {
            System.out.println("\nTesting Look&Feel: " + lnfs[i].getName());
            testLnf(lnfs[i].getClassName());
        }
        if (errors == 1) {
            System.out.println("\nThere was 1 error found.");
        } else {
            System.out.println("\nThere were " + errors + " errors found.");
        }
        System.exit(errors);
    }

    public static void testLnf(String lnfName) {
        try {
            UIManager.setLookAndFeel(lnfName);
        } catch (Throwable t) {
            errors++;
            t.printStackTrace(System.out);
            return;
        }

	testSerialization(JApplet.class);
	testSerialization(JButton.class);
	testSerialization(JCheckBox.class);
	testSerialization(JCheckBoxMenuItem.class);
	testSerialization(JComboBox.class);
	testSerialization(JDialog.class);
	testSerialization(JEditorPane.class);
	testSerialization(JFrame.class);
	testSerialization(JInternalFrame.class);
	testSerialization(JLabel.class);
	testSerialization(JList.class);
	testSerialization(JMenu.class);
	testSerialization(JMenuBar.class);
	testSerialization(JMenuItem.class);
	testSerialization(JOptionPane.class);
	testSerialization(JPanel.class);
	testSerialization(JPasswordField.class);
	testSerialization(JPopupMenu.class);
	testSerialization(JProgressBar.class);
	testSerialization(JRadioButton.class);
	testSerialization(JRadioButtonMenuItem.class);
	testSerialization(JScrollBar.class);
	testSerialization(JScrollPane.class);
	testSerialization(JSeparator.class);
	testSerialization(JSlider.class);
	testSerialization(JSplitPane.class);
	testSerialization(JTabbedPane.class);
	testSerialization(JTable.class);
	testSerialization(JTextArea.class);
	testSerialization(JTextField.class);
	testSerialization(JTextPane.class);
	testSerialization(JToggleButton.class);
	testSerialization(JToolBar.class);
	testSerialization(JTree.class);
	testSerialization(JWindow.class);
    }

    public static void testSerialization(Class c) {
        String className = c.getName();
	try {
            System.out.println("\nTesting " + className);
            Object o = c.newInstance();
	    String file = "Test.ser";
	    ObjectOutputStream os = 
                new ObjectOutputStream(new FileOutputStream(file));
	    System.out.println("Writing " + className);
	    os.writeObject(o);
	    os.close();
	    
	    ObjectInputStream is = 
                new ObjectInputStream(new FileInputStream(file));
	    System.out.println("Reading " + className);
	    o = is.readObject();
	    is.close();
            System.out.println(className + " passed.");
	}
	catch (Throwable e) {
            errors++;
	    e.printStackTrace(System.out);
            System.out.println(className + " failed.");
	}
    }
}

