/*
 * @(#)X11GraphicsDevice.java	1.7 98/07/30
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import java.awt.GraphicsDevice;
import java.awt.GraphicsConfiguration;

/**
 * This is an implementation of a GraphicsDevice object for a single
 * X11 screen.
 *
 * @see GraphicsEnvironment
 * @see GraphicsConfiguration
 * @version 10 Feb 1997
 */
public class X11GraphicsDevice extends GraphicsDevice {
    int screen;

    public X11GraphicsDevice(int screennum) {
	this.screen = screennum;
    }

    /**
     * Returns the X11 screen of the device.
     */
    public int getScreen() {
        return screen;
    }

    /**
     * Returns the X11 Display of this device.
     * This method is also in MDrawingSurfaceInfo but need it here
     * to be able to allow a GraphicsConfigTemplate to get the Display.
     */
    public native int getDisplay();

    /**
     * Returns the type of the graphics device.
     * @see #TYPE_RASTER_SCREEN
     * @see #TYPE_PRINTER
     * @see #TYPE_IMAGE_BUFFER
     */
    public int getType() {
	return TYPE_RASTER_SCREEN;
    }

    /**
     * Returns the identification string associated with this graphics
     * device.
     */
    public String getIDstring() {
	return ":0."+screen;
    }


    GraphicsConfiguration[] configs;
    GraphicsConfiguration defaultConfig;

    /**
     * Returns all of the graphics
     * configurations associated with this graphics device.
     */
    public GraphicsConfiguration[] getConfigurations() {
	GraphicsConfiguration[] ret = configs;
	if (ret == null) {
	    int i = 0;
	    int num = getNumConfigs();
	    ret = new GraphicsConfiguration[num];
	    if (defaultConfig != null) {
	        ret [0] = defaultConfig;
		i++;
	    }
	    for ( ; i < num; i++) {
		int visNum = getConfigVisualId (i);
		ret[i] = X11GraphicsConfig.getConfig(this, visNum);
	    }
	    configs = ret;
	}
	return ret;
    }

    /*
     * Returns the number of X11 visuals representable as an
     * X11GraphicsConfig object.
     */
    private native int getNumConfigs();

    /*
     * Returns the visualid for the given index of graphics configurations.
     */
    private native int getConfigVisualId (int index);

    /**
     * Returns the default graphics configuration
     * associated with this graphics device.
     */
    public GraphicsConfiguration getDefaultConfiguration() {
	if (defaultConfig == null) {
	   int visNum = getConfigVisualId (0);
	   //defaultConfig = new GraphicsConfiguration [1];
	   defaultConfig = X11GraphicsConfig.getConfig (this, visNum);
	}

	return defaultConfig;
    }


    public String toString() {
	return ("X11GraphicsDevice[screen="+screen+"]");
    }

}
