/*
 * @(#)MButtonPeer.java	1.24 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.ActionEvent;

class MButtonPeer extends MComponentPeer implements ButtonPeer {
    native void create(MComponentPeer peer);
    public native void setLabel(String label);
    
    MButtonPeer(Button target) {
	super(target);
    }

    public Dimension getMinimumSize() {
	FontMetrics fm = getFontMetrics(target.getFont());
	return new Dimension(fm.stringWidth(((Button)target).getLabel()) + 14, 
			     fm.getHeight() + 8);
    }

    public boolean isFocusTraversable() {
	return true;
    }

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void action() {
	MToolkit.executeOnEventHandlerThread(target, new Runnable() {
	    public void run() {
                postEvent(new ActionEvent(target, ActionEvent.ACTION_PERFORMED,
                                   ((Button)target).getActionCommand()));
	    }
	});
    }

    /*
     * Print the native component by rendering the Motif look ourselves.
     * ToDo(aim): needs to query native motif for more accurate size and
     * color information.
     */
    public void print(Graphics g) {
        Button b = (Button)target;
	Dimension d = b.size();
	Color bg = b.getBackground();
	Color fg = b.getForeground();

	g.setColor(bg);
	g.fillRect(2, 2, d.width - 3, d.height - 3);
	draw3DRect(g, bg, 1, 1, d.width - 2, d.height - 2, true);

	g.setColor(fg);
	g.setFont(b.getFont());
	FontMetrics fm = g.getFontMetrics();
	String lbl = b.getLabel();
	g.drawString(lbl, (d.width - fm.stringWidth(lbl)) / 2,
		          (d.height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);

	target.print(g);
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

}
