/*
 * @(#)MCheckboxPeer.java	1.19 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.ItemEvent;

public class MCheckboxPeer extends MComponentPeer implements CheckboxPeer {
    native void create(MComponentPeer parent);
    public native void setLabel(String label);

    void initialize() {
	Checkbox t = (Checkbox)target;

	setState(t.getState());
	setCheckboxGroup(t.getCheckboxGroup());
	super.initialize();
    }

    public MCheckboxPeer(Checkbox target) {
	super(target);
    }

    public boolean isFocusTraversable() {
	return true;
    }

    public native void setState(boolean state);
    public native void setCheckboxGroup(CheckboxGroup g);

    public Dimension getMinimumSize() {
	String lbl = ((Checkbox)target).getLabel();
	if (lbl != null) {
	    FontMetrics fm = getFontMetrics(target.getFont());
	    return new Dimension(30 + fm.stringWidth(lbl), 
				 Math.max(fm.getHeight() + 8, 15));
	}
	return new Dimension(20, 15);
    }

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void action(boolean state) {
	final Checkbox cb = (Checkbox)target;
	final boolean newState = state;
	MToolkit.executeOnEventHandlerThread(cb, new Runnable() {
	    public void run() {
                CheckboxGroup cbg = cb.getCheckboxGroup();
                /* Bugid 4039594. If this is the current Checkbox in 
		 * a CheckboxGroup, then return to prevent deselection. 
		 * Otherwise, it's logical state will be turned off, 
		 * but it will appear on.
                 */
                if ((cbg != null) && (cbg.getSelectedCheckbox() == cb) &&
                    cb.getState()) {
                  cb.setState(true);
                  return;
                }
		// All clear - set the new state
	        cb.setState(newState);
                postEvent(new ItemEvent(cb, ItemEvent.ITEM_STATE_CHANGED,
                                cb.getLabel(),
                                newState? ItemEvent.SELECTED : ItemEvent.DESELECTED));
	    } // run()
        });
    } // action()

    static final int SIZE = 19;
    static final int BORDER = 4;
    static final int SIZ = SIZE - BORDER*2 - 1;

    /*
     * Print the native component by rendering the Motif look ourselves.
     * ToDo(aim): needs to query native motif for more accurate size and
     * color information; need to render check mark.
     */
    public void print(Graphics g) {
        Checkbox cb = (Checkbox)target;
	Dimension d = cb.size();
	Color bg = cb.getBackground();
	Color fg = cb.getForeground();
	Color shadow = bg.darker();
	int x = BORDER;
	int y = ((d.height - SIZE) / 2) + BORDER;

	g.setColor(cb.getState()? shadow : bg);

	if (cb.getCheckboxGroup() != null) {
	    g.fillOval(x, y, SIZ, SIZ);
	    draw3DOval(g, bg, x, y, SIZ, SIZ, !(cb.getState()));
	    if (cb.getState()) {
		g.setColor(fg);
		g.fillOval(x + 3, y + 3, SIZ - 6, SIZ - 6);
	    }
	} else {
	    g.fillRect(x, y, SIZ, SIZ);
	    draw3DRect(g, bg, x, y, SIZ, SIZ, !(cb.getState()));
	    if (cb.getState()) {
		g.setColor(fg);
		g.drawLine(x+1, y+1, x+SIZ-1, y+SIZ-1);
		g.drawLine(x+1, y+SIZ-1, x+SIZ-1, y+1);
	    }
	}
	g.setColor(fg);
	String lbl = cb.getLabel();
	if (lbl != null) {
	    // REMIND: align
	    g.setFont(cb.getFont());
	    FontMetrics fm = g.getFontMetrics();
	    g.drawString(lbl, SIZE,
			 (d.height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
	}

	target.print(g);
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	    return getMinimumSize();
    }

}
