/*
 * @(#)MChoicePeer.java	1.34 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.ItemEvent;

class MChoicePeer extends MComponentPeer implements ChoicePeer {
    native void create(MComponentPeer parent);
    native void pReshape(int x, int y, int width, int height);
    public native void select(int index);

    void initialize() {
	Choice opt = (Choice)target;
	int itemCount = opt.countItems();
	for (int i=0; i < itemCount; i++) {
	    add(opt.getItem(i), i);
	}
	if (itemCount > 0) {
	    select(opt.getSelectedIndex());
	}
	super.initialize();
    }

    public MChoicePeer(Choice target) {
	super(target);
    }

    public boolean isFocusTraversable() {
	return true;
    }

    public Dimension getMinimumSize() {
	FontMetrics fm = getFontMetrics(target.getFont());
	Choice c = (Choice)target;
	int w = 0;
	for (int i = c.countItems() ; i-- > 0 ;) {
	    w = Math.max(fm.stringWidth(c.getItem(i)), w);
	}
	return new Dimension(32 + w, Math.max(fm.getHeight() + 8, 15) + 5);
    }

    public native void setFont(Font f);

    public void add(String item, int index) {
      	addItem(item, index);
    }

    public native void remove(int index);

    public void removeAll() {}

    /**
     * DEPRECATED, but for now, called by add(String, int).
     */
    public native void addItem(String item, int index);

    // public native void remove(int index);

    public native void setBackground(Color c);

    public native void setForeground(Color c);

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void action(final int index) {
        final Choice c = (Choice)target;
        MToolkit.executeOnEventHandlerThread(c, new Runnable() {
            public void run() {
                c.select(index);
                postEvent(new ItemEvent(c, ItemEvent.ITEM_STATE_CHANGED,
                                c.getItem(index), ItemEvent.SELECTED));
            }
        });
    }

    /*
     * Print the native component by rendering the Motif look ourselves.
     * ToDo(aim): needs to query native motif for more accurate size and
     * color information.
     */
    public void print(Graphics g) {
	Choice ch = (Choice)target;
	Dimension d = ch.size();
	Color bg = ch.getBackground();
	Color fg = ch.getForeground();

	g.setColor(bg);
	g.fillRect(2, 2, d.width-1, d.height-1);
	draw3DRect(g, bg, 1, 1, d.width-2, d.height-2, true);
	draw3DRect(g, bg, d.width - 18, (d.height / 2) - 3, 10, 6, true);

	g.setColor(fg);
	g.setFont(ch.getFont());
	FontMetrics fm = g.getFontMetrics();
	String lbl = ch.getSelectedItem();
	g.drawString(lbl, 5, (d.height + fm.getMaxAscent()-fm.getMaxDescent())/2);

	target.print(g);
    }	

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	    return getMinimumSize();
    }

}
