/*
 * @(#)MComponentPeer.java	1.113 98/09/09
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.KeyEvent;
import java.awt.event.PaintEvent;
import sun.awt.*;
import sun.awt.image.ImageRepresentation;
import sun.awt.image.OffScreenImage;
import java.awt.image.ImageProducer;
import java.awt.image.ImageObserver;
import java.awt.image.ColorModel;

import java.awt.image.Raster;

import java.awt.dnd.DropTarget;
import java.awt.dnd.peer.DropTargetPeer;

public /* REMIND: should not be public */
abstract class MComponentPeer implements ComponentPeer, UpdateClient, DropTargetPeer {
    Component	target;
    long 	pData;
    long        jniGlobalRef;
    X11GraphicsConfig	graphicsConfig;

    int		cursorSet;

    private boolean disposed = false;

    static {
        initIDs();
    }

    /* initialize the fieldIDs of fields that may be accessed from C */
    private native static void initIDs();

    abstract void create(MComponentPeer parent);
    void create(MComponentPeer parent, Object arg) {
	create(parent);
    }

    void EFcreate(MComponentPeer parent, Object arg, int x){}
    void EF1create(MComponentPeer parent, int x){}

    native void pInitialize();
    native void pShow();
    native void pHide();
    native void pEnable();
    native void pDisable();
    native void pReshape(int x, int y, int width, int height);
    native void pAddRepaint(int x, int y, int width, int height);
    native void pTriggerRepaint();
    native void pDispose();
    native void pMakeCursorVisible();
    native Point pGetLocationOnScreen();

    // Returns the parent of the component, without invoking client
    // code. This must go through native code, because it invokes
    // private methods in the java.awt package, which we cannot
    // do from this package.
    static native Container getParent_NoClientCode(Component component);

    // Returns the parent of the component, without invoking client
    // code. This must go through native code, because it invokes
    // private methods in the java.awt package, which we cannot
    // do from this package.
    static native Component[] getComponents_NoClientCode(Container container);

    void initialize() {
	if (!target.isVisible()) {
	    hide();
	}
	Color c;
	Font  f;
 	Cursor cursor;

	pInitialize();

	if ((c = target.getForeground()) != null) {
	    setForeground(c);
	}
	if ((c = target.getBackground()) != null) {
	    setBackground(c);
	}
	if ((f = target.getFont()) != null) {
	    setFont(f);
	}
	if ((cursor = target.getCursor()) == null) {
	    cursor = Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);
	}
	setCursor(cursor);
	if (cursor.getType() == Cursor.DEFAULT_CURSOR) {
	    cursorSet = 1;	    
        }
	if (!target.isEnabled()) {
	    disable();
	}
	Rectangle r = target.getBounds();
	reshape(r.x, r.y, r.width, r.height);
	if (target.isVisible()) {
	    show();
	}
    }

    public void init(Component target, Object arg) {
	this.target = target;
	Container parent = MToolkit.getNativeContainer(target);
	graphicsConfig = MToolkit.config;
	MComponentPeer parentPeer = (MComponentPeer) MToolkit.targetToPeer(parent);
        if( target instanceof MEmbeddedFrame){
            MEmbeddedFrame tempEF = (MEmbeddedFrame)target;
            EFcreate(parentPeer, arg, (int)tempEF.handle);
        }
        else{
	    create(parentPeer, arg);
	}
	initialize();
    }

    MComponentPeer(Component target, Object arg) {
        init(target, arg);
    }

    MComponentPeer() {}

    public void init(Component target) {
	this.target = target;
	Container parent = MToolkit.getNativeContainer(target);
	graphicsConfig = MToolkit.config;
	MComponentPeer parentPeer = (MComponentPeer) MToolkit.targetToPeer(parent);
        if( target instanceof MEmbeddedFrame){
             MEmbeddedFrame tempEF = (MEmbeddedFrame)target;
             EF1create(parentPeer, (int)tempEF.handle);
        }
        else{
  	    create(parentPeer);
	}
 	if (parent != null && parent instanceof ScrollPane) {
	    MScrollPanePeer speer = (MScrollPanePeer) parentPeer;
	    speer.setScrollChild(this);
        }
	initialize();
    }

    MComponentPeer(Component target) {
        init(target);
    }

    protected void finalize() throws Throwable {
        dispose();
	super.finalize();
    }

    public void setForeground(Color c) {
       pSetForeground(c);
    }

    public void setBackground(Color c) {
       pSetBackground(c);
    }

    public native void pSetForeground(Color c);
    public native void pSetBackground(Color c);
    public native void setFont(Font f);
    public native void setCursor(Cursor c);
    public native boolean isBackgroundInherited();
    public native boolean isForegroundInherited();
    public native void pSetScrollbarBackground(Color c);
    public native void pSetInnerForeground(Color c);


    public boolean isFocusTraversable() {
	return false;
    }

    public ColorModel getColorModel() {
	return graphicsConfig.getColorModel ();
    }

    public ColorModel getColorModel(int transparency) {
	return graphicsConfig.getColorModel (transparency);
    }

    public int updatePriority() {
	return Thread.NORM_PRIORITY;
    }
    public void updateClient(Object arg) {
        if (target.getPeer() != null)
	    pTriggerRepaint();
    }
    public void repaint(long tm, int x, int y, int width, int height) {
	pAddRepaint(x, y, width, height);
	ScreenUpdater.updater.notify(this, tm);
    }
    public void paint(Graphics g) {
	g.setColor(target.getForeground());
	g.setFont(target.getFont());
        target.paint(g);
    }
    public void update(Graphics g) {
	g.setColor(target.getForeground());
	g.setFont(target.getFont());
        target.update(g);
    }
    public void print(Graphics g) {
	Dimension d = target.getSize();
	Color bg = target.getBackground();
	Color fg = target.getForeground();

	g.setColor(bg);
	g.fillRect(0, 0, d.width, d.height);
	g.setColor(fg);
	g.setFont(target.getFont());
	//g.drawRect(0, 0, d.width - 1, d.height - 1);
	target.print(g);
    }

    native void nativeHandleEvent(AWTEvent e);

    public void handleEvent(AWTEvent e) {
        int id = e.getID();

        switch(id) {
          case PaintEvent.PAINT:
          case PaintEvent.UPDATE:
            Rectangle r = ((PaintEvent)e).getUpdateRect();
            Graphics g = ((Component)target).getGraphics();
	    if (g == null) {
	        return;
	    }
	    g.clipRect(r.x, r.y, r.width, r.height);
            if (id == PaintEvent.PAINT) {
                target.paint(g);
            } else {
                target.update(g);
            }
            g.dispose();
        }

        // Call the native code
        nativeHandleEvent(e);
    }

    /* New API for 1.1 */
    public Dimension getMinimumSize() {
	return target.getSize();
    }

    /* New API for 1.1 */
    public Dimension getPreferredSize() {
	return getMinimumSize();
    }

    public java.awt.Toolkit getToolkit() {
	// XXX: bogus
	return Toolkit.getDefaultToolkit();
    }
    public synchronized Graphics getGraphics() {
        if (!disposed) {
	    Graphics g = new X11Graphics(this);
	    g.setColor(target.getForeground());
	    g.setFont(target.getFont());
	    return g;
	}

	return null;
    }
    public Image createImage(ImageProducer producer) {
	return new X11Image(producer);
    }
    public Image createImage(int width, int height) {
	ColorModel model = getColorModel(Transparency.OPAQUE);
	return new OffScreenImage(target, model,
			  model.createCompatibleWritableRaster(width, height),
			  model.isAlphaPremultiplied());
    }
    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return MToolkit.prepareScrImage(img, w, h, o);
    }
    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return MToolkit.checkScrImage(img, w, h, o);
    }
    public FontMetrics getFontMetrics(Font font) {
	return X11FontMetrics.getFontMetrics(font);
    }

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     */
    protected void disposeImpl() {
        ScreenUpdater.updater.removeClient(this);
	MToolkit.targetDisposedPeer(target, this);
	pDispose();
    }
    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
            synchronized (this) {
                if (!disposed) {
                    disposed = call_disposeImpl = true;
                }
            }
        }

        if (call_disposeImpl) {
            disposeImpl();
        }
    }
    public native void requestFocus();

    /*
     * Post an event to the event queue.
     */
    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void postEvent(AWTEvent event) {
        MToolkit.postEvent(MToolkit.targetToAppContext(target), event);
    }

    /* Callbacks for window-system events to the frame 
     *
     * NOTE: This method may be called by privileged threads.
     *       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    */
    void handleExpose(int x, int y, int w, int h) {
        postEvent(new PaintEvent(target, PaintEvent.PAINT, 
                                 new Rectangle(x, y, w, h)));		
    }

    /* Callbacks for window-system events to the frame 
     *
     * NOTE: This method may be called by privileged threads.
     *       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
     */
    void handleRepaint(int x, int y, int w, int h) {
        postEvent(new PaintEvent(target, PaintEvent.UPDATE, 
                                 new Rectangle(x, y, w, h)));
    }

    /* Return the component's z-order position relative to
     * other peer'd siblings (don't count lightweight siblings
     * or siblings who don't yet have valid peers).
     *
     * NOTE: This method may be called by privileged threads.
     *       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
     */
    public int getZOrderPosition_NoClientCode() {
	// SECURITY: use _NoClientCode() methods, because we may
	//           be running on a privileged thread
        Container p = getParent_NoClientCode(target);
        if (p != null) {
	    // SECURITY: use _NoClientCode() methods, because we may
	    //           be running on a privileged thread
            Component children[] = getComponents_NoClientCode(p);
            int i;
            int index = 0;
            for (i = 0; i < children.length; i++) {
                if (children[i] == target) {
                    return index;
                } else {
                    Object cpeer = MToolkit.targetToPeer(children[i]);
                    if (cpeer != null && 
                        !(cpeer instanceof java.awt.peer.LightweightPeer)) {
                        index++;
                    }
                }
            }
        }
        return -1;
    }

    /*
     * drawXXX() methods are used to print the native components by 
     * rendering the Motif look ourselves.
     * ToDo(aim): needs to query native motif for more accurate color
     * information.
     */
    void draw3DOval(Graphics g, Color bg,
		    int x, int y, int w, int h, boolean raised) {
	Color c = g.getColor();
	Color shadow = bg.darker();
	Color highlight = bg.brighter();

	g.setColor(raised ? highlight : shadow);
	g.drawArc(x, y, w, h, 45, 180);
	g.setColor(raised ? shadow : highlight);
	g.drawArc(x, y, w, h, 225, 180);
	g.setColor(c);
    }
    void draw3DRect(Graphics g, Color bg,
			   int x, int y, int width, int height,
			   boolean raised) {
	Color c = g.getColor();
	Color shadow = bg.darker();
	Color highlight = bg.brighter();

	g.setColor(raised ? highlight : shadow);
	g.drawLine(x, y, x, y + height);
	g.drawLine(x + 1, y, x + width - 1, y);
	g.setColor(raised ? shadow : highlight);
	g.drawLine(x + 1, y + height, x + width, y + height);
	g.drawLine(x + width, y, x + width, y + height - 1);
	g.setColor(c);
    }  
    void drawScrollbar(Graphics g, Color bg, int thickness, int length,
		       int min, int max, int val, int vis, boolean horizontal) {
        Color c = g.getColor();
	double f = (double)(length - 2*(thickness-1)) / Math.max(1, ((max - min) + vis));
	int v1 = thickness + (int)(f * (val - min));
	int v2 = (int)(f * vis);
	int w2 = thickness-4;
        int tpts_x[] = new int[3];
        int tpts_y[] = new int[3];

	if (length < 3*w2 ) {
	    v1 = v2 = 0;
	    if (length < 2*w2 + 2) {
		w2 = (length-2)/2;
	    }
	} else 	if (v2 < 7) {
	// enforce a minimum handle size
	    v1 = Math.max(0, v1 - ((7 - v2)>>1));
	    v2 = 7;
	}

	int ctr   = thickness/2;
	int sbmin = ctr - w2/2;
	int sbmax = ctr + w2/2;

	// paint the background slightly darker
	{
	    Color d = new Color((int) (bg.getRed()   * 0.85),
				(int) (bg.getGreen() * 0.85),
				(int) (bg.getBlue()  * 0.85));

	    g.setColor(d);
	    if (horizontal) {
	        g.fillRect(0, 0, length, thickness);
	    } else {
	        g.fillRect(0, 0, thickness, length);
	    }
	}

	// paint the thumb and arrows in the normal background color
	g.setColor(bg);
	if (v1 > 0) {
	    if (horizontal) {
	        g.fillRect(v1, 3, v2, thickness-3);
	    } else {
	        g.fillRect(3, v1, thickness-3, v2);
	    }
	}

	tpts_x[0] = ctr;	tpts_y[0] = 2;
	tpts_x[1] = sbmin;	tpts_y[1] = w2;
	tpts_x[2] = sbmax;	tpts_y[2] = w2;
	if (horizontal) {
	    g.fillPolygon(tpts_y, tpts_x, 3);
	} else {
	    g.fillPolygon(tpts_x, tpts_y, 3);
	}

	tpts_y[0] = length-2;
	tpts_y[1] = length-w2;
	tpts_y[2] = length-w2;
	if (horizontal) {
	    g.fillPolygon(tpts_y, tpts_x, 3);
	} else {
	    g.fillPolygon(tpts_x, tpts_y, 3);
	}

	Color highlight = bg.brighter();

	// // // // draw the "highlighted" edges
	g.setColor(highlight);

	// outline & arrows
	if (horizontal) {
	    g.drawLine(1, thickness, length - 1, thickness);
	    g.drawLine(length - 1, 1, length - 1, thickness);

	    // arrows
	    g.drawLine(1, ctr, w2, sbmin);
	    g.drawLine(length - w2, sbmin, length - w2, sbmax);
	    g.drawLine(length - w2, sbmin, length - 2, ctr);

	} else {
	    g.drawLine(thickness, 1, thickness, length - 1);
	    g.drawLine(1, length - 1, thickness, length - 1);

	    // arrows
	    g.drawLine(ctr, 1, sbmin, w2);
	    g.drawLine(sbmin, length - w2, sbmax, length - w2);
	    g.drawLine(sbmin, length - w2, ctr, length - 2);
	}

	// thumb
	if (v1 > 0) {
	    if (horizontal) {
	        g.drawLine(v1, 2, v1 + v2, 2);
		g.drawLine(v1, 2, v1, thickness-3);
	    } else {
	        g.drawLine(2, v1, 2, v1 + v2);
	        g.drawLine(2, v1, thickness-3, v1);
	    }
	}

	Color shadow = bg.darker();

	// // // // draw the "shadowed" edges
	g.setColor(shadow);

	// outline && arrows
	if (horizontal) {
	    g.drawLine(0, 0, 0, thickness);
	    g.drawLine(0, 0, length - 1, 0);

	    // arrows
	    g.drawLine(w2, sbmin, w2, sbmax);
	    g.drawLine(w2, sbmax, 1, ctr);
	    g.drawLine(length-2, ctr, length-w2, sbmax);

	} else {
	    g.drawLine(0, 0, thickness, 0);
	    g.drawLine(0, 0, 0, length - 1);
	
	    // arrows
	    g.drawLine(sbmin, w2, sbmax, w2);
	    g.drawLine(sbmax, w2, ctr, 1);
	    g.drawLine(ctr, length-2, sbmax, length-w2);
	}

	// thumb
	if (v1 > 0) {
	    if (horizontal) {
	        g.drawLine(v1 + v2, 2, v1 + v2, thickness-2);
		g.drawLine(v1, thickness-2, v1 + v2, thickness-2);
	    } else {
	        g.drawLine(2, v1 + v2, thickness-2, v1 + v2);
	        g.drawLine(thickness-2, v1, thickness-2, v1 + v2);
	    }
	}
	g.setColor(c);
    }

    public String toString() {
	return getClass().getName() + "[" + target + "]";
    }

    /* New 1.1 API */
    public void setVisible(boolean b) {
	if (b) {
    	    pShow();
    	} else {
    	    pHide();
    	}
    }

    /* New 1.1 API */
    public void setEnabled(boolean b) {
    	if (b) {
	    pEnable();
	} else {
	    pDisable();
	}
    }

    /* New 1.1 API */
    public Point getLocationOnScreen() {
	return pGetLocationOnScreen();
    }

    public int serialNum = 0;

    /* New 1.1 API */
    public void setBounds(int x, int y, int width, int height) {
	Container parent = getParent_NoClientCode(target);
	pReshape(x, y, width, height);

	// Note: it would be ideal to NOT execute this if it's
        // merely a Move which is occurring.
        if (parent != null && parent instanceof ScrollPane) {
	    MScrollPanePeer speer = (MScrollPanePeer)parent.getPeer();
	    speer.childResized(width, height);
	}
	serialNum++;
    }

    public void beginValidate() {}
    public void endValidate() {}

    /**
     * DEPRECATED:  Replaced by setVisible(boolean).
     */
    public void show() {
	setVisible(true);
    }

    /**
     * DEPRECATED:  Replaced by setVisible(boolean).
     */
    public void hide() {
	setVisible(false);
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public void enable() {
	setEnabled(true);
    }

    /**
     * DEPRECATED:  Replaced by setEnabled(boolean).
     */
    public void disable() {
	setEnabled(false);
    }

    /**
     * DEPRECATED:  Replaced by setBounds(int, int, int, int).
     */
    public void reshape(int x, int y, int width, int height) {
	setBounds(x, y, width, height);
    }

    /**
     * DEPRECATED:  Replaced by getMinimumSize().
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * DEPRECATED:  Replaced by getPreferredSize().
     */
    public Dimension preferredSize() {
	return getPreferredSize();
    }

    /**
     *
     */

    public void addDropTarget(DropTarget dt) {
	addNativeDropTarget(dt);
    }

    /**
     *
     */

    public void removeDropTarget(DropTarget dt) {
	removeNativeDropTarget(dt);
    }

    public void notifyTextComponentChange(boolean add){
	Container parent = getParent_NoClientCode(target);
	while(!(parent == null ||
		parent instanceof java.awt.Frame ||
		parent instanceof java.awt.Dialog)) {
	    parent = getParent_NoClientCode(parent);
	}

	if (parent instanceof java.awt.Frame ||
	    parent instanceof java.awt.Dialog) {
	    if (add)
	      ((MInputMethodControl)parent.getPeer()).addTextComponent((MComponentPeer)this);
	    else
	      ((MInputMethodControl)parent.getPeer()).removeTextComponent((MComponentPeer)this);
	}
    }

    native void addNativeDropTarget(DropTarget dt);

    native void removeNativeDropTarget(DropTarget dt);
}
