/*
 * @(#)MDialogPeer.java	1.37 98/08/31
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;
import java.awt.event.*;
import sun.awt.motif.X11InputMethod;
import sun.awt.motif.MInputMethodControl;
import sun.awt.im.*;

class MDialogPeer extends MPanelPeer implements DialogPeer, MInputMethodControl {
    Insets insets = new Insets(0, 0, 0, 0);

    static Vector allDialogs = new Vector();

    native void create(MComponentPeer parent, Object arg);
    native void pSetTitle(String title);
    native void pShow();
    native void pHide();
    native void pReshape(int x, int y, int width, int height);
    native void pDispose();
    public native void setResizable(boolean resizable);
    private static native void initIDs();

    native void addTextComponentNative();
    native void removeTextComponentNative();

    public native boolean hasTextComponents();
    static Vector imList = new Vector();
    static Vector tcList = new Vector();

    native void pSetIMMOption(String option);

    static {
        initIDs();
    }

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void notifyIMMOptionChange(){

      // REMIND: IS THIS SECURE??? CAN USER CODE SUBCLASS INPUTMETHODMGR???
      InputMethodManager.getInstance().notifyChangeRequest(target);
    }

    public void addInputMethod(X11InputMethod im){
	if (!imList.contains(im))
	    imList.addElement(im);
    }

    public void removeInputMethod(X11InputMethod im){
	if (imList.contains(im))
	    imList.removeElement(im);
    }

    public void addTextComponent(MComponentPeer tc){
	if (tcList.contains(tc))
	    return;
	if (tcList.isEmpty()){
	  addTextComponentNative();
	  if (!imList.isEmpty()) {
	    for (int i = 0; i < imList.size(); i++) {
		((X11InputMethod)imList.elementAt(i)).reconfigureXIC((MInputMethodControl)this);
	    }
	  }
	}
	tcList.addElement(tc);
    }

    public void removeTextComponent(MComponentPeer tc){
	if (!tcList.contains(tc))
	    return;
	tcList.removeElement(tc);
	if (tcList.isEmpty()){
	  removeTextComponentNative();
	  if (!imList.isEmpty()){
	    for (int i = 0; i < imList.size(); i++) {
		((X11InputMethod)imList.elementAt(i)).reconfigureXIC((MInputMethodControl)this);
	    }
	  }
	}
    }

    // this function is privileged! do not change it to public!
    private static int getInset(final String name, final int def) {
	Integer tmp = (Integer) java.security.AccessController.doPrivileged(
	    new sun.security.action.GetIntegerAction(name, def));
	return tmp.intValue();
    }

    MDialogPeer(Dialog target) {
        init(target, new Insets(
            getInset("awt.frame.topInset", -1),
            getInset("awt.frame.leftInset", -1),
            getInset("awt.frame.bottomInset", -1),
            getInset("awt.frame.rightInset", -1)));

	allDialogs.addElement(this);

	InputMethodManager imm = InputMethodManager.getInstance();
	String menuString = imm.getTriggerMenuString();
	if (menuString != null)
	{ 
	  pSetIMMOption(menuString);
	}

        String title = target.getTitle();
        pSetTitle(title != null? title : " ");

	Font f = target.getFont();
	if (f == null) {
	    f = new Font("Dialog", Font.PLAIN, 12);
	    target.setFont(f);
	    setFont(f);
	}
	Color c = target.getBackground();
	if (c == null) {
	    target.setBackground(SystemColor.window);
	    setBackground(SystemColor.window);
	}
	c = target.getForeground();
	if (c == null) {
	    target.setForeground(SystemColor.windowText);
	    setForeground(SystemColor.windowText);
	}
	setResizable(target.isResizable());
    }

    public void setTitle(String title) {
	pSetTitle(title);
    }

    protected void disposeImpl() {
	allDialogs.removeElement(this);
	super.disposeImpl();
    }

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleMoved(int x, int y) {
        postEvent(new ComponentEvent(target, ComponentEvent.COMPONENT_MOVED));
    }

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleActivate() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_ACTIVATED));
    }

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleDeactivate() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_DEACTIVATED));
    }

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleQuit() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_CLOSING));
    }

    public void show() {
	if (((Dialog)target).isModal()) {
	    ModalThread mt = new ModalThread(this);
	    synchronized(mt) {
	        try {
	    	    mt.start();
		    mt.wait();
	        } catch (InterruptedException e) {
	        }
	    }
	    // End of modal show means dialog is invisible but
	    // has not had visible flag turned off.
	    target.hide();
	} else {
	    pShow();
 	}
    }	    

    public void toFront() {
	pShow();
    }
    public native void toBack();

    /* Handle TAB and Shift-TAB events. */
    public int handleFocusTraversalEvent(KeyEvent e) {

	if (e.getKeyCode() != KeyEvent.VK_TAB || 
	    (e.getSource() instanceof TextArea))
	    return IGNORE_EVENT;

	if ((e.getModifiers() & ~InputEvent.SHIFT_MASK) > 0)
	    return IGNORE_EVENT;

 	int id = e.getID();
  	if (id == KeyEvent.KEY_RELEASED || id == KeyEvent.KEY_TYPED)
	    return CONSUME_EVENT;

  	if (e.isShiftDown())
  	    return FOCUS_PREVIOUS;
	else
 	    return FOCUS_NEXT;
    }

    public Insets insets() {
	return insets;
    }

    /**
     * Called to inform the Dialog that its size has changed and it
     * should layout its children.
     */
    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleResize(int width, int height) {
        postEvent(new ComponentEvent(target, ComponentEvent.COMPONENT_RESIZED));
    }
}

/*
 * We must use a separate thread to post the modal dialog, else
 * there is the potential for the modal loop to do event handling
 * with unsecure classes on the call stack.
 */
class ModalThread extends Thread {
    MDialogPeer d;

    public ModalThread(MDialogPeer d) {
	super("AWT-Modal");
	this.d = d;
    }
    public void run() {
	synchronized(this) {
	    d.pShow();
	    notifyAll();
	}
    }
}
