/*
 * @(#)MDragSourceContextPeer.java	1.28 98/08/28
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.event.InputEvent;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;

import java.awt.peer.ComponentPeer;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceContext;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.InvalidDnDOperationException;

import java.awt.dnd.peer.DragSourceContextPeer;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.StringBufferInputStream;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import sun.awt.motif.MToolkit;
import sun.awt.motif.MDropTargetContextPeer;

import java.lang.reflect.InvocationTargetException;

import javax.swing.SwingUtilities;

import java.rmi.MarshalledObject;
import java.rmi.MarshalException;

/**
 * <p>
 * TBC
 * </p>
 *
 * @version 1.28
 * @since JDK1.2
 *
 */

final class MDragSourceContextPeer implements DragSourceContextPeer {

    /**
     * The Listener Dispatcher
     */

    static private MDragSourceContextPeerDispatcher listenerDispatcher = new MDragSourceContextPeerDispatcher();

    /**
     * provide a MDragSourceContextPeer to the Toolkit object ...
     */

     static DragSourceContextPeer getDragSourceContextPeer(DragGestureEvent dge) {
	return new MDragSourceContextPeer(dge);
     }

    /**
     * construct a new MDragSourceContextPeer
     */

    private MDragSourceContextPeer(DragGestureEvent dge) {
    	super();

	trigger    = dge;
	component  = dge.getComponent();
	dragSource = dge.getDragSource(); 
    }


    /**
     * initiate a DnD operation ...
     */

    public void startDrag(DragSourceContext dsc, Cursor c, Image di, Point ioff) throws InvalidDnDOperationException {
	dragSourceContext = dsc;
	cursor 		  = c;
	actions           = trigger.getSourceAsDragGestureRecognizer().getSourceActions();

	DataFlavor[] dfs     = dsc.getTransferable().getTransferDataFlavors();
	Map	     natives = dragSource.getFlavorMap().getNativesForFlavors(dfs);

	/*
	 * trim the flavors that cannot be externally transferred.
	 */

	for (int i = 0; i < dfs.length; i++) {
	    if (!(dfs[i].isRepresentationClassInputStream()  ||
		  dfs[i].isRepresentationClassSerializable() ||
		  dfs[i].isRepresentationClassRemote()       ||
	          DataFlavor.javaFileListFlavor.equals(dfs[i]))) natives.remove(dfs[i]);
	}

	try {
	    nativeCtxt = startDrag(MToolkit.getNativeContainer(dsc.getComponent()).getPeer(),
			           trigger.getTriggerEvent(),
				   cursor,
				   cursor == null ? 0 : cursor.getType(),
				   dsc.getSourceActions(),
			           natives.values().toArray()
	    );
	} catch (Exception e) {
	    e.printStackTrace();
	    throw new InvalidDnDOperationException();
	}

	if (nativeCtxt == 0) {
	    throw new InvalidDnDOperationException();
	}

	MDropTargetContextPeer.setCurrentJVMLocalSourceTransferable(dsc.getTransferable());
    }

    /**
     * return cursor
     */

    public Cursor getCursor() { return cursor; }

    /**
     *
     */

    public void setCursor(Cursor c) throws InvalidDnDOperationException {
	if (cursor == null || !cursor.equals(c)) {
	    cursor = c;
	    if (nativeCtxt != 0) setCursor(nativeCtxt, c, c == null ? 0 : c.getType());
	}
    }

    /**
     * notify peer that the Transferable has changed
     */

    public void transferablesFlavorsChanged() {
    }

    /**
     * return triggering event
     */

    public DragGestureEvent getTrigger() { return trigger; }

    /**
     * upcall from native code
     */

    private int dragEnter(int dropAction, int targetActions, int modifiers, boolean valid) {
        inValidSite = valid;
	if (inValidSite) {
	    listenerDispatcher.deliverEnter(dragSourceContext, new DragSourceDragEvent(dragSourceContext, mapOperation(dropAction), targetActions, modifiers));
	}
	return actions;
    }
 
    /**
     * upcall from native code
     */
 
    private int dragMotion(int dropAction, int targetActions, int modifiers, boolean valid) {
	if (!valid) {
	    dragExit(modifiers);
	} else if (!inValidSite) {
	    return dragEnter(dropAction, targetActions, modifiers, valid);
	} else {
	    if (inValidSite) {
		listenerDispatcher.deliverMotion(dragSourceContext, new DragSourceDragEvent(dragSourceContext, mapOperation(dropAction), targetActions, modifiers));
	    }
	}

	return actions;
    }

    /**
     * upcall from native code
     */

    private int operationChanged(int dropAction, int targetActions, int modifiers, boolean valid) {
	listenerDispatcher.deliverGestureChanged(dragSourceContext, new DragSourceDragEvent(dragSourceContext, mapOperation(dropAction), targetActions, modifiers));
	return actions;
    }
 
    /**
     * upcall from native code
     */

    private void dragExit(int modifiers) {
	if (inValidSite) {
	    listenerDispatcher.deliverExit(dragSourceContext, new DragSourceEvent(dragSourceContext));
	    inValidSite = false;
	}
    }
 
    /**
     * upcall from native code
     */
 
    private void dropStart(int targetAction, boolean overValidDropTarget, boolean valid) {
	if (overValidDropTarget) {
	    dropOccurred  = true;
	    dropOperation = targetAction;
	}
    }

    /**
     *
     */

    private MDragSourceContextPeerBuffer convertData(String name, int maxLength, long requestor) {
	Long			     requestorID = new Long(requestor);
	MDragSourceContextPeerBuffer mdscpb      = requestor == 0 ? null : (MDragSourceContextPeerBuffer)transfers.get(requestorID);

	if (mdscpb == null) {
	    Map flavors = dragSource.getFlavorMap().getFlavorsForNatives(new String[] { name });
	    if (flavors == null || flavors.isEmpty()) return null;

	    DataFlavor df = (DataFlavor)flavors.get(name);
	    Object     o  = null;

	    if (df == null) return null;

	    try {
	        o = dragSourceContext.getTransferable().getTransferData(df); 
	    } catch (IOException ioe) {
		return null;
	    } catch (UnsupportedFlavorException ufe) {
		return null;
	    }

	    if (df.isRepresentationClassInputStream()) {
		try {
	            mdscpb = new MDragSourceContextPeerBuffer(requestorID, (InputStream)o);
		} catch (Exception e) {
		    e.printStackTrace();
		    return null;
		}
	    } else if (df.isRepresentationClassRemote()) {
		try {
	            mdscpb = new MDragSourceContextPeerBuffer(requestorID, new MarshalledObject(o));
		} catch (Exception e) {
		    e.printStackTrace();
		    return null;
		}
	    } else if (df.isRepresentationClassSerializable()) {
		try {
	            mdscpb = new MDragSourceContextPeerBuffer(requestorID, o);
		} catch (Exception e) {
		    e.printStackTrace();
		    return null;
		}
	    } else {
		Class rc = df.getRepresentationClass();

		if (java.lang.String.class.equals(rc)) {
		    try {
	                mdscpb = new MDragSourceContextPeerBuffer(requestorID,(String)o);
		    } catch (Exception e) {
			e.printStackTrace();
			return null;
		    }
		}
	    }

	    if (mdscpb != null && requestor != 0) {
	        transfers.put(requestorID, mdscpb);
	    }
	}

	return mdscpb;
    }
    

    /**
     * upcall to get list of file names ...
     */

    private String[] convertFileList(String name) {
	Map      flavors = dragSource.getFlavorMap().getFlavorsForNatives(new String[] { name } );

	if (flavors == null || flavors.isEmpty()) return null;

	DataFlavor df = (DataFlavor)flavors.get(name);

	if (df == null) return null;

	final java.util.List list;
	try {
	    list = (java.util.List)dragSourceContext.getTransferable().getTransferData(df);
	} catch (IOException ioe) {
	    return null;
	} catch (UnsupportedFlavorException ufe) {
	    return null;
        } catch (ClassCastException cce) {
	    return null;
        }

	if (list.isEmpty()) return null;

	int nfiles = 0;

	for (int i = 0; i < list.size(); i++) {
	    Object o = list.get(i);
	    if (o instanceof java.io.File || o instanceof String) nfiles++;
	}

	if (nfiles == 0) return null;

	final String[] files = new String[nfiles];

	try {
	    java.security.AccessController.doPrivileged(
				new java.security.PrivilegedExceptionAction() {
		public Object run() throws IOException {

		    for (int i = 0; i < list.size(); i++) {
			Object o = list.get(i);

			if (o instanceof java.io.File) {
			    files[i] = 
				((java.io.File)list.get(i)).getCanonicalPath();
			}
			else if (o instanceof String) files[i] = (String)o;
		    }
		    return null;
		}
	    });
	} catch (java.security.PrivilegedActionException pae) {
	    // give up!
	    return null;
	}
	return files;
    }

    /**
     * upcall from native code
     */

    void freeBuffer(MDragSourceContextPeerBuffer buffer) { 
	try {
	    buffer.close();
	} catch (IOException ioe) {
	    // Oops
	}

	if (buffer.getRequestorID().longValue() != 0)
	    transfers.remove(buffer.getRequestorID());
    }
    
    /**
     * upcall from native code
     */
    
    private void dropFinished(boolean success) {
	dropSuccess = success;
    }

    /**
     * upcall from native code
     */

    private void dragDropFinished() {
	DragSourceDropEvent dsde = dropOccurred ? new DragSourceDropEvent(dragSourceContext, dropOperation, dropSuccess) : new DragSourceDropEvent(dragSourceContext);
	listenerDispatcher.deliverDropFinish(dragSourceContext, dsde);

	MDropTargetContextPeer.setCurrentJVMLocalSourceTransferable(null);
    }

    /**
     * mapOperation
     */
 
    private int mapOperation(int operation) {
        int[] operations = {
                DnDConstants.ACTION_MOVE,
                DnDConstants.ACTION_COPY,
                DnDConstants.ACTION_LINK,
        };
 
        operation &= DnDConstants.ACTION_COPY_OR_MOVE | DnDConstants.ACTION_LINK;
 
        for (int i = 0; i < operations.length; i++) {
            if ((operation & operations[i]) == operations[i]) {
                return operations[i];
            }
        }
 
        return DnDConstants.ACTION_NONE;
    }


    /**
     * downcall into native code
     */

    private native long startDrag(ComponentPeer nativeCont, InputEvent nativeTrigger, Cursor c, int ctype, int actions, Object[] types);

    /**
     * downcall into native code
     */


    private native void setCursor(long nativeCtxt, Cursor c, int cType);

    /*
     * fields
     */

    private DragSource 			     dragSource;

    private Component  			     component;
    private DragGestureEvent		     trigger;
    private Cursor     			     cursor;
    private int        			     actions;

    private Component  			     nativeCont;
    private long       			     nativeCtxt;

    private DragSourceContext		     dragSourceContext;
    private boolean			     dropOccurred  = false;
    private int				     dropOperation = DnDConstants.ACTION_NONE;
    private boolean			     dropSuccess   = false;

    private HashMap  			     transfers     = new HashMap();

    private boolean			     inValidSite   = false;
}

/*****************************************************************************/

final class MDragSourceContextPeerDispatcher implements Runnable {

    private static final int NONE   = 0;
    private static final int ENTER  = 1;
    private static final int MOTION = 2;
    private static final int CHANGE = 3;
    private static final int EXIT   = 4;
    private static final int FINISH = 5;
    
    /**
     * deliver the dragEnter
     */

    synchronized void deliverEnter(DragSourceListener l, DragSourceDragEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = ENTER;

	dispatch();
    }

    /**
     * deliver the dragMotion
     */

    synchronized void deliverMotion(DragSourceListener l, DragSourceDragEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = MOTION;

	dispatch();
    }

    /**
     * deliver the dragGestureChanged
     */

    synchronized void deliverGestureChanged(DragSourceListener l, DragSourceDragEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = CHANGE;

	dispatch();
    }

    /**
     * deliver the dragExit
     */

    synchronized void deliverExit(DragSourceListener l, DragSourceEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = EXIT;

	dispatch();
    }

    /**
     * deliver the dropFinish
     */

    synchronized void deliverDropFinish(DragSourceListener l, DragSourceDropEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = FINISH;

	dispatch();
    }

    /**
     * do the dispatch.
     */

    private void dispatch() {
	boolean done = false;

	while (!done) try {
	    SwingUtilities.invokeAndWait((Runnable)this);
	    done = true;
	} catch (InterruptedException ie) {
	    ie.printStackTrace();
	} catch (InvocationTargetException ite) {
	    ite.printStackTrace();
	    done = true;
	}
    }

    /**
     * deliver the event
     */

    public void run() {
	try {
	    switch (dispatchType) {
		case ENTER:
		    dsl.dragEnter((DragSourceDragEvent)event);
		    break;

		case MOTION:
		    dsl.dragOver((DragSourceDragEvent)event);
		    break;

		case CHANGE:
		    dsl.dropActionChanged((DragSourceDragEvent)event);
		    break;

		case EXIT:
		    dsl.dragExit((DragSourceEvent)event);
		    break;

		case FINISH:
		    dsl.dragDropEnd((DragSourceDropEvent)event);
		    break;

		default:
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	} finally {
	    dsl		 = null;
	    dispatchType = NONE;
	}
    }

    /*
     *
     */

    DragSourceListener	    dsl;

    private int		    dispatchType = NONE;
    private DragSourceEvent event;
}

/*****************************************************************************/

/*
 * for large data transfers this methods is extremely inefficent and should
 * be replaced with a scatter/gather algorithm.
 */

class MDragSourceContextPeerBuffer extends ByteArrayOutputStream {

    /**
     * create a buffer
     */

    MDragSourceContextPeerBuffer(Long requestorKey) {
	super();

	requestorID = requestorKey;
    }

    /**
     * create a buffer
     */

    MDragSourceContextPeerBuffer(Long requestorKey, Object o) throws IOException {
	this(requestorKey);

	ObjectOutputStream oos = new ObjectOutputStream(this);

	oos.writeObject(o);

	oos.close();
	close();
    }

    /**
     * create a buffer
     */

    MDragSourceContextPeerBuffer(Long requestorKey, String s) throws IOException {
	this(requestorKey);
	
	DataOutputStream dos = new DataOutputStream(this);

	dos.writeUTF(s);

	dos.close();
	close();
    }

    /**
     * create a buffer
     */

    MDragSourceContextPeerBuffer(Long requestorKey, byte[] ba) {
	this(requestorKey);

	count = ba.length;

	System.arraycopy(ba, 0, buf = new byte[count], 0, count);

	try {
	    close();
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	}
    }

    /**
     * create a buffer
     */

    MDragSourceContextPeerBuffer(Long requestorKey, InputStream is) throws IOException {
	this(requestorKey);
	fill(is);

	close();
    }

    /**
     * fill the buffer from an InputStream ...
     */

    private void fill(InputStream is) throws IOException {
	boolean eof   = false;
	int     avail = is.available();
	byte[]  tmp   = new byte[avail > 8192 ? avail : 8192];

	do {
	    int ret;

	    if (!(eof = (ret = is.read(tmp, 0, tmp.length)) == -1))
		write(tmp, 0, ret);
	} while (!eof);

	try {
	    is.close();
	} catch (IOException ioe) {
	    // Oops!
	}
    }

    /**
     * return the buffer
     */

    byte[] getBuffer() { return buf; }

    /**
     * update the read Cursors offset, return -1 when at EOF ... 
     * up called from native code.
     */

    int incrReadCursor(int incr) { 
	int t    = readCursor + incr;
	int blen = size();

	readCursor = (t >= blen ? blen : t); 

	return (readCursor == blen ? -1 : readCursor);
    }


    /**
     * get platform transfer ID
     */

    Long getRequestorID() { return requestorID; }

    /*
     * fields
     */

    private int  readCursor = 0;
    private Long requestorID;
}

/*****************************************************************************/
