/*
 * @(#)MFileDialogPeer.java	1.21 98/10/29
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.io.FilenameFilter;
import java.awt.datatransfer.*;

public class MFileDialogPeer extends MDialogPeer implements FileDialogPeer {
    native void create(MComponentPeer parent);
    void create(MComponentPeer parent, Object arg) {
	create(parent);
    }
    public MFileDialogPeer(FileDialog target) {
	super(target);
	FileDialog	fdialog = (FileDialog)target;
	String		dir = fdialog.getDirectory();
        String          file = fdialog.getFile();

	insets = new Insets(0, 0, 0, 0);
	setDirectory(dir);
	if (file != null) {
	    setFile(file);
	}
    }
    native void		pReshape(int x, int y, int width, int height);
    native void         pDispose();
    native void		pShow();
    native void		pHide();
    native void		setFileEntry(String dir, String file);
    native void insertReplaceFileDialogText(String l);
    public native void	setFont(Font f);
    private native void targetSetDirectory_NoClientCode(FileDialog target,
							String dir);
    private native void targetSetFile_NoClientCode(FileDialog target,
						   String file);

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void		handleSelected(String file) {
        FileDialog fileDialog = (FileDialog) target;
	int index = file.lastIndexOf('/');
	String dir;

	if (index == -1) {
	    dir = "./";
	    targetSetFile_NoClientCode(fileDialog, file);
	} else {
	    dir = file.substring(0, index + 1);
	    targetSetFile_NoClientCode(fileDialog, file.substring(index + 1));
	}
	targetSetDirectory_NoClientCode(fileDialog, dir);
    } // handleSelected()

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void		handleCancel() {
        FileDialog fileDialog = (FileDialog) target;

	targetSetFile_NoClientCode(fileDialog, null);
    } // handleCancel()

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void		handleQuit() {
	final FileDialog fileDialog = (FileDialog)target;
	MToolkit.executeOnEventHandlerThread(fileDialog, new Runnable() {
	    public void run() {
	        // We don't actually have to invoke the NoClientCode version of
	        // this function, but doing so is more consistent
	        targetSetFile_NoClientCode(fileDialog, null);

		fileDialog.hide();
	    }
	});
    } // handleQuit()

    public  void setDirectory(String dir) {
        String file = ((FileDialog)target).getFile();
	setFileEntry((dir != null) ? dir : "./", (file != null) ? file : "");
    }


    public  void setFile(String file) {
        String dir = ((FileDialog)target).getDirectory();
	setFileEntry((dir != null) ? dir : "./", file);
    }


    public void		setFilenameFilter(FilenameFilter filter) {
    }

    // Called from native widget when paste key is pressed and we
    // already own the selection (prevents Motif from hanging while
    // waiting for the selection)
    //
    public void pasteFromClipboard() {
        Clipboard clipboard = target.getToolkit().getSystemClipboard();
        
	Transferable content = clipboard.getContents(this);
	if (content != null) {
	    try {
		String data = (String)(content.getTransferData(DataFlavor.stringFlavor));
		insertReplaceFileDialogText(data);
	    } catch (Exception e) {       
	    }
        }
    }
}
