/*
 * @(#)MFramePeer.java	1.80 98/09/13
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.ImageObserver;
import sun.awt.image.ImageRepresentation;
import sun.awt.motif.X11InputMethod;
import sun.awt.motif.MInputMethodControl;
import sun.awt.im.*;

class MFramePeer extends MPanelPeer implements FramePeer, MInputMethodControl {
    Insets insets = new Insets(0, 0, 0, 0);
    int         iconWidth  = -1;
    int         iconHeight = -1;
    static Vector allFrames = new Vector();

    native void create(MComponentPeer parent, Object arg);
    native void pSetTitle(String title);
    native void pShow();
    native void pHide();
    public native void setState(int state);
    public native int getState();

    native void pReshape(int x, int y, int width, int height);
    native void pDispose();
    native boolean pGetIconSize(int widthHint, int heightHint);
    native void pSetIconImage(byte[] byteData, int[] intData,
                              int iconWidth, int iconHeight);
    public native void setResizable(boolean resizable);
    native void addTextComponent();

    native void addTextComponentNative();
    native void removeTextComponentNative();
    public native boolean hasTextComponents();
    private static native void initIDs();
    static Vector imList = new Vector();
    static Vector tcList = new Vector();

    static {
        initIDs();
    }

    native void pSetIMMOption(String option);

    // NOTE: This method is called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void notifyIMMOptionChange(){

      // REMIND: IS THIS SECURE??? CAN USER CODE SUBCLASS INPUTMETHODMGR???
      InputMethodManager.getInstance().notifyChangeRequest(target);
    }

    public void addInputMethod(X11InputMethod im) {
	if (!imList.contains(im))
	    imList.addElement(im);
    }

    public void removeInputMethod(X11InputMethod im) {
	if (imList.contains(im))
	    imList.removeElement(im);
    }

    public void addTextComponent(MComponentPeer tc) {
	if (tcList.contains(tc))
	    return;
	if (tcList.isEmpty()){
	    addTextComponentNative();
	    if (!imList.isEmpty()) {
	      for (int i = 0; i < imList.size(); i++) {
		((X11InputMethod)imList.elementAt(i)).reconfigureXIC((MInputMethodControl)this);
	      }
	    }
	}
	tcList.addElement(tc);

    }

    public void removeTextComponent(MComponentPeer tc) {
	if (!tcList.contains(tc))
	    return;
	tcList.removeElement(tc);
	if (tcList.isEmpty()){
	    removeTextComponentNative();
	    if (!imList.isEmpty()) {
	      for (int i = 0; i < imList.size(); i++) {
		((X11InputMethod)imList.elementAt(i)).reconfigureXIC((MInputMethodControl)this);
	      }
	    }
	}
    }

    // this function is privileged! do not change it to public!
    private static int getInset(final String name, final int def) {
	Integer tmp = (Integer) java.security.AccessController.doPrivileged(
	    new sun.security.action.GetIntegerAction(name, def));
	return tmp.intValue();
    }

    MFramePeer(Frame target) {
	/* [pauly, 10/97].  It is wrong to specify these insets as hardcoded
	   literals.  The actual X/Motif insets are dependent upon both the
	   window manager (WM) running and the hardware display. The values
	   below are usually incorrect (although they are in the ballpark).
	   This leads to problems with spurious frame movement and resizing
	   (e.g., bugtraq 4077874 and others). Moreover, it is not in general
	   possible to ascertain the window manager title bar and border sizes;
	   i.e., there is no published api for obtaining from a given window
	   manager these values.  Given these limitations, a workaround is
	   implemented in shellEH() in awt_Frame.c. Please see comments there.
	   In that code, we correct the window manager insets AFTER the frame
	   window is actually rendered on screen and repositioned by the WM
	   to allow for the WM decorations.  At present, inset correction is
	   done only for Motif and OpenLook window managers (MWM and OLWM).
	   We cannot ascertain the presence of other window managers, and so
	   for these, we will continue to use the (probably incorrect) (25,5).	*/
	/* previous comment now mostly obsolete due to resize dance removal.
	   See awt_Frame.c - 
	   Passing in -1 means "use default for the current window manager"
	*/
        init(target, new Insets(
            getInset("awt.frame.topInset", -1),
            getInset("awt.frame.leftInset", -1),
            getInset("awt.frame.bottomInset", -1),
            getInset("awt.frame.rightInset", -1)));

	allFrames.addElement(this);

	InputMethodManager imm = InputMethodManager.getInstance();
	String menuString = imm.getTriggerMenuString();
	if (menuString != null)
	{ 
	  pSetIMMOption(menuString);
	}

	if (target.getTitle() != null) {
	    pSetTitle(target.getTitle());
	}
	Font f = target.getFont();
	if (f == null) {
	    f = new Font("Dialog", Font.PLAIN, 12);
	    target.setFont(f);
	    setFont(f);
	}
	Color c = target.getBackground();
	if (c == null) {
	    target.setBackground(SystemColor.window);
	    setBackground(SystemColor.window);
	}
	c = target.getForeground();
	if (c == null) {
	    target.setForeground(SystemColor.windowText);
	    setForeground(SystemColor.windowText);
	}
	Image icon = target.getIconImage();
	if (icon != null) {
	    setIconImage(icon);
	}
	setResizable(target.isResizable());
    }

    public void setTitle(String title) {
	pSetTitle(title);
    }

    protected void disposeImpl() {
	allFrames.removeElement(this);
	super.disposeImpl();
    }

    public void setIconImage(Image im) {
        int width;
        int height;
        if (im instanceof X11Image) {
            ImageRepresentation ir = ((X11Image)im).getImageRep();
            ir.reconstruct(ImageObserver.ALLBITS);
            width = ir.getWidth();
            height = ir.getHeight();
        }
        else {
            width = im.getWidth(null);
            height = im.getHeight(null);
        }
        if (pGetIconSize(width, height)) {
            Image image = createImage(iconWidth, iconHeight);
            Graphics g = image.getGraphics();
            g.drawImage(im, 0, 0, iconWidth, iconHeight, null);
            DataBuffer db = ((BufferedImage)image).getRaster().getDataBuffer();
            byte[] bytedata = null;
            int[] intdata = null;
            if (db instanceof DataBufferByte) {
                // Pseudocolor data
                bytedata = ((DataBufferByte)db).getData();
            }
            else if (db instanceof DataBufferInt) {
                // Truecolor data
                intdata = ((DataBufferInt) db).getData();
            }
            pSetIconImage(bytedata, intdata, iconWidth, iconHeight);
        }
    }

    public void setMenuBar(MenuBar mb) {
	MMenuBarPeer mbpeer = (MMenuBarPeer) MToolkit.targetToPeer(mb);
	pSetMenuBar(mbpeer);

	Rectangle r = target.bounds();

	pReshape(r.x, r.y, r.width, r.height);
	if (target.isVisible()) {
	    target.validate();
	}
    }
    native void pSetMenuBar(MMenuBarPeer mbpeer);

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleActivate() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_ACTIVATED));
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleDeactivate() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_DEACTIVATED));
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleQuit() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_CLOSING));
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleIconify() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_ICONIFIED));
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleDeiconify() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_DEICONIFIED));
    }

    public void toFront() {
	pShow();
    }
    public native void toBack();

    /* Handle TAB and Shift-TAB events. */
    public int handleFocusTraversalEvent(KeyEvent e) {

	if (e.getKeyCode() != KeyEvent.VK_TAB || 
	    (e.getSource() instanceof TextArea))
	    return IGNORE_EVENT;

	if ((e.getModifiers() & ~InputEvent.SHIFT_MASK) > 0)
 	    return IGNORE_EVENT;

	int id = e.getID();
 	if (id == KeyEvent.KEY_RELEASED || id == KeyEvent.KEY_TYPED)
 	    return CONSUME_EVENT;

 	if (e.isShiftDown())
 	    return FOCUS_PREVIOUS;
	else
 	    return FOCUS_NEXT;
    }

    /**
     * Called to inform the Frame that it has moved.
     */
    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleMoved(int x, int y) {
        postEvent(new ComponentEvent(target, ComponentEvent.COMPONENT_MOVED));
    }

    /**
     * Called to inform the Frame that its size has changed and it
     * should layout its children.
     */
    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleResize(int width, int height) {

      // REMIND: Is this secure? Can client code subclass input method?
      if (!tcList.isEmpty() &&
	   !imList.isEmpty()){
	int i;
	for (i = 0; i < imList.size(); i++){
	  ((X11InputMethod)imList.elementAt(i)).configureStatus();
	}	
      }
        postEvent(new ComponentEvent(target, ComponentEvent.COMPONENT_RESIZED));
    }

    public Insets getInsets() {
	return insets;
    }

    static final int CROSSHAIR_INSET = 5;

    static final int BUTTON_Y = CROSSHAIR_INSET + 1;
    static final int BUTTON_W = 17;
    static final int BUTTON_H = 17;
  
    static final int SYS_MENU_X = CROSSHAIR_INSET + 1;
    static final int SYS_MENU_CONTAINED_X = SYS_MENU_X + 5;
    static final int SYS_MENU_CONTAINED_Y = BUTTON_Y + 7;
    static final int SYS_MENU_CONTAINED_W = 8;
    static final int SYS_MENU_CONTAINED_H = 3;
  
    static final int MAXIMIZE_X_DIFF = CROSSHAIR_INSET + BUTTON_W;
    static final int MAXIMIZE_CONTAINED_X_DIFF = MAXIMIZE_X_DIFF - 5;
    static final int MAXIMIZE_CONTAINED_Y = BUTTON_Y + 5;
    static final int MAXIMIZE_CONTAINED_W = 8;
    static final int MAXIMIZE_CONTAINED_H = 8;
  
    static final int MINIMIZE_X_DIFF = MAXIMIZE_X_DIFF + BUTTON_W;
    static final int MINIMIZE_CONTAINED_X_DIFF = MINIMIZE_X_DIFF - 7;
    static final int MINIMIZE_CONTAINED_Y = BUTTON_Y + 7;
    static final int MINIMIZE_CONTAINED_W = 3;
    static final int MINIMIZE_CONTAINED_H = 3;
  
    static final int TITLE_X = SYS_MENU_X + BUTTON_W;
    static final int TITLE_W_DIFF = BUTTON_W * 3 + CROSSHAIR_INSET * 2 - 1;
    static final int TITLE_MID_Y = BUTTON_Y + (BUTTON_H / 2);

    static final int MENUBAR_X = CROSSHAIR_INSET + 1;
    static final int MENUBAR_Y = BUTTON_Y + BUTTON_H;

    static final int HORIZ_RESIZE_INSET = CROSSHAIR_INSET + BUTTON_H;
    static final int VERT_RESIZE_INSET = CROSSHAIR_INSET + BUTTON_W;


    /*
     * Print the native component by rendering the Motif look ourselves.
     * We also explicitly print the MenuBar since a MenuBar isn't a subclass
     * of Component (and thus it has no "print" method which gets called by
     * default).
     */
    public void print(Graphics g) {
        super.print(g);

	Frame f = (Frame)target;
	Insets finsets = f.getInsets();
	Dimension fsize = f.getSize();

	Color bg = f.getBackground();
	Color fg = f.getForeground();
	Color highlight = bg.brighter();
	Color shadow = bg.darker();

	// Well, we could query for the currently running window manager
	// and base the look on that, or we could just always do dtwm.
	// aim, tball, and levenson all agree we'll just do dtwm.

	// top outer -- because we'll most likely be drawing on white paper,
	// for aesthetic reasons, don't make any part of the outer border
	// pure white
	if (highlight.equals(Color.white)) {
	    g.setColor(new Color(230, 230, 230));
	}
	else {
	    g.setColor(highlight);
	}
	g.drawLine(0, 0, fsize.width, 0);
	g.drawLine(0, 1, fsize.width - 1, 1);

	// left outer
	// if (highlight.equals(Color.white)) {
	//     g.setColor(new Color(230, 230, 230));
	// }
	// else {
	//     g.setColor(highlight);
	// }
	g.drawLine(0, 0, 0, fsize.height);
	g.drawLine(1, 0, 1, fsize.height - 1);

	// bottom cross-hair
	g.setColor(highlight);
	g.drawLine(CROSSHAIR_INSET + 1, fsize.height - CROSSHAIR_INSET,
		   fsize.width - CROSSHAIR_INSET, 
		   fsize.height - CROSSHAIR_INSET);

	// right cross-hair
	// g.setColor(highlight);
	g.drawLine(fsize.width - CROSSHAIR_INSET, CROSSHAIR_INSET + 1,
		   fsize.width - CROSSHAIR_INSET,
		   fsize.height - CROSSHAIR_INSET);

	// bottom outer
	g.setColor(shadow);
	g.drawLine(1, fsize.height, fsize.width, fsize.height);
	g.drawLine(2, fsize.height - 1, fsize.width, fsize.height - 1);

	// right outer
	// g.setColor(shadow);
	g.drawLine(fsize.width, 1, fsize.width, fsize.height);
	g.drawLine(fsize.width - 1, 2, fsize.width - 1, fsize.height);

	// top cross-hair
	// g.setColor(shadow);
	g.drawLine(CROSSHAIR_INSET, CROSSHAIR_INSET, 
		   fsize.width - CROSSHAIR_INSET, CROSSHAIR_INSET);
	
	// left cross-hair
	// g.setColor(shadow);
	g.drawLine(CROSSHAIR_INSET, CROSSHAIR_INSET, CROSSHAIR_INSET,
		   fsize.height - CROSSHAIR_INSET);

	// system menu
	g.setColor(bg);
	g.fill3DRect(SYS_MENU_X, BUTTON_Y, BUTTON_W, BUTTON_H, true);
	g.fill3DRect(SYS_MENU_CONTAINED_X, SYS_MENU_CONTAINED_Y,
		     SYS_MENU_CONTAINED_W, SYS_MENU_CONTAINED_H, true);

	// title bar
	// g.setColor(bg);
	g.fill3DRect(TITLE_X, BUTTON_Y, fsize.width - TITLE_W_DIFF, BUTTON_H,
		     true);
	
	// minimize button
	// g.setColor(bg);
	g.fill3DRect(fsize.width - MINIMIZE_X_DIFF, BUTTON_Y, BUTTON_W,
		     BUTTON_H, true);
	g.fill3DRect(fsize.width - MINIMIZE_CONTAINED_X_DIFF,
		     MINIMIZE_CONTAINED_Y, MINIMIZE_CONTAINED_W,
		     MINIMIZE_CONTAINED_H, true);

	// maximize button
	// g.setColor(bg);
	g.fill3DRect(fsize.width - MAXIMIZE_X_DIFF, BUTTON_Y, BUTTON_W,
		     BUTTON_H, true);
	g.fill3DRect(fsize.width - MAXIMIZE_CONTAINED_X_DIFF,
		     MAXIMIZE_CONTAINED_Y, MAXIMIZE_CONTAINED_W,
		     MAXIMIZE_CONTAINED_H, true);

	// title bar text
	g.setColor(fg);
	Font sysfont = new Font("SansSerif", Font.PLAIN, 10);
	g.setFont(sysfont);
	FontMetrics sysfm = g.getFontMetrics();
	String ftitle = f.getTitle();
	g.drawString(ftitle, 
		     ((TITLE_X + TITLE_X + fsize.width - TITLE_W_DIFF) / 2) - 
		         (sysfm.stringWidth(ftitle) / 2),
		     TITLE_MID_Y + sysfm.getMaxDescent());
	
	if (f.isResizable()) {
	  // add resize cross hairs

	  // upper-left horiz (shadow)
	  g.setColor(shadow);
	  g.drawLine(1, HORIZ_RESIZE_INSET, CROSSHAIR_INSET,
		     HORIZ_RESIZE_INSET);
	  // upper-left vert (shadow)
	  // g.setColor(shadow);
	  g.drawLine(VERT_RESIZE_INSET, 1, VERT_RESIZE_INSET, CROSSHAIR_INSET);
	  // upper-right horiz (shadow)
	  // g.setColor(shadow);
	  g.drawLine(fsize.width - CROSSHAIR_INSET + 1, HORIZ_RESIZE_INSET,
		     fsize.width, HORIZ_RESIZE_INSET);
	  // upper-right vert (shadow)
	  // g.setColor(shadow);
	  g.drawLine(fsize.width - VERT_RESIZE_INSET - 1, 2,
		     fsize.width - VERT_RESIZE_INSET - 1, CROSSHAIR_INSET + 1);
	  // lower-left horiz (shadow)
	  // g.setColor(shadow);
	  g.drawLine(1, fsize.height - HORIZ_RESIZE_INSET - 1,
		     CROSSHAIR_INSET, fsize.height - HORIZ_RESIZE_INSET - 1);
	  // lower-left vert (shadow)
	  // g.setColor(shadow);
	  g.drawLine(VERT_RESIZE_INSET, fsize.height - CROSSHAIR_INSET + 1,
		     VERT_RESIZE_INSET, fsize.height);
	  // lower-right horiz (shadow)
	  // g.setColor(shadow);
	  g.drawLine(fsize.width - CROSSHAIR_INSET + 1,
		     fsize.height - HORIZ_RESIZE_INSET - 1, fsize.width,
		     fsize.height - HORIZ_RESIZE_INSET - 1);
	  // lower-right vert (shadow)
	  // g.setColor(shadow);
	  g.drawLine(fsize.width - VERT_RESIZE_INSET - 1,
		     fsize.height - CROSSHAIR_INSET + 1,
		     fsize.width - VERT_RESIZE_INSET - 1, fsize.height);

	  // upper-left horiz (highlight)
	  g.setColor(highlight);
	  g.drawLine(2, HORIZ_RESIZE_INSET + 1, CROSSHAIR_INSET,
		     HORIZ_RESIZE_INSET + 1);
	  // upper-left vert (highlight)
	  // g.setColor(highlight);
	  g.drawLine(VERT_RESIZE_INSET + 1, 2, VERT_RESIZE_INSET + 1,
		     CROSSHAIR_INSET);
	  // upper-right horiz (highlight)
	  // g.setColor(highlight);
	  g.drawLine(fsize.width - CROSSHAIR_INSET + 1,
		     HORIZ_RESIZE_INSET + 1, fsize.width - 1,
		     HORIZ_RESIZE_INSET + 1);
	  // upper-right vert (highlight)
	  // g.setColor(highlight);
	  g.drawLine(fsize.width - VERT_RESIZE_INSET, 2, 
		     fsize.width - VERT_RESIZE_INSET, CROSSHAIR_INSET);
	  // lower-left horiz (highlight)
	  // g.setColor(highlight);
	  g.drawLine(2, fsize.height - HORIZ_RESIZE_INSET, CROSSHAIR_INSET,
		     fsize.height - HORIZ_RESIZE_INSET);
	  // lower-left vert (highlight)
	  // g.setColor(highlight);
	  g.drawLine(VERT_RESIZE_INSET + 1,
		     fsize.height - CROSSHAIR_INSET + 1, 
		     VERT_RESIZE_INSET + 1, fsize.height - 1);
	  // lower-right horiz (highlight)
	  // g.setColor(highlight);
	  g.drawLine(fsize.width - CROSSHAIR_INSET + 1,
		     fsize.height - HORIZ_RESIZE_INSET, fsize.width - 1,
		     fsize.height - HORIZ_RESIZE_INSET);
	  // lower-right vert (highlight)
	  // g.setColor(highlight);
	  g.drawLine(fsize.width - VERT_RESIZE_INSET,
		     fsize.height - CROSSHAIR_INSET + 1,
		     fsize.width - VERT_RESIZE_INSET, fsize.height - 1);
	}

	MenuBar mb = f.getMenuBar();
	if (mb != null) {
	    MMenuBarPeer peer = (MMenuBarPeer) MToolkit.targetToPeer(mb);
	    if (peer != null) {
	        Insets insets = insets();
	        Graphics ng = g.create();
		try {
		    ng.translate(MENUBAR_X, MENUBAR_Y);
	            peer.print(ng);
		} finally {
		    ng.dispose();
		}
	    }
	}
    }

    /**
     * DEPRECATED:  Replaced by getInsets().
     */
    public Insets insets() {
	return getInsets();
    }

}

