/*
 * @(#)MInputMethodControl.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import sun.awt.motif.MComponentPeer;
import sun.awt.motif.X11InputMethod;

/**
 * An interface for controlling containment hierarchy configuration to
 * keep track of existence of any TextArea or TextField and to manage
 * input method status area.
 *
 * @version	1.4 09/21/98
 * @auther	JavaSoft International
 */
interface MInputMethodControl {

    /**
     * Informs Frame or Dialog that a text component has been added to
     * the hierarchy.
     * @param	textComponentPeer	peer of the text component
     */
    void addTextComponent(MComponentPeer textComponentPeer);

    /**
     * Informs Frame or Dialog that a text component has been removed
     * from the hierarchy.
     * @param textComponentPeer peer of the text component
     */
    void removeTextComponent(MComponentPeer textComponentPeer);

    /**
     * Returns true if the containment hierarchy has any text
     * components.
     */
    boolean hasTextComponents();

    /**
     * Inform Frame or Dialog that an X11InputMethod has been
     * constructed so that Frame and Dialog can invoke the method in
     * X11InputMethod to reconfigure XICs.
     * @param	inputMethod	an X11InputMethod instance
     */
    void addInputMethod(X11InputMethod inputMethod);

    /**
     * Inform Frame or Dialog that an X11InputMethod is being destroyed.
     * @param	inputMethod	an X11InputMethod instance
     */
    void removeInputMethod(X11InputMethod inputMethod);
}
