/*
 * @(#)MPanelPeer.java	1.23 98/08/31
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;

class MPanelPeer extends MCanvasPeer implements PanelPeer {

    MPanelPeer() {}

    MPanelPeer(Component target) {
	super(target);
    }

    MPanelPeer(Component target, Object arg) {
	super(target, arg);
    }

    public Insets getInsets() {
	return new Insets(0, 0, 0, 0);
    }

    public void print(Graphics g) {
	super.print(g);
	((Container)target).printComponents(g);
    }

    public void setBackground(Color c) {
	Component comp;
	int i;

	Container cont = (Container) target;
	synchronized(target.getTreeLock()) {
	    int n = cont.getComponentCount();
	    for(i=0; i < n; i++) {
	        comp = cont.getComponent(i);
	        MComponentPeer peer = (MComponentPeer) MToolkit.targetToPeer(comp);
	        if (peer != null) {
		    Color color = comp.getBackground();
                    if (color == null || color.equals(c)) {
			peer.setBackground(c);
		        peer.pSetBackground(c);
		    }
		    if ((comp instanceof java.awt.List) ||
			   (comp instanceof java.awt.TextArea) ||
			   (comp instanceof java.awt.ScrollPane)) {
		        peer.pSetScrollbarBackground(c);
		    }
 		}
	    }
	}
	pSetBackground(c);
    }

    public void setForeground(Color c) {
	Component comp;
	int i;

	Container cont = (Container) target;
	synchronized(target.getTreeLock()) {
	    int n = cont.getComponentCount();
	    for(i=0; i < n; i++) {
	        comp = cont.getComponent(i);
	        MComponentPeer peer = (MComponentPeer) MToolkit.targetToPeer(comp);
	        if (peer != null) {
		    Color color = comp.getForeground();
                    if (color == null || color.equals(c)) {
			peer.setForeground(c);
		        peer.pSetForeground(c);
		    }
		    if ((comp instanceof java.awt.List) ||
			   (comp instanceof java.awt.TextArea) ||
			   (comp instanceof java.awt.ScrollPane)) {
		        peer.pSetInnerForeground(c);
		    }
 		}
	    }
	}
	pSetForeground(c);
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void makeCursorsVisible() {
	final Container container = (Container)target;

	MToolkit.executeOnEventHandlerThread(container, new Runnable() {
	    public void run() {
                Component comp;
                int i;

                synchronized(target.getTreeLock()) {
                    int n = container.getComponentCount();
                    for(i=0; i < n; i++) {
                        comp = container.getComponent(i);
                        MComponentPeer peer = (MComponentPeer) MToolkit.targetToPeer(comp);
                        if (peer != null) {
                            if (comp instanceof Container) {
                                ((MPanelPeer)peer).makeCursorsVisible();
                            }
                            if (peer.cursorSet == 0) {
                                peer.pMakeCursorVisible();
                            }
                        }
                    }
                }
            } // run()
        });
    } // makeCursorsVisible()

    /**
     * DEPRECATED:  Replaced by getInsets().
     */
    public Insets insets() {
	return getInsets();
    }

}

