/*
 * @(#)MScrollPanePeer.java	1.26 98/08/31
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;

class MScrollPanePeer extends MPanelPeer implements ScrollPanePeer {

    final static int UNIT_INCREMENT = 0;
    final static int BLOCK_INCREMENT = 1;

    boolean ignore;
    boolean dragInProgress = false;

    native void create(MComponentPeer parent);

    static {
        initIDs();
    }

    /**
     * Initialize JNI field and method IDs
     */
    private static native void initIDs();

    MScrollPanePeer(Component target) {
	init(target);
        scrollPaneInit();
    }

    MScrollPanePeer(Component target, Object arg) {
	init(target, arg);
        scrollPaneInit();
    }

    void scrollPaneInit() {
	ignore = false;
        ScrollPane sp = (ScrollPane)target;
	Adjustable vadj, hadj;
	if ((vadj = sp.getVAdjustable()) != null) {
            pSetIncrement(Adjustable.VERTICAL, UNIT_INCREMENT, vadj.getUnitIncrement());
	}
	if ((hadj = sp.getHAdjustable()) != null) {
	    pSetIncrement(Adjustable.HORIZONTAL, UNIT_INCREMENT, hadj.getUnitIncrement());
	}
	super.pSetScrollbarBackground(sp.getBackground());
    }

    public void setScrollChild(MComponentPeer child) {
        pSetScrollChild(child);
    }

    public void setBackground(Color c) {
        super.setBackground(c);
	pSetScrollbarBackground(c);
    }

    public void setForeground(Color c) {
        super.setForeground(c);
	pSetInnerForeground(c);
    }

    native void pSetScrollChild(MComponentPeer child);
    native void pSetIncrement(int orient, int type, int incr);
    native int pGetScrollbarSpace(int orient);
    native int pGetBlockIncrement(int orient);
    native Insets pInsets(int w, int h, int childw, int childh);

    public int getHScrollbarHeight() {
        ScrollPane sp = (ScrollPane)target;
        if (sp.getScrollbarDisplayPolicy() == ScrollPane.SCROLLBARS_NEVER) {
	    return 0;
	} else {
	    return pGetScrollbarSpace(Adjustable.HORIZONTAL);
	}
    }

    public int getVScrollbarWidth() {
        ScrollPane sp = (ScrollPane)target;
        if (sp.getScrollbarDisplayPolicy() == ScrollPane.SCROLLBARS_NEVER) {
	    return 0;
	} else {
	    return pGetScrollbarSpace(Adjustable.VERTICAL);
	}
    }

    public Insets insets() {
	ScrollPane sp = (ScrollPane)target;
	Dimension d = sp.size();
	Dimension cd;
	try {
	    Component c = getScrollChild();
	    cd = c.size();
	} catch (NullPointerException e) {
	    cd = new Dimension(0, 0);
	}
	return pInsets(d.width, d.height, cd.width, cd.height);
    }

    public void setUnitIncrement(Adjustable adj, int u) {
        ScrollPane sp = (ScrollPane)target;
        if (sp.getScrollbarDisplayPolicy() != ScrollPane.SCROLLBARS_NEVER) {
	    pSetIncrement(adj.getOrientation(), UNIT_INCREMENT, u);
	}
    }

    public void setValue(Adjustable adj, int v) {
	if (! ignore) {
	    Point p;
	    Component c = getScrollChild();
	    p = c.getLocation();
	    switch(adj.getOrientation()) {
	    case Adjustable.VERTICAL:
		setScrollPosition(-(p.x), v);
		break;
	    case Adjustable.HORIZONTAL:
		setScrollPosition(v, -(p.y));
		break;
	    }
	}
    }

    public native void setScrollPosition(int x, int y);

    public void childResized(int w, int h) {
	// REMIND AIM:  May need to revisit this...
        if (((ScrollPane)target).getScrollbarDisplayPolicy() != ScrollPane.SCROLLBARS_NEVER) {
            ScrollPane sp = (ScrollPane)target;
	    Adjustable vAdj = sp.getVAdjustable();
	    Adjustable hAdj = sp.getHAdjustable();
            pSetIncrement(Scrollbar.VERTICAL, UNIT_INCREMENT, vAdj.getUnitIncrement());
            pSetIncrement(Scrollbar.HORIZONTAL, UNIT_INCREMENT, hAdj.getUnitIncrement());
            pSetIncrement(Scrollbar.VERTICAL, BLOCK_INCREMENT, vAdj.getBlockIncrement());
            pSetIncrement(Scrollbar.HORIZONTAL, BLOCK_INCREMENT, hAdj.getBlockIncrement());
        }

    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void scrolledVertical(int yValue, int scrollType) {
	final MScrollPanePeer scrollPanePeer = this;
	final ScrollPane scrollPane = (ScrollPane)target;
	final int newYValue = yValue;

	MToolkit.executeOnEventHandlerThread(scrollPane, new Runnable() {
	    public void run() {
                Adjustable adj = scrollPane.getVAdjustable();
                if (adj != null) {
	            // To avoid an upcall that can potentially deadlock, 
		    // the adjustable will be mutated by the event 
		    // dispatching thread by sending this event.
	            Adjustor e = new Adjustor(adj, newYValue, scrollPanePeer);
                    postEvent(e);
		}
	    }
        });
    } // scrolledVertical()

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void scrolledHorizontal(int xValue, int scrollType) {
	final MScrollPanePeer scrollPanePeer = this;
	final ScrollPane scrollPane = (ScrollPane)target;
	final int newXValue = xValue;

	MToolkit.executeOnEventHandlerThread(scrollPane, new Runnable() {
	    public void run() {
                Adjustable adj = scrollPane.getHAdjustable();
                if (adj != null) {
	            // To avoid an upcall that can potentially deadlock, the adjustable
	            // will be mutated by the event dispatching thread by sending this
	            // event.
	            Adjustor e = new Adjustor(adj, newXValue, scrollPanePeer);
                    postEvent(e);
		}
	    }
        });
    } // scrolledHorizontal()

    /**
     * This is used to change the adjustable on another thread
     * to represent a change made in the native scrollbar.  Since
     * the change was reflected immediately at the native level,
     * notification from the adjustable is temporarily ignored.
     */
    class Adjustor extends AWTEvent implements java.awt.ActiveEvent {

	Adjustor(Adjustable adj, int value, MScrollPanePeer peer) {
	    super(adj, 0);
	    this.adj = adj;
	    this.value = value;
	    this.peer = peer;
	}

        public void dispatch() {
	    peer.ignore = true;
	    adj.setValue(value);
	    peer.ignore = false;
	}

	int value;
	Adjustable adj;
	MScrollPanePeer peer;
    }

    private Component getScrollChild() {
	ScrollPane sp = (ScrollPane)target;
	if (sp.countComponents() > 0) {
	    return sp.getComponent(0);
	}
	throw new NullPointerException("child is null");
    }

    final static int 	MARGIN = 1;
    final static int 	SCROLLBAR = 16;
    int hsbSpace;
    int vsbSpace;
    int vval;
    int hval;
    int vmax;
    int hmax;
    /*
     * Print the native component by rendering the Motif look ourselves.
     * ToDo(aim): needs to query native motif for more accurate size and
     * color information.
     */
    public void print(Graphics g) {
        ScrollPane sp = (ScrollPane)target;
	Dimension d = sp.size();
	Color bg = sp.getBackground();
	Color fg = sp.getForeground();
	Point p = sp.getScrollPosition();
	Component c = getScrollChild();
	Dimension cd = c.size();
	int sbDisplay = sp.getScrollbarDisplayPolicy();
	int vvis, hvis, vmin, hmin, vmax, hmax, vval, hval;

	switch (sbDisplay) {
	  case ScrollPane.SCROLLBARS_NEVER:
	    hsbSpace = vsbSpace = 0;
	    break;
	  case ScrollPane.SCROLLBARS_ALWAYS:
	    hsbSpace = vsbSpace = SCROLLBAR;
	    break;
	  case ScrollPane.SCROLLBARS_AS_NEEDED:
	    hsbSpace = (cd.width <= (d.width - 2*MARGIN)? 0 : SCROLLBAR);
	    vsbSpace = (cd.height <= (d.height - 2*MARGIN)? 0 : SCROLLBAR);

	    if (hsbSpace == 0 && vsbSpace != 0) {
		hsbSpace = (cd.width <= (d.width - SCROLLBAR - 2*MARGIN)? 0 : SCROLLBAR);
	    }
	    if (vsbSpace == 0 && hsbSpace != 0) {
		vsbSpace = (cd.height <= (d.height - SCROLLBAR - 2*MARGIN)? 0 : SCROLLBAR);
	    }
	}

        vvis = hvis = vmin = hmin = vmax = hmax = vval = hval = 0;

	if (vsbSpace > 0) {
	    vmin = 0;
	    vvis = d.height - (2*MARGIN) - hsbSpace;
	    vmax = Math.max(cd.height - vvis, 0);
	    vval = p.y;
	}
	if (hsbSpace > 0) {
	    hmin = 0;
	    hvis = d.width - (2*MARGIN) - vsbSpace;
	    hmax = Math.max(cd.width - hvis, 0);
	    hval = p.x;
	}

	// need to be careful to add the margins back in here because
	// we're drawing the margin border, after all!
	int w = d.width - vsbSpace;
	int h = d.height - hsbSpace;

	g.setColor(bg);
	g.fillRect(0, 0, d.width, d.height);

	if (hsbSpace > 0) {
	    int sbw = d.width - vsbSpace;
	    Graphics ng = g.create();
	    g.fillRect(1, d.height - SCROLLBAR - 3, sbw - 1, SCROLLBAR - 3);
	    try {
	        ng.translate(0, d.height - (SCROLLBAR - 2));
	        drawScrollbar(ng, bg, SCROLLBAR - 2, sbw,
			       hmin, hmax, hval, hvis, true);
	    } finally {
	        ng.dispose();
	    }
	}
	if (vsbSpace > 0) {
	    int sbh = d.height - hsbSpace;
	    Graphics ng = g.create();
	    g.fillRect(d.width - SCROLLBAR - 3, 1, SCROLLBAR - 3, sbh - 1);
	    try {
	        ng.translate(d.width - (SCROLLBAR - 2), 0);
	        drawScrollbar(ng, bg, SCROLLBAR - 2, sbh,
			       vmin, vmax, vval, vvis, false);
	    } finally {
	        ng.dispose();
	    }
	}

	draw3DRect(g, bg, 0, 0, w - 1, h - 1, false);

	target.print(g);
	sp.printComponents(g);

    }

}

