/*
 * @(#)MScrollbarPeer.java	1.20 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.AdjustmentEvent;

class MScrollbarPeer extends MComponentPeer implements ScrollbarPeer {
    native void create(MComponentPeer parent);
    
    MScrollbarPeer(Scrollbar target) {
	super(target);
    }

    public native void setValues(int value, int visible, int minimum, int maximum);    
    public native void setLineIncrement(int l);
    public native void setPageIncrement(int l);

    public Dimension getMinimumSize() {
	if (((Scrollbar)target).getOrientation() == Scrollbar.VERTICAL) {
	    return new Dimension(18, 50);
	} else {
	    return new Dimension(50, 18);
	}
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void lineUp(int value) {
	final Scrollbar sb = (Scrollbar)target;
	final int newValue = value;

	MToolkit.executeOnEventHandlerThread(sb, new Runnable() {
	    public void run() {
	        sb.setValue(newValue);
                postEvent(new AdjustmentEvent(sb, 
				AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
                                AdjustmentEvent.UNIT_DECREMENT, newValue));
	    } //run()
	});
    } // lineUp()

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void lineDown(int value) {
	final Scrollbar sb = (Scrollbar)target;
	final int newValue = value;

	MToolkit.executeOnEventHandlerThread(sb, new Runnable() {
	    public void run() {
	        sb.setValue(newValue);
                postEvent(new AdjustmentEvent(sb, 
				AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
                                AdjustmentEvent.UNIT_INCREMENT, newValue));
	    }
	});
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void pageUp(int value) {
	final Scrollbar sb = (Scrollbar)target;
	final int newValue = value;
	
	MToolkit.executeOnEventHandlerThread(sb, new Runnable() {
	    public void run() {
	        sb.setValue(newValue);
                postEvent(new AdjustmentEvent(sb, 
				AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
                                AdjustmentEvent.BLOCK_DECREMENT, newValue));
	    }
	});
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void pageDown(int value) {
	final Scrollbar sb = (Scrollbar)target;
	final int newValue = value;

	MToolkit.executeOnEventHandlerThread(sb, new Runnable() {
	    public void run() {
	        sb.setValue(newValue);
                postEvent(new AdjustmentEvent(sb, 
				AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
                                AdjustmentEvent.BLOCK_INCREMENT, newValue));
	    }
	});
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void dragAbsolute(int value) {
	final Scrollbar sb = (Scrollbar)target;
	final int newValue = value;

	MToolkit.executeOnEventHandlerThread(sb, new Runnable() {
	    public void run() {
	        sb.setValue(newValue);
                postEvent(new AdjustmentEvent(sb, 
				AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
                                AdjustmentEvent.TRACK, newValue));
	    }
	});
    } // dragAbsolute()

    public void print(Graphics g) {
        Scrollbar sb = (Scrollbar)target;
	Dimension d = sb.size();
	Color bg = sb.getBackground();

	boolean horiz = (sb.getOrientation() == Scrollbar.HORIZONTAL);

	drawScrollbar(g, bg, horiz? d.height : d.width, 
		          horiz? d.width : d.height, 
			  sb.getMinimum(), sb.getMaximum(),
			  sb.getValue(), sb.getVisible(),
			  horiz);
        
	target.print(g);
    }
		

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	    return getMinimumSize();
    }

}
