/*
 * @(#)MTextAreaPeer.java	1.34 98/08/31
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.TextEvent;
import java.awt.datatransfer.*;
import java.io.BufferedReader;
import java.io.StringReader;
import java.io.IOException;
import java.util.Vector;

public class MTextAreaPeer extends MComponentPeer implements TextAreaPeer {
    native void create(MComponentPeer parent);

    void initialize() {
        int start, end;

	TextArea txt = (TextArea)target;
	String	text;

	if ((text = txt.getText()) != null) {
	    setText(text);
	}

	start = txt.getSelectionStart();
	end = txt.getSelectionEnd();

	if (end > start) {
	  select(start, end);
	}

        super.pSetScrollbarBackground(getParent_NoClientCode(target).getBackground());

	if (isBackgroundInherited()) {
	    target.setBackground(SystemColor.text);
	}
	if (isForegroundInherited()) {
	    target.setForeground(SystemColor.textText);
	}
	setEditable(txt.isEditable());

	super.initialize();
    }

    public MTextAreaPeer(TextArea target) {
	super(target);
    }

    public void setEditable(boolean editable) {
	pSetEditable(editable);

	/* 4136955 - Calling setBackground() here works around an Xt
	 * bug by forcing Xt to flush an internal widget cache
	 */
	setBackground(target.getBackground());
    }
    public void setBackground(Color c) {
    	setTextBackground(c);
    }
    public void setForeground(Color c) {
    	pSetInnerForeground(c);
    }

    native int getExtraWidth();
    native int getExtraHeight();
    public native void setTextBackground(Color c);
    public native void pSetEditable(boolean e);
    public native void select(int selStart, int selEnd);
    public native int getSelectionStart();
    public native int getSelectionEnd();
    public native void setText(String txt);
    public native String getText();
    public native void insert(String txt, int pos);
    public native void replaceRange(String txt, int start, int end);
    public native void setFont(Font f);
    public native void setCaretPosition(int pos);
    public native int getCaretPosition();
    public native void setCursor(Cursor c);
    native void pShow2();
    native void pMakeCursorVisible();


    public Dimension getMinimumSize() {
	return getMinimumSize(10, 60);
    }
    public Dimension getPreferredSize(int rows, int cols) {
	return getMinimumSize(rows, cols);
    }
    public Dimension getMinimumSize(int rows, int cols) {
	FontMetrics fm = getFontMetrics(target.getFont());

	/* Calculate proper size for text area plus scrollbars.
	 *   - Motif allows NO leading in its text areas ...
	 *   - extra width and height counts everything outside the
	 *     usable text space.
	 */ 
	int colWidth = fm.charWidth('0');
	int rowHeight = fm.getHeight() - fm.getLeading();
	return new Dimension(cols * colWidth + getExtraWidth(),
			     rows * rowHeight + getExtraHeight());
    }
    public boolean isFocusTraversable() {
	return true;
    }

    // Called from native widget when paste key is pressed and we
    // already own the selection (prevents Motif from hanging while
    // waiting for the selection)
    //
    public void pasteFromClipboard() {
        Clipboard clipboard = target.getToolkit().getSystemClipboard();
        
	Transferable content = clipboard.getContents(this);
	if (content != null) {
	    try {
		String data = (String)(content.getTransferData(DataFlavor.stringFlavor));
                replaceRange(data, getSelectionStart(), getSelectionEnd());
                
	    } catch (Exception e) {       
	    }
        }
    }

    /*
     * Print the native component by rendering the Motif look ourselves.
     * ToDo(aim): needs to query native motif for more accurate size and
     * color information, the top/left text offsets, and selected text.
     */
    static final int MARGIN = 2;
    static final int BORDER = 1;
    static final int SCROLLBAR = 16;
    int fontHeight;
    int fontAscent;
    int fontLeading;
    int topLine = 0;
    int numLines = 0;
    int textLength = 0;
    Vector lines;
    int selStart = 0;
    int selEnd = 0;
    int movedRight = 0;

    public void print(Graphics g) {
        TextArea area = (TextArea)target;
	Dimension d = area.size();
	Color bg = area.getBackground();
	Color fg = area.getForeground();
	FontMetrics fm = getFontMetrics(area.getFont());
	int vmin, vmax, vval, vvis;
	int hmin, hmax, hval, hvis;
	int max = 0;

	/*
	  Doesn't work right yet.
	selStart = area.getSelectionStart();
	selEnd = area.getSelectionEnd();
	*/

	// Figure out number of lines and max line length
	String text = area.getText();
	textLength = text.length();
	BufferedReader is = new BufferedReader(new StringReader(text));
        String line;
	int pos = 0;
	lines = new Vector();
	try {
	    numLines = 0;
            while((line = is.readLine()) != null) {
	        TextLine tline = new TextLine();
		tline.text = line;
		tline.pos = pos;
		lines.addElement(tline);
		pos += line.length() + 1;

	        int len = fm.stringWidth(line);
	        max = Math.max(max, len);
	        numLines++;
	    }
            is.close();

	} catch (IOException e) {
	}

	int w = d.width -  ((2 * BORDER) + SCROLLBAR);
	int h = d.height - ((2 * BORDER) + SCROLLBAR);

	fontHeight = fm.getHeight();
	fontAscent = fm.getAscent();
	fontLeading = fm.getLeading();

	hmin = vmin = 0;

	vvis = linesInWindow(true);
	vmax = Math.max(numLines - vvis, 0);
	h = d.height - SCROLLBAR;
	vval = 0;

	w = d.width - SCROLLBAR;
	hvis = w - (2 * MARGIN);
	hmax = Math.max(max - hvis, 0);
	hval = 0;

	g.setColor(bg);
	g.fillRect(BORDER, BORDER, w, h);

	{
	    int sbh = d.height - SCROLLBAR;
	    Graphics ng = g.create();
	    g.fillRect(d.width - SCROLLBAR - 3, 1, SCROLLBAR - 3, sbh - 1);	
	    try {
	        ng.translate(d.width - (SCROLLBAR - 2), 0);
	        drawScrollbar(ng, bg, SCROLLBAR - 2, sbh, 
			      vmin, vmax, vval, vvis, false);
	    } finally {
	        ng.dispose();
	    }
        }
	{
	    int sbw = d.width - SCROLLBAR;
	    Graphics ng = g.create();
	    g.fillRect(1, d.height - SCROLLBAR - 3, sbw - 1, SCROLLBAR - 3);	
	    try {
	        ng.translate(0, d.height - (SCROLLBAR - 2));
	        drawScrollbar(ng, bg, SCROLLBAR - 2, sbw, 
			      hmin, hmax, hval, hvis, true);
	    } finally {
	        ng.dispose();
	    }
        }
	
	draw3DRect(g, bg, 0, 0, w - 1, h - 1, false);
	
	if (text != null) {
	    int l = linesInWindow(true);
	    h = d.height - ((2 * MARGIN) + SCROLLBAR);
	    int e = Math.min(numLines - 1, (topLine + l) - 1);
	    paintLines(g, bg, fg, topLine, e);
	}
       

	target.print(g);
    }

    int linesInWindow(boolean horizScrollbar) {
        Dimension d = target.size();
	int htotal = d.height - ((2 * MARGIN) + (horizScrollbar? SCROLLBAR : 0));
	return htotal / fontHeight;
    }

    void paintLines(Graphics g, Color bg, Color fg, int s, int e) {
        Dimension d = target.size();
	int w = d.width - ((2 * BORDER) + SCROLLBAR);
	int h = d.height - ((2 * BORDER) + SCROLLBAR);
	int lm = linesInWindow(true) + topLine;
	s = Math.max(topLine, s);
	e = Math.min(e, lm - 1);
	Graphics ng = g.create();
	ng.clipRect(BORDER + MARGIN, MARGIN + BORDER, w - (2*MARGIN), h - (2*MARGIN));
	ng.setFont(target.getFont());
	for (int i = s ; i <= e; i++) {
	    paintLine(ng, bg, fg, i);
	}     
    	ng.dispose();
    }

    void paintLine(Graphics g, Color bg, Color fg, int lnr) {
        Dimension d = target.size();
	int l = linesInWindow(true);

	if((lnr < topLine) || (lnr >= l + topLine)) { 
	    return;
	}
	int w = d.width - ((2 * BORDER) + SCROLLBAR); 
	int y = MARGIN + fontLeading + ((lnr - topLine) * fontHeight);
	String text = ((TextLine)lines.elementAt(lnr)).text;
	int len = text.length();
	
	if (lnr > numLines - 1) {
	    g.setColor(bg);
	    g.fillRect(BORDER, y - fontLeading, w, fontHeight);
	    return;
	}
	int s = 0;
	int e = (lnr < numLines - 1) ? len : textLength;
	int xs = pos2x(selStart) - movedRight;
	int xe = pos2x(selEnd) - movedRight;

	Color highlight = bg.brighter();
	if ((selStart < s) && (selEnd > e)) {
	    g.setColor(highlight);
	    g.fillRect(BORDER, y - fontLeading, w, fontHeight);
	} else {
	    g.setColor(bg);
	    g.fillRect(BORDER, y - fontLeading, w, fontHeight);
		
	    if ((selStart >= s) && (selStart <= e)) {
		g.setColor(highlight);
		    
		if (selEnd > e) {
		    g.fillRect(xs, y - fontLeading, (w + BORDER) - xs, fontHeight);
		} else if (selStart == selEnd) { 
		  //g.fillRect(xs, y - fontLeading, 1, fontHeight);
		} else { 
		    g.fillRect(xs, y - fontLeading, xe - xs, fontHeight);
		}		
	    } else if ((selEnd >= s) && (selEnd <= e)) {
		g.setColor(highlight);
		g.fillRect(BORDER, y - fontLeading, xe - BORDER, fontHeight);
	    }	    
	}
	g.setColor(fg);
	g.drawString(text, MARGIN - movedRight, y + fontAscent);
    } 

    int pos2x(int pos) {
	FontMetrics fm = getFontMetrics(target.getFont());	
	int widths[] = fm.getWidths();
	TextLine tl1 = (TextLine)lines.elementAt(0);
	TextLine tl2;
	int l = 0;
	for (int i = 0; i < lines.size() - 1; i++) {
	    tl2 = (TextLine)lines.elementAt(i+1);
	    if (pos >= tl1.pos && pos < tl2.pos) {
	        l = i;
		break;
	    }
	    tl1 = tl2;
	}
	int x = MARGIN;
	for (int i = 0 ; i < (pos - tl1.pos - 1) ; i++) {
	    x += widths[tl1.text.charAt(i)];
	}
	return x;
    }   

    /**
     * DEPRECATED
     */
    public void insertText(String txt, int pos) {
      	insert(txt, pos);
    }

    /**
     * DEPRECATED
     */
    public void replaceText(String txt, int start, int end) {
      	replaceRange(txt, start, end);
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * DEPRECATED
     */
    public Dimension preferredSize(int rows, int cols) {
    	return getPreferredSize(rows, cols);
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize(int rows, int cols) {
    	return getMinimumSize(rows, cols);
    }

    /*
     * Post a new TextEvent when the value of a text component changes.
     */
    public void valueChanged() {
        postEvent(new TextEvent(target, TextEvent.TEXT_VALUE_CHANGED));
    }

    void pShow(){
      pShow2();
      notifyTextComponentChange(true);
    }

    void pHide(){
      notifyTextComponentChange(false);
      super.pHide();
    }

    void pDispose(){
      notifyTextComponentChange(false);
      super.pDispose();
    }
}

class TextLine {
    String text;
    int pos;
}
