/*
 * @(#)MTinyChoicePeer.java	1.10 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.ItemEvent;

class MChoiceMenu extends Window {
    MTinyChoicePeer peer;
    String items[];
    int selected;
    int h;

    /**
     * Create a choice popup menu
     */
    MChoiceMenu(Frame f, MTinyChoicePeer peer, int x, int y, int w, int h, int selected, String items[]) {
	super(f);
	this.h = h;
	this.peer = peer;
	this.items = items;
	this.selected = selected;
	setBackground(f.getBackground());
	reshape(x - 2, (y - 2) - selected * h, w + 4, items.length * h + 4);
    }

    /**
     * Paint the popup menu
     */
    public void paint(Graphics g) {
	Dimension size = size();

	g.setColor(getBackground());
	g.draw3DRect(0, 0, size.width - 1, size.height - 1, true);
	g.draw3DRect(1, 1, size.width - 3, size.height - 3, true);
	g.draw3DRect(2, 2 + selected * h, size.width - 5, size.height - 1, false);

	g.setFont(getFont());
	FontMetrics fm = g.getFontMetrics();
	g.setColor(getForeground());
	int y = 2 + (h + fm.getAscent()) / 2;
	for (int i = 0 ; i < items.length ; i++, y += h) {
	    g.drawString(items[i], 7, y);
	}
    }

    /**
     * Handle events
     */
    boolean handleWindowEvent(Event evt) {
	int i = selected;
	Dimension size = size();
	if ((evt.x < 0) || (evt.x > size.width) || (evt.y < 0) || (evt.y > size.height)) {
	    i = ((Choice)peer.target).getSelectedIndex();
	} else {
	    i = Math.max(0, Math.min(items.length - 1, (evt.y - 2) / h));
	}
	if (selected != i) {
	    Graphics g = getGraphics();
	    try {
		g.setColor(getBackground());
		g.drawRect(2, 2 + selected * h, size.width - 5, h - 1);
		selected = i;
		g.draw3DRect(2, 2 + selected * h, size.width - 5, h - 1, true);
	    } finally {
		g.dispose();
	    }
	}
	return false;
    }
}

class MTinyChoicePeer extends MCanvasPeer implements ChoicePeer {
    MChoiceMenu menu;

    /**
     * Create the choice
     */
    public MTinyChoicePeer(Choice target) {
	super(target);
    }

    /**
     * Minimum size
     */
    public Dimension getMinimumSize() {
	FontMetrics fm = getFontMetrics(target.getFont());
	Choice c = (Choice)target;
	int w = 0;
	for (int i = c.countItems() ; i-- > 0 ;) {
	    w = Math.max(fm.stringWidth(c.getItem(i)), w);
	}
	return new Dimension(w + 10 + 18, fm.getHeight() + 10);
    }

    /**
     * Paint the choice
     */
    public void paint(Graphics g) {
	Choice ch = (Choice)target;
	Dimension size = target.size();

	g.setColor(Color.lightGray);
	g.fillRect(1, 1, size.width-2, size.height-2);
	g.draw3DRect(0, 0, size.width-1, size.height-1, true);
	g.draw3DRect(size.width - 18, (size.height / 2) - 3, 12, 6, true);

	g.setColor(Color.black);
	g.setFont(ch.getFont());
	FontMetrics fm = g.getFontMetrics();
	String lbl = ch.getSelectedItem();
	g.drawString(lbl, 5, (size.height + fm.getAscent()) / 2);
	target.paint(g);
    }	

    protected void handleMouseDown(long when, int data,  int x, int y,
				   int xroot, int yroot,
				   int clickCount, int flags) {
	if (!target.isEnabled()) {
	    return;
	}
	Dimension size = target.size();
	Choice ch = (Choice)target;
	String items[] = new String[ch.countItems()];
	for (int i = 0 ; i < items.length ; i++) {
	    items[i] = ch.getItem(i);
	}

	Component f;

	for (f=target; f != null && !(f instanceof Frame);
	     f=f.getParent()) {
	}
	menu = new MChoiceMenu((Frame)f, this, xroot, yroot,
			       size.width - 20, size.height, 
			       ch.getSelectedIndex(), items);
	menu.show();
    }

    protected void handleMouseUp(long when, int data, int x, int y,
				 int xroot, int yroot, int flags) {
	if (!target.isEnabled()) {
	    return;
	}
	Choice ch = (Choice)target;
	ch.select(menu.selected);
	menu.hide();
	menu.dispose();
	menu = null;
        postEvent(new ItemEvent(ch, ItemEvent.ITEM_STATE_CHANGED,
                                ch.getSelectedItem(), ItemEvent.SELECTED));
    }

    protected void handleMouseDrag(long when, int data, int x, int y,
				   int xroot, int yroot, int flags) {
	if (!target.isEnabled()) {
	    return;
	}
	Rectangle b = menu.bounds();

	Point pt = new Point(xroot - b.x, yroot - b.y);

	menu.handleWindowEvent(new Event(target, when, Event.MOUSE_DRAG,
					 pt.x, pt.y, 0, flags, null));
    }

    public void select(int index) {
	target.repaint();
    }

    public void add(String item, int index) {
	target.repaint();
    }

    public void remove(int index) {
      	System.out.println("Method remove(int) not implemented yet.");
    }

    public void removeAll() {
      	System.out.println("Method removeAll() not implemented yet.");
    }

    /**
     * DEPRECATED:  Replaced by getMinimumSize().
     */
    public Dimension minimumSize() {
	    return getMinimumSize();
    }

    /**
     * DEPRECATED:  Replaced by add(String, int).
     */
    public void addItem(String item, int index) {
	add(item, index);
    }
}
