/*
 * @(#)MToolkit.java	1.119 98/09/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.util.Properties;
import java.awt.*;
import java.awt.image.*;
import java.awt.peer.*;
import java.awt.datatransfer.Clipboard;
import java.awt.event.*;
import java.io.*;
import sun.awt.image.ImageRepresentation;
import sun.awt.AppContext;
import sun.awt.SunToolkit;
 
import java.awt.dnd.DragSource;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.dnd.MouseDragGestureRecognizer;
import java.awt.dnd.InvalidDnDOperationException;
import java.awt.dnd.peer.DragSourceContextPeer;

import sun.awt.im.InputMethod;
import sun.awt.im.InputMethodHighlightMapping;
import sun.awt.motif.X11InputMethod;
import sun.awt.X11GraphicsConfig;

import sun.java2d.loops.RasterOutputManager;

import sun.awt.motif.MDragSourceContextPeer;

public class MToolkit extends SunToolkit implements Runnable {

    final static boolean useMotifChoiceWidget = true;
    
    // the system Clipboard
    X11Clipboard clipboard;

    /*
     * Note: The MToolkit object depends on the static initializer
     * of X11GraphicsEnvironment to initialize the connection to
     * the X11 server.
     */
    static X11GraphicsConfig config =
	(X11GraphicsConfig) (GraphicsEnvironment.
			     getLocalGraphicsEnvironment().
			     getDefaultScreenDevice().
			     getDefaultConfiguration());

    public MToolkit() {
        super();
	init();
//
// Move to a single thread XInput system, so
// eliminate the creation of the input thread: note 
// the Input Thread class could
// be deleted
//	new InputThread();
//
	new Thread(this, "AWT-Motif").start();
    }

    public native void init();
    public native void run();

    /*
     * Create peer objects.
     */

    public ButtonPeer createButton(Button target) {
	ButtonPeer peer = new MButtonPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public TextFieldPeer createTextField(TextField target) {
	TextFieldPeer peer = new MTextFieldPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public LabelPeer createLabel(Label target) {
	LabelPeer peer = new MLabelPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ListPeer createList(List target) {
	ListPeer peer = new MListPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public CheckboxPeer createCheckbox(Checkbox target) {
	CheckboxPeer peer = new MCheckboxPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ScrollbarPeer createScrollbar(Scrollbar target) {
	ScrollbarPeer peer = new MScrollbarPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ScrollPanePeer createScrollPane(ScrollPane target) {
	ScrollPanePeer peer = new MScrollPanePeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public TextAreaPeer createTextArea(TextArea target) {
	TextAreaPeer peer = new MTextAreaPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ChoicePeer createChoice(Choice target) {
	ChoicePeer peer = null;
	if (useMotifChoiceWidget) {
	    peer = new MChoicePeer(target);
	} else {
	    peer = new MTinyChoicePeer(target);
	}
	peerMap.put(target, peer);
	return peer;
    }

    public FramePeer  createFrame(Frame target) {
	FramePeer peer = new MFramePeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public CanvasPeer createCanvas(Canvas target) {
	CanvasPeer peer = new MCanvasPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public PanelPeer createPanel(Panel target) {
	PanelPeer peer = new MPanelPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public WindowPeer createWindow(Window target) {
	WindowPeer peer = new MWindowPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public DialogPeer createDialog(Dialog target) {
	DialogPeer peer = new MDialogPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public FileDialogPeer createFileDialog(FileDialog target) {
	FileDialogPeer peer = new MFileDialogPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public MenuBarPeer createMenuBar(MenuBar target) {
	MenuBarPeer peer = new MMenuBarPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public MenuPeer createMenu(Menu target) {
	MenuPeer peer = new MMenuPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public PopupMenuPeer createPopupMenu(PopupMenu target) {
	PopupMenuPeer peer = new MPopupMenuPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public MenuItemPeer createMenuItem(MenuItem target) {
	MenuItemPeer peer = new MMenuItemPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public CheckboxMenuItemPeer createCheckboxMenuItem(CheckboxMenuItem target) {
	CheckboxMenuItemPeer peer = new MCheckboxMenuItemPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public MEmbeddedFramePeer createEmbeddedFrame(MEmbeddedFrame target, int handle)
    {
        MEmbeddedFramePeer peer = new MEmbeddedFramePeer(target, handle);
        peerMap.put(target, peer);
        return peer;
    }


    public FontPeer getFontPeer(String name, int style){
	/*----------*
        if (!RasterOutputManager.usesPlatformFont()) {
	    return new FontPeer() {};
        } else {
            return new MFontPeer(name, style);
        }
	*----------*/
	return new MFontPeer(name, style);
    }

    static native ColorModel makeColorModel();
    static ColorModel screenmodel;

    static ColorModel getStaticColorModel() {
	if (screenmodel == null) {
	    screenmodel = config.getColorModel ();
	}
	return screenmodel;
    }

    public ColorModel getColorModel() {
	return getStaticColorModel();
    }

    public native int getScreenResolution();
    protected native int getScreenWidth();
    protected native int getScreenHeight();

    public FontMetrics getFontMetrics(Font font) {
        if (!RasterOutputManager.usesPlatformFont()) {
            return super.getFontMetrics(font);
        } else {
            return X11FontMetrics.getFontMetrics(font);
        }
    }

    public native void sync();

    static boolean prepareScrImage(Image img, int w, int h, ImageObserver o) {
	if (w == 0 || h == 0) {
	    return true;
	}

        // Must be an OffScreenImage
        if (!(img instanceof X11Image)) {
            return true;
        }
            
	X11Image ximg = (X11Image) img;
	if (ximg.hasError()) {
	    if (o != null) {
		o.imageUpdate(img, ImageObserver.ERROR|ImageObserver.ABORT,
			      -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep();
	return ir.prepare(o);
    }

    static int checkScrImage(Image img, int w, int h, ImageObserver o) {

        if (!(img instanceof X11Image)) {
            return ImageObserver.ALLBITS;
        }
            
	X11Image ximg = (X11Image) img;
	int repbits;
	if (w == 0 || h == 0) {
	    repbits = ImageObserver.ALLBITS;
	} else {
	    repbits = ximg.getImageRep().check(o);
	}
	return ximg.check(o) | repbits;
    }

    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return checkScrImage(img, w, h, o);
    }

    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return prepareScrImage(img, w, h, o);
    }

    public Image createImage(ImageProducer producer) {
	return new X11Image(producer);
    }

    public PrintJob getPrintJob(Frame frame, String title, Properties props) {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkPrintJobAccess();
	}
	PSPrintControl printControl = new PSPrintControl(frame);
	return printControl.initJob(title, props, true);
    }

    public native void beep();

    public  Clipboard getSystemClipboard() {
        SecurityManager security = System.getSecurityManager();
        if (security != null) {
	  security.checkSystemClipboardAccess();
	}
        if (clipboard == null) {
	    clipboard = new X11Clipboard("System");
	}
	return clipboard;
    }

    public native void loadSystemColors(int[] systemColors);

    /**
     * Give native peers the ability to query the native container 
     * given a native component (eg the direct parent may be lightweight).
     */
    public static Container getNativeContainer(Component c) {
	return Toolkit.getNativeContainer(c);
    }

    public static final Object targetToPeer(Object target) {
        return SunToolkit.targetToPeer(target);
    }

    public static final void targetDisposedPeer(Object target, Object peer) {
        SunToolkit.targetDisposedPeer(target, peer);
    }

    public DragSourceContextPeer createDragSourceContextPeer(DragGestureEvent dge) throws InvalidDnDOperationException {
	return MDragSourceContextPeer.getDragSourceContextPeer(dge);
    }

    public DragGestureRecognizer createDragGestureRecognizer(Class abstractRecognizerClass, DragSource ds, Component c, int srcActions, DragGestureListener dgl) {
	if (MouseDragGestureRecognizer.class.equals(abstractRecognizerClass))
	    return new MMouseDragGestureRecognizer(ds, c, srcActions, dgl);
	else
            return null;
    }


    /**
     * Returns a new input method adapter for native input methods.
     */
    public InputMethod getInputMethodAdapter() throws AWTException {
	return new X11InputMethod();
    }

    /**
     * Returns the input method highlight mapping interface.
     */
    public InputMethodHighlightMapping getInputMethodHighlightMapping() {
	return X11InputMethod.getInputMethodHighlightMapping();
    }

    /**
     * Returns a new custom cursor.
     */
    public Cursor createCustomCursor(Image cursor, Point hotSpot, String name)
        throws IndexOutOfBoundsException {
        return new X11CustomCursor(cursor, hotSpot, name);
    }

    /**
     * Returns the supported cursor size
     */
    public Dimension getBestCursorSize(int preferredWidth, int preferredHeight) {
        return X11CustomCursor.getBestCursorSize(
            preferredWidth, preferredHeight);
    }

    public int getMaximumCursorColors() {
        return 2;  // Black and white.
    }

    private final static String prefix  = "DnD.Cursor.";
    private final static String postfix = ".32x32";

    protected Object lazilyLoadDesktopProperty(String name) {
	if (name.startsWith(prefix)) {
	    String cursorName = name.substring(prefix.length(), name.length()) + postfix;

	    try {
	        return Cursor.getSystemCustomCursor(cursorName);
	    } catch (AWTException awte) {
	 	System.err.println("cannot load system cursor: " + cursorName);
	
		return null;
	    }
	}

	return super.lazilyLoadDesktopProperty(name);
    }

    protected void initializeDesktopProperties() {
	desktopProperties.put("DnD.Autoscroll.initialDelay",     new Integer(50));
	desktopProperties.put("DnD.Autoscroll.interval",         new Integer(50));
	desktopProperties.put("DnD.Autoscroll.cursorHysteresis", new Integer(5));
    }
}
