/*
 * @(#)MWindowPeer.java	1.21 98/08/31
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.motif;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;
import java.awt.event.*;

class MWindowPeer extends MPanelPeer implements WindowPeer {
    Insets insets = new Insets(0, 0, 0, 0);

    static Vector allWindows = new Vector();

    native void create(MComponentPeer parent);
    native void pShow();
    native void pHide();
    native void pReshape(int x, int y, int width, int height);
    native void pDispose();
    private static native void initIDs();

    static {
        initIDs();
    }

    MWindowPeer(Window target) {
	init(target);
	allWindows.addElement(this);
	Font f = target.getFont();
	if (f == null) {
	    f = new Font("Dialog", Font.PLAIN, 12);
	    target.setFont(f);
	    setFont(f);
	}
	Color c = target.getBackground();
	if (c == null) {
	    target.setBackground(SystemColor.window);
	    setBackground(SystemColor.window);
	}
	c = target.getForeground();
	if (c == null) {
	    target.setForeground(SystemColor.windowText);
	    setForeground(SystemColor.windowText);
	}
    }

    protected void disposeImpl() {
	allWindows.removeElement(this);
	super.disposeImpl();
    }

    public void toFront() {
	pShow();
    }
    public native void toBack();

    /* Handle TAB and Shift-TAB events. */
    public int handleFocusTraversalEvent(KeyEvent e) {

	if (e.getKeyCode() != KeyEvent.VK_TAB ||
	    (e.getSource() instanceof TextArea))
	    return IGNORE_EVENT;

 	if ((e.getModifiers() & ~InputEvent.SHIFT_MASK) > 0)
 	    return IGNORE_EVENT;

	int id = e.getID();
 	if (id == KeyEvent.KEY_RELEASED || id == KeyEvent.KEY_TYPED)
 	    return CONSUME_EVENT;

 	if (e.isShiftDown())
 	    return FOCUS_PREVIOUS;
	else
 	    return FOCUS_NEXT;
    }

    public Insets getInsets() {
	return insets;
    }

    public void handleQuit() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_CLOSING));
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleIconify() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_ICONIFIED));
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public void handleDeiconify() {
        postEvent(new WindowEvent((Window)target, WindowEvent.WINDOW_DEICONIFIED));
    }

    /**
     * Called to inform the Dialog that its size has changed and it
     * should layout its children.
     */
    public synchronized void handleResize(int width, int height) {
        postEvent(new ComponentEvent(target, ComponentEvent.COMPONENT_RESIZED));
    }

    /**
     * DEPRECATED:  Replaced by getInsets().
     */
    public Insets insets() {
	return getInsets();
    }

}

