/*
 * @(#)PSPrintJob.java	1.18 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.*;
import java.io.*;

/**
 * A class which initiates and executes a PostScript print job.
 *
 * @version 	1.18 06/29/98
 * @author 	Amy Fowler
 * @author 	Richard Berlin
 */
public class PSPrintJob extends PrintJob {
    private PrintStream		os;
    private PSPrintStream	stream;
    private PSPrintControl      control;
    private boolean             endInvoked = false;

    String  	 title;
    Dimension 	 pageDimension;
    int		 pageResolution;
    boolean 	 lastFirst = false;
    int		 pageCount = 0;

    /**
     * Creates a print job configured to pipe PostScript output
     * to a PrintStream.
     */
    public PSPrintJob(PrintStream ps, String title, Dimension pageDim,
		      int pageRes) {
        this(ps, title, pageDim, pageRes, null);
    }

    /**
     * Creates a print job configured to pipe PostScript output
     * to a PrintStream. The print job will call back on the PSPrintControl
     * when it has finished printing.
     */
    public PSPrintJob(PrintStream ps, String title, Dimension pageDim,
		      int pageRes, PSPrintControl control) {
	this.title = title;
	pageDimension = pageDim;
	pageResolution = pageRes;
	os = ps;
	this.control = control;
	stream = new PSPrintStream(this, os, title, pageDim, lastFirst);
	stream.sendProlog();
    }

    /**
     * Gets a Graphics object that will draw to the next page.
     * The page is flushed to the printer when the graphics
     * object is disposed.
     */
    public Graphics getGraphics() {
	stream.startPage(++pageCount);
	return new PageGraphics(stream, pageCount);
    }

    /**
     * Returns the dimensions of the page in pixels.
     */
    public Dimension getPageDimension() {
	return pageDimension;
    }

    /**
     * Returns the resolution of the page in pixels per inch.
     */
    public int getPageResolution() {
	return pageResolution;
    }

    /**
     * Returns whether the last page will be printed first.
     */
    public boolean lastPageFirst() {
	return lastFirst;
    }

    /**
     * Ends the print job and does cleanup.
     */
    public void end() {
        if (!endInvoked) {
	    stream.close(pageCount);
	    os.close();
	    os = null;

	    if (control != null) {
	        control.finishedPrinting(this);
	    }

	    endInvoked = true;
	}
    }
}

class PageGraphics extends PSGraphics {
    public PageGraphics(PSPrintStream stream, int pageNumber) {
	super(stream, pageNumber);
    }

    protected void disposeImpl() {
        stream.endPage();
	super.disposeImpl();
    }
}
