/*
 * @(#)X11Clipboard.java	1.7 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.datatransfer.*;
import java.io.IOException;

/**
 * A class which interfaces with the X11 selection service in order
 * to support data transfer via clipboard operations.
 */
public class X11Clipboard extends Clipboard implements X11SelectionHolder {

    private X11Selection clipboardSelection = new X11Selection("CLIPBOARD");

    /**
     * Creates a system clipboard object.
     */
    public X11Clipboard(String name) {
        super(name);
	contents = clipboardSelection;	
    }

    public synchronized void setContents(Transferable contents, ClipboardOwner owner) {
	if (this.owner != null) {
	    this.owner.lostOwnership(this, this.contents);
	}

	if (clipboardSelection.getSelectionOwnership(contents, this)) {
	    this.owner = owner;
	    this.contents = contents;
	} else {
	    // Need to figure out how to inform owner the request failed...
	    this.owner = null;
	    this.contents = clipboardSelection;
	}
    }

    public synchronized Transferable getContents(Object requestor) {
        return contents;
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public synchronized void lostSelectionOwnership() {
        if (this.owner != null) {
	    this.owner.lostOwnership(this, this.contents);
	    this.owner = null;
	    this.contents = clipboardSelection;
	}
    }
}


