/*
 * @(#)X11CustomCursor.java	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import sun.awt.CustomCursor;
import java.awt.*;
import java.awt.image.*;
import sun.awt.image.ImageRepresentation;

/**
 * A class to encapsulate a custom image-based cursor.  
 *
 * @see Component#setCursor
 * @version 	1.7 09/21/98
 * @author 	Thomas Ball
 */
public class X11CustomCursor extends CustomCursor {

    long pData;

    static {
        initIDs();
    }
  
    /**
     * Initialize JNI field and method IDs for fields that may be
       accessed from C.
     */
    private static native void initIDs();    

    public X11CustomCursor(Image cursor, Point hotSpot, String name) 
            throws IndexOutOfBoundsException {
        super(cursor, hotSpot, name);
    }

    /**
     * Returns the supported cursor size
     */
    public static Dimension getBestCursorSize(
        int preferredWidth, int preferredHeight) {
        Dimension d = new Dimension(preferredWidth, preferredHeight);
        queryBestCursor(d);
        return d;
    }

    private static native void queryBestCursor(Dimension d);

    protected void createNativeCursor(int[] pixels, int width, int height,
                                      int xHotSpot, int yHotSpot) {
	// On Solaris 2.5.x, the above code for cursor of any size runs fine
        // but on Solaris 2.6, the width of a cursor has to be 8 divisible, 
        //   otherwise, the cursor could be displayed as garbaged.
  	// To work around the 2.6 problem, the following code pads any cursor 
	//   with a transparent area to make a new cursor of width 8 multiples.
	// --- Bug 4148455
	int wNByte = (width + 7)/8;
	int tNByte = wNByte * height;
        byte[] xorMask = new byte[tNByte];
        byte[] andMask = new byte[tNByte];

	for (int i = 0; i < width; i++) {
	    for (int j = 0; j < height; j++) {
	        int ip = j*width + i;
		int ibyte = j*wNByte + i/8;

                int omask = 1 << (i % 8);
                if ((pixels[ip] & 0xff000000) != 0) {
                    andMask[ibyte] |= omask;
                }
                if ((pixels[ip] & 0xffffff) == 0) {
                    xorMask[ibyte] |= omask;
	        }
	    }
	}
        createCursor(xorMask, andMask, 8*wNByte, height, xHotSpot, yHotSpot);
    }

    private native void createCursor(byte[] xorMask, byte[] andMask, 
                                     int width, int height, 
                                     int xHotSpot, int yHotSpot);

    protected native void finalize() throws Throwable;

    static {
        cacheInit();
    }

    private native static void cacheInit();
}
