/*
 * @(#)X11Selection.java	1.13 98/09/09
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.motif;

import java.awt.datatransfer.*;
import java.io.*;
import java.util.Vector;

/*
 * Implements a general interface to the X11 selection mechanism.
 * This class contains the intelligence to do the format translation
 * between java DataFlavor namespace and the X11 TARGET namespace.
 */
public class X11Selection implements Transferable {

    static Vector selections;
    static Vector targetMap;

    int atom;

    X11SelectionHolder holder;
    Transferable contents;
    String dataString;

    int targetArray[];
    Object data;

    private boolean disposed = false;

    static {
        initIDs();
        init();

	selections = new Vector();

        // for converting flavor to targets
        registerTargetForFlavor(DataFlavor.stringFlavor, "STRING");
	registerTargetForFlavor(DataFlavor.plainTextFlavor, "STRING");

	// for converting target to flavors
	Target t;
	targetMap = new Vector();

	t = new Target("STRING");
	t.addFlavorMap(DataFlavor.stringFlavor);
	t.addFlavorMap(DataFlavor.plainTextFlavor);
	targetMap.addElement(t);

	/*
	t = new Target("TEXT");
	t.addFlavorMap(DataFlavor.stringFlavor);
	t.addFlavorMap(DataFlavor.plainTextFlavor);
	targetMap.addElement(t);
	*/

	t = new Target("COMPOUND_TEXT");
	t.addFlavorMap(DataFlavor.stringFlavor);
	t.addFlavorMap(DataFlavor.plainTextFlavor);
	targetMap.addElement(t);
    }

    private static native void initIDs();
    static native void init();
    static native int getAtomForTarget(String name);
    static native void registerTargetForFlavor(DataFlavor flavor, String target);

    /* Translates a list of X11 selection TARGET atoms into the
     * corresponding list of flavors (not necessarily a one-to-one
     * mapping since different TARGETS may map to the same flavor).
     */
    static DataFlavor[] convertTargetsToFlavors(int[] targets) {
        Vector flavorBuf = new Vector();

        for (int i = 0; i < targets.length; i++) {
	    for (int j = 0; j < targetMap.size(); j++) {
	        Target t = (Target)targetMap.elementAt(j);
	        if (t.atom == targets[i]) {
		    for (int k = 0; k < t.flavors.size(); k++) {
			DataFlavor f = (DataFlavor)t.flavors.elementAt(k);
			boolean exists = false;
		        for (int l = 0; l < flavorBuf.size(); l++) {
			    if (f.equals((DataFlavor)flavorBuf.elementAt(l))) {
			        exists = true;
				break;
			    }
			}
			if (!exists) {
			    flavorBuf.addElement(f);
			}
		    }
		}
		break;
	    }
	}
	DataFlavor flavors[] = new DataFlavor[flavorBuf.size()];
	for (int i = 0; i < flavorBuf.size(); i++) {
	    flavors[i] = (DataFlavor)flavorBuf.elementAt(i);
	    /*
	    System.out.println("flavor: "+flavors[i].getHumanPresentableName());
	    */
	}
	return flavors;
    }

   /* Translates a flavor into a list of the corresponding selection TARGET
    * atoms.
    */
    static int[] convertFlavorToTargets(DataFlavor flavor) {
        int atomBuf[] = new int[24];
	int atomCount = 0;

	for (int i = 0; i < targetMap.size(); i++) {
	    Target t = (Target)targetMap.elementAt(i);
	    for (int j = 0; j < t.flavors.size(); j++) {
	        DataFlavor f = (DataFlavor)t.flavors.elementAt(j);
		if (f.equals(flavor)) {
		    atomBuf[atomCount++] = t.atom;
		    break;
		}
	    }
	}
	int atoms[] = new int[atomCount];
	for (int i = 0; i < atomCount; i++) {
	    atoms[i] = atomBuf[i];
	}
	return atoms;
    }

    public X11Selection(String name) {
        create(name);
        selections.addElement(this);
    }

    protected void finalize() throws Throwable {
        dispose();
	super.finalize();
    }

    native void create(String name);
    native void pDispose();

   /*
    * methods for acting as selection provider
    */
    native boolean pGetSelectionOwnership();

    boolean getSelectionOwnership(Transferable contents, X11SelectionHolder holder) {
        this.holder = holder;
        this.contents = contents;
	if (pGetSelectionOwnership()) {
	    return true;
	}
	this.holder = null;
	this.contents = null;
	return false;
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    void lostSelectionOwnership() {
        holder.lostSelectionOwnership();
	holder = null;
	contents = null;
    }

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public String getStringDataFromOwner(DataFlavor flavor) {
        if (contents != null) {
	    Class repClass = flavor.getRepresentationClass();
	    try {
	        if (repClass == Class.forName("java.lang.String")) {
		    dataString = (String)contents.getTransferData(flavor);
		    return dataString;
	        }
	    } catch (ClassNotFoundException e) {
	        // not likely
	    } catch (Exception e) {

	    }
	}
	return null;
    }

   /*
    * methods for acting as selection consumer
    */
    native void pGetTransferTargets();
    native void pGetTransferData(int atoms[]);

    // NOTE: This method may be called by privileged threads.
    //       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
    public DataFlavor[] getTransferDataFlavors() {
        targetArray = null;

	SelectionThread st = (SelectionThread)
	    java.security.AccessController.doPrivileged(
					new java.security.PrivilegedAction() {
		public Object run() {
		    return new SelectionThread(X11Selection.this, null);
		}
	    });
	synchronized(st) {
	    try {
	        st.start();
		st.wait();
	    } catch (InterruptedException e) {

	    }
	}
	if (targetArray == null || targetArray.length == 0) {
	    return null;
	} else {
	  /*
	    for (int i = 0; i < targetArray.length; i++) {
	        System.out.println("Got target: "+targetArray[i]);
	    }
	    */
	    return convertTargetsToFlavors(targetArray);
        }
    }

    public boolean isDataFlavorSupported(DataFlavor flavor) {
        DataFlavor[] flavors = getTransferDataFlavors();
	if (flavors == null)
	   return false;

	for (int i = 0; i < flavors.length; i++) {
	    if (flavors[i].equals(flavor)) {
	        return true;
	    }
	}
        return false;
    }

    /* Retrieves the X11 selection value for this data flavor
     * (currently only supports string/text flavors)
     */
    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
        data = null;
        final int targets[] = X11Selection.convertFlavorToTargets(flavor);
	if (targets.length == 0) {
	    throw new UnsupportedFlavorException(flavor);
	}
	SelectionThread st = (SelectionThread)
	    java.security.AccessController.doPrivileged(
					new java.security.PrivilegedAction() {
		public Object run() {
		    return new SelectionThread(X11Selection.this, targets);
		}
	    });
	synchronized(st) {
	    try {
	        st.start();
		st.wait();
	    } catch (InterruptedException e) {

	    }
	}
	if (data == null) {
	    throw new IOException("could not get transfer data");
	} else {
	    if (flavor.equals(DataFlavor.plainTextFlavor)) {
	        return (InputStream) new StringBufferInputStream((String)data);

	    } else {
	        Class repClass = flavor.getRepresentationClass();
	        try {
	            if (repClass == Class.forName("java.io.InputStream")) {
		        throw new IOException("can't deal with InputStream yet");
                    } else {
		        return (Object)data;
		    }

	        } catch (ClassNotFoundException e) {
	            // not likely
		    throw new IOException("couldn't get representation class");
	        }
	    }
        }
    }

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     */
    protected void disposeImpl() {
        selections.removeElement(this);
        pDispose();
    }

    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
            synchronized (this) {
                if (!disposed) {
                    disposed = call_disposeImpl = true;
                }
            }
        }

        if (call_disposeImpl) {
            disposeImpl();
        }
    }
}

/*
 * Implements the thread required to deal with X11's asynchronous
 * selection API.  This thread will be responsible for requesting
 * the selection data and not returning until that data has been
 * provided to a native selection callback.
 */
class SelectionThread extends Thread {
    X11Selection s;
    int[] targets;

    public SelectionThread(X11Selection s, int[] targets) {
	super("AWT-Selection");
	this.s = s;
	this.targets = targets;
    }
    public void run() {
	synchronized(this) {
	    if (targets == null) {
	        s.pGetTransferTargets();
	    } else {
	        s.pGetTransferData(targets);
	    }
	    notifyAll();
	}
    }
}

/*
 * Implements a representation for an X11 selection TARGET type and
 * the set of flavors it maps to.
 */
class Target {

    String name;
    int atom;
    Vector flavors;

    public Target(String name) {
        this.name = name;
	this.atom = X11Selection.getAtomForTarget(name);
	flavors = new Vector();
    }

    public void addFlavorMap(DataFlavor flavor) {
        for (int i = 0; i < flavors.size(); i++) {
	    if (((DataFlavor)flavors.elementAt(i)).equals(flavor)) {
	        return; // already exists
	    }
	}
        flavors.addElement(flavor);
    }
}





