/*
 * @(#)TinyButtonPeer.java	1.9 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.awt.peer.*;

class TinyButtonPeer extends TinyComponentPeer implements ButtonPeer {
    boolean down;
    boolean hasFocus;

    /**
     * Create a button.
     */
    TinyButtonPeer(Button target) {
	winCreate((TinyComponentPeer)target.getParent().getPeer());
	init(target);
    }

    /**
     * Minimum size.
     */
    public Dimension getMinimumSize() {
	Button but = (Button)target;
	FontMetrics fm = getFontMetrics(but.getFont());
	return new Dimension(fm.stringWidth(but.getLabel()) + TEXT_XPAD, 
			     fm.getMaxAscent() + fm.getMaxDescent() + TEXT_YPAD);
    }

    /**
     * Set the label.
     */
    public void setLabel(String lbl) {
	repaint();
    }

    /**
     * Paint the button.
     */
    public void paintComponent(Graphics g) {
	Button but = (Button)target;
	Color colors[] = getGUIcolors();

	g.setColor(down ? colors[SHADOW_COLOR] : colors[BACKGROUND_COLOR]);
	g.fillRect(1, 1, width - 2, height - 2);
	g.setColor(colors[BACKGROUND_COLOR]);
	draw3DRect(g, colors, 0, 0, width - 1, height - 1, !down);

	g.setColor(colors[FOREGROUND_COLOR]);
	g.setFont(but.getFont());
	if (hasFocus) {
	    g.drawRect(2, 2, width - 4, height - 4);
	}

	FontMetrics fm = g.getFontMetrics();
	String lbl = but.getLabel();
	g.drawString(lbl, (width - fm.stringWidth(lbl)) / 2, 
		          (height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
    }	

    /**
     * Handle events.
     */
    public boolean handleWindowEvent(Event evt) {
	if (! isEnabled()) {
	    return false;
	}

	switch (evt.id) {
	case Event.MOUSE_DOWN:
	    down = true;
	    repaint();
	    return true;

	case Event.MOUSE_DRAG:
	    if ((evt.x < 0) || (evt.x > width) ||
		(evt.y < 0) || (evt.y > height)) {
		if (down) {
		    down = false;
		    repaint();
		}
	    } else if (!down) {
		down = true;
		repaint();
	    }
	    return true;
	    
	case Event.MOUSE_UP:
	    if (down) {
		targetEvent(Event.ACTION_EVENT, ((Button)target).getLabel());
		down = false;
		repaint();
	    }
	    return true;

	case Event.KEY_PRESS:
	    if (evt.key == '\n' || evt.key == '\r') {

		down = true;
		repaint();
		targetEvent(Event.ACTION_EVENT, ((Button)target).getLabel());
		down = false;
		repaint();
		return true;
	    } else if (evt.key == '\t') {
		target.nextFocus();
		return true;
	    }

	case Event.GOT_FOCUS:
	    hasFocus = true;
	    repaint();
	    return true;

	case Event.LOST_FOCUS:
	    hasFocus = false;
	    repaint();
	    return true;
	}

	return false;
    }

    public void requestFocus() {
	requestFocus(this);
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

}
