/*
 * @(#)TinyChoicePeer.java	1.10 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.awt.peer.*;

class TinyChoiceMenu extends TinyWindow {
    TinyChoicePeer peer;
    String items[];
    int selected;
    int h;

    /**
     * Create a choice popup menu
     */
    TinyChoiceMenu(TinyChoicePeer peer, int x, int y, int w, int h, int selected, String items[]) {
	this.h = h;
	this.peer = peer;
	this.items = items;
	this.selected = selected;
	winCreate(null);
	winBackground(peer.target.getBackground());
	winReshape(x - 2, (y - 2) - selected * h, w + 4, items.length * h + 4);
    }

    /**
     * Paint the popup menu
     */
    public void paint(Graphics g) {
	Color colors[] = peer.getGUIcolors();
	peer.draw3DRect(g, colors, 0, 0, width - 1, height - 1, true);
	peer.draw3DRect(g, colors, 1, 1, width - 3, height - 3, true);
	peer.draw3DRect(g, colors, 2, 2 + selected * h, width - 5, h - 1, false);

	g.setFont(peer.target.getFont());
	FontMetrics fm = g.getFontMetrics();
	g.setColor(peer.target.getForeground());
	int y = 2 + (h + fm.getMaxAscent() - fm.getMaxDescent()) / 2;
	for (int i = 0 ; i < items.length ; i++, y += h) {
	    g.drawString(items[i], 7, y);
	}
    }

    /**
     * Handle events
     */
    boolean handleWindowEvent(Event evt) {
	int i = selected;
	if ((evt.x < 0) || (evt.x > width) || (evt.y < 0) || (evt.y > height)) {
	    i = ((Choice)peer.target).getSelectedIndex();
	} else {
	    i = Math.max(0, Math.min(items.length - 1, (evt.y - 2) / h));
	}
	if (selected != i) {
	    Graphics g = new TinyGraphics(this);
	    Color colors[] = peer.getGUIcolors();
	    try {
		g.setColor(colors[peer.BACKGROUND_COLOR]);
		g.drawRect(2, 2 + selected * h, width - 5, h - 1);
		selected = i;
		peer.draw3DRect(g, colors,
				2, 2 + selected * h, width - 5, h - 1, false);
	    } finally {
		g.dispose();
	    }
	}
	return true;
    }
}

class TinyChoicePeer extends TinyComponentPeer implements ChoicePeer {
    TinyChoiceMenu menu;

    /**
     * Create the choice
     */
    public TinyChoicePeer(Choice target) {
	winCreate((TinyComponentPeer)target.getParent().getPeer());
	init(target);
    }

    /**
     * Minimum size
     */
    public Dimension getMinimumSize() {
	FontMetrics fm = getFontMetrics(target.getFont());
	Choice c = (Choice)target;
	int w = 0;
	for (int i = c.countItems() ; i-- > 0 ;) {
	    w = Math.max(fm.stringWidth(c.getItem(i)), w);
	}
	return new Dimension(w + TEXT_XPAD + 18, 
			     fm.getMaxAscent() + fm.getMaxDescent() + TEXT_YPAD);
    }

    /**
     * Paint the choice
     */
    public void paintComponent(Graphics g) {
	Choice ch = (Choice)target;
	Color colors[] = getGUIcolors();

	g.setColor(colors[BACKGROUND_COLOR]);
	g.fillRect(1, 1, width-2, height-2);
	draw3DRect(g, colors, 0, 0, width-1, height-1, true);
	draw3DRect(g, colors, width - 18, (height / 2) - 3, 12, 6, true);

	g.setColor(target.getForeground());
	g.setFont(ch.getFont());
	FontMetrics fm = g.getFontMetrics();
	String lbl = ch.getSelectedItem();
	g.drawString(lbl, 5, (height + fm.getMaxAscent()-fm.getMaxDescent())/2);
	target.paint(g);
    }	

    /**
     * Handle events
     */
    public boolean handleWindowEvent(Event evt) {
	if (!target.isEnabled()) {
	    return false;
	}

	switch (evt.id) {
	  case Event.MOUSE_DOWN: {
	    Choice ch = (Choice)target;
	    String items[] = new String[ch.countItems()];
	    for (int i = 0 ; i < items.length ; i++) {
		items[i] = ch.getItem(i);
	    }
	    Point pt = toGlobal(0, 0);
	    menu = new TinyChoiceMenu(this, pt.x, pt.y, width - 20, height, 
				      ch.getSelectedIndex(), items);
	    menu.winShow();
	    return true;
	  }

	  case Event.MOUSE_DRAG:
	    Point pt = menu.toLocal(toGlobal(evt.x, evt.y));
	    evt.x = pt.x;
	    evt.y = pt.y;
	    return menu.handleWindowEvent(evt);

	  case Event.MOUSE_UP: {
	    Choice ch = (Choice)target;
	    ch.select(menu.selected);
	    menu.winHide();
	    menu.winDispose();
	    menu = null;
	    targetEvent(Event.ACTION_EVENT, ch.getSelectedItem());
	    return true;
	  }
	}
	return false;
    }

    public void select(int index) {
	repaint();
    }

    public void add(String item, int index) {
	repaint();
    }

    public void remove(int index) {
      	System.out.println("Method remove(int) not implemented yet.");
    }

    public void removeAll() {
      	System.out.println("Method removeAll() not implemented yet.");
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	    return getMinimumSize();
    }

    /**
     * DEPRECATED:  Replaced by add(String, int).
     */
    public void addItem(String item, int index) {
	add(item, index);
    }

}
