/*
 * @(#)TinyFileDialogPeer.java	1.17 98/06/29
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.tiny;
/**
 * @version     1.3 02/23/96
 * @author      Siebe Brouwer
 */

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.*;
import java.io.*;

class TinyFileDialogPeer extends Dialog implements FileDialogPeer {
    FileDialog 		target;
    String 		file;
    String 		dir;
    String 		title;
    int  		mode;
    FilenameFilter 	filter;

    GridBagLayout 	gbl;
    GridBagLayout 	gblButtons;
    GridBagConstraints  gbc;

    // Components in the fileDialogWindow
    TextField 		filterField;
    TextField 		selectionField;
    List 		directoryList;
    List 		fileList;
    Panel 		buttons;
    Button 		openButton;
    Button 		filterButton;
    Button 		cancelButton;

    TinyFileDialogPeer(FileDialog target) {
    	super(((Frame)target.getParent()), target.getTitle(), true);
	this.target = target;
	this.title = target.getTitle();
	this.mode = target.getMode();
	this.target = target;

	gbl = new GridBagLayout();
	gblButtons = new GridBagLayout();
	gbc = new GridBagConstraints();
	setLayout(gbl);

	// create components
	buttons = new Panel();
	buttons.setLayout(gblButtons);
	openButton = new Button(this.title);
	filterButton = new Button("Filter");
	cancelButton = new Button("Cancel");
	directoryList = new List();
	fileList = new List();
        filterField = new TextField();
	selectionField = new TextField();

        // the insets used by the components in the fileDialog
	Insets noInset = new Insets(0, 0, 0, 0);
	Insets textFieldInset = new Insets(0, 8, 0, 8);
	Insets leftListInset = new Insets(0, 8, 0, 4);
	Insets rightListInset = new Insets(0, 4, 0, 8);
	Insets separatorInset = new Insets(8, 0, 0, 0);
	Insets labelInset = new Insets(0, 8, 0, 0);
	Insets buttonsInset = new Insets(10, 8, 10, 8);

	// add components to GridBagLayout "gbl"

	Font f = new Font("Dialog", Font.PLAIN, 12);
	Label label = new Label("Filter");
	label.setFont(f);
	addComponent(label, gbl, gbc, 0, 0, 1,
		     GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.NONE, labelInset);
    	addComponent(filterField, gbl, gbc, 0, 1, 2,
		     GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.HORIZONTAL, textFieldInset);

	label = new Label("Directories");
	label.setFont(f);
	addComponent(label, gbl, gbc, 0, 2, 1,
		     GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.NONE, labelInset);
	label = new Label("Files");
	label.setFont(f);
	addComponent(label, gbl, gbc, 1, 2, 1,
		     GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.NONE, labelInset);
	addComponent(directoryList, gbl, gbc, 0, 3, 1,
		     GridBagConstraints.WEST, (Container)this,
		     1, 1, GridBagConstraints.BOTH, leftListInset);
	addComponent(fileList, gbl, gbc, 1, 3, 1,
		     GridBagConstraints.WEST, (Container)this,
		     1, 1, GridBagConstraints.BOTH, rightListInset);
	label = new Label("Selection");
	label.setFont(f);
	addComponent(label, gbl, gbc, 0, 4, 1,
		     GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.NONE, labelInset);
	addComponent(selectionField, gbl, gbc, 0, 5, 2,
		     GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.HORIZONTAL, textFieldInset);
	addComponent(new Separator(size().width, 2, Separator.HORIZONTAL), gbl, gbc, 0, 6, 15,
		     GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.HORIZONTAL, separatorInset);

        // add buttons to GridBagLayout Buttons
	addComponent(openButton, gblButtons, gbc, 0, 0, 1,
		     GridBagConstraints.WEST, (Container)buttons,
		     1, 0, GridBagConstraints.NONE, noInset);
	addComponent(filterButton, gblButtons, gbc, 1, 0, 1,
		     GridBagConstraints.CENTER, (Container)buttons,
		     1, 0, GridBagConstraints.NONE, noInset);
	addComponent(cancelButton, gblButtons, gbc, 2, 0, 1,
		     GridBagConstraints.EAST, (Container)buttons,
		     1, 0, GridBagConstraints.NONE, noInset);

	// add ButtonPanel to the GridBagLayout of this class
	addComponent(buttons, gbl, gbc, 0, 7, 2, GridBagConstraints.WEST, (Container)this,
		     1, 0, GridBagConstraints.HORIZONTAL, buttonsInset);
	pack();

	String dir = target.getDirectory();

	if (dir != null) {
	    setDirectory(dir);
	} else if ((dir = target.getFile()) != null) {
	    setFile(dir);
	}
    }

    /**
     * add Component comp to the container cont.
     * add the component to the correct GridBagLayout
     */
    void addComponent(Component comp, GridBagLayout gb, GridBagConstraints c, int gridx,
		      int gridy, int gridwidth, int anchor, Container cont, int weightx, int weighty,
		      int fill, Insets in) {
	c.gridx = gridx;
	c.gridy = gridy;
	c.gridwidth = gridwidth;
	c.anchor = anchor;
	c.weightx = weightx;
	c.weighty = weighty;
	c.fill = fill;
	c.insets = in;
   	gb.setConstraints(comp, c);
    	cont.add(comp);
    }

    public void beginValidate() {
    }

    public void endValidate() {
    }

    /**
     * get fileName
     */
    String getFileName(String str) {
	if (str == null) {
	    return "";
	}

	int index = str.lastIndexOf('/');

	if (index == -1) {
	    return str;
	} else {
	    return str.substring(index + 1);
	}
    }

    /** handleFilter
     *
     */
    void handleFilter(String f) {

	if (f == null) {
	    return;
	}

	int index = f.lastIndexOf('/');
	String dir;

	if (index == -1) {
	    dir = "./";
	} else {
	    dir = f.substring(0, index + 1);
	    f = f.substring(index + 1);
	}
    	setFilterEntry(dir, f);
    }

    /**
     * handle the selection event
     */
    void handleSelection(String file) {
	int index = file.lastIndexOf('/');
	String dir;

	if (index == -1) {
	    dir = "./";
	    target.setFile(file);
	} else {
	    dir = file.substring(0, index+1);
	    target.setFile(file.substring(index+1));
	}
	target.setDirectory(dir);
    }

    /**
     * handle the cancel event
     */
    void handleCancel() {
	setSelectionField(null);
	setFilterField(null);
	directoryList.clear();
	fileList.clear();
	target.setFile(null);
	target.setDirectory(null);
    }

    /**
     * handle the quit event
     */
    void handleQuit() {
	handleCancel();
	target.hide();
    }

    /**
     * set the entry of the new dir with f
     */
    void setFilterEntry(String d, String f) {
	File fe = new File(d);

	if (fe.isDirectory() && fe.canRead()) {
	    setSelectionField(d);

	    if (f.equals("")) {
		f = "*";
		setFilterField(d + f);
	    } else {
		setFilterField(d + f);
	    }
	    String l[];

	    if (f.equals("*")) {
		l = fe.list();
	    } else {
                // REMIND: fileDialogFilter is not implemented yet
		FileDialogFilter ff = new FileDialogFilter(f);
		l = fe.list(ff);
	    }
	    directoryList.clear();
	    fileList.clear();

	    for (int i = 0 ; i < l.length ; i++) {
		File file = new File(d + l[i]);
		if (file.isDirectory()) {
		    directoryList.addItem(d + l[i] + "/");
		} else {
		    fileList.addItem(d + l[i]);
		}
	    }
	    this.dir = d;
	    target.setDirectory(this.dir);
	}
    }

    /**
     * set the text in the selectionFiel
     */
    void setSelectionField(String str) {
	 selectionField.setText(str);
    }

    /**
     * set the text in the filterField
     */
    void setFilterField(String str) {
	filterField.setText(str);
    }

    /**
     * handle event
     */
    public boolean handleEvent(Event evt) {
	switch (evt.id) {
	   case Event.LIST_SELECT:
	       if (directoryList == ((List)evt.target)) {
		   int index = ((Integer)evt.arg).intValue();
		   setFilterField(directoryList.getItem(index) + getFileName(filterField.getText()));
		   return true;
	       } else if (fileList == ((List)evt.target)) {
		   int index = ((Integer)evt.arg).intValue();
		   setSelectionField(fileList.getItem(index));
		   return true;
	       }
	       return false;

	    case Event.LIST_DESELECT:
	       if (directoryList == ((List)evt.target)) {
		   setFilterField(dir + getFileName(filterField.getText()));
		   return true;
	       } else if (fileList == ((List)evt.target)) {
		   setSelectionField(dir);
		   return true;
	       }
           case Event.ACTION_EVENT:

	      if (evt.arg.equals("Open"))  {
		  handleSelection(selectionField.getText());
		  handleQuit();
		  return true;
	      }

	      if (evt.arg.equals("Filter")) {
		  handleFilter(filterField.getText());
		  return true;
	      }

	      if (evt.arg.equals("Cancel")) {
		  handleQuit();
		  return true;
	      }

	      if (evt.target instanceof TextField) {
		  if (selectionField == ((TextField)evt.target)) {
		      handleFilter(selectionField.getText());
		      handleQuit();
		      return true;
		  } else if (filterField == ((TextField)evt.target)) {
		      handleFilter(filterField.getText());
		      return true;
		  }
		  return false;
	      }

	      if (evt.target instanceof List) {
		  if (directoryList == ((List)evt.target)) {
		      handleFilter(((String)evt.arg) +  getFileName(filterField.getText()));
		      return true;
		  } else if (fileList == ((List)evt.target)) {
		      handleSelection(((String)evt.arg));
		      handleQuit();
		      return true;
		  }
	      return false;
	      }
	}
    	return false;
    }

    public void handleEvent(AWTEvent e) {
    }

    /**
     * set the file
     */
    public void setFile(String file) {

	if (file == null) {
	    this.file = null;
	    return;
	}

	if (this.dir == null) {
	    String d = "./";
	    File f = new File(d, file);

	    if (f.isFile()) {
		this.file = file;
		setDirectory(d);
	    }
	} else {
	    File f = new File(this.dir, file);
	    if (f.isFile()) {
		this.file = file;
	    }
	}
    }

    /**
     * set the directory
     */
    public void setDirectory(String dir) {

	if (dir == null) {
	    this.dir = null;
	    return;
	}

	if (dir.equals(this.dir)) {
	    return;
	}
	File fe = new File(dir);

	if (!fe.isDirectory()) {
	    dir = "./";
	    fe = new File(dir);

	    if (!fe.isDirectory()) {
		return;
	    }
	}

	if (dir.endsWith("/")) {
		handleFilter(dir);
	} else {
		handleFilter(dir + "/");
	}
    }

    /**
     * set filenameFilter
     * REMIND not implemented yet
     */
    public void setFilenameFilter(FilenameFilter filter) {
    	this.filter = filter;
    }

    public boolean isFocusTraversable() {
        return false;
    }

    /* Handle TAB and Shift-TAB events. */
    public int handleFocusTraversalEvent(KeyEvent e) {

	if (e.getKeyCode() != KeyEvent.VK_TAB ||
	    (e.getSource() instanceof TextArea))
	    return IGNORE_EVENT;

 	if ((e.getModifiers() & ~InputEvent.SHIFT_MASK) > 0)
 	    return IGNORE_EVENT;

	int id = e.getID();
 	if (id == KeyEvent.KEY_RELEASED || id == KeyEvent.KEY_TYPED)
 	    return CONSUME_EVENT;

 	if (e.isShiftDown())
 	    return FOCUS_PREVIOUS;
	else
 	    return FOCUS_NEXT;
    }

}


/**
 * checks the filename and directory with the specified filter
 * checks with multiple "*".
 * the filter has to start with a '*' character.
 * this to keep the search the same as in the motif version
 */
class FileDialogFilter implements FilenameFilter {

    String filter;

    public FileDialogFilter(String f) {
	filter = f;
    }

    /**
     * return true if match
     */
    public boolean accept(File dir, String fileName) {

	File f = new File(dir, fileName);

	if (f.isDirectory()) {
	    return true;
	} else {
	    return searchPattern(fileName, filter);
	}
    }

    /**
     * start searching
     */
    boolean searchPattern(String fileName, String filter) {
	int filterCursor = 0;
	int fileNameCursor = 0;

	int filterChar = filter.charAt(filterCursor);

	if (filterCursor == 0 && filterChar != '*') {
	    return false;
	}
	String ls = filter.substring(filterCursor + 1);
	return handleStar(fileName, ls);
    }

    /**
     * call this method when character was an *
     */
     boolean handleStar(String fileName, String filter) {
	int ftLen = filter.length();
	int flLen = fileName.length();
	char ftChar;
	char flChar;
	int  ftCur = 0;
	int  flCur = 0;
	int c = 0;

	if (ftLen == 0) {
	    return true;
	}

	while (c < flLen) {
	    ftChar = filter.charAt(ftCur);

	    if (ftChar == '*') {
		String ls = filter.substring(ftCur + 1);
		String fs = fileName.substring(flCur);
		if (handleStar(fs, ls)) {
		    return true;
		}
		c++;
		flCur = c;
		ftCur = 0;
		continue;
	    }
	    flChar = fileName.charAt(flCur);

	    if (ftChar == flChar) {
		ftCur++;
		flCur++;

		if (flCur == flLen && ftCur == ftLen) {
		    return true;
		}

		if (flCur < flLen && ftCur == ftLen) {
		    return false;
		}

		if (flCur == flLen) {
		    c = flLen;
		}
	    } else {
		c++;
		flCur = c;
		ftCur = 0;
		if (c == flLen) {
		    return false;
		}
	    }
	}

	for (int i = ftCur ; i < ftLen ; i++) {
	    ftChar = filter.charAt(i);
	    if (ftChar != '*') {
		return false;
	    }
	}
	return true;
     }
}



class Separator extends Canvas {
    public final static int HORIZONTAL = 0;
    public final static int VERTICAL = 1;
    int orientation;

    public Separator(int length, int thickness, int orient) {
	super();
	orientation = orient;
	if (orient == HORIZONTAL) {
	    resize(length, thickness);
	} else { // VERTICAL
	    resize(thickness, length);
   	}
    }

    public void paint(Graphics g) {
	int x1, y1, x2, y2;
	Rectangle bbox = bounds();
	Color c = getBackground();
	Color brighter = c.brighter();
	Color darker = c.darker();

	if (orientation == HORIZONTAL) {
	    x1 = 0;
	    x2 = bbox.width - 1;
	    y1 = y2 = bbox.height/2 - 1;

	} else { // VERTICAL
	    x1 = x2 = bbox.width/2 - 1;
	    y1 = 0;
	    y2 = bbox.height - 1;
	}
	g.setColor(darker);
	g.drawLine(x1, y2, x2, y2);
	g.setColor(brighter);
	if (orientation == HORIZONTAL)
	    g.drawLine(x1, y2+1, x2, y2+1);
	else
	    g.drawLine(x1+1, y2, x2+1, y2);
    }
}
