/*
 * @(#)TinyLabelPeer.java	1.10 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.awt.peer.*;

class TinyLabelPeer extends TinyComponentPeer implements LabelPeer {
    /**
     * Create the label
     */
    TinyLabelPeer(Label target) {
	winCreate((TinyComponentPeer)target.getParent().getPeer());
	init(target);
    }

    /**
     * Minimum size.
     */
    public Dimension getMinimumSize() {
	Label label = (Label)target;
	FontMetrics fm = getFontMetrics(label.getFont());
	int w;
	try {
	    w = fm.stringWidth(label.getText());
	} catch (NullPointerException e) {
	    w = 0;
	}
	return new Dimension(w + TEXT_XPAD, 
			     fm.getAscent() + fm.getMaxDescent() + TEXT_YPAD);
    }

    /**
     * Paint the label
     */
    public void paintComponent(Graphics g) {
	Label label = (Label)target;
	g.setColor(label.getBackground());
	g.fillRect(0, 0, width, height);

	g.setColor(label.getForeground());
	g.setFont(label.getFont());
	FontMetrics fm = g.getFontMetrics();
	String lbl = label.getText();

	switch (label.getAlignment()) {
	case Label.LEFT:
	    g.drawString(lbl, 2,
			 (height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
	    break;
	case Label.RIGHT:
	    g.drawString(lbl, width - (fm.stringWidth(lbl) + 2), 
			 (height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
	    break;
	case Label.CENTER:
	    g.drawString(lbl, (width - fm.stringWidth(lbl)) / 2, 
			 (height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
	    break;
	}
    }	

    public void setText(String text) {
	repaint();
    }
    public void setAlignment(int align) {
	repaint();
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	    return getMinimumSize();
    }

}
