/*
 * @(#)TinyListPeer.java	1.26 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.tiny;
/** 
 * @version     1.2 02/23/96
 * @author      Siebe Brouwer
 */

import java.awt.*;
import java.awt.peer.*;
import java.util.Vector;

class TinyListPeer extends TinyComponentPeer implements ListPeer, TinyScrollbarClient {
    
    public final static int 	MARGIN = 4;
    public final static int 	SPACE = 2;
    public final static int 	SCROLLBAR = 17;
    public final static int	NONE = -1;
    public final static int 	WINDOW = 0;
    public final static int 	VERSCROLLBAR = 1;
    public final static int 	HORSCROLLBAR = 2;
    
    TinyVerticalScrollbar 	vsb = null;
    TinyHorizontalScrollbar 	hsb = null; 
    Vector 			items;
    boolean 			multipleSelections;
    int 			active = NONE;
    int 			selected[];
    int 			fontHeight;
    int 			fontAscent;
    int 			fontLeading;
    int 			currentIndex = -1;
    int				deselectIndex = -1;
    int				selectIndex = -1;
    int 			max; 
    long 			mtm;
    
    /**
     * Create a list
     */
    TinyListPeer(List target) {
	winCreate((TinyComponentPeer)target.getParent().getPeer());
	init(target);
	initFontMetrics();
	
	// get all items from target
	List l = (List)target;
	items = new Vector();
	for (int i = 0 ; i < l.countItems() ; i++) {
	    items.addElement(l.getItem(i));
	}
	
	max = maxLength();
        // get the index containing all indexes to selected items
	int sel[] = l.getSelectedIndexes();
	selected = new int[sel.length];
	for (int i = 0 ; i < sel.length ; i ++) {
	    selected[i] = sel[i];
	}
	multipleSelections = l.allowsMultipleSelections();
	createHorScrollbar();
	createVerScrollbar();
	
    }

    /**
     * add Vertical Scrollbar
     */
    void createVerScrollbar() {
	vsb = new TinyVerticalScrollbar();
	vsb.sb = this;
	vsb.line = 1;
	vsb.min = 0;
    }
    

    /**
     * add Horizontal scrollbar
     */
    void createHorScrollbar() {
	hsb = new TinyHorizontalScrollbar();
	hsb.sb = this;
	hsb.line = 1;
	hsb.page = 10;
	hsb.min = 0;
    }

    /* New method name for 1.1 */
    public void add(String item, int index) {
	addItem(item, index);
    }

    /* New method name for 1.1 */
    public void removeAll() {
	clear();
    }

    /* New method name for 1.1 */
    public void setMultipleMode (boolean b) {
	setMultipleSelections(b);
    }

    /* New method name for 1.1 */
    public Dimension getPreferredSize(int rows) {
	return preferredSize(rows);
    }

    /* New method name for 1.1 */
    public Dimension getMinimumSize(int rows) {
	return minimumSize(rows);
    }

    /**
     * Minimum size.
     */
    public Dimension minimumSize() {
	return minimumSize(4);
    }
    
    /**
     * return the preferredSize
     */
    public Dimension preferredSize(int v) {
	return minimumSize(v);
    }
    
    /**
     * return the minimumsize
     */
    public Dimension minimumSize(int v) {
	FontMetrics fm = getFontMetrics(target.getFont());
	initFontMetrics();
	return new Dimension(20 + fm.stringWidth("0123456789abcde"),
		(fontHeight - fontLeading + (2*SPACE)) * v + ((2*MARGIN)+SCROLLBAR));
    }

    /**
     * Calculate font metrics
     */
    void initFontMetrics() {
	FontMetrics fm = getFontMetrics(target.getFont());
	fontHeight = fm.getHeight();
	fontAscent = fm.getAscent();
	fontLeading = fm.getLeading();
    }


    /**
     * return the length of the largest item in the list
     */
    int maxLength() {
	FontMetrics fm = getFontMetrics(target.getFont());	
	int m = 0;
	int end = items.size();
	for(int i = 0 ; i < end ; i++) {
	    int l = fm.stringWidth(((String)items.elementAt(i)));
	    m = Math.max(m, l);
	}
    	return m;
    }
    
    /**
     * return the length of the given string "str"
     */
    int stringLength(String str) {
	FontMetrics fm = getFontMetrics(target.getFont());	
	return fm.stringWidth(str);
    }


    void paintHorScrollbar(Graphics g, Color colors[]) {
	int w = width - ((vsb.max == 0) ? 0 : (SCROLLBAR-2));
	Graphics ng = g.create();
	g.setColor(colors[BACKGROUND_COLOR]);
	g.fillRect(1, height - 14, w - 2, 13);
	
	try {
	    ng.translate(0, height - (SCROLLBAR - 2));
	    hsb.paint(ng, colors, w + 2, SCROLLBAR - 2);
	} finally {
	    ng.dispose();
	}
    }       

    /**
     * Paint the scrollbar.
     */
    void paintVerScrollbar(Graphics g, Color colors[]) {	
	int h = height - ((hsb.max == 0) ? 0 : (SCROLLBAR-2));
	Graphics ng = g.create();
	g.setColor(colors[BACKGROUND_COLOR]);
	g.fillRect(width - 14, 1, 13, h + 2);
	try {
	    ng.translate(width - (SCROLLBAR - 2), 0);
	    vsb.paint(ng, colors, SCROLLBAR - 2, h);
	} finally {
	    ng.dispose();
	}
    }

    /**
     * repaint the window and scrollbar
     * draw the items in the list in the window
     *
     * Make the scrollbars invisible unless needed
     */
    public synchronized void paintComponent(Graphics g) {
	int w, h;
	List list = (List)target;
	initFontMetrics();
	Color colors[] = getGUIcolors();
	g.setFont(list.getFont());
	g.setColor(colors[BACKGROUND_COLOR]);

	// Compute these on the assumption that there is a
	// horizontal scrollbar
	vsb.vis = itemsInWindow(true);
	vsb.max = Math.max(items.size() - vsb.vis, 0);
	h = height - SCROLLBAR;

	if (vsb.max != 0) {
	    w = width - SCROLLBAR;
	    hsb.vis = w - ((2 * SPACE) + (2 * MARGIN));
	    hsb.max = Math.max(max - hsb.vis, 0);
	} else {
	    // no need for a vertical scrollbar
	    w = width;
	    hsb.vis = w - ((2 * SPACE) + (2 * MARGIN));
	    hsb.max = Math.max(max - hsb.vis, 0);
	}
	if (hsb.max == 0) {
	    h = height;
	    vsb.vis = itemsInWindow(false);
	    vsb.max = Math.max(items.size() - vsb.vis, 0);
	}
	if (vsb.max == 0) {
	    w = width;
	    hsb.vis = w - ((2 * SPACE) + (2 * MARGIN));
	    hsb.max = Math.max(max - hsb.vis, 0);
	}

	hsb.val = 0;
	vsb.val = 0;
	vsb.page = vsb.vis;

	g.fillRect(0, 0, width, height);
	draw3DRect(g, colors, 0, 0, w - 1, h - 1, false);

	if (hsb.max != 0) {
	    paintHorScrollbar(g, colors);
	}
	if (vsb.max != 0) {
	    paintVerScrollbar(g, colors);
	}
	if (items.size() > 0) {
	    int n = itemsInWindow();
	    int e = Math.min(items.size() - 1, (vsb.val + n) - 1);
	    paintItems(g, colors, vsb.val, e);
	}    
    }

    /**
     * paint the item with index "index" to the screen
     * if the index is greater then the index of the last item in the list
     * draw then only a fillrect with the background color
     */
    void paintItem(Graphics g, Color colors[], int index) {
	if ((index < vsb.val) || (index >= vsb.val + itemsInWindow())) {
	    return;
	}
	int w = width - ((2 * MARGIN) + SCROLLBAR);
	int h = (fontHeight - fontLeading);
	int htotal = h + (2 * SPACE);
	int y = index2y(index) - (vsb.val * htotal);
	int x = MARGIN + SPACE;
	Graphics ng = g.create();
	if (index > items.size() - 1) {
	    ng.setColor(colors[BACKGROUND_COLOR]);
	    ng.fillRect(x - 2, y - 2, w, h + 4);
	    return;
	}
	if (isSelected(index)) {
	    ng.setColor(colors[SHADOW_COLOR]);
	    ng.fillRect(x - 1, y - 1, w - 2, h + 2);
	} else {
	    ng.setColor(colors[BACKGROUND_COLOR]);
	    ng.fillRect(x - 1, y - 1, w - 2, h + 2);
	}
	if (index == currentIndex) {
	    ng.setColor(colors[FOREGROUND_COLOR]);
	} else {
	    ng.setColor(colors[BACKGROUND_COLOR]);
	}
	ng.drawRect(x - 2, y - 2, w - 1, h + 3);
	ng.setColor(colors[FOREGROUND_COLOR]);
	String str = (String)items.elementAt(index);
	ng.clipRect(x, y, w - (2 * SPACE), h);
	ng.drawString(str, x - hsb.val, y + fontAscent);
    	ng.dispose();
    }    
    
    /**
     * paint items s to e to the screen
     */
    void paintItems(Graphics g, Color colors[], int s, int e) {
	for (int i = s ; i <= e ; i++) {
	  paintItem(g, colors, i);
	}    
    }	

    /**
     * Handle Events
     */
    public boolean handleWindowEvent(Event evt) {
	switch (evt.id) {
	    case Event.MOUSE_DOWN:
		if (inWindow(evt.x, evt.y)) {
		    active = WINDOW;
		    int i = y2index(evt.y);
		    if (i >= 0) {
			if (isSelected(i) && multipleSelections) {
			    deselect(i);
			    deselectIndex = i;
			    return true;
			} else if (i >= 0) {
			    select(i);
			    selectIndex = i;
			    return true;
			}
		    }
	    	} else if (inVerticalScrollbar(evt.x, evt.y)) {
		    active = VERSCROLLBAR;
		    evt.x -= width - SCROLLBAR;
		    int h = (hsb.max == 0) ? height : height - SCROLLBAR;
		    return vsb.handleWindowEvent(evt, SCROLLBAR - 2, h);
		} else if (inHorizontalScrollbar(evt.x, evt.y)) {
		    active = HORSCROLLBAR;
		    evt.y -= height - SCROLLBAR;
		    int w = (vsb.max == 0) ? width : width - SCROLLBAR;
		    return hsb.handleWindowEvent(evt, w, SCROLLBAR - 2);
		}
		return false;

	    case Event.MOUSE_UP:
	        if(active == VERSCROLLBAR) {
		    return vsb.handleWindowEvent(evt, 0, 0);
	        } else if(active == HORSCROLLBAR) {
		    return hsb.handleWindowEvent(evt, 0, 0);
	        } else if ((currentIndex >= 0) && (evt.when - mtm < 200)) {
		    targetEvent(Event.ACTION_EVENT, items.elementAt(currentIndex));
		} else if (active == WINDOW) {
		    if (deselectIndex >= 0) {
			targetEvent(Event.LIST_DESELECT, new Integer(deselectIndex));
			deselectIndex = -1;
		    } else if (selectIndex >= 0) {
			targetEvent(Event.LIST_SELECT, new Integer(selectIndex));
			selectIndex = -1;
		    }
		}
		mtm = evt.when;
		active = NONE;
		return true;
	    case Event.MOUSE_DRAG:
	    case Event.MOUSE_MOVE:
		if ((active == VERSCROLLBAR)) {
		    evt.x -= width - SCROLLBAR;
		    int h = height - SCROLLBAR;
		    return vsb.handleWindowEvent(evt, 15, h);
		}
		if ((active == HORSCROLLBAR)) {
		    evt.y -= height - SCROLLBAR;
		    int w = width - SCROLLBAR;
		    return hsb.handleWindowEvent(evt, w, 15);
		}
		if ((active == WINDOW) && !multipleSelections) {
		    int i = y2index(evt.y);
		    if (i >= 0 && !isSelected(i)) {
			select(i);
			selectIndex = i;
		    }
		}

	}	
	return false;
    }

    /**
     * return value from the scrollbar
     */
    public void notifyValue(Object obj, int id, int v) {
	if ((TinyVerticalScrollbar)obj == vsb) {
	    scrollVertical(v - vsb.val);
	    Graphics g = getGraphics();
	    paintVerScrollbar(g, getGUIcolors());
	    g.dispose();
   	} else if ((TinyHorizontalScrollbar)obj == hsb) {
	    int i = hsb.val;
	    hsb.val = v;
	    scrollHorizontal(v - i);
	    Graphics g = getGraphics();
	    paintHorScrollbar(g, getGUIcolors());
	    g.dispose();

	    
	}	
    }

    /** 
     * set multiple selections
     */
    public void setMultipleSelections(boolean v) {
   	multipleSelections = v;
    }
     
    /**
     * add an item
     * if the index of the item is < 0 or >= than items.size() 
     * then add the item to the end of the list
     */
    public void addItem(String item, int i) {
	if (i < 0 || i >= items.size()) {
	    i = -1;
	}
	currentIndex = -1;
	if (i == -1) {
	    items.addElement(item);	
	} else {
	    items.insertElementAt(item, i);
	    for (int j = 0 ; j < selected.length ; j++) {
	   	 if (selected[j] >= i) {
		     selected[j] += 1;
		 }
	    }	
	}
	Graphics g = getGraphics();
	Color colors[] = getGUIcolors();
	int e = vsb.val + (itemsInWindow() - 1);
	paintItems(g, colors, vsb.val, e);
	
	int s = stringLength(item);
	max = Math.max(s, max);

	int w = width - ((2 * MARGIN) + (2 * SPACE) + SCROLLBAR);
	hsb.max = max > w ? max - w : 0;
	paintHorScrollbar(g, colors);
	
	vsb.max = Math.max(items.size() - itemsInWindow(), 0);
	paintVerScrollbar(g, colors);
	g.dispose();
    }

    /** 
     * delete items s to e
     * if s < 0 then s = 0
     * if e > the max 
     */
    public void delItems(int s, int e) { 
	
	if (items.size() == 0) {
	    return;
	}

	if (s < 0) {
	    s = 0;
	}
	if (e >= items.size()) {
	    e = items.size() - 1;
	}
	
	// delete the items out of the list items and out of the list selected
	for (int i = s ; i <= e ; i++) {
	    items.removeElementAt(i);
	    int j = posInSel(i);
	    if (j != -1) {
		int newsel[] = new int[selected.length - 1];
		System.arraycopy(selected, 0, newsel, 0, j);
		System.arraycopy(selected, j + 1, newsel, j, selected.length - (j + 1));
		selected = newsel;
	    }
	    
    	}
	// update the indexes in the selected array
	int diff = (e - s) + 1;
	for (int i = 0 ; i < selected.length ; i++) {
	    if (selected[i] > e) {
		selected[i] -= diff;
	    }
	}
	
	// update vsb.val
	if ((vsb.val > e) || ((vsb.val >= s) && (vsb.val <= e))) {
	    vsb.val = Math.max(0, vsb.val - diff);
	}   

	// paint the items in the window
	currentIndex = -1;
	Graphics g = getGraphics();
	Color colors[] = getGUIcolors();
	int nr = itemsInWindow();
	int l = vsb.val + (itemsInWindow() - 1);
	paintItems(g, colors, vsb.val, l);

	max = maxLength();

	int w = width - ((2 * MARGIN) + (2 * SPACE) + SCROLLBAR);
	hsb.max = max > w ? max - w : 0;
	hsb.val = Math.min(hsb.max, hsb.val);
	paintHorScrollbar(g, colors);
	vsb.max = Math.max(items.size() - itemsInWindow(), 0);
	paintVerScrollbar(g, colors);
	g.dispose();
    }
    
    /**
     * select the index and draw the index to the screen
     *
     */
    public void select(int index) {
	if (isSelected(index)) {
	    return;
	}
	Color colors[] = getGUIcolors();
	Graphics g= getGraphics();
	if (!multipleSelections) {
	    int old[] = selected;
	    selected = new int[1];
	    selected[0] = index;
	    for (int i = 0 ; i < old.length ; i++) {
		if ((old[i] != index) && (old[i] != currentIndex)) {
		    paintItem(g, colors, old[i]);
		}	    
	    }
	} else {
	    int newsel[] = new int[selected.length + 1];
	    System.arraycopy(selected, 0, newsel, 0, selected.length);
	    newsel[selected.length] = index;
	    selected = newsel;
	}	
	int old = currentIndex;
	currentIndex = index;
	if (old != index) {
	    paintItem(g, colors, old);	
	}	
	paintItem(g, colors, index);
	g.dispose();    
    }

    /**
     * deselect the index
     * draw the index to the screen
     */
    public void deselect(int index) {
    	if (!isSelected(index)) {
	    return;
	}
	Graphics g = getGraphics();
	Color colors[] = getGUIcolors();
	if (!multipleSelections) {
	    int old[] = selected;
	    selected = new int[0];
	    for (int j = 0 ; j < old.length ; j++) {
		if ((old[j] != index) && (old[j] != currentIndex)) {
		    paintItem(g, colors, old[j]);
		}	    
	    }		  
	} else {
	    int i = posInSel(index);
	    int newsel[] = new int[selected.length - 1];
	    System.arraycopy(selected, 0, newsel, 0, i);
	    System.arraycopy(selected, i+1, newsel, i, selected.length - (i+1));
	    selected = newsel;
	}
	int o = currentIndex;
	currentIndex = index;
	if (o != index) {	
	    paintItem(g, colors, o);
	}		
	paintItem(g, colors, index);	
	g.dispose();
    }	
  
    /**
     * make visible
     */
    public void makeVisible(int index) {
    	if (index < 0 || index >= items.size()) {
	    return;
	}
	if ((index < vsb.val) || (index >= vsb.val + itemsInWindow())) {
	    Graphics g = getGraphics();
	    vsb.val = index;
	    int e = (vsb.val + itemsInWindow()) - 1;
	    paintItems(g, getGUIcolors(), vsb.val, e);
	}  
    }
    
    /**
     * clear 
     */
   public void clear() {
       selected = new int[0];
       items = new Vector();
       currentIndex = -1;
       vsb.val = 0;
       max = 0;
       repaint();
    }

    /**
     * return the selected indexes
     */
    public int[] getSelectedIndexes() {
	return selected;
    }

    /**
     * return the y value of the given index "i".
     * the y value reprecents the top of the text
     */
    int index2y(int i) {
	int h = (fontHeight - fontLeading) + (2 * SPACE);
	return MARGIN + (i * h) + SPACE;
    }

    /**
     * return the position of the index in the selected array
     * if the index isn't in the array selected return -1;
     */
    int posInSel(int index) {
	for (int i = 0 ; i < selected.length ; i++) {
	    if (index == selected[i]) {
		return i;
	    }
	}
	return -1;
    }


    /**
     * scrollVertical
     * y is the number of items to scroll
     */
    void scrollVertical(int y) {
	int k = itemsInWindow();
	int h = (fontHeight - fontLeading) + (2 * SPACE);
	Graphics g = getGraphics();
	Color colors[] = getGUIcolors();
	int j = y * h;
	
	if (vsb.val < -y) {
	    y = -vsb.val;
	}
	vsb.val += y;

	if (y > 0) {

	    if (y < k) {
		g.copyArea(MARGIN, MARGIN + j, width - SCROLLBAR, h * (k - y), 0, -j);  
	    }
	    paintItems(g, colors, vsb.val + (k - y), (vsb.val + k) - 1);
	} else if (y < 0) {

	    if (y + itemsInWindow() > 0) {
		g.copyArea(MARGIN, MARGIN, width - SCROLLBAR, h * (k + y), 0, -j);
	    }
	     int e = Math.min(items.size() - 1, vsb.val + -y);
	     paintItems(g, colors, vsb.val, e);
	}
    }

    /**
     * scrollHorizontal
     * x is the number of pixels to scroll
     */
    void scrollHorizontal(int x) {
    	int w = width - SCROLLBAR;
	w -= ((2 * SPACE) + (2 * MARGIN));
	int h = height - (SCROLLBAR + (2 * MARGIN));
	Graphics g = getGraphics();
	Color colors[] = getGUIcolors();
	
	if (x < 0) {
	    g.copyArea(MARGIN + SPACE, MARGIN, w + x, h, -x, 0);
	    g.clipRect(MARGIN + SPACE, MARGIN, -x, h);
	    int n = itemsInWindow();
	    int e = Math.min(items.size() - 1, (vsb.val + n) - 1);
	    paintItems(g, colors, vsb.val, e);	
	} else if (x > 0) {
	    g.copyArea(MARGIN + SPACE + x, MARGIN, w - x, h, -x, 0);
	    g.clipRect((MARGIN + SPACE + w) - x, MARGIN, x, h);
	    int n = itemsInWindow();
	    int e = Math.min(items.size() - 1, (vsb.val + n) - 1);
	    paintItems(g, colors, vsb.val, e);
	}
	g.dispose();
    }	

    /**
     * return the index 
     */
    int y2index(int y) {
	if (items.size() == 0) {
	    return -1;
	}
	int ln = vsb.val;
	int h = (fontHeight - fontLeading) + (2 * SPACE);
	int ys = MARGIN + h;
	int nr = itemsInWindow() - 1;
	int i = ys + (nr * h);
	if (y > i) {
	    return ln + nr;
	}
	while ((y > ys) && (ln < items.size() - 1)) {
	    ys += h;
	    ln++;
    	}	
    	return ln;
    }

    /**
     * is the index "index" selected
     */
    boolean isSelected(int index) {
	for (int i = 0 ; i < selected.length ; i++) {
	    if (selected[i] == index) {
		return true;
	    }
	}
	return false;
    }
    	
    /**
     * return the number of items that can fit 
     * in the current window
     */
    int itemsInWindow(boolean scrollbarVisible) {
	int h;
	if (scrollbarVisible) {
	    h = height - ((2 * MARGIN) + SCROLLBAR);
	} else {
	    h = height - 2*MARGIN;
	}
	int i = fontHeight - fontLeading;
	return h / (i + (2 * SPACE));
    }
		
    int itemsInWindow() {
	return itemsInWindow(hsb.max != 0);
    }

    /**
     * return true if the x and y position is in the horizontal scrollbar
     */
    boolean inHorizontalScrollbar(int x, int y) {
	int w = (vsb.max == 0) ? width : width - SCROLLBAR;
	int h = (hsb.max == 0) ? height : height - SCROLLBAR;
    	return (hsb.max != 0) &&  (x >= 0) && (x <= w) && (y > h);
    }

    /**
     * return true if the x and y position is in the verticalscrollbar
     */
    boolean inVerticalScrollbar(int x, int y) {
	int w = (vsb.max == 0) ? width : width - SCROLLBAR;
	int h = (hsb.max == 0) ? height : height - SCROLLBAR;
	return (vsb.max != 0) && (x > w) && (y >= 0) && (y <= h);
    }

    /**
     * return true if the x and y position is in the window
     */
    boolean inWindow(int x, int y) {
	int w = (vsb.max == 0) ? width : width - SCROLLBAR;
	int h = (hsb.max == 0) ? height : height - SCROLLBAR;
	return ((x >= 0) && (x <= w)) && ((y >= 0) && (y <= h));
    }
}






