/*
 * @(#)TinyMenuBarPeer.java	1.13 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;

public class TinyMenuBarPeer extends TinyWindow implements MenuBarPeer {
    TinyFramePeer frame;
    TinyMenuPeer menu;
    int menuSelected = -1;
    Vector menus = new Vector(); // local copy of the Menus

    final static int GAP = 10;

    final static Color menuBlack = TinyWindowFrame.frameBlack;
    final static Color menuDark = TinyWindowFrame.frameDark;
    final static Color menuMedium = TinyWindowFrame.frameMedium;
    final static Color menuLight = TinyWindowFrame.frameLight;
    final static Color menuWhite = TinyWindowFrame.frameWhite;

    final static Font menuFont = new Font("Dialog", Font.PLAIN, 12);

    /**
     * Create a menu bar peer
     */
    public TinyMenuBarPeer(MenuBar target) {
	// get menus from the target.
	int nitems = target.countMenus();
	for(int i = 0 ; i < nitems; i++) {
	   menus.addElement(target.getMenu(i));
	}
    }

    /**
     * Show the menu bar in a window, it will be reshaped
     * to the correct size by the window.
     */
    void init(TinyWindow parent) {
	winCreate(parent);
	winBackground(menuLight);
	winShow();
    }

    /**
     * Add a menu.
     */
    public void addMenu(Menu m) {
	menus.addElement(m);
	paint();
    }

    /**
     * Delete a menu.
     */
    public void delMenu(int index) {
	menus.removeElementAt(index);
	paint();
    }

    /**
     * Add help menu (ignore this for now).
     */
    public void addHelpMenu(Menu m) {
	paint();
    }

    /**
     * Dispose the peer. The window will be disposed
     * when the frame is disposed.
     */
    public void dispose() {
    }

    /**
     * Select which menu we are in.
     */
    boolean selectMenu(Event evt, int x, int y) {
	if (y > height) {
	    return menu != null;

	}
	FontMetrics fm = TinyFontMetrics.getFontMetrics(menuFont);
	int nitems = menus.size();
	int mx = 0;
		
	for (int i = 0 ; i < nitems ; i++) {
	    Menu mn = (Menu) menus.elementAt(i);
	    String item = mn.getLabel();
	    int w = fm.stringWidth(item);
	    if (mn.isEnabled() && (x > mx) && (x <= mx + w + GAP*2)) {
		if (menu != mn.getPeer()) {
		    if (menu != null) {
			menu.popdown(evt, false);
		    }
		    Point pt = toGlobal(mx + 1, height - 3);
		    menu = (TinyMenuPeer)mn.getPeer();
		    menu.popup(pt.x, pt.y);
		    menuSelected = i;
		    paint();
		}
		return true;
	    }
	    mx += w + GAP*2;
	}
	return menu != null;
    }

    /**
     * Paint the menu
     */

    void paint(Graphics g) {
	g.setColor(menuMedium);
	g.fillRect(1, 1, width-2, height-2);

        g.setColor(menuWhite);
	g.drawLine(0, 0, width-1, 0);
	g.drawLine(0, 0, 0, height-1);
	g.setColor(menuDark);
	g.drawLine(width-1, 0, width-1, height-1);
	g.drawLine(0, height-1, width-1, height-1);

	g.setFont(menuFont);
	FontMetrics fm = g.getFontMetrics();

	int x = GAP;
	int y = (height + fm.getAscent() - fm.getDescent()) / 2;
	int nitems = menus.size();

	for (int i = 0 ; i < nitems ; i++) {
	    Menu mn = (Menu) menus.elementAt(i);
	    String item = mn.getLabel();
	    int w = fm.stringWidth(item) + GAP*2;
	    if (i == menuSelected) {
		g.setColor(menuLight);
		g.draw3DRect((x - GAP) + 1, 3, w - 2, height - 7, false);
		g.setColor(menuMedium);
		g.fillRect((x - GAP) + 2, 4, w - 4, height - 9);
	    }
	    g.setColor(mn.isEnabled() ? menuBlack : menuDark);
	    g.drawString(item, x, y);
	    x += w;
	    if (x >= width) {
		break;
	    }
	}
    }

    /**
     * Paint the menu
     */
    void paint() {
	Graphics g = new TinyGraphics(this);
	try {
	    g.setColor(menuLight);
	    g.fillRect(1, 1, width-2, height-2);
	    paint(g);
	} finally {
	    g.dispose();
	}
    }

    public void winReshape(int x, int y, int width, int height) {
	super.winReshape(x, y, width, height);

	// resizing a menu usually requires a repaint
	paint();
    }

    /**
     * Event handler.
     */
    boolean handleWindowEvent(Event evt) {
	switch (evt.id) {
	  case Event.MOUSE_DOWN:
	    selectMenu(evt, evt.x, evt.y);
	    return true;

	  case Event.MOUSE_DRAG:
	    if (selectMenu(evt, evt.x, evt.y)) {
		Point pt = menu.toLocal(toGlobal(evt.x, evt.y));
		evt.x = pt.x;
		evt.y = pt.y;
		return menu.handleWindowEvent(evt);
	    }
	    return true;

	  case Event.MOUSE_UP:
	    if (menu != null) {
		menu.popdown(evt, true);
		menu = null;
		menuSelected = -1;
		paint();
	    }
	    return true;
	}
	return false;
    }
}
