/*
 * @(#)TinyMenuPeer.java	1.10 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.awt.peer.*;

public class TinyMenuPeer extends TinyWindow implements MenuPeer {
    TinyMenuPeer menu;
    int selected;
    Menu target;
    int h;
    boolean created;
    private Point pt;

    /**
     * Create a menu
     */
    public TinyMenuPeer(Menu target) {
	this.target = target;
	created = false;
	pt = new Point(0,0);
    }

    /**
     * Compute the width and height of the menu
     */
    void computeSize(int x, int y) {
	FontMetrics fm = TinyFontMetrics.getFontMetrics(TinyMenuBarPeer.menuFont);
	int nitems = target.countItems();
	h = fm.getHeight() + 6;

	int w = 40;
	for (int i = 0 ; i < nitems ; i++) {
	    String str = target.getItem(i).getLabel();
	    w = Math.max(w, fm.stringWidth(str) + 40);
	}
	
	winReshape(x, y, w, 4 + h * nitems);
    }

    /**
     * Show at an x,y location.
     */
    void popup(int x, int y) {
	if (!created) {
	    winCreate(null);
	    winBackground(TinyWindowFrame.frameLight);
	    created = true;
	}

	selected = -1;
	computeSize(x, y);
	winFront();
	winShow();
    }

    /**
     * Hide the menu and execute the action if
     * action is true.
     */
    void popdown(Event evt, boolean action) {
	winHide();
	if (menu != null) {
	    menu.popdown(evt, action);
	    menu = null;
	} else if (action && (selected >= 0)) {
	    MenuItem mi = target.getItem(selected);
	    if (mi instanceof CheckboxMenuItem) {
		CheckboxMenuItem cb = (CheckboxMenuItem)mi;
		cb.setState(!cb.getState());
	    }
	    if (!(mi instanceof Menu)) {
		try {
		    mi.postEvent(new Event(mi, evt.when, Event.ACTION_EVENT, 0, 0, 0, evt.modifiers, mi.getLabel()));
		} catch (ThreadDeath e) {
		    throw e;
		} catch (Throwable e) {
		    e.printStackTrace();
		}
	    }
	}
	selected = -1;
    }

    /**
     * Paint an item
     */
    void paintItem(Graphics g, FontMetrics fm, int i) {
	int y = 2 + (h*i) + (h + fm.getMaxAscent() - fm.getMaxDescent()) / 2;
	MenuItem mi = target.getItem(i);
	String str = mi.getLabel();

	if (str.equals("-")) {
	    g.setColor(TinyWindowFrame.frameLight);
	    g.draw3DRect(1, y - fm.getMaxAscent() / 2, width-3, 1, false);
	} else {
	    if (i == selected) {
		g.setColor(TinyWindowFrame.frameLight);
		g.draw3DRect(2, 2 + i * h, width-5, h-1, false);
		g.setColor(TinyWindowFrame.frameMedium);
		g.fillRect(3, 3 + i * h, width-7, h-3);
	    } else {
		g.setColor(TinyWindowFrame.frameLight);
		g.fillRect(2, 2 + i * h, width-4, h);
	    }
	    g.setColor(mi.isEnabled() ? 
		       TinyWindowFrame.frameBlack : 
		       TinyWindowFrame.frameDark);
	    g.drawString(str, 20, y);

	    if (mi.isEnabled() && (mi instanceof Menu)) {
		int ax = width - 15;
		g.setColor(TinyWindowFrame.frameLight.darker());
		g.drawLine(ax, y - 2, ax + 8, y - 6);
		g.setColor(TinyWindowFrame.frameLight.brighter());
		g.drawLine(ax + 8, y - 6, ax, y - 10);
		g.drawLine(ax, y - 10, ax, y - 2);
	    } else if (mi instanceof CheckboxMenuItem) {
		g.setColor(TinyWindowFrame.frameLight);
		boolean state = ((CheckboxMenuItem)mi).getState();
		g.draw3DRect(8, y - 8, 6, 6, !state);
		g.setColor(state ? TinyWindowFrame.frameMedium : TinyWindowFrame.frameLight);
		g.fillRect(9, y - 7, 5, 5);
	    }
	}
    }

    /**
     * Paint the entire menu.
     */
    void paint(Graphics g) {
	g.setColor(TinyWindowFrame.frameLight);
	g.draw3DRect(0, 0, width-1, height-1, true);
	g.draw3DRect(1, 1, width-3, height-3, true);
	g.setFont(TinyMenuBarPeer.menuFont);
	FontMetrics fm = g.getFontMetrics();

	int nitems = target.countItems();
	for (int i = 0 ; i < nitems ; i++) {
	    paintItem(g, fm, i);
	}
    }

    /**
     * Select an item.
     */
    void select(int i) {
	if (i == selected) {
	    return;
	}

	Graphics g = new TinyGraphics(this);
	try {
	    g.setColor(TinyWindowFrame.frameLight);
	    g.setFont(TinyMenuBarPeer.menuFont);
	    FontMetrics fm = g.getFontMetrics();
	    if (selected >= 0) {	
		int old = selected;
		selected = -1;
		paintItem(g, fm, old);
	    }
	    if (i < 0) {
		return;
	    }

	    MenuItem mi = target.getItem(i);
	    if (mi.isEnabled() && !mi.getLabel().equals("-")) {
		paintItem(g, fm, selected = i);
	    } else {
		selected = -1;
	    }
	} finally {
	    g.dispose();
	}
    }

    /**
     * Handle events.
     */
    boolean handleWindowEvent(Event evt) {
	switch (evt.id) {
	  case Event.MOUSE_DRAG:
	    if (menu != null) {
		if (evt.x < width - 20) {
		    menu.popdown(evt, false);
		    menu = null;
		} else {
		    pt.x = evt.x;
		    pt.y = evt.y;
		    menu.toLocal(toGlobal(pt));
		    evt.x = pt.x;
		    evt.y = pt.y;
		    return menu.handleWindowEvent(evt);
		}
	    }
	    if (   (evt.x < 0) || (evt.x > width)
		|| (evt.y < 0) || (evt.y > height)  ) {
		select(-1);
		return true;
	    }

	    int nitems = target.countItems();
	    int i = Math.max(0, Math.min(nitems-1, (evt.y - 2)/h));
	    select(i);

	    if ((evt.x > width - 20) && (selected >= 0)) {
		MenuItem mi = target.getItem(selected);

		if ((mi instanceof Menu) && (mi.isEnabled())) {
		    menu = (TinyMenuPeer)mi.getPeer();
		    pt.x = width - 20;
		    pt.y = 2 + selected*h;
		    toGlobal(pt);
		    menu.popup(pt.x, pt.y);
		}
	    }
	    return true;
	}
	return false;
    }

    public void addSeparator() {
    }
    public void addItem(MenuItem item) {
    }
    public void delItem(int index) {
    }
    public void setLabel(String label) {
    }
    public void dispose() {
	if (created) {
	    winDispose();
	    created = false;
	}
    }

    public void setEnabled(boolean b) {
    }

    public void enable() {
    }
    public void disable() {
    }
}
