/*
 * @(#)TinyScrollPanePeer.java	1.18 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.tiny;

import java.awt.*;
import java.awt.peer.*;

class TinyScrollPanePeer extends TinyComponentPeer implements ScrollPanePeer, TinyScrollbarClient {
    
    public final static int 	MARGIN = 1;
    public final static int 	SCROLLBAR = 17;
    public final static int 	SPACE = 2;

    public final static int 	VERTICAL = 1 << 0;
    public final static int 	HORIZONTAL = 1 << 1;
    
    TinyVerticalScrollbar 	vsb;
    TinyHorizontalScrollbar 	hsb;
    TinyWindow			clip;
 
    int				active;
    int				hsbSpace;
    int				vsbSpace;
    
    TinyScrollPanePeer(ScrollPane target) {
	winCreate((TinyComponentPeer)target.getParent().getPeer());

	// Create the clip window.  The field "clip" must be null when
	// we call winCreate, or the parent of clip will be set to itself!
	clip = null;
	TinyWindow c = new TinyWindow();
	c.winCreate(this, TINYWINDOWMANAGER);
	clip = c;

	vsb = new TinyVerticalScrollbar();
	vsb.sb = this;
	vsb.line = 1;
	vsb.page = 1;
	vsb.min = 0;
	vsb.val = 0;
  
	hsb = new TinyHorizontalScrollbar();
	hsb.sb = this;
	hsb.line = 1;
	hsb.page = 1;
	hsb.min = 0;
	hsb.val = 0;

	if (target.getScrollbarDisplayPolicy() == ScrollPane.SCROLLBARS_ALWAYS) {
	    vsbSpace = hsbSpace = SCROLLBAR;
	} else {
	    vsbSpace = hsbSpace = 0;
	}
	init(target);

	clip.winShow();
    }

    TinyWindow getParentSubwindow() {
	return (clip == null) ? this : clip;
    }

    public void setBounds(int x, int y, int w, int h) {	
	reshape(x, y, w, h);
    }

    public Insets getInsets() {
	return new Insets(MARGIN, MARGIN, MARGIN+SPACE+hsbSpace, MARGIN+SPACE+vsbSpace);
    }

    public int getHScrollbarHeight() {
	return SCROLLBAR+SPACE;
    }

    public int getVScrollbarWidth() {
	return SCROLLBAR+SPACE;
    }

    public void childResized(int w, int h) {
	if (setScrollbarSpace()) {
	    setViewportSize();
	}
	Graphics g = getGraphics();
	paintComponent(g);
	g.dispose();
    }

    Dimension getChildSize() {
	ScrollPane sp = (ScrollPane)target;
	if (sp.countComponents() > 0) {
	    Component c = sp.getComponent(0);
	    return c.size();
	} else {
	    return new Dimension(0, 0);
	}
    }

    boolean setScrollbarSpace() {
        ScrollPane sp = (ScrollPane)target;
	boolean changed = false;
	int sbDisplayPolicy = sp.getScrollbarDisplayPolicy();

	if (sbDisplayPolicy == ScrollPane.SCROLLBARS_NEVER) {
	    return changed;
	}
	Dimension cSize = getChildSize();

	if (sbDisplayPolicy == ScrollPane.SCROLLBARS_AS_NEEDED) {
	    int oldHsbSpace = hsbSpace;
	    int oldVsbSpace = vsbSpace;
	    hsbSpace = (cSize.width <= (width - 2*MARGIN)? 0 : SCROLLBAR);
	    vsbSpace = (cSize.height <= (height - 2*MARGIN)? 0 : SCROLLBAR);

	    if (hsbSpace == 0 && vsbSpace != 0) {
		hsbSpace = (cSize.width <= (width - SCROLLBAR - 2*MARGIN)? 0 : SCROLLBAR);
	    }
	    if (vsbSpace == 0 && hsbSpace != 0) {
		vsbSpace = (cSize.height <= (height - SCROLLBAR - 2*MARGIN)? 0 : SCROLLBAR);
	    }
	    if (oldHsbSpace != hsbSpace || oldVsbSpace != vsbSpace) {
		changed = true;
	    }
	} 
	if (vsbSpace > 0) {
	    vsb.vis = height - (2*MARGIN) - SPACE - hsbSpace;
	    vsb.max = Math.max(cSize.height - vsb.vis, 0);
            vsb.page = (int)(vsb.vis * .90);
            Adjustable vadj = sp.getVAdjustable();
            vadj.setVisibleAmount(vsb.vis);
            vadj.setMaximum(vsb.max);
            vadj.setBlockIncrement(vsb.page);
	}
	if (hsbSpace > 0) {
	    hsb.vis = width - (2*MARGIN) - SPACE - vsbSpace;
	    hsb.max = Math.max(cSize.width - hsb.vis, 0);
            hsb.page = (int)(hsb.vis * .90);
            Adjustable hadj = sp.getHAdjustable();
            hadj.setVisibleAmount(hsb.vis);
            hadj.setMaximum(hsb.max);
            hadj.setBlockIncrement(hsb.page);
	}

	// Check to see if we hid either of the scrollbars but left
	// ourselves scrolled off of the top and/or right of the pane.
	// If we did, we need to scroll to the top and/or right of of
	// the pane to make it visible.
	//
	// Reminder: see if there is a better place to put this code.
	boolean must_scroll = false;

	// Get the point at which the ScrollPane is currently located
	Point p = ((ScrollPane)target).getComponent(0).location();
	
	if ((vsbSpace == 0) && (p.y < 0)) {
	  p.y = 0;
	  must_scroll = true;
	}
	
	if ((hsbSpace == 0) && (p.x < 0)) {
	  p.x = 0;
	  must_scroll = true;
	}
	
	if (must_scroll)
	  scroll(x, y, VERTICAL | HORIZONTAL);

	return changed;
    }

    void setViewportSize() {
	clip.winReshape(MARGIN, MARGIN,
		 width - (2*MARGIN) - SPACE - vsbSpace,
		 height - (2*MARGIN) - SPACE - hsbSpace);
    }

    public void setUnitIncrement(Adjustable adj, int u) {
        if (adj.getOrientation() == Adjustable.VERTICAL) {
            vsb.line = u;
        } else { // HORIZONTAL
            hsb.line = u;
        }
    }

    public void setValue(Adjustable adj, int v) {
        if (adj.getOrientation() == Adjustable.VERTICAL) {
            scroll(-1, v, VERTICAL);
        } else { // HORIZONTAL
            scroll(v, -1, HORIZONTAL);
        }  
    }

    public void setScrollPosition(int x, int y) {
	scroll(x, y, VERTICAL | HORIZONTAL);
    }

    /**
     * Scroll the contents to position x, y
     */
    void scroll(int x, int y, int flag) {
	ScrollPane sp = (ScrollPane)target;
	if (sp.countComponents() > 0) {
	    int sx, sy;
	    Component c = sp.getComponent(0);
	    Color colors[] = getGUIcolors();

	    if (sp.getScrollbarDisplayPolicy() == ScrollPane.SCROLLBARS_NEVER) {
	        sx = -x;
		sy = -y;
	    } else {
	        Point p = c.location();
	        sx = p.x;
	        sy = p.y;

	        if ((flag & HORIZONTAL) != 0) {
		    hsb.val = Math.min(x, hsb.max);
                    Adjustable hadj = sp.getHAdjustable();
                    hadj.setValue(hsb.val);
		    sx = -(hsb.val);
	            Graphics g = getGraphics();
	            paintHorScrollbar(g, colors);
	            g.dispose();	    
	        }
	        if ((flag & VERTICAL) != 0) {
		    vsb.val = Math.min(y, vsb.max);
                    Adjustable vadj = sp.getVAdjustable();
                    vadj.setValue(vsb.val);                    
		    sy = -(vsb.val);
	    	    Graphics g = getGraphics();
	    	    paintVerScrollbar(g, colors);
	    	    g.dispose();
		}
	    }		
	    //System.out.println("scroll to "+sx+","+sy);
	    // native scrolling would be great here!
	    c.move(sx, sy);
	} else {
	    throw new NullPointerException("scrollpane child is null");
	}
    }

    /**
     * Paint the scrollpane.
     */
    public void paintComponent(Graphics g) {
	//setupScrollPane();

	Color colors[] = getGUIcolors();
	g.setColor(colors[BACKGROUND_COLOR]);
	int h = height - hsbSpace - SPACE;
	int w = width - vsbSpace - SPACE;

	g.fillRect(0, 0, width, height);
	if (MARGIN > 0) {
	    draw3DRect(g, colors, 0, 0, w - 1, h - 1, false);
	}

	if (vsbSpace > 0) {
	    paintVerScrollbar(g, colors);
	}
	if (hsbSpace > 0) {
	    paintHorScrollbar(g, colors);
	}
    }

    void paintHorScrollbar(Graphics g, Color colors[]) {
	if (hsbSpace <= 0) {
	    return;
	}
	Graphics ng = g.create();
	g.setColor(colors[BACKGROUND_COLOR]);
	g.fillRect(0, height - SCROLLBAR, width - vsbSpace, SCROLLBAR);	
	try {
	    ng.translate(MARGIN, height - (SCROLLBAR - 1));
	    hsb.paint(ng, colors, width - vsbSpace - (2*MARGIN), 
		SCROLLBAR - 2);
	} finally {
	    ng.dispose();
	}
    }       

    void paintVerScrollbar(Graphics g, Color colors[]) {
	if (vsbSpace <= 0) {
	    return;
	}	
	Graphics ng = g.create();
	g.setColor(colors[BACKGROUND_COLOR]);
	g.fillRect(width - SCROLLBAR, 0, SCROLLBAR, height - hsbSpace);
	try {
	    ng.translate(width - (SCROLLBAR - 1), MARGIN);
	    vsb.paint(ng, colors, SCROLLBAR - 2,
		      height - hsbSpace - (2*MARGIN));
	} finally {
	    ng.dispose();
	}
    }

    /**
     * Handle Events
     */
    public boolean handleWindowEvent(Event evt) {
	switch (evt.id) {

	    case Event.MOUSE_DOWN:
		if (inVerticalScrollbar(evt.x, evt.y)) {
		    active = VERTICAL;
		    evt.x -= width - SCROLLBAR;
		    int h = height - hsbSpace - (2*MARGIN);
		    return vsb.handleWindowEvent(evt, 15, h);
		} else if (inHorizontalScrollbar(evt.x, evt.y)) {
		    active = HORIZONTAL;
		    evt.y -= height - SCROLLBAR;
		    int w = width - SCROLLBAR;
		    return hsb.handleWindowEvent(evt, w, 15);
		}
		return false;

	    // On mouse up, pass the event through to the scrollbar to stop
	    // scrolling. The x & y passed do not matter.
	    case Event.MOUSE_UP:
	      if (active == VERTICAL)
	      {
		return vsb.handleWindowEvent(evt, 0, 0);
	      }
	      else if (active == HORIZONTAL)
	      {
		return hsb.handleWindowEvent(evt, 0, 0);
	      }
	      return false;


	    case Event.MOUSE_DRAG:
	    case Event.MOUSE_MOVE:
		if ((active == VERTICAL)) {
		    evt.x -= width - SCROLLBAR;
		    int h = height - hsbSpace;
		    return vsb.handleWindowEvent(evt, 15, h);
		}
		if ((active == HORIZONTAL)) {
		    evt.y -= height - SCROLLBAR;
		    int w = width - vsbSpace;
		    return hsb.handleWindowEvent(evt, w, 15);
		}
	}	
	return false;
    }

    /**
     * return value from the scrollbar
     */
    public void notifyValue(Object obj, int id, int v) {
	if ((TinyVerticalScrollbar)obj == vsb) {
	    scroll(-1, v, VERTICAL);
   	} else if ((TinyHorizontalScrollbar)obj == hsb) {
	    scroll(v, -1, HORIZONTAL);
	}	
    }

   /**
     * return true if the x and y position is in the verticalscrollbar
     */
    boolean inVerticalScrollbar(int x, int y) {
	if (vsbSpace <= 0) {
	    return false;
	}
    	int h = height - hsbSpace;
	return (x > width - (SCROLLBAR - 2)) && (x <= width) && (y >= 0) && (y <= h);
    }

    /**
     * return true if the x and y position is in the horizontal scrollbar
     */
    boolean inHorizontalScrollbar(int x, int y) {
	if (hsbSpace <= 0) {
	    return false;
	}
    	int w = width - vsbSpace;
    	return (x >= 0) && (x <= w) && (y > height - SCROLLBAR) && (y <= height);
    }

    /**
     * DEPRECATED:  Replaced by setBounds(int, int, int, int).
     */
    public void reshape(int x, int y, int w, int h) {	
	super.reshape(x, y, w, h);
	setScrollbarSpace();
	setViewportSize();
	Graphics g = getGraphics();
	paintComponent(g);
	g.dispose();
    }

    /**
     * DEPRECATED:  Replaced by getInsets().
     */
    public Insets insets() {
	return getInsets();
    }

}
