/*
 * @(#)TinyTextAreaPeer.java	1.32 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.tiny;
/** 
 * @version     1.2 02/23/96
 * @author      Siebe Brouwer
 */

import java.awt.*;
import java.awt.peer.*;

class TinyTextAreaPeer extends TinyComponentPeer implements TextAreaPeer, TinyScrollbarClient {
    public static final int MARGIN = 4;
    public static final int BORDER = 2;
    public static final int SCROLLBAR = 17;
    public static final int HORSCROLLBAR = 0;
    public static final int VERSCROLLBAR = 1;
    public static final int TEXTAREA = 2;
    
    static String clipboard = "";

    TinyVerticalScrollbar 	vsb;
    TinyHorizontalScrollbar 	hsb;
    
    boolean hasFocus;
    char text[];
    int lines[];
    int nrOfLines;
    int textLength;
    int mpos;
    int selStart;
    int selEnd;
    long mtm;
    int topLine;
    int movedRight;
    int mode;
    int fontAscent;
    int fontLeading;
    int fontHeight;
    
    public static void setClipboard(String contents) {
	clipboard = contents;
    }

    public static String getClipboard() {
	return clipboard;
    }

    /**
     * Create a Text area.
     */
    TinyTextAreaPeer(TextArea target) {

	winCreate((TinyComponentPeer)target.getParent().getPeer());
	init(target);
	target.isEditable();
        
 	FontMetrics fm = getFontMetrics(target.getFont());
	fontAscent = fm.getAscent();
	fontLeading = fm.getLeading();
	fontHeight = fm.getHeight();
	
	// initialize Horizontal Scrollbar
	hsb = new TinyHorizontalScrollbar();
	hsb.sb = this;
	hsb.line = 10;

        // initialize Vertical Scrollbar
	vsb = new TinyVerticalScrollbar();
	vsb.sb = this;
	vsb.line = 1;

	// set the text of this object to the text of its target
	setText(target.getText());
    }

    static boolean scrollbars_optional = false;

    // if we support optional scrollbars, the padding will change dynamically
    int getScrollbarPad(int max) {
	return scrollbars_optional ? ((max == 0) ? 0 : SCROLLBAR) : SCROLLBAR;
    }

    /**
     * Compute minimum size.
     */
    public Dimension getMinimumSize() {
	return getMinimumSize(10, 60);
    }

    public Dimension getPreferredSize(int rows, int cols) {
	return getMinimumSize(rows, cols);
    }

    public Dimension getMinimumSize(int rows, int cols) {
	FontMetrics fm = getFontMetrics(target.getFont());	
	return new Dimension(fm.charWidth('0') * cols + 2*MARGIN + SCROLLBAR,
			     fm.getHeight() * rows + 2*MARGIN + SCROLLBAR);
    }

    public boolean isFocusTraversable() {
	return true;
    }
    
    /**
     * Update the index Array from the specified lineNumber
     */
    protected void updateIndex(int nr) {
    	nrOfLines = nr + 1;
	int offset = 0;
	lines[0] = 0;
	for (int i = lines[nr]; i < textLength; i++) {
	    if (lines.length == nrOfLines) {
	        int tmp[] = new int[nrOfLines + 10];
		System.arraycopy(lines, 0, tmp, 0, nrOfLines);
		lines = tmp;        	
	    }	    
	    if(text[i] == '\n') {
		offset = i + 1;
		if((i+1) < textLength) {
		    if(text[i+1] == '\r') {
			offset++;
			++i;
		    }		    
		}
		lines[nrOfLines++] = offset;
	    }
	    else if(text[i] == '\r') {
		offset = i + 1;
		if((i+1) < textLength) {
		    if(text[i+1] == '\n') {
			offset++;
			++i;
		    }		    
		}
		lines[nrOfLines++] = offset;
	    }
	}
    }

    /** 
     * return the position on the line "ln" given the old xPosition
     * this is used for the UP event
     */
    int xPos(int xo, int ln) {
	FontMetrics fm = getFontMetrics(target.getFont());	
	int pose = (ln + 1 < nrOfLines) ? lines[ln + 1] - 1 : textLength;	
	int x = MARGIN, widths[] = fm.getWidths();
	int posn = lines[ln];
	while((x < xo) && (posn < pose)) {
	    x += widths[text[posn]];
	    posn++;	
	}
	return posn;
    }	

    /**
     * number of lines that can fit in a window
     */
    int linesInWindow(int max) {
	int htotal = height - ((2 * MARGIN) + getScrollbarPad(max));
	return htotal / fontHeight;
    }

    int linesInWindow() {
	return linesInWindow(hsb.max);
    }

    int maxLinesInWindow() {
	return linesInWindow(0);
    }

    /** 
     * return the position in the array "lines" from the values x and y.
     */
    public int xy2pos(int x, int y) {
    	int ln = 0;
	int endl = lineNumber(selEnd);
	if (y < 0) {
	    if (topLine - 1 >= 0) { 
	    	ln = topLine - 1;
	    } else {
		return 0;
	    }	
	} else if (y > height) {
	    if (endl < nrOfLines - 1) {
		ln = endl + 1;
	    } else {
		return textLength;
	    }
	} else {
	    ln = topLine;
	    int ys = MARGIN + fontHeight;
	    while ((y > ys) && (ln < nrOfLines - 1)) {
		ys += fontHeight;
	    	ln++;
	    }
        }
	FontMetrics fm = getFontMetrics(target.getFont());	
	int widths[] = fm.getWidths();
	int begin = lines[ln];
	int end = (ln + 1 < nrOfLines) ? lines[ln + 1] - 1 : textLength;
	x -= MARGIN;
	for (int i = begin; i < end ; i++) {
	    int w = widths[text[i]];
	    if (x < (w / 2)) {
		return i;
	    }
	    x -= w;
	}
	return end;
    }
    
    /**
     * return the x position of the position pos 	
     */
     int pos2x(int pos) {
	FontMetrics fm = getFontMetrics(target.getFont());	
	int widths[] = fm.getWidths();
	int ln = lineNumber(pos);
	int begin = lines[ln];
	int x = MARGIN;
	for (int i = begin ; i < pos ; i++) {
	    x += widths[text[i]];
	}
	return x;
    }
    
    /**
     * return the y position of the lineNumber
     */
    int ln2y(int ln) {
	return MARGIN + fontLeading + ((ln - topLine) * fontHeight);
    }

    /**
     * return the lineNumber
     * use binairy search
     */
    int lineNumber(int pos) {	
	int i = 0, j = nrOfLines - 1;
	while (i < j) {
	    int m = i + ((j - i) >> 1);
	    int s = lines[m];
	    int e = lines[m + 1];
	    if (pos < s) {
		j = m - 1;
	    } else if (pos >= e) {
		i = m + 1;
	    } else {
		return m;
	    }
	}
	return i;
    }

    /**
     * return the length of the largest line
     */
    int maxLength() {
	FontMetrics fm = getFontMetrics(target.getFont());	
	int max = hsb.val;
	int end = Math.min(nrOfLines, topLine + linesInWindow());
		
	for(int i = topLine ; i < end ; i++) {
	    int b = lines[i];
	    int e = (i + 1 == nrOfLines) ? textLength : lines[i + 1];
	    int l = fm.charsWidth(text, b, e - b);
	    max = Math.max(max, l);
	}
    	return max;
    }

    void sizeVerScrollbar(int l) {
	vsb.vis = l;
	vsb.page = l;
	vsb.max = Math.max(0, nrOfLines - l);
	vsb.val = Math.min(topLine, vsb.max);
    }

    /**
     * Paint the vertical scrollbar to the screen
     */
    void paintVerScrollbar(Graphics g) {
	paintVerScrollbar(g, getGUIcolors());
    }

    void paintVerScrollbar(Graphics g, Color colors[]) {
	if (getScrollbarPad(vsb.max) == 0) {
	    return;
	}

	int h = height - getScrollbarPad(hsb.max);

	// paint scrollbar
	Graphics ng = g.create();
	g.setColor(colors[BACKGROUND_COLOR]);
	g.fillRect(width - 14, 1, 13, h + 2);
	
	try {
	    ng.translate(width - (SCROLLBAR - 2), 0);
	    vsb.paint(ng, colors, SCROLLBAR - 2, h);
	} finally {
	    ng.dispose();
	}
    }    
    
    void sizeHorScrollbar(int w) {
        // initialize
	hsb.vis = w;
	hsb.page = hsb.vis;
	hsb.max = Math.max(0, maxLength() - w);
	hsb.val = Math.min(movedRight, hsb.max);
    }


    /**
     * paint the horizontal scrollbar to the screen
     */
    void paintHorScrollbar(Graphics g) {
	paintHorScrollbar(g, getGUIcolors());
    }

    void paintHorScrollbar(Graphics g, Color colors[]) {
	if (getScrollbarPad(hsb.max) == 0) {
	    return;
	}

        // paint scrollbar
	Graphics ng = g.create();
	g.setColor(colors[BACKGROUND_COLOR]);
	int w = width - getScrollbarPad(vsb.max);
	g.fillRect(1, height - 14, w + 2, 13);
	
	try {
	    ng.translate(0, height - (SCROLLBAR - 2));
	    hsb.paint(ng, colors, w, SCROLLBAR - 2);
	} finally {
	    ng.dispose();
	}
    }       

    int resizeScrollbars() {
	int h = hsb.max; 
	int v = vsb.max;

	sizeVerScrollbar(linesInWindow(1));
	sizeHorScrollbar(width - (2*MARGIN + getScrollbarPad(vsb.max)));
	if (getScrollbarPad(hsb.max) == 0) {
	    sizeVerScrollbar(maxLinesInWindow());
	}
	if (getScrollbarPad(vsb.max) == 0) {
	    sizeHorScrollbar(width - (2*MARGIN));
	}
	return (((hsb.max == h) ? 0 : 1) + ((vsb.max == v) ? 0 : 2));
    }
	    
   /**
    * Paint the component
    * this method is called when the repaint instruction has been used
    */
    public void paintComponent(Graphics g) {
	TextArea area = (TextArea)target;
	Color colors[] = getGUIcolors();
	g.setFont(area.getFont());
	g.setColor(backGroundColor(colors));
	g.fillRect(0, 0, width, height);
	
	resizeScrollbars();
	int w = width -  ((2 * BORDER) + getScrollbarPad(vsb.max));
	int h = height - ((2 * BORDER) + getScrollbarPad(hsb.max));

	paintVerScrollbar(g, colors);
	paintHorScrollbar(g, colors);
	
	if (hasFocus) {
	    g.setColor(colors[FOREGROUND_COLOR]);
	    g.drawRect(0, 0, w + (2*BORDER) - 1, h + (2*BORDER) - 1);
	    g.drawRect(1, 1, w + (2*BORDER) - 3, h + (2*BORDER) - 3);
	} else {
	    g.setColor(colors[BACKGROUND_COLOR]);
	    g.drawRect(0, 0, w + (2*BORDER) - 1, h + (2*BORDER) - 1);
	    g.drawRect(1, 1, w + (2*BORDER) - 3, h + (2*BORDER) - 3);
	} 
	
	if (text != null) {
	    int l = linesInWindow();
	    h = height - ((2 * MARGIN) + getScrollbarPad(hsb.max));
	    int e = Math.min(nrOfLines - 1, (topLine + l) - 1);
	    paintLines(g, topLine, e);
	}
    }
    
    /**
     * fill the position on the screen with the desired color
     * the color is given by "s". 
     * true = selectColor
     * false = backGroundColor
     */
    void paintRange(Graphics g, int start, int end) {
	int lnrStart = lineNumber(start);
	int lnrEnd = lineNumber(end);
	Graphics ng = g.create();
	Color colors[] = getGUIcolors();
	
	if (lnrStart == lnrEnd) {
	    int xmax = width - (BORDER + getScrollbarPad(vsb.max));
	    int xs = Math.max(BORDER, pos2x(start) - movedRight);
	    int xe = Math.min(xmax, pos2x(end) - movedRight);
	    int w = (start == end) ? 1 : xe - xs;
	    int y = ln2y(lnrStart) - fontLeading;
	    ng.clipRect(xs, y, w, fontHeight);
	    paintLine(ng, colors, lnrStart);
	} else {
	    int diff = lnrEnd - lnrStart;
	    int w = width - ((2 * BORDER) + getScrollbarPad(vsb.max));
	    int h = height - ((2 * MARGIN) + getScrollbarPad(hsb.max));
	    ng.clipRect(BORDER, MARGIN, w, h);
	    paintLine(ng, colors, lnrStart);
	    lnrStart++;
	    
	    while (diff > 1) {
		paintLine(ng, colors, lnrStart);
		lnrStart++;
		diff--;
	    }
	    paintLine(ng, colors, lnrEnd);
	}
   	ng.dispose();
    }
    	    
    /**
     * draw the lines to the screen
     */
    void paintLines(Graphics g, int s, int e) {
	Color colors[] = getGUIcolors();
	int w = width - ((2 * BORDER) + getScrollbarPad(vsb.max));
	int h = height - ((2 * MARGIN) + getScrollbarPad(hsb.max));
	int lm = linesInWindow() + topLine;
	s = Math.max(topLine, s);
	e = Math.min(e, lm - 1);
	Graphics ng = g.create();
	ng.clipRect(BORDER, MARGIN, w, h);
	
	for (int i = s ; i <= e; i++) {
	    paintLine(ng, colors, i);
	}     
    	ng.dispose();
    }

    /**
     * paint the background color of the line "lnr" to the screen and then paint draw the 
     * line to the screen
     * if the line number is greater than nrOfLines - 1 than do a fillrect to the
     * position of the line, use the background color
     */
    void paintLine(Graphics g, Color colors[], int lnr) {
	int l = linesInWindow();

	if((lnr < topLine) || (lnr >= l + topLine)) { 
	    return;
	}
	int w = width - ((2 * BORDER) + SCROLLBAR); 
	int y = ln2y(lnr);
	
	if (lnr > nrOfLines - 1) {
	    g.setColor(colors[BACKGROUND_COLOR]);
	    g.fillRect(BORDER, y - fontLeading, w, fontHeight);
	    return;
	}
	int s = lines[lnr];
	int e = (lnr < nrOfLines - 1) ? lines[lnr + 1] - 1 : textLength;
	int xs = pos2x(selStart) - movedRight;
	int xe = pos2x(selEnd) - movedRight;

	if ((selStart < s) && (selEnd > e)) {
	    g.setColor(selectColor(colors));
	    g.fillRect(BORDER, y - fontLeading, w, fontHeight);
	} else {
	    g.setColor(backGroundColor(colors));
	    g.fillRect(BORDER, y - fontLeading, w, fontHeight);
		
	    if ((selStart >= s) && (selStart <= e)) {
		g.setColor(selectColor(colors));
		    
		if (selEnd > e) {
		    g.fillRect(xs, y - fontLeading, (w + BORDER) - xs, fontHeight);
		} else if (selStart == selEnd) { 
		    g.fillRect(xs, y - fontLeading, 1, fontHeight);
		} else { 
			g.fillRect(xs, y - fontLeading, xe - xs, fontHeight);
		}		
	    } else if ((selEnd >= s) && (selEnd <= e)) {
		g.setColor(selectColor(colors));
		g.fillRect(BORDER, y - fontLeading, xe - BORDER, fontHeight);
	    }	    
	}
	g.setColor(colors[FOREGROUND_COLOR]);
	g.drawChars(text, s, e - s, MARGIN - movedRight, y + fontAscent);
    }    

    /**
     * return the background color
     */
    Color backGroundColor(Color colors[]) {
	TextArea area = (TextArea)target;
	return (area.isEditable() ? colors[BACKGROUND_COLOR].brighter()
		                  : colors[BACKGROUND_COLOR]);
    }

    /**
     * return the selection color
     */
    Color selectColor(Color colors[]) {
	return (colors[SHADOW_COLOR]);
    }
    
    /**
     * get the value back from the scrollbar
     */
    public void notifyValue(Object obj, int c, int b) {
    	Graphics g = getGraphics();
	if ((TinyVerticalScrollbar)obj == vsb) {
	    scrollVertical(g, b - topLine);
	    resizeScrollbars();
	    paintVerScrollbar(g);
	    paintHorScrollbar(g);
	} else if ((TinyHorizontalScrollbar)obj == hsb) {
	    scrollHorizontal(g, b - movedRight); 
	    resizeScrollbars();
	    paintHorScrollbar(g);
	}
    	g.dispose();
    }

    /**
     * is (x, y) in the horizontal scrollbar
     */
    boolean inHorScrollbar(int x, int y) {
	int w = width - getScrollbarPad(vsb.max);
	int h = height - getScrollbarPad(hsb.max);
    	return (getScrollbarPad(hsb.max) != 0) 
	    &&  (x >= 0) && (x <= w) && (y > h);
    }
    
    /**
     * is (x, y) in the vertical scrollbar
     */
     boolean inVerScrollbar(int x, int y) {
	int w = width - getScrollbarPad(vsb.max);
	int h = height - getScrollbarPad(hsb.max);
	return (getScrollbarPad(vsb.max) != 0)
	    && (x > w) && (y >= 0) && (y <= h);
    }
    
    /**
     * is (x, y) in the Textfield
     */ 
    boolean inTextField(int x, int y) {
	int w = width - getScrollbarPad(vsb.max);
	int h = height - getScrollbarPad(hsb.max);
	return (x < w) && (y < h);
    }

    /**
     * Handle events.
     */
    public boolean handleWindowEvent(Event evt) {
	TextArea area = (TextArea)target;
	switch (evt.id) {
			
	case Event.MOUSE_DOWN:
	    if (inVerScrollbar(evt.x, evt.y)) {
		evt.x -= width - SCROLLBAR;
		mode = VERSCROLLBAR;
		int h = height - getScrollbarPad(hsb.max);
		return vsb.handleWindowEvent(evt, SCROLLBAR, h);
	    } else if (inHorScrollbar(evt.x, evt.y)) {	
		evt.y -= height - SCROLLBAR;
		mode = HORSCROLLBAR;
		int w = width - getScrollbarPad(vsb.max);
		return hsb.handleWindowEvent(evt, w, SCROLLBAR);
	    } else if (inTextField(evt.x, evt.y)) {
		if (!hasFocus() && isEnabled()) {
		    requestFocus();
		}
	    	mode = TEXTAREA;
		mpos = xy2pos(evt.x + movedRight, evt.y);
	    	select(mpos, mpos);
		return true;
	    }
	    break;
	
	case Event.MOUSE_UP:
	    if(mode == VERSCROLLBAR) {
		return vsb.handleWindowEvent(evt, 0, 0);
	    } else if(mode == HORSCROLLBAR) {
		return hsb.handleWindowEvent(evt, 0, 0);
	    } else if ((evt.when - mtm < 200) && (inTextField(evt.x, evt.y))) {
		select(0, textLength);
	    }
	    mtm = evt.when;
	    return true;

	case Event.MOUSE_DRAG:
	    
	    if (mode == VERSCROLLBAR) {
		evt.x -= width - SCROLLBAR;
		return vsb.handleWindowEvent(evt, SCROLLBAR, height - SCROLLBAR);
	    } else if (mode == HORSCROLLBAR) {	
		evt.y -= height - SCROLLBAR;
		return hsb.handleWindowEvent(evt, width - SCROLLBAR, SCROLLBAR);
	    } else if (mode == TEXTAREA) {
		int mpos2 = xy2pos(evt.x + movedRight, evt.y);
	    	if (mpos < mpos2) {
		    select(mpos, mpos2);	
		    scrollTo(selEnd);
		} else {
		    select(mpos2, mpos);
		    scrollTo(selStart); 
		}
		return true;
	    }
	 

	case Event.GOT_FOCUS:
	    hasFocus = true;
	    repaint();
	    return true;
	
	case Event.LOST_FOCUS:
	    hasFocus = false;
	    repaint();
	    return true;
	
	case Event.KEY_ACTION:
	    switch(evt.key) {
	        case Event.UP:
		     int ln = lineNumber(selStart);
		     if(ln > 0) {
			 int xo = pos2x(selStart);
			 int posn = xPos(xo, ln - 1);
		    	 select(posn, posn);
		     	 scrollTo(selStart);
		     }	
		     return false;
	 
	    	case Event.LEFT:
		    if(selStart > 0) {
			select(selStart - 1, selStart - 1);
		        scrollTo(selStart);
		    }
		    return false;
		    
	    	case Event.RIGHT:
		    if(selStart < textLength) {
			select(selStart + 1, selStart + 1);
		        scrollTo(selStart);
		    }
		    return false;
		    	
		case Event.DOWN:
		    int lnr = lineNumber(selStart);
		    if(lnr + 1 < nrOfLines) { 
			int xo = pos2x(selStart);
			int posn = xPos(xo, lnr + 1);
		    	select(posn, posn);
		    	scrollTo(selStart);
		    }
		    return false;

	        case Event.HOME:
		    // move insertion cursor to the beginning of the line
		    int posleft = xPos(0, lineNumber(selStart));
		    select(posleft, posleft);
		    scrollTo(selStart);
		    return false;

	        case Event.END:
		    // move insertion cursor to the beginning of the line
		    int posright = xPos(0, lineNumber(selStart)+1) - 1;
		    select(posright, posright);
		    scrollTo(selStart);
		    return false;
	    	}
         case Event.KEY_PRESS:
	     if(area.isEditable()) {
	    	switch (evt.key) {
	    		
		case '\r':
		    if(selStart != selEnd) {
			delete(selStart, selEnd);
		    }
		    insertText("\r", selStart);
		    return false;
	   	
		case '\n':
	    	    if(selStart != selEnd) {
		       delete(selStart, selEnd);
		    }
		    insertText("\n", selStart);
		    return false;
	   		
	        case 8:
	        case 127:
		    if(selStart != selEnd) {
		        delete(selStart, selEnd);
		    }		    
		    else if(selStart > 0) {
		        delete(selStart-1, selStart);
		    }
		    return false;

		case 24: /* control-X */
		    if (selStart == selEnd) {
			setClipboard("");
		    } else {
			setClipboard(String.valueOf(text, selStart, selEnd - selStart));
			if (((TextArea)target).isEditable()) {
			    delete(selStart, selEnd);
			} else {
			    Toolkit.getDefaultToolkit().beep();
			}
		    }
		    return false;

		case 3:  /* control-C */
		    if (selStart == selEnd) {
			setClipboard("");
		    } else {
			setClipboard(String.valueOf(text, selStart, selEnd - selStart));
		    }
		    return false;

		case 22: /* control-V */
		    if (((TextArea)target).isEditable()) {
			int pos = selStart + getClipboard().length();
			if (selStart == selEnd) {
			    insertText(getClipboard(), selStart);
			} else {
			    replaceText(getClipboard(), selStart, selEnd);
			}
			select(pos, pos);
		    } else {
			Toolkit.getDefaultToolkit().beep();
		    }
		    return false;

	        default:
		    char data[] = {(char)evt.key};
		    if(selStart != selEnd) {
		        replaceRange(new String(data), selStart, selEnd);
		    }
		    else {
		        insert(new String(data), selStart);
		    }
		    return false;
	        }
            }
	}	    
	return false;
    }
    
    /**
     * scroll the text to the position "pos". Pos is the position in the text.
     * if the pos is already on the screen then do nothing
     */
    void scrollTo(int pos) {
	
	int lnr = lineNumber(pos);
	int l = linesInWindow();
	Graphics g = getGraphics();
	
	// scroll vertical
	if (lnr >= l + topLine) {
	    if (selStart < selEnd) {
		scrollVertical(g, 1);
	    } else if (lnr == topLine + l) {
		scrollVertical(g, (l + 1) / 2);
	    } else {
		int i = (topLine + l) - 1;
		int j = (lnr - i) + (l / 2);
		scrollVertical(g, j);
	    }
	} else if (lnr < topLine) {
	    if(selStart < selEnd) {
		scrollVertical(g, -1);
	    } else if(lnr + 1 == topLine) {
		scrollVertical(g, (-l - 1) / 2);
	    } else {
		scrollVertical(g, lnr - (topLine + 1));
	    }
	}
 	
	int x = pos2x(pos);
	int w = width - (MARGIN + SCROLLBAR);
	int w2 = w - MARGIN;
	int xMax = w + movedRight;
	int xMin = movedRight + MARGIN;

	// scroll horizontal
	if (x > xMax) {
	    int diff = x - xMax;
	    scrollHorizontal(g, diff);
	} else if (x < xMin) {
	    int diff = x - xMin;
	    scrollHorizontal(g, diff);
	}
	resizeScrollbars();
	paintHorScrollbar(g);
	paintVerScrollbar(g);
	g.dispose();
    }

    /**
     * scroll horizontal 
     */
    void scrollHorizontal(Graphics g, int x) {
	Graphics ng = g.create();
	int w = width - ((2 * BORDER) + SCROLLBAR);
	int h = height - ((2 * MARGIN) + SCROLLBAR);
	movedRight += x;

	if (x < 0) {
	    ng.copyArea(BORDER, BORDER, w + x, h, -x, 0);
	    ng.clipRect(BORDER, BORDER, -x, h);
	    int n = linesInWindow();
	    int e = Math.min((topLine + n) - 1, nrOfLines - 1);
	    paintLines(ng, vsb.val, e);	
	} else if (x > 0) {
	    ng.copyArea(BORDER + x, MARGIN, w - x, h, -x, 0);
	    ng.clipRect((BORDER + w) - x, MARGIN, x, h);
	    int n = linesInWindow();
	    int e = Math.min((topLine + n) - 1, nrOfLines - 1);
	    paintLines(ng, vsb.val, e);
	}
	ng.dispose();
    }	

    /**
     * scroll the text "ml"  times vertical.
     */
    void scrollVertical(Graphics g, int ml) {
			
	if (topLine < -ml) {
	    ml = -topLine;
	}
	
	if (ml < 0) {
	    int l = linesInWindow();
	    topLine += ml;
            
	    if(ml + l > 0) {
		int hg = ((ml + l) * fontHeight);
		int w = width - ((2 * BORDER) + SCROLLBAR);
		g.copyArea(BORDER, MARGIN, w, hg, 0, -ml * fontHeight);  
	    } 	
	    int le = Math.min(nrOfLines - 1, topLine + -ml); 
	    paintLines(g, topLine, le);
	} else if (ml > 0) {
	    int l = linesInWindow();
	    topLine += ml;
	    	    
	    if (ml < l) {
		int y = MARGIN + (ml * fontHeight); 
		int hg = ((l - ml) * fontHeight);
		int w = width - ((2 * BORDER) + SCROLLBAR);
		g.copyArea(BORDER, y, w, hg, 0, -ml * fontHeight);  
	    }
	    paintLines(g, topLine + (l - ml), (topLine + l) - 1);
	}
    }
    
    /**
     * setEditable
     */
    public void setEditable(boolean editable) {
	repaint();
    }
    
    /**
     * get the selection start
     */
    public int getSelectionStart() {
	return selStart;
    }
    
    /**
     * get the selection end
     */
    public int getSelectionEnd() {
	return selEnd;
    }
    
    /**
     * get the text
     */
    public String getText() {
	return String.valueOf(text, 0, textLength);
    }
    
    /**
     * select the text between selStart and selEnd
     */
    public void select(int s, int e) {
    	
	if ((s > textLength) || (s < 0) || (e > textLength) || (e < 0)) {
	    return;
	}
	
	if ((this.selStart == s) && (this.selEnd == e)) {
	    return;
	}
	
	int startOld = this.selStart;
	int endOld = this.selEnd;
	selEnd = e;
	selStart = s;
 	Graphics g = getGraphics();
		
	if ((selStart >= endOld) || (selEnd <= startOld)) {
	    paintRange(g, startOld, endOld);
	    paintRange(g, selStart, selEnd);	
	} else {
	    int start = Math.min(selStart, startOld);
	    int end = Math.max(selStart, startOld);
	    
	    if (start != end) {
		paintRange(g, start, end);
	    }	
	    start = Math.min(selEnd, endOld);
	    end = Math.max(selEnd, endOld);
				  
	    if (start != end) {
		paintRange(g, start, end);
	    }
   	}   
	g.dispose();
    }

    /**
     * Set the text.
     */
    public void setText(String txt) {
	if (txt == null) {
	    txt = "";
	}
	textLength = txt.length();
	text = new char[textLength+100]; 
	int nr = Math.max(textLength / 20, 1);
	lines = new int[nr];
	txt.getChars(0, textLength, text, 0); 
	updateIndex(0);
	selStart = 0;
	selEnd = selStart;
	topLine = 0;
	movedRight = 0;
	target.repaint(); 
    }
    
    /** 
     * insert the text "txt on position "pos" in the array lines
     */
    public void insert(String txt, int p) {
	TextArea area = (TextArea)target;
	
	if((p >= 0) && (p <= textLength)) {
	    int l = txt.length();
	    int lnr = lineNumber(p);
	    char ln[] = new char[l];
	    txt.getChars(0, l, ln, 0);
	    
	    if (text.length < textLength + l) {
		char tmp[] = new char[textLength + l + 20];
		System.arraycopy(text, 0, tmp, 0, textLength);
		text = tmp;
	    }
	    System.arraycopy(text, p, text, p + l, textLength - p);
	    System.arraycopy(ln, 0, text, p, l);
	    textLength += l;
	    int oldStart = selStart;
	    selStart += l;
	    selEnd = selStart;
	    int nr = nrOfLines;
	    updateIndex(lnr);    
	    int li = linesInWindow();
	    int le = (nrOfLines != nr) ? (topLine + li) - 1 : lnr;
	    Graphics g = getGraphics();
	    if (resizeScrollbars() == 0) {
		 paintLines(g, lnr, le);
		 paintHorScrollbar(g);
		 paintVerScrollbar(g);
	    } else {
		 paintComponent(g);
	    }
	    g.dispose();
	    scrollTo(selStart);
	}
    }
    
    /**
     * replace the text between the position "s" and "e" with "txt"
     */
    public void replaceRange(String txt, int s, int e) {
    		
	if ((e > textLength) || (s > e) || (s < 0)) {
	    return;
	}
	TextArea area = (TextArea)target;
	delete(s, e);
	insert(txt, s);    
    }
    
    /**
     * remove the text from the position start to the position end in the aray lines
     */
    protected void delete(int s, int e) {
    	int lnr = lineNumber(s);
	int length = e - s;	
	System.arraycopy(text, e, text, s, textLength - e); 
	textLength -= length;
	selStart = s;
	selEnd = s;
	int nr = nrOfLines;
	updateIndex(lnr);    
	Graphics g = getGraphics();
	int li = linesInWindow();
	int le = (nrOfLines != nr) ? (topLine + li) - 1 : lnr;
	paintLines(g, lnr, le);
    	g.dispose();
	scrollTo(selStart);
    }

    public void requestFocus() {
	requestFocus(this);
    }

    /**
     * DEPRECATED
     */
    public void insertText(String txt, int pos) {
      	insert(txt, pos);
    }

    /** 
     * to be implemented.
     */
    public void setCaretPosition(int position) {
        select(position, position);
    }
    
    /** 
     * to be implemented.
     */
    public int getCaretPosition() {
        return 0;
    }

    /**
     * DEPRECATED
     */
    public void replaceText(String txt, int start, int end) {
      	replaceRange(txt, start, end);
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize(int rows, int cols) {
	return getMinimumSize(rows, cols);
    }

    /**
     * DEPRECATED
     */
    public Dimension preferredSize(int rows, int cols) {
	return getPreferredSize(rows, cols);
    }

}

