/*
 * @(#)TinyTextFieldPeer.java	1.21 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.tiny;

import java.awt.*;
import java.awt.peer.*;

class TinyTextFieldPeer extends TinyComponentPeer implements TextFieldPeer {
    
    public final static int BORDER = 2;
    public final static int MARGIN = 4;

    boolean hasFocus;
    String text;
    int selStart;
    int selEnd;
    char echoChar;
    int mpos;
    int moved;
    long mtm;

    /**
     * Create the text field.
     */
    TinyTextFieldPeer(TextField target) {
	winCreate((TinyComponentPeer)target.getParent().getPeer());
	init(target);
	text = target.getText();
	if (text == null) {
	    text = "";
	}
        selStart = target.getSelectionStart();
        selEnd = target.getSelectionEnd();
	echoChar = target.getEchoChar();
    }

    /**
     * Minumum size.
     */
    public Dimension getMinimumSize() {
	TextField txt = (TextField)target;
	FontMetrics fm = getFontMetrics(txt.getFont());
	return new Dimension(fm.stringWidth(text) + TEXT_XPAD, 
			     fm.getMaxAscent() + fm.getMaxDescent() + fm.getLeading() + TEXT_YPAD);
    }

    public Dimension getPreferredSize(int cols) {
	return getMinimumSize(cols);
    }

    public Dimension getMinimumSize(int cols) {
	FontMetrics fm = getFontMetrics(target.getFont());
	return new Dimension(fm.charWidth('0') * cols + TEXT_XPAD, 
			     fm.getMaxAscent() + fm.getMaxDescent() + fm.getLeading() + TEXT_YPAD);
    }

    public boolean isFocusTraversable() {
	return true;
    }

    /**
     * Position to x
     */
    int pos2x(int pos) {
	FontMetrics fm = getFontMetrics(target.getFont());
	int x = MARGIN, widths[] = fm.getWidths();
	if (echoChar == 0) {
	    for (int i = 0 ; i < pos ; i++) {
		x += widths[text.charAt(i)];
	    }
	} else {
	    x += widths[echoChar] * pos;
	}
	return x;
    }

    /**
     * x 2 position
     */
    int xy2pos(int x, int y) {
	if (y < 0) {
	    return 0;
	}
	if (y > height) {
	    return text.length();
	}
	FontMetrics fm = getFontMetrics(target.getFont());
	int widths[] = fm.getWidths();
	int len = text.length();
	x -= MARGIN;
	for (int i = 0 ; i < len ; i++) {
	    int w = widths[(echoChar == 0) ? text.charAt(i) : echoChar];
	    if (x < w/2) {
		return i;
	    }
	    x -= w;
	}
	return text.length();
    }

    /**
     * Paint the text field.
     */
    public void paintComponent(Graphics g) {
	TextField txt = (TextField)target;
	Color colors[] = getGUIcolors();
	g.setFont(txt.getFont());
	int w = width - (2 * BORDER); 
	
	g.setColor(txt.isEditable() ? colors[HIGHLIGHT_COLOR]
		   		    : colors[BACKGROUND_COLOR]);
	g.fillRect(BORDER, BORDER, w, height - (2 * BORDER));

	if (hasFocus) {
	    g.setColor(colors[FOREGROUND_COLOR]);
	    g.drawRect(0, 0, width-1, height-1);
	    // g.drawRect(1, 1, width-3, height-3);
	} else {
	    g.setColor(colors[BACKGROUND_COLOR]);
	    g.drawRect(0, 0, width-1, height-1);
	}
	draw3DRect(g, colors, 1, 1, width-3, height-3, false);
	
	if (text != null) {
	   g.clipRect(BORDER, MARGIN, w, height - (2 * MARGIN));
	   paintLine(g, colors);
	}
    }

    /** 
     * Paint a range in the textField given by the values s and e
     */
    void paintRange(Graphics g, int s, int e) {
	TextField txt = (TextField)target;
	int xs = pos2x(s) - moved;
	int xe = pos2x(e) - moved;
	Graphics ng = g.create();
	int w = (s == e) ? 1 : xe - xs;
	ng.clipRect(xs, MARGIN, w, height - (2 * MARGIN));
	paintLine(ng, getGUIcolors());
    	ng.dispose();
    }

    /**
     * draw the text to the textField
     */
    void paintLine(Graphics g, Color colors[]) {
	FontMetrics fm = g.getFontMetrics();
		
	int w = width - BORDER; 
	int h = height - (2 * MARGIN);
	int xs = pos2x(selStart) - moved;
	int xe = pos2x(selEnd) - moved;

	if ((xs < MARGIN) && (xe > w)) {
	   g.setColor(selectColor(colors));
           g.fillRect(BORDER, MARGIN, w - BORDER, h);
	} else {
	   g.setColor(backGroundColor(colors));
           g.fillRect(BORDER, MARGIN, w - BORDER, h);
	   
	   if ((xs >= MARGIN) && (xs <= w)) {
		g.setColor(selectColor(colors));
		
		if (xe > w) {
		    g.fillRect(xs, MARGIN, w - xs, h);
		} else if (xs == xe) {
		    g.fillRect(xs, MARGIN, 1, h);
		} else {
		    g.fillRect(xs, MARGIN, xe - xs, h);
		}
	   } else if ((xe >= MARGIN) && (xe <= w)) {
	       g.setColor(selectColor(colors));
	       g.fillRect(BORDER, MARGIN, xe - BORDER, h);
	   }
	}
	g.setColor(colors[FOREGROUND_COLOR]);  
	int x = MARGIN - moved;
	if (echoChar == 0) {
	    g.drawString(text, x,
			(height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
	} else {
	    char data[] = new char[text.length()];
	    for (int i = 0 ; i < data.length ; i++) {
		data[i] = echoChar;
	    }
	    g.drawChars(data, 0, data.length, x, 
			(height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
	}
    }    
    

    /**
     * return the background color
     */
    Color backGroundColor(Color colors[]) {
	TextField area = (TextField)target;
	return (area.isEditable() ? colors[BACKGROUND_COLOR].brighter()
		                  : colors[BACKGROUND_COLOR]);
    }

    /**
     * return the selection color
     */
    Color selectColor(Color colors[]) {
	return (colors[SHADOW_COLOR]);
    }

    /**
     * Handle events.
     */
    public boolean handleWindowEvent(Event evt) {
	TextField field = (TextField)target;
	switch (evt.id) {
	case Event.MOUSE_DOWN:
	    if (!hasFocus && isEnabled()) {
		requestFocus();
	    }
	    mpos = xy2pos(evt.x + moved, evt.y);
	    select(mpos, mpos);
	    return true;

	case Event.MOUSE_UP:
	    if (evt.when - mtm < 200) {
		select(0, text.length());
	    }
	    mtm = evt.when;
	    return true;

	case Event.MOUSE_DRAG:
	    int mpos2 = xy2pos(evt.x + moved, evt.y);
	    if (mpos < mpos2) {
		select(mpos, mpos2);
	    } else {
		select(mpos2, mpos);
	    }
	    scrollHorizontal(mpos2);
	    return true;

	case Event.GOT_FOCUS:
	    hasFocus = true;
	    repaint();
	    return true;
	case Event.LOST_FOCUS:
	    hasFocus = false;
	    repaint();
	    return true;

	case Event.KEY_ACTION:
	    switch(evt.key) {
	    	case Event.LEFT:
		    if(selStart > 0) {
			select(selStart - 1, selStart - 1);
		        scrollHorizontal(selStart);
		    }
		    return false;

	    	case Event.RIGHT:
		    if(selEnd < text.length()) {
			select(selEnd + 1, selEnd + 1);
		        scrollHorizontal(selStart);
		    }
		    return false;

	        case Event.HOME:
		    select(0, 0);
		    scrollHorizontal(selStart);
		    return false;

	        case Event.END:
		    int n = text.length();
		    select(n, n);
		    scrollHorizontal(selStart);
		    return false;
	    }	
	    return false;	    
	case Event.KEY_PRESS:
	    if (field.isEditable()) {
		switch (evt.key) {
	      	case '\r':
	      	case '\n':
		    targetEvent(Event.ACTION_EVENT, text);
		    return false;

	      	case 8:
	      	case 127:
		    if (selStart != selEnd) {
			delete(selStart, selEnd);
		    } else if (selStart > 0) {
			delete(selStart-1, selStart);
		    }
		    return false;

	    	case 1: // ctrl-a
		    select(0, 0);
		    scrollHorizontal(selStart);
		    return false;
	   	case 2: // ctrl-b
		    if (selStart > 0) {
			select(selStart-1, selStart-1);
		    	scrollHorizontal(selStart);
		    }
		    return false;
		case 5: // ctrl-e
		    select(text.length(), text.length());
		    scrollHorizontal(selStart);
		    return false;
		case 6: // ctrl-f
		    if (selEnd < text.length()) {
			select(selEnd+1, selEnd+1);
			scrollHorizontal(selStart);
		    }
		    return false;

		case 24: /* control-X */
		    if (selStart == selEnd) {
			TinyTextAreaPeer.setClipboard("");
		    } else {
			String str = text.substring(selStart, selEnd);
			TinyTextAreaPeer.setClipboard(str);

			if (((TextField)target).isEditable()) {
			    delete(selStart, selEnd);
			} else {
			    Toolkit.getDefaultToolkit().beep();
			}
		    }
		    return false;

		case 3:  /* control-C */
		    if (selStart == selEnd) {
			TinyTextAreaPeer.setClipboard("");
		    } else {
			String str = text.substring(selStart, selEnd);
			TinyTextAreaPeer.setClipboard(str);
		    }
		    return false;

		case 22: /* control-V */
		    if (((TextField)target).isEditable()) {
			String str = TinyTextAreaPeer.getClipboard();
			int pt = selStart + str.length();
			if (selStart == selEnd) {
			    insert(selStart, str);
			} else {
			    delete(selStart, selEnd);
			    insert(selStart, TinyTextAreaPeer.getClipboard());
			}
			select(pt, pt);
		    } else {
			Toolkit.getDefaultToolkit().beep();
		    }
		    return false;

	      	default:
		    if (selStart != selEnd) {
			delete(selStart, selEnd);
		    }	
		    char data[] = {(char)evt.key};
		    insert(selStart, new String(data));
		    return false;
		}
	    }	    
	}
	return false;
    }

    /** 
     * scroll to the position "pos"
     */
    void scrollHorizontal(int pos) {
	
	int x = pos2x(pos);
	int w = width - (2 * BORDER);
	int xmax = w + moved;
	int xmin = MARGIN + moved;
	
	if ((x > xmax) || (x < xmin)) {
	    int h = height - (2 * MARGIN);
	    Graphics g = getGraphics();
	    moved += (x > xmax) ? x - xmax : x - xmin;
	    g.clipRect(BORDER, MARGIN, w, h);
	    paintLine(g, getGUIcolors());
	    g.dispose();
	}
    }

    public void setEditable(boolean editable) {
	repaint();
    }
    
    public void setEchoChar(char c) {
    }
    
    public int getSelectionStart() {
	return selStart;
    }
    
    public int getSelectionEnd() {
	return selEnd;
    }
    
    public String getText() {
	return text;
    }
    
    void insert(int sel, String str) {
	text = text.substring(0, sel) + str + text.substring(sel);
	selStart += 1;
	selEnd = selStart;
    	Graphics g = getGraphics();
	g.clipRect(BORDER, MARGIN, width - (2 * BORDER), height - (2 * MARGIN));
	paintLine(g, getGUIcolors());
	g.dispose();
	scrollHorizontal(selStart);
		    
    }
    
    void delete(int start, int end) {
	text = text.substring(0, start) + text.substring(end);
	selStart = start;
	selEnd = selStart;
    	Graphics g = getGraphics();
	g.clipRect(BORDER, 4, width - (2 * BORDER), height - (2 * MARGIN));
	paintLine(g, getGUIcolors());
	g.dispose();
	scrollHorizontal(selStart);	    	    
    }

    public void select(int s, int e) {
	
	if ((s > text.length()) || (s < 0) || (e > text.length()) || (e < 0)) {
	    return;
	}
	
	if ((this.selStart == s) && (this.selEnd == e)) {
	    return;
	}
	
	int startOld = this.selStart;
	int endOld = this.selEnd;
	this.selEnd = e;
	this.selStart = s;
 	
	Graphics g = getGraphics();
	g.clipRect(BORDER, MARGIN, width - (2 * BORDER), height - (2 * MARGIN));	
	
	if ((selStart >= endOld) || (selEnd <= startOld)) {
	    paintRange(g, startOld, endOld);
	    paintRange(g, selStart, selEnd);	
	} else {
	    int start = Math.min(selStart, startOld);
	    int end = Math.max(selStart, startOld);
	    
	    if (start != end) {
		paintRange(g, start, end);
	    }	
	    start = Math.min(selEnd, endOld);
	    end = Math.max(selEnd, endOld);
				  
	    if (start != end) {
		paintRange(g, start, end);
	    }
   	}   
	g.dispose();
    }
      
    public void setText(String txt) {
	text = (txt == null) ? "" : txt;
	selStart = text.length();
	selEnd = text.length();
	moved = 0;
	target.repaint();
    }

    /** 
     * to be implemented.
     */
    public void setCaretPosition(int position) {
        select(position, position);
    }
    
    /** 
     * to be implemented.
     */
    public int getCaretPosition() {
        return 0;
    }

    /**
     * DEPRECATED
     */
    public void setEchoCharacter(char c) {
    	setEchoChar(c);
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize(int cols) {
	return getMinimumSize(cols);
    }

    /**
     * DEPRECATED
     */
    public Dimension preferredSize(int cols) {
	return getPreferredSize(cols);
    }

}
