/*
 * @(#)TinyToolkit.java	1.35 98/06/29
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.tiny;

import java.util.Properties;
import java.util.Hashtable;
import java.awt.*;
import java.awt.peer.*;
import java.awt.image.*;
import java.awt.datatransfer.Clipboard;
import java.net.URL;
import sun.awt.image.ByteArrayImageSource;
import sun.awt.image.FileImageSource;
import sun.awt.image.URLImageSource;
import sun.awt.image.ImageRepresentation;
import sun.awt.SunToolkit;
import sun.awt.im.InputMethod;

class TinyInputThread extends Thread {
    TinyInputThread() {
	super("Tiny-AWT-Input");
	start();
    }
    public native void run();
}

class TinyEventThread extends Thread {
    TinyEventThread() {
	super("Tiny-AWT-Event");
	start();
    }
    public native void run();
}

public class TinyToolkit extends SunToolkit {
    static {
	java.security.AccessController.doPrivileged(
			  new sun.security.action.LoadLibraryAction("tawt"));
    }

    public TinyToolkit() {
        super();
	init();
	new TinyInputThread();
	new TinyEventThread();
    }

    public native void init();

    /*
     * Create peer objects.
     */

    public WindowPeer createWindow(Window target) {
	return new TinyWindowPeer(target);
    }

    public FramePeer  createFrame(Frame target) {
	return new TinyFramePeer(target);
    }

    public CanvasPeer createCanvas(Canvas target) {
	return new TinyCanvasPeer(target);
    }

    public PanelPeer createPanel(Panel target) {
	return new TinyPanelPeer(target);
    }

    public ButtonPeer createButton(Button target) {
	return new TinyButtonPeer(target);
    }

    public TextFieldPeer createTextField(TextField target) {
	return new TinyTextFieldPeer(target);
    }

    public ChoicePeer createChoice(Choice target) {
	return new TinyChoicePeer(target);
    }

    public LabelPeer createLabel(Label target) {
	return new TinyLabelPeer(target);
    }

    public ListPeer createList(List target) {
	return new TinyListPeer(target);
    }

    public CheckboxPeer createCheckbox(Checkbox target) {
	return new TinyCheckboxPeer(target);
    }

    public ScrollbarPeer createScrollbar(Scrollbar target) {
	return new TinyScrollbarPeer(target);
    }

    public ScrollPanePeer createScrollPane(ScrollPane target) {
	return new TinyScrollPanePeer(target);
    }

    public TextAreaPeer createTextArea(TextArea target) {
	return new TinyTextAreaPeer(target);
    }

    public DialogPeer createDialog(Dialog target) {
	return new TinyDialogPeer(target);
    }

    public FileDialogPeer createFileDialog(FileDialog target) {
	return new TinyFileDialogPeer(target);
    }

    public MenuBarPeer createMenuBar(MenuBar target) {
	return new TinyMenuBarPeer(target);
    }

    public MenuPeer createMenu(Menu target) {
	return new TinyMenuPeer(target);
    }

    public PopupMenuPeer createPopupMenu(PopupMenu target) {
	System.out.println("TinyToolkit: PopupMenuPeer not implemented yet");
	return null;
    }

    public MenuItemPeer createMenuItem(MenuItem target) {
	return new TinyMenuItemPeer(target);
    }

    public CheckboxMenuItemPeer createCheckboxMenuItem(CheckboxMenuItem target) {
	return new TinyCheckboxMenuItemPeer(target);
    }

    /*
     * Tiny toolkit is not yet developed for JDK1.1.
     */
    public FontPeer getFontPeer(String name, int style){
	return (FontPeer)null;
    }

    static native ColorModel makeColorModel();
    static ColorModel screenmodel;

    static ColorModel getStaticColorModel() {
	if (screenmodel == null) {
	    screenmodel = makeColorModel();
	}
	return screenmodel;
    }

    public ColorModel getColorModel() {
	return getStaticColorModel();
    }

    public native int getScreenResolution();
    protected native int getScreenWidth();
    protected native int getScreenHeight();

    public FontMetrics getFontMetrics(Font font) {
	return TinyFontMetrics.getFontMetrics(font);
    }

    public native void sync();

    static boolean prepareScrImage(Image img, int w, int h, ImageObserver o) {
	if (w == 0 || h == 0) {
	    return true;
	}
	TinyImage ximg = (TinyImage) img;
	if (ximg.hasError()) {
	    if (o != null) {
		o.imageUpdate(img, ImageObserver.ERROR|ImageObserver.ABORT,
			      -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep();
	return ir.prepare(o);
    }

    static int checkScrImage(Image img, int w, int h, ImageObserver o) {
	TinyImage ximg = (TinyImage) img;
	int repbits;
	if (w == 0 || h == 0) {
	    repbits = ImageObserver.ALLBITS;
	} else {
	    repbits = ximg.getImageRep().check(o);
	}
	return ximg.check(o) | repbits;
    }

    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return checkScrImage(img, w, h, o);
    }

    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return prepareScrImage(img, w, h, o);
    }

    public Image createImage(ImageProducer producer) {
	return new TinyImage(producer);
    }

    public PrintJob getPrintJob(Frame frame, String doctitle, Properties props) {
    //REMIND: Tiny can share print implemention with Motif
	throw new InternalError("not implemented yet");
    }

    public native void beep();

    public Clipboard getSystemClipboard() {
        System.out.println("TinyToolkit: system clipboard not implemented yet");
        return null;
    }

    static void postEvent(AWTEvent event) {
        theEventQueue.postEvent(event);
    }

    public static final Object targetToPeer(Object target) {
        return SunToolkit.targetToPeer(target);
    }

    public static final void targetDisposedPeer(Object target, Object peer) {
        SunToolkit.targetDisposedPeer(target, peer);
    }

    /**
     * Returns a new input method adapter for native input methods.
     */
    public InputMethod getInputMethodAdapter() {
        return null;
    }

}
