/*
 * @(#)PlatformMimeTable.java	1.11 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.net.www;

import java.io.*;

/**
 * The platform specific MIME table.  Constructing one of these
 * causes it to be initialized with some default contents.
 * 
 * @version 	1.11, 09/21/98
 * @author Steven B. Byrne
 */

class PlatformMimeTable extends MimeTable {

    PlatformMimeTable() {
	initializeDefaultEntries();

	loadUserMimeTypes();
    }

    // TODO: make this table driven
    private void initializeDefaultEntries() {
	addMimeEntry("content/unknown", "");
	addMimeEntry("application/octet-stream",
		     ".uu,.saveme,.dump,.hqx,.arc,.o,.a,.bin,.exe");
	addMimeEntry("application/octet-stream", ".saveme,.dump,.hqx,.arc,.o,.a,.bin,.exe,.z,.gz");

	/* Temporary only. */
	addMimeEntry("application/octet-stream", ".z,.gz",
		     fullImagePath("compress"));

	addMimeEntry("application/oda", ".oda");
	addMimeEntry("application/pdf", ".pdf");
	addMimeEntry("application/postscript", ".eps,.ai,.ps",
		     fullImagePath("ps"), MimeEntry.LAUNCH_APPLICATION,
		     "imagetool %s");
	addMimeEntry("application/rtf", ".rtf");
	addMimeEntry("application/x-dvi", ".dvi",
		     MimeEntry.LAUNCH_APPLICATION, "xdvi %s");
	addMimeEntry("application/x-hdf", ".hdf", MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-latex", ".latex");
	addMimeEntry("application/x-netcdf", ".cdf", MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-netcdf", ".nc");
	addMimeEntry("application/x-tex", ".tex");
	addMimeEntry("application/x-texinfo", ".texinfo,.texi");
	addMimeEntry("application/x-troff", ".t,.tr,.roff",
		     MimeEntry.LAUNCH_APPLICATION, "xterm -title troff -e sh -c \"nroff %s | col | more -w ; rm %s\"");
	addMimeEntry("application/x-troff-man", ".man",
		     MimeEntry.LAUNCH_APPLICATION, "xterm -title troff -e sh -c \"nroff -man %s | col | more -w ; rm %s\"");
	addMimeEntry("application/x-troff-me", ".me",
		     MimeEntry.LAUNCH_APPLICATION, "xterm -title troff -e sh -c \"nroff -me %s | col | more -w ; rm %s\"");
	addMimeEntry("application/x-troff-ms", ".ms",
		     MimeEntry.LAUNCH_APPLICATION, "xterm -title troff -e sh -c \"nroff -ms %s | col | more -w ; rm %s\"");
	addMimeEntry("application/x-wais-source", ".src,.wsrc");
	addMimeEntry("application/zip", ".zip",
		     fullImagePath("zip"),
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-bcpio", ".bcpio",
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-cpio", ".cpio",
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-gtar", ".gtar",
		     fullImagePath("tar"),
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-shar", ".shar",
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-shar", ".sh,.shar",
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-sv4cpio", ".sv4cpio",
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-sv4crc", ".sv4crc",
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-tar", ".tar",
		     fullImagePath("tar"),
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("application/x-ustar", ".ustar",
		     MimeEntry.SAVE_TO_FILE, null);
	addMimeEntry("audio/basic", ".snd,.au", fullImagePath("audio"),
		     MimeEntry.LAUNCH_APPLICATION, "audiotool %s");
	addMimeEntry("audio/x-aiff", ".aifc,.aif,.aiff", fullImagePath("aiff"));
	addMimeEntry("audio/x-wav", ".wav", fullImagePath("wav"));

	addMimeEntry("image/gif", ".gif", fullImagePath("gif"),
		     MimeEntry.LOAD_INTO_BROWSER, null);
//		     MimeEntry.LAUNCH_APPLICATION, "xv %s");
	addMimeEntry("image/ief", ".ief");
	addMimeEntry("image/jpeg", ".jfif,.jfif-tbnl,.jpe,.jpg,.jpeg",
		     fullImagePath("jpeg"),
		     MimeEntry.LOAD_INTO_BROWSER, null);
//		     MimeEntry.LAUNCH_APPLICATION, "xv %s");

	addMimeEntry("image/tiff", ".tif,.tiff", fullImagePath("tiff"));
	addMimeEntry("image/x-cmu-rast", ".ras");
	addMimeEntry("image/x-portable-anymap", ".pnm");
	addMimeEntry("image/x-portable-bitmap", ".pbm");
	addMimeEntry("image/x-portable-graymap", ".pgm");
	addMimeEntry("image/x-portable-pixmap", ".ppm");
	addMimeEntry("image/x-rgb", ".rgb");
	addMimeEntry("image/x-xbitmap", ".xbm,.xpm");
	addMimeEntry("image/x-xwindowdump", ".xwd");
	addMimeEntry("text/html", ".htm,.html", fullImagePath("html"));
	addMimeEntry("text/plain", ".text,.c,.cc,.c++,.h,.pl,.txt,.java,.el", fullImagePath("text"), MimeEntry.LOAD_INTO_BROWSER, null);
	addMimeEntry("application/rtf", ".rtx");
	addMimeEntry("text/tab-separated-values", ".tsv");
	addMimeEntry("text/x-setext", ".etx");
	addMimeEntry("video/mpeg", ".mpg,.mpe,.mpeg",
		     fullImagePath("mpeg"),
		     MimeEntry.LAUNCH_APPLICATION, "mpeg_play %s");
	addMimeEntry("video/quicktime", ".mov,.qt");
	addMimeEntry("application/x-troff-msvideo", ".avi",
		     fullImagePath("avi"));
	addMimeEntry("video/x-sgi-movie", ".movie,.mv");
	addMimeEntry("message/rfc822", ".mime");


//	add(new MimeEntry ("application/octet-stream", "loadtofile"));

    }


    private void addMimeEntry(String contentType, String ext, int action,
			      String command) {
	MimeEntry e = new MimeEntry(contentType, null, ext);
	e.setAction(action, command);

	add(e);
    }



    private void addMimeEntry(String contentType, String ext) {
	add(new MimeEntry(contentType, null, ext));
    }

    private void addMimeEntry(String contentType, String ext,
			      String imageFileName) {

	add(new MimeEntry(contentType, imageFileName, ext));
    }

    private void addMimeEntry(String contentType, String ext,
			      String imageFileName, int action, String command) {

	MimeEntry e = new MimeEntry(contentType, imageFileName, ext);
	e.setAction(action, command);
	add(e);
    }

    private String fullImagePath(String name) {
	return MimeEntry.defaultImagePath + File.separator + name + ".gif";
    }

    private void loadUserMimeTypes() {
	InputStream is = null;
	String slist[] = {
	    System.getProperty("user.mailcap"),
	    System.getProperty("user.home") + "/.mailcap",
	    "/etc/mailcap",
	    "/usr/etc/mailcap",
	    "/usr/local/etc/mailcap",
	    System.getProperty("hotjava.home", "/usr/local/hotjava") + "/lib/mailcap"
	};
	for (int i = 0; i < slist.length; i++) {
	    if (slist[i] != null) {
		try {
		    is = new FileInputStream(slist[i]);
		    break;
		} catch(Exception e) {
		}
	    }
	}
	if (is != null) {
	    try {
		parseMailcap(is);
		is.close();
	    } catch(IOException e) {}
	}

    }


    // TODO: make this parse a more interesting set of information, including
    //       the image file name to associate with this file type.
    private void parseMailcap(InputStream is) {
	try {
	    char b[] = new char[200];
	    int i = 0;
	    int c;
	    int slot = 0;
	    int eqpos = -1;
	    String mimeType = null;
	    String command = null;
	    String templateFileName = null;
	    while ((c = is.read()) >= 0) {
		if (c == ';' || c == '\n') {
		    while (i > 0 && b[i - 1] < ' ')
			i--;
		    if (slot <= 1) {
			String s;
			if (slot == 0 && i > 0 && b[i - 1] == '*')
			    i--;/* '*' is recognized by the trailing slash */
			if (i == 0)
			    s = "";
			else
			    s = String.copyValueOf(b, 0, i);
			if (slot == 0)
			    mimeType = s;
			else
			    command = s;
		    } else if (i > 0) {
			String key;
			String value;
			if (eqpos >= 0) {
			    key = String.valueOf(b, 0, eqpos);
			    value = String.copyValueOf(b, eqpos + 1, i - eqpos - 1);
			} else {
			    key = String.valueOf(b, 0, i);
			    value = null;
			}
			if (key.equalsIgnoreCase("nametemplate"))
			    templateFileName = value;
		    }
		    slot++;
		    i = 0;
		    if (c == '\n') {
			if (slot >= 2) {
			    int cmd = MimeEntry.LAUNCH_APPLICATION;
			
			    if ("loadtofile".equals(command)) {
				cmd = MimeEntry.SAVE_TO_FILE;
			    } 

			    add(new MimeEntry(mimeType, cmd, command,
					      templateFileName));
			}
			slot = 0;
		    }
		} else if (c == '#' && i == 0) {
		    while ((c = is.read()) >= 0 && c != '\n');
		} else {
		    if (c == '\\') {
			c = is.read();
			if (c == '\n')	/* line  continuation */
			    continue;
		    }
		    if ((c > ' ' || i > 0) && i < b.length) {
			if (c == '=')
			    eqpos = i;
			b[i++] = (char) c;
		    }
		}
	    }
	} catch(Exception e) {
	}
    }
	
    String getTempFileTemplate() {
	return "/tmp/%s";
    }
}

