/*
 * @(#)iomgr.h	1.34 98/09/15
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Very simple IO manager interface.  This needs to get fixed in
 * a big way.
 */

#ifndef _JAVASOFT_GREEN_IOMGR_H_
#define _JAVASOFT_GREEN_IOMGR_H_

int nonblock_io(int,int);

/*
 *  nonblock_io takea a boolean value as its second argument.  When that
 *  boolean is true, the filedescriptor operates in non-blocking mode.
 */
#define IO_DONTBLOCK    1
#define IO_BLOCK        0

/*
 * We keep a reference count of active readers and writers on an
 * fd so that when we receive SIGIO in signals.c we can fill in the
 * pollTable with the right set of POLLIN/POLLOUT flags.  This helps
 * us avoid spurious notifications.
 */

typedef struct {
    short readers;
    short writers;
} fd_reference;

#define FD_INACTIVE(fd) \
    ((fd_ref[(fd)].readers == 0) && (fd_ref[(fd)].writers == 0))
#define FD_READERS(fd) (fd_ref[(fd)].readers)
#define FD_ADD_READER(fd) (fd_ref[(fd)].readers++)
#define FD_REMOVE_READER(fd) \
if (1) {						\
    fd_ref[(fd)].readers--;		\
    sysAssert(fd_ref[(fd)].readers >= 0); \
} else ((void) 0)

#define FD_WRITERS(fd) (fd_ref[(fd)].writers)
#define FD_ADD_WRITER(fd) (fd_ref[(fd)].writers++)
#define FD_REMOVE_WRITER(fd) \
if (1) {						\
    fd_ref[(fd)].writers--;		\
    sysAssert(fd_ref[(fd)].writers >= 0); \
} else ((void) 0)


extern fd_reference *fd_ref;
extern char *fd_flags;
extern sys_mon_t **fdmon;
extern sys_mon_t _io_lock;
extern sys_mon_t _dl_lock;
extern bool_t threadBootstrappedP;

#define IO_LOCK_INIT() VM_CALL(monitorRegister)(&_io_lock, "Monitor IO lock");
#define IO_LOCK(self)    sysMonitorEnter(self, &_io_lock)
#define IO_UNLOCK(self)  sysMonitorExit(self, &_io_lock)
#define IO_LOCKED(self)  sysMonitorEntered(self, &_io_lock)

#define GET_INTERRUPT_IO() \
	sysThreadSelf()?sysThreadSelf()->interruptible_io:FALSE

#define FD_NBINIT       0x1
#define FD_CLOSED       0x2
#define FD_NONBLOCKING  0x4	/* does the user expect non-blocking? */

#define	check_single_fd(fd)		      \
    if ((fd_flags[fd] & FD_NBINIT) == 0) {    \
	nonblock_io(fd, IO_DONTBLOCK);	      \
    }

#define NONBLOCKING_FD(fd) (fd_flags[(fd)] & FD_NONBLOCKING)

#define closed_fd(fd) \
    ((fd_flags[fd] & FD_CLOSED) != 0)

typedef int (*sysfunc_t)();

typedef struct syscall {
    char *sym;
    sysfunc_t addr;
} syscall_t;

extern syscall_t systable[];

#define SYS_WRITE     0
#define SYS_WRITEV    1
#define SYS_READ      2
#define SYS_READV     3
#define SYS_GETMSG    4
#define SYS_PUTMSG    5
#define SYS_POLL      6
#define SYS_OPEN      7
#define SYS_CLOSE     8
#define SYS_PIPE      9
#define SYS_FCNTL     10
#define SYS_DUP       11
#define SYS_CREAT     12
#define SYS_ACCEPT    13
#define SYS_RECV      14
#define SYS_SEND      15
#define SYS_RECVFROM  16
#define SYS_SENDTO    17
#define SYS_SOCKET    18
#define SYS_CONNECT   19
#define SYS_SELECT    20
#define SYS_IOCTL     21
#define SYS_DLOPEN    22
#define SYS_DLCLOSE   23
#define SYS_DLERROR   24
#define SYS_DLSYM     25
#define SYS_DLADDR    26
#define SYS_FILBUF    27

int system_close(int fd);

void PrimordialThreadBootstrapped(bool_t b);
void InitializeAsyncIO(void);

#endif /* !_JAVASOFT_GREEN_IOMGR_H_ */
