/*
 * @(#)itimer.c	1.24 98/07/08
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * The clock interface used when the Green threads API is on Unix.
 */

#include <time.h>
#include <sys/time.h>

#include "hpi_impl.h"

#include "threads_md.h"
#include "monitor_md.h"
#include "itimer_md.h"

extern int gettimeofday(struct timeval *, void *);

/*
 * The minimum clock resolution in uSeconds.
 */
#define CLOCK_RESOLUTION    1000

timeval_t
currentTime()
{
    struct timeval tv;
    timeval_t time;

    (void) gettimeofday(&tv, (void *) 0);
    time.tv_sec = tv.tv_sec;
    time.tv_usec = tv.tv_usec;
    return time;
}

unsigned int
hostClockInit()
{
    asyncNotifier(SYS_ASYNC_MON_ALARM);
    return SYS_ASYNC_MON_ALARM;
}

void
scheduleAlarm(timeval_t when)
{
    struct itimerval itv;

    if (when.tv_sec == 0 && when.tv_usec < CLOCK_RESOLUTION)
	when.tv_usec = CLOCK_RESOLUTION;

    itv.it_value.tv_sec = when.tv_sec;
    itv.it_value.tv_usec = when.tv_usec;
    itv.it_interval.tv_sec = 0;
    itv.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &itv, (struct itimerval *) 0);
    return;
}


void
cancelAlarm()
{
    struct itimerval itv;

    itv.it_value.tv_sec = 0;
    itv.it_value.tv_usec = 0;
    itv.it_interval.tv_sec = 0;
    itv.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &itv, (struct itimerval *) 0);
    return;
}
