/*
 * @(#)rtld.c	1.1 98/09/15
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * This file is a preliminary attempt at implementing the private interface
 * between libthread and ld.so.1 that allows ld.so.1 to use the
 * synchronization primitives that libthread provides.  This appears to be
 * the only way to make dynamic linking safe in the presence of green threads
 * but for now this isn't the default solution.  It will be for 1.2.1.
 */
#ifdef 0

#include "hpi_impl.h"
#include "threads_md.h"
#include "mutex_md.h"
#include "iomgr.h"


/*
 * Threads Interface communication structure for threads library
 */
typedef struct {
        int     ti_tag;
        union {
                int     ti_val;
                int (*  ti_func)();
        } ti_un;
} Thr_interface;

#define TI_NULL         0       /* (void) last entry */
#define TI_MUTEX_LOCK   1       /* _mutex_lock() address */
#define TI_MUTEX_UNLOCK 2       /* _mutex_unlock() address */
#define TI_LRW_RDLOCK   3       /* _llrw_rdlock() address */
#define TI_LRW_WRLOCK   4       /* _llrw_wrlock() address */
#define TI_LRW_UNLOCK   5       /* _llrw_unlock() address */
#define TI_BIND_GUARD   6       /* _bind_guard() address */
#define TI_BIND_CLEAR   7       /* _bind_clear() address */
#define TI_LATFORK      8       /* _lpthread_atfork() */
#define TI_THRSELF      9       /* _thr_self() address */
#define TI_VERSION      10      /* current version of ti_interface */
#define TI_V_CURRENT    1               /* current version of threads */

/*
 * When handling mutex's locally we need to mask signals.  The signal mask
 * is for everything except SIGWAITING.
 */
static const sigset_t	iset = { ~0UL, ~0UL, ~0UL, ~0UL };

static int
green_mutex_lock(mutex_t * mp, sigset_t * oset)
{
	if (oset)
		(void) sigprocmask(SIG_BLOCK, &iset, oset);
	mutex_lock(mp);
	return 0;
}

static int
green_mutex_unlock(mutex_t * mp, sigset_t * oset)
{
	mutex_lock(mp);
	if (oset)
		(void) sigprocmask(SIG_BLOCK, &iset, oset);
	return 0;
}

static int
green_llrw_rdlock(rwlock_t * rwlp)
{
    mutexLock((mutex_t *)rwlp);
    return 0;
}

static int
green_llrw_wrlock(rwlock_t * rwlp)
{
    mutexLock((mutex_t *)rwlp);
    return 0;
}

static int
green_llrw_unlock(rwlock_t * rwlp)
{
    mutexUnlock((mutex_t *)rwlp);
    return 0;
}

static int
green_bind_guard(int bindflags)
{
	sys_thread_t 	*thr;
	extern int _nthreads;

	if (!threadBootstrappedP)
		return (0);
	if ((thr = greenThreadSelf()) == NULL)
		return (1);
	else if ((thr->bindflags & bindflags) == bindflags)
		return (0);
	else {
		thr->bindflags |= bindflags;
		return (1);
	}
}

static int
green_bind_clear(int bindflags)
{
	sys_thread_t *	thr;
	if (threadBootstrappedP && (thr = greenThreadSelf()) != NULL)
		return (thr->t_rtldbind &= ~bindflags);
	else
		return (0);
}

#ifdef notdef
int
green_lpthread_atfork(void (*prepare) (void), void (*parent) (void),
	void (*child) (void))
{
}
#endif

int
green_thr_self()
{
    return (int) greenThreadSelf();
}

static Thr_interface _funcs[] = {
/* 01 */        { TI_MUTEX_LOCK,        (int) green_mutex_lock },
/* 02 */        { TI_MUTEX_UNLOCK,      (int) green_mutex_unlock },
/* 03 */        { TI_LRW_RDLOCK,        (int) green_llrw_rdlock },
/* 04 */        { TI_LRW_WRLOCK,        (int) green_llrw_wrlock },
/* 05 */        { TI_LRW_UNLOCK,        (int) green_llrw_unlock },
/* 06 */        { TI_BIND_GUARD,        (int) green_bind_guard },
/* 07 */        { TI_BIND_CLEAR,        (int) green_bind_clear },
#ifdef notdef
/* 08 */        { TI_LATFORK,           (int) green_lpthread_atfork },
#endif
/* 09 */        { TI_THRSELF,           (int) green_thr_self },
/* 10 */        { TI_VERSION,           TI_V_CURRENT},
/* 11 */        { TI_NULL,              (int) NULL }
};

extern void _ld_concurrency(Thr_interface *);

#pragma	weak	_ld_concurrency
#pragma init(_set_rtld_interface)

static void
green_set_rtld_interface()
{
    void (*fptr)();
    
    if ((fptr = _ld_concurrency) != 0)
	(* fptr)(green_funcs);
}
#endif
