/*
 * @(#)schedule.c	1.28 98/08/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Thread scheduling utilities and data structures.
 */

#include "hpi_impl.h"

#include "threads_md.h"
#include "queue.h"
#include "schedule.h"
#include <memory.h>

int _scheduling_lock;
volatile int _needReschedule;

gt_queue_t runnable_queue;

void
reschedule()
{
    sys_thread_t *self = greenThreadSelf();
    sys_thread_t *tid;

    sysAssert(SCHED_LOCKED());
    sysAssert(_scheduling_lock == 1);

    tid = queueRemoveHead(&runnable_queue, tid);

    sysAssert(tid->state == RUNNABLE);
    sysAssert(!tid->sigmaskdirty);

    Log2(5, "reschedule: from TID 0x%x to TID 0x%x\n",
    	 self, tid);

    updateSliceStart();

    setCurrentThread(tid);

    if (self->sigmaskdirty) {
	green_sigprocmask(SIG_SETMASK, NULL, &MACHDEP(self)->signalMask);
	self->sigmaskdirty = 0;
    }

    if (memcmp(&MACHDEP(tid)->signalMask, &MACHDEP(self)->signalMask,
	sizeof MACHDEP(tid)->signalMask) != 0)
    {
	sigset_t old;
	green_sigprocmask(SIG_SETMASK, &MACHDEP(tid)->signalMask, &old);
	sysAssert(memcmp(&MACHDEP(self)->signalMask, &old, sizeof old) == 0);
    }

    switchContext(CONTEXT(tid));

    /* NOTREACHED */
}
