/*
 * @(#)synch.c	1.24 98/08/14
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <sys/types.h>
#include <wait.h>

#include "hpi_impl.h"

#include "synch.h"
#include "schedule.h"
#include "iomgr.h"

/*
 * SPECIAL CASE: In Solaris, malloc() calls sbrk() for more memory,
 * and both call _mutex_lock(), which we override to make them Green
 * thread-safe.  But either _mutex_lock() call can thus cause monitor
 * cache expansion, which can lead to reentrant malloc calls (we try
 * to expand when entering sbrk() called from malloc()).  malloc()
 * isn't reentrant, so this causes trouble.  Instead, we special-case
 * sbrk() by wrapping it up in a non-cache monitor. 
 */ 
sys_mon_t _sbrk_lock;
#define SBRK_LOCK_INIT()	sysMonitorInit(&_sbrk_lock)
#define SBRK_LOCK(self)		sysMonitorEnter(self, &_sbrk_lock)
#define SBRK_UNLOCK(self)	sysMonitorExit(self, &_sbrk_lock)

static bool_t sbrk_initialized = FALSE;

void InitializeSbrk(void)
{
    if (!sbrk_initialized) {
	/* Should put this elsewhere... */
	SBRK_LOCK_INIT();
	sbrk_initialized = TRUE;
    }
}

#ifndef GPROF
extern void *_sbrk_unlocked(int);

void *_sbrk(int increment)
{
    void *ret;

    if (sbrk_initialized) {
	SBRK_LOCK(sysThreadSelf());
        ret = _sbrk_unlocked(increment);
    	SBRK_UNLOCK(sysThreadSelf());
    } else {
	ret = _sbrk_unlocked(increment);
    }

    return ret;
}
#endif

/* For java.lang.Process */

extern int _waitid(idtype_t idtype, id_t id, siginfo_t *infop, int options);
int waitid(idtype_t idtype, id_t id, siginfo_t *infop, int options)
{
    sys_mon_t *mon = asyncMon(SYS_ASYNC_MON_CHILD);
    sysMonitorEnter(sysThreadSelf(), mon);
    while (_waitid(idtype, id, infop, options | WNOHANG) != 0 ||
	   infop->si_pid == 0) {
        sysMonitorWait(sysThreadSelf(), mon, SYS_TIMEOUT_INFINITY);
    }
    sysMonitorExit(sysThreadSelf(), mon);
    return 0;
}

extern pid_t _fork1(void);
pid_t fork1()
{
    pid_t res = _fork1();
    if (res == 0) {
        PrimordialThreadBootstrapped(FALSE);
    }
    return res;
}
