/*
 * @(#)interrupt.h	1.18 98/09/15
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * Interrupt dispatch interface
 */

#ifndef _JAVASOFT_INTERRUPT_H_
#define _JAVASOFT_INTERRUPT_H_

/*
 * Type definitions.
 */

/*-
 * A function that handles interrupt dispatch requests is of type
 * intr_handler_t.  This type definition is mostly for convenience.
 * A declaration of a handler function, should look like this:
 *
 *   void myHandler(int interrupt, void *siginfo, void *context, void *arg);
 *
 * An intr_handler_t is constrained:
 *
 *	- It runs on the exception stack.
 *	- It cannot yield.
 *	- It cannot allocate/free memory.
 *	- It can only call interrupt-safe routines.
 *
 * "arg" is set to the "handlerArg" specified in intrRegister().
 */
typedef void (*intr_handler_t)(int interrupt, void *siginfo,
			      void *context, void *arg);

/*
 * Routines.
 */

/* Initialize the interrupt system */
void intrInit(void);

/* Set a handler for a particular interrupt */
signal_handler_t intrRegister(int interrupt, intr_handler_t handler,
			      void *handlerArg);

/* Dispatch an interrupt (called from the low-level handlers) */
void intrDispatch(int interrupt, void *siginfo, void *context);

/*-
 * The target specific header file should define the following
 *
 * Constants
 *
 *	N_INTERRUPTS  -	The number of interrupt channels.  These
 *			are numbered from 0 to (N_INTERRUPTS - 1).
 */

#define	N_INTERRUPTS	32	/* 0 to 31 */

/*-
 * Routines/Macros that control whether interrupts are enabled or
 * not.
 *
 *	void intrLock(void)	      -	Disable all interrupts.
 *	void intrUnlock(void)	      -	Enable all interrupts.
 *
 *		Note: intrLock()/intrUnlock() pairs can be nested.
 *
 */

void intrLock(void);
void intrUnlock(void);

/*-
 * intrInitMD() --
 *	Initialize the machine-dependant interrupt software.
 *
 *	This routine should leave the all interrupts disabled as if
 *	one (1) intrLock() had been called.  At the end of the
 *	bootstrap, a single intrUnlock(), will be called to turn
 *	interrupts on.
 */

void intrInitMD(void);

/* Thread implementation dependent interrupt dispatcher. */
void intrDispatchMD(int sig, siginfo_t *info, void *uc);

/* Whether the signal is used by the HPI implementation */
bool_t intrInUse(int sig);

#endif /* !_JAVASOFT_INTERRUPT_H_ */
