ifdef(`EXTRACT_OFFSETS', `divert(-1)' )
/*
 * @(#)executeJava_sparc.s	1.77 98/11/03
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */


/*
 * ExecuteJava: The interpreter core.
 * To replace executeJava.c on SPARC systems.
 * This file uses m4 preprocessing.
 */
	.global	ExecuteJava
	.type	ExecuteJava,2
/*
 * Important variables:
 * 
 * Java PC -- the VM program counter. As we begin interpretation of
 *		a bytecode, this is its address. Is modified during course
 *		of bytecode execution, but only after all error checking
 *		is complete. Initially passed us as our 1st parameter.
 * Ee	   -- the current ExecEnv *. Not often consulted, so needs not
 *		be in a register. Initially passed us as our 2nd parameter.
 * Java SP -- Pointer to our current expression evaluation stack. Points
 *		to the empty memory beyond the top-most memory word actually
 *		occupied.
 *		Initialized from the stack frame at entry and at each call.
 *		Saved in the Java stack frame across calls.
 * Java FP -- Pointer to the current Java stack frame. Initialized at entry
 *		and at calls and returns from calls.
 * Varbase -- Pointer to the array of local parameters in the current
 *		stack frame. Initialized from the stack frame at entry and 
 *		at each call.
 *		Saved in the Java stack frame across calls.
 * ConstP -- Pointer to the current method's constant table. Initialized from
 *		the stack frame at entry and at each call.
 *		Saved in the Java stack frame across calls.
 * Unhand register -- The only regiser ever used to directly address object
 *		data. If the compacting GC actually compacts anything, this
 *		register should always be relocated. I promise it will never
 *		contain "just a number" -- always a pointer or garbage!
 * NextByte -- Fetchahead in the instruction stream. As we begin interpretation
 *		of a bytecode, this is the following byte. It is sometimes
 *		a parameter to this bytecode, and is sometimes the next
 *		bytecode, and may sometimes be garbage! Should always be
 *		fetched ahead before an instruction's execution.
 * TOSA      -- A computational temporary. Formerly the cached top
 *		(or next-to-top) of the expression stack, thus the name.
 * TOSB      -- A computational temporary. Formerly the cached top
 *		of the expression stack, thus the name.
 * R3 temp   -- A computational temporary.  Survives method calls
 * R4 temp   -- A computational temporary.  Survives method calls
 * R5 temp   -- A computational temporary.  Survives method calls
 * Dispatch  -- Address of the bytecode vector.
 *		Initialized at entry. See Opcode.
 * Opcode    -- The current bytecode, scaled appropriately for jump directly
 *		into a bytecode vector. In the case of an instruction without
 *		I-stream operands, this is formed from NextByte by shifting.
 *		For instance, a no-op instruction should contain these
 *		instructions (though not necessarily in this order!):
 *		    sll    NextByte,N,Opcode	! scale next byte for dispatch
 *		    ldub   [PC+1],NextByte	! fetch ahead
 *		    jmp    Dispatch+Opcode 	! dispatch to next instruction
 *		    inc    PC			! increment pc
 *		Each bytecode sequence must dispatch through the
 *		Dispatch table.
 *
 * T1 temp    -- A computational temporary.  Bashed by method calls
 * T2 temp    -- A computational temporary.  Bashed by method calls
 * T3 temp    -- A computational temporary.  Bashed by method calls
 * T4 temp    -- A computational temporary.  Bashed by method calls	
 *
 * Initial_frame
 *	     -- the value of ee->current_frame upon entry.
 *	        popping this at return implies exit. Only used to
 *		compare frame pointer at return instructions. Keep in
 *		C stack.
 * StackT    -- an array of two 64-bit, double-aligned cells on the C stack,
 *		used when shuttling doubles between register sets.
 *		Usually we can use the Java stack for this, but there are
 *		some cases where we need more memory than is promised as 
 *		available there.
 * ErrorBuffer
 *	    -- char[255]. Used for formatting exceptional messages.
 */
	
/*
 * Register Usage:
 * %i0 and %i1 are our input parameters, and we have no better use for
 * them. The %o registers are output parameters, of course, so can only
 * be used for temps. %f registers are used only for expression evaluation
 * temporaries.
 *
 *   %i0 - Java PC	  %l0 - Unhand register %o0            %g0 always zero
 *   %i1 - Ee		  %l1 - TOSA            %o1            %g1 T4 temp
 *   %i2 - Java SP        %l2 - TOSB            %o2            %g2 T5 temp
 *   %i3 - Varbase        %l3 - R3 temp         %o3 - Opcode   %g3 T6 temp
 *   %i4 - ConstP         %l4 - R4 temp	        %o4 - T1 temp
 *   %i5 - Java FP        %l5 - R5 temp	        %o5 - T2 temp
 *   %i6 - %fp		  %l6 - Next Byte       %o6 - %sp
 *   %i7 - return address %l7 - Dispatch        %o7 - T3 temp 
 */
	
define(PC,      `%i0')
define(Ee,      `%i1')
define(SP,      `%i2')
define(Varbase, `%i3')
define(ConstP,  `%i4')
define(FP,      `%i5')	
	
define(Unhand,  `%l0')
define(TOSA,    `%l1')
define(TOSB,    `%l2')
define(R3,      `%l3')
define(R4,      `%l4')
define(R5,      `%l5')
define(NextByte,`%l6')
define(Dispatch, `%l7')
	
define(Opcode, `%o3')
define(T1,     `%o4')
define(T2,     `%o5')
define(T3,     `%o7')
	
define(T4,     `%g1')
define(T5,     `%g2')
define(T6,     `%g3')
	
/*
 * C Stack Usage:
 * 	fp->	|
 * 	fp-16->	|   double StackT[2]
 * 	fp-20->	|   void * Initial_frame
 *	fp-24->	|   char ErrorBuffer[255]
 */
define(`StackT',	16)
define(`Initial_frame', 20)
define(`ErrorBuffer', (24+255))
define(`ErrorBufferSize', 255 )
define(`FRAMESIZE', (96+32+256))

/* Our general-purpose name mangling macros */
define(catName, `$1`.'$2')
define(cat, `$1$2')

/*
 * Stack Management:
 * We do not cache any stack top. In this version, everything
 * is always kept in memory.
 */ 

/*
 * Accessing C struct Elements:
 * Much of the data we need is embedded in data structures defined in
 * the C language. I do not wish to embed in this program numbers
 * representing the offsets of fields within structures. Thus we use m4
 * and awk processing to build symbolic constants to represent these offsets.
 * In this program, C data structure components will always be accessed using
 * the macro STRUCT_OFFSET(struct_tag, fieldname). These references can
 * be programmatically extracted and used to produce an include file,
 * defining appropriately mangled names as little numbers. This file
 * will be included at assembly time, and the STRUCT_OFFSET macro will
 * evaluate to a mangled name ( which then evaluates to a little number ).
 *
 * Similarly, we will use the macro DEFINED_CONSTANT(c) to access C-language
 * #defines. We could just use the cpp and include a lot of the header files
 * here, but they would have to be much more carefully structured than
 * they presently are.
 */
ifdef(`EXTRACT_OFFSETS',
    `divert(-1)
     define( STRUCT_OFFSET, `divert(0)
FIELD			$1	$2	catName($1,$2)
divert(-1)' )
    define( DEFINED_CONSTANT, `divert(0)
CONST			$1
divert(-1)' )',
    `include(`ExecuteJavaStructOffsets.include')
     define( STRUCT_OFFSET, `catName($1,$2)')
     define( DEFINED_CONSTANT, `$1')'
)
	
/*
 * Shorthand for the most frequently typed instructions
 * These are used as part of every bytecodes interpretation.
 */

define(GO, `jmp	Dispatch+Opcode')
define(OPSCALE, 6)
define(SCALE, `sll	$1,OPSCALE,Opcode')
define(SCALENEXT, 	`SCALE(NextByte)')

/* Call a function, but bash the return address to go to the indicated spot 
 * instead of returning normally. 
 * Parameters:	
 *     name of method
 *     number of arguments
 *     label to branch to, instead
 */    
define(CALL_AND_BRANCH,
    `call   $1,$2
     add    %o7, ($3 - (. + 4)), %o7
    ')

/* Call a function, but bash the return address to do an opcode dispatch,
 * instead.  Opcode should contain the next opcode << 6
 * Parameters
 *     name of method
 *     number of arguments
 */

define(CALL_AND_GO, 
    `add    Dispatch,Opcode,Opcode
     call   $1,$2
     add    Opcode,-8,%o7')

/*
 * Entry: load up registers and start interpreting! 
 * On input, have these registers:
 *	PC, Ee
 * And must load these registers:
 *	via Ee: SP, Varbase, ConstP
 *	via PC: Opcode, NextByte
 *	constants: Unhand, Dispatch
 */

	.section ".text",#alloc,#execinstr

ExecuteJava:
	save	%sp,-FRAMESIZE,%sp
	mov	0,Unhand	! Quickly get rid of random trash in reg
	set	TableBase,Dispatch
	ld	[Ee+STRUCT_OFFSET(execenv,current_frame)],FP ! get Java FP
	!	Opcode = PC[0]
	ldub	[PC], Opcode		! fetch first bytecode
	!	Varbase = Ee->current_frame->vars
	ld	[FP+STRUCT_OFFSET(javaframe,vars)], Varbase
	! 	SP = Ee->current_frame->optop
	ld	[FP+STRUCT_OFFSET(javaframe,optop)], SP
	SCALE(Opcode)
	!	ConstP = Ee->current_frame->constant_pool
	ld	[FP+STRUCT_OFFSET(javaframe,constant_pool)],ConstP
	ldub	[PC+1],NextByte		! NextByte = PC[1]

	jmp	Dispatch+Opcode
	st	FP,[%fp-Initial_frame]

	/*
	 * A couple of macros to help me manipulate
	 * the layout of the opcode vector
	 * We allow 16 words (64 bytes) per opcode.
	 *
	 * Automatically padding the code to this size is really tough.
	 * The current, brain-dead SPARC assembler does not let me do dot
	 * arithmetic. It does provide the ".skip" pseudo-op, but allows
	 * only a very simple numeric expression as skip count -- not even
	 * a symbol -- so there's no way I can say something as complex as
	 * .skip 40-(.-lastlabel) ! Here is the process I use:
	 * 1) Macro-process the file using -DEXTRACT_TABLE_PADDING, and 
    	 *    assemble the result.  Each of the padding variables, each of
    	 *    which has a constant value in the range 0 <= i <= 64, to be
         *    assigned an absolute (i.e. nonrelocatable) value. 
	 * 2) Run the resulting .o file through a pipe which
         *     a) Extracts all the names from the symbol table
         *     b) Extracts from those the ones whose names are pad[0-9]*
         *     c) Sorts them (not really necessary)
         *     d) Turns each remaining line from "nm" into an m4 definition
         *         in which the padding variable is defined to be its absolute
         *         value.
	 * 3) Re-macro process and re-assemble this file, without the
	 *    EXTRACT_TABLE_PADDING definition. This will cause .skip pseudo-
	 *    ops to be placed between the table entries, each referencing
	 *    one of the names defined in the padding.include file, which will
	 *    be included. Assuming the program did not change since step (1),
	 *    this should produce a correctly padded and aligned table.
	 * Needless to say, use a make file.
	 */

/* This macro makes it easy to find things in DBX */
define(MARK,
      `.global $1
       .type $1 `,'`2'	
       $1`:'')

define(STARTVECTOR,
    `divert(INLINE_DIVERSION)
     define(`CUROP',$1) 
     define(`ZLABEL', catName(CUROP,`0'))  
     STARTACTION 
     define(`OPNUM', 0 ) 
     MARK($1_0)
     ZLABEL`:'')

define(OP0, 
   `divert(INLINE_DIVERSION)
    define(`CUROP',$1) 
    define(`OPNUM', incr(OPNUM)) 
    PAD0ACTION 
    define(`ZLABEL',catName($1,`0'))
    MARK($1_0)
    ZLABEL`:'')

define(ENDVECTOR, `ILLEGALS')

ifdef( `EXTRACT_TABLE_PADDING',
    `define(`STARTACTION', `')
     define(`PAD0ACTION', `
	 cat(`pad',OPNUM) = 64-(.-ZLABEL)')
    ',
    `ifdef(`EXTRACT_OFFSETS',`',`include(`ExecuteJavaPadding.include')')
     define(`STARTACTION', `' )
     define(`PAD0ACTION', `.skip	cat(`pad',OPNUM )
     ' )
    ')
    
ifdef(`EXTRACT_OFFSETS', `
	define( `OUTLINE', `' )
	define( `INLINE', `' )
	define( `LITERAL_DIVERSION',  `-1')
	define( `INLINE_DIVERSION',  `-1')
	define( `OUTLINE_DIVERSION', `-1')
    ',`
	define( `LITERAL_DIVERSION',  `1')
	define( `INLINE_DIVERSION',  `2')
	define( `OUTLINE_DIVERSION', `3')

	define( `OUTLINE', `divert(OUTLINE_DIVERSION) .align 16')
	define( `INLINE', `divert(INLINE_DIVERSION)')
    ')

/*
 * Some opcodes are unused and thus illegal.
 */
define(`ILLEGALOP',`
    OP0($1)
	b	internalError
	mov	OPNUM,TOSA	! TOSA gets the illegal opcode' )

define(`ILLEGALS',
       `ILLEGALOP(cat(`Illegal', eval(OPNUM+1)))
	ifelse(eval(OPNUM < 255), 1, `ILLEGALS')
       ')


	divert(LITERAL_DIVERSION)
    /*
     * LITERAL POOL.
     * By situating the following data within a few kilobytes of 
     * TableBase, we can use Dispatch as an addressing base register
     * for it. Makes address formation cheaper, and code in less  need
     * of relocation. (Of course, we cannot be totally PIC  unless we
     * get rid of direct calls and global variable references.)  
     * If we had a worthwhile assembler (as if), then the USING
     * directive would help us form these addresses without as much
     * blood on the screen.
     */
	
    /*
     *	
     * First, a table of names of exception
     * These are referenced in the exception signaling code.
     */ 
define(`JAVAPKG', `"cat(`java/lang/',$1)"' )
arithmeticExceptionName:	.asciz	JAVAPKG(ArithmeticException)
nullPointerName:		.asciz	JAVAPKG(NullPointerException)
classCastName:			.asciz	JAVAPKG(ClassCastException)
instantiationErrorName:		.asciz	JAVAPKG(InstantiationError)
illegalAccessName:		.asciz	JAVAPKG(IllegalAccessError)
negativeArrayName:		.asciz	JAVAPKG(NegativeArraySizeException)
outOfMemoryName:		.asciz	JAVAPKG(OutOfMemoryError)
indexBoundsName:		.asciz	JAVAPKG(ArrayIndexOutOfBoundsException)
arrayStoreName:			.asciz	JAVAPKG(ArrayStoreException)
internalErrorName:		.asciz	JAVAPKG(InternalError)
	/*
	 * And other messages, too.
	 * These are also referenced in the exception signaling code.
	 */
divZero:			.asciz	"/ by zero"
undefOpcode:			.asciz	"undefined opcode (0x%x)"
undefWideOpcode:		.asciz	"undefined wide sub-opcode (0x%x)"
percentD:			.asciz	"%d"
	
	/*
	 * Dispatch table for wide opcodes.
         * The file opcodes.wide is automatically generated and has 256 lines, 
         * where each line is either
	 *    NoWideOpcode(<opname>)  or  WideOpcode(<opname>)
	 * depending on whether the corresponding instruction has a wide 
	 * form or not.
	 */
	.align	4
widetab:    	
	define(`NoWideOpcode', `.half wideError - TableBase	! $1')
	define(`WideOpcode',   `.half cat(`wide_',$1) - TableBase')
	include(opcodes.wide)
	undefine(`NoWideOpcode')
	undefine(`WideOpcode')	
	
	/*
	 * And macros for accessing these things.
	 * Macro parameters are label and target register.
	 * Usage example:
	 *	SET_ADDRESS( percentD, %o1 )
	 */
define(`SET_ADDRESS',`add	Dispatch,($1-TableBase),$2')
	/*
	 * Macro parameter is label.
	 * Usage example:
	 *	lduh	BASED_ADDRESS(widetab+2),%o0
	 */
define(`BASED_ADDRESS',`[Dispatch+($1-TableBase)]')

	/*
	 * And here is the opcode vector.
	 */
	.align 128
	
STARTVECTOR(nop)
TableBase:	
	SCALENEXT
	ldub	[PC+2],NextByte
	GO
	inc	PC

/*
 * There are lots of instructions used for pushing one-word constants
 * onto the expression stack. They are all the same. They look like
 * this. Constant word is part of the opcode, so no I-stream parameter
 * is necessary.
 * Parameters are: opcode name and the value to stick in TOS.
 */
define(`PUSH_CONST',
    `OP0($1)
	SCALENEXT
	ldub	[PC+2],NextByte
	ifelse(`$2', 0,
	`st	%g0,[SP]',
	`set	$2,TOSA
	st	TOSA,[SP]')
	inc	PC
	GO
	inc	4,SP')
	
/*
 * And similarly, for pushing two-word constants.
 * Parameters are: opcode name, high-order word, low-order word
 */
define(`PUSH_TWO_WORD_CONST',
    `OP0($1)
	SCALENEXT
	ldub	[PC+2],NextByte
	ifelse(`$2', 0,
	`st	%g0,[SP]',
	`set	$2,TOSA
	st	TOSA,[SP]')
	ifelse(`$3', 0,
	`st	%g0,[SP+4]',
	`set	$3,TOSB
	st	TOSB,[SP+4]')
	inc	PC
	GO
	inc	8,SP')

PUSH_CONST(aconst_null, 0)
PUSH_CONST(iconst_m1, -1)
PUSH_CONST(iconst_0, 0)
PUSH_CONST(iconst_1, 1)
PUSH_CONST(iconst_2, 2)
PUSH_CONST(iconst_3, 3)
PUSH_CONST(iconst_4, 4)
PUSH_CONST(iconst_5, 5)
PUSH_TWO_WORD_CONST(lconst_0, 0, 0)
PUSH_TWO_WORD_CONST(lconst_1, 0, 1)
PUSH_CONST(fconst_0, 0)
PUSH_CONST(fconst_1, (127<<23))
PUSH_CONST(fconst_2, (128<<23))
PUSH_TWO_WORD_CONST(dconst_0, 0, 0)
PUSH_TWO_WORD_CONST(dconst_1, (1023<<20), 0)

/* BIPUSH -- Push sign-extended argument byte onto the stack
 */
	! note use of nonstandard idiom for sign-extending a byte:
	!   v = (byte^128)-128.
	! This groups better on most processors than shifting does!
OP0(bipush)
	xor	NextByte,128,TOSA
	ldub	[PC+2],Opcode
	inc	2,PC
	ldub	[PC+1],NextByte
	SCALE( Opcode )
	sub	TOSA,128,TOSA
	st	TOSA,[SP]
	GO
	inc	4,SP

/*
 * SIPUSH --Take two argument bytes, concatenate into a 16-bit number.
 * Sign extend it and push it on the stack.
 */
OP0(sipush)
	ldub	[PC+3],Opcode
	sll	NextByte,24,T2
	ldub	[PC+2],T1
	sra	T2,16,T2
	inc	3,PC
	ldub	[PC+1],NextByte
	SCALE( Opcode )
	or	T2,T1,TOSA
	st	TOSA,[SP]
	GO
	inc	4,SP

/*
 * LDC: load from constant pool.
 * Resolve the constant value and rewrite as a _quick instruction.
 * This instruction does not affect the stack, directly. However,
 * it will turn into a ldc_quick instruction, which will push a value onto
 * the stack.
 *
 * We share the resolve-and-rewrite
 * code, at resolve_ldc, with some other opcodes. Its parameters are:
 *	R3	-- desired quick opcode.
 *	%o1     -- const pool index.
 *	%o3	-- type bitset.
 *	NextByte -- pc[1]
 * In the case of ldc, %o1 and NextByte are the same.
 * This is not so for other instructions, such as those that are
 * defined using resolve_symbolic_w!
 */
INTorFLOATorSTRING = ( 1<< DEFINED_CONSTANT(CONSTANT_Integer) ) | ( 1 << DEFINED_CONSTANT(CONSTANT_Float) ) | (1 << DEFINED_CONSTANT(CONSTANT_String) )

OP0(ldc)
	mov	DEFINED_CONSTANT(opc_ldc_quick),R3	! how to rewrite pc[0]
	mov	NextByte,%o1		! constant pool index 
	ba	resolve_ldc
	mov	INTorFLOATorSTRING,%o3

/*
 * RESOLVE_SYMBOLIC_W
 * Resolve a symbolic reference instruction, replacing own opcode. 
 * Macro parameters are 
 *      name, 
 *      CONSTANT_type mask for ResolveClassConstant,
 *      replacement opcode. 
 * Following two bytes are concatenated into an unsigned
 * index into the constant pool.  This instruction does not affect the stack,
 * directly. However, it will turn into a _quick instruction, which may.
 */
define(`RESOLVE_SYMBOLIC_W',
`OP0($1)
	ldub	[PC+2],T2	! low-order byte of 16-bit index
	sll	NextByte,8,%o1	! high-order byte of 16-bit index
	mov	DEFINED_CONSTANT($3),R3 ! how to rewrite pc[0]
	mov	$2,%o3		! constant type mask
	ba	resolve_ldc
	or	%o1,T2,%o1	! form index
')

/* LDC_W: load from constant pool (wide index).
 * Just like ldc, but with a 2-byte index.
 * This instruction does not affect the stack, directly. However,
 * it will turn into a ldc_w_quick instruction, which will.
 */
RESOLVE_SYMBOLIC_W(ldc_w, INTorFLOATorSTRING, opc_ldc_w_quick )

/* LDC2_W: load 2 words from constant pool (wide index).
 * Just like ldc_w.
 * Treat accordingly similarly.
 * This instruction does not affect the stack, directly. However,
 * it will turn into a ldc2_w_quick instruction, which will.
 */
DOUBLEorLONG = ( 1<< DEFINED_CONSTANT(CONSTANT_Double) ) | ( 1 << DEFINED_CONSTANT(CONSTANT_Long) )
RESOLVE_SYMBOLIC_W(ldc2_w, DOUBLEorLONG, opc_ldc2_w_quick)

	OUTLINE

/* RESOLVE_LDC: code shared by ldc and clients of resolve_symbolic_w.
 * call ResolveClassConstant. If there is an exception, handle
 * it, else rewrite the current opcode as indicated and re-execute
 * it. Parameters are:
 *	R3	-- desired quick opcode.
 *	%o1     -- const pool index.
 *	%o3	-- type bitset.
 *	NextByte -- pc[1]
 */
resolve_ldc:
	! ResolveClassConstant(constant_pool, index, ee, bitset)!
	mov	ConstP,%o0
	call	ResolveClassConstant,4
	mov	Ee,%o2

	! We're back from ResolveClassConstant. See if it caused an exception.
	! If not, then rewrite own opcode as indicated by register R3, whilst
	! branching to the appropriate piece of code to execute it. Assume
	! that the byte fetched ahead, in NextByte, is still valid.
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T2
	SCALE( R3 )		! (anticipate success)
	tst	T2		
	bgt	handleException	! branch if ResolveClassConstant causes
	nop			! an exception!
	jmp	Dispatch+Opcode	! re-dispatch on new opcode
	stb	R3,[PC]		! while re-writing it in I-stream

	INLINE

/*
 * LOAD_LOCALVAR: There are several opcodes for pushing a local variable, as
 * specified in an operand byte. They are all the same. They
 * all look like this. Parameter is opcode name.
 */
define(`LOAD_LOCALVAR',
    `OP0($1)
	ldub	[PC+2],Opcode		! get the next opcode
	sll	NextByte,2,T2		! offset for Varbase[NextByte]
	ldub	[PC+3],NextByte		! get new NextByte
	ld	[Varbase+T2],TOSA	! load variable value
	SCALE( Opcode )
	inc	2,PC
	st	TOSA,[SP]		! store variable value
	GO
	inc	4,SP')

/*
 * Similarly, we can push a two-word local variable onto the stack.
 * NextByte specifies index of high-order word. Following variable
 * is low-order word.
 */

define(`LOAD_TWO_WORD_LOCALVAR',
    `OP0($1)
	ldub	[PC+2],Opcode		! get the next opcode
	sll	NextByte,2,T2		! offset for Varbase[NextByte]
	ldub	[PC+3],NextByte
	add	Varbase,4,T1		! offset for &Varbase[1]
	ld	[Varbase+T2],TOSA	! get Varbase[NextByte]
	SCALE( Opcode )
	ld	[T1+T2],TOSB		! get Varbase[NextByte + 1]
	inc	2,PC
	st	TOSA,[SP]
	inc	8,SP
	GO
	st	TOSB,[SP-4] ')


LOAD_LOCALVAR(iload)
LOAD_TWO_WORD_LOCALVAR(lload)
LOAD_LOCALVAR(fload)
LOAD_TWO_WORD_LOCALVAR(dload)
LOAD_LOCALVAR(aload)
/*
 * LOAD_LOCALVAR_N
 * There are several opcodes for pushing a local variable, as
 * specified in the opcode itself. They are all the same. They
 * all look like this. Parameters are opcode name and variable
 * index.
 */
define(LOAD_LOCALVAR_N,
    `OP0($1)
	ld	[Varbase+eval(4* $2)],TOSA ! get register $2
	SCALENEXT
	ldub	[PC+2],NextByte
	inc	PC
	st	TOSA,[SP]		! store the value
	GO
	inc	4,SP ' )
/*
 * LOAD_TWO_WORD_LOCALVAR_N
 * Similarly, for pushing two words of local variable, as specified
 * in the opcode. Parameters are opcode name and variable index of
 * high-order word. Low-order word is in following variable.
 */
define(LOAD_TWO_WORD_LOCALVAR_N,
    `OP0($1)
	ld	[Varbase+eval(4* $2)],TOSA ! get register $2
	SCALENEXT
	ldub	[PC+2],NextByte
	ld	[Varbase+eval(4+4* $2)],TOSB ! get register eval($2+1)
	inc	PC
	st	TOSA,[SP]
	inc	8,SP
	GO
	st	TOSB,[SP-4] ')

LOAD_LOCALVAR_N(iload_0, 0)
LOAD_LOCALVAR_N(iload_1, 1)
LOAD_LOCALVAR_N(iload_2, 2)
LOAD_LOCALVAR_N(iload_3, 3)
LOAD_TWO_WORD_LOCALVAR_N(lload_0, 0)
LOAD_TWO_WORD_LOCALVAR_N(lload_1, 1)
LOAD_TWO_WORD_LOCALVAR_N(lload_2, 2)
LOAD_TWO_WORD_LOCALVAR_N(lload_3, 3)
LOAD_LOCALVAR_N(fload_0, 0)
LOAD_LOCALVAR_N(fload_1, 1)
LOAD_LOCALVAR_N(fload_2, 2)
LOAD_LOCALVAR_N(fload_3, 3)
LOAD_TWO_WORD_LOCALVAR_N(dload_0, 0)
LOAD_TWO_WORD_LOCALVAR_N(dload_1, 1)
LOAD_TWO_WORD_LOCALVAR_N(dload_2, 2)
LOAD_TWO_WORD_LOCALVAR_N(dload_3, 3)
LOAD_LOCALVAR_N(aload_0, 0)
LOAD_LOCALVAR_N(aload_1, 1)
LOAD_LOCALVAR_N(aload_2, 2)
LOAD_LOCALVAR_N(aload_3, 3)

/*
 * LOAD_ARRAY
 * Load a word from an array.
 * No I-stream parameters.
 * Takes an index from TOS and an array handle from beneath it. 
 *  Leaves resulting word on TOS.
 * Macro parameters are name, scale factor (shift count)
 * and the load instruction to use to fetch data.
 */
define( `LOAD_ARRAY',
    `OP0($1)
	ld	[SP-8],TOSA		! TOSA = array
	dec	4,SP
	ld	[SP],TOSB		! TOSB = index
	tst	TOSA			! is the array null?
	SCALENEXT
	bz	nullPointerException
    	ldub	[PC+2],NextByte
					! T2 = the word containing length
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,methods)],T2 
					! shift index, if necessary.
	ifelse( eval( $2 != 0 ), 1,
		`sll	TOSB,$2,T1' )
					! T2 = array length
	srl	T2,DEFINED_CONSTANT(METHOD_FLAG_BITS),T2
	cmp	TOSB,T2			! is index < length?
	blu,a	catName(`finish',$1)	! finish if not out of range
					! Unhand = unhand(array)
	  ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand 
	b,a	arrayIndexOutOfBoundsB
	nop
	OUTLINE
    catName(`finish',$1):
    	inc	PC
					! get the data into TOSA
	$3	[Unhand+ifelse(eval( $2 != 0 ), 1, `T1', `TOSB') ],TOSA
	GO
	st	TOSA,[SP-4]
	INLINE
    ')
	
/*
 * LOAD_DOUBLEWORD_ARRAY
 * Loads two words from an array.
 * No I-stream parameters.
 * Takes an index from TOS and an array handle from beneath it. 
 * Leaves resulting word on TOS.
 * Macro parameter is name. The scale factor is always 8. 2 ld instructions
 * are always used.
 */
define( `LOAD_DOUBLEWORD_ARRAY',
    `OP0($1)
	ld	[SP-8],TOSA		! TOSA = array
	ld	[SP-4],TOSB		! TOSB = index
    	tst	TOSA			! is array null?
	SCALENEXT
	bz	nullPointerException
	ldub	[PC+2],NextByte
    	!! T2 = array word containing length
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,methods)],T2 
	sll	TOSB,3,T1	
    	!! T1 = offset into the array
	srl	T2,DEFINED_CONSTANT(METHOD_FLAG_BITS),T2 ! length
	cmp	TOSB,T2			! is index < length?
	blu,a	finish_load_doubleword_array
					! Unhand = unhand(array)
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand 
	b	arrayIndexOutOfBoundsB
	nop')
	OUTLINE
finish_load_doubleword_array:
    	inc	PC
	ld	[Unhand+T1],TOSA	! get the first word of data
	add	T1,4,T2
	ld	[Unhand+T2],TOSB	! get the second word of data
	st	TOSA,[SP-8]		! store the first word
	GO				
	st	TOSB,[SP-4]		! store the second word
	INLINE

LOAD_ARRAY(iaload,2,ld)
LOAD_DOUBLEWORD_ARRAY(laload)
LOAD_ARRAY(faload,2,ld)
LOAD_DOUBLEWORD_ARRAY(daload)
LOAD_ARRAY(aaload,2,ld)
LOAD_ARRAY(baload,0,ldsb)
LOAD_ARRAY(caload,1,lduh)
LOAD_ARRAY(saload,1,ldsh)

/*
 * STORE_LOCALVAR
 * Store a word into a local variable, as indexed by NextByte.
 * Parameter of macro is name
 */
define(`STORE_LOCALVAR',
    `OP0($1)
	ldub	[PC+2],Opcode
	sll	NextByte,2,T2		! T2 = offset of VarBase[NextByte]
	ld	[SP-4],TOSA		! TOSA = value
	dec	4,SP
	ldub	[PC+3],NextByte
	inc	2,PC
	SCALE( Opcode )
	GO
	st	TOSA,[Varbase+T2] ')

/*
 * STORE_TWO_WORD_LOCALVAR
 * Store top stack words into local variables, as indexed by
 * NextByte. Top word is low-order, and goes into local var following
 * the one indexed. Next word is high-order, and goes into the indexed variable.
 * Parameter of macro is name
 */
define(`STORE_TWO_WORD_LOCALVAR',
    `OP0($1)
	ld	[SP-8],TOSA		! TOSA = first word of value
	sll	NextByte,2,T2		! T2 = offset of VarBase[NextByte]
	ldub	[PC+2],Opcode
	dec	8,SP
	ld	[SP+4],TOSB		! TOSB = second word of value
	inc	2,PC
    	st	TOSA,[Varbase+T2]	! store first word
	SCALE( Opcode )
	ldub	[PC+1],NextByte
	add	Varbase,4,T1
	GO
	st	TOSB,[T1+T2]		! store second word
    ')
STORE_LOCALVAR(istore)
STORE_TWO_WORD_LOCALVAR(lstore)
STORE_LOCALVAR(fstore)
STORE_TWO_WORD_LOCALVAR(dstore)
STORE_LOCALVAR(astore)

/*
 * STORE_LOCALVAR_N
 * There are several opcodes for storing a local variable, as
 * specified in the opcode itself. They are all the same. They
 * all look like this. Parameters of macro are opcode name and variable
 * index. Instruction has no Istream parameters.
 */
define(STORE_LOCALVAR_N,
    `OP0($1)
	ld	[SP-4],TOSA		! get value
	SCALENEXT
	ldub	[PC+2],NextByte
	dec	4,SP
	inc	PC
	GO
	st	TOSA,[Varbase+eval(4* $2)] ! store into register $2
    ')
/*
 * STORE_TWO_WORD_LOCALVAR_N
 * Store top stack words into local variables, as indicated by
 * the opcode itself. Top word is low-order, and goes into local var following
 * the one named. Next word is high-order, and goes into the indexed variable.
 * Parameters of macro are opcode name and variable
 * index. Instruction has no Istream parameters.
 */
define(STORE_TWO_WORD_LOCALVAR_N,
    `OP0($1)
	ld	[SP-8],TOSA		! TOSA = first word of value
	SCALENEXT
	ld	[SP-4],TOSB		! TOSB = second word of value
	dec	8,SP
	ldub	[PC+2],NextByte
	inc	PC
	st	TOSA,[Varbase+eval(4* $2)] ! store into register $2
	GO
	st	TOSB,[Varbase+eval(4+4* $2)] ! store into register eval($2+1)
    ')
STORE_LOCALVAR_N(istore_0, 0)
STORE_LOCALVAR_N(istore_1, 1)
STORE_LOCALVAR_N(istore_2, 2)
STORE_LOCALVAR_N(istore_3, 3)
STORE_TWO_WORD_LOCALVAR_N(lstore_0,0)
STORE_TWO_WORD_LOCALVAR_N(lstore_1,1)
STORE_TWO_WORD_LOCALVAR_N(lstore_2,2)
STORE_TWO_WORD_LOCALVAR_N(lstore_3,3)
STORE_LOCALVAR_N(fstore_0, 0)
STORE_LOCALVAR_N(fstore_1, 1)
STORE_LOCALVAR_N(fstore_2, 2)
STORE_LOCALVAR_N(fstore_3, 3)
STORE_TWO_WORD_LOCALVAR_N(dstore_0,0)
STORE_TWO_WORD_LOCALVAR_N(dstore_1,1)
STORE_TWO_WORD_LOCALVAR_N(dstore_2,2)
STORE_TWO_WORD_LOCALVAR_N(dstore_3,3)
STORE_LOCALVAR_N(astore_0, 0)
STORE_LOCALVAR_N(astore_1, 1)
STORE_LOCALVAR_N(astore_2, 2)
STORE_LOCALVAR_N(astore_3, 3)

/*
 * STORE_ARRAY
 * Store a word to an array.
 * No I-stream parameters.
 * Takes a value from TOS, the index beneath it and an array handle from
 * beneath that. Removes all from stack.
 * Macro parameters are name, scale factor (shift count)
 * and store instruction to use to put data.
 */
define( `STORE_ARRAY',
    `OP0($1)
	ld	[SP-12],T2	! T2 has reference
	ld	[SP-8],TOSA	! TOSA has subscript
	dec	12,SP
	ld	[SP+8],TOSB	! TOSB has value to store
    	tst	T2
	bz	nullPointerException
	! scale subscript for indexing to R3
	ifelse( eval( $2 == 0 ), 0,
		`sll	TOSA,$2,R3',
		`mov	TOSA,R3')
    	ld	[T2+STRUCT_OFFSET(Hjava_lang_Object,methods)],T1 ! array length
	ld	[T2+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand ! array data
	srl	T1,DEFINED_CONSTANT(METHOD_FLAG_BITS),T1 ! shift out flag bits
	cmp	TOSA,T1	        ! is index < length
	blu	catName(`finish',$1)
	SCALENEXT
	b,a	arrayIndexOutOfBoundsA

	OUTLINE
    catName(`finish',$1):
    	ldub	[PC+2],NextByte	
	$3	TOSB,[Unhand+R3]	! store the value
	GO
	inc	PC
	INLINE ')
	
/*
 * STORE_DOUBLEWORD_ARRAY
 * Store two words to an array.
 * No I-stream parameters.
 * Takes two words of value from TOS, the index beneath it
 * and an array handle from
 * beneath that. Removes all from stack.
 * Macro parameter is name. Scale factor (shift count)
 * and store instruction are always the same.
 */
define( `STORE_DOUBLEWORD_ARRAY',
    `OP0($1)
	ld	[SP-16],T2	! T2 has reference
	SCALENEXT
	ld	[SP-12],T1	! T1 has subscript
	dec	16,SP
	ld	[SP+8],TOSA	! TOSA and TOSB has value to store
	tst	T2
	bz	nullPointerException
	ld	[SP+12],TOSB
	ld	[T2+STRUCT_OFFSET(Hjava_lang_Object,methods)],T3 ! array length
	sll	T1,3,R3	! scale subscript for indexing
	ld	[T2+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand ! array data
	srl	T3,DEFINED_CONSTANT(METHOD_FLAG_BITS),T3 ! shift out flag bits
	cmp	T1,T3		! is index < length
	blu    finish_store_doubleword_array
    	ldub	[PC+2],NextByte
	b,a    arrayIndexOutOfBoundsD
	.empty ')
	OUTLINE
finish_store_doubleword_array:
	add	R3,4,T2	! offset of second word
	st	TOSA,[Unhand+R3]	! store first word
	inc	PC
	GO
	st	TOSB,[Unhand+T2]	! store second word
	INLINE
STORE_ARRAY(iastore,2,st)
STORE_DOUBLEWORD_ARRAY(lastore)
STORE_ARRAY(fastore,2,st)
STORE_DOUBLEWORD_ARRAY(dastore)

/*
 * AASTORE
 * Has more complex semantics than the scalar stores
 * we have been seeing. Value in TOS, index beneath that,
 * array handle under that. Value must be of a compatible type
 * with the array. And, of course, index must be in range, too.
 * Recall that there is a pointer to the classblock of the array element
 * type in the word at array[array.length]
 */
OP0(aastore)
	ld	[SP-12],T2	! T2 has reference
	ld	[SP-8],TOSA	! TOSA has subscript
	dec	12,SP
	ld	[SP+8],TOSB	! TOSB has value to store
    	tst	T2
	bz	nullPointerException
	sll	TOSA,2,R3	! scale subscript for indexing to R3
    	ld	[T2+STRUCT_OFFSET(Hjava_lang_Object,methods)],T1 ! array length
	ld	[T2+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand ! array data
	srl	T1,DEFINED_CONSTANT(METHOD_FLAG_BITS),T1 ! shift out flag bits
	cmp	TOSA,T1		! is index < length?
	blu	finish_aastore
	sll	T1,2,T1		! scale for indexing!
	b,a	arrayIndexOutOfBoundsA

	OUTLINE
    finish_aastore:
	! is_instance_of(value, array[length], Ee)
	mov	TOSB,%o0	! get proposed value
	ld	[Unhand+T1],%o1	! get array classblock pointer
	call	is_instance_of,3
	mov	Ee,%o2

	! if is_instance_of returns True, then we do the store and
	! continue execution. Else we go have an exception.
	tst	%o0
	SCALENEXT
	bz	arrayStoreException
    	ldub	[PC+2],NextByte
	st	TOSB,[Unhand+R3] ! store the value
	GO
	inc	PC
	INLINE

STORE_ARRAY(bastore,0,stb)
STORE_ARRAY(castore,1,sth)
STORE_ARRAY(sastore,1,sth)

/*
 * POP -- delete word from stack top.
 */
OP0(pop)
	SCALENEXT
	ldub	[PC+2],NextByte
	dec	4,SP
	GO
	inc	PC
/*
 * POP2 -- delete two words from stack top.
 */
OP0(pop2)
	SCALENEXT
	ldub	[PC+2],NextByte
	dec	8,SP
	GO
	inc	PC
/*
 * DUP -- duplicate word on top of stack.
 */
OP0(dup)
	ld	[SP-4],TOSA
	SCALENEXT
	ldub	[PC+2],NextByte
	st	TOSA,[SP]
	inc	PC
	GO
	inc	4,SP
/*
 * DUP_X1 -- duplicate stack top word, insert in stack below 2nd.
 */
OP0(dup_x1)
	ld	[SP-4],TOSB	! load 2, store 3
	SCALENEXT
	ld	[SP-8],TOSA
	inc	4,SP
	ldub	[PC+2],NextByte
	inc	PC
	st	TOSB,[SP-12]
	st	TOSA,[SP-8]
	GO
	st	TOSB,[SP-4]
	
/*
 * DUP_X2 -- duplicate stack top word, insert in stack below 3rd.
 */
OP0(dup_x2)		! load 3, store 4
	ld	[SP-4],TOSB
	SCALENEXT
	ld	[SP-8],TOSA
	inc	PC
	ld	[SP-12],T2
	inc	4,SP
	st	TOSB,[SP-16]
	st	T2,[SP-12]
	st	TOSA,[SP-8]
	st	TOSB,[SP-4]
	GO
	ldub	[PC+1],NextByte
	
/*
 * DUP2 -- duplicate two words from the top of stack.
 */
OP0(dup2)		! load 2 store 2 on top!
	ld	[SP-4],TOSB
	SCALENEXT
	ld	[SP-8],TOSA
	inc	8,SP
	ldub	[PC+2],NextByte
	st	TOSB,[SP-4]
	inc	PC
	GO
	st	TOSA,[SP-8]
	
/*
 * DUP2_X1 -- duplicate two words from the top of stack.
 * Insert them beneath the 3rd word.
 */
OP0(dup2_x1)
	ld	[SP-8],TOSA
	SCALENEXT
	ld	[SP-4],TOSB	! load 3, store 5
	inc	PC
	ld	[SP-12],T2
	inc	8,SP
	ldub	[PC+1],NextByte
	st	TOSA,[SP-20]
	st	TOSB,[SP-16]
	st	T2,[SP-12]
	st	TOSA,[SP-8]
	GO
	st	TOSB,[SP-4]
	
/*
 * DUP2_X2 -- duplicate two words from the top of stack.
 * Insert them beneath the 4th word.
 */
OP0(dup2_x2)
	ld	[SP-4],TOSB	! load 4, store 6
	ld	[SP-8],TOSA
	SCALENEXT
	ld	[SP-12],T1
	ld	[SP-16],T2
	inc	8,SP
	st	TOSA,[SP-24]
	st	TOSB,[SP-20]
   	inc	PC
	st	T2,[SP-16]
	st	T1,[SP-12]
	st	TOSA,[SP-8]
	st	TOSB,[SP-4]
	GO
	ldub	[PC+1],NextByte
	
/*
 * SWAP -- interchange two words on top of stack.
 */
OP0(swap)
	ld	[SP-4],TOSB
	ld	[SP-8],TOSA
	SCALENEXT
	ldub	[PC+2],NextByte
	st	TOSB,[SP-8]
	inc	PC
	GO
	st	TOSA,[SP-4]

	
/*
 * INT_ARITHMETIC
 * The binary integer arithmetic ops, such as add, sub, and, or, xor
 * are all pretty  much the same. Macro parameters are op name and binary
 * instruction opcode.
 * Multiply, division, remainder are done using runtime calls.
 * See call_2_result_1.
 */

define(`INT_ARITHMETIC',
    `OP0($1)
	ld	[SP-4],TOSB	! second operand to opcode
	SCALENEXT
	ld	[SP-8],TOSA	! first operand to opcode
	dec	4,SP
	ldub	[PC+2],NextByte
	inc	PC
	$2	TOSA,TOSB,TOSA	! perform the operation
	GO
	st	TOSA,[SP-4] ' )
	
/*
 * CALL2_RESULT_1
 * Arithmetic call implemented via call to runtime routine.
 * Macro parameters are label and runtime routine name.
 * Not to be called if we need to check for operands of zero!
 */
define(`CALL_2_RESULT_1',
    `OP0($1)
	ld	[SP-4],%o1	! second argument to operation
	dec	4,SP
	call	$2,2		! perform the operation
	ld	[SP-4],%o0	! first argument to operation
	SCALENEXT
	st	%o0,[SP-4]
	ldub	[PC+2],NextByte
	GO
	inc	PC' )
	
/*
 * INT_ARITHMETIC_CHECKED_CALL
 * Arithmetic call implemented via call to runtime routine.
 * Macro parameters are label, runtime routine name.
 * Always checks for 2nd operand of 0 before doing call,
 * and may cause an arithmetic exception.
 */
define(`INT_ARITHMETIC_CHECKED_CALL',
    `OP0($1)
	ld	[SP-4],%o1	! second argument to operand
	ld	[SP-8],%o0	! first argument to operation
	tst	%o1
	bz	arithmeticException
	dec	4,SP
    	call	$2,2		! perform the operation
	inc	PC
	SCALENEXT
	ldub	[PC+1],NextByte
	GO
	st	%o0,[SP-4] ' )
	
/*
 * LONG_ARITHMETIC
 * The binary long-int ops are a bunch more work.
 * This macro will only work on the simple ones
 * such as add, sub, xor, and, or. NOT mult or div or rem.
 * Macro parameters are name, low-order instruction opcode,
 * high-order instruction opcode.  We are doing an operation that
 * allows us to operate on the low order bits first
 * Multiply, division, remainder are done using runtime calls.
 */
define(`LONG_ARITHMETIC',
    `OP0($1)
	ld	[SP-4],TOSB	! low word, second operand
	SCALENEXT
	ld	[SP-12],T2	! low word, first operand
	dec	8,SP
	ld	[SP],TOSA	! high word, second operand
	ld	[SP-8],T1	! high word, first operand
    	$2	T2,TOSB,TOSB	! perform operation of low word
	ldub	[PC+2],NextByte
	inc	PC
	$3	T1,TOSA,TOSA	! perform operation on high words
	st	TOSB,[SP-4]
	GO
	st	TOSA,[SP-8]')
	
/*
 * LONG_ARITHMETIC_CHECKED_CALL
 * Long-int arithmetic operation implemented via call to runtime routine.
 * Always checks second parameter for 0.
 * Macro parameters are label, runtime routine name.
 */
define(`LONG_ARITHMETIC_CHECKED_CALL',
    `OP0($1)
	ld	[SP-8],%o2	! high word, second operand
	ld	[SP-4],%o3	! low word, second operand
	dec	8,SP
	orcc	%o2,%o3,%g0	! is either word non-zero?
	bz	arithmeticException
	ld	[SP-8],%o0	! high word, first operand
	call	$2,4
	ld	[SP-4],%o1	! low word, first operand
	SCALENEXT
	ldub	[PC+2],NextByte
    	st	%o0,[SP-8]
	inc	PC
	GO
	st	%o1,[SP-4] ')
	
/*
 * FLOAT_ARITHMETIC
 * A lot like int_arithmetic, but in float registers.
 */
define(`FLOAT_ARITHMETIC',
    `OP0($1)
	ld	[SP-8],%f0	! first argument
	SCALENEXT		
	ld	[SP-4],%f1	! second argument, or second word
	dec	4,SP
	ifelse(	``$2'', ``fdtos'', `fdtos	%f0,%f2',
		``$2'', ``fdtoi'', `fdtoi	%f0,%f2',
	    	`$2	%f0,%f1,%f2')
	ldub	[PC+2],NextByte
	inc	PC
	GO
	st	%f2,[SP-4]' )
	
/*
 * DOUBLE_ARITHMETIC
 * Double arithmetic is a pain in the butt
 * because it involves so much memory traffic.
 * Macro parameters are name and double-float instruction
 * opcode.
 */
define(`DOUBLE_ARITHMETIC',
    `OP0($1)
	ld	[SP-16],%f0	! first word, first argument
	ld	[SP-12],%f1	! second word, first argument
	ld	[SP-8],%f2	! first word, second argument
	SCALENEXT
	ld	[SP-4],%f3	! second word, second argument
	dec	8,SP
	$2	%f0,%f2,%f4	! do the operation
	ldub	[PC+2],NextByte
	inc	PC
	st	%f4,[SP-8]
	GO
	st	%f5,[SP-4]')
	
INT_ARITHMETIC(iadd, add )
LONG_ARITHMETIC(ladd, addcc, addx )
FLOAT_ARITHMETIC(fadd, fadds)
DOUBLE_ARITHMETIC(dadd, faddd)
INT_ARITHMETIC(isub, sub )
LONG_ARITHMETIC(lsub, subcc, subx )
FLOAT_ARITHMETIC(fsub, fsubs)
DOUBLE_ARITHMETIC(dsub, fsubd)
CALL_2_RESULT_1(imul, .mul )

/*
 * LMUL
 * Long-int multiplication. Call routine __mul64 to do the
 * work. Similar to other call-out situations, but does not
 * check for zero parameter, as this is valid multiplication input.
 */
OP0(lmul)
	ld	[SP-16],%o0
	ld	[SP-12],%o1
	dec	8,SP
	ld	[SP],%o2
	call	__mul64,4
	ld	[SP+4],%o3
	SCALENEXT
	ldub	[PC+2],NextByte
    	st	%o0,[SP-8]
	st	%o1,[SP-4]
	GO
	inc	PC

FLOAT_ARITHMETIC(fmul, fmuls)
DOUBLE_ARITHMETIC( dmul, fmuld )
INT_ARITHMETIC_CHECKED_CALL(idiv, .div )
LONG_ARITHMETIC_CHECKED_CALL(ldiv, __div64 )
FLOAT_ARITHMETIC(fdiv, fdivs )
DOUBLE_ARITHMETIC(ddiv, fdivd )
INT_ARITHMETIC_CHECKED_CALL(irem, .rem )
LONG_ARITHMETIC_CHECKED_CALL(lrem, __rem64 )
	
/*
 * FREM
 * pick two floats off stack, call fmod, stack result.
 * Trouble is, fmod takes double parameters and returns a double,
 * so we have to do converting of everything at all points.
 * ( I bet there is a single fmods somewhere, I just do not know
 * about it! )
 */
OP0(frem)
	ld	[SP-8],%f0
	ld	[SP-4],%f2
	dec	4,SP
	fstod	%f0,%f0
	fstod	%f2,%f2
	std	%f0,[%fp-StackT]
	std	%f2,[%fp-StackT+8]
	ldd	[%fp-StackT],%o0
	call	fmod,4
	ldd	[%fp-StackT+8],%o2
	fdtos	%f0,%f0
	SCALENEXT
	ldub	[PC+2],NextByte
	inc	PC
	GO
	st	%f0,[SP-4]

/*
 * DREM
 * Call fmod with 4 TOS words. A lot like lmul, for instance,
 * but value is returned in %f0/%f1.
 */
OP0(drem)
	ld	[SP-16],%o0
	ld	[SP-12],%o1
	dec	8,SP
	ld	[SP],%o2
	call	fmod,4
	ld	[SP+4],%o3
	SCALENEXT
    	st	%f0,[SP-8]
    	st	%f1,[SP-4]
	inc	PC
	GO
	ldub	[PC+1],NextByte

/*
 * INEG -- single-word integer negation
 */
OP0(ineg)
	ld	[SP-4],TOSA
	SCALENEXT
	ldub	[PC+2],NextByte
	inc	PC
	neg	TOSA
	GO
	st	TOSA,[SP-4]
	
/*
 * LNEG -- Long integer negation
 */
OP0(lneg)
	ld	[SP-4],TOSB
	SCALENEXT
	ld	[SP-8],TOSA
	ldub	[PC+2],NextByte
	subcc	%g0,TOSB,TOSB
	inc	PC
	st	TOSB,[SP-4]
	subx	%g0,TOSA,TOSA
	GO
	st	TOSA,[SP-8]
	
/*
 * UNARY_FLOAT_ARITHMETIC
 * Single-word floating unary operation
 * Macro parameters are name and opcode.
 */
define(UNARY_FLOAT_ARITHMETIC,
    `OP0($1)
	ld	[SP-4],%f0
	SCALENEXT
	ldub	[PC+2],NextByte
	$2	%f0,%f0
	inc	PC
	GO
	st	%f0,[SP-4]')
UNARY_FLOAT_ARITHMETIC(fneg,fnegs)
	
/*
 * DNEG -- Double floating negation
 */
OP0(dneg)
	ld	[SP-8],%f0
	SCALENEXT
	ld	[SP-4],%f1
	inc	PC
	ldub	[PC+1],NextByte
	fnegs	%f0,%f0
	st	%f1,[SP-4]
	GO
	st	%f0,[SP-8]
	
/*
 * Integer shifts.
 * The SPARC 32-bit shift instructions are defined such that only the
 * low-order five bits  of the shift count participate in the operation.
 * Coincidentally, this is the same definition as the Java shift operator. 
 * Thus we need not explicitly mask the shift count before operating.
 * This also means that for a shift count T2 >= 32, the low-order
 * 5 bits of "T2-32" are the same as the low-order 5 bits of
 * "T2", and that the low-order 5 bits of "32-T2" are the same as 
 * the low-order 5 bits of "-T2".
 * We make use of these artifacts in the 64-bit cases.
 */
INT_ARITHMETIC(ishl,sll)
	
/*
 * LSHL -- Long integer left shift.
 * 
 */
OP0(lshl)
	ld	[SP-4],TOSB	! fetch shift count to TOSB
	SCALENEXT
	ld	[SP-8],TOSA	! fetch low-order word to TOSA
	dec	4,SP
	btst	32,TOSB
	bz	finish_small_long_shift_left
	ldub	[PC+2],NextByte
        ! large shift count.  The high order word is immaterial
	sll	TOSA,TOSB,TOSA
	st	%g0,[SP-4]
	inc	PC
	GO
	st	TOSA,[SP-8]
	OUTLINE
finish_small_long_shift_left:
	! small shift count
	! compute into register T1 the high-order bits
	! of the low-order source word that end up in the
	! low-order bits of the high-order result word.
	! There are shift-count of them. They are computed by
	! shifting right by ( 32 - shift-count bits ). I had believed
	! that the low-order bits of the negated shift-count would suffice.
	! This is wrong! in the case of a zero shift count! Thus this nonsense.
	ld	[SP-8],T2	! fetch hight-order word
	mov	31,T3
	sub	T3,TOSB,T3
	srl	TOSA,T3,T1	! shift right by 31 - shift-count
	srl	T1,1,T1		! then shift right by one more!!
	sll	T2,TOSB,R3
	inc	PC
	sll	TOSA,TOSB,TOSB
	or	T1,R3,TOSA
	st	TOSB,[SP-4]
	GO
	st	TOSA,[SP-8]
	INLINE
INT_ARITHMETIC(ishr,sra)
	
/*
 * LONG_RIGHT_SHIFT -- Long integer right shifts.
 * Parameters are: name, sra or srl opcode.
 */
define( LONG_RIGHT_SHIFT,
    `OP0($1)
	ld	[SP-4],TOSB	! fetch shift count to TOSB
	SCALENEXT
	ld	[SP-12],T2	! fetch high-order word
	dec	4,SP
	btst	32,TOSB		! is it a big shift?
	bz	catName(finish_small,$1)
	ldub	[PC+2],NextByte
        ! large shift count.  The low order word is immaterial.
	! The high word of the result is 0 for an unsigned shift, and the
    	! sign of the high word for a signed shift.
	ifelse(``$2'',``sra'',
	    `sra	T2,31,TOSA' ) ! TOSA = sign of high word of argument
	! for purposes of shifting, TOSB == TOSB-32, which
	! is the shift quantity we intend here.
	$2	T2,TOSB,TOSB
	ifelse(``$2'',``sra'',
	    `st		TOSA,[SP-8]',
	    `st		%g0,[SP-8]' )
	inc	PC
	GO
	st	TOSB,[SP-4]
	OUTLINE
    catName(finish_small,$1):
	! small shift count
	! count in TOSB
	! low word in TOSA
	! compute into TOSB the low-order bits
	! of the high-order source word that end up in the
	! high-order bits of the low-order result word.
	! There are shift-count of them. They are computed by
	! shifting left by ( 32 - shift-count bits ). I had believed
	! that the low-order bits of the negated shift-count would suffice.
	! This is wrong! in the case of a zero shift count! Thus this nonsense.
	ld	[SP-4],TOSA	! fetch low-order word to TOSA
	mov	31,T1
	sub	T1,TOSB,T1
	srl	TOSA,TOSB,R3	! part of low result in R3
	$2	T2,TOSB,TOSA	! high result in TOSA
	sll	T2,T1,TOSB	! rest of low result in TOSB: shift by 31-shift-count
	st	TOSA,[SP-8]
	inc	PC
	sll	TOSB,1,TOSB	! ... then shift once more!
	or	TOSB,R3,TOSB	! TOSB <- TOSB | R3
	GO
	st	TOSB,[SP-4]
	INLINE ')
	
LONG_RIGHT_SHIFT(lshr,sra)
INT_ARITHMETIC(iushr,srl)
LONG_RIGHT_SHIFT(lushr,srl)
INT_ARITHMETIC(iand,and)
LONG_ARITHMETIC(land, and, and )
INT_ARITHMETIC(ior,or)
LONG_ARITHMETIC(lor, or, or )
INT_ARITHMETIC(ixor,xor)
LONG_ARITHMETIC(lxor, xor, xor )

/*
 * IINC -- Increment local variable by constant.
 * Has two byte-stream arguments: 
 *    the local variable index (in NextByte)
 *    the signed, 8-bit constant by which to increment.
 * Leaves the stack alone.
 */
OP0(iinc)
	ldsb	[PC+2],T1	! T1 has const value.
	sll	NextByte,2,T2	! T2 has scaled index
	ldub	[PC+3],NextByte ! NextByte has next opcode
	ld	[Varbase+T2],R3 ! R3 has value
	SCALENEXT
	ldub	[PC+4],NextByte ! NextByte has fetchahead
	add	R3,T1,R3	! increment value
	inc	3,PC
	GO
	st	R3,[Varbase+T2]	! put value back where you got it

/*
 * I2L -- convert single int to long int.  
 * We need to sign extend
 */
OP0(i2l)
	ld	[SP-4],TOSB	! get the value
	SCALENEXT
	ldub	[PC+2],NextByte
	sra	TOSB,31,TOSA	! sign extension
	inc	4,SP
	st	TOSB,[SP-4]
	inc	PC
	GO
	st	TOSA,[SP-8]
	
/*
 * Convert single int to float.
 */
UNARY_FLOAT_ARITHMETIC(i2f,fitos)

/*
 * I2D -- Convert single int to double float.
 */
OP0(i2d)
	ld	[SP-4],%f0	! the the value
	SCALENEXT
	inc	PC
	fitod	%f0,%f0		! convert to double
	ldub	[PC+1],NextByte
	inc	4,SP
	st	%f0,[SP-8]
	GO
	st	%f1,[SP-4]
	
/* 
 * L2I --Convert long int to single int, by truncation
 */
OP0(l2i)
	ld	[SP-4],TOSA	! get the low order word
	SCALENEXT
	dec	4,SP
	ldub	[PC+2],NextByte
	inc	PC
	GO
	st	TOSA,[SP-4]	!  and move it
	
/*
 * L2F -- Convert long int to float.
 * Very hard. Call out to helper route to do the dirty work!
 */
OP0(l2f)
	ld	[SP-8],%o0
	dec	4,SP
	call	lltof,2
	ld	[SP],%o1

	st	%f0,[SP-4]
	SCALENEXT
	ldub	[PC+2],NextByte
	GO
	inc	PC
	
/*
 * Convert long int to double.
 * Very hard. Call out to helper route to do the dirty work!
 */
OP0(l2d)
	ld	[SP-8],%o0
	call	lltod,2
	ld	[SP-4],%o1
	st	%f0,[SP-8]
	SCALENEXT
	st	%f1,[SP-4]
	ldub	[PC+2],NextByte
	GO
	inc	PC

OP0(f2i)
	call	float2l,1
	ld	[SP-4],%o0
	SCALENEXT
	ldub	[PC+2],NextByte
	st	%o0,[SP-4]
	GO
	inc	PC
	
/*
 * F2L -- convert float to long int.
 * Done by way of C-library call.
 */

OP0(f2l)
	ld	[SP-4],%o0
	call	float2ll,1
	inc	4,SP
	SCALENEXT
	ldub	[PC+2],NextByte
	st	%o0,[SP-8]
	st	%o1,[SP-4]
	GO
	inc	PC

/*
 * F2D -- convert single float to double float.
 */
OP0(f2d)
	ld	[SP-4],%f0
	SCALENEXT
	ldub	[PC+2],NextByte
	fstod	%f0,%f0
	inc	PC
	inc	4,SP
	st	%f0,[SP-8]
	GO
	st	%f1,[SP-4]

OP0(d2i)
	ld	[SP-8],%o0	! first word
	SCALENEXT		
	ld	[SP-4],%o1	! second word
	call	double2l, 2
	dec	4,SP
	
	ldub	[PC+2],NextByte
	inc	PC
	GO
	st	%o0,[SP-4]

/*
 * Convert double to long int.
 * Call out to C library.
 */
OP0(d2l)
	ld	[SP-8],%o0      ! first word
	call	double2ll,2     ! second word
	ld	[SP-4],%o1
	SCALENEXT
	ldub	[PC+2],NextByte
	st	%o0,[SP-8]
	st	%o1,[SP-4]
	GO
	inc	PC

FLOAT_ARITHMETIC(d2f, fdtos)

/*
 * SHORTEN_INT -- Shorten the TOS word to a byte, or short, by sign
 * extending. Macro arguments are name and number of bits
 * of the target type.
 */
define( `SHORTEN_INT',
    `OP0($1)
	ld	[SP-4],TOSA
	SCALENEXT
	ldub	[PC+2],NextByte
	sll	TOSA,eval(32- $2),TOSA
	inc	PC
	sra	TOSA,eval(32- $2),TOSA
	GO
	st	TOSA,[SP-4] ')

SHORTEN_INT(i2b, 8)

/*
 * I2C -- Shorten the TOS word to a char by zero
 * extending. Avoid use of shifts, which group poorly.
 */
OP0(i2c)
	ld	[SP-4],TOSA
	SCALENEXT
	sethi	%hi(0xffff0000),T2
	ldub	[PC+2],NextByte
	inc	PC
	andn	TOSA,T2,TOSA
	GO
	st	TOSA,[SP-4]

SHORTEN_INT(i2s, 16)

/*
 * LCMP -- Compare two longs from the stack. Replace them
 * with -1, 0, or 1, according to the comparison result.
 */
	
OP0(lcmp)
	ld	[SP-4],TOSB   		! low order second comparand
	ld	[SP-12],T2    		! low order first comparand
	ld	[SP-8],TOSA 		! high order first comparand
	ld	[SP-16],T1   		! high order second comparand
	dec	12,SP
	!! calculate T1:T2 - TOSA:TOSB (i.e. second - first)
	subcc	T2, TOSB, T3	        ! T3 == 0 iff bottom halves the same
	SCALENEXT
	subxcc	T1, TOSA, %g0
	!! At this point, blt or bge would work correctly.. . .
	ldub	[PC+2],NextByte
	bge	lcmp_ge
	  inc	PC
	mov	-1,TOSA
	GO
	st	TOSA,[SP-4]
	
	OUTLINE
lcmp_ge:    
	!! T3 contains 0 iff bottom halves are equal.  Also need to see if 
	!! top halves are equal
    	subcc	T1, TOSA, T4            ! T4 == 0 iff top halves are the same
	mov	1,TOSA		        ! possible result
	orcc	T3,T4,%g0               ! are both halves the same?
	skipnz  
	mov	0,TOSA		    
	GO
	st	TOSA,[SP-4]
	INLINE


/*
 * FCOMPARE
 * Compare two floats from the stack. Replace them
 * with -1, 0, or 1, according to the comparison result.
 * Map unordered to -1 or 1, depending on branch conditions chosen.
 * Macro arguments are name and fbranch opcode
 * Note: SPARC architecture demands a non-fp instruction
 * between the fcmp and the fbXXX!
 * The SPARC ASSEMBLER demands TWO of them, and will insert
 * NOPs, silently!!! It will ALWAYS insert at least one!!!
 */
define( `FCOMPARE',
    `OP0($1)
	ld	[SP-8],%f0
	mov	-1,TOSA	
	ld	[SP-4],%f1
	SCALENEXT
	fcmps	%f0,%f1
	ldub	[PC+2],NextByte
	dec	4,SP
	$2	finish_floating_compare
	inc	PC
	GO 
	st	TOSA,[SP-4] ! < (or unordered) ')

FCOMPARE(fcmpl, fbge )
FCOMPARE(fcmpg, fbuge )

/*
 * DCOMPARE
 * Compare two double floats from the stack. Replace them
 * with -1, 0, or 1, according to the comparison result.
 * Map unordered to -1 or 1, depending on branch conditions chosen.
 * Macro arguments are name and fbranch opcode
 */
define(`DCOMPARE',
    `OP0($1)
	ld	[SP-16],%f0
	ld	[SP-12],%f1
	ld	[SP-8],%f2
	ld	[SP-4],%f3
	SCALENEXT
	mov	-1,TOSA
	fcmpd	%f0,%f2
	ldub	[PC+2],NextByte
	dec	12,SP
	$2	finish_floating_compare
	inc	PC
	GO	
	st	TOSA,[SP-4]	! < ( or unordered ) ')
	OUTLINE
finish_floating_compare:
	fbe,a	1f
	    mov	0,TOSA	! ==
	    mov	1,TOSA	! otherwise
1:	GO
	st	TOSA,[SP-4]
	INLINE

DCOMPARE(dcmpl, fbge )
DCOMPARE(dcmpg, fbuge )

/*
 * IFBRANCH -- Take the TOS word. Compare with zero.
 * Branch on condition, signed 2-byte offset. Macro arguments are name
 * and SPARC branch opcode to emulate.
 */
define(`IFBRANCH',
    `OP0($1)
	ld	[SP-4],TOSA
	dec	4,SP
	tst	TOSA
	$2,a	do.branch.0.X
	    ldub	[PC+2],T2
	! branch not taken.  Perform next operation    	
	ldub	[PC+3],Opcode
	ldub	[PC+4],NextByte
	SCALE(Opcode)
	GO
	inc	3,PC ')

	OUTLINE
    do.branch.0:
	ldub	[PC+2],T2
    do.branch.0.X:
    	sll	NextByte,24,NextByte
	sra	NextByte,16,NextByte
	add	NextByte,PC,PC
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	add	T2,PC,PC
	!
	! FALL THROUGH, to finish_branch.0
	!
/*
 * FINISH_BRANCH.0
 * All successful branches go through here, to check for 
 * asynchronous exceptions on their way across a control arc.
 * I believe that to check conditionally costs more, in the
 * aggregate, than to check unconditionally. So unconditional it is.
 * 
 * Assume:
 *	PC -- contains target PC
 *	T1  -- contains signed byte at [Ee+STRUCT_OFFSET(execenv,exceptionKind)]
 */

 finish_branch.0:
	ldub	[PC],Opcode
	tst	T1
	bgt	handleException
	SCALE(Opcode)	! assume success
	GO
	ldub	[PC+1],NextByte
	
	INLINE

IFBRANCH(ifeq, be)
IFBRANCH(ifne, bne)
IFBRANCH(iflt, bl)
IFBRANCH(ifge, bge)
IFBRANCH(ifgt, bg)
IFBRANCH(ifle, ble)

/*
 * IFCMP -- implementation of
 * if_icmpXXX and if_acmpXXX -- compare two words from the stack
 * and branch on condition. Macro parameters are name and
 * SPARC opcode to emulate.
 */
define(`IFCMP',
    `OP0($1)
	ld	[SP-8],TOSA
	ld	[SP-4],TOSB
	cmp	TOSA,TOSB
	$2	do.branch.0
	dec	8,SP
	ldub	[PC+3],Opcode
	ldub	[PC+4],NextByte
	SCALE(Opcode)
	GO
	inc	3,PC ')
IFCMP(if_icmpeq,be)
IFCMP(if_icmpne,bne)
IFCMP(if_icmplt,bl)
IFCMP(if_icmpge,bge)
IFCMP(if_icmpgt,bg)
IFCMP(if_icmple,ble)
IFCMP(if_acmpeq,be)
IFCMP(if_acmpne,bne)
	
/*
 * GOTO -- Takes two byte signed offset.
 * Leaves stack alone.
 */
OP0(goto)
	ldub	[PC+2],T2
    	sll	NextByte,24,NextByte
	sra	NextByte,16,NextByte
	add	PC,NextByte,PC
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	b	finish_branch.0
	add	PC,T2,PC
	
/*
 * JSR -- Takes two byte signed offset. Leaves
 * PC+3 on TOS.
 */
OP0(jsr)
	ldub	[PC+2],T2
    	sll	NextByte,24,NextByte
	inc	4,SP
	sra	NextByte,16,NextByte
	add	PC,3,TOSA
	add	PC,NextByte,PC
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	st	TOSA,[SP-4]
	b	finish_branch.0
	add	PC,T2,PC
	
/*
 * RET -- Get an address word from a local, and jump to it.
 * Leaves stack alone.
 */
OP0(ret)
	sll	NextByte,2,T2	! scale
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	b	finish_branch.0
	ld	[Varbase+T2],PC	! fetch

/*
 * TABLESWITCH --
 * Fetch default,
 * low, high. Compare TOS word to low, high and 
 * either index or use default as pc-relative displacement.
 */
OP0(tableswitch)
	ld	[SP-4],TOSA	! index in TOSA
	add	PC,4+12,TOSB	! table base in TOSB
	dec	4,SP
	andn	TOSB,3,TOSB	! finish computing table base
	ld	[TOSB-8],T2	! low bound in T2
	ld	[TOSB-4],T1	! high bound in T1
	sub	TOSA,T2,TOSA	! bias index and high bound by low bound
	sub	T1,T2,T1
	cmp	TOSA,T1
	bgu,a	1f
	    ld	[TOSB-12],T1	! fetch default offset into T1
	sll	TOSA,2,TOSA	! scale for indexing
	ld	[TOSB+TOSA],T1	! fetch table entry into T1
1:	add	PC,T1,PC	! PC gets new address
	b	finish_branch.0
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	INLINE
	
/*
 * LOOKUPSWITCH --
 * find key in table of (key,value) pairs, use value as
 * pc-relative branch offset. Uses binary search, straight
 * out of Knuth.
 *
 * This is sticky enough that I want my own local register
 * defs. Thus...
 */
define(LOW,T2)
define(HIGH,T1)
define(MID,R3)
define(BASE,TOSB)

OP0(lookupswitch)
	ld	[SP-4],TOSA	! index in TOSA
	add	PC,4+8,BASE	! table base in TOSB
	andn	BASE,3,BASE	! finish computing table base
	ld	[BASE-4],HIGH	! high bound ...
	dec	4,SP
	mov	0,LOW		! low bound
	deccc	HIGH		! high bound is npairs-1
	sll	HIGH,3,HIGH	! keep indices scaled by 8
	bge	lookupswitch_loop
	mov	HIGH,MID
        !! Yow. npairs <= 0.  Should almost never happen
	b,a	lookupswitch_default

	OUTLINE
lookupswitch_loop:
	srl	MID,1,MID	! "divide" MID by 2
    	! INVARIANT:	 
    	!    LOW <= MID <= HIGH (and each points at some key in the table)
    	! The key, if it is in the table, must be somewhere between LOW and
    	! high, inclusive
	andn	MID,7,MID	! keep it 8-scaled
	ld	[BASE+MID],T3	! get key value
	cmp	TOSA,T3		! compare to lookup key
	be,a	2f
	    add	MID,4,MID	! point MID at value word on match
	bg,a	1f		! else partition range
	    add	MID,8,LOW
    	    sub MID,8,HIGH
    1:	cmp	LOW,HIGH
	ble	lookupswitch_loop
	add	LOW,HIGH,MID	! MID <- LOW+HIGH / 2, approx.
	
lookupswitch_default:		
	! Here, fell out of loop becase LOW, HIGH collided
	mov	-8,MID		! point MID at default value
    2:  ld	[BASE+MID],TOSA	! pick up default or value offset
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	b	finish_branch.0
	add	PC,TOSA,PC
	INLINE
undefine(`LOW')
undefine(`HIGH')
undefine(`MID')
undefine(`BASE')

/*
 * WORDRETURN --
 * Push TOS word onto caller's TOS, then return to caller.
 * Check for monitors and exceptions.
 * Macro parameter is name.
 */
define( `WORDRETURN',
    `OP0($1)
	ld	[SP-4],TOSA	! value to return
	ld	[FP+STRUCT_OFFSET(javaframe,prev)],ConstP
	dec	4,SP
			       ! cur_frame->monitor
	ld	[FP+STRUCT_OFFSET(javaframe,monitor)],%o1
				! top of stack for previos frame
	ld	[ConstP+STRUCT_OFFSET(javaframe,optop)],T1
	sethi	%hi(java_profiler_isOn),R3
	st	TOSA,[T1]	! store value into previous frame
	inc	4,T1		! increment previous frames top
	st	T1,[ConstP+STRUCT_OFFSET(javaframe,optop)]
	b	finish_return
	ld	[R3+%lo(java_profiler_isOn)],R3 ! java_profiler_isOn in R3
')
	
/*
 * DOUBLEWORDRETURN -- 
 * Push two TOS words onto caller's TOS.
 * Then return to caller.
 * Check for monitors and exceptions.
 * Macro parameter is name.
 */
define( `DOUBLEWORDRETURN',
    `OP0($1)
	ld	[SP-8],TOSA	! value to return
	ld	[FP+STRUCT_OFFSET(javaframe,prev)],ConstP
	ld	[SP-4],TOSB	! value to return
				! cur_frame->monitor
	ld	[FP+STRUCT_OFFSET(javaframe,monitor)],%o1
	dec	8,SP
				! top of stack for previous frame
	ld	[ConstP+STRUCT_OFFSET(javaframe,optop)],T1 
	sethi	%hi(java_profiler_isOn),R3
				! store double value into previous frame
	st	TOSA,[T1]
	st	TOSB,[T1+4]
	inc	8,T1		! increment top of previous frame
	st	T1,[ConstP+STRUCT_OFFSET(javaframe,optop)]
	b	finish_return
	ld	[R3+%lo(java_profiler_isOn)],R3 ! java_profiler_isOn in R3
')
WORDRETURN(ireturn)
DOUBLEWORDRETURN(lreturn)
WORDRETURN(freturn)
DOUBLEWORDRETURN(dreturn)
WORDRETURN(areturn)
	
/*
 * RETURN -- 
 * Just return to caller.
 * Check for monitors and exceptions.
 */
OP0(return)
	ld	[FP+STRUCT_OFFSET(javaframe,prev)],ConstP
			    ! cur_frame->monitor
	ld	[FP+STRUCT_OFFSET(javaframe,monitor)],%o1
	sethi	%hi(java_profiler_isOn),R3
	b	finish_return
	ld	[R3+%lo(java_profiler_isOn)],R3 ! java_profiler_isOn in R3

	OUTLINE

finish_return:
	! FP contains the current frame
	! ConstP contains frame->prev
	! %o0 contains frame->monitor
	! R3 contains java_profiler_isOn
	! SP points at first vacant word on current stack.
	! Varbase, TOSA, TOSB irrelevant
	! PC points to the return instruction.

    	! Is there a monitor?
	tst	%o1		! is there monitor?
	bne	returnExitMonitor
				! VarBase = Initial Frame (used as temp)
	ld	[%fp-Initial_frame],Varbase

	! are we at the initial frame
1:  	cmp	FP,Varbase 
  	! If we are returning from the routine in which we started,
	! then we are out of here.
	be	returnInitialFrame
    	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],TOSA

	sethi	%hi(jvmpi_event_flags),R5
	ld	[R5+%lo(jvmpi_event_flags)],R5
	
	! if flag for JVMPI method exit notification is on, deal with it
	and	R5,DEFINED_CONSTANT(JVMPI_EVENT_METHOD_EXIT_ON),R5
	tst	R5
	bne	returnJvmpiMethodExit
	nop
2:	
	! if java_profiler_isOn, we have to deal with that stuff
  	tst	R3		! test java_profiler_isOn
	bne	returnJavaProfiling
    	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],TOSA
3:	
	! We are now committed to popping the java stack.
	! assign FP = FP.prev   [ConstP contains current frame]
	! re-load our context from this frame.
	ld	[ConstP+STRUCT_OFFSET(javaframe,returnpc)],PC
	mov	ConstP,FP
	! If FP.returnpc == 0, it is a special hacking indicating that
	! this is the first execution of this invoker.  We may want to rewrite
	! it again.
	tst	PC
	beq	returnFixingQuickInvocation
	st	ConstP,[Ee+STRUCT_OFFSET(execenv,current_frame)]
    
	! We continue executing at PC, unless an exception occurred.
4:      cmp	TOSA,DEFINED_CONSTANT(EXCKIND_NONE)
    	bg,a	handleException
	    ld	[FP+STRUCT_OFFSET(javaframe,lastpc)],PC
	! Now we can get back to work.
	ldub	[PC],Opcode
	ld	[FP+STRUCT_OFFSET(javaframe,optop)],SP
	ldub	[PC+1],NextByte
	ld	[FP+STRUCT_OFFSET(javaframe,vars)],Varbase
	SCALE(Opcode)
	GO
	ld	[FP+STRUCT_OFFSET(javaframe,constant_pool)],ConstP

returnExitMonitor:
	add	Ee,STRUCT_OFFSET(execenv,sys_thr),%o0
	! calling sysMonitorExit(ee, frame->monitor )
	CALL_AND_BRANCH(_sysMonitorExit, 2, 1b)

returnJvmpiMethodExit:
	mov	Ee,%o0
	! calling jvmpi_method_exit(ee)
	CALL_AND_BRANCH(jvmpi_method_exit,1,2b)

returnJavaProfiling:
	! Record how much time in this routine from the call site.
       	! int time = sysThreadCPUTime() - frame->profiler_info;
	call	_sysThreadCPUTime,0
	nop
	ld	[FP+STRUCT_OFFSET(javaframe,profiler_info)],%o1
	sub	%o0,%o1,%o1
	mov	FP,%o0
	! java_profiler_log_frame(frame, time);
	CALL_AND_BRANCH(java_profiler_log_frame, 2, 3b)



returnFixingQuickInvocation:	
    	! FP.returnPC == 0
    	! We need to recreate frame->lastpc
    	! Also, call the method FixupQuickInvocation() to possibly inline
    	! the method call    	
        ! lastpc = frame->lastpc; 
	ld	[FP+STRUCT_OFFSET(javaframe,lastpc)],%o1
    	! pc = lastpc + 3	
	add	%o1,3,PC	! when result != 0
	
	mov	FP,%o1
	mov	Ee,%o0
	CALL_AND_BRANCH(FixupQuickInvocation,2,4b)

returnInitialFrame:
	! TOSA already contains exceptionKind(Ee)!
	! decache SP to current frame->optop
	st	SP,[FP+STRUCT_OFFSET(javaframe,optop)]
	! return !exceptionOccurred(ee)
	sub     TOSA,DEFINED_CONSTANT(EXCKIND_NONE)+1,%i0
	srl 	%i0,31,%i0	! signof(ee-ExceptionKind-EXEKIND_NONE-1)
	ret
    	restore

	INLINE

/*
 * ACCESS_SYMBOLIC --  Used to form getstatic, putstatic, getfield, putfield.
 * call ResolveClassConstant.
 * branch to finish code, which will call quickXXXXX to rewrite in place
 * check for exceptions.
 * Macro arguments are name, opcode constant name,
 * and refererence type (CONSTANT_Fieldref or CONSTANT_Methodref)
 * Leaves stack alone.
 */
define(`ACCESS_SYMBOLIC',
    `OP0($1)
	! call ResolveClassConstantFromPC(pc, opcode, constant_pool, flags, ee)
	mov	PC,%o0
	mov	DEFINED_CONSTANT($2),%o1
	mov	ConstP,%o2
	mov	Ee,%o3
	call	ResolveClassConstantFromPC,5
	mov	1<<DEFINED_CONSTANT($3),%o4
	
	tst	%o0		            ! result null indicates error/redo
	mov	%o0,%o2		            ! constant_pool[field_index]
	bne,a   catName(`finish',$1)
	  mov	DEFINED_CONSTANT($2),%o0    ! original opcode
		
        ! did an exception occur?
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T2
	tst	T2		! did we garner an exception
	beq,a	redispatch_invoker
	 ldub	[PC],Opcode
	b,a     handleException	
    ')
	
ACCESS_SYMBOLIC(getstatic, opc_getstatic, CONSTANT_Fieldref)
ACCESS_SYMBOLIC(putstatic, opc_putstatic, CONSTANT_Fieldref)
ACCESS_SYMBOLIC(getfield, opc_getfield, CONSTANT_Fieldref)
ACCESS_SYMBOLIC(putfield, opc_putfield, CONSTANT_Fieldref)
ACCESS_SYMBOLIC(invokevirtual, opc_invokevirtual, CONSTANT_Methodref)
ACCESS_SYMBOLIC(invokespecial, opc_invokespecial, CONSTANT_Methodref)
ACCESS_SYMBOLIC(invokestatic, opc_invokestatic, CONSTANT_Methodref)
	OUTLINE
	
/*
 * FINISH_SYMBOLIC_ACCESS. Out-of-line complement to the
 * above. Macro parameter is quickXXX routine name to call.
 * Label is defined outside macro, as there may be multiple of them.
 * Leaves stack as-is.
 * 
 *   %o0 contains the current byte code
 *   %o2 contains the extracted constant pool element
 */
define(`FINISH_SYMBOLIC_ACCESS',`
	MARK(finish_$1)
	! call quickXXXaccess(opcode, pc, fb, ee)
	mov	PC,%o1
	call	$1,4
	mov	Ee,%o3
    	
	ldub	[PC],Opcode	! refetch current PC
	tst	%o0		! -1=error, 0=rewritten, 1=redo
	blt	handleException
	SCALE(Opcode)
	GO
	ldub	[PC+1],NextByte	! do not increment PC, but DO refetch NextByte
')
catName(`finish',`getstatic'):
catName(`finish',`putstatic'):
FINISH_SYMBOLIC_ACCESS( quickStaticAccess )

catName(`finish',`getfield'):
catName(`finish',`putfield'):
FINISH_SYMBOLIC_ACCESS( quickFieldAccess )

define(`FINISH_INVOKERS_A', `
catName(`finish',$1):	
MARK(`finish_'$1)
	/*   %o0 contains the current byte code
	 *   %o2 contains the extracted method
	 */
        lduh	[%o2+STRUCT_OFFSET(methodblock,args_size)],R3 !argsize
	mov	%o2, TOSB	! save this value, which contains method
	sll	R3,2,T2	! bytes used by arguments
 ifelse($1, `invokestatic',
       `! set TOSA to be classOf(method),
	! set R4 to be new stack pointer
	ld	[%o2+STRUCT_OFFSET(methodblock,fb.clazz)],TOSA
	sub	SP,T2,R4	! new stack pointer
	mov	PC, %o1',
       `! set TOSA to be "this" of method call.  Make sure non null
	! set R4 to be new stack pointer
	sub	SP,T2,R4	! new stack pointer
	ld	[R4],TOSA	! "this" of method call
	tst	TOSA		! is "this" null?
    	beq	nullPointerException
	mov	PC, %o1
       ')
	! quickInvocation(opcode, pc, mb, ee)
	call	quickInvocation,4
	mov	Ee,%o3		

    	ldub	[PC],Opcode	! refetch current PC
	tst	%o0		! -1=error, 0=rewritten, 1=redo
	blt	handleException
	nop
	bgt	redispatch_invoker ! need to redispatchg
	mov	TOSA, %o0	! put "this" where it belongs
	mov	TOSB, %o1	! put methodblock where it belongs
	mov	R3, %o2	! put argsize where it belongs
')

FINISH_INVOKERS_A(`invokestatic')
	! if the "current method" is NULL, or if we were rewritten into
	! something other than an opc_invokestatic_quick, then handle
	! this by re-executing it
	! %o0 = handle to class of current method
	! %o1 = current method block
	! %o2 = argsize of current method
	! Opcode = rewritten opcode
	! R4 = new stack pointer
	cmp	Opcode, DEFINED_CONSTANT(opc_invokestatic)
	bne	redispatch_invoker
	mov	0,T1
	ba	callmethod
	mov	R4,SP		! new stack pointer

FINISH_INVOKERS_A(`invokespecial')
        ! The opcode should either have been rewritten to 
	!   opc_invokenonvirtual_quick or opc_invokesuper_quick
        ! Anything else implies there is something strange going on, and
	! we should just re-execute the instruction.
	! %o0 = "this" of method call
	! %o1 = current method block
	! %o2 = argsize of current method
	! Opcode = rewritten opcode
	! R4 = new stack pointer
	cmp	Opcode, DEFINED_CONSTANT(opc_invokespecial)
	bne	redispatch_invoker
	nop
	
	mov	FP,%o0
	call	quickSelectSuperMethod
	mov	TOSB,%o1

	mov	%o0,%o1		! pass the returned method block to callmethod
	mov	TOSA,%o0	! setup "this" for call method 
	mov	R3,%o2		! setup args_size for call method 

	mov	0,T1		! set returnPC to 0
	
	ba	callmethod
	mov	R4,SP		! new stack pointer

FINISH_INVOKERS_A(`invokevirtual')
        ! The opcode should either have been rewritten to 
	!    opc_invokevirtual_quick, opc_invokevirtualobjectquick, 
        !    or opc_invokevirtual_quick_w
        ! Anything else implies there is something strange going on, and
	! we should just re-execute the instruction.  We also don't bother
	! optimizing opc_invokevirtual_quick_w, because we already know that
	! none of them are final and can be inlined.
	! %o0 = "this" of method call
	! %o1 = current method block
	! %o2 = argsize of current method
	! Opcode = rewritten opcode
	! R4 = new stack pointer
	cmp	Opcode, DEFINED_CONSTANT(opc_invokevirtual)
	bne	redispatch_invoker
	mov	0,T1		! set returnPC to 0
	
	! Get the dynamic methodblock, which is 
	!    mt_slot(obj_methodtable(%o0), %o1->fb.u.offset);
  	ld	[%o1+STRUCT_OFFSET(methodblock,fb.u.offset)],T2
	tst	T2
	bz	1f
	nop
	ld	[%o0+STRUCT_OFFSET(Hjava_lang_Object,methods)],%o1
	sll	T2,2,T2
	inc	STRUCT_OFFSET(methodtable,methods),T2 
	ld	[%o1+T2],%o1	
1:
        ! set up for callmethod
	ba	callmethod
	mov	R4,SP		! new stack pointer ;
	
redispatch_invoker:	
	SCALE(Opcode)
	GO
	ldub	[PC+1],NextByte	! do not increment PC, but DO refetch NextByte

	INLINE
	
RESOLVE_SYMBOLIC_W(invokeinterface, `1 << DEFINED_CONSTANT(CONSTANT_InterfaceMethodref)', opc_invokeinterface_quick)

ILLEGALOP(xxxunusedxxx)

/*
 * NEW. Call ResolveClassConstant.
 * Rummage in data structures.
 * Bail on a variety of errors.
 * Rewrite new_quick opcode in place.
 * Leaves stack alone.
 */
OP0(new)
    	! ResolveClassConstant(pool, index, ee, bits) 
	ldub	[PC+2],TOSB
	sll	NextByte,8,NextByte
	mov	ConstP,%o0
	mov	Ee,%o2
	mov	1<<DEFINED_CONSTANT(CONSTANT_Class),%o3
	call	ResolveClassConstant,4
	or	NextByte,TOSB,%o1

	tst	%o0
	bnz	finish_new
	    or	NextByte,TOSB,TOSB	! form scaled index in TOSB
	b	complete_new
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T2
	OUTLINE
finish_new:
	sll	TOSB,2,TOSB	! offset for constant pool
	ld	[ConstP+TOSB],TOSA	! Const pool entry in TOSA
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand ! class access
	lduh	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,access)],R3 ! class access
	andcc	R3,(DEFINED_CONSTANT(ACC_INTERFACE)|DEFINED_CONSTANT(ACC_ABSTRACT)),%g0
	bnz	instantiationError
	mov	TOSA,%o0

	! call InitClass(cb)
	call	InitClass,1
	nop
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T2
complete_new:
	mov	DEFINED_CONSTANT(opc_new_quick),T1
	tst	T2
	bgt	handleException
	nop	/* Only update the opcode if no exception has occurred. */
	stb	T1,[PC]
	b	catName(new_quick,0)
	ldub	[PC+1],NextByte	! waste of bandwidth!
	INLINE
	
/*
 * NEWARRAY --
 * Take length off the stack.
 * Take typecode from NextByte.
 * Allocate. Set to zero. Push handle on stack.
 * Only for arrays of scalars -- no objects here!!
 */
OP0(newarray)
	ld	[SP-4],TOSB		! TOSB gets count
	addcc	TOSB,0,%o2
	blt	negativeArraySize
	mov	NextByte,%o1
	mov	Ee, %o0
	st	PC,[FP+STRUCT_OFFSET(javaframe,lastpc)]
	call	allocArray,3
	and	NextByte,3,NextByte	! NextByte gets size bits from type.

	tst	%o0			! test resulting handle.
	bne	finish_newarray
	st	%o0,[SP-4]		! TOS gets handle
	ba	outOfMemory
	nop
	OUTLINE
finish_newarray:
/* [Sheng 11/08/96] memset has been moved into allocArray.
	sll	TOSB,NextByte,%o2	! %o2 gets byte size of array data
	mov	0,%o1
	ld	[%o0+STRUCT_OFFSET(Hjava_lang_Object,obj)],%o0 ! SHOULD Unhand!
	call	memset,3
*/
	ldub	[PC+2],NextByte
	SCALE(NextByte)
	ldub	[PC+3],NextByte
	GO
	inc	2,PC
	INLINE

RESOLVE_SYMBOLIC_W(anewarray, (1<<DEFINED_CONSTANT(CONSTANT_Class)), opc_anewarray_quick )

/* ARRAYLENGTH --
 * Find array handle on the top of stack.
 * Replace it with length (of major dimension).
 */
OP0(arraylength)
	ld	[SP-4],TOSA
	tst	TOSA
	be	nullPointerException
	SCALENEXT
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,methods)],TOSA
	inc	1,PC
	ldub	[PC+1],NextByte
	srl	TOSA,DEFINED_CONSTANT(METHOD_FLAG_BITS),TOSA
	GO
	st	TOSA,[SP-4]
	
/*
 * ATHROW --
 * Take object off top of stack and throw it.
 * If TOS is null, then throw NullPointerException instead.
 * Otherwise, the handle on TOS is stuck in ee->exception, and the
 * constant value EXECKIND_THROW is stuffed into ee->exceptionKind
 * Effectively empties expression stack.
 */
OP0(athrow)
	ld	[SP-4],TOSA
	dec	4,SP
	mov	DEFINED_CONSTANT(EXCKIND_THROW),T2
	tst	TOSA
	be	nullPointerException
	st	TOSA,[Ee+STRUCT_OFFSET(execenv,exception)]
	b	handleException
	stb	T2,[Ee+STRUCT_OFFSET(execenv,exceptionKind)]

RESOLVE_SYMBOLIC_W(checkcast, (1<<DEFINED_CONSTANT(CONSTANT_Class)), opc_checkcast_quick )
RESOLVE_SYMBOLIC_W(instanceof, (1<<DEFINED_CONSTANT(CONSTANT_Class)), opc_instanceof_quick )

/*
 * Monitorenter and monitorexit are similar.
 * Object handle is popped off the stack and passed as a parameter,
 * though a null handle garners a NullPointerException.
 * The only difference is that monitorExit checks for 
 * IllegalMonitorStateException
 */

OP0(monitorenter)
	ld	[SP-4],%o1
	tst	%o1
	bz	nullPointerException
	dec	4,SP
	SCALENEXT
	ldub	[PC+2],NextByte 
	inc	PC
	mov	Ee, %o0
	CALL_AND_GO(monitorEnter2,2)

OP0(monitorexit)
	ld	[SP-4],%o1
	tst	%o1
	bz	nullPointerException
	dec	4,SP
        call    monitorExit2,2
	mov	Ee, %o0
	tst	%o0
	bnz	handleException
	SCALENEXT
	ldub	[PC+2],NextByte 
	GO
	inc	PC	

/*
 * OPC_WIDE --
 * This never happens!
 * Fetch pc[2] and pc[3],
 * re-dispatch on opcode at pc[1], already in NextByte
 * As the set of legal subsidiary opcodes is sparse, we
 * search with a lookup table. ( I considered hand-coding a
 * binary search, which would be faster, but as this never happens,
 * decided it not to be worth the trouble.)
 */ 
OP0(wide)
	sll	NextByte,1,NextByte ! index into table
	ldub	[PC+2],T2	! high byte of register
	SET_ADDRESS(widetab,T3)	
	ldsh	[T3+NextByte],T3 ! offset from TableBase
	sll	T2,10,T2	
	ldub	[PC+3],T1	! low byte of register
	add	Varbase,T2,T2	
	sll	T1,2,T1			
	jmp	Dispatch+T3
	ldub	[PC+4],Opcode
	OUTLINE
	/*
	 * Upon entry to the specific opcode interpreters we have:
	 *   [T1 + T2] == vars[GET_INDEX(pc[1])]
	 *   Opcode is pc[4]
	 */

wide_aload: 
wide_fload:	
wide_iload: 
	ld	[T2+T1],TOSA	! get variable
	SCALE(Opcode)
	ldub	[PC+5],NextByte
	inc	4,PC
	st	TOSA,[SP]	! store it
	GO
	inc	4,SP

wide_lload:  
wide_dload:    	
	ld	[T2+T1],TOSA	! get first word
	inc	4,T1
	ld	[T2+T1],TOSB	! get second word
	inc	8,SP
	st	TOSA,[SP-8]
	SCALE(Opcode)
	st	TOSB,[SP-4]
	ldub	[PC+5],NextByte
	GO
	inc	4,PC

wide_astore: 
wide_fstore: 
wide_istore:   
	ld	[SP-4],TOSA	! get top of stack
	SCALE(Opcode)
	ldub	[PC+5],NextByte
	dec	4,SP
	st	TOSA,[T2+T1]
	GO
	inc	4,PC

wide_lstore: 
wide_dstore:	
	ld	[SP-8],TOSB
	SCALE(Opcode)
	ld	[SP-4],TOSA
	ldub	[PC+5],NextByte
	dec	8,SP
	st	TOSB,[T2+T1]
	inc	4,T1
	st	TOSA,[T2+T1]
	GO
	inc	4,PC
	
	! wide iinc. Takes signed, 16-bit increment argument
	! leaves stack alone.
wide_iinc:
	ldub	[PC+5],TOSA
	sll	Opcode,24,TOSB	! oops -- did unsigned load.
	ld	[T2+T1],R3
	sra	TOSB,16,TOSB		! so sign extend the mess
	ldub	[PC+6],Opcode
	add	TOSA,TOSB,TOSA
	ldub	[PC+7],NextByte
	add	R3,TOSA,R3
	SCALE(Opcode)
	st	R3,[T2+T1]
	GO
	inc	6,PC

	! wide return. Opcode is not required, so was fetched unnecessarily. 
	! leaves stack alone.
wide_ret:
	ld	[T2+T1],PC	! fetch
	b	finish_branch.0
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	INLINE
	

RESOLVE_SYMBOLIC_W(multianewarray, (1<<DEFINED_CONSTANT(CONSTANT_Class)), opc_multianewarray_quick )

IFBRANCH(ifnull, be)
IFBRANCH(ifnonnull, bne)
	
/*
 * GOTO_W never really happens, but ...
 */
OP0(goto_w)
	ldub	[PC+2],T2
	sll	NextByte,24,NextByte
	ldub	[PC+3],T1
	add	PC,NextByte,T3
	ldub	[PC+4],R3
	sll	T2,16,T2
	add	T3,T2,PC
	sll	T1,8,T1
	add	PC,T1,PC
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	b	finish_branch.0
	add	PC,R3,PC
	
/*
 * JSR_W -- push value PC+5 on the stack, then continue like goto_w.
 */
OP0(jsr_w)
	ldub	[PC+2],T2
	sll	NextByte,24,NextByte
	add	PC,5,TOSA
	ldub	[PC+3],T1
	add	PC,NextByte,T3
	ldub	[PC+4],R3
	sll	T2,16,T2
	add	T3,T2,PC
	sll	T1,8,T1
	add	PC,T1,PC
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T1
	add	PC,R3,PC
	st	TOSA,[SP]
	b	finish_branch.0
	inc	4,SP

/*
 * BREAKPOINT -- treat as a nop?
 */
OP0(breakpoint)
	SCALENEXT
	ldub	[PC+2],NextByte
	GO
	inc	PC
	
/*
 * LDC_QUICK -- NextByte is a constant-pool index.
 * Pick up indicated word, put on stack.
 */
OP0(ldc_quick)
	sll	NextByte,2,T2	! scale for indexing
	ldub	[PC+2],Opcode	! fetch next instruction
	ld	[ConstP+T2],TOSA	! fetch constant-pool entry
	SCALE(Opcode)
	ldub	[PC+3],NextByte
	inc	2,PC
	st	TOSA,[SP]
	GO
	inc	4,SP
	
	
/*
 * LDC_W_QUICK -- next 2 bytes form constant-pool index.
 * push pool word on stack.
 */
OP0(ldc_w_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,T1
	ldub	[PC+3],Opcode
	add	ConstP,T1,T1
	ldub	[PC+4],NextByte
	sll	T2,2,T2
	ld	[T1+T2],TOSA
	SCALE(Opcode)
	inc	3,PC
	st	TOSA,[SP]
	GO
	inc	4,SP
	
	
/*
 * LDC2_W_QUICK -- next 2 bytes form constant-pool index.
 * push 2 words from pool on stack.
 */
OP0(ldc2_w_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,T1
	ldub	[PC+3],Opcode
	add	ConstP,T1,T1
	sll	T2,2,T2
	ld	[T1+T2],TOSA
	add	T2,T1,T2
	ld	[T2+4],TOSB
	ldub	[PC+4],NextByte
	inc	3,PC
	st	TOSA,[SP]
	SCALE(Opcode)
	st	TOSB,[SP+4]
	GO
	inc	8,SP
	
	
/*
 * GETFIELD_QUICK takes a 1-byte index followed
 * by one byte to ignore. Uses object handle on stack top.
 * Checks for null handle, of course, and can cause NullPointerException.
 * Fetch one word to stack top.
 */
OP0(getfield_quick)
	ld	[SP-4],TOSA	! object
	sll	NextByte,2,T2	! scale for indexing.
	tst	TOSA
	be	nullPointerException
	ldub	[PC+3],Opcode	! fetchahead
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand
	SCALE(Opcode)
	ldub	[PC+4],NextByte
	ld	[Unhand+T2],TOSA
	inc	3,PC
	GO
	st	TOSA,[SP-4]
	
	
/*
 * PUTFIELD_QUICK takes a 1-byte index followed by
 * one byte to ignore. Uses value on stack top and
 * object handle beneath it.
 * Checks for null handle, of course, and can cause NullPointerException.
 */
OP0(putfield_quick)
	ld	[SP-8],TOSA	! handle
	sll	NextByte,2,T2
	ld	[SP-4],TOSB	! value
	dec	8,SP
	tst	TOSA
	bz	nullPointerException
	ldub	[PC+3],Opcode	! fetch following opcode.
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand
	SCALE(Opcode)
	ldub	[PC+4],NextByte	! fetchahead
	st	TOSB,[Unhand+T2] ! Store TOS word into object.
	GO
	inc	3,PC
	
	
/*
 * GETFIELD2_QUICK -- like getfield_quick, but fetches
 * two words to stack.
 * Takes a handle from the stack, which, of course, is checked
 * for null.
 */
OP0(getfield2_quick)
	ld	[SP-4],TOSA
	ldub	[PC+3],Opcode	! fetchahead
	tst	TOSA
	bz	nullPointerException
	sll	NextByte,2,T2	! Scale for indexing
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand
	add	T2,4,T1		! indexing of 2nd word
	ld	[Unhand+T2],TOSA
	SCALE(Opcode)
	ld	[Unhand+T1],TOSB
	ldub	[PC+4],NextByte	! fetchahead
	inc	4,SP
	st	TOSA,[SP-8]
	st	TOSB,[SP-4]
	GO
	inc	3,PC
	
	
/*
 * PUTFIELD2_QUICK -- like putfield_quick, but stores
 * two words from stack. Handle is on stack, beneath values.
 * Checks handle for null, of course.
 */
OP0(putfield2_quick)
	ld	[SP-12],R3	! fetch handle
	sll	NextByte,2,T2	! scale for indexing
	ld	[SP-8],TOSA	! fetch first data word
	add	T2,4,T1		! 2nd word index
	ld	[SP-4],TOSB	! fetch second data word
	tst	R3
	ldub	[PC+3],Opcode	! fetchahead
	bz	nullPointerException
	dec	12,SP
	ld	[R3+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand
	ldub	[PC+4],NextByte	
	SCALE(Opcode)
	st	TOSA,[Unhand+T2]
	inc	3,PC
	GO
	st	TOSB,[Unhand+T1]
	
	
/*
 * GETSTATIC_QUICK
 * The following 2 bytes form a constant pool index.
 * The indexed item is a fieldblock pointer. That fieldblock
 * contains a value in u.static_value. Fetch it to the stack.
 */
OP0(getstatic_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,NextByte
	ldub	[PC+3],Opcode	! fetchahead
	add	NextByte,ConstP,T1
	sll	T2,2,T2		! scale for addressing
	ld	[T1+T2],T1	! fieldblock pointer
	SCALE(Opcode)
	ldub	[PC+4],NextByte	! fetchahead
	inc	3,PC
	ld	[T1+STRUCT_OFFSET(fieldblock,u.static_value)],TOSA
	inc	4,SP
	GO
	st	TOSA,[SP-4]
	
/*
 * PUTSTATIC_QUICK
 * The following 2 bytes form a constant pool index.
 * The indexed item is a fieldblock pointer. That fieldblock
 * contains a field u.static_value. Put TOS value there.
 */
OP0(putstatic_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,NextByte
	ld	[SP-4],TOSA
	dec	4,SP
	ldub	[PC+3],Opcode	! fetchahead
	add	NextByte,ConstP,T1
	sll	T2,2,T2		! scale for addressing
	ld	[T1+T2],T1	! fieldblock pointer
	ldub	[PC+4],NextByte	! fetchahead
	st	TOSA,[T1+STRUCT_OFFSET(fieldblock,u.static_value)]
	SCALE(Opcode)
	GO
	inc	3,PC
	
/*
 * GETSTATIC2_QUICK -- like getstatic_quick but
 * transfers 2 words to the stack. Requires an extra
 * indirection, too, as the fieldblock contains, in this case,
 * not the 2-word datum, but a pointer to it. Curses.
 */
OP0(getstatic2_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,NextByte
	ldub	[PC+3],Opcode	! fetchahead
	add	NextByte,ConstP,T1
	sll	T2,2,T2		! scale for addressing
	ld	[T1+T2],T1		! fieldblock pointer
	SCALE(Opcode)
	ld	[T1+STRUCT_OFFSET(fieldblock,u.static_address)],TOSA
	ldub	[PC+4],NextByte	! fetchahead
	! have address, get data.
	ld	[TOSA+4],TOSB
	inc	3,PC
	ld	[TOSA],TOSA
	inc	8,SP
	st	TOSB,[SP-4]
	GO
	st	TOSA,[SP-8]

/*
 * PUTSTATIC2_QUICK -- like putstatic_quick but
 * transfers 2 words from the stack. Requires an extra
 * indirection, too, as the fieldblock contains, in this case,
 * not the 2-word datum, but a pointer to it. Curses.
 */
OP0(putstatic2_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,NextByte
	ldub	[PC+3],Opcode	! fetchahead
	add	NextByte,ConstP,T1
	ld	[SP-8],TOSA
	sll	T2,2,T2		! scale for addressing
	ld	[T1+T2],T1		! fieldblock pointer
	dec	8,SP
	ld	[SP+4],TOSB
	ld	[T1+STRUCT_OFFSET(fieldblock,u.static_address)],T1
	ldub	[PC+4],NextByte	! fetchahead
	SCALE(Opcode)
	! have address, put data.
	st	TOSA,[T1]
	st	TOSB,[T1+4]
	GO
	inc	3,PC

/*
 * START_SHORT_INVOCATION
 * Common code for invokevirtual_quick and
 * invokevirtualobject_quick. These differ from the
 * other quick invocations in that they use pc[2]
 * as args_size, rather than taking it out of the
 * methodblock. This macro computes
 *	%o0 <- object handle. != 0
 *      %o1 <- method table	 
 *	%o2 <- args_size
 *	NextByte-- offset into methodtable
 *	SP	-- pruned back by args_size
 * Then branches elsewhere to finish instruction.
 * Macro arguments are name and branch destination
 */
define( `START_SHORT_INVOCATION',
    `OP0($1)
	ldub	[PC+2],%o2	! get args_size
	sll	%o2,2,R3	! scale args_size for stack adjustment
	sub	SP,R3,SP	! prune stack
	ld	[SP],%o0	! get object handle
	sll	NextByte,2,NextByte	! scale for indexing
	inc	STRUCT_OFFSET(methodtable,methods),NextByte
	tst	%o0
	bne,a	$2
    	  ld	[%o0+STRUCT_OFFSET(Hjava_lang_Object,methods)],%o1
	ba	nullPointerException
	nop ')

	/*
	 * INVOKEVIRTUAL_QUICK
	 */

START_SHORT_INVOCATION(invokevirtual_quick, finish_invokevirtual_quick)
	OUTLINE
finish_invokevirtual_quick:
	/*	%o0	-- object handle. != 0
	 *      %o1     -- method table
    	 *	%o2	-- args_size
	 *	NextByte-- offset into methodtable
	 *	SP	-- pruned back by args_size, this size
	 *	PC	-- points to this instruction
	 */
	ld	[%o1+NextByte],%o1	! methodtable slot for this method.
	add	PC,3,T1
	!
	! fall through
	!
	
callmethod:
	/*
	 * CALLMETHOD.
	 *	%o0	-- object handle. != 0
	 *	%o1	-- methodblock pointer
	 *	%o2	-- args_size
	 *	T1	-- return PC
	 *	PC	-- points to the invoke instruction
	 */	
	!! DANGER
        !!   If you change the following instruction, find all branches to callmethod_1 
        !!   and change the instruction in the delay slot!
	!! REGNAD
	ld	[%o1+STRUCT_OFFSET(methodblock,invoker)],T3
callmethod_1:   	
	.global callmethod_1
	.type callmethod_1,2
	sethi	%hi(invokeJavaMethod),%o3
	st	T1,[FP+STRUCT_OFFSET(javaframe,returnpc)]
	add	%o3,%lo(invokeJavaMethod),%o3
	st	SP,[FP+STRUCT_OFFSET(javaframe,optop)]
	sethi	%hi(invokeSynchronizedJavaMethod),T1
	cmp	T3,%o3
	bne	notSimpleJavaInvocation
	st	PC,[FP+STRUCT_OFFSET(javaframe,lastpc)]

	! Here for a simple, non-synchronized Java method invocation.
	! indicate non-synchronized by setting R3 to 0
	mov	0,R3
	! FALLTHROUGH

	/*
	 * Do Java invocation ourselves.
	 * Set up new java frame, checking to see if we must allocate 
	 * a new stack chunk.
	 * If R3 != null, then %o0 is something we have to
	 * synchronize on.
	 * This code does not do stack allocation -- if it is necessary,
	 * the (T2-implemented) invoker is called. Fortunately, this 
	 * never actually happens.
	 */
doJavaInvocation:
	lduh	[%o1+STRUCT_OFFSET(methodblock,nlocals)],TOSA
	mov	SP,Varbase		! old SP is proposed new Varbase
	ld	[FP+STRUCT_OFFSET(javaframe,javastack)],T3 ! stack chunk
	lduh	[%o1+STRUCT_OFFSET(methodblock,maxstack)],T1
	sll	TOSA,2,TOSA 		! turn n-local-words into byte count
	ld	[%o1+STRUCT_OFFSET(methodblock,code)],PC ! new pc
	add	SP,TOSA,TOSA	  	! proposed new frame pointer
	ld	[T3+STRUCT_OFFSET(javastack,end_data)],%o3
	sll	T1,2,T1			! maxstack from words to bytes
	add	TOSA,STRUCT_OFFSET(javaframe,ostack),SP ! proposed new SP
	add	SP,T1,TOSB		! highest stack address needed
	cmp	TOSB,%o3
	bge,a	notJavaInvocation	! oops -- may have to go do work.
	    ld	[%o1+STRUCT_OFFSET(methodblock,invoker)],T3 ! ( this never happens )

	/*
	 * Now have:
	 *	o0 -- object to synchronize on
	 *	o1 -- new methodblock pointer
	 * 	FP -- previous java frame.
	 *	T3 -- containing javastack.
	 *	Varbase -- new Varbase
	 *	SP	-- new SP
	 *	TOSA	-- new java frame.
	 * 	R3  -- 1 or 0, for synchronized or not.
	 * Fill in current frame info,
	 * synchronize on object, if necessary,
	 * and go
	 */
fillNewJavaFrame:
	ld	[%o1+STRUCT_OFFSET(methodblock,fb.clazz)],%o2 ! class of new method
	st	FP,[TOSA+STRUCT_OFFSET(javaframe,prev)]
	mov	TOSA,FP		! FP gets new FP
	st	Varbase,[TOSA+STRUCT_OFFSET(javaframe,vars)]
	st	T3,[FP+STRUCT_OFFSET(javaframe,javastack)]
	ld	[%o2+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand
	st	%o1,[FP+STRUCT_OFFSET(javaframe,current_method)]
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,constantpool)],ConstP

	sethi	%hi(java_profiler_isOn),TOSB
	ld	[TOSB+%lo(java_profiler_isOn)],TOSB
	
	sethi	%hi(jvmpi_event_flags),R5
	st	FP,[Ee+STRUCT_OFFSET(execenv,current_frame)]
	ld	[R5+%lo(jvmpi_event_flags)],R5
	ldub	[PC],NextByte
	and	R5,DEFINED_CONSTANT(JVMPI_EVENT_METHOD_ENTRY_ON),R5
	tst	R3			! see if we need to synchronize
	bne	synchronizedEntry
	st	ConstP,[FP+STRUCT_OFFSET(javaframe,constant_pool)]
	tst	R5
	st	%g0,[FP+STRUCT_OFFSET(javaframe,monitor)]
	bne	jvmpiMethodEntry
	nop
	tst	TOSB	! i.R3. java_profiler_isOn
	bne	javaProfilerEntry
	SCALENEXT	! anticipate fallthrough
	GO
	ldub	[PC+1],NextByte

jvmpiMethodEntry:	! JVMPI method entry notification
	mov	Ee, %o0
	! calling jvmpi_method_entry(ee, this)
	call	jvmpi_method_entry,2
	ld	[Varbase], %o1

	SCALENEXT	! re-do clobbered shift
	GO
	ldub	[PC+1],NextByte
	

javaProfilerEntry:	! do java profiling
	call	_sysThreadCPUTime,0
	nop
	
	SCALENEXT	! re-do clobbered shift
	ldub	[PC+1],NextByte
	GO
	st	%o0,[FP+STRUCT_OFFSET(javaframe,profiler_info)]
	
synchronizedEntry:
	! %o0 is an object handle we have to synchronize on.
	! other o-registers all junk.
	mov	%o0, %o1
	call	monitorEnter2,2
	mov	Ee, %o0
	st	%o0,[FP+STRUCT_OFFSET(javaframe,monitor)]

	tst	R5	! flag for JVMPI method entry notification
	bne	jvmpiMethodEntry
	nop
	tst	TOSB	! i.R3. java_profiler_isOn
	bne	javaProfilerEntry
	SCALENEXT	! anticipate fallthrough
	GO
	ldub	[PC+1],NextByte

notSimpleJavaInvocation:
	add	T1,%lo(invokeSynchronizedJavaMethod),T1
	cmp	T3,T1
	be,a	doJavaInvocation
	    mov	1,R3
notJavaInvocation:
	/*
	 * We are here because:
	 * (a) we are using an un-recognized invoker, or
	 * (b) we need to allocate another stack chunk, and
	 *     are too lazy to do so in this code.
	 * We have:
	 *	o0 -- object
	 *	o1 -- new methodblock pointer
	 *	o2 -- args_size ( in words )
	 *	FP  -- previous java frame
	 *	T3  -- invoker to call
	 *	Varbase -- old SP, (i.R3. pointer to args we need to move.)
	 */
	call	T3,4		! call invoker
	mov	Ee,%o3

    	! Normally, we continue at FP.returnpc, unless either
        !  . there is an exception, in which case we use the PC of the all
        !  . FP.returnPC == 0, in which case it is a special hack indicating
    	!       this is the first executing of this invoker, and we want to
        !       possibly rewrite it again.
	ld	[Ee+STRUCT_OFFSET(execenv,current_frame)],FP ! get Java FP
	mov	%o0, TOSA	! save value of result
	ld      [FP+STRUCT_OFFSET(javaframe,returnpc)],PC ! get return pc
	tst	PC
	beq     fixupFrame
	.empty
  
	
returnFixupFrame:   
	.global returnFixupFrame
	.type returnFixupFrame,2
	
	! if result == 0, then reset pc to be frame->lastpc and go check for
	! exceptions
        tst     TOSA
  	bz,a	checkException
 	ld	 [FP+STRUCT_OFFSET(javaframe,lastpc)],PC

	!	Varbase = Ee->current_frame->vars
	ld	[FP+STRUCT_OFFSET(javaframe,vars)], Varbase
	!	Opcode = PC[0]
	ldub	[PC], Opcode		! fetch first bytecode
	! 	SP = Ee->current_frame->optop
	ld	[FP+STRUCT_OFFSET(javaframe,optop)], SP
	SCALE(Opcode)
	!	ConstP = Ee->current_frame->constant_pool
	ldub	[PC+1],NextByte		! NextByte = PC[1]
	GO
	ld	[FP+STRUCT_OFFSET(javaframe,constant_pool)],ConstP
    
fixupFrame: 
	.global fixupFrame
	.type fixupFrame,2
    	! FP.returnPC == 0
    	! We need to recreate frame->lastpc
    	! Also, call the method FixupQuickInvocation() to possibly inline
    	! the method call    	
        ! lastpc = frame->lastpc; 
	ld	[FP+STRUCT_OFFSET(javaframe,lastpc)],%o1
    	! pc = lastpc + 3	
	add	%o1,3,PC	! when result != 0
	
	mov	FP,%o1
	mov	Ee,%o0
	CALL_AND_BRANCH(FixupQuickInvocation,2,returnFixupFrame)

	INLINE

	/*
	 * START_INVOCATION.
	 * Common prefix for 3 invoke*quick instructions.
	 * Computes
	 *    %o1 <- constant_pool[GET_INDEX(pc+1)].p
	 *    %o2 <- %o1->args_size
	 *    T1   <- PC+3
	 * Then branches elsewhere to finish instruction.
	 * Macro arguments are name and branch destination
	 */
define(`START_INVOCATION',`
    OP0($1)
	ldub	[PC+2],T1
	sll	NextByte,10,NextByte
	add	NextByte,ConstP,T2
	sll	T1,2,T1
	ld	[T2+T1],%o1	! methodblock pointer
	add	PC,3,T1
	lduh	[%o1+STRUCT_OFFSET(methodblock,args_size)],%o2 ')

	/*
	 * INVOKENONVIRTUAL_QUICK.
	 */
START_INVOCATION(invokenonvirtual_quick)
	/*
	 * 	%o1 -- methodblock pointer
	 *	%o2 -- args_size from the methodblock
	 *	T1   -- PC+3
	 */
	sll	%o2,2,R3	! 
	sub	SP,R3,SP	! trim stack
	ld	[SP],%o0	! get object handle
	tst	%o0
	bnz,a	callmethod_1
	  ld	[%o1+STRUCT_OFFSET(methodblock,invoker)],T3
	b	nullPointerException
	nop
	

/*
 * INVOKESUPER_QUICK
 * A special case of invokenonvirtual_quick, which uses a unique
 * name resolution scheme. Frank will tell you why.
 * Prunes the stack by the number of words specified by the
 * methodblock of the called method.
 */
OP0(invokesuper_quick)
	! currentClass = fieldclass(&frame->current_method->fb);
	! superClass = cbSuperclass(currentClass);
    	! offset = GET_INDEX(pc + 1);
    	! mb = cbMethodTable(superClass)->methods[offset];
	
	ldub	[PC+2],T1
	sll	NextByte,10,NextByte
	ld	[FP+STRUCT_OFFSET(javaframe,current_method)],R3
	sll	T1,2,T1
	add	NextByte,T1,NextByte	! scaled index into method table
	ld	[R3+STRUCT_OFFSET(methodblock,fb.clazz)],R3 ! my Class pointer
	ld	[R3+STRUCT_OFFSET(Hjava_lang_Class,obj)], Unhand ! my class data
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class, superclass)], R3
	add	PC,3,T1
	ld	[R3+STRUCT_OFFSET(Hjava_lang_Class,obj)], Unhand ! my superclass data
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,methodtable)],R3
	inc	STRUCT_OFFSET(methodtable,methods),NextByte
	ld	[R3+NextByte],%o1 ! super->methodtable->methods[index]

	! now have methodblock pointer in %o1,
	! and return PC in T1,
	! and frame pointer in T2
	lduh	[%o1+STRUCT_OFFSET(methodblock,args_size)],%o2
	! args_size in %o2
	b	finish_super_quick
	sll	%o2,2,R3		! scale args_size for stack adjustment
	OUTLINE
finish_super_quick:
	/*
	 * %o1 = unhand(ee->current_frame->current_method->fb.clazz->superclass)
         *            ->methodtable->methods[GET_INDEX(pc+1)]
	 * %o2 has %o1->args_size
	 * T1 has return PC
	 * R3 has args_size << 2
	 * Pop the appropriate number of arguments, then call.
	 */
	sub	SP,R3,SP		! prune stack
	ld	[SP],%o0	! get object handle to %o0
	tst	%o0
	bne,a	callmethod_1
	  ld	[%o1+STRUCT_OFFSET(methodblock,invoker)],T3	
	b,a	nullPointerException
	nop
	/*
	 * for callmethod, need:
	 *	%o0	-- object handle. != 0
	 *	%o1	-- methodblock pointer
	 *	%o2	-- args_size
	 *	T1	-- return PC
	 *	PC	-- points to the invoke instruction
	 */
	INLINE
	
/*
 * INVOKESTATIC_QUICK
 */
START_INVOCATION(invokestatic_quick)
	/*
	 * 	%o1 -- methodblock pointer
	 *	%o2 -- args_size from the methodblock
	 *	T1   -- PC+3
	 *	Stack -- not trimmed.
	 */
	ld	[%o1+STRUCT_OFFSET(methodblock,fb.clazz)],%o0
	sll	%o2,2,R3
	ld	[Ee+STRUCT_OFFSET(execenv,current_frame)],T2 ! get Java FP
	b	callmethod
	sub	SP,R3,SP

/*
 * INVOKEINTERFACE_QUICK
 * In addition to the usual index into the
 * constant pool, the instruction contains a guess
 * as to the last slot used for this call. Good luck.
 */
OP0(invokeinterface_quick)
	ldub	[PC+3],%o2 ! args_size
	sll	NextByte,10,NextByte
	ldub	[PC+2],T2
	sll	%o2,2,R3
	sub	SP,R3,SP		! Trim stack
	ld	[SP],%o0
	add	NextByte,ConstP,T1
	sll	T2,2,T2
	tst	%o0
	bnz	finish_invokeinterface_quick
	ldub	[PC+4],R3
	b,a	nullPointerException
	nop
	
	OUTLINE
finish_invokeinterface_quick:
	/*
	 * %o0 -- this object, != 0
	 * %o2 -- args_size
	 * T2   -- PC[2] << 2
	 * T1   -- ConstP + ( NextByte<<10)
	 * R3   -- PC[4], imethods guess
	 * The stack has already been trimmed
	 */
	! set TOSA = o->methods.  See if this is a reference or an array.	    
	ld	[%o0+STRUCT_OFFSET(Hjava_lang_Object,methods)],TOSA 
	andcc	TOSA,DEFINED_CONSTANT(FLAG_MASK),%g0
	bz	1f
	ld	[T2+T1],R4	! R4 = constant_pool[index].
	    ! This is an array. This should never actually happen,
	    ! since arrays do not support any interfaces (yet).
	    ! 	TOSB = classJavaLangObject;
	    !   TOSA = cbMethodTable(TOSB);
	    sethi  %hi(classJavaLangObject),TOSB
	    ld	   [TOSB+%lo(classJavaLangObject)],TOSB
	    ld	   [TOSB+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand
	    b	   2f
	    ld	   [Unhand+STRUCT_OFFSET(Classjava_lang_Class,methodtable)],TOSA
    
1:	ld	[TOSA+STRUCT_OFFSET(methodtable,classdescriptor)],TOSB
	ld	[TOSB+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand
	! TOSA = o->methods, 
        ! TOSB = current class
	! Unhand = unhand(TOSB)
	! R4 interface method block
	! %o0 -- this object,  R3 = guess
	
2:     	! T1 = imethodtable(current class)
	ld	[Unhand + STRUCT_OFFSET(Classjava_lang_Class,imethodtable)], T1 
	! T2 = interface
	ld	[R4 +STRUCT_OFFSET(methodblock,fb.clazz)], T2
	! T3 = imethodcount	    
	ld	[T1 + STRUCT_OFFSET(imethodtable,icount)], T3 
    	! T1 = imethodtable->itable[0]
	add	T1,STRUCT_OFFSET(imethodtable,itable),T1 ! 
	! R4 = interface offset
	ld	[R4 +STRUCT_OFFSET(methodblock,fb.u.offset)], R4
	cmp	R3,T3		! how does guess compare to icount	
	bgeu,a	iiq_search	! out of bounds
    	  sll	T3,3,R3		! [T1 + R3] is just past the end

	sll	R3,3,R3		! R3 = guess << 3
	ld	[T1+R3],T4	! get potential interface
	cmp	T4,T2	    
	bne,a	iiq_search  ! not found
	 sll	T3,3,R3		!
	
iiq_found:  
	!! item is found.  
	!! The R4 contains the index into the offsets table
	!! [T1 + R3] is the location of the matching interface.
	srl	R3,3,T4		! get new guess to store into array
	add	R3,4,R3		! [T1 + T3] points to offsets array
	stb	T4,[PC+4]	! store guess into appropriate place
    	ld	[T1+R3],R3	! this is the offsets table
	sll	R4,2,R4		! make into an array offset
       	inc	STRUCT_OFFSET(methodtable,methods),TOSA 
	ld	[R3+R4],R3	! offsets[offset], which is offset methodtable
    
  	! %o1 = mt_slot(obj_methodtable(%o0), R3)
	add	PC,5,T1		! return PC for callmethod
    	sll	R3,2,R3		
	ba	callmethod
	ld	[TOSA+R3],%o1	! method block for callmethod
	
iiq_loop_top:	
    	cmp	T4,T2
	beq	iiq_found
	nop
iiq_search:
  	subcc	R3,8,R3		! offset of previous item
	bge,a	iiq_loop_top	! still more to do
	  ld	[T1+R3],T4	! get potential interface

iiq_error:  	
	mov	Ee, %o0
	mov	PC, %o1	    
	mov	TOSB, %o2
	mov	T2, %o3
    	CALL_AND_BRANCH(invokeInterfaceError, 4, handleException) ; 
	
	INLINE
	
/*
 * INVOKEVIRTUAL_QUICK
 */

START_SHORT_INVOCATION(invokevirtualobject_quick, finish_invokevirtualobject_quick)
	OUTLINE
finish_invokevirtualobject_quick:
	/* 	%o0	-- object handle. != 0
	 *      %o1     -- method table
    	 *	%o2	-- args_size
	 *	NextByte-- offset into methodtable
	 *	SP	-- pruned back by args_size, this size
	 *	PC	-- points to this instruction
	 */
	add	PC,3,T1
	! determine if this is a normal object or an array.
	andcc	%o1,DEFINED_CONSTANT(FLAG_MASK),%g0
	bz,a	callmethod
	    ld	[%o1+NextByte],%o1 ! normal methodtable slot for this method.

	! This is an array. Use the methodtable of class Object
	sethi	%hi(classJavaLangObject),%o1
	ld	[%o1+%lo(classJavaLangObject)],%o1
	ld	[%o1+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,methodtable)],%o1
	b	callmethod
	ld	[%o1+NextByte],%o1 ! Object methodtable slot for this method.
	
	/*
	 * callmethod parameters are:
	 *	%o0	-- object handle. != 0
	 *	%o1	-- methodblock pointer
	 *	%o2	-- args_size
	 *	T1	-- return PC
	 *	PC	-- points to the invoke instruction
	 */	
	INLINE

/*
 * INVOKEIGNORED_QUICK
 * A "quick call"
 * to a method that just returns.
 * pc[1] (in NextByte) has number of arguments to pop.
 * pc[2] is non-zero if we must check the first argument
 * ( generally the "this" pointer ) for nullity.
 */
OP0(invokeignored_quick)
	ldub	[PC+2],T2
	sll	NextByte,2,NextByte	! scale word count for stack adjustment
	sub	SP,NextByte,SP		! pop args
	ldub	[PC+3],Opcode
	tst	T2
	be	1f
	ldub	[PC+4],NextByte
	! tst arg[0] for nullity.
	ld	[SP],TOSA
	tst	TOSA
	be	nullPointerException
	.empty
1:	SCALE(Opcode)
	GO
	inc	3,PC

/*
 * NEW_QUICK
 * Next two bytes hold constant pool index of class block
 * pointer. Call allocObject, put result on stack, go check
 * for exceptions.
 */
OP0(new_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,NextByte
	add	ConstP,NextByte,%o1
	sll	T2,2,T2
	mov	Ee,%o0
	st	PC,[FP+STRUCT_OFFSET(javaframe,lastpc)]
	call	allocObject,2	
	ld	[%o1+T2],%o1

	tst	%o0			! test resulting handle.
	bne	finish_new_quick
	st	%o0,[SP]		! TOS gets handle
	ba	outOfMemory
	nop
	OUTLINE
		
finish_new_quick:
	ldub	[PC+3],Opcode
	inc	4,SP
	ldub	[PC+4],NextByte
	SCALE(Opcode)
	GO
	inc	3,PC
	INLINE

/*
 * ANEWARRAY_QUICK
 * Array of objects.
 * TOS has array size.
 * Next two bytes in I-stream have constant pool
 * index of classblock pointer of array.
 * This pointer gets stashed at end of array, of course!
 * Leave array handle on stack.
 */
OP0(anewarray_quick)
	ld	[SP-4],TOSB	! TOSB <- size.
	subcc	TOSB,0,%o2
	blt	negativeArraySize
	mov	Ee, %o0
	st	PC,[FP+STRUCT_OFFSET(javaframe,lastpc)]
	call	allocArray,3
	mov	DEFINED_CONSTANT(T_CLASS),%o1

	tst	%o0		! test object handle
	st	%o0,[SP-4]	! stack object handle
	bne	finish_anewarray_quick
	sll	TOSB,2,TOSB	! TOSB <- scaled size
	b,a	outOfMemory
	nop
	OUTLINE
finish_anewarray_quick:
	ld	[%o0+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand
/* [Sheng 11/08/96] memset has been moved into allocArray.
	mov	0,%o1
	mov	TOSB,%o2
	call	memset,3
	mov	Unhand,%o0
*/	
	ldub	[PC+2],R3
	sll	NextByte,10,NextByte

	ldub	[PC+3],Opcode
	sll	R3,2,R3
	add	ConstP,NextByte,T1
	ld	[T1+R3],T1
	SCALE(Opcode)
	ldub	[PC+4],NextByte
	st	T1,[Unhand+TOSB]
	GO
	inc	3,PC
	INLINE

/*
 * MULTIANEWARRAY_QUICK
 */
OP0(multianewarray_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,NextByte
	add	ConstP,NextByte,NextByte
	sll	T2,2,T2
	ldub	[PC+3],TOSA	   ! Dimensions, for later use
	ld	[T2+NextByte],%o1  ! constant_pool[GET_INDEX(pc+1)] = array_cb
	mov	TOSA,%o0
	sll	TOSA,2,TOSA
	ld	[SP-4],R3           ! least significant dimension
	sub	SP,TOSA,%o2        ! pointer to most significant dimension
	b	finish_multianewarray_quick
	sub	TOSA,4,TOSA     

	OUTLINE
finish_multianewarray_quick:
	! Now have:
        !       %o0  -- dimensions.  Must be >= 1 !!!
	!	%o1  -- constant_pool[GET_INDEX(pc+1)] = array_cb
        !       %o2  -- pointer to most significant dimension
 	!	TOSA -- (dimensions - 1) << 2
        !       R3    -- least significant dimension
        ! Check to make sure all the arguments are >= 0.  
        ! The order of checking doesn't matter.

1:	! ASSERT: R3 = *(int *)(%o2 + TOSA)
	tst	R3
	blt	negativeArraySize
	subcc	TOSA,4,TOSA
	bge,a	1b
	ld	[%o2+TOSA],R3

        ! Trim the stack back to %o2.  Then call 
    	! MultiArrayAlloc(dimensions, array_cb, pointer to first dimension)
	st	PC,[FP+STRUCT_OFFSET(javaframe,lastpc)]
	call	MultiArrayAlloc,3
	mov	%o2,SP

	! %o0 has new handle, or 0
	! go barf on zero,
	! else stack result and dispatch next instruction
	ldub	[PC+4],Opcode
	tst	%o0
	st	%o0,[SP]
	inc	4,SP
	bz	outOfMemory
	ldub	[PC+5],NextByte
	SCALE(Opcode)
	GO
	inc	4,PC
	INLINE

/*
 * CHECKCAST_QUICK
 * 2 bytes in Istream give constant pool index of a class pointer.
 * Handle is on TOS. Leave it there.
 */
OP0(checkcast_quick)
	ldub	[PC+2],T2
	sll	NextByte,10,T1
	ld	[SP-4],%o0	! get object
	add	ConstP,T1,T1	
	sll	T2,2,T2
	ld	[T2+T1],%o1	! get constant pool entry
	mov	Ee,%o2
	call	is_instance_of,3
	ldub	[PC+3],NextByte
	tst	%o0
	SCALENEXT
	bz	checkcast_failure
	ldub	[PC+4],NextByte
	GO
	inc	3,PC

/*
 * INSTANCEOF_QUICK
 * 2 bytes in Istream give constant pool index of a class pointer.
 * Handle is on TOS. Replace it with 0 or 1.
 */
OP0(instanceof_quick)
	ld	[SP-4],%o0	! get object
	sll	NextByte,10,%o1
	ldub	[PC+2],T2
	tst	%o0		! is the object null?
	ldub	[PC+3],R3
	bnz	finish_instanceof_quick
	add	ConstP,%o1,%o1
	! the pointer is null, the answer is 0
	st	%g0,[SP-4]
	SCALE(R3)
	ldub	[PC+4],NextByte
	GO
	inc	3,PC
	OUTLINE
finish_instanceof_quick:
	sll	T2,2,T2
	ld	[%o1+T2],%o1
	call	is_instance_of,3
	mov	Ee,%o2
	! if there was an exception, go handle it.
	! otherwise, continue.
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T2
	st	%o0,[SP-4]
	ldub	[PC+4],NextByte
	tst	T2
	bgt	handleException
	SCALE(R3)
	GO
	inc	3,PC
	INLINE

/*
 * INVOKEVIRTUAL_QUICK_W.
 */
START_INVOCATION(invokevirtual_quick_w)
	/* %o1 -- methodblock pointer
	 * %o2 -- args_size from the methodblock
	 * T1   -- PC+3
	 */
	sll	%o2,2,R3
	ld	[%o1+STRUCT_OFFSET(methodblock,fb.u.offset)],TOSA
	sub	SP,R3,SP		! trim stack
	ld	[SP],%o0	! get object handle
	tst	%o0
	bnz,a   finish_invokevirtual_quick_w
    	  ld	[%o0+STRUCT_OFFSET(Hjava_lang_Object,methods)],%o1
	b,a	nullPointerException
	nop

	
	OUTLINE	    
finish_invokevirtual_quick_w:
	sll	TOSA,2,TOSA	! scale for indexing
	andcc	%o1,DEFINED_CONSTANT(FLAG_MASK),%g0
	inc	STRUCT_OFFSET(methodtable,methods),TOSA ! add methods[] offset
	bz,a	callmethod
	    ld	[%o1+TOSA],%o1	! normal object
	!
	! This is an array. Use the methodtable of class Object
	!
	sethi	%hi(classJavaLangObject),%o1
	ld	[%o1+%lo(classJavaLangObject)],%o1
	ld     	[%o1+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,methodtable)],%o1
	b	callmethod
	ld	[%o1+TOSA],%o1	! Object methodtable slot for this method.
	INLINE
	
/*
 * GETFIELD_QUICK_W
 * A real mess. Must look at the signature of the
 * field to figure out what to fetch.
 */
OP0(getfield_quick_w)
	ld	[SP-4],TOSA	! fetch object handle
	sll	NextByte,10,NextByte
	ldub	[PC+2],T2
	add	ConstP,NextByte,NextByte
	tst	TOSA
	sll	T2,2,T2
	bne	finish_getfield_quick_w
	ld	[NextByte+T2],T2 ! fieldblock pointer from const pool
	b,a	nullPointerException
	OUTLINE
	!
	! TOSA	-- object handle, != 0
	! T2	-- fieldblock pointer from constant_pool 
	!
finish_getfield_quick_w:
	ld	[T2+STRUCT_OFFSET(fieldblock,signature)],T1	   ! sig
	inc	3,PC
	ld	[T2+STRUCT_OFFSET(fieldblock,u.offset)],R3	   ! offset
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand ! data pointer
	ldsb	[T1],T1					! sig[0]
	ldub	[PC],Opcode				! I-fetchahead
	add	R3,4,T3
	ld	[Unhand+R3],TOSA				! data word
	cmp	T1,DEFINED_CONSTANT(SIGNATURE_LONG)
	SCALE(Opcode)
	be	1f
	ldub	[PC+1],NextByte				! I-fetchahead
	cmp	T1,DEFINED_CONSTANT(SIGNATURE_DOUBLE)
	be	2f
	st	TOSA,[SP-4]
	GO
	nop
	!
	! here if we really had a 2-word thingie to fetch
	! first word has been fetched. Fetch 2nd, store both.
1:	ld	[Unhand+T3],TOSB
	st	TOSA,[SP-4]
	inc	4,SP
	GO
	st	TOSB,[SP-4]
	!
	! here if we really had a 2-word thingie to fetch
	! first word has been dealt with. Do 2nd.
2:	ld	[Unhand+T3],TOSB
	inc	4,SP
	GO
	st	TOSB,[SP-4]
	INLINE

/*
 * PUTFIELD_QUICK_W
 * This is especially odious because you have to do
 * 4 memory accesses before even learning where on the stack
 * the object handle is!
 * Quick indeed!
 */
OP0(putfield_quick_w)
	ldub	[PC+2],T2
	sll	NextByte,10,NextByte
	ld	[SP-4],TOSB
	add	ConstP,NextByte,NextByte
	ld	[SP-8],TOSA
	sll	T2,2,T2
	dec	8,SP
	ld	[NextByte+T2],T2
	b	finish_putfield_quick_w
	ldub	[PC+3],Opcode
	OUTLINE
finish_putfield_quick_w:
	!
	! Both TOSA and TOSB are filled from the stack,
	! T2 contains the fieldblock pointer from constant_pool
	! Opcode fetched ahead.
	ld	[T2+STRUCT_OFFSET(fieldblock,signature)],T1
	ld	[T2+STRUCT_OFFSET(fieldblock,u.offset)],R3
	ldsb	[T1],T1
	SCALE(Opcode)
	cmp	T1,DEFINED_CONSTANT(SIGNATURE_LONG)
	be	1f
	ldub	[PC+4],NextByte
	cmp	T1,DEFINED_CONSTANT(SIGNATURE_DOUBLE)
	be,a	2f
	    ld	[SP-4],T1
	tst	TOSA
	be	nullPointerException
	nop
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand
	inc	3,PC
	GO
	st	TOSB,[Unhand+R3]
1:
	ld	[SP-4],T1
2:	add	R3,4,T3
	tst	T1
	be	nullPointerException
	dec	4,SP
	ld	[T1+STRUCT_OFFSET(Hjava_lang_Object,obj)],Unhand
	inc	3,PC
	st	TOSA,[Unhand+R3]
	GO
	st	TOSB,[Unhand+T3]
	INLINE

/*
 * nonnull_quick -- throw exception if top of stack is null
 */
OP0(nonnull_quick)
        ld      [SP-4], TOSA
	SCALENEXT
	dec	4,SP
        tst     TOSA
	be	nullPointerException
	ldub	[PC+2],NextByte
	GO
	inc	PC

ENDVECTOR

OUTLINE	
/*
 * Exception Handling.
 * There are a number of subcases, pretty much the same:
 * Call SignalError, then goto handleException, were we do the real work.
 */
/*
 * Now the entry points, all about the same.
 * Speed is not an issue here, since we are already fu.
 * Macro parameters are label, name, data.
 */

define(`USUAL_EXCEPTION',`$1:
	SET_ADDRESS($2, %o1)
	b	doExceptionProcessing
	ifelse(`$3', 0,
	`mov	0,%o2',
	`SET_ADDRESS($3,%o2)')
')


USUAL_EXCEPTION( nullPointerException,	nullPointerName, 0)
USUAL_EXCEPTION( arithmeticException,	arithmeticExceptionName, divZero)
USUAL_EXCEPTION( negativeArraySize,	negativeArrayName, 0)
USUAL_EXCEPTION( outOfMemory,		outOfMemoryName, 0)
USUAL_EXCEPTION( arrayStoreException,	arrayStoreName, 0)
	/*
	 * Illegal opcode error.
	 * Capture the opcode in the error message
	 */
internalError:
	sub	%fp,ErrorBuffer,%o0
	mov	ErrorBufferSize,%o1
	SET_ADDRESS(undefOpcode,%o2)
	call	jio_snprintf,4
	srl	Opcode,OPSCALE,%o3

	SET_ADDRESS(internalErrorName,%o1)
	b	doExceptionProcessing
	sub	%fp,ErrorBuffer,%o2
	/*
	 * Illegal wide sub-opcode error.
	 * Capture the opcode in the error message
	 */
wideError:
	sub	%fp,ErrorBuffer,%o0
	mov	ErrorBufferSize,%o1
	SET_ADDRESS(undefWideOpcode,%o2)
	call	jio_snprintf,4
	mov	NextByte,%o3

	SET_ADDRESS(internalErrorName,%o1)
	b	doExceptionProcessing
	sub	%fp,ErrorBuffer,%o2
	/*
	 * A checked cast failed, in checkcast_quick.
	 * The handle we could not cast is on TOS
	 * Format a message concerning this occurance.
	 */
checkcast_failure:
	ld	[SP-4],TOSA
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Object,methods)],%o0
	andcc	%o0,DEFINED_CONSTANT(FLAG_MASK),%g0
	bz,a	1f
	    ld	[%o0+STRUCT_OFFSET(methodtable,classdescriptor)],%o0
	    ! oops. Its an array. Use Object
	    sethi	%hi(classJavaLangObject),%o0
	    ld		[%o0+%lo(classJavaLangObject)],%o0
1:	sub	%fp,ErrorBuffer,%o1
	mov	ErrorBufferSize,%o2
	ld	[%o0+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand
	call	classname2string,3
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,name)],%o0

	mov	%o0,%o2
	b	doExceptionProcessing
	SET_ADDRESS(classCastName,%o1)
	/*
	 * An array index was out of bounds.
	 * We have the illegal index in one of our registers,
	 * and want to include it in our message.
	 * Unfortunately, it is not always the same register.
	 * We pay a little more here for the best code in the normal
	 * cases.
	 */
arrayIndexOutOfBoundsD:
	b	1f
	mov	T1,%o3
arrayIndexOutOfBoundsA:
	b	1f
	mov	TOSA,%o3
arrayIndexOutOfBoundsB:
	mov	TOSB,%o3
	!fall through
1:
	sub	%fp,ErrorBuffer,%o0
	SET_ADDRESS(percentD,%o2)
	call	jio_snprintf,4
	mov	ErrorBufferSize,%o1

	SET_ADDRESS(indexBoundsName,%o1)
	b	doExceptionProcessing
	sub	%fp,ErrorBuffer,%o2

	/*
	 * Something bad happened in new.
	 * classblock pointer is in TOSA.
	 */
instantiationError:
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand ! class access
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,name)],%o2
	b	doExceptionProcessing
	SET_ADDRESS(instantiationErrorName,%o1)
	
illegalAccessError:
	ld	[TOSA+STRUCT_OFFSET(Hjava_lang_Class,obj)],Unhand ! class access
	ld	[Unhand+STRUCT_OFFSET(Classjava_lang_Class,name)],%o2
	b	doExceptionProcessing
	SET_ADDRESS(illegalAccessName,%o1)

	/*
	 * Here we think that an exception probably happened.
	 * But be prepared to continue instruction execution,
	 * just in case.
	 */
checkException:
	ldsb	[Ee+STRUCT_OFFSET(execenv,exceptionKind)],T2
	ldub	[PC],Opcode
	tst	T2
	bgt,a	handleException.2
	    ld	[Ee+STRUCT_OFFSET(execenv,exception.exc)],TOSA ! get exception object
	SCALE(Opcode)
	GO
	ldub	[PC+1],NextByte

doExceptionProcessing:
	!
	! have:
	!	%o1 -- points to asciz string of name of exception we want.
	!	%o2 -- points to error message to bury therein.
	! we load up %o0 with Ee, store fp.lastpc = pc,
	! and call SignalError, which is charged with putting an exceptional
	! object in Ee.exception. Then we proceed assuming we got the
	! exception.
	!
	mov	Ee,%o0
	call	SignalError,3
	st	PC,[FP+STRUCT_OFFSET(javaframe,lastpc)]
handleException:
	ld	[Ee+STRUCT_OFFSET(execenv,exception.exc)],TOSA ! get exception object
handleException.2:
	! clear the pending exception before calling ProcedureFindThrowTag
	stb	%g0,[Ee+STRUCT_OFFSET(execenv,exceptionKind)]

        mov     Ee,%o0
	mov	PC,%o3
	mov	FP,%o1
	call	ProcedureFindThrowTag,4
	mov	TOSA,%o2

	ldub	[Ee+STRUCT_OFFSET(execenv,exceptionKind)], T2
	tst	T2
	bgt	exceptionInFindThrowTag

	! restore the exception object.
	mov	DEFINED_CONSTANT(EXCKIND_THROW),T2
	st	TOSA,[Ee+STRUCT_OFFSET(execenv,exception)]
	stb	T2,[Ee+STRUCT_OFFSET(execenv,exceptionKind)]

exceptionInFindThrowTag:		
	addcc	%o0,0,PC
	bz,a	exceptionalFallOut
	    sethi	%hi(java_profiler_isOn),R3
	!
	! ProcedureFindThrowTag found a handler in this routine.
	! clear Ee.exceptionKind. Reload our state, empty the
	! expression stack, push the exception object on it, and go.
	! 
	stb	%g0,[Ee+STRUCT_OFFSET(execenv,exceptionKind)]
	ldub	[PC],Opcode
	ld	[FP+STRUCT_OFFSET(javaframe,vars)],Varbase
	ld	[FP+STRUCT_OFFSET(javaframe,constant_pool)],ConstP
	add	FP,STRUCT_OFFSET(javaframe,ostack),SP
	st	TOSA,[SP]	! TOSA holds exception object
	SCALE(Opcode)
	ldub	[PC+1],NextByte
	jmp	Dispatch+Opcode
	inc	4,SP

	/*
	 * Did not catch exception in this routine.
	 * Return to caller -- perhaps he will know what to do with it!
	 */
exceptionalFallOut:
	ld	[FP+STRUCT_OFFSET(javaframe,monitor)],%o1 ! cur_frame->monitor
	ld	[FP+STRUCT_OFFSET(javaframe,prev)],ConstP
	b	finish_return
	ld	[R3+%lo(java_profiler_isOn)],R3 ! java_profiler_isOn in R3
	! FP contains frame
	! ConstP contains fram->prev
	! %o0 contains frame->monitor
	! R3  contains java_profiler_isOn
	! SP, Varbase, ConstP irrelevant
	!
	
ifdef(`EXTRACT_OFFSETS', `'
      ,`undivert(OUTLINE_DIVERSION)
        /* The literals need to be just before the dispatch table because
	 * the are referenced using an offset from the dispatch table (and
	 * offsets are limited to 13 bits 
         */
	undivert(LITERAL_DIVERSION)
        undivert(INLINE_DIVERSION)
       ')
	.size	ExecuteJava,(.-ExecuteJava)



