/*
 * @(#)invokeNative_sparc.s	1.8 98/03/18
 *
 * Copyright 1996, 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

	.section	".text",#alloc,#execinstr
	.file	"invokeNative_sparc.s"

	.section	".text",#alloc,#execinstr
	.align	4
!
! SUBROUTINE sysInvokeNative
!
! This function translates the "Java" calling convention into the "C"
! calling convention used in native methods. Java VM passes all the 
! arguments in the Java stack, and expects the results to be placed there 
! as well. We therefore have to copy the arguments into the C stack (or 
! registers), and place the return values back into the Java stack.
!
! With a small sacrifise in efficiency, this approach avoids having to
! generate a stub function for every native methods.
!
! See "The SPARC Architecture Manual" for information on native 
! argument-passing conventions.
!
! Since the SPARC does not pass DOUBLEs in floating-point registers, and 
! does not have double-word alignment requirement for DOUBLEs and LONGs, 
! we do not have to parse the method descriptor to obtain the exact type 
! information for each argument. 
!
! The first argument to sysInvokeNative is a pointer to the JNI
! environment, which should be passed unmodified as the first argument
! to the native method.
!
! The second argument is a pointer to the "real" native method function.
!
! The third argument (stk) is a pointer to the Java stack, where all
! the arguments are stored (as stk[0], stk[1], etc.).
!
! The fourth argument is the signature of the native method.
!
! The fifth argument is the total size (in 32-bit words) of the
! arguments on the Java stack. Note that the Java stack does not have
! any alignment requirement, and stores all arguments consecutively in
! words and double words. The argument size includes the "this" pointer
! for non-static methods.
!
! The sixth argument is 0 for non-static methods, or a jclass
! for static methods. Non-static native methods receive an object
! reference as the second argument (which is simply the address of
! stk[0]). The "real" method arguments to non-static methods begin at
! stk[1]. Static native methods receive a class reference as the second
! argument.
!
! The return value of the native method is placed at stk[0] for
! word-sized results, or at stk[0] and stk[1] for
! double-word-sized results. The return value of sysInvokeNative is
! stk if the native method returns void, stk+1 if the native
! method returns a word, or stk+2 if the native method returns a
! double word.

	.global sysInvokeNative

! Arguments:
!
! i0	JNI environment
! i1	native function
! i2	Java stack top
! i3	method descriptor
! i4	number of argument words to be passed to native function
! i5	class for static methods, or 0 for non-static methods
!
! Results:	
! o0	the adjusted Java stack top
!

#define SIGPTR %i3
#define SIGBYTE %g2
#define JSTKPTR %g3
#define CSTKPTR %g4	
	       
sysInvokeNative:

! Set up a big enough stack frame (and make sure it's 8-byte aligned):

	subcc	%o4,4,%g1	! We can freely use %g1.
	bpos	more_than_6_args
	nop
	mov	%g0,%g1

more_than_6_args:
	add	%g1,1,%g1
	sra	%g1,1,%g1
        sll	%g1,3,%g1
	sub	%g0,%g1,%g1	! negate %g1
        sub	%g1,96,%g1
	save	%sp,%g1,%sp

! transferring arguments
!
! first 6 in register

! The first argument is always the JNI env.

	mov	%i0,%o0

! The next one is %i2 or %i5 if it's non-zero (for static methods).

	mov	%i2,JSTKPTR	! keep a backup for %i2.

	add	%sp,76,CSTKPTR	! l0 points to 3rd argument
	ldub	[SIGPTR], SIGBYTE	! preload signature

	set	arg_jumps,%l4	! load the base of the jump table

        orcc	%i5,0,%i5
	be	non_static
        nop
	ba	args_loop
	mov	%i5,%o1

non_static:
	mov	JSTKPTR,%o1
        add     JSTKPTR,4,JSTKPTR

args_loop:
	sll	SIGBYTE,2,SIGBYTE
	ld	[%l4 + SIGBYTE],SIGBYTE
	jmp     SIGBYTE
	inc	SIGPTR

arg_32:			! move a 32-bit value from [JSTKPTR] to [CSTKPTR].
	ld	[JSTKPTR],%o2
	ldub	[SIGPTR], SIGBYTE	! preload
	add	JSTKPTR,4,JSTKPTR
	st	%o2,[CSTKPTR]
	sll	SIGBYTE,2,SIGBYTE
	add	CSTKPTR,4,CSTKPTR	
	ld	[%l4 + SIGBYTE],SIGBYTE
	jmp     SIGBYTE
	inc	SIGPTR

arg_64:
	ld	[JSTKPTR],%o2
	ld	[JSTKPTR+4],%o3
	ldub	[SIGPTR], SIGBYTE
	add	JSTKPTR,8,JSTKPTR
	st	%o2,[CSTKPTR]
	st	%o3,[CSTKPTR+4]
	add	CSTKPTR,8,CSTKPTR
	sll	SIGBYTE,2,SIGBYTE
	ld	[%l4 + SIGBYTE],SIGBYTE
	jmp     SIGBYTE
	inc	SIGPTR
		
arg_object:
	ld	[JSTKPTR],%o2
	cmp	%o2,0
	be	object_checked
	ldub	[SIGPTR],SIGBYTE       ! alreadys executed
	mov	JSTKPTR,%o2
object_checked:

	add	JSTKPTR,4,JSTKPTR
	st	%o2,[CSTKPTR]
	add	CSTKPTR,4,CSTKPTR	
	sll	SIGBYTE,2,SIGBYTE
	ld	[%l4 + SIGBYTE],SIGBYTE
	jmp     SIGBYTE
	inc	SIGPTR

args_done:
	add	%sp,76,CSTKPTR
	ld	[CSTKPTR],%o2
	ld	[CSTKPTR+4],%o3
	ld	[CSTKPTR+8],%o4
	ld	[CSTKPTR+12],%o5	

	ldub	[SIGPTR], %l1	! get the return type
	set	ret_jumps,%l4
	
	jmpl	%i1,%o7
	sll	%l1,2,%l1
	
	ld	[%l4 + %l1],SIGBYTE	! thread the return address to the
	jmp	SIGBYTE			! proper code for our return type
	nop

ret_obj:
	cmp	%o0, 0
	be	ret_null_obj
	nop
	ld	[%o0],%o0
ret_null_obj:	
	st	%o0,[%i2]
	ret
        restore %i2,4,%o0
	
ret_f64:	
	st      %f0,[%i2]
        st      %f1,[%i2+4]
        ret
	restore %i2,8,%o0

ret_f32:
	st      %f0,[%i2]
	ret
        restore %i2,4,%o0

ret_s32:	
	st	%o0,[%i2]
	ret
        restore %i2,4,%o0

ret_s64:
	st	%o0,[%i2]
	st	%o1,[%i2+4]
	ret
        restore %i2,8,%o0

ret_s8:
	sll	%o0,24,%o0
	sra	%o0,24,%o0
	st	%o0,[%i2]
	ret
        restore %i2,4,%o0

ret_u8:
	sll	%o0,24,%o0
	srl	%o0,24,%o0
	st	%o0,[%i2]
	ret
        restore %i2,4,%o0

ret_s16:
	sll	%o0,16,%o0
	sra	%o0,16,%o0
	st	%o0,[%i2]
	ret
        restore %i2,4,%o0
	
ret_u16:
	sll	%o0,16,%o0
	srl	%o0,16,%o0
	st	%o0,[%i2]
	ret
        restore %i2,4,%o0

ret_void:
        ret
        restore %i2,0,%o0

ret_jumps:		
	.word	ret_void	! this is bogus and shouldn't get called
	.word	ret_obj
	.word	ret_s64
	.word	ret_f64
	.word	ret_u8
	.word	ret_s8
	.word	ret_s16
	.word	ret_u16
	.word	ret_s32
	.word	ret_f32
	.word	ret_void

arg_jumps:
	.word	ret_void	! this is bogus and shouldn't get called
	.word	arg_object
	.word	arg_64
	.word	arg_64
	.word	arg_32	! bool
	.word	arg_32	! byte
	.word	arg_32	! short
	.word	arg_32	! char
	.word	arg_32	! int
	.word	arg_32	! float
	.word	args_done   ! for void - remind:	 shouldn't happen
	.word	args_done   ! end-of-args

	.type	sysInvokeNative,2
	.size	sysInvokeNative,(.-sysInvokeNative)
