/
/ @(#)invokeNative_x86.s	1.9 98/03/18
/
/ Copyright 1996, 1997 by Sun Microsystems, Inc.,
/ 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
/ All rights reserved.
/
/ This software is the confidential and proprietary information
/ of Sun Microsystems, Inc. ("Confidential Information").  You
/ shall not disclose such Confidential Information and shall use
/ it only in accordance with the terms of the license agreement
/ you entered into with Sun.
/

	.file "invokeNative_x86.s"
	.section	.text
	.align	4	
	.globl sysInvokeNative

 / This function translates the "Java" calling convention into the
 / C convention used in native methods. Java VM passes all the
 / arguments in the Java stack, and expects the results to be placed there
 / as well. We therefore have to copy the arguments into the C stack (or
 / registers), and place the return values back into the Java stack.
 /
 / With a small sacrifise in efficiency, this approach avoids having to
 / generate a stub function for every native method.
 /
 / The x86 passes all arguments on stack, and returns floating pointer
 / results in ST0, 32-bit results in eax, and 64-bit results in eax and
 / edx.
 /
 / The first argument to sysInvokeNative is a pointer to the JNI
 / environment, which should be passed unmodified as the first argument
 / to the native method.
 /
 / The second argument is a pointer to the "real" native method function.
 /
 / The third argument (stk) is a pointer to the Java stack, where all
 / the arguments are stored (as stk[0], stk[1], etc.).
 /
 / The fourth argument is the "terse" signature of the native method,
 / which basically collapses all objects in the long signature into
 / one byte, since they're all treated the same.  This makes the parsing
 / in this routine simpler and faster.  See classload.c and classruntime.c
 / for details.
 /
 / The fifth argument is the total size (in 32-bit words) of the
 / arguments on the Java stack. Note that the Java stack does not have
 / any alignment requirement, and stores all arguments consecutively in
 / words and double words. The argument size includes the "this" pointer
 / for non-static methods.
 /
 / The sixth argument is 0 for non-static methods, or a jclass
 / for static methods. Non-static native methods receive an object
 / reference as the second argument (passed in the Java stack as
 / stk[0]). The "real" method arguments to non-static methods begin at
 / stk[1]. Static native methods receive a class reference as the second
 / argument.
 /
 / The return value of the native method is placed at stk[0] for
 / word-sized results, or at stk[0] and stk[1] for
 / double-word-sized results. The return value of sysInvokeNative is
 / stk if the native method returns void, stk+1 if the native
 / method returns a word, or stk+2 if the native method returns a
 / double word.

// #define args_again \ // I don't know how to do a macro w/ this assembler
//	movb	(%ebx), %cl \
//	incl	%ebx \
//	jmp	*arg_jumps(,%ecx,4)

	

/ _env$ = 8 ; JNI environment
/ _f$ = 12 ; native method function
/ _stk$ = 16 ; Java stack
/ _sig$ = 20 ; method signature
/ _sz$ = 24 ; total argument size
/ _cls$ = 28 ; class (if static)

sysInvokeNative:

	pushl	%ebp
	movl	%esp, %ebp
	pushl	%esi
	pushl	%edi
	pushl	%ebx

	movl	16(%ebp), %esi / stk
	movl	24(%ebp), %edx / sz
	movl	28(%ebp), %ecx / cls

	cmpl	$0, %ecx
	jne	static_call
	movl	%esi, %ecx 
	addl	$4, %esi 
	jmp	static_done

static_call: 
	addl	$1, %edx 

static_done:
	addl	$1, %edx
	shll	$2, %edx / word address -> byte address

	mov	%edx, 24(%ebp)  / store edx in sz for later use
	
	subl	%edx, %esp
	movl	%esp, %edi

	movl	8(%ebp), %eax / env
	movl	%eax, (%edi)
	addl	$4, %edi
	movl	%ecx, (%edi)
	addl	$4, %edi

	movl	20(%ebp), %ebx
	xorl	%ecx, %ecx / zero it

 // ebx is current index into sig
 // cl (ecx) is current char in sig
 // esi is index into java stack
 // edx is (was) the # args, though we don't
 // need it anymore.
args_loop:
	movb	(%ebx), %cl 
	incl	%ebx 
	jmp	*arg_jumps(,%ecx,4)

arg_32:
	movl	(%esi), %eax
	addl	$4, %esi
	movl	%eax, (%edi)
	addl	$4, %edi
//	args_again 
	movb	(%ebx), %cl 
	incl	%ebx 
	jmp	*arg_jumps(,%ecx,4)

arg_64:
	movl	(%esi), %eax
	addl	$4, %esi
	movl	%eax, (%edi)
	addl	$4, %edi
	movl	(%esi), %eax
	addl	$4, %esi
	movl	%eax, (%edi)
	addl	$4, %edi
//	args_again 
	movb	(%ebx), %cl 
	incl	%ebx 
	jmp	*arg_jumps(,%ecx,4)

arg_class:
	movl	(%esi), %eax
	cmpl	$0, %eax
	je	object_checked
	movl	%esi, %eax
object_checked:
	movl	%eax, (%edi)
	addl	$4, %esi
	addl	$4, %edi
//	args_again 
	movb	(%ebx), %cl 
	incl	%ebx 
	jmp	*arg_jumps(,%ecx,4)

args_done:
	movl	%ebx, %edi
	call	*12(%ebp) / f
	movl	24(%ebp), %ebx	/ sz
	addl	%ebx, %esp	/ pop off arguments.  different from __stdcall
				/ on win32
	movl	%edi, %ebx
	movl	16(%ebp), %esi
	xorl	%ecx, %ecx / c register volatile cross calls - zero it
	movb	(%ebx), %cl
	jmp	*ret_jumps(,%ecx,4)

ret_obj:
	cmp	$0, %eax
	je	ret_null_obj
	mov	(%eax), %eax

ret_null_obj:
	movl	%eax, (%esi)
	leal	4(%esi), %eax
	jmp	done

ret_f64:
	fstpl	(%esi)
	leal	8(%esi), %eax
	jmp	done

ret_f32:
	fstps	(%esi)
	leal	4(%esi), %eax
	jmp	done

ret_s32:
	movl	%eax, (%esi)
	leal	4(%esi), %eax
	jmp	done

ret_s64:
	movl	%eax, (%esi)
	leal	8(%esi), %eax
	movl	%edx, 4(%esi)
	jmp	done

ret_s8:
	shll	$24, %eax
	sarl	$24, %eax
	movl	%eax, (%esi)
	leal	4(%esi), %eax
	jmp	done

ret_u8:
	shll	$24, %eax
	shrl	$24, %eax
	movl	%eax, (%esi)
	leal	4(%esi), %eax
	jmp	done

ret_s16:
	shll	$16, %eax
	sarl	$16, %eax
	movl	%eax, (%esi)
	leal	4(%esi), %eax
	jmp	done

ret_u16:
	shll	$16, %eax
	shrl	$16, %eax
	movl	%eax, (%esi)
	leal	4(%esi), %eax
	jmp	done

ret_void:
	leal	(%esi), %eax

done:
	popl	%ebx
	popl	%edi
	popl	%esi
	movl	%ebp, %esp
	popl	%ebp
	ret

 // keep these offsets in sync with the enum in signature.h
ret_jumps: 
	.long done // this is bogus and shouldn't get called
	.long ret_obj
	.long ret_s64
	.long ret_f64
	.long ret_u8
	.long ret_s8
	.long ret_s16
	.long ret_u16
	.long ret_s32
	.long ret_f32
	.long ret_void

 // keep these offsets in sync with the enum in signature.h
arg_jumps:
	.long done // this is bogus and shouldn't get called
	.long arg_class
	.long arg_64
	.long arg_64
	.long arg_32 / bool
	.long arg_32 / byte
	.long arg_32 / short
	.long arg_32 / char
	.long arg_32 / int
	.long arg_32 / float
	.long args_done / for void - remind:  shouldn't happen
	.long args_done / end-of-args

	.align	4
	.type	sysInvokeNative,@function
	.size sysInvokeNative,.-sysInvokeNative
