/*
 * @(#)javai_md.c	1.7 98/09/15
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <dlfcn.h>
#include <stdlib.h>
#include <math.h>

#include "oobj.h"
#include "interpreter.h"
#include "sys_api.h"

#ifdef DEBUG
#define HPI_DLL "libhpi_g.so"
#else
#define HPI_DLL "libhpi.so"
#endif

void *
InitializeHPI(void *args)
{
    char buf[MAXPATHLEN];
    Dl_info dlinfo;
    void *hpi_handle;
    GetInterfaceFunc getintf = NULL;
    jint (JNICALL * DLL_Initialize)(GetInterfaceFunc *, void *);
    char *thread_type = getenv("_JVM_THREADS_TYPE");

    if (thread_type) {
	/* Set by .java_wrapper.  Remove for the sake of exec'd VMs. */
	thread_type = strdup(thread_type);
        putenv("_JVM_THREADS_TYPE=");
    } else {
	/* Default thread type for the invocation API. */
        thread_type = strdup("native_threads");
    }

    dladdr((void *)InitializeHPI, &dlinfo);
    strcpy(buf, (char *)dlinfo.dli_fname);
    *(strrchr(buf, '/')) = '\0';	/* get rid of /libjvm.so */
    *(strrchr(buf, '/') + 1) = '\0';	/* get rid of classic    */
    strcat(buf, thread_type);
    strcat(buf, "/" HPI_DLL);
    /* we use RTLD_NOW because of bug 4032715 */
    hpi_handle = dlopen(buf, RTLD_NOW);
    if (hpi_handle == NULL) {
	goto bail;
    }
    DLL_Initialize = (jint (JNICALL *)(GetInterfaceFunc *, void *))
        dlsym(hpi_handle, "DLL_Initialize");
    if (DLL_Initialize == NULL ||
        (*DLL_Initialize)(&getintf, args) < 0) {
        goto bail;
    }
bail:
    free(thread_type);
    return (void *)getintf;
}

props_md_t *
GetPropertiesMD()
{
    static props_md_t sprops = {0};
    char *v; /* tmp var */

    if (sprops.dll_dir) {
        return &sprops;
    }

    /* sysclasspath, java_home, dll_dir */
    {
        char *home_path;
	char *sysclasspath;
	char *dll_path;
        char buf[MAXPATHLEN];
	Dl_info dlinfo;
	dladdr((void *)GetPropertiesMD, &dlinfo);
	realpath((char *)dlinfo.dli_fname, buf);
	*(strrchr(buf, '/')) = '\0';	/* get rid of /libjvm.so */
	*(strrchr(buf, '/')) = '\0';    /* get rid of classic    */
	dll_path = malloc(strlen(buf) + 1);
	if (dll_path == 0)
	    return NULL;
	strcpy(dll_path, buf);
	sprops.dll_dir = dll_path;

	*(strrchr(buf, '/')) = '\0';
	*(strrchr(buf, '/')) = '\0';

	home_path = malloc(strlen(buf) + 1);
	if (home_path == 0)
	    return NULL;
	strcpy(home_path, buf);
	sprops.java_home = home_path;

	sysclasspath = malloc(strlen(home_path) * 3 + 75);
	if (sysclasspath == 0) {
          return NULL;
	}
	sprintf(sysclasspath,
		"%s/lib/rt.jar:%s/lib/i18n.jar:%s/classes",
		home_path, home_path, home_path);
	sprops.sysclasspath = sysclasspath;
    }

    /* Where to look for native libraries */
    {
	/*
	 * Get the user setting of LD_LIBRARY_PATH
	 */
	v = getenv("LD_LIBRARY_PATH");

	if (v == NULL) {
	    v = "";
	}
	
	sprops.library_path = strdup(v);
    }

    /* Extensions directories */
    {
	char buf[MAXPATHLEN + 10];
	sprintf(buf, "%s/lib/ext", sprops.java_home);
	sprops.ext_dirs = strdup(buf);
    }

    return &sprops;
}

int32_t float2l(float f)
{
    int64_t lltmp = (int64_t)f;
    int32_t ltmp = (int32_t)lltmp;
    if (ltmp == lltmp) {
        return ltmp;
    } else {
        if (isnan((double)f)) {return 0;}
        if (f < 0) {
            return 0x80000000;
        } else {
            return 0x7fffffff;
        }
    }
}

int32_t double2l(double f)
{
    int64_t lltmp = (int64_t)f;
    int32_t ltmp = (int32_t)lltmp;
    if (ltmp == lltmp) {
        return ltmp;
    } else {
        if (isnan(f)) {return 0;}
        if (f < 0) {
            return 0x80000000;
        } else {
            return 0x7fffffff;
        }
    }
}

int64_t float2ll(float f)
{
#ifdef i386
    int64_t lltmp = (int64_t)f;
    if (lltmp != 0x8000000000000000) {
        return lltmp;
    } else {
        if (isnan((double)f)) {return 0;}
        if (f < 0) {
            return 0x8000000000000000;
        } else {
            return 0x7fffffffffffffff;
        }
    }
#endif /* i386 */

#ifdef sparc
    int64_t lltmp = (int64_t)f;
    if (lltmp != 0x7fffffffffffffff) {
        return lltmp;
    } else {
        if (isnan((double)f)) {return 0;}
        return lltmp;
    }
#endif /* sparc */
}

int64_t double2ll(double f)
{
#ifdef i386
    int64_t lltmp = (int64_t)f;
    if (lltmp != 0x8000000000000000) {
        return lltmp;
    } else {
        if (isnan(f)) {return 0;}
        if (f < 0) {
            return 0x8000000000000000;
        } else {
            return 0x7fffffffffffffff;
        }
    }
#endif /* i386 */

#ifdef sparc
    int64_t lltmp = (int64_t)f;
    if (lltmp != 0x7fffffffffffffff) {
        return lltmp;
    } else {
        if (isnan(f)) {return 0;}
        return lltmp;
    }
#endif /* sparc */
}
