/*
 * @(#)java_props_md.c	1.9 98/10/27
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <pwd.h>
#include <locale.h>
#ifndef ARCH
#include <sys/systeminfo.h>	/* For os_arch */
#endif
#include <sys/utsname.h>	/* For os_name and os_version */
#include <langinfo.h>           /* For nl_langinfo */
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/param.h>
#include <time.h>

#include "locale_str.h"
#include "timezone_map.h"
#include "java_props.h"

#ifdef HAVE_ALTZONE
#else
static long altzone = 0;
#endif

#ifdef ALT_CODESET_KEY
#define CODESET ALT_CODESET_KEY
#endif

/* Take an array of string pairs (map of key->value) and a string (key).
 * Examine each pair in the map to see if the first string (key) matches the
 * string.  If so, store the second string of the pair (value) in the value and
 * return 1.  Otherwise do nothing and return 0.  The end of the map is
 * indicated by an empty string at the start of a pair (key of "").
 */
static int
mapLookup(char* map[], const char* key, char** value) {
    int i;
    for (i = 0; strcmp(map[i], ""); i += 2){
        if (!strcmp(key, map[i])){
            *value = map[i + 1];
            return 1;
        }
    }
    return 0;
}

/* This function gets called very early, before VM_CALLS are setup.
 * Do not use any of the VM_CALLS entries!!!
 */
java_props_t *
GetJavaProperties()
{
    static java_props_t sprops = {0};
    char *v; /* tmp var */

    if (sprops.user_dir) {
        return &sprops;
    }

    /* tmp dir */
    sprops.tmp_dir = P_tmpdir;

    /* Printing properties */
    sprops.printerJob = "sun.awt.motif.PSPrinterJob";

    /* Java 2D properties */
    sprops.graphics_env = "sun.awt.X11GraphicsEnvironment";
    sprops.awt_toolkit = NULL;

    v = getenv("JAVA_FONTS");
    /*
    sprops.font_dir = v ? v :
    "/usr/openwin/lib/X11/fonts/Type1:/usr/openwin/lib/X11/fonts/TrueType";
    */

    /* If JAVA_FONTS is not set, the system-dependent fontpath will
     * be lazy evaluated later
     */
    sprops.font_dir = v ? v : "";

    /* os properties */
    {
        struct utsname name;
	char arch[12];
	uname(&name);
	sprops.os_name = strdup(name.sysname);
	sprops.os_version = strdup(name.release);

#ifdef ARCH
	sprops.os_arch = ARCH;
#else
	sysinfo(SI_ARCHITECTURE, arch, sizeof(arch));
	if (strcmp(arch,"sparc") == 0 ) {
	    sprops.os_arch = "sparc";
	} else if (strcmp(arch,"i386") == 0 ) {
	    /* use x86 to match the value received on win32 */
	    sprops.os_arch = "x86";
	} else if (strcmp(arch,"ppc") == 0 ) {
	    sprops.os_arch = "ppc";
	} else {
            sprops.os_arch = "Unknown";
	}
#endif
    }

    /* Determing the language, country, and encoding from the host,
     * and store these in the user.language, user.region, and
     * file.encoding system properties. */
    {
        char *lc;
        lc = setlocale(LC_CTYPE, "");
        if (lc == NULL) {
            /*
             * 'lc == null' means system doesn't support user's environment
             * variable's locale.
             */
          setlocale(LC_ALL, "C");
          sprops.language = "en";
          sprops.encoding = "8859_1";
        } else {
            /*
             * locale string format in Solaris is
             * <language name>_<region name>.<encoding name>
             * <region name> and <encoding name> are optional.
             */
            char temp[64], *language = NULL, *region = NULL, *encoding = NULL;
            char *std_language = NULL, *std_region = NULL,
                 *std_encoding = NULL;
            char region_variant[64], *variant = NULL, *std_variant = NULL;
            char *p, encoding_variant[64];
            int i, found;

            setlocale(LC_ALL, lc);
            strcpy(temp, lc);

            /* Parse the language, region, encoding, and variant from the
             * locale.  Any of the elements may be missing, but they must occur
             * in the order language_region.encoding@variant, and must be
             * preceded by their delimiter (except for language).
             *
             * If the locale name (without .encoding@variant, if any) matches
             * any of the names in the locale_aliases list, map it to the
             * corresponding full locale name.  Most of the entries in the
             * locale_aliases list are locales that include a language name but
             * no country name, and this facility is used to map each language
             * to a default country if that's possible.  It's also used to map
             * the Solaris locale aliases to their proper Java locale IDs.
             */
            if ((p = strchr(temp, '.')) != NULL) {
                strcpy(encoding_variant, p); /* Copy the leading '.' */
                *p = '\0';
            } else if ((p = strchr(temp, '@')) != NULL) {
                 strcpy(encoding_variant, p); /* Copy the leading '@' */
                 *p = '\0';
            } else {
                *encoding_variant = '\0';
            }
            
            if (mapLookup(locale_aliases, temp, &p)) {
                strcpy(temp, p);
            }
            
            language = temp;
            if ((region = strchr(temp, '_')) != NULL) {
                *region++ = '\0';
            }
            
            p = encoding_variant;
            if ((encoding = strchr(p, '.')) != NULL) {
                p[encoding++ - p] = '\0';
                p = encoding;
            }
            if ((variant = strchr(p, '@')) != NULL) {
                p[variant++ - p] = '\0';
            }

            /* Normalize the language name */
            std_language = "en";
            if (language != NULL) {
                mapLookup(language_names, language, &std_language);
            }
            sprops.language = std_language;

            /* Normalize the variant name.  Do this BEFORE handling the region name,
             * since the variant name will be incorporated into the user.region
             * property.
             */
            if (variant != NULL) {
                mapLookup(variant_names, variant, &std_variant);
            }

            /* Normalize the region name.  If there is a std_variant, then append it
             * to the region.  This applies even if there is no region, in which
             * case the empty string is used for the region.  Note that we only use
             * variants listed in the mapping array; others are ignored.
             */
            *region_variant = '\0';
            if (region != NULL) {
                std_region = region;
                mapLookup(region_names, region, &std_region);
                strcpy(region_variant, std_region);
            }
            if (std_variant != NULL) {
               strcat(region_variant, "_");
               strcat(region_variant, std_variant);
            }
            if ((*region_variant) != '\0') {
               sprops.region = strdup(region_variant);
            }

            /* Normalize the encoding name.  Note that we IGNORE the string
             * 'encoding' extracted from the locale name above.  Instead, we use the
             * more reliable method of calling nl_langinfo(CODESET).  This function
             * returns an empty string if no encoding is set for the given locale
             * (e.g., the C or POSIX locales); we use the default ISO 8859-1
             * converter for such locales.  We don't need to map from the Solaris
             * string to the Java identifier at this point because that mapping is
             * handled by the character converter alias table in
             * CharacterEncoding.java.
             */
            p = nl_langinfo(CODESET);
            std_encoding = (*p != '\0') ? p : "8859_1";
            sprops.encoding = std_encoding;
        }
    }
    
    sprops.unicode_encoding = "UnicodeBig";

    /* user properties */
    {
        struct passwd *pwent = getpwuid(getuid());
	sprops.user_name = pwent ? strdup(pwent->pw_name) : "?";
	sprops.user_home = pwent ? strdup(pwent->pw_dir) : "?";
    }

    /* User TIMEZONE */
    {
        /* In determining the host time zone, we no longer rely on tzname[].
         * Instead, we use the offset 'timezone' and the alternate daylight
         * savings flag 'daylight,' together with the ISO country code, to
         * lookup the zone in a table.
         *
         * If this lookup fails, our fallback is to synthesize and ID of the
         * form GMT[+-]hh:mm.  This allows us, in all cases (assuming the
         * 'timezone' and 'daylight' are correct) to set the Java clock
         * correctly.
         *
         * See also ../include/locale_str.h */

        char *std_timezone = NULL;
        int match, offset, i;
        enum EHasDST hasDST;
        struct tm *local_tm;

        time_t clock = time(NULL);
        tzset();
        local_tm = localtime(&clock);

        hasDST = daylight == 0 ? kStandard : kDaylight;
        offset = timezone / -60; /* Conventions reversed between UNIX & Java */

        match = -1;

        /* Loop to find the beginning of the block of entries matching this
         * offset.  Once we find it, we then start scanning through the entries
         * in that block.
         *
         * If successful, 'match' will be >= 0. 
         */
        i = 0;
        while (i < timezone_map_length
               && timezone_map[i].offset < offset) {
            ++i;
        }
        while (i < timezone_map_length
               && timezone_map[i].offset == offset) {
            /* NULL country (at end of section) matches any country; this
             * provides a default zone. E.g., in Japan, GMT-8 should be
             * recognized as PST, even though there is no JA entry for
             * GMT-8. */
            if (((timezone_map[i].country == NULL)
                 || (sprops.region != NULL
                     && (strcmp(timezone_map[i].country, sprops.region) == 0)))
                && (timezone_map[i].hasDST == kAny
                    || timezone_map[i].hasDST == hasDST)) {
                match = i;
                break;
            }
            ++i;
        }
        
        if (match >= 0) {
            std_timezone = timezone_map[match].zone;
        }
        else {
            /* If we fail to find the offset in our table (VERY unlikely) we
             * then synthesize an ID of the form GMT[+|-]hh:mm.  This offset
             * INCLUDES the effect of DST, so the correct local time can be
             * derived.  However, the local time will be incorrect as soon as we
             * cross a DST/Standard boundary, which could at any time (e.g., in
             * 5 minutes).  */
            int h, m;
            char sign = '+';

            /* Use the 'local_tm' structure, set above by the call to
             * localtime(), to determine if DST is in effect now. */
            if (local_tm->tm_isdst > 0) offset = altzone / -60;
            if (offset < 0) { offset = -offset; sign = '-'; }
            h = offset / 60;
            m = offset % 60;
            std_timezone = malloc(32); /* Allow for pathologically large h */
            sprintf(std_timezone, "GMT%c%02d:%02d", sign, h, m);
        }

        sprops.timezone = std_timezone;
    }

    /* Current directory */
    {
        char buf[MAXPATHLEN];
        getcwd(buf, sizeof(buf));
	sprops.user_dir = strdup(buf);
    }

    sprops.file_separator = "/";
    sprops.path_separator = ":";
    sprops.line_separator = "\n";

    return &sprops;
}
