/*
 * @(#)HAE_API_SolarisOS_Thread.c	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*****************************************************************************/
/*
**	HAE_API_SolarisOS_Thread.c
**
**	This provides platform-specfic frame-thread-managment functions for
**  Windows 95/NT.
**
**	Overview:
**
**	Written by Moe
**
**	Confidential-- Internal use only
**
**	History	-
**	12/1/97		Created
*/
/*****************************************************************************/

#if JAVA_THREAD
#if !JAVA_SOUND
	#error JAVA_THREAD defined when JAVA_SOUND not defined
#endif
#else


#include <thread.h>
#if JAVA_SOUND
#include <jni.h>
#endif

#include "HAE_API.h"


////////////////////////////////////////////////// MACROS:

#if _DEBUG		//CLS?? Why isn't this in some shared header file?
	#define DEBUG_STR(x)	printf(x)
#else
	#define DEBUG_STR(x)
#endif


////////////////////////////////////////////////// STATIC VARIABLES:

// Thread used to build audio frames
static thread_t			theFrameThread = NULL;
// The function that the thread calls
static HAE_FrameThreadProc	theFrameProc = NULL;


////////////////////////////////////////////////// FORWARDS:

static void*	PV_NativeFrameThreadProc(void* arg);
static void*	PV_CreateAudioContext(void);
static void		PV_DestroyAudioContext(void);


////////////////////////////////////////////////// THREAD-API FUNCTIONS:

// Create and start the frame thread
int HAE_CreateFrameThread(void* context, HAE_FrameThreadProc proc)
{
	if (proc)
	{
    long	error;

		theFrameProc = proc;
		error = thr_create(NULL, NULL, PV_SoundOutputDoubleProc, NULL,
							THR_BOUND | THR_NEW_LWP,
							&theFrameThread);
		if (error != 0) 
		{
			error = thr_setprio(gThreadProcessFrame, 127);
			if (error)
			{
				return 0;
			}
			else
			{
				DEBUG_STR("Failure to create Solaris thread, error = %d\n", error);
			}
		}
		else
		{
			DEBUG_STR("Failure to set priority of Solaris thread, error = %d\n", error);
		}
		theFrameProc = NULL;
	}
	return -1;
}
// Stop and destroy the frame thread, wait up to 1/2 second for audio to finish
int HAE_DestroyFrameThread(void* context)
{
	if (theFrameThread)
	{
		thr_kill(theFrameThread, 0);
		theFrameThread = NULL;
		theFrameProc = NULL;
		return 0;
	}
	DEBUG_STR("PV_NativeFrameThreadProc() called when no thread started\n");
	return -1;
}
// Make the frame thread sleep for the given number of milliseconds
int HAE_SleepFrameThread(void* context, long msec)
{
	struct timespec request, actual;
	int sleep_ret;

	request.tv_sec = 0;
	request.tv_nsec = msecs * 1000;
	sleep_ret = nanosleep(&request, &actual);

	return  0;
}

static void* PV_NativeFrameThreadProc(void* arg)
{
	if (theFrameProc)
	{
		(*theFrameProc)(PV_CreateAudioContext());
		PV_DestroyAudioContext();
	}
	else
	{
		DEBUG_STR("PV_NativeFrameThreadProc() called when thread supposed to be dead!");
	}
	return NULL;
}


////////////////////////////////////////////////// HELPER FUNCTIONS:

#if JAVA_SOUND

static JavaVM* PV_GetCurrentVM(void)
{
JavaVM*		vmArray[2];
jsize		count;
jint		result;

	count = 0;
	result = JNI_GetCreatedJavaVMs(vmArray, 2, &count);	// ask for 2
	if ((result == 0) && (count >= 1))						// error if not just 1
	{
		if (count > 1)
		{
			DEBUG_STR("WARNING:  JNI_GetCreatedJavaVMs() returned >1 VMs\n");
		}
		return vmArray[0];
	}
	DEBUG_STR("JNI_GetCreatedJavaVMs() returned error\n");
	return NULL;
}
static void* PV_CreateAudioContext(void)
{
JavaVM*		const vm = PV_GetCurrentVM();

	if (vm)
	{
	JNIEnv*				env;
	JDK1_1AttachArgs	args;	// whatever
	jint				result;

		result = (*vm)->AttachCurrentThread(vm, &env, &args);
		if (result >= 0)
		{
			return (void*)env;
		}
		DEBUG_STR("AttachCurrentThread() failed\n");
	}
	return NULL;
}
static void PV_DestroyAudioContext(void)
{
JavaVM*		const vm = PV_GetCurrentVM();

	if (vm)
	{
	jint		result;

		result = (*vm)->DetachCurrentThread(vm);
		if (result < 0)
		{
			DEBUG_STR("DetachCurrentThread() failed\n");
		}
	}
}

#else

static void * PV_CreateAudioContext(void)
{
	return NULL;
}
static void PV_DestroyAudioContext(void)
{
}

#endif


#endif	// !JAVA_THREAD
// EOF of HAE_API_WinOS_Thread.c
