/*
 * @(#)awt_AWTEvent.c	1.4 98/10/08
 * 
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road, 
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * CopyrightVersion 1.2
 * 
 */
/*
 * Implements the native code for the java.awt.AWTEvent class
 * and all of the classes in the java.awt.event package.
 *
 * THIS FILE DOES NOT IMPLEMENT ANY OF THE OBSOLETE java.awt.Event
 * CLASS. SEE awt_Event.[ch] FOR THAT CLASS' IMPLEMENTATION.
 */

#include "awt_p.h"
#include "java_awt_AWTEvent.h"
#include "java_awt_event_InputEvent.h"
#include "java_awt_event_KeyEvent.h"
#include "jni_util.h"

#include "canvas.h"
#include "awt_AWTEvent.h"

struct AWTEventIDs awtEventIDs;
struct InputEventIDs inputEventIDs;
struct KeyEventIDs keyEventIDs;

JNIEXPORT void JNICALL
Java_java_awt_AWTEvent_initIDs(JNIEnv *env, jclass cls)
{
    awtEventIDs.data = (*env)->GetFieldID(env, cls, "data", "J");
    awtEventIDs.consumed = (*env)->GetFieldID(env, cls, "consumed", "Z");
    awtEventIDs.id = (*env)->GetFieldID(env, cls, "id", "I");
}

/*
 * Class:     java_awt_AWTEvent
 * Method:    copyDataFieldInto
 * Signature: ()V
 *
 * Allocates a copy of the data instance variable and copies it
 * from thisEvent into thatEvent. If data is null, thatEvent's
 * data field is also set to NULL.
 */
JNIEXPORT void JNICALL
Java_java_awt_AWTEvent_copyDataFieldInto
  (JNIEnv *env, jobject thisEvent, jobject thatEvent) {

    XEvent *thisData = NULL;
    XEvent *thatData = NULL;
    AWT_LOCK();
    thisData = (void *)(*env)->GetLongField(env, thisEvent, awtEventIDs.data);
    thatData = (void *)(*env)->GetLongField(env, thatEvent, awtEventIDs.data);
    if ((thisData != NULL) || (thatData != NULL)) {

	// we have work to do!
        if (thatData != NULL) {
	    if (thatData != thisData) {
	        free(thatData);
	    }
	    thatData = NULL;
        }
        if (thisData != NULL) {
	    thatData = awt_copyXEvent(thisData);
        }
        (*env)->SetLongField(env, thatEvent, awtEventIDs.data, (jlong)thatData);
    }
    AWT_UNLOCK();
} /* AWTEvent_copyDataFieldInto() */

/*
 * Class:     java_awt_AWTEvent
 * Method:    freeNativeData
 * Signature: ()V
 *
 * Frees any native data allocated by this object, and sets all
 * references to that data to null. This is normally invoked when
 * the instance is finalized
 */
JNIEXPORT void JNICALL
Java_java_awt_AWTEvent_freeNativeData
  (JNIEnv *env, jobject event) {

    void *data = NULL;
    data = (void *)(*env)->GetLongField(env, event, awtEventIDs.data);
    if (data != NULL) {
        AWT_LOCK();
        free((void *) data);
        (*env)->SetLongField(env, event, awtEventIDs.data, 0);
        AWT_UNLOCK();
    }
} /* AWTEvent_freeNativeData() */
JNIEXPORT void JNICALL
Java_java_awt_event_InputEvent_initIDs(JNIEnv *env, jclass cls)
{
    inputEventIDs.modifiers = (*env)->GetFieldID(env, cls, "modifiers", "I");
}

JNIEXPORT void JNICALL
Java_java_awt_event_KeyEvent_initIDs(JNIEnv *env, jclass cls)
{
    keyEventIDs.keyCode = (*env)->GetFieldID(env, cls, "keyCode", "I");
    keyEventIDs.keyChar = (*env)->GetFieldID(env, cls, "keyChar", "C");
}
