/*
 * @(#)awt_Button.c	1.39 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"

#include <jni.h>
#include <jni_util.h>
#include "multi_font.h"

#include "awt_Component.h"

extern struct MComponentPeerIDs mComponentPeerIDs;

/*
 * When the -jni switch is thrown, these headers can be deleted.
 */
#include "java_awt_Button.h"
#include "sun_awt_motif_MButtonPeer.h"
#include "sun_awt_motif_MComponentPeer.h"

/* fieldIDs for Button fields that may be accessed from C */
static struct ButtonIDs {
    jfieldID label;
} buttonIDs;

/*
 * Class:     java_awt_Button
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for Button.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_Button_initIDs
  (JNIEnv *env, jclass cls)
{
    buttonIDs.label =
      (*env)->GetFieldID(env, cls, "label", "Ljava/lang/String;");
}

/*
 * client_data is MButtonPeer instance
 */
static void
Button_callback (Widget w,
                 XtPointer client_data,
                 XmPushButtonCallbackStruct * call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    JNU_CallMethodByName (env, NULL, (jobject) client_data, "action", "()V");
    if ((*env)->ExceptionOccurred (env)) {
        (*env)->ExceptionDescribe (env);
        (*env)->ExceptionClear (env);
    }
}

/*
 * Class:     sun_awt_motif_MButtonPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MButtonPeer_create
  (JNIEnv * env, jobject this, jobject parent)
{
    jobject target;
    jobject label;
    struct ComponentData *cdata;
    struct ComponentData *wdata;
    char *clabel;
    Pixel bg;
    XmString mfstr = NULL;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef (env, this);
    jobject font = awtJNI_GetFont (env, this);
    jboolean IsMultiFont = awtJNI_IsMultiFont (env, font);

    AWT_LOCK ();

    if (JNU_IsNull (env, parent)) {
        JNU_ThrowNullPointerException (env, "NullPointerException");
        AWT_UNLOCK ();

        return;
    }
    target = (*env)->GetObjectField (env, this, mComponentPeerIDs.target);
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env, parent, mComponentPeerIDs.pData);

    if (JNU_IsNull (env, target) || wdata == NULL) {
        JNU_ThrowNullPointerException (env, "NullPointerException");
        AWT_UNLOCK ();

        return;
    }
    cdata = ZALLOC (ComponentData);
    if (cdata == NULL) {
        JNU_ThrowOutOfMemoryError (env, "OutOfMemoryError");
        AWT_UNLOCK ();
        return;
    }
    (*env)->SetLongField(env, this, mComponentPeerIDs.pData, (jlong) cdata);

    XtVaGetValues (wdata->widget, XmNbackground, &bg, NULL);

    label =
      (*env)->GetObjectField (env, target, buttonIDs.label);

    if (IsMultiFont) {
        /*
         * We don't use makeCString() function here.
         * We create Motif multi-font compound string to display
         * unicode on the platform which is not spporting unicode.
         */
        if (JNU_IsNull (env, label) || ((*env)->GetStringLength (env, label) == 0)) {
            mfstr = XmStringCreateSimple ("");
        } else {
            mfstr = awtJNI_MakeMultiFontString (env, label, font);
        }

        cdata->widget = XtVaCreateManagedWidget
            ("", xmPushButtonWidgetClass,
             wdata->widget,
             XmNlabelString, mfstr,
             XmNrecomputeSize, False,
             XmNbackground, bg,
             XmNhighlightOnEnter, False,
             XmNshowAsDefault, 0,
             XmNdefaultButtonShadowThickness, 0,
             XmNmarginTop, 0,
             XmNmarginBottom, 0,
             XmNmarginLeft, 0,
             XmNmarginRight, 0,
             XmNuserData, (XtPointer) globalRef,
             NULL);
	if (mfstr != NULL) {
	    XmStringFree(mfstr);
	    mfstr = NULL;
	}

    } else {
        if (JNU_IsNull (env, label)) {
            clabel = "";
        } else {
            clabel = (char *) JNU_GetStringPlatformChars (env, label, NULL);

            if (clabel == 0) {        /* Exception? */
                AWT_UNLOCK ();
                return;
            }
        }

        cdata->widget = XtVaCreateManagedWidget
            (clabel, xmPushButtonWidgetClass,
             wdata->widget,
             XmNrecomputeSize, False,
             XmNbackground, bg,
             XmNhighlightOnEnter, False,
             XmNshowAsDefault, 0,
             XmNdefaultButtonShadowThickness, 0,
             XmNmarginTop, 0,
             XmNmarginBottom, 0,
             XmNmarginLeft, 0,
             XmNmarginRight, 0,
             XmNuserData, (XtPointer) globalRef,
             NULL);

        if (clabel != "") {
            JNU_ReleaseStringPlatformChars (env, label, (const char *) clabel);;
        }
    }

    XtSetMappedWhenManaged (cdata->widget, False);
    XtAddCallback (cdata->widget,
                   XmNactivateCallback,
                   (XtCallbackProc) Button_callback,
                   (XtPointer) globalRef);

    AWT_UNLOCK ();
}

/*
 * Class:     sun_awt_motif_MButtonPeer
 * Method:    setLabel
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MButtonPeer_setLabel
  (JNIEnv * env, jobject this, jstring label)
{
    struct ComponentData *wdata;
    char *clabel;
    XmString xim;

    AWT_LOCK ();

    wdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);

    if (wdata == 0) {
        JNU_ThrowNullPointerException (env, "NullPointerException");
        AWT_UNLOCK ();
        return;
    }
    if (JNU_IsNull (env, label) || ((*env)->GetStringLength (env, label) == 0)) {
        xim = XmStringCreateSimple ("");
    } else {
        jobject font = awtJNI_GetFont (env, this);

        if (awtJNI_IsMultiFont (env, font)) {
            xim = awtJNI_MakeMultiFontString (env, label, font);
        } else {
            if (JNU_IsNull (env, label)) {
                clabel = "";
            } else {
                clabel = (char *) JNU_GetStringPlatformChars (env, label, NULL);

                if (clabel == 0) {      /* Exception? */
                    AWT_UNLOCK ();
                    return;
                }
            }

            xim = XmStringCreateLtoR (clabel, "labelFont");

            if (clabel != "") {
                JNU_ReleaseStringPlatformChars (env, label, (const char *) clabel);;
            }
        }
    }

    XtVaSetValues (wdata->widget, XmNlabelString, xim, NULL);
    XmStringFree (xim);
    AWT_FLUSH_UNLOCK ();
}
