/*
 * @(#)awt_Canvas.c	1.29 98/07/13
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "java_awt_Canvas.h"
#include "sun_awt_motif_MCanvasPeer.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "color.h"
#include "canvas.h"
#include "awt_util.h"

#include "awt_Component.h"
#include "awt_GraphicsEnv.h"

#include <jni.h>
#include <jni_util.h>
#include "multi_font.h"

extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct X11GraphicsConfigIDs x11GraphicsConfigIDs;

/*
 * Class:     sun_awt_motif_MCanvasPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MCanvasPeer_create
  (JNIEnv * env, jobject this, jobject parent)
{
    jobject  component_object, gc_object;
    jclass   canvas_class, MPP_class;
    jfieldID component_field, gc_field, vid_field;
    AwtData  *awtData;

    struct CanvasData *wdata;
    struct CanvasData *cdata;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();
    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    cdata = (struct CanvasData *)
      (*env)->GetLongField(env, parent, mComponentPeerIDs.pData);
    if (cdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    wdata = ZALLOC(CanvasData);
    if (wdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    (*env)->SetLongField(env, this, mComponentPeerIDs.pData,(jlong)wdata);

    /* MPanelPeer extends MCanvasPeer but gc_object fails */
    /*  test for MPanelPeer class and do nothing if found */
    MPP_class = (*env)->FindClass(env, "sun/awt/motif/MPanelPeer");
    if ((*env)->IsInstanceOf(env, this, MPP_class)) {
        awtData = defaultConfig;
    }
    else {
        /* Component object from "this" object */
        component_object = (*env)->GetObjectField(env, this, 
					    mComponentPeerIDs.target);
        /* need Canvas class - will "cast" Component to be a Canvas class */
        canvas_class = (*env)->FindClass(env, "java/awt/Canvas");

        /* GraphicsConfiguration object of Canvas */
        gc_field = (*env)->GetFieldID(env, canvas_class,
                    "graphicsConfig", "Ljava/awt/GraphicsConfiguration;");
        gc_object = (*env)->GetObjectField(env, component_object, gc_field);

        if (gc_object != 0) {
	    /* Set graphicsConfig field of MComponentPeer */
	    (*env)->SetObjectField (env, this, mComponentPeerIDs.graphicsConfig,
				    gc_object);
            awtData = (AwtDataPtr) (*env)->GetLongField(env, gc_object, 
					      x11GraphicsConfigIDs.aData);
        }
        else {
            /* Canvas was not constructed with a GraphicsConfiguration object */
            awtData = defaultConfig;
        }
    }

    wdata->comp.widget = awt_canvas_create((XtPointer) globalRef,
                                           cdata->comp.widget,
					   "",
                                           1, 1, False, NULL, awtData);
    XtVaSetValues(wdata->comp.widget,
                  XmNinsertPosition, awt_util_insertCallback,
                  NULL);

    wdata->flags = 0;
    wdata->shell = cdata->shell;
    wdata->comp.awtData = awtData;

    AWT_UNLOCK();
}

