/*
 * @(#)awt_Checkbox.c	1.33 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "sun_awt_motif_MCheckboxPeer.h"
#include "java_awt_Checkbox.h"
#include "java_awt_CheckboxGroup.h"

#include "awt_Component.h"

#include "multi_font.h"
#include <jni.h>
#include <jni_util.h>

extern struct MComponentPeerIDs mComponentPeerIDs;

/* fieldIDs for Checkbox fields that may be accessed from C */
static struct CheckboxIDs {
    jfieldID label;
} checkboxIDs;

/*
 * Class:     java_awt_Checkbox
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for Checkbox.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_Checkbox_initIDs
  (JNIEnv *env, jclass cls)
{
    checkboxIDs.label =
      (*env)->GetFieldID(env, cls, "label", "Ljava/lang/String;");
}

/*
 * client_data is MCheckboxPeer instance pointer
 */
static void
Toggle_callback(Widget w,
                XtPointer client_data,
                XmAnyCallbackStruct * call_data)
{
    Boolean state;
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    XtVaGetValues(w, XmNset, &state, NULL);

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "action", "(Z)V", state);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * Class:     sun_awt_motif_MCheckboxPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MCheckboxPeer_create
  (JNIEnv * env, jobject this, jobject parent)
{
    jobject target;
    struct ComponentData *bdata;
    struct ComponentData *wdata;
    char *clabel;
    Arg args[10];
    int argc;
    Dimension w, h;
    jobject label;
    XmString mfstr = NULL;
    jobject font = awtJNI_GetFont(env, this);
    jboolean isMultiFont = awtJNI_IsMultiFont(env, font);
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env, parent, mComponentPeerIDs.pData);

    if (JNU_IsNull(env, target) || wdata == NULL) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    bdata = ZALLOC(ComponentData);
    if (bdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();

        return;
    }
    (*env)->SetLongField(env, this, mComponentPeerIDs.pData, (jlong)bdata);

    argc = 0;
    XtSetArg(args[argc], XmNrecomputeSize, False);
    argc++;
    XtSetArg(args[argc], XmNvisibleWhenOff, True);
    argc++;
    XtSetArg(args[argc], XmNtraversalOn, True);
    argc++;
    XtSetArg(args[argc], XmNspacing, 0);
    argc++;
    XtSetArg(args[argc], XmNuserData, (XtPointer) globalRef);
    argc++;

    label = (*env)->GetObjectField(env, target, checkboxIDs.label);

    if (isMultiFont) {
        /*
         * We don't use makeCString() function here.
         * We create Motif multi-font compound string to display
         * unicode on the platform which is not spporting unicode.
         */
        if (JNU_IsNull(env, label) || ((*env)->GetStringLength(env, label) == 0)) {
            mfstr = XmStringCreateSimple("");
        } else {
            mfstr = awtJNI_MakeMultiFontString(env, label, font);
        }

        XtSetArg(args[argc], XmNlabelString, mfstr);
        argc++;
        bdata->widget = XmCreateToggleButton(wdata->widget, "", args, argc);

	if (mfstr != NULL) {
	    XmStringFree(mfstr);
	    mfstr = NULL;
	}
    } else {
        if (JNU_IsNull(env, label)) {
            clabel = "";
        } else {
            clabel = (char *) JNU_GetStringPlatformChars(env, label, NULL);

            if (clabel == 0) {        /* Exception? */
                AWT_UNLOCK();
                return;
            }
        }

        bdata->widget = XmCreateToggleButton(wdata->widget, clabel, args, argc);

        if (clabel != "") {
            JNU_ReleaseStringPlatformChars(env, label, (const char *) clabel);;
        }
    }

    XtAddCallback(bdata->widget,
                  XmNvalueChangedCallback,
                  (XtCallbackProc) Toggle_callback,
                  (XtPointer) globalRef);

    XtSetMappedWhenManaged(bdata->widget, False);
    XtManageChild(bdata->widget);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MCheckboxPeer
 * Method:    setLabel
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MCheckboxPeer_setLabel
  (JNIEnv * env, jobject this, jstring label)
{
    struct ComponentData *wdata;
    char *clabel;
    XmString xim;
    jobject font;

    AWT_LOCK();

    wdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (JNU_IsNull(env, label) || ((*env)->GetStringLength(env, label) == 0)) {
        xim = XmStringCreateSimple("");
    } else {
        font = awtJNI_GetFont(env, this);

        if (awtJNI_IsMultiFont(env, font)) {
            xim = awtJNI_MakeMultiFontString(env, label, font);
        } else {
            clabel = (char *) JNU_GetStringPlatformChars(env, label, NULL);

            if (clabel == 0) {
                AWT_UNLOCK();
                return;
            }
            xim = XmStringCreateLtoR(clabel, "labelFont");

            JNU_ReleaseStringPlatformChars(env, label, (const char *) clabel);;
        }
    }

    XtVaSetValues(wdata->widget, XmNlabelString, xim, NULL);
    XmStringFree(xim);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MCheckboxPeer
 * Method:    setState
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MCheckboxPeer_setState
  (JNIEnv * env, jobject this, jboolean state)
{
    struct ComponentData *bdata;

    AWT_LOCK();

    bdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);

    if (bdata == 0 || bdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(bdata->widget, XmNset, (Boolean) state, NULL);
    AWT_FLUSH_UNLOCK();
}


/*
 * Class:     sun_awt_motif_MCheckboxPeer
 * Method:    setCheckboxGroup
 * Signature: (Ljava/awt/CheckboxGroup;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MCheckboxPeer_setCheckboxGroup
  (JNIEnv * env, jobject this, jobject group)
{
    struct ComponentData *bdata;

    AWT_LOCK();

    bdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);

    if (bdata == 0 || bdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (JNU_IsNull(env, group)) {
        XtVaSetValues(bdata->widget,
                      XmNindicatorType, XmN_OF_MANY,
                      NULL);
    } else {
        XtVaSetValues(bdata->widget,
                      XmNindicatorType, XmONE_OF_MANY,
                      NULL);
    }

    AWT_FLUSH_UNLOCK();
}
