/*
 * @(#)awt_Choice.c	1.51 98/07/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "java_awt_Component.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "sun_awt_motif_MChoicePeer.h"

#include "awt_Component.h"

#include "multi_font.h"
#include <jni.h>
#include <jni_util.h>

extern struct MComponentPeerIDs mComponentPeerIDs;

static void
Choice_callback(Widget menu_item,
                jobject this,
                XmAnyCallbackStruct * cbs)
{
    jint index;
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    XtVaGetValues(menu_item, XmNuserData, &index, NULL);
    /* index stored in user-data is 1-based instead of 0-based because */
    /* of a bug in XmNuserData */
    index--;

    JNU_CallMethodByName(env, NULL, this, "action", "(I)V", index);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_create
  (JNIEnv * env, jobject this, jobject parent)
{
    struct ChoiceData *odata;
    struct ComponentData *wdata;
    Arg args[30];
    int argc;
    Pixel bg;
    Pixel fg;
    Widget label;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);

    if (wdata == NULL) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    odata = ZALLOC(ChoiceData);
    if (!odata) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong)odata);

    odata->items = 0;
    odata->maxitems = 0;
    odata->n_items = 0;

    XtVaGetValues(wdata->widget, XmNbackground, &bg, NULL);
    XtVaGetValues(wdata->widget, XmNforeground, &fg, NULL);

    argc = 0;
    XtSetArg(args[argc], XmNx, 0);
    argc++;
    XtSetArg(args[argc], XmNy, 0);
    argc++;
    XtSetArg(args[argc], XmNvisual, defaultConfig->awt_visInfo.visual);
    argc++;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, fg);
    argc++;
    odata->menu = XmCreatePulldownMenu(wdata->widget, "pulldown", args, argc);

    argc = 0;
    XtSetArg(args[argc], XmNx, 0);
    argc++;
    XtSetArg(args[argc], XmNy, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNrecomputeSize, False);
    argc++;
    XtSetArg(args[argc], XmNresizeHeight, False);
    argc++;
    XtSetArg(args[argc], XmNresizeWidth, False);
    argc++;
    XtSetArg(args[argc], XmNspacing, False);
    argc++;
    XtSetArg(args[argc], XmNborderWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNnavigationType, XmTAB_GROUP);
    argc++;
    XtSetArg(args[argc], XmNtraversalOn, True);
    argc++;
    XtSetArg(args[argc], XmNorientation, XmVERTICAL);
    argc++;
    XtSetArg(args[argc], XmNadjustMargin, False);
    argc++;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, fg);
    argc++;
    XtSetArg(args[argc], XmNsubMenuId, odata->menu);
    argc++;
    XtSetArg(args[argc], XmNuserData, (XtPointer) globalRef);
    argc++;
    odata->comp.widget = XmCreateOptionMenu(wdata->widget, "", args, argc);

    label = XmOptionLabelGadget(odata->comp.widget);
    if (label != 0) {
        XtUnmanageChild(label);
    }
    XtSetMappedWhenManaged(odata->comp.widget, False);
    XtManageChild(odata->comp.widget);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    addItem
 * Signature: (Ljava/lang/String;I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_addItem
  (JNIEnv *env, jobject this, jstring item, jint index)
{
    char *citem = NULL;
    struct ChoiceData *odata;
    Widget bw;
    Arg args[10];
    int argc;
    Pixel bg;
    Pixel fg;

    XmString mfstr = NULL;
    XmFontList fontlist;
    jobject font = awtJNI_GetFont(env, this);
    Boolean IsMultiFont = awtJNI_IsMultiFont(env, font);

    if (JNU_IsNull(env, item)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();

    if (IsMultiFont) {
        mfstr = awtJNI_MakeMultiFontString(env, item, font);
        fontlist = awtJNI_GetFontList(env, font);
    } else {
        citem = (char *) JNU_GetStringPlatformChars(env, item, NULL);
    }

    odata = (struct ChoiceData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (odata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        if (citem)
            JNU_ReleaseStringPlatformChars(env, item, (const char *) citem);
        AWT_UNLOCK();

        return;
    }
    if (odata->maxitems == 0 || index > odata->maxitems - 1) {
        odata->maxitems += 20;
        if (odata->n_items > 0) {
            /* grow the list of items */
            odata->items = (Widget *)
                realloc((void *) (odata->items)
                        ,sizeof(Widget) * odata->maxitems);
        } else {
            odata->items = (Widget *) malloc(sizeof(Widget) * odata->maxitems);
        }
        if (odata->items == 0) {
            JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
            if (citem)
                JNU_ReleaseStringPlatformChars(env, item, (const char *) citem);
            AWT_UNLOCK();
            return;
        }
    }
    XtVaGetValues(odata->comp.widget, XmNbackground, &bg, NULL);
    XtVaGetValues(odata->comp.widget, XmNforeground, &fg, NULL);

    argc = 0;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, fg);
    argc++;
    /* XXX: XmNuserData doesn't seem to work when passing in zero */
    /* so we increment the index before passing it in. */
    XtSetArg(args[argc], XmNuserData, index + 1);
    argc++;

    if (IsMultiFont) {
        XtSetArg(args[argc], XmNfontList, fontlist);
        argc++;
        XtSetArg(args[argc], XmNlabelString, mfstr);
        argc++;
        bw = XmCreatePushButton(odata->menu, "", args, argc);
	if (mfstr != NULL) {
	    XmStringFree(mfstr);
	    mfstr = NULL;
	}
    } else {
        bw = XmCreatePushButton(odata->menu, citem, args, argc);
    }

    if (citem)
        JNU_ReleaseStringPlatformChars(env, item, (const char *) citem);

    XtAddCallback(bw,
                  XmNactivateCallback,
                  (XtCallbackProc) Choice_callback,
                  (XtPointer) (*env)->
		    GetLongField(env, this, mComponentPeerIDs.jniGlobalRef));
    odata->items[index] = bw;
    odata->n_items++;
    XtManageChild(bw);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    select
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_select
  (JNIEnv *env, jobject this, jint index)
{
    struct ChoiceData *odata;

    AWT_LOCK();

    odata = (struct ChoiceData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (odata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (index > odata->n_items || index < 0) {
        JNU_ThrowIllegalArgumentException(env, "IllegalArgumentException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(odata->comp.widget,
                  XmNmenuHistory, odata->items[index],
                  NULL);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    setFont
 * Signature: (Ljava/awt/Font;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_setFont
  (JNIEnv *env, jobject this, jobject f)
{
    struct ChoiceData *cdata;
    struct FontData *fdata;
    XmFontList fontlist;
    char *err;

    if (JNU_IsNull(env, f)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();

    fdata = awtJNI_GetFontData(env, f, &err);
    if (fdata == 0) {
        JNU_ThrowInternalError(env, err);
        AWT_UNLOCK();
        return;
    }
    cdata = (struct ChoiceData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (awtJNI_IsMultiFont(env, f)) {
        fontlist = awtJNI_GetFontList(env, f);
    } else {
        fontlist = XmFontListCreate(fdata->xfont, "labelFont");
    }

    if (fontlist != NULL) {
        int i;

        XtVaSetValues(cdata->comp.widget,
                      XmNfontList, fontlist,
                      NULL);
        XtVaSetValues(cdata->menu,
                      XmNfontList, fontlist,
                      NULL);
        for (i = 0; i < cdata->n_items; i++) {
            XtVaSetValues(cdata->items[i],
                          XmNfontList, fontlist,
                          NULL);
        }

        XmFontListFree(fontlist);
    } else {
        JNU_ThrowNullPointerException(env, "NullPointerException");
    }

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    setBackground
 * Signature: (Ljava/awt/Color;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_setBackground
  (JNIEnv *env, jobject this, jobject c)
{
    struct ChoiceData *bdata;
    Pixel bg;
    Pixel fg;
    int i;

    if (JNU_IsNull(env, c)) {
        JNU_ThrowNullPointerException(env, "NullPointerException: null color");
        return;
    }
    AWT_LOCK();

    bdata = (struct ChoiceData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (bdata == 0 || bdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    /* Get background color */
    bg = awtJNI_GetColor(env, c);

    /*
       XmChangeColor(), in addtion to changing the background and
       selection colors, also changes the foreground color to be
       what it thinks should be. However, we want to use the color
       that gets set by setForeground() instead. We therefore need to
       save the current foreground color here, and then set it again
       after the XmChangeColor() occurs.
    */
    XtVaGetValues(bdata->comp.widget, XmNforeground, &fg, NULL);

    /* Set color */
    XmChangeColor(bdata->comp.widget, bg);
    XtVaSetValues(bdata->comp.widget, XmNforeground, fg, NULL);

    XmChangeColor(bdata->menu, bg);
    XtVaSetValues(bdata->menu, XmNforeground, fg, NULL);

    for (i = 0; i < bdata->n_items; i++) {
        XmChangeColor(bdata->items[i], bg);
        XtVaSetValues(bdata->items[i], XmNforeground, fg, NULL);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    setForeground
 * Signature: (Ljava/awt/Color;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_setForeground
  (JNIEnv *env, jobject this, jobject c)
{
    struct ChoiceData *bdata;
    Pixel color;
    int i;

    if (JNU_IsNull(env, c)) {
        JNU_ThrowNullPointerException(env, "NullPointerException: null color");
        return;
    }
    AWT_LOCK();

    bdata = (struct ChoiceData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (bdata == 0 || bdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    color = awtJNI_GetColor(env, c);

    XtVaSetValues(bdata->comp.widget, XmNforeground, color, NULL);
    XtVaSetValues(bdata->menu, XmNforeground, color, NULL);
    for (i = 0; i < bdata->n_items; i++) {
        XtVaSetValues(bdata->items[i], XmNforeground, color, NULL);
    }

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    pReshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_pReshape
  (JNIEnv *env, jobject this, jint x, jint y, jint w, jint h)
{
    struct ComponentData *cdata;
    Widget button;

    AWT_LOCK();

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    button = XmOptionButtonGadget(cdata->widget);
    awt_util_reshape(cdata->widget, x, y, w, h);
    awt_util_reshape(button, x, y, w, h);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MChoicePeer
 * Method:    remove
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MChoicePeer_remove
  (JNIEnv *env, jobject this, jint index)
{
    struct ChoiceData *cdata;
    Widget selected;
    jint i;

    AWT_LOCK();

    cdata = (struct ChoiceData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (index < 0 || index > cdata->n_items) {
        JNU_ThrowIllegalArgumentException(env, "IllegalArgumentException");
        AWT_UNLOCK();
        return;
    }
    XtUnmanageChild(cdata->items[index]);
    awt_util_consumeAllXEvents(cdata->items[index]);
    XtDestroyWidget(cdata->items[index]);
    for (i = index; i < cdata->n_items-1; i++) {
        cdata->items[i] = cdata->items[i + 1];
        /* need to reset stored index value, (adding 1 to disambiguate it */
        /* from an arg list terminator)                                   */
        /* bug fix 4079027 robi.khan@eng                                  */
        XtVaSetValues(cdata->items[i],  XmNuserData, i+1, NULL);
    }
    cdata->items[cdata->n_items-1] = NULL;
    cdata->n_items--;

    AWT_UNLOCK();
}
