/*
 * @(#)awt_Component.c	1.89 98/09/22
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "canvas.h"
#include "awt_AWTEvent.h"
#include "VDrawingArea.h"
#include "java_awt_Color.h"
#include "java_awt_Cursor.h"
#include "java_awt_Font.h"
#include "java_awt_Point.h"
#include "java_awt_Component.h"
#include "java_awt_AWTEvent.h"
#include "java_awt_event_KeyEvent.h"
#include "sun_awt_motif_MComponentPeer.h"
#include <X11/cursorfont.h>


#include "multi_font.h"
#include "jni.h"
#include "jni_util.h"

/* CanvasType widgets: Frame, Dialog, Window, Panel, Canvas, 
 *                     &  all lightweights (Component, Container) 
 */
#define IsCanvasTypeWidget(w) \
        XtIsSubclass(w, xmDrawingAreaWidgetClass) ||\
        XtIsSubclass(w, vDrawingAreaClass)

#include "awt_Component.h"

#include "awt_AWTEvent.h"
#include "awt_Cursor.h"

extern struct CursorIDs cursorIDs;
extern struct X11CustomCursorIDs x11CustomCursorIDs;

/* fieldIDs for Component fields that may be accessed from C */
struct ComponentIDs componentIDs;

/*
 * Class:     java_awt_Component
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for Component.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_Component_initIDs
  (JNIEnv *env, jclass cls)
{
    componentIDs.x = (*env)->GetFieldID(env, cls, "x", "I");
    componentIDs.y = (*env)->GetFieldID(env, cls, "y", "I");
    componentIDs.width = (*env)->GetFieldID(env, cls, "width", "I");
    componentIDs.height = (*env)->GetFieldID(env, cls, "height", "I");
    componentIDs.isPacked = (*env)->GetFieldID(env, cls, "isPacked", "Z");
    componentIDs.peer =
      (*env)->GetFieldID(env, cls, "peer", "Ljava/awt/peer/ComponentPeer;");
    componentIDs.background =
      (*env)->GetFieldID(env, cls, "background", "Ljava/awt/Color;");
    componentIDs.foreground =
      (*env)->GetFieldID(env, cls, "foreground", "Ljava/awt/Color;");

    /* Use _NoClientCode() methods for trusted methods, so that we
     *  know that we are not invoking client code on trusted threads
     */
    componentIDs.getParent = 
      (*env)->GetMethodID(env, cls, "getParent_NoClientCode",
						 "()Ljava/awt/Container;");
}

/* fieldIDs for MComponentPeer fields that may be accessed from C */
struct MComponentPeerIDs mComponentPeerIDs;

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MComponentPeer.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_initIDs
  (JNIEnv *env, jclass cls)
{
    mComponentPeerIDs.pData = (*env)->GetFieldID(env, cls, "pData", "J");
    mComponentPeerIDs.cursorSet =
      (*env)->GetFieldID(env, cls, "cursorSet", "I");
    mComponentPeerIDs.target =
      (*env)->GetFieldID(env, cls, "target", "Ljava/awt/Component;");
    mComponentPeerIDs.jniGlobalRef =
      (*env)->GetFieldID(env, cls, "jniGlobalRef", "J");
    mComponentPeerIDs.graphicsConfig =
      (*env)->GetFieldID(env, cls, "graphicsConfig", 
			 "Lsun/awt/X11GraphicsConfig;");
}

/* field and method IDs for java.awt.Container. */
struct ContainerIDs containerIDs;

/*
 * Class:     java_awt_Container
 * Method:    initIDs
 * Signature: ()V
 */
/* This function gets called from the static initializer for java.awt.Container
   to initialize the fieldIDs for fields that may be accessed from C */
JNIEXPORT void JNICALL
Java_java_awt_Container_initIDs
  (JNIEnv *env, jclass cls)
{
    containerIDs.getComponents = 
      (*env)->GetMethodID(env, cls, "getComponents_NoClientCode",
						 "()[Ljava/awt/Component;");
}

/*
 * Fix for 4090493.  When Motif computes indicator size, it uses
 * (effectively) XmTextExtents, so the size of the indicator depends
 * on the text of the label.  The side effect is that if the label
 * text is rendered using different platform fonts (for a single Java
 * logical font) the display is inconsistent.  E.g. for 12pt font
 * English label will have a check mark, while Japanese label will
 * not, because underlying X11 fonts have different metrics.
 *
 * The fix is to override Motif calculations for the indicatorSize and
 * compute it ourselves based on the font metrics for all the platform
 * fonts given Java font maps onto.  Every time we set XmNfontList we
 * should set XmNindicatorSize as well.
 * 
 * The logic is in awt_computeIndicatorSize that just compute the
 * arithmetic mean of platform fonts by now.  HIE should take a look
 * at this.
 */

struct changeFontInfo {
    XmFontList fontList;	/* value to set */
    Boolean isMultiFont;	/* only need to compute for multifont */
    struct FontData *fontData;	/* need this to compute indicator size */
    Dimension indSize;		/* computed once by changeFont */
};

static void 
changeFont(Widget w, void *info)
{
    struct changeFontInfo *f = (struct changeFontInfo *)info;

    /* Fix for 4090493. */
    if (f->isMultiFont && XmIsToggleButton(w)) {
	Dimension indSize;

	/* Compute indicator size if first time through.  Note, that
	   ToggleButtons that are children of menus live in different
	   hieraqrchy (MenuComponent), so we don't check for this case
	   here.  In fact, the only time the XmNfontList is set on
	   MCheckboxMenuItemPeer widget is when it's created. */
	if (f->indSize == 0)
	    f->indSize = awt_computeIndicatorSize(f->fontData);

	XtVaSetValues(w, XmNfontList, f->fontList, NULL);
	if (f->indSize != MOTIF_XmINVALID_DIMENSION)
	    XtVaSetValues(w, XmNindicatorSize, f->indSize, NULL);
    }
    else {			/* usual case */
	XtVaSetValues(w, XmNfontList, f->fontList, NULL);
    }
}

static void 
changeForeground(Widget w, void *fg)
{
    XtVaSetValues(w, XmNforeground, fg, NULL);
}

static void 
changeBackground(Widget w, void *bg)
{
    Pixel fg;

    XtVaGetValues(w, XmNforeground, &fg, NULL);
    XmChangeColor(w, (Pixel) bg);
    XtVaSetValues(w, XmNforeground, fg, NULL);
}

static void 
setTreeTraversal(Widget w, Boolean trav)
{
    WidgetList children;
    int numChildren;
    int i;

    XtVaGetValues(w,
                  XmNchildren, &children,
                  XmNnumChildren, &numChildren,
                  NULL);

    for (i = 0; i < numChildren; i++) {
        if (!XtIsSubclass(children[i], xmScrollBarWidgetClass)) {
            XtVaSetValues(children[i], XmNtraversalOn, trav, NULL);
        }
        if (IsCanvasTypeWidget(children[i])) {
            setTreeTraversal(children[i], trav);
        }
    }
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pInitialize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pInitialize
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;
    Widget parent;
    jobject target;
    jobject globalRef;
    EventMask xtMask;
    long awtMask = 0;

    globalRef = (jobject)
      (*env)->GetLongField(env, this, mComponentPeerIDs.jniGlobalRef);

    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (JNU_IsNull(env, this) || (cdata == 0)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(cdata->widget,
                  XmNx, (*env)->GetIntField(env, target, componentIDs.x),
                  XmNy, (*env)->GetIntField(env, target, componentIDs.y),
                  XmNvisual, defaultConfig->awt_visInfo.visual,
                  NULL);


    /* For all but canvas-style components, pre-process
     * mouse and keyboard events (which means posting them
     * to the Java EventQueue before dispatching them to Xt).  
     * For canvas-style components ONLY pre-process mouse events 
     * because the input-method currently relies on key events 
     * being processed by Xt first.
     */
    awtMask = java_awt_AWTEvent_MOUSE_EVENT_MASK | 
              java_awt_AWTEvent_MOUSE_MOTION_EVENT_MASK;
    xtMask = ExposureMask | FocusChangeMask;

    if (!IsCanvasTypeWidget(cdata->widget)) {
        awtMask |= java_awt_AWTEvent_KEY_EVENT_MASK;
    } else {
        xtMask |= (KeyPressMask | KeyReleaseMask);
    }
    XtAddEventHandler(cdata->widget, xtMask,
                          True, awt_canvas_event_handler, globalRef);
    
    awt_addWidget(cdata->widget, cdata->widget, globalRef, awtMask);

    cdata->repaintPending = RepaintPending_NONE;
    cdata->cursor = 0;

    /* If the widget was realized as it was instantiated, we
     * may need to adjust the stacking order since the XmNinsertPosition
     * proc doesn't result in proper stacking if the widget is realized.
     */
    if (XtIsRealized(cdata->widget) &&
        (parent = XtParent(cdata->widget)) != NULL) {
        WidgetList children;
        int num_children;
        Window *windows;
        int num_windows = 0;
        int i;

        XtVaGetValues(parent,
                      XmNnumChildren, &num_children,
                      XmNchildren, &children,
                      NULL);

        windows = (Window *) XtMalloc(num_children * sizeof(Window));
        for (i = 0; i < num_children; i++) {
            if (XtIsRealized(children[i])) {
                windows[num_windows++] = XtWindow(children[i]);
            }
        }
        XRestackWindows(awt_display, windows, num_windows);
        XtFree((char *) windows);
    }
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pShow
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pShow
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    awt_util_show(cdata->widget);

    if ((*env)->GetIntField(env,this,mComponentPeerIDs.cursorSet) == 0) {
        (*env)->SetIntField(env,this,mComponentPeerIDs.cursorSet
                            ,(jint) awt_util_setCursor(cdata->widget, cdata->cursor));
    }

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pHide
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pHide
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    awt_util_hide(cdata->widget);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pEnable
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pEnable
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    awt_util_enable(cdata->widget);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pDisable
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pDisable
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    awt_util_disable(cdata->widget);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pReshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pReshape
  (JNIEnv *env, jobject this, jint x, jint y, jint w, jint h)
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    awt_util_reshape(cdata->widget, x, y, w, h);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pAddRepaint
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pAddRepaint
  (JNIEnv *env, jobject this, jint x, jint y, jint w, jint h) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (XtWindow(cdata->widget) == 0) {
        AWT_UNLOCK();
        return;
    }
    /*
      printf("repaint %d,%d,%dx%d\n", x, y, w, h);
    */
    if ((cdata->repaintPending & RepaintPending_REPAINT) == 0) {
        /* remember the repaint area */
        cdata->repaintPending |= RepaintPending_REPAINT;
        cdata->repaintRect.x1 = x;
        cdata->repaintRect.y1 = y;
        cdata->repaintRect.x2 = x + w;
        cdata->repaintRect.y2 = y + h;
    } else {
        /* expand the repaint area */

        if (x < cdata->repaintRect.x1) {
            cdata->repaintRect.x1 = x;
        }
        if (y < cdata->repaintRect.y1) {
            cdata->repaintRect.y1 = y;
        }
        if (x + w > cdata->repaintRect.x2) {
            cdata->repaintRect.x2 = x + w;
        }
        if (y + h > cdata->repaintRect.y2) {
            cdata->repaintRect.y2 = y + h;
        }
    }

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pTriggerRepaint
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pTriggerRepaint
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        AWT_UNLOCK();
        return;
    }
    if (XtWindow(cdata->widget) == 0) {
        AWT_UNLOCK();
        return;
    }
    if (cdata->repaintPending & RepaintPending_REPAINT) {
        /* generate a GraphicsExpose event */
        XEvent evt;

        evt.xgraphicsexpose.type = GraphicsExpose;
        evt.xgraphicsexpose.send_event = False;
        evt.xgraphicsexpose.display = awt_display;
        evt.xgraphicsexpose.drawable = XtWindow(cdata->widget);
        evt.xgraphicsexpose.x = cdata->repaintRect.x1;
        evt.xgraphicsexpose.y = cdata->repaintRect.y1;
        evt.xgraphicsexpose.width = cdata->repaintRect.x2 - cdata->repaintRect.x1;
        evt.xgraphicsexpose.height = cdata->repaintRect.y2 - cdata->repaintRect.y1;
        evt.xgraphicsexpose.count = 0;
        XSendEvent(awt_display, XtWindow(cdata->widget), False, ExposureMask, &evt);
        awt_output_flush();
    }
    AWT_UNLOCK();
    return;
} /* pTriggerRepaint() */

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pDispose
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    XtUnmanageChild(cdata->widget);

    awt_delWidget(cdata->widget);
    awt_util_consumeAllXEvents(cdata->widget);
    XtDestroyWidget(cdata->widget);
    if (cdata->cursor != 0 && cdata->cursor != None && !cdata->customCursor) {
        XFreeCursor(awt_display, cdata->cursor);
    }
    free((void *) cdata);
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,0);

    awtJNI_DeleteGlobalRef(env, this);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pMakeCursorVisible
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pMakeCursorVisible
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    (*env)->SetIntField(env, this, mComponentPeerIDs.cursorSet
                    ,(jint) awt_util_setCursor(cdata->widget, cdata->cursor));

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pGetLocationOnScreen
 * Signature: ()Ljava/awt/Point;
 */
JNIEXPORT jobject JNICALL Java_sun_awt_motif_MComponentPeer_pGetLocationOnScreen
  (JNIEnv *env, jobject this)
{
    struct ComponentData *cdata;
    jobject point = 0;
    jclass clazz;
    jmethodID mid = 0;
    Position rx = 0;
    Position ry = 0;

    AWT_LOCK();
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return NULL;
    }
    if (!XtIsRealized(cdata->widget)) {
        JNU_ThrowInternalError(env, "widget not visible on screen");
        AWT_UNLOCK();
        return NULL;
    }
    /* Translate the component to the screen coordinate system */
    XtTranslateCoords(cdata->widget, 0, 0, &rx, &ry);

    clazz = (*env)->FindClass(env, "java/awt/Point");
    mid = (*env)->GetMethodID(env, clazz, "<init>", "(II)V");
    if (mid != 0) {
        point = (*env)->NewObject(env, clazz, mid,
                                  (jint) rx, (jint) ry);
    }
    if (((*env)->ExceptionOccurred(env)) || JNU_IsNull(env, point)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return NULL;
    }
    AWT_UNLOCK();
    return point;
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    getParent_NoClientCode
 * Signature: (Ljava/awt/Component)Ljava/awt/Container;
 *
 * NOTE: This method may be called by privileged threads. 
 *       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
 */
JNIEXPORT jobject JNICALL Java_sun_awt_motif_MComponentPeer_getParent_1NoClientCode
  (JNIEnv *env, jclass thisClass, jobject component)
{
    jobject parent = NULL;

    /* getParent is actually getParent_NoClientCode() */
    parent = (*env)->CallObjectMethod(env,component,componentIDs.getParent);
    assert(!((*env)->ExceptionOccurred(env)));
    return parent;
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    getComponents_NoClientCode
 * Signature: (Ljava/awt/Container)[Ljava/awt/Component;
 *               REMIND: Signature is incorrect for returned array value
 *
 * NOTE: This method may be called by privileged threads. 
 *       DO NOT INVOKE CLIENT CODE ON THIS THREAD!
 */
JNIEXPORT jobjectArray JNICALL Java_sun_awt_motif_MComponentPeer_getComponents_1NoClientCode
  (JNIEnv *env, jclass thisClass, jobject container)
{
    jobjectArray contents = NULL;
    contents = (*env)->CallObjectMethod(
				env, container, containerIDs.getComponents);
    assert(!((*env)->ExceptionOccurred(env)));
    return contents;
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pSetForeground
 * Signature: (Ljava/awt/Color;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pSetForeground
  (JNIEnv *env, jobject this, jobject c) 
{
    struct ComponentData *bdata;
    Pixel color;

    if (JNU_IsNull(env, c)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    bdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (bdata == 0 || bdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    if (bdata->awtData == (AwtDataPtr) NULL)
	bdata->awtData = defaultConfig;

    color = awtJNI_GetColorForVis (env, c, bdata->awtData);
    XtVaSetValues(bdata->widget, XmNforeground, color, NULL);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pSetBackground
 * Signature: (Ljava/awt/Color;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pSetBackground
  (JNIEnv *env, jobject this, jobject c) 
{
    struct ComponentData *bdata;
    Pixel color;
    Pixel fg;

    if (JNU_IsNull(env, c)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    bdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (bdata == 0 || bdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    if (bdata->awtData == (AwtDataPtr) NULL)
	bdata->awtData = defaultConfig;

    color = awtJNI_GetColorForVis (env, c, bdata->awtData);
    XtVaGetValues(bdata->widget, XmNforeground, &fg, NULL);
    XmChangeColor(bdata->widget, color);
    XtVaSetValues(bdata->widget, XmNforeground, fg, NULL);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pSetScrollbarBackground
 * Signature: (Ljava/awt/Color;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pSetScrollbarBackground
  (JNIEnv *env, jobject this, jobject c) 
{
    struct ComponentData *bdata;
    Pixel color;
    Pixel fg;
    int                 i;
    WidgetList          wlist;
    Cardinal            wlen = 0;


    /* This method propagates the specified background color to the scrollbars in the composite widget.
     * Used to set background scrollbar color in List, TextArea, ScrollPane to its parent.
     */
    if (JNU_IsNull(env, c)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    bdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (bdata == 0 || bdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (!XtIsComposite(bdata->widget)) {
        AWT_UNLOCK();
	return;
    }
    color = awtJNI_GetColor(env, c);

    XtVaGetValues(bdata->widget,
                  XmNchildren, &wlist,
                  XmNnumChildren, &wlen,
                  NULL);
    if (wlen > 0) {
        for (i=0; i < wlen; i++) {
		if (XtIsSubclass(wlist[i], xmScrollBarWidgetClass)) {
    		    XtVaGetValues(wlist[i], XmNforeground, &fg, NULL);
    		    XmChangeColor(wlist[i], color);
    		    XtVaSetValues(wlist[i], XmNforeground, fg, NULL);
		}
        }
        XtVaGetValues(bdata->widget, XmNforeground, &fg, NULL);
        XmChangeColor(bdata->widget, color);
        XtVaSetValues(bdata->widget, XmNforeground, fg, NULL);
    }

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    pSetInnerForeground
 * Signature: (Ljava/awt/Color;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_pSetInnerForeground
  (JNIEnv *env, jobject this, jobject c) 
{
    struct ComponentData *bdata;
    Pixel color;

    /* This method propagates the specified foreground color to all its children.
     * It is called to set foreground color in List, TextArea, ScrollPane.
     */
    if (JNU_IsNull(env, c)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    bdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (bdata == 0 || bdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    color = awtJNI_GetColor(env, c);
    awt_util_mapChildren(bdata->widget, changeForeground, 1, (void *) color);
    AWT_FLUSH_UNLOCK();
}


/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    isBackgroundInherited
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_sun_awt_motif_MComponentPeer_isBackgroundInherited
  (JNIEnv *env, jobject this) 
{
    jobject target;
    jobject color;

    if (JNU_IsNull(env, this)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return JNI_FALSE;
    }
    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    if (JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return JNI_FALSE;
    }
    color = (*env)->GetObjectField(env, target, componentIDs.background);
    if (JNU_IsNull(env, color)) {
	AWT_UNLOCK();
	return JNI_TRUE;
    } else {
	AWT_UNLOCK();
	return JNI_FALSE;
    }
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    isForegroundInherited
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_sun_awt_motif_MComponentPeer_isForegroundInherited
  (JNIEnv *env, jobject this) 
{
    jobject target;
    jobject color;

    if (JNU_IsNull(env, this)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return JNI_FALSE;
    }
    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    if (JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return JNI_FALSE;
    }
    color = (*env)->GetObjectField(env, target, componentIDs.foreground);
    if (JNU_IsNull(env, color)) {
	AWT_UNLOCK();
	return JNI_TRUE;
    } else {
	AWT_UNLOCK();
	return JNI_FALSE;
    }
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    setFont
 * Signature: (Ljava/awt/Font;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_setFont
  (JNIEnv *env, jobject this, jobject f) 
{
    struct ComponentData *cdata;
    struct FontData *fdata;
    XmFontList fontlist;
    char *err;

    struct changeFontInfo finfo = { NULL, FALSE, NULL, 0 };

    if (JNU_IsNull(env, f)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();

    /*Leak baby leak! */
    fdata = awtJNI_GetFontData(env, f, &err);
    if (fdata == 0) {
        JNU_ThrowInternalError(env, err);
        AWT_UNLOCK();
        return;
    }
    cdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (awtJNI_IsMultiFont(env, f)) {
        fontlist = awtJNI_GetFontList(env, f);
	finfo.isMultiFont = TRUE;
    } else {
        fontlist = XmFontListCreate(fdata->xfont, "labelFont");
	finfo.isMultiFont = FALSE;
    }

    if (fontlist != NULL) {
	finfo.fontList = fontlist;
	finfo.fontData = fdata;
	awt_util_mapChildren(cdata->widget, changeFont, 1, (void *)&finfo);
        XmFontListFree(fontlist);
    } else {
        JNU_ThrowNullPointerException(env, "NullPointerException");
    }

    AWT_FLUSH_UNLOCK();
} /* MComponentPeer.setFont() */

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    setCursor
 * Signature: (Ljava/awt/Cursor;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_setCursor
  (JNIEnv *env, jobject this, jobject cursor) 
{
    Cursor xcursor;
    struct ComponentData *cdata;
    int cursorType;

    AWT_LOCK();

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (JNU_IsNull(env, cursor) || cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    cursorType = (*env)->GetIntField(env, cursor, cursorIDs.type);

    if (cursorType == java_awt_Cursor_CUSTOM_CURSOR) {
        xcursor = (Cursor)
	  (*env)->GetLongField(env, cursor, x11CustomCursorIDs.pData);
    } else {
        switch (cursorType) {
          case java_awt_Cursor_DEFAULT_CURSOR:
              cursorType = -1;
              break;
          case java_awt_Cursor_CROSSHAIR_CURSOR:
              cursorType = XC_crosshair;
              break;
          case java_awt_Cursor_TEXT_CURSOR:
              cursorType = XC_xterm;
              break;
          case java_awt_Cursor_WAIT_CURSOR:
              cursorType = XC_watch;
              break;
          case java_awt_Cursor_SW_RESIZE_CURSOR:
              cursorType = XC_bottom_left_corner;
              break;
          case java_awt_Cursor_NW_RESIZE_CURSOR:
              cursorType = XC_top_left_corner;
              break;
          case java_awt_Cursor_SE_RESIZE_CURSOR:
              cursorType = XC_bottom_right_corner;
              break;
          case java_awt_Cursor_NE_RESIZE_CURSOR:
              cursorType = XC_top_right_corner;
              break;
          case java_awt_Cursor_S_RESIZE_CURSOR:
              cursorType = XC_bottom_side;
              break;
          case java_awt_Cursor_N_RESIZE_CURSOR:
              cursorType = XC_top_side;
              break;
          case java_awt_Cursor_W_RESIZE_CURSOR:
              cursorType = XC_left_side;
              break;
          case java_awt_Cursor_E_RESIZE_CURSOR:
              cursorType = XC_right_side;
              break;
          case java_awt_Cursor_HAND_CURSOR:
              cursorType = XC_hand2;
              break;
          case java_awt_Cursor_MOVE_CURSOR:
              cursorType = XC_fleur;
              break;
          default:
              AWT_UNLOCK();
              return;
        }

        if (cursorType != -1) {
            xcursor = XCreateFontCursor(awt_display, cursorType);
        } else {
            xcursor = None;
        }
    }

    if (cdata->cursor != 0 && cdata->cursor != None && !cdata->customCursor) {
        XFreeCursor(awt_display, cdata->cursor);
    }

    cdata->customCursor = (cursorType == java_awt_Cursor_CUSTOM_CURSOR);
    cdata->cursor = xcursor;

    (*env)->SetIntField(env, this, mComponentPeerIDs.cursorSet
                    ,(jint) awt_util_setCursor(cdata->widget, cdata->cursor));

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    nativeHandleEvent
 * Signature: (Ljava/awt/AWTEvent;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_nativeHandleEvent
  (JNIEnv *env, jobject this, jobject event) 
{
    extern void awt_modify_KeyEvent(JNIEnv *env, XEvent * xevent, jobject jevent);
    jlong data;
    Widget widget=0;

    if (JNU_IsNull(env, event)) {
        return;
    }
    AWT_LOCK();

    data = (*env)->GetLongField(env, event, awtEventIDs.data);

    if (data != 0) {
        if (! (*env)->GetBooleanField(env, event, awtEventIDs.consumed) ) {
            switch ((*env)->GetIntField(env, event, awtEventIDs.id)) {
                case java_awt_event_KeyEvent_KEY_RELEASED:
                case java_awt_event_KeyEvent_KEY_PRESSED:
                {
                    awt_modify_KeyEvent(env, (XEvent *) data, event);
                    break;
                }

                default:
                    break;
            }

            widget = XtWindowToWidget(awt_display,
                ((XEvent *)data)->xany.window);

            if (!((widget == 0) || (!XtIsObject(widget)) ||
                  (widget->core.being_destroyed))) {
		/* Queue the event to be handled by the AWT-Motif thread */
		awt_put_back_event(env, (XEvent*)data);
            }

        }
        AWT_UNLOCK();

        return;
    }
    AWT_UNLOCK();
    return;
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    requestFocus
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_requestFocus
  (JNIEnv *env, jobject this)
{
    struct ComponentData *bdata;
    Boolean result;

    AWT_LOCK();
    bdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (bdata == 0 || bdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    /* Motif won't let us set focus to a Panel/Canvas which has
     * traversable children, so temporarily turn off traversal on
     * it's children before requesting the focus.
     */
    if (IsCanvasTypeWidget(bdata->widget)) {
        setTreeTraversal(bdata->widget, FALSE);
    }
    result = XmProcessTraversal(bdata->widget, XmTRAVERSE_CURRENT);

    if (IsCanvasTypeWidget(bdata->widget)) {
        setTreeTraversal(bdata->widget, TRUE);
    }
    AWT_FLUSH_UNLOCK();
}

Dimension
awt_computeIndicatorSize(struct FontData *fdata)
{
    int height;
    int acc;
    int i;

    if (fdata == (struct FontData *) NULL)
	return MOTIF_XmINVALID_DIMENSION;

    /*
     * If Java font maps into single platform font - there's no
     * problem.  Let Motif use it's usual calculations in this case.
     */
    if (fdata->charset_num == 1)
	return MOTIF_XmINVALID_DIMENSION;

    acc = 0;
    for (i = 0; i < fdata->charset_num; ++i) {
	XFontStruct *xfont = fdata->flist[i].xfont;
	acc += xfont->ascent + xfont->descent;
    }

    height = acc / fdata->charset_num;
    if (height < MOTIF_XmDEFAULT_INDICATOR_DIM)
	height = MOTIF_XmDEFAULT_INDICATOR_DIM;

    return height;
}

Dimension
awt_adjustIndicatorSizeForMenu(Dimension indSize)
{
    if (indSize == 0 || indSize == MOTIF_XmINVALID_DIMENSION)
	return MOTIF_XmINVALID_DIMENSION; /* let motif do the job */

    /* Indicators in menus are smaller.
       2/3 is a magic number from Motif internals. */
    indSize = indSize * 2 / 3;
    if (indSize < MOTIF_XmDEFAULT_INDICATOR_DIM)
	indSize = MOTIF_XmDEFAULT_INDICATOR_DIM;

    return indSize;
}
