/*
 *  @(#)awt_Dialog.c	1.87	98/09/15
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include <X11/Shell.h>
#include <Xm/VendorS.h>
#include <Xm/DialogS.h>
#include <Xm/Form.h>
#include <Xm/AtomMgr.h>
#include <Xm/Protocols.h>
#include <Xm/MwmUtil.h>
#include "java_awt_MenuBar.h"
#include "sun_awt_motif_MMenuBarPeer.h"
#include "java_awt_Dialog.h"
#include "sun_awt_motif_MDialogPeer.h"
#include "sun_awt_motif_MFramePeer.h"
#include "java_awt_Color.h"
#include "java_awt_Component.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "java_awt_Image.h"
#include "java_awt_Font.h"
#include "java_awt_Insets.h"
#include "canvas.h"
#include "color.h"
#include "awt_util.h"

#include "awt_Insets.h"
#include "awt_Component.h"
#include "awt_Window.h"

#include <jni.h>
#include <jni_util.h>

extern struct InsetsIDs insetsIDs;
extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct ComponentIDs componentIDs;
extern struct MDialogPeerIDs mDialogPeerIDs;
extern struct WindowIDs windowIDs;

static void reshape();
static int inreshape = 0;
static void awtJNI_ChangeInsets(JNIEnv *env, 
				jobject this, 
				struct FrameData *wdata);
#ifndef NOMODALFIX
extern Boolean awt_isModal();
extern Boolean awt_isWidgetModal(Widget w);
#endif NOMODALFIX

/* fieldIDs for Dialog fields that may be accessed from C */
static struct DialogIDs {
    jfieldID modal;
    jfieldID resizable;
} dialogIDs;

/*
 * Class:     java_awt_Dialog
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for Dialog.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_Dialog_initIDs
  (JNIEnv *env, jclass cls)
{
    dialogIDs.modal = (*env)->GetFieldID(env, cls, "modal", "Z");
    dialogIDs.resizable = (*env)->GetFieldID(env, cls, "resizable", "Z");
}

/* fieldIDs for MDialogPeer fields that may be accessed from C */
struct MDialogPeerIDs {
    jfieldID insets;
} mDialogPeerIDs;

/* This function gets called from the static initializer for MDialogPeer.java
   to initialize the fieldIDs for fields that may be accessed from C */
/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_initIDs
(JNIEnv *env, jclass cls) 
{
  mDialogPeerIDs.insets = (*env)->GetFieldID(env, cls, "insets",
					     "Ljava/awt/Insets;");
}

/*
 * client_data is MDialogPeer instance pointer
 */
static void 
Window_resize(Widget wd, XtPointer client_data, XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (!inreshape) {
        Dimension width;
        Dimension height;
        jobject target;

	if ((*env)->EnsureLocalCapacity(env, 1) < 0)
	    return;

        target = (*env)->GetObjectField(env, (jobject) client_data,
					mComponentPeerIDs.target);

        XtVaGetValues(wd,
                      XmNwidth, &width,
                      XmNheight, &height,
                      NULL);

        (*env)->SetIntField(env, target, componentIDs.width, (jint) width);
        (*env)->SetIntField(env, target, componentIDs.height, (jint) height);

        JNU_CallMethodByName(env, NULL, (jobject) client_data
			     ,"handleResize"
			     ,"(II)V"
			     ,width, height);
        if ((*env)->ExceptionOccurred(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        }
	(*env)->DeleteLocalRef(env, target);
    }
} /* Window_resize() */


/*
 * data is MDialogPeer instance pointer
 */
static void 
Dialog_event_handler(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject) data;
    struct FrameData *wdata;
    jobject target;
    int x, y;
    jboolean isPacked;

    /* Any event handlers which take peer instance pointers as
     * client_data should check to ensure the widget has not been
     * marked as destroyed as a result of a dispose() call on the peer
     * (which can result in the peer instance pointer already haven
     * been gc'd by the time this event is processed)
     */
    if (w->core.being_destroyed) {
        return;
    }
    wdata = (struct FrameData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0) {
        return;
    }
    switch (event->xany.type) {
        case FocusOut:
            if (event->xfocus.mode == NotifyNormal) {
#ifndef NOMODALFIX
		if (!awt_isModal() || awt_isWidgetModal(w)) {
#endif NOMODALFIX
		    awt_setDeactivatedShell(w);

		    JNU_CallMethodByName(env
					 ,NULL
					 ,this
					 ,"handleDeactivate"
					 ,"()V");
		    if ((*env)->ExceptionOccurred(env)) {
			(*env)->ExceptionDescribe(env);
			(*env)->ExceptionClear(env);
		    }
#ifndef NOMODALFIX
		}
#endif NOMODALFIX
            }
            break;

        case MapNotify:
            if (wdata->mappedOnce == 0) {
                wdata->mappedOnce = 1;
                JNU_CallMethodByName(env
				     ,NULL
				     ,this
				     ,"makeCursorsVisible"
				     ,"()V");
                if ((*env)->ExceptionOccurred(env)) {
                    (*env)->ExceptionDescribe(env);
                    (*env)->ExceptionClear(env);
                }
            }
            break;

        case ConfigureNotify:
	    if ((*env)->EnsureLocalCapacity(env, 1) < 0)
	        break;
            target = (*env)->GetObjectField(env, this,
					    mComponentPeerIDs.target);

            /* We can detect the difference between a move and a resize
             * by checking the send_event flag on the event; if it's true,
             * then it's indeed a move, if it's false, then this is a resize
             * and we do not want to process it as a "move"(for resizes
             * the x,y values are misleadingly set to 0,0, and so just
             * checking for an x,y delta won't work).
             */

            x = (*env)->GetIntField(env, target, componentIDs.x);
            y = (*env)->GetIntField(env, target, componentIDs.y);
            if (( (x != event->xconfigure.x  - wdata->left) ||
                  (y != event->xconfigure.y - wdata->top) ) &&
                event->xconfigure.send_event == TRUE) {
                x = event->xconfigure.x - wdata->left;
                y = event->xconfigure.y - wdata->top;

                (*env)->SetIntField(env, target, componentIDs.x, (jint) x);
                (*env)->SetIntField(env, target, componentIDs.y, (jint) y);
                JNU_CallMethodByName(env
				     ,NULL
				     ,this
				     ,"handleMoved"
				     ,"(II)V", x, y);
                if ((*env)->ExceptionOccurred(env)) {
                    (*env)->ExceptionDescribe(env);
                    (*env)->ExceptionClear(env);
                }
            }
	    (*env)->DeleteLocalRef(env, target);
            break;
	case ReparentNotify:
        if ( event->xreparent.parent != awt_root )
        {
	    unsigned		runningWM;
	    XWindowAttributes	winAttr;
	    Position		shellX, shellY;
	    int			correctWMTop;
	    int			correctWMRest;
	    int			topCorrection;
	    int			restCorrection;
	    int			getTargetX, getTargetY;
	    int			width, height;
	    int			correctTargetW;
	    int			correctTargetH;
	    int                     screenX, screenY;
	    Window                  containerWindow;
	    
	    wdata->reparented = True;
	    correctWMTop = -1;
	    correctWMRest = -1;
            target = (*env)->GetObjectField(env, this,
					    mComponentPeerIDs.target);
	    
	    /* N.B. The inset calculation method used for OTHER_WM
	       could actually be used for both MWM, OLWM as well.	*/
	    
	    /* Get an idea what window manager is running.  This iS
	       definitive for Motif, very likely for OpenLook, and
	       otherwise not possible to determine window manager.  */
	    
	    runningWM = awt_util_runningWindowManager();
	    
	    switch (runningWM) {
		
	    case CDE_WM:
	    case MOTIF_WM:
		/* For Motif WM, the correct insets are given by
		   (x,y) coordinates of event window "parent".  */
		XGetWindowAttributes( XtDisplay(wdata->winData.shell),
				      event->xreparent.parent, &winAttr );
		correctWMRest = winAttr.x;
		correctWMTop = winAttr.y;
#ifdef _pauly_debug
		fprintf(stdout,">>>>Motif WM insets: %d, %d\n", correctWMRest, correctWMTop);
		fflush(stdout);
#endif /* _pauly_debug */
		break;
		
	    case OPENLOOK_WM:
		/* For OpenLook WM, correct insets are given by
		   (x,y) coordinates of event window "window".  */
		XGetWindowAttributes( XtDisplay(wdata->winData.shell),
				      event->xreparent.window, &winAttr );
		correctWMRest = winAttr.x;
		correctWMTop = winAttr.y;
#ifdef _pauly_debug
		fprintf(stdout,">>>>OpenLook WM insets: %d, %d\n", correctWMRest, correctWMTop);
		fflush(stdout);
#endif /* _pauly_debug */
		break;
		
	    case OTHER_WM:
	    default:
		if  (XTranslateCoordinates(awt_display,
					   event->xreparent.window,
					   RootWindowOfScreen(XtScreen(wdata->winData.shell)),
					   0,0, &screenX, &screenY, &containerWindow)) {
		    /* Insets given by difference between window's
		       screen coordinates and target coordinates.
		       N.B. THIS "USUALLY" WORKS. BUT NOT ALWAYS.
		       For example, under TWM with shell warping
		       to pointer this fails.  But in any event,
		       for bogus insets, reset to (25,5) below. */
		    
		    getTargetX = (*env)->GetIntField(env, target,
						     componentIDs.x);
		    getTargetY = (*env)->GetIntField(env, target,
						     componentIDs.y);
		    correctWMRest = (int)screenX - getTargetX;
		    correctWMTop = (int)screenY - getTargetY;
		}
                    /* Since preceding code is heuristic
                       rather than deterministic,
                       check that we have reasonable
                       calculations for the insets.
                       If not, just use old (probably
                       incorrect) estimates (25,5). */
                    /* Okay okay but only if we are unsure of the window
                    ** manager, Some of us may be sight impaired with
                    ** some pretty big fonts so only act smarter than
                    ** thou when we have no choice... we can trust
                    ** the numbers under olwm and cde
                    */
                    if  ((correctWMRest < 0) || (correctWMRest > 60))
                        correctWMRest = wdata->left;
                    if  ((correctWMTop < 0) || (correctWMTop > 100))
                        correctWMTop = wdata->top;
 

#ifdef _pauly_debug
		fprintf(stdout,">>>>Other WM insets: %d, %d\n", correctWMRest, correctWMTop);
		fflush(stdout);
#endif /* _pauly_debug */
		break;
		
	    }  /* switch (runningWM) */
	    
            x = (*env)->GetIntField(env, target, componentIDs.x);
            y = (*env)->GetIntField(env, target, componentIDs.y);
            width = (*env)->GetIntField(env, target, componentIDs.width);
            height = (*env)->GetIntField(env, target, componentIDs.height);

            topCorrection = correctWMTop - getTopGuess();
            restCorrection = correctWMRest - getSideGuess();
            if ( topCorrection != 0 || restCorrection != 0 )
            {
 
#ifdef _pauly_debug
fprintf(stdout,">>>>insets fixed: %d, %d\n", correctWMRest, correctWMTop);
fflush(stdout);
#endif /* _pauly_debug */
                wdata->top += topCorrection;
                wdata->bottom += restCorrection;
                wdata->left += restCorrection;
                wdata->right += restCorrection;

                awtJNI_ChangeInsets(env, this, wdata);

                isPacked = (*env)->GetBooleanField( env, target, 
						componentIDs.isPacked);
#ifdef _pauly_debug
fprintf(stdout,">>>>isPacked = %d\n", isPacked );
fflush(stdout);
#endif /* _pauly_debug */

                if  (isPacked)
                {
                    /* if this window has been sized by a pack() the
                    ** user wants to keep the interior (or Client area)
                    ** geometry intact.
                    **
                    ** We must adjust the target dimensions appropriately.
                    ** These were calculated originally with bogus insets.
                    */

                    correctTargetW = width + (2 * restCorrection);
                    correctTargetH = height +(topCorrection+restCorrection);

                    (*env)->SetIntField(env, target, componentIDs.width,
                                        (jint) correctTargetW);
                    (*env)->SetIntField(env, target, componentIDs.height,
                                        (jint) correctTargetH);
                    /*
                    **  Normally you only reconfigure the outerCanvas due to
                    **  handling the ReconfigureNotify on the innerCanvas.
                    **  However, in this case the innerCanvas may not have
                    **  changed, but outterCanvas may still need to, since the
                    **  insets have changed.
                    */
                    reshape(env, this, wdata, x, y,
                                        correctTargetW, correctTargetH);
                } else {
                    reshape(env, this, wdata, x, y, width, height);
                }

            }
/*	    else {
**    		if  ( (!wdata->isResizable) && (!wdata->isFixedSizeSet) ) {
**			reshape(env, this, wdata, x, y, width, height);
**    		}
**	    }
*/
	    (*env)->DeleteLocalRef(env, target);
	}
	break;
        default:
            break;
    }
}

/**  
  static int WaitForUnmap(void *data)

This function is called in the event loop for modal dialogs.  Its
return value indicates whether the event loop for the modal dialog
should continue or return.  We check to see whether the dialog is
still managed, and answer accordingly.  Unfortunately, there seems to
be a bug in event delivery under X11 such that we occasionally get
calls to this function even long after the dialog has been unmanaged
and destroyed.  This can have disastrous consequences if we just
blindly check the dialog's management status.  Instead, we use a
structure to wrap the widget, which lets us gracefully recover in the
case where this function is called after the widget has been
destroyed.  **/

static int 
WaitForUnmap(void *data)
{
        struct FrameData *fdata = (struct FrameData *) data;

        return fdata->DlgIsWithdrawn;
}


/*
 * client_data is MDialogPeer instance pointer
 */
static void 
Dialog_quit(Widget w,
            XtPointer client_data,
            XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    JNU_CallMethodByName(env, NULL, (jobject)client_data, "handleQuit", "()V");
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * client_data is MDialogPeer instance pointer
 */
static void 
Dialog_activate(Widget w,
                XtPointer client_data,
                XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
#ifndef NOMODALFIX
    if (!awt_isModal() || awt_isWidgetModal(w)) {
#endif NOMODALFIX
	awt_setActivatedShell(w);

	JNU_CallMethodByName(env, NULL, (jobject)client_data, "handleActivate", "()V");
	if ((*env)->ExceptionOccurred(env)) {
	    (*env)->ExceptionDescribe(env);
	    (*env)->ExceptionClear(env);
	}
#ifndef NOMODALFIX
    }
#endif NOMODALFIX
}

static void 
setDeleteCallback(jobject this, struct FrameData *wdata)
{
    Atom xa_WM_DELETE_WINDOW;
    Atom xa_WM_TAKE_FOCUS;
    Atom xa_WM_PROTOCOLS;

    XtVaSetValues(wdata->winData.shell,
                  XmNdeleteResponse, XmDO_NOTHING,
                  NULL);
    xa_WM_DELETE_WINDOW = XmInternAtom(XtDisplay(wdata->winData.shell),
                                       "WM_DELETE_WINDOW", False);
    xa_WM_TAKE_FOCUS = XmInternAtom(XtDisplay(wdata->winData.shell),
                                    "WM_TAKE_FOCUS", False);
    xa_WM_PROTOCOLS = XmInternAtom(XtDisplay(wdata->winData.shell),
                                   "WM_PROTOCOLS", False);

    XmAddProtocolCallback(wdata->winData.shell,
                          xa_WM_PROTOCOLS,
                          xa_WM_DELETE_WINDOW,
                          Dialog_quit, (XtPointer) this);
    XmAddProtocolCallback(wdata->winData.shell,
                          xa_WM_PROTOCOLS,
                          xa_WM_TAKE_FOCUS,
                          Dialog_activate, (XtPointer) this);
}

static void 
awtJNI_ChangeInsets(JNIEnv * env, jobject this, struct FrameData *wdata)
{
    jobject insets;

    if ((*env)->EnsureLocalCapacity(env, 1) < 0)
        return;

    insets = (*env)->GetObjectField(env, this, mDialogPeerIDs.insets);

    if (JNU_IsNull(env, insets)) {
        return;
    }
    (*env)->SetIntField(env, insets, insetsIDs.top, (jint) wdata->top);
    (*env)->SetIntField(env, insets, insetsIDs.left, (jint) wdata->left);
    (*env)->SetIntField(env, insets, insetsIDs.bottom, (jint) wdata->bottom);
    (*env)->SetIntField(env, insets, insetsIDs.right, (jint) wdata->right);
    (*env)->DeleteLocalRef(env, insets);
}


/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;Ljava/lang/Object;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_create
  (JNIEnv *env, jobject this, jobject parent, jobject arg) 
{
    Arg args[40];
    int argc;
    struct FrameData *wdata;
    struct FrameData *parentData;
    jobject target;
    jstring warningString;
    int x, y, width, height;
    jboolean resizable;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);
#ifndef NOMODALFIX
    extern void awt_shellPoppedUp(Widget shell, XtPointer c, XtPointer d);
    extern void awt_shellPoppedDown(Widget shell, XtPointer c, XtPointer d);
#endif NOMODALFIX

    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    if (JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (JNU_IsNull(env, parent) || JNU_IsNull(env, arg)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    wdata = ZALLOC(FrameData);
    if (wdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong) wdata);
    (*env)->SetObjectField(env, this, mDialogPeerIDs.insets, arg);

    wdata->top = (*env)->GetIntField(env, arg, insetsIDs.top);
    wdata->left = (*env)->GetIntField(env, arg, insetsIDs.left);
    wdata->bottom = (*env)->GetIntField(env, arg, insetsIDs.bottom);
    wdata->right = (*env)->GetIntField(env, arg, insetsIDs.right);
    awt_Frame_guessInsets( wdata );
    awtJNI_ChangeInsets(env, this, wdata);

    wdata->isModal = (*env)->GetBooleanField(env, target, dialogIDs.modal);
    wdata->mappedOnce = 0;

    wdata->hasTextComponentNative = False;
    wdata->imRemove = False;

    argc = 0;
    parentData = (struct FrameData *) (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);

    x = (*env)->GetIntField(env, target, componentIDs.x);
    y = (*env)->GetIntField(env, target, componentIDs.y);
    width = (*env)->GetIntField(env, target, componentIDs.width);
    height = (*env)->GetIntField(env, target, componentIDs.height);
    resizable = (*env)->GetBooleanField(env, target, dialogIDs.resizable);

    XtSetArg(args[argc], XmNtransientFor, parentData->winData.shell);
    argc++;
    XtSetArg(args[argc], XmNsaveUnder, False);
    argc++;
    XtSetArg(args[argc], XmNx, (XtArgVal) x);
    argc++;
    XtSetArg(args[argc], XmNy, (XtArgVal) y);
    argc++;
    XtSetArg(args[argc], XmNwidth,
             (XtArgVal) width - (wdata->left + wdata->right));
    argc++;

    XtSetArg(args[argc], XmNheight,
             (XtArgVal) height - (wdata->top + wdata->bottom));
    argc++;

    if (resizable) {
        XtSetArg(args[argc], XmNallowShellResize, True);
        argc++;
    } else {
        XtSetArg(args[argc], XmNallowShellResize, False);
        argc++;
    }
    XtSetArg(args[argc], XmNmarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNvisual, defaultConfig->awt_visInfo.visual);
    argc++;
    XtSetArg(args[argc], XmNcolormap, defaultConfig->awt_cmap);
    argc++;
    XtSetArg(args[argc], XmNdepth, defaultConfig->awt_depth);
    argc++;
    XtSetArg(args[argc], XmNnoResize, resizable ? False : True);
    argc++;
    wdata->winData.shell = XtCreatePopupShell("AWTdialog",
                                              xmDialogShellWidgetClass,
                                              parentData->winData.shell,
                                              args,
                                              argc);
    setDeleteCallback(globalRef, wdata);

    XtAddEventHandler(wdata->winData.shell, StructureNotifyMask | FocusChangeMask,
                      FALSE, Dialog_event_handler, globalRef);

    argc = 0;
    if (wdata->isModal) {
#ifndef NOMODALFIX
	XtAddCallback(wdata->winData.shell, 
		      XtNpopupCallback, 
		      awt_shellPoppedUp, 
		      NULL);
	XtAddCallback(wdata->winData.shell, 
		      XtNpopdownCallback, 
		      awt_shellPoppedDown, 
		      NULL);
#endif NOMODALFIX
        XtSetArg(args[argc], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
        argc++;
    }
    XtSetArg(args[argc], XmNwidth,
             (XtArgVal) width - (wdata->left + wdata->right));
    argc++;
    XtSetArg(args[argc], XmNheight,
             (XtArgVal) height - (wdata->top + wdata->bottom));
    argc++;
    XtSetArg(args[argc], XmNmainWindowMarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNmainWindowMarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNspacing, 0);
    argc++;
    wdata->mainWindow = XmCreateForm(wdata->winData.shell, "main",
                                     args, argc);
    wdata->winData.comp.widget = awt_canvas_create((XtPointer) globalRef,
                                                   wdata->mainWindow,
						   "dialog_",
                                                   width,
                                                   height,
						   False,
                                                   wdata,
                                                   defaultConfig);

    warningString = (*env)->GetObjectField(env, target,
					   windowIDs.warningString);
    if (!JNU_IsNull(env, warningString)) {
        Dimension hoffset;
        char *wString;

        wString = (char *) JNU_GetStringPlatformChars(env, warningString, NULL);
        wdata->warningWindow = awt_util_createWarningWindow(wdata->mainWindow, wString);
        JNU_ReleaseStringPlatformChars(env, warningString, (const char *) wString);

        XtVaGetValues(wdata->warningWindow, XmNheight, &hoffset, NULL);
        wdata->top += hoffset;
        XtVaSetValues(wdata->warningWindow,
                      XmNtopAttachment, XmATTACH_FORM,
                      XmNleftAttachment, XmATTACH_FORM,
                      XmNrightAttachment, XmATTACH_FORM,
                      NULL);
        XtVaSetValues(XtParent(wdata->winData.comp.widget),
                      XmNtopAttachment, XmATTACH_WIDGET,
                      XmNtopWidget, wdata->warningWindow,
                      XmNbottomAttachment, XmATTACH_FORM,
                      XmNleftAttachment, XmATTACH_FORM,
                      XmNrightAttachment, XmATTACH_FORM,
                      NULL);
        awtJNI_ChangeInsets(env, this, wdata);
    } else {
        wdata->warningWindow = NULL;
        XtVaSetValues(XtParent(wdata->winData.comp.widget),
                      XmNbottomAttachment, XmATTACH_FORM,
                      XmNtopAttachment, XmATTACH_FORM,
                      XmNleftAttachment, XmATTACH_FORM,
                      XmNrightAttachment, XmATTACH_FORM,
                      NULL);
    }

    XtAddCallback(wdata->winData.comp.widget,
                  XmNresizeCallback,
                  Window_resize,
                  globalRef);

    wdata->menuBar = NULL;

    awt_util_show(wdata->winData.comp.widget);
    XtSetMappedWhenManaged(wdata->winData.shell, False);
    XtManageChild(wdata->mainWindow);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    pSetTitle
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_pSetTitle
  (JNIEnv *env, jobject this, jstring title) 
{
    char *ctitle;
    struct FrameData *wdata;
    XTextProperty text_prop;
    char *c[1];

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 || wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    ctitle = (JNU_IsNull(env, title)) ? "" : (char *) JNU_GetStringPlatformChars(env, title, NULL);
    c[0] = ctitle;
 
    /* need to convert ctitle to CompoundText */
    XmbTextListToTextProperty(awt_display, c, 1,
                XCompoundTextStyle,&text_prop);    
    XtVaSetValues(wdata->winData.shell,
		  XmNtitle, text_prop.value,
                  XmNtitleEncoding, text_prop.encoding,
                  XmNiconName, text_prop.value,
                  XmNiconNameEncoding, text_prop.encoding,
                  XmNname, ctitle,
                  NULL);
    if (ctitle != "")
        JNU_ReleaseStringPlatformChars(env, title, (const char *) ctitle);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    pShow
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_pShow
  (JNIEnv *env, jobject this) 
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(wdata->winData.comp.widget,
                  XmNx, -(wdata->left),
                  XmNy, -(wdata->top), NULL);

    /* a workaround for bug 4062589 that is really a motif problem (see
       bug 4064803). Before popping up a modal dialog, if a pulldown menu
       has the input focus (i.e. user has pulled the menu down), we send
       a fake click event and make sure the click event is processed.
       With this simulation of user clicking, X server will not get
       confused about the modality and a subsequent click on the popup
       modal dialog will not cause system lockup
     */
    if ( wdata->isModal && awt_util_focusIsOnMenu(awt_display) &&
          awt_util_sendButtonClick(awt_display, InputFocus) ) {
       XEvent ev;

       while (True) {
         XtAppPeekEvent(awt_appContext, &ev);
         if ((ev.type == ButtonRelease)
                  && (*(XButtonEvent *)&ev).send_event == True) {
            XtAppProcessEvent (awt_appContext, XtIMAll);
             break;
         } else {
           XtAppProcessEvent (awt_appContext, XtIMAll);
         }
       }
    }

    XtManageChild(wdata->winData.comp.widget);
    XtSetMappedWhenManaged(wdata->winData.shell, True);
    if (wdata->isModal) {

        wdata->DlgIsWithdrawn = False;
        wdata->DlgIsDisposed = False;
        wdata->DlgIsDone = False;

        XtPopup(wdata->winData.shell, XtGrabNonexclusive);
        AWT_FLUSH_UNLOCK();
        awt_MToolkit_modalWait(WaitForUnmap, wdata);
        AWT_LOCK();
        wdata->DlgIsDone = True;
        if ( wdata->DlgIsDisposed )
            free((void *)wdata);
        AWT_UNLOCK();
    } else {
        XtPopup(wdata->winData.shell, XtGrabNone);
        XRaiseWindow(awt_display, XtWindow(wdata->winData.shell));
        AWT_FLUSH_UNLOCK();
    }
}

/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    pHide
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_pHide
  (JNIEnv *env, jobject this) 
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    wdata->DlgIsWithdrawn = True;
    XtUnmanageChild(wdata->winData.comp.widget);
    XtPopdown(wdata->winData.shell);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    pReshape
 * Signature: (IIII)V
 */

/* this functionality  is invoked from both java and native code, and we only want to lock when invoking it from
 * java, so wrap the native method version with the locking */

JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_pReshape
  (JNIEnv *env, jobject this, jint x, jint y, jint w, jint h) 
{
    struct FrameData *wdata;

    AWT_LOCK();

    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    reshape(env, this, wdata, x, y, w, h);
    AWT_FLUSH_UNLOCK();
}

static void
reshape(JNIEnv *env, jobject this, struct FrameData *wdata,
	jint x, jint y, jint w, jint h)
{
    Dimension hoffset;

    inreshape = 1;

    /*
     * The warning window is interior to the main window widget.
     * Since its height is included in the top and bottom insets, we need
     * to subtract them out again before sizing the main window widget. The
     * IM Status Window is exterior to the main window so its height does
     * not need to be removed from the insets.
     */
    hoffset = 0;
    if (wdata->warningWindow != 0) {
        XtVaGetValues(wdata->warningWindow, XmNheight, &hoffset, NULL);
    }

    if (wdata->hasTextComponentNative){
	hoffset +=  wdata->imHeight;
    }
    w -= (wdata->left + wdata->right);
    h += (hoffset - (wdata->top + wdata->bottom));

    if (w < 0) {
        w = 0;
    }
    if (h < 0) {
        h = 0;
    }
    /* In order to get the Shell's location right, we have to manage the 
     * mainWindow - because DialogShell maps the XmNx and XmNy resources
     * of its child to the shells origin.  See XmDialogShell(3X).
     */

    XtManageChild(wdata->mainWindow);

    /* GES: AVH's hack from awt_util.c: 
     * Motif ignores attempts to move a toplevel window to 0,0.
     * Instead we set the position to 1,1. The expected value is
     * returned by Frame.getBounds() since it uses the internally
     * held rectangle rather than querying the peer.
     */
    XtVaSetValues(wdata->mainWindow,
                  XmNx, (XtArgVal) ((x == 0) ? 1 : x),
                  XmNy, (XtArgVal) ((y == 0) ? 1 : y),
                  XmNwidth, (XtArgVal) w,
                  XmNheight, (XtArgVal) h,
                  NULL);

    inreshape = 0;
}

/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_pDispose
  (JNIEnv *env, jobject this) 
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->mainWindow == 0 || wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    XtUnmanageChild(wdata->winData.comp.widget);
    awt_delWidget(wdata->winData.comp.widget);
    awt_util_consumeAllXEvents(wdata->winData.comp.widget);
    XtDestroyWidget(wdata->winData.comp.widget);

    XtUnmanageChild(wdata->winData.shell);
    awt_util_consumeAllXEvents(wdata->mainWindow);
    awt_util_consumeAllXEvents(wdata->winData.shell);
    XtDestroyWidget(wdata->mainWindow);
    XtDestroyWidget(wdata->winData.shell);
    if ( wdata->isModal )
    {
        wdata->DlgIsDisposed = True;
        wdata->DlgIsWithdrawn = True;
        if ( wdata->DlgIsDone )
                free((void *)wdata);
    }
    else
    	free((void *) wdata);
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,0);
    awtJNI_DeleteGlobalRef(env, this);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    setResizable
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_setResizable
  (JNIEnv *env, jobject this, jboolean resizable) 
{
    struct FrameData *wdata;
    jobject target;

    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0 ||
        JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(wdata->winData.shell,
                  XmNnoResize, resizable ? False : True,
                  NULL);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MDialogPeer
 * Method:    toBack
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_toBack
  (JNIEnv *env, jobject this) 
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (XtWindow(wdata->winData.shell) != 0) {
        XLowerWindow(awt_display, XtWindow(wdata->winData.shell));
    }
    AWT_FLUSH_UNLOCK();
}

JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_addTextComponentNative
  (JNIEnv *env, jobject this)
{
    struct FrameData            *wdata;
    jobject                     target;

    if (JNU_IsNull(env, this)){
        return;
    }
    
    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);    

    if (wdata == 0 ||
        wdata->winData.comp.widget==0 ||
        wdata->winData.shell==0 ||
        JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    if (!wdata->hasTextComponentNative) {
	wdata->imHeight = awt_util_getIMStatusHeight(wdata->winData.shell); 
        wdata->bottom += wdata->imHeight;

        awtJNI_ChangeInsets(env, this, wdata);
        reshape(env, this, wdata,
		(*env)->GetIntField(env, target, componentIDs.x),
		(*env)->GetIntField(env, target, componentIDs.y),
		(*env)->GetIntField(env, target, componentIDs.width),
		(*env)->GetIntField(env, target, componentIDs.height));
        wdata->hasTextComponentNative = True;
    } 
    AWT_FLUSH_UNLOCK();
}

JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_removeTextComponentNative
  (JNIEnv *env, jobject this)
{
    struct FrameData            *wdata;
    jobject                     target;

    if (JNU_IsNull(env, this)){
        return;
    }
    
    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);    

    if (wdata == 0 ||
        wdata->winData.comp.widget==0 ||
        wdata->winData.shell==0 ||
        JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    if (!wdata->hasTextComponentNative) {
      AWT_UNLOCK();
      return;
    }

    wdata->bottom -= wdata->imHeight;
    awtJNI_ChangeInsets(env, this, wdata);
    wdata->hasTextComponentNative = False;
    wdata->imHeight = 0;
    reshape(env, this, wdata,
	    (*env)->GetIntField(env, target, componentIDs.x),
	    (*env)->GetIntField(env, target, componentIDs.y),
	    (*env)->GetIntField(env, target, componentIDs.width),
	    (*env)->GetIntField(env, target, componentIDs.height));
    AWT_FLUSH_UNLOCK();
}

JNIEXPORT jboolean JNICALL Java_sun_awt_motif_MDialogPeer_hasTextComponents
  (JNIEnv *env, jobject this)
{
    struct FrameData            *wdata;
    jobject                     target;

    if (JNU_IsNull(env, this)){
        return JNI_FALSE;
    }
    
    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);    

    if (wdata == 0 ||
        wdata->winData.comp.widget==0 ||
        wdata->winData.shell==0 ||
        JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return JNI_FALSE;
    }
    if (wdata->hasTextComponentNative) {
	AWT_UNLOCK();
	return JNI_TRUE;
    }      
    AWT_UNLOCK();
    return JNI_FALSE;
}


static Atom java_protocol = NULL;
static Atom motif_wm_msgs = NULL;
extern char* awt_util_makeWMMenuItem(char *target, Atom protocol);

/*
 * client_data is MDialogPeer instance pointer
 */
static void im_callback(Widget shell, XtPointer client_data, XtPointer call_data){
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    JNU_CallMethodByName(env, NULL,
			 (jobject)client_data,
			 "notifyIMMOptionChange",
			 "()V");
}

JNIEXPORT void JNICALL Java_sun_awt_motif_MDialogPeer_pSetIMMOption
  (JNIEnv *env, jobject this, jstring option)
{
    char        *coption;
    char        *empty = "IMM";
    char        *menuItem;
    jobject	globalRef = awtJNI_CreateAndSetGlobalRef(env, this);
    struct FrameData *wdata;

    AWT_LOCK();

    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 || wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    coption = (JNU_IsNull(env, option)) ? empty : (char *) JNU_GetStringPlatformChars(env, option, NULL);
    if (java_protocol == NULL || motif_wm_msgs == NULL){
      java_protocol = XmInternAtom(awt_display, "_JAVA_IM_MSG", False);
      motif_wm_msgs = XmInternAtom(awt_display, "_MOTIF_WM_MESSAGES", False);
    }
    XmAddProtocols (wdata->winData.shell, motif_wm_msgs, &java_protocol, 1);
    XmAddProtocolCallback(wdata->winData.shell, motif_wm_msgs, java_protocol, im_callback, (XtPointer)globalRef);

    if (menuItem = awt_util_makeWMMenuItem(coption, java_protocol)){
        XtVaSetValues(wdata->winData.shell, 
		      XmNmwmMenu, 
		      menuItem,
		      NULL); 
	free(menuItem); 
    }
    if (coption != empty)
      JNU_ReleaseStringPlatformChars(env, option, (const char *) coption);
    AWT_FLUSH_UNLOCK();
}
