/*
 * @(#)awt_DrawingSurface.c	1.17 98/07/31
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_p.h"
#include "java_awt_Component.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "sun_awt_motif_MDrawingSurfaceInfo.h"

#include "awt_Component.h"

#include <jni.h>
#include <jni_util.h>

extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct ComponentIDs componentIDs;

/* fieldIDs for MDrawingSurfaceInfo fields that may be accessed from C */
static struct MDrawingSurfaceInfoIDs {
    jfieldID state;
    jfieldID peer;
    jfieldID w;
    jfieldID h;
} mDrawingSurfaceInfoIDs;

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MDrawingSurfaceInfo.java to initialize the fieldIDs for fields
   that may be accessed from C */

JNIEXPORT void JNICALL
Java_sun_awt_motif_MDrawingSurfaceInfo_initIDs
  (JNIEnv *env, jclass cls)
{
    mDrawingSurfaceInfoIDs.state =
      (*env)->GetFieldID(env, cls, "state", "I");
    mDrawingSurfaceInfoIDs.peer =
      (*env)->GetFieldID(env, cls, "peer", "Lsun/awt/motif/MComponentPeer;");
    mDrawingSurfaceInfoIDs.w =
      (*env)->GetFieldID(env, cls, "w", "I");
    mDrawingSurfaceInfoIDs.h =
      (*env)->GetFieldID(env, cls, "h", "I");
}

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    lock
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MDrawingSurfaceInfo_lock
  (JNIEnv *env, jobject this) 
{
    jobject hPeer, hComp;
    jint state;
    jint thisW, thisH;
    jint peerW, peerH;

    state = (*env)->GetIntField(env, this, mDrawingSurfaceInfoIDs.state);

    AWT_LOCK();

    hPeer = (*env)->GetObjectField(env, this, mDrawingSurfaceInfoIDs.peer);
    hComp = (*env)->GetObjectField(env, hPeer, mComponentPeerIDs.target);
    
    if (!JNU_IsNull(env, hPeer)) {
        thisW = (*env)->GetIntField(env, this, mDrawingSurfaceInfoIDs.w);
        thisH = (*env)->GetIntField(env, this, mDrawingSurfaceInfoIDs.h);
        peerH = (*env)->GetIntField(env, hComp, componentIDs.width);
        peerH = (*env)->GetIntField(env, hComp, componentIDs.height);

        if (thisW != peerW || thisH != peerH) {
            (*env)->SetIntField(env, this, mDrawingSurfaceInfoIDs.w, peerW);
            (*env)->SetIntField(env, this, mDrawingSurfaceInfoIDs.h, peerH);
            (*env)->SetIntField(env, this, mDrawingSurfaceInfoIDs.state, ++state);
        }
    }
    return state;
}

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    unlock
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MDrawingSurfaceInfo_unlock
  (JNIEnv *env, jobject this) 
{
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    getDisplay
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MDrawingSurfaceInfo_getDisplay
  (JNIEnv *env, jobject this) 
{
    return (jint) awt_display;
}

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    getDrawable
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MDrawingSurfaceInfo_getDrawable
  (JNIEnv *env, jobject this) 
{
    Drawable drawable = 0;
    jobject peer;

    AWT_LOCK();

    peer = (*env)->GetObjectField(env, this, mDrawingSurfaceInfoIDs.peer);

    if (!JNU_IsNull(env, peer)) {
        struct CanvasData *cdata;

        cdata = (struct CanvasData *)
	  (*env)->GetLongField(env,peer,mComponentPeerIDs.pData);
        if (cdata != 0) {
            Widget win = cdata->comp.widget;

            drawable = XtWindow(win);
        }
    }

    AWT_UNLOCK();

    if (drawable == 0) {
        JNU_ThrowNullPointerException(env, "unable to determine Drawable for DrawingSurface");
    }
    return (jint) drawable;
}

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    getDepth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MDrawingSurfaceInfo_getDepth
  (JNIEnv *env, jobject this) 
{
    XWindowAttributes attrs;
    Drawable d = Java_sun_awt_motif_MDrawingSurfaceInfo_getDrawable(env, this);

    /* TODO no visual with a Pixmap !! */
    XGetWindowAttributes(awt_display, (Window)d, &attrs);
    return (jint) attrs.depth;
}

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    getVisualID
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MDrawingSurfaceInfo_getVisualID
  (JNIEnv *env, jobject this)
{
    XWindowAttributes attrs;
    Drawable d = Java_sun_awt_motif_MDrawingSurfaceInfo_getDrawable(env, this);

    /* TODO no visual with a Pixmap !! */
    XGetWindowAttributes(awt_display, (Window)d, &attrs);
    return (jint) XVisualIDFromVisual(attrs.visual);
}

/*
 * Class:     sun_awt_motif_MDrawingSurfaceInfo
 * Method:    getColormapID
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MDrawingSurfaceInfo_getColormapID
  (JNIEnv *env, jobject this)
{
    XWindowAttributes attrs;
    Drawable d = Java_sun_awt_motif_MDrawingSurfaceInfo_getDrawable(env, this);

    /* TODO no visual with a Pixmap !! */
    XGetWindowAttributes(awt_display, (Window)d, &attrs);
    return (jint) attrs.colormap;
}
