/*
 * @(#)awt_FileDialog.c	1.49 98/10/02
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include <Xm/AtomMgr.h>
#include <Xm/Protocols.h>
#include <sys/param.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "awt_p.h"
#include "java_awt_FileDialog.h"
#include "sun_awt_motif_MFileDialogPeer.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "multi_font.h"

#include "awt_Component.h"

#include <jni.h>
#include <jni_util.h>

#define MAX_DIR_PATH_LEN    1024

extern void Text_handlePaste(Widget w, XtPointer client_data, XEvent * event,
                             Boolean * cont);

extern struct MComponentPeerIDs mComponentPeerIDs;

/* fieldIDs for FileDialog fields and methods that may be accessed from C */
static struct FileDialogIDs {
    jfieldID mode;
    jfieldID file;
    jmethodID setDirectoryID;
    jmethodID setFileID;
} fileDialogIDs;

/*
 * Class:     java_awt_FileDialog
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for FileDialog.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_FileDialog_initIDs
  (JNIEnv *env, jclass cls)
{
    fileDialogIDs.mode = (*env)->GetFieldID(env, cls, "mode", "I");
    fileDialogIDs.file =
      (*env)->GetFieldID(env, cls, "file", "Ljava/lang/String;");
    fileDialogIDs.setDirectoryID = 
      (*env)->GetMethodID(env, cls, "setDirectory_NoClientCode",
			  "(Ljava/lang/String;)V");
    fileDialogIDs.setFileID = 
      (*env)->GetMethodID(env, cls, "setFile_NoClientCode",
			  "(Ljava/lang/String;)V");

    assert(fileDialogIDs.mode != NULL);
    assert(fileDialogIDs.file != NULL);
    assert(fileDialogIDs.setDirectoryID != NULL);
    assert(fileDialogIDs.setFileID != NULL);
}

/*
 * client_data is MFileDialogPeer instance pointer
 */
static void
FileDialog_OK(Widget w,
              void *client_data,
              XmFileSelectionBoxCallbackStruct * call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject) client_data;
    struct FrameData *fdata;
    char *file;
    jstring jstr;

    fdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    XmStringGetLtoR(call_data->value, XmSTRING_DEFAULT_CHARSET, &file);

    if ((*env)->EnsureLocalCapacity(env, 1) < 0)
        return;
    jstr = JNU_NewStringPlatform(env, (const char *) file);
    if (jstr != 0) {
	JNU_CallMethodByName(env, NULL, this, "handleSelected",
			     "(Ljava/lang/String;)V", jstr);
	(*env)->DeleteLocalRef(env, jstr);
    }
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    XtUnmanageChild(fdata->winData.comp.widget);
}

/*
 * client_data is MFileDialogPeer instance pointer
 */
static void
FileDialog_CANCEL(Widget w,
                  void *client_data,
                  XmFileSelectionBoxCallbackStruct * call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject) client_data;
    struct FrameData *fdata;

    fdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "handleCancel", "()V");
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    XtUnmanageChild(fdata->winData.comp.widget);
}


/*
 * client_data is MFileDialogPeer instance pointer
 */
static void
FileDialog_quit(Widget w,
                XtPointer client_data,
                XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "handleQuit", "()V");
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

static void
setDeleteCallback(jobject this, struct FrameData *wdata)
{
    Atom xa_WM_DELETE_WINDOW;
    Atom xa_WM_PROTOCOLS;

    XtVaSetValues(wdata->winData.shell,
                  XmNdeleteResponse, XmDO_NOTHING,
                  NULL);
    xa_WM_DELETE_WINDOW = XmInternAtom(XtDisplay(wdata->winData.shell),
                                       "WM_DELETE_WINDOW", False);
    xa_WM_PROTOCOLS = XmInternAtom(XtDisplay(wdata->winData.shell),
                                   "WM_PROTOCOLS", False);

    XmAddProtocolCallback(wdata->winData.shell,
                          xa_WM_PROTOCOLS,
                          xa_WM_DELETE_WINDOW,
                          FileDialog_quit, (XtPointer) this);
}

void
setFSBDirAndFile(Widget w, char *dir, char *file)
{
    Widget textField;
    char dirbuf[MAX_DIR_PATH_LEN];
    XmString xim;
    int lastSelect;

    dirbuf[0] = (char) '\0';

    if (dir != 0)
        strcpy(dirbuf, dir);

    /* -----> make sure dir ends in '/' */
    if (dirbuf[0] != (char) '\0') {
        if (dirbuf[strlen(dirbuf) - 1] != (char) '/')
            strcat(dirbuf, "/");
    } else {
        getcwd(dirbuf, MAX_DIR_PATH_LEN - 16);
        strcat(dirbuf, "/");
    }

    strcat(dirbuf, "[^.]*");
    xim = XmStringCreateLtoR(dirbuf, XmSTRING_DEFAULT_CHARSET);
    XtVaSetValues(w, XmNdirMask, xim, NULL);
    XmStringFree(xim);
    textField = XmFileSelectionBoxGetChild(w, XmDIALOG_TEXT);
    if (textField != 0 && file != 0) {
        lastSelect = strlen(file);
        XtVaSetValues(textField, XmNvalue, file, NULL);
        XmTextFieldSetSelection(textField, 0, lastSelect, CurrentTime);
    }
}

static void
changeBackground(Widget w, void *bg)
{
    XmChangeColor(w, (Pixel) bg);
}

static int
WaitForUnmap(void *data)
{
    return (XtWindow((Widget) data) && !XtIsManaged((Widget) data));
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_create
  (JNIEnv *env, jobject this, jobject parent)
{
    struct FrameData *fdata;
    struct CanvasData *wdata;
    int argc;
    Arg args[10];
    Widget child, textField;
    XmString xim;
    Pixel bg;
    jobject target;
    jstring file;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    if (JNU_IsNull(env, parent) || JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();

    wdata = (struct CanvasData *) (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);
    fdata = ZALLOC(FrameData);
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong)fdata);

    if (fdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    XtVaGetValues(wdata->comp.widget, XmNbackground, &bg, NULL);

    argc = 0;
    XtSetArg(args[argc], XmNmustMatch, False);
    argc++;
    XtSetArg(args[argc], XmNautoUnmanage, False);
    argc++;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNvisual, defaultConfig->awt_visInfo.visual);
    argc++;
    XtSetArg(args[argc], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
    argc++;
    fdata->winData.comp.widget = XmCreateFileSelectionDialog(wdata->shell,
                                                             "",
                                                             args,
                                                             argc);
    fdata->winData.shell = XtParent(fdata->winData.comp.widget);
    awt_util_mapChildren(fdata->winData.shell, changeBackground, 0,
			 (void *) bg);
    child = XmFileSelectionBoxGetChild(fdata->winData.comp.widget,
                                       XmDIALOG_HELP_BUTTON);
    /*
     * Get textfield in FileDialog.
     */
    textField = XmFileSelectionBoxGetChild(fdata->winData.comp.widget,
					   XmDIALOG_TEXT);
    if (child != 0) {
        XtUnmanageChild(child);
    }
    if (!awtJNI_IsMultiFont(env, awtJNI_GetFont(env, this))) {
        /* This process should not be done other than English language
           locale. */
        child = XmFileSelectionBoxGetChild(fdata->winData.comp.widget,
                                           XmDIALOG_DEFAULT_BUTTON);
        if (child != 0) {
            XmString xim;

            switch ((*env)->GetIntField(env, target, fileDialogIDs.mode)) {
                case java_awt_FileDialog_LOAD:
                    xim = XmStringCreateLtoR("Open", "labelFont");
                    XtVaSetValues(child, XmNlabelString, xim, NULL);
                    XmStringFree(xim);
                    break;

                case java_awt_FileDialog_SAVE:
                    xim = XmStringCreateLtoR("Save", "labelFont");
                    XtVaSetValues(child, XmNlabelString, xim, NULL);
                    XmStringFree(xim);
                    break;

                default:
                    break;
            }
        }
    }
    XtAddCallback(fdata->winData.comp.widget,
                  XmNokCallback,
                  (XtCallbackProc) FileDialog_OK,
                  (XtPointer) globalRef);
    XtAddCallback(fdata->winData.comp.widget,
                  XmNcancelCallback,
                  (XtCallbackProc) FileDialog_CANCEL,
                  (XtPointer) globalRef);

    setDeleteCallback(globalRef, fdata);

    if (textField != NULL)  {
	/*
	 * Insert event handler to correctly process cut/copy/paste keys 
	 * such that interaction with our own clipboard mechanism will work 
	 * properly.
	 *
 	 * The Text_handlePaste() event handler is also used by both 
	 * TextField/TextArea.
	 */
        XtInsertEventHandler(textField,
                         KeyPressMask,
                         False, Text_handlePaste, (XtPointer) globalRef,
                         XtListHead);
    }

    file = (*env)->GetObjectField(env, target, fileDialogIDs.file);
    if (JNU_IsNull(env, file)) {
        setFSBDirAndFile(fdata->winData.comp.widget, ".", "");
    } else {
        char *fileString;

        fileString = (char *) JNU_GetStringPlatformChars(env, file, NULL);
        setFSBDirAndFile(fdata->winData.comp.widget, ".", fileString);
        JNU_ReleaseStringPlatformChars(env, file, (const char *) fileString);
    }
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    pReshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_pReshape
  (JNIEnv *env, jobject this, jint x, jint y, jint w, jint h)
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    /* GES: AVH's hack from awt_util.c:
     * Motif ignores attempts to move a toplevel window to 0,0.
     * Instead we set the position to 1,1. The expected value is
     * returned by Frame.getBounds() since it uses the internally
     * held rectangle rather than querying the peer.
     */

    if ((x == 0) && (y == 0)) {
        XtVaSetValues(wdata->winData.shell, XmNx, 1, XmNy, 1, NULL);
    }
    XtVaSetValues(wdata->winData.shell,
                  XtNx, (XtArgVal) x,
                  XtNy, (XtArgVal) y,
                  NULL);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_pDispose
  (JNIEnv *env, jobject this)
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtUnmanageChild(wdata->winData.shell);
    awt_util_consumeAllXEvents(wdata->winData.shell);
    XtDestroyWidget(wdata->winData.shell);
    free((void *) wdata);
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,0);
    awtJNI_DeleteGlobalRef(env, this);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    pShow
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_pShow
  (JNIEnv *env, jobject this)
{
    struct FrameData *wdata;
    XmString dirMask = NULL;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtManageChild(wdata->winData.comp.widget);

    AWT_FLUSH_UNLOCK();

    awt_MToolkit_modalWait(WaitForUnmap, wdata->winData.comp.widget);
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    pHide
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_pHide
  (JNIEnv *env, jobject this)
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (XtIsManaged(wdata->winData.comp.widget)) {
        XtUnmanageChild(wdata->winData.comp.widget);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    setFileEntry
 * Signature: (Ljava/lang/String;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_setFileEntry
  (JNIEnv *env, jobject this, jstring dir, jstring file)
{
    struct ComponentData *cdata;
    char *cdir;
    char *cfile;
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *) 
        (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }

    cdir = (JNU_IsNull(env, dir)) 
               ? NULL
               : (char *) JNU_GetStringPlatformChars(env, dir, NULL);

    cfile = (JNU_IsNull(env, file)) 
               ? NULL
               : (char *) JNU_GetStringPlatformChars(env, file, NULL);

    setFSBDirAndFile(wdata->winData.comp.widget, (cdir) ? cdir : "",
		     (cfile) ? cfile : "");

    if (cdir) {
        JNU_ReleaseStringPlatformChars(env, dir, (const char *) cdir);
    }
    if (cfile) {
        JNU_ReleaseStringPlatformChars(env, file, (const char *) cfile);
    }
    AWT_FLUSH_UNLOCK();
}

static void
changeFont(Widget w, void *fontList)
{
    XtVaSetValues(w, XmNfontList, fontList, NULL);
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    setFont
 * Signature: (Ljava/awt/Font;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_setFont
  (JNIEnv *env, jobject this, jobject f)
{
    struct ComponentData *tdata;
    struct FontData *fdata;
    XmFontListEntry fontentry;
    XmFontList fontlist;
    char *err;

    if (JNU_IsNull(env, f)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    fdata = awtJNI_GetFontData(env, f, &err);
    if (fdata == 0) {
        JNU_ThrowInternalError(env, err);
        AWT_UNLOCK();
        return;
    }
    tdata = (struct ComponentData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (awtJNI_IsMultiFont(env, f)) {
        if (fdata->xfs == NULL) {
            fdata->xfs = awtJNI_MakeFontSet(env, f);
        }
        if (fdata->xfs != NULL) {
            fontentry = XmFontListEntryCreate("labelFont",
                                              XmFONT_IS_FONTSET,
                                              (XtPointer) (fdata->xfs));
            fontlist = XmFontListAppendEntry(NULL, fontentry);
	    /*
	     * Some versions of motif have a bug in 
	     * XmFontListEntryFree() which causes it to free more than it
	     * should.  Use XtFree() instead.  See O'Reilly's
	     * Motif Reference Manual for more information.
	     */
	    XmFontListEntryFree(&fontentry);
        } else {
            fontlist = XmFontListCreate(fdata->xfont, "labelFont");
        }
    } else {
        fontlist = XmFontListCreate(fdata->xfont, "labelFont");
    }

    if (fontlist != NULL) {
     /* setting the fontlist in the FileSelectionBox is not good enough --
        you have to set the resource for all the descendants individually */
        awt_util_mapChildren(tdata->widget, changeFont, 1, (void *)fontlist);
        XmFontListFree(fontlist);
    } else {
        JNU_ThrowNullPointerException(env, "NullPointerException");
    }

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    insertReplaceFileDialogText
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFileDialogPeer_insertReplaceFileDialogText
  (JNIEnv *env, jobject this, jstring l) 
{
    struct ComponentData *cdata;
    char *cl;
    XmTextPosition start, end;
    Widget textField;
    jobject font;

    /*
     * Replaces the text in the FileDialog's textfield with the passed
     * string.
     */

    AWT_LOCK();
    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    textField = XmFileSelectionBoxGetChild(cdata->widget, XmDIALOG_TEXT);

    if (textField == NULL)  {
        JNU_ThrowNullPointerException(env, "Null TextField in FileDialog");
        AWT_UNLOCK();
        return;
    }

    font = awtJNI_GetFont(env, this);

    if (JNU_IsNull(env, l)) {
        cl = "";
    } else {
        /*
         * We use makePlatformCString() to convert unicode to EUC here,
         * although output only components (Label/Button/Menu..)
         * is not using make/allocCString() functions anymore.
         * Because Motif TextFiled widget does not support multi-font
         * compound string.
         */

        cl = (char *) JNU_GetStringPlatformChars(env, l, NULL);
    }

    if (!XmTextGetSelectionPosition(textField, &start, &end)) {
        start = end = XmTextGetCursorPosition(textField);
    }
    XmTextReplace(textField, start, end, cl);

    if (cl && cl != "") {
        JNU_ReleaseStringPlatformChars(env, l, cl);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    targetSetDirectory_NoClientCode
 * Signature: (Ljava/awt/FileDialog;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_motif_MFileDialogPeer_targetSetDirectory_1NoClientCode(
    JNIEnv *env,
    jobject peer,
    jobject target,
    jstring dir)
{
    (*env)->CallVoidMethod(env, target, fileDialogIDs.setDirectoryID, dir);
}

/*
 * Class:     sun_awt_motif_MFileDialogPeer
 * Method:    targetSetFile_NoClientCode
 * Signature: (Ljava/awt/FileDialog;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_motif_MFileDialogPeer_targetSetFile_1NoClientCode(
    JNIEnv *env,
    jobject peer,
    jobject target,
    jstring file)
{
    (*env)->CallVoidMethod(env, target, fileDialogIDs.setFileID, file);
}
