/*
 * @(#)awt_Font.c	1.48 98/09/22
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include <string.h>
#include "java_awt_Component.h"
#include "java_awt_Font.h"
#include "java_awt_FontMetrics.h"
#include "sun_awt_motif_MToolkit.h"
#include "sun_awt_motif_X11FontMetrics.h"
#include "sun_awt_PlatformFont.h"

#include "awt_Font.h"

#include "java_awt_Dimension.h"
#include "multi_font.h"
#include <jni.h>
#include <jni_util.h>

#define defaultXLFD "-*-helvetica-*-*-*-*-*-*-12-*-*-*-iso8859-1"

struct FontIDs fontIDs;
struct PlatformFontIDs platformFontIDs;

/*
 * Class:     java_awt_Font
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for Font.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_Font_initIDs
  (JNIEnv *env, jclass cls)
{
    /** We call "NoClientCode" methods because they won't invoke client
	code on the privileged toolkit thread **/
    fontIDs.pData = (*env)->GetFieldID(env, cls, "pData", "J");
    fontIDs.style = (*env)->GetFieldID(env, cls, "style", "I");
    fontIDs.size = (*env)->GetFieldID(env, cls, "size", "I");
    fontIDs.getPeer = (*env)->GetMethodID(env, cls, "getPeer_NoClientCode",
					   "()Ljava/awt/peer/FontPeer;");
    fontIDs.getFamily =
      (*env)->GetMethodID(env, cls, "getFamily_NoClientCode", 
					    "()Ljava/lang/String;");
}


/* fieldIDs for X11FontMetrics fields that may be accessed from C */
static struct X11FontMetricsIDs {
    jfieldID widths;
    jfieldID font;
    jfieldID ascent;
    jfieldID descent;
    jfieldID leading;
    jfieldID height;
    jfieldID maxAscent;
    jfieldID maxDescent;
    jfieldID maxHeight;
    jfieldID maxAdvance;
} x11FontMetricsIDs;

/*
 * Class:     sun_awt_motif_X11FontMetrics
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   X11FontMetrics.java to initialize the fieldIDs for fields
   that may be accessed from C */

JNIEXPORT void JNICALL
Java_sun_awt_motif_X11FontMetrics_initIDs
  (JNIEnv *env, jclass cls)
{
    x11FontMetricsIDs.widths = (*env)->GetFieldID(env, cls, "widths", "[I");
    x11FontMetricsIDs.font =
      (*env)->GetFieldID(env, cls, "font", "Ljava/awt/Font;");
    x11FontMetricsIDs.ascent =
      (*env)->GetFieldID(env, cls, "ascent", "I");
    x11FontMetricsIDs.descent =
      (*env)->GetFieldID(env, cls, "descent", "I");
    x11FontMetricsIDs.leading =
      (*env)->GetFieldID(env, cls, "leading", "I");
    x11FontMetricsIDs.height =
      (*env)->GetFieldID(env, cls, "height", "I");
    x11FontMetricsIDs.maxAscent =
      (*env)->GetFieldID(env, cls, "maxAscent", "I");
    x11FontMetricsIDs.maxDescent =
      (*env)->GetFieldID(env, cls, "maxDescent", "I");
    x11FontMetricsIDs.maxHeight =
      (*env)->GetFieldID(env, cls, "maxHeight", "I");
    x11FontMetricsIDs.maxAdvance =
      (*env)->GetFieldID(env, cls, "maxAdvance", "I");
}


/* fieldIDs for FontDescriptor fields that may be accessed from C */
static struct FontDescriptorIDs {
    jfieldID nativeName;
    jfieldID fontCharset;
} fontDescriptorIDs;

/*
 * Class:     sun_awt_FontDescriptor
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   FontDescriptor.java to initialize the fieldIDs for fields
   that may be accessed from C */

JNIEXPORT void JNICALL
Java_sun_awt_FontDescriptor_initIDs
  (JNIEnv *env, jclass cls)
{
    fontDescriptorIDs.nativeName =
      (*env)->GetFieldID(env, cls, "nativeName",
			 "Ljava/lang/String;");
    fontDescriptorIDs.fontCharset =
      (*env)->GetFieldID(env, cls, "fontCharset",
			 "Lsun/io/CharToByteConverter;");
}

struct MFontPeerIDs mFontPeerIDs;
/*
 * Class:     sun_awt_motif_MFontPeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MFontPeer.java to initialize the fieldIDs for fields
   that may be accessed from C */

JNIEXPORT void JNICALL
Java_sun_awt_motif_MFontPeer_initIDs
  (JNIEnv *env, jclass cls)
{
    mFontPeerIDs.xfsname =
      (*env)->GetFieldID(env, cls, "xfsname", "Ljava/lang/String;");
}

/*
 * Class:     sun_awt_PlatformFont
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   PlatformFont.java to initialize the fieldIDs for fields
   that may be accessed from C */

JNIEXPORT void JNICALL
Java_sun_awt_PlatformFont_initIDs
  (JNIEnv *env, jclass cls)
{
    platformFontIDs.componentFonts =
      (*env)->GetFieldID(env, cls, "componentFonts",
			 "[Lsun/awt/FontDescriptor;");
    platformFontIDs.props =
      (*env)->GetFieldID(env,cls, "props",
			 "Ljava/util/Properties;");

    platformFontIDs.makeConvertedMultiFontString = 
      (*env)->GetMethodID(env, cls, "makeConvertedMultiFontString",
			  "(Ljava/lang/String;)[Ljava/lang/Object;");

    platformFontIDs.makeConvertedMultiFontChars =
      (*env)->GetMethodID(env, cls, "makeConvertedMultiFontChars",
			  "([CII)[Ljava/lang/Object;");
}

XFontStruct *
loadFont(Display * display, char *name, int pointSize)
{

    XFontStruct *f;

    /*
     * try the exact xlfd name in for.properties<.lang>
     */
    f = XLoadQueryFont(display, name);
    if (f != NULL) {
        return f;
    }
    /*
     * try nearly font
     *
     *  1. specify FAMILY_NAME, WEIGHT_NAME, SLANT, POINT_SIZE,
     *     CHARSET_REGISTRY and CHARSET_ENCODING.
     *  2. change POINT_SIZE to PIXEL_SIZE
     *  3. change FAMILY_NAME to *
     *  4. specify only PIXEL_SIZE and CHARSET_REGISTRY/ENCODING
     *  5. change PIXEL_SIZE +1/-1/+2/-2...+4/-4
     *  6. default font pattern
     */
    {
        /*
         * This code assumes the name contains exactly 14 '-' delimiter.
         * If not use default pattern.
         */
        int i, length, pixelSize;
        Boolean useDefault = FALSE;

        char buffer[BUFSIZ], buffer2[BUFSIZ], *family, *style, *slant, *encoding, *start, *end;

        if (strlen(name) > BUFSIZ - 1) {
            useDefault = TRUE;
        } else {
            strcpy(buffer, name);
        }

#define NEXT_HYPHEN\
        start = end + 1;\
        end = strchr(start, '-');\
        if (end == NULL) {\
                              useDefault = TRUE;\
        break;\
        }\
        *end = '\0'

             do {
                 end = buffer;

                 /* skip FOUNDRY */
                 NEXT_HYPHEN;

                 /* set FAMILY_NAME */
                 NEXT_HYPHEN;
                 family = start;

                 /* set STYLE_NAME */
                 NEXT_HYPHEN;
                 style = start;

                 /* set SLANT */
                 NEXT_HYPHEN;
                 slant = start;

                 /* skip SETWIDTH_NAME, ADD_STYLE_NAME, PIXEL_SIZE
                    POINT_SIZE, RESOLUTION_X, RESOLUTION_Y, SPACING
                    and AVERAGE_WIDTH */
                 NEXT_HYPHEN;
                 NEXT_HYPHEN;
                 NEXT_HYPHEN;
                 NEXT_HYPHEN;
                 NEXT_HYPHEN;
                 NEXT_HYPHEN;
                 NEXT_HYPHEN;
                 NEXT_HYPHEN;

                 /* set CHARSET_REGISTRY and CHARSET_ENCODING */
                 encoding = end + 1;
             }
             while (0);

#define TRY_LOAD\
        f = XLoadQueryFont(display, buffer2);\
        if (f != NULL) {\
                            strcpy(name, buffer2);\
        return f;\
        }

        if (useDefault == FALSE) {
            char *altstyle = NULL;
            /*
             * Regular is the style for TrueType fonts -- Type1, F3 use roman
             */
            if (!strcmp(style, "regular")) {
                altstyle = "roman";
            }
            
            /* try 1. */
            jio_snprintf(buffer2, sizeof(buffer2),
                         "-*-%s-%s-%s-*-*-*-%d-*-*-*-*-%s",
                         family, style, slant, pointSize, encoding);
            TRY_LOAD;

            if (altstyle) {
                jio_snprintf(buffer2, sizeof(buffer2),
                             "-*-%s-%s-%s-*-*-*-%d-*-*-*-*-%s",
                             family, altstyle, slant, pointSize, encoding);
                TRY_LOAD;
            }
                
            /* search bitmap font */
            pixelSize = pointSize / 10;

            /* try 2. */
            jio_snprintf(buffer2, sizeof(buffer2),
                         "-*-%s-%s-%s-*-*-%d-*-*-*-*-*-%s",
                         family, style, slant, pixelSize, encoding);
            TRY_LOAD;

            if (altstyle) {
                jio_snprintf(buffer2, sizeof(buffer2),
                             "-*-%s-%s-%s-*-*-%d-*-*-*-*-*-%s",
                             family, altstyle, slant, pixelSize, encoding);
                TRY_LOAD;
            }
            /* try 3 */
            jio_snprintf(buffer2, sizeof(buffer2),
                         "-*-*-%s-%s-*-*-%d-*-*-*-*-*-%s",
                         style, slant, pixelSize, encoding);

            TRY_LOAD;
            
            if (altstyle) {
                jio_snprintf(buffer2, sizeof(buffer2),
                             "-*-*-%s-%s-*-*-%d-*-*-*-*-*-%s",
                             altstyle, slant, pixelSize, encoding);
                TRY_LOAD;
            }
            
            /* try 4 */
            jio_snprintf(buffer2, sizeof(buffer2),
                         "-*-*-*-%s-*-*-%d-*-*-*-*-*-%s",
                         slant, pixelSize, encoding);

            TRY_LOAD;
            
            /* try 5. */
            jio_snprintf(buffer2, sizeof(buffer2),
                         "-*-*-*-*-*-*-%d-*-*-*-*-*-%s",
                         pixelSize, encoding);
            TRY_LOAD;

            /* try 6. */
            for (i = 1; i < 4; i++) {
                if (pixelSize < i)
                    break;
                jio_snprintf(buffer2, sizeof(buffer2),
                             "-*-%s-%s-%s-*-*-%d-*-*-*-*-*-%s",
                             family, style, slant, pixelSize + i, encoding);
                TRY_LOAD;

                jio_snprintf(buffer2, sizeof(buffer2),
                             "-*-%s-%s-%s-*-*-%d-*-*-*-*-*-%s",
                             family, style, slant, pixelSize - i, encoding);
                TRY_LOAD;

                jio_snprintf(buffer2, sizeof(buffer2),
                             "-*-*-*-*-*-*-%d-*-*-*-*-*-%s",
                             pixelSize + i, encoding);
                TRY_LOAD;

                jio_snprintf(buffer2, sizeof(buffer2),
                             "-*-*-*-*-*-*-%d-*-*-*-*-*-%s",
                             pixelSize - i, encoding);
                TRY_LOAD;
            }
        }
    }

    strcpy(name, defaultXLFD);
    return XLoadQueryFont(display, defaultXLFD);
}

/*
 * Hardwired list of mappings for generic font names "Helvetica",
 * "TimesRoman", "Courier", "Dialog", and "DialogInput".
 */
static char *defaultfontname = "fixed";
static char *defaultfoundry = "misc";
static char *anyfoundry = "*";
static char *anystyle = "*-*";
static char *isolatin1 = "iso8859-1";

static char *
Style(long s)
{
    switch (s) {
        case java_awt_Font_ITALIC:
            return "medium-i";
        case java_awt_Font_BOLD:
            return "bold-r";
        case java_awt_Font_BOLD + java_awt_Font_ITALIC:
            return "bold-i";
        case java_awt_Font_PLAIN:
        default:
            return "medium-r";
    }
}

static int
awtJNI_FontName(JNIEnv * env, jstring name, char **foundry, char **facename, char **encoding)
{
    char *cname;

    if (JNU_IsNull(env, name)) {
        return 0;
    }
    cname = (char *) JNU_GetStringPlatformChars(env, name, NULL);

    /* additional default font names */
    if (!strcmp(cname, "serif")) {
        *foundry = "adobe";
        *facename = "times";
        *encoding = isolatin1;
    } else if (!strcmp(cname, "sansserif")) {
        *foundry = "adobe";
        *facename = "helvetica";
        *encoding = isolatin1;
    } else if (!strcmp(cname, "monospaced")) {
        *foundry = "adobe";
        *facename = "courier";
        *encoding = isolatin1;
    } else if (strcmp(cname, "helvetica") == 0) {
        *foundry = "adobe";
        *facename = "helvetica";
        *encoding = isolatin1;
    } else if (strcmp(cname, "timesroman") == 0) {
        *foundry = "adobe";
        *facename = "times";
        *encoding = isolatin1;
    } else if (strcmp(cname, "courier") == 0) {
        *foundry = "adobe";
        *facename = "courier";
        *encoding = isolatin1;
    } else if (strcmp(cname, "dialog") == 0) {
        *foundry = "b&h";
        *facename = "lucida";
        *encoding = isolatin1;
    } else if (strcmp(cname, "dialoginput") == 0) {
        *foundry = "b&h";
        *facename = "lucidatypewriter";
        *encoding = isolatin1;
    } else if (strcmp(cname, "zapfdingbats") == 0) {
        *foundry = "itc";
        *facename = "zapfdingbats";
        *encoding = "*-*";
    } else {
#ifdef DEBUG
        jio_fprintf(stderr, "Unknown font: %s\n", cname);
#endif
        *foundry = defaultfoundry;
        *facename = defaultfontname;
        *encoding = isolatin1;
    }

    if (cname)
        JNU_ReleaseStringPlatformChars(env, name, (const char *) cname);

    return 1;
}

struct FontData *
awtJNI_GetFontData(JNIEnv * env, jobject font, char **errmsg)
{
    /* We are going to create at most 4 outstanding local refs in this
     * function. */
    if ((*env)->EnsureLocalCapacity(env, 4) < 0)
        return NULL;

    if (!JNU_IsNull(env, font) && awtJNI_IsMultiFont(env, font)) {
        struct FontData *fdata;
        int i, size;
        char *fontsetname;
        char *nativename;
        jobjectArray componentFonts;
        jobject peer;
        jobject fontDescriptor;
        jstring fontDescriptorName;
        jobject charToByteConverter;
        jstring charToByteConverterName;

        fdata = (struct FontData *) (*env)->GetLongField(env,font,
							 fontIDs.pData);

        if (fdata != NULL && fdata->flist != NULL) {
            return fdata;
        }
        size = (*env)->GetIntField(env, font, fontIDs.size);
        fdata = (struct FontData *) malloc(sizeof(struct FontData));

        peer = (*env)->CallObjectMethod(env, font, fontIDs.getPeer);

        componentFonts =
	  (*env)->GetObjectField(env, peer, platformFontIDs.componentFonts);
	/* We no longer need peer */
	(*env)->DeleteLocalRef(env, peer);

        fdata->charset_num = (*env)->GetArrayLength(env, componentFonts);

        fdata->flist = (awtFontList *) malloc(sizeof(awtFontList)
                                              * fdata->charset_num);
        fdata->xfont = NULL;
        for (i = 0; i < fdata->charset_num; i++) {
            /*
             * set xlfd name
             */

            fontDescriptor = (*env)->GetObjectArrayElement(env, componentFonts, i);
            fontDescriptorName =
	      (*env)->GetObjectField(env, fontDescriptor,
				     fontDescriptorIDs.nativeName);

            if (!JNU_IsNull(env, fontDescriptorName)) {
                nativename = (char *) JNU_GetStringPlatformChars(env, fontDescriptorName, NULL);
            } else {
                nativename = "";
            }

            fdata->flist[i].xlfd = malloc(strlen(nativename)
                                          + strlen(defaultXLFD));
            jio_snprintf(fdata->flist[i].xlfd, strlen(nativename) + 10,
                         nativename, size * 10);

            if (nativename && (nativename != ""))
                JNU_ReleaseStringPlatformChars(env, fontDescriptorName, (const char *) nativename);

            /*
             * set charset_name
             */

            charToByteConverter =
	      (*env)->GetObjectField(env, fontDescriptor,
				     fontDescriptorIDs.fontCharset);

            charToByteConverterName =
	        JNU_CallMethodByName(env,
				     NULL,
				     charToByteConverter
				     ,"toString"
				     ,"()Ljava/lang/String;").l;

            fdata->flist[i].charset_name = (char *)
                JNU_GetStringPlatformChars(env, charToByteConverterName, NULL);

	    /* We are done with the objects. */
	    (*env)->DeleteLocalRef(env, fontDescriptor);
	    (*env)->DeleteLocalRef(env, fontDescriptorName);
	    (*env)->DeleteLocalRef(env, charToByteConverter);
	    (*env)->DeleteLocalRef(env, charToByteConverterName);

            /*
             * set load & XFontStruct
             */
            fdata->flist[i].load = 0;

            if (fdata->xfont == NULL &&
                strstr(fdata->flist[i].charset_name, "8859_1")) {
                if (fdata->flist[i].xfont =
                    loadFont(awt_display,
                             fdata->flist[i].xlfd, size * 10)) {
                    fdata->flist[i].load = 1;

                    fdata->xfont = fdata->flist[i].xfont;
                    fdata->flist[i].index_length = 1;
                } else {
                    if (errmsg) {
                        *errmsg = "java/lang" "NullPointerException";
                    }
		    (*env)->DeleteLocalRef(env, componentFonts);
                    return NULL;
                }
            }
        }
	(*env)->DeleteLocalRef(env, componentFonts);
        /*
         * XFontSet will create if the peer of TextField/TextArea
         * are used.
         */
        fdata->xfs = NULL;

        (*env)->SetLongField(env,font,fontIDs.pData,(jlong) fdata);


        return fdata;
    } else {
        Display *display;
        struct FontData *fdata;
        char fontSpec[1024];
        long height;
        long oheight;
        int above = 0;              /* tries above height */
        int below = 0;              /* tries below height */
        char *foundry;
        char *name;
        char *encoding;
        char *style;
        XFontStruct *xfont;
        jstring family;

        if (JNU_IsNull(env, font)) {
            if (errmsg) {
                *errmsg = "java/lang" "NullPointerException";
            }
            return (struct FontData *) NULL;
        }
        display = XDISPLAY;

        fdata = (struct FontData *) (*env)->GetLongField(env,font,fontIDs.pData);
        if (fdata != NULL && fdata->xfont != NULL) {
            return fdata;
        }

        family = (*env)->CallObjectMethod(env, font, fontIDs.getFamily);

        if (!awtJNI_FontName(env, family, &foundry, &name, &encoding)) {
            if (errmsg) {
                *errmsg = "java/lang" "NullPointerException";
            }
	    (*env)->DeleteLocalRef(env, family);
            return (struct FontData *) NULL;
        }
        style = Style((*env)->GetIntField(env, font, fontIDs.style));
        oheight = height = (*env)->GetIntField(env, font, fontIDs.size);

        while (1) {
            jio_snprintf(fontSpec, sizeof(fontSpec), "-%s-%s-%s-*-*-%d-*-*-*-*-*-%s",
                         foundry,
                         name,
                         style,
                         height,
                         encoding);

            /*fprintf(stderr,"LoadFont: %s\n", fontSpec); */
            xfont = XLoadQueryFont(display, fontSpec);

            /* XXX: sometimes XLoadQueryFont returns a bogus font structure */
            /* with negative ascent. */
            if (xfont == (Font) NULL || xfont->ascent < 0) {
                if (xfont) {
                    XFreeFont(display, xfont);
                }
                if (foundry != anyfoundry) {
                    /* Try any other foundry before messing with the sizes */
                    foundry = anyfoundry;
                    continue;
                }
                /* We couldn't find the font. We'll try to find an */
                /* alternate by searching for heights above and below our */
                /* preferred height. We try for 4 heights above and below. */
                /* If we still can't find a font we repeat the algorithm */
                /* using misc-fixed as the font. If we then fail, then we */
                /* give up and signal an error. */
                if (above == below) {
                    above++;
                    height = oheight + above;
                } else {
                    below++;
                    if (below > 4) {
                        if (name != defaultfontname || style != anystyle) {
                            name = defaultfontname;
                            foundry = defaultfoundry;
                            height = oheight;
                            style = anystyle;
                            encoding = isolatin1;
                            above = below = 0;
                            continue;
                        } else {
                            if (errmsg) {
                                *errmsg = "java/io/" "FileNotFoundException";
                            }
			    (*env)->DeleteLocalRef(env, family);
                            return (struct FontData *) NULL;
                        }
                    }
                    height = oheight - below;
                }
                continue;
            } else {
                fdata = ZALLOC(FontData);

                if (fdata == NULL) {
                    if (errmsg) {
                        *errmsg = "java/lang" "OutOfMemoryError";
                    }
                } else {
                    fdata->xfont = xfont;
                    (*env)->SetLongField(env,font,fontIDs.pData,(jlong) fdata);
                }
		(*env)->DeleteLocalRef(env, family);
                return fdata;
            }
        }
	/* not reached */
    }
}

/*
 * Class:     sun_awt_motif_X11FontMetrics
 * Method:    getMFCharsWidth
 * Signature: ([CIILjava/awt/Font;)I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_X11FontMetrics_getMFCharsWidth
  (JNIEnv *env, jobject this, jcharArray data, jint offset, jint length, jobject font)
{
    jint retVal;

    AWT_LOCK();

    retVal = awtJNI_GetMFStringWidth(env, data, offset, length, font);

    AWT_UNLOCK();

    return retVal;
}

/*
 * Class:     sun_awt_motif_X11FontMetrics
 * Method:    bytesWidth
 * Signature: ([BII)I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_X11FontMetrics_bytesWidth
  (JNIEnv *env, jobject this, jbyteArray str, jint off, jint len)
{
    unsigned char *s, *tmpPointer;
    jint w = 0;
    int ch;
    int cnt;
    jobject widths;
    jint tempWidths[256];
    jint maxAdvance;
    int widlen;

    if (JNU_IsNull(env, str)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");

        return 0;
    }
    cnt = (*env)->GetArrayLength(env, str);

    if (cnt == 0) {
        return 0;
    }

    widths = (*env)->GetObjectField(env, this, x11FontMetricsIDs.widths);
    maxAdvance = (*env)->GetIntField(env, this, x11FontMetricsIDs.maxAdvance);
    if (!JNU_IsNull(env, widths)) {
        w = 0;
        widlen = (*env)->GetArrayLength(env, widths);
        (*env)->GetIntArrayRegion(env, widths, 0, widlen, (jint *) tempWidths);

        s = tmpPointer = (unsigned char *) (*env)->GetPrimitiveArrayCritical(env, str, NULL);

        if(s == NULL)
            return 0;

        while (--cnt >= 0) {
            ch = *tmpPointer++;
            if (ch < widlen) {
                w += tempWidths[ch];
            } else {
                w += maxAdvance;
            }
        }

        (*env)->ReleasePrimitiveArrayCritical(env, str, (jchar *) s, JNI_ABORT);
    } else {
        w = maxAdvance * cnt;
    }

    return w;
}

/*
 * Class:     sun_awt_motif_X11FontMetrics
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11FontMetrics_init
  (JNIEnv *env, jobject this)
{
    jobject font;
    struct FontData *fdata;
    jint tempWidths[256];
    jintArray widths;
    int ccount;
    int i;
    int tempWidthsIndex;
    char *err;

    if (JNU_IsNull(env, this)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();

    font = (*env)->GetObjectField(env, this, x11FontMetricsIDs.font);
    if (JNU_IsNull(env, this)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    fdata = awtJNI_GetFontData(env, font, &err);
    if (fdata == NULL) {
        JNU_ThrowInternalError(env, err);
        AWT_UNLOCK();
        return;
    }
    (*env)->SetIntField(env, this, x11FontMetricsIDs.ascent,
			(jint) fdata->xfont->ascent);
    (*env)->SetIntField(env, this, x11FontMetricsIDs.descent,
			(jint) fdata->xfont->descent);
    (*env)->SetIntField(env, this, x11FontMetricsIDs.leading, (jint) 1);
    (*env)->SetIntField(env, this, x11FontMetricsIDs.height,
			(jint) fdata->xfont->ascent + fdata->xfont->descent + 1);
    (*env)->SetIntField(env, this, x11FontMetricsIDs.maxAscent,
			(jint) fdata->xfont->max_bounds.ascent);
    (*env)->SetIntField(env, this, x11FontMetricsIDs.maxDescent,
			(jint) fdata->xfont->max_bounds.descent);
    (*env)->SetIntField(env, this, x11FontMetricsIDs.maxHeight,
			(jint) fdata->xfont->max_bounds.ascent
			+ fdata->xfont->max_bounds.descent + 1);
    (*env)->SetIntField(env, this, x11FontMetricsIDs.maxAdvance,
			(jint) fdata->xfont->max_bounds.width);


    widths = (*env)->NewIntArray(env, 256);
    (*env)->SetObjectField(env, this, x11FontMetricsIDs.widths, widths);
    if (JNU_IsNull(env, widths)) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    /*
     * We could pin the array and then release it, but I believe this method
     * is faster and perturbs the VM less
     *
     */
    memset(tempWidths, 0, 256 * sizeof(jint));

    tempWidthsIndex = fdata->xfont->min_char_or_byte2;

    ccount = fdata->xfont->max_char_or_byte2 - fdata->xfont->min_char_or_byte2;

    if (fdata->xfont->per_char) {
        for (i = 0; i <= ccount; i++) {
            tempWidths[tempWidthsIndex++] = (jint) fdata->xfont->per_char[i].width;
        }
    } else {
        for (i = 0; i <= ccount; i++) {
            tempWidths[tempWidthsIndex++] = (jint) fdata->xfont->max_bounds.width;
        }
    }

    (*env)->SetIntArrayRegion(env, widths, 0, 256, (jint *) tempWidths);

    AWT_UNLOCK();
}

/*
 * Class:     java_awt_Font
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_java_awt_Font_pDispose
  (JNIEnv *env, jobject this)
{
    struct FontData *fdata;
    int i;
    Display *display;

    AWT_LOCK();
    fdata = (struct FontData *) (*env)->GetLongField(env,this,fontIDs.pData);

    if (fdata == NULL) {
        AWT_UNLOCK();
        return;
    }

    if( awtJNI_IsMultiFont(env, this)) {

        for (i = 0; i < fdata->charset_num; i++) {

            /*fdata->flist[i].xlfd = malloc(strlen(nativename)
                                          + strlen(defaultXLFD));*/
            free((void *)fdata->flist[i].xlfd);

            /* fdata->flist[i].charset_name = (char *)
                JNU_GetStringPlatformChars(env, charToByteConverterName, NULL); */
            JNU_ReleaseStringPlatformChars(env,NULL, fdata->flist[i].charset_name);
        }

        /*fdata->flist = (awtFontList *) malloc(sizeof(awtFontList)
                                              * fdata->charset_num);*/
        free((void *)fdata->flist);
    }

    display = XDISPLAY;
    XFreeFont(display, fdata->xfont);
    /* fdata = (struct FontData *) malloc(sizeof(struct FontData)) */;
    free((void *)fdata);

    (*env)->SetLongField(env,this,fontIDs.pData,(jlong) NULL);	
    AWT_UNLOCK();

    return;
}

int
compare (const void *f1, const void *f2)
{
    char **s1 = (char **) (f1);
    char **s2 = (char **) (f2);
    char *p1 = *s1;
    char *p2 = *s2;

    while (*p1 && *p2 && *p1 == *p2)
        p1++, p2++;
    return (*p1 - *p2);
}
 
#define USE_NATIVE_RASTERIZER
/* #undef USE_NATIVE_RASTERIZER  */

JNIEXPORT jstring JNICALL Java_sun_awt_X11GraphicsEnvironment_getNativeFonts 
	(JNIEnv *env, jclass obj) 
{
    int nFonts = 1;
    int i, retFonts;
    char *pattern = "-*-*-*-*-*-*-0-0-*-*-*-*-*-*";
    char *previous;
    char **fonts;
    char **tmpFonts;
    char **tmpFonts2;
    jclass stringClass;
    jobjectArray fontArray;
    int sFonts = 0;

#ifdef USE_NATIVE_RASTERIZER
    fonts = XListFonts (awt_display, pattern, 5000, &retFonts);
    if (retFonts == 0)
	return NULL;

    tmpFonts = (char **) calloc (1, sizeof (char *) * retFonts);

    if (tmpFonts == 0) {
        XFreeFontNames (fonts);
        return NULL;
    }
    
    tmpFonts2 = (char **) calloc (1, sizeof (char *) * retFonts);

    if (tmpFonts2 == 0) {
        free (tmpFonts);
        XFreeFontNames (fonts);
        return NULL;
    }
    
    /* Get rid of everything but scalable fonts. Scalables
     * will have the string "-0-0-0-0" in them.
     */

    for (i = 0; i < retFonts; i++) {
	if (strstr (fonts [i], "-0-0-0-0-") != 0)
            tmpFonts [sFonts++] = fonts [i];
    }

    if (sFonts == 0) {
        free (tmpFonts);
        free (tmpFonts2);
        XFreeFontNames (fonts);
        return NULL;
    }

    qsort (tmpFonts, sFonts, sizeof (char *), compare);

    /* Get rid of duplicates and Speedo fonts (we can't get outlines of Speedos) */
    previous = tmpFonts [0];
    tmpFonts2 [0] = tmpFonts [0];
    for (i = 0; i < sFonts; i++) {
	if ((strcmp (previous, tmpFonts [i]) != 0) &&
	    (strstr (tmpFonts [i], "bitstream-charter") == 0) &&
	    (strstr (tmpFonts [i], "bitstream-courier") == 0)) {
	    tmpFonts2 [nFonts++] = tmpFonts [i];
	    previous = tmpFonts [i];
	}
    }

    stringClass = (*env)->FindClass (env, "java/lang/String");
    fontArray = (*env)->NewObjectArray (env, nFonts, stringClass, NULL);

    for (i = 0; i < nFonts; i++)  
	(*env)->SetObjectArrayElement (env, fontArray, i,
    				(*env)->NewStringUTF (env, tmpFonts2 [i]));

    free (tmpFonts);
    free (tmpFonts2);
    XFreeFontNames (fonts);

    return fontArray;
#else
    return NULL;
#endif
}
