/*
 * @(#)awt_Frame.c	1.132 98/09/23
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "color.h"
#include <X11/Shell.h>
#include <Xm/VendorS.h>
#include <Xm/Form.h>
#include <Xm/DialogS.h>
#include <Xm/AtomMgr.h>
#include <Xm/Protocols.h>
#include <Xm/MwmUtil.h>
#include <X11/IntrinsicP.h>
#include "java_awt_MenuBar.h"
#include "sun_awt_motif_MMenuBarPeer.h"
#include "java_awt_Frame.h"
#include "sun_awt_motif_MFramePeer.h"
#include "java_awt_Color.h"
#include "java_awt_Component.h"
#include "java_awt_Insets.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "java_awt_Image.h"
#include "java_awt_Font.h"
#include "canvas.h"
#include "awt_util.h"
#include "img_util.h"
#include "sun_awt_EmbeddedFrame.h"
#include "sun_awt_motif_MEmbeddedFrame.h"
#include "sun_awt_motif_MEmbeddedFramePeer.h"

#include "awt_Component.h"
#include "awt_Insets.h"
#include "awt_MenuBar.h"
#include "awt_Window.h"

#include <jni.h>
#include <jni_util.h>

static void reshape();
extern struct InsetsIDs insetsIDs;
extern struct ComponentIDs componentIDs;
extern struct MMenuBarPeerIDs mMenuBarPeerIDs;
extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct WindowIDs windowIDs;
 
static int topGuess = 0;
static int sideGuess = 0;
 
void awt_Frame_guessInsets( struct FrameData  *wdata )
{
        char *inset_ptr;
        if ( topGuess == 0 )
        {
                if ( wdata->top >= 0 )
                {       /* they have been set by System Properties */
                        topGuess = wdata->top;
                        sideGuess = wdata->left;
                }
                else switch ( awt_util_runningWindowManager() )
                {
                        case CDE_WM:
                                topGuess=28;
                                sideGuess=6;
                                break;

 			case NO_WM:
 				topGuess = 0;
 				sideGuess = 0;
 				break;

                        case MOTIF_WM:
                        case OPENLOOK_WM:
                        default:
                                topGuess=25;
                                sideGuess=5;
                                break;
                }
                if (inset_ptr = getenv("AWT_INSETS") )
                {
                        topGuess = atol(inset_ptr);
                        sideGuess = topGuess & 0xff;
                        topGuess = (topGuess & 0xff00) >> 8;
                }
                /* don't allow bizarly large insets */
                if ( ( topGuess > 64 ) || ( topGuess < 0 ) )
                        topGuess = 28;
                if ( ( sideGuess > 32 ) || ( sideGuess < 0 ) )
                        sideGuess = 6;
        }
        wdata->top = topGuess;
        wdata->left = sideGuess;
        wdata->bottom = sideGuess;
        wdata->right = sideGuess;
}

int getTopGuess()
{
	return (topGuess);
}

int getSideGuess()
{
	return (sideGuess);
}

/*
 * Since query XmNiconic will not give reliable state, we have to use
 * the following alternative function to return one of
 * IconicState | NormalState | WithdrawnState | NULL.
 *
 * Note: memory free task of pState is left to user.
 */
#define WM_STATE_ELEMENTS 1
static unsigned long *getWindowState(Window win)
{
    unsigned long *pState = NULL;
    unsigned long nitems, leftover;
    Atom xa_WM_STATE, actual_type;
    int actual_format, status;
    
    xa_WM_STATE = XInternAtom (awt_display, "WM_STATE", False);
    status = XGetWindowProperty (awt_display, win,
				 xa_WM_STATE, 0L, WM_STATE_ELEMENTS,
				 False, xa_WM_STATE, &actual_type, &actual_format,
				 &nitems, &leftover, (unsigned char **)&pState);
    
    if (!((status == Success) &&
	  (actual_type == xa_WM_STATE) &&
	  (nitems == WM_STATE_ELEMENTS))) {
	if (pState) {
	    XFree ((char *)pState);
	    pState = NULL;
	}
    }
    return (pState);
} 


/* fieldIDs for Frame fields that may be accessed from C */
static struct FrameIDs {
    jfieldID mbManagement;
    jfieldID resizable;
    jfieldID state;
} frameIDs;

/*
 * Class:     java_awt_Frame
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for Frame.java
   to initialize the fieldIDs for fields that may be accessed from C */
JNIEXPORT void JNICALL
Java_java_awt_Frame_initIDs
  (JNIEnv *env, jclass cls)
{
    frameIDs.mbManagement = (*env)->GetFieldID(env, cls, "mbManagement", "Z");
    frameIDs.resizable = (*env)->GetFieldID(env, cls, "resizable", "Z");
    frameIDs.state = (*env)->GetFieldID(env, cls, "state", "I");
}

/* fieldIDs for MFramePeer fields that may be accessed from C */
static struct MFramePeerIDs {
    jfieldID insets;
    jfieldID iconWidth;
    jfieldID iconHeight;
} mFramePeerIDs;


/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for MFramePeer.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_initIDs
  (JNIEnv *env, jclass clazz)
{
    mFramePeerIDs.insets =
	(*env)->GetFieldID(env, clazz, "insets", "Ljava/awt/Insets;");
    mFramePeerIDs.iconWidth =
	(*env)->GetFieldID(env, clazz, "iconWidth", "I");
    mFramePeerIDs.iconHeight =
	(*env)->GetFieldID(env, clazz, "iconHeight", "I");
}

/* fieldIDs for MEmbeddedFramePeer fields that may be accessed from C */
static struct MEmbeddedFramePeerIDs {
    jfieldID handle;
} mEmbeddedFramePeerIDs;

/*
 * Class:     java_awt_MEmbeddedFramePeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MEmbeddedFramePeer.java
   to initialize the fieldIDs for fields that may be accessed from C */
JNIEXPORT void JNICALL Java_sun_awt_motif_MEmbeddedFramePeer_initIDs
  (JNIEnv *env, jclass cls)
{
    mEmbeddedFramePeerIDs.handle = (*env)->GetFieldID(env, cls, "handle", "J");
}

/* changeInsets() sets target's insets equal to X/Motif values. */

static void
awtJNI_ChangeInsets(JNIEnv * env, jobject this, struct FrameData *wdata)
{
    jobject insets;
    
    if ((*env)->EnsureLocalCapacity(env, 1) < 0)
        return;
    
    insets = (*env)->GetObjectField(env, this, mFramePeerIDs.insets);
    
    if (JNU_IsNull(env, insets)) {
        return;
    }
    
    (*env)->SetIntField(env, insets, insetsIDs.top, wdata->top);
    (*env)->SetIntField(env, insets, insetsIDs.left, wdata->left);
    (*env)->SetIntField(env, insets, insetsIDs.bottom, wdata->bottom);
    (*env)->SetIntField(env, insets, insetsIDs.right, wdata->right);
    
    /* when the insets get set, make sure we set the proper */
    /* max window size (since it's dependent on inset size) */
    if (wdata->isResizable) {
        int screenWidth = XWidthOfScreen( XDefaultScreenOfDisplay(awt_display));        int screenHeight= XHeightOfScreen(XDefaultScreenOfDisplay(awt_display));
        XtVaSetValues(wdata->winData.shell,
                XmNmaxWidth, screenWidth - (wdata->left + wdata->right),
                XmNmaxHeight, screenHeight - (wdata->top + wdata->bottom),
                NULL);
    }
    
    (*env)->DeleteLocalRef(env, insets);
}


/* setMbAndWwHeightAndOffsets() attempts to establish the heights
   of frame's menu bar and warning window (if present in frame).
   setMbAndWwHeightAndOffsets() also adjusts appropriately the
   X/Motif offsets and calls changeInsets() to set target insets.
   A warning window, if present, is established during ...create().
   wdata->warningWindow is set there, wdata->wwHeight is set here.
   Routine pSetMenuBar() sets value of the wdata->menuBar field.
   This routine reads that value. If it is not null, a menubar
   has been added.  In this case, calculate the current height
   of the menu bar.  This may be a partial (incomplete) menubar
   because ths routine may be called before the X/Motif menubar
   is completely realized. In this case, the menubar height may
   be adjusted incrementally.  This routine may be called from
   ...pSetMenuBar(), innerCanvasEH(), and ...pReshape(). It is
   designed to (eventually) obtain the correct menubar height.
   On the other hand, if wdata->menuBar is NULL and the stored
   menubar height is not zero, then we subtract off the height. */

static void
awtJNI_setMbAndWwHeightAndOffsets(JNIEnv * env,
				  jobject this,
				  struct FrameData *wdata )
{
    Dimension   warningHeight,	/* Motif warning window height  */
		labelHeight;	/* Motif warning label's height	*/

    WidgetList	warningChildrenWL; /* warning children widgets	*/

    Dimension   menuBarWidth,   /* Motif menubar width          */
                menuBarHeight,  /* Motif menubar height         */
                menuBarBorderSize, /* Motif menubar border size */
                marginHeight,   /* Motif menubar margin height  */
                menuHeight,     /* Motif menubar's menu height  */
                menuBorderSize, /* Motif menu border size       */
                actualHeight;   /* height: menu+margins+borders */

    WidgetList  menuBarChildrenWL; /* menubar children widgets  */
    Cardinal    numberChildren; /* number of menubar children   */
    
#ifdef _pauly_debug
    fprintf(stdout," ++ setMenuBar\n");
    fflush(stdout);
#endif /* _pauly_debug */

    /* If warning window height not yet known, try to get it now.
       It will be added to top or bottom (iff NETSCAPE) offset. */
    if  (wdata->warningWindow != NULL) {
        XtVaGetValues(wdata->warningWindow,
		      XmNheight, &warningHeight,
                      XmNchildren, &warningChildrenWL,
                      XmNnumChildren, &numberChildren,
                      NULL);

        /* We may be doing this before warning window is realized ! So,
           check for a child label in the warning. If its height is not
           yet accounted for in the warning height, then use it here.	*/
        if  (numberChildren != 0) {
            XtVaGetValues(warningChildrenWL[0],
                          XmNheight, &labelHeight,
                          NULL);
#ifdef _pauly_debug
            fprintf(stdout,"    setMenuBar.... warning label found with height: %d\n", labelHeight);
            fflush(stdout);
#endif /* _pauly_debug */
            if  (warningHeight < labelHeight) {
#ifdef _pauly_debug
    fprintf(stdout,"    setMenuBar.... !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
    fflush(stdout);
#endif /* _pauly_debug */
                warningHeight = labelHeight;
            }
        }

        if  (wdata->wwHeight < warningHeight) {
#ifdef _pauly_debug
            fprintf(stdout, "    setMenuBar.... adding warning height: %d\n", warningHeight);
            fflush(stdout);
#endif /* _pauly_debug */
#ifdef NETSCAPE
            wdata->bottom += (warningHeight - wdata->wwHeight);
#else
            wdata->top += (warningHeight - wdata->wwHeight);
#endif /* NETSCAPE */
            awtJNI_ChangeInsets(env, this, wdata);
            wdata->wwHeight = warningHeight;
        }
    }

    /* Now we adjust offsets for an added or removed menu bar   */
    if  (wdata->menuBar != NULL) {
#ifdef _pauly_debug
        fprintf(stdout,"    setMenuBar.  menu bar: %x\n", wdata->menuBar);
        fflush(stdout);
#endif /* _pauly_debug */
        XtVaGetValues(wdata->menuBar,
                      XmNwidth, &menuBarWidth,
                      XmNheight, &menuBarHeight,
                      XmNchildren, &menuBarChildrenWL,
                      XmNnumChildren, &numberChildren,
                      XmNborderWidth, &menuBarBorderSize,
                      XmNmarginHeight, &marginHeight,
                      NULL);

        /* We may be doing this before menu bar is realized ! Hence,
           check for a menu in the menu bar. If its height is not yet
           accounted for in the menu bar height, then add it in here.   */
        if  (numberChildren != 0) {
            XtVaGetValues(menuBarChildrenWL[0],
                          XmNheight, &menuHeight,
                          XmNborderWidth, &menuBorderSize,
                          NULL);
#ifdef _pauly_debug
            fprintf(stdout,"    setMenuBar.... menu found with height: %d, border: %d, margin: %d, bar border: %d\n", menuHeight, menuBorderSize, marginHeight, menuBarBorderSize);
            fflush(stdout);
#endif /* _pauly_debug */
            /* Calculate real height of menu bar by adding height of its
               child menu and borders, margins, and the menu bar borders*/
            actualHeight = menuHeight + (2 * 1) +
                           (2 * marginHeight) + (2 * 1);
            if  (menuBarHeight < actualHeight) {
#ifdef _pauly_debug
fprintf(stdout,"    setMenuBar.... ****************************************\n");
fflush(stdout);
#endif /* _pauly_debug */
                menuBarHeight = actualHeight;
            }
        }

        if  (wdata->mbHeight < menuBarHeight) {
            /* Adjust the (partially) added menu bar height, top offset.*/
#ifdef _pauly_debug
            fprintf(stdout, "    setMenuBar.... added menuBar height: %d\n", menuBarHeight);
            fflush(stdout);
#endif /* _pauly_debug */
            wdata->top += (menuBarHeight - wdata->mbHeight);
            awtJNI_ChangeInsets(env, this, wdata);
            wdata->mbHeight = menuBarHeight;
        }
    } else if  ((wdata->menuBar == NULL) && (wdata->mbHeight > 0)) {
        /* A menu bar has been removed; subtract height from top offset.*/
        wdata->top -= wdata->mbHeight;
#ifdef _pauly_debug
        fprintf(stdout, "    setMenuBar.... removed menuBar height: %d\n", wdata->mbHeight);
        fflush(stdout);
#endif /* _pauly_debug */
        awtJNI_ChangeInsets(env, this, wdata);
        wdata->mbHeight = 0;
    }
}


/* outerCanvasResizeCB() is Motif resize callback for outer/child canvas.
   It reads width, height of Motif widget, sets java target accordingly,
   and then does a JAVA_UPCALL of handleResize() to effect any changes.
   This upcall is only done for a shell resize or inner/parent resize;
   i.e., it may not be done for a ...pReshape() to avoid doing a loop.  

   client_data is MFramePeer instance
*/
static void
outerCanvasResizeCB(Widget wd, XtPointer client_data, XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject target;
    struct FrameData *wdata;
    Position    screenX;        /* x position of the canvas, screen */
    Position    screenY;        /* y position of the canvas, screen */
    Dimension   width;          /* width of the canvas, target  */
    Dimension   height;         /* height of the canvas, target */

#ifdef _pauly_debug
    fprintf(stdout," ++ WindowResize.\n");
    fflush(stdout);
#endif /* _pauly_debug */

    wdata = (struct FrameData *)
	    (*env)->GetLongField(env, (jobject) client_data,
				 mComponentPeerIDs.pData);
    if (wdata == 0) {
        return;
    }

    if ((*env)->EnsureLocalCapacity(env, 1) < 0)
        return;

    target = (*env)->GetObjectField(env, (jobject) client_data,
				    mComponentPeerIDs.target);
    XtVaGetValues(wd,
		  XmNwidth, &width,
		  XmNheight, &height,
		  NULL);
#ifdef _pauly_debug
    fprintf(stdout,"    outerCanvasResizeCB.  width: %d, height: %d\n", width, height);
    fflush(stdout);
#endif /* _pauly_debug */


    XtTranslateCoords(wd, 0, 0, &screenX, &screenY);

    if  ((wdata->shellResized) || (wdata->canvasResized)) {
#ifdef _pauly_debug
        fprintf(stdout,"    outerCanvasResizeCB. JAVA_UPCALL\n");
        fflush(stdout);
#endif /* _pauly_debug */
	wdata->shellResized = False;
	wdata->canvasResized = False;
        /*
        ** if you are not yet reparented, don't compute the size based on the
        ** widgets, as the window manager shell containg the insets is not yet
        ** there.  Use the size the application has set.
        ** If not reparented, we got here because the application set the size,
        ** so just send them Component.RESIZED event with the size they set.
        **
        ** If the reparenting causes a resize ( only when inset guess is wrong )        ** the new size will be sent in a Component.RESIZED event at that time.
        */
        if (wdata->reparented)
        {
            (*env)->SetIntField(env, target, componentIDs.x, (jint) screenX);
            (*env)->SetIntField(env, target, componentIDs.y, (jint) screenY);
        }
        (*env)->SetIntField(env, target, componentIDs.width, (jint) width);
        (*env)->SetIntField(env, target, componentIDs.height, (jint) height);

	JNU_CallMethodByName(env, NULL, (jobject) client_data
			     ,"handleResize"
			     ,"(II)V"
			     ,width, height);
	if  ((*env)->ExceptionOccurred(env)) {
	    (*env)->ExceptionDescribe(env);
	    (*env)->ExceptionClear(env);
	}
    }

    (*env)->DeleteLocalRef(env, target);

#ifdef _pauly_debug
    fprintf(stdout,"    WindowResize. Done.\n");
    fflush(stdout);
#endif /* _pauly_debug */

} /* outerCanvasResizeCB() */

static void reconfigureOuterCanvas ( JNIEnv *env, jobject target,
                                     jobject this, struct FrameData *wdata )
{
    Dimension   innerDAWidth,   /* width of inner Motif canvas  */
                innerDAHeight,  /* height of inner Motif canvas */
                outerDAWidth,   /* width of outer Motif canvas  */
                outerDAHeight;  /* height of outer Motif canvas */
    int         targetWidth,    /* java target object's width   */
                targetHeight;   /* java target's object height  */
    Dimension   width;          /* width of the canvas, target  */
    Dimension   height;         /* height of the canvas, target */
 
 
    Position    innerX,         /* x loc. of inner Motif canvas */
                innerY,         /* y loc. of inner Motif canvas */
                x, y;
 
    /* canvasW is (visible) inner/parent drawing area (canvas) widget   */
    XtVaGetValues(XtParent(wdata->winData.comp.widget),
                  XmNwidth, &innerDAWidth,
                  XmNheight, &innerDAHeight,
                  XmNx, &innerX,
                  XmNy, &innerY,
                  NULL);
 
    /* This resize may be due to the insertion or removal of a menu bar.
       If so, we appropriately adjust the top offset in wdata, insets.  */
    awtJNI_setMbAndWwHeightAndOffsets(env, this, wdata);
 
    outerDAWidth = innerDAWidth + wdata->left + wdata->right;
    outerDAHeight = innerDAHeight + wdata->top + wdata->bottom;
 
    /* If it's a menu bar reset, do not do resize of outer/child canvas.
       (Another thread problem; we arrest this now before damage done.) */
    if  (wdata->menuBarReset)
    {
	targetWidth = (*env)->GetIntField(env, target, componentIDs.width);
	targetHeight = (*env)->GetIntField(env, target, componentIDs.height);
        if  ((outerDAWidth != targetWidth) || (outerDAHeight != targetHeight))
        {
            return;
        }
    }
 
    wdata->canvasResized = True;
 
    /* The outer/child drawing area (canvas) needs to be configured too.
       If its size changes, its resize callback will thereby be invoked.*/
    x = -wdata->left;
    y = -wdata->top;
    width = innerDAWidth + wdata->left + wdata->right;
    height = innerDAHeight + wdata->top + wdata->bottom;
 
    XtConfigureWidget(wdata->winData.comp.widget, x, y, width, height, 0 );
}



/* innerCanvasEH() is event handler for inner/parent canvas. It handles
   map and configure notify events. It reads width and height, adjusts
   for menubar insertion / removal and configures outer/child canvas.   */

static void
innerCanvasEH(Widget canvasW, caddr_t client_data, XEvent *event)
{
    JNIEnv	*env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject	this = (jobject) client_data;
    jobject	target;
    struct FrameData *wdata;

    Boolean	mbManagement;	/* menubar is being (un)managed	*/

    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if  (wdata == 0) {
	return;
    }

    if  ((*env)->EnsureLocalCapacity(env, 1) < 0)
	return;

    target = (*env)->GetObjectField(env, (jobject) client_data,
				    mComponentPeerIDs.target);

    /* While inside ...pSetMenuBar(), don't react to incomplete resizing
       events supplied by Xt toolkit. Wait for completion of the routine. */

    mbManagement = (Boolean)
	(*env)->GetBooleanField(env, target, frameIDs.mbManagement);
    if  (mbManagement) {
#ifdef _pauly_debug
	fprintf(stdout,">>>>Do not do inner canvas eh because in mbManagement\n");
	fflush(stdout);
#endif /* _pauly_debug */
	(*env)->DeleteLocalRef(env, target);
	return;
    }

    /* For a map or resize, we need to check for the addition or deletion
       of a menu bar to the form which is the of this drawing area (canvas).
       We also must then configure the outer/child canvas appropriately.  */

    if  ( (event->xany.type == MapNotify) ||
	  (event->xany.type == ConfigureNotify) )
    {
	reconfigureOuterCanvas( env, target, this, wdata );
    }

    (*env)->DeleteLocalRef(env, target);

}


/* shellEH() is event handler for the Motif shell widget. It handles
   focus change, map notify, configure notify events for the shell.
   Please see internal comments pertaining to these specific events.    

   data is MFramePeer instance pointer
*/

static void
shellEH(Widget w, XtPointer data, XEvent * event)
{
    JNIEnv	*env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject	this = (jobject) data;
    jobject	target;
    struct FrameData *wdata;
    int		setTargetX,
		setTargetY,
		getTargetX,
		getTargetY;
    
    long        wwHeight;       /* height of any warning window present */
    long        topAdjust;      /* adjust top offset for menu, warning  */
    jboolean    isPacked; 

#ifndef NOMODALFIX
    extern Boolean awt_isModal();
    extern Boolean awt_isWidgetModal(Widget w);
#endif NOMODALFIX

    /* Any event handlers which take peer instance pointers as
     * client_data should check to ensure the widget has not been
     * marked as destroyed as a result of a dispose() call on the peer
     * (which can result in the peer instance pointer already haven
     * been gc'd by the time this event is processed)
     */
    if (w->core.being_destroyed) {
        return;
    }
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0) {
        return;
    }
    switch (event->xany.type) {
    case FocusOut:
	if (event->xfocus.mode == NotifyNormal) {
#ifndef NOMODALFIX
	    if (!awt_isModal() || awt_isWidgetModal(w)) {
#endif NOMODALFIX
	       awt_setDeactivatedShell(w);
	       JNU_CallMethodByName(env, NULL, this, "handleDeactivate", "()V");
	       if ((*env)->ExceptionOccurred(env)) {
		   (*env)->ExceptionDescribe(env);
		   (*env)->ExceptionClear(env);
	       }
#ifndef NOMODALFIX
	   }
#endif NOMODALFIX
	}
	break;
	
    case FocusIn:
	if (event->xfocus.mode == NotifyNormal) {
#ifndef NOMODALFIX
	    if (!awt_isModal() || awt_isWidgetModal(w)) {
#endif NOMODALFIX
		awt_setActivatedShell(w);
		JNU_CallMethodByName(env, NULL, this, "handleActivate", "()V");
		if ((*env)->ExceptionOccurred(env)) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);
		}
#ifndef NOMODALFIX
	    }
#endif NOMODALFIX
	}
	break;
	
    case MapNotify:

#ifdef _pauly_debug
jio_fprintf(stderr , "MapNotify--------------------\n");
#endif
	if (wdata->mappedOnce < 1) {
	    wdata->mappedOnce++;
	    
	    /*wdata->mappedOnce = 1;*/
	    JNU_CallMethodByName(env, NULL, this, "makeCursorsVisible", "()V");
	    if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }
	    
	    if (wdata->isIconic) {
		XIconifyWindow(awt_display, XtWindow(wdata->winData.shell), XDefaultScreen(awt_display));
	    }
	    
	} else {
	    unsigned long *pState = getWindowState(XtWindow(wdata->winData.shell));

	    if (pState != NULL && wdata->isIconic && *pState == NormalState) {
		wdata->isIconic = False;
		JNU_CallMethodByName(env, NULL, this, "handleDeiconify", "()V");
		if ((*env)->ExceptionOccurred(env)) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);
		}
		XFree((char *)pState);	
	    }
	}
	break;
	
    case UnmapNotify:
	
	if (wdata->mappedOnce < 1) {   
	    wdata->mappedOnce++;
	    
	    /*wdata->mappedOnce = 1;*/
	    JNU_CallMethodByName(env, NULL, this, "makeCursorsVisible", "()V");
	    if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }
	} else {
	    unsigned long *pState = getWindowState(XtWindow(wdata->winData.shell));

	    if (pState != NULL && wdata->isShowing && !wdata->isIconic && *pState == IconicState) {
		wdata->isIconic = True;
		JNU_CallMethodByName(env, NULL, this, "handleIconify", "()V");
		if ((*env)->ExceptionOccurred(env)) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);
		} 
		XFree((char *)pState);	
	    }
	}
	break;
	
    case ReparentNotify:
	if  ((*env)->EnsureLocalCapacity(env, 1) < 0)
	    break;
	
	target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
	
	/* The insets were literally hardcoded in the MFramePeer.
	   But they are dependent upon both the window manager (WM)
	   and the hardware display.  So, these are usually wrong.
	   This leads to problems with shell positioning and size.
	   Furthermore, there is not a published interface or way
	   to obtain from any given window manager the dimensions
	   of its decoration windows (i.e., borders and title bar).
	   So, given this problem in design, we must workaround.
	   N.B. (0) This works.  But there is one functional caveat:
	   the frame.insets() function will usually return
	   the wrong values until AFTER the frame is shown.
	   It always did this before; it's just that now,
	   the values will become correct after rendering,
	   whereas before the values were never corrected.
	   (I believe this unavoidable given this design.)
	   (1) Note that we must/have to do this exactly once.
	   (2) The hardcoded values of ...create() (25,5)
	   are also utilized here and must be consistent.
	   This of course could be reworked as desired.
	   (3) Assume top border (title bar) is one width,
	   and other three borders are another width.
	   This, however, could be easily reworked below.	*/

            /*
            ** The above comment is no longer completely true:
            **     The insets are no longer hardcoded but are retrieved
            ** from guessInsets(), either from a per-window manager default,
            ** set in the awt.properties file, or overwritten by the
            ** actual values determined from a previous frames reparenting.
            **
            */

        if ( event->xreparent.parent != awt_root )
        {
	    unsigned		runningWM;
	    XWindowAttributes	winAttr;
	    int			correctWMTop;
	    int			correctWMRest;
	    int			topCorrection;
	    int			restCorrection;
	    int			x, y;
	    int			width, height;
	    int			correctTargetW;
	    int			correctTargetH;
	    int                 screenX, screenY;
	    Window              containerWindow;
	    
	    wdata->reparented = True;
	    correctWMTop = -1;
	    correctWMRest = -1;
	    
            x = (*env)->GetIntField(env, target, componentIDs.x);
            y = (*env)->GetIntField(env, target, componentIDs.y);
            width = (*env)->GetIntField(env, target, componentIDs.width);
            height = (*env)->GetIntField(env, target, componentIDs.height);

	    /* N.B. The inset calculation method used for OTHER_WM
	       could actually be used for both MWM, OLWM as well.	*/
	    
	    /* Get an idea what window manager is running.  This iS
	       definitive for Motif, very likely for OpenLook, and
	       otherwise not possible to determine window manager.  */
	    
	    runningWM = awt_util_runningWindowManager();
	    
	    switch (runningWM) {
		
	    case CDE_WM:
	    case MOTIF_WM:
		/* For Motif WM, the correct insets are given by
		   (x,y) coordinates of event window "parent".  */
		XGetWindowAttributes( XtDisplay(wdata->winData.shell),
				      event->xreparent.parent, &winAttr );
		correctWMRest = winAttr.x;
		correctWMTop = winAttr.y;

		XTranslateCoordinates(awt_display,
	   		event->xreparent.window,
	   		RootWindowOfScreen(XtScreen(wdata->winData.shell)),
	   		0,0, &screenX, &screenY, &containerWindow);
		    
		if ( (screenX != x + sideGuess ) || (screenY != y + topGuess ) )
		{
  		    /* looks like the window manager has placed us somewhere 
  		    ** other than where we asked for, lets respect the window
                    ** and go where he put us, not where we tried to put us
  		    */
  		    x = screenX - correctWMRest;
  		    y = screenY - correctWMTop;
		} 

#ifdef _pauly_debug
		fprintf(stdout,">>>>Motif WM insets: %d, %d\n", correctWMRest, correctWMTop);
		fflush(stdout);
#endif /* _pauly_debug */
		break;
		
	    case OPENLOOK_WM:
		/* For OpenLook WM, correct insets are given by
		   (x,y) coordinates of event window "window".  */
		XGetWindowAttributes( XtDisplay(wdata->winData.shell),
				      event->xreparent.window, &winAttr );
		correctWMRest = winAttr.x;
		correctWMTop = winAttr.y;
#ifdef _pauly_debug
		fprintf(stdout,">>>>OpenLook WM insets: %d, %d\n", correctWMRest, correctWMTop);
		fflush(stdout);
#endif /* _pauly_debug */
		break;
		
	    case OTHER_WM:
	    default:
		if  (XTranslateCoordinates(awt_display,
					   event->xreparent.window,
					   RootWindowOfScreen(XtScreen(wdata->winData.shell)),
					   0,0, &screenX, &screenY, &containerWindow)) {
		    /* Insets given by difference between window's
		       screen coordinates and target coordinates.
		       N.B. THIS "USUALLY" WORKS. BUT NOT ALWAYS.
		       For example, under TWM with shell warping
		       to pointer this fails.  But in any event,
		       for bogus insets, reset to (25,5) below. */
		    
		    getTargetX = (*env)->GetIntField(env, target,
						     componentIDs.x);
		    getTargetY = (*env)->GetIntField(env, target,
						     componentIDs.y);
		    correctWMRest = (int)screenX - getTargetX;
		    correctWMTop = (int)screenY - getTargetY;
		}
                    /* Since preceding code is heuristic
                       rather than deterministic,
                       check that we have reasonable
                       calculations for the insets.
                       If not, just use old (probably
                       incorrect) estimates (25,5). */
                    /* Okay okay but only if we are unsure of the window
                    ** manager, Some of us may be sight impaired with
                    ** some pretty big fonts so only act smarter than
                    ** thou when we have no choice... we can trust
                    ** the numbers under olwm and cde
                    */
                    if  ((correctWMRest < 0) || (correctWMRest > 60))
                        correctWMRest = wdata->left;
                    if  ((correctWMTop < 0) || (correctWMTop > 100))
                        correctWMTop = wdata->top;
 

#ifdef _pauly_debug
		fprintf(stdout,">>>>Other WM insets: %d, %d\n", correctWMRest, correctWMTop);
		fflush(stdout);
#endif /* _pauly_debug */
		break;
		
	    }  /* switch (runningWM) */
	    
            topCorrection = correctWMTop - topGuess;
            restCorrection = correctWMRest - sideGuess;
            if ( topCorrection != 0 || restCorrection != 0 )
            {
 
#ifdef _pauly_debug
fprintf(stdout,">>>>insets fixed: %d, %d\n", correctWMRest, correctWMTop);
fflush(stdout);
#endif /* _pauly_debug */
                topGuess = correctWMTop;
                sideGuess = correctWMRest;
                wdata->top += topCorrection;
                wdata->bottom += restCorrection;
                wdata->left += restCorrection;
                wdata->right += restCorrection;

                awtJNI_ChangeInsets(env, this, wdata);

                isPacked = (*env)->GetBooleanField( env, target, 
						componentIDs.isPacked);
#ifdef _pauly_debug
fprintf(stdout,">>>>isPacked = %d\n", isPacked );
fflush(stdout);
#endif /* _pauly_debug */

                if  (isPacked)
                {
                    /* if this window has been sized by a pack() the
                    ** user wants to keep the interior (or Client area)
                    ** geometry intact.
                    **
                    ** We must adjust the target dimensions appropriately.
                    ** These were calculated originally with bogus insets.
                    */

                    correctTargetW = width + (2 * restCorrection);
                    correctTargetH = height +(topCorrection+restCorrection);

                    (*env)->SetIntField(env, target, componentIDs.width,
                                        (jint) correctTargetW);
                    (*env)->SetIntField(env, target, componentIDs.height,
                                        (jint) correctTargetH);
                    /*
                    **  Normally you only reconfigure the outerCanvas due to
                    **  handling the ReconfigureNotify on the innerCanvas.
                    **  However, in this case the innerCanvas may not have
                    **  changed, but outterCanvas may still need to, since the
                    **  insets have changed.
                    */
                    reshape(env, this, wdata, x, y,
                                        correctTargetW, correctTargetH);
                    reconfigureOuterCanvas ( env, target, this, wdata );
                } else {
                    reshape(env, this, wdata, x, y, width, height);
                    JNU_CallMethodByName(env, NULL, this,
                        "handleResize", "(II)V", width, height );
                }

            } else {
    		if  ( (!wdata->isResizable) && (!wdata->isFixedSizeSet) ) {
			reshape(env, this, wdata, x, y, width, height);
    		}
	    }
	}
	(*env)->DeleteLocalRef(env, target);
	break;
	
    case ConfigureNotify:
	if ((*env)->EnsureLocalCapacity(env, 1) < 0)
	    break;
	
	target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
	
	/* We can detect the difference between a move and a resize
	 * by checking the send_event flag on the event; if it's true,
	 * then it's indeed a move, if it's false, then this is a resize
	 * and we do not want to process it as a "move" (for resizes
	 * the x,y values are misleadingly set to 0,0, and so just
	 * checking for an x,y delta won't work).
	 */
	
	getTargetX = (*env)->GetIntField(env, target, componentIDs.x);
	getTargetY = (*env)->GetIntField(env, target, componentIDs.y);
	    
	    /* N.B. The wdata top offset is the offset from the
	       outside of the entire (bordered window) to the
	       inner/parent drawing area (canvas), NOT to the shell.
	       Thus, if a menubar is present and/or a warning
	       window at the top (not NETSCAPE), the top offset
	       will also include space for these.  In order to
	       position the abstract java window relative to the
	       shell, we must add back in the appropriate space
	       for these when we subtract off the wdata top field.      */
#ifdef NETSCAPE
	wwHeight = 0;
#else /* NETSCAPE */
	if  (wdata->warningWindow != NULL)
		wwHeight = wdata->wwHeight;
	else
		wwHeight = 0;
#endif /* NETSCAPE */
	topAdjust = wdata->mbHeight + wwHeight;
	setTargetX = event->xconfigure.x - wdata->left;
	setTargetY = event->xconfigure.y - wdata->top + topAdjust;

	if ((getTargetX != setTargetX || getTargetY != setTargetY) &&
	    event->xconfigure.send_event == TRUE) {
	    (*env)->SetIntField(env, target, componentIDs.x, (jint) setTargetX);
	    (*env)->SetIntField(env, target, componentIDs.y, (jint) setTargetY);
#ifdef _pauly_debug
	    fprintf(stdout, " ++ shell move. Xevent x,y: %d, %d.\n",
		    event->xconfigure.x, event->xconfigure.y);
	    fprintf(stdout, "    shell move. left: %d, top: %d, but offset: %d\n", wdata->left, wdata->top, topAdjust);
	    fprintf(stdout,"    shell move. target x: %d, target y: %d\n", setTargetX, setTargetY);
	    fprintf(stdout,"    shell move. ww height: %d\n", wwHeight);
	    fflush(stdout);
#endif /* _pauly_debug */
	    JNU_CallMethodByName(env
				 ,NULL
				 ,this
				 ,"handleMoved"
				 ,"(II)V", setTargetX, setTargetY);
	    if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }
#ifdef _pauly_debug
	    fprintf(stdout,"    shell move. Done.\n");
	    fflush(stdout);
#endif /* _pauly_debug */
	} else if  (event->xconfigure.send_event == FALSE) {
#ifdef _pauly_debug
		fprintf(stdout, " ++ shell resize. Xevent x,y,w,h: %d, %d, %d, %d.\n",
			event->xconfigure.x, event->xconfigure.y,event->xconfigure.width,event->xconfigure.height);
		fflush(stdout);
#endif /* _pauly_debug */
		wdata->shellResized = True;
	    }
	
	(*env)->DeleteLocalRef(env, target);
	break;
	
    default:
	break;
    }
    
}


static void
Frame_quit(Widget w,
	   XtPointer client_data,
	   XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    
    JNU_CallMethodByName(env, NULL, (jobject) client_data, "handleQuit", "()V");
    if ((*env)->ExceptionOccurred(env)) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }
}


static void
Frame_activate(Widget w,
	       XtPointer client_data,
	       XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    
    awt_setActivatedShell(w);
    
    JNU_CallMethodByName(env, NULL, (jobject) client_data, "handleActivate", "()V");
    if ((*env)->ExceptionOccurred(env)) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }
}


static void
setDeleteCallback(jobject this, struct FrameData *wdata)
{
    Atom xa_WM_DELETE_WINDOW;
    Atom xa_WM_TAKE_FOCUS;
    Atom xa_WM_PROTOCOLS;
    
    XtVaSetValues(wdata->winData.shell,
		  XmNdeleteResponse, XmDO_NOTHING,
		  NULL);
    xa_WM_DELETE_WINDOW = XmInternAtom(XtDisplay(wdata->winData.shell),
				       "WM_DELETE_WINDOW", False);
    xa_WM_TAKE_FOCUS = XmInternAtom(XtDisplay(wdata->winData.shell),
				    "WM_TAKE_FOCUS", False);
    xa_WM_PROTOCOLS = XmInternAtom(XtDisplay(wdata->winData.shell),
				   "WM_PROTOCOLS", False);
    
    XmAddProtocolCallback(wdata->winData.shell,
			  xa_WM_PROTOCOLS,
			  xa_WM_DELETE_WINDOW,
			  Frame_quit, (XtPointer) this);
    XmAddProtocols(wdata->winData.shell,
		   xa_WM_PROTOCOLS,
		   &xa_WM_TAKE_FOCUS,
		   1);
    
    /* Now using FocusIn event in the Frame_event_handler function.  OpenLook
       Window Managers do not appear to adhere to the ICCCM protocol when it comes
       to WM_TAKE_FOCUS
       XmAddProtocolCallback(wdata->winData.shell,
       xa_WM_PROTOCOLS,
       xa_WM_TAKE_FOCUS,
       Frame_activate, (XtPointer) this);
       */
}


/* sun_awt_motif_MFramePeer_create() is native (X/Motif) create routine */

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;Ljava/lang/Object;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_create
(JNIEnv *env, jobject this, jobject parent, jobject arg)
{
    Arg			args[40];
    int			argc;
    struct FrameData	*wdata;
    int			isEmbedded;
    jobject		target;
    jstring		warningString;
    jboolean		resizable;
    jint		state;
    jclass		clazz;
    jobject		globalRef = awtJNI_CreateAndSetGlobalRef(env, this);
    
    unsigned		runningWM;	/* the running Window Manager	*/
    Widget              innerCanvasW;   /* form's child, parent of the
					   outer canvas (drawing area)  */
    Position            x,y;
    Dimension           w,h;


    AWT_LOCK();
    
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    
    if (JNU_IsNull(env, target)) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    (*env)->SetObjectField(env, this, mFramePeerIDs.insets, arg);
    
    wdata = ZALLOC(FrameData);
    (*env)->SetLongField(env, this, mComponentPeerIDs.pData, (jlong) wdata);
    if (wdata == 0) {
	JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
	AWT_UNLOCK();
	return;
    }
    /* A variation on Netscape's hack for embedded frames: the client area
     * of the browser is a Java Frame for parenting purposes, but really a
     * Motif child window
     */
    clazz = (*env)->FindClass(env, "sun/awt/EmbeddedFrame");
    isEmbedded = (*env)->IsInstanceOf(env, target, clazz);
    
    if (isEmbedded) {
	wdata->winData.flags |= W_IS_EMBEDDED;
	
	(*env)->SetIntField(env, arg, insetsIDs.top, 0);
	(*env)->SetIntField(env, arg, insetsIDs.left, 0);
	(*env)->SetIntField(env, arg, insetsIDs.bottom, 0);
	(*env)->SetIntField(env, arg, insetsIDs.right, 0);
	
	wdata->top = 0;
	wdata->left = 0;
	wdata->bottom = 0;
	wdata->right = 0;
	wdata->reparented = True;
    } else {
	/* The insets will be corrected upon the reparent
	   event in shellEH().  For now, use bogus values.	*/
	wdata->top = (*env)->GetIntField(env, arg, insetsIDs.top);
	wdata->left = (*env)->GetIntField(env, arg, insetsIDs.left);
	wdata->bottom = (*env)->GetIntField(env, arg, insetsIDs.bottom);
	wdata->right = (*env)->GetIntField(env, arg, insetsIDs.right);
	awt_Frame_guessInsets(wdata);
	awtJNI_ChangeInsets(env, this, wdata);
        wdata->reparented = False;
    }
    x = (*env)->GetIntField(env, target, componentIDs.x) + wdata->left;
    y = (*env)->GetIntField(env, target, componentIDs.y) + wdata->top;
    w = (*env)->GetIntField(env, target, componentIDs.width) 
		- ( wdata->left + wdata->right ); 
    h = (*env)->GetIntField(env, target, componentIDs.height)
    		- ( wdata->top + wdata->bottom );

    
    wdata->isModal = 0;
    wdata->mappedOnce = 0;
    wdata->isShowing = False;
    wdata->shellResized = False;
    wdata->canvasResized = False;
    wdata->menuBarReset = False;
    wdata->need_reshape = False;
    
    
    resizable = (*env)->GetBooleanField(env, target, frameIDs.resizable);
    state = (*env)->GetIntField(env, target, frameIDs.state);
    
    wdata->isIconic = (state == java_awt_Frame_ICONIFIED)? True : False;
    
    /* create a top-level shell */
    argc = 0;
    XtSetArg(args[argc], XmNsaveUnder, False); argc++;
    if (resizable) {
	XtSetArg(args[argc], XmNallowShellResize, True); argc++;
    } else {
	XtSetArg(args[argc], XmNallowShellResize, False); argc++;
    }
    XtSetArg(args[argc], XmNvisual, defaultConfig->awt_visInfo.visual); argc++;
    XtSetArg(args[argc], XmNcolormap, defaultConfig->awt_cmap); argc++;
    XtSetArg(args[argc], XmNdepth, defaultConfig->awt_depth); argc++;
    XtSetArg(args[argc], XmNmappedWhenManaged, False); argc++;
    XtSetArg(args[argc], XmNx, x); argc++;
    XtSetArg(args[argc], XmNy, y); argc++;
    XtSetArg(args[argc], XmNwidth, w); argc++;
    XtSetArg(args[argc], XmNheight, h); argc++;
 
    wdata->winData.shell = XtAppCreateShell("AWTapp","XApplication",
					    vendorShellWidgetClass,
					    awt_display,
					    args,
					    argc);
    setDeleteCallback(globalRef, wdata);
    
    /* Establish resizability.  For the case of not resizable, do not
       yet set a fixed size here; we must wait until in the routine
       sun_awt_motif_MFramePeer_pReshape() after insets have been fixed.
       This is because correction of the insets may affect shell size.
       (See comments in shellEH() concerning correction of the insets.  */
    wdata->isResizable = resizable;
    wdata->isFixedSizeSet = False;

    if  (resizable) {
#ifdef _pauly_debug
	    fprintf(stdout,"    create. setting shell resizable\n");
	    fflush(stdout);
#endif /* _pauly_debug */
	    awt_util_setShellResizable(wdata->winData.shell, wdata->isShowing);
    }
    
    XtAddEventHandler(wdata->winData.shell, StructureNotifyMask | FocusChangeMask,
		      FALSE, shellEH, globalRef);
    
    
    argc = 0;
    XtSetArg(args[argc], XmNmarginWidth, 0); argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0); argc++;
    XtSetArg(args[argc], XmNhorizontalSpacing, 0); argc++;
    XtSetArg(args[argc], XmNverticalSpacing, 0); argc++;
    XtSetArg(args[argc], XmNresizePolicy, XmRESIZE_NONE); argc++;
    wdata->mainWindow = XmCreateForm(wdata->winData.shell, "main", args, argc);
    
    /* The widget returned by awt_canvas_create is a drawing area
       (i.e., canvas) which is the child of another drawing area
       parent widget.  The parent is the drawing area within the
       form just created.  The child is an drawing area layer over
       the entire frame window, including the form, any menu bar
       and warning windows present, and also window manager stuff.
       The top, bottom, left, and right fields in wdata maintain
       the respective offsets between these two drawing areas.  */
    
    wdata->winData.comp.widget = awt_canvas_create((XtPointer)globalRef,
						   wdata->mainWindow,
						   "frame_",
						   -1,
						   -1,
						   True,
						   wdata,
                                                   defaultConfig);
    
    wdata->mappedOnce = 0;
    XtAddCallback(wdata->winData.comp.widget,
		  XmNresizeCallback,
		  outerCanvasResizeCB,
		  globalRef);
    
    innerCanvasW = XtParent(wdata->winData.comp.widget);
    XtVaSetValues(innerCanvasW,
		  XmNleftAttachment, XmATTACH_FORM,
		  XmNrightAttachment, XmATTACH_FORM,
		  NULL);
    
    XtAddEventHandler(innerCanvasW, StructureNotifyMask, FALSE,
		      innerCanvasEH, globalRef);
    
    /* No menu bar initially */
    wdata->menuBar = NULL;
    wdata->mbHeight = 0;
    
    /* If a warning window (string) is needed, establish it now.*/
    warningString =
	(*env)->GetObjectField(env, target, windowIDs.warningString);
    if (!JNU_IsNull(env, warningString) && !isEmbedded) {
	char *wString;
	/* Insert a warning window. It's height can't be set yet;
	   it will later be set in setMbAndWwHeightAndOffsets().*/
	wString = (char *) JNU_GetStringPlatformChars(env, warningString, NULL);
	wdata->warningWindow = awt_util_createWarningWindow(wdata->mainWindow, wString);
	JNU_ReleaseStringPlatformChars(env, warningString, (const char *) wString);
	
	wdata->wwHeight = 0;
	XtVaSetValues(wdata->warningWindow,
		      XmNleftAttachment, XmATTACH_FORM,
		      XmNrightAttachment, XmATTACH_FORM,
		      NULL);
#ifdef _pauly_debug
	fprintf(stdout,"    create. warning window\n");
	fflush(stdout);
#endif /* _pauly_debug */
	
#ifdef NETSCAPE
	/* For NETSCAPE, warning window is at bottom of the form*/
	XtVaSetValues(innerCanvasW,
		      XmNtopAttachment, XmATTACH_FORM,
		      NULL);
	XtVaSetValues(wdata->warningWindow,
		      XmNtopAttachment, XmATTACH_WIDGET,
		      XmNtopWidget, innerCanvasW,
		      XmNbottomAttachment, XmATTACH_FORM,
		      NULL);
#else  /* NETSCAPE */
	/* Otherwise (not NETSCAPE), warning is at top of form  */
	XtVaSetValues(wdata->warningWindow,
		      XmNtopAttachment, XmATTACH_FORM,
		      NULL);
	XtVaSetValues(innerCanvasW,
		      XmNtopAttachment, XmATTACH_WIDGET,
		      XmNtopWidget, wdata->warningWindow,
		      XmNbottomAttachment, XmATTACH_FORM,
		      NULL);
#endif /* NETSCAPE */
	
#ifdef _pauly_debug
	fprintf(stdout,"    create. top offset: %d, bottom offset: %d\n",
		wdata->top, wdata->bottom);
	fflush(stdout);
#endif /* _pauly_debug */
    } else {
	/* No warning window present */
	XtVaSetValues(innerCanvasW,
		      XmNtopAttachment, XmATTACH_FORM,
		      XmNbottomAttachment, XmATTACH_FORM,
		      NULL);
	wdata->warningWindow = NULL;
	wdata->wwHeight = 0;
    }
    awt_util_show(wdata->winData.comp.widget);
    
    AWT_FLUSH_UNLOCK();
    
#ifdef _pauly_debug
    fprintf(stdout,"    create. shell: %x, form: %x, inner canvas: %x, outer canvas: %x\n",
	    wdata->winData.shell, wdata->mainWindow, innerCanvasW, wdata->winData.comp.widget);
    fprintf(stdout,"    create. Done.\n");
    fflush(stdout);
#endif /* _pauly_debug */
    
} /* MFramePeer_create() */


/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pSetTitle
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pSetTitle
(JNIEnv *env, jobject this, jstring title)
{
    char *ctitle;
    char *empty_string = "";
    struct FrameData *wdata;
    XTextProperty text_prop;
    char *c[1];

    AWT_LOCK();
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (wdata == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    ctitle = (JNU_IsNull(env, title)) ? empty_string : (char *) JNU_GetStringPlatformChars(env, title, NULL);
    c[0] = ctitle;
    
    /* need to convert ctitle to CompoundText */
    XmbTextListToTextProperty(awt_display, c, 1,
			      XCompoundTextStyle,&text_prop);
    XtVaSetValues(wdata->winData.shell,
		  XmNtitle, text_prop.value,
		  XmNtitleEncoding, text_prop.encoding,
		  XmNiconName, text_prop.value,
		  XmNiconNameEncoding, text_prop.encoding,
		  XmNname, ctitle,
		  NULL);
    if (ctitle != empty_string)
	JNU_ReleaseStringPlatformChars(env, title, (const char *) ctitle);
    
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pShow
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pShow
(JNIEnv *env, jobject this)
{
    struct FrameData *wdata;
    XWMHints *myHints;
    
#ifdef _pauly_debug
    fprintf(stdout," ++ in ..._pShow()\n");
    fflush(stdout);
#endif /* _pauly_debug */
    
    AWT_LOCK();
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.comp.widget == 0 || wdata->winData.shell == 0 ||
	wdata->mainWindow == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    XtVaSetValues(wdata->winData.comp.widget,
		  XmNx, -(wdata->left),
		  XmNy, -(wdata->top), NULL);
    
    if (wdata->menuBar != 0) {
	awt_util_show(wdata->menuBar);
    }
    XtManageChild(wdata->mainWindow);
    if (XtWindow(wdata->winData.shell) == 0) {
	XtRealizeWidget(wdata->winData.shell);
    }

    /* let's honor the iconic state here */
    if ((myHints = XGetWMHints(awt_display, XtWindow(wdata->winData.shell))) 
	!= NULL) {
	if (wdata->isIconic /* && myHints->initial_state != IconicState */ ) {
	    myHints->flags |= StateHint;
	    myHints->initial_state = IconicState;
	    XSetWMHints(awt_display, XtWindow(wdata->winData.shell), myHints);
	} else /* if (myHints->initial_state != NormalState ) */ {
	    myHints->flags |= StateHint;
	    myHints->initial_state = NormalState;
	    XSetWMHints(awt_display, XtWindow(wdata->winData.shell), myHints);
	}
	XFree(myHints); 
    }

    if (wdata->winData.flags & W_IS_EMBEDDED) {
	jobject embeddedFrame;
	jobject embeddedPeer;
	Widget embeddedWidget;

	embeddedFrame = (*env)->GetObjectField(env, this,
					       mComponentPeerIDs.target);
	embeddedPeer = (*env)->GetObjectField(env, embeddedFrame,
					      componentIDs.peer);
	embeddedWidget =
	    (Widget) (*env)->GetLongField(env, embeddedPeer,
					  mEmbeddedFramePeerIDs.handle);
	
	if (embeddedWidget != 0) {
	    XReparentWindow(XtDisplay(wdata->winData.shell),
			    XtWindow(wdata->winData.shell),
			    XtWindow(embeddedWidget), 0, 0);
	}
    }
    XtManageChild(wdata->winData.comp.widget);
    XtSetMappedWhenManaged(wdata->winData.shell, True);
    XtPopup(wdata->winData.shell, XtGrabNone);
    XRaiseWindow(awt_display, XtWindow(wdata->winData.shell));
    wdata->isShowing = True;
    AWT_FLUSH_UNLOCK();
#ifdef _pauly_debug
    fprintf(stdout,"    ..._pShow().  Done.\n");
    fflush(stdout);
#endif /* _pauly_debug */
}

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    getState
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MFramePeer_getState
(JNIEnv *env, jobject this)
{
    struct FrameData *wdata;
    Boolean iconic;
    Boolean dynamic;
    
    AWT_LOCK();
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return JNI_FALSE;
    }
    iconic = wdata->isIconic;
    
    AWT_FLUSH_UNLOCK();
    return (iconic == True)? java_awt_Frame_ICONIFIED : java_awt_Frame_NORMAL;
}

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    setState
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_setState
(JNIEnv *env, jobject this, jint state)
{
    struct FrameData *wdata;
    Window w;
    Bool oldIconic;
    
    AWT_LOCK();
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    
    oldIconic = wdata->isIconic;

    wdata->isIconic = ((state == java_awt_Frame_ICONIFIED)? True : False);
    
    if (oldIconic != wdata->isIconic) {
	if (oldIconic) {
	    JNU_CallMethodByName(env, NULL, this, "handleDeiconify", "()V");
	    if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }
	} else {
	    JNU_CallMethodByName(env, NULL, this, "handleIconify", "()V");
	    if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    } 
	}

	w = XtWindow(wdata->winData.shell);
	/*
	 * calling XmapWindow() on a hidden window will make it visible -- not desired
         */
	if (w != 0 && wdata->isShowing) {
	    if (wdata->isIconic) {
		XIconifyWindow(awt_display, w, XDefaultScreen(awt_display));
	    } else {
		XMapWindow(awt_display, w);
	    }        
	}
    }
    AWT_FLUSH_UNLOCK();
}


/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pHide
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pHide
(JNIEnv *env, jobject this)
{
    struct FrameData *wdata;
    
    AWT_LOCK();
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.comp.widget == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    if (XtWindow(wdata->winData.shell) != 0) {
	XtUnmanageChild(wdata->winData.comp.widget);
	XtPopdown(wdata->winData.shell);
    }
    wdata->isShowing = False;
    AWT_FLUSH_UNLOCK();
}


/* sun_awt_motif_MFramePeer_pReshape() is native (X/Motif) routine that
   is called to effect a reposition and / or resize of the target frame.
   The parameters x,y,w,h specify target's x, y position, width, height.*/

/* this functionality  is invoked from both java and native code, and we only want to lock when invoking it from
 * java, so wrap the native method version with the locking */

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pReshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pReshape
(JNIEnv *env, jobject this, jint x, jint y, jint w, jint h)
{
    struct FrameData	*wdata;
    
    AWT_LOCK();
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (wdata == 0 ||
	wdata->winData.comp.widget == 0 ||
	wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    
    reshape(env, this, wdata, x, y, w, h);
    
    AWT_FLUSH_UNLOCK();
}

static void
reshape(JNIEnv *env, jobject this, struct FrameData *wdata,
	jint x, jint y, jint w, jint h)
{
    long	topAdjust,	/* top adjustment of offset	*/
		bottomAdjust;	/* bottom adjustment of offset	*/
    long	width,		/* of X/Motif shell and form	*/
		height;		/* of X/Motif shell and form	*/	
    long        w1, h1;

    
#ifdef _pauly_debug
    fprintf(stdout," ++ reshape.\n");
    fflush(stdout);
#endif /* _pauly_debug */
    
    
    /* Motif Window Manager (MWM) won't allow toplevel shell at (0,0).
       It's ok in Open Look Window Manger (OLWM).  And other WMs !?     */
    if ( !wdata->reparented )
    {
        int WM = awt_util_runningWindowManager();
        if ( (WM == MOTIF_WM) || (WM == CDE_WM) )
        {
            if ( (x == 0) && (y == 0) )
            {
                x = y = 1;
            }
        }
    }    


    /* Make adjustments in case of a dynamically added/removed menu bar */
    awtJNI_setMbAndWwHeightAndOffsets(env, this, wdata);
    
#ifdef _pauly_debug
    fprintf(stdout,"    reshape. x: %d, y: %d, w: %d, h: %d\n", x, y, w, h);
    fprintf(stdout,"    reshape. offsets - top: %d, bottom: %d, left: %d, right: %d\n",
	    wdata->top, wdata->bottom, wdata->left, wdata->right);
    fflush(stdout);
#endif /* _pauly_debug */
    
    /* The abstract java (target) position coordinates (x,y)
       are for the bordered window.  Eventually(!), the Motif
       (shell) coordinates (XmNx, XmNy) will exclude borders.
       (This is true only AFTER shell is massaged by the WM.)   */
    
    /* The abstract java (target) width and height includes any WM
       borders. But the Motif width and height excludes WM borders.
       The wdata top and bottom fields may include space for menu bar,
       warning window, etc. We must adjust by these values for shell.   */
    topAdjust = 0;
    bottomAdjust = 0;
    /* Surprise - do not(!) check for nonNull MenuBar because that can
       occur separately (in ...pSetMenubar()) from calculation of the
       menu bar height and offsets (in setMbAndWwHeightAndOffsets()).
       In any event, the offsets and wdata mbHeight field should jive.  */
    topAdjust += wdata->mbHeight;
    if  (wdata->warningWindow != NULL) {
#ifdef NETSCAPE
	bottomAdjust += wdata->wwHeight;
#else /* NETSCAPE */
	topAdjust += wdata->wwHeight;
#endif /* NETSCAPE */
    }
    if (wdata->hasTextComponentNative) {
	bottomAdjust +=  wdata->imHeight;
    }
#ifdef _pauly_debug
    fprintf(stdout,"    reshape. adjustments - top: %d, bottom: %d\n", topAdjust, bottomAdjust);
    fflush(stdout);
#endif /* _pauly_debug */
    
    /* Bugtraq 4033151.  If nonpositive size specified (e.g., if no size given),
       establish minimum allowable size.  Note. Motif shell can not be sized 0. */
    
    width = w - (wdata->left + wdata->right);
    height = h - (wdata->top + wdata->bottom) + (topAdjust + bottomAdjust);
    

    /* shell */
    w1 = (width > 0) ? width : 1;
    h1 = (height > 0) ? height : 1;

    if (wdata->imRemove){
	XtVaSetValues(XtParent(wdata->winData.comp.widget),
		      XmNheight, (((h - (wdata->top + wdata->bottom)) > 0) ?
				  (h - (wdata->top + wdata->bottom)) : 1),
		      NULL);
	wdata->imRemove = False;
    }
    
    XtVaSetValues(wdata->winData.shell,
                  XmNx, x,
                  XmNy, y,
                  XmNwidth, w1,
                  XmNheight, h1,
                  NULL);
    
    /* inner/parent drawing area (parent is form) */
    h1 = h - (wdata->top + wdata->bottom);
    h1 = ( h1 > 0 ) ? h1 : 1;
#if 0
    XtConfigureWidget(XtParent(wdata->winData.comp.widget),
                      0, topAdjust, w1, h1, 0 );
#else
    XtVaSetValues(XtParent(wdata->winData.comp.widget),
                  XmNx, 0,
                  XmNy, topAdjust,
                  XmNwidth, w1,
                  XmNheight, h1,
                  NULL);
#endif

#ifdef _pauly_debug
	fprintf(stdout,"    reshape. setting inner canvas to: %d,%d,%d,%d\n",
	0, topAdjust, w1, h1 );
	fflush(stdout);
#endif /* _pauly_debug */

    wdata->menuBarReset = False;
    
    /* If not resizable and the fixed (nonresizable) size not yet set,
       and if the insets have already been corrected during shellEH(),
       and if this is a valid size, then establish the fixed size now.  */
    if  ( (!wdata->isResizable) && (!wdata->isFixedSizeSet) &&
	  (wdata->reparented) && (width > 0) && (height  > 0) )
    {
#ifdef _pauly_debug
	fprintf(stdout,"    reshape. setting shell not resizable\n");
	fflush(stdout);
#endif /* _pauly_debug */
	awt_util_setShellNotResizable(wdata->winData.shell, width, height, wdata->isShowing);
	wdata->isFixedSizeSet = True;
    }
    
#ifdef _pauly_debug
    fprintf(stdout,"   reshape. Done.\n");
    fflush(stdout);
#endif /* _pauly_debug */
    return;
} 


/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pDispose
(JNIEnv *env, jobject this)
{
    struct FrameData *wdata;
    
    AWT_LOCK();
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->mainWindow == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    
    XtUnmanageChild(wdata->winData.comp.widget);
    awt_delWidget(wdata->winData.comp.widget);
    awt_util_consumeAllXEvents(wdata->winData.comp.widget);
    XtDestroyWidget(wdata->winData.comp.widget);
    
    XtUnmanageChild(wdata->mainWindow);
    awt_util_consumeAllXEvents(wdata->mainWindow);
    awt_util_consumeAllXEvents(wdata->winData.shell);
    XtDestroyWidget(wdata->mainWindow);
    XtDestroyWidget(wdata->winData.shell);
    if (wdata->iconPixmap) {
	XFreePixmap(awt_display, wdata->iconPixmap);
    }
    
    free((void *) wdata);
    
    (*env)->SetLongField(env, this, mComponentPeerIDs.pData, 0);
    awtJNI_DeleteGlobalRef(env, this);
    
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pGetIconSize
 * Signature: (Lsun/awt/image/ImageRepresentation;)V
 */
JNIEXPORT jboolean JNICALL Java_sun_awt_motif_MFramePeer_pGetIconSize
(JNIEnv *env, jobject this, jint widthHint, jint heightHint)
{
    struct FrameData *wdata;
    unsigned int width, height, border_width, depth;
    Window win;
    int x, y;
    unsigned long mask;
    XSetWindowAttributes attrs;
    int saveWidth = -1;
    int saveHeight = -1;
    unsigned int dist = 0xffffffff;
    int diff = 0;
    int closestWidth;
    int closestHeight;
    int newDist;
    int found = 0;
    
    AWT_LOCK();
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return FALSE;
    }
    XtVaGetValues(wdata->winData.shell,
		  XmNiconWindow, &win,
		  NULL);
    if (!win) {
	int count;
	int i;
	XIconSize *sizeList;
	if (!XGetIconSizes(awt_display, awt_root,
			   &sizeList, &count)) {
	    AWT_UNLOCK();
	    /* No icon sizes so can't set it -- Should we throw an exception?*/
	    return FALSE;
	}
	for (i=0; i < count; i++) {
	    if (widthHint >= sizeList[i].min_width &&
		widthHint <= sizeList[i].max_width &&
		heightHint >= sizeList[i].min_height &&
		heightHint <= sizeList[i].max_height) {
		found = 1;
		if ((((widthHint-sizeList[i].min_width)
		      % sizeList[i].width_inc) == 0) &&
		    (((heightHint-sizeList[i].min_height)
		      % sizeList[i].height_inc) ==0)) {
		    /* Found an exact match */
		    saveWidth = widthHint;
		    saveHeight = heightHint;
		    dist = 0;
		    break;
		}
		diff = widthHint - sizeList[i].min_width;
		if (diff == 0) {
		    closestWidth = widthHint;
		} else {
		    diff = diff%sizeList[i].width_inc;
		    closestWidth = widthHint - diff;
		}
		diff = heightHint - sizeList[i].min_height;
		if (diff == 0) {
		    closestHeight = heightHint;
		} else {
		    diff = diff%sizeList[i].height_inc;
		    closestHeight = heightHint - diff;
		}
		newDist = closestWidth*closestWidth +
		    closestHeight*closestHeight;
		if (dist > newDist) {
		    saveWidth = closestWidth;
		    saveHeight = closestHeight;
		    dist = newDist;
		}
	    }
	}
	
	if (!found) {
	    /* REMIND: Aspect ratio */
	    if (widthHint  >= sizeList[0].max_width &&
		heightHint >= sizeList[0].max_height) {
		saveWidth = sizeList[0].max_width;
		saveHeight = sizeList[0].max_height;
	    } else if (widthHint  >= sizeList[0].min_width &&
		       heightHint >= sizeList[0].min_height) {
		saveWidth = sizeList[0].min_width;
		saveHeight = sizeList[0].min_height;
	    } else {
		saveWidth = (sizeList[0].min_width+sizeList[0].max_width)/2;
		saveHeight = (sizeList[0].min_height+sizeList[0].max_height)/2;
	    }
	}
	free((void *) sizeList);
    } else {
	Window root;
	if (XGetGeometry(awt_display,
			 win,
			 &root,
			 &x,
			 &y,
			 &saveWidth,
			 &saveHeight,
			 &border_width,
			 &depth)) {
	}
    }
    
    (*env)->SetIntField(env, this, mFramePeerIDs.iconWidth, (jint)saveWidth);
    (*env)->SetIntField(env, this, mFramePeerIDs.iconHeight, (jint)saveHeight);
    
    AWT_UNLOCK();
    
    return TRUE;
}
/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pSetIconImage
 * Signature: (Ljava/awt/image/DataBuffer;II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pSetIconImage
(JNIEnv *env, jobject this, jbyteArray jbyteData, jintArray jintData,
 jint iconWidth, jint iconHeight)
{
    struct FrameData *wdata;
    Window win;
    GC gc;
    int x, y;
    XImage *dst;
    unsigned long mask;
    XSetWindowAttributes attrs;
    jobject jbuf = NULL;
    void *buf = NULL;
    int len = 0;
    int bpp, slp, bpsl;
	    
    if (JNU_IsNull(env, jbyteData)) {
	if (JNU_IsNull(env, jintData)) {
	    JNU_ThrowNullPointerException(env, "NullPointerException");
	    return;
	}
	jbuf = jintData;
    } else {
	jbuf = jbyteData;
	len = (*env)->GetArrayLength(env, jbyteData);
    }
    AWT_LOCK();
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    /* REMIND: Need to figure out how to display image on a pixmap */
    
    if (wdata == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    
    /* Make sure it is the right size */
    if (wdata->iconPixmap &&
	(wdata->iconWidth != iconWidth || wdata->iconHeight != iconHeight)) {
	XFreePixmap(awt_display, wdata->iconPixmap);
	wdata->iconPixmap = NULL;
    }
    
    if (wdata->iconPixmap == NULL) {
	if ((wdata->iconPixmap = XCreatePixmap(awt_display, awt_root,
					       iconWidth, iconHeight,
                                    defaultConfig->awtImage->Depth)) == NULL) {
	    /* REMIND: How to warn that there was a problem? */
	    AWT_UNLOCK();
	    return;
	}
	wdata->iconWidth = iconWidth;
	wdata->iconHeight = iconHeight;
    }
    
    buf = (void *) (*env)->GetPrimitiveArrayCritical(env, jbuf, NULL);
    if (jbyteData != NULL) {
	int i;
	unsigned char *ubuf = (unsigned char *) buf;
	/* Need to map from ICM lut to cmap */
	for (i=0; i < len; i++) {
            ubuf[i] = (ubuf[i] >= defaultConfig->color_data->awt_numICMcolors)
			? 0
                	: defaultConfig->color_data->awt_icmLUT2Colors[ubuf[i]];
	}
    }
    
    bpp = defaultConfig->awtImage->wsImageFormat.bits_per_pixel;
    slp = defaultConfig->awtImage->wsImageFormat.scanline_pad;
    bpsl = paddedwidth(iconWidth * bpp, slp) >> 3;
    if (((bpsl << 3) / bpp) < iconWidth) {
	AWT_UNLOCK();
	return;
    }
    dst = XCreateImage(awt_display, defaultConfig->awt_visInfo.visual,
                       defaultConfig->awtImage->Depth, ZPixmap, 0,
		       buf, iconWidth, iconHeight, 32, bpsl);
    if (dst == NULL) {
	/* REMIND: How to warn that there was a problem? */
	AWT_UNLOCK();
	return;
    }
    
    if ((gc = XCreateGC(awt_display, wdata->iconPixmap, NULL, 0)) == NULL) {
        XDestroyImage (dst);
	(*env)->ReleasePrimitiveArrayCritical(env, jbuf, buf, JNI_ABORT);
	AWT_UNLOCK();
	return;
    }
    
    XPutImage(awt_display, wdata->iconPixmap, gc, dst,
	      0, 0, 0, 0, iconWidth, iconHeight);
    (*env)->ReleasePrimitiveArrayCritical(env, jbuf, buf, JNI_ABORT);
    dst->data=NULL;
    XDestroyImage(dst);
    XFreeGC(awt_display, gc);
    
    XtVaGetValues(wdata->winData.shell,
		  XmNiconWindow, &win,
		  NULL);
    if (!win) {
	mask = CWBorderPixel | CWColormap | CWBackPixmap;
        attrs.border_pixel = awt_defaultFg;
        attrs.colormap = defaultConfig->awt_cmap;
	attrs.background_pixmap = wdata->iconPixmap;
	if (!(win = XCreateWindow(awt_display, awt_root,
				  0, 0, iconWidth, iconHeight,
                                  (unsigned) 0, 
				  defaultConfig->awtImage->Depth,
				  InputOutput,
                                  defaultConfig->awt_visInfo.visual,
				  mask, &attrs))) {
	    /* Still can't create the window so try setting iconPixmap */
	    XtVaSetValues(wdata->winData.shell,
			  XmNiconPixmap, wdata->iconPixmap,
			  NULL);
	    AWT_FLUSH_UNLOCK();
	    return;
	}
    }
    
    XtVaSetValues(wdata->winData.shell,
		  XmNiconWindow, win,
		  NULL);
    
    XSetWindowBackgroundPixmap(awt_display, win, wdata->iconPixmap);
    XClearWindow(awt_display, win);
    AWT_FLUSH_UNLOCK();
}


/* sun_awt_motif_MFramePeer_setResizable() is native (X/Motif) routine
   called to establish the frame's resizability.  Parameter resizable.  */

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    setResizable
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_setResizable
(JNIEnv *env, jobject this, jboolean resizable)
{
    struct FrameData	*wdata;
    jobject		target;
    int			targetWidth,
			targetHeight;
    long                width,          /* fixed width if not resizable */
			height;         /* fixed height if not resizable*/
    long                verticalAdjust; /* menubar, warning window, etc.*/
    
    if ((*env)->EnsureLocalCapacity(env, 1) < 0) {
	return;
    }
    
    AWT_LOCK();
    
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (wdata == 0 ||
	wdata->winData.comp.widget == 0 ||
	wdata->winData.shell == 0 ||
	JNU_IsNull(env, target)) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	if  (!JNU_IsNull(env, target))
	    (*env)->DeleteLocalRef(env, target);
	AWT_UNLOCK();
	return;
    }
    
    XtVaSetValues(wdata->winData.shell,
		  XmNallowShellResize, resizable ? True : False,
		  NULL);
    
    if  ((!wdata->isResizable) && (resizable)) {
#ifdef _pauly_debug
	fprintf(stdout,"    ...setResizable. make nonresizable into resizable\n");
	fflush(stdout);
#endif /* _pauly_debug */
	/* A previously nonresizable frame is now being made resizable */
	awt_util_setShellResizable(wdata->winData.shell, wdata->isShowing);
	wdata->isFixedSizeSet = False;
    } else if  ((wdata->isResizable) && (!resizable)) {
#ifdef _pauly_debug
	fprintf(stdout,"    ...setResizable. make resizable into nonresizable\n");
	fflush(stdout);
#endif /* _pauly_debug */
	/* A previously resizable frame is now being made nonresizable */
	
	/* To calculate fixed window width, height, we must subtract off
	   the window manager borders as stored in the wdata structure.
	   But note that the wdata top and bottom fields may include space
	   for warning window, menubar, IM status; this IS part of shell*/
	verticalAdjust = wdata->mbHeight;
	if  (wdata->warningWindow != NULL) {
	    verticalAdjust += wdata->wwHeight;
	}
	
	if  (wdata->hasTextComponentNative) {
	    verticalAdjust +=  wdata->imHeight;
	}
	
	targetWidth = (*env)->GetIntField(env, target, componentIDs.width);
	targetHeight = (*env)->GetIntField(env, target, componentIDs.height);
	width = targetWidth - (wdata->left + wdata->right);
	height = targetHeight - (wdata->top + wdata->bottom) + verticalAdjust;
	if  ((width > 0) && (height > 0)) {
#ifdef _pauly_debug
	    fprintf(stdout,"    ...setResizable. width: %ld, height: %ld\n", width, height);
	    fflush(stdout);
#endif /* _pauly_debug */
	    awt_util_setShellNotResizable(wdata->winData.shell, width, height, wdata->isShowing);
	    wdata->isFixedSizeSet = True;
	}
    }
    
    wdata->isResizable = (Boolean)resizable;
    
    (*env)->DeleteLocalRef(env, target);
    
    AWT_FLUSH_UNLOCK();
}


/* sun_awt_motif_MFramePeer_pSetMenuBar() is native (X/Motif) routine
   which handles insertion or deletion of a menubar from this frame.    */

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    pSetMenuBar
 * Signature: (Lsun/awt/motif/MMenuBarPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pSetMenuBar
(JNIEnv *env, jobject this, jobject mb)
{
    struct FrameData		*wdata;
    struct ComponentData	*mdata;
    jobject			target;
    Widget                      innerCanvasW;   /* Motif inner canvas   */
#ifdef _pauly_debug
    Dimension                   mbHeight;       /* Motif menubar height */
#endif /* _pauly_debug */
    
#ifdef _pauly_debug
    fprintf(stdout," ++ ...pSetMenuBar.\n");
    fflush(stdout);
#endif /* _pauly_debug */
    
    
    if ((*env)->EnsureLocalCapacity(env, 1) < 0) {
	return;
    }
    AWT_LOCK();
    
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (JNU_IsNull(env, target) || wdata == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	if  (!JNU_IsNull(env, target)) {
	    (*env)->SetBooleanField(env, target, frameIDs.mbManagement,
				    JNI_FALSE);
	    (*env)->DeleteLocalRef(env, target);
	}
	AWT_UNLOCK();
	return;
    }
    
    if (mb == 0) {
#ifdef _pauly_debug
	fprintf(stdout,"    ...pSetMenuBar. mb is null.\n");
	fflush(stdout);
#endif /* _pauly_debug */
	if  (wdata->menuBar != NULL) {
	    /* Redo attachments of other form widgets appropriately now */
	    innerCanvasW = XtParent(wdata->winData.comp.widget);
	    
	    if  (wdata->warningWindow == NULL) {
		/* no warning window: canvas is now attached to form    */
		XtVaSetValues(innerCanvasW,
			      XmNtopAttachment, XmATTACH_FORM,
			      NULL);
	    } else {
		/* warning window present - conditional on #define NETSCAPE:
		   if NETSCAPE, warning window is at bottom, so canvas is
		   attached to the form (as above); otherwise (not NETSCAPE),
		   warning window itself is instead attached to form.   */
#ifdef NETSCAPE
		XtVaSetValues(innerCanvasW,
			      XmNtopAttachment, XmATTACH_FORM,
			      NULL);
#else  /* NETSCAPE */
		XtVaSetValues(wdata->warningWindow,
			      XmNtopAttachment, XmATTACH_FORM,
			      NULL);
#endif /* NETSCAPE */
	    }
	    
	    wdata->menuBarReset = True;
	}
	wdata->menuBar = NULL;
	awtJNI_setMbAndWwHeightAndOffsets(env, this, wdata);
	(*env)->SetBooleanField(env, target, frameIDs.mbManagement, JNI_FALSE);
	(*env)->DeleteLocalRef(env, target);
	AWT_FLUSH_UNLOCK();
#ifdef _pauly_debug
	fprintf(stdout,"    ...pSetMenuBar. Done.\n");
	fflush(stdout);
#endif /* _pauly_debug */
	return;
    }
    
    mdata = (struct ComponentData *)
	(*env)->GetLongField(env, mb,
			     mMenuBarPeerIDs.pData);
    if (mdata == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	(*env)->SetBooleanField(env, target, frameIDs.mbManagement, JNI_FALSE);
	(*env)->DeleteLocalRef(env, target);
	AWT_UNLOCK();
	return;
    }
    
    /* OK - insert the new menu bar into the form (at the top).
       Redo the attachments of other form widgets appropriately.*/
    
    if  (wdata->menuBar == NULL)
	wdata->menuBarReset = True;
    wdata->menuBar = mdata->widget;
    
#ifdef _pauly_debug
    XtVaGetValues(mdata->widget, XmNheight, &mbHeight, NULL);
    fprintf(stdout,"    ...pSetMenuBar. new menu bar (widget %x, parent: %x) - menu bar height: %d\n", wdata->menuBar, XtParent(wdata->menuBar), mbHeight);
    fflush(stdout);
#endif /* _pauly_debug */
    
    XtVaSetValues(mdata->widget,
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_FORM,
		  XmNrightAttachment, XmATTACH_FORM,
		  NULL);
    
    innerCanvasW = XtParent(wdata->winData.comp.widget);
    
    if  (wdata->warningWindow == NULL) {
	/* no warning window: menu bar at top, canvas attached to it    */
	XtVaSetValues(innerCanvasW,
		      XmNtopAttachment, XmATTACH_WIDGET,
		      XmNtopWidget, mdata->widget,
		      NULL);
    } else {
	/* warning window present - conditional on #define NETSCAPE:
	   if NETSCAPE, warning window is at bottom, so canvas is
	   attached to menu bar (as above); otherwise (not NETSCAPE),
	   the warning window is attached just below the menu bar.  */
#ifdef NETSCAPE
	XtVaSetValues(innerCanvasW,
		      XmNtopAttachment, XmATTACH_WIDGET,
		      XmNtopWidget, mdata->widget,
		      NULL);
#else  /* NETSCAPE */
	XtVaSetValues(wdata->warningWindow,
		      XmNtopAttachment, XmATTACH_WIDGET,
		      XmNtopWidget, mdata->widget,
		      NULL);
#endif /* NETSCAPE */
    }
    
    XtManageChild(mdata->widget);
    XtMapWidget(mdata->widget);
    XSync(awt_display, False);
    awtJNI_setMbAndWwHeightAndOffsets(env, this, wdata);
    
#ifdef _pauly_debug
    XtVaGetValues(mdata->widget, XmNheight, &mbHeight, NULL);
    fprintf(stdout,"    ...pSetMenuBar. with menu bar: menu bar height: %d, top offset: %d, bottom offset: %d\n", mbHeight, wdata->top, wdata->bottom);
    fflush(stdout);
#endif /* _pauly_debug */
    
    (*env)->SetBooleanField(env, target, frameIDs.mbManagement, JNI_FALSE);
    (*env)->DeleteLocalRef(env, target);
    
    AWT_FLUSH_UNLOCK();
    
#ifdef _pauly_debug
    fprintf(stdout,"    ...pSetMenuBar. Done\n");
    fflush(stdout);
#endif /* _pauly_debug */
    
}

/*
 * Class:     sun_awt_motif_MFramePeer
 * Method:    toBack
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_toBack
(JNIEnv *env, jobject this)
{
    struct FrameData *wdata;
    
    AWT_LOCK();
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (wdata == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    if (XtWindow(wdata->winData.shell) != 0) {
	XLowerWindow(awt_display, XtWindow(wdata->winData.shell));
    }
    AWT_FLUSH_UNLOCK();
}

JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_addTextComponentNative
(JNIEnv *env, jobject this)
{
    struct FrameData            *wdata;
    jobject                     target;
    
    if (JNU_IsNull(env, this)) {
	return;
    }
    
    AWT_LOCK();
    
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (wdata == 0 ||
	wdata->winData.comp.widget==0 ||
	wdata->winData.shell==0 ||
	JNU_IsNull(env, target)) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    if ( !wdata->hasTextComponentNative) {
	wdata->hasTextComponentNative = True;
	wdata->imHeight = awt_util_getIMStatusHeight(wdata->winData.shell); 
	wdata->bottom += wdata->imHeight;
	
	awtJNI_ChangeInsets(env, this, wdata);
	reshape(env, this, wdata,
		(*env)->GetIntField(env, target, componentIDs.x),
		(*env)->GetIntField(env, target, componentIDs.y),
		(*env)->GetIntField(env, target, componentIDs.width),
		(*env)->GetIntField(env, target, componentIDs.height));
    } 
    AWT_UNLOCK();
}

JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_removeTextComponentNative
(JNIEnv *env, jobject this)
{
    struct FrameData            *wdata;
    jobject                     target;
    
    if (JNU_IsNull(env, this)) {
	return;
    }
    
    AWT_LOCK();
    
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (wdata == 0 ||
	wdata->winData.comp.widget==0 ||
	wdata->winData.shell==0 ||
	JNU_IsNull(env, target)) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    if (!wdata->hasTextComponentNative) {
	AWT_UNLOCK();
	return;
    }
    
    wdata->bottom -= wdata->imHeight;
    awtJNI_ChangeInsets(env, this, wdata);
    wdata->imRemove = True;
    reshape(env, this, wdata,
	    (*env)->GetIntField(env, target, componentIDs.x),
	    (*env)->GetIntField(env, target, componentIDs.y),
	    (*env)->GetIntField(env, target, componentIDs.width),
	    (*env)->GetIntField(env, target, componentIDs.height));
    
    wdata->hasTextComponentNative = False;
    wdata->imHeight = 0;
    
    AWT_UNLOCK();
} /* ...removeTextComponentPeer() */

JNIEXPORT jboolean JNICALL Java_sun_awt_motif_MFramePeer_hasTextComponents
(JNIEnv *env, jobject this)
{
    struct FrameData            *wdata;
    jobject                     target;
    
    if (JNU_IsNull(env, this)){
	return JNI_FALSE;
    }
    
    AWT_LOCK();
    
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    
    if (wdata == 0 ||
	wdata->winData.comp.widget==0 ||
	wdata->winData.shell==0 ||
	JNU_IsNull(env, target)) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return JNI_FALSE;
    }
    if (wdata->hasTextComponentNative) {
	AWT_UNLOCK();
	return JNI_TRUE;
    }
    AWT_UNLOCK();
    return JNI_FALSE;
}


static Atom java_protocol = NULL;
static Atom motif_wm_msgs = NULL;
extern char* awt_util_makeWMMenuItem(char *target, Atom protocol);

static void im_callback(Widget shell, XtPointer client_data, XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    JNU_CallMethodByName(env, NULL,
			 (jobject)client_data,
			 "notifyIMMOptionChange",
			 "()V");
}

JNIEXPORT void JNICALL Java_sun_awt_motif_MFramePeer_pSetIMMOption
(JNIEnv *env, jobject this, jstring option)
{
    char        *coption;
    char        *empty = "InputMethod";
    char        *menuItem;
    jobject	globalRef = awtJNI_CreateAndSetGlobalRef(env, this);
    struct FrameData *wdata;
    
    AWT_LOCK();
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.shell == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    
    coption = (JNU_IsNull(env, option)) ? empty : (char *) JNU_GetStringPlatformChars(env, option, NULL);
    if (java_protocol == NULL || motif_wm_msgs == NULL){
	java_protocol = XmInternAtom(awt_display, "_JAVA_IM_MSG", False);
	motif_wm_msgs = XmInternAtom(awt_display, "_MOTIF_WM_MESSAGES", False);
    }
    XmAddProtocols (wdata->winData.shell, motif_wm_msgs, &java_protocol, 1);
    XmAddProtocolCallback(wdata->winData.shell, motif_wm_msgs, java_protocol, im_callback, (XtPointer)globalRef);
    
    if (menuItem = awt_util_makeWMMenuItem(coption, java_protocol)){
	XtVaSetValues(wdata->winData.shell, 
		      XmNmwmMenu, 
		      menuItem,
		      NULL); 
	free(menuItem); 
    }
    if (coption != empty)
	JNU_ReleaseStringPlatformChars(env, option, (const char *) coption);
    AWT_FLUSH_UNLOCK();
}


JNIEXPORT void JNICALL Java_sun_awt_motif_MEmbeddedFramePeer_NEFcreate(JNIEnv *env, jobject this, jobject parent, jobject arg, jint ladHandle)
{
    
    Arg      args[40];
    int      argc;
    struct   FrameData *wdata;
    int      isEmbedded;
    jobject  target;
    jstring  warningString;
    jboolean resizable;
    jclass   clazz;
    jobject  globalRef = awtJNI_CreateAndSetGlobalRef(env, this);
    unsigned runningWM;  /* the running Window Manager */
    Widget   innerCanvasW;  /* form's child, parent of the outer canvas
			       drawing area */
    
    
    
    AWT_LOCK();
    
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    
    if (JNU_IsNull(env, target)) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    (*env)->SetObjectField(env, this, mFramePeerIDs.insets, arg);
    
    wdata = ZALLOC(FrameData);
    (*env)->SetLongField(env, this ,mComponentPeerIDs.pData ,(jlong) wdata);
    if (wdata == 0) {
	JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
	AWT_UNLOCK();
	return;
    }
    /* A variation on Netscape's hack for embedded frames: the client area
     * of the browser is a Java Frame for parenting purposes, but really a
     * Motif child window
     */
    clazz = (*env)->FindClass(env, "sun/awt/EmbeddedFrame");
    isEmbedded = (*env)->IsInstanceOf(env, target, clazz);
    
    if (isEmbedded) {
	wdata->winData.flags |= W_IS_EMBEDDED;
	
	(*env)->SetIntField(env, arg, insetsIDs.top, 0);
	(*env)->SetIntField(env, arg, insetsIDs.left, 0);
	(*env)->SetIntField(env, arg, insetsIDs.bottom, 0);
	(*env)->SetIntField(env, arg, insetsIDs.right, 0);
	
	wdata->top = 0;
	wdata->left = 0;
	wdata->bottom = 0;
	wdata->right = 0;
    }
    
    
    wdata->isModal = 0;
    wdata->mappedOnce = 0;
    wdata->isShowing = False;
    wdata->mappedOnce = 0;
    wdata->isShowing = False;
    wdata->shellResized = False;
    wdata->canvasResized = False;
    wdata->menuBarReset = False;
    
    
    resizable = (*env)->GetBooleanField(env, target, frameIDs.resizable);
    
    wdata->winData.shell = (Widget)ladHandle;
    
    setDeleteCallback(globalRef, wdata);
    /* Establish resizability.  For the case of not resizable, do not
       yet set a fixed size here; we must wait until in the routine
       sun_awt_motif_MFramePeer_pReshape() after insets have been fixed.
       This is because correction of the insets may affect shell size.
       (See comments in shellEH() concerning correction of the insets.  */
    wdata->isResizable = resizable;
    wdata->isFixedSizeSet = False;
    if  (resizable) {
	awt_util_setShellResizable(wdata->winData.shell, wdata->isShowing);
    }
    
    XtAddEventHandler(wdata->winData.shell, StructureNotifyMask | FocusChangeMask,
		      FALSE, shellEH, globalRef);
    
    
    argc = 0;
    XtSetArg(args[argc], XmNmarginWidth, 0); argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0); argc++;
    XtSetArg(args[argc], XmNhorizontalSpacing, 0); argc++;
    XtSetArg(args[argc], XmNverticalSpacing, 0); argc++;
    
    
    XtSetArg(args[argc], XmNresizePolicy, XmRESIZE_NONE); argc++;
    wdata->mainWindow = XmCreateForm(wdata->winData.shell, "main", args, argc);
    
    /* The widget returned by awt_canvas_create is a drawing area
       (i.e., canvas) which is the child of another drawing area
       parent widget.  The parent is the drawing area within the
       form just created.  The child is an drawing area layer over
       the entire frame window, including the form, any menu bar
       and warning windows present, and also window manager stuff.
       The top, bottom, left, and right fields in wdata maintain
       the respective offsets between these two drawing areas.  */
    
    wdata->winData.comp.widget = awt_canvas_create((XtPointer)globalRef,
                                                   wdata->mainWindow,
                                                   "frame_",
                                                   -1,
                                                   -1,
                                                   True,
                                                   wdata,
                                                   defaultConfig);
    
    XtAddCallback(wdata->winData.comp.widget,
		  XmNresizeCallback,
		  outerCanvasResizeCB,
		  globalRef);
    
    
    innerCanvasW = XtParent(wdata->winData.comp.widget);
    XtVaSetValues(innerCanvasW,
		  XmNleftAttachment, XmATTACH_FORM,
		  XmNrightAttachment, XmATTACH_FORM,
		  NULL);
    
    
    XtAddEventHandler(innerCanvasW, StructureNotifyMask, FALSE,
		      innerCanvasEH, globalRef);
    
    /* No menu bar initially */
    wdata->menuBar = NULL;
    wdata->mbHeight = 0;
    
    /* If a warning window (string) is needed, establish it now.*/
    warningString =
	(*env)->GetObjectField(env, target, windowIDs.warningString);
    
    /* No warning window present */
    XtVaSetValues(innerCanvasW,
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNbottomAttachment, XmATTACH_FORM,
		  NULL);
    wdata->warningWindow = NULL;
    wdata->wwHeight = 0;
    
    
    awt_util_show(wdata->winData.comp.widget);
    
    AWT_FLUSH_UNLOCK();    
}  /* MEmbeddedFramePeer_NEFcreate() */


JNIEXPORT void JNICALL Java_sun_awt_motif_MEmbeddedFramePeer_pShow
(JNIEnv *env, jobject this)
{
    struct FrameData *wdata;
    
    AWT_LOCK();
    
    wdata = (struct FrameData *)
	(*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (wdata == 0 || wdata->winData.comp.widget == 0 || wdata->winData.shell == 0 ||
	wdata->mainWindow == 0) {
	JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
	return;
    }
    XtVaSetValues(wdata->winData.comp.widget,
		  XmNx, -(wdata->left),
		  XmNy, -(wdata->top), NULL);
    
    if (wdata->menuBar != 0) {
	awt_util_show(wdata->menuBar);
    }
    
    XtManageChild(wdata->mainWindow);
    if (XtWindow(wdata->winData.shell) == 0) {
	XtRealizeWidget(wdata->winData.shell);
    }
    if (wdata->winData.flags & W_IS_EMBEDDED) {
	
	jobject embeddedFrame;
	jobject embeddedPeer;
	Widget embeddedWidget;
	
	
	embeddedFrame =
	    (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
	embeddedPeer =
	    (*env)->GetObjectField(env, embeddedFrame, componentIDs.peer);
	
	embeddedWidget = (Widget)
	    (*env)->GetLongField(env, embeddedPeer, mEmbeddedFramePeerIDs.handle);
	
	if (embeddedWidget != 0) {
	    
	    XReparentWindow(XtDisplay(wdata->winData.shell),
			    XtWindow(wdata->winData.shell),
			    XtWindow(embeddedWidget), 0, 0);
	}
	
    }
    XtManageChild(wdata->winData.comp.widget);
    XtSetMappedWhenManaged(wdata->winData.shell, True);
    XtPopup(wdata->winData.shell, XtGrabNone);
    XRaiseWindow(awt_display, XtWindow(wdata->winData.shell));
    wdata->isShowing = True;
    
    AWT_FLUSH_UNLOCK();
    
}
