/*
 * @(#)awt_GraphicsEnv.c	1.19 98/09/28
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_p.h"
#include "awt.h"
#include "color.h"
#include <sun_awt_X11GraphicsEnvironment.h>
#include <sun_awt_X11GraphicsDevice.h>
#include <sun_awt_X11GraphicsConfig.h>

#include <jni.h>
#include <jni_util.h>
#include <jvm.h>

#include <stdlib.h>

#include "awt_GraphicsEnv.h"

#ifdef NETSCAPE
#include <signal.h>
extern int awt_init_xt;
#endif

Display *awt_display;
Window awt_root;
long awt_screen;
int awt_whitepixel;
int awt_blackpixel;

AwtData **graphicsConfigs;
AwtData *defaultConfig;
int totalConfigs;

struct X11GraphicsConfigIDs x11GraphicsConfigIDs;

JNIEXPORT void JNICALL
Java_sun_awt_X11GraphicsConfig_initIDs (JNIEnv *env, jclass cls)
{
    x11GraphicsConfigIDs.aData = (*env)->GetFieldID (env, cls, "aData", "J");
} 

/*
 * error handlers
 */

int
xerror_handler(Display * disp, XErrorEvent * err)
{
#ifdef DEBUG
    char msg[128];
    char buf[128];
    char *ev = getenv("NOISY_AWT");

    if (!ev || !ev[0])
        return 0;
    XGetErrorText(disp, err->error_code, msg, sizeof(msg));
    jio_fprintf(stderr, "Xerror %s\n", msg);
    jio_snprintf(buf, sizeof(buf), "%d", err->request_code);
    XGetErrorDatabaseText(disp, "XRequest", buf, "Unknown", msg, sizeof(msg));
    jio_fprintf(stderr, "Major opcode %d (%s)\n", err->request_code, msg);
    if (err->request_code > 128) {
        jio_fprintf(stderr, "Minor opcode %d\n", err->minor_code);
    }
    if (awt_lock) {
        /*SignalError(lockedee->lastpc, lockedee, "fp/ade/gui/GUIException", msg); */
    }
    if (strcasecmp(ev, "abort") == 0) {
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

        (*env)->FatalError(env, "xerror_handler abort");
    }
#endif
    return 0;
}

static int 
xioerror_handler(Display * disp)
{
    if (awt_lock) {
        /*SignalError(lockedee->lastpc, lockedee, "fp/ade/gui/GUIException", "I/O error"); */
    }
    return 0;
}

Display *awt_init_Display(JNIEnv *env, jobject this)
{
    Display *dpy;
    XVisualInfo viTmp, *pVI;
    int numvi;
    XColor color;
    char errmsg[128];
    Visual *awt_visual;
    int awt_depth;
    XVisualInfo awt_visInfo;
    Colormap awt_cmap;
    AwtData aData;
    jclass fontWrapper;
#ifdef NETSCAPE
    sigset_t alarm_set, oldset;
#endif

    if (awt_display) {
        return awt_display;
    }

#ifdef NETSCAPE
    /* Disable interrupts during XtOpenDisplay to avoid bugs in unix os select
       code: some unix systems don't implement SA_RESTART properly and
       because of this, select returns with EINTR. Most implementations of
       gethostbyname don't cope with EINTR properly and as a result we get
       stuck (forever) in the gethostbyname code
    */
    sigemptyset(&alarm_set);
    sigaddset(&alarm_set, SIGALRM);
    sigprocmask(SIG_BLOCK, &alarm_set, &oldset);
#endif
    fontWrapper = (*env)->FindClass (env, "sun/awt/font/NativeFontWrapper");
    if (fontWrapper != NULL)
	awt_lock = (*env)->NewGlobalRef (env, fontWrapper);
    else
        awt_lock = (*env)->NewGlobalRef(env, this);
    dpy = awt_display = XOpenDisplay(NULL);
#ifdef NETSCAPE
    sigprocmask(SIG_SETMASK, &oldset, 0);
#endif
    if (!dpy) {
	jio_snprintf(errmsg,
		     sizeof(errmsg),
		     "Can't connect to X11 window server using '%s' as the value of the DISPLAY variable.",
		     (getenv("DISPLAY") == 0) ? ":0.0" : getenv("DISPLAY"));
        JNU_ThrowInternalError(env, errmsg);
        return NULL;
    }

    XSetErrorHandler(xerror_handler);
    XSetIOErrorHandler(xioerror_handler);

    awt_screen = DefaultScreen(awt_display);
    awt_root = RootWindow(awt_display, awt_screen);
    if (!getenv("FORCEDEFVIS") && XMatchVisualInfo(dpy, awt_screen,
                                                   24, TrueColor,
                                                   &awt_visInfo)) {
        awt_visual = awt_visInfo.visual;
        awt_depth = awt_visInfo.depth;
        if (awt_visual == DefaultVisual(awt_display, awt_screen)) {
            awt_cmap = DefaultColormap(awt_display, awt_screen);
        } else {
            awt_cmap = XCreateColormap(dpy, awt_root, awt_visual,
                                       AllocNone);
        }
        color.flags = DoRed | DoGreen | DoBlue;
        color.red = color.green = color.blue = 0x0000;
        XAllocColor(dpy, awt_cmap, &color);
        awt_blackpixel = color.pixel;
        color.flags = DoRed | DoGreen | DoBlue;
        color.red = color.green = color.blue = 0xffff;
        XAllocColor(dpy, awt_cmap, &color);
        awt_whitepixel = color.pixel;
    } else {
        awt_visual = DefaultVisual(awt_display, awt_screen);
        awt_depth = DefaultDepth(awt_display, awt_screen);
        awt_cmap = DefaultColormap(awt_display, awt_screen);
        viTmp.visualid = XVisualIDFromVisual(awt_visual);
        viTmp.depth = awt_depth;
        pVI = XGetVisualInfo(dpy, VisualIDMask | VisualDepthMask,
                             &viTmp, &numvi);
        if (!pVI) {
            /* This should never happen. */
            JNU_ThrowInternalError(env, "Can't find default visual information");
	    XCloseDisplay(dpy);
	    awt_display = NULL;
            return NULL;
        }
        awt_visInfo = *pVI;
        XFree(pVI);
        awt_blackpixel = BlackPixel(awt_display, awt_screen);
        awt_whitepixel = WhitePixel(awt_display, awt_screen);
    }

    defaultConfig = ZALLOC (_AwtData);
    defaultConfig->primitiveType = -255;
    defaultConfig->awt_depth = awt_depth;
    memcpy (&defaultConfig->awt_visInfo, &awt_visInfo, sizeof (XVisualInfo));
    defaultConfig->awt_cmap = awt_cmap;

    if (!awt_allocate_colors(defaultConfig)) {
        static char *visnames[] = {"StaticGray", "GrayScale", "StaticColor",
				   "PseudoColor", "TrueColor", "DirectColor"};
	jio_snprintf(errmsg,
		     sizeof(errmsg),
		     "Display type (%s) and depth (%d) not supported.",
		     visnames[awt_visInfo.class], awt_depth);
        JNU_ThrowInternalError(env, errmsg);
	XCloseDisplay(dpy);
	awt_display = NULL;
        return NULL;
    }

    totalConfigs = 0;
    return dpy;
}

static void 
getAllConfigs (JNIEnv *env) {
    int i, n8p, n8s, n8gs, n8sg, n1sg, nTrue, nConfig;
    XVisualInfo *pVI8p, *pVI8s, *pVITrue, *pVI8gs, *pVI8sg, *pVI1sg, viTmp;

    AWT_LOCK ();

    viTmp.depth = 8;
    viTmp.class = PseudoColor;
    viTmp.colormap_size = 256; 
    pVI8p = XGetVisualInfo (awt_display, 
	   	VisualDepthMask | VisualClassMask | VisualColormapSizeMask, 
                            &viTmp, &n8p);
    totalConfigs = n8p;

    viTmp.class = TrueColor;
    pVITrue = XGetVisualInfo (awt_display,  VisualClassMask, &viTmp, &nTrue);
    totalConfigs += nTrue;

    viTmp.depth = 8;
    viTmp.class = StaticColor;
    pVI8s = XGetVisualInfo (awt_display, VisualDepthMask | VisualClassMask,
                            &viTmp, &n8s);
    totalConfigs += n8s;

    viTmp.depth = 8;
    viTmp.class = GrayScale;
    viTmp.colormap_size = 256;
    pVI8gs = XGetVisualInfo (awt_display, 
	   	VisualDepthMask | VisualClassMask | VisualColormapSizeMask,
                             &viTmp, &n8gs);
    totalConfigs += n8gs;

    viTmp.depth = 8;
    viTmp.class = StaticGray;
    viTmp.colormap_size = 256;
    pVI8sg = XGetVisualInfo (awt_display, 
	   	VisualDepthMask | VisualClassMask | VisualColormapSizeMask,
                             &viTmp, &n8sg);
    totalConfigs += n8sg;

    /* TEMPORARY.. remove when support 1-bit StaticGray */
    n1sg = 0;
/*
    viTmp.depth = 1;
    viTmp.class = StaticGray;
    pVI1sg = XGetVisualInfo (awt_display, VisualDepthMask | VisualClassMask,
                             viTmp, &n1sg);
    totalConfigs += n1sg;
*/

    graphicsConfigs = (AwtData **) calloc (totalConfigs, sizeof (AwtData *));
    graphicsConfigs [0] = defaultConfig;
    nConfig = 1;

    for (i = 0; i < nTrue; i++) {
        if (pVITrue [i].visualid == defaultConfig->awt_visInfo.visualid)
           continue;
        graphicsConfigs [nConfig] = ZALLOC (_AwtData);
	graphicsConfigs [nConfig]->primitiveType = -255;
        graphicsConfigs [nConfig]->awt_depth = pVITrue [i].depth;
        memcpy (&graphicsConfigs [nConfig]->awt_visInfo, &pVITrue [i],
                sizeof (XVisualInfo));
        nConfig++;
    }
 
    for (i = 0; i < n8p; i++) {
        if (pVI8p [i].visualid == defaultConfig->awt_visInfo.visualid)
           continue;
        graphicsConfigs [nConfig] = ZALLOC (_AwtData);
	graphicsConfigs [nConfig]->primitiveType = -255;
        graphicsConfigs [nConfig]->awt_depth = pVI8p [i].depth;
        memcpy (&graphicsConfigs [nConfig]->awt_visInfo, &pVI8p [i],
                sizeof (XVisualInfo));
        nConfig++;
    }
       
    for (i = 0; i < n8s; i++) {
        if (pVI8s [i].visualid == defaultConfig->awt_visInfo.visualid)
           continue;
        graphicsConfigs [nConfig] = ZALLOC (_AwtData);
	graphicsConfigs [nConfig]->primitiveType = -255;
        graphicsConfigs [nConfig]->awt_depth = pVI8s [i].depth;
        memcpy (&graphicsConfigs [nConfig]->awt_visInfo, &pVI8s [i],
                sizeof (XVisualInfo));
        nConfig++;
    }

    for (i = 0; i < n8gs; i++) {
        if (pVI8gs [i].visualid == defaultConfig->awt_visInfo.visualid)
           continue;
        graphicsConfigs [nConfig] = ZALLOC (_AwtData);
	graphicsConfigs [nConfig]->primitiveType = -255;
        graphicsConfigs [nConfig]->awt_depth = pVI8gs [i].depth;
        memcpy (&graphicsConfigs [nConfig]->awt_visInfo, &pVI8gs [i],
                sizeof (XVisualInfo));
        nConfig++;
    }

    for (i = 0; i < n8sg; i++) {
        if (pVI8sg [i].visualid == defaultConfig->awt_visInfo.visualid)
           continue;
        graphicsConfigs [nConfig] = ZALLOC (_AwtData);
	graphicsConfigs [nConfig]->primitiveType = -255;
        graphicsConfigs [nConfig]->awt_depth = pVI8sg [i].depth;
        memcpy (&graphicsConfigs [nConfig]->awt_visInfo, &pVI8sg [i],
                sizeof (XVisualInfo));
        nConfig++;
    }

    for (i = 0; i < n1sg; i++) {
        if (pVI1sg [i].visualid == defaultConfig->awt_visInfo.visualid)
           continue;
        graphicsConfigs [nConfig] = ZALLOC (_AwtData);
	graphicsConfigs [nConfig]->primitiveType = -255;
        graphicsConfigs [nConfig]->awt_depth = pVI1sg [i].depth;
        memcpy (&graphicsConfigs [nConfig]->awt_visInfo, &pVI1sg [i],
                sizeof (XVisualInfo));
        nConfig++;
    }

    if (n8p != 0)
       XFree (pVI8p);
    if (n8s != 0)
       XFree (pVI8s);
    if (n8gs != 0)
       XFree (pVI8gs);
    if (n8sg != 0)
       XFree (pVI8sg);
    if (n1sg != 0)
       XFree (pVI1sg);

    AWT_UNLOCK ();
}


/*
 * Class:     sun_awt_X11GraphicsEnvironment
 * Method:    initDisplay
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_X11GraphicsEnvironment_initDisplay
    (JNIEnv *env, jobject this)
{
    (void) awt_init_Display(env, this);
}

/*
 * Class:     sun_awt_X11GraphicsEnvironment
 * Method:    getNumScreens
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_X11GraphicsEnvironment_getNumScreens
    (JNIEnv *env, jobject this)
{
    return 1;
}

/*
 * Class:     sun_awt_X11GraphicsDevice
 * Method:    getDisplay
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_X11GraphicsDevice_getDisplay
    (JNIEnv *env, jobject this)
{
    return (jint) awt_display;
}

/*
 * Class:     sun_awt_X11GraphicsDevice
 * Method:    getNumConfigs
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_X11GraphicsDevice_getNumConfigs
    (JNIEnv *env, jobject this)
{
    if (totalConfigs == 0)
	getAllConfigs (env);

    return totalConfigs;
}

/*
 * Class:     sun_awt_X11GraphicsDevice
 * Method:    getConfigVisualId
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_X11GraphicsDevice_getConfigVisualId
    (JNIEnv *env, jobject this, jint index)
{
    int visNum;
 
    /* Do this in the case where this function is called upon startup,
     * before the graphicsConfigs have been allocated */
    if (index == 0)
        return defaultConfig->awt_visInfo.visualid;
 
    return (graphicsConfigs [index]->awt_visInfo.visualid);
}

/*
 * Class:     sun_awt_X11GraphicsConfig
 * Method:    getXResolution
 * Signature: ()I
 */
JNIEXPORT jdouble JNICALL
Java_sun_awt_X11GraphicsConfig_getXResolution
    (JNIEnv *env, jobject this)
{
    return ((DisplayWidth(awt_display, awt_screen) * 25.4) /
	    DisplayWidthMM(awt_display, awt_screen));
}

/*
 * Class:     sun_awt_X11GraphicsConfig
 * Method:    getYResolution
 * Signature: ()I
 */
JNIEXPORT jdouble JNICALL
Java_sun_awt_X11GraphicsConfig_getYResolution
    (JNIEnv *env, jobject this)
{
    return ((DisplayHeight(awt_display, awt_screen) * 25.4) /
	    DisplayHeightMM(awt_display, awt_screen));
}


/*
 * Class:     sun_awt_X11GraphicsConfig
 * Method:    getNumColors
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_X11GraphicsConfig_getNumColors
    (JNIEnv *env, jobject this)
{
    AwtData *adata;
 
    adata = (AwtData *) (*env)->GetLongField (env, this, 
					x11GraphicsConfigIDs.aData);

    return adata->awt_num_colors;
}

/*
 * Class:     sun_awt_X11GraphicsConfig
 * Method:    init
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_X11GraphicsConfig_init 
    (JNIEnv *env, jobject this, jint visualNum)
{
    AwtData *adata = NULL;
    int i, n;


    /* If this is being called upon startup, we haven't created the awt
     * lock object, so don't try to call AWT_LOCK */
    if ((visualNum == 0) || 
	(visualNum == defaultConfig->awt_visInfo.visualid)) {
	(*env)->SetLongField (env, this, x11GraphicsConfigIDs.aData, 
			      (jlong) defaultConfig);
	return; 
    }
	
    /* If haven't gotten all of the configs yet, do it now. */
    if (totalConfigs == 0)
	getAllConfigs (env);

    /* Check the graphicsConfig for this visual */
    for (i = 0; i < totalConfigs; i++) {
	if (graphicsConfigs [i]->awt_visInfo.visualid == visualNum) {
	   adata = graphicsConfigs [i];
	   break;
	}
    }

    /* If didn't find the visual, throw an exception... */
    if (adata == (AwtData *) NULL) {
        JNU_ThrowIllegalArgumentException(env, "Unknown Visual Specified");
	return;
    }

    (*env)->SetLongField (env, this, x11GraphicsConfigIDs.aData, 
			  (jlong) adata);

}
    

    
/*
 * Class:     sun_awt_X11GraphicsConfig
 * Method:    makeColorModel
 * Signature: ()Ljava/awt/image/ColorModel
 */
JNIEXPORT jobject JNICALL
Java_sun_awt_X11GraphicsConfig_makeColorModel 
    (JNIEnv *env, jobject this)
{
    AwtData *adata;
    jobject colorModel;
    int lock = 1;

    /*
     * If awt is not locked yet, return null since the toolkit is not
     * initialized yet.
     */
    if (awt_lock == 0) {
        return NULL;
    }
    
    AWT_LOCK (); 

    adata = (AwtData *) (*env)->GetLongField (env, this, 
					x11GraphicsConfigIDs.aData);

    if (adata == defaultConfig)
	lock = 0;

    /* If colormap entry of adata is NULL, need to create it now */
    if (adata->awt_cmap == (Colormap) NULL) 
	awtJNI_CreateColorData (env, adata, lock);

    /* Make Color Model object for this GraphicsConfiguration */
    colorModel = awtJNI_GetColorModel (env, adata);

    AWT_UNLOCK (); 

    return colorModel;
}

