/*
 * @(#)awt_MToolkit.c	1.120 98/09/23
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "color.h"
#include "canvas.h"
#include <limits.h>
#include <sys/time.h>
#include "java_awt_AWTEvent.h"
#include "java_awt_SystemColor.h"
#include "sun_awt_motif_MToolkit.h"
#include "sun_awt_motif_InputThread.h"
#include <X11/cursorfont.h>

#include <locale.h>

#include <jni.h>
#include <jni_util.h>
#include <jvm.h>
#include <jlong.h>

/* Initialize the Java VM instance variable when the library is 
   first loaded */
JavaVM *jvm;

JNIEXPORT jint JNICALL
JNI_OnLoad(JavaVM *vm, void *reserved)
{
    jvm = vm;
    return JNI_VERSION_1_2;
}

static jlong awtJNI_TimeMillis(void);
extern void awt_initialize_Xm_DnD(Display*);
 
extern Display *awt_init_Display(JNIEnv *env);
extern jobject currentX11InputMethodInstance;

static void awt_pipe_init(void);
static void processOneEvent(int iMask);
static void waitForEvents(JNIEnv *env, int fdXPipe, int fdAWTPipe);
#ifdef USE_SELECT
static void performSelect(JNIEnv *env, int fdXPipe, int fdAWTPipe);
#else
static void performPoll(JNIEnv *env,int fdXPipe, int fdAWTPipe);
#endif


#include <X11/Intrinsic.h>
#include <dlfcn.h>
#include <fcntl.h>

#ifdef USE_SELECT
#if defined(AIX)
#include <sys/select.h>
#endif
#else
#include <poll.h>
#endif

#ifdef NDEBUG
#undef DEBUG		/* NDEBUG overrides DEBUG */
#endif
#ifdef DEBUG
static Boolean debugEventLoop = False;
#endif

static struct WidgetInfo *awt_winfo = (struct WidgetInfo *) 0;


#define bzero(a,b) memset(a, 0, b)

jobject awt_lock;
XtAppContext awt_appContext;
Pixel awt_defaultBg;
Pixel awt_defaultFg;
int awt_multiclick_time;        /* milliseconds */
int awt_MetaMask = 0;
int awt_AltMask = 0;
int awt_NumLockMask = 0;
Cursor awt_scrollCursor;

/* implement a "putback queue" -- see comments on awt_put_back_event() */
#define PUTBACK_QUEUE_MIN_INCREMENT 5	/* min size increase */
static XEvent *putbackQueue = NULL;	/* the queue -- next event is 0 */
static int putbackQueueCount = 0;	/* # of events available on queue */
static int putbackQueueCapacity = 0;	/* total capacity of queue */
static XtInputMask awt_events_pending(XtAppContext appContext);
static int awt_get_next_put_back_event(XEvent *xev_out);
 

#define AWT_FLUSH_TIMEOUT    ((unsigned long)100) /* milliseconds */
#define AWT_MIN_POLL_TIMEOUT ((unsigned long)0) /* milliseconds */
#define AWT_MAX_POLL_TIMEOUT ((unsigned long)250) /* milliseconds */

#define AWT_POLL_BUFSIZE        100
#define AWT_READPIPE            (awt_pipe_fds[0])
#define AWT_WRITEPIPE           (awt_pipe_fds[1])
#define AWT_FLUSHOUTPUT_NOW()                    \
{                                                \
    XFlush(awt_display);                         \
    awt_next_flush_time = 0;                     \
}

typedef XtIntervalId (*XTFUNC)();

static jobject	awt_MainThread = NULL;
static char     read_buf[AWT_POLL_BUFSIZE + 1];    /* dummy buf to empty pipe */
static int      awt_pipe_fds[2];                   /* fds for wkaeup pipe */
static Boolean	awt_pipe_inited = False;           /* make sure pipe is initialized befor
e write */
static int      def_poll_timeout = AWT_MAX_POLL_TIMEOUT;   /* default value for timeout */
static jlong awt_next_flush_time = 0;	/* 0 == no scheduled flush */
static int      dbg_cnt = 0;
static void     *xt_lib = NULL;
static XTFUNC   xt_timeout = NULL;

#ifdef DEBUG_AWT_LOCK

int awt_locked = 0;
char *lastF = "";
int lastL = -1;

#endif

#ifndef NOMODALFIX
extern Boolean awt_isModal();
extern Boolean awt_isWidgetModal(Widget w);
#endif NOMODALFIX

static unsigned long curPollTimeout = AWT_MAX_POLL_TIMEOUT;

static void 
awt_set_poll_timeout (unsigned long newTimeout)
{
    /*
    #ifdef DEBUG
    if (debugEventLoop) {
	jio_fprintf(stderr,"awt_set_poll_timeout(%lu)\n", newTimeout);
    }
    #endif
    */

    newTimeout = max(AWT_MIN_POLL_TIMEOUT, newTimeout);
    newTimeout = min(AWT_MAX_POLL_TIMEOUT, newTimeout);
    newTimeout = min(newTimeout, curPollTimeout);
    curPollTimeout = newTimeout;

} /* awt_set_poll_timeout */

/*
 * Gets the best timeout for the next call to poll() or select().
 * If timedOut is True, we assume that our previous timeout elapsed
 * with no events/timers arriving. Therefore, we can increase the
 * next timeout slightly.
 */
static unsigned long
awt_get_poll_timeout( Boolean timedOut )
{
    unsigned long timeout = AWT_MAX_POLL_TIMEOUT;

    /*
    #ifdef DEBUG
    if (debugEventLoop) {
	jio_fprintf(stderr, "awt_get_poll_timeout(%s), awt_next_flush_time:%ld\n",
	    (remove?"true":"false"), 
	    awt_next_flush_time);
    }
    #endif
    */

    if (timedOut) {
	/* add 1/16 (plus 1, in case the division truncates to 0) */
	curPollTimeout += ((curPollTimeout>>4) + 1);
        curPollTimeout = min(AWT_MAX_POLL_TIMEOUT, curPollTimeout);
    }
    if (awt_next_flush_time > 0) {
	long flushDiff = (long)(awt_next_flush_time - awtJNI_TimeMillis());
	timeout = min(curPollTimeout, flushDiff);
    } else {
	timeout = curPollTimeout;
    }

    return timeout;
} /* awt_get_poll_timeout() */

/**
 * We override XtAppAddTimeOut, so that we can adapt our polling
 * timeout. Normally, we sit in a poll() [or select()], waiting
 * for events to come across the X pipe. Since timers do not come
 * across the X pipe, we need to timeout from the poll often to
 * check for the timers.
 */
/***** WE CAN'T SEEM TO GET THIS TO WORK EVERYWHERE, SO
 ***** WE ARE TEMPORARILY REMOVING IT
 *****/
#ifdef BOGUS
XtIntervalId 
XtAppAddTimeOut( XtAppContext appContext,
	 	 unsigned long interval,
	 	 XtTimerCallbackProc callbackProc,
	 	 XtPointer data
	       )
{
    unsigned char oneChar = 0;

    #ifdef DEBUG
    if (debugEventLoop) {
        jio_fprintf(stderr, 
		"XtTimeoutTrap incoming value is %d\n", interval);
    }
    #endif

    /* Set the timeout to 1/4 of timer value. This way, our maximum
     * error on timers should be ~25%
     */
    awt_set_poll_timeout(interval>>2);

    /* wake up the event loop, so that it'll see the timer */
    write ( AWT_WRITEPIPE, &oneChar, 1 );

    /* notify the actual XtAppAddTimeOut() */
    if (xt_timeout == NULL) {
	xt_lib = dlopen("libXt.so", RTLD_LAZY);
	assert(xt_lib != NULL);
	xt_timeout = (XTFUNC) dlsym(xt_lib, "XtAppAddTimeOut");
	assert(xt_timeout != NULL);
    }
    if (xt_timeout == NULL) {
        jio_fprintf(stderr, 
	    "FATAL AWT ERROR: cannot link function XtAppAddTimeOut()\n");
	exit(1);
    }
    return (*xt_timeout)(appContext, interval, callbackProc, data);
} /* XtAppAddTimeOut() */
#endif /* BOGUS */

static jlong 
awtJNI_TimeMillis(void)
{
    struct timeval t;

    gettimeofday(&t, 0);

    return jlong_add(jlong_mul(jint_to_jlong(t.tv_sec), jint_to_jlong(1000)),
		     jint_to_jlong(t.tv_usec / 1000));
}

static int 
xtError()
{
#ifdef DEBUG
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    jio_fprintf(stderr, "Xt error\n");
    JNU_ThrowNullPointerException(env, "NullPointerException");
#endif
    return 0;
}

static void 
set_mod_mask(int modnum, int *mask)
{
    switch (modnum) {
        case 1:
            *mask = Mod1Mask;
            break;
        case 2:
            *mask = Mod2Mask;
            break;
        case 3:
            *mask = Mod3Mask;
            break;
        case 4:
            *mask = Mod4Mask;
            break;
        case 5:
            *mask = Mod5Mask;
            break;
        default:
            *mask = 0;
    }
}

static void 
setup_modifier_map(Display * disp)
{
    XModifierKeymap *modmap;
    int i, k, nkeys;
    KeyCode keycode, metaL, metaR, altL, altR, numLock;

    metaL = XKeysymToKeycode(disp, XK_Meta_L);
    metaR = XKeysymToKeycode(disp, XK_Meta_R);
    altR = XKeysymToKeycode(disp, XK_Alt_R);
    altL = XKeysymToKeycode(disp, XK_Alt_L);
    numLock = XKeysymToKeycode(disp, XK_Num_Lock);

    modmap = XGetModifierMapping(disp);
    nkeys = modmap->max_keypermod;
    for (i = 3 * nkeys;
         i < (nkeys * 8) && (awt_MetaMask == 0 || awt_AltMask == 0 || awt_NumLockMask == 0);
         i++) {
        keycode = modmap->modifiermap[i];
        k = (i / nkeys) - 2;
        if (awt_MetaMask == 0 && (keycode == metaL || keycode == metaR)) {
            set_mod_mask(k, &awt_MetaMask);
        } else if (awt_AltMask == 0 && (keycode == altL || keycode == altR)) {
            set_mod_mask(k, &awt_AltMask);
        } else if (awt_NumLockMask == 0 && keycode == numLock) {
            set_mod_mask(k, &awt_NumLockMask);
        }
    }
    XFreeModifiermap(modmap);
}

Boolean scrollBugWorkAround;


void 
awt_output_flush()
{
    char c = 'p';

    if (awt_next_flush_time == 0)
    {
	Boolean needsWakeup = False;
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	if (awt_pipe_inited && (awt_get_poll_timeout(False) > (2*AWT_FLUSH_TIMEOUT))){
	    needsWakeup = True;
	}
	/* awt_next_flush_time affects awt_get_poll_timeout(), so set
	 * the variable *after* calling the function.
	 */
    	awt_next_flush_time = awtJNI_TimeMillis() + AWT_FLUSH_TIMEOUT;
	if (needsWakeup)
	{
	    /* write to the utility pipe to wake up the event
	     * loop, if it's sleeping
	     */
	    write ( AWT_WRITEPIPE, &c, 1 );
	}
    }
#ifdef FLUSHDEBUG
else
jio_fprintf(stderr, "!");
#endif
} /* awt_output_flush() */

static void 
null_event_handler(Widget w, XtPointer client_data,
                   XEvent * event, Boolean * cont)
{
    /* do nothing */
}

void 
awt_addWidget(Widget w, Widget origin, void *peer, long event_flags)
{

    if (!XtIsSubclass(w, xmFileSelectionBoxWidgetClass)) {
        struct WidgetInfo *nw = (struct WidgetInfo *) malloc(sizeof(struct WidgetInfo));

        if (nw) {
            nw->widget	   = w;
            nw->origin     = origin;
            nw->peer       = peer;
            nw->event_mask = event_flags;
            nw->next       = awt_winfo;
            awt_winfo      = nw;

            if (event_flags & java_awt_AWTEvent_MOUSE_EVENT_MASK) {
                XtAddEventHandler(w,
                                  ButtonPressMask | ButtonReleaseMask |
                                  EnterWindowMask | LeaveWindowMask,
                                  False, null_event_handler, NULL);
            }
            if (event_flags & java_awt_AWTEvent_MOUSE_MOTION_EVENT_MASK) {
                XtAddEventHandler(w,
                                  PointerMotionMask,
                                  False, null_event_handler, NULL);
            }
        } else {
            JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

            JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        }

    }
}

void 
awt_delWidget(Widget w)
{
    int p;
    struct WidgetInfo *cw;

    if (awt_winfo != NULL) {
        if (awt_winfo->widget == w) {
            cw = awt_winfo;
            awt_winfo = awt_winfo->next;
            free((void *) cw);
        } else {
            struct WidgetInfo *pw;

            for (pw = awt_winfo, cw = awt_winfo->next;
                 cw != NULL;
                 pw = cw, cw = cw->next) {
                if (cw->widget == w) {
                    pw->next = cw->next;
                    free((void *) cw);
                    break;
                }
            }
        }
    }
}

void 
awt_enableWidgetEvents(Widget widget, long event_flag)
{
    struct WidgetInfo *cw;

    for (cw = awt_winfo; cw != NULL; cw = cw->next) {
        if (cw->widget == widget) {
            cw->event_mask |= event_flag;
        }
    }
}

void 
awt_disableWidgetEvents(Widget widget, long event_flag)
{
    struct WidgetInfo *cw;

    for (cw = awt_winfo; cw != NULL; cw = cw->next) {
        if (cw->widget == widget) {
            cw->event_mask &= ~event_flag;
        }
    }
}

static struct WidgetInfo *
findWidgetInfo(Widget widget)
{
    struct WidgetInfo *cw;

    for (cw = awt_winfo; cw != NULL; cw = cw->next) {
        if (cw->widget == widget) {
            return cw;
        }
    }
    return NULL;
}

static void *
findPeer(Widget * pwidget)
{
    struct WidgetInfo	*cw;
    Widget widgetParent;
    void * peer;

    if ((cw = findWidgetInfo(*pwidget)) != NULL) {
        return cw->peer;
    }
    /* fix for 4053856, robi.khan@eng 
       couldn't find peer corresponding to widget
       but the widget may be child of one with
       a peer, so recurse up the hierarchy */
    widgetParent = XtParent(*pwidget);
    if (widgetParent != NULL ) {
	peer = findPeer(&widgetParent);
	if( peer != NULL ) {
	/* found peer attached to ancestor of given
	   widget, so set widget return value as well */
	    *pwidget = widgetParent;
	    return peer;
	}
    }

    return NULL;
}

static Boolean 
shouldDispatchToWidget(XEvent * xev)
{
    Window win;
    size_t xevent_size;
    Widget widget = 0;
    struct WidgetInfo *winfo;
    void *peer = NULL;
    Boolean cont = FALSE;


    /* Don't pre-post events which were putback by the EventDispatchThread. */
    if (xev->xany.send_event) {
        xev->xany.send_event = False;
        return False;
    }

    switch (xev->type) {
        case KeyPress:
        case KeyRelease:
            win = xev->xkey.window;
            break;
        case FocusIn:
        case FocusOut:
            win = xev->xfocus.window;
            break;
        case ButtonPress:
        case ButtonRelease:
            win = xev->xbutton.window;
            break;
        case MotionNotify:
            win = xev->xmotion.window;
            break;
        case EnterNotify:
        case LeaveNotify:
            win = xev->xcrossing.window;
            break;
        default:
            return False;
    }

    /*
     * Fix for bug 4145193
     *
     * If a popup menu is up, prevent awt components from
     * getting any events until the menu is popped down.
     * Before this fix, the fact that mouse/button events were 
     * preposted to the Java event queue was causing the ButtonRelease
     * (needed to pop menu down) to be seen by the menu's parent and 
     * not the menu.
     */
    if ((awt_util_focusIsOnMenu(awt_display)) == 1) {
	return (False);
    }


    if ((widget = XtWindowToWidget(awt_display, win)) == NULL) {
        return False;
    }

    if (xev->type == KeyPress || xev->type == KeyRelease) {
	Widget focusWidget = XmGetFocusWidget(widget);
	if (focusWidget != 0 && focusWidget != widget &&
	    (peer = findPeer(&focusWidget)) != NULL) {
             widget = focusWidget;
	}
    }

    if ((winfo = findWidgetInfo(widget)) == NULL) {
        return False;
    }

#ifndef NOMODALFIX
    /* 
     * if we have a modal dialog up, and the event is not
     * in the dialog hiearchy, we just discard the event 
     * without calling XtDispatchEvent().
     */
    if (awt_isModal() && !awt_isWidgetModal(widget)) {
        if (xev->type == ButtonRelease || xev->type == KeyRelease) {
             XBell(awt_display, 50);
        }
	return True;
    }
#endif NOMODALFIX

    /* If it's a keyboard event, we need to find the peer associated */
    /* with the widget that has the focus rather than the widget */
    /* associated with the window in the X event. */

    switch (xev->type) {
        case KeyPress:
        case KeyRelease:
            if (!(winfo->event_mask & java_awt_AWTEvent_KEY_EVENT_MASK))
                return False;
            break;
        case FocusIn:
        case FocusOut:
            if (!(winfo->event_mask & java_awt_AWTEvent_FOCUS_EVENT_MASK))
                return False;
            break;
        case ButtonPress:
        case ButtonRelease:
            if (!(winfo->event_mask & java_awt_AWTEvent_MOUSE_EVENT_MASK)) {
                return False;
            }
            break;
        case EnterNotify:
        case LeaveNotify:
            /* 
             * Do not post the enter/leave event if it's on a subwidget
             * within the component.
             */
            if (!(winfo->event_mask & java_awt_AWTEvent_MOUSE_EVENT_MASK) ||
                winfo->widget != winfo->origin)
                return False;
            break;
        case MotionNotify:
            if (!(winfo->event_mask & java_awt_AWTEvent_MOUSE_MOTION_EVENT_MASK))
                return False;
            break;
        default:
            return False;
    }

    peer = winfo->peer;

    /* If we found a widget and a suitable peer (either the focus
       peer above or the one associated with the widget then we
       dispatch to it. */
    if (peer != NULL) {

        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

	/*
          printf("pre-posting event: %d on %x (%s)...\n", xev->type, widget, 
          XtName(widget));
	 */
        
	awt_canvas_handleEvent(widget, peer, xev, winfo, &cont, TRUE);

        if (cont) {
            return False;
        }
        return True;
    } else {
        return False;
    }
} /* shouldDispatchToWidget() */


#ifdef DEBUG
static int debugPrintLineCount = 0;   /* limit debug output per line */
#endif

/*
 * This is the main Xt event loop for the AWT. 
 *
 * Because java applications are multithreaded, but X and Xt 
 * are thread-dumb, we must make special considerations to 
 * make ensure that the X/Xt libraries are not entered by
 * multiple threads simultaneously.
 *
 * The biggest difference between the standard Xt loop
 * and this loop is that we go to great lengths never to block
 * in the X libraries. We poll() on the X event pipe, waiting
 * for events, rather than simply calling XtAppNextEvent() and
 * blocking. If this thread were to block in XtAppNextEvent(),
 * no other thread could enter (e.g., to perform a paint or
 * retrieve data).
 */
static void 
awt_MToolkit_loop(JNIEnv *env)
{
    XtInputMask iMask;
    int  fdXPipe = -1;		/* pipe where X events arrive */
#ifdef DEBUG
    int  numEventsHandled = 0;
#endif

    /* only privileged thread should be running here */
    assert(awt_currentThreadIsPrivileged(env));

    /* The pipe where X events arrive */
    fdXPipe = ConnectionNumber(awt_display) ;

    /* We execute events while locked, unlocking only when waiting
     * for an event
     */
    AWT_LOCK();

    /* Create the AWT utility pipe. See the comments on awt_pipe_init() */
    awt_pipe_init();

    /*
     * Need to flush here in case data on the connection was read
     * before we acquired the monitor.
     *
     * I don't get this, but I'm too chicken to remove it. -jethro 2Sep98
     */
    AWT_FLUSHOUTPUT_NOW();

    /*
     * ACTUALLY PROCESS EVENTS
     */
    while(True) {

	/* process all events in the queue */
	#ifdef DEBUG
	    numEventsHandled = 0;
	#endif
        while (((iMask = awt_events_pending(awt_appContext)) & XtIMAll) > 0) {

	    #ifdef DEBUG
	        ++numEventsHandled;
	    #endif
	    processOneEvent(iMask);

        }  /* end while awt_events_pending() */
	/* At this point, we have exhausted the event queue */

	#ifdef DEBUG
	if (debugEventLoop) {
	    /* print the number of events handled in parens */
	    jio_fprintf(stderr, "(%d events)",(int)numEventsHandled);
	    if (++debugPrintLineCount > 8) {
	        jio_fprintf(stderr, "\n");
	        debugPrintLineCount = 0;
	    }
	}
	#endif

	/* Do some housecleaning */
        AWT_NOTIFY_ALL();		/* wake up modalWait() */
        awtJNI_CleanupGlobalRefs();

	/* Here, we wait for X events, outside of the X libs. When
	 * it's likely that an event is waiting, we process the queue
	 */
	waitForEvents(env, fdXPipe, AWT_READPIPE);

    } /* while(True) */

    /* If we ever exit the loop, must unlock the toolkit */

} /* awt_MToolkit_loop() */

/*
 * Creates the AWT utility pipe. This pipe exists solely so that
 * we can cause the main event thread to wake up from a poll() or
 * select() by writing to this pipe.
 */
static void
awt_pipe_init(void) {

    if (awt_pipe_inited) {
	return;
    }

    if ( pipe ( awt_pipe_fds ) == 0 )
    {
	/* 
	** the write wakes us up from the infinite sleep, which 
	** then we cause a delay of AWT_FLUSHTIME and then we 
	** flush.
	*/
	int flags = 0;
	awt_set_poll_timeout (def_poll_timeout);
	/* set the pipe to be non-blocking */
	flags = fcntl ( AWT_READPIPE, F_GETFL, 0 );
	fcntl( AWT_READPIPE, F_SETFL, flags | O_NDELAY | O_NONBLOCK ); 
	flags = fcntl ( AWT_WRITEPIPE, F_GETFL, 0 );
	fcntl( AWT_WRITEPIPE, F_SETFL, flags | O_NDELAY | O_NONBLOCK );
        awt_pipe_inited = True;
    }
    else
    {
	AWT_READPIPE = -1;
	AWT_WRITEPIPE = -1;
        awt_pipe_inited = False;
    }
} /* awt_pipe_init() */

/**
 * Processes and removes one X/Xt event from the Xt event queue.
 * Handles events pushed back via awt_push_back_event() FIRST,
 * then new events on the X queue
 */
static void
processOneEvent(int iMask) {

            XEvent xev;
	    if (putbackQueueCount > 0) {
		// There is a pushed-back event - handle it first
		if (awt_get_next_put_back_event(&xev) == 0) {
		    XtDispatchEvent(&xev);
		}
		return;
	    }

            if (XtAppPeekEvent(awt_appContext, &xev)) {
             /*
              * Fix for BugTraq ID 4041235, 4100167:
              * First check that the event still has a widget, because
              * the widget may have been destroyed by another thread.
              */
              Widget widget=XtWindowToWidget(awt_display, xev.xany.window);
              if ((widget == 0) || (!XtIsObject(widget)) ||
                  (widget->core.being_destroyed)) {
                /* There is no longer a widget for this event. Skip it. */
                XNextEvent(awt_display, &xev);
                return;
              }

              /* There is an X event on the queue. */
	      switch (xev.type) {
	      case KeyPress:
	      case KeyRelease:
	      case ButtonPress:
	      case ButtonRelease:
	      case MotionNotify:
	      case EnterNotify:
	      case LeaveNotify: {
		/* Fix for BugTraq ID 4048060. Dispatch scrolling events
		   immediately to the ScrollBar widget to prevent spurious
		   continuous scrolling. Otherwise, if the application is busy,
		   the ButtonRelease event is not dispatched in time to prevent
		   a ScrollBar timeout from expiring, and restarting the
		   continuous scrolling timer.
		*/

		Widget w;

		if ((xev.type == ButtonPress				||
		     xev.type == ButtonRelease				||
		     (xev.type == MotionNotify				&&
		      (xev.xmotion.state == Button1Mask			||
		       xev.xmotion.state == Button2Mask			||
		       xev.xmotion.state == Button3Mask)))		&&
		   (w = XtWindowToWidget(awt_display, xev.xany.window)) &&
		   (XtIsSubclass(w, xmScrollBarWidgetClass))) {
		  /* Use XNextEvent instead of XtAppNextEvent, because
		     XtAppNextEvent processes timers before getting the next X
		     event, causing a race condition, since the TimerEvent
		     callback in the ScrollBar widget restarts the continuous
		     scrolling timer.
		  */
		  XNextEvent(awt_display, &xev);
		  XtDispatchEvent(&xev);
		  XSync(awt_display, False);
		}
		else {
		  XtAppNextEvent(awt_appContext, &xev);
		  if (shouldDispatchToWidget(&xev) == 0) {
		      XtDispatchEvent(&xev);
		  }
                }
	      }

	      break;

              default:
                XtAppProcessEvent(awt_appContext, iMask);
                break;
	      }
            }
            else {
              /* There must be a timer, alternate input, or signal event. */
              XtAppProcessEvent(awt_appContext, iMask & ~XtIMXEvent);
	    }
} /* processOneEvent() */

/*
 * Waits for X/Xt events to appear on the pipe. Returns only when
 * it is likely (but not definite) that there are events waiting to
 * be processed.
 *
 * This routine also flushes the outgoing X queue, when the 
 * awt_next_flush_time has been reached.
 */
static void
waitForEvents(JNIEnv *env, int fdXPipe, int fdAWTPipe) {

	int iMask;
        while ( (iMask = awt_events_pending(awt_appContext)) == 0) {
#ifdef USE_SELECT
	    performSelect(env,fdXPipe,fdAWTPipe);
#else
	    performPoll(env,fdXPipe,fdAWTPipe);
#endif
	    if ((awt_next_flush_time > 0) &&
		(awtJNI_TimeMillis() > awt_next_flush_time)) {
                AWT_FLUSHOUTPUT_NOW();
	    }
        }  /* end while awt_events_pending() == 0 */ 
} /* waitForEvents() */

/*************************************************************************
 **									**
 ** WE USE EITHER select() OR poll(), DEPENDING ON THE USE_SELECT       **
 ** COMPILE-TIME CONSTANT.						**
 **									**
 *************************************************************************/

#ifdef USE_SELECT

static struct fd_set rdset;
struct timeval sel_time;

/*
 * Performs select() on both the X pipe and our AWT utility pipe. 
 * Returns when data arrives or the operation times out.
 *
 * Not all Xt events come across the X pipe (e.g., timers
 * and alternate inputs), so we must time out every now and 
 * then to check the Xt event queue.
 *
 * The fdAWTPipe will be empty when this returns.
 */
static void
performSelect(JNIEnv *env, int fdXPipe, int fdAWTPipe) {

	    int result;
	    int count;
	    unsigned long timeout = awt_get_poll_timeout(False);


            FD_ZERO( &rdset );
    	    FD_SET(fdXPipe, &rdset);
	    if (AWT_READPIPE > 0) {
    	    	FD_SET(AWT_READPIPE, &rdset);
	    }
	    if (timeout == 0) {
	        // be sure other threads get a chance
	        awtJNI_ThreadYield(env);
	    }
	    // set the appropriate time values. The assert() in
	    // MToolkit_run() makes sure that this will not overflow
	    sel_time.tv_sec = (timeout * 1000) / (1000 * 1000);
	    sel_time.tv_usec = (timeout * 1000) % (1000 * 1000);
	    AWT_NOFLUSH_UNLOCK();
            result = select(2, &rdset, 0, 0, &sel_time);
	    AWT_LOCK();

	    /* reset tick if this was not a time out */
	    if (result == 0) {
		/* select() timed out -- update timeout value */
		awt_get_poll_timeout(True);
	    }
	    if ( FD_ISSET ( AWT_READPIPE, &rdset ) )
	    {
		/* There is data on the AWT pipe - empty it */
	    	do {
	            count = read(AWT_READPIPE, read_buf, AWT_POLL_BUFSIZE );
	    	} while (count == AWT_POLL_BUFSIZE ); 
	    }
} /* performSelect() */

#else /* !USE_SELECT */

/*
 * Polls both the X pipe and our AWT utility pipe. Returns
 * when there is data on one of the pipes, or the operation times 
 * out.
 *
 * Not all Xt events come across the X pipe (e.g., timers
 * and alternate inputs), so we must time out every now and 
 * then to check the Xt event queue.
 *
 * The fdAWTPipe will be empty when this returns.
 */
static void
performPoll(JNIEnv *env, int fdXPipe, int fdAWTPipe) {

            static struct pollfd pollFds[2];
	    unsigned long timeout = awt_get_poll_timeout(False);
	    int result;
	    int count;

            pollFds[0].fd = fdXPipe;
            pollFds[0].events = POLLRDNORM;
	    pollFds[0].revents = 0;

	    pollFds[1].fd = fdAWTPipe;
            pollFds[1].events = POLLRDNORM;
	    pollFds[1].revents = 0;

	    AWT_NOFLUSH_UNLOCK();
	    #ifdef DEBUG
	    if (debugEventLoop) {
		/* print the poll timeout time in brackets */
		jio_fprintf(stderr, "[%dms]",(int)timeout);
		if (++debugPrintLineCount > 8) {
		    jio_fprintf(stderr, "\n");
		    debugPrintLineCount = 0;
		}
	    }
	    #endif

	    /* ACTUALLY DO THE POLL() */
	    if (timeout == 0) {
	        // be sure other threads get a chance
	        awtJNI_ThreadYield(env);
	    }
	    result = poll( pollFds, 2, timeout );

	    #ifdef DEBUG
	    if (debugEventLoop) {
		jio_fprintf(stderr, "[poll()->%d]", result);
		if (++debugPrintLineCount > 8) {
		    jio_fprintf(stderr, "\n");
		    debugPrintLineCount = 0;
		}
	    }
	    #endif
	    AWT_LOCK();
	    if (result == 0) {
	        /* poll() timed out -- update timeout value */
		awt_get_poll_timeout(True);
	    }
	    if ( pollFds[1].revents )
	    {
		/* There is data on the AWT pipe - empty it */
	        do {
	    	    count = read(AWT_READPIPE, read_buf, AWT_POLL_BUFSIZE );
	        } while (count == AWT_POLL_BUFSIZE ); 
		/*#ifdef DEBUG
		if (debugEventLoop) {
		    jio_fprintf(stderr, "wokeup on AWTPIPE, timeout:%d\n", timeout);
		}
		#endif*/
	    }
	    return;

} /* performPoll() */

#endif /* !USE_SELECT */

/*
 * Pushes an X event back on the queue to be handled
 * later.
 *
 * Ignores the request if event is NULL
 */
void
awt_put_back_event(JNIEnv *env, XEvent *event) {

    Boolean addIt = True;
    if (putbackQueueCount >= putbackQueueCapacity) {
	/* not enough room - alloc 50% more space */
	int newCapacity;
	XEvent *newQueue;
	newCapacity = putbackQueueCapacity * 3 / 2;
	if ((newCapacity - putbackQueueCapacity) 
					< PUTBACK_QUEUE_MIN_INCREMENT) {
	    /* always increase by at least min increment */
	    newCapacity = putbackQueueCapacity + PUTBACK_QUEUE_MIN_INCREMENT;
	}
	newQueue = (XEvent*)realloc(
			putbackQueue, newCapacity*(sizeof(XEvent)));
	if (newQueue == NULL) {
	    JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
	    addIt = False;
	} else {
	    putbackQueue = newQueue;
	    putbackQueueCapacity = newCapacity;
	}
    }
    if (addIt) {
	char oneChar = 'p';
	memcpy(&(putbackQueue[putbackQueueCount]), event, sizeof(XEvent));
	putbackQueueCount++;

	// wake up the event loop, if it's sleeping
        write (AWT_WRITEPIPE, &oneChar, 1);
    }

    return;
} /* awt_put_back_event() */

/*
 * Gets the next event that has been pushed back onto the queue. 
 * Returns 0 and fills in xev_out if successful
 */
static int
awt_get_next_put_back_event(XEvent *xev_out) {

    Boolean err = False;
    if (putbackQueueCount < 1) {
	err = True;
    } else {
	memcpy(xev_out, &(putbackQueue[0]), sizeof(XEvent));
    }
    if (!err) {
	/* remove it from the queue */
	if (putbackQueueCount == 1) {

	    // queue is now empty
	    if (putbackQueueCapacity > PUTBACK_QUEUE_MIN_INCREMENT) {

		/* Too much space -- delete it and rebuild later */
		free(putbackQueue);
		putbackQueue = NULL;
		putbackQueueCapacity = 0;
	    }
	} else {
	    /* more than 1 event in queue - shift all events to the left */
	    /* We don't free the allocated memory until the queue
	       becomes empty, just 'cause it's easier that way. */
	    /* NOTE: use memmove(), because the memory blocks overlap */
	    memmove(&(putbackQueue[0]), &(putbackQueue[1]), 
		(putbackQueueCount-1)*sizeof(XEvent));
	}
	--putbackQueueCount;
    }
    assert(putbackQueueCount >= 0);

    return (err? -1:0);
    
} /* awt_get_next_put_back_event() */

/**
 * Determines whether or not there are X or Xt events pending.
 * Looks at the putbackQueue.
 */
static XtInputMask 
awt_events_pending(XtAppContext appContext) {
    XtInputMask imask = 0;
    imask = XtAppPending(appContext);
    if (putbackQueueCount > 0) {
	imask |= XtIMXEvent;
    }
    return imask;
}


#ifndef NOMODALFIX
#define WIDGET_ARRAY_SIZE 5;
static int arraySize = 0;
static int arrayIndx = 0;
static Widget *dShells = NULL;

void 
awt_shellPoppedUp(Widget shell, 
                   XtPointer modal,
                   XtPointer call_data)
{
    if (arrayIndx == arraySize ) {          
	/* if we have not allocate an array, do it first */
        if (arraySize == 0) {
	    arraySize += WIDGET_ARRAY_SIZE;
	    dShells = (Widget *) malloc(sizeof(Widget) * arraySize);
	} else { 
            arraySize += WIDGET_ARRAY_SIZE;
	    dShells = (Widget *) realloc((void *)dShells, sizeof(Widget) * arraySize);
        }
    } 

    dShells[arrayIndx] = shell;
    arrayIndx++;
}

void 
awt_shellPoppedDown(Widget shell, 
                   XtPointer modal,
                   XtPointer call_data)
{
    arrayIndx--;

    if (dShells[arrayIndx] == shell) {
	dShells[arrayIndx] = NULL;
	return;
    } else {
	int i;

	/* find the position of the shell in the array */
	for (i = arrayIndx; i >= 0; i--) {
	    if (dShells[i] == shell) {
		break;
	    }
	}
	
	/* remove the found element */
	while (i <= arrayIndx-1) {
	    dShells[i] = dShells[i+1];
	    i++;
	}
    }
}

Boolean 
awt_isWidgetModal(Widget widget)
{
    Widget w;	

    for (w = widget; !XtIsShell(w); w = XtParent(w)) { }
    
    while (w != NULL) {
	if (w == dShells[arrayIndx-1]) {
	    return True; 
	}
	w = XtParent(w);
    }
    return False;
}

Boolean 
awt_isModal()
{
    return (arrayIndx > 0);
}
#endif NOMODALFIX


/*
 * Simply waits for terminateFn() to return True. Waits on the
 * awt lock and is notified to check its state by the main event
 * loop whenever the Xt event queue is empty.
 */
void 
awt_MToolkit_modalWait(int (*terminateFn) (void *data), void *data )
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    AWT_LOCK();
    AWT_FLUSHOUTPUT_NOW();
    while ((*terminateFn) (data) == 0) {
        AWT_WAIT(AWT_MAX_POLL_TIMEOUT);
    }
    AWT_NOTIFY_ALL();
    AWT_UNLOCK();
}

static long 
colorToRGB(XColor * color)
{
    long rgb = 0;

    rgb |= ((color->red >> 8) << 16);
    rgb |= ((color->green >> 8) << 8);
    rgb |= ((color->blue >> 8) << 0);

    return rgb;
}

/*
 * fix for bug #4088106 - ugly text boxes and grayed out looking text
 */

XmColorProc oldColorProc;

void 
ColorProc(XColor* bg_color, 
	  XColor* fg_color, 
	  XColor* sel_color, 
	  XColor* ts_color, 
	  XColor* bs_color)
{
    unsigned long plane_masks[1];
    unsigned long colors[5];

    /* use the default procedure to calculate colors */
    oldColorProc(bg_color, fg_color, sel_color, ts_color, bs_color);

    /* check if there is enought free color cells */
    if (XAllocColorCells(awt_display, defaultConfig->awt_cmap, False, 
	plane_masks, 0, colors, 5)) {
        XFreeColors(awt_display, defaultConfig->awt_cmap, colors, 5, 0);
	return;
    }

    /* find the closest matches currently available */
    fg_color->pixel = defaultConfig->AwtColorMatch(fg_color->red   >> 8,
						   fg_color->green >> 8,
						   fg_color->blue  >> 8, 
						   defaultConfig);
    fg_color->flags = DoRed | DoGreen | DoBlue;
    XQueryColor(awt_display, defaultConfig->awt_cmap, fg_color);
    sel_color->pixel = defaultConfig->AwtColorMatch(sel_color->red   >> 8,
						    sel_color->green >> 8,
						    sel_color->blue  >> 8,
						    defaultConfig);
    sel_color->flags = DoRed | DoGreen | DoBlue;
    XQueryColor(awt_display, defaultConfig->awt_cmap, sel_color);
    ts_color->pixel = defaultConfig->AwtColorMatch(ts_color->red   >> 8,
						   ts_color->green >> 8,
						   ts_color->blue  >> 8,
						   defaultConfig);
    ts_color->flags = DoRed | DoGreen | DoBlue;
    XQueryColor(awt_display, defaultConfig->awt_cmap, ts_color);
    bs_color->pixel = defaultConfig->AwtColorMatch(bs_color->red   >> 8,
						   bs_color->green >> 8,
						   bs_color->blue  >> 8,
						   defaultConfig);
    bs_color->flags = DoRed | DoGreen | DoBlue;
    XQueryColor(awt_display, defaultConfig->awt_cmap, bs_color);
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MToolkit_init
  (JNIEnv *env, jobject this) 
{
    int argc = 0;
    char *argv[1];
    /*
     * Note: The MToolkit object depends on the static initializer
     * of X11GraphicsEnvironment to initialize the connection to
     * the X11 server.
     */
    Display *dpy = awt_display;
    char *multiclick_time_query;

    static String fallback_resources[] =
    {
        "*enableThinThickness:                   True",
        "*XmFileSelectionBox.fileFilterStyle:    XmFILTER_HIDDEN_FILES",
        "*XmFileSelectionBox.pathMode:           XmPATH_MODE_RELATIVE",
        "*XmFileSelectionBox.resizePolicy:       XmRESIZE_GROW",
        "*XmFileSelectionBox*dirTextLabelString:         Enter path or folder name:",
        "*XmFileSelectionBox*applyLabelString:           Update",
        "*XmFileSelectionBox*selectionLabelString:       Enter file name:",
        "*XmFileSelectionBox*dirListLabelString:         Folders",
        NULL                        /* Must be NULL terminated */
    };

    /* Need to make sure this is deleted someplace! */
    AWT_LOCK();

    if (!XSupportsLocale()) {
        jio_fprintf(stderr,
                    "current locale is not supported in X11, locale is set to C");
        setlocale(LC_ALL, "C");
    }
    if (!XSetLocaleModifiers("")) {
        jio_fprintf(stderr, "X locale modifiers are not supported, using default");
    }
#ifdef NETSCAPE
    if (awt_init_xt) {
        XtToolkitInitialize();
    }
#else
    XtToolkitInitialize();
#endif

    argv[0] = 0;
    awt_appContext = XtCreateApplicationContext();
    XtAppSetFallbackResources(awt_appContext, fallback_resources);
    XtDisplayInitialize(awt_appContext,
			awt_display,
			"MToolkit app",
			"XApplication",
			NULL,
			0,
			&argc,
			argv);

    XtAppSetErrorHandler(awt_appContext, (XtErrorHandler) xtError);
    /*
     * Find the correct awt_multiclick_time to use. We normally
     * would call XtMultiClickTime() and wouldn't have to do
     * anything special, but because OpenWindows defines its own
     * version (OpenWindows.MultiClickTimeout), we need to
     * determine out which resource to use.
     *
     * We do this by searching in order for:
     *
     *   1) an explicit definition of multiClickTime
     *      (this is the resource that XtGetMultiClickTime uses)
     *
     * if that fails, search for:
     *
     *   2) an explicit definition of Openwindows.MultiClickTimeout
     *
     * if both searches fail:
     *
     *   3) use the fallback provided by XtGetMultiClickTime()
     *      (which is 200 milliseconds... I looked at the source :-)
     *
     */
    multiclick_time_query = XGetDefault(dpy, "*", "multiClickTime");
    if (multiclick_time_query) {
        awt_multiclick_time = XtGetMultiClickTime(awt_display);
    } else {
        multiclick_time_query = XGetDefault(dpy, "OpenWindows",
                                            "MultiClickTimeout");
        if (multiclick_time_query) {
            /* Note: OpenWindows.MultiClickTimeout is in tenths of 
               a second, so we need to multiply by 100 to convert to
               milliseconds */
            awt_multiclick_time = atoi(multiclick_time_query) * 100;
        } else {
            awt_multiclick_time = XtGetMultiClickTime(awt_display);
        }
    }

    /* printf("**** awt_multiclick_time = %d\n", awt_multiclick_time); */


/*
  scrollBugWorkAround = 
  (strcmp(XServerVendor(dpy), "Sun Microsystems, Inc.") == 0
  && XVendorRelease(dpy) == 3400);
*/
    scrollBugWorkAround = TRUE;

    /*
     * Create the cursor for TextArea scrollbars...
     */
    awt_scrollCursor = XCreateFontCursor(awt_display, XC_left_ptr);

    awt_defaultBg = defaultConfig->AwtColorMatch(200, 200, 200, defaultConfig);
    awt_defaultFg = awt_blackpixel;
    setup_modifier_map(awt_display);

    awt_initialize_Xm_DnD(awt_display);

    /*
     * fix for bug #4088106 - ugly text boxes and grayed out looking text
     */
    oldColorProc = XmGetColorCalculation();
    XmSetColorCalculation(ColorProc);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    run
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MToolkit_run
  (JNIEnv *env, jobject this) 
{
    /*
     * in performSelect(), we multiply the timeout by 1000. Make sure
     * that the maximum value will not cause an overflow.
     */
    assert(AWT_MAX_POLL_TIMEOUT <= (ULONG_MAX/1000));

    awt_MainThread = awtJNI_GetCurrentThread(env);
    awt_MToolkit_loop(env); /* never returns */
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    makeColorModel
 * Signature: ()Ljava/awt/image/ColorModel;
 */
JNIEXPORT jobject JNICALL Java_sun_awt_motif_MToolkit_makeColorModel
  (JNIEnv *env, jclass this) 
{
    return awtJNI_GetColorModel(env, defaultConfig);
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    getScreenResolution
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MToolkit_getScreenResolution
  (JNIEnv *env, jobject this) 
{
    return (jint) ((DisplayWidth(awt_display, awt_screen) * 25.4) /
                   DisplayWidthMM(awt_display, awt_screen));
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    getScreenWidth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MToolkit_getScreenWidth
  (JNIEnv *env, jobject this) 
{
    return DisplayWidth(awt_display, awt_screen);
}
/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    getScreenHeight
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MToolkit_getScreenHeight
  (JNIEnv *env, jobject this) 
{
    return DisplayHeight(awt_display, awt_screen);
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    sync
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MToolkit_sync
  (JNIEnv *env, jobject this) 
{
    AWT_LOCK();
    XSync(awt_display, False);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    beep
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MToolkit_beep
  (JNIEnv *env, jobject this) 
{
    XBell(awt_display, 0);
}

/*
 * Class:     sun_awt_motif_MToolkit
 * Method:    loadSystemColors
 * Signature: ([I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MToolkit_loadSystemColors
  (JNIEnv *env, jobject this, jintArray systemColors) 
{
    Widget frame, panel, control, menu, text, scrollbar;
    Colormap cmap;
    Pixel bg, fg, highlight, shadow;
    Pixel pixels[java_awt_SystemColor_NUM_COLORS];
    XColor *colorsPtr;
    jint rgbColors[java_awt_SystemColor_NUM_COLORS];
    int count = 0;
    int i, j;

    AWT_LOCK();

    /* 
     * initialize array of pixels 
     */
    for (i = 0; i < java_awt_SystemColor_NUM_COLORS; i++) {
        pixels[i] = -1;
    }

    /*
     * Create phantom widgets in order to determine the default
     * colors;  this is somewhat inelegant, however it is the simplest
     * and most reliable way to determine the system's default colors
     * for objects.
     */
    frame = XtAppCreateShell("AWTColors", "XApplication",
                             vendorShellWidgetClass,
                             awt_display,
                             NULL,
                             0);
    /*
      XtSetMappedWhenManaged(frame, False);
      XtRealizeWidget(frame);
    */
    panel = XmCreateDrawingArea(frame, "awtPanelColor", NULL, 0);
    control = XmCreatePushButton(panel, "awtControlColor", NULL, 0);
    menu = XmCreatePulldownMenu(control, "awtColorMenu", NULL, 0);
    text = XmCreateText(panel, "awtTextColor", NULL, 0);
    scrollbar = XmCreateScrollBar(panel, "awtScrollbarColor", NULL, 0);

    XtVaGetValues(panel,
                  XmNbackground, &bg,
                  XmNforeground, &fg,
                  XmNcolormap, &cmap,
                  NULL);

    pixels[java_awt_SystemColor_WINDOW] = bg;
    count++;
    pixels[java_awt_SystemColor_INFO] = bg;
    count++;
    pixels[java_awt_SystemColor_WINDOW_TEXT] = fg;
    count++;
    pixels[java_awt_SystemColor_INFO_TEXT] = fg;
    count++;

    XtVaGetValues(menu,
                  XmNbackground, &bg,
                  XmNforeground, &fg,
                  NULL);

    pixels[java_awt_SystemColor_MENU] = bg;
    count++;
    pixels[java_awt_SystemColor_MENU_TEXT] = fg;
    count++;

    XtVaGetValues(text,
                  XmNbackground, &bg,
                  XmNforeground, &fg,
                  NULL);

    pixels[java_awt_SystemColor_TEXT] = bg;
    count++;
    pixels[java_awt_SystemColor_TEXT_TEXT] = fg;
    count++;
    pixels[java_awt_SystemColor_TEXT_HIGHLIGHT] = fg;
    count++;
    pixels[java_awt_SystemColor_TEXT_HIGHLIGHT_TEXT] = bg;
    count++;

    XtVaGetValues(control,
                  XmNbackground, &bg,
                  XmNforeground, &fg,
                  XmNtopShadowColor, &highlight,
                  XmNbottomShadowColor, &shadow,
                  NULL);

    pixels[java_awt_SystemColor_CONTROL] = bg;
    count++;
    pixels[java_awt_SystemColor_CONTROL_TEXT] = fg;
    count++;
    pixels[java_awt_SystemColor_CONTROL_HIGHLIGHT] = highlight;
    count++;
    pixels[java_awt_SystemColor_CONTROL_LT_HIGHLIGHT] = highlight;
    count++;
    pixels[java_awt_SystemColor_CONTROL_SHADOW] = shadow;
    count++;
    pixels[java_awt_SystemColor_CONTROL_DK_SHADOW] = shadow;
    count++;

    XtVaGetValues(scrollbar,
                  XmNbackground, &bg,
                  NULL);
    pixels[java_awt_SystemColor_SCROLLBAR] = bg;
    count++;

    /* 
     * Convert pixel values to RGB 
     */
    colorsPtr = (XColor *) malloc(count * sizeof(XColor));
    j = 0;
    for (i = 0; i < java_awt_SystemColor_NUM_COLORS; i++) {
        if (pixels[i] != -1) {
            colorsPtr[j++].pixel = pixels[i];
        }
    }
    XQueryColors(awt_display, cmap, colorsPtr, count);

    /* Get current System Colors */
 
    (*env)->GetIntArrayRegion (env, systemColors, 0,
                              java_awt_SystemColor_NUM_COLORS,
                              rgbColors);

    /* 
     * Fill systemColor array with new rgb values 
     */

    j = 0;
    for (i = 0; i < java_awt_SystemColor_NUM_COLORS; i++) {
        if (pixels[i] != -1) {
            long rgb = colorToRGB(&colorsPtr[j++]);

            /*
              printf("SystemColor[%d] = %x\n", i, rgb);
            */
            rgbColors[i] = (rgb | 0xFF000000);
        }
    }

    (*env)->SetIntArrayRegion(env,
                              systemColors,
                              0,
                              java_awt_SystemColor_NUM_COLORS,
                              rgbColors);

    /* Duplicate system colors. If color allocation is unsuccessful,
       system colors will be approximated with matched colors */
    if (defaultConfig->awt_depth == 8) 
        awt_allocate_systemcolors(colorsPtr, count, defaultConfig);

    /*
     * Cleanup
     */
    XtDestroyWidget(frame);
    free(colorsPtr);

    AWT_UNLOCK();
}

/* 
 * Returns true if the current thread is privileged. Currently,
 * only the main event loop thread is considered to be privileged.
 */
Boolean 
awt_currentThreadIsPrivileged(JNIEnv *env) {

    return (*env)->IsSameObject(env, 
			awt_MainThread, awtJNI_GetCurrentThread(env));
}
