/*
 * @(#)awt_Menu.c	1.34 98/08/31
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_p.h"
#include "color.h"
#include "java_awt_Menu.h"
#include "sun_awt_motif_MMenuPeer.h"
#include "java_awt_MenuBar.h"
#include "sun_awt_motif_MMenuBarPeer.h"

#include "awt_MenuBar.h"
#include "awt_MenuComponent.h"
#include "awt_MenuItem.h"
#include "awt_Menu.h"

#include "multi_font.h"
#include <jni.h>
#include <jni_util.h>

extern struct MenuComponentIDs menuComponentIDs;
extern struct MenuItemIDs menuItemIDs;
extern struct MMenuItemPeerIDs mMenuItemPeerIDs;
extern struct MMenuBarPeerIDs mMenuBarPeerIDs;

struct MenuIDs menuIDs;

/*
 * Class:     java_awt_Menu
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   Menu.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_java_awt_Menu_initIDs
  (JNIEnv *env, jclass cls)
{
    menuIDs.tearOff = (*env)->GetFieldID(env, cls, "tearOff", "Z");
    menuIDs.isHelpMenu = (*env)->GetFieldID(env, cls, "isHelpMenu", "Z");
}

/*
 * this is a MMenuPeer instance
 */
static void
awtJNI_CreateMenu(JNIEnv * env, jobject this, Widget menuParent)
{
    int argc;
    Arg args[10];
    char *ctitle = NULL;
    struct MenuData *mdata;
    struct FontData *fdata;
    Pixel bg;
    Pixel fg;
    XmFontList fontlist;
    Widget tearOff;
    XmString mfstr = NULL;
    XmString str = NULL;
    jobject target;
    jobject targetFont;
    jobject label;
    jobject font;
    jboolean IsMultiFont;
    jboolean isTearOff;

    /* perhaps this is unncessary, if awtJNI_CreateMenu is only called
     * from a native method.
     */
    if ((*env)->PushLocalFrame(env, 16) < 0) {
        return;
    }

    fdata = NULL;

    target = (*env)->GetObjectField(env, this, mMenuItemPeerIDs.target);
    if (JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
	(*env)->PopLocalFrame(env, 0);
        return;
    }
    font = JNU_CallMethodByName(env, NULL, target, "getFont_NoClientCode",
				"()Ljava/awt/Font;").l;

    mdata = ZALLOC(MenuData);
    (*env)->SetLongField(env, this, mMenuItemPeerIDs.pData, (jlong) mdata);
    if (mdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
	(*env)->PopLocalFrame(env, 0);
        return;
    }
    targetFont = (*env)->GetObjectField(env, target, menuComponentIDs.font);
    if (!JNU_IsNull(env, targetFont) &&
        (fdata = awtJNI_GetFontData(env, targetFont, 0)) != 0) {
        IsMultiFont = awtJNI_IsMultiFont(env, targetFont);
    } else {
        IsMultiFont = awtJNI_IsMultiFont(env, font);
    }

    label = (*env)->GetObjectField(env, target, menuItemIDs.label);
    if (JNU_IsNull(env, label)) {
        mfstr = XmStringCreateSimple("");
        ctitle = "";
    } else {
        if (IsMultiFont) {
            mfstr = awtJNI_MakeMultiFontString(env, label, font);
        } else {
            ctitle = (char *) JNU_GetStringPlatformChars(env, label, NULL);
        }
    }

    XtVaGetValues(menuParent, XmNbackground, &bg, NULL);
    XtVaGetValues(menuParent, XmNforeground, &fg, NULL);

    argc = 0;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, fg);
    argc++;

    isTearOff = (*env)->GetBooleanField(env, target, menuIDs.tearOff);

    if (isTearOff) {
        XtSetArg(args[argc], XmNtearOffModel, XmTEAR_OFF_ENABLED);
        argc++;
    }
    XtSetArg(args[argc], XmNvisual, defaultConfig->awt_visInfo.visual);
    argc++;

    if (IsMultiFont) {
        mdata->itemData.comp.widget = XmCreatePulldownMenu(menuParent,
                                                           "",
                                                           args,
                                                           argc);
    } else {
        mdata->itemData.comp.widget = XmCreatePulldownMenu(menuParent,
                                                           ctitle,
                                                           args,
                                                           argc);
    }

    if (isTearOff) {
        tearOff = XmGetTearOffControl(mdata->itemData.comp.widget);
        XtVaSetValues(tearOff,
                      XmNbackground, bg,
                      XmNforeground, fg,
                      XmNhighlightColor, fg,
                      NULL);
    }
    argc = 0;
    XtSetArg(args[argc], XmNsubMenuId, mdata->itemData.comp.widget);
    argc++;

    if (IsMultiFont) {
        XtSetArg(args[argc], XmNlabelString, mfstr);
    } else {
        str = XmStringCreateLtoR(ctitle, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(args[argc], XmNlabelString, str);
    }
    argc++;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, fg);
    argc++;

    if (!JNU_IsNull(env, targetFont) && (fdata != 0)) {
        if (IsMultiFont) {
            fontlist = awtJNI_GetFontList(env, targetFont);
        } else {
            fontlist = XmFontListCreate(fdata->xfont, "labelFont");
        }
        XtSetArg(args[argc], XmNfontList, fontlist);
        argc++;
    } else {
        if (IsMultiFont) {
            fontlist = awtJNI_GetFontList(env, font);
            XtSetArg(args[argc], XmNfontList, fontlist);
            argc++;
        }
    }

    if (IsMultiFont) {
        mdata->comp.widget = XmCreateCascadeButton(menuParent, "", args, argc);
    } else {
        mdata->comp.widget = XmCreateCascadeButton(menuParent, ctitle, args, argc);
    }

    if ((*env)->GetBooleanField(env, target, menuIDs.isHelpMenu)) {
        XtVaSetValues(menuParent,
                      XmNmenuHelpWidget, mdata->comp.widget,
                      NULL);
    }

    /*
     * Free resources
     */
    if (!JNU_IsNull(env, targetFont)) {
        XmFontListFree(fontlist);
    }

    if (mfstr != NULL) {
      XmStringFree(mfstr); 
      mfstr = NULL;
    }

    if (str) {
      XmStringFree(str);
      str = NULL;
    }

    XtManageChild(mdata->comp.widget);
    XtSetSensitive(mdata->comp.widget,
                   (*env)->GetBooleanField(env, target, menuItemIDs.enabled) ?
		   True : False);

    if (ctitle != NULL && ctitle != "") {
        JNU_ReleaseStringPlatformChars(env, label, (const char *) ctitle);
    }
    (*env)->PopLocalFrame(env, 0);
}

/*
 * Class:     sun_awt_motif_MMenuPeer
 * Method:    createMenu
 * Signature: (Lsun/awt/motif/MMenuBarPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuPeer_createMenu
  (JNIEnv *env, jobject this, jobject parent)
{
    struct ComponentData *mbdata;

    AWT_LOCK();
    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    mbdata = (struct ComponentData *)
      (*env)->GetLongField(env, parent, mMenuBarPeerIDs.pData);
    if (mbdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    awtJNI_CreateMenu(env, this, mbdata->widget);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MMenuPeer
 * Method:    createSubMenu
 * Signature: (Lsun/awt/motif/MMenuPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuPeer_createSubMenu
  (JNIEnv *env, jobject this, jobject parent)
{
    struct MenuData *mpdata;

    AWT_LOCK();
    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    mpdata = (struct MenuData *)
      (*env)->GetLongField(env, parent, mMenuItemPeerIDs.pData);
    if (mpdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    awtJNI_CreateMenu(env, this, mpdata->itemData.comp.widget);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MMenuPeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuPeer_pDispose
  (JNIEnv *env, jobject this)
{
    struct MenuData *mdata;

    AWT_LOCK();

    mdata = (struct MenuData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);
    if (mdata == 0) {
        AWT_UNLOCK();
        return;
    }
    XtUnmanageChild(mdata->comp.widget);
    awt_util_consumeAllXEvents(mdata->itemData.comp.widget);
    awt_util_consumeAllXEvents(mdata->comp.widget);
    XtDestroyWidget(mdata->itemData.comp.widget);
    XtDestroyWidget(mdata->comp.widget);
    free((void *) mdata);
    AWT_UNLOCK();
}
