/*
 * @(#)awt_MenuItem.c	1.43 98/09/15
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include <Xm/Separator.h>
#include "java_awt_MenuItem.h"
#include "sun_awt_motif_MMenuItemPeer.h"
#include "sun_awt_motif_MCheckboxMenuItemPeer.h"
#include "java_awt_Menu.h"
#include "sun_awt_motif_MMenuPeer.h"

#include "awt_MenuComponent.h"
#include "awt_MenuItem.h"

#include "multi_font.h"
#include <jni.h>
#include <jni_util.h>
#include <jlong.h>

extern struct MenuComponentIDs menuComponentIDs;

/* fieldIDs for MenuItem fields that may be accessed from C */
struct MenuItemIDs menuItemIDs;

/*
 * Class:     java_awt_MenuItem
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MenuItem.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_java_awt_MenuItem_initIDs
  (JNIEnv *env, jclass cls)
{
    menuItemIDs.label =
      (*env)->GetFieldID(env, cls, "label", "Ljava/lang/String;");
    menuItemIDs.enabled = 
      (*env)->GetFieldID(env, cls, "enabled", "Z");
    menuItemIDs.shortcut =
      (*env)->GetFieldID(env, cls, "shortcut", "Ljava/awt/MenuShortcut;");
}

/* fieldIDs for MMenuItemPeer fields that may be accessed from C */
struct MMenuItemPeerIDs mMenuItemPeerIDs;

/*
 * Class:     sun_awt_motif_MMenuItemPeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MMenuItemPeer.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuItemPeer_initIDs
  (JNIEnv *env, jclass cls)
{
    mMenuItemPeerIDs.target =
      (*env)->GetFieldID(env, cls, "target", "Ljava/awt/MenuItem;");
    mMenuItemPeerIDs.pData = (*env)->GetFieldID(env, cls, "pData", "J");
    mMenuItemPeerIDs.isCheckbox =
      (*env)->GetFieldID(env, cls, "isCheckbox", "Z");
    mMenuItemPeerIDs.jniGlobalRef =
      (*env)->GetFieldID(env, cls, "jniGlobalRef", "J");
}

/*
 * Class:     sun_awt_motif_MMenuItemPeer
 * Method:    getParent_NoClientCode
 * Signature: (Ljava/awt/MenuComponent;)Ljava/awt/MenuContainer;
 *
 * Gets the MenuContainer parent of this object, without executing client
 * code (e.g., no code in subclasses will be executed).
 */
JNIEXPORT jobject JNICALL Java_sun_awt_motif_MMenuItemPeer_getParent_1NoClientCode
  (JNIEnv *env, jclass thisClass, jobject menuComponent)
{
    jobject parent = NULL;

    /* getParent is actually getParent_NoClientCode() */
    parent = (*env)->CallObjectMethod(
			env,menuComponent,menuComponentIDs.getParent);
    assert(!((*env)->ExceptionOccurred(env)));
    return parent;
}

/*
 *  client_data is MMenuItemPeer instance pointer
 */
static void
MenuItem_selected(Widget w, XtPointer client_data, XmAnyCallbackStruct * s)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    extern int getModifiers(int state);
    int modifiers = getModifiers(s->event->xbutton.state);
    jlong when = jint_to_jlong(s->event->xbutton.time);
    jobject this = (jobject) client_data;

    if ((*env)->GetBooleanField(env, this, mMenuItemPeerIDs.isCheckbox)) {
        Boolean state;
        struct MenuItemData *mdata;

        mdata = (struct MenuItemData *)
	  (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);

        if (mdata != NULL) {
            XtVaGetValues(mdata->comp.widget, XmNset, &state, NULL);

            JNU_CallMethodByName(env, NULL, this
				 ,"action"
				 ,"(JIZ)V"
				 ,when, modifiers, state);
            if ((*env)->ExceptionOccurred(env)) {
                (*env)->ExceptionDescribe(env);
                (*env)->ExceptionClear(env);
            }
        }
    } else {
        JNU_CallMethodByName(env, NULL, this, "action", "(JI)V",
			     when, modifiers);
        if ((*env)->ExceptionOccurred(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        }
    }
}

/*
 * Class:     sun_awt_motif_MMenuItemPeer
 * Method:    createMenuItem
 * Signature: (Lsun/awt/motif/MMenuPeer;)V
 *
 * ASSUMES: This function is never called by a privileged thread
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuItemPeer_createMenuItem
(JNIEnv *env, jobject this, jobject parent)
{
    int argc;
    Arg args[10];
    char *clabel;
    struct MenuData *menuData;
    struct MenuItemData *mdata;
    struct FontData *fdata;
    Pixel bg;
    Pixel fg;
    XmFontList fontlist;
    jobject target;
    jobject targetFont;
    XmString mfstr = NULL;
    XmString shortcut_str = NULL;
    XmString str = NULL;
    jobject font;
    jobject shortcut;
    jboolean IsMultiFont;
    jboolean isCheckbox;
    jstring label;
    jobject globalRef = (*env)->NewGlobalRef(env, this);

    // We call client code on this thread, so it must *NOT* be privileged
    assert(!awt_currentThreadIsPrivileged(env));

    (*env)->SetLongField(env, this,
			 mMenuItemPeerIDs.jniGlobalRef, (jlong) globalRef);

    fdata = NULL;

    fflush(stderr);
    target =
      (*env)->GetObjectField(env, this, mMenuItemPeerIDs.target);
    if (JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    font = JNU_CallMethodByName(env, NULL, target, "getFont_NoClientCode",
				"()Ljava/awt/Font;").l;

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    menuData = (struct MenuData *)
      (*env)->GetLongField(env, parent, mMenuItemPeerIDs.pData);

    targetFont =
      (*env)->GetObjectField(env, target, menuComponentIDs.font);
    if (!JNU_IsNull(env, targetFont) &&
        (fdata = awtJNI_GetFontData(env, targetFont, 0)) != 0) {
        IsMultiFont = awtJNI_IsMultiFont(env, targetFont);
    } else {
        IsMultiFont = awtJNI_IsMultiFont(env, font);
    }

    label = (*env)->GetObjectField(env, target, menuItemIDs.label);
    if (JNU_IsNull(env, label)) {
        mfstr = XmStringCreateSimple("");
        clabel = "";
    } else {
        if (IsMultiFont) {
            mfstr = awtJNI_MakeMultiFontString(env, label, font);
        } else {
            mfstr = XmStringCreateSimple("");
        }
        clabel = (char *) JNU_GetStringPlatformChars(env, label, NULL);
    }

    mdata = ZALLOC(MenuItemData);
    (*env)->SetLongField(env, this, mMenuItemPeerIDs.pData, (jlong) mdata);
    

    argc = 0;
    XtSetArg(args[argc], XmNbackground, &bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, &fg);
    argc++;
    XtGetValues(menuData->itemData.comp.widget, args, argc);

    argc = 0;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, fg);
    argc++;
    if (strcmp(clabel, "-") == 0) {
        mdata->comp.widget = XmCreateSeparator(menuData->itemData.comp.widget,
                                               "", args, argc);
    } else {
        if (IsMultiFont) {
            XtSetArg(args[argc], XmNlabelString, mfstr);
        } else {
 	    str = XmStringCreateLtoR(clabel, XmSTRING_DEFAULT_CHARSET);
	    XtSetArg(args[argc], XmNlabelString, str);
        }
        argc++;

        shortcut =
	  (*env)->GetObjectField(env, target, menuItemIDs.shortcut);
        if (!JNU_IsNull(env, shortcut)) {
            jstring shortcutText;
            char *text = "";

            shortcutText = JNU_CallMethodByName(env, NULL, shortcut,
						"toString",
						"()Ljava/lang/String;").l;

            if (!JNU_IsNull(env, shortcutText)) {
                text = (char *) JNU_GetStringPlatformChars(env, shortcutText, NULL);
            }
	    shortcut_str = XmStringCreateLtoR(text, XmSTRING_DEFAULT_CHARSET);
            XtSetArg(args[argc], XmNacceleratorText, shortcut_str);

            argc++;

            if (!JNU_IsNull(env, shortcutText)) {
                JNU_ReleaseStringPlatformChars(env, shortcutText, (const char *) text);
            }
        }
        if (!JNU_IsNull(env, targetFont) && (fdata != 0)) {
            if (IsMultiFont) {
                fontlist = awtJNI_GetFontList(env, targetFont);
            } else {
                fontlist = XmFontListCreate(fdata->xfont, "labelFont");
            }
            XtSetArg(args[argc], XmNfontList, fontlist);
            argc++;
        } else {
            if (IsMultiFont) {
                fontlist = awtJNI_GetFontList(env, font);
                XtSetArg(args[argc], XmNfontList, fontlist);
                argc++;
            }
        }

        isCheckbox =
	  (*env)->GetBooleanField(env, this, mMenuItemPeerIDs.isCheckbox);
        if (isCheckbox) {
	    /* Fix for 4090493 */
	    if (IsMultiFont) {
		/* FontData that correspond to XmNfontList we just set */
		struct FontData *fdataForIndSize;
		Dimension indSize;
		if (!JNU_IsNull(env, targetFont) && (fdata != 0)) {
		    fdataForIndSize = fdata;
		}
		else {
		    fdataForIndSize = awtJNI_GetFontData(env, font, 0);
		}
		indSize = awt_adjustIndicatorSizeForMenu(awt_computeIndicatorSize(fdataForIndSize));
		if (indSize != MOTIF_XmINVALID_DIMENSION) {
		    XtSetArg(args[argc], XmNindicatorSize, indSize); argc++;
		}
	    }
	    /* End of fix for 4090493 */
            XtSetArg(args[argc], XmNset, False);
            argc++;
            XtSetArg(args[argc], XmNvisibleWhenOff, True);
            argc++;
            mdata->comp.widget = XmCreateToggleButton(menuData->itemData.comp.widget,
                                                      clabel,
                                                      args,
                                                      argc);
        } else {
            mdata->comp.widget = XmCreatePushButton(menuData->itemData.comp.widget,
                                                    clabel,
                                                    args,
                                                    argc);
        }
        XtAddCallback(mdata->comp.widget,
                      ((isCheckbox) ? XmNvalueChangedCallback : XmNactivateCallback),
                      (XtCallbackProc) MenuItem_selected,
                      (XtPointer) globalRef);

        XtSetSensitive(mdata->comp.widget,
	  (*env)->GetBooleanField(env, target, menuItemIDs.enabled) ?
		       True : False);


        if (!JNU_IsNull(env, targetFont)) {
            XmFontListFree(fontlist);
        }
    }

    if (clabel && (clabel != "")) {
        JNU_ReleaseStringPlatformChars(env, label, clabel);
    }

    /*
     * Free up resources after we have created the widget
     */
    if (mfstr != NULL) {
      XmStringFree(mfstr);
      mfstr = NULL;
    }
    if (str) {
      XmStringFree(str);
      str = NULL;
    }
    if (shortcut_str) {
      XmStringFree(shortcut_str);
      shortcut_str = NULL;
    }

    XtManageChild(mdata->comp.widget);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MMenuItemPeer
 * Method:    pSetLabel
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuItemPeer_pSetLabel
(JNIEnv *env, jobject this, jstring label)
{
    struct ComponentData *wdata;
    char *clabel;
    XmString xim;

    AWT_LOCK();
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);
    if (wdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (JNU_IsNull(env, label)) {
        xim = XmStringCreateSimple("");
    } else {
        jobject font;
        jobject target;

        target = (*env)->GetObjectField(env, this, mMenuItemPeerIDs.target);
        if (JNU_IsNull(env, target)) {
            JNU_ThrowNullPointerException(env, "NullPointerException");
            return;
        }
        font = JNU_CallMethodByName(env, NULL, target, "getFont_NoClientCode",
				    "()Ljava/awt/Font;").l;

        if (awtJNI_IsMultiFont(env, font)) {
            xim = awtJNI_MakeMultiFontString(env, label, font);
        } else {
            clabel = (char *) JNU_GetStringPlatformChars(env, label, NULL);
            xim = XmStringCreateLtoR(clabel, "labelFont");
            JNU_ReleaseStringPlatformChars(env, label, clabel);
        }
    }
    XtUnmanageChild(wdata->widget);
    XtVaSetValues(wdata->widget, XmNlabelString, xim, NULL);
    XtManageChild(wdata->widget);
    XmStringFree(xim);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MMenuItemPeer
 * Method:    pEnable
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuItemPeer_pEnable
(JNIEnv *env, jobject this)
{
    struct MenuItemData *mdata;

    AWT_LOCK();

    mdata = (struct MenuItemData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);

    if (mdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtSetSensitive(mdata->comp.widget, True);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MMenuItemPeer
 * Method:    pDisable
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuItemPeer_pDisable
(JNIEnv *env, jobject this)
{
    struct MenuItemData *mdata;

    AWT_LOCK();

    mdata = (struct MenuItemData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);

    if (mdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtSetSensitive(mdata->comp.widget, False);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MMenuItemPeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MMenuItemPeer_pDispose
(JNIEnv *env, jobject this)
{
    struct MenuItemData *mdata;

    AWT_LOCK();

    mdata = (struct MenuItemData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);

    if (mdata != 0) {
        XtUnmanageChild(mdata->comp.widget);
        awt_util_consumeAllXEvents(mdata->comp.widget);
        XtDestroyWidget(mdata->comp.widget);
        free((void *) mdata);
        (*env)->SetLongField(env, this, mMenuItemPeerIDs.pData, 0);
        awtJNI_DeleteGlobalMenuRef(env, this);
    }
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MCheckboxMenuItemPeer
 * Method:    pSetState
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MCheckboxMenuItemPeer_pSetState
  (JNIEnv *env, jobject this, jboolean state)
{
    struct MenuItemData *mdata;

    AWT_LOCK();

    mdata = (struct MenuItemData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);

    if (mdata == 0) {
        JNU_ThrowNullPointerException(env, "menuitem data is null");
	AWT_UNLOCK();
	return;
    }
    XtVaSetValues(mdata->comp.widget, XmNset, (Boolean)state, NULL);
    AWT_UNLOCK();
}
