/*
 * @(#)awt_PopupMenu.c	1.29 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include <Xm/Separator.h>
#include "color.h"
#include "java_awt_PopupMenu.h"
#include "java_awt_Component.h"
#include "java_awt_Event.h"
#include "sun_awt_motif_MPopupMenuPeer.h"
#include "sun_awt_motif_MComponentPeer.h"

#include "awt_PopupMenu.h"
#include "awt_MenuItem.h"
#include "awt_Component.h"
#include "awt_MenuComponent.h"
#include "awt_Menu.h"
#include "awt_Event.h"

#include "multi_font.h"
#include <jni.h>
#include <jni_util.h>

extern struct MMenuItemPeerIDs mMenuItemPeerIDs;
extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct MenuComponentIDs menuComponentIDs;
extern struct MenuItemIDs menuItemIDs;
extern struct MenuIDs menuIDs;

struct MPopupMenuPeerIDs mPopupMenuPeerIDs;

/*
 * Class:     sun_awt_motif_MPopupMenuPeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MPopupMenuPeer.java to initialize the methodIDs for methods that may
   be accessed from C */

JNIEXPORT void JNICALL Java_sun_awt_motif_MPopupMenuPeer_initIDs
  (JNIEnv *env, jclass cls)
{
    mPopupMenuPeerIDs.destroyNativeWidgetAfterGettingTreeLock =
        (*env)->GetMethodID(env, cls,
			    "destroyNativeWidgetAfterGettingTreeLock", "()V");
}

/*
 * client_data is MPopupMenuPeer instance
 */
static void
Popup_popdownCB(Widget w, XtPointer client_data, XtPointer calldata)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject target = NULL;

    XtRemoveCallback(w, XtNpopdownCallback,
                     Popup_popdownCB, (XtPointer) client_data);

    (*env)->CallVoidMethod(env, (jobject) client_data, 
        mPopupMenuPeerIDs.destroyNativeWidgetAfterGettingTreeLock);

    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * Class:     sun_awt_motif_MPopupMenuPeer
 * Method:    createMenu
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MPopupMenuPeer_createMenu
  (JNIEnv *env, jobject this, jobject parent)
{
    struct ComponentData *wdata;
    struct MenuData *mdata;
    struct FontData *fdata;
    char *ctitle;
    int argc;
    Arg args[10];
    Pixel bg;
    Pixel fg;
    XmFontList fontlist;
    XmString mfstr;
    jobject font;
    jobject target;
    jobject targetFont;
    jobject label;
    jboolean IsMultiFont;
    jboolean tearOff;
    jobject globalRef = (*env)->NewGlobalRef(env, this);

    (*env)->SetLongField(env, this,
			 mMenuItemPeerIDs.jniGlobalRef, (jlong) globalRef);
  

    AWT_LOCK();

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    target =
      (*env)->GetObjectField(env, this, mMenuItemPeerIDs.target);
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env, parent, mComponentPeerIDs.pData);

    if (wdata == 0 || JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    mdata = ZALLOC(MenuData);
    if (mdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    (*env)->SetLongField(env, this, mMenuItemPeerIDs.pData, (jlong) mdata);

    /*
     * Why are these different?
     */
    font = JNU_CallMethodByName(env, NULL, target, "getFont_NoClientCode",
				"()Ljava/awt/Font;").l;
    targetFont =
      (*env)->GetObjectField(env, target, menuComponentIDs.font);
    if (!JNU_IsNull(env, targetFont) &&
        (fdata = awtJNI_GetFontData(env, targetFont, 0)) != 0) {
        IsMultiFont = awtJNI_IsMultiFont(env, targetFont);
    } else {
        IsMultiFont = awtJNI_IsMultiFont(env, font);
    }

    label = (*env)->GetObjectField(env, target, menuItemIDs.label);
    if (JNU_IsNull(env, label)) {
        mfstr = XmStringCreateSimple("");
        ctitle = "";
    } else {
        if (IsMultiFont) {
            mfstr = awtJNI_MakeMultiFontString(env, label, font);
        } else {
            ctitle = (char *) JNU_GetStringPlatformChars(env, label, NULL);
        }
    }

    XtVaGetValues(wdata->widget, XmNbackground, &bg, NULL);
    XtVaGetValues(wdata->widget, XmNforeground, &fg, NULL);

    argc = 0;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNforeground, fg);
    argc++;
    tearOff = (*env)->GetBooleanField(env, target, menuIDs.tearOff);
    if (tearOff) {
        XtSetArg(args[argc], XmNtearOffModel, XmTEAR_OFF_ENABLED);
        argc++;
    }
    if (!JNU_IsNull(env, targetFont)
        && (fdata = awtJNI_GetFontData(env, targetFont, 0)) != 0) {
        if (IsMultiFont) {
            fontlist = awtJNI_GetFontList(env, targetFont);
        } else {
            fontlist = XmFontListCreate(fdata->xfont, "labelFont");
        }

        XtSetArg(args[argc], XmNfontList, fontlist);
        argc++;
    } else {
        if (IsMultiFont) {
            fontlist = awtJNI_GetFontList(env, font);
            XtSetArg(args[argc], XmNfontList, fontlist);
            argc++;
        }
    }

    XtSetArg(args[argc], XmNvisual, defaultConfig->awt_visInfo.visual);
    argc++;
    if (IsMultiFont) {
        mdata->itemData.comp.widget = XmCreatePopupMenu(wdata->widget,
                                                        "",
                                                        args,
                                                        argc);
    } else {
        mdata->itemData.comp.widget = XmCreatePopupMenu(wdata->widget,
                                                        ctitle,
                                                        args,
                                                        argc);
    }

    if (!JNU_IsNull(env, label)) {
        if (IsMultiFont) {
            XtVaCreateManagedWidget("",
                                    xmLabelWidgetClass,
                                    mdata->itemData.comp.widget,
                                    XmNfontList, fontlist,
                                    XmNlabelString, mfstr,
                                    XmNbackground, bg,
                                    XmNforeground, fg,
                                    NULL);
            XmStringFree(mfstr);
        } else {
            XmString xmstr = XmStringCreateLocalized(ctitle);

            XtVaCreateManagedWidget(ctitle,
                                    xmLabelWidgetClass,
                                    mdata->itemData.comp.widget,
                                    XmNlabelString, xmstr,
                                    XmNbackground, bg,
                                    XmNforeground, fg,
                                    NULL);
            XmStringFree(xmstr);
            JNU_ReleaseStringPlatformChars(env, label, (const char *) ctitle);
        }
        /* Create separator */
        XtVaCreateManagedWidget("",
                                xmSeparatorWidgetClass,
                                mdata->itemData.comp.widget,
                                XmNbackground, bg,
                                XmNforeground, fg,
                                NULL);
    }
    if (tearOff) {
        Widget tearOffWidget = XmGetTearOffControl(mdata->itemData.comp.widget);

        XtVaSetValues(tearOffWidget,
                      XmNbackground, bg,
                      XmNforeground, fg,
                      XmNhighlightColor, fg,
                      NULL);
    }
    mdata->comp.widget = mdata->itemData.comp.widget;

    if (!JNU_IsNull(env, targetFont)) {
        XmFontListFree(fontlist);
    }
    XtSetSensitive(mdata->comp.widget,
      ((*env)->GetBooleanField(env, target, menuItemIDs.enabled) ?
       True : False));

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MPopupMenuPeer
 * Method:    pShow
 * Signature: (Ljava/awt/Event;IILsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MPopupMenuPeer_pShow
  (JNIEnv *env, jobject this, jobject event, jint x, jint y, jobject origin)
{
    struct MenuData *mdata;
    struct ComponentData *wdata;
    XButtonEvent *bevent;
    XButtonEvent *newEvent = NULL;
    jlong data;

    AWT_LOCK();

    mdata = (struct MenuData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);
    if (mdata == 0 || JNU_IsNull(env, event)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    /* If the raw x event is not available, then we must use an unfortunate
     * round-trip call to XTranslateCoordiates to get the root coordinates.
     */
    data = (*env)->GetLongField(env, event, eventIDs.data);
    if (data == 0 || ((XEvent *) data)->type != ButtonPress) {
        int rx, ry;
        Window root, win;

        wdata = (struct ComponentData *)
	  (*env)->GetLongField(env, origin, mComponentPeerIDs.pData);

        if (!XtIsRealized(wdata->widget)) {
            JNU_ThrowInternalError(env, "widget not visible on screen");
            AWT_UNLOCK();
            return;
        }
        root = RootWindowOfScreen(XtScreen(wdata->widget));
        XTranslateCoordinates(awt_display,
                              XtWindow(wdata->widget),
                              root,
                              (int) x, (int) y,
                              &rx, &ry,
                              &win);
        /*
                printf("translated coords %d,%d to root %d,%d\n", x, y, rx, ry);
        */

        newEvent = (XButtonEvent *) malloc(sizeof(XButtonEvent));
        newEvent->type = ButtonPress;
        newEvent->display = awt_display;
        newEvent->window = XtWindow(wdata->widget);
        newEvent->x = (int) x;
        newEvent->y = (int) y;
        newEvent->x_root = rx;
        newEvent->y_root = ry;
        bevent = newEvent;

    } else {
        bevent = (XButtonEvent *) data;
    }

    XtAddCallback(XtParent(mdata->comp.widget), XtNpopdownCallback,
                  Popup_popdownCB,
		  (XtPointer)
		    (*env)->GetLongField(env,
					 this,
					 mMenuItemPeerIDs.jniGlobalRef));

    XmMenuPosition(mdata->comp.widget, bevent);
    XtManageChild(mdata->comp.widget);

    if (newEvent) {
        free((void *) newEvent);
    }
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MPopupMenuPeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MPopupMenuPeer_pDispose
  (JNIEnv *env, jobject this)
{
    struct MenuData *mdata;

    AWT_LOCK();

    mdata = (struct MenuData *)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.pData);

    if (mdata == 0) {
        AWT_UNLOCK();
        return;
    }
    XtUnmanageChild(mdata->comp.widget);
    awt_util_consumeAllXEvents(mdata->comp.widget);
    XtDestroyWidget(mdata->comp.widget);
    free((void *) mdata);
    (*env)->SetLongField(env, this, mMenuItemPeerIDs.pData, 0);

    awtJNI_DeleteGlobalMenuRef(env, this);

    AWT_UNLOCK();
}
