/*
 * @(#)awt_ScrollPane.c	1.28 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "java_awt_Insets.h"
#include "java_awt_ScrollPane.h"
#include "java_awt_Adjustable.h"
#include "java_awt_event_AdjustmentEvent.h"
#include "sun_awt_motif_MScrollPanePeer.h"
#include "sun_awt_motif_MComponentPeer.h"

#include "awt_Component.h"

#include <jni.h>
#include <jni_util.h>

extern struct MComponentPeerIDs mComponentPeerIDs;

/* fieldIDs for ScrollPane fields that may be accessed from C */
static struct ScrollPaneIDs {
    jfieldID scrollbarDisplayPolicy;
} scrollPaneIDs;

/*
 * Class:     java_awt_ScrollPane
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   ScrollPane.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_java_awt_ScrollPane_initIDs
  (JNIEnv *env, jclass cls)
{
    scrollPaneIDs.scrollbarDisplayPolicy =
      (*env)->GetFieldID(env, cls, "scrollbarDisplayPolicy", "I");
}

/* fieldIDs for MScrollPanePeer fields that may be accessed from C */
static struct MScrollPanePeerIDs {
    jfieldID dragInProgress;
} mScrollPanePeerIDs;

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   MScrollPanePeer.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollPanePeer_initIDs
  (JNIEnv *env, jclass cls)
{
    mScrollPanePeerIDs.dragInProgress =
      (*env)->GetFieldID(env, cls, "dragInProgress", "Z");
}

static void 
dump_scroll_attrs(Widget scrollbar)
{
    unsigned char orient;
    int value, size, incr, pIncr, max, min;

    XtVaGetValues(scrollbar,
                  XmNvalue, &value,
                  XmNincrement, &incr,
                  XmNpageIncrement, &pIncr,
                  XmNsliderSize, &size,
                  XmNmaximum, &max,
                  XmNminimum, &min,
                  XmNorientation, &orient,
                  NULL);

    jio_fprintf(stdout, "%s: min=%d max=%d slider-size=%d incr=%d pageIncr=%d value = %d\n",
                orient == XmVERTICAL ? "VSB" : "HSB", min, max, size,
                incr, pIncr, value);
}

static int 
getScrollType(int reason)
{
    switch (reason) {
        case XmCR_DRAG:
        case XmCR_VALUE_CHANGED:
        case XmCR_TO_TOP:
        case XmCR_TO_BOTTOM:
            return java_awt_event_AdjustmentEvent_TRACK;
        case XmCR_PAGE_INCREMENT:
            return java_awt_event_AdjustmentEvent_BLOCK_INCREMENT;
        case XmCR_PAGE_DECREMENT:
            return java_awt_event_AdjustmentEvent_BLOCK_DECREMENT;
        case XmCR_INCREMENT:
            return java_awt_event_AdjustmentEvent_UNIT_INCREMENT;
        case XmCR_DECREMENT:
            return java_awt_event_AdjustmentEvent_UNIT_DECREMENT;
        default:
            return 0;
    }
}


/*
 * client_data is MScrollPanePeer instance
 */
static void 
ScrollPane_scrollV(Widget w, XtPointer client_data, XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    XmScrollBarCallbackStruct *scrolldata = (XmScrollBarCallbackStruct *) call_data;
    jobject this = (jobject) client_data;

    if (scrolldata->reason == XmCR_DRAG) {
        (*env)->SetBooleanField(env, this, mScrollPanePeerIDs.dragInProgress,
				JNI_TRUE);
    }
    JNU_CallMethodByName(env, NULL, this, "scrolledVertical"
			 ,"(II)V"
			 ,scrolldata->value
			 ,getScrollType(scrolldata->reason));
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    if (scrolldata->reason == XmCR_VALUE_CHANGED) {
        (*env)->SetBooleanField(env, this, mScrollPanePeerIDs.dragInProgress,
				JNI_FALSE);
    }
}

/*
 * client_data is MScrollPanePeer instance
 */
static void 
ScrollPane_scrollH(Widget w, XtPointer client_data, XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    XmScrollBarCallbackStruct *scrolldata = (XmScrollBarCallbackStruct *) call_data;
    jobject this = (jobject) client_data;

    if (scrolldata->reason == XmCR_DRAG) {
        (*env)->SetBooleanField(env, this, mScrollPanePeerIDs.dragInProgress,
				JNI_TRUE);
    }
    JNU_CallMethodByName(env, NULL, this, "scrolledHorizontal"
			 ,"(II)V"
			 ,scrolldata->value
			 ,getScrollType(scrolldata->reason));
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    if (scrolldata->reason == XmCR_VALUE_CHANGED) {
        (*env)->SetBooleanField(env, this, mScrollPanePeerIDs.dragInProgress,
				JNI_FALSE);
    }
}

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollPanePeer_create
  (JNIEnv *env, jobject this, jobject parent) 
{
    int argc;
    Arg args[40];
    struct ComponentData *wdata;
    struct ComponentData *sdata;
    jobject target;
    Pixel bg;
    Widget vsb, hsb;
    jint sbDisplay;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);

    if (JNU_IsNull(env, target) || wdata == NULL) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    sdata = ZALLOC(ComponentData);
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong) sdata);

    if (sdata == NULL) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    XtVaGetValues(wdata->widget, XmNbackground, &bg, NULL);
    argc = 0;

    sbDisplay =
      (*env)->GetIntField(env, target, scrollPaneIDs.scrollbarDisplayPolicy);

    XtSetArg(args[argc], XmNuserData, (XtPointer) globalRef);
    argc++;

    if (sbDisplay == java_awt_ScrollPane_SCROLLBARS_NEVER) {
        sdata->widget = XtCreateWidget("ScrolledWindowClipWindow",
                                       xmManagerWidgetClass, wdata->widget,
                                       args, argc);

    } else {
        if (sbDisplay == java_awt_ScrollPane_SCROLLBARS_ALWAYS) {
            XtSetArg(args[argc], XmNscrollBarDisplayPolicy, XmSTATIC);
            argc++;
        } else {
            XtSetArg(args[argc], XmNscrollBarDisplayPolicy, XmAS_NEEDED);
            argc++;
        }

        XtSetArg(args[argc], XmNscrollingPolicy, XmAUTOMATIC);
        argc++;
        XtSetArg(args[argc], XmNvisualPolicy, XmCONSTANT);
        argc++;
        XtSetArg(args[argc], XmNspacing, 0);
        argc++;

        sdata->widget = XmCreateScrolledWindow(wdata->widget, "scroller", args, argc);

        XtVaGetValues(sdata->widget,
                      XmNverticalScrollBar, &vsb,
                      XmNhorizontalScrollBar, &hsb,
                      NULL);

        if (vsb != NULL) {
            XtAddCallback(vsb, XmNincrementCallback, ScrollPane_scrollV, (XtPointer) globalRef);
            XtAddCallback(vsb, XmNdecrementCallback, ScrollPane_scrollV, (XtPointer) globalRef);
            XtAddCallback(vsb, XmNpageIncrementCallback, ScrollPane_scrollV, (XtPointer) globalRef);
            XtAddCallback(vsb, XmNpageDecrementCallback, ScrollPane_scrollV, (XtPointer) globalRef);
            XtAddCallback(vsb, XmNtoTopCallback, ScrollPane_scrollV, (XtPointer) globalRef);
            XtAddCallback(vsb, XmNtoBottomCallback, ScrollPane_scrollV, (XtPointer) globalRef);
            XtAddCallback(vsb, XmNvalueChangedCallback, ScrollPane_scrollV, (XtPointer) globalRef);
            XtAddCallback(vsb, XmNdragCallback, ScrollPane_scrollV, (XtPointer) globalRef);

            XtVaSetValues(vsb, XmNhighlightThickness, 0, NULL);
        }
        if (hsb != NULL) {
            XtAddCallback(hsb, XmNincrementCallback, ScrollPane_scrollH, (XtPointer) globalRef);
            XtAddCallback(hsb, XmNdecrementCallback, ScrollPane_scrollH, (XtPointer) globalRef);
            XtAddCallback(hsb, XmNpageIncrementCallback, ScrollPane_scrollH, (XtPointer) globalRef);
            XtAddCallback(hsb, XmNpageDecrementCallback, ScrollPane_scrollH, (XtPointer) globalRef);
            XtAddCallback(hsb, XmNtoTopCallback, ScrollPane_scrollH, (XtPointer) globalRef);
            XtAddCallback(hsb, XmNtoBottomCallback, ScrollPane_scrollH, (XtPointer) globalRef);
            XtAddCallback(hsb, XmNvalueChangedCallback, ScrollPane_scrollH, (XtPointer) globalRef);
            XtAddCallback(hsb, XmNdragCallback, ScrollPane_scrollH, (XtPointer) globalRef);

            XtVaSetValues(hsb, XmNhighlightThickness, 0, NULL);
        }
    }

    XtSetMappedWhenManaged(sdata->widget, False);
    XtManageChild(sdata->widget);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    pSetScrollChild
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollPanePeer_pSetScrollChild
  (JNIEnv *env, jobject this, jobject child) 
{
    struct ComponentData *cdata;
    struct ComponentData *sdata;
    jobject target;

    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    if (JNU_IsNull(env, child) || JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,child,mComponentPeerIDs.pData);
    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (sdata == 0 || cdata == 0 || sdata->widget == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    if ((*env)->GetIntField(env, target, scrollPaneIDs.scrollbarDisplayPolicy)
	== java_awt_ScrollPane_SCROLLBARS_NEVER) {
        /* Do Nothing */
    } else {
        XmScrolledWindowSetAreas(sdata->widget, NULL, NULL, cdata->widget);
        /*
          XtInsertEventHandler(cdata->widget, StructureNotifyMask, FALSE, 
          child_event_handler, sdata->widget, XtListHead);
        */
    }

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    pSetIncrement
 * Signature: (III)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollPanePeer_pSetIncrement
  (JNIEnv *env, jobject this, jint orient, jint incrType, jint incr) 
{
    struct ComponentData *sdata;
    Widget scrollbar = NULL;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (sdata == 0 || sdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (!XtIsSubclass(sdata->widget, xmScrolledWindowWidgetClass)) {
        AWT_UNLOCK();
        return;
    }
    if (orient == java_awt_Adjustable_VERTICAL) {
        XtVaGetValues(sdata->widget,
                      XmNverticalScrollBar, &scrollbar,
                      NULL);
    } else {
        XtVaGetValues(sdata->widget,
                      XmNhorizontalScrollBar, &scrollbar,
                      NULL);
    }

    if (scrollbar != NULL) {
        if (incrType == sun_awt_motif_MScrollPanePeer_UNIT_INCREMENT) {
            XtVaSetValues(scrollbar,
                          XmNincrement, (XtArgVal) incr,
                          NULL);

        } else {
            /* BLOCK_INCREMENT */
            XtVaSetValues(scrollbar,
                          XmNpageIncrement, (XtArgVal) incr,
                          NULL);
        }
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    pGetScrollbarSpace
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MScrollPanePeer_pGetScrollbarSpace
  (JNIEnv *env, jobject this, jint orient) 
{
    struct ComponentData *sdata;
    Widget scrollbar;
    Dimension thickness = 0;
    Dimension space = 0;
    Dimension highlight = 0;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (sdata == 0 || sdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    if (orient == java_awt_Adjustable_VERTICAL) {
        XtVaGetValues(sdata->widget,
                      XmNverticalScrollBar, &scrollbar,
                      XmNspacing, &space,
                      NULL);
        XtVaGetValues(scrollbar,
                      XmNwidth, &thickness,
                      XmNhighlightThickness, &highlight,
                      NULL);
    } else {
        XtVaGetValues(sdata->widget,
                      XmNhorizontalScrollBar, &scrollbar,
                      XmNspacing, &space,
                      NULL);
        XtVaGetValues(scrollbar,
                      XmNheight, &thickness,
                      XmNhighlightThickness, &highlight,
                      NULL);
    }

    AWT_UNLOCK();
    return (jint) (thickness + space + 2 * highlight);
}

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    pGetBlockIncrement
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MScrollPanePeer_pGetBlockIncrement
  (JNIEnv *env, jobject this, jint orient) 
{
    struct ComponentData *sdata;
    Widget scrollbar;
    int pageIncr = 0;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (sdata == 0 || sdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    if (orient == java_awt_Adjustable_VERTICAL) {

        XtVaGetValues(sdata->widget,
                      XmNverticalScrollBar, &scrollbar,
                      NULL);
        XtVaGetValues(scrollbar,
                      XmNpageIncrement, &pageIncr,
                      NULL);
    } else {

        XtVaGetValues(sdata->widget,
                      XmNhorizontalScrollBar, &scrollbar,
                      NULL);
        XtVaGetValues(scrollbar,
                      XmNpageIncrement, &pageIncr,
                      NULL);
    }

    AWT_UNLOCK();
    return (jint) (pageIncr);
}

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    pInsets
 * Signature: (IIII)Ljava/awt/Insets;
 */
JNIEXPORT jobject JNICALL Java_sun_awt_motif_MScrollPanePeer_pInsets
  (JNIEnv *env, jobject this, jint width, jint height, jint childWidth, jint childHeight) 
{
    struct ComponentData *sdata;
    jobject target;
    jobject insets;
    Widget hsb, vsb;
    Dimension hsbThickness, hsbHighlight, hsbSpace, vsbThickness, vsbHighlight, vsbSpace, space, border, shadow, hMargin, vMargin;
    unsigned char placement;
    Boolean hsbVisible, vsbVisible;
    jint sbDisplay;
    int top, left, right, bottom;
    jclass clazz;
    jmethodID mid;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    if (JNU_IsNull(env, target) || sdata == 0 || sdata->widget == 0)
    {
        JNU_ThrowNullPointerException(env, "sdata is NULL");
        AWT_UNLOCK();
        return 0;
    }
    sbDisplay =
      (*env)->GetIntField(env, target, scrollPaneIDs.scrollbarDisplayPolicy);

    /* REMIND: investigate caching these values rather than querying for
     * them each time.
     */

    if (sbDisplay == java_awt_ScrollPane_SCROLLBARS_NEVER) {
        XtVaGetValues(sdata->widget,
                      XmNshadowThickness, &shadow,
                      NULL);
        space = border = hMargin = vMargin = 0;

    } else {
        XtVaGetValues(sdata->widget,
                      XmNverticalScrollBar, &vsb,
                      XmNhorizontalScrollBar, &hsb,
                      XmNscrollBarPlacement, &placement,
                      XmNspacing, &space,
                      XmNshadowThickness, &shadow,
                      XmNscrolledWindowMarginHeight, &vMargin,
                      XmNscrolledWindowMarginWidth, &hMargin,
                      XmNborderWidth, &border,
                      NULL);

        XtVaGetValues(vsb,
                      XmNwidth, &vsbThickness,
                      XmNhighlightThickness, &vsbHighlight,
                      NULL);

        XtVaGetValues(hsb,
                      XmNheight, &hsbThickness,
                      XmNhighlightThickness, &hsbHighlight,
                      NULL);

        hsbSpace = hsbThickness + space + hsbHighlight;
        vsbSpace = vsbThickness + space + vsbHighlight;

/*
  XtVaGetValues(clip, 
  XmNwidth, &clipw, XmNheight, &cliph,
  XmNx, &clipx, XmNy, &clipy, 
  NULL);
  printf("insets: spacing=%d shadow=%d swMarginH=%d swMarginW=%d border=%d ; \
  vsb=%d vsbHL=%d ; hsb=%d hsbHL=%d ; %dx%d ->clip=%d,%d %dx%d\n", 
  space, shadow, vMargin, hMargin, border,
  vsbThickness, vsbHighlight, hsbThickness, hsbHighlight,
  w, h, clipx, clipy, clipw, cliph);
*/
    }

    /* We unfortunately have to use the size parameters to determine
     * whether or not "as needed" scrollbars are currently present or
     * not because we can't necessarily rely on getting valid geometry
     * values straight from the Motif widgets until they are mapped. :(
     */
    switch (sbDisplay) {
        case java_awt_ScrollPane_SCROLLBARS_NEVER:
            vsbVisible = hsbVisible = FALSE;
            break;

        case java_awt_ScrollPane_SCROLLBARS_ALWAYS:
            vsbVisible = hsbVisible = TRUE;
            break;

        case java_awt_ScrollPane_SCROLLBARS_AS_NEEDED:
        default:
            vsbVisible = hsbVisible = FALSE;
            if (childWidth > width - 2 * shadow) {
                hsbVisible = TRUE;
            }
            if (childHeight > height - 2 * shadow) {
                vsbVisible = TRUE;
            }
            if (!hsbVisible && vsbVisible && childWidth > width - 2 * shadow - vsbSpace) {
                hsbVisible = TRUE;
            } else if (!vsbVisible && hsbVisible && childHeight > height - 2 * shadow - hsbSpace) {
                vsbVisible = TRUE;
            }
    }

    top = bottom = shadow + vMargin;
    left = right = shadow + hMargin;

    if (sbDisplay != java_awt_ScrollPane_SCROLLBARS_NEVER) {
        switch (placement) {
            case XmBOTTOM_RIGHT:
                bottom += (hsbVisible ? hsbSpace : (vsbVisible ? vsbHighlight : 0));
                right += (vsbVisible ? vsbSpace : (hsbVisible ? hsbHighlight : 0));
                top += (vsbVisible ? vsbHighlight : 0);
                left += (hsbVisible ? hsbHighlight : 0);
                break;

            case XmBOTTOM_LEFT:
                bottom += (hsbVisible ? hsbSpace : (vsbVisible ? vsbHighlight : 0));
                left += (vsbVisible ? hsbSpace : (hsbVisible ? hsbHighlight : 0));
                top += (vsbVisible ? vsbHighlight : 0);
                right += (hsbVisible ? hsbHighlight : 0);
                break;

            case XmTOP_RIGHT:
                top += (hsbVisible ? hsbSpace : (vsbVisible ? vsbHighlight : 0));
                right += (vsbVisible ? vsbSpace : (hsbVisible ? hsbHighlight : 0));
                bottom += (vsbVisible ? vsbHighlight : 0);
                left += (hsbVisible ? hsbHighlight : 0);
                break;

            case XmTOP_LEFT:
                top += (hsbVisible ? hsbSpace : (vsbVisible ? vsbHighlight : 0));
                left += (vsbVisible ? vsbSpace : (hsbVisible ? hsbHighlight : 0));
                bottom += (vsbVisible ? vsbHighlight : 0);
                right += (hsbVisible ? hsbHighlight : 0);
        }
    }
    /* Deadlock prevention:
     * don't hold the toolkit lock while invoking constructor. 
     */
    AWT_UNLOCK();

    clazz = (*env)->FindClass(env, "java/awt/Insets");
    mid = (*env)->GetMethodID(env, clazz, "<init>", "(IIII)V");
    if (mid != 0) {
        insets = (*env)->NewObject(env, clazz, mid,
                                   (jint) top,
                                   (jint) left,
                                   (jint) bottom,
                                   (jint) right);

    }
    /* This should catch both method not found and error exceptions */
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    if (JNU_IsNull(env, insets)) {
        JNU_ThrowNullPointerException(env, "NullPointerException: insets constructor failed");
    }
    return insets;
}

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    setScrollPosition
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollPanePeer_setScrollPosition
  (JNIEnv *env, jobject this, jint x, jint y)
{
    struct ComponentData *sdata;
    jobject target;
    Widget hsb, vsb;
    int size, incr, pIncr;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    if (JNU_IsNull(env, target) || sdata == 0 || sdata->widget == 0)
    {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if ((*env)->GetIntField(env, target, scrollPaneIDs.scrollbarDisplayPolicy)
	== java_awt_ScrollPane_SCROLLBARS_NEVER) {
        WidgetList children;
        int numChildren;

        XtVaGetValues(sdata->widget,
                      XmNchildren, &children,
                      XmNnumChildren, &numChildren,
                      NULL);

        if (numChildren < 1) {
            JNU_ThrowNullPointerException(env, "NullPointerException");
            AWT_UNLOCK();
            return;
        }
        XtMoveWidget(children[0], (int) -x, (int) -y);
    } else {
        XtVaGetValues(sdata->widget,
                      XmNhorizontalScrollBar, &hsb,
                      XmNverticalScrollBar, &vsb,
                      NULL);

        if (vsb) {
            XtVaGetValues(vsb,
                          XmNincrement, &incr,
                          XmNpageIncrement, &pIncr,
                          XmNsliderSize, &size,
                          NULL);
            XmScrollBarSetValues(vsb, (int) y, size, incr, pIncr, TRUE);
        }
        if (hsb) {
            XtVaGetValues(hsb,
                          XmNincrement, &incr,
                          XmNpageIncrement, &pIncr,
                          XmNsliderSize, &size,
                          NULL);
            XmScrollBarSetValues(hsb, (int) x, size, incr, pIncr, TRUE);
        }
    }
    AWT_FLUSH_UNLOCK();
}

