/*
 * @(#)awt_Scrollbar.c	1.41 98/09/22
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "java_awt_Scrollbar.h"
#include "sun_awt_motif_MScrollbarPeer.h"
#include "sun_awt_motif_MComponentPeer.h"

#include "awt_Component.h"

#include <jni.h>
#include <jni_util.h>
#include "multi_font.h"
#include <Xm/ScrollBarP.h>  /* ScrollBar private header. */

extern struct MComponentPeerIDs mComponentPeerIDs;

/* fieldIDs for Scrollbar fields that may be accessed from C */
static struct ScrollbarIDs {
    jfieldID orientation;
    jfieldID visibleAmount;
    jfieldID lineIncrement;
    jfieldID pageIncrement;
    jfieldID value;
    jfieldID minimum;
    jfieldID maximum;
} scrollbarIDs;

/*
 * Class:     java_awt_ScrollBar
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   Scrollbar.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_java_awt_Scrollbar_initIDs
  (JNIEnv *env, jclass cls)
{
    scrollbarIDs.orientation =
      (*env)->GetFieldID(env, cls, "orientation", "I");
    scrollbarIDs.visibleAmount =
      (*env)->GetFieldID(env, cls, "visibleAmount", "I");
    scrollbarIDs.lineIncrement =
      (*env)->GetFieldID(env, cls, "lineIncrement", "I");
    scrollbarIDs.pageIncrement =
      (*env)->GetFieldID(env, cls, "pageIncrement", "I");
    scrollbarIDs.value =
      (*env)->GetFieldID(env, cls, "value", "I");
    scrollbarIDs.minimum =
      (*env)->GetFieldID(env, cls, "minimum", "I");
    scrollbarIDs.maximum =
      (*env)->GetFieldID(env, cls, "maximum", "I");
}

/* Remove the ScrollBar widget's continuous scrolling timeout handler
   on a ButtonRelease to prevent the continuous scrolling that would
   occur if a timeout expired after the ButtonRelease. This must
   prevent a bug in Motif.
*/
static void
ButtonReleaseHandler(Widget w,
                     XtPointer data,
                     XEvent *event,
                     Boolean *cont)
{
  /* Remove the timeout handler. */
#define END_TIMER         (1<<2)
  XmScrollBarWidget sbw = (XmScrollBarWidget) w;
  if (sbw->scrollBar.timer != 0) {
    XtRemoveTimeOut( sbw->scrollBar.timer );
    sbw->scrollBar.timer = (XtIntervalId)NULL;
    sbw->scrollBar.flags |= END_TIMER;
  }
}

/*
 * client_data is MScrollbarPeer instance
 */
static void 
Scrollbar_lineUp(Widget w,
                 XtPointer client_data,
                 XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    int new_value = ((XmScrollBarCallbackStruct *) call_data)->value;

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "lineUp", "(I)V", 
			 new_value);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * client_data is MScrollbarPeer instance
 */
static void 
Scrollbar_lineDown(Widget w,
                   XtPointer client_data,
                   XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    int new_value = ((XmScrollBarCallbackStruct *) call_data)->value;

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "lineDown", "(I)V",
			 new_value);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * client_data is MScrollbarPeer instance
 */
static void 
Scrollbar_pageUp(Widget w,
                 XtPointer client_data,
                 XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    int new_value = ((XmScrollBarCallbackStruct *) call_data)->value;

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "pageUp", "(I)V",
			 new_value);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * client_data is MScrollbarPeer instance
 */
static void 
Scrollbar_pageDown(Widget w,
                   XtPointer client_data,
                   XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    int new_value = ((XmScrollBarCallbackStruct *) call_data)->value;

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "pageDown", "(I)V",
			 new_value);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

static void 
Scrollbar_dragAbsolute(Widget w,
                       XtPointer client_data,
                       XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    int new_value = ((XmScrollBarCallbackStruct *) call_data)->value;

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "dragAbsolute",
			 "(I)V", new_value);
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * Class:     sun_awt_motif_MScrollbarPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollbarPeer_create
  (JNIEnv *env, jobject this, jobject parent) 
{
    int argc;
    Arg args[40];
    struct ComponentData *wdata;
    struct ComponentData *sdata;
    Dimension d;
    jobject target;
    Pixel bg;
    int visibleAmount;
    int lineIncrement;
    int pageIncrement;
    int value;
    int minimum;
    int maximum;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);

    if (JNU_IsNull(env, target) || wdata == NULL) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();

        return;
    }
    sdata = ZALLOC(ComponentData);
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong) sdata);

    if (sdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    XtVaGetValues(wdata->widget, XmNbackground, &bg, NULL);

    switch ((*env)->GetIntField(env, target, scrollbarIDs.orientation)) {
        case java_awt_Scrollbar_HORIZONTAL:
            XtVaGetValues(wdata->widget, XmNwidth, &d, NULL);
            argc = 0;
            XtSetArg(args[argc], XmNorientation, XmHORIZONTAL);
            argc++;
            break;

        case java_awt_Scrollbar_VERTICAL:
            XtVaGetValues(wdata->widget, XmNheight, &d, NULL);
            argc = 0;
            XtSetArg(args[argc], XmNorientation, XmVERTICAL);
            argc++;
            break;

        default:
            free(sdata);
            JNU_ThrowIllegalArgumentException(env, "bad scrollbar orientation");
            AWT_UNLOCK();
            return;
    }

    XtSetArg(args[argc], XmNrecomputeSize, False);
    argc++;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNx, 0);
    argc++;
    XtSetArg(args[argc], XmNy, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginTop, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginBottom, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginLeft, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginRight, 0);
    argc++;
    XtSetArg(args[argc], XmNuserData, (XtPointer) globalRef);
    argc++;

    visibleAmount =
      (int) (*env)->GetIntField(env, target, scrollbarIDs.visibleAmount);
    if (visibleAmount > 0) {
	lineIncrement =
	  (int) (*env)->GetIntField(env, target, scrollbarIDs.lineIncrement);
        pageIncrement =
	  (int) (*env)->GetIntField(env, target, scrollbarIDs.pageIncrement);
        value = (int) (*env)->GetIntField(env, target, scrollbarIDs.value);
        minimum = (int) (*env)->GetIntField(env, target, scrollbarIDs.minimum);
        maximum = (int) (*env)->GetIntField(env, target, scrollbarIDs.maximum);

        XtSetArg(args[argc], XmNincrement, lineIncrement);
        argc++;
        XtSetArg(args[argc], XmNpageIncrement, pageIncrement);
        argc++;
        XtSetArg(args[argc], XmNsliderSize, visibleAmount);
        argc++;
        XtSetArg(args[argc], XmNvalue, value);
        argc++;
        XtSetArg(args[argc], XmNminimum, minimum);
        argc++;
        XtSetArg(args[argc], XmNmaximum, maximum);
        argc++;
    }
    sdata->widget = XmCreateScrollBar(wdata->widget, "scrollbar", args, argc);

    XtSetMappedWhenManaged(sdata->widget, False);
    XtManageChild(sdata->widget);
    XtAddCallback(sdata->widget,
                  XmNdecrementCallback,
                  Scrollbar_lineUp,
                  (XtPointer) globalRef);
    XtAddCallback(sdata->widget,
                  XmNincrementCallback,
                  Scrollbar_lineDown,
                  (XtPointer) globalRef);
    XtAddCallback(sdata->widget,
                  XmNpageDecrementCallback,
                  Scrollbar_pageUp,
                  (XtPointer) globalRef);
    XtAddCallback(sdata->widget,
                  XmNpageIncrementCallback,
                  Scrollbar_pageDown,
                  (XtPointer) globalRef);
    XtAddCallback(sdata->widget,
                  XmNdragCallback,
                  Scrollbar_dragAbsolute,
                  (XtPointer) globalRef);
    XtAddCallback(sdata->widget,
                  XmNtoTopCallback,
                  Scrollbar_dragAbsolute,
                  (XtPointer) globalRef);
    XtAddCallback(sdata->widget,
                  XmNtoBottomCallback,
                  Scrollbar_dragAbsolute,
                  (XtPointer) globalRef);

    XtAddEventHandler(sdata->widget, ButtonReleaseMask,
                      False, ButtonReleaseHandler, NULL);
    AWT_UNLOCK();
}


/*
 * Class:     sun_awt_motif_MScrollbarPeer
 * Method:    setValues
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollbarPeer_setValues
  (JNIEnv *env,
 jobject this,
 jint value,
 jint visible,
 jint minimum,
 jint maximum) 
{
    struct ComponentData *sdata;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (sdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    /* pass in visible for sliderSize since Motif will calculate the */
    /* slider's size for us. */
    XtVaSetValues(sdata->widget,
                  XmNminimum, minimum,
                  XmNmaximum, maximum,
                  XmNvalue, value,
                  XmNsliderSize, visible,
                  NULL);

    AWT_FLUSH_UNLOCK();
}


/*
 * Class:     sun_awt_motif_MScrollbarPeer
 * Method:    setLineIncrement
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollbarPeer_setLineIncrement
  (JNIEnv *env, jobject this, jint value) 
{
    struct ComponentData *sdata;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (sdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(sdata->widget,
                  XmNincrement, value,
                  NULL);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MScrollbarPeer
 * Method:    setPageIncrement
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MScrollbarPeer_setPageIncrement
  (JNIEnv *env, jobject this, jint value)
{
    struct ComponentData *sdata;

    AWT_LOCK();

    sdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (sdata == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(sdata->widget,
                  XmNpageIncrement, value,
                  NULL);
    AWT_FLUSH_UNLOCK();
}
