/*
 * @(#)awt_Selection.c	1.25 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_p.h"
#include "java_awt_datatransfer_Transferable.h"
#include "java_awt_datatransfer_DataFlavor.h"
#include "sun_awt_motif_X11Selection.h"
#include "sun_awt_motif_X11Clipboard.h"
#include <X11/Intrinsic.h>

#include <jni.h>
#include <jni_util.h>

static Atom CLIPBOARD;
static Atom TARGETS;

static jobject selections[10];
static selectionCount = 0;

enum {
    STRING = 0,
    TEXT = 1,
    COMPOUND_TEXT = 2
};

static Atom targetList[24];

static Widget selection_w = NULL;

Boolean selectionProcessed = FALSE;
Boolean gotData = FALSE;


/* fieldIDs for X11Selection fields that may be accessed from C */
static struct X11SelectionIDs {
    jfieldID holder;
    jfieldID atom;
    jfieldID contents;
    jfieldID targetArray;
    jfieldID data;
} x11SelectionIDs;

/*
 * Class:     sun_awt_motif_MScrollPanePeer
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for
   X11Selection.java to initialize the fieldIDs for fields that may
   be accessed from C */

JNIEXPORT void JNICALL Java_sun_awt_motif_X11Selection_initIDs
  (JNIEnv *env, jclass cls)
{
    x11SelectionIDs.holder = (*env)->
      GetFieldID(env, cls, "holder","Lsun/awt/motif/X11SelectionHolder;");
    x11SelectionIDs.atom = (*env)->GetFieldID(env, cls, "atom", "I");
    x11SelectionIDs.contents = (*env)->
      GetFieldID(env, cls, "contents","Ljava/awt/datatransfer/Transferable;");
    x11SelectionIDs.targetArray = (*env)->
      GetFieldID(env, cls, "targetArray","[I");
    x11SelectionIDs.data = (*env)->
      GetFieldID(env, cls, "data","Ljava/lang/Object;");
}

static jobject 
getX11Selection(JNIEnv * env, Atom atom)
{
    int i;

    for (i = 0; i < selectionCount; i++) {
        if ((*env)->GetIntField(env, selections[i], x11SelectionIDs.atom)
	    == atom) {
            return selections[i];
        }
    }

    return NULL;
}

Boolean 
awtJNI_isSelectionOwner(JNIEnv * env, char *sel_str)
{
    Atom selection;
    jobject x11sel;

    selection = XInternAtom(awt_display, sel_str, False);

    x11sel = getX11Selection(env, selection);
    if (!JNU_IsNull(env, x11sel)) {
        jobject holder;

        holder = (*env)->GetObjectField(env, x11sel, x11SelectionIDs.holder);
        if (!JNU_IsNull(env, holder)) {
            return TRUE;
        }
    }
    return FALSE;
}

static void losingSelectionOwnership(Widget w, Atom * selection);

void 
awtJNI_notifySelectionLost(JNIEnv * env, char *sel_str)
{
    Atom selection;

    selection = XInternAtom(awt_display, sel_str, False);
    losingSelectionOwnership(NULL, &selection);
}

static int 
getTargetsForFlavors(JNIEnv * env, jobjectArray flavors, Atom ** targets_ret)
{
    int num_flavors = (*env)->GetArrayLength(env, flavors);
    Atom buf[100];
    int count = 0;
    int i, j;

    for (i = 0; i < num_flavors; i++) {
        Boolean exists = False;
        Atom target;
        jobject flavor;

        flavor = (*env)->GetObjectArrayElement(env, flavors, i);

        target = (Atom) (*env)->GetIntField(env, flavor, x11SelectionIDs.atom);

        for (j = 0; j < count && !exists; j++) {
            if (buf[j] == target) {
                exists = True;
            }
        }
        if (target != 0 && !exists) {
            buf[count++] = target;
            /*
             * CONPOUND_TEXT target is also for the stringFlavor 
             */
            if (target == targetList[STRING]) {
                buf[count++] = targetList[COMPOUND_TEXT];
            }
        }
    }

    if (count > 0) {
        *targets_ret = (Atom *) malloc(count * sizeof(Atom));
        for (i = 0; i < count; i++) {
            (*targets_ret)[i] = buf[i];
        }
    }
    return count;
}

static Boolean 
provideSelectionData(Widget w, Atom * selection, Atom * target, Atom * type,
                     XtPointer * value, unsigned long *length, int *format)
{
    jobject this;
    jobject contents = 0;
    jobject flavor = NULL;
    jobjectArray flavors;
    jbyteArray data;
    Atom *targets;
    int i, j, num_flavors, count;
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if ((*env)->PushLocalFrame(env, 16) < 0)
        return FALSE;

    this = getX11Selection(env, *selection);
    if (JNU_IsNull(env, this)) {
        (*env)->PopLocalFrame(env, 0);
        return FALSE;
    }
    contents = (*env)->GetObjectField(env, this, x11SelectionIDs.contents);
    if (JNU_IsNull(env, contents)) {
        (*env)->PopLocalFrame(env, 0);
        return FALSE;
    }
    //SECURITY: OK to call this on privileged thread - peer does
    //          not call into client code
    flavors = JNU_CallMethodByName(env, NULL, contents,
				   "getTransferDataFlavors",
				   "()[Ljava/awt/datatransfer/DataFlavor;").l;

    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
    num_flavors = (*env)->GetArrayLength(env, flavors);
    count = getTargetsForFlavors(env, flavors, &targets);

    if (*target == TARGETS) {
        /*
          printf("ConvertRequest: TARGETS\n");
        */
        if (count == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return FALSE;
        }
        *type = XA_ATOM;
        *format = 32;
        *value = (XtPointer) targets;
        *length = count;
        (*env)->PopLocalFrame(env, 0);
        return TRUE;
    }
    for (i = 0; i < count; i++) {
        if (*target == targets[i]) {
            for (j = 0; j < num_flavors; j++) {
                jobject f;
                Atom atom;

                f = (*env)->GetObjectArrayElement(env, flavors, j);

                atom =
		  (Atom) (*env)->GetIntField(env, f, x11SelectionIDs.atom);

                if (atom == *target) {
                    flavor = f;
                    break;
                }
                if (atom == targetList[STRING] &&
                    *target == targetList[COMPOUND_TEXT]) {
                    flavor = f;
                    break;
                }
                (*env)->DeleteLocalRef(env, f);
            }

            if (flavor == NULL) {
	        (*env)->PopLocalFrame(env, 0);
                return FALSE;
	    }

            if (*target == targetList[STRING] ||
                *target == targetList[COMPOUND_TEXT]) {
                jstring strData;
                char *str;
                char *data;
                int len;

                /*
                  printf("ConvertRequest: STRING\n");
                */
                //SECURITY: OK to call this on privileged thread - peer does
                //          not call into client code
                strData =
                    JNU_CallMethodByName(env,
					 NULL,
					 this,
					 "getStringDataFromOwner",
					 "(Ljava/awt/datatransfer/DataFlavor;)Ljava/lang/String;",
					 flavor).l;

                if ((*env)->ExceptionOccurred(env)) {
                    (*env)->ExceptionDescribe(env);
                    (*env)->ExceptionClear(env);
                }
                if (JNU_IsNull(env, strData)) {
		    (*env)->PopLocalFrame(env, 0);
                    return FALSE;
                }
                str = (char *) JNU_GetStringPlatformChars(env, strData, NULL);
                if (*target == targetList[STRING]) {
                    len = strlen(str);
                    data = (char *) malloc(len * sizeof(char));

                    strcpy(data, str);

                    *type = XA_STRING;
                    *format = 8;
                    *value = (XtPointer) data;
                    *length = len;
		    (*env)->PopLocalFrame(env, 0);
                    return TRUE;
                } else {
                    XTextProperty text_prop_return;

                    if (XmbTextListToTextProperty(awt_display, &str, 1, XCompoundTextStyle,
                                                  &text_prop_return) == Success) {

                        *type = text_prop_return.encoding;
                        *format = 8;
                        *value = (XtPointer) (text_prop_return.value);
                        *length = strlen((char *) text_prop_return.value);
			(*env)->PopLocalFrame(env, 0);
                        return TRUE;
                    }
                }
                JNU_ReleaseStringPlatformChars(env, strData, (const char *) str);

            } else if (*target == targetList[TEXT]) {
                /*
                  printf("ConvertRequest: TEXT\n");
                */
            } else {
                /*
                  printf("ConvertRequest: UNKNOWN\n");
                */
            }
        }
    }

    (*env)->PopLocalFrame(env, 0);
    return FALSE;
}

static char *data;

static void 
transferDone(Widget w, Atom * selection, Atom * target)
{
    free(data);
}

static void 
losingSelectionOwnership(Widget w, Atom * selection)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = getX11Selection(env, *selection);

    //SECURITY: OK to call this on privileged thread - peer does
    //          not call into client code
    JNU_CallMethodByName(env, NULL, this, "lostSelectionOwnership", "()V");
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

static void 
getSelectionTargets(Widget w, XtPointer client_data, Atom * selection, Atom * type,
                    XtPointer value, unsigned long *length, int *format)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject) client_data;
    jintArray targetArray;
    int count = *length;

    if (*type == TARGETS) {
        Atom *targets = (Atom *) value;

        if (count > 0) {
	    if ((*env)->EnsureLocalCapacity(env, 1) < 0)
	        return;
            targetArray = (*env)->NewIntArray(env, count);
            if (JNU_IsNull(env, targetArray)) {
                return;
            }
            (*env)->SetIntArrayRegion(env, targetArray, 0, count,
                                      (jint *) targets);

            (*env)->SetObjectField(env, this,
				   x11SelectionIDs.targetArray, targetArray);
	    (*env)->DeleteLocalRef(env, targetArray);
        }
    }
    selectionProcessed = TRUE;
}

static void 
getSelectionData(Widget w, XtPointer client_data, Atom * selection, Atom * type,
                 XtPointer value, unsigned long *length, int *format)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject) client_data;
    int count = *length;
    long *atoms;
    int i;
    jstring jstr;

    if (*type == targetList[STRING]) {
        char *str = (char *) value;
	if ((*env)->EnsureLocalCapacity(env, 1) < 0)
	    return;
        jstr = JNU_NewStringPlatform(env, (const char *) value);
        (*env)->SetObjectField(env, this, x11SelectionIDs.data, jstr);
	(*env)->DeleteLocalRef(env, jstr);

        gotData = TRUE;
    } else if (*type == targetList[COMPOUND_TEXT]) {
        XTextProperty text_prop;
        char **list_return;
        int count_return;

        text_prop.value = (unsigned char *) value;
        text_prop.encoding = *type;
        text_prop.format = 8;
        text_prop.nitems = count;
        XmbTextPropertyToTextList(awt_display, &text_prop,
                                  &list_return, &count_return);

        if (list_return != NULL && count_return > 0) {
            jstr = JNU_NewStringPlatform(env, (const char *) list_return[0]);
            (*env)->SetObjectField(env, this, x11SelectionIDs.data, jstr);
	    (*env)->DeleteLocalRef(env, jstr);
            XFreeStringList(list_return);
            gotData = TRUE;
        }
    }
    selectionProcessed = TRUE;
}

static int 
WaitForSelectionEvent(void *data)
{
    return selectionProcessed;
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11Selection_init
  (JNIEnv *env, jclass this) 
{
    AWT_LOCK();

    selection_w = XtAppCreateShell("AWTSelection", "XApplication",
                                   topLevelShellWidgetClass,
                                   awt_display,
                                   NULL, 0);

    XtSetMappedWhenManaged(selection_w, False);
    XtRealizeWidget(selection_w);

    TARGETS = XInternAtom(awt_display, "TARGETS", False);

    targetList[STRING] = XInternAtom(awt_display, "STRING", False);
    targetList[TEXT] = XInternAtom(awt_display, "TEXT", False);
    targetList[COMPOUND_TEXT] = XInternAtom(awt_display, "COMPOUND_TEXT", False);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    getAtomForTarget
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_X11Selection_getAtomForTarget
  (JNIEnv *env, jclass this, jstring targetString) 
{
    Atom target;
    char *target_str;

    AWT_LOCK();

    if (JNU_IsNull(env, targetString)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return -1;
    }
    target_str = (char *) JNU_GetStringPlatformChars(env, targetString, NULL);
    target = XInternAtom(awt_display, target_str, False);
    JNU_ReleaseStringPlatformChars(env, targetString, (const char *) target_str);

    AWT_UNLOCK();
    return (jint) target;
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    registerTargetForFlavor
 * Signature: (Ljava/awt/datatransfer/DataFlavor;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11Selection_registerTargetForFlavor
  (JNIEnv *env, jclass this, jobject flavor, jstring targetString) 
{
    Atom target;
    char *ts;

    AWT_LOCK();

    if (JNU_IsNull(env, targetString) || JNU_IsNull(env, flavor)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    ts = (char *) JNU_GetStringPlatformChars(env, targetString, NULL);
    target = XInternAtom(awt_display, ts, False);
    JNU_ReleaseStringPlatformChars(env, targetString, (const char *) ts);

    (*env)->SetIntField(env, flavor, x11SelectionIDs.atom, (jint) target);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    create
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11Selection_create
  (JNIEnv *env, jobject this, jstring name) 
{
    Atom selection;
    char *sel_str;

    AWT_LOCK();

    if (JNU_IsNull(env, name)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    sel_str = (char *) JNU_GetStringPlatformChars(env, name, NULL);

    selection = XInternAtom(awt_display, sel_str, False);
    JNU_ReleaseStringPlatformChars(env, name, (const char *) sel_str);

    (*env)->SetIntField(env, this, x11SelectionIDs.atom, (jint) selection);

    /* The selection object must be a global ref to prevent */
    /* it from being GC'd */
    selections[selectionCount++] = (*env)->NewGlobalRef(env, this);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11Selection_pDispose
  (JNIEnv *env, jobject this) 
{
    int i;
    int mark;

    AWT_LOCK();

    for (i = 0; i < selectionCount; i++) {
        if ((*env)->IsSameObject(env, selections[i], this)) {
            /* We have to release the object so it can be GC'd */
            (*env)->DeleteGlobalRef(env, selections[i]);
            mark = i;
            break;
        }
    }
    for (i = mark + 1; i < selectionCount; i++) {
        selections[i - 1] = selections[i];
    }
    selectionCount--;

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    pGetSelectionOwnership
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_sun_awt_motif_X11Selection_pGetSelectionOwnership
  (JNIEnv *env, jobject this) 
{
    Atom selection;
    Boolean gotit;

    AWT_LOCK();

    selection = (Atom) (*env)->GetIntField(env, this, x11SelectionIDs.atom);

    gotit = XtOwnSelection(selection_w, selection, XtLastTimestampProcessed(awt_display),
                           provideSelectionData, losingSelectionOwnership, transferDone);

    AWT_UNLOCK();
    return (gotit ? 1 : 0);
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    pGetTransferTargets
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11Selection_pGetTransferTargets
  (JNIEnv *env, jobject this) 
{
    Atom selection;
    jobject gRef;

    AWT_LOCK();

    selection = (Atom) (*env)->GetIntField(env, this, x11SelectionIDs.atom);
    selectionProcessed = FALSE;

    /* We may have callbacks from another thread? Better make */
    /* a global ref to be safe. */
    gRef = (*env)->NewGlobalRef(env, this);

    XtGetSelectionValue(selection_w, selection, TARGETS, getSelectionTargets,
                        (XtPointer) gRef, XtLastTimestampProcessed(awt_display));

    AWT_FLUSH_UNLOCK();
    awt_MToolkit_modalWait(WaitForSelectionEvent, NULL);
    (*env)->DeleteGlobalRef(env, gRef);
}

/*
 * Class:     sun_awt_motif_X11Selection
 * Method:    pGetTransferData
 * Signature: ([I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11Selection_pGetTransferData
  (JNIEnv *env, jobject this, jintArray targets)
{
    Atom selection;
    jint *atoms;
    jobject gRef;
    int count;
    int i;

    AWT_LOCK();

    selection = (Atom) (*env)->GetIntField(env, this, x11SelectionIDs.atom);
    gotData = FALSE;

    atoms = (*env)->GetIntArrayElements(env, targets, NULL);
    if(atoms == NULL) {
        AWT_UNLOCK();
        return;
    }

    count = (*env)->GetArrayLength(env, targets);

    /* We may have callbacks from another thread? Better make */
    /* a global ref to be safe. */
    gRef = (*env)->NewGlobalRef(env, this);

    for (i = 0; i < count && !gotData; i++) {
        selectionProcessed = FALSE;

        XtGetSelectionValue(selection_w, selection, (Atom) atoms[i], getSelectionData,
                            (XtPointer) gRef, XtLastTimestampProcessed(awt_display));

        AWT_FLUSH_UNLOCK();
        awt_MToolkit_modalWait(WaitForSelectionEvent, NULL);
        AWT_LOCK();

    }

    (*env)->DeleteGlobalRef(env, gRef);
    (*env)->ReleaseIntArrayElements(env, targets, atoms, JNI_ABORT);

    AWT_UNLOCK();
}
