/*
 * @(#)awt_TextArea.c	1.49 98/09/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "canvas.h"
#include "java_awt_TextArea.h"
#include "java_awt_Cursor.h"
#include "java_awt_Component.h"
#include "java_awt_Color.h"
#include "java_awt_AWTEvent.h"
#include "java_awt_Font.h"
#include "sun_awt_motif_MTextAreaPeer.h"
#include "sun_awt_motif_MComponentPeer.h"
#include <X11/cursorfont.h>

#include "awt_Component.h"
#include "awt_Cursor.h"

#include <jni.h>
#include <jni_util.h>
#include "multi_font.h"

extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct CursorIDs cursorIDs;
extern struct X11CustomCursorIDs x11CustomCursorIDs;

/* fieldIDs for TextArea fields that may be accessed from C */
static struct TextAreaIDs {
    jfieldID scrollbarVisibility;
} textAreaIDs;

/*
 * Class:     java_awt_TextArea
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for TextArea.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_TextArea_initIDs
  (JNIEnv *env, jclass cls)
{
    textAreaIDs.scrollbarVisibility =
      (*env)->GetFieldID(env, cls, "scrollbarVisibility", "I");
}

extern void Text_handlePaste(Widget w, XtPointer client_data, XEvent * event,
                             Boolean * cont);

extern void Text_valueChanged(Widget w, XtPointer client_data, XtPointer call_data);

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_create
  (JNIEnv *env, jobject this, jobject parent) 
{
    struct TextAreaData *tdata;
    Arg args[30];
    int argc;
    struct ComponentData *wdata;
    jobject target;
    Pixel bg;
    long sbVisibility;
    Boolean wordWrap, hsb, vsb;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);
    if (wdata == NULL) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    tdata = ZALLOC(TextAreaData);
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong) tdata);

    if (!tdata) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    XtVaGetValues(wdata->widget, XmNbackground, &bg, NULL);

    sbVisibility = (*env)->GetIntField(env, target,
				       textAreaIDs.scrollbarVisibility);
    switch (sbVisibility) {
        case java_awt_TextArea_SCROLLBARS_NONE:
            wordWrap = True;
            hsb = False;
            vsb = False;
            break;

        case java_awt_TextArea_SCROLLBARS_VERTICAL_ONLY:
            wordWrap = True;
            hsb = False;
            vsb = True;
            break;

        case java_awt_TextArea_SCROLLBARS_HORIZONTAL_ONLY:
            wordWrap = False;
            hsb = True;
            vsb = False;
            break;

        case java_awt_TextArea_SCROLLBARS_BOTH:
            wordWrap = False;
            hsb = True;
            vsb = True;
            break;
    }

    argc = 0;
    XtSetArg(args[argc], XmNrecomputeSize, False);
    argc++;
    XtSetArg(args[argc], XmNx, 0);
    argc++;
    XtSetArg(args[argc], XmNy, 0);
    argc++;
    XtSetArg(args[argc], XmNbackground, bg);
    argc++;
    XtSetArg(args[argc], XmNeditMode, XmMULTI_LINE_EDIT);
    argc++;
    XtSetArg(args[argc], XmNwordWrap, wordWrap);
    argc++;
    XtSetArg(args[argc], XmNscrollHorizontal, hsb);
    argc++;
    XtSetArg(args[argc], XmNscrollVertical, vsb);
    argc++;
    XtSetArg(args[argc], XmNmarginHeight, 2);
    argc++;
    XtSetArg(args[argc], XmNmarginWidth, 2);
    argc++;
    XtSetArg(args[argc], XmNuserData, (XtPointer) globalRef);
    argc++;

    tdata->txt = XmCreateScrolledText(wdata->widget, "textA",
                                      args, argc);
    tdata->comp.widget = XtParent(tdata->txt);
    XtSetMappedWhenManaged(tdata->comp.widget, False);
    XtManageChild(tdata->txt);
    XtManageChild(tdata->comp.widget);

    XtAddCallback(tdata->txt,
                  XmNvalueChangedCallback,
                  Text_valueChanged,
                  (XtPointer) globalRef);

    XtAddEventHandler(tdata->txt, FocusChangeMask,
                      True, awt_canvas_event_handler, globalRef);

    XtInsertEventHandler(tdata->txt,
                         KeyPressMask,
                         False, Text_handlePaste, (XtPointer) globalRef,
                         XtListHead);

    awt_addWidget(tdata->txt, tdata->comp.widget, globalRef,
                  java_awt_AWTEvent_KEY_EVENT_MASK |
                  java_awt_AWTEvent_MOUSE_EVENT_MASK |
                  java_awt_AWTEvent_MOUSE_MOTION_EVENT_MASK);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    getExtraWidth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextAreaPeer_getExtraWidth
  (JNIEnv *env, jobject this) 
{
    struct TextAreaData *tdata;
    Dimension spacing, shadowThickness, textMarginWidth, sbWidth;
    Widget verticalScrollBar;

    AWT_LOCK();

    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    XtVaGetValues(tdata->txt, XmNmarginWidth, &textMarginWidth, NULL);
    XtVaGetValues(tdata->comp.widget,
                  XmNspacing, &spacing,
                  XmNverticalScrollBar, &verticalScrollBar,
                  NULL);
    if (verticalScrollBar) {
        /* Assumption:  shadowThickness same for scrollbars and text area */
        XtVaGetValues(verticalScrollBar,
                      XmNwidth, &sbWidth,
                      XmNshadowThickness, &shadowThickness,
                      NULL);
    } else {
        sbWidth = 0;
        shadowThickness = 0;
    }

    AWT_UNLOCK();

    return (jint) (sbWidth + spacing + 2 * textMarginWidth + 4 * shadowThickness);
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    getExtraHeight
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextAreaPeer_getExtraHeight
  (JNIEnv *env, jobject this) 
{
    struct TextAreaData *tdata;
    Dimension spacing, shadowThickness, textMarginHeight, sbHeight;
    Widget horizontalScrollBar;

    AWT_LOCK();

    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    XtVaGetValues(tdata->txt, XmNmarginHeight, &textMarginHeight, NULL);
    XtVaGetValues(tdata->comp.widget,
                  XmNspacing, &spacing,
                  XmNhorizontalScrollBar, &horizontalScrollBar,
                  NULL);
    if (horizontalScrollBar) {
        /* Assumption:  shadowThickness same for scrollbars and text area */
        XtVaGetValues(horizontalScrollBar,
                      XmNheight, &sbHeight,
                      XmNshadowThickness, &shadowThickness,
                      NULL);
    } else {
        sbHeight = 0;
        shadowThickness = 0;
    }

    AWT_UNLOCK();

    return (jint) (sbHeight + spacing + 2 * textMarginHeight + 4 * shadowThickness);
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    setTextBackground
 * Signature: (Ljava/awt/Color;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_setTextBackground
  (JNIEnv *env, jobject this, jobject c) 
{
    struct TextAreaData *tdata;
    Pixel color;

    AWT_LOCK();

    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0 || JNU_IsNull(env, c)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    color = awtJNI_GetColor(env, c);
    XtVaSetValues(tdata->txt,
                  XmNbackground, color,
                  NULL);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    pSetEditable
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_pSetEditable
  (JNIEnv *env, jobject this, jboolean editable) 
{
    struct TextAreaData *tdata;

    AWT_LOCK();

    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(tdata->txt,
                  XmNeditable, (editable ? True : False),
                  XmNcursorPositionVisible, (editable ? True : False),
                  NULL);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    select
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_select
  (JNIEnv *env, jobject this, jint start, jint end) 
{
    struct TextAreaData *tdata;

    AWT_LOCK();

    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XmTextSetSelection(tdata->txt, (XmTextPosition) start, (XmTextPosition) end, 0);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    getSelectionStart
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextAreaPeer_getSelectionStart
  (JNIEnv *env, jobject this) 
{
    struct TextAreaData *tdata;
    XmTextPosition start, end, pos;

    AWT_LOCK();

    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    if (XmTextGetSelectionPosition(tdata->txt, &start, &end)) {
        pos = start;
    } else {
        pos = XmTextGetCursorPosition(tdata->txt);
    }
    AWT_UNLOCK();

    return (jint) pos;
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    getSelectionEnd
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextAreaPeer_getSelectionEnd
  (JNIEnv *env, jobject this) 
{
    struct TextAreaData *tdata;
    XmTextPosition start, end, pos;

    AWT_LOCK();

    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    if (XmTextGetSelectionPosition(tdata->txt, &start, &end)) {
        pos = end;
    } else {
        pos = XmTextGetCursorPosition(tdata->txt);
    }
    AWT_UNLOCK();

    return (jint) pos;
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    setText
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_setText
  (JNIEnv *env, jobject this, jstring txt) 
{
    struct TextAreaData *tdata;
    char *cTxt;
    jobject font = awtJNI_GetFont(env, this);

    if (JNU_IsNull(env, txt)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();

    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    cTxt = (char *) JNU_GetStringPlatformChars(env, txt, NULL);

    if (!cTxt) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(tdata->txt, XmNvalue, cTxt, NULL);

    if (cTxt) {
        JNU_ReleaseStringPlatformChars(env, txt, cTxt);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    getText
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_sun_awt_motif_MTextAreaPeer_getText
  (JNIEnv *env, jobject this) 
{
    struct TextAreaData *tdata;
    char *cTxt;
    jstring rval;
    jobject font = awtJNI_GetFont(env, this);

    AWT_LOCK();

    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env,this, mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    cTxt = XmTextGetString(tdata->txt);

    rval = JNU_NewStringPlatform(env, (const char *) cTxt);

    XtFree(cTxt);

    AWT_UNLOCK();

    return rval;
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    insert
 * Signature: (Ljava/lang/String;I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_insert
  (JNIEnv *env, jobject this, jstring txt, jint pos) 
{
    struct TextAreaData *tdata;
    char *cTxt;
    jobject font = awtJNI_GetFont(env, this);

    if (JNU_IsNull(env, txt)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    cTxt = (char *) JNU_GetStringPlatformChars(env, txt, NULL);

    if (!cTxt) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XmTextInsert(tdata->txt, (XmTextPosition) pos, cTxt);

    if (cTxt) {
        JNU_ReleaseStringPlatformChars(env, txt, cTxt);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    replaceRange
 * Signature: (Ljava/lang/String;II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_replaceRange
  (JNIEnv *env, jobject this, jstring txt, jint start, jint end) 
{
    struct TextAreaData *tdata;
    char *cTxt;
    jobject font = awtJNI_GetFont(env, this);

    if (JNU_IsNull(env, txt)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();
    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    cTxt = (char *) JNU_GetStringPlatformChars(env, txt, NULL);

    if (!cTxt) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XmTextReplace(tdata->txt,
                  (XmTextPosition) start,
                  (XmTextPosition) end,
                  cTxt);

    if (cTxt) {
        JNU_ReleaseStringPlatformChars(env, txt, cTxt);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    setFont
 * Signature: (Ljava/awt/Font;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_setFont
  (JNIEnv *env, jobject this, jobject f) 
{
    struct TextAreaData *tdata;
    struct FontData *fdata;
    XmFontList fontlist;
    char *err;
    XmFontListEntry fontentry;

    if (JNU_IsNull(env, f)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }
    AWT_LOCK();

    fdata = awtJNI_GetFontData(env, f, &err);
    if (fdata == 0) {
        JNU_ThrowInternalError(env, err);
        AWT_UNLOCK();
        return;
    }
    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (awtJNI_IsMultiFont(env, f)) {
        if (fdata->xfs == NULL) {
            fdata->xfs = awtJNI_MakeFontSet(env, f);
        }
        if (fdata->xfs != NULL) {
            fontentry = XmFontListEntryCreate("labelFont",
                                              XmFONT_IS_FONTSET,
                                              (XtPointer) (fdata->xfs));
            fontlist = XmFontListAppendEntry(NULL, fontentry);
	    /*
	     * Some versions of motif have a bug in 
	     * XmFontListEntryFree() which causes it to free more than it
	     * should.  Use XtFree() instead.  See O'Reilly's
	     * Motif Reference Manual for more information.
	     */
	    XmFontListEntryFree(&fontentry);

        } else {
            fontlist = XmFontListCreate(fdata->xfont, "labelFont");
        }
    } else {
        fontlist = XmFontListCreate(fdata->xfont, "labelFont");
    }

    if (fontlist != NULL) {
        Dimension textw, texth;
        Dimension w, h;

        XtVaGetValues(tdata->txt,
                      XmNwidth, &textw,
                      XmNheight, &texth,
                      NULL);
        XtVaGetValues(tdata->comp.widget,
                      XmNwidth, &w,
                      XmNheight, &h,
                      NULL);

        /* Must set width/height when we set the font, else
         * Motif resets the text to a single row.
         */
        XtVaSetValues(tdata->txt,
                      XmNfontList, fontlist,
                      XmNwidth, textw,
                      XmNheight, texth,
                      NULL);
        XtVaSetValues(tdata->comp.widget,
                      XmNwidth, w,
                      XmNheight, h,
                      NULL);

        XmFontListFree(fontlist);
    } else {
        JNU_ThrowNullPointerException(env, "NullPointerException");
    }

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    setCaretPosition
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_setCaretPosition
  (JNIEnv *env, jobject this, jint pos) 
{
    struct TextAreaData *tdata;

    AWT_LOCK();

    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XmTextSetCursorPosition(tdata->txt, (XmTextPosition) pos);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    getCaretPosition
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextAreaPeer_getCaretPosition
  (JNIEnv *env, jobject this)
{
    struct TextAreaData *tdata;
    XmTextPosition pos;

    AWT_LOCK();

    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->txt == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    pos = XmTextGetCursorPosition(tdata->txt);

    AWT_UNLOCK();

    return (jint) pos;
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    pShow
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_pShow2
  (JNIEnv *env, jobject this)
{
    struct TextAreaData *tdata;

    AWT_LOCK();
    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    awt_util_show(tdata->comp.widget);
    if ((*env)->GetIntField(env, this, mComponentPeerIDs.cursorSet) == 0) {
    	(*env)->SetIntField(env, this, mComponentPeerIDs.cursorSet, 
           (jint)awt_util_setCursor(tdata->txt, tdata->comp.cursor));
    }

    AWT_FLUSH_UNLOCK();
}



/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    pMakeCursorVisible
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_pMakeCursorVisible
  (JNIEnv *env, jobject this)
{
    struct TextAreaData *tdata;

    AWT_LOCK();
    tdata = (struct TextAreaData *) (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    (*env)->SetIntField(env, this, mComponentPeerIDs.cursorSet, 
           (jint)awt_util_setCursor(tdata->txt, tdata->comp.cursor));

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextAreaPeer
 * Method:    setCursor
 * Signature: (L/java/awt/Cursor;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextAreaPeer_setCursor
  (JNIEnv *env, jobject this, jobject cursor)
{
    Cursor xcursor;
    struct TextAreaData         *tdata;
    int cursorType;

    AWT_LOCK();
    tdata = (struct TextAreaData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->comp.widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    cursorType = (*env)->GetIntField(env, cursor, cursorIDs.type);

    if (cursorType == java_awt_Cursor_CUSTOM_CURSOR) {
        xcursor = (Cursor)
	   (*env)->GetLongField(env, cursor, x11CustomCursorIDs.pData);
    } else {
        switch (cursorType) {
          case java_awt_Cursor_DEFAULT_CURSOR:
              cursorType = -1;
              break;
          case java_awt_Cursor_CROSSHAIR_CURSOR:
              cursorType = XC_crosshair;
              break;
          case java_awt_Cursor_TEXT_CURSOR:
              cursorType = XC_xterm;
              break;
          case java_awt_Cursor_WAIT_CURSOR:
              cursorType = XC_watch;
              break;
          case java_awt_Cursor_SW_RESIZE_CURSOR:
              cursorType = XC_bottom_left_corner;
              break;
          case java_awt_Cursor_NW_RESIZE_CURSOR:
              cursorType = XC_top_left_corner;
              break;
          case java_awt_Cursor_SE_RESIZE_CURSOR:
              cursorType = XC_bottom_right_corner;
              break;
          case java_awt_Cursor_NE_RESIZE_CURSOR:
              cursorType = XC_top_right_corner;
              break;
          case java_awt_Cursor_S_RESIZE_CURSOR:
              cursorType = XC_bottom_side;
              break;
          case java_awt_Cursor_N_RESIZE_CURSOR:
              cursorType = XC_top_side;
              break;
          case java_awt_Cursor_W_RESIZE_CURSOR:
              cursorType = XC_left_side;
              break;
          case java_awt_Cursor_E_RESIZE_CURSOR:
              cursorType = XC_right_side;
              break;
          case java_awt_Cursor_HAND_CURSOR:
              cursorType = XC_hand2;
              break;
          case java_awt_Cursor_MOVE_CURSOR:
              cursorType = XC_fleur;
              break;
          default:
              AWT_UNLOCK();
              return;
        }

        if (cursorType != -1) {
            xcursor = XCreateFontCursor(awt_display, cursorType);
        } else {
            xcursor = None;
        }
    }

    if (tdata->comp.cursor != 0 && tdata->comp.cursor != None
        && !tdata->comp.customCursor) {
        XFreeCursor(awt_display, tdata->comp.cursor);
    }

    tdata->comp.customCursor = (cursorType == java_awt_Cursor_CUSTOM_CURSOR);
    tdata->comp.cursor = xcursor;

    (*env)->SetIntField(env, this, mComponentPeerIDs.cursorSet,
	      (jint) awt_util_setCursor(tdata->txt,tdata->comp.cursor));

    AWT_FLUSH_UNLOCK();
}
