/*
 * @(#)awt_TextField.c	1.66 98/09/22
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <Xm/VirtKeys.h>

#include "awt_p.h"
#include "java_awt_TextField.h"
#include "java_awt_Color.h"
#include "java_awt_AWTEvent.h"
#include "java_awt_Font.h"
#include "java_awt_Canvas.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "sun_awt_motif_MCanvasPeer.h"
#include "sun_awt_motif_MTextFieldPeer.h"

#include "awt_Component.h"

#include "multi_font.h"
#include <jni.h>
#include <jni_util.h>
#include <Xm/TextFP.h>  /* Motif TextField private header. */


#define ECHO_BUFFER_LEN 1024
static int echoContextID = -1;
static Boolean echoContextIDInit = FALSE;

extern struct MComponentPeerIDs mComponentPeerIDs;

static struct TextFielIDs {
    jfieldID echoChar;
} textFieldIDs;

/*
 * Class:     java_awt_TextField
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for TextField.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_TextField_initIDs
  (JNIEnv *env, jclass cls)
{
    textFieldIDs.echoChar =
      (*env)->GetFieldID(env, cls, "echoChar", "C");
}

static void 
echoChar(Widget text_w, XtPointer c, XmTextVerifyCallbackStruct * cbs)
{
    int len;
    char *val;
    struct DPos *dp;
    int ret;

    ret = XFindContext(XtDisplay(text_w), text_w, echoContextID, &dp);
    if ((ret != 0) || (dp == NULL)) { 
        /* no context found or DPos is NULL - shouldn't happen */
        return;
    }
    val = (char *) (dp->data);
    len = strlen(val);
    if (cbs->text->ptr == NULL) {
        if (cbs->text->length == 0 && cbs->startPos == 0) {
            val[0] = 0;
            return;
        } else if (cbs->startPos == (len - 1)) {
            /* handle deletion */
            cbs->endPos = strlen(val);
            val[cbs->startPos] = 0;
            return;
        } else {
            /* disable deletes anywhere but at the end */
            cbs->doit = False;
            return;
        }
    }
    if (cbs->startPos != len) {
        /* disable "paste" or inserts into the middle */
        cbs->doit = False;
        return;
    }
    /* append the value typed in */
    if ((cbs->endPos + cbs->text->length) > ECHO_BUFFER_LEN) {
        val = realloc(val, cbs->endPos + cbs->text->length + 10);
    }
    strncat(val, cbs->text->ptr, cbs->text->length);
    val[cbs->endPos + cbs->text->length] = 0;

    /* modify the output to be the echo character */
    for (len = 0; len < cbs->text->length; len++) {
        cbs->text->ptr[len] = (char) (int) c;
    }
}

/*
 * Event handler used by both TextField/TextArea to correctly process
 * cut/copy/paste keys such that interaction with our own
 * clipboard mechanism will work properly.
 *
 * client_data is MTextFieldPeer instance
 */
void 
Text_handlePaste(Widget w, XtPointer client_data, XEvent * event, Boolean * cont)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    KeySym keysym;
    Modifiers mods;

    /* Any event handlers which take peer instance pointers as
     * client_data should check to ensure the widget has not been
     * marked as destroyed as a result of a dispose() call on the peer
     * (which can result in the peer instance pointer already haven
     * been gc'd by the time this event is processed)
     */
    if (event->type != KeyPress || w->core.being_destroyed)
        return;

    XtTranslateKeycode(event->xkey.display, (KeyCode) event->xkey.keycode,
                       event->xkey.state, &mods, &keysym);

    /* Should be a temporary fix for 4052132 if a cleaner fix is found later */
    if ((event->xkey.state & ControlMask) && (keysym == 'v' || keysym == 'V'))
        keysym = osfXK_Paste;
    if ((event->xkey.state & ShiftMask) && (keysym == osfXK_Insert))
        keysym = osfXK_Paste;
	
    switch (keysym) {
        case osfXK_Paste:
            /* If we own the selection, then paste the data directly */
            if (awtJNI_isSelectionOwner(env, "CLIPBOARD")) {
                JNU_CallMethodByName(env, NULL, (jobject) client_data, 
				     "pasteFromClipboard", "()V");
                if ((*env)->ExceptionOccurred(env)) {
                    (*env)->ExceptionDescribe(env);
                    (*env)->ExceptionClear(env);
                }
                *cont = FALSE;
            }
            break;

        case osfXK_Cut:
        case osfXK_Copy:
            /* For some reason if we own the selection, our loseSelection
             * callback is not automatically called on cut/paste from
             * text widgets.
             */
            if (awtJNI_isSelectionOwner(env, "CLIPBOARD")) {
                awtJNI_notifySelectionLost(env, "CLIPBOARD");
            }
            break;
        default:
            break;
    }
}

/*
 * client_data is MTextFieldPeer instance
 */
void
Text_valueChanged(Widget w, XtPointer client_data, XtPointer call_data)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    JNU_CallMethodByName(env, NULL, (jobject) client_data,
			 "valueChanged", "()V");
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * client_data is MTextFieldPeer instance
 */
static void
TextField_action(Widget w, XtPointer client_data, XmAnyCallbackStruct * s)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    JNU_CallMethodByName(env, NULL, (jobject) client_data, "action", "()V");
    if ((*env)->ExceptionOccurred(env)) {
        (*env)->ExceptionDescribe(env);
        (*env)->ExceptionClear(env);
    }
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_create
  (JNIEnv *env, jobject this, jobject parent) 
{
    struct ComponentData *wdata;
    struct ComponentData *cdata;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();

    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    wdata = (struct ComponentData *)
      (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);
    if (wdata == NULL) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    cdata = ZALLOC(ComponentData);
    if (cdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }
    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong) cdata);

    cdata->widget = XtVaCreateManagedWidget("textfield",
                                            xmTextFieldWidgetClass,
                                            wdata->widget,
                                            XmNrecomputeSize, False,
                                            XmNhighlightThickness, 1,
                                            XmNshadowThickness, 2,
                                            XmNuserData, (XtPointer) globalRef,
                                            NULL);
    XtSetMappedWhenManaged(cdata->widget, False);
    XtAddCallback(cdata->widget,
                  XmNactivateCallback,
                  (XtCallbackProc) TextField_action,
                  (XtPointer) globalRef);
    XtAddCallback(cdata->widget,
                  XmNvalueChangedCallback,
                  (XtCallbackProc) Text_valueChanged,
                  (XtPointer) globalRef);
    XtInsertEventHandler(cdata->widget,
                         KeyPressMask,
                         False, Text_handlePaste, (XtPointer) globalRef,
                         XtListHead);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    pSetEditable
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_pSetEditable
  (JNIEnv *env, jobject this, jboolean editable) 
{
    struct ComponentData *cdata;

    AWT_LOCK();
    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(cdata->widget,
                  XmNeditable, (editable ? True : False),
                  XmNcursorPositionVisible, (editable ? True : False),
                  NULL);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    select
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_select
  (JNIEnv *env, jobject this, jint start, jint end) 
{
    struct ComponentData *tdata;

    AWT_LOCK();

    tdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XmTextSetSelection(tdata->widget, (XmTextPosition) start, (XmTextPosition) end, 0);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    getSelectionStart
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextFieldPeer_getSelectionStart
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *tdata;
    XmTextPosition start, end, pos;

    AWT_LOCK();

    tdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    if (XmTextGetSelectionPosition(tdata->widget, &start, &end)) {
        pos = start;
    } else {
        pos = XmTextGetCursorPosition(tdata->widget);
    }
    AWT_UNLOCK();

    return pos;
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    getSelectionEnd
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextFieldPeer_getSelectionEnd
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *tdata;
    XmTextPosition start, end, pos;

    AWT_LOCK();

    tdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    if (XmTextGetSelectionPosition(tdata->widget, &start, &end)) {
        pos = end;
    } else {
        pos = XmTextGetCursorPosition(tdata->widget);
    }
    AWT_UNLOCK();

    return pos;

}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    setText
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_setText
  (JNIEnv *env, jobject this, jstring l) 
{
    struct ComponentData *cdata;
    char *cl;
    jobject font;
    jobject target;

    AWT_LOCK();

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    font = awtJNI_GetFont(env, this);

    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (JNU_IsNull(env, l)) {
        cl = "";
    } else {
        /*
         * We use makePlatformCString() to convert unicode to EUC here,
         * although output only components (Label/Button/Menu..)
         * is not using make/allocCString() functions anymore.
         * Because Motif TextFiled widget does not support multi-font
         * compound string.
         */


        cl = (char *) JNU_GetStringPlatformChars(env, l, NULL);
    }

    /* Fix for bug 4084454 : setText appears in clear */
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    if ((*env)->GetCharField(env, target, textFieldIDs.echoChar) != 0) {
        XtVaSetValues(cdata->widget,
                      XmNvalue, NULL, NULL);
        XmTextFieldInsert(cdata->widget,0,cl);
    }
    else {
        XtVaSetValues(cdata->widget,
                      XmNvalue, cl,
                      NULL);
    }
    XmTextSetCursorPosition(cdata->widget, (XmTextPosition) strlen(cl));

    if (cl && cl != "") {
        JNU_ReleaseStringPlatformChars(env, l, cl);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    insertReplaceText
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_insertReplaceText
  (JNIEnv *env, jobject this, jstring l) 
{
    struct ComponentData *cdata;
    char *cl;
    XmTextPosition start, end;
    jobject font;

    AWT_LOCK();

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
	AWT_UNLOCK();
        return;
    }
    font = awtJNI_GetFont(env, this);

    if (JNU_IsNull(env, l)) {
        cl = "";
    } else {
        /*
         * We use makePlatformCString() to convert unicode to EUC here,
         * although output only components (Label/Button/Menu..)
         * is not using make/allocCString() functions anymore.
         * Because Motif TextFiled widget does not support multi-font
         * compound string.
         */

        cl = (char *) JNU_GetStringPlatformChars(env, l, NULL);
    }

    if (!XmTextGetSelectionPosition(cdata->widget, &start, &end)) {
        start = end = XmTextGetCursorPosition(cdata->widget);
    }
    XmTextReplace(cdata->widget, start, end, cl);

    if (cl && cl != "") {
        JNU_ReleaseStringPlatformChars(env, l, cl);
    }
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    preDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_preDispose
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;
    struct DPos *dp;
    jobject target;
    XmTextFieldWidget tf = 0;
    int ret;

    AWT_LOCK();

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    if ((*env)->GetCharField(env, target, textFieldIDs.echoChar) != 0) {
        ret = XFindContext(XtDisplay(cdata->widget), (cdata->widget),
                           echoContextID, &dp);
        if ((ret == 0) && dp != NULL) {
            if (dp->data) {
                free(dp->data);
            }
            free(dp);
        }
    }

    /* Remove the BrowseScroll timer, because Motif fails to do it in
     * its Destroy function in Xm/TextF.c. Ask Motif group to fix it.
    */
    tf = (XmTextFieldWidget) cdata->widget;
    if (tf->text.select_id) {
       XtRemoveTimeOut(tf->text.select_id);
       tf->text.select_id = (XtIntervalId) NULL;
    }

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    getText
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_sun_awt_motif_MTextFieldPeer_getText
  (JNIEnv *env, jobject this) 
{
    struct ComponentData *cdata;
    char *val;
    struct DPos *dp;
    jobject target;
    jobject font;
    int ret;

    AWT_LOCK();
    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return NULL;
    }

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    font = awtJNI_GetFont(env, this);

    if ((*env)->GetCharField(env, target, textFieldIDs.echoChar) != 0) {
        ret = XFindContext(XtDisplay(cdata->widget), cdata->widget,
                           echoContextID, &dp);
        if ((ret == 0) && (dp != NULL)) {
            val = (char *)(dp->data);
        } else {
            val = "";
        }
    } else {
        XtVaGetValues(cdata->widget, XmNvalue, &val, NULL);
    }
    AWT_UNLOCK();

    return JNU_NewStringPlatform(env, (const char *) val);
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    setEchoChar
 * Signature: (C)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_setEchoChar
  (JNIEnv *env, jobject this, jchar c) 
{
    char *val;
    char *cval;
    struct ComponentData *cdata;
    struct DPos *dp;
    int i, len;
    int ret;

    AWT_LOCK();
    cdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    XtVaGetValues(cdata->widget,
                  XmNvalue, &cval,
                  NULL);

    if (!echoContextIDInit) {
        echoContextID = XUniqueContext();
        echoContextIDInit = TRUE;
    }
    ret = XFindContext(XtDisplay(cdata->widget), (cdata->widget),
                       echoContextID, &dp);

    if (ret != 0) {
        dp = NULL;
    }

    if (dp) {
        /* Fix bug 4124697: cannot change setEchoChar twice on Motif */
        XtRemoveCallback(cdata->widget, XmNmodifyVerifyCallback,
                        (XtCallbackProc) echoChar, (XtPointer)(dp->echoC));
    } else {
        if ((int) strlen(cval) > ECHO_BUFFER_LEN) {
            val = (char *) malloc(strlen(cval) + 1);
        } else {
            val = (char *) malloc(ECHO_BUFFER_LEN + 1);
        }
        if (val == 0) {
            JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
            AWT_UNLOCK();
            return;
        }
        if (cval) {
            strcpy(val, cval);
        } else {
            *val = 0;
        }
        dp = (struct DPos *) malloc(sizeof(struct DPos));

        dp->x = -1;
        dp->data = (void *) val;
    }

    dp->echoC = c;
    len = strlen(cval);
    for (i = 0; i < len; i++) {
        cval[i] = (char) (c);
    }
    XtVaSetValues(cdata->widget,
                  XmNvalue, cval,
                  NULL);

    ret = XSaveContext(XtDisplay(cdata->widget), cdata->widget,
                       echoContextID, dp);
    if (ret == 0) {
        XtAddCallback(cdata->widget, XmNmodifyVerifyCallback,
                      (XtCallbackProc) echoChar, (XtPointer) (int) c);
    }
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    setFont
 * Signature: (Ljava/awt/Font;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_setFont
  (JNIEnv *env, jobject this, jobject f) 
{
    struct ComponentData *tdata;
    struct FontData *fdata;
    XmFontListEntry fontentry;
    XmFontList fontlist;
    char *err;

    AWT_LOCK();
    if (JNU_IsNull(env, f)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    fdata = awtJNI_GetFontData(env, f, &err);
    if (fdata == 0) {
        JNU_ThrowInternalError(env, err);
        AWT_UNLOCK();
        return;
    }
    tdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (tdata == 0 || tdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (awtJNI_IsMultiFont(env, f)) {
        if (fdata->xfs == NULL) {
            fdata->xfs = awtJNI_MakeFontSet(env, f);
        }
        if (fdata->xfs != NULL) {
            fontentry = XmFontListEntryCreate("labelFont",
                                              XmFONT_IS_FONTSET,
                                              (XtPointer) (fdata->xfs));
            fontlist = XmFontListAppendEntry(NULL, fontentry);
	    /*
	     * Some versions of motif have a bug in 
	     * XmFontListEntryFree() which causes it to free more than it
	     * should.  Use XtFree() instead.  See O'Reilly's
	     * Motif Reference Manual for more information.
	     */
	    XmFontListEntryFree(&fontentry);
        } else {
            fontlist = XmFontListCreate(fdata->xfont, "labelFont");
        }
    } else {
        fontlist = XmFontListCreate(fdata->xfont, "labelFont");
    }

    if (fontlist != NULL) {
        XtVaSetValues(tdata->widget, XmNfontList, fontlist, NULL);
        XmFontListFree(fontlist);
    } else {
        JNU_ThrowNullPointerException(env, "NullPointerException");
    }
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    setCaretPosition
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MTextFieldPeer_setCaretPosition
  (JNIEnv *env, jobject this, jint pos) 
{
    struct ComponentData *tdata;

    AWT_LOCK();

    tdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XmTextSetCursorPosition(tdata->widget, (XmTextPosition) pos);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MTextFieldPeer
 * Method:    getCaretPosition
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_motif_MTextFieldPeer_getCaretPosition
  (JNIEnv *env, jobject this)
{
    struct ComponentData *tdata;
    XmTextPosition pos;

    AWT_LOCK();

    tdata = (struct ComponentData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (tdata == 0 || tdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return 0;
    }
    pos = XmTextGetCursorPosition(tdata->widget);
    AWT_UNLOCK();

    return (jint) pos;
}

