/*
 * @(#)awt_Window.c	1.52 98/09/22
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include <X11/Shell.h>
#include <Xm/RowColumn.h>
#include <Xm/AtomMgr.h>
#include <Xm/MenuShell.h>
#include <Xm/MwmUtil.h>
#include "java_awt_Window.h"
#include "java_awt_Insets.h"
#include "sun_awt_motif_MWindowPeer.h"
#include "java_awt_Component.h"
#include "sun_awt_motif_MComponentPeer.h"
#include "canvas.h"
#include "color.h"

#include "awt_Insets.h"
#include "awt_Component.h"
#include "awt_Window.h"

#include <jni.h>
#include <jni_util.h>

extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct ComponentIDs componentIDs;
extern struct InsetsIDs insetsIDs;

/* fieldIDs for Window fields that may be accessed from C */
struct WindowIDs windowIDs;

/*
 * Class:     java_awt_Window
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for Window.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_java_awt_Window_initIDs
  (JNIEnv *env, jclass cls)
{
    windowIDs.warningString = (*env)->GetFieldID(env, cls, "warningString",
						 "Ljava/lang/String;");
}

/* fieldIDs for MWindowPeer fields that may be accessed from C */
static struct MWindowPeerIDs {
    jfieldID insets;
} mWindowPeerIDs;

/*
 * Class:     sun_motif_awt_Window
 * Method:    initIDs
 * Signature: ()V
 */

/* This function gets called from the static initializer for MWindowPeer.java
   to initialize the fieldIDs for fields that may be accessed from C */

JNIEXPORT void JNICALL
Java_sun_awt_motif_MWindowPeer_initIDs
  (JNIEnv *env, jclass cls)
{
    mWindowPeerIDs.insets = (*env)->GetFieldID(env, cls, "insets",
						 "Ljava/awt/Insets;");
}

static void 
awtJNI_ChangeInsets(JNIEnv *env, jobject this, struct FrameData *wdata)
{
    jobject insets;

    if ((*env)->EnsureLocalCapacity(env, 1) < 0)
        return;

    insets = (*env)->GetObjectField(env, this, mWindowPeerIDs.insets);

    if (JNU_IsNull(env, insets)) {
        return;
    }
    (*env)->SetIntField(env, insets, insetsIDs.top, (jint) wdata->top);
    (*env)->SetIntField(env, insets, insetsIDs.left, (jint) wdata->left);
    (*env)->SetIntField(env, insets, insetsIDs.bottom, (jint) wdata->bottom);
    (*env)->SetIntField(env, insets, insetsIDs.right, (jint) wdata->right);

    (*env)->DeleteLocalRef(env, insets);
}

/*
 * data is MWindowPeer instance
 */
static void
Window_event_handler(Widget w, XtPointer data, XEvent * event, Boolean * cont)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject) data;
    struct FrameData *wdata;
    Position x, y;
    Dimension width, height, hoffset, deltaTop;

    /* Any event handlers which take peer instance pointers as
     * client_data should check to ensure the widget has not been
     * marked as destroyed as a result of a dispose() call on the peer
     * (which can result in the peer instance pointer already haven
     * been gc'd by the time this event is processed)
     */
    if (w->core.being_destroyed) {
        return;
    }
    wdata = (struct FrameData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    if (wdata == 0) {
        return;
    }
    switch (event->xany.type) {
        case MapNotify:
            if (wdata->mappedOnce == 0) {
                wdata->mappedOnce = 1;
                JNU_CallMethodByName(env, NULL, this,
				     "makeCursorsVisible", "()V");
                if ((*env)->ExceptionOccurred(env)) {
                    (*env)->ExceptionDescribe(env);
                    (*env)->ExceptionClear(env);
                }
            } else {
                JNU_CallMethodByName(env, NULL, this, "handleDeiconify", "()V");
                if ((*env)->ExceptionOccurred(env)) {
                    (*env)->ExceptionDescribe(env);
                    (*env)->ExceptionClear(env);
                }
            }
            break;

        case UnmapNotify:
            JNU_CallMethodByName(env, NULL, this, "handleIconify", "()V");
            if ((*env)->ExceptionOccurred(env)) {
                (*env)->ExceptionDescribe(env);
                (*env)->ExceptionClear(env);
            }
            break;

	case ReparentNotify:
	    if (wdata->warningWindow) {
	        jobject target;

	        if ((*env)->EnsureLocalCapacity(env, 1) < 0)
		    return;
	        target = (*env)->GetObjectField(env, this,
						mComponentPeerIDs.target);
                XtVaGetValues(wdata->warningWindow, XmNheight, &hoffset, NULL);
	        deltaTop = hoffset - wdata->top;
                wdata->top += deltaTop;
                awtJNI_ChangeInsets(env, this, wdata);

                x = (*env)->GetIntField(env, target, componentIDs.x);
                y = (*env)->GetIntField(env, target, componentIDs.y);
                width = (*env)->GetIntField(env, target, componentIDs.width);
                height = (*env)->GetIntField(env, target, componentIDs.height);

                if ((*env)->GetBooleanField(env, target, componentIDs.isPacked))
		{
                    height = height + deltaTop;
		    (*env)->SetIntField(env, target, componentIDs.height,
					(jint) height);
                }
	        Java_sun_awt_motif_MWindowPeer_pReshape(env, this, 
						    x, y, width, height);
		(*env)->DeleteLocalRef(env, target);
	    }
	    break;

        default:
            break;
    }
} /* Window_event_handler() */

/*
 * Class:     sun_awt_motif_MWindowPeer
 * Method:    create
 * Signature: (Lsun/awt/motif/MComponentPeer;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MWindowPeer_create
  (JNIEnv *env, jobject this, jobject parent) 
{
    Dimension w, h;
    Position x, y;
    Arg args[20];
    int argc;
    int WM;
    struct FrameData *wdata;
    struct FrameData *parentData;
    jobject target;
    jstring warningString;
    jobject globalRef = awtJNI_CreateAndSetGlobalRef(env, this);

    AWT_LOCK();

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    if (JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (JNU_IsNull(env, parent)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    wdata = ZALLOC(FrameData);
    if (wdata == 0) {
        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
        AWT_UNLOCK();
        return;
    }

    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,(jlong) wdata);
    parentData = (struct FrameData *)
      (*env)->GetLongField(env,parent,mComponentPeerIDs.pData);

    x = (*env)->GetIntField(env, target, componentIDs.x);
    y = (*env)->GetIntField(env, target, componentIDs.y);
    w = (*env)->GetIntField(env, target, componentIDs.width);
    h = (*env)->GetIntField(env, target, componentIDs.height);

	/* Force width and height to be greater than zero */
	/* An Xt Error occurs if they are less than 1 */
	if (w<=0) w = 1;
	if (h<=0) h = 1;

    argc = 0;
    XtSetArg(args[argc], XmNtransientFor, parentData->winData.shell);
    argc++;
    XtSetArg(args[argc], XmNsaveUnder, False);
    argc++;
    XtSetArg(args[argc], XmNx, x);
    argc++;
    XtSetArg(args[argc], XmNy, y);
    argc++;
    XtSetArg(args[argc], XmNwidth, w);
    argc++;
    XtSetArg(args[argc], XmNheight, h);
    argc++;
    XtSetArg(args[argc], XmNmarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNvisual, defaultConfig->awt_visInfo.visual);
    argc++;
    XtSetArg(args[argc], XmNcolormap, defaultConfig->awt_cmap);
    argc++;
    XtSetArg(args[argc], XmNdepth, defaultConfig->awt_depth);
    argc++;
    WM = awt_util_runningWindowManager();
    //if (WM == CDE_WM || WM == MOTIF_WM) {
    if ( 0 ) {
        /*
        ** note - this no decorations dosn't work on openwindows, it does
        ** work on MOTIF and CDE - so I'm only using it there.
        ** The old way (below) creates an override redirect window, 
        ** allways on top, and on every CDE Workspace.
        */
        XtSetArg(args[argc], XmNmwmDecorations, 0); argc++;
        wdata->winData.shell = XtCreatePopupShell("Window",
                                                  transientShellWidgetClass,
                                                  parentData->winData.shell,
                                                  args,
                                                  argc);
 
    } else {
        wdata->winData.shell = XtCreatePopupShell("AWTwindow",
                                                  xmMenuShellWidgetClass,
                                                  parentData->winData.shell,
                                                  args,
                                                  argc);
    }
    XtAddEventHandler(wdata->winData.shell, StructureNotifyMask,
                      FALSE, Window_event_handler, globalRef);

    argc = 0;
    XtSetArg(args[argc], XmNwidth, w);
    argc++;
    XtSetArg(args[argc], XmNheight, h);
    argc++;
    XtSetArg(args[argc], XmNmainWindowMarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNmainWindowMarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginWidth, 0);
    argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0);
    argc++;
    XtSetArg(args[argc], XmNspacing, 0);
    argc++;
    XtSetArg(args[argc], XmNorientation, XmVERTICAL); 
    argc++;
    XtSetArg(args[argc], XmNpacking, XmPACK_NONE); 
    argc++;
 
    wdata->mainWindow = XmCreateRowColumn(wdata->winData.shell, "main",
                                          args, argc);
    wdata->top = wdata->left = wdata->bottom = wdata->right = 0;

    warningString =
      (*env)->GetObjectField(env, target, windowIDs.warningString);
    if (!JNU_IsNull(env, warningString)) {
        Dimension hoffset;
        char *wString;

        wString = (char *) JNU_GetStringPlatformChars(env, warningString, NULL);
        wdata->warningWindow = awt_util_createWarningWindow(wdata->mainWindow, wString);
        JNU_ReleaseStringPlatformChars(env, warningString, (const char *) wString);

        XtVaGetValues(wdata->warningWindow, XmNheight, &hoffset, NULL);
        wdata->top += hoffset;
        awtJNI_ChangeInsets(env, this, wdata);
    } else {
        wdata->warningWindow = NULL;
    }

    wdata->winData.comp.widget = awt_canvas_create((XtPointer) globalRef,
                                                   wdata->mainWindow,
						   "",
                                                   w,
                                                   h,
						   False,
                                                   (wdata->warningWindow
                                                    ? wdata : NULL),
                                                    defaultConfig);

    awt_util_show(wdata->winData.comp.widget);
    XtManageChild(wdata->mainWindow);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MWindowPeer
 * Method:    pShow
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MWindowPeer_pShow
  (JNIEnv *env, jobject this) 
{
    struct FrameData *wdata;
    Dimension width, height;

    AWT_LOCK();

    wdata = (struct FrameData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0 ||
        wdata->mainWindow == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtVaSetValues(wdata->winData.comp.widget,
                  XmNx, -(wdata->left),
                  XmNy, -(wdata->top), NULL);

    XtVaGetValues(wdata->mainWindow,
                  XmNwidth, &width,
                  XmNheight, &height,
                  NULL);

    XtPopup(wdata->winData.shell, XtGrabNone);
    XRaiseWindow(awt_display, XtWindow(wdata->winData.shell));
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MWindowPeer
 * Method:    pHide
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MWindowPeer_pHide
  (JNIEnv *env, jobject this) 
{
    struct FrameData *wdata;

    AWT_LOCK();

    wdata = (struct FrameData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    XtPopdown(wdata->winData.shell);
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MWindowPeer
 * Method:    pReshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MWindowPeer_pReshape
  (JNIEnv *env, jobject this, jint x, jint y, jint w, jint h) 
{
    struct FrameData *wdata;
    Dimension hoffset, aw, ah;
    jobject target;

    AWT_LOCK();

    wdata = (struct FrameData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);
    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);

    if (wdata == 0 ||
        wdata->winData.comp.widget == 0 ||
        wdata->winData.shell == 0 ||
        JNU_IsNull(env, target)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (wdata->warningWindow != 0) {
        XtVaGetValues(wdata->warningWindow, XmNheight, &hoffset, NULL);
    } else {
        hoffset = 0;
    }

    aw = w - (wdata->left + wdata->right);

    ah = h + hoffset - (wdata->top + wdata->bottom);

    aw = (aw > 0) ? aw : 1;
    ah = (ah > 0) ? ah : 1;

    /* GES: AVH's hack from awt_util.c: 
     * Motif ignores attempts to move a toplevel window to 0,0.
     * Instead we set the position to 1,1. The expected value is
     * returned by Frame.getBounds() since it uses the internally
     * held rectangle rather than querying the peer.
     */

    if ((x == 0) && (y == 0)) {
        //XtVaSetValues(wdata->winData.shell, XmNx, 1, XmNy, 1, NULL);
        x = 1;
        y = 1;
    }
    XtVaSetValues(wdata->winData.shell,
                  XtNx, (XtArgVal) x,
                  XtNy, (XtArgVal) y,
                  XtNwidth, (XtArgVal) aw,
                  XtNheight, (XtArgVal) ah,
                  NULL);

    if ( wdata->warningWindow != 0 ) {
        XtConfigureWidget( wdata->warningWindow,
                        wdata->left, wdata->top - hoffset,
                        aw, hoffset, 0);
                          
        XtConfigureWidget( XtParent(wdata->winData.comp.widget),
                        wdata->left, wdata->top,
                        aw, (ah - hoffset), 0 );
                          
        XtConfigureWidget( wdata->winData.comp.widget,
                        -wdata->left, -wdata->top,
                        (w>0) ? w : 0, (h>0) ? h : 0, 0 );
    } else {              
        XtConfigureWidget( wdata->winData.comp.widget,
                        wdata->left, wdata->top, aw, ah, 0 );
    }                     
 
    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MWindowPeer
 * Method:    pDispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MWindowPeer_pDispose
  (JNIEnv *env, jobject this) 
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 || wdata->mainWindow == 0 || wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    XtUnmanageChild(wdata->winData.comp.widget);
    awt_delWidget(wdata->winData.comp.widget);
    awt_util_consumeAllXEvents(wdata->winData.comp.widget);
    XtDestroyWidget(wdata->winData.comp.widget);

    XtUnmanageChild(wdata->mainWindow);
    awt_util_consumeAllXEvents(wdata->mainWindow);
    awt_util_consumeAllXEvents(wdata->winData.shell);
    XtDestroyWidget(wdata->mainWindow);
    XtDestroyWidget(wdata->winData.shell);
    free((void *) wdata);

    (*env)->SetLongField(env,this,mComponentPeerIDs.pData,0);
    awtJNI_DeleteGlobalRef(env, this);

    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_MWindowPeer
 * Method:    toBack
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_MWindowPeer_toBack
  (JNIEnv *env, jobject this)
{
    struct FrameData *wdata;

    AWT_LOCK();
    wdata = (struct FrameData *)
      (*env)->GetLongField(env,this,mComponentPeerIDs.pData);

    if (wdata == 0 || wdata->winData.shell == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }
    if (XtWindow(wdata->winData.shell) != 0) {
        XLowerWindow(awt_display, XtWindow(wdata->winData.shell));
    }
    AWT_FLUSH_UNLOCK();
}
