/*
 * @(#)awt_XmDnD.c	1.34 98/09/29
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <stdio.h>
#include <string.h>

#include "jvm.h"
#include "jni.h"
#include "jni_util.h"

#include "awt_XmDnD.h"

#include "awt_p.h"

#include "java_awt_Cursor.h"
#include "java_awt_dnd_DnDConstants.h"

#include <errno.h>

#include <X11/cursorfont.h>
#include <X11/Xutil.h>

#include <sys/utsname.h>

#include "awt_Component.h"
#include "awt_Cursor.h"
#include "awt_AWTEvent.h"

extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct X11CustomCursorIDs x11CustomCursorIDs;

/* globals */

extern XtAppContext awt_appContext;

/* forwards */

static void awt_XmDropProc(Widget, XtPointer, XmDropProcCallbackStruct*);
static void awt_XmDragProc(Widget, XtPointer, XmDragProcCallbackStruct*);

static void awt_XmTransferProc(Widget, XtPointer, Atom*, Atom*, XtPointer, unsigned long*, int*);

/* for XmDragContext callbacks etc ... */

static Boolean awt_XmConvertProc(Widget, Atom*, Atom*, Atom*, XtPointer*, unsigned long*, int*);

static Boolean awt_XmConvertIncrProc(Widget, Atom*, Atom*, Atom*, XtPointer*, unsigned long*, int*, unsigned long*, XtPointer, XtRequestId*);

static void awt_XmDragEnterProc(Widget, XtPointer, XmDropSiteEnterCallbackStruct*);
static void awt_XmDragMotionProc(Widget, XtPointer, XmDragMotionCallbackStruct*);
static void awt_XmDragLeaveProc(Widget, XtPointer, XmDropSiteLeaveCallbackStruct*);
static void awt_XmDropStartProc(Widget, XtPointer, XmDropStartCallbackStruct*);
static void awt_XmDropOperationChangedProc(Widget, XtPointer, XmDropStartCallbackStruct*);
static void awt_XmDropFinishProc(Widget, XtPointer, XmDropFinishCallbackStruct*);
static void awt_XmDragDropFinishProc(Widget, XtPointer, XmDropFinishCallbackStruct*);

static int  DnDConstantsToXm(jint operations);
static jint XmToDnDConstants(int operations);

static void flush_cache(JNIEnv* env);

static void setCursor(JNIEnv* env, Display* d, jobject c, jint type, Time t);

/* JNI static values for methods in MDropTargetContextPeer */

static jclass dTCClazz;

static const char dTCcbmethodSig[] = "(Ljava/awt/Component;IIII[Ljava/lang/String;J)I";

static jmethodID dTCcreate;
static jmethodID dTCenter;
static jmethodID dTCexit;
static jmethodID dTCmotion;
static jmethodID dTCdrop;
static jmethodID dTCdropDisp;

static jmethodID dTCnewData;
static jmethodID dTCtxFailed;

static jmethodID dTCheartbeat;

static jmethodID dTCgetL;
static jmethodID dTCfilei;


/* JNI static values for methods in MDragSourceContextPeer */

static jclass dSCClazz;

static const char dSCcbmethodSig[] = "(IIIZ)I";

static jmethodID dSCenter;
static jmethodID dSCmotion;
static jmethodID dSCchanged;
static jmethodID dSCexit;

static jmethodID dSCdstart;

static jmethodID dSCconvert;
static jmethodID dSCconvertFiles;

static jmethodID dSCdfinished;
static jmethodID dSCddfinished;

static jmethodID dSCfree;

/* the private Buffer Class */

static jclass    dSCBufferClazz;

static jmethodID dSCBsize;
static jmethodID dSCBgbuf;
static jmethodID dSCBirc;

static int recWindowCount = 0; /*
				* used to simulate when window ids not available
				* provides hints to upper layers regarding 
				* transitions between drop target windows
				*/

/* Cursor Support */

static jclass string; /* used to construct string objects for native cache */
static jclass file;   /* used to construct file objects for transfers */

extern int getModifiers(int); /* in canvas.c */

/**
 * map from atoms to java.lang.String representations for DropTarget cache
 */

typedef struct {
	Atom 	atom;
	jobject string;
	Boolean mark;
} TargetMap, *TargetMapPtr;

/**
 * static cache of DropTarget related info.
 */

static struct {
    Widget  	  w;		/* if NULL, cache invalid */

    jobject	  peer;
    jobject	  component;

    jobject	  dtcpeer;

    Widget	  dt;

    jobject	  targets;
    TargetMapPtr* targetMap;
    Cardinal	  nTargets;

    Boolean	  dropDone;
    int		  transfersPending;
    Widget	  transfer;

    Boolean	  flushPending;
} _cache;

/**
 * global function to initialize this client as a Dynamic-only app.
 *
 * gets called once during toolkit initialization.
 */

void awt_initialize_Xm_DnD(Display* dpy) {
	JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	jclass  clazz;

	XtVaSetValues(XmGetXmDisplay(dpy),
		      XmNdragInitiatorProtocolStyle, XmDRAG_DYNAMIC,
		      XmNdragReceiverProtocolStyle,  XmDRAG_DYNAMIC,
		      NULL
	);

	/*
   	 * load class java.lang.string
	 */

	string = (*env)->FindClass(env, "java/lang/String");

	if (JNU_IsNull(env, string)) {
	    (*env)->ThrowNew(env,
			     (*env)->FindClass(env, "java/lang/ClassNotFoundException"),
			     "Cant find java/lang/String"
	    );
	}

	string = (*env)->NewGlobalRef(env, string); /* never freed! */

	/*
   	 * load class java.io.File
	 */

	file = (*env)->FindClass(env, "java/io/File");

	if (JNU_IsNull(env, string)) {
	    (*env)->ThrowNew(env,
			     (*env)->FindClass(env, "java/lang/ClassNotFoundException"),
			     "Cant find java/io/file"
	    );
	}

	file = (*env)->NewGlobalRef(env, file); /* never freed! */

	dTCfilei = (*env)->GetMethodID(env,
				       file,
				       "<init>",
				       "(Ljava/lang/String;)V"
	);


	if (JNU_IsNull(env, dTCfilei)) return;

	/*
   	 * load class sun.awt.motif.MDropTargetContextPeer and resolve methods
	 */

	dTCClazz = (*env)->FindClass(env, "sun/awt/motif/MDropTargetContextPeer");

	if (JNU_IsNull(env, dTCClazz)) {
	    (*env)->ThrowNew(env,
			     (*env)->FindClass(env, "java/lang/ClassNotFoundException"),
			     "Cant find sun/awt/motif/MDropTargetContextPeer"
	    );
	}

	dTCClazz = (*env)->NewGlobalRef(env, dTCClazz); /* never freed! */

	if (JNU_IsNull(env, dTCClazz)) return;

	dTCcreate = (*env)->GetStaticMethodID(env,
					      dTCClazz,
					      "createMDropTargetContextPeer",
					      "()Lsun/awt/motif/MDropTargetContextPeer;"
		    );

	if (JNU_IsNull(env, dTCcreate)) return;

	dTCenter = (*env)->GetMethodID(env,
				       dTCClazz,
				       "handleEnterMessage",
				        dTCcbmethodSig
		    );


	if (JNU_IsNull(env, dTCenter)) return;

	dTCexit = (*env)->GetMethodID(env,
				      dTCClazz,
				       "handleExitMessage",
				        "(Ljava/awt/Component;J)V"
		    );


	if (JNU_IsNull(env, dTCexit)) return;

	dTCmotion = (*env)->GetMethodID(env,
				       dTCClazz,
				       "handleMotionMessage",
				       "(Ljava/awt/Component;IIIZI[Ljava/lang/String;J)I"
		    );


	if (JNU_IsNull(env, dTCmotion)) return;

	dTCdrop = (*env)->GetMethodID(env,
				      dTCClazz,
				      "handleDropMessage",
				       dTCcbmethodSig
		  );

	if (JNU_IsNull(env, dTCdrop)) return;

	dTCdropDisp = (*env)->GetMethodID(env,
				          dTCClazz,
				      	  "dropDispatched",
				          "()V"
		      );

	if (JNU_IsNull(env, dTCdropDisp)) return;

	/*
         * now load the data method callbacks for MDropTargetContextPeer 
	 */

	dTCnewData = (*env)->GetMethodID(env,
				         dTCClazz,
				         "newData",
				         "(Ljava/lang/String;Ljava/lang/String;[B)V"
		     );

	if (JNU_IsNull(env, dTCnewData)) return;

	dTCtxFailed = (*env)->GetMethodID(env,
                                          dTCClazz,
                                          "transferFailed",
                                          "(Ljava/lang/String;)V"
                      );

	if (JNU_IsNull(env, dTCtxFailed)) return;

	dTCheartbeat = (*env)->GetMethodID(env,
                                          dTCClazz,
                                          "heartbeat",
                                          "()V"
                       );

	if (JNU_IsNull(env, dTCheartbeat)) return;

	dTCgetL = (*env)->GetStaticMethodID(env,
					      dTCClazz,
					      "getList",
					      "([Ljava/lang/Object;)Ljava/util/List;"
		  );

	if (JNU_IsNull(env, dTCgetL)) return;

	/*
   	 * load class sun.awt.motif.MDragSourceContextPeer and resolve methods
	 */

	dSCClazz = (*env)->FindClass(env, "sun/awt/motif/MDragSourceContextPeer");

	if (JNU_IsNull(env, dSCClazz)) {
	    (*env)->ThrowNew(env,
			     (*env)->FindClass(env, "java/lang/ClassNotFoundException"),
			     "Cant find sun/awt/motif/DragSourceContextPeer"
	    );
	}

	dSCClazz = (*env)->NewGlobalRef(env, dSCClazz); /* never freed! */

	if (JNU_IsNull(env, dSCClazz)) return;

	dSCenter = (*env)->GetMethodID(env,
                                       dSCClazz,
                                       "dragEnter",
				       dSCcbmethodSig
                   );

	if (JNU_IsNull(env, dSCenter)) return;

	dSCmotion = (*env)->GetMethodID(env,
                                        dSCClazz,
                                        "dragMotion",
					dSCcbmethodSig
                    );

	if (JNU_IsNull(env, dSCmotion)) return;

	dSCchanged = (*env)->GetMethodID(env,
                                         dSCClazz,
                                         "operationChanged",
					 dSCcbmethodSig
                     );

	if (JNU_IsNull(env, dSCchanged)) return;

	dSCexit = (*env)->GetMethodID(env,
                                      dSCClazz,
                                      "dragExit",
				      "(I)V"
                  );

	if (JNU_IsNull(env, dSCexit)) return;

	dSCdstart = (*env)->GetMethodID(env,
                                      dSCClazz,
                                      "dropStart",
				      "(IZZ)V"
                  );

	if (JNU_IsNull(env, dSCdstart)) return;

	dSCconvert = (*env)->GetMethodID(env,
                                         dSCClazz,
                                         "convertData",
                                         "(Ljava/lang/String;IJ)Lsun/awt/motif/MDragSourceContextPeerBuffer;"
                      );

	if (JNU_IsNull(env, dSCconvert)) return;

	dSCconvertFiles = (*env)->GetMethodID(env,
                                              dSCClazz,
                                              "convertFileList",
                                              "(Ljava/lang/String;)[Ljava/lang/String;"
                      );

	if (JNU_IsNull(env, dSCconvertFiles)) return;

	dSCdfinished = (*env)->GetMethodID(env,
                                          dSCClazz,
                                          "dropFinished",
                                          "(Z)V"
                       );

	if (JNU_IsNull(env, dSCdfinished)) return;

	dSCddfinished = (*env)->GetMethodID(env,
                                          dSCClazz,
                                          "dragDropFinished",
                                          "()V"
                        );

	if (JNU_IsNull(env, dSCddfinished)) return;

	dSCfree = (*env)->GetMethodID(env,
                                      dSCClazz,
                                      "freeBuffer",
				      "(Lsun/awt/motif/MDragSourceContextPeerBuffer;)V"
	);

	if (JNU_IsNull(env, dSCfree)) return;

	/*
   	 * load class sun.awt.motif.MDragSourceContextPeer and resolve methods
	 */

	dSCBufferClazz = (*env)->FindClass(env, "sun/awt/motif/MDragSourceContextPeerBuffer");

	if (JNU_IsNull(env, dSCBufferClazz)) {
	    (*env)->ThrowNew(env,
			     (*env)->FindClass(env, "java/lang/ClassNotFoundException"),
			     "Cant find sun/awt/motif/MDragSourceContextPeerBUffer"
	    );
	}


	dSCBufferClazz = (*env)->NewGlobalRef(env, dSCBufferClazz); /* never freed! */

	if (JNU_IsNull(env, dSCBufferClazz)) return;

	dSCBsize = (*env)->GetMethodID(env,
                                       dSCBufferClazz,
                                       "size",
				       "()I"
                      );

	if (JNU_IsNull(env, dSCBsize)) return;

	dSCBgbuf = (*env)->GetMethodID(env,
                                       dSCBufferClazz,
                                       "getBuffer",
				       "()[B"
	);

	if (JNU_IsNull(env, dSCBgbuf)) return;

	dSCBirc = (*env)->GetMethodID(env,
                                      dSCBufferClazz,
                                      "incrReadCursor",
				      "(I)I"
	);

	if (JNU_IsNull(env, dSCBirc)) return;

	/*
	 * load the Cursor stuff
	 */

	clazz = (*env)->FindClass(env, "sun/awt/motif/X11CustomCursor");

	if (JNU_IsNull(env, clazz)) {
	    (*env)->ThrowNew(env,
			     (*env)->FindClass(env, "java/lang/ClassNotFoundException"),
			     "Cant find sun/awt/motif/X11CustomCursor	"
	    );
	}
}

/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    addNativeDropTarget
 * Signature: (Ljava/awt/dnd/DropTarget;)V
 */

JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_addNativeDropTarget (JNIEnv *env, jobject this, jobject droptarget) {
    struct ComponentData* cdata     = (struct ComponentData *)NULL;
           DropSitePtr	  dropsite  = (DropSitePtr)NULL;

    if (JNU_IsNull(env, droptarget)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }

    AWT_LOCK();

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);

    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    /* introduce a new Component as a root of a set of DropTargets */

    if ((dropsite = cdata->dsi) == (DropSitePtr)NULL) {
	Arg 	     args[10];
	unsigned int nargs = 0;

	dropsite = cdata->dsi = (DropSitePtr)ZALLOC(DropSiteInfo);

	if (dropsite == (DropSitePtr)NULL) {
	    JNU_ThrowOutOfMemoryError (env, "OutOfMemoryError");
	    AWT_UNLOCK ();
            return;
        }

	/*
         * We mark Composite widgets as COMPOSITE DropSite's by default
         */

	dropsite->component = (*env)->
	  NewGlobalRef(env,
		       (*env)->GetObjectField(env,
					      this,
					      mComponentPeerIDs.target));
	dropsite->isComposite =
	  XtIsSubclass(cdata->widget, compositeWidgetClass);
	for (dropsite->tlw = cdata->widget;
	     XtParent(dropsite->tlw) != NULL;
	     dropsite->tlw = XtParent(dropsite->tlw))
	;

	
#define SetArg(n, v) args[nargs].name = n; args[nargs++].value = (XtArgVal)(v);

	SetArg(XmNanimationStyle,	XmDRAG_UNDER_NONE);
	SetArg(XmNdragProc,		awt_XmDragProc);
	SetArg(XmNdropProc,		awt_XmDropProc);
	SetArg(XmNdropSiteActivity,	XmDROP_SITE_ACTIVE);

	SetArg(XmNdropSiteOperations,	XmDROP_LINK |
					XmDROP_MOVE |
					XmDROP_COPY
	);

	SetArg(XmNimportTargets,	NULL);
	SetArg(XmNnumImportTargets,	0);

	if (dropsite->isComposite) {
	    SetArg(XmNdropSiteType,    XmDROP_SITE_COMPOSITE);
	} else {
	    SetArg(XmNdropSiteType,    XmDROP_SITE_SIMPLE);
	    SetArg(XmNdropRectangles, (XRectangle*)NULL);
	}

#undef  SetArg

	XmDropSiteUnregister(cdata->widget);
	XmDropSiteRegister(cdata->widget, args, nargs);
	XmDropSiteConfigureStackingOrder(cdata->widget, (Widget)NULL, XmABOVE);

	/*
 	 *
	 */

	{
	    jobject ref = (jobject)NULL;

	    Arg	    args[1];
	
	    args[0].name = XmNuserData; args[0].value = (XtArgVal)&ref;

	    XtGetValues(cdata->widget, args, 1);

	    if (ref == (jobject)NULL || JNU_IsNull(env, ref)) {
		args[1].value = (XtArgVal)this;

		XtSetValues(cdata->widget, args, 1);
	    }
	}
    }

    dropsite->dsCnt++;

    AWT_UNLOCK();
}
 
/*
 * Class:     sun_awt_motif_MComponentPeer
 * Method:    removeNativeDropTarget
 * Signature: (Ljava/awt/dnd/DropTarget;)V
 */

JNIEXPORT void JNICALL Java_sun_awt_motif_MComponentPeer_removeNativeDropTarget (JNIEnv *env, jobject this, jobject droptarget) {

    struct ComponentData* cdata;
    DropSitePtr		  dropsite;

    if (JNU_IsNull(env, droptarget)) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        return;
    }

    AWT_LOCK();

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env, this, mComponentPeerIDs.pData);

    if (cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    if ((dropsite = cdata->dsi) == (DropSitePtr)NULL) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return;
    }

    if (dropsite->dsCnt-- == 0) {
	XmDropSiteUnregister(cdata->widget);

	(*env)->DeleteGlobalRef(env, dropsite->component);

	free((void *)(cdata->dsi));
	cdata->dsi = (DropSitePtr)NULL;
    }

    AWT_UNLOCK();
}

/**
 *
 */

JNIEXPORT jlong JNICALL Java_sun_awt_motif_MDropTargetContextPeer_startTransfer(JNIEnv *env, jobject this, jlong dragContext, jobject atom) {
    XmDropTransferEntryRec trec;
    Widget		   dropTransfer;
    Arg		           args[3];
    unsigned int	   nargs = 0;
    jboolean		   isCopy;
    const char*		   atomName = (*env)->GetStringUTFChars(env, atom, &isCopy);

    AWT_LOCK();

    trec.target      = XInternAtom(XtDisplayOfObject((Widget)dragContext),
    			       atomName,
    			       False
    		       );

    (*env)->ReleaseStringUTFChars(env, atom, atomName);

    trec.client_data = (XtPointer)trec.target;
    

#define SetArg(n, v) args[nargs].name = n; args[nargs++].value = (XtArgVal)(v);

    SetArg(XmNdropTransfers,    &trec);
    SetArg(XmNnumDropTransfers, 1    );
    SetArg(XmNtransferProc,     awt_XmTransferProc);

#undef SetArg

    _cache.transfer = dropTransfer = XmDropTransferStart((Widget)dragContext, args, nargs);

    AWT_UNLOCK();

    return (jlong)dropTransfer;
}

/**
 *
 */

JNIEXPORT void JNICALL Java_sun_awt_motif_MDropTargetContextPeer_addTransfer(JNIEnv *env, jobject this, jlong dropTransfer, jobject atom) {
    XmDropTransferEntryRec trec;
    jboolean		   isCopy;
    const char*		   atomName = (*env)->GetStringUTFChars(env, atom, &isCopy);

    AWT_LOCK();

    trec.target      = XInternAtom(XtDisplayOfObject((Widget)dropTransfer),
    			       atomName,
    			       False
    		       );

    (*env)->ReleaseStringUTFChars(env, atom, atomName);

    trec.client_data = (XtPointer)trec.target;
    
    XmDropTransferAdd((Widget)dropTransfer, &trec, 1);

    AWT_UNLOCK();
}

/**
 *
 */

JNIEXPORT void JNICALL Java_sun_awt_motif_MDropTargetContextPeer_keepAlive(JNIEnv *env, jobject this, jlong dropTransfer) {
    if (_cache.transfersPending-- == 0) {
        XmDropTransferEntryRec trec;

	AWT_LOCK();

    	trec.target      = XInternAtom(XtDisplayOfObject((Widget)dropTransfer), _XA_JAVA_KEEP_ALIVE, False);
        trec.client_data = (XtPointer)trec.target;
    
        XmDropTransferAdd((Widget)dropTransfer, &trec, 1);

	AWT_UNLOCK();

	_cache.transfersPending++;
    }
}

/**
 *
 */

JNIEXPORT void JNICALL Java_sun_awt_motif_MDropTargetContextPeer_dropDone(JNIEnv *env, jobject this, jlong dragContext, jlong dropTransfer, jboolean isLocal, jboolean success) {
    AWT_LOCK();

    if (_cache.w == (Widget)NULL) {
	AWT_UNLOCK();
	return; 
    }

    if (_cache.dropDone == False) {
	if (dropTransfer != (jlong)NULL) {
            XtVaSetValues((Widget)dropTransfer, 
		      	  XmNtransferStatus,
		      	  success == JNI_TRUE ? XmTRANSFER_SUCCESS : XmTRANSFER_FAILURE,
		          NULL
	    );
	} else {
	    Arg arg;

	    arg.name  = XmNtransferStatus;
	    arg.value = (XtArgVal)(success == JNI_TRUE ? XmTRANSFER_SUCCESS 
						       : XmTRANSFER_FAILURE
			);

	    XmDropTransferStart((Widget)dragContext, &arg, 1);
	}

	/*
         * bugid# 4146717
	 *
	 * If this is a local tx, then we never exec the awt_XmTransferProc,
	 * thus we need to flush the cache here as it is our only chance,
	 * otherwise we leave a mess for the next operation to fail on ....
	 *
	 */

	if (isLocal == JNI_TRUE) 
	    flush_cache(env); /* flush now, last chance */
	else
            _cache.flushPending = True; /* flush pending in transfer proc */
    }

    AWT_UNLOCK();
}

/**
 *
 */
 
JNIEXPORT void JNICALL Java_sun_awt_motif_MDragSourceContextPeer_setCursor(JNIEnv *env, jobject this, jlong nativeCtxt, jobject cursor, jint type) {
    Display*		dpy     = XtDisplayOfObject((Widget)nativeCtxt);

    setCursor(env, dpy, cursor, type, XtLastTimestampProcessed(dpy));
}

/**
 *
 */

JNIEXPORT void JNICALL Java_sun_awt_motif_AWTLockAccess_awtLock(JNIEnv *env, jclass clazz) {
    AWT_LOCK();
}

/**
 *
 */

JNIEXPORT void JNICALL Java_sun_awt_motif_AWTLockAccess_awtUnlock(JNIEnv *env, jclass clazz) {
    AWT_NOTIFY_ALL();
    AWT_UNLOCK();
}

/**
 *
 */

JNIEXPORT jlong JNICALL Java_sun_awt_motif_MDragSourceContextPeer_startDrag(JNIEnv *env, jobject this, jobject nativeCont, jobject trigger, jobject cursor, jint ctype, jint actions, jarray types) {
    struct ComponentData* cdata;
    Display*		  dpy;
    Arg			  args[32];
    int			  nargs = 0;
    jobject		  dscp  = (*env)->NewGlobalRef(env, this);
    int			  i;
    Atom*		  targets;
    int			  nTargets;
    Widget		  dc;
    XtCallbackRec	  dsecbr[2];
    XtCallbackRec	  dmcbr[2];
    XtCallbackRec	  occbr[2];
    XtCallbackRec	  dslcbr[2];
    XtCallbackRec	  dscbr[2];
    XtCallbackRec	  dfcbr[2];
    XtCallbackRec	  ddfcbr[2];
    XEvent*		  event;
    int			  xmActions = DnDConstantsToXm(actions);

    if (xmActions == XmDROP_NOOP) return NULL;

    AWT_LOCK();

#define SetCB(cbr, cb, cl) cbr[0].callback = (XtCallbackProc)cb; cbr[0].closure = (XtPointer)cl; cbr[1].callback = (XtCallbackProc)NULL; cbr[1].closure = (XtPointer)NULL

#define SetArg(n, v) args[nargs].name = n; args[nargs++].value = (XtArgVal)(v);

    cdata = (struct ComponentData *)
      (*env)->GetLongField(env, nativeCont, mComponentPeerIDs.pData);

    if (JNU_IsNull(env, types) || cdata == 0 || cdata->widget == 0) {
        JNU_ThrowNullPointerException(env, "NullPointerException");
        AWT_UNLOCK();
        return(jlong)NULL;
    }

    event = (XEvent*)(*env)->GetLongField(env, trigger, awtEventIDs.data);

    if (event == (XEvent*)NULL || event->type != ButtonPress) {
	AWT_UNLOCK();
	return (jlong)NULL;
    }

    dpy = XtDisplayOfObject(cdata->widget);

    nTargets = (*env)->GetArrayLength(env, types);

    targets = nTargets > 0 ? (Atom*)XtCalloc(nTargets, sizeof(Atom)) : NULL;

    for (i = 0; i < nTargets; i++) {
	jstring  elem = (jstring)(*env)->GetObjectArrayElement(env, types, (jsize)i);
	jboolean isCopy;
	String   s;

	s = (String)(*env)->GetStringUTFChars(env, elem, &isCopy);

	targets[i] = XInternAtom(dpy, s, False);

	(*env)->ReleaseStringUTFChars(env, elem, (const char *)s);
    }

    SetCB(dsecbr, awt_XmDragEnterProc,		  dscp);
    SetCB(dmcbr,  awt_XmDragMotionProc,		  dscp);
    SetCB(occbr,  awt_XmDropOperationChangedProc, dscp);
    SetCB(dslcbr, awt_XmDragLeaveProc,		  dscp);
    SetCB(dscbr,  awt_XmDropStartProc,		  dscp);
    SetCB(dfcbr,  awt_XmDropFinishProc,		  dscp);
    SetCB(ddfcbr, awt_XmDragDropFinishProc,	  dscp);

    SetArg(XmNblendModel,    		XmBLEND_NONE          );
    SetArg(XmNclientData,	        dscp		      );
    SetArg(XmNdragOperations,	        xmActions             );
#if	0
    SetArg(XmNconvertProc,   		awt_XmConvertIncrProc );
    SetArg(XmNincremental,	    	True	 	      );
#else
    SetArg(XmNconvertProc,   		awt_XmConvertProc     );
#endif
    SetArg(XmNdropSiteEnterCallback,    dsecbr		      );
    SetArg(XmNdragMotionCallback, 	dmcbr		      );
    SetArg(XmNoperationChangedCallback,	occbr		      );
    SetArg(XmNdropSiteLeaveCallback,    dslcbr		      );
    SetArg(XmNdropStartCallback, 	dscbr		      );
    SetArg(XmNdropFinishCallback, 	dfcbr		      );
    SetArg(XmNdragDropFinishCallback,   ddfcbr		      );
    SetArg(XmNexportTargets,		targets		      );
    SetArg(XmNnumExportTargets,    	nTargets	      );

    dc = XmDragStart(cdata->widget, event, args, nargs);

    if (dc != (Widget)NULL) {
	setCursor(env, dpy, cursor, ctype, event->xbutton.time);
    }

    if (targets) XtFree((void*)targets);

    AWT_UNLOCK();

    return (jlong)dc;

#undef SetArg
#undef SetCB
}

/******************************************************************************/

/**
 *
 */

static void setCursor(JNIEnv* env, Display* dpy, jobject cursor, jint type, Time time) {
    Cursor xcursor = None;

    if (JNU_IsNull(env, cursor)) return;

    if (type == (jint)-1) {
	xcursor = (Cursor)
	  (*env)->GetLongField(env, cursor, x11CustomCursorIDs.pData);

	if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	    (*env)->ExceptionDescribe(env);
	    (*env)->ExceptionClear(env);
	}
    } else {
        switch (type) {
          default:
          case java_awt_Cursor_DEFAULT_CURSOR:
              break;

          case java_awt_Cursor_CROSSHAIR_CURSOR:
              type = XC_crosshair;
              break;
          case java_awt_Cursor_TEXT_CURSOR:
              type = XC_xterm;
              break;
          case java_awt_Cursor_WAIT_CURSOR:
              type = XC_watch;
              break;
          case java_awt_Cursor_SW_RESIZE_CURSOR:
              type = XC_bottom_left_corner;
              break;
          case java_awt_Cursor_NW_RESIZE_CURSOR:
              type = XC_top_left_corner;
              break;
          case java_awt_Cursor_SE_RESIZE_CURSOR:
              type = XC_bottom_right_corner;
              break;
          case java_awt_Cursor_NE_RESIZE_CURSOR:
              type = XC_top_right_corner;
              break;
          case java_awt_Cursor_S_RESIZE_CURSOR:
              type = XC_bottom_side;
              break;
          case java_awt_Cursor_N_RESIZE_CURSOR:
              type = XC_top_side;
              break;
          case java_awt_Cursor_W_RESIZE_CURSOR:
              type = XC_left_side;
              break;
          case java_awt_Cursor_E_RESIZE_CURSOR:
              type = XC_right_side;
              break;
          case java_awt_Cursor_HAND_CURSOR:
              type = XC_hand2;
              break;
          case java_awt_Cursor_MOVE_CURSOR:
              type = XC_fleur;
              break;
        }
    }

    XChangeActivePointerGrab(dpy,
			     ButtonPressMask   |
			     ButtonMotionMask  |
			     ButtonReleaseMask |
			     EnterWindowMask   |
			     LeaveWindowMask,
			     xcursor,
	    		     time
    );

    XSync(dpy, False);
}

/**
 *
 */

static int _compar(const void *first, const void* second) {
	TargetMapPtr firsttp  = *((TargetMapPtr*)first);
	TargetMapPtr secondtp = *((TargetMapPtr*)second);

	return ((firsttp->atom < secondtp->atom) ? -1 : firsttp->atom != secondtp->atom);
}

/**
 *
 */

static Boolean updateCachedTargets(JNIEnv* env, Widget dt) {
    Atom*    	  targets  = (Atom*)NULL;
    Cardinal 	  nTargets = (Cardinal)0;
    Arg      	  args[2];
    TargetMapPtr* tmp      = (TargetMapPtr*)NULL;
    int 	  changed  = 0;
    Boolean	  OldArry  = !JNU_IsNull(env, _cache.targets);
    int		  i;
    int		  dups     = 0;
    TargetMap	  tm;
    TargetMapPtr  tp       = &tm;

    args[0].name = XmNexportTargets;    args[0].value = (XtArgVal)&targets;
    args[1].name = XmNnumExportTargets; args[1].value = (XtArgVal)&nTargets;

    XtGetValues(_cache.dt = dt, args, 2);


    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) return False; 

    if (nTargets > 0) tmp = (TargetMapPtr*)XtCalloc(nTargets, sizeof(TargetMapPtr));

    tm.string = (jobject)NULL;
    tm.mark   = False;

    for (i = 0; i < nTargets; i++) {
	TargetMapPtr*  p = (TargetMapPtr*)NULL;

	if (OldArry) {
	    tm.atom  = targets[i];

	    p = (TargetMapPtr*)bsearch((const void *)&tp,
				       (const void *)_cache.targetMap,
				       _cache.nTargets,
				       sizeof(TargetMapPtr),
				       _compar
		);
	}

	if (p != (TargetMapPtr*)NULL) {
	    tmp[i] = *p;
	    tmp[i]->mark = OldArry;
	} else {
	    int  	 j;
	    Atom	 a = targets[i];
	    char*	 name;
	    TargetMapPtr tp;

	    for (j = 0;
		 j <= nTargets - 1 && tmp[j] != (TargetMapPtr)NULL && a > tmp[j]->atom;
		 j++)
	    ;

	    if (tmp[j] != (TargetMapPtr)NULL && a == tmp[j]->atom) {
		dups++;
		continue;
	    }

	    tp = (TargetMapPtr)XtCalloc(1, sizeof(TargetMap));

	    name = XGetAtomName(XtDisplayOfObject(dt), a);

	    tp->string = (*env)->NewStringUTF(env, (const char *)name);
	    tp->string = (*env)->NewGlobalRef(env, tp->string);

	    if (tmp[j] != (TargetMapPtr)NULL) {
		 memmove((void *)      (tmp + j + 1),
		         (const void *)(tmp + j),
			 sizeof(TargetMapPtr) * (nTargets - j - 1)
		 );
	    }

	    XFree((void*)name);

	    tp->atom = a;
	    tp->mark = False;
	    tmp[j]   = tp;

	    changed++;
	}
    }

    nTargets -= dups; /* adjust actual # for duplicate entries */

    if (changed) {
	int i;
	int max = nTargets > _cache.nTargets ? nTargets : _cache.nTargets;

        if (_cache.nTargets != nTargets) {
	    if (!JNU_IsNull(env, _cache.targets)) {
	        (*env)->DeleteGlobalRef(env, _cache.targets);

		_cache.targets = (jobject)NULL;
	    }

	    if (nTargets > 0) {
	        _cache.targets = (*env)->NewObjectArray(env, nTargets, string, (jobject)NULL);
	        _cache.targets = (*env)->NewGlobalRef(env, _cache.targets);

		if (JNU_IsNull(env, _cache.targets) || !JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);

		    for (i = 0; i < _cache.nTargets; i++) {
			(*env)->DeleteGlobalRef(env, _cache.targetMap[i]->string);
			XtFree((void*)_cache.targetMap[i]);
		    }
			
	    	    XtFree((void *)_cache.targetMap);
		    XtFree((void *) tmp);

		    _cache.targets  = (jobject)NULL;
		    _cache.nTargets = 0;
		}
	    }

	    OldArry = False;
        }

	for (i = 0; i < max; i++) {
	    if (i < nTargets && (!OldArry || !tmp[i]->mark)) {
		(*env)->SetObjectArrayElement(env, _cache.targets, i, tmp[i]->string);
	    }

	    if (i < _cache.nTargets && !_cache.targetMap[i]->mark) {
		(*env)->DeleteGlobalRef(env, _cache.targetMap[i]->string);
		XtFree((void*)_cache.targetMap[i]);
	    }
	}

	if (_cache.targetMap != (TargetMapPtr*)NULL) {
	    XtFree((void *)_cache.targetMap);
	} 

	_cache.targetMap = tmp;
        _cache.nTargets  = nTargets;

	return True;
    } else XtFree((void *)tmp);

    return False;
}


/**
 *
 */

static jobject findCachedTarget(Atom a) {
    TargetMapPtr* tmp = (TargetMapPtr*)NULL;
    TargetMap     tm;
    TargetMapPtr  tp = &tm;

    tm.mark   = False;
    tm.string = (jobject)NULL;
    tm.atom   = a;

    if (_cache.w != (Widget)NULL) {
	tmp = (TargetMapPtr*)bsearch((const void *)&tp,
				     (const void *)_cache.targetMap,
				     _cache.nTargets,
				     sizeof(TargetMapPtr),
				     _compar
	);
    }

    return tmp != (TargetMapPtr*)NULL ? (*tmp)->string : (jobject)NULL;
}

/**
 *
 */

static void flush_cache(JNIEnv* env) {
    _cache.w  = (Widget)NULL;
    _cache.dt = (Widget)NULL;

    (*env)->DeleteGlobalRef(env, _cache.peer);
    _cache.peer = (jobject)NULL;

    (*env)->DeleteGlobalRef(env, _cache.component);
    _cache.component = (jobject)NULL;

   if (_cache.dtcpeer != (jobject)NULL) {
	(*env)->DeleteGlobalRef(env, _cache.dtcpeer);

	_cache.dtcpeer = (jobject)NULL;
   }

   if (_cache.targetMap != (TargetMapPtr*)NULL) {
	int i;

	for (i = 0; i < _cache.nTargets; i++) {
	    (*env)->DeleteGlobalRef(env, _cache.targetMap[i]->string);
	    XtFree((void *)_cache.targetMap[i]);
	}

	XtFree((void*)_cache.targetMap);
	_cache.targetMap = (TargetMapPtr*)NULL;
	_cache.nTargets  = (Cardinal)0;

	(*env)->DeleteGlobalRef(env, _cache.targets);
	_cache.targets = (jobject)NULL;
    }

    _cache.dropDone         = False;
    _cache.transfersPending = 0;
    _cache.transfer         = (Widget)NULL;
}

/**
 *
 */

static void update_cache(JNIEnv* env, Widget w, Widget dt) {
    if(w != _cache.w) {
        struct ComponentData* cdata   = (struct ComponentData *)NULL;
	Arg		      args[1] = { XmNuserData, (XtArgVal)&_cache.peer };

	flush_cache(env);

	if (w == (Widget)NULL) return;

	XtGetValues(w, args, 1);

	if (JNU_IsNull(env, _cache.peer)) {
	    _cache.w = NULL;

            return;
	}

	cdata = (struct ComponentData *) (*env)->GetLongField(env, _cache.peer, mComponentPeerIDs.pData);

        if (cdata         == 0 ||
	    cdata->widget != w ||
	    cdata->dsi    == (DropSitePtr)NULL) {
	    _cache.w = NULL;

            return;
	}

	_cache.w         = w;
	_cache.component = (*env)->NewGlobalRef(env, cdata->dsi->component);
	_cache.peer      = (*env)->NewGlobalRef(env, _cache.peer);
	//SECURITY: OK to call this on privileged thread - peer is secure
	_cache.dtcpeer   = (*env)->CallStaticObjectMethod(env,
						          dTCClazz,
						  	  dTCcreate						   );

	_cache.dtcpeer          = (*env)->NewGlobalRef(env, _cache.dtcpeer);

        _cache.transfersPending = 0;
        _cache.dropDone         = False;
    }

    if (_cache.w != (Widget)NULL) updateCachedTargets(env, dt);
}


/**
 *
 */

static jint convertModifiers(Display* dpy, XEvent* event) {
    jint 	 modifiers = 0;
    unsigned int xMods	   = 0;

    if (event != (XEvent*)NULL) {
	switch (event->type) {
	    case KeyPress:
	    case KeyRelease:
	        modifiers = getModifiers(event->xkey.state);
	        break;

	    case ButtonPress:
	    case ButtonRelease:
	        modifiers = getModifiers(event->xbutton.state);
	        break;
	}
    } else {
	Window dummyw;
	int    dummyi;

	XQueryPointer(dpy,
		      DefaultRootWindow(dpy),
		      &dummyw,
		      &dummyw,
		      &dummyi,
		      &dummyi,
		      &dummyi,
		      &dummyi,
		      &xMods
	);
    }

    return getModifiers(xMods);
}

/**
 *
 */

static jint XmToDnDConstants(int operations) {
    jint src = java_awt_dnd_DnDConstants_ACTION_NONE;

    if (operations & XmDROP_MOVE) src |= java_awt_dnd_DnDConstants_ACTION_MOVE;
    if (operations & XmDROP_COPY) src |= java_awt_dnd_DnDConstants_ACTION_COPY;
    if (operations & XmDROP_LINK) src |= java_awt_dnd_DnDConstants_ACTION_LINK;

    return src;
}

/**
 *
 */

static int DnDConstantsToXm(jint actions) {
    int ret = XmDROP_NOOP;

    if (actions & java_awt_dnd_DnDConstants_ACTION_COPY) ret |= XmDROP_COPY;
    if (actions & java_awt_dnd_DnDConstants_ACTION_MOVE) ret |= XmDROP_MOVE;
    if (actions & java_awt_dnd_DnDConstants_ACTION_LINK) ret |= XmDROP_LINK;

    return ret;
}

/**
 *
 */

static void awt_XmDragProc(Widget w, XtPointer closure, XmDragProcCallbackStruct* cbstruct) {
    JNIEnv* env       = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject component = (jobject)NULL;
    jint    src       = java_awt_dnd_DnDConstants_ACTION_NONE;
    jint    ret       = java_awt_dnd_DnDConstants_ACTION_NONE;

    (*env)->PushLocalFrame(env, 0);

    src = XmToDnDConstants(cbstruct->operations);

    update_cache(env, w, cbstruct->dragContext);

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	flush_cache(env);
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
	goto wayout;
    }

    AWT_UNLOCK();

    switch (cbstruct->reason) {
	case XmCR_DROP_SITE_ENTER_MESSAGE: {

	    //SECURITY: OK to call this on privileged thread - peer is secure
	    ret = (*env)->CallIntMethod(env,
					_cache.dtcpeer,
					dTCenter,
					_cache.component,
					cbstruct->x,
					cbstruct->y,
					XmToDnDConstants(cbstruct->operation),
					src,
					_cache.targets,
					(jlong)(cbstruct->dragContext)
		   );

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		flush_cache(env);
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }
	}
	break;

	case XmCR_DROP_SITE_LEAVE_MESSAGE: {

	    //SECURITY: OK to call this on privileged thread - peer is secure
	    (*env)->CallVoidMethod(env,
				   _cache.dtcpeer,
				   dTCexit,
				   _cache.component,
				   (jlong)(cbstruct->dragContext)
	    );

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		flush_cache(env);
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }

	    ret = java_awt_dnd_DnDConstants_ACTION_NONE;

	    /* now cleanup */

	   flush_cache(env);
	}
	break;

	case XmCR_DROP_SITE_MOTION_MESSAGE: {

	    //SECURITY: OK to call this on privileged thread - peer is secure
	    ret = (*env)->CallIntMethod(env,
					_cache.dtcpeer,
					dTCmotion,
					_cache.component,
					cbstruct->x,
					cbstruct->y,
					XmToDnDConstants(cbstruct->operation),
					JNI_FALSE,
					src,
					_cache.targets,
					(jlong)(cbstruct->dragContext)
		   );

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		flush_cache(env);
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }

	}
	break;

	case XmCR_OPERATION_CHANGED: {

	    //SECURITY: OK to call this on privileged thread - peer is secure
	    ret = (*env)->CallIntMethod(env,
					_cache.dtcpeer,
					dTCmotion,
					_cache.component,
					cbstruct->x,
					cbstruct->y,
					XmToDnDConstants(cbstruct->operation),
					JNI_TRUE,
					src, 
					_cache.targets,
					(jlong)(cbstruct->dragContext)
		   );

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		flush_cache(env);
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
	    }

	}
	break;

	default: break;
    }

    wayout: AWT_LOCK();

    cbstruct->operation  = DnDConstantsToXm(ret) & cbstruct->operations;

    if (ret != java_awt_dnd_DnDConstants_ACTION_NONE) {
        cbstruct->dropSiteStatus = XmVALID_DROP_SITE;
    }  else {
	cbstruct->dropSiteStatus = XmINVALID_DROP_SITE;
    }

    (*env)->PopLocalFrame(env, 0);
}

/**
 *
 */

typedef struct {
    jobject dtcpeer;
    Widget  dragContext;
    int     dropOperation;
} NotifyDropClosureRec, *NotifyDropClosurePtr;

static void _notifyDropDispatched(XtPointer closure, XtIntervalId* na) {
    NotifyDropClosurePtr ndcp = (NotifyDropClosurePtr)closure;

    /* notify the peer that the drop acknowledgement has been dispatched */

    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    //SECURITY: OK to call this on privileged thread - peer is secure
    (*env)->CallVoidMethod(env, ndcp->dtcpeer, dTCdropDisp);

    /*
     * if we failed the drop start a transfer that notifies failure ...
     * this causes src side callbacks to be processed.
     */

    if (ndcp->dropOperation == XmDROP_NOOP) {
	Arg arg;

	arg.name  = XmNtransferStatus;
	arg.value = (XtArgVal)XmTRANSFER_FAILURE;

	(void)XmDropTransferStart(ndcp->dragContext, &arg, 1);

	flush_cache(env);
    }

    (*env)->DeleteGlobalRef(env, ndcp->dtcpeer);

    XtFree((void*)ndcp);
}

/**
 *
 */

static void awt_XmDropProc(Widget w, XtPointer closure, XmDropProcCallbackStruct* cbstruct) {
    JNIEnv* env	      = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jint    src       = java_awt_dnd_DnDConstants_ACTION_NONE;
    jint    ret       = java_awt_dnd_DnDConstants_ACTION_NONE;
    int     operation = cbstruct->operation;

    jobject dtcpeer;

    (*env)->PushLocalFrame(env, 0);


    src = XmToDnDConstants(cbstruct->operations);

    update_cache(env, w, cbstruct->dragContext);

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);

	goto wayout;
    }

    dtcpeer = (*env)->NewGlobalRef(env, _cache.dtcpeer);

    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    ret = (*env)->CallIntMethod(env,
				dtcpeer,
				dTCdrop,
				_cache.component,
				cbstruct->x,
				cbstruct->y,
				XmToDnDConstants(operation),
				src,
				_cache.targets,
				(jlong)(cbstruct->dragContext)
	   );

    AWT_LOCK();

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	flush_cache(env);
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }

    wayout:

    if (ret != java_awt_dnd_DnDConstants_ACTION_NONE) {
	int operation = DnDConstantsToXm(ret);

        cbstruct->operation       = operation;
        cbstruct->operations     &= operation;
        cbstruct->dropSiteStatus  = XmVALID_DROP_SITE;


    } else {
        cbstruct->operation      = XmDROP_NOOP;
        cbstruct->operations     = XmDROP_NOOP;
        cbstruct->dropSiteStatus = XmINVALID_DROP_SITE;
    }

    {
        NotifyDropClosurePtr ndcp = (NotifyDropClosurePtr)XtCalloc(1, sizeof(NotifyDropClosureRec));

        ndcp->dtcpeer       = dtcpeer;
        ndcp->dragContext   = cbstruct->dragContext;
        ndcp->dropOperation = cbstruct->operation;

        XtAppAddTimeOut(awt_appContext, 0UL, _notifyDropDispatched, (XtPointer)ndcp);
    }


    (*env)->PopLocalFrame(env, 0);
}

/**
 *
 */

static void awt_XmTransferProc(Widget w, XtPointer closure, Atom* selection, Atom* type, XtPointer value, unsigned long* length, int* format) {
    JNIEnv*  env        = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    Atom     req        = (Atom)closure;
    Display* dpy        = XtDisplayOfObject(w);
    Atom     keepAlive  = XInternAtom(dpy, _XA_JAVA_KEEP_ALIVE, False);
    Atom     filename   = XInternAtom(dpy, _XA_FILENAME, False);
    Atom     dtfilename = XInternAtom(dpy, _DT_FILENAME, False);
    jobject  tName;

    /*
     * Note: this method is only called to transfer data between clients
     * in different JVM's or native apps. For Intra-JVM transfers the peer
     * code shares the sources Transferable with the destination.
     */

    if (_cache.w == (Widget)NULL || _cache.transfer != w) return; /* we have already cleaned up ... */

    (*env)->PushLocalFrame(env, 0);

    if (req != keepAlive) {
	if (*type == None || *type == XT_CONVERT_FAIL) {
            //SECURITY: OK to call this on privileged thread - peer is secure
	    (*env)->CallVoidMethod(env,
				   _cache.dtcpeer,
				   dTCtxFailed,
				   findCachedTarget(req)
            );	
	} else if (req == filename || req == dtfilename) {
	    XTextProperty tp;

	    char**        strings  = (char **)NULL;
	    int           nstrings = 0;
	    Status        s;
	    jobject       filenames;
	    int           i;

	    tp.encoding = *type;
	    tp.value    = (unsigned char *)value;
	    tp.nitems   = *length;
	    tp.format   = *format;

	    if (XTextPropertyToStringList(&tp, &strings, &nstrings) == 0 || nstrings == 0) {
                //SECURITY: OK to call this on privileged thread - peer is secure
	        (*env)->CallVoidMethod(env,
				       _cache.dtcpeer,
				       dTCtxFailed,
				       findCachedTarget(req)
                );	

		return;
	    }

	    filenames = (jobjectArray)(*env)->NewObjectArray(env, (jsize)nstrings, file, (jobject)NULL);

	   if (JNU_IsNull(env, filenames) || !JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		XFreeStringList(strings);

		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);

                //SECURITY: OK to call this on privileged thread - peer is secure
	        (*env)->CallVoidMethod(env,
				       _cache.dtcpeer,
				       dTCtxFailed,
				       findCachedTarget(req)
                );	

		return;
	    }

	    for (i = 0; i < nstrings; i++) {
		jobject fileobj;
	   	jstring str = (*env)->NewStringUTF(env, (const char *)strings[i]);
		if (!JNU_IsNull(env, str))
		    str = (jstring)(*env)->NewGlobalRef(env, str);
		

	   	if (JNU_IsNull(env, str) || !JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    XFreeStringList(strings);

		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);

                    //SECURITY: OK to call this on privileged thread - peer is secure
	            (*env)->CallVoidMethod(env,
				           _cache.dtcpeer,
				           dTCtxFailed,
				           findCachedTarget(req)
                    );	

		    return;
		}

		fileobj = (*env)->NewGlobalRef(env, (*env)->NewObject(env, file, dTCfilei, str));


	        (*env)->DeleteGlobalRef(env, str);

	   	if (JNU_IsNull(env, fileobj) || !JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    XFreeStringList(strings);

		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);

                    //SECURITY: OK to call this on privileged thread - peer is secure
	            (*env)->CallVoidMethod(env,
				           _cache.dtcpeer,
				           dTCtxFailed,
				           findCachedTarget(req)
                    );	

		    return;
		}

	        (*env)->SetObjectArrayElement(env, filenames, (jsize)i, fileobj);

	   	if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    XFreeStringList(strings);

		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);

                    //SECURITY: OK to call this on privileged thread - peer is secure
	            (*env)->CallVoidMethod(env,
				           _cache.dtcpeer,
				           dTCtxFailed,
				           findCachedTarget(req)
                    );	

		    return;
		}
	    }

	    XFreeStringList(strings);

            //SECURITY: OK to call this on privileged thread - peer is secure
	    filenames = (*env)->CallStaticObjectMethod(env, dTCClazz, dTCgetL, filenames);

	    if (JNU_IsNull(env, filenames) || !JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	    } else {
	        char* tn = XGetAtomName(dpy, *type);

		tName = (*env)->NewStringUTF(env, (const char *)tn);

                if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		   (*env)->ExceptionDescribe(env);
		   (*env)->ExceptionClear(env);
		}

		XFree((void *)tn);
	    }

            AWT_UNLOCK();
 
            //SECURITY: OK to call this on privileged thread - peer is secure
            (*env)->CallVoidMethod(env,
                                   _cache.dtcpeer,
                                   dTCnewData,
                                   findCachedTarget(req),
                                   tName,
                                   filenames
            );
 
            AWT_LOCK();
 
            if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
                (*env)->ExceptionDescribe(env);
                (*env)->ExceptionClear(env);
            }
	} else switch (*format) {
	    case  8:
	    case 16:
	    case 32: {
	    	jbyteArray arry = (*env)->NewByteArray(env, *length);

                if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);

                    //SECURITY: OK to call this on privileged thread - peer is secure
	    	    (*env)->CallVoidMethod(env,
				           _cache.dtcpeer,
				   	   dTCtxFailed,
				   	   findCachedTarget(req)
            	    );	

		    goto wayout;
	        }

	    	(*env)->SetByteArrayRegion(env, arry, 0, *length, (jbyte*)value);
                if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);

                    //SECURITY: OK to call this on privileged thread - peer is secure
	    	    (*env)->CallVoidMethod(env,
				           _cache.dtcpeer,
				   	   dTCtxFailed,
				   	   findCachedTarget(req)
            	    );	
		    goto wayout;
    	        }

	    	arry = (*env)->NewGlobalRef(env, arry);

                if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);
		}

		{
		    char* tn = XGetAtomName(dpy, *type);

		    tName = (*env)->NewStringUTF(env, (const char *)tn);

                    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    	(*env)->ExceptionDescribe(env);
		    	(*env)->ExceptionClear(env);
		    }

		    XFree((void *)tn);
		}
		

		AWT_UNLOCK();

                //SECURITY: OK to call this on privileged thread - peer is secure
	    	(*env)->CallVoidMethod(env,
				       _cache.dtcpeer,
				       dTCnewData,
				       findCachedTarget(req),
				       tName,
				       arry
	        );

		AWT_LOCK();

                if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);
    	        }
	    }

	    default:
	        break;
        }
    }

    wayout:
	if (_cache.dropDone) {
	    if (_cache.flushPending) flush_cache(env);
	} else {
            //SECURITY: OK to call this on privileged thread - peer is secure

	    AWT_UNLOCK();

	    (*env)->CallVoidMethod(env, _cache.dtcpeer, dTCheartbeat);

	    AWT_LOCK();

            if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	        (*env)->ExceptionDescribe(env);
	        (*env)->ExceptionClear(env);
            }
	}
    (*env)->PopLocalFrame(env, 0);
}
 
/**
 *
 */

static Boolean awt_XmConvertProc(Widget w, Atom* selection, Atom* target, Atom* type, XtPointer* value, unsigned long* length, int* format) {
    unsigned long maxLength = 0x7fffffffL;
    XtRequestId   rid	    = (XtRequestId)0;
    XtPointer     closure   = (XtPointer)NULL;

    XtVaGetValues(w, XmNclientData, (XtArgVal)&closure, NULL);

    return awt_XmConvertIncrProc(w, selection, target, type, value, length, format, &maxLength, closure, &rid);

}

/**
 *
 */

static Boolean awt_XmConvertIncrProc(Widget w, Atom* selection, Atom* target, Atom* type, XtPointer* value, unsigned long* length, int* format, unsigned long* max_length, XtPointer closure, XtRequestId* receiver_id) {
    JNIEnv*  env  = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject  this = (jobject)closure;
    Display* dpy  = XtDisplayOfObject(w);

    (*env)->PushLocalFrame(env, 0);

    if (*target == XInternAtom(dpy, _XA_JAVA_KEEP_ALIVE, False)) {
	*type   = *target;
	*value  = (XtPointer)NULL;
	*length = 0L;
	*format = 32;
    } else if (*target == XInternAtom(dpy, _XA_DELETE, False)) {

	/*
	 * acknowledge the DELETE target here ... the "delete" semantic
	 * of move will take place after the drop is complete.
	 */

	*type   = XInternAtom(dpy, _XA_NULL, False);
	*length = 0;
	*value  = (XtPointer)NULL;
    } else if (*target == XInternAtom(dpy, _XA_HOSTNAME, False)) {
	struct utsname name;
	XTextProperty  tp;

	uname(&name);

	XStringListToTextProperty(&name.nodename, 1, &tp);

	*value = (XtPointer)XtCalloc(tp.nitems, sizeof(char));

	memcpy((void *)*value, (const void *)tp.value, tp.nitems);

	XFree((void *)tp.value);

	*type   = tp.encoding;
	*length = tp.nitems + 1;
	*format = tp.format;
    } else if (*target == XInternAtom(dpy, _XA_FILENAME, False) ||
	       *target == XInternAtom(dpy, _DT_FILENAME, False)) {
	XTextProperty tp;
	jobjectArray  filenames;
	int	      len;
	int	      i;
	char**        stringList;
	Status	      s;

	*type   = XT_CONVERT_FAIL; /* assume forthcoming error */
	*value  = (XtPointer)NULL;
	*length = 0L;
	*format = 8;

	AWT_UNLOCK();

        //SECURITY: OK to call this on privileged thread - peer is secure
	filenames = (jobjectArray)(*env)->CallObjectMethod(env,
					     		   this,
					     		   dSCconvertFiles,
					     		   (*env)->NewStringUTF(env, XGetAtomName(dpy, *target))
	            );

	AWT_LOCK();

	if (JNU_IsNull(env, filenames) || !JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	    (*env)->ExceptionDescribe(env);
	    (*env)->ExceptionClear(env);

    	    (*env)->PopLocalFrame(env, 0);

	    return False;
	}

	len = (*env)->GetArrayLength(env, filenames);

	if (len == 0) {
	    (*env)->ExceptionDescribe(env);
	    (*env)->ExceptionClear(env);

    	    (*env)->PopLocalFrame(env, 0);

	    return False;
	}

	stringList = (char **)XtCalloc(len, sizeof(char *));

	if (stringList == (char**)NULL) {
	    (*env)->ExceptionDescribe(env);
	    (*env)->ExceptionClear(env);

    	    (*env)->PopLocalFrame(env, 0);

	    return False;
	}

	for (i = 0; i < len; i++) {
	    jstring  str  = (*env)->GetObjectArrayElement(env, filenames, (jsize)i);
	    int      slen;
	    char*    utf;
	    jboolean isCopy;

	    if (JNU_IsNull(env, str) || !JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	    	int j;

	        (*env)->ExceptionDescribe(env);
	        (*env)->ExceptionClear(env);

		for (j = 0; j < i; j++) {
		    XtFree((void *)stringList[j]);
		}

		XtFree((void*)stringList);

    	        (*env)->PopLocalFrame(env, 0);

	        return False;
	    }

	    slen = (*env)->GetStringUTFLength(env, str) + 1;
	    utf  = (char *)(*env)->GetStringUTFChars(env, str, &isCopy);

	    stringList[i] = (char*)XtCalloc(slen, sizeof(char));

	    if (stringList[i] == (char *)NULL) {
	    	int j;

		for (j = 0; j < i; j++) {
		    XtFree((void *)stringList[j]);
		}

    	        (*env)->PopLocalFrame(env, 0);

	        return False;
	    }

	    memcpy((void *)stringList[i], (const void*)utf, slen);

	    (*env)->ReleaseStringUTFChars(env, str, (const char *)utf);
	}
	
	s = XStringListToTextProperty(stringList, len, &tp);

	for (i = 0; i < len; i++) {
	    XtFree((void*)stringList[i]);
	}

	XtFree((void*)stringList);

	if (s == 0) return False;

	*value = (XtPointer)XtCalloc(tp.nitems, sizeof(char));

	if (*value == (XtPointer)NULL) {
	    XFree((void*)tp.value);

    	    (*env)->PopLocalFrame(env, 0);

	    return False;
	}

	memcpy((void *)(*value), (const void *)tp.value, tp.nitems);

	XFree((void*)tp.value);

	*length = tp.nitems;
	*type   = tp.encoding;
	*format = tp.format;
    } else {
	jobject ret;
	jint    mlen = (*max_length > 0x7fffffff ? 0x7fffffff : (int)*max_length);

	*type   = XT_CONVERT_FAIL; /* assume forthcoming error */
	*value  = (XtPointer)NULL;
	*length = 0L;
	*format = 8;

	/*
 	 * we get a Buffer object from the MDragSourceContextPeer, then we
 	 * process the Buffer, copying it into a single block of contiguous
	 * memory that we return to Xt.
 	 * 
	 */

	AWT_UNLOCK();

        //SECURITY: OK to call this on privileged thread - peer is secure
	ret = (*env)->CallObjectMethod(env,
				       this,
				       dSCconvert,
				       (*env)->NewStringUTF(env, XGetAtomName(dpy, *target)),
				       (jlong)*receiver_id
	      );
						
	AWT_LOCK();
	
	if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	    (*env)->ExceptionDescribe(env);
	    (*env)->ExceptionClear(env);

    	    (*env)->PopLocalFrame(env, 0);

	    return False;
	}

	if (JNU_IsNull(env, ret)) {

    	    (*env)->PopLocalFrame(env, 0);

	    return False;
	} else {
            jint    len;
	    jint    rcurs;
	    jbyte*  copy;
	    jobject buf;
	    
            //SECURITY: OK to call this on privileged thread - peer is secure
	    len = (*env)->CallIntMethod(env, ret, dSCBsize);

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);

    		(*env)->PopLocalFrame(env, 0);

		return False;
	    }

	    if (len == 0) {
	        *type   = *target;
	     	*format = 8;

    		(*env)->PopLocalFrame(env, 0);

		return True;
	    }

            //SECURITY: OK to call this on privileged thread - peer is secure
	    rcurs = (*env)->CallIntMethod(env, ret, dSCBirc, 0);

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);

    		(*env)->PopLocalFrame(env, 0);

		return False;
	    }

	    if (rcurs == -1) { /* EOF */
                //SECURITY: OK to call this on privileged thread - peer is secure
		(*env)->CallVoidMethod(env, this, dSCfree, ret);

	    	if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);

    		    (*env)->PopLocalFrame(env, 0);

		    return False;
		} else {
	            *type   = *target;
	     	    *format = 8;

    		    (*env)->PopLocalFrame(env, 0);

		    return True;
		}
	    }

	    len = len - rcurs > mlen ? mlen : len - rcurs;

	    copy = (jbyte*)XtMalloc(len * sizeof(jbyte));

	    if (copy == (jbyte*)NULL) {
		JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");

    		(*env)->PopLocalFrame(env, 0);
 
		return False;
	    }

            //SECURITY: OK to call this on privileged thread - peer is secure
	    buf = (*env)->CallObjectMethod(env, ret, dSCBgbuf);

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);

		XtFree((void *)copy);

    		(*env)->PopLocalFrame(env, 0);

	        return False;
	    }

	    (*env)->GetByteArrayRegion(env, (jbyteArray)buf, rcurs, len, copy);

            //SECURITY: OK to call this on privileged thread - peer is secure
	    rcurs = (*env)->CallIntMethod(env, ret, dSCBirc, len);

	    if ((receiver_id== (XtRequestId*)NULL || *receiver_id == (XtRequestId)NULL) && rcurs == -1) { /* EOF */
                //SECURITY: OK to call this on privileged thread - peer is secure
		(*env)->CallVoidMethod(env, this, dSCfree, ret);

	    	if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		    (*env)->ExceptionDescribe(env);
		    (*env)->ExceptionClear(env);
		}
	    }

	    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);

		XtFree((void *)copy);

    		(*env)->PopLocalFrame(env, 0);

	        return False;
	    }

	    *value  = (XtPointer)copy;
	    *type   = *target;
	    *length = len;
	    *format = 8;
	}
    }

    (*env)->PopLocalFrame(env, 0);

    return True;
}

/**
 *
 */

static void awt_XmDragEnterProc(Widget w, XtPointer closure, XmDropSiteEnterCallbackStruct* cbstruct) {
    JNIEnv*  env   = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject  this  = (jobject)closure;
    jint     ret;
    jboolean valid = cbstruct->dropSiteStatus == XmVALID_DROP_SITE ? JNI_TRUE : JNI_FALSE;

    (*env)->PushLocalFrame(env, 0);

    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    ret = (*env)->CallIntMethod(env,
				this,
			  	dSCenter,
				valid == JNI_TRUE ? XmToDnDConstants(cbstruct->operation) : java_awt_dnd_DnDConstants_ACTION_NONE,
				XmToDnDConstants(cbstruct->operations),
				convertModifiers(XtDisplayOfObject(w), cbstruct->event),
				valid
    );

    AWT_LOCK();

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }


    cbstruct->operation = DnDConstantsToXm(ret);

    (*env)->PopLocalFrame(env, 0);
}

/**
 *
 */

static void awt_XmDragMotionProc(Widget w, XtPointer closure, XmDragMotionCallbackStruct* cbstruct) {
    JNIEnv*  env   = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject  this  = (jobject)closure;
    jint     ret;
    jboolean valid = cbstruct->dropSiteStatus == XmVALID_DROP_SITE ? JNI_TRUE : JNI_FALSE;

    if (cbstruct->dropSiteStatus == XmNO_DROP_SITE) return;

    (*env)->PushLocalFrame(env, 0);


    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    ret = (*env)->CallIntMethod(env,
				this,
			  	dSCmotion,
				valid == JNI_TRUE ? XmToDnDConstants(cbstruct->operation) : java_awt_dnd_DnDConstants_ACTION_NONE,
				XmToDnDConstants(cbstruct->operations),
				convertModifiers(XtDisplayOfObject(w), cbstruct->event),
				valid
    );

    AWT_LOCK();

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }

    cbstruct->operations = DnDConstantsToXm(ret);


    (*env)->PopLocalFrame(env, 0);
}

/**
 *
 */

static void awt_XmDragLeaveProc(Widget w, XtPointer closure, XmDropSiteLeaveCallbackStruct* cbstruct) {
    JNIEnv* env  = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject)closure;
    jint    ret;

    (*env)->PushLocalFrame(env, 0);

    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    (*env)->CallVoidMethod(env, this, dSCexit, convertModifiers(XtDisplayOfObject(w), cbstruct->event));

    AWT_LOCK();

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }

    (*env)->PushLocalFrame(env, 0);
}

/**
 *
 */

static void awt_XmDropStartProc(Widget w, XtPointer closure, XmDropStartCallbackStruct* cbstruct) {
    JNIEnv* env  = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject)closure;
    jboolean valid = cbstruct->dropSiteStatus == XmVALID_DROP_SITE ? JNI_TRUE : JNI_FALSE;

    (*env)->PushLocalFrame(env, 0);

    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    (*env)->CallVoidMethod(env,
			   this,
			   dSCdstart,
			   XmToDnDConstants(cbstruct->operation),
			   valid
    );

    AWT_LOCK();

    if (cbstruct->operation == XmDROP_NOOP || cbstruct->dropAction != XmDROP)
	cbstruct->dropAction = XmDROP_CANCEL;

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }

    (*env)->PopLocalFrame(env, 0);
}

/**
 *
 */

static void awt_XmDropOperationChangedProc(Widget w, XtPointer closure, XmDropStartCallbackStruct* cbstruct) {
    JNIEnv*  env   = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject  this  = (jobject)closure;
    jint     ret;
    jboolean valid = cbstruct->dropSiteStatus == XmVALID_DROP_SITE ? JNI_TRUE : JNI_FALSE;

    (*env)->PushLocalFrame(env, 0);


    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    ret = (*env)->CallIntMethod(env,
				this,
			  	dSCchanged,
				XmToDnDConstants(cbstruct->operation),
				XmToDnDConstants(cbstruct->operations),
				convertModifiers(XtDisplayOfObject(w), cbstruct->event),
				valid
    );

    AWT_LOCK();

    if (!JNU_IsNull(env, (*env)->ExceptionOccurred(env))) {
	(*env)->ExceptionDescribe(env);
	(*env)->ExceptionClear(env);
    }


    (*env)->PopLocalFrame(env, 0);
}

/**
 *
 */

static void awt_XmDropFinishProc(Widget w, XtPointer closure, XmDropFinishCallbackStruct* cbstruct) {
    JNIEnv* env  = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject)closure;

    (*env)->PushLocalFrame(env, 0);

    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    (*env)->CallVoidMethod(env, this, dSCdfinished, (jboolean)cbstruct->completionStatus);

    AWT_LOCK();

    (*env)->PopLocalFrame(env, 0);
}

/**
 *
 */

static void awt_XmDragDropFinishProc(Widget w, XtPointer closure, XmDropFinishCallbackStruct* cbstruct) {
    JNIEnv* env  = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject this = (jobject)closure;
    Atom*   targets;

    /* cleanup */

    (*env)->PushLocalFrame(env, 0);

    AWT_UNLOCK();

    //SECURITY: OK to call this on privileged thread - peer is secure
    (*env)->CallVoidMethod(env, this, dSCddfinished);

    AWT_LOCK();

    (*env)->DeleteGlobalRef(env, this);

    (*env)->PopLocalFrame(env, 0);
}

/******************************************************************************/
 
void sun_awt_motif_MComponentPeer_addNativeDropTarget(void* p1, void* p2, void* p3) { }
void sun_awt_motif_MComponentPeer_removeNativeDropTarget(void* p1, void* p2, void* p3) { }
void sun_awt_motif_MDropTargetContextPeer_dropDone(void* p1, void* p2, void* p3) { }
void sun_awt_motif_MDropTargetContextPeer_startTransfer(void* p1, void* p2, void* p3) { }
void sun_awt_motif_MDropTargetContextPeer_addTransfer(void* p1, void* p2, void* p3) { }
void sun_awt_motif_MDropTargetContextPeer_keepAlive(void* p1, void* p2, void* p3) { }
void sun_awt_motif_AWTLockAccess_awtUnlock(void* p1, void* p2, void* p3) { }
void sun_awt_motif_AWTLockAccess_awtLock(void* p1, void* p2, void* p3) { }
void sun_awt_motif_MDragSourceContextPeer_startDrag(void* p1, void* p2, void* p3) { }
void sun_awt_motif_MDragSourceContextPeer_setCursor(void* p1, void* p2, void* p3) { }
