/*
 * @(#)awt_util.c	1.53 98/09/18
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "color.h"
#include <X11/IntrinsicP.h>
#include <X11/Xatom.h>
#include <X11/Xmd.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <Xm/MenuShell.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>

#include <jni.h>
#include <jni_util.h>

extern int xerror_handler();
static int winmgr_running = 0;

/*
 * The following three funtions are to work around menu problems
 */

/* 
 * test if there is a menu that has the current focus 
 * called from awt_Dialog.c and awt_Component.c 
 */
int
awt_util_focusIsOnMenu(Display *display)
{
  Window window;
  Widget widget;
  int rtr;

  XGetInputFocus(display, &window, &rtr);
  if (window == None) {
    return 0;
  }

  widget = XtWindowToWidget(display, window);
  if (widget == NULL) {
    return 0;
  }

  if (XtIsSubclass(widget, xmMenuShellWidgetClass)) {
    return 1;
  } else {
    return 0;
  }
}

static 
void fillButtonEvent(XButtonEvent *ev, int type, Display *display, Window window) {
    ev->type = type;
    ev->display = display;
    ev->window = window;
    ev->send_event = True;
    
    ev->root = RootWindow(display, DefaultScreen(display));
    ev->subwindow = (Window)None;
    ev->time = CurrentTime;
    ev->x = 0;
    ev->y = 0;
    ev->x_root = 0;
    ev->y_root = 0;
    ev->same_screen = True;
    ev->button = Button1;
    ev->state = Button1Mask;
}

/* 
 * generates a mouse press event and a release event 
 * called from awt_Dialog.c
 */
int
awt_util_sendButtonClick(Display *display, Window window)
{
  XButtonEvent ev;
  int status;

  fillButtonEvent(&ev, ButtonPress, display, window);
  status = XSendEvent(display, window, True, ButtonPressMask, &ev);

  if (status != 0) {
      fillButtonEvent(&ev, ButtonRelease, display, window);
      status = XSendEvent(display, window, False, ButtonReleaseMask, &ev);
  }
  return status;
}

/* 
 * create a mouse press and a release event, and dispatch them 
 * called from awt_Component.c
 */
void
awt_util_dispatchButtonClick(Display *display, Window window)
{
    XButtonEvent ev;
 
    fillButtonEvent(&ev, ButtonPress, display, window); 
    XtDispatchEvent(&ev);

    fillButtonEvent(&ev, ButtonRelease, display, window);
    XtDispatchEvent(&ev);
}

Widget
awt_util_createWarningWindow(Widget parent, char *warning)
{
    Widget warningWindow;
#ifdef NETSCAPE
    extern Widget FE_MakeAppletSecurityChrome(Widget parent, char* message);
    warningWindow = FE_MakeAppletSecurityChrome(parent, warning);
#else
    Widget label;
    int argc;
    Arg args[10];
    Pixel yellow = defaultConfig->AwtColorMatch(255, 225, 0, defaultConfig);
    Pixel black = defaultConfig->AwtColorMatch(0, 0, 0, defaultConfig);

    argc = 0;
    XtSetArg(args[argc], XmNbackground, yellow); argc++;
    XtSetArg(args[argc], XmNmarginHeight, 0); argc++;
    XtSetArg(args[argc], XmNmarginWidth, 0); argc++;
    warningWindow =  XmCreateForm(parent, "main", args, argc);

    XtManageChild(warningWindow);
    label = XtVaCreateManagedWidget(warning,
				    xmLabelWidgetClass, warningWindow,
				    XmNhighlightThickness, 0,
				    XmNbackground, yellow,
				    XmNforeground, black,
				    XmNalignment, XmALIGNMENT_CENTER,
				    XmNrecomputeSize, False,
				    NULL);
    XtVaSetValues(label,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_FORM,
		  XmNrightAttachment, XmATTACH_FORM,
		  NULL);
#endif
    return warningWindow;
}

void
awt_setWidgetGravity(Widget w, int gravity)
{
    XSetWindowAttributes    xattr;
    Window  win = XtWindow(w);

    if (win != 0) {
	xattr.bit_gravity = StaticGravity;
	xattr.win_gravity = StaticGravity;
	XChangeWindowAttributes(XtDisplay(w), win,
				CWBitGravity|CWWinGravity,
				&xattr);
    }
}

void
awt_util_reshape(Widget w, long x, long y, long wd, long ht) 
{
    Widget parent;
    Dimension ww, wh;
    Position wx, wy;
    Boolean move = False;
    Boolean resize = True;
    Boolean mapped_when_managed;
    Boolean need_to_unmanage = True;

    if (w == 0) {
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
        JNU_ThrowNullPointerException(env,"NullPointerException");
	return;
    }
    parent = XtParent(w);

    /* Aim: hack to prevent direct children of scrollpane from
     * being unmanaged during a reshape operation (which results
     * in too many expose events).  
     */
    if (parent != NULL && XtParent(parent) != NULL &&
        XtIsSubclass(XtParent(parent), xmScrolledWindowWidgetClass)) {
        need_to_unmanage = False;
    } 

    XtVaGetValues(w, 
		  XmNwidth, &ww, 
		  XmNheight, &wh, 
		  XmNx, &wx, 
		  XmNy, &wy, 
		  NULL);   

    if (x != wx || y != wy) {
        move = True;
    }
    if (wd != ww || ht != wh) {
        resize = True;
    }
    if (!move && !resize) {
        return;
    }

    if (need_to_unmanage) {    
        if (!resize) {
            mapped_when_managed = w->core.mapped_when_managed;
            w->core.mapped_when_managed = False;
        } 
        XtUnmanageChild(w);
    }
       
    /* GES: AVH's hack: 
     * Motif ignores attempts to move a toplevel window to 0,0.
     * Instead we set the position to 1,1. The expected value is
     * returned by Frame.getBounds() since it uses the internally
     * held rectangle rather than querying the peer.
     * N.B. [pauly, 9/97]  This is only required for wm shells
     * under the Motif Window Manager (MWM), not for any others.
     * Note. Utilizes C short-circuiting if w is not a wm shell.
     */
    if ((x == 0) && (y == 0) &&
        (XtIsSubclass(w, wmShellWidgetClass)) &&
        (XmIsMotifWMRunning(w))) {
	XtVaSetValues(w, XmNx, 1, XmNy, 1, NULL);
    }

    if (move && !resize) {
        XtVaSetValues(w, XmNx, x, XmNy, y, NULL);
    
    } else if (resize && !move) {
        XtVaSetValues(w, 
		      XmNwidth, (wd > 0) ? wd : 1,
		      XmNheight, (ht > 0) ? ht : 1,
		      NULL);

    } else  {
        XtVaSetValues(w,
		  XmNx, x,
		  XmNy, y,
		  XmNwidth, (wd > 0) ? wd : 1,
		  XmNheight, (ht > 0) ? ht : 1,
		  NULL);
    }

    if (need_to_unmanage) {    
        XtManageChild(w);
        if (!resize) {   
            w->core.mapped_when_managed = mapped_when_managed;
        }
    }   
}

void
awt_util_hide(Widget w) 
{
    if (w == 0) {
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
        JNU_ThrowNullPointerException(env,"NullPointerException");
	return;
    }
    XtSetMappedWhenManaged(w, False);
}

void
awt_util_show(Widget w) 
{
/*
    extern Boolean  scrollBugWorkAround;
*/
    if (w == 0) {
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
        JNU_ThrowNullPointerException(env,"NullPointerException");
	return;
    }
    XtSetMappedWhenManaged(w, True);
/*
  XXX: causes problems on 2.5
    if (!scrollBugWorkAround) {
	awt_setWidgetGravity(w, StaticGravity);
    }
*/
}

void
awt_util_enable(Widget w) 
{
    XtSetSensitive(w, True);
}

void
awt_util_disable(Widget w) 
{
    XtSetSensitive(w, False);
}

void
awt_util_mapChildren(Widget w, void (*func)(Widget,void *),
		     int applyToCurrent, void *data) {
    WidgetList			wlist;
    Cardinal			wlen = 0;
    int				i;

    /* The widget may have been destroyed by another thread. */
    if ((w == 0) || (!XtIsObject(w)) || (w->core.being_destroyed))
        return;

    if (applyToCurrent != 0) {
	(*func)(w, data);
    }
    if (!XtIsComposite(w)) {
	return;
    }

    XtVaGetValues(w,
		  XmNchildren, &wlist,
		  XmNnumChildren, &wlen,
		  NULL);
    if (wlen > 0) {
	for (i=0; i < wlen; i++) {
	    awt_util_mapChildren(wlist[i], func, 1, data);
	}
    }
}

void
awt_changeAttributes(Display *dpy, Widget w, unsigned long mask,
		     XSetWindowAttributes *xattr)
{
    int			i;
    WidgetList		wlist;
    Cardinal		wlen = 0;

    if (XtWindow(w) && XtIsRealized(w)) {
	XChangeWindowAttributes(dpy,
				XtWindow(w),
				mask,
				xattr);
    } else {
	return;
    }
    XtVaGetValues(w,
		  XmNchildren, &wlist,
		  XmNnumChildren, &wlen,
		  NULL);
    for (i = 0; i < wlen; i++) {
	if (XtWindow(wlist[i]) && XtIsRealized(wlist[i])) {
	    XChangeWindowAttributes(dpy,
				    XtWindow(wlist[i]),
				    mask,
				    xattr);
	}
    }
}

int
awt_util_setCursor(Widget w, Cursor c) {

    if (XtIsRealized(w)) {

	unsigned long 		valuemask = 0;
	XSetWindowAttributes	attributes;

	valuemask = CWCursor;
	attributes.cursor = c;
	XChangeWindowAttributes(awt_display,
				XtWindow(w),
				valuemask,
				&attributes);				
	return 1;
    } else
        return 0;    
}



/*
 * awt_util_getIMStatusHeight is a cut and paste of the ImGetGeo() function
 * found in CDE Motif's Xm/XmIm.c.  It returns the height of the Input Method
 * Status region attached to the given VendorShell.  This is needed in order
 * to calculate geometry for Frames and Dialogs that contain TextField or
 * TextArea widgets.
 *
 * BCB: Copying this function out of the Motif source is a horrible
 * hack. Unfortunately, Motif tries to hide the existence of the IM Status
 * region from us so it does not provide any public way to get this info.
 * Clearly a better long term solution needs to be found.
 */

typedef struct _XmICStruct {
    struct _XmICStruct *next;
    Widget icw;
    Window focus_window;
    XtArgVal foreground;
    XtArgVal background;
    XtArgVal background_pixmap;
    XtArgVal font_list;
    XtArgVal line_space;
    int status_width;
    int status_height;
    int preedit_width;
    int preedit_height;
    Boolean has_focus;
    Boolean need_reset;
}   XmICStruct;

typedef struct {
    Widget im_widget;
    XIMStyle input_style;
    XIC xic;
    int status_width;
    int status_height;
    int preedit_width;
    int preedit_height;
    XmICStruct *iclist;
    XmICStruct *current;
}   XmImInfo;

static XFontSet extract_fontset(XmFontList);
static XmICStruct *get_iclist(Widget);

#define MAXARGS 10
static Arg xic_vlist[MAXARGS];
static Arg status_vlist[MAXARGS];
static Arg preedit_vlist[MAXARGS];

#define NO_ARG_VAL -1
#define SEPARATOR_HEIGHT 2

static XmICStruct *
get_iclist(Widget w)
{
    Widget p;
    XmVendorShellExtObject ve;
    XmWidgetExtData extData;
    XmImInfo *im_info;

    p = w;
    while (!XtIsShell(p))
	p = XtParent(p);

    extData = _XmGetWidgetExtData((Widget) p, XmSHELL_EXTENSION);
    if (extData == NULL)
	return NULL;

    ve = (XmVendorShellExtObject) extData->widget;
    if ((im_info = (XmImInfo *) ve->vendor.im_info) == NULL)
	return NULL;
    else
	return im_info->iclist;
}

int
awt_util_getIMStatusHeight(Widget vw)
{
    XmICStruct *icp;
    XmVendorShellExtObject ve;
    XmWidgetExtData extData;
    XmImInfo *im_info;
    int width = 0;
    int height = 0;
    XRectangle rect;
    XRectangle *rp;
    int old_height;
    Arg args[1];
    int base_height;
    XFontSet fs;
    XFontSet fss = NULL;
    XFontSet fsp = NULL;

    extData = _XmGetWidgetExtData((Widget) vw, XmSHELL_EXTENSION);
    ve = (XmVendorShellExtObject) extData->widget;

    if ((icp = get_iclist(vw)) == NULL) {
	ve->vendor.im_height = 0;
	return 0;
    }
    im_info = (XmImInfo *) ve->vendor.im_info;
    if (im_info->xic == NULL) {
	ve->vendor.im_height = 0;
	return 0;
    }
    status_vlist[0].name = XNFontSet;
    status_vlist[1].name = NULL;
    preedit_vlist[0].name = XNFontSet;
    preedit_vlist[1].name = NULL;

    xic_vlist[0].name = XNAreaNeeded;
    xic_vlist[1].name = NULL;

    im_info->status_width = 0;
    im_info->status_height = 0;
    im_info->preedit_width = 0;
    im_info->preedit_height = 0;
    for (; icp != NULL; icp = icp->next) {
	if (im_info->input_style & XIMStatusArea) {
	    if (icp->status_height == 0) {
		char *ret;

		if (icp->font_list == NO_ARG_VAL ||
		    (fss = extract_fontset((XmFontList) icp->font_list)) == NULL)
		    continue;

		status_vlist[0].value = (XtArgVal) fss;
		XSetICValues(im_info->xic,
			     XNStatusAttributes, &status_vlist[0],
			     NULL);

		xic_vlist[0].value = (XtArgVal) & rp;
		ret = XGetICValues(im_info->xic,
				   XNStatusAttributes, &xic_vlist[0],
				   NULL);

		if (ret) {
		    /* Cannot obtain XIC value. IM server may be gone. */
		    ve->vendor.im_height = 0;
		    return 0;
		} else {
		    icp->status_width = rp->width;
		    icp->status_height = rp->height;
		    XFree(rp);
		}
	    }
	    if (icp->status_width > im_info->status_width)
		im_info->status_width = icp->status_width;
	    if (icp->status_height > im_info->status_height)
		im_info->status_height = icp->status_height;
	}
	if (im_info->input_style & XIMPreeditArea) {
	    if (icp->preedit_height == 0) {
		if (icp->font_list == NO_ARG_VAL ||
		    (fsp = extract_fontset((XmFontList) icp->font_list)) == NULL)
		    continue;

		preedit_vlist[0].value = (XtArgVal) fsp;
		XSetICValues(im_info->xic,
			     XNPreeditAttributes, &preedit_vlist[0],
			     NULL);

		xic_vlist[0].value = (XtArgVal) & rp;
		XGetICValues(im_info->xic,
			     XNPreeditAttributes, &xic_vlist[0],
			     NULL);

		icp->preedit_width = rp->width;
		icp->preedit_height = rp->height;
		XFree(rp);
	    }
	    if (icp->preedit_width > im_info->preedit_width)
		im_info->preedit_width = icp->preedit_width;
	    if (icp->preedit_height > im_info->preedit_height)
		im_info->preedit_height = icp->preedit_height;
	}
    }

    if (im_info->current != NULL && (fss != NULL || fsp != NULL)) {
	if (im_info->current->font_list != NO_ARG_VAL &&
	    (fs = extract_fontset((XmFontList) im_info->current->font_list))
	    != NULL) {
	    if (fss != NULL)
		status_vlist[0].value = (XtArgVal) fs;
	    else
		status_vlist[0].name = NULL;
	    if (fsp != NULL)
		preedit_vlist[0].value = (XtArgVal) fs;
	    else
		preedit_vlist[0].name = NULL;
	    XSetICValues(im_info->xic,
			 XNStatusAttributes, &status_vlist[0],
			 XNPreeditAttributes, &preedit_vlist[0],
			 NULL);
	}
    }
    if (im_info->status_height > im_info->preedit_height)
	height = im_info->status_height;
    else
	height = im_info->preedit_height;
    old_height = ve->vendor.im_height;
    if (height)
	height += SEPARATOR_HEIGHT;

    ve->vendor.im_height = height;

    XtSetArg(args[0], XtNbaseHeight, &base_height);
    XtGetValues(vw, args, 1);
    if (base_height < 0)
	base_height = 0;
    XtSetArg(args[0], XtNbaseHeight, base_height);
    XtSetValues(vw, args, 1);

    return height;
}

Widget
awt_util_getXICStatusAreaWindow(Widget w)
{
    while (!XtIsShell(w)){
	w = XtParent(w);
    }
    return w;
}

static XRectangle geometryRect;
XVaNestedList awt_util_getXICStatusAreaList(Widget w)
{
    Widget p;
    XmVendorShellExtObject ve;
    XmWidgetExtData extData;
    XmImInfo *im_info;
    XmICStruct *icp;

    XVaNestedList list = NULL;
    XRectangle  *ssgeometry = &geometryRect;
    Pixel  bg;
    Pixel  fg;
    Pixmap bpm;
    Dimension height,width;
    Position  x,y;

    p = w;
    while (!XtIsShell(p)){
	p = XtParent(p);
    }

    XtVaGetValues(p,
		  XmNx, &x,
		  XmNy, &y,
		  XmNwidth, &width,
		  XmNheight, &height,
		  NULL);

    extData = _XmGetWidgetExtData((Widget) p, XmSHELL_EXTENSION);
    if (extData == NULL) {
	return NULL;
    }

    ve = (XmVendorShellExtObject) extData->widget;
    if ((im_info = (XmImInfo *) ve->vendor.im_info) == NULL) {
	return NULL;         
    } else
	icp = im_info->iclist;


    if (icp) {
	/*
	 * We hava at least a textfield/textarea in the frame, use the
	 * first one.
	 */
	ssgeometry->x = 0;
	ssgeometry->y = height - icp->status_height;
	ssgeometry->width = icp->status_width;
	ssgeometry->height = icp->status_height;

	/*
	 * use motif TextComponent's resource
	 */
	fg = icp->foreground;
	bg = icp->background;
	bpm = icp->background_pixmap;

	list = XVaCreateNestedList(0, 
			XNFontSet, extract_fontset((XmFontList)icp->font_list),
			XNArea, ssgeometry, 
			XNBackground, bg, 
			XNForeground, fg, 
			XNBackgroundPixmap, bpm, 
			NULL);
   }
    return list;
}

static XFontSet
extract_fontset(XmFontList fl)
{
    XmFontContext context;
    XmFontListEntry next_entry;
    XmFontType type_return;
    XtPointer tmp_font;
    XFontSet first_fs = NULL;
    char *font_tag;

    if (!XmFontListInitFontContext(&context, fl))
	return NULL;

    do {
	next_entry = XmFontListNextEntry(context);
	if (next_entry) {
	    tmp_font = XmFontListEntryGetFont(next_entry, &type_return);
	    if (type_return == XmFONT_IS_FONTSET) {
		font_tag = XmFontListEntryGetTag(next_entry);
		if (!strcmp(font_tag, XmFONTLIST_DEFAULT_TAG)) {
		    XmFontListFreeFontContext(context);
		    XtFree(font_tag);
		    return (XFontSet) tmp_font;
		}
		XtFree(font_tag);
		if (first_fs == NULL)
		    first_fs = (XFontSet) tmp_font;
	    }
	}
    } while (next_entry);

    XmFontListFreeFontContext(context);
    return first_fs;
}


/* For a specified shell widget, make it resizable by setting its
   minimum and maximum width and height to the screen size range.
   Also, set its window manager decorations and menu accordingly.	*/

void
awt_util_setShellResizable( Widget shellW, Boolean isMapped )
{
    int		screenWidth;
    int		screenHeight;
    Atom	OLDecorAddAtom;
    Atom	OLDecorResizeAtom;

    /* Iff shell is mapped, we must unmap and later remap in order to
       effect the changes we make in the window manager decorations.
       N.B.This unmapping / remapping of the shell exposes a bug in
	   X/Motif or the Motif Window Manager.  When you attempt
           to map a widget which is positioned (partially) offscreen,
           the window is relocated to be entirely on screen. Good
           idea.  But if both the x and the y coordinates are less
           than the origin (0,0), the first (re)map will move the
           window to the origin, and any subsequent (re)map will
           relocate the window at some other point on the screen.
           I have written a short Motif test program to discover
           this bug.  Nevertheless, I know of no way to effect the
           changes in the wm decorations other than an unmap and
           remap.  Also, this should occur infrequently and it does
           not cause any real problem.  So for now we'll let it be.     */

    if  (isMapped)
        XtUnmapWidget(shellW);

    /* Make it resizable.  Also, set the resources for MWM decorations.	*/
    screenWidth = XWidthOfScreen(XDefaultScreenOfDisplay(awt_display));
    screenHeight = XHeightOfScreen(XDefaultScreenOfDisplay(awt_display));
   /* subtract decorations when calculating max window size */
    XtVaSetValues(shellW,
		  XmNminWidth, 0,
		  XmNmaxWidth, screenWidth,
		  XmNminHeight, 0,
		  XmNmaxHeight, screenHeight,
		  XmNmwmDecorations, MWM_DECOR_ALL,
		  XmNmwmFunctions, MWM_FUNC_ALL,
		  NULL);

    /* To set OLWM decorations, we must use the appropriate properties.
     * N.B. It is unfortunate that the OLWM does not allow us to adjust
     * these properties dynamically; the following can not be used here.
     * OLDecorAddAtom = XInternAtom( XtDisplay(shellW), "_OL_DECOR_ADD", False);
     * OLDecorResizeAtom = XInternAtom( XtDisplay(shellW), "_OL_DECOR_RESIZE", False);
     * XChangeProperty( XtDisplay(shellW), XtWindow(shellW), OLDecorAddAtom,
     *			XA_ATOM, 32, PropModeReplace,
     *			(unsigned char *)&OLDecorResizeAtom, 1 );
     */

    /* How would we set the appropriate decorations for other WM's ?!	*/

    if  (isMapped)
        XtMapWidget(shellW);
}


/* For a specified shell widget, make it nonresizable by fixing its
   minimum and maximum widths and heights to the values specified.
   Also adjust the window manager decorations and menu accordingly.	*/

void
awt_util_setShellNotResizable( Widget shellW, long width, long height, Boolean isMapped )
{
    Atom	OLDecorDelAtom;
    Atom	OLDecorResizeAtom;

    /* Iff shell is mapped, we must unmap and later remap in order to
       effect the changes we make in the window manager decorations.
       N.B. See the corresponding comment above in setShellResizable().	*/

    if  (isMapped)
        XtUnmapWidget(shellW);

    /* Make it not resizable.  Also, set resources for MWM decorations.	*/
    XtVaSetValues(shellW,
		  XmNminWidth, (XtArgVal)width,
		  XmNmaxWidth, (XtArgVal)width,
		  XmNminHeight, (XtArgVal)height,
		  XmNmaxHeight, (XtArgVal)height,
		  XmNmwmDecorations, 
			MWM_DECOR_BORDER | MWM_DECOR_TITLE | MWM_DECOR_MENU,
		  XmNmwmFunctions, MWM_FUNC_MOVE | MWM_FUNC_CLOSE,
		  NULL);

    /* To set OLWM decorations, we must use the appropriate properties.
     * N.B. It is unfortunate that the OLWM does not allow us to adjust
     * these properties dynamically; the following can not be used here.
     * OLDecorDelAtom = XInternAtom( XtDisplay(shellW), "_OL_DECOR_DEL", False);
     * OLDecorResizeAtom = XInternAtom( XtDisplay(shellW), "_OL_DECOR_RESIZE", False);
     * XChangeProperty( XtDisplay(shellW), XtWindow(shellW), OLDecorDelAtom,
     *			XA_ATOM, 32, PropModeReplace,
     *			(unsigned char *)&OLDecorResizeAtom, 1 );
     */

    /* How would we set the appropriate decorations for other WM's ?!	*/

    if  (isMapped)
        XtMapWidget(shellW);
}


/* 
** awt_winmgrerr: temporary XError handler setup to determine if 
** 		  any windowow manager is running 
*/
static int
awt_winmgrerr( Display *display, XErrorEvent *error )
{
    if ( error->request_code == X_ChangeWindowAttributes )
    {
	winmgr_running = TRUE;
    }
    return 0;
}

/* Determine current running Window Manager (Motif, OpenLook, or other)	*/
unsigned awt_util_runningWindowManager()
{
    Atom      WMAtom;
    static int wmgr = WM_YET_TO_BE_DETERMINED;
    unsigned long       nItems;
    unsigned long       bytesAfter;
    Status      status;
    Atom        actualType;
    int         actualFormat;
    unsigned char       *ptr;
    XSetWindowAttributes attrib;
    XtErrorHandler oldHandler;


    if ( wmgr != WM_YET_TO_BE_DETERMINED )
    {
	return wmgr;
    }

    /* Is any window manager is running? 
    ** It would appear we must make this test first as the winmgr
    ** (at least dtwm) when killed still appears to be running 
    ** presumably as its distinguishing atoms are still lying around
    */
    /* if any window manager is running? */
    XSetErrorHandler(awt_winmgrerr);

    winmgr_running = 0;
    attrib.event_mask = SubstructureRedirectMask;
    XChangeWindowAttributes ( awt_display,
                              DefaultRootWindow(awt_display),
                              CWEventMask,
                              &attrib
                            );
    XSync( awt_display, FALSE );
    (void) XSetErrorHandler(xerror_handler);

    if ( !winmgr_running )
    {
            wmgr = NO_WM;
            attrib.event_mask = 0;
            XChangeWindowAttributes ( awt_display,
                                      DefaultRootWindow(awt_display),
                                      CWEventMask,
                                      &attrib
                                    );
    } else if ( WMAtom = XInternAtom(awt_display, "_DT_SM_WINDOW_INFO" , TRUE))
    {
   	wmgr = CDE_WM;
    } else  if ( WMAtom = XInternAtom(awt_display, "_MOTIF_WM_INFO" , TRUE))
    {
        status = XGetWindowProperty (awt_display, 
				     DefaultRootWindow(awt_display),
                                     WMAtom, 0, 1, False,
                                     WMAtom, &actualType,
                                     &actualFormat, &nItems, &bytesAfter,
                                     &ptr);
 
        if (status == Success && actualFormat != 0) 
	{
            XFree((char *) ptr);
	    wmgr = MOTIF_WM;
        } else if ( WMAtom = XInternAtom (awt_display, 
        				  "_SUN_WM_PROTOCOLS" , TRUE) )
        {
            wmgr = OPENLOOK_WM;
	} else 
	{
    	    wmgr = OTHER_WM;
	}
    } else if ( WMAtom = XInternAtom (awt_display, "_SUN_WM_PROTOCOLS" , TRUE) )
    {
	wmgr = OPENLOOK_WM;
    } else
    {
    	wmgr = OTHER_WM;
    }
    return(wmgr);
}


/*the caller does have the responsibility to free the memory return
  from this function...*/ 
char* awt_util_makeWMMenuItem(char *target, Atom protocol){
    char        *buf = NULL;
    int         buflen = 0;

    /*a label in a menuitem is not supposed to be a FullOfSpaceString... */
    buflen = strlen(target) * 3;
    buf = (char*)malloc(buflen + 20);
    if (buf == NULL){
	JNU_ThrowOutOfMemoryError((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2), NULL);
    }
    else{
      int   off = 0;
      char  *ptr = target;
      while ((off < (buflen - 20)) && (*ptr != '\0')){
	if (*ptr == ' '){
	  *(buf + off++) = 0x5c;
	}
	*(buf + off++) = *ptr++;
      }
      sprintf(buf + off, " f.send_msg %d", protocol);
    }
    return buf;
}

/* 
 * This callback proc is installed via setting the XmNinsertPosition
 * resource on a widget. It ensures that components added
 * to a widget are inserted in the correct z-order position
 * to match up with their peer/target ordering in Container.java
 */
Cardinal 
awt_util_insertCallback(Widget w) 
{
    jobject peer;
    WidgetList children;
    int num_children;
    Widget parent;
    XtPointer userdata;
    Cardinal index;
    long pos;
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    parent = XtParent(w);
    XtVaGetValues(parent,
                  XmNnumChildren, &num_children,
                  XmNchildren, &children,
                  NULL);
    XtVaGetValues(w, XmNuserData, &userdata, NULL);

    index = num_children;         /* default is to add to end */

    if (userdata != NULL) {
        peer = (jobject) userdata;

	// SECURITY: We are running on the privileged toolkit thread.
	//           The peer must *NOT* call into user code
        pos = (long) JNU_CallMethodByName(env
					  ,NULL
					  ,(jobject) peer
					  ,"getZOrderPosition_NoClientCode"
					  ,"()I").i;
        if ((*env)->ExceptionOccurred(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        }
        index = (Cardinal) (pos != -1 ? pos : num_children);
    }
    return index;
}

void
awt_util_consumeAllXEvents(Widget widget)
{
    /* Remove all queued X Events for the window of the widget. */

#define ALL_EVENTS_MASK 0xFFFF

    XEvent xev;

    XFlush(awt_display);
    while (XCheckWindowEvent(awt_display, XtWindow(widget),
               ALL_EVENTS_MASK, &xev)) ;
}

/**
 * Gets the thread we are currently executing on
 */
jobject
awtJNI_GetCurrentThread(JNIEnv *env) {
    static jclass threadClass = NULL;
    static jmethodID currentThreadMethodID = NULL;

    jobject currentThread = NULL;

    /* Initialize our java identifiers once. Checking before locking 
     * is a huge performance win.
     */
    if (threadClass == NULL) {
	// should enter a monitor here...
	Boolean err = FALSE;
	if (threadClass == NULL) {
	    jclass tc = (*env)->FindClass(env, "java/lang/Thread");
	    threadClass = (*env)->NewGlobalRef(env, tc);
	    if (threadClass != NULL) {
	        currentThreadMethodID = (*env)->GetStaticMethodID(env,
					      threadClass,
					      "currentThread",
					      "()Ljava/lang/Thread;"
		    				);
	    }
	}
	if (currentThreadMethodID == NULL) {
	    threadClass = NULL;
	    err = TRUE;
	}
	if (err) {
	    return NULL;
	}
    } /* threadClass == NULL*/

    currentThread = (*env)->CallStaticObjectMethod(
			env, threadClass, currentThreadMethodID);
    assert(!((*env)->ExceptionOccurred(env)));
    /*JNU_PrintString(env, "getCurrentThread() -> ", JNU_ToString(env,currentThread));*/
    return currentThread;
} /* awtJNI_GetCurrentThread() */

void
awtJNI_ThreadYield(JNIEnv *env) {

    static jclass threadClass = NULL;
    static jmethodID yieldMethodID = NULL;

    /* Initialize our java identifiers once. Checking before locking 
     * is a huge performance win.
     */
    if (threadClass == NULL) {
	// should enter a monitor here...
	Boolean err = FALSE;
	if (threadClass == NULL) {
	    threadClass = (*env)->FindClass(env, "java/lang/Thread");
	    if (threadClass != NULL) {
	        yieldMethodID = (*env)->GetStaticMethodID(env,
					      threadClass,
					      "yield",
					      "()V"
		    				);
	    }
	}
	if (yieldMethodID == NULL) {
	    threadClass = NULL;
	    err = TRUE;
	}
	if (err) {
	    return;
	}
    } /* threadClass == NULL*/

    (*env)->CallStaticVoidMethod(env, threadClass, yieldMethodID);
    assert(!((*env)->ExceptionOccurred(env)));
} /* awtJNI_ThreadYield() */
