/*
 * @(#)color.c	1.63 98/10/27
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*-
 *	Image dithering and rendering code for X11.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <alloca.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include "awt_p.h"
#include "java_awt_Color.h"
#include "java_awt_SystemColor.h"
#include "java_awt_color_ColorSpace.h"
#include "java_awt_Transparency.h"
#include "java_awt_image_DataBuffer.h"
#include "img_colors.h"
#include "imageInitIDs.h"

#include <jni.h>
#include <jni_util.h>

#ifdef DEBUG
static int debug_colormap = 0;
#endif

/* returns the absolute value x */
#define ABS(x) ((x) < 0 ? -(x) : (x))

#define CLIP(val,min,max)	((val < min) ? min : ((val > max) ? max : val))

#define RGBTOGRAY(r, g, b) ((int) (.299 * r + .587 * g + .114 * b + 0.5))

static void make_dither_arrays();
static sgn_ordered_dither_array std_img_oda_red;
static sgn_ordered_dither_array std_img_oda_green;
static sgn_ordered_dither_array std_img_oda_blue;
static int std_odas_computed = 0;
void initInverseGrayLut(int* prgb, int rgbsize, ColorData* cData);

enum {
    FREE_COLOR		= 0,
    LIKELY_COLOR	= 1,
    UNAVAILABLE_COLOR	= 2,
    ALLOCATED_COLOR	= 3
};

/*
 * Constants to control the filling of the colormap.
 * By default, try to allocate colors in the default colormap until
 * CMAP_ALLOC_DEFAULT colors are being used (by Java and/or other
 * applications).
 * For cases where the default colormap may already have a large
 * number of colors in it, make sure that we ourselves try to add
 * at least CMAP_ALLOC_MIN new colors, even if we need to allocate
 * more than the DEFAULT to do that.
 * Under no circumstances will the colormap be filled to more than
 * CMAP_ALLOC_MAX colors.
 */
#define CMAP_ALLOC_MIN		100	/* minimum number of colors to "add" */
#define CMAP_ALLOC_DEFAULT	200	/* default number of colors in cmap */
#define CMAP_ALLOC_MAX		245	/* maximum number of colors in cmap */

#ifdef SOLARIS2
#include <sys/utsname.h>

struct {
    char *machine;
    int  cubesize;
} machinemap[] = {
    { "sun4c", LOOKUPSIZE / 4 },
    { "sun4m", LOOKUPSIZE / 2 },
    { "sun4d", LOOKUPSIZE / 2 },
    { "sun4u", LOOKUPSIZE / 1 },
};

#define MACHMAPSIZE	(sizeof(machinemap) / sizeof(machinemap[0]))

int getVirtCubeSize() {
    struct utsname name;
    int i, ret;

    ret = uname(&name);
    if (ret < 0) {
#ifdef DEBUG
#include <errno.h>
	jio_fprintf(stderr, "uname errno = %d, using default cubesize %d\n",
		    errno, LOOKUPSIZE);
#endif
	return LOOKUPSIZE;
    }

    for (i = 0; i < MACHMAPSIZE; i++) {
	if (strcmp(name.machine, machinemap[i].machine) == 0) {
#ifdef DEBUG
	    if (debug_colormap) {
		jio_fprintf(stderr, "'%s'.cubesize = '%d'\n",
			    machinemap[i].machine, machinemap[i].cubesize);
	    }
#endif
	    return machinemap[i].cubesize;
	}
    }

#ifdef DEBUG
    if (debug_colormap) {
	jio_fprintf(stderr, "unknown machine '%s' using cubesize %d\n",
		    name.machine, LOOKUPSIZE);
    }
#endif
    return LOOKUPSIZE;
}
#else /* SOLARIS */
#define getVirtCubeSize()	(LOOKUPSIZE)
#endif /* SOLARIS */

unsigned char img_bwgamma[256];
uns_ordered_dither_array img_oda_alpha;

#ifdef NEED_IMAGE_CONVERT
ImgConvertFcn DirectImageConvert;
ImgConvertFcn Dir16IcmOpqUnsImageConvert;
ImgConvertFcn Dir16IcmTrnUnsImageConvert;
ImgConvertFcn Dir16IcmOpqSclImageConvert;
ImgConvertFcn Dir16DcmOpqUnsImageConvert;
ImgConvertFcn Dir16DcmTrnUnsImageConvert;
ImgConvertFcn Dir16DcmOpqSclImageConvert;
ImgConvertFcn Dir32IcmOpqUnsImageConvert;
ImgConvertFcn Dir32IcmTrnUnsImageConvert;
ImgConvertFcn Dir32IcmOpqSclImageConvert;
ImgConvertFcn Dir32DcmOpqUnsImageConvert;
ImgConvertFcn Dir32DcmTrnUnsImageConvert;
ImgConvertFcn Dir32DcmOpqSclImageConvert;

ImgConvertFcn PseudoImageConvert;
ImgConvertFcn PseudoFSImageConvert;
ImgConvertFcn FSColorIcmOpqUnsImageConvert;
ImgConvertFcn FSColorDcmOpqUnsImageConvert;
ImgConvertFcn OrdColorIcmOpqUnsImageConvert;
ImgConvertFcn OrdColorDcmOpqUnsImageConvert;

#endif /* NEED_IMAGE_CONVERT */

/*
 * Find the best color.
 */
int
awt_color_matchTC(int r, int g, int b, AwtData *awt_data)
{
    r = CLIP(r, 0, 255);
    g = CLIP(g, 0, 255);
    b = CLIP(b, 0, 255);
    return (((r >> awt_data->awtImage->clrdata.rScale) 
		<< awt_data->awtImage->clrdata.rOff) |
	    ((g >> awt_data->awtImage->clrdata.gScale) 
		<< awt_data->awtImage->clrdata.gOff) |
	    ((b >> awt_data->awtImage->clrdata.bScale) 
		<< awt_data->awtImage->clrdata.bOff));
}

int
awt_color_matchGS(int r, int g, int b, AwtData *awt_data)
{
    r = CLIP(r, 0, 255);
    g = CLIP(g, 0, 255);
    b = CLIP(b, 0, 255);
    return awt_data->color_data->img_grays[RGBTOGRAY(r, g, b)];
}

int
awt_color_match(int r, int g, int b, AwtData *awt_data)
{
    int besti = 0;
    int mindist, i, t, d;
    ColorEntry *p = awt_data->color_data->awt_Colors;

    r = CLIP(r, 0, 255);
    g = CLIP(g, 0, 255);
    b = CLIP(b, 0, 255);

    /* look for pure gray match */
    if ((r == g) && (g == b)) {
      mindist = 256;
      for (i = 0 ; i < awt_data->awt_num_colors ; i++, p++)
        if (p->flags == ALLOCATED_COLOR) {
          if (! ((p->r == p->g) && (p->g == p->b)) )
              continue;
          d = ABS(p->r - r);
          if (d == 0)
              return i;
          if (d < mindist) {
              besti = i;
              mindist = d;
          }
        }
      return besti;
    }

    /* look for non-pure gray match */
    mindist = 256 * 256 * 256;
    for (i = 0 ; i < awt_data->awt_num_colors ; i++, p++)
	if (p->flags == ALLOCATED_COLOR) {
	    t = p->r - r;
	    d = t * t;
	    if (d >= mindist)
		continue;
	    t = p->g - g;
	    d += t * t;
	    if (d >= mindist)
		continue;
	    t = p->b - b;
	    d += t * t;
	    if (d >= mindist)
		continue;
	    if (d == 0)
		return i;
	    if (d < mindist) {
		besti = i;
		mindist = d;
	    }
	}
    return besti;
}

/*
 * Allocate a color in the X color map and return the pixel.
 * If the "expected pixel" is non-negative then we will only
 * accept the allocation if we get exactly that pixel value.
 * This prevents us from seeing a bunch of ReadWrite pixels
 * allocated by another imaging application and duplicating
 * that set of inaccessible pixels in our precious remaining
 * ReadOnly colormap cells.
 */
static int
alloc_col(Display *dpy, Colormap cm, int r, int g, int b, int pixel,
	  AwtData *awt_data)
{
    XColor col;

    r = CLIP(r, 0, 255);
    g = CLIP(g, 0, 255);
    b = CLIP(b, 0, 255);

    col.flags = DoRed | DoGreen | DoBlue;
    col.red   = (r << 8) | r;
    col.green = (g << 8) | g;
    col.blue  = (b << 8) | b;
    if (XAllocColor(dpy, cm, &col)) {
#ifdef DEBUG
	if (debug_colormap)
	    jio_fprintf(stdout, "allocated %d (%d,%d, %d)\n", col.pixel, r, g, b);
#endif
	if (pixel >= 0 && col.pixel != pixel) {
	    /*
	     * If we were trying to allocate a shareable "ReadOnly"
	     * color then we would have gotten back the expected
	     * pixel.  If the returned pixel was different, then
	     * the source color that we were attempting to gain
	     * access to must be some other application's ReadWrite
	     * private color.  We free the returned pixel so that
	     * we won't waste precious colormap entries by duplicating
	     * that color in the as yet unallocated entries.  We
	     * return -1 here to indicate the failure to get the
	     * expected pixel.
	     */
	    awt_data->color_data->awt_Colors[pixel].flags = UNAVAILABLE_COLOR;
	    XFreeColors(dpy, cm, &col.pixel, 1, 0);
	    return -1;
	}
	awt_data->color_data->awt_Colors[col.pixel].flags = ALLOCATED_COLOR;
	awt_data->color_data->awt_Colors[col.pixel].r = col.red   >> 8;
	awt_data->color_data->awt_Colors[col.pixel].g = col.green >> 8;
	awt_data->color_data->awt_Colors[col.pixel].b = col.blue  >> 8;
	if (awt_data->color_data->awt_icmLUT != 0) {
            awt_data->color_data->awt_icmLUT2Colors[col.pixel] = col.pixel;
            awt_data->color_data->awt_icmLUT[col.pixel] =
		0xff000000 |
                (awt_data->color_data->awt_Colors[col.pixel].r<<16) |
                (awt_data->color_data->awt_Colors[col.pixel].g<<8) |
                (awt_data->color_data->awt_Colors[col.pixel].b);
	}
	return col.pixel;
    }

    return awt_color_match(r, g, b, awt_data);
}

void 
awt_allocate_systemcolors(XColor *colorsPtr, int num_pixels, AwtData *awtData) {
    int i;
    int r, g, b, pixel;

    for (i=0; i < num_pixels; i++) {
	r = colorsPtr[i].red   >> 8;
	g = colorsPtr[i].green >> 8;
	b = colorsPtr[i].blue  >> 8;
	pixel = alloc_col(awt_display, awtData->awt_cmap, r, g, b, -1, awtData);
    }
}

void
awt_fill_imgcv(ImgConvertFcn **array, int mask, int value, ImgConvertFcn fcn)
{
    int i;

    for (i = 0; i < NUM_IMGCV; i++) {
	if ((i & mask) == value) {
	    array[i] = fcn;
	}
    }
}

/*
 * called from X11Server_create() in xlib.c 
 */
int
awt_allocate_colors(AwtData *awt_data)
{
    Display *dpy;
    unsigned long freecolors[256], plane_masks[1];
    XColor cols[256];
    unsigned char reds[256], greens[256], blues[256], indices[256];
    Colormap cm;
    int i, j, k, cmapsize, nfree, depth, bpp, screen;
    XPixmapFormatValues *pPFV;
    int numpfv;
    XVisualInfo *pVI;
    char *forcemono;
    char *forcegray;

    make_uns_ordered_dither_array(img_oda_alpha, 256);


    forcemono = getenv("FORCEMONO");
    forcegray = getenv("FORCEGRAY");
    if (forcemono && !forcegray)
	forcegray = forcemono;

    /*
     * Get the colormap and make sure we have the right visual
     */
    dpy = awt_display;
    screen = awt_screen;
    cm = awt_data->awt_cmap;
    depth = awt_data->awt_depth;
    pVI = &awt_data->awt_visInfo;
    awt_data->awt_num_colors = awt_data->awt_visInfo.colormap_size;
    awt_data->awtImage = (awtImageData *) calloc (1, sizeof (awtImageData));

    pPFV = XListPixmapFormats(dpy, &numpfv);
    if (pPFV) {
	for (i = 0; i < numpfv; i++) {
	    if (pPFV[i].depth == depth) {
		awt_data->awtImage->wsImageFormat = pPFV[i];
		break;
	    }
	}
	XFree(pPFV);
    }
    bpp = awt_data->awtImage->wsImageFormat.bits_per_pixel;
    if (bpp == 24) {
	bpp = 32;
    }
    awt_data->awtImage->clrdata.bitsperpixel = bpp;
    awt_data->awtImage->Depth = depth;

    if ((bpp == 32 || bpp == 16) && pVI->class == TrueColor) {
	awt_data->AwtColorMatch = awt_color_matchTC;
	awt_data->awtImage->clrdata.rOff = 0;
	for (i = pVI->red_mask; (i & 1) == 0; i >>= 1) {
	    awt_data->awtImage->clrdata.rOff++;
	}
	awt_data->awtImage->clrdata.rScale = 0;
	while (i < 0x80) {
	    awt_data->awtImage->clrdata.rScale++;
	    i <<= 1;
	}
	awt_data->awtImage->clrdata.gOff = 0;
	for (i = pVI->green_mask; (i & 1) == 0; i >>= 1) {
	    awt_data->awtImage->clrdata.gOff++;
	}
	awt_data->awtImage->clrdata.gScale = 0;
	while (i < 0x80) {
	    awt_data->awtImage->clrdata.gScale++;
	    i <<= 1;
	}
	awt_data->awtImage->clrdata.bOff = 0;
	for (i = pVI->blue_mask; (i & 1) == 0; i >>= 1) {
	    awt_data->awtImage->clrdata.bOff++;
	}
	awt_data->awtImage->clrdata.bScale = 0;
	while (i < 0x80) {
	    awt_data->awtImage->clrdata.bScale++;
	    i <<= 1;
	}
#ifdef NEED_IMAGE_CONVERT        
	awt_fill_imgcv(awt_data->awtImage->convert, 0, 0, DirectImageConvert);
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			| IMGCV_ALPHABITS | IMGCV_CMBITS),
		       (IMGCV_UNSCALED | IMGCV_BYTEIN
			| IMGCV_OPAQUE | IMGCV_ICM),
		       (bpp == 32
			? Dir32IcmOpqUnsImageConvert
			: Dir16IcmOpqUnsImageConvert));
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			| IMGCV_ALPHABITS | IMGCV_CMBITS),
		       (IMGCV_UNSCALED | IMGCV_BYTEIN
			| IMGCV_ALPHA | IMGCV_ICM),
		       (bpp == 32
			? Dir32IcmTrnUnsImageConvert
			: Dir16IcmTrnUnsImageConvert));
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			| IMGCV_ALPHABITS | IMGCV_CMBITS),
		       (IMGCV_SCALED | IMGCV_BYTEIN
			| IMGCV_OPAQUE | IMGCV_ICM),
		       (bpp == 32
			? Dir32IcmOpqSclImageConvert
			: Dir16IcmOpqSclImageConvert));
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			| IMGCV_ALPHABITS | IMGCV_CMBITS),
		       (IMGCV_UNSCALED | IMGCV_INTIN
			| IMGCV_OPAQUE | IMGCV_DCM8),
		       (bpp == 32
			? Dir32DcmOpqUnsImageConvert
			: Dir16DcmOpqUnsImageConvert));
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			| IMGCV_ALPHABITS | IMGCV_CMBITS),
		       (IMGCV_UNSCALED | IMGCV_INTIN
			| IMGCV_ALPHA | IMGCV_DCM8),
		       (bpp == 32
			? Dir32DcmTrnUnsImageConvert
			: Dir16DcmTrnUnsImageConvert));
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			| IMGCV_ALPHABITS | IMGCV_CMBITS),
		       (IMGCV_SCALED | IMGCV_INTIN
			| IMGCV_OPAQUE | IMGCV_DCM8),
		       (bpp == 32
			? Dir32DcmOpqSclImageConvert
			: Dir16DcmOpqSclImageConvert));
#endif /* NEED_IMAGE_CONVERT */        
    } else if (bpp <= 8 && (pVI->class == StaticGray
			    || pVI->class == GrayScale
			    || forcegray)) {
	awt_data->AwtColorMatch = awt_color_matchGS;
	awt_data->awtImage->clrdata.grayscale = 1;
	awt_data->awtImage->clrdata.bitsperpixel = 8;
#ifdef NEED_IMAGE_CONVERT        
	awt_fill_imgcv(awt_data->awtImage->convert, 0, 0, PseudoImageConvert);
	if (getenv("NOFSDITHER") == 0) {
	    awt_fill_imgcv(awt_data->awtImage->convert,
			   IMGCV_ORDERBITS, IMGCV_TDLRORDER,
			   PseudoFSImageConvert);
	}
#endif /* NEED_IMAGE_CONVERT */        
    } else if (bpp <= 8 && (pVI->class == PseudoColor
			    || pVI->class == TrueColor
			    || pVI->class == StaticColor)) {
	if (pVI->class == TrueColor)
	   awt_data->awt_num_colors = (1 << pVI->depth);
	awt_data->AwtColorMatch = awt_color_match;
	awt_data->awtImage->clrdata.bitsperpixel = 8;
#ifdef NEED_IMAGE_CONVERT        
	awt_fill_imgcv(awt_data->awtImage->convert, 0, 0, PseudoImageConvert);
	if (getenv("NOFSDITHER") == 0) {
	    awt_fill_imgcv(awt_data->awtImage->convert, IMGCV_ORDERBITS, 
			   IMGCV_TDLRORDER, PseudoFSImageConvert);
	    awt_fill_imgcv(awt_data->awtImage->convert,
			   (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			    | IMGCV_ALPHABITS | IMGCV_ORDERBITS
			    | IMGCV_CMBITS),
			   (IMGCV_UNSCALED | IMGCV_BYTEIN
			    | IMGCV_OPAQUE | IMGCV_TDLRORDER
			    | IMGCV_ICM),
			   FSColorIcmOpqUnsImageConvert);
	    awt_fill_imgcv(awt_data->awtImage->convert,
			   (IMGCV_SCALEBITS | IMGCV_INSIZEBITS
			    | IMGCV_ALPHABITS | IMGCV_ORDERBITS
			    | IMGCV_CMBITS),
			   (IMGCV_UNSCALED | IMGCV_INTIN
			    | IMGCV_OPAQUE | IMGCV_TDLRORDER
			    | IMGCV_DCM8),
			   FSColorDcmOpqUnsImageConvert);
	}
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS | IMGCV_ALPHABITS
			| IMGCV_ORDERBITS | IMGCV_CMBITS),
		       (IMGCV_UNSCALED | IMGCV_BYTEIN | IMGCV_OPAQUE
			| IMGCV_RANDORDER | IMGCV_ICM),
		       OrdColorIcmOpqUnsImageConvert);
	awt_fill_imgcv(awt_data->awtImage->convert,
		       (IMGCV_SCALEBITS | IMGCV_INSIZEBITS | IMGCV_ALPHABITS
			| IMGCV_ORDERBITS | IMGCV_CMBITS),
		       (IMGCV_UNSCALED | IMGCV_INTIN | IMGCV_OPAQUE
			| IMGCV_RANDORDER | IMGCV_DCM8),
		       OrdColorDcmOpqUnsImageConvert);
#endif /* NEED_IMAGE_CONVERT */        
    } else {
 	free (awt_data->awtImage);
	return 0;
    }

    if (depth > 8) {
	return 1;
    }

    if (awt_data->awt_num_colors > 256) {
	free (awt_data->awtImage);
	return 0;
    }

    /* Allocate ColorData structure */
    awt_data->color_data = ZALLOC (_ColorData);
    awt_data->color_data->screendata = 1; /* This ColorData struct corresponds
                                             to some AWT screen/visual, so when
                                             any IndexColorModel using this
                                             struct is finalized, don't free
                                             the struct in freeICMColorData.
                                           */

    /*
     * Initialize colors array
     */
    for (i = 0; i < awt_data->awt_num_colors; i++) {
	cols[i].pixel = i;
    }

    awt_data->color_data->awt_Colors = (ColorEntry *) calloc 
					(256, sizeof (ColorEntry));

    XQueryColors(dpy, cm, cols, awt_data->awt_num_colors);
    for (i = 0; i < awt_data->awt_num_colors; i++) {
	awt_data->color_data->awt_Colors[i].r = cols[i].red >> 8;
	awt_data->color_data->awt_Colors[i].g = cols[i].green >> 8;
	awt_data->color_data->awt_Colors[i].b = cols[i].blue >> 8;
	awt_data->color_data->awt_Colors[i].flags = LIKELY_COLOR;
    }

    /*
     * Determine which colors in the colormap can be allocated and mark
     * them in the colors array
     */
    nfree = 0;
    for (i = 128; i > 0; i >>= 1) {
	if (XAllocColorCells(dpy, cm, False, plane_masks, 0,
			     freecolors + nfree, i)) {
	    nfree += i;
	}
    }

    for (i = 0; i < nfree; i++) {
	awt_data->color_data->awt_Colors[freecolors[i]].flags = FREE_COLOR;
    }

#ifdef DEBUG
    if (debug_colormap) {
	jio_fprintf(stdout, "%d free.\n", nfree);
    }
#endif

    XFreeColors(dpy, cm, freecolors, nfree, 0);

    /*
     * Allocate the colors that are already allocated by other
     * applications
     */
    for (i = 0; i < awt_data->awt_num_colors; i++) {
	if (awt_data->color_data->awt_Colors[i].flags == LIKELY_COLOR) {
	    awt_data->color_data->awt_Colors[i].flags = FREE_COLOR;
	    alloc_col(dpy, cm,
		      awt_data->color_data->awt_Colors[i].r,
		      awt_data->color_data->awt_Colors[i].g,
		      awt_data->color_data->awt_Colors[i].b, i, awt_data);
	}
    }
#ifdef DEBUG
    if (debug_colormap) {
	jio_fprintf(stdout, "got the already allocated ones\n");
    }
#endif

    /*
     * Allocate more colors, filling the color space evenly.
     */

    alloc_col(dpy, cm, 255, 255, 255, -1, awt_data);
    alloc_col(dpy, cm, 0, 0, 0, -1, awt_data);

    if (awt_data->awtImage->clrdata.grayscale) {
	int g;
	ColorEntry *p;

	if (!forcemono) {
	    for (i = 128; i > 0; i >>= 1) {
		for (g = i; g < 256; g += (i * 2)) {
		    alloc_col(dpy, cm, g, g, g, -1, awt_data);
		}
	    }
	}

	awt_data->color_data->img_grays = (unsigned char *) calloc 
						(256, sizeof (unsigned char));
	for (g = 0; g < 256; g++) {
	    int mindist, besti;
	    int d;

	    p = awt_data->color_data->awt_Colors;
	    mindist = 256;
	    besti = 0;
	    for (i = 0 ; i < awt_data->awt_num_colors ; i++, p++) {
		if (forcegray && (p->r != p->g || p->g != p->b))
		    continue;
		if (forcemono && p->g != 0 && p->g != 255)
		    continue;
		if (p->flags == ALLOCATED_COLOR) {
		    d = p->g - g;
		    if (d < 0) d = -d;
		    if (d < mindist) {
			besti = i;
			if (d == 0) {
			    break;
			}
			mindist = d;
		    }
		}
	    }

	    awt_data->color_data->img_grays[g] = besti;
	}


	if (forcemono || (depth == 1)) {
	    char *gammastr = getenv("HJGAMMA");
	    double gamma = atof(gammastr ? gammastr : "1.6");
	    if (gamma < 0.01) gamma = 1.0;
#ifdef DEBUG
	    if (debug_colormap) {
		jio_fprintf(stderr, "gamma = %f\n", gamma);
	    }
#endif
	    for (i = 0; i < 256; i++) {
		img_bwgamma[i] = (int) (pow(i/255.0, gamma) * 255);
#ifdef DEBUG
		if (debug_colormap) {
		    jio_fprintf(stderr, "%3d ", img_bwgamma[i]);
		    if ((i & 7) == 7)
			jio_fprintf(stderr, "\n");
		}
#endif
	    }
	} else {
	    for (i = 0; i < 256; i++) {
		img_bwgamma[i] = i;
	    }
	}

#ifdef DEBUG
	if (debug_colormap) {
	    jio_fprintf(stderr, "GrayScale initialized\n");
	    jio_fprintf(stderr, "color table:\n");
	    for (i = 0; i < awt_data->awt_num_colors; i++) {
		jio_fprintf(stderr, "%3d: %3d %3d %3d\n",
			i, awt_data->color_data->awt_Colors[i].r, 
			awt_data->color_data->awt_Colors[i].g, 
			awt_data->color_data->awt_Colors[i].b);
	    }
	    jio_fprintf(stderr, "gray table:\n");
	    for (i = 0; i < 256; i++) {
		jio_fprintf(stderr, "%3d ", awt_data->color_data->img_grays[i]);
		if ((i & 7) == 7)
		    jio_fprintf(stderr, "\n");
	    }
	}
#endif

    } else {

        alloc_col(dpy, cm, 255, 0, 0, -1, awt_data);
        alloc_col(dpy, cm, 0, 255, 0, -1,awt_data);
        alloc_col(dpy, cm, 0, 0, 255, -1,awt_data);
        alloc_col(dpy, cm, 255, 255, 0, -1,awt_data);
        alloc_col(dpy, cm, 255, 0, 255, -1,awt_data);
        alloc_col(dpy, cm, 0, 255, 255, -1,awt_data);
        alloc_col(dpy, cm, 192, 192, 192, -1,awt_data);
        alloc_col(dpy, cm, 255, 128, 128, -1,awt_data);
        alloc_col(dpy, cm, 128, 255, 128, -1,awt_data);
        alloc_col(dpy, cm, 128, 128, 255, -1,awt_data);
        alloc_col(dpy, cm, 255, 255, 128, -1,awt_data);
        alloc_col(dpy, cm, 255, 128, 255, -1,awt_data);
        alloc_col(dpy, cm, 128, 255, 255, -1,awt_data);
    }

    j = 0;
    k = 0;
    for (i = 0; i < awt_data->awt_num_colors; i++) {
	if (awt_data->color_data->awt_Colors[i].flags == ALLOCATED_COLOR) {
	    reds[j] = awt_data->color_data->awt_Colors[i].r;
	    greens[j] = awt_data->color_data->awt_Colors[i].g;
	    blues[j] = awt_data->color_data->awt_Colors[i].b;
	    indices[j] = i;
	    j++;
	} else if (awt_data->color_data->awt_Colors[i].flags == 
		   					UNAVAILABLE_COLOR) {
	    k++;
	}
    }
    cmapsize = 0;
    if (getenv("CMAPSIZE") != 0) {
	cmapsize = atoi(getenv("CMAPSIZE"));
    }
    if (cmapsize <= 0) {
	cmapsize = CMAP_ALLOC_DEFAULT;
    }
    if (cmapsize < j + k + CMAP_ALLOC_MIN) {
	cmapsize = j + k + CMAP_ALLOC_MIN;
    }
    if (cmapsize > CMAP_ALLOC_MAX) {
	cmapsize = CMAP_ALLOC_MAX;
    }
    if (cmapsize < j) {
	cmapsize = j;
    }
    cmapsize -= k;

    k = 0;
    if (getenv("VIRTCUBESIZE") != 0) {
	k = atoi(getenv("VIRTCUBESIZE"));
    }
    if (k == 0 || (k & (k - 1)) != 0 || k > 32) {
	k = getVirtCubeSize();
    }
    awt_data->color_data->img_clr_tbl = (unsigned char *) calloc 
		(LOOKUPSIZE * LOOKUPSIZE * LOOKUPSIZE, sizeof (unsigned char));
    img_makePalette(cmapsize, k, LOOKUPSIZE, 50, 250,
		    j, TRUE, reds, greens, blues, 
		    awt_data->color_data->img_clr_tbl); 
		    /*img_clr_tbl);*/

    for (i = 0; i < cmapsize; i++) {
	indices[i] = alloc_col(dpy, cm, reds[i], greens[i], blues[i], -1,
			       awt_data);
    }
    for (i = 0; i < LOOKUPSIZE * LOOKUPSIZE * LOOKUPSIZE  ; i++) { 
	awt_data->color_data->img_clr_tbl[i] = 
				indices[awt_data->color_data->img_clr_tbl[i]];
    }

    awt_data->color_data->img_oda_red   = &(std_img_oda_red[0][0]);
    awt_data->color_data->img_oda_green = &(std_img_oda_green[0][0]);
    awt_data->color_data->img_oda_blue  = &(std_img_oda_blue[0][0]);
    make_dither_arrays(cmapsize, awt_data->color_data);
    std_odas_computed = 1;

#ifdef DEBUG
    if (debug_colormap) {
	int alloc_count = 0;
	int reuse_count = 0;
	int free_count = 0;
	for (i = 0; i < awt_data->awt_num_colors; i++) {
	    switch (awt_data->color_data->awt_Colors[i].flags) {
	      case ALLOCATED_COLOR:
		alloc_count++;
		break;
	      case LIKELY_COLOR:
		reuse_count++;
		break;
	      case FREE_COLOR:
		free_count++;
		break;
	    }
	}
	jio_fprintf(stdout, "%d total, %d allocated, %d reused, %d still free.\n",
	       awt_data->awt_num_colors, alloc_count, reuse_count, free_count);
    }
#endif

    /* Fill in the ICM lut and lut2cmap mapping */
    awt_data->color_data->awt_numICMcolors = 0;
    awt_data->color_data->awt_icmLUT2Colors = (unsigned char *) calloc 
					(256, sizeof (unsigned char));
    awt_data->color_data->awt_icmLUT = (int *) calloc (256, sizeof (int));
    for (i=0; i < 256; i++) {
        if (awt_data->color_data->awt_Colors[i].flags == ALLOCATED_COLOR) {
            /* Keep the mapping between this lut and the actual cmap */
            awt_data->color_data->awt_icmLUT2Colors
				[awt_data->color_data->awt_numICMcolors] = i;
            
            /* Screen IndexColorModel LUTS are always xRGB */
            awt_data->color_data->awt_icmLUT
		    [awt_data->color_data->awt_numICMcolors++] = 0xff000000 |
                (awt_data->color_data->awt_Colors[i].r<<16) |
                (awt_data->color_data->awt_Colors[i].g<<8) |
                (awt_data->color_data->awt_Colors[i].b);
	} else {
            /* Keep the mapping between this lut and the actual cmap */
            awt_data->color_data->awt_icmLUT2Colors
				[awt_data->color_data->awt_numICMcolors] = i;
            
            /* Screen IndexColorModel LUTS are always xRGB */
            awt_data->color_data->awt_icmLUT
			[awt_data->color_data->awt_numICMcolors++] = 0xff000000;
        }
    }
    return 1;
}

#define red(v)		(((v) >> 16) & 0xFF)
#define green(v)	(((v) >>  8) & 0xFF)
#define blue(v)		(((v) >>  0) & 0xFF)

jobject awtJNI_GetColorModel(JNIEnv *env, AwtData *aData)
{
    jobject awt_colormodel = NULL;
    jclass clazz;
    jmethodID mid;
    
    if ((*env)->PushLocalFrame(env, 16) < 0)
        return NULL;

    if ((aData->awt_visInfo.class == TrueColor) &&
	(aData->awt_depth >= 15)) 
    {
        clazz = (*env)->FindClass(env,"java/awt/image/DirectColorModel");

        mid = (*env)->GetMethodID(env,clazz,"<init>","(IIIII)V");
        
        if (mid == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
	}

        awt_colormodel = (*env)->NewObject(env,clazz, mid,
                                           aData->awt_visInfo.depth,
                                           aData->awt_visInfo.red_mask,
                                           aData->awt_visInfo.green_mask,
                                           aData->awt_visInfo.blue_mask,
                                           0);

        if(awt_colormodel == 0)
        {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        }

    }
    else if (aData->awt_visInfo.class == StaticGray &&
             aData->awt_num_colors == 256) {
        jclass clazz1;
        jobject cspace = NULL;
        jint bits[1];
        jintArray bitsArray;
        jboolean falseboolean = JNI_FALSE;

        clazz1 = (*env)->FindClass(env,"java/awt/color/ColorSpace");
        mid = (*env)->GetStaticMethodID(env, clazz1, "getInstance",
              "(I)Ljava/awt/color/ColorSpace;");
        if (mid == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
	}
	// SECURITY: This is safe, because static methods cannot
	//           be overridden, and this method does not invoke
	//           client code
        cspace = (*env)->CallStaticObjectMethod(env, clazz1, mid,
            java_awt_color_ColorSpace_CS_GRAY);
        if (cspace == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        }

        bits[0] = 8;
        bitsArray = (*env)->NewIntArray(env, 1);
        if (bitsArray == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        } else {
            (*env)->SetIntArrayRegion(env, bitsArray, 0, 1, bits);
        }

        clazz = (*env)->FindClass(env,"java/awt/image/ComponentColorModel");

        mid = (*env)->GetMethodID(env,clazz,"<init>",
            "(Ljava/awt/color/ColorSpace;[IZZII)V");
        
        if (mid == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
	}

        awt_colormodel = (*env)->NewObject(env,clazz, mid,
                                           cspace,
                                           bitsArray,
                                           falseboolean,
                                           falseboolean,
                                           java_awt_Transparency_OPAQUE,
                                           java_awt_image_DataBuffer_TYPE_BYTE);

        if(awt_colormodel == 0)
        {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        }

    }
    else 
    {
        jbyte hR[256];
        jbyte hG[256];
        jbyte hB[256];
        jbyte hA[256];
        jbyteArray hRArray;
        jbyteArray hGArray;
        jbyteArray hBArray;
        jbyteArray hAArray;
        char *aBase, *a;
        ColorEntry *c;
        jlong pData;
        
        /* These loops copy backwards so that the aBase pointer won't
         * be dropped by the optimizer and allow the garbage collector
         * to move the array being initialized.
         */
        aBase = (char *)hR;
        for (a = aBase + 255, 
	     c = aData->color_data->awt_Colors + 255; 
	     a >= aBase; a--, c--) {
	    if (c->flags == ALLOCATED_COLOR) {
		*a = c->r;
	    } else {
		*a = 0;
	    }
        }
        aBase = (char *)hG;
        for (a = aBase + 255, 
	     c = aData->color_data->awt_Colors + 255; 
	     a >= aBase; a--, c--) {
	    if (c->flags == ALLOCATED_COLOR) {
		*a = c->g;
	    } else {
		*a = 0;
	    }
        }
        aBase = (char *)hB;
        for (a = aBase + 255, 
	     c = aData->color_data->awt_Colors + 255; 
	     a >= aBase; a--, c--) {
	    if (c->flags == ALLOCATED_COLOR) {
		*a = c->b;
	    } else {
		*a = 0;
	    }
        }

        if (aData->awtImage->clrdata.grayscale) {
            int i;
            ColorEntry *p;

            /* For purposes of creating an IndexColorModel, use
               transparent black for non-allocated or non-gray colors.
             */
            p = aData->color_data->awt_Colors;
            for (i = 0; i < aData->awt_num_colors; i++, p++) {
                if ((p->flags != ALLOCATED_COLOR) ||
                    ((p->r != p->g || p->g != p->b))) {
                    hR[i] = hG[i] = hB[i] = hA[i] = 0;
                } else {
                    hA[i] = 0xff;
                }
            }

            if (aData->color_data->pGrayInverseLutData == NULL) {
                /* Compute the inverse gray LUT for this aData->color_data
                   struct, if not already computed.
                 */
                int rgb[256];
                for (i = 0; i < aData->awt_num_colors; i++, p++) {
                    rgb[i] =  (hA[i] << 24) | ((hR[i] & 0xFF) << 16) |
                              ((hG[i] & 0xFF) << 8) | (hB[i] & 0xFF);
                }
                initInverseGrayLut(rgb, aData->awt_num_colors,
                                   aData->color_data);
            }

            hAArray = (*env)->NewByteArray(env,256);
            if (hAArray == 0)
            {
	        (*env)->PopLocalFrame(env, 0);
                return NULL;
            }
            else
            {
                (*env)->SetByteArrayRegion(env,hAArray,0,256,hA);
            }
        
        }
        
        hRArray = (*env)->NewByteArray(env,256);
        if (hRArray == 0)
        {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        }
        else
        {
            (*env)->SetByteArrayRegion(env,hRArray,0,256,hR);
        }
        
        hGArray = (*env)->NewByteArray(env,256);
        if (hGArray == 0)
        {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        }
        else
        {
            (*env)->SetByteArrayRegion(env,hGArray,0,256,hG);
        }
        
        hBArray = (*env)->NewByteArray(env,256);
        if (hBArray == 0)
        {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        }
        else
        {
            (*env)->SetByteArrayRegion(env,hBArray,0,256,hB);
        }

        clazz = (*env)->FindClass(env,"java/awt/image/IndexColorModel");        
        if (aData->awtImage->clrdata.grayscale) {
            mid = (*env)->GetMethodID(env,clazz,"<init>","(II[B[B[B[B)V");
        } else {
            mid = (*env)->GetMethodID(env,clazz,"<init>","(II[B[B[B)V");
        }
        
        if (mid == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
	}

        if (aData->awtImage->clrdata.grayscale) {
            awt_colormodel = (*env)->NewObject(env,clazz, mid,
                                               aData->awt_visInfo.depth,
                                               aData->awt_num_colors,
                                               hRArray,
                                               hGArray,
                                               hBArray,
                                               hAArray);
        } else {
            awt_colormodel = (*env)->NewObject(env,clazz, mid,
                                               aData->awt_visInfo.depth,
                                               256,
                                               hRArray,
                                               hGArray,
                                               hBArray);
        }
        
        if (awt_colormodel == 0)
        {
	    (*env)->PopLocalFrame(env, 0);
            return NULL;
        }

	/* Set pData field of ColorModel to point to ColorData */
	(*env)->SetLongField (env, awt_colormodel, g_CMpDataID,
			      (jlong) aData->color_data);

    }

    return (*env)->PopLocalFrame(env, awt_colormodel);
}

extern jfieldID colorValueID;

int awtJNI_GetColor(JNIEnv *env,jobject this)
{
    return awtJNI_GetColorForVis (env, this, defaultConfig);
}

int awtJNI_GetColorForVis (JNIEnv *env,jobject this, AwtData *awt_data)
{
    int col;
    jclass SYSCLR_class;

    if (!JNU_IsNull(env,this))
    {
	SYSCLR_class = (*env)->FindClass(env, "java/awt/SystemColor");

	if ((*env)->IsInstanceOf(env, this, SYSCLR_class)) {
		// SECURITY: This is safe, because there is no way
		//           for client code to insert an object
		//           that is a subclass of SystemColor
		col = (int) JNU_CallMethodByName(env
                                          ,NULL
                                          ,this
                                          ,"getRGB"
                                          ,"()I").i;
	} else {
		col = (int)(*env)->GetIntField(env,this,colorValueID);
	}

	col = awt_data->AwtColorMatch(red(col), green(col), blue(col), 
				      awt_data);
	return col;
    }
    
    return 0;
}

void
awt_allocate_systemrgbcolors (jint *rgbColors, int num_colors,
                              AwtData *awtData) {
    int i, pixel;
    for (i = 0; i < num_colors; i++)
	pixel = alloc_col (awt_display, awtData->awt_cmap, red (rgbColors [i]),
			   green (rgbColors [i]), blue (rgbColors [i]), -1,
                           awtData);
}

void
awtJNI_CreateColorData (JNIEnv *env, AwtData *adata, int lock) {

    /* Create Colormap */
    if (lock) {
	AWT_LOCK ();
    }

    if (adata->awt_visInfo.visual == DefaultVisual (awt_display, awt_screen))
	adata->awt_cmap = DefaultColormap (awt_display, awt_screen);
    else {
	int share = 0;

	if (adata->awt_visInfo.visual->class % 2) {
	    Atom actual_type;
	    int actual_format;
	    unsigned long nitems, bytes_after;
	    XStandardColormap *scm;
   
	    XGetWindowProperty (awt_display, awt_root, XA_RGB_DEFAULT_MAP, 
				0L, 1L, False, AnyPropertyType, &actual_type,
                                &actual_format, &nitems, &bytes_after,
                                (unsigned char **) &scm);

            XGetWindowProperty (awt_display, awt_root, XA_RGB_DEFAULT_MAP, 0L,
                                bytes_after/4 + 1, False, AnyPropertyType, 
			        &actual_type, &actual_format, &nitems, 
				&bytes_after, (unsigned char **) &scm);

            nitems /= (sizeof (XStandardColormap)/4);
            for (; nitems > 0; ++scm, --nitems)
                if (scm->visualid == adata->awt_visInfo.visualid) {
                    adata->awt_cmap = scm->colormap;
                    share = 1;
                    break;
                }
        }

	if (!share) {
            adata->awt_cmap = XCreateColormap (awt_display, awt_root, 
					       adata->awt_visInfo.visual, 
					       AllocNone);
        }  
    }   

    awt_allocate_colors (adata);

    /* If depth is 8, allocate system colors also...  Here
     * we just get the array of System Colors and allocate
     * it which may be a bit wasteful (if only some were
     * changed). But we don't know which ones were changed
     * and alloc-ing a pixel that is already allocated won't
     * hurt. */

    if (adata->awt_depth == 8) {
	jint colorVals [java_awt_SystemColor_NUM_COLORS];
	jclass sysColors;
	jfieldID colorID;
	jintArray colors;

	/* Unlock now to initialize the SystemColor class */
	if (lock) {
	    AWT_UNLOCK ();
	}
	sysColors = (*env)->FindClass (env, "java/awt/SystemColor");
	if (lock) {
	    AWT_LOCK (); 
	}
	colorID = (*env)->GetStaticFieldID (env, sysColors, 
						   "systemColors",
						   "[I");
	  
	colors = (jintArray) (*env)->GetStaticObjectField 
						(env, sysColors, colorID);

	(*env)->GetIntArrayRegion (env, colors, 0,
				     java_awt_SystemColor_NUM_COLORS,
				     (jint *) colorVals);

	awt_allocate_systemrgbcolors (colorVals, 
			(java_awt_SystemColor_NUM_COLORS - 1), adata); 

    }	

    if (lock) {
	AWT_UNLOCK ();
    }
   
}

void initInverseGrayLut(int* prgb, int rgbsize, ColorData *cData) {
    int *idata;
    int g;

    if (!cData) {
        return;
    }

    idata = calloc(256, sizeof(int));
    if (!idata) {
        return;
    }
    cData->pGrayInverseLutData = idata;
    for (g = 0; g < 256; g++) {
        int i, mindist, d, besti;
        int *pg;
        mindist = 256;
        besti = 0;
        pg = prgb;
        for (i = 0; i < rgbsize; i++, pg++) {
            if (*pg == 0x0) {
                /* ignore transparent black */
                continue;
            }
            d = (*pg & 0xff) - g;
            if (d < 0) d = -d;
            if (d < mindist) {
                besti = i;
                if (d == 0) {
                    break;
                }
                mindist = d;
            }
        }
        *idata++ = besti;
    }
}


void freeICMColorData(ColorData *pData) {
    if (pData  && (pData->screendata == 0)) {
        if (pData->img_clr_tbl) {
            free(pData->img_clr_tbl);
        }
        if (pData->pGrayInverseLutData) {
            free(pData->pGrayInverseLutData);
        }
    free(pData);
    }
}


/*
 * state info needed for breadth-first recursion of color cube from
 * initial palette entries within the cube
 */

typedef struct {
    unsigned int depth;
    unsigned int maxDepth;

    unsigned char *usedFlags;
    unsigned int  activeEntries;
    unsigned short *rgb;
    unsigned char *indices;
    unsigned char *iLUT;
} CubeStateInfo;

#define INSERTNEW(state, rgb, index) do {                           \
        if (!state.usedFlags[rgb]) {                                \
            state.usedFlags[rgb] = 1;                               \
            state.iLUT[rgb] = index;                                \
            state.rgb[state.activeEntries] = rgb;                   \
            state.indices[state.activeEntries] = index;             \
            state.activeEntries++;                                  \
        }                                                           \
} while (0);
    

#define ACTIVATE(code, mask, delta, state, index) do {              \
    if (((rgb & mask) + delta) <= mask) {                           \
        rgb += delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb -= delta;                                               \
    }                                                               \
    if ((rgb & mask) >= delta) {                                    \
        rgb -= delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb += delta;                                               \
    }                                                               \
} while (0);

/* REMIND: does not deal well with bifurcation which happens when two
 * palette entries map to the same cube vertex
 */

static int
recurseLevel(CubeStateInfo *priorState) {
    int i;
    CubeStateInfo currentState;
    memcpy(&currentState, priorState, sizeof(CubeStateInfo));


    currentState.rgb = (unsigned short *)malloc(6
                                                * sizeof(unsigned short)
                                                * priorState->activeEntries);
    if (currentState.rgb == NULL) {
        return 0;
    }

    currentState.indices = (unsigned char *)malloc(6
                                                * sizeof(unsigned char)
                                                * priorState->activeEntries);

    if (currentState.indices == NULL) {
        free(currentState.rgb);
        return 0;
    }

    currentState.depth++;
    if (currentState.depth > priorState->maxDepth) {
        priorState->maxDepth = currentState.depth;
    }
    currentState.activeEntries = 0;
    for (i=priorState->activeEntries - 1; i >= 0; i--) {
        unsigned short rgb = priorState->rgb[i];
        unsigned char  index = priorState->indices[i];
        ACTIVATE(rgb, 0x7c00, 0x0400, currentState, index);
        ACTIVATE(rgb, 0x03e0, 0x0020, currentState, index);
        ACTIVATE(rgb, 0x001f, 0x0001, currentState, index);
    }
    if (currentState.activeEntries) {
        if (!recurseLevel(&currentState)) {
	    free(currentState.rgb);
	    free(currentState.indices);
            return 0;
        }
    }
    if (currentState.maxDepth > priorState->maxDepth) {
        priorState->maxDepth = currentState.maxDepth;
    }
    
    free(currentState.rgb);
    free(currentState.indices);
    return  1;
}

/*
 * REMIND: take core inversedLUT calculation to the shared tree and
 * recode the functions (Win32)awt_Image:initCubemap(),
 * (Win32)awt_Image:make_cubemap(), (Win32)AwtToolkit::GenerateInverseLUT(),
 * (Solaris)color:initCubemap() to call the shared codes.
 */
unsigned char*
initCubemap(int* cmap,
            int  cmap_len,
            int  cube_dim) {
    int i;
    CubeStateInfo currentState;
    int cubesize = cube_dim * cube_dim * cube_dim;
    unsigned char *useFlags;
    unsigned char *newILut = (unsigned char*)malloc(cubesize);
    if (newILut) {

      useFlags = (unsigned char *)calloc(cubesize, 1);
      
      if (useFlags == 0) {
          free(newILut);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in color:initCubemap()1\n");
#endif
          return NULL;
      }
      
        currentState.depth          = 0;
        currentState.maxDepth       = 0;
        currentState.usedFlags      = useFlags;
        currentState.activeEntries  = 0;
        currentState.iLUT           = newILut;

        currentState.rgb = (unsigned short *)
                                malloc(256 * sizeof(unsigned short));
        if (currentState.rgb == NULL) {
            free(newILut);
	    free(useFlags);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in color:initCubemap()2\n");
#endif
            return NULL;
        }

        currentState.indices = (unsigned char *)
                                malloc(256 * sizeof(unsigned char));
        if (currentState.indices == NULL) {
	    free(currentState.rgb);
            free(newILut);
	    free(useFlags);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in color:initCubemap()3\n");
#endif
            return NULL;
        }
 
        for (i = 0; i < 128; i++) {
            unsigned short rgb;
            int pixel = cmap[i];
            rgb = (pixel & 0x00f80000) >> 9;
            rgb |= (pixel & 0x0000f800) >> 6;
            rgb |=  (pixel & 0xf8) >> 3;
            INSERTNEW(currentState, rgb, i);
            pixel = cmap[255-i];
            rgb = (pixel & 0x00f80000) >> 9;
            rgb |= (pixel & 0x0000f800) >> 6;
            rgb |=  (pixel & 0xf8) >> 3;
            INSERTNEW(currentState, rgb, 255-i);
        }

        if (!recurseLevel(&currentState)) {
            free(newILut);
            free(useFlags);
            free(currentState.rgb);
	    free(currentState.indices);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in color:initCubemap()4\n");
#endif
	    return NULL;
	}

	free(useFlags);
	free(currentState.rgb);
	free(currentState.indices);
	
        return newILut;
    }

#ifdef DEBUG
	fprintf(stderr, "Out of memory in color:initCubemap()5\n");
#endif
    return NULL;
}

void
initDitherTables(ColorData* cData) {


    if(std_odas_computed) {
        cData->img_oda_red   = &(std_img_oda_red[0][0]);
        cData->img_oda_green = &(std_img_oda_green[0][0]);
        cData->img_oda_blue  = &(std_img_oda_blue[0][0]);
    } else {
        cData->img_oda_red   = &(std_img_oda_red[0][0]);
        cData->img_oda_green = &(std_img_oda_green[0][0]);
        cData->img_oda_blue  = &(std_img_oda_blue[0][0]);
        make_dither_arrays(256, cData);
        std_odas_computed = 1;
    }

}

static void make_dither_arrays(int cmapsize, ColorData *cData) {
    int i, j, k;

    /*
     * Initialize the per-component ordered dithering arrays
     * Choose a size based on how far between elements in the
     * virtual cube.  Assume the cube has cuberoot(cmapsize)
     * elements per axis and those elements are distributed
     * over 256 colors.
     * The calculation should really divide by (#comp/axis - 1)
     * since the first and last elements are at the extremes of
     * the 256 levels, but in a practical sense this formula
     * produces a smaller error array which results in smoother
     * images that have slightly less color fidelity but much
     * less dithering noise, especially for grayscale images.
     */
    i = (int) (256 / pow(cmapsize, 1.0/3.0));
    make_sgn_ordered_dither_array(cData->img_oda_red, -i / 2, i / 2);
    make_sgn_ordered_dither_array(cData->img_oda_green, -i / 2, i / 2);
    make_sgn_ordered_dither_array(cData->img_oda_blue, -i / 2, i / 2);

    /*
     * Flip green horizontally and blue vertically so that
     * the errors don't line up in the 3 primary components.
     */
    for (i = 0; i < 8; i++) {
        for (j = 0; j < 4; j++) {
            k = cData->img_oda_green[(i<<3)+j];
            cData->img_oda_green[(i<<3)+j] = cData->img_oda_green[(i<<3)+7 - j];
            cData->img_oda_green[(i<<3) + 7 - j] = k;
            k = cData->img_oda_blue[(j<<3)+i];
            cData->img_oda_blue[(j<<3)+i] = cData->img_oda_blue[((7 - j)<<3)+i];
            cData->img_oda_blue[((7 - j)<<3) + i] = k;
        }
    }
}
