/*
 * @(#)cursor.c	1.11 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#include "awt_p.h"
#include "java_awt_Cursor.h"
#include "sun_awt_motif_X11CustomCursor.h"

#include "jni.h"
#include "jni_util.h"

static jfieldID cDataID;
static jfieldID widthID;
static jfieldID heightID;

/*
 * Class:     sun_awt_motif_X11CustomCursor
 * Method:    cacheInit
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11CustomCursor_cacheInit
  (JNIEnv *env, jclass cls)
{
    jclass clsDimension = (*env)->FindClass(env, "java/awt/Dimension");
    cDataID = (*env)->GetFieldID(env, cls, "pData", "J");
    widthID = (*env)->GetFieldID(env, clsDimension, "width", "I");
    heightID = (*env)->GetFieldID(env, clsDimension, "height", "I");
}

/*
 * Class:     sun_awt_motif_X11CustomCursor
 * Method:    queryBestCursor
 * Signature: (Ljava/awt/Dimension;)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11CustomCursor_queryBestCursor
  (JNIEnv *env, jclass cls, jobject dimension)
{
    Window root;
    int width, height;
    
    AWT_LOCK();
    root = RootWindow(awt_display, DefaultScreen(awt_display));
    XQueryBestCursor(awt_display, root, 
                     (*env)->GetIntField(env, dimension, widthID),
                     (*env)->GetIntField(env, dimension, heightID),
                     &width, &height);
    (*env)->SetIntField(env, dimension, widthID, width);
    (*env)->SetIntField(env, dimension, heightID, height);
    AWT_UNLOCK();
}

/*
 * Class:     sun_awt_motif_X11CustomCursor
 * Method:    createCursor
 * Signature: ([B[BIIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11CustomCursor_createCursor
  (JNIEnv *env , jobject this, jbyteArray xorMask, jbyteArray andMask, 
   jint width, jint height, jint xHotSpot, jint yHotSpot)
{
    Cursor cursor;
    char *sourceBits, *maskBits;
    Window root;
    Pixmap source, mask;
    XColor black, white;

    AWT_LOCK();

    root = RootWindow(awt_display, DefaultScreen(awt_display));
    black.flags = DoRed | DoGreen | DoBlue;
    black.red = black.green = black.blue = 0x0000;
    XAllocColor(awt_display, defaultConfig->awt_cmap, &black);
    white.flags = DoRed | DoGreen | DoBlue;
    white.red = white.green = white.blue = 0xffff;
    XAllocColor(awt_display, defaultConfig->awt_cmap, &white);

    /* Create source pixmap. */
    sourceBits = (char *)(*env)->GetPrimitiveArrayCritical(env, xorMask, NULL);
    source = XCreateBitmapFromData(awt_display, root, sourceBits,
                                         width, height);

    /* Create mask pixmap */
    maskBits = (char *)(*env)->GetPrimitiveArrayCritical(env, andMask, NULL);
    mask = XCreateBitmapFromData(awt_display, root, maskBits,
                                       width, height);

    /* Create cursor */
    cursor = XCreatePixmapCursor(awt_display, source, mask, &black, &white,
                                 xHotSpot, yHotSpot);

    /* Free resources */
    XFreePixmap(awt_display, source);
    XFreePixmap(awt_display, mask);

    (*env)->ReleasePrimitiveArrayCritical(env, xorMask, sourceBits, JNI_ABORT);
    (*env)->ReleasePrimitiveArrayCritical(env, andMask, maskBits, JNI_ABORT);

    (*env)->SetLongField(env,this,cDataID,cursor);

    AWT_FLUSH_UNLOCK();
}

/*
 * Class:     sun_awt_motif_X11CustomCursor
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_sun_awt_motif_X11CustomCursor_finalize
  (JNIEnv *env, jobject this)
{
    Cursor cursor;

    AWT_LOCK();
    cursor = (Cursor)(*env)->GetLongField(env,this,cDataID);
    if (cursor != 0) {
        XFreeCursor(awt_display, cursor);
        (*env)->SetLongField(env,this,cDataID,0);
    }
    AWT_FLUSH_UNLOCK();
}
