/* 
 * @(#)fontObject_md.cpp	1.5 98/10/06
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A. 
 * All rights reserved. 
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You 
 * shall not disclose such Confidential Information and shall use 
 * it only in accordance with the terms of the license agreement 
 * you entered into with Sun. 
 */
 
/* 
 * @author Joe Warzecha
 */ 

#include "fontObject_md.h"
#include "hsTemplates.h"
#include <ctype.h>

X11FontObject::~X11FontObject () 
{
    if (fontName != NULL)
	free (fontName);
    if (PSfontName != NULL)
	free (PSfontName);
    if (familyName != NULL)
	free (familyName);
    if (slant != NULL)
	free (slant);
    if (style != NULL)
	free (style); 
}

// unicode char to lowercase
inline UInt16 toLowerU(UInt16 unicodeChar) 
{
    if ((unicodeChar > 0x40) && (unicodeChar <= 0x40 + 26)) {
        return unicodeChar + 0x20;
    }
    return unicodeChar;
}

static int equalUnicodeToAsciiNC (const UInt16 *name, int unicodeNameLen,
                               const char *ascii, int asciiBytes = 0)
 
{
    int retval = 0;
    if (!asciiBytes && ascii) {
        asciiBytes = strlen(ascii);
    }
    if (name && unicodeNameLen && asciiBytes && ascii && *ascii) {
        if (asciiBytes == unicodeNameLen) {
            int i;
            retval = 1;
            for (i=0; i<unicodeNameLen; i++) {
                if ((name[i] > 255) || (name[i] != (0xff & ascii[i]))) {
                    if (toLowerU(name[i]) == toLowerU(0xff & ascii[i])) {
                        continue;
                    }
                    retval = 0;
                    break;
                }
            }
        }
    }
    return retval;
}    

Boolean X11FontObject::MatchName(int nameID, const UInt16 *name,
                                       int nameLen)
{
    Boolean retval;

    if (fontName == NULL)
	MakeFontName ();
	
    retval = equalUnicodeToAsciiNC (name, nameLen, fontName, strlen(fontName));
    return retval;
}

void X11FontObject::MakeFontName () 
{
    int len;
    const char *propVal = GetFontNativeName ();

    char *end, *start;
    char *tmpFamily, *tmpSlant, *tmpStyle, *encoding;
    int useDefault = 0;
    char xlfd [512];

    strcpy (xlfd, propVal);
    end = xlfd;

    do {
	NEXT_HYPHEN; /* skip FOUNDRY */
        NEXT_HYPHEN; /* set FAMILY_NAME */
	tmpFamily = start;
        NEXT_HYPHEN; /* set STYLE_NAME */
	tmpStyle = start;
        NEXT_HYPHEN; /* set SLANT */
	tmpSlant = start;
	NEXT_HYPHEN; /* skip SETWIDTH_NAME */
        NEXT_HYPHEN; /* skip ADD_STYLE_NAME */
        NEXT_HYPHEN; /* skip PIXEL_SIZE */
        NEXT_HYPHEN; /* skip POINT_SIZE */
        NEXT_HYPHEN; /* skip RESOLUTION_X */
        NEXT_HYPHEN; /* skip RESOLUTION_Y */
        NEXT_HYPHEN; /* skip SPACING */
        NEXT_HYPHEN; /* skip AVERAGE_WIDTH */
    } while (0);
        	
    if (useDefault) { /* We didn't find what we wanted */
	tmpFamily = "Unknown";
	encoding = "iso8859-1";
	style = 0;
	slant = 0;
    } else
	encoding = (end + 1);


    char buf [512];	/* BAD, but would a font name ever be longer
			         * than this?? */
    char tmpBuf [100];

    /* Copy family name into buffer, capitalizing after spaces
     * in the family name */
    char *a = tmpFamily;
    char *b = buf;
    *b++ = toupper (*a++);
    while (*a != NULL) {
        if (*a == ' ') {
	    *b++ = toupper (*++a);
            a++;
         } else
            *b++ = *a++;
    }
    *b = '\0';
    familyName = strdup (buf);

    if ((tmpStyle != NULL) &&
	(strstr (tmpStyle, "medium") == 0) &&
	(strstr (tmpStyle, "normal") == 0)) {
	char *boldPart = strstr (tmpStyle, "bold");
        int index = 0;
	if (boldPart != NULL) 
	    index = boldPart - tmpStyle;
	strcpy (tmpBuf, tmpStyle);
	tmpBuf [0] = toupper (tmpBuf[0]);
	if (index != 0)
	    tmpBuf [index] = toupper (tmpBuf[index]);
	strcat (buf, " ");
	strcat (buf, tmpBuf);
	style = strdup (tmpBuf);
    }

    if (tmpSlant != NULL) {
	if (strcmp (tmpSlant, "i") == 0) {
	    strcat (buf, " Italic");
	    slant = strdup ("Italic");
	} else if (strcmp (tmpSlant, "o") == 0) {
	    strcat (buf, " Oblique");
	    slant = strdup ("Oblique");
	}
    }

    /* Check encoding */
    if ((strstr (encoding, "iso") == 0) &&
	(strstr (encoding, "fontspecific") == 0) &&
	(strstr (encoding, "symbol") == 0) &&
	(strstr (encoding, "dingbats") == 0)) {
	strcat (buf, " (");
	strcat (buf, encoding);
	strcat (buf, ")");
    }

    len = strlen (buf);
    fontName = (char *) calloc (1, sizeof (char) * (len + 1));
    strcpy (fontName, buf);

}

int X11FontObject::GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name)
{
    int len;
    if (fontName == NULL) {
	MakeFontName ();
    }

    len = strlen (fontName);
    if (name) {
	strcpy ((char *)name, fontName);
	platformID = 1;
	scriptID = 0;
    }

    return (len);
}
 
static void copyAsciiToUnicode(const char *src, UInt16 *dest, int copySize = 0)
{
    if (copySize == 0 ) {
        // REMIND add some boundary overflow checking
        while (*dest++ = (UInt16)(*src++)) {
        }
    } else {
        for (int i=0; i < copySize; i++) {
            dest[i] = src[i] & 0xff;
        }
    }
}
 
Unicode *strDupASCIItoUNICODE(const char *src, int& retNameLen, int maxInput = 0) {
    if (src) {
        if (!maxInput) {
            retNameLen = strlen(src);
        } else {
            retNameLen = maxInput;
        }
        Unicode *retval = new Unicode[retNameLen];
         
        if( nil != retval ) {
          copyAsciiToUnicode(src, retval, retNameLen);

        } else {
#ifdef DEBUG
          fprintf(
                  stderr,
                  "new returns nil. fontObject_md.cpp, strDupASCIItoUNICODE(const char*,int&,int)\n"
                  );
#endif
          retNameLen = 0;
        }
        return retval;
    } else {
        retNameLen = 0;
        return NULL;
    }
}

const Unicode *X11FontObject::GetFontName(int& nameLen)
{
    if (!fFontname) {
        fFontname = strDupASCIItoUNICODE(fontName, fFontNameLen);
    }
    nameLen = fFontNameLen;
    return fFontname;
}

fileFontObject *GetNativeFontObject () {
	return new X11FontObject ();
}

extern "C" {
extern jobject awt_lock;
}

void NativeLock (JNIEnv *env) 
{
#ifdef USE_NATIVE_RASTERIZER
   if (awt_lock != NULL)
       env->MonitorEnter (awt_lock);
#endif
}

void NativeUnlock (JNIEnv *env) 
{
#ifdef USE_NATIVE_RASTERIZER
   if (awt_lock != NULL)
       env->MonitorExit (awt_lock);
#endif
}
