/*
 * @(#)nrScalerContext.cpp	1.20 98/11/03
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*
 * @author sherman@japan
 * @author Charlton Innovations, Inc.
 * @author Joe Warzecha
 */

#include <stdio.h>

#include "t2kScalerContext.h"
#include "nrScalerContext.h"

#include "fontObject_md.h"
#include "hsTemplates.h"
#include "hsMemory.h"

#include <DPS/dpsXclient.h>
#include "nrOutlines.h"
#include <ctype.h>
#include <sys/utsname.h>

#include <jni.h>
#include <jni_util.h>

#define SHIFTFACTOR 16

#define HANDLE_EXCEPTION()		\
    if (env->ExceptionOccurred()) {	\
	if (debug) {			\
	    env->ExceptionDescribe ();	\
	}				\
	env->ExceptionClear();		\
	return;				\
    }

#define HANDLE_EXCEPTION2()		\
    if (env->ExceptionOccurred()) {	\
	if (debug) {			\
	    env->ExceptionDescribe ();	\
	}				\
	env->ExceptionClear();		\
	free (valsToConvert);		\
	return;				\
    }

#define MOVETO 1
#define LINETO 2
#define CURVETO 3
#define CLOSEPATH 4

extern "C" {
static GC pixmapGC = 0;
static Pixmap pixmap = 0;
static XChar2b *xChar = 0;
static int pixmapWidth = 0;
static int pixmapHeight = 0;
static DPSContext dpsContext = 0;
static short noDPS = 0;
static Atom psAtom = 0;
static Atom fontNameAtom = 0;
static char *osVersion = 0;

extern Window awt_root;
}

class X11nrScalerContext : public hsGScalerContext {
private:
    const fontObject    *fFont;
    Boolean             fDoAntiAlias;
    Boolean             fDoFracEnable;
    double              fMatrix[4];
    double		ptSize;
    int			fNumGlyphs;
    X11CharToGlyphMapper *fMapper;
    Boolean		fDoPaths;
    hsPathSpline	**pathCache;
    hsFixedPoint2	**pathAdvances;
    XFontStruct		*fXFont;
    t2kScalerContext	*t2k;
    Boolean		uset2k;
public:
    X11nrScalerContext(fontObject* fo,
                         const float matrix[],
                         Boolean   doAntiAlias,
                         Boolean   doFractEnable, int style);
    virtual ~X11nrScalerContext();

    virtual UInt32  CountGlyphs();
    virtual void    GenerateMetrics(
                        UInt16 glyphID, hsGGlyph* glyph,
                        hsFixedPoint2* advance);
    virtual void    GenerateMetricsWithImage(
                        UInt16 glyphID, hsGGlyph* glyph,
                        hsFixedPoint2* advance);
    virtual void    GenerateImage(
                        UInt16 glyphID,
                        const hsGGlyph* glyph,
                        void* buffer);
    virtual void    GeneratePath(UInt16 index, hsPathSpline* path,
                        hsFixedPoint2* advance);
    virtual void    GetLineHeight(hsFixedPoint2* ascent,
                        hsFixedPoint2* descent,
                        hsFixedPoint2* baseline,
                        hsFixedPoint2* leading,
                        hsFixedPoint2* maxAdvance);
    virtual void GetCaretAngle(hsFixedPoint2* caret);
    virtual CharToGlyphMapper *getMapper();
    virtual Boolean isStyleSupported(int aStyle) { return false; }
};

//////////////////////////////////////////////////////////////////////////////

#define SINGLE_BYTE 1
#define DOUBLE_BYTE 2

X11CharToGlyphMapper::X11CharToGlyphMapper 
	(char* javaclass, short numBytes, char *psEnc, Unicode maxVal, 
	 Unicode minVal) : nBytes (numBytes), psEncoding (psEnc), 
	 reverseMap (0), convertedCount (65536), reverseCount (65536)
{

#ifdef DEBUG
	Boolean debug = true;
#else
	Boolean debug = false;
#endif

    /* If javaclass is NULL, its an iso8859-1 font, with 256 entries.
     * convertedGlyphs array has 256 entries, reverseMap array has
     * 256 entries also 
     */
    if (javaclass == NULL) {
	convertedGlyphs = (Unicode *) calloc (1, sizeof (Unicode) * 256);
	reverseMap = (Unicode *) calloc (1, sizeof (Unicode) * 256);
	for (int i = 0; i < 256; i++) {
	    convertedGlyphs [i] = i;
	    reverseMap [i] = i;
	}
	convertedCount = 256;
	reverseCount = 256;
 	return;
    }


    /* Create an array to store the unicode->X11 character code values */

    convertedGlyphs = (Unicode *) calloc (1, sizeof (Unicode) * 65536);

    /* If the X11 character code is single byte, the array which maps
     * back from X11 character code to unicode value only needs to
     * hold 256 entries. This array is used when we have to convert
     * an X11 character code back to unicode to pass to the t2k scaler
     * context.
     */
    if (nBytes == SINGLE_BYTE) {
	reverseCount = 256;
        reverseMap = (Unicode *) calloc (1, sizeof (Unicode) * 256);
    } else {
	reverseCount = 65536;
        reverseMap = (Unicode *) calloc (1, sizeof (Unicode) * 65536);
    }

    /* Find the correct CharToByte converter class */

    JNIEnv *env = (JNIEnv *) JNU_GetEnv (jvm, JNI_VERSION_1_2);
    jclass converterClass = env->FindClass (javaclass);
    HANDLE_EXCEPTION();

    if (converterClass == NULL) {
	return;
    }

    /* Get a reference to the constructor we need to use */

    jmethodID converterConstruct = env->GetMethodID (converterClass,
						     "<init>", "()V");
    HANDLE_EXCEPTION();

    if (converterConstruct == NULL) {
	return;
    }

    /* Create the new CharToByte converter object */

    jobject converterObject = env->NewObject (converterClass, 
					      converterConstruct);

    HANDLE_EXCEPTION();

    if (converterObject == NULL) {
        return;
    }

    /* If unicode value can't be converted, set byte to return to 0 */

    jmethodID mid = env->GetMethodID (converterClass, "setSubstitutionBytes",
				      "([B)V");
    HANDLE_EXCEPTION();
    if (mid != NULL) {
	jbyteArray jtmp = env->NewByteArray (1);
	HANDLE_EXCEPTION();
	byte bb [1] = {0};
	env->SetByteArrayRegion (jtmp, 0, 1, (jbyte*) bb);
	HANDLE_EXCEPTION();
	env->CallVoidMethod (converterObject, mid, jtmp);
	HANDLE_EXCEPTION();
    }

    /* Get a reference to the "convert" method of the CharToByteConverter */

    jmethodID converterMethod = env->GetMethodID (converterClass, 
						"convert", "([CII[BII)I");
    HANDLE_EXCEPTION();

    if (converterMethod == NULL) {
	return;
    }
	
    /* Count is total number of unicode values to convert to X11 character
     * code . Used in some cases where the boundaries are well known so 
     * the number of codes converted is lower and faster. Normally though
     * this is the entire unicode range (0 - 0xffff).
     */
    int count = maxVal - minVal + 1;
    Unicode *valsToConvert = (Unicode *) calloc (1, sizeof (Unicode) * count);
    for (int i = 0; i < count; i++) { 
        valsToConvert [i] = minVal + i;
    }

    /* charArray is set to the unicode values to convert,
     * byteArray is the array in which the X11 character codes will be put
     */
    jcharArray charArray = env->NewCharArray (count);
    HANDLE_EXCEPTION2();
    jbyteArray byteArray = env->NewByteArray (count * nBytes);
    HANDLE_EXCEPTION2();

    env->SetCharArrayRegion (charArray, 0, count, valsToConvert);
    HANDLE_EXCEPTION2();

    /* Now loop until all of the unicode values are converted. The
     * 'convert' method may throw an exception if the unicode value
     * can not be converted to X11 character code. If so, we can
     * call 'convert' again, starting with the next unicode value.
     */
    int done = 0;
    int currentChar = -1;
    jmethodID nextCharMethod = NULL;
    jint ret;
    do {
	currentChar++;
        ret = env->CallIntMethod (converterObject, converterMethod,
			          charArray, currentChar, count,
				  byteArray, currentChar * nBytes, 
						count * nBytes);

        if (env->ExceptionOccurred ()) { 
	    // env->ExceptionDescribe ();
            env->ExceptionClear ();
	    if (nextCharMethod == NULL) {
    	        nextCharMethod = env->GetMethodID (converterClass, 
						   "nextCharIndex", "()I");
	    }

	    currentChar = env->CallIntMethod (converterObject, nextCharMethod);
	    HANDLE_EXCEPTION2();
        } else {
	    done = 1;
	}

    } while (!done);

    /* Read the X11 character codes and store in the convertedGlyphs array */

    jbyte *charCodes = (jbyte *) calloc (1, sizeof (jbyte) * (count * nBytes));
    env->GetByteArrayRegion (byteArray, 0, count * nBytes, charCodes);
    if (env->ExceptionOccurred ()) {
	if (debug) {
	    env->ExceptionDescribe ();
	}
	env->ExceptionClear();
    } else {
        for (i = 0; i < count; i++) { 
            if (nBytes == 1) {
	        convertedGlyphs [i + minVal] = 0x00ff & charCodes [i];
	    } else {
	        convertedGlyphs [i + minVal] = (charCodes [i * 2] << 8) +
			           	        charCodes [(i * 2) + 1]; 
	    }
	}
    }

    free (valsToConvert);
    free (charCodes);
}

X11CharToGlyphMapper::~X11CharToGlyphMapper() 
{
    free (convertedGlyphs);
    free (reverseMap);
}
 
void 
X11CharToGlyphMapper::AddAsciiChars() 
{
    /* This check is probably not required, but make sure we
     * allocated enough memory for the ascii characters 
     */
    if ((convertedCount >= 0xff) && (reverseCount >= 0xff)) {
	for (int i = 0x00; i < 0x100; i++) {
	    convertedGlyphs [i] = i;
	    reverseMap [i] = i;
	}
    }
}

void
X11CharToGlyphMapper::CharsToGlyphs(int count, const Unicode unicodes[],
                             UInt32 glyphs[]) const 
{
    for (int i = 0; i < count; i++) {
	if (unicodes [i] >= convertedCount) { 
	    glyphs [i] = 0;
	} else {
	    glyphs [i] = convertedGlyphs [unicodes [i]];
	    if ((glyphs [i] != 0) && (glyphs [i] < reverseCount)) {
	        reverseMap [glyphs [i]] = unicodes [i];
	    }
	}
    }
}

const Unicode
X11CharToGlyphMapper::GlyphToUnicode (UInt16 glyphID)
{
    if (glyphID > reverseCount) {
	return 0;
    }

    return (reverseMap [glyphID]);
}

Boolean 
X11CharToGlyphMapper::canDisplay (Unicode ch) const 
{
    Unicode uVal[1];
    UInt32 xVal [1];
    uVal [0] = ch;
    CharsToGlyphs (1, uVal, xVal);
    return (xVal[0] != 0);
}


struct MapperPair {
    char *encoding;
    X11CharToGlyphMapper *mapper;

    MapperPair () {}
    MapperPair (char *enc, X11CharToGlyphMapper *map) :
			mapper (map) { encoding = strdup (enc); }

    friend int operator==(const MapperPair& a, const MapperPair& b) {
	if (strcmp (a.encoding, b.encoding) == 0)
	    return 1;
	return 0;
    }
};

static hsDynamicArray<MapperPair>* x11Mappers = nil;


void MakeNativeMapper (const char *xlfdString) 
{
#ifdef USE_NATIVE_RASTERIZER

    char *tmpEncoding;
    char *tmpFamily;
    char *family = NULL;
    char *encoding = NULL;

    char *end, *start;
    char xlfd [512];
    int useDefault = 0;

    strcpy (xlfd, xlfdString);
    end = xlfd;

    do {
	NEXT_HYPHEN; /* skip FOUNDRY */
	NEXT_HYPHEN; /* set FAMILY_NAME */
	family = start;
	NEXT_HYPHEN; /* skip STYLE_NAME */
	NEXT_HYPHEN; /* skip SLANT */
	NEXT_HYPHEN; /* skip SETWIDTH_NAME */
	NEXT_HYPHEN; /* skip ADD_STYLE_NAME */
	NEXT_HYPHEN; /* skip PIXEL_SIZE */
	NEXT_HYPHEN; /* skip POINT_SIZE */
	NEXT_HYPHEN; /* skip RESOLUTION_X */
	NEXT_HYPHEN; /* skip RESOLUTION_Y */
 	NEXT_HYPHEN; /* skip SPACING */
	NEXT_HYPHEN; /* skip AVERAGE_WIDTH */
    } while (0);

    end++;

    if (useDefault) {
	encoding = "iso8859-1";
	family = encoding;
    } else
        encoding = end;

    tmpFamily = family;
    tmpEncoding = encoding;

    if (strstr (tmpEncoding, "fontspecific") != 0) {
	if (strstr (tmpFamily, "dingbats") != 0)
	    tmpEncoding = "dingbats";
	else if (strstr (tmpFamily, "symbol") != 0)
	    tmpEncoding = "symbol";
	else
	    tmpEncoding = "iso8859-1";
    } else if (strstr (tmpEncoding, "dingbats") != 0) {
	tmpEncoding = "dingbats";
    } else if (strstr (tmpEncoding, "symbol") != 0) {
        tmpEncoding = "symbol";
    }

    X11CharToGlyphMapper *mapper;
    int found = -1;

    if (x11Mappers == nil) 
        x11Mappers = new hsDynamicArray<MapperPair>;
    else
	found = x11Mappers->Find (MapperPair (tmpEncoding, nil));

    if (found >= 0) {
	return;
    } else {
	char *psPart = NULL;
        char *jclass = NULL;
	int nBytes = SINGLE_BYTE;
      	Unicode maxU = 0xffff;
	Unicode minU = 0;
	Boolean addAscii = false;

    	if (strcmp (tmpEncoding, "iso8859-1") == 0) {
	    maxU = 0xff;
    	} else if (strcmp (tmpEncoding, "iso8859-2") == 0) {
	    jclass = "sun/io/CharToByteISO8859_2";
    	} else if (strcmp (tmpEncoding, "jisx0208.1983-0") == 0) {
	    jclass = "sun/awt/motif/CharToByteX11JIS0208";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-H");
	    /* In this case, need to also create the jisx0201
	       CharToGlyph mapper, because of oddities in the
	       way the font names are mapped.
	     */
	    char *extraEncoding = "jisx0201.1976-0";
	    int isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11JIS0201", SINGLE_BYTE, 
		 	 strdup (".Hankaku"), maxU, minU);
		tmpMapper->AddAsciiChars ();
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
        } else if (strcmp (tmpEncoding, "jisx0201.1976-0") == 0) {
	    jclass = "sun/awt/motif/CharToByteX11JIS0201";
	    addAscii = true;
	    psPart = strdup (".Hankaku");
	    /* In this case, need to also create the jisx0208
	       CharToGlyph mapper, because of oddities in the
	       way the font names are mapped.
	     */
	    char *extraEncoding = "jisx0208.1983-0";
	    int isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11JIS0208", DOUBLE_BYTE, 
			 strdup ("-H"), maxU, minU);
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
    	} else if (strcmp (tmpEncoding, "jisx0212.1990-0") == 0) {
	    jclass = "sun/io/CharToByteJIS0212", DOUBLE_BYTE;
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-Hojo-H");
  	} else if (strcmp (tmpEncoding, "dingbats") == 0) {
	    jclass = "sun/awt/motif/CharToByteX11Dingbats";
	    minU = 0x2701;
	    maxU = 0x27be;
	} else if (strcmp (tmpEncoding, "symbol") == 0) {
	    jclass = "sun/awt/CharToByteSymbol";
	    minU = 0x0391;
	    maxU = 0x22ef;
        } else if (strcmp (tmpEncoding, "iso8859-4") == 0) {
	    jclass = "sun/io/CharToByteISO8859_4";
    	} else if (strcmp (tmpEncoding, "iso8859-5") == 0) {
	    jclass = "sun/io/CharToByteISO8859_5";
    	} else if (strcmp (tmpEncoding, "iso8859-6") == 0) {
	    jclass = "sun/io/CharToByteISO8859_6";
    	} else if (strcmp (tmpEncoding, "iso8859-7") == 0) {
	    jclass = "sun/io/CharToByteISO8859_7";
    	} else if (strcmp (tmpEncoding, "iso8859-8") == 0) {
	    jclass = "sun/io/CharToByteISO8859_8";
        } else if (strcmp (tmpEncoding, "iso8859-9") == 0) {
	    jclass = "sun/io/CharToByteISO8859_9";
    	} else if (strcmp (tmpEncoding, "iso8859-15") == 0) {
	    jclass = "sun/io/CharToByteISO8859_15_FDIS";
        } else if (strcmp (tmpEncoding, "ksc5601.1987-0") == 0) {
	    jclass ="sun/awt/motif/CharToByteX11KSC5601";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-KSC-H");
        } else if (strcmp (tmpEncoding, "ksc5601.1987-1") == 0) {
	    jclass ="sun/io/CharToByteEUC_KR";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-KSC-EUC-H");
    	} else if (strcmp (tmpEncoding, "cns11643-1") == 0) {
	    jclass = "sun/awt/motif/CharToByteX11CNS11643P1";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-CNS1-H");
	    /* In this case, need to also create the cns11643-2
	       and cns11643-3 CharToGlyph mappers, because of oddities 
	       in the way the font names are mapped on X11.
	     */
	    char *extraEncoding = "cns11643-2";
	    int isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11CNS11643P2", DOUBLE_BYTE,
			 strdup ("-CNS2-H"), maxU, minU);
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
	    extraEncoding = "cns11643-3";
	    isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11CNS11643P3", DOUBLE_BYTE,
			 strdup ("-CNS3-H"), maxU, minU);
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
    	} else if (strcmp (tmpEncoding, "cns11643-2") == 0) {
	    jclass = "sun/awt/motif/CharToByteX11CNS11643P2";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-CNS2-H");
	    /* In this case, need to also create the cns11643-2
	       and cns11643-3 CharToGlyph mappers, because of oddities 
	       in the way the font names are mapped on X11.
	     */
	    char *extraEncoding = "cns11643-1";
	    int isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11CNS11643P1", DOUBLE_BYTE,
			 strdup ("-CNS1-H"), maxU, minU);
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
	    extraEncoding = "cns11643-3";
	    isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11CNS11643P3", DOUBLE_BYTE,
			 strdup ("-CNS3-H"), maxU, minU);
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
        } else if (strcmp (tmpEncoding, "cns11643-3") == 0) {
	    jclass = "sun/awt/motif/CharToByteX11CNS11643P3";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-CNS3-H");
	    /* In this case, need to also create the cns11643-1
	       and cns11643-2 CharToGlyph mappers, because of oddities 
	       in the way the font names are mapped on X11.
	     */
	    char *extraEncoding = "cns11643-1";
	    int isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11CNS11643P1", DOUBLE_BYTE,
			 strdup ("-CNS1-H"), maxU, minU);
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
	    extraEncoding = "cns11643-2";
	    isThere = x11Mappers->Find (MapperPair (extraEncoding, nil));

	    if (isThere < 0) {
	        X11CharToGlyphMapper *tmpMapper = new X11CharToGlyphMapper 
			("sun/awt/motif/CharToByteX11CNS11643P2", DOUBLE_BYTE,
			 strdup ("-CNS2-H"), maxU, minU);
	        x11Mappers->Append (MapperPair (extraEncoding, tmpMapper));
	    }
    	} else if (strcmp (tmpEncoding, "gb2312.1980-0") == 0) {
	    jclass = "sun/awt/motif/CharToByteX11GB2312";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-GB-H");
	} else if (strstr (tmpEncoding, "big5") != 0) {
	    jclass = "sun/io/CharToByteBig5";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-B5-H");
	    addAscii = true;
        } else if (strcmp (tmpEncoding, "tis620.2533-0") == 0) {
	    jclass = "sun/io/CharToByteTIS620";
    	} else if (strcmp (tmpEncoding, "gbk-0") == 0) {
	    jclass = "sun/io/CharToByteGBK";
	    nBytes = DOUBLE_BYTE;
	    psPart = strdup ("-GBK-EUC-H");
	}
    
	mapper = new X11CharToGlyphMapper (jclass, nBytes, psPart, maxU, minU);
	if (addAscii) {
	    mapper->AddAsciiChars ();
	}
	x11Mappers->Append (MapperPair (tmpEncoding, mapper));
    }
#endif
}

int
CreatePixmapAndGC (int width, int height)
{
    if (width < 100) {
	width = 100;
    }
    if (height < 100) {
	height = 100;
    }
    pixmapHeight = height;
    pixmapWidth = width;
    if (pixmap != NULL) {
	XFreePixmap (awt_display, pixmap);
    }
    pixmap = XCreatePixmap (awt_display, awt_root, pixmapWidth, 
			    pixmapHeight, 1);
    if (pixmap == NULL) {
	return BadAlloc;
    }

    if (pixmapGC != NULL) {
	XFreeGC (awt_display, pixmapGC);
    }
    pixmapGC = XCreateGC (awt_display, pixmap, 0, 0);
    if (pixmapGC == NULL) {
	return BadAlloc;
    }

    XFillRectangle (awt_display, pixmap, pixmapGC, 0, 0, pixmapWidth, 
		    pixmapHeight);
    XSetForeground (awt_display, pixmapGC, 1);
    if (xChar == (XChar2b *) NULL) {
        xChar = (XChar2b *) calloc (1, sizeof (XChar2b));
	if (xChar == NULL) {
	    return BadAlloc;
 	}
    }

    return Success;
}

#define NO_POINTSIZE -1.0
/* #undef TRUETYPE_WORKS */
#define TRUETYPE_WORKS

X11nrScalerContext::X11nrScalerContext(
                        fontObject *fo, const float matrix[],
                        Boolean doAntiAlias, Boolean doFractEnable, int style)
    : fFont(fo), fDoAntiAlias(doAntiAlias), fDoFracEnable(doFractEnable),
	fNumGlyphs (-1), fMapper (0), fDoPaths (true), fXFont (0), t2k (0),
	uset2k (false), pathCache (0), pathAdvances (0), ptSize (NO_POINTSIZE)
{
    fMatrix[0] = matrix[0];
    fMatrix[1] = matrix[1];
    fMatrix[2] = matrix[2];
    fMatrix[3] = matrix[3];

    /* Determine X11 point size to use for this font. If the matrix
     * is one that X11 can't handle (arbitrary rotation, or different
     * x and y scaling, and the font isn't kNRFontFormat, let the
     * t2k scaler context handle it. If the font is a kNRFontFormat
     * one, we'll do the best we can to figure out some point size
     * to use. Currently, the only case where the nrScalerContext
     * is used for a TrueType font is the case where the matrix
     * is something like [12.0 0.0 0.0 12.0]. Once I add support
     * for simple rotation (90/180/270) and flipping (x or y axis)
     * then this can be expanded.
     */

    float nrPtSize = NO_POINTSIZE;

    if ((fMatrix [1] == 0.0) && (fMatrix [2] == 0.0)) {

        /* First case. Matrix is something like [12.0 0.0 0.0 12.0].
	 * We can handle this. 
	 * NOTE: can change the 'if' to this:
	 *    if (abs (fMatrix [0]) == abs (fMatrix [3])) {
	 * once I can handle flipping of the bitmaps.
	 */

	if ((fMatrix [0] == fMatrix [3]) && (fMatrix [0] > 0)) {
	    ptSize = abs (fMatrix [0]);

	    if (ptSize > 1.0) {
	        nrPtSize = ptSize;
	    }

	/* Else, matrix is something like [14.0 0.0 0.0 12.0] or
	 * [12.0 0.0 0.0 -12.0]. Use fMatrix [0] for point size right 
         * now for F3 fonts.
	 */

	} if (abs (fMatrix [3]) > abs (fMatrix [0])) {
	    ptSize = abs (fMatrix [3]);
	} else {
	    ptSize = abs (fMatrix [0]);
	}

    /* If the matrix is anything like [0.0 12.0 12.0 0.0] or
     * [12.0 12.0 12.0 12.0], let the t2k Scaler handle this.
     * For F3, use the larger of fMatrix [0] and fMatrix [3].
     */
    } else { 
	if (abs (fMatrix [3]) > abs (fMatrix [0])) {
	    ptSize = abs (fMatrix [3]);
	} else {
	    ptSize = abs (fMatrix [0]);
	}
    }
	
    /* If Font Format is not kNRFontFormat (F3), create a t2k Scaler
     * context 
     */

    if (fFont->GetFormat () != kNRFontFormat) {
	t2k = new t2kScalerContext (fo, matrix, doAntiAlias, doFractEnable, style);

	/* REMIND: 9/17/98 For now, TRUETYPE_WORKS is defined, but
	 *	 if you note below, only Japanese TrueType fonts
	 *	 will use the native rasterizer code. Why? Because
	 * 	 if this code is turned on for all TrueType fonts,
	 *	 there is a chance that the X server will crash
	 * 	 (see bug 4045384). However, in JA locale, the
	 * 	 TrueType fonts are the default fonts used and they
	 *	 do not see the crash occurring. So, in order to
	 * 	 increase performance for 2d in Japan, we allow
	 * 	 the TrueType fonts to go through the native
	 * 	 rasterizer codepath.
	 *
	 *	 Another note here is that if there in the case
	 *	 of using the same JA tt font for two different
	 *	 X11 encodings, for one entry, the code will
	 *	 treat the format as kTrueType and for the 2nd,
	 *	 it will treat is as kNR (X11). Why? Because
	 *	 even though you could potentially create more
	 *	 than one fontObject per font file, whenever
	 *       these objects are accessed, it it typically
	 *	 done via a MatchName which only looks at
	 *	 the font name (and each fontObject that references
	 *       the same font file would have the same name).
	 *	 So, in the X11GraphicsEnvironment code, when
	 *	 the 2nd xlfd string is found that references
	 *	 a font file that has already been "registered",
	 *	 it does not register it again with this other
	 *	 encoding. Instead, it allows the "native" entry
	 *	 for this xlfd string to get registered, since
	 *	 this xlfd string is also found via the XListFonts
	 *	 done to find X11 fonts. The only place where
	 *	 this makes a bit of difference is in GeneratePath.
	 *	 In this case, for the one case where the font
	 *	 was registered as a kTrueType, it will retrieve
	 *	 the outline from the t2kScalerContext. The font
	 *	 that was registered as a kNR (native) will retrieve
	 *	 it via DPS. 
	 *
	 *	 This should all be reworked when there is time
	 *	 to architect a better solution.
	 *
	 */

#ifndef TRUETYPE_WORKS
	uset2k = true;
#else 		/* Use this code path for TrueType fonts */

	/* Figure out if we should use t2k rather than native rasterizer. 
	 * First, if anything but TrueType font, use t2k. This includes 
	 * Type1 even though X11 can handle Type 1 fonts. However due to
  	 * bugs in the XServer (DPS) in both 2.5.1 and 2.6, it makes
	 * more sense to use t2k.
	 */
	if (fFont->GetFormat () != kTrueTypeFontFormat) {
	    uset2k = true;	
	} else {
	
	    /* For TrueType fonts, check. If any of these conditions are
	     * true, it will cause t2k to be used instead: 
    	     *    fDoAntiAlias is true
	     *    fDoFracEnable is true
	     *    nrPtSize is equal to NO_POINTSIZE 
	     *    os version is prior to 2.6  (TrueType support wasn't added
	     * 		until Solaris 2.6)
	     *	  xlfd does not contain the string "jisx" (jisx indicates
	     *		one of the JA encodings, jisx0201.1976-0,
	     *		jisx0208.1983-0 or jisx0212.1990-0). REMIND: We
	     * 		currently only use this code path for ja TT fonts.
	     *
	     * If none of these are true, can use X11 get retrieve the
	     * metrics and bitmaps.
	     */

	    /* Check if we've retrived the os version already */
	    if (osVersion == NULL) {
		struct utsname name;
		uname (&name);
		osVersion = strdup (name.release);
	    }

	    if ((fDoAntiAlias) || (fDoFracEnable) || 
		(nrPtSize == NO_POINTSIZE) ||
		(strstr (osVersion, "5.5") != 0) ||
		(strstr (fFont->GetFontNativeName (), "jisx") == 0)) 
	    { 
		uset2k = true;
	    }
	}
#endif

    }
}

X11nrScalerContext::~X11nrScalerContext()
{
    if (fXFont != NULL) {
        XFreeFont (awt_display, fXFont);
	fXFont = NULL;
    }
    if (t2k != NULL) {
	delete t2k;
	t2k = NULL;
    }
    if (pathAdvances != NULL) {
	for (int i = 0; i < fNumGlyphs; i++) {
	    if (pathAdvances [i] != NULL) {
		delete pathAdvances [i];
	    }
	}
	free (pathAdvances);
	pathAdvances = NULL;
    }
    if (pathCache != NULL) {
	for (int i = 0; i < fNumGlyphs; i++) {
	    if (pathCache [i] != NULL) {
		hsPathSpline::Delete (pathCache [i]);
		delete pathCache [i];
	    }
	}
	free (pathCache);
	pathCache = NULL;
    }
}


UInt32
X11nrScalerContext::CountGlyphs()
{

#ifndef USE_NATIVE_RASTERIZER
    return 0;
#endif

    if (fNumGlyphs == -1) {

   	/* If using the t2kScalerContext, return 0, and let that
	 * figure out the number of glyphs.
	 */
        if (uset2k) {
	    return 0;
	    //fNumGlyphs = t2k->CountGlyphs ();
	    //return fNumGlyphs;
	}

	/* If we couldn't figure out a valid X11 point size, return 0
	 * glyphs so the default font can be used instead.
	 */
	if (ptSize == NO_POINTSIZE) {
	    fNumGlyphs = 0;
	    return 0;
	} 

	/* Get the xlfd string */

	char tmpName [512];	/* Should be large enough */
	char xlfd [512];
	strcpy (xlfd, fFont->GetFontNativeName ());

	/* If xlfd has a "%d" in it (for point size, then we can
	 * just set it. If not, then we have to replace the '0'
	 * for point size with the correct value. NOTE: if
	 * fMatrix [0] is zero, just allow that to be set in
	 * the xlfd string. and 12 pt will probably get selected.
	 * Later in GenerateMetrics and GenerateImage, check
	 * for fMatrix[0] equal to zero and if so, return zeros
	 * for metrics and no bitmaps.
	 */

	if (strstr (xlfd, "%d") != 0)
	    sprintf (tmpName, xlfd, (int)(ptSize * 10.0));
	else {
	    char *end, *start;
	    char *lhs, *rhs;
	    char *family, *style, *slant, *encoding, *pixelsize,
	         *setwidth, *addstyle, *pointsize, *spacing, *foundry;
	    int useDefault = 0;

	    end = xlfd;
	    lhs = xlfd;
	    do {
	        SKIP_HYPHEN;	/* Set FOUNDRY */
	        SKIP_HYPHEN;	/* Set FAMILY_NAME */
		SKIP_HYPHEN;	/* Set Style Name */
		SKIP_HYPHEN;	/* Set Slant */
		SKIP_HYPHEN;	/* Set SETWIDTH_NAME */
		NEXT_HYPHEN;	/* Set ADD_STYLE_NAME */
		NEXT_HYPHEN;	/* Get pixel size */
		pixelsize = start;
		NEXT_HYPHEN;	/* Get point size */
		pointsize = start;
	    } while (0);

	    /* This shouldn't happen, but if useDefault = 1, then 
	     * something went wrong parsing this xlfd string. If this
	     * is not  kNRFontFormat, use the t2kScaler, else  
	     * use default font string.
	     */
	    if (useDefault) {
        	if (fFont->GetFormat () != kNRFontFormat) {
		    uset2k = true;
		    return 0;
		    //fNumGlyphs = t2k->CountGlyphs ();
		    //return fNumGlyphs;
		}
		lhs = "-*-*-*-*-*-*";
		pointsize = "*";
	 	rhs = "0-0-p-0-iso8859-1";
	    } else {
	        rhs = end + 1;
	    }

	    /* See if either point or pixel size was valid */
	    if (((strlen (pointsize) == 1) && 
		 ((pointsize[0] == '*') || (pointsize[0] == '0') ||
		  (pointsize[0] == (char) NULL))) &&
		((strlen (pixelsize) == 1) &&
		 ((pixelsize[0] == '*') || (pixelsize[0] == '0') ||
		  (pixelsize[0] == (char) NULL)))) {

		char c= pixelsize [0];
		if (c == (char) NULL)
		    c = '0';
		sprintf (tmpName, "%s-%c-%d-%s", lhs, c, 
			 	((int)(ptSize * 10.0)), rhs);
	    } else {	/* Its ok, use as is */
		strcpy (tmpName, xlfd);
	    }
	}

	/* OK, its a valid X11 font */

	fXFont = XLoadQueryFont (awt_display, tmpName);

	/* If we didn't get the XFontStruct and we're not dealing
	 * with a kNRFontFormat, then use the t2k ScalerContext
	 * instead. 
	 */
	if (fXFont == NULL) {	/* Couldn't find the font */
       	    if (fFont->GetFormat () != kNRFontFormat) {
		uset2k = true;
		return 0;
		//fNumGlyphs = t2k->CountGlyphs ();
		//return fNumGlyphs;
	    }
	    fNumGlyphs = 0;
	    return 0;
	}

	fNumGlyphs = (fXFont->max_byte1 << 8) + fXFont->max_char_or_byte2 + 1;

    }

    return fNumGlyphs;
}

void
X11nrScalerContext::GenerateMetrics(
                UInt16 glyphID,
                hsGGlyph* glyph,
                hsFixedPoint2* advance)
{

    /* If using t2k, convert X11 character code back to Unicode,
     * call t2k CharToGlyphMapper, and t2k GenerateMetrics
     */
    if ((uset2k) && (t2k != NULL)) {
	Unicode *uVal = new Unicode [1];
	uVal [0] = fMapper->GlyphToUnicode (glyphID);
	CharToGlyphMapper *parentMapper = t2k->getMapper();
	UInt32 *index = new UInt32 [1];
	parentMapper->CharsToGlyphs (1, uVal, index);
	t2k->GenerateMetrics (index [0], glyph, advance);
	return;
    } 

    /* IF any of the following conditions are true, we can't retrieve
     * the metrics, so return zeros. Note we don't want to return
     * kNoMetrics_Width in the fWidth field because if we do, we'll
     * keep getting called and since we can't get the metrics, there's
     * no point to that.
     */
    if ((glyphID >= fNumGlyphs) || (fXFont == 0) || (ptSize == NO_POINTSIZE)) {
        if (glyph) {
	    memset(glyph, 0, sizeof(*glyph));
        }
        if (advance) {
            memset(advance, 0, sizeof(*advance));
        }
        return;
    }

    int fMinIndex =  (fXFont->min_byte1 << 8) + fXFont->min_char_or_byte2;

    if (glyphID < fMinIndex) {
	/* Sometimes the default_char field of the XFontStruct isn't 
	 * initialized to anything, so it can be a large number. So, I 
	 * check to see if its less than the largest possible value
	 * (which happens to be fNumGlyphs) and if so, then I use it.
	 * Otherwise, I just use the fMinIndex.
	 */

	if ((fXFont->default_char != 0) && 
	    (fXFont->default_char < fNumGlyphs))
	    glyphID = fXFont->default_char;
	else
	    glyphID = fMinIndex;
    }

    XCharStruct xcs;
	
    /* If number of glyphs is 256 or less, the metrics are
     * stored correctly in the XFontStruct for each
     * character. If the # characters is more (double byte
     * case), then these metrics seem flaky and theres no
     * way to determine if they have been set or not. 
     */
    if ((fNumGlyphs < 256) &&
        (fXFont->per_char != NULL))
        xcs = fXFont->per_char [glyphID - fXFont->min_char_or_byte2];
    else {
        int direction, ascent, descent;

    	if (pixmap == NULL) {
            int retVal = CreatePixmapAndGC (100, 100);
	    if (retVal != Success) {
        	if (glyph) {
	    	    memset(glyph, 0, sizeof(*glyph));
        	}
        	if (advance) {
            	    memset(advance, 0, sizeof(*advance));
        	}
		JNIEnv *env = (JNIEnv *) JNU_GetEnv (jvm, JNI_VERSION_1_2);
		JNU_ThrowOutOfMemoryError (env,
                        "Cannot create native data structure.");
		return;
	    }
	}
	
	xChar->byte1 = (unsigned char) (glyphID >> 8);
	xChar->byte2 = (unsigned char) glyphID;
	XTextExtents16 (fXFont, xChar, 1, &direction, &ascent, 
			&descent, &xcs);
    }
    advance->fY = 0;
/* There is a problem with Type 1 fonts on solaris 2.5.1. The
   metrics are way off. I thought this block of code might help
   fix the problem, but it causes more so its commented out for now.
    int theWidth = xcs.width;
    if ((xcs.lbearing + xcs.rbearing) > theWidth)
	theWidth = xcs.lbearing + xcs.rbearing;
    advance->fX = theWidth << SHIFTFACTOR;
    glyph->fWidth = theWidth;
*/
    advance->fX = xcs.width << SHIFTFACTOR;
    glyph->fWidth = xcs.rbearing - xcs.lbearing;

    glyph->fHeight = xcs.ascent + xcs.descent; 
    glyph->fRowBytes = glyph->fWidth;
    glyph->fTopLeft.fX = xcs.lbearing << SHIFTFACTOR;
    glyph->fTopLeft.fY = -xcs.ascent << SHIFTFACTOR;

}

void
X11nrScalerContext::GenerateMetricsWithImage(
                UInt16 glyphID,
                hsGGlyph* glyph,
                hsFixedPoint2* advance)
{
    this->GenerateMetrics (glyphID, glyph, advance);
    UInt32 imageSize = glyph->fHeight * glyph->fRowBytes;
    void *image = 0;
    if (imageSize > 0) {
        image = HSMemory::SoftNew(imageSize);
        if (image)
            this->GenerateImage (glyphID, glyph, image);
        else {
            HSMemory::Delete(image);
            image = 0;
	}
    } 
    glyph->fImage = image;
}

void
X11nrScalerContext::GenerateImage(
                    UInt16 glyphID,
                    const hsGGlyph* glyph,
                    void* buffer)
{

    /* Check for instances where we won't get a bitmap. If a
     * a very small point size was used, theres a chance that
     * that the RowBytes and Height are zero, in which case
     * return now since we won't get a bitmap.
     */
    if ((uset2k) && (t2k != NULL)) {
	Unicode *uVal = new Unicode [1];
	uVal [0] = fMapper->GlyphToUnicode (glyphID);
	CharToGlyphMapper *parentMapper = t2k->getMapper();
	UInt32 *index = new UInt32 [1];
	parentMapper->CharsToGlyphs (1, uVal, index);
	t2k->GenerateImage (index [0], glyph, buffer);
    }

    /* If any of the following conditions are true, then we aren't
     * going to get a valid image, so return an empty one.
     */
    if ((glyphID >= fNumGlyphs) || (fXFont == 0) || (ptSize == NO_POINTSIZE) ||
	(glyph->fRowBytes == 0) || (glyph->fHeight == 0)) 
    { 
	if (buffer) {
    	    UInt32 imageSize = glyph->fHeight * glyph->fRowBytes;
	    memset(buffer, 0, imageSize);
        }
        return;
    }

    int fMinIndex =  (fXFont->min_byte1 << 8) + fXFont->min_char_or_byte2;

    if (glyphID < fMinIndex) {

	/* Sometimes the default_char field of the XFontStruct isn't 
	 * initialized to anything, so it can be a large number. So, I 
	 * check to see if its less than the largest possible value
	 * (which happens to be fNumGlyphs) and if so, then I use it.
	 * Otherwise, I just use the fMinIndex.
	 */

	if ((fXFont->default_char != 0) && 
	    (fXFont->default_char < fNumGlyphs))
	    glyphID = fXFont->default_char;
	else
	    glyphID = fMinIndex;
    }

    XImage *ximage;
    char *imageData;
    char *tmpImage;
    char *tPtr;
    int nbytes;
    int i, j;
	    
    int startX = -glyph->fTopLeft.fX >> SHIFTFACTOR;
    int theAscent = abs (glyph->fTopLeft.fY) >> SHIFTFACTOR;

    if ((pixmap == NULL) || (glyph->fRowBytes > pixmapWidth) ||
	(glyph->fHeight > pixmapHeight)) 
    {
        int retVal = CreatePixmapAndGC (glyph->fRowBytes, glyph->fHeight);
	if (retVal != Success) {
	    if (buffer) {
    	        UInt32 imageSize = glyph->fHeight * glyph->fRowBytes;
	        memset(buffer, 0, imageSize);
            }
	    JNIEnv *env = (JNIEnv *) JNU_GetEnv (jvm, JNI_VERSION_1_2);
	    JNU_ThrowOutOfMemoryError (env,
                        "Cannot create native data structure.");
	    return;
	}
    }

    XSetFont (awt_display, pixmapGC, fXFont->fid);
    xChar->byte1 = (unsigned char) (glyphID >> 8);
    xChar->byte2 = (unsigned char) glyphID;

    XSetForeground (awt_display, pixmapGC, 0);
    XFillRectangle (awt_display, pixmap, pixmapGC, 0, 0, 
		    pixmapWidth, pixmapHeight);
    XSetForeground (awt_display, pixmapGC, 1);
    XDrawString16 (awt_display, pixmap, pixmapGC, startX, theAscent, xChar, 1);
    ximage = XGetImage (awt_display, pixmap, 0, 0, glyph->fRowBytes, 
			glyph->fHeight, AllPlanes, XYPixmap);
    imageData = ximage->data;
    nbytes = glyph->fRowBytes / 8;
    if ((glyph->fRowBytes % 8) > 0)
	nbytes++;
    tmpImage = (char *) calloc (1, sizeof (char) * (nbytes * glyph->fHeight));

    tPtr = tmpImage;
    for (i = 0; i < glyph->fHeight; i++) {
	for (j = 0; j < nbytes; j++) 
	    *tPtr++ = *imageData++;
	imageData = ximage->data + ((i + 1) * ximage->bytes_per_line);
    }
	    
    /* This is the CopyBW2ToGrey8 function from t2kScalerContext.cpp */

    const UInt8* srcRow = (UInt8*)tmpImage;
    UInt8* dstRow = (UInt8*)buffer;
    int wholeByteCount = glyph->fWidth >> 3;
    int remainingBitsCount = glyph->fWidth & 7;
    int height = glyph->fHeight;

    while (height--) {
        const UInt8* src8 = srcRow;
        UInt8* dstByte = dstRow;
        unsigned srcValue;

        srcRow += nbytes;
        dstRow += glyph->fRowBytes;

        for (i = 0; i < wholeByteCount; i++) {
            srcValue = *src8++;
            for (j = 0; j < 8; j++) {
		if (ximage->bitmap_bit_order == LSBFirst) {
                    *dstByte++ = (srcValue & 0x01) ? 0xFF : 0;
                    srcValue >>= 1;
		} else {		/* MSBFirst */
                    *dstByte++ = (srcValue & 0x80) ? 0xFF : 0;
                    srcValue <<= 1;
		}
            }
        }
        if (remainingBitsCount) {
            srcValue = *src8;
            for (j = 0; j < remainingBitsCount; j++) {
		if (ximage->bitmap_bit_order == LSBFirst) {
                    *dstByte++ = (srcValue & 0x01) ? 0xFF : 0;
                    srcValue >>= 1;
		} else {		/* MSBFirst */
                    *dstByte++ = (srcValue & 0x80) ? 0xFF : 0;
                    srcValue <<= 1;
		}
            }
        }
    }

    free (tmpImage);

}

extern "C" {
/* Text handler for DPS. We define our own rather than using the
 * default so we don't get annoying messages printed out from DPS
 * (like if a font can't be found). 
 */

static Boolean dpsError = false;

void
X11nrDPSTextProc (DPSContext ctxt, char *buf, unsigned long count) 
{
#ifdef DPS_DEBUG
    DPSDefaultTextBackstop (ctxt, buf, count);
#endif
}

void 
X11nrDPSErrorProc (DPSContext ctxt, DPSErrorCode errorCode, 
		   long unsigned int arg1, long unsigned int arg2)
{
    dpsError = true;
    if (dpsContext != NULL) {
	DPSDestroySpace (DPSSpaceFromContext (dpsContext));
        dpsContext = NULL;
    }
#ifdef DPS_DEBUG
    DPSDefaultErrorProc (ctxt, errorCode, arg1, arg2);
#endif
} 

}  /* extern "C" */

Boolean
CheckFontList (char *name, char *style, char *slant, char *psEnc)
{
    /* This function includes a list of known fonts for which the
     * creation of DPS font name typically does not work. There
     * doesnt seem to be any way around this..
     */  
    if (strcmp (name, "LucidaBright-DemiBold") == 0) {
	strcpy (name, "Lucida-BrightDemiBold");
	return true;
    } else if (strcmp (name, "LucidaBright-DemiBoldItalic") == 0) {
	strcpy (name, "Lucida-BrightDemiBoldItalic");
	return true;
    } else if (strstr (name, "Gill") != 0) {
        strcpy (name, "GillSans");
    } else if (strstr (name, "NewCenturySchoolbook") != 0) {
        strcpy (name, "NewCenturySchlbk");
    } else if (strstr (name, "Palatino") != 0) {
        strcpy (name, "Palatino-Roman");
    } else if (strstr (name, "chancery") != 0) {
        strcpy (name, "ZapfChancery-MediumItalic");
        return true;
    } else if (strstr (name, "Gothic") != 0) {
        strcpy (name, "GothicBBB-Medium");
    } else if (strstr (name, "Ryumin") != 0) {
        strcpy (name, "Ryumin");
    } else if (strstr (name, "Mincho") != 0) {
	strcpy (name, "HG-MinchoL");
    } else if (strstr (name, "Heisei") != 0) {
	strcpy (name, "HeiseiMin-W3H");
	return true;
    } else {
        return false;
    }
 
    if ((style != NULL) || (slant != NULL)) {
        strcat (name, "-");
        if (style != NULL)
           strcat (name, style);
        if (slant != NULL)
           strcat (name, slant);
    }
 
    if (psEnc != NULL) {
        strcat (name, psEnc);
    }
 
    return true;
}

void
X11nrScalerContext::GeneratePath(
                    UInt16 index,
                    hsPathSpline* path,
                    hsFixedPoint2* advance)
{

    /* If this is a TrueType or Type1 font, then let the
     * t2kScalerContext geneneate the path.
     */
    if ((fFont->GetFormat () != kNRFontFormat) && (t2k != NULL)) {
	Unicode *uVal = new Unicode [1];
	uVal [0] = fMapper->GlyphToUnicode (index);
	CharToGlyphMapper *parentMapper = t2k->getMapper();
	UInt32 *glyphID = new UInt32 [1];
	parentMapper->CharsToGlyphs (1, uVal, glyphID);
	t2k->GeneratePath (glyphID [0], path, advance);
	return;
    }

    /* NOTE:: This is subject to much error. What we have is
     * xlfd string and what we want is PostScript font name.
     * In some cases, we're lucky and the ADOBE_POSTSCRIPT_NAME
     * was set for the font. If not, we'll try and figure out
     * it out from the FONT_NAME property and do some fun things
     * like remove spaces from font names, try smashing words
     * together, or adding hyphens, capitailizing letters, etc..
     */

    if ((!fDoPaths) || (noDPS)) {
        if (path) {
	    path->fContourCount = 0;
	    path->fContours = nil;
        }
        if (advance) {
            memset(advance, 0, sizeof(*advance));
	}
	return;
    }

    /* See if the path is already in the cache */

    if (pathCache != NULL) {
	if (pathCache [index] != NULL) {
	    pathCache [index]->Copy (path);
	    if (advance) {
		if (pathAdvances [index] != NULL) {
		    advance->fX = pathAdvances [index]->fX;
		    advance->fY = pathAdvances [index]->fY;
		} else {
		    memset (advance, 0, sizeof (*advance));
		}
	    }
	    return;
	}
    }


    dpsError = false;
    X11FontObject *xFont = (X11FontObject *) fFont;
    char *psName = xFont->GetPSName ();


    if ((psName != NULL) && (psName [0] == '\0')) {
	fDoPaths = false;
        if (path) {
	    path->fContourCount = 0;
	    path->fContours = nil;
        }
        if (advance) {
            memset(advance, 0, sizeof(*advance));
	}
	return;
    }
	
    if (dpsContext == NULL) {

	/* We're not doing anything exciting.. just use
	 * a black & white colormap. Note that if we
	 * use XDPSCreateSimpleContext or don't specify
	 * a gray or rgb colormap, DPS will allocate one
	 * even though we don't want it.
	 */
	XStandardColormap grayMap, rgbMap;
	grayMap.colormap = None;
	grayMap.red_max = 1;
	grayMap.red_mult = -1;
	grayMap.base_pixel = 1;
	rgbMap.colormap = None;
	rgbMap.red_max = 0;
	rgbMap.green_max = 0;
	rgbMap.blue_max = 0;
	rgbMap.red_mult = 0;
	rgbMap.green_mult = 0;
	rgbMap.blue_mult = 0;
	rgbMap.base_pixel = 0;

	dpsContext = XDPSCreateContext (awt_display, pixmap, pixmapGC, 0, 
					pixmapHeight, NULL, &grayMap, 
					&rgbMap, NULL, X11nrDPSTextProc,
					X11nrDPSErrorProc, (DPSSpace) NULL);

	/* If dpsContext is NULL, then we're probably running on a
	 * server that doesn't support DPS.
	 */
	if (dpsContext == NULL) {
	    noDPS = 1;
	    fDoPaths = false;
            if (path) {
	        path->fContourCount = 0;
	        path->fContours = nil;
            }
            if (advance) {
                memset(advance, 0, sizeof(*advance));
	    }
	    return;
	}

	DPSSetContext (dpsContext);
	OutlineInit ();
        if (dpsError) {
	    char tempName [1];

	    /* Throw exception, outline not available */
	    fDoPaths = false;
            if (path) {
	        path->fContourCount = 0;
	        path->fContours = nil;
            }
            if (advance) {
                memset(advance, 0, sizeof(*advance));
	    }
    
	    /* Set PSfontName of X11FontObject since we
             * can't do outlines for this font, so we
	     * might as well not try it next time.
	     */
	    tempName [1] = '\0';
	    xFont->SetPSName (tempName); 
	    return;
        }
    }


    /* If font name is null,  we gotta try to figure out some kind of name */
    char tempName [512]; 	/* Should be large enough */
    if (psName == NULL) {
	unsigned long atomValue;
	short foundFont;
	char *fName;
	char *tmpStyle = xFont->GetStyle ();
	char *tmpSlant = xFont->GetSlant ();

	/* Get correct name, use atoms */
	if (psAtom == NULL)
	    psAtom = XInternAtom (awt_display, "_ADOBE_POSTSCRIPT_FONTNAME", 
				  False);
	Bool found = XGetFontProperty (fXFont, psAtom, &atomValue);

	if (!found) {
	    if (fontNameAtom == NULL)
		fontNameAtom = XInternAtom (awt_display, "FULL_NAME", False);
	    found = XGetFontProperty (fXFont, fontNameAtom, &atomValue);
	}
	
	if (found) {
	    fName = XGetAtomName (awt_display, atomValue);
	    strcpy (tempName, fName);
	} else {	/* Construct from info from X11FontObject */
	    strcpy (tempName, xFont->GetFamilyName ());
	    if ((tmpStyle != NULL) || (tmpSlant != NULL)) {
		strcat (tempName, " ");
		if (tmpStyle != NULL)
		    strcat (tempName, tmpStyle);
		if (tmpSlant != NULL)
		    strcat (tempName, tmpSlant);
	    }
	}

	
	/* Now, remove spaces from font names and replace with
	 * hyphens. PostScript font names don't have spaces in
	 * them, but they do have hyphens.
	 */

	char *a = tempName;
	*a = toupper (*a);
	while (*++a != NULL) {
	    if (*a == ' ') {
		*a = '-';
	    }
	}

	/* If theres a PS encoding, then we need to append it */

	char *psPart = fMapper->GetPSEncoding ();
	if (psPart != NULL) {
	    strcat (tempName, psPart);
	}

 	CheckForFont (tempName, &foundFont);
        if (dpsError) {
	    /* Throw exception, outline not available */
	    fDoPaths = false;
            if (path) {
	        path->fContourCount = 0;
	        path->fContours = nil;
            }
            if (advance) {
                memset(advance, 0, sizeof(*advance));
	    }
    
	    /* Set PSfontName of X11FontObject since we
             * can't do outlines for this font, so we
	     * might as well not try it next time.
	     */
	    tempName [0] = '\0';
	    xFont->SetPSName (tempName); 
	    return;
        }

        if (!foundFont) {

	    /* If not found, see if its one of the font names we have
	     * problems with.
	     */
	    foundFont = CheckFontList (tempName, tmpStyle, tmpSlant, psPart);

	    if (foundFont)
		CheckForFont (tempName, &foundFont);

	    if ((!foundFont) || (dpsError)) {

		/* Throw exception ?? */
	        fDoPaths = false;
                if (path) {
	            path->fContourCount = 0;
	            path->fContours = nil;
                }
                if (advance) {
               	    memset(advance, 0, sizeof(*advance));
	        }
	        /* Set PSfontName of X11FontObject since we
 	         * can't do outlines for this font, so we
	         * might as well not try it next time.
	         */
	        tempName [0] = '\0';
	        xFont->SetPSName (tempName); 
	        return;
	    }
	}

	psName = tempName;

        /* Set the PS name of the X11FontObject so we don't have to
         * figure it out again.
         */
        xFont->SetPSName (tempName);
    }

    /* OK now generate the outline */

    unsigned char xByte1 = (unsigned char) (index >> 8);
    unsigned char xByte2 = (unsigned char) index;
    int nOps, nPoints, nPaths;   

    DrawOutline (psName, fMatrix [0], xByte1, xByte2,
		 &nOps, &nPoints, &nPaths);

    if (dpsError) {
	/* Throw exception, outline not available */
	fDoPaths = false;
        if (path) {
	    path->fContourCount = 0;
	    path->fContours = nil;
        }
        if (advance) {
            memset(advance, 0, sizeof(*advance));
	}

	/* Set PSfontName of X11FontObject since we
         * can't do outlines for this font, so we
	 * might as well not try it next time.
	 */
	tempName [0] = '\0';
	xFont->SetPSName (tempName); 
	return;
    }

    int *ops = (int *) calloc (1, sizeof (int) * nOps);
    int *pathPoints = (int *) calloc (1, sizeof (int) * nPaths);
    float *points = (float *) calloc (1, sizeof (float) * nPoints);

    GetOutline (nOps, nPoints, nPaths, ops, pathPoints, points);

    if (dpsError) {
	/* Throw exception, outline not available */
	fDoPaths = false;
        if (path) {
	    path->fContourCount = 0;
	    path->fContours = nil;
        }
        if (advance) {
            memset(advance, 0, sizeof(*advance));
	}

	/* Set PSfontName of X11FontObject since we
         * can't do outlines for this font, so we
	 * might as well not try it next time.
	 */
	tempName [0] = '\0';
	xFont->SetPSName (tempName); 
	return;
    }

    path->fContourCount = nPaths;
    if (nPaths == 0) {
	path->fContours = nil;
        return;
    }
    path->fContours = new hsPathContour[nPaths];

    for (int i = 0; i < nPaths; i++) {
        int ctrlBitLongs = hsPathContour::ControlBitLongs(pathPoints [i]);
        path->fContours[i].fPointCount = pathPoints [i];
        path->fContours[i].fPoints  = new hsPoint2 [pathPoints [i]];
        path->fContours[i].fControlBits  = new UInt32 [ctrlBitLongs];
 
        HSMemory::Clear(path->fContours[i].fControlBits, 
			ctrlBitLongs * sizeof(UInt32));
    }
 
    /* OK, now loop through the ops array. Ops values are:
     *	 1: moveto
     *   2: lineto
     *   3: curveto
     *   4: closepath
     * Associated points are in the points array. Remember that
     * the values are returned in the order they are to be used,
     * ie, points[0] is the X value for the first operation (most
     * likely a moveto) and points[1] is the Y value.
     */

    float curX, curY;
    float newX, newY;
    int iPath = 0;
    int pointCount = 0;
    int pointIndex = 0;
    int ctrlBitIndex;
    int ctrlBitShift;
    for (i = 0; i < nOps; i++) {
	switch (ops [i]) {
	    case MOVETO:
		curX = points [pointIndex++];
		curY = points [pointIndex++];
		break;

	    case LINETO:

		if ((pointCount == 0) || 
		    (path->fContours[iPath].fPoints[pointCount-1].fX != curX) ||
		    (path->fContours[iPath].fPoints[pointCount-1].fY != -curY))
		{
		    path->fContours [iPath].fPoints [pointCount].fX = curX;
		    path->fContours [iPath].fPoints [pointCount].fY = -curY;
		    pointCount++;
		}

		newX = points [pointIndex++];
		newY = points [pointIndex++];

		path->fContours [iPath].fPoints [pointCount].fX = newX;
		path->fContours [iPath].fPoints [pointCount].fY = -newY;
		pointCount++;

		curX = newX;
		curY = newY;
		break;

	    case CURVETO:

		if ((pointCount == 0) || 
		    (path->fContours[iPath].fPoints[pointCount-1].fX != curX) ||
		    (path->fContours[iPath].fPoints[pointCount-1].fY != -curY))
		{
		    path->fContours [iPath].fPoints [pointCount].fX = curX;
		    path->fContours [iPath].fPoints [pointCount].fY = -curY;
		    pointCount++;
		}

		newX = points [pointIndex++];
		newY = points [pointIndex++];

		path->fContours [iPath].fPoints [pointCount].fX = newX;
		path->fContours [iPath].fPoints [pointCount].fY = -newY;
		ctrlBitIndex = pointCount >> 5;
		ctrlBitShift = 31 - (pointCount & 31);
		path->fContours [iPath].fControlBits [ctrlBitIndex] |=
				(1UL << ctrlBitShift);
		pointCount++;

		newX = points [pointIndex++];
		newY = points [pointIndex++];

		path->fContours [iPath].fPoints [pointCount].fX = newX;
		path->fContours [iPath].fPoints [pointCount].fY = -newY;
		ctrlBitIndex = pointCount >> 5;
		ctrlBitShift = 31 - (pointCount & 31);
		path->fContours [iPath].fControlBits [ctrlBitIndex] |=
				(1UL << ctrlBitShift);
		pointCount++;

		newX = points [pointIndex++];
		newY = points [pointIndex++];

		path->fContours [iPath].fPoints [pointCount].fX = newX;
		path->fContours [iPath].fPoints [pointCount].fY = -newY;
		pointCount++;

		curX = newX;
		curY = newY;
		break;

	    case CLOSEPATH:
		/* If pointcount is 0, then this "path" had nothing in it. 
		 * So, decrease the number of paths by one, and delete
		 * the fContour entry.
		 */
		if (pointCount == 0) {
		    if (path->fContourCount > 0) {
			int lastEntry = path->fContourCount - 1;
    		        path->fContourCount = path->fContourCount - 1;
			delete [] path->fContours [lastEntry].fPoints;
			delete [] path->fContours [lastEntry].fControlBits;
			if (path->fContourCount == 0) {
			    delete [] path->fContours;
			    path->fContours = NULL;
			}
		    }
		} else {
		    if ((path->fContours [iPath].fPoints [0].fX ==
		         path->fContours [iPath].fPoints [pointCount -1].fX) &&
		        (path->fContours [iPath].fPoints [0].fY ==
		         path->fContours [iPath].fPoints [pointCount -1].fY)) 
		    {
		        path->fContours [iPath].fPointCount = pointCount - 1;
		    }

		    iPath++;
		}

		pointCount = 0;
		break;
	
	    default:
		break;
	}
    }

    /* At the end, curX, curY is the advance values */
    if (advance) {
	advance->fX = ((int) curX) << SHIFTFACTOR;
	advance->fY = ((int) curY) << SHIFTFACTOR;
    }

    if (pathCache == NULL) {
	pathCache = (hsPathSpline **) calloc (1, sizeof (hsPathSpline *) *
					      fNumGlyphs);
	pathAdvances = (hsFixedPoint2 **) calloc (1, sizeof (hsFixedPoint2 *) *
					      fNumGlyphs);
    }

    pathCache [index] = path->Copy (nil);
    pathAdvances [index] = new hsFixedPoint2;
    pathAdvances [index]->fX = ((int) curX) << SHIFTFACTOR;
    pathAdvances [index]->fY = ((int) curY) << SHIFTFACTOR;
    
}

void
X11nrScalerContext::GetLineHeight(
                    hsFixedPoint2* ascent,
                    hsFixedPoint2* descent,
                    hsFixedPoint2* baseline,
                    hsFixedPoint2* leading,
                    hsFixedPoint2* maxAdvance)
{
    if ((uset2k) && (t2k != NULL)) {
	t2k->GetLineHeight (ascent, descent, baseline,
				leading, maxAdvance);
	return;
    }

    if (fXFont != NULL) {
	if (ascent) {
            ascent->fX = 0;
            ascent->fY = -fXFont->max_bounds.ascent << SHIFTFACTOR;
	}
	if (descent) {
            descent->fX = 0;
            descent->fY = fXFont->max_bounds.descent << SHIFTFACTOR;
	}

	/* I have no idea what this should be! */
	if (baseline) {
            baseline->fX = 16384 << SHIFTFACTOR; 
	    baseline->fY = 0;
	}

	if (leading) {
            leading->fX = 0;
            leading->fY = -1 << SHIFTFACTOR;
	}

	if (maxAdvance) {
            maxAdvance->fY = 0;
            maxAdvance->fX = fXFont->max_bounds.width << SHIFTFACTOR;
	}
	return;
    }

    if (ascent) {
        ascent->fX = 0;
        ascent->fY = 0;
    }

    if (descent) {
        descent->fX = 0;
        descent->fY = 0;
    }

    if (baseline) {
        baseline->fX = 0;
        baseline->fY = 0;
    }

    if (leading) {
        leading->fX = 0;
        leading->fY = 0;
    }

    if (maxAdvance) {
        maxAdvance->fX = 0;
        maxAdvance->fY = 0;
    }
}

void
X11nrScalerContext::GetCaretAngle(hsFixedPoint2* caret)
{
    if (t2k != NULL) {
	t2k->GetCaretAngle (caret);
	return;
    }

    this->GetLineHeight (caret, nil, nil, nil, nil);
}


//////////////////////////////////////////////////////////////////////////////

CharToGlyphMapper*
X11nrScalerContext::getMapper()
{
    if ((uset2k) && (t2k != NULL)) {
	return (t2k->getMapper ());
    }
	
    if (fMapper == NULL) {

	char *end, *start, *family;
	int useDefault = 0;
	char xlfd [512];
	strcpy (xlfd, fFont->GetFontNativeName ());
	end = xlfd;

	do {
	    NEXT_HYPHEN; /* skip FOUNDRY */
	    NEXT_HYPHEN; /* skip FAMILY_NAME */
	    family = start;
	    NEXT_HYPHEN; /* skip STYLE_NAME */
	    NEXT_HYPHEN; /* skip SLANT */
	    NEXT_HYPHEN; /* skip SETWIDTH_NAME */
	    NEXT_HYPHEN; /* skip ADD_STYLE_NAME */
	    NEXT_HYPHEN; /* skip PIXEL_SIZE */
	    NEXT_HYPHEN; /* skip POINT_SIZE */
	    NEXT_HYPHEN; /* skip RESOLUTION_X */
	    NEXT_HYPHEN; /* skip RESOLUTION_Y */
 	    NEXT_HYPHEN; /* skip SPACING */
	    NEXT_HYPHEN; /* skip AVERAGE_WIDTH */
	} while (0);

	end++;

	char *encoding;
	if (useDefault)
	    encoding = "iso8859-1";
	else
	    encoding = end;

	/* Special case for dingbats (and symbol when its available) */
	if (strstr (encoding, "fontspecific") != 0) {
	    if (strstr (family, "dingbats") != 0)
	        encoding = "dingbats";
	    else if (strstr (family, "symbol") != 0)
	        encoding = "symbol";
	    else
		encoding = "iso8859-1";
	} else if (strstr (encoding, "dingbats") != 0) {
	    encoding = "dingbats";
	} else if (strstr (encoding, "symbol") != 0) {
	    encoding = "symbol";
	}
	

        int found = x11Mappers->Find (MapperPair (encoding, nil));
        if (found >= 0)
	    fMapper = x11Mappers->Get (found).mapper;
	else  	/* Dont know how this could happen */
	    fMapper = x11Mappers->Get (0).mapper;
	
    }

    return fMapper;
}

//////////////////////////////////////////////////////////////////////////////

hsGScalerContext*
nrFontScaler::CreateContext(hsConstDescriptor desc)
{
    const FontStrikeDesc *r = (FontStrikeDesc *)hsDescriptor_Find(
                                                        desc,
                                                        kFontStrikeDescTag,
                                                        nil,
                                                        nil);
    return new X11nrScalerContext(
                    r->fFontObj, r->fMatrix,
                    r->fDoAntiAlias, r->fDoFracEnable, r->fStyle);
}

