/*
 * @(#)mlib_v_ImageConv3x3_8nw.c	1.4 98/09/06
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#pragma ident	"@(#)mlib_v_ImageConv3x3_8nw.c	1.13	98/09/02 SMI"

/*
 * FUNCTION
 *      mlib_conv3x3_8nw.c - convolve an 8-bit image with a 3x3 kernel
 *                      (edge = no write)
 *
 *
 * SYNOPSIS
 *  mlib_status
 *  mlib_conv3x3_8nw( mlib_image *src,
 *                    mlib_image *dst,
 *                    mlib_s32 *kernel,
 *                    mlib_s32 scalef_expon
 *                    mlib_s32 cmask )
 * ARGUMENT
 *      src           Pointer to structure of source image
 *      dst           Pointer to structure of destination image
 *      kernel        Convolution kernel
 *      scalef_expons The scaling factor to convert the input integer
 *                    coefficients into floating-point coefficients:
 *                    floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask         Channel mask to indicate the channels to be convolved.
 *                    Each bit of which represents a channel in the image. The
 *                    channels corresponded to 1 bits are those to be processed.
 *
 * DESCRIPTION
 *      A 2-D convolution (3x3 kernel) for 8-bit images.
 */

#include <stdlib.h>
#include <mlib_image.h>
#include <vis_proto.h>

static
int mlib_round_8[16] = { 0x00400040, 0x00200020, 0x00100010, 0x00080008,
        0x00040004, 0x00020002, 0x00010001, 0x00000000,
        0x00000000, 0x00000000, 0x00000000, 0x00000000,
        0x00000000, 0x00000000, 0x00000000, 0x00000000};

/************************************************************************/
#define LOAD_INSERT_STORE_U8(channeld)                                   \
                                                                         \
  sd = *ddst++;                                                          \
                                                                         \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;            \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;            \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;            \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;            \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;            \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;            \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;            \
  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += channeld;
/************************************************************************/
#define GET_SRC_DST_PARAMETERS()                                         \
  adr_src = mlib_ImageGetData(src);                                      \
  adr_dst = mlib_ImageGetData(dst);                                      \
  slb = mlib_ImageGetStride(src);                                        \
  dlb = mlib_ImageGetStride(dst);                                        \
  dh = mlib_ImageGetHeight(dst);                                         \
  dw = mlib_ImageGetWidth(dst);
/************************************************************************/
#define LOAD_KERNEL_INTO_FLOAT()                                         \
  k1k2 = vis_to_float((kernel[0] & 0xFFFF0000) |                         \
                    ((kernel[1] >> 16) & 0xFFFF));                       \
  k3k4 = vis_to_float((kernel[2] & 0xFFFF0000) |                         \
                    ((kernel[3] >> 16) & 0xFFFF));                       \
  k5k6 = vis_to_float((kernel[4] & 0xFFFF0000) |                         \
                    ((kernel[5] >> 16) & 0xFFFF));                       \
  k7k8 = vis_to_float((kernel[6] & 0xFFFF0000) |                         \
                    ((kernel[7] >> 16) & 0xFFFF));                       \
  k9k9 = vis_to_float((kernel[8] & 0xFFFF0000) |                         \
                    ((kernel[8] >> 16) & 0xFFFF));
/************************************************************************/
#define PREPARE_INTERM_BUFFERS()                                         \
  buff_src = (mlib_d64*) mlib_malloc(4 * buf_slb * sizeof(mlib_d64));    \
  if (buff_src == NULL) return MLIB_FAILURE;                             \
                                                                         \
  sbuf1 = buff_src;                                                      \
  sbuf2 = sbuf1 + buf_slb;                                               \
  sbuf3 = sbuf2 + buf_slb;                                               \
  dbuf = sbuf3 + buf_slb;
/************************************************************************/
#define PREPARE_TO_LOAD_LINE(p_buf, p_line)                              \
  prow = p_buf;                                                          \
  dsa = (mlib_d64 *) vis_alignaddr(p_line, 0);                           \
  sd1 = dsa[0]; dsa++;
/************************************************************************/
#define LOAD_LINE_INTO_BUFFER(n)                                         \
  for (i = 0; i < (dw + n); i+= 8) {                                     \
    sd0 = sd1; sd1 = dsa[0];                                             \
    *prow++ = vis_faligndata(sd0, sd1);                                  \
    dsa++;                                                               \
  }
/************************************************************************/
#define LOOP_INI()                                                       \
  ddst = (mlib_d64*)dbuf;                                                \
  da = d_a;                                                              \
  dend = da + (dw - 1);                                                  \
  prow = sbuf1;                                                          \
  sbuf1 = sbuf2; sbuf2 = sbuf3;                                          \
  sbuf3 = prow;                                                          \
  s1 = (mlib_d64*)sbuf1; s2 = (mlib_d64*)sbuf2;                          \
  s3 = (mlib_d64*)sbuf3;
/************************************************************************/
#define PREPARE_TO_COPY_INTERM_BUF_TO_DST()                              \
  ddst = dbuf;                                                           \
  /* prepare the destination addresses */                                \
  dp = (mlib_d64 *)((mlib_addr) da & (~7));                              \
  i = (mlib_addr) dp - (mlib_addr) da;                                   \
  ddst = vis_alignaddr(ddst, i);                                         \
  /* generate edge mask for the start point */                           \
  emask = vis_edge8(da, dend);                                           \
  sd1 = ddst[0];                                                         \
  if (emask != 0xff) {                                                   \
    sd0 = sd1;                                                           \
    sd1 = ddst[1];                                                       \
    sd0 = vis_faligndata(sd0, sd1);                                      \
    vis_pst_8(sd0, dp++, emask);                                         \
    ddst++;                                                              \
    i += 8;                                                              \
  }
/************************************************************************/
#define COPY_INTERM_BUF_TO_DST()                                         \
  for (; i <= (dw - 8); i += 8) {                                        \
    sd0 = sd1;                                                           \
    sd1 = ddst[1];                                                       \
    *dp++ = vis_faligndata(sd0, sd1);                                    \
    ddst++;                                                              \
  }
/************************************************************************/
#define COPY_TAIL()                                                      \
  if (i < dw)  {                                                         \
    sd0 = vis_faligndata(sd1, ddst[1]);                                  \
    emask = vis_edge8(dp, dend);                                         \
    vis_pst_8(sd0, dp, emask);                                           \
  }
/************************************************************************/
#define CONV_AU(dsrc, ka)                                                \
  tmp0 = vis_fmul8x16au(vis_read_hi(dsrc), ka);                          \
  tmp1 = vis_fmul8x16au(vis_read_lo(dsrc), ka);                          \
  out0 = vis_fpadd16(out0, tmp0);                                        \
  out1 = vis_fpadd16(out1, tmp1);
/************************************************************************/
#define CONV_AL(dsrc, ka)                                                \
  tmp0 = vis_fmul8x16al(vis_read_hi(dsrc), ka);                          \
  tmp1 = vis_fmul8x16al(vis_read_lo(dsrc), ka);                          \
  out0 = vis_fpadd16(out0, tmp0);                                        \
  out1 = vis_fpadd16(out1, tmp1);
/************************************************************************/
#define SHIFT_U8                                                         \
  d21 = vis_faligndata(d21, dt_1);                                       \
  dt_1 = vis_faligndata(dt_1, d1);                                       \
  d22 = vis_faligndata(d22, dt_2);                                       \
  dt_2 = vis_faligndata(dt_2, d2);                                       \
  d23 = vis_faligndata(d23, dt_3);                                       \
  dt_3 = vis_faligndata(dt_3, d3);                                       \
  d24 = vis_faligndata(d24, dt_4);                                       \
  dt_4 = vis_faligndata(dt_4, d4);
/************************************************************************/

mlib_status mlib_v_conv3x3_8nw_1( mlib_image *dst,         /* source  image */
                                  mlib_image *src,         /* destination image */
                                  mlib_s32   *kernel,      /* convolution kernel */
                                  mlib_s32   scalef_expon) /* scalefactor exponent */

{
  mlib_u8   *da, *d_a;                                          /* pointers to dst row */
  mlib_u8   *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u8   *sa, *sa1, *sa2;                                    /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow;                   /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3;                                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d_1, d_2, d21, d22;                         /* data */
  mlib_d64  d3, d_3, d23;                                       /* data */
  mlib_d64  dt_1, dt_2, dt_3;                                   /* data */
  mlib_f32  k1k2, k3k4, k5k6, k7k8, k9k9;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1;
  mlib_d64  tmp0, tmp1, rnd;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       rval, gsr_scale, i, j;

  gsr_scale = 31 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));
  rval = mlib_round_8[gsr_scale];
  rnd = vis_freg_pair(vis_to_float(rval), vis_to_float(rval));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (dw+16)>>3;
  PREPARE_INTERM_BUFFERS()

  dw -= 2;
  dh -= 2;

  sa  = adr_src;
  sa1 = sa + slb;
  sa2 = sa1 + slb;
  d_a = adr_dst + dlb + 1;

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf2, sa)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(2)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf3, sa1)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(2)

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    PREPARE_TO_LOAD_LINE(sbuf3, sa2)
#pragma pipeloop(0)
    LOAD_LINE_INTO_BUFFER(2)

    vis_alignaddr(s1, 1);
    d1 = *s1; d2 = *s2; d3 = *s3;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 8) {
      d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
      out0 = out1 = rnd;
      CONV_AU(d1, k1k2)
      CONV_AL(d2, k3k4)
      CONV_AU(d3, k7k8)
      d21 = vis_faligndata(d1, d_1);
      dt_1 = vis_faligndata(d_1, d1);
      d22 = vis_faligndata(d2, d_2);
      dt_2 = vis_faligndata(d_2, d2);
      d23 = vis_faligndata(d3, d_3);
      dt_3 = vis_faligndata(d_3, d3);
      CONV_AL(d21, k1k2)
      CONV_AU(d22, k5k6)
      CONV_AL(d23, k7k8)
      d21 = vis_faligndata(d21, dt_1);
      d22 = vis_faligndata(d22, dt_2);
      d23 = vis_faligndata(d23, dt_3);
      CONV_AU(d21, k3k4)
      CONV_AL(d22, k5k6)
      CONV_AU(d23, k9k9)
      *ddst++ = vis_fpack16_pair(out0, out1);
      d1 = d_1; d2 = d_2; d3 = d_3;
      s1++; s2++; s3++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    sa2 = sa2 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv3x3_8nw_2( mlib_image *dst,         /* source  image */
                                  mlib_image *src,         /* destination image */
                                  mlib_s32   *kernel,      /* convolution kernel */
                                  mlib_s32   scalef_expon) /* scalefactor exponent */
{
  mlib_u8   *da, *d_a;                                          /* pointers to dst row */
  mlib_u8   *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u8   *sa, *sa1, *sa2;                                    /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow;                   /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3;                                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d_1, d_2, d21, d22;                         /* data */
  mlib_d64  d3, d_3, d23;                                       /* data */
  mlib_d64  dt_1, dt_2, dt_3;                                   /* data */
  mlib_f32  k1k2, k3k4, k5k6, k7k8, k9k9;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1;
  mlib_d64  tmp0, tmp1, rnd;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       rval, gsr_scale, i, j;

  gsr_scale = 31 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));
  rval = mlib_round_8[gsr_scale];
  rnd = vis_freg_pair(vis_to_float(rval), vis_to_float(rval));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (2*dw+16)>>3;
  PREPARE_INTERM_BUFFERS()

  dw -= 2;
  dw <<= 1;
  dh -= 2;

  sa  = adr_src;
  sa1 = sa + slb;
  sa2 = sa1 + slb;
  d_a = adr_dst + dlb + 2;

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf2, sa)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(4)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf3, sa1)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(4)

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    PREPARE_TO_LOAD_LINE(sbuf3, sa2)
#pragma pipeloop(0)
    LOAD_LINE_INTO_BUFFER(4)

    vis_alignaddr(s1, 2);
    d1 = *s1; d2 = *s2; d3 = *s3;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 8) {
      d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
      out0 = out1 = rnd;
      CONV_AU(d1, k1k2)
      CONV_AL(d2, k3k4)
      CONV_AU(d3, k7k8)
      d21 = vis_faligndata(d1, d_1);
      dt_1 = vis_faligndata(d_1, d1);
      d22 = vis_faligndata(d2, d_2);
      dt_2 = vis_faligndata(d_2, d2);
      d23 = vis_faligndata(d3, d_3);
      dt_3 = vis_faligndata(d_3, d3);
      CONV_AL(d21, k1k2)
      CONV_AU(d22, k5k6)
      CONV_AL(d23, k7k8)
      d21 = vis_faligndata(d21, dt_1);
      d22 = vis_faligndata(d22, dt_2);
      d23 = vis_faligndata(d23, dt_3);
      CONV_AU(d21, k3k4)
      CONV_AL(d22, k5k6)
      CONV_AU(d23, k9k9)
      *ddst++ = vis_fpack16_pair(out0, out1);
      d1 = d_1; d2 = d_2; d3 = d_3;
      s1++; s2++; s3++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    sa2 = sa2 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv3x3_8nw_3( mlib_image *dst,         /* source  image */
                                  mlib_image *src,         /* destination image */
                                  mlib_s32   *kernel,      /* convolution kernel */
                                  mlib_s32   scalef_expon) /* scalefactor exponent */
{
  mlib_u8   *da, *d_a;                                          /* pointers to dst row */
  mlib_u8   *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u8   *sa, *sa1, *sa2;                                    /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow;                   /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3;                                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d_1, d_2, d21, d22;                         /* data */
  mlib_d64  d3, d_3, d23;                                       /* data */
  mlib_d64  dt_1, dt_2, dt_3;                                   /* data */
  mlib_f32  k1k2, k3k4, k5k6, k7k8, k9k9;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1;
  mlib_d64  tmp0, tmp1, rnd;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       rval, gsr_scale, i, j;

  gsr_scale = 31 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));
  rval = mlib_round_8[gsr_scale];
  rnd = vis_freg_pair(vis_to_float(rval), vis_to_float(rval));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (3*dw+24)>>3;
  PREPARE_INTERM_BUFFERS()

  dw -= 2;
  dw *= 3;
  dh -= 2;

  sa  = adr_src;
  sa1 = sa + slb;
  sa2 = sa1 + slb;
  d_a = adr_dst + dlb + 3;

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf2, sa)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(6)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf3, sa1)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(6)

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    PREPARE_TO_LOAD_LINE(sbuf3, sa2)
#pragma pipeloop(0)
    LOAD_LINE_INTO_BUFFER(6)

    vis_alignaddr(s1, 3);
    d1 = *s1; d2 = *s2; d3 = *s3;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 8) {
      d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
      out0 = out1 = rnd;
      CONV_AU(d1, k1k2)
      CONV_AL(d2, k3k4)
      CONV_AU(d3, k7k8)
      d21 = vis_faligndata(d1, d_1);
      dt_1 = vis_faligndata(d_1, d1);
      d22 = vis_faligndata(d2, d_2);
      dt_2 = vis_faligndata(d_2, d2);
      d23 = vis_faligndata(d3, d_3);
      dt_3 = vis_faligndata(d_3, d3);
      CONV_AL(d21, k1k2)
      CONV_AU(d22, k5k6)
      CONV_AL(d23, k7k8)
      d21 = vis_faligndata(d21, dt_1);
      d22 = vis_faligndata(d22, dt_2);
      d23 = vis_faligndata(d23, dt_3);
      CONV_AU(d21, k3k4)
      CONV_AL(d22, k5k6)
      CONV_AU(d23, k9k9)
      *ddst++ = vis_fpack16_pair(out0, out1);
      d1 = d_1; d2 = d_2; d3 = d_3;
      s1++; s2++; s3++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    sa2 = sa2 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv3x3_8nw_4( mlib_image *dst,         /* source  image */
                                  mlib_image *src,         /* destination image */
                                  mlib_s32   *kernel,      /* convolution kernel */
                                  mlib_s32   scalef_expon, /* scalefactor exponent */
                                  mlib_s32   cmask)        /* mask */
{
  mlib_u8   *da, *d_a;                                          /* pointers to dst row */
  mlib_u8   *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u8   *sa, *sa1, *sa2;                                    /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow;                   /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3;                                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d_1, d_2, d21, d22;                         /* data */
  mlib_d64  d3, d_3, d23;                                       /* data */
  mlib_f32  k1k2, k3k4, k5k6, k7k8, k9k9;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1;
  mlib_d64  tmp0, tmp1, rnd;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1, sd00;
  mlib_s32  emask, cmask1;
  int       rval, gsr_scale, i, j;

  gsr_scale = 31 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));
  rval = mlib_round_8[gsr_scale];
  rnd = vis_freg_pair(vis_to_float(rval), vis_to_float(rval));

  cmask = ((cmask & 0xf) << 4) + (cmask & 0xf);
  cmask = (cmask << 8) + (cmask);

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (4*dw+24)>>3;
  PREPARE_INTERM_BUFFERS()

  dw -= 2;
  dw *= 4;
  dh -= 2;

  sa  = adr_src;
  sa1 = sa + slb;
  sa2 = sa1 + slb;
  d_a = adr_dst + dlb + 4;

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf2, sa)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(8)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf3, sa1)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(8)

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    PREPARE_TO_LOAD_LINE(sbuf3, sa2)
#pragma pipeloop(0)
    LOAD_LINE_INTO_BUFFER(8)

    vis_alignaddr(s1, 4);
    d1 = *s1; d2 = *s2; d3 = *s3;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 8) {
      d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
      out0 = out1 = rnd;
      CONV_AU(d1, k1k2)
      CONV_AL(d2, k3k4)
      CONV_AU(d3, k7k8)
      d21 = vis_faligndata(d1, d_1);
      d22 = vis_faligndata(d2, d_2);
      d23 = vis_faligndata(d3, d_3);
      CONV_AL(d21, k1k2)
      CONV_AU(d22, k5k6)
      CONV_AL(d23, k7k8)
      CONV_AU(d_1, k3k4)
      CONV_AL(d_2, k5k6)
      CONV_AU(d_3, k9k9)
      *ddst++ = vis_fpack16_pair(out0, out1);
      d1 = d_1; d2 = d_2; d3 = d_3;
      s1++; s2++; s3++;
    }

    ddst = dbuf;
    /* prepare the destination addresses */
    dp = (mlib_d64 *)((mlib_addr) da & (~7));
    i = (mlib_addr) dp - (mlib_addr) da;
    cmask1 = cmask >> (-i);
    ddst = vis_alignaddr(ddst, i);
    /* generate edge mask for the start point */
    emask = vis_edge8(da, dend);
    sd1 = ddst[0];
    if (emask != 0xff) {
      sd0 = sd1;
      sd1 = ddst[1];
      sd0 = vis_faligndata(sd0, sd1);
      vis_pst_8(sd0, dp++, emask & cmask1);
      ddst++;
      i += 8;
    }

#pragma pipeloop(0)
    for (; i <= (dw - 8); i += 8) {
      sd0 = sd1;
      sd1 = ddst[1];
      sd00 = vis_faligndata(sd0, sd1);
      vis_pst_8(sd00, dp++, cmask1);
      ddst++;
    }

    if (i < dw)  {
      sd0 = vis_faligndata(sd1, ddst[1]);
      emask = vis_edge8(dp, dend);
      vis_pst_8(sd0, dp, emask & cmask1);
    }

    sa2 = sa2 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv3x3_8nw_mask( mlib_image *dst,         /* source  image */
                                     mlib_image *src,         /* destination image */
                                     mlib_s32   *kernel,      /* convolution kernel */
                                     mlib_s32   scalef_expon, /* scalefactor exponent */
                                     mlib_s32   cmask)        /* mask */
{
    mlib_u8   *da, *d_a;                                          /* pointers to dst row */
    mlib_u8   *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
    mlib_u8   *sa, *sa2, *sa3, *sa4, *sa_4;                       /* pointers to src rows */
    mlib_u8   *buff_src, *sbuf1, *sbuf2, *sbuf3, *s_buf1, *prow;  /* pointers to rows in interm. src buf */
    mlib_u8   *dbuf;                                              /* pointer to row in interm. dst buf */
    mlib_d64  *s1, *s2, *s3;                                      /* mlib_d64 pointers to rows in interm. src buf */
    mlib_d64  *ddst;                                              /* mlib_d64 pointer to row in interm. dst buf */
    mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
    mlib_s32  dh, dw;
    mlib_d64  out0, out1, tmp0, tmp1, rnd, sd;
    mlib_d64  d1, d2, d3, d_1, d_2, d_3;                          /* data */
    mlib_d64  dt_1, dt_2, dt_3;                                   /* temp. data, used in falligndata */
    mlib_d64  d21, d22, d23, d31, d32, d33;                       /* shifted data */
    mlib_f32  k12, k34, k56, k78, k9;                             /* koefficients */
    int       rval, gsr_scale, i, j, nchannel, chan, testchan;

    adr_src = mlib_ImageGetData(src);
    adr_dst = mlib_ImageGetData(dst);
    nchannel = mlib_ImageGetChannels(src);
    slb = mlib_ImageGetStride(src);
    dlb = mlib_ImageGetStride(dst);
    dh = mlib_ImageGetHeight(dst);
    dw = mlib_ImageGetWidth(dst);

    buf_slb = (dw+16) & (~7);                                              /* buf_slb - 8_byte alligned */
    buff_src = (mlib_u8*) mlib_malloc(4 * buf_slb * sizeof(mlib_u8) + 8);  /* alloc. interm. src and dst buffer */
    if (buff_src == NULL) return MLIB_FAILURE;
    dw -= 2;
    dh -= 2;                                                               /* edge - no write */

    /*
     *  The 8x16 mult has built-in 8-bit R shift, and fpack16 has 7-bit
     *  fixed R shift (preceded by variable-bit L shift controlled by GSR
     *  scalefactor field). Thus net R shift = (8+7)-(GSR.scalefactor_field),
     *  so GSR.scalefactor_field = 15-(net R shift):
     */
    gsr_scale = 31 - scalef_expon;
    vis_write_gsr((gsr_scale << 3)+1);
    rval = mlib_round_8[gsr_scale];
    rnd = vis_freg_pair(vis_to_float(rval), vis_to_float(rval));
    sbuf1 = (mlib_u8*) ((mlib_addr)(buff_src +8) & (~7));
    sbuf2 = sbuf1 + buf_slb;
    sbuf3 = sbuf2 + buf_slb;
    dbuf = sbuf3 + buf_slb;

    /* load kernel values into floats */

    k12 = vis_to_float((kernel[0] & 0xFFFF0000)|((kernel[1] >> 16) & 0xFFFF));
    k34 = vis_to_float((kernel[2] & 0xFFFF0000)|((kernel[3] >> 16) & 0xFFFF));
    k56 = vis_to_float((kernel[4] & 0xFFFF0000)|((kernel[5] >> 16) & 0xFFFF));
    k78 = vis_to_float((kernel[6] & 0xFFFF0000)|((kernel[7] >> 16) & 0xFFFF));
    k9  = vis_to_float((kernel[8] & 0xFFFF0000)|((kernel[8] >> 16) & 0xFFFF));

    testchan = 1;

    for ( chan = nchannel - 1; chan >= 0; chan--) {
      if ((cmask & testchan) == 0) {
        testchan <<= 1;
        continue;
      }
      testchan <<= 1;
      sa  = adr_src + chan;
      sa2 = sa + slb;
      sa3 = sa2 + slb;
      sa_4 = sa4 = sa3 + slb;
      d_a = adr_dst + dlb + nchannel + chan;

      /* load interm. src buff */
      for (i = 0, j = 0; j < (dw + 2); i+=nchannel, j++) {
        sbuf1[j] = sa3[i];  sbuf2[j] = sa[i];  sbuf3[j] = sa2[i];
      }

      for (j = 0; j < dh - 1; j++)  {
        ddst = (mlib_d64*)(dbuf);
        da = d_a;
        dend = da + (dw - 1)*nchannel;
        prow = sbuf1;
        sbuf1 = sbuf2; sbuf2 = sbuf3; sbuf3 = prow;
        s1 = (mlib_d64*)sbuf1; s2 = (mlib_d64*)sbuf2; s3 = (mlib_d64*)sbuf3;
        s_buf1 = sbuf1;
        d1 = *s1; d2 = *s2; d3 = *s3 ;

#pragma pipeloop(0)
        /* in each iteraction load data for processing next row */
        for (i = 0; i < dw; i += 8) {
          d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
          out0 = out1 = rnd;
          CONV_AU(d1, k12)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          CONV_AL(d2, k34)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          CONV_AU(d3, k78)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          d22 = vis_faligndata(d2, d_2);
          dt_2 = vis_faligndata(d_2, d2);
          d21 = vis_faligndata(d1, d_1);
          dt_1 = vis_faligndata(d_1, d1);
          d23 = vis_faligndata(d3, d_3);
          dt_3 = vis_faligndata(d_3, d3);
          CONV_AL(d21, k12)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          CONV_AU(d22, k56)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          CONV_AL(d23, k78)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          d31 = vis_faligndata(d21, dt_1);
          d32 = vis_faligndata(d22, dt_2);
          d33 = vis_faligndata(d23, dt_3);
          CONV_AU(d31, k34)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          CONV_AL(d32, k56)
          *s_buf1++ = *sa_4; sa_4 += nchannel;
          CONV_AU(d33, k9)
          *ddst++ = vis_freg_pair(vis_fpack16(out0), vis_fpack16(out1));
          d1 = d_1; d2 = d_2; d3 = d_3;
          s1++; s2++; s3++;
        }

        *s_buf1++ = *sa_4; sa_4 += nchannel; *s_buf1++ = *sa_4; sa_4 += nchannel;
        *s_buf1++ = *sa_4; sa_4 += nchannel; *s_buf1++ = *sa_4; sa_4 += nchannel;
        *s_buf1++ = *sa_4; sa_4 += nchannel; *s_buf1++ = *sa_4; sa_4 += nchannel;
        *s_buf1++ = *sa_4; sa_4 += nchannel; *s_buf1++ = *sa_4;

         ddst = (mlib_d64*)(dbuf);
#pragma pipeloop(0)
        for (i = 0; i <= (dw - 8);  i+= 8) {
          LOAD_INSERT_STORE_U8(nchannel);
        }
        if ((mlib_addr) da <= (mlib_addr) dend) {
          sd = *ddst++;
          vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
          if ((mlib_addr) da <= (mlib_addr) dend) {
            vis_st_u8(sd = vis_faligndata(sd, sd), da);da += nchannel;
            if ((mlib_addr) da <= (mlib_addr) dend) {
              vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
              if ((mlib_addr) da <= (mlib_addr) dend) {
                vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
                if ((mlib_addr) da <= (mlib_addr) dend) {
                  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
                  if ((mlib_addr) da <= (mlib_addr) dend) {
                    vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
                    if ((mlib_addr) da <= (mlib_addr) dend) {
                      vis_st_u8(sd = vis_faligndata(sd, sd), da);
                    }
                  }
                }
              }
            }
          }
       }
        sa_4 = sa4 = sa4 + slb;
        d_a += dlb;
      }

      /* process last row - no need to load data */
      ddst = (mlib_d64*)(dbuf);
      da = d_a;
      dend = da + (dw - 1)*nchannel;
      prow = sbuf1;
      sbuf1 = sbuf2; sbuf2 = sbuf3; sbuf3 = prow;
      s1 = (mlib_d64*)sbuf1; s2 = (mlib_d64*)sbuf2; s3 = (mlib_d64*)sbuf3;
      d1 = *s1; d2 = *s2; d3 = *s3 ;

#pragma pipeloop(0)
      for (i = 0; i < dw; i += 8) {
        d_1 = *(s1+1); d_2 = *(s2+1); d_3 = *(s3+1);
        out0 = out1 = rnd;
        CONV_AU(d1, k12)
        CONV_AL(d2, k34)
        CONV_AU(d3, k78)
        d22 = vis_faligndata(d2, d_2);
        dt_2 = vis_faligndata(d_2, d2);
        d21 = vis_faligndata(d1, d_1);
        dt_1 = vis_faligndata(d_1, d1);
        d23 = vis_faligndata(d3, d_3);
        dt_3 = vis_faligndata(d_3, d3);
        CONV_AL(d21, k12)
        CONV_AU(d22, k56)
        CONV_AL(d23, k78)
        d31 = vis_faligndata(d21, dt_1);
        d32 = vis_faligndata(d22, dt_2);
        d33 = vis_faligndata(d23, dt_3);
        CONV_AU(d31, k34)
        CONV_AL(d32, k56)
        CONV_AU(d33, k9)
        *ddst++ = vis_freg_pair(vis_fpack16(out0), vis_fpack16(out1));
        d1 = d_1; d2 = d_2; d3 = d_3;
        s1++; s2++; s3++;
      }

      ddst = (mlib_d64*)(dbuf);
#pragma pipeloop(0)
      for (i = 0; i <= (dw - 8);  i+= 8) {
        LOAD_INSERT_STORE_U8(nchannel);
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        sd = *ddst++;
        vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
        if ((mlib_addr) da <= (mlib_addr) dend) {
          vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
          if ((mlib_addr) da <= (mlib_addr) dend) {
            vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
            if ((mlib_addr) da <= (mlib_addr) dend) {
              vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
              if ((mlib_addr) da <= (mlib_addr) dend) {
                vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
                if ((mlib_addr) da <= (mlib_addr) dend) {
                  vis_st_u8(sd = vis_faligndata(sd, sd), da); da += nchannel;
                  if ((mlib_addr) da <= (mlib_addr) dend) {
                    vis_st_u8(sd = vis_faligndata(sd, sd), da);
                  }
                }
              }
            }
          }
        }
      }
    }
    mlib_free(buff_src);
    return MLIB_SUCCESS ;
}

/************************************************************************/

mlib_status mlib_v_conv3x3_8nw( mlib_image *dst,         /* source  image */
                                mlib_image *src,         /* destination image */
                                mlib_s32 *kernel,        /* convolution kernel */
                                mlib_s32   scalef_expon, /* scalefactor exponent */
                                mlib_s32   cmask)        /* mask */

{
  mlib_s32 nchannel, chan, cmask1, i;
  mlib_status res;

  cmask1 = cmask;
  nchannel = mlib_ImageGetChannels(src);
  chan = 0;
  for(i = nchannel - 1; i >= 0; i--, cmask1 >>= 1)
    if ((cmask1 & 1) == 1) chan++;

  if ((nchannel == chan) || ((nchannel == 4) && (chan == 3))) {
    switch (nchannel) {
      case 1:
        res = mlib_v_conv3x3_8nw_1(dst, src, kernel, scalef_expon);
        break;
      case 2:
        res = mlib_v_conv3x3_8nw_2(dst, src, kernel, scalef_expon);
        break;
      case 3:
        res = mlib_v_conv3x3_8nw_3(dst, src, kernel, scalef_expon);
        break;
      case 4:
        res = mlib_v_conv3x3_8nw_4(dst, src, kernel, scalef_expon, cmask);
        break;
      default: res = MLIB_FAILURE; break;
    }
  } else {
    res = mlib_v_conv3x3_8nw_mask(dst, src, kernel, scalef_expon, cmask);
  }
  return ( res );
}





