/*
 * @(#)mlib_v_ImageConv5x5_16nw.c	1.4 98/09/06
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#pragma ident	"@(#)mlib_v_ImageConv5x5_16nw.c	1.12	98/09/02 SMI"

/*
 * FUNCTION
 *      mlib_conv5x5_16nw.c - convolve an 16-bit image with a 5x5 kernel
 *                      (edge = no write)
 *
 *
 * SYNOPSIS
 *  mlib_status
 *  mlib_conv5x5_16nw( mlib_image *src,
 *                     mlib_image *dst,
 *                     mlib_s32 *kernel,
 *                     mlib_s32 scalef_expon
 *                     mlib_s32 cmask )
 * ARGUMENT
 *      src           Pointer to structure of source image
 *      dst           Pointer to structure of destination image
 *      kernel        Convolution kernel
 *      scalef_expons The scaling factor to convert the input integer
 *                    coefficients into floating-point coefficients:
 *                    floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask         Channel mask to indicate the channels to be convolved.
 *                    Each bit of which represents a channel in the image. The
 *                    channels corresponded to 1 bits are those to be processed.
 *
 * DESCRIPTION
 *      A 2-D convolution (5x5 kernel) for 16-bit images.
 */

#include <stdlib.h>
#include <mlib_image.h>
#include <vis_proto.h>

/************************************************************************/
#define GET_SRC_DST_PARAMETERS()                                         \
  adr_src = (mlib_u16*)mlib_ImageGetData(src);                           \
  adr_dst = (mlib_u16*)mlib_ImageGetData(dst);                           \
  slb = mlib_ImageGetStride(src) >> 1;                                   \
  dlb = mlib_ImageGetStride(dst) >> 1;                                   \
  dh = mlib_ImageGetHeight(dst);                                         \
  dw = mlib_ImageGetWidth(dst);
/************************************************************************/
#define LOAD_KERNEL_INTO_FLOAT()                                         \
  k1 = vis_to_float((kernel[0] & 0xFFFF0000) |                           \
                   ((kernel[0] >> 16) & 0xFFFF));                        \
  k2 = vis_to_float((kernel[1] & 0xFFFF0000) |                           \
                   ((kernel[1] >> 16) & 0xFFFF));                        \
  k3 = vis_to_float((kernel[2] & 0xFFFF0000) |                           \
                   ((kernel[2] >> 16) & 0xFFFF));                        \
  k4 = vis_to_float((kernel[3] & 0xFFFF0000) |                           \
                   ((kernel[3] >> 16) & 0xFFFF));                        \
  k5 = vis_to_float((kernel[4] & 0xFFFF0000) |                           \
                   ((kernel[4] >> 16) & 0xFFFF));                        \
  k6 = vis_to_float((kernel[5] & 0xFFFF0000) |                           \
                   ((kernel[5] >> 16) & 0xFFFF));                        \
  k7 = vis_to_float((kernel[6] & 0xFFFF0000) |                           \
                   ((kernel[6] >> 16) & 0xFFFF));                        \
  k8 = vis_to_float((kernel[7] & 0xFFFF0000) |                           \
                   ((kernel[7] >> 16) & 0xFFFF));                        \
  k9 = vis_to_float((kernel[8] & 0xFFFF0000) |                           \
                   ((kernel[8] >> 16) & 0xFFFF));                        \
  k10 = vis_to_float((kernel[9] & 0xFFFF0000) |                          \
                    ((kernel[9] >> 16) & 0xFFFF));                       \
  k11 = vis_to_float((kernel[10] & 0xFFFF0000) |                         \
                    ((kernel[10] >> 16) & 0xFFFF));                      \
  k12 = vis_to_float((kernel[11] & 0xFFFF0000) |                         \
                    ((kernel[11] >> 16) & 0xFFFF));                      \
  k13 = vis_to_float((kernel[12] & 0xFFFF0000) |                         \
                    ((kernel[12] >> 16) & 0xFFFF));                      \
  k14 = vis_to_float((kernel[13] & 0xFFFF0000) |                         \
                    ((kernel[13] >> 16) & 0xFFFF));                      \
  k15 = vis_to_float((kernel[14] & 0xFFFF0000) |                         \
                    ((kernel[14] >> 16) & 0xFFFF));                      \
  k16 = vis_to_float((kernel[15] & 0xFFFF0000) |                         \
                    ((kernel[15] >> 16) & 0xFFFF));                      \
  k17 = vis_to_float((kernel[16] & 0xFFFF0000) |                         \
                    ((kernel[16] >> 16) & 0xFFFF));                      \
  k18 = vis_to_float((kernel[17] & 0xFFFF0000) |                         \
                    ((kernel[17] >> 16) & 0xFFFF));                      \
  k19 = vis_to_float((kernel[18] & 0xFFFF0000) |                         \
                    ((kernel[18] >> 16) & 0xFFFF));                      \
  k20 = vis_to_float((kernel[19] & 0xFFFF0000) |                         \
                    ((kernel[19] >> 16) & 0xFFFF));                      \
  k21 = vis_to_float((kernel[20] & 0xFFFF0000) |                         \
                    ((kernel[20] >> 16) & 0xFFFF));                      \
  k22 = vis_to_float((kernel[21] & 0xFFFF0000) |                         \
                    ((kernel[21] >> 16) & 0xFFFF));                      \
  k23 = vis_to_float((kernel[22] & 0xFFFF0000) |                         \
                    ((kernel[22] >> 16) & 0xFFFF));                      \
  k24 = vis_to_float((kernel[23] & 0xFFFF0000) |                         \
                    ((kernel[23] >> 16) & 0xFFFF));                      \
  k25 = vis_to_float((kernel[24] & 0xFFFF0000) |                         \
                    ((kernel[24] >> 16) & 0xFFFF));
/************************************************************************/
#define PREPARE_INTERM_BUFFERS()                                         \
  buff_src = (mlib_d64*) mlib_malloc(7 * buf_slb * sizeof(mlib_d64));    \
  if (buff_src == NULL) return MLIB_FAILURE;                             \
                                                                         \
  sbuf1 = buff_src;                                                      \
  sbuf2 = sbuf1 + buf_slb;                                               \
  sbuf3 = sbuf2 + buf_slb;                                               \
  sbuf4 = sbuf3 + buf_slb;                                               \
  sbuf5 = sbuf4 + buf_slb;                                               \
  dbuf = sbuf5 + buf_slb;
/************************************************************************/
#define PREPARE_TO_LOAD_LINE(p_buf, p_line)                              \
  prow = p_buf;                                                          \
  dsa = (mlib_d64 *) vis_alignaddr(p_line, 0);                           \
  sd1 = dsa[0]; dsa++;
/************************************************************************/
#define LOAD_LINE_INTO_BUFFER(n)                                         \
  for (i = 0; i < (dw + n); i+= 4) {                                     \
    sd0 = sd1; sd1 = dsa[0];                                             \
    *prow++ = vis_faligndata(sd0, sd1);                                  \
    dsa++;                                                               \
  }
/************************************************************************/
#define LOOP_INI()                                                       \
  ddst = (mlib_d64*)dbuf;                                                \
  da = d_a;                                                              \
  dend = da + (dw - 1);                                                  \
  prow = sbuf1;                                                          \
  sbuf1 = sbuf2; sbuf2 = sbuf3;                                          \
  sbuf3 = sbuf4; sbuf4 = sbuf5; sbuf5 = prow;                            \
  s1 = (mlib_d64*)sbuf1; s2 = (mlib_d64*)sbuf2;                          \
  s3 = (mlib_d64*)sbuf3; s4 = (mlib_d64*)sbuf4;                          \
  s5 = (mlib_d64*)sbuf5;
/************************************************************************/
#define PREPARE_TO_COPY_INTERM_BUF_TO_DST()                              \
  ddst = dbuf;                                                           \
  /* prepare the destination addresses */                                \
  dp = (mlib_d64 *)((mlib_addr) da & (~7));                              \
  i = ((mlib_u16*) dp - (mlib_u16*) da);                                 \
  ddst = vis_alignaddr(ddst, 2*i);                                       \
  /* generate edge mask for the start point */                           \
  emask = vis_edge16(da, dend);                                          \
  sd1 = ddst[0];                                                         \
  if (emask != 0xf) {                                                    \
    sd0 = sd1;                                                           \
    sd1 = ddst[1];                                                       \
    sd0 = vis_faligndata(sd0, sd1);                                      \
    vis_pst_16(sd0, dp++, emask);                                        \
    ddst++;                                                              \
    i += 4;                                                              \
  }
/************************************************************************/
#define COPY_INTERM_BUF_TO_DST()                                         \
  for (; i <= (dw - 4); i += 4) {                                        \
    sd0 = sd1;                                                           \
    sd1 = ddst[1];                                                       \
    *dp++ = vis_faligndata(sd0, sd1);                                    \
    ddst++;                                                              \
  }
/************************************************************************/
#define COPY_TAIL()                                                      \
  if (i < dw)  {                                                         \
    sd0 = vis_faligndata(sd1, ddst[1]);                                  \
    emask = vis_edge16(dp, dend);                                        \
    vis_pst_16(sd0, dp, emask);                                          \
  }
/************************************************************************/
#define CONV_16_BEGIN(dsrc, ka)                                          \
  tmp0 = vis_fmuld8sux16(vis_read_hi(dsrc), ka);                         \
  tmp1 = vis_fmuld8ulx16(vis_read_hi(dsrc), ka);                         \
  tmp2 = vis_fmuld8sux16(vis_read_lo(dsrc), ka);                         \
  tmp3 = vis_fmuld8ulx16(vis_read_lo(dsrc), ka);                         \
  out0 = vis_fpadd32(tmp0, tmp1);                                        \
  out1 = vis_fpadd32(tmp2, tmp3);
/************************************************************************/
#define CONV_16(dsrc, ka)                                                \
  tmp0 = vis_fmuld8sux16(vis_read_hi(dsrc), ka);                         \
  tmp1 = vis_fmuld8ulx16(vis_read_hi(dsrc), ka);                         \
  tmp2 = vis_fmuld8sux16(vis_read_lo(dsrc), ka);                         \
  tmp3 = vis_fmuld8ulx16(vis_read_lo(dsrc), ka);                         \
  tmp0 = vis_fpadd32(tmp0, tmp1);                                        \
  out0 = vis_fpadd32(out0, tmp0);                                        \
  tmp2 = vis_fpadd32(tmp2, tmp3);                                        \
  out1 = vis_fpadd32(out1, tmp2);
/************************************************************************/
#define CONV_16_MIX(dsrc1, dsrc2, ka)                                    \
  tmp0 = vis_fmuld8sux16(vis_read_lo(dsrc1), ka);                        \
  tmp1 = vis_fmuld8ulx16(vis_read_lo(dsrc1), ka);                        \
  tmp2 = vis_fmuld8sux16(vis_read_hi(dsrc2), ka);                        \
  tmp3 = vis_fmuld8ulx16(vis_read_hi(dsrc2), ka);                        \
  tmp0 = vis_fpadd32(tmp0, tmp1);                                        \
  out0 = vis_fpadd32(out0, tmp0);                                        \
  tmp2 = vis_fpadd32(tmp2, tmp3);                                        \
  out1 = vis_fpadd32(out1, tmp2);
/************************************************************************/

mlib_status mlib_v_conv5x5_16nw_1( mlib_image *dst,         /* source  image */
                                   mlib_image *src,         /* destination image */
                                   mlib_s32   *kernel,      /* convolution kernel */
                                   mlib_s32   scalef_expon) /* scalefactor exponent */

{
  mlib_u16  *da, *d_a;                                          /* pointers to dst row */
  mlib_u16  *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u16  *sa, *sa1, *sa2, *sa3, *sa4;                        /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow;                   /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3, *sbuf4, *sbuf5;                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3, *s4, *s5;                            /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst, *ddst1;                                      /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d_1, d_2, d21, d22;                         /* data */
  mlib_d64  d3, d4, d_3, d_4, d23, d24;                         /* data */
  mlib_d64  d5, d_5, d25;                                       /* data */
  mlib_d64  dt_1, dt_2, dt_3, dt_4, dt_5;                       /* data */
  mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8;
  mlib_f32  k9, k10, k11, k12, k13, k14, k15, k16;
  mlib_f32  k17, k18, k19, k20, k21, k22, k23, k24, k25;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       gsr_scale, i, j;

  gsr_scale = 32 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (2*dw+26)>>3;
  PREPARE_INTERM_BUFFERS()

  dw -= 4;
  dh -= 4;

  sa  = adr_src;
  sa1 = sa + slb;
  sa2 = sa1 + slb;
  sa3 = sa2 + slb;
  sa4 = sa3 + slb;
  d_a = adr_dst + 2 + 2*dlb;

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf2, sa)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(4)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf3, sa1)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(4)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf4, sa2)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(4)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf5, sa3)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(4)

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    PREPARE_TO_LOAD_LINE(sbuf5, sa4)
#pragma pipeloop(0)
    LOAD_LINE_INTO_BUFFER(4)

    vis_alignaddr(s1, 2);
    d1 = *s1; d2 = *s2;
    ddst1 = (mlib_d64*)dbuf;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d_1 = *(s1+1); d_2 = *(s2+1);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k6)
      CONV_16_MIX(d1, d_1, k3)
      CONV_16_MIX(d2, d_2, k8)
      d21 = vis_faligndata(d1, d_1);
      dt_1 = vis_faligndata(d_1, d1);
      d22 = vis_faligndata(d2, d_2);
      dt_2 = vis_faligndata(d_2, d2);
      CONV_16(d21, k2)
      CONV_16(d22, k7)
      CONV_16_MIX(d21, dt_1, k4)
      CONV_16_MIX(d22, dt_2, k9)
      CONV_16(d_1, k5)
      CONV_16(d_2, k10)

      *ddst1++ = out0; *ddst1++ = out1;
      d1 = d_1; d2 = d_2;
      s1++; s2++;
    }
    ddst1 = (mlib_d64*)dbuf;
    d3= *s3; d4 = *s4;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d_3 = *(s3+1); d_4 = *(s4+1);
      out0 = ddst1[0]; out1 = ddst1[1];
      CONV_16(d3, k11)
      CONV_16(d4, k16)
      CONV_16_MIX(d3, d_3, k13)
      CONV_16_MIX(d4, d_4, k18)
      d23 = vis_faligndata(d3, d_3);
      dt_3 = vis_faligndata(d_3, d3);
      d24 = vis_faligndata(d4, d_4);
      dt_4 = vis_faligndata(d_4, d4);
      CONV_16(d23, k12)
      CONV_16(d24, k17)
      CONV_16_MIX(d23, dt_3, k14)
      CONV_16_MIX(d24, dt_4, k19)
      CONV_16(d_3, k15)
      CONV_16(d_4, k20)

      ddst1[0] = out0; ddst1[1] = out1; ddst1 += 2;
      d3 = d_3; d4 = d_4;
      s3++; s4++;
    }

    ddst1 = (mlib_d64*)dbuf;
    d5 = *s5;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d_5 = *(s5+1);
      out0 = *ddst1++; out1 = *ddst1++;
      CONV_16(d5, k21)
      CONV_16_MIX(d5, d_5, k23)
      d25 = vis_faligndata(d5, d_5);
      dt_5 = vis_faligndata(d_5, d5);
      CONV_16(d25, k22)
      CONV_16_MIX(d25, dt_5, k24)
      CONV_16(d_5, k25)

      *ddst++ = vis_fpackfix_pair(out0, out1);
      d5 = d_5;
      s5++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    sa4 = sa4 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv5x5_16nw_2( mlib_image *dst,         /* source  image */
                                   mlib_image *src,         /* destination image */
                                   mlib_s32   *kernel,      /* convolution kernel */
                                   mlib_s32   scalef_expon) /* scalefactor exponent */

{
  mlib_u16  *da, *d_a;                                          /* pointers to dst row */
  mlib_u16  *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u16  *sa, *sa1, *sa2, *sa3, *sa4;                        /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow;                   /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3, *sbuf4, *sbuf5;                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3, *s4, *s5;                            /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst, *ddst1;                                      /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d11, d12, d21, d22;                         /* data */
  mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8;
  mlib_f32  k9, k10, k11, k12, k13, k14, k15, k16;
  mlib_f32  k17, k18, k19, k20, k21, k22, k23, k24, k25;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       gsr_scale, i, j;

  gsr_scale = 32 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (4*dw+24)>>3;
  PREPARE_INTERM_BUFFERS()

  dw -= 4;
  dw *= 2;
  dh -= 4;

  sa  = adr_src;
  sa1 = sa + slb;
  sa2 = sa1 + slb;
  sa3 = sa2 + slb;
  sa4 = sa3 + slb;
  d_a = adr_dst + 4 + 2*dlb;

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf2, sa)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(8)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf3, sa1)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(8)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf4, sa2)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(8)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf5, sa3)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(8)

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    PREPARE_TO_LOAD_LINE(sbuf5, sa4)
#pragma pipeloop(0)
    LOAD_LINE_INTO_BUFFER(8)

    d1 = *s1; d2 = *s2;
    d11 = *(s1+1); d12 = *(s2+1);
    ddst1 = (mlib_d64*)dbuf;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d21 = *(s1+2); d22 = *(s2+2);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k6)
      CONV_16_MIX(d1, d11, k2)
      CONV_16_MIX(d2, d12, k7)
      CONV_16(d11, k3)
      CONV_16(d12, k8)
      CONV_16_MIX(d11, d21, k4)
      CONV_16_MIX(d12, d22, k9)
      CONV_16(d21, k5)
      CONV_16(d22, k10)

      *ddst1++ = out0; *ddst1++ = out1;
      d1 = d11; d2 = d12;
      d11 = d21; d12 = d22;
      s1++; s2++;
    }
    ddst1 = (mlib_d64*)dbuf;
    d1 = *s3; d2 = *s4;
    d11 = *(s3+1); d12 = *(s4+1);

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d21 = *(s3+2); d22 = *(s4+2);
      out0 = ddst1[0]; out1 = ddst1[1];
      CONV_16(d1, k11)
      CONV_16(d2, k16)
      CONV_16_MIX(d1, d11, k12)
      CONV_16_MIX(d2, d12, k17)
      CONV_16(d11, k13)
      CONV_16(d12, k18)
      CONV_16_MIX(d11, d21, k14)
      CONV_16_MIX(d12, d22, k19)
      CONV_16(d21, k15)
      CONV_16(d22, k20)

      ddst1[0] = out0; ddst1[1] = out1; ddst1 += 2;
      d1 = d11; d2 = d12;
      d11 = d21; d12 = d22;
      s3++; s4++;
    }

    ddst1 = (mlib_d64*)dbuf;
    d1 = *s5; d11 = *(s5+1);

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d21 = *(s5+2);
      out0 = *ddst1++; out1 = *ddst1++;
      CONV_16(d1, k21)
      CONV_16_MIX(d1, d11, k22)
      CONV_16(d11, k23)
      CONV_16_MIX(d11, d21, k24)
      CONV_16(d21, k25)

      *ddst++ = vis_fpackfix_pair(out0, out1);
      d1 = d11; d11 = d21;
      s5++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    sa4 = sa4 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv5x5_16nw_4( mlib_image *dst,         /* source  image */
                                   mlib_image *src,         /* destination image */
                                   mlib_s32   *kernel,      /* convolution kernel */
                                   mlib_s32   scalef_expon) /* scalefactor exponent */

{
  mlib_u16  *da, *d_a;                                          /* pointers to dst row */
  mlib_u16  *adr_dst,  *adr_src, *dend;                         /* pointers to src, dst data */
  mlib_u16  *sa, *sa1, *sa2, *sa3, *sa4;                        /* pointers to src rows */
  mlib_d64  *buff_src, *sbuf1, *sbuf2, *prow;                   /* pointers to rows in interm. src buf */
  mlib_d64  *sbuf3, *sbuf4, *sbuf5;                             /* pointers to rows in interm. src buf */
  mlib_d64  *dbuf;                                              /* pointer to row in interm. dst buf */
  mlib_d64  *s1, *s2, *s3, *s4, *s5;                            /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst, *ddst1;                                      /* mlib_d64 pointer to row in interm. dst buf */
  mlib_d64  d1, d2, d11, d12, d21, d22;                         /* data */
  mlib_d64  d31, d32, d41, d42;                                 /* data */
  mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8;
  mlib_f32  k9, k10, k11, k12, k13, k14, k15, k16;
  mlib_f32  k17, k18, k19, k20, k21, k22, k23, k24, k25;
  mlib_s32  dlb, slb,  buf_slb;                                 /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
  mlib_d64  *dsa, *dp;
  mlib_d64  sd0, sd1;
  mlib_s32  emask;
  int       gsr_scale, i, j;

  gsr_scale = 32 - scalef_expon;
  vis_write_gsr((gsr_scale << 3));

  GET_SRC_DST_PARAMETERS()
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (8*dw+40)>>3;
  PREPARE_INTERM_BUFFERS()

  dw -= 4;
  dw *= 4;
  dh -= 4;

  sa  = adr_src;
  sa1 = sa + slb;
  sa2 = sa1 + slb;
  sa3 = sa2 + slb;
  sa4 = sa3 + slb;
  d_a = adr_dst + 8 + 2*dlb;

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf2, sa)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(16)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf3, sa1)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(16)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf4, sa2)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(16)

  /* load interm. src buff */
  PREPARE_TO_LOAD_LINE(sbuf5, sa3)
#pragma pipeloop(0)
  LOAD_LINE_INTO_BUFFER(16)

#pragma pipeloop(0)
  for (j = 0; j < dh; j++)  {
    LOOP_INI()

    PREPARE_TO_LOAD_LINE(sbuf5, sa4)
#pragma pipeloop(0)
    LOAD_LINE_INTO_BUFFER(16)

    d1 = *s1; d2 = *s2;
    d11 = *(s1+1); d12 = *(s2+1);
    d21 = *(s1+2); d22 = *(s2+2);
    d31 = *(s1+3); d32 = *(s2+3);
    ddst1 = (mlib_d64*)dbuf;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d41 = *(s1+4); d42 = *(s2+4);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k6)
      CONV_16(d11, k2)
      CONV_16(d12, k7)
      CONV_16(d21, k3)
      CONV_16(d22, k8)
      CONV_16(d31, k4)
      CONV_16(d32, k9)
      CONV_16(d41, k5)
      CONV_16(d42, k10)

      *ddst1++ = out0; *ddst1++ = out1;
      d1 = d11; d2 = d12;
      d11 = d21; d12 = d22;
      d21 = d31; d22 = d32;
      d31 = d41; d32 = d42;
      s1++; s2++;
    }
    ddst1 = (mlib_d64*)dbuf;
    d1 = *s3; d2 = *s4;
    d11 = *(s3+1); d12 = *(s4+1);
    d21 = *(s3+2); d22 = *(s4+2);
    d31 = *(s3+3); d32 = *(s4+3);

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d41 = *(s3+4); d42 = *(s4+4);
      out0 = ddst1[0]; out1 = ddst1[1];
      CONV_16(d1, k11)
      CONV_16(d2, k16)
      CONV_16(d11, k12)
      CONV_16(d12, k17)
      CONV_16(d21, k13)
      CONV_16(d22, k18)
      CONV_16(d31, k14)
      CONV_16(d32, k19)
      CONV_16(d41, k15)
      CONV_16(d42, k20)

      ddst1[0] = out0; ddst1[1] = out1; ddst1 += 2;
      d1 = d11; d2 = d12;
      d11 = d21; d12 = d22;
      d21 = d31; d22 = d32;
      d31 = d41; d32 = d42;
      s3++; s4++;
    }

    ddst1 = (mlib_d64*)dbuf;
    d1 = *s5; d11 = *(s5+1);
    d21 = *(s5+2); d31 = *(s5+3);

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d41 = *(s5+4);
      out0 = *ddst1++; out1 = *ddst1++;
      CONV_16(d1, k21)
      CONV_16(d11, k22)
      CONV_16(d21, k23)
      CONV_16(d31, k24)
      CONV_16(d41, k25)

      *ddst++ = vis_fpackfix_pair(out0, out1);
      d1 = d11; d11 = d21;
      d21 = d31; d31 = d41;
      s5++;
    }
    PREPARE_TO_COPY_INTERM_BUF_TO_DST()

#pragma pipeloop(0)
    COPY_INTERM_BUF_TO_DST()
    COPY_TAIL()

    sa4 = sa4 + slb;
    d_a += dlb;
  }
  mlib_free(buff_src);
  return( MLIB_SUCCESS );
}

/************************************************************************/

mlib_status mlib_v_conv5x5_16nw_mask( mlib_image *dst,         /* source  image */
                                      mlib_image *src,         /* destination image */
                                      mlib_s32   *kernel,      /* convolution kernel */
                                      mlib_s32   scalef_expon, /* scalefactor exponent */
                                      mlib_s32   cmask)        /* mask */
{
  mlib_u16  *da, *d_a;                                        /* pointers to dst row */
  mlib_u16  *adr_dst,  *adr_src, *dend;                       /* pointers to src, dst data */
  mlib_u16  *sa, *sa1, *sa2, *sa3, *sa4, *sa5, *sa_5;         /* pointers to src rows */
  mlib_u16  *buff_src, *sbuf1, *sbuf2, *prow;                 /* pointers to rows in interm. src buf */
  mlib_u16  *sbuf3, *sbuf4, *sbuf5;                           /* pointers to rows in interm. src buf */
  mlib_u16  *dbuf, *d_buf;                                    /* pointer to row in interm. dst buf */
  mlib_u16  *s_buf1;
  mlib_d64  *s1, *s2, *s3, *s4, *s5;                          /* mlib_d64 pointers to rows in interm. src buf */
  mlib_d64  *ddst, *ddst1;                                    /* mlib_d64 pointer to row in interm. dst buf */
  mlib_s32  dlb, slb,  buf_slb;                               /* src, dst and interm. buf. strides */
  mlib_s32  dh, dw;
  mlib_d64  out0, out1, tmp0, tmp1, tmp2, tmp3;
  mlib_d64  d1, d2, d_1, d_2, d21, d22;                         /* data */
  mlib_d64  dt_1, dt_2;                                         /* data */
  mlib_f32  k1, k2, k3, k4, k5, k6, k7, k8;
  mlib_f32  k9, k10, k11, k12, k13, k14, k15, k16;
  mlib_f32  k17, k18, k19, k20, k21, k22, k23, k24, k25;                                  /* koefficients */
  int       gsr_scale, i, j, nchannel, nchannel1, chan, testchan;
  mlib_u16  t1, t2, t3, t4;

  gsr_scale = 32 - scalef_expon;
  vis_write_gsr((gsr_scale << 3)+2);

  GET_SRC_DST_PARAMETERS()
  nchannel = mlib_ImageGetChannels(src);
  LOAD_KERNEL_INTO_FLOAT()

  buf_slb = (2*dw+26) & (~7);                                             /* buf_slb - 8_byte alligned */
  buff_src = (mlib_u16*) mlib_malloc(7 * buf_slb * sizeof(mlib_u8) + 8);  /* alloc. interm. src buffer */
  if (buff_src == NULL) return MLIB_FAILURE;

  buf_slb >>= 1;

  sbuf1 = (mlib_u16*) ((mlib_addr)(buff_src +8) & (~7));
  sbuf2 = sbuf1 + buf_slb;
  sbuf3 = sbuf2 + buf_slb;
  sbuf4 = sbuf3 + buf_slb;
  sbuf5 = sbuf4 + buf_slb;
  dbuf = sbuf5 + buf_slb;

  dw -= 4;
  dh -= 4;                                                               /* edge - no write */

  testchan = 1;

  for ( chan = nchannel - 1; chan >= 0; chan--) {
    if ((cmask & testchan) == 0) {
      testchan <<= 1;
      continue;
    }
    testchan <<= 1;
    sa  = adr_src + chan;
    sa1 = sa + slb;
    sa2 = sa1 + slb;
    sa3 = sa2 + slb;
    sa4 = sa3 + slb;
    sa_5 = sa5 = sa4 + slb;
    d_a = adr_dst + (dlb << 1) + (nchannel << 1) + chan;

    /* load interm. src buff */
    for (i = 0, j = 0; j < (dw + 4); i+=nchannel, j++) {
      sbuf1[j] = sa4[i];  sbuf2[j] = sa[i];
      sbuf3[j] = sa1[i];  sbuf4[j] = sa2[i]; sbuf5[j] = sa3[i];
    }

    for (j = 0; j < dh - 1; j++)  {
      LOOP_INI()
      dend = da + (dw - 1)*nchannel;
      s_buf1 = sbuf1;
      d_buf = (dbuf-4);
      nchannel1 = 0;
      ddst1 = (mlib_d64*)dbuf;
      d1 = *s1; d2 = *s2;

     /* in each load data for processing next row */
#pragma pipeloop(0)
      for (i = 0; i < dw; i += 4) {
        d_1 = *(s1+1); d_2 = *(s2+1);
        t1 = *sa_5; sa_5 += nchannel;
        CONV_16_BEGIN(d1, k1)
        t2 = *sa_5; sa_5 += nchannel;
        CONV_16(d2, k6)
        t3 = *sa_5; sa_5 += nchannel;
        CONV_16_MIX(d1, d_1, k3)
        t4 = *sa_5; sa_5 += nchannel;
        CONV_16_MIX(d2, d_2, k8)
        d21 = vis_faligndata(d1, d_1);
        dt_1 = vis_faligndata(d_1, d1);
        d22 = vis_faligndata(d2, d_2);
        *s_buf1++ = t1;
        dt_2 = vis_faligndata(d_2, d2);
        CONV_16(d21, k2)
        *s_buf1++ = t2;
        CONV_16(d22, k7)
        *s_buf1++ = t3;
        CONV_16_MIX(d21, dt_1, k4)
        *s_buf1++ = t4;
        CONV_16_MIX(d22, dt_2, k9)
        CONV_16(d_1, k5)
        CONV_16(d_2, k10)

        *ddst1++ = out0; *ddst1++ = out1;
        d1 = d_1; d2 = d_2;
        s1++; s2++;
      }

      ddst1 = (mlib_d64*)dbuf;
      d1 = *s3;

     /* in each load data for processing next row */
#pragma pipeloop(0)
      for (i = 0; i < dw; i += 4) {
        d_1 = *(s3+1);
        out0 = ddst1[0]; out1 = ddst1[1];
        CONV_16(d1, k11)
        CONV_16_MIX(d1, d_1, k13)
        d21 = vis_faligndata(d1, d_1);
        dt_1 = vis_faligndata(d_1, d1);
        CONV_16(d21, k12)
        CONV_16_MIX(d21, dt_1, k14)
        CONV_16(d_1, k15)

        ddst1[0] = out0; ddst1[1] = out1; ddst1 += 2;
        d1 = d_1;
        s3++;
      }
      ddst1 = (mlib_d64*)dbuf;
      d1= *s4; d2 = *s5;

     /* in each iteraction store result from prev. iteract.*/
#pragma pipeloop(0)
      for (i = 0; i < dw; i += 4) {
        d_1 = *(s4+1); d_2 = *(s5+1);
        out0 = *ddst1++; out1 = *ddst1++;
        CONV_16(d1, k16)
        t1 = *d_buf++;
        CONV_16(d2, k21)
        t2 = *d_buf++;
        CONV_16_MIX(d1, d_1, k18)
        t3 = *d_buf++;
        CONV_16_MIX(d2, d_2, k23)
        t4 = *d_buf++;
        d21 = vis_faligndata(d1, d_1);
        dt_1 = vis_faligndata(d_1, d1);
        d22 = vis_faligndata(d2, d_2);
        *da = t1; da += nchannel1;
        dt_2 = vis_faligndata(d_2, d2);
        CONV_16(d21, k17)
        *da = t2; da += nchannel1;
        CONV_16(d22, k22)
        *da = t3; da += nchannel1;
        CONV_16_MIX(d21, dt_1, k19)
        *da = t4; da += nchannel1;
        CONV_16_MIX(d22, dt_2, k24)
        CONV_16(d_1, k20)
        CONV_16(d_2, k25)

        *ddst++ = vis_fpackfix_pair(out0, out1);
        d1 = d_1; d2 = d_2;
        s4++; s5++;
        nchannel1 = nchannel;
      }

      *s_buf1++ = *sa_5; sa_5 += nchannel; *s_buf1++ = *sa_5; sa_5 += nchannel;
      *s_buf1++ = *sa_5; sa_5 += nchannel; *s_buf1++ = *sa_5;

      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++; da += nchannel;
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++; da += nchannel;
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++; da += nchannel;
      }
      if ((mlib_addr) da <= (mlib_addr) dend) {
        *da = *d_buf++;
      }
      sa_5 = sa5 = sa5 + slb;
      d_a += dlb;
    }

    /* process last row - no need to load data */
    LOOP_INI()
    dend = da + (dw - 1)*nchannel;
    d_buf = (dbuf-4);
    nchannel1 = 0;
    ddst1 = (mlib_d64*)dbuf;
    d1 = *s1; d2 = *s2;

#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d_1 = *(s1+1); d_2 = *(s2+1);
      CONV_16_BEGIN(d1, k1)
      CONV_16(d2, k6)
      CONV_16_MIX(d1, d_1, k3)
      CONV_16_MIX(d2, d_2, k8)
      d21 = vis_faligndata(d1, d_1);
      dt_1 = vis_faligndata(d_1, d1);
      d22 = vis_faligndata(d2, d_2);
      dt_2 = vis_faligndata(d_2, d2);
      CONV_16(d21, k2)
      CONV_16(d22, k7)
      CONV_16_MIX(d21, dt_1, k4)
      CONV_16_MIX(d22, dt_2, k9)
      CONV_16(d_1, k5)
      CONV_16(d_2, k10)

      *ddst1++ = out0; *ddst1++ = out1;
      d1 = d_1; d2 = d_2;
      s1++; s2++;
    }

    ddst1 = (mlib_d64*)dbuf;
    d1 = *s3;

   /* in each load data for processing next row */
#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d_1 = *(s3+1);
      out0 = ddst1[0]; out1 = ddst1[1];
      CONV_16(d1, k11)
      CONV_16_MIX(d1, d_1, k13)
      d21 = vis_faligndata(d1, d_1);
      dt_1 = vis_faligndata(d_1, d1);
      CONV_16(d21, k12)
      CONV_16_MIX(d21, dt_1, k14)
      CONV_16(d_1, k15)

      ddst1[0] = out0; ddst1[1] = out1; ddst1 += 2;
      d1 = d_1;
      s3++;
    }
    ddst1 = (mlib_d64*)dbuf;
    d1= *s4; d2 = *s5;

   /* in each iteraction store result from prev. iteract.*/
#pragma pipeloop(0)
    for (i = 0; i < dw; i += 4) {
      d_1 = *(s4+1); d_2 = *(s5+1);
      out0 = *ddst1++; out1 = *ddst1++;
      CONV_16(d1, k16)
      t1 = *d_buf++;
      CONV_16(d2, k21)
      t2 = *d_buf++;
      CONV_16_MIX(d1, d_1, k18)
      t3 = *d_buf++;
      CONV_16_MIX(d2, d_2, k23)
      t4 = *d_buf++;
      d21 = vis_faligndata(d1, d_1);
      dt_1 = vis_faligndata(d_1, d1);
      d22 = vis_faligndata(d2, d_2);
      *da = t1; da += nchannel1;
      dt_2 = vis_faligndata(d_2, d2);
      CONV_16(d21, k17)
      *da = t2; da += nchannel1;
      CONV_16(d22, k22)
      *da = t3; da += nchannel1;
      CONV_16_MIX(d21, dt_1, k19)
      *da = t4; da += nchannel1;
      CONV_16_MIX(d22, dt_2, k24)
      CONV_16(d_1, k20)
      CONV_16(d_2, k25)

      *ddst++ = vis_fpackfix_pair(out0, out1);
      d1 = d_1; d2 = d_2;
      s4++; s5++;
      nchannel1 = nchannel;
    }

    if ((mlib_addr) da <= (mlib_addr) dend) {
      *da = *d_buf++; da += nchannel;
    }
    if ((mlib_addr) da <= (mlib_addr) dend) {
      *da = *d_buf++; da += nchannel;
    }
    if ((mlib_addr) da <= (mlib_addr) dend) {
      *da = *d_buf++; da += nchannel;
    }
    if ((mlib_addr) da <= (mlib_addr) dend) {
      *da = *d_buf++;
    }
  }
  mlib_free(buff_src);
  return MLIB_SUCCESS ;
}

/************************************************************************/

mlib_status mlib_v_conv5x5_16nw( mlib_image *dst,         /* source  image */
                                 mlib_image *src,         /* destination image */
                                 mlib_s32 *kernel,        /* convolution kernel */
                                 mlib_s32   scalef_expon, /* scalefactor exponent */
                                 mlib_s32   cmask)        /* mask */

{
  mlib_s32 nchannel, chan, cmask1, i;
  mlib_status res;

  cmask1 = cmask;
  nchannel = mlib_ImageGetChannels(src);
  chan = 0;
  for(i = nchannel - 1; i >= 0; i--, cmask1 >>= 1)
    if ((cmask1 & 1) == 1) chan++;

  if ((chan == nchannel) && (nchannel != 3)) {
    switch (nchannel) {
      case 1:
        res = mlib_v_conv5x5_16nw_1(dst, src, kernel, scalef_expon);
        break;
      case 2:
        res = mlib_v_conv5x5_16nw_2(dst, src, kernel, scalef_expon);
        break;
      case 4:
        res = mlib_v_conv5x5_16nw_4(dst, src, kernel, scalef_expon);
        break;
      default: res = MLIB_FAILURE; break;
    }
  } else res = mlib_v_conv5x5_16nw_mask(dst, src, kernel, scalef_expon, cmask);
  return ( res );
}
