/*
 * @(#)mlib_v_ImageConvMxN_32.c	1.1 98/09/07
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#pragma ident	"@(#)mlib_v_ImageConvMxN_32.c	1.1	98/09/02 SMI"


/*
 * FUNCTION
 *      mlib_v_convMxN_32nw - convolve a 32-bit image, MxN kernel,
 *                            edge = no write
 *      mlib_v_convMxN_32ext - convolve a 32-bit image, MxN kernel,
 *                             edge = src extended
 *
 * SYNOPSIS
 *      mlib_status
 *      mlib_convMxN_32nw(mlib_image *dst,
 *                        mlib_image *src,
 *                        mlib_s32   m,
 *                        mlib_s32   n,
 *                        mlib_s32   dm,
 *                        mlib_s32   dn,
 *                        mlib_s32   *kernel,
 *                        mlib_s32   scale,
 *                        mlib_s32   cmask)
 *
 *      mlib_convMxN_32ext(mlib_image *dst,
 *                         mlib_image *src,
 *                         mlib_s32   m,
 *                         mlib_s32   n,
 *                         mlib_s32   dm,
 *                         mlib_s32   dn,
 *                         mlib_s32   *kernel,
 *                         mlib_s32   scale,
 *                         mlib_s32   cmask)
 *
 * ARGUMENTS
 *      dst       Pointer to destination image.
 *      src       Pointer to source image.
 *      m         Kernel width.
 *      n         Kernel height.
 *      kernel    Pointer to convolution kernel.
 *      scale     The scaling factor to convert the input integer
 *                coefficients into floating-point coefficients:
 *                floating-point coefficient = integer coefficient * 2^(-scale)
 *      cmask     Channel mask to indicate the channels to be convolved.
 *                Each bit of which represents a channel in the image. The
 *                channels corresponded to 1 bits are those to be processed.
 *
 * DESCRIPTION
 *      A 2-D convolution (MxN kernel) for 32-bit images.
 *
 *************************************************************************
 */

#include <stdlib.h>
#include <mlib_image.h>


#define SETUP_ROW_KOEF(row, prev_row, ka)               \
   if (n == kcount) {                                   \
     kcount += kwid;                                    \
   row = ibuffp[nrow++];                                \
   } else row = prev_row + 1;                           \
   ka = fk[n++];

/************************************************************************/

mlib_status
mlib_v_convMxN_32nw( mlib_image *dst,       /* source image */
                     mlib_image *src,       /* destination image */
                     mlib_s32 kwid,         /* kernel width */
                     mlib_s32 khgt,         /* kernel height */
                     mlib_s32 dm,
                     mlib_s32 dn,
                     mlib_s32 *kernel,      /* convolution kernel */
                     mlib_s32 scalef_expon, /* scalefactor=2**N */
                     mlib_s32 cmask)
{
    int       i, j, ki, kj;         /* indices for row, col for image and kernel */
    mlib_s32  *sa, *sl, *adr_src;   /* pointers for pixel and line of source */
    mlib_s32  *da, *dl, *adr_dst;   /* pointers for pixel and line of destination */
    int       slb;                  /* src y stride in bytes */
    int       dlb;                  /* dst y stride in bytes */
    int       slpxls;               /* src y stride in pixels */
    int       dlpxls;               /* dst y stride in pixels */
    int       dh;                   /* dst height */
    int       dw;                   /* dst width */
    int       sw, sh;               /* src width, height */
    int       khh;                  /* krnl half height */
    int       khw;                  /* krnl half width */
    int       khgtm1;               /* khgt minus 1 */
    double    *ibuff;               /* buf to divide into rowbufs */
    double    **ibuffp;             /* ptr to rowbuf ptrs */
    double    *itmp;                /* tmp ptrs */
    double    *isum;
    double    *fk;                  /* for floating pt coeff buf */
    double    fscalef;              /* num bits to R shift the accumulator */
    int       n;
    int       chan, testchan, nchannel, tmp;
    double k1, k2, k3, k4, k5, k6, k7, k8;
    double k9, k10, k11, k12;
    double *s, *s1, *s2, *s3, *s4, *s5, *s6, *s7, *s8;
    double *s9, *s10, *s11, *s12;
    int  nrow, kcount;

    /* allocate and fill floating point coeff buf */

    fk = (mlib_d64 *)mlib_malloc( sizeof(mlib_d64) * (khgt * kwid + 1) );
    if ( fk==NULL ) return( MLIB_NULLPOINTER );
    fscalef = 1.0;
    while (scalef_expon > 30) {
      fscalef /= (1 << 30);
      scalef_expon -= 30;
    }

    fscalef /= (1 << scalef_expon);
    n = 0;
    for ( ki = 0; ki < khgt; ki++ ) {
      for ( kj = 0; kj < kwid; kj++ ) {
        fk[n] = fscalef * kernel[n];
        n++;
      }
    }
    fk[n] = 0;

    adr_src = mlib_ImageGetData(src);
    adr_dst = mlib_ImageGetData(dst);
    slb = mlib_ImageGetStride(src);
    dlb = mlib_ImageGetStride(dst);
    slpxls = slb >> 2;
    dlpxls = dlb >> 2;
    sw = mlib_ImageGetWidth(src);
    sh = mlib_ImageGetHeight(src);
    nchannel = mlib_ImageGetChannels(src);
    khh = dn;
    khw = dm;
    dh = sh - (khgt-1);
    dw = sw - (kwid-1);  /* no write to outer rows, cols */

    /* allocate khgt-line integer buffer for intermediate data */

    ibuff = (double *)mlib_malloc(sizeof(double)*khgt*sw);
    if (ibuff == NULL) {
      mlib_free(fk);
      return( MLIB_FAILURE );
    }
    ibuffp = (double **)mlib_malloc(sizeof(double *)*khgt);
    if (ibuffp == NULL) {
      mlib_free(fk);
      mlib_free(ibuff);
      return( MLIB_FAILURE );
    }
    isum = (double*)mlib_malloc(sizeof(double)*sw);
    if (isum == NULL) {
      mlib_free(fk);
      mlib_free(ibuff);
      mlib_free(ibuffp);
      return( MLIB_FAILURE );
    }

    testchan = 1;
    for ( chan = nchannel - 1; chan >= 0; chan--) {
      if ((cmask & testchan) == 0) {
        testchan <<= 1;
        continue;
      }
      testchan <<= 1;
      sl = sa = adr_src + chan;
      dl = da = adr_dst + chan;
      dl = da += ((dlpxls * khh) + (nchannel *khw));

      khgtm1 = khgt - 1;
      for (i = 0; i < khgt; i++)            /* init rowbuf ptrs */
      ibuffp[i] = &ibuff[sw * i];

      for (i = 0; i < khgtm1; i++) {        /* load all but last rowbuf */
        for (j = 0; j < sw; j++) {
          ibuffp[i][j] = *sa;
          sa += nchannel;
        }
        sa = (sl += slpxls);
      }

      for (i = 0; i < dh; i++) {            /* for each destination row */
        for (j = 0; j < sw; j++) {          /* copy next src row -> last rowbuf */
          ibuffp[khgtm1][j] = *sa;
          sa += nchannel;
        }

        n = 0;
        nrow = 0;
        kcount = kwid;
        s1 = ibuffp[nrow++];
        k1 = fk[n++];
        if (n == kcount) {
            kcount += kwid;
            s = ibuffp[nrow++];
          } else s = s1 + 1;

        for (j = 0; j < dw; j++) isum[j] = k1*(*s1++);

        for (ki = khgt*kwid-1; ki >= 12; ki-= 12) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)
          SETUP_ROW_KOEF(s5, s4, k5)
          SETUP_ROW_KOEF(s6, s5, k6)
          SETUP_ROW_KOEF(s7, s6, k7)
          SETUP_ROW_KOEF(s8, s7, k8)
          SETUP_ROW_KOEF(s9, s8, k9)
          SETUP_ROW_KOEF(s10, s9, k10)
          SETUP_ROW_KOEF(s11, s10, k11)
          SETUP_ROW_KOEF(s12, s11, k12)

          if (n == kcount) {
            kcount += kwid;
            s = ibuffp[nrow++];
          } else s = s12 + 1;

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++) +
                       k5 *(*s5++) + k6 * (*s6++) + k7 * (*s7++) + k8 * (*s8++) +
                       k9 *(*s9++) + k10 * (*s10++) + k11 * (*s11++) + k12 * (*s12++);
          }
        }
        if ( ki >= 6) {
          ki -= 6;
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)
          SETUP_ROW_KOEF(s5, s4, k5)
          SETUP_ROW_KOEF(s6, s5, k6)

          if (n == kcount) {
            kcount += kwid;
            s = ibuffp[nrow++];
          } else s = s6 + 1;

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++) +
                       k5 *(*s5++) + k6 * (*s6++);
          }
        }


              if (ki == 5) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)
          SETUP_ROW_KOEF(s5, s4, k5)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++) +
                       k5 *(*s5++);

          }
        }
        else if (ki == 4) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++);

          }
        }
        else if (ki == 3) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++);

          }
        }
        else if (ki == 2) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++);
          }
        }
        else if (ki == 1) {
          k1 = fk[n++];
          s1 = s;

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++);
          }
        }


        for (j = 0; j < dw; j++) {
          tmp = (mlib_s32)isum[j];
          *da = tmp;
          da += nchannel;
        }
        itmp = ibuffp[0];                    /* rotate rowbuf ptrs */
        for (ki = 0; ki < khgtm1; ki++) ibuffp[ki] = ibuffp[ki+1];
        ibuffp[khgtm1] = itmp;
        sa = (sl += slpxls);                 /* setup to copy next src row */
        da = (dl += dlpxls);                 /* setup to write next dst row */
      }
    }

    mlib_free(fk);
    mlib_free(ibuff);
    mlib_free(ibuffp);
    mlib_free(isum);
    return MLIB_SUCCESS;
}

/************************************************************************/

mlib_status
mlib_v_convMxN_32ext( mlib_image *dst,       /* source image */
                      mlib_image *src,       /* destination image */
                      mlib_s32 kwid,         /* kernel width */
                      mlib_s32 khgt,         /* kernel height */
                      mlib_s32 dm,
                      mlib_s32 dn,
                      mlib_s32 *kernel,      /* convolution kernel */
                      mlib_s32 scalef_expon, /* scalefactor=2**N */
                      mlib_s32 cmask)
{
    int       i, j, ki, kj;         /* indices for row, col for image and kernel */
    mlib_s32  *sa, *sl, *adr_src;   /* pointers for pixel and line of source */
    mlib_s32  *da, *dl, *adr_dst;   /* pointers for pixel and line of destination */
    int       slb;                  /* src y stride in bytes */
    int       dlb;                  /* dst y stride in bytes */
    int       slpxls;               /* src y stride in pixels */
    int       dlpxls;               /* dst y stride in pixels */
    int       dh;                   /* dst height */
    int       dw;                   /* dst width */
    int       sw, sh;               /* src width, height */
    int       khh;                  /* krnl half height */
    int       khw;                  /* krnl half width */
    int       khgtm1;               /* khgt minus 1 */
    double    *ibuff;               /* buf to divide into rowbufs */
    double    **ibuffp;             /* ptr to rowbuf ptrs */
    double    *itmp;                /* tmp ptrs */
    double    *isum;
    double    *fk;                  /* for floating pt coeff buf */
    double    fscalef;              /* num bits to R shift the accumulator */
    int       n;
    int       chan, testchan, nchannel, tmp;
    double k1, k2, k3, k4, k5, k6, k7, k8;
    double k9, k10, k11, k12;
    double *s, *s1, *s2, *s3, *s4, *s5, *s6, *s7, *s8;
    double *s9, *s10, *s11, *s12;
    int  nrow, kcount;

    /* allocate and fill floating point coeff buf */

    fk = (mlib_d64 *)mlib_malloc( sizeof(mlib_d64) * (khgt * kwid + 1));
    if ( fk==NULL ) return( MLIB_NULLPOINTER );

    fscalef = 1.0;
    while (scalef_expon > 30) {
      fscalef /= (1 << 30);
      scalef_expon -= 30;
    }

    fscalef /= (1 << scalef_expon);

    n = 0;
    for ( ki = 0; ki < khgt; ki++ ) {
      for ( kj = 0; kj < kwid; kj++ ) {
        fk[n] = fscalef * kernel[n];
        n++;
      }
    }

    fk[n] = 0;

    adr_src = mlib_ImageGetData(src);
    adr_dst = mlib_ImageGetData(dst);
    slb = mlib_ImageGetStride(src);
    dlb = mlib_ImageGetStride(dst);
    slpxls = slb >> 2;
    dlpxls = dlb >> 2;
    sw = mlib_ImageGetWidth(src);
    sh = mlib_ImageGetHeight(src);
    nchannel = mlib_ImageGetChannels(src);
    khh = dn;
    khw = dm;
    dh = sh;
    dw = sw;
    sw += (kwid-1);

    /* allocate khgt-line integer buffer for intermediate data */

    ibuff = (double *)mlib_malloc(sizeof(double)*khgt*sw);
    if (ibuff == NULL) {
      mlib_free(fk);
      return( MLIB_FAILURE );
    }
    ibuffp = (double **)mlib_malloc(sizeof(double *)*khgt);
    if (ibuffp == NULL) {
      mlib_free(fk);
      mlib_free(ibuff);
      return( MLIB_FAILURE );
    }
    isum = (double*)mlib_malloc(sizeof(double)*sw);
    if (isum == NULL) {
      mlib_free(fk);
      mlib_free(ibuff);
      mlib_free(ibuffp);
      return( MLIB_FAILURE );
    }

    testchan = 1;
    for ( chan = nchannel - 1; chan >= 0; chan--) {
      if ((cmask & testchan) == 0) {
        testchan <<= 1;
        continue;
      }
      testchan <<= 1;
      sl = sa = adr_src + chan;
      dl = da = adr_dst + chan;

      khgtm1 = khgt - 1;
      for (i = 0; i < khgt; i++)            /* init rowbuf ptrs */
      ibuffp[i] = &ibuff[sw * i];

      for (i = 0; i < khh; i++) {
        for (j = 0; j < dw; j++) {
          ibuffp[i][j+khw] = *sa;
          sa += nchannel;
        }
        for (j = 0; j < khw; j++) {
          ibuffp[i][j] = ibuffp[i][khw];                          /* extend edges */
        }
        for (j = 0; j < kwid - 1 - khw; j++) {
          ibuffp[i][dw + khw + j] = ibuffp[i][dw + khw - 1];
        }
        sa = sl;
      }

      for (; i < khgtm1; i++) {
        for (j = 0; j < dw; j++) {
          ibuffp[i][j+khw] = *sa;
          sa += nchannel;
        }
        for (j = 0; j < khw; j++) {
          ibuffp[i][j] = ibuffp[i][khw];
        }
        for (j = 0; j < kwid - 1 - khw; j++) {                          /* extend edges */
          ibuffp[i][dw + khw + j] = ibuffp[i][dw + khw - 1];
        }
        sa = (sl += slpxls);
      }

      for (i = 0; i < dh; i++) {            /* for each destination row */
        for (j = 0; j < dw; j++) {
          ibuffp[khgtm1][j+khw] = *sa;
          sa += nchannel;
        }
        for (j = 0; j < khw; j++) {
          ibuffp[khgtm1][j] = ibuffp[khgtm1][khw];
        }
        for (j = 0; j < kwid - 1 - khw; j++) {                          /* extend edges */
          ibuffp[khgtm1][dw + khw + j] = ibuffp[khgtm1][dw + khw - 1];
        }

        n = 0;
        nrow = 0;
        kcount = kwid;
        s1 = ibuffp[nrow++];
        k1 = fk[n++];
        if (n == kcount) {
          kcount += kwid;
          s = ibuffp[nrow++];
        } else s = s1 + 1;

        for (j = 0; j < dw; j++) isum[j] = k1*(*s1++);

        for (ki = khgt*kwid-1; ki >= 12; ki-= 12) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)
          SETUP_ROW_KOEF(s5, s4, k5)
          SETUP_ROW_KOEF(s6, s5, k6)
          SETUP_ROW_KOEF(s7, s6, k7)
          SETUP_ROW_KOEF(s8, s7, k8)
          SETUP_ROW_KOEF(s9, s8, k9)
          SETUP_ROW_KOEF(s10, s9, k10)
          SETUP_ROW_KOEF(s11, s10, k11)
          SETUP_ROW_KOEF(s12, s11, k12)

          if (n == kcount) {
            kcount += kwid;
            s = ibuffp[nrow++];
          } else s = s12 + 1;

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++) +
                       k5 *(*s5++) + k6 * (*s6++) + k7 * (*s7++) + k8 * (*s8++) +
                       k9 *(*s9++) + k10 * (*s10++) + k11 * (*s11++) + k12 * (*s12++);
          }
        }
        if ( ki >= 6) {
          ki -= 6;
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)
          SETUP_ROW_KOEF(s5, s4, k5)
          SETUP_ROW_KOEF(s6, s5, k6)

          if (n == kcount) {
            kcount += kwid;
            s = ibuffp[nrow++];
          } else s = s6 + 1;

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++) +
                       k5 *(*s5++) + k6 * (*s6++);
          }
        }


              if (ki == 5) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)
          SETUP_ROW_KOEF(s5, s4, k5)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++) +
                       k5 *(*s5++);

          }
        }
        else if (ki == 4) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)
          SETUP_ROW_KOEF(s4, s3, k4)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++) + k4 * (*s4++);

          }
        }
        else if (ki == 3) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)
          SETUP_ROW_KOEF(s3, s2, k3)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++) + k3 * (*s3++);

          }
        }
        else if (ki == 2) {
          k1 = fk[n++];
          s1 = s;
          SETUP_ROW_KOEF(s2, s1, k2)

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++) + k2 * (*s2++);
          }
        }
        else if (ki == 1) {
          k1 = fk[n++];
          s1 = s;

          for (j = 0; j < dw; j ++) {
            isum[j] += k1 *(*s1++);
          }
        }

        for (j = 0; j < dw; j++) {
          tmp = (mlib_s32)isum[j];
          *da = tmp;
          da += nchannel;
        }
        itmp = ibuffp[0];                    /* rotate rowbuf ptrs */
        for (ki = 0; ki < khgtm1; ki++) ibuffp[ki] = ibuffp[ki+1];
        ibuffp[khgtm1] = itmp;
        if (i < dh - khgt + khh) {
          sa = (sl += slpxls);
        } else  sa = sl;                       /* setup to copy next src row */
        da = (dl += dlpxls);                   /* setup to write next dst row */
      }
    }

    mlib_free(fk);
    mlib_free(ibuff);
    mlib_free(ibuffp);
    mlib_free(isum);
    return MLIB_SUCCESS;
}
