/*
 * @(#)multi_font.c	1.29 98/08/31
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
/*
 * These routines are used for display string with multi font.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <jni.h>
#include <jni_util.h>
#include <jvm.h>

#include "awt_Font.h"
#include "awt_Component.h"
#include "awt_MenuItem.h"
#include "awt_p.h"
#include "multi_font.h"

extern XFontStruct *loadFont(Display *, char *, int);

extern struct FontIDs fontIDs;
extern struct MComponentPeerIDs mComponentPeerIDs;
extern struct MMenuItemPeerIDs mMenuItemPeerIDs;
extern struct PlatformFontIDs platformFontIDs;
extern struct MFontPeerIDs mFontPeerIDs;


/*
 * make string with str + string representation of num
 * This string is used as tag string of Motif Compound String and FontList. 
 */
static void 
makeTag(char *str, int num, char *buf)
{

    int len = strlen(str);

    strcpy(buf, str);
    buf[len] = '0' + num % 100;
    buf[len + 1] = '\0';
}

jobject 
awtJNI_CreateAndSetGlobalRef(JNIEnv * env, jobject this)
{
    jobject gRef;

    gRef = (*env)->NewGlobalRef(env, this);

    (*env)->SetLongField(env, this,
			 mComponentPeerIDs.jniGlobalRef, (jlong) gRef);

    return gRef;
}

struct gRefStruct
{
    jobject gRef;
    struct gRefStruct *next;
};

static struct gRefStruct *gRefHead = NULL;
static struct gRefStruct *gRefTail = NULL;

/*
 * This function is called by components that
 * are being disposed. It used to invalidate
 * the global ref immediately, but the awt is
 * rather full of thread race conditions involving
 * component disposal and outstanding events.
 * Now we queue up 'to be deleted' global refs
 * as they come in, and don't invalidate them
 * until the X event queue is empty. Callers of
 * either of these functions _must_ have AWT_LOCK'd
 * before using them!
 */

void 
awtJNI_DeleteGlobalRef(JNIEnv * env, jobject this)
{
    jobject gRef;
    struct gRefStruct *newGRef;
    struct gRefStruct *temp;

    gRef = (jobject)
      (*env)->GetLongField(env, this, mComponentPeerIDs.jniGlobalRef);
    (*env)->SetLongField(env, this, mComponentPeerIDs.jniGlobalRef, 0);

    /*
     * Verra handy for tracking down race conditions. If you
     * have a peer getting called after its been disposed...
     */
    /* jio_fprintf(stderr,"%p\n",(void *)gRef); */

    newGRef = (struct gRefStruct *)malloc((size_t)sizeof(struct gRefStruct));

    if(newGRef == NULL)
        (*env)->DeleteGlobalRef(env, gRef);
    else
    {
        newGRef->gRef = gRef;
        newGRef->next = NULL;

        if(gRefHead == NULL)
        {
            gRefTail = newGRef;
            gRefHead = newGRef;
        }
        else
        {
            gRefTail->next = newGRef;
            gRefTail = newGRef;
        }
    }
}

void 
awtJNI_DeleteGlobalMenuRef(JNIEnv * env, jobject this)
{
    jobject gRef;
    struct gRefStruct *newGRef;
    struct gRefStruct *temp;

    gRef = (jobject)
      (*env)->GetLongField(env, this, mMenuItemPeerIDs.jniGlobalRef);
    (*env)->SetLongField(env, this, mMenuItemPeerIDs.jniGlobalRef, 0);

    /*
     * Verra handy for tracking down race conditions. If you
     * have a peer getting called after its been disposed...
     */
    /* jio_fprintf(stderr,"%p\n",(void *)gRef); */

    newGRef = (struct gRefStruct *)malloc((size_t)sizeof(struct gRefStruct));

    if(newGRef == NULL)
        (*env)->DeleteGlobalRef(env, gRef);
    else
    {
        newGRef->gRef = gRef;
        newGRef->next = NULL;

        if(gRefHead == NULL)
        {
            gRefTail = newGRef;
            gRefHead = newGRef;
        }
        else
        {
            gRefTail->next = newGRef;
            gRefTail = newGRef;
        }
    }
}

void awtJNI_CleanupGlobalRefs()
{
    struct gRefStruct *working,*next;
    JNIEnv *env;
    int count = 0;

    if(gRefHead == NULL)
        return;

    env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    working = gRefHead;
    gRefHead = gRefTail = NULL;

    while(working != NULL)
    {
        count++;
        next = working->next;
        (*env)->DeleteGlobalRef(env, working->gRef); 

        free((void *)working);

        working = next;
    }
}

static int 
awtJNI_GetFontDescriptorNumber(JNIEnv * env
                               ,jobject font
                               ,jobject fd)
{
    int i, num;
    /* initialize to 0 so that DeleteLocalRef will work. */
    jobjectArray componentFonts = 0;
    jobject peer = 0;
    jobject temp;
    jboolean validRet = JNI_FALSE;

    if ((*env)->EnsureLocalCapacity(env, 2) < 0)
        goto done;

    peer = (*env)->CallObjectMethod(env,font,fontIDs.getPeer);
    if (!peer)
        goto done;

    componentFonts = (jobjectArray)
        (*env)->GetObjectField(env,peer,platformFontIDs.componentFonts);

    if (!componentFonts)
        goto done;

    num = (*env)->GetArrayLength(env, componentFonts);

    for (i = 0; i < num; i++) {
        temp = (*env)->GetObjectArrayElement(env, componentFonts, i);

        if ((*env)->IsSameObject(env, fd, temp)) {
            validRet = JNI_TRUE;
            break;
        }
        (*env)->DeleteLocalRef(env, temp);
    }

 done:
    (*env)->DeleteLocalRef(env, peer);
    (*env)->DeleteLocalRef(env, componentFonts);

    if (validRet)
        return i;

    return 0;
}

jobject 
awtJNI_GetFont(JNIEnv * env, jobject this)
{
    jobject target;
    jobject font;

    target = (*env)->GetObjectField(env, this, mComponentPeerIDs.target);
    // SECURITY: Must call _NoClientCode() methods to ensure that we
    //           are not invoking client code on the privileged thread
    font = JNU_CallMethodByName(env, 
				NULL, 
				target, 
				"getFont_NoClientCode",
				"()Ljava/awt/Font;").l;
    (*env)->DeleteLocalRef(env, target);
    return font;
}

jobject 
awtJNI_GetFMFont(JNIEnv * env, jobject this)
{
    return JNU_CallMethodByName(env, NULL, this, "getFont_NoClientCode", 
				"()Ljava/awt/Font;").l;
}

jboolean 
awtJNI_IsMultiFont(JNIEnv * env, jobject this)
{
    jobject peer;
    jobject props;

    if (this == 0) {
        return JNI_FALSE;
    }

    if ((*env)->EnsureLocalCapacity(env, 2) < 0)
        return JNI_FALSE;

    peer = (*env)->CallObjectMethod(env,this,fontIDs.getPeer);
    if (peer == 0) {
        return JNI_FALSE;
    }

    props = (*env)->GetObjectField(env,peer,platformFontIDs.props);
    (*env)->DeleteLocalRef(env, peer);

    if (props == 0) {
        return JNI_FALSE;
    }
    (*env)->DeleteLocalRef(env, props);

    return JNI_TRUE;
}

jboolean 
awtJNI_IsMultiFontMetrics(JNIEnv * env, jobject this)
{
    jobject peer;
    jobject props;
    jobject font;

    if (JNU_IsNull(env, this)) {
        return JNI_FALSE;
    }
    if ((*env)->EnsureLocalCapacity(env, 3) < 0)
        return JNI_FALSE;

    font = JNU_CallMethodByName(env, NULL, this, "getFont_NoClientCode",
				"()Ljava/awt/Font;").l;
    if (JNU_IsNull(env, font)) {
        return JNI_FALSE;
    }

    peer = (*env)->CallObjectMethod(env,font,fontIDs.getPeer);
    (*env)->DeleteLocalRef(env, font);

    if (peer == 0) {
        return JNI_FALSE;
    }

    props = (*env)->GetObjectField(env,peer,platformFontIDs.props);
    (*env)->DeleteLocalRef(env, peer);
    if (props == 0) {
        return JNI_FALSE;
    }
    (*env)->DeleteLocalRef(env, props);

    return JNI_TRUE;
}

/*
 * Unicode to Motif Multi Font Compound String converter
 *
 * ASSUMES: We are not running on a privileged thread
 */
XmString 
awtJNI_MakeMultiFontString(JNIEnv * env, jstring s, jobject font)
{
    XmString xmstr = NULL, xmtmp1, xmtmp2;
    jobjectArray dataArray;
    char *err;
    int stringCount,i;
    int fdnumber;
    struct FontData *fdata = awtJNI_GetFontData(env, font, &err);
    jobject fontDescriptor;
    jbyteArray data;
    int length;
    char *stringData;
    char tag[BUFSIZ];

    if ((*env)->PushLocalFrame(env, 16) < 0)
        return NULL;

    if (!JNU_IsNull(env, s) && !JNU_IsNull(env, font)) {
        jobject peer;

        peer = (*env)->CallObjectMethod(env,font,fontIDs.getPeer);

        assert(!awt_currentThreadIsPrivileged(env));
        dataArray =
            (*env)->CallObjectMethod(
			     env,
			     peer,
			     platformFontIDs.makeConvertedMultiFontString,
			     s);

        if ((*env)->ExceptionOccurred(env)) {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);

	    (*env)->PopLocalFrame(env, 0);
            return (XmString) NULL;
        }

        if(dataArray == 0) {
	    (*env)->PopLocalFrame(env, 0);
            return (XmString) NULL;
        }
    } else {
      (*env)->PopLocalFrame(env, 0);
        return (XmString) NULL;
    }

    stringCount = (*env)->GetArrayLength(env, dataArray);

    for (i = 0; i < stringCount; i+=2) {
        fontDescriptor = (*env)->GetObjectArrayElement(env, dataArray, i);
        data = (*env)->GetObjectArrayElement(env, dataArray, i + 1);

        /* Bail if we've finished */
        if(fontDescriptor == 0 || data == 0)
            break;

        fdnumber = awtJNI_GetFontDescriptorNumber(env, font, fontDescriptor);
        fdata = awtJNI_GetFontData(env, font, &err);

        makeTag(fdata->flist[fdnumber].charset_name, fdnumber, tag);

        stringData = (char *)(*env)->GetPrimitiveArrayCritical(env, data, NULL);
        if(stringData != NULL) {
            char *offsetStringData;

            offsetStringData = stringData + (4 * sizeof(char));

            if(xmstr == NULL)
                xmstr = XmStringCreate(offsetStringData, tag);
            else {
                xmtmp1 = XmStringCreate(offsetStringData, tag);
                xmtmp2 = XmStringConcat(xmstr, xmtmp1);
                XmStringFree(xmtmp1);
                XmStringFree(xmstr);
                xmstr = xmtmp2;
            }
        }

        (*env)->ReleasePrimitiveArrayCritical(env, data, stringData, JNI_ABORT);
	(*env)->DeleteLocalRef(env, fontDescriptor);
	(*env)->DeleteLocalRef(env, data);
    }
    (*env)->PopLocalFrame(env, 0);
    return xmstr;
}

XmFontList 
awtJNI_GetFontList(JNIEnv * env, jobject font)
{
    int i;
    XmFontListEntry fle;
    XmFontList fontlist;
    XFontStruct *xf;
    int size;
    struct FontData *fdata;
    char *err, tag[BUFSIZ];
    
    fdata = awtJNI_GetFontData(env, font, &err);

    makeTag(fdata->flist[0].charset_name, 0, tag);

    size = (int) (*env)->GetIntField(env, font, fontIDs.size);

    if (fdata->flist[0].load == 0) {
        xf = loadFont(awt_display, fdata->flist[0].xlfd, size * 10);

	if (xf == NULL) {
	    /*
	    printf("Cannot load font: %s\n", fdata->list[0].xlfd);
	    */
	} else {
            fdata->flist[0].xfont = xf;
            fdata->flist[0].load = 1;

            if (xf->min_byte1 == 0 && xf->max_byte1 == 0)
        	fdata->flist[0].index_length = 1;
            else
        	fdata->flist[0].index_length = 2;
	}
    }
    fle = XmFontListEntryLoad(awt_display,
                              fdata->flist[0].xlfd,
                              XmFONT_IS_FONT,
                              tag);

    fontlist = XmFontListAppendEntry(NULL, fle);
    /*
     * Some versions of motif have a bug in 
     * XmFontListEntryFree() which causes it to free more than it
     * should.  Use XtFree() is used instead.  See O'Reilly's
     * Motif Reference Manual for more information.
     */
    XmFontListEntryFree(&fle);

    for (i = 1; i < fdata->charset_num; i++) {
        makeTag(fdata->flist[i].charset_name, i, tag);

        if (fdata->flist[i].load == 0) {
            xf = loadFont(awt_display, fdata->flist[i].xlfd, size * 10);

	    if (xf == NULL) {
		/*
		printf("Cannot load font: %s\n", fdata->flist[0].xlfd);
		*/
		continue;
	    } 
            fdata->flist[i].xfont = xf;
            fdata->flist[i].load = 1;
            if (xf->min_byte1 == 0 && xf->max_byte1 == 0) {
                fdata->flist[i].index_length = 1;
            } else {
                fdata->flist[i].index_length = 2;
            }
        }
        fle = XmFontListEntryLoad(awt_display,
                                  fdata->flist[i].xlfd,
                                  XmFONT_IS_FONT,
                                  tag);
        fontlist = XmFontListAppendEntry(fontlist, fle);
	/*
	 * Some versions of motif have a bug in 
	 * XmFontListEntryFree() which causes it to free more than it
	 * should.  Use XtFree() instead.  See O'Reilly's
	 * Motif Reference Manual for more information.
	 */
	XmFontListEntryFree(&fle);

    }

    return fontlist;
}

XFontSet 
awtJNI_MakeFontSet(JNIEnv * env, jobject font)
{
    jstring xlfd;
    char *xfontset;
    int size;
    int length = 0;
    char *realxlfd, *ptr, *prev;
    char **missing_list;
    int missing_count;
    char *def_string;
    XFontSet xfs;
    jobject peer;
    jstring xfsname;

    if ((*env)->EnsureLocalCapacity(env, 2) < 0)
        return 0;

    size = (*env)->GetIntField(env, font, fontIDs.size) * 10;

    peer = (*env)->CallObjectMethod(env,font,fontIDs.getPeer);
    xfsname = (*env)->GetObjectField(env, peer, mFontPeerIDs.xfsname);

    if (JNU_IsNull(env, xfsname))
        xfontset = "";
    else
        xfontset = (char *)JNU_GetStringPlatformChars(env, xfsname, NULL);

    realxlfd = malloc(strlen(xfontset) + 50);

    prev = ptr = xfontset;
    while (ptr = strstr(ptr, "%d")) {
        char save = *(ptr + 2);

        *(ptr + 2) = '\0';
        jio_snprintf(realxlfd + length, strlen(xfontset) + 50 - length,
                     prev, size);
        length = strlen(realxlfd);
        *(ptr + 2) = save;

        prev = ptr + 2;
        ptr += 2;
    }
    strcpy(realxlfd + length, prev);

    xfs = XCreateFontSet(awt_display, realxlfd, &missing_list,
                         &missing_count, &def_string);

    free((void *)realxlfd);

    if (xfontset && !JNU_IsNull(env, xfsname))
        JNU_ReleaseStringPlatformChars(env, xfsname, (const char *) xfontset);

    (*env)->DeleteLocalRef(env, peer);
    (*env)->DeleteLocalRef(env, xfsname);
    return xfs;
}

/*
 * ASSUMES: We are not running on a privileged thread
 */
void 
awtJNI_DrawMFString(JNIEnv *env, jcharArray s, struct GraphicsData *gdata
                    , jobject font, int x, int y, int offset, int sLength)
{
    jobjectArray dataArray;
    char *err;
    unsigned char *stringData;
    char *offsetStringData;
    int stringCount,i;
    int size;
    struct FontData *fdata = awtJNI_GetFontData(env, font, &err);
    jobject fontDescriptor;
    jbyteArray data;
    int j;
    int X = x;
    int Y = y;
    int length,copyLength;
    XFontStruct *xf;

    if ((*env)->EnsureLocalCapacity(env, 3) < 0)
        return;

    if (!JNU_IsNull(env, s) && !JNU_IsNull(env, font))
    {
        jobject peer;

        assert(!awt_currentThreadIsPrivileged(env));
        peer = (*env)->CallObjectMethod(env,font,fontIDs.getPeer);
        dataArray = (*env)->CallObjectMethod(
				 env,
				 peer,
				 platformFontIDs.makeConvertedMultiFontChars,
				 s, offset, sLength);

        if ((*env)->ExceptionOccurred(env))
        {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        }

	(*env)->DeleteLocalRef(env, peer);

        if(dataArray == 0)
        {
            return;
        }
    } else {
        return;
    }

    stringCount = (*env)->GetArrayLength(env, dataArray);

    size = (*env)->GetIntField(env, font, fontIDs.size);

    for (i = 0; i < stringCount; i+=2)
    {
        fontDescriptor = (*env)->GetObjectArrayElement(env, dataArray, i);
        data = (*env)->GetObjectArrayElement(env, dataArray, i + 1);

        /* Bail if we've finished */
        if(fontDescriptor == 0 || data == 0) {
	    (*env)->DeleteLocalRef(env, fontDescriptor);
	    (*env)->DeleteLocalRef(env, data);
	    (*env)->DeleteLocalRef(env, dataArray);
            return;
	}

        /*
         * Most multi font strings aren't. So we delay adding
         * the offset until we know that there are additional
         * strings to be drawn
         */
        if(i > 0)
        {
            if (fdata->flist[j].index_length == 2) {
            } else {
            }
        }
        /*
         * Is it more expensive to copy the whole array, or realloc
         * it to the correct size during conversion? Another choice
         * might be to count the data converted and insert the length
         * at the start of the string.
         */

        j = awtJNI_GetFontDescriptorNumber(env, font, fontDescriptor);
	
	/* 
	 * The strange-looking code that follows this comment is
	 * the result of upstream optimizations. In the array of
	 * alternating font descriptor and buffers, the buffers
	 * contain their length in the first four bytes, a la
	 * Pascal arrays. There is another occurence of these below.
	 */
        stringData = 
	    (unsigned char *)(*env)->GetPrimitiveArrayCritical(env, data,NULL);
        length = (stringData[0] << 24) | (stringData[1] << 16) |
	    (stringData[2] << 8) | stringData[3];
        offsetStringData = (char *)(stringData + (4 * sizeof(char)));

        if (fdata->flist[j].load == 0) {
            xf = loadFont(awt_display,
                          fdata->flist[j].xlfd, size * 10);
            if (xf == NULL) {
	        (*env)->ReleasePrimitiveArrayCritical(env, data, stringData, 
						      JNI_ABORT);
		(*env)->DeleteLocalRef(env, fontDescriptor);
		(*env)->DeleteLocalRef(env, data);
                continue;
            }
            fdata->flist[j].load = 1;
            fdata->flist[j].xfont = xf;
            if (xf->min_byte1 == 0 && xf->max_byte1 == 0)
                fdata->flist[j].index_length = 1;
            else
                fdata->flist[j].index_length = 2;
        }
        xf = fdata->flist[j].xfont;
        XSetFont(awt_display, gdata->gc, xf->fid);
        if (fdata->flist[j].index_length == 2) {
            XDrawString16(awt_display,
                          gdata->drawable,
                          gdata->gc,
                          X, Y,
                          (XChar2b *)offsetStringData, length/2);

            X += XTextWidth16(xf, (XChar2b *)offsetStringData, length/2);

        } else {
            XDrawString(awt_display,
                        gdata->drawable,
                        gdata->gc,
                        X, Y,
                        offsetStringData, length);

            X += XTextWidth(xf, offsetStringData, length);
        }

        (*env)->ReleasePrimitiveArrayCritical(env, data, stringData, JNI_ABORT);
	(*env)->DeleteLocalRef(env, fontDescriptor);
	(*env)->DeleteLocalRef(env, data);
    }
    (*env)->DeleteLocalRef(env, dataArray);
}

/*
 * get multi font string width with multiple X11 font
 *
 * ASSUMES: We are not running on a privileged thread
 */
long awtJNI_GetMFStringWidth(JNIEnv * env, jcharArray s, int offset, int sLength, jobject font)
{
    char *err;
    unsigned char *stringData;
    char *offsetStringData;
    int stringCount,i;
    int size;
    struct FontData *fdata;
    jobject fontDescriptor;
    jbyteArray data;
    int j;
    int width = 0;
    int length;
    XFontStruct *xf;
    jobjectArray dataArray;

    assert(!awt_currentThreadIsPrivileged(env));

    if ((*env)->EnsureLocalCapacity(env, 3) < 0)
        return 0;

    
    if (!JNU_IsNull(env, s) && !JNU_IsNull(env, font))
    {
        jobject peer;

        peer = (*env)->CallObjectMethod(env,font,fontIDs.getPeer);
        
        dataArray = (*env)->CallObjectMethod(
				 env,
				 peer,
				 platformFontIDs.makeConvertedMultiFontChars,
				 s, offset, sLength);

        if ((*env)->ExceptionOccurred(env))
        {
            (*env)->ExceptionDescribe(env);
            (*env)->ExceptionClear(env);
        }

	(*env)->DeleteLocalRef(env, peer);

        if(dataArray == 0)
        {
            return 0;
        }
    } else {
        return 0;
    }

    fdata = awtJNI_GetFontData(env, font, &err);

    stringCount = (*env)->GetArrayLength(env, dataArray);

    size = (*env)->GetIntField(env, font, fontIDs.size);

    for (i = 0; i < stringCount; i+=2)
    {
        fontDescriptor = (*env)->GetObjectArrayElement(env, dataArray, i);
        data = (*env)->GetObjectArrayElement(env, dataArray, i + 1);

        /* Bail if we've finished */
        if(fontDescriptor == 0 || data == 0) {
	    (*env)->DeleteLocalRef(env, fontDescriptor);
	    (*env)->DeleteLocalRef(env, data);
            break;
	}
        
        j = awtJNI_GetFontDescriptorNumber(env, font, fontDescriptor);

        if (fdata->flist[j].load == 0) {
            xf = loadFont(awt_display,
                          fdata->flist[j].xlfd, size * 10);
            if (xf == NULL) {
	        (*env)->DeleteLocalRef(env, fontDescriptor);
	        (*env)->DeleteLocalRef(env, data);
                continue;
            }
            fdata->flist[j].load = 1;
            fdata->flist[j].xfont = xf;
            if (xf->min_byte1 == 0 && xf->max_byte1 == 0)
                fdata->flist[j].index_length = 1;
            else
                fdata->flist[j].index_length = 2;
        }
        xf = fdata->flist[j].xfont;

        stringData =
	    (unsigned char *)(*env)->GetPrimitiveArrayCritical(env, data,NULL);
        length = (stringData[0] << 24) | (stringData[1] << 16) |
            (stringData[2] << 8) | stringData[3];
        offsetStringData = (char *)(stringData + (4 * sizeof(char)));

        if (fdata->flist[j].index_length == 2) {
            width += XTextWidth16(xf, (XChar2b *)offsetStringData, length/2);
        } else {
            width += XTextWidth(xf, offsetStringData, length);
        }

        (*env)->ReleasePrimitiveArrayCritical(env, data, stringData, JNI_ABORT);
	(*env)->DeleteLocalRef(env, fontDescriptor);
	(*env)->DeleteLocalRef(env, data);
    }
    (*env)->DeleteLocalRef(env, dataArray);

    return width;
}
