/*
 * @(#)drawsurface.c	1.5 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <DPS/dpsXclient.h>
#include "X11DrawSurfaceTracker.h"

static float matrix[] = {1.0, 0.0, 0.0, 1.0, 0.0, 0.0};

static void
setupTrueColorColormap(Colormap cmap, XStandardColormap *gray,
	XStandardColormap *ccube)
{
	gray->colormap = cmap;
	gray->red_max = 0xFF;
	gray->red_mult = 0x010101;
	gray->base_pixel = 0;

	ccube->colormap = cmap;
	ccube->red_max = 0xFF;
	ccube->red_mult = 1;
	ccube->green_max = 0xFF;
	ccube->green_mult = 0x0100;
	ccube->blue_max = 0xFF;
	ccube->blue_mult = 0x010000;
	ccube->base_pixel = 0;
}

static void
setupStaticGrayColormap(Colormap cmap, XStandardColormap *gray,
	XStandardColormap *ccube, int depth)
{
	if (depth == 1) {
	    /* REMIND: This assumes 0 is white (our hardware) */
	    gray->colormap = cmap;
	    gray->red_max = 1;
	    gray->red_mult = -1;
	    gray->base_pixel = 1;
	} else {
	    gray->colormap = cmap;
	    gray->red_max = 255;
	    gray->red_mult = 1;
	    gray->base_pixel = 0;
	}

	memset(ccube, 0, sizeof(*ccube));
}

static void
setupColorInfo(XVisualInfo *vinfo, Colormap cmap, 
	XStandardColormap *gray, XStandardColormap *ccube)
{
	if (vinfo->depth == 24) {
	    setupTrueColorColormap(cmap, gray, ccube);
	} else if (vinfo->class == StaticGray) {
	    setupStaticGrayColormap(cmap, gray, ccube, vinfo->depth);
	} else if (vinfo->class == PseudoColor) {
	    /* punt */
	}
}

void
X11DrawSurfaceTracker_setup(HX11DrawSurfaceTracker *this,
			    long ldpy, long ldraw,
			    long lvisid, long lcmap,
			    long depth)
{
    Display *dpy;
    Drawable draw;
    Colormap cmap;
    VisualID visid;
    GC gc;
    DPSContext ctx;

    if (ldpy == 0 || ldraw == 0 || lcmap == 0 || lvisid == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }
    dpy = (Display *) ldpy;
    draw = (Drawable) ldraw;
    cmap = (Colormap) lcmap;
    visid = (VisualID) lvisid;

    ctx = (DPSContext) unhand(this)->dpsctx;
    if (ctx != 0) {
	DPSDestroyContext(ctx);
	unhand(this)->dpsctx = 0;
    }
    gc = (GC) unhand(this)->gc;
    if (ctx != 0) {
	XFreeGC(dpy, gc);
	unhand(this)->gc = 0;
    }

    gc = XCreateGC(dpy, draw, 0, NULL);
    if (visid == XVisualIDFromVisual(DefaultVisual(dpy,
						   DefaultScreen(dpy)))) {
	ctx = XDPSCreateSimpleContext(dpy, draw, gc, 0, 0,
				      DPSDefaultTextBackstop,
				      DPSDefaultErrorProc,
				      NULL);
    } else {
	XStandardColormap grayramp, ccube;
	XVisualInfo viTmp, *pVI;
	int numvi;

	viTmp.visualid = visid;
	viTmp.depth = depth;
	pVI = XGetVisualInfo(dpy, VisualIDMask | VisualDepthMask,
			     &viTmp, &numvi);
	if (!pVI) {
	    /* This should never happen. */
	    SignalError(0, JAVAPKG "InternalError",
			"Can't find default visual information");
	    return;
	}
	setupColorInfo(pVI, cmap, &grayramp, &ccube);
	ctx = XDPSCreateContext(dpy, draw, gc, 0, 0,
				0, &grayramp, &ccube, 0,
				DPSDefaultTextBackstop,
				DPSDefaultErrorProc,
				NULL);
	XFree(pVI);
    }
    if (ctx == 0) {
	SignalError(0, JAVAPKG "InternalError", "Cannot get DPS Context");
	return;
    }

    DPSsendfloatarray(ctx, matrix, 6);
    DPSsetmatrix(ctx);

    unhand(this)->gc = (long) gc;
    unhand(this)->dpsctx = (long) ctx;
}

void
X11DrawSurfaceTracker_draw(HX11DrawSurfaceTracker *this,
			   HArrayOfFloat *ctrlpts,
			   float red, float green, float blue)
{
    DPSContext ctx;
    float *cp;

    if (ctrlpts == 0) {
	SignalError(0, JAVAPKG "NullPointerException", 0);
	return;
    }
    if (obj_length(ctrlpts) < 8) {
	SignalError(0, JAVAPKG "ArrayIndexOutOfBoundsException", 0);
	return;
    }
    ctx = (DPSContext) unhand(this)->dpsctx;
    if (ctx == 0) {
	SignalError(0, JAVAPKG "InternalError", "Cannot get DPS Context");
	return;
    }
    cp = unhand(ctrlpts)->body;
    DPSsetrgbcolor(ctx, red, green, blue);
    DPSmoveto(ctx, cp[0], cp[1]);
    DPScurveto(ctx, cp[2], cp[3], cp[4], cp[5], cp[6], cp[7]);
    DPSstroke(ctx);
    DPSFlushContext(ctx);
}
