/*
 * @(#)Win32DrawingSurface.java	1.5 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import sun.awt.PhysicalDrawingSurface;
import java.awt.image.ColorModel;

/**
 * The Win32DrawingSurface interface defines the methods needed to
 * be able to render to a Windows native window.  Only the commonly
 * needed attributes are returned, other attributes can be queried by
 * calling the Windows SDK native API functions.
 *
 * Many of the methods return Java integers which must be cast to the
 * appropriate Windows data types using a cast similar to the pseudocode
 * indicated in the class comments of the various methods.
 *
 * @version 	1.5, 09/21/98
 * @author 	Jim Graham
 */
public interface Win32DrawingSurface extends PhysicalDrawingSurface {
    /**
     * Return the Windows handle for the window
     * represented by this object.
     * If the drawing surface is a DIBSection then this method will
     * return NULL (0).
     * <p>
     * HWND *hWnd = (HWND) getHWnd();
     */
    public int getHWnd();

    /**
     * Return the Windows handle for the bitmap
     * represented by this object.
     * If the drawing surface is a window then this method will
     * return NULL (0).
     * <p>
     * HBITMAP *hBitmap = (HBITMAP) getHBitmap();
     */
    public int getHBitmap();

    /**
     * Return the pointer to the bits for the DIBSection
     * represented by this object.
     * If the drawing surface is not a DIBSection (either a window
     * or another form of Windows bitmap) then this method will
     * return NULL (0).
     * <p>
     * void *ppvBits = (void *) getPBits();
     */
    public int getPBits();

    /**
     * Return the Windows DC handle for the window or bitmap
     * represented by this object.
     * <p>
     * HDC hDC = (HDC) getHDC();
     */
    public int getHDC();

    /**
     * Return the depth of the window or bitmap
     * represented by this object.
     * <p>
     * int depth = getDepth();
     */
    public int getDepth();

    /**
     * Return the Windows palette handle associated with the drawable
     * represented by this object.
     * <p>
     * HPALETTE hPal = (HPALETTE) getHPalette();
     */
    public int getHPalette();

    /**
     * Return the ColorModel of the window or bitmap
     * represented by this object.
     * <p>
     * Hjava_awt_image_ColorModel *cm =
     *     (Hjava_awt_image_ColorModel *) getColorModel();
     */
    public ColorModel getColorModel();
}
