/*
 * @(#)Win32GraphicsDevice.java	1.4 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import java.awt.GraphicsDevice;
import java.awt.GraphicsConfiguration;

/**
 * This is an implementation of a GraphicsDevice object for a single
 * Win32 screen.
 *
 * @see GraphicsEnvironment
 * @see GraphicsConfiguration
 * @version 10 Feb 1997
 */
public class Win32GraphicsDevice extends GraphicsDevice {
    int screen;

    public Win32GraphicsDevice(int screennum) {
	this.screen = screennum;
    }

    /**
     * Returns the type of the graphics device.
     * @see #TYPE_RASTER_SCREEN
     * @see #TYPE_PRINTER
     * @see #TYPE_IMAGE_BUFFER
     */
    public int getType() {
	return TYPE_RASTER_SCREEN;
    }

    /**
     * Returns the identification string associated with this graphics
     * device.
     */
    public String getIDstring() {
	return ":0."+screen;
    }

    GraphicsConfiguration[] configs;

    /**
     * Returns all of the graphics
     * configurations associated with this graphics device.
     */
    public GraphicsConfiguration[] getConfigurations() {
	GraphicsConfiguration[] ret = configs;
	if (ret == null) {
	    int num = getNumConfigs();
	    ret = new GraphicsConfiguration[num];
	    for (int i = 0; i < num; i++) {
		ret[i] = new Win32GraphicsConfig(this, i);
	    }
	    configs = ret;
	}
	return ret;
    }

//    private native int getNumConfigs();
    private int getNumConfigs() {
        return 1;
    }

    /**
     * Returns the default graphics configuration
     * associated with this graphics device.
     */
    public GraphicsConfiguration getDefaultConfiguration() {
	return getConfigurations()[0];
    }
}
