/*
 * @(#)Win32GraphicsEnvironment.java	1.7 98/09/24
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt;

import sun.java2d.SunGraphicsEnvironment;
import java.awt.GraphicsDevice;
import java.awt.Toolkit;
import java.awt.print.PrinterJob;
import sun.awt.windows.WPrinterJob;

/**
 * This is an implementation of a GraphicsEnvironment object for the
 * default local GraphicsEnvironment used by the JavaSoft JDK in Win32
 * environments.
 *
 * @see GraphicsDevice
 * @see GraphicsConfiguration
 * @version 10 Feb 1997
 */

public class Win32GraphicsEnvironment extends SunGraphicsEnvironment {
    static {
	java.security.AccessController.doPrivileged(
		    new sun.security.action.LoadLibraryAction("awt"));
//	initDisplay();
    }

//    private static native void initDisplay();

    public Win32GraphicsEnvironment() {
    }

//    protected native int getNumScreens();
    protected int getNumScreens() {
		return 1;
	}

    protected GraphicsDevice makeScreenDevice(int screennum) {
		return new Win32GraphicsDevice(screennum);
    }

    /* Fix: 4175044
    *  allow platform to override version extension
    *  for font.properties files
    */
    protected String getOsVersion() {
        String osname = System.getProperty("os.name");
        String osversion = System.getProperty("os.version");
        if (osname.equals("Windows NT")) {
            return "NT" + osversion;
        }
        return osversion;
    }

    /**
     * Gets a <code>PrinterJob</code> object suitable for the
     * the current platform.
     * @return    a <code>PrinterJob</code> object.
     * @see       java.awt.PrinterJob
     * @since     JDK1.2
     */
    public PrinterJob getPrinterJob() {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkPrintJobAccess();
	}

	return new WPrinterJob();
    }
}
