/*
 * @(#)MTextAreaPeer.java	1.17 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.win32;

import java.awt.*;
import java.awt.peer.*;

public class MTextAreaPeer extends MComponentPeer implements TextAreaPeer {

    int verticalScrollWidth;		// set by create()
    int horizontalScrollHeight;		// set by create()

    native void create(MComponentPeer parent);

    void initialize() {
	TextArea txt = (TextArea)target;
	String	text;

	if ((text = txt.getText()) != null) {
	    setText(text);
	}
	select(txt.getSelectionStart(), txt.getSelectionEnd());
	setEditable(txt.isEditable());

	super.initialize();
    }

    public MTextAreaPeer(TextArea target) {
	super(target);
    }

    public void setEditable(boolean editable) {
	pSetEditable(editable);
	setBackground(target.getBackground());
    }
    public void setBackground(Color c) {
	TextArea t = (TextArea)target;
	if (t.isEditable()) {
	    c = c.brighter();
	}
	super.setBackground(c);
    }

    public native void pSetEditable(boolean e);
    public native void select(int selStart, int selEnd);
    public native int getSelectionStart();
    public native int getSelectionEnd();
    public native void setText(String txt);
    public native String getText();

    /* This should eventually be a direct native method. */
    public void insert(String txt, int pos) {
	insertText(txt, pos);
    }

    /* This should eventually be a direct native method. */
    public void replaceRange(String txt, int start, int end) {
	replaceText(txt, start, end);
    }

    public Dimension getMinimumSize() {
	return getMinimumSize(10, 60);
    }

    public Dimension getPreferredSize(int rows, int cols) {
	return getMinimumSize(rows, cols);
    }

    public Dimension getMinimumSize(int rows, int cols) {
	FontMetrics fm = getFontMetrics(target.getFont());
	return new Dimension(fm.charWidth('0') * cols + verticalScrollWidth, 
                             fm.getHeight() * rows + horizontalScrollHeight);
    }

    public boolean isFocusTraversable() {
	return true;
    }

    /**
     * DEPRECATED but, for now, still called by insert(String, int).
     */
    public native void insertText(String txt, int pos);

    /**
     * DEPRECATED but, for now, still called by replaceRange(String, int, int).
     */
    public native void replaceText(String txt, int start, int end);

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize(int rows, int cols) {
	return getMinimumSize(rows, cols);
    }

    /**
     * DEPRECATED
     */
    public Dimension preferredSize(int rows, int cols) {
	return getPreferredSize(rows, cols);
    }

}
