/*
 * @(#)MWindowPeer.java	1.12 98/09/21
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.win32;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;

class MWindowPeer extends MPanelPeer implements WindowPeer {
    Insets	insets = new Insets(0,0,0,0);

    static Vector allWindows = new Vector();

    native void create(MComponentPeer parent);
    native void pShow();
    native void pHide();
    native void pReshape(int x, int y, int width, int height);

    // XXX: need separate call to set insets because the insets data member
    // isn't set until after the super constructor is called. This was solved
    // differently on the Motif side but this solution seems cleaner.
    native void setInsets(Insets i);

    native void pDispose();

    MWindowPeer(Window target) {
	super(target);

	// see comment above for setInsets
	setInsets(insets);
	allWindows.addElement(this);
	Font f = target.getFont();
	if (f == null) {
	    f = new Font("Dialog", Font.PLAIN, 10);
	    target.setFont(f);
	    setFont(f);
	}
	Color c = target.getBackground();
	if (c == null) {
	    target.setBackground(Color.lightGray);
	    setBackground(Color.lightGray);
	}
	c = target.getForeground();
	if (c == null) {
	    target.setForeground(Color.black);
	    setForeground(Color.black);
	}
    }

    public void dispose() {
	allWindows.removeElement(this);
	super.dispose();
    }

    public void toFront() {
	show();
    }
    public void toBack() {
    }

/* ----------- INSANITY STARTS BELOW THIS LINE --------------- */
    public Insets getInsets() {
	return insets;
    }

    public void handleQuit() {
	target.postEvent(new Event(target, Event.WINDOW_DESTROY, null));
    }

    public void handleIconify() {
	target.postEvent(new Event(target, Event.WINDOW_ICONIFY, null));
    }

    public void handleDeiconify() {
	target.postEvent(new Event(target, Event.WINDOW_DEICONIFY, null));
    }

    /**
     * Called to inform the Dialog that its size has changed and it
     * should layout its children.
     */
    public synchronized void handleResize(int width, int height) {
	//target.resize(width, height);
	target.invalidate();
	target.validate();
	target.repaint();
    }

    /**
     * DEPRECATED:  Replaced by getInsets().
     */
    public Insets insets() {
	return getInsets();
    }

}
