/*
 * @(#)Win32Graphics.java	1.32 98/06/29
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.win32;


import java.awt.*;
import java.io.*;
import java.util.*;
import java.java2d.AffineTransform;
import java.java2d.BasicStroke;
import java.java2d.BufferedImage;
import java.java2d.BufferedImageOp;
import java.java2d.Composite;
import java.java2d.GraphicsConfiguration;
import java.java2d.Paint;
import java.java2d.Path;
import java.java2d.Stroke;
import java.java2d.Transform;
import java.awt.image.ImageObserver;
import sun.awt.image.ImageRepresentation;

/**
 * Win32Graphics is an object that encapsulates a graphics context for a
 * particular canvas.
 *
 * @version 1.32 06/29/98
 * @author Sami Shaio
 */

public class Win32Graphics extends SunGraphics2D implements AWTFinalizeable {
    int		pData;
    Color	foreground;
    Font	font;
    int		originX;	// REMIND: change to float
    int		originY;	// REMIND: change to float

    private native void createFromComponent(MComponentPeer w);
    private native void createFromGraphics(Win32Graphics g);
    private native void createFromPrintJob(Win32PrintJob pj);
    private native void imageCreate(ImageRepresentation ir);
    private native void pSetFont(Font f);
    private native void pSetForeground(Color c);

    private sun.awt.image.Image		image;

    Win32Graphics(Win32PrintJob pj) {
	createFromPrintJob(pj);
    }
    Win32Graphics(Win32Graphics g) {
	createFromGraphics(g);
    }
    Win32Graphics(MComponentPeer comp) {
	createFromComponent(comp);
    }

    /**
     * Create a new Win32Graphics Object based on this one.
     */
    public Graphics create() {
	Win32Graphics g = new Win32Graphics(this);
	g.foreground = foreground;
	g.font = font;
	g.originX = originX;
	g.originY = originY;
	g.image = image;
	return g;
    }

    /**
     * Translate
     */
    public void translate(int x, int y) {
	originX += x;
	originY += y;
    }

    /**
     * Disposes of this Win32Graphics context. It cannot be used after being
     * disposed.
     */
    public native void dispose();


    AWTFinalizeable finalnext;

    public void finalize() {
	AWTFinalizer.addFinalizeable(this);
    }

    public void doFinalization() {
	dispose();
    }

    public void setNextFinalizeable(AWTFinalizeable o) {
	finalnext = o;
    }

    public AWTFinalizeable getNextFinalizeable() {
	return finalnext;
    }

    public void setFont(Font font) {
	if ((font != null) && (this.font != font)) {
	    this.font = font;
	    pSetFont(font);
	}
    }
    public Font getFont() {
	return font;
    }

    /**
     * Gets font metrics for the given font.
     */
    public FontMetrics getFontMetrics(Font font) {
	return Win32FontMetrics.getFontMetrics(font);
    }

    /**
     * Sets the foreground color.
     */
    public void setColor(Color c) {
	if ((c != null) && (c != foreground)) {
	    foreground = c;
	    pSetForeground(c);
	}
    }
    public Color getColor() {
	return foreground;
    }

    /**
     * Sets the paint mode to overwrite the destination with the
     * current color. This is the default paint mode.
     */
    public native void setPaintMode();

    /**
     * Sets the paint mode to alternate between the current color
     * and the given color.
     */
    public native void setXORMode(Color c1);

    /**
     * Gets the current clipping area
     */
    public Rectangle getClipBounds() {
	Rectangle clip = new Rectangle();
	getClipRect(clip);
	return clip;
    }
    native void getClipRect(Rectangle clip);

    private native void changeClip(int X, int Y, int W, int H, boolean set);

    /** Crops the clipping rectangle for this Win32Graphics context. */
    public void clipRect(int X, int Y, int W, int H) {
	changeClip(X, Y, W, H, false);
    }

    /** Sets the clipping rectangle for this Win32Graphics context. */
    public void setClip(int X, int Y, int W, int H) {
	changeClip(X, Y, W, H, true);
    }

    /** Returns a Shape object representing the clip. */
    public Shape getClip() {
	return getClipBounds();
    }

    /** Sets the clip to a Shape (only Rectangle allowed). */
    public void setClip(Shape clip) {
	if (clip instanceof Rectangle) {
	    Rectangle r = (Rectangle) clip;
	    changeClip(r.x, r.y, r.width, r.height, true);
	} else {
	    throw new IllegalArgumentException("setClip(Shape) only supports Rectangle objects");
	}
    }

    /** Clears the rectangle indicated by x,y,w,h. */
    public native void clearRect(int x, int y, int w, int h);

    /** Fills the given rectangle with the foreground color. */
    public native void fillRect(int X, int Y, int W, int H);

    /** Draws the given rectangle. */
    public native void drawRect(int X, int Y, int W, int H);


    /** Draws the given string. */
    public native void drawString(String str, int x, int y);

    /** Draws the given character array. */
    public native void drawChars(char data[], int offset, int length, int x, int y);

    /** Draws the given byte array. */
    public native void drawBytes(byte data[], int offset, int length, int x, int y);

    /** Draws the given string and returns the length of the drawn
      string in pixels.  If font is not set then returns -1. */
    public native int drawStringWidth(String str, int x, int y);

    /** Draws the given character array and return the width in
      pixels. If font is not set then returns -1. */
    public native int drawCharsWidth(char data[], int offset, int length, int x, int y);

    /** Draws the given character array and return the width in
      pixels. If font is not set then returns -1. */
    public native int drawBytesWidth(byte data[], int offset, int length, int x, int y);

    /** Draws the given line. */
    public native void drawLine(int x1, int y1, int x2, int y2);

    /**
     * Draws an image at x,y in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {
	Win32Image win32img = (Win32Image) img;
	if (win32img.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = win32img.getImageRep(-1, -1);
	return ir.drawImage(this, x, y, null, observer);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}
	Win32Image win32img = (Win32Image) img;
	if (win32img.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = win32img.getImageRep(-1, -1);
	if (width < 0) {
	    if (height < 0) {
		width = ir.width;
		height = ir.height;
	    } else {
		width = height * ir.width / ir.height;
	    }
	} else if (height < 0) {
	    height = width * ir.height / ir.width;
	}
	return ir.drawStretchImage(this,
				   x, y, x+width, y+height,
				   0, 0, ir.width, ir.height,
				   null, observer);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a solid background
     * color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, Color bg,
			     ImageObserver observer) {
	Win32Image win32img = (Win32Image) img;
	if (win32img.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = win32img.getImageRep(-1, -1);
	return ir.drawImage(this, x, y, bg, observer);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * solid background color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     Color bg, ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}
	Win32Image win32img = (Win32Image) img;
	if (win32img.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = win32img.getImageRep(-1, -1);
	if (width < 0) {
	    if (height < 0) {
		width = ir.width;
		height = ir.height;
	    } else {
		width = height * ir.width / ir.height;
	    }
	} else if (height < 0) {
	    height = width * ir.height / ir.width;
	}
	return ir.drawStretchImage(this,
				   x, y, x+width, y+height,
				   0, 0, ir.width, ir.height,
				   bg, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle
     * in nonblocking mode with a callback object.
     */
    /**
     * Draws a subrectangle of an image scaled to a destination rectangle
     * in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer) {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
	Win32Image win32img = (Win32Image) img;
	if (win32img.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = win32img.getImageRep(-1, -1);
	return ir.drawStretchImage(this,
				   dx1, dy1, dx2, dy2,
				   sx1, sy1, sx2, sy2,
				   null, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle in
     * nonblocking mode with a solid background color and a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor,
			     ImageObserver observer)
    {
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}
	Win32Image win32img = (Win32Image) img;
	if (win32img.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = win32img.getImageRep(-1, -1);
	return ir.drawStretchImage(this,
				   dx1, dy1, dx2, dy2,
				   sx1, sy1, sx2, sy2,
				   bgcolor, observer);
    }

    /** REMIND : works only with as special case of TYPE_4BYTE_ABGR
        where the alpha data storage pointer can be casted to an unsigned
        int points to all channels.
        This method needs a native method to bitblt the data onto the
        screen called bitbltBIToScreen.
    **/
    public void drawImage(BufferedImage bufImg,
                          BufferedImageOp op,
                          int x,
                          int y) {

        if (op != null) {
            throw new NoSuchMethodError ("Not implemented to take an OP yet");
        } else {
            int scanStride;
            int pixStride;
            int chanOffs;
            int w = bufImg.getWidth();
            int h = bufImg.getHeight();
            byte[] data;

            Tile t = bufImg.getTile();
            ByteDiscreteChannel dc;
            dc = (ByteDiscreteChannel)t.getChannel(0);   /* the alpha channel */
            //scanStride = dc.getScanlineStride();
            //pixStride  = dc.getPixelStride();
            //chanOffs   = dc.getChannelOffset();
            data       = dc.getDataStorage();

            bitbltBIToScreen(data, x, y, w, h);

        }
    }

    /**
     * Copies an area of the canvas that this graphics context paints to.
     * @param X the x-coordinate of the source.
     * @param Y the y-coordinate of the source.
     * @param W the width.
     * @param H the height.
     * @param dx the x-coordinate of the destination.
     * @param dy the y-coordinate of the destination.
     */
    public native void copyArea(int X, int Y, int W, int H, int dx, int dy);


    /** Draws a rounded rectangle. */
    public native void drawRoundRect(int x, int y, int w, int h,
				     int arcWidth, int arcHeight);

    /** Draws a filled rounded rectangle. */
    public native void fillRoundRect(int x, int y, int w, int h,
				     int arcWidth, int arcHeight);

    /** Draws a polygon defined by an array of x points and y points */
    public native void drawPolygon(int xPoints[], int yPoints[], int nPoints);

    /** Fills a polygon with the current fill mask */
    public native void fillPolygon(int xPoints[], int yPoints[], int nPoints);

    /** Draws an oval to fit in the given rectangle */
    public native void drawOval(int x, int y, int w, int h);

    /** Fills an oval to fit in the given rectangle */
    public native void fillOval(int x, int y, int w, int h);

    /**
     * Draws an arc bounded by the given rectangle from startAngle to
     * endAngle. 0 degrees is a vertical line straight up from the
     * center of the rectangle. Positive angles indicate clockwise
     * rotations, negative angle are counter-clockwise.
     */
    public native void drawArc(int x, int y, int w, int h,
			       int startAngle,
			       int endAngle);

    /** fills an arc. arguments are the same as drawArc. */
    public native void fillArc(int x, int y, int w, int h,
			       int startAngle,
			       int endAngle);

    public native void bitblBIToScreen(byte[] data, int x, int y, int w, int h);


    public String toString() {
	return getClass().getName() + "[" + originX + "," + originY + "]";
    }

    /* Outline the given region. */
    //public native void drawRegion(Region r);

    /* Fill the given region. */
    //public native void fillRegion(Region r);
}
