/*
 * @(#)WChoicePeer.java	1.20 98/07/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.ItemEvent;

class WChoicePeer extends WComponentPeer implements ChoicePeer {

    // WComponentPeer overrides

    public Dimension getMinimumSize() {
	FontMetrics fm = getFontMetrics(((Choice)target).getFont());
	Choice c = (Choice)target;
	int w = 0;
	for (int i = c.getItemCount() ; i-- > 0 ;) {
	    w = Math.max(fm.stringWidth(c.getItem(i)), w);
	}
	return new Dimension(28 + w, Math.max(fm.getHeight() + 6, 15));
    }
    public boolean isFocusTraversable() {
	return true;
    }

    // ChoicePeer implementation

    public native void select(int index);

    public void add(String item, int index) {
      	addItem(item, index);
    }

    public native void remove(int index);

    /**
     * DEPRECATED, but for now, called by add(String, int).
     */
    public native void addItem(String item, int index);

    public native void reshape(int x, int y, int width, int height);

    // Toolkit & peer internals

    WChoicePeer(Choice target) {
	super(target);
    }

    native void create(WComponentPeer parent);

    void initialize() {
	Choice opt = (Choice)target;
	int itemCount = opt.getItemCount();
	for (int i=0; i < itemCount; i++) {
	    add(opt.getItem(i), i);
	}
	if (itemCount > 0 && opt.getSelectedIndex() >= 0) {
	    select(opt.getSelectedIndex());
	}
	super.initialize();
    }

    void clearRectBeforePaint(Graphics g, Rectangle r) {
        // Overload to do nothing for native components
    }

    // native callbacks

    void handleAction(final int index) {
	final Choice c = (Choice)target;
	WToolkit.executeOnEventHandlerThread(c, new Runnable() {
	    public void run() {
		c.select(index);
		postEvent(new ItemEvent(c, ItemEvent.ITEM_STATE_CHANGED,
                                c.getItem(index), ItemEvent.SELECTED));
	    }
	});
    }

    int getDropDownHeight() {
	Choice c = (Choice)target;
	FontMetrics fm = getFontMetrics(c.getFont());
        int maxItems = Math.min(c.getItemCount(), 8);
	return fm.getHeight() * maxItems;
    }

    /**
     * DEPRECATED
     */
    public Dimension minimumSize() {
	    return getMinimumSize();
    }

}
