/*
 * @(#)WClipboard.java	1.10 98/10/02
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.awt.AWTException;
import java.awt.datatransfer.*;

/**
 * A class which interfaces with the Windows clipboard in order
 * to support data transfer via clipboard operations.
 */
public class WClipboard extends Clipboard {

    static {
        init();
    }

    public WClipboard() {
        super("System");
    }

    public synchronized void setContents(Transferable contents,
                                         ClipboardOwner owner) {
	if (this.owner != null && this.owner != owner) {
	    this.owner.lostOwnership(this, this.contents);
	}
	this.owner = owner;
	this.contents = contents;

        if (contents instanceof StringSelection) {
            setClipboardText((StringSelection)contents);
        } else {
            setClipboardObject(contents);
        }
    }

    public synchronized Transferable getContents(Object requestor) {
        if (contents != null) {
            return contents;
        }

        // No Java contents -- see if any text is on system clipboard.
        String s = getClipboardText();
        return (s == null) ? null : new StringSelection(s);
    }

    public synchronized void lostSelectionOwnership() {
        if (this.owner != null) {
	    this.owner.lostOwnership(this, this.contents);
	    this.owner = null;
	}

	// fix 4177171 : always clear cached contents when losing ownership
	this.contents = null;
    }

    // Register Java String clipboard format.
    static private native void init();

    // Replace the current clipboard contents with ss.
    private native void setClipboardText(StringSelection ss);

    // Replace the current clipboard contents with this Java object.
    private native void setClipboardObject(Object obj);

    // Get the current clipboard contents.
    private native String getClipboardText();
}
