/*
 * @(#)WCustomCursor.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import sun.awt.CustomCursor;
import java.awt.*;
import java.awt.image.*;
import sun.awt.image.ImageRepresentation;

/**
 * A class to encapsulate a custom image-based cursor.  
 *
 * @see Component#setCursor
 * @version 	1.3, 09/21/98
 * @author 	ThomasBall
 */
public class WCustomCursor extends CustomCursor {

    static {
        initIDs();
    }

    int pData;

    public WCustomCursor(Image cursor, Point hotSpot, String name) 
            throws IndexOutOfBoundsException {
        super(cursor, hotSpot, name);
    }

    protected void createNativeCursor(int[] pixels, int width, int height,
                                      int xHotSpot, int yHotSpot) {
        // Create two bit masks for a Win32 cursor.
        byte[] xorMask = new byte[width * height / 8];
        byte[] andMask = new byte[width * height / 8];
        int npixels = pixels.length;
        for (int i = 0; i < npixels; i++) {
            int ibyte = i / 8;
            int omask = 1 << (7 - (i % 8));
            if ((pixels[i] & 0xff000000) == 0) {
                // Transparent bit
                andMask[ibyte] |= omask;
            } else if ((pixels[i] & 0xffffff) > 0) {
                // White bit
                xorMask[ibyte] |= omask;
            }
        }
        createCursor(xorMask, andMask, width, height, xHotSpot, yHotSpot);
    }

    private native void createCursor(byte[] xorMask, byte[] andMask, int width, 
                                     int height, int xHotSpot, int yHotSpot);

    protected ImageFilter getImageFilter() {
        return new WCursorImageFilter();
    }
    
    private static class WCursorImageFilter extends RGBImageFilter {
        WCursorImageFilter() {
            canFilterIndexColorModel = true;
        }
        public int filterRGB(int x, int y, int rgb) {
            return rgb;  // we support color cursors.
        }
    }

    /**
     * Return the current value of SM_CXCURSOR.
     */
    static native int getCursorWidth();

    /**
     * Return the current value of SM_CYCURSOR.
     */
    static native int getCursorHeight();

    protected native void finalize() throws Throwable;

    /**
     * Initialize JNI field IDs
     */
    private static native void initIDs();
}
