/*
 * @(#)WDialogPeer.java	1.10 98/07/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;
import sun.awt.im.*;

class WDialogPeer extends WWindowPeer implements DialogPeer {
    class ModalLock extends Object {
	synchronized void waitModal() {
	    try {
		wait();
	    } catch(InterruptedException ie) {
	    }
	}

	synchronized void endModal() {
	    notifyAll();
	}
    }

    ModalLock	modalLock = new ModalLock();
    // DialogPeer implementation

    // Toolkit & peer internals

    WDialogPeer(Dialog target) {
	super(target);

	InputMethodManager imm = InputMethodManager.getInstance();
	String menuString = imm.getTriggerMenuString();
	if (menuString != null)
	{ 
	  pSetIMMOption(menuString);
	}
    }

    native void create(WComponentPeer parent);
    native void showModal();
    native void endModal();

    void initialize() {
        super.initialize();

        Dialog target = (Dialog)this.target;

	if (target.getTitle() != null) {
	    setTitle(target.getTitle());
	}
	setResizable(target.isResizable());
    }

    public void show() {
	if (((Dialog)target).isModal()) {
	    showModal();
	    modalLock.waitModal();
        } else {
            super.show();
        }
    }

    public void hide() {
	if (((Dialog)target).isModal()) {
            endModal();
	    modalLock.endModal();
        } else {
            super.hide();
        }
    }

    /* Set the appropriate default dialog color, depending on whether
     * Win95/NT 4.0 is running or not.  Since these values are class
     * constants, it's easier (and safer) to do this in Java than C++.
     */
    private void setDefaultColor(boolean version4) {
	final Color c = version4 ? SystemColor.control : SystemColor.window;
	WToolkit.executeOnEventHandlerThread(target, new Runnable() {
	    public void run() {
		// Because Dialog.setBackground() may be overloaded by
		// client code, we can't do this synchronously (we mustn't
		// allow the toolkit thread to ever call client code).
		// So, we do it through the client's EventQueue.
		((Dialog)target).setBackground(c);
	    }
	});
    }

    native void pSetIMMOption(String option);
    void notifyIMMOptionChange(){
      InputMethodManager.getInstance().notifyChangeRequest((Component)target);
    }  
}
