/*
 * @(#)WDragSourceContextPeer.java	1.13 98/08/28
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.awt.event.InputEvent;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;

import java.awt.peer.ComponentPeer;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorMap;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceContext;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.InvalidDnDOperationException;

import java.awt.dnd.peer.DragSourceContextPeer;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.StringBufferInputStream;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.Map;

import sun.awt.windows.WToolkit;
import sun.awt.windows.WDropTargetContextPeer;

import java.lang.reflect.InvocationTargetException;

import javax.swing.SwingUtilities;

import java.rmi.MarshalledObject;
import java.rmi.MarshalException;

/**
 * <p>
 * TBC
 * </p>
 *
 * @version 1.13
 * @since JDK1.2
 *
 */

final class WDragSourceContextPeer implements DragSourceContextPeer {

    /**
     * The Listener Dispatcher
     */

    static private WDragSourceContextPeerDispatcher listenerDispatcher = new WDragSourceContextPeerDispatcher();

    /**
     * provide a WDragSourceContextPeer to the Toolkit object ...
     */

     static DragSourceContextPeer getDragSourceContextPeer(DragGestureEvent dge) {
	return new WDragSourceContextPeer(dge);
     }

    /**
     * construct a new WDragSourceContextPeer
     */

    private WDragSourceContextPeer(DragGestureEvent dge) {
    	super();

	trigger    = dge;
	component  = dge.getComponent();
	dragSource = dge.getDragSource(); 
    }


    /**
     * initiate a DnD operation ...
     */

    public void startDrag(DragSourceContext dsc, Cursor c, Image di, Point ioff) throws InvalidDnDOperationException {
	dragSourceContext = dsc;
	cursor 		  = c;
	actions           = trigger.getSourceAsDragGestureRecognizer().getSourceActions();

	synchronized(this) {
	    Transferable trans   = dragSourceContext.getTransferable();
	    DataFlavor[] dfs     = trans.getTransferDataFlavors();
	    Map          natives = dragSource.getFlavorMap().getNativesForFlavors(dfs);

	    /*
	     * trim the flavors that are not externally transferable
	     */

	    for (int i = 0 ; i < dfs.length; i++) {
		if (!(dfs[i].isRepresentationClassInputStream()  ||
		      dfs[i].isRepresentationClassSerializable() ||
		      dfs[i].isRepresentationClassRemote()       ||
		      DataFlavor.javaFileListFlavor.equals(dfs[i]))) natives.remove(dfs[i]);
	    }

	    if (nativeCtxt != 0)
		throw new InvalidDnDOperationException("drag in progress");

	    try {
	        nativeCtxt = createDragSource(trigger.getTriggerEvent(),
					      actions,
					      natives.values().toArray()
			     );
	    } catch (Exception e) {
		e.printStackTrace();
		throw new InvalidDnDOperationException("failed to create native peer");
	    }

	    if (nativeCtxt == 0) 
		throw new InvalidDnDOperationException("failed to create native peer");
	    WDropTargetContextPeer.setCurrentJVMLocalSourceTransferable(trans);

	    try {

		/*
		 * create a new thread (later) to call into the modal windows
		 * code ... 
		 */

	    	((Thread)(new Thread() {

		    /**
  		     * run
		     */

	            public void run() {
		    	try {
	    		    doDragDrop(WDragSourceContextPeer.this.nativeCtxt,
				       WDragSourceContextPeer.this.cursor
			    );
		        } catch (Exception e) {
		    	    e.printStackTrace();
			} finally {
			    release(WDragSourceContextPeer.this.nativeCtxt);

			    WDragSourceContextPeer.this.nativeCtxt = 0;

			    WDropTargetContextPeer.setCurrentJVMLocalSourceTransferable(null);
		        }
		    }
	        })).start();
	    } catch (Exception e) {
	        e.printStackTrace();

	        release(nativeCtxt);
	        nativeCtxt = 0;

	        throw new InvalidDnDOperationException("failed to start Thread");
	    }
	}
    }

    /**
     *
     */

    public Cursor getCursor() { return cursor; }

    /**
     *
     */

    public void setCursor(Cursor c) throws InvalidDnDOperationException {
	if (cursor == null || !cursor.equals(c)) {
	    cursor = c;
	    if (nativeCtxt != 0) setNativeCursor(nativeCtxt, cursor);
	}
    }


    /**
     *
     */

    public void transferablesFlavorsChanged() {
    }

    /**
     *
     */

    private void dragEnter(int dropAction, int targetActions, int modifiers) {
	listenerDispatcher.deliverEnter(dragSourceContext, new DragSourceDragEvent(dragSourceContext, dropAction, targetActions, modifiers));
    }
 
    /**
     *
     */
 
    private void dragMotion(int dropAction, int targetActions, int modifiers) {
	listenerDispatcher.deliverMotion(dragSourceContext, new DragSourceDragEvent(dragSourceContext, dropAction, targetActions, modifiers));
    }

    /**
     *
     */

    private void operationChanged(int dropAction, int targetActions, int modifiers) {
	listenerDispatcher.deliverGestureChanged(dragSourceContext, new DragSourceDragEvent(dragSourceContext, dropAction, targetActions, modifiers));
    }
 
    /**
     *
     */

    private void dragExit(int modifiers) {
	listenerDispatcher.deliverExit(dragSourceContext, new DragSourceEvent(dragSourceContext));
    }
 
    /**
     *
     */

   private WDragSourceContextPeerBuffer convertData(String name) {
	WDragSourceContextPeerBuffer wdscpb  = null;
	Map			     flavors = dragSource.getFlavorMap().getFlavorsForNatives(new String[] { name });

	if (flavors == null || flavors.isEmpty()) return null;

	DataFlavor df = (DataFlavor)flavors.get(name);
	Object     o  = null;

	if (df == null) return null;

	try {
	    o = dragSourceContext.getTransferable().getTransferData(df); 
	} catch (IOException ioe) {
	    return null;
	} catch (UnsupportedFlavorException ufe) {
	    return null;
	}

	if (df.isRepresentationClassInputStream()) {
	    try {
	        wdscpb = new WDragSourceContextPeerBuffer((InputStream)o);
	    } catch (Exception e) {
		e.printStackTrace();
		    return null;
	    }
	} else if (df.isRepresentationClassRemote()) {
	    try {
		wdscpb = new WDragSourceContextPeerBuffer(new MarshalledObject(o));
	    } catch (Exception e) {
		e.printStackTrace();

		return null;
	    }
	} else if (df.isRepresentationClassSerializable()) {
	    try {
	        wdscpb = new WDragSourceContextPeerBuffer(o);
	    } catch (Exception e) {
		e.printStackTrace();
		return null;
	    }
	} else {
	    Class rc = df.getRepresentationClass();

	    if (java.lang.String.class.equals(rc)) {
	        try {
	            wdscpb = new WDragSourceContextPeerBuffer((String)o);
		} catch (Exception e) {
		    e.printStackTrace();
		    return null;
		}
	    }
	} 

	return wdscpb;
    }
    
    /**
     * upcall to get list of file names ...
     */

    private String[] convertFileList(String name) {
	Map      flavors = dragSource.getFlavorMap().getFlavorsForNatives(new String[] { name } );

	if (flavors == null || flavors.isEmpty()) return null;

	DataFlavor df = (DataFlavor)flavors.get(name);

	if (df == null) return null;

	final java.util.List list;
	try {
	    list = (java.util.List)dragSourceContext.getTransferable().getTransferData(df);
	} catch (IOException ioe) {
	    return null;
	} catch (UnsupportedFlavorException ufe) {
	    return null;
        } catch (ClassCastException cce) {
	    return null;
        }

	if (list.isEmpty()) return null;

	int nfiles = 0;

	for (int i = 0; i < list.size(); i++) {
	    Object o = list.get(i);
 
	    if (o instanceof java.io.File || o instanceof String) nfiles++;
	}

	if (nfiles == 0) return null;

	final String[] files = new String[nfiles];

	try {
	    java.security.AccessController.doPrivileged(
				new java.security.PrivilegedExceptionAction() {
		public Object run() throws IOException {

		    for (int i = 0; i < list.size(); i++) {
			Object o = list.get(i);
			if (o instanceof java.io.File) {
			    files[i] = 
			        ((java.io.File)list.get(i)).getCanonicalPath();
				
			} else if (o instanceof String) files[i] = (String)o;
		    }
		    return null;
		}
	    });
	} catch (java.security.PrivilegedActionException pae) {
	    // give up!
	    return null;
	}

	return files;
    }

    /**
     * upcall from native code
     */

    void dragDropFinished(boolean success, int action, int xhint, int yhint, boolean diwcr) {
	DragSourceDropEvent dsde = new DragSourceDropEvent(dragSourceContext, action, success);
	listenerDispatcher.deliverDropFinish(dragSourceContext, dsde);

	// this is used by the native code to synthesize a DROPFILES structure
	// when DnD'ing files.

	dropScreenXHint   = xhint;
	dropScreenYHint   = yhint;
	droppedInWindowCR = diwcr;
    }

    /**
     * upcall from native code
     */

    private int getDropXHint() { return dropScreenXHint; }

    /**
     * upcall from native code
     */

    private int getDropYHint() { return dropScreenYHint; }

    /**
     * upcall from native code
     */

    private boolean getDroppedInWindowCR() { return droppedInWindowCR; }

    /**
     * downcall into native code
     */

    native int createDragSource(InputEvent nativeTrigger, int actions, Object[] types);

    /**
     * downcall into native code
     */

    native void doDragDrop(int nativeCtxt, Cursor cursor);

    /**
     * downcall into native code
     */

    private static native void setNativeCursor(int nativeCtxt, Cursor c);

    /**
     * downcall into native code
     */

    private native void release(int nativeCtxt);

    /**
     * downcall to change source actions
     */

    private native void setActions(int nativeCtxt, int actions);

    /**
     * initialize native code
     */

    static native void Initialize();

    /*
     * fields
     */

       	    DragSource 		     	    dragSource;
    	    DragSourceContext		    dragSourceContext;
    	    DragGestureEvent	  	    trigger;
    	    Cursor     		     	    cursor;

    private Component  			     component;

    private int        			     actions;
    private Component  			     nativeCont;
    private int	       			     nativeCtxt;


    private boolean			     dropOccurred  = false;
    private int				     dropOperation = DnDConstants.ACTION_NONE;
    private boolean			     dropSuccess   = false;

    private int			     	     dropScreenXHint;
    private int			     	     dropScreenYHint;

    private boolean			     droppedInWindowCR;

    private boolean	    		     fakeExitDispatched = false;
}

/*****************************************************************************/

final class WDragSourceContextPeerDispatcher implements Runnable {

    private static final int NONE   = 0;
    private static final int ENTER  = 1;
    private static final int MOTION = 2;
    private static final int CHANGE = 3;
    private static final int EXIT   = 4;
    private static final int FINISH = 5;
    
    /**
     * deliver the dragEnter
     */

    synchronized void deliverEnter(DragSourceListener l, DragSourceDragEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = ENTER;

	dispatch();
    }

    /**
     * deliver the dragMotion
     */

    synchronized void deliverMotion(DragSourceListener l, DragSourceDragEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = MOTION;

	dispatch();
    }

    /**
     * deliver the dragGestureChanged
     */

    synchronized void deliverGestureChanged(DragSourceListener l, DragSourceDragEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = CHANGE;

	dispatch();
    }

    /**
     * deliver the dragExit
     */

    synchronized void deliverExit(DragSourceListener l, DragSourceEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = EXIT;

	dispatch();
    }

    /**
     * deliver the dropFinish
     */

    synchronized void deliverDropFinish(DragSourceListener l, DragSourceDropEvent e) {
	dsl	     = l;
	event 	     = e;
	dispatchType = FINISH;

	dispatch();
    }

    /**
     * do the dispatch.
     */

    private void dispatch() {
	boolean done = false;

	while (!done) try {
	    SwingUtilities.invokeAndWait((Runnable)this);
	    done = true;
	} catch (InterruptedException ie) {
	    ie.printStackTrace();
	} catch (InvocationTargetException ite) {
	    ite.printStackTrace();
	    done = true;
	}
    }

    /**
     * deliver the event
     */

    public void run() {
	try {
	    switch (dispatchType) {
		case ENTER:
		    dsl.dragEnter((DragSourceDragEvent)event);
		    break;

		case MOTION:
		    dsl.dragOver((DragSourceDragEvent)event);
		    break;

		case CHANGE:
		    dsl.dropActionChanged((DragSourceDragEvent)event);
		    break;

		case EXIT:
		    dsl.dragExit((DragSourceEvent)event);
		    break;

		case FINISH:

		    /*
		     * for platform compatibility we call dragExit here
		     * before we notify the end of the drop.
		     */

		    dsl.dragExit((DragSourceEvent)event);

		    dsl.dragDropEnd((DragSourceDropEvent)event);

		    break;

		default:
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	} finally {
	    dsl		 = null;
	    dispatchType = NONE;
	}
    }

    /*
     *
     */

    DragSourceListener	    dsl;

    private int		    dispatchType = NONE;
    private DragSourceEvent event;
}

/*****************************************************************************/

/*
 * for large data transfers this methods is extremely inefficent and should
 * be replaced with a scatter/gather algorithm.
 */

class WDragSourceContextPeerBuffer extends ByteArrayOutputStream {

    /**
     * create a buffer
     */

    WDragSourceContextPeerBuffer(Object o) throws IOException {
	super();

        ObjectOutputStream oos = new ObjectOutputStream(this);

	oos.writeObject(o);

	oos.close();
	close();
    }

    /**
     * create a buffer
     */

    WDragSourceContextPeerBuffer(String s) throws IOException {
	super();

	DataOutputStream dos = new DataOutputStream(this);

	dos.writeUTF(s);

	dos.close();
	close();
    }

    /**
     * create a buffer
     */

    WDragSourceContextPeerBuffer(byte[] ba) {
	super();

	count = ba.length;

	System.arraycopy(ba, 0, buf = new byte[count], 0, count);

	try {
	    close();
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	}
    }

    /**
     * create a buffer
     */

    WDragSourceContextPeerBuffer(InputStream is) throws IOException {
	super();
	fill(is);

	close();
    }

    /**
     * fill the buffer from an InputStream ...
     */

    private void fill(InputStream is) throws IOException {
	boolean eof   = false;
	int     avail = is.available();
	byte[]  tmp   = new byte[avail > 8192 ? avail : 8192];

	do {
	    int ret;

	    if (!(eof = (ret = is.read(tmp, 0, tmp.length)) == -1))
		write(tmp, 0, ret);
	} while (!eof);

	try {
	    is.close();
	} catch (IOException ioe) {
	    // Oops!
	}
    }

    /**
     * return the buffer
     */

    byte[] getBuffer() { return buf; }

    /*
     *
     */

    private int  readCursor = 0;
}

/*****************************************************************************/
