/*
 * @(#)WDrawingSurfaceInfo.java	1.8 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import sun.awt.DrawingSurfaceInfo;
import sun.awt.PhysicalDrawingSurface;
import sun.awt.Win32DrawingSurface;
import sun.awt.image.ImageRepresentation;
import java.awt.Component;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.image.ColorModel;

/**
 * The WDrawingSurfaceInfo object provides direct access for rendering
 * on Components maintained by the native peers of the Windows toolkit.
 *
 * @version 	1.8, 09/21/98
 * @author 	Jim Graham
 */
public class WDrawingSurfaceInfo
    implements DrawingSurfaceInfo, Win32DrawingSurface
{
    static {
        initIDs();
    }

    private int pData;
    int state;
    int w, h;
    WComponentPeer peer;

    /**
     * Construct a new WDrawingSurfaceInfo for the specified peer.
     */
    WDrawingSurfaceInfo(WComponentPeer peer) {
	this.peer = peer;
    }

    /**
     * Lock the current state of the drawing surface in preparation
     * for drawing to it and return an integer "change identifier"
     * which can be used to determine when changes have occured to
     * the drawing surface specification and when to retrieve new
     * extent and physical handle data.  If the identifier has not
     * changed since the previous time lock and unlock were called
     * then the consumer of this information is free to reuse clipping
     * and handle information cached during the previous rendering
     * operation.  If the identifier changes, then all extent,
     * clipping, and handle information must be retrieved again.
     * The unlock() method must be called when rendering is complete
     * or the component may be permanently locked down on the screen
     * and other rendering threads may block indefinately.
     */
    public native int lock();

    /**
     * Unlock the drawing surface and allow other threads to
     * access or change its composition.
     */
    public native void unlock();

    /**
     * Returns a rectangle describing the logical bounds of the given
     * screen object relative to the "drawing handle" information
     * returned by getSurface().  Note that there may be portions
     * of this bounding rectangle that are obscured.  This value
     * defines the location of the logical upper-left origin of
     * the component or screen object and its desired extents
     * (which may differ from the clipping extents due to the
     * presence of other overlapping screen objects).  The getClip()
     * method should be used to determine which portions of the
     * drawing surface can actually be drawn to.
     */
    public Rectangle getBounds() {
	Rectangle r = null;
	if (peer != null) {
	    r = ((Component)peer.target).getBounds();
	    r.setLocation(0, 0);
	}
	return r;
    }

    /**
     * Returns an implementation-specific object which describes a
     * handle to the physical drawing surface on which this screen
     * object resides.  Depending on the type of the return value,
     * this object may describe a Windows hWnd handle, an X11
     * drawable handle, a Mac GrafPtr structure, or some other
     * platform-specific handle.
     */
    public PhysicalDrawingSurface getSurface() {
	return this;
    }

    /**
     * Returns an explicit representation of the renderable portion
     * of the given screen object.  This clipping area will include
     * only those areas of the screen object that are not obscured
     * by other overlapping objects and which fall within the
     * boundaries of the physical drawing surface.  The area described
     * by the return value of this method can be directly used as a
     * graphics clip area to restrict rendering to the given screen
     * object.
     */
    public Shape getClip() {
	return getBounds();
    }

    /**
     * Return the Windows handle for the window
     * represented by this object.
     * If the drawing surface is a DIBSection then this method will
     * return NULL (0).
     * <p>
     * HWND *hWnd = (HWND) getHWnd();
     */
    public native int getHWnd();

    /**
     * Return the Windows handle for the bitmap
     * represented by this object.
     * If the drawing surface is a window then this method will
     * return NULL (0).
     * <p>
     * HBITMAP *hBitmap = (HBITMAP) getHBitmap();
     */
    public native int getHBitmap();

    /**
     * Return the pointer to the bits for the DIBSection
     * represented by this object.
     * If the drawing surface is a window then this method will
     * return NULL (0).
     * <p>
     * void *ppvBits = (void *) getPBits();
     */
    public native int getPBits();

    /**
     * Return the Windows DC handle for the window or bitmap
     * represented by this object.
     * <p>
     * HDC hDC = (HDC) getHDC();
     */
    public native int getHDC();

    /**
     * Return the depth of the window or bitmap
     * represented by this object.
     * <p>
     * int depth = getDepth();
     */
    public native int getDepth();

    /**
     * Return the Windows palette handle associated with the drawable
     * represented by this object.
     * <p>
     * HPALETTE hPal = (HPALETTE) getHPalette();
     */
    public native int getHPalette();

    /**
     * Return the ColorModel of the window or bitmap
     * represented by this object.
     * <p>
     * Hjava_awt_image_ColorModel *cm =
     *     (Hjava_awt_image_ColorModel *) getColorModel();
     */
    public ColorModel getColorModel() {
	return WToolkit.getStaticColorModel();
    }

    /**
     * Initialize JNI field and method ids
     */
    private static native void initIDs();
}
