/*
 * @(#)WFramePeer.java	1.25 98/09/10
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.util.Vector;
import java.awt.*;
import java.awt.peer.*;
import java.awt.image.ImageObserver;
import sun.awt.image.ImageRepresentation;
import sun.awt.image.IntegerComponentRaster;
import java.awt.image.Raster;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.BufferedImage;
import sun.awt.im.*;


class WFramePeer extends WWindowPeer implements FramePeer {

    // FramePeer implementation
    public native void setState(int state);
    public native int getState();

    public Dimension minimumSize() {
        int height = (((Frame)target).getMenuBar() != null) ? 45 : 30;
	return (new Dimension(30, height));
    }

    public void setIconImage(Image im) {
        int w = getSysIconWidth();
        int h = getSysIconHeight();
        BufferedImage bimage = new BufferedImage(w, h,
                                                BufferedImage.TYPE_INT_RGB);
        Graphics g = bimage.getGraphics();
        if (im instanceof WImage) {
	    ImageRepresentation ir = ((WImage)im).getImageRep();
	    ir.reconstruct(ImageObserver.ALLBITS);
        }
        g.drawImage(im, 0, 0, w, h, null);
        Raster  raster = bimage.getRaster();
        DataBuffer buffer = raster.getDataBuffer();
        // REMIND: native code should use ScanStride _AND_ width
        int data[] = ((DataBufferInt)buffer).getData();
        {
            int     ficW = raster.getWidth();
            if( raster instanceof IntegerComponentRaster )
            {
                ficW = ((IntegerComponentRaster)raster).getScanlineStride();
            }
            setIconImageFromIntRasterData(
                ((DataBufferInt)bimage.getRaster().getDataBuffer()).getData(),
                ficW, raster.getWidth(), raster.getHeight());
        }
    }

    // Note: Because this method calls resize(), which may be overridden
    // by client code, this method must not be executed on the toolkit
    // thread.
    public void setMenuBar(MenuBar mb) {
	WMenuBarPeer mbPeer = (WMenuBarPeer) WToolkit.targetToPeer(mb);
        setMenuBar0(mbPeer);
        updateInsets(insets_);
    }

    // Note: Because this method calls resize(), which may be overridden
    // by client code, this method must not be executed on the toolkit
    // thread.
    private native void setMenuBar0(WMenuBarPeer mbPeer);

    // Toolkit & peer internals

    WFramePeer(Frame target) {
	super(target);

	InputMethodManager imm = InputMethodManager.getInstance();
	String menuString = imm.getTriggerMenuString();
	if (menuString != null)
	{ 
	  pSetIMMOption(menuString);
	}
    }

    native void create(WComponentPeer parent);

    void initialize() {
        super.initialize();

        Frame target = (Frame)this.target;

	if (target.getTitle() != null) {
	    setTitle(target.getTitle());
	}
	setResizable(target.isResizable());

	Image icon = target.getIconImage();
	if (icon != null) {
	    setIconImage(icon);
	}
	if (!target.getCursor().equals(Cursor.getDefaultCursor())) {
	    setCursor(target.getCursor());
	}
    }
    private native void
        setIconImageFromIntRasterData(int[] rData, int nScanStride, int nW, int nH);
    private native static int getSysIconWidth();
    private native static int getSysIconHeight();

    native void pSetIMMOption(String option);
    void notifyIMMOptionChange(){
      InputMethodManager.getInstance().notifyChangeRequest((Component)target);
    }
    public synchronized native void reshape(int x, int y, int width, int height);
}
