/*
 * @(#)WGraphics.java	1.79 98/09/30
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.geom.AffineTransform;
import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import sun.awt.image.ByteComponentRaster;
import sun.awt.image.IntegerComponentRaster;
import java.awt.image.ColorModel;
import java.awt.color.ColorSpace;
import java.awt.image.ColorConvertOp;
import java.awt.Composite;
import java.awt.FontMetrics;
import java.awt.GraphicsConfiguration;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.font.GlyphVector;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.ImageObserver;
import sun.awt.PlatformFont;
import sun.awt.AWTFinalizer;
import sun.awt.AWTFinalizeable;
import sun.java2d.SunGraphics2D;
import sun.java2d.OutputManager;
import sun.awt.image.ImageRepresentation;
import sun.java2d.loops.ImageData;
import sun.java2d.loops.RasterOutputManager;
import sun.awt.Win32GraphicsDevice;
import sun.awt.Win32GraphicsConfig;
import sun.java2d.pipe.SimpleRenderPipe;
import sun.java2d.pipe.PixelDrawPipe;
import sun.java2d.pipe.PixelFillPipe;
import sun.java2d.pipe.ShapeDrawPipe;
import sun.java2d.pipe.TranslateablePipe;
import sun.java2d.pipe.PixelToShapeConverter;
import sun.java2d.pipe.ShapeToPolyConverter;
import sun.java2d.pipe.DrawToFillConverter;
import sun.java2d.pipe.TextPipe;
import sun.java2d.pipe.SpanIterator;
import sun.java2d.DeviceClipping;
import sun.java2d.loops.GraphicsPrimitiveMgr;

/**
 * WGraphics is an object that encapsulates a graphics context for a
 * particular canvas. 
 * 
 * @version 1.18 18 Mar 1996
 */

public class WGraphics
extends SunGraphics2D
implements AWTFinalizeable,
    PixelDrawPipe,
    PixelFillPipe,
    TranslateablePipe,
    SimpleRenderPipe,
    TextPipe,
    DeviceClipping
{
    static String myDevID = "WGraphics";

    static {
        initIDs();
        OutputManager manager = RasterOutputManager.getManager();
//        GraphicsPrimitiveMgr.register(new MMXAcceleratorLoops(), myDevID);
    }

    int		   pData;
    WComponentPeer peer;
    private ImageData cachedID;

    private native synchronized void createFromComponent(WComponentPeer w);
    private native synchronized void createFromGraphics(WGraphics g);
    private native synchronized void createFromPrintJob(WPrintJob pj);
    private native synchronized void createFromHDC(int hdc);
    private native void pSetFont(Font f);
    private native void pSetForeground(int c);

    WGraphics(WPrintJob pj) {
        createFromPrintJob(pj);
	Dimension d = pj.getPageDimension();
	setDevClip(0, 0, d.width, d.height);
    }

    WGraphics(WComponentPeer comp) {
	createFromComponent(comp);
        this.peer = comp;
	Component target = ((Component) comp.target);
	backgroundColor = target.getBackground();
        Dimension size = target.getSize();
	setDevClip(0, 0, size.width, size.height);
	if (peer instanceof WPanelPeer) {
	    Insets insets = ((WPanelPeer) peer).getInsets();
	    translate(-insets.left, -insets.top);
	    constrainX = -insets.left;
	    constrainY = -insets.top;
	}
    }

    WGraphics(int hdc) {
        createFromHDC(hdc);
    }

    /**
     * Create a new WGraphics Object based on this one.
     */
    public Object clone() {
	WGraphics g = (WGraphics) super.clone();
	g.pData = 0;
	g.createFromGraphics(this);
	return g;
    }

    public ColorModel getDeviceColorModel() {
        if (peer != null) {
            return peer.getColorModel();
        }
	return WToolkit.config.getColorModel();
    }

    public void validatePipe() {
        RasterOutputManager.getManager().validatePipe(this);
    }
    
    private int serialNum = -1;

    /*
     * Return the composite bounds of the user and device clips
     */
    public Rectangle getCompBounds() {
	if (peer != null && this.serialNum != peer.serialNum) {
	    Dimension size = ((Component) peer.target).getSize();
	    setDevClip(0, 0, size.width, size.height);
	    this.serialNum = peer.serialNum;
	}
	return super.getCompBounds();
    }

    public native void setOrigin(int tx, int ty);

    /**
     * Return the device configuration associated with this Graphics2D.
     */
    public GraphicsConfiguration getDeviceConfiguration() {
        return WToolkit.config;
    }

    /**
     * Return the OutputManager object assigned to manage this Graphics2D.
     */
    public OutputManager getOutputManager() {
        return RasterOutputManager.getManager();
    }

    private boolean disposed = false;

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     *
     * A WGraphics cannot be used after being disposed.
     */
    protected native void disposeImpl();
    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
	    synchronized (this) {
	        if (!disposed) {
		    disposed = call_disposeImpl = true;
		}
	    }
	}

	if (call_disposeImpl) {
	    disposeImpl();
	}
    }

    AWTFinalizeable finalnext;

    public void finalize() {
	if (!disposed) {
	    AWTFinalizer.addFinalizeable(this);
	}
    }

    public void doFinalization() {
	dispose();
    }

    public void setNextFinalizeable(AWTFinalizeable o) {
	finalnext = o;
    }

    public AWTFinalizeable getNextFinalizeable() {
	return finalnext;
    }

    public void setFont(Font font) {
	if ((font != null) && (this.font != font)) {
            super.setFont(font);
	    pSetFont(font);
	}
    }

    public Font getFont() {
	if (font == null)
	    setFont(new Font("Dialog", Font.PLAIN, 12));
	return font;
    }

    /**
     * Gets font metrics for the given font.
     */
    public FontMetrics getFontMetrics(Font font) {
	if (textpipe == invalidpipe) {
	    validatePipe();
	}
        if (textpipe == this) {
	    return WFontMetrics.getFontMetrics(font);
        } else {
            return super.getFontMetrics(font);
        }
    }

    /**
     * Sets the foreground color.
     */
    public void setColor(Color c) {
	if (c != null) {
	    super.setColor(c);
	    pSetForeground(c.getRGB());
	}
    }

    /**
     * Sets the foreground color.
     */
    public void setPaint(Paint p) {
	super.setPaint(p);
	if (p instanceof Color) {	// instanceof protects against null
	    pSetForeground(((Color) p).getRGB());
	}
    }

    /**
     * Sets the paint mode to overwrite the destination with the
     * current color. This is the default paint mode.
     */
    public native void pSetPaintMode();
    
    public void setPaintMode() {
        super.setPaintMode();
        pSetPaintMode();
    }

    /**
     * Sets the paint mode to alternate between the current color
     * and the given color.
     */
    public native void pSetXORMode(Color c);

    public void setXORMode(Color c) {
        super.setXORMode(c);
        pSetXORMode(c);
    }

    public native void changeClip(int X, int Y, int W, int H, boolean set);
    public native void removeClip();

     /** Clears the rectangle indicated by x,y,w,h. */
    public native void devClearRect(int x, int y, int w, int h);
    public void clearRect(int x, int y, int w, int h) {
	if (fillpipe == invalidpipe) {
	    validatePipe();
	}
	if (fillpipe == this) { 
	    devClearRect(x, y, w, h);
	} else {
	    super.clearRect(x, y, w, h);
	}
    }

    /** Fills the given rectangle with the foreground color. */
    public native void devFillRect(int X, int Y, int W, int H);

    public void fillRect(SunGraphics2D sg, int X, int Y, int W, int H) {
	devFillRect(X, Y, W, H);
    }

    /** Draws the given rectangle. */
    public native void devDrawRect(int X, int Y, int W, int H);

    public void drawRect(SunGraphics2D sg, int X, int Y, int W, int H) {
	devDrawRect(X, Y, W, H);
    }

// implementation of textpipe
    public Object drawString(SunGraphics2D g2d, String s,
                            float x, float y, Object cs) {
	drawStringWidth(s, (int)x, (int)y);
        return null;
    }

    public Object drawGlyphVector(
            SunGraphics2D g2d, GlyphVector g,
            float x, float y, Object cs) {
        return super.drawGlyphVector(g2d, g, x, y, cs);
    }

    public Object drawChars(SunGraphics2D g2d,
            char data[], int offset, int length, int x, int y, Object cs) {
	drawCharsWidth(data, offset, length, x, y);
        return null;
    }

    public Object drawBytes(SunGraphics2D g2d,
            byte data[], int offset, int length, int x, int y, Object cs) {
	drawBytesWidth(data, offset, length, x, y);
        return null;
    }
// end of implementation of textpipe

    /** Draws the given string and returns the length of the drawn
      string in pixels.  If font is not set then returns -1. */
    public native int drawStringWidth(String str, int x, int y);

    /** Draws the given character array and return the width in
      pixels. If font is not set then returns -1. */
    public native int drawCharsWidth(char data[], int offset, int length, int x, int y);

    /** Draws the given character array and return the width in
      pixels. If font is not set then returns -1. */
    public native int drawBytesWidth(byte data[], int offset, int length, int x, int y);

    /** Draws the given line. */
    public native void devDrawLine(int x1, int y1, int x2, int y2);

    public void drawLine(SunGraphics2D sg, int x1, int y1, int x2, int y2) {
	devDrawLine(x1, y1, x2, y2);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img, int x, int y, ImageObserver observer) {
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, x, y, null, observer);
        }
                                     
	WImage wimg = (WImage) img;
	if (wimg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wimg.getImageRep(); 
        return ir.drawToBufImage(this, (sun.awt.image.Image)img,
                                 x, y, null, observer);
    }

    

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
			     ImageObserver observer) {

	if (width == 0 || height == 0) {
	    return true;
	}

        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, x, y, width, height,
                                   null, observer);
        }
        
	WImage wimg = (WImage) img;
	if (wimg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
        
	ImageRepresentation ir = wimg.getImageRep();
	return ir.drawToBufImage(this, (sun.awt.image.Image) img, x, y,
                                 width, height, null, observer);
    }

    /**
     * Draws an image at x,y in nonblocking mode with a solid background
     * color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, Color bg,
			     ImageObserver observer) {
        if (img instanceof BufferedImage) {
            super.drawImage((BufferedImage)img, x, y, bg, observer);
            return true;
        }
                                     
	WImage wimg = (WImage) img;
	if (wimg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wimg.getImageRep(); 
        return ir.drawToBufImage(this, (sun.awt.image.Image)img,
                                 x, y, bg, observer);
    }

    /**
     * Draws an image scaled to x,y,w,h in nonblocking mode with a
     * solid background color and a callback object.
     */
    public boolean drawImage(Image img, int x, int y, int width, int height,
 			     Color bg, ImageObserver observer) {
	if (width == 0 || height == 0) {
	    return true;
	}

        // Check if this is actually a buffered image
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, x, y, width, height,
                                   bg, observer);
        }
        
	WImage wimg = (WImage) img;
	if (wimg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wimg.getImageRep();
	return ir.drawToBufImage(this, (sun.awt.image.Image) img, x, y,
                                 width, height, bg, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle
     * in nonblocking mode with a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     ImageObserver observer) {
        return drawImage(img, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2,
                         null, observer);
    }

    /**
     * Draws a subrectangle of an image scaled to a destination rectangle in
     * nonblocking mode with a solid background color and a callback object.
     */
    public boolean drawImage(Image img,
			     int dx1, int dy1, int dx2, int dy2,
			     int sx1, int sy1, int sx2, int sy2,
			     Color bgcolor, ImageObserver observer) {
        if (img instanceof BufferedImage) {
            return super.drawImage((BufferedImage)img, dx1, dy1, dx2, dy2,
                                   sx1, sy1, sx2, sy2, bgcolor, observer);
        }
        // REMIND: Guess we don't draw a line?
	if (dx1 == dx2 || dy1 == dy2) {
	    return true;
	}

        // REMIND: Guess we don't draw a line?
        if (sx1 == sx2 || sy1 == sy2) {
            return true;
        }
        
	WImage wimg = (WImage) img;
	if (wimg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = wimg.getImageRep();
	return ir.drawToBufImage(this, (sun.awt.image.Image) img,
                                 dx1, dy1, dx2, dy2,
                                 sx1, sy1, sx2, sy2,
                                 bgcolor, observer);
    }

    /**
     * Draw an image, applying a transform from image space into user space
     * before drawing.
     * The transformation from user space into device space is done with
     * the current transform in the Graphics2D.
     * The given transformation is applied to the image before the
     * transform attribute in the Graphics2D state is applied.
     * The rendering attributes applied include the clip, transform,
     * paint or color and composite attributes. Note that the result is
     * undefined, if the given transform is non-invertible.
     * @param img The image to be drawn.
     * @param xform The transformation from image space into user space.
     * @param obs The image observer to be notified on the image producing
     * progress.
     * @see #transform
     * @see #setComposite
     * @see #setClip
     */
    public boolean drawImage(Image img,
                             AffineTransform xform,
                             ImageObserver observer) {
        if (img instanceof BufferedImage) {
            super.drawImage((BufferedImage)img, xform, 0, 0, null, observer);
            return true;
        }
        
	WImage wimg = (WImage) img;
	if (wimg.hasError()) {
	    if (observer != null) {
		observer.imageUpdate(img,
				     ImageObserver.ERROR|ImageObserver.ABORT,
				     -1, -1, -1, -1);
                return false;
	    }
	}
	ImageRepresentation ir = wimg.getImageRep();
        return ir.drawToBufImage(this, (sun.awt.image.Image) img, xform,
                                 observer);
    }
    
    /**
     * Copies an area of the canvas that this graphics context paints to.
     * @param X the x-coordinate of the source.
     * @param Y the y-coordinate of the source.
     * @param W the width.
     * @param H the height.
     * @param dx the x-coordinate of the destination.
     * @param dy the y-coordinate of the destination.
     */
    public native void devCopyArea(int X, int Y, int W, int H, int dx, int dy);
    public void copyArea(int X, int Y, int W, int H, int dx, int dy) {
	if (fillpipe == invalidpipe) {
	    validatePipe();
	}
	if (fillpipe == this) {
	    devCopyArea(X, Y, W, H, dx, dy);
	} else {
	    super.copyArea(X, Y, W, H, dx, dy);
	}
    }


    /** Draws a rounded rectangle. */
    public native void devDrawRoundRect(int x, int y, int w, int h,
					int arcWidth, int arcHeight);

    public void drawRoundRect(SunGraphics2D sg,
			      int x, int y, int w, int h,
			      int arcWidth, int arcHeight) {
	devDrawRoundRect(x, y, w, h, arcWidth, arcHeight);
    }

    /** Draws a filled rounded rectangle. */
    public native void devFillRoundRect(int x, int y, int w, int h,
					int arcWidth, int arcHeight);

    public void fillRoundRect(SunGraphics2D sg,
			      int x, int y, int w, int h,
			      int arcWidth, int arcHeight) {
	devFillRoundRect(x, y, w, h, arcWidth, arcHeight);
    }

    /** Draws a bunch of lines defined by an array of x points and y points */
    private native void devDrawPolyline(int xPoints[], int yPoints[],
					int nPoints);

    private native void devDrawPolygon(int xPoints[], int yPoints[],
				       int nPoints);

    public void drawPolys(SunGraphics2D sg,
			  int[] xPoints, int[] yPoints, int nPoints[],
			  int numpolys, boolean close) {
	for (int i = 0; i < numpolys; i++) {
	    int numpts = nPoints[i];
	    if (close) {
		devDrawPolygon(xPoints, yPoints, numpts);
	    } else {
		devDrawPolyline(xPoints, yPoints, numpts);
	    }
	    System.arraycopy(xPoints, numpts, xPoints, 0,
			     xPoints.length - numpts);
	    System.arraycopy(yPoints, numpts, yPoints, 0,
			     yPoints.length - numpts);
	}
    }

    public void drawPolyline(SunGraphics2D sg,
			     int[] xPoints, int[] yPoints, int nPoints) {
	devDrawPolyline(xPoints, yPoints, nPoints);
    }

    public void drawPolygon(SunGraphics2D sg,
			    int[] xPoints, int[] yPoints, int nPoints) {
	devDrawPolygon(xPoints, yPoints, nPoints);
    }

    /** Fills a polygon with the current fill mask */
    public native void devFillPolygon(int xPoints[], int yPoints[],
				      int nPoints, boolean evenodd);

    public void fillPolys(SunGraphics2D sg,
			  int[] xPoints, int[] yPoints, int nPoints[],
			  int numpolys, boolean evenodd) {
	if (numpolys < 1) {
	    return;
	}
	int numpts;
	if (numpolys > 1) {
	    // REMIND: Windows has a PolyPolygon rendering function that
	    // is probably much more efficient.
	    Polygon pg = new Polygon();
	    int index = 0;
	    int x = 0, y = 0;
	    int mx, my;
	    for (int i = 0; i < numpolys; i++) {
		numpts = nPoints[i];
		if (numpts < 2) {
		    index += numpts;
		} else {
		    pg.addPoint(0, 0);
		    mx = xPoints[index];
		    my = yPoints[index];
		    for (int j = 0; j < numpts; j++) {
			x = xPoints[index];
			y = yPoints[index];
			pg.addPoint(x, y);
			index++;
		    }
		    if (x != mx || y != my) {
			pg.addPoint(mx, my);
		    }
		    pg.addPoint(0, 0);
		}
	    }
	    xPoints = pg.xpoints;
	    yPoints = pg.ypoints;
	    numpts = pg.npoints;
	} else {
	    numpts = nPoints[0];
	}
	devFillPolygon(xPoints, yPoints, numpts, evenodd);
    }

    /** Fills a set of spans with the current fill */
    public native void devFillSpans(SpanIterator si, long iterator);

    public void fillSpans(SunGraphics2D sg, SpanIterator si) {
	devFillSpans(si, si.getNativeIterator());
    }

    public void fillPolygon(SunGraphics2D sg,
			    int xPoints[], int yPoints[], int nPoints) {
	devFillPolygon(xPoints, yPoints, nPoints, true);
    }

    /** Draws an oval to fit in the given rectangle */
    public native void devDrawOval(int x, int y, int w, int h);

    public void drawOval(SunGraphics2D sg, int x, int y, int w, int h) {
	devDrawOval(x, y, w, h);
    }

    /** Fills an oval to fit in the given rectangle */
    public native void devFillOval(int x, int y, int w, int h);

    public void fillOval(SunGraphics2D sg, int x, int y, int w, int h) {
	devFillOval(x, y, w, h);
    }

    /**
     * Draws an arc bounded by the given rectangle from startAngle to
     * endAngle. 0 degrees is a vertical line straight up from the
     * center of the rectangle. Positive angles indicate clockwise
     * rotations, negative angle are counter-clockwise.
     */
    public native void devDrawArc(int x, int y, int w, int h,
				  int startAngle, int endAngle);

    public void drawArc(SunGraphics2D sg,
			int x, int y, int w, int h,
			int startAngle, int endAngle) {
	devDrawArc(x, y, w, h, startAngle, endAngle);
    }

    /** fills an arc. arguments are the same as drawArc. */
    public native void devFillArc(int x, int y, int w, int h,
				  int startAngle, int endAngle);

    public void fillArc(SunGraphics2D sg,
			int x, int y, int w, int h,
			int startAngle, int endAngle) {
	devFillArc(x, y, w, h, startAngle, endAngle);
    }

    // Print a peer using this graphics object.
    native void devPrint(WComponentPeer peer);

    public void print(WComponentPeer peer) {
        validatePipe();
        devPrint(peer);
    }

    /* Outline the given region. */
    //public native void drawRegion(Region r);

    /* Fill the given region. */
    //public native void fillRegion(Region r);

    public String toString() {	
	return getClass().getName();
    }

    /*
    * REMIND: DOCS!!!
    */
    public String getDevID() {
        return myDevID;
    }

    private native int W32LockViewResources(ImageData idData,
             int xView, int yView, int dxView, int dyView, int lockFlags);
    private native int W32UnLockViewResources(ImageData idData);

    /**
     * The method prepares locks for a SunGraphics2D to make it accessible
     * for direct operation on graphics content by GraphicsPrimitives
     */
    public ImageData lock(
             int xView, int yView, int dxView, int dyView, int lockFlags)
    {
        ImageData imgData = cachedID;
        if ((null == imgData) || imgData.amILocked()) {
            imgData = new ImageData();
            imgData.setColorModel(getDeviceColorModel());
        }
        imgData.setViewArea(xView, yView, dxView, dyView);
        imgData.setLockMethod(lockFlags);
        int status = W32LockViewResources(imgData,
                                xView, yView, dxView, dyView, lockFlags);
        if (SunGraphics2D.LOCK_OK == status) {
            cachedID = imgData;
            imgData.incLockCount();
            return imgData;
        }

        return null;
    }

    public int unlock(ImageData idData) {
        int nRC = LOCK_DATA_STRUCT_PROBLEM;
        nRC = LOCK_NOT_LOCKED;
        if ((null != idData) && (true == idData.amILocked())) {
            if (UNLOCK_OK == (nRC = W32UnLockViewResources(idData))) {
                idData.decLockCount();
            }
        }
        return nRC;
    }

    /**
     * Initialize JNI field IDs
     */
    private static native void initIDs();

   /**
    * To be called from awt_Graphics.cpp to see if the sun.java2d.noddraw
    * property is defined.
    */
    private static boolean checkNoDDraw() {
        String noddraw = (String) java.security.AccessController.doPrivileged(
            new sun.security.action.GetPropertyAction("sun.java2d.noddraw"));
        if (noddraw == null) {
            return false;
        } else {
            return true;
        }
    }
}
