/*
 * @(#)WListPeer.java	1.16 98/07/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;

class WListPeer extends WComponentPeer implements ListPeer {

    // ComponentPeer overrides

    public Dimension minimumSize() {
	return minimumSize(4);
    }
    public boolean isFocusTraversable() {
	return true;
    }

    // ListPeer implementation

    public int[] getSelectedIndexes() {
	List l = (List)target;
	int len = l.countItems();
	int sel[] = new int[len];
	int nsel = 0;
	for (int i = 0 ; i < len ; i++) {
	    if (isSelected(i)) {
		sel[nsel++] = i;
	    }
	}
	int selected[] = new int[nsel];
	System.arraycopy(sel, 0, selected, 0, nsel);
	return selected;
    }

    /* New method name for 1.1 */
    public void add(String item, int index) {
	addItem(item, index);
    }

    /* New method name for 1.1 */
    public void removeAll() {
	clear();
    }

    /* New method name for 1.1 */
    public void setMultipleMode (boolean b) {
	setMultipleSelections(b);
    }

    /* New method name for 1.1 */
    public Dimension getPreferredSize(int rows) {
	return preferredSize(rows);
    }

    /* New method name for 1.1 */
    public Dimension getMinimumSize(int rows) {
	return minimumSize(rows);
    }

    private FontMetrics   fm;
    public void addItem(String item, int index) {
        addItem0(item, index, fm.stringWidth(item));
    }
    native void addItem0(String item, int index, int width);

    public native void delItems(int start, int end);
    public void clear() {
	List l = (List)target;
	delItems(0, l.countItems());
    }
    public native void select(int index);
    public native void deselect(int index);
    public native void makeVisible(int index);
    public native void setMultipleSelections(boolean v);
    public native int  getMaxWidth();

    public Dimension preferredSize(int v) {
        Dimension d = minimumSize(v);
        d.width = Math.max(d.width, getMaxWidth() + 20);
        return d;
    }
    public Dimension minimumSize(int v) {
	return new Dimension(20 + fm.stringWidth("0123456789abcde"),
                             (fm.getHeight() * v) + 4); // include borders
    }

    // Toolkit & peer internals

    WListPeer(List target) {
	super(target);
    }

    native void create(WComponentPeer parent);

    void initialize() {
	List li = (List)target;

	fm = getFontMetrics( li.getFont() );

	// add any items that were already inserted in the target.
	int  nitems = li.countItems();
	for (int i = 0; i < nitems; i++) {
	    addItem(li.getItem(i), -1);
	}

	// set whether this list should allow multiple selections.
	setMultipleSelections(li.allowsMultipleSelections());

	// make the visible position visible.
	int index = li.getVisibleIndex();
	if (index >= 0) {
	    makeVisible(index);
	}

	// select the item if necessary.
	int sel[] = li.getSelectedIndexes();
	for (int i = 0 ; i < sel.length ; i++) {
	    select(sel[i]);
	}

	super.initialize();
    }

    private native void updateMaxItemWidth();

    /*public*/ native boolean isSelected(int index);

    // update the fontmetrics when the font changes
    public synchronized void setFont(Font f)
    {
        super.setFont( f );
	    fm = getFontMetrics( ((List)target).getFont() );
        updateMaxItemWidth();
    }

    void clearRectBeforePaint(Graphics g, Rectangle r) {
        // Overload to do nothing for native components
    }

    // native callbacks

    void handleAction(final int index) {
	final List l = (List)target;
	WToolkit.executeOnEventHandlerThread(l, new Runnable() {
	    public void run() {
		l.select(index);
		postEvent(new ActionEvent(target, ActionEvent.ACTION_PERFORMED,
                                  l.getItem(index)));
	    }
	});
    }

    void handleListChanged(final int index) {
	final List l = (List)target;
	WToolkit.executeOnEventHandlerThread(l, new Runnable() {
	    public void run() {
		postEvent(new ItemEvent(l, ItemEvent.ITEM_STATE_CHANGED,
                                new Integer(index),
                                isSelected(index)? ItemEvent.SELECTED :
                                                   ItemEvent.DESELECTED));

	    }
	});
    }
}
