/*
 * @(#)WObjectPeer.java	1.9 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.awt.Component;

abstract class WObjectPeer {

    static {
        initIDs();
    }

    int       pData;	// The Windows handle for the native widget.
    Object    target;	// The associated AWT object.

    private boolean disposed = false;

    public WObjectPeer getPeerForTarget(Object t) {
	WObjectPeer peer = (WObjectPeer) WToolkit.targetToPeer(t);
	return peer;
    }

    /*
     * Subclasses should override disposeImpl() instead of dispose(). Client
     * code should always invoke dispose(), never disposeImpl().
     */
    abstract protected void disposeImpl();
    public final void dispose() {
        boolean call_disposeImpl = false;

        if (!disposed) {
	    synchronized (this) {
	        if (!disposed) {
		    disposed = call_disposeImpl = true;
		}
	    }
	}

	if (call_disposeImpl) {
	    disposeImpl();
	}
    }
    protected final boolean isDisposed() {
        return disposed;
    }

    protected void finalize() throws Throwable {
        // Calling dispose() here is essentially a NOP since the current
        // implementation prohibts gc before an explicit call to dispose().
        dispose();
        super.finalize();
    }

    /**
     * Initialize JNI field and method IDs
     */
    private static native void initIDs();
}
