/*
 * @(#)WPathGraphics.java	1.2 98/09/21
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;

import java.awt.geom.PathIterator;
import java.awt.print.PrinterJob;

import sun.java2d.PathGraphics;

class WPathGraphics extends PathGraphics {

    WPathGraphics(Graphics2D graphics, PrinterJob printerJob) {
	super(graphics, printerJob);
    }

    /**
     * Creates a new <code>Graphics</code> object that is 
     * a copy of this <code>Graphics</code> object.
     * @return     a new graphics context that is a copy of 
     *                       this graphics context.
     * @since      JDK1.0
     */
    public Graphics create() {

	return new WPathGraphics((Graphics2D) getDelegate().create(),
				 getPrinterJob());
    }

    /*
     * Fill the path defined by <code>pathIter</code>
     * with the specified color.
     * The path is provided in device coordinates.
     */
    protected void deviceFill(PathIterator pathIter, Color color) {

	WPrinterJob wPrinterJob = (WPrinterJob) getPrinterJob();

	convertToWPath(pathIter);
	wPrinterJob.selectSolidBrush(color);
	wPrinterJob.fillPath();
    }

    /*
     * Set the printer device's clip to be the
     * path defined by <code>pathIter</code>
     * The path is provided in device coordinates.
     */
    protected void deviceClip(PathIterator pathIter) {

	WPrinterJob wPrinterJob = (WPrinterJob) getPrinterJob();

	convertToWPath(pathIter);
	wPrinterJob.selectClipPath();
    }

    /**
     * Given a Java2D <code>PathIterator</code> instance,
     * this method translates that into a Window's path
     * in the printer device context.
     */
    private void convertToWPath(PathIterator pathIter) {

	float[] segment = new float[6];
	int segmentType;

	WPrinterJob wPrinterJob = (WPrinterJob) getPrinterJob();

	/* Map the PathIterator's fill rule into the Window's
	 * polygon fill rule.
	 */
	int polyFillRule;
	if (pathIter.getWindingRule() == PathIterator.WIND_EVEN_ODD) {
	    polyFillRule = WPrinterJob.POLYFILL_ALTERNATE;
	} else {
	    polyFillRule = WPrinterJob.POLYFILL_WINDING;
	}
	wPrinterJob.setPolyFillMode(polyFillRule);

	wPrinterJob.beginPath();

	while (pathIter.isDone() == false) {
	    segmentType = pathIter.currentSegment(segment);

	    switch (segmentType) {
	     case PathIterator.SEG_MOVETO:
		wPrinterJob.moveTo(segment[0], segment[1]);
		break;
	     
	     case PathIterator.SEG_LINETO:
		wPrinterJob.lineTo(segment[0], segment[1]);
		break;

	    /* Convert the quad path to a bezier.
	     */
	     case PathIterator.SEG_QUADTO:
		int lastX = wPrinterJob.getPenX();
		int lastY = wPrinterJob.getPenY();
		float c1x = lastX + (segment[0] - lastX) * 2 / 3;
		float c1y = lastY + (segment[1] - lastY) * 2 / 3;
		float c2x = segment[2] - (segment[2] - segment[0]) * 2/ 3;
		float c2y = segment[3] - (segment[3] - segment[1]) * 2/ 3;
		wPrinterJob.polyBezierTo(c1x, c1y,
				         c2x, c2y,
				         segment[2], segment[3]);
		break;

	     case PathIterator.SEG_CUBICTO:
		wPrinterJob.polyBezierTo(segment[0], segment[1],
				         segment[2], segment[3],
				         segment[4], segment[5]);
		break;

	     case PathIterator.SEG_CLOSE:
		wPrinterJob.closeFigure();
		break;
	    }


	    pathIter.next();
	}

	wPrinterJob.endPath();

    }

}
