/*
 * @(#)WPrintJob.java	1.15 98/07/10
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.awt.*;
import java.io.*;

/**
 * A class which initiates and executes a Win32 print job.
 *
 * @version 1.0 24 Jan 1996
 * @author Erik Gilbert
 */
public class WPrintJob extends PrintJob {
    String  	     title;
    Dimension 	     pageDimension;
    int		     pageResolution;
    boolean 	     lastFirst = false;
    int		     pageCount = 0;
    WPrintGraphics   graphics = null;

    static {
        initIDs();
    }

    public WPrintJob(String title) throws IOException {
	this.title = title;
	pageDimension = new Dimension(612, 792);
	pageResolution = 72;
    }

    PrintJob initJob() {
        /* Instantiating a WPrintGraphics causes a Print Dialog to be
	 * diplayed. If the user presses Cancel, graphics.pData will remain
	 * zero, otherwise it contains a pointer to the AwtGraphics object
	 * that contains the print DC.
         */
        graphics = new WPrintGraphics(this);
	if (graphics.pData != 0) {
	    return this;
	}
	else {
	    graphics = null;
	    return null;
	}
    }

    private native void newPage();

    /**
     * Gets a Graphics object that will draw to the next page.
     * The page is flushed to the printer when the graphics
     * object is disposed.
     */
    public synchronized Graphics getGraphics() {
        if (graphics == null) {
	    return null;
	}
	newPage();
	Graphics g = graphics.create(++pageCount);

	// Fix for bug #4080817 -- Null pointer exception from drawString()
	// on a Graphics object of PrintJob.
	// The problem here was that no default font was set on the graphics
	// object returned.
	// "Dialog, plain, 12 points" matches the default font for Frame.
	g.setFont(new Font("Dialog", Font.PLAIN, 12));

	return g;
    }

    /**
     * Returns the dimensions of the page in pixels.
     */
    public Dimension getPageDimension() {
	return pageDimension;
    }

    /**
     * Returns the resolution of the page in pixels per inch.
     */
    public int getPageResolution() {
	return pageResolution;
    }

    /**
     * Returns true if the last page should be printed first.
     */
    public boolean lastPageFirst() {
	return lastFirst;
    }

    private native void flushPageImpl();

    synchronized void flushPage(int pageNum) {
        // if (pageNum < pageCount), then we already flushed page <pageNum>
        // as part of a call to newPage()
        if (pageNum == pageCount && graphics != null) {
	    flushPageImpl();
	}
    }

    /**
     * Ends the print job and does cleanup.
     */
    private native void endImpl();

    public synchronized void end() {
        if (graphics != null) {
	    endImpl();
	    graphics.dispose();
	    graphics = null;
	}
    }

    /**
     * Initialize JNI field IDs
     */
    private static native void initIDs();
}











