/*
 * @(#)WScrollPanePeer.java	1.17 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.awt.windows;

import java.awt.*;
import java.awt.peer.*;

class WScrollPanePeer extends WPanelPeer implements ScrollPanePeer {

    int scrollbarWidth;
    int scrollbarHeight;
    boolean ignore;

    native void create(WComponentPeer parent);	    
    native int getOffset(int orient);

    WScrollPanePeer(Component target) {
	super(target);
        scrollbarWidth = _getVScrollbarWidth();
        scrollbarHeight = _getHScrollbarHeight();
	ignore = false;
    }

    void initialize() {
        super.initialize();
        setInsets();
    }

    public void setUnitIncrement(Adjustable adj, int p) {
        // The unitIncrement is grabbed from the target as needed.
    }

    public Insets insets() {
        return getInsets();
    }
    private native void setInsets();

    public native synchronized void setScrollPosition(int x, int y);

    public int getHScrollbarHeight() {
        return scrollbarHeight;
    }
    private native int _getHScrollbarHeight();

    public int getVScrollbarWidth() {
        return scrollbarWidth;
    }
    private native int _getVScrollbarWidth();

    public Point getScrollOffset() {
	int x, y;
	x = getOffset(Scrollbar.HORIZONTAL);
	y = getOffset(Scrollbar.VERTICAL);
	return new Point(x, y);
    }

    /**
     * The child component has been resized.  The scrollbars must be
     * updated with the new sizes.  At the native level the sizes of
     * the actual windows may not have changed yet, so the size
     * information from the java-level is passed down and used.
     */
    public void childResized(int width, int height) {
	ScrollPane sp = (ScrollPane)target;
	Dimension vs = sp.getSize();
	setSpans(vs.width, vs.height, width, height);
        setInsets();
    }

    native synchronized void setSpans(int viewWidth, int viewHeight, 
                                      int childWidth, int childHeight);

    /**
     * Called by ScrollPane's internal observer of the scrollpane's adjustables.
     * This is called whenever a scroll position is changed in one
     * of adjustables, whether it was modified externally or from the
     * native scrollbars themselves.  If the change was sourced by the
     * native scrollbars then ignore will be set to true.
     */
    public void setValue(Adjustable adj, int v) {
	if (! ignore) {
	    Component c = getScrollChild();
	    Point p = c.getLocation();
	    switch(adj.getOrientation()) {
	    case Adjustable.VERTICAL:
		setScrollPosition(-(p.x), v);
		break;
	    case Adjustable.HORIZONTAL:
		setScrollPosition(v, -(p.y));
		break;
	    }
	}
    }
	    
    private native Component getScrollChild();

    /**
     * Callback from windows to indicate that the vertical scrollbar was
     * adjusted and a new value is desired.  If this is a valid
     * value the adjustable will change and setValue will be called.
     */
    void scrolledVertical(final int value) {
	final ScrollPane sp = (ScrollPane)target;
	final WScrollPanePeer spp = this;
	WToolkit.executeOnEventHandlerThread(sp, new Runnable() {
	    public void run() {
	        Adjustable adj = sp.getVAdjustable();
	        if (adj != null) {
		    Adjustor e = new Adjustor(adj, value, spp);
	            postEvent(e);
	        }
	    }
	});
    }

    /**
     * Callback from windows to indicate that the horizontal scrollbar was
     * adjusted and a new value is desired.  If this is a valid
     * value the adjustable will change and setValue will be called.
     */
    void scrolledHorizontal(final int value) {
	final ScrollPane sp = (ScrollPane)target;
	final WScrollPanePeer spp = this;
	WToolkit.executeOnEventHandlerThread(sp, new Runnable() {
	    public void run() {
	        Adjustable adj = sp.getHAdjustable();
	        if (adj != null) {
		    Adjustor e = new Adjustor(adj, value, spp);
	            postEvent(e);
	        }
	    }
	});
    }

    /**
     * This is used to change the adjustable on another thread 
     * to represent a change made in the native scrollbar.  Since
     * the change was reflected immediately at the native level,
     * notification from the adjustable is temporarily ignored.
     */
    class Adjustor extends AWTEvent implements java.awt.ActiveEvent {

	Adjustor(Adjustable adj, int value, WScrollPanePeer peer) {
	    super(adj, 0);
	    this.adj = adj;
	    this.value = value;
	    this.peer = peer;
	}

        public void dispatch() {
	    peer.ignore = true;
	    adj.setValue(value);
	    peer.ignore = false;
	}

	int value;
	Adjustable adj;
	WScrollPanePeer peer;
    }

}

