/*
 * @(#)WTextComponentPeer.java	1.15 98/07/07
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.awt.*;
import java.awt.peer.*;
import java.awt.event.TextEvent;

abstract
class WTextComponentPeer extends WComponentPeer implements TextComponentPeer {

    static {
        initIDs();
    }

    // WComponentPeer overrides

    public void setBackground(Color c) {
	TextComponent tc = (TextComponent)target;
	if (tc.isEditable()) {
	    c = c.brighter();
	}
	super.setBackground(c);
    }

    // TextComponentPeer implementation

    public void setEditable(boolean editable) {
	enableEditing(editable);
	setBackground(((TextComponent)target).getBackground());
    }
    public native String getText();
    public native void setText(String txt);
    public native int getSelectionStart();
    public native int getSelectionEnd();
    public native void select(int selStart, int selEnd);

    // Toolkit & peer internals

    WTextComponentPeer(TextComponent target) {
	super(target);
    }

    void initialize() {
	TextComponent tc = (TextComponent)target;
	String text = tc.getText();

	if (text != null) {
	    setText(text);
	}
	select(tc.getSelectionStart(), tc.getSelectionEnd());
	setEditable(tc.isEditable());

	super.initialize();
    }

    native void enableEditing(boolean e);

    public boolean isFocusTraversable() {
	return true;
    }

    /*
     * Set the caret position by doing an empty selection. This
     * unfortunately resets the selection, but seems to be the
     * only way to get this to work.
     */
    public void setCaretPosition(int pos) {
        select(pos,pos);
    }

    /*
     * Get the caret position by looking up the end of the current
     * selection.
     */
    public int getCaretPosition() {
        return getSelectionStart();
    }

    /*
     * Post a new TextEvent when the value of a text component changes.
     */
    public void valueChanged() {
        postEvent(new TextEvent(target, TextEvent.TEXT_VALUE_CHANGED));
    }

    void clearRectBeforePaint(Graphics g, Rectangle r) {
        // Overload to do nothing for native components
    }

    /**
     * Initialize JNI field and method IDs
     */
    private static native void initIDs();

}
