/*
 * @(#)WToolkit.java	1.75 98/09/08
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.awt.windows;

import java.util.Hashtable;
import java.awt.*;
import java.awt.image.*;
import java.awt.peer.*;
import java.awt.event.MouseEvent;
import java.awt.datatransfer.Clipboard;
import java.io.*;
import java.net.URL;
import java.util.Properties;
import java.awt.print.PageFormat;
import sun.awt.image.ByteArrayImageSource;
import sun.awt.image.FileImageSource;
import sun.awt.image.URLImageSource;
import sun.awt.image.ImageRepresentation;
import sun.awt.AppContext;
import sun.awt.EmbeddedFrame;
import sun.awt.SunToolkit;
import sun.awt.im.InputMethod;
import sun.awt.im.InputMethodHighlightMapping;

import sun.java2d.loops.RasterOutputManager;


import java.awt.dnd.DragSource;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.dnd.MouseDragGestureRecognizer;
import java.awt.dnd.InvalidDnDOperationException;
import java.awt.dnd.peer.DragSourceContextPeer;
 
import sun.awt.windows.WDragSourceContextPeer;

public class WToolkit extends SunToolkit implements Runnable {

// Debugging properties
    static final boolean enableDebug = true;
    static GraphicsConfiguration config = (GraphicsEnvironment.
					   getLocalGraphicsEnvironment().
					   getDefaultScreenDevice().
					   getDefaultConfiguration());


    // Print any AWT_TRACE statements
    boolean dbgTrace = false;
    private static final String propTrace = "WINAWT.Trace";

    // Verify all components are in sync with their HWND state
    boolean dbgVerify = false;
    private static final String propVerify = "WINAWT.VerifyComponents";

    // Enter debugger on assertion failure
    boolean dbgBreak = false;
    private static final String propBreak = "WINAWT.BreakOnFailure";

    // Enable heap checking
    boolean dbgHeapCheck = false;
    private static final String propHeapCheck = "WINAWT.HeapCheck";

    // System clipboard.
    WClipboard clipboard;

    // cache of font peers
    private Hashtable cacheFontPeer;

    /**
     * Initialize JNI field and method IDs
     */
    private static native void initIDs();

    static {
	java.security.AccessController.doPrivileged(
		    new sun.security.action.LoadLibraryAction("awt"));
	initIDs();
    }

    public WToolkit() {
        InitializeDebuggingOptions();

        // Startup toolkit threads
        synchronized (this) {

          new Thread(this, "AWT-Windows").start();      // Fix for bug #4046430 -- Race condition
                                                        // where notifyAll can be called before
                                                        // the "AWT-Windows" thread's parent thread is 
                                                        // waiting, resulting in a deadlock on startup.

          try {
            wait();
          }
          catch (InterruptedException x) {
          }
        }

	WDropTargetContextPeer.Initialize();
	WDragSourceContextPeer.Initialize();
    }

    private void InitializeDebuggingOptions() {
        String s = Toolkit.getProperty(propTrace, "false");
        dbgTrace = Boolean.valueOf(s).booleanValue();
        s = Toolkit.getProperty(propVerify, "false");
        dbgVerify = Boolean.valueOf(s).booleanValue();
        s = Toolkit.getProperty(propBreak, "false");
        dbgBreak = Boolean.valueOf(s).booleanValue();
        s = Toolkit.getProperty(propHeapCheck, "false");
        dbgHeapCheck = Boolean.valueOf(s).booleanValue();
    }

    public native void init(Thread t);

    public void run() {
        Thread t = Thread.currentThread();
        init(t);
        synchronized (this) {
            notifyAll();
        }
        eventLoop();
    }

    /* 
     * Notifies the thread that started the server
     * to indicate that initialization is complete.
     * Then enters an infinite loop that retrieves and processes events.  
     */
    native void eventLoop();

    /*
     * Create peer objects.
     */

    public ButtonPeer createButton(Button target) {
	ButtonPeer peer = new WButtonPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public TextFieldPeer createTextField(TextField target) {
	TextFieldPeer peer = new WTextFieldPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public LabelPeer createLabel(Label target) {
	LabelPeer peer = new WLabelPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ListPeer createList(List target) {
	ListPeer peer = new WListPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public CheckboxPeer createCheckbox(Checkbox target) {
	CheckboxPeer peer = new WCheckboxPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ScrollbarPeer createScrollbar(Scrollbar target) {
	ScrollbarPeer peer = new WScrollbarPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ScrollPanePeer createScrollPane(ScrollPane target) {
	ScrollPanePeer peer = new WScrollPanePeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public TextAreaPeer createTextArea(TextArea target) {
	TextAreaPeer peer = new WTextAreaPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public ChoicePeer createChoice(Choice target) {
	ChoicePeer peer = new WChoicePeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public FramePeer  createFrame(Frame target) {
	FramePeer peer = new WFramePeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public CanvasPeer createCanvas(Canvas target) {
	CanvasPeer peer = new WCanvasPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public PanelPeer createPanel(Panel target) {
	PanelPeer peer = new WPanelPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public WindowPeer createWindow(Window target) {
	WindowPeer peer = new WWindowPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public DialogPeer createDialog(Dialog target) {
	DialogPeer peer = new WDialogPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public FileDialogPeer createFileDialog(FileDialog target) {
	FileDialogPeer peer = new WFileDialogPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public MenuBarPeer createMenuBar(MenuBar target) {
	MenuBarPeer peer = new WMenuBarPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public MenuPeer createMenu(Menu target) {
	MenuPeer peer = new WMenuPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public PopupMenuPeer createPopupMenu(PopupMenu target) {
	PopupMenuPeer peer = new WPopupMenuPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public MenuItemPeer createMenuItem(MenuItem target) {
	MenuItemPeer peer = new WMenuItemPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public CheckboxMenuItemPeer createCheckboxMenuItem(CheckboxMenuItem target) {
	CheckboxMenuItemPeer peer = new WCheckboxMenuItemPeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    public WEmbeddedFramePeer createEmbeddedFrame(WEmbeddedFrame target) {
	WEmbeddedFramePeer peer = new WEmbeddedFramePeer(target);
	peerMap.put(target, peer);
	return peer;
    }

    static boolean prepareScrImage(Image img, int w, int h, ImageObserver o) {
	if (w == 0 || h == 0) {
	    return true;
	}
	
	// Must be an OffScreenImage
	if (!(img instanceof WImage)) {
	   return true;
	}

	WImage ximg = (WImage) img;
	if (ximg.hasError()) {
	    if (o != null) {
		o.imageUpdate(img, ImageObserver.ERROR|ImageObserver.ABORT,
			      -1, -1, -1, -1);
	    }
	    return false;
	}
	ImageRepresentation ir = ximg.getImageRep();
	return ir.prepare(o);
    }

    static int checkScrImage(Image img, int w, int h, ImageObserver o) {
        if (!(img instanceof WImage)) {
            return ImageObserver.ALLBITS;
        }
	WImage ximg = (WImage) img;
	int repbits;
	if (w == 0 || h == 0) {
	    repbits = ImageObserver.ALLBITS;
	}
        else {
	    repbits = ximg.getImageRep().check(o);
	}
	return ximg.check(o) | repbits;
    }

    public int checkImage(Image img, int w, int h, ImageObserver o) {
	return checkScrImage(img, w, h, o);
    }

    public boolean prepareImage(Image img, int w, int h, ImageObserver o) {
	return prepareScrImage(img, w, h, o);
    }

    public Image createImage(ImageProducer producer) {
	return new WImage(producer);
    }

    static native ColorModel makeColorModel();
    static ColorModel screenmodel;

    static ColorModel getStaticColorModel() {
	if (screenmodel == null) {
	    screenmodel = makeColorModel();
	}
	return screenmodel;
    }

    public ColorModel getColorModel() {
	return getStaticColorModel();
    }

    public native int getScreenResolution();
    protected native int getScreenWidth();
    protected native int getScreenHeight();

    public FontMetrics getFontMetrics(Font font) {
        if (!RasterOutputManager.usesPlatformFont()) {
            return super.getFontMetrics(font);
        } else {
            return WFontMetrics.getFontMetrics(font);
        }
    }
    
    public FontPeer getFontPeer(String name, int style) {
        FontPeer retval = null;
        String lcName = name.toLowerCase();
        if (null != cacheFontPeer) {
            retval = (FontPeer)cacheFontPeer.get(lcName + style);
            if (null != retval) {
                return retval;
            }
        }
	/*----------*
        if (!RasterOutputManager.usesPlatformFont()) {
	    return new FontPeer() {};
        } else {
            return new WFontPeer(name, style);
	}
	*----------*/
        retval = new WFontPeer(name, style);
        if (retval != null) {
            if (null == cacheFontPeer) {
                cacheFontPeer = new Hashtable(5, (float)0.9);
            }
            if (null != cacheFontPeer) {
                cacheFontPeer.put(lcName + style, retval);
            }
        }
        return retval;
    }

    public native void sync();

    public PrintJob getPrintJob(Frame frame, String doctitle, Properties props) {
	SecurityManager security = System.getSecurityManager();
	if (security != null) {
	    security.checkPrintJobAccess();
	}
        WPrintJob pj = null;

        try {
            pj = new WPrintJob(doctitle);
        } catch (IOException e) {}

        return pj.initJob();
    }

    public native void beep();

    public Clipboard getSystemClipboard() {
        SecurityManager security = System.getSecurityManager();
        if (security != null) {
	  security.checkSystemClipboardAccess();
	}
        if (clipboard == null) {
	    clipboard = new WClipboard();
	}
	return clipboard;
    }

    protected native void loadSystemColors(int[] systemColors);

    public static final Object targetToPeer(Object target) {
        return SunToolkit.targetToPeer(target);
    }

    public static final void targetDisposedPeer(Object target, Object peer) {
        SunToolkit.targetDisposedPeer(target, peer);
    }

    /**
     * Returns a new input method adapter for native input methods.
     */
    public InputMethod getInputMethodAdapter() {
        return new WInputMethod();
    }

    /**
     * Returns the input method highlight mapping interface.
     */
    public InputMethodHighlightMapping getInputMethodHighlightMapping() {
	return WInputMethod.getInputMethodHighlightMapping();
    }

    /**
     * Returns whether enableInputMethods should be set to true for peered
     * TextComponent instances on this platform.
     */
    public boolean enableInputMethodsForTextComponent() {
        return true;
    }

    /**
     * Returns a new custom cursor.
     */
    public Cursor createCustomCursor(Image cursor, Point hotSpot, String name)
        throws IndexOutOfBoundsException {
        return new WCustomCursor(cursor, hotSpot, name);
    }

    /**
     * Returns the supported cursor size (Win32 only has one).
     */
    public Dimension getBestCursorSize(int preferredWidth, int preferredHeight) {
        return new Dimension(WCustomCursor.getCursorWidth(), 
                             WCustomCursor.getCursorHeight());
    }

    public int getMaximumCursorColors() {
        // Black and white for now...
        return 2;
    }

    /**
     * create the peer for a DragSourceContext
     */
 
    public DragSourceContextPeer createDragSourceContextPeer(DragGestureEvent dge) throws InvalidDnDOperationException {
        return WDragSourceContextPeer.getDragSourceContextPeer(dge);
    }

    public DragGestureRecognizer createDragGestureRecognizer(Class abstractRecognizerClass, DragSource ds, Component c, int srcActions, DragGestureListener dgl) {
        if (MouseDragGestureRecognizer.class.equals(abstractRecognizerClass))
            return new WMouseDragGestureRecognizer(ds, c, srcActions, dgl);
        else
            return null;
    }

    /**
     *
     */

    private static final String prefix  = "DnD.Cursor.";
    private static final String postfix = ".32x32";

    protected Object lazilyLoadDesktopProperty(String name) {
	if (name.startsWith(prefix)) {
	    String cursorName = name.substring(prefix.length(), name.length()) + postfix;

	    try {
	        return Cursor.getSystemCustomCursor(cursorName);
	    } catch (AWTException awte) {
	 	throw new RuntimeException("cannot load system cursor: " + cursorName);
	    }
	}

	return super.lazilyLoadDesktopProperty(name);
    }

    protected void initializeDesktopProperties() {
        desktopProperties.put("DnD.Autoscroll.initialDelay",     new Integer(50)
);
        desktopProperties.put("DnD.Autoscroll.interval",         new Integer(50)
);
        desktopProperties.put("DnD.Autoscroll.cursorHysteresis", new Integer(5))
;
    }

}
