/*
 * @(#)socket_md.c	1.9 98/06/30
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <windef.h>
#include <winsock.h>

#include "hpi_impl.h"

#include "mutex_md.h"

struct sockaddr;

#define FN_RECV		  0
#define FN_SEND		  1
#define FN_LISTEN	  2
#define FN_ACCEPT	  3
#define FN_RECVFROM	  4
#define FN_SENDTO	  5
#define FN_SELECT	  6
#define FN_CONNECT	  7
#define FN_CLOSESOCKET	  8
#define FN_SHUTDOWN	  9
#define FN_DOSHUTDOWN	 10
#define FN_GETHOSTBYNAME 11
#define FN_HTONS	 12
#define FN_SOCKET        13
#define FN_WSASENDDISCONNECT 14
#define FN_SOCKETAVAILABLE 15

static int (PASCAL FAR *sockfnptrs[])() = 
    {NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL,
     NULL};

static bool_t sockfnptrs_initialized = FALSE;
static mutex_t sockFnTableMutex;

/* is Winsock2 loaded? better to be explicit than to rely on sockfnptrs */
static bool_t winsock2Available = FALSE;


/* IMPORTANT: whenever possible, we want to use Winsock2 (ws2_32.dll)
 * instead of Winsock (wsock32.dll). Other than the fact that it is
 * newer, less buggy and faster than Winsock, Winsock2 lets us to work
 * around the following problem:
 *
 * Generally speaking, it is important to shutdown a socket before
 * closing it, since failing to do so can sometimes result in a TCP
 * RST (abortive close) which is disturbing to the peer of the
 * connection.
 * 
 * The Winsock way to shutdown a socket is the Berkeley call
 * shutdown(). We do not want to call it on Win95, since it
 * sporadically leads to an OS crash in IFS_MGR.VXD.  Complete hull
 * breach.  Blue screen.  Ugly.
 *
 * So, in initSockTable we look for Winsock 2, and if we find it we
 * assign wsassendisconnectfn function pointer. When we close, we
 * first check to see if it's bound, and if it is, we call it. Winsock
 * 2 will always be there on NT, and we recommend that win95 user
 * install it.
 *
 * - br 10/11/97
 */

static void
initSockFnTable() {
    int (PASCAL FAR* WSAStartupPtr)(WORD, LPWSADATA); 
    WSADATA wsadata;
    OSVERSIONINFO info;

    mutexInit(&sockFnTableMutex);
    mutexLock(&sockFnTableMutex);
    if (sockfnptrs_initialized == FALSE) {
	HANDLE hWinsock;

	/* try to load Winsock2, and if that fails, load Winsock */
	hWinsock = LoadLibrary("ws2_32.dll");
	if (hWinsock == NULL) {
	    hWinsock = LoadLibrary("wsock32.dll");
	    winsock2Available = FALSE;
	} else {
	    winsock2Available = TRUE;
	}

	if (hWinsock == NULL) {
	    VM_CALL(jio_fprintf)(stderr, "Could not load Winsock 1 or 2 (error: %d)\n", 
			GetLastError());
	}

	/* If we loaded a DLL, then we might as well initialize it.  */
	WSAStartupPtr = (int (PASCAL FAR *)(WORD, LPWSADATA))
			    GetProcAddress(hWinsock, "WSAStartup");
	if (WSAStartupPtr(MAKEWORD(1,1), &wsadata) != 0) {
	    VM_CALL(jio_fprintf)(stderr, "Could not initialize Winsock\n"); 
	}

	sockfnptrs[FN_RECV]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "recv");
	sockfnptrs[FN_SEND]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "send");
	sockfnptrs[FN_LISTEN]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "listen");
	sockfnptrs[FN_ACCEPT]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "accept");
	sockfnptrs[FN_RECVFROM]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "recvfrom");
	sockfnptrs[FN_SENDTO]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "sendto");
	sockfnptrs[FN_SELECT]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "select");
	sockfnptrs[FN_CONNECT]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "connect");
	sockfnptrs[FN_CLOSESOCKET]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "closesocket");
	/* we don't use this */
	sockfnptrs[FN_SHUTDOWN]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "shutdown");
	sockfnptrs[FN_GETHOSTBYNAME]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "gethostbyname");
	sockfnptrs[FN_HTONS]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "htons");
	sockfnptrs[FN_SOCKET]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock, "socket");
	/* in winsock 1, this will simply be 0 */
	sockfnptrs[FN_WSASENDDISCONNECT]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock,
						   "WSASendDisconnect");
	sockfnptrs[FN_SOCKETAVAILABLE]
	    = (int (PASCAL FAR *)())GetProcAddress(hWinsock,
						   "ioctlsocket");
    }	    

    sysAssert(sockfnptrs[FN_RECV] != NULL);
    sysAssert(sockfnptrs[FN_SEND] != NULL);
    sysAssert(sockfnptrs[FN_LISTEN] != NULL);
    sysAssert(sockfnptrs[FN_ACCEPT] != NULL);
    sysAssert(sockfnptrs[FN_RECVFROM] != NULL);
    sysAssert(sockfnptrs[FN_SENDTO] != NULL);
    sysAssert(sockfnptrs[FN_SELECT] != NULL);
    sysAssert(sockfnptrs[FN_CONNECT] != NULL);
    sysAssert(sockfnptrs[FN_CLOSESOCKET] != NULL);
    sysAssert(sockfnptrs[FN_SHUTDOWN] != NULL);
    sysAssert(sockfnptrs[FN_GETHOSTBYNAME] != NULL);
    sysAssert(sockfnptrs[FN_HTONS] != NULL);
    sysAssert(sockfnptrs[FN_SOCKET] != NULL);
    
    if (winsock2Available) {
	sysAssert(sockfnptrs[FN_WSASENDDISCONNECT] != NULL);
    }

    sysAssert(sockfnptrs[FN_SOCKETAVAILABLE] != NULL);

    sockfnptrs_initialized = TRUE;
    mutexUnlock(&sockFnTableMutex);
}

/*
 * If we get a nonnull function pointer it might still be the case
 * that some other thread is in the process of initializing the socket
 * function pointer table, but our pointer should still be good.
 */
int
sysListen(int fd, long count) {
    int (PASCAL FAR *listenfn)();
    if ((listenfn = sockfnptrs[FN_LISTEN]) == NULL) {
	initSockFnTable();
	listenfn = sockfnptrs[FN_LISTEN];
    }
    sysAssert(sockfnptrs_initialized == TRUE && listenfn != NULL);
    return (*listenfn)(fd, count);
}

int
sysConnect(int fd, struct sockaddr *name, int namelen) {
    int (PASCAL FAR *connectfn)();
    if ((connectfn = sockfnptrs[FN_CONNECT]) == NULL) {
	initSockFnTable();
	connectfn = sockfnptrs[FN_CONNECT];
    }
    sysAssert(sockfnptrs_initialized == TRUE);
    sysAssert(connectfn != NULL);
    return (*connectfn)(fd, name, namelen);
}

int
sysAccept(int fd, struct sockaddr *name, int *namelen) {
    int (PASCAL FAR *acceptfn)();
    if ((acceptfn = sockfnptrs[FN_ACCEPT]) == NULL) {
	initSockFnTable();
	acceptfn = sockfnptrs[FN_ACCEPT];
    }
    sysAssert(sockfnptrs_initialized == TRUE && acceptfn != NULL);
    return (*acceptfn)(fd, name, namelen);
}

int
sysRecvFrom(int fd, char *buf, int nBytes,
                  int flags, struct sockaddr *from, int *fromlen) {
    int (PASCAL FAR *recvfromfn)();
    if ((recvfromfn = sockfnptrs[FN_RECVFROM]) == NULL) {
	initSockFnTable();
	recvfromfn = sockfnptrs[FN_RECVFROM];
    }
    sysAssert(sockfnptrs_initialized == TRUE && recvfromfn != NULL);
    return (*recvfromfn)(fd, buf, nBytes, flags, from, fromlen);
}

int
sysSendTo(int fd, char *buf, int len,
                int flags, struct sockaddr *to, int tolen) {
    int (PASCAL FAR *sendtofn)();
    if ((sendtofn = sockfnptrs[FN_SENDTO]) == NULL) {
	initSockFnTable();
	sendtofn = sockfnptrs[FN_SENDTO];
    }
    sysAssert(sockfnptrs_initialized == TRUE && sendtofn != NULL);
    return (*sendtofn)(fd, buf, len, flags, to, tolen);
}

int
sysRecv(int fd, char *buf, int nBytes, int flags) {
    int (PASCAL FAR *recvfn)();
    if ((recvfn = sockfnptrs[FN_RECV]) == NULL) {
	initSockFnTable();
	recvfn = sockfnptrs[FN_RECV];
    }
    sysAssert(sockfnptrs_initialized == TRUE && recvfn != NULL);
    return (*recvfn)(fd, buf, nBytes, flags);
}

int
sysSend(int fd, char *buf, int nBytes, int flags) {
    int (PASCAL FAR *sendfn)();
    if ((sendfn = sockfnptrs[FN_SEND]) == NULL) {
	initSockFnTable();
	sendfn = sockfnptrs[FN_SEND];
    }
    sysAssert(sockfnptrs_initialized == TRUE && sendfn != NULL);
    return (*sendfn)(fd, buf, nBytes, flags);
}

struct hostent *
sysGetHostByName(char *hostname) {
    int (PASCAL FAR *sendfn)();
    if ((sendfn = sockfnptrs[FN_GETHOSTBYNAME]) == NULL) {
	initSockFnTable();
	sendfn = sockfnptrs[FN_GETHOSTBYNAME];
    }
    sysAssert(sockfnptrs_initialized == TRUE && sendfn != NULL);
    return (struct hostent *)
	(*(struct hostent *(PASCAL FAR *)(char *))sendfn)(hostname);
}

unsigned short
sysHostToNetworkShort(unsigned short hostshort) {
    int (PASCAL FAR *sendfn)();
    if ((sendfn = sockfnptrs[FN_HTONS]) == NULL) {
	initSockFnTable();
	sendfn = sockfnptrs[FN_HTONS];
    }
    sysAssert(sockfnptrs_initialized == TRUE && sendfn != NULL);
    return (*sendfn)((int)hostshort);
}

int
sysSocket(int domain, int type, int protocol) {
    int (PASCAL FAR *socketfn)();
    if ((socketfn = sockfnptrs[FN_SOCKET]) == NULL) {
	initSockFnTable();
	socketfn = sockfnptrs[FN_SOCKET];
    }
    sysAssert(sockfnptrs_initialized == TRUE && socketfn != NULL);
    return (*socketfn)(domain, type, protocol);
}

/*
 * This function is carefully designed to work around a bug in Windows
 * 95's networking winsock. Please see the beginning of this file for
 * a complete description of the problem.
 */
int sysSocketClose(int fd) {

    if (fd > 0) {
	int (PASCAL FAR *closesocketfn)();
	int (PASCAL FAR *wsasenddisconnectfn)();
	int dynamic_ref = -1;

	if ((closesocketfn = sockfnptrs[FN_CLOSESOCKET]) == NULL) {
	    initSockFnTable();
	}
	/* At this point we are guaranteed the sockfnptrs are initialized */
    	sysAssert(sockfnptrs_initialized == TRUE);

	closesocketfn = sockfnptrs[FN_CLOSESOCKET];
	sysAssert(closesocketfn != NULL);

	if (winsock2Available) {
	    wsasenddisconnectfn = sockfnptrs[FN_WSASENDDISCONNECT];
	    (void) (*wsasenddisconnectfn)(fd, NULL);
	}
        (void) (*closesocketfn)(fd);
    }
    return TRUE;
}

/*
 * Poll the fd for reading for timeout ms.  Returns 1 if something's
 * ready, 0 if it timed out, -1 on error, -2 if interrupted (although
 * interruption isn't implemented yet).  Timeout in milliseconds.  */
int
sysTimeout(int fd, long timeout) {
    int res;
    fd_set tbl;
    struct timeval t;
    int (PASCAL FAR *selectfn)();

    t.tv_sec = timeout / 1000;
    t.tv_usec = (timeout % 1000) * 1000;
    FD_ZERO(&tbl);
    FD_SET(fd, &tbl);

    if ((selectfn = sockfnptrs[FN_SELECT]) == NULL) {
	initSockFnTable();
	selectfn = sockfnptrs[FN_SELECT];
    }
    sysAssert(sockfnptrs_initialized == TRUE && selectfn != NULL); 
    res = (*selectfn)(fd + 1, &tbl, 0, 0, &t);
    return res;
}

long
sysSocketAvailable(int fd, long *pbytes)
{
    int (PASCAL FAR *socketfn)();
    if ((socketfn = sockfnptrs[FN_SOCKETAVAILABLE]) == NULL) {
	initSockFnTable();
	socketfn = sockfnptrs[FN_SOCKETAVAILABLE];
    }
    sysAssert(sockfnptrs_initialized == TRUE && socketfn != NULL);
    return (*socketfn)(fd, FIONREAD, pbytes);
}
