/*
 * @(#)javai_md.c	1.4 98/09/15
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <windows.h>
#include <string.h>

#include "oobj.h"
#include "interpreter.h"
#include "sys_api.h"

#ifdef DEBUG
#define HPI_DLL "hpi_g.dll"
#else
#define HPI_DLL "hpi.dll"
#endif

HINSTANCE hJavaInst;

BOOL WINAPI
DllMain(HINSTANCE hinst, DWORD reason, LPVOID reserved)
{
    switch (reason) {
        case DLL_PROCESS_ATTACH:
            hJavaInst = hinst;
            break;
        default:
            break;
    }
    return TRUE;
}


void * InitializeHPI(void *args)
{
    char buf[MAX_PATH];
    void *hpi_handle;
    jint (JNICALL * DLL_Initialize)(GetInterfaceFunc *, void *);
    GetInterfaceFunc getintf;

    GetModuleFileName(hJavaInst, buf, MAX_PATH);
    *(strrchr(buf, '\\')) = '\0';
    *(strrchr(buf, '\\')) = '\0';
    strcat(buf, "\\" HPI_DLL);
    hpi_handle = LoadLibrary(buf);
    if (hpi_handle == NULL) {
        return NULL;
    }
    DLL_Initialize = (jint (JNICALL *)(GetInterfaceFunc *, void *))
        GetProcAddress(hpi_handle, "DLL_Initialize");
    if (DLL_Initialize == NULL ||
        (*DLL_Initialize)(&getintf, args) < 0) {
        return NULL;
    }
    return getintf;
}

/*
 * THE FOLLOWING DEFINITION IS A FUNCTION CALL IN ORDER TO WORK AROUND
 * AN OPTIMIZER BUG IN MSVC++ 2.1
 */
int64_t ll_shl(int64_t i, int bits)
{
    return i << bits;
}

int32_t float2l(float f)
{
    int64_t lltmp = (int64_t)f;
    int32_t ltmp = (int32_t)lltmp;
    if (ltmp == lltmp) {
        return ltmp;
    } else {
        if (_isnan((double)f)) {return 0;}
        if (f < 0) {
            return 0x80000000;
        } else {
            return 0x7fffffff;
        }
    }
}

int32_t double2l(double f)
{
    int64_t lltmp = (int64_t)f;
    int32_t ltmp = (int32_t)lltmp;
    if (ltmp == lltmp) {
        return ltmp;
    } else {
        if (_isnan(f)) {return 0;}
        if (f < 0) {
            return 0x80000000;
        } else {
            return 0x7fffffff;
        }
    }
}

int64_t float2ll(float f)
{
    int64_t lltmp = (int64_t)f;
    if (lltmp != 0x8000000000000000) {
        return lltmp;
    } else {
        if (_isnan((double)f)) {return 0;}
        if (f < 0) {
            return 0x8000000000000000;
        } else {
            return 0x7fffffffffffffff;
        }
    }
}

int64_t double2ll(double f)
{
    int64_t lltmp = (int64_t)f;
    if (lltmp != 0x8000000000000000) {
        return lltmp;
    } else {
        if (_isnan(f)) {return 0;}
        if (f < 0) {
            return 0x8000000000000000;
        } else {
            return 0x7fffffffffffffff;
        }
    }
}

/* This function gets called very early, before VM_CALLS are setup.
 * Do not use any of the VM_CALLS entries!!!
 */
props_md_t *
GetPropertiesMD()
{
    static props_md_t sprops = {0};
    char *v; /* tmp var */

    if (sprops.dll_dir) {
        return &sprops;
    }

    /* sysclasspath, java_home, dll_dir */
    {
        char *home_path;
        char *sysclasspath;
        char *dll_path;
        char home_dir[MAX_PATH];
        char *alt_java_home = getenv("_ALT_JAVA_HOME_DIR");

        if (alt_java_home) {
            strncpy(home_dir, alt_java_home, MAX_PATH);
        } else {
            GetModuleFileName(hJavaInst, home_dir, MAX_PATH);
            *(strrchr(home_dir, '\\')) = '\0';
            *(strrchr(home_dir, '\\')) = '\0';
            *(strrchr(home_dir, '\\')) = '\0';
        }

        home_path = malloc(strlen(home_dir) + 1);
        if (home_path == NULL)
            return NULL;
        strcpy(home_path, home_dir);
        sprops.java_home = home_path;

        dll_path = malloc(strlen(home_dir) + 10);
        if (dll_path == NULL)
            return NULL;
        strcpy(dll_path, home_dir);
        strcat(dll_path, "\\bin");
        sprops.dll_dir = dll_path;

        sysclasspath = malloc(strlen(home_dir) * 3 + 75);
        if (sysclasspath == NULL)
            return NULL;
        sprintf(sysclasspath,
                "%s\\lib\\rt.jar;%s\\lib\\i18n.jar;%s\\classes",
                home_dir, home_dir, home_dir);
        sprops.sysclasspath = sysclasspath;
    }

    {

        /* Win32 library search order (See the documentation for LoadLibrary):
	 *
	 * 1. The directory from which application is loaded.
	 * 2. The current directory
	 * 3. System directory (GetSystemDirectory)
	 * 4. Windows directory (GetWindowsDirectory)
	 * 5. The PATH environment variable
	 */

        char *library_path;
	char tmp[MAX_PATH];
	char *path_str = getenv("PATH");

	library_path = malloc(MAX_PATH * 4 + (path_str ? strlen(path_str) : 0) + 10);

	library_path[0] = '\0';

	GetModuleFileName(NULL, tmp, sizeof(tmp));
	*(strrchr(tmp, '\\')) = '\0';
	strcat(library_path, tmp);

	strcat(library_path, ";.");

	GetSystemDirectory(tmp, sizeof(tmp));
	strcat(library_path, ";");
	strcat(library_path, tmp);

	GetWindowsDirectory(tmp, sizeof(tmp));
	strcat(library_path, ";");
	strcat(library_path, tmp);
	
	if (path_str) {
	    strcat(library_path, ";");
	    strcat(library_path, path_str);
	}

	sprops.library_path = strdup(library_path);
	free(library_path);
    }

    /* Default extensions directory */
    {
	char buf[MAXPATHLEN + 10];
	sprintf(buf, "%s\\lib\\ext", sprops.java_home);
	sprops.ext_dirs = strdup(buf);
    }

    return &sprops;
}

#ifdef JCOV
void sleep(int t)
{
    Sleep(t * 1000);
}
#endif
