/*
 * @(#)java_props_md.c	1.7 98/09/30
 *
 * Copyright 1994-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <windows.h>
#include <shlobj.h>
#include <objidl.h>
#include <locale.h>
#include <sys/types.h>
#include <sys/timeb.h>

#include "locale_str.h"
#include "timezone_map.h"
#include "java_props.h"

#ifndef VER_PLATFORM_WIN32_WINDOWS
#define VER_PLATFORM_WIN32_WINDOWS 1
#endif

java_props_t *
GetJavaProperties()
{
    static java_props_t sprops = {0};
    char *v; /* tmp var */

    if (sprops.user_dir) {
        return &sprops;
    }

    /* AWT properties */
    sprops.awt_toolkit = "sun.awt.windows.WToolkit";

    /* tmp dir */
    {
        char tmpdir[MAX_PATH + 1];
        /* we might want to check that this succeed */
        GetTempPath(MAX_PATH + 1, tmpdir);
        sprops.tmp_dir = strdup(tmpdir);
    }

    /* Printing properties */
    sprops.printerJob = "sun.awt.windows.WPrinterJob";

    /* Java2D properties */
    sprops.graphics_env = "sun.awt.Win32GraphicsEnvironment";

    {
        char buf[max(512, MAX_PATH+1)];
        char *path = getenv("JAVA_FONTS");
        if (path != 0) {
            if (sprintf(buf, "%s", path) != -1) {
                v = buf; /* sysNativePath(buf); Is this necessary? */
            } else {
                v = "";
            }
        } else {
            v = "";
        }
        sprops.font_dir = strdup(v);
    }

    /* OS properties */
    {
        char buf[100];
        OSVERSIONINFO ver;    
	ver.dwOSVersionInfoSize = sizeof(ver);
	GetVersionEx(&ver);
	switch (ver.dwPlatformId) {
	case VER_PLATFORM_WIN32s:
	    sprops.os_name = "Windows 3.1";
	    break;
	case VER_PLATFORM_WIN32_WINDOWS:
	    sprops.os_name = "Windows 95";
	    break;
	case VER_PLATFORM_WIN32_NT:
	    sprops.os_name = "Windows NT";
	    break;
	}
	sprintf(buf, "%d.%d", ver.dwMajorVersion, ver.dwMinorVersion);
	sprops.os_version = strdup(buf);
#if _X86_
        sprops.os_arch = "x86";
#else
        sprops.os_arch = "unknown";
#endif
    }

    /* User name */
    {
        char buf[100];
        int buflen = sizeof(buf);
        sprops.user_name =
            GetUserName(buf, &buflen) ? strdup(buf) : "unknown";
    }

    /* Home directory (New scheme, see bug 4100238) */
    {
        /* Given user name XXX:
         * On multi-user NT, user.home gets set to c:\winnt\profiles\XXX.
         * On multi-user Win95, user.home gets set to c:\windows\profiles\XXX.
         * On single-user Win95, user.home gets set to c:\windows.
         */
        LPITEMIDLIST itemList;

        if (SUCCEEDED(SHGetSpecialFolderLocation(NULL,
                                                 CSIDL_DESKTOPDIRECTORY,
                                                 &itemList))) {
            char szUserPath[MAX_PATH];
            IMalloc *pMalloc;
            char *p;
            SHGetPathFromIDList(itemList, szUserPath);
            if (SUCCEEDED(SHGetMalloc(&pMalloc))) {
                pMalloc->lpVtbl->Free(pMalloc, itemList);
                pMalloc->lpVtbl->Release(pMalloc);
            }
            /* Get the parent of Desktop directory */
            p = strrchr(szUserPath, '\\');
            if (p) {
                *p = '\0';
            }
            sprops.user_home = strdup(szUserPath);
        } else {
            sprops.user_home = "C:\\";
        }
    }

    /*
     *  user.language
     *  user.region (if user's environmant specify this)
     *  file.encoding
     *  file.encoding.pkg
     */
    {
        /*
         * query the system for the current system default locale
         * (which is a Windows LCID value)
         */
        LCID systemLocale = GetThreadLocale();

        /*
         * binary-search our list of LCID values.  If we don't find the 
	 * one we're looking for, mask out the country code and try again 
	 * with just the language code
         */
        {
            int index = -1;
            int tries = 0;
            char* localeID;
            do {
                int lo, hi, mid;
                lo = 0;
                hi = sizeof(localeIDMap) / sizeof(LCIDtoLocale);
                while (index == -1 && lo < hi) {
                    mid = (lo + hi) / 2;
                    if (localeIDMap[mid].winID == systemLocale)
                        index = mid;
                    else if (localeIDMap[mid].winID > systemLocale)
                        hi = mid;
                    else
                        lo = mid + 1;
                }
                systemLocale &= 0x03ff;
                ++tries;
            } while (index == -1 && tries < 2);

            /*
             * if we didn't find the LCID that the system returned to us,
	     * we don't have a Java locale ID that corresponds to it. 
	     * Fall back on en_US.
             */
            if (index == -1) {
                sprops.language = "en";
                sprops.region = "US";
            }

            /* otherwise, look up the corresponding Java locale ID from the list
             * of Java locale IDs and set up the system properties accordingly.
             * [Two notes: Some of the Java IDs we look up may contain not only
             * a language and country, but also a variant.  If this is the case,
             * we leave the variant with the country (i.e., country_variant) and
             * store it in user.region.  This works because code has been added
             * to Locale for euro support to recognize a user.region value of
             * the form country_variant or _variant during initialization of the
             * default locale from system properties.  We don't use a new
             * property ("user.variant") because we don't want to make the VM to
             * Java property-based communication pipe any wider than it is
             * already.  Second, we are indeed stomping over the actual static
             * tables here to split the locale ID into language and country
             * codes.  Since this function is the only one that uses that table,
             * and it's only called once at system init time, this should be
             * okay, but if either assumption changes, we'll have to re-think
             * this.]  --rtg 5/4/98 --liu 7/27/98 */
            else {
                char* lang;
                char* ctry;

                localeID = localeIDMap[index].javaID;
                lang = localeID;
                ctry = lang;

                sprops.language = lang;
                while (*ctry != '_' && *ctry != 0)
                    ++ctry;

                if (*ctry == '_') {
                    *ctry++ = 0;
                }
                sprops.language = lang;
                sprops.region = ctry;
            }
        }

        {
            char *fallback_encoding = NULL;
            char locale_name[128];
            int i;
            if (*sprops.region != 0)
                sprintf(locale_name, "%s_%s", sprops.language, sprops.region);
            else
                strcpy(locale_name, sprops.language);

            for (i = 0; strcmp(encoding_names[i], ""); i += 2) {
                if (!strcmp(encoding_names[i], locale_name)) {
                    sprops.encoding = encoding_names[i + 1];
                    break;
                }
                else if (fallback_encoding == NULL &&
                         !strcmp(encoding_names[i], sprops.language)) {
                    fallback_encoding = encoding_names[i + 1];
                }
            }

            if (sprops.encoding == NULL)
                sprops.encoding = fallback_encoding;

	    /* If we can't deduce an encoding, fall back to Cp1252 */
            if (sprops.encoding == NULL)
                sprops.encoding = "Cp1252";
        }
    }

    sprops.unicode_encoding = "UnicodeLittle";
    /* User TIMEZONE */
    {
        /* In determining the host time zone, we use three steps.  We try to
         * give the most accurate result possible, and at the very least to get
         * the correct offset, so Java clocks show the right time.
         *
         * First, we try to look up the zone name in a table of English names.
         * This only works on English language installations of Windows.
         * Second, we use a language-independent lookup table that indexes by
         * raw offset, DST, and country code.  If this lookup also fails, our
         * third and final fallback is to synthesize and ID of the form
         * GMT[+-]hh:mm.  This allows us, in all cases (assuming the call to
         * GetTimeZoneInformation() succeeds) to set the Java clock correctly.
         *
         * See also ../include/locale_str.h */

        TIME_ZONE_INFORMATION tzi;
        char *std_timezone = NULL;

        if (GetTimeZoneInformation(&tzi) == -1) {
            /* The call to GetTimeZoneInformation failed.  We can't do anything
             * at all in this case.  We can't even synthesize a GMT+hh:mm name,
             * since we don't have the offset.  */
            std_timezone = "GMT";
        }
        else { /* Call to GetTimeZoneInformation() succeeded. */
            /* First, try to look up the zone name by name in a table (see
             * locale_str.h).  If we can do this, it's more accurate than any
             * other approach.  However, this only works in an English
             * installation of Windows.  */
            char buf[32];
            int i;

            for (i = 0; i < 32; i++) {
                if (tzi.StandardName[i] == 0){
                    buf[i] = '\0';
                    break;
                }
                buf[i] = (char)(tzi.StandardName[i]);
            }

            /* Linear search through name table for exact match.  This only
             * works in English installations of Windows, but is the best
             * approach if it works. */
            for (i = 0; i < timezone_names_length; i += 2) {
                if (!strcmp(timezone_names[i], buf)) {
                    std_timezone = timezone_names[i + 1];
                    break;
                }
            }

            if (std_timezone == NULL) {
                /* We failed to find the zone in our name table.  In that case,
                 * we use our fallback algorithm, which relies only on the
                 * offset, presence/absence of DST, and the country.  These
		 * data are chosen to be independent of OS localization.  For
                 * localized flavors of Windows, we will wind up here. */
                int match, offset;
                enum EHasDST hasDST;

                hasDST = tzi.DaylightBias == 0 ? kStandard : kDaylight;
                offset = -tzi.Bias; /* Conventions reversed between Win32
				       & us */
                match = -1;

                /* Loop to find the beginning of the block of entries matching
                 * this offset.  Once we find it, we then start scanning 
		 * through the entries in that block.
                 *
                 * If successful, 'match' will be >= 0.  */
                i = 0;
                while (i < timezone_map_length
                       && timezone_map[i].offset < offset) {
                    ++i;
                }
                while (i < timezone_map_length
                       && timezone_map[i].offset == offset) {
                    /* NULL country (at end of section) matches any country; this
                     * provides a default zone. E.g., in Japan, GMT-8 should be
                     * recognized as PST, even though there is no JA entry for
                     * GMT-8. */
                    if (((timezone_map[i].country == NULL)
                         || (sprops.region != NULL
                             && (strcmp(timezone_map[i].country, sprops.region) == 0)))
                        && (timezone_map[i].hasDST == kAny
                            || timezone_map[i].hasDST == hasDST)) {
                        match = i;
                        break;
                    }
                    ++i;
                }

                if (match >= 0) {
                    std_timezone = timezone_map[match].zone;
                }
                else {
                    /* If we fail to find the offset in our table (VERY
                     * unlikely) we then synthesize an ID of the form
                     * GMT[+|-]hh:mm.  This offset INCLUDES the effect of DST,
                     * so the correct local time can be derived.  However, the
                     * local time will be incorrect as soon as we cross a
                     * DST/Standard boundary, which could at any time (e.g., in
                     * 5 minutes).
                     *
                     * [A better solution might be to encode the rule, since we
                     * know what it is, in the identifier, in the style of the
                     * UNIX TZ environment variable.  This is currently not
                     * supported.] */
                    int h, m;
                    char sign = '+';
                    struct _timeb tstruct;

                    /* _ftime() is the only way to determine whether or not DST
                     * is in effect right now; there isn't a direct Win32 call
                     * to do this.  */
                    _tzset();
                    _ftime(&tstruct);

                    if (tstruct.dstflag) offset -= tzi.DaylightBias; /* add DST bias */
                    if (offset < 0) { offset = -offset; sign = '-'; }
                    h = offset / 60;
                    m = offset % 60;
                    std_timezone = malloc(32); /* Allow for pathologically large h */
                    sprintf(std_timezone, "GMT%c%02d:%02d", sign, h, m);
                }
            }
        }
        sprops.timezone = std_timezone;
    }

    /* Current directory */
    {
        char buf[MAX_PATH];
        GetCurrentDirectory(sizeof(buf), buf);
        sprops.user_dir = strdup(buf);
    }

    sprops.file_separator = "\\";
    sprops.path_separator = ";";
    sprops.line_separator = "\r\n";

    return &sprops;
}
