/*
 * @(#)PlainDatagramSocketImpl.c	1.29 98/08/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <windows.h>
#include <winsock.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>

#include "java_net_PlainDatagramSocketImpl.h"
#include "java_net_SocketOptions.h"

#include "jvm.h"
#include "jni_util.h"
#include "net_util.h"

#define IN_CLASSD(i)    (((long)(i) & 0xf0000000) == 0xe0000000)
#define IN_MULTICAST(i) IN_CLASSD(i)

/************************************************************************
 * PlainDatagramSocketImpl
 */

static jfieldID IO_fd_fdID;

jfieldID pdsi_fdID;
jfieldID pdsi_timeoutID;
jfieldID pdsi_localPortID;

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_init(JNIEnv *env, jclass cls) {

    pdsi_fdID = (*env)->GetFieldID(env, cls, "fd", "Ljava/io/FileDescriptor;");
    pdsi_timeoutID = (*env)->GetFieldID(env, cls, "timeout", "I");
    pdsi_localPortID = (*env)->GetFieldID(env, cls, "localPort", "I");
    cls = (*env)->FindClass(env, "java/io/FileDescriptor");
    IO_fd_fdID = NET_GetFileDescriptorID(env);
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_bind(JNIEnv *env, jobject this,
					   jint port, jobject addressObj) {
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    struct sockaddr_in lcladdr;
    int lcladdrlen = sizeof(lcladdr);
    int address;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(addressObj)) {
	JNU_ThrowNullPointerException(env, "argument address");
	return;
    } else {
	address = (*env)->GetIntField(env, addressObj, ia_addressID);
    }
    /* bind - pick a port number for local addr*/
    memset((char *) &lcladdr, 0, sizeof(lcladdr));
    lcladdr.sin_family      = AF_INET;
    lcladdr.sin_port        = htons((short)port);
    lcladdr.sin_addr.s_addr = htonl(address);
    if (bind(fd, (struct sockaddr *)&lcladdr, sizeof(lcladdr)) == -1) {
        NET_ThrowCurrent(env, "Cannot bind");
	return;
    }

    /* find what port system picked for us - obviously brain dead interface,
       should have told me in the bind call itself */
    if (getsockname(fd, (struct sockaddr *)&lcladdr, &lcladdrlen) == -1) {
	NET_ThrowCurrent(env, "getsockname");
	return;
    }
    (*env)->SetIntField(env, this, pdsi_localPortID, ntohs(lcladdr.sin_port));
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    send
 * Signature: (Ljava/net/DatagramPacket;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_send(JNIEnv *env, jobject this,
					   jobject packet) {

    char BUF[MAX_BUFFER_LEN];
    char *fullPacket;
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint fd;

    jobject iaObj;
    jint address;
    jint family;

    jint packetBufferOffset, packetBufferLen, packetPort;
    jbyteArray packetBuffer;

    long n;
    struct sockaddr_in rmtaddr;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"Socket closed");
	return;
    }

    if (IS_NULL(packet)) {
	JNU_ThrowNullPointerException(env, "null packet");
	return;
    }

    fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    iaObj = (*env)->GetObjectField(env, packet, dp_addressID);

    packetPort = (*env)->GetIntField(env, packet, dp_portID);
    packetBufferOffset = (*env)->GetIntField(env, packet, dp_offsetID);
    packetBuffer = (jbyteArray)(*env)->GetObjectField(env, packet, dp_bufID);

    if (IS_NULL(iaObj) || IS_NULL(packetBuffer)) {
	JNU_ThrowNullPointerException(env, "null address || null buffer");
	return;
    }

    packetBufferLen = (*env)->GetIntField(env, packet, dp_lengthID);
    address = (*env)->GetIntField(env, iaObj, ia_addressID);
    family = (*env)->GetIntField(env, iaObj, ia_familyID);

    rmtaddr.sin_port = htons((short)packetPort);
    rmtaddr.sin_addr.s_addr = htonl(address);
    rmtaddr.sin_family = family;

    if (packetBufferLen > MAX_BUFFER_LEN) {
	/* When JNI-ifying the JDK's IO routines, we turned
	 * read's and write's of byte arrays of size greater
	 * than 2048 bytes into several operations of size 2048.
	 * This saves a malloc()/memcpy()/free() for big
	 * buffers.  This is OK for file IO and TCP, but that
	 * strategy violates the semantics of a datagram protocol.
	 * (one big send) != (several smaller sends).  So here
	 * we *must* alloc the buffer.  Note it needn't be bigger
	 * than 65,536 (0xFFFF) the max size of an IP packet.
	 * anything bigger is truncated anyway.
	 */
	fullPacket = (char *)malloc(packetBufferLen);
	if (!fullPacket) {
	    JNU_ThrowOutOfMemoryError(env, "heap allocation failed");
	    return;
	}
    } else {
	fullPacket = &(BUF[0]);
    }

    (*env)->GetByteArrayRegion(env, packetBuffer, packetBufferOffset, packetBufferLen,
			       (jbyte *)fullPacket);

    switch (JVM_SendTo(fd, fullPacket, packetBufferLen, 0,
		       (struct sockaddr *)&rmtaddr,
		       sizeof(rmtaddr))) {

        case JVM_IO_ERR:
	    NET_ThrowCurrent(env, "Datagram send failed");
	    break;
        case JVM_IO_INTR:
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "operation interrupted");
    }

    if (packetBufferLen > MAX_BUFFER_LEN) {
	free(fullPacket);
    }
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    peek
 * Signature: (Ljava/net/InetAddress;)I
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_peek(JNIEnv *env, jobject this,
					   jobject addressObj) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint timeout = (*env)->GetIntField(env, this, pdsi_timeoutID);
    jint fd;

    /* The address and family fields of addressObj */
    jint address, family;

    int n;
    struct sockaddr_in remote_addr;
    jint remote_addrsize = sizeof (remote_addr);
    char buf[1];

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
        if (fd < 0) {
           JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
                           "socket closed");
           return -1;
        }
    }
    if (IS_NULL(addressObj)) {
	JNU_ThrowNullPointerException(env, "Null address in peek()");
    } else {
	address = (*env)->GetIntField(env, addressObj, ia_addressID);
	family = (*env)->GetIntField(env, addressObj, ia_familyID);
    }
    if (timeout) {
	int ret = JVM_Timeout(fd, timeout);
	if (ret == 0) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "Peek timed out");
	    return ret;
	} else if (ret == JVM_IO_ERR) {
	    NET_ThrowCurrent(env, "timeout in datagram socket peek");
	    return ret;
	} else if (ret == JVM_IO_INTR) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "operation interrupted");
	    return ret;
	}
    }
    n = JVM_RecvFrom(fd, buf, 1, MSG_PEEK,
		     (struct sockaddr *)&remote_addr, &remote_addrsize);

    if (n == JVM_IO_ERR && WSAGetLastError() != WSAEMSGSIZE) {
	NET_ThrowCurrent(env, "Datagram peek failed");
        return 0;
    }
    if (n == JVM_IO_INTR) {
        JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException", 0);
	return 0;
    }
    (*env)->SetIntField(env, addressObj, ia_addressID,
			ntohl(remote_addr.sin_addr.s_addr));
    (*env)->SetIntField(env, addressObj, ia_familyID, remote_addr.sin_family);

    /* return port */
    return ntohs(remote_addr.sin_port);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    receive
 * Signature: (Ljava/net/DatagramPacket;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_receive(JNIEnv *env, jobject this,
					      jobject packet) {

    char BUF[MAX_BUFFER_LEN];
    char *fullPacket;
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint timeout = (*env)->GetIntField(env, this, pdsi_timeoutID);

    jbyteArray packetBuffer;
    jint packetBufferOffset, packetBufferLen;

    int fd, errorCode;
    jbyteArray data;

    int datalen;
    int n;
    struct sockaddr_in remote_addr;
    jint remote_addrsize = sizeof (remote_addr);

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    }

    fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);

    if (IS_NULL(packet)) {
	JNU_ThrowNullPointerException(env, "packet");
	return;
    }

    packetBuffer = (*env)->GetObjectField(env, packet, dp_bufID);

    if (IS_NULL(packetBuffer)) {
	JNU_ThrowNullPointerException(env, "packet buffer");
        return;
    }

    packetBufferOffset = (*env)->GetIntField(env, packet, dp_offsetID);
    packetBufferLen = (*env)->GetIntField(env, packet, dp_lengthID);

    if (timeout) {
	int ret = JVM_Timeout(fd, timeout);
        if (ret == 0) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "Receive timed out");
	    return;
	} else if (ret == JVM_IO_ERR) {
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			    "Socket closed");
	    return;
	} else if (ret == JVM_IO_INTR) {
	    JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			    "operation interrupted");
	    return;
	}
    }

    if (packetBufferLen > MAX_BUFFER_LEN) {

	/* When JNI-ifying the JDK's IO routines, we turned
	 * read's and write's of byte arrays of size greater
	 * than 2048 bytes into several operations of size 2048.
	 * This saves a malloc()/memcpy()/free() for big
	 * buffers.  This is OK for file IO and TCP, but that
	 * strategy violates the semantics of a datagram protocol.
	 * (one big send) != (several smaller sends).  So here
	 * we *must* alloc the buffer.  Note it needn't be bigger
	 * than 65,536 (0xFFFF) the max size of an IP packet.
	 * anything bigger is truncated anyway.
	 */
	fullPacket = (char *)malloc(packetBufferLen);
	if (!fullPacket) {
	    JNU_ThrowOutOfMemoryError(env, "heap allocation failed");
	    return;
	}
    } else {
	fullPacket = &(BUF[0]);
    }

    n = JVM_RecvFrom(fd, fullPacket, packetBufferLen, 0,
		     (struct sockaddr *)&remote_addr, &remote_addrsize);
    /* truncate the data if the packet's length is too small */
    if (n > packetBufferLen) {
	n = packetBufferLen;
    }
    if (n < 0) {
	errorCode = WSAGetLastError();
	/* check to see if it's because the buffer was too small */
	if (errorCode == WSAEMSGSIZE) {
	    /* it is because the buffer is too small. It's UDP, it's
	     * unreliable, it's all good. discard the rest of the
	     * data..
	     */
	    n = packetBufferLen;
	} else {
	    /* failure */
	    (*env)->SetIntField(env, packet, dp_lengthID, 0);
	}
    }
    if (n == -1) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "socket closed");
    } else if (n == -2) {
	JNU_ThrowByName(env, JNU_JAVAIOPKG "InterruptedIOException",
			"operation interrupted");
    } else if (n < 0) {
	NET_ThrowCurrent(env, "Datagram receive failed");
    } else {
	/* success - fill in received address... */
	jclass ia_clazz = (*env)->FindClass(env, "java/net/InetAddress");
	jmethodID ia_ctor = (*env)->GetMethodID(env, ia_clazz, "<init>", "()V");
	jobject packetAddress = (*env)->NewObject(env, ia_clazz, ia_ctor);
	(*env)->SetIntField(env, packetAddress, ia_addressID,
			    ntohl(remote_addr.sin_addr.s_addr));
	(*env)->SetIntField(env, packetAddress, ia_familyID,
			    remote_addr.sin_family);
	/* stuff the new Inetaddress in the packet */
	(*env)->SetObjectField(env, packet, dp_addressID, packetAddress);
	/* and fill in the data, remote address/port and such */
	(*env)->SetByteArrayRegion(env, packetBuffer, packetBufferOffset, n,
				   (jbyte *)fullPacket);
	(*env)->SetIntField(env, packet, dp_portID,
			    ntohs(remote_addr.sin_port));
	(*env)->SetIntField(env, packet, dp_lengthID, n);
	(*env)->SetIntField(env, packetAddress, ia_addressID,
			    ntohl(remote_addr.sin_addr.s_addr));
	(*env)->SetIntField(env, packetAddress, ia_familyID,
			    remote_addr.sin_family);
    }
    if (packetBufferLen > MAX_BUFFER_LEN) {
	free(fullPacket);
    }
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    datagramSocketCreate
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_datagramSocketCreate(JNIEnv *env,
							   jobject this) {
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    int arg = -1;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return;
    } else {
	fd =  JVM_Socket(AF_INET, SOCK_DGRAM, 0);
    }
    if (fd == JVM_IO_ERR) {
	NET_ThrowCurrent(env, "Socket creation failed");
	return;
    }
    {
    BOOL t = TRUE;
    setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char*)&t, sizeof(BOOL));
    }
    (*env)->SetIntField(env, fdObj, IO_fd_fdID, fd);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    datagramSocketClose
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_datagramSocketClose(JNIEnv *env,
							  jobject this) {
    /*
     * REMIND: PUT A LOCK AROUND THIS CODE
     */
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
        fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    JVM_SocketClose(fd);
    (*env)->SetIntField(env, fdObj, IO_fd_fdID, -1);
}

/*
 * datagramsocket options
 */

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    socketSetOption
 * Signature: (ILjava/lang/Object;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_socketSetOption(JNIEnv *env,jobject this,
						      jint opt,jobject value) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(value)) {
	JNU_ThrowNullPointerException(env, "value argument");
	return;
    }
    if (fd < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "socket closed");
    } else if (opt == java_net_SocketOptions_IP_MULTICAST_IF) {
        int address = (*env)->GetIntField(env, value, ia_addressID);
	struct in_addr in;
	in.s_addr = htonl(address);
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF,
		       (const char*)&in, sizeof(in)) < 0) {
	    NET_ThrowCurrent(env, "set IP_MULTICAST_IF failed");
	}
    } else if (opt == java_net_SocketOptions_SO_REUSEADDR) {
	jclass iCls = (*env)->FindClass(env, "java/lang/Integer");
	jfieldID i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	int tmp = (*env)->GetIntField(env, value, i_valueID);
	int arg = tmp ? -1: 0;
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		       (char *)&arg, sizeof(arg)) < 0) {
	    NET_ThrowCurrent(env, "set REUSEADDR failed");
	}
    } else if (opt == java_net_SocketOptions_SO_SNDBUF) {
	jclass iCls = (*env)->FindClass(env, "java/lang/Integer");
	jfieldID i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	int arg = (*env)->GetIntField(env, value, i_valueID);
	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF,
		       (char *)&arg, sizeof(arg)) < 0) {
	    NET_ThrowCurrent(env, "set SO_SNDBUF failed");
	}
    }  else if (opt == java_net_SocketOptions_SO_RCVBUF) {
	jclass iCls = (*env)->FindClass(env, "java/lang/Integer");
	jfieldID i_valueID = (*env)->GetFieldID(env, iCls, "value", "I");
	int arg = (*env)->GetIntField(env, value, i_valueID);
	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF,
		       (char *)&arg, sizeof(arg)) < 0) {
	    NET_ThrowCurrent(env, "set SO_RCVBUF failed");
	}
    } else {
	JNU_ThrowByName(env, "java.net.SocketException",
			"invalid DatagramSocket option");
    }
    return;
}

/*
 * Returns relevant info as a jint.
 *
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    socketGetOption
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_socketGetOption(JNIEnv *env, jobject this,
						      jint opt) {
    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd, ret;
    ret = 0;

    if (IS_NULL(fdObj)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "Socket closed");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }

    if (fd < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", "socket closed");
    } else if (opt == java_net_SocketOptions_IP_MULTICAST_IF) {
	struct in_addr in;
	int len = sizeof(struct in_addr);
	if (getsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF,
		       (char *)&in, &len) < 0) {
	    NET_ThrowCurrent(env, "get IP_MULTICAST_IF failed");
	    return -1;
	}
	return ntohl(in.s_addr);
    } else if (opt == java_net_SocketOptions_SO_BINDADDR) {
	/* find out local IP address, return as 32 bit value */
	struct sockaddr_in him;
	int len = sizeof(struct sockaddr_in);
	memset(&him, 0, len);
	if (getsockname(fd, (struct sockaddr *)&him, &len) == -1) {
	    NET_ThrowCurrent(env, "getsockopt() SO_BINDADDR");
	    return -1;
	}
	return htonl(him.sin_addr.s_addr);
    } else if (opt == java_net_SocketOptions_SO_SNDBUF) {
	int buflen = 0;
	int len = sizeof(ret);
	if (getsockopt(fd, SOL_SOCKET, SO_SNDBUF,
		       (char *)&buflen, &len) < 0) {
	    NET_ThrowCurrent(env, "getsockopt() SO_SNDBUF");
	    return -1;
	}
	return buflen;
    } else if (opt == java_net_SocketOptions_SO_RCVBUF) {
	int buflen = 0;
	int len = sizeof(ret);
	if (getsockopt(fd, SOL_SOCKET, SO_RCVBUF,
		       (char *)&buflen, &len) < 0) {
	    NET_ThrowCurrent(env, "getsockopt() SO_RCVBUF");
	    return -1;
	}
	return buflen;
    } else {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			"invalid DatagramSocket option");
	return -1;
    }
}

/*
 * Multicast-related calls
 */

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    setTimeToLive
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_setTimeToLive(JNIEnv *env, jobject this,
						    jint ttl) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    int fd;
    int ittl = (int)ttl;

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }

    /* setsockopt to be correct ttl */
    if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (char*)&ittl,
		   sizeof (ittl)) < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(WSAGetLastError()));
    }
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    setTTL
 * Signature: (B)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_setTTL(JNIEnv *env, jobject this,
					     jbyte ttl) {
    Java_java_net_PlainDatagramSocketImpl_setTimeToLive(env, this, (jint)ttl);
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    getTimeToLive
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_getTimeToLive(JNIEnv *env, jobject this) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint fd = -1;

    int ttl = 0;
    int len = sizeof(ttl);
    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return -1;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    /* getsockopt of ttl */
    if (getsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, (char*)&ttl, &len) < 0) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(WSAGetLastError()));
	return -1;
    }
    return (jint)ttl;
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    getTTL
 * Signature: ()B
 */
JNIEXPORT jbyte JNICALL
Java_java_net_PlainDatagramSocketImpl_getTTL(JNIEnv *env, jobject this) {
    int result = Java_java_net_PlainDatagramSocketImpl_getTimeToLive(env, this);
    if (result > 127) {
	return (jbyte)127;
    } else {
	return (jbyte)result;
    }
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    join
 * Signature: (Ljava/net/InetAddress;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_join(JNIEnv *env, jobject this,
					   jobject iaObj) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint fd, address;

    struct sockaddr_in name;
    struct ip_mreq mname;

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(iaObj)) {
	JNU_ThrowNullPointerException(env, "address");
	return;
    } else {
	address = (*env)->GetIntField(env, iaObj, ia_addressID);
    }


    /* Set the multicast group address in the ip_mreq field
     * eventually this check should be done by the security manager
     */
    mname.imr_multiaddr.s_addr = htonl(address);
    if (!IN_MULTICAST(address)) {
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException",
			strerror(WSAGetLastError()));
    }
    /* Set the interface to be default */
    mname.imr_interface.s_addr = INADDR_ANY;

    /* Join the multicast group */
    if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &mname,
		   sizeof (mname)) < 0) {
	NET_ThrowCurrent(env, "set socket option failed");
    }
    return;
}

/*
 * Class:     java_net_PlainDatagramSocketImpl
 * Method:    leave
 * Signature: (Ljava/net/InetAddress;)V
 */
JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_leave(JNIEnv *env, jobject this,
					    jobject address) {

    jobject fdObj = (*env)->GetObjectField(env, this, pdsi_fdID);
    jint fd, addressAddress;

    struct sockaddr_in name;
    struct ip_mreq mname;

    if (IS_NULL(fdObj)) {
	JNU_ThrowNullPointerException(env, "fd object");
	return;
    } else {
	fd = (*env)->GetIntField(env, fdObj, IO_fd_fdID);
    }
    if (IS_NULL(address)) {
	JNU_ThrowNullPointerException(env, "address argument");
	return;
    } else {
	addressAddress = (*env)->GetIntField(env, address, ia_addressID);
    }
    /*
     * Set the multicast group address in the ip_mreq field
     * eventually this check should be done by the security manager
     */
    mname.imr_multiaddr.s_addr = htonl(addressAddress);
    if (!IN_MULTICAST(addressAddress)) {
	NET_ThrowCurrent(env, "multicast leave failed");
    }
    /* Set the interface to be default */
    mname.imr_interface.s_addr = INADDR_ANY;

    /* Join the multicast group */
    if (setsockopt(fd, IPPROTO_IP, IP_DROP_MEMBERSHIP,(char *) &mname,
	   sizeof (mname)) < 0) {
	NET_ThrowCurrent(env, "multicast leave failed");
    }
    return;
}

