/*
 * @(#)net_util_md.c	1.17 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <winsock.h>

#include "net_util.h"
#include "jni.h"

/*
 * Initialize Windows Sockets API support
 */
BOOL WINAPI
DllMain(HINSTANCE hinst, DWORD reason, LPVOID reserved)
{
    WSADATA wsadata;
    OSVERSIONINFO info;
    switch (reason) {
	case DLL_PROCESS_ATTACH:
	    if (WSAStartup(MAKEWORD(1,1), &wsadata) != 0) {
		/* wrong version */
		return FALSE;
	    }
	    break;

	case DLL_PROCESS_DETACH:
	    WSACleanup();
	    break;

	default:
	    break;
    }
    return TRUE;
}

/*
 * Since winsock doesn't have the equivalent of strerror(errno), this
 * function will do a big switch on WSAGetLastError() and we can put a correct
 * description in the message of a SocketException and raise it.
 */
void 
NET_ThrowNew(JNIEnv *env, int errorNum, char *msg) 
{
    char fullMsg[256];
    if ((*env)->ExceptionOccurred(env)) {
	return;
    }
    if (!msg) {
	msg = "no further information";
    }
    switch(errorNum) {
    case WSAEPROTOTYPE:
        JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", 
			"Invalid protocol type");
	break;
    case WSAENOPROTOOPT:
    case WSAEPROTONOSUPPORT:
    case WSAESOCKTNOSUPPORT:
    case WSAEOPNOTSUPP:
    case WSAEPFNOSUPPORT:
    case WSAEAFNOSUPPORT:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Option unsupported by protocol: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAENOTSOCK:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Descriptor not a socket: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAECONNRESET:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Connection reset by peer: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAEADDRINUSE:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Address in use: %s", msg);
        JNU_ThrowByName(env, JNU_JAVANETPKG "BindException", fullMsg);
	break;
    case WSAEADDRNOTAVAIL:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Cannot assign requested address: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "BindException", fullMsg);
	break;
    case WSAESHUTDOWN:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Connection shutdown: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAEISCONN:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Already connected: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAENOTCONN:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Not connected: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAETIMEDOUT:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Operation timed out: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "NoRouteToHostException", fullMsg);
	break;
    case WSAECONNREFUSED:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Connection refused: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "ConnectException", fullMsg);
	break;
    case WSAEHOSTUNREACH:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Host unreachable: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "NoRouteToHostException", fullMsg);
	break;
    case WSAEHOSTDOWN:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "Host down: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAEINPROGRESS:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "blocking winsock call in progress: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAEFAULT:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "argp parameter not part of user address space: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSAENETDOWN:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "network subsystem has failed: %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    case WSANOTINITIALISED:
	jio_snprintf(fullMsg, sizeof(fullMsg),
		     "net dll not initialized (WSAStartup): %s", msg);
	JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", fullMsg);
	break;
    default: {
	    char fullMsg[2048];
	    if (!msg) {
		msg = "Socket error occurred";
	    }
	    
	    jio_snprintf(fullMsg, sizeof(fullMsg), 
			 "%s (code=%d)", msg, GetLastError());
	    JNU_ThrowByName(env, JNU_JAVANETPKG "SocketException", 
			    fullMsg);
	}
    }
    return;
}

void 
NET_ThrowCurrent(JNIEnv *env, char *msg) 
{
    NET_ThrowNew(env, WSAGetLastError(), msg);
}

void
NET_ThrowSocketException(JNIEnv *env, char* msg) 
{
    static jclass cls = NULL;
    if (cls == NULL) {
	cls = (*env)->FindClass(env, "java/net/SocketException");
	if (cls == NULL) {
	    return;
	}
	cls = (*env)->NewGlobalRef(env, cls);
    }
    (*env)->ThrowNew(env, cls, msg);
}
	

jfieldID
NET_GetFileDescriptorID(JNIEnv *env)
{
    jclass cls = (*env)->FindClass(env, "java/io/FileDescriptor");
    return (*env)->GetFieldID(env, cls, "fd", "I");
}
