/*
 * @(#)HAE_API_WinOS_Thread.c	1.5 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

/*****************************************************************************/
/*
**	HAE_API_WinOS_Thread.c
**
**	This provides platform-specfic frame-thread-managment functions for
**  Windows 95/NT.
**
**	Overview:
**
**	Written by Moe
**
**	History	-
**	12/1/97		Created
*/
/*****************************************************************************/

#if JAVA_THREAD
#if !JAVA_SOUND
	#error JAVA_THREAD defined when JAVA_SOUND not defined
#endif
#else


#ifndef WIN32_EXTRA_LEAN
	#define WIN32_EXTRA_LEAN
#endif
#ifndef WIN32_LEAN_AND_MEAN
	#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>

#if JAVA_SOUND
#include <jni.h>
#endif

#include "HAE_API.h"


////////////////////////////////////////////////// MACROS:

#if _DEBUG		//CLS?? Why isn't this in some shared header file?
	#define DEBUG_STR(x)	printf(x)
#else
	#define DEBUG_STR(x)
#endif


////////////////////////////////////////////////// STATIC VARIABLES:

// Thread used to build audio frames
static HANDLE			theFrameThread = NULL;
// The function to call 
static HAE_FrameThreadProc	theFrameProc = NULL;


////////////////////////////////////////////////// FORWARDS:

static DWORD WINAPI PV_NativeFrameThreadProc(LPVOID lpv);
static void*	PV_CreateAudioContext(void);
static void		PV_DestroyAudioContext(void);


////////////////////////////////////////////////// THREAD-API FUNCTIONS:

// Create and start the frame thread
int HAE_CreateFrameThread(void* context, HAE_FrameThreadProc proc)
{
	if (proc)
	{
	DWORD	threadID;

		theFrameProc = proc;
		theFrameThread = CreateThread(NULL, 0L, 
										PV_NativeFrameThreadProc, 0L, 
										0L, &threadID);
		if (theFrameThread)
		{
			// THREAD_PRIORITY_HIGHEST
			// THREAD_PRIORITY_NORMAL
			// THREAD_PRIORITY_TIME_CRITICAL
			SetThreadPriority(theFrameThread, THREAD_PRIORITY_TIME_CRITICAL);
			return 0;
		}
		theFrameProc = NULL;
	}
	return -1;
}
// Stop and destroy the frame thread, 
int HAE_DestroyFrameThread(void* context)
{
	if (theFrameThread)
	{
	DWORD		result;

		result = WaitForSingleObject(theFrameThread, 500);
		// If thread takes too long, kill it
		if (result == WAIT_TIMEOUT)
		{
			TerminateThread(theFrameThread, 0);
		}
		CloseHandle(theFrameThread);
		theFrameThread = NULL;
		theFrameProc = NULL;
		return 0;
	}
	DEBUG_STR("PV_NativeFrameThreadProc() called when no thread started\n");
	return -1;
}
// Make the frame thread sleep for the given number of milliseconds
int HAE_SleepFrameThread(void* context, long msec)
{
	Sleep(msec);
	return  0;
}

static DWORD WINAPI PV_NativeFrameThreadProc(LPVOID lpv)
{
	if (theFrameProc)
	{
		(*theFrameProc)(PV_CreateAudioContext());
		PV_DestroyAudioContext();
	}
	else
	{
		DEBUG_STR("PV_NativeFrameThreadProc() called when thread supposed to be dead!");
	}
	return 0;
}


////////////////////////////////////////////////// HELPER FUNCTIONS:

#if JAVA_SOUND

static JavaVM* PV_GetCurrentVM(void)
{
JavaVM*		vmArray[2];
jsize		count;
jint		result;

	count = 0;
	result = JNI_GetCreatedJavaVMs(vmArray, 2, &count);	// ask for 2
	if ((result == 0) && (count >= 1))						// error if not just 1
	{
		if (count > 1)
		{
			DEBUG_STR("WARNING:  JNI_GetCreatedJavaVMs() returned >1 VMs\n");
		}
		return vmArray[0];
	}
	DEBUG_STR("JNI_GetCreatedJavaVMs() returned error\n");
	return NULL;
}
static void* PV_CreateAudioContext(void)
{
JavaVM*		const vm = PV_GetCurrentVM();

	if (vm)
	{
	JNIEnv*				env;
	JDK1_1AttachArgs	args;	// whatever
	jint				result;

		result = (*vm)->AttachCurrentThread(vm, &env, &args);
		if (result >= 0)
		{
			return (void*)env;
		}
		DEBUG_STR("AttachCurrentThread() failed\n");
	}
	return NULL;
}
static void PV_DestroyAudioContext(void)
{
JavaVM*		const vm = PV_GetCurrentVM();

	if (vm)
	{
	jint		result;

		result = (*vm)->DetachCurrentThread(vm);
		if (result < 0)
		{
			DEBUG_STR("DetachCurrentThread() failed\n");
		}
	}
}

#else

static void * PV_CreateAudioContext(void)
{
	return NULL;
}
static void PV_DestroyAudioContext(void)
{
}

#endif


#endif	// !JAVA_THREAD
// EOF of HAE_API_WinOS_Thread.c
