/*
 * @(#)CmdIDList.cpp	1.4 98/09/15
 *
 * Copyright 1996, 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "CmdIDList.h"

static const UINT MAX_WIN_CMD_ID = 65535;

AwtCmdIDList::AwtCmdIDList() 
{
    m_size = 0;
    m_capacity = 10;
    m_array = (AwtObject**)malloc(m_capacity * sizeof(void*));
    ASSERT(m_array != NULL);
}

AwtCmdIDList::~AwtCmdIDList() 
{
    free(m_array);
}

// Add an object to the end of the vector, increasing its capacity if necessary
UINT AwtCmdIDList::Add(AwtObject* obj)
{
    CriticalSection::Lock l(m_lock);
    if (m_size == m_capacity) {
        m_capacity += 10;
        m_array = (AwtObject**)realloc(m_array, m_capacity * sizeof(void*));
        ASSERT(m_array != NULL);
    }
    UINT id = m_size++;
    if (id >= MAX_WIN_CMD_ID) {
    // part of fix 4133279
    // if we exceed Windows limit of 64K for an id (unlikely, but possible), recycle command id's
	id = m_size = 0;
    }
    m_array[id] = obj;
    return id;
}

// Return the associated object.
AwtObject* AwtCmdIDList::Lookup(UINT id)
{
    CriticalSection::Lock l(m_lock);
    ASSERT(id < m_size);
    return m_array[id];
}

// Erase the object association, but make no attempt to reclaim the space.
// (We'll never allocate that many ids that it will be an issue.)
void AwtCmdIDList::Remove(UINT id)
{
    CriticalSection::Lock l(m_lock);
    m_array[id] = NULL;
}
