/*
 * @(#)awt.h	1.22 98/07/14
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef _AWT_H_
#define _AWT_H_

#include "stdhdrs.h"

#undef ASSERT
#undef VERIFY
#undef UNIMPLEMENTED
#undef INLINE

#if defined(DEBUG)

extern void CallDebugger(char* expr, char* file, unsigned line);
extern void DumpJavaStack();

#define ASSERT(exp) \
    (void)( (exp) || (CallDebugger(#exp, __FILE__, __LINE__), 0) )
#define VERIFY(exp)     ASSERT(exp)
#define UNIMPLEMENTED() ASSERT(FALSE)

#define AWT_TRACE(__args) if (AwtToolkit::GetInstance().Verbose()) printf __args

/* Disable inlining. */
#define INLINE

#else

#define ASSERT(exp) ((void)0)
#define UNIMPLEMENTED() \
    SignalError(0, JAVAPKG "NullPointerException","unimplemented");

/*
 * VERIFY macro -- assertion where expression is always evaluated 
 * (normally used for BOOL functions).
 */
#define VERIFY(exp) ((void)(exp))

/* Enable inlining. */
#define INLINE inline
# define AWT_TRACE(__args) 

#endif  /* !DEBUG */

extern COLORREF DesktopColor2RGB(int colorIndex);

#define JNI_CHECK_NULL_GOTO(obj, msg, where) {                            \
    if (obj == NULL) {                                                    \
        JNU_ThrowNullPointerException(env, msg);                          \
        goto where;                                                       \
    }                                                                     \
}

#define JNI_CHECK_PEER_GOTO(peer, where) {                                \
    JNI_CHECK_NULL_GOTO(peer, "peer", where);                             \
    pData = JNI_GET_PDATA(peer);                                          \
    if (pData == NULL) {                                                  \
        JNU_ThrowNullPointerException(env, "null pData");                 \
        goto where;                                                       \
    }                                                                     \
}

#define JNI_CHECK_NULL_RETURN(obj, msg) {                                 \
    if (obj == NULL) {                                                    \
        JNU_ThrowNullPointerException(env, msg);                          \
        return;                                                           \
    }                                                                     \
}

#define JNI_CHECK_PEER_RETURN(peer) {                                     \
    JNI_CHECK_NULL_RETURN(peer, "peer");                                  \
    pData = JNI_GET_PDATA(peer);                                          \
    if (pData == NULL) {                                                  \
        JNU_ThrowNullPointerException(env, "null pData");                 \
        return;                                                           \
    }                                                                     \
}

#define JNI_CHECK_PEER_CREATION_RETURN(peer) {                            \
    JNI_CHECK_NULL_RETURN(peer, "peer");                                  \
    pData = JNI_GET_PDATA(peer);                                          \
    if (pData == NULL) {                                                  \
        JNU_ThrowOutOfMemoryError(env, "null pData");                     \
        return;                                                           \
    }                                                                     \
}

#define JNI_CHECK_NULL_RETURN_NULL(obj, msg) {                            \
    if (obj == NULL) {                                                    \
        JNU_ThrowNullPointerException(env, msg);                          \
        return 0;                                                         \
    }                                                                     \
}

#define JNI_CHECK_PEER_RETURN_NULL(peer) {                                \
    JNI_CHECK_NULL_RETURN_NULL(peer, "peer");                             \
    pData = JNI_GET_PDATA(peer);                                          \
    if (pData == NULL) {                                                  \
        JNU_ThrowNullPointerException(env, "null pData");                 \
        return 0;                                                         \
    }                                                                     \
}

#define JNI_GET_PDATA(peer) env->GetIntField(peer, AwtObject::pDataID)

#define JNI_SET_PDATA(peer, data) env->SetIntField(peer,                  \
						   AwtObject::pDataID,    \
						   (long)data)
/*  /NEW JNI */


/* (from Generic.c) Make it easier to determine appropriate code paths: */
#if defined (WIN32)
	#define IS_WIN32 TRUE
#else
	#define IS_WIN32 FALSE
#endif
#define IS_NT      (IS_WIN32 && !(::GetVersion() & 0x80000000))
#define IS_WIN32S  (IS_WIN32 && !IS_NT && LOBYTE(LOWORD(::GetVersion())) < 4)
#define IS_WIN95   (IS_WIN32 && !IS_NT && LOBYTE(LOWORD(::GetVersion())) >= 4)
#define IS_WIN4X   (IS_WIN32 && LOBYTE(::GetVersion()) >= 4)
#define IS_WINVER_ATLEAST(maj, min) \
                   ((maj) < LOBYTE(LOWORD(::GetVersion())) || \
                      (maj) == LOBYTE(LOWORD(::GetVersion())) && \
                      (min) <= HIBYTE(LOWORD(::GetVersion())))

/*
 * macros to crack a LPARAM into two ints -- used for signed coordinates,
 * such as with mouse messages.
 */
#define LO_INT(l)           ((int)(short)(l))
#define HI_INT(l)           ((int)(short)(((DWORD)(l) >> 16) & 0xFFFF))

extern JavaVM *jvm;

#endif  /* _AWT_H_ */
