/*
 * @(#)awt_AWTEvent.cpp	1.3 98/10/07
 * 
 * Copyright 1993-1997 Sun Microsystems, Inc. 901 San Antonio Road, 
 * Palo Alto, California, 94303, U.S.A.  All Rights Reserved.
 * 
 * This software is the confidential and proprietary information of Sun
 * Microsystems, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Sun.
 * 
 * CopyrightVersion 1.2
 * 
 */

#include "awt_AWTEvent.h"
#include "awt.h"

/************************************************************************
 * AwtAWTEvent fields
 */

jfieldID AwtAWTEvent::dataID;
jfieldID AwtAWTEvent::idID;
jfieldID AwtAWTEvent::consumedID;

/************************************************************************
 * AwtEvent native methods
 */

extern "C" {

JNIEXPORT void JNICALL 
Java_java_awt_AWTEvent_initIDs(JNIEnv *env, jclass cls) {
    AwtAWTEvent::dataID = env->GetFieldID(cls, "data", "J");
    AwtAWTEvent::idID = env->GetFieldID(cls, "id", "I");
    AwtAWTEvent::consumedID = env->GetFieldID(cls, "consumed", "Z");

    ASSERT(AwtAWTEvent::dataID != NULL);
    ASSERT(AwtAWTEvent::idID != NULL);
    ASSERT(AwtAWTEvent::consumedID != NULL);
}

/*
 * Class:     java_awt_AWTEvent
 * Method:    freeNativeData
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_AWTEvent_freeNativeData(JNIEnv *env, jobject event)
{
    JNI_CHECK_NULL_RETURN(event, "null AWTEvent");

    jlong data = (env)->GetLongField(event, AwtAWTEvent::dataID);
    if (data != 0) {
	MSG* pMsg = (MSG*)data;
	(env)->SetLongField(event, AwtAWTEvent::dataID, 0);
	delete pMsg;
    }
}

/*
 * Class:     java_awt_AWTEvent
 * Method:    copyDataFieldInto
 * Signature: (Ljava/awt/AWTEvent;)V
 *
 * Allocates a copy of the data instance variable and copies it
 * from thisEvent into thatEvent. If data is null, thatEvent's
 * data field is also set to NULL.
 */
JNIEXPORT void JNICALL
Java_java_awt_AWTEvent_copyDataFieldInto
  (JNIEnv *env, jobject thisEvent, jobject thatEvent)
{
    JNI_CHECK_NULL_RETURN(thatEvent, "null AWTEvent");

    MSG *pThisData = (MSG *)(env)->GetLongField(thisEvent, AwtAWTEvent::dataID);
    MSG *pThatData = (MSG *)(env)->GetLongField(thatEvent, AwtAWTEvent::dataID);
    if ((pThisData == NULL) && (pThatData == NULL))
	return; // They're both NULL; we're done
    if (pThatData != NULL) {
	// pThatData points to a MSG struct that should be deleted
	delete pThatData;
	pThatData = NULL;
    }
    if (pThisData != NULL) {
	// Allocate a new MSG struct and copy it into pThatData
	pThatData = new MSG();
	memcpy(pThatData, pThisData, sizeof(MSG));
    }
    (env)->SetLongField(thatEvent, AwtAWTEvent::dataID, (jlong)pThatData);
} /* AWTEvent_copyDataFieldInto() */

} /* extern "C" */
