/*
 * @(#)awt_ActiveEvent.cpp	1.4 98/09/13
 *
 * Copyright 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_ActiveEvent.h"
#include "awt_Object.h"

#include <sun_awt_windows_WActiveEvent.h>

CriticalSection AwtActiveEvent::sm_lock;


void AwtActiveEvent::SendEvent(AwtObject *poster)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    CriticalSection::Lock l(sm_lock);
    if (env->EnsureLocalCapacity(2) < 0) {
        return;
    }

    static jclass wActiveEventCls;
    if (wActiveEventCls == NULL) {
        jclass wActiveEventClsLocal =
	    env->FindClass("sun/awt/windows/WActiveEvent");
	if (!wActiveEventClsLocal) {
	    /* exception already thrown */
	    return;
	}
	wActiveEventCls = (jclass)env->NewGlobalRef(wActiveEventClsLocal);
	env->DeleteLocalRef(wActiveEventClsLocal);
    }

    static jmethodID wActiveEventConst;
    if (wActiveEventConst == NULL) {
        wActiveEventConst = 
            env->GetMethodID(wActiveEventCls, "<init>",
		     "(Ljava/lang/Object;Lsun/awt/windows/WObjectPeer;J)V");
	ASSERT(wActiveEventConst);
    }

    jobject target = poster->GetTarget(env);
    jobject wActiveEvent = env->NewObject(wActiveEventCls, wActiveEventConst,
					  target, poster->GetPeer(env),
					  (jlong) Clone());
    ASSERT(!env->ExceptionOccurred());
    ASSERT(wActiveEvent != NULL);

    poster->SendEvent(wActiveEvent);
    env->DeleteLocalRef(target);
    env->DeleteLocalRef(wActiveEvent);
}

/************************************************************************
 * WActiveEvent native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WActiveEvent
 * Method:    dispatchImpl
 * Signature: (J)V
 */

JNIEXPORT void JNICALL
Java_sun_awt_windows_WActiveEvent_dispatchImpl(JNIEnv *env, jobject self,
					       jlong pData)
{
    ASSERT(!::IsBadReadPtr( (const void *) pData, sizeof(AwtActiveEvent)));
    ((AwtActiveEvent *) pData)->Dispatch();
}

/*
 * Class:     sun_awt_windows_WActiveEvent
 * Method:    finalizeImpl
 * Signature: (J)V
 */

JNIEXPORT void JNICALL
Java_sun_awt_windows_WActiveEvent_finalizeImpl(JNIEnv *env, jobject self,
					       jlong pData)
{
    ASSERT(!::IsBadReadPtr( (const void *) pData, sizeof(AwtActiveEvent)));
    delete ((AwtActiveEvent *) pData);
}

} /* extern "C" */
