/*
 * @(#)awt_Button.cpp	1.27 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <jni.h>

#include "awt_Object.h"    /* wop_pDataID */
#include "awt_Toolkit.h"
#include "awt_Button.h"
#include "awt_Canvas.h"
#include "awt_Graphics.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtButton fields
 */

/* java.awt.Button fields */
jfieldID AwtButton::labelID;


/************************************************************************
 * AwtButton methods
 */

AwtButton::AwtButton() {
}

/* System provided button class */
const char* AwtButton::GetClassName() {
    return "BUTTON";
}

/* Create a new AwtButton object and window. */
AwtButton* AwtButton::Create(jobject self, jobject parent)
{
    /* the result */
    AwtButton *c = NULL;

    jstring label = NULL;
    const char *labelStr;
    DWORD style;
    DWORD exStyle = 0;
    int x, y, height, width;

    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (env->EnsureLocalCapacity(2) < 0) {
	return NULL;
    }

    int pData;
    AwtCanvas* awtParent;
    jobject target = 0;

    JNI_CHECK_PEER_GOTO(parent, done);
    awtParent = (AwtCanvas*)pData;
    JNI_CHECK_NULL_GOTO(awtParent, "awtParent", done)

    target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "target", done);

    c = new AwtButton();
    JNI_CHECK_NULL_GOTO(c, "AwtButton() failed", done);

    label = (jstring)env->GetObjectField(target, AwtButton::labelID);
    x = env->GetIntField(target, AwtComponent::xID);
    y = env->GetIntField(target, AwtComponent::yID);
    width = env->GetIntField(target, AwtComponent::widthID);
    height = env->GetIntField(target, AwtComponent::heightID);

    labelStr = JNU_GetStringPlatformChars(env, label, JNI_FALSE);
    style = 0;

    if (labelStr == NULL) {
	/* out of memory */
	goto done;
    }

    style = WS_CHILD | WS_CLIPSIBLINGS | BS_PUSHBUTTON | BS_OWNERDRAW;
    if (GetRTLReadingOrder())
        exStyle |= WS_EX_RTLREADING;

    c->CreateHWnd(env, labelStr, style, exStyle, x, y, width, height,
		  awtParent->GetHWnd(),
		  (HMENU)awtParent->CreateControlID(),
		  ::GetSysColor(COLOR_BTNTEXT),
		  ::GetSysColor(COLOR_BTNFACE),
		  self);
    c->m_backgroundColorSet = TRUE;  /* suppress inheriting parent's color. */
    JNU_ReleaseStringPlatformChars(env, label, labelStr);

done:
    env->DeleteLocalRef(target);
    env->DeleteLocalRef(label);

    return c;
}

MsgRouting
AwtButton::WmNotify(UINT notifyCode)
{
    if (notifyCode == BN_CLICKED) {
        DoCallback("handleAction", "()V");
    }
    return mrDoDefault;
}

MsgRouting
AwtButton::WmDrawItem(UINT /*ctrlId*/, DRAWITEMSTRUCT far& drawInfo)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (env->EnsureLocalCapacity(3) < 0) {
	/* is this OK? */
	return mrConsume;
    }

    jobject self = GetPeer(env);
    jobject target = env->GetObjectField(self, AwtObject::targetID);

    HDC hDC = drawInfo.hDC;
    RECT rect = drawInfo.rcItem;
    UINT nState;
    SIZE size;

    /* Draw Button */
    nState = DFCS_BUTTONPUSH;
    if (drawInfo.itemState & ODS_SELECTED)
        nState |= DFCS_PUSHED;

    ::FillRect(hDC, &rect, GetBackgroundBrush());
    UINT edgeType = (nState & DFCS_PUSHED) ? EDGE_SUNKEN : EDGE_RAISED;
    ::DrawEdge(hDC, &rect, edgeType, BF_RECT | BF_SOFT);

    /* Draw WindowText */
    jobject font = GET_FONT(target, self);
    jstring str = (jstring)env->GetObjectField(target, AwtButton::labelID);

    size = AwtFont::getMFStringSize(hDC, font, str);

    /* Check whether the button is disabled. */
    BOOL bEnabled = isEnabled();

    int adjust = (nState & DFCS_PUSHED) ? 1 : 0;
    int x = (rect.left + rect.right-size.cx) / 2 + adjust;
    int y = (rect.top + rect.bottom-size.cy) / 2 + adjust;

    if (bEnabled) {
        AwtComponent::DrawWindowText(hDC, font, str, x, y);
    } else {
        AwtComponent::DrawGrayText(hDC, font, str, x, y);
    }

    /* Draw focus rect */
    if (drawInfo.itemState & ODS_FOCUS){
        const int inf = 3; /* heuristic decision */
        RECT focusRect;
        VERIFY(::CopyRect(&focusRect, &rect));
	VERIFY(::InflateRect(&focusRect,-inf,-inf));
        VERIFY(::DrawFocusRect(hDC, &focusRect));
    }

    /* Notify any subclasses */
    DoCallback("handlePaint", "(IIII)V", rect.left, rect.top,
               rect.right-rect.left, rect.bottom-rect.top);

    env->DeleteLocalRef(target);
    env->DeleteLocalRef(font);
    env->DeleteLocalRef(str);

    return mrConsume;
}

MsgRouting AwtButton::WmPaint(HDC)
{
    /* Suppress peer notification, because it's handled in WmDrawItem. */
    return mrDoDefault;
}


/************************************************************************
 * WButtonPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WButtonPeer
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WButtonPeer_initIDs(JNIEnv *env, jclass cls)
{
    cls = env->FindClass("java/awt/Button");
    if (cls == NULL) {
	return;
    }
    AwtButton::labelID = env->GetFieldID(cls, "label", "Ljava/lang/String;");
    ASSERT(AwtButton::labelID != NULL);
}

/*
 * Class:     sun_awt_windows_WButtonPeer
 * Method:    setLabel
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WButtonPeer_setLabel(JNIEnv *env, jobject self,
					  jstring label)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* c = (AwtComponent*)pData;
    const char *labelStr = JNU_GetStringPlatformChars(env, label, JNI_FALSE);
    c->SetText(labelStr);
    JNU_ReleaseStringPlatformChars(env, label, labelStr);
}

/*
 * Class:     sun_awt_windows_WButtonPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WButtonPeer_create(JNIEnv *env, jobject self,
					jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);

    AwtToolkit::CreateComponent(
        self, parent, (AwtToolkit::ComponentFactory)AwtButton::Create);

    JNI_CHECK_PEER_CREATION_RETURN(self);
}

}  /* extern "C" */


