/*
 * @(#)awt_Canvas.cpp	1.49 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_Canvas.h"
#include "awt_Image.h"
#include "awt_Graphics.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtCanvas methods
 */

AwtCanvas::AwtCanvas() {
    m_eraseBackground = FALSE;
}

AwtCanvas::~AwtCanvas() {
}

const char* AwtCanvas::GetClassName() {
  return "SunAwtCanvas";
}

/*
 * Create a new AwtCanvas object and window.
 */
AwtCanvas* AwtCanvas::Create(jobject self, jobject hParent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    AwtComponent* parent;
    jobject target = 0;
    AwtCanvas *canvas;

    JNI_CHECK_NULL_GOTO(hParent, "null hParent", done);

    parent = (AwtComponent*)(env->GetIntField(hParent, AwtObject::pDataID));
    JNI_CHECK_NULL_GOTO(parent, "null parent", done);

    target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    canvas = new AwtCanvas();
    JNI_CHECK_NULL_GOTO(canvas, "AwtCanvas() failed", done);

    {
        jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	DWORD exStyle = 0;

	if (GetRTL())
	    exStyle |= WS_EX_RIGHT;

	canvas->CreateHWnd(env, "",
			   WS_CHILD | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,
			   exStyle,
			   x, y, width, height,
			   parent->GetHWnd(),
			   NULL,
			   ::GetSysColor(COLOR_WINDOWTEXT),
			   ::GetSysColor(COLOR_WINDOW),
			   self);
    }
 done:
    env->DeleteLocalRef(target);

    return canvas;
}

/*
 * This routine is duplicated in AwtWindow.
 */
MsgRouting AwtCanvas::WmEraseBkgnd(HDC hDC, BOOL& didErase)
{
    m_eraseBackground = TRUE;
    didErase = TRUE;
    return mrConsume;
}

/*
 * This routine is duplicated in AwtWindow.
 */
MsgRouting AwtCanvas::WmPaint(HDC)
{
    PAINTSTRUCT ps;
    GetSharedDC()->Lock();
    GetSharedDC()->GrabDC(NULL);
    HDC hDC = ::BeginPaint(GetHWnd(), &ps);// the passed-in HDC is ignored.
    ASSERT(hDC);
    RECT& r = ps.rcPaint;
    GetSharedDC()->Unlock();

    if ((r.right-r.left) > 0 && (r.bottom-r.top) > 0 &&
        m_peerObject != NULL && m_callbacksEnabled) {
        DoCallback("handleExpose",      // BUGID 4070488 repaint
                   "(IIII)V", r.left, r.top,
                   r.right-r.left, r.bottom-r.top);
    }
    m_eraseBackground = FALSE;
    VERIFY(::EndPaint(GetHWnd(), &ps));

    return mrConsume;
}

MsgRouting AwtCanvas::WmMouseDown(UINT flags, int x, int y, int button)
{
    // fixed 4035189: We capture the mouse on button down, since capturing
    // in WM_MOUSEMOVE doesn't work if mouse is already outside window
    // rkhan 4/28/98
    ASSERT(::GetCapture() == NULL || ::GetCapture() == GetHWnd()); // don't want to interfere with other controls
    if (::GetCapture() != GetHWnd()) {
	::SetCapture(GetHWnd());
    }
    return AwtComponent::WmMouseDown(flags, x, y, button);
}

MsgRouting AwtCanvas::WmMouseUp(UINT flags, int x, int y, int button)
{
    BOOL	isMouseCaptured = (::GetCapture() == GetHWnd());
    BOOL	areButtonsReleased = (flags & ALL_MK_BUTTONS) == 0;

    if ( isMouseCaptured && areButtonsReleased ) {
    // user has released all buttons, so release the capture
	::ReleaseCapture();
    }
    return AwtComponent::WmMouseUp(flags, x, y, button);
}

/************************************************************************
 * WCanvasPeer native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_sun_awt_windows_WCanvasPeer_create(JNIEnv *env, jobject self,
					jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtCanvas::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

} /* extern "C" */




