/*
 * @(#)awt_Choice.cpp	1.39 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_Choice.h"

#include "awt_Canvas.h"

#include "awt_Dimension.h"

#include <java_awt_Toolkit.h>
#include <java_awt_FontMetrics.h>

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/*************************************************************************
 * AwtChoice class methods
 */

AwtChoice::AwtChoice() {
}

const char* AwtChoice::GetClassName() {
    return "COMBOBOX";  /* System provided combobox class */
}

AwtChoice* AwtChoice::Create(jobject peer, jobject parent) {


    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }
    AwtCanvas* awtParent;
    jobject target = 0;
    /* the result */
    AwtChoice* c = NULL;

    JNI_CHECK_NULL_GOTO(parent, "null parent", done);

    awtParent = (AwtCanvas*)JNI_GET_PDATA(parent);
    JNI_CHECK_NULL_GOTO(awtParent, "null awtParent", done);

    target = env->GetObjectField(peer, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    c = new AwtChoice();
    JNI_CHECK_NULL_GOTO(c, "AwtChoice() failed", done);

    {
	DWORD style = WS_CHILD | WS_CLIPSIBLINGS | WS_VSCROLL |
	    CBS_DROPDOWNLIST | CBS_OWNERDRAWFIXED;
	DWORD exStyle = 0;
	if (GetRTL()) {
	    exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
	    if (GetRTLReadingOrder())
	        exStyle |= WS_EX_RTLREADING;
	}

	/*
	 * In OWNER_DRAW, the size of the edit control part of the
	 * choice must be determinded in its creation, when the parent
	 * cannot get the choice's instance from its handle.  So
	 * record the pair of the ID and the instance of the choice.
	 */
	UINT myId = awtParent->CreateControlID();
	ASSERT(myId > 0);
	c->m_myControlID = myId;
	awtParent->PushChild(myId, c);

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	c->CreateHWnd(env, "", style, exStyle,
		      x, y, width, height,
		      awtParent->GetHWnd(),
		      (HMENU)myId,
		      ::GetSysColor(COLOR_WINDOWTEXT),
		      ::GetSysColor(COLOR_WINDOW),
		      peer);

	/* suppress inheriting parent's color. */
	c->m_backgroundColorSet = TRUE;
    }

done:
    env->DeleteLocalRef(target);

    return c;
}

// calculate height of drop-down list part of the combobox
// to show all the items up to a maximum of eight
int AwtChoice::GetDropDownHeight()
{
    int itemHeight =(int)::SendMessage(GetHWnd(), CB_GETITEMHEIGHT, (UINT)0,0);
    int numItemsToShow = (int)::SendMessage(GetHWnd(), CB_GETCOUNT, 0,0);
    numItemsToShow = numItemsToShow > 8 ? 8 : numItemsToShow;
    // drop-down height snaps to nearest line, so add a
    // fudge factor of 1/2 line to ensure last line shows
    return itemHeight*numItemsToShow + itemHeight/2;
}

// get the height of the field portion of the combobox
int AwtChoice::GetFieldHeight()
{
    int fieldHeight;
    int borderHeight;
    fieldHeight =(int)::SendMessage(GetHWnd(), CB_GETITEMHEIGHT, (UINT)-1, 0);
    // add top and bottom border lines; border size is different for
    // Win 4.x (3d edge) vs 3.x (1 pixel line)
    borderHeight = ::GetSystemMetrics(IS_WIN4X ? SM_CYEDGE : SM_CYBORDER);
    fieldHeight += borderHeight*2;
    return fieldHeight;
}

// gets the total height of the combobox, including drop down
int AwtChoice::GetTotalHeight()
{
    int dropHeight = GetDropDownHeight();
    int fieldHeight = GetFieldHeight();
    int totalHeight;

    // border on drop-down portion is always non-3d (so don't use SM_CYEDGE)
    int borderHeight = ::GetSystemMetrics(SM_CYBORDER);
    // total height = drop down height + field height + top+bottom drop down border lines
    totalHeight = dropHeight + fieldHeight +borderHeight*2;
    return totalHeight;
}

// Recalculate and set the drop-down height for the Choice.
void AwtChoice::ResetDropDownHeight()
{
    RECT    rcWindow;

    ::GetWindowRect(GetHWnd(), &rcWindow);
    // resize the drop down to accomodate added/removed items
    int	    totalHeight = GetTotalHeight();
    ::SetWindowPos(GetHWnd(), NULL,
		    0, 0, rcWindow.right - rcWindow.left, totalHeight,
		    SWP_NOACTIVATE|SWP_NOMOVE|SWP_NOZORDER);
}

void AwtChoice::Reshape(int x, int y, int w, int h)
{
    // Choice component height is fixed (when rolled up)
    // so vertically center the choice in it's bounding box
    int fieldHeight = GetFieldHeight();
    if (fieldHeight > 0 && fieldHeight < h) {
        y += (h - fieldHeight) / 2;
    }

    int totalHeight = GetTotalHeight();
    AwtComponent::Reshape(x, y, w, totalHeight);
}

jobject AwtChoice::PreferredItemSize(JNIEnv *env)
{
    jobject dimension = JNU_CallMethodByName(env, NULL, GetPeer(env),
					     "preferredSize",
					     "()Ljava/awt/Dimension;").l;
    ASSERT(!env->ExceptionOccurred());
    if (dimension == NULL) {
	return NULL;
    }
    /* This size is window size of choice and it's too big for each
     * drop down item height.
     */
    env->SetIntField(dimension, AwtDimension::heightID,
		       GetFontHeight(env));
    return dimension;
}


MsgRouting AwtChoice::WmNotify(UINT notifyCode)
{
    if (notifyCode == CBN_SELCHANGE) {
        DoCallback("handleAction", "(I)V", SendMessage(CB_GETCURSEL));
    }
    return mrDoDefault;
}

MsgRouting
AwtChoice::WmDrawItem(UINT /*ctrlId*/, DRAWITEMSTRUCT far& drawInfo)
{
    DrawListItem((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2), drawInfo);
    return mrConsume;
}

MsgRouting AwtChoice::WmMeasureItem(UINT /*ctrlId*/, MEASUREITEMSTRUCT far& measureInfo)
{
    MeasureListItem((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2), measureInfo);
    return mrConsume;
}


/************************************************************************
 * WChoicePeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WChoicePeer
 * Method:    select
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WChoicePeer_select(JNIEnv *env, jobject self,
					jint index)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtChoice* c = (AwtChoice*)pData;
    c->SendMessage(CB_SETCURSEL, index);
}

/*
 * Class:     sun_awt_windows_WChoicePeer
 * Method:    remove
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WChoicePeer_remove(JNIEnv *env, jobject self,
					jint index)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtChoice* c = (AwtChoice *)pData;
    c->SendMessage(CB_DELETESTRING, index, 0);
    c->ResetDropDownHeight();
    c->VerifyState();
}

/*
 * Class:     sun_awt_windows_WChoicePeer
 * Method:    addItem
 * Signature: (Ljava/lang/String;I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WChoicePeer_addItem(JNIEnv *env, jobject self,
					 jstring item, jint index)
{
    JNI_CHECK_NULL_RETURN(item, "null item");

    int pData;
    JNI_CHECK_PEER_RETURN(self);

    AwtChoice* c = (AwtChoice*)pData;
    c->SendMessage(CB_INSERTSTRING, index, JavaStringBuffer(env, item));
    c->ResetDropDownHeight();
    c->VerifyState();
}

/*
 * Class:     sun_awt_windows_WChoicePeer
 * Method:    reshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WChoicePeer_reshape(JNIEnv *env, jobject self,
					 jint x, jint y,
					 jint width, jint height)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtChoice* c = (AwtChoice *)pData;
    c->Reshape(x, y, width, height);
    c->VerifyState();
}

/*
 * Class:     sun_awt_windows_WChoicePeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WChoicePeer_create(JNIEnv *env, jobject self,
					jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtChoice::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

} /* extern "C" */


/************************************************************************
 * Diagnostic routines
 */

#ifdef DEBUG

void AwtChoice::VerifyState()
{
    if (AwtToolkit::GetInstance().VerifyComponents() == FALSE) {
        return;
    }

    if (m_callbacksEnabled == FALSE) {
        /* Component is not fully setup yet. */
        return;
    }

    AwtComponent::VerifyState();
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->PushLocalFrame(1) < 0)
        return;

    jobject target = GetTarget(env);

    // To avoid possibly running client code on the toolkit thread, don't
    // do the following checks if we're running on the toolkit thread.
    if (AwtToolkit::MainThread() != ::GetCurrentThreadId()) {
        // Compare number of items.
        int nTargetItems = JNU_CallMethodByName(env, NULL, target,
					        "countItems", "()I").i;
        ASSERT(!env->ExceptionOccurred());
        int nPeerItems = (int)::SendMessage(GetHWnd(), CB_GETCOUNT, 0, 0);
        ASSERT(nTargetItems == nPeerItems);

        // Compare selection
        int targetIndex = JNU_CallMethodByName(env, NULL, target,
					       "getSelectedIndex", "()I").i;
        ASSERT(!env->ExceptionOccurred());
        int peerCurSel = (int)::SendMessage(GetHWnd(), CB_GETCURSEL, 0, 0);
        ASSERT(targetIndex == peerCurSel);
    }
    env->PopLocalFrame(0);
}
#endif //DEBUG
