/*
 * @(#)awt_Clipboard.cpp	1.13 98/10/01
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Clipboard.h"
#include "awt_Font.h"
#include "awt_TextComponent.h"
#include <java_awt_datatransfer_StringSelection.h>


/************************************************************************
 * AwtClipboard utilities
 */

const char* ObjectFormatName = "JavaObject";
static int objectFormat = 0;


/************************************************************************
 * AwtClipboard fields
 */

jobject AwtClipboard::theCurrentClipboard;

/* sun.awt.windows.WClipboard lostSelectionOwnership method ID. */
jmethodID AwtClipboard::lostSelectionOwnershipMID;

/* This flag is set while we call EmptyClipboard to indicate to
   WM_DESTROYCLIPBOARD handler that we are not loosing ownership */
BOOL AwtClipboard::isGettingOwnership = FALSE;

/************************************************************************
 * AwtClipboard methods
 */

#define GALLOCFLG (GMEM_DDESHARE | GMEM_MOVEABLE | GMEM_ZEROINIT)

void AwtClipboard::LostOwnership(JNIEnv *env) {
    if (theCurrentClipboard != NULL) {
        env->CallVoidMethod(theCurrentClipboard,
                            AwtClipboard::lostSelectionOwnershipMID);
        ASSERT(!env->ExceptionOccurred());
        theCurrentClipboard = NULL;
    }
}


/************************************************************************
 * WClipboard native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WClipboard
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WClipboard_init(JNIEnv *env, jclass cls) 
{
    AwtClipboard::lostSelectionOwnershipMID = 
	env->GetMethodID(cls, "lostSelectionOwnership", "()V");
    
    objectFormat = RegisterClipboardFormat(ObjectFormatName);
    if (objectFormat == 0) {
        char buf[80];
        sprintf(buf, "couldn't register clipboard format: %d", 
                GetLastError());
	JNU_ThrowByName(env, "java/awt/AWTException", buf);
    }
}

/*
 * Class:     sun_awt_windows_WClipboard
 * Method:    setClipboardText
 * Signature: (Ljava/awt/datatransfer/StringSelection;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WClipboard_setClipboardText(JNIEnv *env, jobject self,
						 jobject stringSelection) 
{
    JNI_CHECK_NULL_RETURN(stringSelection, "null StringSelection");

    // Convert text into WCHAR array with CRLF line endings.

    /* REMIND: cache and optimize */
    jclass ssCls = env->FindClass("java/awt/datatransfer/StringSelection");
    jfieldID ss_dataID = env->GetFieldID(ssCls, "data", "Ljava/lang/String;");
    jstring text = (jstring)(env->GetObjectField(stringSelection, ss_dataID));

    int length = env->GetStringLength(text)+1;
    WCHAR* ucText = new WCHAR[length];

    jclass stringCls = env->FindClass("java/lang/String");
    jfieldID s_valueID = env->GetFieldID(stringCls, "value", "[C");
    jcharArray textVal = (jcharArray)(env->GetObjectField(text, s_valueID));

    env->GetCharArrayRegion(textVal, 0, length - 1, ucText);
    ucText[length-1] = L'\0';

    ucText = AwtTextComponent::AddCR(ucText, length);
    length = wcslen(ucText) + 1;

    // Open clipboard.
    VERIFY(OpenClipboard(AwtToolkit::GetInstance().GetHWnd()));
    AwtClipboard::GetOwnership();

    if (IS_NT) {
        // Create a global UNICODE copy of the StringSelector.
        HANDLE hGlobalMemory = GlobalAlloc(GALLOCFLG, length * sizeof(WCHAR));
        WCHAR* buffer = (WCHAR*)GlobalLock(hGlobalMemory);
        wcscpy(buffer, ucText);
        GlobalUnlock(hGlobalMemory);
	// fixed 4177171 : pass handle instead of buffer as per API docs
        VERIFY(SetClipboardData(CF_UNICODETEXT, hGlobalMemory));
    } else {
        // Create a multibyte copy of the StringSelector.
        char* multibytes = W2A(ucText);
        DWORD length = strlen(multibytes) + 1;
        HANDLE hGlobalMemory = GlobalAlloc(GALLOCFLG, length);
        char* buffer = (char*)GlobalLock(hGlobalMemory);
        strcpy(buffer, multibytes);
        GlobalUnlock(hGlobalMemory);
	// fixed 4177171 : pass handle instead of buffer as per API docs
        VERIFY(SetClipboardData(CF_TEXT, hGlobalMemory));
    }
    VERIFY(CloseClipboard());
    delete [] ucText;
    AwtClipboard::theCurrentClipboard = env->NewGlobalRef(self);
}

/*
 * Class:     sun_awt_windows_WClipboard
 * Method:    setClipboardObject
 * Signature: (Ljava/lang/Object;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WClipboard_setClipboardObject(JNIEnv *env, jobject self,
                                                   jobject clipboardObject) 
{
    JNI_CHECK_NULL_RETURN(clipboardObject, "null Object");

    // Open clipboard.
    VERIFY(OpenClipboard(AwtToolkit::GetInstance().GetHWnd()));
    AwtClipboard::GetOwnership();
    VERIFY(SetClipboardData(objectFormat, NULL));  // null=delayed rendering
    VERIFY(CloseClipboard());
    AwtClipboard::theCurrentClipboard = env->NewGlobalRef(self);
}

/*
 * Class:     sun_awt_windows_WClipboard
 * Method:    getClipboardText
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL
Java_sun_awt_windows_WClipboard_getClipboardText(JNIEnv *env, jobject self) 
{
    WCHAR* ucText = NULL;
    const jchar *charArr;
    jchar *tmp;

    HANDLE hClipMemory = NULL;
    DWORD length;

    VERIFY(OpenClipboard(AwtToolkit::GetInstance().GetHWnd()));

    if (IS_NT) {
        hClipMemory = GetClipboardData(CF_UNICODETEXT);
        if (hClipMemory == NULL) {
            VERIFY(CloseClipboard());
            return NULL;
        }
        length = (GlobalSize(hClipMemory) / sizeof(unsigned short));
        ucText = new WCHAR[length];
        WCHAR* pClipMemory = (WCHAR*)GlobalLock(hClipMemory);
        wcscpy(ucText, pClipMemory);
    } else {
        HANDLE hClipMemory = GetClipboardData(CF_TEXT);
        if (hClipMemory == NULL) {
            VERIFY(CloseClipboard());
            return NULL;
        }
        WCHAR* pText = A2W((char*)GlobalLock(hClipMemory));
        length = wcslen(pText);
        ucText = new WCHAR[length+1];
        wcscpy(ucText, pText);
    }
    GlobalUnlock(hClipMemory);
    VERIFY(CloseClipboard());

    // Convert WCHAR array to Java String and return it.
    length = AwtTextComponent::RemoveCR(ucText);
    tmp = new jchar[length];
    ASSERT(charArr != NULL);
    wcsncpy(tmp, ucText, length);
    charArr = tmp;
    delete[] ucText;
    jstring jStr = env->NewString(charArr, length);
    delete [] tmp;
    ASSERT(!(env->ExceptionOccurred()));
    return jStr;
}

} /* extern "C" */
