/*
 * @(#)awt_Component.cpp	1.199 98/11/06
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "jlong.h"
#include "awt_AWTEvent.h"
#include "awt_Component.h"
#include "awt_Cursor.h"
#include "awt_Dimension.h"
#include "awt_Frame.h"
#include "awt_Graphics.h"
#include "awt_InputEvent.h"
#include "awt_Insets.h"
#include "awt_KeyEvent.h"
#include "awt_MenuItem.h"
#include "awt_MouseEvent.h"
#include "awt_Toolkit.h"
#include "awt_Window.h"
#ifdef DEBUG
#include "awt_Unicode.h"
#endif

#include <java_awt_Toolkit.h>
#include <java_awt_FontMetrics.h>
#include <java_awt_Color.h>
#include <java_awt_Cursor.h>
#include <java_awt_Event.h>
#include <java_awt_event_KeyEvent.h>
#include <java_awt_Insets.h>
#include <java_awt_Graphics.h>
#include <sun_awt_windows_WPanelPeer.h>
#include <java_awt_event_InputEvent.h>
#include <java_awt_event_InputMethodEvent.h>
#include <sun_awt_windows_WInputMethod.h>

// Begin -- Win32 SDK include files
#include <tchar.h>
#include <imm.h>
#include <ime.h>
// End -- Win32 SDK include files

// Use the one pass input method framework design unless you specify differently
//#define	TWO_PASS_DESIGN

#include <awt_DnDDT.h>

const char* szAwtComponentClassName = "SunAwtComponent";

// Bug #4039858 (Selecting menu item causes bogus mouse click event)
BOOL g_bMenuLoop = FALSE;       
static HWND g_hwndDown = NULL;


//////////////////////////////////////////////////////////////////////////

/*************************************************************************
 * AwtComponent fields
 */

jfieldID AwtComponent::peerID;
jfieldID AwtComponent::xID;
jfieldID AwtComponent::yID;
jfieldID AwtComponent::widthID;
jfieldID AwtComponent::heightID;
jfieldID AwtComponent::visibleID;
jfieldID AwtComponent::backgroundID;
jfieldID AwtComponent::foregroundID;
jfieldID AwtComponent::enabledID;
jfieldID AwtComponent::parentID;

jmethodID AwtComponent::getFontMID;
jmethodID AwtComponent::getToolkitMID;
jmethodID AwtComponent::isEnabledMID;


LANGID AwtComponent::m_idLang = LOWORD(GetKeyboardLayout(0));
UINT AwtComponent::m_CodePage = 
    AwtComponent::LangToCodePage(AwtComponent::m_idLang);
BOOL AwtComponent::m_isWin95 = IS_WIN95;
// Added by waleed to initialize the RTL Flags
BOOL AwtComponent::sm_rtl = PRIMARYLANGID(GetInputLanguage()) == LANG_ARABIC ||
                            PRIMARYLANGID(GetInputLanguage()) == LANG_HEBREW;
BOOL AwtComponent::sm_rtlReadingOrder = 
    PRIMARYLANGID(GetInputLanguage()) == LANG_ARABIC;

static BOOL   bHasDeadKeyDown = false;
static WPARAM uDeadKeyValue   = 0; 

/************************************************************************
 * AwtComponent methods
 */

AwtComponent::AwtComponent()
{
    m_callbacksEnabled = FALSE;
    m_hwnd = NULL;

    m_colorForeground = 0;
    m_colorBackground = 0;
    m_backgroundColorSet = FALSE;
    m_penForeground = NULL;
    m_brushBackground = NULL;
    m_sharedDC = NULL;
    m_DefWindowProc = NULL;
    m_nextControlID = 1;
    m_hCursor = NULL;
    m_childList = NULL;
    m_myControlID = 0;
    m_dragged = FALSE;
    m_hdwp = NULL;
    m_validationNestCount = 0;

    m_dropTarget = NULL;
    m_wprocs     = NULL;

    m_InputMethod = NULL;
    m_useNativeCompWindow = TRUE;
}

AwtComponent::~AwtComponent()
{
    CriticalSection::Lock l(GetLock());

    if (m_hdwp != NULL) {
    // end any deferred window positioning, regardless
    // of m_validationNestCount
	::EndDeferWindowPos(m_hdwp);
    }

    /* Stop message filtering. */
    UnsubclassWindow();

    /* Disconnect all links. */
    UnlinkObjects();

    /* Release global ref to input method */
    SetInputMethod(NULL, TRUE);

    if (m_childList != NULL)
	delete m_childList;

    if (m_myControlID != 0) {
	AwtComponent* parent = GetParent();
	if (parent != NULL)
	    parent->RemoveChild(m_myControlID);
    }

    /* Waiting until this base destructor assumes that we don't care about  
     * getting destroy related messages in derived classes, as this object 
     * is already torn down.
     */
    DestroyHWnd();

    /* Release any allocated resources. */
    if (m_penForeground != NULL) {
        m_penForeground->Release();
        m_penForeground = NULL;
    }
    if (m_brushBackground != NULL) {
        m_brushBackground->Release();
        m_brushBackground = NULL;
    }
#if 0
    /* DestroyCursor is only valid on cursors created with CreateCursor */
    if (m_hCursor != NULL) {
        ::DestroyCursor(m_hCursor);
        m_hCursor = NULL;
    }
#endif

    DiscardMsgWorkProcs();
}

ATOM thisPropertyAtom;

void AwtComponent::Init()
{
    char buf[40];
    VERIFY(sprintf(buf, "%s-0x%x", szAwtComponentClassName, 
                   ::GetCurrentThreadId()) < 40);
    thisPropertyAtom = ::GlobalAddAtom(buf);
    ASSERT(thisPropertyAtom != 0);
}

void AwtComponent::Term()
{
    VERIFY(::GlobalDeleteAtom(thisPropertyAtom) == NULL);
}

/*
 * Single window proc for all the components. Delegates real work to
 * the component's WindowProc() member function.
 */
LRESULT CALLBACK AwtComponent::WndProc(HWND hWnd, UINT message, 
				       WPARAM wParam, LPARAM lParam)
{
    ASSERT(::IsWindow(hWnd));

    AwtComponent* self = GetComponent(hWnd);

    if (!self || self->GetHWnd() != hWnd) {
        return ::DefWindowProc(hWnd, message, wParam, lParam);
    }

    return self->WindowProc(message, wParam, lParam);
}

/************************************************************************
 * AwtComponent dynamic methods
 *
 * Window class registration routines
 */

void AwtComponent::RegisterClass()
{
    const char* className = GetClassName();
    WNDCLASS  wc;

    if (!::GetClassInfo(::GetModuleHandle(NULL), className, &wc)) {
        wc.style         = CS_OWNDC;
        wc.lpfnWndProc   = (WNDPROC)::DefWindowProc;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = 0;
        wc.hInstance     = ::GetModuleHandle(NULL);
        wc.hIcon         = AwtToolkit::GetInstance().GetAwtIcon();
        wc.hCursor       = NULL;
        wc.hbrBackground = NULL;
        wc.lpszMenuName  = NULL;
        wc.lpszClassName = className;

        ATOM ret = ::RegisterClass(&wc);
        ASSERT(ret != 0);
    }
}

void AwtComponent::UnregisterClass()
{
    ::UnregisterClass(GetClassName(), ::GetModuleHandle(NULL));
}

void 
AwtComponent::CreateHWnd(JNIEnv *env, const char* title, long windowStyle, 
			 long windowExStyle,
			 int x, int y, int w, int h,
			 HWND hWndParent, HMENU hMenu, 
			 COLORREF colorForeground, 
			 COLORREF colorBackground,
			 jobject peer) 
{
    if (env->EnsureLocalCapacity(2) < 0) {
	return;
    }

    /*
     * The window class of multifont label must be "BUTTON" because
     * "STATIC" class can't get WM_DRAWITEM message, and m_peerObject
     * member is referred in the GetClassName method of AwtLabel class.
     * So m_peerObject member must be set here.
     */
    m_peerObject = env->NewGlobalRef(peer);
    RegisterClass();

    jobject target = env->GetObjectField(peer, AwtObject::targetID);
    jboolean visible = env->GetBooleanField(target, AwtComponent::visibleID);

    if (visible) {
        windowStyle |= WS_VISIBLE;
    } else {
        windowStyle &= ~WS_VISIBLE;
    }

    HWND hwnd = ::CreateWindowEx(windowExStyle,
				 GetClassName(),
				 title,
				 windowStyle,
				 x, y, w, h,
				 hWndParent,
				 hMenu,
				 ::GetModuleHandle(NULL),
				 NULL);
    ASSERT(hwnd != NULL);
    m_hwnd = hwnd;

    ImmAssociateContext(hwnd,NULL);

    LinkObjects(env, peer);

    /* Subclass the window now so that we can snoop on its messages */
    SubclassWindow();

    /* Set default colors. */
    m_colorForeground = colorForeground;
    m_colorBackground = colorBackground;

    /*
     * Only set background color if the color is actually set on the
     * target -- this avoids inheriting a parent's color unnecessarily, 
     * and has to be done here because there isn't an API to get the
     * real background color from outside the AWT package.
     */
    jobject bkgrd = env->GetObjectField(target, AwtComponent::backgroundID) ;
    if (bkgrd != NULL) {
	JNU_CallMethodByName(env, NULL, peer, "setBackground", 
			     "(Ljava/awt/Color;)V", bkgrd);
	ASSERT(!env->ExceptionOccurred());
    }	
    env->DeleteLocalRef(target);
    env->DeleteLocalRef(bkgrd);
}

/*
 * Destroy this window's HWND
 */
void AwtComponent::DestroyHWnd() {
    if (m_hwnd != NULL) {
        AwtToolkit::DestroyComponent(this);
        m_hwnd = NULL;
    }
}

/*
 * Install our window proc as the proc for our HWND, and save off the 
 * previous proc as the default
 */
void AwtComponent::SubclassWindow()
{
    m_DefWindowProc = (WNDPROC)::SetWindowLong(GetHWnd(), GWL_WNDPROC, 
					       (long)WndProc);
}

/*
 * Reinstall the original window proc as the proc for our HWND
 */
void AwtComponent::UnsubclassWindow()
{
    if (m_DefWindowProc) {
        ::SetWindowLong(GetHWnd(), GWL_WNDPROC, (long)m_DefWindowProc);
        m_DefWindowProc = NULL;
    }
}

/////////////////////////////////////
// (static method)
// Determines the top-level ancestor for a given window. If the given
// window is a top-level window, return itself.
//
// 'Top-level' includes dialogs as well.
//
HWND AwtComponent::GetTopLevelParentForWindow(HWND hwndDescendant) {
    if (hwndDescendant == NULL) {
	return NULL;
    }

    ASSERT(IsWindow(hwndDescendant));
    HWND hwnd = hwndDescendant;
    for(;;) {
	LONG style = ::GetWindowLong(hwnd, GWL_STYLE);
	if ( (style & WS_CHILD) == 0 ) {
	    // found a non-child window so terminate
	    break;
	}
	hwnd = ::GetParent(hwnd);
    }

    ASSERT( !(::GetWindowLong(hwnd, GWL_STYLE) & WS_CHILD) );
    return hwnd;
}
////////////////////

void AwtComponent::SetColor(COLORREF c)
{
    if (AwtImage::m_clrdata.grayscale) {
        int g;

        g = (int) (.299 * (c & 0xFF) + .587 * ((c >> 8) & 0xFF) +
            .114 * ((c >> 16) & 0xFF) + 0.5);
        // c = g | (g << 8) | (g << 16);
        c = PALETTERGB(g, g, g);
    }

    if (m_colorForeground == c) {
        return;
    }
    
    m_colorForeground = c;
    if (m_penForeground != NULL) {
        m_penForeground->Release();
        m_penForeground = NULL;
    }
    VERIFY(::InvalidateRect(GetHWnd(), NULL, FALSE));
}

void AwtComponent::SetBackgroundColor(COLORREF c)
{
    if (AwtImage::m_clrdata.grayscale) {
        int g;

        g = (int) (.299 * (c & 0xFF) + .587 * ((c >> 8) & 0xFF) +
            .114 * ((c >> 16) & 0xFF) + 0.5);
        // c = g | (g << 8) | (g << 16);
        c = PALETTERGB(g, g, g);
    }

    if (m_colorBackground == c) {
        return;
    }
    m_colorBackground = c;
    m_backgroundColorSet = TRUE;
    if (m_brushBackground != NULL) {
        m_brushBackground->Release();
        m_brushBackground = NULL;
    }
    VERIFY(::InvalidateRect(GetHWnd(), NULL, TRUE));
}

HPEN AwtComponent::GetForegroundPen()
{
    if (m_penForeground == NULL) {
        m_penForeground = AwtPen::Get(m_colorForeground);
    }
    return m_penForeground->GetHandle();
}

COLORREF AwtComponent::GetBackgroundColor()
{
    if (m_backgroundColorSet == FALSE) {
        AwtComponent* c = this;
        while ((c = c->GetParent()) != NULL) {
            if (c->IsBackgroundColorSet()) {
                return c->GetBackgroundColor();
            }
        }
    }
    return m_colorBackground;
}

HBRUSH AwtComponent::GetBackgroundBrush()
{
    if (m_backgroundColorSet == FALSE) {
        if (m_brushBackground != NULL) {
            m_brushBackground->Release();
            m_brushBackground = NULL;
        }
          AwtComponent* c = this;
          while ((c = c->GetParent()) != NULL) {
              if (c->IsBackgroundColorSet()) {
                  m_brushBackground = 
                      AwtBrush::Get(c->GetBackgroundColor());
                  break;
              }
          }
    }
    if (m_brushBackground == NULL) {
        m_brushBackground = AwtBrush::Get(m_colorBackground);
    }
    return m_brushBackground->GetHandle();
}

void AwtComponent::SetFont(AwtFont* font)
{
    ASSERT(font != NULL);
    if (font->GetAscent() < 0) {
	AwtFont::SetupAscent(font);
    }
    SendMessage(WM_SETFONT, (WPARAM)font->GetHFont(), MAKELPARAM(FALSE, 0));
    VERIFY(::InvalidateRect(GetHWnd(), NULL, TRUE));
}

AwtSharedDC* AwtComponent::GetSharedDC()
{
    if (m_sharedDC == NULL) {
        HDC hWndDC = ::GetDC(GetHWnd());
        ::SetBkMode(hWndDC, TRANSPARENT);
        ::SelectObject(hWndDC, ::GetStockObject(NULL_BRUSH));
        ::SelectObject(hWndDC, ::GetStockObject(NULL_PEN));
        ::GdiFlush();
        m_sharedDC = new AwtSharedDC(this, hWndDC);
    }
    return m_sharedDC;
}

AwtComponent* AwtComponent::GetParent() 
{
    HWND hwnd = ::GetParent(GetHWnd());
    if (hwnd == NULL) {
        return NULL;
    }
    return GetComponent(hwnd);
}

AwtWindow* AwtComponent::GetContainer() 
{
    AwtComponent* comp = this;
    while (comp != NULL) {
        if (comp->IsContainer()) {
            return (AwtWindow*)comp;
        }
        comp = comp->GetParent();
    }
    return NULL;
}

void AwtComponent::Show()
{
    SendMessage(WM_AWT_COMPONENT_SHOW, SW_SHOWNA);
}

void AwtComponent::Hide()
{
    SendMessage(WM_AWT_COMPONENT_SHOW, SW_HIDE);
}

void AwtComponent::Reshape(int x, int y, int w, int h)
{
    AwtWindow* container = GetContainer();
    AwtComponent* parent = GetParent();
    if (container != NULL && container == parent) {
        container->SubtractInsetPoint(x, y);
    }
    long flags = SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOCOPYBITS;

    if (parent && strcmp(parent->GetClassName(), "SunAwtScrollPane") == 0) {
        if (x > 0) {
            x = 0;
        }
        if (y > 0) {
            y = 0;
        }

	RECT childRect;
	::GetClientRect(GetHWnd(), &childRect);
 
	/* if Reshape() is called and it is not as a result of a move, 
	   suppress repaints.  Given that scrollpane always adjusts
	   the dimensions of its child to be atleast as large as the
	   veiwport,  we can safely assume that these resizes
	   will be outside the viewrect. */

	if ((childRect.right != w) || (childRect.bottom != h)) {
	    flags |= SWP_NOREDRAW;
	}
    }
    if (m_hdwp != NULL) {
	m_hdwp = ::DeferWindowPos(m_hdwp, GetHWnd(), 0, x, y, w, h, flags);
	ASSERT(m_hdwp != NULL);
    } else {
	VERIFY(::SetWindowPos(GetHWnd(), 0, x, y, w, h, flags));
    }
}

void AwtComponent::SetScrollValues(UINT bar, int min, int value, int max)
{
    int minTmp, maxTmp;

    ::GetScrollRange(GetHWnd(), bar, &minTmp, &maxTmp);
    if (min == INT_MAX) {
        min = minTmp;
    }
    if (value == INT_MAX) {
        value = ::GetScrollPos(GetHWnd(), bar);
    }
    if (max == INT_MAX) {
        max = maxTmp;
    }
    if (min == max) {
        max++;
    }
    ::SetScrollRange(GetHWnd(), bar, min, max, FALSE);
    ::SetScrollPos(GetHWnd(), bar, value, TRUE);
}

/*
 * Save Global Reference of sun.awt.windows.WInputMethod object
 */
void AwtComponent::SetInputMethod(jobject im, BOOL useNativeCompWindow) 
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (m_InputMethod!=NULL)    
        env->DeleteGlobalRef(m_InputMethod);

    if (im!=NULL){
        m_InputMethod = env->NewGlobalRef(im);
        m_useNativeCompWindow = useNativeCompWindow;
    } else {
        m_InputMethod = NULL;
        m_useNativeCompWindow = TRUE;
    }

}

/*
 * Opportunity to process and/or eat a message before it is dispatched
 */
MsgRouting AwtComponent::PreProcessMsg(MSG& msg)
{
#ifdef TWO_PASS_DESIGN
	// In Java awt implementation, 
	//  1) KeyTyped Event should generated only when KeyPressed Event
	//      is not consumed.
	//  2) Key Events should be passed to Component's dispatchEvent() 
	//      first, and then passed to InputMethod.
	// For Key Event Message, Win32 TranslateMessage() will do
	//  1) generate and post WM_CHAR message unconditionally 
	//      by WM_KEYDOWN message
	//  2) call IMM32.DLL and make current IME to process key events.
	// These TranslateMessage() actions are only disturbance.
	// So, we cheat TranslateMessage() by changeing message IDs.
	// These special message is re-post afterwords, and then we will
	// change message IDs to their original value and give it to
	// TranslateMessage().

	switch(msg.message){
	//change message IDs to cheat TranslateMessage()
	case WM_KEYDOWN:	msg.message = WM_AWT_PRE_KEYDOWN;	break;
	case WM_KEYUP:		msg.message = WM_AWT_PRE_KEYUP;		break;
	case WM_SYSKEYDOWN:	msg.message = WM_AWT_PRE_SYSKEYDOWN;	break;
	case WM_SYSKEYUP:	msg.message = WM_AWT_PRE_SYSKEYUP;	break;

	//If cheated message re-posted, recover original message IDs 
	// and pass it to TranslateMessage().
	case WM_AWT_PRE_KEYDOWN:	msg.message = WM_KEYDOWN;	break;
	case WM_AWT_PRE_KEYUP:		msg.message = WM_KEYUP;		break;
	case WM_AWT_PRE_SYSKEYDOWN:	msg.message = WM_SYSKEYDOWN;	break;
	case WM_AWT_PRE_SYSKEYUP:	msg.message = WM_SYSKEYUP;	break;
	}
#endif

    return mrPassAlong;
}

static UINT lastMessage = WM_NULL;

/*
 * Dispatch messages for this window class--general component
 */
LRESULT AwtComponent::WindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
    LRESULT retValue = 0;
    MsgRouting mr = mrDoDefault;

    lastMessage = message;

    switch (message) {
      case WM_CREATE:     mr = WmCreate(); break;
      case WM_CLOSE:      mr = WmClose(); break;
      case WM_DESTROY:    mr = WmDestroy(); break;

      case WM_ERASEBKGND: 
          mr = WmEraseBkgnd((HDC)wParam, *(BOOL*)&retValue); break;
      case WM_PAINT:      mr = WmPaint((HDC)wParam); break;
      case WM_MOVE:       mr = WmMove(LOWORD(lParam), HIWORD(lParam)); break;
      case WM_SIZE:       
      {
	  RECT r;
	  // fix 4128317 : use GetClientRect for full 32-bit int precision and
	  // to avoid negative client area dimensions overflowing 16-bit params - robi
	  ::GetClientRect( GetHWnd(), &r );
          mr = WmSize(wParam, r.right - r.left, r.bottom - r.top);
	  //mr = WmSize(wParam, LOWORD(lParam), HIWORD(lParam));
	  break;
      }
      case WM_SHOWWINDOW:
          mr = WmShowWindow(wParam, lParam); break;
      case WM_SYSCOMMAND: 
          mr = WmSysCommand(wParam & 0xFFF0, LOWORD(lParam), HIWORD(lParam));
	  break;
      case WM_EXITSIZEMOVE:
          mr = WmExitSizeMove(); break;

      // Bug #4039858 (Selecting menu item causes bogus mouse click event)
      case WM_ENTERMENULOOP:              
	  mr = WmEnterMenuLoop((BOOL)wParam);
          g_bMenuLoop = TRUE;
          break;     
      case WM_EXITMENULOOP:
	  mr = WmExitMenuLoop((BOOL)wParam);
          g_bMenuLoop = FALSE;
          break;     

      case WM_SETFOCUS:   mr = WmSetFocus((HWND)wParam); break;
      case WM_KILLFOCUS:  mr = WmKillFocus((HWND)wParam); break;
      case WM_ACTIVATE:   
          mr = WmActivate(LOWORD(wParam), (BOOL)HIWORD(wParam)); break;

#if defined(WIN32)
      case WM_CTLCOLORMSGBOX:
      case WM_CTLCOLOREDIT:
      case WM_CTLCOLORLISTBOX:
      case WM_CTLCOLORBTN:
      case WM_CTLCOLORDLG:
      case WM_CTLCOLORSCROLLBAR:
      case WM_CTLCOLORSTATIC:
          mr = WmCtlColor((HDC)wParam, (HWND)lParam, 
                          message-WM_CTLCOLORMSGBOX+CTLCOLOR_MSGBOX, 
                          *(HBRUSH*)&retValue); 
          break;
      case WM_HSCROLL:    
          mr = WmHScroll(LOWORD(wParam), HIWORD(wParam), (HWND)lParam); 
          break;
      case WM_VSCROLL:    
          mr = WmVScroll(LOWORD(wParam), HIWORD(wParam), (HWND)lParam); 
          break;
#else
      case WM_CTLCOLOR:   
          mr = WmCtlColor((HDC)wParam, (HWND)LOWORD(lParam), HIWORD(lParam),
                          *(HBRUSH*)&retValue); 
          break;
      case WM_HSCROLL:    
          mr = WmHScroll(wParam, LOWORD(lParam), (HWND)HIWORD(lParam)); 
          break;
      case WM_VSCROLL:
          mr = WmVScroll(wParam, LOWORD(lParam), (HWND)HIWORD(lParam)); 
          break;
#endif
      case WM_SETCURSOR:  
          mr = WmSetCursor((HWND)wParam, LOWORD(lParam), HIWORD(lParam), 
                           *(BOOL*)&retValue); 
          break;
      case WM_AWT_SETCURSOR:
	  mr = WmAwtSetCursor( (jobject) wParam);
	  break;
      case WM_AWT_MOUSEENTER: 
          mr = WmMouseEnter(wParam, LO_INT(lParam), HI_INT(lParam)); break;
      case WM_LBUTTONDOWN:
      case WM_LBUTTONDBLCLK:
          mr = WmMouseDown(wParam, LO_INT(lParam), HI_INT(lParam), 
			   LEFT_BUTTON); break;
      case WM_MBUTTONDOWN:
      case WM_MBUTTONDBLCLK:
          mr = WmMouseDown(wParam, LO_INT(lParam), HI_INT(lParam), 
			   MIDDLE_BUTTON); break;
      case WM_RBUTTONDOWN:
      case WM_RBUTTONDBLCLK:
          mr = WmMouseDown(wParam, LO_INT(lParam), HI_INT(lParam), 
			   RIGHT_BUTTON); break;
      case WM_LBUTTONUP:
          mr = WmMouseUp(wParam, LO_INT(lParam), HI_INT(lParam), 
			 LEFT_BUTTON); break;
      case WM_RBUTTONUP:
          mr = WmMouseUp(wParam, LO_INT(lParam), HI_INT(lParam), 
			 RIGHT_BUTTON);
          break;
      case WM_MBUTTONUP:      
          mr = WmMouseUp(wParam, LO_INT(lParam), HI_INT(lParam), 
			 MIDDLE_BUTTON);
          break;
      case WM_MOUSEMOVE:
          mr = WmMouseMove(wParam, LO_INT(lParam), HI_INT(lParam)); 
          break;
      case WM_AWT_MOUSEEXIT:
          mr = WmMouseExit(wParam, LO_INT(lParam), HI_INT(lParam)); 
          break;

#ifdef TWO_PASS_DESIGN
      //In 2-pass design, WM_AWT_PRE_XXXX means 1st-pass.
      //WmKeyXXX() is called and Java KeyEvent is generated in 1st-pass.
      //WM_XXXX means 2nd-pass and ignored to passed to DefWindowProc()
      case WM_AWT_PRE_KEYDOWN:
          mr = WmKeyDown(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;
      case WM_AWT_PRE_KEYUP:
          mr = WmKeyUp(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;
      case WM_AWT_PRE_SYSKEYDOWN:
          mr = WmKeyDown(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;
      case WM_AWT_PRE_SYSKEYUP:
          mr = WmKeyUp(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;
#else 
      case WM_KEYDOWN:
          MSG msg;
          if (m_isWin95 && PeekMessage(&msg, GetHWnd(), WM_DEADCHAR, WM_DEADCHAR, PM_NOREMOVE)){
              bHasDeadKeyDown = true;
              uDeadKeyValue = wParam;
              break;
          }
          mr = WmKeyDown(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;
      case WM_KEYUP:
          if (m_isWin95 && bHasDeadKeyDown){
              bHasDeadKeyDown = false;
              if (uDeadKeyValue == wParam)
                  break;
          }
          mr = WmKeyUp(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;
      case WM_SYSKEYDOWN:
          mr = WmKeyDown(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;
      case WM_SYSKEYUP:
          mr = WmKeyUp(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;
#endif
      case WM_IME_SETCONTEXT:
          // lParam is passed as pointer and it can be modified.
          mr = WmImeSetContext(wParam, &lParam); 
          break;
      case WM_IME_NOTIFY:
          mr = WmImeNotify(wParam, lParam); 
          break;
      case WM_IME_STARTCOMPOSITION:
          mr = WmImeStartComposition(); 
          break;
      case WM_IME_ENDCOMPOSITION:
          mr = WmImeEndComposition(); 
          break;
      case WM_IME_COMPOSITION:
          mr = WmImeComposition(wParam, lParam); 
          break;

      case WM_CHAR:
          mr = WmChar(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;
      case WM_SYSCHAR:
          mr = WmChar(wParam, LOWORD(lParam), HIWORD(lParam), TRUE); 
          break;
      case WM_IME_CHAR:
          mr = WmIMEChar(wParam, LOWORD(lParam), HIWORD(lParam), FALSE); 
          break;

      case WM_INPUTLANGCHANGE:
      	  mr = WmInputLangChange(wParam, lParam);
	  // should return non-zero if we process this message
	  retValue = 1;
	  break;

      case WM_AWT_FORWARD_CHAR:
      	  mr = WmForwardChar(LOWORD(wParam), lParam, HIWORD(wParam));
	  break;

      case WM_AWT_FORWARD_BYTE:
          mr = HandleEvent( (MSG *) lParam, (BOOL) wParam);
          break;

      case WM_PASTE:
          mr = WmPaste();
	  break;

      case WM_TIMER:
          mr = WmTimer(wParam);
          break;

      case WM_COMMAND:    
#if defined(WIN32)
          mr = WmCommand(LOWORD(wParam), (HWND)lParam, HIWORD(wParam));
#else
          mr = WmCommand(LOWORD(wParam), (HWND)LOWORD(lParam), HIWORD(lParam));
#endif
          break;

      case WM_COMPAREITEM:
          mr = WmCompareItem(wParam, *(COMPAREITEMSTRUCT*)lParam, retValue);
          break;
      case WM_DELETEITEM: 
          mr = WmDeleteItem(wParam, *(DELETEITEMSTRUCT*)lParam);
          break;
      case WM_DRAWITEM:
          mr = WmDrawItem(wParam, *(DRAWITEMSTRUCT*)lParam);
          break;
      case WM_MEASUREITEM:
          mr = WmMeasureItem(wParam, *(MEASUREITEMSTRUCT*)lParam);
          break;

      case WM_AWT_HANDLE_EVENT:
          mr = HandleEvent( (MSG *) lParam, (BOOL) wParam);
          break;

      case WM_AWT_PRINT_COMPONENT:
	  mr = PrintComponent((AwtGraphics*)wParam);
	  break;
      case WM_PRINT:
	  mr = WmPrint((HDC)wParam, lParam);
	  break;
      case WM_PRINTCLIENT:
	  mr = WmPrintClient((HDC)wParam, lParam);
	  break;

      case WM_NCCALCSIZE:
	  mr = WmNcCalcSize((BOOL)wParam, (LPNCCALCSIZE_PARAMS)lParam, 
			    *(UINT*)&retValue);
	  break;
      case WM_NCPAINT:
	  mr = WmNcPaint((HRGN)wParam);
	  break;
      case WM_NCHITTEST:
	  mr = WmNcHitTest(LOWORD(lParam), HIWORD(lParam), *(UINT*)&retValue);
	  break;

      case WM_AWT_RESHAPE_COMPONENT: {
          RECT* r = (RECT*)lParam;
          Reshape(r->left, r->top, r->right - r->left, r->bottom - r->top);
          delete r;
          mr = mrConsume;
          break;
      }

      case WM_AWT_BEGIN_VALIDATE:
          BeginValidate();
          mr = mrConsume;
          break;
      case WM_AWT_END_VALIDATE:
          EndValidate();
          mr = mrConsume;
          break;
			  
      case WM_QUERYNEWPALETTE:
	  mr = WmQueryNewPalette(*(UINT*)&retValue);
          break;
      case WM_PALETTECHANGED:
          mr = WmPaletteChanged((HWND)wParam);
          break;
      case WM_STYLECHANGED:
	  mr = WmStyleChanged(wParam, (LPSTYLESTRUCT)lParam);
	  break;
      case WM_SETTINGCHANGE:
	  mr = WmSettingChange(wParam, (LPCTSTR)lParam);
	  break;

	  /*
	   * These messages are used to route Win32 calls to the
	   * creating thread, since these calls fail unless executed
	   * there.  
	   */
      case WM_AWT_COMPONENT_SHOW:
          ::ShowWindow(GetHWnd(), wParam);

	  CallMsgWorkProcs(message); // register drop targets

          mr = mrConsume;
          break;

          /*
           * The WM_AWT_COMPONENT_SETFOCUS handler will call ::SetFocus iff
           * the HWND stored in wParam is the currently active window or
           * wParam is NULL. If wParam is set to non-NULL, specify the window
           * that previously had the focus in lParam.
           */
      case WM_AWT_COMPONENT_SETFOCUS:
          if (wParam != NULL) {
              if ( ((HWND) wParam) == ::GetActiveWindow()) {
                  ::SetFocus(GetHWnd());
              }
              else {
                  // Don't call SetFocus if our top-level window is not
                  // active since SetFocus would activate it.
                  ::SendMessage((HWND) lParam, WM_KILLFOCUS, (WPARAM)GetHWnd(),
                                0);
                  ::SendMessage(GetHWnd(), WM_SETFOCUS, lParam, 0);
              }
          }
          else {
              ::SetFocus(GetHWnd());
          }
          mr = mrConsume;
          break;

      case WM_AWT_SET_SCROLL_INFO:
	  SCROLLINFO *si = (SCROLLINFO *) lParam;
	  ::SetScrollInfo(GetHWnd(), (int) wParam, si, TRUE);
	  delete si;
	  mr = mrConsume;
	  break;
    }

    /*
     * If not a specific Consume, it was a specific DoDefault, or a 
     * PassAlong (since the default is the next in chain), then call the 
     * default proc
     */
    if (mr != mrConsume) {
        retValue = DefWindowProc(message, wParam, lParam);
    }
    return retValue;
}

/*
 * Call this instance's default window proc, or if none set, call the stock 
 * Window's one.
 */
LRESULT AwtComponent::DefWindowProc(UINT msg, WPARAM wParam, LPARAM lParam)
{
    return m_DefWindowProc
        ? ::CallWindowProc((FARPROC)m_DefWindowProc, GetHWnd(), msg, 
			   wParam, lParam)
        : ::DefWindowProc(GetHWnd(), msg, wParam, lParam);
}

/*
 * This message should only be received when a window is destroy by
 * Windows, and not Java.  Window termination has been reworked so
 * this method should never be called during termination.
 */
MsgRouting AwtComponent::WmDestroy()
{
    delete this;
    return mrConsume;
}

MsgRouting AwtComponent::WmMove(int x, int y)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSize(int type, int w, int h)
{
    return mrDoDefault;
}
  
MsgRouting AwtComponent::WmSysCommand(UINT uCmdType, UINT xPos, UINT yPos)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmExitSizeMove()
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmEnterMenuLoop(BOOL isTrackPopupMenu)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmExitMenuLoop(BOOL isTrackPopupMenu)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmShowWindow(int show, long status)
{
    // NULL-InputContext is associated to all window just after they created.
    // ( see CreateHWnd() )
    // But to TextField and TextArea on Win95, valid InputContext is associated 
    // by system after that. This is not happen on NT4.0
    // For workaround, force context to NULL here.
    if (ImmGetContext(m_hwnd))
        ImmAssociateContext(m_hwnd, NULL);
 
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSetFocus(HWND hWndLost)
{
    SendFocusEvent(java_awt_event_FocusEvent_FOCUS_GAINED);
    return mrDoDefault;
}

MsgRouting AwtComponent::WmKillFocus(HWND hWndGotFocus)
{
    HWND newTop = NULL;
    HWND oldTop = NULL;

    // Determine new top-level window with component gaining the focus
    newTop = AwtComponent::GetTopLevelParentForWindow(hWndGotFocus);

    // Determine old top-level window with component losing the focus
    oldTop = AwtComponent::GetTopLevelParentForWindow(GetHWnd());

    // FOCUS_LOST is temporary if we are switching to another top-level window
    BOOL isTemporary = (newTop != oldTop);

    // Post the FOCUS_LOST event. It will be redispatched to lightweight
    // components at the java.awt level
    SendFocusEvent(java_awt_event_FocusEvent_FOCUS_LOST, isTemporary);

    return mrDoDefault;
}

MsgRouting AwtComponent::WmCtlColor(HDC hDC, HWND hCtrl,
				    UINT ctlColor, HBRUSH& retBrush)
{
    AwtComponent* child = AwtComponent::GetComponent(hCtrl);
    if (child) {
        ::SetBkColor(hDC, child->GetBackgroundColor());
        ::SetTextColor(hDC, child->GetColor());
        retBrush = child->GetBackgroundBrush();
        return mrConsume;
    }
    return mrDoDefault;
/*
    switch (ctlColor) {
        case CTLCOLOR_MSGBOX:
        case CTLCOLOR_EDIT:
        case CTLCOLOR_LISTBOX:
        case CTLCOLOR_BTN:
        case CTLCOLOR_DLG:
        case CTLCOLOR_SCROLLBAR:
        case CTLCOLOR_STATIC:
    }
*/
}

MsgRouting AwtComponent::WmHScroll(UINT scrollCode, UINT pos,
				   HWND hScrollbar) {
    if (hScrollbar && hScrollbar != GetHWnd()) {  
	/* the last test should never happen */
        AwtComponent* sb = GetComponent(hScrollbar);
        if (sb) {
            sb->WmHScroll(scrollCode, pos, hScrollbar);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmVScroll(UINT scrollCode, UINT pos, HWND hScrollbar)
{
    if (hScrollbar && hScrollbar != GetHWnd()) {  
	/* the last test should never happen */
        AwtComponent* sb = GetComponent(hScrollbar);
        if (sb) {
            sb->WmVScroll(scrollCode, pos, hScrollbar);
        }
    }
    return mrDoDefault;
}

#define FT2INT64(ft) \
	((__int64)(ft).dwHighDateTime << 32 | (__int64)(ft).dwLowDateTime)

__int64 nowMillis()
{
    SYSTEMTIME st, st0;
    FILETIME   ft, ft0;

    (void)memset(&st0, 0, sizeof st0);
    st0.wYear  = 1970;
    st0.wMonth = 1;
    st0.wDay   = 1;
    SystemTimeToFileTime(&st0, &ft0);

    GetSystemTime(&st);
    SystemTimeToFileTime(&st, &ft);

    return (FT2INT64(ft) - FT2INT64(ft0)) / 10000;
    return 0;
}

MsgRouting AwtComponent::WmPaint(HDC)
{
    /* Get the rectangle that covers all update regions, if any exist. */
    RECT r;
    if (::GetUpdateRect(GetHWnd(), &r, FALSE)) {
        if ((r.right-r.left) > 0 && (r.bottom-r.top) > 0 &&
            m_peerObject != NULL && m_callbacksEnabled) {
	    /*
	     * Always call handlePaint, because the underlying control
	     * will have painted itself (the "background") before any
	     * paint method is called.
	     */
            DoCallback("handlePaint", "(IIII)V", 
                       r.left, r.top, r.right-r.left, r.bottom-r.top);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSetCursor(HWND hWnd, UINT hitTest, UINT message, 
                                     BOOL& retVal)
{
    if (hitTest == HTCLIENT) {
        ::SetCursor(GetCursor(TRUE));
        retVal = TRUE;
        return mrConsume;
    } else {
        return mrDoDefault;
    }
}

MsgRouting AwtComponent::WmAwtSetCursor(jobject cursor)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    CriticalSection::Lock l(GetLock());

    jint type = (env)->GetIntField(cursor, AwtCursor::typeID);

    HCURSOR hCursor;

#if 0
    hCursor = GetCursor(FALSE);
    /* DestroyCursor is only valid on cursors created with CreateCursor... */
    jint type;
    /* hania; this GetFieldBy Name is the wrong thing here */
    if (hCursor != NULL && (type = JNU_GetFieldByName(hCursor, "type")) == <custom type when it exists>) {
        VERIFY(::DestroyCursor(hCursor));
    }
#endif

    if (type == java_awt_Cursor_CUSTOM_CURSOR) {
        hCursor = (HCURSOR)
	  (env)->GetIntField(cursor, AwtCursor::pDataID);
    } else {
        LPCTSTR winCursor;
        switch (type) {
          case java_awt_Cursor_DEFAULT_CURSOR:
          default:
              winCursor = IDC_ARROW;
              break;
          case java_awt_Cursor_CROSSHAIR_CURSOR:
              winCursor = IDC_CROSS;
              break;
          case java_awt_Cursor_TEXT_CURSOR:
              winCursor = IDC_IBEAM;
              break;
          case java_awt_Cursor_WAIT_CURSOR:
              winCursor = IDC_WAIT;
              break;
          case java_awt_Cursor_NE_RESIZE_CURSOR:
          case java_awt_Cursor_SW_RESIZE_CURSOR:
              winCursor = IDC_SIZENESW;
              break;
          case java_awt_Cursor_SE_RESIZE_CURSOR:
          case java_awt_Cursor_NW_RESIZE_CURSOR:
              winCursor = IDC_SIZENWSE;
              break;
          case java_awt_Cursor_N_RESIZE_CURSOR:
          case java_awt_Cursor_S_RESIZE_CURSOR:
              winCursor = IDC_SIZENS;
              break;
          case java_awt_Cursor_W_RESIZE_CURSOR:
          case java_awt_Cursor_E_RESIZE_CURSOR:
              winCursor = IDC_SIZEWE;
              break;
          case java_awt_Cursor_HAND_CURSOR:
              winCursor = "HAND_CURSOR";
              break;
          case java_awt_Cursor_MOVE_CURSOR:
              winCursor = IDC_SIZEALL;
              break;
        }
        hCursor = ::LoadCursor(NULL, winCursor);
        if (hCursor == NULL) {
            /* Not a system cursor, check for resource. */
            hCursor = ::LoadCursor(AwtToolkit::GetInstance().GetModuleHandle(), 
                                   winCursor);
        }
    }
    if (hCursor == NULL) {
        /* Not found anywhere, use default. */
        hCursor = ::LoadCursor(NULL, IDC_ARROW);
        ASSERT(hCursor != NULL);
    }
    SetCursor(hCursor);

    /* Force the cursor to change if mouse is over it. */
    POINT pt;
    ::GetCursorPos(&pt);
    AwtComponent* mouseComp = AwtComponent::GetComponent(::WindowFromPoint(pt));
    if (mouseComp != NULL) {
        RECT r;
        ::GetClientRect(mouseComp->GetHWnd(), &r);
        ::MapWindowPoints(HWND_DESKTOP, mouseComp->GetHWnd(), &pt, 1);
        if (::PtInRect(&r, pt)) {
            mouseComp->SendMessage(WM_MOUSEMOVE, 
				   AwtToolkit::GetAsyncMouseKeyState(), 
				   MAKELPARAM((WORD)pt.x, (WORD)pt.y));
        }
    }

    return mrConsume;
}

MsgRouting AwtComponent::WmMouseEnter(UINT flags, int x, int y)
{
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_ENTERED, 
                   nowMillis(), x, y, 0, 0, 0);
    return mrConsume;	/* Don't pass our synthetic event on! */
}

static MSG* CreateMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
    MSG* msg = new MSG();
    msg->message = message;
    msg->wParam = wParam;
    msg->lParam = lParam;
    return msg;
}

/* Double-click variables. */
static ULONG multiClickTime = ::GetDoubleClickTime();
static int multiClickMaxX = ::GetSystemMetrics(SM_CXDOUBLECLK);
static int multiClickMaxY = ::GetSystemMetrics(SM_CYDOUBLECLK);
static AwtComponent* lastClickWnd = NULL;
static ULONG lastTime = 0;
static int lastClickX = 0;
static int lastClickY = 0;
static int clickCount = 0;

MsgRouting AwtComponent::WmMouseDown(UINT flags, int x, int y, int button)
{
    __int64 tm = nowMillis();

    ULONG now = (ULONG)tm;
    if (lastClickWnd == this &&
        (now - lastTime) <= multiClickTime && 
        abs(x - lastClickX) <= multiClickMaxX &&
        abs(y - lastClickY) <= multiClickMaxY) {
        clickCount++;
    }
    else {
        clickCount = 1;
    }
    lastClickWnd = this;
    lastTime = now;
    lastClickX = x;
    lastClickY = y;
    m_dragged = FALSE;

    MSG* msg = CreateMessage(lastMessage, flags, MAKELPARAM(x, y));
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_PRESSED, tm, x, y, 
                   GetJavaModifiers(flags, button), clickCount, 0, msg);

    return mrConsume;
}

MsgRouting AwtComponent::WmMouseUp(UINT flags, int x, int y, int button)
{
    MSG* msg = CreateMessage(lastMessage, flags, MAKELPARAM(x, y));
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_RELEASED, nowMillis(), 
                   x, y, GetJavaModifiers(flags, button), clickCount, 
                   (button == RIGHT_BUTTON? 1 : 0), msg);

    /*
     * If no movement, then report a click following the button release
     */
    if (!m_dragged) {
        SendMouseEvent(java_awt_event_MouseEvent_MOUSE_CLICKED, 
                       nowMillis(), x, y, 
                       GetJavaModifiers(flags, button), clickCount, 0);
    }
    m_dragged = FALSE;

    return mrConsume;
}

MsgRouting AwtComponent::WmMouseMove(UINT flags, int x, int y)
{
    static AwtComponent* lastComp = NULL;
    static int lastX = 0;
    static int lastY = 0;

    /*
     * Only report mouse move and drag events if a move or drag
     * actually happened -- Windows sends a WM_MOUSEMOVE in case the
     * app wants to modify the cursor.  
     */
    if (lastComp != this || x != lastX || y != lastY) {
        lastComp = this;
        lastX = x;
        lastY = y;

        long id;
        if ( (flags & ALL_MK_BUTTONS) != 0 ) {
            m_dragged = TRUE;
            id = java_awt_event_MouseEvent_MOUSE_DRAGGED;
        } else {
            id = java_awt_event_MouseEvent_MOUSE_MOVED;
	}

        MSG* msg = CreateMessage(lastMessage, flags, MAKELPARAM(x, y));
        
		if (flags & MK_LBUTTON) {
			SendMouseEvent(id, nowMillis(), x, y,
						GetJavaModifiers(flags, LEFT_BUTTON), 0, 0, msg);
		}
		else if (flags & MK_MBUTTON) {
            SendMouseEvent(id, nowMillis(), x, y,
                       GetJavaModifiers(flags, MIDDLE_BUTTON), 0, 0, msg);
		}		
		else if (flags & MK_RBUTTON) {
            SendMouseEvent(id, nowMillis(), x, y,
                        GetJavaModifiers(flags, RIGHT_BUTTON), 0, 0, msg);
		}
		else {
			SendMouseEvent(id, nowMillis(), x, y,
                       GetJavaModifiers(flags, 0), 0, 0, msg);
        }

    } else {
        ::SetCursor(GetCursor(TRUE));
    }
    return mrConsume;
}

MsgRouting AwtComponent::WmMouseExit(UINT flags, int x, int y)
{
    SendMouseEvent(java_awt_event_MouseEvent_MOUSE_EXITED, 
                   nowMillis(), x, y, 0, 0, 0);
    return mrConsume;	/* Don't pass our synthetic event on! */
}

long AwtComponent::GetJavaModifiers(UINT flags, int mouseButton)
{
    int modifiers = 0;

    if ((1<<24) & flags) {
        modifiers |= java_awt_event_InputEvent_META_MASK;
    }	
    if (HIBYTE(::GetAsyncKeyState(VK_CONTROL)) || 
        mouseButton == MIDDLE_BUTTON) {
        modifiers |= java_awt_event_InputEvent_CTRL_MASK;
    }
    if (HIBYTE(::GetAsyncKeyState(VK_SHIFT))) {
        modifiers |= java_awt_event_InputEvent_SHIFT_MASK;
    }
    if (HIBYTE(::GetAsyncKeyState(VK_MBUTTON)) || 
        HIBYTE(::GetAsyncKeyState(VK_MENU))) {
        modifiers |= java_awt_event_InputEvent_ALT_MASK;
    }

    /* 
     * set "right mouse" mask depending on whether system is set to be
     * left or right-handed.
     */
    if (::GetSystemMetrics(SM_SWAPBUTTON)) {
	if (HIBYTE(::GetAsyncKeyState(VK_LBUTTON))) {
	    modifiers |= java_awt_event_InputEvent_META_MASK;
	} 

    } else {
	if (HIBYTE(::GetAsyncKeyState(VK_RBUTTON))) {
	    modifiers |= java_awt_event_InputEvent_META_MASK;
	}
    }

    /* Mouse buttons are already set correctly for left/right handedness */
    if (mouseButton == LEFT_BUTTON) {
        modifiers |= java_awt_event_InputEvent_BUTTON1_MASK;
    }
    if (mouseButton == MIDDLE_BUTTON) {
        modifiers |= java_awt_event_InputEvent_ALT_MASK;
    }
    if (mouseButton == RIGHT_BUTTON) {
        modifiers |= java_awt_event_InputEvent_META_MASK;
    }
    return modifiers;
}

// This constant is defined in the Japanese version of VC++5.0, but not the US version
#ifndef VK_CONVERT
#define VK_CONVERT 0x001C
#endif

typedef struct {
    UINT javaKey;
    UINT windowsKey;
} KeyMapEntry;

KeyMapEntry keyMapTable[] = {
	{java_awt_event_KeyEvent_VK_ENTER, VK_RETURN},
	{java_awt_event_KeyEvent_VK_BACK_SPACE, VK_BACK},
	{java_awt_event_KeyEvent_VK_TAB, VK_TAB},
	{java_awt_event_KeyEvent_VK_CANCEL, VK_CANCEL},
	{java_awt_event_KeyEvent_VK_CLEAR, VK_CLEAR},
	{java_awt_event_KeyEvent_VK_SHIFT, VK_SHIFT},
	{java_awt_event_KeyEvent_VK_CONTROL, VK_CONTROL},
	{java_awt_event_KeyEvent_VK_ALT, VK_MENU},
	{java_awt_event_KeyEvent_VK_PAUSE, VK_PAUSE},
	{java_awt_event_KeyEvent_VK_CAPS_LOCK, VK_CAPITAL},
	{java_awt_event_KeyEvent_VK_ESCAPE, VK_ESCAPE},
	{java_awt_event_KeyEvent_VK_SPACE, VK_SPACE},
	{java_awt_event_KeyEvent_VK_PAGE_UP, VK_PRIOR},
	{java_awt_event_KeyEvent_VK_PAGE_DOWN, VK_NEXT},
	{java_awt_event_KeyEvent_VK_END, VK_END},
	{java_awt_event_KeyEvent_VK_HOME, VK_HOME},
	{java_awt_event_KeyEvent_VK_LEFT, VK_LEFT},
	{java_awt_event_KeyEvent_VK_UP, VK_UP},
	{java_awt_event_KeyEvent_VK_RIGHT, VK_RIGHT},
	{java_awt_event_KeyEvent_VK_DOWN, VK_DOWN},
	{java_awt_event_KeyEvent_VK_COMMA, 0x00BC},
	{java_awt_event_KeyEvent_VK_MINUS, 0x00BD},
	{java_awt_event_KeyEvent_VK_PERIOD, 0x00BE},
	{java_awt_event_KeyEvent_VK_SLASH, 0x0BF},
	{java_awt_event_KeyEvent_VK_0, '0'},
	{java_awt_event_KeyEvent_VK_1, '1'},
	{java_awt_event_KeyEvent_VK_2, '2'},
	{java_awt_event_KeyEvent_VK_3, '3'},
	{java_awt_event_KeyEvent_VK_4, '4'},
	{java_awt_event_KeyEvent_VK_5, '5'},
	{java_awt_event_KeyEvent_VK_6, '6'},
	{java_awt_event_KeyEvent_VK_7, '7'},
	{java_awt_event_KeyEvent_VK_8, '8'},
	{java_awt_event_KeyEvent_VK_9, '9'},
	{java_awt_event_KeyEvent_VK_SEMICOLON, 0x00BA},
	{java_awt_event_KeyEvent_VK_EQUALS, 0x00BB},
	{java_awt_event_KeyEvent_VK_A, 'A'},
	{java_awt_event_KeyEvent_VK_B, 'B'},
	{java_awt_event_KeyEvent_VK_C, 'C'},
	{java_awt_event_KeyEvent_VK_D, 'D'},
	{java_awt_event_KeyEvent_VK_E, 'E'},
	{java_awt_event_KeyEvent_VK_F, 'F'},
	{java_awt_event_KeyEvent_VK_G, 'G'},
	{java_awt_event_KeyEvent_VK_H, 'H'},
	{java_awt_event_KeyEvent_VK_I, 'I'},
	{java_awt_event_KeyEvent_VK_J, 'J'},
	{java_awt_event_KeyEvent_VK_K, 'K'},
	{java_awt_event_KeyEvent_VK_L, 'L'},
	{java_awt_event_KeyEvent_VK_M, 'M'},
	{java_awt_event_KeyEvent_VK_N, 'N'},
	{java_awt_event_KeyEvent_VK_O, 'O'},
	{java_awt_event_KeyEvent_VK_P, 'P'},
	{java_awt_event_KeyEvent_VK_Q, 'Q'},
	{java_awt_event_KeyEvent_VK_R, 'R'},
	{java_awt_event_KeyEvent_VK_S, 'S'},
	{java_awt_event_KeyEvent_VK_T, 'T'},
	{java_awt_event_KeyEvent_VK_U, 'U'},
	{java_awt_event_KeyEvent_VK_V, 'V'},
	{java_awt_event_KeyEvent_VK_W, 'W'},
	{java_awt_event_KeyEvent_VK_X, 'X'},
	{java_awt_event_KeyEvent_VK_Y, 'Y'},
	{java_awt_event_KeyEvent_VK_Z, 'Z'},
	{java_awt_event_KeyEvent_VK_OPEN_BRACKET, 0x00DB},
	{java_awt_event_KeyEvent_VK_BACK_SLASH, 0x00DC},
	{java_awt_event_KeyEvent_VK_CLOSE_BRACKET, 0x00DD},
	{java_awt_event_KeyEvent_VK_NUMPAD0, VK_NUMPAD0},
	{java_awt_event_KeyEvent_VK_NUMPAD1, VK_NUMPAD1},
	{java_awt_event_KeyEvent_VK_NUMPAD2, VK_NUMPAD2},
	{java_awt_event_KeyEvent_VK_NUMPAD3, VK_NUMPAD3},
	{java_awt_event_KeyEvent_VK_NUMPAD4, VK_NUMPAD4},
	{java_awt_event_KeyEvent_VK_NUMPAD5, VK_NUMPAD5},
	{java_awt_event_KeyEvent_VK_NUMPAD6, VK_NUMPAD6},
	{java_awt_event_KeyEvent_VK_NUMPAD7, VK_NUMPAD7},
	{java_awt_event_KeyEvent_VK_NUMPAD8, VK_NUMPAD8},
	{java_awt_event_KeyEvent_VK_NUMPAD9, VK_NUMPAD9},
	{java_awt_event_KeyEvent_VK_MULTIPLY, VK_MULTIPLY},
	{java_awt_event_KeyEvent_VK_ADD, VK_ADD},
	{java_awt_event_KeyEvent_VK_SEPARATER, VK_SEPARATOR},
	{java_awt_event_KeyEvent_VK_SUBTRACT, VK_SUBTRACT},
	{java_awt_event_KeyEvent_VK_DECIMAL, VK_DECIMAL},
	{java_awt_event_KeyEvent_VK_DIVIDE, VK_DIVIDE},
	{java_awt_event_KeyEvent_VK_DELETE, VK_DELETE},
	{java_awt_event_KeyEvent_VK_NUM_LOCK, VK_NUMLOCK},
	{java_awt_event_KeyEvent_VK_SCROLL_LOCK, VK_SCROLL},
	{java_awt_event_KeyEvent_VK_F1, VK_F1},
	{java_awt_event_KeyEvent_VK_F2, VK_F2},
	{java_awt_event_KeyEvent_VK_F3, VK_F3},
	{java_awt_event_KeyEvent_VK_F4, VK_F4},
	{java_awt_event_KeyEvent_VK_F5, VK_F5},
	{java_awt_event_KeyEvent_VK_F6, VK_F6},
	{java_awt_event_KeyEvent_VK_F7, VK_F7},
	{java_awt_event_KeyEvent_VK_F8, VK_F8},
	{java_awt_event_KeyEvent_VK_F9, VK_F9},
	{java_awt_event_KeyEvent_VK_F10, VK_F10},
	{java_awt_event_KeyEvent_VK_F11, VK_F11},
	{java_awt_event_KeyEvent_VK_F12, VK_F12},
	{java_awt_event_KeyEvent_VK_F13, VK_F13},
	{java_awt_event_KeyEvent_VK_F14, VK_F14},
	{java_awt_event_KeyEvent_VK_F15, VK_F15},
	{java_awt_event_KeyEvent_VK_F16, VK_F16},
	{java_awt_event_KeyEvent_VK_F17, VK_F17},
	{java_awt_event_KeyEvent_VK_F18, VK_F18},
	{java_awt_event_KeyEvent_VK_F19, VK_F19},
	{java_awt_event_KeyEvent_VK_F20, VK_F20},
	{java_awt_event_KeyEvent_VK_F21, VK_F21},
	{java_awt_event_KeyEvent_VK_F22, VK_F22},
	{java_awt_event_KeyEvent_VK_F23, VK_F23},
	{java_awt_event_KeyEvent_VK_F24, VK_F24},
	{java_awt_event_KeyEvent_VK_PRINTSCREEN, VK_SNAPSHOT},
	{java_awt_event_KeyEvent_VK_INSERT, VK_INSERT},
	{java_awt_event_KeyEvent_VK_HELP, VK_HELP},
	{java_awt_event_KeyEvent_VK_BACK_QUOTE, 0x00C0},
	{java_awt_event_KeyEvent_VK_QUOTE, 0x00DE},
	{java_awt_event_KeyEvent_VK_CONVERT, VK_CONVERT},
	{java_awt_event_KeyEvent_VK_NONCONVERT, 0x001D},
	{java_awt_event_KeyEvent_VK_KANJI, 0x0019},
	{java_awt_event_KeyEvent_VK_ALPHANUMERIC, VK_DBE_ALPHANUMERIC},
	{java_awt_event_KeyEvent_VK_KATAKANA, VK_DBE_KATAKANA},
	{java_awt_event_KeyEvent_VK_HIRAGANA, VK_DBE_HIRAGANA},
	{java_awt_event_KeyEvent_VK_FULL_WIDTH, VK_DBE_DBCSCHAR},
	{java_awt_event_KeyEvent_VK_HALF_WIDTH, VK_DBE_SBCSCHAR},
	{java_awt_event_KeyEvent_VK_ROMAN_CHARACTERS, VK_DBE_ROMAN},
	{java_awt_event_KeyEvent_CHAR_UNDEFINED, 0}
};

void AwtComponent::JavaKeyToWindowsKey(UINT javaKey, UINT *windowsKey, UINT *modifiers)
{
    // Handle the few cases where a Java VK code corresponds to a Windows
    // key/modifier combination
    switch (javaKey) {
        case java_awt_event_KeyEvent_VK_ALL_CANDIDATES:
            *windowsKey = VK_CONVERT;
            *modifiers = java_awt_event_InputEvent_ALT_MASK;
            return;
        case java_awt_event_KeyEvent_VK_PREVIOUS_CANDIDATE:
            *windowsKey = VK_CONVERT;
            *modifiers = java_awt_event_InputEvent_SHIFT_MASK;
            return;
        case java_awt_event_KeyEvent_VK_CODE_INPUT:
            *windowsKey = VK_DBE_ALPHANUMERIC;
            *modifiers = java_awt_event_InputEvent_ALT_MASK;
            return;
    }
    
    // for the general case, use a bi-directional table
    for (int i = 0; keyMapTable[i].windowsKey != 0; i++) {
        if (keyMapTable[i].javaKey == javaKey) {
            *windowsKey = keyMapTable[i].windowsKey;
            *modifiers = 0;
            return;
        }
    }
    
    *windowsKey = 0;
    *modifiers = 0;
    return;
}

UINT AwtComponent::WindowsKeyToJavaKey(UINT windowsKey, UINT modifiers)
{
    // Handle the few cases where we need to take the modifier into
    // consideration for the Java VK code so that function keys can get
    // recognized in a platform-independent way.
    switch (windowsKey) {
        case VK_CONVERT:
            if ((modifiers & java_awt_event_InputEvent_ALT_MASK) != 0) {
                return java_awt_event_KeyEvent_VK_ALL_CANDIDATES;
            }
            if ((modifiers & java_awt_event_InputEvent_SHIFT_MASK) != 0) {
                return java_awt_event_KeyEvent_VK_PREVIOUS_CANDIDATE;
            }
            break;
        case VK_DBE_ALPHANUMERIC:
            if ((modifiers & java_awt_event_InputEvent_ALT_MASK) != 0) {
                return java_awt_event_KeyEvent_VK_CODE_INPUT;
            }
            break;
    };
    
    // for the general case, use a bi-directional table
    for (int i = 0; keyMapTable[i].windowsKey != 0; i++) {
        if (keyMapTable[i].windowsKey == windowsKey) {
            return keyMapTable[i].javaKey;
        }
    }
    
    return java_awt_event_KeyEvent_VK_UNDEFINED;
}

// determine if a key is a numpad numeric key (regardless of NUMLOCK's state)
BOOL AwtComponent::IsNumPadNumberKey( UINT vkey )
{
    UINT	keyScanCode;
    UINT	npKeyScanCode;
    
    keyScanCode = ::MapVirtualKey(vkey, 0);

    // compare scancodes instead of virtual key codes so we don't
    // distinguish between NUMLOCK being on or off
    for( UINT vkNumPad = VK_NUMPAD0; vkNumPad <= VK_NUMPAD9; vkNumPad++ ) {
	npKeyScanCode = ::MapVirtualKey(vkNumPad, 0);
	if (keyScanCode == npKeyScanCode) {
	    return TRUE;
	}
    }
    return FALSE;
}

UINT AwtComponent::WindowsKeyToJavaChar(UINT wkey, UINT modifiers)
{
    BYTE    keyboardState[AwtToolkit::KB_STATE_SIZE];
    WCHAR   unicodeChar[2];
    WORD    mbChar;
    UINT    scancode;
    int     conversionResult;

    scancode = ::MapVirtualKey(wkey, 0);
    AwtToolkit::GetAsyncKeyboardState(keyboardState);
    
    // apply modifiers if necessary
    if (modifiers)
    {
        if (modifiers & java_awt_Event_SHIFT_MASK)
        {
            keyboardState[VK_SHIFT] |= 0x80;
        }
        if (modifiers & java_awt_Event_CTRL_MASK)
        {
 	    //fix for bug #4098210
 	    keyboardState[VK_CONTROL] &= ~0x80;
 	    if (wkey >= java_awt_event_KeyEvent_VK_A &&
 		wkey <= java_awt_event_KeyEvent_VK_Z)
 	    {
 		keyboardState[VK_CONTROL] |= 0x80;
 	    }
 	    if ((((wkey & 0x7F) == java_awt_event_KeyEvent_VK_OPEN_BRACKET ) ||
 		 ((wkey & 0x7F) == java_awt_event_KeyEvent_VK_BACK_SLASH   ) ||
 		 ((wkey & 0x7F) == java_awt_event_KeyEvent_VK_CLOSE_BRACKET)) && 
 		 ( (modifiers & java_awt_Event_SHIFT_MASK) == 0 ) )
 	    {
 		keyboardState[VK_CONTROL] |= 0x80;
 	    }
 	    if ( ( wkey == (UINT)::VkKeyScan('-')        ) && 
 		 ( modifiers & java_awt_Event_SHIFT_MASK ) )
 	    {
 		keyboardState[VK_CONTROL] |= 0x80;
 	    }
        }
        if (modifiers & java_awt_Event_ALT_MASK)
        {
    	    //fix for bug #4098210
    	    if (keyboardState[VK_CONTROL] & 0x80)
    		keyboardState[VK_MENU] &= ~0x80;
   	    else
   	        keyboardState[VK_MENU] |= 0x80;
        }
    }

    // If the windows key is a return, wkey will equal 13 ('\r')
    // In this case, we want to return java_awt_event_KeyEvent_VK_ENTER
    // We'll do that here, since ToAscii will convert the 13 to ASCII 13, 
    // and we want it converted to java_awt_event_KeyEvent_VK_ENTER, or 10. 
    if (wkey == VK_RETURN)
        return java_awt_event_KeyEvent_VK_ENTER;

    // fix 4173679/4122715 - Alt+NumPad combinations
    // can't call ToAscii since this interferes with
    // multi keystroke entry of the character
    if ( ( (modifiers&java_awt_Event_ALT_MASK) != 0) &&
    	  IsNumPadNumberKey(wkey) ) {
	return wkey - VK_NUMPAD0 + '0';
    }

    // instead of creating our own conversion tables, I'll let Win32
    // convert the character for me.
    conversionResult = ToAscii(wkey, scancode, keyboardState, &mbChar, 0);

    if (conversionResult == 0)
    {
        // No translation available -- try known conversions or else punt.
        if (wkey == VK_DELETE)
        {
            return java_awt_event_KeyEvent_VK_DELETE;
        }
        if (wkey >= VK_NUMPAD0 && wkey <= VK_NUMPAD9)
        {
            return wkey - VK_NUMPAD0 + '0';
        }
        return java_awt_event_KeyEvent_CHAR_UNDEFINED;
    }

    // the caller expects a Unicode character.
    VERIFY(::MultiByteToWideChar(GetCodePage(), MB_PRECOMPOSED, (LPCSTR)&mbChar, 1, unicodeChar, 1));
    if (m_isWin95 && conversionResult < 0){
       //For win95 a deadkey should not come here, we block
       //it at WindowProc, the only hole is that we have two 
       //consecutive deadkeys, in this case wee need to reset 
       //the keyboard status to make the next TranslateMessage()
       //work correctly, before I know how to do it, the following
       //lines seem to work.
       scancode = ::MapVirtualKey(0x41, 0);
       keyboardState[0x41] |= 0x80;
       ToAscii(0x41, scancode, keyboardState, &mbChar, 0);
    }
    return unicodeChar[0];
}

MsgRouting AwtComponent::WmKeyDown(UINT wkey, UINT repCnt, 
				   UINT flags, BOOL system)
{

    // VK_PROCESSKEY is a special value which means 
    //		"Current IME wants to consume this KeyEvent"
    // Real key code is saved by IMM32.DLL and can be retrieved by 
    // calling ImmGetVirtualKey();
#ifdef TWO_PASS_DESIGN
    MSG* msg = CreateMessage((system ? WM_AWT_PRE_SYSKEYDOWN : WM_AWT_PRE_KEYDOWN), 
                                wkey, MAKELPARAM(repCnt, flags));
    if(wkey==VK_PROCESSKEY)	wkey=ImmGetVirtualKey(GetHWnd());
#else
    if(wkey==VK_PROCESSKEY)		return mrDoDefault;
    MSG* msg = CreateMessage((system ? WM_SYSKEYDOWN : WM_KEYDOWN), 
                                wkey, MAKELPARAM(repCnt, flags));
#endif

    UINT modifiers = GetJavaModifiers(flags, 0);
    UINT jkey = WindowsKeyToJavaKey(wkey, modifiers);
    UINT character = WindowsKeyToJavaChar(wkey, modifiers);

    SendKeyEvent(java_awt_event_KeyEvent_KEY_PRESSED, nowMillis(), jkey, 
                 character, modifiers, msg);
    return mrConsume;
}

MsgRouting AwtComponent::WmKeyUp(UINT wkey, UINT repCnt, 
				 UINT flags, BOOL system)
{

    // VK_PROCESSKEY is a special value which means 
    //		"Current IME wants to consume this KeyEvent"
    // Real key code is saved by IMM32.DLL and can be retrieved by 
    // calling ImmGetVirtualKey();
#ifdef TWO_PASS_DESIGN
    MSG* msg = CreateMessage((system ? WM_AWT_PRE_SYSKEYUP : WM_AWT_PRE_KEYUP), 
                                wkey, MAKELPARAM(repCnt, flags));
    if(wkey==VK_PROCESSKEY)     wkey=ImmGetVirtualKey(GetHWnd());
#else
    if(wkey==VK_PROCESSKEY)	return mrDoDefault;
    MSG* msg = CreateMessage((system ? WM_SYSKEYUP : WM_KEYUP), 
                                wkey, MAKELPARAM(repCnt, flags));
#endif

    UINT modifiers = GetJavaModifiers(flags, 0);
    UINT jkey = WindowsKeyToJavaKey(wkey, modifiers);
    UINT character = WindowsKeyToJavaChar(wkey, modifiers);

    SendKeyEvent(java_awt_event_KeyEvent_KEY_RELEASED, nowMillis(), jkey,
                 character, modifiers, msg);
    return mrConsume;
}

static BOOL IsDBCSEditClass(AwtComponent *comp)
{
    const char* className = comp->GetClassName();
    if  ( stricmp(className,"Edit")==0 || stricmp(className,"ListBox")==0
        || stricmp(className,"ComboBox")==0 )
      return TRUE;
    else
      return FALSE;
}

MsgRouting AwtComponent::WmInputLangChange(UINT charset, UINT hKeyboardLayout)
{
    // Normally we would be able to use charset and TranslateCharSetInfo
    // to get a code page that should be associated with this keyboard
    // layout change. However, there seems to be an NT 4.0 bug associated
    // with the WM_INPUTLANGCHANGE message, which makes the charset parameter
    // unreliable, especially on Asian systems. Our workaround uses the 
    // keyboard layout handle instead.

    // a LANGID is stored in the bottom word of hKeyboardLayout
    m_idLang = LOWORD(hKeyboardLayout);
    m_CodePage = LangToCodePage(m_idLang);
    return mrConsume;
}

// Convert Language ID to CodePage
UINT AwtComponent::LangToCodePage(LANGID idLang)
{
    TCHAR strCodePage[MAX_ACP_STR_LEN];
    // use the LANGID to create a LCID
    LCID idLocale = MAKELCID(idLang, SORT_DEFAULT);
    // get the ANSI code page associated with this locale
    if (GetLocaleInfo(idLocale, LOCALE_IDEFAULTANSICODEPAGE, strCodePage, sizeof(strCodePage)) > 0 )
        return _ttoi(strCodePage);
    else
        return GetACP();
}


MsgRouting AwtComponent::WmIMEChar(UINT character, UINT repCnt, UINT flags, BOOL system)
{
    // We will simply create Java events here.
    TCHAR mbChar[3] = { _T('\0'), _T('\0'), _T('\0')};
    WCHAR unicodeChar;

    mbChar[0] = HIBYTE(character)?HIBYTE(character):LOBYTE(character);
    mbChar[1] = HIBYTE(character)?LOBYTE(character):_T('\0');

    ::MultiByteToWideChar(GetCodePage(), 0, mbChar, -1, &unicodeChar, 1);
    MSG* pMsg = CreateMessage(WM_IME_CHAR, character, MAKELPARAM(repCnt, flags));
    SendKeyEvent(java_awt_event_KeyEvent_KEY_TYPED, nowMillis(), 
                 java_awt_event_KeyEvent_VK_UNDEFINED, unicodeChar, 
                 GetJavaModifiers(flags, 0), pMsg);
    return mrConsume;
}

MsgRouting AwtComponent::WmChar(UINT character, UINT repCnt, UINT flags, 
				BOOL system)
{

    // Will only get WmChar messages with DBCS if we create them for an Edit class 
    // in the WmForwardChar method. These synthesized DBCS chars are ok to pass on 
    // directly to the default window procedure. They've already been filtered through 
    // the Java key event queue. We will never get the trail byte since the edit class 
    // will PeekMessage(&msg, hwnd, WM_CHAR, WM_CHAR, PM_REMOVE).
    // I would like to be able to pass this character off via WM_AWT_FORWARD_BYTE, but 
    // the Edit classes don't seem to like that.
    // Begin pollution
    if (IsDBCSLeadByteEx(GetCodePage(), BYTE(character)))
    {  
        return mrDoDefault;
    }
    // End pollution
  
    // We will simply create Java events here.
    WCHAR unicodeChar = L'\0';
    UINT message = 0;

    message = system?WM_SYSCHAR:WM_CHAR;
    //fix for bug #4098210
    if (character == java_awt_event_KeyEvent_VK_DELETE) {
 	if ((::GetKeyState(character) & 0x80) == 0)
 	    unicodeChar = java_awt_event_KeyEvent_VK_BACK_SPACE;
    }
    else
	::MultiByteToWideChar(GetCodePage(), 0, (TCHAR*)&character, 1, &unicodeChar, 1);
    MSG* pMsg = CreateMessage(message, character, MAKELPARAM(repCnt, flags));

    // If InputMethod is disabled or character is control code, KeyTyped Event is generated.
    // Otherwise, InputMethod Event is generated.
    if ( m_useNativeCompWindow || unicodeChar<=0x001F) {
	// Enter key generates \r in windows, but \n is required in java
	if(unicodeChar==VK_RETURN)	unicodeChar = java_awt_event_KeyEvent_VK_ENTER;
	SendKeyEvent(java_awt_event_KeyEvent_KEY_TYPED, nowMillis(), 
				 java_awt_event_KeyEvent_VK_UNDEFINED, unicodeChar,
				 GetJavaModifiers(flags, 0), pMsg);
    } else {    
        jstring jstr = ((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2))->NewString(&unicodeChar,1);
        SendInputMethodEvent(java_awt_event_InputMethodEvent_INPUT_METHOD_TEXT_CHANGED, jstr, 
                             0, NULL, NULL, 0, NULL, NULL, 
                             1, 0, 0);
    }

    return mrConsume;
}

MsgRouting AwtComponent::WmForwardChar(WCHAR character, LPARAM lParam, 
				       BOOL synthetic)
{
    // This message is sent from the Java key event handler.
    TCHAR mbChar[2] = {_T('\0'), _T('\0')};

    int cBytes = ::WideCharToMultiByte(GetCodePage(), 0, &character, 1, mbChar, 2, NULL, NULL);
    if (cBytes!=1 && cBytes!=2)    return mrConsume;

    if (IsDBCSEditClass(this) && cBytes==2)
    {
        // The first WM_CHAR message will get handled by the WmChar, but
        // the second WM_CHAR message will get picked off by the Edit class.
        // WmChar will never see it.
        // If an Edit class gets a lead byte, it immediately calls PeekMessage
        // and pulls the trail byte out of the message queue.
        ::PostMessage(GetHWnd(), WM_CHAR, mbChar[0] & 0x00ff, lParam);
        ::PostMessage(GetHWnd(), WM_CHAR, mbChar[1] & 0x00ff, lParam);
    }
    else
    {
        MSG* pMsg;
        pMsg = CreateMessage(WM_CHAR, mbChar[0] & 0x00ff, lParam);
	::PostMessage(GetHWnd(), WM_AWT_FORWARD_BYTE, (WPARAM)synthetic,
		      (LPARAM)pMsg);
        if (mbChar[1])
        {
            pMsg = CreateMessage(WM_CHAR, mbChar[1] & 0x00ff, lParam);
	    ::PostMessage(GetHWnd(), WM_AWT_FORWARD_BYTE, (WPARAM)synthetic,
			  (LPARAM)pMsg);
        }
    }
    return mrConsume;
}

MsgRouting AwtComponent::WmPaste()
{
    return mrDoDefault;
}


// support IME Composition messages
void AwtComponent::SetCompositionWindow()
{
    HIMC hIMC = ImmGetContext(m_hwnd);
    COMPOSITIONFORM cf = { CFS_DEFAULT, {0,0}, {0,0,0,0} };
    ImmSetCompositionWindow(hIMC, &cf);
}

void AwtComponent::SetCandidateWindow(int iCandType,
				      BOOL getLocation)
{
    HIMC hIMC = ImmGetContext(m_hwnd);
    CANDIDATEFORM cf;
    RECT rc;
    cf.dwIndex = iCandType;
    cf.dwStyle = CFS_CANDIDATEPOS;

    // For initialization, the position information is not
    // obtained. This is because the InquireCandidatePosition() call
    // results in calling Component.getLocationOnScreen() that locks
    // the entire AWT (i.e., Component.LOCK).
    if (getLocation == TRUE) {
	InquireCandidatePosition(&cf.ptCurrentPos);
	GetWindowRect(m_hwnd, &rc);
	cf.ptCurrentPos.x -= rc.left;
	cf.ptCurrentPos.y -= rc.top;
    } else {
	cf.ptCurrentPos.x = 0;
	cf.ptCurrentPos.y = 0;
    }

    BOOL ret = ImmSetCandidateWindow(hIMC, &cf);
}

MsgRouting AwtComponent::WmImeSetContext(BOOL fSet, LPARAM *lplParam)
{
    if (fSet) {   
        LPARAM lParam = *lplParam;
        if (m_useNativeCompWindow) {   
            // set position and style of native composing window.
            if (lParam & ISC_SHOWUICOMPOSITIONWINDOW)
                SetCompositionWindow();
        } else {
            // stop to draw native composing window.
            *lplParam &= ~ISC_SHOWUICOMPOSITIONWINDOW;

            // set position and style of candidate window.
            if (lParam & ISC_SHOWUICANDIDATEWINDOW)
                SetCandidateWindow(0, FALSE);
            if (lParam & ISC_SHOWUICANDIDATEWINDOW<<1)
                SetCandidateWindow(1, FALSE);
            if (lParam & ISC_SHOWUICANDIDATEWINDOW<<2)
                SetCandidateWindow(2, FALSE);
            if (lParam & ISC_SHOWUICANDIDATEWINDOW<<3)
                SetCandidateWindow(3, FALSE);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmImeNotify(UINT subMsg, UINT bitsCandType)
{
    if (!m_useNativeCompWindow && subMsg==(unsigned)IMN_OPENCANDIDATE) {
        UINT bits = 1;
        for (int iCandType=0; iCandType<32; iCandType++, bits<<=1) {
            if ( bitsCandType & bits )
		SetCandidateWindow(iCandType, TRUE);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmImeStartComposition()
{
    if (m_useNativeCompWindow) {
        SetCompositionWindow();
        return mrDoDefault;
    } else
        return mrConsume;
}
 
MsgRouting AwtComponent::WmImeEndComposition()
{
    if (m_useNativeCompWindow)   return mrDoDefault;

    MSG* msg = CreateMessage(WM_IME_ENDCOMPOSITION, 0, 0);
    SendInputMethodEvent(java_awt_event_InputMethodEvent_INPUT_METHOD_TEXT_CHANGED, 
			 NULL, 0, NULL, NULL, 0, NULL, NULL, 0, 0, 0 );
    return mrConsume;
}

MsgRouting AwtComponent::WmImeComposition(WORD wChar, long flags)
{
    if (m_useNativeCompWindow)   return mrDoDefault;

 
    HIMC hIMC = ImmGetContext(m_hwnd);
    ASSERT(hIMC!=0);

    //
    // Get Result String (=Committed String)
    //
    int	    cResultStrW = 0;
    jstring jsResultStrW = NULL;
    int	    cResultClauseW = 0;
    int*    bndResultClauseW = NULL;
    jstring* readingResultClauseW = NULL;
    
    if (flags&(GCS_RESULTREADSTR|GCS_RESULTREADCLAUSE|GCS_RESULTSTR|GCS_RESULTCLAUSE)) {

	//
	//Retrieve all information about Result String from InputContext
	//
	LPSTR lpResultStrA;
	int cResultStrA = GetContextData(hIMC, GCS_RESULTSTR, (LPVOID *)&lpResultStrA);
	LPDWORD lpResultClauseA;
	int cResultClauseA = GetContextData(hIMC, GCS_RESULTCLAUSE, (LPVOID *)&lpResultClauseA) / sizeof(DWORD) - 1;
	LPSTR lpResultReadStrA;
	int cResultReadStrA = GetContextData(hIMC, GCS_RESULTREADSTR, (LPVOID *)&lpResultReadStrA);
	LPDWORD lpResultReadClauseA;
	int cResultReadClauseA = GetContextData(hIMC, GCS_RESULTREADCLAUSE, (LPVOID *)&lpResultReadClauseA) / sizeof(DWORD) - 1;

	//
	//Convert string from ANSI to UNICODE and make offset conversion table
	//
	jsResultStrW = MakeJavaString( lpResultStrA, cResultStrA );
	int *tblCharOffResultStrA = MakeCharOffsetTable( lpResultStrA, cResultStrA );
	cResultStrW = tblCharOffResultStrA[cResultStrA];

	//
	//Make Clause and Reading information for UNICODE string
	//
	cResultClauseW = MakeClauseInformation( cResultStrA, tblCharOffResultStrA, 
						cResultClauseA, lpResultClauseA, 
						cResultReadStrA, lpResultReadStrA, 
						cResultReadClauseA, lpResultReadClauseA, 
						&bndResultClauseW, &readingResultClauseW );

	delete [] lpResultStrA;
	delete [] lpResultClauseA;
	delete [] lpResultReadStrA;
	delete [] lpResultReadClauseA;
	delete [] tblCharOffResultStrA;
    }


    //
    // Get Composed String
    //
    int		cCompStrW = 0;
    jstring	jsCompStrW = NULL;
    int		cCompClauseW = 0;
    int*	bndCompClauseW = NULL;
    jstring*	readingCompClauseW = NULL;
    int		cCompAttrW = 0;
    int*	bndCompAttrW = NULL;
    BYTE*	valCompAttrW = NULL;
    int		cursorPosW = cResultStrW;
    int		visiblePosW = cResultStrW;

    if ( flags & (GCS_COMPREADSTR|GCS_COMPREADATTR|GCS_COMPREADCLAUSE|GCS_COMPSTR|GCS_COMPATTR|GCS_COMPCLAUSE|GCS_CURSORPOS|GCS_DELTASTART))
    {
	//
	//Retrieve all information about Composing String from InputContext
	//
	LPSTR lpCompStrA;
	int cCompStrA = GetContextData(hIMC, GCS_COMPSTR, (LPVOID *)&lpCompStrA);
	LPBYTE lpCompAttrA;
	int cCompAttrA = GetContextData(hIMC, GCS_COMPATTR, (LPVOID *)&lpCompAttrA);
	LPDWORD lpCompClauseA;
	int cCompClauseA = GetContextData(hIMC, GCS_COMPCLAUSE, (LPVOID *)&lpCompClauseA) / sizeof(DWORD) - 1;
	LPSTR lpCompReadStrA;
	int cCompReadStrA = GetContextData(hIMC, GCS_COMPREADSTR, (LPVOID *)&lpCompReadStrA);
	LPDWORD lpCompReadClauseA;
	int cCompReadClauseA = GetContextData(hIMC, GCS_COMPREADCLAUSE, (LPVOID *)&lpCompReadClauseA) / sizeof(DWORD) - 1;
	int cursorPosA = ImmGetCompositionString(hIMC, GCS_CURSORPOS, NULL, 0);

	//
	//Convert string from ANSI to UNICODE and make offset conversion table
	//
	jsCompStrW = MakeJavaString( lpCompStrA, cCompStrA );
	int *tblCharOffCompStrA = MakeCharOffsetTable( lpCompStrA, cCompStrA );
	cCompStrW = tblCharOffCompStrA[cCompStrA];

	//
	//Make Clause and Reading information for UNICODE string
	//
	cCompClauseW = MakeClauseInformation( cCompStrA, tblCharOffCompStrA, 
						cCompClauseA, lpCompClauseA, 
						cCompReadStrA, lpCompReadStrA, 
						cCompReadClauseA, lpCompReadClauseA, 
						&bndCompClauseW, &readingCompClauseW );

	//
	//Make attribute run information for UNICODE string
	//
	cCompAttrW = MakeAttributeInformation( cCompStrA, tblCharOffCompStrA, 
						cCompAttrA, lpCompAttrA, 
						&bndCompAttrW, &valCompAttrW );

	//
	//Convert character position information from ANSI to UNICODE.
	//
	if (cCompStrA) {
		cursorPosW = tblCharOffCompStrA[cursorPosA];
		visiblePosW = cursorPosW;
	}

	delete [] lpCompStrA;
	delete [] lpCompAttrA;
	delete [] lpCompClauseA;
	delete [] lpCompReadStrA;
	delete [] lpCompReadClauseA;
	delete [] tblCharOffCompStrA;

    } 


    if (cResultStrW && cCompStrW) { 
	//Both Result and Composing String (partial commit)

	//Concatenate String
	int cMergedStrW = cResultStrW+cCompStrW;
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
        jstring jsMergedStrW = (jstring)JNU_CallMethodByName(env, NULL, jsResultStrW, "concat", 
                                           "(Ljava/lang/String;)Ljava/lang/String;", jsCompStrW).l;
        ASSERT(!env->ExceptionOccurred());
        ASSERT(jsMergedStrW != NULL);

	//Concatenate Clause Information
	int	cMergedClauseW = cResultClauseW+cCompClauseW;
	int*	bndMergedClauseW = new int[cMergedClauseW+1];
	jstring* readingMergedClauseW = new jstring[cMergedClauseW];
        int     i = 0;

	if (cResultClauseW && bndResultClauseW && readingResultClauseW) {
            for (; i < cResultClauseW; i++) {
		bndMergedClauseW[i] = bndResultClauseW[i];
		readingMergedClauseW[i] = readingResultClauseW[i];
	    }
	}
        if (cCompClauseW && bndCompClauseW && readingCompClauseW) {
            for (int j = 0; j < cCompClauseW; j++, i++) {
		bndMergedClauseW[i] = bndCompClauseW[j]+cResultStrW;
		readingMergedClauseW[i] = readingCompClauseW[j];
	    }
	}
	bndMergedClauseW[cMergedClauseW] = cMergedStrW;

	//Concatenate Attribute Information
	int	cMergedAttrW = 1+cCompAttrW;
	int*	bndMergedAttrW = new int[cMergedAttrW+1];
	BYTE*	valMergedAttrW = new BYTE[cMergedAttrW];
	bndMergedAttrW[0] = 0;
	valMergedAttrW[0] = ATTR_CONVERTED;
	for (int j = 0; j < cCompAttrW; j++) {
	    bndMergedAttrW[j+1] = bndCompAttrW[j]+cResultStrW;
	    valMergedAttrW[j+1] = valCompAttrW[j];
	}
	bndMergedAttrW[cMergedAttrW] = cMergedStrW;

        SendInputMethodEvent(java_awt_event_InputMethodEvent_INPUT_METHOD_TEXT_CHANGED, 
                             jsMergedStrW, 
 			     cMergedClauseW, bndMergedClauseW, readingMergedClauseW,
			     cMergedAttrW, bndMergedAttrW, valMergedAttrW,
                             cResultStrW, cursorPosW, visiblePosW);

	delete [] bndMergedClauseW;
	delete [] readingMergedClauseW;
    } 
    else if (cResultStrW) {	
	//Result String Only
        SendInputMethodEvent(java_awt_event_InputMethodEvent_INPUT_METHOD_TEXT_CHANGED, 
                             jsResultStrW,
			     cResultClauseW, bndResultClauseW, readingResultClauseW,
			     0, NULL, NULL,
                             cResultStrW, 
                             cursorPosW, visiblePosW);
    } 
    else if (cCompStrW) {		
	//Composing String Only
        SendInputMethodEvent(java_awt_event_InputMethodEvent_INPUT_METHOD_TEXT_CHANGED, 
                             jsCompStrW,
                             cCompClauseW, bndCompClauseW, readingCompClauseW,
			     cCompAttrW, bndCompAttrW, valCompAttrW,
                             0, cursorPosW, visiblePosW);
    } 


    if (bndResultClauseW)		delete [] bndResultClauseW;
    if (readingResultClauseW)	        delete [] readingResultClauseW;

    if (bndCompClauseW)			delete [] bndCompClauseW;
    if (readingCompClauseW)		delete [] readingCompClauseW;
    if (bndCompAttrW)			delete [] bndCompAttrW;
    if (valCompAttrW)			delete [] valCompAttrW;


    return mrConsume;
}

 
//
//Retrieve information from InputContext
//
//	Array is allocated for *lplpData
//	These array must be deleted by caller.
//
int AwtComponent::GetContextData(HIMC hIMC, DWORD dwIndex, LPVOID* lplpData)
{
	int cbData = ImmGetCompositionString(hIMC, dwIndex, NULL, 0);
	LPBYTE lpData = new BYTE[cbData];
	ImmGetCompositionString(hIMC, dwIndex, lpData, cbData);
	*lplpData = (LPVOID)lpData;
	return cbData;
}


//
//Convert DBCS string to Java string
//	NewStringUTF() cannot be used for DBCS string
//
jstring AwtComponent::MakeJavaString( LPSTR lpStrA, int cStrA )
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (lpStrA==NULL)   cStrA=0;

    int cStrW = MultiByteToWideChar(GetCodePage(), 0, lpStrA, cStrA, NULL, 0);
    LPWSTR lpStrW = new WCHAR[cStrW];
    MultiByteToWideChar(GetCodePage(), 0, lpStrA, cStrA, lpStrW, cStrW);

    jstring jstr = env->NewString(lpStrW, cStrW);
    delete [] lpStrW;
    return jstr;
}


//
//Scan DBCS string and create byte-offset to character-offset table
//	table must be deleted by caller.
//
int* AwtComponent::MakeCharOffsetTable( LPSTR lpStrA, int cStrA )
{
	int* table = new int[cStrA+1];

	int offChar=0;
	for ( int offByte=0; offByte<cStrA; offByte++,offChar++ ) {
		table[offByte] = offChar;
		if ( IsDBCSLeadByteEx(GetCodePage(), lpStrA[offByte]) ) {
			if (++offByte==cStrA)	break;
			table[offByte] = offChar;
		}
	}
	table[cStrA] = offChar;

	return table;
}

//
//Convert Clause and Reading Information for DBCS string to that for Unicode string
//	*lpBndClauseW and *lpReadingClauseW  must be deleted by caller.
//
int AwtComponent::MakeClauseInformation( int cStrA, int *tblCharOffStrA, 
					    int cClauseA, DWORD *lpClauseA, 
					    int cReadStrA, LPSTR lpReadStrA, 
					    int cReadClauseA, DWORD *lpReadClauseA, 
					    int** lpBndClauseW, jstring** lpReadingClauseW )
{

    if ( cStrA==0 || cClauseA==0 || cClauseA!=cReadClauseA 
	            || lpClauseA[0]!=0 || lpClauseA[cClauseA]!=(DWORD)cStrA
	            || lpReadClauseA[0]!=0 || lpReadClauseA[cReadClauseA]!=(DWORD)cReadStrA ) 
    {
        *lpBndClauseW = NULL;
        *lpReadingClauseW = NULL;
        return 0;
    }


    int	    cClauseW;
    int*    bndClauseW;
    jstring* readingClauseW;

    //Convert ANSI string caluse information to UNICODE string clause information.
    cClauseW = cClauseA;
    bndClauseW = new int[cClauseW+1];
    readingClauseW = new jstring[cClauseW];
    for ( int cls=0; cls<cClauseW; cls++ ) 
    {
	bndClauseW[cls] = tblCharOffStrA[lpClauseA[cls]];

	if ( lpReadClauseA[cls]<lpReadClauseA[cls+1] && lpReadClauseA[cls+1]<=(DWORD)cReadStrA ) 
	{
            LPSTR lpHWStrA = lpReadStrA+lpReadClauseA[cls];
            int cHWStrA = lpReadClauseA[cls+1]-lpReadClauseA[cls];
            
            if (PRIMARYLANGID(GetInputLanguage()) == LANG_JAPANESE) {
                LCID lcJPN = MAKELCID(MAKELANGID(LANG_JAPANESE,SUBLANG_DEFAULT),SORT_DEFAULT);
                // Reading string is given in half width katakana in Japanese Windows
                //  Convert it to full width katakana.
                int cFWStrA = LCMapString( lcJPN, LCMAP_FULLWIDTH, lpHWStrA, cHWStrA, NULL, 0 );
                LPSTR lpFWStrA = new CHAR[cFWStrA];
                LCMapString( lcJPN, LCMAP_FULLWIDTH, lpHWStrA, cHWStrA, lpFWStrA, cFWStrA );
                readingClauseW[cls] = MakeJavaString(lpFWStrA, cFWStrA);
                delete [] lpFWStrA;
            } else {
                readingClauseW[cls] = MakeJavaString(lpHWStrA, cHWStrA);
            }
	} else {
	    readingClauseW[cls] = MakeJavaString(NULL, 0);
	}
    }
    bndClauseW[cClauseW] = tblCharOffStrA[cStrA];

    *lpBndClauseW = bndClauseW;
    *lpReadingClauseW = readingClauseW;
    return cClauseW;

}

//
//Convert Attribute Information for DBCS string to that for Unicode string
//	*lpBndAttrW and *lpValAttrW  must be deleted by caller.
//
int AwtComponent::MakeAttributeInformation(int cStrA, int *tblCharOffStrA, 
					   int cAttrA, BYTE* lpAttrA, 
					   int** lpBndAttrW, BYTE** lpValAttrW )
{
    if ( cStrA==0 || cAttrA!=cStrA ) {
        *lpBndAttrW = NULL;
        *lpValAttrW = NULL;
        return 0;
    }


    int* bndAttrW;
    BYTE* valAttrW;
    int	  cAttrW = 0;

    //Scan attribute byte array and make attribute run information.
    bndAttrW = new int[cStrA+1];
    valAttrW = new BYTE[cStrA];
    bndAttrW[cAttrW] = 0;
    valAttrW[cAttrW] = lpAttrA[0];
    for ( int offA=1; offA<cAttrA; offA++ ) {
	    if (lpAttrA[offA]!=valAttrW[cAttrW]) {
		    cAttrW++;
		    bndAttrW[cAttrW] = tblCharOffStrA[offA];
		    valAttrW[cAttrW] = lpAttrA[offA];
	    }
    }
    cAttrW++;
    bndAttrW[cAttrW] =  tblCharOffStrA[cStrA];

    *lpBndAttrW = bndAttrW;
    *lpValAttrW = valAttrW;
    return cAttrW;
}

//
// generate and post InputMethodEvent 
//
void AwtComponent::SendInputMethodEvent(long id, jstring text,
                                        int cClause, int* rgClauseBoundary, jstring* rgClauseReading,
                                        int cAttrBlock, int* rgAttrBoundary, BYTE *rgAttrValue,
                                        int commitedTextLength, int caretPos, int visiblePos)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    // assumption for array type casting
    ASSERT(sizeof(int)==sizeof(jint));
    ASSERT(sizeof(BYTE)==sizeof(jbyte));

    // caluse information
    jintArray clauseBoundary = NULL;
    jobjectArray clauseReading = NULL; 
    if (cClause && rgClauseBoundary && rgClauseReading) {
        // convert clause boundary offset array to java array
        clauseBoundary = env->NewIntArray(cClause+1);
        env->SetIntArrayRegion(clauseBoundary, 0, cClause+1, (jint *)rgClauseBoundary);
        ASSERT(!env->ExceptionOccurred());

        // convert clause reading string array to java array
        clauseReading = env->NewObjectArray(cClause, JNU_ClassString(env), NULL); 
        for (int i=0; i<cClause; i++)   env->SetObjectArrayElement(clauseReading, i, rgClauseReading[i]);
        ASSERT(!env->ExceptionOccurred());
    }


    // attrubute value definition in WInputMethod.java must be equal to that in IMM.H
    ASSERT(ATTR_INPUT==sun_awt_windows_WInputMethod_ATTR_INPUT);
    ASSERT(ATTR_TARGET_CONVERTED==sun_awt_windows_WInputMethod_ATTR_TARGET_CONVERTED);
    ASSERT(ATTR_CONVERTED==sun_awt_windows_WInputMethod_ATTR_CONVERTED);
    ASSERT(ATTR_TARGET_NOTCONVERTED==sun_awt_windows_WInputMethod_ATTR_TARGET_NOTCONVERTED);
    ASSERT(ATTR_INPUT_ERROR==sun_awt_windows_WInputMethod_ATTR_INPUT_ERROR);

    // attribute information
    jintArray attrBoundary = NULL;
    jbyteArray attrValue = NULL;
    if (cAttrBlock && rgAttrBoundary && rgAttrValue) {
        // convert attribute boundary offset array to java array
        attrBoundary = env->NewIntArray(cAttrBlock+1);
        env->SetIntArrayRegion(attrBoundary, 0, cAttrBlock+1, (jint *)rgAttrBoundary);
        ASSERT(!env->ExceptionOccurred());

        // convert attribute value byte array to java array
        attrValue = env->NewByteArray(cAttrBlock);
        env->SetByteArrayRegion(attrValue, 0, cAttrBlock, (jbyte *)rgAttrValue);
        ASSERT(!env->ExceptionOccurred());
    }


    // get global reference of WInputMethod class (run only once)
    static jclass wInputMethodCls = NULL;
    if (wInputMethodCls == NULL) {
	jclass wInputMethodClsLocal = env->FindClass("sun/awt/windows/WInputMethod");
	ASSERT(wInputMethodClsLocal);
	if (wInputMethodClsLocal == NULL) {
	    /* exception already thrown */
	    return;
	}
	wInputMethodCls = (jclass)env->NewGlobalRef(wInputMethodClsLocal);
	env->DeleteLocalRef(wInputMethodClsLocal);
    }

    // get method ID of sendInputMethodEvent() (run only once)
    static jmethodID sendIMEventMid = 0;
    if (sendIMEventMid == 0) {
	sendIMEventMid =  env->GetMethodID(wInputMethodCls, "sendInputMethodEvent", 
					   "(ILjava/lang/String;[I[Ljava/lang/String;[I[BIII)V");
	ASSERT(sendIMEventMid);
    }

    // call m_InputMethod.sendInputMethod()
    env->CallVoidMethod(m_InputMethod, sendIMEventMid, 
                        id, text, clauseBoundary, clauseReading, attrBoundary, attrValue,
                        commitedTextLength, caretPos, visiblePos);
    if (env->ExceptionOccurred())   env->ExceptionDescribe();
    ASSERT(!env->ExceptionOccurred());

}


//
// generate and post InputMethodEvent 
//
void AwtComponent::InquireCandidatePosition(LPPOINT lpPt)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    // get global reference of WInputMethod class (run only once)
    static jclass wInputMethodCls = NULL;
    if (wInputMethodCls == NULL) {
	jclass wInputMethodClsLocal = env->FindClass("sun/awt/windows/WInputMethod");
	ASSERT(wInputMethodClsLocal);
	if (wInputMethodClsLocal == NULL) {
	    /* exception already thrown */
	    return;
	}
	wInputMethodCls = (jclass)env->NewGlobalRef(wInputMethodClsLocal);
	env->DeleteLocalRef(wInputMethodClsLocal);
    }

    // get method ID of sendInputMethodEvent() (run only once)
    static jmethodID inqCandPosMid = 0;
    if (inqCandPosMid == 0) {
	inqCandPosMid =  env->GetMethodID(wInputMethodCls, "inquireCandidatePosition", 
					   "()Ljava/awt/Point;");
        ASSERT(!env->ExceptionOccurred());
	ASSERT(inqCandPosMid);
    }

    // call m_InputMethod.sendInputMethod()
    jobject candPos = env->CallObjectMethod(m_InputMethod, inqCandPosMid);
    ASSERT(!env->ExceptionOccurred());


    // get global reference of Point class (run only once)
    static jclass pointCls = NULL;
    if (pointCls == NULL) {
	jclass pointClsLocal = env->FindClass("java/awt/Point");
	ASSERT(pointClsLocal);
	if (pointClsLocal == NULL) {
	    /* exception already thrown */
	    return;
	}
	pointCls = (jclass)env->NewGlobalRef(pointClsLocal);
	env->DeleteLocalRef(pointClsLocal);
    }

    // get field ID of x and y (run only once)
    static jfieldID xFid = 0;
    if (xFid == 0) {
	xFid =  env->GetFieldID(pointCls, "x", "I");
        ASSERT(!env->ExceptionOccurred());
	ASSERT(xFid);
    }
    static jfieldID yFid = 0;
    if (yFid == 0) {
	yFid =  env->GetFieldID(pointCls, "y", "I");
        ASSERT(!env->ExceptionOccurred());
	ASSERT(yFid);
    }

    // get x and y value of Point object
    lpPt->x = env->GetIntField(candPos, xFid);
    lpPt->y = env->GetIntField(candPos, yFid);
    ASSERT(!env->ExceptionOccurred());

}

MsgRouting AwtComponent::WmCommand(UINT id, HWND hWndChild, UINT notifyCode)
{
    /* Menu/Accelerator */
    if (hWndChild == 0) {
        AwtObject* obj = AwtToolkit::GetInstance().LookupCmdID(id);
        ASSERT(obj);
	ASSERT(((AwtMenuItem*)obj)->GetID() == id);
        obj->DoCommand();
        return mrConsume;
    }
    /* Child id notification */
    else {
        AwtComponent* child = AwtComponent::GetComponent(hWndChild);
        if (child) {
            child->WmNotify(notifyCode);
        }
    }
    return mrDoDefault;
}

MsgRouting AwtComponent::WmNotify(UINT notifyCode)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmCompareItem(UINT ctrlId, 
                                       COMPAREITEMSTRUCT far& compareInfo, 
                                       long& result)
{
    AwtComponent* child = AwtComponent::GetComponent(compareInfo.hwndItem);
    if (child == this) {
        /* DoCallback("handleItemDelete", */
    }
    else if (child) {
        return child->WmCompareItem(ctrlId, compareInfo, result);
    }
    return mrConsume;
}

MsgRouting AwtComponent::WmDeleteItem(UINT ctrlId, 
				      DELETEITEMSTRUCT far& deleteInfo)
{
    /* 
     * Workaround for NT 4.0 bug -- if SetWindowPos is called on a AwtList
     * window, a WM_DELETEITEM message is sent to its parent with a window
     * handle of one of the list's child windows.  The property lookup
     * succeeds, but the HWNDs don't match. 
     */
    if (deleteInfo.hwndItem == NULL) {
        return mrConsume;
    }
    AwtComponent* child = (AwtComponent*)::GetProp(deleteInfo.hwndItem, 
                                                   ThisProperty());
    if (child && child->GetHWnd() != deleteInfo.hwndItem) {
        return mrConsume;
    }

    if (child == this) {
        /*DoCallback("handleItemDelete", */
    }
    else if (child) {
        return child->WmDeleteItem(ctrlId, deleteInfo);
    }
    return mrConsume;
}
                            
MsgRouting AwtComponent::WmDrawItem(UINT ctrlId, DRAWITEMSTRUCT far& drawInfo)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (drawInfo.CtlType == ODT_MENU) {
        if (drawInfo.itemData != 0) {
            AwtMenu* menu = (AwtMenu*)(drawInfo.itemData);
            menu->DrawItem(drawInfo);
        }
    } else {
        AwtComponent* child = AwtComponent::GetComponent(drawInfo.hwndItem);
        if (child == this) {
	    /* DoCallback("handleItemDelete", */
        } else if (child != NULL) {
            return child->WmDrawItem(ctrlId, drawInfo);
        }
    }
    return mrConsume;
}
        
MsgRouting AwtComponent::WmMeasureItem(UINT ctrlId, 
				       MEASUREITEMSTRUCT far& measureInfo)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (measureInfo.CtlType == ODT_MENU) {
        if (measureInfo.itemData != 0) {
            AwtMenu* menu = (AwtMenu*)(measureInfo.itemData);
            HDC hDC = ::GetDC(GetHWnd());
            /* menu->MeasureItem(env, hDC, measureInfo); */
            menu->MeasureItem(hDC, measureInfo);
            ::ReleaseDC(GetHWnd(), hDC);
        }
    } else {
        HWND  hChild = ::GetDlgItem(GetHWnd(), measureInfo.CtlID);
        AwtComponent* child = AwtComponent::GetComponent(hChild);
	/*
	 * If the parent cannot find the child's instance from its handle,
	 * maybe the child is in its creation.  So the child must be searched 
	 * from the list linked before the child's creation.
	 */
        if (child == NULL)
            child = SearchChild((UINT)ctrlId);
        
        if (child == this) {
            /* DoCallback("handleItemDelete",  */
        }
        else if (child)
            return child->WmMeasureItem(ctrlId, measureInfo);
    }
    return mrConsume;
}

/* for WmDrawItem method of Label, Button and Checkbox */
void AwtComponent::DrawWindowText(HDC hDC, jobject font, jstring text,
				  int x, int y)
{
    int nOldBkMode = ::SetBkMode(hDC,TRANSPARENT);
    ASSERT(nOldBkMode != 0);
    AwtFont::drawMFString(hDC, font, text, x, y, GetCodePage());
    VERIFY(::SetBkMode(hDC,nOldBkMode));
}

/*
 * Draw text in gray (the color being set to COLOR_GRAYTEXT) when the
 * component is disabled.  Used only for label, checkbox and button in
 * OWNER_DRAW.  It draws the text in emboss.  
 */
void AwtComponent::DrawGrayText(HDC hDC, jobject font, jstring text,
				int x, int y)
{
    ::SetTextColor(hDC, ::GetSysColor(COLOR_BTNHILIGHT));
    AwtComponent::DrawWindowText(hDC, font, text, x+1, y+1);
    ::SetTextColor(hDC, ::GetSysColor(COLOR_BTNSHADOW));
    AwtComponent::DrawWindowText(hDC, font, text, x, y);
}

/* for WmMeasureItem method of List and Choice */
jstring AwtComponent::GetItemString(JNIEnv *env, jobject target, long index)
{
    jstring str = (jstring)JNU_CallMethodByName(env, NULL, target, "getItemImpl",
						"(I)Ljava/lang/String;",
						index).l;
    ASSERT(!env->ExceptionOccurred());
    return str;
}

/* for WmMeasureItem method of List and Choice */
void AwtComponent::MeasureListItem(JNIEnv *env, 
				   MEASUREITEMSTRUCT far& measureInfo)
{
    if (env->EnsureLocalCapacity(1) < 0) {
	return;
    }
    jobject dimension = PreferredItemSize(env);
    ASSERT(dimension);
    measureInfo.itemWidth =
      env->GetIntField(dimension, AwtDimension::widthID);
    measureInfo.itemHeight =
      env->GetIntField(dimension, AwtDimension::heightID);
    env->DeleteLocalRef(dimension);
}

/* for WmDrawItem method of List and Choice */
void AwtComponent::DrawListItem(JNIEnv *env, DRAWITEMSTRUCT far& drawInfo)
{
    if (env->EnsureLocalCapacity(3) < 0) {
	return;
    }
    jobject peer = GetPeer(env);
    jobject target = env->GetObjectField(peer, AwtObject::targetID);

    HDC hDC = drawInfo.hDC;
    RECT rect = drawInfo.rcItem;
    
    BOOL bEnabled = isEnabled();

    DWORD crBack, crText;
    if (drawInfo.itemState & ODS_SELECTED) {
	/* Set background and text colors for selected item */
	crBack = ::GetSysColor (COLOR_HIGHLIGHT);
	crText = ::GetSysColor (COLOR_HIGHLIGHTTEXT);
    } else {
	/* Set background and text colors for unselected item */
        crBack = GetBackgroundColor();
        crText = bEnabled ? GetColor() : ::GetSysColor(COLOR_GRAYTEXT);
    }
    /* Fill item rectangle with background color */
    HBRUSH hbrBack = ::CreateSolidBrush (crBack);
    ASSERT(hbrBack);
    VERIFY(::FillRect (hDC, &rect, hbrBack));
    VERIFY(::DeleteObject (hbrBack));
    
    /* Set current background and text colors */
    ::SetBkColor (hDC, crBack);
    ::SetTextColor (hDC, crText);
    
    /*draw string (with left mergin of 1 point) */
    if ((int) (drawInfo.itemID) >= 0) {
	    jobject font = GET_FONT(target, peer);
	    jstring text = GetItemString(env, target, drawInfo.itemID);
	    SIZE size = AwtFont::getMFStringSize(hDC, font, text);
	    AwtFont::drawMFString(hDC, font, text,
				  (GetRTL()) ? rect.right - size.cx - 1 
				             : rect.left + 1,
				  (rect.top + rect.bottom - size.cy) / 2,
				  GetCodePage());
	    env->DeleteLocalRef(font);
	    env->DeleteLocalRef(text);
    }
    if ((drawInfo.itemState & ODS_FOCUS)  &&
        (drawInfo.itemAction & (ODA_FOCUS | ODA_DRAWENTIRE))) {
        VERIFY(::DrawFocusRect(hDC, &rect));
    }
    env->DeleteLocalRef(target);
}

/* for MeasureListItem method and WmDrawItem method of Checkbox */
jint AwtComponent::GetFontHeight(JNIEnv *env)
{
    if (env->EnsureLocalCapacity(4) < 0) {
	return NULL;
    }
    jobject self = GetPeer(env);
    jobject target = env->GetObjectField(self, AwtObject::targetID);

    jobject font = GET_FONT(target, self);
    jobject toolkit = env->CallObjectMethod(target, 
					    AwtComponent::getToolkitMID);

    ASSERT(!env->ExceptionOccurred());

    jobject fontMetrics = 
	env->CallObjectMethod(toolkit, AwtToolkit::getFontMetricsMID, font);

    ASSERT(!env->ExceptionOccurred());
    
    jint height = env->CallIntMethod(fontMetrics, AwtFont::getHeightMID);
    ASSERT(!env->ExceptionOccurred());

    env->DeleteLocalRef(target);
    env->DeleteLocalRef(font);
    env->DeleteLocalRef(toolkit);
    env->DeleteLocalRef(fontMetrics);

    return height;
}

/*
 * Forward the print request to the Graphics object.  It's necessary
 * to bounce the print request from WGraphics_print to here and back
 * to avoid race conditions when sending print messages, by running on 
 * the creating thread.
 */
MsgRouting AwtComponent::PrintComponent(AwtGraphics* g)
{
    g->PrintComponent(this);
    return mrConsume;
}

// If you override WmPrint, make sure to save a copy of the DC on the GDI
// stack to be restored in WmPrintClient. Windows mangles the DC in
// ::DefWindowProc.
MsgRouting AwtComponent::WmPrint(HDC hDC, UINT flags)
{
    // Save two copies of the DC: one for WmPrintClient and one for
    // us to use after DefWindowProc returns
    ::SaveDC(hDC);
    ::SaveDC(hDC);

    DefWindowProc(WM_PRINT, (WPARAM) hDC, (LPARAM) flags);

    ::RestoreDC(hDC, -1);

    // Special case for components with a sunken border. Windows does not
    // print the border correctly, so we have to do it ourselves.
    if (IS_WIN4X && (::GetWindowLong(GetHWnd(), GWL_EXSTYLE) &
		     WS_EX_CLIENTEDGE)) {
        RECT r;
	::GetWindowRect(GetHWnd(), &r);
	::OffsetRect(&r, -r.left, -r.top);
	::DrawEdge(hDC, &r, EDGE_SUNKEN, BF_RECT);
    }

    return mrConsume;
}

// If you override WmPrintClient, maker sure to obtain a valid copy of
// the DC from the GDI stack. The copy of the DC should have been placed
// there by WmPrint. Windows mangles the DC in ::DefWindowProc.
MsgRouting AwtComponent::WmPrintClient(HDC hDC, UINT)
{
    // obtain valid DC from GDI stack
    ::RestoreDC(hDC, -1);

    return mrDoDefault;
}

MsgRouting AwtComponent::WmNcCalcSize(BOOL fCalcValidRects, 
				      LPNCCALCSIZE_PARAMS lpncsp, UINT& retVal)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmNcPaint(HRGN hrgn)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmNcHitTest(UINT x, UINT y, UINT& retVal)
{
    return mrDoDefault;
}

MsgRouting AwtComponent::WmQueryNewPalette(UINT& retVal)
{
    if (AwtImage::m_hPal != NULL) {
        GetSharedDC()->Lock();
        GetSharedDC()->GrabDC(NULL);
        HDC hDC = GetSharedDC()->GetDC();
        ASSERT(hDC);
        ::SelectPalette(hDC, AwtImage::m_hPal, FALSE);
        ::RealizePalette(hDC);
        GetSharedDC()->Unlock();
	// Do not invalidate here if the palette
	// has not changed we will get an extra repaint
        retVal = TRUE;
    } else {
        retVal = FALSE;
    }

    return mrDoDefault;
}

MsgRouting AwtComponent::WmPaletteChanged(HWND hwndPalChg)
{
    // Netscape : We need to re-realize our palette here (unless we're the one
    // that was realizing it in the first place).  That will let us match the
    // remaining colors in the system palette as best we can.  We always
    // invalidate because the palette will have changed when we recieve this
    // message.

    if (hwndPalChg == GetHWnd()) {
        GetSharedDC()->Lock();
        GetSharedDC()->GrabDC(NULL);
        HDC hDC = GetSharedDC()->GetDC();
        ASSERT(hDC);
        ASSERT(::RealizePalette(hDC) != GDI_ERROR);
        GetSharedDC()->Unlock();
    }
    AwtToolkit::GetInstance().adjustPalette(NULL);
    Invalidate(NULL);
    return mrDoDefault;
}

MsgRouting AwtComponent::WmStyleChanged(WORD wStyleType, LPSTYLESTRUCT lpss)
{
    ASSERT(!IsBadReadPtr(lpss, sizeof(STYLESTRUCT)));
    return mrDoDefault;
}

MsgRouting AwtComponent::WmSettingChange(WORD wFlag, LPCTSTR pszSection)
{
    ASSERT(!IsBadStringPtr(pszSection, 20));
    AWT_TRACE(("WM_SETTINGCHANGE: wFlag=%d pszSection=%s\n", (int)wFlag, pszSection));
    return mrDoDefault;
}

AwtComponent* AwtComponent::SearchChild(UINT id) {
    ChildListItem* child;
    for (child = m_childList; child != NULL;child = child->m_next) {
	if (child->m_ID == id)
	    return child->m_Component;
    }
    /* 
     * ASSERT(FALSE);
     * This should not be happend if all children are recorded
     */
    return NULL;	/* make compiler happy */
}

void AwtComponent::RemoveChild(UINT id) {
    ChildListItem* child = m_childList;
    ChildListItem* lastChild = NULL;
    while (child != NULL) {
	if (child->m_ID == id) {
	    if (lastChild == NULL) {
		m_childList = child->m_next;
	    } else {
		lastChild->m_next = child->m_next;
	    }
	    child->m_next = NULL;
	    ASSERT(child != NULL);
	    delete child;
	    return;
	}
	lastChild = child;
	child = child->m_next;
    }
}

void AwtComponent::SendKeyEvent(long id, jlong when, long raw, long cooked,
				long modifiers, MSG* msg)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    CriticalSection::Lock l(GetLock());
    if (GetPeer(env) == NULL) {
        /* event received during termination. */
        return;
    }

    static jclass keyEventCls;
    if (keyEventCls == NULL) {
	jclass keyEventClsLocal = env->FindClass("java/awt/event/KeyEvent");
	ASSERT(keyEventClsLocal);
	if (keyEventClsLocal == NULL) {
	    /* exception already thrown */
	    return;
	}
	keyEventCls = (jclass)env->NewGlobalRef(keyEventClsLocal);
	env->DeleteLocalRef(keyEventClsLocal);
    }

    static jmethodID keyEventConst;
    if (keyEventConst == NULL) {
	keyEventConst =  env->GetMethodID(keyEventCls, "<init>", 
					  "(Ljava/awt/Component;IJIIC)V");
	ASSERT(keyEventConst);
    }
    if (env->EnsureLocalCapacity(2) < 0) {
	return;
    }
    jobject target = GetTarget(env);
    jobject keyEvent = env->NewObject(keyEventCls, keyEventConst, target,
				      id, when, modifiers, raw, cooked);
    if (env->ExceptionOccurred()) env->ExceptionDescribe();
    ASSERT(!env->ExceptionOccurred());
    ASSERT(keyEvent != NULL);
    static jfieldID keyDataID;
    if (keyDataID == NULL) {
        keyDataID = env->GetFieldID(keyEventCls, "data", "J");
    }
    env->SetLongField(keyEvent, keyDataID, (long)msg);
    SendEvent(keyEvent);

    env->DeleteLocalRef(keyEvent);
    env->DeleteLocalRef(target);
}

void AwtComponent::SendMouseEvent(long id, jlong when, long x, long y, 
                                  long modifiers, long clickCount, 
                                  long popupTrigger,
                                  MSG* msg)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    CriticalSection::Lock l(GetLock());
    if (GetPeer(env) == NULL) {
        /* event received during termination. */
        return;
    }

    static jclass mouseEventCls;
    if (mouseEventCls == NULL) {
	jclass mouseEventClsLocal = 
	    env->FindClass("java/awt/event/MouseEvent");
	if (!mouseEventClsLocal) {
	    /* exception already thrown */
	    return;
	}
	mouseEventCls = (jclass)env->NewGlobalRef(mouseEventClsLocal);
	env->DeleteLocalRef(mouseEventClsLocal);
    }
    RECT insets;
    GetInsets(&insets);

    static jmethodID mouseEventConst;
    if (mouseEventConst == NULL) {
	mouseEventConst = 
	    env->GetMethodID(mouseEventCls, "<init>", 
			     "(Ljava/awt/Component;IJIIIIZ)V");
	ASSERT(mouseEventConst);
    }
    if (env->EnsureLocalCapacity(2) < 0) {
	return;
    }
    jobject target = GetTarget(env);
    jobject mouseEvent = env->NewObject(mouseEventCls, mouseEventConst, 
					target,
					id, when, modifiers, 
					x+insets.left, y+insets.top, 
					clickCount, popupTrigger);
    ASSERT(!env->ExceptionOccurred());
    ASSERT(mouseEvent != NULL);
    static jfieldID mouseDataID;
    if (mouseDataID == NULL) {
	mouseDataID = env->GetFieldID(mouseEventCls, "data", "J");
	ASSERT(mouseDataID);
    }
    env->SetLongField(mouseEvent, mouseDataID,  (long)msg);
    SendEvent(mouseEvent);

    env->DeleteLocalRef(mouseEvent);
    env->DeleteLocalRef(target);
}

void AwtComponent::SendFocusEvent(long id, BOOL temporary)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    CriticalSection::Lock l(GetLock());
    if (GetPeer(env) == NULL) {
        /* event received during termination. */
        return;
    }

    static jclass focusEventCls;
    if (focusEventCls == NULL) {
	jclass focusEventClsLocal 
	    = env->FindClass("java/awt/event/FocusEvent");
	ASSERT(focusEventClsLocal);
	if (focusEventClsLocal == NULL) {
	    /* exception already thrown */
	    return;
	}
	focusEventCls = (jclass)env->NewGlobalRef(focusEventClsLocal);
	env->DeleteLocalRef(focusEventClsLocal);
    }

    static jmethodID focusEventConst;
    if (focusEventConst == NULL) {
	focusEventConst = env->GetMethodID(focusEventCls, "<init>", 
					   "(Ljava/awt/Component;IZ)V");
	ASSERT(focusEventConst);
    }
    if (env->EnsureLocalCapacity(2) < 0) {
	return;
    }
    jobject target = GetTarget(env);
    jobject focusEvent = env->NewObject(focusEventCls, focusEventConst, 
					target, id, (long)temporary);
    ASSERT(!env->ExceptionOccurred());
    ASSERT(focusEvent != NULL);
    static jfieldID focusDataID;
    if (focusDataID == NULL) {
	focusDataID = env->GetFieldID(focusEventCls, "data", "J");
    }
    SendEvent(focusEvent);

    env->DeleteLocalRef(focusEvent);
    env->DeleteLocalRef(target);
}

/*
 * Forward a filtered event directly to the subclassed window.
 * This method is needed so that DefWindowProc is invoked on the 
 * component's owning thread.
 */
MsgRouting AwtComponent::HandleEvent(MSG *msg, BOOL)
{
    DefWindowProc(msg->message, msg->wParam, msg->lParam);
    delete msg;
    return mrConsume;
}

void AwtComponent::SynthesizeKeyMessage(JNIEnv *env, jobject keyEvent)
{
    jint id = (env)->GetIntField(keyEvent, AwtAWTEvent::idID);
    UINT message;
    switch (id) {
      case java_awt_event_KeyEvent_KEY_PRESSED:
	  message = WM_KEYDOWN; 
	  break;
      case java_awt_event_KeyEvent_KEY_RELEASED:
	  message = WM_KEYUP; 
	  break;
      case java_awt_event_KeyEvent_KEY_TYPED:
	  message = WM_CHAR; 
	  break;
      default:
	  return;
    }

    /*
     * KeyEvent.modifiers aren't supported -- the Java apppwd must send separate
     * KEY_PRESSED and KEY_RELEASED events for the modifier virtual keys.
     */
    if (id == java_awt_event_KeyEvent_KEY_TYPED) {
        // WM_CHAR message must be posted using WM_AWT_FORWARD_CHAR (for Edit control)
	jchar keyChar = (jchar)
	  (env)->GetCharField(keyEvent, AwtKeyEvent::keyCharID);
        ::PostMessage(GetHWnd(), WM_AWT_FORWARD_CHAR,
		      MAKEWPARAM(keyChar, TRUE), 0);
    } else {
	jint keyCode =
	  (env)->GetIntField(keyEvent, AwtKeyEvent::keyCodeID);
	UINT key, modifiers;
        AwtComponent::JavaKeyToWindowsKey(keyCode, &key, &modifiers);
        MSG* msg = CreateMessage(message, key, 0);
	::PostMessage(GetHWnd(), WM_AWT_HANDLE_EVENT, (WPARAM) TRUE,
		      (LPARAM) msg);
    }
}

void AwtComponent::SynthesizeMouseMessage(JNIEnv *env, jobject mouseEvent)
{
    enum Buttons { left, middle, right } button;
    /*    DebugBreak(); */
    jint modifiers =
      (env)->GetIntField(mouseEvent, AwtInputEvent::modifiersID);

    if (modifiers & java_awt_event_InputEvent_META_MASK > 0) {
	button = right;
    } else if (modifiers & java_awt_event_InputEvent_ALT_MASK > 0) {
	button = middle;
    } else {
	button = left;
    }
    jint id = (env)->GetIntField(mouseEvent, AwtAWTEvent::idID);
    UINT message;
    switch (id) {
      case java_awt_event_MouseEvent_MOUSE_PRESSED: {
	  switch (button) {
            case left:   message = WM_LBUTTONDOWN; break;
            case middle: message = WM_MBUTTONDOWN; break;
            case right:  message = WM_RBUTTONDOWN; break;
	  }
	  break;
      }
      case java_awt_event_MouseEvent_MOUSE_RELEASED: {
	  switch (button) {
            case left:   message = WM_LBUTTONUP; break;
            case middle: message = WM_MBUTTONUP; break;
            case right:  message = WM_RBUTTONUP; break;
	  }
	  break;
      }
      case java_awt_event_MouseEvent_MOUSE_MOVED:
	  /* MOUSE_DRAGGED events must first have sent a MOUSE_PRESSED event. */
      case java_awt_event_MouseEvent_MOUSE_DRAGGED:
	  message = WM_MOUSEMOVE;
	  break;
      default:
	  return;
    }
    jint x = (env)->GetIntField(mouseEvent, AwtMouseEvent::xID);
    jint y = (env)->GetIntField(mouseEvent, AwtMouseEvent::yID);
    MSG* msg = CreateMessage(message, 0, MAKELPARAM(x, y));
    ::PostMessage(GetHWnd(), WM_AWT_HANDLE_EVENT, (WPARAM) TRUE, (LPARAM) msg);
}

void AwtComponent::Invalidate(RECT* r)
{
    ::InvalidateRect(GetHWnd(), r, FALSE);
}

void AwtComponent::BeginValidate()
{
    ASSERT(m_validationNestCount >= 0 &&
    	   m_validationNestCount < 1000); // sanity check

    if (m_validationNestCount == 0) {
    // begin deferred window positioning if we're not inside
    // another Begin/EndValidate pair
	ASSERT(m_hdwp == NULL);
	m_hdwp = ::BeginDeferWindowPos(32);
    }

    m_validationNestCount++;
}

void AwtComponent::EndValidate()
{
    ASSERT(m_validationNestCount > 0 &&
    	   m_validationNestCount < 1000); // sanity check
    ASSERT(m_hdwp != NULL);

    m_validationNestCount--;
    if (m_validationNestCount == 0) {
    // if this call to EndValidate is not nested inside another
    // Begin/EndValidate pair, end deferred window positioning
        ::EndDeferWindowPos(m_hdwp);
        m_hdwp = NULL;
    }
}

/**
 * HWND, AwtComponent and Java Peer interaction
 */

/*
 *Link the C++, Java peer, and HWNDs together.
 */
void AwtComponent::LinkObjects(JNIEnv *env, jobject peer)
{
    /*
     * Bind all three objects together thru this C++ object, two-way to each:
     *     JavaPeer <-> C++ <-> HWND
     *
     * C++ -> JavaPeer
     */
    if (m_peerObject == NULL) {
	// This may have already been set up by CreateHWnd
	// And we don't want to create two references so we
	// will leave the prior one alone
	m_peerObject = env->NewGlobalRef(peer);
    }

    /* JavaPeer -> C++ */
    JNI_SET_PDATA(peer, (jint)this);

    /* HWND -> C++ */
    VERIFY(::SetProp(GetHWnd(), ThisProperty(), (HANDLE)this)); 
}

/* Cleanup above linking */
void AwtComponent::UnlinkObjects()
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (m_peerObject) {
	JNI_SET_PDATA(m_peerObject, NULL);
	env->DeleteGlobalRef(m_peerObject);
        m_peerObject = NULL;
    }
    if (m_hwnd) {
        VERIFY(::RemoveProp(m_hwnd, ThisProperty()) != NULL);
    }

    /* Also cleanup any after-the-fact binding */
    if (m_sharedDC) {
        AwtSharedDC* sdc = m_sharedDC;
        m_sharedDC = NULL;
        sdc->Detach(NULL);
    }
}

/*
 * associate an AwtDropTarget with this AwtComponent
 */

AwtDropTarget* AwtComponent::CreateDropTarget(JNIEnv* env) {
    m_dropTarget = new AwtDropTarget(env, this);
 
    return m_dropTarget;
}

/*
 * disassociate an AwtDropTarget with this AwtComponent
 */

void AwtComponent::DestroyDropTarget(JNIEnv* env) {
    if (m_dropTarget != NULL) {
	m_dropTarget->Release();
    }

    m_dropTarget = NULL;
}

/**
 *
 */

void AwtComponent::ScheduleMsgWorkProc(MsgWorkProc wp, void* cls) {
    AddMsgWorkProc(wp, cls);
}

/************************************************************************
 * Component native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_Component_initIDs(JNIEnv *env, jclass cls) 
{
    /* field ids */
    AwtComponent::peerID =
      env->GetFieldID(cls, "peer", "Ljava/awt/peer/ComponentPeer;");
    AwtComponent::xID = env->GetFieldID(cls, "x", "I");
    AwtComponent::yID = env->GetFieldID(cls, "y", "I");
    AwtComponent::heightID = env->GetFieldID(cls, "height", "I");
    AwtComponent::widthID = env->GetFieldID(cls, "width", "I");
    AwtComponent::visibleID = env->GetFieldID(cls, "visible", "Z");
    AwtComponent::backgroundID = 
	env->GetFieldID(cls, "background", "Ljava/awt/Color;");
    AwtComponent::foregroundID = 
	env->GetFieldID(cls, "foreground", "Ljava/awt/Color;");
    AwtComponent::enabledID = env->GetFieldID(cls, "enabled", "Z");
    AwtComponent::parentID = env->GetFieldID(cls, "parent", "Ljava/awt/Container;");

    /* method ids */
    AwtComponent::getFontMID = 
	env->GetMethodID(cls, "getFont_NoClientCode", "()Ljava/awt/Font;");
    AwtComponent::getToolkitMID = 
	env->GetMethodID(cls, "getToolkitImpl", "()Ljava/awt/Toolkit;");
    AwtComponent::isEnabledMID = env->GetMethodID(cls, "isEnabledImpl", "()Z");


    ASSERT(AwtComponent::xID);
    ASSERT(AwtComponent::yID);
    ASSERT(AwtComponent::heightID);
    ASSERT(AwtComponent::widthID);
    ASSERT(AwtComponent::visibleID);
    ASSERT(AwtComponent::backgroundID);
    ASSERT(AwtComponent::foregroundID);
    ASSERT(AwtComponent::enabledID);

    ASSERT(AwtComponent::getFontMID);
    ASSERT(AwtComponent::getToolkitMID);
    ASSERT(AwtComponent::isEnabledMID);
}

} /* extern "C" */


/************************************************************************
 * ComponentPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    show
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_show(JNIEnv *env, jobject self)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent *)pData;
    p->Show();
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    hide
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_hide(JNIEnv *env, jobject self)
{    
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent *)pData;
    p->Hide();
}
    
/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    enable
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_enable(JNIEnv *env, jobject self)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent*)pData;
    ::EnableWindow(p->GetHWnd(), TRUE);
    ::InvalidateRect(p->GetHWnd(), NULL, TRUE); // Bug #4038881 Labels don't enable and disable properly
    CriticalSection::Lock l(p->GetLock());
    p->VerifyState();
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    disable
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_disable(JNIEnv *env, jobject self)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent*)pData;
    ::EnableWindow(p->GetHWnd(), FALSE);
    ::InvalidateRect(p->GetHWnd(), NULL, TRUE); // Bug #4038881 Labels don't enable and disable properly
    CriticalSection::Lock l(p->GetLock());
    p->VerifyState();
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    getLocationOnScreen
 * Signature: ()Ljava/awt/Point;
 */
JNIEXPORT jobject JNICALL
Java_sun_awt_windows_WComponentPeer_getLocationOnScreen(JNIEnv *env, jobject self)
{
    jint pData;
    JNI_CHECK_PEER_RETURN_NULL(self);
    AwtComponent* component = (AwtComponent *)pData;

    RECT rect;
    VERIFY(::GetWindowRect(component->GetHWnd(), &rect));

    jobject point = JNU_NewObjectByName(env, "java/awt/Point", "(II)V", 
					rect.left, rect.top);
    return point;
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    reshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WComponentPeer_reshape(JNIEnv *env, jobject self,
					    jint x, jint y, jint w, jint h)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent *)pData;
    RECT* r = new RECT;
    ::SetRect(r, x, y, x + w, y + h);
    p->SendMessage(WM_AWT_RESHAPE_COMPONENT, 0, (LPARAM)r);
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    handleEvent
 * Signature: (Ljava/awt/AWTEvent;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_nativeHandleEvent(JNIEnv *env,
						      jobject self,
						      jobject event)
{
    jint pData;
    JNI_CHECK_NULL_RETURN(event, "null AWTEvent");

    jlong data = (env)->GetLongField(event, AwtAWTEvent::dataID);
    int id = (env)->GetIntField(event, AwtAWTEvent::idID);
    ASSERT(!env->ExceptionOccurred());
    if (data != 0) {
        static BOOL keyDownConsumed = FALSE;
	static BOOL bCharChanged = FALSE;
	static WCHAR modifiedChar;	
	WCHAR unicodeChar;
        MSG* pMsg = (MSG*)data;

        /* Remember if a KEY_PRESSED event is consumed, as an old model  
	 * program won't consume a subsequent KEY_TYPED event. 
	 */
        jint pData = JNI_GET_PDATA(self);
	jboolean consumed = 
	    (env)->GetBooleanField(event, AwtAWTEvent::consumedID);
	ASSERT(!env->ExceptionOccurred());

        if (consumed) {
            keyDownConsumed = (id == java_awt_event_KeyEvent_KEY_PRESSED);
            return;
        }

        /* Consume a KEY_TYPED event if a KEY_PRESSED had been, to support 
	 * the old model.
	 */
        if ((id == java_awt_event_KeyEvent_KEY_TYPED) && keyDownConsumed) {
            keyDownConsumed = FALSE;
            return;
        }

        /* Modify any event parameters, if necessary. */
        if (self && (pData = JNI_GET_PDATA(self)) && 
            id >= java_awt_event_KeyEvent_KEY_FIRST &&
            id <= java_awt_event_KeyEvent_KEY_LAST) {

            AwtComponent* p = (AwtComponent*)pData;
            CriticalSection::Lock l(p->GetLock());
	    
	    jint keyCode = 
	      (env)->GetIntField(event, AwtKeyEvent::keyCodeID);
	    jchar keyChar = 
	      (env)->GetCharField(event, AwtKeyEvent::keyCharID);
            jint modifiers =
	      (env)->GetIntField(event, AwtInputEvent::modifiersID);

	    ASSERT(!env->ExceptionOccurred());

            switch (id) {
              case java_awt_event_KeyEvent_KEY_PRESSED:
	      {

		  if (pMsg->wParam == VK_PROCESSKEY)
		  {
                      bCharChanged = FALSE;
		  } 
		  else 
		  {
		      UINT key, keyModifiers;
		      p->JavaKeyToWindowsKey(keyCode, &key, &keyModifiers);
		      if (!key) 
		      {
		          key = keyCode;
		      }
		      modifiedChar = p->WindowsKeyToJavaChar(key, modifiers);
		      bCharChanged = (keyChar != modifiedChar);
		  }
	      }

                  break;
              case java_awt_event_KeyEvent_KEY_RELEASED:
                  keyDownConsumed = FALSE;
	 	  bCharChanged = FALSE;
		  break;
              case java_awt_event_KeyEvent_KEY_TYPED: 
              {
		  if (bCharChanged) 
		  {
		      unicodeChar = modifiedChar;
		  } 
		  else
		  {
		      unicodeChar = keyChar;
		  }
		  bCharChanged = FALSE;

                  // send the character back to the native window for 
		  // processing. The WM_AWT_FORWARD_CHAR handler will send
		  // this character to DefWindowProc
                  ::PostMessage(p->GetHWnd(), WM_AWT_FORWARD_CHAR, 
				 MAKEWPARAM(unicodeChar, FALSE), pMsg->lParam);
                  return;
              }
                  break;
              default:
                  break;
            }

            // re-post PRE key event messages
            #ifdef TWO_PASS_DESIGN
	    switch(pMsg->message){
	    case WM_AWT_PRE_KEYDOWN:
	    case WM_AWT_PRE_KEYUP:
	    case WM_AWT_PRE_SYSKEYDOWN:
	    case WM_AWT_PRE_SYSKEYUP:
		    ::PostMessage(p->GetHWnd(), pMsg->message, pMsg->wParam,
				  pMsg->lParam);
		    return;
	    }
            #endif
        }

        // ignore all InputMethodEvents
        if (self && (pData = JNI_GET_PDATA(self)) && 
            id >= java_awt_event_InputMethodEvent_INPUT_METHOD_FIRST &&
            id <= java_awt_event_InputMethodEvent_INPUT_METHOD_LAST) {
            return;
        }

        /* Post the message directly to the subclassed component. */
        if (self && (pData = JNI_GET_PDATA(self))) {
            AwtComponent* p = (AwtComponent*)pData;
            ::PostMessage(p->GetHWnd(), WM_AWT_HANDLE_EVENT, (WPARAM) FALSE,
			  (LPARAM)pMsg);
	    // The handler for WM_AWT_HANDLE_EVENT deletes the pMsg, so we
	    // don't want to do it in freeNativeData, thus we set data to null
	    (env)->SetLongField(event, AwtAWTEvent::dataID, 0);
        }
        return;
    }
    
    /* Forward any valid synthesized events.  Currently only mouse and
     * key events are supported.
     */
    if (self == NULL || (pData = JNI_GET_PDATA(self)) == NULL) {
	return;
    }
    
    AwtComponent* p = (AwtComponent*)pData;
    if (id >= java_awt_event_KeyEvent_KEY_FIRST &&
        id <= java_awt_event_KeyEvent_KEY_LAST) {
	p->SynthesizeKeyMessage(env, event);

    } else if (id >= java_awt_event_MouseEvent_MOUSE_FIRST &&
	       id <= java_awt_event_MouseEvent_MOUSE_LAST) {
	p->SynthesizeMouseMessage(env, event);
    }
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    _dispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WComponentPeer__1dispose(JNIEnv *env, jobject self)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent*)pData;
    AwtToolkit::GetInstance().SendMessage(WM_AWT_DISPOSE, (WPARAM)p, 0);
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    _setForeground
 * Signature: (I)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WComponentPeer__1setForeground(JNIEnv *env, jobject self,
						    jint rgb)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent *)pData;
    p->SetColor(PALETTERGB((rgb>>16)&0xff, 
		           (rgb>>8)&0xff, 
		           (rgb)&0xff));
    p->VerifyState();
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    _setBackground
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer__1setBackground(JNIEnv *env, jobject self,
						    jint rgb)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent *)pData;
    p->SetBackgroundColor(PALETTERGB((rgb>>16)&0xff, 
		                     (rgb>>8)&0xff, 
		                     (rgb)&0xff));
    p->VerifyState();
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    setFont
 * Signature: (Ljava/awt/Font;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_setFont(JNIEnv *env, jobject self, 
					    jobject font)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    ASSERT(font != NULL);
    AwtFont* awtFont = (AwtFont*)jlong_to_ptr(env->GetLongField(font, AwtFont::pDataID)); 
    if (awtFont == NULL) {
	/*arguments of AwtFont::Create are changed for multifont component */
	awtFont = AwtFont::Create(env, font);
    }
    env->SetLongField(font, AwtFont::pDataID,  (long)awtFont);  
    AwtComponent* p = (AwtComponent *)JNI_GET_PDATA(self);
    p->SetFont(awtFont);
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    requestFocus
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_requestFocus(JNIEnv *env, jobject self) 
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent *)pData;
    p->SendMessage(WM_AWT_COMPONENT_SETFOCUS);
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    setCursor
 * Signature: (Ljava/awt/Cursor;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_setCursor(JNIEnv *env, jobject self, 
					      jobject cursor)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(cursor, "null cursor");
    AwtComponent* p = (AwtComponent *)pData;
    p->SendMessage(WM_AWT_SETCURSOR, (WPARAM) cursor);
}
    
/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    start
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_start(JNIEnv *env, jobject self)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* p = (AwtComponent *)pData;
    jobject target = p->GetTarget(env);

    /* Disable window if specified -- windows are enabled by default. */
    jboolean enabled = (jboolean)env->GetBooleanField(target, 
						      AwtComponent::enabledID);
    if (!enabled) {
	::EnableWindow(p->GetHWnd(), FALSE);
    }

    /* The peer is now ready for callbacks, since this is the last  
     * initialization call 
     */
    p->EnableCallbacks(TRUE);
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    _beginValidate
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer__1beginValidate(JNIEnv *env, jobject self)
{
    ASSERT(self);
    jint pData = JNI_GET_PDATA(self);
    if (pData) {
        AwtComponent* p = (AwtComponent *)pData;
        p->SendMessage(WM_AWT_BEGIN_VALIDATE);
    }
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    endValidate
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_endValidate(JNIEnv *env, jobject self)
{
    ASSERT(self);
    jint pData = JNI_GET_PDATA(self);
    if (pData) {
        AwtComponent* p = (AwtComponent *)pData;
        p->SendMessage(WM_AWT_END_VALIDATE);
    }
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    setZOrderPosition
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_setZOrderPosition(JNIEnv *env, 
						      jobject self, 
						      jobject aboveComp)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    if (pData) {
        AwtComponent *p = (AwtComponent *)pData;
        HWND aboveHWnd = NULL;
	
	if (aboveComp != NULL && (pData = JNI_GET_PDATA(aboveComp)) != NULL) {
            AwtComponent* aboveW = (AwtComponent *)pData;
            aboveHWnd = aboveW->GetHWnd();
        }    
        ::SetWindowPos(p->GetHWnd(), aboveHWnd, 
                       0, 0, 0, 0, SWP_NOACTIVATE |SWP_NOMOVE | SWP_NOSIZE);
    }
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    addNativeDropTarget
 * Signature: ()V
 */

JNIEXPORT jint JNICALL
Java_sun_awt_windows_WComponentPeer_addNativeDropTarget(JNIEnv *env, jobject self) {
    return (jint)(((AwtComponent*)JNI_GET_PDATA(self))->CreateDropTarget(env));
}

/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    removeNativeDropTarget
 * Signature: ()V
 */

JNIEXPORT void JNICALL
Java_sun_awt_windows_WComponentPeer_removeNativeDropTarget(JNIEnv *env, jobject self) {
    ((AwtComponent*)JNI_GET_PDATA(self))->DestroyDropTarget(env);
}

} /* extern "C" */


/************************************************************************
 * Diagnostic routines
 */

#ifdef DEBUG

void AwtComponent::VerifyState()
{
    if (AwtToolkit::GetInstance().VerifyComponents() == FALSE) {
        return;
    }

    if (m_callbacksEnabled == FALSE) {
        /* Component is not fully setup yet. */
        return;
    }

    /* Get target bounds. */
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->PushLocalFrame(10) < 0)
        return;

    jobject target = GetTarget(env);

    jint x = env->GetIntField(target, AwtComponent::xID);
    jint y = env->GetIntField(target, AwtComponent::yID);
    jint width = env->GetIntField(target, AwtComponent::widthID);
    jint height = env->GetIntField(target, AwtComponent::heightID);

    /* Convert target origin to absolute coordinates */
    while (TRUE) {

	jobject parent = env->GetObjectField(target, AwtComponent::parentID);
        if (parent == NULL) {
            break;
        }
        x += env->GetIntField(parent, AwtComponent::xID);
        y += env->GetIntField(parent, AwtComponent::yID);

        /* If this component has insets, factor them in, but ignore 
	 * top-level windows.
	 */
	jobject parent2 = env->GetObjectField(parent, AwtComponent::parentID);
        if (parent2 != NULL) {
            jobject peer = GetPeerForTarget(env, parent);
            if (peer != NULL && 
		JNU_IsInstanceOfByName(env, peer, 
				       "sun/awt/windows/WPanelPeer") > 0) {
		jobject insets = 
		    JNU_CallMethodByName(env, NULL, peer,"insets",
					 "()Ljava/awt/Insets;").l;
                x += (env)->GetIntField(insets, AwtInsets::leftID);
                y += (env)->GetIntField(insets, AwtInsets::topID);
            }
        }
	env->DeleteLocalRef(target);
	target = parent;
    }

    // Test whether component's bounds match the native window's
    RECT rect;
    VERIFY(::GetWindowRect(GetHWnd(), &rect));
#if 0
    ASSERT( (x == rect.left) && 
            (y == rect.top) &&
            (width == (rect.right-rect.left)) && 
            (height == (rect.bottom-rect.top)) );
#else
    BOOL fSizeValid = ( (x == rect.left) && 
            (y == rect.top) &&
            (width == (rect.right-rect.left)) && 
            (height == (rect.bottom-rect.top)) );
#endif

    // See if visible state matches
    BOOL wndVisible = ::IsWindowVisible(GetHWnd());
    long targetVisible;
    // To avoid possibly running client code on the toolkit thread, don't
    // do the following check if we're running on the toolkit thread.
    if (AwtToolkit::MainThread() != ::GetCurrentThreadId()) {
        targetVisible = JNU_CallMethodByName(env, NULL, GetTarget(env), 
					          "isShowing", "()Z").z;
        ASSERT(!env->ExceptionOccurred());
    } else {
        targetVisible = wndVisible ? 1 : 0;
    }
#if 0
    ASSERT( (targetVisible && wndVisible) ||
            (!targetVisible && !wndVisible) );
#else
    BOOL fVisibleValid = ( (targetVisible && wndVisible) ||
            (!targetVisible && !wndVisible) );
#endif

    // Check enabled state
    BOOL wndEnabled = ::IsWindowEnabled(GetHWnd());
    jboolean enabled = (jboolean)env->GetBooleanField(target, 
						      AwtComponent::enabledID);
#if 0
    ASSERT( (enabled && wndEnabled) ||
            (!enabled && !wndEnabled) );
#else
    BOOL fEnabledValid = ((enabled && wndEnabled) || 
			  (!(enabled && !wndEnabled) ));

    if (!fSizeValid || !fVisibleValid || !fEnabledValid) {
        printf("AwtComponent::ValidateState() failed:\n");
        // To avoid possibly running client code on the toolkit thread, don't
        // do the following call if we're running on the toolkit thread.
        if (AwtToolkit::MainThread() != ::GetCurrentThreadId()) {
	    jstring targetStr = 
	        (jstring)JNU_CallMethodByName(env, NULL, GetTarget(env), 
					      "getName", 
					      "()Ljava/lang/String;").l;
            ASSERT(!env->ExceptionOccurred());
            printf("\t%S\n", TO_WSTRING(targetStr));
	}
        printf("\twas:       [%d,%d,%dx%d]\n", x, y, width, height);
        if (!fSizeValid) {
            printf("\tshould be: [%d,%d,%dx%d]\n", rect.left, rect.top,
                   rect.right-rect.left, rect.bottom-rect.top);
        }
        if (!fVisibleValid) {
            printf("\tshould be: %s\n", 
                   (targetVisible) ? "visible" : "hidden");
        }
        if (!fEnabledValid) {
            printf("\tshould be: %s\n", 
                   enabled ? "enabled" : "disabled");
        }
    }
#endif
    env->PopLocalFrame(0);
}
#endif //DEBUG
