/*
 * @(#)awt_Component.h	1.85 98/10/07
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_COMPONENT_H
#define AWT_COMPONENT_H

#include "awtmsg.h"
#include "awt_Object.h"
#include "awt_Font.h"
#include "awt_Brush.h"
#include "awt_Pen.h"

#include "java_awt_Component.h"
#include "sun_awt_windows_WComponentPeer.h"
#include "java_awt_event_KeyEvent.h"
#include "java_awt_event_FocusEvent.h"
#include "java_awt_event_MouseEvent.h"
#include "java_awt_event_WindowEvent.h"
#include "java_awt_Dimension.h"

extern const char* szAwtComponentClassName;
extern ATOM thisPropertyAtom;

const UINT IGNORE_KEY = (UINT)-1;
const UINT MAX_ACP_STR_LEN = 7; // ANSI CP identifiers are no longer than this

#define LEFT_BUTTON 1
#define MIDDLE_BUTTON 2
#define RIGHT_BUTTON 3

class AwtSharedDC;
class AwtPopupMenu;

class AwtDropTarget;

/*
 * Message routing codes
 */
enum MsgRouting {
    mrPassAlong,    /* pass along to next in chain */
    mrDoDefault,    /* skip right to underlying default behavior */
    mrConsume,      /* consume msg & terminate routing immediatly, 
		     * don't pass anywhere 
		     */
};

static BOOL IsDBCSEditClass(AwtComponent *comp);

/************************************************************************
 * AwtComponent class
 */

class AwtComponent : public AwtObject {
public:
    enum {
	// combination of all mouse button flags
	ALL_MK_BUTTONS = MK_LBUTTON|MK_MBUTTON|MK_RBUTTON
    };

    /* java.awt.Component fields and method IDs */
    static jfieldID peerID;
    static jfieldID xID;
    static jfieldID yID;
    static jfieldID widthID;
    static jfieldID heightID;
    static jfieldID visibleID;
    static jfieldID backgroundID;
    static jfieldID foregroundID;
    static jfieldID enabledID;
    static jfieldID parentID;

    static jmethodID getFontMID;
    static jmethodID getToolkitMID;
    static jmethodID isEnabledMID;

    AwtComponent();
    virtual ~AwtComponent();

    /*
     * Static component subsystem initialization & termination 
     */

    static void Init();
    static void Term();

    /*
     * Dynamic class registration & creation 
     */
    virtual void RegisterClass();
    virtual void UnregisterClass();
    virtual const char* GetClassName() = 0;

    void CreateHWnd(JNIEnv *env, const char* title, 
		    long windowStyle, long windowExStyle,
                    int x, int y, int w, int h,
                    HWND hWndParent, HMENU hMenu, 
                    COLORREF colorForeground, COLORREF colorBackground,
                    jobject peer);

    void DestroyHWnd();
    void SubclassWindow();
    void UnsubclassWindow();

    static LRESULT CALLBACK WndProc(HWND hWnd, UINT message, 
	WPARAM wParam, LPARAM lParam);
    INLINE static char* ThisProperty() { 
	return (char*)MAKELONG(thisPropertyAtom, 0); 
    }

    /*
     * Access to the various objects of this aggregate component 
     */
    INLINE HWND GetHWnd() { return m_hwnd; }
    INLINE void SetHWnd(HWND hwnd) { m_hwnd = hwnd; }

    INLINE static AwtComponent* GetComponent(HWND hWnd) {
#ifdef DEBUG
        if (hWnd == NULL) {
            return NULL;
        }
        AwtComponent* c = (AwtComponent*)::GetProp(hWnd, ThisProperty());
        ASSERT(c == NULL || c->GetHWnd() == hWnd);
        return c;
#else
        return hWnd ? (AwtComponent*)::GetProp(hWnd, ThisProperty()) : 0;
#endif
    }

    INLINE CriticalSection& GetLock() { return m_lock; }

    /*
     * Access to the properties of the component 
     */
    INLINE COLORREF GetColor() { return m_colorForeground; }
    void SetColor(COLORREF c);
    HPEN GetForegroundPen();

    COLORREF GetBackgroundColor();
    void SetBackgroundColor(COLORREF c);
    HBRUSH GetBackgroundBrush();
    INLINE BOOL IsBackgroundColorSet() { return m_backgroundColorSet; }

    virtual void SetFont(AwtFont *pFont);

    INLINE void SetText(const char* text) { ::SetWindowText(GetHWnd(), text); }
    INLINE int GetText(char* buffer, int size) { 
        return ::GetWindowText(GetHWnd(), buffer, size); 
    }
    INLINE int GetTextLength() { return ::GetWindowTextLength(GetHWnd()); }

    virtual void GetInsets(RECT* rect) { 
        VERIFY(::SetRectEmpty(rect));
    }

    static HWND GetTopLevelParentForWindow(HWND hwndDescendant);

    /*
     * Access to related components & other objects 
     */
    AwtSharedDC* GetSharedDC();

    /*
     * Returns the parent component.  If no parent window, or the
     * parent window isn't an AwtComponent, returns NULL.  
     */
    AwtComponent* GetParent();

    /* Get the component's immediate container. */
    class AwtWindow* GetContainer();

    /* Is a component a container? Used by above method */
    virtual BOOL IsContainer() { return FALSE;} // Plain components can't

    /*
     * Returns an increasing unsigned value used for child control IDs.  
     * There is no attempt to reclaim command ID's.
     */
    INLINE UINT CreateControlID() { return m_nextControlID++; }

    /*
     * Returns the current cursor.  If NULL, returns parent's cursor.
     * Top-level windows always have a default cursor defined.  
     */
    INLINE HCURSOR GetCursor(BOOL checkParent) {
        return (checkParent && m_hCursor == NULL && GetParent()) ? 
            GetParent()->GetCursor(TRUE) : m_hCursor;
    }

    // returns the current code page that should be used in 
    // all MultiByteToWideChar and WideCharToMultiByte calls.
    // This code page should also be use in IsDBCSLeadByteEx.
    INLINE static UINT GetCodePage()
    {
        return m_CodePage;
    }
    
// Added by waleed for BIDI Support
    // returns the right to left status
    INLINE static BOOL GetRTLReadingOrder() {
        return sm_rtlReadingOrder;
    }
    // returns the right to left status
    INLINE static BOOL GetRTL()	{
        return sm_rtl;
    }
    // returns the current sub language
    INLINE static LANGID GetSubLanguage() {
        return SUBLANGID(m_idLang);
    }
// end waleed

    // returns the current input language
    INLINE static LANGID GetInputLanguage()
    {
        return m_idLang;
    }
    // Convert Language ID to CodePage
    static UINT LangToCodePage(LANGID idLang);

    /*
     * methods on this component
     */
    virtual void Show();
    virtual void Hide();
    virtual void Reshape(int x, int y, int w, int h);

    /*
     * Sets the scrollbar values.  'bar' can be either SB_VERT or
     * SB_HORZ.  'min', 'value', and 'max' can have the value INT_MAX
     * which means that the value should not be changed. 
     */
    void SetScrollValues(UINT bar, int min, int value, int max);

    INLINE LRESULT SendMessage(UINT msg, WPARAM wParam=0, LPARAM lParam=0) {
        ASSERT(GetHWnd());
        return ::SendMessage(GetHWnd(), msg, wParam, lParam);
    }
    INLINE long GetStyle() {
        ASSERT(GetHWnd());
        return ::GetWindowLong(GetHWnd(), GWL_STYLE);
    }
    INLINE void SetStyle(long style) {
        ASSERT(GetHWnd());
        // SetWindowLong() error handling as recommended by Win32 API doc.
        ::SetLastError(0);
        LONG ret = ::SetWindowLong(GetHWnd(), GWL_STYLE, style);
        ASSERT(ret != 0 || ::GetLastError() == 0);
    }
    INLINE long GetStyleEx() {
        ASSERT(GetHWnd());
        return ::GetWindowLong(GetHWnd(), GWL_EXSTYLE);
    }
    INLINE void SetStyleEx(long style) {
        ASSERT(GetHWnd());
        // SetWindowLong() error handling as recommended by Win32 API doc.
        ::SetLastError(0);
        LONG ret = ::SetWindowLong(GetHWnd(), GWL_EXSTYLE, style);
        ASSERT(ret != 0 || ::GetLastError() == 0);
    }

    INLINE void SetCursor(HCURSOR hCursor) { 
        ASSERT(hCursor); 
        m_hCursor = hCursor;
        ::SetCursor(m_hCursor);
    }

    virtual BOOL NeedDblClick() { return FALSE; }

    /* for multifont component */
    static void DrawWindowText(HDC hDC, jobject font, jstring text,
			       int x, int y);
    static void DrawGrayText(HDC hDC, jobject font, jstring text,
			     int x, int y);

    void DrawListItem(JNIEnv *env, DRAWITEMSTRUCT far& drawInfo);

    void MeasureListItem(JNIEnv *env, MEASUREITEMSTRUCT far& measureInfo);

    jstring GetItemString(JNIEnv *env, jobject target, long index);

    jint GetFontHeight(JNIEnv *env);

    virtual jobject PreferredItemSize(JNIEnv *env) {ASSERT(FALSE); return NULL; }

    INLINE BOOL isEnabled() {
	JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	if (env->EnsureLocalCapacity(2) < 0) {
	    return NULL;
	}
	jobject self = GetPeer(env);
	jobject target = env->GetObjectField(self, AwtObject::targetID);
	BOOL e = env->CallBooleanMethod(target, AwtComponent::isEnabledMID);
        ASSERT(!env->ExceptionOccurred());
	
	env->DeleteLocalRef(target);

        return e;
    }

    /*
     * Allocate and initialize a new java.awt.event.KeyEvent, and  
     * post it to the peer's target object.  No response is expected 
     * from the target.
     */
    void AwtComponent::SendKeyEvent(long id, jlong when, 
				    long raw, long cooked,
				    long modifiers, MSG* msg=NULL);

    /*
     * Allocate and initialize a new java.awt.event.MouseEvent, and 
     * post it to the peer's target object.  No response is expected 
     * from the target.
     */
    void SendMouseEvent(long id, jlong when, long x, long y, 
                        long modifiers, long clickCount, long
                        popupTrigger, MSG* msg=NULL);

    /*
     * Allocate and initialize a new java.awt.event.FocusEvent, and 
     * post it to the peer's target object.  No response is expected 
     * from the target.
     */
    void SendFocusEvent(long id, BOOL temporary=FALSE);

    /*
     * Allocate and initialize a new java.awt.event.WindowEvent, and 
     * post it to the peer's target object.  No response is expected 
     * from the target.
     */
    void SendWindowEvent(long id, MSG* msg=NULL);

    /* Forward a filtered event directly to the subclassed window.
       synthetic should be TRUE iff the message was generated because
       of a synthetic Java event, rather than a native event. */
    virtual MsgRouting HandleEvent(MSG *msg, BOOL synthetic);

    /* Event->message synthesizer methods. */
    void SynthesizeKeyMessage(JNIEnv *env, jobject keyEvent);
    void SynthesizeMouseMessage(JNIEnv *env, jobject mouseEvent);

    /* Invalidate the specified rectangle. */
    virtual void Invalidate(RECT* r);

    /* Begin and end deferred window positioning. */
    virtual void BeginValidate();
    virtual void EndValidate();

    /* Keyboard conversion routines. */
    static long GetJavaModifiers(UINT flags, int mouseButton);
    static UINT WindowsKeyToJavaKey(UINT windowsKey, UINT modifiers);
    static void JavaKeyToWindowsKey(UINT javaKey, UINT *windowsKey, UINT *modifiers);
    UINT WindowsKeyToJavaChar(UINT wkey, UINT modifiers);

    /* routines used for input method support */
    void SetInputMethod(jobject im, BOOL useNativeCompWindow);
    static int GetContextData(HIMC hIMC, DWORD dwIndex, LPVOID* lplpData);
    static jstring MakeJavaString(LPSTR lpStrA, int cStrA);
    static int* MakeCharOffsetTable(LPSTR lpStrA, int cStrA);
    static int MakeClauseInformation(int cStrA, int *tblCharOffStrA, 
				     int cClauseA, DWORD *lpClauseA, 
				     int cReadStrA, LPSTR lpReadStrA, 
				     int cReadClauseA, DWORD *lpReadClauseA, 
				     int** lpBndClauseW, jstring** lpReadingClauseW );
    static int MakeAttributeInformation(int cStrA, int *tblCharOffStrA, 
					int cAttrA, BYTE* lpAttrA, 
					int** lpBndAttrW, BYTE** lpValAttrW );
    void SendInputMethodEvent(long id, jstring text,
                                int cClause, int* rgClauseBoundary, jstring* rgClauseReading,
                                int cAttrBlock, int* rgAttrBoundary, BYTE *rgAttrValue,
                                int commitedTextLength, int caretPos, int visiblePos);
    void InquireCandidatePosition(LPPOINT lpPt);

    /*
     * Windows message handler functions
     */
    virtual LRESULT WindowProc(UINT message, WPARAM wParam, LPARAM lParam);
    LRESULT DefWindowProc(UINT msg, WPARAM wParam, LPARAM lParam);

    /* return true if msg is processed */
    virtual MsgRouting PreProcessMsg(MSG& msg); 

    virtual MsgRouting WmCreate() {return mrDoDefault;}
    virtual MsgRouting WmClose() {return mrDoDefault;}
    virtual MsgRouting WmDestroy();

    virtual MsgRouting WmActivate(UINT nState, BOOL fMinimized) 
    {
	return mrDoDefault;
    }

    virtual MsgRouting WmEraseBkgnd(HDC hDC, BOOL& didErase) 
    {
	return mrDoDefault;
    }

    virtual MsgRouting WmPaint(HDC hDC);
    virtual MsgRouting WmMove(int x, int y);
    virtual MsgRouting WmSize(int type, int w, int h);
    virtual MsgRouting WmShowWindow(int show, long status);
    virtual MsgRouting WmSetFocus(HWND hWndLost);
    virtual MsgRouting WmKillFocus(HWND hWndGot);

    virtual MsgRouting WmCtlColor(HDC hDC, HWND hCtrl, 
				  UINT ctlColor, HBRUSH& retBrush);
    virtual MsgRouting WmHScroll(UINT scrollCode, UINT pos, HWND hScrollBar);
    virtual MsgRouting WmVScroll(UINT scrollCode, UINT pos, HWND hScrollBar);
    virtual MsgRouting WmSetCursor(HWND hWnd, UINT hitTest, UINT message, 
                                   BOOL& retVal);
    virtual MsgRouting WmAwtSetCursor(jobject cursor);
    
    virtual MsgRouting WmMouseEnter(UINT flags, int x, int y);
    virtual MsgRouting WmMouseDown(UINT flags, int x, int y, int button);
    virtual MsgRouting WmMouseUp(UINT flags, int x, int y, int button);
    virtual MsgRouting WmMouseMove(UINT flags, int x, int y);
    virtual MsgRouting WmMouseExit(UINT flags, int x, int y);

    virtual MsgRouting WmKeyDown(UINT character, UINT repCnt, UINT flags, BOOL system);
    virtual MsgRouting WmKeyUp(UINT character, UINT repCnt, UINT flags, BOOL system);
    virtual MsgRouting WmChar(UINT character, UINT repCnt, UINT flags, BOOL system);
    virtual MsgRouting WmIMEChar(UINT character, UINT repCnt, UINT flags, BOOL system);
    virtual MsgRouting WmInputLangChange(UINT charset, UINT hKeyBoardLayout);
    virtual MsgRouting WmForwardChar(WCHAR character, LPARAM lParam,
				     BOOL synthethic);
    virtual MsgRouting WmPaste();

    virtual void SetCompositionWindow();
    virtual void SetCandidateWindow(int iCandType, BOOL getLocation);
    virtual MsgRouting WmImeSetContext(BOOL fSet, LPARAM *lplParam);
    virtual MsgRouting WmImeNotify(UINT subMsg, UINT bitsCandType);
    virtual MsgRouting WmImeStartComposition();
    virtual MsgRouting WmImeEndComposition();
    virtual MsgRouting WmImeComposition(WORD wChar, long flags);

    virtual MsgRouting WmTimer(UINT timerID) {return mrDoDefault;}

    virtual MsgRouting WmCommand(UINT id, HWND hWndCtrl, UINT notifyCode);

    /* reflected WmCommand from parent */
    virtual MsgRouting WmNotify(UINT notifyCode);  

    virtual MsgRouting WmCompareItem(UINT /*ctrlId*/, 
                                     COMPAREITEMSTRUCT far& compareInfo, 
                                     long& result);
    virtual MsgRouting WmDeleteItem(UINT /*ctrlId*/, 
                                    DELETEITEMSTRUCT far& deleteInfo);
    virtual MsgRouting WmDrawItem(UINT /*ctrlId*/, 
                                  DRAWITEMSTRUCT far& drawInfo);
    virtual MsgRouting WmMeasureItem(UINT ctrlId, 
                                     MEASUREITEMSTRUCT far& measureInfo);

    virtual MsgRouting WmPrint(HDC hDC, UINT flags);
    virtual MsgRouting WmPrintClient(HDC hDC, UINT flags);
    virtual MsgRouting PrintComponent(class AwtGraphics* g);

    virtual MsgRouting WmNcCalcSize(BOOL fCalcValidRects, 
				    LPNCCALCSIZE_PARAMS lpncsp, UINT& retVal);
    virtual MsgRouting WmNcPaint(HRGN hrgn);
    virtual MsgRouting WmNcHitTest(UINT x, UINT y, UINT& retVal);
    virtual MsgRouting WmSysCommand(UINT uCmdType, UINT xPos, UINT yPos);
    virtual MsgRouting WmExitSizeMove();
    virtual MsgRouting WmEnterMenuLoop(BOOL isTrackPopupMenu);
    virtual MsgRouting WmExitMenuLoop(BOOL isTrackPopupMenu);

    virtual MsgRouting WmQueryNewPalette(UINT& retVal);
    virtual MsgRouting WmPaletteChanged(HWND hwndPalChg);
    virtual MsgRouting WmStyleChanged(WORD wStyleType, LPSTYLESTRUCT lpss);
    virtual MsgRouting WmSettingChange(WORD wFlag, LPCTSTR pszSection);

    /*
     * HWND, AwtComponent and Java Peer interaction 
     *
     * Link the C++, Java peer, and HWNDs together. 
     */
    void LinkObjects(JNIEnv *env, jobject peer);

    void UnlinkObjects();

#ifdef DEBUG
    virtual void VerifyState(); /* verify component and peer are in sync. */
#else
    void VerifyState() {}       /* no-op */
#endif

    virtual AwtDropTarget* CreateDropTarget(JNIEnv* env);
    virtual void	   DestroyDropTarget(JNIEnv* env);

protected:
    HWND     m_hwnd;
    UINT     m_myControlID;	/* its own ID from the view point of parent */
    BOOL     m_backgroundColorSet;
    BOOL     m_dragged;         /* in drag operation */

private:
    COLORREF m_colorForeground;
    COLORREF m_colorBackground;

    AwtPen*  m_penForeground;
    AwtBrush* m_brushBackground;

    AwtSharedDC* m_sharedDC;

    WNDPROC  m_DefWindowProc;

    UINT     m_nextControlID;   /* provides a unique ID for child controls */

    HCURSOR  m_hCursor;

    CriticalSection m_lock;

    // DeferWindowPos handle for batched-up window positioning
    HDWP     m_hdwp;
    // Counter to handle nested calls to Begin/EndValidate
    UINT     m_validationNestCount;

    AwtDropTarget* m_dropTarget; // associated DropTarget object

    // when we process WM_INPUTLANGCHANGE we need to change the input
    // codepage. All MultiByteToWideChar and WideCharToMultiByte
    // calls should use this code page
    static UINT    m_CodePage;
    // Current input language (=low word of keyboardlayout handle)
    // m_idLang is shared by all instance of AwtComponent because 
    // keyboardlayout is shared.
    static LANGID m_idLang;    
    static BOOL m_isWin95;  
    static BOOL sm_rtl;
    static BOOL sm_rtlReadingOrder;

    jobject m_InputMethod;
    BOOL    m_useNativeCompWindow;

    // Determines if a given virtual key is a numpad key
    static BOOL IsNumPadNumberKey( UINT vkey );

private:
    /*
     * The association list of children's IDs and corresponding components. 
     * Some components like Choice or List are required their sizes while
     * the creations of themselfs are in progress.
     */
    class ChildListItem {
    public:
        ChildListItem(UINT id, AwtComponent* component) {
            m_ID = id;
            m_Component = component;
            m_next = NULL;
        }
        ~ChildListItem() {
            if (m_next != NULL)
                delete m_next;
        }

        UINT m_ID;
        AwtComponent* m_Component;
        ChildListItem* m_next;
    };

public:
    INLINE void PushChild(UINT id, AwtComponent* component) {
        ChildListItem* child = new ChildListItem(id, component);
        child->m_next = m_childList;
        m_childList = child;
    }
    

private:
    AwtComponent* SearchChild(UINT id);
    void RemoveChild(UINT id) ;

    ChildListItem* m_childList;

    //
    //
    //

public:
   typedef int (*MsgWorkProc)(AwtComponent*, UINT msg, void *);

   virtual void ScheduleMsgWorkProc(MsgWorkProc wp, void* closure);

private:
    typedef struct _MsgWorkProcRec {
	MsgWorkProc		proc;
	void*	 		closure;
	
	struct _MsgWorkProcRec*	next;
    } MsgWorkProcRec, *MsgWorkProcPtr;

    MsgWorkProcPtr	m_wprocs;

    inline void AddMsgWorkProc(MsgWorkProc proc, void* closure) {
	MsgWorkProcPtr tmp = new MsgWorkProcRec();

	tmp->proc    = proc;
	tmp->closure = closure;

	tmp->next = m_wprocs;
	m_wprocs  = tmp;
    }

    inline void CallMsgWorkProcs(UINT msg) {
	MsgWorkProcPtr* wpp = &m_wprocs;

        while (*wpp != (MsgWorkProcPtr)NULL) {
	    if ((*(*wpp)->proc)(this, msg, (*wpp)->closure)) {
		MsgWorkProcPtr tmp = *wpp;

		*wpp = (*wpp)->next;

		delete tmp;
	    } else
		wpp = &((*wpp)->next);
	}
    }

    inline void DiscardMsgWorkProcs() {
	MsgWorkProcPtr wpp = m_wprocs;

	while (wpp != (MsgWorkProcPtr)NULL) {
		MsgWorkProcPtr tmp = wpp;
	
		wpp = tmp->next;

		delete tmp;
	}
    }
};

#include "ObjectList.h"

#endif /* AWT_COMPONENT_H */
