/*
 * @(#)awt_Cursor.cpp	1.4 98/05/06
 *
 * Copyright 1997 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Cursor.h"
#include "awt_Toolkit.h"

/************************************************************************
 * AwtCursor fields
 */
jfieldID AwtCursor::pDataID;
jfieldID AwtCursor::typeID;

extern "C" {

/************************************************************************
 * AwtCursor methods
 */

/*
 * Class:     jave_awt_Cursor
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_Cursor_initIDs(JNIEnv *env, jclass cls)
{
    AwtCursor::typeID = env->GetFieldID(cls, "type", "I");
    ASSERT(AwtCursor::typeID != NULL);
}

/************************************************************************
 * AwtCursor methods
 */

/*
 * Class:     sun_awt_windows_WCustomCursor
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WCustomCursor_initIDs(JNIEnv *env, jclass cls)
{
    AwtCursor::pDataID = env->GetFieldID(cls, "pData", "I");
    ASSERT(AwtCursor::pDataID != NULL);
}

/************************************************************************
 * WCustomCursor native methods
 */

/*
 * Class:     sun_awt_windows_WCustomCursor
 * Method:    createNativeCursor
 * Signature: ([IIIII)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WCustomCursor_createCursor(
    JNIEnv *env, jobject self, jbyteArray xorMask, jbyteArray andMask,
    jint width, jint height, jint xHotSpot, jint yHotSpot)
{
    AWT_TRACE(("%x: WCustomCursor.createCursor()\n", self));

    HCURSOR hCursor = (HCURSOR)env->GetIntField(self, AwtCursor::pDataID);
    if (hCursor != NULL) {
        // Don't verify DestroyIcon() as NT 4.0 returns 0 on success or failure
        ::DestroyIcon(hCursor);
    }

    if (width != ::GetSystemMetrics(SM_CXCURSOR) || 
        height != ::GetSystemMetrics(SM_CYCURSOR)) {
	JNU_ThrowArrayIndexOutOfBoundsException(env, "bad width and/or height");
        return;
    }
    int nBits = width * height;
    if (nBits != env->GetArrayLength(xorMask)*8 ||
        nBits != env->GetArrayLength(andMask)*8) {
	JNU_ThrowArrayIndexOutOfBoundsException(env, "bad bit array length");
        return;
    }

    int length = env->GetArrayLength(xorMask);
    jbyte *xorMaskPtr = new jbyte[length];
    env->GetByteArrayRegion(xorMask, 0, length, xorMaskPtr);

    jbyte *andMaskPtr = new jbyte[length];
    env->GetByteArrayRegion(andMask, 0, length, andMaskPtr);

    HINSTANCE hInst = AwtToolkit::GetInstance().GetModuleHandle();
    hCursor = CreateCursor(hInst, xHotSpot, yHotSpot, width, height, 
                           andMaskPtr, xorMaskPtr);
    ASSERT(hCursor);

    delete[] xorMaskPtr;
    delete[] andMaskPtr;

    env->SetIntField(self, AwtCursor::pDataID, (jint)hCursor);
}

/*
 * Class:     sun_awt_windows_WCustomCursor
 * Method:    getCursorWidth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_windows_WCustomCursor_getCursorWidth
  (JNIEnv *, jclass)
{
    AWT_TRACE(("WCustomCursor.getCursorWidth()\n"));
    return (jint)::GetSystemMetrics(SM_CXCURSOR);
}

/*
 * Class:     sun_awt_windows_WCustomCursor
 * Method:    getCursorHeight
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_sun_awt_windows_WCustomCursor_getCursorHeight
  (JNIEnv *, jclass)
{
    AWT_TRACE(("WCustomCursor.getCursorHeight()\n"));
    return (jint)::GetSystemMetrics(SM_CYCURSOR);
}

/*
 * Class:     sun_awt_windows_WCustomCursor
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WCustomCursor_finalize(JNIEnv *env, jobject self)
{
    AWT_TRACE(("%x: WCustomCursor.finalize()\n", self));
    HCURSOR hCursor = (HCURSOR)env->GetIntField(self, AwtCursor::pDataID);
    VERIFY(DestroyCursor(hCursor));
}

} /* extern "C" */


