/*
 * @(#)awt_DnDDS.h	1.8 98/08/25
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_DND_DS_H
#define AWT_DND_DS_H

#include <Ole2.h>

#include <jni.h>
#include <jvm.h>
#include <jni_util.h>

#include "awt_Object.h"
#include "awt_Component.h"
#include "awt_window.h"

/**
 * Drag Source code
 */

class AwtDragSource : virtual public IDropSource, virtual public IDataObject {
    public:

	AwtDragSource(JNIEnv* env, jobject peer, jobject trigger, jint actions, jobjectArray types);

	virtual ~AwtDragSource();

	// IUnknown

	virtual HRESULT __stdcall QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject);

	virtual ULONG   __stdcall AddRef(void);
	virtual ULONG   __stdcall Release(void);

	// IDropSource

	virtual HRESULT __stdcall QueryContinueDrag(BOOL fEscapeKeyPressed, DWORD grfKeyState);

	virtual HRESULT __stdcall GiveFeedback(DWORD dwEffect);
	
	// IDataObject

	virtual HRESULT __stdcall GetData(FORMATETC __RPC_FAR *pFormatEtc, STGMEDIUM __RPC_FAR *pmedium);
	virtual HRESULT __stdcall GetDataHere(FORMATETC __RPC_FAR *pFormatEtc, STGMEDIUM __RPC_FAR *pmedium);

	virtual HRESULT __stdcall QueryGetData(FORMATETC __RPC_FAR *pFormatEtc);

	virtual HRESULT __stdcall GetCanonicalFormatEtc(FORMATETC __RPC_FAR *pFormatEtcIn, FORMATETC __RPC_FAR *pFormatEtcOut);

	virtual HRESULT __stdcall SetData(FORMATETC __RPC_FAR *pFormatEtc, STGMEDIUM __RPC_FAR *pmedium, BOOL fRelease);

	virtual HRESULT __stdcall EnumFormatEtc(DWORD dwDirection, IEnumFORMATETC *__RPC_FAR *ppenumFormatEtc);

	virtual HRESULT __stdcall DAdvise(FORMATETC __RPC_FAR *pFormatEtc, DWORD advf, IAdviseSink __RPC_FAR *pAdvSink, DWORD __RPC_FAR *pdwConnection);
	virtual HRESULT __stdcall DUnadvise(DWORD dwConnection);
	virtual HRESULT __stdcall EnumDAdvise(IEnumSTATDATA __RPC_FAR *__RPC_FAR *ppenumAdvise);

	
	// AwtDragSource

	static void InitAwtDragSource();

	static void StartDrag(AwtDragSource* self, jobject cursor);

	virtual HCURSOR SetCursor(jobject cursor);

	INLINE unsigned int getNTypes() { return m_ntypes; }

	INLINE FORMATETC getType(unsigned int i) { return (m_types[i]->m_formatEtc); }

	INLINE jobject GetPeer() { return m_peer; }

	INLINE void Signal() { ::ReleaseMutex(m_mutex); }

    protected:
	INLINE void WaitUntilSignalled(BOOL retain) {
	    do {
		// nothing ...
	    } while(::WaitForSingleObject(m_mutex, INFINITE) == WAIT_FAILED);

	    if (!retain) ::ReleaseMutex(m_mutex);
 	}

	static void _DoDragDrop(void* param);

	typedef struct _CacheRec {
	    FORMATETC   m_formatEtc;
	    jobject	m_atomName;
	} CacheRec, *CachePtr;

	HRESULT __stdcall MatchFormatEtc(FORMATETC __RPC_FAR *pFormatEtcIn, CachePtr* cacheEnt);

   private:

	void LoadCache(jobjectArray types);
	void UnloadCache();

	static int __cdecl _compar(const void *, const void *);

    protected:

	class ADSIEnumFormatEtc : public virtual IEnumFORMATETC {
	    public:
		ADSIEnumFormatEtc(AwtDragSource* parent);

		virtual ~ADSIEnumFormatEtc();

		// IUnknown

		virtual HRESULT __stdcall QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject);

		virtual ULONG   __stdcall AddRef(void);
		virtual ULONG   __stdcall Release(void);
	
		// IEnumFORMATETC

		virtual HRESULT _stdcall Next(ULONG celt, FORMATETC __RPC_FAR *rgelt, ULONG __RPC_FAR *pceltFetched);
		virtual HRESULT _stdcall Skip(ULONG celt);
		virtual HRESULT _stdcall Reset();
		virtual HRESULT _stdcall Clone(IEnumFORMATETC __RPC_FAR *__RPC_FAR *ppenum);

	    private:
		AwtDragSource*	m_parent;
		ULONG		m_refs;

		unsigned int	m_idx;
	};

	class ADSIStreamProxy : public virtual IStream {
	    private:
		ADSIStreamProxy(ADSIStreamProxy* cloneof);

	    public:
		ADSIStreamProxy(AwtDragSource* parent, jbyteArray buffer, jint len);

		virtual ~ADSIStreamProxy();

		// IUnknown

		virtual HRESULT __stdcall QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject);

		virtual ULONG   __stdcall AddRef(void);
		virtual ULONG   __stdcall Release(void);
		
		// IStream


		virtual  HRESULT __stdcall Read(void __RPC_FAR *pv, ULONG cb, ULONG __RPC_FAR *pcbRead);

		virtual  HRESULT __stdcall Write(const void __RPC_FAR *pv, ULONG cb, ULONG __RPC_FAR *pcbWritten);

		virtual  HRESULT __stdcall Seek(LARGE_INTEGER dlibMove, DWORD dwOrigin, ULARGE_INTEGER __RPC_FAR *plibNewPosition);

		virtual HRESULT __stdcall SetSize(ULARGE_INTEGER libNewSize);

		virtual  HRESULT __stdcall CopyTo(IStream __RPC_FAR *pstm, ULARGE_INTEGER cb, ULARGE_INTEGER __RPC_FAR *pcbRead, ULARGE_INTEGER __RPC_FAR *pcbWritten);

		virtual HRESULT __stdcall Commit(DWORD grfCommitFlags);

		virtual HRESULT __stdcall Revert();

		virtual HRESULT __stdcall LockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType);

		virtual HRESULT __stdcall UnlockRegion(ULARGE_INTEGER libOffset, ULARGE_INTEGER cb, DWORD dwLockType);

		virtual HRESULT __stdcall Stat(STATSTG __RPC_FAR *pstatstg, DWORD grfStatFlag);

		virtual HRESULT __stdcall Clone(IStream __RPC_FAR *__RPC_FAR *ppstm);
	    protected:
		AwtDragSource*	 m_parent;

  		signed   char*	 m_buffer;
		unsigned int	 m_off;
		unsigned int	 m_blen;

		STATSTG		 m_statstg;

		ADSIStreamProxy* m_cloneof;

		ULONG		 m_refs;
	};

    private:
 
	// instance vars ...

	jobject		m_peer;

	jobject		m_trigger;

	jint		m_initmods;
	jint		m_lastmods;

	jboolean	m_startCond;

	HWND		m_droptarget;
	int		m_enterpending;

	jint		m_actions;

	DWORD		m_effects;

	jboolean	m_cancel;

	CachePtr*	m_types;
	unsigned int	m_ntypes;

	ULONG		m_refs;

 	int		m_cursortype;
	HCURSOR		m_cursor;

	HANDLE		m_mutex;

	// static's ...

	static jclass		dSCClazz;
	static jclass		awtCClazz;
	static jclass		awtIEClazz;
	static jclass		wdscpBClazz;

	static jmethodID	dSCdragenter;
	static jmethodID	dSCdragmotion;
	static jmethodID	dSCopschanged;
	static jmethodID	dSCdragexit;
	static jmethodID	dSCddfinish;

	static jmethodID	dSCgetdxh;
	static jmethodID	dSCgetdyh;
	static jmethodID	dSCgetdiwcr;

	static jmethodID	dSCconvertdata;
	static jmethodID	dSCconvertfiles;

	static jmethodID	wdscpBgetbuffer;
	static jmethodID	wdscpBsize;

	static jfieldID		awtCtype;
	static jfieldID		awtCpdata;
	static jfieldID		awtIEmods;
};

#endif /* AWT_DND_DS_H */
