/*
 * @(#)awt_DnDDT.cpp	1.15 98/08/25
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_DnDDT.h"
#include "awt_DnDDS.h"
#include "awt_Toolkit.h"
#include "java_awt_dnd_DnDConstants.h"
#include "sun_awt_windows_WDropTargetContextPeer.h"

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <shellapi.h>

// forwards

extern "C" {
DWORD __cdecl convertActionsToDROPEFFECT(jint actions);
jint  __cdecl convertDROPEFFECTToActions(DWORD effects);

extern const char *const PredefinedClipboardNames[];

DWORD __cdecl mapModsToDROPEFFECT(DWORD, DWORD);
} // extern "C"

/**
 * static initialization
 */


jclass AwtDropTarget::dTCClazz = (jclass)NULL;

jclass AwtDropTarget::string = (jclass)NULL;
jclass AwtDropTarget::file   = (jclass)NULL;

jmethodID AwtDropTarget::dTCgetp   = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCenter  = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCexit   = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCmotion = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCdrop   = (jmethodID)NULL;
 
jmethodID AwtDropTarget::dTCgetms  = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCgetfs  = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCgetis  = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCfilei  = (jmethodID)NULL;
jmethodID AwtDropTarget::dTCgetl   = (jmethodID)NULL;

void AwtDropTarget::InitAwtDropTarget() {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    
    /*
     * load class java.lang.string
     */
    
    string = env->FindClass("java/lang/String");
    
    if (JNU_IsNull(env, AwtDropTarget::string)) {
        env->ThrowNew(env->FindClass("java/lang/ClassNotFoundException"),
    		      "Cant find java/lang/String"
        );
    }
    
    string = (jclass)env->NewGlobalRef(string); /* never freed! */
    
    /*
     * load class java.io.File
     */
    
    file = env->FindClass("java/io/File");
    
    if (JNU_IsNull(env, AwtDropTarget::file)) {
        env->ThrowNew(env->FindClass("java/lang/ClassNotFoundException"),
    		      "Cant find java/io/File"
        );
    }
    
    file = (jclass)env->NewGlobalRef(file); /* never freed! */
    
    dTCfilei = env->GetMethodID(file,
				"<init>",
				"(Ljava/lang/String;)V"
	       );

    if (JNU_IsNull(env, dTCfilei)) return;

    /*
     * load class sun.awt.motif.WDropTargetContextPeer and resolve methods
     */
    
    dTCClazz = env->FindClass("sun/awt/windows/WDropTargetContextPeer");
    
    if (JNU_IsNull(env, dTCClazz)) {
        env->ThrowNew(env->FindClass("java/lang/ClassNotFoundException"),
    		      "Cant find sun/awt/windows/WDropTargetContextPeer"
        );
    }
    
    dTCClazz = (jclass)env->NewGlobalRef(dTCClazz); /* never freed! */
    
    if (JNU_IsNull(env, dTCClazz)) return;
    
    dTCgetp = env->GetStaticMethodID(dTCClazz,
    				    "getWDropTargetContextPeer",
    				    "()Lsun/awt/windows/WDropTargetContextPeer;"
    	        );
    
    if (JNU_IsNull(env, dTCgetp)) return;
    
    dTCenter = env->GetMethodID(dTCClazz,
    			        "handleEnterMessage",
				"(Ljava/awt/Component;IIII[Ljava/lang/String;I)I"
    	       );
    
    if (JNU_IsNull(env, dTCenter)) return;
    
    dTCexit = env->GetMethodID(dTCClazz,
    			       "handleExitMessage",
    			       "(Ljava/awt/Component;I)V"
    	      );
    
    
    if (JNU_IsNull(env, dTCexit)) return;
    
    dTCmotion = env->GetMethodID(dTCClazz,
    			         "handleMotionMessage",
				 "(Ljava/awt/Component;IIII[Ljava/lang/String;I)I"
    	        );
    
    
    if (JNU_IsNull(env, dTCmotion)) return;
    
    dTCdrop = env->GetMethodID(dTCClazz,
    			       "handleDropMessage",
			       "(Ljava/awt/Component;IIII[Ljava/lang/String;I)V"
    	      );
    
    if (JNU_IsNull(env, dTCdrop)) return;

    dTCgetms = env->GetStaticMethodID(dTCClazz,
				      "getMemoryStream",
				      "([BLjava/lang/String;)Ljava/io/ByteArrayInputStream;"
   	       );

    if (JNU_IsNull(env, dTCgetms)) return;

    dTCgetfs = env->GetStaticMethodID(dTCClazz,
				      "getFileStream",
				      "(Ljava/lang/String;Ljava/lang/String;I)Ljava/io/FileInputStream;"
   	       );

    if (JNU_IsNull(env, dTCgetfs)) return;

    dTCgetis = env->GetStaticMethodID(dTCClazz,
				      "getIStream",
				      "(ILjava/lang/String;)Ljava/lang/Object;"
   	       );

    if (JNU_IsNull(env, dTCgetis)) return;

    dTCgetl = env->GetStaticMethodID(dTCClazz,
				     "getList",
				     "([Ljava/lang/Object;)Ljava/util/List;"
   	       );

    if (JNU_IsNull(env, dTCgetis)) return;
}

/**
 * constructor
 */

AwtDropTarget::AwtDropTarget(JNIEnv* env, AwtComponent* component) {

    m_component     = component;
    m_window	    = component->GetHWnd();
    m_refs          = 0U;
    m_target	    = env->NewGlobalRef(component->GetTarget(env));
    m_registered    = 0;
    m_dataObject    = NULL;
    m_enumFormatEtc = NULL;
    m_formats	    = NULL;
    m_nformats	    = 0;
    m_dtcp    	    = NULL;
    m_natives 	    = NULL;
    m_mutex  	    = ::CreateMutex(NULL, FALSE, NULL);

    AddRef(); // ref == 1

    RegisterTarget(TRUE);

    ::CoLockObjectExternal((IUnknown*)this, TRUE, FALSE);
}

/**
 * destructor
 */

AwtDropTarget::~AwtDropTarget() {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    RegisterTarget(FALSE);

    ::CloseHandle(m_mutex);

    env->DeleteGlobalRef(m_target);

    ::CoLockObjectExternal((IUnknown*)this, FALSE, FALSE);
}

/**
 * QueryInterface
 */

HRESULT __stdcall AwtDropTarget::QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject) {
    if (riid == IID_IUnknown) {
	*ppvObject = (void __RPC_FAR *__RPC_FAR)(IUnknown*)this;
	AddRef();
	return S_OK;
    } else if (riid == IID_IDropTarget) {
	*ppvObject = (void __RPC_FAR *__RPC_FAR)(IDropTarget*)this;
	AddRef();
	return S_OK;
    } else if (riid == IID_IAdviseSink) {
	*ppvObject = (void __RPC_FAR *__RPC_FAR)(IAdviseSink*)this;
	AddRef();
	return S_OK;
    } else {
	*ppvObject = (void __RPC_FAR *__RPC_FAR)NULL;
	return E_NOINTERFACE;
    }
}

/**
 * AddRef
 */

ULONG __stdcall AwtDropTarget::AddRef() {
    return (ULONG)++m_refs;
}

/**
 * Release
 */

ULONG __stdcall AwtDropTarget::Release() {
    int refs;

    if ((refs = --m_refs) == 0) delete this;

    return (ULONG)refs;
}

/**
 * DragEnter
 */

HRESULT __stdcall AwtDropTarget::DragEnter(IDataObject __RPC_FAR *pDataObj, DWORD grfKeyState, POINTL pt, DWORD __RPC_FAR *pdwEffect) {
    JNIEnv*    env       = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    HRESULT    ret       = S_OK;
    DWORD      retEffect = DROPEFFECT_NONE;

    m_dtcp = env->CallStaticObjectMethod(dTCClazz, dTCgetp);

    if (JNU_IsNull(env, m_dtcp) || !JNU_IsNull(env, env->ExceptionOccurred())) {
	return ret;
    }

    LoadCache(pDataObj);

    if (m_nformats > 0) {
        POINT cp;
	RECT  wr;

	::GetWindowRect(m_window, &wr);
	
    	cp.x = pt.x - wr.left;
	cp.y = pt.y - wr.top;

	jint actions = env->CallIntMethod(m_dtcp,
			 	          dTCenter, 
					  m_target,
					  (jint)cp.x,
					  (jint)cp.y,
					  ::convertDROPEFFECTToActions(mapModsToDROPEFFECT(*pdwEffect, grfKeyState)),
					  ::convertDROPEFFECTToActions(*pdwEffect),
					  m_natives,
					  (jint)this
		        );

	if (!JNU_IsNull(env, env->ExceptionOccurred())) {
	    env->ExceptionDescribe();
	    env->ExceptionClear();
	    actions = java_awt_dnd_DnDConstants_ACTION_NONE;
	}

	retEffect = ::convertActionsToDROPEFFECT(actions);
    }

    *pdwEffect = retEffect;

    return ret;
}

/**
 * DragOver
 */

HRESULT __stdcall AwtDropTarget::DragOver(DWORD grfKeyState, POINTL pt, DWORD __RPC_FAR *pdwEffect) {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    HRESULT ret = S_OK;
    POINT   cp;
    RECT    wr;
    jint    actions;
	
    ::GetWindowRect(m_window, &wr);

    cp.x = pt.x - wr.left;
    cp.y = pt.y - wr.top;

    actions = env->CallIntMethod(m_dtcp,
			      	 dTCmotion,
				 m_target,
			         (jint)cp.x,
			         (jint)cp.y,
				 ::convertDROPEFFECTToActions(mapModsToDROPEFFECT(*pdwEffect, grfKeyState)),
			         ::convertDROPEFFECTToActions(*pdwEffect),
			         m_natives,
			         (jint)this
	      );

    if (!JNU_IsNull(env, env->ExceptionOccurred())) {
	env->ExceptionDescribe();
	env->ExceptionClear();
	actions = java_awt_dnd_DnDConstants_ACTION_NONE;
    }

    *pdwEffect = ::convertActionsToDROPEFFECT(actions);

    return ret;
}

/**
 * DragLeave
 */

HRESULT __stdcall AwtDropTarget::DragLeave() {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    HRESULT ret = S_OK;
    
    env->CallVoidMethod(m_dtcp, dTCexit, m_target, (jint)this);

    if (!JNU_IsNull(env, env->ExceptionOccurred())) {
	env->ExceptionDescribe();
	env->ExceptionClear();
    }

    DragCleanup();

    return ret;
}

/**
 * Drop
 */

HRESULT __stdcall AwtDropTarget::Drop(IDataObject __RPC_FAR *pDataObj, DWORD grfKeyState, POINTL pt, DWORD __RPC_FAR *pdwEffect) {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    HRESULT ret = S_OK;
    POINT   cp;
    RECT    wr;
	
    LoadCache(pDataObj);

    if (m_nformats == 0) {
	*pdwEffect = DROPEFFECT_NONE;
	return ret;
    }

    ::GetWindowRect(m_window, &wr);

    cp.x = pt.x - wr.left;
    cp.y = pt.y - wr.top;

    m_dropActions = java_awt_dnd_DnDConstants_ACTION_NONE;

    env->CallVoidMethod(m_dtcp,
		        dTCdrop,
		        m_target,
		        (jint)cp.x,
			(jint)cp.y,
			::convertDROPEFFECTToActions(mapModsToDROPEFFECT(*pdwEffect, grfKeyState)),
			::convertDROPEFFECTToActions(*pdwEffect),
			m_natives,
			(jint)this
    );

    if (!JNU_IsNull(env, env->ExceptionOccurred())) {
	env->ExceptionDescribe();
	env->ExceptionClear();
	ret = E_FAIL;
    }

    AwtToolkit::GetInstance().MessageLoop();

    *pdwEffect = ::convertActionsToDROPEFFECT(m_dropActions);

    DragCleanup();

    return ret;
}

/**
 * DoDropDone
 */

void AwtDropTarget::DoDropDone(jint actions) {
     DropDoneRec ddr = { this, actions };

    AwtToolkit::GetInstance().InvokeFunction(_DropDone, &ddr);
}

/**
 * _DropDone
 */

void AwtDropTarget::_DropDone(void* param) {
	DropDonePtr ddrp = (DropDonePtr)param;

	(ddrp->dropTarget)->DropDone(ddrp->action);
}

/**
 * DropDone
 */

void AwtDropTarget::DropDone(jint actions) {
    m_dropActions = actions;
    AwtToolkit::GetInstance().QuitMessageLoop(AwtToolkit::EXIT_ENCLOSING_LOOP);
}

/**
 * DeferredRegisterTarget
 */

int AwtDropTarget::DeferredRegisterTarget(AwtComponent* cont, UINT msg, void* closure) {
    if (msg != WM_AWT_COMPONENT_SHOW) return FALSE;

    ((AwtDropTarget*)closure)->RegisterTarget(::IsWindowVisible(cont->GetHWnd()));

    return TRUE;
}

/**
 * DoRegisterTarget
 */

void AwtDropTarget::_RegisterTarget(void* param) {
    RegisterTargetPtr rtrp = (RegisterTargetPtr)param;

    rtrp->dropTarget->RegisterTarget(rtrp->show);
}

/**
 * RegisterTarget
 */

void AwtDropTarget::RegisterTarget(WORD show) {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    HRESULT res;
 
    if (!AwtToolkit::IsMainThread()) {
	RegisterTargetRec rtr = { this, show };

	AwtToolkit::GetInstance().InvokeFunction(_RegisterTarget, &rtr);

	return;
    }

    // if we are'nt yet visible, defer until the parent is!

    if (show) {
	if (::IsWindowVisible(m_window) == 0) {
	    m_component->GetContainer()->ScheduleMsgWorkProc(DeferredRegisterTarget, (void*)this);
	    return;
	} else
            res = ::RegisterDragDrop(m_window, (IDropTarget*)this);
    } else
	res = ::RevokeDragDrop(m_window);

    if (res == S_OK) m_registered = show;
}

/**
 * DoGetData
 */

jobject AwtDropTarget::DoGetData(jstring native) {
    jobject    ret = (jobject)NULL;
    GetDataRec gdr = { this, native, &ret };

    AwtToolkit::GetInstance().WaitForSingleObject(m_mutex);

    AwtToolkit::GetInstance().InvokeFunctionLater(_GetData, &gdr);

    WaitUntilSignalled(FALSE);

    return ret;
}

/**
 * _GetData
 */

void AwtDropTarget::_GetData(void* param) {
    GetDataPtr gdrp = (GetDataPtr)param;

    *(gdrp->ret) = gdrp->dropTarget->GetData(gdrp->native);

    gdrp->dropTarget->Signal();
}


/**
 * GetData
 */

const char *const PredefinedClipboardNames[CF_MAX] = {
 	"",
	"TEXT",
	"BITMAP",
	"METAFILEPICT",
	"SYLK",
	"DIF",	
	"TIFF",
	"OEM TEXT",
	"DIB",
	"PALETTE",
	"PENDATA",
	"RIFF",
	"WAVE",
	"UNICODE TEXT",
	"ENHMETAFILE",
	"HDROP",
	"LOCALE"
};

jobject AwtDropTarget::GetData(jstring native) {
          JNIEnv*  env    = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
          jboolean isCopy;
    const char*   name    = env->GetStringUTFChars(native, &isCopy);
          UINT     fmt    = 0;
	  UINT*	   fmtp   = &fmt;
	  HRESULT  res    = S_OK;
	  jobject  stream = (jobject)NULL;

    for (int i = 1; i < CF_MAX; i++) {
	if (strcmp(name, PredefinedClipboardNames[i]) == 0) {
	    fmt = i; 
	    break;
	}
    }

    if (!fmt) fmt = ::RegisterClipboardFormat((LPCTSTR)name);

    env->ReleaseStringUTFChars(native, name);

    CachePtr* cpp = (CachePtr*)bsearch((const void *)&fmtp,
				       (const void *)m_formats,
				       (size_t)      m_nformats,
				       (size_t)	     sizeof(CachePtr),
						     _compar
		    );

    if (cpp == (CachePtr*)NULL) {
	return (jobject)NULL;
    }

    CachePtr cp = *cpp;

    FORMATETC format = cp->m_formatEtc;
    STGMEDIUM stgmedium;

    format.tymed |= TYMED_HGLOBAL | TYMED_FILE | TYMED_ISTREAM;

    if ((res = m_dataObject->GetData(&format, &stgmedium)) != S_OK) {
	return (jobject)NULL;
    }

    switch (stgmedium.tymed) {
	case TYMED_HGLOBAL: {
	    signed char* buffer = (signed char *)::GlobalLock(stgmedium.hGlobal);
	    if (format.cfFormat != CF_HDROP) {
	        DWORD        size   = ::GlobalSize(stgmedium.hGlobal);

		if (size == 0 && ::GetLastError() != 0) {
		    ::SetLastError(0); // clear error

		    ::GlobalUnlock(stgmedium.hGlobal);
		    ::ReleaseStgMedium(&stgmedium);

		    return (jobject)NULL; // failed
		}

	        jbyteArray   arry   = env->NewByteArray((jsize)size);

	        if (JNU_IsNull(env, arry)) {
		    ::GlobalUnlock(stgmedium.hGlobal);
		    ::ReleaseStgMedium(&stgmedium);

		    return (jobject)NULL;
	        }

	        arry = (jbyteArray)env->NewGlobalRef((jobject)arry);

	        env->SetByteArrayRegion(arry, 0, size, buffer);

	        ::GlobalUnlock(stgmedium.hGlobal);
	        ::ReleaseStgMedium(&stgmedium);

	        stream = env->CallStaticObjectMethod(dTCClazz, dTCgetms, arry, native);

	        if (JNU_IsNull(env, stream) || !JNU_IsNull(env, env->ExceptionOccurred())) {
		    env->DeleteGlobalRef(arry);

		    return (jobject)NULL;
	        }

	        stream = env->NewGlobalRef(stream);
	    } else { // its an HDROP
		UINT nFiles = ::DragQueryFile((HDROP)buffer, (UINT)0xffffffff, (LPSTR)NULL, (UINT)0);

		if (nFiles == 0) {
	            ::GlobalUnlock(stgmedium.hGlobal);
	            ::ReleaseStgMedium(&stgmedium);

		    return (jobject)NULL;
		}

		jobjectArray files = env->NewObjectArray((jsize)nFiles, file, (jobject)NULL);

		if (JNU_IsNull(env, files) || !JNU_IsNull(env, env->ExceptionOccurred())) {
	            ::GlobalUnlock(stgmedium.hGlobal);
	            ::ReleaseStgMedium(&stgmedium);

		    return (jobject)NULL;
		}

		files = (jobjectArray)env->NewGlobalRef(files);

		if (JNU_IsNull(env, files) || !JNU_IsNull(env, env->ExceptionOccurred())) {
	            ::GlobalUnlock(stgmedium.hGlobal);
	            ::ReleaseStgMedium(&stgmedium);

		    return (jobject)NULL;
		}

		for (UINT i = 0; i < nFiles; i++) {
		    UINT blen = ::DragQueryFile((HDROP)buffer, i, (LPSTR)NULL, (UINT)0) + 1;

		    static char buf[512];

		    LPSTR bp = (LPSTR)(blen > sizeof(buf) ? calloc(sizeof(char), blen) : buf);

		    ::DragQueryFile((HDROP)buffer, i, bp, blen);

		    jobject fileName = env->NewStringUTF((const char *)bp);

		    if (bp != buf) free((void *)bp);

		    if (JNU_IsNull(env, fileName) || !JNU_IsNull(env, env->ExceptionOccurred())) {
	                ::GlobalUnlock(stgmedium.hGlobal);
	                ::ReleaseStgMedium(&stgmedium);

		        return (jobject)NULL;
		    }

		    jobject fileObj = env->NewObject(file, dTCfilei, env->NewGlobalRef(fileName));

		    if (JNU_IsNull(env, fileObj) || !JNU_IsNull(env, env->ExceptionOccurred())) {
	                ::GlobalUnlock(stgmedium.hGlobal);
	                ::ReleaseStgMedium(&stgmedium);

		        return (jobject)NULL;
		    }

		    env->SetObjectArrayElement(files, (jsize)i, env->NewGlobalRef(fileObj));
		    if (!JNU_IsNull(env, env->ExceptionOccurred())) {
	                ::GlobalUnlock(stgmedium.hGlobal);
	                ::ReleaseStgMedium(&stgmedium);

		        return (jobject)NULL;
		    }
		}

	        ::GlobalUnlock(stgmedium.hGlobal);
	        ::ReleaseStgMedium(&stgmedium);

		files = (jobjectArray)(env)->CallStaticObjectMethod(dTCClazz, dTCgetl, files);

		if (JNU_IsNull(env, files) || !JNU_IsNull(env, env->ExceptionOccurred())) {
		    return (jobject)NULL;
		}

		return env->NewGlobalRef(files);
	    }
	}
        break;

	case TYMED_FILE: {
	    jstring fileName = (jstring)env->NewGlobalRef(env->NewStringUTF((const char *)stgmedium.lpszFileName));

	    STGMEDIUM* stgm = (STGMEDIUM*)calloc(sizeof(STGMEDIUM), 1);
	
	    if (stgm == (STGMEDIUM*)NULL) {
	        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
		::ReleaseStgMedium(&stgmedium);
	    } else
		*stgm = stgmedium;

	    stream = env->CallStaticObjectMethod(dTCClazz, dTCgetfs, fileName, native, stgm);

	    if (JNU_IsNull(env, stream) || !JNU_IsNull(env, env->ExceptionOccurred())) {
		env->DeleteGlobalRef(fileName);

		free((void*)stgm);

		::ReleaseStgMedium(&stgmedium);

		return (jobject)NULL;
	    }
	}
	break;

	case TYMED_ISTREAM: {
	   WDTCPIStreamWrapper* istream = new WDTCPIStreamWrapper(&stgmedium); 

	   stream = env->CallStaticObjectMethod(dTCClazz, dTCgetis, (jint)istream, native);

	    if (JNU_IsNull(env, stream) || !JNU_IsNull(env, env->ExceptionOccurred())) {
		istream->Close();
		return (jobject)NULL;
	    }
	}
	break;

	case TYMED_GDI:
	case TYMED_MFPICT:
	case TYMED_ISTORAGE:
	case TYMED_ENHMF:
	default:
	    ::ReleaseStgMedium(&stgmedium);

	    return (jobject)NULL;
    }

    return stream;
}

/**
 *
 */

int __cdecl AwtDropTarget::_compar(const void* first, const void* second) {
    CachePtr fp = *((CachePtr*)first);
    CachePtr sp = *((CachePtr*)second);

    return fp->m_formatEtc.cfFormat - sp->m_formatEtc.cfFormat;
}

const unsigned int AwtDropTarget::CACHE_INCR = 16;

void AwtDropTarget::LoadCache(IDataObject* pDataObj) {
    JNIEnv*      env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    unsigned int cnt = 0;
    HRESULT  	 res;

    if (m_dataObject != (IDataObject*)NULL) UnloadCache();

    (m_dataObject = pDataObj)->AddRef();

    res = m_dataObject->EnumFormatEtc(DATADIR_GET, &m_enumFormatEtc);

    if (res == S_OK) {
	m_enumFormatEtc->AddRef();
    } else {
	m_dataObject->Release();
	m_dataObject    = (IDataObject*)NULL;
	m_enumFormatEtc = (IEnumFORMATETC*)NULL;
	return;
    }

    for (;;) {
	
	FORMATETC tmp;
	ULONG	  actual = 1;
	jobject   atomName;

	res = m_enumFormatEtc->Next((ULONG)1, &tmp, &actual);

	if (res == S_FALSE) break;

	if (!(tmp.cfFormat  >= 1                &&
	      tmp.ptd       == NULL	        &&
	      tmp.lindex    == -1   	        &&
	      tmp.dwAspect  == DVASPECT_CONTENT &&
	      (tmp.tymed    == TYMED_HGLOBAL    ||
	       tmp.tymed    == TYMED_FILE       ||
	       tmp.tymed    == TYMED_ISTREAM
	      ) // but not ISTORAGE , GDI, MFPICT, ENHMF
	     )
 	) continue;

	if (m_dataObject->QueryGetData(&tmp) != S_OK) continue;

	if (tmp.cfFormat < CF_MAX)
	        atomName = env->NewGlobalRef(env->NewStringUTF(PredefinedClipboardNames[tmp.cfFormat]));
	else {
	    char 	     buf[512];
	    unsigned int nlen;

	    nlen = ::GetClipboardFormatName(tmp.cfFormat, (LPSTR)buf, sizeof(buf));

	    if (nlen != 0) {
	        atomName = env->NewGlobalRef(env->NewStringUTF(buf));
	    } else {
	        continue;
	    }
	}

	if (m_nformats % CACHE_INCR == 0) {
	    m_formats = (CachePtr*)realloc(m_formats, (CACHE_INCR + m_nformats) * sizeof(CachePtr));

	    if (m_formats == (CachePtr*)NULL) {
	        JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
	        return;
	    }
	}

	CachePtr cp = (CachePtr)calloc(sizeof(CacheRec), 1);

	if (cp == (CachePtr)NULL) {
	    JNU_ThrowOutOfMemoryError(env, "OutOfMemoryError");
	    return;
	}

	cp->m_formatEtc = tmp;
	cp->m_atomName  = atomName;

	m_formats[m_nformats++] = cp;
    }

    if (m_nformats > 0)
	qsort((void*)m_formats, m_nformats, sizeof(CachePtr), AwtDropTarget::_compar);


    m_natives = env->NewObjectArray(m_nformats, string, NULL);

    if (JNU_IsNull(env, (jobject)m_natives)) return;

    m_natives = (jobjectArray)env->NewGlobalRef(m_natives);

    for (unsigned int i = 0; i < m_nformats; i++) {
	env->SetObjectArrayElement(m_natives, i, m_formats[i]->m_atomName);
    }
}

/**
 * UnloadCache
 */

void AwtDropTarget::UnloadCache() {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (m_dataObject == (IDataObject*)NULL) return;

    for (unsigned int i = 0; i < m_nformats; i++) {
	env->DeleteGlobalRef(m_formats[i]->m_atomName);

	free((void*)m_formats[i]);
    }

    free((void*)m_formats);

    m_formats  = (CachePtr*)NULL;
    m_nformats = 0;

    env->DeleteGlobalRef(m_natives);
    m_natives = NULL;

    m_enumFormatEtc->Release();
    m_enumFormatEtc = (IEnumFORMATETC*)NULL;

    m_dataObject->Release();
    m_dataObject = (IDataObject*)NULL;
}

/**
 * DragCleanup
 */

void AwtDropTarget::DragCleanup(void) {
    UnloadCache();
}

/*****************************************************************************/

jclass WDTCPIStreamWrapper::javaIOExceptionClazz = (jclass)NULL;

/**
 * construct a wrapper
 */

WDTCPIStreamWrapper::WDTCPIStreamWrapper(STGMEDIUM* stgmedium) {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    m_stgmedium = *stgmedium;
    m_istream   = stgmedium->pstm;
    m_mutex     = ::CreateMutex(NULL, FALSE, NULL);

    if (javaIOExceptionClazz == (jclass)NULL) {
	javaIOExceptionClazz = env->FindClass("java/io/IOException");

        if (JNU_IsNull(env, javaIOExceptionClazz)) {
            env->ThrowNew(env->FindClass("java/lang/ClassNotFoundException"),
    		          "Cant find java/io/IOException"
            );
    	}
    }
}

/**
 * destroy a wrapper
 */

WDTCPIStreamWrapper::~WDTCPIStreamWrapper() {
    ::CloseHandle(m_mutex);

    ::ReleaseStgMedium(&m_stgmedium);
}

/**
 * return available data
 */

jint WDTCPIStreamWrapper::DoAvailable(WDTCPIStreamWrapper* istream) {
    WDTCPIStreamWrapperRec iswr = { istream, 0 };

    AwtToolkit::GetInstance().WaitForSingleObject(istream->m_mutex);

    AwtToolkit::GetInstance().InvokeFunctionLater( _Available, &iswr);

    istream->WaitUntilSignalled(FALSE);

    return iswr.ret;
}

/**
 * return available data
 */

void WDTCPIStreamWrapper::_Available(void *param) {
    WDTCPIStreamWrapperPtr iswrp = (WDTCPIStreamWrapperPtr)param;

    iswrp->ret = (iswrp->istream)->Available();

    iswrp->istream->Signal();
}

/**
 * return available data
 */

jint WDTCPIStreamWrapper::Available() {
    JNIEnv* env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if (m_istream->Stat(&m_statstg, STATFLAG_NONAME) != S_OK) {
	env->ThrowNew(javaIOExceptionClazz, "IStream::Stat() failed");
	return 0;
    }

    if (m_statstg.cbSize.QuadPart > 0x7ffffffL) {
	env->ThrowNew(javaIOExceptionClazz, "IStream::Stat() cbSize > 0x7ffffff");
	return 0;
    }

    return (jint)m_statstg.cbSize.LowPart;
}

/**
 * read 1 byte
 */

jint WDTCPIStreamWrapper::DoRead(WDTCPIStreamWrapper* istream) {
    WDTCPIStreamWrapperRec iswr = { istream, 0 };

    AwtToolkit::GetInstance().WaitForSingleObject(istream->m_mutex);

    AwtToolkit::GetInstance().InvokeFunctionLater(_Read, &iswr);

    istream->WaitUntilSignalled(FALSE);

    return iswr.ret;
}

/**
 * read 1 byte
 */

void WDTCPIStreamWrapper::_Read(void* param) {
    WDTCPIStreamWrapperPtr iswrp = (WDTCPIStreamWrapperPtr)param;

    iswrp->ret = (iswrp->istream)->Read();

    iswrp->istream->Signal();
}

/**
 * read 1 byte
 */

jint WDTCPIStreamWrapper::Read() {
    JNIEnv* env    = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jint    b      = 0;
    ULONG   actual = 0;
    HRESULT res;

    switch (res = m_istream->Read((void *)&b, (ULONG)1, &actual)) {
	case S_FALSE:
	    return (jint)-1;

	case S_OK:
	    return (jint)(actual == 0 ? -1 : b);

	default:
	    env->ThrowNew(javaIOExceptionClazz, "IStream::Read failed");
    }
    return (jint)-1;
}

/**
 * read Buffer
 */

jint WDTCPIStreamWrapper::DoReadBytes(WDTCPIStreamWrapper* istream, jbyteArray array, jint off, jint len) {
    WDTCPIStreamWrapperReadBytesRec iswrbr = { istream, 0, array, off, len };

    AwtToolkit::GetInstance().WaitForSingleObject(istream->m_mutex);

    AwtToolkit::GetInstance().InvokeFunctionLater(_ReadBytes, &iswrbr);

    istream->WaitUntilSignalled(FALSE);

    return iswrbr.ret;
}

/**
 * read buffer
 */

void WDTCPIStreamWrapper::_ReadBytes(void*  param) {
    WDTCPIStreamWrapperReadBytesPtr iswrbrp = (WDTCPIStreamWrapperReadBytesPtr)param;

	iswrbrp->ret = (iswrbrp->istream)->ReadBytes(iswrbrp->array, iswrbrp->off, iswrbrp->len);

	
	iswrbrp->istream->Signal();
}

/**
 * read buffer
 */

jint WDTCPIStreamWrapper::ReadBytes(jbyteArray buf, jint off, jint len) {
    JNIEnv*  env     = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jboolean isCopy  = JNI_FALSE;
    ULONG    actual  = 0;
    jbyte*   local   = env->GetByteArrayElements(buf, &isCopy);
    HRESULT  res;

    switch (res = m_istream->Read((void *)(local + off), (ULONG)len, &actual)) {
	case S_FALSE:
	case S_OK: {
	    int eof = (actual == 0);

       	    env->ReleaseByteArrayElements(buf, local, !eof ? 0 : JNI_ABORT);
	    return (jint)(!eof ? actual : -1);
	}

	default:
       	    env->ReleaseByteArrayElements(buf, local, JNI_ABORT);
	    env->ThrowNew(javaIOExceptionClazz, "IStream::Read failed");
    }

    return (jint)-1;
}

/**
 * close
 */

void WDTCPIStreamWrapper::DoClose(WDTCPIStreamWrapper* istream) {
    AwtToolkit::GetInstance().InvokeFunctionLater(_Close, istream);
}

/**
 * close
 */

void WDTCPIStreamWrapper::_Close(void* param) {
    ((WDTCPIStreamWrapper*)param)->Close();
}

/**
 * close
 */

void WDTCPIStreamWrapper::Close() {
    delete this;
}

/*****************************************************************************/

extern "C" {

/**
 * awt_dnd_initialize: initial DnD system
 */

void awt_dnd_initialize() {
    ::OleInitialize((LPVOID)NULL);
}

/**
 * awt_dnd_uninitialize: deactivate DnD system
 */

void awt_dnd_uninitialize() {
    ::OleUninitialize();
}

/**
 * convertActionsToDROPEFFECT
 */

DWORD convertActionsToDROPEFFECT(jint actions) {
    DWORD effects = DROPEFFECT_NONE;

    if (actions & java_awt_dnd_DnDConstants_ACTION_LINK) effects |= DROPEFFECT_LINK;
    if (actions & java_awt_dnd_DnDConstants_ACTION_MOVE) effects |= DROPEFFECT_MOVE;
    if (actions & java_awt_dnd_DnDConstants_ACTION_COPY) effects |= DROPEFFECT_COPY;
    return effects;
}

/**
 * convertDROPEFFECTToAction
 */

jint convertDROPEFFECTToActions(DWORD effects) {
    jint actions = java_awt_dnd_DnDConstants_ACTION_NONE;

    if (effects & DROPEFFECT_LINK) actions |= java_awt_dnd_DnDConstants_ACTION_LINK;
    if (effects & DROPEFFECT_MOVE) actions |= java_awt_dnd_DnDConstants_ACTION_MOVE;
    if (effects & DROPEFFECT_COPY) actions |= java_awt_dnd_DnDConstants_ACTION_COPY;

    return actions;
}

/**
 * map keyboard modifiers to a DROPEFFECT
 */

DWORD mapModsToDROPEFFECT(DWORD effects, DWORD mods) {
    DWORD ret = DROPEFFECT_NONE;

    switch (mods & (MK_CONTROL | MK_SHIFT)) {
	case MK_CONTROL:
	    ret = DROPEFFECT_COPY;
	break;

	case MK_CONTROL | MK_SHIFT:
	    ret = DROPEFFECT_LINK;
	break;

	default:
	    ret = DROPEFFECT_MOVE;
    }

    return ret & effects;
}

/**
 *
 */

JNIEXPORT void JNICALL Java_sun_awt_windows_WDropTargetContextPeer_Initialize(JNIEnv* env, jclass cls) {
    AwtDropTarget::InitAwtDropTarget();
}

/**
 * downcall to fetch data ... gets scheduled on message thread
 */

JNIEXPORT jobject JNICALL Java_sun_awt_windows_WDropTargetContextPeer_getData(JNIEnv* env, jobject self, jint dropTarget, jstring native) {
    return ((AwtDropTarget*)dropTarget)->DoGetData(native);
}

/**
 * downcall to signal drop done ... gets scheduled on message thread
 */

JNIEXPORT void JNICALL Java_sun_awt_windows_WDropTargetContextPeer_dropDone(JNIEnv* env, jobject self, jint dropTarget, jint actions) {
    ((AwtDropTarget*)dropTarget)->DoDropDone(actions);
}

/**
 * downcall to free up storage medium for FileStream
 */

JNIEXPORT void JNICALL Java_sun_awt_windows_WDropTargetContextPeerFileStream_freeStgMedium(JNIEnv* env, jobject self, jint stgmedium) {
    ::ReleaseStgMedium((STGMEDIUM*)stgmedium);

    free((void*)stgmedium);
}

/**
 * 
 */

JNIEXPORT jint JNICALL Java_sun_awt_windows_WDropTargetContextPeerIStream_Available(JNIEnv* env, jobject self, jobject istream) {
    return WDTCPIStreamWrapper::DoAvailable((WDTCPIStreamWrapper*)istream);
}

/**
 * 
 */

JNIEXPORT jint JNICALL Java_sun_awt_windows_WDropTargetContextPeerIStream_Read(JNIEnv* env, jobject self, jobject istream) {
    return WDTCPIStreamWrapper::DoRead((WDTCPIStreamWrapper*)istream);
}

/**
 * 
 */

JNIEXPORT jint JNICALL Java_sun_awt_windows_WDropTargetContextPeerIStream_ReadBytes(JNIEnv* env, jobject self, jobject istream, jbyteArray buf, jint off, jint len) {
    return WDTCPIStreamWrapper::DoReadBytes((WDTCPIStreamWrapper*)istream, buf, off, len);
}

/**
 * 
 */

JNIEXPORT void JNICALL Java_sun_awt_windows_WDropTargetContextPeerIStream_Close(JNIEnv* env, jobject self, jobject istream) {
    WDTCPIStreamWrapper::DoClose((WDTCPIStreamWrapper*)istream);
}

} /* extern "C" */
