/*
 * @(#)awt_DnDDT.h	1.7 98/08/25
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_DND_DT_H
#define AWT_DND_DT_H

#include <Ole2.h>

#include <jni.h>
#include <jni_util.h>

#include "awt_Object.h"
#include "awt_Component.h"
#include "awt_window.h"

extern "C" void awt_dnd_initialize();

/**
 * AwtDropTarget class: native peer IDropTarget implementation
 */

class AwtDropTarget : virtual public IDropTarget {
    public:
	AwtDropTarget(JNIEnv* env, AwtComponent* component);

	virtual ~AwtDropTarget();

	// IUnknown

	virtual HRESULT __stdcall QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject);

	virtual ULONG   __stdcall AddRef(void);
	virtual ULONG   __stdcall Release(void);

	// IDropTarget

	virtual HRESULT __stdcall DragEnter(IDataObject __RPC_FAR *pDataObject, DWORD grfKeyState, POINTL pt, DWORD __RPC_FAR *pdwEffect);
	virtual HRESULT __stdcall DragOver(DWORD grfKeyState, POINTL pt, DWORD __RPC_FAR *pdwEffect);
	virtual HRESULT __stdcall DragLeave(void);

	virtual HRESULT __stdcall Drop(IDataObject __RPC_FAR *pDataObject, DWORD grfKeyState, POINTL pt, DWORD __RPC_FAR *pdwEffect);

	// AwtDropTarget

	static void InitAwtDropTarget();

	virtual jobject DoGetData(jstring native);

	virtual void DoDropDone(jint actions);

	INLINE void Signal() { ::ReleaseMutex(m_mutex); }

    protected:

	INLINE void WaitUntilSignalled(BOOL retain) {
	    do {
		// nothing ...
	    } while (::WaitForSingleObject(m_mutex, INFINITE) == WAIT_FAILED);

	    if (!retain) ::ReleaseMutex(m_mutex);
	}

	virtual void RegisterTarget(WORD wParam);

	virtual jobject GetData(jstring native);

	virtual void DropDone(jint actions);

	virtual void DragCleanup(void);

	virtual void LoadCache(IDataObject*);

	virtual void UnloadCache();

    private:
	static int DeferredRegisterTarget(AwtComponent*, UINT, void*);

	typedef struct _RegisterTargetRec {
	    AwtDropTarget*	dropTarget;
	    BOOL		show;
	} RegisterTargetRec, *RegisterTargetPtr;

	static void _RegisterTarget(void* param);

	typedef struct _GetDataRec {
	    AwtDropTarget* dropTarget;
	    jstring	   native;
	    jobject*	   ret;
	} GetDataRec, *GetDataPtr;

	static void _GetData(void* param);

	typedef struct _DropDoneRec {
	    AwtDropTarget* dropTarget;
	    jint	   action;
	} DropDoneRec, *DropDonePtr;

	static void _DropDone(void* param);
	
	typedef struct _CacheRec {
	    FORMATETC	m_formatEtc;
	    jobject	m_atomName;
	} CacheRec, *CachePtr;


       	AwtComponent*	      m_component;
	HWND		      m_window;
	jobject		      m_target;

	unsigned int	      m_refs;

	jobject		      m_dtcp;

	WORD	      	      m_registered; // is drop site registered?

	CachePtr*	      m_formats;
	unsigned int	      m_nformats;

 	jobjectArray	      m_natives;

	jint		      m_dropActions;

	HANDLE		      m_mutex;

	// external COM references

	IDataObject    __RPC_FAR *m_dataObject;
	IEnumFORMATETC __RPC_FAR *m_enumFormatEtc;

	// static members

	static jclass dTCClazz; // sun.awt.windows.WDropTargetContextPeer

	static jclass string;   // java.lang.string
	 
	static jclass file;     // java.io.File
	 
	static const char dTCcbmethodSig[];

	// method references

	static jmethodID dTCgetp;
	static jmethodID dTCenter;
	static jmethodID dTCexit;
	static jmethodID dTCmotion;
	static jmethodID dTCdrop;
	 
	static jmethodID dTCgetms;
	static jmethodID dTCgetfs;
	static jmethodID dTCgetis;
	static jmethodID dTCfilei;
	static jmethodID dTCgetl;

	static const unsigned int CACHE_INCR;

	static int __cdecl _compar(const void *, const void *);
};


/**
 * WDTCPIStreamWrapper: cheap wrapper class for incoming IStream drops, maps
 * onto WDropTargetContextPeerIStream class
 */

class WDTCPIStreamWrapper {
    public:
	WDTCPIStreamWrapper(STGMEDIUM* stgmedium);

	virtual ~WDTCPIStreamWrapper();

	static jint DoAvailable(WDTCPIStreamWrapper* istream);
	static jint DoRead(WDTCPIStreamWrapper* istream);
	static jint DoReadBytes(WDTCPIStreamWrapper* istream, jbyteArray buf, jint off, jint len);
	static void DoClose(WDTCPIStreamWrapper* istream);


	virtual jint Available();
	virtual jint Read();
	virtual jint ReadBytes(jbyteArray buf, jint off, jint len);
	virtual void Close();

	INLINE void Signal() { ::ReleaseMutex(m_mutex); }
   protected:

	INLINE void WaitUntilSignalled(BOOL retain) {
	    do {
		// nothing ...
	    } while (::WaitForSingleObject(m_mutex, INFINITE) == WAIT_FAILED);

	    if (!retain) ::ReleaseMutex(m_mutex);
	}

	typedef struct _WDTCPIStreamWrapperRec {
	    WDTCPIStreamWrapper* istream;
	    jint		 ret;
	} WDTCPIStreamWrapperRec, *WDTCPIStreamWrapperPtr;

	static void _Available(void* param);

	static void _Read     (void* Param);

	typedef struct _WDTCPIStreamWrapperReadBytesRec {
	    WDTCPIStreamWrapper* istream;
	    jint		 ret;
	    jbyteArray		 array;
	    jint		 off;
	    jint		 len;
	} WDTCPIStreamWrapperReadBytesRec, *WDTCPIStreamWrapperReadBytesPtr;

	static void _ReadBytes(void* param);

	static void _Close    (void* param);

    private:
	IStream*	m_istream;
	STGMEDIUM	m_stgmedium;
	STATSTG		m_statstg;
	HANDLE		m_mutex;

	static jclass javaIOExceptionClazz;
};

#endif /* AWT_DND_DT_H */
