/*
 * @(#)awt_DrawingSurface.cpp	1.9 98/03/18
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_DrawingSurface.h"

#include "awt_Component.h"
#include "awt_Image.h"
#include "awt_Graphics.h"
#include "sun_awt_windows_WDrawingSurfaceInfo.h"


/************************************************************************
 * AwtDrawingSurface fields
 */

jfieldID AwtDrawingSurface::pDataID;
jfieldID AwtDrawingSurface::stateID;
jfieldID AwtDrawingSurface::wID;
jfieldID AwtDrawingSurface::hID;
jfieldID AwtDrawingSurface::peerID;


/************************************************************************
 * WDrawingSurfaceInfo native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_initIDs(JNIEnv *env, jclass cls) 
{
    AwtDrawingSurface::pDataID = env->GetFieldID(cls, "pData", "I");
    AwtDrawingSurface::stateID = env->GetFieldID(cls, "state", "I");
    AwtDrawingSurface::wID = env->GetFieldID(cls, "w", "I");
    AwtDrawingSurface::hID = env->GetFieldID(cls, "h", "I");
    AwtDrawingSurface::peerID = 
	env->GetFieldID(cls, "peer", "Lsun/awt/windows/WComponentPeer;");
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    lock
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_lock(JNIEnv *env, jobject self) 
{
    //jobject imgrep;

    /* REMIND check peer self? */
    AwtSharedDC *pSharedDC = NULL;
    jobject hComp = env->GetObjectField(self, AwtDrawingSurface::peerID);

    char *str = NULL;
    long state = env->GetIntField(self, AwtDrawingSurface::stateID);

    if (hComp != NULL) {
	AwtComponent *pComp = (AwtComponent *)(JNI_GET_PDATA(hComp));
	jobject hJComp = env->GetObjectField(hComp, AwtObject::targetID);

	if (pComp == NULL || hJComp == NULL) {
	    str = "null component";
	} else {
	    if (::IsWindow(pComp->GetHWnd()) == FALSE) {
		str = "disposed component";
	    } else {
		pSharedDC = pComp->GetSharedDC();
		int w = env->GetIntField(self, AwtDrawingSurface::wID);
		int h = env->GetIntField(self, AwtDrawingSurface::hID);
		int hJCompW = env->GetIntField(hJComp, AwtComponent::widthID);
		int hJCompH = env->GetIntField(hJComp, AwtComponent::heightID);

		if (w != hJCompW || h != hJCompH) {
		    env->SetIntField(self, AwtDrawingSurface::wID, hJCompW);
		    env->SetIntField(self, AwtDrawingSurface::hID, hJCompH);

		    int state = env->GetIntField(self, 
						 AwtDrawingSurface::stateID);
		    env->SetIntField(self, AwtDrawingSurface::stateID, 
				     state++);
		}
	    }
	}
//     } else if ((imgrep = env->GetObjectField(self,AwtDrawingSurface::imgrepID))
// 	       != NULL) {

// 	/* REMIND !!! DEREF */
// 	ImageRep *imgrepHdl = (ImageRep *)DeRef(env, imgrep);
// 	AwtImage *pImage = AwtImage::GetpImage(imgrepHdl);

//         pSharedDC = pImage->m_pSharedDC;
    } else {
	pSharedDC = NULL;
    }

    if (pSharedDC == NULL) {
	JNU_ThrowNullPointerException(env, str);
	return state;
    }

    pSharedDC->Attach();
    pSharedDC->Lock();
    pSharedDC->GrabDC(NULL);
    env->SetIntField(self, AwtDrawingSurface::pDataID, (jint)pSharedDC);

    return state;
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    unlock
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_unlock(JNIEnv *env, jobject self)
{
    AwtSharedDC *pSharedDC = (AwtSharedDC *)(env->GetIntField(self, AwtDrawingSurface::pDataID));
    if (pSharedDC == NULL) {
	JNU_ThrowNullPointerException(env, "null DC");
        return;
    }
    pSharedDC->Unlock();
    pSharedDC->Detach((AwtGraphics *) 1); // Can't be null...
    env->SetIntField(self, AwtDrawingSurface::pDataID, 0);
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    getHWnd
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_getHWnd(JNIEnv *env, jobject self) 
{
    AwtSharedDC *pSharedDC = (AwtSharedDC *)(env->GetIntField(self, AwtDrawingSurface::pDataID));
    if (pSharedDC == NULL) {
	JNU_ThrowNullPointerException(env, "null DC");
	return NULL;
    }
    
    return (long) pSharedDC->GetHWnd();
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    getHBitmap
 * Signature: ()I
 */
JNIEXPORT jint JNICALL 
Java_sun_awt_windows_WDrawingSurfaceInfo_getHBitmap(JNIEnv *env, jobject self) 
{
//     jobject imgrep = env->GetObjectField(self, AwtDrawingSurface::imgrepID);
//     if (imgrep == NULL) {
//         return NULL;
//     }
//     ImageRep *imgrepHdl = (ImageRep *)(DeRef(env, imgrep));
//     AwtImage *pImage = AwtImage::GetpImage(imgrepHdl);
//     return (long) pImage->m_hBitmap;
    return NULL;
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    getPBits
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_getPBits(JNIEnv *env, jobject self)
{
//     jobject imgrep = env->GetObjectField(self, AwtDrawingSurface::imgrepID);
//     if (imgrep == NULL) {
//         return NULL;
//     }
//     ImageRep *imgrepHdl = (ImageRep *)(DeRef(env, imgrep));
//     AwtImage *pImage = AwtImage::GetpImage(imgrepHdl);
//     return (long) pImage->m_hBitmap;
    return NULL;
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    getHDC
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_getHDC(JNIEnv *env, jobject self)
{
    AwtSharedDC *pSharedDC = (AwtSharedDC *)(env->GetIntField(self, AwtDrawingSurface::pDataID));
    if (pSharedDC == NULL) {
	JNU_ThrowNullPointerException(env, "null DC");
        return NULL;
    }
    return (long) pSharedDC->GetDC();
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    getDepth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_getDepth(JNIEnv *env, jobject self)
{
    return AwtImage::GetDepth();
}

/*
 * Class:     sun_awt_windows_WDrawingSurfaceInfo
 * Method:    getHPalette
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WDrawingSurfaceInfo_getHPalette(JNIEnv *env, 
						     jobject self)
{
    return (long)AwtImage::m_hPal;
}

} /* extern "C" */
