/*
 * @(#)awt_FileDialog.cpp	1.14 98/10/02
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt.h"

#include "awt_FileDialog.h"

#include "awt_Dialog.h"

#include <commdlg.h>


/************************************************************************
 * AwtFileDialog fields
 */

/* WFileDialogPeer ids */
jfieldID AwtFileDialog::parentID;
jmethodID AwtFileDialog::handleSelectedMID;
jmethodID AwtFileDialog::handleCancelMID;

/* FileDialog ids */
jfieldID AwtFileDialog::modeID;
jfieldID AwtFileDialog::dirID;
jfieldID AwtFileDialog::fileID;
jfieldID AwtFileDialog::filterID;
jmethodID AwtFileDialog::setDirectoryID;
jmethodID AwtFileDialog::setFileID;


/************************************************************************
 * WFileDialogPeer native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_sun_awt_windows_WFileDialogPeer_initIDs(JNIEnv *env, jclass cls)
{
    AwtFileDialog::parentID = 
	env->GetFieldID(cls, "parent","Lsun/awt/windows/WComponentPeer;");
    AwtFileDialog::handleSelectedMID = 
	env->GetMethodID(cls, "handleSelected", "(Ljava/lang/String;)V");
    AwtFileDialog::handleCancelMID = 
	env->GetMethodID(cls, "handleCancel", "()V");

    /* java.awt.FileDialog fields */
    cls = env->FindClass("java/awt/FileDialog");
    if (cls == NULL) {
	return;
    }
    AwtFileDialog::modeID = env->GetFieldID(cls, "mode", "I");
    AwtFileDialog::dirID = env->GetFieldID(cls, "dir", "Ljava/lang/String;");
    AwtFileDialog::fileID = env->GetFieldID(cls, "file", "Ljava/lang/String;");
    AwtFileDialog::filterID = 
	env->GetFieldID(cls, "filter", "Ljava/io/FilenameFilter;");
    AwtFileDialog::setDirectoryID = 
        env->GetMethodID(cls, "setDirectory_NoClientCode",
			 "(Ljava/lang/String;)V");
    AwtFileDialog::setFileID = 
        env->GetMethodID(cls, "setFile_NoClientCode", "(Ljava/lang/String;)V");

    ASSERT(AwtFileDialog::parentID != NULL);
    ASSERT(AwtFileDialog::handleSelectedMID != NULL);
    ASSERT(AwtFileDialog::handleCancelMID != NULL);

    ASSERT(AwtFileDialog::modeID != NULL);
    ASSERT(AwtFileDialog::dirID != NULL);
    ASSERT(AwtFileDialog::fileID != NULL);
    ASSERT(AwtFileDialog::filterID != NULL);
    ASSERT(AwtFileDialog::setDirectoryID != NULL);
    ASSERT(AwtFileDialog::setFileID != NULL);
}

JNIEXPORT void JNICALL
Java_sun_awt_windows_WFileDialogPeer_show(JNIEnv *env, jobject peer)
{
    ASSERT(peer);
    jobject target = env->GetObjectField(peer, AwtObject::targetID);
    jobject parent = env->GetObjectField(peer, AwtFileDialog::parentID); 
    AwtComponent* awtParent = NULL;
    if (parent != NULL) {
        awtParent = (AwtComponent *)JNI_GET_PDATA(parent);
    }
    ASSERT(awtParent);
    jstring title = (jstring)(env)->GetObjectField(target, AwtDialog::titleID);
    char* titleBuffer = NULL;
    if (title != NULL) {  
	int length = env->GetStringLength(title);
	titleBuffer = new char[length + 1];
	AwtFont::javaString2multibyte(env, title, titleBuffer, length);
    }

    char *directoryBuffer = NULL;
    jstring directory = 
	(jstring)env->GetObjectField(target, AwtFileDialog::dirID);
    if (directory != NULL) {
	int length = env->GetStringLength(directory);
	directoryBuffer = new char[length + 1];
	AwtFont::javaString2multibyte(env, directory, directoryBuffer, length);
    }

    char* fileBuffer = new char[MAX_PATH+1];
    jstring file = (jstring)env->GetObjectField(target, AwtFileDialog::fileID);
    if (file != NULL) {
	int length = env->GetStringLength(file);
	AwtFont::javaString2multibyte(env, file, fileBuffer, length);
    } else {
	fileBuffer[0] = '\0';
    }
    // No one knows what a filename filter is yet (Macs prevent just
    // using extensions), so accept all files for now.
    const char filter[] = "All Files (*.*)\0*.*\0\0";

    OPENFILENAME ofn;
    memset(&ofn, 0, sizeof(ofn));

    ofn.lStructSize = sizeof(ofn);
    ofn.lpstrFilter = filter;
    ofn.nFilterIndex = 1;
    ofn.hwndOwner = awtParent ? awtParent->GetHWnd() : NULL;
    ofn.lpstrFile = fileBuffer;
    ofn.nMaxFile = MAX_PATH;
    ofn.lpstrTitle = titleBuffer;
    ofn.lpstrInitialDir = directoryBuffer;
    ofn.Flags = OFN_LONGNAMES | OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY;

    // Save current directory, so we can reset if it changes.
    char* currentDirectory = new char[MAX_PATH+1];
    VERIFY(::GetCurrentDirectory(MAX_PATH, currentDirectory) > 0);

    BOOL result;

    jint mode = env->GetIntField(target, AwtFileDialog::modeID);

    // disable all top-level windows
    AwtDialog::ModalDisable(NULL);
    // show the Win32 file dialog
    if (mode == java_awt_FileDialog_LOAD) {
	result = GetOpenFileName(&ofn);
    } else {
	result = GetSaveFileName(&ofn);
    }
    // re-enable top-level windows
    AwtDialog::ModalEnable(NULL);
    AwtDialog::ModalNextWindowToFront(ofn.hwndOwner);

    VERIFY(::SetCurrentDirectory(currentDirectory));
    
    // Report result to peer.
    if (result) {
	env->CallVoidMethod(peer, AwtFileDialog::handleSelectedMID,
	                    AwtFont::multibyte2javaString(env,ofn.lpstrFile)); 
    } else {
	env->CallVoidMethod(peer, AwtFileDialog::handleCancelMID);
    }
    ASSERT(!env->ExceptionOccurred());

    delete[] currentDirectory;
    delete[] fileBuffer;
    if (directoryBuffer) {
        delete[] directoryBuffer;
    }
    if (titleBuffer) {
        delete[] titleBuffer;
    }
}

JNIEXPORT void JNICALL
Java_sun_awt_windows_WFileDialogPeer_targetSetDirectory_1NoClientCode(
    JNIEnv *env,
    jobject peer,
    jobject target,
    jstring dir)
{
    env->CallVoidMethod(target, AwtFileDialog::setDirectoryID, dir);
}

JNIEXPORT void JNICALL
Java_sun_awt_windows_WFileDialogPeer_targetSetFile_1NoClientCode(
    JNIEnv *env,
    jobject peer,
    jobject target,
    jstring file)
{
    env->CallVoidMethod(target, AwtFileDialog::setFileID, file);
}

} /* extern "C" */
