/*
 * @(#)awt_Frame.cpp	1.80 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_Frame.h"
#include "awt_MenuBar.h"

#include <sun_awt_windows_WEmbeddedFrame.h>
#include <sun_awt_windows_WEmbeddedFramePeer.h>

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * inner class AwtNativeActivationEvent methods
 */

class AwtFrame::AwtNativeActivationEvent : public AwtActiveEvent
{
public:
    AwtNativeActivationEvent(AwtFrame *frame, HWND prevWnd)
        : AwtActiveEvent(),
	  m_frame(frame),
	  m_prevWnd(prevWnd) {}
    virtual ~AwtNativeActivationEvent() {}
    virtual AwtActiveEvent *Clone() { 
        return new AwtNativeActivationEvent(m_frame, m_prevWnd);
    }
    virtual void Dispatch() {
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	if (env->EnsureLocalCapacity(2) < 0) {
	    return;
	}

	jobject peer =
	  JNU_CallMethodByName(env, NULL, m_frame->GetPeer(env),
			       "getFocusPeer",
			       "()Lsun/awt/windows/WComponentPeer;").l;

	ASSERT(!env->ExceptionOccurred());
	if (peer == NULL) {
	    // Set default focus.
	    JNU_CallMethodByName(env, NULL, m_frame->GetPeer(env),
				 "setDefaultFocus", "()V");
	    ASSERT(!env->ExceptionOccurred());
	} else {
	    AwtComponent* c = (AwtComponent *)JNI_GET_PDATA(peer);
	    if (c != NULL) {
	        c->SendMessage(WM_AWT_COMPONENT_SETFOCUS,
			       (WPARAM) m_frame->GetHWnd(), 
			       (LPARAM) m_prevWnd);
	    }
	}

	env->DeleteLocalRef(peer);
    }

private:
    AwtFrame *m_frame;
    HWND m_prevWnd;
};

/************************************************************************
 * AwtFrame fields
 */

jfieldID AwtFrame::handleID;
jfieldID AwtFrame::stateID;
jmethodID AwtFrame::isCursorAllowedID;


/************************************************************************
 * AwtFrame methods
 */

AwtFrame::AwtFrame() {
    m_hIcon = NULL;
    m_parentWnd = NULL;
    menuBar = NULL;
    m_isEmbedded = FALSE;
    m_extraHeight = 0;
    m_ignoreWmSize = FALSE;
    m_isMenuDropped = FALSE;
}

AwtFrame::~AwtFrame() {
    if (m_hIcon != NULL) {
        ::DestroyIcon(m_hIcon);
    }
}

const char* AwtFrame::GetClassName() {
  return "SunAwtFrame";
}

/*
 * Create a new AwtFrame object and window.
 */
AwtFrame* AwtFrame::Create(jobject self, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    int pData;
    HWND hwndParent = NULL;
    AwtFrame* frame;
    jclass cls = 0;
    jobject target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "target", done);

    if (parent != NULL) {
	JNI_CHECK_PEER_GOTO(parent, done);
	{
	    AwtFrame* parent = (AwtFrame *)pData;
	    hwndParent = parent->GetHWnd();
	}
    }

    frame = new AwtFrame();
    JNI_CHECK_NULL_GOTO(frame, "AwtFrame() failed", done);

    {
        /*
	 * A variation on Netscape's hack for embedded frames: the client area
	 * of the browser is a Java Frame for parenting purposes, but really a
	 * Windows child window
	 */
        cls = env->FindClass("sun/awt/EmbeddedFrame");
	if (cls == NULL) {
	    return NULL;
	}
	jint handle;
	jboolean isEmbeddedInstance = env->IsInstanceOf(target, cls);
	jboolean isEmbedded = FALSE;

	if (isEmbeddedInstance) {
	    handle = env->GetIntField(target, AwtFrame::handleID);
	    if (handle != 0) {
	        isEmbedded = TRUE;
	    }
	}
	frame->m_isEmbedded = isEmbedded;

	if (isEmbedded) {
            hwndParent = (HWND)handle;
	    RECT rect;
	    ::GetClientRect(hwndParent, &rect);
	    frame->RegisterClass();
	    DWORD exStyle = WS_EX_NOPARENTNOTIFY;

	    if (GetRTL()) {
	        exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
		if (GetRTLReadingOrder())
		    exStyle |= WS_EX_RTLREADING;
	    }

	    frame->m_hwnd = ::CreateWindowEx(exStyle,
					     frame->GetClassName(), "",
					     WS_CHILD | WS_CLIPCHILDREN, 0, 0,
					     rect.right, rect.bottom,
					     hwndParent, NULL,
					     ::GetModuleHandle(NULL), NULL);
	    frame->LinkObjects(env, self);
	    frame->SubclassWindow();

	    /* Update target's dimensions to reflect this embedded window. */
	    ::GetClientRect(frame->m_hwnd, &rect);
	    ::MapWindowPoints(frame->m_hwnd, hwndParent, (LPPOINT)&rect, 2);

	    env->SetIntField(target, AwtComponent::xID, rect.left);
	    env->SetIntField(target, AwtComponent::yID, rect.top);
	    env->SetIntField(target, AwtComponent::widthID,
			     rect.right-rect.left);
	    env->SetIntField(target, AwtComponent::heightID,
			     rect.bottom-rect.top);
	} else {
	    jint state = env->GetIntField(target, AwtFrame::stateID);
	    DWORD exStyle = WS_EX_WINDOWEDGE;

	    if (GetRTL()) {
	        exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
		if (GetRTLReadingOrder())
		    exStyle |= WS_EX_RTLREADING;
	    }

	    frame->CreateHWnd(env, "",
			      WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN |
			      ((state == java_awt_Frame_ICONIFIED) ? WS_ICONIC
			                                           : 0),
			      exStyle,
			      0, 0, 0, 0,
			      hwndParent,
			      NULL,
			      ::GetSysColor(COLOR_WINDOWTEXT),
			      ::GetSysColor(COLOR_WINDOWFRAME),
			      self);

	    jint x = env->GetIntField(target, AwtComponent::xID);
	    jint y = env->GetIntField(target, AwtComponent::yID);
	    jint width = env->GetIntField(target, AwtComponent::widthID);
	    jint height = env->GetIntField(target, AwtComponent::heightID);

	    /* 
	     * Reshape here instead of during create, so that a WM_NCCALCSIZE
	     * is sent. 
	     */
	    frame->Reshape(x, y, width, height);
	}
    }
 done:
    env->DeleteLocalRef(target);
    env->DeleteLocalRef(cls);

    return frame;
}

/*
 * Override AwtComponent's Reshape to keep this frame in place if it is
 * an embedded 'frame'
 */
void AwtFrame::Reshape(int x, int y, int width, int height)
{
    if (IsEmbedded()) {
	x = y = 0;
    }
    AwtWindow::Reshape(x, y, width, height);
}

/* Show the frame, and activate it. */
void AwtFrame::Show()
{
    if (isIconic()) {
        SendMessage(WM_AWT_COMPONENT_SHOW, SW_SHOWMINIMIZED);
    } else {
        SendMessage(WM_AWT_COMPONENT_SHOW, SW_SHOW);
    }
}

MsgRouting AwtFrame::WmSize(int type, int w, int h)
{
    if (m_ignoreWmSize) {
        return mrDoDefault;
    }
    else {
        return AwtWindow::WmSize(type, w, h);
    }
}

MsgRouting AwtFrame::WmSetCursor(HWND hWnd, UINT hitTest, UINT message,
                                 BOOL& retVal)
{
    /* Don't pass WM_SETCURSOR messages on to an embedded frame's parent. */
    if(IsEmbedded()) {
	// Don't pass WM_SETCURSOR messages on to an embedded frame's parent
	// unless EmbeddedFrame.isCursorAllowed() is true
	BOOL    isCursorAllowed;
	JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	isCursorAllowed = (BOOL)env->CallBooleanMethod(GetTarget(env),isCursorAllowedID);
	if (isCursorAllowed) {
	    ::SetCursor(GetCursor(TRUE));
	    return mrDoDefault;
	} else {
	    retVal = FALSE;
	    return mrConsume;
	}
    }
    return AwtWindow::WmSetCursor(hWnd, hitTest, message, retVal);
}

MsgRouting AwtFrame::WmActivate(UINT nState, BOOL fMinimized)
{
    if (nState == WA_INACTIVE) {
        SendWindowEvent(java_awt_event_WindowEvent_WINDOW_DEACTIVATED);
        return mrDoDefault;
    }
    else {
        SendWindowEvent(java_awt_event_WindowEvent_WINDOW_ACTIVATED);
	AwtNativeActivationEvent(this, ::GetFocus()).SendEvent(this);
	return mrConsume;
    }
}

MsgRouting AwtFrame::WmEnterMenuLoop(BOOL isTrackPopupMenu)
{
    if ( !isTrackPopupMenu ) {
	m_isMenuDropped = TRUE;
    }
    return mrDoDefault;
}

MsgRouting AwtFrame::WmExitMenuLoop(BOOL isTrackPopupMenu)
{
    if ( !isTrackPopupMenu ) {
	m_isMenuDropped = FALSE;
    }
    return mrDoDefault;
}

AwtMenuBar* AwtFrame::GetMenuBar()
{
    return menuBar;
}

void AwtFrame::SetMenuBar(AwtMenuBar* mb)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0) {
        return;
    }

    jobject frame = GetTarget(env);
    jclass frameClass = env->GetObjectClass(frame);

    jmethodID resizeMID = env->GetMethodID(frameClass, "resize", "(II)V");

    menuBar = mb;
    if (mb == NULL) {
        // Remove existing menu bar, if any.
        ::SetMenu(GetHWnd(), NULL);

	if (m_extraHeight != 0) {
	    RECT outside;
	    ::GetWindowRect(GetHWnd(), &outside);

//	    m_ignoreWmSize = TRUE;
	    ::MoveWindow( GetHWnd(),
			  outside.left,
			  outside.top,
			  outside.right - outside.left,
			  outside.bottom - m_extraHeight - outside.top,
			  TRUE);
//	    m_ignoreWmSize = FALSE;
/*
	    env->CallVoidMethod(frame, resizeMID, outside.right - outside.left,
				outside.bottom - m_extraHeight - outside.top);
*/
	    m_extraHeight = 0;
	}
    } else {
        if (menuBar->GetHMenu() != NULL) {
            // Save the size of original window and client rectangle
            RECT outside;
            RECT inside;
            ::GetWindowRect(GetHWnd(), &outside);
            ::GetClientRect(GetHWnd(), &inside);
            ::MapWindowPoints(GetHWnd(), 0, (LPPOINT)&inside, 2);

            CriticalSection& l = menuBar->GetLock();
	    l.Enter();

            ::SetMenu(GetHWnd(), menuBar->GetHMenu());

            // Get the size of the new client rectangle with the menu bar
            RECT insidewithmenu;
            ::GetClientRect(GetHWnd(), &insidewithmenu);
            ::MapWindowPoints(GetHWnd(), 0, (LPPOINT)&insidewithmenu, 2);

            // If the top of the client rectangle intrudes below the
            // bottom of the window rectangle, we have no choice but to
            // make the window bigger so that insets are computed correctly.
            if (insidewithmenu.top > inside.top && m_extraHeight == 0) {
		RECT exposed = { outside.left, outside.bottom, outside.right,
				 outside.bottom + insidewithmenu.top - 
				 inside.top };

//		m_ignoreWmSize = TRUE;
	        ::MoveWindow( GetHWnd(),
			      outside.left,
			      outside.top,
			      outside.right - outside.left,
			      exposed.bottom - outside.top,
			      TRUE);
//		m_ignoreWmSize = FALSE;
		
		m_extraHeight = insidewithmenu.top - inside.top;

		l.Leave();
/*		
		env->CallVoidMethod(frame, resizeMID, 
				    outside.right - outside.left,
				    exposed.bottom - outside.top);
*/
	    }
	    else {
	        l.Leave();
	    }
        }
    }

    env->DeleteLocalRef(frameClass);
    env->DeleteLocalRef(frame);
}

MsgRouting AwtFrame::WmDrawItem(UINT ctrlId, DRAWITEMSTRUCT far& drawInfo)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    // if the item to be redrawn is the menu bar, then do it
    AwtMenuBar* awtMenubar = GetMenuBar();
    if (drawInfo.CtlType == ODT_MENU && (awtMenubar != NULL) &&
        (::GetMenu( GetHWnd() ) == (HMENU)drawInfo.hwndItem) )
	{
		awtMenubar->DrawItem(drawInfo);
		return mrConsume;
    }

	return AwtComponent::WmDrawItem(ctrlId, drawInfo);
}

MsgRouting AwtFrame::WmMeasureItem(UINT ctrlId, MEASUREITEMSTRUCT far& measureInfo)
{
	JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	AwtMenuBar* awtMenubar = GetMenuBar();
	if ((measureInfo.CtlType == ODT_MENU) && (awtMenubar != NULL))
	{
		// AwtMenu instance is stored in itemData. Use it to check if this
		// menu is the menu bar.
		AwtMenu * pMenu = (AwtMenu *) measureInfo.itemData;
		ASSERT(pMenu != NULL);
		if ( pMenu == awtMenubar )
		{
			HWND hWnd = GetHWnd();
			HDC hDC = ::GetDC(hWnd);
			ASSERT(hDC != NULL);
			awtMenubar->MeasureItem(hDC, measureInfo);
			VERIFY(::ReleaseDC(hWnd, hDC));
			return mrConsume;
		}
	}

	return AwtComponent::WmMeasureItem(ctrlId, measureInfo);
}

typedef struct tagBitmapheader  {
    BITMAPINFOHEADER bmiHeader;
    DWORD            dwMasks[256];
}   Bitmapheader, *LPBITMAPHEADER;

static
HBITMAP create_mask(int hW,int hH)
{
    HBITMAP     hbmRC = NULL;
    int         numBytes = (hW * hH + 7) / 8;
    BYTE *pMask = new BYTE[numBytes];
    memset(pMask, 0, numBytes);
    hbmRC = ::CreateBitmap(hW, hH, 1, 1, pMask);
    ::GdiFlush();
    delete [] pMask;
    return hbmRC;
}

static
HBITMAP create_DIB_BMP(HWND hW,int* imageData,int nSS, int nW, int nH)
{
    Bitmapheader    bmhHeader;
    HDC             hDC;
    char            *ptrImageData;
    HBITMAP         hbmpBitmap;

    int             nNumChannels    = 3;

    if (!hW) {
        hW = ::GetDesktopWindow();
    }
    hDC = ::GetDC(hW);
    if (!hDC) {
        return NULL;
    }

    bmhHeader.bmiHeader.biSize              = sizeof(BITMAPINFOHEADER);
    bmhHeader.bmiHeader.biWidth             = nW;
    bmhHeader.bmiHeader.biHeight            = -nH;
    bmhHeader.bmiHeader.biPlanes            = 1;

    bmhHeader.bmiHeader.biBitCount          = 24;
    bmhHeader.bmiHeader.biCompression       = BI_RGB;

    bmhHeader.bmiHeader.biSizeImage         = 0;
    bmhHeader.bmiHeader.biXPelsPerMeter     = 0;
    bmhHeader.bmiHeader.biYPelsPerMeter     = 0;
    bmhHeader.bmiHeader.biClrUsed           = 0;
    bmhHeader.bmiHeader.biClrImportant      = 0;

    hbmpBitmap = ::CreateDIBSection(hDC, (BITMAPINFO*)&(bmhHeader),
                                    DIB_RGB_COLORS,
                                    (void**)&(ptrImageData),
                                    NULL, 0);
    int  *srcPtr = imageData;
    char *dstPtr = ptrImageData;
    if (!dstPtr) {
        return NULL;
    }
    for (int nOutern = 0; nOutern < nH; nOutern++ ) {
        for (int nInner = 0; nInner < nSS; nInner++ ) {
            dstPtr[2] = (*srcPtr >> 0x10) & 0xFF;
            dstPtr[1] = (*srcPtr >> 0x08) & 0xFF;
            dstPtr[0] = *srcPtr & 0xFF;

            srcPtr++;
            dstPtr += nNumChannels;
        }
    }
    ::ReleaseDC(hW, hDC);
    ::GdiFlush();
    return hbmpBitmap;
}

static
void destroy_BMP(HBITMAP hBMP)
{
    if (hBMP) {
        ::DeleteObject(hBMP);
    }
}

void AwtFrame::MakeSetIcon(int *pINTRGBImageBuffer, int nSS, int nW, int nH)
{
    if (::IsIconic(GetHWnd())) {
        ::InvalidateRect(GetHWnd(), NULL, TRUE);
    }
    if (m_hIcon != NULL) {
        ::DestroyIcon(m_hIcon);
        m_hIcon = NULL;
    }

    if (pINTRGBImageBuffer) {
        HBITMAP     hMask = create_mask(nW, nH);
        HBITMAP     hColor = create_DIB_BMP(m_parentWnd,
                                            pINTRGBImageBuffer, nSS, nW, nH);
        if ( hMask && hColor) {
            ICONINFO icnInfo;
            memset(&icnInfo, 0, sizeof(ICONINFO));
            icnInfo.hbmMask = hMask;
            icnInfo.hbmColor = hColor;
            m_hIcon = ::CreateIconIndirect(&icnInfo);

            destroy_BMP(hColor);
            destroy_BMP(hMask);
        }
    }

    // Save in a local variable, so the toolkit's icon doesn't get deleted.
    HICON hIcon = m_hIcon;

    if (hIcon == NULL) {
        // Get AWT icon.
        hIcon = AwtToolkit::GetInstance().GetAwtIcon();
    }

    // If hIcon is NULL, Windows will use its default icon.
    SendMessage(WM_SETICON, ICON_BIG, (long)hIcon);
    SendMessage(WM_SETICON, ICON_SMALL, (long)hIcon);
    // set icon on owned windows
    ::EnumThreadWindows(
	    AwtToolkit::MainThread(),
	    (WNDENUMPROC)OwnedSetIcon, 
            (LPARAM)this);
}

// set icon on windows owned by this one
BOOL CALLBACK AwtFrame::OwnedSetIcon(HWND hWndDlg, LPARAM lParam)
{
    AwtFrame * frame = (AwtFrame *)lParam;
    HWND    hwndDlgOwner = ::GetWindow(hWndDlg, GW_OWNER);

    while (hwndDlgOwner != NULL) {
	if (hwndDlgOwner == frame->GetHWnd()) {
	    break;
	}
        hwndDlgOwner = ::GetWindow(hwndDlgOwner, GW_OWNER);
    }

    if (hwndDlgOwner == NULL)
	return TRUE;

    // if owned window has an icon, change it to match the frame
    if (::SendMessage(hWndDlg, WM_GETICON, ICON_SMALL, NULL) != NULL) {
	::SendMessage(hWndDlg, WM_SETICON, ICON_BIG, (long)frame->m_hIcon);
	::SendMessage(hWndDlg, WM_SETICON, ICON_SMALL, (long)frame->m_hIcon);
    }

    return TRUE;
}


MsgRouting AwtFrame::WmSysCommand(UINT uCmdType, UINT xPos, UINT yPos)
{
    if (uCmdType == (SYSCOMMAND_IMM & 0xFFF0)){
      JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
      if (env->EnsureLocalCapacity(2) < 0) {
	return mrConsume;
      }
      JNU_CallMethodByName(env, NULL, m_peerObject, "notifyIMMOptionChange", 
			   "()V");
      ASSERT(!env->ExceptionOccurred());  
      return mrConsume;
    }
    return AwtWindow::WmSysCommand(uCmdType, xPos, yPos);
}

LONG AwtFrame::WinThreadExecProc(ExecuteArgs * args)
{
    switch( args->cmdId ) {
	case FRAME_SETMENUBAR:
	{
    	    jobject  mbPeer = (jobject)args->param1;

	    // cancel any currently dropped down menus
	    if (m_isMenuDropped) {
		SendMessage(WM_CANCELMODE);
	    }

	    if (mbPeer == NULL) {
		// Remove existing menu bar, if any
		SetMenuBar(NULL);
	    } else {
		JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
		AwtMenuBar* menuBar = (AwtMenuBar *)JNI_GET_PDATA(mbPeer);
		SetMenuBar(menuBar);
	    }
	    DrawMenuBar();
	    break;
	}

	default:
	    AwtWindow::WinThreadExecProc(args);
	    break;
    }

    return 0L;
}

/************************************************************************
 * WFramePeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_Frame_initIDs(JNIEnv *env, jclass cls)
{
    AwtFrame::stateID = env->GetFieldID(cls, "state", "I");
    ASSERT(AwtFrame::stateID != NULL);
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    setState
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFramePeer_setState(JNIEnv *env, jobject self,
                                                jint state)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);

    AwtFrame* w = (AwtFrame*)pData;
    ASSERT(!IsBadReadPtr(w, sizeof(AwtFrame)));

    if (::IsWindowVisible(w->GetHWnd())) {
        if (state == java_awt_Frame_NORMAL) {
            ::ShowWindow(w->GetHWnd(), SW_RESTORE);
	} else {
            ::ShowWindow(w->GetHWnd(), SW_MINIMIZE);
	}
    } else { 
        if (state == java_awt_Frame_NORMAL) {
           w->setIconic(FALSE);
	} else {
           w->setIconic(TRUE);
        }
    }
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    getState
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WFramePeer_getState(JNIEnv *env, jobject self)
{
    jint pData;
    JNI_CHECK_PEER_RETURN_NULL(self);

    AwtFrame* w = (AwtFrame*)pData;
    ASSERT(!IsBadReadPtr(w, sizeof(AwtFrame)));

    if (w->isIconic()) {
	return java_awt_Frame_ICONIFIED;
    } else {
	return java_awt_Frame_NORMAL;
    }
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    setMenuBar0
 * Signature: (Lsun/awt/windows/WMenuBarPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFramePeer_setMenuBar0(JNIEnv *env, jobject self,
					    jobject mbPeer)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtObject::WinThreadExec(self, AwtFrame::FRAME_SETMENUBAR, (LPARAM)mbPeer);
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFramePeer_create(JNIEnv *env, jobject self,
				       jobject parent)
{
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtFrame::Create);
    int pData;
    JNI_CHECK_PEER_CREATION_RETURN(self);
}
/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    getSysIconHeight
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WFramePeer_getSysIconHeight(JNIEnv *env, jclass self)
{
    return ::GetSystemMetrics(SM_CYICON);
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    getSysIconWidth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WFramePeer_getSysIconWidth(JNIEnv *env, jclass self)
{
    return ::GetSystemMetrics(SM_CXICON);
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    setIconImageFromIntRasterData
 * Signature: ([I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFramePeer_setIconImageFromIntRasterData
  (JNIEnv *env, jobject self, jintArray intRasterData, jint scanStride, jint nW, jint nH)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(intRasterData, "intRasterData argument");

    AwtFrame* frame = (AwtFrame *)pData;

    /* REMIND how does this make sense? -br */
    if (intRasterData == NULL) {
        frame->MakeSetIcon(NULL,(int)scanStride,(int)nW,(int)nH);
    } else {
        jint *intRasterDataPtr = (jint *)env->GetPrimitiveArrayCritical(intRasterData, 0);
        if( NULL == intRasterDataPtr ){
            JNI_CHECK_NULL_RETURN(intRasterDataPtr,"intRasterData data");
        } else  {
            frame->MakeSetIcon((int*)intRasterDataPtr,(int)scanStride,(int)nW,(int)nH);
        }
        env->ReleasePrimitiveArrayCritical(intRasterData, intRasterDataPtr, 0);
    }
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    pSetIMMOption
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFramePeer_pSetIMMOption(JNIEnv *env, jobject self,
					       jstring option) 
{
    jint pData;
    char *coption;
    char *empty = "InputMethod";
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(option, "IMMOption argument");

    AwtFrame* frame = (AwtFrame *)pData;

    coption = (JNU_IsNull(env, option)) ? empty : (char *) JNU_GetStringPlatformChars(env, option, NULL);
    HMENU hSysMenu = ::GetSystemMenu(frame->GetHWnd(), FALSE);
    ::AppendMenu(hSysMenu,  MF_STRING, SYSCOMMAND_IMM, coption);

    if (coption != empty)
        JNU_ReleaseStringPlatformChars(env, option, (const char *) coption);
}

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    reshape
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WFramePeer_reshape(JNIEnv *env, jobject self,
					jint x, jint y, jint w, jint h) 
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);

    if (env->EnsureLocalCapacity(1) < 0) {
	return;
    }

    AwtFrame* p = (AwtFrame *)pData;

    jobject target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_RETURN(target, "null target");

    // enforce tresholds before sending the event
    int minWidth = ::GetSystemMetrics(SM_CXMIN);
    int minHeight = ::GetSystemMetrics(SM_CYMIN);
    if (w < minWidth)
	env->SetIntField(target, AwtComponent::widthID,
			 w = minWidth);
    if (h < minHeight)
	env->SetIntField(target, AwtComponent::heightID,
			 h = minHeight);

    env->DeleteLocalRef(target);

    RECT* r = new RECT;
    ::SetRect(r, x, y, x + w, y + h);
    p->SendMessage(WM_AWT_RESHAPE_COMPONENT, 0, (LPARAM)r);
}

} /* extern "C" */


/************************************************************************
 * WEmbeddedFrame native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WFramePeer
 * Method:    initIDs
 * Signature: (Lsun/awt/windows/WMenuBarPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WEmbeddedFrame_initIDs(JNIEnv *env, jclass cls)
{
    AwtFrame::handleID = env->GetFieldID(cls, "handle", "I");
    ASSERT(AwtFrame::handleID != NULL);
    AwtFrame::isCursorAllowedID = env->GetMethodID(cls, "isCursorAllowedImpl", "()Z");
    ASSERT(AwtFrame::isCursorAllowedID != NULL);
}

/*
 * Class:     sun_awt_windows_WEmbeddedFrame
 * Method:    setThisPeer
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WEmbeddedFrame_setThisPeer(JNIEnv *env, jobject self,
                                                jobject lpeer)
{
    jclass cls;
    jfieldID fid;

    cls = env->GetObjectClass(self);
    fid = env->GetFieldID(cls, "peer", "Ljava/awt/peer/ComponentPeer;");
    env->SetObjectField(self, fid, lpeer);
}

} /* extern "C" */


/************************************************************************
 * WEmbeddedFramePeer native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_sun_awt_windows_WEmbeddedFramePeer_create(JNIEnv *env, jobject self,
					       jobject parent)
{
    JNI_CHECK_NULL_RETURN(self, "peer");
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtFrame::Create);
    int pData;
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

} /* extern "C" */
