/*
 * @(#)awt_Frame.h	1.33 98/09/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_FRAME_H
#define AWT_FRAME_H

#include "awt_Window.h"
#include "awt_ActiveEvent.h"
#include "awt_MenuBar.h" //add for multifont

#include "java_awt_Frame.h"
#include "sun_awt_windows_WFramePeer.h"
#include <awt_image.h>


/************************************************************************
 * AwtFrame class
 */

class AwtFrame : public AwtWindow {
public:
    enum FrameExecIds {
	FRAME_SETMENUBAR
    };

    /* int handle field for sun.awt.windows.WEmbeddedFrame */
    static jfieldID handleID;

    /* int state field for java.awt.Frame */
    static jfieldID stateID;

    /* int handle for sun.awt.EmbeddedFrame.isCursorAllowed method */
    static jmethodID isCursorAllowedID;

    AwtFrame();
    virtual ~AwtFrame();

    virtual const char* GetClassName();
    
    /* Create a new AwtFrame.  This must be run on the main thread. */
    static AwtFrame* Create(jobject self, jobject parent);

    /* Returns whether this frame is embedded in an external native frame. */
    INLINE BOOL IsEmbedded() { return m_isEmbedded; }

    void Reshape(int x, int y, int width, int height);
    void Show();

    INLINE void DrawMenuBar() { VERIFY(::DrawMenuBar(GetHWnd())); }

    void MakeSetIcon(int* intArrayPtr,int ss, int w, int h);

    /*for WmDrawItem and WmMeasureItem method */
    AwtMenuBar* GetMenuBar();
    void SetMenuBar(AwtMenuBar*);

    MsgRouting WmSize(int type, int w, int h);
    MsgRouting WmSetCursor(HWND hWnd, UINT hitTest, UINT message, 
                           BOOL& retVal);
    MsgRouting WmActivate(UINT nState, BOOL fMinimized);
    MsgRouting WmDrawItem(UINT ctrlId, DRAWITEMSTRUCT far& drawInfo);
    MsgRouting WmMeasureItem(UINT ctrlId, MEASUREITEMSTRUCT far& measureInfo);
    MsgRouting WmEnterMenuLoop(BOOL isTrackPopupMenu);
    MsgRouting WmExitMenuLoop(BOOL isTrackPopupMenu);

    virtual MsgRouting WmSysCommand(UINT uCmdType, UINT xPos, UINT yPos);

    LONG WinThreadExecProc(ExecuteArgs * args);

private:
    static BOOL CALLBACK OwnedSetIcon(HWND hWnd, LPARAM lParam);
  /* The frame's icon and its size.  If NULL, just paint the HotJava
       logo. */
    HICON m_hIcon;

    /* The frame's embedding parent (if any) */
    HWND m_parentWnd;

    /* The frame's menubar. */
    AwtMenuBar* menuBar;

    /* The frame is an EmbeddedFrame. */
    BOOL m_isEmbedded;

    /* height added by SetMenuBar */
    int m_extraHeight;

    /* used so that calls to ::MoveWindow in SetMenuBar don't propogate
       because they are immediately followed by calls to Component.resize */
    BOOL m_ignoreWmSize;

    /* tracks whether or not menu on this frame is dropped down */
    BOOL m_isMenuDropped;

//inner classes
    class AwtNativeActivationEvent;
    friend class AwtNativeActivationEvent;
};

#endif /* AWT_FRAME_H */
