/*
 * @(#)awt_GDIObject.h	1.7 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_GDIOBJECT_H
#define AWT_GDIOBJECT_H

#include "awt.h"
#include "Hashtable.h"

/*
 * An AwtGDIObject is a cached, color-based GDI object, such as a pen or
 * brush.
 */
class AwtGDIObject {
public:
    INLINE COLORREF GetColor() { return m_color; }
    INLINE void SetColor(COLORREF color) { m_color = color; }

    INLINE HGDIOBJ GetHandle() { return m_handle; }
    INLINE void SetHandle(HGDIOBJ handle) { m_handle = handle; }

    INLINE int GetRefCount() { return m_refCount; }
    INLINE int IncrRefCount() {
        return ::InterlockedIncrement((LPLONG) &m_refCount);
    }
    INLINE int DecrRefCount() { 
        return ::InterlockedDecrement((LPLONG) &m_refCount);
    }

    /*
     * Decrement the reference count of a cached GDI object.  When it hits 
     * zero, delete both the GDI object and this wrapper.
     */
    INLINE void Release() {
#if MEMORY_OVER_SPEED
        if (DecrRefCount() == 0) {
            RemoveFromCache();
            delete this;
        }
#endif
    }

protected:
    /*
     * Get a GDI object from its respective cache.  If it doesn't exist
     * it gets created, otherwise its reference count gets bumped.
     */
    static AwtGDIObject* Get(COLORREF color);

    virtual void RemoveFromCache() = 0;

    INLINE AwtGDIObject() { 
        m_handle = NULL;
        m_refCount = 0;
    }

    virtual ~AwtGDIObject() {
        if (m_handle != NULL) {
            ::DeleteObject(m_handle);
        }
    }

private:
    COLORREF m_color;
    HGDIOBJ  m_handle;
    int      m_refCount;
};

#endif // AWT_GDIOBJECT_H
