/*
 * @(#)awt_Graphics.h	1.98 98/09/09
 *
 * Copyright 1995-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
#ifndef _AWTGRAPHICS_H_
#define _AWTGRAPHICS_H_

#include "awt_image.h"
#include "awt_Brush.h"
#include "awt_Pen.h"

#include <ddraw.h>
#include <sun_java2d_loops_GraphicsPrimitive.h>

#include "SpanIterator.h"

extern int LogicalSelectClipRgn(HDC hDC, HRGN hClipRgn);


/************************************************************************
 * AwtSharedDC class
 */

class AwtSharedDC
{
public:

    INLINE AwtSharedDC(AwtComponent* pComponent, HDC hDC) {
        m_useCount = 1;
        m_pOwner = NULL;
        m_hDC = hDC;
        m_hOldBmp = NULL;
        m_pComponent = pComponent;
        ::InitializeCriticalSection(&m_lock);

#ifdef DEBUG
        m_lockCount = 0;
#endif
    }

    INLINE ~AwtSharedDC() {
        Lock();
	/* This should be a NOP, but just in case... */
	FreeDC();
        Unlock();
        ::DeleteCriticalSection(&m_lock);

    }

    /* Get the Window being drawn to  (NULL if offscreen graphics) */
    INLINE AwtComponent* GetComponent() {
        return m_pComponent;
    }
    INLINE HWND GetHWnd() {
        return (m_pComponent == NULL) ? NULL : m_pComponent->GetHWnd();
    }

    /* Get the Device Context without "grabbing it"
     * (only used as a source DC for off-screen blits)
     */
    INLINE HDC GetDC() {
	return m_hDC;
    }

    /* Grab this Device Context and assign a new "owner" */
    HDC GrabDC(class AwtGraphics *pNewOwner);

    /*
     * The current DC has become invalid -- release and re-get it
     * (workaround for NT 4.0 bug).
     */
    HDC ResetDC();

    /* Attach this Shared DC to some object (increment reference count) */
    INLINE void Attach() {
	++m_useCount;
    }

    /*
     * Detach this Shared DC from some object
     * (decrement reference count and free if necessary)
     */
    void Detach(class AwtGraphics *pOldOwner);

    INLINE void SetBitmap(HBITMAP hBM) {
	m_hOldBmp = (HBITMAP) ::SelectObject(m_hDC, hBM);
    }

    INLINE void UnsetBitmap() {
	if (m_hOldBmp != NULL) {
	    ::SelectObject(m_hDC, m_hOldBmp);
	}
    }

    /*
     * Lock/unlock this shared DC.  This is necessary so that one
     * thread isn't executing GDI calls when another thread has the
     * DC.
     */
    INLINE void Lock() {
        ::EnterCriticalSection(&m_lock);
#ifdef DEBUG
	ASSERT(m_lockCount >= 0);
	++m_lockCount;
#endif
    }
    INLINE void Unlock() {
#ifdef DEBUG
        ASSERT(m_lockCount > 0);
	--m_lockCount;
#endif
        ::LeaveCriticalSection(&m_lock);
    }

    /* Returns the mutex for this SharedDC. */
private:
    /* Get rid of the Windows DC object */
    void FreeDC();

    int			 m_useCount;
    class AwtGraphics	*m_pOwner;
    HDC			 m_hDC;
    HBITMAP		 m_hOldBmp;
    AwtComponent	*m_pComponent;
    CRITICAL_SECTION     m_lock;

#ifdef DEBUG
    int                  m_lockCount;
#endif
};


/************************************************************************
 * AwtGraphics class
 */

#define METHOD_INVALID  0x00
#define METHOD_DDRAW    0x01
#define METHOD_DIB      0x02
#define METHOD_GDI      0x03

typedef struct tagBitmapheader  {
    BITMAPINFOHEADER bmiHeader;
    DWORD            dwMasks[256];
}   Bitmapheader, *LPBITMAPHEADER;

class VideoInfo {
public :
    // common
    HDC hDC;

    // DDraw
    DDSURFACEDESC       sdPrimary;
    WORD                wVideoMethod;

    // DIB
    Bitmapheader    bmhHeader;
    HBITMAP         hbmpBitmap;
    HPALETTE        hPalette;
    unsigned char*  ptrImageData;

    // GDI ... nil

    //      Lock method
    int     nLockMethod;        //  Do we lock for read or write?

    //      Image related
    void*   vpData;             //  Address of data buffer
    int     nType;              //  Java image type.

    int     m_palSeed;
    unsigned long uLutData[256];//  Color lookup table
    long    lLutDataLength;     //  Length of CLUT

    int     nPixelStride;       //  Distance between pixels in pixel elements
    int     nScanlineStride;    //  Distance between pixel rows in pixelel
    int     nNumChannels;       //  Number of channels
    int     nBytePerChannel;    //  Bytes per channel
    int     nXSize;             //  X size
    int     nYSize;             //  Y size

    RECT    rOutput;

public:
    ~VideoInfo() {
    }
    VideoInfo() {
        memset(this, 0, sizeof(VideoInfo));
    }
};

class AwtGraphics : public AwtObject
{
public:
    /* java.awt.Color foreground field for sun.awt.windows.WGraphics */
    static jfieldID foregroundID;

    /* sun.awt.windows.WGraphics fields */
    static jfieldID pDataID;

    /* java.awt.Font getFont method for sun.awt.windows.WGraphics */
    static jmethodID getFontMID;

    AwtGraphics();
    virtual ~AwtGraphics();

   /*
    * Access methods
    */
    INLINE int GetRop() { return m_rop; }

   /*
    * Operations
    *
    * Creates a graphics context on 'pWindow'
    */
    static AwtGraphics *Create(JNIEnv *env, jobject wGraphics,
			       AwtComponent* pWindow);

    static AwtGraphics *Create(JNIEnv *env, jobject wGraphics, HDC hdc);

    static AwtGraphics *CreateCopy(JNIEnv *env, jobject wGraphics,
				   AwtGraphics *pSource);

    virtual void Dispose();

    INLINE AwtComponent* GetComponent() {
        if (m_pSharedDC == NULL) {return NULL;}
        return m_pSharedDC->GetComponent();
    }

    INLINE HWND GetHWnd() {
        if (m_pSharedDC == NULL) {return NULL;}
        return m_pSharedDC->GetHWnd();
    }

    INLINE HDC GetDC() {
        if (m_pSharedDC != NULL) {
            m_pSharedDC->Lock();
            if (m_hDC == NULL) {
                ValidateDC();
                if (m_hDC == NULL) {
                    m_pSharedDC->Unlock();
                }
            } else if (m_bClipping) {
		VERIFY(LogicalSelectClipRgn(m_hDC, m_hClipRgn) != ERROR);
            }
        }
        return m_hDC;
    }

    INLINE void ReleaseDC() {
        if (m_pSharedDC != NULL) {
            if (m_pSharedDC->GetComponent() != NULL) {
                // Bug #4055623 - Performance enhancements
                if (m_bClipping) {
                    LogicalSelectClipRgn(m_hDC, NULL);
                }
            }
            m_pSharedDC->Unlock();
        }
    }

    void ValidateDC();
    void InvalidateDC();

    /* Sets the m_color instance variable to the absolute COLORREF */
    void ChangeColor(COLORREF color);

    /*
     * Calculates the appropriate new COLORREF according to the
     * current ROP and calls ChangeColor on the new COLORREF
     */
    void SetColor(COLORREF color);

    void SetFont(AwtFont *pFont);

    void SetOrigin(jint tx, jint ty) {
	m_originX = tx;
	m_originY = ty;
    }

    INLINE AwtFont* GetFont() { return m_pFont; }

    void SetRop(int rop, COLORREF fgcolor, COLORREF color);

    void ChangeClip(long nX, long nY, long nW, long nH, BOOL set);
    void RemoveClip();

    /* Returns the clipping rectangle.  If there is none, return NULL. */
    RECT* GetClipRect();

    /*
     * Draws a line using the foreground color between points (nX1, nY1)
     * and (nX2, nY2).  The end-point is painted.
     */
    void DrawLine(JNIEnv* env, long nX1, long nY1, long nX2, long nY2);

    POINT* TransformPoly(long* xpoints, long* ypoints, POINT *points,
			 int *pNpoints, BOOL close);

    BOOL Polyline(long* xpoints, long* ypoints, int nPoints, BOOL close);

    BOOL FillPolygon(long* xpoints, long* ypoints, int nPoints, BOOL evenodd);

    void FillSpans(JNIEnv *env, jobject si, SpanIteratorFuncs *span);

    void ClearRect(JNIEnv* env, long nX, long nY, long nW, long nH);

    void FillRect(JNIEnv* env, long nX, long nY, long nW, long nH);

    void DrawRect(JNIEnv* env, long nX, long nY, long nW, long nH);

    void RoundRect(long nX, long nY, long nW, long nH, long nArcW, long nArcH);

    void FillRoundRect(long nX, long nY, long nW, long nH, long nArcW,
                       long nArcH);

    void Ellipse(long nX, long nY, long nW, long nH);

    void FillEllipse(long nX, long nY, long nW, long nH);

    /*
     * Converts an angle given in 1/64ths of a degree to a pair of
     * coordinates that intersects the line produced by the angle.
     */
    void AngleToCoords(long nAngle, long *nX, long *nY);

    void Arc(long nX, long nY, long nW, long nH, long nStartAngle,
             long nEndAngle);

    void FillArc(long nX, long nY, long nW, long nH, long nStartAngle,
                 long nEndAngle);

    INLINE void SetBrushColor(BOOL bSetBrush) {
        if (bSetBrush != m_bBrushSet) {
            ChangeBrush(bSetBrush);
        }
    };
    void ChangeBrush(BOOL bSetBrush);
    INLINE void SetPenColor(BOOL bSetPen) {
        if (bSetPen != m_bPenSet) {
            ChangePen(bSetPen);
        }
    };
    void ChangePen(BOOL bSetPen);

    /*
     * Draws the string at the specified point.  (nX, nY) designates
     * the reference point of the string - i.e. the top of the string
     * plus the ascent.  The width of the string is returned.
     */
    long DrawStringWidth(wchar_t *pzString, long nX, long nY);
    long DrawStringWidth(jstring string, jobject font, long nX, long nY);

    void CopyArea(long nDstX, long nDstY, long nX, long nY, long nW, long nH);

    /* Delete any AwtGraphics objects which were not disposed. */
    static void Cleanup();

    void PrintComponent(class AwtComponent* c);

    void SetJavaObject(JNIEnv *env, jobject obj);
    int GetOX() {
	return m_originX;
    }
    int GetOY() {
	return m_originY;
    }

// internal methods
    BOOL CanUseDirectDraw(void);
    void ReleaseDDResources(void);

// Variables
public:
    HDC m_hDC;
    class AwtSharedDC *m_pSharedDC;

protected:
    // TRUE iff the underlying DC was obtained from ::PrintDlg, FALSE
    // otherwise
    BOOL m_isPrinter;

private:
    BOOL m_bPaletteSelected;
    RECT m_rectClip;
    HRGN m_hClipRgn;
    BOOL m_bClipping;

    jint m_originX;
    jint m_originY;

    AwtFont *m_pFont;

    COLORREF m_color;
    COLORREF m_xorcolor;
    BOOL m_bBrushSet;
    AwtBrush* m_hBrush;
    BOOL m_bPenSet;
    AwtPen* m_hPen;
    int m_rop;

    jobject m_javaObject;

    HBITMAP     m_cachedDIB;
    BITMAPINFO  m_cachedDIBbi;
    int         m_cachedDIBnUsage;
    void        *m_cachedDIBpImage;

    // members which can be shared across multiple instances
    static long m_DDRefCount;                   // number of instances of this class
    static HINSTANCE m_hLibDDraw;               // DirectDraw Library handle
    static LPDIRECTDRAW m_DDObject;             // DirectDraw COM object
    static LPDIRECTDRAWSURFACE m_lpDDSPrimary;  // DirectDraw primary surface
    static BOOL m_fDDFailed;                    // one-way ratchet ... once failed, never try again

private:
    BOOL SimpleOutputArea(RECT& rClip, const RECT& rView);
    BOOL SimpleViewArea(RECT& rClip, const RECT& rView);
    void PreparePixelFormat(void);
    void GetLUT(VideoInfo& vi);
    HBITMAP getDIBSection( HDC hDC, BITMAPINFO *bi,
                           int nUsage, void **pImage);
    int _dd_lock(const RECT&, RECT&, RECT&);
    int _dd_unlock();
    int _gdi_lock(const RECT&, RECT&, RECT&);
    int _gdi_unlock();
    int _dib_lock(const RECT&, RECT&, RECT&);
    int _dib_unlock();

public:
    VideoInfo m_VideoInfo;

    int  LockScreen(const RECT&, RECT&, RECT&);
    int	 UnLockScreen();
};

class AwtMasterPrinterGraphics : public AwtGraphics
{
public:
    AwtMasterPrinterGraphics();
    virtual ~AwtMasterPrinterGraphics() {}

    BOOL StartPrintJob(jobject pj);
    void StartPage();
    void EndPage();
    void EndPrintJob();

    virtual void Dispose();

    static AwtMasterPrinterGraphics *
        Create(JNIEnv *env, jobject wGraphics, jobject printJob);

private:
    BOOL m_printingPage;
    BOOL m_printingJob;
};


/************************************************************************
 * Miscellaneous Macros
 */

extern int LogicalSelectClipRgn(HDC hDC, HRGN hClipRgn);

#define GET_JAVAFONT(g) \
    (Hjava_awt_Font *)execute_java_dynamic_method \
        (EE(), (HObject *)(g), "getFont", "()Ljava/awt/Font;")

#define JNI_GET_JAVAFONT(wGraphics) \
    env->CallObjectMethod(wGraphics, AwtGraphics::getFontMID)

#endif /* AWTGRAPHICS_H */
