/*
 * @(#)awt_Image.cpp	1.35 98/09/28
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt.h"
#include "awt_Toolkit.h"
#include "awt_Component.h"
#include "awt_Graphics.h"
#include <awt_image.h>
#include "awt_colors.h"
#include "jni.h"
#include <math.h>
#include "java_awt_color_ColorSpace.h"
#include "java_awt_Transparency.h"
#include "java_awt_image_DataBuffer.h"

CriticalSection imageLock;

/************************************************************************
 * AwtImage fields
 */

jmethodID AwtImage::getBackgroundMID;

// Initialize static variables
HPALETTE AwtImage::m_hPal = NULL;
ImgColorData AwtImage::m_clrdata;

extern "C" {

#include "imageInitIDs.h"
const char *cubemapArray = NULL;

uns_ordered_dither_array img_oda_alpha;

static void make_dither_arrays(int cmapsize, ColorData *cData);
static sgn_ordered_dither_array std_img_oda_red;
static sgn_ordered_dither_array std_img_oda_green;
static sgn_ordered_dither_array std_img_oda_blue;
static int std_odas_computed = 0;
void initInverseGrayLut(int* prgb, int rgbsize, ColorData* cData);


// This is data structure holds the common colormap that is used to paint
// all images.  When an image is painted, the BITMAPINFOHEADER of 
// the image is copied into gpBitmapInfo and then gpBitmapInfo is used to
// paint the image.  This strategy eliminates an alloc/free when painting
// an image.
BITMAPINFO *gpBitmapInfo = NULL;
BITMAPINFO *gpTransMaskInfo = NULL;

const int COLOR4 = 0;
const int COLOR8 = 1;

ColorInfo g_colorInfo[2];

} /* extern "C" */

/************************************************************************
 * AwtImage methods
 */

#define SetRGBQUAD(p, r, g, b)	\
    {				\
	(p)->rgbRed = r;	\
	(p)->rgbGreen = g;	\
	(p)->rgbBlue = b;	\
	(p)->rgbReserved = 0;	\
    }

void AwtImage::InitializePalette(HDC hDC)
{
    int i;
    unsigned int ri, gi, bi;
    ColorInfo *pColorInfo;
		
    m_clrdata.bitsperpixel = ::GetDeviceCaps(hDC, BITSPIXEL);
    if (m_clrdata.bitsperpixel < 8) {
		//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
		// Bugs 4032109, 4047966, and 4071991
		// creighto@eng.sun.com 1997-10-07
		//
		// if less than 4, gracefully end
		if (m_clrdata.bitsperpixel < 4) {
			// Windows 95 special case
		    int colorbits = ::GetDeviceCaps(hDC, BITSPIXEL) *
							::GetDeviceCaps(hDC, PLANES);
			if (colorbits < 4) {
				return;
			}
			else {
				m_clrdata.bitsperpixel = 8;
			}
		}
		else {
			// If 4 bit color mode, use 8 bit palette
			m_clrdata.bitsperpixel = 8;
		}
		//////////////////////////////////////////////////
    }

    // Initialize the dithering array used for transparency masks
    make_uns_ordered_dither_array(img_oda_alpha, 256);

    // Create a BitmapInfo object for transparency masks
    gpTransMaskInfo = (BITMAPINFO *)
	malloc(sizeof(BITMAPINFOHEADER) + 2 * sizeof(RGBQUAD));
    gpTransMaskInfo->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    gpTransMaskInfo->bmiHeader.biWidth = 0;
    gpTransMaskInfo->bmiHeader.biHeight = 0;
    gpTransMaskInfo->bmiHeader.biPlanes = 1;
    gpTransMaskInfo->bmiHeader.biBitCount = 1;
    gpTransMaskInfo->bmiHeader.biCompression = BI_RGB;
    gpTransMaskInfo->bmiHeader.biSizeImage = 0;
    gpTransMaskInfo->bmiHeader.biXPelsPerMeter = 0;
    gpTransMaskInfo->bmiHeader.biYPelsPerMeter = 0;
    gpTransMaskInfo->bmiHeader.biClrUsed = 2;
    gpTransMaskInfo->bmiHeader.biClrImportant = 2;
    SetRGBQUAD(&(gpTransMaskInfo->bmiColors[0]), 0, 0, 0);
    SetRGBQUAD(&(gpTransMaskInfo->bmiColors[1]), 255, 255, 255);

    // Create a BitmapInfo object for color data
    gpBitmapInfo = (BITMAPINFO *)
	malloc(sizeof(BITMAPINFOHEADER) + 256 * sizeof(RGBQUAD));
    gpBitmapInfo->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    gpBitmapInfo->bmiHeader.biBitCount = 0;
    HDC hBMDC = ::CreateDC("DISPLAY", NULL, NULL, NULL);
    HBITMAP hBM = ::CreateCompatibleBitmap(hBMDC, 1, 1);
    VERIFY(::GetDIBits(hBMDC, hBM, 0, 1, NULL, gpBitmapInfo, DIB_RGB_COLORS));

    m_clrdata.grayscale = GS_NOTGRAY;
    if (m_clrdata.bitsperpixel > 8) {
	if (m_clrdata.bitsperpixel != 24) {
	    int foo;
	    gpBitmapInfo->bmiHeader.biCompression = BI_BITFIELDS;
	    VERIFY(::GetDIBits(hBMDC, hBM, 0, 1, &foo,
			       gpBitmapInfo, DIB_RGB_COLORS));
	    ri = ((unsigned int *)gpBitmapInfo->bmiColors)[0];
	    m_clrdata.rOff = 0;
	    while ((ri & 1) == 0) {
		m_clrdata.rOff++;
		ri >>= 1;
	    }
	    m_clrdata.rScale = 0;
	    while (ri < 0x80) {
		m_clrdata.rScale++;
		ri <<= 1;
	    }
	    gi = ((unsigned int *)gpBitmapInfo->bmiColors)[1];
	    m_clrdata.gOff = 0;
	    while ((gi & 1) == 0) {
		m_clrdata.gOff++;
		gi >>= 1;
	    }
	    m_clrdata.gScale = 0;
	    while (gi < 0x80) {
		m_clrdata.gScale++;
		gi <<= 1;
	    }
	    bi = ((unsigned int *)gpBitmapInfo->bmiColors)[2];
	    m_clrdata.bOff = 0;
	    while ((bi & 1) == 0) {
		m_clrdata.bOff++;
		bi >>= 1;
	    }
	    m_clrdata.bScale = 0;
	    while (bi < 0x80) {
		m_clrdata.bScale++;
		bi <<= 1;
	    }
            if (       (0 == m_clrdata.bOff)
                    && (5 == m_clrdata.gOff)
                    && (10 == m_clrdata.rOff)
                    && (3 == m_clrdata.bScale)
                    && (3 == m_clrdata.gScale)
                    && (3 == m_clrdata.rScale)) {
                m_clrdata.bitsperpixel = 15;
		gpBitmapInfo->bmiHeader.biCompression = BI_RGB;
            }
	} else {
	    m_clrdata.bitsperpixel = 24;
	    gpBitmapInfo->bmiHeader.biBitCount = 24;
	    gpBitmapInfo->bmiHeader.biCompression = BI_RGB;

	    // Fill these values in as a convenience for the screen
	    // ColorModel construction code below (see getColorModel())
	    ((int *)gpBitmapInfo->bmiColors)[0] = 0x0000ff;
	    ((int *)gpBitmapInfo->bmiColors)[1] = 0x00ff00;
	    ((int *)gpBitmapInfo->bmiColors)[2] = 0xff0000;
	}
    } else {
	gpBitmapInfo->bmiHeader.biBitCount = 8;
	gpBitmapInfo->bmiHeader.biCompression = BI_RGB;
	gpBitmapInfo->bmiHeader.biClrUsed = 256;
	gpBitmapInfo->bmiHeader.biClrImportant = 256;
	pColorInfo = &g_colorInfo[COLOR8];
	pColorInfo->m_pPaletteColors = (RGBQUAD*)((LPSTR)gpBitmapInfo + sizeof(BITMAPINFOHEADER));

	// Create and initialize a palette
	int nEntries = max(256, ::GetDeviceCaps(hDC, SIZEPALETTE));
	char *buf = new char[sizeof(LOGPALETTE) + nEntries * sizeof(PALETTEENTRY)];
	LOGPALETTE *pLogPal = (LOGPALETTE*)buf;
	PALETTEENTRY *pPalEntries = (PALETTEENTRY *)(&(pLogPal->palPalEntry[0])); 
	int result = max(256, ::GetSystemPaletteEntries(hDC, 0, nEntries, pPalEntries));

	if (result == nEntries) {  
	    if (result == 256 ) {
		int cmapsize = 256;
                int staticgray;
                PALETTEENTRY *pPal;
                char ebuf[4];

                staticgray = 1;
                pPal = pPalEntries;
                for (i = 0; i < 256; i++, pPal++) {
                    if ((pPal->peRed != i) ||
                        (pPal->peRed != pPal->peGreen) ||
                        (pPal->peGreen != pPal->peBlue)) {
                        staticgray = 0;
                        break;
                    } else {
                        pPal->peFlags = PC_NOCOLLAPSE;
                    }
                }
                if (staticgray) {
                    m_clrdata.grayscale = GS_STATICGRAY;
                } else if (GetEnvironmentVariable("FORCEGRAY", ebuf, 0)) {
                    float m = 255.0f / 235.0f;
                    float g = 0.5f;
                    m_clrdata.grayscale = GS_INDEXGRAY;
		    for (i = 0; i < 10; i++) {
		        pPalEntries[i].peRed   = awt_reds[i];
		        pPalEntries[i].peGreen = awt_greens[i];
		        pPalEntries[i].peBlue  = awt_blues[i];
		        pPalEntries[i].peFlags = PC_NOCOLLAPSE;
		    }
                    pPal = &pPalEntries[10];
		    for (i = 10; i < 246; i++, pPal++) {
                        pPal->peRed = pPal->peGreen = pPal->peBlue =
                            (int) g;
                        g += m;
                        pPal->peFlags = PC_NOCOLLAPSE;
		    }
		    for (i = 246; i < 256; i++) {
		        pPalEntries[i].peRed   = awt_reds[i];
		        pPalEntries[i].peGreen = awt_greens[i];
		        pPalEntries[i].peBlue  = awt_blues[i];
		        pPalEntries[i].peFlags = PC_NOCOLLAPSE;
		    }
                } else {
		    for (i = 0; i < 256; i++) {
		        pPalEntries[i].peRed   = awt_reds[i];
		        pPalEntries[i].peGreen = awt_greens[i];
		        pPalEntries[i].peBlue  = awt_blues[i];
		        pPalEntries[i].peFlags = PC_NOCOLLAPSE;
		    }
                }

                {
                ColorData colorData;

                colorData.img_oda_red = &(std_img_oda_red[0][0]);
                colorData.img_oda_green = &(std_img_oda_green[0][0]);
                colorData.img_oda_blue = &(std_img_oda_blue[0][0]);
                make_dither_arrays(cmapsize, &colorData);
                std_odas_computed = 1;
                }

		// Initialize m_paletteColors8.
		for (i = 0; i < nEntries; i++) {
		    pColorInfo->m_pPaletteColors[i].rgbRed =
			pPalEntries[i].peRed;
		    pColorInfo->m_pPaletteColors[i].rgbGreen =
			pPalEntries[i].peGreen,
		    pColorInfo->m_pPaletteColors[i].rgbBlue =
			pPalEntries[i].peBlue;
		    pColorInfo->m_pPaletteColors[i].rgbReserved = 0;
		}

		// Now clear flags of system colors.
		for (i=0; i<10; i++) {
		    pPalEntries[i].peFlags = 0;
		    pPalEntries[255-i].peFlags = 0;
		}
	    }
	    pLogPal->palNumEntries = 256;
	    pLogPal->palVersion = 0x300;
	    m_hPal = ::CreatePalette(pLogPal);
	}
	delete [] buf;

	// Now allocate the 4-bit color palette.
	pColorInfo = &g_colorInfo[COLOR4];
	pColorInfo->m_pPaletteColors = (RGBQUAD*)malloc(16*sizeof(RGBQUAD));
	/* 9: 166, 202, 240 	246: 255 251 240 -- non-VGA colors */	
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[0]), 0, 0, 0);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[1]), 128, 0, 0);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[2]), 0, 128, 0);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[3]), 128, 128, 0);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[4]), 0, 0, 128);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[5]), 128, 0, 128);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[6]), 0, 128, 128);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[7]), 192, 192, 192);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[8]), 128, 128, 128);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[9]), 255, 0, 0);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[10]), 0, 255, 0);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[11]), 255, 255, 0);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[12]), 0, 0, 255);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[13]), 255, 0,255);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[14]), 0, 255, 255);
	SetRGBQUAD(&(pColorInfo->m_pPaletteColors[15]), 255, 255, 255);
	/*
	for (bi = 0 ; bi <= CMAP_COLS ; bi += 1) {
	    int b = bi << (8 - CMAP_BITS);
	    for (gi = 0 ; gi <= CMAP_COLS ; gi += 1) {
		int g = gi << (8 - CMAP_BITS);
		for (ri = 0 ; ri <= CMAP_COLS ; ri += 1) {
		    int r = ri << (8 - CMAP_BITS);
		    if (pColorInfo->m_rgbCube[ri][gi][bi] == 0) {
			pColorInfo->m_rgbCube[ri][gi][bi] = 
			    AwtImage::ClosestMatchPalette
			    (pColorInfo->m_pPaletteColors, 16, r, g, b);
		    }
		}
	    }
	}
	*/
    }
    VERIFY(::DeleteObject(hBM));
    VERIFY(::DeleteDC(hBMDC));
}

void AwtImage::DisposePalette()
{
	free(gpBitmapInfo);
	free(gpTransMaskInfo);
	if (m_hPal != NULL) {
		::DeleteObject(m_hPal);
	}
        if (g_colorInfo[COLOR4].m_pPaletteColors != NULL) {
            free(g_colorInfo[COLOR4].m_pPaletteColors);
        }
}

jobject AwtImage::getColorModel(JNIEnv *env)
{
    jobject awt_colormodel;
    int i;
    if (m_clrdata.bitsperpixel == 24) {
        awt_colormodel =
	    JNU_NewObjectByName(env, "sun/awt/Win32ColorModel24", "()V");
    } else if (m_clrdata.bitsperpixel > 8) {
	int *masks = (int *)gpBitmapInfo->bmiColors;
	int numbits = 0;
	unsigned int bits = (masks[0] | masks[1] | masks[2]);
	while (bits) {
	    numbits++;
	    bits >>= 1;
	}
        awt_colormodel = JNU_NewObjectByName(env, 
                                             "java/awt/image/DirectColorModel",
					     "(IIII)V", numbits,
					     masks[0], masks[1], masks[2]);
    } else if (m_clrdata.grayscale == GS_STATICGRAY) {
        jclass clazz;
        jclass clazz1;
        jmethodID mid;
        jobject cspace = NULL;
        jint bits[1];
        jintArray bitsArray;
        jboolean falseboolean = JNI_FALSE;
 
        clazz1 = env->FindClass("java/awt/color/ColorSpace");
        mid = env->GetStaticMethodID(clazz1, "getInstance",
              "(I)Ljava/awt/color/ColorSpace;");
        if (mid == 0) {
            return NULL;
        }
        cspace = env->CallStaticObjectMethod(clazz1, mid,
            java_awt_color_ColorSpace_CS_GRAY);
        if (cspace == 0) {
            return NULL;
        }
 
        bits[0] = 8;
        bitsArray = env->NewIntArray(1);
        if (bitsArray == 0) {
            return NULL;
        } else {
            env->SetIntArrayRegion(bitsArray, 0, 1, bits);
        }
 
        clazz = env->FindClass("java/awt/image/ComponentColorModel");

        mid = env->GetMethodID(clazz,"<init>",
            "(Ljava/awt/color/ColorSpace;[IZZII)V");
     
        if (mid == 0) {
            return NULL;
        }
     
        awt_colormodel = env->NewObject(clazz, mid,
                                        cspace,
                                        bitsArray,
                                        falseboolean,
                                        falseboolean,
                                        java_awt_Transparency_OPAQUE,
                                        java_awt_image_DataBuffer_TYPE_BYTE);        
    } else {
        jintArray hRGB = env->NewIntArray(256);
        unsigned int *rgb, *rgbP;
	RGBQUAD *c, *cBase;
        jboolean hasalpha = JNI_FALSE;

        /* Create the LUT from the color map */
	cBase = g_colorInfo[COLOR8].m_pPaletteColors;
        rgb = (unsigned int *) env->GetPrimitiveArrayCritical(hRGB, 0);
        if (rgb == NULL) {
            return NULL;
        }
        rgbP = rgb;
        if (m_clrdata.grayscale == GS_INDEXGRAY) {
            /* For IndexColorModel, pretend first 10 colors and last
               10 colors are transparent black.  This makes ICM.allgrayopaque
               true.
             */
            hasalpha = JNI_TRUE;
	    for (i=0; i < 10; i++) {
                *rgbP++ = 0x00000000;
            }
	    for (i=10, c = cBase + 10; i < 246; c++, i++) {
                *rgbP++ = 0xff000000 |
                    ((c->rgbRed&0xff)<<16)  |
                    ((c->rgbGreen&0xff)<<8) |
                    (c->rgbBlue&0xff);
	    }
	    for (i=246; i < 256; i++) {
                *rgbP++ = 0x00000000;
            }
        } else {
	    for (i=0, c = cBase; i < 256; c++, i++) {
                *rgbP++ = 0xff000000 |
                    ((c->rgbRed&0xff)<<16)  |
                    ((c->rgbGreen&0xff)<<8) |
                    (c->rgbBlue&0xff);
	    }
        }
        env->ReleasePrimitiveArrayCritical(hRGB, rgb, JNI_COMMIT);

        /* Construct a new color model */
        awt_colormodel = JNU_NewObjectByName(env, 
                                             "java/awt/image/IndexColorModel",
                                             "(II[IIZII)V",
                                             8, 256,
                                             hRGB, 0, 
                                             hasalpha,
                                             -1,
                                             java_awt_image_DataBuffer_TYPE_BYTE);
    }
    return awt_colormodel;
}

/************************************************************************
 * sun.awt.image.Image native methods
 */

extern "C" {

/*
 * Class:     sun_awt_image_Image
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_image_Image_initIDs(JNIEnv *env, jclass cls)
{
//    AwtImage::getBackgroundMID = env->GetMethodID(cls, "getBackgroundColor","()Ljava/awt/Color;");
//    ASSERT(AwtImage::getBackgroundMID != NULL);
}

void color_init()
{
    static const int cmapsize = 256;
    static int initialized = 0;

    cubemapArray = (const char *)(AwtToolkit::GetInstance().getILUT());

    if (initialized) {
        return;
    } else {
        initialized = 1;
    }

}

void initInverseGrayLut(int* prgb, int rgbsize, ColorData *cData) {
    int *idata;
    int g;

    if (!cData) {
        return;
    }

    idata = (int *) calloc(256, sizeof(int));
    if (!idata) {
        return;
    }
    cData->pGrayInverseLutData = idata;
    for (g = 0; g < 256; g++) {
        int i, mindist, d, besti;
        int *pg;
        mindist = 256;
        besti = 0;
        pg = prgb;
        for (i = 0; i < rgbsize; i++, pg++) {
            if (*pg == 0x0) {
                /* ignore transparent black */
                continue;
            }
            d = (*pg & 0xff) - g;
            if (d < 0) d = -d;
            if (d < mindist) {
                besti = i;
                if (d == 0) {
                    break;
                }
                mindist = d;
            }
        }
        *idata++ = besti;
    }
}

void freeICMColorData(ColorData *pData) {
    if (pData) {
        if (pData->img_clr_tbl) {
            free(pData->img_clr_tbl);
        }
        if (pData->pGrayInverseLutData) {
            free(pData->pGrayInverseLutData);
        }
    free(pData);
    }
}

#include <malloc.h>

/*
 * state info needed for breadth-first recursion of color cube from
 * initial palette entries within the cube
 */

typedef struct {
    unsigned int depth;
    unsigned int maxDepth;

    unsigned char *usedFlags;
    unsigned int  activeEntries;
    unsigned short *rgb;
    unsigned char *indices;
    unsigned char *iLUT;
} CubeStateInfo;

#define INSERTNEW(state, rgb, index) do {                           \
        if (!state.usedFlags[rgb]) {                                \
            state.usedFlags[rgb] = 1;                               \
            state.iLUT[rgb] = index;                                \
            state.rgb[state.activeEntries] = rgb;                   \
            state.indices[state.activeEntries] = index;             \
            state.activeEntries++;                                  \
        }                                                           \
} while (0);
    

#define ACTIVATE(code, mask, delta, state, index) do {              \
    if (((rgb & mask) + delta) <= mask) {                           \
        rgb += delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb -= delta;                                               \
    }                                                               \
    if ((rgb & mask) >= delta) {                                    \
        rgb -= delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb += delta;                                               \
    }                                                               \
} while (0);

/* REMIND: does not deal well with bifurcation which happens when two
 * palette entries map to the same cube vertex
 */

static int
recurseLevel(CubeStateInfo *priorState) {
    int i;
    CubeStateInfo currentState;
    memcpy(&currentState, priorState, sizeof(CubeStateInfo));
    currentState.rgb = (unsigned short *)malloc(6
                                                * sizeof(unsigned short)
                                                * priorState->activeEntries);
    if (currentState.rgb == NULL) {
        return 0;
    }
    
    currentState.indices = (unsigned char *)malloc(6
                                                * sizeof(unsigned char)
                                                * priorState->activeEntries);
    if (currentState.indices == NULL) {
        free(currentState.rgb);
        return 0;
    }
    
    currentState.depth++;
    if (currentState.depth > priorState->maxDepth) {
        priorState->maxDepth = currentState.depth;
    }
    currentState.activeEntries = 0;
    for (i=priorState->activeEntries - 1; i >= 0; i--) {
        unsigned short rgb = priorState->rgb[i];
        unsigned char  index = priorState->indices[i];
        ACTIVATE(rgb, 0x7c00, 0x0400, currentState, index);
        ACTIVATE(rgb, 0x03e0, 0x0020, currentState, index);
        ACTIVATE(rgb, 0x001f, 0x0001, currentState, index);
    }
    if (currentState.activeEntries) {
        if (!recurseLevel(&currentState)) {
	    free(currentState.rgb);
	    free(currentState.indices);
            return 0;
        }
    }
    if (currentState.maxDepth > priorState->maxDepth) {
        priorState->maxDepth = currentState.maxDepth;
    }
    free(currentState.rgb);
    free(currentState.indices);
    return  1;
}

/*
 * REMIND: take core inversedLUT calculation to the shared tree and
 * recode the functions (Win32)awt_Image:initCubemap(),
 * (Win32)awt_Image:make_cubemap(), (Win32)AwtToolkit::GenerateInverseLUT(),
 * (Solaris)color:initCubemap() to call the shared codes.
 */
unsigned char*
make_cubemap(int* cmap,
             int  cmap_len,
             int  cube_dim) {
    int i;
    CubeStateInfo currentState;
    int cubesize = cube_dim * cube_dim * cube_dim;
    unsigned char *useFlags;
    unsigned char *newILut = (unsigned char*)malloc(cubesize);
    if (newILut) {
        useFlags = (unsigned char *)calloc(cubesize, 1);
        if (useFlags == 0) {
            free(newILut);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()1\n");
#endif
            return NULL;
        }
        
        /* use existing palette as seed values */
        currentState.depth          = 0;
        currentState.maxDepth       = 0;
        currentState.usedFlags      = useFlags;
        currentState.activeEntries  = 0;
        currentState.iLUT           = newILut;
        currentState.rgb = (unsigned short *)
                                malloc(256 * sizeof(unsigned short));
        if (currentState.rgb == NULL) {
            free(newILut);
	    free(useFlags);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()2\n");
#endif
            return NULL;
        }
        
        currentState.indices = (unsigned char *)
                                malloc(256 * sizeof(unsigned char));
        if (currentState.indices == NULL) {
            free(newILut);
	    free(useFlags);
	    free(currentState.rgb);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()3\n");
#endif
            return NULL;
        }
        
        for (i = 0; i < 128; i++) {
            unsigned short rgb;
            int pixel = cmap[i];
            rgb = (pixel & 0x00f80000) >> 9;
            rgb |= (pixel & 0x0000f800) >> 6;
            rgb |=  (pixel & 0xf8) >> 3;
            INSERTNEW(currentState, rgb, i);
            pixel = cmap[255-i];
            rgb = (pixel & 0x00f80000) >> 9;
            rgb |= (pixel & 0x0000f800) >> 6;
            rgb |=  (pixel & 0xf8) >> 3;
            INSERTNEW(currentState, rgb, 255-i);
        }
        if (!recurseLevel(&currentState)) {
            free(newILut);
            free(useFlags);
            free(currentState.rgb);
	    free(currentState.indices);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()4\n");
#endif
	    return NULL;
	}
        
	free(useFlags);
	free(currentState.rgb);
	free(currentState.indices);
        return newILut;
    }

#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:make_cubemap()5\n");
#endif
    return NULL;
}

/*
 * REMIND: take core inversedLUT calculation to the shared tree and
 * recode the functions (Win32)awt_Image:initCubemap(),
 * (Win32)awt_Image:make_cubemap(), (Win32)AwtToolkit::GenerateInverseLUT(),
 * (Solaris)color:initCubemap() to call the shared codes.
 */
unsigned char*
initCubemap(int* cmap,
            int  cmap_len,
            int  cube_dim) {
    int i;
    CubeStateInfo currentState;
    int cubesize = cube_dim * cube_dim * cube_dim;
    unsigned char *useFlags;
    unsigned char *newILut = (unsigned char*)malloc(cubesize);
    if (newILut) {
        useFlags = (unsigned char *)calloc(cubesize, 1);
        if (useFlags == 0) {
            free(newILut);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()1\n");
#endif
            return NULL;
        }
        /* use existing palette as seed values */
        currentState.depth          = 0;
        currentState.maxDepth       = 0;
        currentState.usedFlags      = useFlags;
        currentState.activeEntries  = 0;
        currentState.iLUT           = newILut;

        currentState.rgb = (unsigned short *)
                                malloc(256 * sizeof(unsigned short));
        if (currentState.rgb == NULL) {
            free(newILut);
	    free(useFlags);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()2\n");
#endif
            return NULL;
        }
        
        currentState.indices = (unsigned char *)
                                malloc(256 * sizeof(unsigned char));
        if (currentState.indices == NULL) {
            free(newILut);
	    free(useFlags);
	    free(currentState.rgb);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()3\n");
#endif
            return NULL;
        }
        
        for (i = 0; i < 128; i++) {
            unsigned short rgb;
            int pixel = cmap[i];
            rgb = (pixel & 0x00f80000) >> 9;
            rgb |= (pixel & 0x0000f800) >> 6;
            rgb |=  (pixel & 0xf8) >> 3;
            INSERTNEW(currentState, rgb, i);
            pixel = cmap[255-i];
            rgb = (pixel & 0x00f80000) >> 9;
            rgb |= (pixel & 0x0000f800) >> 6;
            rgb |=  (pixel & 0xf8) >> 3;
            INSERTNEW(currentState, rgb, 255-i);
        }
        if (!recurseLevel(&currentState)) {
            free(newILut);
            free(useFlags);
            free(currentState.rgb);
	    free(currentState.indices);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()4\n");
#endif
	    return NULL;
	}
        
	free(useFlags);
	free(currentState.rgb);
	free(currentState.indices);
        return newILut;
        
    }
 
#ifdef DEBUG
	fprintf(stderr, "Out of memory in awt_Image:initCubemap()5\n");
#endif
    return NULL;
}

void
initDitherTables(ColorData* cData) {


    if(std_odas_computed) {
        cData->img_oda_red   = &(std_img_oda_red[0][0]);
        cData->img_oda_green = &(std_img_oda_green[0][0]);
        cData->img_oda_blue  = &(std_img_oda_blue[0][0]);
    } else {
        cData->img_oda_red   = &(std_img_oda_red[0][0]);
        cData->img_oda_green = &(std_img_oda_green[0][0]);
        cData->img_oda_blue  = &(std_img_oda_blue[0][0]);
        make_dither_arrays(256, cData);
        std_odas_computed = 1;
    }

}

static void make_dither_arrays(int cmapsize, ColorData *cData) {
    int i, j, k;

    /*
     * Initialize the per-component ordered dithering arrays
     * Choose a size based on how far between elements in the
     * virtual cube.  Assume the cube has cuberoot(cmapsize)
     * elements per axis and those elements are distributed
     * over 256 colors.
     * The calculation should really divide by (#comp/axis - 1)
     * since the first and last elements are at the extremes of
     * the 256 levels, but in a practical sense this formula
     * produces a smaller error array which results in smoother
     * images that have slightly less color fidelity but much
     * less dithering noise, especially for grayscale images.
     */
    i = (int) (256 / pow(cmapsize, 1.0/3.0));
    make_sgn_ordered_dither_array(cData->img_oda_red, -i / 2, i / 2);
    make_sgn_ordered_dither_array(cData->img_oda_green, -i / 2, i / 2);
    make_sgn_ordered_dither_array(cData->img_oda_blue, -i / 2, i / 2);

    /*
     * Flip green horizontally and blue vertically so that
     * the errors don't line up in the 3 primary components.
     */
    for (i = 0; i < 8; i++) {
        for (j = 0; j < 4; j++) {
            k = cData->img_oda_green[(i<<3)+j];
            cData->img_oda_green[(i<<3)+j] = cData->img_oda_green[(i<<3)+7 - j];
            cData->img_oda_green[(i<<3) + 7 - j] = k;
            k = cData->img_oda_blue[(j<<3)+i];
            cData->img_oda_blue[(j<<3)+i] = cData->img_oda_blue[((7 - j)<<3)+i];
            cData->img_oda_blue[((7 - j)<<3) + i] = k;
        }
    }
}

} /* extern "C" */
