/*
 * @(#)awt_InputMethod.cpp	1.15 98/08/17
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "jlong.h"
#include "awt_AWTEvent.h"
#include "awt_Component.h"
#include "awt_Toolkit.h"
#include <sun_awt_windows_WInputMethod.h>
#include <java_awt_event_InputMethodEvent.h>

//#define	TWO_PASS_DESIGN

/************************************************************************
 * WInputMethod native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    createNativeContext
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WInputMethod_createNativeContext(JNIEnv *env, jobject self)
{
    // use special message to call ImmCreateContext() in main thread.
    return (jint)AwtToolkit::GetInstance().SendMessage(WM_AWT_CREATECONTEXT);
}


/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    destroyNativeContext
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WInputMethod_destroyNativeContext(JNIEnv *env, jobject self, jint context)
{
    // use special message to call ImmDestroyContext() in main thread.
    AwtToolkit::GetInstance().SendMessage(WM_AWT_DESTROYCONTEXT, context, 0);
}


/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    enableNativeIME
 * Signature: (Lsun/awt/windows/WComponentPeer;I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WInputMethod_enableNativeIME(JNIEnv *env, jobject self, jobject peer, 
                                                  jint context, jboolean useNativeCompWindow)
{
    //get C++ Class of Focused Component
    if (peer == 0)	return;
    AwtComponent* p = (AwtComponent*)JNI_GET_PDATA(peer);
    if (p == 0)		return;

    p->SetInputMethod(self, useNativeCompWindow);

    // use special message to call ImmAssociateContext() in main thread.
    AwtToolkit::GetInstance().SendMessage(WM_AWT_ASSOCIATECONTEXT,
					  (UINT)p->GetHWnd(), context);
}


/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    disableNativeIME
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WInputMethod_disableNativeIME(JNIEnv *env, jobject self, jobject peer)
{
    //get C++ Class of Focused Component
    if (peer == 0)	return;
    AwtComponent* p = (AwtComponent*)JNI_GET_PDATA(peer);
    if (p == 0)		return;

    p->SetInputMethod(NULL, TRUE);

    // use special message to call ImmAssociateContext() in main thread.
    AwtToolkit::GetInstance().SendMessage(WM_AWT_ASSOCIATECONTEXT,
					  (UINT)p->GetHWnd(), NULL);
}


/*
 * Class:     sun_awt_windows_WComponentPeer
 * Method:    handleEvent
 * Signature: (Lsun/awt/windows/WComponentPeer;Ljava/awt/AWTEvent;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WInputMethod_handleNativeIMEEvent(JNIEnv *env, jobject self, 
                                                       jobject peer, jobject event)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(peer);
    AwtComponent* p = (AwtComponent *)pData;

    JNI_CHECK_NULL_RETURN(event, "null AWTEvent");
    MSG* pMsg =
      (MSG*)jlong_to_ptr((env)->GetLongField(event, AwtAWTEvent::dataID));
    BOOL isConsumed =
      (BOOL)(env)->GetBooleanField(event, AwtAWTEvent::consumedID);
    int id = (env)->GetIntField(event, AwtAWTEvent::idID);
    ASSERT(!env->ExceptionOccurred());

    if (pMsg==NULL || isConsumed || p==NULL)	return;

#ifdef  TWO_PASS_DESIGN     // this method is only for 2-pass design
    switch(pMsg->message){
	case WM_AWT_PRE_KEYDOWN:
	case WM_AWT_PRE_KEYUP:
	case WM_AWT_PRE_SYSKEYDOWN:
	case WM_AWT_PRE_SYSKEYUP:
	    if (pMsg->wParam == VK_PROCESSKEY){
		p->SendMessage(pMsg->message, pMsg->wParam, pMsg->lParam);
                (env)->SetBoolean(event, AwtAWTEvent::consumedID, JNI_TRUE);
	    }
	    break;
    }
#endif

    if (id >= java_awt_event_InputMethodEvent_INPUT_METHOD_FIRST &&
        id <= java_awt_event_InputMethodEvent_INPUT_METHOD_LAST) 
    {
        MSG* pCopiedMsg = new MSG();
	pCopiedMsg->message = pMsg->message;
	pCopiedMsg->wParam = pMsg->wParam;
	pCopiedMsg->lParam = pMsg->lParam;
        if (pMsg->message==WM_CHAR || pMsg->message==WM_SYSCHAR) {
            WCHAR unicodeChar = L'\0';
            ::MultiByteToWideChar(p->GetCodePage(), 0, (TCHAR*)&(pMsg->wParam), 1, &unicodeChar, 1);
	    p->SendKeyEvent(java_awt_event_KeyEvent_KEY_TYPED, 0, //to be fixed nowMillis(), 
				     java_awt_event_KeyEvent_CHAR_UNDEFINED, unicodeChar, 
				     p->GetJavaModifiers(HIWORD(pMsg->lParam), 0), pCopiedMsg);
        } else {
            p->SendMessage(WM_AWT_HANDLE_EVENT, (WPARAM) FALSE,
			(LPARAM) pCopiedMsg);
        }
        (env)->SetBooleanField(event, AwtAWTEvent::consumedID, JNI_TRUE);
    }

}

/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    notifyNativeIME
 * Signature: (Lsun/awt/windows/WComponentPeer;I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WInputMethod_endCompositionNative(JNIEnv *env, jobject self,
						       jint context, jboolean flag)
{
    // TODO: currently the flag parameter is ignored and the outstanding input is
    //       always discarded.
    //       If the flag value is Java_sun_awt_windows_WInputMethod_COMMIT_INPUT,
    //       then input text should be committed. Otherwise, should be discarded.

    // use special message to call ImmNotifyContext() in main thread.
    AwtToolkit::GetInstance().SendMessage(WM_AWT_ENDCOMPOSITION, context, 0);
}

/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    setConversionStatus
 * Signature: (II)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WInputMethod_setConversionStatus(JNIEnv *env, jobject self, jint context, jint request)
{
    // use special message to call ImmSetConversionStatus() in main thread.
    AwtToolkit::GetInstance().SendMessage(WM_AWT_SETCONVERSIONSTATUS, 
					  context, 
					  MAKELPARAM((WORD)request, (WORD)0));
}

/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    getConversionStatus
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WInputMethod_getConversionStatus(JNIEnv *env, jobject self, jint context)
{
    // use special message to call ImmSetConversionStatus() in main thread.
    return (jint) AwtToolkit::GetInstance().SendMessage(
        WM_AWT_GETCONVERSIONSTATUS, context, 0);
}

/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    setOpenStatus
 * Signature: (IZ)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WInputMethod_setOpenStatus(JNIEnv *env, jobject self, jint context, jboolean flag)
{
    // use special message to call ImmSetConversionStatus() in main thread.
    ::SendMessage(AwtToolkit::GetInstance().GetHWnd(), 
		  WM_AWT_SETOPENSTATUS, 
		  context, flag);
}

/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    getConversionStatus
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL
Java_sun_awt_windows_WInputMethod_getOpenStatus(JNIEnv *env, jobject self, jint context)
{
    // use special message to call ImmSetConversionStatus() in main thread.
    return (jboolean)::SendMessage(AwtToolkit::GetInstance().GetHWnd(), 
		                   WM_AWT_GETOPENSTATUS, 
		                   context, 0);
}

/*
 * Class:     sun_awt_windows_WInputMethod
 * Method:    getNativeIMELanguage
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WInputMethod_getNativeIMELanguage(JNIEnv *env, jobject self)
{
    return (jint)AwtComponent::GetInputLanguage();
}
} /* extern "C" */

