/*
 * @(#)awt_Label.cpp	1.31 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_Label.h"
#include "awt_Canvas.h"
#include "awt_Graphics.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtLabel fields
 */

jfieldID AwtLabel::textID;
jfieldID AwtLabel::alignmentID;


/************************************************************************
 * AwtLabel methods
 */

AwtLabel::AwtLabel() {
}

const char* AwtLabel::GetClassName() {
    return "SunAwtLabel";
}

/* Create a new AwtLabel object and window. */
AwtLabel* AwtLabel::Create(jobject labelPeer, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    int pData;
    AwtCanvas* awtParent;
    AwtLabel* awtLabel;
    jobject target;

    JNI_CHECK_PEER_GOTO(parent, done);
    awtParent = (AwtCanvas*)pData;
    JNI_CHECK_NULL_GOTO(awtParent, "awtParent", done);
    target  = env->GetObjectField(labelPeer, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "target", done);
    awtLabel = new AwtLabel();
    JNI_CHECK_NULL_GOTO(awtLabel, "AwtLabel() failed", done);

    {
        DWORD style = WS_CHILD | WS_CLIPSIBLINGS;

	DWORD exStyle = 0;
        if (GetRTLReadingOrder())
	    exStyle |= WS_EX_RTLREADING;

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);
	awtLabel->CreateHWnd(env, "", style, exStyle,
			     x, y, width, height,
			     awtParent->GetHWnd(),
			     NULL,
			     ::GetSysColor(COLOR_WINDOWTEXT),
			     ::GetSysColor(COLOR_BTNFACE),
			     labelPeer);
    }
 done:
    env->DeleteLocalRef(target);
    return awtLabel;
}

void AwtLabel::DoPaint(HDC hDC, RECT& r)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    if ((r.right-r.left) > 0 && (r.bottom-r.top) > 0 &&
        m_peerObject != NULL && m_callbacksEnabled) {

        if (env->EnsureLocalCapacity(3) < 0)
	    return;
        long x,y;
        SIZE size;

	/* self is sun.awt.windows.WLabelPeer  */

	jobject self = GetPeer(env);
	ASSERT(self);

	/* target is java.awt.Label */
	jobject target = env->GetObjectField(self, AwtObject::targetID);
	jobject font = GET_FONT(target, self);
	jstring text = (jstring)env->GetObjectField(target, AwtLabel::textID);

        size = AwtFont::getMFStringSize(hDC, font, text);
        ::SetTextColor(hDC, GetColor());
        /* Redraw whole label to eliminate display noise during resizing. */
        VERIFY(::GetClientRect(GetHWnd(), &r));
        VERIFY(::FillRect (hDC, &r, GetBackgroundBrush()));
        y = (r.top + r.bottom - size.cy) / 2;

	int alignment = env->GetIntField(target, AwtLabel::alignmentID);
        switch (alignment) {
	   case java_awt_Label_LEFT:
              x = 0;
              break;
          case java_awt_Label_CENTER:
              x = (r.left + r.right - size.cx) / 2;
              break;
          case java_awt_Label_RIGHT:
              x = r.right - size.cx;
              break;
        }
        /* draw string */
        if (isEnabled()) {
            AwtComponent::DrawWindowText(hDC, font, text, x, y);
        } else {
            AwtComponent::DrawGrayText(hDC, font, text, x, y);
        }
        DoCallback("handlePaint", "(IIII)V",
                   r.left, r.top, r.right-r.left, r.bottom-r.top);
	env->DeleteLocalRef(target);
	env->DeleteLocalRef(font);
	env->DeleteLocalRef(text);
    }
}

MsgRouting AwtLabel::WmPaint(HDC)
{
    GetSharedDC()->Lock();
    PAINTSTRUCT ps;
    HDC hDC = ::BeginPaint(GetHWnd(), &ps);/* the passed-in HDC is ignored. */
    ASSERT(hDC);
    RECT& r = ps.rcPaint;
    DoPaint(hDC, r);
    GetSharedDC()->Unlock();
    return mrConsume;
}

MsgRouting AwtLabel::WmPrintClient(HDC hDC, UINT)
{
    RECT r;

    // obtain valid DC from GDI stack
    ::RestoreDC(hDC, -1);

    ::GetClipBox(hDC, &r);
    DoPaint(hDC, r);
    return mrConsume;
}


/************************************************************************
 * Label native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_Label_initIDs(JNIEnv *env, jclass cls)
{

    /* init field ids */
    AwtLabel::textID = env->GetFieldID(cls, "text", "Ljava/lang/String;");
    AwtLabel::alignmentID = env->GetFieldID(cls, "alignment", "I");

    ASSERT(AwtLabel::textID != NULL);
    ASSERT(AwtLabel::alignmentID != NULL);
}

} /* extern "C" */


/************************************************************************
 * WLabelPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WLabelPeer
 * Method:    setText
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WLabelPeer_setText(JNIEnv *env, jobject self,
					jstring text)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);

    AwtComponent* c = (AwtComponent*)pData;
    c->SetText(JavaStringBuffer(env, text));
    VERIFY(::InvalidateRect(c->GetHWnd(), NULL, TRUE));
}

/*
 * Class:     sun_awt_windows_WLabelPeer
 * Method:    setAlignment
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WLabelPeer_setAlignment(JNIEnv *env, jobject self,
					     jint alignment)
{

    /*
     * alignment argument of multifont label is referred to in
     * WmDrawItem method
     */
    int pData;
    JNI_CHECK_PEER_RETURN(self);

    AwtComponent* c = (AwtComponent*)pData;
    VERIFY(::InvalidateRect(c->GetHWnd(), NULL, TRUE));
}

/*
 * Class:     sun_awt_windows_WLabelPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WLabelPeer_create(JNIEnv *env, jobject self,
				       jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtLabel::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

} /* export "C" */
