/*
 * @(#)awt_List.cpp	1.40 98/10/08
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_List.h"

#include "awt_Canvas.h"
#include "awt_Dimension.h"
#include "awt_Graphics.h"
#include "awt_Toolkit.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtList methods
 */

AwtList::AwtList() {
    isMultiSelect = FALSE;
}

const char* AwtList::GetClassName() {
    return "LISTBOX";  // System provided listbox class
}

/* Create a new AwtList object and window.   */
AwtList* AwtList::Create(jobject peer, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    int pData;
    AwtCanvas* awtParent;
    jobject target = 0;
    AwtList* c;
    JNI_CHECK_PEER_GOTO(parent, done);

    awtParent = (AwtCanvas*)pData;
    JNI_CHECK_NULL_GOTO(awtParent, "null awtParent", done);

    /* target is Hjava_awt_List * */
    target = env->GetObjectField(peer, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    c = new AwtList();
    JNI_CHECK_NULL_GOTO(c, "AwtList() failed", done);

    {
        long style = WS_CHILD | WS_CLIPSIBLINGS | WS_VSCROLL | WS_HSCROLL |
	  LBS_NOINTEGRALHEIGHT | LBS_NOTIFY | LBS_OWNERDRAWVARIABLE
	  | (IS_WIN4X ? 0 : WS_BORDER);
	long exStyle = IS_WIN4X ? WS_EX_CLIENTEDGE : 0;

	if (GetRTL()) {
	   exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
	   if (GetRTLReadingOrder())
	       exStyle |= WS_EX_RTLREADING;
	}

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	c->CreateHWnd(env, "", style, exStyle,
		      x, y, width, height,
		      awtParent->GetHWnd(),
		      (HMENU)awtParent->CreateControlID(),
		      ::GetSysColor(COLOR_WINDOWTEXT),
		      ::GetSysColor(COLOR_WINDOW),
		      peer
		      );

	/* suppress inheriting awtParent's color. */
	c->m_backgroundColorSet = TRUE;
    }
 done:
    env->DeleteLocalRef(target);
    return c;
}

//Netscape : Override the AwtComponent method so we can set the item height
//for each item in the list.  Modified by echawkes to avoid race condition.  

void AwtList::SetFont(AwtFont* font)
{
    ASSERT(font != NULL);
    if (font->GetAscent() < 0)
    {
        AwtFont::SetupAscent(font);
    }
    HANDLE hFont = font->GetHFont();
    SendMessage(WM_SETFONT, (WPARAM)hFont, MAKELPARAM(FALSE, 0));

    // get the shared DC and lock it
    AwtSharedDC *pDC = GetSharedDC();
    if (pDC == NULL)
        return;
    pDC->Lock();
    pDC->GrabDC(NULL);
    HDC hDC = pDC->GetDC();
    
    TEXTMETRIC tm;
    VERIFY(::SelectObject(hDC, hFont) != NULL);
    VERIFY(::GetTextMetrics(hDC, &tm));

    // Free up the shared DC
    pDC->Unlock();

    long h = tm.tmHeight + tm.tmExternalLeading;

    int nCount = (int)SendMessage(LB_GETCOUNT);
    int i;
    for(i=0; i<nCount; i++)
    {
        VERIFY(SendMessage(LB_SETITEMHEIGHT, i, MAKELPARAM(h, 0)) != LB_ERR);
    }
    VERIFY(::RedrawWindow(GetHWnd(), NULL, NULL, RDW_INVALIDATE |RDW_FRAME |RDW_ERASE));
}

void AwtList::SetMultiSelect(BOOL ms) {
    if (ms == isMultiSelect) {
        return;
    }
    isMultiSelect = ms;

    /* Copy current box's contents to string array */
    int nCount = GetCount();
    char** strings = new char*[nCount];
    for(int i = 0; i < nCount; i++) {
        int len = SendMessage(LB_GETTEXTLEN, i);
        char* text = new char[len + 1];
        VERIFY(SendMessage(LB_GETTEXT, i, (LPARAM)text) != LB_ERR);
        strings[i] = text;
    }
    int nCurSel = GetCurrentSelection();

    /* Save old list box's attributes */
    RECT rect;
    GetWindowRect(GetHWnd(), &rect);
    AwtComponent* pParent = GetParent();
    MapWindowPoints(0, pParent->GetHWnd(), (LPPOINT)&rect, 2);

    HANDLE font = (HANDLE)SendMessage(WM_GETFONT);
    DWORD style = GetStyle() | WS_VSCROLL;
    if (isMultiSelect) {
        style |= LBS_MULTIPLESEL;
    } else {    
        style &= ~LBS_MULTIPLESEL;
    } 
    DWORD exStyle = ::GetWindowLong(GetHWnd(), GWL_EXSTYLE);

    /* Recreate list box peer is sun.awt.windows.WListPeer */
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    jobject peer = GetPeer(env);

    DestroyHWnd();
    CreateHWnd(env, "", style, exStyle, 
               rect.left, rect.top, 
               rect.right-rect.left, rect.bottom-rect.top, 
               pParent->GetHWnd(),
               (HMENU)m_myControlID,
               ::GetSysColor(COLOR_WINDOWTEXT),
               ::GetSysColor(COLOR_WINDOW),
               peer
        );

    SendMessage(WM_SETFONT, (WPARAM)font, (LPARAM)FALSE);
    SendMessage(LB_RESETCONTENT);  
    for (i = 0; i < nCount; i++) {
        InsertString(i, strings[i]);
        delete [] strings[i];
    }
    SendMessage(LB_SETCURSEL, nCurSel);
    delete[] strings;
}

/*
 * There currently is no good place to cache java.awt.Dimension field
 * ids. If this method gets called a lot, one such place should be found.
 * -- br 07/18/97.
 */
jobject AwtList::PreferredItemSize(JNIEnv *env)
{
    jobject peer = GetPeer(env);
    jobject dimension = JNU_CallMethodByName(env, NULL, peer, "preferredSize", 
					     "(I)Ljava/awt/Dimension;",
					     1).l;
					      
    ASSERT(!env->ExceptionOccurred());
    if (dimension == NULL) {
	return NULL;
    }
    /* This size is too big for each item height. */
    (env)->SetIntField(dimension, AwtDimension::heightID, GetFontHeight(env));

    return dimension;
}

// Every time something gets added to the list, we increase the max width 
// of items that have ever been added.  If it surpasses the width of the 
// listbox, we show the scrollbar.  When things get deleted, we shrink 
// the scroll region back down and hide the scrollbar, if needed.
void AwtList::AdjustHorizontalScrollbar()
{
    // The border width is added to the horizontal extent to ensure that we
    // can view all of the text when we move the horz. scrollbar to the end.
    int  cxBorders = GetSystemMetrics( SM_CXBORDER ) * 2;
    RECT rect;
    VERIFY(::GetClientRect(GetHWnd(), &rect));
    int iHorzExt = SendMessage(LB_GETHORIZONTALEXTENT, 0, 0L ) - cxBorders;
    if ( (m_nMaxWidth > rect.left)  // if strings wider than listbox
      || (iHorzExt != m_nMaxWidth) ) //   or scrollbar not needed anymore.
    {
        SendMessage(LB_SETHORIZONTALEXTENT, m_nMaxWidth + cxBorders, 0L);
    }
}

// This function goes through all strings in the list to find the width, 
// in pixels, of the longest string in the list.
void AwtList::UpdateMaxItemWidth()
{
    m_nMaxWidth = 0;

    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0)
        return;

    // get the shared DC and lock it
    AwtSharedDC * pDC = GetSharedDC();
    if ( pDC == NULL )
        return;
    pDC->Lock();
    pDC->GrabDC(NULL);
    HDC hDC = pDC->GetDC();

    jobject self = GetPeer(env);
    ASSERT(self);

    /* target is java.awt.List */
    jobject target = env->GetObjectField(self, AwtObject::targetID);
    jobject font = GET_FONT(target, self);

    int nCount = GetCount();
    for ( int i=0; i < nCount; i++ )
    {
        jstring jstr = GetItemString( env, target, i );
        SIZE size = AwtFont::getMFStringSize( hDC, font, jstr );
        if ( size.cx > m_nMaxWidth )
            m_nMaxWidth = size.cx;
        env->DeleteLocalRef( jstr );
    }

    // free up the shared DC and release local refs
    pDC->Unlock();
    env->DeleteLocalRef( target );
    env->DeleteLocalRef( font );

    // Now adjust the horizontal scrollbar extent
    AdjustHorizontalScrollbar();
}

MsgRouting
AwtList::WmSize(int type, int w, int h)
{
    AdjustHorizontalScrollbar();
    return AwtComponent::WmSize(type, w, h);
}

MsgRouting AwtList::WmNotify(UINT notifyCode)
{
    /*
     * !CQ why not call SendEvent() to get event over to java???  
     * need to get ITEM_CHANGED, ITEMS_SELECTED, ITEMS_DESELECTED over
     * also 
     */
    if (notifyCode == LBN_SELCHANGE) {
        DoCallback("handleListChanged", "(I)V", GetCurrentSelection());
    }
    else if (notifyCode == LBN_DBLCLK) {
        DoCallback("handleAction", "(I)V", GetCurrentSelection());
    }
    return mrDoDefault;
}

MsgRouting
AwtList::WmDrawItem(UINT /*ctrlId*/, DRAWITEMSTRUCT far& drawInfo)
{
    AwtComponent::DrawListItem((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2), drawInfo);
    return mrConsume;
}

MsgRouting 
AwtList::WmMeasureItem(UINT /*ctrlId*/, MEASUREITEMSTRUCT far& measureInfo)
{
    AwtComponent::MeasureListItem((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2), measureInfo);
    return mrConsume;
}


/************************************************************************
 * WListPeer native methods
 *
 * This class seems to have numerous bugs in it, but they are all bugs
 * which were present before conversion to JNI. -br.  
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    getMaxWidth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WListPeer_getMaxWidth(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN_NULL(self);

    AwtList* l = (AwtList*)pData;
    return l->GetMaxWidth();
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    addItem
 * Signature: (Ljava/lang/String;I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WListPeer_updateMaxItemWidth(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);

    AwtList* l = (AwtList*)pData;
    l->UpdateMaxItemWidth();
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    addItem
 * Signature: (Ljava/lang/String;II)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WListPeer_addItem0(JNIEnv *env, jobject self,
				       jstring item, jint index, jint width )
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(item, "null item");

    char *itemPtr = (char *)JNU_GetStringPlatformChars(env, item, 0);
    AwtList* l = (AwtList*)pData;
    l->InsertString(index, itemPtr);
    JNU_ReleaseStringPlatformChars(env, item, itemPtr);
    l->CheckMaxWidth(width);
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    delItems
 * Signature: (II)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WListPeer_delItems(JNIEnv *env, jobject self,
					jint start, jint end)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);

    AwtList* l = (AwtList*)pData;
    int count = l->GetCount();

    if (start == 0 && end == count) {
        l->SendMessage(LB_RESETCONTENT);
    }
    else for (int i = start; i <= end; i++) {
        l->SendMessage(LB_DELETESTRING, start);
    }

    l->UpdateMaxItemWidth();
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    select
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WListPeer_select(JNIEnv *env, jobject self,
				      jint pos)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtList* l = (AwtList*)pData;
    l->Select(pos);
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    deselect
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WListPeer_deselect(JNIEnv *env, jobject self,
					jint pos)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtList* l = (AwtList*)pData;
    l->Deselect(pos);
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    makeVisible
 * Signature: (I)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WListPeer_makeVisible(JNIEnv *env, jobject self,
					   jint pos)
{

    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtList* l = (AwtList*)pData;
    l->SendMessage(LB_SETTOPINDEX, pos);
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    setMultipleSelections
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WListPeer_setMultipleSelections(JNIEnv *env, jobject self,
						     jboolean on)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtList* l = (AwtList*)pData;
    AwtToolkit::GetInstance().SendMessage(WM_AWT_LIST_SETMULTISELECT,
					  (WPARAM) l, on);
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WListPeer_create(JNIEnv *env, jobject self,
				      jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent, 
				(AwtToolkit::ComponentFactory)AwtList::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

/*
 * Class:     sun_awt_windows_WListPeer
 * Method:    isSelected
 * Signature: (I)Z
 */
JNIEXPORT jboolean JNICALL
Java_sun_awt_windows_WListPeer_isSelected(JNIEnv *env, jobject self,
					  jint index)
{
    int pData;
    JNI_CHECK_PEER_RETURN_NULL(self);
    AwtList* l = (AwtList*)pData;
    return l->IsItemSelected(index);
}

} /* extern "C" */





