/*
 * @(#)awt_MenuBar.cpp	1.26 98/09/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_MenuBar.h"
#include "awt_Frame.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtMenuBar fields
 */

jmethodID AwtMenuBar::getMenuMID;
jmethodID AwtMenuBar::getMenuCountMID;


/************************************************************************
 * AwtMenuBar methods
 */


AwtMenuBar::AwtMenuBar() {
    m_frame = NULL;
}

AwtMenuBar::~AwtMenuBar() {
    m_frame = NULL;
}

const char* AwtMenuBar::GetClassName() {
  return "SunAwtMenuBar";
}

/* Create a new AwtMenuBar object and menu.   */
AwtMenuBar* AwtMenuBar::Create(jobject self, jobject framePeer)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    AwtMenuBar* menuBar;
    /* target is a java.awt.MenuBar */
    jobject target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    menuBar = new AwtMenuBar();
    JNI_CHECK_NULL_GOTO(menuBar, "AwtMenuBar() failed", done);
    menuBar->LinkObjects(env, self);
    {
        HMENU hMenu = ::CreateMenu();
	ASSERT(hMenu);

	menuBar->SetHMenu(hMenu);
	if (framePeer != NULL) {
	    int pData;
	    JNI_CHECK_PEER_GOTO(framePeer, done);
	    menuBar->m_frame = (AwtFrame *)pData;
	} else {
	    menuBar->m_frame = NULL;
	}
    }
 done:
    env->DeleteLocalRef(target);

    return menuBar;
}

HWND AwtMenuBar::GetOwnerHWnd()
{
    AwtFrame *myFrame = m_frame;
    if (myFrame == NULL)
        return NULL;
    else
        return myFrame->GetHWnd();
}

void AwtMenuBar::SendDrawItem(AwtMenuItem* awtMenuItem,
			      DRAWITEMSTRUCT far& drawInfo)
{
    awtMenuItem->DrawItem(drawInfo);
}

void AwtMenuBar::SendMeasureItem(AwtMenuItem* awtMenuItem,
				 HDC hDC, MEASUREITEMSTRUCT far& measureInfo)
{
    awtMenuItem->MeasureItem(hDC, measureInfo);
}

int AwtMenuBar::CountItem(jobject menuBar)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    int nCount = env->CallIntMethod(menuBar, AwtMenuBar::getMenuCountMID);
    ASSERT(!env->ExceptionOccurred());

    return nCount;
}

AwtMenuItem* AwtMenuBar::GetItem(jobject target, long index)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0) {
	return NULL;
    }

    jobject menu = env->CallObjectMethod(target, AwtMenuBar::getMenuMID,index);
    ASSERT(!env->ExceptionOccurred());

    jobject menuItemPeer = GetPeerForTarget(env, menu);
    int pData;
    JNI_CHECK_PEER_RETURN_NULL(menuItemPeer);
    AwtMenuItem* awtMenuItem = (AwtMenuItem*)pData;

    env->DeleteLocalRef(menu);
    env->DeleteLocalRef(menuItemPeer);

    return awtMenuItem;
}

void AwtMenuBar::DrawItem(DRAWITEMSTRUCT far& drawInfo)
{
    ASSERT(drawInfo.CtlType == ODT_MENU);
    AwtMenu::DrawItems(drawInfo);
}

void AwtMenuBar::MeasureItem(HDC hDC,
			     MEASUREITEMSTRUCT far& measureInfo)
{
    ASSERT(measureInfo.CtlType == ODT_MENU);
    AwtMenu::MeasureItem(hDC, measureInfo);
}

void AwtMenuBar::AddItem(AwtMenuItem* item)
{
    AwtMenu::AddItem(item);
    HWND hOwnerWnd = GetOwnerHWnd();
    if (hOwnerWnd != NULL) {
        VERIFY(::InvalidateRect(hOwnerWnd,0,TRUE));
    }
}

void AwtMenuBar::DeleteItem(long index)
{
    AwtMenu::DeleteItem(index);
    HWND hOwnerWnd = GetOwnerHWnd();
    if (hOwnerWnd != NULL) {
        VERIFY(::InvalidateRect(hOwnerWnd,0,TRUE));
    }
    ::DrawMenuBar(GetOwnerHWnd());
}

LONG AwtMenuBar::WinThreadExecProc(ExecuteArgs * args)
{
    switch( args->cmdId ) {
	case MENUBAR_DELITEM:
	    this->DeleteItem(args->param1);
	    break;

	default:
	    AwtMenu::WinThreadExecProc(args);
	    break;
    }
    return 0L;
}

/************************************************************************
 * MenuBar native methods
 */

extern "C" {

/*
 * Class:     java_awt_MenuBar
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_MenuBar_initIDs(JNIEnv *env, jclass cls)
{
    AwtMenuBar::getMenuCountMID = env->GetMethodID(cls, "getMenuCountImpl", "()I");
    AwtMenuBar::getMenuMID = env->GetMethodID(cls, "getMenuImpl",
					      "(I)Ljava/awt/Menu;");
    ASSERT(AwtMenuBar::getMenuCountMID != NULL);
    ASSERT(AwtMenuBar::getMenuMID != NULL);
}

} /* extern "C" */


/************************************************************************
 * WMenuBarPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WMenuBarPeer
 * Method:    addMenu
 * Signature: (Ljava/awt/Menu;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuBarPeer_addMenu(JNIEnv *env, jobject self,
					  jobject menu)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtMenuBar* menubar = (AwtMenuBar*)pData;

    /* The menu was already created and added during peer creation -- redraw */
    ::DrawMenuBar(menubar->GetOwnerHWnd());
}

/*
 * Class:     sun_awt_windows_WMenuBarPeer
 * Method:    delMenu
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuBarPeer_delMenu(JNIEnv *env, jobject self,
					  jint index)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtObject::WinThreadExec(self, AwtMenuBar::MENUBAR_DELITEM, (LPARAM)index);
}

/*
 * Class:     sun_awt_windows_WMenuBarPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WFramePeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuBarPeer_create(JNIEnv *env, jobject self,
					 jobject frame)
{
    AwtToolkit::CreateComponent(self, frame,
				(AwtToolkit::ComponentFactory)
				AwtMenuBar::Create);
    int pData;
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

} /* extern "C" */
