/*
 * @(#)awt_MenuItem.cpp	1.37 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_MenuItem.h"
#include "awt_Menu.h"
#include "awt_MenuBar.h"
#include <sun_awt_windows_WCheckboxMenuItemPeer.h>

// Begin -- Win32 SDK include files
#include <tchar.h>
#include <imm.h>
#include <ime.h>
// End -- Win32 SDK include files

//add for multifont menuitem
#include <java_awt_CheckboxMenuItem.h>
#include <java_awt_Toolkit.h>
#include <java_awt_event_InputEvent.h>

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtMenuItem fields
 */

HBITMAP AwtMenuItem::bmpCheck;
jobject AwtMenuItem::systemFont;

jfieldID AwtMenuItem::labelID;
jfieldID AwtMenuItem::enabledID;
jfieldID AwtMenuItem::fontID;
jfieldID AwtMenuItem::shortcutLabelID;
jfieldID AwtMenuItem::isCheckboxID;
jfieldID AwtMenuItem::stateID;

jmethodID AwtMenuItem::getDefaultFontMID;

// Added by waleed to initialize the RTL Flags
LANGID AwtMenuItem::m_idLang = LOWORD(GetKeyboardLayout(0));
UINT AwtMenuItem::m_CodePage =
    AwtMenuItem::LangToCodePage(AwtMenuItem::m_idLang);
BOOL AwtMenuItem::m_isWin95 = IS_WIN95;
BOOL AwtMenuItem::sm_rtl = PRIMARYLANGID(GetInputLanguage()) == LANG_ARABIC ||
                           PRIMARYLANGID(GetInputLanguage()) == LANG_HEBREW;
BOOL AwtMenuItem::sm_rtlReadingOrder = 
    PRIMARYLANGID(GetInputLanguage()) == LANG_ARABIC;

/************************************************************************
 * AwtMenuItem methods
 */

AwtMenuItem::AwtMenuItem() {
    m_peerObject = NULL;
    m_menuContainer = NULL;
    m_Id = (UINT)-1;
}

AwtMenuItem::~AwtMenuItem() {
    CriticalSection::Lock l(GetLock());
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (m_peerObject != NULL) {
	JNI_SET_PDATA(m_peerObject, NULL);
	env->DeleteGlobalRef(m_peerObject);
    }
}

const char* AwtMenuItem::GetClassName() {
  return "SunAwtMenuItem";
}
// Convert Language ID to CodePage
UINT AwtMenuItem::LangToCodePage(LANGID idLang)
{
    TCHAR strCodePage[MAX_ACP_STR_LEN];
    // use the LANGID to create a LCID
    LCID idLocale = MAKELCID(idLang, SORT_DEFAULT);
    // get the ANSI code page associated with this locale
    if (GetLocaleInfo(idLocale, LOCALE_IDEFAULTANSICODEPAGE, strCodePage, sizeof(strCodePage)) > 0 )
        return _ttoi(strCodePage);
    else
        return GetACP();
}

/*
 * Link the C++, Java peer together
 */
void AwtMenuItem::LinkObjects(JNIEnv *env, jobject peer)
{
    m_peerObject = env->NewGlobalRef(peer);
    JNI_SET_PDATA(peer, this);
}

AwtMenuItem* AwtMenuItem::Create(jobject peer, jobject menuPeer)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }
    int pData;
    JNI_CHECK_PEER_RETURN_NULL(menuPeer);

    /* target is a java.awt.MenuItem  */
    jobject target = env->GetObjectField(peer, AwtObject::targetID);

    AwtMenu* menu = (AwtMenu *)pData;
    AwtMenuItem* item = new AwtMenuItem();

    item->LinkObjects(env, peer);
    item->SetMenuContainer(menu);
    item->SetID(AwtToolkit::GetInstance().CreateCmdID(item));
    menu->AddItem(item);

    env->DeleteLocalRef(target);

    return item;
}

MsgRouting AwtMenuItem::WmNotify(UINT notifyCode)
{
    return mrDoDefault;
}

// This function returns a local reference
jobject
AwtMenuItem::GetFont(JNIEnv *env)
{
    if (env->EnsureLocalCapacity(4) < 0) {
	return NULL;
    }
    jobject self = GetPeer(env);
    jobject target = env->GetObjectField(self, AwtObject::targetID);
    jobject font = env->GetObjectField(target, AwtMenuItem::fontID);

    if (font != NULL) {
      goto done;
    }
    if (AwtMenuItem::systemFont == NULL) {
        env->PushLocalFrame(2);
	jclass cls = env->FindClass("sun/awt/windows/WMenuItemPeer");
	ASSERT(cls != NULL);
	AwtMenuItem::systemFont =
	    env->CallStaticObjectMethod(cls, AwtMenuItem::getDefaultFontMID);
	ASSERT(AwtMenuItem::systemFont);
	jstring fontName =
	    (jstring)JNU_CallMethodByName(env, 0, AwtMenuItem::systemFont,
					  "getName",
					  "()Ljava/lang/String;").l;
	AwtMenuItem::systemFont = env->NewGlobalRef(AwtMenuItem::systemFont);
	env->PopLocalFrame(0);
    }
    font = env->NewLocalRef(AwtMenuItem::systemFont);

done:
    env->DeleteLocalRef(target);
    return font;
}

void
AwtMenuItem::DrawSelf(DRAWITEMSTRUCT far& drawInfo)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(4) < 0) {
	return;
    }

    // self is sun.awt.windows.WMenuItemPeer
    jobject self = GetPeer(env);

    //  target is java.awt.MenuItem
    jobject target = env->GetObjectField(self, AwtObject::targetID);

    HDC hDC = drawInfo.hDC;
    RECT rect = drawInfo.rcItem;
    RECT textRect = rect;
    SIZE size;

    DWORD crBack,crText;
    HBRUSH hbrBack;

    jobject font = GetFont(env);
    jstring text = GetJavaString(env);
    size = AwtFont::getMFStringSize(hDC, font, text);

    //Check whether the MenuItem is disabled.
    BOOL bEnabled = (jboolean)env->GetBooleanField(target,
						   AwtMenuItem::enabledID);

    if ((drawInfo.itemState) & (ODS_SELECTED)) {
	// Set background and text colors for selected item
	crBack = ::GetSysColor (COLOR_HIGHLIGHT);
        // Disabled text must be drawn in gray.
	crText = ::GetSysColor(bEnabled? COLOR_HIGHLIGHTTEXT : COLOR_GRAYTEXT);
    } else {
	// Set background and text colors for unselected item
	crBack = ::GetSysColor (COLOR_MENU);
        // Disabled text must be drawn in gray.
	crText = ::GetSysColor (bEnabled ? COLOR_MENUTEXT : COLOR_GRAYTEXT);
    }

    // Fill item rectangle with background color
    hbrBack = ::CreateSolidBrush (crBack);
    ASSERT(hbrBack);
    VERIFY(::FillRect (hDC, &rect, hbrBack));
    VERIFY(::DeleteObject (hbrBack));

    // Set current background and text colors
    ::SetBkColor (hDC, crBack);
    ::SetTextColor (hDC, crText);

    int nOldBkMode = ::SetBkMode(hDC, OPAQUE);
    ASSERT(nOldBkMode != 0);

    //draw check mark
    int checkWidth = ::GetSystemMetrics(SM_CXMENUCHECK);
    jboolean isCheckbox =
	(jboolean)env->GetBooleanField(self, AwtMenuItem::isCheckboxID);

    if (isCheckbox) {
    // means that target is a java.awt.CheckboxMenuItem
	jboolean state = (jboolean)env->GetBooleanField(target,
							AwtMenuItem::stateID);
	if (state) {
            ASSERT(drawInfo.itemState & ODS_CHECKED);
            RECT checkRect;
            ::CopyRect(&checkRect, &textRect);
            if (GetRTL())
	        checkRect.left = checkRect.right - checkWidth;
	    else
	        checkRect.right = checkRect.left + checkWidth;

            DrawCheck(hDC, checkRect);
	}
    }

    ::SetBkMode(hDC, TRANSPARENT);
    int x = 0;
    //draw string
    if (!IsTopMenu()){
        textRect.left += checkWidth;
	x = (GetRTL()) ? textRect.right - checkWidth - size.cx : textRect.left;
    } else {
        x = textRect.left = (textRect.left + textRect.right - size.cx) / 2;
    }
    
    int y = (textRect.top+textRect.bottom-size.cy)/2;

    // Text must be drawn in emboss if the Menu is disabled and not selected.
    BOOL bEmboss = !bEnabled && !(drawInfo.itemState & ODS_SELECTED);
    if (bEmboss) {
        ::SetTextColor(hDC, GetSysColor(COLOR_BTNHILIGHT));
        AwtFont::drawMFString(hDC, font, text, x + 1, y + 1, GetCodePage());
        ::SetTextColor(hDC, GetSysColor(COLOR_BTNSHADOW));
    }
    AwtFont::drawMFString(hDC, font, text, x, y, GetCodePage());

    jstring shortcutLabel =
	(jstring)env->GetObjectField(self, AwtMenuItem::shortcutLabelID);
    if (!IsTopMenu() && shortcutLabel != NULL) {
        UINT oldAlign = 0;
	if (GetRTL()){
	    oldAlign = ::SetTextAlign(hDC, TA_LEFT);
	    AwtFont::drawMFString(hDC, font, shortcutLabel, textRect.left, y,
				  GetCodePage());
	} else {
	    oldAlign = ::SetTextAlign(hDC, TA_RIGHT);
	    AwtFont::drawMFString(hDC, font, shortcutLabel,
				  textRect.right - checkWidth, y, 
				  GetCodePage());
	}

        ::SetTextAlign(hDC, oldAlign);
    }

    VERIFY(::SetBkMode(hDC,nOldBkMode));

    env->DeleteLocalRef(target);
    env->DeleteLocalRef(text);
    env->DeleteLocalRef(font);
    env->DeleteLocalRef(shortcutLabel);
}


void AwtMenuItem::DrawItem(DRAWITEMSTRUCT far& drawInfo)
{
    ASSERT(drawInfo.CtlType == ODT_MENU);

    if (drawInfo.itemID != m_Id)
	return;

    DrawSelf(drawInfo);
}

void AwtMenuItem::MeasureSelf(HDC hDC, MEASUREITEMSTRUCT far& measureInfo)
{
    JNIEnv *env =(JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(4) < 0) {
	return;
    }

    /* self is a sun.awt.windows.WMenuItemPeer */
    jobject self = GetPeer(env);

    /* font is a java.awt.Font */
    jobject font = GetFont(env);

    jstring fontName =
	(jstring)JNU_CallMethodByName(env, 0,font, "getName",
				      "()Ljava/lang/String;").l;
    /* fontMetrics is a Hsun_awt_windows_WFontMetrics */
    jobject fontMetrics =  GetFontMetrics(env, font);
    jstring text = GetJavaString(env);

    SIZE size = AwtFont::getMFStringSize(hDC, font, text);

//     int height = env->GetIntField(fontMetrics, AwtFont::heightID);
    int height = (jint)JNU_CallMethodByName(env, 0, fontMetrics, "getHeight",
                                            "()I").i;

    measureInfo.itemHeight = height;
    measureInfo.itemHeight += measureInfo.itemHeight/3;
    // 3 is a heuristic number
    measureInfo.itemWidth = size.cx;
    if (!IsTopMenu()) {
        int checkWidth = ::GetSystemMetrics(SM_CXMENUCHECK);
	measureInfo.itemWidth += checkWidth;

        // Add in shortcut width, if one exists.
	jstring shortcutLabel =
	    (jstring)env->GetObjectField(self, AwtMenuItem::shortcutLabelID);
        if (shortcutLabel != NULL) {
            size = AwtFont::getMFStringSize(hDC, font, shortcutLabel);
            measureInfo.itemWidth += size.cx + checkWidth;
	    env->DeleteLocalRef(shortcutLabel);
        }
    }
    env->DeleteLocalRef(text);
    env->DeleteLocalRef(font);
    env->DeleteLocalRef(fontName);
    env->DeleteLocalRef(fontMetrics);
}

void AwtMenuItem::MeasureItem(HDC hDC, MEASUREITEMSTRUCT far& measureInfo)
{
    ASSERT(measureInfo.CtlType == ODT_MENU);

    if (measureInfo.itemID != m_Id)
	return;

    MeasureSelf(hDC, measureInfo);
}

jobject AwtMenuItem::GetFontMetrics(JNIEnv *env, jobject font)
{
    static jobject toolkit = NULL;
    if (toolkit == NULL) {
        if (env->PushLocalFrame(2) < 0)
	    return NULL;
	jclass cls = env->FindClass("java/awt/Toolkit");
	jobject toolkitLocal =
	    env->CallStaticObjectMethod(cls, AwtToolkit::getDefaultToolkitMID);
	toolkit = env->NewGlobalRef(toolkitLocal);
	ASSERT(!env->ExceptionOccurred());
	env->PopLocalFrame(0);
    }
    /*
    JNU_PrintClass(env, "toolkit", toolkit);
    JNU_PrintClass(env, "font", font);

    jclass cls = env->FindClass("java/awt/Toolkit");
    jmethodID mid = env->GetMethodID(cls, "getFontMetrics",
				     "(Ljava/awt/Font;)Ljava/awt/FontMetrics;");
    jstring fontName =
	(jstring)JNU_CallMethodByName(env, 0,font, "getName",
				      "()Ljava/lang/String;").l;
    JNU_PrintString(env, "font name", fontName);

    fprintf(stderr, "mid: %x\n", mid);
    fprintf(stderr, "cached mid: %x\n", AwtToolkit::getFontMetricsMID);
    ASSERT(!env->ExceptionOccurred());
    */
    jobject fontMetrics =
      env->CallObjectMethod(toolkit, AwtToolkit::getFontMetricsMID, font);
    ASSERT(!env->ExceptionOccurred());

    return fontMetrics;
}

BOOL AwtMenuItem::IsTopMenu()
{
    return FALSE;
}

void AwtMenuItem::DrawCheck(HDC hDC, RECT rect)
{
    if (bmpCheck == NULL) {
        bmpCheck = ::LoadBitmap(AwtToolkit::GetInstance().GetModuleHandle(),
                                "CHECK_BITMAP");
        ASSERT(bmpCheck != NULL);
    }

#define BM_SIZE 26  /* height and width of check.bmp */

    // Square the rectangle, so the check is proportional.
    int width = rect.right - rect.left;
    int diff = max(rect.bottom - rect.top - width, 0) ;
    int bottom = diff / 2;
    rect.bottom -= bottom;
    rect.top += diff - bottom;

    HDC hdcBitmap = ::CreateCompatibleDC(hDC);
    ASSERT(hdcBitmap != NULL);
    HBITMAP hbmSave = ::SelectObject(hdcBitmap, bmpCheck);
    VERIFY(::StretchBlt(hDC, rect.left, rect.top,
                        rect.right - rect.left, rect.bottom - rect.top,
                        hdcBitmap, 0, 0, BM_SIZE, BM_SIZE, SRCCOPY));
    ::SelectObject(hdcBitmap, hbmSave);
    VERIFY(::DeleteDC(hdcBitmap));
}

void AwtMenuItem::DoCommand()
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    // peer is sun.awt.windows.WMenuItemPeer
    jobject peer = GetPeer(env);

    // Determine which modifier keys have been pressed.
    long modifiers = 0;
    if (HIBYTE(::GetAsyncKeyState(VK_CONTROL))) {
        modifiers |= java_awt_event_InputEvent_CTRL_MASK;
    }
    if (HIBYTE(::GetAsyncKeyState(VK_RBUTTON))) {
        modifiers |= java_awt_event_InputEvent_META_MASK;
    }
    if (HIBYTE(::GetAsyncKeyState(VK_SHIFT))) {
        modifiers |= java_awt_event_InputEvent_SHIFT_MASK;
    }
    if (HIBYTE(::GetAsyncKeyState(VK_MBUTTON))) {
        modifiers |= java_awt_event_InputEvent_ALT_MASK;
    }

    jboolean isCheckbox =
      (jboolean)env->GetBooleanField(peer, AwtMenuItem::isCheckboxID);
    if (isCheckbox) {
        UINT nState = ::GetMenuState(GetMenuContainer()->GetHMenu(),
                                     GetID(), MF_BYCOMMAND);
        ASSERT(nState != 0xFFFFFFFF);
        DoCallback("handleAction", "(ZI)V", ((nState & MF_CHECKED) == 0),
                   modifiers);
    } else {
        DoCallback("handleAction", "(I)V", modifiers);
    }
}

void AwtMenuItem::SetLabel(const char *sb)
{
    AwtMenu* menu = GetMenuContainer();
    ASSERT(menu != NULL && GetID() >= 0);

    MENUITEMINFO mii;
    mii.cbSize = sizeof mii;
    mii.fMask = MIIM_TYPE;
    mii.fType = MFT_OWNERDRAW;
    mii.dwTypeData = (char *)(*sb);
    SetMenuItemInfo(menu->GetHMenu(), GetID(), FALSE, &mii);

    // Redraw menu bar if it was affected.
    if (menu->GetMenuBar() == menu) {
        ::DrawMenuBar(menu->GetOwnerHWnd());
    }
}

void AwtMenuItem::Enable(BOOL isEnabled)
{
    AwtMenu* menu = GetMenuContainer();
    ASSERT(menu != NULL && GetID() >= 0);
    VERIFY(::EnableMenuItem(menu->GetHMenu(), GetID(),
                            MF_BYCOMMAND | (isEnabled ? MF_ENABLED : MF_GRAYED))
           != 0xFFFFFFFF);

    // Redraw menu bar if it was affected.
    if (menu->GetMenuBar() == menu) {
        ::DrawMenuBar(menu->GetOwnerHWnd());
    }
}

void AwtMenuItem::SetState(BOOL isChecked)
{
    AwtMenu* menu = GetMenuContainer();
    ASSERT(menu != NULL && GetID() >= 0);
    VERIFY(::CheckMenuItem(menu->GetHMenu(), GetID(),
                           MF_BYCOMMAND | (isChecked ? MF_CHECKED : MF_UNCHECKED))
           != 0xFFFFFFFF);

    // Redraw menu bar if it was affected.
    if (menu->GetMenuBar() == menu) {
        ::DrawMenuBar(menu->GetOwnerHWnd());
    }
}

void AwtMenuItem::Dispose()
{
    AWT_TRACE(("AwtMenuItem::Dispose %x\n", this));
    if (strcmp(GetClassName(), "SunAwtMenuItem") == 0) {
	AwtToolkit::GetInstance().RemoveCmdID( GetID() );
    }
    delete this;
}

LONG AwtMenuItem::WinThreadExecProc(ExecuteArgs * args)
{
    switch( args->cmdId ) {
	case MENUITEM_SETLABEL:
	{
	    const char * sb = (const char *)args->param1;
	    ASSERT(!IsBadStringPtr(sb, 20));
	    this->SetLabel(sb);
	}
	break;

	case MENUITEM_ENABLE:
	{
	    BOOL	isEnabled = (BOOL)args->param1;
	    this->Enable(isEnabled);
	}
	break;

	case MENUITEM_SETSTATE:
	{
	    BOOL	isChecked = (BOOL)args->param1;
	    this->SetState(isChecked);
	}
	break;

	case MENUITEM_DISPOSE:
	{
	    this->Dispose();
	}
	break;

	default:
	    AwtObject::WinThreadExecProc(args);
	    break;
    }
    return 0L;
}


/************************************************************************
 * MenuComponent native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_MenuComponent_initIDs(JNIEnv *env, jclass cls)
{
    AwtMenuItem::fontID = env->GetFieldID(cls, "font", "Ljava/awt/Font;");
    ASSERT(AwtMenuItem::fontID != NULL);
}

} /* extern "C" */


/************************************************************************
 * MenuItem native methods
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_MenuItem_initIDs(JNIEnv *env, jclass cls)
{
    AwtMenuItem::labelID = env->GetFieldID(cls, "label", "Ljava/lang/String;");
    AwtMenuItem::enabledID = env->GetFieldID(cls, "enabled", "Z");

    ASSERT(AwtMenuItem::labelID != NULL);
    ASSERT(AwtMenuItem::enabledID != NULL);
}

} /* extern "C" */


/************************************************************************
 * CheckboxMenuItem fields
 */

extern "C" {

JNIEXPORT void JNICALL
Java_java_awt_CheckboxMenuItem_initIDs(JNIEnv *env, jclass cls)
{
    AwtMenuItem::stateID = env->GetFieldID(cls, "state", "Z");

    ASSERT(AwtMenuItem::stateID != NULL);
}

} /* extern "C" */


/************************************************************************
 * WMenuItemPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WMenuItemPeer
 * Method:    _setLabel
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuItemPeer_initIDs(JNIEnv *env, jclass cls)
{
    AwtMenuItem::isCheckboxID = env->GetFieldID(cls, "isCheckbox", "Z");
    AwtMenuItem::shortcutLabelID = env->GetFieldID(cls, "shortcutLabel",
						   "Ljava/lang/String;");
    AwtMenuItem::getDefaultFontMID =
	env->GetStaticMethodID(cls, "getDefaultFont", "()Ljava/awt/Font;");

    ASSERT(AwtMenuItem::isCheckboxID != NULL);
    ASSERT(AwtMenuItem::shortcutLabelID != NULL);
    ASSERT(AwtMenuItem::getDefaultFontMID != NULL);
}

/*
 * Class:     sun_awt_windows_WMenuItemPeer
 * Method:    _setLabel
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuItemPeer__1setLabel(JNIEnv *env, jobject self,
					      jstring label)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(label, "null label");
    const char *labelPtr = JNU_GetStringPlatformChars(env, label, 0);
    AwtObject::WinThreadExec(self, AwtMenuItem::MENUITEM_SETLABEL, (LPARAM)labelPtr);
    JNU_ReleaseStringPlatformChars(env, label, labelPtr);
}

/*
 * Class:     sun_awt_windows_WMenuItemPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WMenuPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuItemPeer_create(JNIEnv *env, jobject self,
					  jobject menu)
{
    JNI_CHECK_NULL_RETURN(menu, "null Menu");
    AwtToolkit::CreateComponent(self, menu,
				(AwtToolkit::ComponentFactory)
				AwtMenuItem::Create);
    int pData;
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

/*
 * Class:     sun_awt_windows_WMenuItemPeer
 * Method:    enable
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuItemPeer_enable(JNIEnv *env, jobject self,
					  jboolean on)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtObject::WinThreadExec(self, AwtMenuItem::MENUITEM_ENABLE, (LPARAM)on );
}

/*
 * Class:     sun_awt_windows_WMenuItemPeer
 * Method:    _dispose
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WMenuItemPeer__1dispose(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtMenuItem *p = (AwtMenuItem *)pData;

    if (p != NULL) {
	AwtToolkit::GetInstance().SendMessage(WM_AWT_DISPOSE, (WPARAM)p, 0);
    }
}

} /* extern "C" */

/************************************************************************
 * WCheckboxMenuItemPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WCheckboxMenuItemPeer
 * Method:    setState
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WCheckboxMenuItemPeer_setState(JNIEnv *env, jobject self,
						    jboolean on)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtObject::WinThreadExec(self, AwtMenuItem::MENUITEM_SETSTATE, (LPARAM)on);
}

} /* extern "C" */
