/*
 * @(#)awt_Object.h	1.16 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_OBJECT_H
#define AWT_OBJECT_H

#include "awt.h"
#include "awt_Toolkit.h"

#include "java_awt_Event.h"
#include "java_awt_AWTEvent.h"
#include "sun_awt_windows_WObjectPeer.h"


/************************************************************************
 * AwtObject class
 */

class AwtObject {
public:
    class ExecuteArgs {
	public:
	    UINT	cmdId;
	    LPARAM	param1;
	    LPARAM	param2;
	    LPARAM	param3;
	    LPARAM	param4;
    };


    /* sun.awt.windows.WObjectPeer field and method ids */
    static jfieldID pDataID;
    static jfieldID targetID;

    static jmethodID getPeerForTargetMID;

    AwtObject();
    virtual ~AwtObject();

    // Return the associated AWT peer or target object.
    INLINE jobject GetPeer(JNIEnv *env) { 
        return m_peerObject;
    }

    INLINE jobject GetTarget(JNIEnv *env) { 
        jobject peer = GetPeer(env);
	if (peer != NULL) {
	    return env->GetObjectField(peer, AwtObject::targetID);
	} else {
	    return NULL;
	}
    }

    // Return the peer associated with some target
    jobject GetPeerForTarget(JNIEnv *env, jobject target);

    // Java callback routines
    // Invoke a callback on the java peer object asynchronously
    void DoCallback(const char* methodName, const char* methodSig, ...);

    // Allocate and initialize a new event, and post it to the peer's
    // target object.  No response is expected from the target.
    void SendEvent(jobject event);

    INLINE void EnableCallbacks(BOOL e) { m_callbacksEnabled = e; }

    // Execute any code associated with a command ID -- only classes with
    // DoCommand() defined should associate their instances with cmdIDs.
    virtual void DoCommand(void) {
        ASSERT(FALSE);
    }

    // execute given code on Windows message-pump thread
    static LONG WinThreadExec(jobject peerObject, UINT cmdId, LPARAM param1 = 0L, LPARAM param2 = 0L, LPARAM param3 = 0L, LPARAM param4 = 0L);
    // callback function to execute code on Windows message-pump thread
    virtual LONG WinThreadExecProc(AwtObject::ExecuteArgs * args);

protected:
    jobject                       m_peerObject;
    BOOL                          m_callbacksEnabled;
};

#endif // AWT_OBJECT_H


