/*
 * @(#)awt_PopupMenu.cpp	1.21 98/09/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_PopupMenu.h"

#include "awt_Event.h"

#include <java_awt_PopupMenu.h>
#include <sun_awt_windows_WPopupMenuPeer.h>
#include <java_awt_Event.h>

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtPopupMenu class methods
 */

AwtPopupMenu::AwtPopupMenu() {
    m_parent = NULL;
}

AwtPopupMenu::~AwtPopupMenu() {
    m_parent = NULL;
}

const char* AwtPopupMenu::GetClassName() {
  return "SunAwtPopupMenu";
}

/* Create a new AwtPopupMenu object and menu.   */
AwtPopupMenu* AwtPopupMenu::Create(jobject self, AwtComponent* parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    AwtPopupMenu* popupMenu = 0;
    jobject target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    popupMenu = new AwtPopupMenu();
    JNI_CHECK_NULL_GOTO(popupMenu, "AwtPopupMenu() failed", done);
    {
        popupMenu->LinkObjects(env, self);

	HMENU hMenu = ::CreatePopupMenu();
	ASSERT(hMenu);
	popupMenu->SetHMenu(hMenu);
	popupMenu->SetParent(parent);
    }
 done:
    env->DeleteLocalRef(target);
    return popupMenu;
}

void AwtPopupMenu::Show(JNIEnv *env, jobject event)
{
    /*
     * Convert the event's XY to absolute coordinates.  The XY is
     * relative to the origin component, which is passed by PopupMenu
     * as the event's target.
     */
    if (env->EnsureLocalCapacity(2) < 0) {
	return;
    }
    jobject origin = (env)->GetObjectField(event, AwtEvent::targetID);
    jobject peerOrigin = GetPeerForTarget(env, origin);
    int pData;
    JNI_CHECK_PEER_GOTO(peerOrigin, done);
    {
        AwtComponent* awtOrigin = (AwtComponent*)pData;
        POINT pt;
        pt.x = (env)->GetIntField(event, AwtEvent::xID);
        pt.y = (env)->GetIntField(event, AwtEvent::yID);
        ::MapWindowPoints(awtOrigin->GetHWnd(), 0, (LPPOINT)&pt, 1);

        // Adjust to account for the Inset values
        RECT rctInsets;
        awtOrigin->GetInsets(&rctInsets);
        pt.x -= rctInsets.left;
        pt.y -= rctInsets.top;

        /* Invoke the popup. */
        ::TrackPopupMenu( m_hMenu, TPM_LEFTALIGN | TPM_RIGHTBUTTON,
        pt.x, pt.y, 0, awtOrigin->GetHWnd(), NULL );
    }
 done:
    env->DeleteLocalRef(origin);
    env->DeleteLocalRef(peerOrigin);
}


/************************************************************************
 * WPopupMenuPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WPopupMenuPeer
 * Method:    createMenu
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPopupMenuPeer_createMenu(JNIEnv *env, jobject self,
					       jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtComponent* awtParent = (AwtComponent *)pData;
    AwtToolkit::CreateComponent(
	self, parent, (AwtToolkit::ComponentFactory)AwtPopupMenu::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

/*
 * Class:     sun_awt_windows_WPopupMenuPeer
 * Method:    _show
 * Signature: (Ljava/awt/Event;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPopupMenuPeer__1show(JNIEnv *env, jobject self,
					   jobject event)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(event, "null Event");

    // Invoke popup on toolkit thread -- yet another Win32 restriction.
    AwtPopupMenu* popupMenu = (AwtPopupMenu*)pData;
    AwtToolkit::GetInstance().SendMessage(WM_AWT_POPUPMENU_SHOW,
					  (WPARAM)popupMenu, (LPARAM)event);
}

} /* extern "C" */
