/*
 * @(#)awt_PrintJob.cpp	1.22 98/09/21
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include <windef.h>
#include <wtypes.h>
#include <winuser.h>
#include <commdlg.h>
#include <winspool.h>

#include "awt.h"
#include "awt_Toolkit.h"
#include "awt_Component.h"
#include "awt_Graphics.h"
#include "awt_PrintJob.h"

#include <sun_awt_windows_WPrintJob.h>
#include <sun_awt_windows_WPrinterJob.h>

#define DEBUG_PRINTING	1

/************************************************************************
 * AwtPrintJob fields
 */

jfieldID AwtPrintJob::graphicsID;
jfieldID AwtPrintJob::titleID;
jfieldID AwtPrintJob::pageDimensionID;
jfieldID AwtPrintJob::pageResolutionID;


/************************************************************************
 * WPrintJob native methods
 */   

extern "C" {

/*** Private Constants ***/

static char *kJavaIntStr = "I";

/*
 * Class Names
 */
static const char *PRINTEREXCEPTION_STR = "java/awt/print/PrinterException";

/*
 * The following strings are the names of instance variables in WPrintJob2D.
 */
static const char *PRINTDC_STR = "mPrintDC";	// The print device context.
static const char *XRES_STR = "mPrintXRes";	// The x dpi.
static const char *YRES_STR = "mPrintYRes";	// The y dpi.

static const char *SETRANGE_STR = "setPageRange";
static const char *SETRANGE_SIG = "(II)V";
/*
 * Methods and fields in the interface awt.print.Pageable.
 */

// public int getNumberOfPages()

static const char *GETNUMPAGES_STR = "getNumberOfPages";
static const char *GETNUMPAGES_SIG = "()I";

// public int getCopies()

static const char *GETCOPIES_STR = "getCopies";
static const char *GETCOPIES_SIG = "()I";

/*
 * Methods and fields in awt.print.PageFormat.
 */

// public Paper getPaper()
static const char *GETPAPER_STR = "getPaper";
static const char *GETPAPER_SIG = "()Ljava/awt/print/Paper;";

// public void setPaper(Paper paper)
static const char *SETPAPER_STR = "setPaper";
static const char *SETPAPER_SIG = "(Ljava/awt/print/Paper;)V";

// public int getOrientation()
static const char *GETORIENT_STR = "getOrientation";
static const char *GETORIENT_SIG = "()I";

// public void setOrientation(int orientation)
static const char *SETORIENT_STR = "setOrientation";
static const char *SETORIENT_SIG = "(I)V";

// public static final int PORTRAIT = 0;
static const char *PORTRAIT_STR = "PORTRAIT";
static const char *PORTRAIT_SIG = "I";

// public static final int LANDSCAPE = 1;
static const char *LANDSCAPE_STR = "LANDSCAPE";
static const char *LANDSCAPE_SIG = "I";

/*
 * Methods in awt.print.Paper.
 */

// public void setSize(double width, double height)
static const char *SETSIZE_STR = "setSize";
static const char *SETSIZE_SIG = "(DD)V";

// protected void setImageableArea(double x, double y, double width,
//						    double height)
static const char *SETIMAGEABLE_STR = "setImageableArea";
static const char *SETIMAGEABLE_SIG = "(DDDD)V";

// public double getWidth()
static const char *GETWIDTH_STR = "getWidth";
static const char *GETWIDTH_SIG = "()D";

// public double getHeight()
static const char *GETHEIGHT_STR = "getHeight";
static const char *GETHEIGHT_SIG = "()D";

/*
 * When manipulating 32 bit RGB values the following constants
 * are useful.
 */

// Full on RGB is white.
static const long kRGB32White = 0x00ffffff;

/* Multiply a Window's MM_HIENGLISH value 
 * (1000th of an inch) by this number to
 * get a value in 72nds of an inch.
 */
static const double HIENGLISH_TO_POINTS = (72.0 / 1000.0);

/* Multiply a Window's MM_HIMETRIC value
 * (100ths of a millimeter) by this
 * number to get a value in 72nds of an inch.
 */
static const double HIMETRIC_TO_POINTS = (72.0 / 2540.0);

/* Multiply a measurement in 1/72's of an inch by this
 * value to convert it to Window's MM_HIENGLISH
 * (1000th of an inch) units.
 */
static const double POINTS_TO_HIENGLISH = (1000.0 * 72.0);

/* Multiply a measurement in 1/72's of an inch by this
 * value to convert it to Window's MM_HIMETRIC
 * (100th of an millimeter) units.
 */
static const double POINTS_TO_HIMETRIC = (2540.0 / 72.0);

/* Multiply a measurement in 1/72's of an inch by this
 * value to convert it to Window's MM_LOMETRIC
 * (10th of an millimeter) units.
 */
static const double POINTS_TO_LOMETRIC = (254.0 / 72.0);

/*** Private Macros ***/

/* A Page Setup paint hook passes a word describing the
   orientation and type of page being displayed in the
   dialog. These macros break the word down into meaningful
   values.
*/
#define PRINTER_TYPE_MASK   (0x0003)
#define PORTRAIT_MASK	    (0x0004)
#define ENVELOPE_MASK	    (0x0008)

#define IS_ENVELOPE(param)  (((param) & ENVELOPE_MASK) != 0)
#define IS_PORTRAIT(param)  (((param) & PORTRAIT_MASK) != 0)

/*	If the Pagable does not know the number of pages in the document,
	then we limit the print dialog to this number of pages.
*/
#define MAX_UNKNOWN_PAGES 9999

/*** Private Variable Types ***/

typedef struct {
    jdouble x;
    jdouble y;
    jdouble width;
    jdouble height;
} RectDouble;

/*** Private Prototypes ***/

static UINT CALLBACK pageDlgHook(HWND hDlg, UINT msg,
				 UINT wParam, LONG lParam);
static UINT CALLBACK printDlgHook(HWND hDlg, UINT msg,
				  UINT wParam, LONG lParam);
static HDC getDefaultPrinterDC(void);
static void pageFormatToSetup(JNIEnv *env, jobject page, PAGESETUPDLG *setup);
static void setOrientationInDevMode(HGLOBAL hDevMode, jboolean isPortrait);
static int bitsToDevice(HDC printDC, jint *image, long destX, long destY,
			long width, long height);
static void retrievePaperInfo(const PAGESETUPDLG *setup, POINT *paperSize,
				RECT *margins, jboolean *isPortrait);
static void setRange(JNIEnv *env, jobject self, jint from, jint to);
static jint getNumPages(JNIEnv *env, jobject doc);
static jint getCopies(JNIEnv *env, jobject printerJob);
static jobject getPaper(JNIEnv *env, jobject page);
static void setPaper(JNIEnv *env, jobject page, jobject paper);
static jboolean isPortraitOrientation(JNIEnv *env, jobject page);
static void setOrientation(JNIEnv *env, jobject page, jboolean isPortrait);
static void getPaperValues(JNIEnv *env, jobject paper, RectDouble *paperSize,
			  RectDouble *margins);
static void setPaperValues(JNIEnv *env, jobject paper, const POINT *paperSize,
			    const RECT *margins, int units);
static long convertFromPoints(double value, int units);
static double convertToPoints(long value, int units);
static void setCapabilities(JNIEnv *env, jobject self, HDC printDC);
static HDC getPrintDC(JNIEnv *env, jobject self);
static void setPrintDC(JNIEnv *env, jobject self, HDC printDC);
static jint getIntField(JNIEnv *env, jobject self, const char *fieldName);
static void setIntField(JNIEnv *env, jobject self,
			    const char *fieldName, jint value);
static jfieldID getIdOfIntField(JNIEnv *env, jobject self,
			    const char *fieldName);

static jint *findNonWhite(jint *image, long x, long width, long height, long *numLinesP);
static jint *findWhite(jint *image, long x, long width, long height, long *numLines);
static void dumpDevMode(HGLOBAL hDevMode);
static void dumpPrinterCaps(HANDLE hDevNames);
static void throwPrinterException(JNIEnv *env, DWORD err);

/*** Code ***/

/*
 * Class:     sun_awt_windows_WPrintJob
 * Method:    end
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPrintJob_initIDs(JNIEnv *env, jclass cls)
{
    AwtPrintJob::graphicsID = 
        env->GetFieldID(cls, "graphics", "Lsun/awt/windows/WPrintGraphics;");
    AwtPrintJob::titleID = env->GetFieldID(cls, "title", "Ljava/lang/String;");
    AwtPrintJob::pageDimensionID =
      env->GetFieldID(cls, "pageDimension", "Ljava/awt/Dimension;");
    AwtPrintJob::pageResolutionID =
      env->GetFieldID(cls, "pageResolution", "I");
}

/*
 * Class:     sun_awt_windows_WPrintJob
 * Method:    _flushPage
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPrintJob_flushPageImpl(JNIEnv *env, jobject self)
{
    if (env->EnsureLocalCapacity(1) < 0) {
        return;
    }
    jobject hWGraphics = env->GetObjectField(self, AwtPrintJob::graphicsID);
    AwtMasterPrinterGraphics* pGraphics = (AwtMasterPrinterGraphics*)
        (env->GetIntField(hWGraphics, AwtGraphics::pDataID));

    if (pGraphics != NULL) {
        // if there is currently no page in progress, EndPage is ignored
        pGraphics->EndPage();
    }

    env->DeleteLocalRef(hWGraphics);
}

/*
 * Class:     sun_awt_windows_WPrintJob
 * Method:    newPage
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPrintJob_newPage(JNIEnv *env, jobject self)
{
    if (env->EnsureLocalCapacity(1) < 0) {
        return;
    }
    jobject hWGraphics = env->GetObjectField(self, AwtPrintJob::graphicsID);
    AwtMasterPrinterGraphics* pGraphics = (AwtMasterPrinterGraphics*)
        (env->GetIntField(hWGraphics, AwtGraphics::pDataID));

    if (pGraphics != NULL) {
        // if there is currently no page in progress, EndPage is ignored
        pGraphics->EndPage();
	pGraphics->StartPage();
    }

    env->DeleteLocalRef(hWGraphics);
}

/*
 * Class:     sun_awt_windows_WPrintJob
 * Method:    _end
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPrintJob_endImpl(JNIEnv *env, jobject self)
{
    if (env->EnsureLocalCapacity(1) < 0) {
        return;
    }
    jobject hWGraphics = env->GetObjectField(self, AwtPrintJob::graphicsID);
    AwtMasterPrinterGraphics* pGraphics = (AwtMasterPrinterGraphics*)
        (env->GetIntField(hWGraphics, AwtGraphics::pDataID));

    if (pGraphics != NULL) {
        // End the final page and the document -- if they were already
        // ended, these commands will be ignored. Don't dispose of the
        // awt_Graphics object, though. That's done during finalization
        // of the WGraphics object.
        pGraphics->EndPage();
        pGraphics->EndPrintJob();
    }

    env->DeleteLocalRef(hWGraphics);
}

/************************************************************************
 * WPrinterJob native methods
 */

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    pageSetup
 * Signature: (Ljava/awt/print/PageFormat;Ljava/awt/print/Printable;)Z
 *
 * 11/28/97 Note: Currently unused, the 'painter' parameter is intended one
 * day to allow the creation of an application drawn page preview within
 * the page setup dialog.
 */
JNIEXPORT jboolean JNICALL Java_sun_awt_windows_WPrinterJob_pageSetup
  (JNIEnv *env, jclass clazz, jobject page, jobject painter) {

    jboolean doIt = JNI_FALSE; // Assume the user will cancel the dialog.
    PAGESETUPDLG setup;

    memset(&setup, 0, sizeof(setup));

    setup.lStructSize = sizeof(setup);
    setup.hwndOwner = NULL;
    setup.hDevMode = NULL;
    setup.hDevNames = NULL;
    setup.Flags = PSD_RETURNDEFAULT | PSD_DEFAULTMINMARGINS;
    // setup.ptPaperSize = 
    // setup.rtMinMargin =
    // setup.rtMargin = 
    setup.hInstance = NULL;
    setup.lCustData = 0;
    setup.lpfnPageSetupHook = pageDlgHook;
    setup.lpfnPagePaintHook = NULL;
    setup.lpPageSetupTemplateName = NULL;
    setup.hPageSetupTemplate = NULL;

    /* Because the return default flag is set, this first call
     * will not display the dialog but will return default
     * hDevMode, hDevName, ptPaperSize, and rtMargin values.
     * In particular we want the hDevMode so that we can force
     * the orientation of the page.
     */
    (void) ::PageSetupDlg(&setup);

    /* Move page size and orientation from the PageFormat object
     * into the Windows setup structure so that the format can
     * be displayed in the dialog.
    */
    pageFormatToSetup(env, page, &setup);

    setup.lpfnPageSetupHook = pageDlgHook;
    setup.Flags = PSD_ENABLEPAGESETUPHOOK
		  | PSD_MARGINS;


    if (::PageSetupDlg(&setup)) {

	jobject paper = getPaper(env, page);

	int units = setup.Flags & PSD_INTHOUSANDTHSOFINCHES ?
						MM_HIENGLISH :
						MM_HIMETRIC;
	POINT paperSize;
	RECT margins;
	jboolean isPortrait;

	/* Get the Windows paper and margins description.
	*/
	retrievePaperInfo(&setup, &paperSize, &margins, &isPortrait);

	/* Convert the Windows' paper and margins description
	 * and place them into a Paper instance.
	 */
	setPaperValues(env, paper, &paperSize, &margins, units);

	/* Put the updated Paper instance and the orientation into
	 * the PageFormat.
	 */
	setPaper(env, page, paper);

	setOrientation(env, page, isPortrait);

	doIt = JNI_TRUE;
    }

    return doIt;
}


/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    jobSetup
 * Signature: (Ljava/awt/print/Pageable;)Z
 */
JNIEXPORT jboolean JNICALL Java_sun_awt_windows_WPrinterJob_jobSetup
  (JNIEnv *env, jobject self, jobject doc) {
    jboolean doIt = JNI_FALSE;
    PRINTDLG job;

    /* If the number of pages in the document is bigger than 
     * Window's can use then we'll truncate it - pretty
     * unlikely.
     */
    jint numPages = getNumPages(env, doc);

    unsigned short winNumPages = numPages < (long )USHRT_MAX
				 ? numPages
				 : USHRT_MAX; 

    job.lStructSize = sizeof(job);
    job.hwndOwner = NULL;
    job.hDevMode = NULL;
    job.hDevNames = NULL;
    job.hDC = NULL;
    job.Flags = PD_ENABLEPRINTHOOK | PD_NOSELECTION | PD_RETURNDC;
    job.nFromPage = 1;
    job.nToPage = winNumPages;
    job.nMinPage = 1;

    /* Make sure the maximum page number is not less than the minimum.
     */
    job.nMaxPage = winNumPages >= job.nMinPage ? winNumPages : job.nMinPage;
    
    job.nCopies = (WORD) getCopies(env, self);
    job.hInstance = NULL;
    job.lCustData = NULL;
    job.lpfnPrintHook = printDlgHook;
    job.lpfnSetupHook = NULL;
    job.lpPrintTemplateName = NULL;
    job.lpSetupTemplateName = NULL;
    job.hPrintTemplate = NULL;
    job.hSetupTemplate = NULL;
	
    if (::PrintDlg(&job)){

	setCapabilities(env, self, job.hDC);

	switch (job.Flags & (PD_SELECTION | PD_PAGENUMS)) {

	case PD_ALLPAGES:
	    setRange(env, self, 0, numPages - 1);
	    break;

	case PD_PAGENUMS:
	    setRange(env, self, job.nFromPage - 1, job.nToPage - 1);
	    break;

	/* We disable the selection button in the Print dialog
	 * so we should never see that case. There are only the
	 * three cases aso anything else means something is very
	 * wrong.
	 */
	case PD_SELECTION:
	default:
	    assert(FALSE);
	    break;
	}

	doIt = JNI_TRUE;
    } else {
	job.hDC	= NULL;
    }

    setPrintDC(env, self, job.hDC);

    // We should delete the hDC in finalize in case endDoc is never called.
    // FIX.
    
    return doIt;
}


/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    startDoc
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPrinterJob_startDoc(JNIEnv *env, jobject self) {
    HDC printDC = getPrintDC(env, self);
    DWORD err = 0;

    /*
     * The print device context will be NULL if the
     * user never okayed a print dialog. This
     * will happen most often when the java application
     * decides not to present a print dialog to the user.
     * We create a device context for the default printer.
     */
    if (printDC == NULL) {
	printDC = getDefaultPrinterDC();
	if (printDC){
	    setPrintDC(env, self, printDC);
	    setCapabilities(env, self, printDC);
	}
    }

    if (printDC){
	DOCINFO docInfo;
	memset(&docInfo, 0, sizeof(DOCINFO));

	docInfo.cbSize = sizeof (DOCINFO);
	docInfo.lpszDocName = "Java Printing";
	docInfo.lpszOutput = (LPTSTR)NULL;
	docInfo.fwType = 0;

	err = ::StartDoc(printDC, &docInfo);
	if (err <= 0) {
	    err = GetLastError();
	} else {
	    err = 0;
	}
    }

    if (err) {
	throwPrinterException(env, err);
    }
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    endDoc
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPrinterJob_endDoc(JNIEnv *env, jobject self) {
    HDC printDC = getPrintDC(env, self);

    if (printDC != NULL){
	::EndDoc(printDC);
    }
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    deleteDC
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WPrinterJob_deleteDC(JNIEnv *env, jobject self) {
    HDC printDC = getPrintDC(env, self);

    if (printDC != NULL){
	setPrintDC(env, self, NULL);
	::DeleteDC(printDC);
    }
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    startPage
 * Signature: (Ljava/awt/print/PageFormat;Ljava/awt/print/Printable;I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_startPage
(JNIEnv *env, jobject self, jobject format, jobject painter, jint pageIndex) {

    HDC printDC = getPrintDC(env, self);

    if (printDC != NULL){
	::StartPage(printDC);
    }
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    endPage
 * Signature: (Ljava/awt/print/PageFormat;Ljava/awt/print/Printable;I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_endPage
(JNIEnv *env, jobject self, jobject format, jobject painter, jint pageIndex) {

    HDC printDC = getPrintDC(env, self);

    if (printDC != NULL){
	::EndPage(printDC);
    }
}


/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    printBand
 * Signature: ([IIIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_printBand
  (JNIEnv *env, jobject self, jintArray imageArray, jint x, jint y,
					jint width, jint height) {

    HDC printDC = getPrintDC(env, self);

    jint *image = (jint *)env->GetPrimitiveArrayCritical(imageArray, 0);

    long startY = 0;
    long endY = 0;
    jint *startImage = image + width * (height - 1);
    jint *endImage = NULL;
    long numLines = 0;

    do {
	startImage = findNonWhite(startImage, startY, width, height,
				  &numLines);

	if (startImage != NULL) {
	    startY += numLines;
	    endY = startY;
	    endImage = findWhite(startImage, endY, width, height, &numLines);
 
	    bitsToDevice(printDC, endImage, x, y + startY, width, numLines);

	    startImage = endImage - width;
	    startY = endY + numLines;
	}
    } while (startY < height && startImage != NULL);

    env->ReleasePrimitiveArrayCritical(imageArray, image, 0);
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    beginPath
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_beginPath
(JNIEnv *env , jobject self, jint printDC) {
 
    (void) ::BeginPath((HDC) printDC);

}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    endPath
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_endPath
(JNIEnv *env, jobject self, jint printDC) {

    (void) ::EndPath((HDC) printDC);

}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    fillPath
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_fillPath
(JNIEnv *env, jobject self, jint printDC) {

    (void) ::FillPath((HDC) printDC);

}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    closeFigure
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_closeFigure
(JNIEnv *env, jobject self, jint printDC) {

    (void) ::CloseFigure((HDC) printDC);

}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    lineTo
 * Signature: (IFF)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_lineTo
(JNIEnv *env, jobject self, jint printDC, jfloat x, jfloat y) {

    (void) ::LineTo((HDC) printDC, (long) x, (long) y);

}


/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    moveTo
 * Signature: (IFF)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_moveTo
(JNIEnv *env, jobject self, jint printDC, jfloat x, jfloat y) {

    (void) ::MoveToEx((HDC) printDC, (long) x, (long) y, NULL);

}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    polyBezierTo
 * Signature: (IFFFFFF)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_polyBezierTo
(JNIEnv *env, jobject self, jint printDC,
 jfloat control1x, jfloat control1y,
 jfloat control2x, jfloat control2y,
 jfloat endX, jfloat endY) {

    POINT points[3];

    points[0].x = (long) control1x;
    points[0].y = (long) control1y;
    points[1].x = (long) control2x;
    points[1].y = (long) control2y;
    points[2].x = (long) endX;
    points[2].y = (long) endY;

    (void) ::PolyBezierTo((HDC) printDC, points, 3);

}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    setPolyFillMode
 * Signature: (II)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_setPolyFillMode
(JNIEnv *env, jobject self, jint printDC, jint fillRule) {

    (void) ::SetPolyFillMode((HDC) printDC, fillRule);

}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    selectSolidBrush
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_selectSolidBrush
(JNIEnv *env, jobject self, jint printDC, jint red, jint green, jint blue) {

    HBRUSH colorBrush = ::CreateSolidBrush(RGB(red, green, blue));
    HBRUSH oldBrush = ::SelectObject((HDC) printDC, colorBrush);
    DeleteObject(oldBrush);
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    getPenX
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_sun_awt_windows_WPrinterJob_getPenX
(JNIEnv *env, jobject self, jint printDC) {

    POINT where;
    ::GetCurrentPositionEx((HDC) printDC, &where);

    return (jint) where.x;
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    getPenY
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_sun_awt_windows_WPrinterJob_getPenY
(JNIEnv *env, jobject self, jint printDC) {

    POINT where;
    ::GetCurrentPositionEx((HDC) printDC, &where);

    return (jint) where.y;
}

/*
 * Class:     sun_awt_windows_WPrinterJob
 * Method:    selectClipPath
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_sun_awt_windows_WPrinterJob_selectClipPath
(JNIEnv *env, jobject self, jint printDC) {

    ::SelectClipPath((HDC) printDC, RGN_COPY);
}

/**
 * Scans a DIB image looking for the first non-white line.
 * On entry, 'image' points at the bottom of the DIB, which
 * is where the first scan line is. *'lineP', on entry, is
 * the number of scan lines in the image while 'width' is
 * the number of 32 bit pixels on each line. If a non-white
 * line is found in the DIB, then a pointer to the first,
 * working from the bottom, non-white scan line is returned
 * and the number of remaining scan lines is returned in
 * *'lineP'.
 */
static jint *findNonWhite(jint *image, long x, long width, long height, long *numLinesP) {

    long white = kRGB32White;   
    long found = -1;
    long numLines = 0;
    jint *startLine = image;
    jint *inLine;

    assert(image != NULL);
    assert(0 <= x && x < height);
    assert(0 < width);
    assert(0 < height);
    assert(numLinesP != NULL);

    for (numLines = 0; x < height; numLines++, x++) {

	inLine = startLine;

	for (long col = 0; col < width; col++) {
	    white &= *inLine++;
	    if (white != kRGB32White) {
		found = x;
		break;
	    }
	}

	if(found != -1) {
	    break;
	}

	startLine -= width;
    }

    *numLinesP = numLines;

    return found == -1 ? NULL : startLine;
}

static jint *findWhite(jint *image, long x, long width, long height, long *numLinesP) {

    long numLines;
    jint *startLine = image;
    jint *inLine;
    jint *found = NULL;
    long white = kRGB32White;

    assert(image != NULL);
    assert(0 <= x);
    assert(0 < width);
    assert(0 < height);
    assert(numLinesP != NULL);


    ++x;
    for(numLines = 1; x < height; numLines++, x++) {

	startLine -= width;
	inLine = startLine;
	white = kRGB32White;

	for (long col = 0; col < width; col++) {
	    white &= *inLine++;
	    if (white != kRGB32White) {
		break;
	    }
	}

	if(white == kRGB32White) {
	   found = startLine + width;
	   break;
	}
    }

    *numLinesP = numLines;


    return found == NULL ? startLine : found;

}

static int bitsToDevice(HDC printDC, jint *image, long destX, long destY,
			long width, long height) {
    int result = 0;

    assert(printDC != NULL);
    assert(image != NULL);
    assert(destX >= 0);
    assert(destY >= 0);
    assert(width > 0);
    assert(height > 0);

    static int redMask = 0xff0000;
    static int greenMask = 0x00ff00;
    static int blueMask = 0x0000ff;
    struct {
	BITMAPINFOHEADER bmiHeader; 
	DWORD            bmiMasks[256]; 
    } bitMapHeader;

    memset(&bitMapHeader,0,sizeof(bitMapHeader));
    bitMapHeader.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bitMapHeader.bmiHeader.biWidth = width; 
    bitMapHeader.bmiHeader.biHeight = height; // does -height work ever?
    bitMapHeader.bmiHeader.biPlanes = 1; 
    bitMapHeader.bmiHeader.biBitCount = 32;
    bitMapHeader.bmiHeader.biCompression = BI_BITFIELDS; 
    bitMapHeader.bmiHeader.biSizeImage = 0;	// It's the default size. 
    bitMapHeader.bmiHeader.biXPelsPerMeter = 0;
    bitMapHeader.bmiHeader.biYPelsPerMeter = 0;
    bitMapHeader.bmiHeader.biClrUsed = 0;
    bitMapHeader.bmiHeader.biClrImportant = 0;
    bitMapHeader.bmiMasks[0] = redMask;
    bitMapHeader.bmiMasks[1] = greenMask;
    bitMapHeader.bmiMasks[2] = blueMask;

    if (printDC){
	result = ::SetDIBitsToDevice(printDC,
				destX,	// left of dest rect
				destY,	// top of dest rect	
				width,	// width of dest rect
				height,	// height of dest rect
				0,	// left of source rect
				0,	// top of source rect
				0,	// line number of 1st source scan line
				height,	// number of source scan lines
				image,	// points to the DIB
				(BITMAPINFO *)&bitMapHeader,
				DIB_RGB_COLORS);
    }

    return result;
}

/**
 * Called by the Page Setup dialog this routine makes sure the
 * print dialog becomes the front most window.
 */
static UINT CALLBACK pageDlgHook(HWND hDlg, UINT msg,
				 UINT wParam, LONG lParam) {

    if (msg == WM_INITDIALOG) {
	SetForegroundWindow(hDlg);
    }

    return (UINT) FALSE;
}


/**
 * Called by the Print dialog, this routine makes sure the
 * print dialog becoes the front most window.
 */
static UINT CALLBACK printDlgHook(HWND hDlg, UINT msg,
				  UINT wParam, LONG lParam)
{

    if (msg == WM_INITDIALOG) {
	SetForegroundWindow(hDlg);
    }

    return (UINT) FALSE;
}


/**
 *	Create and return a printer device context for the
 *	default printer. If there is no default printer then
 *	return NULL.
 */
static HDC getDefaultPrinterDC() {
    char printer[80];
    char *device = NULL;
    char *driver = NULL;
    char *output = NULL;
    HDC printDC = NULL;

    /* It's simpler to just use PrintDlg to get the default DC. FIX.
    */

    int len = ::GetProfileString("windows", "device", ",,,", printer,
							sizeof(printer));

    if (len > 0) device = strtok(printer, ",");
    if (device) driver = strtok(NULL, ", ");
    if (driver) output = strtok(NULL, ", ");

    if (output) printDC = ::CreateDC(driver, device, output, NULL);

    return printDC;
}

/**
 * Move the description of the page's size and orientation
 * from the PageFormat object 'page' into the structure,
 * 'setup' used by Windows to display the Page Setup dialog.
 */
static void pageFormatToSetup(JNIEnv *env, jobject page, PAGESETUPDLG *setup) {
    RectDouble paperSize;
    RectDouble margins;

    /* Move the orientation from PageFormat to Windows.
     */
    jboolean isPortrait = isPortraitOrientation(env, page);
    setOrientationInDevMode(setup->hDevMode, isPortrait);

    int units = (setup->Flags & PSD_INTHOUSANDTHSOFINCHES)
						? MM_HIENGLISH
						: MM_HIMETRIC;
    jobject paper = getPaper(env, page);
    getPaperValues(env, paper, &paperSize, &margins);
    setup->ptPaperSize.x = convertFromPoints(paperSize.width, units);
    setup->ptPaperSize.y = convertFromPoints(paperSize.height, units);

    // Set page size here.
}


/**
 * Return an array of POINTS describing the paper sizes supported
 * by the driver identified by 'deviceName' and 'portName'.
 * If there is an error, then NULL is returned.
 */
static POINT *getPaperSizeList(const char *deviceName, const char *portName) {
    DWORD numPaperSizes;
    POINT *paperSizes = NULL;

    numPaperSizes = DeviceCapabilities(deviceName, portName,
				       DC_PAPERSIZE, NULL, NULL);
    
    if (numPaperSizes > 0) {
	paperSizes = (POINT *) malloc(sizeof(*paperSizes) * numPaperSizes);
	if (paperSizes) {
	    DWORD result = DeviceCapabilities(deviceName, portName,
					      DC_PAPERSIZE, (char *) paperSizes,
					      NULL);
	    if (result != -1) {
		free((char *) paperSizes);
		paperSizes = NULL;
	    }
	}
    }

    return paperSizes;
}

/**
 * Set the orientation of the paper described by the printer
 * handle to a device mode structure 'hDevMode'.
 */
static void setOrientationInDevMode(HGLOBAL hDevMode, jboolean isPortrait) {

    if (hDevMode != NULL) {
	LPDEVMODE devMode = (LPDEVMODE) GlobalLock(hDevMode);
	if (devMode != NULL) {
	    devMode->dmOrientation = isPortrait
				    ? DMORIENT_PORTRAIT
				    : DMORIENT_LANDSCAPE;
	    devMode->dmFields |= DM_ORIENTATION;
	    GlobalUnlock(hDevMode);
	}
    }
}

/**
 * Return the paper size and margins for the page
 * adjusted to take into account the portrait or
 * landscape orientation of the page. On entry,
 * 'setup' is a filled in structure as returned
 * by PageSetupDlg(). 'paperSize', 'margins',
 * and 'isPortrait' all point to caller allocated
 * space while will be filled in by this routine
 * with the size, in unknown Windows units, of
 * the paper, of the margins, and an indicator
 * whether the page is in portrait or landscape
 * orientation, respectively.
 */
static void retrievePaperInfo(const PAGESETUPDLG *setup, POINT *paperSize,
					RECT *margins, jboolean *isPortrait) {
    int orientationKnown = FALSE;

    *paperSize = setup->ptPaperSize;
    *isPortrait = TRUE;

    /* Usually the setup dialog will tell us the
     * orientation of the page, but it may not.
     */
    if (setup->hDevMode != NULL) {
	DEVMODE *devMode = (DEVMODE *)GlobalLock(setup->hDevMode);
	if(devMode->dmFields & DM_ORIENTATION){
	    orientationKnown = TRUE;
	    *isPortrait = devMode->dmOrientation == DMORIENT_PORTRAIT;
	}
	GlobalUnlock(setup->hDevMode);
    }

    /* The driver didn't tell us the paper orientation
     * so we declare it portrait if the paper
     * is longer than it is wide. Square paper is
     * declared to be portait.
     */
    if (orientationKnown == FALSE){
	*isPortrait = paperSize->x < paperSize->y;
    }

    /* The Paper class expresses the page size in
     * portait mode while Windows returns the paper
     * size adjusted for the orientation. If the
     * orientation is landscape then we want to
     * flip the width and height to get a portait
     * description of the page.
     */
    if (*isPortrait == FALSE){
	long hold = paperSize->x;
	paperSize->x = paperSize->y;
	paperSize->y = hold;

	margins->left = setup->rtMargin.bottom;
	margins->right = setup->rtMargin.top;
	margins->top = setup->rtMargin.left;
	margins->bottom = setup->rtMargin.right;

    } else {
	*margins = setup->rtMargin;
    }

}

/**
 * Set the range of pages to be printed. Both 'from', the first
 * page and 'to', the last page, are zero based indices.
 */
static void setRange(JNIEnv *env, jobject self, jint from, jint to) {
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());

    jclass wPrintJob2DClass = env->GetObjectClass(self);
    jmethodID setRangeID = env->GetMethodID(wPrintJob2DClass, SETRANGE_STR,
						SETRANGE_SIG);
    env->CallVoidMethod(self, setRangeID, from, to);
}

/**
 * Return the number of pages in a Pageable document.
 */
static jint getNumPages(JNIEnv *env, jobject doc)
{
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());

    jint numPages;
    jclass pageableClass = env->GetObjectClass(doc);
    jmethodID getNumPagesID = env->GetMethodID(pageableClass, GETNUMPAGES_STR,
					       GETNUMPAGES_SIG);
    numPages = env->CallIntMethod(doc, getNumPagesID);
    if (numPages < 0) numPages = MAX_UNKNOWN_PAGES;

    return numPages;
}

/**
 * Return the number of copies to be printed for a printerJob.
 */
static jint getCopies(JNIEnv *env, jobject printerJob)
{
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());


    jclass printerJobClass = env->GetObjectClass(printerJob);
    jmethodID getCopiesID = env->GetMethodID(printerJobClass, GETCOPIES_STR,
					     GETCOPIES_SIG);
    jint copies = env->CallIntMethod(printerJob, getCopiesID);

    return copies;
}

/**
 * Return a copy of the Paper object attached to the
 * PageFormat object 'page.'
 */
static jobject getPaper(JNIEnv *env, jobject page) {
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());


    jclass pageClass = env->GetObjectClass(page);
    jmethodID getPaperID = env->GetMethodID(pageClass, GETPAPER_STR,
							GETPAPER_SIG);
 
    return env->CallObjectMethod(page, getPaperID);
}

/**
 * Set the Paper object for a PageFormat instance.
 * 'paper' is the new Paper object that must be
 * set into 'page'.
 */
static void setPaper(JNIEnv *env, jobject page, jobject paper) {
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());

    jclass pageClass = env->GetObjectClass(page);
    jmethodID setPaperID = env->GetMethodID(pageClass, SETPAPER_STR,
							SETPAPER_SIG);
    env->CallVoidMethod(page, setPaperID, paper);
}

/**
 * Return true if the PageFormat object 'page' describes a
 * page in the portrait orientation.
 */
static jboolean isPortraitOrientation(JNIEnv *env, jobject page) {
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());

    jclass pageClass = env->GetObjectClass(page);
    jmethodID getOrientID = env->GetMethodID(pageClass, GETORIENT_STR,
							GETORIENT_SIG);
   
    jfieldID fieldID = env->GetStaticFieldID(pageClass, PORTRAIT_STR,
							PORTRAIT_SIG);

    jint portrait = env->GetStaticIntField(pageClass, fieldID);

    jint orientation = env->CallIntMethod(page, getOrientID);	
    
    /* We'll return true if the orientation is PORTRAIT. If
     * another PORTRAIT orientation is added, such as
     * PORTRAIT_LANDSCAPE, then this test will need to be
     * extended.
     */
    return orientation == portrait;

}

static void setOrientation(JNIEnv *env, jobject page, jboolean isPortrait) {
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());

    jclass pageClass = env->GetObjectClass(page);
    jmethodID setOrientID = env->GetMethodID(pageClass, SETORIENT_STR,
							SETORIENT_SIG);

    /* We'll either set portrait or landscape, but we need to
     * get the orientation constants out of the PageFormat
     * class. We assume that the two constants have the same
     * signature - PORTRAIT_SIG.
     */
    const char *fieldName = isPortrait ? PORTRAIT_STR : LANDSCAPE_STR;
    jfieldID fieldID = env->GetStaticFieldID(pageClass, fieldName,
							PORTRAIT_SIG);
    jint orientation = env->GetStaticIntField(pageClass, fieldID);

    env->CallVoidMethod(page, setOrientID, orientation);	
}

/**
 * Pull the paper size and margins out of the paper object and
 * return them in points.
 */
static void getPaperValues(JNIEnv *env, jobject paper, RectDouble *paperSize,
			  RectDouble *margins) {
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());

    jmethodID getID;
    
    paperSize->x = 0;
    paperSize->y = 0;

    jclass paperClass = env->GetObjectClass(paper);

    getID = env->GetMethodID(paperClass, GETWIDTH_STR, GETWIDTH_SIG);
    paperSize->width = env->CallDoubleMethod(paper, getID);

    getID = env->GetMethodID(paperClass, GETHEIGHT_STR, GETHEIGHT_SIG);
    paperSize->height = env->CallDoubleMethod(paper, getID);

}

/**
 * Given a RECT specifying the margins
 * for the page and an indication of whether
 * the units are 1000ths of an inch (MM_HIENGLISH)
 * or 100ths of a millimeter (MM_HIMETRIC),
 * convert the margins to 72nds of an inch
 * and set them into the PageFormat insance provided.
 */
static void setPaperValues(JNIEnv *env, jobject paper, const POINT *paperSize,
					 const RECT *margins, int units) {
    // Because this function may call client Java code,
    // we can't run it on the toolkit thread.
    ASSERT(AwtToolkit::MainThread() != ::GetCurrentThreadId());

    jclass paperClass = env->GetObjectClass(paper);
    jmethodID setSizeID = env->GetMethodID(paperClass,
					SETSIZE_STR, SETSIZE_SIG);
    jmethodID setImageableID = env->GetMethodID(paperClass,
					SETIMAGEABLE_STR, SETIMAGEABLE_SIG);
								    
    /* Set the physical size of the paper.
     */
    jdouble paperWidth = convertToPoints(paperSize->x, units);
    jdouble paperHeight = convertToPoints(paperSize->y, units);
    env->CallVoidMethod(paper, setSizeID, paperWidth, paperHeight);

    /* Set the margins of the paper. In Windows' margin RECT,
     * the right and bottom parts of the structure are not
     * really the right and bottom of the imageable rectangle,
     * but rather the right and bottom margins.
     */
    jdouble x = convertToPoints(margins->left, units);
    jdouble y = convertToPoints(margins->top, units);
    long intWidth = paperSize->x - margins->left - margins->right;
    long intHeight = paperSize->y - margins->top - margins->bottom;
    jdouble width = convertToPoints(intWidth, units);
    jdouble height = convertToPoints(intHeight, units);
    env->CallVoidMethod(paper, setImageableID, x, y, width, height);

}

/**
 * Convert 'value' a measurement in 1/72's of an inch to
 * the units specified by 'units' - either MM_HIENGLISH
 * MM_HIMETRIC, or MM_LOMETRIC. The converted value is returned as
 * a long.
 */
static long convertFromPoints(double value, int units) {
    double conversion = 0;

    switch (units){
     case MM_HIENGLISH:
	conversion = POINTS_TO_HIENGLISH;
	break;

     case MM_HIMETRIC:
	conversion = POINTS_TO_HIMETRIC;
	break;

     case MM_LOMETRIC:
	conversion = POINTS_TO_LOMETRIC;
	break;

     default:
	assert(FALSE);	// Unsupported unit.
    }

    return (long) (value * conversion);
}

/**
 * Convert a measurement, 'value', from the units
 * specified by 'units', either MM_HIENGLISH or
 * MM_HIMETRIC to 1/72's of an inch and returned
 * as a double.
 */
static double convertToPoints(long value, int units) {
    double convertedValue = (double)value;

    switch (units){
    case MM_HIENGLISH:
	convertedValue *= HIENGLISH_TO_POINTS;
	break;

    case MM_HIMETRIC:
	convertedValue *= HIMETRIC_TO_POINTS;
	break;

    default:
	assert(FALSE);	// Unsupported unit.
    }

    //Need to round off to the precision of the initial value. FIX.

    return convertedValue;
}

/**
 *	Ask the printer device context, 'printDC' about
 *	its capabilities and set these into the WPrintJob2D
 *	object 'self'.
 */
static void setCapabilities(JNIEnv *env, jobject self, HDC printDC) {
    jint xRes = GetDeviceCaps(printDC, LOGPIXELSX);
    setIntField(env, self, XRES_STR, xRes);

    jint yRes = GetDeviceCaps(printDC, LOGPIXELSY);
    setIntField(env, self, YRES_STR, yRes);
}



static HDC getPrintDC(JNIEnv *env, jobject self) {
    return (HDC)getIntField(env, self, PRINTDC_STR);	
}

static void setPrintDC(JNIEnv *env, jobject self, HDC printDC) {
    setIntField(env, self, PRINTDC_STR, (jint)printDC);
}

/**
 *	Return the java int value of the field 'fieldName' in the
 *	java instance 'self'.
 */
static jint getIntField(JNIEnv *env, jobject self, const char *fieldName) {
    jfieldID fieldId = getIdOfIntField(env, self, fieldName);
    return env->GetIntField(self, fieldId);	
}

/**
 *	Set the int field named 'fieldName' of the java instance
 *	'self' to the value 'value'.
 */
static void setIntField(JNIEnv *env, jobject self, const char *fieldName,
								jint value) {
    jfieldID fieldId = getIdOfIntField(env, self, fieldName);
    env->SetIntField(self, fieldId, value);
}

/**
 *	Return the field id of the java instance 'self' of the
 *	java int field named 'fieldName'.
 */
static jfieldID getIdOfIntField(JNIEnv *env, jobject self,
						const char *fieldName) {
    jclass myClass = env->GetObjectClass(self);
    jfieldID fieldId = env->GetFieldID(myClass, fieldName, kJavaIntStr);
    assert(fieldId != 0);

    return fieldId;

}

/**
 *  Throw a PrinterException with a string describing
 *  the Window's system error 'err'.
 */
static void throwPrinterException(JNIEnv *env, DWORD err) {
    char errStr[256];
    jclass printerException = env->FindClass(PRINTEREXCEPTION_STR);

    errStr[0] = '\0';
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		  NULL,
		  err,
		  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		  (LPTSTR) errStr,
		  sizeof(errStr),
		  NULL );


    env->ThrowNew(printerException, errStr);
}

} /* extern "C" */
