/*
 * @(#)awt_ScrollPane.cpp	1.33 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_ScrollPane.h"

#include "awt_Container.h"
#include "awt_Insets.h"
#include "awt_Panel.h"
#include "awt_Scrollbar.h"   // static #defines
#include "awt_Toolkit.h"

#include <java_awt_ScrollPane.h>
#include <java_awt_Scrollbar.h>
#include <java_awt_ScrollPaneAdjustable.h>

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtScrollPane fields
 */

jfieldID AwtScrollPane::scrollbarDisplayPolicyID;
jfieldID AwtScrollPane::hAdjustableID;
jfieldID AwtScrollPane::vAdjustableID;
jfieldID AwtScrollPane::unitIncrementID;
jfieldID AwtScrollPane::blockIncrementID;


/************************************************************************
 * AwtScrollPane methods
 */

AwtScrollPane::AwtScrollPane() {
}

const char* AwtScrollPane::GetClassName() {
    return "SunAwtScrollPane";
}

/* Create a new AwtScrollPane object and window.   */
AwtScrollPane* AwtScrollPane::Create(jobject self, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    int pData;
    AwtComponent* awtParent;
    jobject target = 0;
    AwtScrollPane* c;
    JNI_CHECK_PEER_GOTO(parent, done);

    awtParent = (AwtComponent*)pData;
    JNI_CHECK_NULL_GOTO(awtParent, "null awtParent", done);

    target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    c = new AwtScrollPane();
    JNI_CHECK_NULL_GOTO(c, "AwtScrollPane() failed", done);

    {
        DWORD style = WS_CHILD | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
	if (!IS_WIN4X) {
	    /* 
	     * It's been decided by the UI folks that 3.X ScrollPanes
	     * should have borders...  
	     */
	    style |= WS_BORDER;
	}
	jint scrollbarDisplayPolicy = 
	    env->GetIntField(target, scrollbarDisplayPolicyID);

	if (scrollbarDisplayPolicy == 
	    Java_java_awt_ScrollPane_SCROLLBARS_ALWAYS) {
	    style |= WS_HSCROLL | WS_VSCROLL;
	}
	DWORD exStyle = IS_WIN4X ? WS_EX_CLIENTEDGE : 0;

	if (GetRTL()) {
	    exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
	    if (GetRTLReadingOrder())
	        exStyle |= WS_EX_RTLREADING;
	}

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);
	c->CreateHWnd(env, "", style, exStyle,
		      x, y, width, height,
		      awtParent->GetHWnd(),
		      (HMENU)awtParent->CreateControlID(),
		      ::GetSysColor(COLOR_WINDOWTEXT),
		      ::GetSysColor(COLOR_WINDOW),
		      self);
    }
 done:
    env->DeleteLocalRef(target);
    return c;
}

void AwtScrollPane::SetPageIncrement(int value, int orient)
{
    SCROLLINFO si;
    si.cbSize = sizeof(SCROLLINFO);
    si.fMask = SIF_PAGE;
    si.nPage = value;
    ::SetScrollInfo(GetHWnd(), orient, &si, TRUE);
}

void AwtScrollPane::SetInsets(JNIEnv *env)
{
    RECT outside;
    RECT inside;
    ::GetWindowRect(GetHWnd(), &outside);
    ::GetClientRect(GetHWnd(), &inside);
    ::MapWindowPoints(GetHWnd(), 0, (LPPOINT)&inside, 2);
    
    if (env->EnsureLocalCapacity(1) < 0) {
	return;
    }
    jobject insets =
      (env)->GetObjectField(GetPeer(env), AwtPanel::insets_ID);

    ASSERT(!env->ExceptionOccurred());

    if (insets == NULL) {
        goto done;
    }
    (env)->SetIntField(insets, AwtInsets.topID, inside.top - outside.top);
    (env)->SetIntField(insets, AwtInsets.leftID, inside.left - outside.left);
    (env)->SetIntField(insets, AwtInsets.bottomID, outside.bottom - inside.bottom);
    (env)->SetIntField(insets, AwtInsets.rightID, outside.right - inside.right);

done:
    env->DeleteLocalRef(insets);
}

void AwtScrollPane::ScrollTo(int x, int y)
{
    int dx = ::GetScrollPos(GetHWnd(), SB_HORZ) - x;
    int dy = ::GetScrollPos(GetHWnd(), SB_VERT) - y;
    ::SetScrollPos(GetHWnd(), SB_HORZ, x, TRUE);
    ::SetScrollPos(GetHWnd(), SB_VERT, y, TRUE);
    VERIFY(::ScrollWindow(GetHWnd(), dx, dy, NULL, NULL));
    VERIFY(::UpdateWindow(GetHWnd()));
}

void AwtScrollPane::SetScrollInfo(int orient, int max, int page, 
                                  BOOL disableNoScroll)
{
   SCROLLINFO si;
    si.cbSize = sizeof(SCROLLINFO);
    si.nMin = 0;
    si.nMax = max;
    si.fMask = SIF_RANGE;
    if (disableNoScroll) {
        si.fMask |= SIF_DISABLENOSCROLL;
    }
    if (page > 0) {
        si.fMask |= SIF_PAGE;
        si.nPage = page;
    }
    ::SetScrollInfo(GetHWnd(), orient, &si, TRUE);
}

void AwtScrollPane::RecalcSizes(int parentWidth, int parentHeight,
				int childWidth, int childHeight)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0) {
	return;
    }

    /* Determine border width without scrollbars. */
    int horzBorder;
    int vertBorder;
    if (IS_WIN4X) {
	horzBorder = ::GetSystemMetrics(SM_CXEDGE);
	vertBorder = ::GetSystemMetrics(SM_CYEDGE);
    } else {
        horzBorder = ::GetSystemMetrics(SM_CXBORDER);
        vertBorder = ::GetSystemMetrics(SM_CYBORDER);
    }
    
    parentWidth -= (horzBorder * 2);
    parentHeight -= (vertBorder * 2);

    /* Enable each scrollbar as needed. */
    jobject target = AwtObject::GetTarget(env);
    jint policy = env->GetIntField(target, 
				   AwtScrollPane::scrollbarDisplayPolicyID);

    BOOL needsHorz=(policy == Java_java_awt_ScrollPane_SCROLLBARS_ALWAYS ||
		    (policy ==Java_java_awt_ScrollPane_SCROLLBARS_AS_NEEDED && 
		     childWidth > parentWidth));
    if (needsHorz) {
	parentHeight -= ::GetSystemMetrics(SM_CYHSCROLL);
    }
    BOOL needsVert=(policy == Java_java_awt_ScrollPane_SCROLLBARS_ALWAYS ||
		    (policy ==Java_java_awt_ScrollPane_SCROLLBARS_AS_NEEDED && 
		     childHeight > parentHeight));
    if (needsVert) {
	parentWidth -= ::GetSystemMetrics(SM_CXVSCROLL);
    }
    /*
     * Since the vertical scrollbar may have reduced the parent width
     * enough to now require a horizontal scrollbar, we need to
     * recalculate the horizontal metrics and scrollbar boolean.
     */
    if (!needsHorz) {
	needsHorz = (policy == java_awt_ScrollPane_SCROLLBARS_ALWAYS ||
		     (policy == java_awt_ScrollPane_SCROLLBARS_AS_NEEDED && 
		      childWidth > parentWidth));
	if (needsHorz) {
	    parentHeight -= ::GetSystemMetrics(SM_CYHSCROLL);
	}
    }

    /* Now set ranges -- setting the min and max the same disables them. */
    if (needsHorz) {
	jobject hAdj = 
	    env->GetObjectField(target, AwtScrollPane::hAdjustableID);
	env->SetIntField(hAdj, AwtScrollPane::blockIncrementID, parentWidth);
        SetScrollInfo(SB_HORZ, childWidth - 1, parentWidth, 
                      (policy == Java_java_awt_ScrollPane_SCROLLBARS_ALWAYS));
	env->DeleteLocalRef(hAdj);
    } else {
        SetScrollInfo(SB_HORZ, 0, 0, 
                      (policy == java_awt_ScrollPane_SCROLLBARS_ALWAYS));
    }

    if (needsVert) {
	jobject vAdj = 
	    env->GetObjectField(target, AwtScrollPane::vAdjustableID);
	env->SetIntField(vAdj, AwtScrollPane::blockIncrementID, parentHeight);
        SetScrollInfo(SB_VERT, childHeight - 1, parentHeight, 
                      (policy == Java_java_awt_ScrollPane_SCROLLBARS_ALWAYS));
	env->DeleteLocalRef(vAdj);
    } else {
        SetScrollInfo(SB_VERT, 0, 0, 
                      (policy == Java_java_awt_ScrollPane_SCROLLBARS_ALWAYS));
    }

    env->DeleteLocalRef(target);
}

void AwtScrollPane::Reshape(int x, int y, int w, int h)
{
    AwtComponent::Reshape(x, y, w, h);
}

void AwtScrollPane::Show(JNIEnv *env)
{
    SetInsets(env);
    AwtComponent::Show();
}

int AwtScrollPane::NewScrollPosition(UINT scrollCode, UINT orient, UINT pos,
                                     UINT unitIncrement, UINT blockIncrement)
{
    int newPos, minPos, maxPos;
    VERIFY(::GetScrollRange(GetHWnd(), orient, &minPos, &maxPos));
    int curPos = ::GetScrollPos(GetHWnd(), orient);

    /* 
     * Restrict range so window isn't displayed beyond the child window's 
     * maximum size. 
     */
    RECT inside;
    ::GetClientRect(GetHWnd(), &inside);
    switch (orient) {
      case SB_HORZ:
          maxPos -= inside.right;  /* GetClientRect always returns left as 0 */
          break;
      case SB_VERT:
          maxPos -= inside.bottom; /* ditto for top */
          break;
#ifdef DEBUG
      default:
          ASSERT(FALSE);
#endif
    }

    switch (scrollCode) {
      case SB_LINEUP:
          newPos = curPos - unitIncrement;
          return (newPos >= minPos) ? newPos : minPos;
      case SB_LINEDOWN:
          newPos = curPos + unitIncrement;
          return (newPos <= maxPos) ? newPos : maxPos;
      case SB_PAGEUP:
          newPos = curPos - blockIncrement;
          return (newPos >= minPos) ? newPos : minPos;
      case SB_PAGEDOWN:
          newPos = curPos + blockIncrement;
          return (newPos <= maxPos) ? newPos : maxPos;
      case SB_TOP:
          return minPos;
      case SB_BOTTOM:
          return maxPos;
      case SB_THUMBPOSITION:
      case SB_THUMBTRACK:
          return pos;
      default:
          return curPos;
    }
}

MsgRouting AwtScrollPane::WmHScroll(UINT scrollCode, UINT pos, 
                                    HWND hScrollPane) 
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0) {
	return mrConsume;
    }

    jobject target = AwtObject::GetTarget(env);

    jobject hAdj = env->GetObjectField(target, AwtScrollPane::hAdjustableID);
    jint unitIncrement = env->GetIntField(hAdj,AwtScrollPane::unitIncrementID);
    jint blockIncrement=env->GetIntField(hAdj,AwtScrollPane::blockIncrementID);

    int newPos = NewScrollPosition(scrollCode, SB_HORZ, pos,
                                   unitIncrement, blockIncrement);
    int curPos = ::GetScrollPos(GetHWnd(), SB_HORZ);
    if (newPos == curPos) {
        env->DeleteLocalRef(target);
	env->DeleteLocalRef(hAdj);
        return mrDoDefault;
    }

    ::SetScrollPos(GetHWnd(), SB_HORZ, newPos, TRUE);
    VERIFY(::ScrollWindow(GetHWnd(), curPos - newPos, 0, NULL, NULL));
    VERIFY(::UpdateWindow(GetHWnd()));

    /* Notify peer object. */
    JNU_CallMethodByName(env, NULL, GetPeer(env), "scrolledHorizontal",
			 "(I)V", newPos);
    ASSERT(!env->ExceptionOccurred());
    
    env->DeleteLocalRef(target);
    env->DeleteLocalRef(hAdj);
    return mrConsume;
}

MsgRouting AwtScrollPane::WmVScroll(UINT scrollCode, UINT pos, 
                                    HWND hScrollPane) 
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0) {
	return mrConsume;
    }

    jobject target = AwtObject::GetTarget(env);
    jobject vAdj = env->GetObjectField(target, AwtScrollPane::vAdjustableID);
    jint unitIncrement = env->GetIntField(vAdj,AwtScrollPane::unitIncrementID);
    jint blockIncrement=env->GetIntField(vAdj,AwtScrollPane::blockIncrementID);

    int newPos = NewScrollPosition(scrollCode, SB_VERT, pos,
                                   unitIncrement, blockIncrement);
    int curPos = ::GetScrollPos(GetHWnd(), SB_VERT);
    if (newPos == curPos) {
        env->DeleteLocalRef(target);
	env->DeleteLocalRef(vAdj);
        return mrDoDefault;
    }

    ::SetScrollPos(GetHWnd(), SB_VERT, newPos, TRUE);
    VERIFY(::ScrollWindow(GetHWnd(), 0, curPos - newPos, NULL, NULL));
    VERIFY(::UpdateWindow(GetHWnd()));

    /* Notify peer object. */
    JNU_CallMethodByName(env, NULL, GetPeer(env), "scrolledVertical",
			 "(I)V", newPos);
    ASSERT(!env->ExceptionOccurred());

    env->DeleteLocalRef(target);
    env->DeleteLocalRef(vAdj);
    return mrConsume;
}

#ifdef DEBUG
void AwtScrollPane::VerifyState()
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(3) < 0) {
	return;
    }
    
    if (AwtToolkit::GetInstance().VerifyComponents() == FALSE) {
        return;
    }

    if (m_callbacksEnabled == FALSE) {
        /* Component is not fully setup yet. */
        return;
    }

    AwtComponent::VerifyState();

    jobject target = AwtObject::GetTarget(env);
    jobject child = JNU_CallMethodByName(env, NULL, GetPeer(env), 
					 "getScrollSchild",
					 "()Ljava/awt/Component;").l;

    ASSERT(!env->ExceptionOccurred());

    if (child != NULL) {
	jobject childPeer = 
	    (env)->GetObjectField(child, AwtComponent::peerID);
	int pData;
	JNI_CHECK_PEER_RETURN(childPeer);
	AwtComponent* awtChild = (AwtComponent *)pData;

	/* Verify child window is positioned correctly. */
	RECT rect, childRect;
	::GetClientRect(GetHWnd(), &rect);
	::MapWindowPoints(GetHWnd(), 0, (LPPOINT)&rect, 2);
	::GetWindowRect(awtChild->GetHWnd(), &childRect);
	ASSERT(childRect.left <= rect.left && childRect.top <= rect.top);

	env->DeleteLocalRef(childPeer);
    }
    env->DeleteLocalRef(target);
    env->DeleteLocalRef(child);
}
#endif


/************************************************************************
 * ScrollPane native methods
 */

extern "C" {

/*
 * Class:     java_awt_ScrollPane
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_java_awt_ScrollPane_initIDs(JNIEnv *env, jclass cls)
{
    AwtScrollPane::scrollbarDisplayPolicyID = 
	env->GetFieldID(cls, "scrollbarDisplayPolicy", "I");
    AwtScrollPane::hAdjustableID =
	env->GetFieldID(cls, "hAdjustable", "Ljava/awt/ScrollPaneAdjustable;");
    AwtScrollPane::vAdjustableID = 
	env->GetFieldID(cls, "vAdjustable", "Ljava/awt/ScrollPaneAdjustable;");

    ASSERT(AwtScrollPane::scrollbarDisplayPolicyID != NULL);
    ASSERT(AwtScrollPane::hAdjustableID != NULL);
    ASSERT(AwtScrollPane::vAdjustableID != NULL); 
}

} /* extern "C" */


/************************************************************************
 * ScrollPaneAdjustable native methods
 */

extern "C" {

/*
 * Class:     java_awt_ScrollPaneAdjustable
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_java_awt_ScrollPaneAdjustable_initIDs(JNIEnv *env, jclass cls)
{
    AwtScrollPane::unitIncrementID = env->GetFieldID(cls,"unitIncrement", "I");
    AwtScrollPane::blockIncrementID = 
	env->GetFieldID(cls,"blockIncrement", "I");

    ASSERT(AwtScrollPane::unitIncrementID != NULL);
    ASSERT(AwtScrollPane::blockIncrementID != NULL);
}

} /* extern "C" */


/************************************************************************
 * ScrollPanePeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WScrollPanePeer_create(JNIEnv *env, jobject self, 
					    jobject parent)
{
    AWT_TRACE(("%x: WScrollPanePeer.create(%x)\n", self, parent));

    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent, 
				(AwtToolkit::ComponentFactory)
				AwtScrollPane::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
    ((AwtScrollPane*)pData)->VerifyState();
}

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    getOffset
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL 
Java_sun_awt_windows_WScrollPanePeer_getOffset(JNIEnv *env, jobject self,
					       jint orient)
{
    AWT_TRACE(("%x: WScrollPanePeer.getOffset(%d)\n", self, orient));

    int pData;
    JNI_CHECK_PEER_RETURN_NULL(self);
    AwtScrollPane* pane = (AwtScrollPane *)pData;
    pane->VerifyState();
    int nBar = (orient == Java_java_awt_Scrollbar_HORIZONTAL)?SB_HORZ:SB_VERT;
    return ::GetScrollPos(pane->GetHWnd(), nBar);
}

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    setInsets
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WScrollPanePeer_setInsets(JNIEnv *env, jobject self)
{
    AWT_TRACE(("%x: WScrollPanePeer.setInsets()\n", self));

    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtScrollPane* pane = (AwtScrollPane *)pData;
    pane->SetInsets(env);
    pane->VerifyState();
}

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    setScrollPosition
 * Signature: (II)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WScrollPanePeer_setScrollPosition(JNIEnv *env, 
						       jobject self,
						       jint x, jint y)
{
    AWT_TRACE(("%x: WScrollPanePeer.setScrollPosition(%d, %d)\n", self, x, y));

    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtScrollPane* pane = (AwtScrollPane *)pData;
    pane->ScrollTo(x, y);
    pane->VerifyState();
}

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    _getHScrollbarHeight
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WScrollPanePeer__1getHScrollbarHeight(JNIEnv *env,
							   jobject self)
{
    AWT_TRACE(("%x: WScrollPanePeer._getHScrollbarHeight()\n", self));
    return ::GetSystemMetrics(SM_CYHSCROLL);
}

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    _getVScrollbarWidth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WScrollPanePeer__1getVScrollbarWidth(JNIEnv *env, 
							  jobject self)
{
    AWT_TRACE(("%x: WScrollPanePeer._getVScrollbarHeight()\n", self));
    return ::GetSystemMetrics(SM_CXVSCROLL);
}

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    setSpans
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WScrollPanePeer_setSpans(JNIEnv *env, jobject self,
					      jint parentWidth,
					      jint parentHeight, 
					      jint childWidth,
					      jint childHeight)
{
    AWT_TRACE(("%x: WScrollPanePeer.setSpans(%d, %d, %d, %d)\n", self, 
               parentWidth, parentHeight, childWidth, childHeight));
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtScrollPane* pane = (AwtScrollPane *)pData;
    pane->RecalcSizes(parentWidth, parentHeight, childWidth, childHeight);
    pane->VerifyState();
}

/*
 * Class:     sun_awt_windows_WScrollPanePeer
 * Method:    getScrollChild
 * Signature: ()Ljava/awt/Component;
 */
JNIEXPORT jobject JNICALL
Java_sun_awt_windows_WScrollPanePeer_getScrollChild(JNIEnv *env, jobject self)
{
    AWT_TRACE(("%x: WScrollPanePeer.getScrollChild()\n", self));

    int pData;
    JNI_CHECK_PEER_RETURN_NULL(self);
    jobject target = env->GetObjectField(self, AwtObject::targetID);
    jint ncomponents = 
        (env)->GetIntField(target, AwtContainer::ncomponentsID);

    if (ncomponents == 0) {
        JNU_ThrowNullPointerException(env, "child is null");
        return NULL;
    }
    jobjectArray components = 
	(jobjectArray)(env)->GetObjectField(target, AwtContainer::componentID);
    return env->GetObjectArrayElement(components, 0);
}

} /* extern "C" */
