/*
 * @(#)awt_Scrollbar.cpp	1.28 98/09/30
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_Scrollbar.h"

#include "awt_Canvas.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtScrollbar fields
 */

jfieldID AwtScrollbar::lineIncrementID;
jfieldID AwtScrollbar::pageIncrementID;
jfieldID AwtScrollbar::orientationID;


/************************************************************************
 * AwtScrollbar methods
 */

AwtScrollbar::AwtScrollbar() {
    m_orientation = SB_HORZ;
    m_lineIncr = 0;
    m_pageIncr = 0;
}

const char* AwtScrollbar::GetClassName() {
    return "SCROLLBAR";  /* System provided scrollbar class */
}

/* Create a new AwtScrollbar object and window.   */
AwtScrollbar* AwtScrollbar::Create(jobject peer, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    int pData;
    AwtCanvas* awtParent;
    jobject target = 0;
    AwtScrollbar* c;
    JNI_CHECK_PEER_GOTO(parent, done);

    awtParent = (AwtCanvas*)pData;
    JNI_CHECK_NULL_GOTO(awtParent, "null awtParent", done);

    target = env->GetObjectField(peer, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    c = new AwtScrollbar();
    JNI_CHECK_NULL_GOTO(c, "AwtScrollbar() failed", done);

    {
        jint orientation =
	    env->GetIntField(target, AwtScrollbar::orientationID);
	c->m_orientation = (orientation == java_awt_Scrollbar_VERTICAL) ?
	    SB_VERT : SB_HORZ;
	c->m_lineIncr =
	    env->GetIntField(target, AwtScrollbar::lineIncrementID);
	c->m_pageIncr =
	    env->GetIntField(target, AwtScrollbar::pageIncrementID);

	DWORD style = WS_CHILD | WS_CLIPSIBLINGS |
	    c->m_orientation;/* Note: SB_ and SBS_ are the same here */

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	c->CreateHWnd(env, "", style, 0,
		      x, y, width, height,
		      awtParent->GetHWnd(),
		      (HMENU)awtParent->CreateControlID(),
		      ::GetSysColor(COLOR_SCROLLBAR),
		      ::GetSysColor(COLOR_SCROLLBAR),
		      peer);
	c->m_backgroundColorSet = TRUE;
	  /* suppress inheriting parent's color. */
    }
 done:
    env->DeleteLocalRef(target);
    return c;
}

LRESULT AwtScrollbar::WindowProc(UINT message, WPARAM wParam, LPARAM lParam)
{
    // Delegate real work to super
    LRESULT retValue = AwtComponent::WindowProc(message, wParam, lParam);

    // After-hooks for workarounds
    switch (message) {

      // Work around a windows bug described in KB article Q73839.
      // Need to update focus indicator on scrollbar if thumb
      // proportion or thumb position was changed.

      case WM_SIZE:
      case SBM_SETSCROLLINFO:
      case SBM_SETRANGE:
      case SBM_SETRANGEREDRAW:
	  if (::GetFocus() == GetHWnd())
	      UpdateFocusIndicator();
          break;
    }

    return retValue;
}


// Fix for a race condition when the WM_LBUTTONUP is picked by the AWT
// message loop before(!) the windows internal message loop for the
// scrollbar is started in response to WM_LBUTTONDOWN.  See KB article
// Q102552.
//
// Note that WM_LBUTTONUP is processed by the windows internal message
// loop.  May be we can synthesize a MOUSE_RELEASED event but that
// seems kludgey, so we'd better left this as is for now.

MsgRouting AwtScrollbar::WmMouseDown(UINT flags, int x, int y, int button)
{
    // We pass the WM_LBUTTONDOWN up to Java, but we process it
    // immediately as well to avoid the race.  Later when this press
    // event returns to us wrapped into a WM_AWT_HANDLE_EVENT we
    // ignore it in the HandleEvent below.  This means that we can not
    // consume the mouse press in the Java world.

    MsgRouting usualRoute = AwtComponent::WmMouseDown(flags, x, y, button);

    if (button == LEFT_BUTTON)
	return mrDoDefault;    // Force immediate processing to avoid the race.
    else
	return usualRoute;
}

MsgRouting AwtScrollbar::HandleEvent(MSG *msg, BOOL synthetic)
{
    if (msg->message == WM_LBUTTONDOWN || msg->message == WM_LBUTTONDBLCLK) {
	// Left button press was already routed to default window
	// procedure in the WmMouseDown above.  Propagating synthetic
	// press seems like a bad idea as internal message loop
	// doesn't know how to unwrap synthetic release.
	delete msg;
	return mrConsume;
    }
    else {
	return AwtComponent::HandleEvent(msg, synthetic);
    }
}


// Work around a windows bug descrbed in KB article Q73839.  Reset
// focus on scrollbars to update focus indicator.  The article advises
// to disable/enable the scrollbar, but simply resetting the focus is
// sufficient.
void AwtScrollbar::UpdateFocusIndicator()
{
    ASSERT(::GetFocus() == GetHWnd());
    m_ignoreFocusEvents = TRUE;
    ::SetFocus(NULL);		// give up focus
    ::SetFocus(GetHWnd());	// force windows to update focus indicator
    m_ignoreFocusEvents = FALSE;
}

MsgRouting AwtScrollbar::WmKillFocus(HWND hWndGot)
{
    if (m_ignoreFocusEvents) {
	// We are voluntary giving up focus and will get it back
	// immediately.  This is necessary to force windows to update
	// the focus indicator.
	return mrDoDefault;
    }
    else {
	return AwtComponent::WmKillFocus(hWndGot);
    }
}

MsgRouting AwtScrollbar::WmSetFocus(HWND hWndLost)
{
    if (m_ignoreFocusEvents) {
	// We have voluntary gave up focus and are getting it back
	// now.  This is necessary to force windows to update the
	// focus indicator.
	return mrDoDefault;
    }
    else {
	return AwtComponent::WmSetFocus(hWndLost);
    }
}


void AwtScrollbar::SetValues(int value, int visible, int minimum, int maximum)
{
    BOOL update_p = ::IsWindowEnabled(GetHWnd());

    if (IS_WINVER_ATLEAST(3,51)) {
	SCROLLINFO si;
	si.cbSize = sizeof si;
	si.fMask = SIF_POS | SIF_PAGE | SIF_RANGE;
	si.nMin = minimum;
	si.nMax = maximum - 1;
	si.nPage = visible;
	si.nPos = value;
        ::SetScrollInfo(GetHWnd(), SB_CTL, &si, update_p);
    }
    else {
        ::SetScrollRange(GetHWnd(), SB_CTL, minimum, maximum - visible, FALSE);
        ::SetScrollPos(GetHWnd(), SB_CTL, value, update_p);
    }
}


// In a windows app one would call SetScrollInfo from WM_[HV]SCROLL
// handler directly.  Since we call SetScrollInfo from Java world
// after scroll handler is over next WM_[HV]SCROLL event can be
// delivered before SetScrollInfo was called in response to the
// previous one and thus we would fire exactly the same event which
// will only contribute to the growth of the backlog of scroll events.

const char * const AwtScrollbar::SbNdragAbsolute = "dragAbsolute";
const char * const AwtScrollbar::SbNdragBegin    = "dragBegin";
const char * const AwtScrollbar::SbNdragEnd      = "dragEnd";
const char * const AwtScrollbar::SbNlineDown     = "lineDown";
const char * const AwtScrollbar::SbNlineUp       = "lineUp";
const char * const AwtScrollbar::SbNpageDown     = "pageDown";
const char * const AwtScrollbar::SbNpageUp       = "pageUp";

inline void
AwtScrollbar::DoScrollCallbackCoalesce(const char* methodName, int newPos)
{
    if (!(methodName == m_prevCallback && newPos == m_prevCallbackPos)) {
	DoCallback(methodName, "(I)V", newPos);
	m_prevCallback = methodName;
	m_prevCallbackPos = newPos;
    }
}


MsgRouting AwtScrollbar::WmVScroll(UINT scrollCode, UINT pos, HWND hScrollbar)
{
    int minVal, maxVal;    // scrollbar range
    int minPos, maxPos;    // thumb positions (max depends on visible amount)
    int curPos, newPos;

    // For drags we have old (static) and new (dynamic) thumb positions
    int dragP = scrollCode == SB_THUMBPOSITION || scrollCode == SB_THUMBTRACK;
    int thumbPos;

    if (IS_WINVER_ATLEAST(3,51)) { // New scrollbar API
	SCROLLINFO si;
	si.cbSize = sizeof si;
	si.fMask = SIF_POS | SIF_PAGE | SIF_RANGE;

	// From, _Win32 Programming_, by Rector and Newcommer, p. 185:
	// "In some of the older documentation on Win32 scroll bars,
	// including that published by Microsoft, you may read that
	// you *cannot* obtain the scroll position while in a handler.
	// The SIF_TRACKPOS flag was added after this documentation
	// was published.  Beware of this older documentation; it may
	// have other obsolete features."
	if (dragP)
	    si.fMask |= SIF_TRACKPOS;

	VERIFY(::GetScrollInfo(GetHWnd(), SB_CTL, &si));
	curPos = si.nPos;
	minPos = minVal = si.nMin;

	// Upper bound of the range.  Note that adding 1 here is safe
	// and won't cause a wrap, since we have substracted 1 in the
	// SetValues above.
	maxVal = si.nMax + 1;

	// Meaningful maximum position is maximum - visible.
	maxPos = maxVal - si.nPage;

	// XXX: Documentation for SBM_SETRANGE says that scrollbar
	// range is limited by MAXLONG, which is 2**31, but when a
	// scroll range is greater than that, thumbPos is reported
	// incorrectly due to integer arithmetic wrap(s).
	if (dragP)
	    thumbPos = si.nTrackPos;
    }
    else {			// Do we still support NT 3.50?
	VERIFY(::GetScrollRange(GetHWnd(), SB_CTL, &minVal, &maxVal));
	curPos = ::GetScrollPos(GetHWnd(), SB_CTL);
	minPos = minVal;
	maxPos = maxVal;	// see SetValues

	// Try to work around 16 bit limit of the `pos'
	if (dragP) {
	    if (minVal >= 0 && maxVal <= 0xffff) {
		// If the range is within [0,2**16] - pos is correct
		thumbPos = pos;
	    }
	    else if ((__int64)maxVal - minVal <= 0xffff) {
		// The range fits into 16 bit so we can unambiguously
		// calculate the real thumb position.
		thumbPos = minPos + LOWORD(pos - LOWORD(minPos));
	    }
	    else {
		// If the range does not fit in 16 bit we loose...
		thumbPos = pos;	// XXX: wrong - but can't do anything about it
	    }
	}
    }    


    // NB: Beware arithmetic wrap when calculating newPos
    switch (scrollCode) {

      case SB_LINEUP:
	  if ((__int64)curPos - m_lineIncr > minPos)
	      newPos = curPos - m_lineIncr;
	  else
	      newPos = minPos;
          if (newPos != curPos)
              DoScrollCallbackCoalesce(SbNlineUp, newPos);
          break;

      case SB_LINEDOWN:
	  if ((__int64)curPos + m_lineIncr < maxPos)
	      newPos = curPos + m_lineIncr;
	  else
	      newPos = maxPos;
          if (newPos != curPos)
              DoScrollCallbackCoalesce(SbNlineDown, newPos);
          break;

      case SB_PAGEUP:
	  if ((__int64)curPos - m_pageIncr > minPos)
	      newPos = curPos - m_pageIncr;
	  else
	      newPos = minPos;
          if (newPos != curPos)
              DoScrollCallbackCoalesce(SbNpageUp, newPos);
          break;

      case SB_PAGEDOWN:
	  if ((__int64)curPos + m_pageIncr < maxPos)
	      newPos = curPos + m_pageIncr;
	  else
	      newPos = maxPos;
          if (newPos != curPos)
              DoScrollCallbackCoalesce(SbNpageDown, newPos);
          break;

      // In ideal world we should use different callbacks for
      // SB_THUMBTRACK and SB_THUMBPOSITION.  Application usually
      // processes SB_THUMBTRACK (real time scrolling) if it can
      // scroll fast.  If redraw required for scrolling is expensive,
      // application will process SB_THUMBPOSITION.  But java program
      // cannot control this choice via AWT API.

      // Since we do process SB_THUMBTRACK events and since the
      // position reported in the SB_THUMBPOSITION event is the same
      // as in the last SB_THUMBTRACK event, the guard expression
      // below will *always* be false and dragAbsolute shall never be
      // called.  The code below is left here just in case.

      case SB_THUMBPOSITION:	// Drag finished, final position
	  if (thumbPos != curPos)
	      DoScrollCallbackCoalesce(SbNdragAbsolute, thumbPos);
          break;

      case SB_THUMBTRACK:	// Track the drag in real time
	  if (thumbPos != curPos)
	      DoScrollCallbackCoalesce(SbNdragAbsolute, thumbPos);
          break;

      case SB_TOP:
	  if (minPos != curPos)
	      DoScrollCallbackCoalesce(SbNdragAbsolute, minPos);
          break;

      case SB_BOTTOM:
	  if (maxPos != curPos)
	      DoScrollCallbackCoalesce(SbNdragAbsolute, maxPos);
          break;

      case SB_ENDSCROLL:
	  // XXX: this workaround appears to be no longer necessary!
	  // Request a redraw at this point to work around problems
	  // with page up/down area not being redrawn correctly after
	  // multiple sequential scroll events, in a case when Java
	  // listeners take long time to complete.
	  ::InvalidateRect(GetHWnd(), NULL, TRUE);
	  m_prevCallback = NULL;
          DoCallback(SbNdragEnd, "(I)V", curPos);
          break;

      case 9 /*SB_BEGINSCROLL*/:
	  m_prevCallback = NULL;
          DoCallback(SbNdragBegin, "(I)V", curPos);
          break;
    }
    return mrDoDefault;
}

MsgRouting AwtScrollbar::WmHScroll(UINT scrollCode, UINT pos, HWND hScrollbar)
{
    return WmVScroll(scrollCode, pos, hScrollbar);
}


/************************************************************************
 * Scrollbar native methods
 */

extern "C" {

/*
 * Class:     java_awt_Scrollbar
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_Scrollbar_initIDs(JNIEnv *env, jclass cls)
{
    AwtScrollbar::lineIncrementID = env->GetFieldID(cls, "lineIncrement", "I");
    AwtScrollbar::pageIncrementID = env->GetFieldID(cls, "pageIncrement", "I");
    AwtScrollbar::orientationID = env->GetFieldID(cls, "orientation", "I");

    ASSERT(AwtScrollbar::lineIncrementID != NULL);
    ASSERT(AwtScrollbar::pageIncrementID != NULL);
    ASSERT(AwtScrollbar::orientationID != NULL);
}

} /* extern "C" */


/************************************************************************
 * WScrollbarPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WScrollbarPeer
 * Method:    setValues
 * Signature: (IIII)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WScrollbarPeer_setValues(JNIEnv *env, jobject self,
					      jint value, jint visible,
					      jint minimum, jint maximum)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtScrollbar* c = (AwtScrollbar*)pData;
    c->SetValues(value, visible, minimum, maximum);
}

/*
 * Class:     sun_awt_windows_WScrollbarPeer
 * Method:    setLineIncrement
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WScrollbarPeer_setLineIncrement(JNIEnv *env, jobject self,
						     jint increment)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtScrollbar* c = (AwtScrollbar*)pData;
    c->SetLineIncrement(increment);
}

/*
 * Class:     sun_awt_windows_WScrollbarPeer
 * Method:    setPageIncrement
 * Signature: (I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WScrollbarPeer_setPageIncrement(JNIEnv *env, jobject self,
						     jint increment)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtScrollbar* c = (AwtScrollbar*)pData;
    c->SetPageIncrement(increment);
}

/*
 * Class:     sun_awt_windows_WScrollbarPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WScrollbarPeer_create(JNIEnv *env, jobject self,
					   jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtScrollbar::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

} /* extern "C" */
