/*
 * @(#)awt_Scrollbar.h	1.14 98/09/30
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_SCROLLBAR_H
#define AWT_SCROLLBAR_H

#include "awt_Component.h"

#include "java_awt_Scrollbar.h"
#include "sun_awt_windows_WScrollbarPeer.h"


#define Java_java_awt_Scrollbar_HORIZONTAL    0
#define Java_java_awt_Scrollbar_VERTICAL      1


/************************************************************************
 * AwtScrollbar class
 */

class AwtScrollbar : public AwtComponent {
public:
  
    /* java.awt.Scrollbar fields */
    static jfieldID lineIncrementID;
    static jfieldID pageIncrementID;
    static jfieldID orientationID;

    AwtScrollbar();

    virtual const char* GetClassName();
    
    static AwtScrollbar* Create(jobject self, jobject parent);

    void SetValue(int value);
    void SetValues(int value, int visible, int minimum, int maximum);
    void SetLineIncrement(int value) { m_lineIncr = value; }
    void SetPageIncrement(int value) { m_pageIncr = value; }

    virtual LRESULT WindowProc(UINT message, WPARAM wParam, LPARAM lParam);

    /*
     * Windows message handler functions
     */
    virtual MsgRouting WmHScroll(UINT scrollCode, UINT pos, HWND hScrollBar);
    virtual MsgRouting WmVScroll(UINT scrollCode, UINT pos, HWND hScrollBar);

    // Work around KB Q73839 bug.
    virtual MsgRouting WmSetFocus(HWND hWndLost);
    virtual MsgRouting WmKillFocus(HWND hWndGot);

    // Prevent KB Q102552 race.
    virtual MsgRouting WmMouseDown(UINT flags, int x, int y, int button);
    virtual MsgRouting HandleEvent(MSG *msg, BOOL synthetic);

private:
    UINT          m_orientation; /* SB_HORZ or SB_VERT */
    AwtComponent* m_hostWindow;  /* if set, contains real hwnd 
				  * we are riding on */
    int           m_lineIncr;
    int           m_pageIncr;

    // Work around KB Q73839 bug.
    BOOL m_ignoreFocusEvents;
    void UpdateFocusIndicator();

    // Don't do redundant callbacks.
    const char *m_prevCallback;
    int m_prevCallbackPos;

    static const char * const SbNdragAbsolute;
    static const char * const SbNdragBegin;
    static const char * const SbNdragEnd;
    static const char * const SbNlineDown;
    static const char * const SbNlineUp;
    static const char * const SbNpageDown;
    static const char * const SbNpageUp;

    void DoScrollCallbackCoalesce(const char* methodName, int newPos);
};

#endif /* AWT_SCROLLBAR_H */

