/*
 * @(#)awt_TextArea.cpp	1.32 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_TextArea.h"
#include "awt_TextComponent.h"

#include "awt_Canvas.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtTextArea fields
 */

jfieldID AwtTextArea::scrollbarVisibilityID;


/************************************************************************
 * AwtTextArea methods
 */

AwtTextArea::AwtTextArea() {
}

/* Create a new AwtTextArea object and window.   */
AwtTextArea* AwtTextArea::Create(jobject peer, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    int pData;
    AwtCanvas* awtParent;
    jobject target = 0;
    AwtTextArea* c;
    JNI_CHECK_PEER_GOTO(parent, done);

    awtParent = (AwtCanvas*)pData;
    JNI_CHECK_NULL_GOTO(awtParent, "null awtParent", done);

    target = env->GetObjectField(peer, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    c = new AwtTextArea();
    JNI_CHECK_NULL_GOTO(c, "AwtTextArea() failed", done);

    {
        /* Adjust style for scrollbar visibility and word wrap  */
        DWORD scroll_style;
	jint scrollbarVisibility =
	    env->GetIntField(target, AwtTextArea::scrollbarVisibilityID);

	switch (scrollbarVisibility) {
	case java_awt_TextArea_SCROLLBARS_NONE:
            scroll_style = ES_AUTOVSCROLL;
	    break;
	case java_awt_TextArea_SCROLLBARS_VERTICAL_ONLY:
	    scroll_style = WS_VSCROLL | ES_AUTOVSCROLL;
	    break;
	case java_awt_TextArea_SCROLLBARS_HORIZONTAL_ONLY:
	    scroll_style = WS_HSCROLL | ES_AUTOHSCROLL | ES_AUTOVSCROLL;
	    break;
	case java_awt_TextArea_SCROLLBARS_BOTH:
	    scroll_style = WS_VSCROLL | WS_HSCROLL |
	        ES_AUTOVSCROLL | ES_AUTOHSCROLL;
	    break;
	}

	DWORD style = WS_CHILD | WS_CLIPSIBLINGS | ES_LEFT | ES_MULTILINE |
	    ES_WANTRETURN | ES_NOHIDESEL | scroll_style |
	    (IS_WIN4X ? 0 : WS_BORDER);
	DWORD exStyle = IS_WIN4X ? WS_EX_CLIENTEDGE : 0;
	if (GetRTL()) {
	    exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
	    if (GetRTLReadingOrder())
	        exStyle |= WS_EX_RTLREADING;
	}

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	c->CreateHWnd(env, "", style, exStyle,
		      x, y, width, height,
		      awtParent->GetHWnd(),
		      (HMENU)awtParent->CreateControlID(),
		      ::GetSysColor(COLOR_WINDOWTEXT),
		      ::GetSysColor(COLOR_WINDOW),
		      peer);

	c->m_backgroundColorSet = TRUE;
	  /* suppress inheriting parent's color. */
    }
 done:
    env->DeleteLocalRef(target);

    return c;
}

/* Count how many '\n's are there in jStr */
int AwtTextArea::CountNewLines(JNIEnv *env, jstring jStr, int maxlen)
{
    int nNewlines = 0;

    if (jStr == NULL) {
	return nNewlines;
    }
    WCHAR* string = TO_WSTRING(jStr);
    for (int i = 0; i < maxlen && i < env->GetStringLength(jStr); i++) {
        if (string[i] == L'\n') {
            nNewlines++;
        }
    }
    return nNewlines;
}


/************************************************************************
 * TextArea native methods
 */

extern "C" {

/*
 * Class:     java_awt_TextArea
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_TextArea_initIDs(JNIEnv *env, jclass cls)
{
    AwtTextArea::scrollbarVisibilityID =
	env->GetFieldID(cls, "scrollbarVisibility", "I");

    ASSERT(AwtTextArea::scrollbarVisibilityID != NULL);
}

} /* extern "C" */


/************************************************************************
 * WTextAreaPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WTextAreaPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextAreaPeer_create(JNIEnv *env, jobject self,
					  jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtTextArea::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

/*
 * Class:     sun_awt_windows_WTextAreaPeer
 * Method:    replaceText
 * Signature: (Ljava/lang/String;II)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextAreaPeer_replaceText(JNIEnv *env, jobject self,
					       jstring text,
					       jint start, jint end)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(text, "null string");

    AwtTextComponent* c = (AwtTextComponent*)pData;
    jstring targetText =
	Java_sun_awt_windows_WTextComponentPeer_getText(env, self);

    /* Adjust positions with the CRs going to be added. */
    int cr_start = AwtTextArea::CountNewLines(env, targetText, start);
    int cr_end = AwtTextArea::CountNewLines(env, targetText, end);

    /*
     * After wrapper functions are made, it must be still remained to
     * distinguish Unicode TextArea (NT) and multibyte TextArea (95).  For
     * multibyte TextArea, the start and end position must be recalculated.
     */
    if (IS_NT) {
	start += cr_start;
	end += cr_end;

        int length = env->GetStringLength(text) + 1;
        // Bugid 4141477 - Can't use TO_WSTRING here because it uses alloca 
        // WCHAR* buffer = TO_WSTRING(text);
        WCHAR *buffer = new WCHAR[length];
        env->GetStringRegion(text, 0, length-1, buffer);
        buffer[length-1] = '\0';
        buffer = AwtTextComponent::AddCR(buffer, length);
        c->SendMessageW(EM_SETSEL, start, end);
        c->SendMessageW(EM_REPLACESEL, FALSE, (LPARAM)buffer);
        delete[] buffer;
    } else {
#ifdef WIN32
	ASSERT(IS_WIN95);
#endif
	/*
	 * If the string in the target text area is multibyte,
	 * the start and end position must be recalculate to get
	 * the right positions.
	 */
        WCHAR* wStartbuffer = new WCHAR[start + 1];
	env->GetStringRegion(targetText, 0, start, wStartbuffer);
        wStartbuffer[start] = L'\0';

        WCHAR* wEndbuffer = new WCHAR[end + 1];
        env->GetStringRegion(targetText, 0, end, wEndbuffer);
        wEndbuffer[end] = L'\0';

        char* aStartbuffer = W2A(wStartbuffer);
	char* aEndbuffer = W2A(wEndbuffer);
	start = strlen(aStartbuffer) + cr_start;
	end = strlen(aEndbuffer) + cr_end;

	delete[] wStartbuffer;
	delete[] wEndbuffer;

        /* Copy Java string into a buffer AddCR can replace if necessary. */
	int length = env->GetStringLength(text);
	char *buffer = new char[2*length + 1];
        AwtFont::javaString2multibyte(env, text, buffer, length);
        buffer = AwtTextComponent::AddCR(buffer, length);

        c->SendMessage(EM_SETSEL, start, end);
        c->SendMessage(EM_REPLACESEL, FALSE, (LPARAM)buffer);

        delete[] buffer;
    }
}

/*
 * Class:     sun_awt_windows_WTextAreaPeer
 * Method:    insertText
 * Signature: (Ljava/lang/String;I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextAreaPeer_insertText(JNIEnv *env, jobject self,
					      jstring text, jint pos)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(text, "null string");
    Java_sun_awt_windows_WTextAreaPeer_replaceText(env, self, text, pos, pos);
}

} /* extern "C" */
