/*
 * @(#)awt_TextComponent.cpp	1.26 98/06/04
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_TextComponent.h"

#include "awt_Canvas.h"

#include "jni.h"
#include "awt_Font.h"


/************************************************************************
 * AwtTextComponent fields
 */

/* java.awt.TextComponent fields */
jfieldID AwtTextComponent::canAccessClipboardID;


/************************************************************************
 * AwtTextComponent methods
 */

AwtTextComponent::AwtTextComponent() {
    m_synthetic = FALSE;
}

const char* AwtTextComponent::GetClassName() {
    return "EDIT";  /* System provided edit control class */
}

/* Set a suitable font to IME against the component font. */
void AwtTextComponent::SetFont(AwtFont* font)
{
    ASSERT(font != NULL);
    if (font->GetAscent() < 0) {
	AwtFont::SetupAscent(font);
    }

    int index = font->GetInputHFontIndex();
    if (index < 0)
	/* In this case, user cannot get any suitable font for input. */
        index = 0;	

    //im --- changed for over the spot composing
    m_hFont = font->GetHFont(index);
    SendMessage(WM_SETFONT, (WPARAM)m_hFont, MAKELPARAM(FALSE, 0));
    VERIFY(::InvalidateRect(GetHWnd(), NULL, TRUE));
    //im --- end

}

/*
 * Returns a copy of pStr where \n have been converted to \r\n.
 * If the result != 'pStr', 'pStr' has been freed.
 */
WCHAR *AwtTextComponent::AddCR(WCHAR *pStr, int nStrLen)
{
    int nNewlines = 0;
    int i, nResultIx, nLen;
    WCHAR *result;

    /* check to see if there are any LF's */
    if (wcschr(pStr, L'\n') == NULL) {
        return pStr;
    }

    /* count newlines */
    for (i=0; pStr[i] != 0; i++) {
        if (pStr[i] == L'\n') {
            nNewlines++;
        }
    }
    nLen = i;

    if (nLen + nNewlines + 1 > nStrLen) {
        result = new WCHAR[nLen + nNewlines + 1];
    } else {
        result = pStr;
    }
    nResultIx = nLen + nNewlines;
    for (i=nLen; i>=0; i--) {
        result[nResultIx--] = pStr[i];
        if (pStr[i] == L'\n') {
            result[nResultIx--] = L'\r';
        }
    }
    ASSERT(nResultIx == -1);

    if (pStr != result) {
        /* We realloc'd, so delete old buffer. */
        delete[] pStr;
    }
    return result;
}

char *AwtTextComponent::AddCR(char *pStr, int nStrLen)
{
    int nNewlines = 0;
    int i, nResultIx, nLen;
    char *result;

    /* check to see if there are any LF's */
    if (strchr(pStr, '\n') == NULL) {
        return pStr;
    }

    /* count newlines */
    for (i=0; pStr[i] != 0; i++) {
        if (pStr[i] == '\n') {
            nNewlines++;
        }
    }
    nLen = i;

    if (nLen + nNewlines + 1 > nStrLen) {
        result = new char[nLen + nNewlines + 1];
    } else {
        result = pStr;
    }
    nResultIx = nLen + nNewlines;
    for (i=nLen; i>=0; i--) {
        result[nResultIx--] = pStr[i];
        if (pStr[i] == '\n') {
            result[nResultIx--] = '\r';
        }
    }
    ASSERT(nResultIx == -1);

    if (pStr != result) {
        /* We realloc'd, so delete old buffer. */
        delete[] pStr;
    }
    return result;
}

/* Converts \r\n pairs to \n. */
int AwtTextComponent::RemoveCR(WCHAR *pStr)
{
    int i, nLen = 0;

    if (pStr) {
        /* check to see if there are any CR's */
        if (wcschr(pStr, L'\r') == NULL) {
            return wcslen(pStr);
        }

        for (i=0; pStr[i] != 0; i++) {
            if (pStr[i] == L'\r' && pStr[i + 1] == L'\n') {
                i++;
            }
            pStr[nLen++] = pStr[i];
        }
        pStr[nLen] = 0;
    }
    return nLen;
}

/*
 * This routine is safe for converting multibyte strings, because
 * control characters such as \n and \r are never part of a multibyte
 * character sequence.
 */
int AwtTextComponent::RemoveCR(char *pStr)
{
    int i, nLen = 0;

    if (pStr) {
        /* check to see if there are any CR's */
        if (strchr(pStr, '\r') == NULL) {
            return strlen(pStr);
        }

        for (i=0; pStr[i] != 0; i++) {
            if (pStr[i] == '\r' && pStr[i + 1] == '\n') {
                i++;
            }
            pStr[nLen++] = pStr[i];
        }
        pStr[nLen] = 0;
    }
    return nLen;
}

MsgRouting
AwtTextComponent::WmNotify(UINT notifyCode)
{
    if (notifyCode == EN_CHANGE) {
      DoCallback("valueChanged", "()V");
    }
    return mrDoDefault;
}

/*
 * Store the 'synthetic' parameter so that the WM_PASTE security check
 * happens only for synthetic events.
 */
MsgRouting
AwtTextComponent::HandleEvent(MSG *msg, BOOL synthetic)
{
    MsgRouting returnVal;

    m_synthetic = synthetic;
    returnVal = AwtComponent::HandleEvent(msg, synthetic);
    m_synthetic = FALSE;
    return returnVal;
}

/*
 * If this Paste is occuring because of a synthetic Java event (e.g.,
 * a synthesized <CTRL>-V KeyEvent), then verify that the TextComponent
 * has permission to access the Clipboard before pasting. If permission
 * is denied, we should throw a SecurityException, but currently do not
 * because when we detect the security violation, we are in the Toolkit
 * thread, not the thread which dispatched the illegal event.
 */
MsgRouting
AwtTextComponent::WmPaste()
{
    if (m_synthetic) {
        JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	if (env->EnsureLocalCapacity(1) < 0) {
	    return mrConsume;
	}
	jobject target = GetTarget(env);
	jboolean canAccessClipboard = 
	    env->GetBooleanField(target,
				 AwtTextComponent::canAccessClipboardID);
	env->DeleteLocalRef(target);
	return (canAccessClipboard) ? mrDoDefault : mrConsume;
    }
    else {
        return mrDoDefault;
    }
}

//im --- override to over the spot composition
void AwtTextComponent::SetCompositionWindow()
{
    HIMC hIMC = ImmGetContext(m_hwnd);
    COMPOSITIONFORM cf = { CFS_POINT, {0,0}, {0,0,0,0} };
    GetCaretPos(&(cf.ptCurrentPos));
    ImmSetCompositionWindow(hIMC, &cf);

    LOGFONT lf;
    GetObject(m_hFont, sizeof(LOGFONT), &lf);
    ImmSetCompositionFont(hIMC, &lf);
}
//im --- end


/************************************************************************
 * WTextComponentPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WTextComponentPeer
 * Method:    getText
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL
Java_sun_awt_windows_WTextComponentPeer_getText(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN_NULL(self);

    AwtTextComponent* c = (AwtTextComponent*)pData;

    BOOL isNT = IS_NT;
    int len;
    /* before wrapper function of GetWindowTextLengthW is made, 
     * we have to distinguish the running environment
     */
    if (isNT) {
	len = ::GetWindowTextLengthW(c->GetHWnd());
    } else {
        len = ::GetWindowTextLengthA(c->GetHWnd());
    }
    
    jstring js;
    if (len == 0) {
        /* Make java null string */
	js = env->NewString(new jchar[0], 0);
    } else {
	/*
	 * before wrapper function of GetWindowText is made so that 2
	 * types of AwtTextComponent::GetText are unified, we have to
	 * distinguish the running environment 
	 */
        if (isNT) {
	    WCHAR* buf = new WCHAR[len + 1];
            c->GetText(buf, len + 1);
            AwtTextComponent::RemoveCR(buf);
            js = env->NewString(buf, wcslen(buf));
            delete [] buf;
        } else {
#ifdef WIN32
            ASSERT(IS_WIN95);
#endif
	    /*
	     * Text in a TextComponent is written in platform-specific
	     * codeset because it may be output from IME, which
	     * doesn't emit unicode.  
	     */
            char* buffer = new char[len + 1];	/* Including null terminater */
            c->GetText(buffer, len + 1);
            AwtTextComponent::RemoveCR(buffer);
	    js = AwtFont::multibyte2javaString(env, buffer);/* convert to unicode. */
            delete[] buffer;
        }
    }
    return js;
}

/*
 * Class:     sun_awt_windows_WTextComponentPeer
 * Method:    setText
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextComponentPeer_setText(JNIEnv *env, jobject self,
						jstring text)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtTextComponent* c = (AwtTextComponent*)pData;
    /*
     * before wrapper function of SetWindowText is made, two versions
     * of SetText will be unified.  Before that we have to distinguish
     * the running environment.
     */
    if (IS_NT) {
    	int length = env->GetStringLength(text);
        WCHAR* buffer = new WCHAR[length + 1];
	env->GetStringRegion(text, 0, length, buffer);
	buffer[length] = 0;
        buffer = AwtTextComponent::AddCR(buffer, length + 1);
        c->SetText(buffer);
        delete[] buffer;

    } else {
#ifdef WIN32
        ASSERT(IS_WIN95);
#endif
	int length = env->GetStringLength(text);
        char* buffer = new char[length * 2 + 1];

	AwtFont::javaString2multibyte(env, text, buffer, length);
        buffer = AwtTextComponent::AddCR(buffer, length);
        c->SetText(buffer);

        delete[] buffer;
    }
}

/*
 * Class:     sun_awt_windows_WTextComponentPeer
 * Method:    getSelectionStart
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WTextComponentPeer_getSelectionStart(JNIEnv *env,
							  jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN_NULL(self);
    AwtComponent* c = (AwtComponent*)pData;
    long start;
    c->SendMessage(EM_GETSEL, (WPARAM)&start);
    return getJavaSelPos((AwtTextComponent*)c, start);
}

/*
 * Class:     sun_awt_windows_WTextComponentPeer
 * Method:    getSelectionEnd
 * Signature: ()I
 */
JNIEXPORT jint JNICALL 
Java_sun_awt_windows_WTextComponentPeer_getSelectionEnd(JNIEnv *env,
							jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN_NULL(self);
    AwtComponent* c = (AwtComponent*)pData;
    long end;
    c->SendMessage(EM_GETSEL, 0, (LPARAM)&end);
    return getJavaSelPos((AwtTextComponent*)c, end);
}

/*
 * Class:     sun_awt_windows_WTextComponentPeer
 * Method:    select
 * Signature: (II)V
 */
JNIEXPORT void JNICALL 
Java_sun_awt_windows_WTextComponentPeer_select(JNIEnv *env, jobject self,
					       jint start, jint end)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* c = (AwtComponent*)pData;

    c->SendMessage(EM_SETSEL, 
		   getWin32SelPos((AwtTextComponent*)c, start),
		   getWin32SelPos((AwtTextComponent*)c, end));
    c->SendMessage(EM_SCROLLCARET);
}

/*
 * Class:     sun_awt_windows_WTextComponentPeer
 * Method:    enableEditing
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextComponentPeer_enableEditing(JNIEnv *env, 
						      jobject self,
						      jboolean on)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtTextComponent* c = (AwtTextComponent*)pData;
    c->SendMessage(EM_SETREADONLY, !on);
}

/*
 * Class:     sun_awt_windows_WTextComponentPeer
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextComponentPeer_initIDs(JNIEnv *env, jclass cls)
{
    cls = env->FindClass("java/awt/TextComponent");
    if (cls != NULL) {
        AwtTextComponent::canAccessClipboardID = 
	    env->GetFieldID(cls, "canAccessClipboard", "Z");
	ASSERT(AwtTextComponent::canAccessClipboardID != NULL);
    }
}

} /* extern "C" */

static long getJavaSelPos(AwtTextComponent *c, long orgPos)
{
    long mblen;
    long pos = 0;
    long cur = 0;
    char *mbbuf;

    if ((mblen = c->GetTextLength()) == 0)
	   return 0;
    mbbuf = new char[mblen + 1];
    c->GetText(mbbuf, mblen + 1);
	while (cur < orgPos && pos++ < mblen) {
	   if (IsDBCSLeadByte(mbbuf[cur])) {
	      cur++;
	   } else if (mbbuf[cur] == '\r' && mbbuf[cur + 1] == '\n') {
	         cur++;
	   }
	  cur++;
	}
    delete[] mbbuf;
	return pos;
}

static long getWin32SelPos(AwtTextComponent *c, long orgPos)
{
    long mblen;
    long pos = 0;
    long cur = 0;
    char *mbbuf;

    if ((mblen = c->GetTextLength()) == 0)
       return 0;
    mbbuf = new char[mblen + 1];
    c->GetText(mbbuf, mblen + 1);

	while (cur < orgPos && pos < mblen) {
       if (IsDBCSLeadByte(mbbuf[pos])) {
	      pos++;
	   } else if (mbbuf[pos] == '\r' && mbbuf[pos + 1] == '\n') {
	    pos++;
	   }
       pos++;
	   cur++;
	}
    delete[] mbbuf;
	return pos;
}
