/*
 * @(#)awt_TextComponent.h	1.18 98/06/04
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#ifndef AWT_TEXTCOMPONENT_H
#define AWT_TEXTCOMPONENT_H

#include "awt_Component.h"

#include "java_awt_TextComponent.h"
#include "sun_awt_windows_WTextComponentPeer.h"


/************************************************************************
 * Native method declarations (until we switch to JNI fully)
 */

extern "C" {

JNIEXPORT jstring JNICALL
Java_sun_awt_windows_WTextComponentPeer_getText(JNIEnv *env, jobject self);

} /* extern "C" */


/************************************************************************
 * AwtTextComponent class
 */

class AwtTextComponent : public AwtComponent {
public:
    /* java.awt.TextComponent canAccessClipboard field ID */
    static jfieldID canAccessClipboardID;

    AwtTextComponent();

    virtual const char* GetClassName();

    static WCHAR *AddCR(WCHAR *pStr, int nStrLen);
    static int RemoveCR(WCHAR *pStr);

    static char *AddCR(char *pStr, int nStrLen);
    static int RemoveCR(char *pStr);

    INLINE void SetText(const char* text) { 
        ::SetWindowText(GetHWnd(), text); 
    }
    INLINE void SetText(const WCHAR* text) {
        ::SetWindowTextW(GetHWnd(), text); 
    }

    INLINE int GetText(char* buffer, int size) { 
        return ::GetWindowText(GetHWnd(), buffer, size); 
    }
    INLINE int GetText(WCHAR* buffer, int size) {
        return ::GetWindowTextW(GetHWnd(), buffer, size); 
    }

    /*
     * For TextComponents that contains WCHAR strings or messages with
     * WCHAR parameters.  
     */
    INLINE LRESULT SendMessageW(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0)
    {
        ASSERT(GetHWnd());
        return ::SendMessageW(GetHWnd(), msg, wParam, lParam);
    }

    void SetFont(AwtFont* font);

    /*
     * Windows message handler functions
     */
    MsgRouting WmNotify(UINT notifyCode);
    MsgRouting HandleEvent(MSG *msg, BOOL synthetic);
    MsgRouting WmPaste();

    //im --- for over the spot composition
    void SetCompositionWindow(); 

private:
    HFONT m_hFont;
    //im --- end

    // Used to prevent untrusted code from synthesizing a WM_PASTE message
    // by posting a <CTRL>-V KeyEvent
    BOOL    m_synthetic; 
};

extern "C" {
static long getJavaSelPos(AwtTextComponent *c, long orgPos);
static long getWin32SelPos(AwtTextComponent *c, long orgPos);
} /* extern "C" */

#endif /* AWT_TEXTCOMPONENT_H */
