/*
 * @(#)awt_TextField.cpp	1.30 98/09/23
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Toolkit.h"
#include "awt_TextField.h"

#include "awt_Canvas.h"

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtTextField methods
 */

AwtTextField::AwtTextField() {
}

/* Create a new AwtTextField object and window.   */
AwtTextField* AwtTextField::Create(jobject peer, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    int pData;
    AwtCanvas* awtParent;
    jobject target = 0;
    AwtTextField* c;
    JNI_CHECK_PEER_GOTO(parent, done);
    awtParent = (AwtCanvas*)pData;

    JNI_CHECK_NULL_GOTO(awtParent, "null awtParent", done);

    target = env->GetObjectField(peer, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    c = new AwtTextField();
    JNI_CHECK_NULL_GOTO(c, "AwtTextField() failed", done);

    {
        DWORD style = WS_CHILD | WS_CLIPSIBLINGS |
	    ES_LEFT | ES_AUTOHSCROLL | ES_NOHIDESEL |
	    (IS_WIN4X ? 0 : WS_BORDER);
	DWORD exStyle = IS_WIN4X ? WS_EX_CLIENTEDGE : 0;
        if (GetRTL()) {
	    exStyle |= WS_EX_RIGHT | WS_EX_LEFTSCROLLBAR;
	    if (GetRTLReadingOrder())
	        exStyle |= WS_EX_RTLREADING;
	}

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	c->CreateHWnd(env, "", style, exStyle,
		      x, y, width, height,
		      awtParent->GetHWnd(),
		      (HMENU)awtParent->CreateControlID(),
		      ::GetSysColor(COLOR_WINDOWTEXT),
		      ::GetSysColor(COLOR_WINDOW),
		      peer);

	c->m_backgroundColorSet = TRUE;  /* suppress inheriting parent's color. */
    }
 done:
    env->DeleteLocalRef(target);

    return c;
}

MsgRouting AwtTextField::WmChar(UINT character, UINT repCnt,
				UINT flags, BOOL system)
{
    if (!system) {
        if (character == VK_RETURN) {
            DoCallback("handleAction", "()V");
            return mrConsume;
        }
        if (character == VK_TAB) {
	  /* illegal character -- handled in Java code. */
            return mrConsume;
        }
    }
    return AwtComponent::WmChar(character, repCnt, flags, system);
}

/************************************************************************
 * WTextFieldPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WTextFieldPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextFieldPeer_create(JNIEnv *env, jobject self,
					   jobject parent)
{
    int pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtTextField::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

/*
 * Class:     sun_awt_windows_WTextFieldPeer
 * Method:    setEchoCharacter
 * Signature: (C)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WTextFieldPeer_setEchoCharacter(JNIEnv *env, jobject self,
						     jchar ch)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtComponent* c = (AwtComponent*)pData;
    c->SendMessage(EM_SETPASSWORDCHAR, ch);
}

} /* extern "C" */
