/*
 * @(#)awt_Toolkit.cpp	1.95 98/10/02
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_AWTEvent.h"
#include "awt_Component.h"
#include "awt_Canvas.h"
#include "awt_Clipboard.h"
#include "awt_Frame.h"
#include "awt_Font.h"
#include "awt_Graphics.h"
#include "awt_InputEvent.h"
#include "awt_KeyEvent.h"
#include "awt_List.h"
#include "awt_PopupMenu.h"
#include "awt_Toolkit.h"
#include "CmdIDList.h"
#include <awt_image.h>

#include <awt_DnDDT.h>
#include <awt_DnDDS.h>

#include <java_awt_Toolkit.h>
#include <java_awt_event_InputMethodEvent.h>
#include <java_awt_peer_ComponentPeer.h>
#include <sun_awt_windows_WEmbeddedFrame.h>
#include <sun_awt_windows_WEmbeddedFramePeer.h>

extern CriticalSection imageLock;

/************************************************************************
 * Utilities
 */

/* Initialize the Java VM instance variable when the library is 
   first loaded */
JavaVM *jvm;

JNIEXPORT jint JNICALL
JNI_OnLoad(JavaVM *vm, void *reserved)
{
    jvm = vm;
    return JNI_VERSION_1_2;
}

#define IDT_AWT_MOUSECHECK 0x101

static const char* szAwtToolkitClassName = "SunAwtToolkit";

extern "C" {
    extern int sysAtexit(void (*func)(void));
}

UINT AwtToolkit::GetAsyncMouseKeyState()
{
    static BOOL mbSwapped = ::GetSystemMetrics(SM_SWAPBUTTON);
    UINT mouseKeyState = 0;

    if (HIBYTE(::GetAsyncKeyState(VK_CONTROL)))
        mouseKeyState |= MK_CONTROL;
    if (HIBYTE(::GetAsyncKeyState(VK_SHIFT)))
        mouseKeyState |= MK_SHIFT;
    if (HIBYTE(::GetAsyncKeyState(VK_LBUTTON)))
        mouseKeyState |= (mbSwapped ? MK_RBUTTON : MK_LBUTTON);
    if (HIBYTE(::GetAsyncKeyState(VK_RBUTTON)))
        mouseKeyState |= (mbSwapped ? MK_LBUTTON : MK_RBUTTON);
    if (HIBYTE(::GetAsyncKeyState(VK_MBUTTON)))
        mouseKeyState |= MK_MBUTTON;
    return mouseKeyState;
}

//
// Normal ::GetKeyboardState call only works if current thread has
// a message pump, so provide a way for other threads to get
// the keyboard state
//
void AwtToolkit::GetAsyncKeyboardState(PBYTE keyboardState)
{
    CriticalSection::Lock	l(AwtToolkit::GetInstance().m_lockKB);
    ASSERT(!IsBadWritePtr(keyboardState, KB_STATE_SIZE));
    memcpy(keyboardState, AwtToolkit::GetInstance().m_lastKeyboardState,
	   KB_STATE_SIZE);
}

/************************************************************************
 * Exported functions
 */

extern "C" BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, 
                                 LPVOID)
{
    AwtToolkit::GetInstance().SetModuleHandle(hModule);
    return TRUE;
}

extern "C" __declspec(dllexport) BOOL AwtInit(void) {
    return AwtToolkit::GetInstance().Initialize(FALSE);
}

extern "C" __declspec(dllexport) jobject
CreateEmbeddedFrame(JNIEnv *env, jint handle)
{
    static jclass frameCls;
    static jmethodID frameConst;
    if (frameCls == NULL) {
	frameCls = env->FindClass("sun/awt/windows/WEmbeddedFrame");
	frameCls = (jclass)env->NewGlobalRef(frameCls);
	if (frameCls == NULL) {
	    return NULL;
	}
	frameConst = env->GetMethodID(frameCls, "<init>", "(I)V");
	if (frameConst == NULL) {
	    return NULL;
	}
    }
    return env->NewObject(frameCls, frameConst, handle);
}

extern "C" __declspec(dllexport) void 
PrintEmbeddedFrame(JNIEnv *env, jobject self, HDC hdc)
{
    JNI_CHECK_NULL_RETURN(self, "null WEmbeddedFrame");
    JNU_CallMethodByName(env, 0, self, "print", "(I)V", hdc);
}


extern "C" void awt_dnd_initialize();
extern "C" void awt_dnd_uninitialize();

/************************************************************************
 * AwtToolkit fields
 */

AwtToolkit AwtToolkit::theInstance;

/* fieldIDs for WToolkit fields accessed from native code */
jfieldID AwtToolkit::dbgTraceID;
jfieldID AwtToolkit::dbgVerifyID;
jfieldID AwtToolkit::dbgBreakID;
jfieldID AwtToolkit::dbgHeapCheckID;

jmethodID AwtToolkit::getDefaultToolkitMID;
jmethodID AwtToolkit::getFontMetricsMID;


/************************************************************************
 * JavaStringBuffer method
 */

JavaStringBuffer::JavaStringBuffer(JNIEnv *env, jstring jstr) {
    if (jstr != NULL) {
        int length = env->GetStringLength(jstr);
        buffer = new char[length * 2 + 1];
        AwtFont::javaString2multibyte(env, jstr, buffer, length);
    } else {
        buffer = new char[1];
        buffer[0] = 0;
    }
}


/************************************************************************
 * AwtToolkit methods
 */

AwtToolkit::AwtToolkit() {
    m_localPump = FALSE;
    m_mainThreadId = 0;
    m_toolkitHWnd = NULL;
    m_verbose = FALSE;
    m_verifyComponents = FALSE;
    m_breakOnError = FALSE;

    m_breakMessageLoop = FALSE;
    m_messageLoopResult = 0;

    m_lastMouseOver = NULL;
    m_mouseDown = FALSE;

    m_hGetMessageHook = 0;
    m_timer = 0;

    m_cmdIDs = new AwtCmdIDList();
    m_pModalDialog = NULL;
    m_peer = NULL;
    m_dllHandle = NULL;

    // initialize kb state array
    ::GetKeyboardState(m_lastKeyboardState);

    m_palEntryCount = 0;
    m_palSeed = 0;
    memset(m_palEntries, 0, 256 * sizeof(PALETTEENTRY));

    m_inverseDesktopLUT = NULL;
    m_inverseDesktopLUTSize = 0;
    m_inverseOffscreenLUT = NULL;
    m_inverseOffscreenLUTSize = 0;
    m_grayInverseLUT = NULL;
    m_offscreen = FALSE;

    /* Register this toolkit's helper window */
    VERIFY(RegisterClass() != NULL);

    AwtComponent::Init();

    HDC hDC = ::GetDC(NULL);
    AwtImage::InitializePalette(hDC);
    adjustPalette(NULL);
    ::ReleaseDC(NULL, hDC);
}

AwtToolkit::~AwtToolkit() {
/*
 *  The code has been moved to AwtToolkit::Finalize() method.
 */
}

HWND AwtToolkit::CreateToolkitWnd(const char* name)
{
    HWND hwnd = CreateWindow(
	szAwtToolkitClassName,
	(LPCTSTR)name,	                  /* window name */
	WS_DISABLED,			  /* window style */
	-1, -1,				  /* position of window */
	0, 0,				  /* width and height */
	NULL, NULL, 			  /* hWndParent and hWndMenu */
	::GetModuleHandle(NULL),	  /* hInstance */
	NULL);				  /* lpParam */
    ASSERT(hwnd != NULL);
    return hwnd;
}

BOOL AwtToolkit::Initialize(BOOL localPump) {
    if (m_mainThreadId != 0) {
	/* Already initialized. */
	return FALSE;
    }

	//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
	// Bugs 4032109, 4047966, and 4071991 to fix AWT
	//	crash in 16 color display mode.  16 color mode is supported.  Less
	//	than 16 color is not.
	// creighto@eng.sun.com 1997-10-07
	//
	// Check for at least 16 colors
    HDC hDC = ::GetDC(NULL);
	if ((::GetDeviceCaps(hDC, BITSPIXEL) * ::GetDeviceCaps(hDC, PLANES)) < 4) {
		::MessageBox(NULL,
			     "Sorry, but this release of Java requires at least 16 colors",
			     "AWT Initialization Error",
			     MB_ICONHAND | MB_APPLMODAL);	
		::DeleteDC(hDC);
		JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
		JNU_ThrowByName(env, "java/lang/InternalError",
				"unsupported screen depth");
		return FALSE;
	}
    ::ReleaseDC(NULL, hDC);
	///////////////////////////////////////////////////////////////////////////

    m_localPump = localPump;
    m_mainThreadId = ::GetCurrentThreadId();

    /* 
     * Create the one-and-only toolkit window.  This window isn't
     * displayed, but is used to route messages to this thread.  
     */
    m_toolkitHWnd = CreateToolkitWnd("theAwtToolkitWindow");
    ASSERT(m_toolkitHWnd != NULL);

    /*
     * Setup a GetMessage filter to watch all messages coming out of our 
     * queue from PreProcessMsg().
     */
    m_hGetMessageHook = ::SetWindowsHookEx(WH_GETMESSAGE,
                                           (HOOKPROC)GetMessageFilter,
                                           0, GetCurrentThreadId());


    awt_dnd_initialize();

    return TRUE;
}

void AwtToolkit::Finalize() {
    AWT_TRACE(("In AwtToolkit::Finalize()\n"));

    Cleanup();

    AwtImage::DisposePalette();

    delete m_cmdIDs;

    awt_dnd_uninitialize();
}

ATOM AwtToolkit::RegisterClass() {
    WNDCLASS  wc;

    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC)WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = ::GetModuleHandle(NULL);
    wc.hIcon         = NULL;
    wc.hCursor       = NULL;
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = szAwtToolkitClassName;

    ATOM ret = ::RegisterClass(&wc);
    ASSERT(ret != NULL);
    return ret;
}

void AwtToolkit::UnregisterClass() {
    VERIFY(::UnregisterClass(szAwtToolkitClassName, ::GetModuleHandle(NULL)));
}

/*
 * Structure holding the information to create a component. This packet is 
 * sent to the toolkit window.
 */
struct ComponentCreatePacket {
    void* hComponent;
    void* hParent;
    void (*factory)(void*, void*);
};

/*
 * Create an AwtXxxx component using a given factory function
 * Implemented by sending a message to the toolkit window to invoke the 
 * factory function from that thread
 */
void AwtToolkit::CreateComponent(void* component, void* parent, 
				 ComponentFactory compFactory)
{
    ComponentCreatePacket ccp = { component, parent, compFactory };
    AwtToolkit::GetInstance().SendMessage(WM_AWT_COMPONENT_CREATE, 0,
					  (long)&ccp);
}

/*
 * Destroy an HWND that was created in the toolkit thread. Can be used on 
 * Components and the toolkit window itself.
 */
void AwtToolkit::DestroyComponent(AwtComponent* comp)
{
    AwtToolkit& tk = AwtToolkit::GetInstance();
    if (comp == tk.m_lastMouseOver) {
        tk.m_lastMouseOver = NULL;
    }

    /* Don't filter any post-destroy messages, such as WM_NCDESTROY */
    comp->UnsubclassWindow();

    tk.SendMessage(WM_AWT_DESTROY_WINDOW, (WPARAM)comp->GetHWnd(), 0);
}

/*
 * An AwtToolkit window is just a means of routing toolkit messages to here.
 */
LRESULT CALLBACK AwtToolkit::WndProc(HWND hWnd, UINT message, 
                                     WPARAM wParam, LPARAM lParam)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    /*
     * Awt widget creation messages are routed here so that all
     * widgets are created on the main thread.  Java allows widgets
     * to live beyond their creating thread -- by creating them on
     * the main thread, a widget can always be properly disposed.
     */
    switch (message) {
      case WM_AWT_EXECUTE_SYNC: {
	  AwtObject * 			object = (AwtObject *)wParam;
	  AwtObject::ExecuteArgs *	args = (AwtObject::ExecuteArgs *)lParam;
	  ASSERT(!IsBadReadPtr(object, sizeof(AwtObject)) );
	  ASSERT(!IsBadReadPtr(args, sizeof(AwtObject::ExecuteArgs)) );
	  return object->WinThreadExecProc(args);
      }
      case WM_AWT_COMPONENT_CREATE: {
          ComponentCreatePacket* ccp = (ComponentCreatePacket*)lParam;
          ASSERT(ccp->factory != NULL);
          ASSERT(ccp->hComponent != NULL);
          (*ccp->factory)(ccp->hComponent, ccp->hParent);
          return 0;
      }
      case WM_AWT_DESTROY_WINDOW: {
          /* Destroy widgets from this same thread that created them */
          VERIFY(::DestroyWindow((HWND)wParam) != NULL);
          return 0;
      }
      case WM_AWT_DISPOSE: {
          AwtObject *p = (AwtObject *)wParam;
          delete p;
          return 0;
      }
      case WM_AWT_TOOLKIT_CLEANUP: {
          /* When this window is being destroyed, clean up the system */
          Cleanup();
          return 0;
      }
      case WM_TIMER: {
          // Create an artifical MouseExit message if the mouse left to 
          // a non-java window (bad mouse!)
          POINT pt;
	  AwtToolkit& tk = AwtToolkit::GetInstance();
	  VERIFY(::GetCursorPos(&pt));
	  
	  HWND hWndOver = ::WindowFromPoint(pt);
	  AwtComponent * last_M;
	  if ( AwtComponent::GetComponent(hWndOver) == NULL && tk.m_lastMouseOver != NULL ) {
	      last_M = tk.m_lastMouseOver;
	      // translate point from screen to target window
	      MapWindowPoints(HWND_DESKTOP, last_M->GetHWnd(), &pt, 1);
	      last_M->SendMessage(WM_AWT_MOUSEEXIT,
				  GetAsyncMouseKeyState(), 
				  POINTTOPOINTS(pt));
	      tk.m_lastMouseOver = 0;
	      ::KillTimer(tk.m_toolkitHWnd, tk.m_timer);
	      tk.m_timer = 0;
	  }
          return 0;
      }
      case WM_AWT_POPUPMENU_SHOW: {
          AwtPopupMenu* popup = (AwtPopupMenu*)wParam;
          popup->Show(env, (jobject)lParam);
          return 0;
      }
      case WM_DESTROYCLIPBOARD: {
	  if (!AwtClipboard::IsGettingOwnership())
	      AwtClipboard::LostOwnership((JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2));
          return 0;
      }
      case WM_AWT_LIST_SETMULTISELECT: {
          AwtList* list = (AwtList*)wParam;
          list->SetMultiSelect(lParam);
          return 0;
      }

      // Special awt message to call Imm APIs.
      // ImmXXXX() API must be used in the main thread.
      // In other thread these APIs does not work correctly even if
      // it returs with no error. (This restriction is not documented)
      // So we must use thse messages to call these APIs in main thread.
      case WM_AWT_CREATECONTEXT: {
          return ImmCreateContext();
      }
      case WM_AWT_DESTROYCONTEXT: {
	  ImmDestroyContext((HIMC)wParam);
          return 0;
      }
      case WM_AWT_ASSOCIATECONTEXT: {
          ImmAssociateContext((HWND)wParam, (HIMC)lParam);
          return 0;
      }
      case WM_AWT_ENDCOMPOSITION: {
	  /*right now we just cancel the composition string
	  may need to commit it in the furture*/
          ImmNotifyIME((HIMC)wParam, NI_COMPOSITIONSTR, CPS_CANCEL, 0);
          return 0;
      }
      case WM_AWT_SETCONVERSIONSTATUS: {
	  DWORD cmode;
	  DWORD smode;
	  ImmGetConversionStatus((HIMC)wParam, (LPDWORD)&cmode, (LPDWORD)&smode);
	  ImmSetConversionStatus((HIMC)wParam, (DWORD)LOWORD(lParam), smode);
	  return 0;
      }
      case WM_AWT_GETCONVERSIONSTATUS: {
	  DWORD cmode;
	  DWORD smode;
	  ImmGetConversionStatus((HIMC)wParam, (LPDWORD)&cmode, (LPDWORD)&smode);
	  return cmode;
      }

      /*
       * send this message via ::SendMessage() and the MPT will acquire the
       * HANDLE synchronized with the sender's thread. The HANDLE must be
       * signalled or deadlock may occur between the MPT and the caller.
       */

      case WM_AWT_WAIT_FOR_SINGLE_OBJECT: {
	return ::WaitForSingleObject((HANDLE)lParam, INFINITE);
      }
      case WM_AWT_INVOKE_METHOD: {
	(*(void(*)(void *))wParam)((void *)lParam);

	return 0;
      }

      case WM_AWT_SETOPENSTATUS: {
	  ImmSetOpenStatus((HIMC)wParam, (BOOL)lParam);
	  return 0;
      }
      case WM_AWT_GETOPENSTATUS: {
	  return (DWORD)ImmGetOpenStatus((HIMC)wParam);
      }
    }
    return DefWindowProc(hWnd, message, wParam, lParam);
}

LRESULT CALLBACK AwtToolkit::GetMessageFilter(int code,
					      WPARAM wParam, LPARAM lParam)
{
    if (code >= 0 && wParam == PM_REMOVE && lParam != 0) {
       if (AwtToolkit::GetInstance().PreProcessMsg(*(MSG*)lParam) !=
	       mrPassAlong) {
           /* PreProcessMsg() wants us to eat it */
           ((MSG*)lParam)->message = WM_NULL;  
       }
    }
    return ::CallNextHookEx(AwtToolkit::GetInstance().m_hGetMessageHook, code, 
			    wParam, lParam);
}

/*
 * Cleanup this toolkit & free all the resources allocated. This function may
 * be called from a number of places an in different situations:
 *
 *   1. Normal sysAtExit() callback, probably on our original thread
 *   2. Error exit causing Toolkit destructor to be called, probably 
 *      different thread
 */
void AwtToolkit::Cleanup() {
    AwtToolkit& tk = AwtToolkit::GetInstance();

    /* Only cleanup if it has not already been done */
    if (tk.m_mainThreadId != 0 && tk.m_toolkitHWnd) {
        AWT_TRACE(("In AwtToolkit::Cleanup: mainThreadId:%d "
                   "toolkitHWnd:%X CurrentThreadId:%d\n",
                   tk.m_mainThreadId, 
                   tk.m_toolkitHWnd, ::GetCurrentThreadId()));

	/*
	 * If this is not the correct thread, use SendMessage to switch to the
	 * toolkit thread before continuing. This will fail, however, if the
	 * main thread was terminated due to an abort.
	 */
        if (tk.m_mainThreadId != ::GetCurrentThreadId()) {
            if (tk.SendMessage(WM_AWT_TOOLKIT_CLEANUP) == 0)
                return;  /* cleaned up in message  */
            AWT_TRACE(("  Cleanup failed due to terminated Toolkit thread\n"));
        }
        /* Now perform the cleanup from within the correct thread */
        else {
            AwtObjectList::Cleanup();
            AwtFont::Cleanup();
            AwtComponent::Term();
            AwtGraphics::Cleanup();

            VERIFY(::DestroyWindow(tk.m_toolkitHWnd) != NULL);
            tk.m_toolkitHWnd = 0;

            UnregisterClass();

            ::UnhookWindowsHookEx(tk.m_hGetMessageHook);

            tk.m_mainThreadId = 0;
        }
    }
}

void AwtToolkit::AtExit() {
    AWT_TRACE(("In AwtToolkit::AtExit\n"));
    Cleanup();
}

/*
 * The main message loop
 */

const int AwtToolkit::EXIT_ENCLOSING_LOOP      = 0;
const int AwtToolkit::EXIT_ALL_ENCLOSING_LOOPS = -1;

UINT AwtToolkit::MessageLoop()
{
    m_messageLoopResult = 0;
    while (!m_breakMessageLoop) {            
        ::WaitMessage();               /* allow system to go idle */
        PumpWaitingMessages();         /* pumps any waiting messages */
    }
    if (m_messageLoopResult == EXIT_ALL_ENCLOSING_LOOPS)
	::PostQuitMessage(EXIT_ALL_ENCLOSING_LOOPS);
    m_breakMessageLoop = FALSE;
    return m_messageLoopResult;
}

/*
 * exit the enclosing message loop(s)
 */

void AwtToolkit::QuitMessageLoop(int status) {
    ::PostMessage(this->GetHWnd(), WM_QUIT, status, 0);
}

/*
 * Called by the message loop to pump the message queue when there are
 * messages waiting. Can also be called anywhere to pump messages.
 */
BOOL AwtToolkit::PumpWaitingMessages()
{
    MSG  msg;
    BOOL foundOne = FALSE;

    while (::PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
        foundOne = TRUE;
        if (msg.message == WM_QUIT) {
            m_breakMessageLoop = TRUE;
            m_messageLoopResult = msg.wParam;
            if (m_messageLoopResult == EXIT_ALL_ENCLOSING_LOOPS)
		::PostQuitMessage(msg.wParam);  // make sure all loops exit
            break;
        }
        else if (msg.message != WM_NULL) {
	    /*
	     * The AWT in standalone mode (that is, dynamically loaded from the
	     * Java VM) doesn't have any translation tables to worry about, so
	     * TranslateAccelerator isn't called.
	     */

            ::TranslateMessage(&msg);
            ::DispatchMessage(&msg);
        }
    }
    return foundOne;
}

/*
 * Perform pre-processing on a message before it is translated &
 * dispatched.  Returns true to eat the message
 */
BOOL AwtToolkit::PreProcessMsg(MSG& msg)
{
    /*
     * Offer preprocessing first to the target component, then call out to
     * specific mouse and key preprocessor methods
     */
    AwtComponent* p = AwtComponent::GetComponent(msg.hwnd);
    if (p && p->PreProcessMsg(msg) == mrConsume)
        return TRUE;

    if (msg.message >= WM_MOUSEFIRST && msg.message <= WM_MOUSELAST ||
        msg.message >= WM_NCMOUSEMOVE && msg.message <= WM_NCMBUTTONDBLCLK) {
        if (PreProcessMouseMsg(p, msg))
            return TRUE;
    }
    else if (msg.message >= WM_KEYFIRST && msg.message <= WM_KEYLAST) {
        if (PreProcessKeyMsg(p, msg))
            return TRUE;
    }
    return FALSE;
}

BOOL AwtToolkit::PreProcessMouseMsg(AwtComponent* p, MSG& msg)
{
    WPARAM mouseWParam;
    LPARAM mouseLParam;
    if (msg.message >= WM_MOUSEFIRST && msg.message <= WM_MOUSELAST) {
        mouseWParam = msg.wParam;
        mouseLParam = msg.lParam;
    } else {
        mouseWParam = GetAsyncMouseKeyState();
    }

    /*
     * Get the window under the mouse, as it will be different if its
     * captured.
     */
    DWORD dwCurPos = ::GetMessagePos();
    DWORD dwScreenPos = dwCurPos;
    POINT curPos;
    curPos.x = LOWORD(dwCurPos);
    curPos.y = HIWORD(dwCurPos);
    HWND hWndFromPoint = ::WindowFromPoint(curPos);
    AwtComponent* mouseComp = 
        AwtComponent::GetComponent(hWndFromPoint);
    // If the point under the mouse isn't in the client area,
    // ignore it to maintain compatibility with Solaris (#4095172)
    RECT windowRect;
    ::GetClientRect(hWndFromPoint, &windowRect);
    POINT topLeft;
    topLeft.x = 0;
    topLeft.y = 0;
    ::ClientToScreen(hWndFromPoint, &topLeft);
    windowRect.top += topLeft.y;
    windowRect.bottom += topLeft.y;
    windowRect.left += topLeft.x;
    windowRect.right += topLeft.x;
    if ((curPos.y < windowRect.top) ||
        (curPos.y >= windowRect.bottom) ||
        (curPos.x < windowRect.left) ||
        (curPos.x >= windowRect.right)) {
        mouseComp = NULL;
	hWndFromPoint = NULL;
    }

    /*
     * Look for mouse transitions between windows & create 
     * MouseExit & MouseEnter messages 
     */
    if (mouseComp != m_lastMouseOver) {
	/*
	 * Send the messages right to the windows so that they are in 
	 * the right sequence.
	 */
        if (m_lastMouseOver) {
            dwCurPos = dwScreenPos;
            curPos.x = LOWORD(dwCurPos);
            curPos.y = HIWORD(dwCurPos);
            ::MapWindowPoints(HWND_DESKTOP, m_lastMouseOver->GetHWnd(),
                              &curPos, 1);
            mouseLParam = MAKELPARAM((WORD)curPos.x, (WORD)curPos.y);
            m_lastMouseOver->SendMessage(WM_AWT_MOUSEEXIT, mouseWParam, 
					 mouseLParam);
        }
        if (mouseComp) {
                dwCurPos = dwScreenPos;
                curPos.x = LOWORD(dwCurPos);
                curPos.y = HIWORD(dwCurPos);
                ::MapWindowPoints(HWND_DESKTOP, mouseComp->GetHWnd(),
                                  &curPos, 1);
                mouseLParam = MAKELPARAM((WORD)curPos.x, (WORD)curPos.y);
            mouseComp->SendMessage(WM_AWT_MOUSEENTER, mouseWParam,
				   mouseLParam);
        }
        m_lastMouseOver = mouseComp;
    }

    /*
     * Make sure we get at least one last chance to check for transitions 
     * before we sleep
     */
    if (m_lastMouseOver && !m_timer) {
        m_timer = ::SetTimer(m_toolkitHWnd, IDT_AWT_MOUSECHECK, 200, 0);
    }
    return FALSE;  /* Now go ahead and process current message as usual */
}

BOOL AwtToolkit::PreProcessKeyMsg(AwtComponent* p, MSG& msg)
{
    // get keyboard state for use in AwtToolkit::GetAsyncKeyboardState
    CriticalSection::Lock	l(m_lockKB);
    ::GetKeyboardState(m_lastKeyboardState);
    return FALSE;
}

UINT AwtToolkit::CreateCmdID(AwtObject* object) 
{ 
    return m_cmdIDs->Add(object);
}

void AwtToolkit::RemoveCmdID(UINT id)
{
    m_cmdIDs->Remove(id);
}

AwtObject* AwtToolkit::LookupCmdID(UINT id)
{
    return m_cmdIDs->Lookup(id);
}

HICON AwtToolkit::GetAwtIcon()
{
    return ::LoadIcon(GetModuleHandle(), "AWT_ICON");
}

/* Enable VC++ 4.X heap checking */
#if defined(_DEBUG) && defined(_MSC_VER) && _MSC_VER >= 1000
#include <crtdbg.h>

void AwtToolkit::SetHeapCheck(long flag) {
    int dbgFlag = _CRTDBG_ALLOC_MEM_DF;
    if (flag != 0) {
        dbgFlag |= _CRTDBG_CHECK_ALWAYS_DF;
    }
    _CrtSetDbgFlag(dbgFlag);
}
#else
void AwtToolkit::SetHeapCheck(long flag) {
    if (flag) {
        printf("heap checking not supported with this build\n");
    }
}
#endif

/* monitor state of screen palette to allow generation of inverse LUT for
 * drawing directly to the framebuffer
 */
void AwtToolkit::adjustPalette(HDC aHDC) {
    PALETTEENTRY pe[256];
    int numEntries = 0;
    HWND dtWND = NULL;
    HDC mainDC;
    int prevNumEntries;
    LPPALETTEENTRY palEntries;

    if (aHDC) {
        mainDC = aHDC;
        prevNumEntries = m_palEntryOffscreenCount;
        palEntries = m_palOffscreenEntries;
    } else {
        dtWND = ::GetDesktopWindow();
        mainDC = ::GetDC(dtWND);
        prevNumEntries = m_palEntryCount;
        palEntries = m_palEntries;
    }
    if (!mainDC) {
        return;
    }

    if (8 != ::GetDeviceCaps(mainDC, BITSPIXEL)) {
        return;
    }

    if (!aHDC) {
        numEntries = ::GetSystemPaletteEntries(mainDC, 0, 256, pe);
    } else {
        numEntries = ::GetPaletteEntries(AwtImage::m_hPal, 0, 256, pe);
    }

//    fprintf(stderr,"New Inverse LUT Entries %d %d\n",numEntries, prevNumEntries);
//    fprintf(stderr,"New Inverse LUT memcmp %d\n", memcmp(pe, palEntries, prevNumEntries * sizeof(PALETTEENTRY)));
    if (    (numEntries == prevNumEntries)
            && (0 == memcmp(pe, palEntries, prevNumEntries * sizeof(PALETTEENTRY)))) {
        return;
    }

    CriticalSection::Lock l1(imageLock); // to prevent adjusting palette
                                         // while inverse Lut is in use
                                         // by rendering loops
    CriticalSection::Lock l2(m_paletteLock);
    numEntries = (numEntries > 256)? 256: numEntries;
    memcpy(palEntries, pe, numEntries * sizeof(PALETTEENTRY));
    if (!aHDC) {
        GenerateInverseLUT(pe, numEntries, m_inverseDesktopLUT, m_inverseDesktopLUTSize);
        m_palEntryCount = numEntries;
        m_palSeed++;
    } else {
        GenerateInverseLUT(pe, numEntries, m_inverseOffscreenLUT, m_inverseOffscreenLUTSize);
        m_palEntryOffscreenCount = numEntries;
    }
    if (AwtImage::m_clrdata.grayscale == GS_INDEXGRAY) {
        GenerateGrayInverseLUT();
    }
    if (!aHDC) {
        ::ReleaseDC(dtWND, mainDC);
    }
}

/*
 * state info needed for breadth-first recursion of color cube from
 * initial palette entries within the cube
 */

struct CubeStateInfo {
    unsigned int depth;
    unsigned int maxDepth;

    unsigned char *usedFlags;
    unsigned int  activeEntries;
    unsigned short *rgb;
    unsigned char *indices;
    unsigned char *iLUT;
};

#define INSERTNEW(state, rgb, index) do {                           \
        if (!state.usedFlags[rgb]) {                                \
            state.usedFlags[rgb] = 1;                               \
            state.iLUT[rgb] = index;                                \
            state.rgb[state.activeEntries] = rgb;                   \
            state.indices[state.activeEntries] = index;             \
            state.activeEntries++;                                  \
        }                                                           \
} while (0);
    

#define ACTIVATE(code, mask, delta, state, index) do {              \
    if (((rgb & mask) + delta) <= mask) {                           \
        rgb += delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb -= delta;                                               \
    }                                                               \
    if ((rgb & mask) >= delta) {                                    \
        rgb -= delta;                                               \
        INSERTNEW(state, rgb, index);                               \
        rgb += delta;                                               \
    }                                                               \
} while (0);

/* REMIND: does not deal well with bifurcation which happens when two
 * palette entries map to the same cube vertex
 */

static int
recurseLevel(CubeStateInfo *priorState) {
//    fprintf(stderr,"Level Recurse %d, entries %d\n", priorState->depth, priorState->activeEntries);
    CubeStateInfo currentState;
    memcpy(&currentState, priorState, sizeof(CubeStateInfo));
    currentState.rgb = (unsigned short *)malloc(6
                                                * sizeof(unsigned short)
                                                * priorState->activeEntries);
    if (currentState.rgb == NULL) {
        return 0;
    }
    
    currentState.indices = (unsigned char *)malloc(6
                                                * sizeof(unsigned char)
                                                * priorState->activeEntries);
    if (currentState.indices == NULL) {
        free(currentState.rgb);
        return 0;
    }
    
    currentState.depth++;
    if (currentState.depth > priorState->maxDepth) {
        priorState->maxDepth = currentState.depth;
    }
    currentState.activeEntries = 0;
    for (int i = priorState->activeEntries - 1; i >= 0; i--) {
        unsigned short rgb = priorState->rgb[i];
        unsigned char  index = priorState->indices[i];
        ACTIVATE(rgb, 0x7c00, 0x0400, currentState, index);
        ACTIVATE(rgb, 0x03e0, 0x0020, currentState, index);
        ACTIVATE(rgb, 0x001f, 0x0001, currentState, index);
    }
    if (currentState.activeEntries) {
        if (!recurseLevel(&currentState)) {
	    free(currentState.rgb);
	    free(currentState.indices);
            return 0;
        }
    }
    
    if (currentState.maxDepth > priorState->maxDepth) {
        priorState->maxDepth = currentState.maxDepth;
    }
    
    free(currentState.rgb);
    free(currentState.indices);
    return  1;
}

/*
 * REMIND: take core inversedLUT calculation to the shared tree and
 * recode the functions (Win32)awt_Image:initCubemap(),
 * (Win32)awt_Image:make_cubemap(), (Win32)AwtToolkit::GenerateInverseLUT(),
 * (Solaris)color:initCubemap() to call the shared codes.
 */
void AwtToolkit::GenerateInverseLUT(
        LPPALETTEENTRY aPalEntries, int aNumPalEntries,
        unsigned char *& aInverseLUT, int& aLutSize)
{
//    fprintf(stderr,"Generating Inverse LUT\n");
    aLutSize = 32*32*32;         // 2^5 * 2^5 * 2^5
    unsigned char *useFlags;
    unsigned char *newILut = new unsigned char[aLutSize];
    if (newILut) {
        unsigned char *tempLUT = aInverseLUT;
        
        useFlags = (unsigned char *)calloc(aLutSize, 1);
        if (useFlags == 0) {
            delete newILut;
#ifdef DEBUG
	fprintf(stderr, "Out of memory in AwtToolkit::GenerateInverseLUT1\n");
#endif
            return;
        }
      
        // use existing palette as seed values
        CubeStateInfo currentState;
        currentState.depth          = 0;
        currentState.maxDepth       = 0;
        currentState.usedFlags      = useFlags;
        currentState.activeEntries  = 0;
        currentState.iLUT           = newILut;
        currentState.rgb = (unsigned short *)
                                malloc(256 * sizeof(unsigned short));
        
        if (currentState.rgb == NULL) {
            delete newILut;
	    free(useFlags);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in AwtToolkit::GenerateInverseLUT2\n");
#endif
            return;
        }

        currentState.indices = (unsigned char *)
                                malloc(256 * sizeof(unsigned char));
        
        if (currentState.indices == NULL) {
            delete newILut;
	    free(useFlags);
	    free(currentState.rgb);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in AwtToolkit::GenerateInverseLUT3\n");
#endif
            return;
        }

        for (int i = 0; i < 128; i++) {
            unsigned short rgb;
            rgb = ((aPalEntries[i].peRed & 0xf8) << 7)
                    | ((aPalEntries[i].peGreen & 0xf8) << 2)
                    | (aPalEntries[i].peBlue >> 3);
            INSERTNEW(currentState, rgb, i);
            rgb = ((aPalEntries[255-i].peRed & 0xf8) << 7)
                    | ((aPalEntries[255-i].peGreen & 0xf8) << 2)
                    | (aPalEntries[255-i].peBlue >> 3);
            INSERTNEW(currentState, rgb, 255-i);
        }

        if (!recurseLevel(&currentState)) {
            delete newILut;
            free(useFlags);
            free(currentState.rgb);
	    free(currentState.indices);
#ifdef DEBUG
	fprintf(stderr, "Out of memory in AwtToolkit::GenerateInverseLUT4\n");
#endif
	    return;
	}

	free(useFlags);
	free(currentState.rgb);
	free(currentState.indices);

        aInverseLUT = newILut;  // atomic insertion operation
        if (tempLUT) {
            delete [] tempLUT;
        }
//        fprintf(stderr,"Inverse LUT maxdepth %d\n",currentState.maxDepth);
        return;
    }
#ifdef DEBUG
	fprintf(stderr, "Out of memory in AwtToolkit::GenerateInverseLUT5\n");
#endif
}

void AwtToolkit::GenerateGrayInverseLUT(void) {
    int *newILUT = new int[256];
    if (newILUT) {
        int *tempILUT = m_grayInverseLUT;
        int grayvals[256];
        int grayindices[256];
        int gcount;
        int mindist, besti, d;
        PALETTEENTRY *pPalEntries;

        /* REMIND: should a more sophisticated algorithm be used to
           map inverse grays, e.g. best luminance match?
         */
        pPalEntries = m_palEntries;
        gcount = 0;
        for (int i = 0; i < m_palEntryCount; i++, pPalEntries++) {
            if ((pPalEntries->peRed == pPalEntries->peGreen) &&
                (pPalEntries->peGreen == pPalEntries->peBlue)) {
                grayvals[gcount] = (unsigned char) pPalEntries->peRed;
                grayindices[gcount++] = i;
            }
        }

        for (int g = 0; g < 256; g++) {
            int *pg = grayvals;
            mindist = 256;
            besti = 0;
            for (int i = 0 ; i < gcount; i++, pg++) {
                d = *pg - g;
                if (d < 0) d = -d;
                if (d < mindist) {
                    besti = i;
                    if (d == 0) {
                        break;
                    }
                    mindist = d;
                }
            }
            newILUT[g] = grayindices[besti];
        }
        m_grayInverseLUT = newILUT;  // atomic insertion operation
        if (tempILUT) {
            delete [] tempILUT;
        }
    }
}

/************************************************************************
 * Toolkit native methods
 */

extern "C" {
  
/*
 * Class:     java_awt_Toolkit
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL 
Java_java_awt_Toolkit_initIDs(JNIEnv *env, jclass cls) {
    
    AwtToolkit::getDefaultToolkitMID = 
	env->GetStaticMethodID(cls,"getDefaultToolkit","()Ljava/awt/Toolkit;");
    AwtToolkit::getFontMetricsMID =
	env->GetMethodID(cls, "getFontMetrics", 
			 "(Ljava/awt/Font;)Ljava/awt/FontMetrics;");
    
    ASSERT(AwtToolkit::getDefaultToolkitMID != NULL);
    ASSERT(AwtToolkit::getFontMetricsMID != NULL);
}

} /* extern "C" */


/************************************************************************
 * WToolkit native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    init
 * Signature: (Ljava/lang/Thread;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WToolkit_init(JNIEnv *env, jobject self, jobject thread)
{
    /* Debug settings */

    jboolean dbgTrace = (env)->GetBooleanField(self,AwtToolkit::dbgTraceID);
    jboolean dbgVerify = (env)->GetBooleanField(self,AwtToolkit::dbgVerifyID);
    jboolean dbgBreak = (env)->GetBooleanField(self,AwtToolkit::dbgBreakID);
    jboolean dbgHeapCheck =
      (env)->GetBooleanField(self,AwtToolkit::dbgHeapCheckID);

    AwtToolkit::GetInstance().SetVerbose(dbgTrace);
    AwtToolkit::GetInstance().SetVerify(dbgVerify);
    AwtToolkit::GetInstance().SetBreak(dbgBreak);
    AwtToolkit::GetInstance().SetHeapCheck(dbgHeapCheck);

    AwtToolkit::GetInstance().SetPeer(self);
    VERIFY(AwtToolkit::GetInstance().Initialize(TRUE));
}


/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WToolkit_initIDs(JNIEnv *env, jclass cls)
{
    AwtToolkit::dbgTraceID = (env)->GetFieldID(cls,"dbgTrace","Z");
    AwtToolkit::dbgVerifyID = (env)->GetFieldID(cls,"dbgVerify","Z");
    AwtToolkit::dbgBreakID = (env)->GetFieldID(cls,"dbgBreak","Z");
    AwtToolkit::dbgHeapCheckID = (env)->GetFieldID(cls,"dbgHeapCheck","Z");

    ASSERT(AwtToolkit::dbgTraceID != NULL);
    ASSERT(AwtToolkit::dbgVerifyID != NULL);
    ASSERT(AwtToolkit::dbgBreakID != NULL);
    ASSERT(AwtToolkit::dbgHeapCheckID != NULL);
}


/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    eventLoop
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WToolkit_eventLoop(JNIEnv *env, jobject self)
{
    /* Add our AtExit handler, now that Java is initialized. */

    //im --- stop to regist exit procedure for awt bug(?) workaround
    // AwtToolkit::AtExit() cause exception on exitting JavaVM.
    // When AwtToolkit::AtExit() is called, awt.dll is already 
    // unloaded.
//    sysAtexit(AwtToolkit::AtExit);

    if (!AwtToolkit::GetInstance().localPump()) {
	return;
    }

    OleInitialize((LPVOID)NULL); // for DnD

    AWT_TRACE(("AWT event loop started\n"));
    AwtToolkit::GetInstance().MessageLoop();
    AWT_TRACE(("AWT event loop ended\n"));

    OleUninitialize(); // for DnD

    AwtToolkit::GetInstance().Finalize();

    /*
     * IMPORTANT NOTES:
     *   The AwtToolkit has been destructed by now.
     * DO NOT CALL any method of AwtToolkit!!!
     */
}

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    makeColorModel
 * Signature: ()Ljava/awt/image/ColorModel;
 */
JNIEXPORT jobject JNICALL 
Java_sun_awt_windows_WToolkit_makeColorModel(JNIEnv *env, jclass cls) 
{
    return AwtImage::getColorModel(env);
}

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    getScreenResolution
 * Signature: ()I
 */
JNIEXPORT jint JNICALL 
Java_sun_awt_windows_WToolkit_getScreenResolution(JNIEnv *env, jobject self) 
{
    HWND hWnd = ::GetDesktopWindow();
    HDC hDC = ::GetDC(hWnd);
    long result = ::GetDeviceCaps(hDC, LOGPIXELSX);
    ::ReleaseDC(hWnd, hDC);
    return result;
}

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    getScreenWidth
 * Signature: ()I
 */
JNIEXPORT jint JNICALL 
Java_sun_awt_windows_WToolkit_getScreenWidth(JNIEnv *env, jobject self)
{
    return ::GetSystemMetrics(SM_CXSCREEN);
}

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    getScreenHeight
 * Signature: ()I
 */
JNIEXPORT jint JNICALL
Java_sun_awt_windows_WToolkit_getScreenHeight(JNIEnv *env, jobject self)
{
    return ::GetSystemMetrics(SM_CYSCREEN);
}

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    sync
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WToolkit_sync(JNIEnv *env, jobject self)
{
    VERIFY(::GdiFlush());
}

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    beep
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WToolkit_beep(JNIEnv *env, jobject self)
{
    VERIFY(::MessageBeep(MB_OK));
}

/*
 * Class:     sun_awt_windows_WToolkit
 * Method:    loadSystemColors
 * Signature: ([I)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WToolkit_loadSystemColors(JNIEnv *env, jobject self,
					       jintArray colors)
{
    static int indexMap[] = {
        COLOR_DESKTOP, /* DESKTOP */
        COLOR_ACTIVECAPTION, /* ACTIVE_CAPTION */
        COLOR_CAPTIONTEXT, /* ACTIVE_CAPTION_TEXT */
        COLOR_ACTIVEBORDER, /* ACTIVE_CAPTION_BORDER */
        COLOR_INACTIVECAPTION, /* INACTIVE_CAPTION */
        COLOR_INACTIVECAPTIONTEXT, /* INACTIVE_CAPTION_TEXT */
        COLOR_INACTIVEBORDER, /* INACTIVE_CAPTION_BORDER */
        COLOR_WINDOW, /* WINDOW */
        COLOR_WINDOWFRAME, /* WINDOW_BORDER */
        COLOR_WINDOWTEXT, /* WINDOW_TEXT */
        COLOR_MENU, /* MENU */
        COLOR_MENUTEXT, /* MENU_TEXT */
        COLOR_3DFACE, /* TEXT */
        COLOR_BTNTEXT, /* TEXT_TEXT */
        COLOR_HIGHLIGHT, /* TEXT_HIGHLIGHT */
        COLOR_HIGHLIGHTTEXT, /* TEXT_HIGHLIGHT_TEXT */
        COLOR_GRAYTEXT, /* TEXT_INACTIVE_TEXT */
        COLOR_3DFACE, /* CONTROL */
        COLOR_BTNTEXT, /* CONTROL_TEXT */
        COLOR_3DLIGHT, /* CONTROL_HIGHLIGHT */
        COLOR_3DHILIGHT, /* CONTROL_LT_HIGHLIGHT */
        COLOR_3DSHADOW, /* CONTROL_SHADOW */
        COLOR_3DDKSHADOW, /* CONTROL_DK_SHADOW */
        COLOR_SCROLLBAR, /* SCROLLBAR */
        COLOR_INFOBK, /* INFO */
        COLOR_INFOTEXT, /* INFO_TEXT */
    };

    jint colorLen = env->GetArrayLength(colors);
    jint* colorsPtr = (jint *)env->GetPrimitiveArrayCritical(colors, 0);
    for (int i = 0; i < sizeof indexMap && i < colorLen; i++) {
	colorsPtr[i] = DesktopColor2RGB(indexMap[i]);
    }
    env->ReleasePrimitiveArrayCritical(colors, colorsPtr, 0);
}

} /* extern "C" */


/* Convert a Windows desktop color index into an RGB value. */
COLORREF DesktopColor2RGB(int colorIndex) {
    long sysColor = ::GetSysColor(colorIndex);
    return ((GetRValue(sysColor)<<16) | (GetGValue(sysColor)<<8) | 
	    (GetBValue(sysColor)) | 0xff000000);
}

/************************************************************************
 * Diagnostic routines
 */
#if defined(DEBUG)

void CallDebugger(char* expr, char* file, unsigned line) {
    DWORD lastError = GetLastError();

    if (AwtToolkit::GetInstance().BreakOnError() == FALSE) {
        printf("Fatal error: %d\n", lastError);
        _assert(expr, file, line);
        return;
    }

    BOOL ignoreError = FALSE;
# if defined(_M_IX86)
    _asm { int 3 };
# else
    DebugBreak();
# endif
    /* Modify ignoreError in the debugger to continue */
    if (!ignoreError) {
        _assert(expr, file, line);
    }
}

#endif /* DEBUG */
