/*
 * @(#)awt_Window.cpp	1.59 98/09/15
 *
 * Copyright 1996-1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

#include "awt_Component.h"
#include "awt_Container.h"
#include "awt_Frame.h"
#include "awt_Graphics.h"
#include "awt_Insets.h"
#include "awt_Panel.h"
#include "awt_Toolkit.h"
#include "awt_Window.h"

#include "java_awt_Insets.h"
#include <java_awt_Container.h>
#include <java_awt_event_ComponentEvent.h>

/* IMPORTANT! Read the README.JNI file for notes on JNI converted AWT code.
 */

/************************************************************************
 * AwtWindow fields
 */

jfieldID AwtWindow::warningStringID;

/************************************************************************
 * AwtWindow class methods
 */

AwtWindow::AwtWindow() {
    m_iconic = FALSE;
    m_eraseBackground = FALSE;
    m_resizing = FALSE;
    m_sizePt.x = m_sizePt.y = 0;
    VERIFY(::SetRectEmpty(&m_insets));
    VERIFY(::SetRectEmpty(&m_old_insets));
    VERIFY(::SetRectEmpty(&m_warningRect));
    SetCursor(::LoadCursor(NULL, IDC_ARROW));
}

AwtWindow::~AwtWindow() {
}

const char* AwtWindow::GetClassName() {
  return "SunAwtWindow";
}

/* Create a new AwtWindow object and window.   */
AwtWindow* AwtWindow::Create(jobject self, jobject parent)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return NULL;
    }

    jint pData;
    AwtWindow* awtParent;
    jobject target = 0;
    AwtWindow* window;
    JNI_CHECK_PEER_GOTO(parent, done);

    awtParent = (AwtWindow*)pData;
    JNI_CHECK_NULL_GOTO(awtParent, "null awtParent", done);

    target = env->GetObjectField(self, AwtObject::targetID);
    JNI_CHECK_NULL_GOTO(target, "null target", done);

    window = new AwtWindow();
    JNI_CHECK_NULL_GOTO(window, "AwtWindow() failed", done);

    {
        DWORD style = WS_CLIPCHILDREN | WS_POPUP;
	window->CreateHWnd(env, "",
			   style, 0,
			   0, 0, 0, 0,
			   awtParent->GetHWnd(),
			   NULL,
			   ::GetSysColor(COLOR_WINDOWTEXT),
			   ::GetSysColor(COLOR_WINDOW),
			   self);

	jint x = env->GetIntField(target, AwtComponent::xID);
	jint y = env->GetIntField(target, AwtComponent::yID);
	jint width = env->GetIntField(target, AwtComponent::widthID);
	jint height = env->GetIntField(target, AwtComponent::heightID);

	/* 
	 * Reshape here instead of during create, so that a WM_NCCALCSIZE
	 * is sent. 
	 */
	window->Reshape(x, y, width, height);
    }
 done:
    env->DeleteLocalRef(target);
    return window;
}

//
// Override AwtComponent's Reshape to handle minimized/maximized
// windows, fixes 4065534 (robi.khan@eng).
// NOTE: See also AwtWindow::WmSize
//
void AwtWindow::Reshape(int x, int y, int width, int height) 
{
    HWND    hWndSelf = GetHWnd();

    if ( IsIconic(hWndSelf)) {
    // normal AwtComponent::Reshape will not work for iconified windows so...
	WINDOWPLACEMENT wp;
	POINT	    ptMinPosition = {x,y};
	POINT	    ptMaxPosition = {0,0};
	RECT	    rcNormalPosition = {x,y,x+width,y+height};
	RECT	    rcWorkspace;
	HWND	    hWndDesktop = GetDesktopWindow();

	// SetWindowPlacement takes workspace coordinates, but
	// if taskbar is at top of screen, workspace coords !=
	// screen coords, so offset by workspace origin
	VERIFY(::SystemParametersInfo(SPI_GETWORKAREA, 0, (PVOID)&rcWorkspace, 0));
	::OffsetRect(&rcNormalPosition, -rcWorkspace.left, -rcWorkspace.top);

	// set the window size for when it is not-iconified
	wp.length = sizeof(wp);
	wp.flags = WPF_SETMINPOSITION;
	wp.showCmd = SW_SHOWNA;
	wp.ptMinPosition = ptMinPosition;
	wp.ptMaxPosition = ptMaxPosition;
	wp.rcNormalPosition = rcNormalPosition;
	SetWindowPlacement(hWndSelf, &wp);
	return;
    }
    
    if (IsZoomed(hWndSelf)) {
    // setting size of maximized window, we remove the
    // maximized state bit (matches Motif behaviour)
    // (calling ShowWindow(SW_RESTORE) would fire an
    //  activation event which we don't want)
	LONG	style = GetStyle();
	ASSERT(style & WS_MAXIMIZE);
	style ^= WS_MAXIMIZE;
	SetStyle(style);
    }

    AwtComponent::Reshape(x, y, width, height);
}

/*
 * Get and return the insets for this window (container, really).
 * Calculate & cache them while we're at it, for use by AwtGraphics
 */
BOOL AwtWindow::UpdateInsets(jobject insets)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    ASSERT(GetPeer(env) != NULL);
    if (env->EnsureLocalCapacity(2) < 0) {
	return FALSE;
    }

    // fix 4167248 : don't update insets when frame is iconified
    // to avoid bizarre window/client rectangles
    if (m_iconic)
	return FALSE;

    /*
     * Code to calculate insets. Stores results in frame's data
     * members, and in the peer's Inset object.
     */
    RECT outside;
    RECT inside;
    ::GetWindowRect(GetHWnd(), &outside);
    ::GetClientRect(GetHWnd(), &inside);

    /* Update our inset member */
    if (outside.right - outside.left > 0 && outside.bottom - outside.top > 0) {
        ::MapWindowPoints(GetHWnd(), 0, (LPPOINT)&inside, 2);
        m_insets.top = inside.top - outside.top;
        m_insets.bottom = outside.bottom - inside.bottom;
        m_insets.left = inside.left - outside.left;
        m_insets.right = outside.right - inside.right;
    } else {
	m_insets.top = -1;
    }
    if (m_insets.left < 0 || m_insets.top < 0 ||
	m_insets.right < 0 || m_insets.bottom < 0)
    {
	/* This window hasn't been sized yet -- use system metrics. */
        jobject target = GetTarget(env);
        if (JNU_IsInstanceOfByName(env, target, "java/awt/Frame") > 0 ||
            JNU_IsInstanceOfByName(env, target, "java/awt/Dialog") > 0) {
            /* Get outer frame sizes. */
            long style = GetStyle();
            if (style & WS_THICKFRAME) {
                m_insets.left = m_insets.right = 
                    ::GetSystemMetrics(SM_CXSIZEFRAME);
                m_insets.top = m_insets.bottom = 
                    ::GetSystemMetrics(SM_CYSIZEFRAME);
            } else {
                m_insets.left = m_insets.right = 
                    ::GetSystemMetrics(SM_CXDLGFRAME);
                m_insets.top = m_insets.bottom = 
                    ::GetSystemMetrics(SM_CYDLGFRAME);
            }

            /* Add in title. */
            m_insets.top += ::GetSystemMetrics(SM_CYCAPTION);

            /* Add in menuBar, if any. */
            if (JNU_IsInstanceOfByName(env, target, "java/awt/Frame") > 0 &&
                ((AwtFrame*)this)->GetMenuBar()) {
                m_insets.top += ::GetSystemMetrics(SM_CYMENU);
            }
        }
	env->DeleteLocalRef(target);
    }

    jobject peer = GetPeer(env);
    /* Get insets into our peer directly */
    jobject peerInsets = (env)->GetObjectField(peer, AwtPanel::insets_ID);
    ASSERT(!env->ExceptionOccurred());
    if (peerInsets != NULL) { // may have been called during creation
	(env)->SetIntField(peerInsets, AwtInsets::topID, m_insets.top);
	(env)->SetIntField(peerInsets, AwtInsets::bottomID, 
			   m_insets.bottom);
	(env)->SetIntField(peerInsets, AwtInsets::leftID, m_insets.left);
	(env)->SetIntField(peerInsets, AwtInsets::rightID, m_insets.right);
    }
    /* Get insets into the Inset object (if any) that was passed */
    if (insets != NULL) {
	(env)->SetIntField(insets, AwtInsets::topID, m_insets.top);
	(env)->SetIntField(insets, AwtInsets::bottomID, m_insets.bottom);
	(env)->SetIntField(insets, AwtInsets::leftID, m_insets.left);
	(env)->SetIntField(insets, AwtInsets::rightID, m_insets.right);
    }
    env->DeleteLocalRef(peerInsets);

    BOOL	insetsChanged = !::EqualRect( &m_old_insets, &m_insets );	
    ::CopyRect( &m_old_insets, &m_insets );
    return insetsChanged;
}


/*
 * Although this function sends ComponentEvents, it needs to be defined
 * here because only top-level windows need to have move and resize
 * events fired from native code.  All contained windows have these events
 * fired from common Java code.
 */
void AwtWindow::SendComponentEvent(long eventId)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    static jclass classEvent = NULL;
    if (classEvent == NULL) {
        if (env->PushLocalFrame(1) < 0)
	    return;
	classEvent = env->FindClass("java/awt/event/ComponentEvent");
	if (classEvent != NULL) {
	    classEvent = (jclass)env->NewGlobalRef(classEvent);
	}
	env->PopLocalFrame(0);
    }
    static jmethodID eventInitMID = NULL;
    if (eventInitMID == NULL) {
	eventInitMID = env->GetMethodID(classEvent, "<init>",
					"(Ljava/awt/Component;I)V");
	if (eventInitMID == NULL) {
	    return;
	}
    }
    if (env->EnsureLocalCapacity(2) < 0) {
	return;
    }
    jobject target = GetTarget(env);
    jobject event = env->NewObject(classEvent, eventInitMID, 
				   target, eventId);
    ASSERT(!env->ExceptionOccurred());
    ASSERT(event != NULL);
    SendEvent(event);

    env->DeleteLocalRef(target);
    env->DeleteLocalRef(event);
}

void AwtWindow::SendWindowEvent(long eventId)
{
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);

    static jclass wClassEvent = NULL;

    if (wClassEvent == NULL) {
        if (env->PushLocalFrame(1) < 0)
	    return;
	wClassEvent = env->FindClass("java/awt/event/WindowEvent");
	if (wClassEvent != NULL) {
	    wClassEvent = (jclass)env->NewGlobalRef(wClassEvent);
	}
	env->PopLocalFrame(0);
        if (wClassEvent == NULL) {
            return;
        }
    }
    static jmethodID wEventInitMID = NULL;
    if (wEventInitMID == NULL) {
	wEventInitMID = env->GetMethodID(wClassEvent, "<init>",
					 "(Ljava/awt/Window;I)V");
        ASSERT(wEventInitMID);
	if (wEventInitMID == NULL) {
	    return;
	}
    }
    if (env->EnsureLocalCapacity(2) < 0)
        return;
    jobject target = GetTarget(env);
    jobject event = env->NewObject(wClassEvent, wEventInitMID, 
				   target, eventId);
    ASSERT(!env->ExceptionOccurred());
    ASSERT(event != NULL);
    SendEvent(event);

    env->DeleteLocalRef(target);
    env->DeleteLocalRef(event);
}

MsgRouting AwtWindow::WmCreate()
{
    return mrDoDefault;  
}

MsgRouting AwtWindow::WmClose()
{
    SendWindowEvent(java_awt_event_WindowEvent_WINDOW_CLOSING);

    /* Rely on above notification to handle quitting as needed */
    return mrConsume;  
}

MsgRouting AwtWindow::WmDestroy()
{
    SendWindowEvent(java_awt_event_WindowEvent_WINDOW_CLOSED);
    return AwtComponent::WmDestroy();
}

/*
 * Override AwtComponent's move handling to first update the
 * java AWT target's position fields directly, since Windows
 * and below can be resized from outside of java (by user)
 */
MsgRouting AwtWindow::WmMove(int x, int y)
{
    if ( IsIconic(GetHWnd()) ) {
    // fixes 4065534 (robi.khan@eng)
    // if a window is iconified we don't want to update
    // it's target's position since minimized Win32 windows
    // move to -32000, -32000 for whatever reason
    // NOTE: See also AwtWindow::Reshape
	return mrDoDefault;
    }

    /* Update the java AWT target component's fields directly */
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(1) < 0) {
	return mrConsume;
    }
    jobject peer = GetPeer(env);
    jobject target = env->GetObjectField(peer, AwtObject::targetID);

    RECT rect;
    ::GetWindowRect(GetHWnd(), &rect);

    jint targX = (env)->GetIntField(target, AwtComponent::xID);
    jint targY = (env)->GetIntField(target, AwtComponent::yID);
    if (targX != rect.left || targY != rect.top) {
	(env)->SetIntField(target, AwtComponent::xID, rect.left);
	(env)->SetIntField(target, AwtComponent::yID, rect.top);
        SendComponentEvent(java_awt_event_ComponentEvent_COMPONENT_MOVED);
    }

    env->DeleteLocalRef(target);
    return AwtComponent::WmMove(x, y);
}

/*
 * Override AwtComponent's size handling to first update the 
 * java AWT target's dimension fields directly, since Windows
 * and below can be resized from outside of java (by user)
 */
MsgRouting AwtWindow::WmSize(int type, int w, int h)
{
    if (type == SIZE_MINIMIZED) {
        SendWindowEvent(java_awt_event_WindowEvent_WINDOW_ICONIFIED);
        m_iconic = TRUE;
    }
    else if (type == SIZE_RESTORED && m_iconic) {
        SendWindowEvent(java_awt_event_WindowEvent_WINDOW_DEICONIFIED);
        m_iconic = FALSE;
    } 
    else if (type == SIZE_MAXIMIZED && m_iconic) {
		SendWindowEvent(java_awt_event_WindowEvent_WINDOW_DEICONIFIED);
		m_iconic = FALSE;
	}
	else {
	JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
	if (env->EnsureLocalCapacity(1) < 0)
	    return mrDoDefault;
	jobject target = GetTarget(env);
	// fix 4167248 : ensure the insets are up-to-date before using
	BOOL insetsChanged = UpdateInsets(NULL);
	jint width = (env)->GetIntField(target, AwtComponent::widthID);
	jint height = (env)->GetIntField(target, AwtComponent::heightID);
        int newWidth = w + m_insets.left + m_insets.right;
        int newHeight = h + m_insets.top + m_insets.bottom;

        if (width != newWidth || height != newHeight || insetsChanged) {
	    (env)->SetIntField(target, AwtComponent::widthID, newWidth);
	    (env)->SetIntField(target, AwtComponent::heightID, newHeight);
            if (!m_resizing) {
                SendComponentEvent(java_awt_event_ComponentEvent_COMPONENT_RESIZED);
            }
        }
	env->DeleteLocalRef(target);
        return AwtComponent::WmSize(type, w, h);
    }

    return mrDoDefault;
}

MsgRouting AwtWindow::WmPaint(HDC)
{
    /* Get the rectangle that covers all update regions, if any exist. */
    RECT r;
    if (::GetUpdateRect(GetHWnd(), &r, FALSE)) {
	/*
	 * Now remove all update regions from this window -- do it
	 * here instead of after the Java upcall, in case any new
	 * updating is requested.  
	 */
        ::ValidateRgn(GetHWnd(), NULL);

	::OffsetRect(&r, m_insets.left, m_insets.top);

        if ((r.right-r.left) > 0 && (r.bottom-r.top) > 0 &&
            m_peerObject != NULL && m_callbacksEnabled) {
            DoCallback("handleExpose", // BUGID 4070488 handlerepaint was a no-op
                       "(IIII)V", r.left, r.top, r.right-r.left, r.bottom-r.top);
        }
    }
    m_eraseBackground = FALSE;

    return mrConsume;
}

MsgRouting AwtWindow::WmSysCommand(UINT uCmdType, UINT xPos, UINT yPos)
{
    if (uCmdType == SC_SIZE) {
        m_resizing = TRUE;
    }
    return mrDoDefault;
}

MsgRouting AwtWindow::WmExitSizeMove()
{
    if (m_resizing) {
        SendComponentEvent(java_awt_event_ComponentEvent_COMPONENT_RESIZED);
        m_resizing = FALSE;
    }
    return mrDoDefault;
}

MsgRouting AwtWindow::WmSettingChange(WORD wFlag, LPCTSTR pszSection)
{
    if (wFlag == SPI_SETNONCLIENTMETRICS) {
    // user changed window metrics in
    // Control Panel->Display->Appearance
    // which may cause window insets to change
	UpdateInsets(NULL);
	return mrConsume;
    }
    return mrDoDefault;
}

MsgRouting AwtWindow::WmNcCalcSize(BOOL fCalcValidRects, 
                                   LPNCCALCSIZE_PARAMS lpncsp, UINT& retVal)
{
    MsgRouting mrRetVal = mrDoDefault;

    if (fCalcValidRects == FALSE) {
        return mrDoDefault;
    }
    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0) {
	return mrConsume;
    }
    jobject target = GetTarget(env);
    jstring warningString = 
      (jstring)(env)->GetObjectField(target, AwtWindow::warningStringID);
    if (warningString != NULL) {
	RECT r;
	::CopyRect(&r, &lpncsp->rgrc[0]);
	retVal = DefWindowProc(WM_NCCALCSIZE, fCalcValidRects, (LPARAM)lpncsp);

	/* Adjust non-client area for warning banner space. */
	m_warningRect.left = lpncsp->rgrc[0].left;
	m_warningRect.right = lpncsp->rgrc[0].right;
	m_warningRect.bottom = lpncsp->rgrc[0].bottom;
	m_warningRect.top = 
            m_warningRect.bottom - ::GetSystemMetrics(SM_CYCAPTION);
        if (GetStyle() & WS_THICKFRAME) {
	    m_warningRect.top -= 2;
        } else {
	    m_warningRect.top += 2;
        }
	lpncsp->rgrc[0].bottom = m_warningRect.top;

	/* Convert to window-relative coordinates. */
	::OffsetRect(&m_warningRect, -r.left, -r.top);

        /* Notify target of Insets change. */
        UpdateInsets(NULL);

	mrRetVal = mrConsume;
    } else {
	// WM_NCCALCSIZE is usually in response to a resize, but
	// also can be triggered by SetWindowPos(SWP_FRAMECHANGED),
	// which means the insets will have changed - rnk 4/7/1998
	retVal = DefWindowProc(WM_NCCALCSIZE, fCalcValidRects, (LPARAM)lpncsp);
	UpdateInsets(NULL);
	mrRetVal = mrConsume;
    }
    env->DeleteLocalRef(target);
    env->DeleteLocalRef(warningString);
    return mrRetVal;
}

MsgRouting AwtWindow::WmNcPaint(HRGN hrgn)
{
    DefWindowProc(WM_NCPAINT, (WPARAM)hrgn, 0);

    JNIEnv *env = (JNIEnv *)JNU_GetEnv(jvm, JNI_VERSION_1_2);
    if (env->EnsureLocalCapacity(2) < 0) {
	return mrConsume;
    }
    jobject target = GetTarget(env);
    jstring warningString =
      (jstring)(env)->GetObjectField(target, AwtWindow::warningStringID);
    if (warningString != NULL) {
	RECT r;
        ::CopyRect(&r, &m_warningRect);
	HDC hDC = ::GetWindowDC(GetHWnd());
	ASSERT(hDC);
	int iSaveDC = ::SaveDC(hDC);
	VERIFY(::SelectClipRgn(hDC, NULL) != NULL);
	VERIFY(::FillRect(hDC, &m_warningRect, ::GetStockObject(BLACK_BRUSH)));

        if (GetStyle() & WS_THICKFRAME) {
            /* draw edge */
            VERIFY(::DrawEdge(hDC, &r, EDGE_RAISED, BF_TOP));
            r.top += 2;
            VERIFY(::DrawEdge(hDC, &r, EDGE_SUNKEN, BF_RECT));
            ::InflateRect(&r, -2, -2);
        }

	/* draw warning text */
	LPWSTR text = TO_WSTRING(warningString);
	VERIFY(::SetBkColor(hDC, RGB(255, 225, 0)) != CLR_INVALID);
	VERIFY(::SetTextColor(hDC, RGB(0, 0, 0)) != CLR_INVALID);
	VERIFY(::SelectObject(hDC, ::GetStockObject(DEFAULT_GUI_FONT)) != NULL);
	VERIFY(::SetTextAlign(hDC, TA_LEFT | TA_BOTTOM) != GDI_ERROR);
        VERIFY(::ExtTextOutW(hDC, r.left, r.bottom, ETO_CLIPPED | ETO_OPAQUE, 
	                     &r, text, wcslen(text), NULL));
	VERIFY(::RestoreDC(hDC, iSaveDC));
	::ReleaseDC(GetHWnd(), hDC);
    }

    env->DeleteLocalRef(target);
    env->DeleteLocalRef(warningString);
    return mrConsume;
}

MsgRouting AwtWindow::WmNcHitTest(UINT x, UINT y, UINT& retVal)
{
    POINT pt = { x, y };
    if (::PtInRect(&m_warningRect, pt)) {
        retVal = HTNOWHERE;
    } else {
        retVal = DefWindowProc(WM_NCHITTEST, 0, MAKELPARAM(x, y));
    }
    return mrConsume;
}

BOOL CALLBACK InvalidateChildRect(HWND hWnd, LPARAM)
{
    ::InvalidateRect(hWnd, NULL, TRUE);
    return TRUE;
}

void AwtWindow::Invalidate(RECT* r)
{
    ::InvalidateRect(GetHWnd(), NULL, TRUE);
    ::EnumChildWindows(GetHWnd(), (WNDENUMPROC)InvalidateChildRect, 0);
}

void AwtWindow::SetResizable(BOOL isResizable)
{
    if (IsEmbedded()) {
	return;
    }
    long style = GetStyle();
    static const long SizeStyles = WS_THICKFRAME|WS_MAXIMIZEBOX|WS_MINIMIZEBOX;

    if (isResizable) {
	style |= SizeStyles;
    } else {
	style &= ~SizeStyles;
    }
    SetStyle(style);
    RedrawNonClient();
}

// SetWindowPos flags to cause frame edge to be recalculated
static const UINT SwpFrameChangeFlags =
    SWP_FRAMECHANGED | /* causes WM_NCCALCSIZE to be called */
    SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER |
    SWP_NOACTIVATE | SWP_NOCOPYBITS | 
    SWP_NOREPOSITION | SWP_NOSENDCHANGING;

//
// Forces WM_NCCALCSIZE to be called to recalculate
// window border (updates insets) without redrawing it
//
void AwtWindow::RecalcNonClient()
{
    ::SetWindowPos(GetHWnd(), (HWND) NULL, 0, 0, 0, 0, SwpFrameChangeFlags|SWP_NOREDRAW);
}

//
// Forces WM_NCCALCSIZE to be called to recalculate
// window border (updates insets) and redraws border to match
//
void AwtWindow::RedrawNonClient()
{
    ::SetWindowPos(GetHWnd(), (HWND) NULL, 0, 0, 0, 0, SwpFrameChangeFlags|SWP_ASYNCWINDOWPOS);
}

extern "C" {

/*
 * Class:     java_awt_Window
 * Method:    initIDs
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_java_awt_Window_initIDs(JNIEnv *env, jclass cls)
{
  AwtWindow::warningStringID =
    env->GetFieldID(cls, "warningString", "Ljava/lang/String;");
}

} /* extern "C" */


/************************************************************************
 * WindowPeer native methods
 */

extern "C" {

/*
 * Class:     sun_awt_windows_WWindowPeer
 * Method:    toFront
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WWindowPeer_toFront(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtWindow* w = (AwtWindow*)pData;

    ::SetWindowPos(w->GetHWnd(), HWND_TOP, 0,0,0,0,
                   SWP_NOMOVE|SWP_NOSIZE);
}

/*
 * Class:     sun_awt_windows_WWindowPeer
 * Method:    toBack
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WWindowPeer_toBack(JNIEnv *env, jobject self)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    AwtWindow* w = (AwtWindow*)pData;

    ::SetWindowPos(w->GetHWnd(), HWND_BOTTOM, 0,0,0,0,
                   SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
}

/*
 * Class:     sun_awt_windows_WWindowPeer
 * Method:    _setTitle
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WWindowPeer__1setTitle(JNIEnv *env, jobject self,
					    jstring title)
{
    int pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(title, "null title");

    AwtWindow* w = (AwtWindow*)pData;

    if (IS_NT) {
	int length = env->GetStringLength(title);
	WCHAR *buffer = new WCHAR[length + 1];
	env->GetStringRegion(title, 0, length, buffer);
	buffer[length] = L'\0';
	VERIFY(::SetWindowTextW(w->GetHWnd(), buffer));
	delete[] buffer;
    } else {
	int length = env->GetStringLength(title);
	char *buffer = new char[length * 2 + 1];
	AwtFont::javaString2multibyte(env, title, buffer, length);	
	VERIFY(::SetWindowText(w->GetHWnd(), buffer));
	delete[] buffer;
    }
}

/*
 * Class:     sun_awt_windows_WWindowPeer
 * Method:    _setResizable
 * Signature: (Z)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WWindowPeer__1setResizable(JNIEnv *env, jobject self,
						jboolean resizable)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);

    AwtWindow* w = (AwtWindow*)pData;
    ASSERT(!IsBadReadPtr(w, sizeof(AwtWindow)));
    w->SetResizable(resizable != 0);
}

/*
 * Class:     sun_awt_windows_WWindowPeer
 * Method:    create
 * Signature: (Lsun/awt/windows/WComponentPeer;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WWindowPeer_create(JNIEnv *env, jobject self, 
					jobject parent)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(parent);
    AwtToolkit::CreateComponent(self, parent,
				(AwtToolkit::ComponentFactory)
				AwtWindow::Create);
    JNI_CHECK_PEER_CREATION_RETURN(self);
}

/*
 * Class:     sun_awt_windows_WWindowPeer
 * Method:    updateInsets
 * Signature: (Ljava/awt/Insets;)V
 */
JNIEXPORT void JNICALL
Java_sun_awt_windows_WWindowPeer_updateInsets(JNIEnv *env, jobject self,
					      jobject insets)
{
    jint pData;
    JNI_CHECK_PEER_RETURN(self);
    JNI_CHECK_NULL_RETURN(insets, "null insets");

    AwtWindow* w = (AwtWindow*)pData;
    w->UpdateInsets(insets);
}

/*
 * Class:     sun_awt_windows_WWindowPeer
 * Method:    getContainerElement
 * Signature: (Ljava/awt/Container;I)Ljava/awt/Component;
 */
JNIEXPORT jobject JNICALL
Java_sun_awt_windows_WWindowPeer_getContainerElement(JNIEnv *env, jobject self,
						     jobject container, 
						     jint index)
{
    static jfieldID Container_ncomponents_FID = 0;
    static jfieldID Container_component_FID = 0;

    JNI_CHECK_NULL_RETURN_NULL(container, "null container");
 
    if ((env)->GetIntField(container, AwtContainer::ncomponentsID) > 0) {
	jobjectArray comps = 
	    (jobjectArray)(env)->GetObjectField(container, 
						AwtContainer::componentID);
        return env->GetObjectArrayElement(comps, index);
    }
    return NULL;
}


} /* extern "C" */
