/*
 * @(#)ByteToCharCp943.java	1.5 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package sun.io;

/**
* Tables and data to convert Cp943 to Unicode.
*
* @author Malcolm Ayres, assisted by UniMap program
*/
public class ByteToCharCp943
	extends ByteToCharDBCS_ASCII

{

	// Return the character set id
	public String getCharacterEncoding()
	{
		return "Cp943";
	}


	private static final boolean leadByte[] = {
		false, false, false, false, false, false, false, false,  // 00 - 07
		false, false, false, false, false, false, false, false,  // 08 - 0F
		false, false, false, false, false, false, false, false,  // 10 - 17
		false, false, false, false, false, false, false, false,  // 18 - 1F
		false, false, false, false, false, false, false, false,  // 20 - 27
		false, false, false, false, false, false, false, false,  // 28 - 2F
		false, false, false, false, false, false, false, false,  // 30 - 37
		false, false, false, false, false, false, false, false,  // 38 - 3F
		false, false, false, false, false, false, false, false,  // 40 - 47
		false, false, false, false, false, false, false, false,  // 48 - 4F
		false, false, false, false, false, false, false, false,  // 50 - 57
		false, false, false, false, false, false, false, false,  // 58 - 5F
		false, false, false, false, false, false, false, false,  // 60 - 67
		false, false, false, false, false, false, false, false,  // 68 - 6F
		false, false, false, false, false, false, false, false,  // 70 - 77
		false, false, false, false, false, false, false, false,  // 78 - 7F
		false, true,  true,  true,  true,  false, false, true,   // 80 - 87
		true,  true,  true,  true,  true,  true,  true,  true,   // 88 - 8F
		true,  true,  true,  true,  true,  true,  true,  true,   // 90 - 97
		true,  true,  true,  true,  true,  true,  true,  true,   // 98 - 9F
		false, false, false, false, false, false, false, false,  // A0 - A7
		false, false, false, false, false, false, false, false,  // A8 - AF
		false, false, false, false, false, false, false, false,  // B0 - B7
		false, false, false, false, false, false, false, false,  // B8 - BF
		false, false, false, false, false, false, false, false,  // C0 - C7
		false, false, false, false, false, false, false, false,  // C8 - CF
		false, false, false, false, false, false, false, false,  // D0 - D7
		false, false, false, false, false, false, false, false,  // D8 - DF
		true,  true,  true,  true,  true,  true,  true,  true,   // E0 - E7
		true,  true,  true,  true,  true,  true,  true,  false,  // E8 - EF
		true,  true,  true,  true,  true,  true,  true,  true,   // F0 - F7
		true,  true,  true,  true,  true,  false, false, false,  // F8 - FF
	};


	private static final String singleByteToChar = 
		"\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007" +
		"\u0008\u0009\n\u000B\u000C\r\u000E\u000F" +
		"\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017" +
		"\u0018\u0019\u001A\u001B\u001C\u001D\u001E\u001F" +
		"\u0020\u0021\"\u0023\u0024\u0025\u0026\u0027" +
		"\u0028\u0029\u002A\u002B\u002C\u002D\u002E\u002F" +
		"\u0030\u0031\u0032\u0033\u0034\u0035\u0036\u0037" +
		"\u0038\u0039\u003A\u003B\u003C\u003D\u003E\u003F" +
		"\u0040\u0041\u0042\u0043\u0044\u0045\u0046\u0047" +
		"\u0048\u0049\u004A\u004B\u004C\u004D\u004E\u004F" +
		"\u0050\u0051\u0052\u0053\u0054\u0055\u0056\u0057" +
		"\u0058\u0059\u005A\u005B\u00A5\u005D\u005E\u005F" +
		"\u0060\u0061\u0062\u0063\u0064\u0065\u0066\u0067" +
		"\u0068\u0069\u006A\u006B\u006C\u006D\u006E\u006F" +
		"\u0070\u0071\u0072\u0073\u0074\u0075\u0076\u0077" +
		"\u0078\u0079\u007A\u007B\u007C\u007D\u203E\u007F" +
		"\u00A2\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\u00A3\uFF61\uFF62\uFF63\uFF64\uFF65\uFF66\uFF67" +
		"\uFF68\uFF69\uFF6A\uFF6B\uFF6C\uFF6D\uFF6E\uFF6F" +
		"\uFF70\uFF71\uFF72\uFF73\uFF74\uFF75\uFF76\uFF77" +
		"\uFF78\uFF79\uFF7A\uFF7B\uFF7C\uFF7D\uFF7E\uFF7F" +
		"\uFF80\uFF81\uFF82\uFF83\uFF84\uFF85\uFF86\uFF87" +
		"\uFF88\uFF89\uFF8A\uFF8B\uFF8C\uFF8D\uFF8E\uFF8F" +
		"\uFF90\uFF91\uFF92\uFF93\uFF94\uFF95\uFF96\uFF97" +
		"\uFF98\uFF99\uFF9A\uFF9B\uFF9C\uFF9D\uFF9E\uFF9F" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\u00AC\\\u007E"
		;
	private static final short index1[] =
	{
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0000 - 00FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0100 - 01FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0200 - 02FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0300 - 03FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0400 - 04FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0500 - 05FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0600 - 06FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0700 - 07FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0800 - 08FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0900 - 09FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0A00 - 0AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0B00 - 0BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0C00 - 0CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0D00 - 0DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0E00 - 0EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 0F00 - 0FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1000 - 10FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1100 - 11FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1200 - 12FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1300 - 13FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1400 - 14FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1500 - 15FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1600 - 16FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1700 - 17FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1800 - 18FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1900 - 19FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1A00 - 1AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1B00 - 1BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1C00 - 1CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1D00 - 1DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1E00 - 1EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 1F00 - 1FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2000 - 20FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2100 - 21FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2200 - 22FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2300 - 23FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2400 - 24FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2500 - 25FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2600 - 26FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2700 - 27FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2800 - 28FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2900 - 29FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2A00 - 2AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2B00 - 2BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2C00 - 2CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2D00 - 2DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2E00 - 2EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 2F00 - 2FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3000 - 30FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3100 - 31FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3200 - 32FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3300 - 33FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3400 - 34FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3500 - 35FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3600 - 36FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3700 - 37FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3800 - 38FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3900 - 39FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3A00 - 3AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3B00 - 3BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3C00 - 3CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3D00 - 3DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3E00 - 3EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 3F00 - 3FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4000 - 40FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4100 - 41FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4200 - 42FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4300 - 43FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4400 - 44FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4500 - 45FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4600 - 46FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4700 - 47FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4800 - 48FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4900 - 49FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4A00 - 4AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4B00 - 4BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4C00 - 4CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4D00 - 4DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4E00 - 4EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 4F00 - 4FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5000 - 50FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5100 - 51FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5200 - 52FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5300 - 53FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5400 - 54FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5500 - 55FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5600 - 56FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5700 - 57FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5800 - 58FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5900 - 59FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5A00 - 5AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5B00 - 5BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5C00 - 5CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5D00 - 5DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5E00 - 5EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 5F00 - 5FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6000 - 60FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6100 - 61FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6200 - 62FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6300 - 63FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6400 - 64FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6500 - 65FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6600 - 66FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6700 - 67FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6800 - 68FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6900 - 69FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6A00 - 6AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6B00 - 6BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6C00 - 6CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6D00 - 6DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6E00 - 6EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 6F00 - 6FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7000 - 70FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7100 - 71FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7200 - 72FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7300 - 73FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7400 - 74FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7500 - 75FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7600 - 76FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7700 - 77FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7800 - 78FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7900 - 79FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7A00 - 7AFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7B00 - 7BFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7C00 - 7CFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7D00 - 7DFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7E00 - 7EFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 7F00 - 7FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 8000 - 80FF
		 4419,  4419,  4964,  6684,  9276,  8412,  5436,  6652, // 8100 - 81FF
		 4419,  4419,  9743,  4932,  9774,  3006,  1228,  2252, // 8200 - 82FF
		 4419,  4419,  3742,  9596,  8028,  6748,  5412,  4419, // 8300 - 83FF
		 4419,  4419,  4126,  2590,  1100, 10126,  4419,  4419, // 8400 - 84FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 8500 - 85FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // 8600 - 86FF
		 4419,  4419,  9870,  9532,  9244,  4419,  4419,  4419, // 8700 - 87FF
		 4419,  4419,  4419,  4419,  4420,  8924,  8668,  8380, // 8800 - 88FF
		 4419,  4419,  8124,  7804,  7516,  7260,  7004,  6620, // 8900 - 89FF
		 4419,  4419,  6364,  6044,  5788,  5532,  5220,  4900, // 8A00 - 8AFF
		 4419,  4419,  4612,  4286,  3838,  3518,  3262,  2974, // 8B00 - 8BFF
		 4419,  4419,  2718,  2366,  1996,  1740,  1484,  1196, // 8C00 - 8CFF
		 4419,  4419,   908,   588,   332,    76,  9148,  2220, // 8D00 - 8DFF
		 4419,  4419,  9628,  9116,  7708,  6268,  4804,  3710, // 8E00 - 8EFF
		 4419,  4419,  2622,  2188,   652, 10158, 10030,  9564, // 8F00 - 8FFF
		 4419,  4419,  9436,  9084,  8828,  8572,  8284,  7996, // 9000 - 90FF
		 4419,  4419,  7836,  7676,  7420,  7164,  6908,  6716, // 9100 - 91FF
		 4419,  4419,  6524,  6236,  6076,  5948,  5692,  5380, // 9200 - 92FF
		 4419,  4419,  5124,  4772,  4516,  4318,  4190,  4094, // 9300 - 93FF
		 4419,  4419,  3998,  3678,  3422,  3166,  2878,  2558, // 9400 - 94FF
		 4419,  4419,  2398,  2156,  1900,  1644,  1388,  1068, // 9500 - 95FF
		 4419,  4419,   812,   620,   492,   236, 10190, 10094, // 9600 - 96FF
		 4419,  4419, 10062,  9998,  9966,  9934,  9902,  9838, // 9700 - 97FF
		 4419,  4419,  9806,  9724,    12,  9692,  9660,  9500, // 9800 - 98FF
		 4419,  4419,  9468,  9404,  9372,  9340,  9308,  9212, // 9900 - 99FF
		 4419,  4419,  9180,  9052,  9020,  8988,  8956,  8892, // 9A00 - 9AFF
		 4419,  4419,  8860,  8796,  8764,  8732,  8700,  8636, // 9B00 - 9BFF
		 4419,  4419,  8604,  8540,  8508,  8476,  8444,  8348, // 9C00 - 9CFF
		 4419,  4419,  8316,  8252,  8220,  8188,  8156,  8092, // 9D00 - 9DFF
		 4419,  4419,  8060,  7964,  7932,  7900,  7868,  7772, // 9E00 - 9EFF
		 4419,  4419,  7740,  7644,  7612,  7580,  7548,  7484, // 9F00 - 9FFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A000 - A0FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A100 - A1FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A200 - A2FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A300 - A3FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A400 - A4FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A500 - A5FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A600 - A6FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A700 - A7FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A800 - A8FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // A900 - A9FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // AA00 - AAFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // AB00 - ABFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // AC00 - ACFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // AD00 - ADFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // AE00 - AEFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // AF00 - AFFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B000 - B0FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B100 - B1FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B200 - B2FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B300 - B3FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B400 - B4FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B500 - B5FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B600 - B6FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B700 - B7FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B800 - B8FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // B900 - B9FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // BA00 - BAFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // BB00 - BBFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // BC00 - BCFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // BD00 - BDFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // BE00 - BEFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // BF00 - BFFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C000 - C0FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C100 - C1FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C200 - C2FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C300 - C3FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C400 - C4FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C500 - C5FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C600 - C6FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C700 - C7FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C800 - C8FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // C900 - C9FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // CA00 - CAFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // CB00 - CBFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // CC00 - CCFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // CD00 - CDFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // CE00 - CEFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // CF00 - CFFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D000 - D0FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D100 - D1FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D200 - D2FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D300 - D3FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D400 - D4FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D500 - D5FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D600 - D6FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D700 - D7FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D800 - D8FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // D900 - D9FF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // DA00 - DAFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // DB00 - DBFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // DC00 - DCFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // DD00 - DDFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // DE00 - DEFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // DF00 - DFFF
		 4419,  4419,  7452,  7388,  7356,  7324,  7292,  7228, // E000 - E0FF
		 4419,  4419,  7196,  7132,  7100,  7068,  7036,  6972, // E100 - E1FF
		 4419,  4419,  6940,  6876,  6844,  6812,  6780,  6588, // E200 - E2FF
		 4419,  4419,  6556,  6492,  6460,  6428,  6396,  6332, // E300 - E3FF
		 4419,  4419,  6300,  6204,  6172,  6140,  6108,  6012, // E400 - E4FF
		 4419,  4419,  5980,  5916,  5884,  5852,  5820,  5756, // E500 - E5FF
		 4419,  4419,  5724,  5660,  5628,  5596,  5564,  5500, // E600 - E6FF
		 4419,  4419,  5468,  5348,  5316,  5284,  5252,  5188, // E700 - E7FF
		 4419,  4419,  5156,  5092,  5060,  5028,  4996,  4868, // E800 - E8FF
		 4419,  4419,  4836,  4740,  4708,  4676,  4644,  4580, // E900 - E9FF
		 4419,  4419,  4548,  4484,  4452,  4414,  4419,  4419, // EA00 - EAFF
		 4419,  4419,  4419,  4419,  2270,  4382,  4350,  4254, // EB00 - EBFF
		 4419,  4419,  4222,  4158,  4062,  4419,  4419,  4419, // EC00 - ECFF
		 4419,  4419,  4030,  3966,  3934,  3902,  3870,  3806, // ED00 - EDFF
		 4419,  4419,  3774,  3646,  3614,  3582,  3550,  3486, // EE00 - EEFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // EF00 - EFFF
		 4419,  4419,  3454,  3390,  3358,  3326,  3294,  3230, // F000 - F0FF
		 4419,  4419,  3198,  3134,  3102,  3070,  3038,  2942, // F100 - F1FF
		 4419,  4419,  2910,  2846,  2814,  2782,  2750,  2686, // F200 - F2FF
		 4419,  4419,  2654,  2526,  2494,  2462,  2430,  2334, // F300 - F3FF
		 4419,  4419,  2302,  2124,  2092,  2060,  2028,  1964, // F400 - F4FF
		 4419,  4419,  1932,  1868,  1836,  1804,  1772,  1708, // F500 - F5FF
		 4419,  4419,  1676,  1612,  1580,  1548,  1516,  1452, // F600 - F6FF
		 4419,  4419,  1420,  1356,  1324,  1292,  1260,  1164, // F700 - F7FF
		 4419,  4419,  1132,  1036,  1004,   972,   940,   876, // F800 - F8FF
		 4419,  4419,   844,   780,   748,   716,   684,   556, // F900 - F9FF
		 4419,  4419,   524,   460,   428,   396,   364,   300, // FA00 - FAFF
		 4419,  4419,   268,   204,   172,   140,   108,    44, // FB00 - FBFF
		 4419,  4419,     0,  4419,  4419,  4419,  4419,  4419, // FC00 - FCFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // FD00 - FDFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, // FE00 - FEFF
		 4419,  4419,  4419,  4419,  4419,  4419,  4419,  4419, 
	};

	private static final String index2 =
		"\u9ADC\u9B75\u9B72\u9B8F\u9BB1\u9BBB\u9C00\u9D70\u9D6B\uFA2D" + //     0 -     9
		"\u9E19\u9ED1\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //    10 -    19
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //    20 -    29
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //    30 -    39
		"\uFFFD\uFFFD\uFFFD\u5F0C\u9357\u93A4\u93C6\u93DE\u93F8\u9431" + //    40 -    49
		"\u9445\u9448\u9592\uF9DC\uFA29\u969D\u96AF\u9733\u973B\u9743" + //    50 -    59
		"\u974D\u974F\u9751\u9755\u9857\u9865\uFA2A\uFA2B\u9927\uFA2C" + //    60 -    69
		"\u999E\u9A4E\u9AD9\uFFFD\uFFFD\uFFFD\u9803\u4ECA\u56F0\u5764" + //    70 -    79
		"\u58BE\u5A5A\u6068\u61C7\u660F\u6606\u6839\u68B1\u6DF7\u75D5" + //    80 -    89
		"\u7D3A\u826E\u9B42\u4E9B\u4F50\u53C9\u5506\u5D6F\u5DE6\u5DEE" + //    90 -    99
		"\u67FB\u6C99\u7473\u7802\u8A50\u9396\u88DF\u5750\u91E5\u9206" + //   100 -   109
		"\u9210\u920A\u923A\u9240\u923C\u924E\u9259\u9251\u9239\u9267" + //   110 -   119
		"\u92A7\u9277\u9278\u92E7\u92D7\u92D9\u92D0\uFA27\u92D5\u92E0" + //   120 -   129
		"\u92D3\u9325\u9321\u92FB\uFA28\u931E\u92FF\u931D\u9302\u9370" + //   130 -   139
		"\uFA21\u8807\u88F5\u8A12\u8A37\u8A79\u8AA7\u8ABE\u8ADF\uFA22" + //   140 -   149
		"\u8AF6\u8B53\u8B7F\u8CF0\u8CF4\u8D12\u8D76\uFA23\u8ECF\uFA24" + //   150 -   159
		"\uFA25\u9067\u90DE\uFA26\u9115\u9127\u91DA\u91D7\u91DE\u91ED" + //   160 -   169
		"\u91EE\u91E4\uFA1A\u7994\uFA1B\u799B\u7AD1\u7AE7\uFA1C\u7AEB" + //   170 -   179
		"\u7B9E\uFA1D\u7D48\u7D5C\u7DB7\u7DA0\u7DD6\u7E52\u7F47\u7FA1" + //   180 -   189
		"\uFA1E\u8301\u8362\u837F\u83C7\u83F6\u8448\u84B4\u8553\u8559" + //   190 -   199
		"\u856B\uFA1F\u85B0\uFA20\u73BD\u73C9\u73D6\u73E3\u73D2\u7407" + //   200 -   209
		"\u73F5\u7426\u742A\u7429\u742E\u7462\u7489\u749F\u7501\u756F" + //   210 -   219
		"\u7682\u769C\u769E\u769B\u76A6\uFA17\u7746\u52AF\u7821\u784E" + //   220 -   229
		"\u7864\u787A\u7930\uFA18\uFA19\uFFFD\u8513\u5473\u672A\u9B45" + //   230 -   239
		"\u5DF3\u7B95\u5CAC\u5BC6\u871C\u6E4A\u84D1\u7A14\u8108\u5999" + //   240 -   249
		"\u7C8D\u6C11\u7720\u52D9\u5922\u7121\u725F\u77DB\u9727\u9D61" + //   250 -   259
		"\u690B\u5A7F\u5A18\u51A5\u540D\u547D\u660E\u76DF\u6D96\u6DAC" + //   260 -   269
		"\u6DCF\u6DF8\u6DF2\u6DFC\u6E39\u6E5C\u6E27\u6E3C\u6EBF\u6F88" + //   270 -   279
		"\u6FB5\u6FF5\u7005\u7007\u7028\u7085\u70AB\u710F\u7104\u715C" + //   280 -   289
		"\u7146\u7147\uFA15\u71C1\u71FE\u72B1\u72BE\u7324\uFA16\u7377" + //   290 -   299
		"\uF929\u6766\u67BB\u6852\u67C0\u6801\u6844\u68CF\uFA13\u6968" + //   300 -   309
		"\uFA14\u6998\u69E2\u6A30\u6A6B\u6A46\u6A73\u6A7E\u6AE2\u6AE4" + //   310 -   319
		"\u6BD6\u6C3F\u6C5C\u6C86\u6C6F\u6CDA\u6D04\u6D87\u6D6F\uFFFD" + //   320 -   329
		"\uFFFD\uFFFD\u9805\u9999\u9AD8\u9D3B\u525B\u52AB\u53F7\u5408" + //   330 -   339
		"\u58D5\u62F7\u6FE0\u8C6A\u8F5F\u9EB9\u514B\u523B\u544A\u56FD" + //   340 -   349
		"\u7A40\u9177\u9D60\u9ED2\u7344\u6F09\u8170\u7511\u5FFD\u60DA" + //   350 -   359
		"\u9AA8\u72DB\u8FBC\u6B64\u60D5\u6120\u60F2\u6111\u6137\u6130" + //   360 -   369
		"\u6198\u6213\u62A6\u63F5\u6460\u649D\u64CE\u654E\u6600\u6615" + //   370 -   379
		"\u663B\u6609\u662E\u661E\u6624\u6665\u6657\u6659\uFA12\u6673" + //   380 -   389
		"\u6699\u66A0\u66B2\u66BF\u66FA\u670E\u5953\u595B\u595D\u5963" + //   390 -   399
		"\u59A4\u59BA\u5B56\u5BC0\u752F\u5BD8\u5BEC\u5C1E\u5CA6\u5CBA" + //   400 -   409
		"\u5CF5\u5D27\u5D53\uFA11\u5D42\u5D6D\u5DB8\u5DB9\u5DD0\u5F21" + //   410 -   419
		"\u5F34\u5F67\u5FB7\u5FDE\u605D\u6085\u608A\u60DE\u5164\u519D" + //   420 -   429
		"\u51BE\u51EC\u5215\u529C\u52A6\u52C0\u52DB\u5300\u5307\u5324" + //   430 -   439
		"\u5372\u5393\u53B2\u53DD\uFA0E\u549C\u548A\u54A9\u54FF\u5586" + //   440 -   449
		"\u5759\u5765\u57AC\u57C8\u57C7\uFA0F\uFA10\u589E\u58B2\u590B" + //   450 -   459
		"\u84DC\u4FC9\u70BB\u6631\u68C8\u92F9\u66FB\u5F45\u4E28\u4EE1" + //   460 -   469
		"\u4EFC\u4F00\u4F03\u4F39\u4F56\u4F92\u4F8A\u4F9A\u4F94\u4FCD" + //   470 -   479
		"\u5040\u5022\u4FFF\u501E\u5046\u5070\u5042\u5094\u50F4\u50D8" + //   480 -   489
		"\u514A\uFFFD\u6469\u78E8\u9B54\u9EBB\u57CB\u59B9\u6627\u679A" + //   490 -   499
		"\u6BCE\u54E9\u69D9\u5E55\u819C\u6795\u9BAA\u67FE\u9C52\u685D" + //   500 -   509
		"\u4EA6\u4FE3\u53C8\u62B9\u672B\u6CAB\u8FC4\u4FAD\u7E6D\u9EBF" + //   510 -   519
		"\u4E07\u6162\u6E80\u6F2B\u2170\u2171\u2172\u2173\u2174\u2175" + //   520 -   529
		"\u2176\u2177\u2178\u2179\u2160\u2161\u2162\u2163\u2164\u2165" + //   530 -   539
		"\u2166\u2167\u2168\u2169\uFFE2\u00A6\uFF07\uFF02\u3231\uF86F" + //   540 -   549
		"\u2121\u2235\u7E8A\u891C\u9348\u9288\uE73B\uE73C\uE73D\uE73E" + //   550 -   559
		"\uE73F\uE740\uE741\uE742\uE743\uE744\uE745\uE746\uE747\uE748" + //   560 -   569
		"\uE749\uE74A\uE74B\uE74C\uE74D\uE74E\uE74F\uE750\uE751\uE752" + //   570 -   579
		"\uE753\uE754\uE755\uE756\uE757\uFFFD\uFFFD\uFFFD\u6E2F\u6E9D" + //   580 -   589
		"\u7532\u7687\u786C\u7A3F\u7CE0\u7D05\u7D18\u7D5E\u7DB1\u8015" + //   590 -   599
		"\u8003\u80AF\u80B1\u8154\u818F\u822A\u8352\u884C\u8861\u8B1B" + //   600 -   609
		"\u8CA2\u8CFC\u90CA\u9175\u9271\u783F\u92FC\u95A4\u964D\uFFFD" + //   610 -   619
		"\u5192\u7D21\u80AA\u81A8\u8B00\u8C8C\u8CBF\u927E\u9632\u5420" + //   620 -   629
		"\u982C\u5317\u50D5\u535C\u58A8\u64B2\u6734\u7267\u7766\u7A46" + //   630 -   639
		"\u91E6\u52C3\u6CA1\u6B86\u5800\u5E4C\u5954\u672C\u7FFB\u51E1" + //   640 -   649
		"\u76C6\uFFFD\u6E96\u6F64\u76FE\u7D14\u5DE1\u9075\u9187\u9806" + //   650 -   659
		"\u51E6\u521D\u6240\u6691\u66D9\u6E1A\u5EB6\u7DD2\u7F72\u66F8" + //   660 -   669
		"\u85AF\u85F7\u8AF8\u52A9\u53D9\u5973\u5E8F\u5F90\u6055\u92E4" + //   670 -   679
		"\u9664\u50B7\u511F\u52DD\uE71B\uE71C\uE71D\uE71E\uE71F\uE720" + //   680 -   689
		"\uE721\uE722\uE723\uE724\uE725\uE726\uE727\uE728\uE729\uE72A" + //   690 -   699
		"\uE72B\uE72C\uE72D\uE72E\uE72F\uE730\uE731\uE732\uE733\uE734" + //   700 -   709
		"\uE735\uE736\uE737\uE738\uE739\uE73A\uE6FB\uE6FC\uE6FD\uE6FE" + //   710 -   719
		"\uE6FF\uE700\uE701\uE702\uE703\uE704\uE705\uE706\uE707\uE708" + //   720 -   729
		"\uE709\uE70A\uE70B\uE70C\uE70D\uE70E\uE70F\uE710\uE711\uE712" + //   730 -   739
		"\uE713\uE714\uE715\uE716\uE717\uE718\uE719\uE71A\uE6DB\uE6DC" + //   740 -   749
		"\uE6DD\uE6DE\uE6DF\uE6E0\uE6E1\uE6E2\uE6E3\uE6E4\uE6E5\uE6E6" + //   750 -   759
		"\uE6E7\uE6E8\uE6E9\uE6EA\uE6EB\uE6EC\uE6ED\uE6EE\uE6EF\uE6F0" + //   760 -   769
		"\uE6F1\uE6F2\uE6F3\uE6F4\uE6F5\uE6F6\uE6F7\uE6F8\uE6F9\uE6FA" + //   770 -   779
		"\uE6BC\uE6BD\uE6BE\uE6BF\uE6C0\uE6C1\uE6C2\uE6C3\uE6C4\uE6C5" + //   780 -   789
		"\uE6C6\uE6C7\uE6C8\uE6C9\uE6CA\uE6CB\uE6CC\uE6CD\uE6CE\uE6CF" + //   790 -   799
		"\uE6D0\uE6D1\uE6D2\uE6D3\uE6D4\uE6D5\uE6D6\uE6D7\uE6D8\uE6D9" + //   800 -   809
		"\uE6DA\uFFFD\u6CD5\u6CE1\u70F9\u7832\u7E2B\u80DE\u82B3\u840C" + //   810 -   819
		"\u84EC\u8702\u8912\u8A2A\u8C4A\u90A6\u92D2\u98FD\u9CF3\u9D6C" + //   820 -   829
		"\u4E4F\u4EA1\u508D\u5256\u574A\u59A8\u5E3D\u5FD8\u5FD9\u623F" + //   830 -   839
		"\u66B4\u671B\u67D0\u68D2\uE69C\uE69D\uE69E\uE69F\uE6A0\uE6A1" + //   840 -   849
		"\uE6A2\uE6A3\uE6A4\uE6A5\uE6A6\uE6A7\uE6A8\uE6A9\uE6AA\uE6AB" + //   850 -   859
		"\uE6AC\uE6AD\uE6AE\uE6AF\uE6B0\uE6B1\uE6B2\uE6B3\uE6B4\uE6B5" + //   860 -   869
		"\uE6B6\uE6B7\uE6B8\uE6B9\uE6BA\uE6BB\uE67F\uE680\uE681\uE682" + //   870 -   879
		"\uE683\uE684\uE685\uE686\uE687\uE688\uE689\uE68A\uE68B\uE68C" + //   880 -   889
		"\uE68D\uE68E\uE68F\uE690\uE691\uE692\uE693\uE694\uE695\uE696" + //   890 -   899
		"\uE697\uE698\uE699\uE69A\uE69B\uFFFD\uFFFD\uFFFD\u540E\u5589" + //   900 -   909
		"\u5751\u57A2\u597D\u5B54\u5B5D\u5B8F\u5DE5\u5DE7\u5DF7\u5E78" + //   910 -   919
		"\u5E83\u5E9A\u5EB7\u5F18\u6052\u614C\u6297\u62D8\u63A7\u653B" + //   920 -   929
		"\u6602\u6643\u66F4\u676D\u6821\u6897\u69CB\u6C5F\u6D2A\u6D69" + //   930 -   939
		"\uE65F\uE660\uE661\uE662\uE663\uE664\uE665\uE666\uE667\uE668" + //   940 -   949
		"\uE669\uE66A\uE66B\uE66C\uE66D\uE66E\uE66F\uE670\uE671\uE672" + //   950 -   959
		"\uE673\uE674\uE675\uE676\uE677\uE678\uE679\uE67A\uE67B\uE67C" + //   960 -   969
		"\uE67D\uE67E\uE63F\uE640\uE641\uE642\uE643\uE644\uE645\uE646" + //   970 -   979
		"\uE647\uE648\uE649\uE64A\uE64B\uE64C\uE64D\uE64E\uE64F\uE650" + //   980 -   989
		"\uE651\uE652\uE653\uE654\uE655\uE656\uE657\uE658\uE659\uE65A" + //   990 -   999
		"\uE65B\uE65C\uE65D\uE65E\uE61F\uE620\uE621\uE622\uE623\uE624" + //  1000 -  1009
		"\uE625\uE626\uE627\uE628\uE629\uE62A\uE62B\uE62C\uE62D\uE62E" + //  1010 -  1019
		"\uE62F\uE630\uE631\uE632\uE633\uE634\uE635\uE636\uE637\uE638" + //  1020 -  1029
		"\uE639\uE63A\uE63B\uE63C\uE63D\uE63E\uE600\uE601\uE602\uE603" + //  1030 -  1039
		"\uE604\uE605\uE606\uE607\uE608\uE609\uE60A\uE60B\uE60C\uE60D" + //  1040 -  1049
		"\uE60E\uE60F\uE610\uE611\uE612\uE613\uE614\uE615\uE616\uE617" + //  1050 -  1059
		"\uE618\uE619\uE61A\uE61B\uE61C\uE61D\uE61E\uFFFD\u6B69\u752B" + //  1060 -  1069
		"\u88DC\u8F14\u7A42\u52DF\u5893\u6155\u620A\u66AE\u6BCD\u7C3F" + //  1070 -  1079
		"\u83E9\u5023\u4FF8\u5305\u5446\u5831\u5949\u5B9D\u5CF0\u5CEF" + //  1080 -  1089
		"\u5D29\u5E96\u62B1\u6367\u653E\u65B9\u670B\uFFFD\uFFFD\uFFFD" + //  1090 -  1099
		"\u043E\u043F\u0440\u0441\u0442\u0443\u0444\u0445\u0446\u0447" + //  1100 -  1109
		"\u0448\u0449\u044A\u044B\u044C\u044D\u044E\u044F\uFFFD\uFFFD" + //  1110 -  1119
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  1120 -  1129
		"\uFFFD\u2500\uE5E0\uE5E1\uE5E2\uE5E3\uE5E4\uE5E5\uE5E6\uE5E7" + //  1130 -  1139
		"\uE5E8\uE5E9\uE5EA\uE5EB\uE5EC\uE5ED\uE5EE\uE5EF\uE5F0\uE5F1" + //  1140 -  1149
		"\uE5F2\uE5F3\uE5F4\uE5F5\uE5F6\uE5F7\uE5F8\uE5F9\uE5FA\uE5FB" + //  1150 -  1159
		"\uE5FC\uE5FD\uE5FE\uE5FF\uE5C3\uE5C4\uE5C5\uE5C6\uE5C7\uE5C8" + //  1160 -  1169
		"\uE5C9\uE5CA\uE5CB\uE5CC\uE5CD\uE5CE\uE5CF\uE5D0\uE5D1\uE5D2" + //  1170 -  1179
		"\uE5D3\uE5D4\uE5D5\uE5D6\uE5D7\uE5D8\uE5D9\uE5DA\uE5DB\uE5DC" + //  1180 -  1189
		"\uE5DD\uE5DE\uE5DF\uFFFD\uFFFD\uFFFD\u5449\u543E\u5A2F\u5F8C" + //  1190 -  1199
		"\u5FA1\u609F\u68A7\u6A8E\u745A\u7881\u8A9E\u8AA4\u8B77\u9190" + //  1200 -  1209
		"\u4E5E\u9BC9\u4EA4\u4F7C\u4FAF\u5019\u5016\u5149\u516C\u529F" + //  1210 -  1219
		"\u52B9\u52FE\u539A\u53E3\u5411\uFFFD\uFFFD\uFFFD\u3062\u3063" + //  1220 -  1229
		"\u3064\u3065\u3066\u3067\u3068\u3069\u306A\u306B\u306C\u306D" + //  1230 -  1239
		"\u306E\u306F\u3070\u3071\u3072\u3073\u3074\u3075\u3076\u3077" + //  1240 -  1249
		"\u3078\u3079\u307A\u307B\u307C\u307D\u307E\u307F\u3080\u3081" + //  1250 -  1259
		"\uE5A3\uE5A4\uE5A5\uE5A6\uE5A7\uE5A8\uE5A9\uE5AA\uE5AB\uE5AC" + //  1260 -  1269
		"\uE5AD\uE5AE\uE5AF\uE5B0\uE5B1\uE5B2\uE5B3\uE5B4\uE5B5\uE5B6" + //  1270 -  1279
		"\uE5B7\uE5B8\uE5B9\uE5BA\uE5BB\uE5BC\uE5BD\uE5BE\uE5BF\uE5C0" + //  1280 -  1289
		"\uE5C1\uE5C2\uE583\uE584\uE585\uE586\uE587\uE588\uE589\uE58A" + //  1290 -  1299
		"\uE58B\uE58C\uE58D\uE58E\uE58F\uE590\uE591\uE592\uE593\uE594" + //  1300 -  1309
		"\uE595\uE596\uE597\uE598\uE599\uE59A\uE59B\uE59C\uE59D\uE59E" + //  1310 -  1319
		"\uE59F\uE5A0\uE5A1\uE5A2\uE563\uE564\uE565\uE566\uE567\uE568" + //  1320 -  1329
		"\uE569\uE56A\uE56B\uE56C\uE56D\uE56E\uE56F\uE570\uE571\uE572" + //  1330 -  1339
		"\uE573\uE574\uE575\uE576\uE577\uE578\uE579\uE57A\uE57B\uE57C" + //  1340 -  1349
		"\uE57D\uE57E\uE57F\uE580\uE581\uE582\uE544\uE545\uE546\uE547" + //  1350 -  1359
		"\uE548\uE549\uE54A\uE54B\uE54C\uE54D\uE54E\uE54F\uE550\uE551" + //  1360 -  1369
		"\uE552\uE553\uE554\uE555\uE556\uE557\uE558\uE559\uE55A\uE55B" + //  1370 -  1379
		"\uE55C\uE55D\uE55E\uE55F\uE560\uE561\uE562\uFFFD\u4E26\u853D" + //  1380 -  1389
		"\u9589\u965B\u7C73\u9801\u50FB\u58C1\u7656\u78A7\u5225\u77A5" + //  1390 -  1399
		"\u8511\u7B86\u504F\u5909\u7247\u7BC7\u7DE8\u8FBA\u8FD4\u904D" + //  1400 -  1409
		"\u4FBF\u52C9\u5A29\u5F01\u97AD\u4FDD\u8217\u92EA\u5703\u6355" + //  1410 -  1419
		"\uE524\uE525\uE526\uE527\uE528\uE529\uE52A\uE52B\uE52C\uE52D" + //  1420 -  1429
		"\uE52E\uE52F\uE530\uE531\uE532\uE533\uE534\uE535\uE536\uE537" + //  1430 -  1439
		"\uE538\uE539\uE53A\uE53B\uE53C\uE53D\uE53E\uE53F\uE540\uE541" + //  1440 -  1449
		"\uE542\uE543\uE507\uE508\uE509\uE50A\uE50B\uE50C\uE50D\uE50E" + //  1450 -  1459
		"\uE50F\uE510\uE511\uE512\uE513\uE514\uE515\uE516\uE517\uE518" + //  1460 -  1469
		"\uE519\uE51A\uE51B\uE51C\uE51D\uE51E\uE51F\uE520\uE521\uE522" + //  1470 -  1479
		"\uE523\uFFFD\uFFFD\uFFFD\u9650\u4E4E\u500B\u53E4\u547C\u56FA" + //  1480 -  1489
		"\u59D1\u5B64\u5DF1\u5EAB\u5F27\u6238\u6545\u67AF\u6E56\u72D0" + //  1490 -  1499
		"\u7CCA\u88B4\u80A1\u80E1\u83F0\u864E\u8A87\u8DE8\u9237\u96C7" + //  1500 -  1509
		"\u9867\u9F13\u4E94\u4E92\u4F0D\u5348\uE4E7\uE4E8\uE4E9\uE4EA" + //  1510 -  1519
		"\uE4EB\uE4EC\uE4ED\uE4EE\uE4EF\uE4F0\uE4F1\uE4F2\uE4F3\uE4F4" + //  1520 -  1529
		"\uE4F5\uE4F6\uE4F7\uE4F8\uE4F9\uE4FA\uE4FB\uE4FC\uE4FD\uE4FE" + //  1530 -  1539
		"\uE4FF\uE500\uE501\uE502\uE503\uE504\uE505\uE506\uE4C7\uE4C8" + //  1540 -  1549
		"\uE4C9\uE4CA\uE4CB\uE4CC\uE4CD\uE4CE\uE4CF\uE4D0\uE4D1\uE4D2" + //  1550 -  1559
		"\uE4D3\uE4D4\uE4D5\uE4D6\uE4D7\uE4D8\uE4D9\uE4DA\uE4DB\uE4DC" + //  1560 -  1569
		"\uE4DD\uE4DE\uE4DF\uE4E0\uE4E1\uE4E2\uE4E3\uE4E4\uE4E5\uE4E6" + //  1570 -  1579
		"\uE4A7\uE4A8\uE4A9\uE4AA\uE4AB\uE4AC\uE4AD\uE4AE\uE4AF\uE4B0" + //  1580 -  1589
		"\uE4B1\uE4B2\uE4B3\uE4B4\uE4B5\uE4B6\uE4B7\uE4B8\uE4B9\uE4BA" + //  1590 -  1599
		"\uE4BB\uE4BC\uE4BD\uE4BE\uE4BF\uE4C0\uE4C1\uE4C2\uE4C3\uE4C4" + //  1600 -  1609
		"\uE4C5\uE4C6\uE488\uE489\uE48A\uE48B\uE48C\uE48D\uE48E\uE48F" + //  1610 -  1619
		"\uE490\uE491\uE492\uE493\uE494\uE495\uE496\uE497\uE498\uE499" + //  1620 -  1629
		"\uE49A\uE49B\uE49C\uE49D\uE49E\uE49F\uE4A0\uE4A1\uE4A2\uE4A3" + //  1630 -  1639
		"\uE4A4\uE4A5\uE4A6\uFFFD\u8179\u8907\u8986\u6DF5\u5F17\u6255" + //  1640 -  1649
		"\u6CB8\u4ECF\u7269\u9B92\u5206\u543B\u5674\u58B3\u61A4\u626E" + //  1650 -  1659
		"\u711A\u596E\u7C89\u7CDE\u7D1B\u96F0\u6587\u805E\u4E19\u4F75" + //  1660 -  1669
		"\u5175\u5840\u5E63\u5E73\u5F0A\u67C4\uE468\uE469\uE46A\uE46B" + //  1670 -  1679
		"\uE46C\uE46D\uE46E\uE46F\uE470\uE471\uE472\uE473\uE474\uE475" + //  1680 -  1689
		"\uE476\uE477\uE478\uE479\uE47A\uE47B\uE47C\uE47D\uE47E\uE47F" + //  1690 -  1699
		"\uE480\uE481\uE482\uE483\uE484\uE485\uE486\uE487\uE44B\uE44C" + //  1700 -  1709
		"\uE44D\uE44E\uE44F\uE450\uE451\uE452\uE453\uE454\uE455\uE456" + //  1710 -  1719
		"\uE457\uE458\uE459\uE45A\uE45B\uE45C\uE45D\uE45E\uE45F\uE460" + //  1720 -  1729
		"\uE461\uE462\uE463\uE464\uE465\uE466\uE467\uFFFD\uFFFD\uFFFD" + //  1730 -  1739
		"\u6A29\u727D\u72AC\u732E\u7814\u786F\u7D79\u770C\u80A9\u898B" + //  1740 -  1749
		"\u8B19\u8CE2\u8ED2\u9063\u9375\u967A\u9855\u9A13\u9E78\u5143" + //  1750 -  1759
		"\u539F\u53B3\u5E7B\u5F26\u6E1B\u6E90\u7384\u73FE\u7D43\u8237" + //  1760 -  1769
		"\u8A00\u8AFA\uE42B\uE42C\uE42D\uE42E\uE42F\uE430\uE431\uE432" + //  1770 -  1779
		"\uE433\uE434\uE435\uE436\uE437\uE438\uE439\uE43A\uE43B\uE43C" + //  1780 -  1789
		"\uE43D\uE43E\uE43F\uE440\uE441\uE442\uE443\uE444\uE445\uE446" + //  1790 -  1799
		"\uE447\uE448\uE449\uE44A\uE40B\uE40C\uE40D\uE40E\uE40F\uE410" + //  1800 -  1809
		"\uE411\uE412\uE413\uE414\uE415\uE416\uE417\uE418\uE419\uE41A" + //  1810 -  1819
		"\uE41B\uE41C\uE41D\uE41E\uE41F\uE420\uE421\uE422\uE423\uE424" + //  1820 -  1829
		"\uE425\uE426\uE427\uE428\uE429\uE42A\uE3EB\uE3EC\uE3ED\uE3EE" + //  1830 -  1839
		"\uE3EF\uE3F0\uE3F1\uE3F2\uE3F3\uE3F4\uE3F5\uE3F6\uE3F7\uE3F8" + //  1840 -  1849
		"\uE3F9\uE3FA\uE3FB\uE3FC\uE3FD\uE3FE\uE3FF\uE400\uE401\uE402" + //  1850 -  1859
		"\uE403\uE404\uE405\uE406\uE407\uE408\uE409\uE40A\uE3CC\uE3CD" + //  1860 -  1869
		"\uE3CE\uE3CF\uE3D0\uE3D1\uE3D2\uE3D3\uE3D4\uE3D5\uE3D6\uE3D7" + //  1870 -  1879
		"\uE3D8\uE3D9\uE3DA\uE3DB\uE3DC\uE3DD\uE3DE\uE3DF\uE3E0\uE3E1" + //  1880 -  1889
		"\uE3E2\uE3E3\uE3E4\uE3E5\uE3E6\uE3E7\uE3E8\uE3E9\uE3EA\uFFFD" + //  1890 -  1899
		"\u65A7\u666E\u6D6E\u7236\u7B26\u8150\u819A\u8299\u8B5C\u8CA0" + //  1900 -  1909
		"\u8CE6\u8D74\u961C\u9644\u4FAE\u64AB\u6B66\u821E\u8461\u856A" + //  1910 -  1919
		"\u90E8\u5C01\u6953\u98A8\u847A\u8557\u4F0F\u526F\u5FA9\u5E45" + //  1920 -  1929
		"\u670D\u798F\uE3AC\uE3AD\uE3AE\uE3AF\uE3B0\uE3B1\uE3B2\uE3B3" + //  1930 -  1939
		"\uE3B4\uE3B5\uE3B6\uE3B7\uE3B8\uE3B9\uE3BA\uE3BB\uE3BC\uE3BD" + //  1940 -  1949
		"\uE3BE\uE3BF\uE3C0\uE3C1\uE3C2\uE3C3\uE3C4\uE3C5\uE3C6\uE3C7" + //  1950 -  1959
		"\uE3C8\uE3C9\uE3CA\uE3CB\uE38F\uE390\uE391\uE392\uE393\uE394" + //  1960 -  1969
		"\uE395\uE396\uE397\uE398\uE399\uE39A\uE39B\uE39C\uE39D\uE39E" + //  1970 -  1979
		"\uE39F\uE3A0\uE3A1\uE3A2\uE3A3\uE3A4\uE3A5\uE3A6\uE3A7\uE3A8" + //  1980 -  1989
		"\uE3A9\uE3AA\uE3AB\uFFFD\uFFFD\uFFFD\u5287\u621F\u6483\u6FC0" + //  1990 -  1999
		"\u9699\u6841\u5091\u6B20\u6C7A\u6F54\u7A74\u7D50\u8840\u8A23" + //  2000 -  2009
		"\u6708\u4EF6\u5039\u5026\u5065\u517C\u5238\u5263\u55A7\u570F" + //  2010 -  2019
		"\u5805\u5ACC\u5EFA\u61B2\u61F8\u62F3\u6372\u691C\uE36F\uE370" + //  2020 -  2029
		"\uE371\uE372\uE373\uE374\uE375\uE376\uE377\uE378\uE379\uE37A" + //  2030 -  2039
		"\uE37B\uE37C\uE37D\uE37E\uE37F\uE380\uE381\uE382\uE383\uE384" + //  2040 -  2049
		"\uE385\uE386\uE387\uE388\uE389\uE38A\uE38B\uE38C\uE38D\uE38E" + //  2050 -  2059
		"\uE34F\uE350\uE351\uE352\uE353\uE354\uE355\uE356\uE357\uE358" + //  2060 -  2069
		"\uE359\uE35A\uE35B\uE35C\uE35D\uE35E\uE35F\uE360\uE361\uE362" + //  2070 -  2079
		"\uE363\uE364\uE365\uE366\uE367\uE368\uE369\uE36A\uE36B\uE36C" + //  2080 -  2089
		"\uE36D\uE36E\uE32F\uE330\uE331\uE332\uE333\uE334\uE335\uE336" + //  2090 -  2099
		"\uE337\uE338\uE339\uE33A\uE33B\uE33C\uE33D\uE33E\uE33F\uE340" + //  2100 -  2109
		"\uE341\uE342\uE343\uE344\uE345\uE346\uE347\uE348\uE349\uE34A" + //  2110 -  2119
		"\uE34B\uE34C\uE34D\uE34E\uE310\uE311\uE312\uE313\uE314\uE315" + //  2120 -  2129
		"\uE316\uE317\uE318\uE319\uE31A\uE31B\uE31C\uE31D\uE31E\uE31F" + //  2130 -  2139
		"\uE320\uE321\uE322\uE323\uE324\uE325\uE326\uE327\uE328\uE329" + //  2140 -  2149
		"\uE32A\uE32B\uE32C\uE32D\uE32E\uFFFD\u63CF\u75C5\u79D2\u82D7" + //  2150 -  2159
		"\u9328\u92F2\u849C\u86ED\u9C2D\u54C1\u5F6C\u658C\u6D5C\u7015" + //  2160 -  2169
		"\u8CA7\u8CD3\u983B\u654F\u74F6\u4E0D\u4ED8\u57E0\u592B\u5A66" + //  2170 -  2179
		"\u5BCC\u51A8\u5E03\u5E9C\u6016\u6276\u6577\uFFFD\u6C41\u6E0B" + //  2180 -  2189
		"\u7363\u7E26\u91CD\u9283\u53D4\u5919\u5BBF\u6DD1\u795D\u7E2E" + //  2190 -  2199
		"\u7C9B\u587E\u719F\u51FA\u8853\u8FF0\u4FCA\u5CFB\u6625\u77AC" + //  2200 -  2209
		"\u7AE3\u821C\u99FF\u51C6\u5FAA\u65EC\u696F\u6B89\u6DF3\uFFFD" + //  2210 -  2219
		"\u8CA1\u51B4\u5742\u962A\u583A\u698A\u80B4\u54B2\u5D0E\u57FC" + //  2220 -  2229
		"\u7895\u9DFA\u4F5C\u524A\u548B\u643E\u6628\u6714\u67F5\u7A84" + //  2230 -  2239
		"\u7B56\u7D22\u932F\u685C\u9BAD\u7B39\u5319\u518A\u5237\uFFFD" + //  2240 -  2249
		"\uFFFD\uFFFD\u3082\u3083\u3084\u3085\u3086\u3087\u3088\u3089" + //  2250 -  2259
		"\u308A\u308B\u308C\u308D\u308E\u308F\u3090\u3091\u3092\u3093" + //  2260 -  2269
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  2270 -  2279
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  2280 -  2289
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  2290 -  2299
		"\uFFFD\uE758\uE2F0\uE2F1\uE2F2\uE2F3\uE2F4\uE2F5\uE2F6\uE2F7" + //  2300 -  2309
		"\uE2F8\uE2F9\uE2FA\uE2FB\uE2FC\uE2FD\uE2FE\uE2FF\uE300\uE301" + //  2310 -  2319
		"\uE302\uE303\uE304\uE305\uE306\uE307\uE308\uE309\uE30A\uE30B" + //  2320 -  2329
		"\uE30C\uE30D\uE30E\uE30F\uE2D3\uE2D4\uE2D5\uE2D6\uE2D7\uE2D8" + //  2330 -  2339
		"\uE2D9\uE2DA\uE2DB\uE2DC\uE2DD\uE2DE\uE2DF\uE2E0\uE2E1\uE2E2" + //  2340 -  2349
		"\uE2E3\uE2E4\uE2E5\uE2E6\uE2E7\uE2E8\uE2E9\uE2EA\uE2EB\uE2EC" + //  2350 -  2359
		"\uE2ED\uE2EE\uE2EF\uFFFD\uFFFD\uFFFD\u5F62\u5F84\u6075\u6176" + //  2360 -  2369
		"\u6167\u61A9\u63B2\u643A\u656C\u666F\u6842\u6E13\u7566\u7A3D" + //  2370 -  2379
		"\u7CFB\u7D4C\u7D99\u7E4B\u7F6B\u830E\u834A\u86CD\u8A08\u8A63" + //  2380 -  2389
		"\u8B66\u8EFD\u981A\u9D8F\u82B8\u8FCE\u9BE8\uFFFD\u9F3B\u67CA" + //  2390 -  2399
		"\u7A17\u5339\u758B\u9AED\u5F66\u819D\u83F1\u8098\u5F3C\u5FC5" + //  2400 -  2409
		"\u7562\u7B46\u903C\u6867\u59EB\u5A9B\u7D10\u767E\u8B2C\u4FF5" + //  2410 -  2419
		"\u5F6A\u6A19\u6C37\u6F02\u74E2\u7968\u8868\u8A55\u8C79\u5EDF" + //  2420 -  2429
		"\uE2B3\uE2B4\uE2B5\uE2B6\uE2B7\uE2B8\uE2B9\uE2BA\uE2BB\uE2BC" + //  2430 -  2439
		"\uE2BD\uE2BE\uE2BF\uE2C0\uE2C1\uE2C2\uE2C3\uE2C4\uE2C5\uE2C6" + //  2440 -  2449
		"\uE2C7\uE2C8\uE2C9\uE2CA\uE2CB\uE2CC\uE2CD\uE2CE\uE2CF\uE2D0" + //  2450 -  2459
		"\uE2D1\uE2D2\uE293\uE294\uE295\uE296\uE297\uE298\uE299\uE29A" + //  2460 -  2469
		"\uE29B\uE29C\uE29D\uE29E\uE29F\uE2A0\uE2A1\uE2A2\uE2A3\uE2A4" + //  2470 -  2479
		"\uE2A5\uE2A6\uE2A7\uE2A8\uE2A9\uE2AA\uE2AB\uE2AC\uE2AD\uE2AE" + //  2480 -  2489
		"\uE2AF\uE2B0\uE2B1\uE2B2\uE273\uE274\uE275\uE276\uE277\uE278" + //  2490 -  2499
		"\uE279\uE27A\uE27B\uE27C\uE27D\uE27E\uE27F\uE280\uE281\uE282" + //  2500 -  2509
		"\uE283\uE284\uE285\uE286\uE287\uE288\uE289\uE28A\uE28B\uE28C" + //  2510 -  2519
		"\uE28D\uE28E\uE28F\uE290\uE291\uE292\uE254\uE255\uE256\uE257" + //  2520 -  2529
		"\uE258\uE259\uE25A\uE25B\uE25C\uE25D\uE25E\uE25F\uE260\uE261" + //  2530 -  2539
		"\uE262\uE263\uE264\uE265\uE266\uE267\uE268\uE269\uE26A\uE26B" + //  2540 -  2549
		"\uE26C\uE26D\uE26E\uE26F\uE270\uE271\uE272\uFFFD\u6249\u6279" + //  2550 -  2559
		"\u62AB\u6590\u6BD4\u6CCC\u75B2\u76AE\u7891\u79D8\u7DCB\u7F77" + //  2560 -  2569
		"\u80A5\u88AB\u8AB9\u8CBB\u907F\u975E\u98DB\u6A0B\u7C38\u5099" + //  2570 -  2579
		"\u5C3E\u5FAE\u6787\u6BD8\u7435\u7709\u7F8E\uFFFD\uFFFD\uFFFD" + //  2580 -  2589
		"\u042F\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  2590 -  2599
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\u0430\u0431\u0432\u0433" + //  2600 -  2609
		"\u0434\u0435\u0451\u0436\u0437\u0438\u0439\u043A\u043B\u043C" + //  2610 -  2619
		"\u043D\uFFFD\u5B97\u5C31\u5DDE\u4FEE\u6101\u62FE\u6D32\u79C0" + //  2620 -  2629
		"\u79CB\u7D42\u7E4D\u7FD2\u81ED\u821F\u8490\u8846\u8972\u8B90" + //  2630 -  2639
		"\u8E74\u8F2F\u9031\u914B\u916C\u96C6\u919C\u4EC0\u4F4F\u5145" + //  2640 -  2649
		"\u5341\u5F93\u620E\u67D4\uE234\uE235\uE236\uE237\uE238\uE239" + //  2650 -  2659
		"\uE23A\uE23B\uE23C\uE23D\uE23E\uE23F\uE240\uE241\uE242\uE243" + //  2660 -  2669
		"\uE244\uE245\uE246\uE247\uE248\uE249\uE24A\uE24B\uE24C\uE24D" + //  2670 -  2679
		"\uE24E\uE24F\uE250\uE251\uE252\uE253\uE217\uE218\uE219\uE21A" + //  2680 -  2689
		"\uE21B\uE21C\uE21D\uE21E\uE21F\uE220\uE221\uE222\uE223\uE224" + //  2690 -  2699
		"\uE225\uE226\uE227\uE228\uE229\uE22A\uE22B\uE22C\uE22D\uE22E" + //  2700 -  2709
		"\uE22F\uE230\uE231\uE232\uE233\uFFFD\uFFFD\uFFFD\u6398\u7A9F" + //  2710 -  2719
		"\u6C93\u9774\u8F61\u7AAA\u718A\u9688\u7C82\u6817\u7E70\u6851" + //  2720 -  2729
		"\u936C\u52F2\u541B\u85AB\u8A13\u7FA4\u8ECD\u90E1\u5366\u8888" + //  2730 -  2739
		"\u7941\u4FC2\u50BE\u5211\u5144\u5553\u572D\u73EA\u578B\u5951" + //  2740 -  2749
		"\uE1F7\uE1F8\uE1F9\uE1FA\uE1FB\uE1FC\uE1FD\uE1FE\uE1FF\uE200" + //  2750 -  2759
		"\uE201\uE202\uE203\uE204\uE205\uE206\uE207\uE208\uE209\uE20A" + //  2760 -  2769
		"\uE20B\uE20C\uE20D\uE20E\uE20F\uE210\uE211\uE212\uE213\uE214" + //  2770 -  2779
		"\uE215\uE216\uE1D7\uE1D8\uE1D9\uE1DA\uE1DB\uE1DC\uE1DD\uE1DE" + //  2780 -  2789
		"\uE1DF\uE1E0\uE1E1\uE1E2\uE1E3\uE1E4\uE1E5\uE1E6\uE1E7\uE1E8" + //  2790 -  2799
		"\uE1E9\uE1EA\uE1EB\uE1EC\uE1ED\uE1EE\uE1EF\uE1F0\uE1F1\uE1F2" + //  2800 -  2809
		"\uE1F3\uE1F4\uE1F5\uE1F6\uE1B7\uE1B8\uE1B9\uE1BA\uE1BB\uE1BC" + //  2810 -  2819
		"\uE1BD\uE1BE\uE1BF\uE1C0\uE1C1\uE1C2\uE1C3\uE1C4\uE1C5\uE1C6" + //  2820 -  2829
		"\uE1C7\uE1C8\uE1C9\uE1CA\uE1CB\uE1CC\uE1CD\uE1CE\uE1CF\uE1D0" + //  2830 -  2839
		"\uE1D1\uE1D2\uE1D3\uE1D4\uE1D5\uE1D6\uE198\uE199\uE19A\uE19B" + //  2840 -  2849
		"\uE19C\uE19D\uE19E\uE19F\uE1A0\uE1A1\uE1A2\uE1A3\uE1A4\uE1A5" + //  2850 -  2859
		"\uE1A6\uE1A7\uE1A8\uE1A9\uE1AA\uE1AB\uE1AC\uE1AD\uE1AE\uE1AF" + //  2860 -  2869
		"\uE1B0\uE1B1\uE1B2\uE1B3\uE1B4\uE1B5\uE1B6\uFFFD\u642C\u6591" + //  2870 -  2879
		"\u677F\u6C3E\u6C4E\u7248\u72AF\u73ED\u7554\u7E41\u822C\u85E9" + //  2880 -  2889
		"\u8CA9\u7BC4\u91C6\u7169\u9812\u98EF\u633D\u6669\u756A\u76E4" + //  2890 -  2899
		"\u78D0\u8543\u86EE\u532A\u5351\u5426\u5983\u5E87\u5F7C\u60B2" + //  2900 -  2909
		"\uE178\uE179\uE17A\uE17B\uE17C\uE17D\uE17E\uE17F\uE180\uE181" + //  2910 -  2919
		"\uE182\uE183\uE184\uE185\uE186\uE187\uE188\uE189\uE18A\uE18B" + //  2920 -  2929
		"\uE18C\uE18D\uE18E\uE18F\uE190\uE191\uE192\uE193\uE194\uE195" + //  2930 -  2939
		"\uE196\uE197\uE15B\uE15C\uE15D\uE15E\uE15F\uE160\uE161\uE162" + //  2940 -  2949
		"\uE163\uE164\uE165\uE166\uE167\uE168\uE169\uE16A\uE16B\uE16C" + //  2950 -  2959
		"\uE16D\uE16E\uE16F\uE170\uE171\uE172\uE173\uE174\uE175\uE176" + //  2960 -  2969
		"\uE177\uFFFD\uFFFD\uFFFD\u91D1\u541F\u9280\u4E5D\u5036\u53E5" + //  2970 -  2979
		"\u533A\u72D7\u7396\u77E9\u82E6\u8EAF\u99C6\u99C8\u99D2\u5177" + //  2980 -  2989
		"\u611A\u865E\u55B0\u7A7A\u5076\u5BD3\u9047\u9685\u4E32\u6ADB" + //  2990 -  2999
		"\u91E7\u5C51\u5C48\uFFFD\uFFFD\uFFFD\u3042\u3043\u3044\u3045" + //  3000 -  3009
		"\u3046\u3047\u3048\u3049\u304A\u304B\u304C\u304D\u304E\u304F" + //  3010 -  3019
		"\u3050\u3051\u3052\u3053\u3054\u3055\u3056\u3057\u3058\u3059" + //  3020 -  3029
		"\u305A\u305B\u305C\u305D\u305E\u305F\u3060\u3061\uE13B\uE13C" + //  3030 -  3039
		"\uE13D\uE13E\uE13F\uE140\uE141\uE142\uE143\uE144\uE145\uE146" + //  3040 -  3049
		"\uE147\uE148\uE149\uE14A\uE14B\uE14C\uE14D\uE14E\uE14F\uE150" + //  3050 -  3059
		"\uE151\uE152\uE153\uE154\uE155\uE156\uE157\uE158\uE159\uE15A" + //  3060 -  3069
		"\uE11B\uE11C\uE11D\uE11E\uE11F\uE120\uE121\uE122\uE123\uE124" + //  3070 -  3079
		"\uE125\uE126\uE127\uE128\uE129\uE12A\uE12B\uE12C\uE12D\uE12E" + //  3080 -  3089
		"\uE12F\uE130\uE131\uE132\uE133\uE134\uE135\uE136\uE137\uE138" + //  3090 -  3099
		"\uE139\uE13A\uE0FB\uE0FC\uE0FD\uE0FE\uE0FF\uE100\uE101\uE102" + //  3100 -  3109
		"\uE103\uE104\uE105\uE106\uE107\uE108\uE109\uE10A\uE10B\uE10C" + //  3110 -  3119
		"\uE10D\uE10E\uE10F\uE110\uE111\uE112\uE113\uE114\uE115\uE116" + //  3120 -  3129
		"\uE117\uE118\uE119\uE11A\uE0DC\uE0DD\uE0DE\uE0DF\uE0E0\uE0E1" + //  3130 -  3139
		"\uE0E2\uE0E3\uE0E4\uE0E5\uE0E6\uE0E7\uE0E8\uE0E9\uE0EA\uE0EB" + //  3140 -  3149
		"\uE0EC\uE0ED\uE0EE\uE0EF\uE0F0\uE0F1\uE0F2\uE0F3\uE0F4\uE0F5" + //  3150 -  3159
		"\uE0F6\uE0F7\uE0F8\uE0F9\uE0FA\uFFFD\u7BB1\u7872\u7BB8\u8087" + //  3160 -  3169
		"\u7B48\u6AE8\u5E61\u808C\u7551\u7560\u516B\u9262\u6E8C\u767A" + //  3170 -  3179
		"\u9197\u9AEA\u4F10\u7F70\u629C\u7B4F\u95A5\u9CE9\u567A\u5859" + //  3180 -  3189
		"\u86E4\u96BC\u4F34\u5224\u534A\u53CD\u53DB\u5E06\uE0BC\uE0BD" + //  3190 -  3199
		"\uE0BE\uE0BF\uE0C0\uE0C1\uE0C2\uE0C3\uE0C4\uE0C5\uE0C6\uE0C7" + //  3200 -  3209
		"\uE0C8\uE0C9\uE0CA\uE0CB\uE0CC\uE0CD\uE0CE\uE0CF\uE0D0\uE0D1" + //  3210 -  3219
		"\uE0D2\uE0D3\uE0D4\uE0D5\uE0D6\uE0D7\uE0D8\uE0D9\uE0DA\uE0DB" + //  3220 -  3229
		"\uE09F\uE0A0\uE0A1\uE0A2\uE0A3\uE0A4\uE0A5\uE0A6\uE0A7\uE0A8" + //  3230 -  3239
		"\uE0A9\uE0AA\uE0AB\uE0AC\uE0AD\uE0AE\uE0AF\uE0B0\uE0B1\uE0B2" + //  3240 -  3249
		"\uE0B3\uE0B4\uE0B5\uE0B6\uE0B7\uE0B8\uE0B9\uE0BA\uE0BB\uFFFD" + //  3250 -  3259
		"\uFFFD\uFFFD\u9957\u9A5A\u4EF0\u51DD\u5C2D\u6681\u696D\u5C40" + //  3260 -  3269
		"\u66F2\u6975\u7389\u6850\u7C81\u50C5\u52E4\u5747\u5DFE\u9326" + //  3270 -  3279
		"\u65A4\u6B23\u6B3D\u7434\u7981\u79BD\u7B4B\u7DCA\u82B9\u83CC" + //  3280 -  3289
		"\u887F\u895F\u8B39\u8FD1\uE07F\uE080\uE081\uE082\uE083\uE084" + //  3290 -  3299
		"\uE085\uE086\uE087\uE088\uE089\uE08A\uE08B\uE08C\uE08D\uE08E" + //  3300 -  3309
		"\uE08F\uE090\uE091\uE092\uE093\uE094\uE095\uE096\uE097\uE098" + //  3310 -  3319
		"\uE099\uE09A\uE09B\uE09C\uE09D\uE09E\uE05F\uE060\uE061\uE062" + //  3320 -  3329
		"\uE063\uE064\uE065\uE066\uE067\uE068\uE069\uE06A\uE06B\uE06C" + //  3330 -  3339
		"\uE06D\uE06E\uE06F\uE070\uE071\uE072\uE073\uE074\uE075\uE076" + //  3340 -  3349
		"\uE077\uE078\uE079\uE07A\uE07B\uE07C\uE07D\uE07E\uE03F\uE040" + //  3350 -  3359
		"\uE041\uE042\uE043\uE044\uE045\uE046\uE047\uE048\uE049\uE04A" + //  3360 -  3369
		"\uE04B\uE04C\uE04D\uE04E\uE04F\uE050\uE051\uE052\uE053\uE054" + //  3370 -  3379
		"\uE055\uE056\uE057\uE058\uE059\uE05A\uE05B\uE05C\uE05D\uE05E" + //  3380 -  3389
		"\uE020\uE021\uE022\uE023\uE024\uE025\uE026\uE027\uE028\uE029" + //  3390 -  3399
		"\uE02A\uE02B\uE02C\uE02D\uE02E\uE02F\uE030\uE031\uE032\uE033" + //  3400 -  3409
		"\uE034\uE035\uE036\uE037\uE038\uE039\uE03A\uE03B\uE03C\uE03D" + //  3410 -  3419
		"\uE03E\uFFFD\u6973\u7164\u72FD\u8CB7\u58F2\u8CE0\u966A\u9019" + //  3420 -  3429
		"\u877F\u79E4\u77E7\u8429\u4F2F\u5265\u535A\u62CD\u67CF\u6CCA" + //  3430 -  3439
		"\u767D\u7B94\u7C95\u8236\u8584\u8FEB\u66DD\u6F20\u7206\u7E1B" + //  3440 -  3449
		"\u83AB\u99C1\u9EA6\u51FD\uE000\uE001\uE002\uE003\uE004\uE005" + //  3450 -  3459
		"\uE006\uE007\uE008\uE009\uE00A\uE00B\uE00C\uE00D\uE00E\uE00F" + //  3460 -  3469
		"\uE010\uE011\uE012\uE013\uE014\uE015\uE016\uE017\uE018\uE019" + //  3470 -  3479
		"\uE01A\uE01B\uE01C\uE01D\uE01E\uE01F\u9AD9\u9ADC\u9B75\u9B72" + //  3480 -  3489
		"\u9B8F\u9BB1\u9BBB\u9C00\u9D70\u9D6B\uFA2D\u9E19\u9ED1\uFFFD" + //  3490 -  3499
		"\uFFFD\u2170\u2171\u2172\u2173\u2174\u2175\u2176\u2177\u2178" + //  3500 -  3509
		"\u2179\uFFE2\u00A6\uFF07\uFF02\uFFFD\uFFFD\uFFFD\u4FA0\u50D1" + //  3510 -  3519
		"\u5147\u7AF6\u5171\u51F6\u5354\u5321\u537F\u53EB\u55AC\u5883" + //  3520 -  3529
		"\u5CE1\u5F37\u5F4A\u602F\u6050\u606D\u631F\u6559\u6A4B\u6CC1" + //  3530 -  3539
		"\u72C2\u72ED\u77EF\u80F8\u8105\u8208\u854E\u90F7\u93E1\u97FF" + //  3540 -  3549
		"\u92FF\u931D\u9302\u9370\u9357\u93A4\u93C6\u93DE\u93F8\u9431" + //  3550 -  3559
		"\u9445\u9448\u9592\uF9DC\uFA29\u969D\u96AF\u9733\u973B\u9743" + //  3560 -  3569
		"\u974D\u974F\u9751\u9755\u9857\u9865\uFA2A\uFA2B\u9927\uFA2C" + //  3570 -  3579
		"\u999E\u9A4E\u91DE\u91ED\u91EE\u91E4\u91E5\u9206\u9210\u920A" + //  3580 -  3589
		"\u923A\u9240\u923C\u924E\u9259\u9251\u9239\u9267\u92A7\u9277" + //  3590 -  3599
		"\u9278\u92E7\u92D7\u92D9\u92D0\uFA27\u92D5\u92E0\u92D3\u9325" + //  3600 -  3609
		"\u9321\u92FB\uFA28\u931E\u856B\uFA1F\u85B0\uFA20\uFA21\u8807" + //  3610 -  3619
		"\u88F5\u8A12\u8A37\u8A79\u8AA7\u8ABE\u8ADF\uFA22\u8AF6\u8B53" + //  3620 -  3629
		"\u8B7F\u8CF0\u8CF4\u8D12\u8D76\uFA23\u8ECF\uFA24\uFA25\u9067" + //  3630 -  3639
		"\u90DE\uFA26\u9115\u9127\u91DA\u91D7\u7930\uFA18\uFA19\uFA1A" + //  3640 -  3649
		"\u7994\uFA1B\u799B\u7AD1\u7AE7\uFA1C\u7AEB\u7B9E\uFA1D\u7D48" + //  3650 -  3659
		"\u7D5C\u7DB7\u7DA0\u7DD6\u7E52\u7F47\u7FA1\uFA1E\u8301\u8362" + //  3660 -  3669
		"\u837F\u83C7\u83F6\u8448\u84B4\u8553\u8559\uFFFD\u8997\u86A4" + //  3670 -  3679
		"\u5DF4\u628A\u64AD\u8987\u6777\u6CE2\u6D3E\u7436\u7834\u5A46" + //  3680 -  3689
		"\u7F75\u82AD\u99AC\u4FF3\u5EC3\u62DD\u6392\u6557\u676F\u76C3" + //  3690 -  3699
		"\u724C\u80CC\u80BA\u8F29\u914D\u500D\u57F9\u5A92\u6885\uFFFD" + //  3700 -  3709
		"\u932B\u82E5\u5BC2\u5F31\u60F9\u4E3B\u53D6\u5B88\u624B\u6731" + //  3710 -  3719
		"\u6B8A\u72E9\u73E0\u7A2E\u816B\u8DA3\u9152\u9996\u5112\u53D7" + //  3720 -  3729
		"\u546A\u5BFF\u6388\u6A39\u7DAC\u9700\u56DA\u53CE\u5468\uFFFD" + //  3730 -  3739
		"\uFFFD\uFFFD\u30A1\u30A2\u30A3\u30A4\u30A5\u30A6\u30A7\u30A8" + //  3740 -  3749
		"\u30A9\u30AA\u30AB\u30AC\u30AD\u30AE\u30AF\u30B0\u30B1\u30B2" + //  3750 -  3759
		"\u30B3\u30B4\u30B5\u30B6\u30B7\u30B8\u30B9\u30BA\u30BB\u30BC" + //  3760 -  3769
		"\u30BD\u30BE\u30BF\u30C0\u72BE\u7324\uFA16\u7377\u73BD\u73C9" + //  3770 -  3779
		"\u73D6\u73E3\u73D2\u7407\u73F5\u7426\u742A\u7429\u742E\u7462" + //  3780 -  3789
		"\u7489\u749F\u7501\u756F\u7682\u769C\u769E\u769B\u76A6\uFA17" + //  3790 -  3799
		"\u7746\u52AF\u7821\u784E\u7864\u787A\u6D6F\u6D96\u6DAC\u6DCF" + //  3800 -  3809
		"\u6DF8\u6DF2\u6DFC\u6E39\u6E5C\u6E27\u6E3C\u6EBF\u6F88\u6FB5" + //  3810 -  3819
		"\u6FF5\u7005\u7007\u7028\u7085\u70AB\u710F\u7104\u715C\u7146" + //  3820 -  3829
		"\u7147\uFA15\u71C1\u71FE\u72B1\uFFFD\uFFFD\uFFFD\u673D\u6C42" + //  3830 -  3839
		"\u6C72\u6CE3\u7078\u7403\u7A76\u7AAE\u7B08\u7D1A\u7CFE\u7D66" + //  3840 -  3849
		"\u65E7\u725B\u53BB\u5C45\u5DE8\u62D2\u62E0\u6319\u6E20\u865A" + //  3850 -  3859
		"\u8A31\u8DDD\u92F8\u6F01\u79A6\u9B5A\u4EA8\u4EAB\u4EAC\u4F9B" + //  3860 -  3869
		"\u66B2\u66BF\u66FA\u670E\uF929\u6766\u67BB\u6852\u67C0\u6801" + //  3870 -  3879
		"\u6844\u68CF\uFA13\u6968\uFA14\u6998\u69E2\u6A30\u6A6B\u6A46" + //  3880 -  3889
		"\u6A73\u6A7E\u6AE2\u6AE4\u6BD6\u6C3F\u6C5C\u6C86\u6C6F\u6CDA" + //  3890 -  3899
		"\u6D04\u6D87\u605D\u6085\u608A\u60DE\u60D5\u6120\u60F2\u6111" + //  3900 -  3909
		"\u6137\u6130\u6198\u6213\u62A6\u63F5\u6460\u649D\u64CE\u654E" + //  3910 -  3919
		"\u6600\u6615\u663B\u6609\u662E\u661E\u6624\u6665\u6657\u6659" + //  3920 -  3929
		"\uFA12\u6673\u6699\u66A0\uFA10\u589E\u58B2\u590B\u5953\u595B" + //  3930 -  3939
		"\u595D\u5963\u59A4\u59BA\u5B56\u5BC0\u752F\u5BD8\u5BEC\u5C1E" + //  3940 -  3949
		"\u5CA6\u5CBA\u5CF5\u5D27\u5D53\uFA11\u5D42\u5D6D\u5DB8\u5DB9" + //  3950 -  3959
		"\u5DD0\u5F21\u5F34\u5F67\u5FB7\u5FDE\u50F4\u50D8\u514A\u5164" + //  3960 -  3969
		"\u519D\u51BE\u51EC\u5215\u529C\u52A6\u52C0\u52DB\u5300\u5307" + //  3970 -  3979
		"\u5324\u5372\u5393\u53B2\u53DD\uFA0E\u549C\u548A\u54A9\u54FF" + //  3980 -  3989
		"\u5586\u5759\u5765\u57AC\u57C8\u57C7\uFA0F\uFFFD\u5982\u5C3F" + //  3990 -  3999
		"\u97EE\u4EFB\u598A\u5FCD\u8A8D\u6FE1\u79B0\u7962\u5BE7\u8471" + //  4000 -  4009
		"\u732B\u71B1\u5E74\u5FF5\u637B\u649A\u71C3\u7C98\u4E43\u5EFC" + //  4010 -  4019
		"\u4E4B\u57DC\u56A2\u60A9\u6FC3\u7D0D\u80FD\u8133\u81BF\u8FB2" + //  4020 -  4029
		"\u7E8A\u891C\u9348\u9288\u84DC\u4FC9\u70BB\u6631\u68C8\u92F9" + //  4030 -  4039
		"\u66FB\u5F45\u4E28\u4EE1\u4EFC\u4F00\u4F03\u4F39\u4F56\u4F92" + //  4040 -  4049
		"\u4F8A\u4F9A\u4F94\u4FCD\u5040\u5022\u4FFF\u501E\u5046\u5070" + //  4050 -  4059
		"\u5042\u5094\uE7F5\uE7F6\uE7F7\uE7F8\uE7F9\uE7FA\uE7FB\uE7FC" + //  4060 -  4069
		"\uE7FD\uE7FE\uE7FF\uE800\uE801\uE802\uE803\uE804\uE805\uE806" + //  4070 -  4079
		"\uE807\uE808\uE809\uE80A\uE80B\uE80C\uE80D\uE80E\uE80F\uE810" + //  4080 -  4089
		"\uE811\uE812\uE813\uFFFD\u5185\u4E4D\u51EA\u8599\u8B0E\u7058" + //  4090 -  4099
		"\u637A\u934B\u6962\u99B4\u7E04\u7577\u5357\u6960\u8EDF\u96E3" + //  4100 -  4109
		"\u6C5D\u4E8C\u5C3C\u5F10\u8FE9\u5302\u8CD1\u8089\u8679\u5EFF" + //  4110 -  4119
		"\u65E5\u4E73\u5165\uFFFD\uFFFD\uFFFD\u0410\u0411\u0412\u0413" + //  4120 -  4129
		"\u0414\u0415\u0401\u0416\u0417\u0418\u0419\u041A\u041B\u041C" + //  4130 -  4139
		"\u041D\u041E\u041F\u0420\u0421\u0422\u0423\u0424\u0425\u0426" + //  4140 -  4149
		"\u0427\u0428\u0429\u042A\u042B\u042C\u042D\u042E\uE7D6\uE7D7" + //  4150 -  4159
		"\uE7D8\uE7D9\uE7DA\uE7DB\uE7DC\uE7DD\uE7DE\uE7DF\uE7E0\uE7E1" + //  4160 -  4169
		"\uE7E2\uE7E3\uE7E4\uE7E5\uE7E6\uE7E7\uE7E8\uE7E9\uE7EA\uE7EB" + //  4170 -  4179
		"\uE7EC\uE7ED\uE7EE\uE7EF\uE7F0\uE7F1\uE7F2\uE7F3\uE7F4\uFFFD" + //  4180 -  4189
		"\u6D9C\u7279\u7763\u79BF\u7BE4\u6BD2\u72EC\u8AAD\u6803\u6A61" + //  4190 -  4199
		"\u51F8\u7A81\u6934\u5C4A\u9CF6\u82EB\u5BC5\u9149\u701E\u5678" + //  4200 -  4209
		"\u5C6F\u60C7\u6566\u6C8C\u8C5A\u9041\u9813\u5451\u66C7\u920D" + //  4210 -  4219
		"\u5948\u90A3\uE7B6\uE7B7\uE7B8\uE7B9\uE7BA\uE7BB\uE7BC\uE7BD" + //  4220 -  4229
		"\uE7BE\uE7BF\uE7C0\uE7C1\uE7C2\uE7C3\uE7C4\uE7C5\uE7C6\uE7C7" + //  4230 -  4239
		"\uE7C8\uE7C9\uE7CA\uE7CB\uE7CC\uE7CD\uE7CE\uE7CF\uE7D0\uE7D1" + //  4240 -  4249
		"\uE7D2\uE7D3\uE7D4\uE7D5\uE799\uE79A\uE79B\uE79C\uE79D\uE79E" + //  4250 -  4259
		"\uE79F\uE7A0\uE7A1\uE7A2\uE7A3\uE7A4\uE7A5\uE7A6\uE7A7\uE7A8" + //  4260 -  4269
		"\uE7A9\uE7AA\uE7AB\uE7AC\uE7AD\uE7AE\uE7AF\uE7B0\uE7B1\uE7B2" + //  4270 -  4279
		"\uE7B3\uE7B4\uE7B5\uFFFD\uFFFD\uFFFD\u7FA9\u87FB\u8ABC\u8B70" + //  4280 -  4289
		"\u63AC\u83CA\u97A0\u5409\u5403\u55AB\u6854\u6A58\u8A70\u7827" + //  4290 -  4299
		"\u6775\u9ECD\u5374\u5BA2\u811A\u8650\u9006\u4E18\u4E45\u4EC7" + //  4300 -  4309
		"\u4F11\u53CA\u5438\u5BAE\u5F13\u6025\u6551\uFFFD\u8569\u85E4" + //  4310 -  4319
		"\u8A0E\u8B04\u8C46\u8E0F\u9003\u900F\u9419\u9676\u982D\u9A30" + //  4320 -  4329
		"\u95D8\u50CD\u52D5\u540C\u5802\u5C0E\u61A7\u649E\u6D1E\u77B3" + //  4330 -  4339
		"\u7AE5\u80F4\u8404\u9053\u9285\u5CE0\u9D07\u533F\u5F97\u5FB3" + //  4340 -  4349
		"\uE779\uE77A\uE77B\uE77C\uE77D\uE77E\uE77F\uE780\uE781\uE782" + //  4350 -  4359
		"\uE783\uE784\uE785\uE786\uE787\uE788\uE789\uE78A\uE78B\uE78C" + //  4360 -  4369
		"\uE78D\uE78E\uE78F\uE790\uE791\uE792\uE793\uE794\uE795\uE796" + //  4370 -  4379
		"\uE797\uE798\uE759\uE75A\uE75B\uE75C\uE75D\uE75E\uE75F\uE760" + //  4380 -  4389
		"\uE761\uE762\uE763\uE764\uE765\uE766\uE767\uE768\uE769\uE76A" + //  4390 -  4399
		"\uE76B\uE76C\uE76D\uE76E\uE76F\uE770\uE771\uE772\uE773\uE774" + //  4400 -  4409
		"\uE775\uE776\uE777\uE778\u69C7\u9059\u7464\u51DC\u7199\uFFFD" + //  4410 -  4419
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  4420 -  4429
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  4430 -  4439
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  4440 -  4449
		"\uFFFD\u4E9C\u9EF4\u9EF6\u9EF7\u9EF9\u9EFB\u9EFC\u9EFD\u9F07" + //  4450 -  4459
		"\u9F08\u76B7\u9F15\u9F21\u9F2C\u9F3E\u9F4A\u9F52\u9F54\u9F63" + //  4460 -  4469
		"\u9F5F\u9F60\u9F61\u9F66\u9F67\u9F6C\u9F6A\u9F77\u9F72\u9F76" + //  4470 -  4479
		"\u9F95\u9F9C\u9FA0\u582F\u9E1B\u9E1E\u9E75\u9E79\u9E7D\u9E81" + //  4480 -  4489
		"\u9E88\u9E8B\u9E8C\u9E92\u9E95\u9E91\u9E9D\u9EA5\u9EA9\u9EB8" + //  4490 -  4499
		"\u9EAA\u9EAD\u9761\u9ECC\u9ECE\u9ECF\u9ED0\u9ED4\u9EDC\u9EDE" + //  4500 -  4509
		"\u9EDD\u9EE0\u9EE5\u9EE8\u9EEF\uFFFD\u51CD\u5200\u5510\u5854" + //  4510 -  4519
		"\u5858\u5957\u5B95\u5CF6\u5D8B\u60BC\u6295\u642D\u6771\u6843" + //  4520 -  4529
		"\u68BC\u68DF\u76D7\u6DD8\u6E6F\u6D9B\u706F\u71C8\u5F53\u75D8" + //  4530 -  4539
		"\u7977\u7B49\u7B54\u7B52\u7CD6\u7D71\u5230\u8463\u9D5D\u9D5E" + //  4540 -  4549
		"\u9D64\u9D51\u9D50\u9D59\u9D72\u9D89\u9D87\u9DAB\u9D6F\u9D7A" + //  4550 -  4559
		"\u9D9A\u9DA4\u9DA9\u9DB2\u9DC4\u9DC1\u9DBB\u9DB8\u9DBA\u9DC6" + //  4560 -  4569
		"\u9DCF\u9DC2\u9DD9\u9DD3\u9DF8\u9DE6\u9DED\u9DEF\u9DFD\u9E1A" + //  4570 -  4579
		"\u9C32\u9C46\u9C3E\u9C5A\u9C60\u9C67\u9C76\u9C78\u9CE7\u9CEC" + //  4580 -  4589
		"\u9CF0\u9D09\u9D08\u9CEB\u9D03\u9D06\u9D2A\u9D26\u9DAF\u9D23" + //  4590 -  4599
		"\u9D1F\u9D44\u9D15\u9D12\u9D41\u9D3F\u9D3E\u9D46\u9D48\uFFFD" + //  4600 -  4609
		"\uFFFD\uFFFD\u6A5F\u5E30\u6BC5\u6C17\u6C7D\u757F\u7948\u5B63" + //  4610 -  4619
		"\u7A00\u7D00\u5FBD\u898F\u8A18\u8CB4\u8D77\u8ECC\u8F1D\u98E2" + //  4620 -  4629
		"\u9A0E\u9B3C\u4E80\u507D\u5100\u5993\u5B9C\u622F\u6280\u64EC" + //  4630 -  4639
		"\u6B3A\u72A0\u7591\u7947\u9BCA\u9BB9\u9BC6\u9BCF\u9BD1\u9BD2" + //  4640 -  4649
		"\u9BE3\u9BE2\u9BE4\u9BD4\u9BE1\u9C3A\u9BF2\u9BF1\u9BF0\u9C15" + //  4650 -  4659
		"\u9C14\u9C09\u9C13\u9C0C\u9C06\u9C08\u9C12\u9C0A\u9C04\u9C2E" + //  4660 -  4669
		"\u9C1B\u9C25\u9C24\u9C21\u9C30\u9C47\u9B06\u9B18\u9B1A\u9B1F" + //  4670 -  4679
		"\u9B22\u9B23\u9B25\u9B27\u9B28\u9B29\u9B2A\u9B2E\u9B2F\u9B32" + //  4680 -  4689
		"\u9B44\u9B43\u9B4F\u9B4D\u9B4E\u9B51\u9B58\u9B74\u9B93\u9B83" + //  4690 -  4699
		"\u9B91\u9B96\u9B97\u9B9F\u9BA0\u9BA8\u9BB4\u9BC0\u9A3E\u9A55" + //  4700 -  4709
		"\u9A4D\u9A5B\u9A57\u9A5F\u9A62\u9A65\u9A64\u9A69\u9A6B\u9A6A" + //  4710 -  4719
		"\u9AAD\u9AB0\u9ABC\u9AC0\u9ACF\u9AD1\u9AD3\u9AD4\u9ADE\u9ADF" + //  4720 -  4729
		"\u9AE2\u9AE3\u9AE6\u9AEF\u9AEB\u9AEE\u9AF4\u9AF1\u9AF7\u9AFB" + //  4730 -  4739
		"\u9952\u994C\u9955\u9997\u9998\u99A5\u99AD\u99AE\u99BC\u99DF" + //  4740 -  4749
		"\u99DB\u99DD\u99D8\u99D1\u99ED\u99EE\u99F1\u99F2\u99FB\u99F8" + //  4750 -  4759
		"\u9A01\u9A0F\u9A05\u99E2\u9A19\u9A2B\u9A37\u9A45\u9A42\u9A40" + //  4760 -  4769
		"\u9A43\uFFFD\u4F1D\u6BBF\u6FB1\u7530\u96FB\u514E\u5410\u5835" + //  4770 -  4779
		"\u5857\u59AC\u5C60\u5F92\u6597\u675C\u6E21\u767B\u83DF\u8CED" + //  4780 -  4789
		"\u9014\u90FD\u934D\u7825\u783A\u52AA\u5EA6\u571F\u5974\u6012" + //  4790 -  4799
		"\u5012\u515A\u51AC\uFFFD\u5B9F\u8500\u7BE0\u5072\u67F4\u829D" + //  4800 -  4809
		"\u5C61\u854A\u7E1E\u820E\u5199\u5C04\u6368\u8D66\u659C\u716E" + //  4810 -  4819
		"\u793E\u7D17\u8005\u8B1D\u8ECA\u906E\u86C7\u90AA\u501F\u52FA" + //  4820 -  4829
		"\u5C3A\u6753\u707C\u7235\u914C\u91C8\u9871\u9874\u9873\u98AA" + //  4830 -  4839
		"\u98AF\u98B1\u98B6\u98C4\u98C3\u98C6\u98E9\u98EB\u9903\u9909" + //  4840 -  4849
		"\u9912\u9914\u9918\u9921\u991D\u991E\u9924\u9920\u992C\u992E" + //  4850 -  4859
		"\u993D\u993E\u9942\u9949\u9945\u9950\u994B\u9951\u97A8\u97A6" + //  4860 -  4869
		"\u97A3\u97B3\u97B4\u97C3\u97C6\u97C8\u97CB\u97DC\u97ED\u9F4F" + //  4870 -  4879
		"\u97F2\u7ADF\u97F6\u97F5\u980F\u980C\u9838\u9824\u9821\u9837" + //  4880 -  4889
		"\u983D\u9846\u984F\u984B\u986B\u986F\u9870\uFFFD\uFFFD\uFFFD" + //  4890 -  4899
		"\u764C\u773C\u5CA9\u7FEB\u8D0B\u96C1\u9811\u9854\u9858\u4F01" + //  4900 -  4909
		"\u4F0E\u5371\u559C\u5668\u57FA\u5947\u5B09\u5BC4\u5C90\u5E0C" + //  4910 -  4919
		"\u5E7E\u5FCC\u63EE\u673A\u65D7\u65E2\u671F\u68CB\u68C4\uFFFD" + //  4920 -  4929
		"\uFFFD\uFFFD\uFF21\uFF22\uFF23\uFF24\uFF25\uFF26\uFF27\uFF28" + //  4930 -  4939
		"\uFF29\uFF2A\uFF2B\uFF2C\uFF2D\uFF2E\uFF2F\uFF30\uFF31\uFF32" + //  4940 -  4949
		"\uFF33\uFF34\uFF35\uFF36\uFF37\uFF38\uFF39\uFF3A\uFFFD\uFFFD" + //  4950 -  4959
		"\uFFFD\uFFFD\uFFFD\uFFFD\u3000\u3001\u3002\uFF0C\uFF0E\u30FB" + //  4960 -  4969
		"\uFF1A\uFF1B\uFF1F\uFF01\u309B\u309C\u00B4\uFF40\u00A8\uFF3E" + //  4970 -  4979
		"\uFFE3\uFF3F\u30FD\u30FE\u309D\u309E\u3003\u4EDD\u3005\u3006" + //  4980 -  4989
		"\u3007\u30FC\u2014\u2010\uFF0F\uFF3C\u970F\u9716\u9719\u9724" + //  4990 -  4999
		"\u972A\u9730\u9739\u973D\u973E\u9744\u9746\u9748\u9742\u9749" + //  5000 -  5009
		"\u975C\u9760\u9764\u9766\u9768\u52D2\u976B\u9771\u9779\u9785" + //  5010 -  5019
		"\u977C\u9781\u977A\u9786\u978B\u978F\u9790\u979C\u965F\u9666" + //  5020 -  5029
		"\u9672\u966C\u968D\u9698\u9695\u9697\u96AA\u96A7\u96B1\u96B2" + //  5030 -  5039
		"\u96B0\u96B4\u96B6\u96B8\u96B9\u96CE\u96CB\u96C9\u96CD\u894D" + //  5040 -  5049
		"\u96DC\u970D\u96D5\u96F9\u9704\u9706\u9708\u9713\u970E\u9711" + //  5050 -  5059
		"\u95A0\u95A8\u95A7\u95AD\u95BC\u95BB\u95B9\u95BE\u95CA\u6FF6" + //  5060 -  5069
		"\u95C3\u95CD\u95CC\u95D5\u95D4\u95D6\u95DC\u95E1\u95E5\u95E2" + //  5070 -  5079
		"\u9621\u9628\u962E\u962F\u9642\u964C\u964F\u964B\u9677\u965C" + //  5080 -  5089
		"\u965E\u965D\u9410\u9436\u942B\u9435\u9421\u943A\u9441\u9452" + //  5090 -  5099
		"\u9444\u945B\u9460\u9462\u945E\u946A\u9229\u9470\u9475\u9477" + //  5100 -  5109
		"\u947D\u945A\u947C\u947E\u9481\u947F\u9582\u9587\u958A\u9594" + //  5110 -  5119
		"\u9596\u9598\u9599\uFFFD\u90B8\u912D\u91D8\u9F0E\u6CE5\u6458" + //  5120 -  5129
		"\u64E2\u6575\u6EF4\u7684\u7B1B\u9069\u93D1\u6EBA\u54F2\u5FB9" + //  5130 -  5139
		"\u64A4\u8F4D\u8FED\u9244\u5178\u586B\u5929\u5C55\u5E97\u6DFB" + //  5140 -  5149
		"\u7E8F\u751C\u8CBC\u8EE2\u985B\u70B9\u9319\u9322\u931A\u9323" + //  5150 -  5159
		"\u933A\u9335\u933B\u935C\u9360\u937C\u936E\u9356\u93B0\u93AC" + //  5160 -  5169
		"\u93AD\u9394\u93B9\u93D6\u93D7\u93E8\u93E5\u93D8\u93C3\u93DD" + //  5170 -  5179
		"\u93D0\u93C8\u93E4\u941A\u9414\u9413\u9403\u9407\u921E\u91FF" + //  5180 -  5189
		"\u9214\u922C\u9215\u9211\u925E\u9257\u9245\u9249\u9264\u9248" + //  5190 -  5199
		"\u9295\u923F\u924B\u9250\u929C\u9296\u9293\u929B\u925A\u92CF" + //  5200 -  5209
		"\u92B9\u92B7\u92E9\u930F\u92FA\u9344\u932E\uFFFD\uFFFD\uFFFD" + //  5210 -  5219
		"\u6F97\u6F45\u74B0\u7518\u76E3\u770B\u7AFF\u7BA1\u7C21\u7DE9" + //  5220 -  5229
		"\u7F36\u7FF0\u809D\u8266\u839E\u89B3\u8ACC\u8CAB\u9084\u9451" + //  5230 -  5239
		"\u9593\u9591\u95A2\u9665\u97D3\u9928\u8218\u4E38\u542B\u5CB8" + //  5240 -  5249
		"\u5DCC\u73A9\u9132\u9130\u914A\u9156\u9158\u9163\u9165\u9169" + //  5250 -  5259
		"\u9173\u9172\u918B\u9189\u9182\u91A2\u91AB\u91AF\u91AA\u91B5" + //  5260 -  5269
		"\u91B4\u91BA\u91C0\u91C1\u91C9\u91CB\u91D0\u91D6\u91DF\u91E1" + //  5270 -  5279
		"\u91DB\u91FC\u91F5\u91F6\u9050\u9051\u9052\u900E\u9049\u903E" + //  5280 -  5289
		"\u9056\u9058\u905E\u9068\u906F\u9076\u96A8\u9072\u9082\u907D" + //  5290 -  5299
		"\u9081\u9080\u908A\u9089\u908F\u90A8\u90AF\u90B1\u90B5\u90E2" + //  5300 -  5309
		"\u90E4\u6248\u90DB\u9102\u9112\u9119\u8F62\u8F63\u8F64\u8F9C" + //  5310 -  5319
		"\u8F9F\u8FA3\u8FAD\u8FAF\u8FB7\u8FDA\u8FE5\u8FE2\u8FEA\u8FEF" + //  5320 -  5329
		"\u9087\u8FF4\u9005\u8FF9\u8FFA\u9011\u9015\u9021\u900D\u901E" + //  5330 -  5339
		"\u9016\u900B\u9027\u9036\u9035\u9039\u8FF8\u904F\u8EC8\u8ECB" + //  5340 -  5349
		"\u8EDB\u8EE3\u8EFC\u8EFB\u8EEB\u8EFE\u8F0A\u8F05\u8F15\u8F12" + //  5350 -  5359
		"\u8F19\u8F13\u8F1C\u8F1F\u8F1B\u8F0C\u8F26\u8F33\u8F3B\u8F39" + //  5360 -  5369
		"\u8F45\u8F42\u8F3E\u8F4C\u8F49\u8F46\u8F4E\u8F57\u8F5C\uFFFD" + //  5370 -  5379
		"\u4EAD\u4F4E\u505C\u5075\u5243\u8C9E\u5448\u5824\u5B9A\u5E1D" + //  5380 -  5389
		"\u5E95\u5EAD\u5EF7\u5F1F\u608C\u62B5\u633A\u63D0\u68AF\u6C40" + //  5390 -  5399
		"\u7887\u798E\u7A0B\u7DE0\u8247\u8A02\u8AE6\u8E44\u9013\uFFFD" + //  5400 -  5409
		"\uFFFD\uFFFD\u03B2\u03B3\u03B4\u03B5\u03B6\u03B7\u03B8\u03B9" + //  5410 -  5419
		"\u03BA\u03BB\u03BC\u03BD\u03BE\u03BF\u03C0\u03C1\u03C3\u03C4" + //  5420 -  5429
		"\u03C5\u03C6\u03C7\u03C8\u03C9\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  5430 -  5439
		"\uFFFD\uFFFD\uFFFD\uFFFD\u2227\u2228\uFFE2\u21D2\u21D4\u2200" + //  5440 -  5449
		"\u2203\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  5450 -  5459
		"\uFFFD\uFFFD\u2220\u22A5\u2312\u2202\u2207\u2261\u8E47\u8E49" + //  5460 -  5469
		"\u8E4C\u8E50\u8E48\u8E59\u8E64\u8E60\u8E2A\u8E63\u8E55\u8E76" + //  5470 -  5479
		"\u8E72\u8E7C\u8E81\u8E87\u8E85\u8E84\u8E8B\u8E8A\u8E93\u8E91" + //  5480 -  5489
		"\u8E94\u8E99\u8EAA\u8EA1\u8EAC\u8EB0\u8EC6\u8EB1\u8EBE\u8EC5" + //  5490 -  5499
		"\u8D73\u8D81\u8D99\u8DC2\u8DBE\u8DBA\u8DCF\u8DDA\u8DD6\u8DCC" + //  5500 -  5509
		"\u8DDB\u8DCB\u8DEA\u8DEB\u8DDF\u8DE3\u8DFC\u8E08\u8E09\u8DFF" + //  5510 -  5519
		"\u8E1D\u8E1E\u8E10\u8E1F\u8E42\u8E35\u8E30\u8E34\u8E4A\uFFFD" + //  5520 -  5529
		"\uFFFD\uFFFD\u5208\u82C5\u74E6\u4E7E\u4F83\u51A0\u5BD2\u520A" + //  5530 -  5539
		"\u52D8\u52E7\u5DFB\u559A\u582A\u59E6\u5B8C\u5B98\u5BDB\u5E72" + //  5540 -  5549
		"\u5E79\u60A3\u611F\u6163\u61BE\u63DB\u6562\u67D1\u6853\u68FA" + //  5550 -  5559
		"\u6B3E\u6B53\u6C57\u6F22\u8C98\u621D\u8CAD\u8CAA\u8CBD\u8CB2" + //  5560 -  5569
		"\u8CB3\u8CAE\u8CB6\u8CC8\u8CC1\u8CE4\u8CE3\u8CDA\u8CFD\u8CFA" + //  5570 -  5579
		"\u8CFB\u8D04\u8D05\u8D0A\u8D07\u8D0F\u8D0D\u8D10\u9F4E\u8D13" + //  5580 -  5589
		"\u8CCD\u8D14\u8D16\u8D67\u8D6D\u8D71\u8B6C\u8B6F\u8B74\u8B7D" + //  5590 -  5599
		"\u8B80\u8B8C\u8B8E\u8B92\u8B93\u8B96\u8B99\u8B9A\u8C3A\u8C41" + //  5600 -  5609
		"\u8C3F\u8C48\u8C4C\u8C4E\u8C50\u8C55\u8C62\u8C6C\u8C78\u8C7A" + //  5610 -  5619
		"\u8C82\u8C89\u8C85\u8C8A\u8C8D\u8C8E\u8C94\u8C7C\u8AE4\u8AF1" + //  5620 -  5629
		"\u8B14\u8AE0\u8AE2\u8AF7\u8ADE\u8ADB\u8B0C\u8B07\u8B1A\u8AE1" + //  5630 -  5639
		"\u8B16\u8B10\u8B17\u8B20\u8B33\u97AB\u8B26\u8B2B\u8B3E\u8B28" + //  5640 -  5649
		"\u8B41\u8B4C\u8B4F\u8B4E\u8B49\u8B56\u8B5B\u8B5A\u8B6B\u8B5F" + //  5650 -  5659
		"\u8A0C\u8A1B\u8A1D\u8A25\u8A36\u8A41\u8A5B\u8A52\u8A46\u8A48" + //  5660 -  5669
		"\u8A7C\u8A6D\u8A6C\u8A62\u8A85\u8A82\u8A84\u8AA8\u8AA1\u8A91" + //  5670 -  5679
		"\u8AA5\u8AA6\u8A9A\u8AA3\u8AC4\u8ACD\u8AC2\u8ADA\u8AEB\u8AF3" + //  5680 -  5689
		"\u8AE7\uFFFD\u8CC3\u93AE\u9673\u6D25\u589C\u690E\u69CC\u8FFD" + //  5690 -  5699
		"\u939A\u75DB\u901A\u585A\u6802\u63B4\u69FB\u4F43\u6F2C\u67D8" + //  5700 -  5709
		"\u8FBB\u8526\u7DB4\u9354\u693F\u6F70\u576A\u58F7\u5B2C\u7D2C" + //  5710 -  5719
		"\u722A\u540A\u91E3\u9DB4\u8966\u8964\u896D\u896A\u896F\u8974" + //  5720 -  5729
		"\u8977\u897E\u8983\u8988\u898A\u8993\u8998\u89A1\u89A9\u89A6" + //  5730 -  5739
		"\u89AC\u89AF\u89B2\u89BA\u89BD\u89BF\u89C0\u89DA\u89DC\u89DD" + //  5740 -  5749
		"\u89E7\u89F4\u89F8\u8A03\u8A16\u8A10\u88C4\u88D4\u88D8\u88D9" + //  5750 -  5759
		"\u88DD\u88F9\u8902\u88FC\u88F4\u88E8\u88F2\u8904\u890C\u890A" + //  5760 -  5769
		"\u8913\u8943\u891E\u8925\u892A\u892B\u8941\u8944\u893B\u8936" + //  5770 -  5779
		"\u8938\u894C\u891D\u8960\u895E\uFFFD\uFFFD\uFFFD\u6A7F\u68B6" + //  5780 -  5789
		"\u9C0D\u6F5F\u5272\u559D\u6070\u62EC\u6D3B\u6E07\u6ED1\u845B" + //  5790 -  5799
		"\u8910\u8F44\u4E14\u9C39\u53F6\u691B\u6A3A\u9784\u682A\u515C" + //  5800 -  5809
		"\u7AC3\u84B2\u91DC\u938C\u565B\u9D28\u6822\u8305\u8431\u7CA5" + //  5810 -  5819
		"\u8821\u8831\u8836\u8839\u8827\u883B\u8844\u8842\u8852\u8859" + //  5820 -  5829
		"\u885E\u8862\u886B\u8881\u887E\u889E\u8875\u887D\u88B5\u8872" + //  5830 -  5839
		"\u8882\u8897\u8892\u88AE\u8899\u88A2\u888D\u88A4\u88B0\u88BF" + //  5840 -  5849
		"\u88B1\u88C3\u8763\u876A\u8805\u87A2\u879F\u8782\u87AF\u87CB" + //  5850 -  5859
		"\u87BD\u87C0\u87D0\u96D6\u87AB\u87C4\u87B3\u87C7\u87C6\u87BB" + //  5860 -  5869
		"\u87EF\u87F2\u87E0\u880F\u880D\u87FE\u87F6\u87F7\u880E\u87D2" + //  5870 -  5879
		"\u8811\u8816\u8815\u8822\u86DF\u86DB\u86EF\u8712\u8706\u8708" + //  5880 -  5889
		"\u8700\u8703\u86FB\u8711\u8709\u870D\u86F9\u870A\u8734\u873F" + //  5890 -  5899
		"\u8737\u873B\u8725\u8729\u871A\u8760\u875F\u8778\u874C\u874E" + //  5900 -  5909
		"\u8774\u8757\u8768\u876E\u8759\u8753\u85FE\u85FA\u8606\u8622" + //  5910 -  5919
		"\u861A\u8630\u863F\u864D\u4E55\u8654\u865F\u8667\u8671\u8693" + //  5920 -  5929
		"\u86A3\u86A9\u86AA\u868B\u868C\u86B6\u86AF\u86C4\u86C6\u86B0" + //  5930 -  5939
		"\u86C9\u8823\u86AB\u86D4\u86DE\u86E9\u86EC\uFFFD\u5E33\u5E81" + //  5940 -  5949
		"\u5F14\u5F35\u5F6B\u5FB4\u61F2\u6311\u66A2\u671D\u6F6E\u7252" + //  5950 -  5959
		"\u753A\u773A\u8074\u8139\u8178\u8776\u8ABF\u8ADC\u8D85\u8DF3" + //  5960 -  5969
		"\u929A\u9577\u9802\u9CE5\u52C5\u6357\u76F4\u6715\u6C88\u73CD" + //  5970 -  5979
		"\u8541\u8602\u854B\u8555\u8580\u85A4\u8588\u8591\u858A\u85A8" + //  5980 -  5989
		"\u856D\u8594\u859B\u85EA\u8587\u859C\u8577\u857E\u8590\u85C9" + //  5990 -  5999
		"\u85BA\u85CF\u85B9\u85D0\u85D5\u85DD\u85E5\u85DC\u85F9\u860A" + //  6000 -  6009
		"\u8613\u860B\u8435\u84CA\u8462\u84B9\u84BF\u849F\u84D9\u84CD" + //  6010 -  6019
		"\u84BB\u84DA\u84D0\u84C1\u84C6\u84D6\u84A1\u8521\u84FF\u84F4" + //  6020 -  6029
		"\u8517\u8518\u852C\u851F\u8515\u8514\u84FC\u8540\u8563\u8558" + //  6030 -  6039
		"\u8548\uFFFD\uFFFD\uFFFD\u67FF\u86CE\u920E\u5283\u5687\u5404" + //  6040 -  6049
		"\u5ED3\u62E1\u64B9\u683C\u6838\u6BBB\u7372\u78BA\u7A6B\u899A" + //  6050 -  6059
		"\u89D2\u8D6B\u8F03\u90ED\u95A3\u9694\u9769\u5B66\u5CB3\u697D" + //  6060 -  6069
		"\u984D\u984E\u639B\u7B20\u6A2B\uFFFD\u9010\u79E9\u7A92\u8336" + //  6070 -  6079
		"\u5AE1\u7740\u4E2D\u4EF2\u5B99\u5FE0\u62BD\u663C\u67F1\u6CE8" + //  6080 -  6089
		"\u866B\u8877\u8A3B\u914E\u92F3\u99D0\u6A17\u7026\u732A\u82E7" + //  6090 -  6099
		"\u8457\u8CAF\u4E01\u5146\u51CB\u558B\u5BF5\u5E16\u83CE\u83FD" + //  6100 -  6109
		"\u8403\u83D8\u840B\u83C1\u83F7\u8407\u83E0\u83F2\u840D\u8422" + //  6110 -  6119
		"\u8420\u83BD\u8438\u8506\u83FB\u846D\u842A\u843C\u855A\u8484" + //  6120 -  6129
		"\u8477\u846B\u84AD\u846E\u8482\u8469\u8446\u842C\u846F\u8479" + //  6130 -  6139
		"\u8334\u8316\u8332\u8331\u8340\u8339\u8350\u8345\u832F\u832B" + //  6140 -  6149
		"\u8317\u8318\u8385\u839A\u83AA\u839F\u83A2\u8396\u8323\u838E" + //  6150 -  6159
		"\u8387\u838A\u837C\u83B5\u8373\u8375\u83A0\u8389\u83A8\u83F4" + //  6160 -  6169
		"\u8413\u83EB\u8262\u8268\u826A\u826B\u822E\u8271\u8277\u8278" + //  6170 -  6179
		"\u827E\u828D\u8292\u82AB\u829F\u82BB\u82AC\u82E1\u82E3\u82DF" + //  6180 -  6189
		"\u82D2\u82F4\u82F3\u82FA\u8393\u8303\u82FB\u82F9\u82DE\u8306" + //  6190 -  6199
		"\u82DC\u8309\u82D9\u8335\u81CD\u81D1\u81D9\u81D8\u81C8\u81DA" + //  6200 -  6209
		"\u81DF\u81E0\u81E7\u81FA\u81FB\u81FE\u8201\u8202\u8205\u8207" + //  6210 -  6219
		"\u820A\u820D\u8210\u8216\u8229\u822B\u8238\u8233\u8240\u8259" + //  6220 -  6229
		"\u8258\u825D\u825A\u825F\u8264\uFFFD\u86CB\u8A95\u935B\u56E3" + //  6230 -  6239
		"\u58C7\u5F3E\u65AD\u6696\u6A80\u6BB5\u7537\u8AC7\u5024\u77E5" + //  6240 -  6249
		"\u5730\u5F1B\u6065\u667A\u6C60\u75F4\u7A1A\u7F6E\u81F4\u8718" + //  6250 -  6259
		"\u9045\u99B3\u7BC9\u755C\u7AF9\u7B51\u84C4\uFFFD\u6ECB\u6CBB" + //  6260 -  6269
		"\u723E\u74BD\u75D4\u78C1\u793A\u800C\u8033\u81EA\u8494\u8F9E" + //  6270 -  6279
		"\u6C50\u9E7F\u5F0F\u8B58\u9D2B\u7AFA\u8EF8\u5B8D\u96EB\u4E03" + //  6280 -  6289
		"\u53F1\u57F7\u5931\u5AC9\u5BA4\u6089\u6E7F\u6F06\u75BE\u8CEA" + //  6290 -  6299
		"\u968B\u8146\u813E\u8153\u8151\u80FC\u8171\u816E\u8165\u8166" + //  6300 -  6309
		"\u8174\u8183\u8188\u818A\u8180\u8182\u81A0\u8195\u81A4\u81A3" + //  6310 -  6319
		"\u815F\u8193\u81A9\u81B0\u81B5\u81BE\u81B8\u81BD\u81C0\u81C2" + //  6320 -  6329
		"\u81BA\u81C9\u8070\u8076\u8079\u807D\u807F\u8084\u8086\u8085" + //  6330 -  6339
		"\u809B\u8093\u809A\u80AD\u5190\u80AC\u80DB\u80E5\u80D9\u80DD" + //  6340 -  6349
		"\u80C4\u80DA\u80D6\u8109\u80EF\u80F1\u811B\u8129\u8123\u812F" + //  6350 -  6359
		"\u814B\uFFFD\uFFFD\uFFFD\u9B41\u6666\u68B0\u6D77\u7070\u754C" + //  6360 -  6369
		"\u7686\u7D75\u82A5\u87F9\u958B\u968E\u8C9D\u51F1\u52BE\u5916" + //  6370 -  6379
		"\u54B3\u5BB3\u5D16\u6168\u6982\u6DAF\u788D\u84CB\u8857\u8A72" + //  6380 -  6389
		"\u93A7\u9AB8\u6D6C\u99A8\u86D9\u57A3\u7FC5\u7FC6\u7FCA\u7FD5" + //  6390 -  6399
		"\u7FD4\u7FE1\u7FE6\u7FE9\u7FF3\u7FF9\u98DC\u8006\u8004\u800B" + //  6400 -  6409
		"\u8012\u8018\u8019\u801C\u8021\u8028\u803F\u803B\u804A\u8046" + //  6410 -  6419
		"\u8052\u8058\u805A\u805F\u8062\u8068\u8073\u8072\u7F4C\u7F4D" + //  6420 -  6429
		"\u7F4E\u7F50\u7F51\u7F55\u7F54\u7F58\u7F5F\u7F60\u7F68\u7F69" + //  6430 -  6439
		"\u7F67\u7F78\u7F82\u7F86\u7F83\u7F88\u7F87\u7F8C\u7F94\u7F9E" + //  6440 -  6449
		"\u7F9D\u7F9A\u7FA3\u7FAF\u7FB2\u7FB9\u7FAE\u7FB6\u7FB8\u8B71" + //  6450 -  6459
		"\u7E32\u7E3A\u7E67\u7E5D\u7E56\u7E5E\u7E59\u7E5A\u7E79\u7E6A" + //  6460 -  6469
		"\u7E69\u7E7C\u7E7B\u7E83\u7DD5\u7E7D\u8FAE\u7E7F\u7E88\u7E89" + //  6470 -  6479
		"\u7E8C\u7E92\u7E90\u7E93\u7E94\u7E96\u7E8E\u7E9B\u7E9C\u7F38" + //  6480 -  6489
		"\u7F3A\u7F45\u7E3D\u7DA2\u7DAF\u7DDC\u7DB8\u7D9F\u7DB0\u7DD8" + //  6490 -  6499
		"\u7DDD\u7DE4\u7DDE\u7DFB\u7DF2\u7DE1\u7E05\u7E0A\u7E23\u7E21" + //  6500 -  6509
		"\u7E12\u7E31\u7E1F\u7E09\u7E0B\u7E22\u7E46\u7E66\u7E3B\u7E35" + //  6510 -  6519
		"\u7E39\u7E43\u7E37\uFFFD\u53E9\u4F46\u9054\u8FB0\u596A\u8131" + //  6520 -  6529
		"\u5DFD\u7AEA\u8FBF\u68DA\u8C37\u72F8\u9C48\u6A3D\u8AB0\u4E39" + //  6530 -  6539
		"\u5358\u5606\u5766\u62C5\u63A2\u65E6\u6B4E\u6DE1\u6E5B\u70AD" + //  6540 -  6549
		"\u77ED\u7AEF\u7BAA\u7DBB\u803D\u80C6\u7D02\u7D1C\u7D15\u7D0A" + //  6550 -  6559
		"\u7D45\u7D4B\u7D2E\u7D32\u7D3F\u7D35\u7D46\u7D73\u7D56\u7D4E" + //  6560 -  6569
		"\u7D72\u7D68\u7D6E\u7D4F\u7D63\u7D93\u7D89\u7D5B\u7D8F\u7D7D" + //  6570 -  6579
		"\u7D9B\u7DBA\u7DAE\u7DA3\u7DB5\u7DC7\u7DBD\u7DAB\u7C75\u7C83" + //  6580 -  6589
		"\u7C90\u7CA4\u7CAD\u7CA2\u7CAB\u7CA1\u7CA8\u7CB3\u7CB2\u7CB1" + //  6590 -  6599
		"\u7CAE\u7CB9\u7CBD\u7CC0\u7CC5\u7CC2\u7CD8\u7CD2\u7CDC\u7CE2" + //  6600 -  6609
		"\u9B3B\u7CEF\u7CF2\u7CF4\u7CF6\u7CFA\u7D06\uFFFD\uFFFD\uFFFD" + //  6610 -  6619
		"\u971E\u868A\u4FC4\u5CE8\u6211\u7259\u753B\u81E5\u82BD\u86FE" + //  6620 -  6629
		"\u8CC0\u96C5\u9913\u99D5\u4ECB\u4F1A\u89E3\u56DE\u584A\u58CA" + //  6630 -  6639
		"\u5EFB\u5FEB\u602A\u6094\u6062\u61D0\u6212\u62D0\u6539\uFFFD" + //  6640 -  6649
		"\uFFFD\uFFFD\u2252\u226A\u226B\u221A\u223D\u221D\u2235\u222B" + //  6650 -  6659
		"\u222C\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\u212B\u2030" + //  6660 -  6669
		"\u266F\u266D\u266A\u2020\u2021\u00B6\uFFFD\uFFFD\uFFFD\uFFFD" + //  6670 -  6679
		"\u25EF\uFFFD\uFFFD\uFFFD\u301C\u2016\uFF5C\u2026\u2025\u2018" + //  6680 -  6689
		"\u2019\u201C\u201D\uFF08\uFF09\u3014\u3015\uFF3B\uFF3D\uFF5B" + //  6690 -  6699
		"\uFF5D\u3008\u3009\u300A\u300B\u300C\u300D\u300E\u300F\u3010" + //  6700 -  6709
		"\u3011\uFF0B\u2212\u00B1\u00D7\uFFFD\u968A\u9EDB\u9BDB\u4EE3" + //  6710 -  6719
		"\u53F0\u5927\u7B2C\u918D\u984C\u9DF9\u6EDD\u7027\u5353\u5544" + //  6720 -  6729
		"\u5B85\u6258\u629E\u62D3\u6CA2\u6FEF\u7422\u8A17\u9438\u6FC1" + //  6730 -  6739
		"\u8AFE\u8338\u51E7\u86F8\u53EA\uFFFD\uFFFD\uFFFD\u0392\u0393" + //  6740 -  6749
		"\u0394\u0395\u0396\u0397\u0398\u0399\u039A\u039B\u039C\u039D" + //  6750 -  6759
		"\u039E\u039F\u03A0\u03A1\u03A3\u03A4\u03A5\u03A6\u03A7\u03A8" + //  6760 -  6769
		"\u03A9\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\u03B1" + //  6770 -  6779
		"\u7C11\u7C14\u7BE6\u7BE5\u7C60\u7C00\u7C07\u7C13\u7BF3\u7BF7" + //  6780 -  6789
		"\u7C17\u7C0D\u7BF6\u7C23\u7C27\u7C2A\u7C1F\u7C37\u7C2B\u7C3D" + //  6790 -  6799
		"\u7C4C\u7C43\u7C54\u7C4F\u7C40\u7C50\u7C58\u7C5F\u7C64\u7C56" + //  6800 -  6809
		"\u7C65\u7C6C\u7B04\u7B4D\u7B0B\u7B4C\u7B45\u7B75\u7B65\u7B74" + //  6810 -  6819
		"\u7B67\u7B70\u7B71\u7B6C\u7B6E\u7B9D\u7B98\u7B9F\u7B8D\u7B9C" + //  6820 -  6829
		"\u7B9A\u7B8B\u7B92\u7B8F\u7B5D\u7B99\u7BCB\u7BC1\u7BCC\u7BCF" + //  6830 -  6839
		"\u7BB4\u7BC6\u7BDD\u7BE9\u7AB6\u7AC5\u7AC4\u7ABF\u9083\u7AC7" + //  6840 -  6849
		"\u7ACA\u7ACD\u7ACF\u7AD5\u7AD3\u7AD9\u7ADA\u7ADD\u7AE1\u7AE2" + //  6850 -  6859
		"\u7AE6\u7AED\u7AF0\u7B02\u7B0F\u7B0A\u7B06\u7B33\u7B18\u7B19" + //  6860 -  6869
		"\u7B1E\u7B35\u7B28\u7B36\u7B50\u7B7A\u79E1\u79E3\u7A08\u7A0D" + //  6870 -  6879
		"\u7A18\u7A19\u7A20\u7A1F\u7980\u7A31\u7A3B\u7A3E\u7A37\u7A43" + //  6880 -  6889
		"\u7A57\u7A49\u7A61\u7A62\u7A69\u9F9D\u7A70\u7A79\u7A7D\u7A88" + //  6890 -  6899
		"\u7A97\u7A95\u7A98\u7A96\u7AA9\u7AC8\u7AB0\uFFFD\u8A51\u553E" + //  6900 -  6909
		"\u5815\u59A5\u60F0\u6253\u67C1\u8235\u6955\u9640\u99C4\u9A28" + //  6910 -  6919
		"\u4F53\u5806\u5BFE\u8010\u5CB1\u5E2F\u5F85\u6020\u614B\u6234" + //  6920 -  6929
		"\u66FF\u6CF0\u6EDE\u80CE\u817F\u82D4\u888B\u8CB8\u9000\u902E" + //  6930 -  6939
		"\u78E7\u78DA\u78FD\u78F4\u7907\u7912\u7911\u7919\u792C\u792B" + //  6940 -  6949
		"\u7940\u7960\u7957\u795F\u795A\u7955\u7953\u797A\u797F\u798A" + //  6950 -  6959
		"\u799D\u79A7\u9F4B\u79AA\u79AE\u79B3\u79B9\u79BA\u79C9\u79D5" + //  6960 -  6969
		"\u79E7\u79EC\u77DC\u77E3\u77EE\u77FC\u780C\u7812\u7926\u7820" + //  6970 -  6979
		"\u792A\u7845\u788E\u7874\u7886\u787C\u789A\u788C\u78A3\u78B5" + //  6980 -  6989
		"\u78AA\u78AF\u78D1\u78C6\u78CB\u78D4\u78BE\u78BC\u78C5\u78CA" + //  6990 -  6999
		"\u78EC\uFFFD\uFFFD\uFFFD\u4F73\u52A0\u53EF\u5609\u590F\u5AC1" + //  7000 -  7009
		"\u5BB6\u5BE1\u79D1\u6687\u679C\u67B6\u6B4C\u6CB3\u706B\u73C2" + //  7010 -  7019
		"\u798D\u79BE\u7A3C\u7B87\u82B1\u82DB\u8304\u8377\u83EF\u83D3" + //  7020 -  7029
		"\u8766\u8AB2\u5629\u8CA8\u8FE6\u904E\u7724\u771E\u7725\u7726" + //  7030 -  7039
		"\u771B\u7737\u7738\u7747\u775A\u7768\u776B\u775B\u7765\u777F" + //  7040 -  7049
		"\u777E\u7779\u778E\u778B\u7791\u77A0\u779E\u77B0\u77B6\u77B9" + //  7050 -  7059
		"\u77BF\u77BC\u77BD\u77BB\u77C7\u77CD\u77D7\u77DA\u7676\u7678" + //  7060 -  7069
		"\u767C\u7680\u7683\u7688\u768B\u768E\u7696\u7693\u7699\u769A" + //  7070 -  7079
		"\u76B0\u76B4\u76B8\u76B9\u76BA\u76C2\u76CD\u76D6\u76D2\u76DE" + //  7080 -  7089
		"\u76E1\u76E5\u76E7\u76EA\u862F\u76FB\u7708\u7707\u7704\u7729" + //  7090 -  7099
		"\u75FC\u7601\u75F0\u75FA\u75F2\u75F3\u760B\u760D\u7609\u761F" + //  7100 -  7109
		"\u7627\u7620\u7621\u7622\u7624\u7634\u7630\u763B\u7647\u7648" + //  7110 -  7119
		"\u7646\u765C\u7658\u7661\u7662\u7668\u7669\u766A\u7667\u766C" + //  7120 -  7129
		"\u7670\u7672\u756B\u756D\u7578\u7576\u7586\u7587\u7574\u758A" + //  7130 -  7139
		"\u7589\u7582\u7594\u759A\u759D\u75A5\u75A3\u75C2\u75B3\u75C3" + //  7140 -  7149
		"\u75B5\u75BD\u75B8\u75BC\u75B1\u75CD\u75CA\u75D2\u75D9\u75E3" + //  7150 -  7159
		"\u75DE\u75FE\u75FF\uFFFD\u8535\u8D08\u9020\u4FC3\u5074\u5247" + //  7160 -  7169
		"\u5373\u606F\u6349\u675F\u6E2C\u8DB3\u901F\u4FD7\u5C5E\u8CCA" + //  7170 -  7179
		"\u65CF\u7D9A\u5352\u8896\u5176\u63C3\u5B58\u5B6B\u5C0A\u640D" + //  7180 -  7189
		"\u6751\u905C\u4ED6\u591A\u592A\u6C70\u74E0\u74E3\u74E7\u74E9" + //  7190 -  7199
		"\u74EE\u74F2\u74F0\u74F1\u74F8\u74F7\u7504\u7503\u7505\u750C" + //  7200 -  7209
		"\u750E\u750D\u7515\u7513\u751E\u7526\u752C\u753C\u7544\u754D" + //  7210 -  7219
		"\u754A\u7549\u755B\u7546\u755A\u7569\u7564\u7567\u73E5\u73EE" + //  7220 -  7229
		"\u73DE\u74A2\u7405\u746F\u7425\u73F8\u7432\u743A\u7455\u743F" + //  7230 -  7239
		"\u745F\u7459\u7441\u745C\u7469\u7470\u7463\u746A\u7476\u747E" + //  7240 -  7249
		"\u748B\u749E\u74A7\u74CA\u74CF\u74D4\u73F1\uFFFD\uFFFD\uFFFD" + //  7250 -  7259
		"\u65FA\u6A2A\u6B27\u6BB4\u738B\u7FC1\u8956\u9D2C\u9D0E\u9EC4" + //  7260 -  7269
		"\u5CA1\u6C96\u837B\u5104\u5C4B\u61B6\u81C6\u6876\u7261\u4E59" + //  7270 -  7279
		"\u4FFA\u5378\u6069\u6E29\u7A4F\u97F3\u4E0B\u5316\u4EEE\u4F55" + //  7280 -  7289
		"\u4F3D\u4FA1\u72E2\u72E0\u72E1\u72F9\u72F7\u500F\u7317\u730A" + //  7290 -  7299
		"\u731C\u7316\u731D\u7334\u732F\u7329\u7325\u733E\u734E\u734F" + //  7300 -  7309
		"\u9ED8\u7357\u736A\u7368\u7370\u7378\u7375\u737B\u737A\u73C8" + //  7310 -  7319
		"\u73B3\u73CE\u73BB\u73C0\u71FF\u720D\u7210\u721B\u7228\u722D" + //  7320 -  7329
		"\u722C\u7230\u7232\u723B\u723C\u723F\u7240\u7246\u724B\u7258" + //  7330 -  7339
		"\u7274\u727E\u7282\u7281\u7287\u7292\u7296\u72A2\u72A7\u72B9" + //  7340 -  7349
		"\u72B2\u72C3\u72C6\u72C4\u72CE\u72D2\u70D9\u7109\u70FD\u711C" + //  7350 -  7359
		"\u7119\u7165\u7155\u7188\u7166\u7162\u714C\u7156\u716C\u718F" + //  7360 -  7369
		"\u71FB\u7184\u7195\u71A8\u71AC\u71D7\u71B9\u71BE\u71D2\u71C9" + //  7370 -  7379
		"\u71D4\u71CE\u71E0\u71EC\u71E7\u71F5\u71FC\u71F9\u6FEE\u6FDB" + //  7380 -  7389
		"\u7009\u700B\u6FFA\u7011\u7001\u700F\u6FFE\u701B\u701A\u6F74" + //  7390 -  7399
		"\u701D\u7018\u701F\u7030\u703E\u7032\u7051\u7063\u7099\u7092" + //  7400 -  7409
		"\u70AF\u70F1\u70AC\u70B8\u70B3\u70AE\u70DF\u70CB\u70DD\uFFFD" + //  7410 -  7419
		"\u64CD\u65E9\u66F9\u5DE3\u69CD\u69FD\u6F15\u71E5\u4E89\u75E9" + //  7420 -  7429
		"\u76F8\u7A93\u7CDF\u7DCF\u7D9C\u8061\u8349\u8358\u846C\u84BC" + //  7430 -  7439
		"\u85FB\u88C5\u8D70\u9001\u906D\u9397\u971C\u9A12\u50CF\u5897" + //  7440 -  7449
		"\u618E\u81D3\u6F3E\u6F13\u6EF7\u6F86\u6F7A\u6F78\u6F81\u6F80" + //  7450 -  7459
		"\u6F6F\u6F5B\u6FF3\u6F6D\u6F82\u6F7C\u6F58\u6F8E\u6F91\u6FC2" + //  7460 -  7469
		"\u6F66\u6FB3\u6FA3\u6FA1\u6FA4\u6FB9\u6FC6\u6FAA\u6FDF\u6FD5" + //  7470 -  7479
		"\u6FEC\u6FD4\u6FD8\u6FF1\u6E38\u6E82\u6EAA\u6E98\u6EC9\u6EB7" + //  7480 -  7489
		"\u6ED3\u6EBD\u6EAF\u6EC4\u6EB2\u6ED4\u6ED5\u6E8F\u6EA5\u6EC2" + //  7490 -  7499
		"\u6E9F\u6F41\u6F11\u704C\u6EEC\u6EF8\u6EFE\u6F3F\u6EF2\u6F31" + //  7500 -  7509
		"\u6EEF\u6F32\u6ECC\uFFFD\uFFFD\uFFFD\u5712\u5830\u5944\u5BB4" + //  7510 -  7519
		"\u5EF6\u6028\u63A9\u63F4\u6CBF\u6F14\u708E\u7114\u7159\u71D5" + //  7520 -  7529
		"\u733F\u7E01\u8276\u82D1\u8597\u9060\u925B\u9D1B\u5869\u65BC" + //  7530 -  7539
		"\u6C5A\u7525\u51F9\u592E\u5965\u5F80\u5FDC\u62BC\u6DC6\u6DEC" + //  7540 -  7549
		"\u6DDE\u6DCC\u6DE8\u6DD2\u6DC5\u6DFA\u6DD9\u6DE4\u6DD5\u6DEA" + //  7550 -  7559
		"\u6DEE\u6E2D\u6E6E\u6E2E\u6E19\u6E72\u6E5F\u6E3E\u6E23\u6E6B" + //  7560 -  7569
		"\u6E2B\u6E76\u6E4D\u6E1F\u6E43\u6E3A\u6E4E\u6E24\u6EFF\u6E1D" + //  7570 -  7579
		"\u6CDB\u6CEF\u6CD9\u6CEA\u6D1F\u884D\u6D36\u6D2B\u6D3D\u6D38" + //  7580 -  7589
		"\u6D19\u6D35\u6D33\u6D12\u6D0C\u6D63\u6D93\u6D64\u6D5A\u6D79" + //  7590 -  7599
		"\u6D59\u6D8E\u6D95\u6FE4\u6D85\u6DF9\u6E15\u6E0A\u6DB5\u6DC7" + //  7600 -  7609
		"\u6DE6\u6DB8\u9EBE\u6C08\u6C13\u6C14\u6C1B\u6C24\u6C23\u6C5E" + //  7610 -  7619
		"\u6C55\u6C62\u6C6A\u6C82\u6C8D\u6C9A\u6C81\u6C9B\u6C7E\u6C68" + //  7620 -  7629
		"\u6C73\u6C92\u6C90\u6CC4\u6CF1\u6CD3\u6CBD\u6CD7\u6CC5\u6CDD" + //  7630 -  7639
		"\u6CAE\u6CB1\u6CBE\u6CBA\u6B54\u6B5B\u6B5F\u6B61\u6B78\u6B79" + //  7640 -  7649
		"\u6B7F\u6B80\u6B84\u6B83\u6B8D\u6B98\u6B95\u6B9E\u6BA4\u6BAA" + //  7650 -  7659
		"\u6BAB\u6BAF\u6BB2\u6BB1\u6BB3\u6BB7\u6BBC\u6BC6\u6BCB\u6BD3" + //  7660 -  7669
		"\u6BDF\u6BEC\u6BEB\u6BF3\u6BEF\uFFFD\u758F\u758E\u790E\u7956" + //  7670 -  7679
		"\u79DF\u7C97\u7D20\u7D44\u8607\u8A34\u963B\u9061\u9F20\u50E7" + //  7680 -  7689
		"\u5275\u53CC\u53E2\u5009\u55AA\u58EE\u594F\u723D\u5B8B\u5C64" + //  7690 -  7699
		"\u531D\u60E3\u60F3\u635C\u6383\u633F\u63BB\uFFFD\u6B7B\u6C0F" + //  7700 -  7709
		"\u7345\u7949\u79C1\u7CF8\u7D19\u7D2B\u80A2\u8102\u81F3\u8996" + //  7710 -  7719
		"\u8A5E\u8A69\u8A66\u8A8C\u8AEE\u8CC7\u8CDC\u96CC\u98FC\u6B6F" + //  7720 -  7729
		"\u4E8B\u4F3C\u4F8D\u5150\u5B57\u5BFA\u6148\u6301\u6642\u6B21" + //  7730 -  7739
		"\u6A97\u8617\u6ABB\u6AC3\u6AC2\u6AB8\u6AB3\u6AAC\u6ADE\u6AD1" + //  7740 -  7749
		"\u6ADF\u6AAA\u6ADA\u6AEA\u6AFB\u6B05\u8616\u6AFA\u6B12\u6B16" + //  7750 -  7759
		"\u9B31\u6B1F\u6B38\u6B37\u76DC\u6B39\u98EE\u6B47\u6B43\u6B49" + //  7760 -  7769
		"\u6B50\u6B59\u6A05\u69B1\u6A1E\u69ED\u6A14\u69EB\u6A0A\u6A12" + //  7770 -  7779
		"\u6AC1\u6A23\u6A13\u6A44\u6A0C\u6A72\u6A36\u6A78\u6A47\u6A62" + //  7780 -  7789
		"\u6A59\u6A66\u6A48\u6A38\u6A22\u6A90\u6A8D\u6AA0\u6A84\u6AA2" + //  7790 -  7799
		"\u6AA3\uFFFD\uFFFD\uFFFD\u834F\u990C\u53E1\u55B6\u5B30\u5F71" + //  7800 -  7809
		"\u6620\u66F3\u6804\u6C38\u6CF3\u6D29\u745B\u76C8\u7A4E\u9834" + //  7810 -  7819
		"\u82F1\u885B\u8A60\u92ED\u6DB2\u75AB\u76CA\u99C5\u60A6\u8B01" + //  7820 -  7829
		"\u8D8A\u95B2\u698E\u53AD\u5186\uFFFD\u7E4A\u7FA8\u817A\u821B" + //  7830 -  7839
		"\u8239\u85A6\u8A6E\u8CCE\u8DF5\u9078\u9077\u92AD\u9291\u9583" + //  7840 -  7849
		"\u9BAE\u524D\u5584\u6F38\u7136\u5168\u7985\u7E55\u81B3\u7CCE" + //  7850 -  7859
		"\u564C\u5851\u5CA8\u63AA\u66FE\u66FD\u695A\u72D9\u695D\u6981" + //  7860 -  7869
		"\u696A\u69B2\u69AE\u69D0\u69BF\u69C1\u69D3\u69BE\u69CE\u5BE8" + //  7870 -  7879
		"\u69CA\u69DD\u69BB\u69C3\u69A7\u6A2E\u6991\u69A0\u699C\u6995" + //  7880 -  7889
		"\u69B4\u69DE\u69E8\u6A02\u6A1B\u69FF\u6B0A\u69F9\u69F2\u69E7" + //  7890 -  7899
		"\u68E7\u68D5\u6936\u6912\u6904\u68D7\u68E3\u6925\u68F9\u68E0" + //  7900 -  7909
		"\u68EF\u6928\u692A\u691A\u6923\u6921\u68C6\u6979\u6977\u695C" + //  7910 -  7919
		"\u6978\u696B\u6954\u697E\u696E\u6939\u6974\u693D\u6959\u6930" + //  7920 -  7929
		"\u6961\u695E\u68B3\u682B\u6859\u6863\u6877\u687F\u689F\u688F" + //  7930 -  7939
		"\u68AD\u6894\u689D\u689B\u6883\u6AAE\u68B9\u6874\u68B5\u68A0" + //  7940 -  7949
		"\u68BA\u690F\u688D\u687E\u6901\u68CA\u6908\u68D8\u6922\u6926" + //  7950 -  7959
		"\u68E1\u690C\u68CD\u68D4\u677C\u676A\u678C\u678B\u67A6\u67A1" + //  7960 -  7969
		"\u6785\u67B7\u67EF\u67B4\u67EC\u67B3\u67E9\u67B8\u67E4\u67DE" + //  7970 -  7979
		"\u67DD\u67E2\u67EE\u67B9\u67CE\u67C6\u67E7\u6A9C\u681E\u6846" + //  7980 -  7989
		"\u6829\u6840\u684D\u6832\u684E\uFFFD\u8AAC\u96EA\u7D76\u820C" + //  7990 -  7999
		"\u8749\u4ED9\u5148\u5343\u5360\u5BA3\u5C02\u5C16\u5DDD\u6226" + //  8000 -  8009
		"\u6247\u64B0\u6813\u6834\u6CC9\u6D45\u6D17\u67D3\u6F5C\u714E" + //  8010 -  8019
		"\u717D\u65CB\u7A7F\u7BAD\u7DDA\uFFFD\uFFFD\uFFFD\u30E0\u30E1" + //  8020 -  8029
		"\u30E2\u30E3\u30E4\u30E5\u30E6\u30E7\u30E8\u30E9\u30EA\u30EB" + //  8030 -  8039
		"\u30EC\u30ED\u30EE\u30EF\u30F0\u30F1\u30F2\u30F3\u30F4\u30F5" + //  8040 -  8049
		"\u30F6\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\u0391" + //  8050 -  8059
		"\u66C4\u66B8\u66D6\u66DA\u66E0\u663F\u66E6\u66E9\u66F0\u66F5" + //  8060 -  8069
		"\u66F7\u670F\u6716\u671E\u6726\u6727\u9738\u672E\u673F\u6736" + //  8070 -  8079
		"\u6741\u6738\u6737\u6746\u675E\u6760\u6759\u6763\u6764\u6789" + //  8080 -  8089
		"\u6770\u67A9\u6773\u6635\u6636\u6634\u661C\u664F\u6644\u6649" + //  8090 -  8099
		"\u6641\u665E\u665D\u6664\u6667\u6668\u665F\u6662\u6670\u6683" + //  8100 -  8109
		"\u6688\u668E\u6689\u6684\u6698\u669D\u66C1\u66B9\u66C9\u66BE" + //  8110 -  8119
		"\u66BC\uFFFD\uFFFD\uFFFD\u9662\u9670\u96A0\u97FB\u540B\u53F3" + //  8120 -  8129
		"\u5B87\u70CF\u7FBD\u8FC2\u96E8\u536F\u9D5C\u7ABA\u4E11\u7893" + //  8130 -  8139
		"\u81FC\u6E26\u5618\u5504\u6B1D\u851A\u9C3B\u59E5\u53A9\u6D66" + //  8140 -  8149
		"\u74DC\u958F\u5642\u4E91\u904B\u96F2\u754B\u6548\u6556\u6555" + //  8150 -  8159
		"\u654D\u6558\u655E\u655D\u6572\u6578\u6582\u6583\u8B8A\u659B" + //  8160 -  8169
		"\u659F\u65AB\u65B7\u65C3\u65C6\u65C1\u65C4\u65CC\u65D2\u65DB" + //  8170 -  8179
		"\u65D9\u65E0\u65E1\u65F1\u6772\u660A\u6603\u65FB\u64D2\u64C5" + //  8180 -  8189
		"\u64C7\u64BB\u64D8\u64C2\u64F1\u64E7\u8209\u64E0\u64E1\u62AC" + //  8190 -  8199
		"\u64E3\u64EF\u652C\u64F6\u64F4\u64F2\u64FA\u6500\u64FD\u6518" + //  8200 -  8209
		"\u651C\u6505\u6524\u6523\u652B\u6534\u6535\u6537\u6536\u6538" + //  8210 -  8219
		"\u6369\u63BE\u63E9\u63C0\u63C6\u63E3\u63C9\u63D2\u63F6\u63C4" + //  8220 -  8229
		"\u6416\u6434\u6406\u6413\u6426\u6436\u651D\u6417\u6428\u640F" + //  8230 -  8239
		"\u6467\u646F\u6476\u644E\u652A\u6495\u6493\u64A5\u64A9\u6488" + //  8240 -  8249
		"\u64BC\u64DA\u62DC\u62CC\u62CA\u62C2\u62C7\u629B\u62C9\u630C" + //  8250 -  8259
		"\u62EE\u62F1\u6327\u6302\u6308\u62EF\u62F5\u6350\u633E\u634D" + //  8260 -  8269
		"\u641C\u634F\u6396\u638E\u6380\u63AB\u6376\u63A3\u638F\u6389" + //  8270 -  8279
		"\u639F\u63B5\u636B\uFFFD\u901D\u9192\u9752\u9759\u6589\u7A0E" + //  8280 -  8289
		"\u8106\u96BB\u5E2D\u60DC\u621A\u65A5\u6614\u6790\u77F3\u7A4D" + //  8290 -  8299
		"\u7C4D\u7E3E\u810A\u8CAC\u8D64\u8DE1\u8E5F\u78A9\u5207\u62D9" + //  8300 -  8309
		"\u63A5\u6442\u6298\u8A2D\u7A83\u7BC0\u621E\u6221\u622A\u622E" + //  8310 -  8319
		"\u6230\u6232\u6233\u6241\u624E\u625E\u6263\u625B\u6260\u6268" + //  8320 -  8329
		"\u627C\u6282\u6289\u627E\u6292\u6293\u6296\u62D4\u6283\u6294" + //  8330 -  8339
		"\u62D7\u62D1\u62BB\u62CF\u62FF\u62C6\u64D4\u62C8\u61AB\u61AE" + //  8340 -  8349
		"\u61CC\u61CA\u61C9\u61F7\u61C8\u61C3\u61C6\u61BA\u61CB\u7F79" + //  8350 -  8359
		"\u61CD\u61E6\u61E3\u61F6\u61FA\u61F4\u61FF\u61FD\u61FC\u61FE" + //  8360 -  8369
		"\u6200\u6208\u6209\u620D\u620C\u6214\u621B\uFFFD\uFFFD\uFFFD" + //  8370 -  8379
		"\u8B02\u9055\u907A\u533B\u4E95\u4EA5\u57DF\u80B2\u90C1\u78EF" + //  8380 -  8389
		"\u4E00\u58F1\u6EA2\u9038\u7A32\u8328\u828B\u9C2F\u5141\u5370" + //  8390 -  8399
		"\u54BD\u54E1\u56E0\u59FB\u5F15\u98F2\u6DEB\u80E4\u852D\uFFFD" + //  8400 -  8409
		"\uFFFD\uFFFD\u25A1\u25A0\u25B3\u25B2\u25BD\u25BC\u203B\u3012" + //  8410 -  8419
		"\u2192\u2190\u2191\u2193\u3013\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  8420 -  8429
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\u2208\u220B\u2286\u2287" + //  8430 -  8439
		"\u2282\u2283\u222A\u2229\u6128\u6127\u614A\u613F\u613C\u612C" + //  8440 -  8449
		"\u6134\u613D\u6142\u6144\u6173\u6177\u6158\u6159\u615A\u616B" + //  8450 -  8459
		"\u6174\u616F\u6165\u6171\u615F\u615D\u6153\u6175\u6199\u6196" + //  8460 -  8469
		"\u6187\u61AC\u6194\u619A\u618A\u6191\u609B\u6096\u6097\u6092" + //  8470 -  8479
		"\u60A7\u608B\u60E1\u60B8\u60E0\u60D3\u60B4\u5FF0\u60BD\u60C6" + //  8480 -  8489
		"\u60B5\u60D8\u614D\u6115\u6106\u60F6\u60F7\u6100\u60F4\u60FA" + //  8490 -  8499
		"\u6103\u6121\u60FB\u60F1\u610D\u610E\u6147\u613E\u6019\u6010" + //  8500 -  8509
		"\u6029\u600E\u6031\u601B\u6015\u602B\u6026\u600F\u603A\u605A" + //  8510 -  8519
		"\u6041\u606A\u6077\u605F\u604A\u6046\u604D\u6063\u6043\u6064" + //  8520 -  8529
		"\u6042\u606C\u606B\u6059\u6081\u608D\u60E7\u6083\u609A\u6084" + //  8530 -  8539
		"\u5F56\u5F57\u5F59\u5F61\u5F6D\u5F73\u5F77\u5F83\u5F82\u5F7F" + //  8540 -  8549
		"\u5F8A\u5F88\u5F91\u5F87\u5F9E\u5F99\u5F98\u5FA0\u5FA8\u5FAD" + //  8550 -  8559
		"\u5FBC\u5FD6\u5FFB\u5FE4\u5FF8\u5FF1\u5FDD\u60B3\u5FFF\u6021" + //  8560 -  8569
		"\u6060\uFFFD\u647A\u5BF8\u4E16\u702C\u755D\u662F\u51C4\u5236" + //  8570 -  8579
		"\u52E2\u59D3\u5F81\u6027\u6210\u653F\u6574\u661F\u6674\u68F2" + //  8580 -  8589
		"\u6816\u6B63\u6E05\u7272\u751F\u76DB\u7CBE\u8056\u58F0\u88FD" + //  8590 -  8599
		"\u897F\u8AA0\u8A93\u8ACB\u5ED6\u5EE3\u5EDD\u5EDA\u5EDB\u5EE2" + //  8600 -  8609
		"\u5EE1\u5EE8\u5EE9\u5EEC\u5EF1\u5EF3\u5EF0\u5EF4\u5EF8\u5EFE" + //  8610 -  8619
		"\u5F03\u5F09\u5F5D\u5F5C\u5F0B\u5F11\u5F16\u5F29\u5F2D\u5F38" + //  8620 -  8629
		"\u5F41\u5F48\u5F4C\u5F4E\u5F2F\u5F51\u5DF5\u5E0B\u5E1A\u5E19" + //  8630 -  8639
		"\u5E11\u5E1B\u5E36\u5E37\u5E44\u5E43\u5E40\u5E4E\u5E57\u5E54" + //  8640 -  8649
		"\u5E5F\u5E62\u5E64\u5E47\u5E75\u5E76\u5E7A\u9EBC\u5E7F\u5EA0" + //  8650 -  8659
		"\u5EC1\u5EC2\u5EC8\u5ED0\u5ECF\uFFFD\uFFFD\uFFFD\u5B89\u5EB5" + //  8660 -  8669
		"\u6309\u6697\u6848\u95C7\u978D\u674F\u4EE5\u4F0A\u4F4D\u4F9D" + //  8670 -  8679
		"\u5049\u56F2\u5937\u59D4\u5A01\u5C09\u60DF\u610F\u6170\u6613" + //  8680 -  8689
		"\u6905\u70BA\u754F\u7570\u79FB\u7DAD\u7DEF\u80C3\u840E\u8863" + //  8690 -  8699
		"\u5D11\u5D14\u5D22\u5D1A\u5D19\u5D18\u5D4C\u5D52\u5D4E\u5D4B" + //  8700 -  8709
		"\u5D6C\u5D73\u5D76\u5D87\u5D84\u5D82\u5DA2\u5D9D\u5DAC\u5DAE" + //  8710 -  8719
		"\u5DBD\u5D90\u5DB7\u5DBC\u5DC9\u5DCD\u5DD3\u5DD2\u5DD6\u5DDB" + //  8720 -  8729
		"\u5DEB\u5DF2\u5C4F\u5B71\u5C6C\u5C6E\u4E62\u5C76\u5C79\u5C8C" + //  8730 -  8739
		"\u5C91\u5C94\u599B\u5CAB\u5CBB\u5CB6\u5CBC\u5CB7\u5CC5\u5CBE" + //  8740 -  8749
		"\u5CC7\u5CD9\u5CE9\u5CFD\u5CFA\u5CED\u5D8C\u5CEA\u5D0B\u5D15" + //  8750 -  8759
		"\u5D17\u5D5C\u5D1F\u5D1B\u5B83\u5BA6\u5BB8\u5BC3\u5BC7\u5BC9" + //  8760 -  8769
		"\u5BD4\u5BD0\u5BE4\u5BE6\u5BE2\u5BDE\u5BE5\u5BEB\u5BF0\u5BF6" + //  8770 -  8779
		"\u5BF3\u5C05\u5C07\u5C08\u5C0D\u5C13\u5C20\u5C22\u5C28\u5C38" + //  8780 -  8789
		"\u5C39\u5C41\u5C46\u5C4E\u5C53\u5C50\u5AE3\u5AD7\u5AE6\u5AE9" + //  8790 -  8799
		"\u5AD6\u5AFA\u5AFB\u5B0C\u5B0B\u5B16\u5B32\u5AD0\u5B2A\u5B36" + //  8800 -  8809
		"\u5B3E\u5B43\u5B45\u5B40\u5B51\u5B55\u5B5A\u5B5B\u5B65\u5B69" + //  8810 -  8819
		"\u5B70\u5B73\u5B75\u5B78\u6588\u5B7A\u5B80\uFFFD\u9017\u5439" + //  8820 -  8829
		"\u5782\u5E25\u63A8\u6C34\u708A\u7761\u7C8B\u7FE0\u8870\u9042" + //  8830 -  8839
		"\u9154\u9310\u9318\u968F\u745E\u9AC4\u5D07\u5D69\u6570\u67A2" + //  8840 -  8849
		"\u8DA8\u96DB\u636E\u6749\u6919\u83C5\u9817\u96C0\u88FE\u6F84" + //  8850 -  8859
		"\u5978\u5981\u599D\u4F5E\u4FAB\u59A3\u59B2\u59C6\u59E8\u59DC" + //  8860 -  8869
		"\u598D\u59D9\u59DA\u5A25\u5A1F\u5A11\u5A1C\u5A09\u5A1A\u5A40" + //  8870 -  8879
		"\u5A6C\u5A49\u5A35\u5A36\u5A62\u5A6A\u5A9A\u5ABC\u5ABE\u5ACB" + //  8880 -  8889
		"\u5AC2\u5ABD\u58DF\u58EF\u58FA\u58F9\u58FB\u58FC\u58FD\u5902" + //  8890 -  8899
		"\u590A\u5910\u591B\u68A6\u5925\u592C\u592D\u5932\u5938\u593E" + //  8900 -  8909
		"\u7AD2\u5955\u5950\u594E\u595A\u5958\u5962\u5960\u5967\u596C" + //  8910 -  8919
		"\u5969\uFFFD\uFFFD\uFFFD\u5516\u5A03\u963F\u54C0\u611B\u6328" + //  8920 -  8929
		"\u59F6\u9022\u8475\u831C\u7A50\u60AA\u63E1\u6E25\u65ED\u8466" + //  8930 -  8939
		"\u82A6\u9BF5\u6893\u5727\u65A1\u6271\u5B9B\u59D0\u867B\u98F4" + //  8940 -  8949
		"\u7D62\u7DBE\u9B8E\u6216\u7C9F\u88B7\u57D6\u57E3\u580B\u5819" + //  8950 -  8959
		"\u581D\u5872\u5821\u5862\u584B\u5870\u6BC0\u5852\u583D\u5879" + //  8960 -  8969
		"\u5885\u58B9\u589F\u58AB\u58BA\u58DE\u58BB\u58B8\u58AE\u58C5" + //  8970 -  8979
		"\u58D3\u58D1\u58D7\u58D9\u58D8\u58E5\u58DC\u58E4\u570B\u570D" + //  8980 -  8989
		"\u5713\u5718\u5716\u55C7\u571C\u5726\u5737\u5738\u574E\u573B" + //  8990 -  8999
		"\u5740\u574F\u5769\u57C0\u5788\u5761\u577F\u5789\u5793\u57A0" + //  9000 -  9009
		"\u57B3\u57A4\u57AA\u57B0\u57C3\u57C6\u57D4\u57D2\u57D3\u580A" + //  9010 -  9019
		"\u566B\u5664\u562F\u566C\u566A\u5686\u5680\u568A\u56A0\u5694" + //  9020 -  9029
		"\u568F\u56A5\u56AE\u56B6\u56B4\u56C2\u56BC\u56C1\u56C3\u56C0" + //  9030 -  9039
		"\u56C8\u56CE\u56D1\u56D3\u56D7\u56EE\u56F9\u5700\u56FF\u5704" + //  9040 -  9049
		"\u5709\u5708\u557B\u557E\u5598\u559E\u55AE\u557C\u5583\u55A9" + //  9050 -  9059
		"\u5587\u55A8\u55DA\u55C5\u55DF\u55C4\u55DC\u55E4\u55D4\u5614" + //  9060 -  9069
		"\u55F7\u5616\u55FE\u55FD\u561B\u55F9\u564E\u5650\u71DF\u5634" + //  9070 -  9079
		"\u5636\u5632\u5638\uFFFD\u79E6\u7D33\u81E3\u82AF\u85AA\u89AA" + //  9080 -  9089
		"\u8A3A\u8EAB\u8F9B\u9032\u91DD\u9707\u4EBA\u4EC1\u5203\u5875" + //  9090 -  9099
		"\u58EC\u5C0B\u751A\u5C3D\u814E\u8A0A\u8FC5\u9663\u976D\u7B25" + //  9100 -  9109
		"\u8ACF\u9808\u9162\u56F3\u53A8\uFFFD\u9910\u65AC\u66AB\u6B8B" + //  9110 -  9119
		"\u4ED5\u4ED4\u4F3A\u4F7F\u523A\u53F8\u53F2\u55E3\u56DB\u58EB" + //  9120 -  9129
		"\u59CB\u59C9\u59FF\u5B50\u5C4D\u5E02\u5E2B\u5FD7\u601D\u6307" + //  9130 -  9139
		"\u652F\u5B5C\u65AF\u65BD\u65E8\u679D\u6B62\uFFFD\u5EA7\u632B" + //  9140 -  9149
		"\u50B5\u50AC\u518D\u6700\u54C9\u585E\u59BB\u5BB0\u5F69\u624D" + //  9150 -  9159
		"\u63A1\u683D\u6B73\u6E08\u707D\u91C7\u7280\u7815\u7826\u796D" + //  9160 -  9169
		"\u658E\u7D30\u83DC\u88C1\u8F09\u969B\u5264\u5728\u6750\u7F6A" + //  9170 -  9179
		"\u54AB\u54C2\u54A4\u54BE\u54BC\u54D8\u54E5\u54E6\u550F\u5514" + //  9180 -  9189
		"\u54FD\u54EE\u54ED\u54FA\u54E2\u5539\u5540\u5563\u554C\u552E" + //  9190 -  9199
		"\u555C\u5545\u5556\u5557\u5538\u5533\u555D\u5599\u5580\u54AF" + //  9200 -  9209
		"\u558A\u559F\u542D\u543C\u542E\u5436\u5429\u541D\u544E\u548F" + //  9210 -  9219
		"\u5475\u548E\u545F\u5471\u5477\u5470\u5492\u547B\u5480\u5476" + //  9220 -  9229
		"\u5484\u5490\u5486\u54C7\u54A2\u54B8\u54A5\u54AC\u54C4\u54C8" + //  9230 -  9239
		"\u54A8\uFFFD\uFFFD\uFFFD\u301D\u301F\u2116\u33CD\u2121\u32A4" + //  9240 -  9249
		"\u32A5\u32A6\u32A7\u32A8\u3231\u3232\u3239\u337E\u337D\u337C" + //  9250 -  9259
		"\u2252\u2261\u222B\u222E\u2211\u221A\u22A5\u2220\u221F\u22BF" + //  9260 -  9269
		"\u2235\u2229\u222A\uFFFD\uFFFD\uFFFD\u00F7\uFF1D\u2260\uFF1C" + //  9270 -  9279
		"\uFF1E\u2266\u2267\u221E\u2234\u2642\u2640\u00B0\u2032\u2033" + //  9280 -  9289
		"\u2103\uFFE5\uFF04\uFFE0\uFFE1\uFF05\uFF03\uFF06\uFF0A\uFF20" + //  9290 -  9299
		"\u00A7\u2606\u2605\u25CB\u25CF\u25CE\u25C7\u25C6\u4E17\u5349" + //  9300 -  9309
		"\u534D\u51D6\u535E\u5369\u536E\u5918\u537B\u5377\u5382\u5396" + //  9310 -  9319
		"\u53A0\u53A6\u53A5\u53AE\u53B0\u53B6\u53C3\u7C12\u96D9\u53DF" + //  9320 -  9329
		"\u66FC\u71EE\u53EE\u53E8\u53ED\u53FA\u5401\u543D\u5440\u542C" + //  9330 -  9339
		"\u52AC\u52AD\u52BC\u52B5\u52C1\u52CD\u52D7\u52DE\u52E3\u52E6" + //  9340 -  9349
		"\u98ED\u52E0\u52F3\u52F5\u52F8\u52F9\u5306\u5308\u7538\u530D" + //  9350 -  9359
		"\u5310\u530F\u5315\u531A\u5323\u532F\u5331\u5333\u5338\u5340" + //  9360 -  9369
		"\u5346\u5345\u51F0\u51F5\u51FE\u5204\u520B\u5214\u520E\u5227" + //  9370 -  9379
		"\u522A\u522E\u5233\u5239\u524F\u5244\u524B\u524C\u525E\u5254" + //  9380 -  9389
		"\u526A\u5274\u5269\u5273\u527F\u527D\u528D\u5294\u5292\u5271" + //  9390 -  9399
		"\u5288\u5291\u8FA8\u8FA7\u516A\u516E\u5180\u5182\u56D8\u518C" + //  9400 -  9409
		"\u5189\u518F\u5191\u5193\u5195\u5196\u51A4\u51A6\u51A2\u51A9" + //  9410 -  9419
		"\u51AA\u51AB\u51B3\u51B1\u51B2\u51B0\u51B5\u51BD\u51C5\u51C9" + //  9420 -  9429
		"\u51DB\u51E0\u8655\u51E9\u51ED\uFFFD\u62ED\u690D\u6B96\u71ED" + //  9430 -  9439
		"\u7E54\u8077\u8272\u89E6\u98DF\u8755\u8FB1\u5C3B\u4F38\u4FE1" + //  9440 -  9449
		"\u4FB5\u5507\u5A20\u5BDD\u5BE9\u5FC3\u614E\u632F\u65B0\u664B" + //  9450 -  9459
		"\u68EE\u699B\u6D78\u6DF1\u7533\u75B9\u771F\u795E\u50C9\u50CA" + //  9460 -  9469
		"\u50B3\u50C2\u50D6\u50DE\u50E5\u50ED\u50E3\u50EE\u50F9\u50F5" + //  9470 -  9479
		"\u5109\u5101\u5102\u5116\u5115\u5114\u511A\u5121\u513A\u5137" + //  9480 -  9489
		"\u513C\u513B\u513F\u5140\u5152\u514C\u5154\u5162\u7AF8\u5169" + //  9490 -  9499
		"\u5028\u5014\u502A\u5025\u5005\u4F1C\u4FF6\u5021\u5029\u502C" + //  9500 -  9509
		"\u4FFE\u4FEF\u5011\u5006\u5043\u5047\u6703\u5055\u5050\u5048" + //  9510 -  9519
		"\u505A\u5056\u506C\u5078\u5080\u509A\u5085\u50B4\u50B2\uFFFD" + //  9520 -  9529
		"\uFFFD\uFFFD\u3314\u3322\u334D\u3318\u3327\u3303\u3336\u3351" + //  9530 -  9539
		"\u3357\u330D\u3326\u3323\u332B\u334A\u333B\u339C\u339D\u339E" + //  9540 -  9549
		"\u338E\u338F\u33C4\u33A1\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  9550 -  9559
		"\uFFFD\uFFFD\u337B\uFFFD\u9418\u969C\u9798\u4E0A\u4E08\u4E1E" + //  9560 -  9569
		"\u4E57\u5197\u5270\u57CE\u5834\u58CC\u5B22\u5E38\u60C5\u64FE" + //  9570 -  9579
		"\u6761\u6756\u6D44\u72B6\u7573\u7A63\u84B8\u8B72\u91B8\u9320" + //  9580 -  9589
		"\u5631\u57F4\u98FE\uFFFD\uFFFD\uFFFD\u30C1\u30C2\u30C3\u30C4" + //  9590 -  9599
		"\u30C5\u30C6\u30C7\u30C8\u30C9\u30CA\u30CB\u30CC\u30CD\u30CE" + //  9600 -  9609
		"\u30CF\u30D0\u30D1\u30D2\u30D3\u30D4\u30D5\u30D6\u30D7\u30D8" + //  9610 -  9619
		"\u30D9\u30DA\u30DB\u30DC\u30DD\u30DE\u30DF\uFFFD\u5BDF\u62F6" + //  9620 -  9629
		"\u64AE\u64E6\u672D\u6BBA\u85A9\u96D1\u7690\u9BD6\u634C\u9306" + //  9630 -  9639
		"\u9BAB\u76BF\u6652\u4E09\u5098\u53C2\u5C71\u60E8\u6492\u6563" + //  9640 -  9649
		"\u685F\u71E6\u73CA\u7523\u7B97\u7E82\u8695\u8B83\u8CDB\u9178" + //  9650 -  9659
		"\u4EDF\u4EF7\u4F09\u4F5A\u4F30\u4F5B\u4F5D\u4F57\u4F47\u4F76" + //  9660 -  9669
		"\u4F88\u4F8F\u4F98\u4F7B\u4F69\u4F70\u4F91\u4F6F\u4F86\u4F96" + //  9670 -  9679
		"\u5118\u4FD4\u4FDF\u4FCE\u4FD8\u4FDB\u4FD1\u4FDA\u4FD0\u4FE4" + //  9680 -  9689
		"\u4FE5\u501A\u4E10\u4E15\u4E2A\u4E31\u4E36\u4E3C\u4E3F\u4E42" + //  9690 -  9699
		"\u4E56\u4E58\u4E82\u4E85\u8C6B\u4E8A\u8212\u5F0D\u4E8E\u4E9E" + //  9700 -  9709
		"\u4E9F\u4EA0\u4EA2\u4EB0\u4EB3\u4EB6\u4ECE\u4ECD\u4EC4\u4EC6" + //  9710 -  9719
		"\u4EC2\u4ED7\u4EDE\u4EED\u502D\u548C\u8A71\u6B6A\u8CC4\u8107" + //  9720 -  9729
		"\u60D1\u67A0\u9DF2\u4E99\u4E98\u9C10\u8A6B\u85C1\u8568\u6900" + //  9730 -  9739
		"\u6E7E\u7897\u8155\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  9740 -  9749
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFF10\uFF11" + //  9750 -  9759
		"\uFF12\uFF13\uFF14\uFF15\uFF16\uFF17\uFF18\uFF19\uFFFD\uFFFD" + //  9760 -  9769
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFF41\uFF42\uFF43\uFF44\uFF45" + //  9770 -  9779
		"\uFF46\uFF47\uFF48\uFF49\uFF4A\uFF4B\uFF4C\uFF4D\uFF4E\uFF4F" + //  9780 -  9789
		"\uFF50\uFF51\uFF52\uFF53\uFF54\uFF55\uFF56\uFF57\uFF58\uFF59" + //  9790 -  9799
		"\uFF5A\uFFFD\uFFFD\uFFFD\uFFFD\u3041\u84EE\u9023\u932C\u5442" + //  9800 -  9809
		"\u9B6F\u6AD3\u7089\u8CC2\u8DEF\u9732\u52B4\u5A41\u5ECA\u5F04" + //  9810 -  9819
		"\u6717\u697C\u6994\u6D6A\u6F0F\u7262\u72FC\u7BED\u8001\u807E" + //  9820 -  9829
		"\u874B\u90CE\u516D\u9E93\u7984\u808B\u9332\u8AD6\u4F36\u4F8B" + //  9830 -  9839
		"\u51B7\u52B1\u5DBA\u601C\u73B2\u793C\u82D3\u9234\u96B7\u96F6" + //  9840 -  9849
		"\u970A\u9E97\u9F62\u66A6\u6B74\u5217\u52A3\u70C8\u88C2\u5EC9" + //  9850 -  9859
		"\u604B\u6190\u6F23\u7149\u7C3E\u7DF4\u806F\uFFFD\uFFFD\uFFFD" + //  9860 -  9869
		"\u2460\u2461\u2462\u2463\u2464\u2465\u2466\u2467\u2468\u2469" + //  9870 -  9879
		"\u246A\u246B\u246C\u246D\u246E\u246F\u2470\u2471\u2472\u2473" + //  9880 -  9889
		"\u2160\u2161\u2162\u2163\u2164\u2165\u2166\u2167\u2168\u2169" + //  9890 -  9899
		"\uFFFD\u3349\u6881\u6DBC\u731F\u7642\u77AD\u7A1C\u7CE7\u826F" + //  9900 -  9909
		"\u8AD2\u907C\u91CF\u9675\u9818\u529B\u7DD1\u502B\u5398\u6797" + //  9910 -  9919
		"\u6DCB\u71D0\u7433\u81E8\u8F2A\u96A3\u9C57\u9E9F\u7460\u5841" + //  9920 -  9929
		"\u6D99\u7D2F\u985E\u4EE4\u88CF\u88E1\u91CC\u96E2\u9678\u5F8B" + //  9930 -  9939
		"\u7387\u7ACB\u844E\u63A0\u7565\u5289\u6D41\u6E9C\u7409\u7559" + //  9940 -  9949
		"\u786B\u7C92\u9686\u7ADC\u9F8D\u4FB6\u616E\u65C5\u865C\u4E86" + //  9950 -  9959
		"\u4EAE\u50DA\u4E21\u51CC\u5BEE\u6599\u6C83\u6D74\u7FCC\u7FFC" + //  9960 -  9969
		"\u6DC0\u7F85\u87BA\u88F8\u6765\u83B1\u983C\u96F7\u6D1B\u7D61" + //  9970 -  9979
		"\u843D\u916A\u4E71\u5375\u5D50\u6B04\u6FEB\u85CD\u862D\u89A7" + //  9980 -  9989
		"\u5229\u540F\u5C65\u674E\u68A8\u7406\u7483\u75E2\u8F3F\u9810" + //  9990 -  9999
		"\u50AD\u5E7C\u5996\u5BB9\u5EB8\u63DA\u63FA\u64C1\u66DC\u694A" + // 10000 - 10009
		"\u69D8\u6D0B\u6EB6\u7194\u7528\u7AAF\u7F8A\u8000\u8449\u84C9" + // 10010 - 10019
		"\u8981\u8B21\u8E0A\u9065\u967D\u990A\u617E\u6291\u6B32\uFFFD" + // 10020 - 10029
		"\u6CBC\u6D88\u6E09\u6E58\u713C\u7126\u7167\u75C7\u7701\u785D" + // 10030 - 10039
		"\u7901\u7965\u79F0\u7AE0\u7B11\u7CA7\u7D39\u8096\u83D6\u848B" + // 10040 - 10049
		"\u8549\u885D\u88F3\u8A1F\u8A3C\u8A54\u8A73\u8C61\u8CDE\u91A4" + // 10050 - 10059
		"\u9266\u937E\u8AED\u8F38\u552F\u4F51\u512A\u52C7\u53CB\u5BA5" + // 10060 - 10069
		"\u5E7D\u60A0\u6182\u63D6\u6709\u67DA\u6E67\u6D8C\u7336\u7337" + // 10070 - 10079
		"\u7531\u7950\u88D5\u8A98\u904A\u9091\u90F5\u96C4\u878D\u5915" + // 10080 - 10089
		"\u4E88\u4F59\u4E0E\u8A89\u7C7E\u8CB0\u554F\u60B6\u7D0B\u9580" + // 10090 - 10099
		"\u5301\u4E5F\u51B6\u591C\u723A\u8036\u91CE\u5F25\u77E2\u5384" + // 10100 - 10109
		"\u5F79\u7D04\u85AC\u8A33\u8E8D\u9756\u67F3\u85AE\u9453\u6109" + // 10110 - 10119
		"\u6108\u6CB9\u7652\uFFFD\uFFFD\uFFFD\u2502\u250C\u2510\u2518" + // 10120 - 10129
		"\u2514\u251C\u252C\u2524\u2534\u253C\u2501\u2503\u250F\u2513" + // 10130 - 10139
		"\u251B\u2517\u2523\u2533\u252B\u253B\u254B\u2520\u252F\u2528" + // 10140 - 10149
		"\u2537\u253F\u251D\u2530\u2525\u2538\u2542\uFFFD\u5320\u5347" + // 10150 - 10159
		"\u53EC\u54E8\u5546\u5531\u5617\u5968\u59BE\u5A3C\u5BB5\u5C06" + // 10160 - 10169
		"\u5C0F\u5C11\u5C1A\u5E84\u5E8A\u5EE0\u5F70\u627F\u6284\u62DB" + // 10170 - 10179
		"\u638C\u6377\u6607\u660C\u662D\u6676\u677E\u68A2\u6A1F\u6A35" + // 10180 - 10189
		"\u8FF7\u9298\u9CF4\u59EA\u725D\u6EC5\u514D\u68C9\u7DBF\u7DEC" + // 10190 - 10199
		"\u9762\u9EBA\u6478\u6A21\u8302\u5984\u5B5F\u6BDB\u731B\u76F2" + // 10200 - 10209
		"\u7DB2\u8017\u8499\u5132\u6728\u9ED9\u76EE\u6762\u52FF\u9905" + // 10210 - 10219
		"\u5C24\u623B"
		;


	public ByteToCharCp943() {
		super();
		super.mask1 = 0xFFE0;
		super.mask2 = 0x001F;
		super.shift = 5;
		super.leadByte = this.leadByte;
		super.singleByteToChar = this.singleByteToChar;
		super.index1 = this.index1;
		super.index2 = this.index2;
	}
}
