/*
 * @(#)ByteToCharCp949.java	1.7 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.io;

/**
* Tables and data to convert Cp949 to Unicode.
*
* @author Malcolm Ayres, assisted by UniMap program
*/
public class ByteToCharCp949
	extends ByteToCharDBCS_ASCII

{

	// Return the character set id
	public String getCharacterEncoding()
	{
		return "Cp949";
	}


	private static final boolean leadByte[] = {
		false, false, false, false, false, false, false, false,  // 00 - 07
		false, false, false, false, false, false, false, false,  // 08 - 0F
		false, false, false, false, false, false, false, false,  // 10 - 17
		false, false, false, false, false, false, false, false,  // 18 - 1F
		false, false, false, false, false, false, false, false,  // 20 - 27
		false, false, false, false, false, false, false, false,  // 28 - 2F
		false, false, false, false, false, false, false, false,  // 30 - 37
		false, false, false, false, false, false, false, false,  // 38 - 3F
		false, false, false, false, false, false, false, false,  // 40 - 47
		false, false, false, false, false, false, false, false,  // 48 - 4F
		false, false, false, false, false, false, false, false,  // 50 - 57
		false, false, false, false, false, false, false, false,  // 58 - 5F
		false, false, false, false, false, false, false, false,  // 60 - 67
		false, false, false, false, false, false, false, false,  // 68 - 6F
		false, false, false, false, false, false, false, false,  // 70 - 77
		false, false, false, false, false, false, false, false,  // 78 - 7F
		false, false, false, false, false, false, false, false,  // 80 - 87
		false, false, false, false, false, false, false, true,   // 88 - 8F
		true,  true,  true,  true,  true,  true,  true,  true,   // 90 - 97
		true,  true,  true,  true,  true,  true,  true,  true,   // 98 - 9F
		true,  true,  true,  true,  true,  true,  true,  true,   // A0 - A7
		true,  true,  true,  true,  true,  false, false, false,  // A8 - AF
		true,  true,  true,  true,  true,  true,  true,  true,   // B0 - B7
		true,  true,  true,  true,  true,  true,  true,  true,   // B8 - BF
		true,  true,  true,  true,  true,  true,  true,  true,   // C0 - C7
		true,  true,  true,  true,  true,  true,  true,  true,   // C8 - CF
		true,  true,  true,  true,  true,  true,  true,  true,   // D0 - D7
		true,  true,  true,  true,  true,  true,  true,  true,   // D8 - DF
		true,  true,  true,  true,  true,  true,  true,  true,   // E0 - E7
		true,  true,  true,  true,  true,  true,  true,  true,   // E8 - EF
		true,  true,  true,  true,  true,  true,  true,  true,   // F0 - F7
		true,  true,  true,  true,  true,  true,  true,  false,  // F8 - FF
	};


	private static final String singleByteToChar = 
		"\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007" +
		"\u0008\u0009\n\u000B\u000C\r\u000E\u000F" +
		"\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017" +
		"\u0018\u0019\u001A\u001B\u001C\u001D\u001E\u001F" +
		"\u0020\u0021\"\u0023\u0024\u0025\u0026\u0027" +
		"\u0028\u0029\u002A\u002B\u002C\u002D\u002E\u002F" +
		"\u0030\u0031\u0032\u0033\u0034\u0035\u0036\u0037" +
		"\u0038\u0039\u003A\u003B\u003C\u003D\u003E\u003F" +
		"\u0040\u0041\u0042\u0043\u0044\u0045\u0046\u0047" +
		"\u0048\u0049\u004A\u004B\u004C\u004D\u004E\u004F" +
		"\u0050\u0051\u0052\u0053\u0054\u0055\u0056\u0057" +
		"\u0058\u0059\u005A\u005B\u20A9\u005D\u005E\u005F" +
		"\u0060\u0061\u0062\u0063\u0064\u0065\u0066\u0067" +
		"\u0068\u0069\u006A\u006B\u006C\u006D\u006E\u006F" +
		"\u0070\u0071\u0072\u0073\u0074\u0075\u0076\u0077" +
		"\u0078\u0079\u007A\u007B\u007C\u007D\u007E\u007F" +
		"\u00A2\u00AC\\\u203E\u00A6\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" +
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD"
		;
	private static final short index1[] =
	{
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0000 - 00FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0100 - 01FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0200 - 02FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0300 - 03FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0400 - 04FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0500 - 05FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0600 - 06FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0700 - 07FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0800 - 08FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0900 - 09FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0A00 - 0AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0B00 - 0BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0C00 - 0CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0D00 - 0DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0E00 - 0EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 0F00 - 0FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1000 - 10FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1100 - 11FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1200 - 12FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1300 - 13FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1400 - 14FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1500 - 15FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1600 - 16FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1700 - 17FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1800 - 18FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1900 - 19FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1A00 - 1AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1B00 - 1BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1C00 - 1CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1D00 - 1DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1E00 - 1EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 1F00 - 1FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2000 - 20FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2100 - 21FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2200 - 22FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2300 - 23FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2400 - 24FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2500 - 25FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2600 - 26FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2700 - 27FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2800 - 28FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2900 - 29FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2A00 - 2AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2B00 - 2BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2C00 - 2CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2D00 - 2DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2E00 - 2EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 2F00 - 2FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3000 - 30FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3100 - 31FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3200 - 32FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3300 - 33FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3400 - 34FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3500 - 35FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3600 - 36FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3700 - 37FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3800 - 38FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3900 - 39FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3A00 - 3AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3B00 - 3BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3C00 - 3CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3D00 - 3DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3E00 - 3EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 3F00 - 3FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4000 - 40FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4100 - 41FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4200 - 42FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4300 - 43FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4400 - 44FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4500 - 45FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4600 - 46FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4700 - 47FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4800 - 48FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4900 - 49FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4A00 - 4AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4B00 - 4BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4C00 - 4CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4D00 - 4DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4E00 - 4EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 4F00 - 4FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5000 - 50FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5100 - 51FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5200 - 52FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5300 - 53FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5400 - 54FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5500 - 55FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5600 - 56FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5700 - 57FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5800 - 58FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5900 - 59FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5A00 - 5AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5B00 - 5BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5C00 - 5CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5D00 - 5DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5E00 - 5EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 5F00 - 5FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6000 - 60FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6100 - 61FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6200 - 62FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6300 - 63FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6400 - 64FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6500 - 65FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6600 - 66FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6700 - 67FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6800 - 68FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6900 - 69FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6A00 - 6AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6B00 - 6BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6C00 - 6CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6D00 - 6DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6E00 - 6EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 6F00 - 6FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7000 - 70FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7100 - 71FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7200 - 72FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7300 - 73FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7400 - 74FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7500 - 75FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7600 - 76FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7700 - 77FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7800 - 78FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7900 - 79FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7A00 - 7AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7B00 - 7BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7C00 - 7CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7D00 - 7DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7E00 - 7EFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 7F00 - 7FFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8000 - 80FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8100 - 81FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8200 - 82FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8300 - 83FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8400 - 84FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8500 - 85FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8600 - 86FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8700 - 87FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8800 - 88FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8900 - 89FF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8A00 - 8AFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8B00 - 8BFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8C00 - 8CFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8D00 - 8DFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // 8E00 - 8EFF
		 8080,  8080,  8080,  8080,  8080,  8111, 10326, 10263, // 8F00 - 8FFF
		 8080,  8080,  8080,  8080,  8080,  8396, 10231, 10168, // 9000 - 90FF
		 8080,  8080,  8080,  8080,  8080,  8011, 10136, 10073, // 9100 - 91FF
		 8080,  8080,  8080,  8080,  8080,  7536, 10041,  9978, // 9200 - 92FF
		 8080,  8080,  8080,  8080,  8080,  7726,  9946,  9883, // 9300 - 93FF
		 8080,  8080,  8080,  8080,  8080,  7631,  9851,  9788, // 9400 - 94FF
		 8080,  8080,  8080,  8080,  8080,  8933,  9756,  9693, // 9500 - 95FF
		 8080,  8080,  8080,  8080,  8080,  8807,  9661,  9598, // 9600 - 96FF
		 8080,  8080,  8080,  8080,  8080,  9059,  9566,  9503, // 9700 - 97FF
		 8080,  8080,  8080,  8080,  8080, 10294,  9471,  9408, // 9800 - 98FF
		 8080,  8080,  8080,  8080,  8080, 10199,  9376,  9313, // 9900 - 99FF
		 8080,  8080,  8080,  8080,  8080, 10104,  9281,  9218, // 9A00 - 9AFF
		 8080,  8080,  8080,  8080,  8080, 10009,  9186,  9123, // 9B00 - 9BFF
		 8080,  8080,  8080,  8080,  8080,  9914,  9091,  8997, // 9C00 - 9CFF
		 8080,  8080,  8080,  8080,  8080,  9819,  8965,  8871, // 9D00 - 9DFF
		 8080,  8080,  8080,  8080,  8080,  9724,  8839,  8745, // 9E00 - 9EFF
		 8080,  8080,  8080,  8080,  8080,  9629,  8713,  8650, // 9F00 - 9FFF
		 8080,  8080,  8080,  8080,  8080,  9534,  8618,  8555, // A000 - A0FF
		 8080,  8080,  8080,  8080,  8080,  9439,  8523,  8460, // A100 - A1FF
		 8080,  8080,  8080,  8080,  8080,  9344,  8428,  8365, // A200 - A2FF
		 8080,  8080,  8080,  8080,  8080,  9249,  8333,  8270, // A300 - A3FF
		 8080,  8080,  8080,  8080,  8080,  9154,  8238,  8175, // A400 - A4FF
		 8080,  8080,  8080,  8080,  8080,  9028,  8902,  8776, // A500 - A5FF
		 8080,  8080,  8080,  8080,  8080,  8681,  8143,  8075, // A600 - A6FF
		 8080,  8080,  8080,  8080,  8080,  8586,  8043,  7980, // A700 - A7FF
		 8080,  8080,  8080,  8080,  8080,  8491,  7948,  7885, // A800 - A8FF
		 8080,  8080,  8080,  8080,  8080,  8301,  7853,  7790, // A900 - A9FF
		 8080,  8080,  8080,  8080,  8080,  8206,  7758,  7695, // AA00 - AAFF
		 8080,  8080,  8080,  8080,  8080,  7916,  7663,  7600, // AB00 - ABFF
		 8080,  8080,  8080,  8080,  8080,  7821,  7568,  7505, // AC00 - ACFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // AD00 - ADFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // AE00 - AEFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, // AF00 - AFFF
		 8080,  8080,  8080,  8080,  8080,  7441,  7473,  7410, // B000 - B0FF
		 8080,  8080,  8080,  8080,  8080,  7346,  7378,  7315, // B100 - B1FF
		 8080,  8080,  8080,  8080,  8080,  7251,  7283,  7220, // B200 - B2FF
		 8080,  8080,  8080,  8080,  8080,  7156,  7188,  7125, // B300 - B3FF
		 8080,  8080,  8080,  8080,  8080,  7061,  7093,  7030, // B400 - B4FF
		 8080,  8080,  8080,  8080,  8080,  6966,  6998,  6935, // B500 - B5FF
		 8080,  8080,  8080,  8080,  8080,  6871,  6903,  6840, // B600 - B6FF
		 8080,  8080,  8080,  8080,  8080,  6776,  6808,  6745, // B700 - B7FF
		 8080,  8080,  8080,  8080,  8080,  6681,  6713,  6650, // B800 - B8FF
		 8080,  8080,  8080,  8080,  8080,  6586,  6618,  6555, // B900 - B9FF
		 8080,  8080,  8080,  8080,  8080,  6491,  6523,  6460, // BA00 - BAFF
		 8080,  8080,  8080,  8080,  8080,  6396,  6428,  6365, // BB00 - BBFF
		 8080,  8080,  8080,  8080,  8080,  6301,  6333,  6270, // BC00 - BCFF
		 8080,  8080,  8080,  8080,  8080,  6206,  6238,  6175, // BD00 - BDFF
		 8080,  8080,  8080,  8080,  8080,  6111,  6143,  6080, // BE00 - BEFF
		 8080,  8080,  8080,  8080,  8080,  6016,  6048,  5985, // BF00 - BFFF
		 8080,  8080,  8080,  8080,  8080,  5921,  5953,  5890, // C000 - C0FF
		 8080,  8080,  8080,  8080,  8080,  5826,  5858,  5795, // C100 - C1FF
		 8080,  8080,  8080,  8080,  8080,  5731,  5763,  5700, // C200 - C2FF
		 8080,  8080,  8080,  8080,  8080,  5636,  5668,  5605, // C300 - C3FF
		 8080,  8080,  8080,  8080,  8080,  5541,  5573,  5510, // C400 - C4FF
		 8080,  8080,  8080,  8080,  8080,  5446,  5478,  5415, // C500 - C5FF
		 8080,  8080,  8080,  8080,  8080,  5351,  5383,  5320, // C600 - C6FF
		 8080,  8080,  8080,  8080,  8080,  5256,  5288,  5225, // C700 - C7FF
		 8080,  8080,  8080,  8080,  8080,  5161,  5193,  5130, // C800 - C8FF
		 8080,  8080,  8080,  8080,  8080,  5066,  5098,  5035, // C900 - C9FF
		 8080,  8080,  8080,  8080,  8080,  4971,  5003,  4940, // CA00 - CAFF
		 8080,  8080,  8080,  8080,  8080,  4876,  4908,  4845, // CB00 - CBFF
		 8080,  8080,  8080,  8080,  8080,  4781,  4813,  4750, // CC00 - CCFF
		 8080,  8080,  8080,  8080,  8080,  4686,  4718,  4655, // CD00 - CDFF
		 8080,  8080,  8080,  8080,  8080,  4591,  4623,  4560, // CE00 - CEFF
		 8080,  8080,  8080,  8080,  8080,  4496,  4528,  4465, // CF00 - CFFF
		 8080,  8080,  8080,  8080,  8080,  4401,  4433,  4370, // D000 - D0FF
		 8080,  8080,  8080,  8080,  8080,  4306,  4338,  4275, // D100 - D1FF
		 8080,  8080,  8080,  8080,  8080,  4211,  4243,  4180, // D200 - D2FF
		 8080,  8080,  8080,  8080,  8080,  4116,  4148,  4085, // D300 - D3FF
		 8080,  8080,  8080,  8080,  8080,  4021,  4053,  3990, // D400 - D4FF
		 8080,  8080,  8080,  8080,  8080,  3926,  3958,  3895, // D500 - D5FF
		 8080,  8080,  8080,  8080,  8080,  3831,  3863,  3800, // D600 - D6FF
		 8080,  8080,  8080,  8080,  8080,  3736,  3768,  3705, // D700 - D7FF
		 8080,  8080,  8080,  8080,  8080,  3641,  3673,  3610, // D800 - D8FF
		 8080,  8080,  8080,  8080,  8080,  3546,  3578,  3515, // D900 - D9FF
		 8080,  8080,  8080,  8080,  8080,  3451,  3483,  3420, // DA00 - DAFF
		 8080,  8080,  8080,  8080,  8080,  3356,  3388,  3325, // DB00 - DBFF
		 8080,  8080,  8080,  8080,  8080,  3261,  3293,  3230, // DC00 - DCFF
		 8080,  8080,  8080,  8080,  8080,  3166,  3198,  3135, // DD00 - DDFF
		 8080,  8080,  8080,  8080,  8080,  3071,  3103,  3040, // DE00 - DEFF
		 8080,  8080,  8080,  8080,  8080,  2976,  3008,  2945, // DF00 - DFFF
		 8080,  8080,  8080,  8080,  8080,  2881,  2913,  2850, // E000 - E0FF
		 8080,  8080,  8080,  8080,  8080,  2786,  2818,  2755, // E100 - E1FF
		 8080,  8080,  8080,  8080,  8080,  2691,  2723,  2660, // E200 - E2FF
		 8080,  8080,  8080,  8080,  8080,  2596,  2628,  2565, // E300 - E3FF
		 8080,  8080,  8080,  8080,  8080,  2501,  2533,  2470, // E400 - E4FF
		 8080,  8080,  8080,  8080,  8080,  2406,  2438,  2375, // E500 - E5FF
		 8080,  8080,  8080,  8080,  8080,  2311,  2343,  2280, // E600 - E6FF
		 8080,  8080,  8080,  8080,  8080,  2216,  2248,  2185, // E700 - E7FF
		 8080,  8080,  8080,  8080,  8080,  2121,  2153,  2090, // E800 - E8FF
		 8080,  8080,  8080,  8080,  8080,  2026,  2058,  1995, // E900 - E9FF
		 8080,  8080,  8080,  8080,  8080,  1931,  1963,  1900, // EA00 - EAFF
		 8080,  8080,  8080,  8080,  8080,  1836,  1868,  1805, // EB00 - EBFF
		 8080,  8080,  8080,  8080,  8080,  1741,  1773,  1710, // EC00 - ECFF
		 8080,  8080,  8080,  8080,  8080,  1646,  1678,  1615, // ED00 - EDFF
		 8080,  8080,  8080,  8080,  8080,  1551,  1583,  1520, // EE00 - EEFF
		 8080,  8080,  8080,  8080,  8080,  1456,  1488,  1425, // EF00 - EFFF
		 8080,  8080,  8080,  8080,  8080,  1361,  1393,  1330, // F000 - F0FF
		 8080,  8080,  8080,  8080,  8080,  1266,  1298,  1235, // F100 - F1FF
		 8080,  8080,  8080,  8080,  8080,  1171,  1203,  1140, // F200 - F2FF
		 8080,  8080,  8080,  8080,  8080,  1076,  1108,  1045, // F300 - F3FF
		 8080,  8080,  8080,  8080,  8080,   981,  1013,   950, // F400 - F4FF
		 8080,  8080,  8080,  8080,  8080,   886,   918,   855, // F500 - F5FF
		 8080,  8080,  8080,  8080,  8080,   791,   823,   760, // F600 - F6FF
		 8080,  8080,  8080,  8080,  8080,   696,   728,   665, // F700 - F7FF
		 8080,  8080,  8080,  8080,  8080,   601,   633,   570, // F800 - F8FF
		 8080,  8080,  8080,  8080,  8080,   506,   538,   475, // F900 - F9FF
		 8080,  8080,  8080,  8080,  8080,   411,   443,   380, // FA00 - FAFF
		 8080,  8080,  8080,  8080,  8080,   316,   348,   285, // FB00 - FBFF
		 8080,  8080,  8080,  8080,  8080,   221,   253,   190, // FC00 - FCFF
		 8080,  8080,  8080,  8080,  8080,   126,   158,    95, // FD00 - FDFF
		 8080,  8080,  8080,  8080,  8080,    31,    63,     0, // FE00 - FEFF
		 8080,  8080,  8080,  8080,  8080,  8080,  8080,  8080, 
	};

	private static final String index2 =
		"\uE09D\uE09E\uE09F\uE0A0\uE0A1\uE0A2\uE0A3\uE0A4\uE0A5\uE0A6" + //     0 -     9
		"\uE0A7\uE0A8\uE0A9\uE0AA\uE0AB\uE0AC\uE0AD\uE0AE\uE0AF\uE0B0" + //    10 -    19
		"\uE0B1\uE0B2\uE0B3\uE0B4\uE0B5\uE0B6\uE0B7\uE0B8\uE0B9\uE0BA" + //    20 -    29
		"\uE0BB\uFFFD\uE05E\uE05F\uE060\uE061\uE062\uE063\uE064\uE065" + //    30 -    39
		"\uE066\uE067\uE068\uE069\uE06A\uE06B\uE06C\uE06D\uE06E\uE06F" + //    40 -    49
		"\uE070\uE071\uE072\uE073\uE074\uE075\uE076\uE077\uE078\uE079" + //    50 -    59
		"\uE07A\uE07B\uE07C\uE07D\uE07E\uE07F\uE080\uE081\uE082\uE083" + //    60 -    69
		"\uE084\uE085\uE086\uE087\uE088\uE089\uE08A\uE08B\uE08C\uE08D" + //    70 -    79
		"\uE08E\uE08F\uE090\uE091\uE092\uE093\uE094\uE095\uE096\uE097" + //    80 -    89
		"\uE098\uE099\uE09A\uE09B\uE09C\u7D07\u8A16\u6B20\u6B3D\u6B46" + //    90 -    99
		"\u5438\u6070\u6D3D\u7FD5\u8208\u50D6\u51DE\u559C\u566B\u56CD" + //   100 -   109
		"\u59EC\u5B09\u5E0C\u6199\u6198\u6231\u665E\u66E6\u7199\u71B9" + //   110 -   119
		"\u71BA\u72A7\u79A7\u7A00\u7FB2\u8A70\uFFFD\u723B\u80B4\u9175" + //   120 -   129
		"\u9A4D\u4FAF\u5019\u539A\u540E\u543C\u5589\u55C5\u5E3F\u5F8C" + //   130 -   139
		"\u673D\u7166\u73DD\u9005\u52DB\u52F3\u5864\u58CE\u7104\u718F" + //   140 -   149
		"\u71FB\u85B0\u8A13\u6688\u85A8\u55A7\u6684\u714A\u8431\u5349" + //   150 -   159
		"\u5599\u6BC1\u5F59\u5FBD\u63EE\u6689\u7147\u8AF1\u8F1D\u9EBE" + //   160 -   169
		"\u4F11\u643A\u70CB\u7566\u8667\u6064\u8B4E\u9DF8\u5147\u51F6" + //   170 -   179
		"\u5308\u6D36\u80F8\u9ED1\u6615\u6B23\u7098\u75D5\u5403\u5C79" + //   180 -   189
		"\u5F8A\u6062\u6094\u61F7\u6666\u6703\u6A9C\u6DEE\u6FAE\u7070" + //   190 -   199
		"\u736A\u7E6A\u81BE\u8334\u86D4\u8AA8\u8CC4\u5283\u7372\u5B96" + //   200 -   209
		"\u6A6B\u9404\u54EE\u5686\u5B5D\u6548\u6585\u66C9\u689F\u6D8D" + //   210 -   219
		"\u6DC6\uFFFD\u798D\u79BE\u82B1\u83EF\u8A71\u8B41\u8CA8\u9774" + //   220 -   229
		"\uFA0B\u64F4\u652B\u78BA\u78BB\u7A6B\u4E38\u559A\u5950\u5BA6" + //   230 -   239
		"\u5E7B\u60A3\u63DB\u6B61\u6665\u6853\u6E19\u7165\u74B0\u7D08" + //   240 -   249
		"\u9084\u9A69\u9C25\u6D3B\u6ED1\u733E\u8C41\u95CA\u51F0\u5E4C" + //   250 -   259
		"\u5FA8\u604D\u60F6\u6130\u614C\u6643\u6644\u69A5\u6CC1\u6E5F" + //   260 -   269
		"\u6EC9\u6F62\u714C\u749C\u7687\u7BC1\u7C27\u8352\u8757\u9051" + //   270 -   279
		"\u968D\u9EC3\u532F\u56DE\u5EFB\u93AC\u9800\u9865\u60D1\u6216" + //   280 -   289
		"\u9177\u5A5A\u660F\u6DF7\u6E3E\u743F\u9B42\u5FFD\u60DA\u7B0F" + //   290 -   299
		"\u54C4\u5F18\u6C5E\u6CD3\u6D2A\u70D8\u7D05\u8679\u8A0C\u9D3B" + //   300 -   309
		"\u5316\u548C\u5B05\u6A3A\u706B\u7575\uFFFD\u5F62\u6CC2\u6ECE" + //   310 -   319
		"\u7005\u7050\u70AF\u7192\u73E9\u7469\u834A\u87A2\u8861\u9008" + //   320 -   329
		"\u90A2\u93A3\u99A8\u516E\u5F57\u60E0\u6167\u66B3\u8559\u8E4A" + //   330 -   339
		"\u91AF\u978B\u4E4E\u4E92\u547C\u58D5\u58FA\u597D\u5CB5\u5F27" + //   340 -   349
		"\u6236\u6248\u660A\u6667\u6BEB\u6D69\u6DCF\u6E56\u6EF8\u6F94" + //   350 -   359
		"\u6FE0\u6FE9\u705D\u72D0\u7425\u745A\u74E0\u7693\u795C\u7CCA" + //   360 -   369
		"\u7E1E\u80E1\u82A6\u846B\u84BF\u864E\u865F\u8774\u8B77\u8C6A" + //   370 -   379
		"\u774D\u7D43\u7D62\u7E23\u8237\u8852\uFA0A\u8CE2\u9249\u986F" + //   380 -   389
		"\u5B51\u7A74\u8840\u9801\u5ACC\u4FE0\u5354\u593E\u5CFD\u633E" + //   390 -   399
		"\u6D79\u72F9\u8105\u8107\u83A2\u92CF\u9830\u4EA8\u5144\u5211" + //   400 -   409
		"\u578B\uFFFD\uFA08\uFA09\u9805\u4EA5\u5055\u54B3\u5793\u595A" + //   410 -   419
		"\u5B69\u5BB3\u61C8\u6977\u6D77\u7023\u87F9\u89E3\u8A72\u8AE7" + //   420 -   429
		"\u9082\u99ED\u9AB8\u52BE\u6838\u5016\u5E78\u674F\u8347\u884C" + //   430 -   439
		"\u4EAB\u5411\u56AE\u73E6\u9115\u97FF\u9909\u9957\u9999\u5653" + //   440 -   449
		"\u589F\u865B\u8A31\u61B2\u6AF6\u737B\u8ED2\u6B47\u96AA\u9A57" + //   450 -   459
		"\u5955\u7200\u8D6B\u9769\u4FD4\u5CF4\u5F26\u61F8\u665B\u6CEB" + //   460 -   469
		"\u70AB\u7384\u73B9\u73FE\u7729\u54B8\u5563\u558A\u6ABB\u6DB5" + //   470 -   479
		"\u7DD8\u8266\u929C\u9677\u9E79\u5408\u54C8\u76D2\u86E4\u95A4" + //   480 -   489
		"\u95D4\u965C\u4EA2\u4F09\u59EE\u5AE6\u5DF7\u6052\u6297\u676D" + //   490 -   499
		"\u6841\u6C86\u6E2F\u7F38\u809B\u822A\uFFFD\u54C1\u7A1F\u6953" + //   500 -   509
		"\u8AF7\u8C4A\u98A8\u99AE\u5F7C\u62AB\u75B2\u76AE\u88AB\u907F" + //   510 -   519
		"\u9642\u5339\u5F3C\u5FC5\u6CCC\u73CC\u7562\u758B\u7B46\u82FE" + //   520 -   529
		"\u999D\u4E4F\u903C\u4E0B\u4F55\u53A6\u590F\u5EC8\u6630\u6CB3" + //   530 -   539
		"\u7455\u8377\u8766\u8CC0\u9050\u971E\u9C15\u58D1\u5B78\u8650" + //   540 -   549
		"\u8B14\u9DB4\u5BD2\u6068\u608D\u65F1\u6C57\u6F22\u6FA3\u701A" + //   550 -   559
		"\u7F55\u7FF0\u9591\u9592\u9650\u97D3\u5272\u8F44\u51FD\u542B" + //   560 -   569
		"\u80DE\u812F\u82DE\u8461\u84B2\u888D\u8912\u900B\u92EA\u98FD" + //   570 -   579
		"\u9B91\u5E45\u66B4\u66DD\u7011\u7206\uFA07\u4FF5\u527D\u5F6A" + //   580 -   589
		"\u6153\u6753\u6A19\u6F02\u74E2\u7968\u8868\u8C79\u98C7\u98C4" + //   590 -   599
		"\u9A43\uFFFD\u962A\u516B\u53ED\u634C\u4F69\u5504\u6096\u6557" + //   600 -   609
		"\u6C9B\u6D7F\u724C\u72FD\u7A17\u8987\u8C9D\u5F6D\u6F8E\u70F9" + //   610 -   619
		"\u81A8\u610E\u4FBF\u504F\u6241\u7247\u7BC7\u7DE8\u7FE9\u904D" + //   620 -   629
		"\u97AD\u9A19\u8CB6\u576A\u5E73\u67B0\u840D\u8A55\u5420\u5B16" + //   630 -   639
		"\u5E63\u5EE2\u5F0A\u6583\u80BA\u853D\u9589\u965B\u4F48\u5305" + //   640 -   649
		"\u530D\u530F\u5486\u54FA\u5703\u5E03\u6016\u629B\u62B1\u6355" + //   650 -   659
		"\uFA06\u6CE1\u6D66\u75B1\u7832\u59AC\u6295\u900F\u9B2A\u615D" + //   660 -   669
		"\u7279\u95D6\u5761\u5A46\u5DF4\u628A\u64AD\u64FA\u6777\u6CE2" + //   670 -   679
		"\u6D3E\u722C\u7436\u7834\u7F77\u82AD\u8DDB\u9817\u5224\u5742" + //   680 -   689
		"\u677F\u7248\u74E3\u8CA9\u8FA6\u9211\uFFFD\u9438\u5451\u5606" + //   690 -   699
		"\u5766\u5F48\u619A\u6B4E\u7058\u70AD\u7DBB\u8A95\u596A\u812B" + //   700 -   709
		"\u63A2\u7708\u803D\u8CAA\u5854\u642D\u69BB\u5B95\u5E11\u6E6F" + //   710 -   719
		"\uFA03\u8569\u514C\u53F0\u592A\u6020\u614B\u6B86\u6C70\u6CF0" + //   720 -   729
		"\u7B1E\u80CE\u82D4\u8DC6\u90B0\u98B1\uFA04\u64C7\u6FA4\u6491" + //   730 -   739
		"\u6504\u514E\u5410\u571F\u8A0E\u615F\u6876\uFA05\u75DB\u7B52" + //   740 -   749
		"\u7D71\u901A\u5806\u69CC\u817F\u892A\u9000\u9839\u5078\u5957" + //   750 -   759
		"\u7A31\u5FEB\u4ED6\u54A4\u553E\u58AE\u59A5\u60F0\u6253\u62D6" + //   760 -   769
		"\u6736\u6955\u8235\u9640\u99B1\u99DD\u502C\u5353\u5544\u577C" + //   770 -   779
		"\uFA01\u6258\uFA02\u64E2\u666B\u67DD\u6FC1\u6FEF\u7422\u7438" + //   780 -   789
		"\u8A17\uFFFD\u8D05\u53D6\u5439\u5634\u5A36\u5C31\u708A\u7FE0" + //   790 -   799
		"\u805A\u8106\u81ED\u8DA3\u9189\u9A5F\u9DF2\u5074\u4EC4\u53A0" + //   800 -   809
		"\u60FB\u6E2C\u5C64\u4F88\u5024\u55E4\u5CD9\u5E5F\u6065\u6894" + //   810 -   819
		"\u6CBB\u6DC4\u71BE\u75D4\u75F4\u7661\u7A1A\u7A49\u7DC7\u7DFB" + //   820 -   829
		"\u7F6E\u81F4\u86A9\u8F1C\u96C9\u99B3\u9F52\u5247\u52C5\u98ED" + //   830 -   839
		"\u89AA\u4E03\u67D2\u6F06\u4FB5\u5BE2\u6795\u6C88\u6D78\u741B" + //   840 -   849
		"\u7827\u91DD\u937C\u87C4\u79E4\u939A\u96DB\u9A36\u9C0D\u4E11" + //   850 -   859
		"\u755C\u795D\u7AFA\u7B51\u7BC9\u7E2E\u84C4\u8E59\u8E74\u8EF8" + //   860 -   869
		"\u9010\u6625\u693F\u7443\u51FA\u672E\u9EDC\u5145\u5FE0\u6C96" + //   870 -   879
		"\u87F2\u885D\u8877\u60B4\u81B5\u8403\uFFFD\u6912\u695A\u6A35" + //   880 -   889
		"\u7092\u7126\u785D\u7901\u790E\u79D2\u7A0D\u8096\u8278\u82D5" + //   890 -   899
		"\u8349\u8549\u8C82\u8D85\u9162\u918B\u91AE\u4FC3\u56D1\u71ED" + //   900 -   909
		"\u77D7\u8700\u89F8\u5BF8\u5FD6\u6751\u90A8\u53E2\u585A\u5BF5" + //   910 -   919
		"\u60A4\u6181\u6460\u7E3D\u8070\u8525\u9283\u64AE\u50AC\u5D14" + //   920 -   929
		"\u6700\u589C\u62BD\u63A8\u690E\u6978\u6A1E\u6E6B\u76BA\u79CB" + //   930 -   939
		"\u82BB\u8429\u8ACF\u8DA8\u8FFD\u9112\u914B\u919C\u9310\u9318" + //   940 -   949
		"\u7252\u758A\u776B\u8ADC\u8CBC\u8F12\u5EF3\u6674\u6DF8\u807D" + //   950 -   959
		"\u83C1\u8ACB\u9751\u9BD6\uFA00\u5243\u66FF\u6D95\u6EEF\u7DE0" + //   960 -   969
		"\u8AE6\u902E\u905E\u9AD4\u521D\u527F\u54E8\u6194\u6284\u62DB" + //   970 -   979
		"\u68A2\uFFFD\u8CAC\u51C4\u59BB\u60BD\u8655\u501C\uF9FF\u5254" + //   980 -   989
		"\u5C3A\u617D\u621A\u62D3\u64F2\u65A5\u6ECC\u7620\u810A\u8E60" + //   990 -   999
		"\u965F\u96BB\u4EDF\u5343\u5598\u5929\u5DDD\u64C5\u6CC9\u6DFA" + //  1000 -  1009
		"\u7394\u7A7F\u821B\u85A6\u8CE4\u8E10\u9077\u91E7\u95E1\u9621" + //  1010 -  1019
		"\u97C6\u51F8\u54F2\u5586\u5FB9\u64A4\u6F88\u7DB4\u8F1F\u8F4D" + //  1020 -  1029
		"\u9435\u50C9\u5C16\u6CBE\u6DFB\u751B\u77BB\u7C3D\u7C64\u8A79" + //  1030 -  1039
		"\u8AC2\u581E\u59BE\u5E16\u6377\u5F70\u6134\u655E\u660C\u6636" + //  1040 -  1049
		"\u66A2\u69CD\u6EC4\u6F32\u7316\u7621\u7A93\u8139\u8259\u83D6" + //  1050 -  1059
		"\u84BC\u50B5\u57F0\u5BC0\u5BE8\u5F69\u63A1\u7826\u7DB5\u83DC" + //  1060 -  1069
		"\u8521\u91C7\u91F5\u518A\u67F5\u7B56\uFFFD\u93F6\u96C6\u5FB5" + //  1070 -  1079
		"\u61F2\u6F84\u4E14\u4F98\u501F\u53C9\u55DF\u5D6F\u5DEE\u6B21" + //  1080 -  1089
		"\u6B64\u78CB\u7B9A\uF9FE\u8E49\u8ECA\u906E\u6349\u643E\u7740" + //  1090 -  1099
		"\u7A84\u932F\u947F\u9F6A\u64B0\u6FAF\u71E6\u74A8\u74DA\u7AC4" + //  1100 -  1109
		"\u7C12\u7E82\u7CB2\u7E98\u8B9A\u8D0A\u947D\u9910\u994C\u5239" + //  1110 -  1119
		"\u5BDF\u64E6\u672D\u7D2E\u50ED\u53C3\u5879\u6158\u6159\u61FA" + //  1120 -  1129
		"\u65AC\u7AD9\u8B92\u8B96\u5009\u5021\u5275\u5531\u5A3C\u5EE0" + //  1130 -  1139
		"\u8A3A\u8CD1\u8EEB\u8FB0\u9032\u93AD\u9663\u9673\u9707\u4F84" + //  1140 -  1149
		"\u53F1\u59EA\u5AC9\u5E19\u684E\u74C6\u75BE\u79E9\u7A92\u81A3" + //  1150 -  1159
		"\u86ED\u8CEA\u8DCC\u8FED\u659F\u6715\uF9FD\u57F7\u6F57\u7DDD" + //  1160 -  1169
		"\u8F2F\uFFFD\u54AB\u5730\u5740\u5FD7\u6301\u6307\u646F\u652F" + //  1170 -  1179
		"\u65E8\u667A\u679D\u67B3\u6B62\u6C60\u6C9A\u6F2C\u77E5\u7825" + //  1180 -  1189
		"\u7949\u7957\u7D19\u80A2\u8102\u81F3\u829D\u82B7\u8718\u8A8C" + //  1190 -  1199
		"\uF9FC\u8D04\u8DBE\u9072\u76F4\u7A19\u7A37\u7E54\u8077\u5507" + //  1200 -  1209
		"\u55D4\u5875\u632F\u6422\u6649\u664B\u686D\u699B\u6B84\u6D25" + //  1210 -  1219
		"\u6EB1\u73CD\u7468\u74A1\u755B\u75B9\u76E1\u771E\u778B\u79E6" + //  1220 -  1229
		"\u7E09\u7E1D\u81FB\u852F\u8897\u710C\u756F\u7AE3\u8822\u9021" + //  1230 -  1239
		"\u9075\u96CB\u99FF\u8301\u4E2D\u4EF2\u8846\u91CD\u537D\u6ADB" + //  1240 -  1249
		"\u696B\u6C41\u847A\u589E\u618E\u66FE\u62EF\u70DD\u7511\u75C7" + //  1250 -  1259
		"\u7E52\u84B8\u8B49\u8D08\u4E4B\u53EA\uFFFD\u8E2A\u8E35\u937E" + //  1260 -  1269
		"\u9418\u4F50\u5750\u5DE6\u5EA7\u632B\u7F6A\u4E3B\u4F4F\u4F8F" + //  1270 -  1279
		"\u505A\u59DD\u80C4\u546A\u5468\u55FE\u594F\u5B99\u5DDE\u5EDA" + //  1280 -  1289
		"\u665D\u6731\u67F1\u682A\u6CE8\u6D32\u6E4A\u6F8D\u70B7\u73E0" + //  1290 -  1299
		"\u7587\u7C4C\u7D02\u7D2C\u7DA2\u821F\u86DB\u8A3B\u8A85\u8D70" + //  1300 -  1309
		"\u8E8A\u8F33\u9031\u914E\u9152\u9444\u99D0\u7AF9\u7CA5\u4FCA" + //  1310 -  1319
		"\u5101\u51C6\u57C8\u5BEF\u5CFB\u6659\u6A3D\u6D5A\u6E96\u6FEC" + //  1320 -  1329
		"\u8ABF\u8D99\u8E81\u9020\u906D\u91E3\u963B\u96D5\u9CE5\u65CF" + //  1330 -  1339
		"\u7C07\u8DB3\u93C3\u5B58\u5C0A\u5352\u62D9\u731D\u5027\u5B97" + //  1340 -  1349
		"\u5F9E\u60B0\u616B\u68D5\u6DD9\u742E\u7A2E\u7D42\u7D9C\u7E31" + //  1350 -  1359
		"\u816B\uFFFD\u975C\u9802\u9F0E\u5236\u5291\u557C\u5824\u5E1D" + //  1360 -  1369
		"\u5F1F\u608C\u63D0\u68AF\u6FDF\u796D\u7B2C\u81CD\u85BA\u88FD" + //  1370 -  1379
		"\u8AF8\u8E44\u918D\u9664\u969B\u973D\u984C\u9F4A\u4FCE\u5146" + //  1380 -  1389
		"\u51CB\u52A9\u5632\u5F14\u5F6B\u63AA\u64CD\u65E9\u6641\u66FA" + //  1390 -  1399
		"\u66F9\u671D\u689D\u68D7\u69FD\u6F15\u6F6E\u7167\u71E5\u722A" + //  1400 -  1409
		"\u74AA\u773A\u7956\u795A\u79DF\u7A20\u7A95\u7C97\u7CDF\u7D44" + //  1410 -  1419
		"\u7E70\u8087\u85FB\u86A4\u8A54\u6A89\u6B63\u6C40\u6DC0\u6DE8" + //  1420 -  1429
		"\u6E1F\u6E5E\u701E\u70A1\u738E\u73FD\u753A\u775B\u7887\u798E" + //  1430 -  1439
		"\u7A0B\u7A7D\u7CBE\u7D8E\u8247\u8A02\u8AEA\u8C9E\u912D\u914A" + //  1440 -  1449
		"\u91D8\u9266\u92CC\u9320\u9706\u9756\uFFFD\u714E\u7420\u7530" + //  1450 -  1459
		"\u7538\u7551\u7672\u7B4C\u7B8B\u7BAD\u7BC6\u7E8F\u8A6E\u8F3E" + //  1460 -  1469
		"\u8F49\u923F\u9293\u9322\u942B\u96FB\u985A\u986B\u991E\u5207" + //  1470 -  1479
		"\u622A\u6298\u6D59\u7664\u7ACA\u7BC0\u7D76\u5360\u5CBE\u5E97" + //  1480 -  1489
		"\u6F38\u70B9\u7C98\u9711\u9B8E\u9EDE\u63A5\u647A\u8776\u4E01" + //  1490 -  1499
		"\u4E95\u4EAD\u505C\u5075\u5448\u59C3\u5B9A\u5E40\u5EAD\u5EF7" + //  1500 -  1509
		"\u5F81\u60C5\u633A\u653F\u6574\u65CC\u6676\u6678\u67FE\u6968" + //  1510 -  1519
		"\u7E3E\u7FDF\u837B\u8B2B\u8CCA\u8D64\u8DE1\u8E5F\u8FEA\u8FF9" + //  1520 -  1529
		"\u9069\u93D1\u4F43\u4F7A\u50B3\u5168\u5178\u524D\u526A\u5861" + //  1530 -  1539
		"\u587C\u5960\u5C08\u5C55\u5EDB\u609B\u6230\u6813\u6BBF\u6C08" + //  1540 -  1549
		"\u6FB1\uFFFD\u969C\u518D\u54C9\u5728\u5BB0\u624D\u6750\u683D" + //  1550 -  1559
		"\u6893\u6E3D\u6ED3\u707D\u7E21\u88C1\u8CA1\u8F09\u9F4B\u9F4E" + //  1560 -  1569
		"\u722D\u7B8F\u8ACD\u931A\u4F47\u4F4E\u5132\u5480\u59D0\u5E95" + //  1570 -  1579
		"\u62B5\u6775\u696E\u6A17\u6CAE\u6E1A\u72D9\u732A\u75BD\u7BB8" + //  1580 -  1589
		"\u7D35\u82E7\u83F9\u8457\u85F7\u8A5B\u8CAF\u8E87\u9019\u90B8" + //  1590 -  1599
		"\u96CE\u9F5F\u52E3\u540A\u5AE1\u5BC2\u6458\u6575\u6EF4\u72C4" + //  1600 -  1609
		"\uF9FB\u7684\u7A4D\u7B1B\u7C4D\u58EF\u596C\u5C07\u5E33\u5E84" + //  1610 -  1619
		"\u5F35\u638C\u66B2\u6756\u6A1F\u6AA3\u6B0C\u6F3F\u7246\uF9FA" + //  1620 -  1629
		"\u7350\u748B\u7AE0\u7CA7\u8178\u81DF\u81E7\u838A\u846C\u8523" + //  1630 -  1639
		"\u8594\u85CF\u88DD\u8D13\u91AC\u9577\uFFFD\uF9F7\uF9F8\uF9F9" + //  1640 -  1649
		"\u4ECD\u5269\u5B55\u82BF\u4ED4\u523A\u54A8\u59C9\u59FF\u5B50" + //  1650 -  1659
		"\u5B57\u5B5C\u6063\u6148\u6ECB\u7099\u716E\u7386\u74F7\u75B5" + //  1660 -  1669
		"\u78C1\u7D2B\u8005\u81EA\u8328\u8517\u85C9\u8AEE\u8CC7\u96CC" + //  1670 -  1679
		"\u4F5C\u52FA\u56BC\u65AB\u6628\u707C\u70B8\u7235\u7DBD\u828D" + //  1680 -  1689
		"\u914C\u96C0\u9D72\u5B71\u68E7\u6B98\u6F7A\u76DE\u5C91\u66AB" + //  1690 -  1699
		"\u6F5B\u7BB4\u7C2A\u8836\u96DC\u4E08\u4ED7\u5320\u5834\u58BB" + //  1700 -  1709
		"\u8335\uF9F0\u8693\u8A8D\uF9F1\u976D\u9777\uF9F2\uF9F3\u4E00" + //  1710 -  1719
		"\u4F5A\u4F7E\u58F9\u65E5\u6EA2\u9038\u93B0\u99B9\u4EFB\u58EC" + //  1720 -  1729
		"\u598A\u59D9\u6041\uF9F4\uF9F5\u7A14\uF9F6\u834F\u8CC3\u5165" + //  1730 -  1739
		"\u5344\uFFFD\u8B70\u91AB\u4E8C\u4EE5\u4F0A\uF9DD\uF9DE\u5937" + //  1740 -  1749
		"\u59E8\uF9DF\u5DF2\u5F1B\u5F5B\u6021\uF9E0\uF9E1\uF9E2\uF9E3" + //  1750 -  1759
		"\u723E\u73E5\uF9E4\u7570\u75CD\uF9E5\u79FB\uF9E6\u800C\u8033" + //  1760 -  1769
		"\u8084\u82E1\u8351\uF9E7\uF9E8\u8CBD\u8CB3\u9087\uF9E9\uF9EA" + //  1770 -  1779
		"\u98F4\u990C\uF9EB\uF9EC\u7037\u76CA\u7FCA\u7FCC\u7FFC\u8B1A" + //  1780 -  1789
		"\u4EBA\u4EC1\u5203\u5370\uF9ED\u54BD\u56E0\u59FB\u5BC5\u5F15" + //  1790 -  1799
		"\u5FCD\u6E6E\uF9EE\uF9EF\u7D6A\u4E59\u541F\u6DEB\u852D\u9670" + //  1800 -  1809
		"\u97F3\u98EE\u63D6\u6CE3\u9091\u51DD\u61C9\u81BA\u9DF9\u4F9D" + //  1810 -  1819
		"\u501A\u5100\u5B9C\u610F\u61FF\u64EC\u6905\u6BC5\u7591\u77E3" + //  1820 -  1829
		"\u7FA9\u8264\u858F\u87FB\u8863\u8ABC\uFFFD\u6FE1\u7336\u7337" + //  1830 -  1839
		"\uF9CC\u745C\u7531\uF9CD\u7652\uF9CE\uF9CF\u7DAD\u81FE\u8438" + //  1840 -  1849
		"\u88D5\u8A98\u8ADB\u8AED\u8E30\u8E42\u904A\u903E\u907A\u9149" + //  1850 -  1859
		"\u91C9\u936E\uF9D0\uF9D1\u5809\uF9D2\u6BD3\u8089\u80B2\uF9D3" + //  1860 -  1869
		"\uF9D4\u5141\u596B\u5C39\uF9D5\uF9D6\u6F64\u73A7\u80E4\u8D07" + //  1870 -  1879
		"\uF9D7\u9217\u958F\uF9D8\uF9D9\uF9DA\uF9DB\u807F\u620E\u701C" + //  1880 -  1889
		"\u7D68\u878D\uF9DC\u57A0\u6069\u6147\u6BB7\u8ABE\u9280\u96B1" + //  1890 -  1899
		"\u9B4F\u4E73\u4F91\u5112\u516A\uF9C7\u552F\u55A9\u5B7A\u5BA5" + //  1900 -  1909
		"\u5E7C\u5E7D\u5EBE\u60A0\u60DF\u6108\u6109\u63C4\u6538\u6709" + //  1910 -  1919
		"\uF9C8\u67D4\u67DA\uF9C9\u6961\u6962\u6CB9\u6D27\uF9CA\u6E38" + //  1920 -  1929
		"\uF9CB\uFFFD\u904B\u9695\u96F2\u97FB\u851A\u9B31\u4E90\u718A" + //  1930 -  1939
		"\u96C4\u5143\u539F\u54E1\u5713\u5712\u57A3\u5A9B\u5AC4\u5BC3" + //  1940 -  1949
		"\u6028\u613F\u63F4\u6C85\u6D39\u6E72\u6E90\u7230\u733F\u7457" + //  1950 -  1959
		"\u82D1\u8881\u8F45\u9060\uF9C6\u9662\u9858\u9D1B\u6708\u8D8A" + //  1960 -  1969
		"\u925E\u4F4D\u5049\u50DE\u5371\u570D\u59D4\u5A01\u5C09\u6170" + //  1970 -  1979
		"\u6690\u6E2D\u7232\u744B\u7DEF\u80C3\u840E\u8466\u853F\u875F" + //  1980 -  1989
		"\u885B\u8918\u8B02\u9055\u97CB\u79B9\u7D06\u7FBD\u828B\u85D5" + //  1990 -  1999
		"\u865E\u8FC2\u9047\u90F5\u91EA\u9685\u96E8\u96E9\u52D6\u5F67" + //  2000 -  2009
		"\u65ED\u6631\u682F\u715C\u7A36\u90C1\u980A\u4E91\uF9C5\u6A52" + //  2010 -  2019
		"\u6B9E\u6F90\u7189\u8018\u82B8\u8553\uFFFD\u7A88\u7AAF\u7E47" + //  2020 -  2029
		"\u7E5E\u8000\u8170\uF9C2\u87EF\u8981\u8B20\u9059\uF9C3\u9080" + //  2030 -  2039
		"\u9952\u617E\u6B32\u6D74\u7E1F\u8925\u8FB1\u4FD1\u50AD\u5197" + //  2040 -  2049
		"\u52C7\u57C7\u5889\u5BB9\u5EB8\u6142\u6995\u6D8C\u6E67\u6EB6" + //  2050 -  2059
		"\u7194\u7462\u7528\u752C\u8073\u8338\u84C9\u8E0A\u9394\u93DE" + //  2060 -  2069
		"\uF9C4\u4E8E\u4F51\u5076\u512A\u53C8\u53CB\u53F3\u5B87\u5BD3" + //  2070 -  2079
		"\u5C24\u611A\u6182\u65F4\u725B\u7397\u7440\u76C2\u7950\u7991" + //  2080 -  2089
		"\u6B6A\u77EE\u5916\u5D6C\u5DCD\u7325\u754F\uF9BA\uF9BB\u50E5" + //  2090 -  2099
		"\u51F9\u582F\u592D\u5996\u59DA\u5BE5\uF9BC\uF9BD\u5DA2\u62D7" + //  2100 -  2109
		"\u6416\u6493\u64FE\uF9BE\u66DC\uF9BF\u6A48\uF9C0\u71FF\u7464" + //  2110 -  2119
		"\uF9C1\uFFFD\u70CF\u71AC\u7352\u7B7D\u8708\u8AA4\u9C32\u9F07" + //  2120 -  2129
		"\u5C4B\u6C83\u7344\u7389\u923A\u6EAB\u7465\u761F\u7A69\u7E15" + //  2130 -  2139
		"\u860A\u5140\u58C5\u64C1\u74EE\u7515\u7670\u7FC1\u9095\u96CD" + //  2140 -  2149
		"\u9954\u6E26\u74E6\u7AA9\u7AAA\u81E5\u86D9\u8778\u8A1B\u5A49" + //  2150 -  2159
		"\u5B8C\u5B9B\u68A1\u6900\u6D63\u73A9\u7413\u742C\u7897\u7DE9" + //  2160 -  2169
		"\u7FEB\u8118\u8155\u839E\u8C4C\u962E\u9811\u66F0\u5F80\u65FA" + //  2170 -  2179
		"\u6789\u6C6A\u738B\u502D\u5A03\u88D4\u8A63\u8B7D\u8C6B\uF9B7" + //  2180 -  2189
		"\u92B3\uF9B8\u9713\u9810\u4E94\u4F0D\u4FC9\u50B2\u5348\u543E" + //  2190 -  2199
		"\u5433\u55DA\u5862\u58BA\u5967\u5A1B\u5BE4\u609F\uF9B9\u61CA" + //  2200 -  2209
		"\u6556\u65FF\u6664\u68A7\u6C5A\u6FB3\uFFFD\uF9A6\u95BB\u9AE5" + //  2210 -  2219
		"\u9E7D\u66C4\uF9A7\u71C1\u8449\uF9A8\uF9A9\u584B\uF9AA\uF9AB" + //  2220 -  2229
		"\u5DB8\u5F71\uF9AC\u6620\u668E\u6979\u69AE\u6C38\u6CF3\u6E36" + //  2230 -  2239
		"\u6F41\u6FDA\u701B\u702F\u7150\u71DF\u7370\uF9AD\u745B\uF9AE" + //  2240 -  2249
		"\u74D4\u76C8\u7A4E\u7E93\uF9AF\uF9B0\u82F1\u8A60\u8FCE\uF9B1" + //  2250 -  2259
		"\u9348\uF9B2\u9719\uF9B3\uF9B4\u4E42\u502A\uF9B5\u5208\u53E1" + //  2260 -  2269
		"\u66F3\u6C6D\u6FCA\u730A\u777F\u7A62\u82AE\u85DD\u8602\uF9B6" + //  2270 -  2279
		"\u7E2F\uF997\u884D\u8EDF\uF998\uF999\uF99A\u925B\uF99B\u9CF6" + //  2280 -  2289
		"\uF99C\uF99D\uF99E\u6085\u6D85\uF99F\u71B1\uF9A0\uF9A1\u95B1" + //  2290 -  2299
		"\u53AD\uF9A2\uF9A3\uF9A4\u67D3\uF9A5\u708E\u7130\u7430\u8276" + //  2300 -  2309
		"\u82D2\uFFFD\uF983\u6B5F\u6C5D\uF984\u74B5\u7916\uF985\u8207" + //  2310 -  2319
		"\u8245\u8339\u8F3F\u8F5D\uF986\u9918\uF987\uF988\uF989\u4EA6" + //  2320 -  2329
		"\uF98A\u57DF\u5F79\u6613\uF98B\uF98C\u75AB\u7E79\u8B6F\uF98D" + //  2330 -  2339
		"\u9006\u9A5B\u56A5\u5827\u59F8\u5A1F\u5BB4\uF98E\u5EF6\uF98F" + //  2340 -  2349
		"\uF990\u6350\u633B\uF991\u693D\u6C87\u6CBF\u6D8E\u6D93\u6DF5" + //  2350 -  2359
		"\u6F14\uF992\u70DF\u7136\u7159\uF993\u71C3\u71D5\uF994\u784F" + //  2360 -  2369
		"\u786F\uF995\u7B75\u7DE3\uF996\u9B5A\u9F6C\u5104\u61B6\u6291" + //  2370 -  2379
		"\u6A8D\u81C6\u5043\u5830\u5F66\u7109\u8A00\u8AFA\u5B7C\u8616" + //  2380 -  2389
		"\u4FFA\u513C\u56B4\u5944\u63A9\u6DF9\u5DAA\u696D\u5186\u4E88" + //  2390 -  2399
		"\u4F59\uF97F\uF980\uF981\u5982\uF982\uFFFD\u6AFB\u7F4C\u9DAF" + //  2400 -  2409
		"\u9E1A\u4E5F\u503B\u51B6\u591C\u60F9\u63F6\u6930\u723A\u8036" + //  2410 -  2419
		"\uF974\u91CE\u5F31\uF975\uF976\u7D04\u82E5\u846F\u84BB\u85E5" + //  2420 -  2429
		"\u8E8D\uF977\u4F6F\uF978\uF979\u58E4\u5B43\u6059\u63DA\u6518" + //  2430 -  2439
		"\u656D\u6698\uF97A\u694A\u6A23\u6D0B\u7001\u716C\u75D2\u760D" + //  2440 -  2449
		"\u79B3\u7A70\uF97B\u7F8A\uF97C\u8944\uF97D\u8B93\u91C0\u967D" + //  2450 -  2459
		"\uF97E\u990A\u5704\u5FA1\u65BC\u6F01\u7600\u79A6\u8A9E\u99AD" + //  2460 -  2469
		"\u83F4\u95C7\u58D3\u62BC\u72CE\u9D28\u4EF0\u592E\u600F\u663B" + //  2470 -  2479
		"\u6B83\u79E7\u9D26\u5393\u54C0\u57C3\u5D16\u611B\u66D6\u6DAF" + //  2480 -  2489
		"\u788D\u827E\u9698\u9744\u5384\u627C\u6396\u6DB2\u7E0A\u814B" + //  2490 -  2499
		"\u984D\uFFFD\uF972\u6DF1\u700B\u751A\u82AF\u8AF6\u4EC0\u5341" + //  2500 -  2509
		"\uF973\u96D9\u6C0F\u4E9E\u4FC4\u5152\u555E\u5A25\u5CE8\u6211" + //  2510 -  2519
		"\u7259\u82BD\u83AA\u86FE\u8859\u8A1D\u963F\u96C5\u9913\u9D09" + //  2520 -  2529
		"\u9D5D\u580A\u5CB3\u5DBD\u5E44\u60E1\u6115\u63E1\u6A02\u6E25" + //  2530 -  2539
		"\u9102\u9354\u984E\u9C10\u9F77\u5B89\u5CB8\u6309\u664F\u6848" + //  2540 -  2549
		"\u773C\u96C1\u978D\u9854\u9B9F\u65A1\u8B01\u8ECB\u95BC\u5535" + //  2550 -  2559
		"\u5CA9\u5DD6\u5EB5\u6697\u764C\u4F81\u4FE1\u547B\u5A20\u5BB8" + //  2560 -  2569
		"\u613C\u65B0\u6668\u71FC\u7533\u795E\u7D33\u814E\u81E3\u8398" + //  2570 -  2579
		"\u85AA\u85CE\u8703\u8A0A\u8EAB\u8F9B\uF971\u8FC5\u5931\u5BA4" + //  2580 -  2589
		"\u5BE6\u6089\u5BE9\u5C0B\u5FC3\u6C81\uFFFD\u5D69\u745F\u819D" + //  2590 -  2599
		"\u8768\u6FD5\u62FE\u7FD2\u8936\u8972\u4E1E\u4E58\u50E7\u52DD" + //  2600 -  2609
		"\u5347\u627F\u6607\u7E69\u8805\u965E\u4F8D\u5319\u5636\u59CB" + //  2610 -  2619
		"\u5AA4\u5C38\u5C4E\u5C4D\u5E02\u5F11\u6043\u65BD\u662F\u6642" + //  2620 -  2629
		"\u67BE\u67F4\u731C\u77E2\u793A\u7FC5\u8494\u84CD\u8996\u8A66" + //  2630 -  2639
		"\u8A69\u8AE1\u8C55\u8C7A\u57F4\u5BD4\u5F0F\u606F\u62ED\u690D" + //  2640 -  2649
		"\u6B96\u6E5C\u7184\u7BD2\u8755\u8B58\u8EFE\u98DF\u98FE\u4F38" + //  2650 -  2659
		"\u5FAA\u6042\u65EC\u6812\u696F\u6A53\u6B89\u6D35\u6DF3\u73E3" + //  2660 -  2669
		"\u76FE\u77AC\u7B4D\u7D14\u8123\u821C\u8340\u84F4\u8563\u8A62" + //  2670 -  2679
		"\u8AC4\u9187\u931E\u9806\u99B4\u620C\u8853\u8FF0\u9265\u5D07" + //  2680 -  2689
		"\u5D27\uFFFD\u620D\u624B\u6388\u641C\u6536\u6578\u6A39\u6B8A" + //  2690 -  2699
		"\u6C34\u6D19\u6F31\u71E7\u72E9\u7378\u7407\u74B2\u7626\u7761" + //  2700 -  2709
		"\u79C0\u7A57\u7AEA\u7CB9\u7D8F\u7DAC\u7E61\u7F9E\u8129\u8331" + //  2710 -  2719
		"\u8490\u84DA\u85EA\u8896\u8AB0\u8B90\u8F38\u9042\u9083\u916C" + //  2720 -  2729
		"\u9296\u92B9\u968B\u96A7\u96A8\u96D6\u9700\u9808\u9996\u9AD3" + //  2730 -  2739
		"\u9B1A\u53D4\u587E\u5919\u5B70\u5BBF\u6DD1\u6F5A\u719F\u7421" + //  2740 -  2749
		"\u74B9\u8085\u83FD\u5DE1\u5F87\u84C0\u905C\u98E1\u7387\u5B8B" + //  2750 -  2759
		"\u609A\u677E\u6DDE\u8A1F\u8AA6\u9001\u980C\u5237\uF970\u7051" + //  2760 -  2769
		"\u788E\u9396\u8870\u91D7\u4FEE\u53D7\u55FD\u56DA\u5782\u58FD" + //  2770 -  2779
		"\u5AC2\u5B88\u5CAB\u5CC0\u5E25\u6101\uFFFD\u8056\u8072\u8165" + //  2780 -  2789
		"\u8AA0\u9192\u4E16\u52E2\u6B72\u6D17\u7A05\u7B39\u7D30\uF96F" + //  2790 -  2799
		"\u8CB0\u53EC\u562F\u5851\u5BB5\u5C0F\u5C11\u5DE2\u6240\u6383" + //  2800 -  2809
		"\u6414\u662D\u68B3\u6CBC\u6D88\u6EAF\u701F\u70A4\u71D2\u7526" + //  2810 -  2819
		"\u758F\u758E\u7619\u7B11\u7BE0\u7C2B\u7D20\u7D39\u852C\u856D" + //  2820 -  2829
		"\u8607\u8A34\u900D\u9061\u90B5\u92B7\u97F6\u9A37\u4FD7\u5C6C" + //  2830 -  2839
		"\u675F\u6D91\u7C9F\u7E8C\u8B16\u8D16\u901F\u5B6B\u5DFD\u640D" + //  2840 -  2849
		"\u859B\u893B\u8A2D\u8AAA\u96EA\u9F67\u5261\u66B9\u6BB2\u7E96" + //  2850 -  2859
		"\u87FE\u8D0D\u9583\u965D\u651D\u6D89\u71EE\uF96E\u57CE\u59D3" + //  2860 -  2869
		"\u5BAC\u6027\u60FA\u6210\u661F\u665F\u7329\u73F9\u76DB\u7701" + //  2870 -  2879
		"\u7B6C\uFFFD\u80E5\u8212\u85AF\u897F\u8A93\u901D\u92E4\u9ECD" + //  2880 -  2889
		"\u9F20\u5915\u596D\u5E2D\u60DC\u6614\u6673\u6790\u6C50\u6DC5" + //  2890 -  2899
		"\u6F5F\u77F3\u78A9\u84C6\u91CB\u932B\u4ED9\u50CA\u5148\u5584" + //  2900 -  2909
		"\u5B0B\u5BA3\u6247\u657E\u65CB\u6E32\u717D\u7401\u7444\u7487" + //  2910 -  2919
		"\u74BF\u766C\u79AA\u7DDA\u7E55\u7FA8\u817A\u81B3\u8239\u861A" + //  2920 -  2929
		"\u87EC\u8A75\u8DE3\u9078\u9291\u9425\u994D\u9BAE\u5368\u5C51" + //  2930 -  2939
		"\u6954\u6CC4\u6D29\u6E2B\u820C\u55C7\uF96C\u7A61\u7D22\u8272" + //  2940 -  2949
		"\u7272\u751F\u7525\uF96D\u7B19\u5885\u58FB\u5DBC\u5E8F\u5EB6" + //  2950 -  2959
		"\u5F90\u6055\u6292\u637F\u654D\u6691\u66D9\u66F8\u6816\u68F2" + //  2960 -  2969
		"\u7280\u745E\u7B6E\u7D6E\u7DD6\u7F72\uFFFD\u5098\u522A\u5C71" + //  2970 -  2979
		"\u6563\u6C55\u73CA\u7523\u759D\u7B97\u849C\u9178\u9730\u4E77" + //  2980 -  2989
		"\u6492\u6BBA\u715E\u85A9\u4E09\uF96B\u6749\u68EE\u6E17\u829F" + //  2990 -  2999
		"\u8518\u886B\u63F7\u6F81\u9212\u98AF\u4E0A\u50B7\u50CF\u511F" + //  3000 -  3009
		"\u5546\u55AA\u5617\u5B40\u5C19\u5CE0\u5E38\u5E8A\u5EA0\u5EC2" + //  3010 -  3019
		"\u60F3\u6851\u6A61\u6E58\u723D\u7240\u72C0\u76F8\u7965\u7BB1" + //  3020 -  3029
		"\u7FD4\u88F3\u89F4\u8A73\u8C61\u8CDE\u971C\u585E\u74BD\u8CFD" + //  3030 -  3039
		"\u6E23\u7009\u7345\u7802\u793E\u7940\u7960\u79C1\u7BE9\u7D17" + //  3040 -  3049
		"\u7D72\u8086\u820D\u838E\u84D1\u86C7\u88DF\u8A50\u8A5E\u8B1D" + //  3050 -  3059
		"\u8CDC\u8D66\u8FAD\u90AA\u98FC\u99DF\u9E9D\u524A\uF969\u6714" + //  3060 -  3069
		"\uF96A\uFFFD\u813E\u81C2\u83F2\u871A\u88E8\u8AB9\u8B6C\u8CBB" + //  3070 -  3079
		"\u9119\u975E\u98DB\u9F3B\u56AC\u5B2A\u5F6C\u658C\u6AB3\u6BAF" + //  3080 -  3089
		"\u6D5C\u6FF1\u7015\u725D\u73AD\u8CA7\u8CD3\u983B\u6191\u6C37" + //  3090 -  3099
		"\u8058\u9A01\u4E4D\u4E8B\u4E9B\u4ED5\u4F3A\u4F3C\u4F7F\u4FDF" + //  3100 -  3109
		"\u50FF\u53F2\u53F8\u5506\u55E3\u56DB\u58EB\u5962\u5A11\u5BEB" + //  3110 -  3119
		"\u5BFA\u5C04\u5DF3\u5E2B\u5F99\u601D\u6368\u659C\u65AF\u67F6" + //  3120 -  3129
		"\u67FB\u68AD\u6B7B\u6C99\u6CD7\u4E15\u5099\u5315\u532A\u5351" + //  3130 -  3139
		"\u5983\u5A62\u5E87\u60B2\u618A\u6249\u6279\u6590\u6787\u69A7" + //  3140 -  3149
		"\u6BD4\u6BD6\u6BD7\u6BD8\u6CB8\uF968\u7435\u75FA\u7812\u7891" + //  3150 -  3159
		"\u79D5\u79D8\u7C83\u7DCB\u7FE1\u80A5\uFFFD\u5B5A\u5B75\u5BCC" + //  3160 -  3169
		"\u5E9C\uF966\u6276\u6577\u65A7\u6D6E\u6EA5\u7236\u7B26\u7C3F" + //  3170 -  3179
		"\u7F36\u8150\u8151\u819A\u8240\u8299\u83A9\u8A03\u8CA0\u8CE6" + //  3180 -  3189
		"\u8CFB\u8D74\u8DBA\u90E8\u91DC\u961C\u9644\u99D9\u9CE7\u5317" + //  3190 -  3199
		"\u5206\u5429\u5674\u58B3\u5954\u596E\u5FFF\u61A4\u626E\u6610" + //  3200 -  3209
		"\u6C7E\u711A\u76C6\u7C89\u7CDE\u7D1B\u82AC\u8CC1\u96F0\uF967" + //  3210 -  3219
		"\u4F5B\u5F17\u5F7F\u62C2\u5D29\u670B\u68DA\u787C\u7E43\u9D6C" + //  3220 -  3229
		"\u99A5\u9C12\u672C\u4E76\u4FF8\u5949\u5C01\u5CEF\u5CF0\u6367" + //  3230 -  3239
		"\u68D2\u70FD\u71A2\u742B\u7E2B\u84EC\u8702\u9022\u92D2\u9CF3" + //  3240 -  3249
		"\u4E0D\u4ED8\u4FEF\u5085\u5256\u526F\u5426\u5490\u57E0\u592B" + //  3250 -  3259
		"\u5A66\uFFFD\u78A7\u8617\u95E2\u9739\uF965\u535E\u5F01\u8B8A" + //  3260 -  3269
		"\u8FA8\u8FAF\u908A\u5225\u77A5\u9C49\u9F08\u4E19\u5002\u5175" + //  3270 -  3279
		"\u5C5B\u5E77\u661E\u663A\u67C4\u68C5\u70B3\u7501\u75C5\u79C9" + //  3280 -  3289
		"\u7ADD\u8F27\u9920\u9A08\u4FDD\u5821\u5831\u5BF6\u666E\u6B65" + //  3290 -  3299
		"\u6D11\u6E7A\u6F7D\u73E4\u752B\u83E9\u88DC\u8913\u8B5C\u8F14" + //  3300 -  3309
		"\u4F0F\u50D5\u5310\u535C\u5B93\u5FA9\u670D\u798F\u8179\u832F" + //  3310 -  3319
		"\u8514\u8907\u8986\u8F39\u8F3B\u6A0A\u7169\u71D4\u756A\uF964" + //  3320 -  3329
		"\u7E41\u8543\u85E9\u98DC\u4F10\u7B4F\u7F70\u95A5\u51E1\u5E06" + //  3330 -  3339
		"\u68B5\u6C3E\u6C4E\u6CDB\u72AF\u7BC4\u8303\u6CD5\u743A\u50FB" + //  3340 -  3349
		"\u5288\u58C1\u64D8\u6A97\u74A7\u7656\uFFFD\u767C\u8DCB\u91B1" + //  3350 -  3359
		"\u9262\u9AEE\u9B43\u5023\u508D\u574A\u59A8\u5C28\u5E47\u5F77" + //  3360 -  3369
		"\u623F\u653E\u65B9\u65C1\u6609\u678B\u699C\u6EC2\u78C5\u7D21" + //  3370 -  3379
		"\u80AA\u8180\u822B\u82B3\u84A1\u868C\u8A2A\u8B17\u90A6\u9632" + //  3380 -  3389
		"\u9F90\u500D\u4FF3\uF963\u57F9\u5F98\u62DC\u6392\u676F\u6E43" + //  3390 -  3399
		"\u7119\u76C3\u80CC\u80DA\u88F4\u88F5\u8919\u8CE0\u8F29\u914D" + //  3400 -  3409
		"\u966A\u4F2F\u4F70\u5E1B\u67CF\u6822\u767D\u767E\u9B44\u5E61" + //  3410 -  3419
		"\u99C1\u4F34\u534A\u53CD\u53DB\u62CC\u642C\u6500\u6591\u69C3" + //  3420 -  3429
		"\u6CEE\u6F58\u73ED\u7554\u7622\u76E4\u76FC\u78D0\u78FB\u792C" + //  3430 -  3439
		"\u7D46\u822C\u87E0\u8FD4\u9812\u98EF\u52C3\u62D4\u64A5\u6E24" + //  3440 -  3449
		"\u6F51\uFFFD\u6C76\u7D0A\u7D0B\u805E\u868A\u9580\u96EF\u52FF" + //  3450 -  3459
		"\u6C95\u7269\u5473\u5A9A\u5C3E\u5D4B\u5F4C\u5FAE\u672A\u68B6" + //  3460 -  3469
		"\u6963\u6E3C\u6E44\u7709\u7C73\u7F8E\u8587\u8B0E\u8FF7\u9761" + //  3470 -  3479
		"\u9EF4\u5CB7\u60B6\u610D\u61AB\u654F\u65FB\u65FC\u6C11\u6CEF" + //  3480 -  3489
		"\u739F\u73C9\u7DE1\u9594\u5BC6\u871C\u8B10\u525D\u535A\u62CD" + //  3490 -  3499
		"\u640F\u64B2\u6734\u6A38\u6CCA\u73C0\u749E\u7B94\u7C95\u7E1B" + //  3500 -  3509
		"\u818A\u8236\u8584\u8FEB\u96F9\u82D7\u9328\u52D9\u5DEB\u61AE" + //  3510 -  3519
		"\u61CB\u620A\u62C7\u64AB\u65E0\u6959\u6B66\u6BCB\u7121\u73F7" + //  3520 -  3529
		"\u755D\u7E46\u821E\u8302\u856A\u8AA3\u8CBF\u9727\u9D61\u58A8" + //  3530 -  3539
		"\u9ED8\u5011\u520E\u543B\u554F\u6587\uFFFD\u8511\u51A5\u540D" + //  3540 -  3549
		"\u547D\u660E\u669D\u6927\u6E9F\u76BF\u7791\u8317\u84C2\u879F" + //  3550 -  3559
		"\u9169\u9298\u9CF4\u8882\u4FAE\u5192\u52DF\u59C6\u5E3D\u6155" + //  3560 -  3569
		"\u6478\u6479\u66AE\u67D0\u6A21\u6BCD\u6BDB\u725F\u7261\u7441" + //  3570 -  3579
		"\u7738\u77DB\u8017\u82BC\u8305\u8B00\u8B28\u8C8C\u6728\u6C90" + //  3580 -  3589
		"\u7267\u76EE\u7766\u7A46\u9DA9\u6B7F\u6C92\u5922\u6726\u8499" + //  3590 -  3599
		"\u536F\u5893\u5999\u5EDF\u63CF\u6634\u6773\u6E3A\u732B\u7AD7" + //  3600 -  3609
		"\u7164\u7F75\u8CB7\u8CE3\u9081\u9B45\u8108\u8C8A\u964C\u9A40" + //  3610 -  3619
		"\u9EA5\u5B5F\u6C13\u731B\u76F2\u76DF\u840C\u51AA\u8993\u514D" + //  3620 -  3629
		"\u5195\u52C9\u68C9\u6C94\u7704\u7720\u7DBF\u7DEC\u9762\u9EB5" + //  3630 -  3639
		"\u6EC5\uFFFD\u7ACB\u7B20\u7C92\u6469\u746A\u75F2\u78BC\u78E8" + //  3640 -  3649
		"\u99AC\u9B54\u9EBB\u5BDE\u5E55\u6F20\u819C\u83AB\u9088\u4E07" + //  3650 -  3659
		"\u534D\u5A29\u5DD2\u5F4E\u6162\u633D\u6669\u66FC\u6EFF\u6F2B" + //  3660 -  3669
		"\u7063\u779E\u842C\u8513\u883B\u8F13\u9945\u9C3B\u551C\u62B9" + //  3670 -  3679
		"\u672B\u6CAB\u8309\u896A\u977A\u4EA1\u5984\u5FD8\u5FD9\u671B" + //  3680 -  3689
		"\u7DB2\u7F54\u8292\u832B\u83BD\u8F1E\u9099\u57CB\u59B9\u5A92" + //  3690 -  3699
		"\u5BD0\u6627\u679A\u6885\u6BCF\u7281\u72F8\u7406\u7483\uF962" + //  3700 -  3709
		"\u75E2\u7C6C\u7F79\u7FB8\u8389\u88CF\u88E1\u91CC\u91D0\u96E2" + //  3710 -  3719
		"\u9BC9\u541D\u6F7E\u71D0\u7498\u85FA\u8EAA\u96A3\u9C57\u9E9F" + //  3720 -  3729
		"\u6797\u6DCB\u7433\u81E8\u9716\u782C\uFFFD\u907C\u9B27\u9F8D" + //  3730 -  3739
		"\u58D8\u5A41\u5C62\u6A13\u6DDA\u6F0F\u763B\u7D2F\u7E37\u851E" + //  3740 -  3749
		"\u8938\u93E4\u964B\u5289\u65D2\u67F3\u69B4\u6D41\u6E9C\u700F" + //  3750 -  3759
		"\u7409\u7460\u7559\u7624\u786B\u8B2C\u985E\u516D\u622E\u9678" + //  3760 -  3769
		"\u4F96\u502B\u5D19\u6DEA\u7DB8\u8F2A\u5F8B\u6144\u6817\uF961" + //  3770 -  3779
		"\u9686\u52D2\u808B\u51DC\u51CC\u695E\u7A1C\u7DBE\u83F1\u9675" + //  3780 -  3789
		"\u4FDA\u5229\u5398\u540F\u550E\u5C65\u60A7\u674E\u68A8\u6D6C" + //  3790 -  3799
		"\u7DA0\u83C9\u9304\u9E7F\u9E93\u8AD6\u58DF\u5F04\u6727\u7027" + //  3800 -  3809
		"\u74CF\u7C60\u807E\u5121\u7028\u7262\u78CA\u8CC2\u8CDA\u8CF4" + //  3810 -  3819
		"\u96F7\u4E86\u50DA\u5BEE\u5ED6\u6599\u71CE\u7642\u77AD\u804A" + //  3820 -  3829
		"\u84FC\uFFFD\u7149\u7489\u7DF4\u806F\u84EE\u8F26\u9023\u934A" + //  3830 -  3839
		"\u51BD\u5217\u52A3\u6D0C\u70C8\u88C2\u5EC9\u6582\u6BAE\u6FC2" + //  3840 -  3849
		"\u7C3E\u7375\u4EE4\u4F36\u56F9\uF95F\u5CBA\u5DBA\u601C\u73B2" + //  3850 -  3859
		"\u7B2D\u7F9A\u7FCE\u8046\u901E\u9234\u96F6\u9748\u9818\u9F61" + //  3860 -  3869
		"\u4F8B\u6FA7\u79AE\u91B4\u96B7\u52DE\uF960\u6488\u64C4\u6AD3" + //  3870 -  3879
		"\u6F5E\u7018\u7210\u76E7\u8001\u8606\u865C\u8DEF\u8F05\u9732" + //  3880 -  3889
		"\u9B6F\u9DFA\u9E75\u788C\u797F\u8F1B\u91CF\u4FB6\u5137\u52F5" + //  3890 -  3899
		"\u5442\u5EEC\u616E\u623E\u65C5\u6ADA\u6FFE\u792A\u85DC\u8823" + //  3900 -  3909
		"\u95AD\u9A62\u9A6A\u9E97\u9ECE\u529B\u66C6\u6B77\u701D\u792B" + //  3910 -  3919
		"\u8F62\u9742\u6190\u6200\u6523\u6F23\uFFFD\u863F\u87BA\u88F8" + //  3920 -  3929
		"\u908F\uF95C\u6D1B\u70D9\u73DE\u7D61\u843D\uF95D\u916A\u99F1" + //  3930 -  3939
		"\uF95E\u4E82\u5375\u6B04\u6B12\u703E\u721B\u862D\u9E1E\u524C" + //  3940 -  3949
		"\u8FA3\u5D50\u64E5\u652C\u6B16\u6FEB\u7C43\u7E9C\u85CD\u8964" + //  3950 -  3959
		"\u89BD\u62C9\u81D8\u881F\u5ECA\u6717\u6D6A\u72FC\u7405\u746F" + //  3960 -  3969
		"\u8782\u90DE\u4F86\u5D0D\u5FA0\u840A\u51B7\u63A0\u7565\u4EAE" + //  3970 -  3979
		"\u5006\u5169\u51C9\u6881\u6A11\u7CAE\u7CB1\u7CE7\u826F\u8AD2" + //  3980 -  3989
		"\u6597\u675C\u6793\u75D8\u7AC7\u8373\uF95A\u8C46\u9017\u982D" + //  3990 -  3999
		"\u5C6F\u81C0\u829A\u9041\u906F\u920D\u5F97\u5D9D\u6A59\u71C8" + //  4000 -  4009
		"\u767B\u7B49\u85E4\u8B04\u9127\u9A30\u5587\u61F6\uF95B\u7669" + //  4010 -  4019
		"\u7F85\uFFFD\u68F9\u6AC2\u6DD8\u6E21\u6ED4\u6FE4\u71FE\u76DC" + //  4020 -  4029
		"\u7779\u79B1\u7A3B\u8404\u89A9\u8CED\u8DF3\u8E48\u9003\u9014" + //  4030 -  4039
		"\u9053\u90FD\u934D\u9676\u97DC\u6BD2\u7006\u7258\u72A2\u7368" + //  4040 -  4049
		"\u7763\u79BF\u7BE4\u7E9B\u8B80\u58A9\u60C7\u6566\u65FD\u66BE" + //  4050 -  4059
		"\u6C8C\u711E\u71C9\u8C5A\u9813\u4E6D\u7A81\u4EDD\u51AC\u51CD" + //  4060 -  4069
		"\u52D5\u540C\u61A7\u6771\u6850\u68DF\u6D1E\u6F7C\u75BC\u77B3" + //  4070 -  4079
		"\u7AE5\u80F4\u8463\u9285\u515C\u5CB1\u5E36\u5F85\u6234\u64E1" + //  4080 -  4089
		"\u73B3\u81FA\u888B\u8CB8\u968A\u9EDB\u5B85\u5FB7\u60B3\u5012" + //  4090 -  4099
		"\u5200\u5230\u5716\u5835\u5857\u5C0E\u5C60\u5CF6\u5D8B\u5EA6" + //  4100 -  4109
		"\u5F92\u60BC\u6311\u6389\u6417\u6843\uFFFD\u4E39\u4EB6\u4F46" + //  4110 -  4119
		"\u55AE\u5718\u58C7\u5F56\u65B7\u65E6\u6A80\u6BB5\u6E4D\u77ED" + //  4120 -  4129
		"\u7AEF\u7C1E\u7DDE\u86CB\u8892\u9132\u935B\u64BB\u6FBE\u737A" + //  4130 -  4139
		"\u75B8\u9054\u5556\u574D\u61BA\u64D4\u66C7\u6DE1\u6E5B\u6F6D" + //  4140 -  4149
		"\u6FB9\u75F0\u8043\u81BD\u8541\u8983\u8AC7\u8B5A\u931F\u6C93" + //  4150 -  4159
		"\u7553\u7B54\u8E0F\u905D\u5510\u5802\u5858\u5E62\u6207\u649E" + //  4160 -  4169
		"\u68E0\u7576\u7CD6\u87B3\u9EE8\u4EE3\u5788\u576E\u5927\u5C0D" + //  4170 -  4179
		"\u8166\uF948\uF949\u5C3F\uF94A\uF94B\uF94C\uF94D\uF94E\uF94F" + //  4180 -  4189
		"\uF950\uF951\u5AE9\u8A25\u677B\u7D10\uF952\uF953\uF954\uF955" + //  4190 -  4199
		"\uF956\uF957\u80FD\uF958\uF959\u5C3C\u6CE5\u533F\u6EBA\u591A" + //  4200 -  4209
		"\u8336\uFFFD\u7D0D\uF926\uF927\u8872\u56CA\u5A18\uF928\uF929" + //  4210 -  4219
		"\uF92A\uF92B\uF92C\u4E43\uF92D\u5167\u5948\u67F0\u8010\uF92E" + //  4220 -  4229
		"\u5973\u5E74\u649A\u79CA\u5FF5\u606C\u62C8\u637B\u5BE7\u5BD7" + //  4230 -  4239
		"\u52AA\uF92F\u5974\u5F29\u6012\uF930\uF931\uF932\u7459\uF933" + //  4240 -  4249
		"\uF934\uF935\uF936\uF937\uF938\u99D1\uF939\uF93A\uF93B\uF93C" + //  4250 -  4259
		"\uF93D\uF93E\uF93F\uF940\uF941\uF942\uF943\u6FC3\uF944\uF945" + //  4260 -  4269
		"\u81BF\u8FB2\u60F1\uF946\uF947\uF913\u90A3\uF914\uF915\uF916" + //  4270 -  4279
		"\uF917\uF918\u8AFE\uF919\uF91A\uF91B\uF91C\u6696\uF91D\u7156" + //  4280 -  4289
		"\uF91E\uF91F\u96E3\uF920\u634F\u637A\u5357\uF921\u678F\u6960" + //  4290 -  4299
		"\u6E73\uF922\u7537\uF923\uF924\uF925\uFFFD\u671E\u671F\u675E" + //  4300 -  4309
		"\u68CB\u68C4\u6A5F\u6B3A\u6C23\u6C7D\u6C82\u6DC7\u7398\u7426" + //  4310 -  4319
		"\u742A\u7482\u74A3\u7578\u757F\u7881\u78EF\u7941\u7947\u7948" + //  4320 -  4329
		"\u797A\u7B95\u7D00\u7DBA\u7F88\u8006\u802D\u808C\u8A18\u8B4F" + //  4330 -  4339
		"\u8C48\u8D77\u9321\u9324\u98E2\u9951\u9A0E\u9A0F\u9A65\u9E92" + //  4340 -  4349
		"\u7DCA\u4F76\u5409\u62EE\u6854\u91D1\u55AB\u513A\uF90B\uF90C" + //  4350 -  4359
		"\u5A1C\u61E6\uF90D\u62CF\u62FF\uF90E\uF90F\uF910\uF911\uF912" + //  4360 -  4369
		"\u53CA\u6025\u6271\u6C72\u7D1A\u7D66\u4E98\u5162\u77DC\u80AF" + //  4370 -  4379
		"\u4F01\u4F0E\u5176\u5180\u55DC\u5668\u573B\u57FA\u57FC\u5914" + //  4380 -  4389
		"\u5947\u5993\u5BC4\u5C90\u5D0E\u5DF1\u5E7E\u5FCC\u6280\u65D7" + //  4390 -  4399
		"\u65E3\uFFFD\u9B3C\uF907\u53EB\u572D\u594E\u63C6\u69FB\u73EA" + //  4400 -  4409
		"\u7845\u7ABA\u7AC5\u7CFE\u8475\u898F\u8D73\u9035\u95A8\u52FB" + //  4410 -  4419
		"\u5747\u7547\u7B60\u83CC\u921E\uF908\u6A58\u514B\u524B\u5287" + //  4420 -  4429
		"\u621F\u68D8\u6975\u9699\u50C5\u52A4\u52E4\u61C3\u65A4\u6839" + //  4430 -  4439
		"\u69FF\u747E\u7B4B\u82B9\u83EB\u89B2\u8B39\u8FD1\u9949\uF909" + //  4440 -  4449
		"\u4ECA\u5997\u64D2\u6611\u6A8E\u7434\u7981\u79BD\u82A9\u887E" + //  4450 -  4459
		"\u887F\u895F\uF90A\u9326\u4F0B\u5BAE\u5F13\u7A79\u7AAE\u828E" + //  4460 -  4469
		"\u8EAC\u5026\u5238\u52F8\u5377\u5708\u62F3\u6372\u6B0A\u6DC3" + //  4470 -  4479
		"\u7737\u53A5\u7357\u8568\u8E76\u95D5\u673A\u6AC3\u6F70\u8A6D" + //  4480 -  4489
		"\u8ECC\u994B\uF906\u6677\u6B78\u8CB4\uFFFD\u5340\u53E3\u53E5" + //  4490 -  4499
		"\u548E\u5614\u5775\u57A2\u5BC7\u5D87\u5ED0\u61FC\u62D8\u6551" + //  4500 -  4509
		"\u67B8\u67E9\u69CB\u6B50\u6BC6\u6BEC\u6C42\u6E9D\u7078\u72D7" + //  4510 -  4519
		"\u7396\u7403\u77BF\u77E9\u7A76\u7D7F\u8009\u81FC\u8205\u820A" + //  4520 -  4529
		"\u82DF\u8862\u8B33\u8CFC\u8EC0\u9011\u90B1\u9264\u92B6\u99D2" + //  4530 -  4539
		"\u9A45\u9CE9\u9DD7\u9F9C\u570B\u5C40\u83CA\u97A0\u97AB\u9EB4" + //  4540 -  4549
		"\u541B\u7A98\u7FA4\u88D9\u8ECD\u90E1\u5800\u5C48\u6398\u7A9F" + //  4550 -  4559
		"\u50D1\u54AC\u55AC\u5B0C\u5DA0\u5DE7\u652A\u654E\u6821\u6A4B" + //  4560 -  4569
		"\u72E1\u768E\u77EF\u7D5E\u7FF9\u81A0\u854E\u86DF\u8F03\u8F4E" + //  4570 -  4579
		"\u90CA\u9903\u9A55\u9BAB\u4E18\u4E45\u4E5D\u4EC7\u4FF1\u5177" + //  4580 -  4589
		"\u52FE\uFFFD\u79D1\u83D3\u8A87\u8AB2\u8DE8\u904E\u934B\u9846" + //  4590 -  4599
		"\u5ED3\u69E8\u85FF\u90ED\uF905\u51A0\u5B98\u5BEC\u6163\u68FA" + //  4600 -  4609
		"\u6B3E\u704C\u742F\u74D8\u7BA1\u7F50\u83C5\u89C0\u8CAB\u95DC" + //  4610 -  4619
		"\u9928\u522E\u605D\u62EC\u9002\u4F8A\u5149\u5321\u58D9\u5EE3" + //  4620 -  4629
		"\u66E0\u6D38\u709A\u72C2\u73D6\u7B50\u80F1\u945B\u5366\u639B" + //  4630 -  4639
		"\u7F6B\u4E56\u5080\u584A\u58DE\u602A\u6127\u62D0\u69D0\u9B41" + //  4640 -  4649
		"\u5B8F\u7D18\u80B1\u8F5F\u4EA4\u6606\u68B1\u68CD\u6EFE\u7428" + //  4650 -  4659
		"\u889E\u9BE4\u6C68\uF904\u9AA8\u4F9B\u516C\u5171\u529F\u5B54" + //  4660 -  4669
		"\u5DE5\u6050\u606D\u62F1\u63A7\u653B\u73D9\u7A7A\u86A3\u8CA2" + //  4670 -  4679
		"\u978F\u4E32\u5BE1\u6208\u679C\u74DC\uFFFD\u68E8\u6EAA\u754C" + //  4680 -  4689
		"\u7678\u78CE\u7A3D\u7CFB\u7E6B\u7E7C\u8A08\u8AA1\u8C3F\u968E" + //  4690 -  4699
		"\u9DC4\u53E4\u53E9\u544A\u5471\u56FA\u59D1\u5B64\u5C3B\u5EAB" + //  4700 -  4709
		"\u62F7\u6537\u6545\u6572\u66A0\u67AF\u69C1\u6CBD\u75FC\u7690" + //  4710 -  4719
		"\u777E\u7A3F\u7F94\u8003\u80A1\u818F\u82E6\u82FD\u83F0\u85C1" + //  4720 -  4729
		"\u8831\u88B4\u8AA5\uF903\u8F9C\u932E\u96C7\u9867\u9AD8\u9F13" + //  4730 -  4739
		"\u54ED\u659B\u66F2\u688F\u7A40\u8C37\u9D60\u56F0\u5764\u5D11" + //  4740 -  4749
		"\u74A5\u74CA\u75D9\u786C\u78EC\u7ADF\u7AF6\u7D45\u7D93\u8015" + //  4750 -  4759
		"\u803F\u811B\u8396\u8B66\u8F15\u9015\u93E1\u9803\u9838\u9A5A" + //  4760 -  4769
		"\u9BE8\u4FC2\u5553\u583A\u5951\u5B63\u5C46\u60B8\u6212\u6842" + //  4770 -  4779
		"\u68B0\uFFFD\u77BC\u9210\u9ED4\u52AB\u602F\u8FF2\u5048\u61A9" + //  4780 -  4789
		"\u63ED\u64CA\u683C\u6A84\u6FC0\u8188\u89A1\u9694\u5805\u727D" + //  4790 -  4799
		"\u72AC\u7504\u7D79\u7E6D\u80A9\u898B\u8B74\u9063\u9D51\u6289" + //  4800 -  4809
		"\u6C7A\u6F54\u7D50\u7F3A\u8A23\u517C\u614A\u7B9D\u8B19\u9257" + //  4810 -  4819
		"\u938C\u4EAC\u4FD3\u501E\u50BE\u5106\u52C1\u52CD\u537F\u5770" + //  4820 -  4829
		"\u5883\u5E9A\u5F91\u6176\u61AC\u64CE\u656C\u666F\u66BB\u66F4" + //  4830 -  4839
		"\u6897\u6D87\u7085\u70F1\u749F\u64DA\u64E7\u6E20\u70AC\u795B" + //  4840 -  4849
		"\u8DDD\u8E1E\uF902\u907D\u9245\u92F8\u4E7E\u4EF6\u5065\u5DFE" + //  4850 -  4859
		"\u5EFA\u6106\u6957\u8171\u8654\u8E47\u9375\u9A2B\u4E5E\u5091" + //  4860 -  4869
		"\u6770\u6840\u5109\u528D\u5292\u6AA2\uFFFD\u5323\u5CAC\u7532" + //  4870 -  4879
		"\u80DB\u9240\u9598\u525B\u5808\u59DC\u5CA1\u5D17\u5EB7\u5F3A" + //  4880 -  4889
		"\u5F4A\u6177\u6C5F\u757A\u7586\u7CE0\u7D73\u7DB1\u7F8C\u8154" + //  4890 -  4899
		"\u8221\u8591\u8941\u8B1B\u92FC\u964D\u9C47\u4ECB\u4EF7\u500B" + //  4900 -  4909
		"\u51F1\u584F\u6137\u613E\u6168\u6539\u69EA\u6F11\u75A5\u7686" + //  4910 -  4919
		"\u76D6\u7B87\u82A5\u84CB\uF900\u93A7\u958B\u5580\u5BA2\u5751" + //  4920 -  4929
		"\uF901\u7CB3\u7FB9\u91B5\u5028\u53BB\u5C45\u5DE8\u62D2\u636E" + //  4930 -  4939
		"\u9593\u4E6B\u559D\u66F7\u6E34\u78A3\u7AED\u845B\u8910\u874E" + //  4940 -  4949
		"\u97A8\u52D8\u574E\u582A\u5D4C\u611F\u61BE\u6221\u6562\u67D1" + //  4950 -  4959
		"\u6A44\u6E1B\u7518\u75B3\u76E3\u77B0\u7D3A\u90AF\u9451\u9452" + //  4960 -  4969
		"\u9F95\uFFFD\u4F3D\u4F73\u5047\u50F9\u52A0\u53EF\u5475\u54E5" + //  4970 -  4979
		"\u5609\u5AC1\u5BB6\u6687\u67B6\u67B7\u67EF\u6B4C\u73C2\u75C2" + //  4980 -  4989
		"\u7A3C\u82DB\u8304\u8857\u8888\u8A36\u8CC8\u8DCF\u8EFB\u8FE6" + //  4990 -  4999
		"\u99D5\u523B\u5374\u5404\u606A\u6164\u6BBC\u73CF\u811A\u89BA" + //  5000 -  5009
		"\u89D2\u95A3\u4F83\u520A\u58BE\u5978\u59E6\u5E72\u5E79\u61C7" + //  5010 -  5019
		"\u63C0\u6746\u67EC\u687F\u6F97\u764E\u770B\u78F5\u7A08\u7AFF" + //  5020 -  5029
		"\u7C21\u809D\u826E\u8271\u8AEB\uE03F\uE040\uE041\uE042\uE043" + //  5030 -  5039
		"\uE044\uE045\uE046\uE047\uE048\uE049\uE04A\uE04B\uE04C\uE04D" + //  5040 -  5049
		"\uE04E\uE04F\uE050\uE051\uE052\uE053\uE054\uE055\uE056\uE057" + //  5050 -  5059
		"\uE058\uE059\uE05A\uE05B\uE05C\uE05D\uFFFD\uE000\uE001\uE002" + //  5060 -  5069
		"\uE003\uE004\uE005\uE006\uE007\uE008\uE009\uE00A\uE00B\uE00C" + //  5070 -  5079
		"\uE00D\uE00E\uE00F\uE010\uE011\uE012\uE013\uE014\uE015\uE016" + //  5080 -  5089
		"\uE017\uE018\uE019\uE01A\uE01B\uE01C\uE01D\uE01E\uE01F\uE020" + //  5090 -  5099
		"\uE021\uE022\uE023\uE024\uE025\uE026\uE027\uE028\uE029\uE02A" + //  5100 -  5109
		"\uE02B\uE02C\uE02D\uE02E\uE02F\uE030\uE031\uE032\uE033\uE034" + //  5110 -  5119
		"\uE035\uE036\uE037\uE038\uE039\uE03A\uE03B\uE03C\uE03D\uE03E" + //  5120 -  5129
		"\uD738\uD73C\uD744\uD747\uD749\uD750\uD751\uD754\uD756\uD757" + //  5130 -  5139
		"\uD758\uD759\uD760\uD761\uD763\uD765\uD769\uD76C\uD770\uD774" + //  5140 -  5149
		"\uD77C\uD77D\uD781\uD788\uD789\uD78C\uD790\uD798\uD799\uD79B" + //  5150 -  5159
		"\uD79D\uFFFD\uD624\uD62D\uD638\uD639\uD63C\uD640\uD645\uD648" + //  5160 -  5169
		"\uD649\uD64B\uD64D\uD651\uD654\uD655\uD658\uD65C\uD667\uD669" + //  5170 -  5179
		"\uD670\uD671\uD674\uD683\uD685\uD68C\uD68D\uD690\uD694\uD69D" + //  5180 -  5189
		"\uD69F\uD6A1\uD6A8\uD6AC\uD6B0\uD6B9\uD6BB\uD6C4\uD6C5\uD6C8" + //  5190 -  5199
		"\uD6CC\uD6D1\uD6D4\uD6D7\uD6D9\uD6E0\uD6E4\uD6E8\uD6F0\uD6F5" + //  5200 -  5209
		"\uD6FC\uD6FD\uD700\uD704\uD711\uD718\uD719\uD71C\uD720\uD728" + //  5210 -  5219
		"\uD729\uD72B\uD72D\uD734\uD735\uD589\uD590\uD5A5\uD5C8\uD5C9" + //  5220 -  5229
		"\uD5CC\uD5D0\uD5D2\uD5D8\uD5D9\uD5DB\uD5DD\uD5E4\uD5E5\uD5E8" + //  5230 -  5239
		"\uD5EC\uD5F4\uD5F5\uD5F7\uD5F9\uD600\uD601\uD604\uD608\uD610" + //  5240 -  5249
		"\uD611\uD613\uD614\uD615\uD61C\uD620\uFFFD\uD408\uD41D\uD440" + //  5250 -  5259
		"\uD444\uD45C\uD460\uD464\uD46D\uD46F\uD478\uD479\uD47C\uD47F" + //  5260 -  5269
		"\uD480\uD482\uD488\uD489\uD48B\uD48D\uD494\uD4A9\uD4CC\uD4D0" + //  5270 -  5279
		"\uD4D4\uD4DC\uD4DF\uD4E8\uD4EC\uD4F0\uD4F8\uD4FB\uD4FD\uD504" + //  5280 -  5289
		"\uD508\uD50C\uD514\uD515\uD517\uD53C\uD53D\uD540\uD544\uD54C" + //  5290 -  5299
		"\uD54D\uD54F\uD551\uD558\uD559\uD55C\uD560\uD565\uD568\uD569" + //  5300 -  5309
		"\uD56B\uD56D\uD574\uD575\uD578\uD57C\uD584\uD585\uD587\uD588" + //  5310 -  5319
		"\uD38D\uD38F\uD390\uD391\uD398\uD399\uD39C\uD3A0\uD3A8\uD3A9" + //  5320 -  5329
		"\uD3AB\uD3AD\uD3B4\uD3B8\uD3BC\uD3C4\uD3C5\uD3C8\uD3C9\uD3D0" + //  5330 -  5339
		"\uD3D8\uD3E1\uD3E3\uD3EC\uD3ED\uD3F0\uD3F4\uD3FC\uD3FD\uD3FF" + //  5340 -  5349
		"\uD401\uFFFD\uD264\uD280\uD281\uD284\uD288\uD290\uD291\uD295" + //  5350 -  5359
		"\uD29C\uD2A0\uD2A4\uD2AC\uD2B1\uD2B8\uD2B9\uD2BC\uD2BF\uD2C0" + //  5360 -  5369
		"\uD2C2\uD2C8\uD2C9\uD2CB\uD2D4\uD2D8\uD2DC\uD2E4\uD2E5\uD2F0" + //  5370 -  5379
		"\uD2F1\uD2F4\uD2F8\uD300\uD301\uD303\uD305\uD30C\uD30D\uD30E" + //  5380 -  5389
		"\uD310\uD314\uD316\uD31C\uD31D\uD31F\uD320\uD321\uD325\uD328" + //  5390 -  5399
		"\uD329\uD32C\uD330\uD338\uD339\uD33B\uD33C\uD33D\uD344\uD345" + //  5400 -  5409
		"\uD37C\uD37D\uD380\uD384\uD38C\uD16C\uD17C\uD184\uD188\uD1A0" + //  5410 -  5419
		"\uD1A1\uD1A4\uD1A8\uD1B0\uD1B1\uD1B3\uD1B5\uD1BA\uD1BC\uD1C0" + //  5420 -  5429
		"\uD1D8\uD1F4\uD1F8\uD207\uD209\uD210\uD22C\uD22D\uD230\uD234" + //  5430 -  5439
		"\uD23C\uD23D\uD23F\uD241\uD248\uD25C\uFFFD\uD044\uD045\uD047" + //  5440 -  5449
		"\uD049\uD050\uD054\uD058\uD060\uD06C\uD06D\uD070\uD074\uD07C" + //  5450 -  5459
		"\uD07D\uD081\uD0A4\uD0A5\uD0A8\uD0AC\uD0B4\uD0B5\uD0B7\uD0B9" + //  5460 -  5469
		"\uD0C0\uD0C1\uD0C4\uD0C8\uD0C9\uD0D0\uD0D1\uD0D3\uD0D4\uD0D5" + //  5470 -  5479
		"\uD0DC\uD0DD\uD0E0\uD0E4\uD0EC\uD0ED\uD0EF\uD0F0\uD0F1\uD0F8" + //  5480 -  5489
		"\uD10D\uD130\uD131\uD134\uD138\uD13A\uD140\uD141\uD143\uD144" + //  5490 -  5499
		"\uD145\uD14C\uD14D\uD150\uD154\uD15C\uD15D\uD15F\uD161\uD168" + //  5500 -  5509
		"\uCF67\uCF69\uCF70\uCF71\uCF74\uCF78\uCF80\uCF85\uCF8C\uCFA1" + //  5510 -  5519
		"\uCFA8\uCFB0\uCFC4\uCFE0\uCFE1\uCFE4\uCFE8\uCFF0\uCFF1\uCFF3" + //  5520 -  5529
		"\uCFF5\uCFFC\uD000\uD004\uD011\uD018\uD02D\uD034\uD035\uD038" + //  5530 -  5539
		"\uD03C\uFFFD\uCE58\uCE59\uCE5C\uCE5F\uCE60\uCE61\uCE68\uCE69" + //  5540 -  5549
		"\uCE6B\uCE6D\uCE74\uCE75\uCE78\uCE7C\uCE84\uCE85\uCE87\uCE89" + //  5550 -  5559
		"\uCE90\uCE91\uCE94\uCE98\uCEA0\uCEA1\uCEA3\uCEA4\uCEA5\uCEAC" + //  5560 -  5569
		"\uCEAD\uCEC1\uCEE4\uCEE5\uCEE8\uCEEB\uCEEC\uCEF4\uCEF5\uCEF7" + //  5570 -  5579
		"\uCEF8\uCEF9\uCF00\uCF01\uCF04\uCF08\uCF10\uCF11\uCF13\uCF15" + //  5580 -  5589
		"\uCF1C\uCF20\uCF24\uCF2C\uCF2D\uCF2F\uCF30\uCF31\uCF38\uCF54" + //  5590 -  5599
		"\uCF55\uCF58\uCF5C\uCF64\uCF65\uCD95\uCD98\uCD9C\uCDA4\uCDA5" + //  5600 -  5609
		"\uCDA7\uCDA9\uCDB0\uCDC4\uCDCC\uCDD0\uCDE8\uCDEC\uCDF0\uCDF8" + //  5610 -  5619
		"\uCDF9\uCDFB\uCDFD\uCE04\uCE08\uCE0C\uCE14\uCE19\uCE20\uCE21" + //  5620 -  5629
		"\uCE24\uCE28\uCE30\uCE31\uCE33\uCE35\uFFFD\uCC3C\uCC3D\uCC3E" + //  5630 -  5639
		"\uCC44\uCC45\uCC48\uCC4C\uCC54\uCC55\uCC57\uCC58\uCC59\uCC60" + //  5640 -  5649
		"\uCC64\uCC66\uCC68\uCC70\uCC75\uCC98\uCC99\uCC9C\uCCA0\uCCA8" + //  5650 -  5659
		"\uCCA9\uCCAB\uCCAC\uCCAD\uCCB4\uCCB5\uCCB8\uCCBC\uCCC4\uCCC5" + //  5660 -  5669
		"\uCCC7\uCCC9\uCCD0\uCCD4\uCCE4\uCCEC\uCCF0\uCD01\uCD08\uCD09" + //  5670 -  5679
		"\uCD0C\uCD10\uCD18\uCD19\uCD1B\uCD1D\uCD24\uCD28\uCD2C\uCD39" + //  5680 -  5689
		"\uCD5C\uCD60\uCD64\uCD6C\uCD6D\uCD6F\uCD71\uCD78\uCD88\uCD94" + //  5690 -  5699
		"\uCB4C\uCB50\uCB58\uCB59\uCB5D\uCB64\uCB78\uCB79\uCB9C\uCBB8" + //  5700 -  5709
		"\uCBD4\uCBE4\uCBE7\uCBE9\uCC0C\uCC0D\uCC10\uCC14\uCC1C\uCC1D" + //  5710 -  5719
		"\uCC21\uCC22\uCC27\uCC28\uCC29\uCC2C\uCC2E\uCC30\uCC38\uCC39" + //  5720 -  5729
		"\uCC3B\uFFFD\uC9D5\uC9D6\uC9D9\uC9DA\uC9DC\uC9DD\uC9E0\uC9E2" + //  5730 -  5739
		"\uC9E4\uC9E7\uC9EC\uC9ED\uC9EF\uC9F0\uC9F1\uC9F8\uC9F9\uC9FC" + //  5740 -  5749
		"\uCA00\uCA08\uCA09\uCA0B\uCA0C\uCA0D\uCA14\uCA18\uCA29\uCA4C" + //  5750 -  5759
		"\uCA4D\uCA50\uCA54\uCA5C\uCA5D\uCA5F\uCA60\uCA61\uCA68\uCA7D" + //  5760 -  5769
		"\uCA84\uCA98\uCABC\uCABD\uCAC0\uCAC4\uCACC\uCACD\uCACF\uCAD1" + //  5770 -  5779
		"\uCAD3\uCAD8\uCAD9\uCAE0\uCAEC\uCAF4\uCB08\uCB10\uCB14\uCB18" + //  5780 -  5789
		"\uCB20\uCB21\uCB41\uCB48\uCB49\uC918\uC92C\uC934\uC950\uC951" + //  5790 -  5799
		"\uC954\uC958\uC960\uC961\uC963\uC96C\uC970\uC974\uC97C\uC988" + //  5800 -  5809
		"\uC989\uC98C\uC990\uC998\uC999\uC99B\uC99D\uC9C0\uC9C1\uC9C4" + //  5810 -  5819
		"\uC9C7\uC9C8\uC9CA\uC9D0\uC9D1\uC9D3\uFFFD\uC810\uC811\uC813" + //  5820 -  5829
		"\uC815\uC816\uC81C\uC81D\uC820\uC824\uC82C\uC82D\uC82F\uC831" + //  5830 -  5839
		"\uC838\uC83C\uC840\uC848\uC849\uC84C\uC84D\uC854\uC870\uC871" + //  5840 -  5849
		"\uC874\uC878\uC87A\uC880\uC881\uC883\uC885\uC886\uC887\uC88B" + //  5850 -  5859
		"\uC88C\uC88D\uC894\uC89D\uC89F\uC8A1\uC8A8\uC8BC\uC8BD\uC8C4" + //  5860 -  5869
		"\uC8C8\uC8CC\uC8D4\uC8D5\uC8D7\uC8D9\uC8E0\uC8E1\uC8E4\uC8F5" + //  5870 -  5879
		"\uC8FC\uC8FD\uC900\uC904\uC905\uC906\uC90C\uC90D\uC90F\uC911" + //  5880 -  5889
		"\uC79A\uC7A0\uC7A1\uC7A3\uC7A4\uC7A5\uC7A6\uC7AC\uC7AD\uC7B0" + //  5890 -  5899
		"\uC7B4\uC7BC\uC7BD\uC7BF\uC7C0\uC7C1\uC7C8\uC7C9\uC7CC\uC7CE" + //  5900 -  5909
		"\uC7D0\uC7D8\uC7DD\uC7E4\uC7E8\uC7EC\uC800\uC801\uC804\uC808" + //  5910 -  5919
		"\uC80A\uFFFD\uC6E9\uC6EC\uC6F0\uC6F8\uC6F9\uC6FD\uC704\uC705" + //  5920 -  5929
		"\uC708\uC70C\uC714\uC715\uC717\uC719\uC720\uC721\uC724\uC728" + //  5930 -  5939
		"\uC730\uC731\uC733\uC735\uC737\uC73C\uC73D\uC740\uC744\uC74A" + //  5940 -  5949
		"\uC74C\uC74D\uC74F\uC751\uC752\uC753\uC754\uC755\uC756\uC757" + //  5950 -  5959
		"\uC758\uC75C\uC760\uC768\uC76B\uC774\uC775\uC778\uC77C\uC77D" + //  5960 -  5969
		"\uC77E\uC783\uC784\uC785\uC787\uC788\uC789\uC78A\uC78E\uC790" + //  5970 -  5979
		"\uC791\uC794\uC796\uC797\uC798\uC688\uC689\uC68B\uC68D\uC694" + //  5980 -  5989
		"\uC695\uC698\uC69C\uC6A4\uC6A5\uC6A7\uC6A9\uC6B0\uC6B1\uC6B4" + //  5990 -  5999
		"\uC6B8\uC6B9\uC6BA\uC6C0\uC6C1\uC6C3\uC6C5\uC6CC\uC6CD\uC6D0" + //  6000 -  6009
		"\uC6D4\uC6DC\uC6DD\uC6E0\uC6E1\uC6E8\uFFFD\uC5D0\uC5D1\uC5D4" + //  6010 -  6019
		"\uC5D8\uC5E0\uC5E1\uC5E3\uC5E5\uC5EC\uC5ED\uC5EE\uC5F0\uC5F4" + //  6020 -  6029
		"\uC5F6\uC5F7\uC5FC\uC5FD\uC5FE\uC5FF\uC600\uC601\uC605\uC606" + //  6030 -  6039
		"\uC607\uC608\uC60C\uC610\uC618\uC619\uC61B\uC61C\uC624\uC625" + //  6040 -  6049
		"\uC628\uC62C\uC62D\uC62E\uC630\uC633\uC634\uC635\uC637\uC639" + //  6050 -  6059
		"\uC63B\uC640\uC641\uC644\uC648\uC650\uC651\uC653\uC654\uC655" + //  6060 -  6069
		"\uC65C\uC65D\uC660\uC66C\uC66F\uC671\uC678\uC679\uC67C\uC680" + //  6070 -  6079
		"\uC57D\uC580\uC584\uC587\uC58C\uC58D\uC58F\uC591\uC595\uC597" + //  6080 -  6089
		"\uC598\uC59C\uC5A0\uC5A9\uC5B4\uC5B5\uC5B8\uC5B9\uC5BB\uC5BC" + //  6090 -  6099
		"\uC5BD\uC5BE\uC5C4\uC5C5\uC5C6\uC5C7\uC5C8\uC5C9\uC5CA\uC5CC" + //  6100 -  6109
		"\uC5CE\uFFFD\uC434\uC43C\uC43D\uC448\uC464\uC465\uC468\uC46C" + //  6110 -  6119
		"\uC474\uC475\uC479\uC480\uC494\uC49C\uC4B8\uC4BC\uC4E9\uC4F0" + //  6120 -  6129
		"\uC4F1\uC4F4\uC4F8\uC4FA\uC4FF\uC500\uC501\uC50C\uC510\uC514" + //  6130 -  6139
		"\uC51C\uC528\uC529\uC52C\uC530\uC538\uC539\uC53B\uC53D\uC544" + //  6140 -  6149
		"\uC545\uC548\uC549\uC54A\uC54C\uC54D\uC54E\uC553\uC554\uC555" + //  6150 -  6159
		"\uC557\uC558\uC559\uC55D\uC55E\uC560\uC561\uC564\uC568\uC570" + //  6160 -  6169
		"\uC571\uC573\uC574\uC575\uC57C\uC345\uC368\uC369\uC36C\uC370" + //  6170 -  6179
		"\uC372\uC378\uC379\uC37C\uC37D\uC384\uC388\uC38C\uC3C0\uC3D8" + //  6180 -  6189
		"\uC3D9\uC3DC\uC3DF\uC3E0\uC3E2\uC3E8\uC3E9\uC3ED\uC3F4\uC3F5" + //  6190 -  6199
		"\uC3F8\uC408\uC410\uC424\uC42C\uC430\uFFFD\uC22F\uC231\uC232" + //  6200 -  6209
		"\uC234\uC248\uC250\uC251\uC254\uC258\uC260\uC265\uC26C\uC26D" + //  6210 -  6219
		"\uC270\uC274\uC27C\uC27D\uC27F\uC281\uC288\uC289\uC290\uC298" + //  6220 -  6229
		"\uC29B\uC29D\uC2A4\uC2A5\uC2A8\uC2AC\uC2AD\uC2B4\uC2B5\uC2B7" + //  6230 -  6239
		"\uC2B9\uC2DC\uC2DD\uC2E0\uC2E3\uC2E4\uC2EB\uC2EC\uC2ED\uC2EF" + //  6240 -  6249
		"\uC2F1\uC2F6\uC2F8\uC2F9\uC2FB\uC2FC\uC300\uC308\uC309\uC30C" + //  6250 -  6259
		"\uC30D\uC313\uC314\uC315\uC318\uC31C\uC324\uC325\uC328\uC329" + //  6260 -  6269
		"\uC1B0\uC1BD\uC1C4\uC1C8\uC1CC\uC1D4\uC1D7\uC1D8\uC1E0\uC1E4" + //  6270 -  6279
		"\uC1E8\uC1F0\uC1F1\uC1F3\uC1FC\uC1FD\uC200\uC204\uC20C\uC20D" + //  6280 -  6289
		"\uC20F\uC211\uC218\uC219\uC21C\uC21F\uC220\uC228\uC229\uC22B" + //  6290 -  6299
		"\uC22D\uFFFD\uC0E5\uC0E8\uC0EC\uC0F4\uC0F5\uC0F7\uC0F9\uC100" + //  6300 -  6309
		"\uC104\uC108\uC110\uC115\uC11C\uC11D\uC11E\uC11F\uC120\uC123" + //  6310 -  6319
		"\uC124\uC126\uC127\uC12C\uC12D\uC12F\uC130\uC131\uC136\uC138" + //  6320 -  6329
		"\uC139\uC13C\uC140\uC148\uC149\uC14B\uC14C\uC14D\uC154\uC155" + //  6330 -  6339
		"\uC158\uC15C\uC164\uC165\uC167\uC168\uC169\uC170\uC174\uC178" + //  6340 -  6349
		"\uC185\uC18C\uC18D\uC18E\uC190\uC194\uC196\uC19C\uC19D\uC19F" + //  6350 -  6359
		"\uC1A1\uC1A5\uC1A8\uC1A9\uC1AC\uC091\uC094\uC098\uC0A0\uC0A1" + //  6360 -  6369
		"\uC0A3\uC0A5\uC0AC\uC0AD\uC0AF\uC0B0\uC0B3\uC0B4\uC0B5\uC0B6" + //  6370 -  6379
		"\uC0BC\uC0BD\uC0BF\uC0C0\uC0C1\uC0C5\uC0C8\uC0C9\uC0CC\uC0D0" + //  6380 -  6389
		"\uC0D8\uC0D9\uC0DB\uC0DC\uC0DD\uC0E4\uFFFD\uBE68\uBE6A\uBE70" + //  6390 -  6399
		"\uBE71\uBE73\uBE74\uBE75\uBE7B\uBE7C\uBE7D\uBE80\uBE84\uBE8C" + //  6400 -  6409
		"\uBE8D\uBE8F\uBE90\uBE91\uBE98\uBE99\uBEA8\uBED0\uBED1\uBED4" + //  6410 -  6419
		"\uBED7\uBED8\uBEE0\uBEE3\uBEE4\uBEE5\uBEEC\uBF01\uBF08\uBF09" + //  6420 -  6429
		"\uBF18\uBF19\uBF1B\uBF1C\uBF1D\uBF40\uBF41\uBF44\uBF48\uBF50" + //  6430 -  6439
		"\uBF51\uBF55\uBF94\uBFB0\uBFC5\uBFCC\uBFCD\uBFD0\uBFD4\uBFDC" + //  6440 -  6449
		"\uBFDF\uBFE1\uC03C\uC051\uC058\uC05C\uC060\uC068\uC069\uC090" + //  6450 -  6459
		"\uBDD5\uBDD8\uBDDC\uBDE9\uBDF0\uBDF4\uBDF8\uBE00\uBE03\uBE05" + //  6460 -  6469
		"\uBE0C\uBE0D\uBE10\uBE14\uBE1C\uBE1D\uBE1F\uBE44\uBE45\uBE48" + //  6470 -  6479
		"\uBE4C\uBE4E\uBE54\uBE55\uBE57\uBE59\uBE5A\uBE5B\uBE60\uBE61" + //  6480 -  6489
		"\uBE64\uFFFD\uBC99\uBC9A\uBCA0\uBCA1\uBCA4\uBCA7\uBCA8\uBCB0" + //  6490 -  6499
		"\uBCB1\uBCB3\uBCB4\uBCB5\uBCBC\uBCBD\uBCC0\uBCC4\uBCCD\uBCCF" + //  6500 -  6509
		"\uBCD0\uBCD1\uBCD5\uBCD8\uBCDC\uBCF4\uBCF5\uBCF6\uBCF8\uBCFC" + //  6510 -  6519
		"\uBD04\uBD05\uBD07\uBD09\uBD10\uBD14\uBD24\uBD2C\uBD40\uBD48" + //  6520 -  6529
		"\uBD49\uBD4C\uBD50\uBD58\uBD59\uBD64\uBD68\uBD80\uBD81\uBD84" + //  6530 -  6539
		"\uBD87\uBD88\uBD89\uBD8A\uBD90\uBD91\uBD93\uBD95\uBD99\uBD9A" + //  6540 -  6549
		"\uBD9C\uBDA4\uBDB0\uBDB8\uBDD4\uBC1D\uBC1E\uBC1F\uBC24\uBC25" + //  6550 -  6559
		"\uBC27\uBC29\uBC2D\uBC30\uBC31\uBC34\uBC38\uBC40\uBC41\uBC43" + //  6560 -  6569
		"\uBC44\uBC45\uBC49\uBC4C\uBC4D\uBC50\uBC5D\uBC84\uBC85\uBC88" + //  6570 -  6579
		"\uBC8B\uBC8C\uBC8E\uBC94\uBC95\uBC97\uFFFD\uBB00\uBB04\uBB0D" + //  6580 -  6589
		"\uBB0F\uBB11\uBB18\uBB1C\uBB20\uBB29\uBB2B\uBB34\uBB35\uBB36" + //  6590 -  6599
		"\uBB38\uBB3B\uBB3C\uBB3D\uBB3E\uBB44\uBB45\uBB47\uBB49\uBB4D" + //  6600 -  6609
		"\uBB4F\uBB50\uBB54\uBB58\uBB61\uBB63\uBB6C\uBB88\uBB8C\uBB90" + //  6610 -  6619
		"\uBBA4\uBBA8\uBBAC\uBBB4\uBBB7\uBBC0\uBBC4\uBBC8\uBBD0\uBBD3" + //  6620 -  6629
		"\uBBF8\uBBF9\uBBFC\uBBFF\uBC00\uBC02\uBC08\uBC09\uBC0B\uBC0C" + //  6630 -  6639
		"\uBC0D\uBC0F\uBC11\uBC14\uBC15\uBC16\uBC17\uBC18\uBC1B\uBC1C" + //  6640 -  6649
		"\uBA58\uBA5C\uBA64\uBA65\uBA67\uBA68\uBA69\uBA70\uBA71\uBA74" + //  6650 -  6659
		"\uBA78\uBA83\uBA84\uBA85\uBA87\uBA8C\uBAA8\uBAA9\uBAAB\uBAAC" + //  6660 -  6669
		"\uBAB0\uBAB2\uBAB8\uBAB9\uBABB\uBABD\uBAC4\uBAC8\uBAD8\uBAD9" + //  6670 -  6679
		"\uBAFC\uFFFD\uB96B\uB96D\uB974\uB975\uB978\uB97C\uB984\uB985" + //  6680 -  6689
		"\uB987\uB989\uB98A\uB98D\uB98E\uB9AC\uB9AD\uB9B0\uB9B4\uB9BC" + //  6690 -  6699
		"\uB9BD\uB9BF\uB9C1\uB9C8\uB9C9\uB9CC\uB9CE\uB9CF\uB9D0\uB9D1" + //  6700 -  6709
		"\uB9D2\uB9D8\uB9D9\uB9DB\uB9DD\uB9DE\uB9E1\uB9E3\uB9E4\uB9E5" + //  6710 -  6719
		"\uB9E8\uB9EC\uB9F4\uB9F5\uB9F7\uB9F8\uB9F9\uB9FA\uBA00\uBA01" + //  6720 -  6729
		"\uBA08\uBA15\uBA38\uBA39\uBA3C\uBA40\uBA42\uBA48\uBA49\uBA4B" + //  6730 -  6739
		"\uBA4D\uBA4E\uBA53\uBA54\uBA55\uB8C5\uB8CC\uB8D0\uB8D4\uB8DD" + //  6740 -  6749
		"\uB8DF\uB8E1\uB8E8\uB8E9\uB8EC\uB8F0\uB8F8\uB8F9\uB8FB\uB8FD" + //  6750 -  6759
		"\uB904\uB918\uB920\uB93C\uB93D\uB940\uB944\uB94C\uB94F\uB951" + //  6760 -  6769
		"\uB958\uB959\uB95C\uB960\uB968\uB969\uFFFD\uB798\uB799\uB79C" + //  6770 -  6779
		"\uB7A0\uB7A8\uB7A9\uB7AB\uB7AC\uB7AD\uB7B4\uB7B5\uB7B8\uB7C7" + //  6780 -  6789
		"\uB7C9\uB7EC\uB7ED\uB7F0\uB7F4\uB7FC\uB7FD\uB7FF\uB800\uB801" + //  6790 -  6799
		"\uB807\uB808\uB809\uB80C\uB810\uB818\uB819\uB81B\uB81D\uB824" + //  6800 -  6809
		"\uB825\uB828\uB82C\uB834\uB835\uB837\uB838\uB839\uB840\uB844" + //  6810 -  6819
		"\uB851\uB853\uB85C\uB85D\uB860\uB864\uB86C\uB86D\uB86F\uB871" + //  6820 -  6829
		"\uB878\uB87C\uB88D\uB8A8\uB8B0\uB8B4\uB8B8\uB8C0\uB8C1\uB8C3" + //  6830 -  6839
		"\uB729\uB72C\uB72F\uB730\uB738\uB739\uB73B\uB744\uB748\uB74C" + //  6840 -  6849
		"\uB754\uB755\uB760\uB764\uB768\uB770\uB771\uB773\uB775\uB77C" + //  6850 -  6859
		"\uB77D\uB780\uB784\uB78C\uB78D\uB78F\uB790\uB791\uB792\uB796" + //  6860 -  6869
		"\uB797\uFFFD\uB540\uB541\uB543\uB544\uB545\uB54B\uB54C\uB54D" + //  6870 -  6879
		"\uB550\uB554\uB55C\uB55D\uB55F\uB560\uB561\uB5A0\uB5A1\uB5A4" + //  6880 -  6889
		"\uB5A8\uB5AA\uB5AB\uB5B0\uB5B1\uB5B3\uB5B4\uB5B5\uB5BB\uB5BC" + //  6890 -  6899
		"\uB5BD\uB5C0\uB5C4\uB5CC\uB5CD\uB5CF\uB5D0\uB5D1\uB5D8\uB5EC" + //  6900 -  6909
		"\uB610\uB611\uB614\uB618\uB625\uB62C\uB634\uB648\uB664\uB668" + //  6910 -  6919
		"\uB69C\uB69D\uB6A0\uB6A4\uB6AB\uB6AC\uB6B1\uB6D4\uB6F0\uB6F4" + //  6920 -  6929
		"\uB6F8\uB700\uB701\uB705\uB728\uB4C0\uB4C4\uB4C8\uB4D0\uB4D5" + //  6930 -  6939
		"\uB4DC\uB4DD\uB4E0\uB4E3\uB4E4\uB4E6\uB4EC\uB4ED\uB4EF\uB4F1" + //  6940 -  6949
		"\uB4F8\uB514\uB515\uB518\uB51B\uB51C\uB524\uB525\uB527\uB528" + //  6950 -  6959
		"\uB529\uB52A\uB530\uB531\uB534\uB538\uFFFD\uB367\uB369\uB36B" + //  6960 -  6969
		"\uB36E\uB370\uB371\uB374\uB378\uB380\uB381\uB383\uB384\uB385" + //  6970 -  6979
		"\uB38C\uB390\uB394\uB3A0\uB3A1\uB3A8\uB3AC\uB3C4\uB3C5\uB3C8" + //  6980 -  6989
		"\uB3CB\uB3CC\uB3CE\uB3D0\uB3D4\uB3D5\uB3D7\uB3D9\uB3DB\uB3DD" + //  6990 -  6999
		"\uB3E0\uB3E4\uB3E8\uB3FC\uB410\uB418\uB41C\uB420\uB428\uB429" + //  7000 -  7009
		"\uB42B\uB434\uB450\uB451\uB454\uB458\uB460\uB461\uB463\uB465" + //  7010 -  7019
		"\uB46C\uB480\uB488\uB49D\uB4A4\uB4A8\uB4AC\uB4B5\uB4B7\uB4B9" + //  7020 -  7029
		"\uB2EE\uB2EF\uB2F3\uB2F4\uB2F5\uB2F7\uB2F8\uB2F9\uB2FA\uB2FB" + //  7030 -  7039
		"\uB2FF\uB300\uB301\uB304\uB308\uB310\uB311\uB313\uB314\uB315" + //  7040 -  7049
		"\uB31C\uB354\uB355\uB356\uB358\uB35B\uB35C\uB35E\uB35F\uB364" + //  7050 -  7059
		"\uB365\uFFFD\uB1DF\uB1E8\uB1E9\uB1EC\uB1F0\uB1F9\uB1FB\uB1FD" + //  7060 -  7069
		"\uB204\uB205\uB208\uB20B\uB20C\uB214\uB215\uB217\uB219\uB220" + //  7070 -  7079
		"\uB234\uB23C\uB258\uB25C\uB260\uB268\uB269\uB274\uB275\uB27C" + //  7080 -  7089
		"\uB284\uB285\uB289\uB290\uB291\uB294\uB298\uB299\uB29A\uB2A0" + //  7090 -  7099
		"\uB2A1\uB2A3\uB2A5\uB2A6\uB2AA\uB2AC\uB2B0\uB2B4\uB2C8\uB2C9" + //  7100 -  7109
		"\uB2CC\uB2D0\uB2D2\uB2D8\uB2D9\uB2DB\uB2DD\uB2E2\uB2E4\uB2E5" + //  7110 -  7119
		"\uB2E6\uB2E8\uB2EB\uB2EC\uB2ED\uB140\uB141\uB144\uB148\uB150" + //  7120 -  7129
		"\uB151\uB154\uB155\uB158\uB15C\uB160\uB178\uB179\uB17C\uB180" + //  7130 -  7139
		"\uB182\uB188\uB189\uB18B\uB18D\uB192\uB193\uB194\uB198\uB19C" + //  7140 -  7149
		"\uB1A8\uB1CC\uB1D0\uB1D4\uB1DC\uB1DD\uFFFD\uB05D\uB07C\uB07D" + //  7150 -  7159
		"\uB080\uB084\uB08C\uB08D\uB08F\uB091\uB098\uB099\uB09A\uB09C" + //  7160 -  7169
		"\uB09F\uB0A0\uB0A1\uB0A2\uB0A8\uB0A9\uB0AB\uB0AC\uB0AD\uB0AE" + //  7170 -  7179
		"\uB0AF\uB0B1\uB0B3\uB0B4\uB0B5\uB0B8\uB0BC\uB0C4\uB0C5\uB0C7" + //  7180 -  7189
		"\uB0C8\uB0C9\uB0D0\uB0D1\uB0D4\uB0D8\uB0E0\uB0E5\uB108\uB109" + //  7190 -  7199
		"\uB10B\uB10C\uB110\uB112\uB113\uB118\uB119\uB11B\uB11C\uB11D" + //  7200 -  7209
		"\uB123\uB124\uB125\uB128\uB12C\uB134\uB135\uB137\uB138\uB139" + //  7210 -  7219
		"\uAFCB\uAFCD\uAFCE\uAFD4\uAFDC\uAFE8\uAFE9\uAFF0\uAFF1\uAFF4" + //  7220 -  7229
		"\uAFF8\uB000\uB001\uB004\uB00C\uB010\uB014\uB01C\uB01D\uB028" + //  7230 -  7239
		"\uB044\uB045\uB048\uB04A\uB04C\uB04E\uB053\uB054\uB055\uB057" + //  7240 -  7249
		"\uB059\uFFFD\uAE79\uAE7B\uAE7C\uAE7D\uAE84\uAE85\uAE8C\uAEBC" + //  7250 -  7259
		"\uAEBD\uAEBE\uAEC0\uAEC4\uAECC\uAECD\uAECF\uAED0\uAED1\uAED8" + //  7260 -  7269
		"\uAED9\uAEDC\uAEE8\uAEEB\uAEED\uAEF4\uAEF8\uAEFC\uAF07\uAF08" + //  7270 -  7279
		"\uAF0D\uAF10\uAF2C\uAF2D\uAF30\uAF32\uAF34\uAF3C\uAF3D\uAF3F" + //  7280 -  7289
		"\uAF41\uAF42\uAF43\uAF48\uAF49\uAF50\uAF5C\uAF5D\uAF64\uAF65" + //  7290 -  7299
		"\uAF79\uAF80\uAF84\uAF88\uAF90\uAF91\uAF95\uAF9C\uAFB8\uAFB9" + //  7300 -  7309
		"\uAFBC\uAFC0\uAFC7\uAFC8\uAFC9\uAE0D\uAE14\uAE30\uAE31\uAE34" + //  7310 -  7319
		"\uAE37\uAE38\uAE3A\uAE40\uAE41\uAE43\uAE45\uAE46\uAE4A\uAE4C" + //  7320 -  7329
		"\uAE4D\uAE4E\uAE50\uAE54\uAE56\uAE5C\uAE5D\uAE5F\uAE60\uAE61" + //  7330 -  7339
		"\uAE65\uAE68\uAE69\uAE6C\uAE70\uAE78\uFFFD\uAD0C\uAD0D\uAD0F" + //  7340 -  7349
		"\uAD11\uAD18\uAD1C\uAD20\uAD29\uAD2C\uAD2D\uAD34\uAD35\uAD38" + //  7350 -  7359
		"\uAD3C\uAD44\uAD45\uAD47\uAD49\uAD50\uAD54\uAD58\uAD61\uAD63" + //  7360 -  7369
		"\uAD6C\uAD6D\uAD70\uAD73\uAD74\uAD75\uAD76\uAD7B\uAD7C\uAD7D" + //  7370 -  7379
		"\uAD7F\uAD81\uAD82\uAD88\uAD89\uAD8C\uAD90\uAD9C\uAD9D\uADA4" + //  7380 -  7389
		"\uADB7\uADC0\uADC1\uADC4\uADC8\uADD0\uADD1\uADD3\uADDC\uADE0" + //  7390 -  7399
		"\uADE4\uADF8\uADF9\uADFC\uADFF\uAE00\uAE01\uAE08\uAE09\uAE0B" + //  7400 -  7409
		"\uACAF\uACB0\uACB8\uACB9\uACBB\uACBC\uACBD\uACC1\uACC4\uACC8" + //  7410 -  7419
		"\uACCC\uACD5\uACD7\uACE0\uACE1\uACE4\uACE7\uACE8\uACEA\uACEC" + //  7420 -  7429
		"\uACEF\uACF0\uACF1\uACF3\uACF5\uACF6\uACFC\uACFD\uAD00\uAD04" + //  7430 -  7439
		"\uAD06\uFFFD\uAC00\uAC01\uAC04\uAC07\uAC08\uAC09\uAC0A\uAC10" + //  7440 -  7449
		"\uAC11\uAC12\uAC13\uAC14\uAC15\uAC16\uAC17\uAC19\uAC1A\uAC1B" + //  7450 -  7459
		"\uAC1C\uAC1D\uAC20\uAC24\uAC2C\uAC2D\uAC2F\uAC30\uAC31\uAC38" + //  7460 -  7469
		"\uAC39\uAC3C\uAC40\uAC4B\uAC4D\uAC54\uAC58\uAC5C\uAC70\uAC71" + //  7470 -  7479
		"\uAC74\uAC77\uAC78\uAC7A\uAC80\uAC81\uAC83\uAC84\uAC85\uAC86" + //  7480 -  7489
		"\uAC89\uAC8A\uAC8B\uAC8C\uAC90\uAC94\uAC9C\uAC9D\uAC9F\uACA0" + //  7490 -  7499
		"\uACA1\uACA8\uACA9\uACAA\uACAC\u043E\u043F\u0440\u0441\u0442" + //  7500 -  7509
		"\u0443\u0444\u0445\u0446\u0447\u0448\u0449\u044A\u044B\u044C" + //  7510 -  7519
		"\u044D\u044E\u044F\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  7520 -  7529
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uE1D6\uE1D7\uE1D8" + //  7530 -  7539
		"\uE1D9\uE1DA\uE1DB\uE1DC\uE1DD\uE1DE\uE1DF\uE1E0\uE1E1\uE1E2" + //  7540 -  7549
		"\uE1E3\uE1E4\uE1E5\uE1E6\uE1E7\uE1E8\uE1E9\uE1EA\uE1EB\uE1EC" + //  7550 -  7559
		"\uE1ED\uE1EE\uE1EF\uE1F0\uE1F1\uE1F2\uE1F3\uE1F4\u042E\u042F" + //  7560 -  7569
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  7570 -  7579
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\u0430\u0431\u0432\u0433\u0434" + //  7580 -  7589
		"\u0435\u0451\u0436\u0437\u0438\u0439\u043A\u043B\u043C\u043D" + //  7590 -  7599
		"\u30E0\u30E1\u30E2\u30E3\u30E4\u30E5\u30E6\u30E7\u30E8\u30E9" + //  7600 -  7609
		"\u30EA\u30EB\u30EC\u30ED\u30EE\u30EF\u30F0\u30F1\u30F2\u30F3" + //  7610 -  7619
		"\u30F4\u30F5\u30F6\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  7620 -  7629
		"\uFFFD\uFFFD\uE292\uE293\uE294\uE295\uE296\uE297\uE298\uE299" + //  7630 -  7639
		"\uE29A\uE29B\uE29C\uE29D\uE29E\uE29F\uE2A0\uE2A1\uE2A2\uE2A3" + //  7640 -  7649
		"\uE2A4\uE2A5\uE2A6\uE2A7\uE2A8\uE2A9\uE2AA\uE2AB\uE2AC\uE2AD" + //  7650 -  7659
		"\uE2AE\uE2AF\uE2B0\u30C0\u30C1\u30C2\u30C3\u30C4\u30C5\u30C6" + //  7660 -  7669
		"\u30C7\u30C8\u30C9\u30CA\u30CB\u30CC\u30CD\u30CE\u30CF\u30D0" + //  7670 -  7679
		"\u30D1\u30D2\u30D3\u30D4\u30D5\u30D6\u30D7\u30D8\u30D9\u30DA" + //  7680 -  7689
		"\u30DB\u30DC\u30DD\u30DE\u30DF\u3080\u3081\u3082\u3083\u3084" + //  7690 -  7699
		"\u3085\u3086\u3087\u3088\u3089\u308A\u308B\u308C\u308D\u308E" + //  7700 -  7709
		"\u308F\u3090\u3091\u3092\u3093\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  7710 -  7719
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uE234\uE235\uE236" + //  7720 -  7729
		"\uE237\uE238\uE239\uE23A\uE23B\uE23C\uE23D\uE23E\uE23F\uE240" + //  7730 -  7739
		"\uE241\uE242\uE243\uE244\uE245\uE246\uE247\uE248\uE249\uE24A" + //  7740 -  7749
		"\uE24B\uE24C\uE24D\uE24E\uE24F\uE250\uE251\uE252\u3060\u3061" + //  7750 -  7759
		"\u3062\u3063\u3064\u3065\u3066\u3067\u3068\u3069\u306A\u306B" + //  7760 -  7769
		"\u306C\u306D\u306E\u306F\u3070\u3071\u3072\u3073\u3074\u3075" + //  7770 -  7779
		"\u3076\u3077\u3078\u3079\u307A\u307B\u307C\u307D\u307E\u307F" + //  7780 -  7789
		"\u24AF\u24B0\u24B1\u24B2\u24B3\u24B4\u24B5\u2474\u2475\u2476" + //  7790 -  7799
		"\u2477\u2478\u2479\u247A\u247B\u247C\u247D\u247E\u247F\u2480" + //  7800 -  7809
		"\u2481\u2482\u00B9\u00B2\u00B3\u2074\u207F\u2081\u2082\u2083" + //  7810 -  7819
		"\u2084\uFFFD\u0410\u0411\u0412\u0413\u0414\u0415\u0401\u0416" + //  7820 -  7829
		"\u0417\u0418\u0419\u041A\u041B\u041C\u041D\u041E\u041F\u0420" + //  7830 -  7839
		"\u0421\u0422\u0423\u0424\u0425\u0426\u0427\u0428\u0429\u042A" + //  7840 -  7849
		"\u042B\u042C\u042D\u320F\u3210\u3211\u3212\u3213\u3214\u3215" + //  7850 -  7859
		"\u3216\u3217\u3218\u3219\u321A\u321B\u249C\u249D\u249E\u249F" + //  7860 -  7869
		"\u24A0\u24A1\u24A2\u24A3\u24A4\u24A5\u24A6\u24A7\u24A8\u24A9" + //  7870 -  7879
		"\u24AA\u24AB\u24AC\u24AD\u24AE\u24E3\u24E4\u24E5\u24E6\u24E7" + //  7880 -  7889
		"\u24E8\u24E9\u2460\u2461\u2462\u2463\u2464\u2465\u2466\u2467" + //  7890 -  7899
		"\u2468\u2469\u246A\u246B\u246C\u246D\u246E\u00BD\u2153\u2154" + //  7900 -  7909
		"\u00BC\u00BE\u215B\u215C\u215D\u215E\uFFFD\u30A1\u30A2\u30A3" + //  7910 -  7919
		"\u30A4\u30A5\u30A6\u30A7\u30A8\u30A9\u30AA\u30AB\u30AC\u30AD" + //  7920 -  7929
		"\u30AE\u30AF\u30B0\u30B1\u30B2\u30B3\u30B4\u30B5\u30B6\u30B7" + //  7930 -  7939
		"\u30B8\u30B9\u30BA\u30BB\u30BC\u30BD\u30BE\u30BF\u326F\u3270" + //  7940 -  7949
		"\u3271\u3272\u3273\u3274\u3275\u3276\u3277\u3278\u3279\u327A" + //  7950 -  7959
		"\u327B\u24D0\u24D1\u24D2\u24D3\u24D4\u24D5\u24D6\u24D7\u24D8" + //  7960 -  7969
		"\u24D9\u24DA\u24DB\u24DC\u24DD\u24DE\u24DF\u24E0\u24E1\u24E2" + //  7970 -  7979
		"\u33C5\u33AD\u33AE\u33AF\u33DB\u33A9\u33AA\u33AB\u33AC\u33DD" + //  7980 -  7989
		"\u33D0\u33D3\u33C3\u33C9\u33DC\u33C6\uFFFD\uFFFD\uFFFD\uFFFD" + //  7990 -  7999
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  8000 -  8009
		"\uFFFD\uFFFD\uE178\uE179\uE17A\uE17B\uE17C\uE17D\uE17E\uE17F" + //  8010 -  8019
		"\uE180\uE181\uE182\uE183\uE184\uE185\uE186\uE187\uE188\uE189" + //  8020 -  8029
		"\uE18A\uE18B\uE18C\uE18D\uE18E\uE18F\uE190\uE191\uE192\uE193" + //  8030 -  8039
		"\uE194\uE195\uE196\u33B1\u33B2\u33B3\u33B4\u33B5\u33B6\u33B7" + //  8040 -  8049
		"\u33B8\u33B9\u3380\u3381\u3382\u3383\u3384\u33BA\u33BB\u33BC" + //  8050 -  8059
		"\u33BD\u33BE\u33BF\u3390\u3391\u3392\u3393\u3394\u2126\u33C0" + //  8060 -  8069
		"\u33C1\u338A\u338B\u338C\u33D6\u2546\u2547\u2548\u2549\u254A" + //  8070 -  8079
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  8080 -  8089
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  8090 -  8099
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  8100 -  8109
		"\uFFFD\uFFFD\uE0BC\uE0BD\uE0BE\uE0BF\uE0C0\uE0C1\uE0C2\uE0C3" + //  8110 -  8119
		"\uE0C4\uE0C5\uE0C6\uE0C7\uE0C8\uE0C9\uE0CA\uE0CB\uE0CC\uE0CD" + //  8120 -  8129
		"\uE0CE\uE0CF\uE0D0\uE0D1\uE0D2\uE0D3\uE0D4\uE0D5\uE0D6\uE0D7" + //  8130 -  8139
		"\uE0D8\uE0D9\uE0DA\u2542\u2512\u2511\u251A\u2519\u2516\u2515" + //  8140 -  8149
		"\u250E\u250D\u251E\u251F\u2521\u2522\u2526\u2527\u2529\u252A" + //  8150 -  8159
		"\u252D\u252E\u2531\u2532\u2535\u2536\u2539\u253A\u253D\u253E" + //  8160 -  8169
		"\u2540\u2541\u2543\u2544\u2545\u3170\u3171\u3172\u3173\u3174" + //  8170 -  8179
		"\u3175\u3176\u3177\u3178\u3179\u317A\u317B\u317C\u317D\u317E" + //  8180 -  8189
		"\u317F\u3180\u3181\u3182\u3183\u3184\u3185\u3186\u3187\u3188" + //  8190 -  8199
		"\u3189\u318A\u318B\u318C\u318D\u318E\uFFFD\u3041\u3042\u3043" + //  8200 -  8209
		"\u3044\u3045\u3046\u3047\u3048\u3049\u304A\u304B\u304C\u304D" + //  8210 -  8219
		"\u304E\u304F\u3050\u3051\u3052\u3053\u3054\u3055\u3056\u3057" + //  8220 -  8229
		"\u3058\u3059\u305A\u305B\u305C\u305D\u305E\u305F\u3150\u3151" + //  8230 -  8239
		"\u3152\u3153\u3154\u3155\u3156\u3157\u3158\u3159\u315A\u315B" + //  8240 -  8249
		"\u315C\u315D\u315E\u315F\u3160\u3161\u3162\u3163\u3164\u3165" + //  8250 -  8259
		"\u3166\u3167\u3168\u3169\u316A\u316B\u316C\u316D\u316E\u316F" + //  8260 -  8269
		"\uFF40\uFF41\uFF42\uFF43\uFF44\uFF45\uFF46\uFF47\uFF48\uFF49" + //  8270 -  8279
		"\uFF4A\uFF4B\uFF4C\uFF4D\uFF4E\uFF4F\uFF50\uFF51\uFF52\uFF53" + //  8280 -  8289
		"\uFF54\uFF55\uFF56\uFF57\uFF58\uFF59\uFF5A\uFF5B\uFF5C\uFF5D" + //  8290 -  8299
		"\uFFE3\uFFFD\u00E6\u0111\u00F0\u0127\u0131\u0133\u0138\u0140" + //  8300 -  8309
		"\u0142\u00F8\u0153\u00DF\u00FE\u0167\u014B\u0149\u3200\u3201" + //  8310 -  8319
		"\u3202\u3203\u3204\u3205\u3206\u3207\u3208\u3209\u320A\u320B" + //  8320 -  8329
		"\u320C\u320D\u320E\uFF20\uFF21\uFF22\uFF23\uFF24\uFF25\uFF26" + //  8330 -  8339
		"\uFF27\uFF28\uFF29\uFF2A\uFF2B\uFF2C\uFF2D\uFF2E\uFF2F\uFF30" + //  8340 -  8349
		"\uFF31\uFF32\uFF33\uFF34\uFF35\uFF36\uFF37\uFF38\uFF39\uFF3A" + //  8350 -  8359
		"\uFF3B\uFFE6\uFF3D\uFF3E\uFF3F\u2116\u33C7\u2122\u33C2\u33D8" + //  8360 -  8369
		"\u2121\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  8370 -  8379
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  8380 -  8389
		"\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uE11A\uE11B\uE11C" + //  8390 -  8399
		"\uE11D\uE11E\uE11F\uE120\uE121\uE122\uE123\uE124\uE125\uE126" + //  8400 -  8409
		"\uE127\uE128\uE129\uE12A\uE12B\uE12C\uE12D\uE12E\uE12F\uE130" + //  8410 -  8419
		"\uE131\uE132\uE133\uE134\uE135\uE136\uE137\uE138\u2663\u25C9" + //  8420 -  8429
		"\u25C8\u25A3\u25D0\u25D1\u2592\u25A4\u25A5\u25A8\u25A7\u25A6" + //  8430 -  8439
		"\u25A9\u2668\u260F\u260E\u261C\u261E\u00B6\u2020\u2021\u2195" + //  8440 -  8449
		"\u2197\u2199\u2196\u2198\u266D\u2669\u266A\u266C\u327F\u321C" + //  8450 -  8459
		"\u25A1\u25A0\u25B3\u25B2\u25BD\u25BC\u2192\u2190\u2191\u2193" + //  8460 -  8469
		"\u2194\u3013\u226A\u226B\u221A\u223D\u221D\u2235\u222B\u222C" + //  8470 -  8479
		"\u2208\u220B\u2286\u2287\u2282\u2283\u222A\u2229\u2227\u2228" + //  8480 -  8489
		"\uFFE2\uFFFD\u00C6\u00D0\u00AA\u0126\uFFFD\u0132\uFFFD\u013F" + //  8490 -  8499
		"\u0141\u00D8\u0152\u00BA\u00DE\u0166\u014A\uFFFD\u3260\u3261" + //  8500 -  8509
		"\u3262\u3263\u3264\u3265\u3266\u3267\u3268\u3269\u326A\u326B" + //  8510 -  8519
		"\u326C\u326D\u326E\u00F7\u2260\u2264\u2265\u221E\u2234\u00B0" + //  8520 -  8529
		"\u2032\u2033\u2103\u212B\uFFE0\uFFE1\uFFE5\u2642\u2640\u2220" + //  8530 -  8539
		"\u22A5\u2312\u2202\u2207\u2261\u2252\u00A7\u203B\u2606\u2605" + //  8540 -  8549
		"\u25CB\u25CF\u25CE\u25C7\u25C6\uD277\uD293\uD2CD\uD2E7\uD30A" + //  8550 -  8559
		"\uD326\uD359\uD360\uD3B2\uD3B5\uD3C7\uD424\uD4B0\uD4E9\uD505" + //  8560 -  8569
		"\uD510\uD519\uD520\uD524\uD55F\uD561\uD56C\uD571\uD5AC\uD5CF" + //  8570 -  8579
		"\uD62C\uD6A9\uD6B8\uD6D5\uD70C\uD76D\uFFFD\u3395\u3396\u3397" + //  8580 -  8589
		"\u2113\u3398\u33C4\u33A3\u33A4\u33A5\u33A6\u3399\u339A\u339B" + //  8590 -  8599
		"\u339C\u339D\u339E\u339F\u33A0\u33A1\u33A2\u33CA\u338D\u338E" + //  8600 -  8609
		"\u338F\u33CF\u3388\u3389\u33C8\u33A7\u33A8\u33B0\uCAD2\uCB2C" + //  8610 -  8619
		"\uCB80\uCBE5\uCBF0\uCC1F\uCC26\uCC2F\uCC3F\uCC42\uCC71\uCC7C" + //  8620 -  8629
		"\uCCE3\uCCE5\uCD40\uCDC3\uCE3C\uCE7B\uCE97\uCEA9\uCEC8\uCEFD" + //  8630 -  8639
		"\uCF19\uCF8D\uCF90\uCF9F\uCFAC\uCFBD\uD088\uD114\uD160\uD169" + //  8640 -  8649
		"\uC2B3\uC2C0\uC2E6\uC302\uC30B\uC327\uC330\uC343\uC34C\uC37B" + //  8650 -  8659
		"\uC385\uC399\uC3A0\uC3BC\uC3FC\uC43F\uC477\uC493\uC4D3\uC4D4" + //  8660 -  8669
		"\uC53C\uC53F\uC54F\uC55F\uC590\uC5AB\uC5B6\uC5F1\uC5F3\uC61D" + //  8670 -  8679
		"\uC62B\uFFFD\u2500\u2502\u250C\u2510\u2518\u2514\u251C\u252C" + //  8680 -  8689
		"\u2524\u2534\u253C\u2501\u2503\u250F\u2513\u251B\u2517\u2523" + //  8690 -  8699
		"\u2533\u252B\u253B\u254B\u2520\u252F\u2528\u2537\u253F\u251D" + //  8700 -  8709
		"\u2530\u2525\u2538\uBE28\uBE4B\uBE9C\uBEB4\uBEED\uBEF0\uBEF4" + //  8710 -  8719
		"\uBEFF\uBF24\uBF5C\uBF78\uBFC0\uBFD5\uBFDD\uBFE8\uC004\uC020" + //  8720 -  8729
		"\uC059\uC074\uC0AE\uC0B7\uC0BB\uC0C3\uC0C7\uC0CF\uC125\uC13F" + //  8730 -  8739
		"\uC151\uC157\uC193\uC1BB\uC28C\uB2D1\uB2E0\uB331\uB338\uB368" + //  8740 -  8749
		"\uB36A\uB39C\uB3D3\uB400\uB40F\uB42C\uB457\uB47F\uB4B4\uB4C1" + //  8750 -  8759
		"\uB4E7\uB52E\uB532\uB537\uB53F\uB568\uB584\uB5F4\uB680\uB6B8" + //  8760 -  8769
		"\uB70C\uB7D0\uB80F\uB894\uB8DC\uB917\uFFFD\u03B1\u03B2\u03B3" + //  8770 -  8779
		"\u03B4\u03B5\u03B6\u03B7\u03B8\u03B9\u03BA\u03BB\u03BC\u03BD" + //  8780 -  8789
		"\u03BE\u03BF\u03C0\u03C1\u03C3\u03C4\u03C5\u03C6\u03C7\u03C8" + //  8790 -  8799
		"\u03C9\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uE34E\uE34F" + //  8800 -  8809
		"\uE350\uE351\uE352\uE353\uE354\uE355\uE356\uE357\uE358\uE359" + //  8810 -  8819
		"\uE35A\uE35B\uE35C\uE35D\uE35E\uE35F\uE360\uE361\uE362\uE363" + //  8820 -  8829
		"\uE364\uE365\uE366\uE367\uE368\uE369\uE36A\uE36B\uE36C\uAE49" + //  8830 -  8839
		"\uAE62\uAEA0\uAF04\uAF33\uAF4C\uAF58\uAF5B\uAF68\uAF93\uAFB2" + //  8840 -  8849
		"\uAFBF\uAFD8\uAFE7\uB00D\uB021\uB060\uB090\uB0BB\uB0EC\uB10F" + //  8850 -  8859
		"\uB11E\uB147\uB153\uB159\uB16F\uB17A\uB1A7\uB1B0\uB233\uB2A7" + //  8860 -  8869
		"\uB2C1\u51B2\u6A47\uF869\u5DF5\u5FB4\u9D44\u5FF1\u62C6\u6A50" + //  8870 -  8879
		"\u99C4\uF86A\u8759\u5E96\u70AE\u8216\u924B\u9784\uF86B\u84D6" + //  8880 -  8889
		"\u8E55\u7627\uF86C\u9DF3\u7095\u5EE8\uF86D\u7BCB\uF86E\u769E" + //  8890 -  8899
		"\u9190\u9DBB\uFFFD\u0391\u0392\u0393\u0394\u0395\u0396\u0397" + //  8900 -  8909
		"\u0398\u0399\u039A\u039B\u039C\u039D\u039E\u039F\u03A0\u03A1" + //  8910 -  8919
		"\u03A3\u03A4\u03A5\u03A6\u03A7\u03A8\u03A9\uFFFD\uFFFD\uFFFD" + //  8920 -  8929
		"\uFFFD\uFFFD\uFFFD\uFFFD\uE2F0\uE2F1\uE2F2\uE2F3\uE2F4\uE2F5" + //  8930 -  8939
		"\uE2F6\uE2F7\uE2F8\uE2F9\uE2FA\uE2FB\uE2FC\uE2FD\uE2FE\uE2FF" + //  8940 -  8949
		"\uE300\uE301\uE302\uE303\uE304\uE305\uE306\uE307\uE308\uE309" + //  8950 -  8959
		"\uE30A\uE30B\uE30C\uE30D\uE30E\u75E3\uF865\u615A\u5231\u60B5" + //  8960 -  8969
		"\u6C05\u7C00\u8734\u8E91\u6FFA\u7C37\u873B\u780C\u9746\u5CED" + //  8970 -  8979
		"\u7D83\u9214\u9798\uF866\u8E85\u9AD1\u6031\u8471\u6467\uF867" + //  8980 -  8989
		"\u7503\u7B92\u97A6\u9E81\u9EA4\uF868\u8233\u9C2E\u8558\u8202" + //  8990 -  8999
		"\u86F9\u5401\u71A8\u873F\u5E43\u885E\u56FF\u5E37\u8564\u9EDD" + //  9000 -  9009
		"\u9B3B\u6ABC\u73E2\u9F66\u6339\u682E\u9823\u4EDE\u7725\u7CA2" + //  9010 -  9019
		"\u8014\u89DC\u8D6D\u67DE\u6F5C\u8695\u5D82\u7634\uFFFD\u2170" + //  9020 -  9029
		"\u2171\u2172\u2173\u2174\u2175\u2176\u2177\u2178\u2179\uFFFD" + //  9030 -  9039
		"\uFFFD\uFFFD\uFFFD\uFFFD\u2160\u2161\u2162\u2163\u2164\u2165" + //  9040 -  9049
		"\u2166\u2167\u2168\u2169\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD\uFFFD" + //  9050 -  9059
		"\uE3AC\uE3AD\uE3AE\uE3AF\uE3B0\uE3B1\uE3B2\uE3B3\uE3B4\uE3B5" + //  9060 -  9069
		"\uE3B6\uE3B7\uE3B8\uE3B9\uE3BA\uE3BB\uE3BC\uE3BD\uE3BE\uE3BF" + //  9070 -  9079
		"\uE3C0\uE3C1\uE3C2\uE3C3\uE3C4\uE3C5\uE3C6\uE3C7\uE3C8\uE3C9" + //  9080 -  9089
		"\uE3CA\u9D08\u621E\u904F\u5D52\u8AF3\u9EEF\u9785\u6B38\u769A" + //  9090 -  9099
		"\u7919\u9749\u9628\uF860\u7BDB\u7C65\u7F98\u6554\u605A\uF861" + //  9100 -  9109
		"\uF862\u81D9\u8815\u8B8C\u5869\u995C\u5B30\u7768\u7FF3\u854B" + //  9110 -  9119
		"\u9068\u5ABC\u8580\uF851\u8561\uF852\u9AF4\u9EFB\u59A3\uF853" + //  9120 -  9129
		"\u6C98\u7765\u7BE6\u8153\u8F61\u9AC0\u64EF\u860B\uF854\u9870" + //  9130 -  9139
		"\u9B22\u59D2\uF855\u6942\u69CE\u7B25\u69CA\u9460\u6B43\u9364" + //  9140 -  9149
		"\u970E\u6BA4\u9C13\u566C\uFFFD\u3131\u3132\u3133\u3134\u3135" + //  9150 -  9159
		"\u3136\u3137\u3138\u3139\u313A\u313B\u313C\u313D\u313E\u313F" + //  9160 -  9169
		"\u3140\u3141\u3142\u3143\u3144\u3145\u3146\u3147\u3148\u3149" + //  9170 -  9179
		"\u314A\u314B\u314C\u314D\u314E\u314F\u927E\u6FDB\u77C7\u7030" + //  9180 -  9189
		"\u7CDC\u95A9\uF84E\u6B02\u7254\u80D6\u9AE3\u9B74\uF84F\u7FFB" + //  9190 -  9199
		"\u8F9F\u6C74\u8FAE\uF850\u99E2\u5F46\u8FF8\u9D07\u9EFC\u8760" + //  9200 -  9209
		"\u4E30\u8451\u4EC6\u7F58\u82FB\u8709\u982B\u9B92\u95E5\u97C3" + //  9210 -  9219
		"\u515A\u87F7\u7893\u83DF\u5484\u578C\u809A\u86AA\u6ED5\u706F" + //  9220 -  9229
		"\u9419\u7296\u5E71\u57D3\u6994\u6DBC\u9B4E\u7658\u8182\u8821" + //  9230 -  9239
		"\u9462\u6ADF\u9B23\u6624\u6CE0\u82D3\u86C9\u6F66\u826B\uFFFD" + //  9240 -  9249
		"\uFF01\uFF02\uFF03\uFF04\uFF05\uFF06\uFF07\uFF08\uFF09\uFF0A" + //  9250 -  9259
		"\uFF0B\uFF0C\uFF0D\uFF0E\uFF0F\uFF10\uFF11\uFF12\uFF13\uFF14" + //  9260 -  9269
		"\uFF15\uFF16\uFF17\uFF18\uFF19\uFF1A\uFF1B\uFF1C\uFF1D\uFF1E" + //  9270 -  9279
		"\uFF1F\u874C\u970D\u76E5\u9E1B\u9278\u4F5D\u50B4\u5ABE\u5AD7" + //  9280 -  9289
		"\uF846\u750C\u89AF\u98B6\u63AC\u8DEA\u5DF9\u6F0C\u5C8C\u7B08" + //  9290 -  9299
		"\uF847\u9C2D\uF848\u7CEF\u5583\u66E9\u8FFA\u4F5E\uF849\u5B65" + //  9300 -  9309
		"\uF84A\u977C\u601B\uE4A7\uE4A8\uE4A9\uE4AA\uE4AB\uE4AC\uE4AD" + //  9310 -  9319
		"\uE4AE\uE4AF\uE4B0\uE4B1\uE4B2\uE4B3\uE4B4\uE4B5\uE4B6\uE4B7" + //  9320 -  9329
		"\uE4B8\uE4B9\uE4BA\uE4BB\uE4BC\uE4BD\uE4BE\uE4BF\uE4C0\uE4C1" + //  9330 -  9339
		"\uE4C2\uE4C3\uE4C4\uE4C5\uFFFD\u21D2\u21D4\u2200\u2203\u00B4" + //  9340 -  9349
		"\u02DC\u02C7\u02D8\u02DD\u02DA\u02D9\u00B8\u02DB\u00A1\u00BF" + //  9350 -  9359
		"\u02D0\u222E\u2211\u220F\u00A4\u2109\u2030\u25C1\u25C0\u25B7" + //  9360 -  9369
		"\u25B6\u2664\u2660\u2661\u2665\u2667\uE487\uE488\uE489\uE48A" + //  9370 -  9379
		"\uE48B\uE48C\uE48D\uE48E\uE48F\uE490\uE491\uE492\uE493\uE494" + //  9380 -  9389
		"\uE495\uE496\uE497\uE498\uE499\uE49A\uE49B\uE49C\uE49D\uE49E" + //  9390 -  9399
		"\uE49F\uE4A0\uE4A1\uE4A2\uE4A3\uE4A4\uE4A5\uE4A6\uE449\uE44A" + //  9400 -  9409
		"\uE44B\uE44C\uE44D\uE44E\uE44F\uE450\uE451\uE452\uE453\uE454" + //  9410 -  9419
		"\uE455\uE456\uE457\uE458\uE459\uE45A\uE45B\uE45C\uE45D\uE45E" + //  9420 -  9429
		"\uE45F\uE460\uE461\uE462\uE463\uE464\uE465\uE466\uE467\uFFFD" + //  9430 -  9439
		"\u3000\u3001\u3002\u30FB\u2025\u2026\u00A8\u3003\u2010\u2014" + //  9440 -  9449
		"\u2225\uFF3C\u301C\u2018\u2019\u201C\u201D\u3014\u3015\u3008" + //  9450 -  9459
		"\u3009\u300A\u300B\u300C\u300D\u300E\u300F\u3010\u3011\u00B1" + //  9460 -  9469
		"\u00D7\uE429\uE42A\uE42B\uE42C\uE42D\uE42E\uE42F\uE430\uE431" + //  9470 -  9479
		"\uE432\uE433\uE434\uE435\uE436\uE437\uE438\uE439\uE43A\uE43B" + //  9480 -  9489
		"\uE43C\uE43D\uE43E\uE43F\uE440\uE441\uE442\uE443\uE444\uE445" + //  9490 -  9499
		"\uE446\uE447\uE448\uE3EB\uE3EC\uE3ED\uE3EE\uE3EF\uE3F0\uE3F1" + //  9500 -  9509
		"\uE3F2\uE3F3\uE3F4\uE3F5\uE3F6\uE3F7\uE3F8\uE3F9\uE3FA\uE3FB" + //  9510 -  9519
		"\uE3FC\uE3FD\uE3FE\uE3FF\uE400\uE401\uE402\uE403\uE404\uE405" + //  9520 -  9529
		"\uE406\uE407\uE408\uE409\uFFFD\uC63A\uC6B7\uC6DF\uC70B\uC736" + //  9530 -  9539
		"\uC77B\uC7A7\uC7AA\uC807\uC814\uC81B\uC839\uC84B\uC890\uC89C" + //  9540 -  9549
		"\uC8A0\uC8AC\uC8B0\uC8B8\uC8E8\uC8F0\uC8F1\uC92B\uC96D\uC9A4" + //  9550 -  9559
		"\uC9D4\uCA30\uCA57\uCA70\uCA97\uCAA0\uE3CB\uE3CC\uE3CD\uE3CE" + //  9560 -  9569
		"\uE3CF\uE3D0\uE3D1\uE3D2\uE3D3\uE3D4\uE3D5\uE3D6\uE3D7\uE3D8" + //  9570 -  9579
		"\uE3D9\uE3DA\uE3DB\uE3DC\uE3DD\uE3DE\uE3DF\uE3E0\uE3E1\uE3E2" + //  9580 -  9589
		"\uE3E3\uE3E4\uE3E5\uE3E6\uE3E7\uE3E8\uE3E9\uE3EA\uE38D\uE38E" + //  9590 -  9599
		"\uE38F\uE390\uE391\uE392\uE393\uE394\uE395\uE396\uE397\uE398" + //  9600 -  9609
		"\uE399\uE39A\uE39B\uE39C\uE39D\uE39E\uE39F\uE3A0\uE3A1\uE3A2" + //  9610 -  9619
		"\uE3A3\uE3A4\uE3A5\uE3A6\uE3A7\uE3A8\uE3A9\uE3AA\uE3AB\uFFFD" + //  9620 -  9629
		"\uB990\uB9DF\uB9FB\uBA1C\uBA6B\uBA6D\uBA80\uBAAF\uBAC3\uBAE0" + //  9630 -  9639
		"\uBBC1\uBBD5\uBBDC\uBBE0\uBC0E\uBC28\uBC37\uBC5C\uBC68\uBC98" + //  9640 -  9649
		"\uBC9C\uBCB9\uBCCC\uBCD2\uBCD3\uBCD4\uBD23\uBD97\uBDB4\uBE18" + //  9650 -  9659
		"\uBE21\uE36D\uE36E\uE36F\uE370\uE371\uE372\uE373\uE374\uE375" + //  9660 -  9669
		"\uE376\uE377\uE378\uE379\uE37A\uE37B\uE37C\uE37D\uE37E\uE37F" + //  9670 -  9679
		"\uE380\uE381\uE382\uE383\uE384\uE385\uE386\uE387\uE388\uE389" + //  9680 -  9689
		"\uE38A\uE38B\uE38C\uE32F\uE330\uE331\uE332\uE333\uE334\uE335" + //  9690 -  9699
		"\uE336\uE337\uE338\uE339\uE33A\uE33B\uE33C\uE33D\uE33E\uE33F" + //  9700 -  9709
		"\uE340\uE341\uE342\uE343\uE344\uE345\uE346\uE347\uE348\uE349" + //  9710 -  9719
		"\uE34A\uE34B\uE34C\uE34D\uFFFD\u944A\u571C\u61FD\u9B1F\u5A93" + //  9720 -  9729
		"\u6033\u56C2\u7334\u7BCC\u5FFB\u8FC4\u9821\uAC02\uAC0B\uAC79" + //  9730 -  9739
		"\uAC87\uAC93\uACE9\uACFA\uAD19\uAD28\uAD2B\uAD9B\uADD5\uADEC" + //  9740 -  9749
		"\uAE02\uAE0F\uAE11\uAE27\uAE3C\uAE44\uE30F\uE310\uE311\uE312" + //  9750 -  9759
		"\uE313\uE314\uE315\uE316\uE317\uE318\uE319\uE31A\uE31B\uE31C" + //  9760 -  9769
		"\uE31D\uE31E\uE31F\uE320\uE321\uE322\uE323\uE324\uE325\uE326" + //  9770 -  9779
		"\uE327\uE328\uE329\uE32A\uE32B\uE32C\uE32D\uE32E\uE2D1\uE2D2" + //  9780 -  9789
		"\uE2D3\uE2D4\uE2D5\uE2D6\uE2D7\uE2D8\uE2D9\uE2DA\uE2DB\uE2DC" + //  9790 -  9799
		"\uE2DD\uE2DE\uE2DF\uE2E0\uE2E1\uE2E2\uE2E3\uE2E4\uE2E5\uE2E6" + //  9800 -  9809
		"\uE2E7\uE2E8\uE2E9\uE2EA\uE2EB\uE2EC\uE2ED\uE2EE\uE2EF\uFFFD" + //  9810 -  9819
		"\u88C5\u7E94\u67E2\u86C6\u8C6C\u7CF4\u56C0\u5DD3\u78DA\u7FE6" + //  9820 -  9829
		"\u7A83\u6904\u6883\u6662\u7445\u8E36\uF863\u566A\u7681\u7AC8" + //  9830 -  9839
		"\u7B0A\u7CF6\u7D5B\u9BDB\u6A05\u8E64\u851F\u8098\u96BC\uF864" + //  9840 -  9849
		"\u8A3C\uE2B1\uE2B2\uE2B3\uE2B4\uE2B5\uE2B6\uE2B7\uE2B8\uE2B9" + //  9850 -  9859
		"\uE2BA\uE2BB\uE2BC\uE2BD\uE2BE\uE2BF\uE2C0\uE2C1\uE2C2\uE2C3" + //  9860 -  9869
		"\uE2C4\uE2C5\uE2C6\uE2C7\uE2C8\uE2C9\uE2CA\uE2CB\uE2CC\uE2CD" + //  9870 -  9879
		"\uE2CE\uE2CF\uE2D0\uE273\uE274\uE275\uE276\uE277\uE278\uE279" + //  9880 -  9889
		"\uE27A\uE27B\uE27C\uE27D\uE27E\uE27F\uE280\uE281\uE282\uE283" + //  9890 -  9899
		"\uE284\uE285\uE286\uE287\uE288\uE289\uE28A\uE28B\uE28C\uE28D" + //  9900 -  9909
		"\uE28E\uE28F\uE290\uE291\uFFFD\u5A7F\uF856\uF857\uF858\u5C20" + //  9910 -  9919
		"\u6103\uF859\u71F9\uF85A\u5070\uF85B\u6308\u8258\u9704\u87C0" + //  9920 -  9929
		"\u7463\u53DF\uF85C\u666C\u6EB2\u795F\uF85D\u9D89\u8671\u557B" + //  9930 -  9939
		"\uF85E\u7DE6\u77E7\uF85F\u843C\u8D0B\uE253\uE254\uE255\uE256" + //  9940 -  9949
		"\uE257\uE258\uE259\uE25A\uE25B\uE25C\uE25D\uE25E\uE25F\uE260" + //  9950 -  9959
		"\uE261\uE262\uE263\uE264\uE265\uE266\uE267\uE268\uE269\uE26A" + //  9960 -  9969
		"\uE26B\uE26C\uE26D\uE26E\uE26F\uE270\uE271\uE272\uE215\uE216" + //  9970 -  9979
		"\uE217\uE218\uE219\uE21A\uE21B\uE21C\uE21D\uE21E\uE21F\uE220" + //  9980 -  9989
		"\uE221\uE222\uE223\uE224\uE225\uE226\uE227\uE228\uE229\uE22A" + //  9990 -  9999
		"\uE22B\uE22C\uE22D\uE22E\uE22F\uE230\uE231\uE232\uE233\uFFFD" + // 10000 - 10009
		"\u8F64\u6F09\uF84B\u8F46\u7C5F\u857E\u8A84\uF84C\u50C2\u9ACF" + // 10010 - 10019
		"\u7ABF\u51DB\u5EE9\uF84D\u6F13\u79BB\u87AD\u9B51\u75F3\u5CA6" + // 10020 - 10029
		"\u5ABD\u87C7\u8B3E\u93DD\u9B18\u9B4D\u771B\u82FA\u8109\u4FDB" + // 10030 - 10039
		"\u8004\uE1F5\uE1F6\uE1F7\uE1F8\uE1F9\uE1FA\uE1FB\uE1FC\uE1FD" + // 10040 - 10049
		"\uE1FE\uE1FF\uE200\uE201\uE202\uE203\uE204\uE205\uE206\uE207" + // 10050 - 10059
		"\uE208\uE209\uE20A\uE20B\uE20C\uE20D\uE20E\uE20F\uE210\uE211" + // 10060 - 10069
		"\uE212\uE213\uE214\uE1B7\uE1B8\uE1B9\uE1BA\uE1BB\uE1BC\uE1BD" + // 10070 - 10079
		"\uE1BE\uE1BF\uE1C0\uE1C1\uE1C2\uE1C3\uE1C4\uE1C5\uE1C6\uE1C7" + // 10080 - 10089
		"\uE1C8\uE1C9\uE1CA\uE1CB\uE1CC\uE1CD\uE1CE\uE1CF\uE1D0\uE1D1" + // 10090 - 10099
		"\uE1D2\uE1D3\uE1D4\uE1D5\uFFFD\uE4C6\uE4C7\uE4C8\uE4C9\uE4CA" + // 10100 - 10109
		"\u01C2\u2266\u2267\u212A\uFFE4\u02BA\uF843\u64F1\u7FAF\u9163" + // 10110 - 10119
		"\uF844\u9ABC\u84B9\u54FD\u6243\u6AA0\u71B2\u754A\uF845\u96DE" + // 10120 - 10129
		"\u6772\u77BD\u8A41\u6831\u69D3\u7B9C\uE197\uE198\uE199\uE19A" + // 10130 - 10139
		"\uE19B\uE19C\uE19D\uE19E\uE19F\uE1A0\uE1A1\uE1A2\uE1A3\uE1A4" + // 10140 - 10149
		"\uE1A5\uE1A6\uE1A7\uE1A8\uE1A9\uE1AA\uE1AB\uE1AC\uE1AD\uE1AE" + // 10150 - 10159
		"\uE1AF\uE1B0\uE1B1\uE1B2\uE1B3\uE1B4\uE1B5\uE1B6\uE159\uE15A" + // 10160 - 10169
		"\uE15B\uE15C\uE15D\uE15E\uE15F\uE160\uE161\uE162\uE163\uE164" + // 10170 - 10179
		"\uE165\uE166\uE167\uE168\uE169\uE16A\uE16B\uE16C\uE16D\uE16E" + // 10180 - 10189
		"\uE16F\uE170\uE171\uE172\uE173\uE174\uE175\uE176\uE177\uFFFD" + // 10190 - 10199
		"\uE468\uE469\uE46A\uE46B\uE46C\uE46D\uE46E\uE46F\uE470\uE471" + // 10200 - 10209
		"\uE472\uE473\uE474\uE475\uE476\uE477\uE478\uE479\uE47A\uE47B" + // 10210 - 10219
		"\uE47C\uE47D\uE47E\uE47F\uE480\uE481\uE482\uE483\uE484\uE485" + // 10220 - 10229
		"\uE486\uE139\uE13A\uE13B\uE13C\uE13D\uE13E\uE13F\uE140\uE141" + // 10230 - 10239
		"\uE142\uE143\uE144\uE145\uE146\uE147\uE148\uE149\uE14A\uE14B" + // 10240 - 10249
		"\uE14C\uE14D\uE14E\uE14F\uE150\uE151\uE152\uE153\uE154\uE155" + // 10250 - 10259
		"\uE156\uE157\uE158\uE0FB\uE0FC\uE0FD\uE0FE\uE0FF\uE100\uE101" + // 10260 - 10269
		"\uE102\uE103\uE104\uE105\uE106\uE107\uE108\uE109\uE10A\uE10B" + // 10270 - 10279
		"\uE10C\uE10D\uE10E\uE10F\uE110\uE111\uE112\uE113\uE114\uE115" + // 10280 - 10289
		"\uE116\uE117\uE118\uE119\uFFFD\uE40A\uE40B\uE40C\uE40D\uE40E" + // 10290 - 10299
		"\uE40F\uE410\uE411\uE412\uE413\uE414\uE415\uE416\uE417\uE418" + // 10300 - 10309
		"\uE419\uE41A\uE41B\uE41C\uE41D\uE41E\uE41F\uE420\uE421\uE422" + // 10310 - 10319
		"\uE423\uE424\uE425\uE426\uE427\uE428\uE0DB\uE0DC\uE0DD\uE0DE" + // 10320 - 10329
		"\uE0DF\uE0E0\uE0E1\uE0E2\uE0E3\uE0E4\uE0E5\uE0E6\uE0E7\uE0E8" + // 10330 - 10339
		"\uE0E9\uE0EA\uE0EB\uE0EC\uE0ED\uE0EE\uE0EF\uE0F0\uE0F1\uE0F2" + // 10340 - 10349
		"\uE0F3\uE0F4\uE0F5\uE0F6\uE0F7\uE0F8\uE0F9\uE0FA"
		;


	public ByteToCharCp949() {
		super();
		super.mask1 = 0xFFE0;
		super.mask2 = 0x001F;
		super.shift = 5;
		super.leadByte = this.leadByte;
		super.singleByteToChar = this.singleByteToChar;
		super.index1 = this.index1;
		super.index2 = this.index2;
	}
}
