/*
 * @(#)CharToByteCp970.java	1.6 98/03/18
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */
package sun.io;

/**
* @author Malcolm Ayres
*/

public class CharToByteCp970 extends CharToByteConverter
{
    private static final char SBase = '\uAC00';
    private static final char LBase = '\u1100';
    private static final char VBase = '\u1161';
    private static final char TBase = '\u11A7';
    private static final int  VCount = 21;
    private static final int  TCount = 28;
    private static final byte G0 = 0;
    private static final byte G1 = 1;
    private static final byte G2 = 2;
    private static final byte G3 = 3;
    private byte   charState = G0;
    private char   l, v, t;

    private byte[] outputByte;

    private char highHalfZoneCode;
    private int  mask1;
    private int  mask2;
    private int  shift;

    public CharToByteCp970() {
       super();
       highHalfZoneCode = 0;
       outputByte = new byte[2];
       mask1 = 0xFFF8;
       mask2 = 0x0007;
       shift = 3;
    }

    /**
      * flush out any residual data and reset the buffer state
      */
    public int flush(byte[] output, int outStart, int outEnd)
        throws MalformedInputException,
               ConversionBufferFullException
    {
       int bytesOut;

       byteOff = outStart;

       if (highHalfZoneCode != 0) {
           reset();
           badInputLength = 0;
           throw new MalformedInputException();
       }

       if (charState != G0) {
           try {
              unicodeToBuffer(composeHangul() ,output, outEnd);
           }
           catch(UnknownCharacterException e) {
              reset();
              badInputLength = 0;
              throw new MalformedInputException();
           }
           charState = G0;
       }

       bytesOut = byteOff - outStart;

       reset();
       return bytesOut;
    }

    /**
     * Resets converter to its initial state.
     */
    public void reset() {
       highHalfZoneCode = 0;
       charState = G0;
       charOff = byteOff = 0;
    }

    /**
     * Returns true if the given character can be converted to the
     * target character encoding.
     */
    public boolean canConvert(char ch) {
       int  index;
       int  theBytes;

       index = index1[((ch & mask1) >> shift)] + (ch & mask2);
       if (index < 15000)
         theBytes = (int)(index2.charAt(index));
       else
         theBytes = (int)(index2a.charAt(index-15000));

       if (theBytes != 0)
          return (true);

       // only return true if input char was unicode null - all others are
       //    undefined
       return( ch == '\u0000');
    }

    /**
     * Character conversion
     */

    public int convert(char[] input, int inOff, int inEnd,
                       byte[] output, int outOff, int outEnd)
        throws UnknownCharacterException, MalformedInputException,
               ConversionBufferFullException
    {
       char    inputChar;
       int     inputSize;

       charOff = inOff;
       byteOff = outOff;

       while (charOff < inEnd) {

          if (highHalfZoneCode == 0) {
             inputChar = input[charOff];
             inputSize = 1;
          } else {
             inputChar = highHalfZoneCode;
             inputSize = 0;
             highHalfZoneCode = 0;
          }

          switch (charState) {
          case G0:

             l = LBase;
             v = VBase;
             t = TBase;

             if ( isLeadingC(inputChar) ) {     // Leading Consonant
                l = inputChar;
                charState = G1;
                break;
             }

             if ( isVowel(inputChar) ) {        // Vowel
                v = inputChar;
                charState = G2;
                break;
             }

             if ( isTrailingC(inputChar) ) {    // Trailing Consonant
                t = inputChar;
                charState = G3;
                break;
             }

             break;

          case G1:
             if ( isLeadingC(inputChar) ) {     // Leading Consonant
                l = composeLL(l, inputChar);
                break;
             }

             if ( isVowel(inputChar) ) {        // Vowel
                v = inputChar;
                charState = G2;
                break;
             }

             if ( isTrailingC(inputChar) ) {    // Trailing Consonant
                t = inputChar;
                charState = G3;
                break;
             }

             unicodeToBuffer(composeHangul(), output, outEnd);

             charState = G0;
             break;

          case G2:
             if ( isLeadingC(inputChar) ) {     // Leading Consonant

                unicodeToBuffer(composeHangul(), output, outEnd);

                l = inputChar;
                v = VBase;
                t = TBase;
                charState = G1;
                break;
             }

             if ( isVowel(inputChar) ) {        // Vowel
                v = composeVV(l, inputChar);
                charState = G2;
                break;
             }

             if ( isTrailingC(inputChar) ) {    // Trailing Consonant
                t = inputChar;
                charState = G3;
                break;
             }

             unicodeToBuffer(composeHangul(), output, outEnd);

             charState = G0;

             break;

          case G3:
             if ( isTrailingC(inputChar) ) {    // Trailing Consonant
                t = composeTT(t, inputChar);
                charState = G3;
                break;
             }

             unicodeToBuffer(composeHangul(), output, outEnd);

             charState = G0;

             break;
          }

          if (charState != G0)
             charOff++;
          else {

             // Is this a high surrogate?
             if(inputChar >= '\ud800' && inputChar <= '\udbff') {
                // Is this the last character of the input?
                if (charOff + inputSize >= inEnd) {
                   highHalfZoneCode = inputChar;
                   charOff += inputSize;
                   break;
                }

                // Is there a low surrogate following?
                inputChar = input[charOff + inputSize];
                if (inputChar >= '\udc00' && inputChar <= '\udfff') {
                   // We have a valid surrogate pair.  Too bad we don't do
                   // surrogates.  Is substitution enabled?
                   if (subMode) {
                      if (subBytes.length == 1) {
                         outputByte[0] = 0x00;
                         outputByte[1] = subBytes[0];
                      } else {
                         outputByte[0] = subBytes[0];
                         outputByte[1] = subBytes[1];
                      }

                      bytesToBuffer(outputByte, output, outEnd);
                      inputSize++;
                   } else {
                      badInputLength = 2;
                      throw new UnknownCharacterException();
                   }
                } else {
                   // We have a malformed surrogate pair
                   badInputLength = 1;
                   throw new MalformedInputException();
                }
             }

               // Is this an unaccompanied low surrogate?
             else
                if (inputChar >= '\uDC00' && inputChar <= '\uDFFF') {
                   badInputLength = 1;
                   throw new MalformedInputException();
                } else {
                   unicodeToBuffer(inputChar, output, outEnd);
                }

             charOff += inputSize;

          }

       }

       return byteOff - outOff;

    }

    private char composeHangul() {
       int lIndex, vIndex, tIndex;

       lIndex = l - LBase;
       vIndex = v - VBase;
       tIndex = t - TBase;

       return (char)((lIndex * VCount + vIndex) * TCount + tIndex + SBase);
    }

    private char composeLL(char l1, char l2) {
       return l2;
    }

    private char composeVV(char v1, char v2) {
       return v2;
    }

    private char composeTT(char t1, char t2) {
       return t2;
    }

    private boolean isLeadingC(char c) {
       return (c >= LBase && c <= '\u1159');
    }

    private boolean isVowel(char c) {
       return (c >= VBase && c <= '\u11a2');
    }

    private boolean isTrailingC(char c) {
       return (c >= TBase && c <= '\u11f9');
    }

    /**
     * returns the maximum number of bytes needed to convert a char
     */
    public int getMaxBytesPerChar() {
       return 2;
    }


    /**
     * Return the character set ID
     */
    public String getCharacterEncoding() {
       return "Cp970";
    }

    /**
     * private function to add the bytes to the output buffer
     */
    private void bytesToBuffer(byte[] theBytes, byte[] output, int outEnd)
        throws ConversionBufferFullException,
               UnknownCharacterException {

       int spaceNeeded;

       // ensure sufficient space for the bytes(s)

       if (theBytes[0] == 0x00)
          spaceNeeded = 1;
       else
          spaceNeeded = 2;

       if (byteOff + spaceNeeded > outEnd)
          throw new ConversionBufferFullException();

       // move the data into the buffer

       if (spaceNeeded == 1)
          output[byteOff++] = theBytes[1];
       else {
          output[byteOff++] = theBytes[0];
          output[byteOff++] = theBytes[1];
       }

    }

    /**
     * private function to add a unicode character to the output buffer
     */
    private void unicodeToBuffer(char unicode, byte[] output, int outEnd)
        throws ConversionBufferFullException,
               UnknownCharacterException {

       int index;
       int theBytes;

       // first we convert the unicode to its byte representation

       index = index1[((unicode & mask1) >> shift)] + (unicode & mask2);
       if (index < 15000)
         theBytes = (int)(index2.charAt(index));
       else
         theBytes = (int)(index2a.charAt(index-15000));
       outputByte[0] = (byte)((theBytes & 0x0000ff00)>>8);
       outputByte[1] = (byte)(theBytes & 0x000000ff);

       // if the unicode was not mappable - look for the substitution bytes

       if (outputByte[0] == 0x00 && outputByte[1] == 0x00
                          && unicode != '\u0000') {
          if (subMode) {
             if (subBytes.length == 1) {
                outputByte[0] = 0x00;
                outputByte[1] = subBytes[0];
             } else {
                outputByte[0] = subBytes[0];
                outputByte[1] = subBytes[1];
             }
          } else {
             badInputLength = 1;
             throw new UnknownCharacterException();
          }
       }

       // now put the bytes in the buffer

       bytesToBuffer(outputByte, output, outEnd);

    }

        private static final short index1[] =
        {
                 2751,   771,   710,   642,    41, 20287, 19487, 19075, // 0000 - 003F
                15804, 13732, 12911, 12503, 12212, 11716, 11086,  8904, // 0040 - 007F
                 6712,  4859,  3337,  1411,  2695, 20484, 20279, 20234, // 0080 - 00BF
                14112, 19002, 20155, 19929, 13922, 19002, 19844, 19479, // 00C0 - 00FF
                19002, 19002,  2553, 19002, 13857, 19002,  2491, 19270, // 0100 - 013F
                19058,  2376,  9341, 19002, 13804, 19002, 19002, 19002, // 0140 - 017F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0180 - 01BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 01C0 - 01FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0200 - 023F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0240 - 027F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0280 - 02BF
                19003, 19002, 19001, 18783, 19002, 19002, 19002, 19002, // 02C0 - 02FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0300 - 033F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0340 - 037F
                19002, 19002,  2328, 18663, 18523, 18472,  2313, 18383, // 0380 - 03BF
                18272, 18138, 19002, 19002, 19002, 19002, 19002, 19002, // 03C0 - 03FF
                 2201, 19002, 18020, 17958, 17671, 17506, 17138, 16569, // 0400 - 043F
                16474, 16386,  2137, 19002, 19002, 19002, 19002, 19002, // 0440 - 047F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0480 - 04BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 04C0 - 04FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0500 - 053F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0540 - 057F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0580 - 05BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 05C0 - 05FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0600 - 063F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0640 - 067F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0680 - 06BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 06C0 - 06FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0700 - 073F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0740 - 077F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0780 - 07BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 07C0 - 07FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0800 - 083F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0840 - 087F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0880 - 08BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 08C0 - 08FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0900 - 093F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0940 - 097F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0980 - 09BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 09C0 - 09FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0A00 - 0A3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0A40 - 0A7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0A80 - 0ABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0AC0 - 0AFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0B00 - 0B3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0B40 - 0B7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0B80 - 0BBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0BC0 - 0BFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0C00 - 0C3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0C40 - 0C7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0C80 - 0CBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0CC0 - 0CFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0D00 - 0D3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0D40 - 0D7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0D80 - 0DBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0DC0 - 0DFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0E00 - 0E3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0E40 - 0E7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0E80 - 0EBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0EC0 - 0EFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0F00 - 0F3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0F40 - 0F7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0F80 - 0FBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 0FC0 - 0FFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1000 - 103F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1040 - 107F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1080 - 10BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 10C0 - 10FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1100 - 113F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1140 - 117F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1180 - 11BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 11C0 - 11FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1200 - 123F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1240 - 127F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1280 - 12BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 12C0 - 12FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1300 - 133F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1340 - 137F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1380 - 13BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 13C0 - 13FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1400 - 143F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1440 - 147F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1480 - 14BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 14C0 - 14FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1500 - 153F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1540 - 157F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1580 - 15BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 15C0 - 15FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1600 - 163F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1640 - 167F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1680 - 16BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 16C0 - 16FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1700 - 173F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1740 - 177F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1780 - 17BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 17C0 - 17FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1800 - 183F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1840 - 187F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1880 - 18BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 18C0 - 18FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1900 - 193F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1940 - 197F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1980 - 19BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 19C0 - 19FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1A00 - 1A3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1A40 - 1A7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1A80 - 1ABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1AC0 - 1AFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1B00 - 1B3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1B40 - 1B7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1B80 - 1BBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1BC0 - 1BFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1C00 - 1C3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1C40 - 1C7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1C80 - 1CBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1CC0 - 1CFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1D00 - 1D3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1D40 - 1D7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1D80 - 1DBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1DC0 - 1DFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1E00 - 1E3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1E40 - 1E7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1E80 - 1EBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1EC0 - 1EFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1F00 - 1F3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1F40 - 1F7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1F80 - 1FBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 1FC0 - 1FFF
                19002, 19002, 16346, 16100, 15796, 19002, 15230, 11945, // 2000 - 203F
                19002, 19002, 19002, 19002, 19002, 19002,  8657, 18834, // 2040 - 207F
                 2052, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2080 - 20BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 20C0 - 20FF
                11959,  1989,  9248, 19002,  1958,  9001, 19002, 19002, // 2100 - 213F
                19002, 19002, 18361,  8885, 14927, 14669, 14506, 14423, // 2140 - 217F
                19002, 19002, 14007, 13714, 19002, 19002, 19002, 19002, // 2180 - 21BF
                19002, 19002,  9179, 19002, 19002, 19002, 19002, 19002, // 21C0 - 21FF
                13605, 13116,  1861,  4692, 13037, 12903,  3903,  3439, // 2200 - 223F
                19002, 19002, 18335, 19002, 12566,  9927, 19002, 19002, // 2240 - 227F
                19638, 19002, 19002, 19002,  3396, 19002, 19002, 19002, // 2280 - 22BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 22C0 - 22FF
                19002, 19002,  4529, 19002, 19002, 19002, 19002, 19002, // 2300 - 233F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2340 - 237F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2380 - 23BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 23C0 - 23FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2400 - 243F
                19002, 19002, 19002, 19002, 12529, 12495,  8234, 12405, // 2440 - 247F
                12190, 19002, 19002,  8215, 12024, 11881, 11697, 19002, // 2480 - 24BF
                19002, 19002, 11531, 11480, 11382, 11214, 19002, 19002, // 24C0 - 24FF
                11067,  8156, 10970, 10466, 10369, 10258, 10194, 10028, // 2500 - 253F
                 9903,  9776, 19002, 19002, 19002, 19002, 19002, 19002, // 2540 - 257F
                19002, 19002, 19599, 19002,  9752,  9554, 18173, 16194, // 2580 - 25BF
                 9256,  9072,  8893, 19002, 19002, 19002, 19002, 19002, // 25C0 - 25FF
                 3377, 13771, 19002, 15873, 19002, 19002, 19002, 19002, // 2600 - 263F
                 8728, 19002, 19002, 19002,  8684,  8571, 19002, 19002, // 2640 - 267F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2680 - 26BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 26C0 - 26FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2700 - 273F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2740 - 277F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2780 - 27BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 27C0 - 27FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2800 - 283F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2840 - 287F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2880 - 28BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 28C0 - 28FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2900 - 293F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2940 - 297F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2980 - 29BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 29C0 - 29FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2A00 - 2A3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2A40 - 2A7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2A80 - 2ABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2AC0 - 2AFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2B00 - 2B3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2B40 - 2B7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2B80 - 2BBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2BC0 - 2BFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2C00 - 2C3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2C40 - 2C7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2C80 - 2CBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2CC0 - 2CFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2D00 - 2D3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2D40 - 2D7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2D80 - 2DBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2DC0 - 2DFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2E00 - 2E3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2E40 - 2E7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2E80 - 2EBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2EC0 - 2EFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2F00 - 2F3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2F40 - 2F7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2F80 - 2FBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 2FC0 - 2FFF
                 8426,  8260,  8040, 15735, 19002, 19002, 19002, 19002, // 3000 - 303F
                 1649,  7754,  7652,  7598,  7450,  7334,  7152,  7086, // 3040 - 307F
                 7017,  6818,  6677, 19002,  1807,  6624,  6502,  6324, // 3080 - 30BF
                 6233,  5853,  5749,  5332,  5244,  5010,  4851,  7296, // 30C0 - 30FF
                19002, 19002, 19002, 19002, 19002, 19002,  1792,  4708, // 3100 - 313F
                 4627,  4460,  4430,  4250,  4224,  4172,  3785,  3664, // 3140 - 317F
                 3578,  3329, 19002, 19002, 19002, 19002, 19002, 19002, // 3180 - 31BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 31C0 - 31FF
                 3224,  3029,  2572,  2460, 19002, 19002, 19002, 19002, // 3200 - 323F
                19002, 19002, 19002, 19002,  2419,  2257,  2175,  1974, // 3240 - 327F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3280 - 32BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 32C0 - 32FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3300 - 333F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3340 - 337F
                 1382,  1276,  1140,   948,   914,   831,   442, 20683, // 3380 - 33BF
                20614, 20562, 20476, 20435, 19002, 19002, 19002, 19002, // 33C0 - 33FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3400 - 343F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3440 - 347F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3480 - 34BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 34C0 - 34FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3500 - 353F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3540 - 357F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3580 - 35BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 35C0 - 35FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3600 - 363F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3640 - 367F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3680 - 36BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 36C0 - 36FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3700 - 373F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3740 - 377F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3780 - 37BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 37C0 - 37FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3800 - 383F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3840 - 387F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3880 - 38BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 38C0 - 38FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3900 - 393F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3940 - 397F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3980 - 39BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 39C0 - 39FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3A00 - 3A3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3A40 - 3A7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3A80 - 3ABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3AC0 - 3AFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3B00 - 3B3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3B40 - 3B7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3B80 - 3BBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3BC0 - 3BFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3C00 - 3C3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3C40 - 3C7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3C80 - 3CBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3CC0 - 3CFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3D00 - 3D3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3D40 - 3D7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3D80 - 3DBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3DC0 - 3DFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3E00 - 3E3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3E40 - 3E7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3E80 - 3EBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3EC0 - 3EFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3F00 - 3F3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3F40 - 3F7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3F80 - 3FBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 3FC0 - 3FFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4000 - 403F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4040 - 407F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4080 - 40BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 40C0 - 40FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4100 - 413F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4140 - 417F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4180 - 41BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 41C0 - 41FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4200 - 423F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4240 - 427F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4280 - 42BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 42C0 - 42FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4300 - 433F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4340 - 437F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4380 - 43BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 43C0 - 43FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4400 - 443F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4440 - 447F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4480 - 44BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 44C0 - 44FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4500 - 453F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4540 - 457F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4580 - 45BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 45C0 - 45FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4600 - 463F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4640 - 467F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4680 - 46BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 46C0 - 46FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4700 - 473F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4740 - 477F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4780 - 47BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 47C0 - 47FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4800 - 483F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4840 - 487F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4880 - 48BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 48C0 - 48FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4900 - 493F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4940 - 497F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4980 - 49BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 49C0 - 49FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4A00 - 4A3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4A40 - 4A7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4A80 - 4ABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4AC0 - 4AFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4B00 - 4B3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4B40 - 4B7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4B80 - 4BBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4BC0 - 4BFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4C00 - 4C3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4C40 - 4C7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4C80 - 4CBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4CC0 - 4CFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4D00 - 4D3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4D40 - 4D7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4D80 - 4DBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 4DC0 - 4DFF
                20419, 20303,  1697, 20271, 19002,  3278,  9727, 20215, // 4E00 - 4E3F
                17471,  8801, 13421, 20182, 19002, 15114, 18218, 13347, // 4E40 - 4E7F
                19254, 20147, 20124, 19921,  1664, 19836, 13194,  4030, // 4E80 - 4EBF
                19773, 13078,  3620, 19757, 15394, 19002, 19699, 18111, // 4EC0 - 4EFF
                 1568,  1553, 19680, 19002, 19002, 16907,  7542, 19454, // 4F00 - 4F3F
                18064, 19415, 19342,  1509, 19002,  1538, 19262, 12793, // 4F40 - 4F7F
                 1449, 19205,  1359, 19162, 19002, 13184,  3168, 16210, // 4F80 - 4FBF
                17257,  1291,  1483, 17196, 19110, 13160,  1228, 19036, // 4FC0 - 4FFF
                17122,  1155,  1113,  1052,  1332, 18983, 19002, 14850, // 5000 - 503F
                18033, 18913,  3087,  8505,  3069, 19002, 15343, 18833, // 5040 - 507F
                18768,  2874,  1034, 18715, 19002, 15017,  8459, 13095, // 5080 - 50BF
                 2706,  1019,  1246,  8365,  2680,  2634, 19002,  1004, // 50C0 - 50FF
                18655,   963,  3850, 15772,   929,  9876, 18588,  4359, // 5100 - 513F
                18604, 18515, 18545, 14834,  9103, 18464,   846, 18438, // 5140 - 517F
                18375,  8115, 15998, 19002, 18234, 15701, 13047,  2612, // 5180 - 51BF
                 7070,   804, 19002, 14600,   752, 19002, 18130, 18005, // 51C0 - 51FF
                17981, 17950,   729,  2594,  6893,   868, 17919, 17889, // 5200 - 523F
                 2099,  7676, 14407,  6662,   691,   676,  3609,  2583, // 5240 - 527F
                 6361, 17861,  4866, 13411, 17803,  4877, 19002, 13021, // 5280 - 52BF
                18790, 18482, 18367, 17787, 18313, 19002, 14035, 17748, // 52C0 - 52FF
                 2517, 17726, 17687, 18489, 17659,  7622, 19002, 18499, // 5300 - 533F
                17579, 17522, 18148,  7582, 17498, 17350, 17298,  2476, // 5340 - 537F
                14209, 19002, 13908, 17227, 17130,  2287, 19002,  5830, // 5380 - 53BF
                13305, 17085,  6792,  5214, 16107, 16114, 17026, 16906, // 53C0 - 53FF
                 5702, 16778, 16717,  5642, 16561, 15241, 17240, 16466, // 5400 - 543F
                 7510, 16420, 14679, 19002, 19002, 16367, 14689, 17183, // 5440 - 547F
                16338, 13744, 16209, 19002,  6763, 16139,  5189, 16023, // 5480 - 54BF
                15953, 15881, 19002, 19002, 14696, 15788,  7306, 14757, // 54C0 - 54FF
                13271, 12960, 15771,  6390, 19002, 15593, 14703, 12883, // 5500 - 553F
                 6123, 15547, 19828, 12803, 20664, 19002, 19002, 12715, // 5540 - 557F
                15749, 14710, 19002, 15636, 15331, 13724, 19002, 19002, // 5580 - 55BF
                 2186, 19002, 12606,  3267, 20399, 19002, 19002,  2028, // 55C0 - 55FF
                12586, 12573,  6082, 19002, 19002, 14813,  3310, 19002, // 5600 - 563F
                19002, 19002, 16836, 19002, 19002, 15609,  5990, 19002, // 5640 - 567F
                12513, 19002, 19002, 19002,  1916, 11843,  5884,  5774, // 5680 - 56BF
                19002,  6842, 12201, 17853, 15592, 19002, 15546, 11704, // 56C0 - 56FF
                12429, 15488, 17779, 15449, 19002, 15351, 15330,  3999, // 5700 - 573F
                15287, 17740, 15261, 19002, 11078, 17718, 15215,  2835, // 5740 - 577F
                 6482, 15050, 10315, 19002, 14985, 19002, 19002, 19002, // 5780 - 57BF
                   25, 14919, 19002, 12812, 14812, 19002, 14781,  9564, // 57C0 - 57FF
                14661, 14608, 19002, 12472,  8757,  6458, 14574, 17633, // 5800 - 583F
                19002,  6209,  8578, 14498,  8047, 19002,  1886,  1403, // 5840 - 587F
                20647, 20442, 16487,  5547, 19002, 14415,  1765,  5662, // 5880 - 58BF
                20452, 12444, 20310, 14395, 11192, 16442, 19002, 20317, // 58C0 - 58FF
                19002, 12375, 11098, 20324, 12839, 20226, 20131, 12366, // 5900 - 593F
                10885, 14353, 14311,  5117, 14120,  5034, 11744, 13930, // 5940 - 597F
                11902, 16961,  1589, 19461,  1818, 13842, 19002, 19471, // 5980 - 59BF
                 1268, 19349, 13752, 19169, 12324, 13706, 19002, 13643, // 59C0 - 59FF
                19176, 19002, 19120, 13627, 13565, 18990, 12240,  2791, // 5A00 - 5A3F
                18775, 18725, 19002, 16938, 10954, 19002, 19002, 19002, // 5A40 - 5A7F
                19002, 19002, 10920, 10820,  4676, 19002, 19002, 19002, // 5A80 - 5ABF
                18735, 18241, 19002, 19002, 18012, 17900, 19002, 19002, // 5AC0 - 5AFF
                 1631, 17868, 12108, 19002, 19002,  4274, 19002, 19002, // 5B00 - 5B3F
                13451, 19002, 13429, 13355,   429, 17529, 13327, 13240, // 5B40 - 5B7F
                 1422, 13202, 19624, 13103, 10693, 10070, 13029, 13011, // 5B80 - 5BBF
                12968,  4513, 12950, 12080, 17092, 12927,  1193, 12891, // 5BC0 - 5BFF
                17037, 12873, 16378, 16030,  9597, 12811, 16040, 12594, // 5C00 - 5C3F
                12558, 12521, 16050, 20580, 12480,  9585, 16057, 16067, // 5C40 - 5C7F
                19002, 19002, 12452, 19002, 16077, 16087, 15969, 12397, // 5C80 - 5CBF
                12374, 19002, 19002, 15643, 12332, 12248, 12182, 19516, // 5CC0 - 5CFF
                12333,   977, 15620, 15495, 12117, 15505, 19002, 19002, // 5D00 - 5D3F
                19002,  7480, 12116, 19002, 19002, 15515, 19002, 19002, // 5D40 - 5D7F
                10720, 15414, 19002,   887, 12007, 16766, 19002, 11924, // 5D80 - 5DBF
                19002,   782,  2403,   653, 15358, 11873, 15271,  2556, // 5DC0 - 5DFF
                16709,  4333, 15222, 15061, 18475,  7249, 14303, 11851, // 5E00 - 5E3F
                11762,  9466, 18141, 10556, 15068, 19002, 16695, 11689, // 5E40 - 5E7F
                 4083,  4921,  2204, 16631, 11553, 13143,  2140, 11523, // 5E80 - 5EBF
                 3823, 11463, 11374, 16545, 11320,  4014, 12658,  9694, // 5EC0 - 5EFF
                15084,  9454, 15005, 11263, 11978, 14619, 14629,  8663, // 5F00 - 5F3F
                19002, 11206, 11658, 14636,  3491, 14581, 11144, 14588, // 5F40 - 5F7F
                11128,  3361, 11106, 11059, 11034, 10962,  1992, 13937, // 5F80 - 5FBF
                12293,  9146, 11630, 10755, 10719,  5924, 14672, 14426, // 5FC0 - 5FFF
                19002,  9510,  3201,  3941, 10701, 10647, 19002, 19002, // 6000 - 603F
                13944, 13717, 10618, 13964, 16162, 10574, 10555, 19002, // 6040 - 607F
                 1864, 13971,  8623,  8014, 10488, 19002, 10458, 10430, // 6080 - 60BF
                11217, 19002, 13978,  1939, 10387, 19002, 10361, 13988, // 60C0 - 60FF
                13999, 10337,  9557, 15898,  8632, 19002, 10281,  3873, // 6100 - 613F
                15861, 10245,  5417, 10223, 15838, 10186, 10101,  8896, // 6140 - 617F
                13572,  7738, 10078, 10016,  3759, 13597, 15763, 15741, // 6180 - 61BF
                 5352,  9989, 19002, 19002, 11515, 19002,  7689,  9971, // 61C0 - 61FF
                 9953,  9895,  9768, 15676, 13334, 15584,  9744, 11490, // 6200 - 623F
                 9629,  9536, 10537,  9509, 19002, 11455, 12487, 12462, // 6240 - 627F
                 9417, 11931, 11345,  9368, 19002,  4490, 10765, 10772, // 6280 - 62BF
                15561,  9296,  9278,  9238, 19002,  8186, 10779, 11417, // 62C0 - 62FF
                10625, 10437, 10397, 19002, 19002,  8391, 19002, 15538, // 6300 - 633F
                19002, 10407,  9199, 19002,  8597,  9064,  7486, 15472, // 6340 - 637F
                 3290,  8986, 15441,  8871,  8849,  8787, 19002, 19002, // 6380 - 63BF
                 8720,  8543,  8676, 15380,  9543,  1571,  3726, 19002, // 63C0 - 63FF
                19002, 19683,  7975,  3597, 15154,  3450, 19002, 15042, // 6400 - 643F
                19002, 19002, 19002,  8631,  8596,  9433, 19002,  8559, // 6440 - 647F
                19002,  8542,  9379, 14977,  7531,  6308,  8477,  5578, // 6480 - 64BF
                 9395, 14911, 14868,  8409,  9206,  3407, 14840, 14804, // 64C0 - 64FF
                 8373, 19002, 19002,  8326,  2665, 14736, 11180,  8300, // 6500 - 653F
                19113,  8282,  9012, 11026, 14566,  3243, 14543,  8252, // 6540 - 657F
                14529,  2088,  8174,  8333,  7890,  2341,  8091,  7883, // 6580 - 65BF
                 7876,  1620, 14482, 19002,  8032,  7869,  7815, 14440, // 65C0 - 65FF
                10895,  7827,  7808, 10789,  7790,  7772,  7746, 14376, // 6600 - 663F
                 7437,  7366, 19002,  7189,  6594,  7707,  1182,  7640, // 6640 - 667F
                 3098,  7590,  7550,  7430,  7350,   411, 14325,  7144, // 6680 - 66BF
                 6189,  7179, 10711,  7137,  7326, 19002,  7267,  7207, // 66C0 - 66FF
                 7168,  7130,  3058,   134, 10657,  7078,  6669, 14236, // 6700 - 673F
                10566,  6582,  7009,  2947, 19002,  1037,  6973, 20493, // 6740 - 677F
                 6154,  6553,  6921, 14215, 19002,  5961,  6874,  6810, // 6780 - 67BF
                 2885,  5522,  6647, 14190, 19002,  6416,  6616, 19067, // 67C0 - 67FF
                19002, 19002, 14134, 19002,  6041, 14102, 19002,  6575, // 6800 - 683F
                 6542,  6494,  6404, 19002, 19002, 18718, 10450,  5071, // 6840 - 687F
                 6034,  3980,  9350,   966,  6177,  6342,  6316, 19002, // 6880 - 68BF
                 2824,  6027, 14088,  6295,  6279,  6225,  2979,  5841, // 68C0 - 68FF
                 6170,   932, 14021, 19002,  3177, 19002,  6153,   755, // 6900 - 693F
                19002, 13894, 16351,  5741,  6131,  6104,   694,  6017, // 6940 - 697F
                19002, 19002, 18492,  2057, 16720, 10379,  5395,  2062, // 6980 - 69BF
                 5714,  9184,  5960, 19002, 19002,  5816, 19002,  9785, // 69C0 - 69FF
                13871, 13818,  5614, 20017,  5200, 19002, 14682,  5734, // 6A00 - 6A3F
                 2717,  5688, 13785,  5670,  5049,  8737, 19002, 19002, // 6A40 - 6A7F
                 5632, 20675,  2037,  5274,  6769, 19002,  8746,  8435, // 6A80 - 6ABF
                13684, 19002,  6686, 13661, 19002, 19002, 10329,  6695, // 6AC0 - 6AFF
                 5101,  2923, 13619, 19002,  5603, 19002, 13548, 13443, // 6B00 - 6B3F
                10125,  4971,  5521, 20338,  3656, 13369, 13285,  5451, // 6B40 - 6B7F
                 6704,  4753, 10093,  5324, 19002, 10058, 13259, 13216, // 6B80 - 6BBF
                20411,  2465, 13174,  5302, 19002,  1387, 19002, 19002, // 6BC0 - 6BFF
                19002,  5262, 20006, 19002,  1392, 19002,  2645,  5236, // 6C00 - 6C3F
                 5175,  9981,  5089, 13130,  5070,  5042,  5002, 13003, // 6C40 - 6C7F
                11831,  4979,  4843, 19873, 19002,  9736,  1514,  4796, // 6C80 - 6CBF
                19818, 20654,  1523, 17262,  3649,  4780,  4742, 19002, // 6CC0 - 6CFF
                19002, 19045, 19399,  3216, 15884,  3153,  6396,  4700, // 6D00 - 6D3F
                19323, 19002, 19002, 20389, 19050, 18905,  4893,  4643, // 6D40 - 6D7F
                12576,  4619, 19999, 19002, 19002, 20092, 12982, 19002, // 6D80 - 6DBF
                 4551,  8510,  2246,  4452, 11821,  4422,  1682,  4382, // 6DC0 - 6DFF
                19002, 19002, 20070, 11620,  4242,  4364, 12865,  4216, // 6E00 - 6E3F
                18554, 12826,  9963,  4198, 20042, 18559, 12751, 12721, // 6E40 - 6E7F
                19002, 19002,  4164,  2623, 12672, 12630, 19913, 12612, // 6E80 - 6EBF
                12543,  1374,  1208, 19002, 19002, 19782,  4655,  4146, // 6EC0 - 6EFF
                 1132, 19741, 19866, 19002,  4101, 18443, 19807,  4064, // 6F00 - 6F3F
                19721, 19002, 15192,  4042, 12419, 12204,  3979, 12389, // 6F40 - 6F7F
                   15,  3777,  3697, 19002, 18448,  9835, 20633, 20606, // 6F80 - 6FBF
                 3638, 12276, 15264, 12262,  3565, 20591, 19523,  9528, // 6FC0 - 6FFF
                19446, 20538, 19392,  3527, 15706,  3385, 19664,  9501, // 7000 - 703F
                19002,  1840,  3318,  3209, 17808, 18318,  3176,  3146, // 7040 - 707F
                20445, 12174, 12131,  3120, 14285,  7519, 16429,  3018, // 7080 - 70BF
                19002,  2996, 19002,  2863, 19002, 19002, 19313, 20382, // 70C0 - 70FF
                 2528, 20358, 19002, 17571, 17490, 19002,  2564, 19002, // 7100 - 713F
                19655, 20263,  2452, 20174,  2506, 20116, 19002, 19464, // 7140 - 717F
                 2438, 12703,  5996, 20061, 12046,  2391, 18895, 18825, // 7180 - 71BF
                19988,  2411,  2239, 19424, 19123, 18993, 19002, 16144, // 71C0 - 71FF
                 2167, 19002,  2036, 15958, 19002, 11992,  1966, 11865, // 7200 - 723F
                 1905,  1755, 11776,  1705, 19962, 19944, 11730, 11814, // 7240 - 727F
                 1672, 19002, 19002, 19002,  5890,  2298, 19002, 19002, // 7280 - 72BF
                 1579,  9360,  1457, 11610, 19903, 10849, 19002,  1367, // 72C0 - 72FF
                19002, 11672,  9288, 14766, 18728, 16810,  9216,  9113, // 7300 - 733F
                 3885, 19002,  1299, 19002, 19002,  1254,  1201,  1121, // 7340 - 737F
                 2276, 10178,  2223,  1079, 19279, 19859, 11644, 19797, // 7380 - 73BF
                  985,  9669,  8978,  9191,   940, 19714, 19084, 15945, // 73C0 - 73FF
                 7412, 15182, 14994,  9569,   906,   876,   823,   661, // 7400 - 743F
                  419, 17642, 17903, 15131,     8, 20699, 19002,  8841, // 7440 - 747F
                11567, 19610, 19002, 20691, 14963, 20622, 20599, 19561, // 7480 - 74BF
                 8712,  5780,  3709, 20570, 20554,  8588, 19012, 19002, // 74C0 - 74FF
                19506, 19002, 19439, 20546, 17651, 20526, 20518, 20501, // 7500 - 753F
                18851,  3539, 18093, 19385, 11545, 11445, 20468, 20460, // 7540 - 757F
                 8520,  2841, 14275, 16033, 16043,  8052, 19306, 20427, // 7580 - 75BF
                11431, 16060, 11359, 20371, 11308, 19002, 20346, 11281, // 75C0 - 75FF
                20337, 16070, 19002, 14167, 20295, 19002, 19002,  8061, // 7600 - 763F
                11243,  2159, 11198, 19002, 17559, 17478, 20242, 20207, // 7640 - 767F
                 2122,  8469, 20190, 19002, 19002,  8401, 19002, 11170, // 7680 - 76BF
                11120, 20163, 11051, 20329, 19228,  8318, 10984, 20139, // 76C0 - 76FF
                19138, 20100, 19002,  8292, 20091, 17413, 18799, 20078, // 7700 - 773F
                20069, 12455, 19002,  5126, 19102,  5135, 19002, 13085, // 7740 - 777F
                19002,  5144, 12693,  8244, 16080,  2077, 20050,  5153, // 7780 - 77BF
                19002, 19002, 18407, 11907, 10866, 12314, 11912, 19002, // 77C0 - 77FF
                 2797, 19002, 10803, 19002, 15646,  2011, 10734, 19002, // 7800 - 783F
                15498, 18348, 19002, 15508, 19002, 13847, 19002,  1927, // 7840 - 787F
                18959,  1897, 18932, 19002, 19358, 18885, 19002, 10671, // 7880 - 78BF
                18818, 10639, 20041, 19002, 19002, 20033, 14622, 13757, // 78C0 - 78FF
                18704,  8166,  8024, 19002, 19002, 10588, 19002, 10547, // 7900 - 793F
                20025, 19978, 19970, 10480, 19955, 19937, 19002, 10295, // 7940 - 797F
                18686, 11037, 11804, 19002,  7908, 10237, 11599, 11016, // 7980 - 79BF
                19893, 10838, 16803, 19885,  2959, 10168, 19002, 19185, // 79C0 - 79FF
                19852, 19790,  2903, 16595, 19781,  7800, 16505, 16947, // 7A00 - 7A3F
                19765,  9887, 18299, 19002,  9658,  9331, 19749, 18574, // 7A40 - 7A7F
                 9173, 19740, 10208, 19732, 19002,  8968, 19002,  4682, // 7A80 - 7ABF
                 1875, 10139, 18290, 16299, 19707, 10115,  7782, 16241, // 7AC0 - 7AFF
                19002, 18281, 16179, 18430, 19691,  1829, 19002, 15935, // 7B00 - 7B3F
                 7764, 18398, 19672, 19002, 19663,  1747, 10758, 13981, // 7B40 - 7B7F
                18182, 10929,  1724,  4590,  7402, 10390, 18329, 19654, // 7B80 - 7BBF
                19646, 15172,  9917, 19002, 19632, 15121, 19002, 19002, // 7BC0 - 7BFF
                18051, 19002,  4519,  7699, 18197,  9849, 19002, 13337, // 7C00 - 7C3F
                18740,  2111, 19002, 19002, 19593,  1951, 18246, 19002, // 7C40 - 7C7F
                18255, 14953,  9799, 19585, 10440,  7632, 18163, 18122, // 7C80 - 7CBF
                19002,  9717,  7422,  7277, 19577, 19002, 19002, 18264, // 7CC0 - 7CFF
                19569, 19554, 19546, 19534, 19495, 17873, 19432, 18082, // 7D00 - 7D3F
                 9621, 19002, 19423,  7199, 14773, 19407, 17973, 17934, // 7D40 - 7D7F
                19002,  7096,  4283,  1642, 19374,  1608, 14268, 19366, // 7D80 - 7DBF
                18042,  9603,  7001, 19334, 19295, 19287,  1502,  4292, // 7DC0 - 7DFF
                19002, 17831, 10400, 13460, 14156, 13245, 14052,  9546, // 7E00 - 7E3F
                13534, 19002,  4339,  6913, 13502, 17552, 19278, 17465, // 7E40 - 7E7F
                 9472,  1434, 13108, 19248, 19002, 19002, 19002, 19002, // 7E80 - 7EBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 7EC0 - 7EFF
                19002, 19002, 19002, 19002, 19002, 19002,  6866, 19236, // 7F00 - 7F3F
                19002,  1476, 19221, 19002, 19002,  9409, 19213, 13316, // 7F40 - 7F7F
                 7182, 19197,  1325,  9270,  1239, 19151,  4020, 19131, // 7F80 - 7FBF
                17406,  9230,  9127, 17757, 19092, 17387, 19083, 13072, // 7FC0 - 7FFF
                19028, 12787, 19020, 19011, 19002,  6020, 16092,  5052, // 8000 - 803F
                12016,  9026,  6802, 18975, 19002, 17696, 18967, 19326, // 8040 - 807F
                 1098, 17251,  6534, 15077, 12686, 17211, 17190,  8918, // 8080 - 80BF
                16636,  1071, 19002,  8863, 12307, 19002, 17116, 18952, // 80C0 - 80FF
                 8771, 18940, 19002, 18921, 16641, 12152, 19002, 17077, // 8100 - 813F
                19002, 11472, 18878, 19002,  2249, 11329, 18867, 18859, // 8140 - 817F
                18850, 18842, 19002,  8646, 18807, 18798, 11338,  8611, // 8180 - 81BF
                18760, 11824, 19002, 18752,  9703, 18697, 15089,  8534, // 81C0 - 81FF
                19724, 18679,  8495, 15094, 12067,  8453, 19526, 11794, // 8200 - 823F
                18671, 19002, 19002, 11589,   898,  6434, 17018, 18647, // 8240 - 827F
                19002,  8668,  8355, 16995,  3321, 11407, 16980, 18639, // 8280 - 82BF
                19002, 19002,  5503, 14645, 16882, 19002, 11006,  3021, // 82C0 - 82FF
                10831, 16793, 17442, 19002, 19002, 18631, 10610, 18620, // 8300 - 833F
                18612, 16735, 10157, 19002, 19002, 19002,  3366, 13953, // 8340 - 837F
                19002,  9945,  6334, 18596,  3840, 16584, 19002, 19316, // 8380 - 83BF
                16498,  9870, 10250,   722, 19002,  9825, 18582,  9651, // 83C0 - 83FF
                13581,  8274, 19002, 19002, 19002,  4353,  9321, 18567, // 8400 - 843F
                19002,  9163, 17425, 13586, 16330, 10083, 18898,  8200, // 8440 - 847F
                19002, 19002, 18539,  9097,  8958, 19002,  8105, 18531, // 8480 - 84BF
                18507, 16292,  3958,  8079, 19002,   815, 20488, 19062, // 84C0 - 84FF
                19002, 19002,  8831, 18456, 16234, 18339, 19002, 19947, // 8500 - 853F
                 3910, 16201,  9422, 16169, 11940, 18423, 19002, 19002, // 8540 - 857F
                 4533, 17368, 15992, 11954,  6145, 18415, 18406,  7989, // 8580 - 85BF
                15925,  7849,  1675, 12194, 19603, 15691, 17359,  7963, // 85C0 - 85FF
                 7922,  7721,  5938,  7666, 19002, 11613, 19002, 17101, // 8600 - 863F
                19002,  5808, 18391, 11968, 17062, 19002, 19002,  3464, // 8640 - 867F
                19002,  3603,  9243, 19002,  8991,  7392, 19002, 19002, // 8680 - 86BF
                16924,  8996,  8732, 15401,  9731, 19906, 19002,  5680, // 86C0 - 86FF
                18356, 18347, 19002, 18307, 19002, 19002, 19002, 19002, // 8700 - 873F
                19002,  5624, 10852, 16867,  5513, 18298,  4034, 18289, // 8740 - 877F
                 7612, 19800, 19002, 16858,  7572, 19002,  8880,  7500, // 8780 - 87BF
                19040, 19002, 19002, 19002, 18280,  3854,  7464, 15322, // 87C0 - 87FF
                15185, 19002, 19002, 16849,  7291, 19002, 15207,  8796, // 8800 - 883F
                18226,  9880,  7221, 15165, 15109, 18213, 18205,  5362, // 8840 - 887F
                15024, 18190,  3413,  5316, 19002, 15389, 18549, 19002, // 8880 - 88BF
                14942, 16823,  4870, 14897,  7256, 18181,  9388, 18156, // 88C0 - 88FF
                16752, 19002, 18106, 18072, 14278,  7110,  5254, 18059, // 8900 - 893F
                14751, 19002, 19002, 16613,  7514,  7031,  6987, 16604, // 8940 - 897F
                 6949, 14873,  8418, 19002,  6900,  3257,  3300,  6935, // 8980 - 89BF
                18050, 19002,  6832, 19002, 14845, 19002, 16424, 18041, // 89C0 - 89FF
                18028, 17997,  8378, 17989,  7820,  6752, 14387,  2094, // 8A00 - 8A3F
                19002, 19002, 17966,  7442, 17942, 14345, 17927, 14258, // 8A40 - 8A7F
                17416, 16371,  7359, 17911, 17881, 17843, 17820, 14243, // 8A80 - 8ABF
                 6738, 14330,  6608,  6652, 14222,  6567, 14197, 17795, // 8AC0 - 8AFF
                17773,  5228, 17765, 14149, 17756, 17734,  6657, 14067, // 8B00 - 8B3F
                14042,  6856, 19002, 17712,  5081,  7310, 17704, 12696, // 8B40 - 8B7F
                17695,  6472, 17679,  6448, 19002, 19002, 19002, 19002, // 8B80 - 8BBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 8BC0 - 8BFF
                19002, 19002, 19002, 19002, 19002, 19002, 16523, 16514, // 8C00 - 8C3F
                13834, 17627, 18888,  6375, 13524,  6356, 19002, 13491, // 8C40 - 8C7F
                 6251,  6203, 19002, 18707, 17619, 17611, 17603, 17595, // 8C80 - 8CBF
                17587, 17541, 13406,  6096, 17514,  6071, 14761,  6409, // 8CC0 - 8CFF
                14714, 17458,  2988, 19002, 19002, 19002, 19002, 19002, // 8D00 - 8D3F
                19002, 19002, 19002, 19002, 11708, 14030, 17450, 19002, // 8D40 - 8D7F
                19981,  5975, 19002, 13380, 13903, 17441,  5825,  5952, // 8D80 - 8DBF
                 5062, 13880, 19002, 13827, 13300, 17433,  5209, 19002, // 8DC0 - 8DFF
                19002,  5904, 17424,  4994, 19002,  5794, 17399, 19002, // 8E00 - 8E3F
                 2891, 17376, 19002,  6524, 17367, 19002,  6486, 19002, // 8E40 - 8E7F
                13150,  5763, 19002, 19002, 19002,  5656, 19002, 19002, // 8E80 - 8EBF
                17358,  5588,  5561, 16404, 19002,  5697, 19002, 17342, // 8EC0 - 8EFF
                 5707, 13062,  5536, 13794,  4931, 12989,  5637, 17334, // 8F00 - 8F3F
                14612, 12942, 19002, 18689,  5401, 19002, 19002, 19002, // 8F40 - 8F7F
                19002, 19002, 19002, 13693, 13698, 17326, 17314, 19002, // 8F80 - 8FBF
                 5376,  4835, 12833, 19002,  4772,  2723, 17306, 12780, // 8FC0 - 8FFF
                17290, 17282, 17274, 12737, 17235,  4611, 12679, 17219, // 9000 - 903F
                12645,  5294, 17204, 12550, 17178, 12356,  5280, 17170, // 9040 - 907F
                17162, 17154, 12300, 12227,  5167, 17146, 17109, 17100, // 9080 - 90BF
                12142,  5107, 19002,  4543, 12095, 17070, 11807, 19896, // 90C0 - 90FF
                 5024, 19002,  4945, 12057, 16395, 10171,  4810, 19002, // 9100 - 913F
                19002, 11999,  4722, 19002,  4661, 11896, 16182, 17061, // 9140 - 917F
                16277, 11787,  4565, 17637, 19002, 13670, 11582, 19002, // 9180 - 91BF
                17053, 11505, 17045, 17011,  2928,  4474, 15938, 19002, // 91C0 - 91FF
                19002,  7405, 17003,  4414, 19002, 19002,  5121,  1846, // 9200 - 923F
                16988, 11397, 16268, 13557,  4444, 19002, 19002, 19002, // 9240 - 927F
                16973, 19002, 11366, 16955, 19002, 19002,  4762,  5493, // 9280 - 92BF
                19002, 16965,  2534, 19002, 16942,  4400, 19002, 16932, // 92C0 - 92FF
                10924, 19002, 16923, 16915, 16898, 16890, 19002, 19002, // 9300 - 933F
                19002, 16875,  4278,  5184, 19002,  4374, 15175, 12895, // 9340 - 937F
                19002, 12011, 16770,  4310,  4984,  3827, 16866, 19002, // 9380 - 93BF
                19823, 19002, 10996,  4234, 10948, 19002,  4208, 19002, // 93C0 - 93FF
                11467, 19002, 19002, 16857, 15124, 20659, 14956, 16848, // 9400 - 943F
                13576, 19002, 10910, 20394, 19002, 19002, 19002, 13505, // 9440 - 947F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 9480 - 94BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 94C0 - 94FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 9500 - 953F
                19002, 19002, 19002, 19002, 19002, 19002, 16259, 19002, // 9540 - 957F
                16831, 10873, 10814, 16822, 20404, 16786,  5424, 12760, // 9580 - 95BF
                16250,  4264, 10020, 11935, 10683, 19002, 19002, 19002, // 95C0 - 95FF
                19002, 19002, 19002,  8791, 10600,  2444,  4186, 12765, // 9600 - 963F
                16760, 12730, 16751,  1213,  2397,  4124, 16743, 16728, // 9640 - 967F
                13319,  4056, 15158, 16703,  4110, 16685,  4959, 12424, // 9680 - 96BF
                16677, 10422, 19154,  4915,  3994, 16669, 16661, 10310, // 96C0 - 96FF
                16653, 19002, 10273, 10215, 16219, 19002, 16621, 10150, // 9700 - 973F
                 3817, 16612, 10008,  8563,  4597,  9938,  8481,  3799, // 9740 - 977F
                19002, 12434, 19002, 19002, 16603, 16577, 19002, 19002, // 9780 - 97BF
                 4156,    20, 20642,  8413, 19002, 19002,  3570, 14290, // 97C0 - 97FF
                16553, 16539, 16531, 16522, 19002, 19095, 16513, 16482, // 9800 - 983F
                 4138,  7644,  7354, 16458, 18870, 20363, 19002, 19002, // 9840 - 987F
                19002, 19002, 19002, 19002, 19002, 16450,  9860, 19002, // 9880 - 98BF
                 6045, 19002, 19002,  6001,  9814, 12070,  2983,  5845, // 98C0 - 98FF
                 1760,  9688, 16437, 16412, 16403, 16394, 19002, 19002, // 9900 - 993F
                11797,  9644,  9491, 19002, 19002, 19002, 19002, 19002, // 9940 - 997F
                19002, 19002,  4093,  4346, 11592, 16359,  9448,  9311, // 9980 - 99BF
                 9153, 19002, 16323,  9134, 19002, 11009,  9087, 16123, // 99C0 - 99FF
                 9045, 16315, 19002,  8948, 19002, 11739, 16307, 19002, // 9A00 - 9A3F
                16285, 16796, 18623,  3740,  3715,  8929, 19002, 19002, // 9A40 - 9A7F
                19002, 19002, 19002, 19002, 19002, 16276, 19002, 16267, // 9A80 - 9ABF
                19002, 19002, 11749, 16258,  9324,  3971, 19002, 19002, // 9AC0 - 9AFF
                19002, 19002, 19002,  3678, 16007,  3553,  3948, 14025, // 9B00 - 9B3F
                 8824, 15847, 13898,  3505, 19002, 15715, 19002, 19002, // 9B40 - 9B7F
                19002,  3769,  8699, 15570, 19002, 11754, 19002, 19002, // 9B80 - 9BBF
                19002,  8653,  3630, 19002,  5820, 16249, 19002, 19002, // 9BC0 - 9BFF
                19002,  9166, 16227, 19002,  8961, 19002,  3481,  1584, // 9C00 - 9C3F
                15524,  3899, 15458, 19002, 19002, 19002, 19002, 19002, // 9C40 - 9C7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 9C80 - 9CBF
                19002, 19002, 19002, 19002,  3961,  8230, 11681, 19002, // 9CC0 - 9CFF
                19002,  8211, 19002,  1263,  3519, 16218, 19002,   424, // 9D00 - 9D3F
                19002, 19002,  8152, 16172, 16190, 13875,  3351, 19002, // 9D40 - 9D7F
                19002, 19002, 19002, 19002, 19002,  8137, 13822, 19002, // 9D80 - 9DBF
                13552, 19002, 15427, 19002, 19002, 19002,  3191, 16152, // 9DC0 - 9DFF
                19002, 19002, 19002,  3138, 19002, 19002, 19002, 19002, // 9E00 - 9E3F
                19002, 19002, 19002, 19002, 19002, 19002, 15928,  8122, // 9E40 - 9E7F
                19002, 19002,  2017,  3467,  7395, 19002, 13373,   434, // 9E80 - 9EBF
                19619,  7259,  8008, 16131, 19002, 16122,  4757, 19002, // 9EC0 - 9EFF
                15140, 16015, 20575, 19002, 16006, 19002, 19002, 19511, // 9F00 - 9F3F
                19002,  3112,  3043, 14882,  7937,  1933, 14822, 19002, // 9F40 - 9F7F
                19002, 18075, 15985,  5179, 19002, 19002, 19002, 19002, // 9F80 - 9FBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // 9FC0 - 9FFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A000 - A03F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A040 - A07F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A080 - A0BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A0C0 - A0FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A100 - A13F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A140 - A17F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A180 - A1BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A1C0 - A1FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A200 - A23F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A240 - A27F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A280 - A2BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A2C0 - A2FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A300 - A33F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A340 - A37F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A380 - A3BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A3C0 - A3FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A400 - A43F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A440 - A47F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A480 - A4BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A4C0 - A4FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A500 - A53F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A540 - A57F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A580 - A5BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A5C0 - A5FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A600 - A63F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A640 - A67F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A680 - A6BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A6C0 - A6FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A700 - A73F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A740 - A77F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A780 - A7BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A7C0 - A7FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A800 - A83F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A840 - A87F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A880 - A8BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A8C0 - A8FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A900 - A93F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A940 - A97F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A980 - A9BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // A9C0 - A9FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // AA00 - AA3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // AA40 - AA7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // AA80 - AABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // AAC0 - AAFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // AB00 - AB3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // AB40 - AB7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // AB80 - ABBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // ABC0 - ABFF
                15977, 15914, 15906,  7842, 15892,  3157, 15869, 15855, // AC00 - AC3F
                15846, 20531, 12725, 15832, 19002, 19002, 15824, 15812, // AC40 - AC7F
                15780,  7732, 15757, 12616, 15731, 15723, 15714, 15684, // AC80 - ACBF
                 7683, 15670,  6903, 19002, 15662, 15654, 15628, 19811, // ACC0 - ACFF
                15601, 12280,  3616, 15578, 15569,  3457, 12135, 15555, // AD00 - AD3F
                 3124,  7538, 15532, 15523,  7475, 12050, 15480, 20510, // AD40 - AD7F
                 7381, 15466, 15457, 11780, 11734, 19002, 14790, 19002, // AD80 - ADBF
                15435, 15426, 15409, 11676, 15374, 19002, 19002, 15366, // ADC0 - ADFF
                15339, 15315,  1258, 19002, 19002, 19002, 15307, 15295, // AE00 - AE3F
                15279,  3010, 15253,   989, 15200, 15148, 15139, 15102, // AE40 - AE7F
                20626, 20505, 19002, 19002, 19002, 19002, 19002, 11312, // AE80 - AEBF
                15036, 11247, 15013, 14971, 19002, 14935, 20246, 14905, // AEC0 - AEFF
                14552, 14890, 14881, 19002, 19002, 20167, 14862,  2801, // AF00 - AF3F
                 7244, 14830, 14821, 10807,  9662, 19002, 19002,  7066, // AF40 - AF7F
                14798, 14789, 14744,  4686, 19002, 19002, 19002, 14730, // AF80 - AFBF
                14722, 14653,  9921,  4523, 19002, 14596, 14560, 14551, // AFC0 - AFFF
                14537,  9721, 14523, 19499, 19002, 14514, 19002, 19002, // B000 - B03F
                18086, 14490, 18098,  6942, 19002, 19002, 19002, 19378, // B040 - B07F
                14476, 17835,  6889, 14468, 14456, 14448,  3250, 14434, // B080 - B0BF
                 9476, 14403, 14370, 14361, 14338, 19002, 19002, 19002, // B0C0 - B0FF
                19002, 14319, 14298, 14251, 20351, 14230, 19240, 14205, // B100 - B13F
                14184, 14175, 14142, 14128, 14110, 19002, 19002, 14096, // B140 - B17F
                14078, 14060,  2973, 14015, 19002, 13920, 19002, 19002, // B180 - B1BF
                19002,  4024, 13888,  9030, 19002, 13865, 13855,  6849, // B1C0 - B1FF
                 8922, 13812, 18944,  6788, 13802, 19002,  8499,  8359, // B200 - B23F
                19002, 19002, 19002, 13779, 13769, 13740, 10161,  3844, // B240 - B27F
                 8204,  6759, 13678, 13651, 13635,  2917, 13613, 19002, // B280 - B2BF
                19002, 13542, 13513, 13484,  2855, 13476, 11286, 13468, // B2C0 - B2FF
                13437, 13419, 13399,  8109, 19002, 19002, 19002, 19002, // B300 - B33F
                19002, 19002,  8083, 13391,  7993,  6639, 13363, 13345, // B340 - B37F
                13293, 15695, 13279, 19002, 13267, 13253, 19002, 19002, // B380 - B3BF
                 7725, 13232, 13224,  6517, 13210, 13192, 19002,  7670, // B3C0 - B3FF
                19002, 19002, 13182, 13168, 13158, 13138,  7616, 19002, // B400 - B43F
                19002, 19002, 13124, 13093, 13055,  7576, 19002, 19002, // B440 - B47F
                13045, 13019, 19002, 14261,  7504, 12997, 14070,  6386, // B480 - B4BF
                12976, 12958, 12935,  7468, 12919, 15028,  6119, 12881, // B4C0 - B4FF
                19002, 19002, 14946, 12859,  7114, 12847, 12820, 12801, // B500 - B53F
                12773, 17564, 12745,  7035, 12711, 19002, 19002, 19002, // B540 - B57F
                19002, 19002, 19002, 19002, 12666, 12653, 12638, 17483, // B580 - B5BF
                12624,  6991, 12602, 12584, 19002,  3261, 19002, 19002, // B5C0 - B5FF
                19002, 19002, 12537, 12511, 14045,  3304,  6836, 19002, // B600 - B63F
                19002, 12470, 19002, 19002, 17847, 12442, 19002, 19002, // B640 - B67F
                19002, 19002, 19002, 17824, 12413, 20251,  6078, 19002, // B680 - B6BF
                19002, 19002,  6476, 19002, 19002, 19002, 12383, 12364, // B6C0 - B6FF
                12349, 19002, 19002, 19002, 19002, 12341, 12322, 12288, // B700 - B73F
                 6452, 12270,  6379, 19002, 12256, 12238, 12220, 13495, // B740 - B77F
                12168,  6255, 12160, 12125, 12106, 12088, 17545, 12078, // B780 - B7BF
                14515,  5986, 19002, 19002, 19002,  5979, 12040,  5798, // B7C0 - B7FF
                12032, 11986, 11976, 11889, 17380, 11859,  5565, 11839, // B800 - B83F
                11770, 19002,  5919,  5405, 11724, 17318,  5880, 11666, // B840 - B87F
                19002, 13527, 19002, 19002, 19002, 11656, 11638, 11628, // B880 - B8BF
                11575,  5111, 11561, 13383,  5770, 11539, 11513, 11498, // B8C0 - B8FF
                 5028, 19002, 19002, 11488, 11453, 19002, 19002,  4814, // B900 - B93F
                11439,  4726,  2831, 11425, 11415, 11390,  4665, 11353, // B940 - B97F
                 4569,  5595, 19002, 19002, 19002,  4478, 11302,  2538, // B980 - B9BF
                 5543, 11294, 11271, 11255,  5486, 11237,  4404, 11225, // B9C0 - B9FF
                11188, 11178, 13065, 19002, 19002, 19002, 19002, 11164, // BA00 - BA3F
                11152, 11136, 20256, 11114,  4314, 11094, 11045, 11024, // BA40 - BA7F
                20199, 10914, 19002, 19002, 19002, 10978, 10937, 10903, // BA80 - BABF
                 4268, 10893, 19002, 10881, 19002, 19002, 19002, 10687, // BAC0 - BAFF
                10860, 12230,  2787, 10797, 10787,  5412,  4190, 10747, // BB00 - BB3F
                 4128,  5383, 10728, 10709,  5347, 16689, 19002, 19002, // BB40 - BB7F
                19002, 10665, 10655, 19002, 16625, 10633,  4601, 19002, // BB80 - BBBF
                10582, 10564, 10532, 19002, 19002, 19002, 19002, 10524, // BBC0 - BBFF
                10512, 10504,  2730, 10496,  3803,  4952, 10474, 10448, // BC00 - BC3F
                10415,  4908, 10377, 12145, 19002, 19002, 19002, 19002, // BC40 - BC7F
                 9818, 10353,  8933,  4821, 10345, 10327, 10303,  3682, // BC80 - BCBF
                10289, 12098, 10266, 10231, 19002, 19002,  3557, 10202, // BCC0 - BCFF
                 3509,  4672, 10133, 19002,  3485,  3355, 19002, 19002, // BD00 - BD3F
                10123, 10109, 10091, 10066,  3195, 10056, 19002, 19002, // BD40 - BD7F
                10048, 10036, 10001,  4584, 16156, 19002,  9979,  9961, // BD80 - BDBF
                19002, 19002,  3047,  9911, 19002,  4509,  9843,  9833, // BDC0 - BDFF
                 9807, 15918,  9793, 15816, 19002, 19002, 19002, 19002, // BE00 - BE3F
                 7385,  9760, 15299,  4485,  9711,  9677,  9637, 10989, // BE40 - BE7F
                 9615, 14460,  9593,  9581, 19002,  9526, 19002, 19002, // BE80 - BEBF
                19002, 19002,  9518,  9499,  9484, 14082, 19002, 19002, // BEC0 - BEFF
                 4329,  9462, 19002,  9441, 19002, 19002, 19002, 19002, // BF00 - BF3F
                 9403,  9358,  9304, 19002, 19002, 19002, 19002, 19002, // BF40 - BF7F
                19002, 19002, 13655, 19002, 19002, 19002,  9286, 19002, // BF80 - BFBF
                12060, 13517,  9264, 12851,  4079, 19002, 19002, 19002, // BFC0 - BFFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 11275, // C000 - C03F
                19002, 19002,  4010,  9224,  9214,  9142, 19002, 19002, // C040 - C07F
                19002, 19002,  9121,  9111,  9080, 11229,  9056, 11156, // C080 - C0BF
                 9038,  9020,  8976,  8941, 10941,  8912, 10516,  3937, // C0C0 - C0FF
                 8857,  8839,  8817,  4825,  8809, 10040,  8779,  8765, // C100 - C13F
                 8710,  8692,  9681,  8640,  7949,  8619,  8605,  8586, // C140 - C17F
                11400,  7861,  8551,  6724,  3892,  8528,  8518,  5496, // C180 - C1BF
                 6245,  8489,  6057,  8467,  8447,  8399,  8386,  5873, // C1C0 - C1FF
                 8349,  5463,  3869,  8341,  8316,  8308,  3834, 19002, // C200 - C23F
                19002,  8290,  8268,  8242,  8223,  2813,  8194,  2353, // C240 - C27F
                 3755,  8182,  8164,  8145,  1736,  8130,  1344,  3722, // C280 - C2BF
                19002, 19002, 19002,  2495,  8099, 19143,  3689,  8073, // C2C0 - C2FF
                 8022,  8001, 20109,  7983,  2380,  7971, 19002, 19002, // C300 - C33F
                10999, 19002, 19002, 19002, 19002,  7957,  7945,  7930, // C340 - C37F
                 9345,  7916, 19002, 19002, 19002, 19002, 19002, 19002, // C380 - C3BF
                 7906, 19002, 19002,  7898,  7857,  7835, 15234,  7798, // C3C0 - C3FF
                19002,  7780,  7762, 19002, 11949, 11963,  7715,  9005, // C400 - C43F
                19002,  7697, 19002, 19002,  9931,  7660, 11071,  3593, // C440 - C47F
                 7630, 19002,  9780,  8430, 19002, 19002, 19002,  7606, // C480 - C4BF
                19002, 19002, 19002, 19002, 19002,  3446,  7566,  7558, // C4C0 - C4FF
                 7527,  6681,  7494,  7300, 19002,  7458,  7420,  7374, // C500 - C53F
                20219,  7342, 20083,  7318,  7285,  7275,  7237, 18115, // C540 - C57F
                 7229, 14854,  3428,  7215,  7197,  3403, 17893,  7160, // C580 - C5BF
                 4881,  7122,  7104,  7094,  7059, 17663,  7051, 13912, // C5C0 - C5FF
                 7043,  7025,  6999,  6981,  5834,  6965,  6957,  3285, // C600 - C63F
                 6929,  6911,  6882, 13309,  6864,  5218,  3239,  6826, // C640 - C67F
                 6800,  6781, 17030,  6746, 15245,  2084,  6732,  6720, // C680 - C6BF
                 6632, 17244,  6602,  5193,  6590,  6561,  6532,  6510, // C6C0 - C6FF
                20668,  6466, 16840,  3094,  6442,  6432,  6424, 15613, // C700 - C73F
                 6369,  2765,  3076,  6350,  6332,  6303,  4003,  6287, // C740 - C77F
                10739,  6271,  6263,  6241,  6217, 15054,  6197, 10319, // C780 - C7BF
                 6185,  6162,  6143,  6112, 14989,  6090, 19002, 19002, // C7C0 - C7FF
                 6065,  6053,  6009, 16491,  5969,  1769,  3054,  5946, // C800 - C83F
                 5936,  5912, 19353, 19002, 19002, 19002,  5898,  5869, // C840 - C87F
                 5861, 10676, 19180,  5427,  2943,  5806, 19002, 10824, // C880 - C8BF
                13455,  5788, 17533,  2881,  5757, 19002, 10603, 20584, // C8C0 - C8FF
                 5726, 15418,  2820,  5678, 19002, 15072, 11324, 19002, // C900 - C93F
                19002, 19002,  5650,  5622,  5573,  9698,  5555, 14640, // C940 - C97F
                19002,  5530,  5511,  5479, 19002, 19002, 19002, 19002, // C980 - C9BF
                 5471,  5459,  5443,  2780,  5435,  5928,  5391,  5370, // C9C0 - C9FF
                 5360,  5340, 13948,  5314, 19002,  2713, 19002, 19002, // CA00 - CA3F
                19002, 13992,  5288,  9993,  5270,  5252, 19002,  9863, // CA40 - CA7F
                10541, 19002, 19002,  5226, 19002, 19002, 19002,  9372, // CA80 - CABF
                 5161,  4494,  2660,  5097,  5079,  3294,  8875, 19002, // CAC0 - CAFF
                19002,  5060,  5018,  4992,  4967, 19002, 19002, 19002, // CB00 - CB3F
                 2641,  4939,  4929,  4901, 15384, 19002, 19002,  4889, // CB40 - CB7F
                19002, 19002, 19002,  9383, 19002, 19002, 19002,  4833, // CB80 - CBBF
                19002, 19002,  5582, 19002,  2669,  2619, 19002, 19002, // CBC0 - CBFF
                19002, 14380,  4804,  7172,  2601,  4788,  4770,  4734, // CC00 - CC3F
                 6546,  4716,  6135,  4651,  4635,  4609,  4577, 19002, // CC40 - CC7F
                19002, 19002, 19002,  4559,  4541,  4502,  2066,  4468, // CC80 - CCBF
                 5718,  1836,  4438, 19002,  5204,  5692,  4412, 19002, // CCC0 - CCFF
                 2524,  4394,  4372,  4322, 13789,  4304, 19002,  2502, // CD00 - CD3F
                19002, 19002, 19002,  8741,  4258,  6773,  2434,  4232, // CD40 - CD7F
                19002,  4206,  8750,  4180,  8439,  2387,  4154, 19002, // CD80 - CDBF
                13688,  6690,  4136, 19002, 19002,  4118,  4091,  4072, // CDC0 - CDFF
                13665,  4050,  6699,  2294,  3988,  3969,  3930, 19002, // CE00 - CE3F
                19002, 19002, 19002,  3922,  3881,  3862,  5607,  3811, // CE40 - CE7F
                 5306,  2272,  3793,  3767,  3748, 20010, 19002, 19002, // CE80 - CEBF
                 2219, 19002, 19002, 19002,  1396,  3734, 19877,  3705, // CEC0 - CEFF
                 3672,  3628,  3586,  1518,  3547, 17266,  3535,  3517, // CF00 - CF3F
                19002, 19002,  4746,  3499,  4386,  2155,  3475,  3436, // CF40 - CF7F
                 3421, 12755, 19002, 19002,  2118,  3393,  3374, 19002, // CF80 - CFBF
                 4105, 19002, 19002, 19002,  3345,  3275,  3232, 20637, // CFC0 - CFFF
                 3185, 19002,  2073,  3165, 19002,  9314,  3642,  3132, // D000 - D03F
                17812,  2007,  3106,  3084,  3066, 18322,  3037, 19992, // D040 - D07F
                 1923, 19002, 19002, 19002, 15962,  3004, 16814,  1893, // D080 - D0BF
                 2967,  2955,  2936,  1125,  2911, 11648,  2899,  2871, // D0C0 - D0FF
                19002,  9156, 19002, 19002, 19002, 19002,  2849,  2809, // D100 - D13F
                 2773, 14998,  2759,  9573,  1871,  2738, 19002, 17646, // D140 - D17F
                19614,  2703, 19002, 19002,  2688,  2677,  2653,  2744, // D180 - D1BF
                 2631, 19002, 19002,  2609, 19002, 19002,  8056,  2591, // D1C0 - D1FF
                14362,  1825,  2580, 19002, 19002, 20375,  2546,  8065, // D200 - D23F
                 1743,  2514, 19002, 20194, 20104, 19002, 19002, 19002, // D240 - D27F
                 2484,  2473,  2427,  5130,  2369,  5139,  1720,  2361, // D280 - D2BF
                 2349,  2336,  5148,  2321, 20054, 19002,  2306,  2284, // D2C0 - D2FF
                 2265, 11916,  2231, 13761,  2212,  2194,  2183,  2148, // D300 - D33F
                11603, 19002, 19002, 19002, 19002, 19002, 19002, 10842, // D340 - D37F
                 2130, 19189,  2107,  2045,  2025,  2000,  9335,  1982, // D380 - D3BF
                10143,  1947,  1913,  1883,  1615,  9853,  1854, 18744, // D3C0 - D3FF
                 1638,  1815, 19002,  9090, 19002, 19002, 19002, 19002, // D400 - D43F
                 1800, 19002, 19002, 18250,  1785,  9048, 19002,  1777, // D440 - D47F
                 1732,  1713, 18259, 19002, 19002,  1604, 19002, 19002, // D480 - D4BF
                19002, 18167,  1690, 19538, 19002,  1657,  1628,  1597, // D4C0 - D4FF
                 4287,  1561,  9607, 19002, 19002, 19002, 19002, 19299, // D500 - D53F
                 1546,  4296,  1498,  1531,  1491,  1465, 14160,  1442, // D540 - D57F
                17391,  1430,  1419, 19002,  8951, 19002, 19002, 19002, // D580 - D5BF
                19002,  1352,  1340,  1307, 18925,  1284, 16645,  1472, // D5C0 - D5FF
                 1221,  1190,  1170, 11333,  1148,  3951, 19002,  1106, // D600 - D63F
                 1087,  1060,  1314,  1045, 14176,  1321,  1027, 19002, // D640 - D67F
                10593, 18811,  1012,  8702,  1235,   997,   974,  1177, // D680 - D6BF
                13957,   956,  1162,  1094,   922,   884,   854, 16588, // D6C0 - D6FF
                  839, 19002,  1067,   797,   779,   763, 13590,   745, // D700 - D73F
                 3914,   894,   737,   718,   702,   861,   684,  9426, // D740 - D77F
                  811,   669,   650,   634, 19002, 19002, 19002, 19002, // D780 - D7BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D7C0 - D7FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D800 - D83F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D840 - D87F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D880 - D8BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D8C0 - D8FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D900 - D93F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D940 - D97F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D980 - D9BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // D9C0 - D9FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DA00 - DA3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DA40 - DA7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DA80 - DABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DAC0 - DAFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DB00 - DB3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DB40 - DB7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DB80 - DBBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DBC0 - DBFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DC00 - DC3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DC40 - DC7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DC80 - DCBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DCC0 - DCFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DD00 - DD3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DD40 - DD7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DD80 - DDBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DDC0 - DDFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DE00 - DE3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DE40 - DE7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DE80 - DEBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DEC0 - DEFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DF00 - DF3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DF40 - DF7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DF80 - DFBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // DFC0 - DFFF
                  626,   618,   610,   602,   594,   586,   578,   570, // E000 - E03F
                  562,   554,   546,   538,   530,   522,   514,   506, // E040 - E07F
                  498,   490,   482,   474,   466,   458,   450,   406, // E080 - E0BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E0C0 - E0FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E100 - E13F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E140 - E17F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E180 - E1BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E1C0 - E1FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E200 - E23F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E240 - E27F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E280 - E2BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E2C0 - E2FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E300 - E33F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E340 - E37F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E380 - E3BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E3C0 - E3FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E400 - E43F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E440 - E47F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E480 - E4BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E4C0 - E4FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E500 - E53F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E540 - E57F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E580 - E5BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E5C0 - E5FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E600 - E63F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E640 - E67F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E680 - E6BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E6C0 - E6FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E700 - E73F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E740 - E77F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E780 - E7BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E7C0 - E7FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E800 - E83F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E840 - E87F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E880 - E8BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E8C0 - E8FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E900 - E93F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E940 - E97F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E980 - E9BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // E9C0 - E9FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EA00 - EA3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EA40 - EA7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EA80 - EABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EAC0 - EAFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EB00 - EB3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EB40 - EB7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EB80 - EBBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EBC0 - EBFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EC00 - EC3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EC40 - EC7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EC80 - ECBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // ECC0 - ECFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // ED00 - ED3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // ED40 - ED7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // ED80 - EDBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EDC0 - EDFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EE00 - EE3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EE40 - EE7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EE80 - EEBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EEC0 - EEFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EF00 - EF3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EF40 - EF7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EF80 - EFBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // EFC0 - EFFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F000 - F03F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F040 - F07F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F080 - F0BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F0C0 - F0FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F100 - F13F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F140 - F17F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F180 - F1BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F1C0 - F1FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F200 - F23F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F240 - F27F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F280 - F2BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F2C0 - F2FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F300 - F33F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F340 - F37F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F380 - F3BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F3C0 - F3FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F400 - F43F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F440 - F47F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F480 - F4BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F4C0 - F4FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F500 - F53F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F540 - F57F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F580 - F5BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F5C0 - F5FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F600 - F63F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F640 - F67F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F680 - F6BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F6C0 - F6FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F700 - F73F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F740 - F77F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F780 - F7BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F7C0 - F7FF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F800 - F83F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F840 - F87F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F880 - F8BF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // F8C0 - F8FF
                  398,   390,   382,   374,   366,   358,   350,   342, // F900 - F93F
                  334,   326,   318,   310,   302,   294,   286,   278, // F940 - F97F
                  270,   262,   254,   246,   238,   230,   222,   214, // F980 - F9BF
                  206,   198,   190,   182,   174,   166,   158,   150, // F9C0 - F9FF
                  142,   129, 19002, 19002, 19002, 19002, 19002, 19002, // FA00 - FA3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FA40 - FA7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FA80 - FABF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FAC0 - FAFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FB00 - FB3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FB40 - FB7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FB80 - FBBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FBC0 - FBFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FC00 - FC3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FC40 - FC7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FC80 - FCBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FCC0 - FCFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FD00 - FD3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FD40 - FD7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FD80 - FDBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FDC0 - FDFF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FE00 - FE3F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FE40 - FE7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FE80 - FEBF
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FEC0 - FEFF
                  789,   121,   113,   105,    97,    89,    81,    73, // FF00 - FF3F
                   65,    57,    49,    33, 19002, 19002, 19002, 19002, // FF40 - FF7F
                19002, 19002, 19002, 19002, 19002, 19002, 19002, 19002, // FF80 - FFBF
                19002, 19002, 19002, 19002,     0, 19002, 19002, 19002,
        };

        private static final String index2 =
                "\uA1CB\uA1CC\uA1FE\uA3FE\u0000\uA1CD\uA3DC\u0000\uD7B9\u0000" + //     0 -     9
                "\uE9C3\u0000\uE8FD\uE8AF\u0000\u0000\uDFBB\u0000\u0000\uF3A5" + //    10 -    19
                "\u0000\u0000\u0000\uEADF\u0000\u0000\u0000\u0000\uE4EF\u0000" + //    20 -    29
                "\u0000\u0000\uE9B9\uA3F8\uA3F9\uA3FA\uA3FB\uA3FC\uA3FD\u0000" + //    30 -    39
                "\u0000\u0020\u0021\"\u0023\u0024\u0025\u0026\u0027\uA3F0" + //    40 -    49
                "\uA3F1\uA3F2\uA3F3\uA3F4\uA3F5\uA3F6\uA3F7\uA3E8\uA3E9\uA3EA" + //    50 -    59
                "\uA3EB\uA3EC\uA3ED\uA3EE\uA3EF\uA3E0\uA3E1\uA3E2\uA3E3\uA3E4" + //    60 -    69
                "\uA3E5\uA3E6\uA3E7\uA3D8\uA3D9\uA3DA\uA3DB\uA1AC\uA3DD\uA3DE" + //    70 -    79
                "\uA3DF\uA3D0\uA3D1\uA3D2\uA3D3\uA3D4\uA3D5\uA3D6\uA3D7\uA3C8" + //    80 -    89
                "\uA3C9\uA3CA\uA3CB\uA3CC\uA3CD\uA3CE\uA3CF\uA3C0\uA3C1\uA3C2" + //    90 -    99
                "\uA3C3\uA3C4\uA3C5\uA3C6\uA3C7\uA3B8\uA3B9\uA3BA\uA3BB\uA3BC" + //   100 -   109
                "\uA3BD\uA3BE\uA3BF\uA3B0\uA3B1\uA3B2\uA3B3\uA3B4\uA3B5\uA3B6" + //   110 -   119
                "\uA3B7\uA3A8\uA3A9\uA3AA\uA3AB\uA3AC\uA3AD\uA3AE\uA3AF\uFAA1" + //   120 -   129
                "\uFAA2\uFAE6\uFCA9\u0000\u0000\u0000\u0000\uD8D0\u0000\uF0C8" + //   130 -   139
                "\uD1A1\uD1A2\uF4EE\uF6F4\uF6F6\uF7B8\uF7C8\uF7D3\uF8DB\uF8F0" + //   140 -   149
                "\uEDA2\uEDA3\uEDEE\uEEDB\uF2BD\uF2FA\uF3B1\uF4A7\uECE1\uECE4" + //   150 -   159
                "\uECE7\uECE8\uECF7\uECF8\uECFA\uEDA1\uECC1\uECC5\uECC6\uECC9" + //   160 -   169
                "\uECCA\uECD5\uECDD\uECDE\uECAF\uECB0\uECB1\uECB2\uECB5\uECB8" + //   170 -   179
                "\uECBA\uECC0\uEBCF\uEBD0\uEBD1\uEBD2\uEBD8\uECA6\uECA7\uECAA" + //   180 -   189
                "\uEBBA\uEBBB\uEBBD\uEBC1\uEBC2\uEBC6\uEBC7\uEBCC\uEAF4\uEAF7" + //   190 -   199
                "\uEAFC\uEAFE\uEBA4\uEBA7\uEBA9\uEBAA\uE8FB\uE8FE\uE9A7\uE9AC" + //   200 -   209
                "\uE9CC\uE9F7\uEAC1\uEAE5\uE7E6\uE7F7\uE8E7\uE8E8\uE8F0\uE8F1" + //   210 -   219
                "\uE8F7\uE8F9\uE7C7\uE7CB\uE7CD\uE7CF\uE7D0\uE7D3\uE7DF\uE7E4" + //   220 -   229
                "\uE7A9\uE7AA\uE7AC\uE7AD\uE7B0\uE7BF\uE7C1\uE7C6\uE6F1\uE6F2" + //   230 -   239
                "\uE6F5\uE6F6\uE6F7\uE6F9\uE7A1\uE7A6\uE6E4\uE6E5\uE6E6\uE6E8" + //   240 -   249
                "\uE6EA\uE6EB\uE6EC\uE6EF\uE6C7\uE6CA\uE6D2\uE6D6\uE6D9\uE6DC" + //   250 -   259
                "\uE6DF\uE6E1\uE6B0\uE6B1\uE6B3\uE6B7\uE6B8\uE6BC\uE6C4\uE6C6" + //   260 -   269
                "\uE5FB\uE5FC\uE5FE\uE6A1\uE6A4\uE6A7\uE6AD\uE6AF\uE5BB\uE5BC" + //   270 -   279
                "\uE5C4\uE5CE\uE5D0\uE5D2\uE5D6\uE5FA\uE1ED\uE3F5\uE4A1\uE4A9" + //   280 -   289
                "\uE5AE\uE5B1\uE5B2\uE5B9\uDDF4\uDEFC\uDEFE\uDFB3\uDFE1\uDFE8" + //   290 -   299
                "\uE0F1\uE1AD\uD6CD\uD7CB\uD7E4\uDBC5\uDBE4\uDCA5\uDDA5\uDDD5" + //   300 -   309
                "\uD2F7\uD2F8\uD4E6\uD4FC\uD5A5\uD5AB\uD5AE\uD6B8\uD2EA\uD2EB" + //   310 -   319
                "\uD2F0\uD2F1\uD2F2\uD2F3\uD2F4\uD2F5\uD2E1\uD2E2\uD2E4\uD2E5" + //   320 -   329
                "\uD2E6\uD2E7\uD2E8\uD2E9\uD2D4\uD2D5\uD2D6\uD2D7\uD2D9\uD2DA" + //   330 -   339
                "\uD2DE\uD2DF\uD2CB\uD2CD\uD2CE\uD2CF\uD2D0\uD2D1\uD2D2\uD2D3" + //   340 -   349
                "\uD2C2\uD2C3\uD2C4\uD2C6\uD2C7\uD2C8\uD2C9\uD2CA\uD2A7\uD2A8" + //   350 -   359
                "\uD2A9\uD2AA\uD2AB\uD2AD\uD2B2\uD2BE\uD1F2\uD1F6\uD1FA\uD1FC" + //   360 -   369
                "\uD1FD\uD1FE\uD2A2\uD2A3\uD1E6\uD1E8\uD1E9\uD1EA\uD1EB\uD1ED" + //   370 -   379
                "\uD1EF\uD1F0\uD1DD\uD1DE\uD1DF\uD1E0\uD1E2\uD1E3\uD1E4\uD1E5" + //   380 -   389
                "\uD0B8\uD0D0\uD0DD\uD1D4\uD1D5\uD1D8\uD1DB\uD1DC\uCBD0\uCBD6" + //   390 -   399
                "\uCBE7\uCDCF\uCDE8\uCEAD\uCFFB\uD0A2\uFEFB\uFEFC\uFEFD\uFEFE" + //   400 -   409
                "\u0000\u0000\u0000\u0000\uEDD5\u0000\u0000\uD9BA\u0000\uE9DC" + //   410 -   419
                "\uD9C1\u0000\uF5F2\uE0C5\u0000\u0000\u0000\uFBF8\u0000\u0000" + //   420 -   429
                "\u0000\u0000\uCCF9\uCDB5\u0000\u0000\u0000\uD8AB\u0000\u0000" + //   430 -   439
                "\uFDCB\u0000\uA7BF\uA7C0\uA7C1\uA7C2\uA7C3\uA7C4\uA7C5\uA7C6" + //   440 -   449
                "\uFEF3\uFEF4\uFEF5\uFEF6\uFEF7\uFEF8\uFEF9\uFEFA\uFEEB\uFEEC" + //   450 -   459
                "\uFEED\uFEEE\uFEEF\uFEF0\uFEF1\uFEF2\uFEE3\uFEE4\uFEE5\uFEE6" + //   460 -   469
                "\uFEE7\uFEE8\uFEE9\uFEEA\uFEDB\uFEDC\uFEDD\uFEDE\uFEDF\uFEE0" + //   470 -   479
                "\uFEE1\uFEE2\uFED3\uFED4\uFED5\uFED6\uFED7\uFED8\uFED9\uFEDA" + //   480 -   489
                "\uFECB\uFECC\uFECD\uFECE\uFECF\uFED0\uFED1\uFED2\uFEC3\uFEC4" + //   490 -   499
                "\uFEC5\uFEC6\uFEC7\uFEC8\uFEC9\uFECA\uFEBB\uFEBC\uFEBD\uFEBE" + //   500 -   509
                "\uFEBF\uFEC0\uFEC1\uFEC2\uFEB3\uFEB4\uFEB5\uFEB6\uFEB7\uFEB8" + //   510 -   519
                "\uFEB9\uFEBA\uFEAB\uFEAC\uFEAD\uFEAE\uFEAF\uFEB0\uFEB1\uFEB2" + //   520 -   529
                "\uFEA3\uFEA4\uFEA5\uFEA6\uFEA7\uFEA8\uFEA9\uFEAA\uC9F9\uC9FA" + //   530 -   539
                "\uC9FB\uC9FC\uC9FD\uC9FE\uFEA1\uFEA2\uC9F1\uC9F2\uC9F3\uC9F4" + //   540 -   549
                "\uC9F5\uC9F6\uC9F7\uC9F8\uC9E9\uC9EA\uC9EB\uC9EC\uC9ED\uC9EE" + //   550 -   559
                "\uC9EF\uC9F0\uC9E1\uC9E2\uC9E3\uC9E4\uC9E5\uC9E6\uC9E7\uC9E8" + //   560 -   569
                "\uC9D9\uC9DA\uC9DB\uC9DC\uC9DD\uC9DE\uC9DF\uC9E0\uC9D1\uC9D2" + //   570 -   579
                "\uC9D3\uC9D4\uC9D5\uC9D6\uC9D7\uC9D8\uC9C9\uC9CA\uC9CB\uC9CC" + //   580 -   589
                "\uC9CD\uC9CE\uC9CF\uC9D0\uC9C1\uC9C2\uC9C3\uC9C4\uC9C5\uC9C6" + //   590 -   599
                "\uC9C7\uC9C8\uC9B9\uC9BA\uC9BB\uC9BC\uC9BD\uC9BE\uC9BF\uC9C0" + //   600 -   609
                "\uC9B1\uC9B2\uC9B3\uC9B4\uC9B5\uC9B6\uC9B7\uC9B8\uC9A9\uC9AA" + //   610 -   619
                "\uC9AB\uC9AC\uC9AD\uC9AE\uC9AF\uC9B0\uC9A1\uC9A2\uC9A3\uC9A4" + //   620 -   629
                "\uC9A5\uC9A6\uC9A7\uC9A8\uC8FB\uC8FC\u0000\uC8FD\u0000\uC8FE" + //   630 -   639
                "\u0000\u0000\u0018\u0019\u001A\u001B\u001C\u001D\u001E\u001F" + //   640 -   649
                "\uC8FA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF4B9\uF1B6" + //   650 -   659
                "\u0000\uF6FD\u0000\uDBF7\u0000\u0000\u0000\u0000\uFBEA\uC8F7" + //   660 -   669
                "\uC8F8\u0000\u0000\uC8F9\u0000\u0000\u0000\uEDA5\uEEF2\u0000" + //   670 -   679
                "\u0000\u0000\u0000\uDCF9\uC8F2\u0000\u0000\u0000\uC8F3\u0000" + //   680 -   689
                "\u0000\u0000\uE0E6\u0000\u0000\u0000\u0000\u0000\u0000\uD0BF" + //   690 -   699
                "\u0000\uFAAC\uC8EC\uC8ED\u0000\uC8EE\u0000\uC8EF\u0000\u0000" + //   700 -   709
                "\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017\uC8EA\uC8EB" + //   710 -   719
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF3F8\u0000\u0000\u0000" + //   720 -   729
                "\uFAFD\u0000\u0000\u0000\u0000\u0000\uD6AA\uC8E5\uC8E6\u0000" + //   730 -   739
                "\u0000\uC8E7\u0000\uC8E8\uC8E9\uC8E0\u0000\u0000\u0000\uC8E1" + //   740 -   749
                "\u0000\u0000\u0000\uDBED\u0000\u0000\u0000\u0000\u0000\u0000" + //   750 -   759
                "\uE6CB\u0000\uF5F1\uC8DA\uC8DB\u0000\uC8DC\u0000\uC8DD\u0000" + //   760 -   769
                "\u0000\u0008\u0009\n\u000B\u000C\r\u000E\u000F\uC8D9" + //   770 -   779
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE8E4\u0000\u0000" + //   780 -   789
                "\uA3A1\uA3A2\uA3A3\uA3A4\uA3A5\uA3A6\uA3A7\uC8D6\uC8D7\u0000" + //   790 -   799
                "\u0000\uC8D8\u0000\u0000\u0000\uD5D8\u0000\uF0BD\uD7D0\uD4D0" + //   800 -   809
                "\u0000\u0000\uC8F6\u0000\u0000\u0000\u0000\u0000\u0000\uDCEF" + //   810 -   819
                "\u0000\uD6A5\u0000\uE6FC\u0000\u0000\uD7FB\uD0D6\uDDF5\uF7F1" + //   820 -   829
                "\u0000\uA7BE\uA7E5\uA7E6\uA7E7\uA7E8\uA7E1\uA7E2\uA7E3\uC8D3" + //   830 -   839
                "\u0000\u0000\u0000\uC8D4\u0000\u0000\u0000\uCDEC\u0000\u0000" + //   840 -   849
                "\u0000\uDCB2\uD0EC\uCEFD\uC8CF\u0000\u0000\u0000\u0000\uC8D0" + //   850 -   859
                "\u0000\u0000\uC8F0\u0000\u0000\uC8F1\u0000\u0000\u0000\uD7D7" + //   860 -   869
                "\uDFA2\u0000\u0000\u0000\uCEBE\u0000\uCDE4\u0000\uD1AE\uDCED" + //   870 -   879
                "\uE8CE\u0000\uF0F9\uCEB5\uC8CE\u0000\u0000\u0000\u0000\u0000" + //   880 -   889
                "\u0000\u0000\uD4F1\u0000\u0000\uC8E4\u0000\u0000\u0000\u0000" + //   890 -   899
                "\u0000\u0000\uEBFA\u0000\uF9E6\u0000\uEFA2\uE2DA\uF6FC\u0000" + //   900 -   909
                "\u0000\uFBD0\uD1AD\u0000\uA7B2\uA7B3\uA7B4\uA7A7\uA7A8\uA7A9" + //   910 -   919
                "\uA7AA\uA7BD\uC8CC\u0000\u0000\u0000\uC8CD\u0000\u0000\u0000" + //   920 -   929
                "\uD6ED\u0000\u0000\u0000\u0000\u0000\u0000\uE3D5\uF5D0\u0000" + //   930 -   939
                "\uF1C1\u0000\u0000\uE2E9\uDCCA\uECB4\uFAC0\u0000\uA7A5\uA7AB" + //   940 -   949
                "\uA7AC\uA7AD\uA7AE\uA7AF\uA7B0\uA7B1\uC8C6\u0000\u0000\u0000" + //   950 -   959
                "\uC8C7\u0000\u0000\u0000\uCBFB\u0000\u0000\u0000\u0000\u0000" + //   960 -   969
                "\u0000\uF0C9\u0000\uFCFC\uC8C1\u0000\u0000\u0000\u0000\u0000" + //   970 -   979
                "\u0000\u0000\uD5CF\uD0F8\u0000\uDAD6\u0000\uCAB1\u0000\u0000" + //   980 -   989
                "\u0000\u0000\u0000\uB1F4\uB1F5\u0000\uB1F6\uC8BF\u0000\u0000" + //   990 -   999
                "\u0000\uC8C0\u0000\u0000\u0000\uCAA4\u0000\uDBF8\u0000\u0000" + //  1000 -  1009
                "\u0000\uDEC7\uC8BA\u0000\u0000\u0000\uC8BB\u0000\u0000\u0000" + //  1010 -  1019
                "\uF4D2\uE0BA\u0000\u0000\u0000\u0000\uDFC0\uC8B3\uC8B4\u0000" + //  1020 -  1029
                "\u0000\uC8B5\u0000\u0000\u0000\uCBF8\u0000\u0000\u0000\u0000" + //  1030 -  1039
                "\u0000\u0000\uF9F8\u0000\uDBCA\uC8AF\u0000\u0000\u0000\uC8B0" + //  1040 -  1049
                "\u0000\u0000\u0000\uFDA6\uEBEF\u0000\uF4A6\u0000\uCCCA\uF3A8" + //  1050 -  1059
                "\uC8A8\uC8A9\u0000\uC8AA\u0000\uC8AB\u0000\u0000\uC8D5\u0000" + //  1060 -  1069
                "\u0000\u0000\u0000\u0000\u0000\uDBCE\u0000\uF7C3\u0000\uD1AC" + //  1070 -  1079
                "\u0000\u0000\u0000\u0000\u0000\u0000\uDAC7\uC8A6\u0000\u0000" + //  1080 -  1089
                "\u0000\u0000\uC8A7\u0000\u0000\uC8CB\u0000\u0000\u0000\u0000" + //  1090 -  1099
                "\u0000\u0000\uECBD\uE2DC\uDEEB\uF0DC\uC8A3\uC8A4\u0000\u0000" + //  1100 -  1109
                "\uC8A5\u0000\u0000\u0000\uD9FA\uD3EE\u0000\u0000\u0000\uFAB8" + //  1110 -  1119
                "\u0000\uE2AE\u0000\uD3B7\uFACC\u0000\u0000\u0000\u0000\uC5C2" + //  1120 -  1129
                "\uC5C3\u0000\u0000\uE5DB\uF8F7\u0000\u0000\u0000\uF6D4\u0000" + //  1130 -  1139
                "\uA7D4\uA7D5\uA7D6\uA7D7\uA7D8\uA7A1\uA7A2\uA7A3\uC7FE\u0000" + //  1140 -  1149
                "\u0000\u0000\uC8A1\u0000\u0000\u0000\uF3DA\u0000\uCBC1\u0000" + //  1150 -  1159
                "\uDBC3\u0000\u0000\uC8C8\u0000\u0000\uC8C9\u0000\u0000\uC8CA" + //  1160 -  1169
                "\uC7F8\uC7F9\u0000\uC7FA\uC7FB\uC7FC\u0000\u0000\uC8C2\u0000" + //  1170 -  1179
                "\uC8C3\u0000\u0000\u0000\u0000\uE0AF\uF4E7\u0000\uEFDC\uCFFC" + //  1180 -  1189
                "\uC7F7\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF5C1\uDCC4" + //  1190 -  1199
                "\u0000\uE7BE\u0000\uFCF2\u0000\u0000\uD6B4\u0000\u0000\uFCC1" + //  1200 -  1209
                "\u0000\uEEAB\uD4A5\u0000\u0000\u0000\uF8CE\uF9F0\uE0ED\uE3B3" + //  1210 -  1219
                "\uF4B3\uC7F4\uC7F5\u0000\u0000\uC7F6\u0000\u0000\u0000\uCEFC" + //  1220 -  1229
                "\u0000\uDBC4\u0000\uF8F1\u0000\u0000\uC8BE\u0000\u0000\u0000" + //  1230 -  1239
                "\u0000\u0000\u0000\uCFD8\u0000\u0000\u0000\uCEE0\u0000\u0000" + //  1240 -  1249
                "\u0000\uDCD2\uFDEA\u0000\uD4BC\u0000\uFCEA\u0000\u0000\u0000" + //  1250 -  1259
                "\u0000\u0000\uB1E1\u0000\u0000\u0000\uEAC4\u0000\u0000\u0000" + //  1260 -  1269
                "\u0000\uEFD1\u0000\u0000\uD9B5\u0000\uA7BA\uA7BB\uA7DC\uA7DD" + //  1270 -  1279
                "\uA7DE\uA7B6\uA7B7\uA7B8\uC7EE\u0000\u0000\u0000\uC7EF\u0000" + //  1280 -  1289
                "\u0000\u0000\uE7EB\uF1D5\u0000\u0000\u0000\uF0BB\u0000\uEDEF" + //  1290 -  1299
                "\u0000\uE8A3\u0000\u0000\u0000\u0000\uCFF1\uC7E8\uC7E9\u0000" + //  1300 -  1309
                "\uC7EA\u0000\uC7EB\u0000\u0000\uC8AC\u0000\u0000\uC8AD\uC8AE" + //  1310 -  1319
                "\u0000\u0000\uC8B2\u0000\u0000\u0000\u0000\u0000\u0000\uCDC4" + //  1320 -  1329
                "\u0000\u0000\u0000\uF3DB\u0000\uDBA7\uF6B7\u0000\uCFE6\uF0F2" + //  1330 -  1339
                "\uC7E6\u0000\uC7E7\u0000\u0000\u0000\u0000\u0000\uBDBF\uBDC0" + //  1340 -  1349
                "\u0000\uBDC1\uC7E3\uC7E4\u0000\u0000\uC7E5\u0000\u0000\u0000" + //  1350 -  1359
                "\uEAE2\u0000\u0000\u0000\u0000\uD7C2\u0000\uD7E1\uFAF5\u0000" + //  1360 -  1369
                "\u0000\uD5C9\uF8AC\u0000\u0000\uFCD1\u0000\uEDB2\uF4AF\u0000" + //  1370 -  1379
                "\uFBA3\u0000\uA7C9\uA7CA\uA7CB\uA7CC\uA7CD\u0000\u0000\u0000" + //  1380 -  1389
                "\uFBC6\uCFB3\u0000\u0000\u0000\uD1A8\u0000\u0000\u0000\u0000" + //  1390 -  1399
                "\uC4BF\uC4C0\u0000\u0000\uF3D2\u0000\u0000\uEEF4\u0000\uE2D3" + //  1400 -  1409
                "\u0000\u0098\u0099\u009A\u009B\u009C\u009D\u009E\u009F\uC7E1" + //  1410 -  1419
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD3EB\u0000\uE9D4" + //  1420 -  1429
                "\uC7DF\uC7E0\u0000\u0000\u0000\u0000\u0000\u0000\uE1D9\u0000" + //  1430 -  1439
                "\u0000\uEFAB\uC7DA\u0000\u0000\u0000\uC7DB\u0000\u0000\u0000" + //  1440 -  1449
                "\uE3E0\u0000\uCAC9\uF2E9\u0000\uD5CE\u0000\uFBCF\u0000\u0000" + //  1450 -  1459
                "\u0000\u0000\u0000\u0000\uCFB7\uC7D4\uC7D5\u0000\uC7D6\u0000" + //  1460 -  1469
                "\uC7D7\u0000\u0000\uC7F3\u0000\u0000\u0000\u0000\u0000\u0000" + //  1470 -  1479
                "\uE5A2\u0000\u0000\u0000\uE9B5\u0000\uCCC9\uFAD5\u0000\u0000" + //  1480 -  1489
                "\uE1D4\uC7D2\u0000\u0000\u0000\u0000\uC7D3\u0000\u0000\uC7CE" + //  1490 -  1499
                "\u0000\u0000\u0000\u0000\u0000\u0000\uD6A3\u0000\u0000\u0000" + //  1500 -  1509
                "\uE5F9\uECEA\uDDD6\uEDC2\u0000\u0000\u0000\uF9C1\u0000\u0000" + //  1510 -  1519
                "\u0000\u0000\uC4D1\u0000\u0000\u0000\uFBF2\u0000\uDBF6\u0000" + //  1520 -  1529
                "\uDEDF\uC7CF\uC7D0\u0000\u0000\uC7D1\u0000\u0000\u0000\uF8A5" + //  1530 -  1539
                "\u0000\u0000\u0000\u0000\u0000\uE5BA\uC7C9\u0000\u0000\u0000" + //  1540 -  1549
                "\uC7CA\u0000\u0000\u0000\uF9F2\uECA5\uD0DF\u0000\uE7EA\uD0EB" + //  1550 -  1559
                "\uDCD1\uC7C2\u0000\u0000\u0000\uC7C3\u0000\u0000\u0000\uD0EA" + //  1560 -  1569
                "\u0000\u0000\u0000\u0000\u0000\u0000\uCCA9\uFDC6\u0000\uDFD2" + //  1570 -  1579
                "\u0000\uCECA\u0000\uEEDA\u0000\u0000\u0000\uD8C4\u0000\u0000" + //  1580 -  1589
                "\u0000\u0000\uD0F5\u0000\u0000\uE8ED\uD0D2\uC7BE\u0000\u0000" + //  1590 -  1599
                "\uC7BF\u0000\uC7C0\u0000\u0000\uC7B5\u0000\u0000\u0000\u0000" + //  1600 -  1609
                "\u0000\u0000\uE2B8\uEBAB\u0000\u0000\uC6F5\u0000\uC6F6\u0000" + //  1610 -  1619
                "\u0000\u0000\u0000\uE0C1\uEFDB\u0000\u0000\uF0E9\uC7BD\u0000" + //  1620 -  1629
                "\u0000\u0000\u0000\u0000\u0000\u0000\uFBFB\u0000\u0000\uC6FE" + //  1630 -  1639
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF0FC\u0000\u0000\u0000" + //  1640 -  1649
                "\uAAA1\uAAA2\uAAA3\uAAA4\uAAA5\uAAA6\uAAA7\uC7BB\u0000\u0000" + //  1650 -  1659
                "\u0000\uC7BC\u0000\u0000\u0000\uD8CC\uF9F1\u0000\uCEDF\uFAA4" + //  1660 -  1669
                "\uE6B2\u0000\uDFF9\uD7E0\u0000\u0000\u0000\u0000\u0000\u0000" + //  1670 -  1679
                "\uE9E4\u0000\u0000\uE4A2\u0000\uE2E8\u0000\uE6D0\u0000\uFBE8" + //  1680 -  1689
                "\uC7B7\u0000\u0000\u0000\uC7B8\u0000\u0000\u0000\uF5E4\u0000" + //  1690 -  1699
                "\u0000\uF3A6\uDDE0\uE1A6\u0000\uD4BA\uE4B3\u0000\uE9DA\u0000" + //  1700 -  1709
                "\uDEB6\u0000\uD9BF\uC7B0\uC7B1\u0000\uC7B2\u0000\uC7B3\u0000" + //  1710 -  1719
                "\u0000\uC6AD\u0000\u0000\u0000\u0000\u0000\u0000\uDAD8\uD1B9" + //  1720 -  1729
                "\u0000\uDFA9\uC7AE\u0000\uC7AF\u0000\u0000\u0000\u0000\u0000" + //  1730 -  1739
                "\uBDBA\uBDBB\u0000\u0000\uC5FC\u0000\u0000\u0000\u0000\u0000" + //  1740 -  1749
                "\u0000\uE0FE\u0000\uDFFB\u0000\uF7FA\u0000\u0000\u0000\uF8AB" + //  1750 -  1759
                "\u0000\u0000\u0000\uCEF5\u0000\u0000\u0000\u0000\uDDC5\u0000" + //  1760 -  1769
                "\u0000\u0000\u0000\uC1AA\uC1AB\u0000\uC1AC\uC7AA\uC7AB\u0000" + //  1770 -  1779
                "\u0000\uC7AC\u0000\u0000\uC7AD\uC7A6\u0000\u0000\u0000\uC7A7" + //  1780 -  1789
                "\u0000\u0000\u0000\uA4A1\uA4A2\uA4A3\uA4A4\uA4A5\uA4A6\uA4A7" + //  1790 -  1799
                "\uC7A3\u0000\u0000\u0000\uC7A4\u0000\u0000\u0000\uABA1\uABA2" + //  1800 -  1809
                "\uABA3\uABA4\uABA5\uABA6\uABA7\uC7A1\u0000\u0000\u0000\u0000" + //  1810 -  1819
                "\u0000\u0000\u0000\uF6E6\u0000\u0000\uC5F3\u0000\u0000\u0000" + //  1820 -  1829
                "\u0000\u0000\u0000\uF0AF\uD6BD\u0000\u0000\uC3C3\u0000\u0000" + //  1830 -  1839
                "\u0000\u0000\u0000\u0000\uCEB4\u0000\u0000\u0000\uE8AD\u0000" + //  1840 -  1849
                "\u0000\u0000\u0000\uEFAF\uC6F9\u0000\u0000\u0000\uC6FA\u0000" + //  1850 -  1859
                "\u0000\u0000\uA2B2\u0000\u0000\u0000\u0000\u0000\u0000\uE6ED" + //  1860 -  1869
                "\u0000\u0000\uC5DE\u0000\u0000\u0000\u0000\u0000\u0000\uF3C1" + //  1870 -  1879
                "\uD0AB\u0000\uD4E4\uC6F4\u0000\u0000\u0000\u0000\u0000\u0000" + //  1880 -  1889
                "\u0000\uF2C8\u0000\u0000\uC5B7\u0000\u0000\u0000\u0000\u0000" + //  1890 -  1899
                "\u0000\uD6DE\uE4F4\uE1EF\u0000\uDFD1\u0000\u0000\u0000\u0000" + //  1900 -  1909
                "\u0000\uEDED\uF8B8\uC6F3\u0000\u0000\u0000\u0000\u0000\u0000" + //  1910 -  1919
                "\u0000\uE6BF\u0000\u0000\uC5AF\u0000\u0000\u0000\u0000\u0000" + //  1920 -  1929
                "\u0000\uDDDD\u0000\u0000\u0000\uF3BB\u0000\uE5E1\u0000\u0000" + //  1930 -  1939
                "\u0000\uFBED\u0000\uE0AD\u0000\u0000\uEAEE\uC6F1\uC6F2\u0000" + //  1940 -  1949
                "\u0000\u0000\u0000\u0000\u0000\uD7E6\u0000\u0000\u0000\uA2E5" + //  1950 -  1959
                "\uA2E2\u0000\u0000\u0000\uA7D9\u0000\uEABA\u0000\uEAD3\u0000" + //  1960 -  1969
                "\u0000\uEDC9\uDDAB\u0000\uA8C9\uA8CA\uA8CB\uA8CC\u0000\u0000" + //  1970 -  1979
                "\u0000\uA2DE\uC6ED\u0000\u0000\u0000\uC6EE\u0000\u0000\u0000" + //  1980 -  1989
                "\uA2B5\u0000\u0000\u0000\u0000\u0000\u0000\uF3A3\u0000\uD3EC" + //  1990 -  1999
                "\uC6E8\uC6E9\u0000\uC6EA\u0000\uC6EB\u0000\u0000\uC5A4\u0000" + //  2000 -  2009
                "\u0000\u0000\u0000\u0000\u0000\uD7FE\u0000\u0000\u0000\uD1CB" + //  2010 -  2019
                "\uD6E4\u0000\u0000\u0000\uD5F2\uC6E7\u0000\u0000\u0000\u0000" + //  2020 -  2029
                "\u0000\u0000\u0000\uE1F5\uF1B3\u0000\uD6D3\u0000\u0000\u0000" + //  2030 -  2039
                "\u0000\u0000\u0000\u0000\uDBFC\uC6E4\uC6E5\u0000\u0000\uC6E6" + //  2040 -  2049
                "\u0000\u0000\u0000\uA9FB\uA9FC\uA9FD\uA9FE\u0000\u0000\u0000" + //  2050 -  2059
                "\uF2CE\uDBB4\u0000\u0000\u0000\uF7B4\u0000\u0000\u0000\u0000" + //  2060 -  2069
                "\uC3BC\uC3BD\u0000\u0000\uC4F8\u0000\u0000\u0000\u0000\u0000" + //  2070 -  2079
                "\u0000\uE2EB\uD6FC\u0000\u0000\uBFEB\u0000\u0000\u0000\u0000" + //  2080 -  2089
                "\u0000\u0000\uDEB0\u0000\u0000\u0000\uF2E0\uF1C9\u0000\u0000" + //  2090 -  2099
                "\u0000\u0000\uF4EF\u0000\u0000\u0000\uF6CE\uC6E2\uC6E3\u0000" + //  2100 -  2109
                "\u0000\u0000\u0000\u0000\u0000\uF1C3\uEEDF\u0000\u0000\uC4E9" + //  2110 -  2119
                "\u0000\u0000\u0000\u0000\u0000\u0000\uEEDC\u0000\uCBCB\uFCD5" + //  2120 -  2129
                "\uC6DD\u0000\u0000\u0000\uC6DE\u0000\u0000\u0000\uACD7\u0000" + //  2130 -  2139
                "\u0000\u0000\u0000\u0000\u0000\uE4DD\uDFEE\uCBAC\uC6D4\uC6D5" + //  2140 -  2149
                "\u0000\uC6D6\uC6D7\uC6D8\u0000\u0000\uC4E1\u0000\u0000\u0000" + //  2150 -  2159
                "\u0000\u0000\u0000\uE4DF\u0000\uCAD6\u0000\uFAD2\u0000\u0000" + //  2160 -  2169
                "\u0000\u0000\u0000\uF8EF\u0000\uA8C1\uA8C2\uA8C3\uA8C4\uA8C5" + //  2170 -  2179
                "\uA8C6\uA8C7\uA8C8\uC6D3\u0000\u0000\u0000\u0000\u0000\u0000" + //  2180 -  2189
                "\u0000\uFDAB\u0000\uDFE0\uC6D0\uC6D1\u0000\u0000\uC6D2\u0000" + //  2190 -  2199
                "\u0000\u0000\uACA7\u0000\u0000\u0000\u0000\u0000\u0000\uEEBC" + //  2200 -  2209
                "\u0000\uEFC1\uC6CD\uC6CE\u0000\u0000\u0000\uC6CF\u0000\u0000" + //  2210 -  2219
                "\uC4BE\u0000\u0000\u0000\u0000\u0000\u0000\uF4BD\u0000\uCFB8" + //  2220 -  2229
                "\uE9DB\uC6C7\u0000\u0000\u0000\uC6C8\u0000\uC6C9\u0000\uD7F2" + //  2230 -  2239
                "\u0000\uE1C0\u0000\uDBE2\uE6D8\u0000\u0000\uE2D7\u0000\u0000" + //  2240 -  2249
                "\u0000\u0000\u0000\u0000\uE1A3\uD2E0\u0000\uA8B9\uA8BA\uA8BB" + //  2250 -  2259
                "\uA8BC\uA8BD\uA8BE\uA8BF\uA8C0\uC6C0\uC6C1\u0000\uC6C2\u0000" + //  2260 -  2269
                "\uC6C3\u0000\u0000\uC4B2\u0000\u0000\u0000\u0000\u0000\u0000" + //  2270 -  2279
                "\uFADC\u0000\uEDB5\uE1E3\uC6BF\u0000\u0000\u0000\u0000\u0000" + //  2280 -  2289
                "\u0000\u0000\uE6F4\u0000\u0000\uC3F6\u0000\u0000\u0000\u0000" + //  2290 -  2299
                "\u0000\u0000\uCCB3\u0000\u0000\uDBF3\uC6BC\uC6BD\u0000\u0000" + //  2300 -  2309
                "\uC6BE\u0000\u0000\u0000\uA5E1\uA5E2\uA5E3\uA5E4\uA5E5\uA5E6" + //  2310 -  2319
                "\uA5E7\uC6B8\u0000\u0000\u0000\uC6B9\u0000\u0000\u0000\uA5C1" + //  2320 -  2329
                "\uA5C2\uA5C3\uA5C4\uA5C5\uA5C6\uA5C7\uC6B4\uC6B5\u0000\uC6B6" + //  2330 -  2339
                "\u0000\u0000\u0000\u0000\uEDC5\uF3D6\u0000\u0000\uDED9\uC6B2" + //  2340 -  2349
                "\u0000\uC6B3\u0000\u0000\u0000\u0000\u0000\uBDB0\uBDB1\u0000" + //  2350 -  2359
                "\uBDB2\uC6AE\uC6AF\u0000\u0000\uC6B0\u0000\u0000\uC6B1\uC6AA" + //  2360 -  2369
                "\u0000\u0000\u0000\uC6AB\u0000\u0000\u0000\uA9B0\uA8AF\uA9AF" + //  2370 -  2379
                "\u0000\u0000\u0000\u0000\uBDDC\uBDDD\u0000\u0000\uC3E6\u0000" + //  2380 -  2389
                "\u0000\u0000\u0000\u0000\u0000\uE8A2\u0000\u0000\u0000\uEAC2" + //  2390 -  2399
                "\uF2E6\uF0B6\u0000\u0000\u0000\uD8B5\u0000\u0000\u0000\uE4DC" + //  2400 -  2409
                "\u0000\uD4F3\uD4C9\u0000\u0000\u0000\u0000\uD6FA\u0000\uA8B1" + //  2410 -  2419
                "\uA8B2\uA8B3\uA8B4\uA8B5\uA8B6\uA8B7\uA8B8\uC6A6\uC6A7\u0000" + //  2420 -  2429
                "\u0000\u0000\uC6A8\u0000\u0000\uC3DC\u0000\u0000\u0000\u0000" + //  2430 -  2439
                "\u0000\u0000\uE3D8\u0000\u0000\u0000\uF8A1\u0000\u0000\u0000" + //  2440 -  2449
                "\uE8D6\u0000\uE7BC\u0000\u0000\u0000\u0000\u0000\uD1EE\u0000" + //  2450 -  2459
                "\uA9C9\uA9CA\uA9CB\uA9CC\uA2DF\u0000\u0000\u0000\uD9EC\u0000" + //  2460 -  2469
                "\uD9BD\u0000\uD8DF\uC6A5\u0000\u0000\u0000\u0000\u0000\u0000" + //  2470 -  2479
                "\u0000\uF1ED\u0000\uCCCF\uC6A2\uC6A3\u0000\u0000\uC6A4\u0000" + //  2480 -  2489
                "\u0000\u0000\uA9A5\uA8A6\uA9A6\u0000\u0000\u0000\u0000\uBDC3" + //  2490 -  2499
                "\uBDC4\u0000\u0000\uC3D5\u0000\u0000\u0000\u0000\u0000\u0000" + //  2500 -  2509
                "\uD8E0\uFCBA\uFDAF\uF0CE\uC5FD\u0000\u0000\u0000\u0000\u0000" + //  2510 -  2519
                "\u0000\u0000\uF8D0\u0000\u0000\uC3C9\u0000\u0000\u0000\u0000" + //  2520 -  2529
                "\u0000\u0000\uFDB6\u0000\u0000\u0000\uDCF2\u0000\u0000\u0000" + //  2530 -  2539
                "\u0000\u0000\uB8B2\uB8B3\u0000\uB8B4\uC5F7\u0000\u0000\u0000" + //  2540 -  2549
                "\uC5F8\u0000\u0000\u0000\uA9A2\u0000\u0000\u0000\u0000\u0000" + //  2550 -  2559
                "\u0000\uE1DE\uCBEE\u0000\uE6FB\u0000\u0000\u0000\u0000\u0000" + //  2560 -  2569
                "\uE6D4\u0000\uA9C1\uA9C2\uA9C3\uA9C4\uA9C5\uA9C6\uA9C7\uA9C8" + //  2570 -  2579
                "\uC5F4\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF8F2\u0000" + //  2580 -  2589
                "\uF4F9\uC5F1\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF4F8" + //  2590 -  2599
                "\u0000\u0000\uC2F4\uC2F5\u0000\u0000\u0000\u0000\uC2F6\uC5EF" + //  2600 -  2609
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD6A9\u0000\u0000" + //  2610 -  2619
                "\uC2ED\u0000\u0000\u0000\u0000\u0000\u0000\uD7B6\uCFB5\u0000" + //  2620 -  2629
                "\uD9A8\uC5EE\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF3D0" + //  2630 -  2639
                "\u0000\u0000\uC2DD\u0000\u0000\u0000\u0000\u0000\u0000\uE2A9" + //  2640 -  2649
                "\u0000\u0000\uDEBC\uC5E8\uC5E9\u0000\uC5EA\u0000\uC5EB\u0000" + //  2650 -  2659
                "\u0000\uC2D0\u0000\uC2D1\u0000\u0000\u0000\u0000\uD5FD\u0000" + //  2660 -  2669
                "\u0000\u0000\u0000\uC2EB\u0000\u0000\uC2EC\uC5E7\u0000\u0000" + //  2670 -  2679
                "\u0000\u0000\u0000\u0000\u0000\uE8E9\u0000\uE3AC\uC5E4\uC5E5" + //  2680 -  2689
                "\u0000\u0000\uC5E6\u0000\u0000\u0000\uA2AE\u0000\u0000\uA2B4" + //  2690 -  2699
                "\u0000\u0000\uA1D7\uC5E3\u0000\u0000\u0000\u0000\u0000\u0000" + //  2700 -  2709
                "\u0000\uD0C1\u0000\u0000\uC2BB\u0000\u0000\u0000\u0000\u0000" + //  2710 -  2719
                "\u0000\uCAF4\u0000\u0000\u0000\uEEE8\uDADE\u0000\uF2F7\u0000" + //  2720 -  2729
                "\u0000\uB9D8\u0000\u0000\uB9D9\uB9DA\uB9DB\uB9DC\uC5DF\u0000" + //  2730 -  2739
                "\u0000\u0000\uC5E0\u0000\u0000\u0000\uC5EC\u0000\uC5ED\u0000" + //  2740 -  2749
                "\u0000\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007\uC5D9" + //  2750 -  2759
                "\u0000\u0000\u0000\uC5DA\u0000\u0000\u0000\uC0BC\u0000\uC0BD" + //  2760 -  2769
                "\uC0BE\u0000\uC0BF\uC5D2\uC5D3\u0000\uC5D4\uC5D5\uC5D6\u0000" + //  2770 -  2779
                "\u0000\uC2A3\uC2A4\u0000\uC2A5\uC2A6\u0000\u0000\uB9A5\u0000" + //  2780 -  2789
                "\u0000\u0000\u0000\u0000\u0000\uF3DE\u0000\u0000\u0000\uDEE3" + //  2790 -  2799
                "\u0000\u0000\u0000\u0000\u0000\uB2C4\uB2C5\u0000\uB2C6\uC5D0" + //  2800 -  2809
                "\u0000\uC5D1\u0000\u0000\u0000\u0000\u0000\uBDAC\uBDAD\u0000" + //  2810 -  2819
                "\u0000\uC1DF\u0000\u0000\u0000\u0000\u0000\u0000\uD1A5\uDCB8" + //  2820 -  2829
                "\u0000\u0000\uB7F8\u0000\u0000\u0000\u0000\u0000\u0000\uF6F3" + //  2830 -  2839
                "\u0000\u0000\u0000\uF4E1\uF9B5\u0000\u0000\uE1C3\uE1C2\uC5CD" + //  2840 -  2849
                "\uC5CE\u0000\u0000\uC5CF\u0000\u0000\u0000\uB4D8\u0000\uB4D9" + //  2850 -  2859
                "\uB4DA\uB4DB\u0000\uFBF4\uD5A7\u0000\u0000\u0000\uF1F6\u0000" + //  2860 -  2869
                "\uE6D3\uC5CB\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDBA8" + //  2870 -  2879
                "\u0000\u0000\uC1D1\u0000\u0000\u0000\u0000\u0000\u0000\uDCB7" + //  2880 -  2889
                "\u0000\u0000\u0000\uEBB3\u0000\uF0B4\u0000\u0000\uCBF4\uC5C9" + //  2890 -  2899
                "\uC5CA\u0000\u0000\u0000\u0000\u0000\u0000\uECF9\u0000\u0000" + //  2900 -  2909
                "\uF8AD\uC5C4\u0000\u0000\u0000\uC5C5\u0000\u0000\u0000\uB4CB" + //  2910 -  2919
                "\u0000\uB4CC\u0000\u0000\u0000\uCFED\u0000\uEDEB\u0000\u0000" + //  2920 -  2929
                "\u0000\uF0E5\u0000\u0000\u0000\uF4C4\uC5BD\uC5BE\u0000\uC5BF" + //  2930 -  2939
                "\uC5C0\uC5C1\u0000\u0000\uC1C7\u0000\u0000\u0000\u0000\u0000" + //  2940 -  2949
                "\u0000\uD4E1\u0000\uD1A3\uE1D6\uC5BB\uC5BC\u0000\u0000\u0000" + //  2950 -  2959
                "\u0000\u0000\u0000\uF6DF\u0000\uF2DA\uE4EB\uC5B8\uC5B9\u0000" + //  2960 -  2969
                "\u0000\uC5BA\u0000\u0000\u0000\uB3F4\uB3F5\uB3F6\u0000\u0000" + //  2970 -  2979
                "\u0000\uDFF8\u0000\u0000\u0000\u0000\u0000\uECC7\u0000\u0000" + //  2980 -  2989
                "\u0000\uEDFC\u0000\u0000\uE1DB\u0000\uD6AD\u0000\u0000\uFDCE" + //  2990 -  2999
                "\u0000\u0000\u0000\uE8A1\uC5B2\u0000\u0000\u0000\uC5B3\u0000" + //  3000 -  3009
                "\u0000\u0000\uB1ED\u0000\uB1EE\uB1EF\uB1F0\u0000\uEDC8\uEFC3" + //  3010 -  3019
                "\u0000\u0000\u0000\u0000\u0000\u0000\uCDC9\uF9B7\u0000\uA9B9" + //  3020 -  3029
                "\uA9BA\uA9BB\uA9BC\uA9BD\uA9BE\uA9BF\uA9C0\uC5AB\u0000\u0000" + //  3030 -  3039
                "\u0000\uC5AC\u0000\u0000\u0000\uF6CD\u0000\u0000\u0000\u0000" + //  3040 -  3049
                "\u0000\uBADF\uBAE0\u0000\u0000\uC1AD\u0000\u0000\u0000\u0000" + //  3050 -  3059
                "\u0000\u0000\uDEFD\uF2F9\u0000\uD5C7\uC5A8\u0000\u0000\u0000" + //  3060 -  3069
                "\u0000\u0000\u0000\u0000\uCBED\u0000\u0000\uC0C0\uC0C1\uC0C2" + //  3070 -  3079
                "\uC0C3\uC0C4\uC0C5\uC0C6\uC5A7\u0000\u0000\u0000\u0000\u0000" + //  3080 -  3089
                "\u0000\u0000\uFAA5\u0000\u0000\uC0AE\u0000\u0000\u0000\u0000" + //  3090 -  3099
                "\u0000\u0000\uFDBE\u0000\u0000\uCAAC\uC5A5\u0000\u0000\u0000" + //  3100 -  3109
                "\uC5A6\u0000\u0000\u0000\uF0BA\uEEB1\u0000\u0000\uEEB2\u0000" + //  3110 -  3119
                "\uFDDC\uEDB3\uCEC9\u0000\u0000\u0000\u0000\u0000\uB1AF\uB1B0" + //  3120 -  3129
                "\u0000\uB1B1\uC4FD\u0000\u0000\u0000\uC4FE\u0000\u0000\u0000" + //  3130 -  3139
                "\uE5A4\u0000\u0000\u0000\uD5B6\u0000\uCFB6\u0000\u0000\u0000" + //  3140 -  3149
                "\uEDC7\uEEAC\u0000\u0000\uE0DD\uFBF3\u0000\u0000\u0000\u0000" + //  3150 -  3159
                "\u0000\uB0B7\uB0B8\u0000\uB0B9\uC4F9\u0000\u0000\u0000\u0000" + //  3160 -  3169
                "\u0000\u0000\u0000\uF6D5\uD5E2\u0000\uFCE9\u0000\u0000\u0000" + //  3170 -  3179
                "\u0000\u0000\u0000\u0000\uD9A7\uC4F6\u0000\u0000\u0000\uC4F7" + //  3180 -  3189
                "\u0000\u0000\u0000\uF6AF\u0000\u0000\u0000\u0000\u0000\uBACC" + //  3190 -  3199
                "\u0000\u0000\u0000\uD2C1\u0000\u0000\u0000\uF8D7\u0000\uF7A8" + //  3200 -  3209
                "\u0000\u0000\u0000\u0000\uFBCE\u0000\u0000\uE2AA\u0000\uD5A6" + //  3210 -  3219
                "\u0000\u0000\uD4D7\u0000\uA9B1\uA9B2\uA9B3\uA9B4\uA9B5\uA9B6" + //  3220 -  3229
                "\uA9B7\uA9B8\uC4F1\uC4F2\u0000\uC4F3\u0000\uC4F4\u0000\u0000" + //  3230 -  3239
                "\uBFDB\u0000\u0000\u0000\u0000\u0000\u0000\uCCD7\uE5C2\u0000" + //  3240 -  3249
                "\u0000\uB3B9\u0000\uB3BA\uB3BB\uB3BC\u0000\u0000\uD4AD\uF6D1" + //  3250 -  3259
                "\u0000\u0000\u0000\u0000\u0000\uB6C6\u0000\u0000\u0000\uE7F0" + //  3260 -  3269
                "\u0000\uD0EE\u0000\u0000\uF3AA\uC4F0\u0000\u0000\u0000\u0000" + //  3270 -  3279
                "\u0000\u0000\u0000\uF1E9\u0000\u0000\uBFCB\u0000\uBFCC\u0000" + //  3280 -  3289
                "\u0000\u0000\u0000\uE1B7\u0000\u0000\u0000\u0000\uC2D5\u0000" + //  3290 -  3299
                "\u0000\u0000\uD0CC\u0000\u0000\u0000\u0000\u0000\uB6CC\u0000" + //  3300 -  3309
                "\u0000\u0000\uF0BF\u0000\uF6A4\u0000\uE3B6\u0000\uFBA5\uE1EE" + //  3310 -  3319
                "\u0000\u0000\u0000\u0000\u0000\u0000\uCBCE\uFBD8\u0000\uA4F8" + //  3320 -  3329
                "\uA4F9\uA4FA\uA4FB\uA4FC\uA4FD\uA4FE\u0000\u0090\u0091\u0092" + //  3330 -  3339
                "\u0093\u0094\u0095\u0096\u0097\uC4ED\uC4EE\u0000\u0000\uC4EF" + //  3340 -  3349
                "\u0000\u0000\u0000\uEDCE\u0000\u0000\u0000\u0000\u0000\uBAC4" + //  3350 -  3359
                "\u0000\u0000\u0000\uFCE0\uD7C8\uFDAD\u0000\u0000\u0000\uD4E5" + //  3360 -  3369
                "\u0000\u0000\u0000\uF9C3\uC4EB\u0000\u0000\u0000\u0000\u0000" + //  3370 -  3379
                "\u0000\u0000\uA1DA\uA1D9\u0000\uD6EE\u0000\u0000\u0000\u0000" + //  3380 -  3389
                "\u0000\u0000\uE7BB\uC4EA\u0000\u0000\u0000\u0000\u0000\u0000" + //  3390 -  3399
                "\u0000\uA1D1\u0000\u0000\uBEED\u0000\u0000\u0000\u0000\u0000" + //  3400 -  3409
                "\u0000\uEBF4\u0000\u0000\u0000\uD3B2\u0000\u0000\u0000\uE2C0" + //  3410 -  3419
                "\uF2DF\uC4E6\u0000\u0000\u0000\u0000\uC4E7\u0000\u0000\uBEE7" + //  3420 -  3429
                "\u0000\u0000\u0000\uBEE8\u0000\uBEE9\uC4E5\u0000\u0000\u0000" + //  3430 -  3439
                "\u0000\u0000\u0000\u0000\uA1EF\u0000\u0000\uBEB1\u0000\u0000" + //  3440 -  3449
                "\u0000\u0000\u0000\u0000\uDAE6\uF7B3\u0000\u0000\uB1A8\u0000" + //  3450 -  3459
                "\u0000\uB1A9\uB1AA\u0000\u0000\uFBF6\u0000\u0000\u0000\u0000" + //  3460 -  3469
                "\u0000\u0000\uDEFA\u0000\uD7F8\uC4E2\uC4E3\u0000\u0000\uC4E4" + //  3470 -  3479
                "\u0000\u0000\u0000\uE8A7\u0000\u0000\u0000\u0000\u0000\uBAC3" + //  3480 -  3489
                "\u0000\u0000\u0000\uFBA1\u0000\u0000\u0000\uE5E9\uE9EE\uC4DC" + //  3490 -  3499
                "\u0000\u0000\u0000\uC4DD\u0000\u0000\u0000\uE5E0\u0000\u0000" + //  3500 -  3509
                "\u0000\u0000\u0000\uBABD\uBABE\u0000\uBABF\uC4D9\u0000\u0000" + //  3510 -  3519
                "\u0000\u0000\u0000\u0000\u0000\uE4EC\u0000\uD6D2\u0000\uF9D5" + //  3520 -  3529
                "\uE7BA\uEBD5\uD5F7\uEFE7\uE1BE\uC4D7\uC4D8\u0000\u0000\u0000" + //  3530 -  3539
                "\u0000\u0000\u0000\uCDA3\u0000\u0000\uE8E6\uC4D2\u0000\u0000" + //  3540 -  3549
                "\u0000\uC4D3\u0000\u0000\u0000\uF7E3\u0000\u0000\u0000\u0000" + //  3550 -  3559
                "\u0000\uBAB8\uBAB9\uBABA\u0000\uFBCC\uEBA1\u0000\u0000\uD4A6" + //  3560 -  3569
                "\u0000\u0000\u0000\uEBE5\u0000\u0000\uE1D2\u0000\uA4F0\uA4F1" + //  3570 -  3579
                "\uA4F2\uA4F3\uA4F4\uA4F5\uA4F6\uA4F7\uC4CD\uC4CE\u0000\uC4CF" + //  3580 -  3589
                "\u0000\uC4D0\u0000\u0000\uBEAB\u0000\u0000\u0000\u0000\u0000" + //  3590 -  3599
                "\u0000\uE2A4\u0000\u0000\u0000\uDAA5\u0000\uDBBD\u0000\u0000" + //  3600 -  3609
                "\u0000\uF9DC\u0000\u0000\uF3DC\u0000\u0000\uB1A4\u0000\u0000" + //  3610 -  3619
                "\u0000\u0000\u0000\u0000\uEDA8\uDEC2\uF6E2\uEDDC\uC4CC\u0000" + //  3620 -  3629
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF4ED\u0000\uCCAD\uF6FA" + //  3630 -  3639
                "\uD6B2\uD2D8\u0000\u0000\u0000\u0000\uC4FB\uC4FC\u0000\u0000" + //  3640 -  3649
                "\uF8DC\uF7EE\uEBE8\u0000\uD2FA\u0000\u0000\uFCB6\uF2AD\uEFE1" + //  3650 -  3659
                "\uF3AE\uDCC6\uD9EB\u0000\uA4E8\uA4E9\uA4EA\uA4EB\uA4EC\uA4ED" + //  3660 -  3669
                "\uA4EE\uA4EF\uC4C9\uC4CA\u0000\u0000\uC4CB\u0000\u0000\u0000" + //  3670 -  3679
                "\uE2D1\u0000\u0000\u0000\u0000\u0000\uBAAD\uBAAE\u0000\u0000" + //  3680 -  3689
                "\uBDCC\u0000\u0000\u0000\u0000\uBDCD\u0000\uE9FA\u0000\u0000" + //  3690 -  3699
                "\u0000\uFBCB\u0000\u0000\uCAD5\uC4C7\uC4C8\u0000\u0000\u0000" + //  3700 -  3709
                "\u0000\u0000\u0000\uE7C2\u0000\u0000\u0000\uD5F0\u0000\u0000" + //  3710 -  3719
                "\uD1CA\u0000\u0000\uBDC2\u0000\u0000\u0000\u0000\u0000\u0000" + //  3720 -  3729
                "\uEAB5\u0000\uE5AA\uDFBA\uC4C1\u0000\u0000\uC4C2\uC4C3\u0000" + //  3730 -  3739
                "\u0000\u0000\uCCF3\uE6BE\u0000\u0000\u0000\uF6AE\uC4B7\uC4B8" + //  3740 -  3749
                "\u0000\uC4B9\uC4BA\uC4BB\u0000\u0000\uBDB3\u0000\u0000\u0000" + //  3750 -  3759
                "\u0000\u0000\u0000\uDDC9\u0000\u0000\uD4D3\uC4B6\u0000\u0000" + //  3760 -  3769
                "\u0000\u0000\u0000\u0000\u0000\uEFC6\u0000\uF4CD\u0000\u0000" + //  3770 -  3779
                "\u0000\u0000\uF1BF\uF8B1\u0000\uA4E0\uA4E1\uA4E2\uA4E3\uA4E4" + //  3780 -  3789
                "\uA4E5\uA4E6\uA4E7\uC4B3\uC4B4\u0000\u0000\uC4B5\u0000\u0000" + //  3790 -  3799
                "\u0000\uD8CB\u0000\u0000\u0000\u0000\u0000\uB9E3\uB9E4\u0000" + //  3800 -  3809
                "\uB9E5\uC4AD\u0000\u0000\u0000\uC4AE\u0000\u0000\u0000\uD5FA" + //  3810 -  3819
                "\u0000\uE4F7\u0000\u0000\u0000\uDFCB\u0000\u0000\u0000\u0000" + //  3820 -  3829
                "\u0000\uFBE0\uF2E5\u0000\u0000\uBDA2\uBDA3\u0000\uBDA4\u0000" + //  3830 -  3839
                "\u0000\u0000\uFAF8\u0000\u0000\u0000\u0000\u0000\uB4BC\u0000" + //  3840 -  3849
                "\u0000\u0000\uEAE3\u0000\u0000\u0000\u0000\u0000\uE0D1\u0000" + //  3850 -  3859
                "\u0000\uE9A8\uC4A7\uC4A8\u0000\uC4A9\u0000\uC4AA\u0000\u0000" + //  3860 -  3869
                "\uBCF5\u0000\u0000\u0000\u0000\u0000\u0000\uE3E5\u0000\uCBC5" + //  3870 -  3879
                "\uEAB4\uC4A5\uC4A6\u0000\u0000\u0000\u0000\u0000\u0000\uE8AB" + //  3880 -  3889
                "\uDEE2\u0000\u0000\uBCDB\u0000\u0000\u0000\uBCDC\u0000\u0000" + //  3890 -  3899
                "\uDCAE\u0000\u0000\u0000\u0000\u0000\u0000\uA1C5\uA1F1\u0000" + //  3900 -  3909
                "\u0000\uD3C6\u0000\uDBE6\u0000\u0000\u0000\u0000\uC8E2\u0000" + //  3910 -  3919
                "\u0000\uC8E3\uC4A1\uC4A2\u0000\u0000\uC4A3\u0000\u0000\uC4A4" + //  3920 -  3929
                "\uC3FB\uC3FC\u0000\uC3FD\u0000\uC3FE\u0000\u0000\uBCA7\u0000" + //  3930 -  3939
                "\u0000\u0000\u0000\u0000\u0000\uD6BB\uDED6\u0000\u0000\uEAA6" + //  3940 -  3949
                "\u0000\u0000\u0000\u0000\u0000\u0000\uC8A2\u0000\u0000\uDEEE" + //  3950 -  3959
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF0E8\u0000\uDDC0\uC3FA" + //  3960 -  3969
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDBA5\u0000\uCFF7" + //  3970 -  3979
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCDD9\uC3F7\uC3F8" + //  3980 -  3989
                "\u0000\u0000\uC3F9\u0000\u0000\u0000\uD7EE\uD1F1\u0000\u0000" + //  3990 -  3999
                "\u0000\u0000\uD0F0\u0000\u0000\u0000\u0000\uC0CC\uC0CD\u0000" + //  4000 -  4009
                "\u0000\uBBD9\u0000\u0000\u0000\u0000\u0000\u0000\uD5E6\u0000" + //  4010 -  4019
                "\u0000\u0000\uFDFD\u0000\u0000\u0000\u0000\u0000\uB3FA\u0000" + //  4020 -  4029
                "\u0000\u0000\uECD1\u0000\u0000\u0000\u0000\u0000\uFBDD\u0000" + //  4030 -  4039
                "\uEFCA\u0000\uDAEB\u0000\uE2D8\uEDD6\u0000\u0000\uD6D1\uE0B3" + //  4040 -  4049
                "\uC3F3\u0000\u0000\u0000\uC3F4\u0000\u0000\u0000\uD3E9\uE2C9" + //  4050 -  4059
                "\u0000\uFCDB\uCDAD\u0000\uEFC2\u0000\u0000\u0000\u0000\u0000" + //  4060 -  4069
                "\u0000\uEDEC\uC3EE\uC3EF\u0000\uC3F0\u0000\uC3F1\u0000\u0000" + //  4070 -  4079
                "\uBBD7\u0000\u0000\u0000\u0000\u0000\u0000\uEDE4\u0000\u0000" + //  4080 -  4089
                "\uDDE7\uC3ED\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE2CF" + //  4090 -  4099
                "\u0000\uD8AE\u0000\uF9D3\uD5FE\u0000\u0000\u0000\u0000\uC4EC" + //  4100 -  4109
                "\u0000\u0000\u0000\uD7F6\u0000\u0000\u0000\uE2CA\uC3EB\u0000" + //  4110 -  4119
                "\u0000\u0000\uC3EC\u0000\u0000\u0000\uDBD6\u0000\u0000\u0000" + //  4120 -  4129
                "\u0000\u0000\uB9B3\uB9B4\u0000\uB9B5\uC3EA\u0000\u0000\u0000" + //  4130 -  4139
                "\u0000\u0000\u0000\u0000\uCEA8\u0000\uFBCA\u0000\u0000\u0000" + //  4140 -  4149
                "\u0000\u0000\uCDE3\uD8BB\uC3E7\u0000\u0000\u0000\u0000\u0000" + //  4150 -  4159
                "\u0000\u0000\uF4C7\u0000\uEAB9\u0000\u0000\u0000\u0000\u0000" + //  4160 -  4169
                "\uF1DE\u0000\uA4D8\uA4D9\uA4DA\uA4DB\uA4DC\uA4DD\uA4DE\uA4DF" + //  4170 -  4179
                "\uC3E1\u0000\u0000\u0000\uC3E2\u0000\u0000\u0000\uDBC1\u0000" + //  4180 -  4189
                "\u0000\u0000\u0000\u0000\uB9AB\uB9AC\uB9AD\u0000\uDFCF\u0000" + //  4190 -  4199
                "\u0000\uD3C0\uE3D7\u0000\uEFE6\uFCD0\uC3DE\u0000\u0000\u0000" + //  4200 -  4209
                "\u0000\u0000\u0000\u0000\uF3A1\u0000\uEAFD\u0000\uD9DD\u0000" + //  4210 -  4219
                "\uDAB4\uEEAA\uFBE9\u0000\uA4D0\uA4D1\uA4D2\uA4D3\uA4D4\uA4D5" + //  4220 -  4229
                "\uA4D6\uA4D7\uC3DD\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  4230 -  4239
                "\uE9CB\u0000\uCBE2\uD4A4\u0000\uDEE0\uDAFD\uE4C6\uE8BE\u0000" + //  4240 -  4249
                "\uA4C8\uA4C9\uA4CA\uA4CB\uA4CC\uA4CD\uA4CE\uA4CF\uC3D7\u0000" + //  4250 -  4259
                "\u0000\u0000\uC3D8\u0000\u0000\u0000\uFCC4\u0000\u0000\u0000" + //  4260 -  4269
                "\u0000\u0000\uB8FA\u0000\u0000\u0000\uDEAE\u0000\u0000\u0000" + //  4270 -  4279
                "\u0000\u0000\uE4C8\u0000\u0000\u0000\uCCE8\u0000\u0000\u0000" + //  4280 -  4289
                "\u0000\uC7C1\u0000\u0000\u0000\uF6C6\u0000\u0000\u0000\u0000" + //  4290 -  4299
                "\uC7CB\uC7CC\u0000\uC7CD\uC3D3\u0000\u0000\u0000\uC3D4\u0000" + //  4300 -  4309
                "\u0000\u0000\uF5E0\u0000\u0000\u0000\u0000\u0000\uB8E2\uB8E3" + //  4310 -  4319
                "\u0000\uB8E4\uC3CE\uC3CF\u0000\uC3D0\u0000\uC3D1\u0000\u0000" + //  4320 -  4329
                "\uBBBF\u0000\u0000\u0000\u0000\u0000\u0000\uFDF1\u0000\u0000" + //  4330 -  4339
                "\u0000\uF1F9\u0000\uF2C4\uE0CB\u0000\u0000\uFAC5\u0000\u0000" + //  4340 -  4349
                "\u0000\uF9B8\u0000\u0000\uF5D7\u0000\u0000\uD8BF\u0000\u0000" + //  4350 -  4359
                "\u0000\uD1D3\u0000\uE5F0\u0000\u0000\u0000\uE0DE\uF6B4\uEAD2" + //  4360 -  4369
                "\u0000\uF9FB\uC3CD\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  4370 -  4379
                "\uEBB9\u0000\uF4E8\uE5F4\uF4BC\uF4D5\u0000\u0000\u0000\u0000" + //  4380 -  4389
                "\uC4DE\uC4DF\u0000\uC4E0\uC3CA\uC3CB\u0000\u0000\uC3CC\u0000" + //  4390 -  4399
                "\u0000\u0000\uF8E8\u0000\u0000\u0000\u0000\u0000\uB8C9\uB8CA" + //  4400 -  4409
                "\u0000\uB8CB\uC3C8\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  4410 -  4419
                "\uD0B7\u0000\uEFE4\u0000\uD7C5\uEBE2\u0000\u0000\uFCE7\u0000" + //  4420 -  4429
                "\uA4C0\uA4C1\uA4C2\uA4C3\uA4C4\uA4C5\uA4C6\uA4C7\uC3C4\u0000" + //  4430 -  4439
                "\u0000\u0000\uC3C5\u0000\u0000\u0000\uDBA4\u0000\uCFC9\uE2FC" + //  4440 -  4449
                "\uEFFA\u0000\uD4A3\uF0F8\uD7A8\u0000\u0000\u0000\uE1E7\u0000" + //  4450 -  4459
                "\uA4B8\uA4B9\uA4BA\uA4BB\uA4BC\uA4BD\uA4BE\uA4BF\uC3BE\u0000" + //  4460 -  4469
                "\u0000\u0000\uC3BF\u0000\u0000\u0000\uE9E9\u0000\u0000\u0000" + //  4470 -  4479
                "\u0000\u0000\uB8AE\uB8AF\u0000\u0000\uBAF9\uBAFA\uBAFB\u0000" + //  4480 -  4489
                "\u0000\u0000\u0000\uF9A9\u0000\u0000\u0000\u0000\uC2CD\uC2CE" + //  4490 -  4499
                "\u0000\uC2CF\uC3B7\uC3B8\u0000\uC3B9\uC3BA\uC3BB\u0000\u0000" + //  4500 -  4509
                "\uBAE3\u0000\u0000\u0000\u0000\u0000\u0000\uDDA3\u0000\u0000" + //  4510 -  4519
                "\u0000\uF3C2\u0000\u0000\u0000\u0000\u0000\uB2E4\u0000\u0000" + //  4520 -  4529
                "\u0000\uA1D2\u0000\u0000\u0000\u0000\u0000\uDADD\u0000\u0000" + //  4530 -  4539
                "\uDAB9\uC3B6\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD5CD" + //  4540 -  4549
                "\u0000\uEFE3\u0000\u0000\uCFEE\uF6BE\uE0B2\uFCFE\uD1AB\uC3B3" + //  4550 -  4559
                "\uC3B4\u0000\u0000\uC3B5\u0000\u0000\u0000\uE1A5\u0000\u0000" + //  4560 -  4569
                "\u0000\u0000\u0000\uB8A7\uB8A8\u0000\uB8A9\uC3B1\u0000\u0000" + //  4570 -  4579
                "\u0000\u0000\uC3B2\u0000\u0000\uBAD9\uBADA\u0000\uBADB\u0000" + //  4580 -  4589
                "\u0000\u0000\uF3B0\u0000\u0000\uCCC4\u0000\u0000\uDABC\uD8FC" + //  4590 -  4599
                "\u0000\u0000\u0000\u0000\u0000\uB9C5\u0000\u0000\uB9C6\uC3B0" + //  4600 -  4609
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF4F5\u0000\uE1BC" + //  4610 -  4619
                "\uE0EF\u0000\u0000\uE9BF\uFCFD\uE6CE\u0000\uA4B0\uA4B1\uA4B2" + //  4620 -  4629
                "\uA4B3\uA4B4\uA4B5\uA4B6\uA4B7\uC3AD\u0000\u0000\u0000\uC3AE" + //  4630 -  4639
                "\u0000\uC3AF\u0000\uF6D9\uFAF4\u0000\u0000\u0000\u0000\u0000" + //  4640 -  4649
                "\uF8AA\uC3AB\uC3AC\u0000\u0000\u0000\u0000\u0000\u0000\uEED9" + //  4650 -  4659
                "\u0000\u0000\u0000\uF5B2\u0000\u0000\u0000\u0000\u0000\uB8A3" + //  4660 -  4669
                "\uB8A4\u0000\u0000\uBAC0\u0000\u0000\u0000\u0000\u0000\u0000" + //  4670 -  4679
                "\uE3B8\u0000\u0000\u0000\uD0AA\u0000\u0000\u0000\u0000\u0000" + //  4680 -  4689
                "\uB2D8\u0000\u0000\u0000\uA1EE\u0000\u0000\uA1F0\uA1C4\u0000" + //  4690 -  4699
                "\uCEC8\uEAB7\u0000\uFCC0\u0000\uFDE7\uF7EF\u0000\uA4A8\uA4A9" + //  4700 -  4709
                "\uA4AA\uA4AB\uA4AC\uA4AD\uA4AE\uA4AF\uC3A6\u0000\u0000\u0000" + //  4710 -  4719
                "\uC3A7\u0000\u0000\u0000\uF1D0\u0000\u0000\u0000\u0000\u0000" + //  4720 -  4729
                "\uB7F6\u0000\u0000\uB7F7\uC2FC\uC2FD\u0000\uC2FE\uC3A1\uC3A2" + //  4730 -  4739
                "\uC3A3\u0000\uF7C1\u0000\u0000\uE7B6\u0000\u0000\u0000\u0000" + //  4740 -  4749
                "\uC4DA\uC4DB\u0000\u0000\uE2E6\uE2A8\u0000\u0000\u0000\u0000" + //  4750 -  4759
                "\u0000\uDABD\u0000\u0000\u0000\uE7E5\u0000\u0000\uCFCA\uE1D1" + //  4760 -  4769
                "\uC2FB\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCABC\u0000" + //  4770 -  4779
                "\uF1BC\u0000\u0000\uFADA\u0000\u0000\uDAEA\uDAC6\uC2F7\uC2F8" + //  4780 -  4789
                "\u0000\u0000\uC2F9\u0000\uC2FA\u0000\uDDF3\uEAFA\u0000\uF6BD" + //  4790 -  4799
                "\uE1BB\uCDBF\uF4D4\uE6CD\uC2F0\u0000\u0000\u0000\uC2F1\u0000" + //  4800 -  4809
                "\u0000\u0000\uD3B3\u0000\u0000\u0000\u0000\u0000\uB7F2\uB7F3" + //  4810 -  4819
                "\u0000\u0000\uBAA1\uBAA2\u0000\u0000\u0000\u0000\u0000\uBCAD" + //  4820 -  4829
                "\uBCAE\uBCAF\uBCB0\uC2E9\u0000\u0000\u0000\u0000\u0000\u0000" + //  4830 -  4839
                "\u0000\uE7CA\u0000\uD9CB\u0000\uD9D2\uD3CB\uD8F7\uDAA9\uF5F8" + //  4840 -  4849
                "\u0000\uABF0\uABF1\uABF2\uABF3\uABF4\uABF5\uABF6\u0000\u0088" + //  4850 -  4859
                "\u0089\u008A\u008B\u008C\u008D\u0000\u0000\uF0A5\uCBFD\u0000" + //  4860 -  4869
                "\u0000\u0000\u0000\u0000\uE7E0\uEBAE\u0000\u0000\uF0BE\uD2BD" + //  4870 -  4879
                "\uCCA4\u0000\u0000\u0000\u0000\uBEF6\uBEF7\uBEF8\uBEF9\uC2E6" + //  4880 -  4889
                "\uC2E7\u0000\u0000\u0000\u0000\u0000\u0000\uE9B1\u0000\u0000" + //  4890 -  4899
                "\uFAAD\uC2E2\uC2E3\u0000\u0000\u0000\uC2E4\u0000\u0000\uB9F1" + //  4900 -  4909
                "\u0000\u0000\uB9F2\uB9F3\u0000\u0000\uE4AA\u0000\uF5E1\uEDDA" + //  4910 -  4919
                "\u0000\u0000\u0000\uDFC9\u0000\u0000\u0000\u0000\uDFED\uC2E1" + //  4920 -  4929
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD6A6\uDCBE\uC2DE" + //  4930 -  4939
                "\uC2DF\u0000\u0000\uC2E0\u0000\u0000\u0000\uF5DB\u0000\u0000" + //  4940 -  4949
                "\uFAC1\u0000\u0000\uB9E6\u0000\u0000\u0000\uB9E7\u0000\u0000" + //  4950 -  4959
                "\uEBDF\u0000\u0000\u0000\u0000\u0000\uD6CB\uC2DB\uC2DC\u0000" + //  4960 -  4969
                "\u0000\u0000\u0000\u0000\u0000\uCAB0\u0000\uF7A7\u0000\uF6D8" + //  4970 -  4979
                "\u0000\u0000\u0000\uD4C7\u0000\u0000\u0000\uFBAF\u0000\u0000" + //  4980 -  4989
                "\u0000\uCBD1\uC2DA\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  4990 -  4999
                "\uCBE6\u0000\uF7C0\u0000\uD0E3\u0000\u0000\u0000\uDAA1\u0000" + //  5000 -  5009
                "\uABE8\uABE9\uABEA\uABEB\uABEC\uABED\uABEE\uABEF\uC2D8\u0000" + //  5010 -  5019
                "\u0000\u0000\uC2D9\u0000\u0000\u0000\uE4C7\u0000\u0000\u0000" + //  5020 -  5029
                "\u0000\u0000\uB7EF\u0000\u0000\u0000\uF7AC\uEBC4\uEDE1\uE0AB" + //  5030 -  5039
                "\uDDC7\u0000\uCDE7\u0000\uE8DC\u0000\u0000\uE7D7\u0000\u0000" + //  5040 -  5049
                "\uDFCE\u0000\u0000\u0000\u0000\u0000\u0000\uF7B0\u0000\uCCEA" + //  5050 -  5059
                "\uC2D7\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF7C5\u0000" + //  5060 -  5069
                "\uF2AE\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCAD4\uC2D4" + //  5070 -  5079
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCCED\u0000\uE0B1" + //  5080 -  5089
                "\u0000\u0000\u0000\u0000\uDFA5\u0000\uF9D2\uC2D2\uC2D3\u0000" + //  5090 -  5099
                "\u0000\u0000\u0000\u0000\u0000\uD5B1\u0000\u0000\u0000\uCEF4" + //  5100 -  5109
                "\u0000\u0000\u0000\u0000\u0000\uB7E1\u0000\u0000\u0000\uFAA8" + //  5110 -  5119
                "\u0000\u0000\u0000\u0000\u0000\uD6C2\u0000\u0000\u0000\uEFEC" + //  5120 -  5129
                "\u0000\u0000\u0000\u0000\uC6A9\u0000\u0000\u0000\uF4E2\u0000" + //  5130 -  5139
                "\u0000\u0000\u0000\uC6AC\u0000\u0000\u0000\uF2D9\u0000\u0000" + //  5140 -  5149
                "\u0000\u0000\uC6B7\u0000\u0000\u0000\uF4D7\uCCA1\u0000\u0000" + //  5150 -  5159
                "\uCFBA\uC2CB\u0000\u0000\u0000\uC2CC\u0000\u0000\u0000\uFBAE" + //  5160 -  5169
                "\uD1E1\u0000\u0000\uDBC0\u0000\uEFE2\uF1F0\uCFB4\u0000\u0000" + //  5170 -  5179
                "\u0000\u0000\u0000\uCFCF\u0000\u0000\u0000\uD3B4\u0000\u0000" + //  5180 -  5189
                "\u0000\u0000\uFAA6\u0000\u0000\u0000\u0000\uBFFA\uBFFB\u0000" + //  5190 -  5199
                "\u0000\uD9BC\u0000\uE5C6\u0000\u0000\u0000\u0000\uC3C6\u0000" + //  5200 -  5209
                "\u0000\u0000\uD4AF\u0000\u0000\u0000\u0000\uDAE4\u0000\u0000" + //  5210 -  5219
                "\u0000\u0000\uBFD9\u0000\u0000\uBFDA\uC2C8\u0000\u0000\u0000" + //  5220 -  5229
                "\u0000\u0000\u0000\u0000\uDABA\u0000\uE7B5\u0000\u0000\u0000" + //  5230 -  5239
                "\u0000\u0000\uDBF0\u0000\uABE0\uABE1\uABE2\uABE3\uABE4\uABE5" + //  5240 -  5249
                "\uABE6\uABE7\uC2C5\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  5250 -  5259
                "\uE3A8\u0000\uEEFD\u0000\u0000\u0000\u0000\u0000\u0000\uE4AB" + //  5260 -  5269
                "\uC2C3\uC2C4\u0000\u0000\u0000\u0000\u0000\u0000\uFCE6\u0000" + //  5270 -  5279
                "\u0000\u0000\uF2C0\u0000\u0000\uF1E5\u0000\uF4C3\uC2BE\u0000" + //  5280 -  5289
                "\u0000\u0000\uC2BF\u0000\u0000\u0000\uEBB4\uEAA1\u0000\uF8BC" + //  5290 -  5299
                "\uCEA6\u0000\uDDF2\u0000\u0000\uD9BE\u0000\u0000\u0000\u0000" + //  5300 -  5309
                "\uC4AF\uC4B0\u0000\uC4B1\uC2BA\u0000\u0000\u0000\u0000\u0000" + //  5310 -  5319
                "\u0000\u0000\uCDE5\u0000\uEDD1\u0000\u0000\u0000\u0000\u0000" + //  5320 -  5329
                "\uE9F9\u0000\uABD8\uABD9\uABDA\uABDB\uABDC\uABDD\uABDE\uABDF" + //  5330 -  5339
                "\uC2B4\uC2B5\u0000\uC2B6\uC2B7\uC2B8\u0000\u0000\uB9BC\u0000" + //  5340 -  5349
                "\uB9BD\u0000\u0000\u0000\u0000\uD0C4\u0000\u0000\u0000\uCAD0" + //  5350 -  5359
                "\uC2B3\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD0DA\uD0DB" + //  5360 -  5369
                "\uC2B0\uC2B1\u0000\u0000\uC2B2\u0000\u0000\u0000\uE9E6\u0000" + //  5370 -  5379
                "\u0000\uE3F6\u0000\u0000\uB9B6\u0000\u0000\u0000\uB9B7\u0000" + //  5380 -  5389
                "\uB9B8\uC2AE\uC2AF\u0000\u0000\u0000\u0000\u0000\u0000\uD7B4" + //  5390 -  5399
                "\u0000\u0000\u0000\uD5F9\u0000\u0000\u0000\u0000\u0000\uB7CE" + //  5400 -  5409
                "\uB7CF\u0000\u0000\uB9A9\u0000\uB9AA\u0000\u0000\u0000\u0000" + //  5410 -  5419
                "\uF8F4\u0000\uD9B7\u0000\u0000\uE6F3\u0000\u0000\u0000\u0000" + //  5420 -  5429
                "\u0000\u0000\uC1C5\u0000\uC1C6\uC2A7\u0000\uC2A8\u0000\uC2A9" + //  5430 -  5439
                "\u0000\u0000\uC2AA\uC1FC\uC1FD\u0000\uC1FE\u0000\uC2A1\uC2A2" + //  5440 -  5449
                "\u0000\uCFFD\u0000\u0000\uDEDD\u0000\u0000\u0000\uD9D1\uC1FA" + //  5450 -  5459
                "\u0000\uC1FB\u0000\u0000\u0000\u0000\u0000\uBCF2\uBCF3\u0000" + //  5460 -  5469
                "\uBCF4\uC1F6\uC1F7\u0000\u0000\uC1F8\u0000\u0000\uC1F9\uC1F2" + //  5470 -  5479
                "\uC1F3\u0000\uC1F4\u0000\uC1F5\u0000\u0000\uB8C3\u0000\uB8C4" + //  5480 -  5489
                "\uB8C5\uB8C6\u0000\u0000\uE2C8\u0000\u0000\u0000\u0000\u0000" + //  5490 -  5499
                "\u0000\uBCE1\u0000\u0000\uEABD\uE6FE\u0000\uF7C4\uF5AD\u0000" + //  5500 -  5509
                "\uD9E0\uC1F1\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF9C4" + //  5510 -  5519
                "\u0000\uCFB1\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDBDA" + //  5520 -  5529
                "\uC1EE\uC1EF\u0000\u0000\uC1F0\u0000\u0000\u0000\uF4E5\uD8C2" + //  5530 -  5539
                "\uDCD0\uCCEE\u0000\u0000\uB8B5\u0000\u0000\u0000\u0000\u0000" + //  5540 -  5549
                "\u0000\uF5CD\u0000\uF1F2\uFAC7\uC1EB\u0000\u0000\u0000\uC1EC" + //  5550 -  5559
                "\u0000\u0000\u0000\uFACD\u0000\u0000\u0000\u0000\u0000\uB7C5" + //  5560 -  5569
                "\uB7C6\u0000\uB7C7\uC1E7\uC1E8\u0000\uC1E9\u0000\u0000\u0000" + //  5570 -  5579
                "\u0000\uD3B5\u0000\u0000\u0000\u0000\uC2EA\u0000\u0000\u0000" + //  5580 -  5589
                "\uF3B3\uE4D8\uCFF9\uCFDA\u0000\u0000\uB8AA\uB8AB\u0000\u0000" + //  5590 -  5599
                "\uB8AC\uB8AD\u0000\uFDE2\uF3AD\u0000\uFDDB\u0000\u0000\u0000" + //  5600 -  5609
                "\u0000\uC4AB\uC4AC\u0000\u0000\uD5DA\u0000\uD7A7\u0000\u0000" + //  5610 -  5619
                "\u0000\uEEC0\uC1E6\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  5620 -  5629
                "\uCAE9\u0000\uD3AA\u0000\u0000\u0000\uCCAC\u0000\u0000\u0000" + //  5630 -  5639
                "\uF1CD\u0000\u0000\u0000\u0000\uCFD6\u0000\uD7F0\u0000\uEBE1" + //  5640 -  5649
                "\uC1E3\uC1E4\u0000\u0000\uC1E5\u0000\u0000\u0000\uD7F5\uE3F3" + //  5650 -  5659
                "\uCFE5\u0000\u0000\u0000\uE7F2\uEDDF\u0000\u0000\uCACB\u0000" + //  5660 -  5669
                "\uD0B9\uD4F2\u0000\u0000\u0000\u0000\u0000\uD1A6\uC1E0\u0000" + //  5670 -  5679
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE4B6\u0000\uE8FA\u0000" + //  5680 -  5689
                "\u0000\uCEE9\u0000\u0000\u0000\u0000\uC3C7\u0000\u0000\u0000" + //  5690 -  5699
                "\uF2E2\u0000\u0000\u0000\u0000\uFDDE\uCAC0\u0000\u0000\u0000" + //  5700 -  5709
                "\uCEF2\u0000\uD6D9\u0000\u0000\uCDBE\u0000\uDAE9\u0000\u0000" + //  5710 -  5719
                "\u0000\u0000\uC3C0\uC3C1\u0000\uC3C2\uC1D8\u0000\u0000\u0000" + //  5720 -  5729
                "\uC1D9\uC1DA\uC1DB\u0000\uDAD4\uE2A7\uFBFC\u0000\u0000\uF1DC" + //  5730 -  5739
                "\u0000\u0000\uD9EA\uF5A2\u0000\u0000\u0000\uD7D1\u0000\uABD0" + //  5740 -  5749
                "\uABD1\uABD2\uABD3\uABD4\uABD5\uABD6\uABD7\uC1D2\uC1D3\u0000" + //  5750 -  5759
                "\u0000\uC1D4\u0000\u0000\u0000\uF1CC\u0000\u0000\uE5B8\u0000" + //  5760 -  5769
                "\u0000\uB7E6\u0000\u0000\u0000\u0000\u0000\u0000\uEDC4\u0000" + //  5770 -  5779
                "\u0000\u0000\uCCE1\u0000\u0000\u0000\u0000\uD6EA\uC1CC\u0000" + //  5780 -  5789
                "\u0000\u0000\uC1CD\u0000\u0000\u0000\uF1A1\u0000\u0000\u0000" + //  5790 -  5799
                "\u0000\u0000\uB7B3\uB7B4\u0000\uB7B5\uC1C8\u0000\u0000\u0000" + //  5800 -  5809
                "\u0000\u0000\u0000\u0000\uFBDB\u0000\uCEAA\u0000\uCBC8\u0000" + //  5810 -  5819
                "\u0000\u0000\u0000\u0000\uCDE6\u0000\u0000\u0000\uF0EB\u0000" + //  5820 -  5829
                "\u0000\u0000\u0000\uCBDB\u0000\u0000\u0000\u0000\uBFC0\uBFC1" + //  5830 -  5839
                "\u0000\u0000\uD4A1\uCEB2\u0000\u0000\u0000\u0000\u0000\uDEF8" + //  5840 -  5849
                "\uF8E9\uE3DE\u0000\uABC8\uABC9\uABCA\uABCB\uABCC\uABCD\uABCE" + //  5850 -  5859
                "\uABCF\uC1BB\uC1BC\u0000\uC1BD\u0000\uC1BE\uC1BF\uC1C0\uC1B9" + //  5860 -  5869
                "\u0000\uC1BA\u0000\u0000\u0000\u0000\u0000\uBCEE\uBCEF\u0000" + //  5870 -  5879
                "\u0000\uB7D5\u0000\u0000\u0000\u0000\u0000\u0000\uE5F1\u0000" + //  5880 -  5889
                "\u0000\u0000\uD4BB\u0000\u0000\u0000\u0000\uFDFA\uC1B6\uC1B7" + //  5890 -  5899
                "\u0000\u0000\uC1B8\u0000\u0000\u0000\uE9C9\u0000\u0000\u0000" + //  5900 -  5909
                "\u0000\uD3CE\uC1B1\uC1B2\u0000\u0000\uC1B3\uC1B4\u0000\u0000" + //  5910 -  5919
                "\uB7CC\u0000\uB7CD\u0000\u0000\u0000\u0000\uF6E1\u0000\u0000" + //  5920 -  5929
                "\u0000\u0000\uC2AB\uC2AC\u0000\uC2AD\uC1B0\u0000\u0000\u0000" + //  5930 -  5939
                "\u0000\u0000\u0000\u0000\uE5EE\uDCA2\uC1AE\u0000\u0000\u0000" + //  5940 -  5949
                "\uC1AF\u0000\u0000\u0000\uDDBA\u0000\u0000\u0000\uF2BF\u0000" + //  5950 -  5959
                "\uCED9\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCDBD\uC1A8" + //  5960 -  5969
                "\u0000\u0000\u0000\uC1A9\u0000\u0000\u0000\uEAC6\u0000\u0000" + //  5970 -  5979
                "\u0000\u0000\u0000\uB7AF\uB7B0\u0000\u0000\uB7AE\u0000\u0000" + //  5980 -  5989
                "\u0000\u0000\u0000\u0000\uDDC4\u0000\u0000\u0000\uFBA7\u0000" + //  5990 -  5999
                "\uE9C2\u0000\u0000\u0000\uDEAB\uDBE8\u0000\u0000\uE3DD\uC1A1" + //  6000 -  6009
                "\uC1A2\u0000\uC1A3\u0000\uC1A4\uC1A5\u0000\uF5D1\uE7B3\u0000" + //  6010 -  6019
                "\u0000\u0000\u0000\u0000\u0000\uD1BE\u0000\u0000\uD8F6\u0000" + //  6020 -  6029
                "\uD1A4\u0000\uCDE2\u0000\u0000\uD5D9\u0000\u0000\u0000\uD8DE" + //  6030 -  6039
                "\u0000\u0000\uCEE8\uDBDB\u0000\u0000\u0000\u0000\u0000\uF8FD" + //  6040 -  6049
                "\u0000\u0000\uF8FC\uC0FD\u0000\uC0FE\u0000\u0000\u0000\u0000" + //  6050 -  6059
                "\u0000\uBCE5\u0000\u0000\uBCE6\uC0FA\uC0FB\u0000\u0000\uC0FC" + //  6060 -  6069
                "\u0000\u0000\u0000\uF2F5\u0000\u0000\uD4AE\u0000\u0000\uB6D7" + //  6070 -  6079
                "\u0000\u0000\u0000\u0000\u0000\u0000\uCFA5\u0000\u0000\uDFC4" + //  6080 -  6089
                "\uC0F8\u0000\u0000\u0000\uC0F9\u0000\u0000\u0000\uD6F2\u0000" + //  6090 -  6099
                "\uDEF4\u0000\uDFDB\u0000\uEFDF\u0000\u0000\uF1EF\u0000\uE5F6" + //  6100 -  6109
                "\uEEBF\uE2E4\uC0F5\u0000\u0000\u0000\u0000\uC0F6\u0000\u0000" + //  6110 -  6119
                "\uB5EE\u0000\u0000\u0000\u0000\u0000\u0000\uF6F2\u0000\uDFC2" + //  6120 -  6129
                "\u0000\uD1F8\uEAF8\uEAF9\uDAB3\u0000\u0000\u0000\u0000\uC3A8" + //  6130 -  6139
                "\uC3A9\u0000\uC3AA\uC0F4\u0000\u0000\u0000\u0000\u0000\u0000" + //  6140 -  6149
                "\u0000\uF4C0\u0000\uE5AB\u0000\u0000\u0000\u0000\u0000\u0000" + //  6150 -  6159
                "\u0000\uDDED\uC0F0\uC0F1\u0000\u0000\uC0F2\u0000\uC0F3\u0000" + //  6160 -  6169
                "\uE8CA\u0000\u0000\u0000\u0000\uEBF5\u0000\u0000\uE8C9\uF4FE" + //  6170 -  6179
                "\u0000\u0000\u0000\u0000\uE7FC\uC0EE\uC0EF\u0000\u0000\u0000" + //  6180 -  6189
                "\u0000\u0000\u0000\uE7A5\u0000\uD5F5\uD3BE\uC0E9\u0000\u0000" + //  6190 -  6199
                "\u0000\uC0EA\u0000\u0000\u0000\uD8E7\u0000\uD9C9\u0000\u0000" + //  6200 -  6209
                "\u0000\uCED4\uE7AB\u0000\u0000\u0000\uCBC3\uC0E1\uC0E2\u0000" + //  6210 -  6219
                "\uC0E3\uC0E4\uC0E5\uC0E6\u0000\uCDA1\u0000\u0000\u0000\u0000" + //  6220 -  6229
                "\u0000\uDFB5\u0000\uABC0\uABC1\uABC2\uABC3\uABC4\uABC5\uABC6" + //  6230 -  6239
                "\uABC7\uC0DF\u0000\uC0E0\u0000\u0000\u0000\u0000\u0000\uBCE2" + //  6240 -  6249
                "\u0000\u0000\u0000\uF5B0\u0000\u0000\u0000\u0000\u0000\uB6F7" + //  6250 -  6259
                "\uB6F8\u0000\uB6F9\uC0DA\uC0DB\u0000\u0000\uC0DC\u0000\uC0DD" + //  6260 -  6269
                "\uC0DE\uC0D6\uC0D7\uC0D8\u0000\u0000\u0000\uC0D9\u0000\uD3D6" + //  6270 -  6279
                "\u0000\u0000\u0000\u0000\u0000\u0000\uEDD0\uC0CE\u0000\u0000" + //  6280 -  6289
                "\u0000\uC0CF\uC0D0\uC0D1\u0000\uD0BE\u0000\uDDDC\u0000\u0000" + //  6290 -  6299
                "\u0000\u0000\uD4D6\uC0CA\u0000\u0000\uC0CB\u0000\u0000\u0000" + //  6300 -  6309
                "\u0000\uD9E8\u0000\uF7EB\uF5C9\u0000\uCCFE\uCDE1\u0000\uE1BA" + //  6310 -  6319
                "\u0000\uDBEF\uDAB2\u0000\uABB8\uABB9\uABBA\uABBB\uABBC\uABBD" + //  6320 -  6329
                "\uABBE\uABBF\uC0C9\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  6330 -  6339
                "\uCCEC\u0000\uD7DE\u0000\u0000\u0000\u0000\uDEDC\u0000\uF0AC" + //  6340 -  6349
                "\uC0C7\u0000\u0000\u0000\uC0C8\u0000\u0000\u0000\uFBDF\uE7E3" + //  6350 -  6359
                "\u0000\u0000\u0000\u0000\uFCF1\u0000\u0000\u0000\uD0BC\uC0BA" + //  6360 -  6369
                "\u0000\u0000\u0000\uC0BB\u0000\u0000\u0000\uD4CA\u0000\u0000" + //  6370 -  6379
                "\u0000\u0000\u0000\uB6EA\uB6EB\u0000\u0000\uB5DF\u0000\u0000" + //  6380 -  6389
                "\u0000\u0000\u0000\u0000\uD8C5\u0000\u0000\u0000\uF1BD\u0000" + //  6390 -  6399
                "\u0000\uE2E7\uFDD7\u0000\uD4D5\uDFCD\u0000\uFCB8\uD1D0\u0000" + //  6400 -  6409
                "\u0000\u0000\uDDB8\uCFC5\uDFDF\u0000\u0000\uCFAF\u0000\u0000" + //  6410 -  6419
                "\uCAD3\u0000\u0000\uCAAF\uC0B3\uC0B4\u0000\uC0B5\u0000\uC0B6" + //  6420 -  6429
                "\u0000\uC0B7\uC0B2\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  6430 -  6439
                "\uCADD\uD5DE\uC0AF\uC0B0\u0000\u0000\uC0B1\u0000\u0000\u0000" + //  6440 -  6449
                "\uF3C6\u0000\u0000\u0000\u0000\u0000\uB6E7\u0000\u0000\u0000" + //  6450 -  6459
                "\uCAED\u0000\u0000\u0000\u0000\uE8EB\uC0A9\u0000\u0000\u0000" + //  6460 -  6469
                "\uC0AA\u0000\u0000\u0000\uDCA8\u0000\u0000\u0000\u0000\u0000" + //  6470 -  6479
                "\uB6D8\u0000\u0000\u0000\uE1F7\u0000\u0000\u0000\u0000\u0000" + //  6480 -  6489
                "\uF5ED\u0000\uCFF3\u0000\uE4D0\u0000\u0000\u0000\u0000\u0000" + //  6490 -  6499
                "\uF2EE\u0000\uABB0\uABB1\uABB2\uABB3\uABB4\uABB5\uABB6\uABB7" + //  6500 -  6509
                "\uC0A4\uC0A5\u0000\u0000\u0000\uC0A6\u0000\u0000\uB5BF\u0000" + //  6510 -  6519
                "\uB5C0\u0000\uB5C1\u0000\u0000\uF5EC\u0000\u0000\u0000\u0000" + //  6520 -  6529
                "\u0000\uEEE7\uC0A3\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  6530 -  6539
                "\uF5AB\u0000\uCBFA\uF9F9\uCCFD\uD3FE\u0000\u0000\u0000\u0000" + //  6540 -  6549
                "\uC3A4\uC3A5\u0000\u0000\uE8DB\u0000\uDBB3\u0000\u0000\u0000" + //  6550 -  6559
                "\uD1F7\uBFFE\uC0A1\u0000\u0000\uC0A2\u0000\u0000\u0000\uEFF5" + //  6560 -  6569
                "\uCADF\u0000\uEBB1\uEDBF\u0000\uFAB7\uD0C6\u0000\u0000\uCCAB" + //  6570 -  6579
                "\uEEA8\u0000\u0000\uDFB4\u0000\u0000\u0000\u0000\uD7DD\uFABA" + //  6580 -  6589
                "\uBFFC\uBFFD\u0000\u0000\u0000\u0000\u0000\u0000\uE7FB\uFCB7" + //  6590 -  6599
                "\uFCE4\uFBC5\uBFF8\u0000\u0000\u0000\uBFF9\u0000\u0000\u0000" + //  6600 -  6609
                "\uD5DF\u0000\u0000\u0000\uD6E5\u0000\uD2B0\uF1BA\u0000\uD7B3" + //  6610 -  6619
                "\uE3C3\uF3FD\uDEDA\u0000\uABA8\uABA9\uABAA\uABAB\uABAC\uABAD" + //  6620 -  6629
                "\uABAE\uABAF\uBFF2\uBFF3\u0000\uBFF4\u0000\uBFF5\u0000\u0000" + //  6630 -  6639
                "\uB5A2\u0000\uB5A3\u0000\u0000\uB5A4\u0000\uD9BB\uCAF3\uF6D3" + //  6640 -  6649
                "\uE6F8\uEAF5\u0000\u0000\u0000\uEBB0\uF4E3\u0000\u0000\u0000" + //  6650 -  6659
                "\uCFC4\u0000\u0000\u0000\u0000\uCBA7\u0000\uDACE\u0000\u0000" + //  6660 -  6669
                "\uF1B9\u0000\u0000\uDAD3\u0000\uF6EA\u0000\uAAF0\uAAF1\uAAF2" + //  6670 -  6679
                "\uAAF3\u0000\u0000\u0000\u0000\uBEBA\u0000\u0000\u0000\uD6D0" + //  6680 -  6689
                "\u0000\u0000\u0000\u0000\uC3E9\u0000\u0000\u0000\uE5A1\u0000" + //  6690 -  6699
                "\u0000\u0000\u0000\uC3F5\u0000\u0000\u0000\uE4EA\uF2CF\u0000" + //  6700 -  6709
                "\uF7BF\u0000\u0080\u0081\u0082\u0083\u0084\u0085\u0086\u0087" + //  6710 -  6719
                "\uBFEF\uBFF0\uBFF1\u0000\u0000\u0000\u0000\u0000\uBCD8\uBCD9" + //  6720 -  6729
                "\u0000\uBCDA\uBFEC\uBFED\u0000\u0000\uBFEE\u0000\u0000\u0000" + //  6730 -  6739
                "\uF4DB\u0000\uE2F4\u0000\u0000\uD3C8\uBFE6\u0000\u0000\u0000" + //  6740 -  6749
                "\uBFE7\u0000\u0000\u0000\uDBBE\u0000\u0000\uE0E2\u0000\u0000" + //  6750 -  6759
                "\uB4BF\u0000\u0000\u0000\u0000\u0000\u0000\uF6E3\u0000\u0000" + //  6760 -  6769
                "\u0000\uCBFE\uEDEA\u0000\u0000\u0000\u0000\uC3D9\uC3DA\u0000" + //  6770 -  6779
                "\uC3DB\uBFE0\uBFE1\u0000\uBFE2\u0000\uBFE3\u0000\u0000\uB4B1" + //  6780 -  6789
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE2D2\u0000\uF6A2\uE1F4" + //  6790 -  6799
                "\uBFDF\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE1A1\u0000" + //  6800 -  6809
                "\uCFAE\u0000\u0000\u0000\u0000\u0000\uE3C2\u0000\uAAE8\uAAE9" + //  6810 -  6819
                "\uAAEA\uAAEB\uAAEC\uAAED\uAAEE\uAAEF\uBFDC\uBFDD\u0000\u0000" + //  6820 -  6829
                "\uBFDE\u0000\u0000\u0000\uCAC7\u0000\u0000\u0000\u0000\u0000" + //  6830 -  6839
                "\uB6CD\u0000\u0000\u0000\uD2A5\u0000\u0000\uFDEE\u0000\u0000" + //  6840 -  6849
                "\uB4A6\u0000\uB4A7\u0000\uB4A8\u0000\u0000\uF1FB\u0000\u0000" + //  6850 -  6859
                "\u0000\u0000\uFDD2\uD1C1\uBFD8\u0000\u0000\u0000\u0000\u0000" + //  6860 -  6869
                "\u0000\u0000\uDDAE\u0000\uF8C2\u0000\u0000\uF2AC\u0000\u0000" + //  6870 -  6879
                "\uCAAD\uCAAE\uBFD1\uBFD2\u0000\uBFD3\uBFD4\uBFD5\u0000\u0000" + //  6880 -  6889
                "\uB3A9\u0000\u0000\u0000\u0000\u0000\u0000\uF7F7\uDCAC\u0000" + //  6890 -  6899
                "\u0000\uCCAF\u0000\u0000\u0000\u0000\u0000\u0000\uB0EB\u0000" + //  6900 -  6909
                "\uB0EC\uBFD0\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE9A4" + //  6910 -  6919
                "\u0000\uE0B0\u0000\u0000\uD4E2\u0000\uF6D7\u0000\uD7F9\uBFCD" + //  6920 -  6929
                "\uBFCE\u0000\u0000\uBFCF\u0000\u0000\u0000\uCAC6\u0000\u0000" + //  6930 -  6939
                "\uD5C2\u0000\u0000\uB2FE\u0000\u0000\u0000\uB3A1\u0000\u0000" + //  6940 -  6949
                "\uE9A9\u0000\uD3C7\u0000\u0000\uDCDD\uF8AE\uBFC6\u0000\u0000" + //  6950 -  6959
                "\uBFC7\uBFC8\uBFC9\u0000\uBFCA\uBFC2\u0000\u0000\u0000\uBFC3" + //  6960 -  6969
                "\uBFC4\uBFC5\u0000\uCBF9\uD4D4\u0000\uD9DC\u0000\uEEBE\u0000" + //  6970 -  6979
                "\uF7ED\uBFBC\uBFBD\u0000\uBFBE\uBFBF\u0000\u0000\u0000\uE3A9" + //  6980 -  6989
                "\u0000\u0000\u0000\u0000\u0000\uB6C0\uB6C1\u0000\uB6C2\uBFBB" + //  6990 -  6999
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDFFD\u0000\uEEA7" + //  7000 -  7009
                "\uF5BD\u0000\uF8F5\u0000\u0000\uEDE8\u0000\uAAE0\uAAE1\uAAE2" + //  7010 -  7019
                "\uAAE3\uAAE4\uAAE5\uAAE6\uAAE7\uBFB9\u0000\u0000\u0000\uBFBA" + //  7020 -  7029
                "\u0000\u0000\u0000\uD8CA\u0000\u0000\u0000\u0000\u0000\uB6AB" + //  7030 -  7039
                "\uB6AC\u0000\uB6AD\uBFB4\uBFB5\u0000\u0000\u0000\uBFB6\uBFB7" + //  7040 -  7049
                "\uBFB8\uBFAC\u0000\u0000\u0000\uBFAD\u0000\uBFAE\uBFAF\uBFA5" + //  7050 -  7059
                "\uBFA6\u0000\uBFA7\u0000\uBFA8\u0000\u0000\uB2D1\u0000\u0000" + //  7060 -  7069
                "\u0000\u0000\u0000\u0000\uF4A2\u0000\uF1D7\u0000\uD9CA\u0000" + //  7070 -  7079
                "\uDAB1\uD8C7\uDCE2\uF3CE\uF5F4\u0000\uAAD8\uAAD9\uAADA\uAADB" + //  7080 -  7089
                "\uAADC\uAADD\uAADE\uAADF\uBFA4\u0000\u0000\u0000\u0000\u0000" + //  7090 -  7099
                "\u0000\u0000\uEFF2\uE2B7\uBFA1\uBFA2\u0000\u0000\uBFA3\u0000" + //  7100 -  7109
                "\u0000\u0000\uF7DB\u0000\u0000\u0000\u0000\u0000\uB5F5\uB5F6" + //  7110 -  7119
                "\u0000\uB5F7\uBEFA\uBEFB\uBEFC\u0000\uBEFD\u0000\uBEFE\u0000" + //  7120 -  7129
                "\uEAC5\uEAF3\u0000\uDDDB\u0000\uDCD7\u0000\u0000\uDFF5\u0000" + //  7130 -  7139
                "\u0000\uE8F8\uF8ED\u0000\u0000\uE0E7\u0000\uCCD9\u0000\u0000" + //  7140 -  7149
                "\uD4C6\u0000\uAAD0\uAAD1\uAAD2\uAAD3\uAAD4\uAAD5\uAAD6\uAAD7" + //  7150 -  7159
                "\uBEF0\uBEF1\u0000\uBEF2\uBEF3\uBEF4\uBEF5\u0000\uF5CC\u0000" + //  7160 -  7169
                "\u0000\uFCE5\u0000\u0000\u0000\u0000\uC2F2\uC2F3\u0000\u0000" + //  7170 -  7179
                "\uFCFB\u0000\u0000\u0000\u0000\u0000\u0000\uD4FE\u0000\u0000" + //  7180 -  7189
                "\uF1DB\u0000\uFAD9\u0000\uF1B8\uFDF5\uE0F9\uBEEC\u0000\u0000" + //  7190 -  7199
                "\u0000\u0000\u0000\u0000\u0000\uCEED\u0000\uDFF6\uF0C7\uF0C6" + //  7200 -  7209
                "\u0000\uD8BA\u0000\uF1F4\uF4F0\uBEEA\u0000\u0000\u0000\uBEEB" + //  7210 -  7219
                "\u0000\u0000\u0000\uFAE5\uE2FA\u0000\u0000\u0000\uCAB6\uBEE1" + //  7220 -  7229
                "\u0000\u0000\u0000\uBEE2\u0000\u0000\uBEE3\uBEDA\uBEDB\u0000" + //  7230 -  7239
                "\uBEDC\uBEDD\uBEDE\u0000\u0000\uB2C7\uB2C8\uB2C9\u0000\u0000" + //  7240 -  7249
                "\u0000\u0000\uDED4\u0000\uE0AC\u0000\u0000\uD7EB\u0000\u0000" + //  7250 -  7259
                "\u0000\u0000\u0000\u0000\uE0A8\uD5F3\u0000\uE8D8\u0000\uCDD8" + //  7260 -  7269
                "\uE7D6\uCCDA\u0000\u0000\uCAE3\uBED9\u0000\u0000\u0000\u0000" + //  7270 -  7279
                "\u0000\u0000\u0000\uDDD0\uF0D9\uBED6\uBED7\u0000\u0000\uBED8" + //  7280 -  7289
                "\u0000\u0000\u0000\uF1E3\uD5EE\u0000\u0000\u0000\u0000\uA1A4" + //  7290 -  7299
                "\u0000\u0000\u0000\u0000\uBEBD\u0000\u0000\u0000\uF4C9\u0000" + //  7300 -  7309
                "\u0000\u0000\u0000\u0000\uDEA7\u0000\u0000\uE6BB\uBED2\uBED3" + //  7310 -  7319
                "\u0000\u0000\u0000\uBED4\uBED5\u0000\uCEC7\u0000\u0000\u0000" + //  7320 -  7329
                "\u0000\u0000\uFDF6\u0000\uAAC8\uAAC9\uAACA\uAACB\uAACC\uAACD" + //  7330 -  7339
                "\uAACE\uAACF\uBEC8\uBEC9\uBECA\u0000\uBECB\uBECC\uBECD\u0000" + //  7340 -  7349
                "\uCDBC\u0000\uF3E5\u0000\u0000\u0000\u0000\u0000\uE4D4\u0000" + //  7350 -  7359
                "\u0000\u0000\uE0A5\u0000\uF7AB\u0000\u0000\uF2CB\u0000\uF2CC" + //  7360 -  7369
                "\u0000\u0000\u0000\uE4CF\uBEC2\uBEC3\u0000\uBEC4\u0000\uBEC5" + //  7370 -  7379
                "\u0000\u0000\uB1C3\uB1C4\u0000\u0000\u0000\u0000\u0000\uBAF1" + //  7380 -  7389
                "\uBAF2\u0000\u0000\uF6C9\u0000\u0000\u0000\u0000\u0000\u0000" + //  7390 -  7399
                "\uD8EA\u0000\u0000\uCEB7\u0000\u0000\u0000\u0000\u0000\u0000" + //  7400 -  7409
                "\uD4EF\u0000\u0000\uE0C4\u0000\uCFB9\u0000\uD5CA\uD7E2\uE2AF" + //  7410 -  7419
                "\uBEC1\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD3D8\u0000" + //  7420 -  7429
                "\uE5C3\u0000\u0000\u0000\u0000\uD9A6\u0000\u0000\uF0C5\uE3C1" + //  7430 -  7439
                "\uFCCC\uFCCD\u0000\u0000\u0000\uEECC\u0000\u0000\uDEF2\u0000" + //  7440 -  7449
                "\uAAC0\uAAC1\uAAC2\uAAC3\uAAC4\uAAC5\uAAC6\uAAC7\uBEBE\uBEBF" + //  7450 -  7459
                "\u0000\u0000\uBEC0\u0000\u0000\u0000\uF5F9\u0000\u0000\u0000" + //  7460 -  7469
                "\u0000\u0000\uB5E5\uB5E6\u0000\u0000\uB1B6\u0000\uB1B7\u0000" + //  7470 -  7479
                "\u0000\u0000\u0000\uDAAE\uCAEE\u0000\u0000\u0000\uCFEC\u0000" + //  7480 -  7489
                "\u0000\u0000\u0000\uF4DF\uBEBB\u0000\u0000\u0000\uBEBC\u0000" + //  7490 -  7499
                "\u0000\u0000\uD5A2\u0000\u0000\u0000\u0000\u0000\uB5DA\u0000" + //  7500 -  7509
                "\u0000\u0000\uD5E5\u0000\u0000\u0000\u0000\u0000\uD5C1\u0000" + //  7510 -  7519
                "\u0000\u0000\uFADB\uCBE3\uF7A9\u0000\uFBA6\uBEB8\uBEB9\u0000" + //  7520 -  7529
                "\u0000\u0000\u0000\u0000\u0000\uF4CC\uDAFC\u0000\u0000\uB1B2" + //  7530 -  7539
                "\u0000\u0000\u0000\u0000\u0000\u0000\uDAE1\u0000\uD6B6\u0000" + //  7540 -  7549
                "\uEAD1\uDFF4\u0000\u0000\u0000\u0000\uD1EC\uE4DE\uBEB5\u0000" + //  7550 -  7559
                "\uBEB6\u0000\u0000\u0000\u0000\uBEB7\uBEB2\uBEB3\u0000\u0000" + //  7560 -  7569
                "\uBEB4\u0000\u0000\u0000\uFBAB\u0000\u0000\u0000\u0000\u0000" + //  7570 -  7579
                "\uB5D6\u0000\u0000\u0000\uDACF\u0000\uDCD4\u0000\uDCA6\u0000" + //  7580 -  7589
                "\uFDBB\uFDC7\u0000\u0000\u0000\u0000\uE7B2\u0000\uAAB8\uAAB9" + //  7590 -  7599
                "\uAABA\uAABB\uAABC\uAABD\uAABE\uAABF\uBEAF\u0000\u0000\u0000" + //  7600 -  7609
                "\uBEB0\u0000\u0000\u0000\uD5CC\u0000\u0000\u0000\u0000\u0000" + //  7610 -  7619
                "\uB5CD\u0000\u0000\u0000\uDDE3\u0000\u0000\u0000\u0000\uFCDD" + //  7620 -  7629
                "\uBEAC\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD5DB\u0000" + //  7630 -  7639
                "\uEFDD\u0000\uF2AA\u0000\u0000\u0000\u0000\u0000\uF0B9\uE4FE" + //  7640 -  7649
                "\uE4C9\u0000\uAAB0\uAAB1\uAAB2\uAAB3\uAAB4\uAAB5\uAAB6\uAAB7" + //  7650 -  7659
                "\uBEA7\u0000\u0000\u0000\uBEA8\u0000\u0000\u0000\uE0D0\u0000" + //  7660 -  7669
                "\u0000\u0000\u0000\u0000\uB5C5\u0000\u0000\u0000\uDEFB\uD0BB" + //  7670 -  7679
                "\uD5B7\uEEF1\u0000\u0000\uB0E7\u0000\u0000\uB0E8\u0000\u0000" + //  7680 -  7689
                "\u0000\uF3A4\u0000\u0000\u0000\uD4FB\uFCE3\uBEA4\u0000\u0000" + //  7690 -  7699
                "\u0000\u0000\u0000\u0000\u0000\uD3AF\u0000\uE3E7\uD8B9\u0000" + //  7700 -  7709
                "\uF6F8\u0000\u0000\uDCC5\uCCD8\uBDFE\u0000\u0000\u0000\uBEA1" + //  7710 -  7719
                "\u0000\u0000\u0000\uE8B3\u0000\u0000\u0000\u0000\u0000\uB5B5" + //  7720 -  7729
                "\uB5B6\u0000\u0000\uB0D1\uB0D2\uB0D3\uB0D4\u0000\u0000\u0000" + //  7730 -  7739
                "\uDDE9\u0000\u0000\u0000\uF1F3\u0000\uF9C0\uE9F0\u0000\u0000" + //  7740 -  7749
                "\uD9DB\u0000\uF3E4\u0000\uAAA8\uAAA9\uAAAA\uAAAB\uAAAC\uAAAD" + //  7750 -  7759
                "\uAAAE\uAAAF\uBDFB\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  7760 -  7769
                "\uF9B6\u0000\uEDC6\u0000\u0000\u0000\u0000\uE1B9\u0000\uE3C0" + //  7770 -  7779
                "\uBDFA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCCE6\u0000" + //  7780 -  7789
                "\uE7B1\u0000\u0000\u0000\u0000\uF5F0\u0000\uD8DC\uBDF9\u0000" + //  7790 -  7799
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF0FA\u0000\uDDCB\uD0D4" + //  7800 -  7809
                "\u0000\uE6B6\uE0AE\uFDDA\u0000\u0000\uF9D1\u0000\u0000\uE9D9" + //  7810 -  7819
                "\u0000\u0000\u0000\uCCC1\u0000\uD2ED\u0000\u0000\uDBB2\uFBC4" + //  7820 -  7829
                "\u0000\uF3E3\u0000\uD9A5\uFBE7\uBDF4\uBDF5\u0000\u0000\u0000" + //  7830 -  7839
                "\uBDF6\u0000\u0000\uB0B0\uB0B1\uB0B2\uB0B3\uB0B4\u0000\u0000" + //  7840 -  7849
                "\uEDBE\u0000\u0000\u0000\uD5C0\uE3F0\uEDFA\uBDF2\u0000\uBDF3" + //  7850 -  7859
                "\u0000\u0000\u0000\u0000\u0000\uBCD2\uBCD3\uBCD4\u0000\uF2A9" + //  7860 -  7869
                "\uF0C4\u0000\u0000\uE2E2\uE9EF\u0000\u0000\uDBB1\u0000\u0000" + //  7870 -  7879
                "\u0000\uD5E9\u0000\u0000\uDBB0\u0000\u0000\uE5DA\uE3BF\u0000" + //  7880 -  7889
                "\u0000\uE4D6\u0000\u0000\uD0C5\uF4AE\u0000\uDDA8\uBDEE\uBDEF" + //  7890 -  7899
                "\u0000\u0000\uBDF0\u0000\u0000\uBDF1\uBDED\u0000\u0000\u0000" + //  7900 -  7909
                "\u0000\u0000\u0000\u0000\uE5DD\uFDFB\uBDEB\u0000\u0000\u0000" + //  7910 -  7919
                "\uBDEC\u0000\u0000\u0000\uE7DE\u0000\u0000\u0000\uD6D6\uE1CC" + //  7920 -  7929
                "\uBDE6\uBDE7\u0000\u0000\uBDE8\uBDE9\u0000\u0000\uD6C6\u0000" + //  7930 -  7939
                "\u0000\u0000\u0000\u0000\uE0E5\uBDE4\u0000\uBDE5\u0000\u0000" + //  7940 -  7949
                "\u0000\u0000\u0000\uBCC9\uBCCA\u0000\uBCCB\uBDE1\uBDE2\u0000" + //  7950 -  7959
                "\u0000\uBDE3\u0000\u0000\u0000\uD7F4\uF0DD\u0000\u0000\u0000" + //  7960 -  7969
                "\uCEAB\uBDDE\uBDDF\u0000\u0000\u0000\u0000\u0000\u0000\uE1B8" + //  7970 -  7979
                "\u0000\uE8F4\uD3FD\uBDDA\u0000\u0000\u0000\uBDDB\u0000\u0000" + //  7980 -  7989
                "\u0000\uF0B1\u0000\u0000\u0000\u0000\u0000\uB4FD\uB4FE\u0000" + //  7990 -  7999
                "\uB5A1\uBDD3\uBDD4\u0000\u0000\uBDD5\uBDD6\u0000\u0000\uFDD9" + //  8000 -  8009
                "\u0000\u0000\uCCA3\u0000\u0000\u0000\uE1E5\uEEF9\u0000\u0000" + //  8010 -  8019
                "\u0000\uE7F6\uBDD2\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  8020 -  8029
                "\uE6A6\u0000\uD9E9\u0000\u0000\uD0FE\u0000\uECED\uD3A9\u0000" + //  8030 -  8039
                "\uA1BC\uA1BD\u0000\uA1EB\uA1B2\uA1B3\u0000\u0000\uEEF3\uE7F1" + //  8040 -  8049
                "\u0000\uFDB4\u0000\u0000\u0000\uE6B9\u0000\u0000\u0000\u0000" + //  8050 -  8059
                "\uC5F0\u0000\u0000\u0000\uD7AA\u0000\u0000\u0000\u0000\uC5F9" + //  8060 -  8069
                "\uC5FA\u0000\uC5FB\uBDCE\uBDCF\u0000\uBDD0\uBDD1\u0000\u0000" + //  8070 -  8079
                "\u0000\uE2BE\u0000\u0000\u0000\u0000\u0000\uB4F5\uB4F6\uB4F7" + //  8080 -  8089
                "\u0000\uE3E6\u0000\u0000\u0000\u0000\u0000\u0000\uD3A8\uBDC5" + //  8090 -  8099
                "\u0000\u0000\uBDC6\uBDC7\u0000\u0000\u0000\uF8E4\u0000\u0000" + //  8100 -  8109
                "\u0000\u0000\u0000\uB4F4\u0000\u0000\u0000\uF3FC\u0000\u0000" + //  8110 -  8119
                "\uEEA2\u0000\u0000\uF9E9\u0000\u0000\u0000\uE7A4\u0000\uD6E3" + //  8120 -  8129
                "\uBDBC\u0000\u0000\u0000\uBDBD\uBDBE\u0000\u0000\uD9D0\u0000" + //  8130 -  8139
                "\u0000\u0000\u0000\u0000\uE5A3\uBDB7\u0000\u0000\uBDB8\u0000" + //  8140 -  8149
                "\uBDB9\u0000\u0000\uCCBB\u0000\u0000\u0000\u0000\u0000\u0000" + //  8150 -  8159
                "\uA6A3\uA6C8\uA6C7\uA6AE\uBDB6\u0000\u0000\u0000\u0000\u0000" + //  8160 -  8169
                "\u0000\u0000\uF5A8\u0000\uDDEC\uDAE8\u0000\u0000\u0000\u0000" + //  8170 -  8179
                "\u0000\uD4E0\uBDB4\uBDB5\u0000\u0000\u0000\u0000\u0000\u0000" + //  8180 -  8189
                "\uCEC0\uE3D4\uD1CF\uF1F5\uBDAE\u0000\u0000\u0000\uBDAF\u0000" + //  8190 -  8199
                "\u0000\u0000\uF1F1\u0000\u0000\u0000\u0000\u0000\uB4BD\uB4BE" + //  8200 -  8209
                "\u0000\u0000\uE4BC\u0000\u0000\u0000\u0000\u0000\u0000\uA9CD" + //  8210 -  8219
                "\uA9CE\uA9CF\uA9D0\uBDAA\u0000\u0000\u0000\u0000\uBDAB\u0000" + //  8220 -  8229
                "\u0000\uCFCD\u0000\u0000\u0000\u0000\u0000\u0000\uA9E7\uA9E8" + //  8230 -  8239
                "\uA9E9\uA9EA\uBDA9\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  8240 -  8249
                "\uD8BE\u0000\uE2A6\u0000\u0000\u0000\u0000\u0000\uE0C0\u0000" + //  8250 -  8259
                "\uA1B4\uA1B5\uA1B6\uA1B7\uA1B8\uA1B9\uA1BA\uA1BB\uBDA6\uBDA7" + //  8260 -  8269
                "\u0000\u0000\uBDA8\u0000\u0000\u0000\uD5D1\u0000\uD8F0\uF8C3" + //  8270 -  8279
                "\uEAD7\u0000\uFCF9\u0000\u0000\u0000\u0000\uDFF3\uCEE7\uDAC2" + //  8280 -  8289
                "\uBDA5\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF2D8\u0000" + //  8290 -  8299
                "\uEAF2\uCBC7\u0000\uCDF4\u0000\u0000\uDBAF\uEFD9\uBCFB\uBCFC" + //  8300 -  8309
                "\u0000\uBCFD\u0000\uBCFE\u0000\uBDA1\uBCFA\u0000\u0000\u0000" + //  8310 -  8319
                "\u0000\u0000\u0000\u0000\uD9CD\u0000\uE5C1\u0000\u0000\u0000" + //  8320 -  8329
                "\u0000\uE0EE\u0000\u0000\uD6F9\u0000\uCDD7\uDED8\u0000\u0000" + //  8330 -  8339
                "\uF2F8\uBCF6\uBCF7\u0000\u0000\uBCF8\u0000\u0000\uBCF9\uBCF0" + //  8340 -  8349
                "\u0000\u0000\u0000\uBCF1\u0000\u0000\u0000\uD8D3\u0000\u0000" + //  8350 -  8359
                "\u0000\u0000\u0000\uB4B4\u0000\u0000\u0000\uD6F6\u0000\u0000" + //  8360 -  8369
                "\u0000\uEACA\u0000\uDAE7\u0000\u0000\u0000\uF7CC\u0000\u0000" + //  8370 -  8379
                "\u0000\uFDBA\u0000\u0000\uFDE1\uF6FE\uBCEB\uBCEC\u0000\uBCED" + //  8380 -  8389
                "\u0000\u0000\u0000\u0000\uF1A9\u0000\u0000\u0000\uF2C9\uBCEA" + //  8390 -  8399
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF9AB\u0000\uDBFB" + //  8400 -  8409
                "\u0000\uCBE0\u0000\u0000\u0000\u0000\u0000\uD4B7\u0000\u0000" + //  8410 -  8419
                "\u0000\uD8F2\u0000\u0000\uE3CA\u0000\uA1A1\uA1A2\uA1A3\uA1A8" + //  8420 -  8429
                "\u0000\u0000\u0000\u0000\uBEAE\u0000\u0000\u0000\uF9E3\u0000" + //  8430 -  8439
                "\u0000\u0000\u0000\uC3E3\uC3E4\u0000\uC3E5\uBCE8\u0000\u0000" + //  8440 -  8449
                "\u0000\uBCE9\u0000\u0000\u0000\uF9FE\uDBBA\uDAF5\u0000\u0000" + //  8450 -  8459
                "\u0000\uE7EC\uEEEE\u0000\uF3F0\u0000\uDFBF\uBCE7\u0000\u0000" + //  8460 -  8469
                "\u0000\u0000\u0000\u0000\u0000\uCEEB\u0000\uF3BC\u0000\uDAD2" + //  8470 -  8479
                "\u0000\u0000\u0000\u0000\u0000\uFCA8\u0000\u0000\uECE6\uBCE3" + //  8480 -  8489
                "\u0000\u0000\u0000\uBCE4\u0000\u0000\u0000\uE0A2\u0000\u0000" + //  8490 -  8499
                "\u0000\u0000\u0000\uB4B3\u0000\u0000\u0000\uF1AE\u0000\uEFCE" + //  8500 -  8509
                "\u0000\u0000\u0000\uD7FA\u0000\u0000\u0000\uFBC8\uBCE0\u0000" + //  8510 -  8519
                "\u0000\u0000\u0000\u0000\u0000\u0000\uCBB2\uF1C2\uBCDD\uBCDE" + //  8520 -  8529
                "\u0000\u0000\uBCDF\u0000\u0000\u0000\uD3E6\uF2DD\uCFBF\u0000" + //  8530 -  8539
                "\uEBAC\u0000\uD6CE\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + //  8540 -  8549
                "\uD9DA\uBCD5\u0000\u0000\u0000\uBCD6\u0000\uBCD7\u0000\uD9B8" + //  8550 -  8559
                "\uD9B9\uEFC9\u0000\u0000\u0000\u0000\u0000\uF0A1\u0000\uDEAA" + //  8560 -  8569
                "\u0000\uA2CD\uA2DB\uA2DC\u0000\uA2DD\uA2DA\u0000\u0000\uE1B1" + //  8570 -  8579
                "\u0000\u0000\uF7B2\u0000\u0000\uD3F3\uBCD0\u0000\u0000\u0000" + //  8580 -  8589
                "\u0000\u0000\u0000\u0000\uE8B7\u0000\uF5C4\u0000\u0000\u0000" + //  8590 -  8599
                "\u0000\u0000\u0000\u0000\uDCE9\uBCCE\u0000\u0000\u0000\uBCCF" + //  8600 -  8609
                "\u0000\u0000\u0000\uEBEC\u0000\u0000\uD3C5\uFCEC\uD2DB\uBCCC" + //  8610 -  8619
                "\uBCCD\u0000\u0000\u0000\u0000\u0000\u0000\uFCE2\u0000\uF8A7" + //  8620 -  8629
                "\u0000\uEED7\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCED7" + //  8630 -  8639
                "\uBCC7\u0000\u0000\u0000\uBCC8\u0000\u0000\u0000\uDDB1\u0000" + //  8640 -  8649
                "\uD8AF\uE3A3\u0000\u0000\uD7EF\u0000\u0000\u0000\u0000\u0000" + //  8650 -  8659
                "\u0000\uA9F9\u0000\u0000\u0000\uCBAD\u0000\uF9B0\u0000\u0000" + //  8660 -  8669
                "\u0000\uE9E3\u0000\uEDCB\uCFE4\u0000\uF0AB\u0000\u0000\u0000" + //  8670 -  8679
                "\u0000\u0000\uEBE7\u0000\uA2BC\uA2BD\u0000\uA2C0\uA2BB\uA2BE" + //  8680 -  8689
                "\u0000\uA2BF\uBCC0\uBCC1\u0000\uBCC2\uBCC3\uBCC4\u0000\u0000" + //  8690 -  8699
                "\uF8EA\u0000\u0000\u0000\u0000\u0000\u0000\uC8BC\u0000\uC8BD" + //  8700 -  8709
                "\uBCBF\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF2EF\u0000" + //  8710 -  8719
                "\uCAD1\u0000\u0000\u0000\uEAF1\u0000\uD0A6\u0000\uA1CF\u0000" + //  8720 -  8729
                "\uA1CE\u0000\u0000\u0000\u0000\u0000\uFCEE\u0000\u0000\u0000" + //  8730 -  8739
                "\uFCF4\u0000\u0000\u0000\u0000\uC3D6\u0000\u0000\u0000\uDEB1" + //  8740 -  8749
                "\u0000\u0000\u0000\u0000\uC3DF\uC3E0\u0000\u0000\uDCC2\u0000" + //  8750 -  8759
                "\u0000\uF0A7\u0000\u0000\uE6C0\uBCBC\uBCBD\u0000\u0000\uBCBE" + //  8760 -  8769
                "\u0000\u0000\u0000\uF2B7\u0000\u0000\uFAF6\uF6AA\uFAF7\uBCB9" + //  8770 -  8779
                "\uBCBA\u0000\u0000\u0000\u0000\uBCBB\u0000\uF5CF\uE5F3\uF0C2" + //  8780 -  8789
                "\u0000\u0000\u0000\u0000\u0000\uDDBD\u0000\u0000\u0000\uD8C1" + //  8790 -  8799
                "\u0000\u0000\u0000\u0000\uF1FD\u0000\uDEBF\uFBBA\uF9B9\uBCB1" + //  8800 -  8809
                "\u0000\u0000\uBCB2\uBCB3\u0000\uBCB4\uBCB5\uBCAB\u0000\u0000" + //  8810 -  8819
                "\u0000\u0000\uBCAC\u0000\u0000\uCEDA\uFBEB\uDBA6\uDBDE\uD8E5" + //  8820 -  8829
                "\u0000\u0000\uD9A1\u0000\uD8C0\uDCDB\u0000\u0000\uEDBD\uBCAA" + //  8830 -  8839
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD0C8\u0000\uD5D3" + //  8840 -  8849
                "\uF3F5\uF7AE\u0000\u0000\uEFC8\u0000\uCDF3\uBCA8\u0000\u0000" + //  8850 -  8859
                "\u0000\uBCA9\u0000\u0000\u0000\uDBCF\uCBA4\u0000\u0000\uF8E0" + //  8860 -  8869
                "\u0000\uCFDE\u0000\u0000\uCED0\u0000\u0000\u0000\u0000\uC2D6" + //  8870 -  8879
                "\u0000\u0000\u0000\uD3D9\u0000\u0000\u0000\u0000\uA8FB\uA8FC" + //  8880 -  8889
                "\uA8FD\uA8FE\u0000\uA2C4\uA2C5\u0000\u0000\u0000\u0000\u0000" + //  8890 -  8899
                "\u0000\uF4AA\uE9AF\u0000\u0078\u0079\u007A\u007B\u007C\u007D" + //  8900 -  8909
                "\u007E\u007F\uBCA2\u0000\u0000\u0000\uBCA3\u0000\u0000\u0000" + //  8910 -  8919
                "\uF8CB\u0000\u0000\u0000\u0000\u0000\uB4A9\uB4AA\u0000\u0000" + //  8920 -  8929
                "\uFCBE\uD5F1\u0000\u0000\u0000\u0000\u0000\uB9FC\uB9FD\u0000" + //  8930 -  8939
                "\uB9FE\uBBF9\uBBFA\u0000\uBBFB\uBBFC\uBBFD\u0000\u0000\uF8BE" + //  8940 -  8949
                "\u0000\u0000\u0000\u0000\u0000\u0000\uC7E2\u0000\u0000\uDBBC" + //  8950 -  8959
                "\u0000\u0000\u0000\u0000\u0000\u0000\uFCBF\u0000\u0000\uE8C0" + //  8960 -  8969
                "\uE8C1\u0000\u0000\u0000\uCFE3\uE9A2\uBBF8\u0000\u0000\u0000" + //  8970 -  8979
                "\u0000\u0000\u0000\u0000\uCECB\u0000\uE2A3\uD3FC\u0000\u0000" + //  8980 -  8989
                "\uEDE6\u0000\u0000\u0000\uCDF7\uF0DE\u0000\u0000\u0000\uD3B1" + //  8990 -  8999
                "\u0000\u0000\u0000\u0000\uA1CA\u0000\u0000\u0000\u0000\uBEA2" + //  9000 -  9009
                "\uBEA3\u0000\u0000\uCFAD\u0000\u0000\u0000\u0000\uE7F9\uF8A8" + //  9010 -  9019
                "\uBBF5\uBBF6\u0000\u0000\uBBF7\u0000\u0000\u0000\uD6FD\u0000" + //  9020 -  9029
                "\u0000\u0000\u0000\u0000\uB3FD\uB3FE\u0000\uB4A1\uBBF2\uBBF3" + //  9030 -  9039
                "\u0000\u0000\u0000\uBBF4\u0000\u0000\uDEBE\u0000\u0000\u0000" + //  9040 -  9049
                "\u0000\u0000\u0000\uC7A8\u0000\uC7A9\uBBEA\u0000\u0000\uBBEB" + //  9050 -  9059
                "\uBBEC\uBBED\uBBEE\u0000\uDED7\u0000\u0000\u0000\u0000\u0000" + //  9060 -  9069
                "\uCBDF\u0000\uA2C2\uA2C1\u0000\uA1DB\u0000\u0000\uA1DD\uA1DC" + //  9070 -  9079
                "\uBBE3\uBBE4\u0000\uBBE5\u0000\uBBE6\u0000\u0000\uD5AD\u0000" + //  9080 -  9089
                "\u0000\u0000\u0000\u0000\u0000\uC7A2\u0000\u0000\uD9D5\u0000" + //  9090 -  9099
                "\u0000\uDFAA\u0000\u0000\u0000\uD0E7\u0000\u0000\uECFD\u0000" + //  9100 -  9109
                "\uD2AE\uBBE2\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uFCC2" + //  9110 -  9119
                "\uEABB\uBBDF\uBBE0\u0000\u0000\uBBE1\u0000\u0000\u0000\uE3A7" + //  9120 -  9129
                "\u0000\uDFD6\uFDE8\u0000\u0000\uDDBF\u0000\u0000\u0000\uF6EF" + //  9130 -  9139
                "\u0000\uDEF9\uBBDD\uBBDE\u0000\u0000\u0000\u0000\u0000\u0000" + //  9140 -  9149
                "\uD0FB\uECDB\u0000\u0000\uDAE0\u0000\u0000\u0000\u0000\u0000" + //  9150 -  9159
                "\u0000\uC5CC\u0000\u0000\uE7A8\u0000\u0000\u0000\u0000\u0000" + //  9160 -  9169
                "\u0000\uF5E3\u0000\u0000\uD4CD\u0000\u0000\uF3B8\u0000\u0000" + //  9170 -  9179
                "\u0000\uA2A1\u0000\uA2A2\u0000\u0000\u0000\uCFB0\uF7D9\uF3E6" + //  9180 -  9189
                "\u0000\u0000\uCDF5\u0000\u0000\u0000\uFDB0\uD5A8\u0000\uE6C8" + //  9190 -  9199
                "\u0000\u0000\u0000\u0000\uF8DA\u0000\u0000\uD3E4\uF6F7\u0000" + //  9200 -  9209
                "\u0000\uD5BA\uF3CD\uCBE1\uBBDC\u0000\u0000\u0000\u0000\u0000" + //  9210 -  9219
                "\u0000\u0000\uEBA2\uEBA3\uBBDA\u0000\u0000\u0000\uBBDB\u0000" + //  9220 -  9229
                "\u0000\u0000\uECCD\u0000\uECCE\u0000\uD6BF\u0000\uCFAC\uF0F0" + //  9230 -  9239
                "\u0000\uF4FD\uDBC8\u0000\u0000\u0000\uECE2\u0000\u0000\u0000" + //  9240 -  9249
                "\u0000\uA7A4\u0000\u0000\uA2E0\u0000\uA2B8\uA2B7\u0000\u0000" + //  9250 -  9259
                "\u0000\u0000\uA1DF\uA1DE\uBBD3\u0000\u0000\u0000\uBBD4\u0000" + //  9260 -  9269
                "\u0000\u0000\uD6BE\u0000\u0000\u0000\uE2BA\u0000\uCED8\u0000" + //  9270 -  9279
                "\uCBDE\uF4AC\uDAFB\u0000\uF6E9\uE8F3\uBBCF\u0000\u0000\u0000" + //  9280 -  9289
                "\u0000\u0000\u0000\u0000\uF3E9\u0000\uD2B9\uD5C3\u0000\u0000" + //  9290 -  9299
                "\uDAE5\uDAD0\u0000\uD1D9\uBBCB\uBBCC\u0000\u0000\u0000\uBBCD" + //  9300 -  9309
                "\u0000\u0000\uECF1\u0000\u0000\u0000\u0000\u0000\u0000\uC4FA" + //  9310 -  9319
                "\u0000\u0000\uFDC0\u0000\u0000\u0000\u0000\u0000\u0000\uE7A3" + //  9320 -  9329
                "\u0000\u0000\uE8B1\u0000\uFCAE\u0000\u0000\u0000\u0000\uC6EC" + //  9330 -  9339
                "\u0000\u0000\u0000\uA8AB\uA9AB\u0000\u0000\u0000\u0000\uBDEA" + //  9340 -  9349
                "\u0000\u0000\u0000\uEEA9\uF6BC\u0000\u0000\uCCDB\uBBCA\u0000" + //  9350 -  9359
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE4E4\u0000\uEFB9\u0000" + //  9360 -  9369
                "\u0000\uF8D8\u0000\u0000\u0000\u0000\uC2C9\uC2CA\u0000\u0000" + //  9370 -  9379
                "\uF7CB\uDFAE\uE8F5\u0000\u0000\u0000\u0000\uC2E8\u0000\u0000" + //  9380 -  9389
                "\u0000\uDFD7\uDBD0\uDBD1\u0000\u0000\uE8B6\u0000\u0000\uD6CF" + //  9390 -  9399
                "\uF4BA\u0000\uF7C9\uBBC7\uBBC8\u0000\u0000\uBBC9\u0000\u0000" + //  9400 -  9409
                "\u0000\uF1AA\uCED1\u0000\u0000\uF6C7\u0000\uD0FC\u0000\u0000" + //  9410 -  9419
                "\u0000\uF4FC\u0000\u0000\u0000\uE9FE\u0000\u0000\u0000\u0000" + //  9420 -  9429
                "\uC8F4\uC8F5\u0000\u0000\uD8A4\u0000\u0000\u0000\u0000\u0000" + //  9430 -  9439
                "\uF2A7\uBBC2\uBBC3\u0000\uBBC4\uBBC5\uBBC6\u0000\u0000\uF6EE" + //  9440 -  9449
                "\u0000\uF6CC\uE2F8\u0000\u0000\u0000\uF8C9\u0000\u0000\u0000" + //  9450 -  9459
                "\u0000\uE3D2\uBBC0\uBBC1\u0000\u0000\u0000\u0000\u0000\u0000" + //  9460 -  9469
                "\uFCC6\u0000\u0000\u0000\uF3C3\u0000\u0000\u0000\u0000\u0000" + //  9470 -  9479
                "\uB3BF\uB3C0\u0000\uB3C1\uBBBA\u0000\u0000\uBBBB\uBBBC\uBBBD" + //  9480 -  9489
                "\u0000\u0000\uD1C7\uE9AE\u0000\uE8BD\u0000\u0000\uFAC4\uBBB9" + //  9490 -  9499
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD5B3\u0000\uF6F5" + //  9500 -  9509
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE4E8\uBBB5\uBBB6" + //  9510 -  9519
                "\u0000\u0000\uBBB7\u0000\u0000\uBBB8\uBBB4\u0000\u0000\u0000" + //  9520 -  9529
                "\u0000\u0000\u0000\u0000\uD5EB\u0000\uFBC3\uDDEA\u0000\uE2A2" + //  9530 -  9539
                "\u0000\uEEA6\u0000\u0000\uE4C4\u0000\u0000\u0000\u0000\u0000" + //  9540 -  9549
                "\u0000\uF5C5\uEEE0\u0000\uA2C9\uA2CC\u0000\u0000\u0000\u0000" + //  9550 -  9559
                "\u0000\u0000\uE4C3\u0000\u0000\uDBC6\uD0F1\u0000\uD0F2\u0000" + //  9560 -  9569
                "\u0000\u0000\uF6DA\u0000\u0000\u0000\u0000\uC5DB\uC5DC\u0000" + //  9570 -  9579
                "\uC5DD\uBBB2\uBBB3\u0000\u0000\u0000\u0000\u0000\u0000\uE1D5" + //  9580 -  9589
                "\u0000\u0000\uD4EA\uBBB0\uBBB1\u0000\u0000\u0000\u0000\u0000" + //  9590 -  9599
                "\u0000\uE9D6\u0000\u0000\u0000\uD1CC\uDDFC\u0000\u0000\u0000" + //  9600 -  9609
                "\u0000\uC7C4\uC7C5\u0000\uC7C6\uBBAB\u0000\u0000\u0000\uBBAC" + //  9610 -  9619
                "\u0000\u0000\u0000\uF0FB\uFAE1\uF0DA\uCCE7\uDAF4\u0000\uE1B6" + //  9620 -  9629
                "\uF8B7\u0000\u0000\u0000\u0000\u0000\uE0BF\uBBA3\uBBA4\u0000" + //  9630 -  9639
                "\uBBA5\uBBA6\uBBA7\u0000\u0000\uD0CF\u0000\uCFFA\uF3CA\uE0D7" + //  9640 -  9649
                "\u0000\u0000\uEEC9\u0000\u0000\u0000\uE2DD\u0000\u0000\uDFE2" + //  9650 -  9659
                "\uE7DB\u0000\u0000\u0000\u0000\u0000\uB2CF\uB2D0\u0000\u0000" + //  9660 -  9669
                "\uDAC8\uDFA6\u0000\uF9B3\uF2D2\u0000\uCAC4\uBBA1\u0000\uBBA2" + //  9670 -  9679
                "\u0000\u0000\u0000\u0000\u0000\uBCC5\uBCC6\u0000\u0000\uFAC3" + //  9680 -  9689
                "\uE5D7\u0000\uECC8\u0000\u0000\u0000\uCBEF\uFCDF\u0000\u0000" + //  9690 -  9699
                "\u0000\u0000\uC1EA\u0000\u0000\u0000\uE3ED\u0000\uE8C2\u0000" + //  9700 -  9709
                "\uEDF5\uBAFC\uBAFD\u0000\u0000\uBAFE\u0000\u0000\u0000\uFBD5" + //  9710 -  9719
                "\u0000\u0000\u0000\u0000\u0000\uB2EE\u0000\u0000\u0000\uCDFA" + //  9720 -  9729
                "\u0000\u0000\u0000\u0000\u0000\uF9ED\u0000\u0000\u0000\uD8C8" + //  9730 -  9739
                "\u0000\u0000\uEEC1\u0000\uEEFA\uFDF4\u0000\u0000\uD3E3\u0000" + //  9740 -  9749
                "\uFBC2\u0000\uA1E1\uA1E0\u0000\uA2C3\uA2C7\uA2C8\uA2CB\uA2CA" + //  9750 -  9759
                "\uBAF3\u0000\u0000\u0000\uBAF4\u0000\uBAF5\u0000\uE0F7\uE4B2" + //  9760 -  9769
                "\uCCFC\u0000\u0000\u0000\uFBE4\u0000\uA6E2\uA6E3\uA6E4\uA6B6" + //  9770 -  9779
                "\u0000\u0000\u0000\u0000\uBEAD\u0000\u0000\u0000\uD0A7\u0000" + //  9780 -  9789
                "\uF0CB\u0000\uD0C7\uBAEC\u0000\u0000\u0000\uBAED\u0000\u0000" + //  9790 -  9799
                "\u0000\uD8A3\u0000\u0000\uDAD9\u0000\uF0D8\uBAE7\u0000\u0000" + //  9800 -  9809
                "\uBAE8\u0000\uBAE9\u0000\u0000\uE1E2\uD1C6\u0000\u0000\u0000" + //  9810 -  9819
                "\u0000\u0000\uB9F6\uB9F7\u0000\u0000\uDCCC\u0000\uD0CB\u0000" + //  9820 -  9829
                "\u0000\u0000\uFCA4\uBAE6\u0000\u0000\u0000\u0000\u0000\u0000" + //  9830 -  9839
                "\u0000\uFCE8\uF3BD\uBAE4\u0000\u0000\u0000\uBAE5\u0000\u0000" + //  9840 -  9849
                "\u0000\uEDD8\uE1C7\u0000\u0000\u0000\u0000\uC6F7\uC6F8\u0000" + //  9850 -  9859
                "\u0000\uF7C7\u0000\u0000\u0000\u0000\u0000\u0000\uC2C6\u0000" + //  9860 -  9869
                "\u0000\uD6E1\uCFD2\u0000\uD0B6\u0000\u0000\u0000\uE9D0\u0000" + //  9870 -  9879
                "\u0000\u0000\u0000\u0000\uFABC\uE6E2\u0000\u0000\uF6C4\u0000" + //  9880 -  9889
                "\u0000\u0000\uEEDD\uE7C4\u0000\uCDFC\u0000\uD9E6\u0000\uE2F9" + //  9890 -  9899
                "\uE2A1\uEBD4\u0000\uA6DB\uA6DC\uA6C0\uA6DD\uA6DE\uA6DF\uA6E0" + //  9900 -  9909
                "\uA6E1\uBAE1\u0000\u0000\u0000\uBAE2\u0000\u0000\u0000\uE3D9" + //  9910 -  9919
                "\u0000\u0000\u0000\u0000\u0000\uB2E3\u0000\u0000\u0000\uA1EC" + //  9920 -  9929
                "\uA1ED\u0000\u0000\u0000\u0000\uBEA5\uBEA6\u0000\u0000\uFAD4" + //  9930 -  9939
                "\u0000\u0000\u0000\uECE5\u0000\u0000\uD7E9\uEDF6\u0000\u0000" + //  9940 -  9949
                "\u0000\uDEED\u0000\uD5FC\u0000\u0000\u0000\u0000\u0000\u0000" + //  9950 -  9959
                "\uD3D4\uBADE\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uFBC9" + //  9960 -  9969
                "\u0000\uFAD8\u0000\uF3D5\u0000\uCFAB\u0000\u0000\uEBF3\uBADD" + //  9970 -  9979
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDBF1\u0000\uFAAB" + //  9980 -  9989
                "\uEBEB\uE7F8\uD9E5\u0000\u0000\u0000\u0000\uC2C0\uC2C1\u0000" + //  9990 -  9999
                "\uC2C2\uBAD5\uBAD6\u0000\uBAD7\u0000\uBAD8\u0000\u0000\uF4EC" + // 10000 - 10009
                "\u0000\u0000\u0000\u0000\uEFFE\u0000\uFDF3\uFDF2\uF7A6\u0000" + // 10010 - 10019
                "\u0000\u0000\u0000\u0000\uF9EF\uCFF4\uF7E6\u0000\uA6BF\uA6D7" + // 10020 - 10029
                "\uA6D8\uA6B5\uA6AB\uA6D9\uA6DA\uA6BB\uBAD2\uBAD3\uBAD4\u0000" + // 10030 - 10039
                "\u0000\u0000\u0000\u0000\uBCB6\uBCB7\u0000\uBCB8\uBACE\uBACF" + // 10040 - 10049
                "\u0000\u0000\uBAD0\u0000\u0000\uBAD1\uBACD\u0000\u0000\u0000" + // 10050 - 10059
                "\u0000\u0000\u0000\u0000\uD6B1\uDEB2\uBACA\uBACB\u0000\u0000" + // 10060 - 10069
                "\u0000\u0000\u0000\u0000\uE0F4\u0000\uCFE0\u0000\uD5FB\uDEBB" + // 10070 - 10079
                "\u0000\u0000\uF4FB\u0000\u0000\u0000\uFBD9\uEDF7\u0000\u0000" + // 10080 - 10089
                "\uE5B5\uBAC9\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE3D6" + // 10090 - 10099
                "\u0000\uEAD0\u0000\u0000\u0000\u0000\u0000\uCCD4\uCBAF\uBAC6" + // 10100 - 10109
                "\uBAC7\u0000\u0000\uBAC8\u0000\u0000\u0000\uE2B5\u0000\u0000" + // 10110 - 10119
                "\uCAE6\u0000\uD3AE\uBAC5\u0000\u0000\u0000\u0000\u0000\u0000" + // 10120 - 10129
                "\u0000\uFDE4\uFACE\uBAC1\u0000\u0000\u0000\uBAC2\u0000\u0000" + // 10130 - 10139
                "\u0000\uEFBC\uD8A1\u0000\u0000\u0000\u0000\uC6EF\uC6F0\u0000" + // 10140 - 10149
                "\u0000\uDCA4\u0000\u0000\u0000\uF0B8\u0000\u0000\uECBF\uFCD8" + // 10150 - 10159
                "\u0000\u0000\u0000\u0000\u0000\uB4BA\uB4BB\u0000\u0000\uF2F1" + // 10160 - 10169
                "\u0000\u0000\u0000\u0000\u0000\u0000\uEFF7\u0000\u0000\uE8AC" + // 10170 - 10179
                "\u0000\uE8DD\u0000\u0000\uEFE9\u0000\uCBC6\u0000\u0000\uF0F6" + // 10180 - 10189
                "\u0000\u0000\uD5E7\u0000\uA6BD\uA6D3\uA6D4\uA6B3\uA6AA\uA6D5" + // 10190 - 10199
                "\uA6D6\uA6BA\uBABB\u0000\u0000\u0000\uBABC\u0000\u0000\u0000" + // 10200 - 10209
                "\uF2F2\uF3EB\u0000\uF0D7\u0000\u0000\uE7CE\u0000\u0000\uDFDC" + // 10210 - 10219
                "\u0000\uF9C7\u0000\uF3D3\uF3D4\u0000\u0000\u0000\uF7E4\u0000" + // 10220 - 10229
                "\uF7D1\uBAB6\u0000\u0000\u0000\uBAB7\u0000\u0000\u0000\uE0C9" + // 10230 - 10239
                "\u0000\u0000\u0000\uD6C9\u0000\uEDB1\u0000\uCCC3\uF7BE\uFCCB" + // 10240 - 10249
                "\u0000\u0000\u0000\uCEA2\u0000\u0000\uF3EE\u0000\uA6B9\uA6CF" + // 10250 - 10259
                "\uA6D0\uA6B4\uA6A8\uA6D1\uA6D2\uA6B8\uBAB3\uBAB4\u0000\u0000" + // 10260 - 10269
                "\u0000\uBAB5\u0000\u0000\uEFC5\u0000\uE7E7\u0000\u0000\uD7FD" + // 10270 - 10279
                "\u0000\uFCCA\u0000\u0000\u0000\uF3E1\u0000\u0000\uCBC4\uBAAF" + // 10280 - 10289
                "\u0000\u0000\u0000\uBAB0\u0000\u0000\u0000\uD1B8\u0000\u0000" + // 10290 - 10299
                "\u0000\u0000\uD6DF\uBAA8\uBAA9\u0000\uBAAA\uBAAB\uBAAC\u0000" + // 10300 - 10309
                "\u0000\uDADF\u0000\uEFB3\u0000\u0000\u0000\u0000\uFAA7\u0000" + // 10310 - 10319
                "\u0000\u0000\u0000\uC0EB\uC0EC\u0000\uC0ED\uBAA7\u0000\u0000" + // 10320 - 10329
                "\u0000\u0000\u0000\u0000\u0000\uFACB\u0000\uEAEF\uEAF0\u0000" + // 10330 - 10339
                "\u0000\u0000\uDAC0\uF8B4\uEBF2\uBAA3\uBAA4\u0000\u0000\uBAA5" + // 10340 - 10349
                "\u0000\u0000\uBAA6\uB9F8\u0000\u0000\uB9F9\uB9FA\u0000\uB9FB" + // 10350 - 10359
                "\u0000\uF6E7\uD2DD\u0000\uDFCC\u0000\u0000\uFCC9\u0000\uA6B7" + // 10360 - 10369
                "\uA6CB\uA6CC\uA6B2\uA6A9\uA6BE\uA6CD\uA6CE\uB9F4\u0000\u0000" + // 10370 - 10379
                "\u0000\u0000\u0000\u0000\u0000\uE7B4\u0000\uFBB3\uE4C2\u0000" + // 10380 - 10389
                "\u0000\u0000\u0000\u0000\u0000\uEFA9\u0000\u0000\uD3FB\u0000" + // 10390 - 10399
                "\u0000\u0000\u0000\u0000\u0000\uE8B2\u0000\u0000\uF3B5\u0000" + // 10400 - 10409
                "\u0000\uF8A4\u0000\u0000\uD1F3\uB9EC\uB9ED\u0000\uB9EE\uB9EF" + // 10410 - 10419
                "\uB9F0\u0000\u0000\uF6CB\u0000\uF1E6\uEDC1\uE8BC\uEED1\u0000" + // 10420 - 10429
                "\uCCFB\u0000\u0000\u0000\uD3FA\uF4A4\u0000\u0000\uE4CE\u0000" + // 10430 - 10439
                "\u0000\u0000\u0000\u0000\u0000\uF1D4\u0000\uEDF2\uB9EB\u0000" + // 10440 - 10449
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF7D2\u0000\uF0F5\u0000" + // 10450 - 10459
                "\uDDE8\uD3ED\uF5FC\u0000\uDABF\u0000\uA6A5\uA6C4\uA6C3\uA6B0" + // 10460 - 10469
                "\uA6A7\uA6BC\uA6C9\uA6CA\uB9E8\uB9E9\u0000\u0000\uB9EA\u0000" + // 10470 - 10479
                "\u0000\u0000\uF0D4\uCBE4\uFBD4\uF5E6\uE3EA\u0000\uEAED\u0000" + // 10480 - 10489
                "\u0000\uFCB4\uF5C2\u0000\u0000\uD7DC\uB9DD\u0000\u0000\uB9DE" + // 10490 - 10499
                "\uB9DF\uB9E0\uB9E1\uB9E2\uB9D2\uB9D3\u0000\uB9D4\uB9D5\uB9D6" + // 10500 - 10509
                "\u0000\uB9D7\uB9D0\u0000\uB9D1\u0000\u0000\u0000\u0000\u0000" + // 10510 - 10519
                "\uBCA4\uBCA5\u0000\uBCA6\uB9CC\uB9CD\u0000\u0000\uB9CE\u0000" + // 10520 - 10529
                "\u0000\uB9CF\uB9CA\u0000\u0000\uB9CB\u0000\u0000\u0000\u0000" + // 10530 - 10539
                "\uF6E8\u0000\u0000\u0000\u0000\uC2C7\u0000\u0000\u0000\uE3C6" + // 10540 - 10549
                "\u0000\u0000\u0000\uDEE4\u0000\uFDE6\u0000\u0000\u0000\u0000" + // 10550 - 10559
                "\u0000\u0000\u0000\uF6BA\uB9C9\u0000\u0000\u0000\u0000\u0000" + // 10560 - 10569
                "\u0000\u0000\uCAD2\u0000\uF9CF\uEBDA\uCAC1\u0000\uD2B8\uCDF1" + // 10570 - 10579
                "\u0000\uE3D3\uB9C7\u0000\u0000\u0000\uB9C8\u0000\u0000\u0000" + // 10580 - 10589
                "\uD5EC\uD5F8\uDAF3\u0000\u0000\u0000\uC8B6\u0000\uC8B7\u0000" + // 10590 - 10599
                "\u0000\uF4C6\u0000\u0000\u0000\u0000\u0000\u0000\uC1D5\u0000" + // 10600 - 10609
                "\u0000\uE2BC\u0000\u0000\uFCED\uECE0\uD2FE\u0000\uCDF0\u0000" + // 10610 - 10619
                "\uF9F6\u0000\u0000\uDFF0\u0000\u0000\uF2A5\u0000\u0000\u0000" + // 10620 - 10629
                "\u0000\u0000\uF2A6\uB9C3\u0000\u0000\u0000\uB9C4\u0000\u0000" + // 10630 - 10639
                "\u0000\uD6F0\uF3AF\u0000\u0000\uCDA5\u0000\uEAB3\u0000\uCED6" + // 10640 - 10649
                "\u0000\u0000\u0000\u0000\uCCA5\uB9C1\u0000\u0000\u0000\u0000" + // 10650 - 10659
                "\u0000\u0000\u0000\uD9D4\uD6E8\uB9BF\u0000\u0000\u0000\uB9C0" + // 10660 - 10669
                "\u0000\u0000\u0000\uFCAC\uFCAD\uD8A7\u0000\u0000\u0000\uC1C1" + // 10670 - 10679
                "\uC1C2\uC1C3\u0000\u0000\uF4C5\uDCA3\u0000\u0000\u0000\u0000" + // 10680 - 10689
                "\u0000\uB8FE\u0000\u0000\u0000\uCBD4\uE0BE\uE3F8\uEAE9\uFCB2" + // 10690 - 10699
                "\u0000\uF7BD\uECAE\u0000\u0000\u0000\uD0E1\u0000\uE0F5\uB9BB" + // 10700 - 10709
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE4F2\u0000\uF5F7" + // 10710 - 10719
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCFA9\uB9B9\u0000" + // 10720 - 10729
                "\u0000\u0000\uB9BA\u0000\u0000\u0000\uF8DF\u0000\uF7F2\u0000" + // 10730 - 10739
                "\u0000\u0000\uC0D2\uC0D3\uC0D4\u0000\uC0D5\uB9AE\u0000\u0000" + // 10740 - 10749
                "\uB9AF\uB9B0\uB9B1\uB9B2\u0000\uD8CE\uD8CF\u0000\u0000\u0000" + // 10750 - 10759
                "\u0000\u0000\u0000\uE6DD\u0000\u0000\uF8D9\u0000\u0000\u0000" + // 10760 - 10769
                "\uEEBD\u0000\u0000\uD8C6\u0000\u0000\uE4E3\uF5CE\u0000\u0000" + // 10770 - 10779
                "\uCDF2\u0000\uCFEB\u0000\u0000\u0000\uCDB8\uB9A8\u0000\u0000" + // 10780 - 10789
                "\u0000\u0000\u0000\u0000\u0000\uDCB5\uE0F8\uB9A6\u0000\u0000" + // 10790 - 10799
                "\u0000\uB9A7\u0000\u0000\u0000\uDDF7\u0000\u0000\u0000\u0000" + // 10800 - 10809
                "\u0000\uB2CD\uB2CE\u0000\u0000\uF9D8\uF9D9\uCAE0\uDACA\u0000" + // 10810 - 10819
                "\u0000\u0000\uDAAC\uEAB0\u0000\u0000\u0000\u0000\uC1C9\uC1CA" + // 10820 - 10829
                "\u0000\u0000\uF1E8\uD9F2\uDBF5\uCAB5\uD9C6\u0000\u0000\uDCBC" + // 10830 - 10839
                "\uD2B6\uF5D5\u0000\u0000\u0000\u0000\uC6DB\uC6DC\u0000\u0000" + // 10840 - 10849
                "\uE2AD\u0000\u0000\u0000\u0000\u0000\u0000\uE3DA\u0000\uFCD9" + // 10850 - 10859
                "\uB9A1\u0000\u0000\u0000\uB9A2\u0000\u0000\u0000\uE3C5\uEBF8" + // 10860 - 10869
                "\u0000\uF2B1\u0000\u0000\uF8CD\u0000\uCBD2\u0000\u0000\u0000" + // 10870 - 10879
                "\uEBCE\uB8FC\uB8FD\u0000\u0000\u0000\u0000\u0000\u0000\uE5F2" + // 10880 - 10889
                "\u0000\u0000\uD0F4\uB8FB\u0000\u0000\u0000\u0000\u0000\u0000" + // 10890 - 10899
                "\u0000\uCDE0\uE3B0\uB8F6\uB8F7\u0000\uB8F8\u0000\uB8F9\u0000" + // 10900 - 10909
                "\u0000\uCAFC\uCAFD\u0000\u0000\u0000\u0000\u0000\uB8EF\u0000" + // 10910 - 10919
                "\u0000\u0000\uD8DA\u0000\u0000\u0000\u0000\u0000\uD6E2\u0000" + // 10920 - 10929
                "\u0000\u0000\uEFA8\u0000\u0000\u0000\uEEB4\uB8F4\u0000\uB8F5" + // 10930 - 10939
                "\u0000\u0000\u0000\u0000\u0000\uBBFE\uBCA1\u0000\u0000\uCCF0" + // 10940 - 10949
                "\u0000\u0000\uD7AF\u0000\u0000\u0000\uDDE6\u0000\u0000\u0000" + // 10950 - 10959
                "\uDCFE\u0000\uFCC7\uDCD6\uE2E0\u0000\u0000\u0000\uDAB0\u0000" + // 10960 - 10969
                "\uA6A4\uA6C2\uA6C1\uA6AF\uA6A6\uA6C6\uA6C5\uA6B1\uB8F0\uB8F1" + // 10970 - 10979
                "\u0000\uB8F2\uB8F3\u0000\u0000\u0000\uD8EE\u0000\uF2C1\u0000" + // 10980 - 10989
                "\u0000\u0000\uBBA8\uBBA9\uBBAA\u0000\u0000\uEEEB\u0000\u0000" + // 10990 - 10999
                "\u0000\u0000\u0000\u0000\uBDE0\u0000\u0000\uE7C8\u0000\u0000" + // 11000 - 11009
                "\u0000\u0000\u0000\u0000\uFAB4\u0000\u0000\uE9E0\u0000\u0000" + // 11010 - 11019
                "\u0000\uD0D8\uFCA2\uD4BE\uB8EA\u0000\u0000\u0000\u0000\u0000" + // 11020 - 11029
                "\u0000\u0000\uF3E2\u0000\uD5D0\uE5D9\u0000\u0000\u0000\u0000" + // 11030 - 11039
                "\u0000\u0000\uFCA1\uEFEE\uDCD8\uB8E7\uB8E8\u0000\u0000\uB8E9" + // 11040 - 11049
                "\u0000\u0000\u0000\uF9EC\u0000\u0000\u0000\uCBCC\u0000\uDBC7" + // 11050 - 11059
                "\uDED5\u0000\u0000\u0000\u0000\uF0F4\u0000\uA6A1\uA6AC\uA6A2" + // 11060 - 11069
                "\uA6AD\u0000\u0000\u0000\u0000\uBEA9\uBEAA\u0000\u0000\uF7E7" + // 11070 - 11079
                "\u0000\u0000\uCDDE\u0000\uF7A4\u0000\u0070\u0071\u0072\u0073" + // 11080 - 11089
                "\u0074\u0075\u0076\u0077\uB8E5\uB8E6\u0000\u0000\u0000\u0000" + // 11090 - 11099
                "\u0000\u0000\uD0F3\uE0AA\uE8E2\u0000\uDFEF\uCCD3\uD3F9\u0000" + // 11100 - 11109
                "\u0000\u0000\u0000\uD4F0\uB8E0\u0000\u0000\u0000\uB8E1\u0000" + // 11110 - 11119
                "\u0000\u0000\uE9DD\uDBCD\u0000\u0000\uDDCE\u0000\uE8D9\uEFD6" + // 11120 - 11129
                "\u0000\u0000\u0000\uD3E2\u0000\uE2DF\uB8D8\uB8D9\u0000\uB8DA" + // 11130 - 11139
                "\u0000\uB8DB\uB8DC\u0000\uF3E0\uE7AF\u0000\u0000\u0000\u0000" + // 11140 - 11149
                "\u0000\uDBAD\uB8D6\u0000\uB8D7\u0000\u0000\u0000\u0000\u0000" + // 11150 - 11159
                "\uBBEF\uBBF0\u0000\uBBF1\uB8D3\uB8D4\u0000\u0000\uB8D5\u0000" + // 11160 - 11169
                "\u0000\u0000\uF5D4\u0000\u0000\u0000\u0000\uD9A9\uB8D1\u0000" + // 11170 - 11179
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE2A5\uCDB9\uB8CF\uB8D0" + // 11180 - 11189
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE5BD\u0000\u0000\u0000" + // 11190 - 11199
                "\uEBA8\u0000\u0000\u0000\uDBFE\u0000\uF7A5\u0000\uCBAE\u0000" + // 11200 - 11209
                "\uDAAF\u0000\uD8B6\u0000\uA8E5\uA8E6\u0000\u0000\u0000\u0000" + // 11210 - 11219
                "\u0000\u0000\uEFD7\u0000\uD4C3\uB8CC\uB8CD\uB8CE\u0000\u0000" + // 11220 - 11229
                "\u0000\u0000\u0000\uBBE7\uBBE8\u0000\uBBE9\uB8C7\u0000\u0000" + // 11230 - 11239
                "\u0000\uB8C8\u0000\u0000\u0000\uD6FB\u0000\u0000\u0000\u0000" + // 11240 - 11249
                "\u0000\uB2AD\uB2AE\u0000\uB2AF\uB8BE\uB8BF\u0000\uB8C0\u0000" + // 11250 - 11259
                "\uB8C1\uB8C2\u0000\uFBF0\u0000\u0000\uECAC\u0000\u0000\u0000" + // 11260 - 11269
                "\uF0A9\uB8BB\uB8BC\uB8BD\u0000\u0000\u0000\u0000\u0000\uBBD8" + // 11270 - 11279
                "\u0000\u0000\u0000\uDDF6\u0000\uCDC0\u0000\u0000\u0000\uB4E2" + // 11280 - 11289
                "\uB4E3\uB4E4\u0000\uB4E5\uB8B6\uB8B7\u0000\u0000\uB8B8\u0000" + // 11290 - 11299
                "\uB8B9\uB8BA\uB8B0\u0000\u0000\u0000\uB8B1\u0000\u0000\u0000" + // 11300 - 11309
                "\uD7E5\u0000\u0000\u0000\u0000\u0000\uB2A8\uB2A9\uB2AA\u0000" + // 11310 - 11319
                "\uF3DF\u0000\uF8C8\uCEC6\u0000\u0000\u0000\u0000\uC1E2\u0000" + // 11320 - 11329
                "\u0000\u0000\uF0FE\u0000\u0000\u0000\u0000\uC7FD\u0000\u0000" + // 11330 - 11339
                "\u0000\uE0CE\u0000\uF5FD\u0000\u0000\uE5E4\uDFF1\u0000\u0000" + // 11340 - 11349
                "\uF7E1\u0000\uF9F7\uB8A5\u0000\u0000\u0000\uB8A6\u0000\u0000" + // 11350 - 11359
                "\u0000\uE5CA\u0000\uF6C0\uFDDD\u0000\u0000\uE0D5\u0000\uEFB0" + // 11360 - 11369
                "\u0000\u0000\uE2C7\u0000\uCFAA\u0000\u0000\uCEA9\u0000\u0000" + // 11370 - 11379
                "\uD6F8\u0000\uA8DD\uA8DE\uA8DF\uA8E0\uA8E1\uA8E2\uA8E3\uA8E4" + // 11380 - 11389
                "\uB7FD\uB7FE\u0000\uB8A1\u0000\uB8A2\u0000\u0000\uFAE8\u0000" + // 11390 - 11399
                "\u0000\u0000\u0000\u0000\u0000\uBCD1\u0000\u0000\uD0D9\u0000" + // 11400 - 11409
                "\u0000\uDDD2\uF7F4\uE7DC\uE4A5\uB7FC\u0000\u0000\u0000\u0000" + // 11410 - 11419
                "\u0000\u0000\u0000\uE3A6\uD1DA\uB7F9\uB7FA\u0000\u0000\uB7FB" + // 11420 - 11429
                "\u0000\u0000\u0000\uCAB2\u0000\u0000\uDCBB\u0000\uF1F8\uB7F4" + // 11430 - 11439
                "\u0000\u0000\u0000\uB7F5\u0000\u0000\u0000\uDBE3\u0000\u0000" + // 11440 - 11449
                "\u0000\u0000\uF1E1\uB7F1\u0000\u0000\u0000\u0000\u0000\u0000" + // 11450 - 11459
                "\u0000\uDDCA\u0000\uF9BF\uD6AF\uD5C6\u0000\u0000\u0000\u0000" + // 11460 - 11469
                "\u0000\uFCF5\u0000\u0000\u0000\uE4FD\u0000\u0000\uE3EC\u0000" + // 11470 - 11479
                "\uA8D5\uA8D6\uA8D7\uA8D8\uA8D9\uA8DA\uA8DB\uA8DC\uB7F0\u0000" + // 11480 - 11489
                "\u0000\u0000\u0000\u0000\u0000\u0000\uD5E8\uDBAE\uB7EB\uB7EC" + // 11490 - 11499
                "\u0000\uB7ED\u0000\uB7EE\u0000\u0000\uEBB8\u0000\uE0B7\uD7EC" + // 11500 - 11509
                "\uF1EC\uE5AF\uD5E1\uB7EA\u0000\u0000\u0000\u0000\u0000\u0000" + // 11510 - 11519
                "\u0000\uD1D7\u0000\uE9BC\u0000\u0000\u0000\u0000\u0000\uEAEC" + // 11520 - 11529
                "\u0000\uA8CD\uA8CE\uA8CF\uA8D0\uA8D1\uA8D2\uA8D3\uA8D4\uB7E7" + // 11530 - 11539
                "\uB7E8\u0000\u0000\uB7E9\u0000\u0000\u0000\uF9B4\u0000\u0000" + // 11540 - 11549
                "\uD5D4\uFDCF\u0000\uDFCA\u0000\u0000\u0000\u0000\u0000\uD3F8" + // 11550 - 11559
                "\uF1A8\uB7E2\u0000\u0000\u0000\uB7E3\u0000\u0000\u0000\uD1AF" + // 11560 - 11569
                "\uD7E3\u0000\u0000\u0000\uE0C6\uB7DD\uB7DE\u0000\uB7DF\u0000" + // 11570 - 11579
                "\uB7E0\u0000\u0000\uDBA3\u0000\u0000\uD6CA\uCBD9\u0000\u0000" + // 11580 - 11589
                "\uF3ED\u0000\u0000\u0000\u0000\u0000\u0000\uDCE0\u0000\u0000" + // 11590 - 11599
                "\uD4AA\u0000\uE5CC\u0000\u0000\u0000\u0000\uC6D9\uC6DA\u0000" + // 11600 - 11609
                "\u0000\uEEC3\u0000\u0000\u0000\u0000\u0000\u0000\uD5B5\u0000" + // 11610 - 11619
                "\u0000\uFCB9\uEEC2\uCAF5\u0000\u0000\u0000\uEFE5\uB7DC\u0000" + // 11620 - 11629
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF5BC\uF2A4\uB7DA\u0000" + // 11630 - 11639
                "\u0000\u0000\uB7DB\u0000\u0000\u0000\uD6BC\uD3E5\u0000\u0000" + // 11640 - 11649
                "\u0000\u0000\uC5C6\uC5C7\u0000\uC5C8\uB7D9\u0000\u0000\u0000" + // 11650 - 11659
                "\u0000\u0000\u0000\u0000\uD3A7\uFBB2\uB7D6\u0000\u0000\u0000" + // 11660 - 11669
                "\uB7D7\u0000\u0000\u0000\uE7D9\u0000\u0000\u0000\u0000\u0000" + // 11670 - 11679
                "\uB1D4\u0000\u0000\u0000\uDCF3\uD9B0\u0000\uE6E9\u0000\uFAB9" + // 11680 - 11689
                "\uCACF\u0000\uFCB3\uEAEA\uEAEB\uD0FA\u0000\uA9E1\uA9E2\uA9E3" + // 11690 - 11699
                "\uA9E4\uA9E5\uA9E6\u0000\u0000\uD6B7\uCDB3\u0000\u0000\u0000" + // 11700 - 11709
                "\u0000\u0000\uEEE5\u0000\uDEF5\u0000\u0068\u0069\u006A\u006B" + // 11710 - 11719
                "\u006C\u006D\u006E\u006F\uB7D0\u0000\u0000\u0000\uB7D1\u0000" + // 11720 - 11729
                "\u0000\u0000\uDFE5\u0000\u0000\u0000\u0000\u0000\uB1CB\u0000" + // 11730 - 11739
                "\u0000\u0000\uCBF6\u0000\u0000\u0000\u0000\uD2B3\uD2BF\u0000" + // 11740 - 11749
                "\u0000\u0000\uE2D0\uF4F7\u0000\u0000\u0000\uCEF7\u0000\u0000" + // 11750 - 11759
                "\uE0D8\u0000\uEFD3\u0000\u0000\u0000\uE4C1\uF8EB\u0000\uDBAC" + // 11760 - 11769
                "\uB7CA\u0000\u0000\u0000\uB7CB\u0000\u0000\u0000\uF4E0\u0000" + // 11770 - 11779
                "\u0000\u0000\u0000\u0000\uB1C9\uB1CA\u0000\u0000\uF6AD\u0000" + // 11780 - 11789
                "\uF5B3\u0000\uF0B5\u0000\u0000\uE0CF\u0000\u0000\u0000\u0000" + // 11790 - 11799
                "\u0000\u0000\uD8C3\u0000\u0000\uE9DF\u0000\u0000\u0000\u0000" + // 11800 - 11809
                "\u0000\u0000\uE9E8\u0000\u0000\uF7E5\u0000\u0000\u0000\uCCB2" + // 11810 - 11819
                "\u0000\u0000\uD3BF\u0000\u0000\u0000\u0000\u0000\u0000\uF0B0" + // 11820 - 11829
                "\u0000\u0000\uE3FE\uD1AA\uE8AA\u0000\uEAB6\uF9FA\uE6CC\uB7C8" + // 11830 - 11839
                "\uB7C9\u0000\u0000\u0000\u0000\u0000\u0000\uDEAD\u0000\uFABF" + // 11840 - 11849
                "\u0000\uDFC8\u0000\u0000\u0000\u0000\uD9B6\u0000\uFDAC\uB7C3" + // 11850 - 11859
                "\u0000\u0000\u0000\uB7C4\u0000\u0000\u0000\uE5AC\uFDA1\u0000" + // 11860 - 11869
                "\uDFD0\uECB3\u0000\uCBDD\u0000\u0000\uD9E3\u0000\u0000\uF3AC" + // 11870 - 11879
                "\u0000\uA9D9\uA9DA\uA9DB\uA9DC\uA9DD\uA9DE\uA9DF\uA9E0\uB7BD" + // 11880 - 11889
                "\uB7BE\u0000\uB7BF\u0000\uB7C0\u0000\u0000\uD9AE\uD5AC\u0000" + // 11890 - 11899
                "\uE2C6\u0000\u0000\u0000\uE5FD\uDDE5\uD8CD\u0000\u0000\u0000" + // 11900 - 11909
                "\uD9C3\uD0E8\u0000\u0000\u0000\uE0B4\u0000\u0000\u0000\u0000" + // 11910 - 11919
                "\uC6C4\uC6C5\uC6C6\u0000\uE7AE\u0000\uD6BA\u0000\uDFEC\uE4C0" + // 11920 - 11929
                "\u0000\u0000\uCCBC\uF7EA\u0000\u0000\u0000\u0000\u0000\uCEBC" + // 11930 - 11939
                "\u0000\u0000\u0000\uE2F2\u0000\u0000\u0000\u0000\uA1D8\u0000" + // 11940 - 11949
                "\u0000\u0000\u0000\uBDFC\u0000\u0000\u0000\uE0E0\u0000\u0000" + // 11950 - 11959
                "\u0000\u0000\uA1C9\u0000\u0000\u0000\u0000\uBDFD\u0000\u0000" + // 11960 - 11969
                "\u0000\uFAC8\uD6D7\u0000\uE9E5\uFBDC\uB7BC\u0000\u0000\u0000" + // 11970 - 11979
                "\u0000\u0000\u0000\u0000\uFAD7\uFBC1\uB7B9\uB7BA\u0000\u0000" + // 11980 - 11989
                "\uB7BB\u0000\u0000\u0000\uF0D0\u0000\uF7F0\uEEB3\u0000\u0000" + // 11990 - 11999
                "\uEBB7\uEFF8\uF5DC\uEDCC\uDBD5\uF1CF\u0000\uCEE4\u0000\uE8F2" + // 12000 - 12009
                "\u0000\u0000\u0000\u0000\u0000\uCCC7\u0000\u0000\u0000\uD3C4" + // 12010 - 12019
                "\u0000\u0000\uD6C0\u0000\uA9D1\uA9D2\uA9D3\uA9D4\uA9D5\uA9D6" + // 12020 - 12029
                "\uA9D7\uA9D8\uB7B6\uB7B7\u0000\u0000\u0000\u0000\u0000\uB7B8" + // 12030 - 12039
                "\uB7B1\u0000\u0000\u0000\uB7B2\u0000\u0000\u0000\uDCEC\u0000" + // 12040 - 12049
                "\u0000\u0000\u0000\u0000\uB1B8\uB1B9\u0000\u0000\uDEA9\u0000" + // 12050 - 12059
                "\u0000\u0000\u0000\u0000\u0000\uBBD0\u0000\u0000\uCBB8\u0000" + // 12060 - 12069
                "\u0000\u0000\u0000\u0000\u0000\uF6D0\uEBE6\uDAF9\uB7AC\u0000" + // 12070 - 12079
                "\u0000\u0000\u0000\u0000\u0000\u0000\uD8AC\uF3CC\uB7A5\uB7A6" + // 12080 - 12089
                "\u0000\uB7A7\uB7A8\uB7A9\u0000\u0000\uCFDB\u0000\u0000\u0000" + // 12090 - 12099
                "\u0000\u0000\u0000\uBAB1\u0000\uBAB2\uB7A4\u0000\u0000\u0000" + // 12100 - 12109
                "\u0000\u0000\u0000\u0000\uF8C6\u0000\uD5B9\u0000\u0000\u0000" + // 12110 - 12119
                "\u0000\u0000\u0000\u0000\uE2FE\uB7A1\uB7A2\u0000\u0000\uB7A3" + // 12120 - 12129
                "\u0000\u0000\u0000\uF5A4\u0000\u0000\u0000\u0000\u0000\uB1AB" + // 12130 - 12139
                "\uB1AC\u0000\u0000\uE9F4\u0000\u0000\u0000\u0000\u0000\u0000" + // 12140 - 12149
                "\uB9F5\u0000\u0000\uE2BB\u0000\uF7AD\u0000\u0000\u0000\uF8E1" + // 12150 - 12159
                "\uB6FA\uB6FB\uB6FC\u0000\u0000\u0000\uB6FD\uB6FE\uB6F5\u0000" + // 12160 - 12169
                "\u0000\u0000\uB6F6\u0000\u0000\u0000\uF6A7\u0000\u0000\u0000" + // 12170 - 12179
                "\uE6FA\u0000\uDCE8\u0000\u0000\u0000\uFAD6\u0000\uD3F6\u0000" + // 12180 - 12189
                "\uA9F3\uA9F4\uA9F5\u0000\u0000\u0000\u0000\u0000\uD5ED\uE7DD" + // 12190 - 12199
                "\u0000\u0000\uF5B6\u0000\u0000\u0000\u0000\u0000\u0000\uD3C1" + // 12200 - 12209
                "\uF0CD\u0000\u0060\u0061\u0062\u0063\u0064\u0065\u0066\u0067" + // 12210 - 12219
                "\uB6EF\uB6F0\u0000\uB6F1\u0000\uB6F2\u0000\u0000\uD8D7\u0000" + // 12220 - 12229
                "\u0000\u0000\u0000\u0000\u0000\uB9A3\u0000\uB9A4\uB6EE\u0000" + // 12230 - 12239
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF6A5\u0000\uE4B1\u0000" + // 12240 - 12249
                "\u0000\u0000\u0000\u0000\u0000\uDCE7\uB6EC\u0000\u0000\u0000" + // 12250 - 12259
                "\uB6ED\u0000\u0000\u0000\uE7B9\u0000\u0000\u0000\u0000\uF0AD" + // 12260 - 12269
                "\uB6E8\u0000\u0000\u0000\uB6E9\u0000\u0000\u0000\uE7D8\u0000" + // 12270 - 12279
                "\u0000\u0000\u0000\u0000\uB1A1\uB1A2\u0000\uB1A3\uB6E4\uB6E5" + // 12280 - 12289
                "\u0000\uB6E6\u0000\u0000\u0000\u0000\uE3FD\u0000\uF9B1\u0000" + // 12290 - 12299
                "\u0000\uEBE9\u0000\u0000\u0000\uE8BB\u0000\u0000\uFBD7\u0000" + // 12300 - 12309
                "\u0000\uEBCA\uE0A1\u0000\u0000\uCFBB\u0000\u0000\u0000\uD3AD" + // 12310 - 12319
                "\uE8E1\uCEEC\uB6E3\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 12320 - 12329
                "\uCACD\u0000\uDFC7\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 12330 - 12339
                "\uE2FD\uB6DF\uB6E0\u0000\u0000\uB6E1\u0000\u0000\uB6E2\uB6DC" + // 12340 - 12349
                "\uB6DD\u0000\u0000\u0000\uB6DE\u0000\u0000\uEEEA\u0000\u0000" + // 12350 - 12359
                "\u0000\uF0E4\uF3B4\uD4EE\uB6DB\u0000\u0000\u0000\u0000\u0000" + // 12360 - 12369
                "\u0000\u0000\uFAF1\u0000\uE1FC\u0000\u0000\u0000\u0000\u0000" + // 12370 - 12379
                "\u0000\u0000\uF9BE\uB6D9\u0000\u0000\u0000\uB6DA\u0000\u0000" + // 12380 - 12389
                "\u0000\uEDD2\u0000\uD4D8\uDCC9\uD7F1\u0000\uE4CD\u0000\uD6B9" + // 12390 - 12399
                "\u0000\u0000\u0000\uEFC0\u0000\uA9EB\uA9EC\uA9ED\uA9EE\uA9EF" + // 12400 - 12409
                "\uA9F0\uA9F1\uA9F2\uB6D3\u0000\u0000\u0000\uB6D4\u0000\u0000" + // 12410 - 12419
                "\u0000\uFCD2\u0000\uEBC8\u0000\u0000\u0000\uF4B4\u0000\u0000" + // 12420 - 12429
                "\u0000\u0000\uF8D5\uE5D8\u0000\u0000\u0000\uFBB9\u0000\uE4D3" + // 12430 - 12439
                "\u0000\uCDF9\uB6D0\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 12440 - 12449
                "\uFDB5\u0000\uD0F7\uEDD4\u0000\u0000\u0000\u0000\u0000\u0000" + // 12450 - 12459
                "\uFAE0\u0000\u0000\uDDEB\u0000\u0000\uE4F9\u0000\u0000\uE3AF" + // 12460 - 12469
                "\uB6CE\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF4DC\u0000" + // 12470 - 12479
                "\uD3F5\u0000\uD7A6\u0000\uF6B5\uD7DB\u0000\u0000\uD0E2\u0000" + // 12480 - 12489
                "\u0000\u0000\u0000\uDDA6\u0000\uA8EF\uA8F0\uA8F1\uA8F2\uA8F3" + // 12490 - 12499
                "\uA8F4\uA8F5\u0000\u0058\u0059\u005A\u005B\\\u005D\u005E" + // 12500 - 12509
                "\u005F\uB6CA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uFCF7" + // 12510 - 12519
                "\u0000\uCFDD\u0000\u0000\uE8A9\u0000\uE3BB\uE3BA\u0000\uA8E7" + // 12520 - 12529
                "\uA8E8\uA8E9\uA8EA\uA8EB\uA8EC\uA8ED\uA8EE\uB6C7\uB6C8\u0000" + // 12530 - 12539
                "\u0000\uB6C9\u0000\u0000\u0000\uDBB5\u0000\uF3E7\uD8FE\u0000" + // 12540 - 12549
                "\u0000\uE9AB\u0000\u0000\uE1E1\uD3CF\uF4F6\u0000\uCFD1\u0000" + // 12550 - 12559
                "\u0000\u0000\u0000\uCBDC\uCCFA\u0000\uA1C1\uA1D5\u0000\u0000" + // 12560 - 12569
                "\uA1C2\uA1C3\u0000\u0000\uCAA9\u0000\u0000\u0000\u0000\u0000" + // 12570 - 12579
                "\u0000\uE6EE\u0000\uCCDC\uB6C5\u0000\u0000\u0000\u0000\u0000" + // 12580 - 12589
                "\u0000\u0000\uF7A3\u0000\uE3B9\uEBC5\uF4A9\uCDB6\uD2F9\u0000" + // 12590 - 12599
                "\uDAAD\uD2E3\uB6C3\uB6C4\u0000\u0000\u0000\u0000\u0000\u0000" + // 12600 - 12609
                "\uF2C7\u0000\u0000\u0000\uD2FC\u0000\u0000\u0000\u0000\u0000" + // 12610 - 12619
                "\uB0D7\uB0D8\u0000\uB0D9\uB6BE\u0000\u0000\u0000\uB6BF\u0000" + // 12620 - 12629
                "\u0000\u0000\uCDA2\uE8AE\u0000\u0000\u0000\uE1BD\uB6B6\uB6B7" + // 12630 - 12639
                "\u0000\uB6B8\uB6B9\uB6BA\u0000\u0000\uD4ED\uE2C4\u0000\u0000" + // 12640 - 12649
                "\u0000\u0000\uE9E7\uB6B3\u0000\uB6B4\uB6B5\u0000\u0000\u0000" + // 12650 - 12659
                "\u0000\uF4E6\u0000\u0000\uE6C5\uEFD5\uB6B0\uB6B1\u0000\u0000" + // 12660 - 12669
                "\uB6B2\u0000\u0000\u0000\uECEE\u0000\u0000\uDDAA\u0000\u0000" + // 12670 - 12679
                "\uF1CE\uF2E4\u0000\u0000\uD0B0\u0000\u0000\uCDC6\uF2B6\u0000" + // 12680 - 12689
                "\u0000\uDDFE\u0000\u0000\uD9AA\u0000\u0000\u0000\u0000\u0000" + // 12690 - 12699
                "\u0000\uE7E2\u0000\u0000\uE9FB\uEAA8\u0000\u0000\u0000\u0000" + // 12700 - 12709
                "\uFDB7\uB6AE\uB6AF\u0000\u0000\u0000\u0000\u0000\u0000\uF0A6" + // 12710 - 12719
                "\u0000\u0000\u0000\uDCC8\u0000\u0000\u0000\u0000\u0000\uB0C2" + // 12720 - 12729
                "\u0000\u0000\u0000\uD7B0\uD8E8\uCBBD\u0000\u0000\uEECF\uF7D7" + // 12730 - 12739
                "\u0000\u0000\uE0A6\uD6C1\uE1DC\uB6A9\u0000\u0000\u0000\uB6AA" + // 12740 - 12749
                "\u0000\u0000\u0000\uEAB8\uD1F9\u0000\u0000\u0000\u0000\uC4E8" + // 12750 - 12759
                "\u0000\u0000\u0000\uE7A2\uE4D9\u0000\u0000\u0000\uF0E6\u0000" + // 12760 - 12769
                "\u0000\u0000\uE4B9\uB6A1\uB6A2\u0000\uB6A3\uB6A4\uB6A5\u0000" + // 12770 - 12779
                "\u0000\uEEE9\u0000\u0000\u0000\uF5DA\u0000\u0000\uCFBE\u0000" + // 12780 - 12789
                "\u0000\uECBB\u0000\u0000\u0000\uEEED\u0000\u0000\u0000\uECEB" + // 12790 - 12799
                "\uDEC5\uB5FE\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE4AF" + // 12800 - 12809
                "\u0000\uDBAB\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE6B4" + // 12810 - 12819
                "\uB5FB\uB5FC\u0000\u0000\uB5FD\u0000\u0000\u0000\uF1BE\u0000" + // 12820 - 12829
                "\u0000\uD3AC\u0000\u0000\uD0CE\u0000\u0000\uDAF7\u0000\u0000" + // 12830 - 12839
                "\u0000\uD9D3\u0000\u0000\u0000\u0000\uD3DE\uB5F8\uB5F9\uB5FA" + // 12840 - 12849
                "\u0000\u0000\u0000\u0000\u0000\uBBD5\u0000\u0000\uBBD6\uB5F2" + // 12850 - 12859
                "\u0000\u0000\uB5F3\uB5F4\u0000\u0000\u0000\uE0C2\u0000\uCAE4" + // 12860 - 12869
                "\u0000\uE7B7\u0000\uEEF6\uEACF\uF0EE\uE3FC\u0000\uD3DF\uD3F4" + // 12870 - 12879
                "\uE1B3\uB5EF\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF6E4" + // 12880 - 12889
                "\u0000\uF5BB\u0000\uDED1\u0000\u0000\u0000\u0000\u0000\uF6DD" + // 12890 - 12899
                "\u0000\uF1A3\u0000\uA1FD\uA1FB\uA1FA\uA1F2\uA1F3\u0000\uA2B1" + // 12900 - 12909
                "\u0000\u0050\u0051\u0052\u0053\u0054\u0055\u0056\u0057\uB5E7" + // 12910 - 12919
                "\u0000\u0000\uB5E8\uB5E9\u0000\uB5EA\u0000\uF3F3\uE3FB\u0000" + // 12920 - 12929
                "\uDED0\uCEB0\u0000\uD6F7\uF1D9\uB5E3\u0000\u0000\u0000\u0000" + // 12930 - 12939
                "\uB5E4\u0000\u0000\uEFAE\u0000\u0000\u0000\uF4D0\uCEF3\u0000" + // 12940 - 12949
                "\uD8DB\u0000\uF9CE\uE9D5\uE3D1\u0000\u0000\uD2BC\uB5E2\u0000" + // 12950 - 12959
                "\u0000\u0000\u0000\u0000\u0000\u0000\uD7DA\u0000\uF3F2\u0000" + // 12960 - 12969
                "\uEED6\uEAB2\uD0F6\uECD9\uDACB\uCFA8\uB5E0\u0000\u0000\u0000" + // 12970 - 12979
                "\uB5E1\u0000\u0000\u0000\uE4FB\u0000\u0000\uF9E4\u0000\u0000" + // 12980 - 12989
                "\uDBD4\uD7C7\u0000\u0000\u0000\u0000\uF2FE\uB5DB\u0000\u0000" + // 12990 - 12999
                "\u0000\uB5DC\u0000\u0000\u0000\uCCBD\u0000\u0000\uD1A9\uDDCC" + // 13000 - 13009
                "\u0000\uE3E4\uE9BB\u0000\u0000\u0000\u0000\u0000\uE2D6\uB5D8" + // 13010 - 13019
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uFAB6\u0000\uEEA5" + // 13020 - 13029
                "\u0000\u0000\uFAAA\uE6C3\uE1B2\uCAAB\u0000\uA1D0\u0000\u0000" + // 13030 - 13039
                "\u0000\u0000\uA1AB\u0000\uA1FC\uB5D7\u0000\u0000\u0000\u0000" + // 13040 - 13049
                "\u0000\u0000\u0000\uE5A7\uD5D2\uB5D2\uB5D3\u0000\uB5D4\u0000" + // 13050 - 13059
                "\uB5D5\u0000\u0000\uEEB0\u0000\u0000\u0000\u0000\u0000\u0000" + // 13060 - 13069
                "\uB8D2\u0000\u0000\uCEEE\u0000\u0000\uECCF\u0000\u0000\u0000" + // 13070 - 13079
                "\uD0D1\uCBBF\u0000\uEDA4\u0000\u0000\uD4A9\u0000\u0000\u0000" + // 13080 - 13089
                "\u0000\uCDC2\uE7DA\uB5D1\u0000\u0000\u0000\u0000\u0000\u0000" + // 13090 - 13099
                "\u0000\uCCCB\u0000\uCEAF\uF1B5\uEFD2\uE8C8\uEBF1\u0000\u0000" + // 13100 - 13109
                "\u0000\uE7C5\u0000\u0000\uE0E9\u0000\uA1F4\u0000\u0000\uA1F5" + // 13110 - 13119
                "\u0000\u0000\u0000\uA2B3\uB5CE\uB5CF\u0000\u0000\uB5D0\u0000" + // 13120 - 13129
                "\u0000\u0000\uE7FD\u0000\u0000\uE6A3\uFBF1\uCBB0\uB5CA\uB5CB" + // 13130 - 13139
                "\u0000\uB5CC\u0000\u0000\u0000\u0000\uCDB7\u0000\uEFD4\u0000" + // 13140 - 13149
                "\u0000\uF0E2\u0000\u0000\u0000\u0000\u0000\uEECE\uB5C9\u0000" + // 13150 - 13159
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE1F3\uDCF6\uB5C7\u0000" + // 13160 - 13169
                "\u0000\u0000\uB5C8\u0000\u0000\u0000\uD4B8\uEBBE\uDDEF\u0000" + // 13170 - 13179
                "\uDDF0\uDDF1\uB5C6\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 13180 - 13189
                "\uD9B2\uFDA5\uB5C4\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 13190 - 13199
                "\uD3A2\u0000\uE1FA\uE4CC\u0000\uE1E4\uE8C7\u0000\u0000\uCEDB" + // 13200 - 13209
                "\uB5C2\u0000\u0000\u0000\uB5C3\u0000\u0000\u0000\uDFAF\u0000" + // 13210 - 13219
                "\uCAC3\u0000\u0000\uEEFC\uB5BB\u0000\u0000\u0000\uB5BC\uB5BD" + // 13220 - 13229
                "\u0000\uB5BE\uB5B7\u0000\u0000\uB5B8\uB5B9\u0000\uB5BA\u0000" + // 13230 - 13239
                "\uF9CA\u0000\uEAE8\u0000\uE5ED\u0000\u0000\u0000\uDCEE\u0000" + // 13240 - 13249
                "\u0000\uF5EA\uE6E0\uB5B3\u0000\u0000\u0000\uB5B4\u0000\u0000" + // 13250 - 13259
                "\u0000\uE0E8\u0000\u0000\uD3AB\u0000\uEBDC\uB5B1\uB5B2\u0000" + // 13260 - 13269
                "\u0000\u0000\u0000\u0000\u0000\uF8A6\u0000\uDECA\uF2C6\uB5AF" + // 13270 - 13279
                "\u0000\u0000\u0000\uB5B0\u0000\u0000\u0000\uE1A8\u0000\u0000" + // 13280 - 13289
                "\u0000\u0000\uD5F6\uB5A9\uB5AA\u0000\uB5AB\uB5AC\uB5AD\u0000" + // 13290 - 13299
                "\u0000\uEEE6\u0000\uE0D3\u0000\u0000\u0000\u0000\uF3D1\u0000" + // 13300 - 13309
                "\u0000\u0000\u0000\uBFD6\uBFD7\u0000\u0000\uD7E7\u0000\u0000" + // 13310 - 13319
                "\u0000\u0000\u0000\u0000\uE9EA\uD7CC\u0000\uE2D5\uEDCF\u0000" + // 13320 - 13329
                "\u0000\u0000\uDDA2\u0000\u0000\uCAF1\u0000\u0000\u0000\u0000" + // 13330 - 13339
                "\u0000\u0000\uF4D8\uD6B3\uDDAD\uB5A8\u0000\u0000\u0000\u0000" + // 13340 - 13349
                "\u0000\u0000\u0000\uCBEB\u0000\uF0ED\u0000\uDDA1\u0000\uEDAF" + // 13350 - 13359
                "\uFCF8\u0000\uD8EB\uB5A5\uB5A6\u0000\u0000\uB5A7\u0000\u0000" + // 13360 - 13369
                "\u0000\uE8E0\u0000\u0000\u0000\u0000\u0000\uCFD5\uD8FD\u0000" + // 13370 - 13379
                "\u0000\uF0E1\u0000\u0000\u0000\u0000\u0000\u0000\uB7E4\u0000" + // 13380 - 13389
                "\uB7E5\uB4F8\u0000\u0000\uB4F9\uB4FA\u0000\uB4FB\uB4FC\uB4EF" + // 13390 - 13399
                "\uB4F0\u0000\uB4F1\uB4F2\uB4F3\u0000\u0000\uF2E1\u0000\uDEB9" + // 13400 - 13409
                "\u0000\u0000\u0000\u0000\uD5F4\u0000\u0000\u0000\uCDED\uB4EE" + // 13410 - 13419
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCED2\u0000\uEDAD" + // 13420 - 13429
                "\uFAEA\u0000\u0000\uCDEE\uEDA6\u0000\uEDAE\uB4EB\uB4EC\u0000" + // 13430 - 13439
                "\u0000\uB4ED\u0000\u0000\u0000\uD1A7\u0000\u0000\uFDE3\uCEB3" + // 13440 - 13449
                "\u0000\uDFC5\u0000\u0000\uE5BE\u0000\u0000\u0000\u0000\uC1CB" + // 13450 - 13459
                "\u0000\u0000\u0000\uDADA\u0000\uF2DC\uFBD6\uE9B2\uB4E6\uB4E7" + // 13460 - 13469
                "\uB4E8\uB4E9\u0000\u0000\u0000\uB4EA\uB4DC\u0000\u0000\uB4DD" + // 13470 - 13479
                "\uB4DE\uB4DF\uB4E0\uB4E1\uB4D4\uB4D5\u0000\uB4D6\u0000\uB4D7" + // 13480 - 13489
                "\u0000\u0000\uF8FB\uE3CF\u0000\u0000\u0000\u0000\u0000\uB6F3" + // 13490 - 13499
                "\uB6F4\u0000\u0000\uE2B9\u0000\u0000\u0000\u0000\u0000\u0000" + // 13500 - 13509
                "\uF3C8\u0000\uF3BA\uB4D2\u0000\uB4D3\u0000\u0000\u0000\u0000" + // 13510 - 13519
                "\u0000\uBBD1\uBBD2\u0000\u0000\uDFDA\u0000\u0000\u0000\u0000" + // 13520 - 13529
                "\u0000\u0000\uB7D8\u0000\u0000\uDBE5\u0000\uDDDE\u0000\u0000" + // 13530 - 13539
                "\uD9F0\uE9A3\uB4CF\uB4D0\u0000\u0000\uB4D1\u0000\u0000\u0000" + // 13540 - 13549
                "\uE9B0\u0000\u0000\u0000\u0000\u0000\uCDAE\u0000\u0000\u0000" + // 13550 - 13559
                "\uE6E7\u0000\u0000\uEAC7\u0000\uE3E3\u0000\u0000\u0000\u0000" + // 13560 - 13569
                "\uE4B0\u0000\u0000\uF5C3\uE9D8\u0000\u0000\u0000\u0000\u0000" + // 13570 - 13579
                "\uF1D1\u0000\u0000\u0000\uF5FE\uD4AC\u0000\u0000\u0000\uCAE7" + // 13580 - 13589
                "\u0000\u0000\u0000\u0000\uC8DE\uC8DF\u0000\u0000\uCCA8\u0000" + // 13590 - 13599
                "\uDAC1\uCCD5\u0000\uD9E4\u0000\uA2A3\u0000\uA1D3\uA2A4\u0000" + // 13600 - 13609
                "\u0000\u0000\uA1D4\uB4CD\u0000\u0000\u0000\uB4CE\u0000\u0000" + // 13610 - 13619
                "\u0000\uD5B2\u0000\u0000\u0000\uD5BC\u0000\uD2A6\u0000\u0000" + // 13620 - 13629
                "\uE7F4\uD1D6\u0000\u0000\uE6C2\uB4C6\uB4C7\u0000\uB4C8\u0000" + // 13630 - 13639
                "\uB4C9\uB4CA\u0000\uE6C1\u0000\u0000\uECD8\u0000\u0000\u0000" + // 13640 - 13649
                "\uEDAC\uB4C3\uB4C4\uB4C5\u0000\u0000\u0000\u0000\u0000\uBBCE" + // 13650 - 13659
                "\u0000\u0000\u0000\uD5EA\uF1EE\u0000\u0000\u0000\u0000\uC3F2" + // 13660 - 13669
                "\u0000\u0000\u0000\uECA2\uEDFD\u0000\uF5B4\uFBB8\uB4C0\uB4C1" + // 13670 - 13679
                "\u0000\u0000\uB4C2\u0000\u0000\u0000\uD4A2\uCFF6\u0000\u0000" + // 13680 - 13689
                "\u0000\u0000\uC3E8\u0000\u0000\u0000\uE3F4\uCDD0\u0000\u0000" + // 13690 - 13699
                "\u0000\uD5B8\u0000\u0000\uF7FD\u0000\uECA9\u0000\uF2EB\u0000" + // 13700 - 13709
                "\uFDEF\u0000\uF9F3\u0000\uA2D9\uA2D7\u0000\u0000\u0000\u0000" + // 13710 - 13719
                "\u0000\u0000\uFCC8\u0000\u0000\uEAE7\uDFC3\uD1D2\uCEE2\u0000" + // 13720 - 13729
                "\uD3A4\u0000\u0048\u0049\u004A\u004B\u004C\u004D\u004E\u004F" + // 13730 - 13739
                "\uB4B8\uB4B9\u0000\u0000\u0000\u0000\u0000\u0000\uFBFA\u0000" + // 13740 - 13749
                "\uCFA4\u0000\uEEBB\uCDB4\u0000\uE0F3\uEACD\u0000\u0000\u0000" + // 13750 - 13759
                "\uDAF2\u0000\u0000\u0000\u0000\uC6CA\uC6CB\u0000\uC6CC\uB4B7" + // 13760 - 13769
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uA2CF\uA2CE\uB4B5" + // 13770 - 13779
                "\u0000\u0000\u0000\uB4B6\u0000\u0000\u0000\uE9F8\uE2E5\u0000" + // 13780 - 13789
                "\u0000\u0000\u0000\uC3D2\u0000\u0000\u0000\uD5E0\uF6CA\uFDCA" + // 13790 - 13799
                "\uD8D6\uF4CF\uB4B2\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 13800 - 13809
                "\uA8AE\uA9AE\uB4AB\u0000\u0000\uB4AC\uB4AD\u0000\u0000\u0000" + // 13810 - 13819
                "\uDBE0\u0000\u0000\u0000\u0000\u0000\uF9CD\u0000\u0000\u0000" + // 13820 - 13829
                "\uF7F5\u0000\uCBE5\u0000\u0000\uFCC3\u0000\u0000\u0000\u0000" + // 13830 - 13839
                "\uD4E7\u0000\uDBAA\u0000\u0000\u0000\uF7E0\u0000\u0000\u0000" + // 13840 - 13849
                "\uD7BC\uCCE3\u0000\u0000\uE6DB\uB4A5\u0000\u0000\u0000\u0000" + // 13850 - 13859
                "\u0000\u0000\u0000\uA8A4\uA9A4\uB4A2\uB4A3\u0000\u0000\uB4A4" + // 13860 - 13869
                "\u0000\u0000\u0000\uE4C5\u0000\u0000\u0000\u0000\u0000\uDDDF" + // 13870 - 13879
                "\u0000\u0000\u0000\uDBA2\uF2F6\u0000\u0000\uCABA\uB3FB\u0000" + // 13880 - 13889
                "\u0000\u0000\uB3FC\u0000\u0000\u0000\uE5C5\u0000\u0000\u0000" + // 13890 - 13899
                "\u0000\u0000\uD8AA\u0000\u0000\u0000\uF6AC\u0000\u0000\u0000" + // 13900 - 13909
                "\u0000\uE4ED\u0000\u0000\u0000\u0000\uBFB0\uBFB1\uBFB2\uBFB3" + // 13910 - 13919
                "\uB3F9\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uA9A1\u0000" + // 13920 - 13929
                "\uCACC\u0000\u0000\u0000\u0000\uFBBF\u0000\u0000\uF4CB\u0000" + // 13930 - 13939
                "\u0000\u0000\uFDC5\u0000\u0000\uECF6\uE2E1\uE3BE\u0000\u0000" + // 13940 - 13949
                "\u0000\u0000\uC2B9\u0000\u0000\u0000\uEEE2\u0000\u0000\u0000" + // 13950 - 13959
                "\u0000\uC8C4\uC8C5\u0000\u0000\uE5BF\u0000\u0000\u0000\uCEBF" + // 13960 - 13969
                "\u0000\u0000\uE3FA\u0000\u0000\uF0AA\uF9D0\u0000\u0000\uFBE3" + // 13970 - 13979
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE8A4\u0000\u0000\uE5A9" + // 13980 - 13989
                "\uE0F6\uF6B3\u0000\u0000\u0000\u0000\uC2BC\uC2BD\u0000\u0000" + // 13990 - 13999
                "\uE1FE\u0000\u0000\u0000\u0000\uCBF0\u0000\uA1E7\uA1E8\uA1E6" + // 14000 - 14009
                "\uA1E9\uA1EA\uA2D5\uA2D8\uA2D6\uB3F7\u0000\u0000\u0000\uB3F8" + // 14010 - 14019
                "\u0000\u0000\u0000\uF5A1\u0000\u0000\u0000\u0000\u0000\uD0A1" + // 14020 - 14029
                "\u0000\u0000\u0000\uFAD3\u0000\u0000\u0000\u0000\uFDB3\u0000" + // 14030 - 14039
                "\uD5E4\u0000\u0000\uFCA6\u0000\u0000\u0000\u0000\u0000\u0000" + // 14040 - 14049
                "\uB6CB\u0000\u0000\uF0FD\u0000\u0000\u0000\u0000\u0000\uD7AC" + // 14050 - 14059
                "\uB3F0\uB3F1\u0000\uB3F2\u0000\uB3F3\u0000\u0000\uD0CD\u0000" + // 14060 - 14069
                "\u0000\u0000\u0000\u0000\u0000\uB5DD\u0000\uB5DE\uB3EE\u0000" + // 14070 - 14079
                "\uB3EF\u0000\u0000\u0000\u0000\u0000\uBBBE\u0000\u0000\u0000" + // 14080 - 14089
                "\uDCEA\u0000\u0000\uF0F7\u0000\uF0CA\uB3EB\uB3EC\u0000\u0000" + // 14090 - 14099
                "\uB3ED\u0000\u0000\u0000\uF1BB\u0000\u0000\u0000\u0000\uE9F1" + // 14100 - 14109
                "\uB3EA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uA8A1\u0000" + // 14110 - 14119
                "\uEEF5\u0000\uDECE\u0000\u0000\u0000\u0000\uE7F3\uB3E8\u0000" + // 14120 - 14129
                "\u0000\u0000\uB3E9\u0000\u0000\u0000\uE2E3\uEEFB\u0000\u0000" + // 14130 - 14139
                "\uDFF7\uD7CA\uB3E4\uB3E5\u0000\u0000\uB3E6\uB3E7\u0000\u0000" + // 14140 - 14149
                "\uCCC5\uECD0\uCBBB\u0000\uDEF3\u0000\u0000\uEEAD\u0000\uFAE3" + // 14150 - 14159
                "\u0000\u0000\u0000\u0000\uC7D8\uC7D9\u0000\u0000\uE1C4\u0000" + // 14160 - 14169
                "\u0000\u0000\u0000\u0000\uE8B0\uB3E3\u0000\u0000\u0000\u0000" + // 14170 - 14179
                "\u0000\u0000\u0000\uC8B1\uB3E0\uB3E1\u0000\u0000\uB3E2\u0000" + // 14180 - 14189
                "\u0000\u0000\uEAF6\u0000\u0000\uF6F9\u0000\u0000\uFDC9\u0000" + // 14190 - 14199
                "\u0000\u0000\u0000\uE4A6\uF9A4\uB3DE\uB3DF\u0000\u0000\u0000" + // 14200 - 14209
                "\u0000\u0000\u0000\uE4F8\u0000\u0000\u0000\uD8DD\u0000\uCDFD" + // 14210 - 14219
                "\uF2AB\u0000\u0000\uE3CD\u0000\u0000\u0000\u0000\uF4F4\uFAB2" + // 14220 - 14229
                "\uB3D9\u0000\u0000\u0000\uB3DA\u0000\u0000\u0000\uCFF5\u0000" + // 14230 - 14239
                "\u0000\uFDAE\u0000\u0000\uDEA6\u0000\u0000\uEBFE\u0000\uEBDD" + // 14240 - 14249
                "\uF0E0\uB3D1\uB3D2\u0000\uB3D3\uB3D4\uB3D5\u0000\u0000\uF4DA" + // 14250 - 14259
                "\u0000\u0000\u0000\u0000\u0000\u0000\uB5D9\u0000\u0000\uCBB5" + // 14260 - 14269
                "\uD8D1\u0000\uF4CE\uF3F7\u0000\u0000\uEBF7\u0000\u0000\u0000" + // 14270 - 14279
                "\u0000\u0000\u0000\uE9B3\u0000\u0000\uEFE8\u0000\u0000\uE1BF" + // 14280 - 14289
                "\u0000\u0000\u0000\uEAA4\u0000\u0000\u0000\uFAC2\uB3CE\u0000" + // 14290 - 14299
                "\uB3CF\uB3D0\u0000\u0000\u0000\u0000\uEDE3\u0000\u0000\uD3E1" + // 14300 - 14309
                "\u0000\uFCB1\uCCF8\u0000\u0000\uDDC6\uFAD1\u0000\uF7DF\uB3CA" + // 14310 - 14319
                "\uB3CB\u0000\uB3CC\uB3CD\u0000\u0000\u0000\uEDE7\uFBB5\uF8EC" + // 14320 - 14329
                "\u0000\u0000\u0000\uF4EB\u0000\uEEB5\u0000\uF5D8\uB3C8\u0000" + // 14330 - 14339
                "\u0000\u0000\u0000\uB3C9\u0000\u0000\uE3CC\u0000\u0000\u0000" + // 14340 - 14349
                "\uCFF8\uEFAC\u0000\uD2AF\uDCE5\u0000\u0000\u0000\u0000\uD0A5" + // 14350 - 14359
                "\uF1B4\uB3C7\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uC5F2" + // 14360 - 14369
                "\uB3C4\uB3C5\u0000\u0000\uB3C6\u0000\u0000\u0000\uDCB6\uE4E9" + // 14370 - 14379
                "\u0000\u0000\u0000\u0000\uC2EE\uC2EF\u0000\u0000\uFAC9\u0000" + // 14380 - 14389
                "\u0000\uE1CD\u0000\uCAB8\u0000\uD7A4\uCEC5\u0000\u0000\u0000" + // 14390 - 14399
                "\u0000\uCED5\uD6E6\uB3C2\uB3C3\u0000\u0000\u0000\u0000\u0000" + // 14400 - 14409
                "\u0000\uF4A8\u0000\uDCF8\u0000\uD9F8\uD4C2\u0000\u0000\u0000" + // 14410 - 14419
                "\u0000\uF6E5\u0000\uA5A9\uA5AA\u0000\u0000\u0000\u0000\u0000" + // 14420 - 14429
                "\u0000\uFBEC\u0000\uDDC8\uB3BD\u0000\u0000\u0000\uB3BE\u0000" + // 14430 - 14439
                "\u0000\u0000\uE8DA\uDAC3\uDAC4\uD4C5\u0000\uE7FA\uB3B2\uB3B3" + // 14440 - 14449
                "\u0000\uB3B4\uB3B5\uB3B6\uB3B7\uB3B8\uB3AF\uB3B0\uB3B1\u0000" + // 14450 - 14459
                "\u0000\u0000\u0000\u0000\uBBAD\uBBAE\u0000\uBBAF\uB3AA\uB3AB" + // 14460 - 14469
                "\uB3AC\u0000\uB3AD\u0000\u0000\uB3AE\uB3A4\u0000\u0000\u0000" + // 14470 - 14479
                "\uB3A5\u0000\u0000\u0000\uD7B2\u0000\u0000\u0000\u0000\uD0FD" + // 14480 - 14489
                "\uB2F6\u0000\uB2F7\u0000\uB2F8\u0000\uB2F9\u0000\uD3D2\u0000" + // 14490 - 14499
                "\uF5C0\u0000\u0000\u0000\uDFDD\u0000\uA5A1\uA5A2\uA5A3\uA5A4" + // 14500 - 14509
                "\uA5A5\uA5A6\uA5A7\uA5A8\uB2F3\u0000\u0000\u0000\u0000\u0000" + // 14510 - 14519
                "\u0000\u0000\uB7AD\uB2EF\u0000\u0000\u0000\uB2F0\u0000\u0000" + // 14520 - 14529
                "\u0000\uD6B0\uF8CA\u0000\uFCFA\u0000\uD9FE\uB2EB\uB2EC\u0000" + // 14530 - 14539
                "\u0000\uB2ED\u0000\u0000\u0000\uCDBB\u0000\uEFDA\uEED8\u0000" + // 14540 - 14549
                "\uDDA7\uB2EA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uB2BB" + // 14550 - 14559
                "\uB2E7\uB2E8\u0000\u0000\uB2E9\u0000\u0000\u0000\uCAF2\uDFA4" + // 14560 - 14569
                "\u0000\u0000\uD4C4\u0000\uE5E8\uDCC3\u0000\u0000\uEDDE\uD3F2" + // 14570 - 14579
                "\u0000\u0000\uF3F4\uF8F3\uF0C1\uDEAF\uF8B0\u0000\u0000\uE6B5" + // 14580 - 14589
                "\u0000\u0000\uF9A8\u0000\u0000\uDDD8\uB2E5\uB2E6\u0000\u0000" + // 14590 - 14599
                "\u0000\u0000\u0000\u0000\uD7CF\uEBEA\uFDEB\u0000\uCBA8\uEBBC" + // 14600 - 14609
                "\uE4BE\u0000\u0000\u0000\u0000\u0000\uF9DD\uEABF\u0000\u0000" + // 14610 - 14619
                "\uD2C0\u0000\u0000\u0000\u0000\u0000\u0000\uCAD8\u0000\u0000" + // 14620 - 14629
                "\uE5B0\u0000\u0000\u0000\uEDE5\u0000\u0000\uFDC4\u0000\uECAD" + // 14630 - 14639
                "\u0000\u0000\u0000\u0000\uC1ED\u0000\u0000\u0000\uCAB4\u0000" + // 14640 - 14649
                "\u0000\uF8E2\uCFC2\uB2DE\uB2DF\u0000\uB2E0\u0000\uB2E1\uB2E2" + // 14650 - 14659
                "\u0000\uCFDC\u0000\uD3D1\u0000\u0000\uCCB1\uF7D8\u0000\uA5B8" + // 14660 - 14669
                "\uA5B9\u0000\u0000\u0000\u0000\u0000\u0000\uD2B7\u0000\u0000" + // 14670 - 14679
                "\uF7A2\u0000\u0000\u0000\u0000\u0000\u0000\uF5A3\u0000\u0000" + // 14680 - 14689
                "\uCDB2\u0000\uDAAB\u0000\uCAA7\u0000\u0000\uEAAC\u0000\u0000" + // 14690 - 14699
                "\u0000\uCAA8\u0000\u0000\uF3DD\u0000\u0000\u0000\uE4DA\u0000" + // 14700 - 14709
                "\u0000\uFDAA\uF9E2\u0000\u0000\u0000\u0000\u0000\uF2BE\uF6A1" + // 14710 - 14719
                "\u0000\uEBCB\uB2DC\u0000\u0000\u0000\u0000\u0000\u0000\uB2DD" + // 14720 - 14729
                "\uB2D9\uB2DA\u0000\u0000\uB2DB\u0000\u0000\u0000\uCEE6\uFCAB" + // 14730 - 14739
                "\uD5BB\u0000\u0000\uF2A8\uB2D5\uB2D6\u0000\u0000\u0000\uB2D7" + // 14740 - 14749
                "\u0000\u0000\uCBBA\u0000\u0000\uE5D1\u0000\u0000\u0000\uF8D4" + // 14750 - 14759
                "\u0000\u0000\u0000\u0000\u0000\uD6F3\u0000\u0000\u0000\uD8ED" + // 14760 - 14769
                "\uE3C4\uF0F1\u0000\u0000\uD5A9\uFAE2\u0000\u0000\u0000\uD0E5" + // 14770 - 14779
                "\u0000\uF3F1\u0000\u0000\u0000\uE3D0\u0000\u0000\uF2FB\uB2D4" + // 14780 - 14789
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uB1CC\uB2D2\u0000" + // 14790 - 14799
                "\u0000\u0000\uB2D3\u0000\u0000\u0000\uF7EC\u0000\u0000\u0000" + // 14800 - 14809
                "\uE8F6\u0000\uDCFC\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 14810 - 14819
                "\uE1B0\uB2CC\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE4CB" + // 14820 - 14829
                "\uB2CA\uB2CB\u0000\u0000\u0000\u0000\u0000\u0000\uD4DF\u0000" + // 14830 - 14839
                "\u0000\u0000\uF4AD\u0000\uFCAA\u0000\u0000\u0000\uFAB0\u0000" + // 14840 - 14849
                "\u0000\u0000\u0000\uE5A6\u0000\u0000\u0000\u0000\uBEE4\uBEE5" + // 14850 - 14859
                "\u0000\uBEE6\uB2C1\u0000\uB2C2\u0000\uB2C3\u0000\u0000\u0000" + // 14860 - 14869
                "\uD0D3\u0000\uD3BD\u0000\u0000\u0000\uCCB8\u0000\u0000\u0000" + // 14870 - 14879
                "\uD0AE\uB2BE\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uEED2" + // 14880 - 14889
                "\uB2BC\u0000\u0000\u0000\u0000\uB2BD\u0000\u0000\uCFD9\u0000" + // 14890 - 14899
                "\u0000\uDCCD\uEDFB\u0000\uDEF0\uB2B9\u0000\u0000\u0000\uB2BA" + // 14900 - 14909
                "\u0000\u0000\u0000\uCCAA\u0000\u0000\uF0C3\uCCD6\u0000\uF1D8" + // 14910 - 14919
                "\u0000\u0000\uD8D8\u0000\u0000\uE0F2\u0000\uA5B0\uA5B1\uA5B2" + // 14920 - 14929
                "\uA5B3\uA5B4\uA5B5\uA5B6\uA5B7\uB2B5\u0000\u0000\uB2B6\u0000" + // 14930 - 14939
                "\uB2B7\u0000\u0000\uEEAE\uD6AE\u0000\u0000\u0000\u0000\u0000" + // 14940 - 14949
                "\uB5F0\uB5F1\u0000\u0000\uDDCF\u0000\u0000\u0000\u0000\u0000" + // 14950 - 14959
                "\u0000\uF4D1\u0000\u0000\uF2D4\u0000\uD1B0\u0000\uCCE0\u0000" + // 14960 - 14969
                "\uDBFD\uB2B2\uB2B3\u0000\u0000\uB2B4\u0000\u0000\u0000\uD2B5" + // 14970 - 14979
                "\u0000\u0000\u0000\uD3D5\u0000\uEBD9\u0000\uCFA7\uEAAF\u0000" + // 14980 - 14989
                "\u0000\u0000\u0000\uC0F7\u0000\u0000\u0000\uE8CD\u0000\u0000"   // 14990 - 14999
                ;

        private static final String index2a =
                "\u0000\u0000\uC5D7\uC5D8\u0000\u0000\uE3BD\u0000\uCFE1\uF0C0" + // 15000 - 15009
                "\uECDA\u0000\uDDD7\uB2B0\uB2B1\u0000\u0000\u0000\u0000\u0000" + // 15010 - 15019
                "\u0000\uF5CA\uE9B6\u0000\u0000\uEABE\uD9B1\u0000\u0000\u0000" + // 15020 - 15029
                "\u0000\u0000\uB5EB\uB5EC\u0000\uB5ED\uB2AB\u0000\u0000\u0000" + // 15030 - 15039
                "\uB2AC\u0000\u0000\u0000\uFDCD\u0000\u0000\u0000\uF3B6\u0000" + // 15040 - 15049
                "\uD3DC\u0000\u0000\uFAFE\u0000\u0000\u0000\u0000\uC0E7\uC0E8" + // 15050 - 15059
                "\u0000\u0000\uF2ED\u0000\uDBD9\u0000\uF0A8\u0000\u0000\uDBDF" + // 15060 - 15069
                "\uD3D3\uF8C7\u0000\u0000\u0000\u0000\uC1E1\u0000\u0000\u0000" + // 15070 - 15079
                "\uF9FD\u0000\uCADC\u0000\u0000\uDCA7\u0000\u0000\uD6E7\u0000" + // 15080 - 15089
                "\u0000\u0000\uF2B8\uF6C8\u0000\u0000\u0000\uF4BF\uE2EF\u0000" + // 15090 - 15099
                "\uD9F1\uF1C7\uB1FE\uB2A1\u0000\uB2A2\uB2A3\uB2A4\u0000\u0000" + // 15100 - 15109
                "\uFBAC\uCFC3\uEBFD\u0000\u0000\u0000\u0000\uCAE1\u0000\uD4CC" + // 15110 - 15119
                "\u0000\u0000\uDEE8\u0000\u0000\u0000\u0000\u0000\u0000\uE0D6" + // 15120 - 15129
                "\u0000\u0000\uD2C5\uFBD1\uE7C0\uEBA5\u0000\uDFFA\uE3A2\uB1FD" + // 15130 - 15139
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE8A8\uB1FA\uB1FB" + // 15140 - 15149
                "\u0000\u0000\uB1FC\u0000\u0000\u0000\uF2CA\u0000\u0000\u0000" + // 15150 - 15159
                "\u0000\u0000\uCCB0\uEAA2\u0000\u0000\uE4B7\u0000\uEADB\u0000" + // 15160 - 15169
                "\uF5FA\u0000\u0000\uF5E9\u0000\u0000\u0000\u0000\u0000\u0000" + // 15170 - 15179
                "\uCBF5\u0000\u0000\uD7B8\u0000\u0000\u0000\u0000\u0000\u0000" + // 15180 - 15189
                "\uE3B2\u0000\u0000\uDAFE\u0000\u0000\uCCBE\u0000\u0000\uF2FC" + // 15190 - 15199
                "\uB1F7\uB1F8\u0000\u0000\u0000\uB1F9\u0000\u0000\uCDCC\u0000" + // 15200 - 15209
                "\u0000\u0000\u0000\uEDD9\u0000\uCCD0\u0000\u0000\u0000\u0000" + // 15210 - 15219
                "\uCFA6\u0000\u0000\uF7B6\u0000\u0000\u0000\u0000\uF4DE\u0000" + // 15220 - 15229
                "\uA2B6\u0000\uA1C7\uA1C8\u0000\u0000\u0000\u0000\uBDF7\uBDF8" + // 15230 - 15239
                "\u0000\u0000\uDDC3\u0000\uF9DF\u0000\u0000\u0000\u0000\uBFE8" + // 15240 - 15249
                "\uBFE9\u0000\uBFEA\uB1F1\u0000\u0000\u0000\uB1F2\u0000\uB1F3" + // 15250 - 15259
                "\u0000\uF1A6\uCBD5\u0000\u0000\u0000\u0000\u0000\u0000\uE3A5" + // 15260 - 15269
                "\u0000\u0000\uD0F9\uECAB\uDED3\uF7E9\u0000\u0000\uF9F5\uB1E8" + // 15270 - 15279
                "\uB1E9\u0000\uB1EA\u0000\uB1EB\uB1EC\u0000\uF2A3\u0000\uF7F8" + // 15280 - 15289
                "\u0000\u0000\u0000\u0000\uD0B3\uB1E6\u0000\uB1E7\u0000\u0000" + // 15290 - 15299
                "\u0000\u0000\u0000\uBAF6\uBAF7\u0000\uBAF8\uB1E2\uB1E3\u0000" + // 15300 - 15309
                "\u0000\uB1E4\u0000\u0000\uB1E5\uB1DD\uB1DE\u0000\uB1DF\u0000" + // 15310 - 15319
                "\uB1E0\u0000\u0000\uFAAF\u0000\uEBFC\u0000\u0000\uE0EA\u0000" + // 15320 - 15329
                "\uF2A2\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uFDBD\uB1DB" + // 15330 - 15339
                "\uB1DC\u0000\u0000\u0000\u0000\u0000\u0000\uF6B0\uEFCF\uE9CF" + // 15340 - 15349
                "\u0000\uEEA4\u0000\u0000\u0000\u0000\uD0A4\u0000\u0000\uE2DE" + // 15350 - 15359
                "\uE1B5\u0000\u0000\uCDEF\uF1A7\uCEE5\uB1D7\uB1D8\u0000\u0000" + // 15360 - 15369
                "\uB1D9\u0000\u0000\uB1DA\uB1D5\u0000\u0000\u0000\uB1D6\u0000" + // 15370 - 15379
                "\u0000\u0000\uE5C0\uFCB5\u0000\u0000\u0000\u0000\uC2E5\u0000" + // 15380 - 15389
                "\u0000\u0000\uF9AC\u0000\u0000\u0000\u0000\uD3DB\uD6B5\uECA4" + // 15390 - 15399
                "\u0000\u0000\uE8C3\u0000\uF1C8\u0000\u0000\u0000\uCEF1\uB1D1" + // 15400 - 15409
                "\uB1D2\u0000\uB1D3\u0000\u0000\u0000\u0000\uD3F7\u0000\u0000" + // 15410 - 15419
                "\u0000\u0000\uC1DC\uC1DD\u0000\uC1DE\uB1D0\u0000\u0000\u0000" + // 15420 - 15429
                "\u0000\u0000\u0000\u0000\uCFCE\uB1CD\uB1CE\u0000\u0000\uB1CF" + // 15430 - 15439
                "\u0000\u0000\u0000\uDBC9\u0000\u0000\u0000\uE4FA\u0000\uD3A5" + // 15440 - 15449
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF7CF\uB1C8\u0000\u0000" + // 15450 - 15459
                "\u0000\u0000\u0000\u0000\u0000\uD7F7\uB1C5\uB1C6\u0000\u0000" + // 15460 - 15469
                "\uB1C7\u0000\u0000\u0000\uD1F4\uD2BA\u0000\u0000\u0000\uDFF2" + // 15470 - 15479
                "\uB1BA\u0000\u0000\uB1BB\uB1BC\uB1BD\uB1BE\u0000\uCFEA\u0000" + // 15480 - 15489
                "\u0000\uCFD0\u0000\uEACC\u0000\u0000\uD7C4\u0000\u0000\u0000" + // 15490 - 15499
                "\u0000\u0000\u0000\uD0A9\u0000\u0000\uDDDA\u0000\u0000\u0000" + // 15500 - 15509
                "\u0000\u0000\u0000\uF5A6\u0000\u0000\uE3A1\u0000\u0000\uE8E3" + // 15510 - 15519
                "\u0000\u0000\uF3AB\uB1B5\u0000\u0000\u0000\u0000\u0000\u0000" + // 15520 - 15529
                "\u0000\uCBBE\uB1B3\u0000\u0000\u0000\uB1B4\u0000\u0000\u0000" + // 15530 - 15539
                "\uEFD8\uE6C9\u0000\uD8B8\uFAF3\u0000\uCDDD\u0000\u0000\u0000" + // 15540 - 15549
                "\u0000\u0000\u0000\u0000\uD9FD\uB1AD\u0000\u0000\u0000\uB1AE" + // 15550 - 15559
                "\u0000\u0000\u0000\uDDD9\u0000\u0000\u0000\u0000\uD9E7\uB1A7" + // 15560 - 15569
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE4D5\uB1A5\u0000" + // 15570 - 15579
                "\u0000\u0000\uB1A6\u0000\u0000\u0000\uEFB8\u0000\u0000\u0000" + // 15580 - 15589
                "\uD7C0\u0000\uECD7\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 15590 - 15599
                "\uEAE6\uB0FC\u0000\u0000\u0000\uB0FD\u0000\uB0FE\u0000\uD0EF" + // 15600 - 15609
                "\u0000\u0000\uFDED\u0000\u0000\u0000\u0000\uC0B8\uC0B9\u0000" + // 15610 - 15619
                "\u0000\uCDDF\u0000\u0000\uF5CB\u0000\uE4F0\uCBAB\uB0F5\uB0F6" + // 15620 - 15629
                "\u0000\uB0F7\u0000\uB0F8\uB0F9\u0000\uF4B7\uFDC2\uFCB0\u0000" + // 15630 - 15639
                "\uFDEC\uCAE2\u0000\u0000\uF6B9\u0000\u0000\u0000\u0000\u0000" + // 15640 - 15649
                "\u0000\uF2B2\uF3F6\uF6DB\uB0F1\u0000\uB0F2\u0000\uB0F3\u0000" + // 15650 - 15659
                "\u0000\uB0F4\uB0ED\uB0EE\u0000\u0000\uB0EF\u0000\u0000\uB0F0" + // 15660 - 15669
                "\uB0E9\u0000\u0000\u0000\uB0EA\u0000\u0000\u0000\uF4AB\u0000" + // 15670 - 15679
                "\u0000\u0000\u0000\uD0BD\uB0E2\uB0E3\u0000\uB0E4\uB0E5\uB0E6" + // 15680 - 15689
                "\u0000\u0000\uDBE7\uE2BF\u0000\u0000\u0000\u0000\u0000\uB5AE" + // 15690 - 15699
                "\u0000\u0000\u0000\uD8F1\u0000\uD4CF\u0000\u0000\u0000\uFAAE" + // 15700 - 15709
                "\u0000\u0000\u0000\uD6E9\uB0E1\u0000\u0000\u0000\u0000\u0000" + // 15710 - 15719
                "\u0000\u0000\uD6DB\uB0DC\uB0DD\uB0DE\u0000\uB0DF\u0000\u0000" + // 15720 - 15729
                "\uB0E0\uB0DA\uB0DB\u0000\u0000\u0000\u0000\u0000\u0000\uA1AD" + // 15730 - 15739
                "\u0000\u0000\u0000\uD3BC\u0000\u0000\u0000\uCAF0\u0000\uCBD3" + // 15740 - 15749
                "\u0000\u0000\u0000\uE0BC\u0000\uF4CA\uD4FA\uB0D5\u0000\u0000" + // 15750 - 15759
                "\u0000\uB0D6\u0000\u0000\u0000\uFACA\u0000\u0000\u0000\uE5E3" + // 15760 - 15769
                "\u0000\uD3D0\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDFC1" + // 15770 - 15779
                "\uB0CB\uB0CC\u0000\uB0CD\uB0CE\uB0CF\uB0D0\u0000\uF4FA\u0000" + // 15780 - 15789
                "\u0000\u0000\u0000\uCDD6\uFCF6\u0000\uA2D3\uA2D4\u0000\u0000" + // 15790 - 15799
                "\u0000\uA1A5\uA1A6\u0000\u0040\u0041\u0042\u0043\u0044\u0045" + // 15800 - 15809
                "\u0046\u0047\uB0C9\u0000\uB0CA\u0000\u0000\u0000\u0000\u0000" + // 15810 - 15819
                "\uBAEE\uBAEF\u0000\uBAF0\uB0C5\uB0C6\u0000\u0000\uB0C7\u0000" + // 15820 - 15829
                "\u0000\uB0C8\uB0C3\u0000\u0000\u0000\uB0C4\u0000\u0000\u0000" + // 15830 - 15839
                "\uD8B7\uCEB1\uCAC2\u0000\u0000\uFBB4\uB0BF\u0000\u0000\u0000" + // 15840 - 15849
                "\u0000\u0000\u0000\u0000\uEAE0\uB0BC\uB0BD\u0000\u0000\uB0BE" + // 15850 - 15859
                "\u0000\u0000\u0000\uE9BD\u0000\uD7C9\u0000\u0000\uEBDB\uB0BA" + // 15860 - 15869
                "\uB0BB\u0000\u0000\u0000\u0000\u0000\u0000\uA2D0\u0000\uA2D1" + // 15870 - 15879
                "\u0000\uF9EB\uEEA3\u0000\u0000\u0000\u0000\u0000\u0000\uF2D0" + // 15880 - 15889
                "\u0000\uEAFB\uB0B5\u0000\u0000\u0000\uB0B6\u0000\u0000\u0000" + // 15890 - 15899
                "\uE9D7\uE4F1\u0000\u0000\u0000\uCAEF\uB0A8\uB0A9\uB0AA\uB0AB" + // 15900 - 15909
                "\uB0AC\uB0AD\uB0AE\uB0AF\uB0A5\uB0A6\uB0A7\u0000\u0000\u0000" + // 15910 - 15919
                "\u0000\u0000\uBAEA\uBAEB\u0000\u0000\uCDCB\u0000\u0000\u0000" + // 15920 - 15929
                "\u0000\u0000\u0000\uD6DD\u0000\u0000\uE1AB\u0000\u0000\u0000" + // 15930 - 15939
                "\u0000\u0000\u0000\uF3FB\u0000\u0000\uE0FB\u0000\u0000\u0000" + // 15940 - 15949
                "\uEFEA\uFADE\u0000\uE4EE\uF9A1\u0000\u0000\uFBEF\u0000\u0000" + // 15950 - 15959
                "\u0000\uD5B4\u0000\u0000\u0000\u0000\uC5B0\uC5B1\u0000\u0000" + // 15960 - 15969
                "\uD3E0\u0000\uE4BF\u0000\uFBC0\u0000\uDABE\uB0A1\uB0A2\u0000" + // 15970 - 15979
                "\u0000\uB0A3\u0000\u0000\uB0A4\uDBC2\u0000\u0000\u0000\u0000" + // 15980 - 15989
                "\uCAFE\u0000\u0000\uCBB9\u0000\u0000\uEDF9\u0000\u0000\u0000" + // 15990 - 15999
                "\uD9B3\u0000\u0000\uD8F4\u0000\uE9B7\uE0A9\u0000\u0000\u0000" + // 16000 - 16009
                "\u0000\u0000\u0000\u0000\uD7A2\uDCAF\u0000\u0000\u0000\u0000" + // 16010 - 16019
                "\u0000\uF0A3\u0000\uF9E0\u0000\u0000\u0000\u0000\uECD6\u0000" + // 16020 - 16029
                "\u0000\uDFC6\u0000\u0000\u0000\u0000\u0000\u0000\uDFA8\u0000" + // 16030 - 16039
                "\u0000\uF6A6\u0000\u0000\u0000\u0000\u0000\u0000\uCBCA\u0000" + // 16040 - 16049
                "\u0000\uE0DA\u0000\u0000\u0000\uEEF7\u0000\u0000\uDFA3\u0000" + // 16050 - 16059
                "\u0000\u0000\u0000\u0000\u0000\uECB7\u0000\u0000\uFDDF\u0000" + // 16060 - 16069
                "\u0000\u0000\u0000\u0000\u0000\uE5CB\u0000\u0000\uCBAA\u0000" + // 16070 - 16079
                "\u0000\u0000\u0000\u0000\u0000\uDCAD\u0000\u0000\uE4DB\u0000" + // 16080 - 16089
                "\uE1FB\uCBA2\u0000\u0000\u0000\uECBC\u0000\u0000\uE5AD\u0000" + // 16090 - 16099
                "\uA1AE\uA1AF\u0000\u0000\uA1B0\uA1B1\u0000\u0000\uE7D5\uF5BF" + // 16100 - 16109
                "\uCFA2\uCDAF\uCFA3\u0000\u0000\uCDB0\uF1FE\uD0A3\uE1AF\uF8A3" + // 16110 - 16119
                "\u0000\uCAA6\uD3DA\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 16120 - 16129
                "\uF1E7\uD9F9\u0000\u0000\uD3EA\uF5F5\u0000\uEFC7\u0000\uEDAA" + // 16130 - 16139
                "\u0000\u0000\uF2A1\uCEE1\u0000\u0000\u0000\uFDB8\uE3E8\u0000" + // 16140 - 16149
                "\uD4A7\uE8FC\uFDD3\uEBED\uD6DC\u0000\u0000\u0000\u0000\u0000" + // 16150 - 16159
                "\uBADC\u0000\u0000\u0000\uFCE1\uEDB0\uFDD1\uF6BB\u0000\u0000" + // 16160 - 16169
                "\uFBB6\u0000\u0000\u0000\u0000\u0000\u0000\uE4BD\u0000\u0000" + // 16170 - 16179
                "\uE1C5\u0000\u0000\u0000\u0000\u0000\u0000\uFDA3\u0000\uFBE5" + // 16180 - 16189
                "\uCDDC\uD9F7\u0000\u0000\u0000\u0000\u0000\u0000\uA1E5\uA1E4" + // 16190 - 16199
                "\u0000\u0000\uF5AF\u0000\u0000\u0000\u0000\uCEF0\u0000\uDCFB" + // 16200 - 16209
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF8B5\uE4E5\u0000" + // 16210 - 16219
                "\u0000\u0000\u0000\u0000\u0000\u0000\uD9F6\uE4CA\u0000\uDCE1" + // 16220 - 16229
                "\u0000\u0000\uF9C8\u0000\u0000\uF3F9\u0000\uEDF8\u0000\uF5C7" + // 16230 - 16239
                "\u0000\u0000\uF1D3\uF5E7\u0000\u0000\u0000\u0000\uCADA\uCCF4" + // 16240 - 16249
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE4E1\uCDD4\u0000" + // 16250 - 16259
                "\u0000\u0000\u0000\u0000\u0000\u0000\uEDFE\uFAB5\u0000\u0000" + // 16260 - 16269
                "\u0000\u0000\u0000\u0000\u0000\uCCC6\uCDE9\u0000\u0000\u0000" + // 16270 - 16279
                "\u0000\u0000\u0000\u0000\uE2F5\uD8E9\u0000\u0000\uF8FE\u0000" + // 16280 - 16289
                "\uCFCC\u0000\u0000\uE9C8\u0000\uCBCF\u0000\uE3C9\u0000\u0000" + // 16290 - 16299
                "\uF3D7\u0000\u0000\u0000\uDCBD\u0000\uCCE5\uD4F9\u0000\u0000" + // 16300 - 16309
                "\u0000\u0000\u0000\uF5E2\uE1D3\uDCC0\u0000\u0000\u0000\u0000" + // 16310 - 16319
                "\u0000\uD1C8\uD1C9\uF1D2\uD2CC\uCFCB\u0000\u0000\uCABD\u0000" + // 16320 - 16329
                "\u0000\uF8E3\u0000\uD4DD\u0000\u0000\uEAD8\u0000\uEEBA\u0000" + // 16330 - 16339
                "\u0000\u0000\u0000\u0000\uF8D3\u0000\uA1A9\u0000\u0000\u0000" + // 16340 - 16349
                "\uA1AA\u0000\u0000\u0000\uF9A3\uE0DB\uF6EB\u0000\uCBF1\uFBB0" + // 16350 - 16359
                "\u0000\u0000\u0000\uD8A9\uE5DF\uF9A7\u0000\uF1B2\u0000\uF1B1" + // 16360 - 16369
                "\u0000\u0000\u0000\u0000\u0000\uF2BC\uECE3\u0000\u0000\uE1B4" + // 16370 - 16379
                "\u0000\u0000\u0000\u0000\uF4D3\u0000\uACEA\uACEB\uACEC\uACED" + // 16380 - 16389
                "\uACEE\uACEF\uACF0\uACF1\uCEBD\u0000\u0000\u0000\u0000\u0000" + // 16390 - 16399
                "\u0000\u0000\uD4F8\uDCBF\u0000\u0000\u0000\u0000\u0000\u0000" + // 16400 - 16409
                "\u0000\uE6E3\uE6AE\u0000\u0000\u0000\u0000\u0000\uEFB6\u0000" + // 16410 - 16419
                "\uEFD0\u0000\uCDB1\u0000\u0000\u0000\u0000\u0000\uDFD8\u0000" + // 16420 - 16429
                "\u0000\u0000\uDCB9\u0000\u0000\u0000\uF1C0\uF3C9\u0000\u0000" + // 16430 - 16439
                "\uE4BB\u0000\u0000\u0000\u0000\uDECD\uECF3\u0000\u0000\uEDE0" + // 16440 - 16449
                "\uF9A6\u0000\u0000\u0000\u0000\u0000\u0000\uDFBD\uEAC3\u0000" + // 16450 - 16459
                "\uEFB4\u0000\u0000\u0000\uD7BE\u0000\uFDE5\uF6A3\u0000\uD9FC" + // 16460 - 16469
                "\uFDA9\u0000\uE7EE\u0000\uACE2\uACE3\uACE4\uACE5\uACE6\uACE7" + // 16470 - 16479
                "\uACE8\uACE9\uCCF2\uF7DD\u0000\uDEBA\u0000\u0000\u0000\u0000" + // 16480 - 16489
                "\uD9D7\u0000\u0000\u0000\u0000\uC1A6\uC1A7\u0000\u0000\uF4EA" + // 16490 - 16499
                "\u0000\u0000\u0000\uCEB9\u0000\u0000\uF6E0\u0000\u0000\u0000" + // 16500 - 16509
                "\u0000\uE9F3\uF2C3\uFAFA\u0000\u0000\u0000\u0000\u0000\u0000" + // 16510 - 16519
                "\u0000\uCDAC\uD6C5\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 16520 - 16529
                "\uCDDB\uE7E8\uE8D7\uDAF8\uD4CB\u0000\u0000\u0000\uF7F6\uE2CE" + // 16530 - 16539
                "\u0000\uE9F5\u0000\uE1EB\u0000\u0000\u0000\uF1B7\uEEF8\u0000" + // 16540 - 16549
                "\u0000\u0000\uD9D9\uFBE1\uFAED\uF0A2\uCCF1\u0000\uFAA3\uE2F7" + // 16550 - 16559
                "\u0000\uF8C5\u0000\u0000\u0000\u0000\u0000\uDCFA\u0000\uACDA" + // 16560 - 16569
                "\uACDB\uACDC\uACDD\uACDE\uACDF\uACE0\uACE1\uCAEA\u0000\u0000" + // 16570 - 16579
                "\uCFD4\u0000\uF8BD\u0000\u0000\uDDB4\uE4B5\uD8B0\u0000\u0000" + // 16580 - 16589
                "\u0000\u0000\uC8D1\uC8D2\u0000\u0000\uF2C2\uF6C3\u0000\uD7D2" + // 16590 - 16599
                "\u0000\u0000\uF9A2\uCFD3\u0000\u0000\u0000\u0000\u0000\u0000" + // 16600 - 16609
                "\u0000\uE0A4\uD6C4\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 16610 - 16619
                "\uD0DC\uDFAC\u0000\uD6DA\u0000\u0000\u0000\u0000\u0000\uB9C2" + // 16620 - 16629
                "\u0000\u0000\u0000\uCCD2\u0000\uDDA4\u0000\u0000\u0000\uEAD6" + // 16630 - 16639
                "\uF1B0\u0000\u0000\u0000\uE2EE\u0000\u0000\u0000\u0000\uC7F0" + // 16640 - 16649
                "\uC7F1\u0000\uC7F2\uE2CD\u0000\u0000\u0000\u0000\u0000\uEFFD" + // 16650 - 16659
                "\uF2E8\uDDD4\u0000\uEAA3\u0000\u0000\u0000\uD6C3\uD6F4\uE9EB" + // 16660 - 16669
                "\uE9EC\uE0E4\u0000\u0000\u0000\u0000\uDAA7\uEDCD\uE4D2\u0000" + // 16670 - 16679
                "\u0000\uEAA9\uE4BA\uF3A2\uCDD2\uE2CB\u0000\uFACF\u0000\u0000" + // 16680 - 16689
                "\u0000\u0000\u0000\uB9BE\u0000\u0000\u0000\uCACE\uF8C1\uD2B4" + // 16690 - 16699
                "\u0000\u0000\uDCB4\uE4F6\uD0C0\u0000\uF0B7\uEEA1\u0000\u0000" + // 16700 - 16709
                "\u0000\uE3BC\uF8D6\u0000\u0000\uDBEE\u0000\uF7CE\uFABE\u0000" + // 16710 - 16719
                "\u0000\u0000\u0000\u0000\u0000\uFCCE\u0000\uDDEE\uD7C1\u0000" + // 16720 - 16729
                "\u0000\u0000\u0000\uE5D5\u0000\u0000\uF5AE\uFBAA\u0000\u0000" + // 16730 - 16739
                "\u0000\u0000\uECFB\uEBE4\u0000\u0000\uF2E7\u0000\uD7D5\uD4B6" + // 16740 - 16749
                "\uF9E8\uF9DA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDCDC" + // 16750 - 16759
                "\uF6ED\u0000\uF9AE\u0000\uDDBE\u0000\u0000\u0000\uE5F5\u0000" + // 16760 - 16769
                "\u0000\u0000\u0000\u0000\uE9CA\u0000\uE1F0\u0000\uF9EA\uD1CE" + // 16770 - 16779
                "\uEED4\u0000\uD4D2\uD9A3\uFDA8\uD7D9\uD0B1\u0000\u0000\u0000" + // 16780 - 16789
                "\u0000\uD5EF\u0000\u0000\uD8C9\u0000\u0000\u0000\u0000\u0000" + // 16790 - 16799
                "\u0000\uFDA4\u0000\u0000\uCEA1\uF5A9\u0000\u0000\uDDF9\u0000" + // 16800 - 16809
                "\u0000\uE0FA\uEEC4\uD9DE\u0000\u0000\u0000\u0000\uC5B4\uC5B5" + // 16810 - 16819
                "\u0000\uC5B6\uCBA6\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 16820 - 16829
                "\uD7EA\uDAA6\u0000\u0000\uE0EC\u0000\u0000\u0000\u0000\uFAC6" + // 16830 - 16839
                "\u0000\u0000\u0000\u0000\uC0AB\uC0AC\u0000\uC0AD\uF7A1\u0000" + // 16840 - 16849
                "\u0000\u0000\u0000\u0000\u0000\u0000\uD5C5\uF1A4\u0000\u0000" + // 16850 - 16859
                "\u0000\u0000\u0000\u0000\u0000\uD9AD\uECF0\u0000\u0000\u0000" + // 16860 - 16869
                "\u0000\u0000\u0000\u0000\uEADA\uE7CC\u0000\uD6A8\uCEA7\u0000" + // 16870 - 16879
                "\uD4B5\u0000\u0000\uECBE\u0000\u0000\u0000\uE5B4\uCDC8\uEEC8" + // 16880 - 16889
                "\uD9E1\u0000\u0000\uE0B8\u0000\u0000\uCDD1\uF3B9\uEFFC\uD1C4" + // 16890 - 16899
                "\uEFB1\u0000\uD1C5\u0000\uD0DE\u0000\uDEC9\u0000\u0000\u0000" + // 16900 - 16909
                "\u0000\u0000\u0000\u0000\uDBD7\uF5DF\u0000\uEEB6\u0000\u0000" + // 16910 - 16919
                "\u0000\uE2F6\uD3CA\uF5DE\u0000\u0000\u0000\u0000\u0000\u0000" + // 16920 - 16929
                "\u0000\uDEEF\uCBEA\u0000\u0000\u0000\uCBBC\u0000\u0000\u0000" + // 16930 - 16939
                "\uFBE6\u0000\u0000\u0000\u0000\u0000\uE0A7\u0000\u0000\u0000" + // 16940 - 16949
                "\uD4AB\uCAB3\uCDA6\u0000\uCDC3\uD9AF\u0000\u0000\u0000\uF9E7" + // 16950 - 16959
                "\u0000\u0000\u0000\uECF4\u0000\u0000\u0000\u0000\u0000\uEFFB" + // 16960 - 16969
                "\u0000\u0000\uFAF9\uEBDE\u0000\u0000\uF5C8\u0000\uD4DE\u0000" + // 16970 - 16979
                "\u0000\uFCA3\u0000\uDBBB\u0000\u0000\u0000\uF2BA\uCBA5\u0000" + // 16980 - 16989
                "\u0000\u0000\u0000\uCBE9\u0000\u0000\uDDB3\uD4EC\u0000\u0000" + // 16990 - 16999
                "\uF2B9\u0000\uDFB7\uCCA2\uF7FE\uDFBC\u0000\u0000\u0000\u0000" + // 17000 - 17009
                "\uEBCD\uEFF9\u0000\u0000\u0000\uDDBC\uF6DC\u0000\u0000\uCADE" + // 17010 - 17019
                "\uDFE4\u0000\u0000\u0000\uE6FD\u0000\uF7BB\uF2EA\uDEC8\uE9D3" + // 17020 - 17029
                "\u0000\u0000\u0000\u0000\uBFE4\uBFE5\u0000\u0000\uDCE6\u0000" + // 17030 - 17039
                "\u0000\uDED2\u0000\u0000\uEDE2\uD7ED\uD1D1\u0000\u0000\u0000" + // 17040 - 17049
                "\u0000\u0000\uE1F2\uE5D4\u0000\u0000\u0000\u0000\u0000\u0000" + // 17050 - 17059
                "\uF3FA\uDFAB\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uFDD0" + // 17060 - 17069
                "\uDDBB\u0000\u0000\u0000\u0000\uCEAC\u0000\u0000\uF3EC\u0000" + // 17070 - 17079
                "\u0000\u0000\u0000\uDEA1\u0000\uE9D1\uF3A9\uD0E0\uE9D2\u0000" + // 17080 - 17089
                "\uDAE3\u0000\u0000\uCDFB\uF6D6\u0000\uE7F5\uE8EF\uE3F9\uD2BB" + // 17090 - 17099
                "\uEED0\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD5A1\uF7C6" + // 17100 - 17109
                "\uCFC8\u0000\u0000\u0000\uE1D0\u0000\u0000\uCECD\u0000\u0000" + // 17110 - 17119
                "\uD4DC\u0000\u0000\u0000\uDCB1\u0000\u0000\u0000\uD5D6\u0000" + // 17120 - 17129
                "\uF6B2\u0000\u0000\u0000\u0000\uCFF0\uF9BD\u0000\uACD1\uACD2" + // 17130 - 17139
                "\uACD3\uACD4\uACD5\uACD6\uACD8\uACD9\uF5BE\u0000\uDEF7\u0000" + // 17140 - 17149
                "\u0000\u0000\u0000\uCAFB\uD8B1\u0000\uDCAB\u0000\u0000\u0000" + // 17150 - 17159
                "\u0000\uD5A4\uE9AD\uD8E4\uFAB3\uE2C5\uFCBD\u0000\u0000\uECC4" + // 17160 - 17169
                "\uE0D4\u0000\uEBB6\u0000\uD7A1\uCBE8\u0000\uF9AD\uEAC0\uE1CF" + // 17170 - 17179
                "\u0000\uCCBA\u0000\u0000\u0000\u0000\uE3E2\uFBBC\uD9A4\u0000" + // 17180 - 17189
                "\u0000\uCEDD\uEBC0\u0000\uFDA2\u0000\u0000\u0000\uD7D6\u0000" + // 17190 - 17199
                "\u0000\uDCC1\u0000\uDEC6\uF9C6\uFCDA\u0000\uD4B3\uD3B9\uEADE" + // 17200 - 17209
                "\u0000\u0000\uCCB7\uDBB8\u0000\u0000\u0000\u0000\uD0E9\uECEF" + // 17210 - 17219
                "\u0000\u0000\u0000\uF9BA\u0000\uEBB5\u0000\uD7D8\u0000\uFDA7" + // 17220 - 17229
                "\u0000\u0000\u0000\u0000\uEAAB\uF0E3\uF1E4\uDCF1\uD6A7\u0000" + // 17230 - 17239
                "\u0000\u0000\u0000\uE7EF\u0000\u0000\u0000\u0000\uBFF6\uBFF7" + // 17240 - 17249
                "\u0000\u0000\uEBBF\u0000\uD7CE\uD1BF\u0000\u0000\u0000\uCCF5" + // 17250 - 17259
                "\uF5B5\uE4AD\u0000\u0000\u0000\uDBF2\u0000\u0000\u0000\u0000" + // 17260 - 17269
                "\uC4D4\uC4D5\u0000\uC4D6\uF5EF\uCFC7\u0000\u0000\uD4B2\uCCEF" + // 17270 - 17279
                "\u0000\uD4E8\uFBAD\u0000\u0000\uF8E7\u0000\uE1CE\u0000\uF7E2" + // 17280 - 17289
                "\uF7DC\uE1EA\uCEC1\uD4B1\u0000\uFDB1\uE6BD\u0000\uECD4\uEACB" + // 17290 - 17299
                "\u0000\u0000\uCABF\uD5B0\u0000\uCFE9\uE2FB\u0000\uCCA6\u0000" + // 17300 - 17309
                "\u0000\u0000\u0000\uDABB\uF2E3\uE9B4\uD2DC\u0000\u0000\u0000" + // 17310 - 17319
                "\u0000\u0000\uB7D2\uB7D3\u0000\uB7D4\uDCA9\u0000\u0000\u0000" + // 17320 - 17329
                "\u0000\uDEF6\u0000\uDCAA\uE2C3\uDCDE\u0000\uDCDF\u0000\u0000" + // 17330 - 17339
                "\uEFAD\uE6AB\uF5EE\u0000\u0000\uCABB\u0000\u0000\uE3DC\u0000" + // 17340 - 17349
                "\uE0D9\u0000\u0000\u0000\u0000\u0000\u0000\uD9D6\uCFC6\u0000" + // 17350 - 17359
                "\u0000\u0000\u0000\u0000\u0000\u0000\uEECB\uF4B2\u0000\u0000" + // 17360 - 17369
                "\u0000\u0000\u0000\u0000\u0000\uEBFB\uD4B0\uF3B2\uFBB7\u0000" + // 17370 - 17379
                "\u0000\u0000\u0000\u0000\uB7C1\uB7C2\u0000\u0000\uF8BB\u0000" + // 17380 - 17389
                "\uE8D1\u0000\u0000\u0000\u0000\uC7DC\uC7DD\u0000\uC7DE\uEBB2" + // 17390 - 17399
                "\u0000\u0000\u0000\u0000\uF1A2\u0000\u0000\uE8BA\u0000\u0000" + // 17400 - 17409
                "\u0000\uE3C7\u0000\u0000\uFADF\u0000\u0000\u0000\u0000\u0000" + // 17410 - 17419
                "\u0000\uF1CA\u0000\uCEA3\uF4C2\u0000\u0000\u0000\u0000\u0000" + // 17420 - 17429
                "\u0000\u0000\uEECA\uCEA5\u0000\u0000\u0000\u0000\u0000\u0000" + // 17430 - 17439
                "\uD6D8\uF5D9\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD9AB" + // 17440 - 17449
                "\uF1CB\u0000\u0000\uD0AF\uDDB9\u0000\u0000\uD1C3\uF1FC\u0000" + // 17450 - 17459
                "\uF3C7\u0000\u0000\uE0EB\u0000\u0000\uE6BA\u0000\u0000\uCDA9" + // 17460 - 17469
                "\u0000\u0000\u0000\uE7D1\uD2AC\u0000\uCEF9\u0000\u0000\uD4FD" + // 17470 - 17479
                "\u0000\u0000\uE0C8\u0000\u0000\u0000\uB6BB\uB6BC\uB6BD\u0000" + // 17480 - 17489
                "\u0000\uD9ED\u0000\u0000\u0000\u0000\uF5A5\u0000\uEFBF\u0000" + // 17490 - 17499
                "\u0000\u0000\u0000\u0000\uCECF\u0000\uACBA\uACBB\uACBC\uACBD" + // 17500 - 17509
                "\uACBE\uACBF\uACC0\uACC1\uDBD3\u0000\uFAE7\uD8E3\uF4C1\u0000" + // 17510 - 17519
                "\uDDB7\u0000\uE7ED\uFDC1\uDAE2\u0000\u0000\uD8B3\u0000\u0000" + // 17520 - 17529
                "\uFAA9\u0000\uE1DD\u0000\u0000\u0000\u0000\uC1CE\uC1CF\u0000" + // 17530 - 17539
                "\uC1D0\uCAB9\u0000\uEEE4\u0000\u0000\u0000\u0000\u0000\uB7AA" + // 17540 - 17549
                "\uB7AB\u0000\u0000\uE3B1\uFCEB\uCDA8\u0000\uCCB6\u0000\u0000" + // 17550 - 17559
                "\uF6C2\u0000\u0000\uEFBB\u0000\u0000\u0000\uB6A6\uB6A7\uB6A8" + // 17560 - 17569
                "\u0000\u0000\uDBCC\uDDCD\u0000\u0000\u0000\uD4C8\u0000\uCFA1" + // 17570 - 17579
                "\uE4A8\u0000\uF4B6\uECFE\u0000\u0000\uE3AE\uF9C5\uDDD3\uD6F1" + // 17580 - 17589
                "\uECFC\uFCF0\u0000\u0000\uEDC0\uD3E8\u0000\u0000\uDEA8\uF4E4" + // 17590 - 17599
                "\uECC2\u0000\uD9F5\uE1AE\u0000\u0000\uECC3\uCFFE\u0000\uF8BF" + // 17600 - 17609
                "\uD8E2\uFCA7\uF7FC\uF7B1\uCEBB\uF4A1\u0000\u0000\uEECD\uDDB6" + // 17610 - 17619
                "\uEEAF\uCDF8\u0000\u0000\u0000\u0000\uDEB8\uD1C2\u0000\uF9A5" + // 17620 - 17629
                "\u0000\uE8D5\u0000\u0000\u0000\uCCF7\u0000\u0000\u0000\u0000" + // 17630 - 17639
                "\u0000\uF5DD\u0000\u0000\u0000\uEAD4\u0000\u0000\u0000\u0000" + // 17640 - 17649
                "\uC5E1\u0000\u0000\u0000\uDFA7\u0000\uDFE7\uE1C1\u0000\uEDDD" + // 17650 - 17659
                "\uCEC4\u0000\uCBA1\u0000\u0000\u0000\u0000\uBFA9\uBFAA\uBFAB" + // 17660 - 17669
                "\u0000\uACB2\uACB3\uACB4\uACB5\uACB6\uACB7\uACB8\uACB9\uE2C2" + // 17670 - 17679
                "\u0000\uF3D8\uE5D3\u0000\u0000\uF3D9\u0000\uDCD3\u0000\u0000" + // 17680 - 17689
                "\u0000\u0000\uDDE2\uFBF9\uDDC1\uD4C1\u0000\u0000\u0000\u0000" + // 17690 - 17699
                "\u0000\u0000\u0000\uD6A4\uECA1\u0000\u0000\u0000\uCCB9\u0000" + // 17700 - 17709
                "\u0000\uFBDE\uE3DB\u0000\uD3C9\u0000\uDCCF\u0000\u0000\u0000" + // 17710 - 17719
                "\uF8C0\u0000\u0000\u0000\uD3DD\u0000\uFDD6\u0000\u0000\u0000" + // 17720 - 17729
                "\u0000\uF8D1\u0000\uF8D2\uD9C8\u0000\u0000\uEEE3\uD7BD\u0000" + // 17730 - 17739
                "\u0000\u0000\uDBA9\u0000\u0000\uD3BB\uCAEC\u0000\uCFE8\u0000" + // 17740 - 17749
                "\uEDC3\uD0B2\u0000\u0000\uCEFE\uDAA8\uE9AA\u0000\u0000\u0000" + // 17750 - 17759
                "\u0000\u0000\u0000\u0000\uEEE1\uDACD\u0000\u0000\u0000\uF9CC" + // 17760 - 17769
                "\u0000\uE1DA\uDBBF\uD9C7\uE4D7\uEADD\u0000\uD4F7\u0000\u0000" + // 17770 - 17779
                "\u0000\uEAAE\uEAAD\u0000\u0000\uD3F1\u0000\uCAEB\uD9E2\u0000" + // 17780 - 17789
                "\uFDB2\u0000\uE3AD\uD6CC\uD9B4\uF0B3\u0000\uE5EC\u0000\u0000" + // 17790 - 17799
                "\u0000\uD1E7\u0000\uCAA5\u0000\u0000\uD6AB\uD0C2\u0000\u0000" + // 17800 - 17809
                "\u0000\uD8BD\u0000\u0000\u0000\u0000\uC5A1\uC5A2\u0000\uC5A3" + // 17810 - 17819
                "\uE2C1\u0000\uCEA4\u0000\u0000\u0000\u0000\u0000\uB6D1\uB6D2" + // 17820 - 17829
                "\u0000\u0000\uF2DB\uE4FC\u0000\u0000\u0000\u0000\u0000\uB3A6" + // 17830 - 17839
                "\uB3A7\u0000\uB3A8\uFCEF\u0000\uE0E3\u0000\u0000\u0000\u0000" + // 17840 - 17849
                "\u0000\uB6CF\u0000\u0000\u0000\uE1F6\uDECC\u0000\u0000\uFCDE" + // 17850 - 17859
                "\u0000\uDBF9\uD7B1\u0000\u0000\u0000\uCBFC\u0000\u0000\uFDF0" + // 17860 - 17869
                "\u0000\uE0BD\uCEE3\u0000\u0000\u0000\uEDB9\uF1C5\u0000\uF3CF" + // 17870 - 17879
                "\uD7AB\uE1A4\uCDAB\u0000\uD9F4\uE8A6\uCDCE\uE1E9\u0000\uCFE7" + // 17880 - 17889
                "\uF3CB\uEDA9\uCABE\u0000\u0000\u0000\u0000\uBEEE\uBEEF\u0000" + // 17890 - 17899
                "\u0000\uD2EC\u0000\u0000\u0000\u0000\u0000\u0000\uF9C2\u0000" + // 17900 - 17909
                "\uEABC\uEBAF\u0000\u0000\u0000\u0000\u0000\uE5DE\u0000\uD3F0" + // 17910 - 17919
                "\u0000\u0000\u0000\u0000\u0000\uF0A4\uE1EC\uFDFE\uFCA5\uFAB1" + // 17920 - 17929
                "\uDFD9\u0000\uE0D2\u0000\u0000\uCCB5\u0000\u0000\u0000\u0000" + // 17930 - 17939
                "\u0000\uCFBD\uE7C9\u0000\uE2F3\uE7E1\u0000\u0000\uE3CB\u0000" + // 17940 - 17949
                "\uE7D4\u0000\uCACA\u0000\u0000\u0000\uD9FB\u0000\uACAA\uACAB" + // 17950 - 17959
                "\uACAC\uACAD\uACAE\uACAF\uACB0\uACB1\uDEF1\u0000\u0000\u0000" + // 17960 - 17969
                "\uF0DF\uF8C4\u0000\u0000\uF7D6\uDEEA\uCBB4\u0000\u0000\uEFBE" + // 17970 - 17979
                "\u0000\uD3EF\u0000\u0000\uECD3\u0000\u0000\uDDC2\uEFB7\uD1C0" + // 17980 - 17989
                "\u0000\u0000\uE8C5\u0000\uE4B8\u0000\uE1E8\uCDAA\u0000\uE3F2" + // 17990 - 17999
                "\u0000\uFBF7\u0000\uF7D0\u0000\uF4C8\uE8EA\uF5F3\u0000\u0000" + // 18000 - 18009
                "\uF9DE\u0000\u0000\uEED5\u0000\u0000\u0000\u0000\uF9F4\u0000" + // 18010 - 18019
                "\uACA1\uACA2\uACA3\uACA4\uACA5\uACA6\uACA8\uACA9\uE5EB\u0000" + // 18020 - 18029
                "\uEFF4\uDDB5\u0000\u0000\u0000\u0000\uE5E7\u0000\u0000\u0000" + // 18030 - 18039
                "\uCAA3\uF5BA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF6C5" + // 18040 - 18049
                "\uCEBA\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF0EA\uD7AE" + // 18050 - 18059
                "\u0000\u0000\uE0E1\u0000\u0000\u0000\u0000\uEEEC\u0000\u0000" + // 18060 - 18069
                "\uD3A3\uEEB7\uEADC\uDBD2\u0000\u0000\u0000\u0000\u0000\u0000" + // 18070 - 18079
                "\uD7A3\u0000\u0000\uE1C9\uCAFA\u0000\u0000\u0000\u0000\u0000" + // 18080 - 18089
                "\uB2F4\uB2F5\u0000\u0000\uEFA5\u0000\uD3CC\uDAED\u0000\u0000" + // 18090 - 18099
                "\u0000\uB2FA\uB2FB\uB2FC\u0000\uB2FD\uCAE8\u0000\uF8E6\uDCCE" + // 18100 - 18109
                "\u0000\u0000\u0000\u0000\uECF2\u0000\u0000\u0000\u0000\uBEDF" + // 18110 - 18119
                "\uBEE0\u0000\u0000\uE2B6\u0000\u0000\u0000\u0000\uEFF1\u0000" + // 18120 - 18129
                "\uFCC5\uCBC2\u0000\u0000\u0000\u0000\uFDD5\u0000\uA5F7\uA5F8" + // 18130 - 18139
                "\u0000\u0000\u0000\u0000\u0000\u0000\uD8AD\u0000\u0000\uDDE4" + // 18140 - 18149
                "\uF0EF\uF6F1\uFAF0\u0000\u0000\uD1F5\uD5A3\u0000\u0000\u0000" + // 18150 - 18159
                "\u0000\uF0B2\u0000\u0000\uD5DC\uF3C4\uCBD7\u0000\u0000\u0000" + // 18160 - 18169
                "\u0000\uC7B6\u0000\u0000\u0000\uA1E3\uA1E2\u0000\u0000\uA2BA" + // 18170 - 18179
                "\uA2B9\uDEA5\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uCBCD" + // 18180 - 18189
                "\uCAB7\u0000\u0000\uD3E7\u0000\uF8E5\u0000\u0000\uCADB\u0000" + // 18190 - 18199
                "\u0000\u0000\u0000\u0000\uFCD7\uE1F1\u0000\uD2A4\u0000\u0000" + // 18200 - 18209
                "\u0000\u0000\uF5FB\uF8FA\u0000\u0000\uDFB9\u0000\u0000\u0000" + // 18210 - 18219
                "\u0000\uEAE1\u0000\u0000\uDCE3\uDFAD\uFAEC\u0000\u0000\u0000" + // 18220 - 18229
                "\u0000\u0000\uF1EB\u0000\uCEAE\u0000\u0000\u0000\u0000\uD9A2" + // 18230 - 18239
                "\u0000\u0000\uF2EC\u0000\u0000\uFAEE\u0000\u0000\u0000\uDAB7" + // 18240 - 18249
                "\u0000\u0000\u0000\u0000\uC7A5\u0000\u0000\u0000\uDDFB\u0000" + // 18250 - 18259
                "\u0000\u0000\u0000\uC7B4\u0000\u0000\u0000\uCDA7\u0000\u0000" + // 18260 - 18269
                "\uD0AC\u0000\uA5F0\uA5F1\u0000\uA5F2\uA5F3\uA5F4\uA5F5\uA5F6" + // 18270 - 18279
                "\uDAF6\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uFBEE\uE8C4" + // 18280 - 18289
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD9DF\uE3A4\u0000" + // 18290 - 18299
                "\u0000\u0000\u0000\u0000\u0000\u0000\uE2B4\uF2BB\u0000\uDEA4" + // 18300 - 18309
                "\u0000\uDACC\u0000\u0000\u0000\uE1A7\uEED3\uD0C3\u0000\u0000" + // 18310 - 18319
                "\u0000\uFBFD\u0000\u0000\u0000\u0000\uC5A9\uC5AA\u0000\u0000" + // 18320 - 18329
                "\uDFD5\u0000\u0000\uEDD7\u0000\u0000\u0000\uA1D6\u0000\u0000" + // 18330 - 18339
                "\u0000\u0000\u0000\uE1CA\uEBE3\u0000\uF2DE\uE8A5\u0000\u0000" + // 18340 - 18349
                "\u0000\u0000\u0000\u0000\u0000\uE6DA\uF5B9\u0000\uDCF0\uE3F1" + // 18350 - 18359
                "\u0000\u0000\u0000\u0000\uA8F7\uA8F8\u0000\u0000\u0000\uD7CD" + // 18360 - 18369
                "\u0000\u0000\uD4D1\uE9ED\u0000\uD0ED\u0000\u0000\u0000\u0000" + // 18370 - 18379
                "\u0000\uE5F7\u0000\uA5E8\uA5E9\uA5EA\uA5EB\uA5EC\uA5ED\uA5EE" + // 18380 - 18389
                "\uA5EF\uF9CB\u0000\u0000\u0000\uCBF3\uF4A5\u0000\u0000\uD4F5" + // 18390 - 18399
                "\u0000\uD0C9\uEFA7\uE2EC\u0000\uDBEA\uFDB9\u0000\u0000\u0000" + // 18400 - 18409
                "\u0000\u0000\u0000\u0000\uF5B8\uFDBC\uDFB1\uE3EF\u0000\u0000" + // 18410 - 18419
                "\u0000\u0000\uE0A3\uCFF2\uF7B9\uD9F3\u0000\u0000\uE1CB\u0000" + // 18420 - 18429
                "\u0000\uDFE9\u0000\uEEDE\u0000\u0000\uF7C2\u0000\uEEF0\u0000" + // 18430 - 18439
                "\u0000\u0000\uCCC2\u0000\u0000\u0000\uD8BC\uF2B0\u0000\u0000" + // 18440 - 18449
                "\u0000\uF9D4\uF7CA\u0000\u0000\uD6C8\uDFB8\u0000\uEAA5\u0000" + // 18450 - 18459
                "\u0000\u0000\uD7AD\u0000\uEEEF\uD5D7\uEAE4\uF8A2\uCDEB\uD7BF" + // 18460 - 18469
                "\uFBB1\u0000\uA5D7\uA5D8\u0000\u0000\u0000\u0000\u0000\u0000" + // 18470 - 18479
                "\uE1FD\u0000\u0000\uD8F5\u0000\u0000\u0000\uCCCE\u0000\u0000" + // 18480 - 18489
                "\uE3B5\u0000\u0000\u0000\u0000\u0000\u0000\uE9BE\u0000\u0000" + // 18490 - 18499
                "\uF9AF\u0000\u0000\u0000\u0000\u0000\uD2FB\uE1E0\u0000\uD9AC" + // 18500 - 18509
                "\u0000\uF5EB\u0000\uE0B6\u0000\uE0BB\uCEC3\u0000\uD0BA\uF7BA" + // 18510 - 18519
                "\uD8F3\uF7CD\u0000\uA5D0\uA5D1\u0000\uA5D2\uA5D3\uA5D4\uA5D5" + // 18520 - 18529
                "\uA5D6\uF1FA\u0000\u0000\uE5B6\uF3EF\u0000\u0000\uFBDA\uE2BD" + // 18530 - 18539
                "\u0000\u0000\u0000\uE3C8\u0000\u0000\u0000\uE4AE\u0000\u0000" + // 18540 - 18549
                "\u0000\u0000\u0000\uCDCD\u0000\u0000\u0000\uDBCB\uDAB5\u0000" + // 18550 - 18559
                "\u0000\u0000\uF5D3\u0000\u0000\uECDC\uF7B7\uEBAD\u0000\u0000" + // 18560 - 18569
                "\u0000\u0000\uD5AA\u0000\u0000\uCFE2\uCDF6\u0000\u0000\uEFF0" + // 18570 - 18579
                "\u0000\uF4BE\uCDCA\uD7D4\uDEA3\u0000\uE4E0\u0000\u0000\u0000" + // 18580 - 18589
                "\uEEB9\u0000\u0000\u0000\u0000\uD5E3\uE3EE\u0000\u0000\u0000" + // 18590 - 18599
                "\u0000\u0000\uE8D4\u0000\uE8B4\uEBC3\u0000\uEAAA\uFAFC\uF5F6" + // 18600 - 18609
                "\uF0BC\uFDD4\uE2F0\u0000\u0000\u0000\u0000\u0000\u0000\uFABB" + // 18610 - 18619
                "\uE9C7\uE6AA\u0000\u0000\u0000\u0000\u0000\u0000\uCEF6\u0000" + // 18620 - 18629
                "\uFAD0\uEDBC\u0000\u0000\uD8D4\u0000\u0000\u0000\uDCDA\uE9FD" + // 18630 - 18639
                "\uD0CA\u0000\uF5D6\uD9C5\uE4B4\u0000\uEDA7\uF5AC\u0000\u0000" + // 18640 - 18649
                "\u0000\u0000\u0000\uE4F5\u0000\uEBF0\uF1D6\u0000\u0000\uE5E2" + // 18650 - 18659
                "\u0000\uCCCC\u0000\uA5C8\uA5C9\uA5CA\uA5CB\uA5CC\uA5CD\uA5CE" + // 18660 - 18669
                "\uA5CF\uDDB2\u0000\u0000\u0000\u0000\uE6A9\u0000\uEFF3\uFDE9" + // 18670 - 18679
                "\u0000\uCFC1\u0000\uE0DF\uDEEC\u0000\u0000\uD0D7\u0000\u0000" + // 18680 - 18689
                "\u0000\u0000\u0000\u0000\uE6AC\u0000\uCEDE\uD7FC\u0000\uEDBB" + // 18690 - 18699
                "\u0000\u0000\uF6AB\u0000\u0000\uF5A7\u0000\u0000\u0000\u0000" + // 18700 - 18709
                "\u0000\u0000\uF8AF\uEFF6\u0000\uDFA1\uDDE1\u0000\u0000\u0000" + // 18710 - 18719
                "\u0000\u0000\u0000\uF2CD\u0000\u0000\uE8C6\u0000\u0000\u0000" + // 18720 - 18729
                "\u0000\u0000\u0000\uE8E5\u0000\u0000\uCAAA\uE1F9\u0000\uEAB1" + // 18730 - 18739
                "\u0000\u0000\u0000\uD5BE\u0000\u0000\u0000\u0000\uC6FB\uC6FC" + // 18740 - 18749
                "\u0000\uC6FD\uD5C4\u0000\u0000\u0000\u0000\u0000\u0000\uEDF4" + // 18750 - 18759
                "\uD4EB\u0000\uDEA2\u0000\u0000\u0000\uE5E6\u0000\uCED3\u0000" + // 18760 - 18769
                "\u0000\u0000\u0000\uDCF7\u0000\u0000\uD7A5\u0000\u0000\u0000" + // 18770 - 18779
                "\u0000\uF7E8\u0000\uA2A8\uA2AB\uA2AA\uA2AD\uA2A6\uA2A9\u0000" + // 18780 - 18789
                "\u0000\uCCCD\u0000\uDAFA\u0000\uF6CF\u0000\uE9B8\uF8B3\u0000" + // 18790 - 18799
                "\u0000\u0000\u0000\u0000\u0000\u0000\uCFEF\uCEEF\u0000\u0000" + // 18800 - 18809
                "\uF2F3\u0000\u0000\u0000\u0000\uC8B8\uC8B9\u0000\u0000\uEDB8" + // 18810 - 18819
                "\u0000\u0000\u0000\uDBB6\u0000\u0000\uFDF8\uFDF9\u0000\u0000" + // 18820 - 18829
                "\u0000\uF6BF\u0000\uF7DE\u0000\u0000\u0000\u0000\u0000\u0000" + // 18830 - 18839
                "\u0000\uA9FA\uCCAE\u0000\uDADB\u0000\u0000\u0000\u0000\uCDC7" + // 18840 - 18849
                "\uDBB9\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD0B4\uEDF3" + // 18850 - 18859
                "\uDCD9\uE0CD\u0000\u0000\u0000\u0000\uF7DA\uE9A6\uCBF2\u0000" + // 18860 - 18869
                "\u0000\u0000\u0000\u0000\u0000\uFBE2\u0000\uCDD3\uDDAF\uDDB0" + // 18870 - 18879
                "\u0000\u0000\uCBB7\uE8D3\u0000\u0000\uE0B5\u0000\u0000\u0000" + // 18880 - 18889
                "\u0000\u0000\u0000\uE3CE\u0000\u0000\uE6F0\u0000\u0000\u0000" + // 18890 - 18899
                "\u0000\u0000\u0000\uD0AD\u0000\u0000\uFBC7\uD5C8\u0000\uD7DF" + // 18900 - 18909
                "\u0000\uDDA9\u0000\uCCA7\uEAC9\u0000\u0000\u0000\u0000\u0000" + // 18910 - 18919
                "\uF8B6\uE8D2\u0000\uCAC5\uCCEB\u0000\u0000\u0000\u0000\uC7EC" + // 18920 - 18929
                "\uC7ED\u0000\u0000\uDDF8\u0000\u0000\u0000\u0000\u0000\uE8CF" + // 18930 - 18939
                "\uD8E6\u0000\uF4B1\u0000\u0000\u0000\u0000\u0000\uB4AE\uB4AF" + // 18940 - 18949
                "\u0000\uB4B0\uFDD8\u0000\u0000\u0000\u0000\uD2F6\u0000\u0000" + // 18950 - 18959
                "\uD1B3\u0000\u0000\u0000\u0000\u0000\uEFED\uF5C6\u0000\uE1A2" + // 18960 - 18969
                "\uE9C6\u0000\u0000\u0000\uF2C5\uDEBD\u0000\uF6A9\u0000\u0000" + // 18970 - 18979
                "\u0000\uDAA4\u0000\uCBDA\u0000\uE7D2\uD7C3\uF6F0\uE8DE\u0000" + // 18980 - 18989
                "\u0000\uD8B4\u0000\u0000\u0000\u0000\u0000\u0000\uF5B7\uE0F0" + // 18990 - 18999
                "\u0000\uA2B0\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000" + // 19000 - 19009
                "\uA2A7\uE9FC\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uEDB6" + // 19010 - 19019
                "\uD2B1\u0000\u0000\u0000\u0000\uCCE9\u0000\uD9C4\uE9A5\uD6D5" + // 19020 - 19029
                "\u0000\uCDC5\u0000\uEDBA\uD1BD\u0000\uDCE4\u0000\uE5EF\u0000" + // 19030 - 19039
                "\u0000\u0000\u0000\u0000\uF6DE\u0000\u0000\u0000\uE5C7\uD6AC" + // 19040 - 19049
                "\u0000\u0000\u0000\uE8CB\u0000\u0000\uF8DD\u0000\uA9A8\uA8A9" + // 19050 - 19059
                "\uA9A9\u0000\u0000\u0000\u0000\u0000\uD6FE\u0000\u0000\u0000" + // 19060 - 19069
                "\uDEDB\u0000\u0000\uEFDE\u0000\u0038\u0039\u003A\u003B\u003C" + // 19070 - 19079
                "\u003D\u003E\u003F\uF9D7\u0000\u0000\u0000\u0000\u0000\u0000" + // 19080 - 19089
                "\u0000\uD9EE\uF6A8\uDDFD\u0000\u0000\u0000\u0000\u0000\u0000" + // 19090 - 19099
                "\uD4E9\u0000\u0000\uE2B2\u0000\uD4BD\u0000\u0000\uD9CE\u0000" + // 19100 - 19109
                "\uFAEF\uE3E1\u0000\u0000\u0000\u0000\u0000\u0000\uCDBA\u0000" + // 19110 - 19119
                "\u0000\uDECF\u0000\u0000\u0000\u0000\u0000\u0000\uF0CF\uF3BE" + // 19120 - 19129
                "\uE2AC\uD7E8\uCBD8\u0000\u0000\u0000\uE9E2\u0000\u0000\uE0FD" + // 19130 - 19139
                "\u0000\u0000\uD8F8\u0000\u0000\u0000\uBDC8\uBDC9\uBDCA\u0000" + // 19140 - 19149
                "\uBDCB\uE0CC\uEBF9\u0000\u0000\u0000\u0000\u0000\u0000\uF0E7" + // 19150 - 19159
                "\uE2CC\u0000\uF3A7\u0000\u0000\uCDEA\u0000\uEBEE\u0000\u0000" + // 19160 - 19169
                "\uECF5\uE8EE\u0000\uCBA9\uF1AF\u0000\u0000\uEACE\u0000\uE8DF" + // 19170 - 19179
                "\u0000\u0000\u0000\u0000\uC1C4\u0000\u0000\u0000\uECB9\u0000" + // 19180 - 19189
                "\u0000\u0000\u0000\uC6DF\uC6E0\u0000\uC6E1\uD1BC\u0000\uE5CF" + // 19190 - 19199
                "\u0000\uCBB6\u0000\uDAB8\u0000\uF6B6\u0000\uCEC2\uD6C7\u0000" + // 19200 - 19209
                "\uE3B4\u0000\uF1AD\uDBEB\u0000\uDFFE\u0000\u0000\uD8E1\u0000" + // 19210 - 19219
                "\uF7F3\uCEB8\u0000\u0000\u0000\uD8D2\uF9D6\u0000\u0000\uF2D7" + // 19220 - 19229
                "\u0000\uCAF8\uDAEF\u0000\u0000\uD6D4\uF9FC\u0000\uCCC0\u0000" + // 19230 - 19239
                "\u0000\u0000\u0000\u0000\uB3DB\uB3DC\u0000\uB3DD\uF3C5\u0000" + // 19240 - 19249
                "\u0000\uD4C0\uD5BF\u0000\u0000\u0000\uD5AF\u0000\u0000\u0000" + // 19250 - 19259
                "\uD6F5\u0000\uDBD8\u0000\u0000\uCAA2\u0000\u0000\uD1CD\u0000" + // 19260 - 19269
                "\uA9A7\u0000\u0000\u0000\u0000\u0000\u0000\uA8A8\uF0DB\u0000" + // 19270 - 19279
                "\u0000\u0000\u0000\u0000\u0000\u0000\uEBC9\uF8BA\uE8D0\u0000" + // 19280 - 19289
                "\u0000\uD8FB\u0000\u0000\uEAD5\uF4F3\uDAC9\u0000\uE6DE\u0000" + // 19290 - 19299
                "\u0000\u0000\u0000\uC7C7\uC7C8\u0000\u0000\uF8DE\uF9AA\uCAF7" + // 19300 - 19309
                "\u0000\uEDB7\u0000\u0000\uCCDE\u0000\u0000\u0000\u0000\u0000" + // 19310 - 19319
                "\u0000\uD8D5\u0000\u0000\uD7B5\u0000\u0000\u0000\u0000\u0000" + // 19320 - 19329
                "\u0000\uF4E9\uD6EC\uEBD3\uF9E5\u0000\uE0CA\u0000\u0000\uF2FD" + // 19330 - 19339
                "\uD3B0\u0000\uF1A5\uE9CE\u0000\u0000\u0000\uF9BC\u0000\u0000" + // 19340 - 19349
                "\uEDAB\u0000\uE3B7\u0000\u0000\u0000\u0000\uC1B5\u0000\u0000" + // 19350 - 19359
                "\u0000\uCAE5\u0000\u0000\u0000\uDCA1\uD7C6\u0000\uD1BB\uF7AA" + // 19360 - 19369
                "\u0000\uEDCA\uD7D3\uD8FA\uD6E0\u0000\uF1C6\u0000\u0000\u0000" + // 19370 - 19379
                "\u0000\u0000\uB3A2\uB3A3\u0000\u0000\uD7BA\u0000\uF2D5\uF5E5" + // 19380 - 19389
                "\uD9EF\u0000\u0000\uF8EE\u0000\u0000\u0000\uDEB5\u0000\u0000" + // 19390 - 19399
                "\uDCC7\u0000\u0000\u0000\u0000\u0000\uE1A9\uEBD6\u0000\uECDF" + // 19400 - 19409
                "\u0000\u0000\u0000\uDFFC\u0000\uF8CF\u0000\u0000\u0000\u0000" + // 19410 - 19419
                "\uEAC8\uEEB8\uF1AC\uCCBF\u0000\u0000\u0000\u0000\u0000\u0000" + // 19420 - 19429
                "\u0000\uE7BD\uE1AC\u0000\u0000\uE3EB\u0000\uEEC7\u0000\u0000" + // 19430 - 19439
                "\uF1F7\u0000\u0000\u0000\uE8B8\u0000\u0000\uE5C8\u0000\u0000" + // 19440 - 19449
                "\u0000\uFBA4\uD4B9\u0000\uE3DF\u0000\uDEC3\u0000\uDEC4\uCAA1" + // 19450 - 19459
                "\u0000\u0000\uD9D8\u0000\u0000\u0000\u0000\u0000\u0000\uE0C3" + // 19460 - 19469
                "\u0000\u0000\uD8D9\u0000\uF4A3\u0000\u0000\uF4DD\u0000\uA9AA" + // 19470 - 19479
                "\u0000\u0000\u0000\u0000\u0000\uA9AD\u0000\u0030\u0031\u0032" + // 19480 - 19489
                "\u0033\u0034\u0035\u0036\u0037\uE1C8\uDBB7\uDFE3\u0000\u0000" + // 19490 - 19499
                "\u0000\u0000\u0000\uB2F1\uB2F2\u0000\u0000\uDCBA\u0000\u0000" + // 19500 - 19509
                "\uCCB4\u0000\u0000\u0000\uDEAC\u0000\u0000\u0000\u0000\uF1DA" + // 19510 - 19519
                "\u0000\uFAF2\u0000\u0000\uDEB4\u0000\u0000\u0000\u0000\u0000" + // 19520 - 19529
                "\u0000\uF6EC\uDADC\uFAE4\uCEDC\uF2B5\uD0E4\uDDD1\u0000\u0000" + // 19530 - 19539
                "\u0000\u0000\uC7B9\u0000\u0000\uC7BA\uD2EF\u0000\u0000\u0000" + // 19540 - 19549
                "\uE2ED\u0000\u0000\uDEE9\uFCBC\u0000\uDAA2\uDAA3\u0000\uD2A1" + // 19550 - 19559
                "\u0000\u0000\uE2DB\u0000\u0000\u0000\uDFDE\u0000\uE0C7\uD1BA" + // 19560 - 19569
                "\u0000\uF1C4\u0000\uE5B3\uFBF5\uE9E1\uFDE0\uCBB3\u0000\u0000" + // 19570 - 19579
                "\u0000\u0000\u0000\u0000\uD5DD\uEFC4\u0000\u0000\u0000\u0000" + // 19580 - 19589
                "\u0000\u0000\uE1D8\uD6EB\u0000\u0000\u0000\uF4D9\u0000\u0000" + // 19590 - 19599
                "\u0000\uA2C6\u0000\u0000\u0000\u0000\u0000\uD4F6\uE5B7\u0000" + // 19600 - 19609
                "\u0000\uD6A2\u0000\uEDF0\u0000\u0000\u0000\u0000\uC5E2\u0000" + // 19610 - 19619
                "\u0000\u0000\uFCDC\u0000\u0000\u0000\u0000\uDCD5\u0000\uF7B5" + // 19620 - 19629
                "\uFCF3\uF0F3\uE1C6\u0000\u0000\u0000\uD4BF\u0000\u0000\u0000" + // 19630 - 19639
                "\uA1F8\uA1F9\u0000\u0000\uA1F6\uA1F7\uEFBD\uFCD6\u0000\u0000" + // 19640 - 19649
                "\uDBF4\u0000\uEFAA\uF8B9\uEEC6\u0000\u0000\u0000\u0000\u0000" + // 19650 - 19659
                "\u0000\u0000\uFDC8\uD0B5\u0000\u0000\u0000\u0000\u0000\u0000" + // 19660 - 19669
                "\u0000\uECCB\uCECC\uF5E8\uF7D5\u0000\uD3CD\u0000\uF3FE\u0000" + // 19670 - 19679
                "\uDBE9\uFDCC\u0000\u0000\u0000\u0000\u0000\u0000\uE1DF\u0000" + // 19680 - 19689
                "\uDAD1\uD8A2\u0000\u0000\u0000\u0000\u0000\uDDAC\u0000\uE4E6" + // 19690 - 19699
                "\u0000\uF1EA\u0000\u0000\u0000\uCBEC\uCBC0\uEDF1\u0000\u0000" + // 19700 - 19709
                "\uF1E2\u0000\uD4DB\u0000\u0000\uFBA8\uD0A8\u0000\u0000\uDAEC" + // 19710 - 19719
                "\u0000\u0000\uE7B8\u0000\u0000\u0000\u0000\u0000\u0000\uCFC0" + // 19720 - 19729
                "\u0000\uE6A8\uCFD7\u0000\u0000\u0000\u0000\u0000\u0000\uCFDF" + // 19730 - 19739
                "\uE9A1\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uD7A9\uE5CD" + // 19740 - 19749
                "\u0000\u0000\u0000\uFAEB\u0000\uCFBC\u0000\uDCF5\uE0B9\u0000" + // 19750 - 19759
                "\u0000\u0000\uD4CE\u0000\uF4B5\uCDDA\u0000\u0000\u0000\u0000" + // 19760 - 19769
                "\u0000\uD9CF\u0000\uE4A7\uECD2\u0000\u0000\uF6B1\u0000\u0000" + // 19770 - 19779
                "\uCEFB\uF0D6\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uF4F2" + // 19780 - 19789
                "\uCAD9\u0000\u0000\uEFEF\u0000\uF5AA\u0000\u0000\uFADD\u0000" + // 19790 - 19799
                "\u0000\u0000\u0000\u0000\u0000\uEBD7\u0000\u0000\uE2AB\uF3E8" + // 19800 - 19809
                "\u0000\u0000\u0000\u0000\u0000\uB0FA\uB0FB\u0000\u0000\uFCCF" + // 19810 - 19819
                "\uFBA2\u0000\uE0DC\u0000\u0000\u0000\uF0EC\u0000\u0000\u0000" + // 19820 - 19829
                "\u0000\uCCF6\u0000\u0000\uD3BA\u0000\uFAFB\u0000\u0000\uFABD" + // 19830 - 19839
                "\uCCC8\uEFCD\uD5D5\u0000\uA9A3\u0000\u0000\u0000\u0000\u0000" + // 19840 - 19849
                "\u0000\uA1C0\uFDFC\u0000\u0000\u0000\u0000\uE1AA\u0000\u0000" + // 19850 - 19859
                "\uE8CC\u0000\u0000\u0000\uDEB7\u0000\u0000\uCBC9\u0000\u0000" + // 19860 - 19869
                "\uE6D1\uF0CC\u0000\u0000\uDEDE\uF2AF\uF8A9\u0000\u0000\u0000" + // 19870 - 19879
                "\u0000\uC4C4\uC4C5\u0000\uC4C6\uDDFA\u0000\u0000\u0000\u0000" + // 19880 - 19889
                "\u0000\u0000\uF0D5\uE2B3\uDEE7\u0000\u0000\u0000\u0000\u0000" + // 19890 - 19899
                "\u0000\uD4B4\u0000\u0000\uCEEA\u0000\u0000\u0000\u0000\u0000" + // 19900 - 19909
                "\u0000\uF2F4\u0000\u0000\uF2D1\u0000\u0000\u0000\u0000\uE9C1" + // 19910 - 19919
                "\u0000\uD0E6\u0000\u0000\uDEC1\u0000\u0000\uE4AC\u0000\uA8AA" + // 19920 - 19929
                "\u0000\u0000\u0000\u0000\u0000\uA8AD\uA9AC\uF8F9\u0000\u0000" + // 19930 - 19939
                "\u0000\u0000\uF0AE\u0000\u0000\uDAAA\u0000\u0000\u0000\u0000" + // 19940 - 19949
                "\u0000\u0000\uF8CC\u0000\uEAD9\uDEE6\u0000\u0000\u0000\u0000" + // 19950 - 19959
                "\uDFD4\u0000\u0000\uD9C0\uD6EF\u0000\u0000\u0000\u0000\uD9CC" + // 19960 - 19969
                "\uE9DE\u0000\u0000\u0000\u0000\u0000\uF0D3\uF2B4\uD1B7\uF2B3" + // 19970 - 19979
                "\u0000\u0000\u0000\u0000\u0000\u0000\uF5B1\u0000\u0000\uE7A7" + // 19980 - 19989
                "\u0000\uE6D7\u0000\u0000\u0000\u0000\uC5AD\uC5AE\u0000\u0000" + // 19990 - 19999
                "\uE1D7\u0000\uE6CF\u0000\uF4F1\u0000\u0000\uDAC5\u0000\uD8EC" + // 20000 - 20009
                "\u0000\u0000\u0000\u0000\uC4BC\uC4BD\u0000\u0000\uF8F6\u0000" + // 20010 - 20019
                "\u0000\u0000\u0000\uF5D2\uEDE9\uDEE5\uD1B5\u0000\u0000\u0000" + // 20020 - 20029
                "\u0000\u0000\uD1B6\uD8A8\u0000\u0000\u0000\uCCE4\u0000\u0000" + // 20030 - 20039
                "\uD1B4\uDAF1\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE9C0" + // 20040 - 20049
                "\uCAF9\u0000\u0000\uD4DA\u0000\u0000\u0000\u0000\uC6BA\uC6BB" + // 20050 - 20059
                "\u0000\u0000\uFDF7\u0000\u0000\u0000\u0000\u0000\uE2D9\uF3B7" + // 20060 - 20069
                "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uDFB6\uD9C2\u0000" + // 20070 - 20079
                "\uF0D2\u0000\uE4D1\u0000\u0000\u0000\uBECE\uBECF\uBED0\u0000" + // 20080 - 20089
                "\uBED1\uD8F9\u0000\u0000\u0000\u0000\u0000\u0000\u0000\uE4F3" + // 20090 - 20099
                "\uF7AF\uDAB6\u0000\uCAD7\u0000\u0000\u0000\u0000\uC6A1\u0000" + // 20100 - 20109
                "\u0000\u0000\uBDD7\uBDD8\uBDD9\u0000\u0000\uDBE1\u0000\u0000" + // 20110 - 20119
                "\uE5C9\u0000\uEDB4\u0000\uEAA7\uE9F6\uFBBB\u0000\uE7E9\uEFCC" + // 20120 - 20129
                "\u0000\u0000\uE3F7\u0000\u0000\u0000\u0000\u0000\uECA8\uDFD3" + // 20130 - 20139
                "\u0000\u0000\u0000\uDAF0\u0000\uE2EA\u0000\uE5F8\u0000\u0000" + // 20140 - 20149
                "\uDEC0\uECA3\u0000\uE9CD\u0000\uA8A2\u0000\u0000\u0000\u0000" + // 20150 - 20159
                "\u0000\u0000\uA1BF\uE7C3\u0000\uECCC\u0000\u0000\u0000\u0000" + // 20160 - 20169
                "\u0000\uB2BF\uB2C0\u0000\u0000\uE6D5\u0000\u0000\uE9F2\u0000" + // 20170 - 20179
                "\uDFB0\u0000\uE3AB\uEBE0\u0000\u0000\u0000\uCEFA\uCBF7\uE5A5" + // 20180 - 20189
                "\uCDC1\u0000\u0000\uFBD3\u0000\u0000\u0000\u0000\uC5FE\u0000" + // 20190 - 20199
                "\u0000\u0000\uB8EB\uB8EC\uB8ED\u0000\uB8EE\uCDA4\u0000\u0000" + // 20200 - 20209
                "\uD4F4\uDBA1\uDBDC\uDBDD\u0000\uFCAF\uD3A1\u0000\uF1AB\u0000" + // 20210 - 20219
                "\u0000\u0000\u0000\uBEC6\uBEC7\u0000\u0000\uF4B8\uF7BC\uDCFD" + // 20220 - 20229
                "\u0000\uE8EC\uE4E7\u0000\uA2AC\uA9F6\uA8AC\u0000\uA8F9\uA8F6" + // 20230 - 20239
                "\uA8FA\uA2AF\uE8B9\u0000\uEFA6\u0000\u0000\u0000\u0000\u0000" + // 20240 - 20249
                "\uB2B8\u0000\u0000\u0000\uB6D5\uB6D6\u0000\u0000\u0000\uB8DD" + // 20250 - 20259
                "\uB8DE\uB8DF\u0000\u0000\uD6A1\uFDBF\u0000\uFCD3\u0000\uEFA1" + // 20260 - 20269
                "\u0000\uCEF8\uDCB0\u0000\u0000\u0000\u0000\uE3AA\u0000\uA1C6" + // 20270 - 20279
                "\uA1BE\uA9F7\uA9F8\uA2A5\u0000\uA2D2\u0000\u0028\u0029\u002A" + // 20280 - 20289
                "\u002B\u002C\u002D\u002E\u002F\uF4B0\uF3EA\uDAEE\u0000\uD7BB" + // 20290 - 20299
                "\u0000\uE2B1\u0000\uEDDB\uDFB2\uDFBE\uF9BB\u0000\uDCF4\u0000" + // 20300 - 20309
                "\u0000\uF9C9\u0000\uE4E2\u0000\uFBBD\u0000\u0000\uECEC\uFBBE" + // 20310 - 20319
                "\uDFEB\u0000\uE1F8\u0000\u0000\uE2D4\uD2FD\u0000\uE5A8\u0000" + // 20320 - 20329
                "\u0000\u0000\uE0FC\uD4A8\u0000\uEDD3\uD8EF\uE5DC\u0000\u0000" + // 20330 - 20339
                "\u0000\u0000\u0000\u0000\u0000\uE6A2\uD3C3\u0000\uD8A6\u0000" + // 20340 - 20349
                "\uF6C1\u0000\u0000\u0000\uB3D6\uB3D7\uB3D8\u0000\u0000\uE5EA" + // 20350 - 20359
                "\u0000\u0000\uF1E0\u0000\u0000\u0000\uEFB5\u0000\u0000\u0000" + // 20360 - 20369
                "\uFAE9\uD4E3\uCCE2\u0000\uF7D4\u0000\u0000\u0000\u0000\uC5F5" + // 20370 - 20379
                "\uC5F6\u0000\u0000\uF8B2\u0000\u0000\u0000\uDCEB\u0000\u0000" + // 20380 - 20389
                "\uEFBA\uF1DD\u0000\uDEB3\u0000\u0000\u0000\uCECE\u0000\u0000" + // 20390 - 20399
                "\u0000\u0000\uDECB\uF6B8\u0000\u0000\u0000\uCAC8\uF9EE\uDBEC" + // 20400 - 20409
                "\u0000\u0000\uFDC3\u0000\u0000\u0000\uEBF6\uCFB2\u0000\uECE9" + // 20410 - 20419
                "\uEFCB\u0000\uF6D2\u0000\u0000\u0000\uD8B2\uD3B8\uF2D6\u0000" + // 20420 - 20429
                "\u0000\uD4D9\uEEC5\uF2F0\u0000\uA2E4\u0000\u0000\uA7E4\uA7EE" + // 20430 - 20439
                "\uA7E9\u0000\u0000\uE9BA\u0000\u0000\u0000\u0000\u0000\u0000" + // 20440 - 20449
                "\uCCDD\u0000\u0000\uDBFA\u0000\u0000\u0000\uE8B5\u0000\uD3A6" + // 20450 - 20459
                "\uD1B1\u0000\uCBB1\u0000\u0000\u0000\u0000\uD1B2\uECB6\u0000" + // 20460 - 20469
                "\u0000\u0000\u0000\uFBFE\uD3D7\u0000\uA7EA\u0000\u0000\uA7EB" + // 20470 - 20479
                "\u0000\u0000\uA7DF\u0000\uA1A7\u0000\uA8A3\u0000\u0000\u0000" + // 20480 - 20489
                "\u0000\u0000\uE2F1\u0000\u0000\u0000\uD2EE\u0000\u0000\uE1E6" + // 20490 - 20499
                "\uF7F9\uEFA4\u0000\uEFEB\u0000\u0000\u0000\u0000\u0000\uB2A7" + // 20500 - 20509
                "\u0000\u0000\u0000\uB1BF\uB1C0\uB1C1\u0000\uB1C2\uEFA3\uEBA6" + // 20510 - 20519
                "\uCBA3\uE3E9\u0000\u0000\u0000\uD1FB\uE9C4\u0000\u0000\uDCCB" + // 20520 - 20529
                "\uE9C5\u0000\u0000\u0000\uB0C0\u0000\uB0C1\u0000\u0000\uDEE1" + // 20530 - 20539
                "\u0000\uE4A3\u0000\u0000\u0000\uD7B7\uCAF6\u0000\uE4A4\uF4D6" + // 20540 - 20549
                "\u0000\u0000\u0000\uDFE6\uFBD2\u0000\uF8F8\uF7FB\u0000\u0000" + // 20550 - 20559
                "\uE8BF\u0000\uA7BC\uA7ED\uA7B5\u0000\u0000\u0000\u0000\uA7B9" + // 20560 - 20569
                "\uCEB6\u0000\uF3C0\u0000\uCDFE\u0000\u0000\u0000\uCDD5\u0000" + // 20570 - 20579
                "\u0000\u0000\u0000\uDCB3\u0000\u0000\u0000\u0000\uC1D6\uC1D7" + // 20580 - 20589
                "\u0000\u0000\uFBCD\u0000\uD5BD\uF1DF\u0000\u0000\uF6FB\uFCBB" + // 20590 - 20599
                "\u0000\uE2B0\u0000\u0000\uE6A5\u0000\u0000\uD3C2\u0000\u0000" + // 20600 - 20609
                "\u0000\u0000\uD3B6\u0000\uA7DA\uA7DB\uA2E3\uA7EC\uA7A6\uA7E0" + // 20610 - 20619
                "\uA7EF\uA2E1\uF3BF\u0000\uF0D1\u0000\u0000\u0000\u0000\u0000" + // 20620 - 20629
                "\uB2A5\uB2A6\u0000\u0000\uEEFE\u0000\uE7FE\u0000\u0000\u0000" + // 20630 - 20639
                "\u0000\uC4F5\u0000\u0000\u0000\uF9DB\u0000\u0000\u0000\u0000" + // 20640 - 20649
                "\uCCD1\u0000\uDFEA\u0000\u0000\uF4BB\uDAD5\u0000\uF9B2\u0000" + // 20650 - 20659
                "\u0000\u0000\uEFB2\u0000\u0000\u0000\u0000\uF9E1\u0000\u0000" + // 20660 - 20669
                "\u0000\u0000\uC0A7\uC0A8\u0000\u0000\uEFE0\u0000\u0000\u0000" + // 20670 - 20679
                "\uE5E5\uD0D5\u0000\uA7C7\uA7C8\uA7CE\uA7CF\uA7D0\uA7D1\uA7D2" + // 20680 - 20689
                "\uA7D3\uD7F3\u0000\u0000\u0000\uFCD4\u0000\uDAD7\uCCDF\uF2D3" + // 20690 - 20699
                "\uFBA9\uD8A5\u0000\u0000\u0000\u0000\uD5CB"
                ;


}
