/*
 * @(#)Box.java	1.3 98/09/21
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 */

package visualBenchmarks;

import java.awt.*;

/**
 * A boxed panel with an etched border and a label. Useful to group 
 * components on a GUI. The component is centered in the box with a fixed 
 * inset of 10 pixels.
 */

public class Box extends Panel {

	private EtchedRectangle box = new EtchedRectangle(this);
	private Label titleLabel;

	public Box(Component surrounded, String title) {
		this(surrounded, new Label(title, Label.CENTER));
	}

	public Box(Component surrounded, Label label) {
		titleLabel = label;

		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();

		setLayout(gbl);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.anchor = GridBagConstraints.NORTH;
		gbl.setConstraints(titleLabel,gbc);

		add(titleLabel);
		
		//The top inset is set to 0, since the component is 
		//anchored to the south . 

		gbc.insets = new Insets(0,10,10,10);
		gbc.anchor = GridBagConstraints.SOUTH;
		gbc.weighty = 1.0;
		gbc.weightx = 1.0;
		gbc.fill = GridBagConstraints.BOTH;
		gbl.setConstraints(surrounded, gbc);

		add(surrounded);
	}


	/**
	 * Draw the box with an etched <em>in</em> border
	 */

	public void etchedIn() {
		box.etchedIn();
	}

	
	/**
	 * Draw the box with an etched <em>out</em> border
	 */

	public void etchedOut() {
		box.etchedOut();
	}

	
	public void paint(Graphics g) {
		box.paint();
	}

	
	/**
	 * Set the width and height for the box
	 */

	public void setsize(int w, int h) {
		setBounds(getLocation().x, getLocation().y, w, h);
	}


	/**
	 * Set the width and the height, and the location of the box
	 */

	public void setBounds(int x, int y, int w, int h) {
		super.setBounds(x,y,w,h);
		
		FontMetrics fm = titleLabel.getFontMetrics(
			titleLabel.getFont());
		int top = getInsets().top + fm.getAscent();
		Dimension size = getSize();
		
		//Center the box around the top label.
		box.setBounds(0, top, size.width -1, size.height - top - 1);
	}

	public String paramString() {
		return super.paramString() + ",etching=" +
			(box.isEtchedIn()?EtchedRectangle.ETCHED_IN:
		 	EtchedRectangle.ETCHED_OUT) + ",title=" + titleLabel;
	}
}
		
